(function(e){var n={};function t(r){if(n[r])return n[r].exports;var a=n[r]={i:r,l:!1,exports:{}};return e[r].call(a.exports,a,a.exports,t),a.l=!0,a.exports}t.m=e,t.c=n,t.d=function(e,n,r){t.o(e,n)||Object.defineProperty(e,n,{enumerable:!0,get:r})},t.r=function(e){"undefined"!==typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(e,"__esModule",{value:!0})},t.t=function(e,n){if(1&n&&(e=t(e)),8&n)return e;if(4&n&&"object"===typeof e&&e&&e.__esModule)return e;var r=Object.create(null);if(t.r(r),Object.defineProperty(r,"default",{enumerable:!0,value:e}),2&n&&"string"!=typeof e)for(var a in e)t.d(r,a,function(n){return e[n]}.bind(null,a));return r},t.n=function(e){var n=e&&e.__esModule?function(){return e["default"]}:function(){return e};return t.d(n,"a",n),n},t.o=function(e,n){return Object.prototype.hasOwnProperty.call(e,n)},t.p="/ui/static/",t(t.s="d0c7")})({"30db":function(e,n,t){"use strict";(function(e,r){t.d(n,"c",function(){return g}),t.d(n,"b",function(){return v}),t.d(n,"a",function(){return h});var a="en",i=!1,o=!1,s=!1,c=!1,l=!1,u=void 0,d="undefined"!==typeof e&&"undefined"!==typeof e.versions&&"undefined"!==typeof e.versions.electron&&"renderer"===e.type;if("object"!==typeof navigator||d){if("object"===typeof e){i="win32"===e.platform,o="darwin"===e.platform,s="linux"===e.platform,u=a,a;var p=Object({NODE_ENV:"production",BASE_URL:"/ui/static/"})["VSCODE_NLS_CONFIG"];if(p)try{var m=JSON.parse(p),f=m.availableLanguages["*"];u=m.locale,f||a,m._translationsConfigFile}catch(b){}c=!0}}else{var _=navigator.userAgent;i=_.indexOf("Windows")>=0,o=_.indexOf("Macintosh")>=0,s=_.indexOf("Linux")>=0,l=!0,u=navigator.language,u}var g=i,v=l,y="object"===typeof self?self:"object"===typeof r?r:{},h=y}).call(this,t("f28c"),t("c8ba"))},5110:function(e,n,t){(function(e){
/*!
Copyright (c) 2014 Taylor Hakes
Copyright (c) 2014 Forbes Lindesay
 */
(function(e,n){n()})(0,function(){"use strict";function n(e){var n=this.constructor;return this.then(function(t){return n.resolve(e()).then(function(){return t})},function(t){return n.resolve(e()).then(function(){return n.reject(t)})})}var t=setTimeout;function r(){}function a(e,n){return function(){e.apply(n,arguments)}}function i(e){if(!(this instanceof i))throw new TypeError("Promises must be constructed via new");if("function"!==typeof e)throw new TypeError("not a function");this._state=0,this._handled=!1,this._value=void 0,this._deferreds=[],d(e,this)}function o(e,n){while(3===e._state)e=e._value;0!==e._state?(e._handled=!0,i._immediateFn(function(){var t=1===e._state?n.onFulfilled:n.onRejected;if(null!==t){var r;try{r=t(e._value)}catch(a){return void c(n.promise,a)}s(n.promise,r)}else(1===e._state?s:c)(n.promise,e._value)})):e._deferreds.push(n)}function s(e,n){try{if(n===e)throw new TypeError("A promise cannot be resolved with itself.");if(n&&("object"===typeof n||"function"===typeof n)){var t=n.then;if(n instanceof i)return e._state=3,e._value=n,void l(e);if("function"===typeof t)return void d(a(t,n),e)}e._state=1,e._value=n,l(e)}catch(r){c(e,r)}}function c(e,n){e._state=2,e._value=n,l(e)}function l(e){2===e._state&&0===e._deferreds.length&&i._immediateFn(function(){e._handled||i._unhandledRejectionFn(e._value)});for(var n=0,t=e._deferreds.length;n<t;n++)o(e,e._deferreds[n]);e._deferreds=null}function u(e,n,t){this.onFulfilled="function"===typeof e?e:null,this.onRejected="function"===typeof n?n:null,this.promise=t}function d(e,n){var t=!1;try{e(function(e){t||(t=!0,s(n,e))},function(e){t||(t=!0,c(n,e))})}catch(r){if(t)return;t=!0,c(n,r)}}i.prototype["catch"]=function(e){return this.then(null,e)},i.prototype.then=function(e,n){var t=new this.constructor(r);return o(this,new u(e,n,t)),t},i.prototype["finally"]=n,i.all=function(e){return new i(function(n,t){if(!e||"undefined"===typeof e.length)throw new TypeError("Promise.all accepts an array");var r=Array.prototype.slice.call(e);if(0===r.length)return n([]);var a=r.length;function i(e,o){try{if(o&&("object"===typeof o||"function"===typeof o)){var s=o.then;if("function"===typeof s)return void s.call(o,function(n){i(e,n)},t)}r[e]=o,0===--a&&n(r)}catch(c){t(c)}}for(var o=0;o<r.length;o++)i(o,r[o])})},i.resolve=function(e){return e&&"object"===typeof e&&e.constructor===i?e:new i(function(n){n(e)})},i.reject=function(e){return new i(function(n,t){t(e)})},i.race=function(e){return new i(function(n,t){for(var r=0,a=e.length;r<a;r++)e[r].then(n,t)})},i._immediateFn="function"===typeof setImmediate&&function(e){setImmediate(e)}||function(e){t(e,0)},i._unhandledRejectionFn=function(e){"undefined"!==typeof console&&console&&console.warn("Possible Unhandled Promise Rejection:",e)};var p=function(){if("undefined"!==typeof self)return self;if("undefined"!==typeof window)return window;if("undefined"!==typeof e)return e;throw new Error("unable to locate global object")}();"Promise"in p?p.Promise.prototype["finally"]||(p.Promise.prototype["finally"]=n):p["Promise"]=i})}).call(this,t("c8ba"))},"973d":function(e,n,t){"use strict";(function(e,r,a,i){t.d(n,"c",function(){return m}),t.d(n,"a",function(){return f}),t.d(n,"b",function(){return _});
/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0  
 
THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, 
MERCHANTABLITY OR NON-INFRINGEMENT. 
 
See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
var o=function(){return o=Object.assign||function(e){for(var n,t=1,r=arguments.length;t<r;t++)for(var a in n=arguments[t],n)Object.prototype.hasOwnProperty.call(n,a)&&(e[a]=n[a]);return e},o.apply(this,arguments)},s=function(e,n){var t,r,a,i,o={label:0,sent:function(){if(1&a[0])throw a[1];return a[1]},trys:[],ops:[]};return i={next:s(0),throw:s(1),return:s(2)},"function"===typeof Symbol&&(i[Symbol.iterator]=function(){return this}),i;function s(e){return function(n){return c([e,n])}}function c(i){if(t)throw new TypeError("Generator is already executing.");while(o)try{if(t=1,r&&(a=2&i[0]?r["return"]:i[0]?r["throw"]||((a=r["return"])&&a.call(r),0):r.next)&&!(a=a.call(r,i[1])).done)return a;switch(r=0,a&&(i=[2&i[0],a.value]),i[0]){case 0:case 1:a=i;break;case 4:return o.label++,{value:i[1],done:!1};case 5:o.label++,r=i[1],i=[0];continue;case 7:i=o.ops.pop(),o.trys.pop();continue;default:if(a=o.trys,!(a=a.length>0&&a[a.length-1])&&(6===i[0]||2===i[0])){o=0;continue}if(3===i[0]&&(!a||i[1]>a[0]&&i[1]<a[3])){o.label=i[1];break}if(6===i[0]&&o.label<a[1]){o.label=a[1],a=i;break}if(a&&o.label<a[2]){o.label=a[2],o.ops.push(i);break}a[2]&&o.ops.pop(),o.trys.pop();continue}i=n.call(e,o)}catch(s){i=[6,s],r=0}finally{t=a=0}if(5&i[0])throw i[1];return{value:i[0]?i[1]:void 0,done:!0}}},c=function(e,n){return Object.defineProperty?Object.defineProperty(e,"raw",{value:n}):e.raw=n,e},l=function(){var e=function(n,t){return e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,n){e.__proto__=n}||function(e,n){for(var t in n)n.hasOwnProperty(t)&&(e[t]=n[t])},e(n,t)};return function(n,t){function r(){this.constructor=n}e(n,t),n.prototype=null===t?Object.create(t):(r.prototype=t.prototype,new r)}}();(function(e){e.versionMajorMinor="3.3",e.version=e.versionMajorMinor+".3333"})(u||(u={})),function(e){(function(e){e[e["LessThan"]=-1]="LessThan",e[e["EqualTo"]=0]="EqualTo",e[e["GreaterThan"]=1]="GreaterThan"})(e.Comparison||(e.Comparison={}))}(u||(u={})),function(e){function n(){var e=Object.create(null);return e.__=void 0,delete e.__,e}function t(){return new e.MapCtr}function r(e){for(var n=t(),r=0,a=e;r<a.length;r++){var i=a[r],o=i[0],s=i[1];n.set(o,s)}return n}function a(n){var t=new e.MapCtr;for(var r in n)Ee.call(n,r)&&t.set(r,n[r]);return t}function i(){var e=function(){function e(e,n){this.index=0,this.data=e,this.selector=n,this.keys=Object.keys(e)}return e.prototype.next=function(){var e=this.index;return e<this.keys.length?(this.index++,{value:this.selector(this.data,this.keys[e]),done:!1}):{value:void 0,done:!0}},e}();return function(){function t(){this.data=n(),this.size=0}return t.prototype.get=function(e){return this.data[e]},t.prototype.set=function(e,n){return this.has(e)||this.size++,this.data[e]=n,this},t.prototype.has=function(e){return e in this.data},t.prototype.delete=function(e){return!!this.has(e)&&(this.size--,delete this.data[e],!0)},t.prototype.clear=function(){this.data=n(),this.size=0},t.prototype.keys=function(){return new e(this.data,function(e,n){return n})},t.prototype.values=function(){return new e(this.data,function(e,n){return e[n]})},t.prototype.entries=function(){return new e(this.data,function(e,n){return[n,e[n]]})},t.prototype.forEach=function(e){for(var n in this.data)e(this.data[n],n)},t}()}function o(e){return e?e.length:0}function s(e,n){if(e)for(var t=0;t<e.length;t++){var r=n(e[t],t);if(r)return r}}function c(e,n){if(void 0!==e)for(var t=0;t<e.length;t++){var r=n(e[t],t);if(void 0!==r)return r}}function l(e,n){while(1){var t=e.next(),r=t.value,a=t.done;if(a)return;var i=n(r);if(void 0!==i)return i}}function u(e,n,t){var r=[];be.assertEqual(e.length,n.length);for(var a=0;a<e.length;a++)r.push(t(e[a],n[a],a));return r}function d(e,n){be.assertEqual(e.length,n.length);var t=0;return{next:function(){return t===e.length?{value:void 0,done:!0}:(t++,{value:[e[t-1],n[t-1]],done:!1})}}}function p(e,n){be.assert(e.length===n.length);for(var r=t(),a=0;a<e.length;++a)r.set(e[a],n[a]);return r}function m(e,n){if(e)for(var t=0;t<e.length;t++)if(!n(e[t],t))return!1;return!0}function f(e,n){for(var t=0;t<e.length;t++){var r=e[t];if(n(r,t))return r}}function _(e,n){for(var t=e.length-1;t>=0;t--){var r=e[t];if(n(r,t))return r}}function g(e,n,t){for(var r=t||0;r<e.length;r++)if(n(e[r],r))return r;return-1}function v(e,n,t){for(var r=void 0===t?e.length-1:t;r>=0;r--)if(n(e[r],r))return r;return-1}function y(e,n){for(var t=0;t<e.length;t++){var r=n(e[t],t);if(r)return r}return be.fail()}function h(e,n,t){if(void 0===t&&(t=nn),e)for(var r=0,a=e;r<a.length;r++){var i=a[r];if(t(i,n))return!0}return!1}function b(e,n,t){return void 0===t&&(t=nn),e.length===n.length&&e.every(function(e,r){return t(e,n[r])})}function E(e,n,t){for(var r=t||0;r<e.length;r++)if(h(n,e.charCodeAt(r)))return r;return-1}function T(e,n){var t=0;if(e)for(var r=0;r<e.length;r++){var a=e[r];n(a,r)&&t++}return t}function S(e,n){if(e){var t=e.length,r=0;while(r<t&&n(e[r]))r++;if(r<t){var a=e.slice(0,r);r++;while(r<t){var i=e[r];n(i)&&a.push(i),r++}return a}}return e}function x(e,n){for(var t=0,r=0;r<e.length;r++)n(e[r],r,e)&&(e[t]=e[r],t++);e.length=t}function C(e){e.length=0}function A(e,n){var t;if(e){t=[];for(var r=0;r<e.length;r++)t.push(n(e[r],r))}return t}function D(e,n){return{next:function(){var t=e.next();return t.done?t:{value:n(t.value),done:!1}}}}function L(e,n){if(e)for(var t=0;t<e.length;t++){var r=e[t],a=n(r,t);if(r!==a){var i=e.slice(0,t);for(i.push(a),t++;t<e.length;t++)i.push(n(e[t],t));return i}}return e}function k(e){var n;if(e){n=[];for(var t=0,r=e;t<r.length;t++){var a=r[t];a&&(Be(a)?ne(n,a):n.push(a))}}return n}function N(n,t){var r;if(n)for(var a=0;a<n.length;a++){var i=t(n[a],a);i&&(r=Be(i)?ne(r,i):$(r,i))}return r||e.emptyArray}function I(e,n){var t=[];if(e)for(var r=0;r<e.length;r++){var a=n(e[r],r);a&&(Be(a)?ne(t,a):t.push(a))}return t}function w(n,t){var r=n.next();if(r.done)return e.emptyIterator;var a=i(r.value);return{next:function(){while(1){var e=a.next();if(!e.done)return e;var t=n.next();if(t.done)return t;a=i(t.value)}}};function i(n){var r=t(n);return void 0===r?e.emptyIterator:Be(r)?oe(r):r}}function M(e,n){var t;if(e)for(var r=0;r<e.length;r++){var a=e[r],i=n(a,r);(t||a!==i||Be(i))&&(t||(t=e.slice(0,r)),Be(i)?ne(t,i):t.push(i))}return t||e}function O(e,n){for(var t=[],r=0;r<e.length;r++){var a=n(e[r],r);if(void 0===a)return;t.push(a)}return t}function P(e,n){var t=[];if(e)for(var r=0;r<e.length;r++){var a=n(e[r],r);void 0!==a&&t.push(a)}return t}function R(e,n){return{next:function(){while(1){var t=e.next();if(t.done)return t;var r=n(t.value);if(void 0!==r)return{value:r,done:!1}}}}}function F(e){var n=!1;return{next:function(){var t=n;return n=!0,t?{value:void 0,done:!0}:{value:e,done:!1}}}}function G(e,n,t){var r;if(e){r=[];var a=e.length,i=void 0,o=void 0,s=0,c=0;while(s<a){while(c<a){var l=e[c];if(o=n(l,c),0===c)i=o;else if(o!==i)break;c++}if(s<c){var u=t(e.slice(s,c),i,s,c);u&&r.push(u),s=c}i=o,c++}}return r}function V(e,n){if(e){var r=t();return e.forEach(function(e,t){var a=n(t,e),i=a[0],o=a[1];r.set(i,o)}),r}}function B(e,n){if(e){if(!n)return e.length>0;for(var t=0,r=e;t<r.length;t++){var a=r[t];if(n(a))return!0}}return!1}function K(e,n,t){for(var r,a=0;a<e.length;a++)n(e[a])?r=void 0===r?a:r:void 0!==r&&(t(r,a),r=void 0);void 0!==r&&t(r,e.length)}function j(e,n){return B(n)?B(e)?e.concat(n):n:e}function U(e,n,t){var r=e.map(function(e,n){return n});ae(e,r,t);for(var a=e[r[0]],i=[r[0]],o=1;o<r.length;o++){var s=r[o],c=e[s];n(a,c)||(i.push(s),a=c)}return i.sort(),i.map(function(n){return e[n]})}function H(e,n){for(var t=[],r=0,a=e;r<a.length;r++){var i=a[r];te(t,i,n)}return t}function W(e,n,t){return 0===e.length?[]:1===e.length?e.slice():t?U(e,n,t):H(e,n)}function q(n,t){if(0===n.length)return e.emptyArray;for(var r=n[0],a=[r],i=1;i<n.length;i++){var o=n[i];switch(t(o,r)){case!0:case 0:continue;case-1:return be.fail("Array is unsorted.")}a.push(r=o)}return a}function z(e,n,t){if(0!==e.length){var r=ve(e,n,Xe,t);r<0&&e.splice(~r,0,n)}else e.push(n)}function J(e,n,t){return q(ie(e,n),t||n||ln)}function X(e,n,t){if(void 0===t&&(t=nn),!e||!n)return e===n;if(e.length!==n.length)return!1;for(var r=0;r<e.length;r++)if(!t(e[r],n[r],r))return!1;return!0}function Y(e){var n;if(e)for(var t=0;t<e.length;t++){var r=e[t];!n&&r||(n||(n=e.slice(0,t)),r&&n.push(r))}return n||e}function Q(e,n,t){if(!n||!e||0===n.length||0===e.length)return n;var r=[];e:for(var a=0,i=0;i<n.length;i++){i>0&&be.assertGreaterThanOrEqual(t(n[i],n[i-1]),0);n:for(var o=a;a<e.length;a++)switch(a>o&&be.assertGreaterThanOrEqual(t(e[a],e[a-1]),0),t(n[i],e[a])){case-1:r.push(n[i]);continue e;case 0:continue e;case 1:continue n}}return r}function Z(e,n){for(var t=0,r=0,a=e;r<a.length;r++){var i=a[r];t+=i[n]}return t}function $(e,n){return void 0===n?e:void 0===e?[n]:(e.push(n),e)}function ee(e,n){return n<0?e.length+n:n}function ne(e,n,t,r){if(void 0===n||0===n.length)return e;if(void 0===e)return n.slice(t,r);t=void 0===t?0:ee(n,t),r=void 0===r?n.length:ee(n,r);for(var a=t;a<r&&a<n.length;a++)void 0!==n[a]&&e.push(n[a]);return e}function te(e,n,t){return!h(e,n,t)&&(e.push(n),!0)}function re(e,n,t){return e?(te(e,n,t),e):[n]}function ae(e,n,t){n.sort(function(n,r){return t(e[n],e[r])||on(n,r)})}function ie(e,n){return 0===e.length?e:e.slice().sort(n)}function oe(e){var n=0;return{next:function(){return n===e.length?{value:void 0,done:!0}:(n++,{value:e[n-1],done:!1})}}}function se(e,n){var t=e.map(function(e,n){return n});return ae(e,t,n),t.map(function(n){return e[n]})}function ce(e,n,t,r){while(t<r){if(e[t]!==n[t])return!1;t++}return!0}function le(e,n){if(e&&(n=ee(e,n),n<e.length))return e[n]}function ue(e){return 0===e.length?void 0:e[0]}function de(e){return be.assert(0!==e.length),e[0]}function pe(e){return 0===e.length?void 0:e[e.length-1]}function me(e){return be.assert(0!==e.length),e[e.length-1]}function fe(e){return e&&1===e.length?e[0]:void 0}function _e(e){return e&&1===e.length?e[0]:e}function ge(e,n,t){var r=e.slice(0);return r[n]=t,r}function ve(e,n,t,r,a){return ye(e,t(n),t,r,a)}function ye(e,n,t,r,a){if(!B(e))return-1;var i=a||0,o=e.length-1;while(i<=o){var s=i+(o-i>>1),c=t(e[s]);switch(r(c,n)){case-1:i=s+1;break;case 0:return s;case 1:o=s-1;break}}return~i}function he(e,n,t,r,a){if(e&&e.length>0){var i=e.length;if(i>0){var o=void 0===r||r<0?0:r,s=void 0===a||o+a>i-1?i-1:o+a,c=void 0;arguments.length<=2?(c=e[o],o++):c=t;while(o<=s)c=n(c,e[o],o),o++;return c}}return t}e.emptyArray=[],e.createMap=t,e.createMapFromEntries=r,e.createMapFromTemplate=a,e.MapCtr="undefined"!==typeof Map&&"entries"in Map.prototype?Map:i(),e.length=o,e.forEach=s,e.firstDefined=c,e.firstDefinedIterator=l,e.zipWith=u,e.zipToIterator=d,e.zipToMap=p,e.every=m,e.find=f,e.findLast=_,e.findIndex=g,e.findLastIndex=v,e.findMap=y,e.contains=h,e.arraysEqual=b,e.indexOfAnyCharCode=E,e.countWhere=T,e.filter=S,e.filterMutate=x,e.clear=C,e.map=A,e.mapIterator=D,e.sameMap=L,e.flatten=k,e.flatMap=N,e.flatMapToMutable=I,e.flatMapIterator=w,e.sameFlatMap=M,e.mapAllOrFail=O,e.mapDefined=P,e.mapDefinedIterator=R,e.emptyIterator={next:function(){return{value:void 0,done:!0}}},e.singleIterator=F,e.spanMap=G,e.mapEntries=V,e.some=B,e.getRangesWhere=K,e.concatenate=j,e.deduplicate=W,e.insertSorted=z,e.sortAndDeduplicate=J,e.arrayIsEqualTo=X,e.compact=Y,e.relativeComplement=Q,e.sum=Z,e.append=$,e.addRange=ne,e.pushIfUnique=te,e.appendIfUnique=re,e.sort=ie,e.arrayIterator=oe,e.stableSort=se,e.rangeEquals=ce,e.elementAt=le,e.firstOrUndefined=ue,e.first=de,e.lastOrUndefined=pe,e.last=me,e.singleOrUndefined=fe,e.singleOrMany=_e,e.replaceElement=ge,e.binarySearch=ve,e.binarySearchKey=ye,e.reduceLeft=he;var be,Ee=Object.prototype.hasOwnProperty;function Te(e,n){return Ee.call(e,n)}function Se(e,n){return Ee.call(e,n)?e[n]:void 0}function xe(e){var n=[];for(var t in e)Ee.call(e,t)&&n.push(t);return n}function Ce(e){var n=[];for(var t in e)Ee.call(e,t)&&n.push(e[t]);return n}function Ae(e,n){for(var t,r=[],a=e.next(),i=a.value,o=a.done;!o;t=e.next(),i=t.value,o=t.done,t)r.push(n?n(i):i);return r}function De(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];for(var r=0,a=n;r<a.length;r++){var i=a[r];for(var o in i)Te(i,o)&&(e[o]=i[o])}return e}function Le(e,n,t){if(void 0===t&&(t=nn),e===n)return!0;if(!e||!n)return!1;for(var r in e)if(Ee.call(e,r)){if(!Ee.call(n,r))return!1;if(!t(e[r],n[r]))return!1}for(var r in n)if(Ee.call(n,r)&&!Ee.call(e,r))return!1;return!0}function ke(e,n,r){void 0===r&&(r=Xe);for(var a=t(),i=0,o=e;i<o.length;i++){var s=o[i],c=n(s);void 0!==c&&a.set(c,r(s))}return a}function Ne(e,n,t){void 0===t&&(t=Xe);for(var r=[],a=0,i=e;a<i.length;a++){var o=i[a];r[n(o)]=t(o)}return r}function Ie(e,n,t){void 0===t&&(t=Xe);for(var r=Fe(),a=0,i=e;a<i.length;a++){var o=i[a];r.add(n(o),t(o))}return r}function we(e,n){return Ae(Ie(e,n).values())}function Me(e){var n={};for(var t in e)Ee.call(e,t)&&(n[t]=e[t]);return n}function Oe(e,n){var t={};for(var r in n)Ee.call(n,r)&&(t[r]=n[r]);for(var r in e)Ee.call(e,r)&&(t[r]=e[r]);return t}function Pe(e,n){for(var t in n)Ee.call(n,t)&&(e[t]=n[t])}function Re(e,n){return n?n.bind(e):void 0}function Fe(){var e=t();return e.add=Ge,e.remove=Ve,e}function Ge(e,n){var t=this.get(e);return t?t.push(n):this.set(e,t=[n]),t}function Ve(e,n){var t=this.get(e);t&&(In(t,n),t.length||this.delete(e))}function Be(e){return Array.isArray?Array.isArray(e):e instanceof Array}function Ke(e){return Be(e)?e:[e]}function je(e){return"string"===typeof e}function Ue(e){return"number"===typeof e}function He(e,n){return void 0!==e&&n(e)?e:void 0}function We(e,n){return void 0!==e&&n(e)?e:be.fail("Invalid cast. The supplied value "+e+" did not pass the test '"+be.getFunctionName(n)+"'.")}function qe(e){}function ze(){return!1}function Je(){return!0}function Xe(e){return e}function Ye(e){return e.toLowerCase()}function Qe(){throw new Error("Not implemented")}function Ze(e){var n;return function(){return e&&(n=e(),e=void 0),n}}function $e(e,n,t,r,a){if(a){for(var i=[],o=0;o<arguments.length;o++)i[o]=arguments[o];return function(e){return en.apply(void 0,A(i,function(n){return n(e)}))}}return r?function(a){return en(e(a),n(a),t(a),r(a))}:t?function(r){return en(e(r),n(r),t(r))}:n?function(t){return en(e(t),n(t))}:e?function(n){return en(e(n))}:function(e){return function(e){return e}}}function en(e,n,t,r,a){if(a){for(var i=[],o=0;o<arguments.length;o++)i[o]=arguments[o];return function(e){return he(i,function(e,n){return n(e)},e)}}return r?function(a){return r(t(n(e(a))))}:t?function(r){return t(n(e(r)))}:n?function(t){return n(e(t))}:e?function(n){return e(n)}:function(e){return e}}function nn(e,n){return e===n}function tn(e,n){return e===n||void 0!==e&&void 0!==n&&e.toUpperCase()===n.toUpperCase()}function rn(e,n){return nn(e,n)}function an(e,n){return e===n?0:void 0===e?-1:void 0===n?1:e<n?-1:1}function on(e,n){return an(e,n)}function sn(e,n,t){return-1===t(e,n)?e:n}function cn(e,n){return e===n?0:void 0===e?-1:void 0===n?1:(e=e.toUpperCase(),n=n.toUpperCase(),e<n?-1:e>n?1:0)}function ln(e,n){return an(e,n)}function un(e){return e?cn:ln}e.hasProperty=Te,e.getProperty=Se,e.getOwnKeys=xe,e.getOwnValues=Ce,e.arrayFrom=Ae,e.assign=De,e.equalOwnProperties=Le,e.arrayToMap=ke,e.arrayToNumericMap=Ne,e.arrayToMultiMap=Ie,e.group=we,e.clone=Me,e.extend=Oe,e.copyProperties=Pe,e.maybeBind=Re,e.createMultiMap=Fe,e.isArray=Be,e.toArray=Ke,e.isString=je,e.isNumber=Ue,e.tryCast=He,e.cast=We,e.noop=qe,e.returnFalse=ze,e.returnTrue=Je,e.identity=Xe,e.toLowerCase=Ye,e.notImplemented=Qe,e.memoize=Ze,e.chain=$e,e.compose=en,function(e){e[e["None"]=0]="None",e[e["Normal"]=1]="Normal",e[e["Aggressive"]=2]="Aggressive",e[e["VeryAggressive"]=3]="VeryAggressive"}(e.AssertionLevel||(e.AssertionLevel={})),function(e){function n(n){return e.currentAssertionLevel>=n}function t(e,n,r,a){e||(r&&(n+="\r\nVerbose Debug Information: "+("string"===typeof r?r:r())),s(n?"False expression: "+n:"False expression.",a||t))}function r(e,n,t,r){if(e!==n){var a=t?r?t+" "+r:t:"";s("Expected "+e+" === "+n+". "+a)}}function a(e,n,t){e>=n&&s("Expected "+e+" < "+n+". "+(t||""))}function i(e,n){e>n&&s("Expected "+e+" <= "+n)}function o(e,n){e<n&&s("Expected "+e+" >= "+n)}function s(e,n){var t=new Error(e?"Debug Failure. "+e:"Debug Failure.");throw Error.captureStackTrace&&Error.captureStackTrace(t,n||s),t}function c(e,n){return void 0===e||null===e?s(n):e}function l(e,n){for(var t=0,r=e;t<r.length;t++){var a=r[t];c(a,n)}return e}function u(n,t,r){void 0===t&&(t="Illegal value:");var a="kind"in n&&"pos"in n?"SyntaxKind: "+e.showSyntaxKind(n):JSON.stringify(n);return s(t+" "+a,r||u)}function d(e){if("function"!==typeof e)return"";if(e.hasOwnProperty("name"))return e.name;var n=Function.prototype.toString.call(e),t=/^function\s+([\w\$]+)\s*\(/.exec(n);return t?t[1]:""}e.currentAssertionLevel=0,e.isDebugging=!1,e.shouldAssert=n,e.assert=t,e.assertEqual=r,e.assertLessThan=a,e.assertLessThanOrEqual=i,e.assertGreaterThanOrEqual=o,e.fail=s,e.assertDefined=c,e.assertEachDefined=l,e.assertNever=u,e.getFunctionName=d}(be=e.Debug||(e.Debug={})),e.equateValues=nn,e.equateStringsCaseInsensitive=tn,e.equateStringsCaseSensitive=rn,e.compareValues=on,e.min=sn,e.compareStringsCaseInsensitive=cn,e.compareStringsCaseSensitive=ln,e.getStringComparer=un;var dn,pn,mn=function(){var e,n,t=s();return c;function r(e,n,t){if(e===n)return 0;if(void 0===e)return-1;if(void 0===n)return 1;var r=t(e,n);return r<0?-1:r>0?1:0}function a(e){var n=new Intl.Collator(e,{usage:"sort",sensitivity:"variant"}).compare;return function(e,t){return r(e,t,n)}}function i(e){return void 0!==e?o():function(e,t){return r(e,t,n)};function n(e,n){return e.localeCompare(n)}}function o(){return function(n,t){return r(n,t,e)};function e(e,t){return n(e.toUpperCase(),t.toUpperCase())||n(e,t)}function n(e,n){return e<n?-1:e>n?1:0}}function s(){return"object"===typeof Intl&&"function"===typeof Intl.Collator?a:"function"===typeof String.prototype.localeCompare&&"function"===typeof String.prototype.toLocaleUpperCase&&"a".localeCompare("B")<0?i:o}function c(r){return void 0===r?e||(e=t(r)):"en-US"===r?n||(n=t(r)):t(r)}}();function fn(){return pn}function _n(e){pn!==e&&(pn=e,dn=void 0)}function gn(e,n){var t=dn||(dn=mn(pn));return t(e,n)}function vn(e,n,t,r){return e===n?0:void 0===e?-1:void 0===n?1:r(e[t],n[t])}function yn(e,n){return on(e?1:0,n?1:0)}function hn(e,n,t){for(var r,a=Math.min(2,Math.floor(.34*e.length)),i=Math.floor(.4*e.length)+1,o=!1,s=e.toLowerCase(),c=0,l=n;c<l.length;c++){var u=l[c],d=t(u);if(void 0!==d&&Math.abs(d.length-s.length)<=a){var p=d.toLowerCase();if(p===s){if(d===e)continue;return u}if(o)continue;if(d.length<3)continue;var m=bn(s,p,i-1);if(void 0===m)continue;m<3?(o=!0,r=u):(be.assert(m<i),i=m,r=u)}}return r}function bn(e,n,t){for(var r=new Array(n.length+1),a=new Array(n.length+1),i=t+1,o=0;o<=n.length;o++)r[o]=o;for(o=1;o<=e.length;o++){var s=e.charCodeAt(o-1),c=o>t?o-t:1,l=n.length>t+o?t+o:n.length;a[0]=o;for(var u=o,d=1;d<c;d++)a[d]=i;for(d=c;d<=l;d++){var p=s===n.charCodeAt(d-1)?r[d-1]:Math.min(r[d]+1,a[d-1]+1,r[d-1]+2);a[d]=p,u=Math.min(u,p)}for(d=l+1;d<=n.length;d++)a[d]=i;if(u>t)return;var m=r;r=a,a=m}var f=r[n.length];return f>t?void 0:f}function En(e,n){var t=e.length-n.length;return t>=0&&e.indexOf(n,t)===t}function Tn(e,n){return En(e,n)?e.slice(0,e.length-n.length):e}function Sn(e,n){return En(e,n)?e.slice(0,e.length-n.length):void 0}function xn(e,n){return-1!==e.indexOf(n)}function Cn(e,n){return e.length>n.length&&En(e,n)}function An(e,n){for(var t=0,r=n;t<r.length;t++){var a=r[t];if(Cn(e,a))return!0}return!1}function Dn(e){var n=/[.-]((min)|(\d+(\.\d+)*))$/;return e.replace(n,"").replace(n,"")}function Ln(e,n){for(var t=0;t<e.length;t++)if(e[t]===n)return kn(e,t),!0;return!1}function kn(e,n){for(var t=n;t<e.length-1;t++)e[t]=e[t+1];e.pop()}function Nn(e,n){e[n]=e[e.length-1],e.pop()}function In(e,n){return wn(e,function(e){return e===n})}function wn(e,n){for(var t=0;t<e.length;t++)if(n(e[t]))return Nn(e,t),!0;return!1}function Mn(e){return e?Xe:Ye}function On(e){var n=e.prefix,t=e.suffix;return n+"*"+t}function Pn(e,n){return be.assert(Bn(e,n)),n.substring(e.prefix.length,n.length-e.suffix.length)}function Rn(e,n,t){for(var r,a=-1,i=0,o=e;i<o.length;i++){var s=o[i],c=n(s);Bn(c,t)&&c.prefix.length>a&&(a=c.prefix.length,r=s)}return r}function Fn(e,n){return 0===e.lastIndexOf(n,0)}function Gn(e,n){return Fn(e,n)?e.substr(n.length):e}function Vn(e,n,t){return void 0===t&&(t=Xe),Fn(t(e),t(n))?e.substring(n.length):void 0}function Bn(e,n){var t=e.prefix,r=e.suffix;return n.length>=t.length+r.length&&Fn(n,t)&&En(n,r)}function Kn(e,n){return function(t){return e(t)&&n(t)}}function jn(e,n){return function(t){return e(t)||n(t)}}function Un(e){}function Hn(e){return void 0===e?void 0:[e]}function Wn(e,n,t,r,a,i){i=i||qe;var o=0,s=0,c=e.length,l=n.length;while(o<c&&s<l){var u=e[o],d=n[s],p=t(u,d);-1===p?(r(u),o++):1===p?(a(d),s++):(i(d,u),o++,s++)}while(o<c)r(e[o++]);while(s<l)a(n[s++])}function qn(e,n){for(var t=Array(e),r=0;r<e;r++)t[r]=n(r);return t}e.getUILocale=fn,e.setUILocale=_n,e.compareStringsCaseSensitiveUI=gn,e.compareProperties=vn,e.compareBooleans=yn,e.getSpellingSuggestion=hn,e.endsWith=En,e.removeSuffix=Tn,e.tryRemoveSuffix=Sn,e.stringContains=xn,e.fileExtensionIs=Cn,e.fileExtensionIsOneOf=An,e.removeMinAndVersionNumbers=Dn,e.orderedRemoveItem=Ln,e.orderedRemoveItemAt=kn,e.unorderedRemoveItemAt=Nn,e.unorderedRemoveItem=In,e.createGetCanonicalFileName=Mn,e.patternText=On,e.matchedText=Pn,e.findBestPatternMatch=Rn,e.startsWith=Fn,e.removePrefix=Gn,e.tryRemovePrefix=Vn,e.and=Kn,e.or=jn,e.assertType=Un,e.singleElementArray=Hn,e.enumerateInsertsAndDeletes=Wn,e.fill=qn}(u||(u={})),function(e){e.timestamp="undefined"!==typeof performance&&performance.now?function(){return performance.now()}:Date.now?Date.now:function(){return+new Date}}(u||(u={})),function(e){(function(n){var t,r,a,i="function"===typeof onProfilerEvent&&!0===onProfilerEvent.profiler?onProfilerEvent:function(){},o=!1,s=0;function c(e,t,r,a){return e?l(t,r,a):n.nullTimer}function l(n,t,r){var a=0;return{enter:i,exit:o};function i(){1===++a&&u(t)}function o(){0===--a?(u(r),d(n,t,r)):a<0&&e.Debug.fail("enter/exit count does not match.")}}function u(n){o&&(r.set(n,e.timestamp()),t.set(n,(t.get(n)||0)+1),i(n))}function d(n,t,i){if(o){var c=i&&r.get(i)||e.timestamp(),l=t&&r.get(t)||s;a.set(n,(a.get(n)||0)+(c-l))}}function p(e){return t&&t.get(e)||0}function m(e){return a&&a.get(e)||0}function f(e){a.forEach(function(n,t){e(t,n)})}function _(){t=e.createMap(),r=e.createMap(),a=e.createMap(),o=!0,s=e.timestamp()}function g(){o=!1}n.createTimerIf=c,n.createTimer=l,n.nullTimer={enter:e.noop,exit:e.noop},n.mark=u,n.measure=d,n.getCount=p,n.getDuration=m,n.forEachMeasure=f,n.enable=_,n.disable=g})(e.performance||(e.performance={}))}(u||(u={})),function(e){var n=/^(0|[1-9]\d*)(?:\.(0|[1-9]\d*)(?:\.(0|[1-9]\d*)(?:\-([a-z0-9-.]+))?(?:\+([a-z0-9-.]+))?)?)?$/i,t=/^(?:0|[1-9]\d*|[a-z-][a-z0-9-]*)(?:\.(?:0|[1-9]\d*|[a-z-][a-z0-9-]*))*$/i,r=/^[a-z0-9-]+(?:\.[a-z0-9-]+)*$/i,a=/^(0|[1-9]\d*)$/,i=function(){function n(n,a,i,s,c){if(void 0===a&&(a=0),void 0===i&&(i=0),void 0===s&&(s=""),void 0===c&&(c=""),"string"===typeof n){var l=e.Debug.assertDefined(o(n),"Invalid version");n=l.major,a=l.minor,i=l.patch,s=l.prerelease,c=l.build}e.Debug.assert(n>=0,"Invalid argument: major"),e.Debug.assert(a>=0,"Invalid argument: minor"),e.Debug.assert(i>=0,"Invalid argument: patch"),e.Debug.assert(!s||t.test(s),"Invalid argument: prerelease"),e.Debug.assert(!c||r.test(c),"Invalid argument: build"),this.major=n,this.minor=a,this.patch=i,this.prerelease=s?s.split("."):e.emptyArray,this.build=c?c.split("."):e.emptyArray}return n.tryParse=function(e){var t=o(e);if(t){var r=t.major,a=t.minor,i=t.patch,s=t.prerelease,c=t.build;return new n(r,a,i,s,c)}},n.prototype.compareTo=function(n){return this===n?0:void 0===n?1:e.compareValues(this.major,n.major)||e.compareValues(this.minor,n.minor)||e.compareValues(this.patch,n.patch)||s(this.prerelease,n.prerelease)},n.prototype.increment=function(t){switch(t){case"major":return new n(this.major+1,0,0);case"minor":return new n(this.major,this.minor+1,0);case"patch":return new n(this.major,this.minor,this.patch+1);default:return e.Debug.assertNever(t)}},n.prototype.toString=function(){var n=this.major+"."+this.minor+"."+this.patch;return e.some(this.prerelease)&&(n+="-"+this.prerelease.join(".")),e.some(this.build)&&(n+="+"+this.build.join(".")),n},n.zero=new n(0,0,0),n}();function o(e){var a=n.exec(e);if(a){var i=a[1],o=a[2],s=void 0===o?"0":o,c=a[3],l=void 0===c?"0":c,u=a[4],d=void 0===u?"":u,p=a[5],m=void 0===p?"":p;if((!d||t.test(d))&&(!m||r.test(m)))return{major:parseInt(i,10),minor:parseInt(s,10),patch:parseInt(l,10),prerelease:d,build:m}}}function s(n,t){if(n===t)return 0;if(0===n.length)return 0===t.length?0:1;if(0===t.length)return-1;for(var r=Math.min(n.length,t.length),i=0;i<r;i++){var o=n[i],s=t[i];if(o!==s){var c=a.test(o),l=a.test(s);if(c||l){if(c!==l)return c?-1:1;var u=e.compareValues(+o,+s);if(u)return u}else{u=e.compareStringsCaseSensitive(o,s);if(u)return u}}}return e.compareValues(n.length,t.length)}e.Version=i;var c=function(){function n(n){this._alternatives=n?e.Debug.assertDefined(f(n),"Invalid range spec."):e.emptyArray}return n.tryParse=function(e){var t=f(e);if(t){var r=new n("");return r._alternatives=t,r}},n.prototype.test=function(e){return"string"===typeof e&&(e=new i(e)),b(e,this._alternatives)},n.prototype.toString=function(){return S(this._alternatives)},n}();e.VersionRange=c;var l=/\s*\|\|\s*/g,u=/\s+/g,d=/^([xX*0]|[1-9]\d*)(?:\.([xX*0]|[1-9]\d*)(?:\.([xX*0]|[1-9]\d*)(?:-([a-z0-9-.]+))?(?:\+([a-z0-9-.]+))?)?)?$/i,p=/^\s*([a-z0-9-+.*]+)\s+-\s+([a-z0-9-+.*]+)\s*$/i,m=/^\s*(~|\^|<|<=|>|>=|=)?\s*([a-z0-9-+.*]+)$/i;function f(e){for(var n=[],t=0,r=e.trim().split(l);t<r.length;t++){var a=r[t];if(a){var i=[],o=p.exec(a);if(o){if(!g(o[1],o[2],i))return}else for(var s=0,c=a.split(u);s<c.length;s++){var d=c[s],f=m.exec(d);if(!f||!v(f[1],f[2],i))return}n.push(i)}}return n}function _(e){var n=d.exec(e);if(n){var t=n[1],r=n[2],a=void 0===r?"*":r,o=n[3],s=void 0===o?"*":o,c=n[4],l=n[5],u=new i(y(t)?0:parseInt(t,10),y(t)||y(a)?0:parseInt(a,10),y(t)||y(a)||y(s)?0:parseInt(s,10),c,l);return{version:u,major:t,minor:a,patch:s}}}function g(e,n,t){var r=_(e);if(!r)return!1;var a=_(n);return!!a&&(y(r.major)||t.push(h(">=",r.version)),y(a.major)||t.push(y(a.minor)?h("<",a.version.increment("major")):y(a.patch)?h("<",a.version.increment("minor")):h("<=",a.version)),!0)}function v(e,n,t){var r=_(n);if(!r)return!1;var a=r.version,o=r.major,s=r.minor,c=r.patch;if(y(o))"<"!==e&&">"!==e||t.push(h("<",i.zero));else switch(e){case"~":t.push(h(">=",a)),t.push(h("<",a.increment(y(s)?"major":"minor")));break;case"^":t.push(h(">=",a)),t.push(h("<",a.increment(a.major>0||y(s)?"major":a.minor>0||y(c)?"minor":"patch")));break;case"<":case">=":t.push(h(e,a));break;case"<=":case">":t.push(y(s)?h("<="===e?"<":">=",a.increment("major")):y(c)?h("<="===e?"<":">=",a.increment("minor")):h(e,a));break;case"=":case void 0:y(s)||y(c)?(t.push(h(">=",a)),t.push(h("<",a.increment(y(s)?"major":"minor")))):t.push(h("=",a));break;default:return!1}return!0}function y(e){return"*"===e||"x"===e||"X"===e}function h(e,n){return{operator:e,operand:n}}function b(e,n){if(0===n.length)return!0;for(var t=0,r=n;t<r.length;t++){var a=r[t];if(E(e,a))return!0}return!1}function E(e,n){for(var t=0,r=n;t<r.length;t++){var a=r[t];if(!T(e,a.operator,a.operand))return!1}return!0}function T(n,t,r){var a=n.compareTo(r);switch(t){case"<":return a<0;case"<=":return a<=0;case">":return a>0;case">=":return a>=0;case"=":return 0===a;default:return e.Debug.assertNever(t)}}function S(n){return e.map(n,x).join(" || ")||"*"}function x(n){return e.map(n,C).join(" ")}function C(e){return""+e.operator+e.operand}}(u||(u={})),function(e){(function(e){e[e["Unknown"]=0]="Unknown",e[e["EndOfFileToken"]=1]="EndOfFileToken",e[e["SingleLineCommentTrivia"]=2]="SingleLineCommentTrivia",e[e["MultiLineCommentTrivia"]=3]="MultiLineCommentTrivia",e[e["NewLineTrivia"]=4]="NewLineTrivia",e[e["WhitespaceTrivia"]=5]="WhitespaceTrivia",e[e["ShebangTrivia"]=6]="ShebangTrivia",e[e["ConflictMarkerTrivia"]=7]="ConflictMarkerTrivia",e[e["NumericLiteral"]=8]="NumericLiteral",e[e["BigIntLiteral"]=9]="BigIntLiteral",e[e["StringLiteral"]=10]="StringLiteral",e[e["JsxText"]=11]="JsxText",e[e["JsxTextAllWhiteSpaces"]=12]="JsxTextAllWhiteSpaces",e[e["RegularExpressionLiteral"]=13]="RegularExpressionLiteral",e[e["NoSubstitutionTemplateLiteral"]=14]="NoSubstitutionTemplateLiteral",e[e["TemplateHead"]=15]="TemplateHead",e[e["TemplateMiddle"]=16]="TemplateMiddle",e[e["TemplateTail"]=17]="TemplateTail",e[e["OpenBraceToken"]=18]="OpenBraceToken",e[e["CloseBraceToken"]=19]="CloseBraceToken",e[e["OpenParenToken"]=20]="OpenParenToken",e[e["CloseParenToken"]=21]="CloseParenToken",e[e["OpenBracketToken"]=22]="OpenBracketToken",e[e["CloseBracketToken"]=23]="CloseBracketToken",e[e["DotToken"]=24]="DotToken",e[e["DotDotDotToken"]=25]="DotDotDotToken",e[e["SemicolonToken"]=26]="SemicolonToken",e[e["CommaToken"]=27]="CommaToken",e[e["LessThanToken"]=28]="LessThanToken",e[e["LessThanSlashToken"]=29]="LessThanSlashToken",e[e["GreaterThanToken"]=30]="GreaterThanToken",e[e["LessThanEqualsToken"]=31]="LessThanEqualsToken",e[e["GreaterThanEqualsToken"]=32]="GreaterThanEqualsToken",e[e["EqualsEqualsToken"]=33]="EqualsEqualsToken",e[e["ExclamationEqualsToken"]=34]="ExclamationEqualsToken",e[e["EqualsEqualsEqualsToken"]=35]="EqualsEqualsEqualsToken",e[e["ExclamationEqualsEqualsToken"]=36]="ExclamationEqualsEqualsToken",e[e["EqualsGreaterThanToken"]=37]="EqualsGreaterThanToken",e[e["PlusToken"]=38]="PlusToken",e[e["MinusToken"]=39]="MinusToken",e[e["AsteriskToken"]=40]="AsteriskToken",e[e["AsteriskAsteriskToken"]=41]="AsteriskAsteriskToken",e[e["SlashToken"]=42]="SlashToken",e[e["PercentToken"]=43]="PercentToken",e[e["PlusPlusToken"]=44]="PlusPlusToken",e[e["MinusMinusToken"]=45]="MinusMinusToken",e[e["LessThanLessThanToken"]=46]="LessThanLessThanToken",e[e["GreaterThanGreaterThanToken"]=47]="GreaterThanGreaterThanToken",e[e["GreaterThanGreaterThanGreaterThanToken"]=48]="GreaterThanGreaterThanGreaterThanToken",e[e["AmpersandToken"]=49]="AmpersandToken",e[e["BarToken"]=50]="BarToken",e[e["CaretToken"]=51]="CaretToken",e[e["ExclamationToken"]=52]="ExclamationToken",e[e["TildeToken"]=53]="TildeToken",e[e["AmpersandAmpersandToken"]=54]="AmpersandAmpersandToken",e[e["BarBarToken"]=55]="BarBarToken",e[e["QuestionToken"]=56]="QuestionToken",e[e["ColonToken"]=57]="ColonToken",e[e["AtToken"]=58]="AtToken",e[e["EqualsToken"]=59]="EqualsToken",e[e["PlusEqualsToken"]=60]="PlusEqualsToken",e[e["MinusEqualsToken"]=61]="MinusEqualsToken",e[e["AsteriskEqualsToken"]=62]="AsteriskEqualsToken",e[e["AsteriskAsteriskEqualsToken"]=63]="AsteriskAsteriskEqualsToken",e[e["SlashEqualsToken"]=64]="SlashEqualsToken",e[e["PercentEqualsToken"]=65]="PercentEqualsToken",e[e["LessThanLessThanEqualsToken"]=66]="LessThanLessThanEqualsToken",e[e["GreaterThanGreaterThanEqualsToken"]=67]="GreaterThanGreaterThanEqualsToken",e[e["GreaterThanGreaterThanGreaterThanEqualsToken"]=68]="GreaterThanGreaterThanGreaterThanEqualsToken",e[e["AmpersandEqualsToken"]=69]="AmpersandEqualsToken",e[e["BarEqualsToken"]=70]="BarEqualsToken",e[e["CaretEqualsToken"]=71]="CaretEqualsToken",e[e["Identifier"]=72]="Identifier",e[e["BreakKeyword"]=73]="BreakKeyword",e[e["CaseKeyword"]=74]="CaseKeyword",e[e["CatchKeyword"]=75]="CatchKeyword",e[e["ClassKeyword"]=76]="ClassKeyword",e[e["ConstKeyword"]=77]="ConstKeyword",e[e["ContinueKeyword"]=78]="ContinueKeyword",e[e["DebuggerKeyword"]=79]="DebuggerKeyword",e[e["DefaultKeyword"]=80]="DefaultKeyword",e[e["DeleteKeyword"]=81]="DeleteKeyword",e[e["DoKeyword"]=82]="DoKeyword",e[e["ElseKeyword"]=83]="ElseKeyword",e[e["EnumKeyword"]=84]="EnumKeyword",e[e["ExportKeyword"]=85]="ExportKeyword",e[e["ExtendsKeyword"]=86]="ExtendsKeyword",e[e["FalseKeyword"]=87]="FalseKeyword",e[e["FinallyKeyword"]=88]="FinallyKeyword",e[e["ForKeyword"]=89]="ForKeyword",e[e["FunctionKeyword"]=90]="FunctionKeyword",e[e["IfKeyword"]=91]="IfKeyword",e[e["ImportKeyword"]=92]="ImportKeyword",e[e["InKeyword"]=93]="InKeyword",e[e["InstanceOfKeyword"]=94]="InstanceOfKeyword",e[e["NewKeyword"]=95]="NewKeyword",e[e["NullKeyword"]=96]="NullKeyword",e[e["ReturnKeyword"]=97]="ReturnKeyword",e[e["SuperKeyword"]=98]="SuperKeyword",e[e["SwitchKeyword"]=99]="SwitchKeyword",e[e["ThisKeyword"]=100]="ThisKeyword",e[e["ThrowKeyword"]=101]="ThrowKeyword",e[e["TrueKeyword"]=102]="TrueKeyword",e[e["TryKeyword"]=103]="TryKeyword",e[e["TypeOfKeyword"]=104]="TypeOfKeyword",e[e["VarKeyword"]=105]="VarKeyword",e[e["VoidKeyword"]=106]="VoidKeyword",e[e["WhileKeyword"]=107]="WhileKeyword",e[e["WithKeyword"]=108]="WithKeyword",e[e["ImplementsKeyword"]=109]="ImplementsKeyword",e[e["InterfaceKeyword"]=110]="InterfaceKeyword",e[e["LetKeyword"]=111]="LetKeyword",e[e["PackageKeyword"]=112]="PackageKeyword",e[e["PrivateKeyword"]=113]="PrivateKeyword",e[e["ProtectedKeyword"]=114]="ProtectedKeyword",e[e["PublicKeyword"]=115]="PublicKeyword",e[e["StaticKeyword"]=116]="StaticKeyword",e[e["YieldKeyword"]=117]="YieldKeyword",e[e["AbstractKeyword"]=118]="AbstractKeyword",e[e["AsKeyword"]=119]="AsKeyword",e[e["AnyKeyword"]=120]="AnyKeyword",e[e["AsyncKeyword"]=121]="AsyncKeyword",e[e["AwaitKeyword"]=122]="AwaitKeyword",e[e["BooleanKeyword"]=123]="BooleanKeyword",e[e["ConstructorKeyword"]=124]="ConstructorKeyword",e[e["DeclareKeyword"]=125]="DeclareKeyword",e[e["GetKeyword"]=126]="GetKeyword",e[e["InferKeyword"]=127]="InferKeyword",e[e["IsKeyword"]=128]="IsKeyword",e[e["KeyOfKeyword"]=129]="KeyOfKeyword",e[e["ModuleKeyword"]=130]="ModuleKeyword",e[e["NamespaceKeyword"]=131]="NamespaceKeyword",e[e["NeverKeyword"]=132]="NeverKeyword",e[e["ReadonlyKeyword"]=133]="ReadonlyKeyword",e[e["RequireKeyword"]=134]="RequireKeyword",e[e["NumberKeyword"]=135]="NumberKeyword",e[e["ObjectKeyword"]=136]="ObjectKeyword",e[e["SetKeyword"]=137]="SetKeyword",e[e["StringKeyword"]=138]="StringKeyword",e[e["SymbolKeyword"]=139]="SymbolKeyword",e[e["TypeKeyword"]=140]="TypeKeyword",e[e["UndefinedKeyword"]=141]="UndefinedKeyword",e[e["UniqueKeyword"]=142]="UniqueKeyword",e[e["UnknownKeyword"]=143]="UnknownKeyword",e[e["FromKeyword"]=144]="FromKeyword",e[e["GlobalKeyword"]=145]="GlobalKeyword",e[e["BigIntKeyword"]=146]="BigIntKeyword",e[e["OfKeyword"]=147]="OfKeyword",e[e["QualifiedName"]=148]="QualifiedName",e[e["ComputedPropertyName"]=149]="ComputedPropertyName",e[e["TypeParameter"]=150]="TypeParameter",e[e["Parameter"]=151]="Parameter",e[e["Decorator"]=152]="Decorator",e[e["PropertySignature"]=153]="PropertySignature",e[e["PropertyDeclaration"]=154]="PropertyDeclaration",e[e["MethodSignature"]=155]="MethodSignature",e[e["MethodDeclaration"]=156]="MethodDeclaration",e[e["Constructor"]=157]="Constructor",e[e["GetAccessor"]=158]="GetAccessor",e[e["SetAccessor"]=159]="SetAccessor",e[e["CallSignature"]=160]="CallSignature",e[e["ConstructSignature"]=161]="ConstructSignature",e[e["IndexSignature"]=162]="IndexSignature",e[e["TypePredicate"]=163]="TypePredicate",e[e["TypeReference"]=164]="TypeReference",e[e["FunctionType"]=165]="FunctionType",e[e["ConstructorType"]=166]="ConstructorType",e[e["TypeQuery"]=167]="TypeQuery",e[e["TypeLiteral"]=168]="TypeLiteral",e[e["ArrayType"]=169]="ArrayType",e[e["TupleType"]=170]="TupleType",e[e["OptionalType"]=171]="OptionalType",e[e["RestType"]=172]="RestType",e[e["UnionType"]=173]="UnionType",e[e["IntersectionType"]=174]="IntersectionType",e[e["ConditionalType"]=175]="ConditionalType",e[e["InferType"]=176]="InferType",e[e["ParenthesizedType"]=177]="ParenthesizedType",e[e["ThisType"]=178]="ThisType",e[e["TypeOperator"]=179]="TypeOperator",e[e["IndexedAccessType"]=180]="IndexedAccessType",e[e["MappedType"]=181]="MappedType",e[e["LiteralType"]=182]="LiteralType",e[e["ImportType"]=183]="ImportType",e[e["ObjectBindingPattern"]=184]="ObjectBindingPattern",e[e["ArrayBindingPattern"]=185]="ArrayBindingPattern",e[e["BindingElement"]=186]="BindingElement",e[e["ArrayLiteralExpression"]=187]="ArrayLiteralExpression",e[e["ObjectLiteralExpression"]=188]="ObjectLiteralExpression",e[e["PropertyAccessExpression"]=189]="PropertyAccessExpression",e[e["ElementAccessExpression"]=190]="ElementAccessExpression",e[e["CallExpression"]=191]="CallExpression",e[e["NewExpression"]=192]="NewExpression",e[e["TaggedTemplateExpression"]=193]="TaggedTemplateExpression",e[e["TypeAssertionExpression"]=194]="TypeAssertionExpression",e[e["ParenthesizedExpression"]=195]="ParenthesizedExpression",e[e["FunctionExpression"]=196]="FunctionExpression",e[e["ArrowFunction"]=197]="ArrowFunction",e[e["DeleteExpression"]=198]="DeleteExpression",e[e["TypeOfExpression"]=199]="TypeOfExpression",e[e["VoidExpression"]=200]="VoidExpression",e[e["AwaitExpression"]=201]="AwaitExpression",e[e["PrefixUnaryExpression"]=202]="PrefixUnaryExpression",e[e["PostfixUnaryExpression"]=203]="PostfixUnaryExpression",e[e["BinaryExpression"]=204]="BinaryExpression",e[e["ConditionalExpression"]=205]="ConditionalExpression",e[e["TemplateExpression"]=206]="TemplateExpression",e[e["YieldExpression"]=207]="YieldExpression",e[e["SpreadElement"]=208]="SpreadElement",e[e["ClassExpression"]=209]="ClassExpression",e[e["OmittedExpression"]=210]="OmittedExpression",e[e["ExpressionWithTypeArguments"]=211]="ExpressionWithTypeArguments",e[e["AsExpression"]=212]="AsExpression",e[e["NonNullExpression"]=213]="NonNullExpression",e[e["MetaProperty"]=214]="MetaProperty",e[e["SyntheticExpression"]=215]="SyntheticExpression",e[e["TemplateSpan"]=216]="TemplateSpan",e[e["SemicolonClassElement"]=217]="SemicolonClassElement",e[e["Block"]=218]="Block",e[e["VariableStatement"]=219]="VariableStatement",e[e["EmptyStatement"]=220]="EmptyStatement",e[e["ExpressionStatement"]=221]="ExpressionStatement",e[e["IfStatement"]=222]="IfStatement",e[e["DoStatement"]=223]="DoStatement",e[e["WhileStatement"]=224]="WhileStatement",e[e["ForStatement"]=225]="ForStatement",e[e["ForInStatement"]=226]="ForInStatement",e[e["ForOfStatement"]=227]="ForOfStatement",e[e["ContinueStatement"]=228]="ContinueStatement",e[e["BreakStatement"]=229]="BreakStatement",e[e["ReturnStatement"]=230]="ReturnStatement",e[e["WithStatement"]=231]="WithStatement",e[e["SwitchStatement"]=232]="SwitchStatement",e[e["LabeledStatement"]=233]="LabeledStatement",e[e["ThrowStatement"]=234]="ThrowStatement",e[e["TryStatement"]=235]="TryStatement",e[e["DebuggerStatement"]=236]="DebuggerStatement",e[e["VariableDeclaration"]=237]="VariableDeclaration",e[e["VariableDeclarationList"]=238]="VariableDeclarationList",e[e["FunctionDeclaration"]=239]="FunctionDeclaration",e[e["ClassDeclaration"]=240]="ClassDeclaration",e[e["InterfaceDeclaration"]=241]="InterfaceDeclaration",e[e["TypeAliasDeclaration"]=242]="TypeAliasDeclaration",e[e["EnumDeclaration"]=243]="EnumDeclaration",e[e["ModuleDeclaration"]=244]="ModuleDeclaration",e[e["ModuleBlock"]=245]="ModuleBlock",e[e["CaseBlock"]=246]="CaseBlock",e[e["NamespaceExportDeclaration"]=247]="NamespaceExportDeclaration",e[e["ImportEqualsDeclaration"]=248]="ImportEqualsDeclaration",e[e["ImportDeclaration"]=249]="ImportDeclaration",e[e["ImportClause"]=250]="ImportClause",e[e["NamespaceImport"]=251]="NamespaceImport",e[e["NamedImports"]=252]="NamedImports",e[e["ImportSpecifier"]=253]="ImportSpecifier",e[e["ExportAssignment"]=254]="ExportAssignment",e[e["ExportDeclaration"]=255]="ExportDeclaration",e[e["NamedExports"]=256]="NamedExports",e[e["ExportSpecifier"]=257]="ExportSpecifier",e[e["MissingDeclaration"]=258]="MissingDeclaration",e[e["ExternalModuleReference"]=259]="ExternalModuleReference",e[e["JsxElement"]=260]="JsxElement",e[e["JsxSelfClosingElement"]=261]="JsxSelfClosingElement",e[e["JsxOpeningElement"]=262]="JsxOpeningElement",e[e["JsxClosingElement"]=263]="JsxClosingElement",e[e["JsxFragment"]=264]="JsxFragment",e[e["JsxOpeningFragment"]=265]="JsxOpeningFragment",e[e["JsxClosingFragment"]=266]="JsxClosingFragment",e[e["JsxAttribute"]=267]="JsxAttribute",e[e["JsxAttributes"]=268]="JsxAttributes",e[e["JsxSpreadAttribute"]=269]="JsxSpreadAttribute",e[e["JsxExpression"]=270]="JsxExpression",e[e["CaseClause"]=271]="CaseClause",e[e["DefaultClause"]=272]="DefaultClause",e[e["HeritageClause"]=273]="HeritageClause",e[e["CatchClause"]=274]="CatchClause",e[e["PropertyAssignment"]=275]="PropertyAssignment",e[e["ShorthandPropertyAssignment"]=276]="ShorthandPropertyAssignment",e[e["SpreadAssignment"]=277]="SpreadAssignment",e[e["EnumMember"]=278]="EnumMember",e[e["SourceFile"]=279]="SourceFile",e[e["Bundle"]=280]="Bundle",e[e["UnparsedSource"]=281]="UnparsedSource",e[e["InputFiles"]=282]="InputFiles",e[e["JSDocTypeExpression"]=283]="JSDocTypeExpression",e[e["JSDocAllType"]=284]="JSDocAllType",e[e["JSDocUnknownType"]=285]="JSDocUnknownType",e[e["JSDocNullableType"]=286]="JSDocNullableType",e[e["JSDocNonNullableType"]=287]="JSDocNonNullableType",e[e["JSDocOptionalType"]=288]="JSDocOptionalType",e[e["JSDocFunctionType"]=289]="JSDocFunctionType",e[e["JSDocVariadicType"]=290]="JSDocVariadicType",e[e["JSDocComment"]=291]="JSDocComment",e[e["JSDocTypeLiteral"]=292]="JSDocTypeLiteral",e[e["JSDocSignature"]=293]="JSDocSignature",e[e["JSDocTag"]=294]="JSDocTag",e[e["JSDocAugmentsTag"]=295]="JSDocAugmentsTag",e[e["JSDocClassTag"]=296]="JSDocClassTag",e[e["JSDocCallbackTag"]=297]="JSDocCallbackTag",e[e["JSDocEnumTag"]=298]="JSDocEnumTag",e[e["JSDocParameterTag"]=299]="JSDocParameterTag",e[e["JSDocReturnTag"]=300]="JSDocReturnTag",e[e["JSDocThisTag"]=301]="JSDocThisTag",e[e["JSDocTypeTag"]=302]="JSDocTypeTag",e[e["JSDocTemplateTag"]=303]="JSDocTemplateTag",e[e["JSDocTypedefTag"]=304]="JSDocTypedefTag",e[e["JSDocPropertyTag"]=305]="JSDocPropertyTag",e[e["SyntaxList"]=306]="SyntaxList",e[e["NotEmittedStatement"]=307]="NotEmittedStatement",e[e["PartiallyEmittedExpression"]=308]="PartiallyEmittedExpression",e[e["CommaListExpression"]=309]="CommaListExpression",e[e["MergeDeclarationMarker"]=310]="MergeDeclarationMarker",e[e["EndOfDeclarationMarker"]=311]="EndOfDeclarationMarker",e[e["Count"]=312]="Count",e[e["FirstAssignment"]=59]="FirstAssignment",e[e["LastAssignment"]=71]="LastAssignment",e[e["FirstCompoundAssignment"]=60]="FirstCompoundAssignment",e[e["LastCompoundAssignment"]=71]="LastCompoundAssignment",e[e["FirstReservedWord"]=73]="FirstReservedWord",e[e["LastReservedWord"]=108]="LastReservedWord",e[e["FirstKeyword"]=73]="FirstKeyword",e[e["LastKeyword"]=147]="LastKeyword",e[e["FirstFutureReservedWord"]=109]="FirstFutureReservedWord",e[e["LastFutureReservedWord"]=117]="LastFutureReservedWord",e[e["FirstTypeNode"]=163]="FirstTypeNode",e[e["LastTypeNode"]=183]="LastTypeNode",e[e["FirstPunctuation"]=18]="FirstPunctuation",e[e["LastPunctuation"]=71]="LastPunctuation",e[e["FirstToken"]=0]="FirstToken",e[e["LastToken"]=147]="LastToken",e[e["FirstTriviaToken"]=2]="FirstTriviaToken",e[e["LastTriviaToken"]=7]="LastTriviaToken",e[e["FirstLiteralToken"]=8]="FirstLiteralToken",e[e["LastLiteralToken"]=14]="LastLiteralToken",e[e["FirstTemplateToken"]=14]="FirstTemplateToken",e[e["LastTemplateToken"]=17]="LastTemplateToken",e[e["FirstBinaryOperator"]=28]="FirstBinaryOperator",e[e["LastBinaryOperator"]=71]="LastBinaryOperator",e[e["FirstNode"]=148]="FirstNode",e[e["FirstJSDocNode"]=283]="FirstJSDocNode",e[e["LastJSDocNode"]=305]="LastJSDocNode",e[e["FirstJSDocTagNode"]=294]="FirstJSDocTagNode",e[e["LastJSDocTagNode"]=305]="LastJSDocTagNode",e[e["FirstContextualKeyword"]=118]="FirstContextualKeyword",e[e["LastContextualKeyword"]=147]="LastContextualKeyword"})(e.SyntaxKind||(e.SyntaxKind={})),function(e){e[e["None"]=0]="None",e[e["Let"]=1]="Let",e[e["Const"]=2]="Const",e[e["NestedNamespace"]=4]="NestedNamespace",e[e["Synthesized"]=8]="Synthesized",e[e["Namespace"]=16]="Namespace",e[e["ExportContext"]=32]="ExportContext",e[e["ContainsThis"]=64]="ContainsThis",e[e["HasImplicitReturn"]=128]="HasImplicitReturn",e[e["HasExplicitReturn"]=256]="HasExplicitReturn",e[e["GlobalAugmentation"]=512]="GlobalAugmentation",e[e["HasAsyncFunctions"]=1024]="HasAsyncFunctions",e[e["DisallowInContext"]=2048]="DisallowInContext",e[e["YieldContext"]=4096]="YieldContext",e[e["DecoratorContext"]=8192]="DecoratorContext",e[e["AwaitContext"]=16384]="AwaitContext",e[e["ThisNodeHasError"]=32768]="ThisNodeHasError",e[e["JavaScriptFile"]=65536]="JavaScriptFile",e[e["ThisNodeOrAnySubNodesHasError"]=131072]="ThisNodeOrAnySubNodesHasError",e[e["HasAggregatedChildData"]=262144]="HasAggregatedChildData",e[e["PossiblyContainsDynamicImport"]=524288]="PossiblyContainsDynamicImport",e[e["PossiblyContainsImportMeta"]=1048576]="PossiblyContainsImportMeta",e[e["JSDoc"]=2097152]="JSDoc",e[e["Ambient"]=4194304]="Ambient",e[e["InWithStatement"]=8388608]="InWithStatement",e[e["JsonFile"]=16777216]="JsonFile",e[e["BlockScoped"]=3]="BlockScoped",e[e["ReachabilityCheckFlags"]=384]="ReachabilityCheckFlags",e[e["ReachabilityAndEmitFlags"]=1408]="ReachabilityAndEmitFlags",e[e["ContextFlags"]=12679168]="ContextFlags",e[e["TypeExcludesFlags"]=20480]="TypeExcludesFlags",e[e["PermanentlySetIncrementalFlags"]=1572864]="PermanentlySetIncrementalFlags"}(e.NodeFlags||(e.NodeFlags={})),function(e){e[e["None"]=0]="None",e[e["Export"]=1]="Export",e[e["Ambient"]=2]="Ambient",e[e["Public"]=4]="Public",e[e["Private"]=8]="Private",e[e["Protected"]=16]="Protected",e[e["Static"]=32]="Static",e[e["Readonly"]=64]="Readonly",e[e["Abstract"]=128]="Abstract",e[e["Async"]=256]="Async",e[e["Default"]=512]="Default",e[e["Const"]=2048]="Const",e[e["HasComputedFlags"]=536870912]="HasComputedFlags",e[e["AccessibilityModifier"]=28]="AccessibilityModifier",e[e["ParameterPropertyModifier"]=92]="ParameterPropertyModifier",e[e["NonPublicAccessibilityModifier"]=24]="NonPublicAccessibilityModifier",e[e["TypeScriptModifier"]=2270]="TypeScriptModifier",e[e["ExportDefault"]=513]="ExportDefault",e[e["All"]=3071]="All"}(e.ModifierFlags||(e.ModifierFlags={})),function(e){e[e["None"]=0]="None",e[e["IntrinsicNamedElement"]=1]="IntrinsicNamedElement",e[e["IntrinsicIndexedElement"]=2]="IntrinsicIndexedElement",e[e["IntrinsicElement"]=3]="IntrinsicElement"}(e.JsxFlags||(e.JsxFlags={})),function(e){e[e["Succeeded"]=1]="Succeeded",e[e["Failed"]=2]="Failed",e[e["FailedAndReported"]=3]="FailedAndReported"}(e.RelationComparisonResult||(e.RelationComparisonResult={})),function(e){e[e["None"]=0]="None",e[e["Auto"]=1]="Auto",e[e["Loop"]=2]="Loop",e[e["Unique"]=3]="Unique",e[e["Node"]=4]="Node",e[e["KindMask"]=7]="KindMask",e[e["ReservedInNestedScopes"]=8]="ReservedInNestedScopes",e[e["Optimistic"]=16]="Optimistic",e[e["FileLevel"]=32]="FileLevel"}(e.GeneratedIdentifierFlags||(e.GeneratedIdentifierFlags={})),function(e){e[e["None"]=0]="None",e[e["PrecedingLineBreak"]=1]="PrecedingLineBreak",e[e["PrecedingJSDocComment"]=2]="PrecedingJSDocComment",e[e["Unterminated"]=4]="Unterminated",e[e["ExtendedUnicodeEscape"]=8]="ExtendedUnicodeEscape",e[e["Scientific"]=16]="Scientific",e[e["Octal"]=32]="Octal",e[e["HexSpecifier"]=64]="HexSpecifier",e[e["BinarySpecifier"]=128]="BinarySpecifier",e[e["OctalSpecifier"]=256]="OctalSpecifier",e[e["ContainsSeparator"]=512]="ContainsSeparator",e[e["BinaryOrOctalSpecifier"]=384]="BinaryOrOctalSpecifier",e[e["NumericLiteralFlags"]=1008]="NumericLiteralFlags"}(e.TokenFlags||(e.TokenFlags={})),function(e){e[e["Unreachable"]=1]="Unreachable",e[e["Start"]=2]="Start",e[e["BranchLabel"]=4]="BranchLabel",e[e["LoopLabel"]=8]="LoopLabel",e[e["Assignment"]=16]="Assignment",e[e["TrueCondition"]=32]="TrueCondition",e[e["FalseCondition"]=64]="FalseCondition",e[e["SwitchClause"]=128]="SwitchClause",e[e["ArrayMutation"]=256]="ArrayMutation",e[e["Referenced"]=512]="Referenced",e[e["Shared"]=1024]="Shared",e[e["PreFinally"]=2048]="PreFinally",e[e["AfterFinally"]=4096]="AfterFinally",e[e["Label"]=12]="Label",e[e["Condition"]=96]="Condition"}(e.FlowFlags||(e.FlowFlags={}));var n,t=function(){function e(){}return e}();function r(e,t){void 0===t&&(t=!0);var r=n[e.category];return t?r.toLowerCase():r}function a(e){return e}e.OperationCanceledException=t,function(e){e[e["Not"]=0]="Not",e[e["SafeModules"]=1]="SafeModules",e[e["Completely"]=2]="Completely"}(e.StructureIsReused||(e.StructureIsReused={})),function(e){e[e["Success"]=0]="Success",e[e["DiagnosticsPresent_OutputsSkipped"]=1]="DiagnosticsPresent_OutputsSkipped",e[e["DiagnosticsPresent_OutputsGenerated"]=2]="DiagnosticsPresent_OutputsGenerated"}(e.ExitStatus||(e.ExitStatus={})),function(e){e[e["None"]=0]="None",e[e["Literal"]=1]="Literal",e[e["Subtype"]=2]="Subtype"}(e.UnionReduction||(e.UnionReduction={})),function(e){e[e["None"]=0]="None",e[e["NoTruncation"]=1]="NoTruncation",e[e["WriteArrayAsGenericType"]=2]="WriteArrayAsGenericType",e[e["GenerateNamesForShadowedTypeParams"]=4]="GenerateNamesForShadowedTypeParams",e[e["UseStructuralFallback"]=8]="UseStructuralFallback",e[e["ForbidIndexedAccessSymbolReferences"]=16]="ForbidIndexedAccessSymbolReferences",e[e["WriteTypeArgumentsOfSignature"]=32]="WriteTypeArgumentsOfSignature",e[e["UseFullyQualifiedType"]=64]="UseFullyQualifiedType",e[e["UseOnlyExternalAliasing"]=128]="UseOnlyExternalAliasing",e[e["SuppressAnyReturnType"]=256]="SuppressAnyReturnType",e[e["WriteTypeParametersInQualifiedName"]=512]="WriteTypeParametersInQualifiedName",e[e["MultilineObjectLiterals"]=1024]="MultilineObjectLiterals",e[e["WriteClassExpressionAsTypeLiteral"]=2048]="WriteClassExpressionAsTypeLiteral",e[e["UseTypeOfFunction"]=4096]="UseTypeOfFunction",e[e["OmitParameterModifiers"]=8192]="OmitParameterModifiers",e[e["UseAliasDefinedOutsideCurrentScope"]=16384]="UseAliasDefinedOutsideCurrentScope",e[e["AllowThisInObjectLiteral"]=32768]="AllowThisInObjectLiteral",e[e["AllowQualifedNameInPlaceOfIdentifier"]=65536]="AllowQualifedNameInPlaceOfIdentifier",e[e["AllowAnonymousIdentifier"]=131072]="AllowAnonymousIdentifier",e[e["AllowEmptyUnionOrIntersection"]=262144]="AllowEmptyUnionOrIntersection",e[e["AllowEmptyTuple"]=524288]="AllowEmptyTuple",e[e["AllowUniqueESSymbolType"]=1048576]="AllowUniqueESSymbolType",e[e["AllowEmptyIndexInfoType"]=2097152]="AllowEmptyIndexInfoType",e[e["AllowNodeModulesRelativePaths"]=67108864]="AllowNodeModulesRelativePaths",e[e["DoNotIncludeSymbolChain"]=134217728]="DoNotIncludeSymbolChain",e[e["IgnoreErrors"]=70221824]="IgnoreErrors",e[e["InObjectTypeLiteral"]=4194304]="InObjectTypeLiteral",e[e["InTypeAlias"]=8388608]="InTypeAlias",e[e["InInitialEntityName"]=16777216]="InInitialEntityName",e[e["InReverseMappedType"]=33554432]="InReverseMappedType"}(e.NodeBuilderFlags||(e.NodeBuilderFlags={})),function(e){e[e["None"]=0]="None",e[e["NoTruncation"]=1]="NoTruncation",e[e["WriteArrayAsGenericType"]=2]="WriteArrayAsGenericType",e[e["UseStructuralFallback"]=8]="UseStructuralFallback",e[e["WriteTypeArgumentsOfSignature"]=32]="WriteTypeArgumentsOfSignature",e[e["UseFullyQualifiedType"]=64]="UseFullyQualifiedType",e[e["SuppressAnyReturnType"]=256]="SuppressAnyReturnType",e[e["MultilineObjectLiterals"]=1024]="MultilineObjectLiterals",e[e["WriteClassExpressionAsTypeLiteral"]=2048]="WriteClassExpressionAsTypeLiteral",e[e["UseTypeOfFunction"]=4096]="UseTypeOfFunction",e[e["OmitParameterModifiers"]=8192]="OmitParameterModifiers",e[e["UseAliasDefinedOutsideCurrentScope"]=16384]="UseAliasDefinedOutsideCurrentScope",e[e["AllowUniqueESSymbolType"]=1048576]="AllowUniqueESSymbolType",e[e["AddUndefined"]=131072]="AddUndefined",e[e["WriteArrowStyleSignature"]=262144]="WriteArrowStyleSignature",e[e["InArrayType"]=524288]="InArrayType",e[e["InElementType"]=2097152]="InElementType",e[e["InFirstTypeArgument"]=4194304]="InFirstTypeArgument",e[e["InTypeAlias"]=8388608]="InTypeAlias",e[e["WriteOwnNameForAnyLike"]=0]="WriteOwnNameForAnyLike",e[e["NodeBuilderFlagsMask"]=9469291]="NodeBuilderFlagsMask"}(e.TypeFormatFlags||(e.TypeFormatFlags={})),function(e){e[e["None"]=0]="None",e[e["WriteTypeParametersOrArguments"]=1]="WriteTypeParametersOrArguments",e[e["UseOnlyExternalAliasing"]=2]="UseOnlyExternalAliasing",e[e["AllowAnyNodeKind"]=4]="AllowAnyNodeKind",e[e["UseAliasDefinedOutsideCurrentScope"]=8]="UseAliasDefinedOutsideCurrentScope",e[e["DoNotIncludeSymbolChain"]=16]="DoNotIncludeSymbolChain"}(e.SymbolFormatFlags||(e.SymbolFormatFlags={})),function(e){e[e["Accessible"]=0]="Accessible",e[e["NotAccessible"]=1]="NotAccessible",e[e["CannotBeNamed"]=2]="CannotBeNamed"}(e.SymbolAccessibility||(e.SymbolAccessibility={})),function(e){e[e["UnionOrIntersection"]=0]="UnionOrIntersection",e[e["Spread"]=1]="Spread"}(e.SyntheticSymbolKind||(e.SyntheticSymbolKind={})),function(e){e[e["This"]=0]="This",e[e["Identifier"]=1]="Identifier"}(e.TypePredicateKind||(e.TypePredicateKind={})),function(e){e[e["Unknown"]=0]="Unknown",e[e["TypeWithConstructSignatureAndValue"]=1]="TypeWithConstructSignatureAndValue",e[e["VoidNullableOrNeverType"]=2]="VoidNullableOrNeverType",e[e["NumberLikeType"]=3]="NumberLikeType",e[e["BigIntLikeType"]=4]="BigIntLikeType",e[e["StringLikeType"]=5]="StringLikeType",e[e["BooleanType"]=6]="BooleanType",e[e["ArrayLikeType"]=7]="ArrayLikeType",e[e["ESSymbolType"]=8]="ESSymbolType",e[e["Promise"]=9]="Promise",e[e["TypeWithCallSignature"]=10]="TypeWithCallSignature",e[e["ObjectType"]=11]="ObjectType"}(e.TypeReferenceSerializationKind||(e.TypeReferenceSerializationKind={})),function(e){e[e["None"]=0]="None",e[e["FunctionScopedVariable"]=1]="FunctionScopedVariable",e[e["BlockScopedVariable"]=2]="BlockScopedVariable",e[e["Property"]=4]="Property",e[e["EnumMember"]=8]="EnumMember",e[e["Function"]=16]="Function",e[e["Class"]=32]="Class",e[e["Interface"]=64]="Interface",e[e["ConstEnum"]=128]="ConstEnum",e[e["RegularEnum"]=256]="RegularEnum",e[e["ValueModule"]=512]="ValueModule",e[e["NamespaceModule"]=1024]="NamespaceModule",e[e["TypeLiteral"]=2048]="TypeLiteral",e[e["ObjectLiteral"]=4096]="ObjectLiteral",e[e["Method"]=8192]="Method",e[e["Constructor"]=16384]="Constructor",e[e["GetAccessor"]=32768]="GetAccessor",e[e["SetAccessor"]=65536]="SetAccessor",e[e["Signature"]=131072]="Signature",e[e["TypeParameter"]=262144]="TypeParameter",e[e["TypeAlias"]=524288]="TypeAlias",e[e["ExportValue"]=1048576]="ExportValue",e[e["Alias"]=2097152]="Alias",e[e["Prototype"]=4194304]="Prototype",e[e["ExportStar"]=8388608]="ExportStar",e[e["Optional"]=16777216]="Optional",e[e["Transient"]=33554432]="Transient",e[e["Assignment"]=67108864]="Assignment",e[e["ModuleExports"]=134217728]="ModuleExports",e[e["All"]=67108863]="All",e[e["Enum"]=384]="Enum",e[e["Variable"]=3]="Variable",e[e["Value"]=67220415]="Value",e[e["Type"]=67897832]="Type",e[e["Namespace"]=1920]="Namespace",e[e["Module"]=1536]="Module",e[e["Accessor"]=98304]="Accessor",e[e["FunctionScopedVariableExcludes"]=67220414]="FunctionScopedVariableExcludes",e[e["BlockScopedVariableExcludes"]=67220415]="BlockScopedVariableExcludes",e[e["ParameterExcludes"]=67220415]="ParameterExcludes",e[e["PropertyExcludes"]=0]="PropertyExcludes",e[e["EnumMemberExcludes"]=68008959]="EnumMemberExcludes",e[e["FunctionExcludes"]=67219887]="FunctionExcludes",e[e["ClassExcludes"]=68008383]="ClassExcludes",e[e["InterfaceExcludes"]=67897736]="InterfaceExcludes",e[e["RegularEnumExcludes"]=68008191]="RegularEnumExcludes",e[e["ConstEnumExcludes"]=68008831]="ConstEnumExcludes",e[e["ValueModuleExcludes"]=110735]="ValueModuleExcludes",e[e["NamespaceModuleExcludes"]=0]="NamespaceModuleExcludes",e[e["MethodExcludes"]=67212223]="MethodExcludes",e[e["GetAccessorExcludes"]=67154879]="GetAccessorExcludes",e[e["SetAccessorExcludes"]=67187647]="SetAccessorExcludes",e[e["TypeParameterExcludes"]=67635688]="TypeParameterExcludes",e[e["TypeAliasExcludes"]=67897832]="TypeAliasExcludes",e[e["AliasExcludes"]=2097152]="AliasExcludes",e[e["ModuleMember"]=2623475]="ModuleMember",e[e["ExportHasLocal"]=944]="ExportHasLocal",e[e["BlockScoped"]=418]="BlockScoped",e[e["PropertyOrAccessor"]=98308]="PropertyOrAccessor",e[e["ClassMember"]=106500]="ClassMember",e[e["Classifiable"]=2885600]="Classifiable",e[e["LateBindingContainer"]=6240]="LateBindingContainer"}(e.SymbolFlags||(e.SymbolFlags={})),function(e){e[e["Numeric"]=0]="Numeric",e[e["Literal"]=1]="Literal"}(e.EnumKind||(e.EnumKind={})),function(e){e[e["Instantiated"]=1]="Instantiated",e[e["SyntheticProperty"]=2]="SyntheticProperty",e[e["SyntheticMethod"]=4]="SyntheticMethod",e[e["Readonly"]=8]="Readonly",e[e["Partial"]=16]="Partial",e[e["HasNonUniformType"]=32]="HasNonUniformType",e[e["HasLiteralType"]=64]="HasLiteralType",e[e["ContainsPublic"]=128]="ContainsPublic",e[e["ContainsProtected"]=256]="ContainsProtected",e[e["ContainsPrivate"]=512]="ContainsPrivate",e[e["ContainsStatic"]=1024]="ContainsStatic",e[e["Late"]=2048]="Late",e[e["ReverseMapped"]=4096]="ReverseMapped",e[e["OptionalParameter"]=8192]="OptionalParameter",e[e["RestParameter"]=16384]="RestParameter",e[e["Synthetic"]=6]="Synthetic",e[e["Discriminant"]=96]="Discriminant"}(e.CheckFlags||(e.CheckFlags={})),function(e){e["Call"]="__call",e["Constructor"]="__constructor",e["New"]="__new",e["Index"]="__index",e["ExportStar"]="__export",e["Global"]="__global",e["Missing"]="__missing",e["Type"]="__type",e["Object"]="__object",e["JSXAttributes"]="__jsxAttributes",e["Class"]="__class",e["Function"]="__function",e["Computed"]="__computed",e["Resolving"]="__resolving__",e["ExportEquals"]="export=",e["Default"]="default",e["This"]="this"}(e.InternalSymbolName||(e.InternalSymbolName={})),function(e){e[e["TypeChecked"]=1]="TypeChecked",e[e["LexicalThis"]=2]="LexicalThis",e[e["CaptureThis"]=4]="CaptureThis",e[e["CaptureNewTarget"]=8]="CaptureNewTarget",e[e["SuperInstance"]=256]="SuperInstance",e[e["SuperStatic"]=512]="SuperStatic",e[e["ContextChecked"]=1024]="ContextChecked",e[e["AsyncMethodWithSuper"]=2048]="AsyncMethodWithSuper",e[e["AsyncMethodWithSuperBinding"]=4096]="AsyncMethodWithSuperBinding",e[e["CaptureArguments"]=8192]="CaptureArguments",e[e["EnumValuesComputed"]=16384]="EnumValuesComputed",e[e["LexicalModuleMergesWithClass"]=32768]="LexicalModuleMergesWithClass",e[e["LoopWithCapturedBlockScopedBinding"]=65536]="LoopWithCapturedBlockScopedBinding",e[e["ContainsCapturedBlockScopeBinding"]=131072]="ContainsCapturedBlockScopeBinding",e[e["CapturedBlockScopedBinding"]=262144]="CapturedBlockScopedBinding",e[e["BlockScopedBindingInLoop"]=524288]="BlockScopedBindingInLoop",e[e["ClassWithBodyScopedClassBinding"]=1048576]="ClassWithBodyScopedClassBinding",e[e["BodyScopedClassBinding"]=2097152]="BodyScopedClassBinding",e[e["NeedsLoopOutParameter"]=4194304]="NeedsLoopOutParameter",e[e["AssignmentsMarked"]=8388608]="AssignmentsMarked",e[e["ClassWithConstructorReference"]=16777216]="ClassWithConstructorReference",e[e["ConstructorReferenceInClass"]=33554432]="ConstructorReferenceInClass"}(e.NodeCheckFlags||(e.NodeCheckFlags={})),function(e){e[e["Any"]=1]="Any",e[e["Unknown"]=2]="Unknown",e[e["String"]=4]="String",e[e["Number"]=8]="Number",e[e["Boolean"]=16]="Boolean",e[e["Enum"]=32]="Enum",e[e["BigInt"]=64]="BigInt",e[e["StringLiteral"]=128]="StringLiteral",e[e["NumberLiteral"]=256]="NumberLiteral",e[e["BooleanLiteral"]=512]="BooleanLiteral",e[e["EnumLiteral"]=1024]="EnumLiteral",e[e["BigIntLiteral"]=2048]="BigIntLiteral",e[e["ESSymbol"]=4096]="ESSymbol",e[e["UniqueESSymbol"]=8192]="UniqueESSymbol",e[e["Void"]=16384]="Void",e[e["Undefined"]=32768]="Undefined",e[e["Null"]=65536]="Null",e[e["Never"]=131072]="Never",e[e["TypeParameter"]=262144]="TypeParameter",e[e["Object"]=524288]="Object",e[e["Union"]=1048576]="Union",e[e["Intersection"]=2097152]="Intersection",e[e["Index"]=4194304]="Index",e[e["IndexedAccess"]=8388608]="IndexedAccess",e[e["Conditional"]=16777216]="Conditional",e[e["Substitution"]=33554432]="Substitution",e[e["NonPrimitive"]=67108864]="NonPrimitive",e[e["ContainsWideningType"]=134217728]="ContainsWideningType",e[e["ContainsObjectLiteral"]=268435456]="ContainsObjectLiteral",e[e["ContainsAnyFunctionType"]=536870912]="ContainsAnyFunctionType",e[e["AnyOrUnknown"]=3]="AnyOrUnknown",e[e["Nullable"]=98304]="Nullable",e[e["Literal"]=2944]="Literal",e[e["Unit"]=109440]="Unit",e[e["StringOrNumberLiteral"]=384]="StringOrNumberLiteral",e[e["StringOrNumberLiteralOrUnique"]=8576]="StringOrNumberLiteralOrUnique",e[e["DefinitelyFalsy"]=117632]="DefinitelyFalsy",e[e["PossiblyFalsy"]=117724]="PossiblyFalsy",e[e["Intrinsic"]=67359327]="Intrinsic",e[e["Primitive"]=131068]="Primitive",e[e["StringLike"]=132]="StringLike",e[e["NumberLike"]=296]="NumberLike",e[e["BigIntLike"]=2112]="BigIntLike",e[e["BooleanLike"]=528]="BooleanLike",e[e["EnumLike"]=1056]="EnumLike",e[e["ESSymbolLike"]=12288]="ESSymbolLike",e[e["VoidLike"]=49152]="VoidLike",e[e["DisjointDomains"]=67238908]="DisjointDomains",e[e["UnionOrIntersection"]=3145728]="UnionOrIntersection",e[e["StructuredType"]=3670016]="StructuredType",e[e["TypeVariable"]=8650752]="TypeVariable",e[e["InstantiableNonPrimitive"]=58982400]="InstantiableNonPrimitive",e[e["InstantiablePrimitive"]=4194304]="InstantiablePrimitive",e[e["Instantiable"]=63176704]="Instantiable",e[e["StructuredOrInstantiable"]=66846720]="StructuredOrInstantiable",e[e["Narrowable"]=133970943]="Narrowable",e[e["NotUnionOrUnit"]=67637251]="NotUnionOrUnit",e[e["NotPrimitiveUnion"]=66994211]="NotPrimitiveUnion",e[e["RequiresWidening"]=402653184]="RequiresWidening",e[e["PropagatingFlags"]=939524096]="PropagatingFlags",e[e["NonWideningType"]=134217728]="NonWideningType",e[e["Wildcard"]=268435456]="Wildcard",e[e["EmptyObject"]=536870912]="EmptyObject",e[e["ConstructionFlags"]=939524096]="ConstructionFlags",e[e["GenericMappedType"]=134217728]="GenericMappedType"}(e.TypeFlags||(e.TypeFlags={})),function(e){e[e["Class"]=1]="Class",e[e["Interface"]=2]="Interface",e[e["Reference"]=4]="Reference",e[e["Tuple"]=8]="Tuple",e[e["Anonymous"]=16]="Anonymous",e[e["Mapped"]=32]="Mapped",e[e["Instantiated"]=64]="Instantiated",e[e["ObjectLiteral"]=128]="ObjectLiteral",e[e["EvolvingArray"]=256]="EvolvingArray",e[e["ObjectLiteralPatternWithComputedProperties"]=512]="ObjectLiteralPatternWithComputedProperties",e[e["ContainsSpread"]=1024]="ContainsSpread",e[e["ReverseMapped"]=2048]="ReverseMapped",e[e["JsxAttributes"]=4096]="JsxAttributes",e[e["MarkerType"]=8192]="MarkerType",e[e["JSLiteral"]=16384]="JSLiteral",e[e["FreshLiteral"]=32768]="FreshLiteral",e[e["ClassOrInterface"]=3]="ClassOrInterface"}(e.ObjectFlags||(e.ObjectFlags={})),function(e){e[e["Invariant"]=0]="Invariant",e[e["Covariant"]=1]="Covariant",e[e["Contravariant"]=2]="Contravariant",e[e["Bivariant"]=3]="Bivariant",e[e["Independent"]=4]="Independent"}(e.Variance||(e.Variance={})),function(e){e[e["Component"]=0]="Component",e[e["Function"]=1]="Function",e[e["Mixed"]=2]="Mixed"}(e.JsxReferenceKind||(e.JsxReferenceKind={})),function(e){e[e["Call"]=0]="Call",e[e["Construct"]=1]="Construct"}(e.SignatureKind||(e.SignatureKind={})),function(e){e[e["String"]=0]="String",e[e["Number"]=1]="Number"}(e.IndexKind||(e.IndexKind={})),function(e){e[e["NakedTypeVariable"]=1]="NakedTypeVariable",e[e["HomomorphicMappedType"]=2]="HomomorphicMappedType",e[e["MappedTypeConstraint"]=4]="MappedTypeConstraint",e[e["ReturnType"]=8]="ReturnType",e[e["LiteralKeyof"]=16]="LiteralKeyof",e[e["NoConstraints"]=32]="NoConstraints",e[e["AlwaysStrict"]=64]="AlwaysStrict",e[e["PriorityImpliesCombination"]=28]="PriorityImpliesCombination"}(e.InferencePriority||(e.InferencePriority={})),function(e){e[e["None"]=0]="None",e[e["NoDefault"]=1]="NoDefault",e[e["AnyDefault"]=2]="AnyDefault"}(e.InferenceFlags||(e.InferenceFlags={})),function(e){e[e["False"]=0]="False",e[e["Maybe"]=1]="Maybe",e[e["True"]=-1]="True"}(e.Ternary||(e.Ternary={})),function(e){e[e["None"]=0]="None",e[e["ExportsProperty"]=1]="ExportsProperty",e[e["ModuleExports"]=2]="ModuleExports",e[e["PrototypeProperty"]=3]="PrototypeProperty",e[e["ThisProperty"]=4]="ThisProperty",e[e["Property"]=5]="Property",e[e["Prototype"]=6]="Prototype",e[e["ObjectDefinePropertyValue"]=7]="ObjectDefinePropertyValue",e[e["ObjectDefinePropertyExports"]=8]="ObjectDefinePropertyExports",e[e["ObjectDefinePrototypeProperty"]=9]="ObjectDefinePrototypeProperty"}(e.AssignmentDeclarationKind||(e.AssignmentDeclarationKind={})),function(e){e[e["Warning"]=0]="Warning",e[e["Error"]=1]="Error",e[e["Suggestion"]=2]="Suggestion",e[e["Message"]=3]="Message"}(n=e.DiagnosticCategory||(e.DiagnosticCategory={})),e.diagnosticCategoryName=r,function(e){e[e["Classic"]=1]="Classic",e[e["NodeJs"]=2]="NodeJs"}(e.ModuleResolutionKind||(e.ModuleResolutionKind={})),function(e){e[e["None"]=0]="None",e[e["CommonJS"]=1]="CommonJS",e[e["AMD"]=2]="AMD",e[e["UMD"]=3]="UMD",e[e["System"]=4]="System",e[e["ES2015"]=5]="ES2015",e[e["ESNext"]=6]="ESNext"}(e.ModuleKind||(e.ModuleKind={})),function(e){e[e["None"]=0]="None",e[e["Preserve"]=1]="Preserve",e[e["React"]=2]="React",e[e["ReactNative"]=3]="ReactNative"}(e.JsxEmit||(e.JsxEmit={})),function(e){e[e["CarriageReturnLineFeed"]=0]="CarriageReturnLineFeed",e[e["LineFeed"]=1]="LineFeed"}(e.NewLineKind||(e.NewLineKind={})),function(e){e[e["Unknown"]=0]="Unknown",e[e["JS"]=1]="JS",e[e["JSX"]=2]="JSX",e[e["TS"]=3]="TS",e[e["TSX"]=4]="TSX",e[e["External"]=5]="External",e[e["JSON"]=6]="JSON",e[e["Deferred"]=7]="Deferred"}(e.ScriptKind||(e.ScriptKind={})),function(e){e[e["ES3"]=0]="ES3",e[e["ES5"]=1]="ES5",e[e["ES2015"]=2]="ES2015",e[e["ES2016"]=3]="ES2016",e[e["ES2017"]=4]="ES2017",e[e["ES2018"]=5]="ES2018",e[e["ESNext"]=6]="ESNext",e[e["JSON"]=100]="JSON",e[e["Latest"]=6]="Latest"}(e.ScriptTarget||(e.ScriptTarget={})),function(e){e[e["Standard"]=0]="Standard",e[e["JSX"]=1]="JSX"}(e.LanguageVariant||(e.LanguageVariant={})),function(e){e[e["None"]=0]="None",e[e["Recursive"]=1]="Recursive"}(e.WatchDirectoryFlags||(e.WatchDirectoryFlags={})),function(e){e[e["nullCharacter"]=0]="nullCharacter",e[e["maxAsciiCharacter"]=127]="maxAsciiCharacter",e[e["lineFeed"]=10]="lineFeed",e[e["carriageReturn"]=13]="carriageReturn",e[e["lineSeparator"]=8232]="lineSeparator",e[e["paragraphSeparator"]=8233]="paragraphSeparator",e[e["nextLine"]=133]="nextLine",e[e["space"]=32]="space",e[e["nonBreakingSpace"]=160]="nonBreakingSpace",e[e["enQuad"]=8192]="enQuad",e[e["emQuad"]=8193]="emQuad",e[e["enSpace"]=8194]="enSpace",e[e["emSpace"]=8195]="emSpace",e[e["threePerEmSpace"]=8196]="threePerEmSpace",e[e["fourPerEmSpace"]=8197]="fourPerEmSpace",e[e["sixPerEmSpace"]=8198]="sixPerEmSpace",e[e["figureSpace"]=8199]="figureSpace",e[e["punctuationSpace"]=8200]="punctuationSpace",e[e["thinSpace"]=8201]="thinSpace",e[e["hairSpace"]=8202]="hairSpace",e[e["zeroWidthSpace"]=8203]="zeroWidthSpace",e[e["narrowNoBreakSpace"]=8239]="narrowNoBreakSpace",e[e["ideographicSpace"]=12288]="ideographicSpace",e[e["mathematicalSpace"]=8287]="mathematicalSpace",e[e["ogham"]=5760]="ogham",e[e["_"]=95]="_",e[e["$"]=36]="$",e[e["_0"]=48]="_0",e[e["_1"]=49]="_1",e[e["_2"]=50]="_2",e[e["_3"]=51]="_3",e[e["_4"]=52]="_4",e[e["_5"]=53]="_5",e[e["_6"]=54]="_6",e[e["_7"]=55]="_7",e[e["_8"]=56]="_8",e[e["_9"]=57]="_9",e[e["a"]=97]="a",e[e["b"]=98]="b",e[e["c"]=99]="c",e[e["d"]=100]="d",e[e["e"]=101]="e",e[e["f"]=102]="f",e[e["g"]=103]="g",e[e["h"]=104]="h",e[e["i"]=105]="i",e[e["j"]=106]="j",e[e["k"]=107]="k",e[e["l"]=108]="l",e[e["m"]=109]="m",e[e["n"]=110]="n",e[e["o"]=111]="o",e[e["p"]=112]="p",e[e["q"]=113]="q",e[e["r"]=114]="r",e[e["s"]=115]="s",e[e["t"]=116]="t",e[e["u"]=117]="u",e[e["v"]=118]="v",e[e["w"]=119]="w",e[e["x"]=120]="x",e[e["y"]=121]="y",e[e["z"]=122]="z",e[e["A"]=65]="A",e[e["B"]=66]="B",e[e["C"]=67]="C",e[e["D"]=68]="D",e[e["E"]=69]="E",e[e["F"]=70]="F",e[e["G"]=71]="G",e[e["H"]=72]="H",e[e["I"]=73]="I",e[e["J"]=74]="J",e[e["K"]=75]="K",e[e["L"]=76]="L",e[e["M"]=77]="M",e[e["N"]=78]="N",e[e["O"]=79]="O",e[e["P"]=80]="P",e[e["Q"]=81]="Q",e[e["R"]=82]="R",e[e["S"]=83]="S",e[e["T"]=84]="T",e[e["U"]=85]="U",e[e["V"]=86]="V",e[e["W"]=87]="W",e[e["X"]=88]="X",e[e["Y"]=89]="Y",e[e["Z"]=90]="Z",e[e["ampersand"]=38]="ampersand",e[e["asterisk"]=42]="asterisk",e[e["at"]=64]="at",e[e["backslash"]=92]="backslash",e[e["backtick"]=96]="backtick",e[e["bar"]=124]="bar",e[e["caret"]=94]="caret",e[e["closeBrace"]=125]="closeBrace",e[e["closeBracket"]=93]="closeBracket",e[e["closeParen"]=41]="closeParen",e[e["colon"]=58]="colon",e[e["comma"]=44]="comma",e[e["dot"]=46]="dot",e[e["doubleQuote"]=34]="doubleQuote",e[e["equals"]=61]="equals",e[e["exclamation"]=33]="exclamation",e[e["greaterThan"]=62]="greaterThan",e[e["hash"]=35]="hash",e[e["lessThan"]=60]="lessThan",e[e["minus"]=45]="minus",e[e["openBrace"]=123]="openBrace",e[e["openBracket"]=91]="openBracket",e[e["openParen"]=40]="openParen",e[e["percent"]=37]="percent",e[e["plus"]=43]="plus",e[e["question"]=63]="question",e[e["semicolon"]=59]="semicolon",e[e["singleQuote"]=39]="singleQuote",e[e["slash"]=47]="slash",e[e["tilde"]=126]="tilde",e[e["backspace"]=8]="backspace",e[e["formFeed"]=12]="formFeed",e[e["byteOrderMark"]=65279]="byteOrderMark",e[e["tab"]=9]="tab",e[e["verticalTab"]=11]="verticalTab"}(e.CharacterCodes||(e.CharacterCodes={})),function(e){e["Ts"]=".ts",e["Tsx"]=".tsx",e["Dts"]=".d.ts",e["Js"]=".js",e["Jsx"]=".jsx",e["Json"]=".json"}(e.Extension||(e.Extension={})),function(e){e[e["None"]=0]="None",e[e["TypeScript"]=1]="TypeScript",e[e["ContainsTypeScript"]=2]="ContainsTypeScript",e[e["ContainsJsx"]=4]="ContainsJsx",e[e["ContainsESNext"]=8]="ContainsESNext",e[e["ContainsES2017"]=16]="ContainsES2017",e[e["ContainsES2016"]=32]="ContainsES2016",e[e["ES2015"]=64]="ES2015",e[e["ContainsES2015"]=128]="ContainsES2015",e[e["Generator"]=256]="Generator",e[e["ContainsGenerator"]=512]="ContainsGenerator",e[e["DestructuringAssignment"]=1024]="DestructuringAssignment",e[e["ContainsDestructuringAssignment"]=2048]="ContainsDestructuringAssignment",e[e["ContainsTypeScriptClassSyntax"]=4096]="ContainsTypeScriptClassSyntax",e[e["ContainsLexicalThis"]=8192]="ContainsLexicalThis",e[e["ContainsCapturedLexicalThis"]=16384]="ContainsCapturedLexicalThis",e[e["ContainsLexicalThisInComputedPropertyName"]=32768]="ContainsLexicalThisInComputedPropertyName",e[e["ContainsDefaultValueAssignments"]=65536]="ContainsDefaultValueAssignments",e[e["ContainsRestOrSpread"]=131072]="ContainsRestOrSpread",e[e["ContainsObjectRestOrSpread"]=262144]="ContainsObjectRestOrSpread",e[e["ContainsComputedPropertyName"]=524288]="ContainsComputedPropertyName",e[e["ContainsBlockScopedBinding"]=1048576]="ContainsBlockScopedBinding",e[e["ContainsBindingPattern"]=2097152]="ContainsBindingPattern",e[e["ContainsYield"]=4194304]="ContainsYield",e[e["ContainsHoistedDeclarationOrCompletion"]=8388608]="ContainsHoistedDeclarationOrCompletion",e[e["ContainsDynamicImport"]=16777216]="ContainsDynamicImport",e[e["Super"]=33554432]="Super",e[e["ContainsSuper"]=67108864]="ContainsSuper",e[e["HasComputedFlags"]=536870912]="HasComputedFlags",e[e["AssertTypeScript"]=3]="AssertTypeScript",e[e["AssertJsx"]=4]="AssertJsx",e[e["AssertESNext"]=8]="AssertESNext",e[e["AssertES2017"]=16]="AssertES2017",e[e["AssertES2016"]=32]="AssertES2016",e[e["AssertES2015"]=192]="AssertES2015",e[e["AssertGenerator"]=768]="AssertGenerator",e[e["AssertDestructuringAssignment"]=3072]="AssertDestructuringAssignment",e[e["OuterExpressionExcludes"]=536872257]="OuterExpressionExcludes",e[e["PropertyAccessExcludes"]=570426689]="PropertyAccessExcludes",e[e["NodeExcludes"]=637535553]="NodeExcludes",e[e["ArrowFunctionExcludes"]=653604161]="ArrowFunctionExcludes",e[e["FunctionExcludes"]=653620545]="FunctionExcludes",e[e["ConstructorExcludes"]=653616449]="ConstructorExcludes",e[e["MethodOrAccessorExcludes"]=653616449]="MethodOrAccessorExcludes",e[e["ClassExcludes"]=638121281]="ClassExcludes",e[e["ModuleExcludes"]=647001409]="ModuleExcludes",e[e["TypeExcludes"]=-3]="TypeExcludes",e[e["ObjectLiteralExcludes"]=638358849]="ObjectLiteralExcludes",e[e["ArrayLiteralOrCallOrNewExcludes"]=637666625]="ArrayLiteralOrCallOrNewExcludes",e[e["VariableDeclarationListExcludes"]=639894849]="VariableDeclarationListExcludes",e[e["ParameterExcludes"]=637535553]="ParameterExcludes",e[e["CatchClauseExcludes"]=637797697]="CatchClauseExcludes",e[e["BindingPatternExcludes"]=637666625]="BindingPatternExcludes",e[e["ES2015FunctionSyntaxMask"]=81920]="ES2015FunctionSyntaxMask"}(e.TransformFlags||(e.TransformFlags={})),function(e){e[e["None"]=0]="None",e[e["SingleLine"]=1]="SingleLine",e[e["AdviseOnEmitNode"]=2]="AdviseOnEmitNode",e[e["NoSubstitution"]=4]="NoSubstitution",e[e["CapturesThis"]=8]="CapturesThis",e[e["NoLeadingSourceMap"]=16]="NoLeadingSourceMap",e[e["NoTrailingSourceMap"]=32]="NoTrailingSourceMap",e[e["NoSourceMap"]=48]="NoSourceMap",e[e["NoNestedSourceMaps"]=64]="NoNestedSourceMaps",e[e["NoTokenLeadingSourceMaps"]=128]="NoTokenLeadingSourceMaps",e[e["NoTokenTrailingSourceMaps"]=256]="NoTokenTrailingSourceMaps",e[e["NoTokenSourceMaps"]=384]="NoTokenSourceMaps",e[e["NoLeadingComments"]=512]="NoLeadingComments",e[e["NoTrailingComments"]=1024]="NoTrailingComments",e[e["NoComments"]=1536]="NoComments",e[e["NoNestedComments"]=2048]="NoNestedComments",e[e["HelperName"]=4096]="HelperName",e[e["ExportName"]=8192]="ExportName",e[e["LocalName"]=16384]="LocalName",e[e["InternalName"]=32768]="InternalName",e[e["Indented"]=65536]="Indented",e[e["NoIndentation"]=131072]="NoIndentation",e[e["AsyncFunctionBody"]=262144]="AsyncFunctionBody",e[e["ReuseTempVariableScope"]=524288]="ReuseTempVariableScope",e[e["CustomPrologue"]=1048576]="CustomPrologue",e[e["NoHoisting"]=2097152]="NoHoisting",e[e["HasEndOfDeclarationMarker"]=4194304]="HasEndOfDeclarationMarker",e[e["Iterator"]=8388608]="Iterator",e[e["NoAsciiEscaping"]=16777216]="NoAsciiEscaping",e[e["TypeScriptClassWrapper"]=33554432]="TypeScriptClassWrapper",e[e["NeverApplyImportHelper"]=67108864]="NeverApplyImportHelper"}(e.EmitFlags||(e.EmitFlags={})),function(e){e[e["Extends"]=1]="Extends",e[e["Assign"]=2]="Assign",e[e["Rest"]=4]="Rest",e[e["Decorate"]=8]="Decorate",e[e["Metadata"]=16]="Metadata",e[e["Param"]=32]="Param",e[e["Awaiter"]=64]="Awaiter",e[e["Generator"]=128]="Generator",e[e["Values"]=256]="Values",e[e["Read"]=512]="Read",e[e["Spread"]=1024]="Spread",e[e["Await"]=2048]="Await",e[e["AsyncGenerator"]=4096]="AsyncGenerator",e[e["AsyncDelegator"]=8192]="AsyncDelegator",e[e["AsyncValues"]=16384]="AsyncValues",e[e["ExportStar"]=32768]="ExportStar",e[e["MakeTemplateObject"]=65536]="MakeTemplateObject",e[e["FirstEmitHelper"]=1]="FirstEmitHelper",e[e["LastEmitHelper"]=65536]="LastEmitHelper",e[e["ForOfIncludes"]=256]="ForOfIncludes",e[e["ForAwaitOfIncludes"]=16384]="ForAwaitOfIncludes",e[e["AsyncGeneratorIncludes"]=6144]="AsyncGeneratorIncludes",e[e["AsyncDelegatorIncludes"]=26624]="AsyncDelegatorIncludes",e[e["SpreadIncludes"]=1536]="SpreadIncludes"}(e.ExternalEmitHelpers||(e.ExternalEmitHelpers={})),function(e){e[e["SourceFile"]=0]="SourceFile",e[e["Expression"]=1]="Expression",e[e["IdentifierName"]=2]="IdentifierName",e[e["MappedTypeParameter"]=3]="MappedTypeParameter",e[e["Unspecified"]=4]="Unspecified",e[e["EmbeddedStatement"]=5]="EmbeddedStatement"}(e.EmitHint||(e.EmitHint={})),function(e){e[e["None"]=0]="None",e[e["SingleLine"]=0]="SingleLine",e[e["MultiLine"]=1]="MultiLine",e[e["PreserveLines"]=2]="PreserveLines",e[e["LinesMask"]=3]="LinesMask",e[e["NotDelimited"]=0]="NotDelimited",e[e["BarDelimited"]=4]="BarDelimited",e[e["AmpersandDelimited"]=8]="AmpersandDelimited",e[e["CommaDelimited"]=16]="CommaDelimited",e[e["AsteriskDelimited"]=32]="AsteriskDelimited",e[e["DelimitersMask"]=60]="DelimitersMask",e[e["AllowTrailingComma"]=64]="AllowTrailingComma",e[e["Indented"]=128]="Indented",e[e["SpaceBetweenBraces"]=256]="SpaceBetweenBraces",e[e["SpaceBetweenSiblings"]=512]="SpaceBetweenSiblings",e[e["Braces"]=1024]="Braces",e[e["Parenthesis"]=2048]="Parenthesis",e[e["AngleBrackets"]=4096]="AngleBrackets",e[e["SquareBrackets"]=8192]="SquareBrackets",e[e["BracketsMask"]=15360]="BracketsMask",e[e["OptionalIfUndefined"]=16384]="OptionalIfUndefined",e[e["OptionalIfEmpty"]=32768]="OptionalIfEmpty",e[e["Optional"]=49152]="Optional",e[e["PreferNewLine"]=65536]="PreferNewLine",e[e["NoTrailingNewLine"]=131072]="NoTrailingNewLine",e[e["NoInterveningComments"]=262144]="NoInterveningComments",e[e["NoSpaceIfEmpty"]=524288]="NoSpaceIfEmpty",e[e["SingleElement"]=1048576]="SingleElement",e[e["Modifiers"]=262656]="Modifiers",e[e["HeritageClauses"]=512]="HeritageClauses",e[e["SingleLineTypeLiteralMembers"]=768]="SingleLineTypeLiteralMembers",e[e["MultiLineTypeLiteralMembers"]=32897]="MultiLineTypeLiteralMembers",e[e["TupleTypeElements"]=528]="TupleTypeElements",e[e["UnionTypeConstituents"]=516]="UnionTypeConstituents",e[e["IntersectionTypeConstituents"]=520]="IntersectionTypeConstituents",e[e["ObjectBindingPatternElements"]=525136]="ObjectBindingPatternElements",e[e["ArrayBindingPatternElements"]=524880]="ArrayBindingPatternElements",e[e["ObjectLiteralExpressionProperties"]=526226]="ObjectLiteralExpressionProperties",e[e["ArrayLiteralExpressionElements"]=8914]="ArrayLiteralExpressionElements",e[e["CommaListElements"]=528]="CommaListElements",e[e["CallExpressionArguments"]=2576]="CallExpressionArguments",e[e["NewExpressionArguments"]=18960]="NewExpressionArguments",e[e["TemplateExpressionSpans"]=262144]="TemplateExpressionSpans",e[e["SingleLineBlockStatements"]=768]="SingleLineBlockStatements",e[e["MultiLineBlockStatements"]=129]="MultiLineBlockStatements",e[e["VariableDeclarationList"]=528]="VariableDeclarationList",e[e["SingleLineFunctionBodyStatements"]=768]="SingleLineFunctionBodyStatements",e[e["MultiLineFunctionBodyStatements"]=1]="MultiLineFunctionBodyStatements",e[e["ClassHeritageClauses"]=0]="ClassHeritageClauses",e[e["ClassMembers"]=129]="ClassMembers",e[e["InterfaceMembers"]=129]="InterfaceMembers",e[e["EnumMembers"]=145]="EnumMembers",e[e["CaseBlockClauses"]=129]="CaseBlockClauses",e[e["NamedImportsOrExportsElements"]=525136]="NamedImportsOrExportsElements",e[e["JsxElementOrFragmentChildren"]=262144]="JsxElementOrFragmentChildren",e[e["JsxElementAttributes"]=262656]="JsxElementAttributes",e[e["CaseOrDefaultClauseStatements"]=163969]="CaseOrDefaultClauseStatements",e[e["HeritageClauseTypes"]=528]="HeritageClauseTypes",e[e["SourceFileStatements"]=131073]="SourceFileStatements",e[e["Decorators"]=49153]="Decorators",e[e["TypeArguments"]=53776]="TypeArguments",e[e["TypeParameters"]=53776]="TypeParameters",e[e["Parameters"]=2576]="Parameters",e[e["IndexSignatureParameters"]=8848]="IndexSignatureParameters",e[e["JSDocComment"]=33]="JSDocComment"}(e.ListFormat||(e.ListFormat={})),function(e){e[e["None"]=0]="None",e[e["TripleSlashXML"]=1]="TripleSlashXML",e[e["SingleLine"]=2]="SingleLine",e[e["MultiLine"]=4]="MultiLine",e[e["All"]=7]="All",e[e["Default"]=7]="Default"}(e.PragmaKindFlags||(e.PragmaKindFlags={})),e.commentPragmas=a({reference:{args:[{name:"types",optional:!0,captureSpan:!0},{name:"lib",optional:!0,captureSpan:!0},{name:"path",optional:!0,captureSpan:!0},{name:"no-default-lib",optional:!0}],kind:1},"amd-dependency":{args:[{name:"path"},{name:"name",optional:!0}],kind:1},"amd-module":{args:[{name:"name"}],kind:1},"ts-check":{kind:2},"ts-nocheck":{kind:2},jsx:{args:[{name:"factory"}],kind:4}})}(u||(u={})),function(n){function t(e){var n=e.split("").map(function(e){return e.charCodeAt(0)});return""+n.reduce(function(e,n){return(e<<5)+e+n},5381)}function r(){Error.stackTraceLimit<100&&(Error.stackTraceLimit=100)}var a,i;function s(e){var n;return n={},n[i.Low]=e.Low,n[i.Medium]=e.Medium,n[i.High]=e.High,n}n.generateDjb2Hash=t,n.setStackTraceLimit=r,function(e){e[e["Created"]=0]="Created",e[e["Changed"]=1]="Changed",e[e["Deleted"]=2]="Deleted"}(a=n.FileWatcherEventKind||(n.FileWatcherEventKind={})),function(e){e[e["High"]=2e3]="High",e[e["Medium"]=500]="Medium",e[e["Low"]=250]="Low"}(i=n.PollingInterval||(n.PollingInterval={})),n.missingFileModifiedTime=new Date(0);var c={Low:32,Medium:64,High:256},l=s(c);function u(e){if(e.getEnvironmentVariable){var t=u("TSC_WATCH_POLLINGINTERVAL",i);l=d("TSC_WATCH_POLLINGCHUNKSIZE",c)||l,n.unchangedPollThresholds=d("TSC_WATCH_UNCHANGEDPOLLTHRESHOLDS",c)||n.unchangedPollThresholds}function r(n,t){return e.getEnvironmentVariable(n+"_"+t.toUpperCase())}function a(e){var n;return t("Low"),t("Medium"),t("High"),n;function t(t){var a=r(e,t);a&&((n||(n={}))[t]=Number(a))}}function u(e,n){var t=a(e);return!!t&&(r("Low"),r("Medium"),r("High"),!0);function r(e){n[e]=t[e]||n[e]}}function d(e,n){var r=a(e);return(t||r)&&s(r?o({},n,r):n)}}function d(e){var t=[],r=[],a=u(i.Low),o=u(i.Medium),s=u(i.High);return c;function c(e,r,a){var i={fileName:e,callback:r,unchangedPolls:0,mtime:b(e)};return t.push(i),g(i,a),{close:function(){i.isClosed=!0,n.unorderedRemoveItem(t,i)}}}function u(e){var n=[];return n.pollingInterval=e,n.pollIndex=0,n.pollScheduled=!1,n}function d(e){e.pollIndex=f(e,e.pollingInterval,e.pollIndex,l[e.pollingInterval]),e.length?h(e.pollingInterval):(n.Debug.assert(0===e.pollIndex),e.pollScheduled=!1)}function m(e){f(r,i.Low,0,r.length),d(e),!e.pollScheduled&&r.length&&h(i.Low)}function f(e,t,a,o){for(var s=e.length,c=a,l=0;l<o&&s>0;m(),s--){var u=e[a];if(u)if(u.isClosed)e[a]=void 0;else{l++;var d=p(u,b(u.fileName));u.isClosed?e[a]=void 0:d?(u.unchangedPolls=0,e!==r&&(e[a]=void 0,v(u))):u.unchangedPolls!==n.unchangedPollThresholds[t]?u.unchangedPolls++:e===r?(u.unchangedPolls=1,e[a]=void 0,g(u,i.Low)):t!==i.High&&(u.unchangedPolls++,e[a]=void 0,g(u,t===i.Low?i.Medium:i.High)),e[a]&&(c<a&&(e[c]=u,e[a]=void 0),c++)}}return a;function m(){a++,a===e.length&&(c<a&&(e.length=c),a=0,c=0)}}function _(e){switch(e){case i.Low:return a;case i.Medium:return o;case i.High:return s}}function g(e,n){_(n).push(e),y(n)}function v(e){r.push(e),y(i.Low)}function y(e){_(e).pollScheduled||h(e)}function h(n){_(n).pollScheduled=e.setTimeout(n===i.Low?m:d,n,_(n))}function b(t){return e.getModifiedTime(t)||n.missingFileModifiedTime}}function p(e,n){var t=e.mtime.getTime(),r=n.getTime();return t!==r&&(e.mtime=n,e.callback(e.fileName,m(t,r)),!0)}function m(e,n){return 0===e?a.Created:0===n?a.Deleted:a.Changed}function f(e){var t=n.createMap(),r=n.createMultiMap(),a=n.getStringComparer(!e.useCaseSensitiveFileNames),i=n.createGetCanonicalFileName(e.useCaseSensitiveFileNames);return o;function o(a,o){var c=i(a),l=t.get(c);return l?l.refCount++:(l={watcher:e.watchDirectory(a,function(e){r.forEach(function(t,r){(r===c||n.startsWith(c,r)&&c[r.length]===n.directorySeparator)&&t.forEach(function(n){return n(e)})}),s(a,c)}),refCount:1,childWatches:n.emptyArray},t.set(c,l),s(a,c)),o&&r.add(c,o),{dirName:a,close:function(){var e=n.Debug.assertDefined(t.get(c));o&&r.remove(c,o),e.refCount--,e.refCount||(t.delete(c),n.closeFileWatcherOf(e),e.childWatches.forEach(n.closeFileWatcher))}}}function s(e,n){var r=t.get(n);r&&(r.childWatches=c(e,r.childWatches))}function c(t,r){var i;return n.enumerateInsertsAndDeletes(e.directoryExists(t)?n.mapDefined(e.getAccessibleSortedChildDirectories(t),function(r){var i=n.getNormalizedAbsolutePath(r,t);return 0===a(i,n.normalizePath(e.realpath(i)))?i:void 0}):n.emptyArray,r,function(e,n){return a(e,n.dirName)},s,n.closeFileWatcher,c),i||n.emptyArray;function s(e){var n=o(e);c(n)}function c(e){(i||(i=[])).push(e)}}}function _(){if("undefined"!==typeof e){var n=e.version;if(n){var t=n.indexOf(".");if(-1!==t)return parseInt(n.substring(1,t))}}}n.unchangedPollThresholds=s(c),n.setCustomPollingValues=u,n.createDynamicPriorityPollingWatchFile=d,n.onWatchedFileStat=p,n.getFileWatcherEventKind=m,n.createRecursiveDirectoryWatcher=f,n.getNodeMajorVersion=_,n.sys=void 0,n.sys&&n.sys.getEnvironmentVariable&&(u(n.sys),n.Debug.currentAssertionLevel=/^development$/i.test(n.sys.getEnvironmentVariable("NODE_ENV"))?1:0),n.sys&&n.sys.debugMode&&(n.Debug.isDebugging=!0)}(u||(u={})),function(e){function n(e,n,t,r,a){return{code:e,category:n,key:t,message:r,reportsUnnecessary:a}}e.Diagnostics={Unterminated_string_literal:n(1002,e.DiagnosticCategory.Error,"Unterminated_string_literal_1002","Unterminated string literal."),Identifier_expected:n(1003,e.DiagnosticCategory.Error,"Identifier_expected_1003","Identifier expected."),_0_expected:n(1005,e.DiagnosticCategory.Error,"_0_expected_1005","'{0}' expected."),A_file_cannot_have_a_reference_to_itself:n(1006,e.DiagnosticCategory.Error,"A_file_cannot_have_a_reference_to_itself_1006","A file cannot have a reference to itself."),Trailing_comma_not_allowed:n(1009,e.DiagnosticCategory.Error,"Trailing_comma_not_allowed_1009","Trailing comma not allowed."),Asterisk_Slash_expected:n(1010,e.DiagnosticCategory.Error,"Asterisk_Slash_expected_1010","'*/' expected."),An_element_access_expression_should_take_an_argument:n(1011,e.DiagnosticCategory.Error,"An_element_access_expression_should_take_an_argument_1011","An element access expression should take an argument."),Unexpected_token:n(1012,e.DiagnosticCategory.Error,"Unexpected_token_1012","Unexpected token."),A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma:n(1013,e.DiagnosticCategory.Error,"A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma_1013","A rest parameter or binding pattern may not have a trailing comma."),A_rest_parameter_must_be_last_in_a_parameter_list:n(1014,e.DiagnosticCategory.Error,"A_rest_parameter_must_be_last_in_a_parameter_list_1014","A rest parameter must be last in a parameter list."),Parameter_cannot_have_question_mark_and_initializer:n(1015,e.DiagnosticCategory.Error,"Parameter_cannot_have_question_mark_and_initializer_1015","Parameter cannot have question mark and initializer."),A_required_parameter_cannot_follow_an_optional_parameter:n(1016,e.DiagnosticCategory.Error,"A_required_parameter_cannot_follow_an_optional_parameter_1016","A required parameter cannot follow an optional parameter."),An_index_signature_cannot_have_a_rest_parameter:n(1017,e.DiagnosticCategory.Error,"An_index_signature_cannot_have_a_rest_parameter_1017","An index signature cannot have a rest parameter."),An_index_signature_parameter_cannot_have_an_accessibility_modifier:n(1018,e.DiagnosticCategory.Error,"An_index_signature_parameter_cannot_have_an_accessibility_modifier_1018","An index signature parameter cannot have an accessibility modifier."),An_index_signature_parameter_cannot_have_a_question_mark:n(1019,e.DiagnosticCategory.Error,"An_index_signature_parameter_cannot_have_a_question_mark_1019","An index signature parameter cannot have a question mark."),An_index_signature_parameter_cannot_have_an_initializer:n(1020,e.DiagnosticCategory.Error,"An_index_signature_parameter_cannot_have_an_initializer_1020","An index signature parameter cannot have an initializer."),An_index_signature_must_have_a_type_annotation:n(1021,e.DiagnosticCategory.Error,"An_index_signature_must_have_a_type_annotation_1021","An index signature must have a type annotation."),An_index_signature_parameter_must_have_a_type_annotation:n(1022,e.DiagnosticCategory.Error,"An_index_signature_parameter_must_have_a_type_annotation_1022","An index signature parameter must have a type annotation."),An_index_signature_parameter_type_must_be_string_or_number:n(1023,e.DiagnosticCategory.Error,"An_index_signature_parameter_type_must_be_string_or_number_1023","An index signature parameter type must be 'string' or 'number'."),readonly_modifier_can_only_appear_on_a_property_declaration_or_index_signature:n(1024,e.DiagnosticCategory.Error,"readonly_modifier_can_only_appear_on_a_property_declaration_or_index_signature_1024","'readonly' modifier can only appear on a property declaration or index signature."),Accessibility_modifier_already_seen:n(1028,e.DiagnosticCategory.Error,"Accessibility_modifier_already_seen_1028","Accessibility modifier already seen."),_0_modifier_must_precede_1_modifier:n(1029,e.DiagnosticCategory.Error,"_0_modifier_must_precede_1_modifier_1029","'{0}' modifier must precede '{1}' modifier."),_0_modifier_already_seen:n(1030,e.DiagnosticCategory.Error,"_0_modifier_already_seen_1030","'{0}' modifier already seen."),_0_modifier_cannot_appear_on_a_class_element:n(1031,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_a_class_element_1031","'{0}' modifier cannot appear on a class element."),super_must_be_followed_by_an_argument_list_or_member_access:n(1034,e.DiagnosticCategory.Error,"super_must_be_followed_by_an_argument_list_or_member_access_1034","'super' must be followed by an argument list or member access."),Only_ambient_modules_can_use_quoted_names:n(1035,e.DiagnosticCategory.Error,"Only_ambient_modules_can_use_quoted_names_1035","Only ambient modules can use quoted names."),Statements_are_not_allowed_in_ambient_contexts:n(1036,e.DiagnosticCategory.Error,"Statements_are_not_allowed_in_ambient_contexts_1036","Statements are not allowed in ambient contexts."),A_declare_modifier_cannot_be_used_in_an_already_ambient_context:n(1038,e.DiagnosticCategory.Error,"A_declare_modifier_cannot_be_used_in_an_already_ambient_context_1038","A 'declare' modifier cannot be used in an already ambient context."),Initializers_are_not_allowed_in_ambient_contexts:n(1039,e.DiagnosticCategory.Error,"Initializers_are_not_allowed_in_ambient_contexts_1039","Initializers are not allowed in ambient contexts."),_0_modifier_cannot_be_used_in_an_ambient_context:n(1040,e.DiagnosticCategory.Error,"_0_modifier_cannot_be_used_in_an_ambient_context_1040","'{0}' modifier cannot be used in an ambient context."),_0_modifier_cannot_be_used_with_a_class_declaration:n(1041,e.DiagnosticCategory.Error,"_0_modifier_cannot_be_used_with_a_class_declaration_1041","'{0}' modifier cannot be used with a class declaration."),_0_modifier_cannot_be_used_here:n(1042,e.DiagnosticCategory.Error,"_0_modifier_cannot_be_used_here_1042","'{0}' modifier cannot be used here."),_0_modifier_cannot_appear_on_a_data_property:n(1043,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_a_data_property_1043","'{0}' modifier cannot appear on a data property."),_0_modifier_cannot_appear_on_a_module_or_namespace_element:n(1044,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_a_module_or_namespace_element_1044","'{0}' modifier cannot appear on a module or namespace element."),A_0_modifier_cannot_be_used_with_an_interface_declaration:n(1045,e.DiagnosticCategory.Error,"A_0_modifier_cannot_be_used_with_an_interface_declaration_1045","A '{0}' modifier cannot be used with an interface declaration."),A_declare_modifier_is_required_for_a_top_level_declaration_in_a_d_ts_file:n(1046,e.DiagnosticCategory.Error,"A_declare_modifier_is_required_for_a_top_level_declaration_in_a_d_ts_file_1046","A 'declare' modifier is required for a top level declaration in a .d.ts file."),A_rest_parameter_cannot_be_optional:n(1047,e.DiagnosticCategory.Error,"A_rest_parameter_cannot_be_optional_1047","A rest parameter cannot be optional."),A_rest_parameter_cannot_have_an_initializer:n(1048,e.DiagnosticCategory.Error,"A_rest_parameter_cannot_have_an_initializer_1048","A rest parameter cannot have an initializer."),A_set_accessor_must_have_exactly_one_parameter:n(1049,e.DiagnosticCategory.Error,"A_set_accessor_must_have_exactly_one_parameter_1049","A 'set' accessor must have exactly one parameter."),A_set_accessor_cannot_have_an_optional_parameter:n(1051,e.DiagnosticCategory.Error,"A_set_accessor_cannot_have_an_optional_parameter_1051","A 'set' accessor cannot have an optional parameter."),A_set_accessor_parameter_cannot_have_an_initializer:n(1052,e.DiagnosticCategory.Error,"A_set_accessor_parameter_cannot_have_an_initializer_1052","A 'set' accessor parameter cannot have an initializer."),A_set_accessor_cannot_have_rest_parameter:n(1053,e.DiagnosticCategory.Error,"A_set_accessor_cannot_have_rest_parameter_1053","A 'set' accessor cannot have rest parameter."),A_get_accessor_cannot_have_parameters:n(1054,e.DiagnosticCategory.Error,"A_get_accessor_cannot_have_parameters_1054","A 'get' accessor cannot have parameters."),Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value:n(1055,e.DiagnosticCategory.Error,"Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Prom_1055","Type '{0}' is not a valid async function return type in ES5/ES3 because it does not refer to a Promise-compatible constructor value."),Accessors_are_only_available_when_targeting_ECMAScript_5_and_higher:n(1056,e.DiagnosticCategory.Error,"Accessors_are_only_available_when_targeting_ECMAScript_5_and_higher_1056","Accessors are only available when targeting ECMAScript 5 and higher."),An_async_function_or_method_must_have_a_valid_awaitable_return_type:n(1057,e.DiagnosticCategory.Error,"An_async_function_or_method_must_have_a_valid_awaitable_return_type_1057","An async function or method must have a valid awaitable return type."),The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member:n(1058,e.DiagnosticCategory.Error,"The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_t_1058","The return type of an async function must either be a valid promise or must not contain a callable 'then' member."),A_promise_must_have_a_then_method:n(1059,e.DiagnosticCategory.Error,"A_promise_must_have_a_then_method_1059","A promise must have a 'then' method."),The_first_parameter_of_the_then_method_of_a_promise_must_be_a_callback:n(1060,e.DiagnosticCategory.Error,"The_first_parameter_of_the_then_method_of_a_promise_must_be_a_callback_1060","The first parameter of the 'then' method of a promise must be a callback."),Enum_member_must_have_initializer:n(1061,e.DiagnosticCategory.Error,"Enum_member_must_have_initializer_1061","Enum member must have initializer."),Type_is_referenced_directly_or_indirectly_in_the_fulfillment_callback_of_its_own_then_method:n(1062,e.DiagnosticCategory.Error,"Type_is_referenced_directly_or_indirectly_in_the_fulfillment_callback_of_its_own_then_method_1062","Type is referenced directly or indirectly in the fulfillment callback of its own 'then' method."),An_export_assignment_cannot_be_used_in_a_namespace:n(1063,e.DiagnosticCategory.Error,"An_export_assignment_cannot_be_used_in_a_namespace_1063","An export assignment cannot be used in a namespace."),The_return_type_of_an_async_function_or_method_must_be_the_global_Promise_T_type:n(1064,e.DiagnosticCategory.Error,"The_return_type_of_an_async_function_or_method_must_be_the_global_Promise_T_type_1064","The return type of an async function or method must be the global Promise<T> type."),In_ambient_enum_declarations_member_initializer_must_be_constant_expression:n(1066,e.DiagnosticCategory.Error,"In_ambient_enum_declarations_member_initializer_must_be_constant_expression_1066","In ambient enum declarations member initializer must be constant expression."),Unexpected_token_A_constructor_method_accessor_or_property_was_expected:n(1068,e.DiagnosticCategory.Error,"Unexpected_token_A_constructor_method_accessor_or_property_was_expected_1068","Unexpected token. A constructor, method, accessor, or property was expected."),Unexpected_token_A_type_parameter_name_was_expected_without_curly_braces:n(1069,e.DiagnosticCategory.Error,"Unexpected_token_A_type_parameter_name_was_expected_without_curly_braces_1069","Unexpected token. A type parameter name was expected without curly braces."),_0_modifier_cannot_appear_on_a_type_member:n(1070,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_a_type_member_1070","'{0}' modifier cannot appear on a type member."),_0_modifier_cannot_appear_on_an_index_signature:n(1071,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_an_index_signature_1071","'{0}' modifier cannot appear on an index signature."),A_0_modifier_cannot_be_used_with_an_import_declaration:n(1079,e.DiagnosticCategory.Error,"A_0_modifier_cannot_be_used_with_an_import_declaration_1079","A '{0}' modifier cannot be used with an import declaration."),Invalid_reference_directive_syntax:n(1084,e.DiagnosticCategory.Error,"Invalid_reference_directive_syntax_1084","Invalid 'reference' directive syntax."),Octal_literals_are_not_available_when_targeting_ECMAScript_5_and_higher_Use_the_syntax_0:n(1085,e.DiagnosticCategory.Error,"Octal_literals_are_not_available_when_targeting_ECMAScript_5_and_higher_Use_the_syntax_0_1085","Octal literals are not available when targeting ECMAScript 5 and higher. Use the syntax '{0}'."),An_accessor_cannot_be_declared_in_an_ambient_context:n(1086,e.DiagnosticCategory.Error,"An_accessor_cannot_be_declared_in_an_ambient_context_1086","An accessor cannot be declared in an ambient context."),_0_modifier_cannot_appear_on_a_constructor_declaration:n(1089,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_a_constructor_declaration_1089","'{0}' modifier cannot appear on a constructor declaration."),_0_modifier_cannot_appear_on_a_parameter:n(1090,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_a_parameter_1090","'{0}' modifier cannot appear on a parameter."),Only_a_single_variable_declaration_is_allowed_in_a_for_in_statement:n(1091,e.DiagnosticCategory.Error,"Only_a_single_variable_declaration_is_allowed_in_a_for_in_statement_1091","Only a single variable declaration is allowed in a 'for...in' statement."),Type_parameters_cannot_appear_on_a_constructor_declaration:n(1092,e.DiagnosticCategory.Error,"Type_parameters_cannot_appear_on_a_constructor_declaration_1092","Type parameters cannot appear on a constructor declaration."),Type_annotation_cannot_appear_on_a_constructor_declaration:n(1093,e.DiagnosticCategory.Error,"Type_annotation_cannot_appear_on_a_constructor_declaration_1093","Type annotation cannot appear on a constructor declaration."),An_accessor_cannot_have_type_parameters:n(1094,e.DiagnosticCategory.Error,"An_accessor_cannot_have_type_parameters_1094","An accessor cannot have type parameters."),A_set_accessor_cannot_have_a_return_type_annotation:n(1095,e.DiagnosticCategory.Error,"A_set_accessor_cannot_have_a_return_type_annotation_1095","A 'set' accessor cannot have a return type annotation."),An_index_signature_must_have_exactly_one_parameter:n(1096,e.DiagnosticCategory.Error,"An_index_signature_must_have_exactly_one_parameter_1096","An index signature must have exactly one parameter."),_0_list_cannot_be_empty:n(1097,e.DiagnosticCategory.Error,"_0_list_cannot_be_empty_1097","'{0}' list cannot be empty."),Type_parameter_list_cannot_be_empty:n(1098,e.DiagnosticCategory.Error,"Type_parameter_list_cannot_be_empty_1098","Type parameter list cannot be empty."),Type_argument_list_cannot_be_empty:n(1099,e.DiagnosticCategory.Error,"Type_argument_list_cannot_be_empty_1099","Type argument list cannot be empty."),Invalid_use_of_0_in_strict_mode:n(1100,e.DiagnosticCategory.Error,"Invalid_use_of_0_in_strict_mode_1100","Invalid use of '{0}' in strict mode."),with_statements_are_not_allowed_in_strict_mode:n(1101,e.DiagnosticCategory.Error,"with_statements_are_not_allowed_in_strict_mode_1101","'with' statements are not allowed in strict mode."),delete_cannot_be_called_on_an_identifier_in_strict_mode:n(1102,e.DiagnosticCategory.Error,"delete_cannot_be_called_on_an_identifier_in_strict_mode_1102","'delete' cannot be called on an identifier in strict mode."),A_for_await_of_statement_is_only_allowed_within_an_async_function_or_async_generator:n(1103,e.DiagnosticCategory.Error,"A_for_await_of_statement_is_only_allowed_within_an_async_function_or_async_generator_1103","A 'for-await-of' statement is only allowed within an async function or async generator."),A_continue_statement_can_only_be_used_within_an_enclosing_iteration_statement:n(1104,e.DiagnosticCategory.Error,"A_continue_statement_can_only_be_used_within_an_enclosing_iteration_statement_1104","A 'continue' statement can only be used within an enclosing iteration statement."),A_break_statement_can_only_be_used_within_an_enclosing_iteration_or_switch_statement:n(1105,e.DiagnosticCategory.Error,"A_break_statement_can_only_be_used_within_an_enclosing_iteration_or_switch_statement_1105","A 'break' statement can only be used within an enclosing iteration or switch statement."),Jump_target_cannot_cross_function_boundary:n(1107,e.DiagnosticCategory.Error,"Jump_target_cannot_cross_function_boundary_1107","Jump target cannot cross function boundary."),A_return_statement_can_only_be_used_within_a_function_body:n(1108,e.DiagnosticCategory.Error,"A_return_statement_can_only_be_used_within_a_function_body_1108","A 'return' statement can only be used within a function body."),Expression_expected:n(1109,e.DiagnosticCategory.Error,"Expression_expected_1109","Expression expected."),Type_expected:n(1110,e.DiagnosticCategory.Error,"Type_expected_1110","Type expected."),A_default_clause_cannot_appear_more_than_once_in_a_switch_statement:n(1113,e.DiagnosticCategory.Error,"A_default_clause_cannot_appear_more_than_once_in_a_switch_statement_1113","A 'default' clause cannot appear more than once in a 'switch' statement."),Duplicate_label_0:n(1114,e.DiagnosticCategory.Error,"Duplicate_label_0_1114","Duplicate label '{0}'."),A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement:n(1115,e.DiagnosticCategory.Error,"A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement_1115","A 'continue' statement can only jump to a label of an enclosing iteration statement."),A_break_statement_can_only_jump_to_a_label_of_an_enclosing_statement:n(1116,e.DiagnosticCategory.Error,"A_break_statement_can_only_jump_to_a_label_of_an_enclosing_statement_1116","A 'break' statement can only jump to a label of an enclosing statement."),An_object_literal_cannot_have_multiple_properties_with_the_same_name_in_strict_mode:n(1117,e.DiagnosticCategory.Error,"An_object_literal_cannot_have_multiple_properties_with_the_same_name_in_strict_mode_1117","An object literal cannot have multiple properties with the same name in strict mode."),An_object_literal_cannot_have_multiple_get_Slashset_accessors_with_the_same_name:n(1118,e.DiagnosticCategory.Error,"An_object_literal_cannot_have_multiple_get_Slashset_accessors_with_the_same_name_1118","An object literal cannot have multiple get/set accessors with the same name."),An_object_literal_cannot_have_property_and_accessor_with_the_same_name:n(1119,e.DiagnosticCategory.Error,"An_object_literal_cannot_have_property_and_accessor_with_the_same_name_1119","An object literal cannot have property and accessor with the same name."),An_export_assignment_cannot_have_modifiers:n(1120,e.DiagnosticCategory.Error,"An_export_assignment_cannot_have_modifiers_1120","An export assignment cannot have modifiers."),Octal_literals_are_not_allowed_in_strict_mode:n(1121,e.DiagnosticCategory.Error,"Octal_literals_are_not_allowed_in_strict_mode_1121","Octal literals are not allowed in strict mode."),Variable_declaration_list_cannot_be_empty:n(1123,e.DiagnosticCategory.Error,"Variable_declaration_list_cannot_be_empty_1123","Variable declaration list cannot be empty."),Digit_expected:n(1124,e.DiagnosticCategory.Error,"Digit_expected_1124","Digit expected."),Hexadecimal_digit_expected:n(1125,e.DiagnosticCategory.Error,"Hexadecimal_digit_expected_1125","Hexadecimal digit expected."),Unexpected_end_of_text:n(1126,e.DiagnosticCategory.Error,"Unexpected_end_of_text_1126","Unexpected end of text."),Invalid_character:n(1127,e.DiagnosticCategory.Error,"Invalid_character_1127","Invalid character."),Declaration_or_statement_expected:n(1128,e.DiagnosticCategory.Error,"Declaration_or_statement_expected_1128","Declaration or statement expected."),Statement_expected:n(1129,e.DiagnosticCategory.Error,"Statement_expected_1129","Statement expected."),case_or_default_expected:n(1130,e.DiagnosticCategory.Error,"case_or_default_expected_1130","'case' or 'default' expected."),Property_or_signature_expected:n(1131,e.DiagnosticCategory.Error,"Property_or_signature_expected_1131","Property or signature expected."),Enum_member_expected:n(1132,e.DiagnosticCategory.Error,"Enum_member_expected_1132","Enum member expected."),Variable_declaration_expected:n(1134,e.DiagnosticCategory.Error,"Variable_declaration_expected_1134","Variable declaration expected."),Argument_expression_expected:n(1135,e.DiagnosticCategory.Error,"Argument_expression_expected_1135","Argument expression expected."),Property_assignment_expected:n(1136,e.DiagnosticCategory.Error,"Property_assignment_expected_1136","Property assignment expected."),Expression_or_comma_expected:n(1137,e.DiagnosticCategory.Error,"Expression_or_comma_expected_1137","Expression or comma expected."),Parameter_declaration_expected:n(1138,e.DiagnosticCategory.Error,"Parameter_declaration_expected_1138","Parameter declaration expected."),Type_parameter_declaration_expected:n(1139,e.DiagnosticCategory.Error,"Type_parameter_declaration_expected_1139","Type parameter declaration expected."),Type_argument_expected:n(1140,e.DiagnosticCategory.Error,"Type_argument_expected_1140","Type argument expected."),String_literal_expected:n(1141,e.DiagnosticCategory.Error,"String_literal_expected_1141","String literal expected."),Line_break_not_permitted_here:n(1142,e.DiagnosticCategory.Error,"Line_break_not_permitted_here_1142","Line break not permitted here."),or_expected:n(1144,e.DiagnosticCategory.Error,"or_expected_1144","'{' or ';' expected."),Declaration_expected:n(1146,e.DiagnosticCategory.Error,"Declaration_expected_1146","Declaration expected."),Import_declarations_in_a_namespace_cannot_reference_a_module:n(1147,e.DiagnosticCategory.Error,"Import_declarations_in_a_namespace_cannot_reference_a_module_1147","Import declarations in a namespace cannot reference a module."),Cannot_use_imports_exports_or_module_augmentations_when_module_is_none:n(1148,e.DiagnosticCategory.Error,"Cannot_use_imports_exports_or_module_augmentations_when_module_is_none_1148","Cannot use imports, exports, or module augmentations when '--module' is 'none'."),File_name_0_differs_from_already_included_file_name_1_only_in_casing:n(1149,e.DiagnosticCategory.Error,"File_name_0_differs_from_already_included_file_name_1_only_in_casing_1149","File name '{0}' differs from already included file name '{1}' only in casing."),new_T_cannot_be_used_to_create_an_array_Use_new_Array_T_instead:n(1150,e.DiagnosticCategory.Error,"new_T_cannot_be_used_to_create_an_array_Use_new_Array_T_instead_1150","'new T[]' cannot be used to create an array. Use 'new Array<T>()' instead."),const_declarations_must_be_initialized:n(1155,e.DiagnosticCategory.Error,"const_declarations_must_be_initialized_1155","'const' declarations must be initialized."),const_declarations_can_only_be_declared_inside_a_block:n(1156,e.DiagnosticCategory.Error,"const_declarations_can_only_be_declared_inside_a_block_1156","'const' declarations can only be declared inside a block."),let_declarations_can_only_be_declared_inside_a_block:n(1157,e.DiagnosticCategory.Error,"let_declarations_can_only_be_declared_inside_a_block_1157","'let' declarations can only be declared inside a block."),Unterminated_template_literal:n(1160,e.DiagnosticCategory.Error,"Unterminated_template_literal_1160","Unterminated template literal."),Unterminated_regular_expression_literal:n(1161,e.DiagnosticCategory.Error,"Unterminated_regular_expression_literal_1161","Unterminated regular expression literal."),An_object_member_cannot_be_declared_optional:n(1162,e.DiagnosticCategory.Error,"An_object_member_cannot_be_declared_optional_1162","An object member cannot be declared optional."),A_yield_expression_is_only_allowed_in_a_generator_body:n(1163,e.DiagnosticCategory.Error,"A_yield_expression_is_only_allowed_in_a_generator_body_1163","A 'yield' expression is only allowed in a generator body."),Computed_property_names_are_not_allowed_in_enums:n(1164,e.DiagnosticCategory.Error,"Computed_property_names_are_not_allowed_in_enums_1164","Computed property names are not allowed in enums."),A_computed_property_name_in_an_ambient_context_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type:n(1165,e.DiagnosticCategory.Error,"A_computed_property_name_in_an_ambient_context_must_refer_to_an_expression_whose_type_is_a_literal_t_1165","A computed property name in an ambient context must refer to an expression whose type is a literal type or a 'unique symbol' type."),A_computed_property_name_in_a_class_property_declaration_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type:n(1166,e.DiagnosticCategory.Error,"A_computed_property_name_in_a_class_property_declaration_must_refer_to_an_expression_whose_type_is_a_1166","A computed property name in a class property declaration must refer to an expression whose type is a literal type or a 'unique symbol' type."),A_computed_property_name_in_a_method_overload_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type:n(1168,e.DiagnosticCategory.Error,"A_computed_property_name_in_a_method_overload_must_refer_to_an_expression_whose_type_is_a_literal_ty_1168","A computed property name in a method overload must refer to an expression whose type is a literal type or a 'unique symbol' type."),A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type:n(1169,e.DiagnosticCategory.Error,"A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_1169","A computed property name in an interface must refer to an expression whose type is a literal type or a 'unique symbol' type."),A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type:n(1170,e.DiagnosticCategory.Error,"A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type__1170","A computed property name in a type literal must refer to an expression whose type is a literal type or a 'unique symbol' type."),A_comma_expression_is_not_allowed_in_a_computed_property_name:n(1171,e.DiagnosticCategory.Error,"A_comma_expression_is_not_allowed_in_a_computed_property_name_1171","A comma expression is not allowed in a computed property name."),extends_clause_already_seen:n(1172,e.DiagnosticCategory.Error,"extends_clause_already_seen_1172","'extends' clause already seen."),extends_clause_must_precede_implements_clause:n(1173,e.DiagnosticCategory.Error,"extends_clause_must_precede_implements_clause_1173","'extends' clause must precede 'implements' clause."),Classes_can_only_extend_a_single_class:n(1174,e.DiagnosticCategory.Error,"Classes_can_only_extend_a_single_class_1174","Classes can only extend a single class."),implements_clause_already_seen:n(1175,e.DiagnosticCategory.Error,"implements_clause_already_seen_1175","'implements' clause already seen."),Interface_declaration_cannot_have_implements_clause:n(1176,e.DiagnosticCategory.Error,"Interface_declaration_cannot_have_implements_clause_1176","Interface declaration cannot have 'implements' clause."),Binary_digit_expected:n(1177,e.DiagnosticCategory.Error,"Binary_digit_expected_1177","Binary digit expected."),Octal_digit_expected:n(1178,e.DiagnosticCategory.Error,"Octal_digit_expected_1178","Octal digit expected."),Unexpected_token_expected:n(1179,e.DiagnosticCategory.Error,"Unexpected_token_expected_1179","Unexpected token. '{' expected."),Property_destructuring_pattern_expected:n(1180,e.DiagnosticCategory.Error,"Property_destructuring_pattern_expected_1180","Property destructuring pattern expected."),Array_element_destructuring_pattern_expected:n(1181,e.DiagnosticCategory.Error,"Array_element_destructuring_pattern_expected_1181","Array element destructuring pattern expected."),A_destructuring_declaration_must_have_an_initializer:n(1182,e.DiagnosticCategory.Error,"A_destructuring_declaration_must_have_an_initializer_1182","A destructuring declaration must have an initializer."),An_implementation_cannot_be_declared_in_ambient_contexts:n(1183,e.DiagnosticCategory.Error,"An_implementation_cannot_be_declared_in_ambient_contexts_1183","An implementation cannot be declared in ambient contexts."),Modifiers_cannot_appear_here:n(1184,e.DiagnosticCategory.Error,"Modifiers_cannot_appear_here_1184","Modifiers cannot appear here."),Merge_conflict_marker_encountered:n(1185,e.DiagnosticCategory.Error,"Merge_conflict_marker_encountered_1185","Merge conflict marker encountered."),A_rest_element_cannot_have_an_initializer:n(1186,e.DiagnosticCategory.Error,"A_rest_element_cannot_have_an_initializer_1186","A rest element cannot have an initializer."),A_parameter_property_may_not_be_declared_using_a_binding_pattern:n(1187,e.DiagnosticCategory.Error,"A_parameter_property_may_not_be_declared_using_a_binding_pattern_1187","A parameter property may not be declared using a binding pattern."),Only_a_single_variable_declaration_is_allowed_in_a_for_of_statement:n(1188,e.DiagnosticCategory.Error,"Only_a_single_variable_declaration_is_allowed_in_a_for_of_statement_1188","Only a single variable declaration is allowed in a 'for...of' statement."),The_variable_declaration_of_a_for_in_statement_cannot_have_an_initializer:n(1189,e.DiagnosticCategory.Error,"The_variable_declaration_of_a_for_in_statement_cannot_have_an_initializer_1189","The variable declaration of a 'for...in' statement cannot have an initializer."),The_variable_declaration_of_a_for_of_statement_cannot_have_an_initializer:n(1190,e.DiagnosticCategory.Error,"The_variable_declaration_of_a_for_of_statement_cannot_have_an_initializer_1190","The variable declaration of a 'for...of' statement cannot have an initializer."),An_import_declaration_cannot_have_modifiers:n(1191,e.DiagnosticCategory.Error,"An_import_declaration_cannot_have_modifiers_1191","An import declaration cannot have modifiers."),Module_0_has_no_default_export:n(1192,e.DiagnosticCategory.Error,"Module_0_has_no_default_export_1192","Module '{0}' has no default export."),An_export_declaration_cannot_have_modifiers:n(1193,e.DiagnosticCategory.Error,"An_export_declaration_cannot_have_modifiers_1193","An export declaration cannot have modifiers."),Export_declarations_are_not_permitted_in_a_namespace:n(1194,e.DiagnosticCategory.Error,"Export_declarations_are_not_permitted_in_a_namespace_1194","Export declarations are not permitted in a namespace."),Catch_clause_variable_cannot_have_a_type_annotation:n(1196,e.DiagnosticCategory.Error,"Catch_clause_variable_cannot_have_a_type_annotation_1196","Catch clause variable cannot have a type annotation."),Catch_clause_variable_cannot_have_an_initializer:n(1197,e.DiagnosticCategory.Error,"Catch_clause_variable_cannot_have_an_initializer_1197","Catch clause variable cannot have an initializer."),An_extended_Unicode_escape_value_must_be_between_0x0_and_0x10FFFF_inclusive:n(1198,e.DiagnosticCategory.Error,"An_extended_Unicode_escape_value_must_be_between_0x0_and_0x10FFFF_inclusive_1198","An extended Unicode escape value must be between 0x0 and 0x10FFFF inclusive."),Unterminated_Unicode_escape_sequence:n(1199,e.DiagnosticCategory.Error,"Unterminated_Unicode_escape_sequence_1199","Unterminated Unicode escape sequence."),Line_terminator_not_permitted_before_arrow:n(1200,e.DiagnosticCategory.Error,"Line_terminator_not_permitted_before_arrow_1200","Line terminator not permitted before arrow."),Import_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_import_Asterisk_as_ns_from_mod_import_a_from_mod_import_d_from_mod_or_another_module_format_instead:n(1202,e.DiagnosticCategory.Error,"Import_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_import_Asterisk_as_1202","Import assignment cannot be used when targeting ECMAScript modules. Consider using 'import * as ns from \"mod\"', 'import {a} from \"mod\"', 'import d from \"mod\"', or another module format instead."),Export_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_export_default_or_another_module_format_instead:n(1203,e.DiagnosticCategory.Error,"Export_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_export_default_or__1203","Export assignment cannot be used when targeting ECMAScript modules. Consider using 'export default' or another module format instead."),Cannot_re_export_a_type_when_the_isolatedModules_flag_is_provided:n(1205,e.DiagnosticCategory.Error,"Cannot_re_export_a_type_when_the_isolatedModules_flag_is_provided_1205","Cannot re-export a type when the '--isolatedModules' flag is provided."),Decorators_are_not_valid_here:n(1206,e.DiagnosticCategory.Error,"Decorators_are_not_valid_here_1206","Decorators are not valid here."),Decorators_cannot_be_applied_to_multiple_get_Slashset_accessors_of_the_same_name:n(1207,e.DiagnosticCategory.Error,"Decorators_cannot_be_applied_to_multiple_get_Slashset_accessors_of_the_same_name_1207","Decorators cannot be applied to multiple get/set accessors of the same name."),Cannot_compile_namespaces_when_the_isolatedModules_flag_is_provided:n(1208,e.DiagnosticCategory.Error,"Cannot_compile_namespaces_when_the_isolatedModules_flag_is_provided_1208","Cannot compile namespaces when the '--isolatedModules' flag is provided."),Ambient_const_enums_are_not_allowed_when_the_isolatedModules_flag_is_provided:n(1209,e.DiagnosticCategory.Error,"Ambient_const_enums_are_not_allowed_when_the_isolatedModules_flag_is_provided_1209","Ambient const enums are not allowed when the '--isolatedModules' flag is provided."),Invalid_use_of_0_Class_definitions_are_automatically_in_strict_mode:n(1210,e.DiagnosticCategory.Error,"Invalid_use_of_0_Class_definitions_are_automatically_in_strict_mode_1210","Invalid use of '{0}'. Class definitions are automatically in strict mode."),A_class_declaration_without_the_default_modifier_must_have_a_name:n(1211,e.DiagnosticCategory.Error,"A_class_declaration_without_the_default_modifier_must_have_a_name_1211","A class declaration without the 'default' modifier must have a name."),Identifier_expected_0_is_a_reserved_word_in_strict_mode:n(1212,e.DiagnosticCategory.Error,"Identifier_expected_0_is_a_reserved_word_in_strict_mode_1212","Identifier expected. '{0}' is a reserved word in strict mode."),Identifier_expected_0_is_a_reserved_word_in_strict_mode_Class_definitions_are_automatically_in_strict_mode:n(1213,e.DiagnosticCategory.Error,"Identifier_expected_0_is_a_reserved_word_in_strict_mode_Class_definitions_are_automatically_in_stric_1213","Identifier expected. '{0}' is a reserved word in strict mode. Class definitions are automatically in strict mode."),Identifier_expected_0_is_a_reserved_word_in_strict_mode_Modules_are_automatically_in_strict_mode:n(1214,e.DiagnosticCategory.Error,"Identifier_expected_0_is_a_reserved_word_in_strict_mode_Modules_are_automatically_in_strict_mode_1214","Identifier expected. '{0}' is a reserved word in strict mode. Modules are automatically in strict mode."),Invalid_use_of_0_Modules_are_automatically_in_strict_mode:n(1215,e.DiagnosticCategory.Error,"Invalid_use_of_0_Modules_are_automatically_in_strict_mode_1215","Invalid use of '{0}'. Modules are automatically in strict mode."),Identifier_expected_esModule_is_reserved_as_an_exported_marker_when_transforming_ECMAScript_modules:n(1216,e.DiagnosticCategory.Error,"Identifier_expected_esModule_is_reserved_as_an_exported_marker_when_transforming_ECMAScript_modules_1216","Identifier expected. '__esModule' is reserved as an exported marker when transforming ECMAScript modules."),Export_assignment_is_not_supported_when_module_flag_is_system:n(1218,e.DiagnosticCategory.Error,"Export_assignment_is_not_supported_when_module_flag_is_system_1218","Export assignment is not supported when '--module' flag is 'system'."),Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_to_remove_this_warning:n(1219,e.DiagnosticCategory.Error,"Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_t_1219","Experimental support for decorators is a feature that is subject to change in a future release. Set the 'experimentalDecorators' option to remove this warning."),Generators_are_only_available_when_targeting_ECMAScript_2015_or_higher:n(1220,e.DiagnosticCategory.Error,"Generators_are_only_available_when_targeting_ECMAScript_2015_or_higher_1220","Generators are only available when targeting ECMAScript 2015 or higher."),Generators_are_not_allowed_in_an_ambient_context:n(1221,e.DiagnosticCategory.Error,"Generators_are_not_allowed_in_an_ambient_context_1221","Generators are not allowed in an ambient context."),An_overload_signature_cannot_be_declared_as_a_generator:n(1222,e.DiagnosticCategory.Error,"An_overload_signature_cannot_be_declared_as_a_generator_1222","An overload signature cannot be declared as a generator."),_0_tag_already_specified:n(1223,e.DiagnosticCategory.Error,"_0_tag_already_specified_1223","'{0}' tag already specified."),Signature_0_must_be_a_type_predicate:n(1224,e.DiagnosticCategory.Error,"Signature_0_must_be_a_type_predicate_1224","Signature '{0}' must be a type predicate."),Cannot_find_parameter_0:n(1225,e.DiagnosticCategory.Error,"Cannot_find_parameter_0_1225","Cannot find parameter '{0}'."),Type_predicate_0_is_not_assignable_to_1:n(1226,e.DiagnosticCategory.Error,"Type_predicate_0_is_not_assignable_to_1_1226","Type predicate '{0}' is not assignable to '{1}'."),Parameter_0_is_not_in_the_same_position_as_parameter_1:n(1227,e.DiagnosticCategory.Error,"Parameter_0_is_not_in_the_same_position_as_parameter_1_1227","Parameter '{0}' is not in the same position as parameter '{1}'."),A_type_predicate_is_only_allowed_in_return_type_position_for_functions_and_methods:n(1228,e.DiagnosticCategory.Error,"A_type_predicate_is_only_allowed_in_return_type_position_for_functions_and_methods_1228","A type predicate is only allowed in return type position for functions and methods."),A_type_predicate_cannot_reference_a_rest_parameter:n(1229,e.DiagnosticCategory.Error,"A_type_predicate_cannot_reference_a_rest_parameter_1229","A type predicate cannot reference a rest parameter."),A_type_predicate_cannot_reference_element_0_in_a_binding_pattern:n(1230,e.DiagnosticCategory.Error,"A_type_predicate_cannot_reference_element_0_in_a_binding_pattern_1230","A type predicate cannot reference element '{0}' in a binding pattern."),An_export_assignment_can_only_be_used_in_a_module:n(1231,e.DiagnosticCategory.Error,"An_export_assignment_can_only_be_used_in_a_module_1231","An export assignment can only be used in a module."),An_import_declaration_can_only_be_used_in_a_namespace_or_module:n(1232,e.DiagnosticCategory.Error,"An_import_declaration_can_only_be_used_in_a_namespace_or_module_1232","An import declaration can only be used in a namespace or module."),An_export_declaration_can_only_be_used_in_a_module:n(1233,e.DiagnosticCategory.Error,"An_export_declaration_can_only_be_used_in_a_module_1233","An export declaration can only be used in a module."),An_ambient_module_declaration_is_only_allowed_at_the_top_level_in_a_file:n(1234,e.DiagnosticCategory.Error,"An_ambient_module_declaration_is_only_allowed_at_the_top_level_in_a_file_1234","An ambient module declaration is only allowed at the top level in a file."),A_namespace_declaration_is_only_allowed_in_a_namespace_or_module:n(1235,e.DiagnosticCategory.Error,"A_namespace_declaration_is_only_allowed_in_a_namespace_or_module_1235","A namespace declaration is only allowed in a namespace or module."),The_return_type_of_a_property_decorator_function_must_be_either_void_or_any:n(1236,e.DiagnosticCategory.Error,"The_return_type_of_a_property_decorator_function_must_be_either_void_or_any_1236","The return type of a property decorator function must be either 'void' or 'any'."),The_return_type_of_a_parameter_decorator_function_must_be_either_void_or_any:n(1237,e.DiagnosticCategory.Error,"The_return_type_of_a_parameter_decorator_function_must_be_either_void_or_any_1237","The return type of a parameter decorator function must be either 'void' or 'any'."),Unable_to_resolve_signature_of_class_decorator_when_called_as_an_expression:n(1238,e.DiagnosticCategory.Error,"Unable_to_resolve_signature_of_class_decorator_when_called_as_an_expression_1238","Unable to resolve signature of class decorator when called as an expression."),Unable_to_resolve_signature_of_parameter_decorator_when_called_as_an_expression:n(1239,e.DiagnosticCategory.Error,"Unable_to_resolve_signature_of_parameter_decorator_when_called_as_an_expression_1239","Unable to resolve signature of parameter decorator when called as an expression."),Unable_to_resolve_signature_of_property_decorator_when_called_as_an_expression:n(1240,e.DiagnosticCategory.Error,"Unable_to_resolve_signature_of_property_decorator_when_called_as_an_expression_1240","Unable to resolve signature of property decorator when called as an expression."),Unable_to_resolve_signature_of_method_decorator_when_called_as_an_expression:n(1241,e.DiagnosticCategory.Error,"Unable_to_resolve_signature_of_method_decorator_when_called_as_an_expression_1241","Unable to resolve signature of method decorator when called as an expression."),abstract_modifier_can_only_appear_on_a_class_method_or_property_declaration:n(1242,e.DiagnosticCategory.Error,"abstract_modifier_can_only_appear_on_a_class_method_or_property_declaration_1242","'abstract' modifier can only appear on a class, method, or property declaration."),_0_modifier_cannot_be_used_with_1_modifier:n(1243,e.DiagnosticCategory.Error,"_0_modifier_cannot_be_used_with_1_modifier_1243","'{0}' modifier cannot be used with '{1}' modifier."),Abstract_methods_can_only_appear_within_an_abstract_class:n(1244,e.DiagnosticCategory.Error,"Abstract_methods_can_only_appear_within_an_abstract_class_1244","Abstract methods can only appear within an abstract class."),Method_0_cannot_have_an_implementation_because_it_is_marked_abstract:n(1245,e.DiagnosticCategory.Error,"Method_0_cannot_have_an_implementation_because_it_is_marked_abstract_1245","Method '{0}' cannot have an implementation because it is marked abstract."),An_interface_property_cannot_have_an_initializer:n(1246,e.DiagnosticCategory.Error,"An_interface_property_cannot_have_an_initializer_1246","An interface property cannot have an initializer."),A_type_literal_property_cannot_have_an_initializer:n(1247,e.DiagnosticCategory.Error,"A_type_literal_property_cannot_have_an_initializer_1247","A type literal property cannot have an initializer."),A_class_member_cannot_have_the_0_keyword:n(1248,e.DiagnosticCategory.Error,"A_class_member_cannot_have_the_0_keyword_1248","A class member cannot have the '{0}' keyword."),A_decorator_can_only_decorate_a_method_implementation_not_an_overload:n(1249,e.DiagnosticCategory.Error,"A_decorator_can_only_decorate_a_method_implementation_not_an_overload_1249","A decorator can only decorate a method implementation, not an overload."),Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5:n(1250,e.DiagnosticCategory.Error,"Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_1250","Function declarations are not allowed inside blocks in strict mode when targeting 'ES3' or 'ES5'."),Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Class_definitions_are_automatically_in_strict_mode:n(1251,e.DiagnosticCategory.Error,"Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Class_d_1251","Function declarations are not allowed inside blocks in strict mode when targeting 'ES3' or 'ES5'. Class definitions are automatically in strict mode."),Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Modules_are_automatically_in_strict_mode:n(1252,e.DiagnosticCategory.Error,"Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Modules_1252","Function declarations are not allowed inside blocks in strict mode when targeting 'ES3' or 'ES5'. Modules are automatically in strict mode."),_0_tag_cannot_be_used_independently_as_a_top_level_JSDoc_tag:n(1253,e.DiagnosticCategory.Error,"_0_tag_cannot_be_used_independently_as_a_top_level_JSDoc_tag_1253","'{0}' tag cannot be used independently as a top level JSDoc tag."),A_const_initializer_in_an_ambient_context_must_be_a_string_or_numeric_literal_or_literal_enum_reference:n(1254,e.DiagnosticCategory.Error,"A_const_initializer_in_an_ambient_context_must_be_a_string_or_numeric_literal_or_literal_enum_refere_1254","A 'const' initializer in an ambient context must be a string or numeric literal or literal enum reference."),A_definite_assignment_assertion_is_not_permitted_in_this_context:n(1255,e.DiagnosticCategory.Error,"A_definite_assignment_assertion_is_not_permitted_in_this_context_1255","A definite assignment assertion '!' is not permitted in this context."),A_rest_element_must_be_last_in_a_tuple_type:n(1256,e.DiagnosticCategory.Error,"A_rest_element_must_be_last_in_a_tuple_type_1256","A rest element must be last in a tuple type."),A_required_element_cannot_follow_an_optional_element:n(1257,e.DiagnosticCategory.Error,"A_required_element_cannot_follow_an_optional_element_1257","A required element cannot follow an optional element."),with_statements_are_not_allowed_in_an_async_function_block:n(1300,e.DiagnosticCategory.Error,"with_statements_are_not_allowed_in_an_async_function_block_1300","'with' statements are not allowed in an async function block."),await_expression_is_only_allowed_within_an_async_function:n(1308,e.DiagnosticCategory.Error,"await_expression_is_only_allowed_within_an_async_function_1308","'await' expression is only allowed within an async function."),can_only_be_used_in_an_object_literal_property_inside_a_destructuring_assignment:n(1312,e.DiagnosticCategory.Error,"can_only_be_used_in_an_object_literal_property_inside_a_destructuring_assignment_1312","'=' can only be used in an object literal property inside a destructuring assignment."),The_body_of_an_if_statement_cannot_be_the_empty_statement:n(1313,e.DiagnosticCategory.Error,"The_body_of_an_if_statement_cannot_be_the_empty_statement_1313","The body of an 'if' statement cannot be the empty statement."),Global_module_exports_may_only_appear_in_module_files:n(1314,e.DiagnosticCategory.Error,"Global_module_exports_may_only_appear_in_module_files_1314","Global module exports may only appear in module files."),Global_module_exports_may_only_appear_in_declaration_files:n(1315,e.DiagnosticCategory.Error,"Global_module_exports_may_only_appear_in_declaration_files_1315","Global module exports may only appear in declaration files."),Global_module_exports_may_only_appear_at_top_level:n(1316,e.DiagnosticCategory.Error,"Global_module_exports_may_only_appear_at_top_level_1316","Global module exports may only appear at top level."),A_parameter_property_cannot_be_declared_using_a_rest_parameter:n(1317,e.DiagnosticCategory.Error,"A_parameter_property_cannot_be_declared_using_a_rest_parameter_1317","A parameter property cannot be declared using a rest parameter."),An_abstract_accessor_cannot_have_an_implementation:n(1318,e.DiagnosticCategory.Error,"An_abstract_accessor_cannot_have_an_implementation_1318","An abstract accessor cannot have an implementation."),A_default_export_can_only_be_used_in_an_ECMAScript_style_module:n(1319,e.DiagnosticCategory.Error,"A_default_export_can_only_be_used_in_an_ECMAScript_style_module_1319","A default export can only be used in an ECMAScript-style module."),Type_of_await_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member:n(1320,e.DiagnosticCategory.Error,"Type_of_await_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member_1320","Type of 'await' operand must either be a valid promise or must not contain a callable 'then' member."),Type_of_yield_operand_in_an_async_generator_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member:n(1321,e.DiagnosticCategory.Error,"Type_of_yield_operand_in_an_async_generator_must_either_be_a_valid_promise_or_must_not_contain_a_cal_1321","Type of 'yield' operand in an async generator must either be a valid promise or must not contain a callable 'then' member."),Type_of_iterated_elements_of_a_yield_Asterisk_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member:n(1322,e.DiagnosticCategory.Error,"Type_of_iterated_elements_of_a_yield_Asterisk_operand_must_either_be_a_valid_promise_or_must_not_con_1322","Type of iterated elements of a 'yield*' operand must either be a valid promise or must not contain a callable 'then' member."),Dynamic_import_is_only_supported_when_module_flag_is_commonjs_or_esNext:n(1323,e.DiagnosticCategory.Error,"Dynamic_import_is_only_supported_when_module_flag_is_commonjs_or_esNext_1323","Dynamic import is only supported when '--module' flag is 'commonjs' or 'esNext'."),Dynamic_import_must_have_one_specifier_as_an_argument:n(1324,e.DiagnosticCategory.Error,"Dynamic_import_must_have_one_specifier_as_an_argument_1324","Dynamic import must have one specifier as an argument."),Specifier_of_dynamic_import_cannot_be_spread_element:n(1325,e.DiagnosticCategory.Error,"Specifier_of_dynamic_import_cannot_be_spread_element_1325","Specifier of dynamic import cannot be spread element."),Dynamic_import_cannot_have_type_arguments:n(1326,e.DiagnosticCategory.Error,"Dynamic_import_cannot_have_type_arguments_1326","Dynamic import cannot have type arguments"),String_literal_with_double_quotes_expected:n(1327,e.DiagnosticCategory.Error,"String_literal_with_double_quotes_expected_1327","String literal with double quotes expected."),Property_value_can_only_be_string_literal_numeric_literal_true_false_null_object_literal_or_array_literal:n(1328,e.DiagnosticCategory.Error,"Property_value_can_only_be_string_literal_numeric_literal_true_false_null_object_literal_or_array_li_1328","Property value can only be string literal, numeric literal, 'true', 'false', 'null', object literal or array literal."),_0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write_0:n(1329,e.DiagnosticCategory.Error,"_0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write__1329","'{0}' accepts too few arguments to be used as a decorator here. Did you mean to call it first and write '@{0}()'?"),A_property_of_an_interface_or_type_literal_whose_type_is_a_unique_symbol_type_must_be_readonly:n(1330,e.DiagnosticCategory.Error,"A_property_of_an_interface_or_type_literal_whose_type_is_a_unique_symbol_type_must_be_readonly_1330","A property of an interface or type literal whose type is a 'unique symbol' type must be 'readonly'."),A_property_of_a_class_whose_type_is_a_unique_symbol_type_must_be_both_static_and_readonly:n(1331,e.DiagnosticCategory.Error,"A_property_of_a_class_whose_type_is_a_unique_symbol_type_must_be_both_static_and_readonly_1331","A property of a class whose type is a 'unique symbol' type must be both 'static' and 'readonly'."),A_variable_whose_type_is_a_unique_symbol_type_must_be_const:n(1332,e.DiagnosticCategory.Error,"A_variable_whose_type_is_a_unique_symbol_type_must_be_const_1332","A variable whose type is a 'unique symbol' type must be 'const'."),unique_symbol_types_may_not_be_used_on_a_variable_declaration_with_a_binding_name:n(1333,e.DiagnosticCategory.Error,"unique_symbol_types_may_not_be_used_on_a_variable_declaration_with_a_binding_name_1333","'unique symbol' types may not be used on a variable declaration with a binding name."),unique_symbol_types_are_only_allowed_on_variables_in_a_variable_statement:n(1334,e.DiagnosticCategory.Error,"unique_symbol_types_are_only_allowed_on_variables_in_a_variable_statement_1334","'unique symbol' types are only allowed on variables in a variable statement."),unique_symbol_types_are_not_allowed_here:n(1335,e.DiagnosticCategory.Error,"unique_symbol_types_are_not_allowed_here_1335","'unique symbol' types are not allowed here."),An_index_signature_parameter_type_cannot_be_a_type_alias_Consider_writing_0_Colon_1_Colon_2_instead:n(1336,e.DiagnosticCategory.Error,"An_index_signature_parameter_type_cannot_be_a_type_alias_Consider_writing_0_Colon_1_Colon_2_instead_1336","An index signature parameter type cannot be a type alias. Consider writing '[{0}: {1}]: {2}' instead."),An_index_signature_parameter_type_cannot_be_a_union_type_Consider_using_a_mapped_object_type_instead:n(1337,e.DiagnosticCategory.Error,"An_index_signature_parameter_type_cannot_be_a_union_type_Consider_using_a_mapped_object_type_instead_1337","An index signature parameter type cannot be a union type. Consider using a mapped object type instead."),infer_declarations_are_only_permitted_in_the_extends_clause_of_a_conditional_type:n(1338,e.DiagnosticCategory.Error,"infer_declarations_are_only_permitted_in_the_extends_clause_of_a_conditional_type_1338","'infer' declarations are only permitted in the 'extends' clause of a conditional type."),Module_0_does_not_refer_to_a_value_but_is_used_as_a_value_here:n(1339,e.DiagnosticCategory.Error,"Module_0_does_not_refer_to_a_value_but_is_used_as_a_value_here_1339","Module '{0}' does not refer to a value, but is used as a value here."),Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0:n(1340,e.DiagnosticCategory.Error,"Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0_1340","Module '{0}' does not refer to a type, but is used as a type here. Did you mean 'typeof import('{0}')'?"),Type_arguments_cannot_be_used_here:n(1342,e.DiagnosticCategory.Error,"Type_arguments_cannot_be_used_here_1342","Type arguments cannot be used here."),The_import_meta_meta_property_is_only_allowed_using_ESNext_for_the_target_and_module_compiler_options:n(1343,e.DiagnosticCategory.Error,"The_import_meta_meta_property_is_only_allowed_using_ESNext_for_the_target_and_module_compiler_option_1343","The 'import.meta' meta-property is only allowed using 'ESNext' for the 'target' and 'module' compiler options."),A_label_is_not_allowed_here:n(1344,e.DiagnosticCategory.Error,"A_label_is_not_allowed_here_1344","'A label is not allowed here."),An_expression_of_type_void_cannot_be_tested_for_truthiness:n(1345,e.DiagnosticCategory.Error,"An_expression_of_type_void_cannot_be_tested_for_truthiness_1345","An expression of type 'void' cannot be tested for truthiness"),This_parameter_is_not_allowed_with_use_strict_directive:n(1346,e.DiagnosticCategory.Error,"This_parameter_is_not_allowed_with_use_strict_directive_1346","This parameter is not allowed with 'use strict' directive."),use_strict_directive_cannot_be_used_with_non_simple_parameter_list:n(1347,e.DiagnosticCategory.Error,"use_strict_directive_cannot_be_used_with_non_simple_parameter_list_1347","'use strict' directive cannot be used with non-simple parameter list."),Non_simple_parameter_declared_here:n(1348,e.DiagnosticCategory.Error,"Non_simple_parameter_declared_here_1348","Non-simple parameter declared here."),use_strict_directive_used_here:n(1349,e.DiagnosticCategory.Error,"use_strict_directive_used_here_1349","'use strict' directive used here."),Print_the_final_configuration_instead_of_building:n(1350,e.DiagnosticCategory.Message,"Print_the_final_configuration_instead_of_building_1350","Print the final configuration instead of building."),An_identifier_or_keyword_cannot_immediately_follow_a_numeric_literal:n(1351,e.DiagnosticCategory.Error,"An_identifier_or_keyword_cannot_immediately_follow_a_numeric_literal_1351","An identifier or keyword cannot immediately follow a numeric literal."),A_bigint_literal_cannot_use_exponential_notation:n(1352,e.DiagnosticCategory.Error,"A_bigint_literal_cannot_use_exponential_notation_1352","A bigint literal cannot use exponential notation."),A_bigint_literal_must_be_an_integer:n(1353,e.DiagnosticCategory.Error,"A_bigint_literal_must_be_an_integer_1353","A bigint literal must be an integer."),Duplicate_identifier_0:n(2300,e.DiagnosticCategory.Error,"Duplicate_identifier_0_2300","Duplicate identifier '{0}'."),Initializer_of_instance_member_variable_0_cannot_reference_identifier_1_declared_in_the_constructor:n(2301,e.DiagnosticCategory.Error,"Initializer_of_instance_member_variable_0_cannot_reference_identifier_1_declared_in_the_constructor_2301","Initializer of instance member variable '{0}' cannot reference identifier '{1}' declared in the constructor."),Static_members_cannot_reference_class_type_parameters:n(2302,e.DiagnosticCategory.Error,"Static_members_cannot_reference_class_type_parameters_2302","Static members cannot reference class type parameters."),Circular_definition_of_import_alias_0:n(2303,e.DiagnosticCategory.Error,"Circular_definition_of_import_alias_0_2303","Circular definition of import alias '{0}'."),Cannot_find_name_0:n(2304,e.DiagnosticCategory.Error,"Cannot_find_name_0_2304","Cannot find name '{0}'."),Module_0_has_no_exported_member_1:n(2305,e.DiagnosticCategory.Error,"Module_0_has_no_exported_member_1_2305","Module '{0}' has no exported member '{1}'."),File_0_is_not_a_module:n(2306,e.DiagnosticCategory.Error,"File_0_is_not_a_module_2306","File '{0}' is not a module."),Cannot_find_module_0:n(2307,e.DiagnosticCategory.Error,"Cannot_find_module_0_2307","Cannot find module '{0}'."),Module_0_has_already_exported_a_member_named_1_Consider_explicitly_re_exporting_to_resolve_the_ambiguity:n(2308,e.DiagnosticCategory.Error,"Module_0_has_already_exported_a_member_named_1_Consider_explicitly_re_exporting_to_resolve_the_ambig_2308","Module {0} has already exported a member named '{1}'. Consider explicitly re-exporting to resolve the ambiguity."),An_export_assignment_cannot_be_used_in_a_module_with_other_exported_elements:n(2309,e.DiagnosticCategory.Error,"An_export_assignment_cannot_be_used_in_a_module_with_other_exported_elements_2309","An export assignment cannot be used in a module with other exported elements."),Type_0_recursively_references_itself_as_a_base_type:n(2310,e.DiagnosticCategory.Error,"Type_0_recursively_references_itself_as_a_base_type_2310","Type '{0}' recursively references itself as a base type."),A_class_may_only_extend_another_class:n(2311,e.DiagnosticCategory.Error,"A_class_may_only_extend_another_class_2311","A class may only extend another class."),An_interface_can_only_extend_an_object_type_or_intersection_of_object_types_with_statically_known_members:n(2312,e.DiagnosticCategory.Error,"An_interface_can_only_extend_an_object_type_or_intersection_of_object_types_with_statically_known_me_2312","An interface can only extend an object type or intersection of object types with statically known members."),Type_parameter_0_has_a_circular_constraint:n(2313,e.DiagnosticCategory.Error,"Type_parameter_0_has_a_circular_constraint_2313","Type parameter '{0}' has a circular constraint."),Generic_type_0_requires_1_type_argument_s:n(2314,e.DiagnosticCategory.Error,"Generic_type_0_requires_1_type_argument_s_2314","Generic type '{0}' requires {1} type argument(s)."),Type_0_is_not_generic:n(2315,e.DiagnosticCategory.Error,"Type_0_is_not_generic_2315","Type '{0}' is not generic."),Global_type_0_must_be_a_class_or_interface_type:n(2316,e.DiagnosticCategory.Error,"Global_type_0_must_be_a_class_or_interface_type_2316","Global type '{0}' must be a class or interface type."),Global_type_0_must_have_1_type_parameter_s:n(2317,e.DiagnosticCategory.Error,"Global_type_0_must_have_1_type_parameter_s_2317","Global type '{0}' must have {1} type parameter(s)."),Cannot_find_global_type_0:n(2318,e.DiagnosticCategory.Error,"Cannot_find_global_type_0_2318","Cannot find global type '{0}'."),Named_property_0_of_types_1_and_2_are_not_identical:n(2319,e.DiagnosticCategory.Error,"Named_property_0_of_types_1_and_2_are_not_identical_2319","Named property '{0}' of types '{1}' and '{2}' are not identical."),Interface_0_cannot_simultaneously_extend_types_1_and_2:n(2320,e.DiagnosticCategory.Error,"Interface_0_cannot_simultaneously_extend_types_1_and_2_2320","Interface '{0}' cannot simultaneously extend types '{1}' and '{2}'."),Excessive_stack_depth_comparing_types_0_and_1:n(2321,e.DiagnosticCategory.Error,"Excessive_stack_depth_comparing_types_0_and_1_2321","Excessive stack depth comparing types '{0}' and '{1}'."),Type_0_is_not_assignable_to_type_1:n(2322,e.DiagnosticCategory.Error,"Type_0_is_not_assignable_to_type_1_2322","Type '{0}' is not assignable to type '{1}'."),Cannot_redeclare_exported_variable_0:n(2323,e.DiagnosticCategory.Error,"Cannot_redeclare_exported_variable_0_2323","Cannot redeclare exported variable '{0}'."),Property_0_is_missing_in_type_1:n(2324,e.DiagnosticCategory.Error,"Property_0_is_missing_in_type_1_2324","Property '{0}' is missing in type '{1}'."),Property_0_is_private_in_type_1_but_not_in_type_2:n(2325,e.DiagnosticCategory.Error,"Property_0_is_private_in_type_1_but_not_in_type_2_2325","Property '{0}' is private in type '{1}' but not in type '{2}'."),Types_of_property_0_are_incompatible:n(2326,e.DiagnosticCategory.Error,"Types_of_property_0_are_incompatible_2326","Types of property '{0}' are incompatible."),Property_0_is_optional_in_type_1_but_required_in_type_2:n(2327,e.DiagnosticCategory.Error,"Property_0_is_optional_in_type_1_but_required_in_type_2_2327","Property '{0}' is optional in type '{1}' but required in type '{2}'."),Types_of_parameters_0_and_1_are_incompatible:n(2328,e.DiagnosticCategory.Error,"Types_of_parameters_0_and_1_are_incompatible_2328","Types of parameters '{0}' and '{1}' are incompatible."),Index_signature_is_missing_in_type_0:n(2329,e.DiagnosticCategory.Error,"Index_signature_is_missing_in_type_0_2329","Index signature is missing in type '{0}'."),Index_signatures_are_incompatible:n(2330,e.DiagnosticCategory.Error,"Index_signatures_are_incompatible_2330","Index signatures are incompatible."),this_cannot_be_referenced_in_a_module_or_namespace_body:n(2331,e.DiagnosticCategory.Error,"this_cannot_be_referenced_in_a_module_or_namespace_body_2331","'this' cannot be referenced in a module or namespace body."),this_cannot_be_referenced_in_current_location:n(2332,e.DiagnosticCategory.Error,"this_cannot_be_referenced_in_current_location_2332","'this' cannot be referenced in current location."),this_cannot_be_referenced_in_constructor_arguments:n(2333,e.DiagnosticCategory.Error,"this_cannot_be_referenced_in_constructor_arguments_2333","'this' cannot be referenced in constructor arguments."),this_cannot_be_referenced_in_a_static_property_initializer:n(2334,e.DiagnosticCategory.Error,"this_cannot_be_referenced_in_a_static_property_initializer_2334","'this' cannot be referenced in a static property initializer."),super_can_only_be_referenced_in_a_derived_class:n(2335,e.DiagnosticCategory.Error,"super_can_only_be_referenced_in_a_derived_class_2335","'super' can only be referenced in a derived class."),super_cannot_be_referenced_in_constructor_arguments:n(2336,e.DiagnosticCategory.Error,"super_cannot_be_referenced_in_constructor_arguments_2336","'super' cannot be referenced in constructor arguments."),Super_calls_are_not_permitted_outside_constructors_or_in_nested_functions_inside_constructors:n(2337,e.DiagnosticCategory.Error,"Super_calls_are_not_permitted_outside_constructors_or_in_nested_functions_inside_constructors_2337","Super calls are not permitted outside constructors or in nested functions inside constructors."),super_property_access_is_permitted_only_in_a_constructor_member_function_or_member_accessor_of_a_derived_class:n(2338,e.DiagnosticCategory.Error,"super_property_access_is_permitted_only_in_a_constructor_member_function_or_member_accessor_of_a_der_2338","'super' property access is permitted only in a constructor, member function, or member accessor of a derived class."),Property_0_does_not_exist_on_type_1:n(2339,e.DiagnosticCategory.Error,"Property_0_does_not_exist_on_type_1_2339","Property '{0}' does not exist on type '{1}'."),Only_public_and_protected_methods_of_the_base_class_are_accessible_via_the_super_keyword:n(2340,e.DiagnosticCategory.Error,"Only_public_and_protected_methods_of_the_base_class_are_accessible_via_the_super_keyword_2340","Only public and protected methods of the base class are accessible via the 'super' keyword."),Property_0_is_private_and_only_accessible_within_class_1:n(2341,e.DiagnosticCategory.Error,"Property_0_is_private_and_only_accessible_within_class_1_2341","Property '{0}' is private and only accessible within class '{1}'."),An_index_expression_argument_must_be_of_type_string_number_symbol_or_any:n(2342,e.DiagnosticCategory.Error,"An_index_expression_argument_must_be_of_type_string_number_symbol_or_any_2342","An index expression argument must be of type 'string', 'number', 'symbol', or 'any'."),This_syntax_requires_an_imported_helper_named_1_but_module_0_has_no_exported_member_1:n(2343,e.DiagnosticCategory.Error,"This_syntax_requires_an_imported_helper_named_1_but_module_0_has_no_exported_member_1_2343","This syntax requires an imported helper named '{1}', but module '{0}' has no exported member '{1}'."),Type_0_does_not_satisfy_the_constraint_1:n(2344,e.DiagnosticCategory.Error,"Type_0_does_not_satisfy_the_constraint_1_2344","Type '{0}' does not satisfy the constraint '{1}'."),Argument_of_type_0_is_not_assignable_to_parameter_of_type_1:n(2345,e.DiagnosticCategory.Error,"Argument_of_type_0_is_not_assignable_to_parameter_of_type_1_2345","Argument of type '{0}' is not assignable to parameter of type '{1}'."),Call_target_does_not_contain_any_signatures:n(2346,e.DiagnosticCategory.Error,"Call_target_does_not_contain_any_signatures_2346","Call target does not contain any signatures."),Untyped_function_calls_may_not_accept_type_arguments:n(2347,e.DiagnosticCategory.Error,"Untyped_function_calls_may_not_accept_type_arguments_2347","Untyped function calls may not accept type arguments."),Value_of_type_0_is_not_callable_Did_you_mean_to_include_new:n(2348,e.DiagnosticCategory.Error,"Value_of_type_0_is_not_callable_Did_you_mean_to_include_new_2348","Value of type '{0}' is not callable. Did you mean to include 'new'?"),Cannot_invoke_an_expression_whose_type_lacks_a_call_signature_Type_0_has_no_compatible_call_signatures:n(2349,e.DiagnosticCategory.Error,"Cannot_invoke_an_expression_whose_type_lacks_a_call_signature_Type_0_has_no_compatible_call_signatur_2349","Cannot invoke an expression whose type lacks a call signature. Type '{0}' has no compatible call signatures."),Only_a_void_function_can_be_called_with_the_new_keyword:n(2350,e.DiagnosticCategory.Error,"Only_a_void_function_can_be_called_with_the_new_keyword_2350","Only a void function can be called with the 'new' keyword."),Cannot_use_new_with_an_expression_whose_type_lacks_a_call_or_construct_signature:n(2351,e.DiagnosticCategory.Error,"Cannot_use_new_with_an_expression_whose_type_lacks_a_call_or_construct_signature_2351","Cannot use 'new' with an expression whose type lacks a call or construct signature."),Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the_other_If_this_was_intentional_convert_the_expression_to_unknown_first:n(2352,e.DiagnosticCategory.Error,"Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the__2352","Conversion of type '{0}' to type '{1}' may be a mistake because neither type sufficiently overlaps with the other. If this was intentional, convert the expression to 'unknown' first."),Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1:n(2353,e.DiagnosticCategory.Error,"Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1_2353","Object literal may only specify known properties, and '{0}' does not exist in type '{1}'."),This_syntax_requires_an_imported_helper_but_module_0_cannot_be_found:n(2354,e.DiagnosticCategory.Error,"This_syntax_requires_an_imported_helper_but_module_0_cannot_be_found_2354","This syntax requires an imported helper but module '{0}' cannot be found."),A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value:n(2355,e.DiagnosticCategory.Error,"A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value_2355","A function whose declared type is neither 'void' nor 'any' must return a value."),An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type:n(2356,e.DiagnosticCategory.Error,"An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type_2356","An arithmetic operand must be of type 'any', 'number', 'bigint' or an enum type."),The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access:n(2357,e.DiagnosticCategory.Error,"The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access_2357","The operand of an increment or decrement operator must be a variable or a property access."),The_left_hand_side_of_an_instanceof_expression_must_be_of_type_any_an_object_type_or_a_type_parameter:n(2358,e.DiagnosticCategory.Error,"The_left_hand_side_of_an_instanceof_expression_must_be_of_type_any_an_object_type_or_a_type_paramete_2358","The left-hand side of an 'instanceof' expression must be of type 'any', an object type or a type parameter."),The_right_hand_side_of_an_instanceof_expression_must_be_of_type_any_or_of_a_type_assignable_to_the_Function_interface_type:n(2359,e.DiagnosticCategory.Error,"The_right_hand_side_of_an_instanceof_expression_must_be_of_type_any_or_of_a_type_assignable_to_the_F_2359","The right-hand side of an 'instanceof' expression must be of type 'any' or of a type assignable to the 'Function' interface type."),The_left_hand_side_of_an_in_expression_must_be_of_type_any_string_number_or_symbol:n(2360,e.DiagnosticCategory.Error,"The_left_hand_side_of_an_in_expression_must_be_of_type_any_string_number_or_symbol_2360","The left-hand side of an 'in' expression must be of type 'any', 'string', 'number', or 'symbol'."),The_right_hand_side_of_an_in_expression_must_be_of_type_any_an_object_type_or_a_type_parameter:n(2361,e.DiagnosticCategory.Error,"The_right_hand_side_of_an_in_expression_must_be_of_type_any_an_object_type_or_a_type_parameter_2361","The right-hand side of an 'in' expression must be of type 'any', an object type or a type parameter."),The_left_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type:n(2362,e.DiagnosticCategory.Error,"The_left_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type_2362","The left-hand side of an arithmetic operation must be of type 'any', 'number', 'bigint' or an enum type."),The_right_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type:n(2363,e.DiagnosticCategory.Error,"The_right_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type_2363","The right-hand side of an arithmetic operation must be of type 'any', 'number', 'bigint' or an enum type."),The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access:n(2364,e.DiagnosticCategory.Error,"The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access_2364","The left-hand side of an assignment expression must be a variable or a property access."),Operator_0_cannot_be_applied_to_types_1_and_2:n(2365,e.DiagnosticCategory.Error,"Operator_0_cannot_be_applied_to_types_1_and_2_2365","Operator '{0}' cannot be applied to types '{1}' and '{2}'."),Function_lacks_ending_return_statement_and_return_type_does_not_include_undefined:n(2366,e.DiagnosticCategory.Error,"Function_lacks_ending_return_statement_and_return_type_does_not_include_undefined_2366","Function lacks ending return statement and return type does not include 'undefined'."),This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap:n(2367,e.DiagnosticCategory.Error,"This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap_2367","This condition will always return '{0}' since the types '{1}' and '{2}' have no overlap."),Type_parameter_name_cannot_be_0:n(2368,e.DiagnosticCategory.Error,"Type_parameter_name_cannot_be_0_2368","Type parameter name cannot be '{0}'."),A_parameter_property_is_only_allowed_in_a_constructor_implementation:n(2369,e.DiagnosticCategory.Error,"A_parameter_property_is_only_allowed_in_a_constructor_implementation_2369","A parameter property is only allowed in a constructor implementation."),A_rest_parameter_must_be_of_an_array_type:n(2370,e.DiagnosticCategory.Error,"A_rest_parameter_must_be_of_an_array_type_2370","A rest parameter must be of an array type."),A_parameter_initializer_is_only_allowed_in_a_function_or_constructor_implementation:n(2371,e.DiagnosticCategory.Error,"A_parameter_initializer_is_only_allowed_in_a_function_or_constructor_implementation_2371","A parameter initializer is only allowed in a function or constructor implementation."),Parameter_0_cannot_be_referenced_in_its_initializer:n(2372,e.DiagnosticCategory.Error,"Parameter_0_cannot_be_referenced_in_its_initializer_2372","Parameter '{0}' cannot be referenced in its initializer."),Initializer_of_parameter_0_cannot_reference_identifier_1_declared_after_it:n(2373,e.DiagnosticCategory.Error,"Initializer_of_parameter_0_cannot_reference_identifier_1_declared_after_it_2373","Initializer of parameter '{0}' cannot reference identifier '{1}' declared after it."),Duplicate_string_index_signature:n(2374,e.DiagnosticCategory.Error,"Duplicate_string_index_signature_2374","Duplicate string index signature."),Duplicate_number_index_signature:n(2375,e.DiagnosticCategory.Error,"Duplicate_number_index_signature_2375","Duplicate number index signature."),A_super_call_must_be_the_first_statement_in_the_constructor_when_a_class_contains_initialized_properties_or_has_parameter_properties:n(2376,e.DiagnosticCategory.Error,"A_super_call_must_be_the_first_statement_in_the_constructor_when_a_class_contains_initialized_proper_2376","A 'super' call must be the first statement in the constructor when a class contains initialized properties or has parameter properties."),Constructors_for_derived_classes_must_contain_a_super_call:n(2377,e.DiagnosticCategory.Error,"Constructors_for_derived_classes_must_contain_a_super_call_2377","Constructors for derived classes must contain a 'super' call."),A_get_accessor_must_return_a_value:n(2378,e.DiagnosticCategory.Error,"A_get_accessor_must_return_a_value_2378","A 'get' accessor must return a value."),Getter_and_setter_accessors_do_not_agree_in_visibility:n(2379,e.DiagnosticCategory.Error,"Getter_and_setter_accessors_do_not_agree_in_visibility_2379","Getter and setter accessors do not agree in visibility."),get_and_set_accessor_must_have_the_same_type:n(2380,e.DiagnosticCategory.Error,"get_and_set_accessor_must_have_the_same_type_2380","'get' and 'set' accessor must have the same type."),A_signature_with_an_implementation_cannot_use_a_string_literal_type:n(2381,e.DiagnosticCategory.Error,"A_signature_with_an_implementation_cannot_use_a_string_literal_type_2381","A signature with an implementation cannot use a string literal type."),Specialized_overload_signature_is_not_assignable_to_any_non_specialized_signature:n(2382,e.DiagnosticCategory.Error,"Specialized_overload_signature_is_not_assignable_to_any_non_specialized_signature_2382","Specialized overload signature is not assignable to any non-specialized signature."),Overload_signatures_must_all_be_exported_or_non_exported:n(2383,e.DiagnosticCategory.Error,"Overload_signatures_must_all_be_exported_or_non_exported_2383","Overload signatures must all be exported or non-exported."),Overload_signatures_must_all_be_ambient_or_non_ambient:n(2384,e.DiagnosticCategory.Error,"Overload_signatures_must_all_be_ambient_or_non_ambient_2384","Overload signatures must all be ambient or non-ambient."),Overload_signatures_must_all_be_public_private_or_protected:n(2385,e.DiagnosticCategory.Error,"Overload_signatures_must_all_be_public_private_or_protected_2385","Overload signatures must all be public, private or protected."),Overload_signatures_must_all_be_optional_or_required:n(2386,e.DiagnosticCategory.Error,"Overload_signatures_must_all_be_optional_or_required_2386","Overload signatures must all be optional or required."),Function_overload_must_be_static:n(2387,e.DiagnosticCategory.Error,"Function_overload_must_be_static_2387","Function overload must be static."),Function_overload_must_not_be_static:n(2388,e.DiagnosticCategory.Error,"Function_overload_must_not_be_static_2388","Function overload must not be static."),Function_implementation_name_must_be_0:n(2389,e.DiagnosticCategory.Error,"Function_implementation_name_must_be_0_2389","Function implementation name must be '{0}'."),Constructor_implementation_is_missing:n(2390,e.DiagnosticCategory.Error,"Constructor_implementation_is_missing_2390","Constructor implementation is missing."),Function_implementation_is_missing_or_not_immediately_following_the_declaration:n(2391,e.DiagnosticCategory.Error,"Function_implementation_is_missing_or_not_immediately_following_the_declaration_2391","Function implementation is missing or not immediately following the declaration."),Multiple_constructor_implementations_are_not_allowed:n(2392,e.DiagnosticCategory.Error,"Multiple_constructor_implementations_are_not_allowed_2392","Multiple constructor implementations are not allowed."),Duplicate_function_implementation:n(2393,e.DiagnosticCategory.Error,"Duplicate_function_implementation_2393","Duplicate function implementation."),Overload_signature_is_not_compatible_with_function_implementation:n(2394,e.DiagnosticCategory.Error,"Overload_signature_is_not_compatible_with_function_implementation_2394","Overload signature is not compatible with function implementation."),Individual_declarations_in_merged_declaration_0_must_be_all_exported_or_all_local:n(2395,e.DiagnosticCategory.Error,"Individual_declarations_in_merged_declaration_0_must_be_all_exported_or_all_local_2395","Individual declarations in merged declaration '{0}' must be all exported or all local."),Duplicate_identifier_arguments_Compiler_uses_arguments_to_initialize_rest_parameters:n(2396,e.DiagnosticCategory.Error,"Duplicate_identifier_arguments_Compiler_uses_arguments_to_initialize_rest_parameters_2396","Duplicate identifier 'arguments'. Compiler uses 'arguments' to initialize rest parameters."),Declaration_name_conflicts_with_built_in_global_identifier_0:n(2397,e.DiagnosticCategory.Error,"Declaration_name_conflicts_with_built_in_global_identifier_0_2397","Declaration name conflicts with built-in global identifier '{0}'."),Duplicate_identifier_this_Compiler_uses_variable_declaration_this_to_capture_this_reference:n(2399,e.DiagnosticCategory.Error,"Duplicate_identifier_this_Compiler_uses_variable_declaration_this_to_capture_this_reference_2399","Duplicate identifier '_this'. Compiler uses variable declaration '_this' to capture 'this' reference."),Expression_resolves_to_variable_declaration_this_that_compiler_uses_to_capture_this_reference:n(2400,e.DiagnosticCategory.Error,"Expression_resolves_to_variable_declaration_this_that_compiler_uses_to_capture_this_reference_2400","Expression resolves to variable declaration '_this' that compiler uses to capture 'this' reference."),Duplicate_identifier_super_Compiler_uses_super_to_capture_base_class_reference:n(2401,e.DiagnosticCategory.Error,"Duplicate_identifier_super_Compiler_uses_super_to_capture_base_class_reference_2401","Duplicate identifier '_super'. Compiler uses '_super' to capture base class reference."),Expression_resolves_to_super_that_compiler_uses_to_capture_base_class_reference:n(2402,e.DiagnosticCategory.Error,"Expression_resolves_to_super_that_compiler_uses_to_capture_base_class_reference_2402","Expression resolves to '_super' that compiler uses to capture base class reference."),Subsequent_variable_declarations_must_have_the_same_type_Variable_0_must_be_of_type_1_but_here_has_type_2:n(2403,e.DiagnosticCategory.Error,"Subsequent_variable_declarations_must_have_the_same_type_Variable_0_must_be_of_type_1_but_here_has_t_2403","Subsequent variable declarations must have the same type.  Variable '{0}' must be of type '{1}', but here has type '{2}'."),The_left_hand_side_of_a_for_in_statement_cannot_use_a_type_annotation:n(2404,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_in_statement_cannot_use_a_type_annotation_2404","The left-hand side of a 'for...in' statement cannot use a type annotation."),The_left_hand_side_of_a_for_in_statement_must_be_of_type_string_or_any:n(2405,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_in_statement_must_be_of_type_string_or_any_2405","The left-hand side of a 'for...in' statement must be of type 'string' or 'any'."),The_left_hand_side_of_a_for_in_statement_must_be_a_variable_or_a_property_access:n(2406,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_in_statement_must_be_a_variable_or_a_property_access_2406","The left-hand side of a 'for...in' statement must be a variable or a property access."),The_right_hand_side_of_a_for_in_statement_must_be_of_type_any_an_object_type_or_a_type_parameter_but_here_has_type_0:n(2407,e.DiagnosticCategory.Error,"The_right_hand_side_of_a_for_in_statement_must_be_of_type_any_an_object_type_or_a_type_parameter_but_2407","The right-hand side of a 'for...in' statement must be of type 'any', an object type or a type parameter, but here has type '{0}'."),Setters_cannot_return_a_value:n(2408,e.DiagnosticCategory.Error,"Setters_cannot_return_a_value_2408","Setters cannot return a value."),Return_type_of_constructor_signature_must_be_assignable_to_the_instance_type_of_the_class:n(2409,e.DiagnosticCategory.Error,"Return_type_of_constructor_signature_must_be_assignable_to_the_instance_type_of_the_class_2409","Return type of constructor signature must be assignable to the instance type of the class."),The_with_statement_is_not_supported_All_symbols_in_a_with_block_will_have_type_any:n(2410,e.DiagnosticCategory.Error,"The_with_statement_is_not_supported_All_symbols_in_a_with_block_will_have_type_any_2410","The 'with' statement is not supported. All symbols in a 'with' block will have type 'any'."),Property_0_of_type_1_is_not_assignable_to_string_index_type_2:n(2411,e.DiagnosticCategory.Error,"Property_0_of_type_1_is_not_assignable_to_string_index_type_2_2411","Property '{0}' of type '{1}' is not assignable to string index type '{2}'."),Property_0_of_type_1_is_not_assignable_to_numeric_index_type_2:n(2412,e.DiagnosticCategory.Error,"Property_0_of_type_1_is_not_assignable_to_numeric_index_type_2_2412","Property '{0}' of type '{1}' is not assignable to numeric index type '{2}'."),Numeric_index_type_0_is_not_assignable_to_string_index_type_1:n(2413,e.DiagnosticCategory.Error,"Numeric_index_type_0_is_not_assignable_to_string_index_type_1_2413","Numeric index type '{0}' is not assignable to string index type '{1}'."),Class_name_cannot_be_0:n(2414,e.DiagnosticCategory.Error,"Class_name_cannot_be_0_2414","Class name cannot be '{0}'."),Class_0_incorrectly_extends_base_class_1:n(2415,e.DiagnosticCategory.Error,"Class_0_incorrectly_extends_base_class_1_2415","Class '{0}' incorrectly extends base class '{1}'."),Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2:n(2416,e.DiagnosticCategory.Error,"Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2_2416","Property '{0}' in type '{1}' is not assignable to the same property in base type '{2}'."),Class_static_side_0_incorrectly_extends_base_class_static_side_1:n(2417,e.DiagnosticCategory.Error,"Class_static_side_0_incorrectly_extends_base_class_static_side_1_2417","Class static side '{0}' incorrectly extends base class static side '{1}'."),Type_of_computed_property_s_value_is_0_which_is_not_assignable_to_type_1:n(2418,e.DiagnosticCategory.Error,"Type_of_computed_property_s_value_is_0_which_is_not_assignable_to_type_1_2418","Type of computed property's value is '{0}', which is not assignable to type '{1}'."),Class_0_incorrectly_implements_interface_1:n(2420,e.DiagnosticCategory.Error,"Class_0_incorrectly_implements_interface_1_2420","Class '{0}' incorrectly implements interface '{1}'."),A_class_can_only_implement_an_object_type_or_intersection_of_object_types_with_statically_known_members:n(2422,e.DiagnosticCategory.Error,"A_class_can_only_implement_an_object_type_or_intersection_of_object_types_with_statically_known_memb_2422","A class can only implement an object type or intersection of object types with statically known members."),Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_accessor:n(2423,e.DiagnosticCategory.Error,"Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_access_2423","Class '{0}' defines instance member function '{1}', but extended class '{2}' defines it as instance member accessor."),Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_property:n(2424,e.DiagnosticCategory.Error,"Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_proper_2424","Class '{0}' defines instance member function '{1}', but extended class '{2}' defines it as instance member property."),Class_0_defines_instance_member_property_1_but_extended_class_2_defines_it_as_instance_member_function:n(2425,e.DiagnosticCategory.Error,"Class_0_defines_instance_member_property_1_but_extended_class_2_defines_it_as_instance_member_functi_2425","Class '{0}' defines instance member property '{1}', but extended class '{2}' defines it as instance member function."),Class_0_defines_instance_member_accessor_1_but_extended_class_2_defines_it_as_instance_member_function:n(2426,e.DiagnosticCategory.Error,"Class_0_defines_instance_member_accessor_1_but_extended_class_2_defines_it_as_instance_member_functi_2426","Class '{0}' defines instance member accessor '{1}', but extended class '{2}' defines it as instance member function."),Interface_name_cannot_be_0:n(2427,e.DiagnosticCategory.Error,"Interface_name_cannot_be_0_2427","Interface name cannot be '{0}'."),All_declarations_of_0_must_have_identical_type_parameters:n(2428,e.DiagnosticCategory.Error,"All_declarations_of_0_must_have_identical_type_parameters_2428","All declarations of '{0}' must have identical type parameters."),Interface_0_incorrectly_extends_interface_1:n(2430,e.DiagnosticCategory.Error,"Interface_0_incorrectly_extends_interface_1_2430","Interface '{0}' incorrectly extends interface '{1}'."),Enum_name_cannot_be_0:n(2431,e.DiagnosticCategory.Error,"Enum_name_cannot_be_0_2431","Enum name cannot be '{0}'."),In_an_enum_with_multiple_declarations_only_one_declaration_can_omit_an_initializer_for_its_first_enum_element:n(2432,e.DiagnosticCategory.Error,"In_an_enum_with_multiple_declarations_only_one_declaration_can_omit_an_initializer_for_its_first_enu_2432","In an enum with multiple declarations, only one declaration can omit an initializer for its first enum element."),A_namespace_declaration_cannot_be_in_a_different_file_from_a_class_or_function_with_which_it_is_merged:n(2433,e.DiagnosticCategory.Error,"A_namespace_declaration_cannot_be_in_a_different_file_from_a_class_or_function_with_which_it_is_merg_2433","A namespace declaration cannot be in a different file from a class or function with which it is merged."),A_namespace_declaration_cannot_be_located_prior_to_a_class_or_function_with_which_it_is_merged:n(2434,e.DiagnosticCategory.Error,"A_namespace_declaration_cannot_be_located_prior_to_a_class_or_function_with_which_it_is_merged_2434","A namespace declaration cannot be located prior to a class or function with which it is merged."),Ambient_modules_cannot_be_nested_in_other_modules_or_namespaces:n(2435,e.DiagnosticCategory.Error,"Ambient_modules_cannot_be_nested_in_other_modules_or_namespaces_2435","Ambient modules cannot be nested in other modules or namespaces."),Ambient_module_declaration_cannot_specify_relative_module_name:n(2436,e.DiagnosticCategory.Error,"Ambient_module_declaration_cannot_specify_relative_module_name_2436","Ambient module declaration cannot specify relative module name."),Module_0_is_hidden_by_a_local_declaration_with_the_same_name:n(2437,e.DiagnosticCategory.Error,"Module_0_is_hidden_by_a_local_declaration_with_the_same_name_2437","Module '{0}' is hidden by a local declaration with the same name."),Import_name_cannot_be_0:n(2438,e.DiagnosticCategory.Error,"Import_name_cannot_be_0_2438","Import name cannot be '{0}'."),Import_or_export_declaration_in_an_ambient_module_declaration_cannot_reference_module_through_relative_module_name:n(2439,e.DiagnosticCategory.Error,"Import_or_export_declaration_in_an_ambient_module_declaration_cannot_reference_module_through_relati_2439","Import or export declaration in an ambient module declaration cannot reference module through relative module name."),Import_declaration_conflicts_with_local_declaration_of_0:n(2440,e.DiagnosticCategory.Error,"Import_declaration_conflicts_with_local_declaration_of_0_2440","Import declaration conflicts with local declaration of '{0}'."),Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module:n(2441,e.DiagnosticCategory.Error,"Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_2441","Duplicate identifier '{0}'. Compiler reserves name '{1}' in top level scope of a module."),Types_have_separate_declarations_of_a_private_property_0:n(2442,e.DiagnosticCategory.Error,"Types_have_separate_declarations_of_a_private_property_0_2442","Types have separate declarations of a private property '{0}'."),Property_0_is_protected_but_type_1_is_not_a_class_derived_from_2:n(2443,e.DiagnosticCategory.Error,"Property_0_is_protected_but_type_1_is_not_a_class_derived_from_2_2443","Property '{0}' is protected but type '{1}' is not a class derived from '{2}'."),Property_0_is_protected_in_type_1_but_public_in_type_2:n(2444,e.DiagnosticCategory.Error,"Property_0_is_protected_in_type_1_but_public_in_type_2_2444","Property '{0}' is protected in type '{1}' but public in type '{2}'."),Property_0_is_protected_and_only_accessible_within_class_1_and_its_subclasses:n(2445,e.DiagnosticCategory.Error,"Property_0_is_protected_and_only_accessible_within_class_1_and_its_subclasses_2445","Property '{0}' is protected and only accessible within class '{1}' and its subclasses."),Property_0_is_protected_and_only_accessible_through_an_instance_of_class_1:n(2446,e.DiagnosticCategory.Error,"Property_0_is_protected_and_only_accessible_through_an_instance_of_class_1_2446","Property '{0}' is protected and only accessible through an instance of class '{1}'."),The_0_operator_is_not_allowed_for_boolean_types_Consider_using_1_instead:n(2447,e.DiagnosticCategory.Error,"The_0_operator_is_not_allowed_for_boolean_types_Consider_using_1_instead_2447","The '{0}' operator is not allowed for boolean types. Consider using '{1}' instead."),Block_scoped_variable_0_used_before_its_declaration:n(2448,e.DiagnosticCategory.Error,"Block_scoped_variable_0_used_before_its_declaration_2448","Block-scoped variable '{0}' used before its declaration."),Class_0_used_before_its_declaration:n(2449,e.DiagnosticCategory.Error,"Class_0_used_before_its_declaration_2449","Class '{0}' used before its declaration."),Enum_0_used_before_its_declaration:n(2450,e.DiagnosticCategory.Error,"Enum_0_used_before_its_declaration_2450","Enum '{0}' used before its declaration."),Cannot_redeclare_block_scoped_variable_0:n(2451,e.DiagnosticCategory.Error,"Cannot_redeclare_block_scoped_variable_0_2451","Cannot redeclare block-scoped variable '{0}'."),An_enum_member_cannot_have_a_numeric_name:n(2452,e.DiagnosticCategory.Error,"An_enum_member_cannot_have_a_numeric_name_2452","An enum member cannot have a numeric name."),The_type_argument_for_type_parameter_0_cannot_be_inferred_from_the_usage_Consider_specifying_the_type_arguments_explicitly:n(2453,e.DiagnosticCategory.Error,"The_type_argument_for_type_parameter_0_cannot_be_inferred_from_the_usage_Consider_specifying_the_typ_2453","The type argument for type parameter '{0}' cannot be inferred from the usage. Consider specifying the type arguments explicitly."),Variable_0_is_used_before_being_assigned:n(2454,e.DiagnosticCategory.Error,"Variable_0_is_used_before_being_assigned_2454","Variable '{0}' is used before being assigned."),Type_argument_candidate_1_is_not_a_valid_type_argument_because_it_is_not_a_supertype_of_candidate_0:n(2455,e.DiagnosticCategory.Error,"Type_argument_candidate_1_is_not_a_valid_type_argument_because_it_is_not_a_supertype_of_candidate_0_2455","Type argument candidate '{1}' is not a valid type argument because it is not a supertype of candidate '{0}'."),Type_alias_0_circularly_references_itself:n(2456,e.DiagnosticCategory.Error,"Type_alias_0_circularly_references_itself_2456","Type alias '{0}' circularly references itself."),Type_alias_name_cannot_be_0:n(2457,e.DiagnosticCategory.Error,"Type_alias_name_cannot_be_0_2457","Type alias name cannot be '{0}'."),An_AMD_module_cannot_have_multiple_name_assignments:n(2458,e.DiagnosticCategory.Error,"An_AMD_module_cannot_have_multiple_name_assignments_2458","An AMD module cannot have multiple name assignments."),Type_0_is_not_an_array_type:n(2461,e.DiagnosticCategory.Error,"Type_0_is_not_an_array_type_2461","Type '{0}' is not an array type."),A_rest_element_must_be_last_in_a_destructuring_pattern:n(2462,e.DiagnosticCategory.Error,"A_rest_element_must_be_last_in_a_destructuring_pattern_2462","A rest element must be last in a destructuring pattern."),A_binding_pattern_parameter_cannot_be_optional_in_an_implementation_signature:n(2463,e.DiagnosticCategory.Error,"A_binding_pattern_parameter_cannot_be_optional_in_an_implementation_signature_2463","A binding pattern parameter cannot be optional in an implementation signature."),A_computed_property_name_must_be_of_type_string_number_symbol_or_any:n(2464,e.DiagnosticCategory.Error,"A_computed_property_name_must_be_of_type_string_number_symbol_or_any_2464","A computed property name must be of type 'string', 'number', 'symbol', or 'any'."),this_cannot_be_referenced_in_a_computed_property_name:n(2465,e.DiagnosticCategory.Error,"this_cannot_be_referenced_in_a_computed_property_name_2465","'this' cannot be referenced in a computed property name."),super_cannot_be_referenced_in_a_computed_property_name:n(2466,e.DiagnosticCategory.Error,"super_cannot_be_referenced_in_a_computed_property_name_2466","'super' cannot be referenced in a computed property name."),A_computed_property_name_cannot_reference_a_type_parameter_from_its_containing_type:n(2467,e.DiagnosticCategory.Error,"A_computed_property_name_cannot_reference_a_type_parameter_from_its_containing_type_2467","A computed property name cannot reference a type parameter from its containing type."),Cannot_find_global_value_0:n(2468,e.DiagnosticCategory.Error,"Cannot_find_global_value_0_2468","Cannot find global value '{0}'."),The_0_operator_cannot_be_applied_to_type_symbol:n(2469,e.DiagnosticCategory.Error,"The_0_operator_cannot_be_applied_to_type_symbol_2469","The '{0}' operator cannot be applied to type 'symbol'."),Symbol_reference_does_not_refer_to_the_global_Symbol_constructor_object:n(2470,e.DiagnosticCategory.Error,"Symbol_reference_does_not_refer_to_the_global_Symbol_constructor_object_2470","'Symbol' reference does not refer to the global Symbol constructor object."),A_computed_property_name_of_the_form_0_must_be_of_type_symbol:n(2471,e.DiagnosticCategory.Error,"A_computed_property_name_of_the_form_0_must_be_of_type_symbol_2471","A computed property name of the form '{0}' must be of type 'symbol'."),Spread_operator_in_new_expressions_is_only_available_when_targeting_ECMAScript_5_and_higher:n(2472,e.DiagnosticCategory.Error,"Spread_operator_in_new_expressions_is_only_available_when_targeting_ECMAScript_5_and_higher_2472","Spread operator in 'new' expressions is only available when targeting ECMAScript 5 and higher."),Enum_declarations_must_all_be_const_or_non_const:n(2473,e.DiagnosticCategory.Error,"Enum_declarations_must_all_be_const_or_non_const_2473","Enum declarations must all be const or non-const."),const_enum_member_initializers_can_only_contain_literal_values_and_other_computed_enum_values:n(2474,e.DiagnosticCategory.Error,"const_enum_member_initializers_can_only_contain_literal_values_and_other_computed_enum_values_2474","const enum member initializers can only contain literal values and other computed enum values."),const_enums_can_only_be_used_in_property_or_index_access_expressions_or_the_right_hand_side_of_an_import_declaration_or_export_assignment_or_type_query:n(2475,e.DiagnosticCategory.Error,"const_enums_can_only_be_used_in_property_or_index_access_expressions_or_the_right_hand_side_of_an_im_2475","'const' enums can only be used in property or index access expressions or the right hand side of an import declaration or export assignment or type query."),A_const_enum_member_can_only_be_accessed_using_a_string_literal:n(2476,e.DiagnosticCategory.Error,"A_const_enum_member_can_only_be_accessed_using_a_string_literal_2476","A const enum member can only be accessed using a string literal."),const_enum_member_initializer_was_evaluated_to_a_non_finite_value:n(2477,e.DiagnosticCategory.Error,"const_enum_member_initializer_was_evaluated_to_a_non_finite_value_2477","'const' enum member initializer was evaluated to a non-finite value."),const_enum_member_initializer_was_evaluated_to_disallowed_value_NaN:n(2478,e.DiagnosticCategory.Error,"const_enum_member_initializer_was_evaluated_to_disallowed_value_NaN_2478","'const' enum member initializer was evaluated to disallowed value 'NaN'."),Property_0_does_not_exist_on_const_enum_1:n(2479,e.DiagnosticCategory.Error,"Property_0_does_not_exist_on_const_enum_1_2479","Property '{0}' does not exist on 'const' enum '{1}'."),let_is_not_allowed_to_be_used_as_a_name_in_let_or_const_declarations:n(2480,e.DiagnosticCategory.Error,"let_is_not_allowed_to_be_used_as_a_name_in_let_or_const_declarations_2480","'let' is not allowed to be used as a name in 'let' or 'const' declarations."),Cannot_initialize_outer_scoped_variable_0_in_the_same_scope_as_block_scoped_declaration_1:n(2481,e.DiagnosticCategory.Error,"Cannot_initialize_outer_scoped_variable_0_in_the_same_scope_as_block_scoped_declaration_1_2481","Cannot initialize outer scoped variable '{0}' in the same scope as block scoped declaration '{1}'."),The_left_hand_side_of_a_for_of_statement_cannot_use_a_type_annotation:n(2483,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_of_statement_cannot_use_a_type_annotation_2483","The left-hand side of a 'for...of' statement cannot use a type annotation."),Export_declaration_conflicts_with_exported_declaration_of_0:n(2484,e.DiagnosticCategory.Error,"Export_declaration_conflicts_with_exported_declaration_of_0_2484","Export declaration conflicts with exported declaration of '{0}'."),The_left_hand_side_of_a_for_of_statement_must_be_a_variable_or_a_property_access:n(2487,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_of_statement_must_be_a_variable_or_a_property_access_2487","The left-hand side of a 'for...of' statement must be a variable or a property access."),Type_0_must_have_a_Symbol_iterator_method_that_returns_an_iterator:n(2488,e.DiagnosticCategory.Error,"Type_0_must_have_a_Symbol_iterator_method_that_returns_an_iterator_2488","Type '{0}' must have a '[Symbol.iterator]()' method that returns an iterator."),An_iterator_must_have_a_next_method:n(2489,e.DiagnosticCategory.Error,"An_iterator_must_have_a_next_method_2489","An iterator must have a 'next()' method."),The_type_returned_by_the_next_method_of_an_iterator_must_have_a_value_property:n(2490,e.DiagnosticCategory.Error,"The_type_returned_by_the_next_method_of_an_iterator_must_have_a_value_property_2490","The type returned by the 'next()' method of an iterator must have a 'value' property."),The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern:n(2491,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern_2491","The left-hand side of a 'for...in' statement cannot be a destructuring pattern."),Cannot_redeclare_identifier_0_in_catch_clause:n(2492,e.DiagnosticCategory.Error,"Cannot_redeclare_identifier_0_in_catch_clause_2492","Cannot redeclare identifier '{0}' in catch clause."),Tuple_type_0_of_length_1_has_no_element_at_index_2:n(2493,e.DiagnosticCategory.Error,"Tuple_type_0_of_length_1_has_no_element_at_index_2_2493","Tuple type '{0}' of length '{1}' has no element at index '{2}'."),Using_a_string_in_a_for_of_statement_is_only_supported_in_ECMAScript_5_and_higher:n(2494,e.DiagnosticCategory.Error,"Using_a_string_in_a_for_of_statement_is_only_supported_in_ECMAScript_5_and_higher_2494","Using a string in a 'for...of' statement is only supported in ECMAScript 5 and higher."),Type_0_is_not_an_array_type_or_a_string_type:n(2495,e.DiagnosticCategory.Error,"Type_0_is_not_an_array_type_or_a_string_type_2495","Type '{0}' is not an array type or a string type."),The_arguments_object_cannot_be_referenced_in_an_arrow_function_in_ES3_and_ES5_Consider_using_a_standard_function_expression:n(2496,e.DiagnosticCategory.Error,"The_arguments_object_cannot_be_referenced_in_an_arrow_function_in_ES3_and_ES5_Consider_using_a_stand_2496","The 'arguments' object cannot be referenced in an arrow function in ES3 and ES5. Consider using a standard function expression."),This_module_can_only_be_referenced_with_ECMAScript_imports_Slashexports_by_turning_on_the_0_flag_and_referencing_its_default_export:n(2497,e.DiagnosticCategory.Error,"This_module_can_only_be_referenced_with_ECMAScript_imports_Slashexports_by_turning_on_the_0_flag_and_2497","This module can only be referenced with ECMAScript imports/exports by turning on the '{0}' flag and referencing its default export."),Module_0_uses_export_and_cannot_be_used_with_export_Asterisk:n(2498,e.DiagnosticCategory.Error,"Module_0_uses_export_and_cannot_be_used_with_export_Asterisk_2498","Module '{0}' uses 'export =' and cannot be used with 'export *'."),An_interface_can_only_extend_an_identifier_Slashqualified_name_with_optional_type_arguments:n(2499,e.DiagnosticCategory.Error,"An_interface_can_only_extend_an_identifier_Slashqualified_name_with_optional_type_arguments_2499","An interface can only extend an identifier/qualified-name with optional type arguments."),A_class_can_only_implement_an_identifier_Slashqualified_name_with_optional_type_arguments:n(2500,e.DiagnosticCategory.Error,"A_class_can_only_implement_an_identifier_Slashqualified_name_with_optional_type_arguments_2500","A class can only implement an identifier/qualified-name with optional type arguments."),A_rest_element_cannot_contain_a_binding_pattern:n(2501,e.DiagnosticCategory.Error,"A_rest_element_cannot_contain_a_binding_pattern_2501","A rest element cannot contain a binding pattern."),_0_is_referenced_directly_or_indirectly_in_its_own_type_annotation:n(2502,e.DiagnosticCategory.Error,"_0_is_referenced_directly_or_indirectly_in_its_own_type_annotation_2502","'{0}' is referenced directly or indirectly in its own type annotation."),Cannot_find_namespace_0:n(2503,e.DiagnosticCategory.Error,"Cannot_find_namespace_0_2503","Cannot find namespace '{0}'."),Type_0_must_have_a_Symbol_asyncIterator_method_that_returns_an_async_iterator:n(2504,e.DiagnosticCategory.Error,"Type_0_must_have_a_Symbol_asyncIterator_method_that_returns_an_async_iterator_2504","Type '{0}' must have a '[Symbol.asyncIterator]()' method that returns an async iterator."),A_generator_cannot_have_a_void_type_annotation:n(2505,e.DiagnosticCategory.Error,"A_generator_cannot_have_a_void_type_annotation_2505","A generator cannot have a 'void' type annotation."),_0_is_referenced_directly_or_indirectly_in_its_own_base_expression:n(2506,e.DiagnosticCategory.Error,"_0_is_referenced_directly_or_indirectly_in_its_own_base_expression_2506","'{0}' is referenced directly or indirectly in its own base expression."),Type_0_is_not_a_constructor_function_type:n(2507,e.DiagnosticCategory.Error,"Type_0_is_not_a_constructor_function_type_2507","Type '{0}' is not a constructor function type."),No_base_constructor_has_the_specified_number_of_type_arguments:n(2508,e.DiagnosticCategory.Error,"No_base_constructor_has_the_specified_number_of_type_arguments_2508","No base constructor has the specified number of type arguments."),Base_constructor_return_type_0_is_not_an_object_type_or_intersection_of_object_types_with_statically_known_members:n(2509,e.DiagnosticCategory.Error,"Base_constructor_return_type_0_is_not_an_object_type_or_intersection_of_object_types_with_statically_2509","Base constructor return type '{0}' is not an object type or intersection of object types with statically known members."),Base_constructors_must_all_have_the_same_return_type:n(2510,e.DiagnosticCategory.Error,"Base_constructors_must_all_have_the_same_return_type_2510","Base constructors must all have the same return type."),Cannot_create_an_instance_of_an_abstract_class:n(2511,e.DiagnosticCategory.Error,"Cannot_create_an_instance_of_an_abstract_class_2511","Cannot create an instance of an abstract class."),Overload_signatures_must_all_be_abstract_or_non_abstract:n(2512,e.DiagnosticCategory.Error,"Overload_signatures_must_all_be_abstract_or_non_abstract_2512","Overload signatures must all be abstract or non-abstract."),Abstract_method_0_in_class_1_cannot_be_accessed_via_super_expression:n(2513,e.DiagnosticCategory.Error,"Abstract_method_0_in_class_1_cannot_be_accessed_via_super_expression_2513","Abstract method '{0}' in class '{1}' cannot be accessed via super expression."),Classes_containing_abstract_methods_must_be_marked_abstract:n(2514,e.DiagnosticCategory.Error,"Classes_containing_abstract_methods_must_be_marked_abstract_2514","Classes containing abstract methods must be marked abstract."),Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2:n(2515,e.DiagnosticCategory.Error,"Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2_2515","Non-abstract class '{0}' does not implement inherited abstract member '{1}' from class '{2}'."),All_declarations_of_an_abstract_method_must_be_consecutive:n(2516,e.DiagnosticCategory.Error,"All_declarations_of_an_abstract_method_must_be_consecutive_2516","All declarations of an abstract method must be consecutive."),Cannot_assign_an_abstract_constructor_type_to_a_non_abstract_constructor_type:n(2517,e.DiagnosticCategory.Error,"Cannot_assign_an_abstract_constructor_type_to_a_non_abstract_constructor_type_2517","Cannot assign an abstract constructor type to a non-abstract constructor type."),A_this_based_type_guard_is_not_compatible_with_a_parameter_based_type_guard:n(2518,e.DiagnosticCategory.Error,"A_this_based_type_guard_is_not_compatible_with_a_parameter_based_type_guard_2518","A 'this'-based type guard is not compatible with a parameter-based type guard."),An_async_iterator_must_have_a_next_method:n(2519,e.DiagnosticCategory.Error,"An_async_iterator_must_have_a_next_method_2519","An async iterator must have a 'next()' method."),Duplicate_identifier_0_Compiler_uses_declaration_1_to_support_async_functions:n(2520,e.DiagnosticCategory.Error,"Duplicate_identifier_0_Compiler_uses_declaration_1_to_support_async_functions_2520","Duplicate identifier '{0}'. Compiler uses declaration '{1}' to support async functions."),Expression_resolves_to_variable_declaration_0_that_compiler_uses_to_support_async_functions:n(2521,e.DiagnosticCategory.Error,"Expression_resolves_to_variable_declaration_0_that_compiler_uses_to_support_async_functions_2521","Expression resolves to variable declaration '{0}' that compiler uses to support async functions."),The_arguments_object_cannot_be_referenced_in_an_async_function_or_method_in_ES3_and_ES5_Consider_using_a_standard_function_or_method:n(2522,e.DiagnosticCategory.Error,"The_arguments_object_cannot_be_referenced_in_an_async_function_or_method_in_ES3_and_ES5_Consider_usi_2522","The 'arguments' object cannot be referenced in an async function or method in ES3 and ES5. Consider using a standard function or method."),yield_expressions_cannot_be_used_in_a_parameter_initializer:n(2523,e.DiagnosticCategory.Error,"yield_expressions_cannot_be_used_in_a_parameter_initializer_2523","'yield' expressions cannot be used in a parameter initializer."),await_expressions_cannot_be_used_in_a_parameter_initializer:n(2524,e.DiagnosticCategory.Error,"await_expressions_cannot_be_used_in_a_parameter_initializer_2524","'await' expressions cannot be used in a parameter initializer."),Initializer_provides_no_value_for_this_binding_element_and_the_binding_element_has_no_default_value:n(2525,e.DiagnosticCategory.Error,"Initializer_provides_no_value_for_this_binding_element_and_the_binding_element_has_no_default_value_2525","Initializer provides no value for this binding element and the binding element has no default value."),A_this_type_is_available_only_in_a_non_static_member_of_a_class_or_interface:n(2526,e.DiagnosticCategory.Error,"A_this_type_is_available_only_in_a_non_static_member_of_a_class_or_interface_2526","A 'this' type is available only in a non-static member of a class or interface."),The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary:n(2527,e.DiagnosticCategory.Error,"The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary_2527","The inferred type of '{0}' references an inaccessible '{1}' type. A type annotation is necessary."),A_module_cannot_have_multiple_default_exports:n(2528,e.DiagnosticCategory.Error,"A_module_cannot_have_multiple_default_exports_2528","A module cannot have multiple default exports."),Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_containing_async_functions:n(2529,e.DiagnosticCategory.Error,"Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_containing_async_func_2529","Duplicate identifier '{0}'. Compiler reserves name '{1}' in top level scope of a module containing async functions."),Property_0_is_incompatible_with_index_signature:n(2530,e.DiagnosticCategory.Error,"Property_0_is_incompatible_with_index_signature_2530","Property '{0}' is incompatible with index signature."),Object_is_possibly_null:n(2531,e.DiagnosticCategory.Error,"Object_is_possibly_null_2531","Object is possibly 'null'."),Object_is_possibly_undefined:n(2532,e.DiagnosticCategory.Error,"Object_is_possibly_undefined_2532","Object is possibly 'undefined'."),Object_is_possibly_null_or_undefined:n(2533,e.DiagnosticCategory.Error,"Object_is_possibly_null_or_undefined_2533","Object is possibly 'null' or 'undefined'."),A_function_returning_never_cannot_have_a_reachable_end_point:n(2534,e.DiagnosticCategory.Error,"A_function_returning_never_cannot_have_a_reachable_end_point_2534","A function returning 'never' cannot have a reachable end point."),Enum_type_0_has_members_with_initializers_that_are_not_literals:n(2535,e.DiagnosticCategory.Error,"Enum_type_0_has_members_with_initializers_that_are_not_literals_2535","Enum type '{0}' has members with initializers that are not literals."),Type_0_cannot_be_used_to_index_type_1:n(2536,e.DiagnosticCategory.Error,"Type_0_cannot_be_used_to_index_type_1_2536","Type '{0}' cannot be used to index type '{1}'."),Type_0_has_no_matching_index_signature_for_type_1:n(2537,e.DiagnosticCategory.Error,"Type_0_has_no_matching_index_signature_for_type_1_2537","Type '{0}' has no matching index signature for type '{1}'."),Type_0_cannot_be_used_as_an_index_type:n(2538,e.DiagnosticCategory.Error,"Type_0_cannot_be_used_as_an_index_type_2538","Type '{0}' cannot be used as an index type."),Cannot_assign_to_0_because_it_is_not_a_variable:n(2539,e.DiagnosticCategory.Error,"Cannot_assign_to_0_because_it_is_not_a_variable_2539","Cannot assign to '{0}' because it is not a variable."),Cannot_assign_to_0_because_it_is_a_read_only_property:n(2540,e.DiagnosticCategory.Error,"Cannot_assign_to_0_because_it_is_a_read_only_property_2540","Cannot assign to '{0}' because it is a read-only property."),The_target_of_an_assignment_must_be_a_variable_or_a_property_access:n(2541,e.DiagnosticCategory.Error,"The_target_of_an_assignment_must_be_a_variable_or_a_property_access_2541","The target of an assignment must be a variable or a property access."),Index_signature_in_type_0_only_permits_reading:n(2542,e.DiagnosticCategory.Error,"Index_signature_in_type_0_only_permits_reading_2542","Index signature in type '{0}' only permits reading."),Duplicate_identifier_newTarget_Compiler_uses_variable_declaration_newTarget_to_capture_new_target_meta_property_reference:n(2543,e.DiagnosticCategory.Error,"Duplicate_identifier_newTarget_Compiler_uses_variable_declaration_newTarget_to_capture_new_target_me_2543","Duplicate identifier '_newTarget'. Compiler uses variable declaration '_newTarget' to capture 'new.target' meta-property reference."),Expression_resolves_to_variable_declaration_newTarget_that_compiler_uses_to_capture_new_target_meta_property_reference:n(2544,e.DiagnosticCategory.Error,"Expression_resolves_to_variable_declaration_newTarget_that_compiler_uses_to_capture_new_target_meta__2544","Expression resolves to variable declaration '_newTarget' that compiler uses to capture 'new.target' meta-property reference."),A_mixin_class_must_have_a_constructor_with_a_single_rest_parameter_of_type_any:n(2545,e.DiagnosticCategory.Error,"A_mixin_class_must_have_a_constructor_with_a_single_rest_parameter_of_type_any_2545","A mixin class must have a constructor with a single rest parameter of type 'any[]'."),Property_0_has_conflicting_declarations_and_is_inaccessible_in_type_1:n(2546,e.DiagnosticCategory.Error,"Property_0_has_conflicting_declarations_and_is_inaccessible_in_type_1_2546","Property '{0}' has conflicting declarations and is inaccessible in type '{1}'."),The_type_returned_by_the_next_method_of_an_async_iterator_must_be_a_promise_for_a_type_with_a_value_property:n(2547,e.DiagnosticCategory.Error,"The_type_returned_by_the_next_method_of_an_async_iterator_must_be_a_promise_for_a_type_with_a_value__2547","The type returned by the 'next()' method of an async iterator must be a promise for a type with a 'value' property."),Type_0_is_not_an_array_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator:n(2548,e.DiagnosticCategory.Error,"Type_0_is_not_an_array_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator_2548","Type '{0}' is not an array type or does not have a '[Symbol.iterator]()' method that returns an iterator."),Type_0_is_not_an_array_type_or_a_string_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator:n(2549,e.DiagnosticCategory.Error,"Type_0_is_not_an_array_type_or_a_string_type_or_does_not_have_a_Symbol_iterator_method_that_returns__2549","Type '{0}' is not an array type or a string type or does not have a '[Symbol.iterator]()' method that returns an iterator."),Property_0_does_not_exist_on_type_1_Did_you_mean_2:n(2551,e.DiagnosticCategory.Error,"Property_0_does_not_exist_on_type_1_Did_you_mean_2_2551","Property '{0}' does not exist on type '{1}'. Did you mean '{2}'?"),Cannot_find_name_0_Did_you_mean_1:n(2552,e.DiagnosticCategory.Error,"Cannot_find_name_0_Did_you_mean_1_2552","Cannot find name '{0}'. Did you mean '{1}'?"),Computed_values_are_not_permitted_in_an_enum_with_string_valued_members:n(2553,e.DiagnosticCategory.Error,"Computed_values_are_not_permitted_in_an_enum_with_string_valued_members_2553","Computed values are not permitted in an enum with string valued members."),Expected_0_arguments_but_got_1:n(2554,e.DiagnosticCategory.Error,"Expected_0_arguments_but_got_1_2554","Expected {0} arguments, but got {1}."),Expected_at_least_0_arguments_but_got_1:n(2555,e.DiagnosticCategory.Error,"Expected_at_least_0_arguments_but_got_1_2555","Expected at least {0} arguments, but got {1}."),Expected_0_arguments_but_got_1_or_more:n(2556,e.DiagnosticCategory.Error,"Expected_0_arguments_but_got_1_or_more_2556","Expected {0} arguments, but got {1} or more."),Expected_at_least_0_arguments_but_got_1_or_more:n(2557,e.DiagnosticCategory.Error,"Expected_at_least_0_arguments_but_got_1_or_more_2557","Expected at least {0} arguments, but got {1} or more."),Expected_0_type_arguments_but_got_1:n(2558,e.DiagnosticCategory.Error,"Expected_0_type_arguments_but_got_1_2558","Expected {0} type arguments, but got {1}."),Type_0_has_no_properties_in_common_with_type_1:n(2559,e.DiagnosticCategory.Error,"Type_0_has_no_properties_in_common_with_type_1_2559","Type '{0}' has no properties in common with type '{1}'."),Value_of_type_0_has_no_properties_in_common_with_type_1_Did_you_mean_to_call_it:n(2560,e.DiagnosticCategory.Error,"Value_of_type_0_has_no_properties_in_common_with_type_1_Did_you_mean_to_call_it_2560","Value of type '{0}' has no properties in common with type '{1}'. Did you mean to call it?"),Object_literal_may_only_specify_known_properties_but_0_does_not_exist_in_type_1_Did_you_mean_to_write_2:n(2561,e.DiagnosticCategory.Error,"Object_literal_may_only_specify_known_properties_but_0_does_not_exist_in_type_1_Did_you_mean_to_writ_2561","Object literal may only specify known properties, but '{0}' does not exist in type '{1}'. Did you mean to write '{2}'?"),Base_class_expressions_cannot_reference_class_type_parameters:n(2562,e.DiagnosticCategory.Error,"Base_class_expressions_cannot_reference_class_type_parameters_2562","Base class expressions cannot reference class type parameters."),The_containing_function_or_module_body_is_too_large_for_control_flow_analysis:n(2563,e.DiagnosticCategory.Error,"The_containing_function_or_module_body_is_too_large_for_control_flow_analysis_2563","The containing function or module body is too large for control flow analysis."),Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor:n(2564,e.DiagnosticCategory.Error,"Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor_2564","Property '{0}' has no initializer and is not definitely assigned in the constructor."),Property_0_is_used_before_being_assigned:n(2565,e.DiagnosticCategory.Error,"Property_0_is_used_before_being_assigned_2565","Property '{0}' is used before being assigned."),A_rest_element_cannot_have_a_property_name:n(2566,e.DiagnosticCategory.Error,"A_rest_element_cannot_have_a_property_name_2566","A rest element cannot have a property name."),Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations:n(2567,e.DiagnosticCategory.Error,"Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations_2567","Enum declarations can only merge with namespace or other enum declarations."),Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterating_of_iterators:n(2569,e.DiagnosticCategory.Error,"Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterati_2569","Type '{0}' is not an array type or a string type. Use compiler option '--downlevelIteration' to allow iterating of iterators."),Property_0_does_not_exist_on_type_1_Did_you_forget_to_use_await:n(2570,e.DiagnosticCategory.Error,"Property_0_does_not_exist_on_type_1_Did_you_forget_to_use_await_2570","Property '{0}' does not exist on type '{1}'. Did you forget to use 'await'?"),Object_is_of_type_unknown:n(2571,e.DiagnosticCategory.Error,"Object_is_of_type_unknown_2571","Object is of type 'unknown'."),Rest_signatures_are_incompatible:n(2572,e.DiagnosticCategory.Error,"Rest_signatures_are_incompatible_2572","Rest signatures are incompatible."),Property_0_is_incompatible_with_rest_element_type:n(2573,e.DiagnosticCategory.Error,"Property_0_is_incompatible_with_rest_element_type_2573","Property '{0}' is incompatible with rest element type."),A_rest_element_type_must_be_an_array_type:n(2574,e.DiagnosticCategory.Error,"A_rest_element_type_must_be_an_array_type_2574","A rest element type must be an array type."),No_overload_expects_0_arguments_but_overloads_do_exist_that_expect_either_1_or_2_arguments:n(2575,e.DiagnosticCategory.Error,"No_overload_expects_0_arguments_but_overloads_do_exist_that_expect_either_1_or_2_arguments_2575","No overload expects {0} arguments, but overloads do exist that expect either {1} or {2} arguments."),Property_0_is_a_static_member_of_type_1:n(2576,e.DiagnosticCategory.Error,"Property_0_is_a_static_member_of_type_1_2576","Property '{0}' is a static member of type '{1}'"),Return_type_annotation_circularly_references_itself:n(2577,e.DiagnosticCategory.Error,"Return_type_annotation_circularly_references_itself_2577","Return type annotation circularly references itself."),Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_types_Slashnode_and_then_add_node_to_the_types_field_in_your_tsconfig:n(2580,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_types_Slashnode_and_th_2580","Cannot find name '{0}'. Do you need to install type definitions for node? Try `npm i @types/node` and then add `node` to the types field in your tsconfig."),Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_types_Slashjquery_and_then_add_jquery_to_the_types_field_in_your_tsconfig:n(2581,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_types_Slashjquery_an_2581","Cannot find name '{0}'. Do you need to install type definitions for jQuery? Try `npm i @types/jquery` and then add `jquery` to the types field in your tsconfig."),Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_types_Slashjest_or_npm_i_types_Slashmocha_and_then_add_jest_or_mocha_to_the_types_field_in_your_tsconfig:n(2582,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_types_Slashje_2582","Cannot find name '{0}'. Do you need to install type definitions for a test runner? Try `npm i @types/jest` or `npm i @types/mocha`  and then add `jest` or `mocha` to the types field in your tsconfig."),Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_es2015_or_later:n(2583,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_2583","Cannot find name '{0}'. Do you need to change your target library? Try changing the `lib` compiler option to es2015 or later."),Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_include_dom:n(2584,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_2584","Cannot find name '{0}'. Do you need to change your target library? Try changing the `lib` compiler option to include 'dom'."),_0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_es2015_or_later:n(2585,e.DiagnosticCategory.Error,"_0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Do_you_need_to_change_your_target_library_2585","'{0}' only refers to a type, but is being used as a value here. Do you need to change your target library? Try changing the `lib` compiler option to es2015 or later."),Enum_type_0_circularly_references_itself:n(2586,e.DiagnosticCategory.Error,"Enum_type_0_circularly_references_itself_2586","Enum type '{0}' circularly references itself."),JSDoc_type_0_circularly_references_itself:n(2587,e.DiagnosticCategory.Error,"JSDoc_type_0_circularly_references_itself_2587","JSDoc type '{0}' circularly references itself."),Cannot_assign_to_0_because_it_is_a_constant:n(2588,e.DiagnosticCategory.Error,"Cannot_assign_to_0_because_it_is_a_constant_2588","Cannot assign to '{0}' because it is a constant."),JSX_element_attributes_type_0_may_not_be_a_union_type:n(2600,e.DiagnosticCategory.Error,"JSX_element_attributes_type_0_may_not_be_a_union_type_2600","JSX element attributes type '{0}' may not be a union type."),The_return_type_of_a_JSX_element_constructor_must_return_an_object_type:n(2601,e.DiagnosticCategory.Error,"The_return_type_of_a_JSX_element_constructor_must_return_an_object_type_2601","The return type of a JSX element constructor must return an object type."),JSX_element_implicitly_has_type_any_because_the_global_type_JSX_Element_does_not_exist:n(2602,e.DiagnosticCategory.Error,"JSX_element_implicitly_has_type_any_because_the_global_type_JSX_Element_does_not_exist_2602","JSX element implicitly has type 'any' because the global type 'JSX.Element' does not exist."),Property_0_in_type_1_is_not_assignable_to_type_2:n(2603,e.DiagnosticCategory.Error,"Property_0_in_type_1_is_not_assignable_to_type_2_2603","Property '{0}' in type '{1}' is not assignable to type '{2}'."),JSX_element_type_0_does_not_have_any_construct_or_call_signatures:n(2604,e.DiagnosticCategory.Error,"JSX_element_type_0_does_not_have_any_construct_or_call_signatures_2604","JSX element type '{0}' does not have any construct or call signatures."),JSX_element_type_0_is_not_a_constructor_function_for_JSX_elements:n(2605,e.DiagnosticCategory.Error,"JSX_element_type_0_is_not_a_constructor_function_for_JSX_elements_2605","JSX element type '{0}' is not a constructor function for JSX elements."),Property_0_of_JSX_spread_attribute_is_not_assignable_to_target_property:n(2606,e.DiagnosticCategory.Error,"Property_0_of_JSX_spread_attribute_is_not_assignable_to_target_property_2606","Property '{0}' of JSX spread attribute is not assignable to target property."),JSX_element_class_does_not_support_attributes_because_it_does_not_have_a_0_property:n(2607,e.DiagnosticCategory.Error,"JSX_element_class_does_not_support_attributes_because_it_does_not_have_a_0_property_2607","JSX element class does not support attributes because it does not have a '{0}' property."),The_global_type_JSX_0_may_not_have_more_than_one_property:n(2608,e.DiagnosticCategory.Error,"The_global_type_JSX_0_may_not_have_more_than_one_property_2608","The global type 'JSX.{0}' may not have more than one property."),JSX_spread_child_must_be_an_array_type:n(2609,e.DiagnosticCategory.Error,"JSX_spread_child_must_be_an_array_type_2609","JSX spread child must be an array type."),Cannot_augment_module_0_with_value_exports_because_it_resolves_to_a_non_module_entity:n(2649,e.DiagnosticCategory.Error,"Cannot_augment_module_0_with_value_exports_because_it_resolves_to_a_non_module_entity_2649","Cannot augment module '{0}' with value exports because it resolves to a non-module entity."),A_member_initializer_in_a_enum_declaration_cannot_reference_members_declared_after_it_including_members_defined_in_other_enums:n(2651,e.DiagnosticCategory.Error,"A_member_initializer_in_a_enum_declaration_cannot_reference_members_declared_after_it_including_memb_2651","A member initializer in a enum declaration cannot reference members declared after it, including members defined in other enums."),Merged_declaration_0_cannot_include_a_default_export_declaration_Consider_adding_a_separate_export_default_0_declaration_instead:n(2652,e.DiagnosticCategory.Error,"Merged_declaration_0_cannot_include_a_default_export_declaration_Consider_adding_a_separate_export_d_2652","Merged declaration '{0}' cannot include a default export declaration. Consider adding a separate 'export default {0}' declaration instead."),Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1:n(2653,e.DiagnosticCategory.Error,"Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1_2653","Non-abstract class expression does not implement inherited abstract member '{0}' from class '{1}'."),Exported_external_package_typings_file_cannot_contain_tripleslash_references_Please_contact_the_package_author_to_update_the_package_definition:n(2654,e.DiagnosticCategory.Error,"Exported_external_package_typings_file_cannot_contain_tripleslash_references_Please_contact_the_pack_2654","Exported external package typings file cannot contain tripleslash references. Please contact the package author to update the package definition."),Exported_external_package_typings_file_0_is_not_a_module_Please_contact_the_package_author_to_update_the_package_definition:n(2656,e.DiagnosticCategory.Error,"Exported_external_package_typings_file_0_is_not_a_module_Please_contact_the_package_author_to_update_2656","Exported external package typings file '{0}' is not a module. Please contact the package author to update the package definition."),JSX_expressions_must_have_one_parent_element:n(2657,e.DiagnosticCategory.Error,"JSX_expressions_must_have_one_parent_element_2657","JSX expressions must have one parent element."),Type_0_provides_no_match_for_the_signature_1:n(2658,e.DiagnosticCategory.Error,"Type_0_provides_no_match_for_the_signature_1_2658","Type '{0}' provides no match for the signature '{1}'."),super_is_only_allowed_in_members_of_object_literal_expressions_when_option_target_is_ES2015_or_higher:n(2659,e.DiagnosticCategory.Error,"super_is_only_allowed_in_members_of_object_literal_expressions_when_option_target_is_ES2015_or_highe_2659","'super' is only allowed in members of object literal expressions when option 'target' is 'ES2015' or higher."),super_can_only_be_referenced_in_members_of_derived_classes_or_object_literal_expressions:n(2660,e.DiagnosticCategory.Error,"super_can_only_be_referenced_in_members_of_derived_classes_or_object_literal_expressions_2660","'super' can only be referenced in members of derived classes or object literal expressions."),Cannot_export_0_Only_local_declarations_can_be_exported_from_a_module:n(2661,e.DiagnosticCategory.Error,"Cannot_export_0_Only_local_declarations_can_be_exported_from_a_module_2661","Cannot export '{0}'. Only local declarations can be exported from a module."),Cannot_find_name_0_Did_you_mean_the_static_member_1_0:n(2662,e.DiagnosticCategory.Error,"Cannot_find_name_0_Did_you_mean_the_static_member_1_0_2662","Cannot find name '{0}'. Did you mean the static member '{1}.{0}'?"),Cannot_find_name_0_Did_you_mean_the_instance_member_this_0:n(2663,e.DiagnosticCategory.Error,"Cannot_find_name_0_Did_you_mean_the_instance_member_this_0_2663","Cannot find name '{0}'. Did you mean the instance member 'this.{0}'?"),Invalid_module_name_in_augmentation_module_0_cannot_be_found:n(2664,e.DiagnosticCategory.Error,"Invalid_module_name_in_augmentation_module_0_cannot_be_found_2664","Invalid module name in augmentation, module '{0}' cannot be found."),Invalid_module_name_in_augmentation_Module_0_resolves_to_an_untyped_module_at_1_which_cannot_be_augmented:n(2665,e.DiagnosticCategory.Error,"Invalid_module_name_in_augmentation_Module_0_resolves_to_an_untyped_module_at_1_which_cannot_be_augm_2665","Invalid module name in augmentation. Module '{0}' resolves to an untyped module at '{1}', which cannot be augmented."),Exports_and_export_assignments_are_not_permitted_in_module_augmentations:n(2666,e.DiagnosticCategory.Error,"Exports_and_export_assignments_are_not_permitted_in_module_augmentations_2666","Exports and export assignments are not permitted in module augmentations."),Imports_are_not_permitted_in_module_augmentations_Consider_moving_them_to_the_enclosing_external_module:n(2667,e.DiagnosticCategory.Error,"Imports_are_not_permitted_in_module_augmentations_Consider_moving_them_to_the_enclosing_external_mod_2667","Imports are not permitted in module augmentations. Consider moving them to the enclosing external module."),export_modifier_cannot_be_applied_to_ambient_modules_and_module_augmentations_since_they_are_always_visible:n(2668,e.DiagnosticCategory.Error,"export_modifier_cannot_be_applied_to_ambient_modules_and_module_augmentations_since_they_are_always__2668","'export' modifier cannot be applied to ambient modules and module augmentations since they are always visible."),Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_declarations:n(2669,e.DiagnosticCategory.Error,"Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_2669","Augmentations for the global scope can only be directly nested in external modules or ambient module declarations."),Augmentations_for_the_global_scope_should_have_declare_modifier_unless_they_appear_in_already_ambient_context:n(2670,e.DiagnosticCategory.Error,"Augmentations_for_the_global_scope_should_have_declare_modifier_unless_they_appear_in_already_ambien_2670","Augmentations for the global scope should have 'declare' modifier unless they appear in already ambient context."),Cannot_augment_module_0_because_it_resolves_to_a_non_module_entity:n(2671,e.DiagnosticCategory.Error,"Cannot_augment_module_0_because_it_resolves_to_a_non_module_entity_2671","Cannot augment module '{0}' because it resolves to a non-module entity."),Cannot_assign_a_0_constructor_type_to_a_1_constructor_type:n(2672,e.DiagnosticCategory.Error,"Cannot_assign_a_0_constructor_type_to_a_1_constructor_type_2672","Cannot assign a '{0}' constructor type to a '{1}' constructor type."),Constructor_of_class_0_is_private_and_only_accessible_within_the_class_declaration:n(2673,e.DiagnosticCategory.Error,"Constructor_of_class_0_is_private_and_only_accessible_within_the_class_declaration_2673","Constructor of class '{0}' is private and only accessible within the class declaration."),Constructor_of_class_0_is_protected_and_only_accessible_within_the_class_declaration:n(2674,e.DiagnosticCategory.Error,"Constructor_of_class_0_is_protected_and_only_accessible_within_the_class_declaration_2674","Constructor of class '{0}' is protected and only accessible within the class declaration."),Cannot_extend_a_class_0_Class_constructor_is_marked_as_private:n(2675,e.DiagnosticCategory.Error,"Cannot_extend_a_class_0_Class_constructor_is_marked_as_private_2675","Cannot extend a class '{0}'. Class constructor is marked as private."),Accessors_must_both_be_abstract_or_non_abstract:n(2676,e.DiagnosticCategory.Error,"Accessors_must_both_be_abstract_or_non_abstract_2676","Accessors must both be abstract or non-abstract."),A_type_predicate_s_type_must_be_assignable_to_its_parameter_s_type:n(2677,e.DiagnosticCategory.Error,"A_type_predicate_s_type_must_be_assignable_to_its_parameter_s_type_2677","A type predicate's type must be assignable to its parameter's type."),Type_0_is_not_comparable_to_type_1:n(2678,e.DiagnosticCategory.Error,"Type_0_is_not_comparable_to_type_1_2678","Type '{0}' is not comparable to type '{1}'."),A_function_that_is_called_with_the_new_keyword_cannot_have_a_this_type_that_is_void:n(2679,e.DiagnosticCategory.Error,"A_function_that_is_called_with_the_new_keyword_cannot_have_a_this_type_that_is_void_2679","A function that is called with the 'new' keyword cannot have a 'this' type that is 'void'."),A_0_parameter_must_be_the_first_parameter:n(2680,e.DiagnosticCategory.Error,"A_0_parameter_must_be_the_first_parameter_2680","A '{0}' parameter must be the first parameter."),A_constructor_cannot_have_a_this_parameter:n(2681,e.DiagnosticCategory.Error,"A_constructor_cannot_have_a_this_parameter_2681","A constructor cannot have a 'this' parameter."),get_and_set_accessor_must_have_the_same_this_type:n(2682,e.DiagnosticCategory.Error,"get_and_set_accessor_must_have_the_same_this_type_2682","'get' and 'set' accessor must have the same 'this' type."),this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation:n(2683,e.DiagnosticCategory.Error,"this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_2683","'this' implicitly has type 'any' because it does not have a type annotation."),The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1:n(2684,e.DiagnosticCategory.Error,"The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1_2684","The 'this' context of type '{0}' is not assignable to method's 'this' of type '{1}'."),The_this_types_of_each_signature_are_incompatible:n(2685,e.DiagnosticCategory.Error,"The_this_types_of_each_signature_are_incompatible_2685","The 'this' types of each signature are incompatible."),_0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead:n(2686,e.DiagnosticCategory.Error,"_0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead_2686","'{0}' refers to a UMD global, but the current file is a module. Consider adding an import instead."),All_declarations_of_0_must_have_identical_modifiers:n(2687,e.DiagnosticCategory.Error,"All_declarations_of_0_must_have_identical_modifiers_2687","All declarations of '{0}' must have identical modifiers."),Cannot_find_type_definition_file_for_0:n(2688,e.DiagnosticCategory.Error,"Cannot_find_type_definition_file_for_0_2688","Cannot find type definition file for '{0}'."),Cannot_extend_an_interface_0_Did_you_mean_implements:n(2689,e.DiagnosticCategory.Error,"Cannot_extend_an_interface_0_Did_you_mean_implements_2689","Cannot extend an interface '{0}'. Did you mean 'implements'?"),An_import_path_cannot_end_with_a_0_extension_Consider_importing_1_instead:n(2691,e.DiagnosticCategory.Error,"An_import_path_cannot_end_with_a_0_extension_Consider_importing_1_instead_2691","An import path cannot end with a '{0}' extension. Consider importing '{1}' instead."),_0_is_a_primitive_but_1_is_a_wrapper_object_Prefer_using_0_when_possible:n(2692,e.DiagnosticCategory.Error,"_0_is_a_primitive_but_1_is_a_wrapper_object_Prefer_using_0_when_possible_2692","'{0}' is a primitive, but '{1}' is a wrapper object. Prefer using '{0}' when possible."),_0_only_refers_to_a_type_but_is_being_used_as_a_value_here:n(2693,e.DiagnosticCategory.Error,"_0_only_refers_to_a_type_but_is_being_used_as_a_value_here_2693","'{0}' only refers to a type, but is being used as a value here."),Namespace_0_has_no_exported_member_1:n(2694,e.DiagnosticCategory.Error,"Namespace_0_has_no_exported_member_1_2694","Namespace '{0}' has no exported member '{1}'."),Left_side_of_comma_operator_is_unused_and_has_no_side_effects:n(2695,e.DiagnosticCategory.Error,"Left_side_of_comma_operator_is_unused_and_has_no_side_effects_2695","Left side of comma operator is unused and has no side effects.",!0),The_Object_type_is_assignable_to_very_few_other_types_Did_you_mean_to_use_the_any_type_instead:n(2696,e.DiagnosticCategory.Error,"The_Object_type_is_assignable_to_very_few_other_types_Did_you_mean_to_use_the_any_type_instead_2696","The 'Object' type is assignable to very few other types. Did you mean to use the 'any' type instead?"),An_async_function_or_method_must_return_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option:n(2697,e.DiagnosticCategory.Error,"An_async_function_or_method_must_return_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_in_2697","An async function or method must return a 'Promise'. Make sure you have a declaration for 'Promise' or include 'ES2015' in your `--lib` option."),Spread_types_may_only_be_created_from_object_types:n(2698,e.DiagnosticCategory.Error,"Spread_types_may_only_be_created_from_object_types_2698","Spread types may only be created from object types."),Static_property_0_conflicts_with_built_in_property_Function_0_of_constructor_function_1:n(2699,e.DiagnosticCategory.Error,"Static_property_0_conflicts_with_built_in_property_Function_0_of_constructor_function_1_2699","Static property '{0}' conflicts with built-in property 'Function.{0}' of constructor function '{1}'."),Rest_types_may_only_be_created_from_object_types:n(2700,e.DiagnosticCategory.Error,"Rest_types_may_only_be_created_from_object_types_2700","Rest types may only be created from object types."),The_target_of_an_object_rest_assignment_must_be_a_variable_or_a_property_access:n(2701,e.DiagnosticCategory.Error,"The_target_of_an_object_rest_assignment_must_be_a_variable_or_a_property_access_2701","The target of an object rest assignment must be a variable or a property access."),_0_only_refers_to_a_type_but_is_being_used_as_a_namespace_here:n(2702,e.DiagnosticCategory.Error,"_0_only_refers_to_a_type_but_is_being_used_as_a_namespace_here_2702","'{0}' only refers to a type, but is being used as a namespace here."),The_operand_of_a_delete_operator_must_be_a_property_reference:n(2703,e.DiagnosticCategory.Error,"The_operand_of_a_delete_operator_must_be_a_property_reference_2703","The operand of a delete operator must be a property reference."),The_operand_of_a_delete_operator_cannot_be_a_read_only_property:n(2704,e.DiagnosticCategory.Error,"The_operand_of_a_delete_operator_cannot_be_a_read_only_property_2704","The operand of a delete operator cannot be a read-only property."),An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option:n(2705,e.DiagnosticCategory.Error,"An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_de_2705","An async function or method in ES5/ES3 requires the 'Promise' constructor.  Make sure you have a declaration for the 'Promise' constructor or include 'ES2015' in your `--lib` option."),Required_type_parameters_may_not_follow_optional_type_parameters:n(2706,e.DiagnosticCategory.Error,"Required_type_parameters_may_not_follow_optional_type_parameters_2706","Required type parameters may not follow optional type parameters."),Generic_type_0_requires_between_1_and_2_type_arguments:n(2707,e.DiagnosticCategory.Error,"Generic_type_0_requires_between_1_and_2_type_arguments_2707","Generic type '{0}' requires between {1} and {2} type arguments."),Cannot_use_namespace_0_as_a_value:n(2708,e.DiagnosticCategory.Error,"Cannot_use_namespace_0_as_a_value_2708","Cannot use namespace '{0}' as a value."),Cannot_use_namespace_0_as_a_type:n(2709,e.DiagnosticCategory.Error,"Cannot_use_namespace_0_as_a_type_2709","Cannot use namespace '{0}' as a type."),_0_are_specified_twice_The_attribute_named_0_will_be_overwritten:n(2710,e.DiagnosticCategory.Error,"_0_are_specified_twice_The_attribute_named_0_will_be_overwritten_2710","'{0}' are specified twice. The attribute named '{0}' will be overwritten."),A_dynamic_import_call_returns_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option:n(2711,e.DiagnosticCategory.Error,"A_dynamic_import_call_returns_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES20_2711","A dynamic import call returns a 'Promise'. Make sure you have a declaration for 'Promise' or include 'ES2015' in your `--lib` option."),A_dynamic_import_call_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option:n(2712,e.DiagnosticCategory.Error,"A_dynamic_import_call_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declarat_2712","A dynamic import call in ES5/ES3 requires the 'Promise' constructor.  Make sure you have a declaration for the 'Promise' constructor or include 'ES2015' in your `--lib` option."),Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_property_1_in_0_with_0_1:n(2713,e.DiagnosticCategory.Error,"Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_p_2713","Cannot access '{0}.{1}' because '{0}' is a type, but not a namespace. Did you mean to retrieve the type of the property '{1}' in '{0}' with '{0}[\"{1}\"]'?"),The_expression_of_an_export_assignment_must_be_an_identifier_or_qualified_name_in_an_ambient_context:n(2714,e.DiagnosticCategory.Error,"The_expression_of_an_export_assignment_must_be_an_identifier_or_qualified_name_in_an_ambient_context_2714","The expression of an export assignment must be an identifier or qualified name in an ambient context."),Abstract_property_0_in_class_1_cannot_be_accessed_in_the_constructor:n(2715,e.DiagnosticCategory.Error,"Abstract_property_0_in_class_1_cannot_be_accessed_in_the_constructor_2715","Abstract property '{0}' in class '{1}' cannot be accessed in the constructor."),Type_parameter_0_has_a_circular_default:n(2716,e.DiagnosticCategory.Error,"Type_parameter_0_has_a_circular_default_2716","Type parameter '{0}' has a circular default."),Subsequent_property_declarations_must_have_the_same_type_Property_0_must_be_of_type_1_but_here_has_type_2:n(2717,e.DiagnosticCategory.Error,"Subsequent_property_declarations_must_have_the_same_type_Property_0_must_be_of_type_1_but_here_has_t_2717","Subsequent property declarations must have the same type.  Property '{0}' must be of type '{1}', but here has type '{2}'."),Duplicate_property_0:n(2718,e.DiagnosticCategory.Error,"Duplicate_property_0_2718","Duplicate property '{0}'."),Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated:n(2719,e.DiagnosticCategory.Error,"Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated_2719","Type '{0}' is not assignable to type '{1}'. Two different types with this name exist, but they are unrelated."),Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass:n(2720,e.DiagnosticCategory.Error,"Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclas_2720","Class '{0}' incorrectly implements class '{1}'. Did you mean to extend '{1}' and inherit its members as a subclass?"),Cannot_invoke_an_object_which_is_possibly_null:n(2721,e.DiagnosticCategory.Error,"Cannot_invoke_an_object_which_is_possibly_null_2721","Cannot invoke an object which is possibly 'null'."),Cannot_invoke_an_object_which_is_possibly_undefined:n(2722,e.DiagnosticCategory.Error,"Cannot_invoke_an_object_which_is_possibly_undefined_2722","Cannot invoke an object which is possibly 'undefined'."),Cannot_invoke_an_object_which_is_possibly_null_or_undefined:n(2723,e.DiagnosticCategory.Error,"Cannot_invoke_an_object_which_is_possibly_null_or_undefined_2723","Cannot invoke an object which is possibly 'null' or 'undefined'."),Module_0_has_no_exported_member_1_Did_you_mean_2:n(2724,e.DiagnosticCategory.Error,"Module_0_has_no_exported_member_1_Did_you_mean_2_2724","Module '{0}' has no exported member '{1}'. Did you mean '{2}'?"),Class_name_cannot_be_Object_when_targeting_ES5_with_module_0:n(2725,e.DiagnosticCategory.Error,"Class_name_cannot_be_Object_when_targeting_ES5_with_module_0_2725","Class name cannot be 'Object' when targeting ES5 with module {0}."),Cannot_find_lib_definition_for_0:n(2726,e.DiagnosticCategory.Error,"Cannot_find_lib_definition_for_0_2726","Cannot find lib definition for '{0}'."),Cannot_find_lib_definition_for_0_Did_you_mean_1:n(2727,e.DiagnosticCategory.Error,"Cannot_find_lib_definition_for_0_Did_you_mean_1_2727","Cannot find lib definition for '{0}'. Did you mean '{1}'?"),_0_is_declared_here:n(2728,e.DiagnosticCategory.Message,"_0_is_declared_here_2728","'{0}' is declared here."),Property_0_is_used_before_its_initialization:n(2729,e.DiagnosticCategory.Error,"Property_0_is_used_before_its_initialization_2729","Property '{0}' is used before its initialization."),An_arrow_function_cannot_have_a_this_parameter:n(2730,e.DiagnosticCategory.Error,"An_arrow_function_cannot_have_a_this_parameter_2730","An arrow function cannot have a 'this' parameter."),Implicit_conversion_of_a_symbol_to_a_string_will_fail_at_runtime_Consider_wrapping_this_expression_in_String:n(2731,e.DiagnosticCategory.Error,"Implicit_conversion_of_a_symbol_to_a_string_will_fail_at_runtime_Consider_wrapping_this_expression_i_2731","Implicit conversion of a 'symbol' to a 'string' will fail at runtime. Consider wrapping this expression in 'String(...)'."),Cannot_find_module_0_Consider_using_resolveJsonModule_to_import_module_with_json_extension:n(2732,e.DiagnosticCategory.Error,"Cannot_find_module_0_Consider_using_resolveJsonModule_to_import_module_with_json_extension_2732","Cannot find module '{0}'. Consider using '--resolveJsonModule' to import module with '.json' extension"),Property_0_was_also_declared_here:n(2733,e.DiagnosticCategory.Error,"Property_0_was_also_declared_here_2733","Property '{0}' was also declared here."),It_is_highly_likely_that_you_are_missing_a_semicolon:n(2734,e.DiagnosticCategory.Error,"It_is_highly_likely_that_you_are_missing_a_semicolon_2734","It is highly likely that you are missing a semicolon."),Did_you_mean_for_0_to_be_constrained_to_type_new_args_Colon_any_1:n(2735,e.DiagnosticCategory.Error,"Did_you_mean_for_0_to_be_constrained_to_type_new_args_Colon_any_1_2735","Did you mean for '{0}' to be constrained to type 'new (...args: any[]) => {1}'?"),Operator_0_cannot_be_applied_to_type_1:n(2736,e.DiagnosticCategory.Error,"Operator_0_cannot_be_applied_to_type_1_2736","Operator '{0}' cannot be applied to type '{1}'."),BigInt_literals_are_not_available_when_targeting_lower_than_ESNext:n(2737,e.DiagnosticCategory.Error,"BigInt_literals_are_not_available_when_targeting_lower_than_ESNext_2737","BigInt literals are not available when targeting lower than ESNext."),An_outer_value_of_this_is_shadowed_by_this_container:n(2738,e.DiagnosticCategory.Message,"An_outer_value_of_this_is_shadowed_by_this_container_2738","An outer value of 'this' is shadowed by this container."),Type_0_is_missing_the_following_properties_from_type_1_Colon_2:n(2739,e.DiagnosticCategory.Error,"Type_0_is_missing_the_following_properties_from_type_1_Colon_2_2739","Type '{0}' is missing the following properties from type '{1}': {2}"),Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more:n(2740,e.DiagnosticCategory.Error,"Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more_2740","Type '{0}' is missing the following properties from type '{1}': {2}, and {3} more."),Property_0_is_missing_in_type_1_but_required_in_type_2:n(2741,e.DiagnosticCategory.Error,"Property_0_is_missing_in_type_1_but_required_in_type_2_2741","Property '{0}' is missing in type '{1}' but required in type '{2}'."),The_inferred_type_of_0_cannot_be_named_without_a_reference_to_1_This_is_likely_not_portable_A_type_annotation_is_necessary:n(2742,e.DiagnosticCategory.Error,"The_inferred_type_of_0_cannot_be_named_without_a_reference_to_1_This_is_likely_not_portable_A_type_a_2742","The inferred type of '{0}' cannot be named without a reference to '{1}'. This is likely not portable. A type annotation is necessary."),No_overload_expects_0_type_arguments_but_overloads_do_exist_that_expect_either_1_or_2_type_arguments:n(2743,e.DiagnosticCategory.Error,"No_overload_expects_0_type_arguments_but_overloads_do_exist_that_expect_either_1_or_2_type_arguments_2743","No overload expects {0} type arguments, but overloads do exist that expect either {1} or {2} type arguments."),Type_parameter_defaults_can_only_reference_previously_declared_type_parameters:n(2744,e.DiagnosticCategory.Error,"Type_parameter_defaults_can_only_reference_previously_declared_type_parameters_2744","Type parameter defaults can only reference previously declared type parameters."),This_JSX_tag_s_0_prop_expects_type_1_which_requires_multiple_children_but_only_a_single_child_was_provided:n(2745,e.DiagnosticCategory.Error,"This_JSX_tag_s_0_prop_expects_type_1_which_requires_multiple_children_but_only_a_single_child_was_pr_2745","This JSX tag's '{0}' prop expects type '{1}' which requires multiple children, but only a single child was provided."),This_JSX_tag_s_0_prop_expects_a_single_child_of_type_1_but_multiple_children_were_provided:n(2746,e.DiagnosticCategory.Error,"This_JSX_tag_s_0_prop_expects_a_single_child_of_type_1_but_multiple_children_were_provided_2746","This JSX tag's '{0}' prop expects a single child of type '{1}', but multiple children were provided."),_0_components_don_t_accept_text_as_child_elements_Text_in_JSX_has_the_type_string_but_the_expected_type_of_1_is_2:n(2747,e.DiagnosticCategory.Error,"_0_components_don_t_accept_text_as_child_elements_Text_in_JSX_has_the_type_string_but_the_expected_t_2747","'{0}' components don't accept text as child elements. Text in JSX has the type 'string', but the expected type of '{1}' is '{2}'."),Import_declaration_0_is_using_private_name_1:n(4e3,e.DiagnosticCategory.Error,"Import_declaration_0_is_using_private_name_1_4000","Import declaration '{0}' is using private name '{1}'."),Type_parameter_0_of_exported_class_has_or_is_using_private_name_1:n(4002,e.DiagnosticCategory.Error,"Type_parameter_0_of_exported_class_has_or_is_using_private_name_1_4002","Type parameter '{0}' of exported class has or is using private name '{1}'."),Type_parameter_0_of_exported_interface_has_or_is_using_private_name_1:n(4004,e.DiagnosticCategory.Error,"Type_parameter_0_of_exported_interface_has_or_is_using_private_name_1_4004","Type parameter '{0}' of exported interface has or is using private name '{1}'."),Type_parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1:n(4006,e.DiagnosticCategory.Error,"Type_parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1_4006","Type parameter '{0}' of constructor signature from exported interface has or is using private name '{1}'."),Type_parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1:n(4008,e.DiagnosticCategory.Error,"Type_parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1_4008","Type parameter '{0}' of call signature from exported interface has or is using private name '{1}'."),Type_parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1:n(4010,e.DiagnosticCategory.Error,"Type_parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1_4010","Type parameter '{0}' of public static method from exported class has or is using private name '{1}'."),Type_parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1:n(4012,e.DiagnosticCategory.Error,"Type_parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1_4012","Type parameter '{0}' of public method from exported class has or is using private name '{1}'."),Type_parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1:n(4014,e.DiagnosticCategory.Error,"Type_parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1_4014","Type parameter '{0}' of method from exported interface has or is using private name '{1}'."),Type_parameter_0_of_exported_function_has_or_is_using_private_name_1:n(4016,e.DiagnosticCategory.Error,"Type_parameter_0_of_exported_function_has_or_is_using_private_name_1_4016","Type parameter '{0}' of exported function has or is using private name '{1}'."),Implements_clause_of_exported_class_0_has_or_is_using_private_name_1:n(4019,e.DiagnosticCategory.Error,"Implements_clause_of_exported_class_0_has_or_is_using_private_name_1_4019","Implements clause of exported class '{0}' has or is using private name '{1}'."),extends_clause_of_exported_class_0_has_or_is_using_private_name_1:n(4020,e.DiagnosticCategory.Error,"extends_clause_of_exported_class_0_has_or_is_using_private_name_1_4020","'extends' clause of exported class '{0}' has or is using private name '{1}'."),extends_clause_of_exported_interface_0_has_or_is_using_private_name_1:n(4022,e.DiagnosticCategory.Error,"extends_clause_of_exported_interface_0_has_or_is_using_private_name_1_4022","'extends' clause of exported interface '{0}' has or is using private name '{1}'."),Exported_variable_0_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4023,e.DiagnosticCategory.Error,"Exported_variable_0_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named_4023","Exported variable '{0}' has or is using name '{1}' from external module {2} but cannot be named."),Exported_variable_0_has_or_is_using_name_1_from_private_module_2:n(4024,e.DiagnosticCategory.Error,"Exported_variable_0_has_or_is_using_name_1_from_private_module_2_4024","Exported variable '{0}' has or is using name '{1}' from private module '{2}'."),Exported_variable_0_has_or_is_using_private_name_1:n(4025,e.DiagnosticCategory.Error,"Exported_variable_0_has_or_is_using_private_name_1_4025","Exported variable '{0}' has or is using private name '{1}'."),Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4026,e.DiagnosticCategory.Error,"Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot__4026","Public static property '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:n(4027,e.DiagnosticCategory.Error,"Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4027","Public static property '{0}' of exported class has or is using name '{1}' from private module '{2}'."),Public_static_property_0_of_exported_class_has_or_is_using_private_name_1:n(4028,e.DiagnosticCategory.Error,"Public_static_property_0_of_exported_class_has_or_is_using_private_name_1_4028","Public static property '{0}' of exported class has or is using private name '{1}'."),Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4029,e.DiagnosticCategory.Error,"Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_name_4029","Public property '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:n(4030,e.DiagnosticCategory.Error,"Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4030","Public property '{0}' of exported class has or is using name '{1}' from private module '{2}'."),Public_property_0_of_exported_class_has_or_is_using_private_name_1:n(4031,e.DiagnosticCategory.Error,"Public_property_0_of_exported_class_has_or_is_using_private_name_1_4031","Public property '{0}' of exported class has or is using private name '{1}'."),Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4032,e.DiagnosticCategory.Error,"Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2_4032","Property '{0}' of exported interface has or is using name '{1}' from private module '{2}'."),Property_0_of_exported_interface_has_or_is_using_private_name_1:n(4033,e.DiagnosticCategory.Error,"Property_0_of_exported_interface_has_or_is_using_private_name_1_4033","Property '{0}' of exported interface has or is using private name '{1}'."),Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4034,e.DiagnosticCategory.Error,"Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_name_1_from_private_mod_4034","Parameter type of public static setter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_private_name_1:n(4035,e.DiagnosticCategory.Error,"Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_private_name_1_4035","Parameter type of public static setter '{0}' from exported class has or is using private name '{1}'."),Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4036,e.DiagnosticCategory.Error,"Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2_4036","Parameter type of public setter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_private_name_1:n(4037,e.DiagnosticCategory.Error,"Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_private_name_1_4037","Parameter type of public setter '{0}' from exported class has or is using private name '{1}'."),Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4038,e.DiagnosticCategory.Error,"Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_external_modul_4038","Return type of public static getter '{0}' from exported class has or is using name '{1}' from external module {2} but cannot be named."),Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4039,e.DiagnosticCategory.Error,"Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_4039","Return type of public static getter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_private_name_1:n(4040,e.DiagnosticCategory.Error,"Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_private_name_1_4040","Return type of public static getter '{0}' from exported class has or is using private name '{1}'."),Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4041,e.DiagnosticCategory.Error,"Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_4041","Return type of public getter '{0}' from exported class has or is using name '{1}' from external module {2} but cannot be named."),Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4042,e.DiagnosticCategory.Error,"Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2_4042","Return type of public getter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),Return_type_of_public_getter_0_from_exported_class_has_or_is_using_private_name_1:n(4043,e.DiagnosticCategory.Error,"Return_type_of_public_getter_0_from_exported_class_has_or_is_using_private_name_1_4043","Return type of public getter '{0}' from exported class has or is using private name '{1}'."),Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:n(4044,e.DiagnosticCategory.Error,"Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_name_0_from_private_mod_4044","Return type of constructor signature from exported interface has or is using name '{0}' from private module '{1}'."),Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_0:n(4045,e.DiagnosticCategory.Error,"Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_0_4045","Return type of constructor signature from exported interface has or is using private name '{0}'."),Return_type_of_call_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:n(4046,e.DiagnosticCategory.Error,"Return_type_of_call_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1_4046","Return type of call signature from exported interface has or is using name '{0}' from private module '{1}'."),Return_type_of_call_signature_from_exported_interface_has_or_is_using_private_name_0:n(4047,e.DiagnosticCategory.Error,"Return_type_of_call_signature_from_exported_interface_has_or_is_using_private_name_0_4047","Return type of call signature from exported interface has or is using private name '{0}'."),Return_type_of_index_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:n(4048,e.DiagnosticCategory.Error,"Return_type_of_index_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1_4048","Return type of index signature from exported interface has or is using name '{0}' from private module '{1}'."),Return_type_of_index_signature_from_exported_interface_has_or_is_using_private_name_0:n(4049,e.DiagnosticCategory.Error,"Return_type_of_index_signature_from_exported_interface_has_or_is_using_private_name_0_4049","Return type of index signature from exported interface has or is using private name '{0}'."),Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:n(4050,e.DiagnosticCategory.Error,"Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_external_module__4050","Return type of public static method from exported class has or is using name '{0}' from external module {1} but cannot be named."),Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_private_module_1:n(4051,e.DiagnosticCategory.Error,"Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_private_module_1_4051","Return type of public static method from exported class has or is using name '{0}' from private module '{1}'."),Return_type_of_public_static_method_from_exported_class_has_or_is_using_private_name_0:n(4052,e.DiagnosticCategory.Error,"Return_type_of_public_static_method_from_exported_class_has_or_is_using_private_name_0_4052","Return type of public static method from exported class has or is using private name '{0}'."),Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:n(4053,e.DiagnosticCategory.Error,"Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_c_4053","Return type of public method from exported class has or is using name '{0}' from external module {1} but cannot be named."),Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_private_module_1:n(4054,e.DiagnosticCategory.Error,"Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_private_module_1_4054","Return type of public method from exported class has or is using name '{0}' from private module '{1}'."),Return_type_of_public_method_from_exported_class_has_or_is_using_private_name_0:n(4055,e.DiagnosticCategory.Error,"Return_type_of_public_method_from_exported_class_has_or_is_using_private_name_0_4055","Return type of public method from exported class has or is using private name '{0}'."),Return_type_of_method_from_exported_interface_has_or_is_using_name_0_from_private_module_1:n(4056,e.DiagnosticCategory.Error,"Return_type_of_method_from_exported_interface_has_or_is_using_name_0_from_private_module_1_4056","Return type of method from exported interface has or is using name '{0}' from private module '{1}'."),Return_type_of_method_from_exported_interface_has_or_is_using_private_name_0:n(4057,e.DiagnosticCategory.Error,"Return_type_of_method_from_exported_interface_has_or_is_using_private_name_0_4057","Return type of method from exported interface has or is using private name '{0}'."),Return_type_of_exported_function_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:n(4058,e.DiagnosticCategory.Error,"Return_type_of_exported_function_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named_4058","Return type of exported function has or is using name '{0}' from external module {1} but cannot be named."),Return_type_of_exported_function_has_or_is_using_name_0_from_private_module_1:n(4059,e.DiagnosticCategory.Error,"Return_type_of_exported_function_has_or_is_using_name_0_from_private_module_1_4059","Return type of exported function has or is using name '{0}' from private module '{1}'."),Return_type_of_exported_function_has_or_is_using_private_name_0:n(4060,e.DiagnosticCategory.Error,"Return_type_of_exported_function_has_or_is_using_private_name_0_4060","Return type of exported function has or is using private name '{0}'."),Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4061,e.DiagnosticCategory.Error,"Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_can_4061","Parameter '{0}' of constructor from exported class has or is using name '{1}' from external module {2} but cannot be named."),Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4062,e.DiagnosticCategory.Error,"Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_private_module_2_4062","Parameter '{0}' of constructor from exported class has or is using name '{1}' from private module '{2}'."),Parameter_0_of_constructor_from_exported_class_has_or_is_using_private_name_1:n(4063,e.DiagnosticCategory.Error,"Parameter_0_of_constructor_from_exported_class_has_or_is_using_private_name_1_4063","Parameter '{0}' of constructor from exported class has or is using private name '{1}'."),Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4064,e.DiagnosticCategory.Error,"Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_name_1_from_private_mod_4064","Parameter '{0}' of constructor signature from exported interface has or is using name '{1}' from private module '{2}'."),Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1:n(4065,e.DiagnosticCategory.Error,"Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1_4065","Parameter '{0}' of constructor signature from exported interface has or is using private name '{1}'."),Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4066,e.DiagnosticCategory.Error,"Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2_4066","Parameter '{0}' of call signature from exported interface has or is using name '{1}' from private module '{2}'."),Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1:n(4067,e.DiagnosticCategory.Error,"Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1_4067","Parameter '{0}' of call signature from exported interface has or is using private name '{1}'."),Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4068,e.DiagnosticCategory.Error,"Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_external_module__4068","Parameter '{0}' of public static method from exported class has or is using name '{1}' from external module {2} but cannot be named."),Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4069,e.DiagnosticCategory.Error,"Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_private_module_2_4069","Parameter '{0}' of public static method from exported class has or is using name '{1}' from private module '{2}'."),Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1:n(4070,e.DiagnosticCategory.Error,"Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1_4070","Parameter '{0}' of public static method from exported class has or is using private name '{1}'."),Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4071,e.DiagnosticCategory.Error,"Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_c_4071","Parameter '{0}' of public method from exported class has or is using name '{1}' from external module {2} but cannot be named."),Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4072,e.DiagnosticCategory.Error,"Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_private_module_2_4072","Parameter '{0}' of public method from exported class has or is using name '{1}' from private module '{2}'."),Parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1:n(4073,e.DiagnosticCategory.Error,"Parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1_4073","Parameter '{0}' of public method from exported class has or is using private name '{1}'."),Parameter_0_of_method_from_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4074,e.DiagnosticCategory.Error,"Parameter_0_of_method_from_exported_interface_has_or_is_using_name_1_from_private_module_2_4074","Parameter '{0}' of method from exported interface has or is using name '{1}' from private module '{2}'."),Parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1:n(4075,e.DiagnosticCategory.Error,"Parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1_4075","Parameter '{0}' of method from exported interface has or is using private name '{1}'."),Parameter_0_of_exported_function_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4076,e.DiagnosticCategory.Error,"Parameter_0_of_exported_function_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named_4076","Parameter '{0}' of exported function has or is using name '{1}' from external module {2} but cannot be named."),Parameter_0_of_exported_function_has_or_is_using_name_1_from_private_module_2:n(4077,e.DiagnosticCategory.Error,"Parameter_0_of_exported_function_has_or_is_using_name_1_from_private_module_2_4077","Parameter '{0}' of exported function has or is using name '{1}' from private module '{2}'."),Parameter_0_of_exported_function_has_or_is_using_private_name_1:n(4078,e.DiagnosticCategory.Error,"Parameter_0_of_exported_function_has_or_is_using_private_name_1_4078","Parameter '{0}' of exported function has or is using private name '{1}'."),Exported_type_alias_0_has_or_is_using_private_name_1:n(4081,e.DiagnosticCategory.Error,"Exported_type_alias_0_has_or_is_using_private_name_1_4081","Exported type alias '{0}' has or is using private name '{1}'."),Default_export_of_the_module_has_or_is_using_private_name_0:n(4082,e.DiagnosticCategory.Error,"Default_export_of_the_module_has_or_is_using_private_name_0_4082","Default export of the module has or is using private name '{0}'."),Type_parameter_0_of_exported_type_alias_has_or_is_using_private_name_1:n(4083,e.DiagnosticCategory.Error,"Type_parameter_0_of_exported_type_alias_has_or_is_using_private_name_1_4083","Type parameter '{0}' of exported type alias has or is using private name '{1}'."),Conflicting_definitions_for_0_found_at_1_and_2_Consider_installing_a_specific_version_of_this_library_to_resolve_the_conflict:n(4090,e.DiagnosticCategory.Error,"Conflicting_definitions_for_0_found_at_1_and_2_Consider_installing_a_specific_version_of_this_librar_4090","Conflicting definitions for '{0}' found at '{1}' and '{2}'. Consider installing a specific version of this library to resolve the conflict."),Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4091,e.DiagnosticCategory.Error,"Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2_4091","Parameter '{0}' of index signature from exported interface has or is using name '{1}' from private module '{2}'."),Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_private_name_1:n(4092,e.DiagnosticCategory.Error,"Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_private_name_1_4092","Parameter '{0}' of index signature from exported interface has or is using private name '{1}'."),Property_0_of_exported_class_expression_may_not_be_private_or_protected:n(4094,e.DiagnosticCategory.Error,"Property_0_of_exported_class_expression_may_not_be_private_or_protected_4094","Property '{0}' of exported class expression may not be private or protected."),Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4095,e.DiagnosticCategory.Error,"Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_4095","Public static method '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:n(4096,e.DiagnosticCategory.Error,"Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4096","Public static method '{0}' of exported class has or is using name '{1}' from private module '{2}'."),Public_static_method_0_of_exported_class_has_or_is_using_private_name_1:n(4097,e.DiagnosticCategory.Error,"Public_static_method_0_of_exported_class_has_or_is_using_private_name_1_4097","Public static method '{0}' of exported class has or is using private name '{1}'."),Public_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4098,e.DiagnosticCategory.Error,"Public_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named_4098","Public method '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),Public_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:n(4099,e.DiagnosticCategory.Error,"Public_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4099","Public method '{0}' of exported class has or is using name '{1}' from private module '{2}'."),Public_method_0_of_exported_class_has_or_is_using_private_name_1:n(4100,e.DiagnosticCategory.Error,"Public_method_0_of_exported_class_has_or_is_using_private_name_1_4100","Public method '{0}' of exported class has or is using private name '{1}'."),Method_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4101,e.DiagnosticCategory.Error,"Method_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2_4101","Method '{0}' of exported interface has or is using name '{1}' from private module '{2}'."),Method_0_of_exported_interface_has_or_is_using_private_name_1:n(4102,e.DiagnosticCategory.Error,"Method_0_of_exported_interface_has_or_is_using_private_name_1_4102","Method '{0}' of exported interface has or is using private name '{1}'."),The_current_host_does_not_support_the_0_option:n(5001,e.DiagnosticCategory.Error,"The_current_host_does_not_support_the_0_option_5001","The current host does not support the '{0}' option."),Cannot_find_the_common_subdirectory_path_for_the_input_files:n(5009,e.DiagnosticCategory.Error,"Cannot_find_the_common_subdirectory_path_for_the_input_files_5009","Cannot find the common subdirectory path for the input files."),File_specification_cannot_end_in_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0:n(5010,e.DiagnosticCategory.Error,"File_specification_cannot_end_in_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0_5010","File specification cannot end in a recursive directory wildcard ('**'): '{0}'."),Cannot_read_file_0_Colon_1:n(5012,e.DiagnosticCategory.Error,"Cannot_read_file_0_Colon_1_5012","Cannot read file '{0}': {1}."),Failed_to_parse_file_0_Colon_1:n(5014,e.DiagnosticCategory.Error,"Failed_to_parse_file_0_Colon_1_5014","Failed to parse file '{0}': {1}."),Unknown_compiler_option_0:n(5023,e.DiagnosticCategory.Error,"Unknown_compiler_option_0_5023","Unknown compiler option '{0}'."),Compiler_option_0_requires_a_value_of_type_1:n(5024,e.DiagnosticCategory.Error,"Compiler_option_0_requires_a_value_of_type_1_5024","Compiler option '{0}' requires a value of type {1}."),Could_not_write_file_0_Colon_1:n(5033,e.DiagnosticCategory.Error,"Could_not_write_file_0_Colon_1_5033","Could not write file '{0}': {1}."),Option_project_cannot_be_mixed_with_source_files_on_a_command_line:n(5042,e.DiagnosticCategory.Error,"Option_project_cannot_be_mixed_with_source_files_on_a_command_line_5042","Option 'project' cannot be mixed with source files on a command line."),Option_isolatedModules_can_only_be_used_when_either_option_module_is_provided_or_option_target_is_ES2015_or_higher:n(5047,e.DiagnosticCategory.Error,"Option_isolatedModules_can_only_be_used_when_either_option_module_is_provided_or_option_target_is_ES_5047","Option 'isolatedModules' can only be used when either option '--module' is provided or option 'target' is 'ES2015' or higher."),Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided:n(5051,e.DiagnosticCategory.Error,"Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided_5051","Option '{0} can only be used when either option '--inlineSourceMap' or option '--sourceMap' is provided."),Option_0_cannot_be_specified_without_specifying_option_1:n(5052,e.DiagnosticCategory.Error,"Option_0_cannot_be_specified_without_specifying_option_1_5052","Option '{0}' cannot be specified without specifying option '{1}'."),Option_0_cannot_be_specified_with_option_1:n(5053,e.DiagnosticCategory.Error,"Option_0_cannot_be_specified_with_option_1_5053","Option '{0}' cannot be specified with option '{1}'."),A_tsconfig_json_file_is_already_defined_at_Colon_0:n(5054,e.DiagnosticCategory.Error,"A_tsconfig_json_file_is_already_defined_at_Colon_0_5054","A 'tsconfig.json' file is already defined at: '{0}'."),Cannot_write_file_0_because_it_would_overwrite_input_file:n(5055,e.DiagnosticCategory.Error,"Cannot_write_file_0_because_it_would_overwrite_input_file_5055","Cannot write file '{0}' because it would overwrite input file."),Cannot_write_file_0_because_it_would_be_overwritten_by_multiple_input_files:n(5056,e.DiagnosticCategory.Error,"Cannot_write_file_0_because_it_would_be_overwritten_by_multiple_input_files_5056","Cannot write file '{0}' because it would be overwritten by multiple input files."),Cannot_find_a_tsconfig_json_file_at_the_specified_directory_Colon_0:n(5057,e.DiagnosticCategory.Error,"Cannot_find_a_tsconfig_json_file_at_the_specified_directory_Colon_0_5057","Cannot find a tsconfig.json file at the specified directory: '{0}'."),The_specified_path_does_not_exist_Colon_0:n(5058,e.DiagnosticCategory.Error,"The_specified_path_does_not_exist_Colon_0_5058","The specified path does not exist: '{0}'."),Invalid_value_for_reactNamespace_0_is_not_a_valid_identifier:n(5059,e.DiagnosticCategory.Error,"Invalid_value_for_reactNamespace_0_is_not_a_valid_identifier_5059","Invalid value for '--reactNamespace'. '{0}' is not a valid identifier."),Option_paths_cannot_be_used_without_specifying_baseUrl_option:n(5060,e.DiagnosticCategory.Error,"Option_paths_cannot_be_used_without_specifying_baseUrl_option_5060","Option 'paths' cannot be used without specifying '--baseUrl' option."),Pattern_0_can_have_at_most_one_Asterisk_character:n(5061,e.DiagnosticCategory.Error,"Pattern_0_can_have_at_most_one_Asterisk_character_5061","Pattern '{0}' can have at most one '*' character."),Substitution_0_in_pattern_1_in_can_have_at_most_one_Asterisk_character:n(5062,e.DiagnosticCategory.Error,"Substitution_0_in_pattern_1_in_can_have_at_most_one_Asterisk_character_5062","Substitution '{0}' in pattern '{1}' in can have at most one '*' character."),Substitutions_for_pattern_0_should_be_an_array:n(5063,e.DiagnosticCategory.Error,"Substitutions_for_pattern_0_should_be_an_array_5063","Substitutions for pattern '{0}' should be an array."),Substitution_0_for_pattern_1_has_incorrect_type_expected_string_got_2:n(5064,e.DiagnosticCategory.Error,"Substitution_0_for_pattern_1_has_incorrect_type_expected_string_got_2_5064","Substitution '{0}' for pattern '{1}' has incorrect type, expected 'string', got '{2}'."),File_specification_cannot_contain_a_parent_directory_that_appears_after_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0:n(5065,e.DiagnosticCategory.Error,"File_specification_cannot_contain_a_parent_directory_that_appears_after_a_recursive_directory_wildca_5065","File specification cannot contain a parent directory ('..') that appears after a recursive directory wildcard ('**'): '{0}'."),Substitutions_for_pattern_0_shouldn_t_be_an_empty_array:n(5066,e.DiagnosticCategory.Error,"Substitutions_for_pattern_0_shouldn_t_be_an_empty_array_5066","Substitutions for pattern '{0}' shouldn't be an empty array."),Invalid_value_for_jsxFactory_0_is_not_a_valid_identifier_or_qualified_name:n(5067,e.DiagnosticCategory.Error,"Invalid_value_for_jsxFactory_0_is_not_a_valid_identifier_or_qualified_name_5067","Invalid value for 'jsxFactory'. '{0}' is not a valid identifier or qualified-name."),Adding_a_tsconfig_json_file_will_help_organize_projects_that_contain_both_TypeScript_and_JavaScript_files_Learn_more_at_https_Colon_Slash_Slashaka_ms_Slashtsconfig:n(5068,e.DiagnosticCategory.Error,"Adding_a_tsconfig_json_file_will_help_organize_projects_that_contain_both_TypeScript_and_JavaScript__5068","Adding a tsconfig.json file will help organize projects that contain both TypeScript and JavaScript files. Learn more at https://aka.ms/tsconfig."),Option_0_cannot_be_specified_without_specifying_option_1_or_option_2:n(5069,e.DiagnosticCategory.Error,"Option_0_cannot_be_specified_without_specifying_option_1_or_option_2_5069","Option '{0}' cannot be specified without specifying option '{1}' or option '{2}'."),Option_resolveJsonModule_cannot_be_specified_without_node_module_resolution_strategy:n(5070,e.DiagnosticCategory.Error,"Option_resolveJsonModule_cannot_be_specified_without_node_module_resolution_strategy_5070","Option '--resolveJsonModule' cannot be specified without 'node' module resolution strategy."),Option_resolveJsonModule_can_only_be_specified_when_module_code_generation_is_commonjs_amd_es2015_or_esNext:n(5071,e.DiagnosticCategory.Error,"Option_resolveJsonModule_can_only_be_specified_when_module_code_generation_is_commonjs_amd_es2015_or_5071","Option '--resolveJsonModule' can only be specified when module code generation is 'commonjs', 'amd', 'es2015' or 'esNext'."),Unknown_build_option_0:n(5072,e.DiagnosticCategory.Error,"Unknown_build_option_0_5072","Unknown build option '{0}'."),Build_option_0_requires_a_value_of_type_1:n(5073,e.DiagnosticCategory.Error,"Build_option_0_requires_a_value_of_type_1_5073","Build option '{0}' requires a value of type {1}."),Generates_a_sourcemap_for_each_corresponding_d_ts_file:n(6e3,e.DiagnosticCategory.Message,"Generates_a_sourcemap_for_each_corresponding_d_ts_file_6000","Generates a sourcemap for each corresponding '.d.ts' file."),Concatenate_and_emit_output_to_single_file:n(6001,e.DiagnosticCategory.Message,"Concatenate_and_emit_output_to_single_file_6001","Concatenate and emit output to single file."),Generates_corresponding_d_ts_file:n(6002,e.DiagnosticCategory.Message,"Generates_corresponding_d_ts_file_6002","Generates corresponding '.d.ts' file."),Specify_the_location_where_debugger_should_locate_map_files_instead_of_generated_locations:n(6003,e.DiagnosticCategory.Message,"Specify_the_location_where_debugger_should_locate_map_files_instead_of_generated_locations_6003","Specify the location where debugger should locate map files instead of generated locations."),Specify_the_location_where_debugger_should_locate_TypeScript_files_instead_of_source_locations:n(6004,e.DiagnosticCategory.Message,"Specify_the_location_where_debugger_should_locate_TypeScript_files_instead_of_source_locations_6004","Specify the location where debugger should locate TypeScript files instead of source locations."),Watch_input_files:n(6005,e.DiagnosticCategory.Message,"Watch_input_files_6005","Watch input files."),Redirect_output_structure_to_the_directory:n(6006,e.DiagnosticCategory.Message,"Redirect_output_structure_to_the_directory_6006","Redirect output structure to the directory."),Do_not_erase_const_enum_declarations_in_generated_code:n(6007,e.DiagnosticCategory.Message,"Do_not_erase_const_enum_declarations_in_generated_code_6007","Do not erase const enum declarations in generated code."),Do_not_emit_outputs_if_any_errors_were_reported:n(6008,e.DiagnosticCategory.Message,"Do_not_emit_outputs_if_any_errors_were_reported_6008","Do not emit outputs if any errors were reported."),Do_not_emit_comments_to_output:n(6009,e.DiagnosticCategory.Message,"Do_not_emit_comments_to_output_6009","Do not emit comments to output."),Do_not_emit_outputs:n(6010,e.DiagnosticCategory.Message,"Do_not_emit_outputs_6010","Do not emit outputs."),Allow_default_imports_from_modules_with_no_default_export_This_does_not_affect_code_emit_just_typechecking:n(6011,e.DiagnosticCategory.Message,"Allow_default_imports_from_modules_with_no_default_export_This_does_not_affect_code_emit_just_typech_6011","Allow default imports from modules with no default export. This does not affect code emit, just typechecking."),Skip_type_checking_of_declaration_files:n(6012,e.DiagnosticCategory.Message,"Skip_type_checking_of_declaration_files_6012","Skip type checking of declaration files."),Do_not_resolve_the_real_path_of_symlinks:n(6013,e.DiagnosticCategory.Message,"Do_not_resolve_the_real_path_of_symlinks_6013","Do not resolve the real path of symlinks."),Only_emit_d_ts_declaration_files:n(6014,e.DiagnosticCategory.Message,"Only_emit_d_ts_declaration_files_6014","Only emit '.d.ts' declaration files."),Specify_ECMAScript_target_version_Colon_ES3_default_ES5_ES2015_ES2016_ES2017_ES2018_or_ESNEXT:n(6015,e.DiagnosticCategory.Message,"Specify_ECMAScript_target_version_Colon_ES3_default_ES5_ES2015_ES2016_ES2017_ES2018_or_ESNEXT_6015","Specify ECMAScript target version: 'ES3' (default), 'ES5', 'ES2015', 'ES2016', 'ES2017','ES2018' or 'ESNEXT'."),Specify_module_code_generation_Colon_none_commonjs_amd_system_umd_es2015_or_ESNext:n(6016,e.DiagnosticCategory.Message,"Specify_module_code_generation_Colon_none_commonjs_amd_system_umd_es2015_or_ESNext_6016","Specify module code generation: 'none', 'commonjs', 'amd', 'system', 'umd', 'es2015', or 'ESNext'."),Print_this_message:n(6017,e.DiagnosticCategory.Message,"Print_this_message_6017","Print this message."),Print_the_compiler_s_version:n(6019,e.DiagnosticCategory.Message,"Print_the_compiler_s_version_6019","Print the compiler's version."),Compile_the_project_given_the_path_to_its_configuration_file_or_to_a_folder_with_a_tsconfig_json:n(6020,e.DiagnosticCategory.Message,"Compile_the_project_given_the_path_to_its_configuration_file_or_to_a_folder_with_a_tsconfig_json_6020","Compile the project given the path to its configuration file, or to a folder with a 'tsconfig.json'."),Syntax_Colon_0:n(6023,e.DiagnosticCategory.Message,"Syntax_Colon_0_6023","Syntax: {0}"),options:n(6024,e.DiagnosticCategory.Message,"options_6024","options"),file:n(6025,e.DiagnosticCategory.Message,"file_6025","file"),Examples_Colon_0:n(6026,e.DiagnosticCategory.Message,"Examples_Colon_0_6026","Examples: {0}"),Options_Colon:n(6027,e.DiagnosticCategory.Message,"Options_Colon_6027","Options:"),Version_0:n(6029,e.DiagnosticCategory.Message,"Version_0_6029","Version {0}"),Insert_command_line_options_and_files_from_a_file:n(6030,e.DiagnosticCategory.Message,"Insert_command_line_options_and_files_from_a_file_6030","Insert command line options and files from a file."),Starting_compilation_in_watch_mode:n(6031,e.DiagnosticCategory.Message,"Starting_compilation_in_watch_mode_6031","Starting compilation in watch mode..."),File_change_detected_Starting_incremental_compilation:n(6032,e.DiagnosticCategory.Message,"File_change_detected_Starting_incremental_compilation_6032","File change detected. Starting incremental compilation..."),KIND:n(6034,e.DiagnosticCategory.Message,"KIND_6034","KIND"),FILE:n(6035,e.DiagnosticCategory.Message,"FILE_6035","FILE"),VERSION:n(6036,e.DiagnosticCategory.Message,"VERSION_6036","VERSION"),LOCATION:n(6037,e.DiagnosticCategory.Message,"LOCATION_6037","LOCATION"),DIRECTORY:n(6038,e.DiagnosticCategory.Message,"DIRECTORY_6038","DIRECTORY"),STRATEGY:n(6039,e.DiagnosticCategory.Message,"STRATEGY_6039","STRATEGY"),FILE_OR_DIRECTORY:n(6040,e.DiagnosticCategory.Message,"FILE_OR_DIRECTORY_6040","FILE OR DIRECTORY"),Generates_corresponding_map_file:n(6043,e.DiagnosticCategory.Message,"Generates_corresponding_map_file_6043","Generates corresponding '.map' file."),Compiler_option_0_expects_an_argument:n(6044,e.DiagnosticCategory.Error,"Compiler_option_0_expects_an_argument_6044","Compiler option '{0}' expects an argument."),Unterminated_quoted_string_in_response_file_0:n(6045,e.DiagnosticCategory.Error,"Unterminated_quoted_string_in_response_file_0_6045","Unterminated quoted string in response file '{0}'."),Argument_for_0_option_must_be_Colon_1:n(6046,e.DiagnosticCategory.Error,"Argument_for_0_option_must_be_Colon_1_6046","Argument for '{0}' option must be: {1}."),Locale_must_be_of_the_form_language_or_language_territory_For_example_0_or_1:n(6048,e.DiagnosticCategory.Error,"Locale_must_be_of_the_form_language_or_language_territory_For_example_0_or_1_6048","Locale must be of the form <language> or <language>-<territory>. For example '{0}' or '{1}'."),Unsupported_locale_0:n(6049,e.DiagnosticCategory.Error,"Unsupported_locale_0_6049","Unsupported locale '{0}'."),Unable_to_open_file_0:n(6050,e.DiagnosticCategory.Error,"Unable_to_open_file_0_6050","Unable to open file '{0}'."),Corrupted_locale_file_0:n(6051,e.DiagnosticCategory.Error,"Corrupted_locale_file_0_6051","Corrupted locale file {0}."),Raise_error_on_expressions_and_declarations_with_an_implied_any_type:n(6052,e.DiagnosticCategory.Message,"Raise_error_on_expressions_and_declarations_with_an_implied_any_type_6052","Raise error on expressions and declarations with an implied 'any' type."),File_0_not_found:n(6053,e.DiagnosticCategory.Error,"File_0_not_found_6053","File '{0}' not found."),File_0_has_unsupported_extension_The_only_supported_extensions_are_1:n(6054,e.DiagnosticCategory.Error,"File_0_has_unsupported_extension_The_only_supported_extensions_are_1_6054","File '{0}' has unsupported extension. The only supported extensions are {1}."),Suppress_noImplicitAny_errors_for_indexing_objects_lacking_index_signatures:n(6055,e.DiagnosticCategory.Message,"Suppress_noImplicitAny_errors_for_indexing_objects_lacking_index_signatures_6055","Suppress noImplicitAny errors for indexing objects lacking index signatures."),Do_not_emit_declarations_for_code_that_has_an_internal_annotation:n(6056,e.DiagnosticCategory.Message,"Do_not_emit_declarations_for_code_that_has_an_internal_annotation_6056","Do not emit declarations for code that has an '@internal' annotation."),Specify_the_root_directory_of_input_files_Use_to_control_the_output_directory_structure_with_outDir:n(6058,e.DiagnosticCategory.Message,"Specify_the_root_directory_of_input_files_Use_to_control_the_output_directory_structure_with_outDir_6058","Specify the root directory of input files. Use to control the output directory structure with --outDir."),File_0_is_not_under_rootDir_1_rootDir_is_expected_to_contain_all_source_files:n(6059,e.DiagnosticCategory.Error,"File_0_is_not_under_rootDir_1_rootDir_is_expected_to_contain_all_source_files_6059","File '{0}' is not under 'rootDir' '{1}'. 'rootDir' is expected to contain all source files."),Specify_the_end_of_line_sequence_to_be_used_when_emitting_files_Colon_CRLF_dos_or_LF_unix:n(6060,e.DiagnosticCategory.Message,"Specify_the_end_of_line_sequence_to_be_used_when_emitting_files_Colon_CRLF_dos_or_LF_unix_6060","Specify the end of line sequence to be used when emitting files: 'CRLF' (dos) or 'LF' (unix)."),NEWLINE:n(6061,e.DiagnosticCategory.Message,"NEWLINE_6061","NEWLINE"),Option_0_can_only_be_specified_in_tsconfig_json_file:n(6064,e.DiagnosticCategory.Error,"Option_0_can_only_be_specified_in_tsconfig_json_file_6064","Option '{0}' can only be specified in 'tsconfig.json' file."),Enables_experimental_support_for_ES7_decorators:n(6065,e.DiagnosticCategory.Message,"Enables_experimental_support_for_ES7_decorators_6065","Enables experimental support for ES7 decorators."),Enables_experimental_support_for_emitting_type_metadata_for_decorators:n(6066,e.DiagnosticCategory.Message,"Enables_experimental_support_for_emitting_type_metadata_for_decorators_6066","Enables experimental support for emitting type metadata for decorators."),Enables_experimental_support_for_ES7_async_functions:n(6068,e.DiagnosticCategory.Message,"Enables_experimental_support_for_ES7_async_functions_6068","Enables experimental support for ES7 async functions."),Specify_module_resolution_strategy_Colon_node_Node_js_or_classic_TypeScript_pre_1_6:n(6069,e.DiagnosticCategory.Message,"Specify_module_resolution_strategy_Colon_node_Node_js_or_classic_TypeScript_pre_1_6_6069","Specify module resolution strategy: 'node' (Node.js) or 'classic' (TypeScript pre-1.6)."),Initializes_a_TypeScript_project_and_creates_a_tsconfig_json_file:n(6070,e.DiagnosticCategory.Message,"Initializes_a_TypeScript_project_and_creates_a_tsconfig_json_file_6070","Initializes a TypeScript project and creates a tsconfig.json file."),Successfully_created_a_tsconfig_json_file:n(6071,e.DiagnosticCategory.Message,"Successfully_created_a_tsconfig_json_file_6071","Successfully created a tsconfig.json file."),Suppress_excess_property_checks_for_object_literals:n(6072,e.DiagnosticCategory.Message,"Suppress_excess_property_checks_for_object_literals_6072","Suppress excess property checks for object literals."),Stylize_errors_and_messages_using_color_and_context_experimental:n(6073,e.DiagnosticCategory.Message,"Stylize_errors_and_messages_using_color_and_context_experimental_6073","Stylize errors and messages using color and context (experimental)."),Do_not_report_errors_on_unused_labels:n(6074,e.DiagnosticCategory.Message,"Do_not_report_errors_on_unused_labels_6074","Do not report errors on unused labels."),Report_error_when_not_all_code_paths_in_function_return_a_value:n(6075,e.DiagnosticCategory.Message,"Report_error_when_not_all_code_paths_in_function_return_a_value_6075","Report error when not all code paths in function return a value."),Report_errors_for_fallthrough_cases_in_switch_statement:n(6076,e.DiagnosticCategory.Message,"Report_errors_for_fallthrough_cases_in_switch_statement_6076","Report errors for fallthrough cases in switch statement."),Do_not_report_errors_on_unreachable_code:n(6077,e.DiagnosticCategory.Message,"Do_not_report_errors_on_unreachable_code_6077","Do not report errors on unreachable code."),Disallow_inconsistently_cased_references_to_the_same_file:n(6078,e.DiagnosticCategory.Message,"Disallow_inconsistently_cased_references_to_the_same_file_6078","Disallow inconsistently-cased references to the same file."),Specify_library_files_to_be_included_in_the_compilation:n(6079,e.DiagnosticCategory.Message,"Specify_library_files_to_be_included_in_the_compilation_6079","Specify library files to be included in the compilation."),Specify_JSX_code_generation_Colon_preserve_react_native_or_react:n(6080,e.DiagnosticCategory.Message,"Specify_JSX_code_generation_Colon_preserve_react_native_or_react_6080","Specify JSX code generation: 'preserve', 'react-native', or 'react'."),File_0_has_an_unsupported_extension_so_skipping_it:n(6081,e.DiagnosticCategory.Message,"File_0_has_an_unsupported_extension_so_skipping_it_6081","File '{0}' has an unsupported extension, so skipping it."),Only_amd_and_system_modules_are_supported_alongside_0:n(6082,e.DiagnosticCategory.Error,"Only_amd_and_system_modules_are_supported_alongside_0_6082","Only 'amd' and 'system' modules are supported alongside --{0}."),Base_directory_to_resolve_non_absolute_module_names:n(6083,e.DiagnosticCategory.Message,"Base_directory_to_resolve_non_absolute_module_names_6083","Base directory to resolve non-absolute module names."),Deprecated_Use_jsxFactory_instead_Specify_the_object_invoked_for_createElement_when_targeting_react_JSX_emit:n(6084,e.DiagnosticCategory.Message,"Deprecated_Use_jsxFactory_instead_Specify_the_object_invoked_for_createElement_when_targeting_react__6084","[Deprecated] Use '--jsxFactory' instead. Specify the object invoked for createElement when targeting 'react' JSX emit"),Enable_tracing_of_the_name_resolution_process:n(6085,e.DiagnosticCategory.Message,"Enable_tracing_of_the_name_resolution_process_6085","Enable tracing of the name resolution process."),Resolving_module_0_from_1:n(6086,e.DiagnosticCategory.Message,"Resolving_module_0_from_1_6086","======== Resolving module '{0}' from '{1}'. ========"),Explicitly_specified_module_resolution_kind_Colon_0:n(6087,e.DiagnosticCategory.Message,"Explicitly_specified_module_resolution_kind_Colon_0_6087","Explicitly specified module resolution kind: '{0}'."),Module_resolution_kind_is_not_specified_using_0:n(6088,e.DiagnosticCategory.Message,"Module_resolution_kind_is_not_specified_using_0_6088","Module resolution kind is not specified, using '{0}'."),Module_name_0_was_successfully_resolved_to_1:n(6089,e.DiagnosticCategory.Message,"Module_name_0_was_successfully_resolved_to_1_6089","======== Module name '{0}' was successfully resolved to '{1}'. ========"),Module_name_0_was_not_resolved:n(6090,e.DiagnosticCategory.Message,"Module_name_0_was_not_resolved_6090","======== Module name '{0}' was not resolved. ========"),paths_option_is_specified_looking_for_a_pattern_to_match_module_name_0:n(6091,e.DiagnosticCategory.Message,"paths_option_is_specified_looking_for_a_pattern_to_match_module_name_0_6091","'paths' option is specified, looking for a pattern to match module name '{0}'."),Module_name_0_matched_pattern_1:n(6092,e.DiagnosticCategory.Message,"Module_name_0_matched_pattern_1_6092","Module name '{0}', matched pattern '{1}'."),Trying_substitution_0_candidate_module_location_Colon_1:n(6093,e.DiagnosticCategory.Message,"Trying_substitution_0_candidate_module_location_Colon_1_6093","Trying substitution '{0}', candidate module location: '{1}'."),Resolving_module_name_0_relative_to_base_url_1_2:n(6094,e.DiagnosticCategory.Message,"Resolving_module_name_0_relative_to_base_url_1_2_6094","Resolving module name '{0}' relative to base url '{1}' - '{2}'."),Loading_module_as_file_Slash_folder_candidate_module_location_0_target_file_type_1:n(6095,e.DiagnosticCategory.Message,"Loading_module_as_file_Slash_folder_candidate_module_location_0_target_file_type_1_6095","Loading module as file / folder, candidate module location '{0}', target file type '{1}'."),File_0_does_not_exist:n(6096,e.DiagnosticCategory.Message,"File_0_does_not_exist_6096","File '{0}' does not exist."),File_0_exist_use_it_as_a_name_resolution_result:n(6097,e.DiagnosticCategory.Message,"File_0_exist_use_it_as_a_name_resolution_result_6097","File '{0}' exist - use it as a name resolution result."),Loading_module_0_from_node_modules_folder_target_file_type_1:n(6098,e.DiagnosticCategory.Message,"Loading_module_0_from_node_modules_folder_target_file_type_1_6098","Loading module '{0}' from 'node_modules' folder, target file type '{1}'."),Found_package_json_at_0:n(6099,e.DiagnosticCategory.Message,"Found_package_json_at_0_6099","Found 'package.json' at '{0}'."),package_json_does_not_have_a_0_field:n(6100,e.DiagnosticCategory.Message,"package_json_does_not_have_a_0_field_6100","'package.json' does not have a '{0}' field."),package_json_has_0_field_1_that_references_2:n(6101,e.DiagnosticCategory.Message,"package_json_has_0_field_1_that_references_2_6101","'package.json' has '{0}' field '{1}' that references '{2}'."),Allow_javascript_files_to_be_compiled:n(6102,e.DiagnosticCategory.Message,"Allow_javascript_files_to_be_compiled_6102","Allow javascript files to be compiled."),Option_0_should_have_array_of_strings_as_a_value:n(6103,e.DiagnosticCategory.Error,"Option_0_should_have_array_of_strings_as_a_value_6103","Option '{0}' should have array of strings as a value."),Checking_if_0_is_the_longest_matching_prefix_for_1_2:n(6104,e.DiagnosticCategory.Message,"Checking_if_0_is_the_longest_matching_prefix_for_1_2_6104","Checking if '{0}' is the longest matching prefix for '{1}' - '{2}'."),Expected_type_of_0_field_in_package_json_to_be_1_got_2:n(6105,e.DiagnosticCategory.Message,"Expected_type_of_0_field_in_package_json_to_be_1_got_2_6105","Expected type of '{0}' field in 'package.json' to be '{1}', got '{2}'."),baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1:n(6106,e.DiagnosticCategory.Message,"baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1_6106","'baseUrl' option is set to '{0}', using this value to resolve non-relative module name '{1}'."),rootDirs_option_is_set_using_it_to_resolve_relative_module_name_0:n(6107,e.DiagnosticCategory.Message,"rootDirs_option_is_set_using_it_to_resolve_relative_module_name_0_6107","'rootDirs' option is set, using it to resolve relative module name '{0}'."),Longest_matching_prefix_for_0_is_1:n(6108,e.DiagnosticCategory.Message,"Longest_matching_prefix_for_0_is_1_6108","Longest matching prefix for '{0}' is '{1}'."),Loading_0_from_the_root_dir_1_candidate_location_2:n(6109,e.DiagnosticCategory.Message,"Loading_0_from_the_root_dir_1_candidate_location_2_6109","Loading '{0}' from the root dir '{1}', candidate location '{2}'."),Trying_other_entries_in_rootDirs:n(6110,e.DiagnosticCategory.Message,"Trying_other_entries_in_rootDirs_6110","Trying other entries in 'rootDirs'."),Module_resolution_using_rootDirs_has_failed:n(6111,e.DiagnosticCategory.Message,"Module_resolution_using_rootDirs_has_failed_6111","Module resolution using 'rootDirs' has failed."),Do_not_emit_use_strict_directives_in_module_output:n(6112,e.DiagnosticCategory.Message,"Do_not_emit_use_strict_directives_in_module_output_6112","Do not emit 'use strict' directives in module output."),Enable_strict_null_checks:n(6113,e.DiagnosticCategory.Message,"Enable_strict_null_checks_6113","Enable strict null checks."),Unknown_option_excludes_Did_you_mean_exclude:n(6114,e.DiagnosticCategory.Error,"Unknown_option_excludes_Did_you_mean_exclude_6114","Unknown option 'excludes'. Did you mean 'exclude'?"),Raise_error_on_this_expressions_with_an_implied_any_type:n(6115,e.DiagnosticCategory.Message,"Raise_error_on_this_expressions_with_an_implied_any_type_6115","Raise error on 'this' expressions with an implied 'any' type."),Resolving_type_reference_directive_0_containing_file_1_root_directory_2:n(6116,e.DiagnosticCategory.Message,"Resolving_type_reference_directive_0_containing_file_1_root_directory_2_6116","======== Resolving type reference directive '{0}', containing file '{1}', root directory '{2}'. ========"),Resolving_using_primary_search_paths:n(6117,e.DiagnosticCategory.Message,"Resolving_using_primary_search_paths_6117","Resolving using primary search paths..."),Resolving_from_node_modules_folder:n(6118,e.DiagnosticCategory.Message,"Resolving_from_node_modules_folder_6118","Resolving from node_modules folder..."),Type_reference_directive_0_was_successfully_resolved_to_1_primary_Colon_2:n(6119,e.DiagnosticCategory.Message,"Type_reference_directive_0_was_successfully_resolved_to_1_primary_Colon_2_6119","======== Type reference directive '{0}' was successfully resolved to '{1}', primary: {2}. ========"),Type_reference_directive_0_was_not_resolved:n(6120,e.DiagnosticCategory.Message,"Type_reference_directive_0_was_not_resolved_6120","======== Type reference directive '{0}' was not resolved. ========"),Resolving_with_primary_search_path_0:n(6121,e.DiagnosticCategory.Message,"Resolving_with_primary_search_path_0_6121","Resolving with primary search path '{0}'."),Root_directory_cannot_be_determined_skipping_primary_search_paths:n(6122,e.DiagnosticCategory.Message,"Root_directory_cannot_be_determined_skipping_primary_search_paths_6122","Root directory cannot be determined, skipping primary search paths."),Resolving_type_reference_directive_0_containing_file_1_root_directory_not_set:n(6123,e.DiagnosticCategory.Message,"Resolving_type_reference_directive_0_containing_file_1_root_directory_not_set_6123","======== Resolving type reference directive '{0}', containing file '{1}', root directory not set. ========"),Type_declaration_files_to_be_included_in_compilation:n(6124,e.DiagnosticCategory.Message,"Type_declaration_files_to_be_included_in_compilation_6124","Type declaration files to be included in compilation."),Looking_up_in_node_modules_folder_initial_location_0:n(6125,e.DiagnosticCategory.Message,"Looking_up_in_node_modules_folder_initial_location_0_6125","Looking up in 'node_modules' folder, initial location '{0}'."),Containing_file_is_not_specified_and_root_directory_cannot_be_determined_skipping_lookup_in_node_modules_folder:n(6126,e.DiagnosticCategory.Message,"Containing_file_is_not_specified_and_root_directory_cannot_be_determined_skipping_lookup_in_node_mod_6126","Containing file is not specified and root directory cannot be determined, skipping lookup in 'node_modules' folder."),Resolving_type_reference_directive_0_containing_file_not_set_root_directory_1:n(6127,e.DiagnosticCategory.Message,"Resolving_type_reference_directive_0_containing_file_not_set_root_directory_1_6127","======== Resolving type reference directive '{0}', containing file not set, root directory '{1}'. ========"),Resolving_type_reference_directive_0_containing_file_not_set_root_directory_not_set:n(6128,e.DiagnosticCategory.Message,"Resolving_type_reference_directive_0_containing_file_not_set_root_directory_not_set_6128","======== Resolving type reference directive '{0}', containing file not set, root directory not set. ========"),Resolving_real_path_for_0_result_1:n(6130,e.DiagnosticCategory.Message,"Resolving_real_path_for_0_result_1_6130","Resolving real path for '{0}', result '{1}'."),Cannot_compile_modules_using_option_0_unless_the_module_flag_is_amd_or_system:n(6131,e.DiagnosticCategory.Error,"Cannot_compile_modules_using_option_0_unless_the_module_flag_is_amd_or_system_6131","Cannot compile modules using option '{0}' unless the '--module' flag is 'amd' or 'system'."),File_name_0_has_a_1_extension_stripping_it:n(6132,e.DiagnosticCategory.Message,"File_name_0_has_a_1_extension_stripping_it_6132","File name '{0}' has a '{1}' extension - stripping it."),_0_is_declared_but_its_value_is_never_read:n(6133,e.DiagnosticCategory.Error,"_0_is_declared_but_its_value_is_never_read_6133","'{0}' is declared but its value is never read.",!0),Report_errors_on_unused_locals:n(6134,e.DiagnosticCategory.Message,"Report_errors_on_unused_locals_6134","Report errors on unused locals."),Report_errors_on_unused_parameters:n(6135,e.DiagnosticCategory.Message,"Report_errors_on_unused_parameters_6135","Report errors on unused parameters."),The_maximum_dependency_depth_to_search_under_node_modules_and_load_JavaScript_files:n(6136,e.DiagnosticCategory.Message,"The_maximum_dependency_depth_to_search_under_node_modules_and_load_JavaScript_files_6136","The maximum dependency depth to search under node_modules and load JavaScript files."),Cannot_import_type_declaration_files_Consider_importing_0_instead_of_1:n(6137,e.DiagnosticCategory.Error,"Cannot_import_type_declaration_files_Consider_importing_0_instead_of_1_6137","Cannot import type declaration files. Consider importing '{0}' instead of '{1}'."),Property_0_is_declared_but_its_value_is_never_read:n(6138,e.DiagnosticCategory.Error,"Property_0_is_declared_but_its_value_is_never_read_6138","Property '{0}' is declared but its value is never read.",!0),Import_emit_helpers_from_tslib:n(6139,e.DiagnosticCategory.Message,"Import_emit_helpers_from_tslib_6139","Import emit helpers from 'tslib'."),Auto_discovery_for_typings_is_enabled_in_project_0_Running_extra_resolution_pass_for_module_1_using_cache_location_2:n(6140,e.DiagnosticCategory.Error,"Auto_discovery_for_typings_is_enabled_in_project_0_Running_extra_resolution_pass_for_module_1_using__6140","Auto discovery for typings is enabled in project '{0}'. Running extra resolution pass for module '{1}' using cache location '{2}'."),Parse_in_strict_mode_and_emit_use_strict_for_each_source_file:n(6141,e.DiagnosticCategory.Message,"Parse_in_strict_mode_and_emit_use_strict_for_each_source_file_6141",'Parse in strict mode and emit "use strict" for each source file.'),Module_0_was_resolved_to_1_but_jsx_is_not_set:n(6142,e.DiagnosticCategory.Error,"Module_0_was_resolved_to_1_but_jsx_is_not_set_6142","Module '{0}' was resolved to '{1}', but '--jsx' is not set."),Module_0_was_resolved_as_locally_declared_ambient_module_in_file_1:n(6144,e.DiagnosticCategory.Message,"Module_0_was_resolved_as_locally_declared_ambient_module_in_file_1_6144","Module '{0}' was resolved as locally declared ambient module in file '{1}'."),Module_0_was_resolved_as_ambient_module_declared_in_1_since_this_file_was_not_modified:n(6145,e.DiagnosticCategory.Message,"Module_0_was_resolved_as_ambient_module_declared_in_1_since_this_file_was_not_modified_6145","Module '{0}' was resolved as ambient module declared in '{1}' since this file was not modified."),Specify_the_JSX_factory_function_to_use_when_targeting_react_JSX_emit_e_g_React_createElement_or_h:n(6146,e.DiagnosticCategory.Message,"Specify_the_JSX_factory_function_to_use_when_targeting_react_JSX_emit_e_g_React_createElement_or_h_6146","Specify the JSX factory function to use when targeting 'react' JSX emit, e.g. 'React.createElement' or 'h'."),Resolution_for_module_0_was_found_in_cache_from_location_1:n(6147,e.DiagnosticCategory.Message,"Resolution_for_module_0_was_found_in_cache_from_location_1_6147","Resolution for module '{0}' was found in cache from location '{1}'."),Directory_0_does_not_exist_skipping_all_lookups_in_it:n(6148,e.DiagnosticCategory.Message,"Directory_0_does_not_exist_skipping_all_lookups_in_it_6148","Directory '{0}' does not exist, skipping all lookups in it."),Show_diagnostic_information:n(6149,e.DiagnosticCategory.Message,"Show_diagnostic_information_6149","Show diagnostic information."),Show_verbose_diagnostic_information:n(6150,e.DiagnosticCategory.Message,"Show_verbose_diagnostic_information_6150","Show verbose diagnostic information."),Emit_a_single_file_with_source_maps_instead_of_having_a_separate_file:n(6151,e.DiagnosticCategory.Message,"Emit_a_single_file_with_source_maps_instead_of_having_a_separate_file_6151","Emit a single file with source maps instead of having a separate file."),Emit_the_source_alongside_the_sourcemaps_within_a_single_file_requires_inlineSourceMap_or_sourceMap_to_be_set:n(6152,e.DiagnosticCategory.Message,"Emit_the_source_alongside_the_sourcemaps_within_a_single_file_requires_inlineSourceMap_or_sourceMap__6152","Emit the source alongside the sourcemaps within a single file; requires '--inlineSourceMap' or '--sourceMap' to be set."),Transpile_each_file_as_a_separate_module_similar_to_ts_transpileModule:n(6153,e.DiagnosticCategory.Message,"Transpile_each_file_as_a_separate_module_similar_to_ts_transpileModule_6153","Transpile each file as a separate module (similar to 'ts.transpileModule')."),Print_names_of_generated_files_part_of_the_compilation:n(6154,e.DiagnosticCategory.Message,"Print_names_of_generated_files_part_of_the_compilation_6154","Print names of generated files part of the compilation."),Print_names_of_files_part_of_the_compilation:n(6155,e.DiagnosticCategory.Message,"Print_names_of_files_part_of_the_compilation_6155","Print names of files part of the compilation."),The_locale_used_when_displaying_messages_to_the_user_e_g_en_us:n(6156,e.DiagnosticCategory.Message,"The_locale_used_when_displaying_messages_to_the_user_e_g_en_us_6156","The locale used when displaying messages to the user (e.g. 'en-us')"),Do_not_generate_custom_helper_functions_like_extends_in_compiled_output:n(6157,e.DiagnosticCategory.Message,"Do_not_generate_custom_helper_functions_like_extends_in_compiled_output_6157","Do not generate custom helper functions like '__extends' in compiled output."),Do_not_include_the_default_library_file_lib_d_ts:n(6158,e.DiagnosticCategory.Message,"Do_not_include_the_default_library_file_lib_d_ts_6158","Do not include the default library file (lib.d.ts)."),Do_not_add_triple_slash_references_or_imported_modules_to_the_list_of_compiled_files:n(6159,e.DiagnosticCategory.Message,"Do_not_add_triple_slash_references_or_imported_modules_to_the_list_of_compiled_files_6159","Do not add triple-slash references or imported modules to the list of compiled files."),Deprecated_Use_skipLibCheck_instead_Skip_type_checking_of_default_library_declaration_files:n(6160,e.DiagnosticCategory.Message,"Deprecated_Use_skipLibCheck_instead_Skip_type_checking_of_default_library_declaration_files_6160","[Deprecated] Use '--skipLibCheck' instead. Skip type checking of default library declaration files."),List_of_folders_to_include_type_definitions_from:n(6161,e.DiagnosticCategory.Message,"List_of_folders_to_include_type_definitions_from_6161","List of folders to include type definitions from."),Disable_size_limitations_on_JavaScript_projects:n(6162,e.DiagnosticCategory.Message,"Disable_size_limitations_on_JavaScript_projects_6162","Disable size limitations on JavaScript projects."),The_character_set_of_the_input_files:n(6163,e.DiagnosticCategory.Message,"The_character_set_of_the_input_files_6163","The character set of the input files."),Emit_a_UTF_8_Byte_Order_Mark_BOM_in_the_beginning_of_output_files:n(6164,e.DiagnosticCategory.Message,"Emit_a_UTF_8_Byte_Order_Mark_BOM_in_the_beginning_of_output_files_6164","Emit a UTF-8 Byte Order Mark (BOM) in the beginning of output files."),Do_not_truncate_error_messages:n(6165,e.DiagnosticCategory.Message,"Do_not_truncate_error_messages_6165","Do not truncate error messages."),Output_directory_for_generated_declaration_files:n(6166,e.DiagnosticCategory.Message,"Output_directory_for_generated_declaration_files_6166","Output directory for generated declaration files."),A_series_of_entries_which_re_map_imports_to_lookup_locations_relative_to_the_baseUrl:n(6167,e.DiagnosticCategory.Message,"A_series_of_entries_which_re_map_imports_to_lookup_locations_relative_to_the_baseUrl_6167","A series of entries which re-map imports to lookup locations relative to the 'baseUrl'."),List_of_root_folders_whose_combined_content_represents_the_structure_of_the_project_at_runtime:n(6168,e.DiagnosticCategory.Message,"List_of_root_folders_whose_combined_content_represents_the_structure_of_the_project_at_runtime_6168","List of root folders whose combined content represents the structure of the project at runtime."),Show_all_compiler_options:n(6169,e.DiagnosticCategory.Message,"Show_all_compiler_options_6169","Show all compiler options."),Deprecated_Use_outFile_instead_Concatenate_and_emit_output_to_single_file:n(6170,e.DiagnosticCategory.Message,"Deprecated_Use_outFile_instead_Concatenate_and_emit_output_to_single_file_6170","[Deprecated] Use '--outFile' instead. Concatenate and emit output to single file"),Command_line_Options:n(6171,e.DiagnosticCategory.Message,"Command_line_Options_6171","Command-line Options"),Basic_Options:n(6172,e.DiagnosticCategory.Message,"Basic_Options_6172","Basic Options"),Strict_Type_Checking_Options:n(6173,e.DiagnosticCategory.Message,"Strict_Type_Checking_Options_6173","Strict Type-Checking Options"),Module_Resolution_Options:n(6174,e.DiagnosticCategory.Message,"Module_Resolution_Options_6174","Module Resolution Options"),Source_Map_Options:n(6175,e.DiagnosticCategory.Message,"Source_Map_Options_6175","Source Map Options"),Additional_Checks:n(6176,e.DiagnosticCategory.Message,"Additional_Checks_6176","Additional Checks"),Experimental_Options:n(6177,e.DiagnosticCategory.Message,"Experimental_Options_6177","Experimental Options"),Advanced_Options:n(6178,e.DiagnosticCategory.Message,"Advanced_Options_6178","Advanced Options"),Provide_full_support_for_iterables_in_for_of_spread_and_destructuring_when_targeting_ES5_or_ES3:n(6179,e.DiagnosticCategory.Message,"Provide_full_support_for_iterables_in_for_of_spread_and_destructuring_when_targeting_ES5_or_ES3_6179","Provide full support for iterables in 'for-of', spread, and destructuring when targeting 'ES5' or 'ES3'."),Enable_all_strict_type_checking_options:n(6180,e.DiagnosticCategory.Message,"Enable_all_strict_type_checking_options_6180","Enable all strict type-checking options."),List_of_language_service_plugins:n(6181,e.DiagnosticCategory.Message,"List_of_language_service_plugins_6181","List of language service plugins."),Scoped_package_detected_looking_in_0:n(6182,e.DiagnosticCategory.Message,"Scoped_package_detected_looking_in_0_6182","Scoped package detected, looking in '{0}'"),Reusing_resolution_of_module_0_to_file_1_from_old_program:n(6183,e.DiagnosticCategory.Message,"Reusing_resolution_of_module_0_to_file_1_from_old_program_6183","Reusing resolution of module '{0}' to file '{1}' from old program."),Reusing_module_resolutions_originating_in_0_since_resolutions_are_unchanged_from_old_program:n(6184,e.DiagnosticCategory.Message,"Reusing_module_resolutions_originating_in_0_since_resolutions_are_unchanged_from_old_program_6184","Reusing module resolutions originating in '{0}' since resolutions are unchanged from old program."),Disable_strict_checking_of_generic_signatures_in_function_types:n(6185,e.DiagnosticCategory.Message,"Disable_strict_checking_of_generic_signatures_in_function_types_6185","Disable strict checking of generic signatures in function types."),Enable_strict_checking_of_function_types:n(6186,e.DiagnosticCategory.Message,"Enable_strict_checking_of_function_types_6186","Enable strict checking of function types."),Enable_strict_checking_of_property_initialization_in_classes:n(6187,e.DiagnosticCategory.Message,"Enable_strict_checking_of_property_initialization_in_classes_6187","Enable strict checking of property initialization in classes."),Numeric_separators_are_not_allowed_here:n(6188,e.DiagnosticCategory.Error,"Numeric_separators_are_not_allowed_here_6188","Numeric separators are not allowed here."),Multiple_consecutive_numeric_separators_are_not_permitted:n(6189,e.DiagnosticCategory.Error,"Multiple_consecutive_numeric_separators_are_not_permitted_6189","Multiple consecutive numeric separators are not permitted."),Found_package_json_at_0_Package_ID_is_1:n(6190,e.DiagnosticCategory.Message,"Found_package_json_at_0_Package_ID_is_1_6190","Found 'package.json' at '{0}'. Package ID is '{1}'."),Whether_to_keep_outdated_console_output_in_watch_mode_instead_of_clearing_the_screen:n(6191,e.DiagnosticCategory.Message,"Whether_to_keep_outdated_console_output_in_watch_mode_instead_of_clearing_the_screen_6191","Whether to keep outdated console output in watch mode instead of clearing the screen."),All_imports_in_import_declaration_are_unused:n(6192,e.DiagnosticCategory.Error,"All_imports_in_import_declaration_are_unused_6192","All imports in import declaration are unused.",!0),Found_1_error_Watching_for_file_changes:n(6193,e.DiagnosticCategory.Message,"Found_1_error_Watching_for_file_changes_6193","Found 1 error. Watching for file changes."),Found_0_errors_Watching_for_file_changes:n(6194,e.DiagnosticCategory.Message,"Found_0_errors_Watching_for_file_changes_6194","Found {0} errors. Watching for file changes."),Resolve_keyof_to_string_valued_property_names_only_no_numbers_or_symbols:n(6195,e.DiagnosticCategory.Message,"Resolve_keyof_to_string_valued_property_names_only_no_numbers_or_symbols_6195","Resolve 'keyof' to string valued property names only (no numbers or symbols)."),_0_is_declared_but_never_used:n(6196,e.DiagnosticCategory.Error,"_0_is_declared_but_never_used_6196","'{0}' is declared but never used.",!0),Include_modules_imported_with_json_extension:n(6197,e.DiagnosticCategory.Message,"Include_modules_imported_with_json_extension_6197","Include modules imported with '.json' extension"),All_destructured_elements_are_unused:n(6198,e.DiagnosticCategory.Error,"All_destructured_elements_are_unused_6198","All destructured elements are unused.",!0),All_variables_are_unused:n(6199,e.DiagnosticCategory.Error,"All_variables_are_unused_6199","All variables are unused.",!0),Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0:n(6200,e.DiagnosticCategory.Error,"Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0_6200","Definitions of the following identifiers conflict with those in another file: {0}"),Conflicts_are_in_this_file:n(6201,e.DiagnosticCategory.Message,"Conflicts_are_in_this_file_6201","Conflicts are in this file."),_0_was_also_declared_here:n(6203,e.DiagnosticCategory.Message,"_0_was_also_declared_here_6203","'{0}' was also declared here."),and_here:n(6204,e.DiagnosticCategory.Message,"and_here_6204","and here."),All_type_parameters_are_unused:n(6205,e.DiagnosticCategory.Error,"All_type_parameters_are_unused_6205","All type parameters are unused"),package_json_has_a_typesVersions_field_with_version_specific_path_mappings:n(6206,e.DiagnosticCategory.Message,"package_json_has_a_typesVersions_field_with_version_specific_path_mappings_6206","'package.json' has a 'typesVersions' field with version-specific path mappings."),package_json_does_not_have_a_typesVersions_entry_that_matches_version_0:n(6207,e.DiagnosticCategory.Message,"package_json_does_not_have_a_typesVersions_entry_that_matches_version_0_6207","'package.json' does not have a 'typesVersions' entry that matches version '{0}'."),package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_match_module_name_2:n(6208,e.DiagnosticCategory.Message,"package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_ma_6208","'package.json' has a 'typesVersions' entry '{0}' that matches compiler version '{1}', looking for a pattern to match module name '{2}'."),package_json_has_a_typesVersions_entry_0_that_is_not_a_valid_semver_range:n(6209,e.DiagnosticCategory.Message,"package_json_has_a_typesVersions_entry_0_that_is_not_a_valid_semver_range_6209","'package.json' has a 'typesVersions' entry '{0}' that is not a valid semver range."),An_argument_for_0_was_not_provided:n(6210,e.DiagnosticCategory.Message,"An_argument_for_0_was_not_provided_6210","An argument for '{0}' was not provided."),An_argument_matching_this_binding_pattern_was_not_provided:n(6211,e.DiagnosticCategory.Message,"An_argument_matching_this_binding_pattern_was_not_provided_6211","An argument matching this binding pattern was not provided."),Did_you_mean_to_call_this_expression:n(6212,e.DiagnosticCategory.Message,"Did_you_mean_to_call_this_expression_6212","Did you mean to call this expression?"),Did_you_mean_to_use_new_with_this_expression:n(6213,e.DiagnosticCategory.Message,"Did_you_mean_to_use_new_with_this_expression_6213","Did you mean to use 'new' with this expression?"),Enable_strict_bind_call_and_apply_methods_on_functions:n(6214,e.DiagnosticCategory.Message,"Enable_strict_bind_call_and_apply_methods_on_functions_6214","Enable strict 'bind', 'call', and 'apply' methods on functions."),Using_compiler_options_of_project_reference_redirect_0:n(6215,e.DiagnosticCategory.Message,"Using_compiler_options_of_project_reference_redirect_0_6215","Using compiler options of project reference redirect '{0}'."),Found_1_error:n(6216,e.DiagnosticCategory.Message,"Found_1_error_6216","Found 1 error."),Found_0_errors:n(6217,e.DiagnosticCategory.Message,"Found_0_errors_6217","Found {0} errors."),Projects_to_reference:n(6300,e.DiagnosticCategory.Message,"Projects_to_reference_6300","Projects to reference"),Enable_project_compilation:n(6302,e.DiagnosticCategory.Message,"Enable_project_compilation_6302","Enable project compilation"),Project_references_may_not_form_a_circular_graph_Cycle_detected_Colon_0:n(6202,e.DiagnosticCategory.Error,"Project_references_may_not_form_a_circular_graph_Cycle_detected_Colon_0_6202","Project references may not form a circular graph. Cycle detected: {0}"),Composite_projects_may_not_disable_declaration_emit:n(6304,e.DiagnosticCategory.Error,"Composite_projects_may_not_disable_declaration_emit_6304","Composite projects may not disable declaration emit."),Output_file_0_has_not_been_built_from_source_file_1:n(6305,e.DiagnosticCategory.Error,"Output_file_0_has_not_been_built_from_source_file_1_6305","Output file '{0}' has not been built from source file '{1}'."),Referenced_project_0_must_have_setting_composite_Colon_true:n(6306,e.DiagnosticCategory.Error,"Referenced_project_0_must_have_setting_composite_Colon_true_6306","Referenced project '{0}' must have setting \"composite\": true."),File_0_is_not_in_project_file_list_Projects_must_list_all_files_or_use_an_include_pattern:n(6307,e.DiagnosticCategory.Error,"File_0_is_not_in_project_file_list_Projects_must_list_all_files_or_use_an_include_pattern_6307","File '{0}' is not in project file list. Projects must list all files or use an 'include' pattern."),Cannot_prepend_project_0_because_it_does_not_have_outFile_set:n(6308,e.DiagnosticCategory.Error,"Cannot_prepend_project_0_because_it_does_not_have_outFile_set_6308","Cannot prepend project '{0}' because it does not have 'outFile' set"),Output_file_0_from_project_1_does_not_exist:n(6309,e.DiagnosticCategory.Error,"Output_file_0_from_project_1_does_not_exist_6309","Output file '{0}' from project '{1}' does not exist"),Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2:n(6350,e.DiagnosticCategory.Message,"Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2_6350","Project '{0}' is out of date because oldest output '{1}' is older than newest input '{2}'"),Project_0_is_up_to_date_because_newest_input_1_is_older_than_oldest_output_2:n(6351,e.DiagnosticCategory.Message,"Project_0_is_up_to_date_because_newest_input_1_is_older_than_oldest_output_2_6351","Project '{0}' is up to date because newest input '{1}' is older than oldest output '{2}'"),Project_0_is_out_of_date_because_output_file_1_does_not_exist:n(6352,e.DiagnosticCategory.Message,"Project_0_is_out_of_date_because_output_file_1_does_not_exist_6352","Project '{0}' is out of date because output file '{1}' does not exist"),Project_0_is_out_of_date_because_its_dependency_1_is_out_of_date:n(6353,e.DiagnosticCategory.Message,"Project_0_is_out_of_date_because_its_dependency_1_is_out_of_date_6353","Project '{0}' is out of date because its dependency '{1}' is out of date"),Project_0_is_up_to_date_with_d_ts_files_from_its_dependencies:n(6354,e.DiagnosticCategory.Message,"Project_0_is_up_to_date_with_d_ts_files_from_its_dependencies_6354","Project '{0}' is up to date with .d.ts files from its dependencies"),Projects_in_this_build_Colon_0:n(6355,e.DiagnosticCategory.Message,"Projects_in_this_build_Colon_0_6355","Projects in this build: {0}"),A_non_dry_build_would_delete_the_following_files_Colon_0:n(6356,e.DiagnosticCategory.Message,"A_non_dry_build_would_delete_the_following_files_Colon_0_6356","A non-dry build would delete the following files: {0}"),A_non_dry_build_would_build_project_0:n(6357,e.DiagnosticCategory.Message,"A_non_dry_build_would_build_project_0_6357","A non-dry build would build project '{0}'"),Building_project_0:n(6358,e.DiagnosticCategory.Message,"Building_project_0_6358","Building project '{0}'..."),Updating_output_timestamps_of_project_0:n(6359,e.DiagnosticCategory.Message,"Updating_output_timestamps_of_project_0_6359","Updating output timestamps of project '{0}'..."),delete_this_Project_0_is_up_to_date_because_it_was_previously_built:n(6360,e.DiagnosticCategory.Message,"delete_this_Project_0_is_up_to_date_because_it_was_previously_built_6360","delete this - Project '{0}' is up to date because it was previously built"),Project_0_is_up_to_date:n(6361,e.DiagnosticCategory.Message,"Project_0_is_up_to_date_6361","Project '{0}' is up to date"),Skipping_build_of_project_0_because_its_dependency_1_has_errors:n(6362,e.DiagnosticCategory.Message,"Skipping_build_of_project_0_because_its_dependency_1_has_errors_6362","Skipping build of project '{0}' because its dependency '{1}' has errors"),Project_0_can_t_be_built_because_its_dependency_1_has_errors:n(6363,e.DiagnosticCategory.Message,"Project_0_can_t_be_built_because_its_dependency_1_has_errors_6363","Project '{0}' can't be built because its dependency '{1}' has errors"),Build_one_or_more_projects_and_their_dependencies_if_out_of_date:n(6364,e.DiagnosticCategory.Message,"Build_one_or_more_projects_and_their_dependencies_if_out_of_date_6364","Build one or more projects and their dependencies, if out of date"),Delete_the_outputs_of_all_projects:n(6365,e.DiagnosticCategory.Message,"Delete_the_outputs_of_all_projects_6365","Delete the outputs of all projects"),Enable_verbose_logging:n(6366,e.DiagnosticCategory.Message,"Enable_verbose_logging_6366","Enable verbose logging"),Show_what_would_be_built_or_deleted_if_specified_with_clean:n(6367,e.DiagnosticCategory.Message,"Show_what_would_be_built_or_deleted_if_specified_with_clean_6367","Show what would be built (or deleted, if specified with '--clean')"),Build_all_projects_including_those_that_appear_to_be_up_to_date:n(6368,e.DiagnosticCategory.Message,"Build_all_projects_including_those_that_appear_to_be_up_to_date_6368","Build all projects, including those that appear to be up to date"),Option_build_must_be_the_first_command_line_argument:n(6369,e.DiagnosticCategory.Error,"Option_build_must_be_the_first_command_line_argument_6369","Option '--build' must be the first command line argument."),Options_0_and_1_cannot_be_combined:n(6370,e.DiagnosticCategory.Error,"Options_0_and_1_cannot_be_combined_6370","Options '{0}' and '{1}' cannot be combined."),Updating_unchanged_output_timestamps_of_project_0:n(6371,e.DiagnosticCategory.Message,"Updating_unchanged_output_timestamps_of_project_0_6371","Updating unchanged output timestamps of project '{0}'..."),The_expected_type_comes_from_property_0_which_is_declared_here_on_type_1:n(6500,e.DiagnosticCategory.Message,"The_expected_type_comes_from_property_0_which_is_declared_here_on_type_1_6500","The expected type comes from property '{0}' which is declared here on type '{1}'"),The_expected_type_comes_from_this_index_signature:n(6501,e.DiagnosticCategory.Message,"The_expected_type_comes_from_this_index_signature_6501","The expected type comes from this index signature."),The_expected_type_comes_from_the_return_type_of_this_signature:n(6502,e.DiagnosticCategory.Message,"The_expected_type_comes_from_the_return_type_of_this_signature_6502","The expected type comes from the return type of this signature."),Variable_0_implicitly_has_an_1_type:n(7005,e.DiagnosticCategory.Error,"Variable_0_implicitly_has_an_1_type_7005","Variable '{0}' implicitly has an '{1}' type."),Parameter_0_implicitly_has_an_1_type:n(7006,e.DiagnosticCategory.Error,"Parameter_0_implicitly_has_an_1_type_7006","Parameter '{0}' implicitly has an '{1}' type."),Member_0_implicitly_has_an_1_type:n(7008,e.DiagnosticCategory.Error,"Member_0_implicitly_has_an_1_type_7008","Member '{0}' implicitly has an '{1}' type."),new_expression_whose_target_lacks_a_construct_signature_implicitly_has_an_any_type:n(7009,e.DiagnosticCategory.Error,"new_expression_whose_target_lacks_a_construct_signature_implicitly_has_an_any_type_7009","'new' expression, whose target lacks a construct signature, implicitly has an 'any' type."),_0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type:n(7010,e.DiagnosticCategory.Error,"_0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type_7010","'{0}', which lacks return-type annotation, implicitly has an '{1}' return type."),Function_expression_which_lacks_return_type_annotation_implicitly_has_an_0_return_type:n(7011,e.DiagnosticCategory.Error,"Function_expression_which_lacks_return_type_annotation_implicitly_has_an_0_return_type_7011","Function expression, which lacks return-type annotation, implicitly has an '{0}' return type."),Construct_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type:n(7013,e.DiagnosticCategory.Error,"Construct_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type_7013","Construct signature, which lacks return-type annotation, implicitly has an 'any' return type."),Function_type_which_lacks_return_type_annotation_implicitly_has_an_0_return_type:n(7014,e.DiagnosticCategory.Error,"Function_type_which_lacks_return_type_annotation_implicitly_has_an_0_return_type_7014","Function type, which lacks return-type annotation, implicitly has an '{0}' return type."),Element_implicitly_has_an_any_type_because_index_expression_is_not_of_type_number:n(7015,e.DiagnosticCategory.Error,"Element_implicitly_has_an_any_type_because_index_expression_is_not_of_type_number_7015","Element implicitly has an 'any' type because index expression is not of type 'number'."),Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type:n(7016,e.DiagnosticCategory.Error,"Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type_7016","Could not find a declaration file for module '{0}'. '{1}' implicitly has an 'any' type."),Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature:n(7017,e.DiagnosticCategory.Error,"Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature_7017","Element implicitly has an 'any' type because type '{0}' has no index signature."),Object_literal_s_property_0_implicitly_has_an_1_type:n(7018,e.DiagnosticCategory.Error,"Object_literal_s_property_0_implicitly_has_an_1_type_7018","Object literal's property '{0}' implicitly has an '{1}' type."),Rest_parameter_0_implicitly_has_an_any_type:n(7019,e.DiagnosticCategory.Error,"Rest_parameter_0_implicitly_has_an_any_type_7019","Rest parameter '{0}' implicitly has an 'any[]' type."),Call_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type:n(7020,e.DiagnosticCategory.Error,"Call_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type_7020","Call signature, which lacks return-type annotation, implicitly has an 'any' return type."),_0_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_and_is_referenced_directly_or_indirectly_in_its_own_initializer:n(7022,e.DiagnosticCategory.Error,"_0_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_and_is_referenced_directly_or__7022","'{0}' implicitly has type 'any' because it does not have a type annotation and is referenced directly or indirectly in its own initializer."),_0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions:n(7023,e.DiagnosticCategory.Error,"_0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_reference_7023","'{0}' implicitly has return type 'any' because it does not have a return type annotation and is referenced directly or indirectly in one of its return expressions."),Function_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions:n(7024,e.DiagnosticCategory.Error,"Function_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_ref_7024","Function implicitly has return type 'any' because it does not have a return type annotation and is referenced directly or indirectly in one of its return expressions."),Generator_implicitly_has_type_0_because_it_does_not_yield_any_values_Consider_supplying_a_return_type:n(7025,e.DiagnosticCategory.Error,"Generator_implicitly_has_type_0_because_it_does_not_yield_any_values_Consider_supplying_a_return_typ_7025","Generator implicitly has type '{0}' because it does not yield any values. Consider supplying a return type."),JSX_element_implicitly_has_type_any_because_no_interface_JSX_0_exists:n(7026,e.DiagnosticCategory.Error,"JSX_element_implicitly_has_type_any_because_no_interface_JSX_0_exists_7026","JSX element implicitly has type 'any' because no interface 'JSX.{0}' exists."),Unreachable_code_detected:n(7027,e.DiagnosticCategory.Error,"Unreachable_code_detected_7027","Unreachable code detected.",!0),Unused_label:n(7028,e.DiagnosticCategory.Error,"Unused_label_7028","Unused label.",!0),Fallthrough_case_in_switch:n(7029,e.DiagnosticCategory.Error,"Fallthrough_case_in_switch_7029","Fallthrough case in switch."),Not_all_code_paths_return_a_value:n(7030,e.DiagnosticCategory.Error,"Not_all_code_paths_return_a_value_7030","Not all code paths return a value."),Binding_element_0_implicitly_has_an_1_type:n(7031,e.DiagnosticCategory.Error,"Binding_element_0_implicitly_has_an_1_type_7031","Binding element '{0}' implicitly has an '{1}' type."),Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation:n(7032,e.DiagnosticCategory.Error,"Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation_7032","Property '{0}' implicitly has type 'any', because its set accessor lacks a parameter type annotation."),Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation:n(7033,e.DiagnosticCategory.Error,"Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation_7033","Property '{0}' implicitly has type 'any', because its get accessor lacks a return type annotation."),Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined:n(7034,e.DiagnosticCategory.Error,"Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined_7034","Variable '{0}' implicitly has type '{1}' in some locations where its type cannot be determined."),Try_npm_install_types_Slash_1_if_it_exists_or_add_a_new_declaration_d_ts_file_containing_declare_module_0:n(7035,e.DiagnosticCategory.Error,"Try_npm_install_types_Slash_1_if_it_exists_or_add_a_new_declaration_d_ts_file_containing_declare_mod_7035","Try `npm install @types/{1}` if it exists or add a new declaration (.d.ts) file containing `declare module '{0}';`"),Dynamic_import_s_specifier_must_be_of_type_string_but_here_has_type_0:n(7036,e.DiagnosticCategory.Error,"Dynamic_import_s_specifier_must_be_of_type_string_but_here_has_type_0_7036","Dynamic import's specifier must be of type 'string', but here has type '{0}'."),Enables_emit_interoperability_between_CommonJS_and_ES_Modules_via_creation_of_namespace_objects_for_all_imports_Implies_allowSyntheticDefaultImports:n(7037,e.DiagnosticCategory.Message,"Enables_emit_interoperability_between_CommonJS_and_ES_Modules_via_creation_of_namespace_objects_for__7037","Enables emit interoperability between CommonJS and ES Modules via creation of namespace objects for all imports. Implies 'allowSyntheticDefaultImports'."),Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cause_a_failure_at_runtime_Consider_using_a_default_import_or_import_require_here_instead:n(7038,e.DiagnosticCategory.Message,"Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cau_7038","Type originates at this import. A namespace-style import cannot be called or constructed, and will cause a failure at runtime. Consider using a default import or import require here instead."),Mapped_object_type_implicitly_has_an_any_template_type:n(7039,e.DiagnosticCategory.Error,"Mapped_object_type_implicitly_has_an_any_template_type_7039","Mapped object type implicitly has an 'any' template type."),If_the_0_package_actually_exposes_this_module_consider_sending_a_pull_request_to_amend_https_Colon_Slash_Slashgithub_com_SlashDefinitelyTyped_SlashDefinitelyTyped_Slashtree_Slashmaster_Slashtypes_Slash_1:n(7040,e.DiagnosticCategory.Error,"If_the_0_package_actually_exposes_this_module_consider_sending_a_pull_request_to_amend_https_Colon_S_7040","If the '{0}' package actually exposes this module, consider sending a pull request to amend 'https://github.com/DefinitelyTyped/DefinitelyTyped/tree/master/types/{1}`"),The_containing_arrow_function_captures_the_global_value_of_this_which_implicitly_has_type_any:n(7041,e.DiagnosticCategory.Error,"The_containing_arrow_function_captures_the_global_value_of_this_which_implicitly_has_type_any_7041","The containing arrow function captures the global value of 'this' which implicitly has type 'any'."),Module_0_was_resolved_to_1_but_resolveJsonModule_is_not_used:n(7042,e.DiagnosticCategory.Error,"Module_0_was_resolved_to_1_but_resolveJsonModule_is_not_used_7042","Module '{0}' was resolved to '{1}', but '--resolveJsonModule' is not used."),Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage:n(7043,e.DiagnosticCategory.Suggestion,"Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage_7043","Variable '{0}' implicitly has an '{1}' type, but a better type may be inferred from usage."),Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage:n(7044,e.DiagnosticCategory.Suggestion,"Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage_7044","Parameter '{0}' implicitly has an '{1}' type, but a better type may be inferred from usage."),Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage:n(7045,e.DiagnosticCategory.Suggestion,"Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage_7045","Member '{0}' implicitly has an '{1}' type, but a better type may be inferred from usage."),Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage:n(7046,e.DiagnosticCategory.Suggestion,"Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage_7046","Variable '{0}' implicitly has type '{1}' in some locations, but a better type may be inferred from usage."),Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage:n(7047,e.DiagnosticCategory.Suggestion,"Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage_7047","Rest parameter '{0}' implicitly has an 'any[]' type, but a better type may be inferred from usage."),Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage:n(7048,e.DiagnosticCategory.Suggestion,"Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage_7048","Property '{0}' implicitly has type 'any', but a better type for its get accessor may be inferred from usage."),Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage:n(7049,e.DiagnosticCategory.Suggestion,"Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage_7049","Property '{0}' implicitly has type 'any', but a better type for its set accessor may be inferred from usage."),_0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage:n(7050,e.DiagnosticCategory.Suggestion,"_0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage_7050","'{0}' implicitly has an '{1}' return type, but a better type may be inferred from usage."),Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1:n(7051,e.DiagnosticCategory.Error,"Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1_7051","Parameter has a name but no type. Did you mean '{0}: {1}'?"),You_cannot_rename_this_element:n(8e3,e.DiagnosticCategory.Error,"You_cannot_rename_this_element_8000","You cannot rename this element."),You_cannot_rename_elements_that_are_defined_in_the_standard_TypeScript_library:n(8001,e.DiagnosticCategory.Error,"You_cannot_rename_elements_that_are_defined_in_the_standard_TypeScript_library_8001","You cannot rename elements that are defined in the standard TypeScript library."),import_can_only_be_used_in_a_ts_file:n(8002,e.DiagnosticCategory.Error,"import_can_only_be_used_in_a_ts_file_8002","'import ... =' can only be used in a .ts file."),export_can_only_be_used_in_a_ts_file:n(8003,e.DiagnosticCategory.Error,"export_can_only_be_used_in_a_ts_file_8003","'export=' can only be used in a .ts file."),type_parameter_declarations_can_only_be_used_in_a_ts_file:n(8004,e.DiagnosticCategory.Error,"type_parameter_declarations_can_only_be_used_in_a_ts_file_8004","'type parameter declarations' can only be used in a .ts file."),implements_clauses_can_only_be_used_in_a_ts_file:n(8005,e.DiagnosticCategory.Error,"implements_clauses_can_only_be_used_in_a_ts_file_8005","'implements clauses' can only be used in a .ts file."),interface_declarations_can_only_be_used_in_a_ts_file:n(8006,e.DiagnosticCategory.Error,"interface_declarations_can_only_be_used_in_a_ts_file_8006","'interface declarations' can only be used in a .ts file."),module_declarations_can_only_be_used_in_a_ts_file:n(8007,e.DiagnosticCategory.Error,"module_declarations_can_only_be_used_in_a_ts_file_8007","'module declarations' can only be used in a .ts file."),type_aliases_can_only_be_used_in_a_ts_file:n(8008,e.DiagnosticCategory.Error,"type_aliases_can_only_be_used_in_a_ts_file_8008","'type aliases' can only be used in a .ts file."),_0_can_only_be_used_in_a_ts_file:n(8009,e.DiagnosticCategory.Error,"_0_can_only_be_used_in_a_ts_file_8009","'{0}' can only be used in a .ts file."),types_can_only_be_used_in_a_ts_file:n(8010,e.DiagnosticCategory.Error,"types_can_only_be_used_in_a_ts_file_8010","'types' can only be used in a .ts file."),type_arguments_can_only_be_used_in_a_ts_file:n(8011,e.DiagnosticCategory.Error,"type_arguments_can_only_be_used_in_a_ts_file_8011","'type arguments' can only be used in a .ts file."),parameter_modifiers_can_only_be_used_in_a_ts_file:n(8012,e.DiagnosticCategory.Error,"parameter_modifiers_can_only_be_used_in_a_ts_file_8012","'parameter modifiers' can only be used in a .ts file."),non_null_assertions_can_only_be_used_in_a_ts_file:n(8013,e.DiagnosticCategory.Error,"non_null_assertions_can_only_be_used_in_a_ts_file_8013","'non-null assertions' can only be used in a .ts file."),enum_declarations_can_only_be_used_in_a_ts_file:n(8015,e.DiagnosticCategory.Error,"enum_declarations_can_only_be_used_in_a_ts_file_8015","'enum declarations' can only be used in a .ts file."),type_assertion_expressions_can_only_be_used_in_a_ts_file:n(8016,e.DiagnosticCategory.Error,"type_assertion_expressions_can_only_be_used_in_a_ts_file_8016","'type assertion expressions' can only be used in a .ts file."),Octal_literal_types_must_use_ES2015_syntax_Use_the_syntax_0:n(8017,e.DiagnosticCategory.Error,"Octal_literal_types_must_use_ES2015_syntax_Use_the_syntax_0_8017","Octal literal types must use ES2015 syntax. Use the syntax '{0}'."),Octal_literals_are_not_allowed_in_enums_members_initializer_Use_the_syntax_0:n(8018,e.DiagnosticCategory.Error,"Octal_literals_are_not_allowed_in_enums_members_initializer_Use_the_syntax_0_8018","Octal literals are not allowed in enums members initializer. Use the syntax '{0}'."),Report_errors_in_js_files:n(8019,e.DiagnosticCategory.Message,"Report_errors_in_js_files_8019","Report errors in .js files."),JSDoc_types_can_only_be_used_inside_documentation_comments:n(8020,e.DiagnosticCategory.Error,"JSDoc_types_can_only_be_used_inside_documentation_comments_8020","JSDoc types can only be used inside documentation comments."),JSDoc_typedef_tag_should_either_have_a_type_annotation_or_be_followed_by_property_or_member_tags:n(8021,e.DiagnosticCategory.Error,"JSDoc_typedef_tag_should_either_have_a_type_annotation_or_be_followed_by_property_or_member_tags_8021","JSDoc '@typedef' tag should either have a type annotation or be followed by '@property' or '@member' tags."),JSDoc_0_is_not_attached_to_a_class:n(8022,e.DiagnosticCategory.Error,"JSDoc_0_is_not_attached_to_a_class_8022","JSDoc '@{0}' is not attached to a class."),JSDoc_0_1_does_not_match_the_extends_2_clause:n(8023,e.DiagnosticCategory.Error,"JSDoc_0_1_does_not_match_the_extends_2_clause_8023","JSDoc '@{0} {1}' does not match the 'extends {2}' clause."),JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name:n(8024,e.DiagnosticCategory.Error,"JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_8024","JSDoc '@param' tag has name '{0}', but there is no parameter with that name."),Class_declarations_cannot_have_more_than_one_augments_or_extends_tag:n(8025,e.DiagnosticCategory.Error,"Class_declarations_cannot_have_more_than_one_augments_or_extends_tag_8025","Class declarations cannot have more than one `@augments` or `@extends` tag."),Expected_0_type_arguments_provide_these_with_an_extends_tag:n(8026,e.DiagnosticCategory.Error,"Expected_0_type_arguments_provide_these_with_an_extends_tag_8026","Expected {0} type arguments; provide these with an '@extends' tag."),Expected_0_1_type_arguments_provide_these_with_an_extends_tag:n(8027,e.DiagnosticCategory.Error,"Expected_0_1_type_arguments_provide_these_with_an_extends_tag_8027","Expected {0}-{1} type arguments; provide these with an '@extends' tag."),JSDoc_may_only_appear_in_the_last_parameter_of_a_signature:n(8028,e.DiagnosticCategory.Error,"JSDoc_may_only_appear_in_the_last_parameter_of_a_signature_8028","JSDoc '...' may only appear in the last parameter of a signature."),JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_It_would_match_arguments_if_it_had_an_array_type:n(8029,e.DiagnosticCategory.Error,"JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_It_would_match_arguments_if_it_h_8029","JSDoc '@param' tag has name '{0}', but there is no parameter with that name. It would match 'arguments' if it had an array type."),The_type_of_a_function_declaration_must_match_the_function_s_signature:n(8030,e.DiagnosticCategory.Error,"The_type_of_a_function_declaration_must_match_the_function_s_signature_8030","The type of a function declaration must match the function's signature."),You_cannot_rename_a_module_via_a_global_import:n(8031,e.DiagnosticCategory.Error,"You_cannot_rename_a_module_via_a_global_import_8031","You cannot rename a module via a global import."),Qualified_name_0_is_not_allowed_without_a_leading_param_object_1:n(8032,e.DiagnosticCategory.Error,"Qualified_name_0_is_not_allowed_without_a_leading_param_object_1_8032","Qualified name '{0}' is not allowed without a leading '@param {object} {1}'."),Only_identifiers_Slashqualified_names_with_optional_type_arguments_are_currently_supported_in_a_class_extends_clause:n(9002,e.DiagnosticCategory.Error,"Only_identifiers_Slashqualified_names_with_optional_type_arguments_are_currently_supported_in_a_clas_9002","Only identifiers/qualified-names with optional type arguments are currently supported in a class 'extends' clause."),class_expressions_are_not_currently_supported:n(9003,e.DiagnosticCategory.Error,"class_expressions_are_not_currently_supported_9003","'class' expressions are not currently supported."),Language_service_is_disabled:n(9004,e.DiagnosticCategory.Error,"Language_service_is_disabled_9004","Language service is disabled."),JSX_attributes_must_only_be_assigned_a_non_empty_expression:n(17e3,e.DiagnosticCategory.Error,"JSX_attributes_must_only_be_assigned_a_non_empty_expression_17000","JSX attributes must only be assigned a non-empty 'expression'."),JSX_elements_cannot_have_multiple_attributes_with_the_same_name:n(17001,e.DiagnosticCategory.Error,"JSX_elements_cannot_have_multiple_attributes_with_the_same_name_17001","JSX elements cannot have multiple attributes with the same name."),Expected_corresponding_JSX_closing_tag_for_0:n(17002,e.DiagnosticCategory.Error,"Expected_corresponding_JSX_closing_tag_for_0_17002","Expected corresponding JSX closing tag for '{0}'."),JSX_attribute_expected:n(17003,e.DiagnosticCategory.Error,"JSX_attribute_expected_17003","JSX attribute expected."),Cannot_use_JSX_unless_the_jsx_flag_is_provided:n(17004,e.DiagnosticCategory.Error,"Cannot_use_JSX_unless_the_jsx_flag_is_provided_17004","Cannot use JSX unless the '--jsx' flag is provided."),A_constructor_cannot_contain_a_super_call_when_its_class_extends_null:n(17005,e.DiagnosticCategory.Error,"A_constructor_cannot_contain_a_super_call_when_its_class_extends_null_17005","A constructor cannot contain a 'super' call when its class extends 'null'."),An_unary_expression_with_the_0_operator_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses:n(17006,e.DiagnosticCategory.Error,"An_unary_expression_with_the_0_operator_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_ex_17006","An unary expression with the '{0}' operator is not allowed in the left-hand side of an exponentiation expression. Consider enclosing the expression in parentheses."),A_type_assertion_expression_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses:n(17007,e.DiagnosticCategory.Error,"A_type_assertion_expression_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Con_17007","A type assertion expression is not allowed in the left-hand side of an exponentiation expression. Consider enclosing the expression in parentheses."),JSX_element_0_has_no_corresponding_closing_tag:n(17008,e.DiagnosticCategory.Error,"JSX_element_0_has_no_corresponding_closing_tag_17008","JSX element '{0}' has no corresponding closing tag."),super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class:n(17009,e.DiagnosticCategory.Error,"super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class_17009","'super' must be called before accessing 'this' in the constructor of a derived class."),Unknown_type_acquisition_option_0:n(17010,e.DiagnosticCategory.Error,"Unknown_type_acquisition_option_0_17010","Unknown type acquisition option '{0}'."),super_must_be_called_before_accessing_a_property_of_super_in_the_constructor_of_a_derived_class:n(17011,e.DiagnosticCategory.Error,"super_must_be_called_before_accessing_a_property_of_super_in_the_constructor_of_a_derived_class_17011","'super' must be called before accessing a property of 'super' in the constructor of a derived class."),_0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2:n(17012,e.DiagnosticCategory.Error,"_0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2_17012","'{0}' is not a valid meta-property for keyword '{1}'. Did you mean '{2}'?"),Meta_property_0_is_only_allowed_in_the_body_of_a_function_declaration_function_expression_or_constructor:n(17013,e.DiagnosticCategory.Error,"Meta_property_0_is_only_allowed_in_the_body_of_a_function_declaration_function_expression_or_constru_17013","Meta-property '{0}' is only allowed in the body of a function declaration, function expression, or constructor."),JSX_fragment_has_no_corresponding_closing_tag:n(17014,e.DiagnosticCategory.Error,"JSX_fragment_has_no_corresponding_closing_tag_17014","JSX fragment has no corresponding closing tag."),Expected_corresponding_closing_tag_for_JSX_fragment:n(17015,e.DiagnosticCategory.Error,"Expected_corresponding_closing_tag_for_JSX_fragment_17015","Expected corresponding closing tag for JSX fragment."),JSX_fragment_is_not_supported_when_using_jsxFactory:n(17016,e.DiagnosticCategory.Error,"JSX_fragment_is_not_supported_when_using_jsxFactory_17016","JSX fragment is not supported when using --jsxFactory"),JSX_fragment_is_not_supported_when_using_an_inline_JSX_factory_pragma:n(17017,e.DiagnosticCategory.Error,"JSX_fragment_is_not_supported_when_using_an_inline_JSX_factory_pragma_17017","JSX fragment is not supported when using an inline JSX factory pragma"),Circularity_detected_while_resolving_configuration_Colon_0:n(18e3,e.DiagnosticCategory.Error,"Circularity_detected_while_resolving_configuration_Colon_0_18000","Circularity detected while resolving configuration: {0}"),A_path_in_an_extends_option_must_be_relative_or_rooted_but_0_is_not:n(18001,e.DiagnosticCategory.Error,"A_path_in_an_extends_option_must_be_relative_or_rooted_but_0_is_not_18001","A path in an 'extends' option must be relative or rooted, but '{0}' is not."),The_files_list_in_config_file_0_is_empty:n(18002,e.DiagnosticCategory.Error,"The_files_list_in_config_file_0_is_empty_18002","The 'files' list in config file '{0}' is empty."),No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2:n(18003,e.DiagnosticCategory.Error,"No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2_18003","No inputs were found in config file '{0}'. Specified 'include' paths were '{1}' and 'exclude' paths were '{2}'."),File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module:n(80001,e.DiagnosticCategory.Suggestion,"File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module_80001","File is a CommonJS module; it may be converted to an ES6 module."),This_constructor_function_may_be_converted_to_a_class_declaration:n(80002,e.DiagnosticCategory.Suggestion,"This_constructor_function_may_be_converted_to_a_class_declaration_80002","This constructor function may be converted to a class declaration."),Import_may_be_converted_to_a_default_import:n(80003,e.DiagnosticCategory.Suggestion,"Import_may_be_converted_to_a_default_import_80003","Import may be converted to a default import."),JSDoc_types_may_be_moved_to_TypeScript_types:n(80004,e.DiagnosticCategory.Suggestion,"JSDoc_types_may_be_moved_to_TypeScript_types_80004","JSDoc types may be moved to TypeScript types."),require_call_may_be_converted_to_an_import:n(80005,e.DiagnosticCategory.Suggestion,"require_call_may_be_converted_to_an_import_80005","'require' call may be converted to an import."),This_may_be_converted_to_an_async_function:n(80006,e.DiagnosticCategory.Suggestion,"This_may_be_converted_to_an_async_function_80006","This may be converted to an async function."),Add_missing_super_call:n(90001,e.DiagnosticCategory.Message,"Add_missing_super_call_90001","Add missing 'super()' call"),Make_super_call_the_first_statement_in_the_constructor:n(90002,e.DiagnosticCategory.Message,"Make_super_call_the_first_statement_in_the_constructor_90002","Make 'super()' call the first statement in the constructor"),Change_extends_to_implements:n(90003,e.DiagnosticCategory.Message,"Change_extends_to_implements_90003","Change 'extends' to 'implements'"),Remove_declaration_for_Colon_0:n(90004,e.DiagnosticCategory.Message,"Remove_declaration_for_Colon_0_90004","Remove declaration for: '{0}'"),Remove_import_from_0:n(90005,e.DiagnosticCategory.Message,"Remove_import_from_0_90005","Remove import from '{0}'"),Implement_interface_0:n(90006,e.DiagnosticCategory.Message,"Implement_interface_0_90006","Implement interface '{0}'"),Implement_inherited_abstract_class:n(90007,e.DiagnosticCategory.Message,"Implement_inherited_abstract_class_90007","Implement inherited abstract class"),Add_0_to_unresolved_variable:n(90008,e.DiagnosticCategory.Message,"Add_0_to_unresolved_variable_90008","Add '{0}.' to unresolved variable"),Remove_destructuring:n(90009,e.DiagnosticCategory.Message,"Remove_destructuring_90009","Remove destructuring"),Remove_variable_statement:n(90010,e.DiagnosticCategory.Message,"Remove_variable_statement_90010","Remove variable statement"),Remove_template_tag:n(90011,e.DiagnosticCategory.Message,"Remove_template_tag_90011","Remove template tag"),Remove_type_parameters:n(90012,e.DiagnosticCategory.Message,"Remove_type_parameters_90012","Remove type parameters"),Import_0_from_module_1:n(90013,e.DiagnosticCategory.Message,"Import_0_from_module_1_90013","Import '{0}' from module \"{1}\""),Change_0_to_1:n(90014,e.DiagnosticCategory.Message,"Change_0_to_1_90014","Change '{0}' to '{1}'"),Add_0_to_existing_import_declaration_from_1:n(90015,e.DiagnosticCategory.Message,"Add_0_to_existing_import_declaration_from_1_90015","Add '{0}' to existing import declaration from \"{1}\""),Declare_property_0:n(90016,e.DiagnosticCategory.Message,"Declare_property_0_90016","Declare property '{0}'"),Add_index_signature_for_property_0:n(90017,e.DiagnosticCategory.Message,"Add_index_signature_for_property_0_90017","Add index signature for property '{0}'"),Disable_checking_for_this_file:n(90018,e.DiagnosticCategory.Message,"Disable_checking_for_this_file_90018","Disable checking for this file"),Ignore_this_error_message:n(90019,e.DiagnosticCategory.Message,"Ignore_this_error_message_90019","Ignore this error message"),Initialize_property_0_in_the_constructor:n(90020,e.DiagnosticCategory.Message,"Initialize_property_0_in_the_constructor_90020","Initialize property '{0}' in the constructor"),Initialize_static_property_0:n(90021,e.DiagnosticCategory.Message,"Initialize_static_property_0_90021","Initialize static property '{0}'"),Change_spelling_to_0:n(90022,e.DiagnosticCategory.Message,"Change_spelling_to_0_90022","Change spelling to '{0}'"),Declare_method_0:n(90023,e.DiagnosticCategory.Message,"Declare_method_0_90023","Declare method '{0}'"),Declare_static_method_0:n(90024,e.DiagnosticCategory.Message,"Declare_static_method_0_90024","Declare static method '{0}'"),Prefix_0_with_an_underscore:n(90025,e.DiagnosticCategory.Message,"Prefix_0_with_an_underscore_90025","Prefix '{0}' with an underscore"),Rewrite_as_the_indexed_access_type_0:n(90026,e.DiagnosticCategory.Message,"Rewrite_as_the_indexed_access_type_0_90026","Rewrite as the indexed access type '{0}'"),Declare_static_property_0:n(90027,e.DiagnosticCategory.Message,"Declare_static_property_0_90027","Declare static property '{0}'"),Call_decorator_expression:n(90028,e.DiagnosticCategory.Message,"Call_decorator_expression_90028","Call decorator expression"),Add_async_modifier_to_containing_function:n(90029,e.DiagnosticCategory.Message,"Add_async_modifier_to_containing_function_90029","Add async modifier to containing function"),Replace_infer_0_with_unknown:n(90030,e.DiagnosticCategory.Message,"Replace_infer_0_with_unknown_90030","Replace 'infer {0}' with 'unknown'"),Replace_all_unused_infer_with_unknown:n(90031,e.DiagnosticCategory.Message,"Replace_all_unused_infer_with_unknown_90031","Replace all unused 'infer' with 'unknown'"),Import_default_0_from_module_1:n(90032,e.DiagnosticCategory.Message,"Import_default_0_from_module_1_90032","Import default '{0}' from module \"{1}\""),Add_default_import_0_to_existing_import_declaration_from_1:n(90033,e.DiagnosticCategory.Message,"Add_default_import_0_to_existing_import_declaration_from_1_90033","Add default import '{0}' to existing import declaration from \"{1}\""),Add_parameter_name:n(90034,e.DiagnosticCategory.Message,"Add_parameter_name_90034","Add parameter name"),Convert_function_to_an_ES2015_class:n(95001,e.DiagnosticCategory.Message,"Convert_function_to_an_ES2015_class_95001","Convert function to an ES2015 class"),Convert_function_0_to_class:n(95002,e.DiagnosticCategory.Message,"Convert_function_0_to_class_95002","Convert function '{0}' to class"),Extract_to_0_in_1:n(95004,e.DiagnosticCategory.Message,"Extract_to_0_in_1_95004","Extract to {0} in {1}"),Extract_function:n(95005,e.DiagnosticCategory.Message,"Extract_function_95005","Extract function"),Extract_constant:n(95006,e.DiagnosticCategory.Message,"Extract_constant_95006","Extract constant"),Extract_to_0_in_enclosing_scope:n(95007,e.DiagnosticCategory.Message,"Extract_to_0_in_enclosing_scope_95007","Extract to {0} in enclosing scope"),Extract_to_0_in_1_scope:n(95008,e.DiagnosticCategory.Message,"Extract_to_0_in_1_scope_95008","Extract to {0} in {1} scope"),Annotate_with_type_from_JSDoc:n(95009,e.DiagnosticCategory.Message,"Annotate_with_type_from_JSDoc_95009","Annotate with type from JSDoc"),Annotate_with_types_from_JSDoc:n(95010,e.DiagnosticCategory.Message,"Annotate_with_types_from_JSDoc_95010","Annotate with types from JSDoc"),Infer_type_of_0_from_usage:n(95011,e.DiagnosticCategory.Message,"Infer_type_of_0_from_usage_95011","Infer type of '{0}' from usage"),Infer_parameter_types_from_usage:n(95012,e.DiagnosticCategory.Message,"Infer_parameter_types_from_usage_95012","Infer parameter types from usage"),Convert_to_default_import:n(95013,e.DiagnosticCategory.Message,"Convert_to_default_import_95013","Convert to default import"),Install_0:n(95014,e.DiagnosticCategory.Message,"Install_0_95014","Install '{0}'"),Replace_import_with_0:n(95015,e.DiagnosticCategory.Message,"Replace_import_with_0_95015","Replace import with '{0}'."),Use_synthetic_default_member:n(95016,e.DiagnosticCategory.Message,"Use_synthetic_default_member_95016","Use synthetic 'default' member."),Convert_to_ES6_module:n(95017,e.DiagnosticCategory.Message,"Convert_to_ES6_module_95017","Convert to ES6 module"),Add_undefined_type_to_property_0:n(95018,e.DiagnosticCategory.Message,"Add_undefined_type_to_property_0_95018","Add 'undefined' type to property '{0}'"),Add_initializer_to_property_0:n(95019,e.DiagnosticCategory.Message,"Add_initializer_to_property_0_95019","Add initializer to property '{0}'"),Add_definite_assignment_assertion_to_property_0:n(95020,e.DiagnosticCategory.Message,"Add_definite_assignment_assertion_to_property_0_95020","Add definite assignment assertion to property '{0}'"),Add_all_missing_members:n(95022,e.DiagnosticCategory.Message,"Add_all_missing_members_95022","Add all missing members"),Infer_all_types_from_usage:n(95023,e.DiagnosticCategory.Message,"Infer_all_types_from_usage_95023","Infer all types from usage"),Delete_all_unused_declarations:n(95024,e.DiagnosticCategory.Message,"Delete_all_unused_declarations_95024","Delete all unused declarations"),Prefix_all_unused_declarations_with_where_possible:n(95025,e.DiagnosticCategory.Message,"Prefix_all_unused_declarations_with_where_possible_95025","Prefix all unused declarations with '_' where possible"),Fix_all_detected_spelling_errors:n(95026,e.DiagnosticCategory.Message,"Fix_all_detected_spelling_errors_95026","Fix all detected spelling errors"),Add_initializers_to_all_uninitialized_properties:n(95027,e.DiagnosticCategory.Message,"Add_initializers_to_all_uninitialized_properties_95027","Add initializers to all uninitialized properties"),Add_definite_assignment_assertions_to_all_uninitialized_properties:n(95028,e.DiagnosticCategory.Message,"Add_definite_assignment_assertions_to_all_uninitialized_properties_95028","Add definite assignment assertions to all uninitialized properties"),Add_undefined_type_to_all_uninitialized_properties:n(95029,e.DiagnosticCategory.Message,"Add_undefined_type_to_all_uninitialized_properties_95029","Add undefined type to all uninitialized properties"),Change_all_jsdoc_style_types_to_TypeScript:n(95030,e.DiagnosticCategory.Message,"Change_all_jsdoc_style_types_to_TypeScript_95030","Change all jsdoc-style types to TypeScript"),Change_all_jsdoc_style_types_to_TypeScript_and_add_undefined_to_nullable_types:n(95031,e.DiagnosticCategory.Message,"Change_all_jsdoc_style_types_to_TypeScript_and_add_undefined_to_nullable_types_95031","Change all jsdoc-style types to TypeScript (and add '| undefined' to nullable types)"),Implement_all_unimplemented_interfaces:n(95032,e.DiagnosticCategory.Message,"Implement_all_unimplemented_interfaces_95032","Implement all unimplemented interfaces"),Install_all_missing_types_packages:n(95033,e.DiagnosticCategory.Message,"Install_all_missing_types_packages_95033","Install all missing types packages"),Rewrite_all_as_indexed_access_types:n(95034,e.DiagnosticCategory.Message,"Rewrite_all_as_indexed_access_types_95034","Rewrite all as indexed access types"),Convert_all_to_default_imports:n(95035,e.DiagnosticCategory.Message,"Convert_all_to_default_imports_95035","Convert all to default imports"),Make_all_super_calls_the_first_statement_in_their_constructor:n(95036,e.DiagnosticCategory.Message,"Make_all_super_calls_the_first_statement_in_their_constructor_95036","Make all 'super()' calls the first statement in their constructor"),Add_qualifier_to_all_unresolved_variables_matching_a_member_name:n(95037,e.DiagnosticCategory.Message,"Add_qualifier_to_all_unresolved_variables_matching_a_member_name_95037","Add qualifier to all unresolved variables matching a member name"),Change_all_extended_interfaces_to_implements:n(95038,e.DiagnosticCategory.Message,"Change_all_extended_interfaces_to_implements_95038","Change all extended interfaces to 'implements'"),Add_all_missing_super_calls:n(95039,e.DiagnosticCategory.Message,"Add_all_missing_super_calls_95039","Add all missing super calls"),Implement_all_inherited_abstract_classes:n(95040,e.DiagnosticCategory.Message,"Implement_all_inherited_abstract_classes_95040","Implement all inherited abstract classes"),Add_all_missing_async_modifiers:n(95041,e.DiagnosticCategory.Message,"Add_all_missing_async_modifiers_95041","Add all missing 'async' modifiers"),Add_ts_ignore_to_all_error_messages:n(95042,e.DiagnosticCategory.Message,"Add_ts_ignore_to_all_error_messages_95042","Add '@ts-ignore' to all error messages"),Annotate_everything_with_types_from_JSDoc:n(95043,e.DiagnosticCategory.Message,"Annotate_everything_with_types_from_JSDoc_95043","Annotate everything with types from JSDoc"),Add_to_all_uncalled_decorators:n(95044,e.DiagnosticCategory.Message,"Add_to_all_uncalled_decorators_95044","Add '()' to all uncalled decorators"),Convert_all_constructor_functions_to_classes:n(95045,e.DiagnosticCategory.Message,"Convert_all_constructor_functions_to_classes_95045","Convert all constructor functions to classes"),Generate_get_and_set_accessors:n(95046,e.DiagnosticCategory.Message,"Generate_get_and_set_accessors_95046","Generate 'get' and 'set' accessors"),Convert_require_to_import:n(95047,e.DiagnosticCategory.Message,"Convert_require_to_import_95047","Convert 'require' to 'import'"),Convert_all_require_to_import:n(95048,e.DiagnosticCategory.Message,"Convert_all_require_to_import_95048","Convert all 'require' to 'import'"),Move_to_a_new_file:n(95049,e.DiagnosticCategory.Message,"Move_to_a_new_file_95049","Move to a new file"),Remove_unreachable_code:n(95050,e.DiagnosticCategory.Message,"Remove_unreachable_code_95050","Remove unreachable code"),Remove_all_unreachable_code:n(95051,e.DiagnosticCategory.Message,"Remove_all_unreachable_code_95051","Remove all unreachable code"),Add_missing_typeof:n(95052,e.DiagnosticCategory.Message,"Add_missing_typeof_95052","Add missing 'typeof'"),Remove_unused_label:n(95053,e.DiagnosticCategory.Message,"Remove_unused_label_95053","Remove unused label"),Remove_all_unused_labels:n(95054,e.DiagnosticCategory.Message,"Remove_all_unused_labels_95054","Remove all unused labels"),Convert_0_to_mapped_object_type:n(95055,e.DiagnosticCategory.Message,"Convert_0_to_mapped_object_type_95055","Convert '{0}' to mapped object type"),Convert_namespace_import_to_named_imports:n(95056,e.DiagnosticCategory.Message,"Convert_namespace_import_to_named_imports_95056","Convert namespace import to named imports"),Convert_named_imports_to_namespace_import:n(95057,e.DiagnosticCategory.Message,"Convert_named_imports_to_namespace_import_95057","Convert named imports to namespace import"),Add_or_remove_braces_in_an_arrow_function:n(95058,e.DiagnosticCategory.Message,"Add_or_remove_braces_in_an_arrow_function_95058","Add or remove braces in an arrow function"),Add_braces_to_arrow_function:n(95059,e.DiagnosticCategory.Message,"Add_braces_to_arrow_function_95059","Add braces to arrow function"),Remove_braces_from_arrow_function:n(95060,e.DiagnosticCategory.Message,"Remove_braces_from_arrow_function_95060","Remove braces from arrow function"),Convert_default_export_to_named_export:n(95061,e.DiagnosticCategory.Message,"Convert_default_export_to_named_export_95061","Convert default export to named export"),Convert_named_export_to_default_export:n(95062,e.DiagnosticCategory.Message,"Convert_named_export_to_default_export_95062","Convert named export to default export"),Add_missing_enum_member_0:n(95063,e.DiagnosticCategory.Message,"Add_missing_enum_member_0_95063","Add missing enum member '{0}'"),Add_all_missing_imports:n(95064,e.DiagnosticCategory.Message,"Add_all_missing_imports_95064","Add all missing imports"),Convert_to_async_function:n(95065,e.DiagnosticCategory.Message,"Convert_to_async_function_95065","Convert to async function"),Convert_all_to_async_functions:n(95066,e.DiagnosticCategory.Message,"Convert_all_to_async_functions_95066","Convert all to async functions"),Generate_types_for_0:n(95067,e.DiagnosticCategory.Message,"Generate_types_for_0_95067","Generate types for '{0}'"),Generate_types_for_all_packages_without_types:n(95068,e.DiagnosticCategory.Message,"Generate_types_for_all_packages_without_types_95068","Generate types for all packages without types"),Add_unknown_conversion_for_non_overlapping_types:n(95069,e.DiagnosticCategory.Message,"Add_unknown_conversion_for_non_overlapping_types_95069","Add 'unknown' conversion for non-overlapping types"),Add_unknown_to_all_conversions_of_non_overlapping_types:n(95070,e.DiagnosticCategory.Message,"Add_unknown_to_all_conversions_of_non_overlapping_types_95070","Add 'unknown' to all conversions of non-overlapping types"),Add_missing_new_operator_to_call:n(95071,e.DiagnosticCategory.Message,"Add_missing_new_operator_to_call_95071","Add missing 'new' operator to call"),Add_missing_new_operator_to_all_calls:n(95072,e.DiagnosticCategory.Message,"Add_missing_new_operator_to_all_calls_95072","Add missing 'new' operator to all calls"),Add_names_to_all_parameters_without_names:n(95073,e.DiagnosticCategory.Message,"Add_names_to_all_parameters_without_names_95073","Add names to all parameters without names"),Enable_the_experimentalDecorators_option_in_your_configuration_file:n(95074,e.DiagnosticCategory.Message,"Enable_the_experimentalDecorators_option_in_your_configuration_file_95074","Enable the 'experimentalDecorators' option in your configuration file")}}(u||(u={})),function(e){var n;function t(e){return e>=72}function r(e){return 30===e||t(e)}e.tokenIsIdentifierOrKeyword=t,e.tokenIsIdentifierOrKeywordOrGreaterThan=r;var a=(n={abstract:118,any:120,as:119,bigint:146,boolean:123,break:73,case:74,catch:75,class:76,continue:78,const:77},n["constructor"]=124,n.debugger=79,n.declare=125,n.default=80,n.delete=81,n.do=82,n.else=83,n.enum=84,n.export=85,n.extends=86,n.false=87,n.finally=88,n.for=89,n.from=144,n.function=90,n.get=126,n.if=91,n.implements=109,n.import=92,n.in=93,n.infer=127,n.instanceof=94,n.interface=110,n.is=128,n.keyof=129,n.let=111,n.module=130,n.namespace=131,n.never=132,n.new=95,n.null=96,n.number=135,n.object=136,n.package=112,n.private=113,n.protected=114,n.public=115,n.readonly=133,n.require=134,n.global=145,n.return=97,n.set=137,n.static=116,n.string=138,n.super=98,n.switch=99,n.symbol=139,n.this=100,n.throw=101,n.true=102,n.try=103,n.type=140,n.typeof=104,n.undefined=141,n.unique=142,n.unknown=143,n.var=105,n.void=106,n.while=107,n.with=108,n.yield=117,n.async=121,n.await=122,n.of=147,n),i=e.createMapFromTemplate(a),s=e.createMapFromTemplate(o({},a,{"{":18,"}":19,"(":20,")":21,"[":22,"]":23,".":24,"...":25,";":26,",":27,"<":28,">":30,"<=":31,">=":32,"==":33,"!=":34,"===":35,"!==":36,"=>":37,"+":38,"-":39,"**":41,"*":40,"/":42,"%":43,"++":44,"--":45,"<<":46,"</":29,">>":47,">>>":48,"&":49,"|":50,"^":51,"!":52,"~":53,"&&":54,"||":55,"?":56,":":57,"=":59,"+=":60,"-=":61,"*=":62,"**=":63,"/=":64,"%=":65,"<<=":66,">>=":67,">>>=":68,"&=":69,"|=":70,"^=":71,"@":58})),c=[170,170,181,181,186,186,192,214,216,246,248,543,546,563,592,685,688,696,699,705,720,721,736,740,750,750,890,890,902,902,904,906,908,908,910,929,931,974,976,983,986,1011,1024,1153,1164,1220,1223,1224,1227,1228,1232,1269,1272,1273,1329,1366,1369,1369,1377,1415,1488,1514,1520,1522,1569,1594,1600,1610,1649,1747,1749,1749,1765,1766,1786,1788,1808,1808,1810,1836,1920,1957,2309,2361,2365,2365,2384,2384,2392,2401,2437,2444,2447,2448,2451,2472,2474,2480,2482,2482,2486,2489,2524,2525,2527,2529,2544,2545,2565,2570,2575,2576,2579,2600,2602,2608,2610,2611,2613,2614,2616,2617,2649,2652,2654,2654,2674,2676,2693,2699,2701,2701,2703,2705,2707,2728,2730,2736,2738,2739,2741,2745,2749,2749,2768,2768,2784,2784,2821,2828,2831,2832,2835,2856,2858,2864,2866,2867,2870,2873,2877,2877,2908,2909,2911,2913,2949,2954,2958,2960,2962,2965,2969,2970,2972,2972,2974,2975,2979,2980,2984,2986,2990,2997,2999,3001,3077,3084,3086,3088,3090,3112,3114,3123,3125,3129,3168,3169,3205,3212,3214,3216,3218,3240,3242,3251,3253,3257,3294,3294,3296,3297,3333,3340,3342,3344,3346,3368,3370,3385,3424,3425,3461,3478,3482,3505,3507,3515,3517,3517,3520,3526,3585,3632,3634,3635,3648,3654,3713,3714,3716,3716,3719,3720,3722,3722,3725,3725,3732,3735,3737,3743,3745,3747,3749,3749,3751,3751,3754,3755,3757,3760,3762,3763,3773,3773,3776,3780,3782,3782,3804,3805,3840,3840,3904,3911,3913,3946,3976,3979,4096,4129,4131,4135,4137,4138,4176,4181,4256,4293,4304,4342,4352,4441,4447,4514,4520,4601,4608,4614,4616,4678,4680,4680,4682,4685,4688,4694,4696,4696,4698,4701,4704,4742,4744,4744,4746,4749,4752,4782,4784,4784,4786,4789,4792,4798,4800,4800,4802,4805,4808,4814,4816,4822,4824,4846,4848,4878,4880,4880,4882,4885,4888,4894,4896,4934,4936,4954,5024,5108,5121,5740,5743,5750,5761,5786,5792,5866,6016,6067,6176,6263,6272,6312,7680,7835,7840,7929,7936,7957,7960,7965,7968,8005,8008,8013,8016,8023,8025,8025,8027,8027,8029,8029,8031,8061,8064,8116,8118,8124,8126,8126,8130,8132,8134,8140,8144,8147,8150,8155,8160,8172,8178,8180,8182,8188,8319,8319,8450,8450,8455,8455,8458,8467,8469,8469,8473,8477,8484,8484,8486,8486,8488,8488,8490,8493,8495,8497,8499,8505,8544,8579,12293,12295,12321,12329,12337,12341,12344,12346,12353,12436,12445,12446,12449,12538,12540,12542,12549,12588,12593,12686,12704,12727,13312,19893,19968,40869,40960,42124,44032,55203,63744,64045,64256,64262,64275,64279,64285,64285,64287,64296,64298,64310,64312,64316,64318,64318,64320,64321,64323,64324,64326,64433,64467,64829,64848,64911,64914,64967,65008,65019,65136,65138,65140,65140,65142,65276,65313,65338,65345,65370,65382,65470,65474,65479,65482,65487,65490,65495,65498,65500],l=[170,170,181,181,186,186,192,214,216,246,248,543,546,563,592,685,688,696,699,705,720,721,736,740,750,750,768,846,864,866,890,890,902,902,904,906,908,908,910,929,931,974,976,983,986,1011,1024,1153,1155,1158,1164,1220,1223,1224,1227,1228,1232,1269,1272,1273,1329,1366,1369,1369,1377,1415,1425,1441,1443,1465,1467,1469,1471,1471,1473,1474,1476,1476,1488,1514,1520,1522,1569,1594,1600,1621,1632,1641,1648,1747,1749,1756,1759,1768,1770,1773,1776,1788,1808,1836,1840,1866,1920,1968,2305,2307,2309,2361,2364,2381,2384,2388,2392,2403,2406,2415,2433,2435,2437,2444,2447,2448,2451,2472,2474,2480,2482,2482,2486,2489,2492,2492,2494,2500,2503,2504,2507,2509,2519,2519,2524,2525,2527,2531,2534,2545,2562,2562,2565,2570,2575,2576,2579,2600,2602,2608,2610,2611,2613,2614,2616,2617,2620,2620,2622,2626,2631,2632,2635,2637,2649,2652,2654,2654,2662,2676,2689,2691,2693,2699,2701,2701,2703,2705,2707,2728,2730,2736,2738,2739,2741,2745,2748,2757,2759,2761,2763,2765,2768,2768,2784,2784,2790,2799,2817,2819,2821,2828,2831,2832,2835,2856,2858,2864,2866,2867,2870,2873,2876,2883,2887,2888,2891,2893,2902,2903,2908,2909,2911,2913,2918,2927,2946,2947,2949,2954,2958,2960,2962,2965,2969,2970,2972,2972,2974,2975,2979,2980,2984,2986,2990,2997,2999,3001,3006,3010,3014,3016,3018,3021,3031,3031,3047,3055,3073,3075,3077,3084,3086,3088,3090,3112,3114,3123,3125,3129,3134,3140,3142,3144,3146,3149,3157,3158,3168,3169,3174,3183,3202,3203,3205,3212,3214,3216,3218,3240,3242,3251,3253,3257,3262,3268,3270,3272,3274,3277,3285,3286,3294,3294,3296,3297,3302,3311,3330,3331,3333,3340,3342,3344,3346,3368,3370,3385,3390,3395,3398,3400,3402,3405,3415,3415,3424,3425,3430,3439,3458,3459,3461,3478,3482,3505,3507,3515,3517,3517,3520,3526,3530,3530,3535,3540,3542,3542,3544,3551,3570,3571,3585,3642,3648,3662,3664,3673,3713,3714,3716,3716,3719,3720,3722,3722,3725,3725,3732,3735,3737,3743,3745,3747,3749,3749,3751,3751,3754,3755,3757,3769,3771,3773,3776,3780,3782,3782,3784,3789,3792,3801,3804,3805,3840,3840,3864,3865,3872,3881,3893,3893,3895,3895,3897,3897,3902,3911,3913,3946,3953,3972,3974,3979,3984,3991,3993,4028,4038,4038,4096,4129,4131,4135,4137,4138,4140,4146,4150,4153,4160,4169,4176,4185,4256,4293,4304,4342,4352,4441,4447,4514,4520,4601,4608,4614,4616,4678,4680,4680,4682,4685,4688,4694,4696,4696,4698,4701,4704,4742,4744,4744,4746,4749,4752,4782,4784,4784,4786,4789,4792,4798,4800,4800,4802,4805,4808,4814,4816,4822,4824,4846,4848,4878,4880,4880,4882,4885,4888,4894,4896,4934,4936,4954,4969,4977,5024,5108,5121,5740,5743,5750,5761,5786,5792,5866,6016,6099,6112,6121,6160,6169,6176,6263,6272,6313,7680,7835,7840,7929,7936,7957,7960,7965,7968,8005,8008,8013,8016,8023,8025,8025,8027,8027,8029,8029,8031,8061,8064,8116,8118,8124,8126,8126,8130,8132,8134,8140,8144,8147,8150,8155,8160,8172,8178,8180,8182,8188,8255,8256,8319,8319,8400,8412,8417,8417,8450,8450,8455,8455,8458,8467,8469,8469,8473,8477,8484,8484,8486,8486,8488,8488,8490,8493,8495,8497,8499,8505,8544,8579,12293,12295,12321,12335,12337,12341,12344,12346,12353,12436,12441,12442,12445,12446,12449,12542,12549,12588,12593,12686,12704,12727,13312,19893,19968,40869,40960,42124,44032,55203,63744,64045,64256,64262,64275,64279,64285,64296,64298,64310,64312,64316,64318,64318,64320,64321,64323,64324,64326,64433,64467,64829,64848,64911,64914,64967,65008,65019,65056,65059,65075,65076,65101,65103,65136,65138,65140,65140,65142,65276,65296,65305,65313,65338,65343,65343,65345,65370,65381,65470,65474,65479,65482,65487,65490,65495,65498,65500],u=[170,170,181,181,186,186,192,214,216,246,248,705,710,721,736,740,748,748,750,750,880,884,886,887,890,893,902,902,904,906,908,908,910,929,931,1013,1015,1153,1162,1319,1329,1366,1369,1369,1377,1415,1488,1514,1520,1522,1568,1610,1646,1647,1649,1747,1749,1749,1765,1766,1774,1775,1786,1788,1791,1791,1808,1808,1810,1839,1869,1957,1969,1969,1994,2026,2036,2037,2042,2042,2048,2069,2074,2074,2084,2084,2088,2088,2112,2136,2208,2208,2210,2220,2308,2361,2365,2365,2384,2384,2392,2401,2417,2423,2425,2431,2437,2444,2447,2448,2451,2472,2474,2480,2482,2482,2486,2489,2493,2493,2510,2510,2524,2525,2527,2529,2544,2545,2565,2570,2575,2576,2579,2600,2602,2608,2610,2611,2613,2614,2616,2617,2649,2652,2654,2654,2674,2676,2693,2701,2703,2705,2707,2728,2730,2736,2738,2739,2741,2745,2749,2749,2768,2768,2784,2785,2821,2828,2831,2832,2835,2856,2858,2864,2866,2867,2869,2873,2877,2877,2908,2909,2911,2913,2929,2929,2947,2947,2949,2954,2958,2960,2962,2965,2969,2970,2972,2972,2974,2975,2979,2980,2984,2986,2990,3001,3024,3024,3077,3084,3086,3088,3090,3112,3114,3123,3125,3129,3133,3133,3160,3161,3168,3169,3205,3212,3214,3216,3218,3240,3242,3251,3253,3257,3261,3261,3294,3294,3296,3297,3313,3314,3333,3340,3342,3344,3346,3386,3389,3389,3406,3406,3424,3425,3450,3455,3461,3478,3482,3505,3507,3515,3517,3517,3520,3526,3585,3632,3634,3635,3648,3654,3713,3714,3716,3716,3719,3720,3722,3722,3725,3725,3732,3735,3737,3743,3745,3747,3749,3749,3751,3751,3754,3755,3757,3760,3762,3763,3773,3773,3776,3780,3782,3782,3804,3807,3840,3840,3904,3911,3913,3948,3976,3980,4096,4138,4159,4159,4176,4181,4186,4189,4193,4193,4197,4198,4206,4208,4213,4225,4238,4238,4256,4293,4295,4295,4301,4301,4304,4346,4348,4680,4682,4685,4688,4694,4696,4696,4698,4701,4704,4744,4746,4749,4752,4784,4786,4789,4792,4798,4800,4800,4802,4805,4808,4822,4824,4880,4882,4885,4888,4954,4992,5007,5024,5108,5121,5740,5743,5759,5761,5786,5792,5866,5870,5872,5888,5900,5902,5905,5920,5937,5952,5969,5984,5996,5998,6e3,6016,6067,6103,6103,6108,6108,6176,6263,6272,6312,6314,6314,6320,6389,6400,6428,6480,6509,6512,6516,6528,6571,6593,6599,6656,6678,6688,6740,6823,6823,6917,6963,6981,6987,7043,7072,7086,7087,7098,7141,7168,7203,7245,7247,7258,7293,7401,7404,7406,7409,7413,7414,7424,7615,7680,7957,7960,7965,7968,8005,8008,8013,8016,8023,8025,8025,8027,8027,8029,8029,8031,8061,8064,8116,8118,8124,8126,8126,8130,8132,8134,8140,8144,8147,8150,8155,8160,8172,8178,8180,8182,8188,8305,8305,8319,8319,8336,8348,8450,8450,8455,8455,8458,8467,8469,8469,8473,8477,8484,8484,8486,8486,8488,8488,8490,8493,8495,8505,8508,8511,8517,8521,8526,8526,8544,8584,11264,11310,11312,11358,11360,11492,11499,11502,11506,11507,11520,11557,11559,11559,11565,11565,11568,11623,11631,11631,11648,11670,11680,11686,11688,11694,11696,11702,11704,11710,11712,11718,11720,11726,11728,11734,11736,11742,11823,11823,12293,12295,12321,12329,12337,12341,12344,12348,12353,12438,12445,12447,12449,12538,12540,12543,12549,12589,12593,12686,12704,12730,12784,12799,13312,19893,19968,40908,40960,42124,42192,42237,42240,42508,42512,42527,42538,42539,42560,42606,42623,42647,42656,42735,42775,42783,42786,42888,42891,42894,42896,42899,42912,42922,43e3,43009,43011,43013,43015,43018,43020,43042,43072,43123,43138,43187,43250,43255,43259,43259,43274,43301,43312,43334,43360,43388,43396,43442,43471,43471,43520,43560,43584,43586,43588,43595,43616,43638,43642,43642,43648,43695,43697,43697,43701,43702,43705,43709,43712,43712,43714,43714,43739,43741,43744,43754,43762,43764,43777,43782,43785,43790,43793,43798,43808,43814,43816,43822,43968,44002,44032,55203,55216,55238,55243,55291,63744,64109,64112,64217,64256,64262,64275,64279,64285,64285,64287,64296,64298,64310,64312,64316,64318,64318,64320,64321,64323,64324,64326,64433,64467,64829,64848,64911,64914,64967,65008,65019,65136,65140,65142,65276,65313,65338,65345,65370,65382,65470,65474,65479,65482,65487,65490,65495,65498,65500],d=[170,170,181,181,186,186,192,214,216,246,248,705,710,721,736,740,748,748,750,750,768,884,886,887,890,893,902,902,904,906,908,908,910,929,931,1013,1015,1153,1155,1159,1162,1319,1329,1366,1369,1369,1377,1415,1425,1469,1471,1471,1473,1474,1476,1477,1479,1479,1488,1514,1520,1522,1552,1562,1568,1641,1646,1747,1749,1756,1759,1768,1770,1788,1791,1791,1808,1866,1869,1969,1984,2037,2042,2042,2048,2093,2112,2139,2208,2208,2210,2220,2276,2302,2304,2403,2406,2415,2417,2423,2425,2431,2433,2435,2437,2444,2447,2448,2451,2472,2474,2480,2482,2482,2486,2489,2492,2500,2503,2504,2507,2510,2519,2519,2524,2525,2527,2531,2534,2545,2561,2563,2565,2570,2575,2576,2579,2600,2602,2608,2610,2611,2613,2614,2616,2617,2620,2620,2622,2626,2631,2632,2635,2637,2641,2641,2649,2652,2654,2654,2662,2677,2689,2691,2693,2701,2703,2705,2707,2728,2730,2736,2738,2739,2741,2745,2748,2757,2759,2761,2763,2765,2768,2768,2784,2787,2790,2799,2817,2819,2821,2828,2831,2832,2835,2856,2858,2864,2866,2867,2869,2873,2876,2884,2887,2888,2891,2893,2902,2903,2908,2909,2911,2915,2918,2927,2929,2929,2946,2947,2949,2954,2958,2960,2962,2965,2969,2970,2972,2972,2974,2975,2979,2980,2984,2986,2990,3001,3006,3010,3014,3016,3018,3021,3024,3024,3031,3031,3046,3055,3073,3075,3077,3084,3086,3088,3090,3112,3114,3123,3125,3129,3133,3140,3142,3144,3146,3149,3157,3158,3160,3161,3168,3171,3174,3183,3202,3203,3205,3212,3214,3216,3218,3240,3242,3251,3253,3257,3260,3268,3270,3272,3274,3277,3285,3286,3294,3294,3296,3299,3302,3311,3313,3314,3330,3331,3333,3340,3342,3344,3346,3386,3389,3396,3398,3400,3402,3406,3415,3415,3424,3427,3430,3439,3450,3455,3458,3459,3461,3478,3482,3505,3507,3515,3517,3517,3520,3526,3530,3530,3535,3540,3542,3542,3544,3551,3570,3571,3585,3642,3648,3662,3664,3673,3713,3714,3716,3716,3719,3720,3722,3722,3725,3725,3732,3735,3737,3743,3745,3747,3749,3749,3751,3751,3754,3755,3757,3769,3771,3773,3776,3780,3782,3782,3784,3789,3792,3801,3804,3807,3840,3840,3864,3865,3872,3881,3893,3893,3895,3895,3897,3897,3902,3911,3913,3948,3953,3972,3974,3991,3993,4028,4038,4038,4096,4169,4176,4253,4256,4293,4295,4295,4301,4301,4304,4346,4348,4680,4682,4685,4688,4694,4696,4696,4698,4701,4704,4744,4746,4749,4752,4784,4786,4789,4792,4798,4800,4800,4802,4805,4808,4822,4824,4880,4882,4885,4888,4954,4957,4959,4992,5007,5024,5108,5121,5740,5743,5759,5761,5786,5792,5866,5870,5872,5888,5900,5902,5908,5920,5940,5952,5971,5984,5996,5998,6e3,6002,6003,6016,6099,6103,6103,6108,6109,6112,6121,6155,6157,6160,6169,6176,6263,6272,6314,6320,6389,6400,6428,6432,6443,6448,6459,6470,6509,6512,6516,6528,6571,6576,6601,6608,6617,6656,6683,6688,6750,6752,6780,6783,6793,6800,6809,6823,6823,6912,6987,6992,7001,7019,7027,7040,7155,7168,7223,7232,7241,7245,7293,7376,7378,7380,7414,7424,7654,7676,7957,7960,7965,7968,8005,8008,8013,8016,8023,8025,8025,8027,8027,8029,8029,8031,8061,8064,8116,8118,8124,8126,8126,8130,8132,8134,8140,8144,8147,8150,8155,8160,8172,8178,8180,8182,8188,8204,8205,8255,8256,8276,8276,8305,8305,8319,8319,8336,8348,8400,8412,8417,8417,8421,8432,8450,8450,8455,8455,8458,8467,8469,8469,8473,8477,8484,8484,8486,8486,8488,8488,8490,8493,8495,8505,8508,8511,8517,8521,8526,8526,8544,8584,11264,11310,11312,11358,11360,11492,11499,11507,11520,11557,11559,11559,11565,11565,11568,11623,11631,11631,11647,11670,11680,11686,11688,11694,11696,11702,11704,11710,11712,11718,11720,11726,11728,11734,11736,11742,11744,11775,11823,11823,12293,12295,12321,12335,12337,12341,12344,12348,12353,12438,12441,12442,12445,12447,12449,12538,12540,12543,12549,12589,12593,12686,12704,12730,12784,12799,13312,19893,19968,40908,40960,42124,42192,42237,42240,42508,42512,42539,42560,42607,42612,42621,42623,42647,42655,42737,42775,42783,42786,42888,42891,42894,42896,42899,42912,42922,43e3,43047,43072,43123,43136,43204,43216,43225,43232,43255,43259,43259,43264,43309,43312,43347,43360,43388,43392,43456,43471,43481,43520,43574,43584,43597,43600,43609,43616,43638,43642,43643,43648,43714,43739,43741,43744,43759,43762,43766,43777,43782,43785,43790,43793,43798,43808,43814,43816,43822,43968,44010,44012,44013,44016,44025,44032,55203,55216,55238,55243,55291,63744,64109,64112,64217,64256,64262,64275,64279,64285,64296,64298,64310,64312,64316,64318,64318,64320,64321,64323,64324,64326,64433,64467,64829,64848,64911,64914,64967,65008,65019,65024,65039,65056,65062,65075,65076,65101,65103,65136,65140,65142,65276,65296,65305,65313,65338,65343,65343,65345,65370,65382,65470,65474,65479,65482,65487,65490,65495,65498,65500];function p(e,n){if(e<n[0])return!1;var t,r=0,a=n.length;while(r+1<a){if(t=r+(a-r)/2,t-=t%2,n[t]<=e&&e<=n[t+1])return!0;e<n[t]?a=t:r=t+2}return!1}function m(e,n){return p(e,n>=1?u:c)}function f(e,n){return p(e,n>=1?d:l)}function _(e){var n=[];return e.forEach(function(e,t){n[e]=t}),n}e.isUnicodeIdentifierStart=m;var g=_(s);function v(e){return g[e]}function y(e){return s.get(e)}function h(e){var n=new Array,t=0,r=0;while(t<e.length){var a=e.charCodeAt(t);switch(t++,a){case 13:10===e.charCodeAt(t)&&t++;case 10:n.push(r),r=t;break;default:a>127&&D(a)&&(n.push(r),r=t);break}}return n.push(r),n}function b(e,n,t,r){return e.getPositionOfLineAndCharacter?e.getPositionOfLineAndCharacter(n,t,r):E(T(e),n,t,e.text,r)}function E(n,t,r,a,i){(t<0||t>=n.length)&&(i?t=t<0?0:t>=n.length?n.length-1:t:e.Debug.fail("Bad line number. Line: "+t+", lineStarts.length: "+n.length+" , line map is correct? "+(void 0!==a?e.arraysEqual(n,h(a)):"unknown")));var o=n[t]+r;return i?o>n[t+1]?n[t+1]:"string"===typeof a&&o>a.length?a.length:o:(t<n.length-1?e.Debug.assert(o<n[t+1]):void 0!==a&&e.Debug.assert(o<=a.length),o)}function T(e){return e.lineMap||(e.lineMap=h(e.text))}function S(n,t){var r=e.binarySearch(n,t,e.identity,e.compareValues);return r<0&&(r=~r-1,e.Debug.assert(-1!==r,"position cannot precede the beginning of the file")),{line:r,character:t-n[r]}}function x(e,n){return S(T(e),n)}function C(e){return A(e)||D(e)}function A(e){return 32===e||9===e||11===e||12===e||160===e||133===e||5760===e||e>=8192&&e<=8203||8239===e||8287===e||12288===e||65279===e}function D(e){return 10===e||13===e||8232===e||8233===e}function L(e){return e>=48&&e<=57}function k(e){return e>=48&&e<=55}function N(e,n){var t=e.charCodeAt(n);switch(t){case 13:case 10:case 9:case 11:case 12:case 32:case 47:case 60:case 124:case 61:case 62:return!0;case 35:return 0===n;default:return t>127}}function I(n,t,r,a){if(void 0===a&&(a=!1),e.positionIsSynthesized(t))return t;while(1){var i=n.charCodeAt(t);switch(i){case 13:10===n.charCodeAt(t+1)&&t++;case 10:if(t++,r)return t;continue;case 9:case 11:case 12:case 32:t++;continue;case 47:if(a)break;if(47===n.charCodeAt(t+1)){t+=2;while(t<n.length){if(D(n.charCodeAt(t)))break;t++}continue}if(42===n.charCodeAt(t+1)){t+=2;while(t<n.length){if(42===n.charCodeAt(t)&&47===n.charCodeAt(t+1)){t+=2;break}t++}continue}break;case 60:case 124:case 61:case 62:if(M(n,t)){t=O(n,t);continue}break;case 35:if(0===t&&R(n,t)){t=F(n,t);continue}break;default:if(i>127&&C(i)){t++;continue}break}return t}}e.tokenToString=v,e.stringToToken=y,e.computeLineStarts=h,e.getPositionOfLineAndCharacter=b,e.computePositionOfLineAndCharacter=E,e.getLineStarts=T,e.computeLineAndCharacterOfPosition=S,e.getLineAndCharacterOfPosition=x,e.isWhiteSpaceLike=C,e.isWhiteSpaceSingleLine=A,e.isLineBreak=D,e.isOctalDigit=k,e.couldStartTrivia=N,e.skipTrivia=I;var w="<<<<<<<".length;function M(n,t){if(e.Debug.assert(t>=0),0===t||D(n.charCodeAt(t-1))){var r=n.charCodeAt(t);if(t+w<n.length){for(var a=0;a<w;a++)if(n.charCodeAt(t+a)!==r)return!1;return 61===r||32===n.charCodeAt(t+w)}}return!1}function O(n,t,r){r&&r(e.Diagnostics.Merge_conflict_marker_encountered,t,w);var a=n.charCodeAt(t),i=n.length;if(60===a||62===a)while(t<i&&!D(n.charCodeAt(t)))t++;else{e.Debug.assert(124===a||61===a);while(t<i){var o=n.charCodeAt(t);if((61===o||62===o)&&o!==a&&M(n,t))break;t++}}return t}var P=/^#!.*/;function R(n,t){return e.Debug.assert(0===t),P.test(n)}function F(e,n){var t=P.exec(e)[0];return n+=t.length,n}function G(e,n,t,r,a,i,o){var s,c,l,u,d=!1,p=r||0===t,m=o;e:while(t>=0&&t<n.length){var f=n.charCodeAt(t);switch(f){case 13:10===n.charCodeAt(t+1)&&t++;case 10:if(t++,r)break e;p=!0,d&&(u=!0);continue;case 9:case 11:case 12:case 32:t++;continue;case 47:var _=n.charCodeAt(t+1),g=!1;if(47===_||42===_){var v=47===_?2:3,y=t;if(t+=2,47===_)while(t<n.length){if(D(n.charCodeAt(t))){g=!0;break}t++}else while(t<n.length){if(42===n.charCodeAt(t)&&47===n.charCodeAt(t+1)){t+=2;break}t++}if(p){if(d&&(m=a(s,c,l,u,i,m),!e&&m))return m;s=y,c=t,l=v,u=g,d=!0}continue}break e;default:if(f>127&&C(f)){d&&D(f)&&(u=!0),t++;continue}break e}}return d&&(m=a(s,c,l,u,i,m)),m}function V(e,n,t,r){return G(!1,e,n,!1,t,r)}function B(e,n,t,r){return G(!1,e,n,!0,t,r)}function K(e,n,t,r,a){return G(!0,e,n,!1,t,r,a)}function j(e,n,t,r,a){return G(!0,e,n,!0,t,r,a)}function U(e,n,t,r,a,i){return i||(i=[]),i.push({kind:t,pos:e,end:n,hasTrailingNewLine:r}),i}function H(e,n){return K(e,n,U,void 0,void 0)}function W(e,n){return j(e,n,U,void 0,void 0)}function q(e){var n=P.exec(e);if(n)return n[0]}function z(e,n){return e>=65&&e<=90||e>=97&&e<=122||36===e||95===e||e>127&&m(e,n)}function J(e,n){return e>=65&&e<=90||e>=97&&e<=122||e>=48&&e<=57||36===e||95===e||e>127&&f(e,n)}function X(e,n){if(!z(e.charCodeAt(0),n))return!1;for(var t=1;t<e.length;t++)if(!J(e.charCodeAt(t),n))return!1;return!0}function Y(n,r,a,o,s,c,l){void 0===a&&(a=0);var u,d,p,m,f,_,g,v=o,y=0;return ue(v,c,l),{getStartPos:function(){return p},getTextPos:function(){return u},getToken:function(){return f},getTokenPos:function(){return m},getTokenText:function(){return v.substring(m,u)},getTokenValue:function(){return _},hasExtendedUnicodeEscape:function(){return 0!==(8&g)},hasPrecedingLineBreak:function(){return 0!==(1&g)},isIdentifier:function(){return 72===f||f>108},isReservedWord:function(){return f>=73&&f<=108},isUnterminated:function(){return 0!==(4&g)},getTokenFlags:function(){return g},reScanGreaterToken:Y,reScanSlashToken:Q,reScanTemplateToken:Z,scanJsxIdentifier:te,scanJsxAttributeValue:re,reScanJsxToken:$,reScanLessThanToken:ee,scanJsxToken:ne,scanJSDocToken:ae,scan:X,getText:le,setText:ue,setScriptTarget:pe,setLanguageVariant:me,setOnError:de,setTextPos:fe,setInJSDocType:_e,tryScan:ce,lookAhead:se,scanRange:oe};function h(e,n,t){if(void 0===n&&(n=u),s){var r=u;u=n,s(e,t||0),u=r}}function b(){var n=u,t=!1,r=!1,a="";while(1){var i=v.charCodeAt(u);if(95!==i){if(!L(i))break;t=!0,r=!1,u++}else g|=512,t?(t=!1,r=!0,a+=v.substring(n,u)):h(r?e.Diagnostics.Multiple_consecutive_numeric_separators_are_not_permitted:e.Diagnostics.Numeric_separators_are_not_allowed_here,u,1),u++,n=u}return 95===v.charCodeAt(u-1)&&h(e.Diagnostics.Numeric_separators_are_not_allowed_here,u-1,1),a+v.substring(n,u)}function E(){var n,t,r=u,a=b();46===v.charCodeAt(u)&&(u++,n=b());var i,o=u;if(69===v.charCodeAt(u)||101===v.charCodeAt(u)){u++,g|=16,43!==v.charCodeAt(u)&&45!==v.charCodeAt(u)||u++;var s=u,c=b();c?(t=v.substring(o,s)+c,o=u):h(e.Diagnostics.Digit_expected)}if(512&g?(i=a,n&&(i+="."+n),t&&(i+=t)):i=v.substring(r,o),void 0!==n||16&g)return T(r,void 0===n&&!!(16&g)),{type:8,value:""+ +i};_=i;var l=q();return T(r),{type:l,value:_}}function T(t,r){if(z(v.charCodeAt(u),n)){var a=u,i=U().length;1===i&&"n"===v[a]?h(r?e.Diagnostics.A_bigint_literal_cannot_use_exponential_notation:e.Diagnostics.A_bigint_literal_must_be_an_integer,t,a-t+1):(h(e.Diagnostics.An_identifier_or_keyword_cannot_immediately_follow_a_numeric_literal,a,i),u=a)}}function S(){var e=u;while(k(v.charCodeAt(u)))u++;return+v.substring(e,u)}function x(e,n){var t=I(e,!1,n);return t?parseInt(t,16):-1}function N(e,n){return I(e,!0,n)}function I(n,t,r){var a=[],i=!1,o=!1;while(a.length<n||t){var s=v.charCodeAt(u);if(r&&95===s)g|=512,i?(i=!1,o=!0):h(o?e.Diagnostics.Multiple_consecutive_numeric_separators_are_not_permitted:e.Diagnostics.Numeric_separators_are_not_allowed_here,u,1),u++;else{if(i=r,s>=65&&s<=70)s+=32;else if(!(s>=48&&s<=57||s>=97&&s<=102))break;a.push(s),u++,o=!1}}return a.length<n&&(a=[]),95===v.charCodeAt(u-1)&&h(e.Diagnostics.Numeric_separators_are_not_allowed_here,u-1,1),String.fromCharCode.apply(String,a)}function w(n){void 0===n&&(n=!1);var t=v.charCodeAt(u);u++;var r="",a=u;while(1){if(u>=d){r+=v.substring(a,u),g|=4,h(e.Diagnostics.Unterminated_string_literal);break}var i=v.charCodeAt(u);if(i===t){r+=v.substring(a,u),u++;break}if(92!==i||n){if(D(i)&&!n){r+=v.substring(a,u),g|=4,h(e.Diagnostics.Unterminated_string_literal);break}u++}else r+=v.substring(a,u),r+=G(),a=u}return r}function P(){var n=96===v.charCodeAt(u);u++;var t,r=u,a="";while(1){if(u>=d){a+=v.substring(r,u),g|=4,h(e.Diagnostics.Unterminated_template_literal),t=n?14:17;break}var i=v.charCodeAt(u);if(96===i){a+=v.substring(r,u),u++,t=n?14:17;break}if(36===i&&u+1<d&&123===v.charCodeAt(u+1)){a+=v.substring(r,u),u+=2,t=n?15:16;break}92!==i?13!==i?u++:(a+=v.substring(r,u),u++,u<d&&10===v.charCodeAt(u)&&u++,a+="\n",r=u):(a+=v.substring(r,u),a+=G(),r=u)}return e.Debug.assert(void 0!==t),_=a,t}function G(){if(u++,u>=d)return h(e.Diagnostics.Unexpected_end_of_text),"";var n=v.charCodeAt(u);switch(u++,n){case 48:return"\0";case 98:return"\b";case 116:return"\t";case 110:return"\n";case 118:return"\v";case 102:return"\f";case 114:return"\r";case 39:return"'";case 34:return'"';case 117:return u<d&&123===v.charCodeAt(u)?(g|=8,u++,B()):V(4);case 120:return V(2);case 13:u<d&&10===v.charCodeAt(u)&&u++;case 10:case 8232:case 8233:return"";default:return String.fromCharCode(n)}}function V(n){var t=x(n,!1);return t>=0?String.fromCharCode(t):(h(e.Diagnostics.Hexadecimal_digit_expected),"")}function B(){var n=N(1,!1),t=n?parseInt(n,16):-1,r=!1;return t<0?(h(e.Diagnostics.Hexadecimal_digit_expected),r=!0):t>1114111&&(h(e.Diagnostics.An_extended_Unicode_escape_value_must_be_between_0x0_and_0x10FFFF_inclusive),r=!0),u>=d?(h(e.Diagnostics.Unexpected_end_of_text),r=!0):125===v.charCodeAt(u)?u++:(h(e.Diagnostics.Unterminated_Unicode_escape_sequence),r=!0),r?"":K(t)}function K(n){if(e.Debug.assert(0<=n&&n<=1114111),n<=65535)return String.fromCharCode(n);var t=Math.floor((n-65536)/1024)+55296,r=(n-65536)%1024+56320;return String.fromCharCode(t,r)}function j(){if(u+5<d&&117===v.charCodeAt(u+1)){var e=u;u+=2;var n=x(4,!1);return u=e,n}return-1}function U(){var e="",t=u;while(u<d){var r=v.charCodeAt(u);if(J(r,n))u++;else{if(92!==r)break;if(r=j(),!(r>=0&&J(r,n)))break;e+=v.substring(t,u),e+=String.fromCharCode(r),u+=6,t=u}}return e+=v.substring(t,u),e}function H(){var e=_.length;if(e>=2&&e<=11){var n=_.charCodeAt(0);if(n>=97&&n<=122){var t=i.get(_);if(void 0!==t)return f=t}}return f=72}function W(n){var t="",r=!1,a=!1;while(1){var i=v.charCodeAt(u);if(95!==i){if(r=!0,!L(i)||i-48>=n)break;t+=v[u],u++,a=!1}else g|=512,r?(r=!1,a=!0):h(a?e.Diagnostics.Multiple_consecutive_numeric_separators_are_not_permitted:e.Diagnostics.Numeric_separators_are_not_allowed_here,u,1),u++}return 95===v.charCodeAt(u-1)&&h(e.Diagnostics.Numeric_separators_are_not_allowed_here,u-1,1),t}function q(){if(110===v.charCodeAt(u))return _+="n",384&g&&(_=e.parsePseudoBigInt(_)+"n"),u++,9;var n=128&g?parseInt(_.slice(2),2):256&g?parseInt(_.slice(2),8):+_;return _=""+n,8}function X(){var t;p=u,g=0;var i=!1;while(1){if(m=u,u>=d)return f=1;var o=v.charCodeAt(u);if(35===o&&0===u&&R(v,u)){if(u=F(v,u),r)continue;return f=6}switch(o){case 10:case 13:if(g|=1,r){u++;continue}return 13===o&&u+1<d&&10===v.charCodeAt(u+1)?u+=2:u++,f=4;case 9:case 11:case 12:case 32:case 160:case 5760:case 8192:case 8193:case 8194:case 8195:case 8196:case 8197:case 8198:case 8199:case 8200:case 8201:case 8202:case 8203:case 8239:case 8287:case 12288:case 65279:if(r){u++;continue}while(u<d&&A(v.charCodeAt(u)))u++;return f=5;case 33:return 61===v.charCodeAt(u+1)?61===v.charCodeAt(u+2)?(u+=3,f=36):(u+=2,f=34):(u++,f=52);case 34:case 39:return _=w(),f=10;case 96:return f=P();case 37:return 61===v.charCodeAt(u+1)?(u+=2,f=65):(u++,f=43);case 38:return 38===v.charCodeAt(u+1)?(u+=2,f=54):61===v.charCodeAt(u+1)?(u+=2,f=69):(u++,f=49);case 40:return u++,f=20;case 41:return u++,f=21;case 42:if(61===v.charCodeAt(u+1))return u+=2,f=62;if(42===v.charCodeAt(u+1))return 61===v.charCodeAt(u+2)?(u+=3,f=63):(u+=2,f=41);if(u++,y&&!i&&1&g){i=!0;continue}return f=40;case 43:return 43===v.charCodeAt(u+1)?(u+=2,f=44):61===v.charCodeAt(u+1)?(u+=2,f=60):(u++,f=38);case 44:return u++,f=27;case 45:return 45===v.charCodeAt(u+1)?(u+=2,f=45):61===v.charCodeAt(u+1)?(u+=2,f=61):(u++,f=39);case 46:return L(v.charCodeAt(u+1))?(_=E().value,f=8):46===v.charCodeAt(u+1)&&46===v.charCodeAt(u+2)?(u+=3,f=25):(u++,f=24);case 47:if(47===v.charCodeAt(u+1)){u+=2;while(u<d){if(D(v.charCodeAt(u)))break;u++}if(r)continue;return f=2}if(42===v.charCodeAt(u+1)){u+=2,42===v.charCodeAt(u)&&47!==v.charCodeAt(u+1)&&(g|=2);var s=!1;while(u<d){var c=v.charCodeAt(u);if(42===c&&47===v.charCodeAt(u+1)){u+=2,s=!0;break}D(c)&&(g|=1),u++}if(s||h(e.Diagnostics.Asterisk_Slash_expected),r)continue;return s||(g|=4),f=3}return 61===v.charCodeAt(u+1)?(u+=2,f=64):(u++,f=42);case 48:if(u+2<d&&(88===v.charCodeAt(u+1)||120===v.charCodeAt(u+1)))return u+=2,_=N(1,!0),_||(h(e.Diagnostics.Hexadecimal_digit_expected),_="0"),_="0x"+_,g|=64,f=q();if(u+2<d&&(66===v.charCodeAt(u+1)||98===v.charCodeAt(u+1)))return u+=2,_=W(2),_||(h(e.Diagnostics.Binary_digit_expected),_="0"),_="0b"+_,g|=128,f=q();if(u+2<d&&(79===v.charCodeAt(u+1)||111===v.charCodeAt(u+1)))return u+=2,_=W(8),_||(h(e.Diagnostics.Octal_digit_expected),_="0"),_="0o"+_,g|=256,f=q();if(u+1<d&&k(v.charCodeAt(u+1)))return _=""+S(),g|=32,f=8;case 49:case 50:case 51:case 52:case 53:case 54:case 55:case 56:case 57:return t=E(),f=t.type,_=t.value,f;case 58:return u++,f=57;case 59:return u++,f=26;case 60:if(M(v,u)){if(u=O(v,u,h),r)continue;return f=7}return 60===v.charCodeAt(u+1)?61===v.charCodeAt(u+2)?(u+=3,f=66):(u+=2,f=46):61===v.charCodeAt(u+1)?(u+=2,f=31):1===a&&47===v.charCodeAt(u+1)&&42!==v.charCodeAt(u+2)?(u+=2,f=29):(u++,f=28);case 61:if(M(v,u)){if(u=O(v,u,h),r)continue;return f=7}return 61===v.charCodeAt(u+1)?61===v.charCodeAt(u+2)?(u+=3,f=35):(u+=2,f=33):62===v.charCodeAt(u+1)?(u+=2,f=37):(u++,f=59);case 62:if(M(v,u)){if(u=O(v,u,h),r)continue;return f=7}return u++,f=30;case 63:return u++,f=56;case 91:return u++,f=22;case 93:return u++,f=23;case 94:return 61===v.charCodeAt(u+1)?(u+=2,f=71):(u++,f=51);case 123:return u++,f=18;case 124:if(M(v,u)){if(u=O(v,u,h),r)continue;return f=7}return 124===v.charCodeAt(u+1)?(u+=2,f=55):61===v.charCodeAt(u+1)?(u+=2,f=70):(u++,f=50);case 125:return u++,f=19;case 126:return u++,f=53;case 64:return u++,f=58;case 92:var l=j();return l>=0&&z(l,n)?(u+=6,_=String.fromCharCode(l)+U(),f=H()):(h(e.Diagnostics.Invalid_character),u++,f=0);default:if(z(o,n)){u++;while(u<d&&J(o=v.charCodeAt(u),n))u++;return _=v.substring(m,u),92===o&&(_+=U()),f=H()}if(A(o)){u++;continue}if(D(o)){g|=1,u++;continue}return h(e.Diagnostics.Invalid_character),u++,f=0}}}function Y(){if(30===f){if(62===v.charCodeAt(u))return 62===v.charCodeAt(u+1)?61===v.charCodeAt(u+2)?(u+=3,f=68):(u+=2,f=48):61===v.charCodeAt(u+1)?(u+=2,f=67):(u++,f=47);if(61===v.charCodeAt(u))return u++,f=32}return f}function Q(){if(42===f||64===f){var t=m+1,r=!1,a=!1;while(1){if(t>=d){g|=4,h(e.Diagnostics.Unterminated_regular_expression_literal);break}var i=v.charCodeAt(t);if(D(i)){g|=4,h(e.Diagnostics.Unterminated_regular_expression_literal);break}if(r)r=!1;else{if(47===i&&!a){t++;break}91===i?a=!0:92===i?r=!0:93===i&&(a=!1)}t++}while(t<d&&J(v.charCodeAt(t),n))t++;u=t,_=v.substring(m,u),f=13}return f}function Z(){return e.Debug.assert(19===f,"'reScanTemplateToken' should only be called on a '}'"),u=m,f=P()}function $(){return u=m=p,f=ne()}function ee(){return 46===f?(u=m+1,f=28):f}function ne(){if(p=m=u,u>=d)return f=1;var e=v.charCodeAt(u);if(60===e)return 47===v.charCodeAt(u+1)?(u+=2,f=29):(u++,f=28);if(123===e)return u++,f=18;var n=0;while(u<d){if(e=v.charCodeAt(u),123===e)break;if(60===e){if(M(v,u))return u=O(v,u,h),f=7;break}D(e)&&0===n?n=-1:C(e)||(n=u),u++}return-1===n?12:11}function te(){if(t(f)){var e=u;while(u<d){var r=v.charCodeAt(u);if(45!==r&&(e===u?!z(r,n):!J(r,n)))break;u++}_+=v.substring(e,u)}return f}function re(){switch(p=u,v.charCodeAt(u)){case 34:case 39:return _=w(!0),f=10;default:return X()}}function ae(){if(p=m=u,g=0,u>=d)return f=1;var e=v.charCodeAt(u);switch(u++,e){case 9:case 11:case 12:case 32:while(u<d&&A(v.charCodeAt(u)))u++;return f=5;case 64:return f=58;case 10:case 13:return g|=1,f=4;case 42:return f=40;case 123:return f=18;case 125:return f=19;case 91:return f=22;case 93:return f=23;case 60:return f=28;case 61:return f=59;case 44:return f=27;case 46:return f=24;case 96:while(u<d&&96!==v.charCodeAt(u))u++;return _=v.substring(m+1,u),u++,f=14}if(z(e,6)){while(J(v.charCodeAt(u),6)&&u<d)u++;return _=v.substring(m,u),f=H()}return f=0}function ie(e,n){var t=u,r=p,a=m,i=f,o=_,s=g,c=e();return c&&!n||(u=t,p=r,m=a,f=i,_=o,g=s),c}function oe(e,n,t){var r=d,a=u,i=p,o=m,s=f,c=_,l=g;ue(v,e,n);var y=t();return d=r,u=a,p=i,m=o,f=s,_=c,g=l,y}function se(e){return ie(e,!0)}function ce(e){return ie(e,!1)}function le(){return v}function ue(e,n,t){v=e||"",d=void 0===t?v.length:n+t,fe(n||0)}function de(e){s=e}function pe(e){n=e}function me(e){a=e}function fe(n){e.Debug.assert(n>=0),u=n,p=n,m=n,f=0,_=void 0,g=0}function _e(e){y+=e?1:-1}}e.forEachLeadingCommentRange=V,e.forEachTrailingCommentRange=B,e.reduceEachLeadingCommentRange=K,e.reduceEachTrailingCommentRange=j,e.getLeadingCommentRanges=H,e.getTrailingCommentRanges=W,e.getShebang=q,e.isIdentifierStart=z,e.isIdentifierPart=J,e.isIdentifierText=X,e.createScanner=Y}(u||(u={})),function(e){function n(n){return e.pathIsRelative(n)||e.isRootedDiskPath(n)}function t(n){return e.sortAndDeduplicate(n,e.compareDiagnostics)}e.isExternalModuleNameRelative=n,e.sortAndDeduplicateDiagnostics=t}(u||(u={})),function(e){function n(e,n){var t=e.declarations;if(t)for(var r=0,a=t;r<a.length;r++){var i=a[r];if(i.kind===n)return i}}function t(){return new e.MapCtr}function r(e){return!!e&&!!e.size}function a(n){var t=e.createMap();if(n)for(var r=0,a=n;r<a.length;r++){var i=a[r];t.set(i.escapedName,i)}return t}e.resolvingEmptyArray=[],e.emptyMap=e.createMap(),e.emptyUnderscoreEscapedMap=e.emptyMap,e.externalHelpersModuleNameText="tslib",e.defaultMaximumTruncationLength=160,e.getDeclarationOfKind=n,e.createUnderscoreEscapedMap=t,e.hasEntries=r,e.createSymbolTable=a;var i=s();function s(){var n="",t=function(e){return n+=e};return{getText:function(){return n},write:t,rawWrite:t,writeKeyword:t,writeOperator:t,writePunctuation:t,writeSpace:t,writeStringLiteral:t,writeLiteral:t,writeParameter:t,writeProperty:t,writeSymbol:function(e,n){return t(e)},writeTrailingSemicolon:t,writeComment:t,getTextPos:function(){return n.length},getLine:function(){return 0},getColumn:function(){return 0},getIndent:function(){return 0},isAtStartOfLine:function(){return!1},writeLine:function(){return n+=" "},increaseIndent:e.noop,decreaseIndent:e.noop,clear:function(){return n=""},trackSymbol:e.noop,reportInaccessibleThisError:e.noop,reportInaccessibleUniqueSymbolError:e.noop,reportPrivateInBaseOfClassExpression:e.noop}}function c(n,t,r){var a=e.isRootedDiskPath(n)?e.normalizePath(n):e.getNormalizedAbsolutePath(n,t);return r(a)}function l(n,t){return n.configFilePath!==t.configFilePath||e.moduleResolutionOptionDeclarations.some(function(r){return!e.isJsonEqual(e.getCompilerOptionValue(n,r),e.getCompilerOptionValue(t,r))})}function u(e,n){while(e){var t=n(e);if("quit"===t)return;if(t)return e;e=e.parent}}function d(n,t){while(1){var r=t(n);if("quit"===r)return;if(void 0!==r)return r;if(e.isSourceFile(n))return;n=n.parent}}function p(e,n){for(var t,r=e.entries(),a=r.next(),i=a.value,o=a.done;!o;t=r.next(),i=t.value,o=t.done,t){var s=i[0],c=i[1],l=n(c,s);if(l)return l}}function m(e,n){for(var t,r=e.keys(),a=r.next(),i=a.value,o=a.done;!o;t=r.next(),i=t.value,o=t.done,t){var s=n(i);if(s)return s}}function f(e,n){e.forEach(function(e,t){n.set(t,e)})}function _(n,t){return e.arrayToMap(n,t||function(e){return e},function(){return!0})}function g(n){var t=e.createMap();return f(n,t),t}function v(e){var n=i.getText();try{return e(i),i.getText()}finally{i.clear(),i.writeKeyword(n)}}function y(e){return e.end-e.pos}function h(e,n){return e&&e.resolvedModules&&e.resolvedModules.get(n)}function b(n,t,r){n.resolvedModules||(n.resolvedModules=e.createMap()),n.resolvedModules.set(t,r)}function E(n,t,r){n.resolvedTypeReferenceDirectiveNames||(n.resolvedTypeReferenceDirectiveNames=e.createMap()),n.resolvedTypeReferenceDirectiveNames.set(t,r)}function T(e,n){return e.path===n.path&&!e.prepend===!n.prepend&&!e.circular===!n.circular}function S(e,n){return e.isExternalLibraryImport===n.isExternalLibraryImport&&e.extension===n.extension&&e.resolvedFileName===n.resolvedFileName&&e.originalPath===n.originalPath&&x(e.packageId,n.packageId)}function x(e,n){return e===n||!!e&&!!n&&e.name===n.name&&e.subModuleName===n.subModuleName&&e.version===n.version}function C(e){var n=e.name,t=e.subModuleName,r=e.version,a=t?n+"/"+t:n;return a+"@"+r}function A(e,n){return e.resolvedFileName===n.resolvedFileName&&e.primary===n.primary}function D(n,t,r,a){e.Debug.assert(n.length===t.length);for(var i=0;i<n.length;i++){var o=t[i],s=r&&r.get(n[i]),c=s?!o||!a(s,o):o;if(c)return!0}return!1}function L(e){return k(e),0!==(131072&e.flags)}function k(n){if(!(262144&n.flags)){var t=0!==(32768&n.flags)||e.forEachChild(n,L);t&&(n.flags|=131072),n.flags|=262144}}function N(e){while(e&&279!==e.kind)e=e.parent;return e}function I(e){switch(e.kind){case 218:case 246:case 225:case 226:case 227:return!0}return!1}function w(n,t){return e.Debug.assert(n>=0),e.getLineStarts(t)[n]}function M(n){var t=N(n),r=e.getLineAndCharacterOfPosition(t,n.pos);return t.fileName+"("+(r.line+1)+","+(r.character+1)+")"}function O(n,t){e.Debug.assert(n>=0);var r=e.getLineStarts(t),a=n,i=t.text;if(a+1===r.length)return i.length-1;var o=r[a],s=r[a+1]-1;e.Debug.assert(e.isLineBreak(i.charCodeAt(s)));while(o<=s&&e.isLineBreak(i.charCodeAt(s)))s--;return s}function P(e,n,t){return!(t&&t(n))&&!e.identifiers.has(n)}function R(e){return void 0===e||e.pos===e.end&&e.pos>=0&&1!==e.kind}function F(e){return!R(e)}function G(e,n){if(void 0===n||0===n.length)return e;for(var t=0;t<e.length;++t)if(!Be(e[t]))break;return e.splice.apply(e,[t,0].concat(n)),e}function V(n,t,r){if(47===n.charCodeAt(t+1)&&t+2<r&&47===n.charCodeAt(t+2)){var a=n.substring(t,r);return!!(a.match(e.fullTripleSlashReferencePathRegEx)||a.match(e.fullTripleSlashAMDReferencePathRegEx)||a.match(Ue)||a.match(He))}return!1}function B(e,n){return 42===e.charCodeAt(n+1)&&33===e.charCodeAt(n+2)}function K(n,t,r){return R(n)?n.pos:e.isJSDocNode(n)?e.skipTrivia((t||N(n)).text,n.pos,!1,!0):r&&e.hasJSDocNodes(n)?K(n.jsDoc[0]):306===n.kind&&n._children.length>0?K(n._children[0],t,r):e.skipTrivia((t||N(n)).text,n.pos)}function j(n,t){return R(n)||!n.decorators?K(n,t):e.skipTrivia((t||N(n)).text,n.decorators.end)}function U(e,n,t){return void 0===t&&(t=!1),W(e.text,n,t)}function H(e){return 283===e.kind||e.parent&&H(e.parent)}function W(n,t,r){if(void 0===r&&(r=!1),R(t))return"";var a=n.substring(r?t.pos:e.skipTrivia(n,t.pos),t.end);return H(t)&&(a=a.replace(/(^|\r?\n|\r)\s*\*\s*/g,"$1")),a}function q(e,n){return void 0===n&&(n=!1),U(N(e),e,n)}function z(e){return e.pos}function J(n,t){return e.binarySearch(n,t,z,e.compareValues)}function X(e){var n=e.emitNode;return n&&n.flags||0}function Y(n,t,r){if(!Dr(n)&&n.parent&&!(e.isNumericLiteral(n)&&512&n.numericLiteralFlags||e.isBigIntLiteral(n)))return U(t,n);var a=r||16777216&X(n)?Br:Wr;switch(n.kind){case 10:return n.singleQuote?"'"+a(n.text,39)+"'":'"'+a(n.text,34)+'"';case 14:return"`"+a(n.text,96)+"`";case 15:return"`"+a(n.text,96)+"${";case 16:return"}"+a(n.text,96)+"${";case 17:return"}"+a(n.text,96)+"`";case 8:case 9:case 13:return n.text}return e.Debug.fail("Literal kind '"+n.kind+"' not accounted for.")}function Q(n){return e.isString(n)?'"'+Wr(n)+'"':""+n}function Z(n){return e.getBaseFileName(n).replace(/^(\d)/,"_$1").replace(/\W/g,"_")}function $(n){return 0!==(3&e.getCombinedNodeFlags(n))||ee(n)}function ee(e){var n=Cr(e);return 237===n.kind&&274===n.parent.kind}function ne(n){return e.isModuleDeclaration(n)&&(10===n.name.kind||ce(n))}function te(n){return e.isModuleDeclaration(n)&&10===n.name.kind}function re(n){return e.isModuleDeclaration(n)&&e.isStringLiteral(n.name)}function ae(n){return e.isModuleDeclaration(n)||e.isIdentifier(n)}function ie(e){return oe(e.valueDeclaration)}function oe(e){return e&&244===e.kind&&!e.body}function se(n){return 279===n.kind||244===n.kind||e.isFunctionLike(n)}function ce(e){return!!(512&e.flags)}function le(e){return ne(e)&&ue(e)}function ue(n){switch(n.parent.kind){case 279:return e.isExternalModule(n.parent);case 245:return ne(n.parent.parent)&&e.isSourceFile(n.parent.parent.parent)&&!e.isExternalModule(n.parent.parent.parent)}return!1}function de(n){return e.find(n.declarations,function(n){return!le(n)&&!(e.isModuleDeclaration(n)&&ce(n))})}function pe(n,t){return e.isExternalModule(n)||t.isolatedModules||e.getEmitModuleKind(t)===e.ModuleKind.CommonJS&&!!n.commonJsModuleIndicator}function me(n,t){switch(n.kind){case 279:case 246:case 274:case 244:case 225:case 226:case 227:case 157:case 156:case 158:case 159:case 239:case 196:case 197:return!0;case 218:return!e.isFunctionLike(t)}return!1}function fe(n){switch(n.kind){case 297:case 304:case 293:return!0;default:return e.assertType(n),_e(n)}}function _e(n){switch(n.kind){case 160:case 161:case 155:case 162:case 165:case 166:case 289:case 240:case 209:case 241:case 242:case 303:case 239:case 156:case 157:case 158:case 159:case 196:case 197:return!0;default:return e.assertType(n),!1}}function ge(e){switch(e.kind){case 249:case 248:return!0;default:return!1}}function ve(e){switch(e.kind){case 249:case 248:case 219:case 240:case 239:case 244:case 242:case 241:case 243:return!0;default:return!1}}function ye(n){return ge(n)||e.isExportDeclaration(n)}function he(e){return u(e.parent,function(e){return me(e,e.parent)})}function be(e){return e&&0!==y(e)?q(e):"(Missing)"}function Ee(e){return e.declaration?be(e.declaration.parameters[0].name):void 0}function Te(n){switch(n.kind){case 72:return n.escapedText;case 10:case 8:case 14:return e.escapeLeadingUnderscores(n.text);case 149:return pr(n.expression)?e.escapeLeadingUnderscores(n.expression.text):e.Debug.fail("Text of property name cannot be read from non-literal-valued ComputedPropertyNames");default:return e.Debug.assertNever(n)}}function Se(n){switch(n.kind){case 72:return 0===y(n)?e.idText(n):q(n);case 148:return Se(n.left)+"."+Se(n.right);case 189:return Se(n.expression)+"."+Se(n.name);default:throw e.Debug.assertNever(n)}}function xe(e,n,t,r,a,i){var o=N(e);return Ae(o,e,n,t,r,a,i)}function Ce(n,t,r,a,i,o,s){var c=e.skipTrivia(n.text,t.pos);return e.createFileDiagnostic(n,c,t.end-c,r,a,i,o,s)}function Ae(n,t,r,a,i,o,s){var c=Ne(n,t);return e.createFileDiagnostic(n,c.start,c.length,r,a,i,o,s)}function De(e,n,t){var r=N(e),a=Ne(r,e);return{file:r,start:a.start,length:a.length,code:n.code,category:n.category,messageText:n.next?n:n.messageText,relatedInformation:t}}function Le(n,t){var r=e.createScanner(n.languageVersion,!0,n.languageVariant,n.text,void 0,t);r.scan();var a=r.getTokenPos();return e.createTextSpanFromBounds(a,r.getTextPos())}function ke(n,t){var r=e.skipTrivia(n.text,t.pos);if(t.body&&218===t.body.kind){var a=e.getLineAndCharacterOfPosition(n,t.body.pos).line,i=e.getLineAndCharacterOfPosition(n,t.body.end).line;if(a<i)return e.createTextSpan(r,O(a,n)-r+1)}return e.createTextSpanFromBounds(r,t.end)}function Ne(n,t){var r=t;switch(t.kind){case 279:var a=e.skipTrivia(n.text,0,!1);return a===n.text.length?e.createTextSpan(0,0):Le(n,a);case 237:case 186:case 240:case 209:case 241:case 244:case 243:case 278:case 239:case 196:case 156:case 158:case 159:case 242:case 154:case 153:r=t.name;break;case 197:return ke(n,t)}if(void 0===r)return Le(n,t.pos);var i=R(r),o=i||e.isJsxText(t)?r.pos:e.skipTrivia(n.text,r.pos);return i?(e.Debug.assert(o===r.pos,"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809"),e.Debug.assert(o===r.end,"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809")):(e.Debug.assert(o>=r.pos,"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809"),e.Debug.assert(o<=r.end,"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809")),e.createTextSpanFromBounds(o,r.end)}function Ie(e){return void 0!==(e.externalModuleIndicator||e.commonJsModuleIndicator)}function we(e){return 6===e.scriptKind}function Me(n){return!!(2048&e.getCombinedModifierFlags(n))}function Oe(n){return!(!(64&e.getCombinedModifierFlags(n))||e.isParameterPropertyDeclaration(n))}function Pe(n){return!!(2&e.getCombinedNodeFlags(n))}function Re(n){return!!(1&e.getCombinedNodeFlags(n))}function Fe(e){return 191===e.kind&&98===e.expression.kind}function Ge(e){return 191===e.kind&&92===e.expression.kind}function Ve(n){return e.isImportTypeNode(n)&&e.isLiteralTypeNode(n.argument)&&e.isStringLiteral(n.argument.literal)}function Be(e){return 221===e.kind&&10===e.expression.kind}function Ke(n,t){return 11!==n.kind?e.getLeadingCommentRanges(t.text,n.pos):void 0}function je(n,t){var r=151===n.kind||150===n.kind||196===n.kind||197===n.kind||195===n.kind?e.concatenate(e.getTrailingCommentRanges(t,n.pos),e.getLeadingCommentRanges(t,n.pos)):e.getLeadingCommentRanges(t,n.pos);return e.filter(r,function(e){return 42===t.charCodeAt(e.pos+1)&&42===t.charCodeAt(e.pos+2)&&47!==t.charCodeAt(e.pos+3)})}e.toPath=c,e.changesAffectModuleResolution=l,e.findAncestor=u,e.forEachAncestor=d,e.forEachEntry=p,e.forEachKey=m,e.copyEntries=f,e.arrayToSet=_,e.cloneMap=g,e.usingSingleLineStringWriter=v,e.getFullWidth=y,e.getResolvedModule=h,e.setResolvedModule=b,e.setResolvedTypeReferenceDirective=E,e.projectReferenceIsEqualTo=T,e.moduleResolutionIsEqualTo=S,e.packageIdToString=C,e.typeDirectiveIsEqualTo=A,e.hasChangesInResolutions=D,e.containsParseError=L,e.getSourceFileOfNode=N,e.isStatementWithLocals=I,e.getStartPositionOfLine=w,e.nodePosToString=M,e.getEndLinePosition=O,e.isFileLevelUniqueName=P,e.nodeIsMissing=R,e.nodeIsPresent=F,e.addStatementsAfterPrologue=G,e.isRecognizedTripleSlashComment=V,e.isPinnedComment=B,e.getTokenPosOfNode=K,e.getNonDecoratorTokenPosOfNode=j,e.getSourceTextOfNodeFromSourceFile=U,e.getTextOfNodeFromSourceText=W,e.getTextOfNode=q,e.indexOfNode=J,e.getEmitFlags=X,e.getLiteralText=Y,e.getTextOfConstantValue=Q,e.makeIdentifierFromModuleName=Z,e.isBlockOrCatchScoped=$,e.isCatchClauseVariableDeclarationOrBindingElement=ee,e.isAmbientModule=ne,e.isModuleWithStringLiteralName=te,e.isNonGlobalAmbientModule=re,e.isEffectiveModuleDeclaration=ae,e.isShorthandAmbientModuleSymbol=ie,e.isBlockScopedContainerTopLevel=se,e.isGlobalScopeAugmentation=ce,e.isExternalModuleAugmentation=le,e.isModuleAugmentationExternal=ue,e.getNonAugmentationDeclaration=de,e.isEffectiveExternalModule=pe,e.isBlockScope=me,e.isDeclarationWithTypeParameters=fe,e.isDeclarationWithTypeParameterChildren=_e,e.isAnyImportSyntax=ge,e.isLateVisibilityPaintedStatement=ve,e.isAnyImportOrReExport=ye,e.getEnclosingBlockScopeContainer=he,e.declarationNameToString=be,e.getNameFromIndexInfo=Ee,e.getTextOfPropertyName=Te,e.entityNameToString=Se,e.createDiagnosticForNode=xe,e.createDiagnosticForNodeArray=Ce,e.createDiagnosticForNodeInSourceFile=Ae,e.createDiagnosticForNodeFromMessageChain=De,e.getSpanOfTokenAtPosition=Le,e.getErrorSpanForNode=Ne,e.isExternalOrCommonJsModule=Ie,e.isJsonSourceFile=we,e.isEnumConst=Me,e.isDeclarationReadonly=Oe,e.isVarConst=Pe,e.isLet=Re,e.isSuperCall=Fe,e.isImportCall=Ge,e.isLiteralImportTypeNode=Ve,e.isPrologueDirective=Be,e.getLeadingCommentRangesOfNode=Ke,e.getJSDocCommentRanges=je,e.fullTripleSlashReferencePathRegEx=/^(\/\/\/\s*<reference\s+path\s*=\s*)('|")(.+?)\2.*?\/>/;var Ue=/^(\/\/\/\s*<reference\s+types\s*=\s*)('|")(.+?)\2.*?\/>/;e.fullTripleSlashAMDReferencePathRegEx=/^(\/\/\/\s*<amd-dependency\s+path\s*=\s*)('|")(.+?)\2.*?\/>/;var He=/^(\/\/\/\s*<reference\s+no-default-lib\s*=\s*)('|")(.+?)\2\s*\/>/;function We(n){if(163<=n.kind&&n.kind<=183)return!0;switch(n.kind){case 120:case 143:case 135:case 146:case 138:case 123:case 139:case 136:case 141:case 132:return!0;case 106:return 200!==n.parent.kind;case 211:return!qa(n);case 150:return 181===n.parent.kind||176===n.parent.kind;case 72:148===n.parent.kind&&n.parent.right===n?n=n.parent:189===n.parent.kind&&n.parent.name===n&&(n=n.parent),e.Debug.assert(72===n.kind||148===n.kind||189===n.kind,"'node' was expected to be a qualified name, identifier or property access in 'isPartOfTypeNode'.");case 148:case 189:case 100:var t=n.parent;if(167===t.kind)return!1;if(183===t.kind)return!t.isTypeOf;if(163<=t.kind&&t.kind<=183)return!0;switch(t.kind){case 211:return!qa(t);case 150:return n===t.constraint;case 303:return n===t.constraint;case 154:case 153:case 151:case 237:return n===t.type;case 239:case 196:case 197:case 157:case 156:case 155:case 158:case 159:return n===t.type;case 160:case 161:case 162:return n===t.type;case 194:return n===t.type;case 191:case 192:return e.contains(t.typeArguments,n);case 193:return!1}}return!1}function qe(e,n){while(e){if(e.kind===n)return!0;e=e.parent}return!1}function ze(n,t){return r(n);function r(n){switch(n.kind){case 230:return t(n);case 246:case 218:case 222:case 223:case 224:case 225:case 226:case 227:case 231:case 232:case 271:case 272:case 233:case 235:case 274:return e.forEachChild(n,r)}}}function Je(n,t){return r(n);function r(n){switch(n.kind){case 207:t(n);var a=n.expression;return void(a&&r(a));case 243:case 241:case 244:case 242:case 240:case 209:return;default:if(e.isFunctionLike(n)){if(n.name&&149===n.name.kind)return void r(n.name.expression)}else We(n)||e.forEachChild(n,r)}}}function Xe(n){return n&&169===n.kind?n.elementType:n&&164===n.kind?e.singleOrUndefined(n.typeArguments):void 0}function Ye(e){switch(e.kind){case 241:case 240:case 209:case 168:return e.members;case 188:return e.properties}}function Qe(e){if(e)switch(e.kind){case 186:case 278:case 151:case 275:case 154:case 153:case 276:case 237:return!0}return!1}function Ze(n){return Qe(n)||e.isAccessor(n)}function $e(e){return 238===e.parent.kind&&219===e.parent.parent.kind}function en(n){return e.isVariableDeclaration(n)?Pe(n)&&e.isIdentifier(n.name)&&$e(n):e.isPropertyDeclaration(n)?Pa(n)&&Oa(n):e.isPropertySignature(n)&&Pa(n)}function nn(e){switch(e.kind){case 156:case 155:case 157:case 158:case 159:case 239:case 196:return!0}return!1}function tn(e,n){while(1){if(n&&n(e),233!==e.statement.kind)return e.statement;e=e.statement}}function rn(n){return n&&218===n.kind&&e.isFunctionLike(n.parent)}function an(e){return e&&156===e.kind&&188===e.parent.kind}function on(e){return 156===e.kind&&(188===e.parent.kind||209===e.parent.kind)}function sn(e){return e&&1===e.kind}function cn(e){return e&&0===e.kind}function ln(e,n,t){return e.properties.filter(function(e){if(275===e.kind){var r=Te(e.name);return n===r||!!t&&t===r}return!1})}function un(n){if(n&&n.statements.length){var t=n.statements[0].expression;return e.tryCast(t,e.isObjectLiteralExpression)}}function dn(n,t,r){return e.firstDefined(pn(n,t),function(n){return e.isArrayLiteralExpression(n.initializer)?e.find(n.initializer.elements,function(n){return e.isStringLiteral(n)&&n.text===r}):void 0})}function pn(n,t){var r=un(n);return r?ln(r,t):e.emptyArray}function mn(n){return u(n.parent,e.isFunctionLike)}function fn(n){return u(n.parent,e.isClassLike)}function _n(n,t){e.Debug.assert(279!==n.kind);while(1){if(n=n.parent,!n)return e.Debug.fail();switch(n.kind){case 149:if(e.isClassLike(n.parent.parent))return n;n=n.parent;break;case 152:151===n.parent.kind&&e.isClassElement(n.parent.parent)?n=n.parent.parent:e.isClassElement(n.parent)&&(n=n.parent);break;case 197:if(!t)continue;case 239:case 196:case 244:case 154:case 153:case 156:case 155:case 157:case 158:case 159:case 160:case 161:case 162:case 243:case 279:return n}}}function gn(e){var n=_n(e,!1);if(n)switch(n.kind){case 157:case 239:case 196:return n}}function vn(n,t){while(1){if(n=n.parent,!n)return n;switch(n.kind){case 149:n=n.parent;break;case 239:case 196:case 197:if(!t)continue;case 154:case 153:case 156:case 155:case 157:case 158:case 159:return n;case 152:151===n.parent.kind&&e.isClassElement(n.parent.parent)?n=n.parent.parent:e.isClassElement(n.parent)&&(n=n.parent);break}}}function yn(e){if(196===e.kind||197===e.kind){var n=e,t=e.parent;while(195===t.kind)n=t,t=t.parent;if(191===t.kind&&t.expression===n)return t}}function hn(e){var n=e.kind;return(189===n||190===n)&&98===e.expression.kind}function bn(e){var n=e.kind;return(189===n||190===n)&&100===e.expression.kind}function En(e){switch(e.kind){case 164:return e.typeName;case 211:return za(e.expression)?e.expression:void 0;case 72:case 148:return e}}function Tn(e){switch(e.kind){case 193:return e.tag;case 262:case 261:return e.tagName;default:return e.expression}}function Sn(e,n,t){switch(e.kind){case 240:return!0;case 154:return 240===n.kind;case 158:case 159:case 156:return void 0!==e.body&&240===n.kind;case 151:return void 0!==n.body&&(157===n.kind||156===n.kind||159===n.kind)&&240===t.kind}return!1}function xn(e,n,t){return void 0!==e.decorators&&Sn(e,n,t)}function Cn(e,n,t){return xn(e,n,t)||An(e,n)}function An(n,t){switch(n.kind){case 240:return e.some(n.members,function(e){return Cn(e,n,t)});case 156:case 159:return e.some(n.parameters,function(e){return xn(e,n,t)});default:return!1}}function Dn(e){var n=e.parent;return(262===n.kind||261===n.kind||263===n.kind)&&n.tagName===e}function Ln(e){switch(e.kind){case 98:case 96:case 102:case 87:case 13:case 187:case 188:case 189:case 190:case 191:case 192:case 193:case 212:case 194:case 213:case 195:case 196:case 209:case 197:case 200:case 198:case 199:case 202:case 203:case 204:case 205:case 208:case 206:case 14:case 210:case 260:case 261:case 264:case 207:case 201:case 214:return!0;case 148:while(148===e.parent.kind)e=e.parent;return 167===e.parent.kind||Dn(e);case 72:if(167===e.parent.kind||Dn(e))return!0;case 8:case 9:case 10:case 100:return kn(e);default:return!1}}function kn(e){var n=e.parent;switch(n.kind){case 237:case 151:case 154:case 153:case 278:case 275:case 186:return n.initializer===e;case 221:case 222:case 223:case 224:case 230:case 231:case 232:case 271:case 234:return n.expression===e;case 225:var t=n;return t.initializer===e&&238!==t.initializer.kind||t.condition===e||t.incrementor===e;case 226:case 227:var r=n;return r.initializer===e&&238!==r.initializer.kind||r.expression===e;case 194:case 212:return e===n.expression;case 216:return e===n.expression;case 149:return e===n.expression;case 152:case 270:case 269:case 277:return!0;case 211:return n.expression===e&&qa(n);case 276:return n.objectAssignmentInitializer===e;default:return Ln(n)}}function Nn(e){return 248===e.kind&&259===e.moduleReference.kind}function In(n){return e.Debug.assert(Nn(n)),n.moduleReference.expression}function wn(e){return 248===e.kind&&259!==e.moduleReference.kind}function Mn(e){return Pn(e)}function On(e){return!Pn(e)}function Pn(e){return!!e&&!!(65536&e.flags)}function Rn(e){return!!e&&!!(16777216&e.flags)}function Fn(e){return!!e&&!!(2097152&e.flags)}function Gn(n){return e.isTypeReferenceNode(n)&&e.isIdentifier(n.typeName)&&"Object"===n.typeName.escapedText&&n.typeArguments&&2===n.typeArguments.length&&(138===n.typeArguments[0].kind||135===n.typeArguments[0].kind)}function Vn(n,t){if(191!==n.kind)return!1;var r=n,a=r.expression,i=r.arguments;if(72!==a.kind||"require"!==a.escapedText)return!1;if(1!==i.length)return!1;var o=i[0];return!t||e.isStringLiteralLike(o)}function Bn(e){return 39===e||34===e}function Kn(e,n){return 34===U(n,e).charCodeAt(0)}function jn(n){if(n.parent){var t,r;if(e.isVariableDeclaration(n.parent)&&n.parent.initializer===n){if(!Pn(n)&&!Pe(n.parent))return;t=n.parent.name,r=n.parent}else if(e.isBinaryExpression(n.parent)&&59===n.parent.operatorToken.kind&&n.parent.right===n)t=n.parent.left,r=t;else if(e.isBinaryExpression(n.parent)&&55===n.parent.operatorToken.kind&&(e.isVariableDeclaration(n.parent.parent)&&n.parent.parent.initializer===n.parent?(t=n.parent.parent.name,r=n.parent.parent):e.isBinaryExpression(n.parent.parent)&&59===n.parent.parent.operatorToken.kind&&n.parent.parent.right===n.parent&&(t=n.parent.parent.left,r=t),!t||!za(t)||!Zn(t,n.parent.left)))return;if(t&&Jn(n,Xa(t)))return r}}function Un(n){return e.isBinaryExpression(n)||e.isPropertyAccessExpression(n)||e.isIdentifier(n)||e.isCallExpression(n)}function Hn(n){return Pn(n)&&n.initializer&&e.isBinaryExpression(n.initializer)&&55===n.initializer.operatorToken.kind&&n.name&&za(n.name)&&Zn(n.name,n.initializer.left)?n.initializer.right:n.initializer}function Wn(e){var n=Hn(e);return n&&Jn(n,Xa(e.name))}function qn(n,t){return e.forEach(n.properties,function(n){return e.isPropertyAssignment(n)&&e.isIdentifier(n.name)&&"value"===n.name.escapedText&&n.initializer&&Jn(n.initializer,t)})}function zn(n){if(n&&n.parent&&e.isBinaryExpression(n.parent)&&59===n.parent.operatorToken.kind){var t=Xa(n.parent.left);return Jn(n.parent.right,t)||Xn(n.parent.left,n.parent.right,t)}if(n&&e.isCallExpression(n)&&rt(n)){var r=qn(n.arguments[2],"prototype"===n.arguments[1].text);if(r)return r}}function Jn(n,t){if(e.isCallExpression(n)){var r=Bt(n.expression);return 196===r.kind||197===r.kind?n:void 0}return 196===n.kind||209===n.kind||197===n.kind?n:e.isObjectLiteralExpression(n)&&(0===n.properties.length||t)?n:void 0}function Xn(n,t,r){var a=e.isBinaryExpression(t)&&55===t.operatorToken.kind&&Jn(t.right,r);if(a&&Zn(n,t.left))return a}function Yn(n){var t=e.isVariableDeclaration(n.parent)?n.parent.name:e.isBinaryExpression(n.parent)&&59===n.parent.operatorToken.kind?n.parent.left:void 0;return t&&Jn(n.right,Xa(t))&&za(t)&&Zn(t,n.left)}function Qn(n){if(e.isBinaryExpression(n.parent)){var t=55===n.parent.operatorToken.kind&&e.isBinaryExpression(n.parent.parent)?n.parent.parent:n.parent;if(59===t.operatorToken.kind&&e.isIdentifier(t.left))return t.left}else if(e.isVariableDeclaration(n.parent))return n.parent.name}function Zn(n,t){return e.isIdentifier(n)&&e.isIdentifier(t)?n.escapedText===t.escapedText:e.isIdentifier(n)&&e.isPropertyAccessExpression(t)?(100===t.expression.kind||e.isIdentifier(t.expression)&&("window"===t.expression.escapedText||"self"===t.expression.escapedText||"global"===t.expression.escapedText))&&Zn(n,t.name):!(!e.isPropertyAccessExpression(n)||!e.isPropertyAccessExpression(t))&&(n.name.escapedText===t.name.escapedText&&Zn(n.expression,t.expression))}function $n(e){while(Ha(e,!0))e=e.right;return e}function et(n){return e.isIdentifier(n)&&"exports"===n.escapedText}function nt(n){return e.isPropertyAccessExpression(n)&&e.isIdentifier(n.expression)&&"module"===n.expression.escapedText&&"exports"===n.name.escapedText}function tt(e){var n=at(e);return 5===n||Pn(e)?n:0}function rt(n){return 3===e.length(n.arguments)&&e.isPropertyAccessExpression(n.expression)&&e.isIdentifier(n.expression.expression)&&"Object"===e.idText(n.expression.expression)&&"defineProperty"===e.idText(n.expression.name)&&pr(n.arguments[1])&&za(n.arguments[0])}function at(n){if(e.isCallExpression(n)){if(!rt(n))return 0;var t=n.arguments[0];return et(t)||nt(t)?8:e.isPropertyAccessExpression(t)&&"prototype"===t.name.escapedText&&za(t.expression)?9:7}if(59!==n.operatorToken.kind||!e.isPropertyAccessExpression(n.left))return 0;var r=n.left;return za(r.expression)&&"prototype"===r.name.escapedText&&e.isObjectLiteralExpression(ot(n))?6:it(r)}function it(n){if(100===n.expression.kind)return 4;if(nt(n))return 2;if(za(n.expression)){if(Xa(n.expression))return 3;var t=n;while(e.isPropertyAccessExpression(t.expression))t=t.expression;e.Debug.assert(e.isIdentifier(t.expression));var r=t.expression;return"exports"===r.escapedText||"module"===r.escapedText&&"exports"===t.name.escapedText?1:5}return 0}function ot(n){while(e.isBinaryExpression(n.right))n=n.right;return n.right}function st(n){return e.isBinaryExpression(n)&&3===tt(n)}function ct(n){return Pn(n)&&n.parent&&221===n.parent.kind&&!!e.getJSDocTypeTag(n.parent)}function lt(n){if(!n||!n.valueDeclaration)return!1;var t=n.valueDeclaration;return 239===t.kind||e.isVariableDeclaration(t)&&t.initializer&&e.isFunctionLike(t.initializer)}function ut(n){return dt(n)||e.Debug.fail(e.Debug.showSyntaxKind(n.parent))}function dt(n){switch(n.parent.kind){case 249:case 255:return n.parent;case 259:return n.parent.parent;case 191:return Ge(n.parent)||Vn(n.parent,!1)?n.parent:void 0;case 182:return e.Debug.assert(e.isStringLiteral(n)),e.tryCast(n.parent.parent,e.isImportTypeNode);default:return}}function pt(n){switch(n.kind){case 249:case 255:return n.moduleSpecifier;case 248:return 259===n.moduleReference.kind?n.moduleReference.expression:void 0;case 183:return Ve(n)?n.argument.literal:void 0;default:return e.Debug.assertNever(n)}}function mt(n){switch(n.kind){case 249:return n.importClause&&e.tryCast(n.importClause.namedBindings,e.isNamespaceImport);case 248:return n;case 255:return;default:return e.Debug.assertNever(n)}}function ft(e){return 249===e.kind&&!!e.importClause&&!!e.importClause.name}function _t(e){if(e)switch(e.kind){case 151:case 156:case 155:case 276:case 275:case 154:case 153:return void 0!==e.questionToken}return!1}function gt(n){var t=e.isJSDocFunctionType(n)?e.firstOrUndefined(n.parameters):void 0,r=e.tryCast(t&&t.name,e.isIdentifier);return!!r&&"new"===r.escapedText}function vt(e){return 304===e.kind||297===e.kind}function yt(n){return vt(n)||e.isTypeAliasDeclaration(n)}function ht(n){return e.isExpressionStatement(n)&&n.expression&&e.isBinaryExpression(n.expression)&&59===n.expression.operatorToken.kind?n.expression.right:void 0}function bt(n){return e.isExpressionStatement(n)&&e.isBinaryExpression(n.expression)&&0!==tt(n.expression)&&e.isBinaryExpression(n.expression.right)&&55===n.expression.right.operatorToken.kind?n.expression.right.right:void 0}function Et(e){switch(e.kind){case 219:var n=Tt(e);return n&&n.initializer;case 154:return e.initializer;case 275:return e.initializer}}function Tt(n){return e.isVariableStatement(n)?e.firstOrUndefined(n.declarationList.declarations):void 0}function St(n){return e.isModuleDeclaration(n)&&n.body&&244===n.body.kind?n.body:void 0}function xt(n){var t;Qe(n)&&e.hasInitializer(n)&&e.hasJSDocNodes(n.initializer)&&(t=e.addRange(t,n.initializer.jsDoc));var r=n;while(r&&r.parent){if(e.hasJSDocNodes(r)&&(t=e.addRange(t,r.jsDoc)),151===r.kind){t=e.addRange(t,e.getJSDocParameterTags(r));break}if(150===r.kind){t=e.addRange(t,e.getJSDocTypeParameterTags(r));break}r=Ct(r)}return t||e.emptyArray}function Ct(n){var t=n.parent;return 275===t.kind||154===t.kind||221===t.kind&&189===n.kind||St(t)||e.isBinaryExpression(n)&&59===n.operatorToken.kind?t:t.parent&&(Tt(t.parent)===n||e.isBinaryExpression(t)&&59===t.operatorToken.kind)?t.parent:t.parent&&t.parent.parent&&(Tt(t.parent.parent)||Et(t.parent.parent)===n||bt(t.parent.parent))?t.parent.parent:void 0}function At(n){if(n.symbol)return n.symbol;if(e.isIdentifier(n.name)){var t=n.name.escapedText,r=Dt(n);if(r){var a=e.find(r.parameters,function(e){return 72===e.name.kind&&e.name.escapedText===t});return a&&a.symbol}}}function Dt(e){return Lt(kt(e))}function Lt(n){var t=bt(n)||ht(n)||Et(n)||Tt(n)||St(n)||n;return t&&e.isFunctionLike(t)?t:void 0}function kt(n){return e.Debug.assertDefined(u(n.parent,e.isJSDoc)).parent}function Nt(n){var t=n.name.escapedText,r=n.parent.parent.parent.typeParameters;return e.find(r,function(e){return e.name.escapedText===t})}function It(n){var t=e.lastOrUndefined(n.parameters);return!!t&&wt(t)}function wt(n){var t=e.isJSDocParameterTag(n)?n.typeExpression&&n.typeExpression.type:n.type;return void 0!==n.dotDotDotToken||!!t&&290===t.kind}function Mt(e){var n=e.parent;while(1){switch(n.kind){case 204:var t=n.operatorToken.kind;return Ka(t)&&n.left===e?59===t?1:2:0;case 202:case 203:var r=n.operator;return 44===r||45===r?2:0;case 226:case 227:return n.initializer===e?1:0;case 195:case 187:case 208:case 213:e=n;break;case 276:if(n.name!==e)return 0;e=n.parent;break;case 275:if(n.name===e)return 0;e=n.parent;break;default:return 0}n=e.parent}}function Ot(e){return 0!==Mt(e)}function Pt(e){switch(e.kind){case 218:case 219:case 231:case 222:case 232:case 246:case 271:case 272:case 233:case 225:case 226:case 227:case 223:case 224:case 235:case 274:return!0}return!1}function Rt(n){return e.isFunctionExpression(n)||e.isArrowFunction(n)||e.isMethodOrAccessor(n)||e.isFunctionDeclaration(n)||e.isConstructorDeclaration(n)}function Ft(e,n){while(e&&e.kind===n)e=e.parent;return e}function Gt(e){return Ft(e,177)}function Vt(e){return Ft(e,195)}function Bt(e){while(195===e.kind)e=e.expression;return e}function Kt(e){while(195===e.kind)e=e.parent;return e}function jt(e){return(189===e.kind||190===e.kind)&&(e=Vt(e.parent),e&&198===e.kind)}function Ut(e,n){while(e){if(e===n)return!0;e=e.parent}return!1}function Ht(n){return!e.isSourceFile(n)&&!e.isBindingPattern(n)&&e.isDeclaration(n.parent)&&n.parent.name===n}function Wt(n){var t=n.parent;switch(n.kind){case 10:case 8:if(e.isComputedPropertyName(t))return t.parent;case 72:if(e.isDeclaration(t))return t.name===n?t:void 0;if(e.isQualifiedName(t)){var r=t.parent;return e.isJSDocParameterTag(r)&&r.name===t?r:void 0}var a=t.parent;return e.isBinaryExpression(a)&&0!==tt(a)&&(a.left.symbol||a.symbol)&&e.getNameOfDeclaration(a)===n?a:void 0;default:return}}function qt(n){return(10===n.kind||8===n.kind)&&149===n.parent.kind&&e.isDeclaration(n.parent.parent)}function zt(e){var n=e.parent;switch(n.kind){case 154:case 153:case 156:case 155:case 158:case 159:case 278:case 275:case 189:return n.name===e;case 148:if(n.right===e){while(148===n.kind)n=n.parent;return 167===n.kind||164===n.kind}return!1;case 186:case 253:return n.propertyName===e;case 257:case 267:return!0}return!1}function Jt(n){return 248===n.kind||247===n.kind||250===n.kind&&!!n.name||251===n.kind||253===n.kind||257===n.kind||254===n.kind&&Xt(n)||e.isBinaryExpression(n)&&2===tt(n)&&Xt(n)}function Xt(n){var t=e.isExportAssignment(n)?n.expression:n.right;return za(t)||e.isClassExpression(t)}function Yt(n){if(Pn(n)){var t=e.getJSDocAugmentsTag(n);if(t)return t.class}return Qt(n)}function Qt(e){var n=nr(e.heritageClauses,86);return n&&n.types.length>0?n.types[0]:void 0}function Zt(e){var n=nr(e.heritageClauses,109);return n?n.types:void 0}function $t(n){return e.isInterfaceDeclaration(n)?er(n)||e.emptyArray:e.isClassLike(n)&&e.concatenate(e.singleElementArray(Yt(n)),Zt(n))||e.emptyArray}function er(e){var n=nr(e.heritageClauses,86);return n?n.types:void 0}function nr(e,n){if(e)for(var t=0,r=e;t<r.length;t++){var a=r[t];if(a.token===n)return a}}function tr(n,t,r){if(!n.getCompilerOptions().noResolve){var a=e.isRootedDiskPath(r.fileName)?r.fileName:e.combinePaths(e.getDirectoryPath(t.fileName),r.fileName);return n.getSourceFile(a)}}function rr(e,n){while(e){if(e.kind===n)return e;e=e.parent}}function ar(e){return 73<=e&&e<=147}function ir(e){return 118<=e&&e<=147}function or(e){return ar(e)&&!ir(e)}function sr(n){var t=e.stringToToken(n);return void 0!==t&&or(t)}function cr(e){var n=e.originalKeywordKind;return!!n&&!ir(n)}function lr(e){return 2<=e&&e<=7}function ur(e){if(!e)return 4;var n=0;switch(e.kind){case 239:case 196:case 156:e.asteriskToken&&(n|=1);case 197:Ma(e,256)&&(n|=2);break}return e.body||(n|=4),n}function dr(e){switch(e.kind){case 239:case 196:case 197:case 156:return void 0!==e.body&&void 0===e.asteriskToken&&Ma(e,256)}return!1}function pr(n){return e.isStringLiteralLike(n)||e.isNumericLiteral(n)}function mr(n){var t=e.getNameOfDeclaration(n);return!!t&&fr(t)}function fr(e){return 149===e.kind&&!pr(e.expression)&&!_r(e.expression)}function _r(n){return e.isPropertyAccessExpression(n)&&Tr(n.expression)}function gr(n){switch(n.kind){case 72:return n.escapedText;case 10:case 8:return e.escapeLeadingUnderscores(n.text);case 149:var t=n.expression;return _r(t)?br(e.idText(t.name)):pr(t)?e.escapeLeadingUnderscores(t.text):void 0;default:return e.Debug.assertNever(n)}}function vr(e){switch(e.kind){case 72:case 10:case 14:case 8:return!0;default:return!1}}function yr(n){return 72===n.kind?e.idText(n):n.text}function hr(n){return 72===n.kind?n.escapedText:e.escapeLeadingUnderscores(n.text)}function br(e){return"__@"+e}function Er(n){return e.startsWith(n.escapedName,"__@")}function Tr(e){return 72===e.kind&&"Symbol"===e.escapedText}function Sr(e){return"push"===e.escapedText||"unshift"===e.escapedText}function xr(e){var n=Cr(e);return 151===n.kind}function Cr(e){while(186===e.kind)e=e.parent.parent;return e}function Ar(e){var n=e.kind;return 157===n||196===n||239===n||197===n||156===n||158===n||159===n||244===n||279===n}function Dr(n){return e.positionIsSynthesized(n.pos)||e.positionIsSynthesized(n.end)}function Lr(n){return e.getParseTreeNode(n,e.isSourceFile)||n}function kr(e){var n=wr(e),t=192===e.kind&&void 0!==e.arguments;return Nr(e.kind,n,t)}function Nr(e,n,t){switch(e){case 192:return t?0:1;case 202:case 199:case 200:case 198:case 201:case 205:case 207:return 1;case 204:switch(n){case 41:case 59:case 60:case 61:case 63:case 62:case 64:case 65:case 66:case 67:case 68:case 69:case 71:case 70:return 1}}return 0}function Ir(e){var n=wr(e),t=192===e.kind&&void 0!==e.arguments;return Mr(e.kind,n,t)}function wr(e){return 204===e.kind?e.operatorToken.kind:202===e.kind||203===e.kind?e.operator:e.kind}function Mr(e,n,t){switch(e){case 309:return 0;case 208:return 1;case 207:return 2;case 205:return 4;case 204:switch(n){case 27:return 0;case 59:case 60:case 61:case 63:case 62:case 64:case 65:case 66:case 67:case 68:case 69:case 71:case 70:return 3;default:return Or(n)}case 202:case 199:case 200:case 198:case 201:return 16;case 203:return 17;case 191:return 18;case 192:return t?19:18;case 193:case 189:case 190:return 19;case 100:case 98:case 72:case 96:case 102:case 87:case 8:case 9:case 10:case 187:case 188:case 196:case 197:case 209:case 260:case 261:case 264:case 13:case 14:case 206:case 195:case 210:return 20;default:return-1}}function Or(e){switch(e){case 55:return 5;case 54:return 6;case 50:return 7;case 51:return 8;case 49:return 9;case 33:case 34:case 35:case 36:return 10;case 28:case 30:case 31:case 32:case 94:case 93:case 119:return 11;case 46:case 47:case 48:return 12;case 38:case 39:return 13;case 40:case 42:case 43:return 14;case 41:return 15}return-1}function Pr(){var n=[],t=[],r=e.createMap(),a=!1;return{add:s,lookup:o,getGlobalDiagnostics:c,getDiagnostics:l,reattachFileDiagnostics:i};function i(n){e.forEach(r.get(n.fileName),function(e){return e.file=n})}function o(t){var a;if(a=t.file?r.get(t.file.fileName):n,a){var i=e.binarySearch(a,t,e.identity,e.compareDiagnosticsSkipRelatedInformation);return i>=0?a[i]:void 0}}function s(i){var o;i.file?(o=r.get(i.file.fileName),o||(o=[],r.set(i.file.fileName,o),e.insertSorted(t,i.file.fileName,e.compareStringsCaseSensitive))):(a&&(a=!1,n=n.slice()),o=n),e.insertSorted(o,i,e.compareDiagnostics)}function c(){return a=!0,n}function l(a){if(a)return r.get(a)||[];var i=e.flatMapToMutable(t,function(e){return r.get(e)});return n.length?(i.unshift.apply(i,n),i):i}}e.isPartOfTypeNode=We,e.isChildOfNodeWithKind=qe,e.forEachReturnStatement=ze,e.forEachYieldExpression=Je,e.getRestParameterElementType=Xe,e.getMembersOfDeclaration=Ye,e.isVariableLike=Qe,e.isVariableLikeOrAccessor=Ze,e.isVariableDeclarationInVariableStatement=$e,e.isValidESSymbolDeclaration=en,e.introducesArgumentsExoticObject=nn,e.unwrapInnermostStatementOfLabel=tn,e.isFunctionBlock=rn,e.isObjectLiteralMethod=an,e.isObjectLiteralOrClassExpressionMethod=on,e.isIdentifierTypePredicate=sn,e.isThisTypePredicate=cn,e.getPropertyAssignment=ln,e.getTsConfigObjectLiteralExpression=un,e.getTsConfigPropArrayElementValue=dn,e.getTsConfigPropArray=pn,e.getContainingFunction=mn,e.getContainingClass=fn,e.getThisContainer=_n,e.getNewTargetContainer=gn,e.getSuperContainer=vn,e.getImmediatelyInvokedFunctionExpression=yn,e.isSuperProperty=hn,e.isThisProperty=bn,e.getEntityNameFromTypeNode=En,e.getInvokedExpression=Tn,e.nodeCanBeDecorated=Sn,e.nodeIsDecorated=xn,e.nodeOrChildIsDecorated=Cn,e.childIsDecorated=An,e.isJSXTagName=Dn,e.isExpressionNode=Ln,e.isInExpressionContext=kn,e.isExternalModuleImportEqualsDeclaration=Nn,e.getExternalModuleImportEqualsDeclarationExpression=In,e.isInternalModuleImportEqualsDeclaration=wn,e.isSourceFileJS=Mn,e.isSourceFileNotJS=On,e.isInJSFile=Pn,e.isInJsonFile=Rn,e.isInJSDoc=Fn,e.isJSDocIndexSignature=Gn,e.isRequireCall=Vn,e.isSingleOrDoubleQuote=Bn,e.isStringDoubleQuoted=Kn,e.getDeclarationOfExpando=jn,e.isAssignmentDeclaration=Un,e.getEffectiveInitializer=Hn,e.getDeclaredExpandoInitializer=Wn,e.getAssignedExpandoInitializer=zn,e.getExpandoInitializer=Jn,e.isDefaultedExpandoInitializer=Yn,e.getNameOfExpando=Qn,e.getRightMostAssignedExpression=$n,e.isExportsIdentifier=et,e.isModuleExportsPropertyAccessExpression=nt,e.getAssignmentDeclarationKind=tt,e.isBindableObjectDefinePropertyCall=rt,e.getAssignmentDeclarationPropertyAccessKind=it,e.getInitializerOfBinaryExpression=ot,e.isPrototypePropertyAssignment=st,e.isSpecialPropertyDeclaration=ct,e.isFunctionSymbol=lt,e.importFromModuleSpecifier=ut,e.tryGetImportFromModuleSpecifier=dt,e.getExternalModuleName=pt,e.getNamespaceDeclarationNode=mt,e.isDefaultImport=ft,e.hasQuestionToken=_t,e.isJSDocConstructSignature=gt,e.isJSDocTypeAlias=vt,e.isTypeAlias=yt,e.getSingleInitializerOfVariableStatementOrPropertyDeclaration=Et,e.getJSDocCommentsAndTags=xt,e.getParameterSymbolFromJSDoc=At,e.getHostSignatureFromJSDoc=Dt,e.getHostSignatureFromJSDocHost=Lt,e.getJSDocHost=kt,e.getTypeParameterFromJsDoc=Nt,e.hasRestParameter=It,e.isRestParameter=wt,function(e){e[e["None"]=0]="None",e[e["Definite"]=1]="Definite",e[e["Compound"]=2]="Compound"}(e.AssignmentKind||(e.AssignmentKind={})),e.getAssignmentTargetKind=Mt,e.isAssignmentTarget=Ot,e.isNodeWithPossibleHoistedDeclaration=Pt,e.isValueSignatureDeclaration=Rt,e.walkUpParenthesizedTypes=Gt,e.walkUpParenthesizedExpressions=Vt,e.skipParentheses=Bt,e.isDeleteTarget=jt,e.isNodeDescendantOf=Ut,e.isDeclarationName=Ht,e.getDeclarationFromName=Wt,e.isLiteralComputedPropertyDeclarationName=qt,e.isIdentifierName=zt,e.isAliasSymbolDeclaration=Jt,e.exportAssignmentIsAlias=Xt,e.getEffectiveBaseTypeNode=Yt,e.getClassExtendsHeritageElement=Qt,e.getClassImplementsHeritageClauseElements=Zt,e.getAllSuperTypeNodes=$t,e.getInterfaceBaseTypeNodes=er,e.getHeritageClause=nr,e.tryResolveScriptReference=tr,e.getAncestor=rr,e.isKeyword=ar,e.isContextualKeyword=ir,e.isNonContextualKeyword=or,e.isStringANonContextualKeyword=sr,e.isIdentifierANonContextualKeyword=cr,e.isTrivia=lr,function(e){e[e["Normal"]=0]="Normal",e[e["Generator"]=1]="Generator",e[e["Async"]=2]="Async",e[e["Invalid"]=4]="Invalid",e[e["AsyncGenerator"]=3]="AsyncGenerator"}(e.FunctionFlags||(e.FunctionFlags={})),e.getFunctionFlags=ur,e.isAsyncFunction=dr,e.isStringOrNumericLiteralLike=pr,e.hasDynamicName=mr,e.isDynamicName=fr,e.isWellKnownSymbolSyntactically=_r,e.getPropertyNameForPropertyNameNode=gr,e.isPropertyNameLiteral=vr,e.getTextOfIdentifierOrLiteral=yr,e.getEscapedTextOfIdentifierOrLiteral=hr,e.getPropertyNameForKnownSymbolName=br,e.isKnownSymbol=Er,e.isESSymbolIdentifier=Tr,e.isPushOrUnshiftIdentifier=Sr,e.isParameterDeclaration=xr,e.getRootDeclaration=Cr,e.nodeStartsNewLexicalEnvironment=Ar,e.nodeIsSynthesized=Dr,e.getOriginalSourceFile=Lr,function(e){e[e["Left"]=0]="Left",e[e["Right"]=1]="Right"}(e.Associativity||(e.Associativity={})),e.getExpressionAssociativity=kr,e.getOperatorAssociativity=Nr,e.getExpressionPrecedence=Ir,e.getOperator=wr,e.getOperatorPrecedence=Mr,e.getBinaryOperatorPrecedence=Or,e.createDiagnosticCollection=Pr;var Rr=/[\\\"\u0000-\u001f\t\v\f\b\r\n\u2028\u2029\u0085]/g,Fr=/[\\\'\u0000-\u001f\t\v\f\b\r\n\u2028\u2029\u0085]/g,Gr=/[\\\`\u0000-\u001f\t\v\f\b\r\n\u2028\u2029\u0085]/g,Vr=e.createMapFromTemplate({"\t":"\\t","\v":"\\v","\f":"\\f","\b":"\\b","\r":"\\r","\n":"\\n","\\":"\\\\",'"':'\\"',"'":"\\'","`":"\\`","\u2028":"\\u2028","\u2029":"\\u2029","":"\\u0085"});function Br(e,n){var t=96===n?Gr:39===n?Fr:Rr;return e.replace(t,Kr)}function Kr(e,n,t){if(0===e.charCodeAt(0)){var r=t.charCodeAt(n+e.length);return r>=48&&r<=57?"\\x00":"\\0"}return Vr.get(e)||Ur(e.charCodeAt(0))}function jr(n){var t=n.charCodeAt(0);return t>=97&&t<=122||e.stringContains(n,"-")}function Ur(e){var n=e.toString(16).toUpperCase(),t=("0000"+n).slice(-4);return"\\u"+t}e.escapeString=Br,e.isIntrinsicJsxName=jr;var Hr=/[^\u0000-\u007F]/g;function Wr(e,n){return e=Br(e,n),Hr.test(e)?e.replace(Hr,function(e){return Ur(e.charCodeAt(0))}):e}e.escapeNonAsciiString=Wr;var qr=["","    "];function zr(e){return void 0===qr[e]&&(qr[e]=zr(e-1)+qr[1]),qr[e]}function Jr(){return qr[1].length}function Xr(n){var t,r,a,i,o;function s(n){var r=e.computeLineStarts(n);r.length>1?(i=i+r.length-1,o=t.length-n.length+e.last(r),a=o-t.length===0):a=!1}function c(e){e&&e.length&&(a&&(e=zr(r)+e,a=!1),t+=e,s(e))}function l(){t="",r=0,a=!0,i=0,o=0}function u(e){void 0!==e&&(t+=e,s(e))}function d(e){e&&e.length&&c(e)}function p(){a||(t+=n,i++,o=t.length,a=!0)}return l(),{write:c,rawWrite:u,writeLiteral:d,writeLine:p,increaseIndent:function(){r++},decreaseIndent:function(){r--},getIndent:function(){return r},getTextPos:function(){return t.length},getLine:function(){return i},getColumn:function(){return a?r*Jr():t.length-o},getText:function(){return t},isAtStartOfLine:function(){return a},clear:l,reportInaccessibleThisError:e.noop,reportPrivateInBaseOfClassExpression:e.noop,reportInaccessibleUniqueSymbolError:e.noop,trackSymbol:e.noop,writeKeyword:c,writeOperator:c,writeParameter:c,writeProperty:c,writePunctuation:c,writeSpace:c,writeStringLiteral:c,writeSymbol:function(e,n){return c(e)},writeTrailingSemicolon:c,writeComment:c}}function Yr(e){var n=!1;function t(){n&&(e.writeTrailingSemicolon(";"),n=!1)}return o({},e,{writeTrailingSemicolon:function(){n=!0},writeLiteral:function(n){t(),e.writeLiteral(n)},writeStringLiteral:function(n){t(),e.writeStringLiteral(n)},writeSymbol:function(n,r){t(),e.writeSymbol(n,r)},writePunctuation:function(n){t(),e.writePunctuation(n)},writeKeyword:function(n){t(),e.writeKeyword(n)},writeOperator:function(n){t(),e.writeOperator(n)},writeParameter:function(n){t(),e.writeParameter(n)},writeSpace:function(n){t(),e.writeSpace(n)},writeProperty:function(n){t(),e.writeProperty(n)},writeComment:function(n){t(),e.writeComment(n)},writeLine:function(){t(),e.writeLine()},increaseIndent:function(){t(),e.increaseIndent()},decreaseIndent:function(){t(),e.decreaseIndent()}})}function Qr(e,n,t){return n.moduleName||$r(e,n.fileName,t&&t.fileName)}function Zr(e,n,t){var r=n.getExternalModuleFileFromDeclaration(t);if(r&&!r.isDeclarationFile)return Qr(e,r)}function $r(n,t,r){var a=function(e){return n.getCanonicalFileName(e)},i=c(r?e.getDirectoryPath(r):n.getCommonSourceDirectory(),n.getCurrentDirectory(),a),o=e.getNormalizedAbsolutePath(t,n.getCurrentDirectory()),s=e.getRelativePathToDirectoryOrUrl(i,o,i,a,!1),l=e.removeFileExtension(s);return r?e.ensurePathIsNonModuleName(l):l}function ea(n,t,r){var a,i=t.getCompilerOptions();return a=i.outDir?e.removeFileExtension(ia(n,t,i.outDir)):e.removeFileExtension(n),a+r}function na(e,n){return ta(e,n.getCompilerOptions(),n.getCurrentDirectory(),n.getCommonSourceDirectory(),function(e){return n.getCanonicalFileName(e)})}function ta(n,t,r,a,i){var o=t.declarationDir||t.outDir,s=o?oa(n,o,r,a,i):n;return e.removeFileExtension(s)+".d.ts"}function ra(n,t){var r=n.getCompilerOptions(),a=function(e){return n.isSourceFileFromExternalLibrary(e)};if(r.outFile||r.out){var i=e.getEmitModuleKind(r),o=r.emitDeclarationOnly||i===e.ModuleKind.AMD||i===e.ModuleKind.System;return e.filter(n.getSourceFiles(),function(n){return(o||!e.isExternalModule(n))&&aa(n,r,a)})}var s=void 0===t?n.getSourceFiles():[t];return e.filter(s,function(e){return aa(e,r,a)})}function aa(e,n,t){return!(n.noEmitForJsFiles&&Mn(e))&&!e.isDeclarationFile&&!t(e)}function ia(e,n,t){return oa(e,t,n.getCurrentDirectory(),n.getCommonSourceDirectory(),function(e){return n.getCanonicalFileName(e)})}function oa(n,t,r,a,i){var o=e.getNormalizedAbsolutePath(n,r),s=0===i(o).indexOf(i(a));return o=s?o.substring(a.length):o,e.combinePaths(t,o)}function sa(n,t,r,a,i,o){n.writeFile(r,a,i,function(n){t.add(e.createCompilerDiagnostic(e.Diagnostics.Could_not_write_file_0_Colon_1,r,n))},o)}function ca(n,t){return e.getLineAndCharacterOfPosition(n,t).line}function la(n,t){return e.computeLineAndCharacterOfPosition(n,t).line}function ua(n){return e.find(n.members,function(n){return e.isConstructorDeclaration(n)&&F(n.body)})}function da(e){if(e&&e.parameters.length>0){var n=2===e.parameters.length&&fa(e.parameters[0]);return e.parameters[n?1:0]}}function pa(e){var n=da(e);return n&&n.type}function ma(n){if(n.parameters.length&&!e.isJSDocSignature(n)){var t=n.parameters[0];if(fa(t))return t}}function fa(e){return _a(e.name)}function _a(e){return!!e&&72===e.kind&&ga(e)}function ga(e){return 100===e.originalKeywordKind}function va(n,t){var r,a,i,o;return mr(t)?(r=t,158===t.kind?i=t:159===t.kind?o=t:e.Debug.fail("Accessor has wrong kind")):e.forEach(n,function(n){if(e.isAccessor(n)&&Ma(n,32)===Ma(t,32)){var s=gr(n.name),c=gr(t.name);s===c&&(r?a||(a=n):r=n,158!==n.kind||i||(i=n),159!==n.kind||o||(o=n))}}),{firstAccessor:r,secondAccessor:a,getAccessor:i,setAccessor:o}}function ya(n){var t=n.type;return t||!Pn(n)?t:e.isJSDocPropertyLikeTag(n)?n.typeExpression&&n.typeExpression.type:e.getJSDocType(n)}function ha(e){return e.type}function ba(n){return e.isJSDocSignature(n)?n.type&&n.type.typeExpression&&n.type.typeExpression.type:n.type||(Pn(n)?e.getJSDocReturnType(n):void 0)}function Ea(n){return e.flatMap(e.getJSDocTags(n),function(e){return Ta(e)?e.typeParameters:void 0})}function Ta(n){return e.isJSDocTemplateTag(n)&&!(291===n.parent.kind&&n.parent.tags.some(vt))}function Sa(e){var n=da(e);return n&&ya(n)}function xa(e,n,t,r){Ca(e,n,t.pos,r)}function Ca(e,n,t,r){r&&r.length&&t!==r[0].pos&&la(e,t)!==la(e,r[0].pos)&&n.writeLine()}function Aa(e,n,t,r){t!==r&&la(e,t)!==la(e,r)&&n.writeLine()}function Da(e,n,t,r,a,i,o,s){if(r&&r.length>0){a&&t.writeSpace(" ");for(var c=!1,l=0,u=r;l<u.length;l++){var d=u[l];c&&(t.writeSpace(" "),c=!1),s(e,n,t,d.pos,d.end,o),d.hasTrailingNewLine?t.writeLine():c=!0}c&&i&&t.writeSpace(" ")}}function La(n,t,r,a,i,o,s){var c,l;if(s?0===i.pos&&(c=e.filter(e.getLeadingCommentRanges(n,i.pos),y)):c=e.getLeadingCommentRanges(n,i.pos),c){for(var u=[],d=void 0,p=0,m=c;p<m.length;p++){var f=m[p];if(d){var _=la(t,d.end),g=la(t,f.pos);if(g>=_+2)break}u.push(f),d=f}if(u.length){_=la(t,e.last(u).end);var v=la(t,e.skipTrivia(n,i.pos));v>=_+2&&(xa(t,r,i,c),Da(n,t,r,u,!1,!0,o,a),l={nodePos:i.pos,detachedCommentEndPos:e.last(u).end})}}return l;function y(e){return B(n,e.pos)}}function ka(n,t,r,a,i,o){if(42===n.charCodeAt(a+1))for(var s=e.computeLineAndCharacterOfPosition(t,a),c=t.length,l=void 0,u=a,d=s.line;u<i;d++){var p=d+1===c?n.length+1:t[d+1];if(u!==a){void 0===l&&(l=Ia(n,t[s.line],a));var m=r.getIndent()*Jr(),f=m-l+Ia(n,u,p);if(f>0){var _=f%Jr(),g=zr((f-_)/Jr());r.rawWrite(g);while(_)r.rawWrite(" "),_--}else r.rawWrite("")}Na(n,i,r,o,u,p),u=p}else r.writeComment(n.substring(a,i))}function Na(e,n,t,r,a,i){var o=Math.min(n,i-1),s=e.substring(a,o).replace(/^\s+|\s+$/g,"");s?(t.writeComment(s),o!==n&&t.writeLine()):t.rawWrite(r)}function Ia(n,t,r){for(var a=0;t<r&&e.isWhiteSpaceSingleLine(n.charCodeAt(t));t++)9===n.charCodeAt(t)?a+=Jr()-a%Jr():a++;return a}function wa(e){return 0!==Fa(e)}function Ma(e,n){return!!Ra(e,n)}function Oa(e){return Ma(e,32)}function Pa(e){return Ma(e,64)}function Ra(e,n){return Fa(e)&n}function Fa(e){if(536870912&e.modifierFlagsCache)return-536870913&e.modifierFlagsCache;var n=Ga(e);return e.modifierFlagsCache=536870912|n,n}function Ga(e){var n=0;if(e.modifiers)for(var t=0,r=e.modifiers;t<r.length;t++){var a=r[t];n|=Va(a.kind)}return(4&e.flags||72===e.kind&&e.isInJSDocNamespace)&&(n|=1),n}function Va(e){switch(e){case 116:return 32;case 115:return 4;case 114:return 16;case 113:return 8;case 118:return 128;case 85:return 1;case 125:return 2;case 77:return 2048;case 80:return 512;case 121:return 256;case 133:return 64}return 0}function Ba(e){return 55===e||54===e||52===e}function Ka(e){return e>=59&&e<=71}function ja(e){var n=Ua(e);return n&&!n.isImplements?n.class:void 0}function Ua(n){return e.isExpressionWithTypeArguments(n)&&e.isHeritageClause(n.parent)&&e.isClassLike(n.parent.parent)?{class:n.parent.parent,isImplements:109===n.parent.token}:void 0}function Ha(n,t){return e.isBinaryExpression(n)&&(t?59===n.operatorToken.kind:Ka(n.operatorToken.kind))&&e.isLeftHandSideExpression(n.left)}function Wa(e){if(Ha(e,!0)){var n=e.left.kind;return 188===n||187===n}return!1}function qa(e){return void 0!==ja(e)}function za(e){return 72===e.kind||Ja(e)}function Ja(n){return e.isPropertyAccessExpression(n)&&za(n.expression)}function Xa(n){return e.isPropertyAccessExpression(n)&&"prototype"===n.name.escapedText}function Ya(e){return 148===e.parent.kind&&e.parent.right===e||189===e.parent.kind&&e.parent.name===e}function Qa(e){return 188===e.kind&&0===e.properties.length}function Za(e){return 187===e.kind&&0===e.elements.length}function $a(e){return ei(e)?e.declarations[0].localSymbol:void 0}function ei(n){return n&&e.length(n.declarations)>0&&Ma(n.declarations[0],512)}function ni(n){return e.find(e.supportedTSExtensionsForExtractExtension,function(t){return e.fileExtensionIs(n,t)})}function ti(n){for(var t=[],r=n.length,a=0;a<r;a++){var i=n.charCodeAt(a);i<128?t.push(i):i<2048?(t.push(i>>6|192),t.push(63&i|128)):i<65536?(t.push(i>>12|224),t.push(i>>6&63|128),t.push(63&i|128)):i<131072?(t.push(i>>18|240),t.push(i>>12&63|128),t.push(i>>6&63|128),t.push(63&i|128)):e.Debug.assert(!1,"Unexpected code point")}return t}e.getIndentString=zr,e.getIndentSize=Jr,e.createTextWriter=Xr,e.getTrailingSemicolonOmittingWriter=Yr,e.getResolvedExternalModuleName=Qr,e.getExternalModuleNameFromDeclaration=Zr,e.getExternalModuleNameFromPath=$r,e.getOwnEmitOutputFilePath=ea,e.getDeclarationEmitOutputFilePath=na,e.getDeclarationEmitOutputFilePathWorker=ta,e.getSourceFilesToEmit=ra,e.sourceFileMayBeEmitted=aa,e.getSourceFilePathInNewDir=ia,e.getSourceFilePathInNewDirWorker=oa,e.writeFile=sa,e.getLineOfLocalPosition=ca,e.getLineOfLocalPositionFromLineMap=la,e.getFirstConstructorWithBody=ua,e.getSetAccessorTypeAnnotationNode=pa,e.getThisParameter=ma,e.parameterIsThisKeyword=fa,e.isThisIdentifier=_a,e.identifierIsThisKeyword=ga,e.getAllAccessorDeclarations=va,e.getEffectiveTypeAnnotationNode=ya,e.getTypeAnnotationNode=ha,e.getEffectiveReturnTypeNode=ba,e.getJSDocTypeParameterDeclarations=Ea,e.getEffectiveSetAccessorTypeAnnotationNode=Sa,e.emitNewLineBeforeLeadingComments=xa,e.emitNewLineBeforeLeadingCommentsOfPosition=Ca,e.emitNewLineBeforeLeadingCommentOfPosition=Aa,e.emitComments=Da,e.emitDetachedComments=La,e.writeCommentRange=ka,e.hasModifiers=wa,e.hasModifier=Ma,e.hasStaticModifier=Oa,e.hasReadonlyModifier=Pa,e.getSelectedModifierFlags=Ra,e.getModifierFlags=Fa,e.getModifierFlagsNoCache=Ga,e.modifierToFlag=Va,e.isLogicalOperator=Ba,e.isAssignmentOperator=Ka,e.tryGetClassExtendingExpressionWithTypeArguments=ja,e.tryGetClassImplementingOrExtendingExpressionWithTypeArguments=Ua,e.isAssignmentExpression=Ha,e.isDestructuringAssignment=Wa,e.isExpressionWithTypeArgumentsInClassExtendsClause=qa,e.isEntityNameExpression=za,e.isPropertyAccessEntityNameExpression=Ja,e.isPrototypeAccess=Xa,e.isRightSideOfQualifiedNameOrPropertyAccess=Ya,e.isEmptyObjectLiteral=Qa,e.isEmptyArrayLiteral=Za,e.getLocalSymbolForExportDefault=$a,e.tryExtractTSExtension=ni;var ri="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";function ai(e){var n,t,r,a,i="",o=ti(e),s=0,c=o.length;while(s<c)n=o[s]>>2,t=(3&o[s])<<4|o[s+1]>>4,r=(15&o[s+1])<<2|o[s+2]>>6,a=63&o[s+2],s+1>=c?r=a=64:s+2>=c&&(a=64),i+=ri.charAt(n)+ri.charAt(t)+ri.charAt(r)+ri.charAt(a),s+=3;return i}function ii(e){var n="",t=0,r=e.length;while(t<r){var a=e[t];if(a<128)n+=String.fromCharCode(a),t++;else if(192===(192&a)){var i=63&a;t++;var o=e[t];while(128===(192&o))i=i<<6|63&o,t++,o=e[t];n+=String.fromCharCode(i)}else n+=String.fromCharCode(a),t++}return n}function oi(e,n){return e&&e.base64encode?e.base64encode(n):ai(n)}function si(e,n){if(e&&e.base64decode)return e.base64decode(n);var t=n.length,r=[],a=0;while(a<t){if(n.charCodeAt(a)===ri.charCodeAt(64))break;var i=ri.indexOf(n[a]),o=ri.indexOf(n[a+1]),s=ri.indexOf(n[a+2]),c=ri.indexOf(n[a+3]),l=(63&i)<<2|o>>4&3,u=(15&o)<<4|s>>2&15,d=(3&s)<<6|63&c;0===u&&0!==s?r.push(l):0===d&&0!==c?r.push(l,u):r.push(l,u,d),a+=4}return ii(r)}function ci(n,t){try{var r=t.readFile(n);if(!r)return{};var a=e.parseConfigFileTextToJson(n,r);return a.error?{}:a.config}catch(i){return{}}}function li(e,n){return!n.directoryExists||n.directoryExists(e)}e.convertToBase64=ai,e.base64encode=oi,e.base64decode=si,e.readJson=ci,e.directoryProbablyExists=li;var ui,di="\r\n",pi="\n";function mi(n,t){switch(n.newLine){case 0:return di;case 1:return pi}return t?t():e.sys?e.sys.newLine:di}function fi(e,n,t){void 0===e&&(e=0);var r=_i(n);if(0===e)return r.length>0&&0===r[0][0]?r[0][1]:"0";if(t){for(var a="",i=e,o=r.length-1;o>=0&&0!==i;o--){var s=r[o],c=s[0],l=s[1];0!==c&&(i&c)===c&&(i&=~c,a=l+(a?", ":"")+a)}if(0===i)return a}else for(var u=0,d=r;u<d.length;u++){var p=d[u];c=p[0],l=p[1];if(c===e)return l}return e.toString()}function _i(n){var t=[];for(var r in n){var a=n[r];"number"===typeof a&&t.push([a,r])}return e.stableSort(t,function(n,t){return e.compareValues(n[0],t[0])})}function gi(n){return fi(n,e.SyntaxKind,!1)}function vi(n){return fi(n,e.ModifierFlags,!0)}function yi(n){return fi(n,e.TransformFlags,!0)}function hi(n){return fi(n,e.EmitFlags,!0)}function bi(n){return fi(n,e.SymbolFlags,!0)}function Ei(n){return fi(n,e.TypeFlags,!0)}function Ti(n){return fi(n,e.ObjectFlags,!0)}function Si(n,t){return void 0===t&&(t=n),e.Debug.assert(t>=n||-1===t),{pos:n,end:t}}function xi(e,n){return Si(e.pos,n)}function Ci(e,n){return Si(n,e.end)}function Ai(e){return e.decorators&&e.decorators.length>0?Ci(e,e.decorators.end):e}function Di(e){return e.modifiers&&e.modifiers.length>0?Ci(e,e.modifiers.end):Ai(e)}function Li(e){return e.pos===e.end}function ki(n,t){return Si(n,n+e.tokenToString(t).length)}function Ni(e,n){return Mi(e,e,n)}function Ii(e,n,t){return Pi(Ri(e,t),Ri(n,t),t)}function wi(e,n,t){return Pi(e.end,n.end,t)}function Mi(e,n,t){return Pi(Ri(e,t),n.end,t)}function Oi(e,n,t){return Pi(e.end,Ri(n,t),t)}function Pi(e,n,t){return e===n||ca(t,e)===ca(t,n)}function Ri(n,t){return e.positionIsSynthesized(n.pos)?-1:e.skipTrivia(t.text,n.pos)}function Fi(n){var t=e.getParseTreeNode(n);if(t)switch(t.parent.kind){case 243:case 244:return t===t.parent.name}return!1}function Gi(n){return e.filter(n.declarations,Vi)}function Vi(e){return void 0!==e.initializer}function Bi(e){return e.watch&&e.hasOwnProperty("watch")}function Ki(e){e.close()}function ji(e){return 33554432&e.flags?e.checkFlags:0}function Ui(n){if(n.valueDeclaration){var t=e.getCombinedModifierFlags(n.valueDeclaration);return n.parent&&32&n.parent.flags?t:-29&t}if(6&ji(n)){var r=n.checkFlags,a=512&r?8:128&r?4:16,i=1024&r?32:0;return a|i}return 4194304&n.flags?36:0}function Hi(e,n){return 2097152&e.flags?n.getAliasedSymbol(e):e}function Wi(e){return e.exportSymbol?e.exportSymbol.flags|e.flags:e.flags}function qi(e){return 1===Ji(e)}function zi(e){return 0!==Ji(e)}function Ji(e){var n=e.parent;if(!n)return 0;switch(n.kind){case 195:return Ji(n);case 203:case 202:var t=n.operator;return 44===t||45===t?s():0;case 204:var r=n,a=r.left,i=r.operatorToken;return a===e&&Ka(i.kind)?59===i.kind?1:s():0;case 189:return n.name!==e?0:Ji(n);case 275:var o=Ji(n.parent);return e===n.name?Xi(o):o;case 276:return e===n.objectAssignmentInitializer?0:Ji(n.parent);case 187:return Ji(n);default:return 0}function s(){return n.parent&&221===Kt(n.parent).kind?1:2}}function Xi(n){switch(n){case 0:return 1;case 1:return 0;case 2:return 2;default:return e.Debug.assertNever(n)}}function Yi(e,n){if(!e||!n||Object.keys(e).length!==Object.keys(n).length)return!1;for(var t in e)if("object"===typeof e[t]){if(!Yi(e[t],n[t]))return!1}else if("function"!==typeof e[t]&&e[t]!==n[t])return!1;return!0}function Qi(e,n){e.forEach(n),e.clear()}function Zi(e,n,t){var r=t.createNewValue,a=t.onDeleteValue,i=t.onExistingValue;e.forEach(function(t,r){var o=n.get(r);void 0===o?(e.delete(r),a(t,r)):i&&i(t,o,r)}),n.forEach(function(n,t){e.has(t)||e.set(t,r(t,n))})}function $i(n,t){while(1){var r=t(n);if(void 0!==r)return r;var a=e.getDirectoryPath(n);if(a===n)return;n=a}}function eo(e){return!!(16&ro(e))&&!!e.symbol&&no(e.symbol)}function no(e){if(32&e.flags){var n=to(e);return!!n&&Ma(n,128)}return!1}function to(n){return e.find(n.declarations,e.isClassLike)}function ro(e){return 524288&e.flags?e.objectFlags:0}function ao(e,n){return 0!==n.getSignaturesOfType(e,0).length||0!==n.getSignaturesOfType(e,1).length}function io(e,n){return!!$i(e,function(e){return!!n(e)||void 0})}function oo(n){return!!n&&!!n.declarations&&!!n.declarations[0]&&e.isNamespaceExportDeclaration(n.declarations[0])}function so(n){var t=n.moduleSpecifier;return e.isStringLiteral(t)?t.text:q(t)}function co(n){var t;return e.forEachChild(n,function(e){F(e)&&(t=e)},function(e){for(var n=e.length-1;n>=0;n--)if(F(e[n])){t=e[n];break}}),t}function lo(e,n,t){return void 0===t&&(t=!0),n=String(n),!e.has(n)&&(e.set(n,t),!0)}function uo(n){return e.isClassLike(n)||e.isInterfaceDeclaration(n)||e.isTypeLiteralNode(n)}function po(e){return e>=163&&e<=183||120===e||143===e||135===e||146===e||136===e||123===e||138===e||139===e||100===e||106===e||141===e||96===e||132===e||211===e||284===e||285===e||286===e||287===e||288===e||289===e||290===e}function mo(e){return 189===e.kind||190===e.kind}e.getNewLineCharacter=mi,e.formatSyntaxKind=gi,e.formatModifierFlags=vi,e.formatTransformFlags=yi,e.formatEmitFlags=hi,e.formatSymbolFlags=bi,e.formatTypeFlags=Ei,e.formatObjectFlags=Ti,e.createRange=Si,e.moveRangeEnd=xi,e.moveRangePos=Ci,e.moveRangePastDecorators=Ai,e.moveRangePastModifiers=Di,e.isCollapsedRange=Li,e.createTokenRange=ki,e.rangeIsOnSingleLine=Ni,e.rangeStartPositionsAreOnSameLine=Ii,e.rangeEndPositionsAreOnSameLine=wi,e.rangeStartIsOnSameLineAsRangeEnd=Mi,e.rangeEndIsOnSameLineAsRangeStart=Oi,e.positionsAreOnSameLine=Pi,e.getStartPositionOfRange=Ri,e.isDeclarationNameOfEnumOrNamespace=Fi,e.getInitializedVariables=Gi,e.isWatchSet=Bi,e.closeFileWatcher=Ki,e.getCheckFlags=ji,e.getDeclarationModifierFlagsFromSymbol=Ui,e.skipAlias=Hi,e.getCombinedLocalAndExportSymbolFlags=Wi,e.isWriteOnlyAccess=qi,e.isWriteAccess=zi,function(e){e[e["Read"]=0]="Read",e[e["Write"]=1]="Write",e[e["ReadWrite"]=2]="ReadWrite"}(ui||(ui={})),e.compareDataObjects=Yi,e.clearMap=Qi,e.mutateMap=Zi,e.forEachAncestorDirectory=$i,e.isAbstractConstructorType=eo,e.isAbstractConstructorSymbol=no,e.getClassLikeDeclarationOfSymbol=to,e.getObjectFlags=ro,e.typeHasCallOrConstructSignatures=ao,e.forSomeAncestorDirectory=io,e.isUMDExportSymbol=oo,e.showModuleSpecifier=so,e.getLastChild=co,e.addToSeen=lo,e.isObjectTypeDeclaration=uo,e.isTypeNodeKind=po,e.isAccessExpression=mo}(u||(u={})),function(e){function n(e){switch(e.target){case 6:return"lib.esnext.full.d.ts";case 5:return"lib.es2018.full.d.ts";case 4:return"lib.es2017.full.d.ts";case 3:return"lib.es2016.full.d.ts";case 2:return"lib.es6.d.ts";default:return"lib.d.ts"}}function t(e){return e.start+e.length}function r(e){return 0===e.length}function a(e,n){return n>=e.start&&n<t(e)}function i(e,n){return n>=e.pos&&n<=e.end}function o(e,n){return n.start>=e.start&&t(n)<=t(e)}function s(e,n){return void 0!==c(e,n)}function c(e,n){var t=m(e,n);return t&&0===t.length?void 0:t}function l(e,n){return d(e.start,e.length,n.start,n.length)}function u(e,n,t){return d(e.start,e.length,n,t)}function d(e,n,t,r){var a=e+n,i=t+r;return t<=a&&i>=e}function p(e,n){return n<=t(e)&&n>=e.start}function m(e,n){var r=Math.max(e.start,n.start),a=Math.min(t(e),t(n));return r<=a?_(r,a):void 0}function f(e,n){if(e<0)throw new Error("start < 0");if(n<0)throw new Error("length < 0");return{start:e,length:n}}function _(e,n){return f(e,n-e)}function g(e){return f(e.span.start,e.newLength)}function v(e){return r(e.span)&&0===e.newLength}function y(e,n){if(n<0)throw new Error("newLength < 0");return{span:e,newLength:n}}function h(n){if(0===n.length)return e.unchangedTextChangeRange;if(1===n.length)return n[0];for(var r=n[0],a=r.span.start,i=t(r.span),o=a+r.newLength,s=1;s<n.length;s++){var c=n[s],l=a,u=i,d=o,p=c.span.start,m=t(c.span),f=p+c.newLength;a=Math.min(l,p),i=Math.max(u,u+(m-d)),o=Math.max(f,f+(d-m))}return y(_(a,i),o-a)}function b(n){if(n&&150===n.kind)for(var t=n;t;t=t.parent)if(e.isFunctionLike(t)||e.isClassLike(t)||241===t.kind)return t}function E(n){return e.hasModifier(n,92)&&157===n.parent.kind}function T(n){return!!e.isBindingPattern(n)&&e.every(n.elements,S)}function S(n){return!!e.isOmittedExpression(n)||T(n.name)}function x(n){var t=n.parent;while(e.isBindingElement(t.parent))t=t.parent.parent;return t.parent}function C(n,t){e.isBindingElement(n)&&(n=x(n));var r=t(n);return 237===n.kind&&(n=n.parent),n&&238===n.kind&&(r|=t(n),n=n.parent),n&&219===n.kind&&(r|=t(n)),r}function A(n){return C(n,e.getModifierFlags)}function D(e){return C(e,function(e){return e.flags})}function L(n,t,r){var a=/^([a-z]+)([_\-]([a-z]+))?$/.exec(n.toLowerCase());if(a){var i=a[1],o=a[3];s(i,o,r)||s(i,void 0,r),e.setUILocale(n)}else r&&r.push(e.createCompilerDiagnostic(e.Diagnostics.Locale_must_be_of_the_form_language_or_language_territory_For_example_0_or_1,"en","ja-jp"));function s(n,r,a){var i=e.normalizePath(t.getExecutingFilePath()),o=e.getDirectoryPath(i),s=e.combinePaths(o,n);if(r&&(s=s+"-"+r),s=t.resolvePath(e.combinePaths(s,"diagnosticMessages.generated.json")),!t.fileExists(s))return!1;var c="";try{c=t.readFile(s)}catch(l){return a&&a.push(e.createCompilerDiagnostic(e.Diagnostics.Unable_to_open_file_0,s)),!1}try{e.localizedDiagnosticMessages=JSON.parse(c)}catch(u){return a&&a.push(e.createCompilerDiagnostic(e.Diagnostics.Corrupted_locale_file_0,s)),!1}return!0}}function k(e,n){if(e)while(void 0!==e.original)e=e.original;return!n||n(e)?e:void 0}function N(e){return 0===(8&e.flags)}function I(e,n){return void 0===e||N(e)?e:(e=k(e),!N(e)||n&&!n(e)?void 0:e)}function w(e){return e.length>=2&&95===e.charCodeAt(0)&&95===e.charCodeAt(1)?"_"+e:e}function M(e){var n=e;return n.length>=3&&95===n.charCodeAt(0)&&95===n.charCodeAt(1)&&95===n.charCodeAt(2)?n.substr(1):n}function O(e){return M(e.escapedText)}function P(e){return M(e.escapedName)}function R(n){var t=n.parent.parent;if(t){if(e.isDeclaration(t))return F(t);switch(t.kind){case 219:if(t.declarationList&&t.declarationList.declarations[0])return F(t.declarationList.declarations[0]);break;case 221:var r=t.expression;switch(r.kind){case 189:return r.name;case 190:var a=r.argumentExpression;if(e.isIdentifier(a))return a}break;case 195:return F(t.expression);case 233:if(e.isDeclaration(t.statement)||e.isExpression(t.statement))return F(t.statement);break}}}function F(n){var t=K(n);return t&&e.isIdentifier(t)?t:void 0}function G(e){return e.name||R(e)}function V(e){return!!e.name}function B(n){switch(n.kind){case 72:return n;case 305:case 299:var t=n.name;if(148===t.kind)return t.right;break;case 191:case 204:var r=n;switch(e.getAssignmentDeclarationKind(r)){case 1:case 4:case 5:case 3:return r.left.name;case 7:case 8:case 9:return r.arguments[1];default:return}case 304:return G(n);case 254:var a=n.expression;return e.isIdentifier(a)?a:void 0}return n.name}function K(n){if(void 0!==n)return B(n)||(e.isFunctionExpression(n)||e.isClassExpression(n)?j(n):void 0)}function j(n){if(n.parent){if(e.isPropertyAssignment(n.parent)||e.isBindingElement(n.parent))return n.parent.name;if(e.isBinaryExpression(n.parent)&&n===n.parent.right){if(e.isIdentifier(n.parent.left))return n.parent.left;if(e.isPropertyAccessExpression(n.parent.left))return n.parent.left.name}}}function U(n){if(n.name){if(e.isIdentifier(n.name)){var t=n.name.escapedText;return ne(n.parent).filter(function(n){return e.isJSDocParameterTag(n)&&e.isIdentifier(n.name)&&n.name.escapedText===t})}var r=n.parent.parameters.indexOf(n);e.Debug.assert(r>-1,"Parameters should always be in their parents' parameter list");var a=ne(n.parent).filter(e.isJSDocParameterTag);if(r<a.length)return[a[r]]}return e.emptyArray}function H(n){var t=n.name.escapedText;return ne(n.parent).filter(function(n){return e.isJSDocTemplateTag(n)&&n.typeParameters.some(function(e){return e.name.escapedText===t})})}function W(n){return!!te(n,e.isJSDocParameterTag)}function q(n){return te(n,e.isJSDocAugmentsTag)}function z(n){return te(n,e.isJSDocClassTag)}function J(n){return te(n,e.isJSDocEnumTag)}function X(n){return te(n,e.isJSDocThisTag)}function Y(n){return te(n,e.isJSDocReturnTag)}function Q(n){return te(n,e.isJSDocTemplateTag)}function Z(n){var t=te(n,e.isJSDocTypeTag);if(t&&t.typeExpression&&t.typeExpression.type)return t}function $(n){var t=te(n,e.isJSDocTypeTag);return!t&&e.isParameter(n)&&(t=e.find(U(n),function(e){return!!e.typeExpression})),t&&t.typeExpression&&t.typeExpression.type}function ee(n){var t=Y(n);if(t&&t.typeExpression)return t.typeExpression.type;var r=Z(n);if(r&&r.typeExpression){var a=r.typeExpression.type;if(e.isTypeLiteralNode(a)){var i=e.find(a.members,e.isCallSignatureDeclaration);return i&&i.type}if(e.isFunctionTypeNode(a))return a.type}}function ne(n){var t=n.jsDocCache;if(void 0===t){var r=e.getJSDocCommentsAndTags(n);e.Debug.assert(r.length<2||r[0]!==r[1]),n.jsDocCache=t=e.flatMap(r,function(n){return e.isJSDoc(n)?n.tags:n})}return t}function te(n,t){return e.find(ne(n),t)}function re(e,n){return ne(e).filter(function(e){return e.kind===n})}function ae(n){if(e.isJSDocSignature(n))return e.emptyArray;if(e.isJSDocTypeAlias(n))return e.Debug.assert(291===n.parent.kind),e.flatMap(n.parent.tags,function(n){return e.isJSDocTemplateTag(n)?n.typeParameters:void 0});if(n.typeParameters)return n.typeParameters;if(e.isInJSFile(n)){var t=e.getJSDocTypeParameterDeclarations(n);if(t.length)return t;var r=$(n);if(r&&e.isFunctionTypeNode(r)&&r.typeParameters)return r.typeParameters}return e.emptyArray}function ie(n){return n.constraint?n.constraint:e.isJSDocTemplateTag(n.parent)&&n===n.parent.typeParameters[0]?n.parent.constraint:void 0}e.getDefaultLibFileName=n,e.textSpanEnd=t,e.textSpanIsEmpty=r,e.textSpanContainsPosition=a,e.textRangeContainsPositionInclusive=i,e.textSpanContainsTextSpan=o,e.textSpanOverlapsWith=s,e.textSpanOverlap=c,e.textSpanIntersectsWithTextSpan=l,e.textSpanIntersectsWith=u,e.decodedTextSpanIntersectsWith=d,e.textSpanIntersectsWithPosition=p,e.textSpanIntersection=m,e.createTextSpan=f,e.createTextSpanFromBounds=_,e.textChangeRangeNewSpan=g,e.textChangeRangeIsUnchanged=v,e.createTextChangeRange=y,e.unchangedTextChangeRange=y(f(0,0),0),e.collapseTextChangeRangesAcrossMultipleVersions=h,e.getTypeParameterOwner=b,e.isParameterPropertyDeclaration=E,e.isEmptyBindingPattern=T,e.isEmptyBindingElement=S,e.walkUpBindingElementsAndPatterns=x,e.getCombinedModifierFlags=A,e.getCombinedNodeFlags=D,e.validateLocaleAndSetLanguage=L,e.getOriginalNode=k,e.isParseTreeNode=N,e.getParseTreeNode=I,e.escapeLeadingUnderscores=w,e.unescapeLeadingUnderscores=M,e.idText=O,e.symbolName=P,e.getNameOfJSDocTypedef=G,e.isNamedDeclaration=V,e.getNonAssignedNameOfDeclaration=B,e.getNameOfDeclaration=K,e.getJSDocParameterTags=U,e.getJSDocTypeParameterTags=H,e.hasJSDocParameterTags=W,e.getJSDocAugmentsTag=q,e.getJSDocClassTag=z,e.getJSDocEnumTag=J,e.getJSDocThisTag=X,e.getJSDocReturnTag=Y,e.getJSDocTemplateTag=Q,e.getJSDocTypeTag=Z,e.getJSDocType=$,e.getJSDocReturnType=ee,e.getJSDocTags=ne,e.getAllJSDocTagsOfKind=re,e.getEffectiveTypeParameterDeclarations=ae,e.getEffectiveConstraintOfTypeParameter=ie}(u||(u={})),function(e){function n(e){return 8===e.kind}function t(e){return 9===e.kind}function r(e){return 10===e.kind}function a(e){return 11===e.kind}function i(e){return 13===e.kind}function o(e){return 14===e.kind}function s(e){return 15===e.kind}function c(e){return 16===e.kind}function l(e){return 17===e.kind}function u(e){return 72===e.kind}function d(e){return 148===e.kind}function p(e){return 149===e.kind}function m(e){return 150===e.kind}function f(e){return 151===e.kind}function _(e){return 152===e.kind}function g(e){return 153===e.kind}function v(e){return 154===e.kind}function y(e){return 155===e.kind}function h(e){return 156===e.kind}function b(e){return 157===e.kind}function E(e){return 158===e.kind}function T(e){return 159===e.kind}function S(e){return 160===e.kind}function x(e){return 161===e.kind}function C(e){return 162===e.kind}function A(e){return 159===e.kind||158===e.kind}function D(e){return 163===e.kind}function L(e){return 164===e.kind}function k(e){return 165===e.kind}function N(e){return 166===e.kind}function I(e){return 167===e.kind}function w(e){return 168===e.kind}function M(e){return 169===e.kind}function O(e){return 170===e.kind}function P(e){return 173===e.kind}function R(e){return 174===e.kind}function F(e){return 175===e.kind}function G(e){return 176===e.kind}function V(e){return 177===e.kind}function B(e){return 178===e.kind}function K(e){return 179===e.kind}function j(e){return 180===e.kind}function U(e){return 181===e.kind}function H(e){return 182===e.kind}function W(e){return 183===e.kind}function q(e){return 184===e.kind}function z(e){return 185===e.kind}function J(e){return 186===e.kind}function X(e){return 187===e.kind}function Y(e){return 188===e.kind}function Q(e){return 189===e.kind}function Z(e){return 190===e.kind}function $(e){return 191===e.kind}function ee(e){return 192===e.kind}function ne(e){return 193===e.kind}function te(e){return 194===e.kind}function re(e){return 195===e.kind}function ae(e){while(308===e.kind)e=e.expression;return e}function ie(e){return 196===e.kind}function oe(e){return 197===e.kind}function se(e){return 198===e.kind}function ce(e){return 199===e.kind}function le(e){return 200===e.kind}function ue(e){return 201===e.kind}function de(e){return 202===e.kind}function pe(e){return 203===e.kind}function me(e){return 204===e.kind}function fe(e){return 205===e.kind}function _e(e){return 206===e.kind}function ge(e){return 207===e.kind}function ve(e){return 208===e.kind}function ye(e){return 209===e.kind}function he(e){return 210===e.kind}function be(e){return 211===e.kind}function Ee(e){return 212===e.kind}function Te(e){return 213===e.kind}function Se(e){return 214===e.kind}function xe(e){return 216===e.kind}function Ce(e){return 217===e.kind}function Ae(e){return 218===e.kind}function De(e){return 219===e.kind}function Le(e){return 220===e.kind}function ke(e){return 221===e.kind}function Ne(e){return 222===e.kind}function Ie(e){return 223===e.kind}function we(e){return 224===e.kind}function Me(e){return 225===e.kind}function Oe(e){return 226===e.kind}function Pe(e){return 227===e.kind}function Re(e){return 228===e.kind}function Fe(e){return 229===e.kind}function Ge(e){return 229===e.kind||228===e.kind}function Ve(e){return 230===e.kind}function Be(e){return 231===e.kind}function Ke(e){return 232===e.kind}function je(e){return 233===e.kind}function Ue(e){return 234===e.kind}function He(e){return 235===e.kind}function We(e){return 236===e.kind}function qe(e){return 237===e.kind}function ze(e){return 238===e.kind}function Je(e){return 239===e.kind}function Xe(e){return 240===e.kind}function Ye(e){return 241===e.kind}function Qe(e){return 242===e.kind}function Ze(e){return 243===e.kind}function $e(e){return 244===e.kind}function en(e){return 245===e.kind}function nn(e){return 246===e.kind}function tn(e){return 247===e.kind}function rn(e){return 248===e.kind}function an(e){return 249===e.kind}function on(e){return 250===e.kind}function sn(e){return 251===e.kind}function cn(e){return 252===e.kind}function ln(e){return 253===e.kind}function un(e){return 254===e.kind}function dn(e){return 255===e.kind}function pn(e){return 256===e.kind}function mn(e){return 257===e.kind}function fn(e){return 258===e.kind}function _n(e){return 259===e.kind}function gn(e){return 260===e.kind}function vn(e){return 261===e.kind}function yn(e){return 262===e.kind}function hn(e){return 263===e.kind}function bn(e){return 264===e.kind}function En(e){return 265===e.kind}function Tn(e){return 266===e.kind}function Sn(e){return 267===e.kind}function xn(e){return 268===e.kind}function Cn(e){return 269===e.kind}function An(e){return 270===e.kind}function Dn(e){return 271===e.kind}function Ln(e){return 272===e.kind}function kn(e){return 273===e.kind}function Nn(e){return 274===e.kind}function In(e){return 275===e.kind}function wn(e){return 276===e.kind}function Mn(e){return 277===e.kind}function On(e){return 278===e.kind}function Pn(e){return 279===e.kind}function Rn(e){return 280===e.kind}function Fn(e){return 281===e.kind}function Gn(e){return 283===e.kind}function Vn(e){return 284===e.kind}function Bn(e){return 285===e.kind}function Kn(e){return 286===e.kind}function jn(e){return 287===e.kind}function Un(e){return 288===e.kind}function Hn(e){return 289===e.kind}function Wn(e){return 290===e.kind}function qn(e){return 291===e.kind}function zn(e){return 295===e.kind}function Jn(e){return 296===e.kind}function Xn(e){return 298===e.kind}function Yn(e){return 301===e.kind}function Qn(e){return 299===e.kind}function Zn(e){return 300===e.kind}function $n(e){return 302===e.kind}function et(e){return 303===e.kind}function nt(e){return 304===e.kind}function tt(e){return 305===e.kind}function rt(e){return 305===e.kind||299===e.kind}function at(e){return 292===e.kind}function it(e){return 297===e.kind}function ot(e){return 293===e.kind}e.isNumericLiteral=n,e.isBigIntLiteral=t,e.isStringLiteral=r,e.isJsxText=a,e.isRegularExpressionLiteral=i,e.isNoSubstitutionTemplateLiteral=o,e.isTemplateHead=s,e.isTemplateMiddle=c,e.isTemplateTail=l,e.isIdentifier=u,e.isQualifiedName=d,e.isComputedPropertyName=p,e.isTypeParameterDeclaration=m,e.isParameter=f,e.isDecorator=_,e.isPropertySignature=g,e.isPropertyDeclaration=v,e.isMethodSignature=y,e.isMethodDeclaration=h,e.isConstructorDeclaration=b,e.isGetAccessorDeclaration=E,e.isSetAccessorDeclaration=T,e.isCallSignatureDeclaration=S,e.isConstructSignatureDeclaration=x,e.isIndexSignatureDeclaration=C,e.isGetOrSetAccessorDeclaration=A,e.isTypePredicateNode=D,e.isTypeReferenceNode=L,e.isFunctionTypeNode=k,e.isConstructorTypeNode=N,e.isTypeQueryNode=I,e.isTypeLiteralNode=w,e.isArrayTypeNode=M,e.isTupleTypeNode=O,e.isUnionTypeNode=P,e.isIntersectionTypeNode=R,e.isConditionalTypeNode=F,e.isInferTypeNode=G,e.isParenthesizedTypeNode=V,e.isThisTypeNode=B,e.isTypeOperatorNode=K,e.isIndexedAccessTypeNode=j,e.isMappedTypeNode=U,e.isLiteralTypeNode=H,e.isImportTypeNode=W,e.isObjectBindingPattern=q,e.isArrayBindingPattern=z,e.isBindingElement=J,e.isArrayLiteralExpression=X,e.isObjectLiteralExpression=Y,e.isPropertyAccessExpression=Q,e.isElementAccessExpression=Z,e.isCallExpression=$,e.isNewExpression=ee,e.isTaggedTemplateExpression=ne,e.isTypeAssertion=te,e.isParenthesizedExpression=re,e.skipPartiallyEmittedExpressions=ae,e.isFunctionExpression=ie,e.isArrowFunction=oe,e.isDeleteExpression=se,e.isTypeOfExpression=ce,e.isVoidExpression=le,e.isAwaitExpression=ue,e.isPrefixUnaryExpression=de,e.isPostfixUnaryExpression=pe,e.isBinaryExpression=me,e.isConditionalExpression=fe,e.isTemplateExpression=_e,e.isYieldExpression=ge,e.isSpreadElement=ve,e.isClassExpression=ye,e.isOmittedExpression=he,e.isExpressionWithTypeArguments=be,e.isAsExpression=Ee,e.isNonNullExpression=Te,e.isMetaProperty=Se,e.isTemplateSpan=xe,e.isSemicolonClassElement=Ce,e.isBlock=Ae,e.isVariableStatement=De,e.isEmptyStatement=Le,e.isExpressionStatement=ke,e.isIfStatement=Ne,e.isDoStatement=Ie,e.isWhileStatement=we,e.isForStatement=Me,e.isForInStatement=Oe,e.isForOfStatement=Pe,e.isContinueStatement=Re,e.isBreakStatement=Fe,e.isBreakOrContinueStatement=Ge,e.isReturnStatement=Ve,e.isWithStatement=Be,e.isSwitchStatement=Ke,e.isLabeledStatement=je,e.isThrowStatement=Ue,e.isTryStatement=He,e.isDebuggerStatement=We,e.isVariableDeclaration=qe,e.isVariableDeclarationList=ze,e.isFunctionDeclaration=Je,e.isClassDeclaration=Xe,e.isInterfaceDeclaration=Ye,e.isTypeAliasDeclaration=Qe,e.isEnumDeclaration=Ze,e.isModuleDeclaration=$e,e.isModuleBlock=en,e.isCaseBlock=nn,e.isNamespaceExportDeclaration=tn,e.isImportEqualsDeclaration=rn,e.isImportDeclaration=an,e.isImportClause=on,e.isNamespaceImport=sn,e.isNamedImports=cn,e.isImportSpecifier=ln,e.isExportAssignment=un,e.isExportDeclaration=dn,e.isNamedExports=pn,e.isExportSpecifier=mn,e.isMissingDeclaration=fn,e.isExternalModuleReference=_n,e.isJsxElement=gn,e.isJsxSelfClosingElement=vn,e.isJsxOpeningElement=yn,e.isJsxClosingElement=hn,e.isJsxFragment=bn,e.isJsxOpeningFragment=En,e.isJsxClosingFragment=Tn,e.isJsxAttribute=Sn,e.isJsxAttributes=xn,e.isJsxSpreadAttribute=Cn,e.isJsxExpression=An,e.isCaseClause=Dn,e.isDefaultClause=Ln,e.isHeritageClause=kn,e.isCatchClause=Nn,e.isPropertyAssignment=In,e.isShorthandPropertyAssignment=wn,e.isSpreadAssignment=Mn,e.isEnumMember=On,e.isSourceFile=Pn,e.isBundle=Rn,e.isUnparsedSource=Fn,e.isJSDocTypeExpression=Gn,e.isJSDocAllType=Vn,e.isJSDocUnknownType=Bn,e.isJSDocNullableType=Kn,e.isJSDocNonNullableType=jn,e.isJSDocOptionalType=Un,e.isJSDocFunctionType=Hn,e.isJSDocVariadicType=Wn,e.isJSDoc=qn,e.isJSDocAugmentsTag=zn,e.isJSDocClassTag=Jn,e.isJSDocEnumTag=Xn,e.isJSDocThisTag=Yn,e.isJSDocParameterTag=Qn,e.isJSDocReturnTag=Zn,e.isJSDocTypeTag=$n,e.isJSDocTemplateTag=et,e.isJSDocTypedefTag=nt,e.isJSDocPropertyTag=tt,e.isJSDocPropertyLikeTag=rt,e.isJSDocTypeLiteral=at,e.isJSDocCallbackTag=it,e.isJSDocSignature=ot}(u||(u={})),function(e){function n(e){return 306===e.kind}function t(e){return r(e.kind)}function r(e){return e>=148}function a(e){return e.kind>=0&&e.kind<=147}function i(e){return e.hasOwnProperty("pos")&&e.hasOwnProperty("end")}function o(e){return 8<=e&&e<=14}function s(e){return o(e.kind)}function c(e){return 14<=e&&e<=17}function l(e){return c(e.kind)}function u(e){var n=e.kind;return 16===n||17===n}function d(n){return e.isImportSpecifier(n)||e.isExportSpecifier(n)}function p(e){return 10===e.kind||c(e.kind)}function m(n){return e.isIdentifier(n)&&(7&n.autoGenerateFlags)>0}function f(e){switch(e){case 118:case 121:case 77:case 125:case 80:case 85:case 115:case 113:case 114:case 133:case 116:return!0}return!1}function _(n){return!!(92&e.modifierToFlag(n))}function g(e){return _(e)||116===e}function v(e){return f(e.kind)}function y(e){var n=e.kind;return 148===n||72===n}function h(e){var n=e.kind;return 72===n||10===n||8===n||149===n}function b(e){var n=e.kind;return 72===n||184===n||185===n}function E(e){return e&&x(e.kind)}function T(e){return e&&S(e.kind)}function S(e){switch(e){case 239:case 156:case 157:case 158:case 159:case 196:case 197:return!0;default:return!1}}function x(e){switch(e){case 155:case 160:case 293:case 161:case 162:case 165:case 289:case 166:return!0;default:return S(e)}}function C(n){return e.isSourceFile(n)||e.isModuleBlock(n)||e.isBlock(n)&&E(n.parent)}function A(e){var n=e.kind;return 157===n||154===n||156===n||158===n||159===n||162===n||217===n}function D(e){return e&&(240===e.kind||209===e.kind)}function L(e){return e&&(158===e.kind||159===e.kind)}function k(e){switch(e.kind){case 156:case 158:case 159:return!0;default:return!1}}function N(e){var n=e.kind;return 161===n||160===n||153===n||155===n||162===n}function I(e){return N(e)||A(e)}function w(e){var n=e.kind;return 275===n||276===n||277===n||156===n||158===n||159===n}function M(n){return e.isTypeNodeKind(n.kind)}function O(e){switch(e.kind){case 165:case 166:return!0}return!1}function P(e){if(e){var n=e.kind;return 185===n||184===n}return!1}function R(e){var n=e.kind;return 187===n||188===n}function F(e){var n=e.kind;return 186===n||210===n}function G(e){switch(e.kind){case 237:case 151:case 186:return!0}return!1}function V(e){return B(e)||K(e)}function B(e){switch(e.kind){case 184:case 188:return!0}return!1}function K(e){switch(e.kind){case 185:case 187:return!0}return!1}function j(e){var n=e.kind;return 189===n||148===n||183===n}function U(e){var n=e.kind;return 189===n||148===n}function H(e){switch(e.kind){case 262:case 261:case 191:case 192:case 193:case 152:return!0;default:return!1}}function W(e){return 191===e.kind||192===e.kind}function q(e){var n=e.kind;return 206===n||14===n}function z(n){return J(e.skipPartiallyEmittedExpressions(n).kind)}function J(e){switch(e){case 189:case 190:case 192:case 191:case 260:case 261:case 264:case 193:case 187:case 195:case 188:case 209:case 196:case 72:case 13:case 8:case 9:case 10:case 14:case 206:case 87:case 96:case 100:case 102:case 98:case 213:case 214:case 92:return!0;default:return!1}}function X(n){return Y(e.skipPartiallyEmittedExpressions(n).kind)}function Y(e){switch(e){case 202:case 203:case 198:case 199:case 200:case 201:case 194:return!0;default:return J(e)}}function Q(e){switch(e.kind){case 203:return!0;case 202:return 44===e.operator||45===e.operator;default:return!1}}function Z(n){return $(e.skipPartiallyEmittedExpressions(n).kind)}function $(e){switch(e){case 205:case 207:case 197:case 204:case 208:case 212:case 210:case 309:case 308:return!0;default:return Y(e)}}function ee(e){var n=e.kind;return 194===n||212===n}function ne(e){return 308===e.kind}function te(e){return 307===e.kind}function re(e){return te(e)||ne(e)}function ae(e,n){switch(e.kind){case 225:case 226:case 227:case 223:case 224:return!0;case 233:return n&&ae(e.statement,n)}return!1}function ie(e){return 226===e.kind||227===e.kind}function oe(n){return e.isBlock(n)||Z(n)}function se(n){return e.isBlock(n)}function ce(n){return e.isVariableDeclarationList(n)||Z(n)}function le(e){var n=e.kind;return 245===n||244===n||72===n}function ue(e){var n=e.kind;return 245===n||244===n}function de(e){var n=e.kind;return 72===n||244===n}function pe(e){var n=e.kind;return 252===n||251===n}function me(e){return 244===e.kind||243===e.kind}function fe(e){return 197===e||186===e||240===e||209===e||157===e||243===e||278===e||257===e||239===e||196===e||158===e||250===e||248===e||253===e||241===e||267===e||156===e||155===e||244===e||247===e||251===e||151===e||275===e||154===e||153===e||159===e||276===e||242===e||150===e||237===e||304===e||297===e||305===e}function _e(e){return 239===e||258===e||240===e||241===e||242===e||243===e||244===e||249===e||248===e||255===e||254===e||247===e}function ge(e){return 229===e||228===e||236===e||223===e||221===e||220===e||226===e||227===e||225===e||222===e||233===e||230===e||232===e||234===e||235===e||219===e||224===e||231===e||307===e||311===e||310===e}function ve(n){return 150===n.kind?303!==n.parent.kind||e.isInJSFile(n):fe(n.kind)}function ye(e){return _e(e.kind)}function he(e){return ge(e.kind)}function be(e){var n=e.kind;return ge(n)||_e(n)||Ee(e)}function Ee(n){return 218===n.kind&&((void 0===n.parent||235!==n.parent.kind&&274!==n.parent.kind)&&!e.isFunctionBlock(n))}function Te(e){var n=e.kind;return 259===n||148===n||72===n}function Se(e){var n=e.kind;return 100===n||72===n||189===n}function xe(e){var n=e.kind;return 260===n||270===n||261===n||11===n||264===n}function Ce(e){var n=e.kind;return 267===n||269===n}function Ae(e){var n=e.kind;return 10===n||270===n}function De(e){var n=e.kind;return 262===n||261===n}function Le(e){var n=e.kind;return 271===n||272===n}function ke(e){return e.kind>=283&&e.kind<=305}function Ne(n){return 291===n.kind||Ie(n)||e.isJSDocTypeLiteral(n)||e.isJSDocSignature(n)}function Ie(e){return e.kind>=294&&e.kind<=305}function we(e){return 159===e.kind}function Me(e){return 158===e.kind}function Oe(e){var n=e.jsDoc;return!!n&&n.length>0}function Pe(e){return!!e.type}function Re(e){return!!e.initializer}function Fe(n){return Re(n)&&!e.isForStatement(n)&&!e.isForInStatement(n)&&!e.isForOfStatement(n)&&!e.isJsxAttribute(n)}function Ge(e){return 267===e.kind||269===e.kind||w(e)}function Ve(e){return 164===e.kind||211===e.kind}e.isSyntaxList=n,e.isNode=t,e.isNodeKind=r,e.isToken=a,e.isNodeArray=i,e.isLiteralKind=o,e.isLiteralExpression=s,e.isTemplateLiteralKind=c,e.isTemplateLiteralToken=l,e.isTemplateMiddleOrTemplateTail=u,e.isImportOrExportSpecifier=d,e.isStringTextContainingNode=p,e.isGeneratedIdentifier=m,e.isModifierKind=f,e.isParameterPropertyModifier=_,e.isClassMemberModifier=g,e.isModifier=v,e.isEntityName=y,e.isPropertyName=h,e.isBindingName=b,e.isFunctionLike=E,e.isFunctionLikeDeclaration=T,e.isFunctionLikeKind=x,e.isFunctionOrModuleBlock=C,e.isClassElement=A,e.isClassLike=D,e.isAccessor=L,e.isMethodOrAccessor=k,e.isTypeElement=N,e.isClassOrTypeElement=I,e.isObjectLiteralElementLike=w,e.isTypeNode=M,e.isFunctionOrConstructorTypeNode=O,e.isBindingPattern=P,e.isAssignmentPattern=R,e.isArrayBindingElement=F,e.isDeclarationBindingElement=G,e.isBindingOrAssignmentPattern=V,e.isObjectBindingOrAssignmentPattern=B,e.isArrayBindingOrAssignmentPattern=K,e.isPropertyAccessOrQualifiedNameOrImportTypeNode=j,e.isPropertyAccessOrQualifiedName=U,e.isCallLikeExpression=H,e.isCallOrNewExpression=W,e.isTemplateLiteral=q,e.isLeftHandSideExpression=z,e.isUnaryExpression=X,e.isUnaryExpressionWithWrite=Q,e.isExpression=Z,e.isAssertionExpression=ee,e.isPartiallyEmittedExpression=ne,e.isNotEmittedStatement=te,e.isNotEmittedOrPartiallyEmittedNode=re,e.isIterationStatement=ae,e.isForInOrOfStatement=ie,e.isConciseBody=oe,e.isFunctionBody=se,e.isForInitializer=ce,e.isModuleBody=le,e.isNamespaceBody=ue,e.isJSDocNamespaceBody=de,e.isNamedImportBindings=pe,e.isModuleOrEnumDeclaration=me,e.isDeclaration=ve,e.isDeclarationStatement=ye,e.isStatementButNotDeclaration=he,e.isStatement=be,e.isModuleReference=Te,e.isJsxTagNameExpression=Se,e.isJsxChild=xe,e.isJsxAttributeLike=Ce,e.isStringLiteralOrJsxExpression=Ae,e.isJsxOpeningLikeElement=De,e.isCaseOrDefaultClause=Le,e.isJSDocNode=ke,e.isJSDocCommentContainingNode=Ne,e.isJSDocTag=Ie,e.isSetAccessor=we,e.isGetAccessor=Me,e.hasJSDocNodes=Oe,e.hasType=Pe,e.hasInitializer=Re,e.hasOnlyExpressionInitializer=Fe,e.isObjectLiteralElement=Ge,e.isTypeReferenceType=Ve;var Be=1073741823;function Ke(n){for(var t=Be,r=0,a=n;r<a.length;r++){var i=a[r];if(i.length){for(var o=0;o<i.length&&o<t;o++)if(!e.isWhiteSpaceLike(i.charCodeAt(o)))break;if(o<t&&(t=o),0===t)return 0}}return t===Be?void 0:t}function je(e){return 10===e.kind||14===e.kind}e.guessIndentation=Ke,e.isStringLiteralLike=je}(u||(u={})),function(e){function n(e){return 252===e.kind||256===e.kind}function t(e,n){this.flags=e,this.escapedName=n,this.declarations=void 0,this.valueDeclaration=void 0,this.id=void 0,this.mergeId=void 0,this.parent=void 0}function r(n,t){this.flags=t,e.Debug.isDebugging&&(this.checker=n)}function a(){}function i(e,n,t){this.pos=n,this.end=t,this.kind=e,this.id=0,this.flags=0,this.modifierFlagsCache=0,this.transformFlags=0,this.parent=void 0,this.original=void 0}function o(e,n,t){this.fileName=e,this.text=n,this.skipTrivia=t||function(e){return e}}function s(n,t,r){return void 0===r&&(r=0),n.replace(/{(\d+)}/g,function(n,a){return""+e.Debug.assertDefined(t[+a+r])})}function c(n){return e.localizedDiagnosticMessages&&e.localizedDiagnosticMessages[n.key]||n.message}function l(n,t,r,a){e.Debug.assertGreaterThanOrEqual(t,0),e.Debug.assertGreaterThanOrEqual(r,0),n&&(e.Debug.assertLessThanOrEqual(t,n.text.length),e.Debug.assertLessThanOrEqual(t+r,n.text.length));var i=c(a);return arguments.length>4&&(i=s(i,arguments,4)),{file:n,start:t,length:r,messageText:i,category:a.category,code:a.code,reportsUnnecessary:a.reportsUnnecessary}}function u(e,n){var t=c(n);return arguments.length>2&&(t=s(t,arguments,2)),t}function d(e){var n=c(e);return arguments.length>1&&(n=s(n,arguments,1)),{file:void 0,start:void 0,length:void 0,messageText:n,category:e.category,code:e.code,reportsUnnecessary:e.reportsUnnecessary}}function p(e){return{file:void 0,start:void 0,length:void 0,code:e.code,category:e.category,messageText:e.next?e:e.messageText}}function m(e,n){var t=c(n);return arguments.length>2&&(t=s(t,arguments,2)),{messageText:t,category:n.category,code:n.code,next:e}}function f(e,n){var t=e;while(t.next)t=t.next;return t.next=n,e}function _(e){return e.file?e.file.path:void 0}function g(e,n){return v(e,n)||y(e,n)||0}function v(n,t){return e.compareStringsCaseSensitive(_(n),_(t))||e.compareValues(n.start,t.start)||e.compareValues(n.length,t.length)||e.compareValues(n.code,t.code)||h(n.messageText,t.messageText)||0}function y(n,t){return n.relatedInformation||t.relatedInformation?n.relatedInformation&&t.relatedInformation?e.compareValues(n.relatedInformation.length,t.relatedInformation.length)||e.forEach(n.relatedInformation,function(e,n){var r=t.relatedInformation[n];return g(e,r)})||0:n.relatedInformation?-1:1:0}function h(n,t){var r=n,a=t;while(r&&a){var i=e.isString(r)?r:r.messageText,o=e.isString(a)?a:a.messageText,s=e.compareStringsCaseSensitive(i,o);if(s)return s;r=e.isString(r)?void 0:r.next,a=e.isString(a)?void 0:a.next}return r||a?r?1:-1:0}function b(e){return e.target||0}function E(n){return"number"===typeof n.module?n.module:b(n)>=2?e.ModuleKind.ES2015:e.ModuleKind.CommonJS}function T(n){var t=n.moduleResolution;return void 0===t&&(t=E(n)===e.ModuleKind.CommonJS?e.ModuleResolutionKind.NodeJs:e.ModuleResolutionKind.Classic),t}function S(n){switch(E(n)){case e.ModuleKind.CommonJS:case e.ModuleKind.AMD:case e.ModuleKind.ES2015:case e.ModuleKind.ESNext:return!0;default:return!1}}function x(e){return!1===e.allowUnreachableCode}function C(e){return!1===e.allowUnusedLabels}function A(e){return!(!L(e)||!e.declarationMap)}function D(n){var t=E(n);return void 0!==n.allowSyntheticDefaultImports?n.allowSyntheticDefaultImports:n.esModuleInterop||t===e.ModuleKind.System}function L(e){return!(!e.declaration&&!e.composite)}function k(e,n){return void 0===e[n]?!!e.strict:!!e[n]}function N(n,t){return t!==n&&e.semanticDiagnosticsOptionDeclarations.some(function(r){return!e.isJsonEqual(I(t,r),I(n,r))})}function I(e,n){return n.strictFlag?k(e,n.name):e[n.name]}function w(e){for(var n=!1,t=0;t<e.length;t++)if(42===e.charCodeAt(t)){if(n)return!1;n=!0}return!0}e.isNamedImportsOrExports=n,e.objectAllocator={getNodeConstructor:function(){return i},getTokenConstructor:function(){return i},getIdentifierConstructor:function(){return i},getSourceFileConstructor:function(){return i},getSymbolConstructor:function(){return t},getTypeConstructor:function(){return r},getSignatureConstructor:function(){return a},getSourceMapSourceConstructor:function(){return o}},e.formatStringFromArgs=s,e.getLocaleSpecificMessage=c,e.createFileDiagnostic=l,e.formatMessage=u,e.createCompilerDiagnostic=d,e.createCompilerDiagnosticFromMessageChain=p,e.chainDiagnosticMessages=m,e.concatenateDiagnosticMessageChains=f,e.compareDiagnostics=g,e.compareDiagnosticsSkipRelatedInformation=v,e.getEmitScriptTarget=b,e.getEmitModuleKind=E,e.getEmitModuleResolutionKind=T,e.hasJsonModuleEmitEnabled=S,e.unreachableCodeIsError=x,e.unusedLabelIsError=C,e.getAreDeclarationMapsEnabled=A,e.getAllowSyntheticDefaultImports=D,e.getEmitDeclarations=L,e.getStrictOptionValue=k,e.compilerOptionsAffectSemanticDiagnostics=N,e.getCompilerOptionValue=I,e.hasZeroOrOneAsteriskCharacter=w,e.directorySeparator="/";var M="\\",O="://",P=/\\/g;function R(n){return n.replace(P,e.directorySeparator)}function F(e){return e>=97&&e<=122||e>=65&&e<=90}function G(e,n){var t=e.charCodeAt(n);if(58===t)return n+1;if(37===t&&51===e.charCodeAt(n+1)){var r=e.charCodeAt(n+2);if(97===r||65===r)return n+3}return-1}function V(n){if(!n)return 0;var t=n.charCodeAt(0);if(47===t||92===t){if(n.charCodeAt(1)!==t)return 1;var r=n.indexOf(47===t?e.directorySeparator:M,2);return r<0?n.length:r+1}if(F(t)&&58===n.charCodeAt(1)){var a=n.charCodeAt(2);if(47===a||92===a)return 3;if(2===n.length)return 2}var i=n.indexOf(O);if(-1!==i){var o=i+O.length,s=n.indexOf(e.directorySeparator,o);if(-1!==s){var c=n.slice(0,i),l=n.slice(o,s);if("file"===c&&(""===l||"localhost"===l)&&F(n.charCodeAt(s+1))){var u=G(n,s+2);if(-1!==u){if(47===n.charCodeAt(u))return~(u+1);if(u===n.length)return~u}}return~(s+1)}return~n.length}return 0}function B(e){var n=V(e);return n<0?~n:n}function K(n){return e.resolvePath(n)}function j(n){n=R(n);var t=Z(Q(n)),r=t[0],a=t.slice(1);if(a.length){var i=r+a.join(e.directorySeparator);return{path:e.hasTrailingDirectorySeparator(n)?e.ensureTrailingDirectorySeparator(i):i,parts:a}}return{path:r,parts:a}}function U(n){n=R(n);var t=B(n);return t===n.length?n:(n=e.removeTrailingDirectorySeparator(n),n.slice(0,Math.max(t,n.lastIndexOf(e.directorySeparator))))}function H(n,t,r){var a=r(n),i=r(t);return e.startsWith(a,i+"/")||e.startsWith(a,i+"\\")}function W(e){return V(e)<0}function q(e){return/^\.\.?($|[\\/])/.test(e)}function z(e){return V(e)>0}function J(e){var n=V(e);return n>0&&n===e.length}function X(n,t,r){return z(n)?e.getRelativePathToDirectoryOrUrl(t,n,t,r,!1):n}function Y(n,t){var r=n.substring(0,t),a=n.substring(t).split(e.directorySeparator);return a.length&&!e.lastOrUndefined(a)&&a.pop(),[r].concat(a)}function Q(n,t){void 0===t&&(t=""),n=e.combinePaths(t,n);var r=B(n);return Y(n,r)}function Z(n){if(!e.some(n))return[];for(var t=[n[0]],r=1;r<n.length;r++){var a=n[r];if(a&&"."!==a){if(".."===a)if(t.length>1){if(".."!==t[t.length-1]){t.pop();continue}}else if(t[0])continue;t.push(a)}}return t}function $(e,n){return Z(Q(e,n))}function ee(e,n){return ne($(e,n))}function ne(n){if(0===n.length)return"";var t=n[0]&&e.ensureTrailingDirectorySeparator(n[0]);return t+n.slice(1).join(e.directorySeparator)}e.normalizeSlashes=R,e.getRootLength=B,e.normalizePath=K,e.normalizePathAndParts=j,e.getDirectoryPath=U,e.startsWithDirectory=H,e.isUrl=W,e.pathIsRelative=q,e.isRootedDiskPath=z,e.isDiskPathRoot=J,e.convertToRelativePath=X,e.getPathComponents=Q,e.reducePathComponents=Z,e.getNormalizedPathComponents=$,e.getNormalizedAbsolutePath=ee,e.getPathFromPathComponents=ne}(u||(u={})),function(e){function n(n,t,r,a){var i,o=e.reducePathComponents(e.getPathComponents(n)),s=e.reducePathComponents(e.getPathComponents(t));for(i=0;i<o.length&&i<s.length;i++){var c=a(o[i]),l=a(s[i]),u=0===i?e.equateStringsCaseInsensitive:r;if(!u(c,l))break}if(0===i)return s;for(var d=s.slice(i),p=[];i<o.length;i++)p.push("..");return[""].concat(p,d)}function t(n,t,a){return i(r(e.getDirectoryPath(n),t,a))}function r(t,r,a){ne.assert(e.getRootLength(t)>0===e.getRootLength(r)>0,"Paths must either both be absolute or both be relative");var i="function"===typeof a?a:e.identity,o="boolean"===typeof a&&a,s=n(t,r,o?e.equateStringsCaseInsensitive:e.equateStringsCaseSensitive,i);return e.getPathFromPathComponents(s)}function a(t,r,a,i,o){var s=n(c(a,t),c(a,r),e.equateStringsCaseSensitive,i),l=s[0];if(o&&e.isRootedDiskPath(l)){var u=l.charAt(0)===e.directorySeparator?"file://":"file:///";s[0]=u+l}return e.getPathFromPathComponents(s)}function i(n){return 0!==e.getRootLength(n)||e.pathIsRelative(n)?n:"./"+n}function o(n,t,r){n=e.normalizeSlashes(n);var a=e.getRootLength(n);if(a===n.length)return"";n=u(n);var i=n.slice(Math.max(e.getRootLength(n),n.lastIndexOf(e.directorySeparator)+1)),o=void 0!==t&&void 0!==r?ge(i,t,r):void 0;return o?i.slice(0,i.length-o.length):i}function s(n){for(var t=[],r=1;r<arguments.length;r++)t[r-1]=arguments[r];n&&(n=e.normalizeSlashes(n));for(var a=0,i=t;a<i.length;a++){var o=i[a];o&&(o=e.normalizeSlashes(o),n=n&&0===e.getRootLength(o)?d(n)+o:o)}return n}function c(n){for(var t=[],r=1;r<arguments.length;r++)t[r-1]=arguments[r];var a=e.some(t)?s.apply(void 0,[n].concat(t)):e.normalizeSlashes(n),i=e.getPathFromPathComponents(e.reducePathComponents(e.getPathComponents(a)));return i&&l(a)?d(i):i}function l(e){if(0===e.length)return!1;var n=e.charCodeAt(e.length-1);return 47===n||92===n}function u(e){return l(e)?e.substr(0,e.length-1):e}function d(n){return l(n)?n:n+e.directorySeparator}e.getPathComponentsRelativeTo=n,e.getRelativePathFromFile=t,e.getRelativePathFromDirectory=r,e.getRelativePathToDirectoryOrUrl=a,e.ensurePathIsNonModuleName=i,e.getBaseFileName=o,e.combinePaths=s,e.resolvePath=c,e.hasTrailingDirectorySeparator=l,e.removeTrailingDirectorySeparator=u,e.ensureTrailingDirectorySeparator=d;var p=/(^|\/)\.{0,2}($|\/)/;function m(n,t,r){if(n===t)return 0;if(void 0===n)return-1;if(void 0===t)return 1;var a=n.substring(0,e.getRootLength(n)),i=t.substring(0,e.getRootLength(t)),o=e.compareStringsCaseInsensitive(a,i);if(0!==o)return o;var s=n.substring(a.length),c=t.substring(i.length);if(!p.test(s)&&!p.test(c))return r(s,c);for(var l=e.reducePathComponents(e.getPathComponents(n)),u=e.reducePathComponents(e.getPathComponents(t)),d=Math.min(l.length,u.length),m=1;m<d;m++){var f=r(l[m],u[m]);if(0!==f)return f}return e.compareValues(l.length,u.length)}function f(n,t){return m(n,t,e.compareStringsCaseSensitive)}function _(n,t){return m(n,t,e.compareStringsCaseInsensitive)}function g(n,t,r,a){return"string"===typeof r?(n=s(r,n),t=s(r,t)):"boolean"===typeof r&&(a=r),m(n,t,e.getStringComparer(a))}function v(n,t,r,a){if("string"===typeof r?(n=s(r,n),t=s(r,t)):"boolean"===typeof r&&(a=r),void 0===n||void 0===t)return!1;if(n===t)return!0;var i=e.reducePathComponents(e.getPathComponents(n)),o=e.reducePathComponents(e.getPathComponents(t));if(o.length<i.length)return!1;for(var c=a?e.equateStringsCaseInsensitive:e.equateStringsCaseSensitive,l=0;l<i.length;l++){var u=0===l?e.equateStringsCaseInsensitive:c;if(!u(i[l],o[l]))return!1}return!0}function y(e){return 47===e||92===e}function h(e){return y(e.charCodeAt(0))?e.slice(1):void 0}function b(n,t,r){var a=e.tryRemovePrefix(n,t,r);return void 0===a?void 0:h(a)}e.comparePathsCaseSensitive=f,e.comparePathsCaseInsensitive=_,e.comparePaths=g,e.containsPath=v,e.tryRemoveDirectoryPrefix=b;var E=/[^\w\s\/]/g;function T(e){return e.replace(E,S)}function S(e){return"\\"+e}e.regExpEscape=T;var x=[42,63];function C(n){return e.stringContains(o(n),".")}e.hasExtension=C,e.commonPackageFolders=["node_modules","bower_components","jspm_packages"];var A="(?!("+e.commonPackageFolders.join("|")+")(/|$))",D={singleAsteriskRegexFragment:"([^./]|(\\.(?!min\\.js$))?)*",doubleAsteriskRegexFragment:"(/"+A+"[^/.][^/]*)*?",replaceWildcardCharacter:function(e){return P(e,D.singleAsteriskRegexFragment)}},L={singleAsteriskRegexFragment:"[^/]*",doubleAsteriskRegexFragment:"(/"+A+"[^/.][^/]*)*?",replaceWildcardCharacter:function(e){return P(e,L.singleAsteriskRegexFragment)}},k={singleAsteriskRegexFragment:"[^/]*",doubleAsteriskRegexFragment:"(/.+?)?",replaceWildcardCharacter:function(e){return P(e,k.singleAsteriskRegexFragment)}},N={files:D,directories:L,exclude:k};function I(e,n,t){var r=w(e,n,t);if(r&&r.length){var a=r.map(function(e){return"("+e+")"}).join("|"),i="exclude"===t?"($|/)":"$";return"^("+a+")"+i}}function w(n,t,r){if(void 0!==n&&0!==n.length)return e.flatMap(n,function(e){return e&&O(e,t,r,N[r])})}function M(e){return!/[.*?]/.test(e)}function O(n,t,r,a){var i=a.singleAsteriskRegexFragment,o=a.doubleAsteriskRegexFragment,s=a.replaceWildcardCharacter,c="",l=!1,d=e.getNormalizedPathComponents(n,t),p=e.last(d);if("exclude"===r||"**"!==p){d[0]=u(d[0]),M(p)&&d.push("**","*");for(var m=0,f=0,_=d;f<_.length;f++){var g=_[f];if("**"===g)c+=o;else if("directories"===r&&(c+="(",m++),l&&(c+=e.directorySeparator),"exclude"!==r){var v="";42===g.charCodeAt(0)?(v+="([^./]"+i+")?",g=g.substr(1)):63===g.charCodeAt(0)&&(v+="[^./]",g=g.substr(1)),v+=g.replace(E,s),v!==g&&(c+=A),c+=v}else c+=g.replace(E,s);l=!0}while(m>0)c+=")?",m--;return c}}function P(e,n){return"*"===e?n:"?"===e?"[^/]":"\\"+e}function R(n,t,r,a,i){n=e.normalizePath(n),i=e.normalizePath(i);var o=s(i,n);return{includeFilePatterns:e.map(w(r,o,"files"),function(e){return"^"+e+"$"}),includeFilePattern:I(r,o,"files"),includeDirectoryPattern:I(r,o,"directories"),excludePattern:I(t,o,"exclude"),basePaths:V(n,r,a)}}function F(e,n){return new RegExp(e,n?"":"i")}function G(n,t,r,a,i,o,c,l){n=e.normalizePath(n),o=e.normalizePath(o);for(var u=R(n,r,a,i,o),d=u.includeFilePatterns&&u.includeFilePatterns.map(function(e){return F(e,i)}),p=u.includeDirectoryPattern&&F(u.includeDirectoryPattern,i),m=u.excludePattern&&F(u.excludePattern,i),f=d?d.map(function(){return[]}):[[]],_=0,g=u.basePaths;_<g.length;_++){var v=g[_];y(v,s(o,v),c)}return e.flatten(f);function y(n,r,a){for(var i=l(n),o=i.files,c=i.directories,u=function(a){var i=s(n,a),o=s(r,a);if(t&&!e.fileExtensionIsOneOf(i,t))return"continue";if(m&&m.test(o))return"continue";if(d){var c=e.findIndex(d,function(e){return e.test(o)});-1!==c&&f[c].push(i)}else f[0].push(i)},_=0,g=e.sort(o,e.compareStringsCaseSensitive);_<g.length;_++){var v=g[_];u(v)}if(void 0===a||(a--,0!==a))for(var h=0,b=e.sort(c,e.compareStringsCaseSensitive);h<b.length;h++){v=b[h];var E=s(n,v),T=s(r,v);p&&!p.test(T)||m&&m.test(T)||y(E,T,a)}}}function V(n,t,r){var a=[n];if(t){for(var i=[],o=0,c=t;o<c.length;o++){var l=c[o],u=e.isRootedDiskPath(l)?l:e.normalizePath(s(n,l));i.push(B(u))}i.sort(e.getStringComparer(!r));for(var d=function(t){e.every(a,function(e){return!v(e,t,n,!r)})&&a.push(t)},p=0,m=i;p<m.length;p++){var f=m[p];d(f)}}return a}function B(n){var t=e.indexOfAnyCharCode(n,x);return t<0?C(n)?u(e.getDirectoryPath(n)):n:n.substring(0,n.lastIndexOf(e.directorySeparator,t))}function K(e,n){return n||j(e)||3}function j(e){var n=e.substr(e.lastIndexOf("."));switch(n.toLowerCase()){case".js":return 1;case".jsx":return 2;case".ts":return 3;case".tsx":return 4;case".json":return 6;default:return 0}}e.getRegularExpressionForWildcard=I,e.getRegularExpressionsForWildcards=w,e.isImplicitGlob=M,e.getFileMatcherPatterns=R,e.getRegexFromPattern=F,e.matchFiles=G,e.ensureScriptKind=K,e.getScriptKindFromFileName=j,e.supportedTSExtensions=[".ts",".tsx",".d.ts"],e.supportedTSExtensionsWithJson=[".ts",".tsx",".d.ts",".json"],e.supportedTSExtensionsForExtractExtension=[".d.ts",".ts",".tsx"],e.supportedJSExtensions=[".js",".jsx"],e.supportedJSAndJsonExtensions=[".js",".jsx",".json"];var U=e.supportedTSExtensions.concat(e.supportedJSExtensions),H=e.supportedTSExtensions.concat(e.supportedJSExtensions,[".json"]);function W(n,t){var r=n&&n.allowJs;if(!t||0===t.length)return r?U:e.supportedTSExtensions;var a=(r?U:e.supportedTSExtensions).concat(e.mapDefined(t,function(e){return 7===e.scriptKind||r&&z(e.scriptKind)?e.extension:void 0}));return e.deduplicate(a,e.equateStringsCaseSensitive,e.compareStringsCaseSensitive)}function q(n,t){return n&&n.resolveJsonModule?t===U?H:t===e.supportedTSExtensions?e.supportedTSExtensionsWithJson:t.concat([".json"]):t}function z(e){return 1===e||2===e}function J(n){return e.some(e.supportedJSExtensions,function(t){return e.fileExtensionIs(n,t)})}function X(n){return e.supportedJSAndJsonExtensions.some(function(t){return e.fileExtensionIs(n,t)})}function Y(n){return e.some(e.supportedTSExtensions,function(t){return e.fileExtensionIs(n,t)})}function Q(n,t,r){if(!n)return!1;for(var a=W(t,r),i=0,o=q(t,a);i<o.length;i++){var s=o[i];if(e.fileExtensionIs(n,s))return!0}return!1}function Z(n,t){for(var r=t.length-1;r>=0;r--)if(e.fileExtensionIs(n,t[r]))return $(r,t);return 0}function $(e,n){return e<2?0:e<n.length?2:n.length}function ee(e,n){return e<2?2:n.length}e.getSupportedExtensions=W,e.getSuppoertedExtensionsWithJsonIfResolveJsonModule=q,e.hasJSFileExtension=J,e.hasJSOrJsonFileExtension=X,e.hasTSFileExtension=Y,e.isSupportedSourceFileName=Q,function(e){e[e["TypeScriptFiles"]=0]="TypeScriptFiles",e[e["DeclarationAndJavaScriptFiles"]=2]="DeclarationAndJavaScriptFiles",e[e["Highest"]=0]="Highest",e[e["Lowest"]=2]="Lowest"}(e.ExtensionPriority||(e.ExtensionPriority={})),e.getExtensionPriority=Z,e.adjustExtensionPriority=$,e.getNextLowestExtensionPriority=ee;var ne,te=[".d.ts",".ts",".js",".tsx",".jsx",".json"];function re(e){for(var n=0,t=te;n<t.length;n++){var r=t[n],a=ae(e,r);if(void 0!==a)return a}return e}function ae(n,t){return e.fileExtensionIs(n,t)?ie(n,t):void 0}function ie(e,n){return e.substring(0,e.length-n.length)}function oe(e,n){return se(e,n,te,!1)}function se(n,t,r,a){var i=void 0!==r&&void 0!==a?ge(n,r,a):ge(n);return i?n.slice(0,n.length-i.length)+(e.startsWith(t,".")?t:"."+t):n}function ce(n){ne.assert(e.hasZeroOrOneAsteriskCharacter(n));var t=n.indexOf("*");return-1===t?void 0:{prefix:n.substr(0,t),suffix:n.substr(t+1)}}function le(e){return!(e>=0)}function ue(e){return".ts"===e||".tsx"===e||".d.ts"===e}function de(e){return ue(e)||".json"===e}function pe(e){var n=fe(e);return void 0!==n?n:ne.fail("File "+e+" has unknown extension.")}function me(e){return void 0!==fe(e)}function fe(n){return e.find(te,function(t){return e.fileExtensionIs(n,t)})}function _e(n,t,r){"string"===typeof t&&(t=[t]);for(var a=0,i=t;a<i.length;a++){var o=i[a];if(e.startsWith(o,".")||(o="."+o),n.length>=o.length&&"."===n.charAt(n.length-o.length)){var s=n.slice(n.length-o.length);if(r(s,o))return s}}return""}function ge(n,t,r){if(t)return _e(n,t,r?e.equateStringsCaseInsensitive:e.equateStringsCaseSensitive);var a=o(n),i=a.lastIndexOf(".");return i>=0?a.substring(i):""}function ve(e,n){return e.checkJsDirective?e.checkJsDirective.enabled:n.checkJs}function ye(n,t){for(var r=[],a=0,i=n;a<i.length;a++){var o=i[a],s=ce(o);if(s)r.push(s);else if(o===t)return o}return e.findBestPatternMatch(r,function(e){return e},t)}function he(e,n){var t=e.indexOf(n);return ne.assert(-1!==t),e.slice(t)}function be(e,n){ne.assert(0!==e.length);for(var t=n(e[0]),r=t,a=1;a<e.length;a++){var i=n(e[a]);i<t?t=i:i>r&&(r=i)}return{min:t,max:r}}e.removeFileExtension=re,e.tryRemoveExtension=ae,e.removeExtension=ie,e.changeExtension=oe,e.changeAnyExtension=se,function(n){function t(n){var t=e.SymbolFlags;return"{ flags: "+(t?r(n.flags,t):n.flags)+"; declarations: "+e.map(n.declarations,a)+" }"}function r(e,n){for(var t=[],r=0;r<=30;r++){var a=1<<r;e&a&&t.push(n[a])}return t.join("|")}function a(n){var t=e.SyntaxKind;return t?t[n.kind]:n.kind.toString()}n.showSymbol=t,n.showSyntaxKind=a}(ne=e.Debug||(e.Debug={})),e.tryParsePattern=ce,e.positionIsSynthesized=le,e.extensionIsTS=ue,e.resolutionExtensionIsTSOrJson=de,e.extensionFromPath=pe,e.isAnySupportedFileExtension=me,e.tryGetExtensionFromPath=fe,e.getAnyExtensionFromPath=ge,e.isCheckJsEnabledForFile=ve,e.emptyFileSystemEntries={files:e.emptyArray,directories:e.emptyArray},e.matchPatternOrExact=ye,e.sliceAfter=he,e.minAndMax=be;var Ee=function(){function n(){this.map=e.createMap()}return n.prototype.add=function(n){this.map.set(String(e.getNodeId(n)),n)},n.prototype.tryAdd=function(e){return!this.has(e)&&(this.add(e),!0)},n.prototype.has=function(n){return this.map.has(String(e.getNodeId(n)))},n.prototype.forEach=function(e){this.map.forEach(e)},n.prototype.some=function(n){return e.forEachEntry(this.map,n)||!1},n}();e.NodeSet=Ee;var Te=function(){function n(){this.map=e.createMap()}return n.prototype.get=function(n){var t=this.map.get(String(e.getNodeId(n)));return t&&t.value},n.prototype.getOrUpdate=function(e,n){var t=this.get(e);if(t)return t;var r=n();return this.set(e,r),r},n.prototype.set=function(n,t){this.map.set(String(e.getNodeId(n)),{node:n,value:t})},n.prototype.has=function(n){return this.map.has(String(e.getNodeId(n)))},n.prototype.forEach=function(e){this.map.forEach(function(n){var t=n.node,r=n.value;return e(r,t)})},n}();function Se(n){return{pos:e.getTokenPosOfNode(n),end:n.end}}function xe(e){return{pos:e.pos-1,end:e.end+1}}function Ce(e,n){return n.skipLibCheck&&e.isDeclarationFile||n.skipDefaultLibCheck&&e.hasNoDefaultLib}function Ae(n,t){return n===t||"object"===typeof n&&null!==n&&"object"===typeof t&&null!==t&&e.equalOwnProperties(n,t,Ae)}function De(e,n,t){var r=e.get(n);if(void 0===r){var a=t();return e.set(n,a),a}return r}function Le(e){var n;switch(e.charCodeAt(1)){case 98:case 66:n=1;break;case 111:case 79:n=3;break;case 120:case 88:n=4;break;default:var t=e.length-1,r=0;while(48===e.charCodeAt(r))r++;return e.slice(r,t)||"0"}for(var a=2,i=e.length-1,o=(i-a)*n,s=new Uint16Array((o>>>4)+(15&o?1:0)),c=i-1,l=0;c>=a;c--,l+=n){var u=l>>>4,d=e.charCodeAt(c),p=d<=57?d-48:10+d-(d<=70?65:97),m=p<<(15&l);s[u]|=m;var f=m>>>16;f&&(s[u+1]|=f)}var _="",g=s.length-1,v=!0;while(v){var y=0;v=!1;for(u=g;u>=0;u--){var h=y<<16|s[u],b=h/10|0;s[u]=b,y=h-10*b,b&&!v&&(g=u,v=!0)}_=y+_}return _}function ke(e){var n=e.negative,t=e.base10Value;return(n&&"0"!==t?"-":"")+t}e.NodeMap=Te,e.rangeOfNode=Se,e.rangeOfTypeParameters=xe,e.skipTypeChecking=Ce,e.isJsonEqual=Ae,e.getOrUpdate=De,e.parsePseudoBigInt=Le,e.pseudoBigIntToString=ke}(u||(u={})),function(e){var n,t,r,a,i,o,s;function c(n,o,s){return 279===n?new(i||(i=e.objectAllocator.getSourceFileConstructor()))(n,o,s):72===n?new(a||(a=e.objectAllocator.getIdentifierConstructor()))(n,o,s):e.isNodeKind(n)?new(t||(t=e.objectAllocator.getNodeConstructor()))(n,o,s):new(r||(r=e.objectAllocator.getTokenConstructor()))(n,o,s)}function l(e,n){return n&&e(n)}function u(e,n,t){if(t){if(n)return n(t);for(var r=0,a=t;r<a.length;r++){var i=a[r],o=e(i);if(o)return o}}}function d(e,n){return 42===e.charCodeAt(n+1)&&42===e.charCodeAt(n+2)&&47!==e.charCodeAt(n+3)}function p(n,t,r){if(n&&!(n.kind<=147))switch(n.kind){case 148:return l(t,n.left)||l(t,n.right);case 150:return l(t,n.name)||l(t,n.constraint)||l(t,n.default)||l(t,n.expression);case 276:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.name)||l(t,n.questionToken)||l(t,n.exclamationToken)||l(t,n.equalsToken)||l(t,n.objectAssignmentInitializer);case 277:return l(t,n.expression);case 151:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.dotDotDotToken)||l(t,n.name)||l(t,n.questionToken)||l(t,n.type)||l(t,n.initializer);case 154:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.name)||l(t,n.questionToken)||l(t,n.exclamationToken)||l(t,n.type)||l(t,n.initializer);case 153:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.name)||l(t,n.questionToken)||l(t,n.type)||l(t,n.initializer);case 275:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.name)||l(t,n.questionToken)||l(t,n.initializer);case 237:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.name)||l(t,n.exclamationToken)||l(t,n.type)||l(t,n.initializer);case 186:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.dotDotDotToken)||l(t,n.propertyName)||l(t,n.name)||l(t,n.initializer);case 165:case 166:case 160:case 161:case 162:return u(t,r,n.decorators)||u(t,r,n.modifiers)||u(t,r,n.typeParameters)||u(t,r,n.parameters)||l(t,n.type);case 156:case 155:case 157:case 158:case 159:case 196:case 239:case 197:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.asteriskToken)||l(t,n.name)||l(t,n.questionToken)||l(t,n.exclamationToken)||u(t,r,n.typeParameters)||u(t,r,n.parameters)||l(t,n.type)||l(t,n.equalsGreaterThanToken)||l(t,n.body);case 164:return l(t,n.typeName)||u(t,r,n.typeArguments);case 163:return l(t,n.parameterName)||l(t,n.type);case 167:return l(t,n.exprName);case 168:return u(t,r,n.members);case 169:return l(t,n.elementType);case 170:return u(t,r,n.elementTypes);case 173:case 174:return u(t,r,n.types);case 175:return l(t,n.checkType)||l(t,n.extendsType)||l(t,n.trueType)||l(t,n.falseType);case 176:return l(t,n.typeParameter);case 183:return l(t,n.argument)||l(t,n.qualifier)||u(t,r,n.typeArguments);case 177:case 179:return l(t,n.type);case 180:return l(t,n.objectType)||l(t,n.indexType);case 181:return l(t,n.readonlyToken)||l(t,n.typeParameter)||l(t,n.questionToken)||l(t,n.type);case 182:return l(t,n.literal);case 184:case 185:return u(t,r,n.elements);case 187:return u(t,r,n.elements);case 188:return u(t,r,n.properties);case 189:return l(t,n.expression)||l(t,n.name);case 190:return l(t,n.expression)||l(t,n.argumentExpression);case 191:case 192:return l(t,n.expression)||u(t,r,n.typeArguments)||u(t,r,n.arguments);case 193:return l(t,n.tag)||u(t,r,n.typeArguments)||l(t,n.template);case 194:return l(t,n.type)||l(t,n.expression);case 195:return l(t,n.expression);case 198:return l(t,n.expression);case 199:return l(t,n.expression);case 200:return l(t,n.expression);case 202:return l(t,n.operand);case 207:return l(t,n.asteriskToken)||l(t,n.expression);case 201:return l(t,n.expression);case 203:return l(t,n.operand);case 204:return l(t,n.left)||l(t,n.operatorToken)||l(t,n.right);case 212:return l(t,n.expression)||l(t,n.type);case 213:return l(t,n.expression);case 214:return l(t,n.name);case 205:return l(t,n.condition)||l(t,n.questionToken)||l(t,n.whenTrue)||l(t,n.colonToken)||l(t,n.whenFalse);case 208:return l(t,n.expression);case 218:case 245:return u(t,r,n.statements);case 279:return u(t,r,n.statements)||l(t,n.endOfFileToken);case 219:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.declarationList);case 238:return u(t,r,n.declarations);case 221:return l(t,n.expression);case 222:return l(t,n.expression)||l(t,n.thenStatement)||l(t,n.elseStatement);case 223:return l(t,n.statement)||l(t,n.expression);case 224:return l(t,n.expression)||l(t,n.statement);case 225:return l(t,n.initializer)||l(t,n.condition)||l(t,n.incrementor)||l(t,n.statement);case 226:return l(t,n.initializer)||l(t,n.expression)||l(t,n.statement);case 227:return l(t,n.awaitModifier)||l(t,n.initializer)||l(t,n.expression)||l(t,n.statement);case 228:case 229:return l(t,n.label);case 230:return l(t,n.expression);case 231:return l(t,n.expression)||l(t,n.statement);case 232:return l(t,n.expression)||l(t,n.caseBlock);case 246:return u(t,r,n.clauses);case 271:return l(t,n.expression)||u(t,r,n.statements);case 272:return u(t,r,n.statements);case 233:return l(t,n.label)||l(t,n.statement);case 234:return l(t,n.expression);case 235:return l(t,n.tryBlock)||l(t,n.catchClause)||l(t,n.finallyBlock);case 274:return l(t,n.variableDeclaration)||l(t,n.block);case 152:return l(t,n.expression);case 240:case 209:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.name)||u(t,r,n.typeParameters)||u(t,r,n.heritageClauses)||u(t,r,n.members);case 241:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.name)||u(t,r,n.typeParameters)||u(t,r,n.heritageClauses)||u(t,r,n.members);case 242:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.name)||u(t,r,n.typeParameters)||l(t,n.type);case 243:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.name)||u(t,r,n.members);case 278:return l(t,n.name)||l(t,n.initializer);case 244:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.name)||l(t,n.body);case 248:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.name)||l(t,n.moduleReference);case 249:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.importClause)||l(t,n.moduleSpecifier);case 250:return l(t,n.name)||l(t,n.namedBindings);case 247:return l(t,n.name);case 251:return l(t,n.name);case 252:case 256:return u(t,r,n.elements);case 255:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.exportClause)||l(t,n.moduleSpecifier);case 253:case 257:return l(t,n.propertyName)||l(t,n.name);case 254:return u(t,r,n.decorators)||u(t,r,n.modifiers)||l(t,n.expression);case 206:return l(t,n.head)||u(t,r,n.templateSpans);case 216:return l(t,n.expression)||l(t,n.literal);case 149:return l(t,n.expression);case 273:return u(t,r,n.types);case 211:return l(t,n.expression)||u(t,r,n.typeArguments);case 259:return l(t,n.expression);case 258:return u(t,r,n.decorators);case 309:return u(t,r,n.elements);case 260:return l(t,n.openingElement)||u(t,r,n.children)||l(t,n.closingElement);case 264:return l(t,n.openingFragment)||u(t,r,n.children)||l(t,n.closingFragment);case 261:case 262:return l(t,n.tagName)||u(t,r,n.typeArguments)||l(t,n.attributes);case 268:return u(t,r,n.properties);case 267:return l(t,n.name)||l(t,n.initializer);case 269:return l(t,n.expression);case 270:return l(t,n.dotDotDotToken)||l(t,n.expression);case 263:return l(t,n.tagName);case 171:case 172:case 283:case 287:case 286:case 288:case 290:return l(t,n.type);case 289:return u(t,r,n.parameters)||l(t,n.type);case 291:return u(t,r,n.tags);case 299:case 305:return l(t,n.tagName)||(n.isNameFirst?l(t,n.name)||l(t,n.typeExpression):l(t,n.typeExpression)||l(t,n.name));case 295:return l(t,n.tagName)||l(t,n.class);case 303:return l(t,n.tagName)||l(t,n.constraint)||u(t,r,n.typeParameters);case 304:return l(t,n.tagName)||(n.typeExpression&&283===n.typeExpression.kind?l(t,n.typeExpression)||l(t,n.fullName):l(t,n.fullName)||l(t,n.typeExpression));case 297:return l(t,n.tagName)||l(t,n.fullName)||l(t,n.typeExpression);case 300:case 302:case 301:case 298:return l(t,n.tagName)||l(t,n.typeExpression);case 293:return e.forEach(n.typeParameters,t)||e.forEach(n.parameters,t)||l(t,n.type);case 292:return e.forEach(n.jsDocPropertyTags,t);case 294:case 296:return l(t,n.tagName);case 308:return l(t,n.expression)}}function m(n,t,r,a,i){var s;return void 0===a&&(a=!1),e.performance.mark("beforeParse"),s=100===r?o.parseSourceFile(n,t,r,void 0,a,6):o.parseSourceFile(n,t,r,void 0,a,i),e.performance.mark("afterParse"),e.performance.measure("Parse","beforeParse","afterParse"),s}function f(e,n){return o.parseIsolatedEntityName(e,n)}function _(e,n){return o.parseJsonText(e,n)}function g(e){return void 0!==e.externalModuleIndicator}function v(e,n,t,r){void 0===r&&(r=!1);var a=s.updateSourceFile(e,n,t,r);return a.flags|=1572864&e.flags,a}function y(e,n,t){var r=o.JSDocParser.parseIsolatedJSDocComment(e,n,t);return r&&r.jsDoc&&o.fixupParentReferences(r.jsDoc),r}function h(e,n,t){return o.JSDocParser.parseJSDocTypeExpressionForTests(e,n,t)}function b(n){return e.fileExtensionIs(n,".d.ts")}function E(n,t){var r=e.createScanner(n.languageVersion,!1,0,t),a=[];while(1){var i=r.scan();if(!e.isTrivia(i))break;var o={kind:r.getToken(),pos:r.getTokenPos(),end:r.getTextPos()},s=t.substring(o.pos,o.end);D(a,o,s)}n.pragmas=e.createMap();for(var c=0,l=a;c<l.length;c++){var u=l[c];if(n.pragmas.has(u.name)){var d=n.pragmas.get(u.name);d instanceof Array?d.push(u.args):n.pragmas.set(u.name,[d,u.args])}else n.pragmas.set(u.name,u.args)}}function T(n,t){n.checkJsDirective=void 0,n.referencedFiles=[],n.typeReferenceDirectives=[],n.libReferenceDirectives=[],n.amdDependencies=[],n.hasNoDefaultLib=!1,n.pragmas.forEach(function(r,a){switch(a){case"reference":var i=n.referencedFiles,o=n.typeReferenceDirectives,s=n.libReferenceDirectives;e.forEach(e.toArray(r),function(r){var a=r.arguments,c=a.types,l=a.lib,u=a.path;r.arguments["no-default-lib"]?n.hasNoDefaultLib=!0:c?o.push({pos:c.pos,end:c.end,fileName:c.value}):l?s.push({pos:l.pos,end:l.end,fileName:l.value}):u?i.push({pos:u.pos,end:u.end,fileName:u.value}):t(r.range.pos,r.range.end-r.range.pos,e.Diagnostics.Invalid_reference_directive_syntax)});break;case"amd-dependency":n.amdDependencies=e.map(e.toArray(r),function(e){return{name:e.arguments.name,path:e.arguments.path}});break;case"amd-module":if(r instanceof Array)for(var c=0,l=r;c<l.length;c++){var u=l[c];n.moduleName&&t(u.range.pos,u.range.end-u.range.pos,e.Diagnostics.An_AMD_module_cannot_have_multiple_name_assignments),n.moduleName=u.arguments.name}else n.moduleName=r.arguments.name;break;case"ts-nocheck":case"ts-check":e.forEach(e.toArray(r),function(e){(!n.checkJsDirective||e.range.pos>n.checkJsDirective.pos)&&(n.checkJsDirective={enabled:"ts-check"===a,end:e.range.end,pos:e.range.pos})});break;case"jsx":return;default:e.Debug.fail("Unhandled pragma kind")}})}(function(e){e[e["None"]=0]="None",e[e["Yield"]=1]="Yield",e[e["Await"]=2]="Await",e[e["Type"]=4]="Type",e[e["IgnoreMissingOpenBrace"]=16]="IgnoreMissingOpenBrace",e[e["JSDoc"]=32]="JSDoc"})(n||(n={})),e.createNode=c,e.isJSDocLikeText=d,e.forEachChild=p,e.createSourceFile=m,e.parseIsolatedEntityName=f,e.parseJsonText=_,e.isExternalModule=g,e.updateSourceFile=v,e.parseIsolatedJSDocComment=y,e.parseJSDocTypeExpressionForTests=h,function(n){var t,r,a,i,o,s,c,l,u,m,f,_,g,v,y,h,S,x=e.createScanner(6,!0),C=10240,A=!1;function D(n,t,r,a,i,o){if(void 0===i&&(i=!1),o=e.ensureScriptKind(n,o),6===o){var s=k(n,t,r,a,i);return e.convertToObjectWorker(s,s.parseDiagnostics,!1,void 0,void 0),s.referencedFiles=e.emptyArray,s.typeReferenceDirectives=e.emptyArray,s.libReferenceDirectives=e.emptyArray,s.amdDependencies=e.emptyArray,s.hasNoDefaultLib=!1,s.pragmas=e.emptyMap,s}w(t,r,a,o);var c=O(n,r,i,o);return M(),c}function L(e,n){w(e,n,void 0,1),ue();var t=Cn(!0),r=1===le()&&!s.length;return M(),r?t:void 0}function k(n,t,r,a,i){void 0===r&&(r=2),w(t,r,a,6),o=F(n,2,6,!1),o.flags=v,ue();var c=ce();if(1===le())o.statements=Ie([],c,c),o.endOfFileToken=Ae();else{var l=ke(221);switch(le()){case 22:l.expression=Zr();break;case 102:case 87:case 96:l.expression=Ae();break;case 39:he(function(){return 8===ue()&&57!==ue()})?l.expression=fr():l.expression=ea();break;case 8:case 10:if(he(function(){return 57!==ue()})){l.expression=Nn();break}default:l.expression=ea();break}we(l),o.statements=Ie([l],c),o.endOfFileToken=Ce(1,e.Diagnostics.Unexpected_token)}i&&R(o),o.parseDiagnostics=s;var u=o;return M(),u}function I(e){return 4===e||2===e||1===e||6===e?1:0}function w(n,o,l,d){switch(t=e.objectAllocator.getNodeConstructor(),r=e.objectAllocator.getTokenConstructor(),a=e.objectAllocator.getIdentifierConstructor(),i=e.objectAllocator.getSourceFileConstructor(),u=n,c=l,s=[],g=0,f=e.createMap(),_=0,m=0,d){case 1:case 2:v=65536;break;case 6:v=16842752;break;default:v=0;break}A=!1,x.setText(u),x.setOnError(se),x.setScriptTarget(o),x.setLanguageVariant(I(d))}function M(){x.setText(""),x.setOnError(void 0),s=void 0,o=void 0,f=void 0,c=void 0,u=void 0}function O(n,t,r,a){var i=b(n);return i&&(v|=4194304),o=F(n,t,a,i),o.flags=v,ue(),E(o,u),T(o,c),o.statements=on(0,wa),e.Debug.assert(1===le()),o.endOfFileToken=P(Ae()),ji(o),o.nodeCount=m,o.identifierCount=_,o.identifiers=f,o.parseDiagnostics=s,r&&R(o),o;function c(n,t,r){s.push(e.createFileDiagnostic(o,n,t,r))}}function P(n){e.Debug.assert(!n.jsDoc);var t=e.mapDefined(e.getJSDocCommentRanges(n,o.text),function(e){return S.parseJSDocComment(n,e.pos,e.end-e.pos)});return t.length&&(n.jsDoc=t),n}function R(n){var t=n;return void p(n,r);function r(n){if(n.parent!==t){n.parent=t;var a=t;if(t=n,p(n,r),e.hasJSDocNodes(n))for(var i=0,o=n.jsDoc;i<o.length;i++){var s=o[i];s.parent=n,t=s,p(s,r)}t=a}}}function F(n,t,r,a){var o=new i(279,0,u.length);return m++,o.text=u,o.bindDiagnostics=[],o.bindSuggestionDiagnostics=void 0,o.languageVersion=t,o.fileName=e.normalizePath(n),o.languageVariant=I(r),o.isDeclarationFile=a,o.scriptKind=r,o}function G(e,n){e?v|=n:v&=~n}function V(e){G(e,2048)}function B(e){G(e,4096)}function K(e){G(e,8192)}function j(e){G(e,16384)}function U(e,n){var t=e&v;if(t){G(!1,t);var r=n();return G(!0,t),r}return n()}function H(e,n){var t=e&~v;if(t){G(!0,t);var r=n();return G(!1,t),r}return n()}function W(e){return U(2048,e)}function q(e){return H(2048,e)}function z(e){return H(4096,e)}function J(e){return H(8192,e)}function X(e){return H(16384,e)}function Y(e){return U(16384,e)}function Q(e){return H(20480,e)}function Z(e){return 0!==(v&e)}function $(){return Z(4096)}function ee(){return Z(2048)}function ne(){return Z(8192)}function te(){return Z(16384)}function re(e,n){ie(x.getTokenPos(),x.getTextPos(),e,n)}function ae(n,t,r,a){var i=e.lastOrUndefined(s);i&&n===i.start||s.push(e.createFileDiagnostic(o,n,t,r,a)),A=!0}function ie(e,n,t,r){ae(e,n-e,t,r)}function oe(e,n,t){ie(e.pos,e.end,n,t)}function se(e,n){ae(x.getTextPos(),n,e)}function ce(){return x.getStartPos()}function le(){return l}function ue(){return l=x.scan()}function de(){return l=x.reScanGreaterToken()}function pe(){return l=x.reScanSlashToken()}function me(){return l=x.reScanTemplateToken()}function fe(){return l=x.reScanLessThanToken()}function _e(){return l=x.scanJsxIdentifier()}function ge(){return l=x.scanJsxToken()}function ve(){return l=x.scanJsxAttributeValue()}function ye(n,t){var r=l,a=s.length,i=A,o=v,c=t?x.lookAhead(n):x.tryScan(n);return e.Debug.assert(o===v),c&&!t||(l=r,s.length=a,A=i),c}function he(e){return ye(e,!0)}function be(e){return ye(e,!1)}function Ee(){return 72===le()||(117!==le()||!$())&&((122!==le()||!te())&&le()>108)}function Te(n,t,r){return void 0===r&&(r=!0),le()===n?(r&&ue(),!0):(t?re(t):re(e.Diagnostics._0_expected,e.tokenToString(n)),!1)}function Se(e){return le()===e&&(ue(),!0)}function xe(e){if(le()===e)return Ae()}function Ce(n,t,r){return xe(n)||Me(n,!1,t||e.Diagnostics._0_expected,r||e.tokenToString(n))}function Ae(){var e=ke(le());return ue(),we(e)}function De(){return 26===le()||(19===le()||1===le()||x.hasPrecedingLineBreak())}function Le(){return De()?(26===le()&&ue(),!0):Te(26)}function ke(n,i){m++;var o=i>=0?i:x.getStartPos();return e.isNodeKind(n)||0===n?new t(n,o,o):72===n?new a(n,o,o):new r(n,o,o)}function Ne(e,n){var t=ke(e,n);return 2&x.getTokenFlags()&&P(t),t}function Ie(e,n,t){var r=e.length,a=r>=1&&r<=4?e.slice():e;return a.pos=n,a.end=void 0===t?x.getStartPos():t,a}function we(e,n){return e.end=void 0===n?x.getStartPos():n,v&&(e.flags|=v),A&&(A=!1,e.flags|=32768),e}function Me(n,t,r,a){t?ae(x.getStartPos(),0,r,a):r&&re(r,a);var i=ke(n);return 72===n?i.escapedText="":(e.isLiteralKind(n)||e.isTemplateLiteralKind(n))&&(i.text=""),we(i)}function Oe(e){var n=f.get(e);return void 0===n&&f.set(e,n=e),n}function Pe(n,t){if(_++,n){var r=ke(72);return 72!==le()&&(r.originalKeywordKind=le()),r.escapedText=e.escapeLeadingUnderscores(Oe(x.getTokenValue())),ue(),we(r)}var a=1===le();return Me(72,a,t||e.Diagnostics.Identifier_expected)}function Re(e){return Pe(Ee(),e)}function Fe(n){return Pe(e.tokenIsIdentifierOrKeyword(le()),n)}function Ge(){return e.tokenIsIdentifierOrKeyword(le())||10===le()||8===le()}function Ve(e){if(10===le()||8===le()){var n=Nn();return n.text=Oe(n.text),n}return e&&22===le()?Ke():Fe()}function Be(){return Ve(!0)}function Ke(){var e=ke(149);return Te(22),e.expression=W(qt),Te(23),we(e)}function je(e){return le()===e&&be(He)}function Ue(){return ue(),!x.hasPrecedingLineBreak()&&qe()}function He(){switch(le()){case 77:return 84===ue();case 85:return ue(),80===le()?he(ze):40!==le()&&119!==le()&&18!==le()&&qe();case 80:return ze();case 116:case 126:case 137:return ue(),qe();default:return Ue()}}function We(){return e.isModifierKind(le())&&be(He)}function qe(){return 22===le()||18===le()||40===le()||25===le()||Ge()}function ze(){return ue(),76===le()||90===le()||110===le()||118===le()&&he(xa)||121===le()&&he(Ca)}function Je(n,t){var r=cn(n);if(r)return!0;switch(n){case 0:case 1:case 3:return!(26===le()&&t)&&ka();case 2:return 74===le()||80===le();case 4:return he(it);case 5:return he(ni)||26===le()&&!t;case 6:return 22===le()||Ge();case 12:switch(le()){case 22:case 40:case 25:case 24:return!0;default:return Ge()}case 18:return Ge();case 9:return 22===le()||25===le()||Ge();case 7:return 18===le()?he(Xe):t?Ee()&&!$e():Ut()&&!$e();case 8:return ja();case 10:return 27===le()||25===le()||ja();case 19:return Ee();case 15:switch(le()){case 27:case 24:return!0}case 11:return 25===le()||Ht();case 16:return Jn(!1);case 17:return Jn(!0);case 20:case 21:return 27===le()||Ct();case 22:return _i();case 23:return e.tokenIsIdentifierOrKeyword(le());case 13:return e.tokenIsIdentifierOrKeyword(le())||18===le();case 14:return!0}return e.Debug.fail("Non-exhaustive case in 'isListElement'.")}function Xe(){if(e.Debug.assert(18===le()),19===ue()){var n=ue();return 27===n||18===n||86===n||109===n}return!0}function Ye(){return ue(),Ee()}function Qe(){return ue(),e.tokenIsIdentifierOrKeyword(le())}function Ze(){return ue(),e.tokenIsIdentifierOrKeywordOrGreaterThan(le())}function $e(){return(109===le()||86===le())&&he(en)}function en(){return ue(),Ht()}function nn(){return ue(),Ct()}function tn(e){if(1===le())return!0;switch(e){case 1:case 2:case 4:case 5:case 6:case 12:case 9:case 23:return 19===le();case 3:return 19===le()||74===le()||80===le();case 7:return 18===le()||86===le()||109===le();case 8:return rn();case 19:return 30===le()||20===le()||18===le()||86===le()||109===le();case 11:return 21===le()||26===le();case 15:case 21:case 10:return 23===le();case 17:case 16:case 18:return 21===le()||23===le();case 20:return 27!==le();case 22:return 18===le()||19===le();case 13:return 30===le()||42===le();case 14:return 28===le()&&he(Di);default:return!1}}function rn(){return!!De()||(!!lr(le())||37===le())}function an(){for(var e=0;e<24;e++)if(g&1<<e&&(Je(e,!0)||tn(e)))return!0;return!1}function on(e,n){var t=g;g|=1<<e;var r=[],a=ce();while(!tn(e))if(Je(e,!1)){var i=sn(e,n);r.push(i)}else if(hn(e))break;return g=t,Ie(r,a)}function sn(e,n){var t=cn(e);return t?ln(t):n()}function cn(n){if(c&&un(n)&&!A){var t=c.currentNode(x.getStartPos());if(!(e.nodeIsMissing(t)||t.intersectsChange||e.containsParseError(t))){var r=12679168&t.flags;if(r===v&&dn(t,n))return t.jsDocCache&&(t.jsDocCache=void 0),t}}}function ln(e){return x.setTextPos(e.end),ue(),e}function un(e){switch(e){case 5:case 2:case 0:case 1:case 3:case 6:case 4:case 8:case 17:case 16:return!0}return!1}function dn(e,n){switch(n){case 5:return pn(e);case 2:return mn(e);case 0:case 1:case 3:return fn(e);case 6:return _n(e);case 4:return gn(e);case 8:return vn(e);case 17:case 16:return yn(e)}return!1}function pn(e){if(e)switch(e.kind){case 157:case 162:case 158:case 159:case 154:case 217:return!0;case 156:var n=e,t=72===n.name.kind&&124===n.name.originalKeywordKind;return!t}return!1}function mn(e){if(e)switch(e.kind){case 271:case 272:return!0}return!1}function fn(e){if(e)switch(e.kind){case 239:case 219:case 218:case 222:case 221:case 234:case 230:case 232:case 229:case 228:case 226:case 227:case 225:case 224:case 231:case 220:case 235:case 233:case 223:case 236:case 249:case 248:case 255:case 254:case 244:case 240:case 241:case 243:case 242:return!0}return!1}function _n(e){return 278===e.kind}function gn(e){if(e)switch(e.kind){case 161:case 155:case 162:case 153:case 160:return!0}return!1}function vn(e){if(237!==e.kind)return!1;var n=e;return void 0===n.initializer}function yn(e){if(151!==e.kind)return!1;var n=e;return void 0===n.initializer}function hn(e){return re(bn(e)),!!an()||(ue(),!1)}function bn(n){switch(n){case 0:return e.Diagnostics.Declaration_or_statement_expected;case 1:return e.Diagnostics.Declaration_or_statement_expected;case 2:return e.Diagnostics.case_or_default_expected;case 3:return e.Diagnostics.Statement_expected;case 18:case 4:return e.Diagnostics.Property_or_signature_expected;case 5:return e.Diagnostics.Unexpected_token_A_constructor_method_accessor_or_property_was_expected;case 6:return e.Diagnostics.Enum_member_expected;case 7:return e.Diagnostics.Expression_expected;case 8:return e.Diagnostics.Variable_declaration_expected;case 9:return e.Diagnostics.Property_destructuring_pattern_expected;case 10:return e.Diagnostics.Array_element_destructuring_pattern_expected;case 11:return e.Diagnostics.Argument_expression_expected;case 12:return e.Diagnostics.Property_assignment_expected;case 15:return e.Diagnostics.Expression_or_comma_expected;case 17:return e.Diagnostics.Parameter_declaration_expected;case 16:return e.Diagnostics.Parameter_declaration_expected;case 19:return e.Diagnostics.Type_parameter_declaration_expected;case 20:return e.Diagnostics.Type_argument_expected;case 21:return e.Diagnostics.Type_expected;case 22:return e.Diagnostics.Unexpected_token_expected;case 23:return e.Diagnostics.Identifier_expected;case 13:return e.Diagnostics.Identifier_expected;case 14:return e.Diagnostics.Identifier_expected;default:return}}function En(e,n,t){var r=g;g|=1<<e;var a=[],i=ce(),o=-1;while(1)if(Je(e,!1)){var s=x.getStartPos();if(a.push(sn(e,n)),o=x.getTokenPos(),Se(27))continue;if(o=-1,tn(e))break;Te(27),t&&26===le()&&!x.hasPrecedingLineBreak()&&ue(),s===x.getStartPos()&&ue()}else{if(tn(e))break;if(hn(e))break}g=r;var c=Ie(a,i);return o>=0&&(c.hasTrailingComma=!0),c}function Tn(){var e=Ie([],ce());return e.isMissingList=!0,e}function Sn(e){return!!e.isMissingList}function xn(e,n,t,r){if(Te(t)){var a=En(e,n);return Te(r),a}return Tn()}function Cn(e,n){var t=e?Fe(n):Re(n),r=x.getStartPos();while(Se(24)){if(28===le()){t.jsdocDotPos=r;break}r=x.getStartPos(),t=An(t,Dn(e))}return t}function An(e,n){var t=ke(148,e.pos);return t.left=e,t.right=n,we(t)}function Dn(n){if(x.hasPrecedingLineBreak()&&e.tokenIsIdentifierOrKeyword(le())){var t=he(Sa);if(t)return Me(72,!0,e.Diagnostics.Identifier_expected)}return n?Fe():Re()}function Ln(){var n=ke(206);n.head=In(),e.Debug.assert(15===n.head.kind,"Template head has wrong token kind");var t=[],r=ce();do{t.push(kn())}while(16===e.last(t).literal.kind);return n.templateSpans=Ie(t,r),we(n)}function kn(){var n,t=ke(216);return t.expression=W(qt),19===le()?(me(),n=wn()):n=Ce(17,e.Diagnostics._0_expected,e.tokenToString(19)),t.literal=n,we(t)}function Nn(){return Mn(le())}function In(){var n=Mn(le());return e.Debug.assert(15===n.kind,"Template head has wrong token kind"),n}function wn(){var n=Mn(le());return e.Debug.assert(16===n.kind||17===n.kind,"Template fragment has wrong token kind"),n}function Mn(e){var n=ke(e);return n.text=x.getTokenValue(),x.hasExtendedUnicodeEscape()&&(n.hasExtendedUnicodeEscape=!0),x.isUnterminated()&&(n.isUnterminated=!0),8===n.kind&&(n.numericLiteralFlags=1008&x.getTokenFlags()),ue(),we(n),n}function On(){var n=ke(164);return n.typeName=Cn(!0,e.Diagnostics.Type_expected),x.hasPrecedingLineBreak()||28!==fe()||(n.typeArguments=xn(20,Bt,28,30)),we(n)}function Pn(n){switch(n.kind){case 164:return e.nodeIsMissing(n.typeName);case 165:case 166:var t=n,r=t.parameters,a=t.type;return Sn(r)||Pn(a);case 177:return Pn(n.type);default:return!1}}function Rn(e){ue();var n=ke(163,e.pos);return n.parameterName=e,n.type=Bt(),we(n)}function Fn(){var e=ke(178);return ue(),we(e)}function Gn(e){var n=ke(284);return e?Lt(288,n):(ue(),we(n))}function Vn(){var e=ke(287);return ue(),e.type=xt(),we(e)}function Bn(){var e=x.getStartPos();if(ue(),27===le()||19===le()||21===le()||30===le()||59===le()||50===le()){var n=ke(285,e);return we(n)}n=ke(286,e);return n.type=Bt(),we(n)}function Kn(){if(he(Ai)){var e=Ne(289);return ue(),Yn(57,36,e),we(e)}var n=ke(164);return n.typeName=Fe(),we(n)}function jn(){var e=ke(151);return 100!==le()&&95!==le()||(e.name=Fe(),Te(57)),e.type=Un(),we(e)}function Un(){x.setInJSDocType(!0);var e=xe(25),n=Gt();if(x.setInJSDocType(!1),e){var t=ke(290,e.pos);t.type=n,n=we(t)}return 59===le()?Lt(288,n):n}function Hn(){var e=ke(167);return Te(104),e.exprName=Cn(!0),we(e)}function Wn(){var e=ke(150);return e.name=Re(),Se(86)&&(Ct()||!Ht()?e.constraint=Bt():e.expression=br()),Se(59)&&(e.default=Bt()),we(e)}function qn(){if(28===le())return xn(19,Wn,28,30)}function zn(){if(Se(57))return Bt()}function Jn(n){return 25===le()||ja()||e.isModifierKind(le())||58===le()||Ct(!n)}function Xn(){var n=Ne(151);return 100===le()?(n.name=Pe(!0),n.type=zn(),we(n)):(n.decorators=ti(),n.modifiers=ri(),n.dotDotDotToken=xe(25),n.name=Ua(),0===e.getFullWidth(n.name)&&!e.hasModifiers(n)&&e.isModifierKind(le())&&ue(),n.questionToken=xe(56),n.type=zn(),n.initializer=zt(),we(n))}function Yn(e,n,t){32&n||(t.typeParameters=qn());var r=Zn(t,n);return(!Qn(e,!!(4&n))||(t.type=Gt(),!Pn(t.type)))&&r}function Qn(n,t){return 37===n?(Te(n),!0):!!Se(57)||!(!t||37!==le())&&(re(e.Diagnostics._0_expected,e.tokenToString(57)),ue(),!0)}function Zn(e,n){if(!Te(20))return e.parameters=Tn(),!1;var t=$(),r=te();return B(!!(1&n)),j(!!(2&n)),e.parameters=32&n?En(17,jn):En(16,Xn),B(t),j(r),Te(21)}function $n(){Se(27)||Le()}function et(e){var n=Ne(e);return 161===e&&Te(95),Yn(57,4,n),$n(),we(n)}function nt(){return 22===le()&&he(tt)}function tt(){if(ue(),25===le()||23===le())return!0;if(e.isModifierKind(le())){if(ue(),Ee())return!0}else{if(!Ee())return!1;ue()}return 57===le()||27===le()||56===le()&&(ue(),57===le()||27===le()||23===le())}function rt(e){return e.kind=162,e.parameters=xn(16,Xn,22,23),e.type=jt(),$n(),we(e)}function at(e){return e.name=Be(),e.questionToken=xe(56),20===le()||28===le()?(e.kind=155,Yn(57,4,e)):(e.kind=153,e.type=jt(),59===le()&&(e.initializer=zt())),$n(),we(e)}function it(){if(20===le()||28===le())return!0;var n=!1;while(e.isModifierKind(le()))n=!0,ue();return 22===le()||(Ge()&&(n=!0,ue()),!!n&&(20===le()||28===le()||56===le()||57===le()||27===le()||De()))}function ot(){if(20===le()||28===le())return et(160);if(95===le()&&he(st))return et(161);var e=Ne(0);return e.modifiers=ri(),nt()?rt(e):at(e)}function st(){return ue(),20===le()||28===le()}function ct(){return 24===ue()}function lt(){switch(ue()){case 20:case 28:case 24:return!0}return!1}function ut(){var e=ke(168);return e.members=dt(),we(e)}function dt(){var e;return Te(18)?(e=on(4,ot),Te(19)):e=Tn(),e}function pt(){return ue(),38===le()||39===le()?133===ue():(133===le()&&ue(),22===le()&&Ye()&&93===ue())}function mt(){var e=ke(150);return e.name=Re(),Te(93),e.constraint=Bt(),we(e)}function ft(){var e=ke(181);return Te(18),133!==le()&&38!==le()&&39!==le()||(e.readonlyToken=Ae(),133!==e.readonlyToken.kind&&Ce(133)),Te(22),e.typeParameter=mt(),Te(23),56!==le()&&38!==le()&&39!==le()||(e.questionToken=Ae(),56!==e.questionToken.kind&&Ce(56)),e.type=jt(),Le(),Te(19),we(e)}function _t(){var e=ce();if(Se(25)){var n=ke(172,e);return n.type=Bt(),we(n)}var t=Bt();return 2097152&v||286!==t.kind||t.pos!==t.type.pos||(t.kind=171),t}function gt(){var e=ke(170);return e.elementTypes=xn(21,_t,22,23),we(e)}function vt(){var e=ke(177);return Te(20),e.type=Bt(),Te(21),we(e)}function yt(){var e=ce(),n=Se(95)?166:165,t=Ne(n,e);return Yn(37,4,t),we(t)}function ht(){var e=Ae();return 24===le()?void 0:e}function bt(e){var n,t=ke(182);e&&(n=ke(202),n.operator=39,ue());var r=102===le()||87===le()?Ae():Mn(le());return e&&(n.operand=r,we(n),r=n),t.literal=r,we(t)}function Et(){return ue(),92===le()}function Tt(){o.flags|=524288;var n=ke(183);return Se(104)&&(n.isTypeOf=!0),Te(92),Te(20),n.argument=Bt(),Te(21),Se(24)&&(n.qualifier=Cn(!0,e.Diagnostics.Type_expected)),n.typeArguments=fi(),we(n)}function St(){return ue(),8===le()||9===le()}function xt(){switch(le()){case 120:case 143:case 138:case 135:case 146:case 139:case 123:case 141:case 132:case 136:return be(ht)||On();case 40:return Gn(!1);case 62:return Gn(!0);case 56:return Bn();case 90:return Kn();case 52:return Vn();case 14:case 10:case 8:case 9:case 102:case 87:return bt();case 39:return he(St)?bt(!0):On();case 106:case 96:return Ae();case 100:var e=Fn();return 128!==le()||x.hasPrecedingLineBreak()?e:Rn(e);case 104:return he(Et)?Tt():Hn();case 18:return he(pt)?ft():ut();case 22:return gt();case 20:return vt();case 92:return Tt();default:return On()}}function Ct(e){switch(le()){case 120:case 143:case 138:case 135:case 146:case 123:case 139:case 142:case 106:case 141:case 96:case 100:case 104:case 132:case 18:case 22:case 28:case 50:case 49:case 95:case 10:case 8:case 9:case 102:case 87:case 136:case 40:case 56:case 52:case 25:case 127:case 92:return!0;case 90:return!e;case 39:return!e&&he(St);case 20:return!e&&he(At);default:return Ee()}}function At(){return ue(),21===le()||Jn(!1)||Ct()}function Dt(){var e=xt();while(!x.hasPrecedingLineBreak())switch(le()){case 52:e=Lt(287,e);break;case 56:if(!(2097152&v)&&he(nn))return e;e=Lt(286,e);break;case 22:if(Te(22),Ct()){var n=ke(180,e.pos);n.objectType=e,n.indexType=Bt(),Te(23),e=we(n)}else{n=ke(169,e.pos);n.elementType=e,Te(23),e=we(n)}break;default:return e}return e}function Lt(e,n){ue();var t=ke(e,n.pos);return t.type=n,we(t)}function kt(e){var n=ke(179);return Te(e),n.operator=e,n.type=It(),we(n)}function Nt(){var e=ke(176);Te(127);var n=ke(150);return n.name=Re(),e.typeParameter=we(n),we(e)}function It(){var e=le();switch(e){case 129:case 142:return kt(e);case 127:return Nt()}return Dt()}function wt(e,n,t){Se(t);var r=n();if(le()===t){var a=[r];while(Se(t))a.push(n());var i=ke(e,r.pos);i.types=Ie(a,r.pos),r=we(i)}return r}function Mt(){return wt(174,It,49)}function Ot(){return wt(173,Mt,50)}function Pt(){return 28===le()||20===le()&&he(Ft)}function Rt(){if(e.isModifierKind(le())&&ri(),Ee()||100===le())return ue(),!0;if(22===le()||18===le()){var n=s.length;return Ua(),n===s.length}return!1}function Ft(){if(ue(),21===le()||25===le())return!0;if(Rt()){if(57===le()||27===le()||56===le()||59===le())return!0;if(21===le()&&(ue(),37===le()))return!0}return!1}function Gt(){var e=Ee()&&be(Vt),n=Bt();if(e){var t=ke(163,e.pos);return t.parameterName=e,t.type=n,we(t)}return n}function Vt(){var e=Re();if(128===le()&&!x.hasPrecedingLineBreak())return ue(),e}function Bt(){return U(20480,Kt)}function Kt(e){if(Pt()||95===le())return yt();var n=Ot();if(!e&&!x.hasPrecedingLineBreak()&&Se(86)){var t=ke(175,n.pos);return t.checkType=n,t.extendsType=Kt(!0),Te(56),t.trueType=Kt(),Te(57),t.falseType=Kt(),we(t)}return n}function jt(){return Se(57)?Bt():void 0}function Ut(){switch(le()){case 100:case 98:case 96:case 102:case 87:case 8:case 9:case 10:case 14:case 15:case 20:case 22:case 18:case 90:case 76:case 95:case 42:case 64:case 72:return!0;case 92:return he(lt);default:return Ee()}}function Ht(){if(Ut())return!0;switch(le()){case 38:case 39:case 53:case 52:case 81:case 104:case 106:case 44:case 45:case 28:case 122:case 117:return!0;default:return!!dr()||Ee()}}function Wt(){return 18!==le()&&90!==le()&&76!==le()&&58!==le()&&Ht()}function qt(){var e=ne();e&&K(!1);var n,t=Jt();while(n=xe(27))t=pr(t,n,Jt());return e&&K(!0),t}function zt(){return Se(59)?Jt():void 0}function Jt(){if(Xt())return Qt();var n=$t()||rr();if(n)return n;var t=cr(0);return 72===t.kind&&37===le()?Zt(t):e.isLeftHandSideExpression(t)&&e.isAssignmentOperator(de())?pr(t,Ae(),Jt()):sr(t)}function Xt(){return 117===le()&&(!!$()||he(Aa))}function Yt(){return ue(),!x.hasPrecedingLineBreak()&&Ee()}function Qt(){var e=ke(207);return ue(),x.hasPrecedingLineBreak()||40!==le()&&!Ht()?we(e):(e.asteriskToken=xe(40),e.expression=Jt(),we(e))}function Zt(n,t){var r;e.Debug.assert(37===le(),"parseSimpleArrowFunctionExpression should only have been called if we had a =>"),t?(r=ke(197,t.pos),r.modifiers=t):r=ke(197,n.pos);var a=ke(151,n.pos);return a.name=n,we(a),r.parameters=Ie([a],a.pos,a.end),r.equalsGreaterThanToken=Ce(37),r.body=or(!!t),P(we(r))}function $t(){var n=er();if(0!==n){var t=1===n?ir(!0):be(tr);if(t){var r=e.hasModifier(t,256),a=le();return t.equalsGreaterThanToken=Ce(37),t.body=37===a||18===a?or(r):Re(),we(t)}}}function er(){return 20===le()||28===le()||121===le()?he(nr):37===le()?1:0}function nr(){if(121===le()){if(ue(),x.hasPrecedingLineBreak())return 0;if(20!==le()&&28!==le())return 0}var n=le(),t=ue();if(20===n){if(21===t){var r=ue();switch(r){case 37:case 57:case 18:return 1;default:return 0}}if(22===t||18===t)return 2;if(25===t)return 1;if(e.isModifierKind(t)&&121!==t&&he(Ye))return 1;if(!Ee()&&100!==t)return 0;switch(ue()){case 57:return 1;case 56:return ue(),57===le()||27===le()||59===le()||21===le()?1:0;case 27:case 59:case 21:return 2}return 0}if(e.Debug.assert(28===n),!Ee())return 0;if(1===o.languageVariant){var a=he(function(){var e=ue();if(86===e){var n=ue();switch(n){case 59:case 30:return!1;default:return!0}}else if(27===e)return!0;return!1});return a?1:0}return 2}function tr(){return ir(!1)}function rr(){if(121===le()&&1===he(ar)){var e=ai(),n=cr(0);return Zt(n,e)}}function ar(){if(121===le()){if(ue(),x.hasPrecedingLineBreak()||37===le())return 0;var e=cr(0);if(!x.hasPrecedingLineBreak()&&72===e.kind&&37===le())return 1}return 0}function ir(n){var t=Ne(197);t.modifiers=ai();var r=e.hasModifier(t,256)?2:0;if((Yn(57,r,t)||n)&&(n||37===le()||18===le()))return t}function or(e){return 18===le()?ia(e?2:0):26!==le()&&90!==le()&&76!==le()&&ka()&&!Wt()?ia(16|(e?2:0)):e?X(Jt):Y(Jt)}function sr(n){var t=xe(56);if(!t)return n;var r=ke(205,n.pos);return r.condition=n,r.questionToken=t,r.whenTrue=U(C,Jt),r.colonToken=Ce(57),r.whenFalse=e.nodeIsPresent(r.colonToken)?Jt():Me(72,!1,e.Diagnostics._0_expected,e.tokenToString(57)),we(r)}function cr(e){var n=br();return ur(e,n)}function lr(e){return 93===e||147===e}function ur(n,t){while(1){de();var r=e.getBinaryOperatorPrecedence(le()),a=41===le()?r>=n:r>n;if(!a)break;if(93===le()&&ee())break;if(119===le()){if(x.hasPrecedingLineBreak())break;ue(),t=mr(t,Bt())}else t=pr(t,Ae(),cr(r))}return t}function dr(){return(!ee()||93!==le())&&e.getBinaryOperatorPrecedence(le())>0}function pr(e,n,t){var r=ke(204,e.pos);return r.left=e,r.operatorToken=n,r.right=t,we(r)}function mr(e,n){var t=ke(212,e.pos);return t.expression=e,t.type=n,we(t)}function fr(){var e=ke(202);return e.operator=le(),ue(),e.operand=Er(),we(e)}function _r(){var e=ke(198);return ue(),e.expression=Er(),we(e)}function gr(){var e=ke(199);return ue(),e.expression=Er(),we(e)}function vr(){var e=ke(200);return ue(),e.expression=Er(),we(e)}function yr(){return 122===le()&&(!!te()||he(Aa))}function hr(){var e=ke(201);return ue(),e.expression=Er(),we(e)}function br(){if(Tr()){var n=Sr();return 41===le()?ur(e.getBinaryOperatorPrecedence(le()),n):n}var t=le(),r=Er();if(41===le()){var a=e.skipTrivia(u,r.pos),i=r.end;194===r.kind?ie(a,i,e.Diagnostics.A_type_assertion_expression_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses):ie(a,i,e.Diagnostics.An_unary_expression_with_the_0_operator_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses,e.tokenToString(t))}return r}function Er(){switch(le()){case 38:case 39:case 53:case 52:return fr();case 81:return _r();case 104:return gr();case 106:return vr();case 28:return Vr();case 122:if(yr())return hr();default:return Sr()}}function Tr(){switch(le()){case 38:case 39:case 53:case 52:case 81:case 104:case 106:case 122:return!1;case 28:if(1!==o.languageVariant)return!1;default:return!0}}function Sr(){if(44===le()||45===le()){var n=ke(202);return n.operator=le(),ue(),n.operand=xr(),we(n)}if(1===o.languageVariant&&28===le()&&he(Ze))return Dr(!0);var t=xr();if(e.Debug.assert(e.isLeftHandSideExpression(t)),(44===le()||45===le())&&!x.hasPrecedingLineBreak()){n=ke(203,t.pos);return n.operand=t,n.operator=le(),ue(),we(n)}return t}function xr(){var e;if(92===le())if(he(st))o.flags|=524288,e=Ae();else if(he(ct)){var n=x.getStartPos();ue(),ue();var t=ke(214,n);t.keywordToken=92,t.name=Fe(),e=we(t),o.flags|=1048576}else e=Cr();else e=98===le()?Ar():Cr();return Ur(e)}function Cr(){var e=zr();return Br(e)}function Ar(){var n=Ae();if(20===le()||24===le()||22===le())return n;var t=ke(189,n.pos);return t.expression=n,Ce(24,e.Diagnostics.super_must_be_followed_by_an_argument_list_or_member_access),t.name=Dn(!0),we(t)}function Dr(n){var t,r=wr(n);if(262===r.kind){var a=ke(260,r.pos);a.openingElement=r,a.children=Nr(a.openingElement),a.closingElement=Fr(n),N(a.openingElement.tagName,a.closingElement.tagName)||oe(a.closingElement,e.Diagnostics.Expected_corresponding_JSX_closing_tag_for_0,e.getTextOfNodeFromSourceText(u,a.openingElement.tagName)),t=we(a)}else if(265===r.kind){a=ke(264,r.pos);a.openingFragment=r,a.children=Nr(a.openingFragment),a.closingFragment=Gr(n),t=we(a)}else e.Debug.assert(261===r.kind),t=r;if(n&&28===le()){var i=be(function(){return Dr(!0)});if(i){re(e.Diagnostics.JSX_expressions_must_have_one_parent_element);var o=ke(204,t.pos);return o.end=i.end,o.left=t,o.right=i,o.operatorToken=Me(27,!1,void 0),o.operatorToken.pos=o.operatorToken.end=o.right.pos,o}}return t}function Lr(){var e=ke(11);return e.containsOnlyWhiteSpaces=12===l,l=x.scanJsxToken(),we(e)}function kr(n,t){switch(t){case 1:return void(e.isJsxOpeningFragment(n)?oe(n,e.Diagnostics.JSX_fragment_has_no_corresponding_closing_tag):oe(n.tagName,e.Diagnostics.JSX_element_0_has_no_corresponding_closing_tag,e.getTextOfNodeFromSourceText(u,n.tagName)));case 29:case 7:return;case 11:case 12:return Lr();case 18:return Or(!1);case 28:return Dr(!1);default:return e.Debug.assertNever(t)}}function Nr(e){var n=[],t=ce(),r=g;g|=16384;while(1){var a=kr(e,l=x.reScanJsxToken());if(!a)break;n.push(a)}return g=r,Ie(n,t)}function Ir(){var e=ke(268);return e.properties=on(13,Pr),we(e)}function wr(e){var n=x.getStartPos();if(Te(28),30===le()){var t=ke(265,n);return ge(),we(t)}var r,a=Mr(),i=fi(),o=Ir();return 30===le()?(r=ke(262,n),ge()):(Te(42),e?Te(30):(Te(30,void 0,!1),ge()),r=ke(261,n)),r.tagName=a,r.typeArguments=i,r.attributes=o,we(r)}function Mr(){_e();var e=100===le()?Ae():Fe();while(Se(24)){var n=ke(189,e.pos);n.expression=e,n.name=Dn(!0),e=we(n)}return e}function Or(e){var n=ke(270);if(Te(18))return 19!==le()&&(n.dotDotDotToken=xe(25),n.expression=Jt()),e?Te(19):(Te(19,void 0,!1),ge()),we(n)}function Pr(){if(18===le())return Rr();_e();var e=ke(267);if(e.name=Fe(),59===le())switch(ve()){case 10:e.initializer=Nn();break;default:e.initializer=Or(!0);break}return we(e)}function Rr(){var e=ke(269);return Te(18),Te(25),e.expression=qt(),Te(19),we(e)}function Fr(e){var n=ke(263);return Te(29),n.tagName=Mr(),e?Te(30):(Te(30,void 0,!1),ge()),we(n)}function Gr(n){var t=ke(266);return Te(29),e.tokenIsIdentifierOrKeyword(le())&&oe(Mr(),e.Diagnostics.Expected_corresponding_closing_tag_for_JSX_fragment),n?Te(30):(Te(30,void 0,!1),ge()),we(t)}function Vr(){var e=ke(194);return Te(28),e.type=Bt(),Te(30),e.expression=Er(),we(e)}function Br(n){while(1){var t=xe(24);if(t){var r=ke(189,n.pos);r.expression=n,r.name=Dn(!0),n=we(r)}else if(52!==le()||x.hasPrecedingLineBreak())if(ne()||!Se(22)){if(!Kr())return n;n=jr(n,void 0)}else{var a=ke(190,n.pos);if(a.expression=n,23===le())a.argumentExpression=Me(72,!0,e.Diagnostics.An_element_access_expression_should_take_an_argument);else{var i=W(qt);e.isStringOrNumericLiteralLike(i)&&(i.text=Oe(i.text)),a.argumentExpression=i}Te(23),n=we(a)}else{ue();var o=ke(213,n.pos);o.expression=n,n=we(o)}}}function Kr(){return 14===le()||15===le()}function jr(e,n){var t=ke(193,e.pos);return t.tag=e,t.typeArguments=n,t.template=14===le()?Nn():Ln(),we(t)}function Ur(e){while(1)if(e=Br(e),28!==le()&&46!==le()){if(20!==le())return e;t=ke(191,e.pos);t.expression=e,t.arguments=Hr(),e=we(t)}else{var n=be(Wr);if(!n)return e;if(Kr()){e=jr(e,n);continue}var t=ke(191,e.pos);t.expression=e,t.typeArguments=n,t.arguments=Hr(),e=we(t)}}function Hr(){Te(20);var e=En(11,Qr);return Te(21),e}function Wr(){if(28===fe()){ue();var e=En(20,Bt);if(Te(30))return e&&qr()?e:void 0}}function qr(){switch(le()){case 20:case 14:case 15:case 24:case 21:case 23:case 57:case 26:case 56:case 33:case 35:case 34:case 36:case 54:case 55:case 51:case 49:case 50:case 19:case 1:return!0;case 27:case 18:default:return!1}}function zr(){switch(le()){case 8:case 9:case 10:case 14:return Nn();case 100:case 98:case 96:case 102:case 87:return Ae();case 20:return Jr();case 22:return Zr();case 18:return ea();case 121:if(!he(Ca))break;return na();case 76:return oi();case 90:return na();case 95:return ra();case 42:case 64:if(13===pe())return Nn();break;case 15:return Ln()}return Re(e.Diagnostics.Expression_expected)}function Jr(){var e=Ne(195);return Te(20),e.expression=W(qt),Te(21),we(e)}function Xr(){var e=ke(208);return Te(25),e.expression=Jt(),we(e)}function Yr(){return 25===le()?Xr():27===le()?ke(210):Jt()}function Qr(){return U(C,Yr)}function Zr(){var e=ke(187);return Te(22),x.hasPrecedingLineBreak()&&(e.multiLine=!0),e.elements=En(15,Yr),Te(23),we(e)}function $r(){var e=Ne(0);if(xe(25))return e.kind=277,e.expression=Jt(),we(e);if(e.decorators=ti(),e.modifiers=ri(),je(126))return ei(e,158);if(je(137))return ei(e,159);var n=xe(40),t=Ee();if(e.name=Be(),e.questionToken=xe(56),e.exclamationToken=xe(52),n||20===le()||28===le())return Qa(e,n);var r=t&&57!==le();if(r){e.kind=276;var a=xe(59);a&&(e.equalsToken=a,e.objectAssignmentInitializer=W(Jt))}else e.kind=275,Te(57),e.initializer=W(Jt);return we(e)}function ea(){var e=ke(188);return Te(18),x.hasPrecedingLineBreak()&&(e.multiLine=!0),e.properties=En(12,$r,!0),Te(19),we(e)}function na(){var n=ne();n&&K(!1);var t=Ne(196);t.modifiers=ri(),Te(90),t.asteriskToken=xe(40);var r=t.asteriskToken?1:0,a=e.hasModifier(t,256)?2:0;return t.name=r&&a?Q(ta):r?z(ta):a?X(ta):ta(),Yn(57,r|a,t),t.body=ia(r|a),n&&K(!0),we(t)}function ta(){return Ee()?Re():void 0}function ra(){var n=x.getStartPos();if(Te(95),Se(24)){var t=ke(214,n);return t.keywordToken=95,t.name=Fe(),we(t)}var r,a=zr();while(1){a=Br(a),r=be(Wr),Kr()&&(e.Debug.assert(!!r,"Expected a type argument list; all plain tagged template starts should be consumed in 'parseMemberExpressionRest'"),a=jr(a,r),r=void 0);break}var i=ke(192,n);return i.expression=a,i.typeArguments=r,(i.typeArguments||20===le())&&(i.arguments=Hr()),we(i)}function aa(e,n){var t=ke(218);return Te(18,n)||e?(x.hasPrecedingLineBreak()&&(t.multiLine=!0),t.statements=on(1,wa),Te(19)):t.statements=Tn(),we(t)}function ia(e,n){var t=$();B(!!(1&e));var r=te();j(!!(2&e));var a=ne();a&&K(!1);var i=aa(!!(16&e),n);return a&&K(!0),B(t),j(r),i}function oa(){var e=ke(220);return Te(26),we(e)}function sa(){var e=ke(222);return Te(91),Te(20),e.expression=W(qt),Te(21),e.thenStatement=wa(),e.elseStatement=Se(83)?wa():void 0,we(e)}function ca(){var e=ke(223);return Te(82),e.statement=wa(),Te(107),Te(20),e.expression=W(qt),Te(21),Se(26),we(e)}function la(){var e=ke(224);return Te(107),Te(20),e.expression=W(qt),Te(21),e.statement=wa(),we(e)}function ua(){var e=ce();Te(89);var n,t,r=xe(122);if(Te(20),26!==le()&&(n=105===le()||111===le()||77===le()?qa(!0):q(qt)),r?Te(147):Se(147)){var a=ke(227,e);a.awaitModifier=r,a.initializer=n,a.expression=W(Jt),Te(21),t=a}else if(Se(93)){var i=ke(226,e);i.initializer=n,i.expression=W(qt),Te(21),t=i}else{var o=ke(225,e);o.initializer=n,Te(26),26!==le()&&21!==le()&&(o.condition=W(qt)),Te(26),21!==le()&&(o.incrementor=W(qt)),Te(21),t=o}return t.statement=wa(),we(t)}function da(e){var n=ke(e);return Te(229===e?73:78),De()||(n.label=Re()),Le(),we(n)}function pa(){var e=ke(230);return Te(97),De()||(e.expression=W(qt)),Le(),we(e)}function ma(){var e=ke(231);return Te(108),Te(20),e.expression=W(qt),Te(21),e.statement=H(8388608,wa),we(e)}function fa(){var e=ke(271);return Te(74),e.expression=W(qt),Te(57),e.statements=on(3,wa),we(e)}function _a(){var e=ke(272);return Te(80),Te(57),e.statements=on(3,wa),we(e)}function ga(){return 74===le()?fa():_a()}function va(){var e=ke(232);Te(99),Te(20),e.expression=W(qt),Te(21);var n=ke(246);return Te(18),n.clauses=on(2,ga),Te(19),e.caseBlock=we(n),we(e)}function ya(){var e=ke(234);return Te(101),e.expression=x.hasPrecedingLineBreak()?void 0:W(qt),Le(),we(e)}function ha(){var e=ke(235);return Te(103),e.tryBlock=aa(!1),e.catchClause=75===le()?ba():void 0,e.catchClause&&88!==le()||(Te(88),e.finallyBlock=aa(!1)),we(e)}function ba(){var e=ke(274);return Te(75),Se(20)?(e.variableDeclaration=Wa(),Te(21)):e.variableDeclaration=void 0,e.block=aa(!1),we(e)}function Ea(){var e=ke(236);return Te(79),Le(),we(e)}function Ta(){var e=Ne(0),n=W(qt);return 72===n.kind&&Se(57)?(e.kind=233,e.label=n,e.statement=wa()):(e.kind=221,e.expression=n,Le()),we(e)}function Sa(){return ue(),e.tokenIsIdentifierOrKeyword(le())&&!x.hasPrecedingLineBreak()}function xa(){return ue(),76===le()&&!x.hasPrecedingLineBreak()}function Ca(){return ue(),90===le()&&!x.hasPrecedingLineBreak()}function Aa(){return ue(),(e.tokenIsIdentifierOrKeyword(le())||8===le()||9===le()||10===le())&&!x.hasPrecedingLineBreak()}function Da(){while(1)switch(le()){case 105:case 111:case 77:case 90:case 76:case 84:return!0;case 110:case 140:return Yt();case 130:case 131:return Ra();case 118:case 121:case 125:case 113:case 114:case 115:case 133:if(ue(),x.hasPrecedingLineBreak())return!1;continue;case 145:return ue(),18===le()||72===le()||85===le();case 92:return ue(),10===le()||40===le()||18===le()||e.tokenIsIdentifierOrKeyword(le());case 85:if(ue(),59===le()||40===le()||18===le()||80===le()||119===le())return!0;continue;case 116:ue();continue;default:return!1}}function La(){return he(Da)}function ka(){switch(le()){case 58:case 26:case 18:case 105:case 111:case 90:case 76:case 84:case 91:case 82:case 107:case 89:case 78:case 73:case 97:case 108:case 99:case 101:case 103:case 79:case 75:case 88:return!0;case 92:return La()||he(lt);case 77:case 85:return La();case 121:case 125:case 110:case 130:case 131:case 140:case 145:return!0;case 115:case 113:case 114:case 116:case 133:return La()||!he(Sa);default:return Ht()}}function Na(){return ue(),Ee()||18===le()||22===le()}function Ia(){return he(Na)}function wa(){switch(le()){case 26:return oa();case 18:return aa(!1);case 105:return Ja(Ne(237));case 111:if(Ia())return Ja(Ne(237));break;case 90:return Xa(Ne(239));case 76:return si(Ne(240));case 91:return sa();case 82:return ca();case 107:return la();case 89:return ua();case 78:return da(228);case 73:return da(229);case 97:return pa();case 108:return ma();case 99:return va();case 101:return ya();case 103:case 75:case 88:return ha();case 79:return Ea();case 58:return Oa();case 121:case 110:case 140:case 130:case 131:case 125:case 77:case 84:case 85:case 92:case 113:case 114:case 115:case 118:case 116:case 133:case 145:if(La())return Oa();break}return Ta()}function Ma(e){return 125===e.kind}function Oa(){var n=Ne(0);if(n.decorators=ti(),n.modifiers=ri(),e.some(n.modifiers,Ma)){for(var t=0,r=n.modifiers;t<r.length;t++){var a=r[t];a.flags|=4194304}return H(4194304,function(){return Pa(n)})}return Pa(n)}function Pa(n){switch(le()){case 105:case 111:case 77:return Ja(n);case 90:return Xa(n);case 76:return si(n);case 110:return vi(n);case 140:return yi(n);case 84:return bi(n);case 145:case 130:case 131:return xi(n);case 92:return ki(n);case 85:switch(ue(),le()){case 80:case 59:return Ki(n);case 119:return Li(n);default:return Bi(n)}default:if(n.decorators||n.modifiers){var t=Me(258,!0,e.Diagnostics.Declaration_expected);return t.pos=n.pos,t.decorators=n.decorators,t.modifiers=n.modifiers,we(t)}return}}function Ra(){return ue(),!x.hasPrecedingLineBreak()&&(Ee()||10===le())}function Fa(e,n){if(18===le()||!De())return ia(e,n);Le()}function Ga(){if(27===le())return ke(210);var e=ke(186);return e.dotDotDotToken=xe(25),e.name=Ua(),e.initializer=zt(),we(e)}function Va(){var e=ke(186);e.dotDotDotToken=xe(25);var n=Ee(),t=Be();return n&&57!==le()?e.name=t:(Te(57),e.propertyName=t,e.name=Ua()),e.initializer=zt(),we(e)}function Ba(){var e=ke(184);return Te(18),e.elements=En(9,Va),Te(19),we(e)}function Ka(){var e=ke(185);return Te(22),e.elements=En(10,Ga),Te(23),we(e)}function ja(){return 18===le()||22===le()||Ee()}function Ua(){return 22===le()?Ka():18===le()?Ba():Re()}function Ha(){return Wa(!0)}function Wa(e){var n=ke(237);return n.name=Ua(),e&&72===n.name.kind&&52===le()&&!x.hasPrecedingLineBreak()&&(n.exclamationToken=Ae()),n.type=jt(),lr(le())||(n.initializer=zt()),we(n)}function qa(n){var t=ke(238);switch(le()){case 105:break;case 111:t.flags|=1;break;case 77:t.flags|=2;break;default:e.Debug.fail()}if(ue(),147===le()&&he(za))t.declarations=Tn();else{var r=ee();V(n),t.declarations=En(8,n?Wa:Ha),V(r)}return we(t)}function za(){return Ye()&&21===ue()}function Ja(e){return e.kind=219,e.declarationList=qa(!1),Le(),we(e)}function Xa(n){n.kind=239,Te(90),n.asteriskToken=xe(40),n.name=e.hasModifier(n,512)?ta():Re();var t=n.asteriskToken?1:0,r=e.hasModifier(n,256)?2:0;return Yn(57,t|r,n),n.body=Fa(t|r,e.Diagnostics.or_expected),we(n)}function Ya(n){return n.kind=157,Te(124),Yn(57,0,n),n.body=Fa(0,e.Diagnostics.or_expected),we(n)}function Qa(n,t,r){n.kind=156,n.asteriskToken=t;var a=t?1:0,i=e.hasModifier(n,256)?2:0;return Yn(57,a|i,n),n.body=Fa(a|i,r),we(n)}function Za(n){return n.kind=154,n.questionToken||52!==le()||x.hasPrecedingLineBreak()||(n.exclamationToken=Ae()),n.type=jt(),n.initializer=e.hasModifier(n,32)?W(zt):U(6144,zt),Le(),we(n)}function $a(n){var t=xe(40);return n.name=Be(),n.questionToken=xe(56),t||20===le()||28===le()?Qa(n,t,e.Diagnostics.or_expected):Za(n)}function ei(e,n){return e.kind=n,e.name=Be(),Yn(57,0,e),e.body=Fa(0),we(e)}function ni(){var n;if(58===le())return!0;while(e.isModifierKind(le())){if(n=le(),e.isClassMemberModifier(n))return!0;ue()}if(40===le())return!0;if(Ge()&&(n=le(),ue()),22===le())return!0;if(void 0!==n){if(!e.isKeyword(n)||137===n||126===n)return!0;switch(le()){case 20:case 28:case 52:case 57:case 59:case 56:return!0;default:return De()}}return!1}function ti(){var e,n=ce();while(1){var t=ce();if(!Se(58))break;var r=ke(152,t);r.expression=J(xr),we(r),(e||(e=[])).push(r)}return e&&Ie(e,n)}function ri(e){var n,t=ce();while(1){var r=x.getStartPos(),a=le();if(77===le()&&e){if(!be(Ue))break}else if(!We())break;var i=we(ke(a,r));(n||(n=[])).push(i)}return n&&Ie(n,t)}function ai(){var e;if(121===le()){var n=x.getStartPos(),t=le();ue();var r=we(ke(t,n));e=Ie([r],n)}return e}function ii(){if(26===le()){var n=ke(217);return ue(),we(n)}var t=Ne(0);return t.decorators=ti(),t.modifiers=ri(!0),je(126)?ei(t,158):je(137)?ei(t,159):124===le()?Ya(t):nt()?rt(t):e.tokenIsIdentifierOrKeyword(le())||10===le()||8===le()||40===le()||22===le()?$a(t):t.decorators||t.modifiers?(t.name=Me(72,!0,e.Diagnostics.Declaration_expected),Za(t)):e.Debug.fail("Should not have attempted to parse class member declaration.")}function oi(){return ci(Ne(0),209)}function si(e){return ci(e,240)}function ci(e,n){return e.kind=n,Te(76),e.name=li(),e.typeParameters=qn(),e.heritageClauses=di(),Te(18)?(e.members=gi(),Te(19)):e.members=Tn(),we(e)}function li(){return Ee()&&!ui()?Re():void 0}function ui(){return 109===le()&&he(Qe)}function di(){if(_i())return on(22,pi)}function pi(){var n=le();e.Debug.assert(86===n||109===n);var t=ke(273);return t.token=n,ue(),t.types=En(7,mi),we(t)}function mi(){var e=ke(211);return e.expression=xr(),e.typeArguments=fi(),we(e)}function fi(){return 28===le()?xn(20,Bt,28,30):void 0}function _i(){return 86===le()||109===le()}function gi(){return on(5,ii)}function vi(e){return e.kind=241,Te(110),e.name=Re(),e.typeParameters=qn(),e.heritageClauses=di(),e.members=dt(),we(e)}function yi(e){return e.kind=242,Te(140),e.name=Re(),e.typeParameters=qn(),Te(59),e.type=Bt(),Le(),we(e)}function hi(){var e=Ne(278);return e.name=Be(),e.initializer=W(zt),we(e)}function bi(e){return e.kind=243,Te(84),e.name=Re(),Te(18)?(e.members=En(6,hi),Te(19)):e.members=Tn(),we(e)}function Ei(){var e=ke(245);return Te(18)?(e.statements=on(1,wa),Te(19)):e.statements=Tn(),we(e)}function Ti(e,n){e.kind=244;var t=16&n;return e.flags|=n,e.name=Re(),e.body=Se(24)?Ti(ke(0),4|t):Ei(),we(e)}function Si(e){return e.kind=244,145===le()?(e.name=Re(),e.flags|=512):(e.name=Nn(),e.name.text=Oe(e.name.text)),18===le()?e.body=Ei():Le(),we(e)}function xi(e){var n=0;if(145===le())return Si(e);if(Se(131))n|=16;else if(Te(130),10===le())return Si(e);return Ti(e,n)}function Ci(){return 134===le()&&he(Ai)}function Ai(){return 20===ue()}function Di(){return 42===ue()}function Li(e){return e.kind=247,Te(119),Te(131),e.name=Re(),Le(),we(e)}function ki(e){Te(92);var n,t=x.getStartPos();return Ee()&&(n=Re(),27!==le()&&144!==le())?Ni(e,n):(e.kind=249,(n||40===le()||18===le())&&(e.importClause=Ii(n,t),Te(144)),e.moduleSpecifier=Oi(),Le(),we(e))}function Ni(e,n){return e.kind=248,e.name=n,Te(59),e.moduleReference=wi(),Le(),we(e)}function Ii(e,n){var t=ke(250,n);return e&&(t.name=e),t.name&&!Se(27)||(t.namedBindings=40===le()?Pi():Ri(252)),we(t)}function wi(){return Ci()?Mi():Cn(!1)}function Mi(){var e=ke(259);return Te(134),Te(20),e.expression=Oi(),Te(21),we(e)}function Oi(){if(10===le()){var e=Nn();return e.text=Oe(e.text),e}return qt()}function Pi(){var e=ke(251);return Te(40),Te(119),e.name=Re(),we(e)}function Ri(e){var n=ke(e);return n.elements=xn(23,252===e?Gi:Fi,18,19),we(n)}function Fi(){return Vi(257)}function Gi(){return Vi(253)}function Vi(n){var t=ke(n),r=e.isKeyword(le())&&!Ee(),a=x.getTokenPos(),i=x.getTextPos(),o=Fe();return 119===le()?(t.propertyName=o,Te(119),r=e.isKeyword(le())&&!Ee(),a=x.getTokenPos(),i=x.getTextPos(),t.name=Fe()):t.name=o,253===n&&r&&ie(a,i,e.Diagnostics.Identifier_expected),we(t)}function Bi(e){return e.kind=255,Se(40)?(Te(144),e.moduleSpecifier=Oi()):(e.exportClause=Ri(256),(144===le()||10===le()&&!x.hasPrecedingLineBreak())&&(Te(144),e.moduleSpecifier=Oi())),Le(),we(e)}function Ki(e){return e.kind=254,Se(59)?e.isExportEquals=!0:Te(80),e.expression=Jt(),Le(),we(e)}function ji(n){n.externalModuleIndicator=e.forEach(n.statements,Ui)||Hi(n)}function Ui(n){return e.hasModifier(n,1)||248===n.kind&&259===n.moduleReference.kind||249===n.kind||254===n.kind||255===n.kind?n:void 0}function Hi(e){return 1048576&e.flags?Wi(e):void 0}function Wi(e){return qi(e)?e:p(e,Wi)}function qi(n){return e.isMetaProperty(n)&&92===n.keywordToken&&"meta"===n.name.escapedText}n.parseSourceFile=D,n.parseIsolatedEntityName=L,n.parseJsonText=k,n.fixupParentReferences=R,function(e){e[e["SourceElements"]=0]="SourceElements",e[e["BlockStatements"]=1]="BlockStatements",e[e["SwitchClauses"]=2]="SwitchClauses",e[e["SwitchClauseStatements"]=3]="SwitchClauseStatements",e[e["TypeMembers"]=4]="TypeMembers",e[e["ClassMembers"]=5]="ClassMembers",e[e["EnumMembers"]=6]="EnumMembers",e[e["HeritageClauseElement"]=7]="HeritageClauseElement",e[e["VariableDeclarations"]=8]="VariableDeclarations",e[e["ObjectBindingElements"]=9]="ObjectBindingElements",e[e["ArrayBindingElements"]=10]="ArrayBindingElements",e[e["ArgumentExpressions"]=11]="ArgumentExpressions",e[e["ObjectLiteralMembers"]=12]="ObjectLiteralMembers",e[e["JsxAttributes"]=13]="JsxAttributes",e[e["JsxChildren"]=14]="JsxChildren",e[e["ArrayLiteralMembers"]=15]="ArrayLiteralMembers",e[e["Parameters"]=16]="Parameters",e[e["JSDocParameters"]=17]="JSDocParameters",e[e["RestProperties"]=18]="RestProperties",e[e["TypeParameters"]=19]="TypeParameters",e[e["TypeArguments"]=20]="TypeArguments",e[e["TupleElementTypes"]=21]="TupleElementTypes",e[e["HeritageClauses"]=22]="HeritageClauses",e[e["ImportOrExportSpecifiers"]=23]="ImportOrExportSpecifiers",e[e["Count"]=24]="Count"}(y||(y={})),function(e){e[e["False"]=0]="False",e[e["True"]=1]="True",e[e["Unknown"]=2]="Unknown"}(h||(h={})),function(n){function t(e,n,t){w(e,6,void 0,1),o=F("file.js",6,1,!1),x.setText(e,n,t),l=x.scan();var a=r(),i=s;return M(),a?{jsDocTypeExpression:a,diagnostics:i}:void 0}function r(e){var n=ke(283),t=(e?Se:Te)(18);return n.type=H(2097152,Un),e&&!t||Te(19),R(n),we(n)}function a(e,n,t){w(e,6,void 0,1),o={languageVariant:0,text:e};var r=m(n,t),a=s;return M(),r?{jsDoc:r,diagnostics:a}:void 0}function i(e,n,t){var r,a=l,i=s.length,c=A,u=m(n,t);return u&&(u.parent=e),65536&v&&(o.jsDocDiagnostics||(o.jsDocDiagnostics=[]),(r=o.jsDocDiagnostics).push.apply(r,s)),l=a,s.length=i,A=c,u}var c,p;function m(n,t){void 0===n&&(n=0);var a=u,i=void 0===t?a.length:n+t;if(t=i-n,e.Debug.assert(n>=0),e.Debug.assert(n<=i),e.Debug.assert(i<=a.length),d(a,n)){var o,s,c,p=[];return x.scanRange(n+3,t-5,function(){var e,t=1,r=n-Math.max(a.lastIndexOf("\n",n),0)+4;function i(n){e||(e=r),p.push(n),r+=n.length}q();while(z(5));z(4)&&(t=0,r=0);e:while(1){switch(le()){case 58:0===t||1===t?(f(p),T(h(r)),t=0,e=void 0,r++):i(x.getTokenText());break;case 4:p.push(x.getTokenText()),t=0,r=0;break;case 40:var o=x.getTokenText();1===t||2===t?(t=2,i(o)):(t=1,r+=o.length);break;case 5:var s=x.getTokenText();2===t?p.push(s):void 0!==e&&r+s.length>e&&p.push(s.slice(e-r-1)),r+=s.length;break;case 1:break e;default:t=2,i(x.getTokenText());break}q()}return m(p),f(p),_()})}function m(e){while(e.length&&("\n"===e[0]||"\r"===e[0]))e.shift()}function f(e){while(e.length&&""===e[e.length-1].trim())e.pop()}function _(){var e=ke(291,n);return e.tags=o&&Ie(o,s,c),e.comment=p.length?p.join(""):void 0,we(e,i)}function g(){while(1){if(q(),1===le())return!0;if(5!==le()&&4!==le())return!1}}function v(){if(5!==le()&&4!==le()||!he(g))while(5===le()||4===le())q()}function y(){if(5!==le()&&4!==le()||!he(g)){var e=x.hasPrecedingLineBreak();while(e&&40===le()||5===le()||4===le())4===le()?e=!0:40===le()&&(e=!1),q()}}function h(n){e.Debug.assert(58===le());var t=x.getTokenPos();q();var r,a=X(void 0);switch(y(),a.escapedText){case"augments":case"extends":r=I(t,a);break;case"class":case"constructor":r=O(t,a);break;case"this":r=P(t,a);break;case"enum":r=R(t,a);break;case"arg":case"argument":case"param":return D(t,a,2,n);case"return":case"returns":r=k(t,a);break;case"template":r=W(t,a);break;case"type":r=N(t,a);break;case"typedef":r=F(t,a,n);break;case"callback":r=V(t,a,n);break;default:r=E(t,a);break}return r.comment||(r.comment=b(n+r.end-r.pos)),r}function b(n){var t,r=[],a=0;function i(e){t||(t=n),r.push(e),n+=e.length}var o=le();e:while(1){switch(o){case 4:a>=1&&(a=0,r.push(x.getTokenText())),n=0;break;case 58:x.setTextPos(x.getTextPos()-1);case 1:break e;case 5:if(2===a)i(x.getTokenText());else{var s=x.getTokenText();void 0!==t&&n+s.length>t&&r.push(s.slice(t-n-1)),n+=s.length}break;case 18:a=2,he(function(){return 58===q()&&e.tokenIsIdentifierOrKeyword(q())&&"link"===x.getTokenText()})&&(i(x.getTokenText()),q(),i(x.getTokenText()),q()),i(x.getTokenText());break;case 40:if(0===a){a=1,n+=1;break}default:a=2,i(x.getTokenText());break}o=q()}return m(r),f(r),0===r.length?void 0:r.join("")}function E(e,n){var t=ke(294,e);return t.tagName=n,we(t)}function T(e){e&&(o?o.push(e):(o=[e],s=e.pos),c=e.end)}function S(){return y(),18===le()?r():void 0}function C(){if(14===le())return{name:Pe(!0),isBracketed:!1};var e=Se(22),n=J();return e&&(v(),xe(59)&&qt(),Te(23)),{name:n,isBracketed:e}}function A(n){switch(n.kind){case 136:return!0;case 169:return A(n.elementType);default:return e.isTypeReferenceNode(n)&&e.isIdentifier(n.typeName)&&"Object"===n.typeName.escapedText}}function D(e,n,t,r){var a=S(),i=!a;y();var o=C(),s=o.name,c=o.isBracketed;v(),i&&(a=S());var l=ke(1===t?305:299,e),u=b(r+x.getStartPos()-e),d=4!==t&&L(a,s,t,r);return d&&(a=d,i=!0),l.tagName=n,l.typeExpression=a,l.name=s,l.isNameFirst=i,l.isBracketed=c,l.comment=u,we(l)}function L(n,t,r,a){if(n&&A(n.type)){var i=ke(283,x.getTokenPos()),o=void 0,s=void 0,c=x.getStartPos(),l=void 0;while(o=be(function(){return U(r,a,t)}))299!==o.kind&&305!==o.kind||(l=e.append(l,o));if(l)return s=ke(292,c),s.jsDocPropertyTags=l,169===n.type.kind&&(s.isArrayType=!0),i.type=we(s),we(i)}}function k(n,t){e.forEach(o,function(e){return 300===e.kind})&&ie(t.pos,x.getTokenPos(),e.Diagnostics._0_tag_already_specified,t.escapedText);var r=ke(300,n);return r.tagName=t,r.typeExpression=S(),we(r)}function N(n,t){e.forEach(o,function(e){return 302===e.kind})&&ie(t.pos,x.getTokenPos(),e.Diagnostics._0_tag_already_specified,t.escapedText);var a=ke(302,n);return a.tagName=t,a.typeExpression=r(!0),we(a)}function I(e,n){var t=ke(295,e);return t.tagName=n,t.class=w(),we(t)}function w(){var e=Se(18),n=ke(211);n.expression=M(),n.typeArguments=fi();var t=we(n);return e&&Te(19),t}function M(){var e=X();while(Se(24)){var n=ke(189,e.pos);n.expression=e,n.name=X(),e=we(n)}return e}function O(e,n){var t=ke(296,e);return t.tagName=n,we(t)}function P(e,n){var t=ke(301,e);return t.tagName=n,t.typeExpression=r(!0),v(),we(t)}function R(e,n){var t=ke(298,e);return t.tagName=n,t.typeExpression=r(!0),v(),we(t)}function F(n,t,r){var a=S();y();var i,o=ke(304,n);if(o.tagName=t,o.fullName=G(),o.name=B(o.fullName),v(),o.comment=b(r),o.typeExpression=a,!a||A(a.type)){var s=void 0,c=void 0,l=void 0;while(s=be(function(){return j(r)}))if(c||(c=ke(292,n)),302===s.kind){if(l)break;l=s}else c.jsDocPropertyTags=e.append(c.jsDocPropertyTags,s);c&&(a&&169===a.type.kind&&(c.isArrayType=!0),o.typeExpression=l&&l.typeExpression&&!A(l.typeExpression.type)?l.typeExpression:we(c),i=o.typeExpression.end)}return we(o,i||void 0!==o.comment?x.getStartPos():(o.fullName||o.typeExpression||o.tagName).end)}function G(n){var t=x.getTokenPos();if(e.tokenIsIdentifierOrKeyword(le())){var r=X();if(Se(24)){var a=ke(244,t);return n&&(a.flags|=4),a.name=r,a.body=G(!0),we(a)}return n&&(r.isInJSDocNamespace=!0),r}}function V(n,t,r){var a,i=ke(297,n);i.tagName=t,i.fullName=G(),i.name=B(i.fullName),v(),i.comment=b(r);var o=ke(293,n);o.parameters=[];while(a=be(function(){return U(4,r)}))o.parameters=e.append(o.parameters,a);var s=be(function(){if(z(58)){var e=h(r);if(e&&300===e.kind)return e}});return s&&(o.type=s),i.typeExpression=we(o),we(i)}function B(n){if(n){var t=n;while(1){if(e.isIdentifier(t)||!t.body)return e.isIdentifier(t)?t:t.name;t=t.body}}}function K(n,t){while(!e.isIdentifier(n)||!e.isIdentifier(t)){if(e.isIdentifier(n)||e.isIdentifier(t)||n.right.escapedText!==t.right.escapedText)return!1;n=n.left,t=t.left}return n.escapedText===t.escapedText}function j(e){return U(1,e)}function U(n,t,r){var a=!0,i=!1;while(1)switch(q()){case 58:if(a){var o=H(n,t);return!(o&&(299===o.kind||305===o.kind)&&4!==n&&r&&(e.isIdentifier(o.name)||!K(r,o.name.left)))&&o}i=!1;break;case 4:a=!0,i=!1;break;case 40:i&&(a=!1),i=!0;break;case 72:a=!1;break;case 1:return!1}}function H(n,t){e.Debug.assert(58===le());var r=x.getStartPos();q();var a,i=X();switch(v(),i.escapedText){case"type":return 1===n&&N(r,i);case"prop":case"property":a=1;break;case"arg":case"argument":case"param":a=6;break;default:return!1}return!!(n&a)&&D(r,i,n,t)}function W(n,t){var a;18===le()&&(a=r());var i=[],o=ce();do{v();var s=ke(150);s.name=X(e.Diagnostics.Unexpected_token_A_type_parameter_name_was_expected_without_curly_braces),we(s),v(),i.push(s)}while(z(27));var c=ke(303,n);return c.tagName=t,c.constraint=a,c.typeParameters=Ie(i,o),we(c),c}function q(){return l=x.scanJSDocToken()}function z(e){return le()===e&&(q(),!0)}function J(){var e=X();Se(22)&&Te(23);while(Se(24)){var n=X();Se(22)&&Te(23),e=An(e,n)}return e}function X(n){if(!e.tokenIsIdentifierOrKeyword(le()))return Me(72,!n,n||e.Diagnostics.Identifier_expected);var t=x.getTokenPos(),r=x.getTextPos(),a=ke(72,t);return a.escapedText=e.escapeLeadingUnderscores(x.getTokenText()),we(a,r),q(),a}}n.parseJSDocTypeExpressionForTests=t,n.parseJSDocTypeExpression=r,n.parseIsolatedJSDocComment=a,n.parseJSDocComment=i,function(e){e[e["BeginningOfLine"]=0]="BeginningOfLine",e[e["SawAsterisk"]=1]="SawAsterisk",e[e["SavingComments"]=2]="SavingComments"}(c||(c={})),function(e){e[e["Property"]=1]="Property",e[e["Parameter"]=2]="Parameter",e[e["CallbackParameter"]=4]="CallbackParameter"}(p||(p={})),n.parseJSDocCommentWorker=m}(S=n.JSDocParser||(n.JSDocParser={}))}(o||(o={})),function(n){function t(n,t,r,a){if(a=a||e.Debug.shouldAssert(2),d(n,t,r,a),e.textChangeRangeIsUnchanged(r))return n;if(0===n.statements.length)return o.parseSourceFile(n.fileName,t,n.languageVersion,void 0,!0,n.scriptKind);var i=n;e.Debug.assert(!i.hasBeenIncrementallyParsed),i.hasBeenIncrementallyParsed=!0;var s=n.text,u=m(n),p=l(n,r);d(n,t,p,a),e.Debug.assert(p.span.start<=r.span.start),e.Debug.assert(e.textSpanEnd(p.span)===e.textSpanEnd(r.span)),e.Debug.assert(e.textSpanEnd(e.textChangeRangeNewSpan(p))===e.textSpanEnd(e.textChangeRangeNewSpan(r)));var f=e.textChangeRangeNewSpan(p).length-p.span.length;c(i,p.span.start,e.textSpanEnd(p.span),e.textSpanEnd(e.textChangeRangeNewSpan(p)),f,s,t,a);var _=o.parseSourceFile(n.fileName,t,n.languageVersion,u,!0,n.scriptKind);return _}function r(n,t,r,i,o,c){return void(t?u(n):l(n));function l(n){var t="";if(c&&a(n)&&(t=i.substring(n.pos,n.end)),n._children&&(n._children=void 0),n.pos+=r,n.end+=r,c&&a(n)&&e.Debug.assert(t===o.substring(n.pos,n.end)),p(n,l,u),e.hasJSDocNodes(n))for(var d=0,m=n.jsDoc;d<m.length;d++){var f=m[d];l(f)}s(n,c)}function u(e){e._children=void 0,e.pos+=r,e.end+=r;for(var n=0,t=e;n<t.length;n++){var a=t[n];l(a)}}}function a(e){switch(e.kind){case 10:case 8:case 72:return!0}return!1}function i(n,t,r,a,i){e.Debug.assert(n.end>=t,"Adjusting an element that was entirely before the change range"),e.Debug.assert(n.pos<=r,"Adjusting an element that was entirely after the change range"),e.Debug.assert(n.pos<=n.end),n.pos=Math.min(n.pos,a),n.end>=r?n.end+=i:n.end=Math.min(n.end,a),e.Debug.assert(n.pos<=n.end),n.parent&&(e.Debug.assert(n.pos>=n.parent.pos),e.Debug.assert(n.end<=n.parent.end))}function s(n,t){if(t){var r=n.pos,a=function(n){e.Debug.assert(n.pos>=r),r=n.end};if(e.hasJSDocNodes(n))for(var i=0,o=n.jsDoc;i<o.length;i++){var s=o[i];a(s)}p(n,a),e.Debug.assert(r<=n.end)}}function c(n,t,a,o,c,l,u,d){return void m(n);function m(n){if(e.Debug.assert(n.pos<=n.end),n.pos>a)r(n,!1,c,l,u,d);else{var _=n.end;if(_>=t){if(n.intersectsChange=!0,n._children=void 0,i(n,t,a,o,c),p(n,m,f),e.hasJSDocNodes(n))for(var g=0,v=n.jsDoc;g<v.length;g++){var y=v[g];m(y)}s(n,d)}else e.Debug.assert(_<t)}}function f(n){if(e.Debug.assert(n.pos<=n.end),n.pos>a)r(n,!0,c,l,u,d);else{var s=n.end;if(s>=t){n.intersectsChange=!0,n._children=void 0,i(n,t,a,o,c);for(var p=0,f=n;p<f.length;p++){var _=f[p];m(_)}}else e.Debug.assert(s<t)}}}function l(n,t){for(var r=1,a=t.span.start,i=0;a>0&&i<=r;i++){var o=u(n,a);e.Debug.assert(o.pos<=a);var s=o.pos;a=Math.max(0,s-1)}var c=e.createTextSpanFromBounds(a,e.textSpanEnd(t.span)),l=t.newLength+(t.span.start-a);return e.createTextChangeRange(c,l)}function u(n,t){var r,a=n;if(p(n,s),r){var i=o(r);i.pos>a.pos&&(a=i)}return a;function o(n){while(1){var t=e.getLastChild(n);if(!t)return n;n=t}}function s(n){if(!e.nodeIsMissing(n))return n.pos<=t?(n.pos>=a.pos&&(a=n),t<n.end?(p(n,s),!0):(e.Debug.assert(n.end<=t),void(r=n))):(e.Debug.assert(n.pos>t),!0)}}function d(n,t,r,a){var i=n.text;if(r&&(e.Debug.assert(i.length-r.span.length+r.newLength===t.length),a||e.Debug.shouldAssert(3))){var o=i.substr(0,r.span.start),s=t.substr(0,r.span.start);e.Debug.assert(o===s);var c=i.substring(e.textSpanEnd(r.span),i.length),l=t.substring(e.textSpanEnd(e.textChangeRangeNewSpan(r)),t.length);e.Debug.assert(c===l)}}function m(n){var t=n.statements,r=0;e.Debug.assert(r<t.length);var a=t[r],i=-1;return{currentNode:function(n){return n!==i&&(a&&a.end===n&&r<t.length-1&&(r++,a=t[r]),a&&a.pos===n||o(n)),i=n,e.Debug.assert(!a||a.pos===n),a}};function o(e){return t=void 0,r=-1,a=void 0,void p(n,i,o);function i(n){return e>=n.pos&&e<n.end&&(p(n,i,o),!0)}function o(n){if(e>=n.pos&&e<n.end)for(var s=0;s<n.length;s++){var c=n[s];if(c){if(c.pos===e)return t=n,r=s,a=c,!0;if(c.pos<e&&e<c.end)return p(c,i,o),!0}}return!1}}}var f;n.updateSourceFile=t,function(e){e[e["Value"]=-1]="Value"}(f||(f={}))}(s||(s={})),e.isDeclarationFileName=b,e.processCommentPragmas=E,e.processPragmasIntoFields=T;var S=e.createMap();function x(e){if(S.has(e))return S.get(e);var n=new RegExp("(\\s"+e+"\\s*=\\s*)('|\")(.+?)\\2","im");return S.set(e,n),n}var C=/^\/\/\/\s*<(\S+)\s.*?\/>/im,A=/^\/\/\/?\s*@(\S+)\s*(.*)\s*$/im;function D(n,t,r){var a=2===t.kind&&C.exec(r);if(a){var i=a[1].toLowerCase(),o=e.commentPragmas[i];if(!o||!(1&o.kind))return;if(o.args){for(var s={},c=0,l=o.args;c<l.length;c++){var u=l[c],d=x(u.name),p=d.exec(r);if(!p&&!u.optional)return;if(p)if(u.captureSpan){var m=t.pos+p.index+p[1].length+p[2].length;s[u.name]={value:p[3],pos:m,end:m+p[3].length}}else s[u.name]=p[3]}n.push({name:i,args:{arguments:s,range:t}})}else n.push({name:i,args:{arguments:{},range:t}})}else{var f=2===t.kind&&A.exec(r);if(f)return L(n,t,2,f);if(3===t.kind){var _=/\s*@(\S+)\s*(.*)\s*$/gim,g=void 0;while(g=_.exec(r))L(n,t,4,g)}}}function L(n,t,r,a){if(a){var i=a[1].toLowerCase(),o=e.commentPragmas[i];if(o&&o.kind&r){var s=a[2],c=k(o,s);"fail"!==c&&n.push({name:i,args:{arguments:c,range:t}})}}}function k(n,t){if(!t)return{};if(!n.args)return{};for(var r=t.split(/\s+/),a={},i=0;i<n.args.length;i++){var o=n.args[i];if(!r[i]&&!o.optional)return"fail";if(o.captureSpan)return e.Debug.fail("Capture spans not yet implemented for non-xml pragmas");a[o.name]=r[i]}return a}function N(e,n){return e.kind===n.kind&&(72===e.kind?e.escapedText===n.escapedText:100===e.kind||e.name.escapedText===n.name.escapedText&&N(e.expression,n.expression))}e.tagNamesAreEquivalent=N}(u||(u={})),function(e){e.compileOnSaveCommandLineOption={name:"compileOnSave",type:"boolean"};var n,t,r=[["es5","lib.es5.d.ts"],["es6","lib.es2015.d.ts"],["es2015","lib.es2015.d.ts"],["es7","lib.es2016.d.ts"],["es2016","lib.es2016.d.ts"],["es2017","lib.es2017.d.ts"],["es2018","lib.es2018.d.ts"],["esnext","lib.esnext.d.ts"],["dom","lib.dom.d.ts"],["dom.iterable","lib.dom.iterable.d.ts"],["webworker","lib.webworker.d.ts"],["webworker.importscripts","lib.webworker.importscripts.d.ts"],["scripthost","lib.scripthost.d.ts"],["es2015.core","lib.es2015.core.d.ts"],["es2015.collection","lib.es2015.collection.d.ts"],["es2015.generator","lib.es2015.generator.d.ts"],["es2015.iterable","lib.es2015.iterable.d.ts"],["es2015.promise","lib.es2015.promise.d.ts"],["es2015.proxy","lib.es2015.proxy.d.ts"],["es2015.reflect","lib.es2015.reflect.d.ts"],["es2015.symbol","lib.es2015.symbol.d.ts"],["es2015.symbol.wellknown","lib.es2015.symbol.wellknown.d.ts"],["es2016.array.include","lib.es2016.array.include.d.ts"],["es2017.object","lib.es2017.object.d.ts"],["es2017.sharedmemory","lib.es2017.sharedmemory.d.ts"],["es2017.string","lib.es2017.string.d.ts"],["es2017.intl","lib.es2017.intl.d.ts"],["es2017.typedarrays","lib.es2017.typedarrays.d.ts"],["es2018.intl","lib.es2018.intl.d.ts"],["es2018.promise","lib.es2018.promise.d.ts"],["es2018.regexp","lib.es2018.regexp.d.ts"],["esnext.array","lib.esnext.array.d.ts"],["esnext.symbol","lib.esnext.symbol.d.ts"],["esnext.asynciterable","lib.esnext.asynciterable.d.ts"],["esnext.intl","lib.esnext.intl.d.ts"],["esnext.bigint","lib.esnext.bigint.d.ts"]];function a(e){return e&&void 0!==e.enableAutoDiscovery&&void 0===e.enable?{enable:e.enableAutoDiscovery,include:e.include||[],exclude:e.exclude||[]}:e}function i(){return n||(n=s(e.optionDeclarations))}function s(n){var t=e.createMap(),r=e.createMap();return e.forEach(n,function(e){t.set(e.name.toLowerCase(),e),e.shortName&&r.set(e.shortName,e.name)}),{optionNameMap:t,shortOptionNames:r}}function c(n){return l(n,e.createCompilerDiagnostic)}function l(n,t){var r=e.arrayFrom(n.type.keys()).map(function(e){return"'"+e+"'"}).join(", ");return t(e.Diagnostics.Argument_for_0_option_must_be_Colon_1,"--"+n.name,r)}function u(e,n,t){return pe(e,fe(n||""),t)}function d(n,t,r){if(void 0===t&&(t=""),t=fe(t),!e.startsWith(t,"-")){if(""===t)return[];var a=t.split(",");switch(n.element.type){case"number":return e.map(a,parseInt);case"string":return e.map(a,function(e){return e||""});default:return e.mapDefined(a,function(e){return u(n.element,e,r)})}}}function p(n,t,r,a){var i=t[0],o=t[1],s={},c=[],l=[];return p(r),{options:s,fileNames:c,errors:l};function p(t){var r=0;while(r<t.length){var a=t[r];if(r++,64===a.charCodeAt(0))m(a.slice(1));else if(45===a.charCodeAt(0)){var p=_(n,a.slice(45===a.charCodeAt(1)?2:1),!0);if(p)if(p.isTSConfigOnly)l.push(e.createCompilerDiagnostic(e.Diagnostics.Option_0_can_only_be_specified_in_tsconfig_json_file,p.name));else switch(t[r]||"boolean"===p.type||l.push(e.createCompilerDiagnostic(o,p.name)),p.type){case"number":s[p.name]=parseInt(t[r]),r++;break;case"boolean":var f=t[r];s[p.name]="false"!==f,"false"!==f&&"true"!==f||r++;break;case"string":s[p.name]=t[r]||"",r++;break;case"list":var g=d(p,t[r],l);s[p.name]=g||[],g&&r++;break;default:s[p.name]=u(p,t[r],l),r++;break}else l.push(e.createCompilerDiagnostic(i,a))}else c.push(a)}}function m(n){var t=a?a(n):e.sys.readFile(n);if(t){var r=[],i=0;while(1){while(i<t.length&&t.charCodeAt(i)<=32)i++;if(i>=t.length)break;var o=i;if(34===t.charCodeAt(o)){i++;while(i<t.length&&34!==t.charCodeAt(i))i++;i<t.length?(r.push(t.substring(o+1,i)),i++):l.push(e.createCompilerDiagnostic(e.Diagnostics.Unterminated_quoted_string_in_response_file_0,n))}else{while(t.charCodeAt(i)>32)i++;r.push(t.substring(o,i))}}p(r)}else l.push(e.createCompilerDiagnostic(e.Diagnostics.File_0_not_found,n))}}function m(n,t){return p(i,[e.Diagnostics.Unknown_compiler_option_0,e.Diagnostics.Compiler_option_0_expects_an_argument],n,t)}function f(e,n){return _(i,e,n)}function _(e,n,t){void 0===t&&(t=!1),n=n.toLowerCase();var r=e(),a=r.optionNameMap,i=r.shortOptionNames;if(t){var o=i.get(n);void 0!==o&&(n=o)}return a.get(n)}function g(n){var t,r=function(){return t||(t=s(e.buildOpts))},a=p(r,[e.Diagnostics.Unknown_build_option_0,e.Diagnostics.Build_option_0_requires_a_value_of_type_1],n),i=a.options,o=a.fileNames,c=a.errors,l=i;return 0===o.length&&o.push("."),l.clean&&l.force&&c.push(e.createCompilerDiagnostic(e.Diagnostics.Options_0_and_1_cannot_be_combined,"clean","force")),l.clean&&l.verbose&&c.push(e.createCompilerDiagnostic(e.Diagnostics.Options_0_and_1_cannot_be_combined,"clean","verbose")),l.clean&&l.watch&&c.push(e.createCompilerDiagnostic(e.Diagnostics.Options_0_and_1_cannot_be_combined,"clean","watch")),l.watch&&l.dry&&c.push(e.createCompilerDiagnostic(e.Diagnostics.Options_0_and_1_cannot_be_combined,"watch","dry")),{buildOptions:l,projects:o,errors:c}}function v(n){for(var t=[],r=1;r<arguments.length;r++)t[r-1]=arguments[r];var a=e.createCompilerDiagnostic.apply(void 0,arguments);return a.messageText}function y(){e.sys.write(v(e.Diagnostics.Version_0,e.version)+e.sys.newLine)}function h(n,t){void 0===t&&(t="");var r=[],a=v(e.Diagnostics.Syntax_Colon_0,"").length,i=v(e.Diagnostics.Examples_Colon_0,"").length,o=Math.max(a,i),s=I(o-a);s+="tsc "+t+"["+v(e.Diagnostics.options)+"] ["+v(e.Diagnostics.file)+"...]",r.push(v(e.Diagnostics.Syntax_Colon_0,s)),r.push(e.sys.newLine+e.sys.newLine);var c=I(o);r.push(v(e.Diagnostics.Examples_Colon_0,I(o-i)+"tsc hello.ts")+e.sys.newLine),r.push(c+"tsc --outFile file.js file.ts"+e.sys.newLine),r.push(c+"tsc @args.txt"+e.sys.newLine),r.push(c+"tsc --build tsconfig.json"+e.sys.newLine),r.push(e.sys.newLine),r.push(v(e.Diagnostics.Options_Colon)+e.sys.newLine),o=0;for(var l=[],u=[],d=e.createMap(),p=0,m=n;p<m.length;p++){var f=m[p];if(f.description){var _=" ";f.shortName&&(_+="-"+f.shortName,_+=N(f),_+=", "),_+="--"+f.name,_+=N(f),l.push(_);var g=void 0;if("lib"===f.name){g=v(f.description);var y=f.element,h=y.type;d.set(g,e.arrayFrom(h.keys()).map(function(e){return"'"+e+"'"}))}else g=v(f.description);u.push(g),o=Math.max(_.length,o)}}var b=" @<"+v(e.Diagnostics.file)+">";l.push(b),u.push(v(e.Diagnostics.Insert_command_line_options_and_files_from_a_file)),o=Math.max(b.length,o);for(var E=0;E<l.length;E++){var T=l[E],S=(g=u[E],d.get(g));if(r.push(T+I(o-T.length+2)+g+e.sys.newLine),S){r.push(I(o+4));for(var x=0,C=S;x<C.length;x++){var A=C[x];r.push(A+" ")}r.push(e.sys.newLine)}}for(var D=0,L=r;D<L.length;D++){var k=L[D];e.sys.write(k)}return;function N(e){return void 0!==e.paramType?" "+v(e.paramType):""}function I(e){return Array(e+1).join(" ")}}function b(n,t,r){var a;try{a=r.readFile(n)}catch(c){var i=e.createCompilerDiagnostic(e.Diagnostics.Cannot_read_file_0_Colon_1,n,c.message);return void r.onUnRecoverableConfigFileDiagnostic(i)}if(a){var o=e.parseJsonText(n,a),s=r.getCurrentDirectory();return o.path=e.toPath(n,s,e.createGetCanonicalFileName(r.useCaseSensitiveFileNames)),o.resolvedPath=o.path,o.originalFileName=o.fileName,V(o,r,e.getNormalizedAbsolutePath(e.getDirectoryPath(n),s),t,e.getNormalizedAbsolutePath(n,s))}i=e.createCompilerDiagnostic(e.Diagnostics.File_0_not_found,n);r.onUnRecoverableConfigFileDiagnostic(i)}function E(n,t){var r=x(n,t);return e.isString(r)?T(n,r):{config:{},error:r}}function T(n,t){var r=e.parseJsonText(n,t);return{config:D(r,r.parseDiagnostics),error:r.parseDiagnostics.length?r.parseDiagnostics[0]:void 0}}function S(n,t){var r=x(n,t);return e.isString(r)?e.parseJsonText(n,r):{parseDiagnostics:[r]}}function x(n,t){var r;try{r=t(n)}catch(a){return e.createCompilerDiagnostic(e.Diagnostics.Cannot_read_file_0_Colon_1,n,a.message)}return void 0===r?e.createCompilerDiagnostic(e.Diagnostics.The_specified_path_does_not_exist_Colon_0,n):r}function C(n){return e.arrayToMap(n,function(e){return e.name})}function A(){return void 0===t&&(t={name:void 0,type:"object",elementOptions:C([{name:"compilerOptions",type:"object",elementOptions:C(e.optionDeclarations),extraKeyDiagnosticMessage:e.Diagnostics.Unknown_compiler_option_0},{name:"typingOptions",type:"object",elementOptions:C(e.typeAcquisitionDeclarations),extraKeyDiagnosticMessage:e.Diagnostics.Unknown_type_acquisition_option_0},{name:"typeAcquisition",type:"object",elementOptions:C(e.typeAcquisitionDeclarations),extraKeyDiagnosticMessage:e.Diagnostics.Unknown_type_acquisition_option_0},{name:"extends",type:"string"},{name:"references",type:"list",element:{name:"references",type:"object"}},{name:"files",type:"list",element:{name:"files",type:"string"}},{name:"include",type:"list",element:{name:"include",type:"string"}},{name:"exclude",type:"list",element:{name:"exclude",type:"string"}},e.compileOnSaveCommandLineOption])}),t}function D(e,n){return L(e,n,!0,void 0,void 0)}function L(n,t,r,a,i){return n.statements.length?u(n.statements[0].expression,a):r?{}:void 0;function o(e){return a&&a.elementOptions===e}function s(a,s,c,l){for(var p=r?{}:void 0,m=0,f=a.properties;m<f.length;m++){var _=f[m];if(275===_.kind){_.questionToken&&t.push(e.createDiagnosticForNodeInSourceFile(n,_.questionToken,e.Diagnostics._0_can_only_be_used_in_a_ts_file,"?")),d(_.name)||t.push(e.createDiagnosticForNodeInSourceFile(n,_.name,e.Diagnostics.String_literal_with_double_quotes_expected));var g=e.getTextOfPropertyName(_.name),v=g&&e.unescapeLeadingUnderscores(g),y=v&&s?s.get(v):void 0;v&&c&&!y&&t.push(e.createDiagnosticForNodeInSourceFile(n,_.name,c,v));var h=u(_.initializer,y);if("undefined"!==typeof v&&(r&&(p[v]=h),i&&(l||o(s)))){var b=N(y,h);l?b&&i.onSetValidOptionKeyValueInParent(l,y,h):o(s)&&(b?i.onSetValidOptionKeyValueInRoot(v,_.name,h,_.initializer):y||i.onSetUnknownOptionKeyValueInRoot(v,_.name,h,_.initializer))}}else t.push(e.createDiagnosticForNodeInSourceFile(n,_,e.Diagnostics.Property_assignment_expected))}return p}function c(n,t){return r?e.filter(n.map(function(e){return u(e,t)}),function(e){return void 0!==e}):n.forEach(function(e){return u(e,t)})}function u(r,a){switch(r.kind){case 102:return g(a&&"boolean"!==a.type),!0;case 87:return g(a&&"boolean"!==a.type),!1;case 96:return g(a&&"extends"===a.name),null;case 10:d(r)||t.push(e.createDiagnosticForNodeInSourceFile(n,r,e.Diagnostics.String_literal_with_double_quotes_expected)),g(a&&e.isString(a.type)&&"string"!==a.type);var i=r.text;if(a&&!e.isString(a.type)){var o=a;o.type.has(i.toLowerCase())||t.push(l(o,function(t,a,i){return e.createDiagnosticForNodeInSourceFile(n,r,t,a,i)}))}return i;case 8:return g(a&&"number"!==a.type),Number(r.text);case 202:if(39!==r.operator||8!==r.operand.kind)break;return g(a&&"number"!==a.type),-Number(r.operand.text);case 188:g(a&&"object"!==a.type);var u=r;if(a){var p=a,m=p.elementOptions,f=p.extraKeyDiagnosticMessage,_=p.name;return s(u,m,f,_)}return s(u,void 0,void 0,void 0);case 187:return g(a&&"list"!==a.type),c(r.elements,a&&a.element)}return void(a?g(!0):t.push(e.createDiagnosticForNodeInSourceFile(n,r,e.Diagnostics.Property_value_can_only_be_string_literal_numeric_literal_true_false_null_object_literal_or_array_literal)));function g(i){i&&t.push(e.createDiagnosticForNodeInSourceFile(n,r,e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,a.name,k(a)))}}function d(t){return e.isStringLiteral(t)&&e.isStringDoubleQuoted(t,n)}}function k(n){return"list"===n.type?"Array":e.isString(n.type)?n.type:"string"}function N(n,t){if(n){if(K(t))return!0;if("list"===n.type)return e.isArray(t);var r=e.isString(n.type)?n.type:"string";return typeof t===r}return!1}function I(n,t,r){var a=e.createGetCanonicalFileName(r.useCaseSensitiveFileNames),i=e.map(e.filter(n.fileNames,n.configFileSpecs&&n.configFileSpecs.validatedIncludeSpecs?M(t,n.configFileSpecs.validatedIncludeSpecs,n.configFileSpecs.validatedExcludeSpecs):function(e){return!0}),function(n){return e.getRelativePathFromFile(e.getNormalizedAbsolutePath(t,r.getCurrentDirectory()),e.getNormalizedAbsolutePath(n,r.getCurrentDirectory()),a)}),s=R(n.options,{configFilePath:e.getNormalizedAbsolutePath(t,r.getCurrentDirectory()),useCaseSensitiveFileNames:r.useCaseSensitiveFileNames}),c=o({compilerOptions:o({},e.arrayFrom(s.entries()).reduce(function(e,n){var t;return o({},e,(t={},t[n[0]]=n[1],t))},{}),{showConfig:void 0,configFile:void 0,configFilePath:void 0,help:void 0,init:void 0,listFiles:void 0,listEmittedFiles:void 0,project:void 0,build:void 0,version:void 0}),references:e.map(n.projectReferences,function(e){return o({},e,{path:e.originalPath,originalPath:void 0})}),files:e.length(i)?i:void 0},n.configFileSpecs?{include:w(n.configFileSpecs.validatedIncludeSpecs),exclude:n.configFileSpecs.validatedExcludeSpecs}:{},{compilerOnSave:!!n.compileOnSave||void 0});return c}function w(n){if(e.length(n)){if(1!==e.length(n))return n;if("**/*"!==n[0])return n}}function M(n,t,r){if(!t)return function(e){return!0};var a=e.getFileMatcherPatterns(n,r,t,e.sys.useCaseSensitiveFileNames,e.sys.getCurrentDirectory()),i=a.excludePattern&&e.getRegexFromPattern(a.excludePattern,e.sys.useCaseSensitiveFileNames),o=a.includeFilePattern&&e.getRegexFromPattern(a.includeFilePattern,e.sys.useCaseSensitiveFileNames);return o?i?function(e){return!(o.test(e)&&!i.test(e))}:function(e){return!o.test(e)}:i?function(e){return i.test(e)}:function(e){return!0}}function O(e){return"string"===e.type||"number"===e.type||"boolean"===e.type||"object"===e.type?void 0:"list"===e.type?O(e.element):e.type}function P(n,t){return e.forEachEntry(t,function(e,t){if(e===n)return t})}function R(n,t){var r=e.createMap(),a=i().optionNameMap,o=t&&e.createGetCanonicalFileName(t.useCaseSensitiveFileNames),s=function(i){if(e.hasProperty(n,i)){if(a.has(i)&&a.get(i).category===e.Diagnostics.Command_line_Options)return"continue";var s=n[i],c=a.get(i.toLowerCase());if(c){var l=O(c);l?"list"===c.type?r.set(i,s.map(function(e){return P(e,l)})):r.set(i,P(s,l)):t&&c.isFilePath?r.set(i,e.getRelativePathFromFile(t.configFilePath,e.getNormalizedAbsolutePath(s,e.getDirectoryPath(t.configFilePath)),o)):r.set(i,s)}}};for(var c in n)s(c);return r}function F(n,t,r){var a=e.extend(n,e.defaultInitCompilerOptions),i=R(a);return l();function o(e){switch(e.type){case"number":return 1;case"boolean":return!0;case"string":return e.isFilePath?"./":"";case"list":return[];case"object":return{};default:return e.type.keys().next().value}}function s(e){return Array(e+1).join(" ")}function c(n){var t=n.category,r=n.name;return void 0!==t&&t!==e.Diagnostics.Command_line_Options&&(t!==e.Diagnostics.Advanced_Options||i.has(r))}function l(){for(var n=e.createMultiMap(),a=0,l=e.optionDeclarations;a<l.length;a++){var u=l[a],d=u.category;c(u)&&n.add(e.getLocaleSpecificMessage(d),u)}var p=0,m=0,f=[],_=[];n.forEach(function(n,t){0!==f.length&&(f.push(""),_.push("")),f.push("/* "+t+" */"),_.push("");for(var r=0,a=n;r<a.length;r++){var s=a[r],c=void 0;c=i.has(s.name)?'"'+s.name+'": '+JSON.stringify(i.get(s.name))+((m+=1)===i.size?"":","):'// "'+s.name+'": '+JSON.stringify(o(s))+",",f.push(c),_.push("/* "+(s.description&&e.getLocaleSpecificMessage(s.description)||s.name)+" */"),p=Math.max(c.length,p)}});var g=s(2),v=[];v.push("{"),v.push(g+'"compilerOptions": {');for(var y=0;y<f.length;y++){var h=f[y],b=_[y];v.push(h&&""+g+g+h+(b&&s(p-h.length+2)+b))}if(t.length){v.push(g+"},"),v.push(g+'"files": [');for(y=0;y<t.length;y++)v.push(""+g+g+JSON.stringify(t[y])+(y===t.length-1?"":","));v.push(g+"]")}else v.push(g+"}");return v.push("}"),v.join(r)+r}}function G(e,n,t,r,a,i,o){return U(e,void 0,n,t,r,a,i,o)}function V(e,n,t,r,a,i,o){return U(void 0,e,n,t,r,a,i,o)}function B(e,n){n&&Object.defineProperty(e,"configFile",{enumerable:!1,writable:!1,value:n})}function K(e){return void 0===e||null===e}function j(n,t){return e.getDirectoryPath(e.getNormalizedAbsolutePath(n,t))}function U(n,t,r,a,i,o,s,c){void 0===i&&(i={}),void 0===s&&(s=[]),void 0===c&&(c=[]),e.Debug.assert(void 0===n&&void 0!==t||void 0!==n&&void 0===t);var l,u=[],d=Y(n,t,r,a,o,s,u),p=d.raw,m=e.extend(i,d.options||{});m.configFilePath=o&&e.normalizeSlashes(o),B(m,t);var f=y(),_=f.fileNames,g=f.wildcardDirectories,v=f.spec;return{options:m,fileNames:_,projectReferences:l,typeAcquisition:d.typeAcquisition||oe(),raw:p,errors:u,wildcardDirectories:g,compileOnSave:!!p.compileOnSave,configFileSpecs:v};function y(){var n,i,d;if(e.hasProperty(p,"files")&&!K(p.files))if(e.isArray(p.files)){n=p.files;var f=e.hasProperty(p,"references")&&!K(p.references),_=!f||0===p.references.length,g=e.hasProperty(p,"extends");if(0===n.length&&_&&!g)if(t){var v=o||"tsconfig.json",y=e.Diagnostics.The_files_list_in_config_file_0_is_empty,b=e.firstDefined(e.getTsConfigPropArray(t,"files"),function(e){return e.initializer}),E=b?e.createDiagnosticForNodeInSourceFile(t,b,y,v):e.createCompilerDiagnostic(y,v);u.push(E)}else h(e.Diagnostics.The_files_list_in_config_file_0_is_empty,o||"tsconfig.json")}else h(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,"files","Array");if(e.hasProperty(p,"include")&&!K(p.include)&&(e.isArray(p.include)?i=p.include:h(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,"include","Array")),e.hasProperty(p,"exclude")&&!K(p.exclude))e.isArray(p.exclude)?d=p.exclude:h(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,"exclude","Array");else if(p.compilerOptions){var T=p.compilerOptions.outDir,S=p.compilerOptions.declarationDir;(T||S)&&(d=[T,S].filter(function(e){return!!e}))}void 0===n&&void 0===i&&(i=["**/*"]);var x=he(n,i,d,o?j(o,a):a,m,r,u,c,t);if(q(x,z(p),s)&&u.push(W(x.spec,o)),e.hasProperty(p,"references")&&!K(p.references))if(e.isArray(p.references))for(var C=0,A=p.references;C<A.length;C++){var D=A[C];"string"!==typeof D.path?h(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,"reference.path","string"):(l||(l=[])).push({path:e.getNormalizedAbsolutePath(D.path,a),originalPath:D.path,prepend:D.prepend,circular:D.circular})}else h(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,"references","Array");return x}function h(n,r,a){t||u.push(e.createCompilerDiagnostic(n,r,a))}}function H(n){return n.code===e.Diagnostics.No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2.code}function W(n,t){var r=n.includeSpecs,a=n.excludeSpecs;return e.createCompilerDiagnostic(e.Diagnostics.No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2,t||"tsconfig.json",JSON.stringify(r||[]),JSON.stringify(a||[]))}function q(e,n,t){return 0===e.fileNames.length&&n&&(!t||0===t.length)}function z(n){return!e.hasProperty(n,"files")&&!e.hasProperty(n,"references")}function J(n,t,r,a,i){var o=a.length;return q(n,i)?a.push(W(r,t)):e.filterMutate(a,function(e){return!H(e)}),o!==a.length}function X(e){return!!e.options}function Y(n,t,r,a,i,o,s){a=e.normalizeSlashes(a);var c=e.getNormalizedAbsolutePath(i||"",a);if(o.indexOf(c)>=0)return s.push(e.createCompilerDiagnostic(e.Diagnostics.Circularity_detected_while_resolving_configuration_Colon_0,o.concat([c]).join(" -> "))),{raw:n||D(t,s)};var l=n?Q(n,r,a,i,s):Z(t,r,a,i,s);if(l.extendedConfigPath){o=o.concat([c]);var u=ee(t,l.extendedConfigPath,r,a,o,s);if(u&&X(u)){var d=u.raw,p=l.raw,m=function(e){var n=p[e]||d[e];n&&(p[e]=n)};m("include"),m("exclude"),m("files"),void 0===p.compileOnSave&&(p.compileOnSave=d.compileOnSave),l.options=e.assign({},u.options,l.options)}}return l}function Q(n,t,r,a,i){e.hasProperty(n,"excludes")&&i.push(e.createCompilerDiagnostic(e.Diagnostics.Unknown_option_excludes_Did_you_mean_exclude));var o,s=ie(n.compilerOptions,r,i,a),c=se(n.typeAcquisition||n.typingOptions,r,i,a);if(n.compileOnSave=ne(n,r,i),n.extends)if(e.isString(n.extends)){var l=a?j(a,r):r;o=$(n.extends,t,l,i,e.createCompilerDiagnostic)}else i.push(e.createCompilerDiagnostic(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,"extends","string"));return{raw:n,options:s,typeAcquisition:c,extendedConfigPath:o}}function Z(n,t,r,a,i){var o,s,c,l=ae(a),u={onSetValidOptionKeyValueInParent:function(n,t,i){e.Debug.assert("compilerOptions"===n||"typeAcquisition"===n||"typingOptions"===n);var c="compilerOptions"===n?l:"typeAcquisition"===n?o||(o=oe(a)):s||(s=oe(a));c[t.name]=ue(t,r,i)},onSetValidOptionKeyValueInRoot:function(o,s,l,u){switch(o){case"extends":var d=a?j(a,r):r;return void(c=$(l,t,d,i,function(t,r){return e.createDiagnosticForNodeInSourceFile(n,u,t,r)}))}},onSetUnknownOptionKeyValueInRoot:function(t,r,a,o){"excludes"===t&&i.push(e.createDiagnosticForNodeInSourceFile(n,r,e.Diagnostics.Unknown_option_excludes_Did_you_mean_exclude))}},d=L(n,i,!0,A(),u);return o||(o=s?void 0!==s.enableAutoDiscovery?{enable:s.enableAutoDiscovery,include:s.include,exclude:s.exclude}:s:oe(a)),{raw:d,options:l,typeAcquisition:o,extendedConfigPath:c}}function $(n,t,r,a,i){if(n=e.normalizeSlashes(n),e.isRootedDiskPath(n)||e.startsWith(n,"./")||e.startsWith(n,"../")){var o=e.getNormalizedAbsolutePath(n,r);return t.fileExists(o)||e.endsWith(o,".json")||(o+=".json",t.fileExists(o))?o:void a.push(i(e.Diagnostics.File_0_does_not_exist,n))}var s=e.nodeModuleNameResolver(n,e.combinePaths(r,"tsconfig.json"),{moduleResolution:e.ModuleResolutionKind.NodeJs},t,void 0,void 0,!0);if(s.resolvedModule)return s.resolvedModule.resolvedFileName;a.push(i(e.Diagnostics.File_0_does_not_exist,n))}function ee(n,t,r,a,i,o){var s,c=S(t,function(e){return r.readFile(e)});if(n&&(n.extendedSourceFiles=[c.fileName]),!c.parseDiagnostics.length){var l=e.getDirectoryPath(t),u=Y(void 0,c,r,l,e.getBaseFileName(t),i,o);if(n&&c.extendedSourceFiles&&(s=n.extendedSourceFiles).push.apply(s,c.extendedSourceFiles),X(u)){var d=e.convertToRelativePath(l,a,e.identity),p=function(n){return e.isRootedDiskPath(n)?n:e.combinePaths(d,n)},m=function(n){f[n]&&(f[n]=e.map(f[n],p))},f=u.raw;m("include"),m("exclude"),m("files")}return u}o.push.apply(o,c.parseDiagnostics)}function ne(n,t,r){if(!e.hasProperty(n,e.compileOnSaveCommandLineOption.name))return!1;var a=le(e.compileOnSaveCommandLineOption,n.compileOnSave,t,r);return"boolean"===typeof a&&a}function te(e,n,t){var r=[],a=ie(e,n,r,t);return{options:a,errors:r}}function re(e,n,t){var r=[],a=se(e,n,r,t);return{options:a,errors:r}}function ae(n){var t=n&&"jsconfig.json"===e.getBaseFileName(n)?{allowJs:!0,maxNodeModuleJsDepth:2,allowSyntheticDefaultImports:!0,skipLibCheck:!0,noEmit:!0}:{};return t}function ie(n,t,r,a){var i=ae(a);return ce(e.optionDeclarations,n,t,i,e.Diagnostics.Unknown_compiler_option_0,r),a&&(i.configFilePath=e.normalizeSlashes(a)),i}function oe(n){return{enable:!!n&&"jsconfig.json"===e.getBaseFileName(n),include:[],exclude:[]}}function se(n,t,r,i){var o=oe(i),s=a(n);return ce(e.typeAcquisitionDeclarations,s,t,o,e.Diagnostics.Unknown_type_acquisition_option_0,r),o}function ce(n,t,r,a,i,o){if(t){var s=C(n);for(var c in t){var l=s.get(c);l?a[l.name]=le(l,t[c],r,o):o.push(e.createCompilerDiagnostic(i,c))}}}function le(n,t,r,a){if(N(n,t)){var i=n.type;return"list"===i&&e.isArray(t)?me(n,t,r,a):e.isString(i)?de(n,r,t):pe(n,t,a)}a.push(e.createCompilerDiagnostic(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,n.name,k(n)))}function ue(n,t,r){if(!K(r)){if("list"===n.type){var a=n;return a.element.isFilePath||!e.isString(a.element.type)?e.filter(e.map(r,function(e){return ue(a.element,t,e)}),function(e){return!!e}):r}return e.isString(n.type)?de(n,t,r):n.type.get(e.isString(r)?r.toLowerCase():r)}}function de(n,t,r){return n.isFilePath&&(r=e.normalizePath(e.combinePaths(t,r)),""===r&&(r=".")),r}function pe(e,n,t){if(!K(n)){var r=n.toLowerCase(),a=e.type.get(r);if(void 0!==a)return a;t.push(c(e))}}function me(n,t,r,a){return e.filter(e.map(t,function(e){return le(n.element,e,r,a)}),function(e){return!!e})}function fe(e){return"function"===typeof e.trim?e.trim():e.replace(/^[\s]+|[\s]+$/g,"")}e.libs=r.map(function(e){return e[0]}),e.libMap=e.createMapFromEntries(r),e.commonOptionsWithBuild=[{name:"help",shortName:"h",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Print_this_message},{name:"help",shortName:"?",type:"boolean"},{name:"watch",shortName:"w",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Watch_input_files},{name:"preserveWatchOutput",type:"boolean",showInSimplifiedHelpView:!1,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Whether_to_keep_outdated_console_output_in_watch_mode_instead_of_clearing_the_screen},{name:"listFiles",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Print_names_of_files_part_of_the_compilation},{name:"listEmittedFiles",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Print_names_of_generated_files_part_of_the_compilation},{name:"pretty",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Stylize_errors_and_messages_using_color_and_context_experimental},{name:"traceResolution",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Enable_tracing_of_the_name_resolution_process},{name:"diagnostics",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Show_diagnostic_information},{name:"extendedDiagnostics",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Show_verbose_diagnostic_information}],e.optionDeclarations=e.commonOptionsWithBuild.concat([{name:"all",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Show_all_compiler_options},{name:"version",shortName:"v",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Print_the_compiler_s_version},{name:"init",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Initializes_a_TypeScript_project_and_creates_a_tsconfig_json_file},{name:"project",shortName:"p",type:"string",isFilePath:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,paramType:e.Diagnostics.FILE_OR_DIRECTORY,description:e.Diagnostics.Compile_the_project_given_the_path_to_its_configuration_file_or_to_a_folder_with_a_tsconfig_json},{name:"build",type:"boolean",shortName:"b",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Build_one_or_more_projects_and_their_dependencies_if_out_of_date},{name:"showConfig",type:"boolean",category:e.Diagnostics.Command_line_Options,isCommandLineOnly:!0,description:e.Diagnostics.Print_the_final_configuration_instead_of_building},{name:"target",shortName:"t",type:e.createMapFromTemplate({es3:0,es5:1,es6:2,es2015:2,es2016:3,es2017:4,es2018:5,esnext:6}),affectsSourceFile:!0,affectsModuleResolution:!0,paramType:e.Diagnostics.VERSION,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Specify_ECMAScript_target_version_Colon_ES3_default_ES5_ES2015_ES2016_ES2017_ES2018_or_ESNEXT},{name:"module",shortName:"m",type:e.createMapFromTemplate({none:e.ModuleKind.None,commonjs:e.ModuleKind.CommonJS,amd:e.ModuleKind.AMD,system:e.ModuleKind.System,umd:e.ModuleKind.UMD,es6:e.ModuleKind.ES2015,es2015:e.ModuleKind.ES2015,esnext:e.ModuleKind.ESNext}),affectsModuleResolution:!0,paramType:e.Diagnostics.KIND,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Specify_module_code_generation_Colon_none_commonjs_amd_system_umd_es2015_or_ESNext},{name:"lib",type:"list",element:{name:"lib",type:e.libMap},affectsModuleResolution:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Specify_library_files_to_be_included_in_the_compilation},{name:"allowJs",type:"boolean",affectsModuleResolution:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Allow_javascript_files_to_be_compiled},{name:"checkJs",type:"boolean",category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Report_errors_in_js_files},{name:"jsx",type:e.createMapFromTemplate({preserve:1,"react-native":3,react:2}),affectsSourceFile:!0,paramType:e.Diagnostics.KIND,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Specify_JSX_code_generation_Colon_preserve_react_native_or_react},{name:"declaration",shortName:"d",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Generates_corresponding_d_ts_file},{name:"declarationMap",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Generates_a_sourcemap_for_each_corresponding_d_ts_file},{name:"emitDeclarationOnly",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Only_emit_d_ts_declaration_files},{name:"sourceMap",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Generates_corresponding_map_file},{name:"outFile",type:"string",isFilePath:!0,paramType:e.Diagnostics.FILE,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Concatenate_and_emit_output_to_single_file},{name:"outDir",type:"string",isFilePath:!0,paramType:e.Diagnostics.DIRECTORY,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Redirect_output_structure_to_the_directory},{name:"rootDir",type:"string",isFilePath:!0,paramType:e.Diagnostics.LOCATION,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Specify_the_root_directory_of_input_files_Use_to_control_the_output_directory_structure_with_outDir},{name:"composite",type:"boolean",isTSConfigOnly:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Enable_project_compilation},{name:"removeComments",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Do_not_emit_comments_to_output},{name:"noEmit",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Do_not_emit_outputs},{name:"importHelpers",type:"boolean",category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Import_emit_helpers_from_tslib},{name:"downlevelIteration",type:"boolean",category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Provide_full_support_for_iterables_in_for_of_spread_and_destructuring_when_targeting_ES5_or_ES3},{name:"isolatedModules",type:"boolean",category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Transpile_each_file_as_a_separate_module_similar_to_ts_transpileModule},{name:"strict",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Enable_all_strict_type_checking_options},{name:"noImplicitAny",type:"boolean",affectsSemanticDiagnostics:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Raise_error_on_expressions_and_declarations_with_an_implied_any_type},{name:"strictNullChecks",type:"boolean",affectsSemanticDiagnostics:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Enable_strict_null_checks},{name:"strictFunctionTypes",type:"boolean",affectsSemanticDiagnostics:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Enable_strict_checking_of_function_types},{name:"strictBindCallApply",type:"boolean",strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Enable_strict_bind_call_and_apply_methods_on_functions},{name:"strictPropertyInitialization",type:"boolean",affectsSemanticDiagnostics:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Enable_strict_checking_of_property_initialization_in_classes},{name:"noImplicitThis",type:"boolean",affectsSemanticDiagnostics:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Raise_error_on_this_expressions_with_an_implied_any_type},{name:"alwaysStrict",type:"boolean",affectsSourceFile:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Parse_in_strict_mode_and_emit_use_strict_for_each_source_file},{name:"noUnusedLocals",type:"boolean",affectsSemanticDiagnostics:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Additional_Checks,description:e.Diagnostics.Report_errors_on_unused_locals},{name:"noUnusedParameters",type:"boolean",affectsSemanticDiagnostics:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Additional_Checks,description:e.Diagnostics.Report_errors_on_unused_parameters},{name:"noImplicitReturns",type:"boolean",affectsSemanticDiagnostics:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Additional_Checks,description:e.Diagnostics.Report_error_when_not_all_code_paths_in_function_return_a_value},{name:"noFallthroughCasesInSwitch",type:"boolean",affectsBindDiagnostics:!0,affectsSemanticDiagnostics:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Additional_Checks,description:e.Diagnostics.Report_errors_for_fallthrough_cases_in_switch_statement},{name:"moduleResolution",type:e.createMapFromTemplate({node:e.ModuleResolutionKind.NodeJs,classic:e.ModuleResolutionKind.Classic}),affectsModuleResolution:!0,paramType:e.Diagnostics.STRATEGY,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Specify_module_resolution_strategy_Colon_node_Node_js_or_classic_TypeScript_pre_1_6},{name:"baseUrl",type:"string",affectsModuleResolution:!0,isFilePath:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Base_directory_to_resolve_non_absolute_module_names},{name:"paths",type:"object",affectsModuleResolution:!0,isTSConfigOnly:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.A_series_of_entries_which_re_map_imports_to_lookup_locations_relative_to_the_baseUrl},{name:"rootDirs",type:"list",isTSConfigOnly:!0,element:{name:"rootDirs",type:"string",isFilePath:!0},affectsModuleResolution:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.List_of_root_folders_whose_combined_content_represents_the_structure_of_the_project_at_runtime},{name:"typeRoots",type:"list",element:{name:"typeRoots",type:"string",isFilePath:!0},affectsModuleResolution:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.List_of_folders_to_include_type_definitions_from},{name:"types",type:"list",element:{name:"types",type:"string"},affectsModuleResolution:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Type_declaration_files_to_be_included_in_compilation},{name:"allowSyntheticDefaultImports",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Allow_default_imports_from_modules_with_no_default_export_This_does_not_affect_code_emit_just_typechecking},{name:"esModuleInterop",type:"boolean",affectsSemanticDiagnostics:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Enables_emit_interoperability_between_CommonJS_and_ES_Modules_via_creation_of_namespace_objects_for_all_imports_Implies_allowSyntheticDefaultImports},{name:"preserveSymlinks",type:"boolean",category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Do_not_resolve_the_real_path_of_symlinks},{name:"sourceRoot",type:"string",paramType:e.Diagnostics.LOCATION,category:e.Diagnostics.Source_Map_Options,description:e.Diagnostics.Specify_the_location_where_debugger_should_locate_TypeScript_files_instead_of_source_locations},{name:"mapRoot",type:"string",paramType:e.Diagnostics.LOCATION,category:e.Diagnostics.Source_Map_Options,description:e.Diagnostics.Specify_the_location_where_debugger_should_locate_map_files_instead_of_generated_locations},{name:"inlineSourceMap",type:"boolean",category:e.Diagnostics.Source_Map_Options,description:e.Diagnostics.Emit_a_single_file_with_source_maps_instead_of_having_a_separate_file},{name:"inlineSources",type:"boolean",category:e.Diagnostics.Source_Map_Options,description:e.Diagnostics.Emit_the_source_alongside_the_sourcemaps_within_a_single_file_requires_inlineSourceMap_or_sourceMap_to_be_set},{name:"experimentalDecorators",type:"boolean",category:e.Diagnostics.Experimental_Options,description:e.Diagnostics.Enables_experimental_support_for_ES7_decorators},{name:"emitDecoratorMetadata",type:"boolean",category:e.Diagnostics.Experimental_Options,description:e.Diagnostics.Enables_experimental_support_for_emitting_type_metadata_for_decorators},{name:"jsxFactory",type:"string",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Specify_the_JSX_factory_function_to_use_when_targeting_react_JSX_emit_e_g_React_createElement_or_h},{name:"resolveJsonModule",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Include_modules_imported_with_json_extension},{name:"out",type:"string",isFilePath:!1,category:e.Diagnostics.Advanced_Options,paramType:e.Diagnostics.FILE,description:e.Diagnostics.Deprecated_Use_outFile_instead_Concatenate_and_emit_output_to_single_file},{name:"reactNamespace",type:"string",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Deprecated_Use_jsxFactory_instead_Specify_the_object_invoked_for_createElement_when_targeting_react_JSX_emit},{name:"skipDefaultLibCheck",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Deprecated_Use_skipLibCheck_instead_Skip_type_checking_of_default_library_declaration_files},{name:"charset",type:"string",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.The_character_set_of_the_input_files},{name:"emitBOM",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Emit_a_UTF_8_Byte_Order_Mark_BOM_in_the_beginning_of_output_files},{name:"locale",type:"string",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.The_locale_used_when_displaying_messages_to_the_user_e_g_en_us},{name:"newLine",type:e.createMapFromTemplate({crlf:0,lf:1}),paramType:e.Diagnostics.NEWLINE,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Specify_the_end_of_line_sequence_to_be_used_when_emitting_files_Colon_CRLF_dos_or_LF_unix},{name:"noErrorTruncation",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_truncate_error_messages},{name:"noLib",type:"boolean",affectsModuleResolution:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_include_the_default_library_file_lib_d_ts},{name:"noResolve",type:"boolean",affectsModuleResolution:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_add_triple_slash_references_or_imported_modules_to_the_list_of_compiled_files},{name:"stripInternal",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_emit_declarations_for_code_that_has_an_internal_annotation},{name:"disableSizeLimit",type:"boolean",affectsSourceFile:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Disable_size_limitations_on_JavaScript_projects},{name:"noImplicitUseStrict",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_emit_use_strict_directives_in_module_output},{name:"noEmitHelpers",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_generate_custom_helper_functions_like_extends_in_compiled_output},{name:"noEmitOnError",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_emit_outputs_if_any_errors_were_reported},{name:"preserveConstEnums",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_erase_const_enum_declarations_in_generated_code},{name:"declarationDir",type:"string",isFilePath:!0,paramType:e.Diagnostics.DIRECTORY,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Output_directory_for_generated_declaration_files},{name:"skipLibCheck",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Skip_type_checking_of_declaration_files},{name:"allowUnusedLabels",type:"boolean",affectsBindDiagnostics:!0,affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_report_errors_on_unused_labels},{name:"allowUnreachableCode",type:"boolean",affectsBindDiagnostics:!0,affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_report_errors_on_unreachable_code},{name:"suppressExcessPropertyErrors",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Suppress_excess_property_checks_for_object_literals},{name:"suppressImplicitAnyIndexErrors",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Suppress_noImplicitAny_errors_for_indexing_objects_lacking_index_signatures},{name:"forceConsistentCasingInFileNames",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Disallow_inconsistently_cased_references_to_the_same_file},{name:"maxNodeModuleJsDepth",type:"number",affectsModuleResolution:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.The_maximum_dependency_depth_to_search_under_node_modules_and_load_JavaScript_files},{name:"noStrictGenericChecks",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Disable_strict_checking_of_generic_signatures_in_function_types},{name:"keyofStringsOnly",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Resolve_keyof_to_string_valued_property_names_only_no_numbers_or_symbols},{name:"plugins",type:"list",isTSConfigOnly:!0,element:{name:"plugin",type:"object"},description:e.Diagnostics.List_of_language_service_plugins}]),e.semanticDiagnosticsOptionDeclarations=e.optionDeclarations.filter(function(e){return!!e.affectsSemanticDiagnostics}),e.moduleResolutionOptionDeclarations=e.optionDeclarations.filter(function(e){return!!e.affectsModuleResolution}),e.sourceFileAffectingCompilerOptions=e.optionDeclarations.filter(function(e){return!!e.affectsSourceFile||!!e.affectsModuleResolution||!!e.affectsBindDiagnostics}),e.buildOpts=e.commonOptionsWithBuild.concat([{name:"verbose",shortName:"v",category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Enable_verbose_logging,type:"boolean"},{name:"dry",shortName:"d",category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Show_what_would_be_built_or_deleted_if_specified_with_clean,type:"boolean"},{name:"force",shortName:"f",category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Build_all_projects_including_those_that_appear_to_be_up_to_date,type:"boolean"},{name:"clean",category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Delete_the_outputs_of_all_projects,type:"boolean"}]),e.typeAcquisitionDeclarations=[{name:"enableAutoDiscovery",type:"boolean"},{name:"enable",type:"boolean"},{name:"include",type:"list",element:{name:"include",type:"string"}},{name:"exclude",type:"list",element:{name:"exclude",type:"string"}}],e.defaultInitCompilerOptions={module:e.ModuleKind.CommonJS,target:1,strict:!0,esModuleInterop:!0},e.convertEnableAutoDiscoveryToEnable=a,e.createOptionNameMap=s,e.createCompilerDiagnosticForInvalidCustomType=c,e.parseCustomTypeOption=u,e.parseListTypeOption=d,e.parseCommandLine=m,e.getOptionFromName=f,e.parseBuildCommand=g,e.printVersion=y,e.printHelp=h,e.getParsedCommandLineOfConfigFile=b,e.readConfigFile=E,e.parseConfigFileTextToJson=T,e.readJsonConfigFile=S,e.convertToObject=D,e.convertToObjectWorker=L,e.convertToTSConfig=I,e.generateTSConfig=F,e.parseJsonConfigFileContent=G,e.parseJsonSourceFileConfigFileContent=V,e.setConfigFileInOptions=B,e.canJsonReportNoInutFiles=z,e.updateErrorForNoInputFiles=J,e.convertCompilerOptionsFromJson=te,e.convertTypeAcquisitionFromJson=re;var _e=/(^|\/)\*\*\/?$/,ge=/(^|\/)\*\*\/(.*\/)?\.\.($|\/)/,ve=/\/[^/]*?[*?][^/]*\//,ye=/^[^*?]*(?=\/[^/]*[*?])/;function he(n,t,r,a,i,o,s,c,l){var u,d;a=e.normalizePath(a),t&&(u=Ee(t,s,!1,l,"include")),r&&(d=Ee(r,s,!0,l,"exclude"));var p=Se(u,d,a,o.useCaseSensitiveFileNames),m={filesSpecs:n,includeSpecs:t,excludeSpecs:r,validatedIncludeSpecs:u,validatedExcludeSpecs:d,wildcardDirectories:p};return be(m,a,i,o,c)}function be(n,t,r,a,i){void 0===i&&(i=[]),t=e.normalizePath(t);var o,s=a.useCaseSensitiveFileNames?e.identity:e.toLowerCase,c=e.createMap(),l=e.createMap(),u=e.createMap(),d=n.filesSpecs,p=n.validatedIncludeSpecs,m=n.validatedExcludeSpecs,f=n.wildcardDirectories,_=e.getSupportedExtensions(r,i),g=e.getSuppoertedExtensionsWithJsonIfResolveJsonModule(r,_);if(d)for(var v=0,y=d;v<y.length;v++){var h=y[v],b=e.getNormalizedAbsolutePath(h,t);c.set(s(b),b)}if(p&&p.length>0)for(var E=function(n){if(e.fileExtensionIs(n,".json")){if(!o){var r=p.filter(function(n){return e.endsWith(n,".json")}),i=e.map(e.getRegularExpressionsForWildcards(r,t,"files"),function(e){return"^"+e+"$"});o=i?i.map(function(n){return e.getRegexFromPattern(n,a.useCaseSensitiveFileNames)}):e.emptyArray}var d=e.findIndex(o,function(e){return e.test(n)});if(-1!==d){var m=s(n);c.has(m)||u.has(m)||u.set(m,n)}return"continue"}if(Ce(n,c,l,_,s))return"continue";Ae(n,l,_,s);var f=s(n);c.has(f)||l.has(f)||l.set(f,n)},T=0,S=a.readDirectory(t,g,m,p,void 0);T<S.length;T++){b=S[T];E(b)}var x=e.arrayFrom(c.values()),C=e.arrayFrom(l.values());return{fileNames:x.concat(C,e.arrayFrom(u.values())),wildcardDirectories:f,spec:n}}function Ee(n,t,r,a,i){return n.filter(function(e){var n=Te(e,r);return void 0!==n&&t.push(o(n,e)),void 0===n});function o(n,t){var r=e.getTsConfigPropArrayElementValue(a,i,t);return r?e.createDiagnosticForNodeInSourceFile(a,r,n,t):e.createCompilerDiagnostic(n,t)}}function Te(n,t){return!t&&_e.test(n)?e.Diagnostics.File_specification_cannot_end_in_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0:ge.test(n)?e.Diagnostics.File_specification_cannot_contain_a_parent_directory_that_appears_after_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0:void 0}function Se(n,t,r,a){var i=e.getRegularExpressionForWildcard(t,r,"exclude"),o=i&&new RegExp(i,a?"":"i"),s={};if(void 0!==n){for(var c=[],l=0,u=n;l<u.length;l++){var d=u[l],p=e.normalizePath(e.combinePaths(r,d));if(!o||!o.test(p)){var m=xe(p,a);if(m){var f=m.key,_=m.flags,g=s[f];(void 0===g||g<_)&&(s[f]=_,1===_&&c.push(f))}}}for(var f in s)if(e.hasProperty(s,f))for(var v=0,y=c;v<y.length;v++){var h=y[v];f!==h&&e.containsPath(h,f,r,!a)&&delete s[f]}}return s}function xe(n,t){var r=ye.exec(n);return r?{key:t?r[0]:r[0].toLowerCase(),flags:ve.test(n)?1:0}:e.isImplicitGlob(n)?{key:n,flags:1}:void 0}function Ce(n,t,r,a,i){for(var o=e.getExtensionPriority(n,a),s=e.adjustExtensionPriority(o,a),c=0;c<s;c++){var l=a[c],u=i(e.changeExtension(n,l));if(t.has(u)||r.has(u))return!0}return!1}function Ae(n,t,r,a){for(var i=e.getExtensionPriority(n,r),o=e.getNextLowestExtensionPriority(i,r),s=o;s<r.length;s++){var c=r[s],l=a(e.changeExtension(n,c));t.delete(l)}}function De(e){var n={};for(var t in e)if(e.hasOwnProperty(t)){var r=f(t);void 0!==r&&(n[t]=Le(e[t],r))}return n}function Le(n,t){switch(t.type){case"object":return"";case"string":return"";case"number":return"number"===typeof n?n:"";case"boolean":return"boolean"===typeof n?n:"";case"list":var r=t.element;return e.isArray(n)?n.map(function(e){return Le(e,r)}):"";default:return e.forEachEntry(t.type,function(e,t){if(e===n)return t})}}e.getFileNamesFromConfigSpecs=be,e.convertCompilerOptionsForTelemetry=De}(u||(u={})),function(e){function n(n){n.trace(e.formatMessage.apply(void 0,arguments))}function t(e,n){return!!e.traceResolution&&void 0!==n.trace}function r(e,n){return n&&{path:n.path,extension:n.ext,packageId:e}}function a(e){return r(void 0,e)}function i(n){if(n)return e.Debug.assert(void 0===n.packageId),{path:n.path,ext:n.extension}}var s,c;function l(n){if(n)return e.Debug.assert(e.extensionIsTS(n.extension)),{fileName:n.path,packageId:n.packageId}}function u(e,n,t){return{resolvedModule:e&&{resolvedFileName:e.path,originalPath:!0===e.originalPath?void 0:e.originalPath,extension:e.extension,isExternalLibraryImport:n,packageId:e.packageId},failedLookupLocations:t}}function d(t,r,a,i){if(e.hasProperty(t,r)){var o=t[r];if(typeof o===a&&null!==o)return o;i.traceEnabled&&n(i.host,e.Diagnostics.Expected_type_of_0_field_in_package_json_to_be_1_got_2,r,a,null===o?"null":typeof o)}else i.traceEnabled&&n(i.host,e.Diagnostics.package_json_does_not_have_a_0_field,r)}function p(t,r,a,i){var o=d(t,r,"string",i);if(void 0!==o){var s=e.normalizePath(e.combinePaths(a,o));return i.traceEnabled&&n(i.host,e.Diagnostics.package_json_has_0_field_1_that_references_2,r,o,s),s}}function m(e,n,t){return p(e,"typings",n,t)||p(e,"types",n,t)}function f(e,n,t){return p(e,"tsconfig",n,t)}function _(e,n,t){return p(e,"main",n,t)}function g(t,r){var a=d(t,"typesVersions","object",r);if(void 0!==a)return r.traceEnabled&&n(r.host,e.Diagnostics.package_json_has_a_typesVersions_field_with_version_specific_path_mappings),a}function v(t,r){var a=g(t,r);if(void 0!==a){if(r.traceEnabled)for(var i in a)e.hasProperty(a,i)&&!e.VersionRange.tryParse(i)&&n(r.host,e.Diagnostics.package_json_has_a_typesVersions_entry_0_that_is_not_a_valid_semver_range,i);var o=y(a);if(o){var s=o.version,c=o.paths;if("object"===typeof c)return o;r.traceEnabled&&n(r.host,e.Diagnostics.Expected_type_of_0_field_in_package_json_to_be_1_got_2,"typesVersions['"+s+"']","object",typeof c)}else r.traceEnabled&&n(r.host,e.Diagnostics.package_json_does_not_have_a_typesVersions_entry_that_matches_version_0,e.versionMajorMinor)}}function y(n){for(var t in c||(c=new e.Version(e.version)),n)if(e.hasProperty(n,t)){var r=e.VersionRange.tryParse(t);if(void 0!==r&&r.test(c))return{version:t,paths:n[t]}}}function h(n,t){return n.typeRoots?n.typeRoots:(n.configFilePath?r=e.getDirectoryPath(n.configFilePath):t.getCurrentDirectory&&(r=t.getCurrentDirectory()),void 0!==r?b(r,t):void 0);var r}function b(n,t){return t.directoryExists?(e.forEachAncestorDirectory(e.normalizePath(n),function(n){var a=e.combinePaths(n,E);t.directoryExists(a)&&(r||(r=[])).push(a)}),r):[e.combinePaths(n,E)];var r}e.trace=n,e.isTraceEnabled=t,function(e){e[e["TypeScript"]=0]="TypeScript",e[e["JavaScript"]=1]="JavaScript",e[e["Json"]=2]="Json",e[e["TSConfig"]=3]="TSConfig",e[e["DtsOnly"]=4]="DtsOnly"}(s||(s={})),e.getPackageJsonTypesVersionsPaths=y,e.getEffectiveTypeRoots=h;var E=e.combinePaths("node_modules","@types");function T(r,a,i,o,c){var u=t(i,o);c&&(i=c.commandLine.options);var d=[],p={compilerOptions:i,host:o,traceEnabled:u,failedLookupLocations:d},m=h(i,o);u&&(void 0===a?void 0===m?n(o,e.Diagnostics.Resolving_type_reference_directive_0_containing_file_not_set_root_directory_not_set,r):n(o,e.Diagnostics.Resolving_type_reference_directive_0_containing_file_not_set_root_directory_1,r,m):void 0===m?n(o,e.Diagnostics.Resolving_type_reference_directive_0_containing_file_1_root_directory_not_set,r,a):n(o,e.Diagnostics.Resolving_type_reference_directive_0_containing_file_1_root_directory_2,r,a,m),c&&n(o,e.Diagnostics.Using_compiler_options_of_project_reference_redirect_0,c.sourceFile.fileName));var f,_=E(),g=!0;if(_||(_=T(),g=!1),_){var v=_.fileName,y=_.packageId,b=i.preserveSymlinks?v:j(v,o,u);u&&n(o,e.Diagnostics.Type_reference_directive_0_was_successfully_resolved_to_1_primary_Colon_2,r,b,g),f={primary:g,resolvedFileName:b,packageId:y,isExternalLibraryImport:H(v)}}return{resolvedTypeReferenceDirective:f,failedLookupLocations:d};function E(){if(m&&m.length)return u&&n(o,e.Diagnostics.Resolving_with_primary_search_path_0,m.join(", ")),e.firstDefined(m,function(t){var a=e.combinePaths(t,r),i=e.getDirectoryPath(a),c=e.directoryProbablyExists(i,o);return!c&&u&&n(o,e.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it,i),l(Z(s.DtsOnly,a,!c,p))});u&&n(o,e.Diagnostics.Root_directory_cannot_be_determined_skipping_primary_search_paths)}function T(){var t=a&&e.getDirectoryPath(a);if(void 0!==t){u&&n(o,e.Diagnostics.Looking_up_in_node_modules_folder_initial_location_0,t);var i=void 0;if(e.isExternalModuleNameRelative(r)){var c=e.normalizePathAndParts(e.combinePaths(t,r)).path;i=U(s.DtsOnly,c,!1,p,!0)}else{var d=ae(s.DtsOnly,r,t,p,void 0,void 0);i=d&&d.value}var m=l(i);return!m&&u&&n(o,e.Diagnostics.Type_reference_directive_0_was_not_resolved,r),m}u&&n(o,e.Diagnostics.Containing_file_is_not_specified_and_root_directory_cannot_be_determined_skipping_lookup_in_node_modules_folder)}}function S(n,t){if(n.types)return n.types;var r=[];if(t.directoryExists&&t.getDirectories){var a=h(n,t);if(a)for(var i=0,o=a;i<o.length;i++){var s=o[i];if(t.directoryExists(s))for(var c=0,l=t.getDirectories(s);c<l.length;c++){var u=l[c],d=e.normalizePath(u),p=e.combinePaths(s,d,"package.json"),m=t.fileExists(p)&&null===e.readJson(p,t).typings;if(!m){var f=e.getBaseFileName(d);46!==f.charCodeAt(0)&&r.push(f)}}}}return r}function x(e,n){return A(C(),C(),e,n)}function C(){var n=e.createMap(),t=e.createMap();return{ownMap:n,redirectsMap:t,getOrCreateMapOfCacheRedirects:r,clear:a};function r(r){if(!r)return n;var a=r.sourceFile.path,i=t.get(a);return i||(i=e.createMap(),t.set(a,i)),i}function a(){n.clear(),t.clear()}}function A(n,t,r,a){return{getOrCreateCacheForDirectory:i,getOrCreateCacheForModuleName:o};function i(t,i){var o=e.toPath(t,r,a);return s(n,i,o,e.createMap)}function o(n,r){return e.Debug.assert(!e.isExternalModuleNameRelative(n)),s(t,r,n,c)}function s(e,n,t,r){var a=e.getOrCreateMapOfCacheRedirects(n),i=a.get(t);return i||(i=r(),a.set(t,i)),i}function c(){var n=e.createMap();return{get:t,set:i};function t(t){return n.get(e.toPath(t,r,a))}function i(t,i){var s=e.toPath(t,r,a);if(!n.has(s)){n.set(s,i);var c=i.resolvedModule&&(i.resolvedModule.originalPath||i.resolvedModule.resolvedFileName),l=c&&o(s,c),u=s;while(u!==l){var d=e.getDirectoryPath(u);if(d===u||n.has(d))break;n.set(d,i),u=d}}}function o(n,t){var i=e.toPath(e.getDirectoryPath(t),r,a),o=0,s=Math.min(n.length,i.length);while(o<s&&n.charCodeAt(o)===i.charCodeAt(o))o++;if(o===n.length&&(i.length===o||i[o]===e.directorySeparator))return n;var c=e.getRootLength(n);if(!(o<c)){var l=n.lastIndexOf(e.directorySeparator,o-1);if(-1!==l)return n.substr(0,Math.max(l,c))}}}}function D(n,t,r){var a=e.getDirectoryPath(t),i=r&&r.getOrCreateCacheForDirectory(a);return i&&i.get(n)}function L(r,a,i,o,s,c){var l=t(i,o);c&&(i=c.commandLine.options),l&&(n(o,e.Diagnostics.Resolving_module_0_from_1,r,a),c&&n(o,e.Diagnostics.Using_compiler_options_of_project_reference_redirect_0,c.sourceFile.fileName));var u=e.getDirectoryPath(a),d=s&&s.getOrCreateCacheForDirectory(u,c),p=d&&d.get(r);if(p)l&&n(o,e.Diagnostics.Resolution_for_module_0_was_found_in_cache_from_location_1,r,u);else{var m=i.moduleResolution;switch(void 0===m?(m=e.getEmitModuleKind(i)===e.ModuleKind.CommonJS?e.ModuleResolutionKind.NodeJs:e.ModuleResolutionKind.Classic,l&&n(o,e.Diagnostics.Module_resolution_kind_is_not_specified_using_0,e.ModuleResolutionKind[m])):l&&n(o,e.Diagnostics.Explicitly_specified_module_resolution_kind_Colon_0,e.ModuleResolutionKind[m]),m){case e.ModuleResolutionKind.NodeJs:p=B(r,a,i,o,s,c);break;case e.ModuleResolutionKind.Classic:p=ve(r,a,i,o,s,c);break;default:return e.Debug.fail("Unexpected moduleResolution: "+m)}d&&(d.set(r,p),e.isExternalModuleNameRelative(r)||s.getOrCreateCacheForModuleName(r,c).set(u,p))}return l&&(p.resolvedModule?n(o,e.Diagnostics.Module_name_0_was_successfully_resolved_to_1,r,p.resolvedModule.resolvedFileName):n(o,e.Diagnostics.Module_name_0_was_not_resolved,r)),p}function k(n,t,r,a,i){var o=N(n,t,a,i);return o?o.value:e.isExternalModuleNameRelative(t)?I(n,t,r,a,i):w(n,t,a,i)}function N(t,r,a,i){var o=i.compilerOptions,s=o.baseUrl,c=o.paths;if(s&&c&&!e.pathIsRelative(r))return i.traceEnabled&&(n(i.host,e.Diagnostics.baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1,s,r),n(i.host,e.Diagnostics.paths_option_is_specified_looking_for_a_pattern_to_match_module_name_0,r)),le(t,r,s,c,a,!1,i)}function I(t,r,a,i,o){if(o.compilerOptions.rootDirs){o.traceEnabled&&n(o.host,e.Diagnostics.rootDirs_option_is_set_using_it_to_resolve_relative_module_name_0,r);for(var s,c,l=e.normalizePath(e.combinePaths(a,r)),u=0,d=o.compilerOptions.rootDirs;u<d.length;u++){var p=d[u],m=e.normalizePath(p);e.endsWith(m,e.directorySeparator)||(m+=e.directorySeparator);var f=e.startsWith(l,m)&&(void 0===c||c.length<m.length);o.traceEnabled&&n(o.host,e.Diagnostics.Checking_if_0_is_the_longest_matching_prefix_for_1_2,m,l,f),f&&(c=m,s=p)}if(c){o.traceEnabled&&n(o.host,e.Diagnostics.Longest_matching_prefix_for_0_is_1,l,c);var _=l.substr(c.length);o.traceEnabled&&n(o.host,e.Diagnostics.Loading_0_from_the_root_dir_1_candidate_location_2,_,c,l);var g=i(t,l,!e.directoryProbablyExists(a,o.host),o);if(g)return g;o.traceEnabled&&n(o.host,e.Diagnostics.Trying_other_entries_in_rootDirs);for(var v=0,y=o.compilerOptions.rootDirs;v<y.length;v++){p=y[v];if(p!==s){var h=e.combinePaths(e.normalizePath(p),_);o.traceEnabled&&n(o.host,e.Diagnostics.Loading_0_from_the_root_dir_1_candidate_location_2,_,p,h);var b=e.getDirectoryPath(h),E=i(t,h,!e.directoryProbablyExists(b,o.host),o);if(E)return E}}o.traceEnabled&&n(o.host,e.Diagnostics.Module_resolution_using_rootDirs_has_failed)}}}function w(t,r,a,i){var o=i.compilerOptions.baseUrl;if(o){i.traceEnabled&&n(i.host,e.Diagnostics.baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1,o,r);var s=e.normalizePath(e.combinePaths(o,r));return i.traceEnabled&&n(i.host,e.Diagnostics.Resolving_module_name_0_relative_to_base_url_1_2,r,o,s),a(t,s,!e.directoryProbablyExists(e.getDirectoryPath(s),i.host),i)}}function M(e,n,t){var r=V(e,n,t),a=r.resolvedModule,i=r.failedLookupLocations;if(!a)throw new Error("Could not resolve JS module '"+e+"' starting at '"+n+"'. Looked in: "+i.join(", "));return a.resolvedFileName}function O(e,n,t){var r=V(e,n,t).resolvedModule;return r&&r.resolvedFileName}e.resolveTypeReferenceDirective=T,e.getAutomaticTypeDirectiveNames=S,e.createModuleResolutionCache=x,e.createCacheWithRedirects=C,e.createModuleResolutionCacheWithMaps=A,e.resolveModuleNameFromCache=D,e.resolveModuleName=L,e.resolveJSModule=M,e.tryResolveJSModule=O;var P=[s.JavaScript],R=[s.TypeScript,s.JavaScript],F=R.concat([s.Json]),G=[s.TSConfig];function V(n,t,r){return K(n,t,{moduleResolution:e.ModuleResolutionKind.NodeJs,allowJs:!0},r,void 0,P,void 0)}function B(n,t,r,a,i,o,s){return K(n,e.getDirectoryPath(t),r,a,i,s?G:r.resolveJsonModule?F:R,o)}function K(r,a,i,c,l,d,p){var m=t(i,c),f=[],_={compilerOptions:i,host:c,traceEnabled:m,failedLookupLocations:f},g=e.forEach(d,function(e){return b(e)});if(g&&g.value){var v=g.value,y=v.resolved,h=v.isExternalLibraryImport;return u(y,h,f)}return{resolvedModule:void 0,failedLookupLocations:f};function b(t){var u=function(e,n,t,r){return U(e,n,t,r,!0)},d=k(t,r,a,u,_);if(d)return he({resolved:d,isExternalLibraryImport:H(d.path)});if(e.isExternalModuleNameRelative(r)){var f=e.normalizePathAndParts(e.combinePaths(a,r)),g=f.path,v=f.parts,y=U(t,g,!1,_,!0);return y&&he({resolved:y,isExternalLibraryImport:e.contains(v,"node_modules")})}m&&n(c,e.Diagnostics.Loading_module_0_from_node_modules_folder_target_file_type_1,r,s[t]);var h=ae(t,r,a,_,l,p);if(h){var b=h.value;if(!i.preserveSymlinks&&b&&!b.originalPath){var E=j(b.path,c,m),T=E===b.path?void 0:b.path;b=o({},b,{path:E,originalPath:T})}return{value:b&&{resolved:b,isExternalLibraryImport:!0}}}}}function j(t,r,a){if(!r.realpath)return t;var i=e.normalizePath(r.realpath(t));return a&&n(r,e.Diagnostics.Resolving_real_path_for_0_result_1,t,i),e.Debug.assert(r.fileExists(i),t+" linked to nonexistent file "+i),i}function U(t,a,i,o,c){if(o.traceEnabled&&n(o.host,e.Diagnostics.Loading_module_as_file_Slash_folder_candidate_module_location_0_target_file_type_1,a,s[t]),!e.hasTrailingDirectorySeparator(a)){if(!i){var l=e.getDirectoryPath(a);e.directoryProbablyExists(l,o.host)||(o.traceEnabled&&n(o.host,e.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it,l),i=!0)}var u=X(t,a,i,o);if(u){var d=c?W(u):void 0,p=d&&$(d.packageDirectory,d.subModuleName,!1,o),m=p&&p.packageId;return r(m,u)}}if(!i){var f=e.directoryProbablyExists(a,o.host);f||(o.traceEnabled&&n(o.host,e.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it,a),i=!0)}return Z(t,a,i,o,c)}function H(n){return e.stringContains(n,e.nodeModulesPathPart)}function W(n){var t=e.normalizePath(n.path),r=t.lastIndexOf(e.nodeModulesPathPart);if(-1!==r){var a=r+e.nodeModulesPathPart.length,i=q(t,a);64===t.charCodeAt(a)&&(i=q(t,i));var o=t.slice(0,i),s=e.removeExtension(t.slice(i+1),n.ext)+".d.ts";return{packageDirectory:o,subModuleName:s}}}function q(n,t){var r=n.indexOf(e.directorySeparator,t+1);return-1===r?t:r}function z(n){return""===n?"index.d.ts":e.endsWith(n,".d.ts")?n:"index"===n||e.endsWith(n,"/index")?n+".d.ts":n+"/index.d.ts"}function J(e,n,t,r){return a(X(e,n,t,r))}function X(t,r,a,i){if(t===s.Json||t===s.TSConfig){var o=e.tryRemoveExtension(r,".json");return void 0===o&&t===s.Json?void 0:Y(o||r,t,a,i)}var c=Y(r,t,a,i);if(c)return c;if(e.hasJSFileExtension(r)){var l=e.removeFileExtension(r);if(i.traceEnabled){var u=r.substring(l.length);n(i.host,e.Diagnostics.File_name_0_has_a_1_extension_stripping_it,r,u)}return Y(l,t,a,i)}}function Y(n,t,r,a){if(!r){var i=e.getDirectoryPath(n);i&&(r=!e.directoryProbablyExists(i,a.host))}switch(t){case s.DtsOnly:return o(".d.ts");case s.TypeScript:return o(".ts")||o(".tsx")||o(".d.ts");case s.JavaScript:return o(".js")||o(".jsx");case s.TSConfig:case s.Json:return o(".json")}function o(e){var t=Q(n+e,r,a);return void 0===t?void 0:{path:t,ext:e}}}function Q(t,r,a){if(!r){if(a.host.fileExists(t))return a.traceEnabled&&n(a.host,e.Diagnostics.File_0_exist_use_it_as_a_name_resolution_result,t),t;a.traceEnabled&&n(a.host,e.Diagnostics.File_0_does_not_exist,t)}a.failedLookupLocations.push(t)}function Z(e,n,t,a,i){void 0===i&&(i=!0);var o=i?$(n,"",t,a):void 0,s=o&&o.packageId,c=o&&o.packageJsonContent,l=c&&v(c,a);return r(s,ee(e,n,t,a,c,l))}function $(t,r,a,i){var o=i.host,s=i.traceEnabled,c=!a&&e.directoryProbablyExists(t,o),l=e.combinePaths(t,"package.json");if(c&&o.fileExists(l)){var u=e.readJson(l,o);if(""===r){var d=m(u,t,i);if("string"===typeof d)r=z(d.substring(t.length+1));else{var p=_(u,t,i);if("string"===typeof p&&p.length>t.length){var f=p.substring(t.length+1);r=(e.forEach(e.supportedJSExtensions,function(n){return e.tryRemoveExtension(f,n)})||f)+".d.ts"}else r="index.d.ts"}}e.endsWith(r,".d.ts")||(r=z(r));var g=v(u,i),y="string"===typeof u.name&&"string"===typeof u.version?{name:u.name,subModuleName:r,version:u.version}:void 0;return s&&(y?n(o,e.Diagnostics.Found_package_json_at_0_Package_ID_is_1,l,e.packageIdToString(y)):n(o,e.Diagnostics.Found_package_json_at_0,l)),{packageJsonContent:u,packageId:y,versionPaths:g}}c&&s&&n(o,e.Diagnostics.File_0_does_not_exist,l),i.failedLookupLocations.push(l)}function ee(t,r,o,c,l,u){var d;if(l)switch(t){case s.JavaScript:case s.Json:d=_(l,r,c);break;case s.TypeScript:d=m(l,r,c)||_(l,r,c);break;case s.DtsOnly:d=m(l,r,c);break;case s.TSConfig:d=f(l,r,c);break;default:return e.Debug.assertNever(t)}var p=function(t,r,i,o){var c=Q(r,i,o);if(c){var l=ne(t,c);if(l)return a(l);o.traceEnabled&&n(o.host,e.Diagnostics.File_0_has_an_unsupported_extension_so_skipping_it,c)}var u=t===s.DtsOnly?s.TypeScript:t;return U(u,r,i,o,!1)},g=d?!e.directoryProbablyExists(e.getDirectoryPath(d),c.host):void 0,v=o||!e.directoryProbablyExists(r,c.host),y=e.combinePaths(r,t===s.TSConfig?"tsconfig":"index");if(u&&(!d||e.containsPath(r,d))){var h=e.getRelativePathFromDirectory(r,d||y,!1);c.traceEnabled&&n(c.host,e.Diagnostics.package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_match_module_name_2,u.version,e.version,h);var b=le(t,h,r,u.paths,p,g||v,c);if(b)return i(b.value)}var E=d&&i(p(t,d,g,c));return E||X(t,y,v,c)}function ne(n,t){var r=e.tryGetExtensionFromPath(t);return void 0!==r&&te(n,r)?{path:t,ext:r}:void 0}function te(e,n){switch(e){case s.JavaScript:return".js"===n||".jsx"===n;case s.TSConfig:case s.Json:return".json"===n;case s.TypeScript:return".ts"===n||".tsx"===n||".d.ts"===n;case s.DtsOnly:return".d.ts"===n}}function re(n){var t=n.indexOf(e.directorySeparator);return"@"===n[0]&&(t=n.indexOf(e.directorySeparator,t+1)),-1===t?{packageName:n,rest:""}:{packageName:n.slice(0,t),rest:n.slice(t+1)}}function ae(e,n,t,r,a,i){return oe(e,n,t,r,!1,a,i)}function ie(e,n,t){return oe(s.DtsOnly,e,n,t,!0,void 0,void 0)}function oe(n,t,r,a,i,o,s){var c=o&&o.getOrCreateCacheForModuleName(t,s);return e.forEachAncestorDirectory(e.normalizeSlashes(r),function(r){if("node_modules"!==e.getBaseFileName(r)){var o=ge(c,t,r,a);return o||he(se(n,t,r,a,i))}})}function se(t,r,a,i,o){var c=e.combinePaths(a,"node_modules"),l=e.directoryProbablyExists(c,i.host);!l&&i.traceEnabled&&n(i.host,e.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it,c);var u=o?void 0:ce(t,r,c,l,i);if(u)return u;if(t===s.TypeScript||t===s.DtsOnly){var d=e.combinePaths(c,"@types"),p=l;return l&&!e.directoryProbablyExists(d,i.host)&&(i.traceEnabled&&n(i.host,e.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it,d),p=!1),ce(s.DtsOnly,de(r,i),d,p,i)}}function ce(t,i,o,s,c){var l,u,d,p=e.normalizePath(e.combinePaths(o,i)),m=$(p,"",!s,c);if(m){l=m.packageJsonContent,u=m.packageId,d=m.versionPaths;var f=X(t,p,!s,c);if(f)return a(f);var _=ee(t,p,!s,c,l,d);return r(u,_)}var g=function(e,n,t,a){var i=X(e,n,t,a)||ee(e,n,t,a,l,d);return r(u,i)},v=re(i),y=v.packageName,h=v.rest;if(""!==h){var b=e.combinePaths(o,y),E=$(b,h,!s,c);if(E&&(u=E.packageId,d=E.versionPaths),d){c.traceEnabled&&n(c.host,e.Diagnostics.package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_match_module_name_2,d.version,e.version,h);var T=s&&e.directoryProbablyExists(b,c.host),S=le(t,h,b,d.paths,g,!T,c);if(S)return S.value}}return g(t,p,!s,c)}function le(t,r,i,o,s,c,l){var u=e.matchPatternOrExact(e.getOwnKeys(o),r);if(u){var d=e.isString(u)?void 0:e.matchedText(u,r),p=e.isString(u)?u:e.patternText(u);l.traceEnabled&&n(l.host,e.Diagnostics.Module_name_0_matched_pattern_1,r,p);var m=e.forEach(o[p],function(r){var o=d?r.replace("*",d):r,u=e.normalizePath(e.combinePaths(i,o));l.traceEnabled&&n(l.host,e.Diagnostics.Trying_substitution_0_candidate_module_location_Colon_1,r,o);var p=e.tryGetExtensionFromPath(u);if(void 0!==p){var m=Q(u,c,l);if(void 0!==m)return a({path:m,ext:p})}return s(t,u,c||!e.directoryProbablyExists(e.getDirectoryPath(u),l.host),l)});return{value:m}}}e.nodeModuleNameResolver=B,e.nodeModulesPathPart="/node_modules/",e.pathContainsNodeModules=H,e.parsePackageName=re;var ue="__";function de(t,r){var a=me(t);return r.traceEnabled&&a!==t&&n(r.host,e.Diagnostics.Scoped_package_detected_looking_in_0,a),a}function pe(e){return"@types/"+me(e)}function me(n){if(e.startsWith(n,"@")){var t=n.replace(e.directorySeparator,ue);if(t!==n)return t.slice(1)}return n}function fe(n){var t=e.removePrefix(n,"@types/");return t!==n?_e(t):n}function _e(n){return e.stringContains(n,ue)?"@"+n.replace(ue,e.directorySeparator):n}function ge(t,r,a,i){var o,s=t&&t.get(a);if(s)return i.traceEnabled&&n(i.host,e.Diagnostics.Resolution_for_module_0_was_found_in_cache_from_location_1,r,a),(o=i.failedLookupLocations).push.apply(o,s.failedLookupLocations),{value:s.resolvedModule&&{path:s.resolvedModule.resolvedFileName,originalPath:s.resolvedModule.originalPath||!0,extension:s.resolvedModule.extension,packageId:s.resolvedModule.packageId}}}function ve(n,r,a,i,o,c){var l=t(a,i),d=[],p={compilerOptions:a,host:i,traceEnabled:l,failedLookupLocations:d},m=e.getDirectoryPath(r),f=_(s.TypeScript)||_(s.JavaScript);return u(f&&f.value,!1,d);function _(t){var r=k(t,n,m,J,p);if(r)return{value:r};if(e.isExternalModuleNameRelative(n)){var a=e.normalizePath(e.combinePaths(m,n));return he(J(t,a,!1,p))}var i=o&&o.getOrCreateCacheForModuleName(n,c),l=e.forEachAncestorDirectory(m,function(r){var a=ge(i,n,r,p);if(a)return a;var o=e.normalizePath(e.combinePaths(r,n));return he(J(t,o,!1,p))});return l||(t===s.TypeScript?ie(n,m,p):void 0)}}function ye(r,a,i,o,c){var l=t(i,o);l&&n(o,e.Diagnostics.Auto_discovery_for_typings_is_enabled_in_project_0_Running_extra_resolution_pass_for_module_1_using_cache_location_2,a,r,c);var d=[],p={compilerOptions:i,host:o,traceEnabled:l,failedLookupLocations:d},m=se(s.DtsOnly,r,c,p,!1);return u(m,!0,d)}function he(e){return void 0!==e?{value:e}:void 0}e.getTypesPackageName=pe,e.mangleScopedPackageName=me,e.getPackageNameFromTypesPackageName=fe,e.unmangleScopedPackageName=_e,e.classicNameResolver=ve,e.loadModuleFromGlobalCache=ye}(u||(u={})),function(e){var n;function t(e){return e.body?r(e.body):1}function r(n){switch(n.kind){case 241:case 242:return 0;case 243:if(e.isEnumConst(n))return 2;break;case 249:case 248:if(!e.hasModifier(n,1))return 0;break;case 245:var a=0;return e.forEachChild(n,function(n){var t=r(n);switch(t){case 0:return;case 2:return void(a=2);case 1:return a=1,!0;default:e.Debug.assertNever(t)}}),a;case 244:return t(n);case 72:if(n.isInJSDocNamespace)return 0}return 1}(function(e){e[e["NonInstantiated"]=0]="NonInstantiated",e[e["Instantiated"]=1]="Instantiated",e[e["ConstEnumOnly"]=2]="ConstEnumOnly"})(e.ModuleInstanceState||(e.ModuleInstanceState={})),e.getModuleInstanceState=t,function(e){e[e["None"]=0]="None",e[e["IsContainer"]=1]="IsContainer",e[e["IsBlockScopedContainer"]=2]="IsBlockScopedContainer",e[e["IsControlFlowContainer"]=4]="IsControlFlowContainer",e[e["IsFunctionLike"]=8]="IsFunctionLike",e[e["IsFunctionExpression"]=16]="IsFunctionExpression",e[e["HasLocals"]=32]="HasLocals",e[e["IsInterface"]=64]="IsInterface",e[e["IsObjectLiteralOrClassExpressionMethod"]=128]="IsObjectLiteralOrClassExpressionMethod"}(n||(n={}));var a=e.identity,i=c();function s(n,t){e.performance.mark("beforeBind"),i(n,t),e.performance.mark("afterBind"),e.performance.measure("Bind","beforeBind","afterBind")}function c(){var n,r,i,s,c,u,d,f,v,y,h,b,E,T,S,x,C,A,D,L,k,N,I,w,M=0,O={flags:1},P={flags:1},R=0;function F(t,r,a,i,o){return e.createDiagnosticForNodeInSourceFile(e.getSourceFileOfNode(t)||n,t,r,a,i,o)}function G(t,a){n=t,r=a,i=e.getEmitScriptTarget(r),k=V(n,a),I=e.createUnderscoreEscapedMap(),M=0,w=n.isDeclarationFile,N=e.objectAllocator.getSymbolConstructor(),n.locals||(Mn(n),n.symbolCount=M,n.classifiableNames=I,pn()),n=void 0,r=void 0,i=void 0,s=void 0,c=void 0,u=void 0,d=void 0,f=void 0,v=void 0,y=!1,h=void 0,b=void 0,E=void 0,T=void 0,S=void 0,x=void 0,A=void 0,D=!1,L=0,R=0}return G;function V(n,t){return!(!e.getStrictOptionValue(t,"alwaysStrict")||n.isDeclarationFile)||!!n.externalModuleIndicator}function B(e,n){return M++,new N(e,n)}function K(n,t,r){n.flags|=r,t.symbol=n,n.declarations=e.append(n.declarations,t),1955&r&&!n.exports&&(n.exports=e.createSymbolTable()),6240&r&&!n.members&&(n.members=e.createSymbolTable()),n.constEnumOnlyModule&&304&n.flags&&(n.constEnumOnlyModule=!1),67220415&r&&j(n,t)}function j(n,t){var r=n.valueDeclaration;(!r||e.isAssignmentDeclaration(r)&&!e.isAssignmentDeclaration(t)||r.kind!==t.kind&&e.isEffectiveModuleDeclaration(r))&&(n.valueDeclaration=t)}function U(n){if(254===n.kind)return n.isExportEquals?"export=":"default";var t=e.getNameOfDeclaration(n);if(t){if(e.isAmbientModule(n)){var r=e.getTextOfIdentifierOrLiteral(t);return e.isGlobalScopeAugmentation(n)?"__global":'"'+r+'"'}if(149===t.kind){var a=t.expression;return e.isStringOrNumericLiteralLike(a)?e.escapeLeadingUnderscores(a.text):(e.Debug.assert(e.isWellKnownSymbolSyntactically(a)),e.getPropertyNameForKnownSymbolName(e.idText(a.name)))}return e.isPropertyNameLiteral(t)?e.getEscapedTextOfIdentifierOrLiteral(t):void 0}switch(n.kind){case 157:return"__constructor";case 165:case 160:case 293:return"__call";case 166:case 161:return"__new";case 162:return"__index";case 255:return"__export";case 279:return"export=";case 204:if(2===e.getAssignmentDeclarationKind(n))return"export=";e.Debug.fail("Unknown binary declaration kind");break;case 289:return e.isJSDocConstructSignature(n)?"__new":"__call";case 151:e.Debug.assert(289===n.parent.kind,"Impossible parameter parent kind",function(){return"parent is: "+(e.SyntaxKind?e.SyntaxKind[n.parent.kind]:n.parent.kind)+", expected JSDocFunctionType"});var i=n.parent,o=i.parameters.indexOf(n);return"arg"+o}}function q(n){return e.isNamedDeclaration(n)?e.declarationNameToString(n.name):e.unescapeLeadingUnderscores(e.Debug.assertDefined(U(n)))}function z(t,r,a,i,o,s){e.Debug.assert(!e.hasDynamicName(a));var c,l=e.hasModifier(a,512),u=l&&r?"default":U(a);if(void 0===u)c=B(0,"__missing");else if(c=t.get(u),2885600&i&&I.set(u,!0),c){if(s&&!c.isReplaceableByMethod)return c;if(c.flags&o)if(c.isReplaceableByMethod)t.set(u,c=B(0,u));else if(!(3&i&&67108864&c.flags)){e.isNamedDeclaration(a)&&(a.name.parent=a);var d=2&c.flags?e.Diagnostics.Cannot_redeclare_block_scoped_variable_0:e.Diagnostics.Duplicate_identifier_0,p=!0;(384&c.flags||384&i)&&(d=e.Diagnostics.Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations,p=!1),c.declarations&&c.declarations.length&&(l?(d=e.Diagnostics.A_module_cannot_have_multiple_default_exports,p=!1):c.declarations&&c.declarations.length&&254===a.kind&&!a.isExportEquals&&(d=e.Diagnostics.A_module_cannot_have_multiple_default_exports,p=!1));var m=function(t){n.bindDiagnostics.push(F(e.getNameOfDeclaration(t)||t,d,p?q(t):void 0))};e.forEach(c.declarations,m),m(a),c=B(0,u)}}else t.set(u,c=B(0,u)),s&&(c.isReplaceableByMethod=!0);return K(c,a,i),c.parent?e.Debug.assert(c.parent===r,"Existing symbol parent should match new one"):c.parent=r,c}function J(n,t,r){var a=1&e.getCombinedModifierFlags(n);if(2097152&t)return 257===n.kind||248===n.kind&&a?z(c.symbol.exports,c.symbol,n,t,r):z(c.locals,void 0,n,t,r);if(e.isJSDocTypeAlias(n)&&e.Debug.assert(e.isInJSFile(n)),!e.isAmbientModule(n)&&(a||32&c.flags)||e.isJSDocTypeAlias(n)){if(e.hasModifier(n,512)&&!U(n))return z(c.symbol.exports,c.symbol,n,t,r);var i=67220415&t?1048576:0,o=z(c.locals,void 0,n,i,r);return o.exportSymbol=z(c.symbol.exports,c.symbol,n,t,r),n.localSymbol=o,o}return z(c.locals,void 0,n,t,r)}function X(n,t){var r=c,i=u,o=d;if(1&t?(197!==n.kind&&(u=c),c=d=n,32&t&&(c.locals=e.createSymbolTable()),Qe(c)):2&t&&(d=n,d.locals=void 0),4&t){var s=a,l=h,p=b,m=E,f=T,_=A,g=D,v=16&t&&!e.hasModifier(n,256)&&!n.asteriskToken&&!!e.getImmediatelyInvokedFunctionExpression(n);v||(h={flags:2},144&t&&(h.container=n)),T=v||157===n.kind?ce():void 0,b=void 0,E=void 0,A=void 0,D=!1,a=e.identity,Y(n),n.flags&=-1409,!(1&h.flags)&&8&t&&e.nodeIsPresent(n.body)&&(n.flags|=128,D&&(n.flags|=256)),279===n.kind&&(n.flags|=L),T&&(de(T,h),h=ge(T),157===n.kind&&(n.returnFlowNode=h)),v||(h=l),b=p,E=m,T=f,A=_,D=g,a=s}else 64&t?(y=!1,Y(n),n.flags=y?64|n.flags:-65&n.flags):Y(n);c=r,u=i,d=o}function Y(e){if(w)ee(e);else if(536870912&e.transformFlags)w=!0,ee(e),w=!1,R|=e.transformFlags&~H(e.kind);else{var n=R;R=0,ee(e),R=n|g(e,R)}}function Q(e){Z(e,function(e){return 239===e.kind?Mn(e):void 0}),Z(e,function(e){return 239!==e.kind?Mn(e):void 0})}function Z(n,t){if(void 0===t&&(t=Mn),void 0!==n)if(w)e.forEach(n,t);else{var r=R;R=0;for(var a=0,i=0,o=n;i<o.length;i++){var s=o[i];t(s),a|=-536870913&s.transformFlags}n.transformFlags=536870912|a,R|=r}}function $(n){e.forEachChild(n,Mn,Z)}function ee(e){if(Tt(e))return $(e),void On(e);switch(e.kind){case 224:Te(e);break;case 223:Se(e);break;case 225:xe(e);break;case 226:case 227:Ce(e);break;case 222:Ae(e);break;case 230:case 234:De(e);break;case 229:case 228:Ne(e);break;case 235:Ie(e);break;case 232:we(e);break;case 246:Me(e);break;case 271:Oe(e);break;case 233:Fe(e);break;case 202:Ke(e);break;case 203:je(e);break;case 204:Ue(e);break;case 198:He(e);break;case 205:We(e);break;case 237:ze(e);break;case 191:Xe(e);break;case 304:case 297:Je(e);break;case 279:Q(e.statements),Mn(e.endOfFileToken);break;case 218:case 245:Q(e.statements);break;default:$(e);break}On(e)}function ne(e){switch(e.kind){case 72:case 100:case 189:case 190:return te(e);case 191:return re(e);case 195:return ne(e.expression);case 204:return oe(e);case 202:return 52===e.operator&&ne(e.operand);case 199:return ne(e.expression)}return!1}function te(n){return 72===n.kind||100===n.kind||98===n.kind||(e.isPropertyAccessExpression(n)||e.isNonNullExpression(n)||e.isParenthesizedExpression(n))&&te(n.expression)||e.isElementAccessExpression(n)&&n.argumentExpression&&(e.isStringLiteral(n.argumentExpression)||e.isNumericLiteral(n.argumentExpression))&&te(n.expression)}function re(e){if(e.arguments)for(var n=0,t=e.arguments;n<t.length;n++){var r=t[n];if(te(r))return!0}return!(189!==e.expression.kind||!te(e.expression.expression))}function ae(n,t){return e.isTypeOfExpression(n)&&se(n.expression)&&e.isStringLiteralLike(t)}function ie(n,t){return e.isStringLiteralLike(n)&&ne(t)}function oe(e){switch(e.operatorToken.kind){case 59:return te(e.left);case 33:case 34:case 35:case 36:return se(e.left)||se(e.right)||ae(e.right,e.left)||ae(e.left,e.right);case 94:return se(e.left);case 93:return ie(e.left,e.right);case 27:return ne(e.right)}return!1}function se(e){switch(e.kind){case 195:return se(e.expression);case 204:switch(e.operatorToken.kind){case 59:return se(e.left);case 27:return se(e.right)}}return te(e)}function ce(){return{flags:4,antecedents:void 0}}function le(){return{flags:8,antecedents:void 0}}function ue(e){e.flags|=512&e.flags?1024:512}function de(n,t){1&t.flags||e.contains(n.antecedents,t)||((n.antecedents||(n.antecedents=[])).push(t),ue(t))}function pe(e,n,t){return 1&n.flags?n:t?102===t.kind&&64&e||87===t.kind&&32&e?O:ne(t)?(ue(n),a({flags:e,expression:t,antecedent:n})):n:32&e?n:O}function me(e,n,t,r){return ne(n.expression)?(ue(e),a({flags:128,switchStatement:n,clauseStart:t,clauseEnd:r,antecedent:e})):e}function fe(e,n){return ue(e),a({flags:16,antecedent:e,node:n})}function _e(e,n){ue(e);var t=a({flags:256,antecedent:e,node:n});return t}function ge(e){var n=e.antecedents;return n?1===n.length?n[0]:e:O}function ve(e){var n=e.parent;switch(n.kind){case 222:case 224:case 223:return n.expression===e;case 225:case 205:return n.condition===e}return!1}function ye(e){while(1)if(195===e.kind)e=e.expression;else{if(202!==e.kind||52!==e.operator)return 204===e.kind&&(54===e.operatorToken.kind||55===e.operatorToken.kind);e=e.operand}}function he(e){while(195===e.parent.kind||202===e.parent.kind&&52===e.parent.operator)e=e.parent;return!ve(e)&&!ye(e.parent)}function be(e,n,t){var r=S,a=x;S=n,x=t,Mn(e),S=r,x=a,e&&ye(e)||(de(n,pe(32,h,e)),de(t,pe(64,h,e)))}function Ee(e,n,t){var r=b,a=E;b=n,E=t,Mn(e),b=r,E=a}function Te(e){var n=le(),t=ce(),r=ce();de(n,h),h=n,be(e.expression,t,r),h=ge(t),Ee(e.statement,r,n),de(n,h),h=ge(r)}function Se(n){var t=le(),r=233===n.parent.kind?e.lastOrUndefined(A):void 0,a=r?r.continueTarget:ce(),i=r?r.breakTarget:ce();de(t,h),h=t,Ee(n.statement,i,a),de(a,h),h=ge(a),be(n.expression,t,i),h=ge(i)}function xe(e){var n=le(),t=ce(),r=ce();Mn(e.initializer),de(n,h),h=n,be(e.condition,t,r),h=ge(t),Ee(e.statement,r,n),Mn(e.incrementor),de(n,h),h=ge(r)}function Ce(e){var n=le(),t=ce();de(n,h),h=n,227===e.kind&&Mn(e.awaitModifier),Mn(e.expression),de(t,h),Mn(e.initializer),238!==e.initializer.kind&&Ve(e.initializer),Ee(e.statement,t,n),de(n,h),h=ge(t)}function Ae(e){var n=ce(),t=ce(),r=ce();be(e.expression,n,t),h=ge(n),Mn(e.thenStatement),de(r,h),h=ge(t),Mn(e.elseStatement),de(r,h),h=ge(r)}function De(e){Mn(e.expression),230===e.kind&&(D=!0,T&&de(T,h)),h=O}function Le(e){if(A)for(var n=0,t=A;n<t.length;n++){var r=t[n];if(r.name===e)return r}}function ke(e,n,t){var r=229===e.kind?n:t;r&&(de(r,h),h=O)}function Ne(e){if(Mn(e.label),e.label){var n=Le(e.label.escapedText);n&&(n.referenced=!0,ke(e,n.breakTarget,n.continueTarget))}else ke(e,b,E)}function Ie(e){var n=ce(),t=h,r=[],i=a;(e.catchClause||e.finallyBlock)&&(a=function(e){return r.push(e),e}),Mn(e.tryBlock),a=i,de(n,h);var o=h,s=O;if(e.catchClause){if(h=t,r.length){var c=ce();de(c,h);for(var l=0,u=r;l<u.length;l++){var d=u[l];de(c,d)}h=ge(c)}Mn(e.catchClause),de(n,h),s=h}if(e.finallyBlock){var p=t;if(!e.catchClause&&r.length){var m=ce();de(m,t);for(var f=0,_=r;f<_.length;f++){d=_[f];de(m,d)}p=ge(m)}var g={flags:2048,antecedent:p,lock:{}};if(de(n,g),h=ge(n),Mn(e.finallyBlock),1&h.flags||1&o.flags&&1&s.flags&&(h=o===P||s===P?P:O),!(1&h.flags)){var v=a({flags:4096,antecedent:h});g.lock=v,h=v}}else h=ge(n)}function we(n){var t=ce();Mn(n.expression);var r=b,a=C;b=t,C=h,Mn(n.caseBlock),de(t,h);var i=e.forEach(n.caseBlock.clauses,function(e){return 272===e.kind});n.possiblyExhaustive=!i&&!t.antecedents,i||de(t,me(C,n,0,0)),b=r,C=a,h=ge(t)}function Me(n){var t=R;R=0;for(var a=n.clauses,i=O,o=0;o<a.length;o++){var s=o;while(!a[o].statements.length&&o+1<a.length)Mn(a[o]),o++;var c=ce();de(c,me(C,n.parent,s,o+1)),de(c,i),h=ge(c);var l=a[o];Mn(l),i=h,1&h.flags||o===a.length-1||!r.noFallthroughCasesInSwitch||kn(l,e.Diagnostics.Fallthrough_case_in_switch)}a.transformFlags=536870912|R,R|=t}function Oe(e){var n=h;h=C,Mn(e.expression),h=n,Z(e.statements)}function Pe(e,n,t){var r={name:e,breakTarget:n,continueTarget:t,referenced:!1};return(A||(A=[])).push(r),r}function Re(){A.pop()}function Fe(n){var t=le(),a=ce();Mn(n.label),de(t,h);var i=Pe(n.label.escapedText,a,t);Mn(n.statement),Re(),i.referenced||r.allowUnusedLabels||Nn(e.unusedLabelIsError(r),n.label,e.Diagnostics.Unused_label),n.statement&&223===n.statement.kind||(de(a,h),h=ge(a))}function Ge(e){204===e.kind&&59===e.operatorToken.kind?Ve(e.left):Ve(e)}function Ve(e){if(te(e))h=fe(h,e);else if(187===e.kind)for(var n=0,t=e.elements;n<t.length;n++){var r=t[n];208===r.kind?Ve(r.expression):Ge(r)}else if(188===e.kind)for(var a=0,i=e.properties;a<i.length;a++){var o=i[a];275===o.kind?Ge(o.initializer):276===o.kind?Ve(o.name):277===o.kind&&Ve(o.expression)}}function Be(e,n,t){var r=ce();54===e.operatorToken.kind?be(e.left,r,t):be(e.left,n,r),h=ge(r),Mn(e.operatorToken),be(e.right,n,t)}function Ke(e){if(52===e.operator){var n=S;S=x,x=n,$(e),x=S,S=n}else $(e),44!==e.operator&&45!==e.operator||Ve(e.operand)}function je(e){$(e),44!==e.operator&&45!==e.operator||Ve(e.operand)}function Ue(n){var t=n.operatorToken.kind;if(54===t||55===t)if(he(n)){var r=ce();Be(n,r,r),h=ge(r)}else Be(n,S,x);else if($(n),e.isAssignmentOperator(t)&&!e.isAssignmentTarget(n)&&(Ve(n.left),59===t&&190===n.left.kind)){var a=n.left;se(a.expression)&&(h=_e(h,n))}}function He(e){$(e),189===e.expression.kind&&Ve(e.expression)}function We(e){var n=ce(),t=ce(),r=ce();be(e.condition,n,t),h=ge(n),Mn(e.questionToken),Mn(e.whenTrue),de(r,h),h=ge(t),Mn(e.colonToken),Mn(e.whenFalse),de(r,h),h=ge(r)}function qe(n){var t=e.isOmittedExpression(n)?void 0:n.name;if(e.isBindingPattern(t))for(var r=0,a=t.elements;r<a.length;r++){var i=a[r];qe(i)}else h=fe(h,n)}function ze(n){$(n),(n.initializer||e.isForInOrOfStatement(n.parent.parent))&&qe(n)}function Je(e){e.tagName.parent=e,e.fullName&&W(e,e.fullName)}function Xe(n){var t=n.expression;while(195===t.kind)t=t.expression;if(196===t.kind||197===t.kind?(Z(n.typeArguments),Z(n.arguments),Mn(n.expression)):$(n),189===n.expression.kind){var r=n.expression;se(r.expression)&&e.isPushOrUnshiftIdentifier(r.name)&&(h=_e(h,n))}}function Ye(n){switch(n.kind){case 209:case 240:case 243:case 188:case 168:case 292:case 268:return 1;case 241:return 65;case 244:case 242:case 181:return 33;case 279:return 37;case 156:if(e.isObjectLiteralOrClassExpressionMethod(n))return 173;case 157:case 239:case 155:case 158:case 159:case 160:case 293:case 289:case 165:case 161:case 162:case 166:return 45;case 196:case 197:return 61;case 245:return 4;case 154:return n.initializer?4:0;case 274:case 225:case 226:case 227:case 246:return 2;case 218:return e.isFunctionLike(n.parent)?0:2}return 0}function Qe(e){f&&(f.nextContainer=e),f=e}function Ze(e,n,t){switch(c.kind){case 244:return J(e,n,t);case 279:return en(e,n,t);case 209:case 240:return $e(e,n,t);case 243:return z(c.symbol.exports,c.symbol,e,n,t);case 168:case 292:case 188:case 241:case 268:return z(c.symbol.members,c.symbol,e,n,t);case 165:case 166:case 160:case 161:case 293:case 162:case 156:case 155:case 157:case 158:case 159:case 239:case 196:case 197:case 289:case 304:case 297:case 242:case 181:return z(c.locals,void 0,e,n,t)}}function $e(n,t,r){return e.hasModifier(n,32)?z(c.symbol.exports,c.symbol,n,t,r):z(c.symbol.members,c.symbol,n,t,r)}function en(t,r,a){return e.isExternalModule(n)?J(t,r,a):z(n.locals,void 0,t,r,a)}function nn(e){var n=279===e.kind?e:e.body;if(n&&(279===n.kind||245===n.kind))for(var t=0,r=n.statements;t<r.length;t++){var a=r[t];if(255===a.kind||254===a.kind)return!0}return!1}function tn(e){4194304&e.flags&&!nn(e)?e.flags|=32:e.flags&=-33}function rn(t){if(tn(t),e.isAmbientModule(t))if(e.hasModifier(t,1)&&kn(t,e.Diagnostics.export_modifier_cannot_be_applied_to_ambient_modules_and_module_augmentations_since_they_are_always_visible),e.isModuleAugmentationExternal(t))an(t);else{var r=void 0;if(10===t.name.kind){var a=t.name.text;e.hasZeroOrOneAsteriskCharacter(a)?r=e.tryParsePattern(a):kn(t.name,e.Diagnostics.Pattern_0_can_have_at_most_one_Asterisk_character,a)}var i=Ze(t,512,110735);n.patternAmbientModules=e.append(n.patternAmbientModules,r&&{pattern:r,symbol:i})}else{var o=an(t);if(0!==o){i=t.symbol;i.constEnumOnlyModule=!(304&i.flags)&&2===o&&!1!==i.constEnumOnlyModule}}}function an(e){var n=t(e),r=0!==n;return Ze(e,r?512:1024,r?110735:0),n}function on(n){var t=B(131072,U(n));K(t,n,131072);var r=B(2048,"__type");K(r,n,2048),r.members=e.createSymbolTable(),r.members.set(t.escapedName,t)}function sn(t){var r;if(function(e){e[e["Property"]=1]="Property",e[e["Accessor"]=2]="Accessor"}(r||(r={})),k)for(var a=e.createUnderscoreEscapedMap(),i=0,o=t.properties;i<o.length;i++){var s=o[i];if(277!==s.kind&&72===s.name.kind){var c=s.name,l=275===s.kind||276===s.kind||156===s.kind?1:2,u=a.get(c.escapedText);if(u){if(1===l&&1===u){var d=e.getErrorSpanForNode(n,c);n.bindDiagnostics.push(e.createFileDiagnostic(n,d.start,d.length,e.Diagnostics.An_object_literal_cannot_have_multiple_properties_with_the_same_name_in_strict_mode))}}else a.set(c.escapedText,l)}}return un(t,4096,"__object")}function cn(e){return un(e,4096,"__jsxAttributes")}function ln(e,n,t){return Ze(e,n,t)}function un(e,n,t){var r=B(n,t);return 106508&n&&(r.parent=c.symbol),K(r,e,n),r}function dn(n,t,r){switch(d.kind){case 244:J(n,t,r);break;case 279:if(e.isExternalOrCommonJsModule(c)){J(n,t,r);break}default:d.locals||(d.locals=e.createSymbolTable(),Qe(d)),z(d.locals,void 0,n,t,r)}}function pn(){if(v){for(var t=c,r=f,a=d,i=s,o=h,l=0,u=v;l<u.length;l++){var p=u[l],m=e.getJSDocHost(p);c=e.findAncestor(m.parent,function(e){return!!(1&Ye(e))})||n,d=e.getEnclosingBlockScopeContainer(m)||n,h={flags:2},s=p,Mn(p.typeExpression),p.fullName&&72!==p.fullName.kind?Mn(p.fullName):(s=p.parent,dn(p,524288,67897832))}c=t,f=r,d=a,s=i,h=o}}function mn(t){!(k&&t.originalKeywordKind>=109&&t.originalKeywordKind<=117)||e.isIdentifierName(t)||4194304&t.flags||n.parseDiagnostics.length||n.bindDiagnostics.push(F(t,fn(t),e.declarationNameToString(t)))}function fn(t){return e.getContainingClass(t)?e.Diagnostics.Identifier_expected_0_is_a_reserved_word_in_strict_mode_Class_definitions_are_automatically_in_strict_mode:n.externalModuleIndicator?e.Diagnostics.Identifier_expected_0_is_a_reserved_word_in_strict_mode_Modules_are_automatically_in_strict_mode:e.Diagnostics.Identifier_expected_0_is_a_reserved_word_in_strict_mode}function _n(n){k&&e.isLeftHandSideExpression(n.left)&&e.isAssignmentOperator(n.operatorToken.kind)&&hn(n,n.left)}function gn(e){k&&e.variableDeclaration&&hn(e,e.variableDeclaration.name)}function vn(t){if(k&&72===t.expression.kind){var r=e.getErrorSpanForNode(n,t.expression);n.bindDiagnostics.push(e.createFileDiagnostic(n,r.start,r.length,e.Diagnostics.delete_cannot_be_called_on_an_identifier_in_strict_mode))}}function yn(n){return e.isIdentifier(n)&&("eval"===n.escapedText||"arguments"===n.escapedText)}function hn(t,r){if(r&&72===r.kind){var a=r;if(yn(a)){var i=e.getErrorSpanForNode(n,r);n.bindDiagnostics.push(e.createFileDiagnostic(n,i.start,i.length,bn(t),e.idText(a)))}}}function bn(t){return e.getContainingClass(t)?e.Diagnostics.Invalid_use_of_0_Class_definitions_are_automatically_in_strict_mode:n.externalModuleIndicator?e.Diagnostics.Invalid_use_of_0_Modules_are_automatically_in_strict_mode:e.Diagnostics.Invalid_use_of_0_in_strict_mode}function En(e){k&&hn(e,e.name)}function Tn(t){return e.getContainingClass(t)?e.Diagnostics.Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Class_definitions_are_automatically_in_strict_mode:n.externalModuleIndicator?e.Diagnostics.Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Modules_are_automatically_in_strict_mode:e.Diagnostics.Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5}function Sn(t){if(i<2&&279!==d.kind&&244!==d.kind&&!e.isFunctionLike(d)){var r=e.getErrorSpanForNode(n,t);n.bindDiagnostics.push(e.createFileDiagnostic(n,r.start,r.length,Tn(t)))}}function xn(t){k&&32&t.numericLiteralFlags&&n.bindDiagnostics.push(F(t,e.Diagnostics.Octal_literals_are_not_allowed_in_strict_mode))}function Cn(e){k&&hn(e,e.operand)}function An(e){k&&(44!==e.operator&&45!==e.operator||hn(e,e.operand))}function Dn(n){k&&kn(n,e.Diagnostics.with_statements_are_not_allowed_in_strict_mode)}function Ln(n){k&&r.target>=2&&(e.isDeclarationStatement(n.statement)||e.isVariableStatement(n.statement))&&kn(n.label,e.Diagnostics.A_label_is_not_allowed_here)}function kn(t,r,a,i,o){var s=e.getSpanOfTokenAtPosition(n,t.pos);n.bindDiagnostics.push(e.createFileDiagnostic(n,s.start,s.length,r,a,i,o))}function Nn(e,n,t){In(e,n,n,t)}function In(t,r,a,i){wn(t,{pos:e.getTokenPosOfNode(r,n),end:a.end},i)}function wn(t,r,a){var i=e.createFileDiagnostic(n,r.pos,r.end-r.pos,a);t?n.bindDiagnostics.push(i):n.bindSuggestionDiagnostics=e.append(n.bindSuggestionDiagnostics,o({},i,{category:e.DiagnosticCategory.Suggestion}))}function Mn(e){if(e){e.parent=s;var n=k;if(Fn(e),e.kind>147){var t=s;s=e;var r=Ye(e);0===r?Y(e):X(e,r),s=t}else if(!w&&0===(536870912&e.transformFlags)){R|=g(e,0);t=s;1===e.kind&&(s=e),On(e),s=t}k=n}}function On(n){if(e.hasJSDocNodes(n))if(e.isInJSFile(n))for(var t=0,r=n.jsDoc;t<r.length;t++){var a=r[t];Mn(a)}else for(var i=0,o=n.jsDoc;i<o.length;i++){a=o[i];W(n,a)}}function Pn(n){if(!k)for(var t=0,r=n;t<r.length;t++){var a=r[t];if(!e.isPrologueDirective(a))return;if(Rn(a))return void(k=!0)}}function Rn(t){var r=e.getSourceTextOfNodeFromSourceFile(n,t.expression);return'"use strict"'===r||"'use strict'"===r}function Fn(t){switch(t.kind){case 72:if(t.isInJSDocNamespace){var r=t.parent;while(r&&!e.isJSDocTypeAlias(r))r=r.parent;dn(r,524288,67897832);break}case 100:return h&&(e.isExpression(t)||276===s.kind)&&(t.flowNode=h),mn(t);case 189:case 190:h&&te(t)&&(t.flowNode=h),e.isSpecialPropertyDeclaration(t)&&Qn(t),e.isInJSFile(t)&&n.commonJsModuleIndicator&&e.isModuleExportsPropertyAccessExpression(t)&&!_(d,"module")&&z(n.locals,void 0,t.expression,134217729,67220414);break;case 204:var a=e.getAssignmentDeclarationKind(t);switch(a){case 1:Jn(t);break;case 2:Xn(t);break;case 3:et(t.left,t);break;case 6:Zn(t);break;case 4:Yn(t);break;case 5:tt(t);break;case 0:break;default:e.Debug.fail("Unknown binary expression special property assignment kind")}return _n(t);case 274:return gn(t);case 198:return vn(t);case 8:return xn(t);case 203:return Cn(t);case 202:return An(t);case 231:return Dn(t);case 233:return Ln(t);case 178:return void(y=!0);case 163:break;case 150:return bt(t);case 151:return _t(t);case 237:return ft(t);case 186:return t.flowNode=h,ft(t);case 154:case 153:return Gn(t);case 275:case 276:return yt(t,4,0);case 278:return yt(t,8,68008959);case 160:case 161:case 162:return Ze(t,131072,0);case 156:case 155:return yt(t,8192|(t.questionToken?16777216:0),e.isObjectLiteralMethod(t)?0:67212223);case 239:return gt(t);case 157:return Ze(t,16384,0);case 158:return yt(t,32768,67154879);case 159:return yt(t,65536,67187647);case 165:case 289:case 293:case 166:return on(t);case 168:case 292:case 181:return Vn(t);case 188:return sn(t);case 196:case 197:return vt(t);case 191:var i=e.getAssignmentDeclarationKind(t);switch(i){case 7:return nt(t);case 8:return zn(t);case 9:return $n(t);case 0:break;default:return e.Debug.fail("Unknown call expression assignment declaration kind")}e.isInJSFile(t)&&dt(t);break;case 209:case 240:return k=!0,pt(t);case 241:return dn(t,64,67897736);case 242:return dn(t,524288,67897832);case 243:return mt(t);case 244:return rn(t);case 268:return cn(t);case 267:return ln(t,4,0);case 248:case 251:case 253:case 257:return Ze(t,2097152,2097152);case 247:return Un(t);case 250:return Wn(t);case 255:return Hn(t);case 254:return jn(t);case 279:return Pn(t.statements),Bn();case 218:if(!e.isFunctionLike(t.parent))return;case 245:return Pn(t.statements);case 299:if(293===t.parent.kind)return _t(t);if(292!==t.parent.kind)break;case 305:var o=t,c=o.isBracketed||o.typeExpression&&288===o.typeExpression.type.kind?16777220:4;return Ze(o,c,0);case 304:case 297:return(v||(v=[])).push(t)}}function Gn(e){return yt(e,4|(e.questionToken?16777216:0),0)}function Vn(e){return un(e,2048,"__type")}function Bn(){if(tn(n),e.isExternalModule(n))Kn();else if(e.isJsonSourceFile(n)){Kn();var t=n.symbol;z(n.symbol.exports,n.symbol,n,4,67108863),n.symbol=t}}function Kn(){un(n,512,'"'+e.removeFileExtension(n.fileName)+'"')}function jn(n){if(c.symbol&&c.symbol.exports){var t=e.exportAssignmentIsAlias(n)?2097152:4,r=z(c.symbol.exports,c.symbol,n,t,67108863);n.isExportEquals&&j(r,n)}else un(n,2097152,U(n))}function Un(t){t.modifiers&&t.modifiers.length&&n.bindDiagnostics.push(F(t,e.Diagnostics.Modifiers_cannot_appear_here));var r=e.isSourceFile(t.parent)?e.isExternalModule(t.parent)?t.parent.isDeclarationFile?void 0:e.Diagnostics.Global_module_exports_may_only_appear_in_declaration_files:e.Diagnostics.Global_module_exports_may_only_appear_in_module_files:e.Diagnostics.Global_module_exports_may_only_appear_at_top_level;r?n.bindDiagnostics.push(F(t,r)):(n.symbol.globalExports=n.symbol.globalExports||e.createSymbolTable(),z(n.symbol.globalExports,n.symbol,t,2097152,2097152))}function Hn(e){c.symbol&&c.symbol.exports?e.exportClause||z(c.symbol.exports,c.symbol,e,8388608,0):un(e,8388608,U(e))}function Wn(e){e.name&&Ze(e,2097152,2097152)}function qn(e){return!n.externalModuleIndicator&&(n.commonJsModuleIndicator||(n.commonJsModuleIndicator=e,Kn()),!0)}function zn(e){if(qn(e)){var n=ut(e.arguments[0],void 0,function(e,n){return n&&K(n,e,67110400),n});if(n){var t=1048580;z(n.exports,n,e,t,0)}}}function Jn(n){if(qn(n)){var t=n.left,r=ut(t.expression,void 0,function(e,n){return n&&K(n,e,67110400),n});if(r){var a=e.isClassExpression(n.right)?1048612:1048580;z(r.exports,r,t,a,0)}}}function Xn(t){if(qn(t)){var r=e.getRightMostAssignedExpression(t.right);if(!(e.isEmptyObjectLiteral(r)||c===n&&p(n,r))){var a=e.exportAssignmentIsAlias(t)?2097152:1049092;z(n.symbol.exports,n.symbol,t,67108864|a,0)}}}function Yn(n){e.Debug.assert(e.isInJSFile(n));var t=e.getThisContainer(n,!1);switch(t.kind){case 239:case 196:var r=t.symbol;if(e.isBinaryExpression(t.parent)&&59===t.parent.operatorToken.kind){var a=t.parent.left;e.isPropertyAccessEntityNameExpression(a)&&e.isPrototypeAccess(a.expression)&&(r=lt(a.expression.expression,u))}r&&(r.members=r.members||e.createSymbolTable(),z(r.members,r,n,4,0));break;case 157:case 154:case 156:case 158:case 159:var i=t.parent,o=e.hasModifier(t,32)?i.symbol.exports:i.symbol.members;z(o,i.symbol,n,4,0,!0);break;case 279:break;default:e.Debug.fail(e.Debug.showSyntaxKind(t))}}function Qn(n){100===n.expression.kind?Yn(n):e.isPropertyAccessEntityNameExpression(n)&&279===n.parent.parent.kind&&(e.isPrototypeAccess(n.expression)?et(n,n.parent):rt(n))}function Zn(e){e.left.parent=e,e.right.parent=e;var n=e.left;ot(n.expression,n,!1)}function $n(e){var n=lt(e.arguments[0].expression);it(e,n,!0)}function et(e,n){var t=e.expression,r=t.expression;e.parent=n,r.parent=t,t.parent=e,ot(r,e,!0)}function nt(e){var n=lt(e.arguments[0]),t=279===e.parent.parent.kind;n=at(n,e.arguments[0],t,!1),it(e,n,!1)}function tt(t){var r=t.left,a=lt(r.expression);(e.isInJSFile(t)||e.isFunctionSymbol(a))&&(t.left.parent=t,t.right.parent=t,e.isIdentifier(r.expression)&&c===n&&m(n,r.expression)?Jn(t):rt(r))}function rt(e){e.expression.parent=e,ot(e.expression,e,!1)}function at(t,r,a,i){if(a&&!i&&(!t||!(1920&t.flags))){var o=67110400,s=110735;t=ut(r,t,function(t,r,a){if(r)return K(r,t,o),r;var i=a?a.exports:n.jsGlobalAugmentations||(n.jsGlobalAugmentations=e.createSymbolTable());return z(i,a,t,o,s)})}return t}function it(n,t,r){if(t&&st(t)){var a=r?t.members||(t.members=e.createSymbolTable()):t.exports||(t.exports=e.createSymbolTable()),i=e.isFunctionLikeDeclaration(e.getAssignedExpandoInitializer(n)),o=i?8192:4,s=i?67212223:0;z(a,t,n,67108864|o,-67108865&s)}}function ot(n,t,r){var a=lt(n),i=e.isBinaryExpression(t.parent)?279===ct(t.parent).parent.kind:279===t.parent.parent.kind;a=at(a,t.expression,i,r),it(t,a,r)}function st(n){if(1072&n.flags)return!0;var t=n.valueDeclaration;if(t&&e.isCallExpression(t))return!!e.getAssignedExpandoInitializer(t);var r=t?e.isVariableDeclaration(t)?t.initializer:e.isBinaryExpression(t)?t.right:e.isPropertyAccessExpression(t)&&e.isBinaryExpression(t.parent)?t.parent.right:void 0:void 0;if(r=r&&e.getRightMostAssignedExpression(r),r){var a=e.isPrototypeAccess(e.isVariableDeclaration(t)?t.name:e.isBinaryExpression(t)?t.left:t);return!!e.getExpandoInitializer(e.isBinaryExpression(r)&&55===r.operatorToken.kind?r.right:r,a)}return!1}function ct(n){while(e.isBinaryExpression(n.parent))n=n.parent;return n.parent}function lt(n,t){if(void 0===t&&(t=c),e.isIdentifier(n))return _(t,n.escapedText);var r=lt(n.expression);return r&&r.exports&&r.exports.get(n.name.escapedText)}function ut(t,r,a){if(p(n,t))return n.symbol;if(e.isIdentifier(t))return a(t,lt(t),r);var i=ut(t.expression,r,a);return i&&i.exports?a(t.name,i.exports.get(t.name.escapedText),i):e.Debug.fail()}function dt(t){!n.commonJsModuleIndicator&&e.isRequireCall(t,!1)&&qn(t)}function pt(t){if(240===t.kind)dn(t,32,68008383);else{var r=t.name?t.name.escapedText:"__class";un(t,32,r),t.name&&I.set(t.name.escapedText,!0)}var a=t.symbol,i=B(4194308,"prototype"),o=a.exports.get(i.escapedName);o&&(t.name&&(t.name.parent=t),n.bindDiagnostics.push(F(o.declarations[0],e.Diagnostics.Duplicate_identifier_0,e.symbolName(i)))),a.exports.set(i.escapedName,i),i.parent=a}function mt(n){return e.isEnumConst(n)?dn(n,128,68008831):dn(n,256,68008191)}function ft(n){if(k&&hn(n,n.name),!e.isBindingPattern(n.name)){var t=e.isInJSFile(n)&&!!e.getJSDocEnumTag(n),r=t?256:0,a=t?68008191:0;e.isBlockOrCatchScoped(n)?dn(n,2|r,67220415|a):e.isParameterDeclaration(n)?Ze(n,1,67220415):Ze(n,1|r,67220414|a)}}function _t(n){if((299!==n.kind||293===c.kind)&&(!k||4194304&n.flags||hn(n,n.name),e.isBindingPattern(n.name)?un(n,1,"__"+n.parent.parameters.indexOf(n)):Ze(n,1,67220415),e.isParameterPropertyDeclaration(n))){var t=n.parent.parent;z(t.symbol.members,t.symbol,n,4|(n.questionToken?16777216:0),0)}}function gt(t){n.isDeclarationFile||4194304&t.flags||e.isAsyncFunction(t)&&(L|=1024),En(t),k?(Sn(t),dn(t,16,67219887)):Ze(t,16,67219887)}function vt(t){n.isDeclarationFile||4194304&t.flags||e.isAsyncFunction(t)&&(L|=1024),h&&(t.flowNode=h),En(t);var r=t.name?t.name.escapedText:"__function";return un(t,16,r)}function yt(t,r,a){return n.isDeclarationFile||4194304&t.flags||!e.isAsyncFunction(t)||(L|=1024),h&&e.isObjectLiteralOrClassExpressionMethod(t)&&(t.flowNode=h),e.hasDynamicName(t)?un(t,r,"__computed"):Ze(t,r,a)}function ht(n){var t=e.findAncestor(n,function(n){return n.parent&&e.isConditionalTypeNode(n.parent)&&n.parent.extendsType===n});return t&&t.parent}function bt(n){if(e.isJSDocTemplateTag(n.parent)){var t=e.find(n.parent.parent.tags,e.isJSDocTypeAlias)||e.getHostSignatureFromJSDoc(n.parent);t?(t.locals||(t.locals=e.createSymbolTable()),z(t.locals,void 0,n,262144,67635688)):Ze(n,262144,67635688)}else if(176===n.parent.kind){var r=ht(n.parent);r?(r.locals||(r.locals=e.createSymbolTable()),z(r.locals,void 0,n,262144,67635688)):un(n,262144,U(n))}else Ze(n,262144,67635688)}function Et(e){var n=t(e);return 1===n||2===n&&!!r.preserveConstEnums}function Tt(n){if(!(1&h.flags))return!1;if(h===O){var t=e.isStatementButNotDeclaration(n)&&220!==n.kind||240===n.kind||244===n.kind&&Et(n);if(t&&(h=P,!r.allowUnreachableCode)){var a=e.unreachableCodeIsError(r)&&!(4194304&n.flags)&&(!e.isVariableStatement(n)||!!(3&e.getCombinedNodeFlags(n.declarationList))||n.declarationList.declarations.some(function(e){return!!e.initializer}));l(n,function(n,t){return In(a,n,t,e.Diagnostics.Unreachable_code_detected)})}}return!0}}function l(n,t){if(e.isStatement(n)&&u(n)&&e.isBlock(n.parent)){var r=n.parent.statements,a=e.sliceAfter(r,n);e.getRangesWhere(a,u,function(e,n){return t(a[e],a[n-1])})}else t(n,n)}function u(n){return!e.isFunctionDeclaration(n)&&!d(n)&&!e.isEnumDeclaration(n)&&!(e.isVariableStatement(n)&&!(3&e.getCombinedNodeFlags(n))&&n.declarationList.declarations.some(function(e){return!e.initializer}))}function d(n){switch(n.kind){case 241:case 242:return!0;case 244:return 1!==t(n);case 243:return e.hasModifier(n,2048);default:return!1}}function p(n,t){return e.isExportsIdentifier(t)||e.isModuleExportsPropertyAccessExpression(t)||e.isIdentifier(t)&&m(n,t)}function m(n,t){var r=_(n,t.escapedText);return!!r&&!!r.valueDeclaration&&e.isVariableDeclaration(r.valueDeclaration)&&!!r.valueDeclaration.initializer&&f(n,r.valueDeclaration.initializer)}function f(n,t){return p(n,t)||e.isAssignmentExpression(t,!0)&&(f(n,t.left)||f(n,t.right))}function _(n,t){var r=n.locals&&n.locals.get(t);return r?r.exportSymbol||r:e.isSourceFile(n)&&n.jsGlobalAugmentations&&n.jsGlobalAugmentations.has(t)?n.jsGlobalAugmentations.get(t):n.symbol&&n.symbol.exports&&n.symbol.exports.get(t)}function g(e,n){var t=e.kind;switch(t){case 191:return v(e,n);case 192:return y(e,n);case 244:return K(e,n);case 195:return E(e,n);case 204:return h(e,n);case 221:return B(e,n);case 151:return b(e,n);case 197:return M(e,n);case 196:return w(e,n);case 239:return I(e,n);case 237:return R(e,n);case 238:return j(e,n);case 219:return F(e,n);case 233:return G(e,n);case 240:return T(e,n);case 209:return S(e,n);case 273:return x(e,n);case 274:return C(e,n);case 211:return A(e,n);case 157:return D(e,n);case 154:return N(e,n);case 156:return L(e,n);case 158:case 159:return k(e,n);case 248:return V(e,n);case 189:return O(e,n);case 190:return P(e,n);default:return U(e,t,n)}}function v(e,n){var t=n,r=e.expression;return e.typeArguments&&(t|=3),(131072&n||100663296&r.transformFlags)&&(t|=192,67108864&r.transformFlags&&(t|=8192)),92===r.kind&&(t|=16777216,8192&n&&(t|=16384)),e.transformFlags=536870912|t,-637666626&t}function y(e,n){var t=n;return e.typeArguments&&(t|=3),131072&n&&(t|=192),e.transformFlags=536870912|t,-637666626&t}function h(e,n){var t=n,r=e.operatorToken.kind,a=e.left.kind;return 59===r&&188===a?t|=3272:59===r&&187===a?t|=3264:41!==r&&63!==r||(t|=32),e.transformFlags=536870912|t,-637535554&t}function b(n,t){var r=t,a=n.name,i=n.initializer,o=n.dotDotDotToken;return(n.questionToken||n.type||4096&t&&e.some(n.decorators)||e.isThisIdentifier(a))&&(r|=3),e.hasModifier(n,92)&&(r|=4099),262144&t&&(r|=8),(2097152&t||i||o)&&(r|=65728),n.transformFlags=536870912|r,-637535554&r}function E(e,n){var t=n,r=e.expression,a=r.kind,i=r.transformFlags;return 212!==a&&194!==a||(t|=3),1024&i&&(t|=1024),e.transformFlags=536870912|t,-536872258&t}function T(n,t){var r;return e.hasModifier(n,2)?r=3:(r=192|t,(4096&t||n.typeParameters)&&(r|=3),32768&t&&(r|=8192)),n.transformFlags=536870912|r,-638121282&r}function S(e,n){var t=192|n;return(4096&n||e.typeParameters)&&(t|=3),32768&n&&(t|=8192),e.transformFlags=536870912|t,-638121282&t}function x(n,t){var r=t;switch(n.token){case 86:r|=192;break;case 109:r|=3;break;default:e.Debug.fail("Unexpected token for heritage clause");break}return n.transformFlags=536870912|r,-637535554&r}function C(n,t){var r=t;return n.variableDeclaration?e.isBindingPattern(n.variableDeclaration.name)&&(r|=192):r|=8,n.transformFlags=536870912|r,-637797698&r}function A(e,n){var t=192|n;return e.typeArguments&&(t|=3),e.transformFlags=536870912|t,-637535554&t}function D(n,t){var r=t;return!e.hasModifier(n,2270)&&n.body||(r|=3),262144&t&&(r|=8),n.transformFlags=536870912|r,-653616450&r}function L(n,t){var r=192|t;return(n.decorators||e.hasModifier(n,2270)||n.typeParameters||n.type||n.name&&e.isComputedPropertyName(n.name)||!n.body)&&(r|=3),262144&t&&(r|=8),e.hasModifier(n,256)&&(r|=n.asteriskToken?8:16),n.asteriskToken&&(r|=768),n.transformFlags=536870912|r,-653616450&r}function k(n,t){var r=t;return(n.decorators||e.hasModifier(n,2270)||n.type||n.name&&e.isComputedPropertyName(n.name)||!n.body)&&(r|=3),262144&t&&(r|=8),n.transformFlags=536870912|r,-653616450&r}function N(n,t){var r=3|t;return(n.initializer||e.isComputedPropertyName(n.name))&&(r|=4096),n.transformFlags=536870912|r,-637535554&r}function I(n,t){var r,a=e.getModifierFlags(n),i=n.body;return!i||2&a?r=3:(r=8388608|t,(2270&a||n.typeParameters||n.type)&&(r|=3),256&a&&(r|=n.asteriskToken?8:16),262144&t&&(r|=8),81920&t&&(r|=192),n.asteriskToken&&(r|=768)),n.transformFlags=536870912|r,-653620546&r}function w(n,t){var r=t;return(e.hasModifier(n,2270)||n.typeParameters||n.type)&&(r|=3),e.hasModifier(n,256)&&(r|=n.asteriskToken?8:16),262144&t&&(r|=8),81920&t&&(r|=192),n.asteriskToken&&(r|=768),n.transformFlags=536870912|r,-653620546&r}function M(n,t){var r=192|t;return(e.hasModifier(n,2270)||n.typeParameters||n.type)&&(r|=3),e.hasModifier(n,256)&&(r|=16),262144&t&&(r|=8),8192&t&&(r|=16384),n.transformFlags=536870912|r,-653604162&r}function O(e,n){var t=n;return 33554432&t&&(t^=33554432,t|=67108888),e.transformFlags=536870912|t,-570426690&t}function P(e,n){var t=n,r=e.expression,a=r.transformFlags;return 33554432&a&&(t&=-33554433,t|=67108888),e.transformFlags=536870912|t,-570426690&t}function R(e,n){var t=n;return t|=2097344,262144&n&&(t|=8),e.type&&(t|=3),e.transformFlags=536870912|t,-637535554&t}function F(n,t){var r,a=n.declarationList.transformFlags;return e.hasModifier(n,2)?r=3:(r=t,2097152&a&&(r|=192)),n.transformFlags=536870912|r,-637535554&r}function G(n,t){var r=t;return 1048576&t&&e.isIterationStatement(n,!0)&&(r|=192),n.transformFlags=536870912|r,-637535554&r}function V(n,t){var r=t;return e.isExternalModuleImportEqualsDeclaration(n)||(r|=3),n.transformFlags=536870912|r,-637535554&r}function B(e,n){var t=n;return 1024&e.expression.transformFlags&&(t|=192),e.transformFlags=536870912|t,-637535554&t}function K(n,t){var r=3,a=e.getModifierFlags(n);return 0===(2&a)&&(r|=t),n.transformFlags=536870912|r,-647001410&r}function j(e,n){var t=8388608|n;return 2097152&n&&(t|=192),3&e.flags&&(t|=1048768),e.transformFlags=536870912|t,-639894850&t}function U(e,n,t){var r=t,a=637535553;switch(n){case 121:case 201:r|=24;break;case 194:case 212:case 308:r|=3,a=536872257;break;case 115:case 113:case 114:case 118:case 125:case 77:case 243:case 278:case 213:case 133:r|=3;break;case 260:case 261:case 262:case 11:case 263:case 264:case 265:case 266:case 267:case 268:case 269:case 270:r|=4;break;case 14:case 15:case 16:case 17:case 206:case 193:case 276:case 116:case 214:r|=192;break;case 10:e.hasExtendedUnicodeEscape&&(r|=192);break;case 8:384&e.numericLiteralFlags&&(r|=192);break;case 9:r|=8;break;case 227:e.awaitModifier&&(r|=8),r|=192;break;case 207:r|=4194504;break;case 120:case 135:case 146:case 132:case 136:case 138:case 123:case 139:case 106:case 150:case 153:case 155:case 160:case 161:case 162:case 163:case 164:case 165:case 166:case 167:case 168:case 169:case 170:case 171:case 172:case 173:case 174:case 175:case 176:case 177:case 241:case 242:case 178:case 179:case 180:case 181:case 182:case 247:r=3,a=-3;break;case 149:r|=524288,8192&t&&(r|=32768);break;case 208:r|=131264;break;case 277:r|=262152;break;case 98:r|=33554624,a=536872257;break;case 100:r|=8192;break;case 184:r|=2097344,131072&t&&(r|=262152),a=637666625;break;case 185:r|=2097344,a=637666625;break;case 186:r|=192,e.dotDotDotToken&&(r|=131072);break;case 152:r|=4099;break;case 188:a=638358849,524288&t&&(r|=192),32768&t&&(r|=8192),262144&t&&(r|=8);break;case 187:case 192:a=637666625,131072&t&&(r|=192);break;case 223:case 224:case 225:case 226:1048576&t&&(r|=192);break;case 279:16384&t&&(r|=192);break;case 230:r|=8388616;break;case 228:case 229:r|=8388608;break}return e.transformFlags=536870912|r,r&~a}function H(e){if(e>=163&&e<=183)return-3;switch(e){case 191:case 192:case 187:return 637666625;case 244:return 647001409;case 151:return 637535553;case 197:return 653604161;case 196:case 239:return 653620545;case 238:return 639894849;case 240:case 209:return 638121281;case 157:return 653616449;case 156:case 158:case 159:return 653616449;case 120:case 135:case 146:case 132:case 138:case 136:case 123:case 139:case 106:case 150:case 153:case 155:case 160:case 161:case 162:case 241:case 242:return-3;case 188:return 638358849;case 274:return 637797697;case 184:case 185:return 637666625;case 194:case 212:case 308:case 195:case 98:return 536872257;case 189:case 190:return 570426689;default:return 637535553}}function W(n,t){t.parent=n,e.forEachChild(t,function(e){return W(t,e)})}e.bindSourceFile=s,e.isExportsOrModuleExportsOrAlias=p,e.computeTransformFlagsForNode=g,e.getTransformFlagsSubtreeExclusions=H}(u||(u={})),function(e){function n(n,t,r,a,i,o,s,c,l,u){return d;function d(d){void 0===d&&(d=function(){return!0});var p=[],m=[];return{walkType:function(n){try{return f(n),{visitedTypes:e.getOwnValues(p),visitedSymbols:e.getOwnValues(m)}}finally{e.clear(p),e.clear(m)}},walkSymbol:function(n){try{return x(n),{visitedTypes:e.getOwnValues(p),visitedSymbols:e.getOwnValues(m)}}finally{e.clear(p),e.clear(m)}}};function f(e){if(e&&!p[e.id]){p[e.id]=e;var n=x(e.symbol);if(!n){if(524288&e.flags){var t=e,r=t.objectFlags;4&r&&_(e),32&r&&b(e),3&r&&T(e),24&r&&S(t)}262144&e.flags&&g(e),3145728&e.flags&&v(e),4194304&e.flags&&y(e),8388608&e.flags&&h(e)}}}function _(n){f(n.target),e.forEach(n.typeArguments,f)}function g(e){f(l(e))}function v(n){e.forEach(n.types,f)}function y(e){f(e.type)}function h(e){f(e.objectType),f(e.indexType),f(e.constraint)}function b(e){f(e.typeParameter),f(e.constraintType),f(e.templateType),f(e.modifiersType)}function E(a){var i=t(a);i&&f(i.type),e.forEach(a.typeParameters,f);for(var o=0,s=a.parameters;o<s.length;o++){var c=s[o];x(c)}f(n(a)),f(r(a))}function T(n){S(n),e.forEach(n.typeParameters,f),e.forEach(a(n),f),f(n.thisType)}function S(e){var n=c(e,0);f(n);var t=c(e,1);f(t);for(var r=i(e),a=0,o=r.callSignatures;a<o.length;a++){var s=o[a];E(s)}for(var l=0,u=r.constructSignatures;l<u.length;l++){s=u[l];E(s)}for(var d=0,p=r.properties;d<p.length;d++){var m=p[d];x(m)}}function x(n){if(!n)return!1;var t=e.getSymbolId(n);if(m[t])return!1;if(m[t]=n,!d(n))return!0;var r=o(n);return f(r),n.exports&&n.exports.forEach(x),e.forEach(n.declarations,function(e){if(e.type&&167===e.type.kind){var n=e.type,t=s(u(n.exprName));x(t)}}),!1}}}e.createGetSymbolWalker=n}(u||(u={})),function(e){var n,t=/^".+"$/,r=1,a=1,i=1,c=1;function l(e){return e.id||(e.id=a,a++),e.id}function u(e){return e.id||(e.id=r,r++),e.id}function d(n,t){var r=e.getModuleInstanceState(n);return 1===r||t&&2===r}function p(r,a){var p,_,g,v=e.memoize(function(){var n=e.createMap();return r.getSourceFiles().forEach(function(t){t.resolvedModules&&e.forEachEntry(t.resolvedModules,function(e){e&&e.packageId&&n.set(e.packageId.name,!0)})}),n}),y=e.objectAllocator.getSymbolConstructor(),h=e.objectAllocator.getTypeConstructor(),b=e.objectAllocator.getSignatureConstructor(),E=0,T=0,S=0,x=0,C=0,A=e.createSymbolTable(),D=e.identity,L=r.getCompilerOptions(),k=e.getEmitScriptTarget(L),N=e.getEmitModuleKind(L),I=e.getAllowSyntheticDefaultImports(L),w=e.getStrictOptionValue(L,"strictNullChecks"),M=e.getStrictOptionValue(L,"strictFunctionTypes"),O=e.getStrictOptionValue(L,"strictBindCallApply"),P=e.getStrictOptionValue(L,"strictPropertyInitialization"),R=e.getStrictOptionValue(L,"noImplicitAny"),F=e.getStrictOptionValue(L,"noImplicitThis"),G=!!L.keyofStringsOnly,V=L.suppressExcessPropertyErrors?0:32768,B=pA(),K=Va(),j=Ot(4,"undefined");j.declarations=[];var U,H=Ot(4,"arguments"),W=Ot(4,"require"),q={getNodeCount:function(){return e.sum(r.getSourceFiles(),"nodeCount")},getIdentifierCount:function(){return e.sum(r.getSourceFiles(),"identifierCount")},getSymbolCount:function(){return e.sum(r.getSourceFiles(),"symbolCount")+T},getTypeCount:function(){return E},isUndefinedSymbol:function(e){return e===j},isArgumentsSymbol:function(e){return e===H},isUnknownSymbol:function(e){return e===ne},getMergedSymbol:ta,getDiagnostics:tC,getGlobalDiagnostics:aC,getTypeOfSymbolAtLocation:function(n,t){return t=e.getParseTreeNode(t),t?_g(n,t):oe},getSymbolsOfParameterPropertyDeclaration:function(n,t){var r=e.getParseTreeNode(n,e.isParameter);return void 0===r?e.Debug.fail("Cannot get symbols of a synthetic parameter that cannot be resolved to a parse-tree node."):Yt(r,e.escapeLeadingUnderscores(t))},getDeclaredTypeOfSymbol:Do,getPropertiesOfType:Rs,getPropertyOfType:function(n,t){return sc(n,e.escapeLeadingUnderscores(t))},getTypeOfPropertyOfType:function(n,t){return ti(n,e.escapeLeadingUnderscores(t))},getIndexInfoOfType:pc,getSignaturesOfType:lc,getIndexTypeOfType:mc,getBaseTypes:uo,getBaseTypeOfLiteralType:Wm,getWidenedType:bf,getTypeFromTypeNode:function(n){var t=e.getParseTreeNode(n,e.isTypeNode);return t?Rd(t):oe},getParameterType:Cb,getPromisedTypeOfPromise:OT,getReturnTypeOfSignature:Fc,getNullableType:of,getNonNullableType:lf,typeToTypeNode:K.typeToTypeNode,indexInfoToIndexSignatureDeclaration:K.indexInfoToIndexSignatureDeclaration,signatureToSignatureDeclaration:K.signatureToSignatureDeclaration,symbolToEntityName:K.symbolToEntityName,symbolToExpression:K.symbolToExpression,symbolToTypeParameterDeclarations:K.symbolToTypeParameterDeclarations,symbolToParameterDeclaration:K.symbolToParameterDeclaration,typeParameterToDeclaration:K.typeParameterToDeclaration,getSymbolsInScope:function(n,t){return n=e.getParseTreeNode(n),n?oC(n,t):[]},getSymbolAtLocation:function(n){return n=e.getParseTreeNode(n),n?hC(n):void 0},getShorthandAssignmentValueSymbol:function(n){return n=e.getParseTreeNode(n),n?bC(n):void 0},getExportSpecifierLocalTargetSymbol:function(n){var t=e.getParseTreeNode(n,e.isExportSpecifier);return t?EC(t):void 0},getExportSymbolOfSymbol:function(e){return ta(e.exportSymbol||e)},getTypeAtLocation:function(n){return n=e.getParseTreeNode(n),n?TC(n):oe},getPropertySymbolOfDestructuringAssignment:function(n){var t=e.getParseTreeNode(n,e.isIdentifier);return t?xC(t):void 0},signatureToString:function(n,t,r,a){return Ra(n,e.getParseTreeNode(t),r,a)},typeToString:function(n,t,r){return Fa(n,e.getParseTreeNode(t),r)},symbolToString:function(n,t,r,a){return Pa(n,e.getParseTreeNode(t),r,a)},typePredicateToString:function(n,t,r){return Ba(n,e.getParseTreeNode(t),r)},writeSignature:function(n,t,r,a,i){return Ra(n,e.getParseTreeNode(t),r,a,i)},writeType:function(n,t,r,a){return Fa(n,e.getParseTreeNode(t),r,a)},writeSymbol:function(n,t,r,a,i){return Pa(n,e.getParseTreeNode(t),r,a,i)},writeTypePredicate:function(n,t,r,a){return Ba(n,e.getParseTreeNode(t),r,a)},getAugmentedPropertiesOfType:LC,getRootSymbols:NC,getContextualType:function(n){var t=e.getParseTreeNode(n,e.isExpression);return t?bv(t):void 0},getContextualTypeForObjectLiteralElement:function(n){var t=e.getParseTreeNode(n,e.isObjectLiteralElementLike);return t?uv(t):void 0},getContextualTypeForArgumentAtIndex:function(n,t){var r=e.getParseTreeNode(n,e.isCallLikeExpression);return r&&rv(r,t)},getContextualTypeForJsxAttribute:function(n){var t=e.getParseTreeNode(n,e.isJsxAttributeLike);return t&&_v(t)},isContextSensitive:yp,getFullyQualifiedName:Or,getResolvedSignature:function(e,n,t){return z(e,n,t,!1)},getResolvedSignatureForSignatureHelp:function(e,n,t){return z(e,n,t,!0)},getConstantValue:function(n){var t=e.getParseTreeNode(n,$C);return t?eA(t):void 0},isValidPropertyAccess:function(n,t){var r=e.getParseTreeNode(n,e.isPropertyAccessOrQualifiedNameOrImportTypeNode);return!!r&&Qy(r,e.escapeLeadingUnderscores(t))},isValidPropertyAccessForCompletions:function(n,t,r){var a=e.getParseTreeNode(n,e.isPropertyAccessExpression);return!!a&&Zy(a,t,r)},getSignatureFromDeclaration:function(n){var t=e.getParseTreeNode(n,e.isFunctionLike);return t?Cc(t):void 0},isImplementationOfOverload:function(n){var t=e.getParseTreeNode(n,e.isFunctionLike);return t?qC(t):void 0},getImmediateAliasedSymbol:qv,getAliasedSymbol:Nr,getEmitResolver:Lt,getExportsOfModule:Jr,getExportsAndPropertiesOfModule:Xr,getSymbolWalker:e.createGetSymbolWalker(Bc,Oc,Fc,uo,ws,Xi,e_,dc,Bs,wx),getAmbientModules:kD,getJsxIntrinsicTagNamesAt:by,isOptionalParameter:function(n){var t=e.getParseTreeNode(n,e.isParameter);return!!t&&hc(t)},tryGetMemberInModuleExports:function(n,t){return Yr(e.escapeLeadingUnderscores(n),t)},tryGetMemberInModuleExportsAndProperties:function(n,t){return Qr(e.escapeLeadingUnderscores(n),t)},tryFindAmbientModuleWithoutAugmentations:function(e){return yc(e,!1)},getApparentType:rc,getUnionType:Au,createAnonymousType:ba,createSignature:$o,createSymbol:Ot,createIndexInfo:el,getAnyType:function(){return re},getStringType:function(){return pe},getNumberType:function(){return me},createPromiseType:Vb,createArrayType:su,getElementTypeOfArrayType:Rm,getBooleanType:function(){return he},getFalseType:function(e){return e?_e:ge},getTrueType:function(e){return e?ve:ye},getVoidType:function(){return Ee},getUndefinedType:function(){return ce},getNullType:function(){return ue},getESSymbolType:function(){return be},getNeverType:function(){return Te},isSymbolAccessible:ka,getObjectFlags:e.getObjectFlags,isArrayLikeType:Fm,isTypeInvalidDueToUnionDiscriminant:Fs,getAllPossiblePropertiesOfTypes:Gs,getSuggestionForNonexistentProperty:function(e,n){return Hy(e,n)},getSuggestionForNonexistentSymbol:function(n,t,r){return qy(n,e.escapeLeadingUnderscores(t),r)},getSuggestionForNonexistentExport:function(e,n){return Jy(e,n)},getBaseConstraintOfType:qs,getDefaultFromTypeParameter:function(e){return e&&262144&e.flags?Zs(e):void 0},resolveName:function(n,t,r,a){return Zt(t,e.escapeLeadingUnderscores(n),r,void 0,void 0,!1,a)},getJsxNamespace:function(n){return e.unescapeLeadingUnderscores(Dt(n))},getAccessibleSymbolChain:Sa,getTypePredicateOfSignature:Oc,resolveExternalModuleSymbol:Hr,tryGetThisTypeAt:function(n){return n=e.getParseTreeNode(n),n&&Gg(n)},getTypeArgumentConstraint:function(n){var t=e.getParseTreeNode(n,e.isTypeNode);return t&&_T(t)},getSuggestionDiagnostics:function(n,t){if(e.skipTypeChecking(n,L))return e.emptyArray;var r;try{return p=t,Zx(n),e.Debug.assert(!!(1&zt(n).flags)),r=e.addRange(r,ot.get(n.fileName)),n.isDeclarationFile||$x(0)&&$x(1)||a(),r||e.emptyArray}finally{p=void 0}function a(){tS(eC(n),function(n,t,a){e.containsParseError(n)||$x(t)||(r||(r=[])).push(o({},a,{category:e.DiagnosticCategory.Suggestion}))})}},runWithCancellationToken:function(e,n){try{return p=e,n(q)}finally{p=void 0}},getLocalTypeParametersOfClassOrInterfaceOrTypeAlias:to};function z(n,t,r,a){var i=e.getParseTreeNode(n,e.isCallLikeExpression);U=r;var o=i?ab(i,t,a):void 0;return U=void 0,o}var J=e.createMap(),X=e.createMap(),Y=e.createMap(),Q=e.createMap(),Z=e.createMap(),$=[],ee=e.createMap(),ne=Ot(4,"unknown"),te=Ot(0,"__resolving__"),re=pa(1,"any"),ae=pa(1,"any"),ie=pa(1,"any"),oe=pa(1,"error"),se=pa(2,"unknown"),ce=pa(32768,"undefined"),le=w?ce:pa(134250496,"undefined"),ue=pa(65536,"null"),de=w?ue:pa(134283264,"null"),pe=pa(4,"string"),me=pa(8,"number"),fe=pa(64,"bigint"),_e=pa(512,"false"),ge=pa(512,"false"),ve=pa(512,"true"),ye=pa(512,"true");ve.regularType=ye,ve.freshType=ve,ye.regularType=ye,ye.freshType=ve,_e.regularType=ge,_e.freshType=_e,ge.regularType=ge,ge.freshType=_e;var he=ma([ge,ye]);ma([ge,ve]),ma([_e,ye]),ma([_e,ve]);var be=pa(4096,"symbol"),Ee=pa(16384,"void"),Te=pa(131072,"never"),Se=pa(131072,"never"),xe=pa(131072,"never"),Ce=pa(67108864,"object"),Ae=Au([pe,me,be]),De=G?pe:Ae,Le=Au([me,fe]),ke=ba(void 0,A,e.emptyArray,e.emptyArray,void 0,void 0),Ne=ba(void 0,A,e.emptyArray,e.emptyArray,void 0,void 0);Ne.objectFlags|=4096;var Ie=Ot(2048,"__type");Ie.members=e.createSymbolTable();var we=ba(Ie,A,e.emptyArray,e.emptyArray,void 0,void 0),Me=ba(void 0,A,e.emptyArray,e.emptyArray,void 0,void 0);Me.instantiations=e.createMap();var Oe=ba(void 0,A,e.emptyArray,e.emptyArray,void 0,void 0);Oe.flags|=536870912;var Pe=ba(void 0,A,e.emptyArray,e.emptyArray,void 0,void 0),Re=ba(void 0,A,e.emptyArray,e.emptyArray,void 0,void 0),Fe=ba(void 0,A,e.emptyArray,e.emptyArray,void 0,void 0),Ge=ga(),Ve=ga();Ve.constraint=Ge;var Be,Ke,je,Ue,He,We,qe,ze,Je,Xe,Ye,Qe,Ze,$e,en,nn,tn,rn,an,on,sn,cn,ln,un,dn,pn,mn,fn,_n,gn,vn,yn,hn,bn,En,Tn,Sn,xn,Cn=ga(),An=Ec("<<unresolved>>",0,re),Dn=$o(void 0,void 0,void 0,e.emptyArray,re,void 0,0,!1,!1),Ln=$o(void 0,void 0,void 0,e.emptyArray,oe,void 0,0,!1,!1),kn=$o(void 0,void 0,void 0,e.emptyArray,re,void 0,0,!1,!1),Nn=$o(void 0,void 0,void 0,e.emptyArray,Se,void 0,0,!1,!1),In=el(pe,!0),wn=e.createSymbolTable(),Mn=e.createMap(),On=e.createMap(),Pn=0,Rn=0,Fn=0,Gn=!1,Vn=Nd(""),Bn=Nd(0),Kn=Nd({negative:!1,base10Value:"0"}),jn=[],Un=[],Hn=[],Wn=0,qn=10,zn=[],Jn=[],Xn=[],Yn=[],Qn=[],Zn=[],$n=[],et=[],nt=[],tt=[],rt=[],at=[],it=e.createDiagnosticCollection(),ot=e.createMultiMap();(function(e){e[e["None"]=0]="None",e[e["TypeofEQString"]=1]="TypeofEQString",e[e["TypeofEQNumber"]=2]="TypeofEQNumber",e[e["TypeofEQBigInt"]=4]="TypeofEQBigInt",e[e["TypeofEQBoolean"]=8]="TypeofEQBoolean",e[e["TypeofEQSymbol"]=16]="TypeofEQSymbol",e[e["TypeofEQObject"]=32]="TypeofEQObject",e[e["TypeofEQFunction"]=64]="TypeofEQFunction",e[e["TypeofEQHostObject"]=128]="TypeofEQHostObject",e[e["TypeofNEString"]=256]="TypeofNEString",e[e["TypeofNENumber"]=512]="TypeofNENumber",e[e["TypeofNEBigInt"]=1024]="TypeofNEBigInt",e[e["TypeofNEBoolean"]=2048]="TypeofNEBoolean",e[e["TypeofNESymbol"]=4096]="TypeofNESymbol",e[e["TypeofNEObject"]=8192]="TypeofNEObject",e[e["TypeofNEFunction"]=16384]="TypeofNEFunction",e[e["TypeofNEHostObject"]=32768]="TypeofNEHostObject",e[e["EQUndefined"]=65536]="EQUndefined",e[e["EQNull"]=131072]="EQNull",e[e["EQUndefinedOrNull"]=262144]="EQUndefinedOrNull",e[e["NEUndefined"]=524288]="NEUndefined",e[e["NENull"]=1048576]="NENull",e[e["NEUndefinedOrNull"]=2097152]="NEUndefinedOrNull",e[e["Truthy"]=4194304]="Truthy",e[e["Falsy"]=8388608]="Falsy",e[e["All"]=16777215]="All",e[e["BaseStringStrictFacts"]=3735041]="BaseStringStrictFacts",e[e["BaseStringFacts"]=12582401]="BaseStringFacts",e[e["StringStrictFacts"]=16317953]="StringStrictFacts",e[e["StringFacts"]=16776705]="StringFacts",e[e["EmptyStringStrictFacts"]=12123649]="EmptyStringStrictFacts",e[e["EmptyStringFacts"]=12582401]="EmptyStringFacts",e[e["NonEmptyStringStrictFacts"]=7929345]="NonEmptyStringStrictFacts",e[e["NonEmptyStringFacts"]=16776705]="NonEmptyStringFacts",e[e["BaseNumberStrictFacts"]=3734786]="BaseNumberStrictFacts",e[e["BaseNumberFacts"]=12582146]="BaseNumberFacts",e[e["NumberStrictFacts"]=16317698]="NumberStrictFacts",e[e["NumberFacts"]=16776450]="NumberFacts",e[e["ZeroNumberStrictFacts"]=12123394]="ZeroNumberStrictFacts",e[e["ZeroNumberFacts"]=12582146]="ZeroNumberFacts",e[e["NonZeroNumberStrictFacts"]=7929090]="NonZeroNumberStrictFacts",e[e["NonZeroNumberFacts"]=16776450]="NonZeroNumberFacts",e[e["BaseBigIntStrictFacts"]=3734276]="BaseBigIntStrictFacts",e[e["BaseBigIntFacts"]=12581636]="BaseBigIntFacts",e[e["BigIntStrictFacts"]=16317188]="BigIntStrictFacts",e[e["BigIntFacts"]=16775940]="BigIntFacts",e[e["ZeroBigIntStrictFacts"]=12122884]="ZeroBigIntStrictFacts",e[e["ZeroBigIntFacts"]=12581636]="ZeroBigIntFacts",e[e["NonZeroBigIntStrictFacts"]=7928580]="NonZeroBigIntStrictFacts",e[e["NonZeroBigIntFacts"]=16775940]="NonZeroBigIntFacts",e[e["BaseBooleanStrictFacts"]=3733256]="BaseBooleanStrictFacts",e[e["BaseBooleanFacts"]=12580616]="BaseBooleanFacts",e[e["BooleanStrictFacts"]=16316168]="BooleanStrictFacts",e[e["BooleanFacts"]=16774920]="BooleanFacts",e[e["FalseStrictFacts"]=12121864]="FalseStrictFacts",e[e["FalseFacts"]=12580616]="FalseFacts",e[e["TrueStrictFacts"]=7927560]="TrueStrictFacts",e[e["TrueFacts"]=16774920]="TrueFacts",e[e["SymbolStrictFacts"]=7925520]="SymbolStrictFacts",e[e["SymbolFacts"]=16772880]="SymbolFacts",e[e["ObjectStrictFacts"]=7888800]="ObjectStrictFacts",e[e["ObjectFacts"]=16736160]="ObjectFacts",e[e["FunctionStrictFacts"]=7880640]="FunctionStrictFacts",e[e["FunctionFacts"]=16728e3]="FunctionFacts",e[e["UndefinedFacts"]=9830144]="UndefinedFacts",e[e["NullFacts"]=9363232]="NullFacts",e[e["EmptyObjectStrictFacts"]=16318463]="EmptyObjectStrictFacts",e[e["EmptyObjectFacts"]=16777215]="EmptyObjectFacts"})(xn||(xn={}));var st,ct,lt,ut,dt,pt,mt,ft,_t,gt=e.createMapFromTemplate({string:1,number:2,bigint:4,boolean:8,symbol:16,undefined:65536,object:32,function:64}),vt=e.createMapFromTemplate({string:256,number:512,bigint:1024,boolean:2048,symbol:4096,undefined:524288,object:8192,function:16384}),yt=e.createMapFromTemplate({string:pe,number:me,bigint:fe,boolean:he,symbol:be,undefined:ce}),ht=_a(),bt=e.createMap(),Et=e.createMap(),Tt=e.createMap(),St=e.createMap(),xt=e.createMap();(function(e){e[e["Type"]=0]="Type",e[e["ResolvedBaseConstructorType"]=1]="ResolvedBaseConstructorType",e[e["DeclaredType"]=2]="DeclaredType",e[e["ResolvedReturnType"]=3]="ResolvedReturnType",e[e["ImmediateBaseConstraint"]=4]="ImmediateBaseConstraint",e[e["EnumTagType"]=5]="EnumTagType",e[e["JSDocTypeReference"]=6]="JSDocTypeReference"})(lt||(lt={})),function(e){e[e["Normal"]=0]="Normal",e[e["SkipContextSensitive"]=1]="SkipContextSensitive",e[e["Inferential"]=2]="Inferential",e[e["Contextual"]=3]="Contextual"}(ut||(ut={})),function(e){e[e["None"]=0]="None",e[e["Bivariant"]=1]="Bivariant",e[e["Strict"]=2]="Strict"}(dt||(dt={})),function(e){e[e["IncludeReadonly"]=1]="IncludeReadonly",e[e["ExcludeReadonly"]=2]="ExcludeReadonly",e[e["IncludeOptional"]=4]="IncludeOptional",e[e["ExcludeOptional"]=8]="ExcludeOptional"}(pt||(pt={})),function(e){e[e["None"]=0]="None",e[e["Source"]=1]="Source",e[e["Target"]=2]="Target",e[e["Both"]=3]="Both"}(mt||(mt={})),function(e){e["resolvedExports"]="resolvedExports",e["resolvedMembers"]="resolvedMembers"}(ft||(ft={})),function(e){e[e["Local"]=0]="Local",e[e["Parameter"]=1]="Parameter"}(_t||(_t={}));var Ct=e.createSymbolTable();Ct.set(j.escapedName,j);var At=e.and(Wx,Hx);return fA(),q;function Dt(n){if(n){var t=e.getSourceFileOfNode(n);if(t){if(t.localJsxNamespace)return t.localJsxNamespace;var r=t.pragmas.get("jsx");if(r){var a=e.isArray(r)?r[0]:r;if(t.localJsxFactory=e.parseIsolatedEntityName(a.arguments.factory,k),t.localJsxFactory)return t.localJsxNamespace=wx(t.localJsxFactory).escapedText}}}return st||(st="React",L.jsxFactory?(ct=e.parseIsolatedEntityName(L.jsxFactory,k),ct&&(st=wx(ct).escapedText)):L.reactNamespace&&(st=e.escapeLeadingUnderscores(L.reactNamespace))),st}function Lt(e,n){return tC(e,n),B}function kt(n,t,r,a,i,o){var s=n?e.createDiagnosticForNode(n,t,r,a,i,o):e.createCompilerDiagnostic(t,r,a,i,o),c=it.lookup(s);return c||(it.add(s),s)}function Nt(e){for(var n,t=[],r=1;r<arguments.length;r++)t[r-1]=arguments[r];return e.relatedInformation||(e.relatedInformation=[]),(n=e.relatedInformation).push.apply(n,t),e}function It(n,t,r,a,i,o){var s=n?e.createDiagnosticForNode(n,t,r,a,i,o):e.createCompilerDiagnostic(t,r,a,i,o);return it.add(s),s}function wt(n,t){n?it.add(t):ot.add(t.file.fileName,o({},t,{category:e.DiagnosticCategory.Suggestion}))}function Mt(n,t,r,a,i,o,s){wt(n,"message"in r?e.createDiagnosticForNode(t,r,a,i,o,s):e.createDiagnosticForNodeFromMessageChain(t,r))}function Ot(e,n,t){T++;var r=new y(33554432|e,n);return r.checkFlags=t||0,r}function Pt(e){return 0!==(33554432&e.flags)}function Rt(e){var n=0;return 2&e&&(n|=67220415),1&e&&(n|=67220414),4&e&&(n|=0),8&e&&(n|=68008959),16&e&&(n|=67219887),32&e&&(n|=68008383),64&e&&(n|=67897736),256&e&&(n|=68008191),128&e&&(n|=68008831),512&e&&(n|=110735),8192&e&&(n|=67212223),32768&e&&(n|=67154879),65536&e&&(n|=67187647),262144&e&&(n|=67635688),524288&e&&(n|=67897832),2097152&e&&(n|=2097152),n}function Ft(e,n){n.mergeId||(n.mergeId=i,i++),zn[n.mergeId]=e}function Gt(n){var t=Ot(n.flags,n.escapedName);return t.declarations=n.declarations?n.declarations.slice():[],t.parent=n.parent,n.valueDeclaration&&(t.valueDeclaration=n.valueDeclaration),n.constEnumOnlyModule&&(t.constEnumOnlyModule=!0),n.members&&(t.members=e.cloneMap(n.members)),n.exports&&(t.exports=e.cloneMap(n.exports)),Ft(t,n),t}function Vt(n,t){if(!(n.flags&Rt(t.flags))||67108864&(t.flags|n.flags)){if(e.Debug.assert(t!==n),!(33554432&n.flags)){var r=kr(n);if(r===ne)return t;n=Gt(r)}512&t.flags&&512&n.flags&&n.constEnumOnlyModule&&!t.constEnumOnlyModule&&(n.constEnumOnlyModule=!1),n.flags|=t.flags,t.valueDeclaration&&(!n.valueDeclaration||e.isAssignmentDeclaration(n.valueDeclaration)&&!e.isAssignmentDeclaration(t.valueDeclaration)||e.isEffectiveModuleDeclaration(n.valueDeclaration)&&!e.isEffectiveModuleDeclaration(t.valueDeclaration))&&(n.valueDeclaration=t.valueDeclaration),e.addRange(n.declarations,t.declarations),t.members&&(n.members||(n.members=e.createSymbolTable()),Ut(n.members,t.members)),t.exports&&(n.exports||(n.exports=e.createSymbolTable()),Ut(n.exports,t.exports)),Ft(n,t)}else if(1024&n.flags)It(e.getNameOfDeclaration(t.declarations[0]),e.Diagnostics.Cannot_augment_module_0_with_value_exports_because_it_resolves_to_a_non_module_entity,Pa(n));else{var a=!!(384&n.flags||384&t.flags),i=!!(2&n.flags||2&t.flags),o=a?e.Diagnostics.Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations:i?e.Diagnostics.Cannot_redeclare_block_scoped_variable_0:e.Diagnostics.Duplicate_identifier_0,s=t.declarations&&e.getSourceFileOfNode(t.declarations[0]),c=n.declarations&&e.getSourceFileOfNode(n.declarations[0]),l=Pa(t);if(s&&c&&Be&&!a&&s!==c){var u=-1===e.comparePaths(s.path,c.path)?s:c,d=u===s?c:s,p=e.getOrUpdate(Be,u.path+"|"+d.path,function(){return{firstFile:u,secondFile:d,conflictingSymbols:e.createMap()}}),m=e.getOrUpdate(p.conflictingSymbols,l,function(){return{isBlockScoped:i,firstFileLocations:[],secondFileLocations:[]}});f(m.firstFileLocations,t),f(m.secondFileLocations,n)}else Bt(t,o,l,n),Bt(n,o,l,t)}return n;function f(n,t){for(var r=0,a=t.declarations;r<a.length;r++){var i=a[r];e.pushIfUnique(n,(e.getExpandoInitializer(i,!1)?e.getNameOfExpando(i):e.getNameOfDeclaration(i))||i)}}}function Bt(n,t,r,a){e.forEach(n.declarations,function(n){var i=(e.getExpandoInitializer(n,!1)?e.getNameOfExpando(n):e.getNameOfDeclaration(n))||n;Kt(i,t,r,a.declarations)})}function Kt(n,t,r,a){for(var i=kt(n,t,r),o=0,s=a||e.emptyArray;o<s.length;o++){var c=s[o];i.relatedInformation=i.relatedInformation||[],e.length(i.relatedInformation)>=5||Nt(i,e.length(i.relatedInformation)?e.createDiagnosticForNode(c,e.Diagnostics.and_here):e.createDiagnosticForNode(c,e.Diagnostics._0_was_also_declared_here,r))}}function jt(n,t){if(!e.hasEntries(n))return t;if(!e.hasEntries(t))return n;var r=e.createSymbolTable();return Ut(r,n),Ut(r,t),r}function Ut(e,n){n.forEach(function(n,t){var r=e.get(t);e.set(t,r?Vt(r,n):n)})}function Ht(n){var t=n.parent;if(t.symbol.declarations[0]===t)if(e.isGlobalScopeAugmentation(t))Ut(wn,t.symbol.exports);else{var r=4194304&n.parent.parent.flags?void 0:e.Diagnostics.Invalid_module_name_in_augmentation_module_0_cannot_be_found,a=Br(n,n,r,!0);if(!a)return;a=Hr(a),1920&a.flags?a=Vt(a,t.symbol):It(n,e.Diagnostics.Cannot_augment_module_0_because_it_resolves_to_a_non_module_entity,n.text)}else e.Debug.assert(t.symbol.declarations.length>1)}function Wt(n,t,r){function a(n,t){return function(r){return it.add(e.createDiagnosticForNode(r,t,n))}}t.forEach(function(t,i){var o=n.get(i);o?e.forEach(o.declarations,a(e.unescapeLeadingUnderscores(i),r)):n.set(i,t)})}function qt(e){if(33554432&e.flags)return e;var n=u(e);return Jn[n]||(Jn[n]={})}function zt(e){var n=l(e);return Xn[n]||(Xn[n]={flags:0})}function Jt(n){return 279===n.kind&&!e.isExternalOrCommonJsModule(n)}function Xt(n,t,r){if(r){var a=n.get(t);if(a){if(e.Debug.assert(0===(1&e.getCheckFlags(a)),"Should never get an instantiated symbol here."),a.flags&r)return a;if(2097152&a.flags){var i=Nr(a);if(i===ne||i.flags&r)return a}}}}function Yt(n,t){var r=n.parent,a=n.parent.parent,i=Xt(r.locals,t,67220415),o=Xt(zo(a.symbol),t,67220415);return i&&o?[i,o]:e.Debug.fail("There should exist two symbols, one as property declaration and one as parameter declaration")}function Qt(n,t){var a=e.getSourceFileOfNode(n),i=e.getSourceFileOfNode(t);if(a!==i){if(N&&(a.externalModuleIndicator||i.externalModuleIndicator)||!L.outFile&&!L.out||n_(t)||4194304&n.flags)return!0;if(u(t,n))return!0;var o=r.getSourceFiles();return o.indexOf(a)<=o.indexOf(i)}if(n.pos<=t.pos){if(186===n.kind){var s=e.getAncestor(t,186);return s?e.findAncestor(s,e.isBindingElement)!==e.findAncestor(n,e.isBindingElement)||n.pos<s.pos:Qt(e.getAncestor(n,237),t)}return 237===n.kind?!l(n,t):!e.isClassDeclaration(n)||!e.findAncestor(t,function(t){return e.isComputedPropertyName(t)&&t.parent.parent===n})}if(257===t.parent.kind||254===t.parent.kind&&t.parent.isExportEquals)return!0;if(254===t.kind&&t.isExportEquals)return!0;var c=e.getEnclosingBlockScopeContainer(n);return!!(2097152&t.flags)||n_(t)||u(t,n,c);function l(n,t){var r=e.getEnclosingBlockScopeContainer(n);switch(n.parent.parent.kind){case 219:case 225:case 227:if(ur(t,n,r))return!0;break}var a=n.parent.parent;return e.isForInOrOfStatement(a)&&ur(t,a.expression,r)}function u(n,t,r){return!!e.findAncestor(n,function(a){if(a===r)return"quit";if(e.isFunctionLike(a))return!0;var i=a.parent&&154===a.parent.kind&&a.parent.initializer===a;if(i)if(e.hasModifier(a.parent,32)){if(156===t.kind)return!0}else{var o=154===t.kind&&!e.hasModifier(t,32);if(!o||e.getContainingClass(n)!==e.getContainingClass(t))return!0}return!1})}}function Zt(e,n,t,r,a,i,o,s){return void 0===o&&(o=!1),$t(e,n,t,r,a,i,o,Xt,s)}function $t(n,t,r,a,i,o,s,c,l){var u,d,p,m,f,_=n,g=n,v=!1;e:while(n){if(n.locals&&!Jt(n)&&(u=c(n.locals,t,r))){var y=!0;if(e.isFunctionLike(n)&&d&&d!==n.body){if(r&u.flags&67897832&&291!==d.kind&&(y=!!(262144&u.flags)&&(d===n.type||151===d.kind||150===d.kind)),r&u.flags&3){var h=n;L.target&&L.target>=2&&e.isParameter(d)&&h.body&&u.valueDeclaration.pos>=h.body.pos&&u.valueDeclaration.end<=h.body.end?y=!1:1&u.flags&&(y=151===d.kind||d===n.type&&!!e.findAncestor(u.valueDeclaration,e.isParameter))}}else 175===n.kind&&(y=d===n.trueType);if(y)break e;u=void 0}switch(n.kind){case 279:if(!e.isExternalOrCommonJsModule(n))break;v=!0;case 244:var b=ra(n).exports;if(279===n.kind||e.isAmbientModule(n)){if(u=b.get("default")){var E=e.getLocalSymbolForExportDefault(u);if(E&&u.flags&r&&E.escapedName===t)break e;u=void 0}var T=b.get(t);if(T&&2097152===T.flags&&e.getDeclarationOfKind(T,257))break}if("default"!==t&&(u=c(b,t,2623475&r))){if(!e.isSourceFile(n)||!n.commonJsModuleIndicator||u.declarations.some(e.isJSDocTypeAlias))break e;u=void 0}break;case 243:if(u=c(ra(n).exports,t,8&r))break e;break;case 154:case 153:if(e.isClassLike(n.parent)&&!e.hasModifier(n,32)){var S=ua(n.parent);S&&S.locals&&c(S.locals,t,67220415&r)&&(m=n)}break;case 240:case 209:case 241:if(u=c(ra(n).members||A,t,67897832&r)){if(!tr(u,n)){u=void 0;break}if(d&&e.hasModifier(d,32))return void It(g,e.Diagnostics.Static_members_cannot_reference_class_type_parameters);break e}if(209===n.kind&&32&r){var x=n.name;if(x&&t===x.escapedText){u=n.symbol;break e}}break;case 211:if(d===n.expression&&86===n.parent.token){var C=n.parent.parent;if(e.isClassLike(C)&&(u=c(ra(C).members,t,67897832&r)))return void(a&&It(g,e.Diagnostics.Base_class_expressions_cannot_reference_class_type_parameters))}break;case 149:if(f=n.parent.parent,(e.isClassLike(f)||241===f.kind)&&(u=c(ra(f).members,t,67897832&r)))return void It(g,e.Diagnostics.A_computed_property_name_cannot_reference_a_type_parameter_from_its_containing_type);break;case 197:if(L.target>=2)break;case 156:case 157:case 158:case 159:case 239:if(3&r&&"arguments"===t){u=H;break e}break;case 196:if(3&r&&"arguments"===t){u=H;break e}if(16&r){var D=n.name;if(D&&t===D.escapedText){u=n.symbol;break e}}break;case 152:n.parent&&151===n.parent.kind&&(n=n.parent),n.parent&&e.isClassElement(n.parent)&&(n=n.parent);break;case 304:case 297:n=e.getJSDocHost(n);break}er(n)&&(p=n),d=n,n=n.parent}if(!o||!u||p&&u===p.symbol||(u.isReferenced|=r),!u){if(d&&(e.Debug.assert(279===d.kind),d.commonJsModuleIndicator&&"exports"===t&&r&d.symbol.flags))return d.symbol;s||(u=c(wn,t,r))}if(!u&&_&&e.isInJSFile(_)&&_.parent&&e.isRequireCall(_.parent,!1))return W;if(u){if(a){if(m){var k=m.name;return void It(g,e.Diagnostics.Initializer_of_instance_member_variable_0_cannot_reference_identifier_1_declared_in_the_constructor,e.declarationNameToString(k),nr(i))}if(g&&(2&r||(32&r||384&r)&&67220415===(67220415&r))){var N=ca(u);(2&N.flags||32&N.flags||384&N.flags)&&lr(N,g)}if(u&&v&&67220415===(67220415&r)&&!(2097152&_.flags)){var I=ta(u);e.length(I.declarations)&&e.every(I.declarations,function(n){return e.isNamespaceExportDeclaration(n)||e.isSourceFile(n)&&!!n.symbol.globalExports})&&It(g,e.Diagnostics._0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead,e.unescapeLeadingUnderscores(t))}}return u}if(a&&(!g||!rr(g,t,i)&&!ar(g)&&!or(g,t,r)&&!sr(g,t,r)&&!cr(g,t,r))){var w=void 0;if(l&&Wn<qn&&(w=Wy(_,t,r),w)){var M=Pa(w),O=It(g,l,nr(i),M);w.valueDeclaration&&Nt(O,e.createDiagnosticForNode(w.valueDeclaration,e.Diagnostics._0_is_declared_here,M))}w||It(g,a,nr(i)),Wn++}}function er(e){switch(e.kind){case 239:case 240:case 241:case 243:case 242:case 244:return!0;default:return!1}}function nr(n){return e.isString(n)?e.unescapeLeadingUnderscores(n):e.declarationNameToString(n)}function tr(n,t){for(var r=0,a=n.declarations;r<a.length;r++){var i=a[r];if(150===i.kind){var o=e.isJSDocTemplateTag(i.parent)?e.getJSDocHost(i.parent):i.parent;if(o===t)return!(e.isJSDocTemplateTag(i.parent)&&e.find(i.parent.parent.tags,e.isJSDocTypeAlias))}}return!1}function rr(n,t,r){if(!e.isIdentifier(n)||n.escapedText!==t||lC(n)||n_(n))return!1;var a=e.getThisContainer(n,!1),i=a;while(i){if(e.isClassLike(i.parent)){var o=ra(i.parent);if(!o)break;var s=Xi(o);if(sc(s,t))return It(n,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0,nr(r),Pa(o)),!0;if(i===a&&!e.hasModifier(i,32)){var c=Do(o).thisType;if(sc(c,t))return It(n,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0,nr(r)),!0}}i=i.parent}return!1}function ar(n){var t=ir(n);return!(!t||!Pr(t,64,!0))&&(It(n,e.Diagnostics.Cannot_extend_an_interface_0_Did_you_mean_implements,e.getTextOfNode(t)),!0)}function ir(n){switch(n.kind){case 72:case 189:return n.parent?ir(n.parent):void 0;case 211:if(e.isEntityNameExpression(n.expression))return n.expression;default:return}}function or(n,t,r){var a=1920|(e.isInJSFile(n)?67220415:0);if(r===a){var i=kr(Zt(n,t,67897832&~a,void 0,void 0,!1)),o=n.parent;if(i){if(e.isQualifiedName(o)){e.Debug.assert(o.left===n,"Should only be resolving left side of qualified name as a namespace");var s=o.right.escapedText,c=sc(Do(i),s);if(c)return It(o,e.Diagnostics.Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_property_1_in_0_with_0_1,e.unescapeLeadingUnderscores(t),e.unescapeLeadingUnderscores(s)),!0}return It(n,e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_namespace_here,e.unescapeLeadingUnderscores(t)),!0}}return!1}function sr(n,t,r){if(67220415&r){if("any"===t||"string"===t||"number"===t||"boolean"===t||"never"===t)return It(n,e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here,e.unescapeLeadingUnderscores(t)),!0;var a=kr(Zt(n,t,788544,void 0,void 0,!1));if(a&&!(1024&a.flags)){var i="Promise"===t||"Symbol"===t?e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_es2015_or_later:e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here;return It(n,i,e.unescapeLeadingUnderscores(t)),!0}}return!1}function cr(n,t,r){if(111127&r){var a=kr(Zt(n,t,1024,void 0,void 0,!1));if(a)return It(n,e.Diagnostics.Cannot_use_namespace_0_as_a_value,e.unescapeLeadingUnderscores(t)),!0}else if(788544&r){a=kr(Zt(n,t,1536,void 0,void 0,!1));if(a)return It(n,e.Diagnostics.Cannot_use_namespace_0_as_a_type,e.unescapeLeadingUnderscores(t)),!0}return!1}function lr(n,t){e.Debug.assert(!!(2&n.flags||32&n.flags||384&n.flags));var r=e.find(n.declarations,function(n){return e.isBlockOrCatchScoped(n)||e.isClassLike(n)||243===n.kind||e.isInJSFile(n)&&!!e.getJSDocEnumTag(n)});if(void 0===r)return e.Debug.fail("Declaration to checkResolvedBlockScopedVariable is undefined");if(!(4194304&r.flags)&&!Qt(r,t)){var a=void 0,i=e.declarationNameToString(e.getNameOfDeclaration(r));2&n.flags?a=It(t,e.Diagnostics.Block_scoped_variable_0_used_before_its_declaration,i):32&n.flags?a=It(t,e.Diagnostics.Class_0_used_before_its_declaration,i):256&n.flags?a=It(t,e.Diagnostics.Enum_0_used_before_its_declaration,i):(e.Debug.assert(!!(128&n.flags)),L.preserveConstEnums&&(a=It(t,e.Diagnostics.Class_0_used_before_its_declaration,i))),a&&Nt(a,e.createDiagnosticForNode(r,e.Diagnostics._0_is_declared_here,i))}}function ur(n,t,r){return!!t&&!!e.findAncestor(n,function(n){return n===r||e.isFunctionLike(n)?"quit":n===t})}function dr(e){switch(e.kind){case 248:return e;case 250:return e.parent;case 251:return e.parent.parent;case 253:return e.parent.parent.parent;default:return}}function pr(n){return e.find(n.declarations,e.isAliasSymbolDeclaration)}function mr(n,t){return 259===n.moduleReference.kind?Hr(Vr(n,e.getExternalModuleImportEqualsDeclarationExpression(n))):Mr(n.moduleReference,t)}function fr(e,n,t){var r=e.exports.get("export=");return r?sc(Xi(r),n):kr(e.exports.get(n),t)}function _r(n){return e.isExportAssignment(n)&&!n.isExportEquals||e.hasModifier(n,512)||e.isExportSpecifier(n)}function gr(n,t,r){if(!I)return!1;if(!n||n.isDeclarationFile){var a=fr(t,"default",!0);return(!a||!e.some(a.declarations,_r))&&!fr(t,e.escapeLeadingUnderscores("__esModule"),r)}return e.isSourceFileJS(n)?!n.externalModuleIndicator&&!fr(t,e.escapeLeadingUnderscores("__esModule"),r):zr(t)}function vr(n,t){var r=Vr(n,n.parent.moduleSpecifier);if(r){var a=void 0;a=e.isShorthandAmbientModuleSymbol(r)?r:fr(r,"default",t);var i=e.find(r.declarations,e.isSourceFile),o=gr(i,r,t);if(a||o){if(o)return Hr(r,t)||kr(r,t)}else It(n.name,e.Diagnostics.Module_0_has_no_default_export,Pa(r));return a}}function yr(e,n){var t=e.parent.parent.moduleSpecifier;return qr(Vr(e,t),t,n)}function hr(n,t){if(n===ne&&t===ne)return ne;if(67899368&n.flags)return n;var r=Ot(n.flags|t.flags,n.escapedName);return r.declarations=e.deduplicate(e.concatenate(n.declarations,t.declarations),e.equateValues),r.parent=n.parent||t.parent,n.valueDeclaration&&(r.valueDeclaration=n.valueDeclaration),t.members&&(r.members=t.members),n.exports&&(r.exports=n.exports),r}function br(e,n,t){if(1536&e.flags)return kr(Zr(e).get(n),t)}function Er(e,n){if(3&e.flags){var t=e.valueDeclaration.type;if(t)return kr(sc(Rd(t),n))}}function Tr(n,t,r){void 0===r&&(r=!1);var a=Vr(n,n.moduleSpecifier),i=qr(a,n.moduleSpecifier,r);if(i){var o=t.propertyName||t.name;if(o.escapedText){if(e.isShorthandAmbientModuleSymbol(a))return a;var s=void 0;s=a&&a.exports&&a.exports.get("export=")?sc(Xi(i),o.escapedText):Er(i,o.escapedText),s=kr(s,r);var c=br(i,o.escapedText,r);!c&&I&&"default"===o.escapedText&&(c=Hr(a,r)||kr(a,r));var l=c&&s&&c!==s?hr(s,c):c||s;if(!l){var u=Or(a,n),d=e.declarationNameToString(o),p=zy(o,i);if(void 0!==p){var m=Pa(p),f=It(o,e.Diagnostics.Module_0_has_no_exported_member_1_Did_you_mean_2,u,d,m);p.valueDeclaration&&Nt(f,e.createDiagnosticForNode(p.valueDeclaration,e.Diagnostics._0_is_declared_here,m))}else It(o,e.Diagnostics.Module_0_has_no_exported_member_1,u,d)}return l}}}function Sr(e,n){return Tr(e.parent.parent.parent,e,n)}function xr(e,n){return Hr(e.parent.symbol,n)}function Cr(e,n,t){return e.parent.parent.moduleSpecifier?Tr(e.parent.parent,e,t):Pr(e.propertyName||e.name,n,!1,t)}function Ar(n,t){var r=e.isExportAssignment(n)?n.expression:n.right;if(e.isClassExpression(r))return qE(r).symbol;var a=Pr(r,68009983,!0,t);return a||(qE(r),zt(r).resolvedSymbol)}function Dr(n,t){switch(void 0===t&&(t=!1),n.kind){case 248:return mr(n,t);case 250:return vr(n,t);case 251:return yr(n,t);case 253:return Sr(n,t);case 257:return Cr(n,68009983,t);case 254:case 204:return Ar(n,t);case 247:return xr(n,t);default:return e.Debug.fail()}}function Lr(e,n){return void 0===n&&(n=68009983),!!e&&(2097152===(e.flags&(2097152|n))||!!(2097152&e.flags&&67108864&e.flags))}function kr(e,n){return!n&&Lr(e)?Nr(e):e}function Nr(n){e.Debug.assert(0!==(2097152&n.flags),"Should only get Alias here.");var t=qt(n);if(t.target)t.target===te&&(t.target=ne);else{t.target=te;var r=pr(n);if(!r)return e.Debug.fail();var a=Dr(r);t.target===te?t.target=a||ne:It(r,e.Diagnostics.Circular_definition_of_import_alias_0,Pa(n))}return t.target}function Ir(e){var n=ra(e),t=Nr(n);if(t){var r=t===ne||67220415&t.flags&&!HC(t);r&&wr(n)}}function wr(n){var t=qt(n);if(!t.referenced){t.referenced=!0;var r=pr(n);if(!r)return e.Debug.fail();254===r.kind?RE(r.expression):257===r.kind?RE(r.propertyName||r.name):e.isInternalModuleImportEqualsDeclaration(r)&&RE(r.moduleReference)}}function Mr(n,t){return 72===n.kind&&e.isRightSideOfQualifiedNameOrPropertyAccess(n)&&(n=n.parent),72===n.kind||148===n.parent.kind?Pr(n,1920,!1,t):(e.Debug.assert(248===n.parent.kind),Pr(n,68009983,!1,t))}function Or(e,n){return e.parent?Or(e.parent,n)+"."+Pa(e):Pa(e,n,void 0,20)}function Pr(n,t,r,a,i){if(!e.nodeIsMissing(n)){var o,s=1920|(e.isInJSFile(n)?67220415&t:0);if(72===n.kind){var c=t===s?e.Diagnostics.Cannot_find_namespace_0:$f(wx(n).escapedText),l=e.isInJSFile(n)?Rr(n,t):void 0;if(o=Zt(i||n,n.escapedText,t,r||l?void 0:c,n,!0),!o)return l}else{if(148!==n.kind&&189!==n.kind)throw e.Debug.assertNever(n,"Unknown entity name kind.");var u=148===n.kind?n.left:n.expression,d=148===n.kind?n.right:n.name,p=Pr(u,s,r,!1,i);if(!p||e.nodeIsMissing(d))return;if(p===ne)return p;if(e.isInJSFile(n)&&p.valueDeclaration&&e.isVariableDeclaration(p.valueDeclaration)&&p.valueDeclaration.initializer&&_b(p.valueDeclaration.initializer)){var m=p.valueDeclaration.initializer.arguments[0],f=Vr(m,m);if(f){var _=Hr(f);_&&(p=_)}}if(o=Xt(Zr(p),d.escapedText,t),!o)return void(r||It(d,e.Diagnostics.Namespace_0_has_no_exported_member_1,Or(p),e.declarationNameToString(d)))}return e.Debug.assert(0===(1&e.getCheckFlags(o)),"Should never get an instantiated symbol here."),o.flags&t||a?o:Nr(o)}}function Rr(e,n){if(Sl(e.parent)){var t=Fr(e.parent);if(t)return Zt(t,e.escapedText,n,void 0,e,!0)}}function Fr(n){var t=e.findAncestor(n,function(n){return e.isJSDocNode(n)||2097152&n.flags?e.isJSDocTypeAlias(n):"quit"});if(!t){var r=e.getJSDocHost(n);if(e.isExpressionStatement(r)&&e.isBinaryExpression(r.expression)&&3===e.getAssignmentDeclarationKind(r.expression)){var a=ra(r.expression.left);if(a)return Gr(a)}if((e.isObjectLiteralMethod(r)||e.isPropertyAssignment(r))&&e.isBinaryExpression(r.parent.parent)&&6===e.getAssignmentDeclarationKind(r.parent.parent)){a=ra(r.parent.parent.left);if(a)return Gr(a)}var i=e.getHostSignatureFromJSDocHost(r);if(i){a=ra(i);return a&&a.valueDeclaration}}}function Gr(n){var t=n.parent.valueDeclaration;if(t){var r=e.isAssignmentDeclaration(t)?e.getAssignedExpandoInitializer(t):e.hasOnlyExpressionInitializer(t)?e.getDeclaredExpandoInitializer(t):void 0;return r||t}}function Vr(n,t,r){return Br(n,t,r?void 0:e.Diagnostics.Cannot_find_module_0)}function Br(n,t,r,a){return void 0===a&&(a=!1),e.isStringLiteralLike(t)?Kr(n,t.text,r,t,a):void 0}function Kr(n,t,a,i,o){if(void 0===o&&(o=!1),void 0!==t){if(e.startsWith(t,"@types/")){var s=e.Diagnostics.Cannot_import_type_declaration_files_Consider_importing_0_instead_of_1,c=e.removePrefix(t,"@types/");It(i,s,c,t)}var l=yc(t,!0);if(l)return l;var u=e.getSourceFileOfNode(n),d=e.getResolvedModule(u,t),p=d&&e.getResolutionDiagnostic(L,d),m=d&&!p&&r.getSourceFile(d.resolvedFileName);if(m)return m.symbol?(d.isExternalLibraryImport&&!e.resolutionExtensionIsTSOrJson(d.extension)&&jr(!1,i,d,t),ta(m.symbol)):void(a&&It(i,e.Diagnostics.File_0_is_not_a_module,m.fileName));if(je){var f=e.findBestPatternMatch(je,function(e){return e.pattern},t);if(f)return ta(f.symbol)}if(d&&!e.resolutionExtensionIsTSOrJson(d.extension)&&void 0===p||p===e.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type)if(o){s=e.Diagnostics.Invalid_module_name_in_augmentation_Module_0_resolves_to_an_untyped_module_at_1_which_cannot_be_augmented;It(i,s,t,d.resolvedFileName)}else jr(R&&!!a,i,d,t);else if(a){if(e.pathIsRelative(t)){var _=e.getSourceFileOfNode(n),g=_.redirectedReferences;if(g)for(var v=e.getNormalizedAbsolutePath(t,e.getDirectoryPath(_.fileName)),y=0,h=[".ts",".tsx"];y<h.length;y++){var b=h[y],E=v+b;if(g.indexOf(E)>=0)return void It(i,e.Diagnostics.Output_file_0_has_not_been_built_from_source_file_1,t,E)}}if(p)It(i,p,t,d.resolvedFileName);else{var T=e.tryExtractTSExtension(t);if(T){s=e.Diagnostics.An_import_path_cannot_end_with_a_0_extension_Consider_importing_1_instead;It(i,s,T,e.removeExtension(t,T))}else!L.resolveJsonModule&&e.fileExtensionIs(t,".json")&&e.getEmitModuleResolutionKind(L)===e.ModuleResolutionKind.NodeJs&&e.hasJsonModuleEmitEnabled(L)?It(i,e.Diagnostics.Cannot_find_module_0_Consider_using_resolveJsonModule_to_import_module_with_json_extension,t):It(i,a,t)}}}}function jr(n,t,r,a){var i=r.packageId,o=r.resolvedFileName,s=!e.isExternalModuleNameRelative(a)&&i?Ur(i.name)?e.chainDiagnosticMessages(void 0,e.Diagnostics.If_the_0_package_actually_exposes_this_module_consider_sending_a_pull_request_to_amend_https_Colon_Slash_Slashgithub_com_SlashDefinitelyTyped_SlashDefinitelyTyped_Slashtree_Slashmaster_Slashtypes_Slash_1,i.name,e.mangleScopedPackageName(i.name)):e.chainDiagnosticMessages(void 0,e.Diagnostics.Try_npm_install_types_Slash_1_if_it_exists_or_add_a_new_declaration_d_ts_file_containing_declare_module_0,a,e.mangleScopedPackageName(i.name)):void 0;Mt(n,t,e.chainDiagnosticMessages(s,e.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type,a,o))}function Ur(n){return v().has(e.getTypesPackageName(n))}function Hr(e,n){if(e){var t=kr(e.exports.get("export="),n),r=Wr(t,e);return ta(r)||e}}function Wr(n,t){if(!n||n===ne||n===t||1===t.exports.size||2097152&n.flags)return n;var r=Gt(n);return void 0===r.exports&&(r.flags=512|r.flags,r.exports=e.createSymbolTable()),t.exports.forEach(function(e,n){"export="!==n&&r.exports.set(n,r.exports.has(n)?Vt(r.exports.get(n),e):e)}),r}function qr(n,t,r){var a=Hr(n,r);if(!r&&a){if(!(1539&a.flags)&&!e.getDeclarationOfKind(a,279)){var i=N>=e.ModuleKind.ES2015?"allowSyntheticDefaultImports":"esModuleInterop";return It(t,e.Diagnostics.This_module_can_only_be_referenced_with_ECMAScript_imports_Slashexports_by_turning_on_the_0_flag_and_referencing_its_default_export,i),a}if(L.esModuleInterop){var o=t.parent;if(e.isImportDeclaration(o)&&e.getNamespaceDeclarationNode(o)||e.isImportCall(o)){var s=Xi(a),c=cc(s,0);if(c&&c.length||(c=cc(s,1)),c&&c.length){var l=fb(s,a,n),u=Ot(a.flags,a.escapedName);u.declarations=a.declarations?a.declarations.slice():[],u.parent=a.parent,u.target=a,u.originatingImport=o,a.valueDeclaration&&(u.valueDeclaration=a.valueDeclaration),a.constEnumOnlyModule&&(u.constEnumOnlyModule=!0),a.members&&(u.members=e.cloneMap(a.members)),a.exports&&(u.exports=e.cloneMap(a.exports));var d=ws(l);return u.type=ba(u,d.members,e.emptyArray,e.emptyArray,d.stringIndexInfo,d.numberIndexInfo),u}}}}return a}function zr(e){return void 0!==e.exports.get("export=")}function Jr(e){return gc($r(e))}function Xr(n){var t=Jr(n),r=Hr(n);return r!==n&&e.addRange(t,Rs(Xi(r))),t}function Yr(e,n){var t=$r(n);if(t)return t.get(e)}function Qr(e,n){var t=Yr(e,n);if(t)return t;var r=Hr(n);if(r!==n){var a=Xi(r);return 131068&a.flags?void 0:sc(a,e)}}function Zr(e){return 32&e.flags?qo(e,"resolvedExports"):1536&e.flags?$r(e):e.exports||A}function $r(e){var n=qt(e);return n.resolvedExports||(n.resolvedExports=na(e))}function ea(n,t,r,a){t&&t.forEach(function(t,i){if("default"!==i){var o=n.get(i);if(o){if(r&&a&&o&&kr(o)!==kr(t)){var s=r.get(i);s.exportsWithDuplicate?s.exportsWithDuplicate.push(a):s.exportsWithDuplicate=[a]}}else n.set(i,t),r&&a&&r.set(i,{specifierText:e.getTextOfNode(a.moduleSpecifier)})}})}function na(n){var t=[];return n=Hr(n),r(n)||A;function r(n){if(n&&n.exports&&e.pushIfUnique(t,n)){var a=e.cloneMap(n.exports),i=n.exports.get("__export");if(i){for(var o=e.createSymbolTable(),s=e.createMap(),c=0,l=i.declarations;c<l.length;c++){var u=l[c],d=Vr(u,u.moduleSpecifier),p=r(d);ea(o,p,s,u)}s.forEach(function(n,t){var r=n.exportsWithDuplicate;if("export="!==t&&r&&r.length&&!a.has(t))for(var i=0,o=r;i<o.length;i++){var c=o[i];it.add(e.createDiagnosticForNode(c,e.Diagnostics.Module_0_has_already_exported_a_member_named_1_Consider_explicitly_re_exporting_to_resolve_the_ambiguity,s.get(t).specifierText,e.unescapeLeadingUnderscores(t)))}}),ea(a,o)}return a}}}function ta(e){var n;return e&&e.mergeId&&(n=zn[e.mergeId])?n:e}function ra(e){return ta(e.symbol&&Jo(e.symbol))}function aa(e){return ta(e.parent&&Jo(e.parent))}function ia(n,t){var a,i=e.getSourceFileOfNode(t),o=""+l(i),s=qt(n);if(s.extendedContainersByFile&&(a=s.extendedContainersByFile.get(o)))return a;if(i&&i.imports){for(var c=0,u=i.imports;c<u.length;c++){var d=u[c];if(!e.nodeIsSynthesized(d)){var p=Vr(t,d,!0);if(p){var m=sa(p,n);m&&(a=e.append(a,p))}}}if(e.length(a))return(s.extendedContainersByFile||(s.extendedContainersByFile=e.createMap())).set(o,a),a}if(s.extendedContainers)return s.extendedContainers;for(var f=r.getSourceFiles(),_=0,g=f;_<g.length;_++){var v=g[_];if(e.isExternalModule(v)){var y=ra(v);m=sa(y,n);m&&(a=e.append(a,y))}}return s.extendedContainers=a||e.emptyArray}function oa(n,t){var r=aa(n);if(r){var a=e.mapDefined(r.declarations,c),i=t&&ia(n,t);if(t&&Sa(r,t,1920,!1))return e.concatenate(e.concatenate([r],a),i);var o=e.append(a,r);return e.concatenate(o,i)}var s=e.mapDefined(n.declarations,function(n){return!e.isAmbientModule(n)&&n.parent&&wa(n.parent)?ra(n.parent):void 0});if(e.length(s))return e.mapDefined(s,function(e){return sa(e,n)?e:void 0});function c(e){var n=Na(e),t=n&&n.exports&&n.exports.get("export=");return kr(t)===kr(r)?n:void 0}}function sa(n,t){if(n===aa(t))return t;var r=Zr(n),a=r.get(t.escapedName);return a&&i(a)?a:e.forEachEntry(r,function(e){if(i(e))return e});function i(e){if(e===t||kr(e)===t||kr(e)===kr(t))return e}}function ca(e){return ta(e&&0!==(1048576&e.flags)?e.exportSymbol:e)}function la(e){return!!(67220415&e.flags||2097152&e.flags&&67220415&Nr(e).flags)}function ua(n){for(var t=n.members,r=0,a=t;r<a.length;r++){var i=a[r];if(157===i.kind&&e.nodeIsPresent(i.body))return i}}function da(e){var n=new h(q,e);return E++,n.id=E,n}function pa(e,n){var t=da(e);return t.intrinsicName=n,t}function ma(e){var n=Au(e);return n.flags|=16,n.intrinsicName="boolean",n}function fa(e,n){var t=da(524288);return t.objectFlags=e,t.symbol=n,t.members=void 0,t.properties=void 0,t.callSignatures=void 0,t.constructSignatures=void 0,t.stringIndexInfo=void 0,t.numberIndexInfo=void 0,t}function _a(){return Au(e.arrayFrom(gt.keys(),Nd))}function ga(e){var n=da(262144);return e&&(n.symbol=e),n}function va(e){return 95===e.charCodeAt(0)&&95===e.charCodeAt(1)&&95!==e.charCodeAt(2)&&64!==e.charCodeAt(2)}function ya(n){var t;return n.forEach(function(e,n){!va(n)&&la(e)&&(t||(t=[])).push(e)}),t||e.emptyArray}function ha(n,t,r,a,i,o){return n.members=t,n.properties=t===A?e.emptyArray:ya(t),n.callSignatures=r,n.constructSignatures=a,n.stringIndexInfo=i,n.numberIndexInfo=o,n}function ba(e,n,t,r,a,i){return ha(fa(16,e),n,t,r,a,i)}function Ea(n,t){for(var r,a=n;a;a=a.parent){if(a.locals&&!Jt(a)&&(r=t(a.locals)))return r;switch(a.kind){case 279:if(!e.isExternalOrCommonJsModule(a))break;case 244:if(r=t(ra(a).exports))return r;break}}return t(wn)}function Ta(e){return 67220415===e?67220415:1920}function Sa(n,t,r,a,i){if(void 0===i&&(i=e.createMap()),n&&!Ca(n)){var o=""+u(n),s=i.get(o);return s||i.set(o,s=[]),Ea(t,c)}function c(n,t){if(e.pushIfUnique(s,n)){var r=p(n,t);return s.pop(),r}}function l(e,n){return!xa(e,t,n)||!!Sa(e.parent,t,Ta(n),a,i)}function d(t,a,i){return n===(a||t)&&!e.some(t.declarations,wa)&&(i||l(t,r))}function p(i,o){return d(i.get(n.escapedName),void 0,o)?[n]:e.forEachEntry(i,function(i){if(2097152&i.flags&&"export="!==i.escapedName&&"default"!==i.escapedName&&!(e.isUMDExportSymbol(i)&&t&&e.isExternalModule(e.getSourceFileOfNode(t)))&&(!a||e.some(i.declarations,e.isExternalModuleImportEqualsDeclaration))&&(o||!e.getDeclarationOfKind(i,257))){var s=Nr(i);if(d(i,s,o))return[i];var u=Zr(s),p=u&&c(u,!0);if(p&&l(i,Ta(r)))return[i].concat(p)}if(i.escapedName===n.escapedName&&i.exportSymbol&&d(ta(i.exportSymbol),void 0,o))return[n]})}}function xa(n,t,r){var a=!1;return Ea(t,function(t){var i=ta(t.get(n.escapedName));return!!i&&(i===n||(i=2097152&i.flags&&!e.getDeclarationOfKind(i,257)?Nr(i):i,!!(i.flags&r)&&(a=!0,!0)))}),a}function Ca(e){if(e.declarations&&e.declarations.length){for(var n=0,t=e.declarations;n<t.length;n++){var r=t[n];switch(r.kind){case 154:case 156:case 158:case 159:continue;default:return!1}}return!0}return!1}function Aa(e,n){var t=ka(e,n,67897832,!1);return 0===t.accessibility}function Da(e,n){var t=ka(e,n,67220415,!1);return 0===t.accessibility}function La(n,t,r,a,i){if(e.length(n)){for(var o,s=0,c=n;s<c.length;s++){var l=c[s],u=Sa(l,t,a,!1);if(u){o=l;var d=Ma(u[0],i);if(d)return d}else if(e.some(l.declarations,wa))return{accessibility:0};var p=oa(l,t),m=e.first(l.declarations);!e.length(p)&&67220415&a&&m&&e.isObjectLiteralExpression(m)&&m.parent&&e.isVariableDeclaration(m.parent)&&m===m.parent.initializer&&(p=[ra(m.parent)]);var f=La(p,t,r,r===l?Ta(a):a,i);if(f)return f}return o?{accessibility:1,errorSymbolName:Pa(r,t,a),errorModuleName:o!==r?Pa(o,t,1920):void 0}:void 0}}function ka(n,t,r,a){if(n&&t){var i=La([n],t,n,r,a);if(i)return i;var o=e.forEach(n.declarations,Na);if(o){var s=Na(t);if(o!==s)return{accessibility:2,errorSymbolName:Pa(n,t,r),errorModuleName:Pa(o)}}return{accessibility:1,errorSymbolName:Pa(n,t,r)}}return{accessibility:0}}function Na(n){var t=e.findAncestor(n,Ia);return t&&ra(t)}function Ia(n){return e.isAmbientModule(n)||279===n.kind&&e.isExternalOrCommonJsModule(n)}function wa(n){return e.isModuleWithStringLiteralName(n)||279===n.kind&&e.isExternalOrCommonJsModule(n)}function Ma(n,t){var r;if(e.every(n.declarations,a))return{accessibility:0,aliasesToMakeVisible:r};function a(n){if(!Ja(n)){var t=dr(n);return t&&!e.hasModifier(t,1)&&Ja(t.parent)?i(n,t):e.isVariableDeclaration(n)&&e.isVariableStatement(n.parent.parent)&&!e.hasModifier(n.parent.parent,1)&&Ja(n.parent.parent.parent)?i(n,n.parent.parent):!(!e.isLateVisibilityPaintedStatement(n)||e.hasModifier(n,1)||!Ja(n.parent))&&i(n,n)}return!0}function i(n,a){return t&&(zt(n).isVisible=!0,r=e.appendIfUnique(r,a)),!0}}function Oa(n,t){var r;r=167===n.parent.kind||e.isExpressionWithTypeArgumentsInClassExtendsClause(n.parent)||149===n.parent.kind?68268991:148===n.kind||189===n.kind||248===n.parent.kind?1920:67897832;var a=wx(n),i=Zt(t,a.escapedText,r,void 0,void 0,!1);return i&&Ma(i,!0)||{accessibility:1,errorSymbolName:e.getTextOfNode(a),errorNode:a}}function Pa(n,t,r,a,i){void 0===a&&(a=4);var o=70221824;2&a&&(o|=128),1&a&&(o|=512),8&a&&(o|=16384),16&a&&(o|=134217728);var s=4&a?K.symbolToExpression:K.symbolToEntityName;return i?c(i).getText():e.usingSingleLineStringWriter(c);function c(a){var i=s(n,r,t,o),c=e.createPrinter({removeComments:!0}),l=t&&e.getSourceFileOfNode(t);return c.writeNode(4,i,l,a),a}}function Ra(n,t,r,a,i){return void 0===r&&(r=0),i?o(i).getText():e.usingSingleLineStringWriter(o);function o(i){var o;o=262144&r?1===a?166:165:1===a?161:160;var s=K.signatureToSignatureDeclaration(n,o,t,70222336|Ga(r)),c=e.createPrinter({removeComments:!0,omitTrailingSemicolon:!0}),l=t&&e.getSourceFileOfNode(t);return c.writeNode(4,s,l,e.getTrailingSemicolonOmittingWriter(i)),i}}function Fa(n,t,r,a){void 0===r&&(r=1064960),void 0===a&&(a=e.createTextWriter(""));var i=L.noErrorTruncation||1&r,o=K.typeToTypeNode(n,t,70221824|Ga(r)|(i?1:0),a);if(void 0===o)return e.Debug.fail("should always get typenode");var s={removeComments:!0},c=e.createPrinter(s),l=t&&e.getSourceFileOfNode(t);c.writeNode(4,o,l,a);var u=a.getText(),d=i?void 0:2*e.defaultMaximumTruncationLength;return d&&u&&u.length>=d?u.substr(0,d-"...".length)+"...":u}function Ga(e){return void 0===e&&(e=0),9469291&e}function Va(){return{typeToTypeNode:function(e,t,r,a){return n(t,r,a,function(n){return i(e,n)})},indexInfoToIndexSignatureDeclaration:function(e,t,r,a,i){return n(r,a,i,function(n){return d(e,t,n)})},signatureToSignatureDeclaration:function(e,t,r,a,i){return n(r,a,i,function(n){return m(e,t,n)})},symbolToEntityName:function(e,t,r,a,i){return n(r,a,i,function(n){return C(e,n,t,!1)})},symbolToExpression:function(e,t,r,a,i){return n(r,a,i,function(n){return A(e,n,t)})},symbolToTypeParameterDeclarations:function(e,t,r,a){return n(t,r,a,function(n){return b(e,n)})},symbolToParameterDeclaration:function(e,t,r,a){return n(t,r,a,function(n){return v(e,n)})},typeParameterToDeclaration:function(e,t,r,a){return n(t,r,a,function(n){return g(e,n)})}};function n(n,t,a,i){e.Debug.assert(void 0===n||0===(8&n.flags));var o={enclosingDeclaration:n,flags:t||0,tracker:a&&a.trackSymbol?a:{trackSymbol:e.noop,moduleResolverHost:134217728&t?{getCommonSourceDirectory:r.getCommonSourceDirectory?function(){return r.getCommonSourceDirectory()}:function(){return""},getSourceFiles:function(){return r.getSourceFiles()},getCurrentDirectory:r.getCurrentDirectory&&function(){return r.getCurrentDirectory()}}:void 0},encounteredError:!1,visitedTypes:void 0,symbolDepth:void 0,inferTypeParameters:void 0,approximateLength:0},s=i(o);return o.encounteredError?void 0:s}function a(n){return n.truncating?n.truncating:n.truncating=!(1&n.flags)&&n.approximateLength>e.defaultMaximumTruncationLength}function i(n,t){p&&p.throwIfCancellationRequested&&p.throwIfCancellationRequested();var r=8388608&t.flags;if(t.flags&=-8388609,n){if(1&n.flags)return t.approximateLength+=3,e.createKeywordTypeNode(120);if(2&n.flags)return e.createKeywordTypeNode(143);if(4&n.flags)return t.approximateLength+=6,e.createKeywordTypeNode(138);if(8&n.flags)return t.approximateLength+=6,e.createKeywordTypeNode(135);if(64&n.flags)return t.approximateLength+=6,e.createKeywordTypeNode(146);if(16&n.flags)return t.approximateLength+=7,e.createKeywordTypeNode(123);if(1024&n.flags&&!(1048576&n.flags)){var o=aa(n.symbol),g=x(o,t,67897832),v=Do(o)===n?g:F(g,e.createTypeReferenceNode(e.symbolName(n.symbol),void 0));return v}if(1056&n.flags)return x(n.symbol,t,67897832);if(128&n.flags)return t.approximateLength+=n.value.length+2,e.createLiteralTypeNode(e.setEmitFlags(e.createLiteral(n.value),16777216));if(256&n.flags)return t.approximateLength+=(""+n.value).length,e.createLiteralTypeNode(e.createLiteral(n.value));if(2048&n.flags)return t.approximateLength+=e.pseudoBigIntToString(n.value).length+1,e.createLiteralTypeNode(e.createLiteral(n.value));if(512&n.flags)return t.approximateLength+=n.intrinsicName.length,"true"===n.intrinsicName?e.createTrue():e.createFalse();if(8192&n.flags){if(!(1048576&t.flags)){if(Da(n.symbol,t.enclosingDeclaration))return t.approximateLength+=6,x(n.symbol,t,67220415);t.tracker.reportInaccessibleUniqueSymbolError&&t.tracker.reportInaccessibleUniqueSymbolError()}return t.approximateLength+=13,e.createTypeOperatorNode(142,e.createKeywordTypeNode(139))}if(16384&n.flags)return t.approximateLength+=4,e.createKeywordTypeNode(106);if(32768&n.flags)return t.approximateLength+=9,e.createKeywordTypeNode(141);if(65536&n.flags)return t.approximateLength+=4,e.createKeywordTypeNode(96);if(131072&n.flags)return t.approximateLength+=5,e.createKeywordTypeNode(132);if(4096&n.flags)return t.approximateLength+=6,e.createKeywordTypeNode(139);if(67108864&n.flags)return t.approximateLength+=6,e.createKeywordTypeNode(136);if(262144&n.flags&&n.isThisType)return 4194304&t.flags&&(t.encounteredError||32768&t.flags||(t.encounteredError=!0),t.tracker.reportInaccessibleThisError&&t.tracker.reportInaccessibleThisError()),t.approximateLength+=4,e.createThis();var y=e.getObjectFlags(n);if(4&y)return e.Debug.assert(!!(524288&n.flags)),R(n);if(262144&n.flags||3&y){if(262144&n.flags&&e.contains(t.inferTypeParameters,n))return t.approximateLength+=e.symbolName(n.symbol).length+6,e.createInferTypeNode(_(n,t,void 0));if(4&t.flags&&262144&n.flags&&e.length(n.symbol.declarations)&&e.isTypeParameterDeclaration(n.symbol.declarations[0])&&f(n,t)&&!Aa(n.symbol,t.enclosingDeclaration)){var h=n.symbol.declarations[0].name;return t.approximateLength+=e.idText(h).length,e.createTypeReferenceNode(e.getGeneratedNameForNode(h,24),void 0)}return n.symbol?x(n.symbol,t,67897832):e.createTypeReferenceNode(e.createIdentifier("?"),void 0)}if(!r&&n.aliasSymbol&&(16384&t.flags||Aa(n.aliasSymbol,t.enclosingDeclaration))){var b=l(n.aliasTypeArguments,t);return!va(n.aliasSymbol.escapedName)||32&n.aliasSymbol.flags?x(n.aliasSymbol,t,67897832,b):e.createTypeReferenceNode(e.createIdentifier(""),b)}if(!(3145728&n.flags)){if(48&y)return e.Debug.assert(!!(524288&n.flags)),O(n);if(4194304&n.flags){var E=n.type;t.approximateLength+=6;var T=i(E,t);return e.createTypeOperatorNode(T)}if(8388608&n.flags){var S=i(n.objectType,t);T=i(n.indexType,t);return t.approximateLength+=2,e.createIndexedAccessTypeNode(S,T)}if(16777216&n.flags){var C=i(n.checkType,t),A=t.inferTypeParameters;t.inferTypeParameters=n.root.inferTypeParameters;var D=i(n.extendsType,t);t.inferTypeParameters=A;var L=i(ld(n),t),k=i(ud(n),t);return t.approximateLength+=15,e.createConditionalTypeNode(C,D,L,k)}return 33554432&n.flags?i(n.typeVariable,t):e.Debug.fail("Should be unreachable.")}var N=1048576&n.flags?Ka(n.types):n.types;if(1===e.length(N))return i(N[0],t);var I=l(N,t,!0);if(I&&I.length>0){var w=e.createUnionOrIntersectionTypeNode(1048576&n.flags?173:174,I);return w}t.encounteredError||262144&t.flags||(t.encounteredError=!0)}else t.encounteredError=!0;function M(n){e.Debug.assert(!!(524288&n.flags));var r,a=n.declaration.readonlyToken?e.createToken(n.declaration.readonlyToken.kind):void 0,o=n.declaration.questionToken?e.createToken(n.declaration.questionToken.kind):void 0;r=Cs(n)?e.createTypeOperatorNode(i(As(n),t)):i(Ts(n),t);var s=_(Es(n),t,r),c=i(Ss(n),t),l=e.createMappedTypeNode(a,s,o,c);return t.approximateLength+=10,e.setEmitFlags(l,1)}function O(n){var r,a=""+n.id,i=n.symbol;if(i){var o=16&e.getObjectFlags(n)&&n.symbol&&32&n.symbol.flags;if(r=(o?"+":"")+u(i),ib(i.valueDeclaration)){var c=n===ub(i)?67897832:67220415;return x(i,t,c)}if(32&i.flags&&!ji(i)&&!(209===i.valueDeclaration.kind&&2048&t.flags)||896&i.flags||m())return x(i,t,67220415);if(t.visitedTypes&&t.visitedTypes.has(a)){var l=Ua(n);return l?x(l,t,67897832):s(t)}t.visitedTypes||(t.visitedTypes=e.createMap()),t.symbolDepth||(t.symbolDepth=e.createMap());var d=t.symbolDepth.get(r)||0;if(d>10)return s(t);t.symbolDepth.set(r,d+1),t.visitedTypes.set(a,!0);var p=P(n);return t.visitedTypes.delete(a),t.symbolDepth.set(r,d),p}return P(n);function m(){var n=!!(8192&i.flags)&&e.some(i.declarations,function(n){return e.hasModifier(n,32)}),r=!!(16&i.flags)&&(i.parent||e.forEach(i.declarations,function(e){return 279===e.parent.kind||245===e.parent.kind}));if(n||r)return(!!(4096&t.flags)||t.visitedTypes&&t.visitedTypes.has(a))&&(!(8&t.flags)||Da(i,t.enclosingDeclaration))}}function P(n){if(Is(n))return M(n);var r=ws(n);if(!r.properties.length&&!r.stringIndexInfo&&!r.numberIndexInfo){if(!r.callSignatures.length&&!r.constructSignatures.length)return t.approximateLength+=2,e.setEmitFlags(e.createTypeLiteralNode(void 0),1);if(1===r.callSignatures.length&&!r.constructSignatures.length){var a=r.callSignatures[0],i=m(a,165,t);return i}if(1===r.constructSignatures.length&&!r.callSignatures.length){a=r.constructSignatures[0],i=m(a,166,t);return i}}var o=t.flags;t.flags|=4194304;var s=V(r);t.flags=o;var c=e.createTypeLiteralNode(s);return t.approximateLength+=2,e.setEmitFlags(c,1024&t.flags?0:1)}function R(n){var r=n.typeArguments||e.emptyArray;if(n.target===ze){if(2&t.flags){var a=i(r[0],t);return e.createTypeReferenceNode("Array",[a])}var o=i(r[0],t);return e.createArrayTypeNode(o)}if(8&n.target.objectFlags){if(r.length>0){var s=ul(n),c=l(r.slice(0,s),t),u=n.target.hasRestElement;if(c){for(var d=n.target.minLength;d<Math.min(s,c.length);d++)c[d]=u&&d===s-1?e.createRestTypeNode(e.createArrayTypeNode(c[d])):e.createOptionalTypeNode(c[d]);return e.createTupleTypeNode(c)}}return t.encounteredError||524288&t.flags?e.createTupleTypeNode([]):void(t.encounteredError=!0)}if(2048&t.flags&&n.symbol.valueDeclaration&&e.isClassLike(n.symbol.valueDeclaration)&&!Da(n.symbol,t.enclosingDeclaration))return O(n);var p=n.target.outerTypeParameters,m=(d=0,void 0);if(p){var f=p.length;while(d<f){var _=d,g=il(p[d]);do{d++}while(d<f&&il(p[d])===g);if(!e.rangeEquals(p,r,_,d)){var v=l(r.slice(_,d),t),y=t.flags;t.flags|=16;var h=x(g,t,67897832,v);t.flags=y,m=m?F(m,h):h}}}var b=void 0;if(r.length>0){var E=(n.target.typeParameters||e.emptyArray).length;b=l(r.slice(d,E),t)}var T=t.flags;t.flags|=16;var S=x(n.symbol,t,67897832,b);return t.flags=T,m?F(m,S):S}function F(n,t){if(e.isImportTypeNode(n)){var r=n.typeArguments;n.qualifier&&((e.isIdentifier(n.qualifier)?n.qualifier:n.qualifier.right).typeArguments=r),n.typeArguments=t.typeArguments;for(var a=G(t),i=0,o=a;i<o.length;i++){var s=o[i];n.qualifier=n.qualifier?e.createQualifiedName(n.qualifier,s):s}return n}r=n.typeArguments;(e.isIdentifier(n.typeName)?n.typeName:n.typeName.right).typeArguments=r,n.typeArguments=t.typeArguments;a=G(t);for(var c=0,l=a;c<l.length;c++){s=l[c];n.typeName=e.createQualifiedName(n.typeName,s)}return n}function G(n){var t=n.typeName,r=[];while(!e.isIdentifier(t))r.unshift(t.right),t=t.left;return r.unshift(t),r}function V(n){if(a(t))return[e.createPropertySignature(void 0,"...",void 0,void 0,void 0)];for(var r=[],i=0,o=n.callSignatures;i<o.length;i++){var l=o[i];r.push(m(l,160,t))}for(var u=0,p=n.constructSignatures;u<p.length;u++){l=p[u];r.push(m(l,161,t))}if(n.stringIndexInfo){var f=void 0;2048&n.objectFlags?(f=d(el(re,n.stringIndexInfo.isReadonly,n.stringIndexInfo.declaration),0,t),f.type=s(t)):f=d(n.stringIndexInfo,0,t),r.push(f)}n.numberIndexInfo&&r.push(d(n.numberIndexInfo,1,t));var _=n.properties;if(!_)return r;for(var g=0,v=0,y=_;v<y.length;v++){var h=y[v];if(g++,2048&t.flags){if(4194304&h.flags)continue;24&e.getDeclarationModifierFlagsFromSymbol(h)&&t.tracker.reportPrivateInBaseOfClassExpression&&t.tracker.reportPrivateInBaseOfClassExpression(e.unescapeLeadingUnderscores(h.escapedName))}if(a(t)&&g+2<_.length-1){r.push(e.createPropertySignature(void 0,"... "+(_.length-g)+" more ...",void 0,void 0,void 0)),c(_[_.length-1],t,r);break}c(h,t,r)}return r.length?r:void 0}}function s(n){return n.approximateLength+=3,1&n.flags?e.createKeywordTypeNode(120):e.createTypeReferenceNode(e.createIdentifier("..."),void 0)}function c(n,t,r){var a=!!(4096&e.getCheckFlags(n)),o=a&&33554432&t.flags?re:Xi(n),c=t.enclosingDeclaration;if(t.enclosingDeclaration=void 0,t.tracker.trackSymbol&&2048&e.getCheckFlags(n)){var l=e.first(n.declarations);Bo(l)&&y(l.name,c,t)}var u=C(n,t,67220415,!0);t.approximateLength+=e.symbolName(n).length+1,t.enclosingDeclaration=c;var d=16777216&n.flags?e.createToken(56):void 0;if(8208&n.flags&&!Ms(o).length)for(var p=lc(o,0),f=0,_=p;f<_.length;f++){var g=_[f],v=m(g,155,t);v.name=u,v.questionToken=d,n.valueDeclaration&&e.setCommentRange(v,n.valueDeclaration),r.push(v)}else{var h=t.flags;t.flags|=a?33554432:0;var b=void 0;b=a&&33554432&h?s(t):o?i(o,t):e.createKeywordTypeNode(120),t.flags=h;var E=tE(n)?[e.createToken(133)]:void 0;E&&(t.approximateLength+=9);var T=e.createPropertySignature(E,u,d,b,void 0);n.valueDeclaration&&e.setCommentRange(T,n.valueDeclaration),r.push(T)}}function l(n,t,r){if(e.some(n)){if(a(t)){if(!r)return[e.createTypeReferenceNode("...",void 0)];if(n.length>2)return[i(n[0],t),e.createTypeReferenceNode("... "+(n.length-2)+" more ...",void 0),i(n[n.length-1],t)]}for(var o=[],s=0,c=0,l=n;c<l.length;c++){var u=l[c];if(s++,a(t)&&s+2<n.length-1){o.push(e.createTypeReferenceNode("... "+(n.length-s)+" more ...",void 0));var d=i(n[n.length-1],t);d&&o.push(d);break}t.approximateLength+=2;var p=i(u,t);p&&o.push(p)}return o}}function d(n,t,r){var a=e.getNameFromIndexInfo(n)||"x",o=e.createKeywordTypeNode(0===t?138:135),s=e.createParameter(void 0,void 0,void 0,a,void 0,o,void 0),c=i(n.type||re,r);return n.type||2097152&r.flags||(r.encounteredError=!0),r.approximateLength+=a.length+4,e.createIndexSignature(void 0,n.isReadonly?[e.createToken(133)]:void 0,[s],c)}function m(n,t,r){var a,o;32&r.flags&&n.target&&n.mapper&&n.target.typeParameters?o=n.target.typeParameters.map(function(e){return i(mp(e,n.mapper),r)}):a=n.typeParameters&&n.typeParameters.map(function(e){return g(e,r)});var s,c=ns(n).map(function(e){return v(e,r,157===t)});if(n.thisParameter){var l=v(n.thisParameter,r);c.unshift(l)}var u=Oc(n);if(u){var d=1===u.kind?e.setEmitFlags(e.createIdentifier(u.parameterName),16777216):e.createThisTypeNode(),p=i(u.type,r);s=e.createTypePredicateNode(d,p)}else{var m=Fc(n);s=m&&i(m,r)}return 256&r.flags?s&&120===s.kind&&(s=void 0):s||(s=e.createKeywordTypeNode(120)),r.approximateLength+=3,e.createSignatureDeclaration(t,a,c,s,o)}function f(e,n){return!!Zt(n.enclosingDeclaration,e.symbol.escapedName,67897832,void 0,e.symbol.escapedName,!1)}function _(n,t,r){var a=t.flags;t.flags&=-513;var o=4&t.flags&&n.symbol.declarations[0]&&e.isTypeParameterDeclaration(n.symbol.declarations[0])&&f(n,t),s=o?e.getGeneratedNameForNode(n.symbol.declarations[0].name,24):C(n.symbol,t,67897832,!0),c=Zs(n),l=c&&i(c,t);return t.flags=a,e.createTypeParameterDeclaration(s,r,l)}function g(e,n,t){void 0===t&&(t=Bs(e));var r=t&&i(t,n);return _(e,n,r)}function v(n,t,r){var a=e.getDeclarationOfKind(n,151);a||Pt(n)||(a=e.getDeclarationOfKind(n,299));var o=Xi(n);a&&zC(a)&&(o=sf(o));var s=i(o,t),c=!(8192&t.flags)&&r&&a&&a.modifiers?a.modifiers.map(e.getSynthesizedClone):void 0,l=a&&e.isRestParameter(a)||16384&e.getCheckFlags(n),u=l?e.createToken(25):void 0,d=a&&a.name?72===a.name.kind?e.setEmitFlags(e.getSynthesizedClone(a.name),16777216):148===a.name.kind?e.setEmitFlags(e.getSynthesizedClone(a.name.right),16777216):_(a.name):e.symbolName(n),p=a&&hc(a)||8192&e.getCheckFlags(n),m=p?e.createToken(56):void 0,f=e.createParameter(void 0,c,u,d,m,s,void 0);return t.approximateLength+=e.symbolName(n).length+3,f;function _(n){return r(n);function r(n){t.tracker.trackSymbol&&e.isComputedPropertyName(n)&&Go(n)&&y(n,t.enclosingDeclaration,t);var a=e.visitEachChild(n,r,e.nullTransformationContext,void 0,r),i=e.nodeIsSynthesized(a)?a:e.getSynthesizedClone(a);return 186===i.kind&&(i.initializer=void 0),e.setEmitFlags(i,16777217)}}}function y(e,n,t){if(t.tracker.trackSymbol){var r=wx(e.expression),a=Zt(r,r.escapedText,68268991,void 0,void 0,!0);a&&t.tracker.trackSymbol(a,n,67220415)}}function h(n,t,r,a){var i;t.tracker.trackSymbol(n,t.enclosingDeclaration,r);var o=262144&n.flags;return o||!(t.enclosingDeclaration||64&t.flags)||134217728&t.flags?i=[n]:(i=e.Debug.assertDefined(s(n,r,!0)),e.Debug.assert(i&&i.length>0)),i;function s(n,r,i){var o,c=Sa(n,t.enclosingDeclaration,r,!!(128&t.flags));if(!c||xa(c[0],t.enclosingDeclaration,1===c.length?r:Ta(r))){var l=oa(c?c[0]:n,t.enclosingDeclaration);if(e.length(l)){o=l.map(function(n){return e.some(n.declarations,wa)?S(n,t):void 0});var u=l.map(function(e,n){return n});u.sort(g);for(var d=u.map(function(e){return l[e]}),p=0,m=d;p<m.length;p++){var f=m[p],_=s(f,Ta(r),!1);if(_){c=_.concat(c||[sa(f,n)||n]);break}}}}if(c)return c;if(i||!(6144&n.flags)){if(!i&&!a&&e.forEach(n.declarations,wa))return;return[n]}function g(n,t){var r=o[n],a=o[t];if(r&&a){var i=e.pathIsRelative(a);return e.pathIsRelative(r)===i?e.moduleSpecifiers.countPathComponents(r)-e.moduleSpecifiers.countPathComponents(a):i?-1:1}return 0}}}function b(n,t){var r,a=fx(n);return 524384&a.flags&&(r=e.createNodeArray(e.map(to(n),function(e){return g(e,t)}))),r}function E(n,t,r){e.Debug.assert(n&&0<=t&&t<n.length);var a,i=n[t];if(512&r.flags&&t<n.length-1){var o=i,s=n[t+1];if(1&e.getCheckFlags(s)){var c=ro(2097152&o.flags?Nr(o):o);a=l(e.map(c,s.mapper),r)}else a=b(i,r)}return a}function T(n){return e.isIndexedAccessTypeNode(n.objectType)?T(n.objectType):n}function S(n,a){var i=e.getDeclarationOfKind(n,279);if(i&&void 0!==i.moduleName)return i.moduleName;if(!i){if(a.tracker.trackReferencedAmbientModule){var s=e.filter(n.declarations,e.isAmbientModule);if(e.length(s))for(var c=0,l=s;c<l.length;c++){var u=l[c];a.tracker.trackReferencedAmbientModule(u,n)}}if(t.test(n.escapedName))return n.escapedName.substring(1,n.escapedName.length-1)}if(!a.enclosingDeclaration||!a.tracker.moduleResolverHost)return t.test(n.escapedName)?n.escapedName.substring(1,n.escapedName.length-1):e.getSourceFileOfNode(e.getNonAugmentationDeclaration(n)).fileName;var d=e.getSourceFileOfNode(e.getOriginalNode(a.enclosingDeclaration)),p=qt(n),m=p.specifierCache&&p.specifierCache.get(d.path);if(!m){var f=L.out||L.outFile,_=a.tracker.moduleResolverHost,g=f?o({},L,{baseUrl:_.getCommonSourceDirectory()}):L;m=e.first(e.moduleSpecifiers.getModuleSpecifiers(n,g,d,_,r.getSourceFiles(),{importModuleSpecifierPreference:f?"non-relative":"relative"},r.redirectTargetsMap)),p.specifierCache=p.specifierCache||e.createMap(),p.specifierCache.set(d.path,m)}return m}function x(n,t,r,a){var i=h(n,t,r,!(16384&t.flags)),o=67220415===r;if(e.some(i[0].declarations,wa)){var s=i.length>1?g(i,i.length-1,1):void 0,c=a||E(i,0,t),l=S(i[0],t);!(67108864&t.flags)&&e.getEmitModuleResolutionKind(L)===e.ModuleResolutionKind.NodeJs&&l.indexOf("/node_modules/")>=0&&(t.encounteredError=!0,t.tracker.reportLikelyUnsafeImportRequiredError&&t.tracker.reportLikelyUnsafeImportRequiredError(l));var u=e.createLiteralTypeNode(e.createLiteral(l));if(t.tracker.trackExternalModuleSymbolOfImportTypeNode&&t.tracker.trackExternalModuleSymbolOfImportTypeNode(i[0]),t.approximateLength+=l.length+10,!s||e.isEntityName(s)){if(s){var d=e.isIdentifier(s)?s:s.right;d.typeArguments=void 0}return e.createImportTypeNode(u,s,c,o)}var p=T(s),m=p.objectType.typeName;return e.createIndexedAccessTypeNode(e.createImportTypeNode(u,m,c,o),p.indexType)}var f=g(i,i.length-1,0);if(e.isIndexedAccessTypeNode(f))return f;if(o)return e.createTypeQueryNode(f);d=e.isIdentifier(f)?f:f.right;var _=d.typeArguments;return d.typeArguments=void 0,e.createTypeReferenceNode(f,_);function g(n,r,i){var o=r===n.length-1?a:E(n,r,t),s=n[r];0===r&&(t.flags|=16777216);var c=za(s,t);t.approximateLength+=c.length+1,0===r&&(t.flags^=16777216);var l=n[r-1];if(!(16&t.flags)&&l&&zo(l)&&zo(l).get(s.escapedName)===s){var u=g(n,r-1,i);return e.isIndexedAccessTypeNode(u)?e.createIndexedAccessTypeNode(u,e.createLiteralTypeNode(e.createLiteral(c))):e.createIndexedAccessTypeNode(e.createTypeReferenceNode(u,o),e.createLiteralTypeNode(e.createLiteral(c)))}var d=e.setEmitFlags(e.createIdentifier(c,o),16777216);if(d.symbol=s,r>i){u=g(n,r-1,i);return e.isEntityName(u)?e.createQualifiedName(u,d):e.Debug.fail("Impossible construct - an export of an indexed access cannot be reachable")}return d}}function C(n,t,r,a){var i=h(n,t,r);return!a||1===i.length||t.encounteredError||65536&t.flags||(t.encounteredError=!0),o(i,i.length-1);function o(n,r){var a=E(n,r,t),i=n[r];0===r&&(t.flags|=16777216);var s=za(i,t);0===r&&(t.flags^=16777216);var c=e.setEmitFlags(e.createIdentifier(s,a),16777216);return c.symbol=i,r>0?e.createQualifiedName(o(n,r-1),c):c}}function A(n,t,r){var a=h(n,t,r);return i(a,a.length-1);function i(n,r){var a=E(n,r,t),o=n[r];if(e.some(o.declarations,wa))return e.createLiteral(S(o,t));0===r&&(t.flags|=16777216);var s=za(o,t);0===r&&(t.flags^=16777216);var c=s.charCodeAt(0),l=e.isIdentifierStart(c,k);if(0===r||l){var u=e.setEmitFlags(e.createIdentifier(s,a),16777216);return u.symbol=o,r>0?e.createPropertyAccess(i(n,r-1),u):u}91===c&&(s=s.substring(1,s.length-1),c=s.charCodeAt(0));var d=void 0;return e.isSingleOrDoubleQuote(c)?(d=e.createLiteral(s.substring(1,s.length-1).replace(/\\./g,function(e){return e.substring(1)})),d.singleQuote=39===c):""+ +s===s&&(d=e.createLiteral(+s)),d||(d=e.setEmitFlags(e.createIdentifier(s,a),16777216),d.symbol=o),e.createElementAccess(i(n,r-1),d)}}}function Ba(n,t,r,a){return void 0===r&&(r=16384),a?i(a).getText():e.usingSingleLineStringWriter(i);function i(a){var i=e.createTypePredicateNode(1===n.kind?e.createIdentifier(n.parameterName):e.createThisTypeNode(),K.typeToTypeNode(n.type,t,70222336|Ga(r))),o=e.createPrinter({removeComments:!0}),s=t&&e.getSourceFileOfNode(t);return o.writeNode(4,i,s,a),a}}function Ka(e){for(var n=[],t=0,r=0;r<e.length;r++){var a=e[r];if(t|=a.flags,!(98304&a.flags)){if(1536&a.flags){var i=512&a.flags?he:To(a);if(1048576&i.flags){var o=i.types.length;if(r+o<=e.length&&Ld(e[r+o-1])===Ld(i.types[o-1])){n.push(i),r+=o-1;continue}}}n.push(a)}}return 65536&t&&n.push(ue),32768&t&&n.push(ce),n||e}function ja(e){return 8===e?"private":16===e?"protected":"public"}function Ua(n){if(n.symbol&&2048&n.symbol.flags){var t=e.findAncestor(n.symbol.declarations[0].parent,function(e){return 177!==e.kind});if(242===t.kind)return ra(t)}}function Ha(n){return n&&n.parent&&245===n.parent.kind&&e.isExternalModuleAugmentation(n.parent.parent)}function Wa(n){return 279===n.kind||e.isAmbientModule(n)}function qa(n,t){var r=n.nameType;if(r){if(384&r.flags){var a=""+r.value;return e.isIdentifierText(a,L.target)||Uv(a)?a:'"'+e.escapeString(a,34)+'"'}if(8192&r.flags)return"["+za(r.symbol,t)+"]"}}function za(n,t){if(t&&"default"===n.escapedName&&!(16384&t.flags)&&(!(16777216&t.flags)||!n.declarations||t.enclosingDeclaration&&e.findAncestor(n.declarations[0],Wa)!==e.findAncestor(t.enclosingDeclaration,Wa)))return"default";if(n.declarations&&n.declarations.length){var r=n.declarations[0],a=e.getNameOfDeclaration(r);if(a){if(e.isCallExpression(r)&&e.isBindableObjectDefinePropertyCall(r))return e.symbolName(n);if(e.isComputedPropertyName(a)&&!(2048&e.getCheckFlags(n))&&n.nameType&&384&n.nameType.flags){var i=qa(n,t);if(void 0!==i)return i}return e.declarationNameToString(a)}if(r.parent&&237===r.parent.kind)return e.declarationNameToString(r.parent.name);switch(r.kind){case 209:case 196:case 197:return!t||t.encounteredError||131072&t.flags||(t.encounteredError=!0),209===r.kind?"(Anonymous class)":"(Anonymous function)"}}var o=qa(n,t);return void 0!==o?o:e.symbolName(n)}function Ja(n){if(n){var t=zt(n);return void 0===t.isVisible&&(t.isVisible=!!r()),t.isVisible}return!1;function r(){switch(n.kind){case 297:case 304:return!!(n.parent&&n.parent.parent&&n.parent.parent.parent&&e.isSourceFile(n.parent.parent.parent));case 186:return Ja(n.parent.parent);case 237:if(e.isBindingPattern(n.name)&&!n.name.elements.length)return!1;case 244:case 240:case 241:case 242:case 239:case 243:case 248:if(e.isExternalModuleAugmentation(n))return!0;var t=ei(n);return 1&e.getCombinedModifierFlags(n)||248!==n.kind&&279!==t.kind&&4194304&t.flags?Ja(t):Jt(t);case 154:case 153:case 158:case 159:case 156:case 155:if(e.hasModifier(n,24))return!1;case 157:case 161:case 160:case 162:case 151:case 245:case 165:case 166:case 168:case 164:case 169:case 170:case 173:case 174:case 177:return Ja(n.parent);case 250:case 251:case 253:return!1;case 150:case 279:case 247:return!0;case 254:return!1;default:return!1}}}function Xa(n,t){var r,a;return n.parent&&254===n.parent.kind?r=Zt(n,n.escapedText,70107135,void 0,n,!1):257===n.parent.kind&&(r=Cr(n.parent,70107135)),r&&i(r.declarations),a;function i(n){e.forEach(n,function(n){var r=dr(n)||n;if(t?zt(n).isVisible=!0:(a=a||[],e.pushIfUnique(a,r)),e.isInternalModuleImportEqualsDeclaration(n)){var o=n.moduleReference,s=wx(o),c=Zt(n,s.escapedText,68009983,void 0,void 0,!1);c&&i(c.declarations)}})}}function Ya(e,n){var t=Qa(e,n);if(t>=0){for(var r=jn.length,a=t;a<r;a++)Un[a]=!1;return!1}return jn.push(e),Un.push(!0),Hn.push(n),!0}function Qa(e,n){for(var t=jn.length-1;t>=0;t--){if(Za(jn[t],Hn[t]))return-1;if(jn[t]===e&&Hn[t]===n)return t}return-1}function Za(n,t){switch(t){case 0:return!!qt(n).type;case 5:return!!zt(n).resolvedEnumType;case 2:return!!qt(n).declaredType;case 1:return!!n.resolvedBaseConstructorType;case 3:return!!n.resolvedReturnType;case 4:return!!n.immediateBaseConstraint;case 6:return!!qt(n).resolvedJSDocType}return e.Debug.assertNever(t)}function $a(){return jn.pop(),Hn.pop(),Un.pop()}function ei(n){return e.findAncestor(e.getRootDeclaration(n),function(e){switch(e.kind){case 237:case 238:case 253:case 252:case 251:case 250:return!1;default:return!0}}).parent}function ni(n){var t=Do(aa(n));return t.typeParameters?cl(t,e.map(t.typeParameters,function(e){return re})):t}function ti(e,n){var t=sc(e,n);return t?Xi(t):void 0}function ri(e,n){return ti(e,n)||Uv(n)&&mc(e,1)||mc(e,0)||se}function ai(e){return e&&0!==(1&e.flags)}function ii(e){var n=ra(e);return n&&qt(n).type||yi(e,!1)}function oi(n){return 149===n.kind&&!e.isStringOrNumericLiteralLike(n.expression)}function si(n,t,r){if(n=J_(n,function(e){return!(98304&e.flags)}),131072&n.flags)return ke;if(1048576&n.flags)return X_(n,function(e){return si(e,t,r)});var a=Au(e.map(t,Bu));if($u(n)||ed(a)){if(131072&a.flags)return n;var i=$l(),o=Zl();if(!i||!o)return oe;var s=pl(o,[Wu(n),a]);return pl(i,[n,s])}for(var c=e.createSymbolTable(),l=0,u=Rs(n);l<u.length;l++){var d=u[l];Lp(ju(d,8576),a)||24&e.getDeclarationModifierFlagsFromSymbol(d)||!Sd(d)||c.set(d.escapedName,xd(d))}var p=pc(n,0),m=pc(n,1);return ba(r,c,e.emptyArray,e.emptyArray,p,m)}function ci(e,n){var t=li(e);return t?fg(t,n):n}function li(n){var t=ui(n);if(t&&t.flowNode){var r=di(n);if(r){var a=e.createNode(190,n.pos,n.end);a.parent=n,a.expression=t;var i=e.createNode(10,n.pos,n.end);return i.parent=a,i.text=r,a.argumentExpression=i,a.flowNode=t.flowNode,a}}}function ui(e){var n=e.parent.parent;switch(n.kind){case 186:case 275:return li(n);case 187:return li(e.parent);case 237:return n.initializer;case 204:return n.right}}function di(e){var n=e.parent;return 186===e.kind&&184===n.kind?pi(e.propertyName||e.name):275===e.kind||276===e.kind?pi(e.name):""+n.elements.indexOf(e)}function pi(e){var n=Bu(e);return 384&n.flags?""+n.value:void 0}function mi(n){var t,r=n.parent,a=ii(r.parent);if(a===oe)return oe;if(!a)return n.initializer?GE(n):a;if(ai(a))return a;if(w&&4194304&n.flags&&e.isParameterDeclaration(n)&&(a=lf(a)),184===r.kind)if(n.dotDotDotToken){if(2&a.flags||!Jv(a))return It(n,e.Diagnostics.Rest_types_may_only_be_created_from_object_types),oe;for(var i=[],o=0,s=r.elements;o<s.length;o++){var c=s[o];c.dotDotDotToken||i.push(c.propertyName||c.name)}t=si(a,i,n.symbol)}else{var l=n.propertyName||n.name,u=Bu(l),d=xg(ad(a,u,l),n.name);t=ci(n,d)}else{var p=KS(a,r,!1,!1),m=r.elements.indexOf(n);if(n.dotDotDotToken)t=z_(a,Xm)?X_(a,function(e){return fu(e,m)}):su(p);else if(Fm(a)){u=Nd(m),d=xg(ad(a,u,n.name),n.name);t=ci(n,d)}else t=p}return!w||!n.initializer||32768&nf(GE(n))||(t=b_(t,524288)),n.initializer&&!e.getEffectiveTypeAnnotationNode(e.walkUpBindingElementsAndPatterns(n))?Au([t,GE(n)],2):t}function fi(n){var t=e.getJSDocType(n);if(t)return Rd(t)}function _i(n){var t=e.skipParentheses(n);return 96===t.kind||72===t.kind&&e_(t)===j}function gi(n){var t=e.skipParentheses(n);return 187===t.kind&&0===t.elements.length}function vi(e,n){return void 0===n&&(n=!0),w&&n?sf(e):e}function yi(n,t){if(e.isVariableDeclaration(n)&&226===n.parent.parent.kind){var r=Wu(wy(qE(n.parent.parent.expression)));return 4456448&r.flags?qu(r):pe}if(e.isVariableDeclaration(n)&&227===n.parent.parent.kind){var a=n.parent.parent;return BS(a.expression,a.awaitModifier)||re}if(e.isBindingPattern(n.parent))return mi(n);var i=t&&(e.isParameter(n)&&vc(n)||!e.isBindingElement(n)&&!e.isVariableDeclaration(n)&&!!n.questionToken),o=Mi(n);if(o)return vi(o,i);if((R||e.isInJSFile(n))&&237===n.kind&&!e.isBindingPattern(n.name)&&!(1&e.getCombinedModifierFlags(n))&&!(4194304&n.flags)){if(!(2&e.getCombinedNodeFlags(n))&&(!n.initializer||_i(n.initializer)))return ae;if(n.initializer&&gi(n.initializer))return nn}if(151===n.kind){var s=n.parent;if(159===s.kind&&!Ko(s)){var c=e.getDeclarationOfKind(ra(n.parent),158);if(c){var l=Cc(c),u=ZA(s);return u&&n===u?(e.Debug.assert(!u.type),Xi(l.thisParameter)):Fc(l)}}if(e.isInJSFile(n)){var d=e.getJSDocType(s);if(d&&e.isFunctionTypeNode(d))return Cb(Cc(d),s.parameters.indexOf(n))}var p="this"===n.symbol.escapedName?qg(s):zg(n);if(p)return vi(p,i)}else if(e.isInJSFile(n)){var m=bi(n,ra(n),e.getDeclaredExpandoInitializer(n));if(m)return m}if(n.initializer){p=GE(n);return vi(p,i)}return e.isJsxAttribute(n)?ve:e.isBindingPattern(n.name)?ki(n.name,!1,!0):void 0}function hi(n,t){var r=e.getAssignedExpandoInitializer(n.valueDeclaration);if(r){var a=e.getJSDocTypeTag(r);if(a&&a.typeExpression)return Rd(a.typeExpression);var i=bi(n.valueDeclaration,n,r);return i||qm(RE(r))}for(var o,s,c=!1,l=!1,u=0,d=n.declarations;u<d.length;u++){var p=d[u],m=e.isBinaryExpression(p)||e.isCallExpression(p)?p:e.isPropertyAccessExpression(p)?e.isBinaryExpression(p.parent)?p.parent:p:void 0;if(!m)return oe;var f=e.isPropertyAccessExpression(m)?e.getAssignmentDeclarationPropertyAccessKind(m):e.getAssignmentDeclarationKind(m);4===f&&(Si(m)?c=!0:l=!0),e.isCallExpression(m)||(o=Ei(o,m,n,p)),o||(s||(s=[])).push(e.isBinaryExpression(m)||e.isCallExpression(m)?Ti(n,t,m,f):Te)}var _=o;if(!_){var g=c?xi(s,n.declarations):void 0;if(l){var v=Ci(n);v&&((g||(g=[])).push(v),c=!0)}var y=e.some(g,function(e){return!!(-134316033&e.flags)})?g:s;_=Au(y,2)}var h=bf(vi(_,l&&!c));return J_(h,function(e){return!!(-98305&e.flags)})===Te?(Sf(n.valueDeclaration,re),re):h}function bi(n,t,r){if(e.isInJSFile(n)&&r&&e.isObjectLiteralExpression(r)&&!r.properties.length){var a=e.createSymbolTable();while(e.isBinaryExpression(n)||e.isPropertyAccessExpression(n)){var i=ra(n);i&&e.hasEntries(i.exports)&&Ut(a,i.exports),n=e.isBinaryExpression(n)?n.parent:n.parent.parent}var o=ra(n);o&&e.hasEntries(o.exports)&&Ut(a,o.exports);var s=ba(t,a,e.emptyArray,e.emptyArray,void 0,void 0);return s.objectFlags|=16384,s}}function Ei(n,t,r,a){var i=e.getJSDocType(t.parent);if(i){var o=bf(Rd(i));if(!n)return o;n===oe||o===oe||Sp(n,o)||AS(n,a,o)}return n}function Ti(n,t,r,a){if(e.isCallExpression(r)){if(t)return Xi(t);var i=RE(r.arguments[2]),o=ti(i,"value");if(o)return o;var s=ti(i,"get");if(s){var c=gh(s);if(c)return Fc(c)}var l=ti(i,"set");if(l){var u=gh(l);if(u)return Mb(u)}return re}var d=t?Xi(t):qm(RE(r.right));if(524288&d.flags&&2===a&&"export="===n.escapedName){var p=ws(d),m=e.createSymbolTable();e.copyEntries(p.members,m),t&&!t.exports&&(t.exports=e.createSymbolTable()),(t||n).exports.forEach(function(e,n){if(m.has(n)){var t=p.members.get(n),r=Ot(e.flags|t.flags,n);r.type=Au([Xi(e),Xi(t)]),m.set(n,r)}else m.set(n,e)});var f=ba(p.symbol,m,p.callSignatures,p.constructSignatures,p.stringIndexInfo,p.numberIndexInfo);return f.objectFlags|=16384&e.getObjectFlags(d),f}return Gm(d)?(Sf(r,en),en):d}function Si(n){var t=e.getThisContainer(n,!1);return 157===t.kind||239===t.kind||196===t.kind&&!e.isPrototypePropertyAssignment(t.parent)}function xi(n,t){return e.Debug.assert(n.length===t.length),n.filter(function(n,r){var a=t[r],i=e.isBinaryExpression(a)?a:e.isBinaryExpression(a.parent)?a.parent:void 0;return i&&Si(i)})}function Ci(n){var t=e.forEach(n.declarations,function(n){var t=e.getThisContainer(n,!1).parent;return e.isClassLike(t)&&t});if(t){var r=Do(ra(t)),a=r&&uo(r)[0];if(a)return ti(a,n.escapedName)}}function Ai(n,t,r){return n.initializer?vi(GE(n)):e.isBindingPattern(n.name)?ki(n.name,t,r):(r&&!wi(n)&&Sf(n,re),re)}function Di(n,t,r){var a,i=e.createSymbolTable(),o=128;e.forEach(n.elements,function(e){var n=e.propertyName||e.name;if(e.dotDotDotToken)a=el(re,!1);else{var s=Bu(n);if(Fo(s)){var c=Uo(s),l=4|(e.initializer?16777216:0),u=Ot(l,c);u.type=Ai(e,t,r),u.bindingElement=e,i.set(u.escapedName,u)}else o|=512}});var s=ba(void 0,i,e.emptyArray,e.emptyArray,a,void 0);return s.flags|=268435456,s.objectFlags|=o,t&&(s.pattern=n),s}function Li(n,t,r){var a=n.elements,i=e.lastOrUndefined(a),o=!(!i||186!==i.kind||!i.dotDotDotToken);if(0===a.length||1===a.length&&o)return k>=2?iu(re):en;var s=e.map(a,function(n){return e.isOmittedExpression(n)?re:Ai(n,t,r)}),c=e.findLastIndex(a,function(n){return!e.isOmittedExpression(n)&&!Rv(n)},a.length-(o?2:1))+1,l=pu(s,c,o);return t&&(l=ll(l),l.pattern=n),l}function ki(e,n,t){return void 0===n&&(n=!1),void 0===t&&(t=!1),184===e.kind?Di(e,n,t):Li(e,n,t)}function Ni(e,n){return Ii(yi(e,!0),e,n)}function Ii(n,t,r){return n?(r&&xf(t,n),8192&n.flags&&(e.isBindingElement(t)||!t.type)&&n.symbol!==ra(t)&&(n=be),bf(n)):(n=e.isParameter(t)&&t.dotDotDotToken?en:re,r&&(wi(t)||Sf(t,n)),n)}function wi(n){var t=e.getRootDeclaration(n),r=151===t.kind?t.parent:t;return kT(r)}function Mi(n){var t=e.getEffectiveTypeAnnotationNode(n);if(t)return Rd(t)}function Oi(e){var n=qt(e);return n.type||(n.type=Pi(e))}function Pi(n){if(4194304&n.flags)return ni(n);if(n===W)return re;if(134217728&n.flags){var t=ra(e.getSourceFileOfNode(n.valueDeclaration)),r=e.createSymbolTable();return r.set("exports",t),ba(n,r,e.emptyArray,e.emptyArray,void 0,void 0)}var a,i=n.valueDeclaration;if(e.isCatchClauseVariableDeclarationOrBindingElement(i))return re;if(e.isSourceFile(i)&&e.isJsonSourceFile(i)){if(!i.statements.length)return ke;var o=qm(qE(i.statements[0].expression));return 524288&o.flags?mf(o):o}if(254===i.kind)return Ii(RE(i.expression),i);if(!Ya(n,0))return 512&n.flags?Ui(n):oe;if(e.isInJSFile(i)&&(e.isCallExpression(i)||e.isBinaryExpression(i)||e.isPropertyAccessExpression(i)&&e.isBinaryExpression(i.parent)))a=hi(n);else if(e.isJSDocPropertyLikeTag(i)||e.isPropertyAccessExpression(i)||e.isIdentifier(i)||e.isClassDeclaration(i)||e.isFunctionDeclaration(i)||e.isMethodDeclaration(i)&&!e.isObjectLiteralMethod(i)||e.isMethodSignature(i)||e.isSourceFile(i)){if(9136&n.flags)return Ui(n);a=e.isBinaryExpression(i.parent)?hi(n):Mi(i)||re}else if(e.isPropertyAssignment(i))a=Mi(i)||KE(i);else if(e.isJsxAttribute(i))a=Mi(i)||ty(i);else if(e.isShorthandPropertyAssignment(i))a=Mi(i)||BE(i.name,0);else if(e.isObjectLiteralMethod(i))a=Mi(i)||jE(i,0);else if(e.isParameter(i)||e.isPropertyDeclaration(i)||e.isPropertySignature(i)||e.isVariableDeclaration(i)||e.isBindingElement(i))a=Ni(i,!0);else if(e.isEnumDeclaration(i))a=Ui(n);else{if(!e.isEnumMember(i))return e.Debug.fail("Unhandled declaration kind! "+e.Debug.showSyntaxKind(i)+" for "+e.Debug.showSymbol(n));a=Wi(n)}if(!$a()){if(512&n.flags)return Ui(n);a=Ji(n)}return a}function Ri(n){if(n){if(158===n.kind){var t=e.getEffectiveReturnTypeNode(n);return t}var r=e.getEffectiveSetAccessorTypeAnnotationNode(n);return r}}function Fi(e){var n=Ri(e);return n&&Rd(n)}function Gi(e){var n=ZA(e);return n&&n.symbol}function Vi(e){return wc(Cc(e))}function Bi(e){var n=qt(e);return n.type||(n.type=Ki(e))}function Ki(n){var t,r=e.getDeclarationOfKind(n,158),a=e.getDeclarationOfKind(n,159);if(r&&e.isInJSFile(r)){var i=fi(r);if(i)return i}if(!Ya(n,0))return oe;var o=Fi(r);if(o)t=o;else{var s=Fi(a);s?t=s:r&&r.body?t=jb(r):(a?Mt(R,a,e.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation,Pa(n)):(e.Debug.assert(!!r,"there must existed getter as we are current checking either setter or getter in this function"),Mt(R,r,e.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation,Pa(n))),t=re)}if(!$a()&&(t=re,R)){var c=e.getDeclarationOfKind(n,158);It(c,e.Diagnostics._0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions,Pa(n))}return t}function ji(e){var n=lo(vo(e));return 8650752&n.flags?n:void 0}function Ui(n){var t=qt(n),r=t;if(!t.type){var a=e.getDeclarationOfExpando(n.valueDeclaration);if(a){var i=ra(a);i&&(e.hasEntries(i.exports)||e.hasEntries(i.members))&&(n=Gt(n),t=n,e.hasEntries(i.exports)&&(n.exports=n.exports||e.createSymbolTable(),Ut(n.exports,i.exports)),e.hasEntries(i.members)&&(n.members=n.members||e.createSymbolTable(),Ut(n.members,i.members)))}r.type=t.type=Hi(n)}return t.type}function Hi(n){var t=n.valueDeclaration;if(1536&n.flags&&e.isShorthandAmbientModuleSymbol(n))return re;if(204===t.kind||189===t.kind&&204===t.parent.kind)return hi(n);if(512&n.flags&&t&&e.isSourceFile(t)&&t.commonJsModuleIndicator){var r=Hr(n);if(r!==n){if(!Ya(n,0))return oe;var a=ta(n.exports.get("export=")),i=hi(a,a===r?void 0:r);return $a()?i:Ji(n)}}var o=fa(16,n);if(32&n.flags){var s=ji(n);return s?Ru([o,s]):o}return w&&16777216&n.flags?sf(o):o}function Wi(e){var n=qt(e);return n.type||(n.type=xo(e))}function qi(e){var n=qt(e);if(!n.type){var t=Nr(e);n.type=67220415&t.flags?Xi(t):oe}return n.type}function zi(e){var n=qt(e);if(!n.type){if(!Ya(e,0))return n.type=oe;var t=mp(Xi(n.target),n.mapper);$a()||(t=Ji(e)),n.type=t}return n.type}function Ji(n){return e.getEffectiveTypeAnnotationNode(n.valueDeclaration)?(It(n.valueDeclaration,e.Diagnostics._0_is_referenced_directly_or_indirectly_in_its_own_type_annotation,Pa(n)),oe):(R&&It(n.valueDeclaration,e.Diagnostics._0_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_and_is_referenced_directly_or_indirectly_in_its_own_initializer,Pa(n)),re)}function Xi(n){return 1&e.getCheckFlags(n)?zi(n):4096&e.getCheckFlags(n)?Pf(n):7&n.flags?Oi(n):9136&n.flags?Ui(n):8&n.flags?Wi(n):98304&n.flags?Bi(n):2097152&n.flags?qi(n):oe}function Yi(n,t){return void 0!==n&&void 0!==t&&0!==(4&e.getObjectFlags(n))&&n.target===t}function Qi(n){return 4&e.getObjectFlags(n)?n.target:n}function Zi(n,t){return r(n);function r(n){if(7&e.getObjectFlags(n)){var a=Qi(n);return a===t||e.some(uo(a),r)}return!!(2097152&n.flags)&&e.some(n.types,r)}}function $i(n,t){for(var r=0,a=t;r<a.length;r++){var i=a[r];n=e.appendIfUnique(n,Co(ra(i)))}return n}function eo(n,t){while(1){if(n=n.parent,!n)return;switch(n.kind){case 240:case 209:case 241:case 160:case 161:case 155:case 165:case 166:case 289:case 239:case 156:case 196:case 197:case 242:case 303:case 304:case 297:case 181:case 175:var r=eo(n,t);if(181===n.kind)return e.append(r,Co(ra(n.typeParameter)));if(175===n.kind)return e.concatenate(r,dd(n));var a=$i(r,e.getEffectiveTypeParameterDeclarations(n)),i=t&&(240===n.kind||209===n.kind||241===n.kind)&&vo(ra(n)).thisType;return i?e.append(a,i):a}}}function no(n){var t=32&n.flags?n.valueDeclaration:e.getDeclarationOfKind(n,241);return eo(t)}function to(n){for(var t,r=0,a=n.declarations;r<a.length;r++){var i=a[r];if(241===i.kind||240===i.kind||209===i.kind||e.isTypeAlias(i)){var o=i;t=$i(t,e.getEffectiveTypeParameterDeclarations(o))}}return t}function ro(n){return e.concatenate(no(n),to(n))}function ao(e){var n=lc(e,1);if(1===n.length){var t=n[0];return!t.typeParameters&&1===t.parameters.length&&t.hasRestParameter&&Sb(t.parameters[0])===en}return!1}function io(e){if(fo(e)&&lc(e,1).length>0)return!0;if(8650752&e.flags){var n=qs(e);return!!n&&fo(n)&&ao(n)}return ob(e)}function oo(n){return e.getEffectiveBaseTypeNode(n.symbol.valueDeclaration)}function so(n,t,r){var a=e.length(t),i=e.isInJSFile(r);return e.filter(lc(n,1),function(n){return(i||a>=Sc(n.typeParameters))&&a<=e.length(n.typeParameters)})}function co(n,t,r){var a=so(n,t,r),i=e.map(t,Rd);return e.sameMap(a,function(n){return e.some(n.typeParameters)?jc(n,i,e.isInJSFile(r)):n})}function lo(n){if(!n.resolvedBaseConstructorType){var t=n.symbol.valueDeclaration,r=e.getEffectiveBaseTypeNode(t),a=oo(n);if(!a)return n.resolvedBaseConstructorType=ce;if(!Ya(n,1))return oe;var i=qE(a.expression);if(r&&a!==r&&(e.Debug.assert(!r.typeArguments),qE(r.expression)),2621440&i.flags&&ws(i),!$a())return It(n.symbol.valueDeclaration,e.Diagnostics._0_is_referenced_directly_or_indirectly_in_its_own_base_expression,Pa(n.symbol)),n.resolvedBaseConstructorType=oe;if(!(1&i.flags)&&i!==de&&!io(i)){var o=It(a.expression,e.Diagnostics.Type_0_is_not_a_constructor_function_type,Fa(i));if(262144&i.flags){var s=al(i),c=se;if(s){var l=lc(s,1);l[0]&&(c=Fc(l[0]))}Nt(o,e.createDiagnosticForNode(i.symbol.declarations[0],e.Diagnostics.Did_you_mean_for_0_to_be_constrained_to_type_new_args_Colon_any_1,Pa(i.symbol),Fa(c)))}return n.resolvedBaseConstructorType=oe}n.resolvedBaseConstructorType=i}return n.resolvedBaseConstructorType}function uo(n){return n.resolvedBaseTypes||(8&n.objectFlags?n.resolvedBaseTypes=[su(Au(n.typeParameters||e.emptyArray))]:96&n.symbol.flags?(32&n.symbol.flags&&po(n),64&n.symbol.flags&&_o(n)):e.Debug.fail("type must be class or interface")),n.resolvedBaseTypes}function po(n){n.resolvedBaseTypes=e.resolvingEmptyArray;var t=rc(lo(n));if(!(2621441&t.flags))return n.resolvedBaseTypes=e.emptyArray;var r,a=oo(n),i=Ll(a),o=ob(t)?t:t.symbol?Do(t.symbol):void 0;if(t.symbol&&32&t.symbol.flags&&mo(o))r=dl(a,t.symbol,i);else if(1&t.flags)r=t;else if(ob(t))r=!a.typeArguments&&sb(t.symbol)||re;else{var s=co(t,a.typeArguments,a);if(!s.length)return It(a.expression,e.Diagnostics.No_base_constructor_has_the_specified_number_of_type_arguments),n.resolvedBaseTypes=e.emptyArray;r=Fc(s[0])}return r===oe?n.resolvedBaseTypes=e.emptyArray:fo(r)?n===r||Zi(r,n)?(It(n.symbol.valueDeclaration,e.Diagnostics.Type_0_recursively_references_itself_as_a_base_type,Fa(n,void 0,2)),n.resolvedBaseTypes=e.emptyArray):(n.resolvedBaseTypes===e.resolvingEmptyArray&&(n.members=void 0),n.resolvedBaseTypes=[r]):(It(a.expression,e.Diagnostics.Base_constructor_return_type_0_is_not_an_object_type_or_intersection_of_object_types_with_statically_known_members,Fa(r)),n.resolvedBaseTypes=e.emptyArray)}function mo(e){var n=e.outerTypeParameters;if(n){var t=n.length-1,r=e.typeArguments;return n[t].symbol!==r[t].symbol}return!0}function fo(n){return!!(67633153&n.flags)&&!Is(n)||!!(2097152&n.flags)&&e.every(n.types,fo)}function _o(n){n.resolvedBaseTypes=n.resolvedBaseTypes||e.emptyArray;for(var t=0,r=n.symbol.declarations;t<r.length;t++){var a=r[t];if(241===a.kind&&e.getInterfaceBaseTypeNodes(a))for(var i=0,o=e.getInterfaceBaseTypeNodes(a);i<o.length;i++){var s=o[i],c=Rd(s);c!==oe&&(fo(c)?n===c||Zi(c,n)?It(a,e.Diagnostics.Type_0_recursively_references_itself_as_a_base_type,Fa(n,void 0,2)):n.resolvedBaseTypes===e.emptyArray?n.resolvedBaseTypes=[c]:n.resolvedBaseTypes.push(c):It(s,e.Diagnostics.An_interface_can_only_extend_an_object_type_or_intersection_of_object_types_with_statically_known_members))}}}function go(n){for(var t=0,r=n.declarations;t<r.length;t++){var a=r[t];if(241===a.kind){if(64&a.flags)return!1;var i=e.getInterfaceBaseTypeNodes(a);if(i)for(var o=0,s=i;o<s.length;o++){var c=s[o];if(e.isEntityNameExpression(c.expression)){var l=Pr(c.expression,67897832,!0);if(!l||!(64&l.flags)||vo(l).thisType)return!1}}}}return!0}function vo(n){var t=qt(n);if(!t.declaredType){var r=32&n.flags?1:2,a=t.declaredType=fa(r,n),i=no(n),o=to(n);(i||o||1===r||!go(n))&&(a.objectFlags|=4,a.typeParameters=e.concatenate(i,o),a.outerTypeParameters=i,a.localTypeParameters=o,a.instantiations=e.createMap(),a.instantiations.set(ol(a.typeParameters),a),a.target=a,a.typeArguments=a.typeParameters,a.thisType=ga(n),a.thisType.isThisType=!0,a.thisType.constraint=a)}return t.declaredType}function yo(n){var t=qt(n);if(!t.declaredType){if(!Ya(n,2))return oe;var r=e.find(n.declarations,function(n){return e.isJSDocTypeAlias(n)||242===n.kind}),a=e.isJSDocTypeAlias(r)?r.typeExpression:r.type,i=a?Rd(a):oe;if($a()){var o=to(n);o&&(t.typeParameters=o,t.instantiations=e.createMap(),t.instantiations.set(ol(o),i))}else i=oe,It(r.name,e.Diagnostics.Type_alias_0_circularly_references_itself,Pa(n));t.declaredType=i}return t.declaredType}function ho(e){return 10===e.kind||204===e.kind&&(ho(e.left)&&ho(e.right))}function bo(n){var t=n.initializer;if(!t)return!(4194304&n.flags);switch(t.kind){case 10:case 8:return!0;case 202:return 39===t.operator&&8===t.operand.kind;case 72:return e.nodeIsMissing(t)||!!ra(n.parent).exports.get(t.escapedText);case 204:return ho(t);default:return!1}}function Eo(e){var n=qt(e);if(void 0!==n.enumKind)return n.enumKind;for(var t=!1,r=0,a=e.declarations;r<a.length;r++){var i=a[r];if(243===i.kind)for(var o=0,s=i.members;o<s.length;o++){var c=s[o];if(c.initializer&&10===c.initializer.kind)return n.enumKind=1;bo(c)||(t=!0)}}return n.enumKind=t?0:1}function To(e){return 1024&e.flags&&!(1048576&e.flags)?Do(aa(e.symbol)):e}function So(e){var n=qt(e);if(n.declaredType)return n.declaredType;if(1===Eo(e)){S++;for(var t=[],r=0,a=e.declarations;r<a.length;r++){var i=a[r];if(243===i.kind)for(var o=0,s=i.members;o<s.length;o++){var c=s[o],l=Dd(Nd(ZC(c),S,ra(c)));qt(ra(c)).declaredType=l,t.push(Ld(l))}}if(t.length){var u=Au(t,1,e,void 0);return 1048576&u.flags&&(u.flags|=1024,u.symbol=e),n.declaredType=u}}var d=da(32);return d.symbol=e,n.declaredType=d}function xo(e){var n=qt(e);if(!n.declaredType){var t=So(aa(e));n.declaredType||(n.declaredType=t)}return n.declaredType}function Co(e){var n=qt(e);return n.declaredType||(n.declaredType=ga(e))}function Ao(e){var n=qt(e);return n.declaredType||(n.declaredType=Do(Nr(e)))}function Do(e){return Lo(e)||oe}function Lo(e){return 96&e.flags?vo(e):524288&e.flags?yo(e):262144&e.flags?Co(e):384&e.flags?So(e):8&e.flags?xo(e):2097152&e.flags?Ao(e):void 0}function ko(e){switch(e.kind){case 120:case 143:case 138:case 135:case 146:case 123:case 139:case 136:case 106:case 141:case 96:case 132:case 182:return!0;case 169:return ko(e.elementType);case 164:return!e.typeArguments||e.typeArguments.every(ko)}return!1}function No(n){var t=e.getEffectiveConstraintOfTypeParameter(n);return!t||ko(t)}function Io(n){var t=e.getEffectiveTypeAnnotationNode(n);return t?ko(t):!e.hasInitializer(n)}function wo(n){var t=e.getEffectiveReturnTypeNode(n),r=e.getEffectiveTypeParameterDeclarations(n);return(157===n.kind||!!t&&ko(t))&&n.parameters.every(Io)&&r.every(No)}function Mo(e){if(e.declarations&&1===e.declarations.length){var n=e.declarations[0];if(n)switch(n.kind){case 154:case 153:return Io(n);case 156:case 155:case 157:return wo(n)}}return!1}function Oo(n,t,r){for(var a=e.createSymbolTable(),i=0,o=n;i<o.length;i++){var s=o[i];a.set(s.escapedName,r&&Mo(s)?s:tp(s,t))}return a}function Po(e,n){for(var t=0,r=n;t<r.length;t++){var a=r[t];e.has(a.escapedName)||e.set(a.escapedName,a)}}function Ro(n){if(!n.declaredProperties){var t=n.symbol,r=zo(t);n.declaredProperties=ya(r),n.declaredCallSignatures=e.emptyArray,n.declaredConstructSignatures=e.emptyArray,n.declaredCallSignatures=Nc(r.get("__call")),n.declaredConstructSignatures=Nc(r.get("__new")),n.declaredStringIndexInfo=nl(t,0),n.declaredNumberIndexInfo=nl(t,1)}return n}function Fo(e){return!!(8576&e.flags)}function Go(n){return e.isComputedPropertyName(n)&&e.isEntityNameExpression(n.expression)&&Fo(Hv(n))}function Vo(e){return 95===e.charCodeAt(0)&&95===e.charCodeAt(1)&&64===e.charCodeAt(2)}function Bo(n){var t=e.getNameOfDeclaration(n);return!!t&&Go(t)}function Ko(n){return e.hasDynamicName(n)&&!Bo(n)}function jo(n){return e.isDynamicName(n)&&!Go(n)}function Uo(n){return 8192&n.flags?n.escapedName:384&n.flags?e.escapeLeadingUnderscores(""+n.value):e.Debug.fail()}function Ho(n,t,r){e.Debug.assert(!!(2048&e.getCheckFlags(n)),"Expected a late-bound symbol."),n.flags|=r,qt(t.symbol).lateSymbol=n,n.declarations?n.declarations.push(t):n.declarations=[t],67220415&r&&(n.valueDeclaration&&n.valueDeclaration.kind===t.kind||(n.valueDeclaration=t))}function Wo(n,t,r,a){e.Debug.assert(!!a.symbol,"The member is expected to have a symbol.");var i=zt(a);if(!i.resolvedSymbol){i.resolvedSymbol=a.symbol;var o=Hv(a.name);if(Fo(o)){var s=Uo(o),c=a.symbol.flags,l=r.get(s);l||r.set(s,l=Ot(0,s,2048));var u=t&&t.get(s);if(l.flags&Rt(c)||u){var d=u?e.concatenate(u.declarations,l.declarations):l.declarations,p=!(8192&o.flags)&&e.unescapeLeadingUnderscores(s)||e.declarationNameToString(a.name);e.forEach(d,function(n){return It(e.getNameOfDeclaration(n)||n,e.Diagnostics.Property_0_was_also_declared_here,p)}),It(a.name||a,e.Diagnostics.Duplicate_property_0,p),l=Ot(0,s,2048)}return l.nameType=o,Ho(l,a,c),l.parent?e.Debug.assert(l.parent===n,"Existing symbol parent should match new one"):l.parent=n,i.resolvedSymbol=l}}return i.resolvedSymbol}function qo(n,t){var r=qt(n);if(!r[t]){var a="resolvedExports"===t,i=a?1536&n.flags?na(n):n.exports:n.members;r[t]=i||A;for(var o=e.createSymbolTable(),s=0,c=n.declarations;s<c.length;s++){var l=c[s],u=e.getMembersOfDeclaration(l);if(u)for(var d=0,p=u;d<p.length;d++){var m=p[d];a===e.hasStaticModifier(m)&&Bo(m)&&Wo(n,i,o,m)}}r[t]=jt(i,o)||A}return r[t]}function zo(e){return 6240&e.flags?qo(e,"resolvedMembers"):e.members||A}function Jo(n){if(106500&n.flags&&"__computed"===n.escapedName){var t=qt(n);return!t.lateSymbol&&e.some(n.declarations,Bo)&&(e.some(n.declarations,e.hasStaticModifier)?Zr(n.parent):zo(n.parent)),t.lateSymbol||(t.lateSymbol=n)}return n}function Xo(n,t,r){if(4&e.getObjectFlags(n)){var a=n.target,i=n.typeArguments;if(e.length(a.typeParameters)===e.length(i)){var o=cl(a,e.concatenate(i,[t||a.thisType]));return r?rc(o):o}}else if(2097152&n.flags)return Ru(e.map(n.types,function(e){return Xo(e,t,r)}));return r?rc(n):n}function Yo(n,t,r,a){var i,o,s,c,l,u;e.rangeEquals(r,a,0,r.length)?(i=D,o=t.symbol?zo(t.symbol):e.createSymbolTable(t.declaredProperties),s=t.declaredCallSignatures,c=t.declaredConstructSignatures,l=t.declaredStringIndexInfo,u=t.declaredNumberIndexInfo):(i=Ud(r,a),o=Oo(t.declaredProperties,i,1===r.length),s=Vd(t.declaredCallSignatures,i),c=Vd(t.declaredConstructSignatures,i),l=vp(t.declaredStringIndexInfo,i),u=vp(t.declaredNumberIndexInfo,i));var d=uo(t);if(d.length){t.symbol&&o===zo(t.symbol)&&(o=e.createSymbolTable(t.declaredProperties)),ha(n,o,s,c,l,u);for(var p=e.lastOrUndefined(a),m=0,f=d;m<f.length;m++){var _=f[m],g=p?Xo(mp(_,i),p):_;Po(o,Rs(g)),s=e.concatenate(s,lc(g,0)),c=e.concatenate(c,lc(g,1)),l||(l=g===re?el(re,!1):pc(g,0)),u=u||pc(g,1)}}ha(n,o,s,c,l,u)}function Qo(n){Yo(n,Ro(n),e.emptyArray,e.emptyArray)}function Zo(n){var t=Ro(n.target),r=e.concatenate(t.typeParameters,[t.thisType]),a=n.typeArguments&&n.typeArguments.length===r.length?n.typeArguments:e.concatenate(n.typeArguments,[n]);Yo(n,t,r,a)}function $o(e,n,t,r,a,i,o,s,c){var l=new b(q);return l.declaration=e,l.typeParameters=n,l.parameters=r,l.thisParameter=t,l.resolvedReturnType=a,l.resolvedTypePredicate=i,l.minArgumentCount=o,l.hasRestParameter=s,l.hasLiteralTypes=c,l.target=void 0,l.mapper=void 0,l}function es(e){return $o(e.declaration,e.typeParameters,e.thisParameter,e.parameters,void 0,void 0,e.minArgumentCount,e.hasRestParameter,e.hasLiteralTypes)}function ns(n){if(n.hasRestParameter){var t=n.parameters.length-1,r=n.parameters[t],a=Xi(r);if(Xm(a)){var i=a.typeArguments||e.emptyArray,o=a.target.minLength,s=a.target.hasRestElement?i.length-1:-1,c=e.map(i,function(e,r){var a=xb(n,t+r),i=r===s?16384:r>=o?8192:0,c=Ot(1,a,i);return c.type=r===s?su(e):e,c});return e.concatenate(n.parameters.slice(0,t),c)}}return n.parameters}function ts(n){var t=lo(n),r=lc(t,1);if(0===r.length)return[$o(void 0,n.localTypeParameters,void 0,e.emptyArray,n,void 0,0,!1,!1)];for(var a=oo(n),i=e.isInJSFile(a),o=Ll(a),s=e.length(o),c=[],l=0,u=r;l<u.length;l++){var d=u[l],p=Sc(d.typeParameters),m=e.length(d.typeParameters);if(i||s>=p&&s<=m){var f=m?Hc(d,xc(o,d.typeParameters,p,i)):es(d);f.typeParameters=n.localTypeParameters,f.resolvedReturnType=n,c.push(f)}}return c}function rs(e,n,t,r,a){for(var i=0,o=e;i<o.length;i++){var s=o[i];if(Lm(s,n,t,r,a,t?Ap:xp))return s}}function as(n,t,r){if(t.typeParameters){if(r>0)return;for(var a=1;a<n.length;a++)if(!rs(n[a],t,!1,!1,!1))return;return[t]}var i;for(a=0;a<n.length;a++){var o=a===r?t:rs(n[a],t,!0,!0,!0);if(!o)return;i=e.appendIfUnique(i,o)}return i}function is(n){for(var t,r,a=0;a<n.length;a++){if(0===n[a].length)return e.emptyArray;n[a].length>1&&(r=void 0===r?a:-1);for(var i=0,o=n[a];i<o.length;i++){var s=o[i];if(!t||!rs(t,s,!1,!0,!0)){var c=as(n,s,a);if(c){var l=s;if(c.length>1){var u=s.thisParameter;if(e.forEach(c,function(e){return e.thisParameter})){var d=Au(e.map(c,function(e){return e.thisParameter?Xi(e.thisParameter):re}),2);u=df(s.thisParameter,d)}l=es(s),l.thisParameter=u,l.unionSignatures=c}(t||(t=[])).push(l)}}}}if(!e.length(t)&&-1!==r){for(var p=n[void 0!==r?r:0],m=p.slice(),f=function(n){if(n!==p){var t=n[0];if(e.Debug.assert(!!t,"getUnionSignatures bails early on empty signature lists and should not have empty lists on second pass"),m=t.typeParameters&&e.some(m,function(e){return!!e.typeParameters})?void 0:e.map(m,function(e){return cs(e,t)}),!m)return"break"}},_=0,g=n;_<g.length;_++){var v=g[_],y=f(v);if("break"===y)break}t=m}return t||e.emptyArray}function os(e,n){if(!e||!n)return e||n;var t=Au([Xi(e),Xi(n)],2);return df(e,t)}function ss(e,n){for(var t=Lb(e)>=Lb(n)?e:n,r=t===e?n:e,a=Lb(t),i=Nb(e)||Nb(n),o=i&&!Nb(t),s=new Array(a+(o?1:0)),c=0;c<a;c++){var l=Ab(t,c),u=Ab(r,c)||se,d=Ru([l,u]),p=i&&!o&&c===a-1,m=c>=kb(t)&&c>=kb(r),f=xb(e,c),_=xb(n,c),g=Ot(1|(m&&!p?16777216:0),f===_?f:"arg"+c);g.type=p?su(d):d,s[c]=g}if(o){var v=Ot(1,"args");v.type=su(Cb(r,a)),s[a]=v}return s}function cs(n,t){var r=n.declaration,a=ss(n,t),i=os(n.thisParameter,t.thisParameter),o=Math.max(n.minArgumentCount,t.minArgumentCount),s=n.hasRestParameter||t.hasRestParameter,c=n.hasLiteralTypes||t.hasLiteralTypes,l=$o(r,n.typeParameters||t.typeParameters,i,a,void 0,void 0,o,s,c);return l.unionSignatures=e.concatenate(n.unionSignatures||[n],[t]),l}function ls(e,n){for(var t=[],r=!1,a=0,i=e;a<i.length;a++){var o=i[a],s=pc(o,n);if(!s)return;t.push(s.type),r=r||s.isReadonly}return el(Au(t,2),r)}function us(n){var t=is(e.map(n.types,function(e){return e===He?[Ln]:lc(e,0)})),r=is(e.map(n.types,function(e){return lc(e,1)})),a=ls(n.types,0),i=ls(n.types,1);ha(n,A,t,r,a,i)}function ds(e,n){return e?n?Ru([e,n]):e:n}function ps(e,n){return e?n?el(Ru([e.type,n.type]),e.isReadonly&&n.isReadonly):e:n}function ms(e,n){return e&&n&&el(Au([e.type,n.type]),e.isReadonly||n.isReadonly)}function fs(e,n,t){for(var r=[],a=0;a<n.length;a++)a===t?r.push(e):ao(n[a])&&r.push(Fc(lc(n[a],1)[0]));return Ru(r)}function _s(n){for(var t,r,a=e.emptyArray,i=e.emptyArray,o=n.types,s=e.countWhere(o,ao),c=function(c){var l=n.types[c];if(0===s||s===o.length&&0===c||!ao(l)){var u=lc(l,1);u.length&&s>0&&(u=e.map(u,function(e){var n=es(e);return n.resolvedReturnType=fs(Fc(e),o,c),n})),i=e.concatenate(i,u)}a=e.concatenate(a,lc(l,0)),t=ps(t,pc(l,0)),r=ps(r,pc(l,1))},l=0;l<o.length;l++)c(l);ha(n,A,a,i,t,r)}function gs(n){var t=n.symbol;if(n.target){ha(n,A,e.emptyArray,e.emptyArray,void 0,void 0);var r=Oo(Ms(n.target),n.mapper,!1),a=Vd(lc(n.target,0),n.mapper),i=Vd(lc(n.target,1),n.mapper),o=vp(pc(n.target,0),n.mapper),s=vp(pc(n.target,1),n.mapper);ha(n,r,a,i,o,s)}else if(2048&t.flags){ha(n,A,e.emptyArray,e.emptyArray,void 0,void 0);r=zo(t),a=Nc(r.get("__call")),i=Nc(r.get("__new")),o=nl(t,0),s=nl(t,1);ha(n,r,a,i,o,s)}else{r=A,o=void 0;if(t.exports&&(r=Zr(t)),ha(n,r,e.emptyArray,e.emptyArray,void 0,void 0),32&t.flags){var c=vo(t),l=lo(c);11272192&l.flags?(r=e.createSymbolTable(ya(r)),Po(r,Rs(l))):l===re&&(o=el(re,!1))}s=384&t.flags?In:void 0;if(ha(n,r,e.emptyArray,e.emptyArray,o,s),8208&t.flags&&(n.callSignatures=Nc(t),n.constructSignatures=e.filter(n.callSignatures,function(e){return ib(e.declaration)})),32&t.flags){c=vo(t),i=Nc(t.members.get("__constructor"));i.length||(i=ts(c)),n.constructSignatures=i}}}function vs(n){for(var t=pc(n.source,0),r=Ds(n.mappedType),a=!(1&r),i=4&r?0:16777216,o=t&&el(Rf(t.type,n.mappedType,n.constraintType),a&&t.isReadonly),s=e.createSymbolTable(),c=0,l=Rs(n.source);c<l.length;c++){var u=l[c],d=4096|(a&&tE(u)?8:0),p=Ot(4|u.flags&i,u.escapedName,d);p.declarations=u.declarations,p.nameType=u.nameType,p.propertyType=Xi(u),p.mappedType=n.mappedType,p.constraintType=n.constraintType,s.set(u.escapedName,p)}ha(n,s,e.emptyArray,e.emptyArray,o,void 0)}function ys(n){return 131069&n.flags?n:4194304&n.flags?Wu(rc(n.type)):16777216&n.flags?hs(n):1048576&n.flags?Au(e.sameMap(n.types,ys)):2097152&n.flags?Ru(e.sameMap(n.types,ys)):Te}function hs(e){if(e.root.isDistributive){var n=ys(e.checkType);if(n!==e.checkType){var t=Bd(e.root.checkType,n);return dp(e,Jd(t,e.mapper))}}return e}function bs(n){var t,r,a=e.createSymbolTable();ha(n,A,e.emptyArray,e.emptyArray,void 0,void 0);var i=Es(n),o=Ts(n),s=Ss(n.target||n),c=rc(As(n)),l=Ds(n),u=G?128:8576;if(Cs(n)){for(var d=0,p=Rs(c);d<p.length;d++){var m=p[d];f(ju(m,u))}(1&c.flags||pc(c,0))&&f(pe),!G&&pc(c,1)&&f(me)}else q_(ys(o),f);function f(e){var o=Jd(n.mapper,Ud([i],[e])),u=mp(s,o);if(Fo(e)){var d=Uo(e),p=sc(c,d),m=!!(4&l||!(8&l)&&p&&16777216&p.flags),f=!!(1&l||!(2&l)&&p&&tE(p)),_=Ot(4|(m?16777216:0),d,f?8:0);_.type=w&&m&&!Lp(ce,u)?sf(u):w&&!m&&p&&16777216&p.flags?b_(u,524288):u,p&&(_.syntheticOrigin=p,_.declarations=p.declarations),_.nameType=e,a.set(d,_)}else 5&e.flags?t=el(u,!!(1&l)):8&e.flags&&(r=el(u,!!(1&l)))}ha(n,a,e.emptyArray,e.emptyArray,t,r)}function Es(e){return e.typeParameter||(e.typeParameter=Co(ra(e.declaration.typeParameter)))}function Ts(e){return e.constraintType||(e.constraintType=Bs(Es(e))||oe)}function Ss(e){return e.templateType||(e.templateType=e.declaration.type?mp(vi(Rd(e.declaration.type),!!(4&Ds(e))),e.mapper||D):oe)}function xs(n){return e.getEffectiveConstraintOfTypeParameter(n.declaration.typeParameter)}function Cs(e){var n=xs(e);return 179===n.kind&&129===n.operator}function As(e){if(!e.modifiersType)if(Cs(e))e.modifiersType=mp(Rd(xs(e).type),e.mapper||D);else{var n=od(e.declaration),t=Ts(n),r=t&&262144&t.flags?Bs(t):t;e.modifiersType=r&&4194304&r.flags?mp(r.type,e.mapper||D):ke}return e.modifiersType}function Ds(e){var n=e.declaration;return(n.readonlyToken?39===n.readonlyToken.kind?2:1:0)|(n.questionToken?39===n.questionToken.kind?8:4:0)}function Ls(e){var n=Ds(e);return 8&n?-1:4&n?1:0}function ks(e){var n=Ls(e),t=As(e);return n||(Is(t)?Ls(t):0)}function Ns(n){return!!(32&e.getObjectFlags(n)&&4&Ds(n))}function Is(n){return!!(32&e.getObjectFlags(n))&&ed(Ts(n))}function ws(e){return e.members||(524288&e.flags?4&e.objectFlags?Zo(e):3&e.objectFlags?Qo(e):2048&e.objectFlags?vs(e):16&e.objectFlags?gs(e):32&e.objectFlags&&bs(e):1048576&e.flags?us(e):2097152&e.flags&&_s(e)),e}function Ms(n){return 524288&n.flags?ws(n).properties:e.emptyArray}function Os(e,n){if(524288&e.flags){var t=ws(e),r=t.members.get(n);if(r&&la(r))return r}}function Ps(n){if(!n.resolvedProperties){for(var t=e.createSymbolTable(),r=0,a=n.types;r<a.length;r++){for(var i=a[r],o=0,s=Rs(i);o<s.length;o++){var c=s[o];if(!t.has(c.escapedName)){var l=oc(n,c.escapedName);l&&t.set(c.escapedName,l)}}if(1048576&n.flags)break}n.resolvedProperties=ya(t)}return n.resolvedProperties}function Rs(e){return e=rc(e),3145728&e.flags?Ps(e):Ms(e)}function Fs(e,n){var t=n.properties;return t.some(function(n){var t=n.name&&Bu(n.name),r=t&&Fo(t)?Uo(t):void 0,a=void 0===r?void 0:ti(e,r);return!!a&&Hm(a)&&!Sp(TC(n),a)})}function Gs(n){var t=Au(n);if(!(1048576&t.flags))return LC(t);for(var r=e.createSymbolTable(),a=0,i=n;a<i.length;a++)for(var o=i[a],s=0,c=LC(o);s<c.length;s++){var l=c[s].escapedName;if(!r.has(l)){var u=ac(t,l);u&&r.set(l,u)}}return e.arrayFrom(r.values())}function Vs(e){return 262144&e.flags?Bs(e):8388608&e.flags?Ks(e):16777216&e.flags?Hs(e):qs(e)}function Bs(e){return Js(e)?al(e):void 0}function Ks(e){var n=Vs(e.objectType)||e.objectType;if(n!==e.objectType){var t=ad(n,e.indexType,void 0,oe);if(t&&t!==oe)return t}var r=qs(e);return r&&r!==e?r:void 0}function js(e){if(!e.resolvedDefaultConstraint){var n=e.root.trueType,t=33554432&n.flags?3&n.substitute.flags?n.typeVariable:Ru([n.substitute,n.typeVariable]):n;e.resolvedDefaultConstraint=Au([mp(t,e.combinedMapper||e.mapper),ud(e)])}return e.resolvedDefaultConstraint}function Us(e){if(e.root.isDistributive){var n=nd(e.checkType),t=n===e.checkType?Vs(n):n;if(t&&t!==e.checkType){var r=Bd(e.root.checkType,t),a=dp(e,Jd(r,e.mapper));if(!(131072&a.flags))return a}}}function Hs(e){return Us(e)||js(e)}function Ws(n,t){for(var r,a=!1,i=0,o=n.types;i<o.length;i++){var s=o[i];if(63176704&s.flags){var c=Vs(s);while(c&&21233664&c.flags)c=Vs(c);if(c){if(!(1048576&c.flags))return;r=e.append(r,c)}}else 67238908&s.flags&&(a=!0)}if(r&&(t||a)){if(a)for(var l=0,u=n.types;l<u.length;l++){s=u[l];67238908&s.flags&&(r=e.append(r,s))}return Ru(r)}}function qs(e){if(62128128&e.flags){var n=Xs(e);return n!==Pe&&n!==Re?n:void 0}return 4194304&e.flags?De:void 0}function zs(e){return qs(e)||e}function Js(e){return Xs(e)!==Re}function Xs(e){var n=!1;return e.resolvedBaseConstraint||(e.resolvedBaseConstraint=Xo(t(e),e));function t(e){if(!e.immediateBaseConstraint){if(!Ya(e,4))return Re;if(50===C)return n=!0,e.immediateBaseConstraint=Pe;C++;var t=a(nd(e));C--,$a()&&!n||(t=Re),e.immediateBaseConstraint=t||Pe}return e.immediateBaseConstraint}function r(e){var n=t(e);return n!==Pe&&n!==Re?n:void 0}function a(e){if(262144&e.flags){var n=al(e);return e.isThisType||!n?n:r(n)}if(3145728&e.flags){for(var t=e.types,a=[],i=0,o=t;i<o.length;i++){var s=o[i],c=r(s);c&&a.push(c)}return 1048576&e.flags&&a.length===t.length?Au(a):2097152&e.flags&&a.length?Ru(a):void 0}if(4194304&e.flags)return De;if(8388608&e.flags){var l=r(e.objectType),u=r(e.indexType),d=l&&u?ad(l,u,void 0,oe):void 0;return d&&d!==oe?r(d):void 0}if(16777216&e.flags){n=Hs(e);return n&&r(n)}return 33554432&e.flags?r(e.substitute):e}}function Ys(e){return e.resolvedApparentType||(e.resolvedApparentType=Xo(e,e,!0))}function Qs(n){if(n.default)n.default===Fe&&(n.default=Re);else if(n.target){var t=Qs(n.target);n.default=t?mp(t,n.mapper):Pe}else{n.default=Fe;var r=n.symbol&&e.forEach(n.symbol.declarations,function(n){return e.isTypeParameterDeclaration(n)&&n.default}),a=r?Rd(r):Pe;n.default===Fe&&(n.default=a)}return n.default}function Zs(e){var n=Qs(e);return n!==Pe&&n!==Re?n:void 0}function $s(e){return Qs(e)!==Re}function ec(n){return!(!n.symbol||!e.forEach(n.symbol.declarations,function(n){return e.isTypeParameterDeclaration(n)&&n.default}))}function nc(e){return e.resolvedApparentType||(e.resolvedApparentType=tc(e))}function tc(e){var n=op(e);if(n){var t=Bs(n);if(t&&(Om(t)||Pm(t)||Xm(t))){var r=Bd(n,t);return mp(e,Jd(r,e.mapper))}}return e}function rc(n){var t=63176704&n.flags?qs(n)||ke:n;return 32&e.getObjectFlags(t)?nc(t):2097152&t.flags?Ys(t):132&t.flags?Xe:296&t.flags?Ye:2112&t.flags?eu(k>=6):528&t.flags?Qe:12288&t.flags?Vl(k>=2):67108864&t.flags?ke:4194304&t.flags?De:t}function ac(n,t){for(var r,a,i=1048576&n.flags,o=i?24:0,s=i?0:16777216,c=4,l=0,u=0,d=n.types;u<d.length;u++){var p=d[u],m=rc(p);if(m!==oe){var f=sc(m,t),_=f?e.getDeclarationModifierFlagsFromSymbol(f):0;if(!f||_&o){if(i){var g=!Vo(t)&&(Uv(t)&&pc(m,1)||pc(m,0));g?(l|=g.isReadonly?8:0,a=e.append(a,Xm(m)?Ym(m)||ce:g.type)):l|=16}}else s&=f.flags,r=e.appendIfUnique(r,f),l|=(tE(f)?8:0)|(24&_?0:128)|(16&_?256:0)|(8&_?512:0)|(32&_?1024:0),Dy(f)||(c=2)}}if(r){if(1===r.length&&!(16&l)&&!a)return r[0];for(var v,y,h,b,E=[],T=!1,S=0,x=r;S<x.length;S++){f=x[S];b?f.valueDeclaration!==b&&(T=!0):b=f.valueDeclaration,v=e.addRange(v,f.declarations);m=Xi(f);y?m!==y&&(l|=32):(y=m,h=f.nameType),Hm(m)&&(l|=64),E.push(m)}e.addRange(E,a);var C=Ot(4|s,t,c|l);return C.containingType=n,!T&&b&&(C.valueDeclaration=b,b.symbol.parent&&(C.parent=b.symbol.parent)),C.declarations=v,C.nameType=h,C.type=i?Au(E):Ru(E),C}}function ic(n,t){var r=n.propertyCache||(n.propertyCache=e.createSymbolTable()),a=r.get(t);return a||(a=ac(n,t),a&&r.set(t,a)),a}function oc(n,t){var r=ic(n,t);return!r||16&e.getCheckFlags(r)?void 0:r}function sc(e,n){if(e=rc(e),524288&e.flags){var t=ws(e),r=t.members.get(n);if(r&&la(r))return r;var a=t===Oe?He:t.callSignatures.length?We:t.constructSignatures.length?qe:void 0;if(a){var i=Os(a,n);if(i)return i}return Os(Ue,n)}if(3145728&e.flags)return oc(e,n)}function cc(n,t){if(3670016&n.flags){var r=ws(n);return 0===t?r.callSignatures:r.constructSignatures}return e.emptyArray}function lc(e,n){return cc(rc(e),n)}function uc(e,n){if(3670016&e.flags){var t=ws(e);return 0===n?t.stringIndexInfo:t.numberIndexInfo}}function dc(e,n){var t=uc(e,n);return t&&t.type}function pc(e,n){return uc(rc(e),n)}function mc(e,n){return dc(rc(e),n)}function fc(e,n){if(uf(e)){for(var t=[],r=0,a=Rs(e);r<a.length;r++){var i=a[r];(0===n||Uv(i.escapedName))&&t.push(Xi(i))}if(t.length)return Au(t,2)}}function _c(n){for(var t,r=0,a=e.getEffectiveTypeParameterDeclarations(n);r<a.length;r++){var i=a[r];t=e.appendIfUnique(t,Co(i.symbol))}return t}function gc(e){var n=[];return e.forEach(function(e,t){va(t)||n.push(e)}),n}function vc(n){return e.isInJSFile(n)&&(n.type&&288===n.type.kind||e.getJSDocParameterTags(n).some(function(e){var n=e.isBracketed,t=e.typeExpression;return n||!!t&&288===t.type.kind}))}function yc(n,t){if(!e.isExternalModuleNameRelative(n)){var r=Xt(wn,'"'+n+'"',512);return r&&t?ta(r):r}}function hc(n){if(e.hasQuestionToken(n)||bc(n)||vc(n))return!0;if(n.initializer){var t=Cc(n.parent),r=n.parent.parameters.indexOf(n);return e.Debug.assert(r>=0),r>=kb(t)}var a=e.getImmediatelyInvokedFunctionExpression(n.parent);return!!a&&(!n.type&&!n.dotDotDotToken&&n.parent.parameters.indexOf(n)>=a.arguments.length)}function bc(n){if(!e.isJSDocParameterTag(n))return!1;var t=n.isBracketed,r=n.typeExpression;return t||!!r&&288===r.type.kind}function Ec(e,n,t){return{kind:1,parameterName:e,parameterIndex:n,type:t}}function Tc(e){return{kind:0,type:e}}function Sc(e){var n=0;if(e)for(var t=0;t<e.length;t++)ec(e[t])||(n=t+1);return n}function xc(n,t,r,a){var i=e.length(t);if(!i)return[];var o=e.length(n);if(a||o>=r&&o<=i){for(var s=n?n.slice():[],c=o;c<i;c++)s[c]=oe;var l=Qf(a);for(c=o;c<i;c++){var u=Zs(t[c]);a&&u&&Sp(u,ke)&&(u=re),s[c]=u?mp(u,Ud(t,s)):l}return s.length=t.length,s}return n&&n.slice()}function Cc(n){var t=zt(n);if(!t.resolvedSignature){for(var r=[],a=!1,i=0,o=void 0,s=!1,c=e.getImmediatelyInvokedFunctionExpression(n),l=e.isJSDocConstructSignature(n),u=!c&&e.isInJSFile(n)&&e.isValueSignatureDeclaration(n)&&!e.hasJSDocParameterTags(n)&&!e.getJSDocType(n),d=l?1:0;d<n.parameters.length;d++){var p=n.parameters[d],m=p.symbol,f=e.isJSDocParameterTag(p)?p.typeExpression&&p.typeExpression.type:p.type;if(m&&4&m.flags&&!e.isBindingPattern(p.name)){var _=Zt(p,m.escapedName,67220415,void 0,void 0,!1);m=_}0===d&&"this"===m.escapedName?(s=!0,o=p.symbol):r.push(m),f&&182===f.kind&&(a=!0);var g=bc(p)||p.initializer||p.questionToken||p.dotDotDotToken||c&&r.length>c.arguments.length&&!f||u||vc(p);g||(i=r.length)}if((158===n.kind||159===n.kind)&&!Ko(n)&&(!s||!o)){var v=158===n.kind?159:158,y=e.getDeclarationOfKind(ra(n),v);y&&(o=Gi(y))}var h=157===n.kind?vo(ta(n.parent.symbol)):void 0,b=h?h.localTypeParameters:_c(n),E=e.hasRestParameter(n)||e.isInJSFile(n)&&Ac(n,r);t.resolvedSignature=$o(n,b,o,r,void 0,void 0,i,E,a)}return t.resolvedSignature}function Ac(n,t){if(e.isJSDocSignature(n)||!kc(n))return!1;var r=e.lastOrUndefined(n.parameters),a=r?e.getJSDocParameterTags(r):e.getJSDocTags(n).filter(e.isJSDocParameterTag),i=e.firstDefined(a,function(n){return n.typeExpression&&e.isJSDocVariadicType(n.typeExpression.type)?n.typeExpression.type:void 0}),o=Ot(3,"args",16384);return o.type=i?su(Rd(i.type)):en,i&&t.pop(),t.push(o),!0}function Dc(n){var t=e.isInJSFile(n)?e.getJSDocTypeTag(n):void 0,r=t&&t.typeExpression&&gh(Rd(t.typeExpression));return r&&qc(r)}function Lc(e){var n=Dc(e);return n&&Fc(n)}function kc(n){var t=zt(n);return void 0===t.containsArgumentsReference&&(8192&t.flags?t.containsArgumentsReference=!0:t.containsArgumentsReference=r(n.body)),t.containsArgumentsReference;function r(n){if(!n)return!1;switch(n.kind){case 72:return"arguments"===n.escapedText&&e.isExpressionNode(n);case 154:case 156:case 158:case 159:return 149===n.name.kind&&r(n.name);default:return!e.nodeStartsNewLexicalEnvironment(n)&&!e.isPartOfTypeNode(n)&&!!e.forEachChild(n,r)}}}function Nc(n){if(!n)return e.emptyArray;for(var t=[],r=0;r<n.declarations.length;r++){var a=n.declarations[r];if(e.isFunctionLike(a)){if(r>0&&a.body){var i=n.declarations[r-1];if(a.parent===i.parent&&a.kind===i.kind&&a.pos===i.end)continue}t.push(Cc(a))}}return t}function Ic(e){var n=Vr(e,e);if(n){var t=Hr(n);if(t)return Xi(t)}return re}function wc(e){if(e.thisParameter)return Xi(e.thisParameter)}function Mc(e){return void 0!==Oc(e)}function Oc(n){if(!n.resolvedTypePredicate){if(n.target){var t=Oc(n.target);n.resolvedTypePredicate=t?ep(t,n.mapper):An}else if(n.unionSignatures)n.resolvedTypePredicate=Du(n.unionSignatures)||An;else{var r=n.declaration&&e.getEffectiveReturnTypeNode(n.declaration),a=void 0;if(!r&&e.isInJSFile(n.declaration)){var i=Dc(n.declaration);i&&n!==i&&(a=Oc(i))}n.resolvedTypePredicate=r&&e.isTypePredicateNode(r)?Pc(r,n.declaration):a||An}e.Debug.assert(!!n.resolvedTypePredicate)}return n.resolvedTypePredicate===An?void 0:n.resolvedTypePredicate}function Pc(e,n){var t=e.parameterName,r=Rd(e.type);return 72===t.kind?Ec(t.escapedText,Rc(n.parameters,t),r):Tc(r)}function Rc(e,n){for(var t=0;t<e.length;t++){var r=e[t];if(72===r.name.kind&&r.name.escapedText===n.escapedText)return t}return-1}function Fc(n){if(!n.resolvedReturnType){if(!Ya(n,3))return oe;var t=n.target?mp(Fc(n.target),n.mapper):n.unionSignatures?Au(e.map(n.unionSignatures,Fc),2):Gc(n.declaration)||ib(n.declaration)&&sb(ra(n.declaration))||(e.nodeIsMissing(n.declaration.body)?re:jb(n.declaration));if(!$a()){if(n.declaration){var r=e.getEffectiveReturnTypeNode(n.declaration);if(r)It(r,e.Diagnostics.Return_type_annotation_circularly_references_itself);else if(R){var a=n.declaration,i=e.getNameOfDeclaration(a);i?It(i,e.Diagnostics._0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions,e.declarationNameToString(i)):It(a,e.Diagnostics.Function_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions)}}t=re}n.resolvedReturnType=t}return n.resolvedReturnType}function Gc(n){if(157===n.kind)return vo(ta(n.parent.symbol));if(e.isJSDocConstructSignature(n))return Rd(n.parameters[0].type);var t=e.getEffectiveReturnTypeNode(n);if(t)return Rd(t);if(158===n.kind&&!Ko(n)){var r=e.isInJSFile(n)&&fi(n);if(r)return r;var a=e.getDeclarationOfKind(ra(n),159),i=Fi(a);if(i)return i}return Lc(n)}function Vc(e){return!e.resolvedReturnType&&Qa(e,3)>=0}function Bc(e){return Kc(e)||re}function Kc(e){if(e.hasRestParameter){var n=Xi(e.parameters[e.parameters.length-1]),t=Xm(n)?Ym(n):n;return t&&mc(t,1)}}function jc(e,n,t){return Uc(e,xc(n,e.typeParameters,Sc(e.typeParameters),t))}function Uc(n,t){var r=n.instantiations||(n.instantiations=e.createMap()),a=ol(t),i=r.get(a);return i||r.set(a,i=Hc(n,t)),i}function Hc(e,n){return np(e,Wc(e,n),!0)}function Wc(e,n){return Ud(e.typeParameters,n)}function qc(e){return e.typeParameters?e.erasedSignatureCache||(e.erasedSignatureCache=zc(e)):e}function zc(e){return np(e,Hd(e.typeParameters),!0)}function Jc(e){return e.typeParameters?e.canonicalSignatureCache||(e.canonicalSignatureCache=Xc(e)):e}function Xc(n){return jc(n,e.map(n.typeParameters,function(e){return e.target&&!Bs(e.target)?e.target:e}),e.isInJSFile(n.declaration))}function Yc(n){var t=n.typeParameters;if(t){var r=Hd(t),a=e.map(t,function(e){return mp(qs(e),r)||ke});return np(n,Ud(t,a),!0)}return n}function Qc(n){if(!n.isolatedSignatureType){var t=157===n.declaration.kind||161===n.declaration.kind,r=fa(16);r.members=A,r.properties=e.emptyArray,r.callSignatures=t?e.emptyArray:[n],r.constructSignatures=t?[n]:e.emptyArray,n.isolatedSignatureType=r}return n.isolatedSignatureType}function Zc(e){return e.members.get("__index")}function $c(n,t){var r=1===t?135:138,a=Zc(n);if(a)for(var i=0,o=a.declarations;i<o.length;i++){var s=o[i],c=e.cast(s,e.isIndexSignatureDeclaration);if(1===c.parameters.length){var l=c.parameters[0];if(l.type&&l.type.kind===r)return c}}}function el(e,n,t){return{type:e,isReadonly:n,declaration:t}}function nl(n,t){var r=$c(n,t);if(r)return el(r.type?Rd(r.type):re,e.hasModifier(r,64),r)}function tl(n){var t=n.symbol&&e.getDeclarationOfKind(n.symbol,150);return t&&e.getEffectiveConstraintOfTypeParameter(t)}function rl(n){var t;if(n.symbol)for(var r=0,a=n.symbol.declarations;r<a.length;r++){var i=a[r];if(176===i.parent.kind){var o=i.parent.parent;if(164===o.kind){var s=o,c=mT(s);if(c){var l=s.typeArguments.indexOf(i.parent);if(l<c.length){var u=Bs(c[l]);if(u){var d=Ud(c,dT(s,c)),p=mp(u,d);p!==n&&(t=e.append(t,p))}}}}else 151===o.kind&&o.dotDotDotToken&&(t=e.append(t,su(se)))}}return t&&Ru(t)}function al(e){if(!e.constraint)if(e.target){var n=Bs(e.target);e.constraint=n?mp(n,e.mapper):Pe}else{var t=tl(e);e.constraint=t?Rd(t):rl(e)||Pe}return e.constraint===Pe?void 0:e.constraint}function il(n){var t=e.getDeclarationOfKind(n.symbol,150),r=e.isJSDocTemplateTag(t.parent)?e.getHostSignatureFromJSDoc(t.parent):t.parent;return r&&ra(r)}function ol(e){var n="";if(e){var t=e.length,r=0;while(r<t){var a=e[r].id,i=1;while(r+i<t&&e[r+i].id===a+i)i++;n.length&&(n+=","),n+=a,i>1&&(n+=":"+i),r+=i}}return n}function sl(e,n){for(var t=0,r=0,a=e;r<a.length;r++){var i=a[r];i.flags&n||(t|=i.flags)}return 939524096&t}function cl(e,n){var t=ol(n),r=e.instantiations.get(t);return r||(r=fa(4,e.symbol),e.instantiations.set(t,r),r.flags|=n?sl(n,0):0,r.target=e,r.typeArguments=n),r}function ll(e){var n=da(e.flags);return n.symbol=e.symbol,n.objectFlags=e.objectFlags,n.target=e.target,n.typeArguments=e.typeArguments,n}function ul(n){return e.length(n.target.typeParameters)}function dl(n,t,r){var a=Do(ta(t)),i=a.localTypeParameters;if(i){var o=e.length(n.typeArguments),s=Sc(i),c=e.isInJSFile(n),l=!R&&c;if(!l&&(o<s||o>i.length)){var u=c&&e.isExpressionWithTypeArguments(n)&&!e.isJSDocAugmentsTag(n.parent),d=s===i.length?u?e.Diagnostics.Expected_0_type_arguments_provide_these_with_an_extends_tag:e.Diagnostics.Generic_type_0_requires_1_type_argument_s:u?e.Diagnostics.Expected_0_1_type_arguments_provide_these_with_an_extends_tag:e.Diagnostics.Generic_type_0_requires_between_1_and_2_type_arguments,p=Fa(a,void 0,2);if(It(n,d,p,s,i.length),!c)return oe}var m=e.concatenate(a.outerTypeParameters,xc(r,i,s,c));return cl(a,m)}return xl(n,t)?a:oe}function pl(n,t){var r=Do(n),a=qt(n),i=a.typeParameters,o=ol(t),s=a.instantiations.get(o);return s||a.instantiations.set(o,s=mp(r,Ud(i,xc(t,i,Sc(i),e.isInJSFile(n.valueDeclaration))))),s}function ml(n,t,r){var a=Do(t),i=qt(t).typeParameters;if(i){var o=e.length(n.typeArguments),s=Sc(i);return o<s||o>i.length?(It(n,s===i.length?e.Diagnostics.Generic_type_0_requires_1_type_argument_s:e.Diagnostics.Generic_type_0_requires_between_1_and_2_type_arguments,Pa(t),s,i.length),oe):pl(t,r)}return xl(n,t)?a:oe}function fl(n){switch(n.kind){case 164:return n.typeName;case 211:var t=n.expression;if(e.isEntityNameExpression(t))return t}}function _l(e,n){return e&&Pr(e,n)||ne}function gl(n,t){var r=Ll(n);if(t===ne)return oe;var a=yl(n,t,r);if(a)return a;var i=e.isInJSFile(n)&&t.valueDeclaration&&e.getJSDocEnumTag(t.valueDeclaration);if(i){var o=zt(i);if(!Ya(i,5))return oe;var s=i.typeExpression?Rd(i.typeExpression):oe;return $a()||(s=oe,It(n,e.Diagnostics.Enum_type_0_circularly_references_itself,Pa(t))),o.resolvedEnumType=s}var c=Lo(t);if(c)return xl(n,t)?262144&c.flags?Tl(c,n):Ld(c):oe;if(!(67220415&t.flags&&Sl(n)))return oe;var l=vl(n,t,r);return l||(_l(fl(n),67897832),Xi(t))}function vl(e,n,t){var r=Xi(n),a=r.symbol&&r.symbol!==n&&yl(e,r.symbol,t);if(a)return qt(n).resolvedJSDocType=a}function yl(n,t,r){if(96&t.flags){if(t.valueDeclaration&&e.isBinaryExpression(t.valueDeclaration.parent)){var a=vl(n,t,r);if(a)return a}return dl(n,t,r)}if(524288&t.flags)return ml(n,t,r);if(16&t.flags&&Sl(n)&&ib(t.valueDeclaration)){var i=ws(Xi(t));if(1===i.callSignatures.length)return Fc(i.callSignatures[0])}}function hl(e,n){var t=da(33554432);return t.typeVariable=e,t.substitute=n,t}function bl(e){return 170===e.kind&&1===e.elementTypes.length}function El(e,n,t){return bl(n)&&bl(t)?El(e,n.elementTypes[0],t.elementTypes[0]):sd(Rd(n))===e?Rd(t):void 0}function Tl(n,t){var r;while(t&&!e.isStatement(t)&&291!==t.kind){var a=t.parent;if(175===a.kind&&t===a.trueType){var i=El(n,a.checkType,a.extendsType);i&&(r=e.append(r,i))}t=a}return r?hl(n,Ru(e.append(r,n))):n}function Sl(e){return!!(2097152&e.flags)&&(164===e.kind||183===e.kind)}function xl(n,t){return!n.typeArguments||(It(n,e.Diagnostics.Type_0_is_not_generic,t?Pa(t):n.typeName?e.declarationNameToString(n.typeName):"(anonymous)"),!1)}function Cl(n){if(e.isIdentifier(n.typeName)){var t=n.typeArguments;switch(n.typeName.escapedText){case"String":return xl(n),pe;case"Number":return xl(n),me;case"Boolean":return xl(n),he;case"Void":return xl(n),Ee;case"Undefined":return xl(n),ce;case"Null":return xl(n),ue;case"Function":case"function":return xl(n),He;case"Array":case"array":return t&&t.length?void 0:en;case"Promise":case"promise":return t&&t.length?void 0:Vb(re);case"Object":if(t&&2===t.length){if(e.isJSDocIndexSignature(n)){var r=Rd(t[0]),a=Rd(t[1]),i=el(a,!1);return ba(void 0,A,e.emptyArray,e.emptyArray,r===pe?i:void 0,r===me?i:void 0)}return re}return xl(n),re}}}function Al(e){var n=Rd(e.type);return w?of(n,65536):n}function Dl(e){var n=zt(e);if(!n.resolvedType){var t=void 0,r=void 0,a=67897832;Sl(e)&&(r=Cl(e),a|=67220415),r||(t=_l(fl(e),a),r=gl(e,t)),n.resolvedSymbol=t,n.resolvedType=r}return n.resolvedType}function Ll(n){return e.map(n.typeArguments,Rd)}function kl(e){var n=zt(e);return n.resolvedType||(n.resolvedType=Ld(bf(qE(e.exprName)))),n.resolvedType}function Nl(n,t){function r(e){for(var n=e.declarations,t=0,r=n;t<r.length;t++){var a=r[t];switch(a.kind){case 240:case 241:case 243:return a}}}if(!n)return t?Me:ke;var a=Do(n);return 524288&a.flags?e.length(a.typeParameters)!==t?(It(r(n),e.Diagnostics.Global_type_0_must_have_1_type_parameter_s,e.symbolName(n),t),t?Me:ke):a:(It(r(n),e.Diagnostics.Global_type_0_must_be_a_class_or_interface_type,e.symbolName(n)),t?Me:ke)}function Il(n,t){return Ml(n,67220415,t?e.Diagnostics.Cannot_find_global_value_0:void 0)}function wl(n,t){return Ml(n,67897832,t?e.Diagnostics.Cannot_find_global_type_0:void 0)}function Ml(e,n,t){return Zt(void 0,e,n,t,e,!1)}function Ol(e,n,t){var r=wl(e,t);return r||t?Nl(r,n):void 0}function Pl(){return sn||(sn=Ol("TypedPropertyDescriptor",1,!0))||Me}function Rl(){return yn||(yn=Ol("TemplateStringsArray",0,!0))||ke}function Fl(){return hn||(hn=Ol("ImportMeta",0,!0))||ke}function Gl(e){return an||(an=Il("Symbol",e))}function Vl(e){return on||(on=Ol("Symbol",0,e))||ke}function Bl(e){return cn||(cn=Ol("Promise",1,e))||Me}function Kl(e){return ln||(ln=Ol("PromiseLike",1,e))||Me}function jl(e){return un||(un=Il("Promise",e))}function Ul(e){return dn||(dn=Ol("PromiseConstructorLike",0,e))||ke}function Hl(e){return _n||(_n=Ol("AsyncIterable",1,e))||Me}function Wl(e){return gn||(gn=Ol("AsyncIterator",1,e))||Me}function ql(e){return vn||(vn=Ol("AsyncIterableIterator",1,e))||Me}function zl(e){return pn||(pn=Ol("Iterable",1,e))||Me}function Jl(e){return mn||(mn=Ol("Iterator",1,e))||Me}function Xl(e){return fn||(fn=Ol("IterableIterator",1,e))||Me}function Yl(e,n){void 0===n&&(n=0);var t=Ml(e,67897832,void 0);return t&&Nl(t,n)}function Ql(){return bn||(bn=Ml("Extract",524288,e.Diagnostics.Cannot_find_global_type_0))}function Zl(){return En||(En=Ml("Exclude",524288,e.Diagnostics.Cannot_find_global_type_0))}function $l(){return Tn||(Tn=Ml("Pick",524288,e.Diagnostics.Cannot_find_global_type_0))}function eu(e){return Sn||(Sn=Ol("BigInt",0,e))||ke}function nu(e,n){return e!==Me?cl(e,n):ke}function tu(e){return nu(Pl(),[e])}function ru(e){return nu(Hl(!0),[e])}function au(e){return nu(ql(!0),[e])}function iu(e){return nu(zl(!0),[e])}function ou(e){return nu(Xl(!0),[e])}function su(e){return nu(ze,[e])}function cu(e){return nu(Je,[e])}function lu(e){var n=zt(e);return n.resolvedType||(n.resolvedType=su(Rd(e.elementType))),n.resolvedType}function uu(n,t,r,a){var i,o=[],s=r?n-1:n;if(n){i=new Array(n);for(var c=0;c<n;c++){var l=i[c]=ga();if(c<s){var u=Ot(4|(c>=t?16777216:0),""+c);u.type=l,o.push(u)}}}var d=[];for(c=t;c<=s;c++)d.push(Nd(c));var p=Ot(4,"length");p.type=r?me:Au(d),o.push(p);var m=fa(12);return m.typeParameters=i,m.outerTypeParameters=void 0,m.localTypeParameters=i,m.instantiations=e.createMap(),m.instantiations.set(ol(m.typeParameters),m),m.target=m,m.typeArguments=m.typeParameters,m.thisType=ga(),m.thisType.isThisType=!0,m.thisType.constraint=m,m.declaredProperties=o,m.declaredCallSignatures=e.emptyArray,m.declaredConstructSignatures=e.emptyArray,m.declaredStringIndexInfo=void 0,m.declaredNumberIndexInfo=void 0,m.minLength=t,m.hasRestElement=r,m.associatedNames=a,m}function du(e,n,t,r){var a=e+(t?"+":",")+n+(r&&r.length?","+r.join(","):""),i=J.get(a);return i||J.set(a,i=uu(e,n,t,r)),i}function pu(e,n,t,r){void 0===n&&(n=e.length),void 0===t&&(t=!1);var a=e.length;if(1===a&&t)return su(e[0]);var i=du(a,n,a>0&&t,r);return e.length?cl(i,e):i}function mu(n){var t=zt(n);if(!t.resolvedType){var r=e.lastOrUndefined(n.elementTypes),a=r&&172===r.kind?r:void 0,i=e.findLastIndex(n.elementTypes,function(e){return 171!==e.kind&&e!==a})+1,o=e.map(n.elementTypes,function(e){var n=Rd(e);return e===a&&mc(n,1)||n});t.resolvedType=pu(o,i,!!a)}return t.resolvedType}function fu(n,t){var r=n.target;return r.hasRestElement&&(t=Math.min(t,ul(n)-1)),pu((n.typeArguments||e.emptyArray).slice(t),Math.max(0,r.minLength-t),r.hasRestElement,r.associatedNames&&r.associatedNames.slice(t))}function _u(e){var n=Rd(e.type);return w?sf(n):n}function gu(e){return e.id}function vu(n,t){return e.binarySearch(n,t,gu,e.compareValues)>=0}function yu(n,t){var r=e.binarySearch(n,t,gu,e.compareValues);return r<0&&(n.splice(~r,0,t),!0)}function hu(e){for(var n=0,t=0,r=e.types;t<r.length;t++){var a=r[t];if(109440&a.flags&&109440&n)return!0;if(n|=a.flags,98304&n&&67633152&n||67108864&n&&130044&n||132&n&&67238776&n||296&n&&67238612&n||2112&n&&67236796&n||12288&n&&67226620&n||49152&n&&67189756&n)return!0}return!1}function bu(n,t,r){var a=r.flags;if(1048576&a)return Eu(n,t,r.types);if(!(131072&a||2097152&a&&hu(r)))if(t|=-939524097&a,r===ie&&(t|=268435456),!w&&98304&a)134217728&a||(t|=134217728);else{var i=n.length,o=i&&r.id>n[i-1].id?~i:e.binarySearch(n,r,gu,e.compareValues);o<0&&n.splice(~o,0,r)}return t}function Eu(e,n,t){for(var r=0,a=t;r<a.length;r++){var i=a[r];n=bu(e,n,i)}return n}function Tu(n,t){for(var r=0,a=t;r<a.length;r++){var i=a[r];if(n!==i&&Dp(n,i)&&(!(1&e.getObjectFlags(Qi(n)))||!(1&e.getObjectFlags(Qi(i)))||kp(n,i)))return!0}return!1}function Su(e){var n=e[0];if(1024&n.flags){for(var t=aa(n.symbol),r=1;r<e.length;r++){var a=e[r];if(!(1024&a.flags)||t!==aa(a.symbol))return!1}return!0}return!1}function xu(n){if(0!==n.length&&!Su(n)){var t=n.length;while(t>0)t--,Tu(n[t],n)&&e.orderedRemoveItemAt(n,t)}}function Cu(n,t){var r=n.length;while(r>0){r--;var a=n[r],i=128&a.flags&&4&t||256&a.flags&&8&t||2048&a.flags&&64&t||8192&a.flags&&4096&t||kd(a)&&vu(n,a.regularType);i&&e.orderedRemoveItemAt(n,r)}}function Au(e,n,t,r){if(void 0===n&&(n=1),0===e.length)return Te;if(1===e.length)return e[0];var a=[],i=Eu(a,0,e);if(0!==n){if(3&i)return 1&i?268435456&i?ie:re:se;switch(n){case 1:8576&i|512&&Cu(a,i);break;case 2:xu(a);break}if(0===a.length)return 65536&i?134217728&i?ue:de:32768&i?134217728&i?ce:le:Te}return ku(a,!(66994211&i),t,r)}function Du(n){for(var t,r=[],a=0,i=n;a<i.length;a++){var o=i[a],s=Oc(o);if(s){if(t){if(!Lu(t,s))return}else t=s;r.push(s.type)}}if(t){var c=Au(r);return e.isIdentifierTypePredicate(t)?Ec(t.parameterName,t.parameterIndex,c):Tc(c)}}function Lu(n,t){return e.isIdentifierTypePredicate(n)?e.isIdentifierTypePredicate(t)&&n.parameterIndex===t.parameterIndex:!e.isIdentifierTypePredicate(t)}function ku(e,n,t,r){if(0===e.length)return Te;if(1===e.length)return e[0];var a=ol(e),i=X.get(a);if(!i){var o=sl(e,98304);i=da(1048576|o),X.set(a,i),i.types=e,i.primitiveTypesOnly=n,i.aliasSymbol=t,i.aliasTypeArguments=r}return i}function Nu(n){var t=zt(n);if(!t.resolvedType){var r=hd(n);t.resolvedType=Au(e.map(n.types,Rd),1,r,bd(r))}return t.resolvedType}function Iu(n,t,r){var a=r.flags;return 2097152&a?wu(n,t,r.types):(nm(r)?536870912&t||(t|=536870912,n.push(r)):(t|=-939524097&a,3&a?r===ie&&(t|=268435456):!w&&98304&a||e.contains(n,r)||n.push(r)),t)}function wu(e,n,t){for(var r=0,a=t;r<a.length;r++){var i=a[r];n=Iu(e,n,Ld(i))}return n}function Mu(n,t){var r=n.length;while(r>0){r--;var a=n[r],i=4&a.flags&&128&t||8&a.flags&&256&t||64&a.flags&&2048&t||4096&a.flags&&8192&t;i&&e.orderedRemoveItemAt(n,r)}}function Ou(e,n){for(var t=0,r=e;t<r.length;t++){var a=r[t];if(!vu(a.types,n)){var i=128&n.flags?pe:256&n.flags?me:2048&n.flags?fe:8192&n.flags?be:void 0;if(!i||!vu(a.types,i))return!1}}return!0}function Pu(n){var t,r=e.findIndex(n,function(e){return!!(1048576&e.flags)&&e.primitiveTypesOnly});if(r<0)return!1;var a=r+1;while(a<n.length){var i=n[a];1048576&i.flags&&i.primitiveTypesOnly?((t||(t=[n[r]])).push(i),e.orderedRemoveItemAt(n,a)):a++}if(!t)return!1;for(var o=[],s=[],c=0,l=t;c<l.length;c++)for(var u=l[c],d=0,p=u.types;d<p.length;d++){i=p[d];yu(o,i)&&Ou(t,i)&&yu(s,i)}return n[r]=ku(s,!0),!0}function Ru(n,t,r){var a=[],i=wu(a,0,n);if(131072&i)return Te;if(1&i)return 268435456&i?ie:re;if(!w&&98304&i)return 32768&i?ce:ue;if((4&i&&128&i||8&i&&256&i||64&i&&2048&i||4096&i&&8192&i)&&Mu(a,i),536870912&i&&524288&i&&e.orderedRemoveItemAt(a,e.findIndex(a,nm)),0===a.length)return se;if(1===a.length)return a[0];if(1048576&i){if(Pu(a))return Ru(a,t,r);var o=e.findIndex(a,function(e){return 0!==(1048576&e.flags)}),s=a[o];return Au(e.map(s.types,function(n){return Ru(e.replaceElement(a,o,n))}),1,t,r)}var c=ol(a),l=Y.get(c);if(!l){var u=sl(a,98304);l=da(2097152|u),Y.set(c,l),l.types=a,l.aliasSymbol=t,l.aliasTypeArguments=r}return l}function Fu(n){var t=zt(n);if(!t.resolvedType){var r=hd(n);t.resolvedType=Ru(e.map(n.types,Rd),r,bd(r))}return t.resolvedType}function Gu(e,n){var t=da(4194304);return t.type=e,t.stringsOnly=n,t}function Vu(e,n){return n?e.resolvedStringIndexType||(e.resolvedStringIndexType=Gu(e,!0)):e.resolvedIndexType||(e.resolvedIndexType=Gu(e,!1))}function Bu(n){return e.isIdentifier(n)?Nd(e.unescapeLeadingUnderscores(n.escapedText)):Ld(e.isComputedPropertyName(n)?Hv(n):qE(n))}function Ku(n){return Nd({negative:!1,base10Value:e.parsePseudoBigInt(n.text)})}function ju(n,t){if(!(24&e.getDeclarationModifierFlagsFromSymbol(n))){var r=Jo(n).nameType;if(!r&&!e.isKnownSymbol(n))if("default"===n.escapedName)r=Nd("default");else{var a=n.valueDeclaration&&e.getNameOfDeclaration(n.valueDeclaration);r=a&&Bu(a)||Nd(e.symbolName(n))}if(r&&r.flags&t)return r}return Te}function Uu(n,t){return Au(e.map(Rs(n),function(e){return ju(e,t)}))}function Hu(e){var n=pc(e,1);return n!==In?n:void 0}function Wu(n,t){return void 0===t&&(t=G),1048576&n.flags?Ru(e.map(n.types,function(e){return Wu(e,t)})):2097152&n.flags?Au(e.map(n.types,function(e){return Wu(e,t)})):mE(n,58982400)?Vu(n,t):32&e.getObjectFlags(n)?Ts(n):n===ie?ie:1&n.flags?De:t?pc(n,0)?pe:Uu(n,128):pc(n,0)?Au([pe,me,Uu(n,8192)]):Hu(n)?Au([me,Uu(n,8320)]):Uu(n,8576)}function qu(e){if(G)return e;var n=Ql();return n?pl(n,[e,pe]):pe}function zu(e){var n=qu(Wu(e));return 131072&n.flags?pe:n}function Ju(n){var t=zt(n);if(!t.resolvedType)switch(n.operator){case 129:t.resolvedType=Wu(Rd(n.type));break;case 142:t.resolvedType=139===n.type.kind?Md(e.walkUpParenthesizedTypes(n.parent)):oe;break}return t.resolvedType}function Xu(e,n){var t=da(8388608);return t.objectType=e,t.indexType=n,t}function Yu(n){return!R&&(!!(16384&e.getObjectFlags(n))||(1048576&n.flags?e.every(n.types,Yu):2097152&n.flags?e.some(n.types,Yu):!!(63176704&n.flags)&&Yu(Xs(n))))}function Qu(n,t,r,a,i){var o=r&&190===r.kind?r:void 0,s=Fo(t)?Uo(t):o&&oh(o.argumentExpression,t,!1)?e.getPropertyNameForKnownSymbolName(e.idText(o.argumentExpression.name)):r&&e.isPropertyName(r)?e.getPropertyNameForPropertyNameNode(r):void 0;if(void 0!==s){var c=sc(n,s);if(c){if(o){if(Yy(c,o,100===o.expression.kind),e.isAssignmentTarget(o)&&(rE(o,c)||aE(o)))return It(o.argumentExpression,e.Diagnostics.Cannot_assign_to_0_because_it_is_a_read_only_property,Pa(c)),i;a&&(zt(r).resolvedSymbol=c)}var l=Xi(c);return o&&1!==e.getAssignmentTargetKind(o)?fg(o,l):l}if(z_(n,Xm)&&Uv(s)&&+s>=0){if(r&&z_(n,function(e){return!e.target.hasRestElement})){var u=Zu(r);Xm(n)?It(u,e.Diagnostics.Tuple_type_0_of_length_1_has_no_element_at_index_2,Fa(n),ul(n),e.unescapeLeadingUnderscores(s)):It(u,e.Diagnostics.Property_0_does_not_exist_on_type_1,e.unescapeLeadingUnderscores(s),Fa(n))}return X_(n,function(e){return Ym(e)||ce})}}if(!(98304&t.flags)&&fE(t,12716)){if(131073&n.flags)return n;var d=fE(t,296)&&pc(n,1)||pc(n,0)||void 0;if(d){if(r&&!fE(t,12)){u=Zu(r);It(u,e.Diagnostics.Type_0_cannot_be_used_as_an_index_type,Fa(t))}else o&&d.isReadonly&&(e.isAssignmentTarget(o)||e.isDeleteTarget(o))&&It(o,e.Diagnostics.Index_signature_in_type_0_only_permits_reading,Fa(n));return d.type}if(131072&t.flags)return Te;if(Yu(n))return re;if(o&&!gE(n)){if(R&&!L.suppressImplicitAnyIndexErrors)if(void 0!==s&&jy(s,n))It(o,e.Diagnostics.Property_0_is_a_static_member_of_type_1,s,Fa(n));else if(mc(n,1))It(o.argumentExpression,e.Diagnostics.Element_implicitly_has_an_any_type_because_index_expression_is_not_of_type_number);else{var p=void 0;void 0!==s&&(p=Hy(s,n))?void 0!==p&&It(o.argumentExpression,e.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2,s,Fa(n),p):It(o,e.Diagnostics.Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature,Fa(n))}return i}}if(Yu(n))return re;if(r){u=Zu(r);384&t.flags?It(u,e.Diagnostics.Property_0_does_not_exist_on_type_1,""+t.value,Fa(n)):12&t.flags?It(u,e.Diagnostics.Type_0_has_no_matching_index_signature_for_type_1,Fa(n),Fa(t)):It(u,e.Diagnostics.Type_0_cannot_be_used_as_an_index_type,Fa(t))}return ai(t)?t:i}function Zu(e){return 190===e.kind?e.argumentExpression:180===e.kind?e.indexType:149===e.kind?e.expression:e}function $u(e){return mE(e,193200128)}function ed(e){return mE(e,63176704)}function nd(e){return 8388608&e.flags?rd(e):e}function td(n,t){return 1048576&n.flags?X_(n,function(e){return nd(ad(e,t))}):2097152&n.flags?Ru(e.map(n.types,function(e){return nd(ad(e,t))})):void 0}function rd(e){if(e.simplified)return e.simplified===Re?e:e.simplified;e.simplified=Re;var n=nd(e.objectType),t=nd(e.indexType);if(1048576&t.flags)return e.simplified=X_(t,function(e){return nd(ad(n,e))});if(!(63176704&t.flags)){var r=td(n,t);if(r)return e.simplified=r}if(Is(n)){var a=Ud([Es(n)],[e.indexType]),i=Jd(n.mapper,a);return e.simplified=X_(mp(Ss(n),i),nd)}return e.simplified=e}function ad(e,n,t,r){if(void 0===r&&(r=t?oe:se),e===ie||n===ie)return ie;if(ed(n)||(!t||180===t.kind)&&$u(e)){if(3&e.flags)return e;var a=e.id+","+n.id,i=Z.get(a);return i||Z.set(a,i=Xu(e,n)),i}var o=rc(e);if(1048576&n.flags&&!(16&n.flags)){for(var s=[],c=!1,l=0,u=n.types;l<u.length;l++){var d=u[l],p=Qu(o,d,t,!1,r);if(p===r){if(!t)return r;c=!0}s.push(p)}return c?r:Au(s)}return Qu(o,n,t,!0,r)}function id(e){var n=zt(e);if(!n.resolvedType){var t=Rd(e.objectType),r=Rd(e.indexType),a=ad(t,r,e);n.resolvedType=8388608&a.flags&&a.objectType===t&&a.indexType===r?Tl(a,e):a}return n.resolvedType}function od(e){var n=zt(e);if(!n.resolvedType){var t=fa(32,e.symbol);t.declaration=e,t.aliasSymbol=hd(e),t.aliasTypeArguments=bd(t.aliasSymbol),n.resolvedType=t,Ts(t)}return n.resolvedType}function sd(e){return 33554432&e.flags?e.typeVariable:e}function cd(e,n){var t=mp(e.checkType,n),r=mp(e.extendsType,n);if(t===ie||r===ie)return ie;var a,i=mE(t,197394432);if(e.inferTypeParameters){var o=Af(e.inferTypeParameters,void 0,0);i||jf(o.inferences,t,r,96),a=Jd(n,o)}var s=a?mp(e.extendsType,a):r;if(!i&&!mE(s,197394432)){if(3&s.flags)return mp(e.trueType,n);if(1&t.flags)return Au([mp(e.trueType,a||n),mp(e.falseType,n)]);if(!Lp(_p(t),_p(s)))return mp(e.falseType,n);if(Lp(gp(t),gp(s)))return mp(e.trueType,a||n)}var c=sd(t),l=da(16777216);return l.root=e,l.checkType=c,l.extendsType=r,l.mapper=n,l.combinedMapper=a,l.aliasSymbol=e.aliasSymbol,l.aliasTypeArguments=Gd(e.aliasTypeArguments,n),l}function ld(e){return e.resolvedTrueType||(e.resolvedTrueType=mp(e.root.trueType,e.mapper))}function ud(e){return e.resolvedFalseType||(e.resolvedFalseType=mp(e.root.falseType,e.mapper))}function dd(n){var t;return n.locals&&n.locals.forEach(function(n){262144&n.flags&&(t=e.append(t,Do(n)))}),t}function pd(e,n){if(ip(e,n))return!0;while(n){if(175===n.kind&&ip(e,n.extendsType))return!0;n=n.parent}return!1}function md(n){var t=zt(n);if(!t.resolvedType){var r=Rd(n.checkType),a=hd(n),i=bd(a),o=eo(n,!0),s=i?o:e.filter(o,function(e){return pd(e,n)}),c={node:n,checkType:r,extendsType:Rd(n.extendsType),trueType:Rd(n.trueType),falseType:Rd(n.falseType),isDistributive:!!(262144&r.flags),inferTypeParameters:dd(n),outerTypeParameters:s,instantiations:void 0,aliasSymbol:a,aliasTypeArguments:i};t.resolvedType=cd(c,void 0),s&&(c.instantiations=e.createMap(),c.instantiations.set(ol(s),t.resolvedType))}return t.resolvedType}function fd(e){var n=zt(e);return n.resolvedType||(n.resolvedType=Co(ra(e.typeParameter))),n.resolvedType}function _d(n){return e.isIdentifier(n)?[n]:e.append(_d(n.left),n.right)}function gd(n){var t=zt(n);if(!t.resolvedType){if(n.isTypeOf&&n.typeArguments)return It(n,e.Diagnostics.Type_arguments_cannot_be_used_here),t.resolvedSymbol=ne,t.resolvedType=oe;if(!e.isLiteralImportTypeNode(n))return It(n.argument,e.Diagnostics.String_literal_expected),t.resolvedSymbol=ne,t.resolvedType=oe;var r=n.isTypeOf?67220415:2097152&n.flags?68008959:67897832,a=Vr(n,n.argument.literal);if(!a)return t.resolvedSymbol=ne,t.resolvedType=oe;var i=Hr(a,!1);if(e.nodeIsMissing(n.qualifier))if(i.flags&r)vd(n,t,i,r);else{var o=67220415===r?e.Diagnostics.Module_0_does_not_refer_to_a_value_but_is_used_as_a_value_here:e.Diagnostics.Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0;It(n,o,n.argument.literal.text),t.resolvedSymbol=ne,t.resolvedType=oe}else{var s=_d(n.qualifier),c=i,l=void 0;while(l=s.shift()){var u=s.length?1920:r,d=Xt(Zr(ta(kr(c))),l.escapedText,u);if(!d)return It(l,e.Diagnostics.Namespace_0_has_no_exported_member_1,Or(c),e.declarationNameToString(l)),t.resolvedType=oe;zt(l).resolvedSymbol=d,zt(l.parent).resolvedSymbol=d,c=d}vd(n,t,c,r)}}return t.resolvedType}function vd(e,n,t,r){var a=kr(t);return n.resolvedSymbol=a,n.resolvedType=67220415===r?Xi(t):gl(e,a)}function yd(n){var t=zt(n);if(!t.resolvedType){var r=hd(n);if(0!==zo(n.symbol).size||r){var a=fa(16,n.symbol);a.aliasSymbol=r,a.aliasTypeArguments=bd(r),e.isJSDocTypeLiteral(n)&&n.isArrayType&&(a=su(a)),t.resolvedType=a}else t.resolvedType=we}return t.resolvedType}function hd(n){return e.isTypeAlias(n.parent)?ra(n.parent):void 0}function bd(e){return e?to(e):void 0}function Ed(e){return!!(524288&e.flags)&&!Is(e)}function Td(n,t,r,a,i){if(1&n.flags||1&t.flags)return re;if(2&n.flags||2&t.flags)return se;if(131072&n.flags)return t;if(131072&t.flags)return n;if(1048576&n.flags)return X_(n,function(e){return Td(e,t,r,a,i)});if(1048576&t.flags)return X_(t,function(e){return Td(n,e,r,a,i)});if(71307260&t.flags)return n;if($u(n)||$u(t)){if(em(n))return t;if(2097152&n.flags){var o=n.types,s=o[o.length-1];if(Ed(s)&&Ed(t))return Ru(e.concatenate(o.slice(0,o.length-1),[Td(s,t,r,a,i)]))}return Ru([n,t])}var c,l,u=e.createSymbolTable(),d=e.createUnderscoreEscapedMap();n===ke?(c=pc(t,0),l=pc(t,1)):(c=ms(pc(n,0),pc(t,0)),l=ms(pc(n,1),pc(t,1)));for(var p=0,m=Rs(t);p<m.length;p++){var f=m[p];24&e.getDeclarationModifierFlagsFromSymbol(f)?d.set(f.escapedName,!0):Sd(f)&&u.set(f.escapedName,xd(f))}for(var _=0,g=Rs(n);_<g.length;_++){var v=g[_];if(!d.has(v.escapedName)&&Sd(v))if(u.has(v.escapedName)){f=u.get(v.escapedName);var y=Xi(f);if(16777216&f.flags){var h=e.concatenate(v.declarations,f.declarations),b=4|16777216&v.flags,E=Ot(b,v.escapedName);E.type=Au([Xi(v),b_(y,524288)]),E.leftSpread=v,E.rightSpread=f,E.declarations=h,E.nameType=v.nameType,u.set(v.escapedName,E)}}else u.set(v.escapedName,xd(v))}var T=ba(r,u,e.emptyArray,e.emptyArray,Cd(c),Cd(l));return T.flags|=268435456|a,T.objectFlags|=1152|i,T}function Sd(n){return!(106496&n.flags)||!n.declarations.some(function(n){return e.isClassLike(n.parent)})}function xd(e){var n=tE(e),t=65536&e.flags&&!(32768&e.flags);if(!n&&!t)return e;var r=4|16777216&e.flags,a=Ot(r,e.escapedName);return a.type=t?ce:Xi(e),a.declarations=e.declarations,a.nameType=e.nameType,a.syntheticOrigin=e,a}function Cd(e){return e&&e.isReadonly?el(e.type,!1,e.declaration):e}function Ad(e,n,t){var r=da(e);return r.symbol=t,r.value=n,r}function Dd(e){if(2944&e.flags){if(!e.freshType){var n=Ad(e.flags,e.value,e.symbol);n.regularType=e,n.freshType=n,e.freshType=n}return e.freshType}return e}function Ld(n){return 2944&n.flags?n.regularType:1048576&n.flags?Au(e.sameMap(n.types,Ld)):n}function kd(e){return!!(2944&e.flags)&&e.freshType===e}function Nd(n,t,r){var a="number"===typeof n?"#":"string"===typeof n?"@":"n",i=(t||"")+a+("object"===typeof n?e.pseudoBigIntToString(n):n),o=Q.get(i);if(!o){var s=("number"===typeof n?256:"string"===typeof n?128:2048)|(t?1024:0);Q.set(i,o=Ad(s,n,r)),o.regularType=o}return o}function Id(e){var n=zt(e);return n.resolvedType||(n.resolvedType=Ld(qE(e.literal))),n.resolvedType}function wd(e){var n=da(8192);return n.symbol=e,n.escapedName="__@"+n.symbol.escapedName+"@"+u(n.symbol),n}function Md(n){if(e.isValidESSymbolDeclaration(n)){var t=ra(n),r=qt(t);return r.uniqueESSymbolType||(r.uniqueESSymbolType=wd(t))}return be}function Od(n){var t=e.getThisContainer(n,!1),r=t&&t.parent;return!r||!e.isClassLike(r)&&241!==r.kind||e.hasModifier(t,32)||157===t.kind&&!e.isNodeDescendantOf(n,t.body)?(It(n,e.Diagnostics.A_this_type_is_available_only_in_a_non_static_member_of_a_class_or_interface),oe):vo(ra(r)).thisType}function Pd(e){var n=zt(e);return n.resolvedType||(n.resolvedType=Od(e)),n.resolvedType}function Rd(e){switch(e.kind){case 120:case 284:case 285:return re;case 143:return se;case 138:return pe;case 135:return me;case 146:return fe;case 123:return he;case 139:return be;case 106:return Ee;case 141:return ce;case 96:return ue;case 132:return Te;case 136:return 65536&e.flags?re:Ce;case 178:case 100:return Pd(e);case 182:return Id(e);case 164:return Dl(e);case 163:return he;case 211:return Dl(e);case 167:return kl(e);case 169:return lu(e);case 170:return mu(e);case 171:return _u(e);case 173:return Nu(e);case 174:return Fu(e);case 286:return Al(e);case 288:return vi(Rd(e.type));case 177:case 172:case 287:case 283:return Rd(e.type);case 290:return Xx(e);case 165:case 166:case 168:case 292:case 289:case 293:return yd(e);case 179:return Ju(e);case 180:return id(e);case 181:return od(e);case 175:return md(e);case 176:return fd(e);case 183:return gd(e);case 72:case 148:var n=hC(e);return n?Do(n):oe;default:return oe}}function Fd(e,n,t){if(e&&e.length)for(var r=0;r<e.length;r++){var a=e[r],i=t(a,n);if(a!==i){var o=0===r?[]:e.slice(0,r);for(o.push(i),r++;r<e.length;r++)o.push(t(e[r],n));return o}}return e}function Gd(e,n){return Fd(e,n,mp)}function Vd(e,n){return Fd(e,n,np)}function Bd(e,n){return function(t){return t===e?n:t}}function Kd(e,n,t,r){return function(a){return a===e?n:a===t?r:a}}function jd(e,n){return function(t){for(var r=0;r<e.length;r++)if(t===e[r])return n?n[r]:re;return t}}function Ud(n,t){return e.Debug.assert(void 0===t||n.length===t.length),1===n.length?Bd(n[0],t?t[0]:re):2===n.length?Kd(n[0],t?t[0]:re,n[1],t?t[1]:re):jd(n,t)}function Hd(e){return Ud(e,void 0)}function Wd(e,n){return function(t){return e.indexOf(t)>=n?ke:t}}function qd(e){return!!e.typeParameters}function zd(e){return e&&qd(e)?Af(e.typeParameters,e.signature,1|e.flags,e.compareTypes,e.inferences):e}function Jd(e,n){return e?n?function(t){return mp(e(t),n)}:e:n}function Xd(e,n,t){return function(r){return r===e?n:t(r)}}function Yd(e){return 262144&e.flags?ie:e}function Qd(e){return e.constraint===se?e:e.restrictiveInstantiation||(e.restrictiveInstantiation=ga(e.symbol),e.restrictiveInstantiation.constraint=se,e.restrictiveInstantiation)}function Zd(e){return 262144&e.flags?Qd(e):e}function $d(e){var n=ga(e.symbol);return n.target=e,n}function ep(n,t){return e.isIdentifierTypePredicate(n)?{kind:1,parameterName:n.parameterName,parameterIndex:n.parameterIndex,type:mp(n.type,t)}:{kind:0,type:mp(n.type,t)}}function np(n,t,r){var a;if(n.typeParameters&&!r){a=e.map(n.typeParameters,$d),t=Jd(Ud(n.typeParameters,a),t);for(var i=0,o=a;i<o.length;i++){var s=o[i];s.mapper=t}}var c=$o(n.declaration,a,n.thisParameter&&tp(n.thisParameter,t),Fd(n.parameters,t,tp),void 0,void 0,n.minArgumentCount,n.hasRestParameter,n.hasLiteralTypes);return c.target=n,c.mapper=t,c}function tp(n,t){var r=qt(n);if(r.type&&!mE(r.type,63700992))return n;1&e.getCheckFlags(n)&&(n=r.target,t=Jd(r.mapper,t));var a=Ot(n.flags,n.escapedName,1|26624&e.getCheckFlags(n));return a.declarations=n.declarations,a.parent=n.parent,a.target=n,a.mapper=t,n.valueDeclaration&&(a.valueDeclaration=n.valueDeclaration),n.nameType&&(a.nameType=n.nameType),a}function rp(n,t){var r=64&n.objectFlags?n.target:n,a=r.symbol,i=qt(a),o=i.outerTypeParameters;if(!o){var s=a.declarations[0];if(e.isInJSFile(s)){var c=e.findAncestor(s,e.isJSDocParameterTag);if(c){var l=e.getParameterSymbolFromJSDoc(c);l&&(s=l.valueDeclaration)}}var u=eo(s,!0);if(ib(s)){var d=_c(s);u=e.addRange(u,d)}o=u||e.emptyArray,o=2048&a.flags&&!r.aliasTypeArguments?e.filter(o,function(e){return ip(e,s)}):o,i.outerTypeParameters=o,o.length&&(i.instantiations=e.createMap(),i.instantiations.set(ol(o),r))}if(o.length){var p=64&n.objectFlags?Jd(n.mapper,t):t,m=e.map(o,p),f=ol(m),_=i.instantiations.get(f);if(!_){var g=Ud(o,m);_=32&r.objectFlags?sp(r,g):up(r,g),i.instantiations.set(f,_)}return _}return n}function ap(e){return!(148===e.kind||164===e.parent.kind&&e.parent.typeArguments&&e===e.parent.typeName)}function ip(n,t){if(n.symbol&&n.symbol.declarations&&1===n.symbol.declarations.length){var r=n.symbol.declarations[0].parent;if(e.findAncestor(t,function(e){return 218===e.kind?"quit":e===r}))return!!e.forEachChild(t,a)}return!0;function a(t){switch(t.kind){case 178:return!!n.isThisType;case 72:return!n.isThisType&&e.isPartOfTypeNode(t)&&ap(t)&&Rd(t)===n;case 167:return!0}return!!e.forEachChild(t,a)}}function op(e){var n=Ts(e);if(4194304&n.flags){var t=n.type;if(262144&t.flags)return t}}function sp(e,n){var t=op(e);if(t){var r=mp(t,n);if(t!==r){if(e.instantiating)return oe;e.instantiating=!0;var a=X_(r,function(r){if(61603843&r.flags&&r!==ie){var a=Xd(t,r,n);return Om(r)?su(lp(e,me,!0,a)):Pm(r)?cu(lp(e,me,!0,a)):Xm(r)?cp(r,e,a):up(e,a)}return r});return e.instantiating=!1,a}}return up(e,n)}function cp(n,t,r){var a=n.target.minLength,i=e.map(n.typeArguments||e.emptyArray,function(e,n){return lp(t,Nd(""+n),n>=a,r)}),o=Ds(t),s=4&o?0:8&o?ul(n)-(n.target.hasRestElement?1:0):a;return pu(i,s,n.target.hasRestElement,n.target.associatedNames)}function lp(e,n,t,r){var a=Jd(r,Ud([Es(e)],[n])),i=mp(Ss(e.target||e),a),o=Ds(e);return w&&4&o&&!Lp(ce,i)?sf(i):w&&8&o&&t?b_(i,524288):i}function up(e,n){var t=fa(64|e.objectFlags,e.symbol);if(32&e.objectFlags){t.declaration=e.declaration;var r=Es(e),a=$d(r);t.typeParameter=a,n=Jd(Bd(r,a),n),a.mapper=n}return t.target=e,t.mapper=n,t.aliasSymbol=e.aliasSymbol,t.aliasTypeArguments=Gd(e.aliasTypeArguments,n),t}function dp(n,t){var r=n.root;if(r.outerTypeParameters){var a=e.map(r.outerTypeParameters,t),i=ol(a),o=r.instantiations.get(i);if(!o){var s=Ud(r.outerTypeParameters,a);o=pp(r,s),r.instantiations.set(i,o)}return o}return n}function pp(e,n){if(e.isDistributive){var t=e.checkType,r=n(t);if(t!==r&&1179648&r.flags)return X_(r,function(r){return cd(e,Xd(t,r,n))})}return cd(e,n)}function mp(e,n){if(!e||!n||n===D)return e;if(50===x)return oe;x++;var t=fp(e,n);return x--,t}function fp(e,n){var t=e.flags;if(262144&t)return n(e);if(524288&t){var r=e.objectFlags;if(16&r)return e.symbol&&14384&e.symbol.flags&&e.symbol.declarations?rp(e,n):e;if(32&r)return rp(e,n);if(4&r){var a=e.typeArguments,i=Gd(a,n);return i!==a?cl(e.target,i):e}return e}if(1048576&t&&!(131068&t)){var o=e.types,s=Gd(o,n);return s!==o?Au(s,1,e.aliasSymbol,Gd(e.aliasTypeArguments,n)):e}if(2097152&t){o=e.types,s=Gd(o,n);return s!==o?Ru(s,e.aliasSymbol,Gd(e.aliasTypeArguments,n)):e}return 4194304&t?Wu(mp(e.type,n)):8388608&t?ad(mp(e.objectType,n),mp(e.indexType,n)):16777216&t?dp(e,Jd(e.mapper,n)):33554432&t?mp(e.typeVariable,n):e}function _p(e){return 262143&e.flags?e:e.permissiveInstantiation||(e.permissiveInstantiation=mp(e,Yd))}function gp(e){return 262143&e.flags?e:e.restrictiveInstantiation||(e.restrictiveInstantiation=mp(e,Zd))}function vp(e,n){return e&&el(mp(e.type,n),e.isReadonly,e.declaration)}function yp(n){switch(e.Debug.assert(156!==n.kind||e.isObjectLiteralMethod(n)),n.kind){case 196:case 197:case 156:return hp(n);case 188:return e.some(n.properties,yp);case 187:return e.some(n.elements,yp);case 205:return yp(n.whenTrue)||yp(n.whenFalse);case 204:return 55===n.operatorToken.kind&&(yp(n.left)||yp(n.right));case 275:return yp(n.initializer);case 195:return yp(n.expression);case 268:return e.some(n.properties,yp)||e.isJsxOpeningElement(n.parent)&&e.some(n.parent.parent.children,yp);case 267:var t=n.initializer;return!!t&&yp(t);case 270:var r=n.expression;return!!r&&yp(r)}return!1}function hp(n){if(n.typeParameters)return!1;if(e.some(n.parameters,function(n){return!e.getEffectiveTypeAnnotationNode(n)}))return!0;if(197!==n.kind){var t=e.firstOrUndefined(n.parameters);if(!t||!e.parameterIsThisKeyword(t))return!0}return bp(n)}function bp(e){return!!e.body&&218!==e.body.kind&&yp(e.body)}function Ep(n){return(e.isInJSFile(n)&&e.isFunctionDeclaration(n)||Iv(n)||e.isObjectLiteralMethod(n))&&hp(n)}function Tp(n){if(524288&n.flags){var t=ws(n);if(t.constructSignatures.length||t.callSignatures.length){var r=fa(16,n.symbol);return r.members=t.members,r.properties=t.properties,r.callSignatures=e.emptyArray,r.constructSignatures=e.emptyArray,r}}else if(2097152&n.flags)return Ru(e.map(n.types,Tp));return n}function Sp(e,n){return am(e,n,St)}function xp(e,n){return am(e,n,St)?-1:0}function Cp(e,n){return am(e,n,Et)?-1:0}function Ap(e,n){return am(e,n,bt)?-1:0}function Dp(e,n){return am(e,n,bt)}function Lp(e,n){return am(e,n,Et)}function kp(n,t){return 1048576&n.flags?e.every(n.types,function(e){return kp(e,t)}):1048576&t.flags?e.some(t.types,function(e){return kp(n,e)}):58982400&n.flags?kp(qs(n)||ke,t):t===Ue?!!(67633152&n.flags):t===He?!!(524288&n.flags)&&y_(n):Zi(n,Qi(t))}function Np(e,n){return am(e,n,Tt)}function Ip(e,n){return Np(e,n)||Np(n,e)}function wp(e,n,t,r,a,i){return om(e,n,Et,t,r,a,i)}function Mp(e,n,t,r,a,i){return Op(e,n,Et,t,r,a,i)}function Op(e,n,t,r,a,i,o){return!!am(e,n,t)||(!r||!Rp(a,e,n,t,i))&&om(e,n,t,r,i,o)}function Pp(n){return!!(16777216&n.flags||2097152&n.flags&&e.some(n.types,Pp))}function Rp(e,n,t,r,a){if(!e||Pp(t))return!1;if(!om(n,t,r,void 0)&&Fp(e,n,t,r,a))return!0;switch(e.kind){case 270:case 195:return Rp(e.expression,n,t,r,a);case 204:switch(e.operatorToken.kind){case 59:case 27:return Rp(e.right,n,t,r,a)}break;case 188:return zp(e,n,t,r);case 187:return Wp(e,n,t,r);case 268:return Up(e,n,t,r);case 197:return Gp(e,n,t,r)}return!1}function Fp(n,t,r,a,i){for(var o=lc(t,0),s=lc(t,1),c=0,l=[s,o];c<l.length;c++){var u=l[c];if(e.some(u,function(e){var n=Fc(e);return!(131073&n.flags)&&om(n,r,a,void 0)})){var d={};wp(t,r,n,i,void 0,d);var p=d.error;return Nt(p,e.createDiagnosticForNode(n,u===s?e.Diagnostics.Did_you_mean_to_use_new_with_this_expression:e.Diagnostics.Did_you_mean_to_call_this_expression)),!0}}return!1}function Gp(n,t,r,a){if(e.isBlock(n.body))return!1;if(e.some(n.parameters,e.hasType))return!1;var i=gh(t);if(!i)return!1;var o=lc(r,0);if(!e.length(o))return!1;var s=n.body,c=Fc(i),l=Au(e.map(o,Fc));if(!om(c,l,a,void 0)){var u=s&&Rp(s,c,l,a,void 0);if(u)return u;var d={};if(om(c,l,a,s,void 0,void 0,d),d.error)return r.symbol&&e.length(r.symbol.declarations)&&Nt(d.error,e.createDiagnosticForNode(r.symbol.declarations[0],e.Diagnostics.The_expected_type_comes_from_the_return_type_of_this_signature)),!0}return!1}function Vp(n,t,r,a){for(var i=!1,o=n.next();!o.done;o=n.next()){var s=o.value,c=s.errorNode,l=s.innerExpression,u=s.nameType,d=s.errorMessage,p=ad(r,u,void 0,oe);if(!(p===oe||8388608&p.flags)){var m=ad(t,u,void 0,oe);if(m!==oe&&p!==oe&&!om(m,p,a,void 0)){var f=l&&Rp(l,m,p,a,void 0);if(f)i=!0;else{var _={},g=l?BE(l,0,m):m,v=om(g,p,a,c,d,void 0,_);if(v&&g!==m&&om(m,p,a,c,d,void 0,_),_.error){var y=_.error,h=Fo(u)?Uo(u):void 0,b=void 0!==h?sc(r,h):void 0,E=!1;if(!b){var T=fE(u,296)&&pc(r,1)||pc(r,0)||void 0;T&&T.declaration&&!e.getSourceFileOfNode(T.declaration).hasNoDefaultLib&&(E=!0,Nt(y,e.createDiagnosticForNode(T.declaration,e.Diagnostics.The_expected_type_comes_from_this_index_signature)))}if(!E&&(b&&e.length(b.declarations)||r.symbol&&e.length(r.symbol.declarations))){var S=b&&e.length(b.declarations)?b.declarations[0]:r.symbol.declarations[0];e.getSourceFileOfNode(S).hasNoDefaultLib||Nt(y,e.createDiagnosticForNode(S,e.Diagnostics.The_expected_type_comes_from_property_0_which_is_declared_here_on_type_1,!h||8192&u.flags?Fa(u):e.unescapeLeadingUnderscores(h),Fa(r)))}}i=!0}}}}return i}function Bp(n){var t,r,a;return s(this,function(i){switch(i.label){case 0:if(!e.length(n.properties))return[2];t=0,r=n.properties,i.label=1;case 1:return t<r.length?(a=r[t],e.isJsxSpreadAttribute(a)?[3,3]:[4,{errorNode:a.name,innerExpression:a.initializer,nameType:Nd(e.idText(a.name))}]):[3,4];case 2:i.sent(),i.label=3;case 3:return t++,[3,1];case 4:return[2]}})}function Kp(n,t){var r,a,i,o,c;return s(this,function(s){switch(s.label){case 0:if(!e.length(n.children))return[2];r=0,a=0,s.label=1;case 1:return a<n.children.length?(i=n.children[a],o=Nd(a-r),c=jp(i,o,t),c?[4,c]:[3,3]):[3,5];case 2:return s.sent(),[3,4];case 3:r++,s.label=4;case 4:return a++,[3,1];case 5:return[2]}})}function jp(n,t,r){switch(n.kind){case 270:return{errorNode:n,innerExpression:n.expression,nameType:t};case 11:if(n.containsOnlyWhiteSpaces)break;return{errorNode:n,innerExpression:void 0,nameType:t,errorMessage:r()};case 260:case 261:case 264:return{errorNode:n,innerExpression:n,nameType:t};default:return e.Debug.assertNever(n,"Found invalid jsx child")}}function Up(n,t,r,a){var i,c=Vp(Bp(n),t,r,a);if(e.isJsxOpeningElement(n.parent)&&e.isJsxElement(n.parent.parent)){var l=n.parent.parent,u=py(cy(n)),d=void 0===u?"children":e.unescapeLeadingUnderscores(u),p=Nd(d),m=ad(r,p),f=e.filter(l.children,function(n){return!e.isJsxText(n)||!n.containsOnlyWhiteSpaces});if(!e.length(f))return c;var _=e.length(f)>1,g=J_(m,Bm),v=J_(m,function(e){return!Bm(e)});if(_)if(g!==Te){var y=pu(ay(l,0));c=Vp(Kp(l,E),y,g,a)||c}else am(ad(t,p),m,a)||(c=!0,It(l.openingElement.tagName,e.Diagnostics.This_JSX_tag_s_0_prop_expects_a_single_child_of_type_1_but_multiple_children_were_provided,d,Fa(m)));else if(v!==Te){var h=f[0],b=jp(h,p,E);b&&(c=Vp(function(){return s(this,function(e){switch(e.label){case 0:return[4,b];case 1:return e.sent(),[2]}})}(),t,r,a)||c)}else am(ad(t,p),m,a)||(c=!0,It(l.openingElement.tagName,e.Diagnostics.This_JSX_tag_s_0_prop_expects_type_1_which_requires_multiple_children_but_only_a_single_child_was_provided,d,Fa(m)))}return c;function E(){if(!i){var t=e.getTextOfNode(n.parent.tagName),a=py(cy(n)),s=void 0===a?"children":e.unescapeLeadingUnderscores(a),c=ad(r,Nd(s)),l=e.Diagnostics._0_components_don_t_accept_text_as_child_elements_Text_in_JSX_has_the_type_string_but_the_expected_type_of_1_is_2;i=o({},l,{key:"!!ALREADY FORMATTED!!",message:e.formatMessage(void 0,l,t,s,Fa(c))})}return i}}function Hp(n,t){var r,a,i,o;return s(this,function(s){switch(s.label){case 0:if(r=e.length(n.elements),!r)return[2];a=0,s.label=1;case 1:return a<r?Vm(t)&&!sc(t,""+a)?[3,3]:(i=n.elements[a],e.isOmittedExpression(i)?[3,3]:(o=Nd(a),[4,{errorNode:i,innerExpression:i,nameType:o}])):[3,4];case 2:s.sent(),s.label=3;case 3:return a++,[3,1];case 4:return[2]}})}function Wp(e,n,t,r){if(Vm(n))return Vp(Hp(e,t),n,t,r);var a=Fv(e,3,!0);return!!Vm(a)&&Vp(Hp(e,t),a,t,r)}function qp(n){var t,r,a,i,o;return s(this,function(s){switch(s.label){case 0:if(!e.length(n.properties))return[2];t=0,r=n.properties,s.label=1;case 1:if(!(t<r.length))return[3,8];if(a=r[t],e.isSpreadAssignment(a))return[3,7];if(i=ju(ra(a),8576),!i||131072&i.flags)return[3,7];switch(o=a.kind,o){case 159:return[3,2];case 158:return[3,2];case 156:return[3,2];case 276:return[3,2];case 275:return[3,4]}return[3,6];case 2:return[4,{errorNode:a.name,innerExpression:void 0,nameType:i}];case 3:return s.sent(),[3,7];case 4:return[4,{errorNode:a.name,innerExpression:a.initializer,nameType:i,errorMessage:oi(a.name)?e.Diagnostics.Type_of_computed_property_s_value_is_0_which_is_not_assignable_to_type_1:void 0}];case 5:return s.sent(),[3,7];case 6:e.Debug.assertNever(a),s.label=7;case 7:return t++,[3,1];case 8:return[2]}})}function zp(e,n,t,r){return Vp(qp(e),n,t,r)}function Jp(e,n,t,r,a){return om(e,n,Tt,t,r,a)}function Xp(e,n,t){return 0!==Yp(e,n,0,t,!1,void 0,Cp)}function Yp(n,t,r,a,i,o,s){if(n===t)return-1;var c=Lb(t);if(!Nb(t)&&kb(n)>c)return 0;n.typeParameters&&n.typeParameters!==t.typeParameters&&(t=Jc(t),n=vh(n,t,void 0,s));var l=Lb(n),u=wb(n),d=wb(t);if(u&&d&&l!==c)return 0;var p=t.declaration?t.declaration.kind:0,m=!r&&M&&156!==p&&155!==p&&157!==p,f=-1,_=wc(n);if(_&&_!==Ee){var g=wc(t);if(g){var v=!m&&s(_,g,!1)||s(g,_,i);if(!v)return i&&o(e.Diagnostics.The_this_types_of_each_signature_are_incompatible),0;f&=v}}for(var y=u||d?Math.min(l,c):Math.max(l,c),h=u||d?y-1:-1,b=0;b<y;b++){var E=b===h?Db(n,b):Cb(n,b),T=b===h?Db(t,b):Cb(t,b),S=r?void 0:gh(lf(E)),x=r?void 0:gh(lf(T)),C=S&&x&&!Mc(S)&&!Mc(x)&&(98304&nf(E))===(98304&nf(T));v=C?Yp(x,S,m?2:1,!1,i,o,s):!r&&!m&&s(E,T,!1)||s(T,E,i);if(!v)return i&&o(e.Diagnostics.Types_of_parameters_0_and_1_are_incompatible,e.unescapeLeadingUnderscores(xb(n,b)),e.unescapeLeadingUnderscores(xb(t,b))),0;f&=v}if(!a){var A=t.declaration&&ib(t.declaration)?sb(t.declaration.symbol):Fc(t);if(A===Ee)return f;var D=n.declaration&&ib(n.declaration)?sb(n.declaration.symbol):Fc(n),L=Oc(t);if(L){var k=Oc(n);if(k)f&=Qp(k,L,n.declaration,t.declaration,i,o,s);else if(e.isIdentifierTypePredicate(L))return i&&o(e.Diagnostics.Signature_0_must_be_a_type_predicate,Ra(n)),0}else f&=1===r&&s(A,D,!1)||s(D,A,i)}return f}function Qp(n,t,r,a,i,o,s){if(n.kind!==t.kind)return i&&(o(e.Diagnostics.A_this_based_type_guard_is_not_compatible_with_a_parameter_based_type_guard),o(e.Diagnostics.Type_predicate_0_is_not_assignable_to_1,Ba(n),Ba(t))),0;if(1===n.kind){var c=t,l=n.parameterIndex-(e.getThisParameter(r)?1:0),u=c.parameterIndex-(e.getThisParameter(a)?1:0);if(l!==u)return i&&(o(e.Diagnostics.Parameter_0_is_not_in_the_same_position_as_parameter_1,n.parameterName,c.parameterName),o(e.Diagnostics.Type_predicate_0_is_not_assignable_to_1,Ba(n),Ba(t))),0}var d=s(n.type,t.type,i);return 0===d&&i&&o(e.Diagnostics.Type_predicate_0_is_not_assignable_to_1,Ba(n),Ba(t)),d}function Zp(e,n){var t=qc(e),r=qc(n),a=Fc(t),i=Fc(r);return!(i!==Ee&&!am(i,a,Et)&&!am(a,i,Et))&&Xp(t,r,!0)}function $p(e){return 0===e.properties.length&&0===e.callSignatures.length&&0===e.constructSignatures.length&&!e.stringIndexInfo&&!e.numberIndexInfo}function em(n){return 524288&n.flags?!Is(n)&&$p(ws(n)):!!(67108864&n.flags)||(1048576&n.flags?e.some(n.types,em):!!(2097152&n.flags)&&e.every(n.types,em))}function nm(n){return!!(16&e.getObjectFlags(n))&&em(n)}function tm(n,t,r){if(n===t)return!0;var a=u(n)+","+u(t),i=xt.get(a);if(void 0!==i&&(2!==i||!r))return 1===i;if(n.escapedName!==t.escapedName||!(256&n.flags)||!(256&t.flags))return xt.set(a,3),!1;for(var o=Xi(t),s=0,c=Rs(Xi(n));s<c.length;s++){var l=c[s];if(8&l.flags){var d=sc(o,l.escapedName);if(!d||!(8&d.flags))return r?(r(e.Diagnostics.Property_0_is_missing_in_type_1,e.symbolName(l),Fa(Do(t),void 0,64)),xt.set(a,3)):xt.set(a,2),!1}}return xt.set(a,1),!0}function rm(e,n,t,r){var a=e.flags,i=n.flags;if(3&i||131072&a||e===ie)return!0;if(131072&i)return!1;if(132&a&&4&i)return!0;if(128&a&&1024&a&&128&i&&!(1024&i)&&e.value===n.value)return!0;if(296&a&&8&i)return!0;if(256&a&&1024&a&&256&i&&!(1024&i)&&e.value===n.value)return!0;if(2112&a&&64&i)return!0;if(528&a&&16&i)return!0;if(12288&a&&4096&i)return!0;if(32&a&&32&i&&tm(e.symbol,n.symbol,r))return!0;if(1024&a&&1024&i){if(1048576&a&&1048576&i&&tm(e.symbol,n.symbol,r))return!0;if(2944&a&&2944&i&&e.value===n.value&&tm(aa(e.symbol),aa(n.symbol),r))return!0}if(32768&a&&(!w||49152&i))return!0;if(65536&a&&(!w||65536&i))return!0;if(524288&a&&67108864&i)return!0;if(t===Et||t===Tt){if(1&a)return!0;if(264&a&&!(1024&a)&&(32&i||256&i&&1024&i))return!0}return!1}function am(e,n,t){if(kd(e)&&(e=e.regularType),kd(n)&&(n=n.regularType),e===n||t===Tt&&!(131072&n.flags)&&rm(n,e,t)||t!==St&&rm(e,n,t))return!0;if(524288&e.flags&&524288&n.flags){var r=t.get(ym(e,n,t));if(void 0!==r)return 1===r}return!!(66846720&e.flags||66846720&n.flags)&&om(e,n,t,void 0)}function im(n,t,r){return 4096&e.getObjectFlags(n)&&!(ey(t.escapedName)||r)}function om(t,r,a,i,o,s,c){var l,u,d,p,m,f=0,_=0,g=0,v=!1,y=!1;e.Debug.assert(a!==St||!i,"no error reporting in identity checking");var h=w(t,r,!!i,o);if(v)It(i,e.Diagnostics.Excessive_stack_depth_comparing_types_0_and_1,Fa(t),Fa(r));else if(l){if(s){var b=s();b&&(l=e.concatenateDiagnosticMessageChains(b,l))}var E=void 0;if(o&&i&&!h&&t.symbol){var T=qt(t.symbol);if(T.originatingImport&&!e.isImportCall(T.originatingImport)){var S=om(Xi(T.target),r,a,void 0);if(S){var x=e.createDiagnosticForNode(T.originatingImport,e.Diagnostics.Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cause_a_failure_at_runtime_Consider_using_a_default_import_or_import_require_here_instead);E=e.append(E,x)}}}var C=e.createDiagnosticForNodeFromMessageChain(i,l,E);u&&Nt.apply(void 0,[C].concat(u)),c&&(c.error=C),it.add(C)}return 0!==h;function A(n,t,r,a,o){e.Debug.assert(!!i),l=e.chainDiagnosticMessages(l,n,t,r,a,o)}function D(n){e.Debug.assert(!!l),u?u.push(n):u=[n]}function k(n,t,r){var i=Fa(t),o=Fa(r);i===o&&(i=Fa(t,void 0,64),o=Fa(r,void 0,64)),n||(n=a===Tt?e.Diagnostics.Type_0_is_not_comparable_to_type_1:i===o?e.Diagnostics.Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated:e.Diagnostics.Type_0_is_not_assignable_to_type_1),A(n,i,o)}function N(n,t){var r=Fa(n),a=Fa(t);(Xe===n&&pe===t||Ye===n&&me===t||Qe===n&&he===t||Vl(!1)===n&&be===t)&&A(e.Diagnostics._0_is_a_primitive_but_1_is_a_wrapper_object_Prefer_using_0_when_possible,a,r)}function I(e){if(!(3145728&e.flags))return!1;for(var n=!1,t=0,r=e.types;t<r.length;t++){var a=r[t];if(!(98304&a.flags)){if(n)return!0;n=!0}}return!1}function w(t,r,o,s,c){if(void 0===o&&(o=!1),kd(t)&&(t=t.regularType),kd(r)&&(r=r.regularType),33554432&t.flags&&(t=t.substitute),33554432&r.flags&&(r=r.typeVariable),8388608&t.flags&&(t=nd(t)),8388608&r.flags&&(r=nd(r)),1048576&r.flags&&524288&t.flags&&r.types.length<=3&&mE(r,98304)){var u=Y_(r,-98305);1179648&u.flags||(r=u)}if(t===r)return-1;if(a===St)return M(t,r);if(a===Tt&&!(131072&r.flags)&&rm(r,t,a)||rm(t,r,a,o?A:void 0))return-1;var d=!!(4096&e.getObjectFlags(t));if(qf(t)&&32768&e.getObjectFlags(t)){var p=1048576&r.flags?U(t,r):void 0;if(O(t,r,p,o))return o&&k(s,t,r),0;I(r)&&!p&&(t=mf(t))}if(a!==Tt&&!c&&2752508&t.flags&&t!==Ue&&2621440&r.flags&&cm(r)&&(Rs(t).length>0||kC(t))&&!lm(t,r,d)){if(o){var m=lc(t,0),f=lc(t,1);m.length>0&&w(Fc(m[0]),r,!1)||f.length>0&&w(Fc(f[0]),r,!1)?A(e.Diagnostics.Value_of_type_0_has_no_properties_in_common_with_type_1_Did_you_mean_to_call_it,Fa(t),Fa(r)):A(e.Diagnostics.Type_0_has_no_properties_in_common_with_type_1,Fa(t),Fa(r))}return 0}var _=0,g=l,v=!!c;if(1048576&t.flags?_=a===Tt?W(t,r,o&&!(131068&t.flags)):q(t,r,o&&!(131068&t.flags)):(1048576&r.flags?_=G(t,r,o&&!(131068&t.flags)&&!(131068&r.flags)):2097152&r.flags?(v=!0,_=H(t,r,o)):2097152&t.flags&&(_=W(t,r,!1)),!_&&(66846720&t.flags||66846720&r.flags)&&(_=J(t,r,o,v))&&(l=g)),!_&&2097152&t.flags){var h=Ws(t,!!(1048576&r.flags));h&&(_=w(h,r,o,void 0,v))&&(l=g)}if(!_&&o){var b=y;if(y=!1,524288&t.flags&&131068&r.flags)N(t,r);else if(t.symbol&&524288&t.flags&&Ue===t)A(e.Diagnostics.The_Object_type_is_assignable_to_very_few_other_types_Did_you_mean_to_use_the_any_type_instead);else if(d&&2097152&r.flags){var E=r.types,T=oy(n.IntrinsicAttributes,i),S=oy(n.IntrinsicClassAttributes,i);if(T!==oe&&S!==oe&&(e.contains(E,T)||e.contains(E,S)))return _}if(!s&&b)return _;k(s,t,r)}return _}function M(e,n){var t,r=e.flags&n.flags;return 524288&r||8388608&r||16777216&r||4194304&r||33554432&r?J(e,n,!1,!1):3145728&r&&(t=F(e,n))&&(t&=F(n,e))?t:0}function O(n,t,r,o){if(!R&&16384&e.getObjectFlags(t))return!1;if(xy(t)){var s=!!(4096&e.getObjectFlags(n));if((a===Et||a===Tt)&&(H_(Ue,t)||!s&&em(t)))return!1;if(r)return O(n,r,void 0,o);for(var c=function(r){if(P(r,n.symbol)&&!Sy(t,r.escapedName,s)){if(o){var a=J_(t,xy);if(!i)return{value:e.Debug.fail()};if(e.isJsxAttributes(i)||e.isJsxOpeningLikeElement(i)||e.isJsxOpeningLikeElement(i.parent))A(e.Diagnostics.Property_0_does_not_exist_on_type_1,Pa(r),Fa(a));else{var c=n.symbol&&e.firstOrUndefined(n.symbol.declarations),l=void 0;if(r.valueDeclaration&&e.findAncestor(r.valueDeclaration,function(e){return e===c})){var u=r.valueDeclaration;e.Debug.assertNode(u,e.isObjectLiteralElementLike),i=u;var d=u.name;e.isIdentifier(d)&&(l=Hy(d,a))}void 0!==l?A(e.Diagnostics.Object_literal_may_only_specify_known_properties_but_0_does_not_exist_in_type_1_Did_you_mean_to_write_2,Pa(r),Fa(a),l):A(e.Diagnostics.Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1,Pa(r),Fa(a))}}return{value:!0}}},l=0,u=Ms(n);l<u.length;l++){var d=u[l],p=c(d);if("object"===typeof p)return p.value}}return!1}function P(e,n){return e.valueDeclaration&&n.valueDeclaration&&e.valueDeclaration.parent===n.valueDeclaration}function F(e,n){for(var t=-1,r=e.types,a=0,i=r;a<i.length;a++){var o=i[a],s=G(o,n,!1);if(!s)return 0;t&=s}return t}function G(e,n,t){var r=n.types;if(1048576&n.flags&&vu(r,e))return-1;for(var a=0,i=r;a<i.length;a++){var o=i[a],s=w(e,o,!1);if(s)return s}if(t){var c=U(e,n)||V(e,n)||B(e,n)||K(e,n)||j(e,n);w(e,c||r[r.length-1],!0)}return 0}function V(n,t){var r=e.getObjectFlags(n);if(20&r&&1048576&t.flags)return e.find(t.types,function(t){if(524288&t.flags){var a=r&e.getObjectFlags(t);if(4&a)return n.target===t.target;if(16&a)return!!n.aliasSymbol&&n.aliasSymbol===t.aliasSymbol}return!1})}function B(n,t){if(128&e.getObjectFlags(n)&&q_(t,Fm))return e.find(t.types,function(e){return!Fm(e)})}function K(n,t){var r=0,a=lc(n,r).length>0||(r=1,lc(n,r).length>0);if(a)return e.find(t.types,function(e){return lc(e,r).length>0})}function j(n,t){for(var r,a=0,i=0,o=t.types;i<o.length;i++){var s=o[i],c=Ru([Wu(n),Wu(s)]);if(4194304&c.flags)r=s,a=1/0;else if(1048576&c.flags){var l=e.length(e.filter(c.types,Um));l>=a&&(r=s,a=l)}else Um(c)&&1>=a&&(r=s,a=1)}return r}function U(n,t){if(1048576&t.flags){var r=Ms(n);if(r){var a=d_(r,t);if(a)return sm(t,e.map(a,function(e){return[function(){return Xi(e)},e.escapedName]}),w)}}}function H(e,n,t){for(var r=-1,a=n.types,i=0,o=a;i<o.length;i++){var s=o[i],c=w(e,s,t,void 0,!0);if(!c)return 0;r&=c}return r}function W(e,n,t){var r=e.types;if(1048576&e.flags&&vu(r,n))return-1;for(var a=r.length,i=0;i<a;i++){var o=w(r[i],n,t&&i===a-1);if(o)return o}return 0}function q(e,n,t){for(var r=-1,a=e.types,i=0,o=a;i<o.length;i++){var s=o[i],c=w(s,n,t);if(!c)return 0;r&=c}return r}function z(n,t,r,i){if(void 0===n&&(n=e.emptyArray),void 0===t&&(t=e.emptyArray),void 0===r&&(r=e.emptyArray),n.length!==t.length&&a===St)return 0;for(var o=n.length<=t.length?n.length:t.length,s=-1,c=0;c<o;c++){var l=c<r.length?r[c]:1;if(4!==l){var u=n[c],d=t[c],p=-1;if(1===l?p=w(u,d,i):2===l?p=w(d,u,i):3===l?(p=w(d,u,!1),p||(p=w(u,d,i))):(p=w(u,d,i),p&&(p&=w(d,u,i))),!p)return 0;s&=p}}return s}function J(e,n,t,r){if(v)return 0;var i=ym(e,n,a),o=a.get(i);if(void 0!==o&&(!t||2!==o))return 1===o?-1:0;if(d){for(var s=0;s<f;s++)if(i===d[s])return 1;if(100===_)return v=!0,0}else d=[],p=[],m=[];var c=f;d[f]=i,f++,p[_]=e,m[_]=n,_++;var l=g;1&g||!xm(e,p,_)||(g|=1),2&g||!xm(n,m,_)||(g|=2);var u=3!==g?X(e,n,t,r):1;if(g=l,_--,u){if(-1===u||0===_){for(s=c;s<f;s++)a.set(d[s],1);f=c}}else a.set(i,t?3:2),f=c;return u}function X(n,t,r,i){var o,s,c=n.flags&t.flags;if(a===St&&!(524288&c)){if(4194304&c)return w(n.type,t.type,!1);var u=0;return 8388608&c&&(u=w(n.objectType,t.objectType,!1))&&(u&=w(n.indexType,t.indexType,!1))?u:16777216&c&&n.root.isDistributive===t.root.isDistributive&&(u=w(n.checkType,t.checkType,!1))&&(u&=w(n.extendsType,t.extendsType,!1))&&(u&=w(ld(n),ld(t),!1))&&(u&=w(ud(n),ud(t),!1))?u:33554432&c?w(n.substitute,t.substitute,!1):0}var d=l;if(17301504&n.flags&&n.aliasSymbol&&n.aliasTypeArguments&&n.aliasSymbol===t.aliasSymbol&&!n.aliasTypeArgumentsContainsMarker&&!t.aliasTypeArgumentsContainsMarker){var p=dm(n.aliasSymbol);if(o=z(n.aliasTypeArguments,t.aliasTypeArguments,p,r))return o;s=l,l=d}if(262144&t.flags){if(32&e.getObjectFlags(n)&&w(Wu(t),Ts(n))&&!(4&Ds(n))){var m=Ss(n),f=ad(t,Es(n));if(o=w(m,f,r))return o}}else if(4194304&t.flags){if(4194304&n.flags&&(o=w(t.type,n.type,!1)))return o;var _=nd(t.type),g=_!==t.type?_:Vs(t.type);if(g&&-1===w(n,Wu(g,t.stringsOnly),r))return-1}else if(8388608&t.flags){if(a!==St&&(!$u(t.objectType)||!ed(t.indexType))){g=qs(t);if(g&&g!==t&&(o=w(n,g,r)))return o}}else if(Is(t)){var v=Ss(t),y=Ds(t);if(!(8&y)){if(8388608&v.flags&&v.objectType===n&&v.indexType===Es(t))return-1;if(!Is(n)&&w(Ts(t),Wu(n))){f=ad(n,Es(t)),m=Ss(t);if(o=w(f,m,r))return o}s=l,l=d}}if(8650752&n.flags){if(8388608&n.flags&&8388608&t.flags&&((o=w(n.objectType,t.objectType,r))&&(o&=w(n.indexType,t.indexType,r)),o))return l=d,o;g=Vs(n);if(!g||262144&n.flags&&1&g.flags){if(o=w(ke,Y_(t,-67108865)))return l=d,o}else{if(o=w(g,t,!1,void 0,i))return l=d,o;if(o=w(Xo(g,n),t,r,void 0,i))return l=d,o}}else if(4194304&n.flags){if(o=w(De,t,r))return l=d,o}else if(16777216&n.flags)if(16777216&t.flags){if(Sp(n.extendsType,t.extendsType)&&(w(n.checkType,t.checkType)||w(t.checkType,n.checkType))&&((o=w(ld(n),ld(t),r))&&(o&=w(ud(n),ud(t),r)),o))return l=d,o}else{var h=Us(n);if(h&&(o=w(h,t,r)))return l=d,o;var b=js(n);if(b&&(o=w(b,t,r)))return l=d,o}else{if(a!==bt&&Ns(t)&&em(n))return-1;if(Is(t))return Is(n)&&(o=Y(n,t,r))?(l=d,o):0;var E=!!(131068&n.flags);if(a!==St&&(n=rc(n)),4&e.getObjectFlags(n)&&4&e.getObjectFlags(t)&&n.target===t.target&&!(8192&e.getObjectFlags(n)||8192&e.getObjectFlags(t))){p=mm(n.target);if(o=z(n.typeArguments,t.typeArguments,p,r))return o;if(p!==e.emptyArray&&!fm(t,p)){if(!r||!e.some(p,function(e){return 0===e}))return 0;s=l,l=d}}else if(Xm(n)&&(Om(t)||Pm(t))||Om(n)&&Pm(t))return w(mc(n,1)||re,mc(t,1)||re,r);if(2621440&n.flags&&524288&t.flags){var T=r&&l===d&&!E;if(o=Q(n,t,T),o&&(o&=$(n,t,0,T),o&&(o&=$(n,t,1,T),o&&(o&=ie(n,t,0,E,T),o&&(o&=ie(n,t,1,E,T))))),o){if(!s)return l=d,o;l=s}}}return 0}function Y(e,n,t){var r,i=a===Tt||(a===St?Ds(e)===Ds(n):ks(e)<=ks(n));if(i&&(r=w(Ts(n),Ts(e),t))){var o=Ud([Es(e)],[Es(n)]);return r&w(mp(Ss(e),o),Ss(n),t)}return 0}function Q(n,t,r){if(a===St)return Z(n,t);var i=a===bt&&!qf(n)&&!Gm(n)&&!Xm(n),s=Gf(n,t,i,!1);if(s){if(r){var c=e.arrayFrom(Ff(n,t,i,!1));if((!o||o.code!==e.Diagnostics.Class_0_incorrectly_implements_interface_1.code&&o.code!==e.Diagnostics.Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass.code)&&(y=!0),1===c.length){var l=Pa(s);A(e.Diagnostics.Property_0_is_missing_in_type_1_but_required_in_type_2,l,Fa(n),Fa(t)),e.length(s.declarations)&&D(e.createDiagnosticForNode(s.declarations[0],e.Diagnostics._0_is_declared_here,l))}else c.length>5?A(e.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more,Fa(n),Fa(t),e.map(c.slice(0,4),function(e){return Pa(e)}).join(", "),c.length-4):A(e.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2,Fa(n),Fa(t),e.map(c,function(e){return Pa(e)}).join(", "))}return 0}if(qf(t))for(var u=0,d=Rs(n);u<d.length;u++){var p=d[u];if(!Os(t,p.escapedName)){var m=Xi(p);if(m!==ce&&m!==le)return r&&A(e.Diagnostics.Property_0_does_not_exist_on_type_1,Pa(p),Fa(t)),0}}var f=-1;if(Xm(t)){var _=Ym(t);if(_){if(!Xm(n))return 0;var g=Ym(n);if(g&&!w(g,_,r))return r&&A(e.Diagnostics.Rest_signatures_are_incompatible),0;for(var v=ul(t)-1,h=ul(n)-(g?1:0),b=v;b<h;b++){var E=w(n.typeArguments[b],_,r);if(!E)return r&&A(e.Diagnostics.Property_0_is_incompatible_with_rest_element_type,""+b),0;f&=E}}}for(var T=Ms(t),S=0,x=T;S<x.length;S++){var C=x[S];if(!(4194304&C.flags)){p=sc(n,C.escapedName);if(p&&p!==C){if(im(n,p,Xi(C)))continue;var L=e.getDeclarationModifierFlagsFromSymbol(p),k=e.getDeclarationModifierFlagsFromSymbol(C);if(8&L||8&k){var N=p.valueDeclaration!==C.valueDeclaration;if(512&e.getCheckFlags(p)&&N)return r&&A(e.Diagnostics.Property_0_has_conflicting_declarations_and_is_inaccessible_in_type_1,Pa(p),Fa(n)),0;if(N)return r&&(8&L&&8&k?A(e.Diagnostics.Types_have_separate_declarations_of_a_private_property_0,Pa(C)):A(e.Diagnostics.Property_0_is_private_in_type_1_but_not_in_type_2,Pa(C),Fa(8&L?n:t),Fa(8&L?t:n))),0}else if(16&k){if(!Tm(p,C))return r&&A(e.Diagnostics.Property_0_is_protected_but_type_1_is_not_a_class_derived_from_2,Pa(C),Fa(bm(p)||n),Fa(bm(C)||t)),0}else if(16&L)return r&&A(e.Diagnostics.Property_0_is_protected_in_type_1_but_public_in_type_2,Pa(C),Fa(n),Fa(t)),0;E=w(Xi(p),Xi(C),r);if(!E)return r&&A(e.Diagnostics.Types_of_property_0_are_incompatible,Pa(C)),0;if(f&=E,a!==Tt&&16777216&p.flags&&!(16777216&C.flags))return r&&A(e.Diagnostics.Property_0_is_optional_in_type_1_but_required_in_type_2,Pa(C),Fa(n),Fa(t)),0}}}return f}function Z(e,n){if(!(524288&e.flags&&524288&n.flags))return 0;var t=Ms(e),r=Ms(n);if(t.length!==r.length)return 0;for(var a=-1,i=0,o=t;i<o.length;i++){var s=o[i],c=Os(n,s.escapedName);if(!c)return 0;var l=Am(s,c,w);if(!l)return 0;a&=l}return a}function $(n,t,r,i){if(a===St)return ne(n,t,r);if(t===Oe||n===Oe)return-1;var o=n.symbol&&ib(n.symbol.valueDeclaration),s=t.symbol&&ib(t.symbol.valueDeclaration),c=lc(n,o&&1===r?0:r),u=lc(t,s&&1===r?0:r);if(1===r&&c.length&&u.length){if(e.isAbstractConstructorType(n)&&!e.isAbstractConstructorType(t))return i&&A(e.Diagnostics.Cannot_assign_an_abstract_constructor_type_to_a_non_abstract_constructor_type),0;if(!ue(c[0],u[0],i))return 0}var d=-1,p=l;if(64&e.getObjectFlags(n)&&64&e.getObjectFlags(t)&&n.symbol===t.symbol)for(var m=0;m<u.length;m++){var f=ee(c[m],u[m],!0,i);if(!f)return 0;d&=f}else if(1===c.length&&1===u.length){var _=a===Tt||!!L.noStrictGenericChecks;d=ee(c[0],u[0],_,i)}else e:for(var g=0,v=u;g<v.length;g++){for(var y=v[g],h=i,b=0,E=c;b<E.length;b++){var T=E[b];f=ee(T,y,!0,h);if(f){d&=f,l=p;continue e}h=!1}return h&&A(e.Diagnostics.Type_0_provides_no_match_for_the_signature_1,Fa(n),Ra(y,void 0,void 0,r)),0}return d}function ee(e,n,t,r){return Yp(t?qc(e):e,t?qc(n):n,0,!1,r,A,w)}function ne(e,n,t){var r=lc(e,t),a=lc(n,t);if(r.length!==a.length)return 0;for(var i=-1,o=0;o<r.length;o++){var s=Lm(r[o],a[o],!1,!1,!1,w);if(!s)return 0;i&=s}return i}function te(n,t,r,a){for(var i=-1,o=0,s=Ms(n);o<s.length;o++){var c=s[o];if(!im(n,c,void 0)&&(!(c.nameType&&8192&c.nameType.flags)&&(0===r||Uv(c.escapedName)))){var l=w(Xi(c),t,a);if(!l)return a&&A(e.Diagnostics.Property_0_is_incompatible_with_index_signature,Pa(c)),0;i&=l}}return i}function ae(n,t,r){var a=w(n.type,t.type,r);return!a&&r&&A(e.Diagnostics.Index_signatures_are_incompatible),a}function ie(n,t,r,i,o){if(a===St)return se(n,t,r);var s=pc(t,r);if(!s||3&s.type.flags&&!i)return-1;var c=pc(n,r)||1===r&&pc(n,0);if(c)return ae(c,s,o);if(Is(n))return 0===r&&w(Ss(n),s.type,o);if(uf(n)){var l=-1;if(0===r){var u=pc(n,1);u&&(l=ae(u,s,o))}return l&&(l&=te(n,s.type,r,o)),l}return o&&A(e.Diagnostics.Index_signature_is_missing_in_type_0,Fa(n)),0}function se(e,n,t){var r=pc(n,t),a=pc(e,t);return a||r?a&&r&&a.isReadonly===r.isReadonly?w(a.type,r.type):0:-1}function ue(n,t,r){if(!n.declaration||!t.declaration)return!0;var a=e.getSelectedModifierFlags(n.declaration,24),i=e.getSelectedModifierFlags(t.declaration,24);return 8===i||(16===i&&8!==a||(16!==i&&!a||(r&&A(e.Diagnostics.Cannot_assign_a_0_constructor_type_to_a_1_constructor_type,ja(a),ja(i)),!1)))}}function sm(e,n,t,r){for(var a,i=0,o=n;i<o.length;i++)for(var s=o[i],c=s[0],l=s[1],u=0,d=e.types;u<d.length;u++){var p=d[u],m=ti(p,l);if(m&&t(c(),m)){if(a){if(p===a)continue;return r}a=p}}return a||r}function cm(n){if(524288&n.flags){var t=ws(n);return 0===t.callSignatures.length&&0===t.constructSignatures.length&&!t.stringIndexInfo&&!t.numberIndexInfo&&t.properties.length>0&&e.every(t.properties,function(e){return!!(16777216&e.flags)})}return!!(2097152&n.flags)&&e.every(n.types,cm)}function lm(e,n,t){for(var r=0,a=Rs(e);r<a.length;r++){var i=a[r];if(Sy(n,i.escapedName,t))return!0}return!1}function um(n,t,r){var a=cl(n,e.map(n.typeParameters,function(e){return e===t?r:e}));return a.objectFlags|=8192,a}function dm(e){var n=qt(e);return pm(n.typeParameters,n,function(t,r,a){var i=pl(e,Gd(n.typeParameters,Bd(r,a)));return i.aliasTypeArgumentsContainsMarker=!0,i})}function pm(n,t,r){void 0===n&&(n=e.emptyArray);var a=t.variances;if(!a){t.variances=e.emptyArray,a=[];for(var i=0,o=n;i<o.length;i++){var s=o[i],c=r(t,s,Ge),l=r(t,s,Ve),u=(Lp(l,c)?1:0)|(Lp(c,l)?2:0);3===u&&Lp(r(t,s,Cn),c)&&(u=4),a.push(u)}t.variances=a}return a}function mm(n){return M?n===ze||n===Je?e.emptyArray:pm(n.typeParameters,n,um):e.emptyArray}function fm(e,n){for(var t=0;t<n.length;t++)if(1===n[t]&&16384&e.typeArguments[t].flags)return!0;return!1}function _m(e){return 262144&e.flags&&!Bs(e)}function gm(n){return!!(4&e.getObjectFlags(n))&&e.some(n.typeArguments,function(e){return _m(e)||gm(e)})}function vm(e,n,t){void 0===t&&(t=0);for(var r=""+e.target.id,a=0,i=e.typeArguments;a<i.length;a++){var o=i[a];if(_m(o)){var s=n.indexOf(o);s<0&&(s=n.length,n.push(o)),r+="="+s}else t<4&&gm(o)?r+="<"+vm(o,n,t+1)+">":r+="-"+o.id}return r}function ym(e,n,t){if(t===St&&e.id>n.id){var r=e;e=n,n=r}if(gm(e)&&gm(n)){var a=[];return vm(e,a)+","+vm(n,a)}return e.id+","+n.id}function hm(n,t){if(!(6&e.getCheckFlags(n)))return t(n);for(var r=0,a=n.containingType.types;r<a.length;r++){var i=a[r],o=sc(i,n.escapedName),s=o&&hm(o,t);if(s)return s}}function bm(e){return e.parent&&32&e.parent.flags?Do(aa(e)):void 0}function Em(e,n){return hm(e,function(e){var t=bm(e);return!!t&&Zi(t,n)})}function Tm(n,t){return!hm(t,function(t){return!!(16&e.getDeclarationModifierFlagsFromSymbol(t))&&!Em(n,bm(t))})}function Sm(n,t){return hm(t,function(t){return!!(16&e.getDeclarationModifierFlagsFromSymbol(t))&&!Zi(n,bm(t))})?void 0:n}function xm(e,n,t){if(t>=5&&524288&e.flags){var r=e.symbol;if(r)for(var a=0,i=0;i<t;i++){var o=n[i];if(524288&o.flags&&o.symbol===r&&(a++,a>=5))return!0}}return!1}function Cm(e,n){return 0!==Am(e,n,xp)}function Am(n,t,r){if(n===t)return-1;var a=24&e.getDeclarationModifierFlagsFromSymbol(n),i=24&e.getDeclarationModifierFlagsFromSymbol(t);if(a!==i)return 0;if(a){if(fx(n)!==fx(t))return 0}else if((16777216&n.flags)!==(16777216&t.flags))return 0;return tE(n)!==tE(t)?0:r(Xi(n),Xi(t))}function Dm(e,n,t){var r=Lb(e),a=Lb(n),i=kb(e),o=kb(n),s=Nb(e),c=Nb(n);return r===a&&i===o&&s===c||!!(t&&i<=o)}function Lm(n,t,r,a,i,o){if(n===t)return-1;if(!Dm(n,t,r))return 0;if(e.length(n.typeParameters)!==e.length(t.typeParameters))return 0;n=qc(n),t=qc(t);var s=-1;if(!a){var c=wc(n);if(c){var l=wc(t);if(l){var u=o(c,l);if(!u)return 0;s&=u}}}for(var d=Lb(t),p=0;p<d;p++){var m=Cb(n,p),f=Cb(t,p);u=o(f,m);if(!u)return 0;s&=u}if(!i){var _=Oc(n),g=Oc(t);s&=void 0!==_||void 0!==g?km(_,g,o):o(Fc(n),Fc(t))}return s}function km(e,n,t){return void 0!==e&&void 0!==n&&Lu(e,n)?t(e.type,n.type):0}function Nm(e){for(var n,t=0,r=e;t<r.length;t++){var a=r[t],i=Wm(a);if(n||(n=i),i===a||i!==n)return!1}return!0}function Im(n){return Nm(n)?Au(n):e.reduceLeft(n,function(e,n){return Dp(e,n)?n:e})}function wm(n){if(!w)return Im(n);var t=e.filter(n,function(e){return!(98304&e.flags)});return t.length?of(Im(t),98304&ef(n)):Au(n,2)}function Mm(n){return e.reduceLeft(n,function(e,n){return Dp(n,e)?n:e})}function Om(n){return!!(4&e.getObjectFlags(n))&&n.target===ze}function Pm(n){return!!(4&e.getObjectFlags(n))&&n.target===Je}function Rm(e){return Om(e)&&e.typeArguments?e.typeArguments[0]:void 0}function Fm(n){return 4&e.getObjectFlags(n)&&(n.target===ze||n.target===Je)||!(98304&n.flags)&&Lp(n,tn)}function Gm(e){var n=Om(e)?e.typeArguments[0]:void 0;return n===le||n===xe}function Vm(e){return Xm(e)||!!sc(e,"0")}function Bm(e){return Fm(e)||Vm(e)}function Km(e,n){var t=ti(e,""+n);return t||(z_(e,Xm)&&!z_(e,function(e){return!e.target.hasRestElement})?X_(e,function(e){return Ym(e)||ce}):void 0)}function jm(e){return!(240512&e.flags)}function Um(e){return!!(109440&e.flags)}function Hm(n){return!!(16&n.flags)||(1048576&n.flags?!!(1024&n.flags)||e.every(n.types,Um):Um(n))}function Wm(n){return 1024&n.flags?To(n):128&n.flags?pe:256&n.flags?me:2048&n.flags?fe:512&n.flags?he:1048576&n.flags?Au(e.sameMap(n.types,Wm)):n}function qm(n){return 1024&n.flags&&kd(n)?To(n):128&n.flags&&kd(n)?pe:256&n.flags&&kd(n)?me:2048&n.flags&&kd(n)?fe:512&n.flags&&kd(n)?he:1048576&n.flags?Au(e.sameMap(n.types,qm)):n}function zm(n){return 8192&n.flags?be:1048576&n.flags?Au(e.sameMap(n.types,zm)):n}function Jm(e,n){return VE(e,n)||(e=zm(qm(e))),e}function Xm(n){return!!(4&e.getObjectFlags(n)&&8&n.target.objectFlags)}function Ym(e){return e.target.hasRestElement?e.typeArguments[e.target.typeParameters.length-1]:void 0}function Qm(e){var n=Ym(e);return n&&su(n)}function Zm(e){return ul(e)-(e.target.hasRestElement?1:0)}function $m(e){var n=e.value;return"0"===n.base10Value}function ef(e){for(var n=0,t=0,r=e;t<r.length;t++){var a=r[t];n|=nf(a)}return n}function nf(e){return 1048576&e.flags?ef(e.types):128&e.flags?""===e.value?128:0:256&e.flags?0===e.value?256:0:2048&e.flags?$m(e)?2048:0:512&e.flags?e===_e||e===ge?512:0:117724&e.flags}function tf(e){return 117632&nf(e)?J_(e,function(e){return!(117632&nf(e))}):e}function rf(e){return X_(e,af)}function af(e){return 4&e.flags?Vn:8&e.flags?Bn:64&e.flags?Kn:e===ge||e===_e||114688&e.flags||128&e.flags&&""===e.value||256&e.flags&&0===e.value||2048&e.flags&&$m(e)?e:Te}function of(e,n){var t=n&~e.flags&98304;return 0===t?e:Au(32768===t?[e,ce]:65536===t?[e,ue]:[e,ce,ue])}function sf(n){return e.Debug.assert(w),32768&n.flags?n:Au([n,ce])}function cf(e){return rn||(rn=Ml("NonNullable",524288,void 0)||ne),rn!==ne?pl(rn,[e]):b_(e,2097152)}function lf(e){return w?cf(e):e}function uf(e){return e.symbol&&0!==(6656&e.symbol.flags)&&!kC(e)}function df(e,n){var t=Ot(e.flags,e.escapedName);return t.declarations=e.declarations,t.parent=e.parent,t.type=n,t.target=e,e.valueDeclaration&&(t.valueDeclaration=e.valueDeclaration),e.nameType&&(t.nameType=e.nameType),t}function pf(n,t){for(var r=e.createSymbolTable(),a=0,i=Ms(n);a<i.length;a++){var o=i[a],s=Xi(o),c=t(s);r.set(o.escapedName,c===s?o:df(o,c))}return r}function mf(n){if(!(qf(n)&&32768&e.getObjectFlags(n)))return n;var t=n.regularType;if(t)return t;var r=n,a=pf(n,mf),i=ba(r.symbol,a,r.callSignatures,r.constructSignatures,r.stringIndexInfo,r.numberIndexInfo);return i.flags=r.flags,i.objectFlags|=128|16384&e.getObjectFlags(r),n.regularType=i,i}function ff(e,n,t){return{parent:e,propertyName:n,siblings:t,resolvedProperties:void 0}}function _f(e){if(!e.siblings){for(var n=[],t=0,r=_f(e.parent);t<r.length;t++){var a=r[t];if(qf(a)){var i=Os(a,e.propertyName);i&&q_(Xi(i),function(e){n.push(e)})}}e.siblings=n}return e.siblings}function gf(n){if(!n.resolvedProperties){for(var t=e.createMap(),r=0,a=_f(n);r<a.length;r++){var i=a[r];if(qf(i)&&!(1024&e.getObjectFlags(i)))for(var o=0,s=Rs(i);o<s.length;o++){var c=s[o];t.set(c.escapedName,c)}}n.resolvedProperties=e.arrayFrom(t.values())}return n.resolvedProperties}function vf(e,n){if(!(4&e.flags))return e;var t=Xi(e),r=n&&ff(n,e.escapedName,void 0),a=Ef(t,r);return a===t?e:df(e,a)}function yf(e){var n=ee.get(e.escapedName);if(n)return n;var t=df(e,ce);return t.flags|=16777216,ee.set(e.escapedName,t),t}function hf(n,t){for(var r=e.createSymbolTable(),a=0,i=Ms(n);a<i.length;a++){var o=i[a];r.set(o.escapedName,vf(o,t))}if(t)for(var s=0,c=gf(t);s<c.length;s++){o=c[s];r.has(o.escapedName)||r.set(o.escapedName,yf(o))}var l=pc(n,0),u=pc(n,1),d=ba(n.symbol,r,e.emptyArray,e.emptyArray,l&&el(bf(l.type),l.isReadonly),u&&el(bf(u.type),u.isReadonly));return d.objectFlags|=16384&e.getObjectFlags(n),d}function bf(e){return Ef(e,void 0)}function Ef(n,t){if(402653184&n.flags){if(98304&n.flags)return re;if(qf(n))return hf(n,t);if(1048576&n.flags){var r=t||ff(void 0,void 0,n.types),a=e.sameMap(n.types,function(e){return 98304&e.flags?e:Ef(e,r)});return Au(a,e.some(a,em)?2:1)}if(2097152&n.flags)return Ru(e.sameMap(n.types,bf));if(Om(n)||Xm(n))return cl(n.target,e.sameMap(n.typeArguments,bf))}return n}function Tf(n){var t=!1;if(134217728&n.flags){if(1048576&n.flags)if(e.some(n.types,em))t=!0;else for(var r=0,a=n.types;r<a.length;r++){var i=a[r];Tf(i)&&(t=!0)}if(Om(n)||Xm(n))for(var o=0,s=n.typeArguments;o<s.length;o++){i=s[o];Tf(i)&&(t=!0)}if(qf(n))for(var c=0,l=Ms(n);c<l.length;c++){var u=l[c];i=Xi(u);134217728&i.flags&&(Tf(i)||It(u.valueDeclaration,e.Diagnostics.Object_literal_s_property_0_implicitly_has_an_1_type,Pa(u),Fa(bf(i))),t=!0)}}return t}function Sf(n,t){var r=Fa(bf(t));if(!e.isInJSFile(n)||e.isCheckJsEnabledForFile(e.getSourceFileOfNode(n),L)){var a;switch(n.kind){case 204:case 154:case 153:a=R?e.Diagnostics.Member_0_implicitly_has_an_1_type:e.Diagnostics.Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage;break;case 151:var i=n;if(e.isIdentifier(i.name)&&(e.isCallSignatureDeclaration(i.parent)||e.isMethodSignature(i.parent)||e.isFunctionTypeNode(i.parent))&&i.parent.parameters.indexOf(i)>-1&&(Zt(i,i.name.escapedText,67897832,void 0,i.name.escapedText,!0)||i.name.originalKeywordKind&&e.isTypeNodeKind(i.name.originalKeywordKind))){var o="arg"+i.parent.parameters.indexOf(i);return void Mt(R,n,e.Diagnostics.Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1,o,e.declarationNameToString(i.name))}a=n.dotDotDotToken?R?e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type:e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage:R?e.Diagnostics.Parameter_0_implicitly_has_an_1_type:e.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage;break;case 186:if(a=e.Diagnostics.Binding_element_0_implicitly_has_an_1_type,!R)return;break;case 289:return void It(n,e.Diagnostics.Function_type_which_lacks_return_type_annotation_implicitly_has_an_0_return_type,r);case 239:case 156:case 155:case 158:case 159:case 196:case 197:if(R&&!n.name)return void It(n,e.Diagnostics.Function_expression_which_lacks_return_type_annotation_implicitly_has_an_0_return_type,r);a=R?e.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type:e.Diagnostics._0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage;break;case 181:return void(R&&It(n,e.Diagnostics.Mapped_object_type_implicitly_has_an_any_template_type));default:a=R?e.Diagnostics.Variable_0_implicitly_has_an_1_type:e.Diagnostics.Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage}Mt(R,n,a,e.declarationNameToString(e.getNameOfDeclaration(n)),r)}}function xf(e,n){a&&R&&134217728&n.flags&&(Tf(n)||Sf(e,n))}function Cf(e,n,t){var r=Lb(e),a=Lb(n),i=Ib(e),o=Ib(n),s=o?a-1:a,c=i?s:Math.min(r,s),l=wc(e);if(l){var u=wc(n);u&&t(l,u)}for(var d=0;d<c;d++)t(Cb(e,d),Cb(n,d));o&&t(Db(e,c),o)}function Af(e,n,t,r,a){var i=a?a.map(Lf):e.map(Df),o=s;return o.typeParameters=e,o.signature=n,o.inferences=i,o.flags=t,o.compareTypes=r||Cp,o;function s(e){for(var n=0;n<i.length;n++)if(e===i[n].typeParameter)return i[n].isFixed=!0,Yf(o,n);return e}}function Df(e){return{typeParameter:e,candidates:void 0,contraCandidates:void 0,inferredType:void 0,priority:void 0,topLevel:!0,isFixed:!1}}function Lf(e){return{typeParameter:e.typeParameter,candidates:e.candidates&&e.candidates.slice(),contraCandidates:e.contraCandidates&&e.contraCandidates.slice(),inferredType:e.inferredType,priority:e.priority,topLevel:e.topLevel,isFixed:e.isFixed}}function kf(n){var t=e.getObjectFlags(n);return!!(63176704&n.flags||4&t&&e.forEach(n.typeArguments,kf)||16&t&&n.symbol&&10288&n.symbol.flags||32&t||3145728&n.flags&&Nf(n))}function Nf(n){return void 0===n.couldContainTypeVariables&&(n.couldContainTypeVariables=e.some(n.types,kf)),n.couldContainTypeVariables}function If(n,t){return n===t||!!(3145728&n.flags)&&e.some(n.types,function(e){return If(e,t)})}function wf(n){var t=e.createSymbolTable();q_(n,function(n){if(128&n.flags){var r=e.escapeLeadingUnderscores(n.value),a=Ot(4,r);a.type=re,n.symbol&&(a.declarations=n.symbol.declarations,a.valueDeclaration=n.symbol.valueDeclaration),t.set(r,a)}});var r=4&n.flags?el(ke,!1):void 0;return ba(void 0,t,e.emptyArray,e.emptyArray,r,void 0)}function Mf(e,n,t){var r=e.id+","+n.id+","+t.id;if(Mn.has(r))return Mn.get(r);Mn.set(r,void 0);var a=Of(e,n,t);return Mn.set(r,a),a}function Of(n,t,r){var a=Rs(n);if(0!==a.length||pc(n,0)){for(var i=0,o=a;i<o.length;i++){var s=o[i];if(536870912&Xi(s).flags)return}if(Om(n))return su(Rf(n.typeArguments[0],t,r));if(Pm(n))return cu(Rf(n.typeArguments[0],t,r));if(Xm(n)){var c=e.map(n.typeArguments||e.emptyArray,function(e){return Rf(e,t,r)}),l=4&Ds(t)?ul(n)-(n.target.hasRestElement?1:0):n.target.minLength;return pu(c,l,n.target.hasRestElement,n.target.associatedNames)}var u=fa(2064,void 0);return u.source=n,u.mappedType=t,u.constraintType=r,u}}function Pf(e){return Rf(e.propertyType,e.mappedType,e.constraintType)}function Rf(e,n,t){var r=ad(t.type,Es(n)),a=Ss(n),i=Df(r);return jf([i],e,a),Kf(i)}function Ff(e,n,t,r){var a,i,o,c,l,u,d;return s(this,function(s){switch(s.label){case 0:a=2097152&n.flags?Ps(n):Ms(n),i=0,o=a,s.label=1;case 1:return i<o.length?(c=o[i],!t&&16777216&c.flags?[3,5]:(l=sc(e,c.escapedName),l?[3,3]:[4,c])):[3,6];case 2:return s.sent(),[3,5];case 3:return r?(u=Xi(c),109440&u.flags?(d=Xi(l),1&d.flags||Ld(d)===Ld(u)?[3,5]:[4,c]):[3,5]):[3,5];case 4:s.sent(),s.label=5;case 5:return i++,[3,1];case 6:return[2]}})}function Gf(e,n,t,r){return Ff(e,n,t,r).next().value}function Vf(e,n){return n.target.minLength>e.target.minLength||!Ym(n)&&(!!Ym(e)||Zm(n)<Zm(e))}function Bf(e,n){return Xm(e)&&Xm(n)&&Vf(e,n)||!!Gf(e,n,!1,!0)&&!!Gf(n,e,!1,!0)}function Kf(e){return e.candidates?Au(e.candidates,2):e.contraCandidates?Ru(e.contraCandidates):ke}function jf(n,t,r,a){var i,o;void 0===a&&(a=0);var s,c=!1,l=!1,u=!0;function d(n,t){if(kf(t)){if(n===ie){var f=s;return s=n,d(t,t),void(s=f)}if(n.aliasSymbol&&n.aliasTypeArguments&&n.aliasSymbol===t.aliasSymbol)for(var g=n.aliasTypeArguments,v=t.aliasTypeArguments,y=0;y<g.length;y++)d(g[y],v[y]);else{if(1048576&n.flags&&1048576&t.flags&&!(1024&n.flags&&1024&t.flags)||2097152&n.flags&&2097152&t.flags){if(n===t){for(var h=0,b=n.types;h<b.length;h++){var E=b[h];d(E,E)}return}for(var T=void 0,S=0,x=n.types;S<x.length;S++){E=x[S];if(Uf(E,t.types))(T||(T=[])).push(E),d(E,E);else if(384&E.flags){var C=Wm(E);Uf(C,t.types)&&(T||(T=[])).push(E,C)}}T&&(n=Hf(n,T),t=Hf(t,T))}if(8650752&t.flags){if(536870912&n.flags||n===Se||8&a&&(n===ae||n===nn))return;var A=m(t);if(A){if(!A.isFixed){if((void 0===A.priority||a<A.priority)&&(A.candidates=void 0,A.contraCandidates=void 0,A.priority=a),a===A.priority){var D=s||n;c&&!l?A.contraCandidates=e.appendIfUnique(A.contraCandidates,D):A.candidates=e.appendIfUnique(A.candidates,D)}8&a||!(262144&t.flags)||If(r,t)||(A.topLevel=!1)}return}var L=nd(t);if(L!==t)q(n,L);else if(8388608&t.flags){var k=nd(t.indexType);if(63176704&k.flags){var N=td(nd(t.objectType),k);N&&N!==t&&q(n,N)}}}if(4&e.getObjectFlags(n)&&4&e.getObjectFlags(t)&&n.target===t.target){g=n.typeArguments||e.emptyArray,v=t.typeArguments||e.emptyArray;var I=g.length<v.length?g.length:v.length,w=mm(n.target);for(y=0;y<I;y++)y<w.length&&2===w[y]?p(g[y],v[y]):d(g[y],v[y])}else if(4194304&n.flags&&4194304&t.flags)c=!c,d(n.type,t.type),c=!c;else if((Hm(n)||4&n.flags)&&4194304&t.flags){var M=wf(n);c=!c;var O=a;a|=16,d(M,t.type),a=O,c=!c}else if(8388608&n.flags&&8388608&t.flags)d(n.objectType,t.objectType),d(n.indexType,t.indexType);else if(16777216&n.flags&&16777216&t.flags)d(n.checkType,t.checkType),d(n.extendsType,t.extendsType),d(ld(n),ld(t)),d(ud(n),ud(t));else if(16777216&t.flags)d(n,Au([ld(t),ud(t)]));else if(3145728&t.flags){v=t.types;for(var P=0,R=void 0,F=0,G=v;F<G.length;F++){E=G[F];m(E)?(R=E,P++):d(n,E)}if(1===P){O=a;a|=1,d(n,R),a=O}}else if(1048576&n.flags){g=n.types;for(var V=0,B=g;V<B.length;V++){var K=B[V];d(K,t)}}else{if(!(32&a&&65273856&n.flags)){var j=rc(n);if(j!==n&&u&&!(2621440&j.flags))return u=!1,d(j,t);n=j}if(2621440&n.flags){var U=n.id+","+t.id;if(o&&o.get(U))return;(o||(o=e.createMap())).set(U,!0);var H=524288&t.flags&&!(16&e.getObjectFlags(t)&&t.symbol&&32&t.symbol.flags),W=H?t.symbol:void 0;if(W){if(e.contains(i,W))return;(i||(i=[])).push(W),_(n,t),i.pop()}else _(n,t)}}}}function q(n,t){var r=n.id+","+t.id;o&&o.get(r)||((o||(o=e.createMap())).set(r,!0),d(n,t))}}function p(e,n){M||64&a?(c=!c,d(e,n),c=!c):d(e,n)}function m(e){if(8650752&e.flags)for(var t=0,r=n;t<r.length;t++){var a=r[t];if(e===a.typeParameter)return a}}function f(n,t,r){if(1048576&r.flags){for(var i=!1,o=0,s=r.types;o<s.length;o++){var c=s[o];i=f(n,t,c)||i}return i}if(4194304&r.flags){var l=m(r.type);if(l&&!l.isFixed){var u=Mf(n,t,r);if(u){var p=a;a|=2,d(u,l.typeParameter),a=p}}return!0}if(262144&r.flags){p=a;a|=4,d(Wu(n),r),a=p;var _=e.compact([mc(n,0),mc(n,1)].concat(e.map(Rs(n),Xi)));return d(Au(_),Ss(t)),!0}return!1}function _(n,t){if(Is(n)&&Is(t)&&(d(Ts(n),Ts(t)),d(Ss(n),Ss(t))),32&e.getObjectFlags(t)){var r=Ts(t);if(f(n,t,r))return}Bf(n,t)||(g(n,t),v(n,t,0),v(n,t,1),h(n,t))}function g(e,n){if(Xm(e)){if(Xm(n)){for(var t=Zm(e),r=Zm(n),a=Ym(e),i=Ym(n),o=r<t||a?r:t,s=0;s<o;s++)d(s<t?e.typeArguments[s]:a,n.typeArguments[s]);if(i){var c=o<t?e.typeArguments.slice(o,t):[];a&&c.push(a),c.length&&d(Au(c),i)}return}if(Om(n))return void h(e,n)}for(var l=Ms(n),u=0,p=l;u<p.length;u++){var m=p[u],f=sc(e,m.escapedName);f&&d(Xi(f),Xi(m))}}function v(e,n,t){for(var r=lc(e,t),a=lc(n,t),i=r.length,o=a.length,s=i<o?i:o,c=!!(536870912&e.flags),l=0;l<s;l++)y(Yc(r[i-s+l]),Yc(a[o-s+l]),c)}function y(e,n,t){if(!t){var r=l,a=n.declaration?n.declaration.kind:0;l=l||156===a||155===a||157===a,Cf(e,n,p),l=r}var i=Oc(e),o=Oc(n);i&&o&&i.kind===o.kind?d(i.type,o.type):d(Fc(e),Fc(n))}function h(e,n){var t=mc(n,0);if(t){var r=mc(e,0)||fc(e,0);r&&d(r,t)}var a=mc(n,1);if(a){r=mc(e,1)||mc(e,0)||fc(e,1);r&&d(r,a)}}d(t,r)}function Uf(e,n){for(var t=0,r=n;t<r.length;t++){var a=r[t];if(Sp(a,e))return!0}return!1}function Hf(e,n){for(var t=[],r=0,a=e.types;r<a.length;r++){var i=a[r];Uf(i,n)||t.push(i)}return 1048576&e.flags?Au(t):Ru(t)}function Wf(e){var n=Bs(e);return!!n&&mE(16777216&n.flags?js(n):n,4325372)}function qf(n){return!!(128&e.getObjectFlags(n))}function zf(n){if(n.length>1){var t=e.filter(n,qf);if(t.length){var r=bf(Au(t,2));return e.concatenate(e.filter(n,function(e){return!qf(e)}),[r])}}return n}function Jf(e){return 28&e.priority?Ru(e.contraCandidates):Mm(e.contraCandidates)}function Xf(n,t){var r=zf(n.candidates),a=Wf(n.typeParameter),i=!a&&n.topLevel&&(n.isFixed||!If(Fc(t),n.typeParameter)),o=a?e.sameMap(r,Ld):i?e.sameMap(r,qm):r,s=28&n.priority?Au(o,2):wm(o);return bf(s)}function Yf(e,n){var t=e.inferences[n],r=t.inferredType;if(!r){var a=e.signature;if(a){var i=t.candidates?Xf(t,a):void 0;if(t.contraCandidates){var o=Jf(t);r=!i||131072&i.flags||!Dp(i,o)?o:i}else if(i)r=i;else if(1&e.flags)r=Se;else{var s=Zs(t.typeParameter);r=s?mp(s,Jd(Wd(e.signature.typeParameters,n),e)):Qf(!!(2&e.flags))}}else r=Kf(t);t.inferredType=r;var c=Bs(t.typeParameter);if(c){var l=mp(c,e);e.compareTypes(r,Xo(l,r))||(t.inferredType=r=l)}}return r}function Qf(e){return e?re:ke}function Zf(e){for(var n=[],t=0;t<e.inferences.length;t++)n.push(Yf(e,t));return n}function $f(n){switch(n){case"document":case"console":return e.Diagnostics.Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_include_dom;case"$":return e.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_types_Slashjquery_and_then_add_jquery_to_the_types_field_in_your_tsconfig;case"describe":case"suite":case"it":case"test":return e.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_types_Slashjest_or_npm_i_types_Slashmocha_and_then_add_jest_or_mocha_to_the_types_field_in_your_tsconfig;case"process":case"require":case"Buffer":case"module":return e.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_types_Slashnode_and_then_add_node_to_the_types_field_in_your_tsconfig;case"Map":case"Set":case"Promise":case"Symbol":case"WeakMap":case"WeakSet":case"Iterator":case"AsyncIterator":return e.Diagnostics.Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_es2015_or_later;default:return e.Diagnostics.Cannot_find_name_0}}function e_(n){var t=zt(n);return t.resolvedSymbol||(t.resolvedSymbol=!e.nodeIsMissing(n)&&Zt(n,n.escapedText,68268991,$f(n.escapedText),n,!e.isWriteOnlyAccess(n),!1,e.Diagnostics.Cannot_find_name_0_Did_you_mean_1)||ne),t.resolvedSymbol}function n_(n){return!!e.findAncestor(n,function(e){return 167===e.kind||72!==e.kind&&148!==e.kind&&"quit"})}function t_(e){switch(e.kind){case 72:var n=e_(e);return n!==ne?(Tg(e)?"@":"")+u(n):void 0;case 100:return"0";case 213:case 195:return t_(e.expression);case 189:case 190:var t=a_(e);if(void 0!==t){var r=t_(e.expression);return r&&r+"."+t}}}function r_(n,t){switch(t.kind){case 195:case 213:return r_(n,t.expression)}switch(n.kind){case 72:return 72===t.kind&&e_(n)===e_(t)||(237===t.kind||186===t.kind)&&ca(e_(n))===ra(t);case 100:return 100===t.kind;case 98:return 98===t.kind;case 213:case 195:return r_(n.expression,t);case 189:case 190:return e.isAccessExpression(t)&&a_(n)===a_(t)&&r_(n.expression,t.expression)}return!1}function a_(n){return 189===n.kind?n.name.escapedText:e.isStringLiteral(n.argumentExpression)||e.isNumericLiteral(n.argumentExpression)?e.escapeLeadingUnderscores(n.argumentExpression.text):void 0}function i_(n,t){while(e.isAccessExpression(n))if(n=n.expression,r_(n,t))return!0;return!1}function o_(n,t){var r;return e.isAccessExpression(t)&&i_(n,t.expression)&&void 0!==(r=a_(t))&&l_(s_(t.expression),r)}function s_(n){if(72===n.kind)return Xi(e_(n));if(e.isAccessExpression(n)){var t=s_(n.expression);if(t){var r=a_(n);return void 0!==r?ti(t,r):void 0}}}function c_(e){return!!(1048576&e.flags&&(1040&e.flags||!ed(e)))}function l_(n,t){if(n&&1048576&n.flags){var r=ic(n,t);if(r&&2&e.getCheckFlags(r))return void 0===r.isDiscriminantProperty&&(r.isDiscriminantProperty=96===(96&r.checkFlags)&&c_(Xi(r))),!!r.isDiscriminantProperty}return!1}function u_(e){var n=s_(e);return!!(n&&1048576&n.flags)}function d_(e,n){for(var t,r=0,a=e;r<a.length;r++){var i=a[r];if(l_(n,i.escapedName)){if(t){t.push(i);continue}t=[i]}}return t}function p_(e,n){return r_(e,n)||i_(e,n)}function m_(e,n){if(e.arguments)for(var t=0,r=e.arguments;t<r.length;t++){var a=r[t];if(p_(n,a))return!0}return!(189!==e.expression.kind||!p_(n,e.expression.expression))}function f_(e){return e.id||(e.id=c,c++),e.id}function __(e,n){if(!(1048576&e.flags))return Lp(e,n);for(var t=0,r=e.types;t<r.length;t++){var a=r[t];if(Lp(a,n))return!0}return!1}function g_(e,n){if(e!==n){if(131072&n.flags)return n;var t=J_(e,function(e){return __(n,e)});if(512&n.flags&&kd(n)&&(t=X_(t,Dd)),Lp(n,t))return t}return e}function v_(e){for(var n=0,t=0,r=e;t<r.length;t++){var a=r[t];n|=h_(a)}return n}function y_(e){var n=ws(e);return!!(n.callSignatures.length||n.constructSignatures.length||n.members.get("bind")&&Dp(e,He))}function h_(n){var t=n.flags;if(4&t)return w?16317953:16776705;if(128&t){var r=""===n.value;return w?r?12123649:7929345:r?12582401:16776705}if(40&t)return w?16317698:16776450;if(256&t){var a=0===n.value;return w?a?12123394:7929090:a?12582146:16776450}if(64&t)return w?16317188:16775940;if(2048&t){a=$m(n);return w?a?12122884:7928580:a?12581636:16775940}return 16&t?w?16316168:16774920:528&t?w?n===_e||n===ge?12121864:7927560:n===_e||n===ge?12580616:16774920:524288&t?16&e.getObjectFlags(n)&&em(n)?w?16318463:16777215:y_(n)?w?7880640:16728e3:w?7888800:16736160:49152&t?9830144:65536&t?9363232:12288&t?w?7925520:16772880:67108864&t?w?7888800:16736160:63176704&t?h_(qs(n)||ke):3145728&t?v_(n.types):16777215}function b_(e,n){return J_(e,function(e){return 0!==(h_(e)&n)})}function E_(e,n){if(n){var t=HE(n);return Au([b_(e,524288),t])}return e}function T_(e,n){var t=Bu(n);if(!Fo(t))return oe;var r=Uo(t);return xg(ti(e,r),n)||Uv(r)&&mc(e,1)||mc(e,0)||oe}function S_(e,n){return z_(e,Vm)&&Km(e,n)||KS(e,void 0,!1,!1)||oe}function x_(e){return su(KS(e,void 0,!1,!1)||oe)}function C_(e){var n=187===e.parent.kind&&A_(e.parent)||275===e.parent.kind&&A_(e.parent.parent);return n?E_(I_(e),e.right):HE(e.right)}function A_(e){return 204===e.parent.kind&&e.parent.left===e||227===e.parent.kind&&e.parent.initializer===e}function D_(e,n){return S_(I_(e),e.elements.indexOf(n))}function L_(e){return x_(I_(e.parent))}function k_(e){return T_(I_(e.parent),e.name)}function N_(e){return E_(k_(e),e.objectAssignmentInitializer)}function I_(e){var n=e.parent;switch(n.kind){case 226:return pe;case 227:return BS(n.expression,n.awaitModifier)||oe;case 204:return C_(n);case 198:return ce;case 187:return D_(n,e);case 208:return L_(n);case 275:return k_(n);case 276:return N_(n)}return oe}function w_(e){var n=e.parent,t=P_(n.parent),r=184===n.kind?T_(t,e.propertyName||e.name):e.dotDotDotToken?x_(t):S_(t,n.elements.indexOf(e));return E_(r,e.initializer)}function M_(e){var n=zt(e);return n.resolvedType||HE(e)}function O_(e){return e.initializer?M_(e.initializer):226===e.parent.parent.kind?pe:227===e.parent.parent.kind&&BS(e.parent.parent.expression,e.parent.parent.awaitModifier)||oe}function P_(e){return 237===e.kind?O_(e):w_(e)}function R_(e,n){return xg(237===e.kind||186===e.kind?P_(e):I_(e),n)}function F_(e){return 237===e.kind&&e.initializer&&gi(e.initializer)||186!==e.kind&&204===e.parent.kind&&gi(e.parent.right)}function G_(e){switch(e.kind){case 195:return G_(e.expression);case 204:switch(e.operatorToken.kind){case 59:return G_(e.left);case 27:return G_(e.right)}}return e}function V_(e){var n=e.parent;return 195===n.kind||204===n.kind&&59===n.operatorToken.kind&&n.left===e||204===n.kind&&27===n.operatorToken.kind&&n.right===e?V_(n):e}function B_(e){return 271===e.kind?Ld(HE(e.expression)):Te}function K_(e){var n=zt(e);if(!n.switchTypes){n.switchTypes=[];for(var t=0,r=e.caseBlock.clauses;t<r.length;t++){var a=r[t];n.switchTypes.push(B_(a))}}return n.switchTypes}function j_(n){for(var t=[],r=0,a=n.caseBlock.clauses;r<a.length;r++){var i=a[r];if(271===i.kind){if(10===i.expression.kind){t.push(i.expression.text);continue}return e.emptyArray}t.push(void 0)}return t}function U_(n,t){return 1048576&n.flags?!e.forEach(n.types,function(n){return!e.contains(t,n)}):e.contains(t,n)}function H_(e,n){return e===n||1048576&n.flags&&W_(e,n)}function W_(e,n){if(1048576&e.flags){for(var t=0,r=e.types;t<r.length;t++){var a=r[t];if(!vu(n.types,a))return!1}return!0}return!!(1024&e.flags&&To(e)===n)||vu(n.types,e)}function q_(n,t){return 1048576&n.flags?e.forEach(n.types,t):t(n)}function z_(n,t){return 1048576&n.flags?e.every(n.types,t):t(n)}function J_(n,t){if(1048576&n.flags){var r=n.types,a=e.filter(r,t);return a===r?n:ku(a,n.primitiveTypesOnly)}return t(n)?n:Te}function X_(e,n,t){if(131072&e.flags)return e;if(!(1048576&e.flags))return n(e);for(var r,a,i=e.types,o=0,s=i;o<s.length;o++){var c=s[o],l=n(c);l&&(r?a?a.push(l):a=[r,l]:r=l)}return a?Au(a,t?0:1):r}function Y_(e,n){return J_(e,function(e){return 0!==(e.flags&n)})}function Q_(e,n){return H_(pe,e)&&mE(n,128)||H_(me,e)&&mE(n,256)||H_(fe,e)&&mE(n,2048)?X_(e,function(e){return 4&e.flags?Y_(n,132):8&e.flags?Y_(n,264):64&e.flags?Y_(n,2112):e}):e}function Z_(e){return 0===e.flags}function $_(e){return 0===e.flags?e.type:e}function eg(e,n){return n?{flags:0,type:e}:e}function ng(e){var n=fa(256);return n.elementType=e,n}function tg(e){return $[e.id]||($[e.id]=ng(e))}function rg(e,n){var t=Wm(WE(n));return H_(t,e.elementType)?e:tg(Au([e.elementType,t]))}function ag(e){return 131072&e.flags?nn:su(1048576&e.flags?Au(e.types,2):e)}function ig(e){return e.finalArrayType||(e.finalArrayType=ag(e.elementType))}function og(n){return 256&e.getObjectFlags(n)?ig(n):n}function sg(n){return 256&e.getObjectFlags(n)?n.elementType:Te}function cg(n){for(var t=!1,r=0,a=n;r<a.length;r++){var i=a[r];if(!(131072&i.flags)){if(!(256&e.getObjectFlags(i)))return!1;t=!0}}return t}function lg(n,t){return cg(n)?tg(Au(e.map(n,sg))):Au(e.sameMap(n,og),t)}function ug(n){var t=V_(n),r=t.parent,a=189===r.kind&&("length"===r.name.escapedText||191===r.parent.kind&&e.isPushOrUnshiftIdentifier(r.name)),i=190===r.kind&&r.expression===t&&204===r.parent.kind&&59===r.parent.operatorToken.kind&&r.parent.left===r&&!e.isAssignmentTarget(r.parent)&&fE(HE(r.argumentExpression),296);return a||i}function dg(e){var n=zt(e);return void 0===n.maybeTypePredicate&&(n.maybeTypePredicate=pg(e)),n.maybeTypePredicate}function pg(n){if(98!==n.expression.kind){var t=Iy(n.expression);if(t!==Se){var r=rc(t);return r!==oe&&e.some(lc(r,0),Mc)}}return!1}function mg(n){var t=e.findAncestor(n,e.isFunctionOrModuleBlock),r=e.getSourceFileOfNode(n),a=e.getSpanOfTokenAtPosition(r,t.statements.pos);it.add(e.createFileDiagnostic(r,a.start,a.length,e.Diagnostics.The_containing_function_or_module_body_is_too_large_for_control_flow_analysis))}function fg(n,t,r,a,i){var o;void 0===r&&(r=t);var s=0;if(Gn)return oe;if(!n.flowNode||!i&&!(133970943&t.flags))return t;var c=Fn,l=$_(d(n.flowNode));Fn=c;var u=256&e.getObjectFlags(l)&&ug(n)?nn:og(l);return n.parent&&213===n.parent.kind&&131072&b_(u,2097152).flags?t:u;function d(e){if(2e3===s)return Gn=!0,mg(n),oe;s++;while(1){var i=e.flags;if(1024&i)for(var o=c;o<Fn;o++)if(et[o]===e)return s--,nt[o];var l=void 0;if(4096&i)e.locked=!0,l=d(e.antecedent),e.locked=!1;else{if(2048&i){e=e.antecedent;continue}if(16&i){if(l=p(e),!l){e=e.antecedent;continue}}else if(96&i)l=f(e);else if(128&i)l=_(e);else if(12&i){if(1===e.antecedents.length){e=e.antecedents[0];continue}l=4&i?g(e):v(e)}else if(256&i){if(l=m(e),!l){e=e.antecedent;continue}}else if(2&i){var u=e.container;if(u&&u!==a&&189!==n.kind&&190!==n.kind&&100!==n.kind){e=u.flowNode;continue}l=r}else l=xS(t)}return 1024&i&&(et[Fn]=e,nt[Fn]=l,Fn++),s--,l}}function p(r){var a=r.node;if(r_(n,a)){if(2===e.getAssignmentTargetKind(a)){var i=d(r.antecedent);return eg(Wm($_(i)),Z_(i))}if(t===ae||t===nn){if(F_(a))return tg(Te);var o=Wm(R_(a,n));return Lp(o,t)?o:en}return 1048576&t.flags?g_(t,R_(a,n)):t}if(i_(n,a)){if(e.isVariableDeclaration(a)&&(e.isInJSFile(a)||e.isVarConst(a))){var s=e.getDeclaredExpandoInitializer(a);if(s&&(196===s.kind||197===s.kind))return d(r.antecedent)}return t}if(e.isVariableDeclaration(a)&&226===a.parent.parent.kind&&r_(n,a.parent.parent.expression))return wy($_(d(r.antecedent)))}function m(r){if(t===ae||t===nn){var a=r.node,i=191===a.kind?a.expression.expression:a.left.expression;if(r_(n,G_(i))){var o=d(r.antecedent),s=$_(o);if(256&e.getObjectFlags(s)){var c=s;if(191===a.kind)for(var l=0,u=a.arguments;l<u.length;l++){var p=u[l];c=rg(c,p)}else{var m=WE(a.left.argumentExpression);fE(m,296)&&(c=rg(c,a.right))}return c===s?o:eg(c,Z_(o))}return o}}}function f(e){var n=d(e.antecedent),t=$_(n);if(131072&t.flags)return n;var r=0!==(32&e.flags),a=og(t),i=O(a,e.expression,r);if(i===a)return n;var o=Z_(n),s=o&&131072&i.flags?Se:i;return eg(s,o)}function _(e){var r=e.switchStatement.expression,a=d(e.antecedent),i=$_(a);return r_(n,r)?i=A(i,e.switchStatement,e.clauseStart,e.clauseEnd):y(r,i)?i=h(i,r,function(n){return A(n,e.switchStatement,e.clauseStart,e.clauseEnd)}):199===r.kind&&r_(n,r.expression)?i=k(i,e.switchStatement,e.clauseStart,e.clauseEnd):o_(n,r)&&(i=t),eg(i,Z_(a))}function g(n){for(var a=[],i=!1,o=!1,s=0,c=n.antecedents;s<c.length;s++){var l=c[s];if(!(2048&l.flags&&l.lock.locked)){var u=d(l),p=$_(u);if(p===t&&t===r)return p;e.pushIfUnique(a,p),H_(p,t)||(i=!0),Z_(u)&&(o=!0)}}return eg(lg(a,i?2:1),o)}function v(r){var a=f_(r),i=Yn[a]||(Yn[a]=e.createMap());if(!o&&(o=t_(n),!o))return t;var s=i.get(o);if(s)return s;for(var c=Pn;c<Rn;c++)if(Qn[c]===r&&Zn[c]===o&&$n[c].length)return eg(lg($n[c],1),!0);var l,u=[],p=!1;Qn[Rn]=r,Zn[Rn]=o,$n[Rn]=u;for(var m=0,f=r.antecedents;m<f.length;m++){var _=f[m];Rn++;var g=d(_);Rn--,l||(l=g);var v=$_(g),y=i.get(o);if(y)return y;if(e.pushIfUnique(u,v),H_(v,t)||(p=!0),v===t)break}var h=lg(u,p?2:1);return Z_(l)?eg(h,!0):(i.set(o,h),h)}function y(t,r){if(!(1048576&r.flags)||!e.isAccessExpression(t))return!1;var a=a_(t);return void 0!==a&&(r_(n,t.expression)&&l_(r,a))}function h(e,n,t){var r=a_(n);if(void 0===r)return e;var a=ti(e,r),i=a&&t(a);return a===i?e:J_(e,function(e){return Np(ri(e,r),i)})}function b(e,r,a){return r_(n,r)?b_(e,a?4194304:8388608):y(r,t)?h(e,r,function(e){return b_(e,a?4194304:8388608)}):o_(n,r)?t:e}function E(e,n,t){if(pc(e,0))return!0;var r=sc(e,n);return r?!!(16777216&r.flags)||t:!t}function T(n,t,r){if(1572864&n.flags||262144&n.flags&&n.isThisType){var a=e.escapeLeadingUnderscores(t.text);return J_(n,function(e){return E(e,a,r)})}return n}function S(r,a,i){switch(a.operatorToken.kind){case 59:return b(r,a.left,i);case 33:case 34:case 35:case 36:var o=a.operatorToken.kind,s=G_(a.left),c=G_(a.right);if(199===s.kind&&e.isStringLiteralLike(c))return C(r,s,o,c,i);if(199===c.kind&&e.isStringLiteralLike(s))return C(r,c,o,s,i);if(r_(n,s))return x(r,o,c,i);if(r_(n,c))return x(r,o,s,i);if(y(s,t))return h(r,s,function(e){return x(e,o,c,i)});if(y(c,t))return h(r,c,function(e){return x(e,o,s,i)});if(o_(n,s)||o_(n,c))return t;break;case 94:return N(r,a,i);case 93:var l=G_(a.right);if(e.isStringLiteralLike(a.left)&&r_(n,l))return T(r,a.left,i);break;case 27:return O(r,a.right,i)}return r}function x(e,n,t,r){if(1&e.flags)return e;34!==n&&36!==n||(r=!r);var a=HE(t);if(98304&a.flags){if(!w)return e;var i=33===n||34===n,o=i?r?262144:2097152:65536&a.flags?r?131072:1048576:r?65536:524288;return b_(e,o)}if(67637251&e.flags)return e;if(r){var s=J_(e,function(e){return Ip(e,a)});return 131072&s.flags?e:Q_(s,a)}if(Um(a)){var c=Ld(a);return J_(e,function(e){return Ld(e)!==c})}return e}function C(e,r,a,i,o){var s=G_(r.expression);if(!r_(n,s))return i_(n,s)&&u_(s)?t:e;if(34!==a&&36!==a||(o=!o),1&e.flags&&"function"===i.text)return e;var c=o?gt.get(i.text)||128:vt.get(i.text)||32768;return b_(o?X_(e,l):e,c);function l(e){if(2&e.flags&&"object"===i.text)return Au([Ce,ue]);var n="function"===i.text?He:yt.get(i.text);if(n){if(Dp(e,n))return e;if(Dp(n,e))return n;if(63176704&e.flags){var t=qs(e)||re;if(Dp(n,t))return Ru([e,n])}}return e}}function A(n,t,r,a){var i=K_(t);if(!i.length)return n;var o=i.slice(r,a),s=r===a||e.contains(o,Te),c=Au(o),l=131072&c.flags?Te:Q_(J_(n,function(e){return Ip(c,e)}),c);if(!s)return l;var u=J_(n,function(n){return!(Um(n)&&e.contains(i,Ld(n)))});return 131072&l.flags?u:Au([l,u])}function D(e,n){switch(n){case"function":return 1&e.flags?e:He;case"object":return 2&e.flags?Au([Ce,ue]):e;default:return yt.get(n)||e}}function L(e){return function(n){if(Dp(e,n))return e;if(63176704&n.flags){var t=qs(n)||re;if(Dp(e,t))return Ru([n,e])}return n}}function k(n,t,r,a){var i=j_(t);if(!i.length)return n;var o,s,c=e.findIndex(i,function(e){return void 0===e}),l=r===a||c>=r&&c<a;if(c>-1){var u=i.filter(function(e){return void 0!==e}),d=c<r?r-1:r,p=c<a?a-1:a;o=u.slice(d,p),s=Wb(d,p,u,l)}else o=i.slice(r,a),s=Wb(r,a,i,l);if(l)return J_(n,function(e){return(h_(e)&s)===s});var m=b_(Au(o.map(function(e){return D(n,e)})),s);return 1048576&m.flags&&(m=g_(m,zs(n))),b_(X_(n,L(m)),s)}function N(r,a,i){var o=G_(a.left);if(!r_(n,o))return i_(n,o)&&u_(o)?t:r;var s,c=HE(a.right);if(!kp(c,He))return r;var l=sc(c,"prototype");if(l){var u=Xi(l);ai(u)||(s=u)}if(ai(r)&&(s===Ue||s===He))return r;if(!s){var d=lc(c,1);s=d.length?Au(e.map(d,function(e){return Fc(qc(e))})):ke}return I(r,s,i,kp)}function I(e,n,t,r){if(!t)return J_(e,function(e){return!r(e,n)});if(1048576&e.flags){var a=J_(e,function(e){return r(e,n)});if(!(131072&a.flags))return a}return Dp(n,e)?n:Lp(e,n)?e:Lp(n,e)?n:Ru([e,n])}function M(r,a,i){if(!m_(a,n)||!dg(a))return r;var o=ab(a),s=Oc(o);if(!s)return r;if(ai(r)&&(s.type===Ue||s.type===He))return r;if(e.isIdentifierTypePredicate(s)){var c=a.arguments[s.parameterIndex-(o.thisParameter?1:0)];if(c){if(r_(n,c))return I(r,s.type,i,Dp);if(i_(n,c))return t}}else{var l=e.skipParentheses(a.expression);if(e.isAccessExpression(l)){var u=e.skipParentheses(l.expression);if(r_(n,u))return I(r,s.type,i,Dp);if(i_(n,u))return t}}return r}function O(e,n,t){switch(n.kind){case 72:case 100:case 98:case 189:case 190:return b(e,n,t);case 191:return M(e,n,t);case 195:return O(e,n.expression,t);case 204:return S(e,n,t);case 202:if(52===n.operator)return O(e,n.operand,!t);break}return e}}function _g(n,t){if(n=n.exportSymbol||n,72===t.kind&&(e.isRightSideOfQualifiedNameOrPropertyAccess(t)&&(t=t.parent),e.isExpressionNode(t)&&!e.isAssignmentTarget(t))){var r=HE(t);if(ca(zt(t).resolvedSymbol)===n)return r}return Xi(n)}function gg(n){return e.findAncestor(n.parent,function(n){return e.isFunctionLike(n)&&!e.getImmediatelyInvokedFunctionExpression(n)||245===n.kind||279===n.kind||154===n.kind})}function vg(n){var t=e.getRootDeclaration(n.valueDeclaration).parent,r=zt(t);return 8388608&r.flags||(r.flags|=8388608,yg(t)||hg(t)),n.isAssigned||!1}function yg(n){return!!e.findAncestor(n.parent,function(n){return e.isFunctionLike(n)&&!!(8388608&zt(n).flags)})}function hg(n){if(72===n.kind){if(e.isAssignmentTarget(n)){var t=e_(n);t.valueDeclaration&&151===e.getRootDeclaration(t.valueDeclaration).kind&&(t.isAssigned=!0)}}else e.forEachChild(n,hg)}function bg(e){return 3&e.flags&&0!==(2&Ay(e))&&Xi(e)!==nn}function Eg(e,n){var t=w&&151===n.kind&&n.initializer&&32768&nf(e)&&!(32768&nf(qE(n.initializer)));return t?b_(e,524288):e}function Tg(e){var n=e.parent;return 189===n.kind||191===n.kind&&n.expression===e||190===n.kind&&n.expression===e||186===n.kind&&n.name===e&&!!n.initializer}function Sg(e){return 58982400&e.flags&&mE(qs(e)||ke,98304)}function xg(e,n){return e&&Tg(n)&&q_(e,Sg)?X_(bf(e),zs):e}function Cg(e,n){!Lr(e,67220415)||n_(n)||HC(Nr(e))||wr(e)}function Ag(n){var t=e_(n);if(t===ne)return oe;if(t===H){var r=e.getContainingFunction(n);return k<2&&(197===r.kind?It(n,e.Diagnostics.The_arguments_object_cannot_be_referenced_in_an_arrow_function_in_ES3_and_ES5_Consider_using_a_standard_function_expression):e.hasModifier(r,256)&&It(n,e.Diagnostics.The_arguments_object_cannot_be_referenced_in_an_async_function_or_method_in_ES3_and_ES5_Consider_using_a_standard_function_or_method)),zt(r).flags|=8192,Xi(t)}n.parent&&e.isPropertyAccessExpression(n.parent)&&n.parent.expression===n||Cg(t,n);var a=ca(t),i=a.valueDeclaration;if(32&a.flags)if(240===i.kind&&e.nodeIsDecorated(i)){r=e.getContainingClass(n);while(void 0!==r){if(r===i&&r.name!==n){zt(i).flags|=16777216,zt(n).flags|=33554432;break}r=e.getContainingClass(r)}}else if(209===i.kind){r=e.getThisContainer(n,!1);while(279!==r.kind){if(r.parent===i){154===r.kind&&e.hasModifier(r,32)&&(zt(i).flags|=16777216,zt(n).flags|=33554432);break}r=e.getThisContainer(r,!1)}}kg(n,t);var o=xg(Xi(a),n),s=e.getAssignmentTargetKind(n);if(s){if(!(3&a.flags)&&!(e.isInJSFile(n)&&512&a.flags))return It(n,e.Diagnostics.Cannot_assign_to_0_because_it_is_not_a_variable,Pa(t)),oe;if(tE(a))return 3&a.flags?It(n,e.Diagnostics.Cannot_assign_to_0_because_it_is_a_constant,Pa(t)):It(n,e.Diagnostics.Cannot_assign_to_0_because_it_is_a_read_only_property,Pa(t)),oe}var c=2097152&a.flags;if(3&a.flags){if(1===s)return o}else{if(!c)return o;i=e.find(t.declarations,f)}if(!i)return o;var l=151===e.getRootDeclaration(i).kind,u=gg(i),d=gg(n),p=d!==u,m=n.parent&&n.parent.parent&&e.isSpreadAssignment(n.parent)&&A_(n.parent.parent),_=134217728&t.flags;while(d!==u&&(196===d.kind||197===d.kind||e.isObjectLiteralOrClassExpressionMethod(d))&&(bg(a)||l&&!vg(a)))d=gg(d);var g=l||c||p||m||_||o!==ae&&o!==nn&&(!w||0!==(3&o.flags)||n_(n)||257===n.parent.kind)||213===n.parent.kind||237===i.kind&&i.exclamationToken||4194304&i.flags,v=g?l?Eg(o,i):o:o===ae||o===nn?ce:sf(o),y=fg(n,o,v,d,!g);if(ug(n)||o!==ae&&o!==nn){if(!g&&!(32768&nf(o))&&32768&nf(y))return It(n,e.Diagnostics.Variable_0_is_used_before_being_assigned,Pa(t)),o}else if(y===ae||y===nn)return R&&(It(e.getNameOfDeclaration(i),e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined,Pa(t),Fa(y)),It(n,e.Diagnostics.Variable_0_implicitly_has_an_1_type,Pa(t),Fa(y))),xS(y);return s?Wm(y):y}function Dg(n,t){return!!e.findAncestor(n,function(n){return n===t?"quit":e.isFunctionLike(n)})}function Lg(n,t){return e.findAncestor(n,function(e){return e===t?"quit":e===t.initializer||e===t.condition||e===t.incrementor||e===t.statement})}function kg(n,t){if(!(k>=2||0===(34&t.flags)||274===t.valueDeclaration.parent.kind)){var r=e.getEnclosingBlockScopeContainer(t.valueDeclaration),a=Dg(n.parent,r),i=r,o=!1;while(i&&!e.nodeStartsNewLexicalEnvironment(i)){if(e.isIterationStatement(i,!1)){o=!0;break}i=i.parent}if(o){if(a){var s=!0;if(e.isForStatement(r)&&e.getAncestor(t.valueDeclaration,238).parent===r){var c=Lg(n.parent,r);if(c){var l=zt(c);l.flags|=131072;var u=l.capturedBlockScopeBindings||(l.capturedBlockScopeBindings=[]);e.pushIfUnique(u,t),c===r.initializer&&(s=!1)}}s&&(zt(i).flags|=65536)}225===r.kind&&e.getAncestor(t.valueDeclaration,238).parent===r&&Ig(n,r)&&(zt(t.valueDeclaration).flags|=4194304),zt(t.valueDeclaration).flags|=524288}a&&(zt(t.valueDeclaration).flags|=262144)}}function Ng(n,t){var r=zt(n);return!!r&&e.contains(r.capturedBlockScopeBindings,ra(t))}function Ig(n,t){var r=n;while(195===r.parent.kind)r=r.parent;var a=!1;if(e.isAssignmentTarget(r))a=!0;else if(202===r.parent.kind||203===r.parent.kind){var i=r.parent;a=44===i.operator||45===i.operator}return!!a&&!!e.findAncestor(r,function(e){return e===t?"quit":e===t.statement})}function wg(e,n){if(zt(e).flags|=2,154===n.kind||157===n.kind){var t=n.parent;zt(t).flags|=4}else zt(n).flags|=4}function Mg(n){return e.isSuperCall(n)?n:e.isFunctionLike(n)?void 0:e.forEachChild(n,Mg)}function Og(e){var n=zt(e);return void 0===n.hasSuperCall&&(n.superCall=Mg(e.body),n.hasSuperCall=!!n.superCall),n.superCall}function Pg(e){var n=ra(e),t=Do(n),r=lo(t);return r===de}function Rg(n,t,r){var a=t.parent,i=e.getClassExtendsHeritageElement(a);if(i&&!Pg(a)){var o=Og(t);(!o||o.end>n.pos)&&It(n,r)}}function Fg(n){var t=e.getThisContainer(n,!0),r=!1;switch(157===t.kind&&Rg(n,t,e.Diagnostics.super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class),197===t.kind&&(t=e.getThisContainer(t,!1),r=!0),t.kind){case 244:It(n,e.Diagnostics.this_cannot_be_referenced_in_a_module_or_namespace_body);break;case 243:It(n,e.Diagnostics.this_cannot_be_referenced_in_current_location);break;case 157:Kg(n,t)&&It(n,e.Diagnostics.this_cannot_be_referenced_in_constructor_arguments);break;case 154:case 153:e.hasModifier(t,32)&&It(n,e.Diagnostics.this_cannot_be_referenced_in_a_static_property_initializer);break;case 149:It(n,e.Diagnostics.this_cannot_be_referenced_in_a_computed_property_name);break}r&&k<2&&wg(n,t);var a=Gg(n,t);if(!a&&F){var i=It(n,r&&279===t.kind?e.Diagnostics.The_containing_arrow_function_captures_the_global_value_of_this_which_implicitly_has_type_any:e.Diagnostics.this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation);if(!e.isSourceFile(t)){var o=Gg(t);o&&Nt(i,e.createDiagnosticForNode(t,e.Diagnostics.An_outer_value_of_this_is_shadowed_by_this_container))}}return a||re}function Gg(n,t){void 0===t&&(t=e.getThisContainer(n,!1));var r=e.isInJSFile(n);if(e.isFunctionLike(t)&&(!ev(n)||e.getThisParameter(t))){var a=Vg(t);if(r&&a){var i=qE(a).symbol;if(i&&i.members&&16&i.flags){var o=sb(i);if(o)return fg(n,o)}}else if(r&&(196===t.kind||239===t.kind)&&e.getJSDocClassTag(t)){o=sb(ta(t.symbol));if(o)return fg(n,o)}var s=Vi(t)||qg(t);if(s)return fg(n,s)}if(e.isClassLike(t.parent)){var c=ra(t.parent),l=e.hasModifier(t,32)?Xi(c):Do(c).thisType;return fg(n,l)}if(r){l=Bg(t);if(l&&l!==oe)return fg(n,l)}}function Vg(n){return 196===n.kind&&e.isBinaryExpression(n.parent)&&3===e.getAssignmentDeclarationKind(n.parent)?n.parent.left.expression.expression:156===n.kind&&188===n.parent.kind&&e.isBinaryExpression(n.parent.parent)&&6===e.getAssignmentDeclarationKind(n.parent.parent)?n.parent.parent.left.expression:196===n.kind&&275===n.parent.kind&&188===n.parent.parent.kind&&e.isBinaryExpression(n.parent.parent.parent)&&6===e.getAssignmentDeclarationKind(n.parent.parent.parent)?n.parent.parent.parent.left.expression:196===n.kind&&e.isPropertyAssignment(n.parent)&&e.isIdentifier(n.parent.name)&&("value"===n.parent.name.escapedText||"get"===n.parent.name.escapedText||"set"===n.parent.name.escapedText)&&e.isObjectLiteralExpression(n.parent.parent)&&e.isCallExpression(n.parent.parent.parent)&&n.parent.parent.parent.arguments[2]===n.parent.parent&&9===e.getAssignmentDeclarationKind(n.parent.parent.parent)?n.parent.parent.parent.arguments[0].expression:e.isMethodDeclaration(n)&&e.isIdentifier(n.name)&&("value"===n.name.escapedText||"get"===n.name.escapedText||"set"===n.name.escapedText)&&e.isObjectLiteralExpression(n.parent)&&e.isCallExpression(n.parent.parent)&&n.parent.parent.arguments[2]===n.parent&&9===e.getAssignmentDeclarationKind(n.parent.parent)?n.parent.parent.arguments[0].expression:void 0}function Bg(n){var t=e.getJSDocType(n);if(t&&289===t.kind){var r=t;if(r.parameters.length>0&&r.parameters[0].name&&"this"===r.parameters[0].name.escapedText)return Rd(r.parameters[0].type)}var a=e.getJSDocThisTag(n);if(a&&a.typeExpression)return Rd(a.typeExpression)}function Kg(n,t){return!!e.findAncestor(n,function(e){return e===t?"quit":151===e.kind})}function jg(n){var t=191===n.parent.kind&&n.parent.expression===n,r=e.getSuperContainer(n,!0),a=!1;if(!t)while(r&&197===r.kind)r=e.getSuperContainer(r,!0),a=k<2;var i=d(r),o=0;if(!i){var s=e.findAncestor(n,function(e){return e===r?"quit":149===e.kind});return s&&149===s.kind?It(n,e.Diagnostics.super_cannot_be_referenced_in_a_computed_property_name):t?It(n,e.Diagnostics.Super_calls_are_not_permitted_outside_constructors_or_in_nested_functions_inside_constructors):r&&r.parent&&(e.isClassLike(r.parent)||188===r.parent.kind)?It(n,e.Diagnostics.super_property_access_is_permitted_only_in_a_constructor_member_function_or_member_accessor_of_a_derived_class):It(n,e.Diagnostics.super_can_only_be_referenced_in_members_of_derived_classes_or_object_literal_expressions),oe}if(t||157!==r.kind||Rg(n,r,e.Diagnostics.super_must_be_called_before_accessing_a_property_of_super_in_the_constructor_of_a_derived_class),o=e.hasModifier(r,32)||t?512:256,zt(n).flags|=o,156===r.kind&&e.hasModifier(r,256)&&(e.isSuperProperty(n.parent)&&e.isAssignmentTarget(n.parent)?zt(r).flags|=4096:zt(r).flags|=2048),a&&wg(n.parent,r),188===r.parent.kind)return k<2?(It(n,e.Diagnostics.super_is_only_allowed_in_members_of_object_literal_expressions_when_option_target_is_ES2015_or_higher),oe):re;var c=r.parent;if(!e.getClassExtendsHeritageElement(c))return It(n,e.Diagnostics.super_can_only_be_referenced_in_a_derived_class),oe;var l=Do(ra(c)),u=l&&uo(l)[0];return u?157===r.kind&&Kg(n,r)?(It(n,e.Diagnostics.super_cannot_be_referenced_in_constructor_arguments),oe):512===o?lo(l):Xo(u,l.thisType):oe;function d(n){return!!n&&(t?157===n.kind:!(!e.isClassLike(n.parent)&&188!==n.parent.kind)&&(e.hasModifier(n,32)?156===n.kind||155===n.kind||158===n.kind||159===n.kind:156===n.kind||155===n.kind||158===n.kind||159===n.kind||154===n.kind||153===n.kind||157===n.kind))}}function Ug(e){return 156!==e.kind&&158!==e.kind&&159!==e.kind||188!==e.parent.kind?196===e.kind&&275===e.parent.kind?e.parent.parent:void 0:e.parent}function Hg(n){return 4&e.getObjectFlags(n)&&n.target===$e?n.typeArguments[0]:void 0}function Wg(n){return X_(n,function(n){return 2097152&n.flags?e.forEach(n.types,Hg):Hg(n)})}function qg(n){if(197!==n.kind){if(Ep(n)){var t=Ov(n);if(t){var r=t.thisParameter;if(r)return Xi(r)}}var a=e.isInJSFile(n);if(F||a){var i=Ug(n);if(i){var o=hv(i),s=i,c=o;while(c){var l=Wg(c);if(l)return mp(l,Ev(i));if(275!==s.parent.kind)break;s=s.parent.parent,c=hv(s)}return o?lf(o):RE(i)}var u=n.parent;if(204===u.kind&&59===u.operatorToken.kind){var d=u.left;if(189===d.kind||190===d.kind){var p=d.expression;if(a&&e.isIdentifier(p)){var m=e.getSourceFileOfNode(u);if(m.commonJsModuleIndicator&&e_(p)===m.symbol)return}return RE(p)}}}}}function zg(n){var t=n.parent;if(Ep(t)){var r=e.getImmediatelyInvokedFunctionExpression(t);if(r&&r.arguments){var a=Lh(r),i=t.parameters.indexOf(n);if(n.dotDotDotToken)return Eh(a,i,a.length,re,void 0);var o=zt(r),s=o.resolvedSignature;o.resolvedSignature=Dn;var c=i<a.length?qm(qE(a[i])):n.initializer?void 0:le;return o.resolvedSignature=s,c}var l=Ov(t);if(l){var u=t.parameters.indexOf(n)-(e.getThisParameter(t)?1:0);return n.dotDotDotToken&&e.lastOrUndefined(t.parameters)===n?Db(l,u):Ab(l,u)}}}function Jg(n){var t=e.getEffectiveTypeAnnotationNode(n);if(t)return Rd(t);switch(n.kind){case 151:return zg(n);case 186:return Xg(n)}}function Xg(n){var t=n.parent.parent,r=n.propertyName||n.name,a=Jg(t);if(a&&!e.isBindingPattern(r)&&!oi(r)){var i=Bu(r);if(Fo(i)){var o=Uo(i);return ti(a,o)}}}function Yg(n){var t=n.parent;if(e.hasInitializer(t)&&n===t.initializer){var r=Jg(t);if(r)return r;if(e.isBindingPattern(t.name))return ki(t.name,!0,!1)}}function Qg(n){var t=e.getContainingFunction(n);if(t){var r=e.getFunctionFlags(t);if(1&r)return;var a=nv(t);if(a){if(2&r){var i=MT(a);return i&&Au([i,Bb(i)])}return a}}}function Zg(e){var n=bv(e);if(n){var t=RT(n);return t&&Au([t,Bb(t)])}}function $g(n){var t=e.getContainingFunction(n);if(t){var r=e.getFunctionFlags(t),a=nv(t);if(a)return n.asteriskToken?a:qS(a,0!==(2&r))}}function ev(n){var t=!1;while(n.parent&&!e.isFunctionLike(n.parent)){if(e.isParameter(n.parent)&&(t||n.parent.initializer===n))return!0;e.isBindingElement(n.parent)&&n.parent.initializer===n&&(t=!0),n=n.parent}return!1}function nv(e){var n=Gc(e);if(n)return n;var t=wv(e);return t&&!Vc(t)?Fc(t):void 0}function tv(e,n){var t=Lh(e),r=t.indexOf(n);return-1===r?void 0:rv(e,r)}function rv(n,t){var r=zt(n).resolvedSignature===kn?kn:ab(n);return e.isJsxOpeningLikeElement(n)&&0===t?Sv(r,n):Cb(r,t)}function av(e,n){if(193===e.parent.kind)return tv(e.parent,n)}function iv(n){var t=n.parent,r=t.left,a=t.operatorToken,i=t.right;switch(a.kind){case 59:if(n!==i)return;var o=ov(t);if(!o)return;return!0===o?HE(r):o;case 55:var s=bv(t);return s||n!==i||e.isDefaultedExpandoInitializer(t)?s:HE(r);case 54:case 27:return n===i?bv(t):void 0;default:return}}function ov(n){var t=e.getAssignmentDeclarationKind(n);switch(t){case 0:return!0;case 5:case 1:case 6:case 3:if(n.left.symbol){var r=n.left.symbol.valueDeclaration;if(!r)return!1;var a=n.left,i=e.getEffectiveTypeAnnotationNode(r);if(i)return Rd(i);if(e.isIdentifier(a.expression)){var o=a.expression,s=Zt(o,o.escapedText,67220415,void 0,o.escapedText,!0);if(s){var c=e.getEffectiveTypeAnnotationNode(s.valueDeclaration);if(c){var l=sv(Rd(c),a.name.escapedText);return l||!1}return!1}}return!e.isInJSFile(r)}return!0;case 2:case 4:if(!n.symbol)return!0;if(n.symbol.valueDeclaration){c=e.getEffectiveTypeAnnotationNode(n.symbol.valueDeclaration);if(c){l=Rd(c);if(l)return l}}if(2===t)return!1;var u=n.left;if(!e.isObjectLiteralMethod(e.getThisContainer(u.expression,!1)))return!1;var d=Fg(u.expression);return d&&sv(d,u.name.escapedText)||!1;case 7:case 8:case 9:return e.Debug.fail("Does not apply");default:return e.Debug.assertNever(t)}}function sv(e,n){return X_(e,function(e){if(3670016&e.flags){var t=sc(e,n);if(t)return Xi(t);if(Xm(e)){var r=Ym(e);if(r&&Uv(n)&&+n>=0)return r}return Uv(n)&&cv(e,1)||cv(e,0)}},!0)}function cv(e,n){return X_(e,function(e){return dc(e,n)},!0)}function lv(n){if(e.Debug.assert(e.isObjectLiteralMethod(n)),!(8388608&n.flags))return uv(n)}function uv(e){var n=e.parent,t=hv(n);if(t){if(!Ko(e)){var r=ra(e).escapedName,a=sv(t,r);if(a)return a}return Bv(e.name)&&cv(t,1)||cv(t,0)}}function dv(e,n){return e&&(sv(e,""+n)||jS(e,void 0,!1,!1,!1))}function pv(e){var n=e.parent;return e===n.whenTrue||e===n.whenFalse?bv(n):void 0}function mv(e,n){var t=hv(e.openingElement.tagName),r=py(cy(e));if(t&&!ai(t)&&r&&""!==r){var a=e.children.indexOf(n),i=sv(t,r);return i&&X_(i,function(e){return Fm(e)?ad(e,Nd(a)):e},!0)}}function fv(n){var t=n.parent;return e.isJsxAttributeLike(t)?bv(n):e.isJsxElement(t)?mv(t,n):void 0}function _v(n){if(e.isJsxAttribute(n)){var t=hv(n.parent);if(!t||ai(t))return;return sv(t,n.name.escapedText)}return bv(n.parent)}function gv(e){switch(e.kind){case 10:case 8:case 9:case 14:case 102:case 87:case 96:case 72:case 141:return!0;case 189:case 195:return gv(e.expression);case 270:return!e.expression||gv(e.expression)}return!1}function vv(n,t){return sm(t,e.map(e.filter(n.properties,function(e){return!!e.symbol&&275===e.kind&&gv(e.initializer)&&l_(t,e.symbol.escapedName)}),function(e){return[function(){return qE(e.initializer)},e.symbol.escapedName]}),Lp,t)}function yv(n,t){return sm(t,e.map(e.filter(n.properties,function(e){return!!e.symbol&&267===e.kind&&l_(t,e.symbol.escapedName)&&(!e.initializer||gv(e.initializer))}),function(e){return[e.initializer?function(){return qE(e.initializer)}:function(){return ve},e.symbol.escapedName]}),Lp,t)}function hv(n){var t=bv(n);if(t=t&&X_(t,rc),t&&1048576&t.flags){if(e.isObjectLiteralExpression(n))return vv(n,t);if(e.isJsxAttributes(n))return yv(n,t)}return t}function bv(n){if(!(8388608&n.flags)){if(n.contextualType)return n.contextualType;var t=n.parent;switch(t.kind){case 237:case 151:case 154:case 153:case 186:return Yg(n);case 197:case 230:return Qg(n);case 207:return $g(t);case 201:return Zg(t);case 191:case 192:return tv(t,n);case 194:case 212:return Rd(t.type);case 204:return iv(n);case 275:case 276:return uv(t);case 277:return hv(t.parent);case 187:var r=t,a=hv(r);return dv(a,e.indexOfNode(r.elements,n));case 205:return pv(n);case 216:return e.Debug.assert(206===t.parent.kind),av(t.parent,n);case 195:var i=e.isInJSFile(t)?e.getJSDocTypeTag(t):void 0;return i?Rd(i.typeExpression.type):bv(t);case 270:return fv(t);case 267:case 269:return _v(t);case 262:case 261:return Tv(t)}}}function Ev(n){var t=e.findAncestor(n,function(e){return!!e.contextualMapper});return t?t.contextualMapper:D}function Tv(n){return e.isJsxOpeningElement(n)&&n.parent.contextualType?n.parent.contextualType:rv(n,0)}function Sv(e,n){return 0!==Sh(n)?xv(e,n):Lv(e,n)}function xv(e,t){var r=Ob(e,ke);r=Dv(t,cy(t),r);var a=oy(n.IntrinsicAttributes,t);return a!==oe&&(r=ds(a,r)),r}function Cv(e,n){if(e.unionSignatures){for(var t=[],r=0,a=e.unionSignatures;r<a.length;r++){var i=a[r],o=Fc(i);if(ai(o))return o;var s=ti(o,n);if(!s)return;t.push(s)}return Ru(t)}var c=Fc(e);return ai(c)?c:ti(c,n)}function Av(e){if(ny(e.tagName)){var n=gy(e),t=eb(e,n);return Qc(t)}var r=RE(e.tagName);if(128&r.flags){n=fy(r,e);if(!n)return oe;t=eb(e,n);return Qc(t)}return r}function Dv(n,t,r){var a=uy(t);if(a){var i=Do(a),o=Av(n);if(e.length(i.typeParameters)>=2){var s=xc([o,r],i.typeParameters,2,e.isInJSFile(n));return cl(i,s)}if(e.length(i.aliasTypeArguments)>=2){s=xc([o,r],i.aliasTypeArguments,2,e.isInJSFile(n));return pl(i.aliasSymbol,s)}}return r}function Lv(t,r){var a=cy(r),i=dy(a),o=void 0===i?Ob(t,ke):""===i?Fc(t):Cv(t,i);if(!o)return i&&e.length(r.attributes.properties)&&It(r,e.Diagnostics.JSX_element_class_does_not_support_attributes_because_it_does_not_have_a_0_property,e.unescapeLeadingUnderscores(i)),ke;if(o=Dv(r,a,o),ai(o))return o;var s=o,c=oy(n.IntrinsicClassAttributes,r);if(c!==oe){var l=to(c.symbol),u=Fc(t);s=ds(l?cl(c,xc([u],l,Sc(l),e.isInJSFile(r))):c,s)}var d=oy(n.IntrinsicAttributes,r);return d!==oe&&(s=ds(d,s)),s}function kv(e,n){var t=lc(e,0);if(1===t.length){var r=t[0];if(!Nv(r,n))return r}}function Nv(n,t){for(var r=0;r<t.parameters.length;r++){var a=t.parameters[r];if(a.initializer||a.questionToken||a.dotDotDotToken||vc(a))break}return t.parameters.length&&e.parameterIsThisKeyword(t.parameters[0])&&r--,!Nb(n)&&Lb(n)<r}function Iv(e){return 196===e.kind||197===e.kind}function wv(n){return Iv(n)||e.isObjectLiteralMethod(n)?Ov(n):void 0}function Mv(n){return e.isObjectLiteralMethod(n)?lv(n):hv(n)}function Ov(n){e.Debug.assert(156!==n.kind||e.isObjectLiteralMethod(n));var t=Dc(n);if(t)return t;var r=Mv(n);if(r){if(!(1048576&r.flags))return kv(r,n);for(var a,i,o=r.types,s=0,c=o;s<c.length;s++){var l=c[s],u=kv(l,n);if(u)if(a){if(!Lm(a[0],u,!1,!0,!0,xp))return;a.push(u)}else a=[u]}return a&&(i=es(a[0]),i.unionSignatures=a),i}}function Pv(e,n){k<2&&L.downlevelIteration&&_A(e,1536);var t=qE(e.expression,n);return KS(t,e.expression,!1,!1)}function Rv(e){return 186===e.kind&&!!e.initializer||204===e.kind&&59===e.operatorToken.kind}function Fv(n,t,r){for(var a=n.elements,i=a.length,o=!1,s=[],c=e.isAssignmentTarget(n),l=hv(n),u=0;u<i;u++){var d=a[u];if(c&&208===d.kind){var p=qE(d.expression,t,r),m=mc(p,1)||jS(p,void 0,!1,!1,!1);m&&s.push(m)}else{var f=dv(l,u),_=BE(d,t,f,r);s.push(_)}u<i-1&&208===d.kind&&(o=!0)}if(!o){var g=i>0&&208===a[i-1].kind,v=i-(g?1:0),y=void 0;if(c&&v>0){_=ll(pu(s,v,g));return _.pattern=n,_}if(y=Gv(s,l,g,i))return y;if(r)return pu(s,v,g)}return Vv(s,2)}function Gv(n,t,r,a){if(void 0===a&&(a=n.length),t&&q_(t,Vm)){var i=a-(r?1:0),o=t.pattern;if(!r&&o&&(185===o.kind||187===o.kind))for(var s=o.elements,c=a;c<s.length;c++){var l=s[c];Rv(l)?n.push(t.typeArguments[c]):(c<s.length-1||!(186===l.kind&&l.dotDotDotToken||208===l.kind))&&(210!==l.kind&&It(l,e.Diagnostics.Initializer_provides_no_value_for_this_binding_element_and_the_binding_element_has_no_default_value),n.push(w?xe:le))}return pu(n,i,r)}}function Vv(e,n){return void 0===n&&(n=1),su(e.length?Au(e,n):w?xe:le)}function Bv(e){switch(e.kind){case 149:return Kv(e);case 72:return Uv(e.escapedText);case 8:case 10:return Uv(e.text);default:return!1}}function Kv(e){return fE(Hv(e),296)}function jv(e){return"Infinity"===e||"-Infinity"===e||"NaN"===e}function Uv(e){return(+e).toString()===e}function Hv(n){var t=zt(n.expression);return t.resolvedType||(t.resolvedType=qE(n.expression),98304&t.resolvedType.flags||!fE(t.resolvedType,12716)&&!Lp(t.resolvedType,Ae)?It(n,e.Diagnostics.A_computed_property_name_must_be_of_type_string_number_symbol_or_any):oh(n.expression,t.resolvedType,!0)),t.resolvedType}function Wv(e,n,t,r){for(var a=[],i=0;i<t.length;i++)(0===r||Bv(e[i+n].name))&&a.push(Xi(t[i]));var o=a.length?Au(a,2):ce;return el(o,!1)}function qv(n){e.Debug.assert(0!==(2097152&n.flags),"Should only get Alias here.");var t=qt(n);if(!t.immediateTarget){var r=pr(n);if(!r)return e.Debug.fail();t.immediateTarget=Dr(r,!0)}return t.immediateTarget}function zv(n,t){var r,a=e.isAssignmentTarget(n);zA(n,a);var i=[],o=ke,s=0,c=hv(n),l=c&&c.pattern&&(184===c.pattern.kind||188===c.pattern.kind),u=e.isInJSFile(n)&&!e.isInJsonFile(n),d=e.getJSDocEnumTag(n),p=!c&&u&&!d,m=0,f=!1,_=!1,g=!1;r=e.createSymbolTable();for(var v=0,y=0;y<n.properties.length;y++){var h=n.properties[y],b=ra(h),E=h.name&&149===h.name.kind&&!e.isWellKnownSymbolSyntactically(h.name.expression)?Hv(h.name):void 0;if(275===h.kind||276===h.kind||e.isObjectLiteralMethod(h)){var T=275===h.kind?KE(h,t):276===h.kind?BE(h.name,t):jE(h,t);if(u){var S=fi(h);S?(wp(T,S,h),T=S):d&&d.typeExpression&&wp(T,Rd(d.typeExpression),h)}m|=T.flags;var x=E&&Fo(E)?E:void 0,C=x?Ot(4|b.flags,Uo(x),2048):Ot(4|b.flags,b.escapedName);if(x&&(C.nameType=x),a){var A=275===h.kind&&Rv(h.initializer)||276===h.kind&&h.objectAssignmentInitializer;A&&(C.flags|=16777216)}else if(l&&!(512&e.getObjectFlags(c))){var D=sc(c,b.escapedName);D?C.flags|=16777216&D.flags:L.suppressExcessPropertyErrors||pc(c,0)||It(h.name,e.Diagnostics.Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1,Pa(b),Fa(c))}C.declarations=b.declarations,C.parent=b.parent,b.valueDeclaration&&(C.valueDeclaration=b.valueDeclaration),C.type=T,C.target=b,b=C}else{if(277===h.kind){k<2&&_A(h,2),i.length>0&&(o=Td(o,w(),n.symbol,s,32768),i=[],r=e.createSymbolTable(),_=!1,g=!1,m=0);T=qE(h.expression);if(!Jv(T))return It(h,e.Diagnostics.Spread_types_may_only_be_created_from_object_types),oe;o=Td(o,T,n.symbol,s,32768),v=y+1;continue}e.Debug.assert(158===h.kind||159===h.kind),Yx(h)}!E||8576&E.flags?r.set(b.escapedName,b):Lp(E,Ae)&&(Lp(E,me)?g=!0:_=!0,a&&(f=!0)),i.push(b)}if(l)for(var N=0,I=Rs(c);N<I.length;N++){C=I[N];r.get(C.escapedName)||o&&sc(o,C.escapedName)||(16777216&C.flags||It(C.valueDeclaration||C.bindingElement,e.Diagnostics.Initializer_provides_no_value_for_this_binding_element_and_the_binding_element_has_no_default_value),r.set(C.escapedName,C),i.push(C))}return o!==ke?(i.length>0&&(o=Td(o,w(),n.symbol,s,32768)),o):w();function w(){var t=_?Wv(n.properties,v,i,0):void 0,o=g?Wv(n.properties,v,i,1):void 0,c=ba(n.symbol,r,e.emptyArray,e.emptyArray,t,o);return c.flags|=268435456|939524096&m,c.objectFlags|=128|V,p&&(c.objectFlags|=16384),f&&(c.objectFlags|=512),a&&(c.pattern=n),s|=939524096&c.flags,c}}function Jv(n){return!!(126615555&n.flags||117632&nf(n)&&Jv(tf(n))||3145728&n.flags&&e.every(n.types,Jv))}function Xv(e){Ty(e)}function Yv(e,n){return Yx(e),yy(e)||re}function Qv(e){Ty(e.openingElement),ny(e.closingElement.tagName)?sy(e.closingElement):qE(e.closingElement.tagName),ay(e)}function Zv(e,n){return Yx(e),yy(e)||re}function $v(n){return Ty(n.openingFragment),2===L.jsx&&(L.jsxFactory||e.getSourceFileOfNode(n).pragmas.has("jsx"))&&It(n,L.jsxFactory?e.Diagnostics.JSX_fragment_is_not_supported_when_using_jsxFactory:e.Diagnostics.JSX_fragment_is_not_supported_when_using_an_inline_JSX_factory_pragma),ay(n),yy(n)||re}function ey(n){return!e.stringContains(n,"-")}function ny(n){return 72===n.kind&&e.isIntrinsicJsxName(n.escapedText)}function ty(e,n){return e.initializer?BE(e.initializer,n):ve}function ry(n,t){for(var r,a=n.attributes,i=e.createSymbolTable(),o=Ne,s=!1,c=!1,l=0,u=4096,d=py(cy(n)),p=0,m=a.properties;p<m.length;p++){var f=m[p],_=f.symbol;if(e.isJsxAttribute(f)){var g=ty(f,t);l|=939524096&g.flags;var v=Ot(33554436|_.flags,_.escapedName);v.declarations=_.declarations,v.parent=_.parent,_.valueDeclaration&&(v.valueDeclaration=_.valueDeclaration),v.type=g,v.target=_,i.set(v.escapedName,v),f.name.escapedText===d&&(c=!0)}else{e.Debug.assert(269===f.kind),i.size>0&&(o=Td(o,x(),a.symbol,l,u),i=e.createSymbolTable());g=RE(f.expression,t);ai(g)&&(s=!0),Jv(g)?o=Td(o,g,a.symbol,l,u):r=r?Ru([r,g]):g}}s||i.size>0&&(o=Td(o,x(),a.symbol,l,u));var y=260===n.parent.kind?n.parent:void 0;if(y&&y.openingElement===n&&y.children.length>0){var h=ay(y,t);if(!s&&d&&""!==d){c&&It(a,e.Diagnostics._0_are_specified_twice_The_attribute_named_0_will_be_overwritten,e.unescapeLeadingUnderscores(d));var b=hv(n.attributes),E=b&&sv(b,d),T=Ot(33554436,d);T.type=1===h.length?h[0]:Gv(h,E,!1)||su(Au(h)),T.valueDeclaration=e.createPropertySignature(void 0,e.unescapeLeadingUnderscores(d),void 0,void 0,void 0),T.valueDeclaration.parent=a,T.valueDeclaration.symbol=T;var S=e.createSymbolTable();S.set(d,T),o=Td(o,ba(a.symbol,S,e.emptyArray,e.emptyArray,void 0,void 0),a.symbol,l,u)}}return s?re:r&&o!==Ne?Ru([r,o]):r||(o===Ne?x():o);function x(){u|=V;var n=ba(a.symbol,i,e.emptyArray,e.emptyArray,void 0,void 0);return n.flags|=268435456|l,n.objectFlags|=128|u,n}}function ay(e,n){for(var t=[],r=0,a=e.children;r<a.length;r++){var i=a[r];11===i.kind?i.containsOnlyWhiteSpaces||t.push(pe):t.push(BE(i,n))}return t}function iy(e,n){return ry(e.parent,n)}function oy(e,n){var t=cy(n),r=t&&Zr(t),a=r&&Xt(r,e,67897832);return a?Do(a):oe}function sy(t){var r=zt(t);if(!r.resolvedSymbol){var a=oy(n.IntrinsicElements,t);if(a!==oe){if(!e.isIdentifier(t.tagName))return e.Debug.fail();var i=sc(a,t.tagName.escapedText);if(i)return r.jsxFlags|=1,r.resolvedSymbol=i;var o=mc(a,0);return o?(r.jsxFlags|=2,r.resolvedSymbol=a.symbol):(It(t,e.Diagnostics.Property_0_does_not_exist_on_type_1,e.idText(t.tagName),"JSX."+n.IntrinsicElements),r.resolvedSymbol=ne)}return R&&It(t,e.Diagnostics.JSX_element_implicitly_has_type_any_because_no_interface_JSX_0_exists,e.unescapeLeadingUnderscores(n.IntrinsicElements)),r.resolvedSymbol=ne}return r.resolvedSymbol}function cy(e){var t=e&&zt(e);if(t&&t.jsxNamespace)return t.jsxNamespace;if(!t||!1!==t.jsxNamespace){var r=Dt(e),a=Zt(e,r,1920,void 0,r,!1);if(a){var i=Xt(Zr(kr(a)),n.JSX,1920);if(i)return t&&(t.jsxNamespace=i),i;t&&(t.jsxNamespace=!1)}}return Ml(n.JSX,1920,void 0)}function ly(n,t){var r=t&&Xt(t.exports,n,67897832),a=r&&Do(r),i=a&&Rs(a);if(i){if(0===i.length)return"";if(1===i.length)return i[0].escapedName;i.length>1&&It(r.declarations[0],e.Diagnostics.The_global_type_JSX_0_may_not_have_more_than_one_property,e.unescapeLeadingUnderscores(n))}}function uy(e){return e&&Xt(e.exports,n.LibraryManagedAttributes,67897832)}function dy(e){return ly(n.ElementAttributesPropertyNameContainer,e)}function py(e){return ly(n.ElementChildrenAttributeNameContainer,e)}function my(t,r){if(4&t.flags)return[Dn];if(128&t.flags){var a=fy(t,r);if(a){var i=eb(r,a);return[i]}return It(r,e.Diagnostics.Property_0_does_not_exist_on_type_1,t.value,"JSX."+n.IntrinsicElements),e.emptyArray}var o=rc(t),s=lc(o,1);return 0===s.length&&(s=lc(o,0)),0===s.length&&1048576&o.flags&&(s=is(e.map(o.types,function(e){return my(e,r)}))),s}function fy(t,r){var a=oy(n.IntrinsicElements,r);if(a!==oe){var i=t.value,o=sc(a,e.escapeLeadingUnderscores(i));if(o)return Xi(o);var s=mc(a,0);return s||void 0}return re}function _y(n,t,r){if(1===n){var a=hy(r);a&&om(t,a,Et,r,e.Diagnostics.JSX_element_type_0_is_not_a_constructor_function_for_JSX_elements)}else if(0===n){var i=vy(r);i&&om(t,i,Et,r,e.Diagnostics.JSX_element_type_0_is_not_a_constructor_function_for_JSX_elements)}else{a=hy(r),i=vy(r);if(!a||!i)return;var o=Au([a,i]);om(t,o,Et,r,e.Diagnostics.JSX_element_type_0_is_not_a_constructor_function_for_JSX_elements)}}function gy(n){e.Debug.assert(ny(n.tagName));var t=zt(n);if(!t.resolvedJsxElementAttributesType){var r=sy(n);return 1&t.jsxFlags?t.resolvedJsxElementAttributesType=Xi(r):2&t.jsxFlags?t.resolvedJsxElementAttributesType=nl(r,0).type:t.resolvedJsxElementAttributesType=oe}return t.resolvedJsxElementAttributesType}function vy(e){var t=oy(n.ElementClass,e);if(t!==oe)return t}function yy(e){return oy(n.Element,e)}function hy(e){var n=yy(e);if(n)return Au([n,ue])}function by(t){var r=oy(n.IntrinsicElements,t);return r?Rs(r):e.emptyArray}function Ey(n){0===(L.jsx||0)&&It(n,e.Diagnostics.Cannot_use_JSX_unless_the_jsx_flag_is_provided),void 0===yy(n)&&R&&It(n,e.Diagnostics.JSX_element_implicitly_has_type_any_because_the_global_type_JSX_Element_does_not_exist)}function Ty(n){var t=e.isJsxOpeningLikeElement(n);t&&JA(n),Ey(n);var r=it&&2===L.jsx?e.Diagnostics.Cannot_find_name_0:void 0,a=Dt(n),i=t?n.tagName:n,o=Zt(i,a,67220415,r,a,!0);if(o&&(o.isReferenced=67108863,2097152&o.flags&&!HC(Nr(o))&&wr(o)),t){var s=ab(n);_y(Sh(n),Fc(s),n)}}function Sy(e,n,t){if(524288&e.flags){var r=ws(e);if(r.stringIndexInfo||r.numberIndexInfo&&Uv(n)||Os(e,n)||t&&!ey(n))return!0}else if(3145728&e.flags&&xy(e))for(var a=0,i=e.types;a<i.length;a++){var o=i[a];if(Sy(o,n,t))return!0}return!1}function xy(n){return!!(524288&n.flags&&!(512&e.getObjectFlags(n))||67108864&n.flags||1048576&n.flags&&e.some(n.types,xy)||2097152&n.flags&&e.every(n.types,xy))}function Cy(n,t){if(n.expression){var r=qE(n.expression,t);return n.dotDotDotToken&&r!==re&&!Om(r)&&It(n,e.Diagnostics.JSX_spread_child_must_be_an_array_type),r}return oe}function Ay(n){return n.valueDeclaration?e.getCombinedNodeFlags(n.valueDeclaration):0}function Dy(n){if(8192&n.flags||4&e.getCheckFlags(n))return!0;if(e.isInJSFile(n.valueDeclaration)){var t=n.valueDeclaration.parent;return t&&e.isBinaryExpression(t)&&3===e.getAssignmentDeclarationKind(t)}}function Ly(n,t,r,a){var i=e.getDeclarationModifierFlagsFromSymbol(a),o=148===n.kind?n.right:183===n.kind?n:n.name;if(512&e.getCheckFlags(a))return It(o,e.Diagnostics.Property_0_has_conflicting_declarations_and_is_inaccessible_in_type_1,Pa(a),Fa(r)),!1;if(t){if(k<2&&Ny(a))return It(o,e.Diagnostics.Only_public_and_protected_methods_of_the_base_class_are_accessible_via_the_super_keyword),!1;if(128&i)return It(o,e.Diagnostics.Abstract_method_0_in_class_1_cannot_be_accessed_via_super_expression,Pa(a),Fa(bm(a))),!1}if(128&i&&e.isThisProperty(n)&&Ny(a)){var s=e.getClassLikeDeclarationOfSymbol(aa(a));if(s&&pC(n))return It(o,e.Diagnostics.Abstract_property_0_in_class_1_cannot_be_accessed_in_the_constructor,Pa(a),e.getTextOfIdentifierOrLiteral(s.name)),!1}if(!(24&i))return!0;if(8&i){s=e.getClassLikeDeclarationOfSymbol(aa(a));return!!mC(n,s)||(It(o,e.Diagnostics.Property_0_is_private_and_only_accessible_within_class_1,Pa(a),Fa(bm(a))),!1)}if(t)return!0;var c=dC(n,function(e){var n=Do(ra(e));return Sm(n,a)?n:void 0});if(!c){var l=void 0;if(32&i||!(l=ky(n))||!l.type)return It(o,e.Diagnostics.Property_0_is_protected_and_only_accessible_within_class_1_and_its_subclasses,Pa(a),Fa(bm(a)||r)),!1;var u=Rd(l.type);c=262144&u.flags?Bs(u):u}return!!(32&i)||(262144&r.flags&&(r=r.isThisType?Bs(r):qs(r)),!(!r||!Zi(r,c))||(It(o,e.Diagnostics.Property_0_is_protected_and_only_accessible_through_an_instance_of_class_1,Pa(a),Fa(c)),!1))}function ky(n){var t=e.getThisContainer(n,!1);return t&&e.isFunctionLike(t)?e.getThisParameter(t):void 0}function Ny(e){return!!hm(e,function(e){return!(8192&e.flags)})}function Iy(e,n,t,r){return My(qE(e),e,n,t,r)}function wy(e){var n=98304&(w?nf(e):e.flags);return n?lf(e):e}function My(n,t,r,a,i){if(2&n.flags)return It(t,e.Diagnostics.Object_is_of_type_unknown),oe;var o=98304&(w?nf(n):n.flags);if(o){It(t,32768&o?65536&o?i||e.Diagnostics.Object_is_possibly_null_or_undefined:a||e.Diagnostics.Object_is_possibly_undefined:r||e.Diagnostics.Object_is_possibly_null);var s=lf(n);return 229376&s.flags?oe:s}return n}function Oy(e){return Ry(e,e.expression,e.name)}function Py(e){return Ry(e,e.left,e.right)}function Ry(n,t,r){var a,i=Iy(t),o=zt(t).resolvedSymbol,s=rc(bf(i));if(ai(s)||s===Se)return e.isIdentifier(t)&&o&&Cg(o,n),s;var c=e.getAssignmentTargetKind(n),l=sc(s,r.escapedText);if(!e.isIdentifier(t)||!o||l&&HC(l)||Cg(o,n),l){if(Fy(l,n,r),Yy(l,n,100===t.kind),zt(n).resolvedSymbol=l,Ly(n,98===t.kind,s,l),c&&(rE(n,l)||aE(n)))return It(r,e.Diagnostics.Cannot_assign_to_0_because_it_is_a_read_only_property,e.idText(r)),oe;a=xg(Xi(l),n)}else{var u=pc(s,0);if(!u||!u.type)return Yu(i)?re:(r.escapedText&&!ar(n)&&Ky(r,262144&i.flags&&i.isThisType?s:i),oe);u.isReadonly&&(e.isAssignmentTarget(n)||e.isDeleteTarget(n))&&It(n,e.Diagnostics.Index_signature_in_type_0_only_permits_reading,Fa(s)),a=u.type}if(189!==n.kind||1===c||l&&!(98311&l.flags)&&!(8192&l.flags&&1048576&a.flags))return a;var d=!1;if(w&&P&&100===t.kind){var p=l&&l.valueDeclaration;if(p&&hx(p)){var m=gg(n);157===m.kind&&m.parent===p.parent&&(d=!0)}}else w&&l&&l.valueDeclaration&&e.isPropertyAccessExpression(l.valueDeclaration)&&e.getAssignmentDeclarationPropertyAccessKind(l.valueDeclaration)&&gg(n)===gg(l.valueDeclaration)&&(d=!0);var f=fg(n,a,d?sf(a):a);return d&&!(32768&nf(a))&&32768&nf(f)?(It(r,e.Diagnostics.Property_0_is_used_before_being_assigned,Pa(l)),a):c?Wm(f):f}function Fy(n,t,r){var a=n.valueDeclaration;if(a){var i,o=e.idText(r);!Gy(t)||Qt(a,r)||Vy(n)?240!==a.kind||164===t.parent.kind||4194304&a.flags||Qt(a,r)||(i=It(r,e.Diagnostics.Class_0_used_before_its_declaration,o)):i=It(r,e.Diagnostics.Property_0_is_used_before_its_initialization,o),i&&Nt(i,e.createDiagnosticForNode(a,e.Diagnostics._0_is_declared_here,o))}}function Gy(n){return!!e.findAncestor(n,function(n){switch(n.kind){case 154:return!0;case 275:return!1;default:return!e.isExpressionNode(n)&&"quit"}})}function Vy(e){if(!(32&e.parent.flags))return!1;var n=Xi(e.parent);while(1){if(n=n.symbol&&By(n),!n)return!1;var t=sc(n,e.escapedName);if(t&&t.valueDeclaration)return!0}}function By(e){var n=uo(e);if(0!==n.length)return Ru(n)}function Ky(n,t){var r,a;if(1048576&t.flags&&!(131068&t.flags))for(var i=0,o=t.types;i<o.length;i++){var s=o[i];if(!sc(s,n.escapedText)){r=e.chainDiagnosticMessages(r,e.Diagnostics.Property_0_does_not_exist_on_type_1,e.declarationNameToString(n),Fa(s));break}}if(jy(n.escapedText,t))r=e.chainDiagnosticMessages(r,e.Diagnostics.Property_0_is_a_static_member_of_type_1,e.declarationNameToString(n),Fa(t));else{var c=OT(t);if(c&&sc(c,n.escapedText))r=e.chainDiagnosticMessages(r,e.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_forget_to_use_await,e.declarationNameToString(n),Fa(t));else{var l=Uy(n,t);if(void 0!==l){var u=e.symbolName(l);r=e.chainDiagnosticMessages(r,e.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2,e.declarationNameToString(n),Fa(t),u),a=l.valueDeclaration&&e.createDiagnosticForNode(l.valueDeclaration,e.Diagnostics._0_is_declared_here,u)}else r=e.chainDiagnosticMessages(r,e.Diagnostics.Property_0_does_not_exist_on_type_1,e.declarationNameToString(n),Fa(t))}}var d=e.createDiagnosticForNodeFromMessageChain(n,r);a&&Nt(d,a),it.add(d)}function jy(n,t){var r=t.symbol&&sc(Xi(t.symbol),n);return void 0!==r&&r.valueDeclaration&&e.hasModifier(r.valueDeclaration,32)}function Uy(n,t){return Xy(e.isString(n)?n:e.idText(n),Rs(t),67220415)}function Hy(n,t){var r=Uy(n,t);return r&&e.symbolName(r)}function Wy(n,t,r){e.Debug.assert(void 0!==t,"outername should always be defined");var a=$t(n,t,r,void 0,t,!1,!1,function(n,r,a){e.Debug.assertEqual(t,r,"name should equal outerName");var i=Xt(n,r,a);return i||Xy(e.unescapeLeadingUnderscores(r),e.arrayFrom(n.values()),a)});return a}function qy(n,t,r){var a=Wy(n,t,r);return a&&e.symbolName(a)}function zy(n,t){return t.exports&&Xy(e.idText(n),Jr(t),2623475)}function Jy(n,t){var r=zy(n,t);return r&&e.symbolName(r)}function Xy(n,t,r){return e.getSpellingSuggestion(n,t,a);function a(n){var t=e.symbolName(n);return!e.startsWith(t,'"')&&n.flags&r?t:void 0}}function Yy(n,t,r){if(n&&106500&n.flags&&n.valueDeclaration&&e.hasModifier(n.valueDeclaration,8)&&(!t||!e.isWriteOnlyAccess(t)||65536&n.flags&&!(32768&n.flags))){if(r){var a=e.findAncestor(t,e.isFunctionLikeDeclaration);if(a&&a.symbol===n)return}(1&e.getCheckFlags(n)?qt(n).target:n).isReferenced=67108863}}function Qy(e,n){switch(e.kind){case 189:return nh(e,98===e.expression.kind,n,bf(qE(e.expression)));case 148:return nh(e,!1,n,bf(qE(e.left)));case 183:return nh(e,!1,n,Rd(e))}}function Zy(e,n,t){return nh(e,189===e.kind&&98===e.expression.kind,t.escapedName,n)&&(!(8192&t.flags)||$y(t,n))}function $y(n,t){var r=ti(t,n.escapedName),a=lc(lf(r),0);return e.Debug.assert(0!==a.length),a.some(function(e){var n=wc(e);return!n||Lp(t,eh(e,n,t))})}function eh(e,n,t){if(!e.typeParameters)return n;var r=Af(e.typeParameters,e,0);return jf(r.inferences,t,n),mp(n,Wc(e,Zf(r)))}function nh(n,t,r,a){if(a===oe||ai(a))return!0;var i=sc(a,r);return i?Ly(n,t,a,i):e.isInJSFile(n)&&0!==(1048576&a.flags)&&a.types.some(function(e){return nh(n,t,r,e)})}function th(n){var t=n.initializer;if(238===t.kind){var r=t.declarations[0];if(r&&!e.isBindingPattern(r.name))return ra(r)}else if(72===t.kind)return e_(t)}function rh(e){return mc(e,1)&&!mc(e,0)}function ah(n){var t=e.skipParentheses(n);if(72===t.kind){var r=e_(t);if(3&r.flags){var a=n,i=n.parent;while(i){if(226===i.kind&&a===i.statement&&th(i)===r&&rh(HE(i.expression)))return!0;a=i,i=i.parent}}}return!1}function ih(n){var t=Iy(n.expression),r=n.argumentExpression;if(!r){var a=e.getSourceFileOfNode(n);if(192===n.parent.kind&&n.parent.expression===n){var i=e.skipTrivia(a.text,n.expression.end),o=n.end;vD(a,i,o-i,e.Diagnostics.new_T_cannot_be_used_to_create_an_array_Use_new_Array_T_instead)}else{i=n.end-"]".length,o=n.end;vD(a,i,o-i,e.Diagnostics.Expression_expected)}return oe}var s=qE(r);return t===oe||t===Se?t:gE(t)&&10!==r.kind?(It(r,e.Diagnostics.A_const_enum_member_can_only_be_accessed_using_a_string_literal),oe):ET(ad(t,ah(r)?me:s,n),n)}function oh(n,t,r){if(t===oe)return!1;if(!e.isWellKnownSymbolSyntactically(n))return!1;if(0===(12288&t.flags))return r&&It(n,e.Diagnostics.A_computed_property_name_of_the_form_0_must_be_of_type_symbol,e.getTextOfNode(n)),!1;var a=n.expression,i=e_(a);if(!i)return!1;var o=Gl(!0);return!!o&&(i===o||(r&&It(a,e.Diagnostics.Symbol_reference_does_not_refer_to_the_global_Symbol_constructor_object),!1))}function sh(n){return e.isCallOrNewExpression(n)||e.isTaggedTemplateExpression(n)||e.isJsxOpeningLikeElement(n)}function ch(n){return sh(n)&&e.forEach(n.typeArguments,qx),193===n.kind?qE(n.template):e.isJsxOpeningLikeElement(n)?qE(n.attributes):152!==n.kind&&e.forEach(n.arguments,function(e){qE(e)}),Dn}function lh(e){return ch(e),Ln}function uh(n,t){var r,a,i,o,s=0,c=-1;e.Debug.assert(!t.length);for(var l=0,u=n;l<u.length;l++){var d=u[l],p=d.declaration&&ra(d.declaration),m=d.declaration&&d.declaration.parent;a&&p!==a?(i=s=t.length,r=m):r&&m===r?i+=1:(r=m,i=s),a=p,d.hasLiteralTypes?(c++,o=c,s++):o=i,t.splice(o,0,d)}}function dh(e){return!!e&&(208===e.kind||215===e.kind&&e.isSpread)}function ph(n){return e.findIndex(n,dh)}function mh(e){return!!(16384&e.flags)}function fh(n,t,r,a){var i;void 0===a&&(a=!1);var o=!1,s=Lb(r),c=kb(r);if(193===n.kind)if(i=t.length,206===n.template.kind){var l=e.last(n.template.templateSpans);o=e.nodeIsMissing(l.literal)||!!l.literal.isUnterminated}else{var u=n.template;e.Debug.assert(14===u.kind),o=!!u.isUnterminated}else if(152===n.kind)i=Nh(n,r);else if(e.isJsxOpeningLikeElement(n)){if(o=n.attributes.end===n.end,o)return!0;i=0===c?t.length:1,s=0===t.length?s:1,c=Math.min(c,1)}else{if(!n.arguments)return e.Debug.assert(192===n.kind),0===kb(r);i=a?t.length+1:t.length,o=n.arguments.end===n.end;var d=ph(t);if(d>=0)return d>=kb(r)&&(Nb(r)||d<Lb(r))}if(!Nb(r)&&i>s)return!1;if(o||i>=c)return!0;for(var p=i;p<c;p++){var m=Cb(r,p);if(131072&J_(m,mh).flags)return!1}return!0}function _h(n,t){var r=e.length(n.typeParameters),a=Sc(n.typeParameters);return!t||t.length>=a&&t.length<=r}function gh(e){if(524288&e.flags){var n=ws(e);if(1===n.callSignatures.length&&0===n.constructSignatures.length&&0===n.properties.length&&!n.stringIndexInfo&&!n.numberIndexInfo)return n.callSignatures[0]}}function vh(n,t,r,a){var i=Af(n.typeParameters,n,0,a),o=r?np(t,r):t;return Cf(o,n,function(e,n){jf(i.inferences,e,n)}),r||jf(i.inferences,Fc(t),Fc(n),8),jc(n,Zf(i),e.isInJSFile(t.declaration))}function yh(e,n,t,r){var a=Sv(n,e),i=PE(e.attributes,a,t&&void 0!==t[0]?D:r);return jf(r.inferences,i,a),Zf(r)}function hh(n,t,r,a,i){for(var o=0,s=i.inferences;o<s.length;o++){var c=s[o];c.isFixed||(c.inferredType=void 0)}if(e.isJsxOpeningLikeElement(n))return yh(n,t,a,i);if(152!==n.kind){var l=bv(n);if(l){var u=mp(l,zd(Ev(n))),d=gh(u),p=d&&d.typeParameters?Qc(Uc(d,d.typeParameters)):u,m=Fc(t);jf(i.inferences,p,m,8)}}var f=wc(t);if(f){var _=Ah(n),g=_?qE(_):Ee;jf(i.inferences,g,f)}for(var v=wb(t),y=v?Math.min(Lb(t)-1,r.length):r.length,h=0;h<y;h++){var b=r[h];if(210!==b.kind){var E=Cb(t,h),T=a&&void 0!==a[h]?D:i,S=PE(b,E,T);jf(i.inferences,S,E)}}if(v){var x=Eh(r,y,r.length,v,i);jf(i.inferences,x,v)}return Zf(i)}function bh(e){return q_(e,function(e){return!(63176705&e.flags||Om(e)||Xm(e))})?su(mc(e,1)||oe):e}function Eh(n,t,r,a,i){if(t>=r-1){var o=n[r-1];if(dh(o))return 215===o.kind?su(o.type):bh(PE(o.expression,a,i))}for(var s=mc(a,1)||re,c=mE(s,4325372),l=[],u=-1,d=t;d<r;d++){var p=PE(n[d],s,i);u<0&&dh(n[d])&&(u=d-t),l.push(c?Ld(p):qm(p))}return u<0?pu(l):pu(e.append(l.slice(0,u),Au(l.slice(u))),u,!0)}function Th(n,t,r,a){for(var i,o=e.isInJSFile(n.declaration),s=n.typeParameters,c=xc(e.map(t,Rd),s,Sc(s),o),l=0;l<t.length;l++){e.Debug.assert(void 0!==s[l],"Should not call checkTypeArguments with too many type arguments");var u=Bs(s[l]);if(u){var d=r&&a?function(){return e.chainDiagnosticMessages(void 0,e.Diagnostics.Type_0_does_not_satisfy_the_constraint_1)}:void 0,p=a||e.Diagnostics.Type_0_does_not_satisfy_the_constraint_1;i||(i=Ud(s,c));var m=c[l];if(!wp(m,Xo(mp(u,i),m),r?t[l]:void 0,p,d))return}}return c}function Sh(n){if(ny(n.tagName))return 2;var t=rc(qE(n.tagName));return e.length(lc(t,1))?0:e.length(lc(t,0))?1:2}function xh(e,n,t,r,a){var i=Sv(n,e),o=PE(e.attributes,i,r&&r[0]?D:void 0);return Op(o,i,t,a?e.tagName:void 0,e.attributes)}function Ch(n,t,r,a,i,o){if(e.isJsxOpeningLikeElement(n))return xh(n,r,a,i,o);var s=wc(r);if(s&&s!==Ee&&192!==n.kind){var c=Ah(n),l=c?qE(c):Ee,u=o?c||n:void 0,d=e.Diagnostics.The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1;if(!om(l,s,a,u,d))return!1}for(var p=e.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1,m=wb(r),f=m?Math.min(Lb(r)-1,t.length):t.length,_=0;_<f;_++){var g=t[_];if(210!==g.kind){var v=Cb(r,_),y=PE(g,v,i&&i[_]?D:void 0),h=i?mf(y):y;if(!Op(h,v,a,o?g:void 0,g,p))return!1}}if(m){var b=Eh(t,f,t.length,m,void 0);u=o?f<t.length?t[f]:n:void 0;return om(b,m,a,u,p)}return!0}function Ah(n){if(191===n.kind){var t=e.skipOuterExpressions(n.expression);if(189===t.kind||190===t.kind)return t.expression}}function Dh(n,t,r){var a=e.createNode(215,n.pos,n.end);return a.parent=n,a.type=t,a.isSpread=r||!1,a}function Lh(n){if(193===n.kind){var t=n.template,r=[Dh(t,Rl())];return 206===t.kind&&e.forEach(t.templateSpans,function(e){r.push(e.expression)}),r}if(152===n.kind)return kh(n);if(e.isJsxOpeningLikeElement(n))return n.attributes.properties.length>0||e.isJsxOpeningElement(n)&&n.parent.children.length>0?[n.attributes]:e.emptyArray;var a=n.arguments||e.emptyArray,i=a.length;if(i&&dh(a[i-1])&&ph(a)===i-1){var o=a[i-1],s=RE(o.expression);if(Xm(s)){var c=s.typeArguments||e.emptyArray,l=s.target.hasRestElement?c.length-1:-1,u=e.map(c,function(e,n){return Dh(o,e,n===l)});return e.concatenate(a.slice(0,i-1),u)}}return a}function kh(n){var t=n.parent,r=n.expression;switch(t.kind){case 240:case 209:return[Dh(r,Xi(ra(t)))];case 151:var a=t.parent;return[Dh(r,157===t.parent.kind?Xi(ra(a)):oe),Dh(r,re),Dh(r,me)];case 154:case 156:case 158:case 159:var i=154!==t.kind&&0!==k;return[Dh(r,AC(t)),Dh(r,DC(t)),Dh(r,i?tu(TC(t)):re)]}return e.Debug.fail()}function Nh(n,t){switch(n.parent.kind){case 240:case 209:return 1;case 154:return 2;case 156:case 158:case 159:return 0===k||t.parameters.length<=2?2:3;case 151:return 3;default:return e.Debug.fail()}}function Ih(n,t,r){for(var a,i=Number.POSITIVE_INFINITY,o=Number.NEGATIVE_INFINITY,s=Number.NEGATIVE_INFINITY,c=Number.POSITIVE_INFINITY,l=r.length,u=0,d=t;u<d.length;u++){var p=d[u],m=kb(p),f=Lb(p);m<l&&m>s&&(s=m),l<f&&f<c&&(c=f),m<i&&(i=m,a=p),o=Math.max(o,f)}var _,g=e.some(t,Nb),v=g?i:i<o?i+"-"+o:i,y=ph(r)>-1;if(l<=o&&y&&l--,a&&kb(a)>l&&a.declaration){var h=a.declaration.parameters[a.thisParameter?l+1:l];h&&(_=e.createDiagnosticForNode(h,e.isBindingPattern(h.name)?e.Diagnostics.An_argument_matching_this_binding_pattern_was_not_provided:e.Diagnostics.An_argument_for_0_was_not_provided,h.name?e.isBindingPattern(h.name)?void 0:e.idText(wx(h.name)):l))}if(g||y){var b=g&&y?e.Diagnostics.Expected_at_least_0_arguments_but_got_1_or_more:g?e.Diagnostics.Expected_at_least_0_arguments_but_got_1:e.Diagnostics.Expected_0_arguments_but_got_1_or_more,E=e.createDiagnosticForNode(n,b,v,l);return _?Nt(E,_):E}if(i<l&&l<o)return e.createDiagnosticForNode(n,e.Diagnostics.No_overload_expects_0_arguments_but_overloads_do_exist_that_expect_either_1_or_2_arguments,l,s,c);var T=e.createDiagnosticForNode(n,e.Diagnostics.Expected_0_arguments_but_got_1,v,l);return _?Nt(T,_):T}function wh(n,t,r){var a=r.length;if(1===t.length){var i=t[0],o=Sc(i.typeParameters),s=e.length(i.typeParameters);return e.createDiagnosticForNodeArray(e.getSourceFileOfNode(n),r,e.Diagnostics.Expected_0_type_arguments_but_got_1,o<s?o+"-"+s:o,a)}for(var c=-1/0,l=1/0,u=0,d=t;u<d.length;u++){i=d[u];var p=Sc(i.typeParameters);s=e.length(i.typeParameters);p>a?l=Math.min(l,p):s<a&&(c=Math.max(c,s))}return c!==-1/0&&l!==1/0?e.createDiagnosticForNodeArray(e.getSourceFileOfNode(n),r,e.Diagnostics.No_overload_expects_0_type_arguments_but_overloads_do_exist_that_expect_either_1_or_2_type_arguments,a,c,l):e.createDiagnosticForNodeArray(e.getSourceFileOfNode(n),r,e.Diagnostics.Expected_0_type_arguments_but_got_1,c===-1/0?l:c,a)}function Mh(n,t,r,i,o){var s,c=193===n.kind,l=152===n.kind,u=e.isJsxOpeningLikeElement(n),d=!r;l||(s=n.typeArguments,(c||u||98!==n.expression.kind)&&e.forEach(s,qx));var p=r||[];if(uh(t,p),!p.length)return d&&it.add(e.createDiagnosticForNode(n,e.Diagnostics.Call_target_does_not_contain_any_signatures)),lh(n);var m,f,_,g,v=Lh(n),y=1===p.length&&!p[0].typeParameters,h=l||y?void 0:Oh(v),b=i&&191===n.kind&&n.arguments.hasTrailingComma;if(p.length>1&&(g=T(p,bt,b)),g||(g=T(p,Et,b)),g)return g;if(d)if(m)Ch(n,v,m,Et,void 0,!0);else if(f)it.add(Ih(n,[f],v));else if(_)Th(_,n.typeArguments,!0,o);else{var E=e.filter(t,function(e){return _h(e,s)});0===E.length?it.add(wh(n,t,s)):l?o&&it.add(e.createDiagnosticForNode(n,o)):it.add(Ih(n,E,v))}return a||!v?lh(n):Ph(n,p,v,!!r);function T(t,r,a){if(void 0===a&&(a=!1),m=void 0,f=void 0,_=void 0,y){var i=t[0];if(s||!fh(n,v,i,a))return;return Ch(n,v,i,r,h,!1)?i:void(m=i)}for(var o=0;o<t.length;o++){i=t[o];if(_h(i,s)&&fh(n,v,i,a)){var c=void 0,l=void 0;if(i.typeParameters){var u=void 0;if(s){if(u=Th(i,s,!1),!u){_=i;continue}}else l=Af(i.typeParameters,i,e.isInJSFile(n)?2:0),u=hh(n,i,v,h,l);if(c=jc(i,u,e.isInJSFile(i.declaration)),wb(i)&&!fh(n,v,c,a)){f=c;continue}}else c=i;if(Ch(n,v,c,r,h,!1)){if(h){if(h=void 0,l){u=hh(n,i,v,h,l);if(c=jc(i,u,e.isInJSFile(i.declaration)),wb(i)&&!fh(n,v,c,a)){f=c;continue}}if(!Ch(n,v,c,r,h,!1)){m&&!Ib(m)&&Ib(c)||(m=c);continue}}return t[o]=c,c}m&&!Ib(m)&&Ib(c)||(m=c)}}}}function Oh(e){for(var n,t=0;t<e.length;t++)yp(e[t])&&(n||(n=new Array(e.length)),n[t]=!0);return n}function Ph(n,t,r,a){return e.Debug.assert(t.length>0),a||1===t.length||t.some(function(e){return!!e.typeParameters})?Bh(n,t,r):Rh(t)}function Rh(n){var t,r=e.mapDefined(n,function(e){return e.thisParameter});r.length&&(t=Gh(r,r.map(Sb)));for(var a=e.minAndMax(n,Fh),i=a.min,o=a.max,s=[],c=function(t){var r=e.mapDefined(n,function(n){var r=n.parameters,a=n.hasRestParameter;return a?t<r.length-1?r[t]:e.last(r):t<r.length?r[t]:void 0});e.Debug.assert(0!==r.length),s.push(Gh(r,e.mapDefined(n,function(e){return Ab(e,t)})))},l=0;l<o;l++)c(l);var u=e.mapDefined(n,function(n){return n.hasRestParameter?e.last(n.parameters):void 0}),d=0!==u.length;if(d){var p=su(Au(e.mapDefined(n,Kc),2));s.push(Vh(u,p))}return $o(n[0].declaration,void 0,t,s,Ru(n.map(Fc)),void 0,i,d,n.some(function(e){return e.hasLiteralTypes}))}function Fh(e){var n=e.parameters.length;return e.hasRestParameter?n-1:n}function Gh(e,n){return Vh(e,Au(n,2))}function Vh(n,t){return df(e.first(n),t)}function Bh(n,t,r){var a=Uh(t,void 0===U?r.length:U),i=t[a],o=i.typeParameters;if(!o)return i;var s=sh(n)?n.typeArguments:void 0,c=s?Hc(i,Kh(s,o,e.isInJSFile(n))):jh(n,o,i,r);return t[a]=c,c}function Kh(e,n,t){var r=e.map(TC);while(r.length>n.length)r.pop();while(r.length<n.length)r.push(Bs(n[r.length])||Qf(t));return r}function jh(n,t,r,a){var i=Af(t,r,e.isInJSFile(n)?2:0),o=hh(n,r,a,Oh(a),i);return Hc(r,o)}function Uh(e,n){for(var t=-1,r=-1,a=0;a<e.length;a++){var i=e[a],o=Lb(i);if(Nb(i)||o>=n)return a;o>r&&(r=o,t=a)}return t}function Hh(n,t,r){if(98===n.expression.kind){var a=jg(n.expression);if(ai(a)){for(var i=0,o=n.arguments;i<o.length;i++){var s=o[i];qE(s)}return Dn}if(a!==oe){var c=e.getEffectiveBaseTypeNode(e.getContainingClass(n));if(c){var l=co(a,c.typeArguments,c);return Mh(n,l,t,r)}}return ch(n)}var u=Iy(n.expression,e.Diagnostics.Cannot_invoke_an_object_which_is_possibly_null,e.Diagnostics.Cannot_invoke_an_object_which_is_possibly_undefined,e.Diagnostics.Cannot_invoke_an_object_which_is_possibly_null_or_undefined);if(u===Se)return Nn;var d=rc(u);if(d===oe)return lh(n);var p=lc(d,0),m=lc(d,1).length;if(Wh(u,d,p.length,m))return u!==oe&&n.typeArguments&&It(n,e.Diagnostics.Untyped_function_calls_may_not_accept_type_arguments),ch(n);if(!p.length){if(m)It(n,e.Diagnostics.Value_of_type_0_is_not_callable_Did_you_mean_to_include_new,Fa(u));else{var f=void 0;if(1===n.arguments.length){var _=e.getSourceFileOfNode(n).text;e.isLineBreak(_.charCodeAt(e.skipTrivia(_,n.expression.end,!0)-1))&&(f=e.createDiagnosticForNode(n.expression,e.Diagnostics.It_is_highly_likely_that_you_are_missing_a_semicolon))}Xh(n,d,0,f)}return lh(n)}return p.some(function(n){return e.isInJSFile(n.declaration)&&!!e.getJSDocClassTag(n.declaration)})?(It(n,e.Diagnostics.Value_of_type_0_is_not_callable_Did_you_mean_to_include_new,Fa(u)),lh(n)):Mh(n,p,t,r)}function Wh(e,n,t,r){return ai(e)||ai(n)&&!!(262144&e.flags)||!t&&!r&&!(1179648&n.flags)&&Lp(e,He)}function qh(n,t,r){if(n.arguments&&k<1){var a=ph(n.arguments);a>=0&&It(n.arguments[a],e.Diagnostics.Spread_operator_in_new_expressions_is_only_available_when_targeting_ECMAScript_5_and_higher)}var i=Iy(n.expression);if(i===Se)return Nn;if(i=rc(i),i===oe)return lh(n);if(ai(i))return n.typeArguments&&It(n,e.Diagnostics.Untyped_function_calls_may_not_accept_type_arguments),ch(n);var o=lc(i,1);if(o.length){if(!Jh(n,o[0]))return lh(n);var s=i.symbol&&e.getClassLikeDeclarationOfSymbol(i.symbol);return s&&e.hasModifier(s,128)?(It(n,e.Diagnostics.Cannot_create_an_instance_of_an_abstract_class),lh(n)):Mh(n,o,t,r)}var c=lc(i,0);if(c.length){var l=Mh(n,c,t,r);return R||(l.declaration&&!ib(l.declaration)&&Fc(l)!==Ee&&It(n,e.Diagnostics.Only_a_void_function_can_be_called_with_the_new_keyword),wc(l)===Ee&&It(n,e.Diagnostics.A_function_that_is_called_with_the_new_keyword_cannot_have_a_this_type_that_is_void)),l}return Xh(n,i,1),lh(n)}function zh(n,t){var r=uo(t);if(!e.length(r))return!1;var a=r[0];if(2097152&a.flags){for(var i=a.types,o=e.countWhere(i,ao),s=0,c=0,l=a.types;c<l.length;c++){var u=l[c];if(s++,(0===o||o===i.length&&0===s||!ao(u))&&3&e.getObjectFlags(u)){if(u.symbol===n)return!0;if(zh(n,u))return!0}}return!1}return a.symbol===n||zh(n,a)}function Jh(n,t){if(!t||!t.declaration)return!0;var r=t.declaration,a=e.getSelectedModifierFlags(r,24);if(!a)return!0;var i=e.getClassLikeDeclarationOfSymbol(r.parent.symbol),o=Do(r.parent.symbol);if(!mC(n,i)){var s=e.getContainingClass(n);if(s&&16&a){var c=TC(s);if(zh(r.parent.symbol,c))return!0}return 8&a&&It(n,e.Diagnostics.Constructor_of_class_0_is_private_and_only_accessible_within_the_class_declaration,Fa(o)),16&a&&It(n,e.Diagnostics.Constructor_of_class_0_is_protected_and_only_accessible_within_the_class_declaration,Fa(o)),!1}return!0}function Xh(n,t,r,a){var i=It(n,0===r?e.Diagnostics.Cannot_invoke_an_expression_whose_type_lacks_a_call_signature_Type_0_has_no_compatible_call_signatures:e.Diagnostics.Cannot_use_new_with_an_expression_whose_type_lacks_a_call_or_construct_signature,Fa(t));Yh(t,r,a?Nt(i,a):i)}function Yh(n,t,r){if(n.symbol){var a=qt(n.symbol).originatingImport;if(a&&!e.isImportCall(a)){var i=lc(Xi(qt(n.symbol).target),t);if(!i||!i.length)return;Nt(r,e.createDiagnosticForNode(a,e.Diagnostics.Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cause_a_failure_at_runtime_Consider_using_a_default_import_or_import_require_here_instead))}}}function Qh(e,n,t){var r=qE(e.tag),a=rc(r);if(a===oe)return lh(e);var i=lc(a,0),o=lc(a,1).length;return Wh(r,a,i.length,o)?ch(e):i.length?Mh(e,i,n,t):(Xh(e,a,0),lh(e))}function Zh(n){switch(n.parent.kind){case 240:case 209:return e.Diagnostics.Unable_to_resolve_signature_of_class_decorator_when_called_as_an_expression;case 151:return e.Diagnostics.Unable_to_resolve_signature_of_parameter_decorator_when_called_as_an_expression;case 154:return e.Diagnostics.Unable_to_resolve_signature_of_property_decorator_when_called_as_an_expression;case 156:case 158:case 159:return e.Diagnostics.Unable_to_resolve_signature_of_method_decorator_when_called_as_an_expression;default:return e.Debug.fail()}}function $h(n,t,r){var a=qE(n.expression),i=rc(a);if(i===oe)return lh(n);var o=lc(i,0),s=lc(i,1).length;if(Wh(a,i,o.length,s))return ch(n);if(tb(n,o)){var c=e.getTextOfNode(n.expression,!1);return It(n,e.Diagnostics._0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write_0,c),lh(n)}var l=Zh(n);if(!o.length){var u=e.chainDiagnosticMessages(void 0,e.Diagnostics.Cannot_invoke_an_expression_whose_type_lacks_a_call_signature_Type_0_has_no_compatible_call_signatures,Fa(i));u=e.chainDiagnosticMessages(u,l);var d=e.createDiagnosticForNodeFromMessageChain(n,u);return it.add(d),Yh(i,0,d),lh(n)}return Mh(n,o,t,r,l)}function eb(t,r){var a=cy(t),i=a&&Zr(a),o=i&&Xt(i,n.Element,67897832),s=o&&K.symbolToEntityName(o,67897832,t),c=e.createFunctionTypeNode(void 0,[e.createParameter(void 0,void 0,void 0,"props",void 0,K.typeToTypeNode(r,t))],s?e.createTypeReferenceNode(s,void 0):e.createKeywordTypeNode(120)),l=Ot(1,"props");return l.type=r,$o(c,void 0,void 0,[l],o?Do(o):oe,void 0,1,!1,!1)}function nb(n,t,r){if(ny(n.tagName)){var a=gy(n),i=eb(n,a);return Mp(PE(n.attributes,Sv(i,n),void 0),a,n.tagName,n.attributes),i}var o=qE(n.tagName),s=rc(o);if(s===oe)return lh(n);var c=my(o,n);return Wh(o,s,c.length,0)?ch(n):0===c.length?(It(n.tagName,e.Diagnostics.JSX_element_type_0_does_not_have_any_construct_or_call_signatures,e.getTextOfNode(n.tagName)),lh(n)):Mh(n,c,t,r)}function tb(n,t){return t.length&&e.every(t,function(e){return 0===e.minArgumentCount&&!e.hasRestParameter&&e.parameters.length<Nh(n,e)})}function rb(n,t,r){switch(n.kind){case 191:return Hh(n,t,r);case 192:return qh(n,t,r);case 193:return Qh(n,t,r);case 152:return $h(n,t,r);case 262:case 261:return nb(n,t,r)}throw e.Debug.assertNever(n,"Branch in 'resolveSignature' should be unreachable.")}function ab(e,n,t){void 0===t&&(t=!1);var r=zt(e),a=r.resolvedSignature;if(a&&a!==kn&&!n)return a;r.resolvedSignature=kn;var i=rb(e,n,t);return r.resolvedSignature=Pn===Rn?i:a,i}function ib(n){if(!n||!e.isInJSFile(n))return!1;var t=e.isFunctionDeclaration(n)||e.isFunctionExpression(n)?n:e.isVariableDeclaration(n)&&n.initializer&&e.isFunctionExpression(n.initializer)?n.initializer:void 0;if(t){if(e.getJSDocClassTag(n))return!0;var r=ra(t);return!!r&&(void 0!==r.members||void 0!==r.exports&&void 0!==r.exports.get("prototype"))}return!1}function ob(e){if(524288&e.flags){var n=ws(e);return 1===n.callSignatures.length&&ib(n.callSignatures[0].declaration)}return!1}function sb(e){var n;ib(e.valueDeclaration)&&(n=ub(e));var t=cb(e);return t&&n?Ru([n,t]):t||n}function cb(n){var t=n.valueDeclaration,r=t&&t.parent&&(e.isFunctionDeclaration(t)&&ra(t)||e.isBinaryExpression(t.parent)&&ra(t.parent.left)||e.isVariableDeclaration(t.parent)&&ra(t.parent)),a=r&&r.exports&&r.exports.get("prototype"),i=a&&a.valueDeclaration&&lb(a.valueDeclaration);return i?qE(i):void 0}function lb(n){if(!n.parent)return!1;var t=n.parent;while(t&&189===t.kind)t=t.parent;if(t&&e.isBinaryExpression(t)&&e.isPrototypeAccess(t.left)&&59===t.operatorToken.kind){var r=e.getInitializerOfBinaryExpression(t);return e.isObjectLiteralExpression(r)&&r}}function ub(n){var t=qt(n);return t.inferredClassType||(t.inferredClassType=ba(n,zo(n)||A,e.emptyArray,e.emptyArray,void 0,void 0)),t.inferredClassType}function db(n){RA(n,n.typeArguments)||GA(n.arguments);var t=ab(n);if(98===n.expression.kind)return Ee;if(192===n.kind){var r=t.declaration;if(r&&157!==r.kind&&161!==r.kind&&166!==r.kind&&!e.isJSDocConstructSignature(r)&&!ib(r))return R&&It(n,e.Diagnostics.new_expression_whose_target_lacks_a_construct_signature_implicitly_has_an_any_type),re}if(e.isInJSFile(n)&&_b(n))return Ic(n.arguments[0]);var a,i=Fc(t);if(12288&i.flags&&pb(n))return Md(e.walkUpParenthesizedExpressions(n.parent));if(e.isInJSFile(n)){var o=e.getDeclarationOfExpando(n);if(o){var s=ra(o);s&&e.hasEntries(s.exports)&&(a=ba(s,s.exports,e.emptyArray,e.emptyArray,void 0,void 0),a.objectFlags|=16384)}}return a?Ru([i,a]):i}function pb(n){if(!e.isCallExpression(n))return!1;var t=n.expression;if(e.isPropertyAccessExpression(t)&&"for"===t.name.escapedText&&(t=t.expression),!e.isIdentifier(t)||"Symbol"!==t.escapedText)return!1;var r=Gl(!1);return!!r&&r===Zt(t,"Symbol",67220415,void 0,void 0,!1)}function mb(n){if(GA(n.arguments)||ND(n),0===n.arguments.length)return Kb(n,re);for(var t=n.arguments[0],r=RE(t),a=1;a<n.arguments.length;++a)RE(n.arguments[a]);(32768&r.flags||65536&r.flags||!Lp(r,pe))&&It(t,e.Diagnostics.Dynamic_import_s_specifier_must_be_of_type_string_but_here_has_type_0,Fa(r));var i=Vr(n,t);if(i){var o=qr(i,t,!0);if(o)return Kb(n,fb(Xi(o),o,i))}return Kb(n,re)}function fb(n,t,r){if(I&&n&&n!==oe){var a=n;if(!a.syntheticType){var i=e.find(r.declarations,e.isSourceFile),o=gr(i,r,!1);if(o){var s=e.createSymbolTable(),c=Ot(2097152,"default");c.nameType=Nd("default"),c.target=kr(t),s.set("default",c);var l=Ot(2048,"__type"),u=ba(l,s,e.emptyArray,e.emptyArray,void 0,void 0);l.type=u,a.syntheticType=Jv(n)?Td(n,u,l,0,0):u}else a.syntheticType=n}return a.syntheticType}return n}function _b(n){if(!e.isRequireCall(n,!0))return!1;if(!e.isIdentifier(n.expression))return e.Debug.fail();var t=Zt(n.expression,n.expression.escapedText,67220415,void 0,void 0,!0);if(t===W)return!0;if(2097152&t.flags)return!1;var r=16&t.flags?239:3&t.flags?237:0;if(0!==r){var a=e.getDeclarationOfKind(t,r);return!!a&&!!(4194304&a.flags)}return!1}function gb(e){return RA(e,e.typeArguments),k<2&&_A(e,65536),Fc(ab(e))}function vb(e){return yb(e,e.type,e.expression)}function yb(n,t,r,i){var o=mf(Wm(qE(r,i)));qx(t);var s=Rd(t);if(a&&s!==oe){var c=bf(o);Np(s,c)||Jp(o,s,n,e.Diagnostics.Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the_other_If_this_was_intentional_convert_the_expression_to_unknown_first)}return s}function hb(e){return lf(qE(e.expression))}function bb(n){return fD(n),95===n.keywordToken?Eb(n):92===n.keywordToken?Tb(n):e.Debug.assertNever(n.keywordToken)}function Eb(n){var t=e.getNewTargetContainer(n);if(t){if(157===t.kind){var r=ra(t.parent);return Xi(r)}r=ra(t);return Xi(r)}return It(n,e.Diagnostics.Meta_property_0_is_only_allowed_in_the_body_of_a_function_declaration_function_expression_or_constructor,"new.target"),oe}function Tb(n){(k<6||N<e.ModuleKind.ESNext)&&It(n,e.Diagnostics.The_import_meta_meta_property_is_only_allowed_using_ESNext_for_the_target_and_module_compiler_options);var t=e.getSourceFileOfNode(n);return e.Debug.assert(!!(1048576&t.flags),"Containing file is missing import meta node flag."),e.Debug.assert(!!t.externalModuleIndicator,"Containing file should be a module."),"meta"===n.name.escapedText?Fl():oe}function Sb(n){var t=Xi(n);if(w){var r=n.valueDeclaration;if(r&&e.hasInitializer(r))return sf(t)}return t}function xb(e,n){var t=e.parameters.length-(e.hasRestParameter?1:0);if(n<t)return e.parameters[n].escapedName;var r=e.parameters[t]||ne,a=Xi(r);if(Xm(a)){var i=a.target.associatedNames,o=n-t;return i?i[o]:r.escapedName+"_"+o}return r.escapedName}function Cb(e,n){return Ab(e,n)||re}function Ab(e,n){var t=e.parameters.length-(e.hasRestParameter?1:0);if(n<t)return Sb(e.parameters[n]);if(e.hasRestParameter){var r=Xi(e.parameters[t]);return Xm(r)?n-t<Zm(r)?r.typeArguments[n-t]:Ym(r):mc(r,1)}}function Db(e,n){var t=Lb(e),r=Ib(e);if(r&&n===t-1)return r;for(var a=r?Math.min(n,t-1):n,i=[],o=[],s=a;s<t;s++)i.push(Cb(e,s)),o.push(xb(e,s));var c=kb(e),l=c<a?0:c-a;return pu(i,l,!!r,o)}function Lb(n){var t=n.parameters.length;if(n.hasRestParameter){var r=Xi(n.parameters[t-1]);if(Xm(r))return t+(r.typeArguments||e.emptyArray).length-1}return t}function kb(e){if(e.hasRestParameter){var n=Xi(e.parameters[e.parameters.length-1]);if(Xm(n)){var t=n.target.minLength;if(t>0)return e.parameters.length-1+t}}return e.minArgumentCount}function Nb(e){if(e.hasRestParameter){var n=Xi(e.parameters[e.parameters.length-1]);return!Xm(n)||n.target.hasRestElement}return!1}function Ib(e){if(e.hasRestParameter){var n=Xi(e.parameters[e.parameters.length-1]);return Xm(n)?Qm(n):n}}function wb(e){var n=Ib(e);return!n||Om(n)||ai(n)?void 0:n}function Mb(e){return Ob(e,Te)}function Ob(e,n){return e.parameters.length>0?Cb(e,0):n}function Pb(n,t,r){for(var a=n.parameters.length-(n.hasRestParameter?1:0),i=0;i<a;i++){var o=n.parameters[i].valueDeclaration;if(o.type){var s=e.getEffectiveTypeAnnotationNode(o);s&&jf(r.inferences,Rd(s),Cb(t,i))}}}function Rb(n,t){if(n.typeParameters=t.typeParameters,t.thisParameter){var r=n.thisParameter;(!r||r.valueDeclaration&&!r.valueDeclaration.type)&&(r||(n.thisParameter=df(t.thisParameter,void 0)),Gb(n.thisParameter,Xi(t.thisParameter)))}for(var a=n.parameters.length-(n.hasRestParameter?1:0),i=0;i<a;i++){r=n.parameters[i];if(!e.getEffectiveTypeAnnotationNode(r.valueDeclaration)){var o=Cb(t,i);Gb(r,o)}}if(n.hasRestParameter){r=e.last(n.parameters);if(Pt(r)||!e.getEffectiveTypeAnnotationNode(r.valueDeclaration)){o=Db(t,a);Gb(r,o)}}}function Fb(n){for(var t=0,r=n.elements;t<r.length;t++){var a=r[t];e.isOmittedExpression(a)||(72===a.name.kind?qt(ra(a)).type=mi(a):Fb(a.name))}}function Gb(e,n){var t=qt(e);if(!t.type){t.type=n;var r=e.valueDeclaration;72!==r.name.kind&&(t.type===ke&&(t.type=ki(r.name)),Fb(r.name))}}function Vb(e){var n=Bl(!0);return n!==Me?(e=RT(e)||ke,cl(n,[e])):ke}function Bb(e){var n=Kl(!0);return n!==Me?(e=RT(e)||ke,cl(n,[e])):ke}function Kb(n,t){var r=Vb(t);return r===ke?(It(n,e.isImportCall(n)?e.Diagnostics.A_dynamic_import_call_returns_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option:e.Diagnostics.An_async_function_or_method_must_return_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option),oe):(jl(!0)||It(n,e.isImportCall(n)?e.Diagnostics.A_dynamic_import_call_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option:e.Diagnostics.An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option),r)}function jb(n,t){if(!n.body)return oe;var r,a=e.getFunctionFlags(n);if(218!==n.body.kind)r=RE(n.body,t),2&a&&(r=PT(r,n,e.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member));else{var i=Jb(n,t);if(1&a){if(i=e.concatenate(Ub(n,t),i),!i||0===i.length){var o=2&a?au(re):ou(re);return R&&It(n.asteriskToken,e.Diagnostics.Generator_implicitly_has_type_0_because_it_does_not_yield_any_values_Consider_supplying_a_return_type,Fa(o)),o}}else{if(!i)return 2&a?Kb(n,Te):Te;if(0===i.length)return 2&a?Kb(n,Ee):Ee}r=Au(i,2)}var s=wv(n);if(s||xf(n,r),Um(r)){var c=s?s===Cc(n)?r:Fc(s):void 0;if(c)switch(3&a){case 3:c=qS(c,!0);break;case 1:c=qS(c,!1);break;case 2:c=OT(c);break}r=Jm(r,c)}var l=bf(r);switch(3&a){case 3:return au(l);case 1:return ou(l);case 2:return Vb(l);default:return l}}function Ub(n,t){var r=[],a=0!==(2&e.getFunctionFlags(n));return e.forEachYieldExpression(n.body,function(n){e.pushIfUnique(r,Hb(n,a,t))}),r}function Hb(n,t,r){var a=n.expression||n,i=n.expression?qE(n.expression,r):le,o=n.asteriskToken?KS(i,a,!1,t):i;return t?RT(o,a,n.asteriskToken?e.Diagnostics.Type_of_iterated_elements_of_a_yield_Asterisk_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member:e.Diagnostics.Type_of_yield_operand_in_an_async_generator_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member):o}function Wb(e,n,t,r){var a=0;if(r){for(var i=n;i<t.length;i++)a|=vt.get(t[i])||32768;for(i=e;i<n;i++)a&=~(vt.get(t[i])||0);for(i=0;i<e;i++)a|=vt.get(t[i])||32768}else{for(i=e;i<n;i++)a|=gt.get(t[i])||128;for(i=0;i<e;i++)a&=~(gt.get(t[i])||0)}return a}function qb(n){if(!n.possiblyExhaustive)return!1;if(199===n.expression.kind){var t=HE(n.expression.expression),r=j_(n),a=Wb(0,0,r,!0),i=qs(t)||t;return!!(131072&J_(i,function(e){return(h_(e)&a)===a}).flags)}var o=HE(n.expression);if(!Hm(o))return!1;var s=K_(n);return!(!s.length||e.some(s,jm))&&U_(X_(o,Ld),s)}function zb(n){return!!(128&n.flags)&&!e.some(n.body.statements,function(e){return 232===e.kind&&qb(e)})}function Jb(n,t){var r=e.getFunctionFlags(n),a=[],i=zb(n),o=!1;if(e.forEachReturnStatement(n.body,function(s){var c=s.expression;if(c){var l=RE(c,t);2&r&&(l=PT(l,n,e.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member)),131072&l.flags&&(o=!0),e.pushIfUnique(a,l)}else i=!0}),0!==a.length||i||!o&&!Xb(n))return!(w&&a.length&&i)||ib(n)&&a.some(function(e){return e.symbol===n.symbol})||e.pushIfUnique(a,ce),a}function Xb(e){switch(e.kind){case 196:case 197:return!0;case 156:return 188===e.parent.kind;default:return!1}}function Yb(n,t){if(a&&(!t||!mE(t,16385))&&155!==n.kind&&!e.nodeIsMissing(n.body)&&218===n.body.kind&&zb(n)){var r=256&n.flags;if(t&&131072&t.flags)It(e.getEffectiveReturnTypeNode(n),e.Diagnostics.A_function_returning_never_cannot_have_a_reachable_end_point);else if(t&&!r)It(e.getEffectiveReturnTypeNode(n),e.Diagnostics.A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value);else if(t&&w&&!Lp(ce,t))It(e.getEffectiveReturnTypeNode(n),e.Diagnostics.Function_lacks_ending_return_statement_and_return_type_does_not_include_undefined);else if(L.noImplicitReturns){if(!t){if(!r)return;var i=Fc(Cc(n));if(JS(n,i))return}It(e.getEffectiveReturnTypeNode(n)||n,e.Diagnostics.Not_all_code_paths_return_a_value)}}}function Qb(n,t){if(e.Debug.assert(156!==n.kind||e.isObjectLiteralMethod(n)),Yx(n),1===t&&yp(n)){if(!e.getEffectiveReturnTypeNode(n)&&bp(n)){var r=zt(n);if(r.contextFreeType)return r.contextFreeType;var a=jb(n,t),i=$o(void 0,void 0,void 0,e.emptyArray,a,void 0,0,!1,!1),o=ba(n.symbol,A,[i],e.emptyArray,void 0,void 0);return o.flags|=536870912,r.contextFreeType=o}return Oe}var s=NA(n);s||196!==n.kind||HA(n);var c=zt(n),l=Xi(ta(n.symbol));if(ai(l))return l;if(!(1024&c.flags)){var u=Ov(n);if(!(1024&c.flags)){if(c.flags|=1024,u){var d=lc(l,0)[0];if(yp(n)){var p=Ev(n);2===t&&Pb(d,u,p);var m=p===D?u:np(u,p);Rb(d,m)}if(!Gc(n)&&!d.resolvedReturnType){a=jb(n,t);d.resolvedReturnType||(d.resolvedReturnType=a)}}eT(n)}}return l}function Zb(e,n){var t=Gc(e);return t&&2===(3&n)?RT(t)||oe:t}function $b(n){e.Debug.assert(156!==n.kind||e.isObjectLiteralMethod(n));var t=e.getFunctionFlags(n),r=Zb(n,t);if(0===(1&t)&&Yb(n,r),n.body)if(e.getEffectiveReturnTypeNode(n)||Fc(Cc(n)),218===n.body.kind)qx(n.body);else{var a=qE(n.body);if(r)if(2===(3&t)){var i=PT(a,n.body,e.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member);Mp(i,r,n.body,n.body)}else Mp(a,r,n.body,n.body)}}function eE(e,n,t){return!!Lp(n,Le)||(It(e,t),!1)}function nE(n){if(!e.isCallExpression(n))return!1;if(!e.isBindableObjectDefinePropertyCall(n))return!1;var t=RE(n.arguments[2]),r=ti(t,"value");if(r){var a=sc(t,"writable"),i=a&&Xi(a);if(!i||i===_e||i===ge)return!0;if(a&&a.valueDeclaration&&e.isPropertyAssignment(a.valueDeclaration)){var o=a.valueDeclaration.initializer,s=qE(o);if(s===_e||s===ge)return!0}return!1}var c=sc(t,"set");return!c}function tE(n){return!!(8&e.getCheckFlags(n)||4&n.flags&&64&e.getDeclarationModifierFlagsFromSymbol(n)||3&n.flags&&2&Ay(n)||98304&n.flags&&!(65536&n.flags)||8&n.flags||e.some(n.declarations,nE))}function rE(n,t){if(tE(t)){if(4&t.flags&&(189===n.kind||190===n.kind)&&100===n.expression.kind){var r=e.getContainingFunction(n);return!r||157!==r.kind||(!t.valueDeclaration||!(r.parent===t.valueDeclaration.parent||r===t.valueDeclaration.parent))}return!0}return!1}function aE(n){if(189===n.kind||190===n.kind){var t=e.skipParentheses(n.expression);if(72===t.kind){var r=zt(t).resolvedSymbol;if(2097152&r.flags){var a=pr(r);return!!a&&251===a.kind}}}return!1}function iE(n,t){var r=e.skipOuterExpressions(n,3);return 72===r.kind||189===r.kind||190===r.kind||(It(n,t),!1)}function oE(n){qE(n.expression);var t=e.skipParentheses(n.expression);if(189!==t.kind&&190!==t.kind)return It(t,e.Diagnostics.The_operand_of_a_delete_operator_must_be_a_property_reference),he;var r=zt(t),a=ca(r.resolvedSymbol);return a&&tE(a)&&It(t,e.Diagnostics.The_operand_of_a_delete_operator_cannot_be_a_read_only_property),he}function sE(e){return qE(e.expression),ht}function cE(e){return qE(e.expression),le}function lE(n){a&&(16384&n.flags||gD(n,e.Diagnostics.await_expression_is_only_allowed_within_an_async_function),ev(n)&&It(n,e.Diagnostics.await_expressions_cannot_be_used_in_a_parameter_initializer));var t=qE(n.expression);return PT(t,n,e.Diagnostics.Type_of_await_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member)}function uE(n){var t=qE(n.operand);if(t===Se)return Se;switch(n.operand.kind){case 8:switch(n.operator){case 39:return Dd(Nd(-n.operand.text));case 38:return Dd(Nd(+n.operand.text))}break;case 9:if(39===n.operator)return Dd(Nd({negative:!0,base10Value:e.parsePseudoBigInt(n.operand.text)}))}switch(n.operator){case 38:case 39:case 53:return My(t,n.operand),mE(t,12288)&&It(n.operand,e.Diagnostics.The_0_operator_cannot_be_applied_to_type_symbol,e.tokenToString(n.operator)),38===n.operator?(mE(t,2112)&&It(n.operand,e.Diagnostics.Operator_0_cannot_be_applied_to_type_1,e.tokenToString(n.operator),Fa(t)),me):pE(t);case 52:PS(n.operand);var r=12582912&h_(t);return 4194304===r?_e:8388608===r?ve:he;case 44:case 45:var a=eE(n.operand,My(t,n.operand),e.Diagnostics.An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type);return a&&iE(n.operand,e.Diagnostics.The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access),pE(t)}return oe}function dE(n){var t=qE(n.operand);if(t===Se)return Se;var r=eE(n.operand,My(t,n.operand),e.Diagnostics.An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type);return r&&iE(n.operand,e.Diagnostics.The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access),pE(t)}function pE(e){return mE(e,2112)?fE(e,3)||mE(e,296)?Le:fe:me}function mE(e,n){if(e.flags&n&-134217729||134217728&n&&Is(e))return!0;if(3145728&e.flags)for(var t=e.types,r=0,a=t;r<a.length;r++){var i=a[r];if(mE(i,n))return!0}return!1}function fE(e,n,t){return!!(e.flags&n)||!(t&&114691&e.flags)&&(!!(296&n)&&Lp(e,me)||!!(2112&n)&&Lp(e,fe)||!!(132&n)&&Lp(e,pe)||!!(528&n)&&Lp(e,he)||!!(16384&n)&&Lp(e,Ee)||!!(131072&n)&&Lp(e,Te)||!!(65536&n)&&Lp(e,ue)||!!(32768&n)&&Lp(e,ce)||!!(4096&n)&&Lp(e,be)||!!(67108864&n)&&Lp(e,Ce))}function _E(n,t,r){return 1048576&n.flags?e.every(n.types,function(e){return _E(e,t,r)}):fE(n,t,r)}function gE(n){return!!(16&e.getObjectFlags(n))&&!!n.symbol&&vE(n.symbol)}function vE(e){return 0!==(128&e.flags)}function yE(n,t,r,a){return r===Se||a===Se?Se:(!ai(r)&&_E(r,131068)&&It(n,e.Diagnostics.The_left_hand_side_of_an_instanceof_expression_must_be_of_type_any_an_object_type_or_a_type_parameter),ai(a)||kC(a)||Dp(a,He)||It(t,e.Diagnostics.The_right_hand_side_of_an_instanceof_expression_must_be_of_type_any_or_of_a_type_assignable_to_the_Function_interface_type),he)}function hE(n,t,r,a){return r===Se||a===Se?Se:(r=My(r,n),a=My(a,t),Np(r,pe)||fE(r,12584)||It(n,e.Diagnostics.The_left_hand_side_of_an_in_expression_must_be_of_type_any_string_number_or_symbol),_E(a,126091264)||It(t,e.Diagnostics.The_right_hand_side_of_an_in_expression_must_be_of_type_any_an_object_type_or_a_type_parameter),he)}function bE(e,n,t){var r=e.properties;if(w&&0===r.length)return My(n,e);for(var a=0,i=r;a<i.length;a++){var o=i[a];EE(n,o,r,t)}return n}function EE(n,t,r,a){if(void 0===a&&(a=!1),275===t.kind||276===t.kind){var i=t.name,o=Bu(i);if(Fo(o)){var s=Uo(o),c=sc(n,s);c&&(Yy(c,t,a),Ly(t,!1,n,c))}var l=ad(n,o,i),u=ci(t,l);return xE(276===t.kind?t:t.initializer,u)}if(277===t.kind){k<6&&_A(t,4);var d=[];if(r)for(var p=0;p<r.length-1;p++)d.push(r[p].name);u=si(n,d,n.symbol);return CA(r,e.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma),xE(t.expression,u)}It(t,e.Diagnostics.Property_assignment_expected)}function TE(e,n,t){var r=e.elements;k<2&&L.downlevelIteration&&_A(e,512);for(var a=KS(n,e,!1,!1)||oe,i=0;i<r.length;i++)SE(e,n,i,a,t);return n}function SE(n,t,r,a,i){var o=n.elements,s=o[r];if(210!==s.kind){if(208!==s.kind){var c=Nd(r);if(Fm(t)){var l=ad(t,c,Dh(s,c)),u=ci(s,l);return xE(s,u,i)}return xE(s,a,i)}if(r<o.length-1)It(s,e.Diagnostics.A_rest_element_must_be_last_in_a_destructuring_pattern);else{var d=s.expression;if(204!==d.kind||59!==d.operatorToken.kind){CA(n.elements,e.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma);u=z_(t,Xm)?X_(t,function(e){return fu(e,r)}):su(a);return xE(d,u,i)}It(d.operatorToken,e.Diagnostics.A_rest_element_cannot_have_an_initializer)}}}function xE(e,n,t,r){var a;if(276===e.kind){var i=e;i.objectAssignmentInitializer&&(!w||32768&nf(qE(i.objectAssignmentInitializer))||(n=b_(n,524288)),kE(i.name,i.equalsToken,i.objectAssignmentInitializer,t)),a=e.name}else a=e;return 204===a.kind&&59===a.operatorToken.kind&&(LE(a,t),a=a.left),188===a.kind?bE(a,n,r):187===a.kind?TE(a,n,t):CE(a,n,t)}function CE(n,t,r){var a=qE(n,r),i=277===n.parent.kind?e.Diagnostics.The_target_of_an_object_rest_assignment_must_be_a_variable_or_a_property_access:e.Diagnostics.The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access;return iE(n,i)&&Mp(t,a,n,n),t}function AE(n){switch(n=e.skipParentheses(n),n.kind){case 72:case 10:case 13:case 193:case 206:case 14:case 8:case 9:case 102:case 87:case 96:case 141:case 196:case 209:case 197:case 187:case 188:case 199:case 213:case 261:case 260:return!0;case 205:return AE(n.whenTrue)&&AE(n.whenFalse);case 204:return!e.isAssignmentOperator(n.operatorToken.kind)&&(AE(n.left)&&AE(n.right));case 202:case 203:switch(n.operator){case 52:case 38:case 39:case 53:return!0}return!1;case 200:case 194:case 212:default:return!1}}function DE(e,n){return 0!==(98304&n.flags)||Np(e,n)}function LE(n,t){return e.isInJSFile(n)&&e.getAssignedExpandoInitializer(n)?qE(n.right,t):kE(n.left,n.operatorToken,n.right,t,n)}function kE(n,t,r,i,o){var s,c=t.kind;if(59===c&&(188===n.kind||187===n.kind))return xE(n,qE(r,i),i,100===r.kind);s=54===c||55===c?PS(n,i):qE(n,i);var l=qE(r,i);switch(c){case 40:case 41:case 62:case 63:case 42:case 64:case 43:case 65:case 39:case 61:case 46:case 66:case 47:case 67:case 48:case 68:case 50:case 70:case 51:case 71:case 49:case 69:if(s===Se||l===Se)return Se;s=My(s,n),l=My(l,r);var u=void 0;if(528&s.flags&&528&l.flags&&void 0!==(u=E(t.kind)))return It(o||t,e.Diagnostics.The_0_operator_is_not_allowed_for_boolean_types_Consider_using_1_instead,e.tokenToString(t.kind),e.tokenToString(u)),me;var d,p=eE(n,s,e.Diagnostics.The_left_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type),m=eE(r,l,e.Diagnostics.The_right_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type);if(fE(s,3)&&fE(l,3)||!mE(s,2112)&&!mE(l,2112))d=me;else if(fE(s,2112)&&fE(l,2112)){switch(c){case 48:case 68:x()}d=fe}else x(),d=oe;return p&&m&&T(d),d;case 38:case 60:if(s===Se||l===Se)return Se;fE(s,132)||fE(l,132)||(s=My(s,n),l=My(l,r));var f=void 0;return fE(s,296,!0)&&fE(l,296,!0)?f=me:fE(s,2112,!0)&&fE(l,2112,!0)?f=fe:fE(s,132,!0)||fE(l,132,!0)?f=pe:(ai(s)||ai(l))&&(f=s===oe||l===oe?oe:re),f&&!b(c)?f:f?(60===c&&T(f),f):(x(),re);case 28:case 30:case 31:case 32:return b(c)&&(s=Wm(My(s,n)),l=Wm(My(l,r)),Np(s,l)||Np(l,s)||Lp(s,Le)&&Lp(l,Le)||x()),he;case 33:case 34:case 35:case 36:var _=Hm(s),g=Hm(l);return _&&g||(s=_?Wm(s):s,l=g?Wm(l):l),DE(s,l)||DE(l,s)||x(),he;case 94:return yE(n,r,s,l);case 93:return hE(n,r,s,l);case 54:return 4194304&h_(s)?Au([rf(w?s:Wm(l)),l]):s;case 55:return 8388608&h_(s)?Au([tf(s),l],2):s;case 59:var v=e.isBinaryExpression(n.parent)?e.getAssignmentDeclarationKind(n.parent):0;return y(v,l),S(v)?(524288&l.flags&&(2===v||6===v||em(l)||y_(l)||1&e.getObjectFlags(l))||T(l),s):(T(l),mf(l));case 27:return L.allowUnreachableCode||!AE(n)||h(r)||It(n,e.Diagnostics.Left_side_of_comma_operator_is_unused_and_has_no_side_effects),l;default:return e.Debug.fail()}function y(n,t){if(2===n)for(var r=0,a=Ms(t);r<a.length;r++){var i=a[r],o=Xi(i);if(o.symbol&&32&o.symbol.flags){var s=i.escapedName,c=Zt(i.valueDeclaration,s,67897832,void 0,s,!1);if(c&&c.declarations.some(e.isJSDocTypedefTag))return yD(c.declarations[0],e.Diagnostics.Duplicate_identifier_0,e.unescapeLeadingUnderscores(s)),yD(i.valueDeclaration,e.Diagnostics.Duplicate_identifier_0,e.unescapeLeadingUnderscores(s))}}}function h(e){return 72===e.kind&&"eval"===e.escapedText}function b(t){var a=mE(s,12288)?n:mE(l,12288)?r:void 0;return!a||(It(a,e.Diagnostics.The_0_operator_cannot_be_applied_to_type_symbol,e.tokenToString(t)),!1)}function E(e){switch(e){case 50:case 70:return 55;case 51:case 71:return 36;case 49:case 69:return 54;default:return}}function T(t){a&&e.isAssignmentOperator(c)&&(!iE(n,e.Diagnostics.The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access)||e.isIdentifier(n)&&"exports"===e.unescapeLeadingUnderscores(n.escapedText)||Mp(t,s,n,r))}function S(t){switch(t){case 2:return!0;case 1:case 5:case 6:case 3:case 4:var a=ra(n),i=e.getAssignedExpandoInitializer(r);return i&&e.isObjectLiteralExpression(i)&&a&&e.hasEntries(a.exports);default:return!1}}function x(){var n=Fa(s),r=Fa(l),a=o||t;C(a,n,r)||It(a,e.Diagnostics.Operator_0_cannot_be_applied_to_types_1_and_2,e.tokenToString(t.kind),n,r)}function C(n,r,a){switch(t.kind){case 35:case 33:return It(n,e.Diagnostics.This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap,"false",r,a);case 36:case 34:return It(n,e.Diagnostics.This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap,"true",r,a)}}}function NE(n){var t=n,r=n.parent;while(r){if(e.isFunctionLike(r)&&t===r.body)return!1;if(e.isClassLike(t))return!0;t=r,r=r.parent}return!1}function IE(n){a&&(4096&n.flags&&!NE(n)||gD(n,e.Diagnostics.A_yield_expression_is_only_allowed_in_a_generator_body),ev(n)&&It(n,e.Diagnostics.yield_expressions_cannot_be_used_in_a_parameter_initializer));var t=e.getContainingFunction(n);if(!t)return re;var r=e.getFunctionFlags(t);if(!(1&r))return re;n.asteriskToken&&(3===(3&r)&&k<6&&_A(n,26624),1===(3&r)&&k<2&&L.downlevelIteration&&_A(n,256));var i=0!==(2&r),o=Hb(n,i),s=Gc(t);if(s){var c=qS(s,i)||re;Mp(o,c,n.expression||n,n.expression)}return re}function wE(e,n){PS(e.condition);var t=qE(e.whenTrue,n),r=qE(e.whenFalse,n);return Au([t,r],2)}function ME(n){return e.forEach(n.templateSpans,function(n){mE(qE(n.expression),12288)&&It(n.expression,e.Diagnostics.Implicit_conversion_of_a_symbol_to_a_string_will_fail_at_runtime_Consider_wrapping_this_expression_in_String)}),pe}function OE(n){return 268!==n.kind||e.isJsxSelfClosingElement(n.parent)?n:n.parent.parent}function PE(e,n,t){var r=OE(e),a=r.contextualType,i=r.contextualMapper;r.contextualType=n,r.contextualMapper=t;var o=t===D?1:t?2:3,s=qE(e,o);return r.contextualType=a,r.contextualMapper=i,s}function RE(e,n){var t=zt(e);if(!t.resolvedType){if(n)return qE(e,n);var r=Pn;Pn=Rn,t.resolvedType=qE(e,n),Pn=r}return t.resolvedType}function FE(n){return n=e.skipParentheses(n),194===n.kind||212===n.kind}function GE(n){var t=e.getEffectiveInitializer(n),r=HE(t,!0),a=2&e.getCombinedNodeFlags(n)||e.isDeclarationReadonly(n)||FE(t)||VE(r,bv(t))?r:qm(r);if(e.isInJSFile(n)){if(98304&a.flags)return Sf(n,re),re;if(Gm(a))return Sf(n,en),en}return a}function VE(n,t){if(t){if(3145728&t.flags){var r=t.types;return e.some(r,function(e){return VE(n,e)})}if(58982400&t.flags){var a=qs(t)||ke;return mE(a,4)&&mE(n,128)||mE(a,8)&&mE(n,256)||mE(a,64)&&mE(n,2048)||mE(a,4096)&&mE(n,8192)||VE(n,a)}return!!(4194432&t.flags&&mE(n,128)||256&t.flags&&mE(n,256)||2048&t.flags&&mE(n,2048)||512&t.flags&&mE(n,512)||8192&t.flags&&mE(n,8192))}return!1}function BE(e,n,t,r){2===arguments.length&&(t=bv(e));var a=qE(e,n,r);return FE(e)?a:Jm(a,t)}function KE(e,n){return 149===e.name.kind&&Hv(e.name),BE(e.initializer,n)}function jE(e,n){nD(e),149===e.name.kind&&Hv(e.name);var t=Qb(e,n);return UE(e,t,n)}function UE(e,n,t){if(2===t){var r=gh(n);if(r&&r.typeParameters){var a=hv(e);if(a){var i=gh(lf(a));if(i&&!i.typeParameters)return Qc(vh(r,i,Ev(e)))}}}return n}function HE(n,t){var r=e.skipParentheses(n);if(191!==r.kind||98===r.expression.kind||e.isRequireCall(r,!0)||pb(r)){if(194===r.kind||212===r.kind)return Rd(r.type)}else{var a=Iy(r.expression),i=gh(a);if(i&&!i.typeParameters)return Fc(i)}return t?RE(n):qE(n)}function WE(e){var n=zt(e);if(n.contextFreeType)return n.contextFreeType;var t=e.contextualType;e.contextualType=re;var r=n.contextFreeType=qE(e,1);return e.contextualType=t,r}function qE(n,t,r){var a;if(148===n.kind)a=Py(n);else{var i=JE(n,t,r);a=UE(n,i,t)}if(gE(a)){var o=189===n.parent.kind&&n.parent.expression===n||190===n.parent.kind&&n.parent.expression===n||(72===n.kind||148===n.kind)&&_C(n)||167===n.parent.kind&&n.parent.exprName===n;o||It(n,e.Diagnostics.const_enums_can_only_be_used_in_property_or_index_access_expressions_or_the_right_hand_side_of_an_import_declaration_or_export_assignment_or_type_query)}return a}function zE(n,t){var r=e.isInJSFile(n)?e.getJSDocTypeTag(n):void 0;return r?yb(r,r.typeExpression.type,n.expression,t):qE(n.expression,t)}function JE(n,t,r){switch(n.kind){case 72:return Ag(n);case 100:return Fg(n);case 98:return jg(n);case 96:return de;case 14:case 10:return Dd(Nd(n.text));case 8:return AD(n),Dd(Nd(+n.text));case 9:return DD(n),Dd(Ku(n));case 102:return ve;case 87:return _e;case 206:return ME(n);case 13:return Ze;case 187:return Fv(n,t,r);case 188:return zv(n,t);case 189:return Oy(n);case 190:return ih(n);case 191:if(92===n.expression.kind)return mb(n);case 192:return db(n);case 193:return gb(n);case 195:return zE(n,t);case 209:return cx(n);case 196:case 197:return Qb(n,t);case 199:return sE(n);case 194:case 212:return vb(n);case 213:return hb(n);case 214:return bb(n);case 198:return oE(n);case 200:return cE(n);case 201:return lE(n);case 202:return uE(n);case 203:return dE(n);case 204:return LE(n,t);case 205:return wE(n,t);case 208:return Pv(n,t);case 210:return le;case 207:return IE(n);case 215:return n.type;case 270:return Cy(n,t);case 260:return Zv(n,t);case 261:return Yv(n,t);case 264:return $v(n);case 268:return iy(n,t);case 262:e.Debug.fail("Shouldn't ever directly check a JsxOpeningElement")}return oe}function XE(n){n.expression&&gD(n.expression,e.Diagnostics.Type_expected),qx(n.constraint),qx(n.default);var t=Co(ra(n));Js(t)||It(e.getEffectiveConstraintOfTypeParameter(n),e.Diagnostics.Type_parameter_0_has_a_circular_constraint,Fa(t)),$s(t)||It(n.default,e.Diagnostics.Type_parameter_0_has_a_circular_default,Fa(t));var r=Bs(t),i=Zs(t);r&&i&&wp(i,Xo(r,i),n.default,e.Diagnostics.Type_0_does_not_satisfy_the_constraint_1),a&&tx(n.name,e.Diagnostics.Type_parameter_name_cannot_be_0)}function YE(n){yA(n),CS(n);var t=e.getContainingFunction(n);e.hasModifier(n,92)&&(157===t.kind&&e.nodeIsPresent(t.body)||It(n,e.Diagnostics.A_parameter_property_is_only_allowed_in_a_constructor_implementation)),n.questionToken&&e.isBindingPattern(n.name)&&t.body&&It(n,e.Diagnostics.A_binding_pattern_parameter_cannot_be_optional_in_an_implementation_signature),n.name&&e.isIdentifier(n.name)&&("this"===n.name.escapedText||"new"===n.name.escapedText)&&(0!==t.parameters.indexOf(n)&&It(n,e.Diagnostics.A_0_parameter_must_be_the_first_parameter,n.name.escapedText),157!==t.kind&&161!==t.kind&&166!==t.kind||It(n,e.Diagnostics.A_constructor_cannot_have_a_this_parameter),197===t.kind&&It(n,e.Diagnostics.An_arrow_function_cannot_have_a_this_parameter)),!n.dotDotDotToken||e.isBindingPattern(n.name)||Lp(Xi(n.symbol),en)||It(n,e.Diagnostics.A_rest_parameter_must_be_of_an_array_type)}function QE(n){var t=ZE(n);if(t){var r=Oc(Cc(t));if(r){qx(n.type);var a=n.parameterName;if(e.isThisTypePredicate(r))Pd(a);else if(r.parameterIndex>=0)if(t.parameters[r.parameterIndex].dotDotDotToken)It(a,e.Diagnostics.A_type_predicate_cannot_reference_a_rest_parameter);else{var i=function(){return e.chainDiagnosticMessages(void 0,e.Diagnostics.A_type_predicate_s_type_must_be_assignable_to_its_parameter_s_type)};wp(r.type,TC(t.parameters[r.parameterIndex]),n.type,void 0,i)}else if(a){for(var o=!1,s=0,c=t.parameters;s<c.length;s++){var l=c[s].name;if(e.isBindingPattern(l)&&$E(l,a,r.parameterName)){o=!0;break}}o||It(n.parameterName,e.Diagnostics.Cannot_find_parameter_0,r.parameterName)}}}else It(n,e.Diagnostics.A_type_predicate_is_only_allowed_in_return_type_position_for_functions_and_methods)}function ZE(e){switch(e.parent.kind){case 197:case 160:case 239:case 196:case 165:case 156:case 155:var n=e.parent;if(e===n.type)return n}}function $E(n,t,r){for(var a=0,i=n.elements;a<i.length;a++){var o=i[a];if(!e.isOmittedExpression(o)){var s=o.name;if(72===s.kind&&s.escapedText===r)return It(t,e.Diagnostics.A_type_predicate_cannot_reference_element_0_in_a_binding_pattern,r),!0;if((185===s.kind||184===s.kind)&&$E(s,t,r))return!0}}}function eT(n){162===n.kind?OA(n):165!==n.kind&&239!==n.kind&&166!==n.kind&&160!==n.kind&&157!==n.kind&&161!==n.kind||NA(n);var t=e.getFunctionFlags(n);if(4&t||(3===(3&t)&&k<6&&_A(n,6144),2===(3&t)&&k<4&&_A(n,64),0!==(3&t)&&k<2&&_A(n,128)),ax(n.typeParameters),e.forEach(n.parameters,YE),n.type&&qx(n.type),a){gS(n);var r=e.getEffectiveReturnTypeNode(n);if(R&&!r)switch(n.kind){case 161:It(n,e.Diagnostics.Construct_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type);break;case 160:It(n,e.Diagnostics.Call_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type);break}if(r){var i=e.getFunctionFlags(n);if(1===(5&i)){var o=Rd(r);if(o===Ee)It(r,e.Diagnostics.A_generator_cannot_have_a_void_type_annotation);else{var s=qS(o,0!==(2&i))||re,c=2&i?au(s):ou(s);wp(c,o,r)}}else 2===(3&i)&&FT(n,r)}162!==n.kind&&289!==n.kind&&nS(n)}}function nT(n){var t;(function(e){e[e["Getter"]=1]="Getter",e[e["Setter"]=2]="Setter",e[e["Method"]=4]="Method",e[e["Property"]=3]="Property"})(t||(t={}));for(var r=e.createUnderscoreEscapedMap(),a=e.createUnderscoreEscapedMap(),i=0,o=n.members;i<o.length;i++){var s=o[i];if(157===s.kind)for(var c=0,l=s.parameters;c<l.length;c++){var u=l[c];e.isParameterPropertyDeclaration(u)&&!e.isBindingPattern(u.name)&&_(r,u.name,u.name.escapedText,3)}else{var d=e.hasModifier(s,32),p=d?a:r,m=s.name,f=m&&e.getPropertyNameForPropertyNameNode(m);if(m&&f)switch(s.kind){case 158:_(p,m,f,1);break;case 159:_(p,m,f,2);break;case 154:_(p,m,f,3);break;case 156:_(p,m,f,4);break}}}function _(n,t,r,a){var i=n.get(r);i?4&i?4!==a&&It(t,e.Diagnostics.Duplicate_identifier_0,e.getTextOfNode(t)):i&a?It(t,e.Diagnostics.Duplicate_identifier_0,e.getTextOfNode(t)):n.set(r,i|a):n.set(r,a)}}function tT(n){for(var t=0,r=n.members;t<r.length;t++){var a=r[t],i=a.name,o=e.hasModifier(a,32);if(o&&i){var s=e.getPropertyNameForPropertyNameNode(i);switch(s){case"name":case"length":case"caller":case"arguments":case"prototype":var c=e.Diagnostics.Static_property_0_conflicts_with_built_in_property_Function_0_of_constructor_function_1,l=za(ra(n));It(i,c,s,l);break}}}}function rT(n){for(var t=e.createMap(),r=0,a=n.members;r<a.length;r++){var i=a[r];if(153===i.kind){var o=void 0,s=i.name;switch(s.kind){case 10:case 8:o=s.text;break;case 72:o=e.idText(s);break;default:continue}t.get(o)?(It(e.getNameOfDeclaration(i.symbol.valueDeclaration),e.Diagnostics.Duplicate_identifier_0,o),It(i.name,e.Diagnostics.Duplicate_identifier_0,o)):t.set(o,!0)}}}function aT(n){if(241===n.kind){var t=ra(n);if(t.declarations.length>0&&t.declarations[0]!==n)return}var r=Zc(ra(n));if(r)for(var a=!1,i=!1,o=0,s=r.declarations;o<s.length;o++){var c=s[o],l=c;if(1===l.parameters.length&&l.parameters[0].type)switch(l.parameters[0].type.kind){case 138:i?It(l,e.Diagnostics.Duplicate_string_index_signature):i=!0;break;case 135:a?It(l,e.Diagnostics.Duplicate_number_index_signature):a=!0;break}}}function iT(e){yA(e)||ED(e)||UA(e.name),CS(e)}function oT(n){nD(n)||UA(n.name),eS(n),e.hasModifier(n,128)&&156===n.kind&&n.body&&It(n,e.Diagnostics.Method_0_cannot_have_an_implementation_because_it_is_marked_abstract,e.declarationNameToString(n.name))}function sT(n){eT(n),hD(n)||bD(n),qx(n.body);var t=ra(n),r=e.getDeclarationOfKind(t,n.kind);if(n===r&&IT(t),!e.nodeIsMissing(n.body)&&a){var i=n.parent;if(e.getClassExtendsHeritageElement(i)){wg(n.parent,i);var o=Pg(i),s=Og(n);if(s){o&&It(s,e.Diagnostics.A_constructor_cannot_contain_a_super_call_when_its_class_extends_null);var c=e.some(n.parent.members,f)||e.some(n.parameters,function(n){return e.hasModifier(n,92)});if(c){for(var l=n.body.statements,u=void 0,d=0,p=l;d<p.length;d++){var m=p[d];if(221===m.kind&&e.isSuperCall(m.expression)){u=m;break}if(!e.isPrologueDirective(m))break}u||It(n,e.Diagnostics.A_super_call_must_be_the_first_statement_in_the_constructor_when_a_class_contains_initialized_properties_or_has_parameter_properties)}}else o||It(n,e.Diagnostics.Constructors_for_derived_classes_must_contain_a_super_call)}}function f(n){return 154===n.kind&&!e.hasModifier(n,32)&&!!n.initializer}}function cT(n){if(a){if(NA(n)||YA(n)||UA(n.name),WT(n),eT(n),158===n.kind&&!(4194304&n.flags)&&e.nodeIsPresent(n.body)&&128&n.flags&&(256&n.flags||It(n.name,e.Diagnostics.A_get_accessor_must_return_a_value)),149===n.name.kind&&Hv(n.name),!Ko(n)){var t=158===n.kind?159:158,r=e.getDeclarationOfKind(ra(n),t);if(r){var i=e.getModifierFlags(n),o=e.getModifierFlags(r);(28&i)!==(28&o)&&It(n.name,e.Diagnostics.Getter_and_setter_accessors_do_not_agree_in_visibility),(128&i)!==(128&o)&&It(n.name,e.Diagnostics.Accessors_must_both_be_abstract_or_non_abstract),lT(n,r,Fi,e.Diagnostics.get_and_set_accessor_must_have_the_same_type),lT(n,r,Vi,e.Diagnostics.get_and_set_accessor_must_have_the_same_this_type)}}var s=Bi(ra(n));158===n.kind&&Yb(n,s)}qx(n.body)}function lT(e,n,t,r){var a=t(e),i=t(n);a&&i&&!Sp(a,i)&&It(e,r)}function uT(e){WT(e)}function dT(n,t){return xc(e.map(n.typeArguments,Rd),t,Sc(t),e.isInJSFile(n))}function pT(n,t){for(var r,a,i=!0,o=0;o<t.length;o++){var s=Bs(t[o]);s&&(r||(r=dT(n,t),a=Ud(t,r)),i=i&&wp(r[o],mp(s,a),n.typeArguments[o],e.Diagnostics.Type_0_does_not_satisfy_the_constraint_1))}return i}function mT(n){var t=Dl(n);if(t!==oe){var r=zt(n).resolvedSymbol;if(r)return 524288&r.flags&&qt(r).typeParameters||(4&e.getObjectFlags(t)?t.target.localTypeParameters:void 0)}}function fT(n){RA(n,n.typeArguments),164!==n.kind||void 0===n.typeName.jsdocDotPos||e.isInJSFile(n)||e.isInJSDoc(n)||vD(n,n.typeName.jsdocDotPos,1,e.Diagnostics.JSDoc_types_can_only_be_used_inside_documentation_comments);var t=Dl(n);if(t!==oe){if(n.typeArguments&&(e.forEach(n.typeArguments,qx),a)){var r=mT(n);r&&pT(n,r)}32&t.flags&&8&zt(n).resolvedSymbol.flags&&It(n,e.Diagnostics.Enum_type_0_has_members_with_initializers_that_are_not_literals,Fa(t))}}function _T(n){var t=e.tryCast(n.parent,e.isTypeReferenceType);if(t){var r=mT(t),a=Bs(r[t.typeArguments.indexOf(n)]);return a&&mp(a,Ud(r,dT(t,r)))}}function gT(e){kl(e)}function vT(n){if(e.forEach(n.members,qx),a){var t=yd(n);nx(t),aT(n),rT(n)}}function yT(e){qx(e.elementType)}function hT(n){for(var t=n.elementTypes,r=!1,a=0;a<t.length;a++){var i=t[a];if(172===i.kind){if(a!==t.length-1){yD(i,e.Diagnostics.A_rest_element_must_be_last_in_a_tuple_type);break}Om(Rd(i))||It(i,e.Diagnostics.A_rest_element_type_must_be_an_array_type)}else if(171===i.kind)r=!0;else if(r){yD(i,e.Diagnostics.A_required_element_cannot_follow_an_optional_element);break}}e.forEach(n.elementTypes,qx)}function bT(n){e.forEach(n.types,qx)}function ET(n,t){if(!(8388608&n.flags))return n;var r=n.objectType,a=n.indexType;return Lp(a,Wu(r,!1))?(190===t.kind&&e.isAssignmentTarget(t)&&32&e.getObjectFlags(r)&&1&Ds(r)&&It(t,e.Diagnostics.Index_signature_in_type_0_only_permits_reading,Fa(r)),n):pc(rc(r),1)&&fE(a,296)?n:(It(t,e.Diagnostics.Type_0_cannot_be_used_to_index_type_1,Fa(a),Fa(r)),n)}function TT(e){qx(e.objectType),qx(e.indexType),ET(id(e),e)}function ST(n){qx(n.typeParameter),qx(n.type),n.type||Sf(n,re);var t=od(n),r=Ts(t);wp(r,De,e.getEffectiveConstraintOfTypeParameter(n.typeParameter))}function xT(e){Pd(e)}function CT(e){$A(e),qx(e.type)}function AT(n){e.forEachChild(n,qx)}function DT(n){e.findAncestor(n,function(e){return e.parent&&175===e.parent.kind&&e.parent.extendsType===e})||yD(n,e.Diagnostics.infer_declarations_are_only_permitted_in_the_extends_clause_of_a_conditional_type),qx(n.typeParameter),nS(n)}function LT(e){qx(e.argument),Rd(e)}function kT(n){return e.hasModifier(n,8)&&!!(4194304&n.flags)}function NT(n,t){var r=e.getCombinedModifierFlags(n);return 241!==n.parent.kind&&240!==n.parent.kind&&209!==n.parent.kind&&4194304&n.flags&&(2&r||e.isModuleBlock(n.parent)&&e.isModuleDeclaration(n.parent.parent)&&e.isGlobalScopeAugmentation(n.parent.parent)||(r|=1),r|=2),r&t}function IT(n){if(a){for(var t,r,i,o=155,s=0,c=o,l=!1,u=!0,d=!1,p=n.declarations,m=0!==(16384&n.flags),f=!1,_=!1,g=0,v=p;g<v.length;g++){var y=v[g],h=y,b=4194304&h.flags,E=241===h.parent.kind||168===h.parent.kind||b;if(E&&(i=void 0),239===h.kind||156===h.kind||155===h.kind||157===h.kind){var T=NT(h,o);s|=T,c&=T,l=l||e.hasQuestionToken(h),u=u&&e.hasQuestionToken(h),e.nodeIsPresent(h.body)&&t?m?_=!0:f=!0:i&&i.parent===h.parent&&i.end!==h.pos&&I(i),e.nodeIsPresent(h.body)?t||(t=h):d=!0,i=h,E||(r=h)}}if(_&&e.forEach(p,function(n){It(n,e.Diagnostics.Multiple_constructor_implementations_are_not_allowed)}),f&&e.forEach(p,function(n){It(e.getNameOfDeclaration(n),e.Diagnostics.Duplicate_function_implementation)}),!r||r.body||e.hasModifier(r,128)||r.questionToken||I(r),d&&(k(p,t,o,s,c),N(p,t,l,u),t))for(var S=Nc(n),x=Cc(t),C=0,A=S;C<A.length;C++){var D=A[C];if(!Zp(x,D)){It(D.declaration,e.Diagnostics.Overload_signature_is_not_compatible_with_function_implementation);break}}}function L(e,n){var t=void 0!==n&&n.parent===e[0].parent;return t?n:e[0]}function k(n,t,r,a,i){var o=a^i;if(0!==o){var s=NT(L(n,t),r);e.forEach(n,function(n){var t=NT(n,r)^s;1&t?It(e.getNameOfDeclaration(n),e.Diagnostics.Overload_signatures_must_all_be_exported_or_non_exported):2&t?It(e.getNameOfDeclaration(n),e.Diagnostics.Overload_signatures_must_all_be_ambient_or_non_ambient):24&t?It(e.getNameOfDeclaration(n)||n,e.Diagnostics.Overload_signatures_must_all_be_public_private_or_protected):128&t&&It(e.getNameOfDeclaration(n),e.Diagnostics.Overload_signatures_must_all_be_abstract_or_non_abstract)})}}function N(n,t,r,a){if(r!==a){var i=e.hasQuestionToken(L(n,t));e.forEach(n,function(n){var t=e.hasQuestionToken(n)!==i;t&&It(e.getNameOfDeclaration(n),e.Diagnostics.Overload_signatures_must_all_be_optional_or_required)})}}function I(n){if(!n.name||!e.nodeIsMissing(n.name)){var t=!1,r=e.forEachChild(n.parent,function(e){if(t)return e;t=e===n});if(r&&r.pos===n.end&&r.kind===n.kind){var a=r.name||r,i=r.name;if(n.name&&i&&(e.isComputedPropertyName(n.name)&&e.isComputedPropertyName(i)||!e.isComputedPropertyName(n.name)&&!e.isComputedPropertyName(i)&&e.getEscapedTextOfIdentifierOrLiteral(n.name)===e.getEscapedTextOfIdentifierOrLiteral(i))){var o=(156===n.kind||155===n.kind)&&e.hasModifier(n,32)!==e.hasModifier(r,32);if(o){var s=e.hasModifier(n,32)?e.Diagnostics.Function_overload_must_be_static:e.Diagnostics.Function_overload_must_not_be_static;It(a,s)}return}if(e.nodeIsPresent(r.body))return void It(a,e.Diagnostics.Function_implementation_name_must_be_0,e.declarationNameToString(n.name))}var c=n.name||n;m?It(c,e.Diagnostics.Constructor_implementation_is_missing):e.hasModifier(n,128)?It(c,e.Diagnostics.All_declarations_of_an_abstract_method_must_be_consecutive):It(c,e.Diagnostics.Function_implementation_is_missing_or_not_immediately_following_the_declaration)}}}function wT(n){if(a){var t=n.localSymbol;if((t||(t=ra(n),t.exportSymbol))&&e.getDeclarationOfKind(t,n.kind)===n){for(var r=0,i=0,o=0,s=0,c=t.declarations;s<c.length;s++){var l=c[s],u=y(l),d=NT(l,513);1&d?512&d?o|=u:r|=u:i|=u}var p=r|i,m=r&i,f=o&p;if(m||f)for(var _=0,g=t.declarations;_<g.length;_++){l=g[_],u=y(l);var v=e.getNameOfDeclaration(l);u&f?It(v,e.Diagnostics.Merged_declaration_0_cannot_include_a_default_export_declaration_Consider_adding_a_separate_export_default_0_declaration_instead,e.declarationNameToString(v)):u&m&&It(v,e.Diagnostics.Individual_declarations_in_merged_declaration_0_must_be_all_exported_or_all_local,e.declarationNameToString(v))}}}function y(n){var t=n;switch(t.kind){case 241:case 242:case 304:case 297:return 2;case 244:return e.isAmbientModule(t)||0!==e.getModuleInstanceState(t)?5:4;case 240:case 243:return 3;case 279:return 7;case 254:if(!e.isEntityNameExpression(t.expression))return 1;t=t.expression;case 248:case 251:case 250:var r=0,a=Nr(ra(t));return e.forEach(a.declarations,function(e){r|=y(e)}),r;case 237:case 186:case 239:case 253:return 1;default:return e.Debug.fail(e.Debug.showSyntaxKind(t))}}}function MT(e,n,t){var r=OT(e,n);return r&&RT(r,n,t)}function OT(n,t){if(!ai(n)){var r=n;if(r.promisedTypeOfPromise)return r.promisedTypeOfPromise;if(Yi(n,Bl(!1)))return r.promisedTypeOfPromise=n.typeArguments[0];var a=ti(n,"then");if(!ai(a)){var i=a?lc(a,0):e.emptyArray;if(0!==i.length){var o=b_(Au(e.map(i,Mb)),2097152);if(!ai(o)){var s=lc(o,0);if(0!==s.length)return r.promisedTypeOfPromise=Au(e.map(s,Mb),2);t&&It(t,e.Diagnostics.The_first_parameter_of_the_then_method_of_a_promise_must_be_a_callback)}}else t&&It(t,e.Diagnostics.A_promise_must_have_a_then_method)}}}function PT(e,n,t){return RT(e,n,t)||oe}function RT(n,t,r){var a=n;if(a.awaitedTypeOfType)return a.awaitedTypeOfType;if(ai(n))return a.awaitedTypeOfType=n;if(1048576&n.flags){for(var i=void 0,o=0,s=n.types;o<s.length;o++){var c=s[o];i=e.append(i,RT(c,t,r))}if(!i)return;return a.awaitedTypeOfType=Au(i)}var l=OT(n);if(l){if(n.id===l.id||at.indexOf(l.id)>=0)return void(t&&It(t,e.Diagnostics.Type_is_referenced_directly_or_indirectly_in_the_fulfillment_callback_of_its_own_then_method));at.push(n.id);var u=RT(l,t,r);if(at.pop(),!u)return;return a.awaitedTypeOfType=u}var d=ti(n,"then");if(!(d&&lc(d,0).length>0))return a.awaitedTypeOfType=n;if(t){if(!r)return e.Debug.fail();It(t,r)}}function FT(n,t){var r=Rd(t);if(k>=2){if(r===oe)return;var a=Bl(!0);if(a!==Me&&!Yi(r,a))return void It(t,e.Diagnostics.The_return_type_of_an_async_function_or_method_must_be_the_global_Promise_T_type)}else{if(VT(t),r===oe)return;var i=e.getEntityNameFromTypeNode(t);if(void 0===i)return void It(t,e.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value,Fa(r));var o=Pr(i,67220415,!0),s=o?Xi(o):oe;if(s===oe)return void(72===i.kind&&"Promise"===i.escapedText&&Qi(r)===Bl(!1)?It(t,e.Diagnostics.An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option):It(t,e.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value,e.entityNameToString(i)));var c=Ul(!0);if(c===ke)return void It(t,e.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value,e.entityNameToString(i));if(!wp(s,c,t,e.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value))return;var l=i&&wx(i),u=Xt(n.locals,l.escapedText,67220415);if(u)return void It(u.valueDeclaration,e.Diagnostics.Duplicate_identifier_0_Compiler_uses_declaration_1_to_support_async_functions,e.idText(l),e.entityNameToString(i))}PT(r,n,e.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member)}function GT(n){var t=ab(n),r=Fc(t);if(!(1&r.flags)){var a,i,o=Zh(n);switch(n.parent.kind){case 240:var s=ra(n.parent),c=Xi(s);a=Au([c,Ee]);break;case 151:a=Ee,i=e.chainDiagnosticMessages(void 0,e.Diagnostics.The_return_type_of_a_parameter_decorator_function_must_be_either_void_or_any);break;case 154:a=Ee,i=e.chainDiagnosticMessages(void 0,e.Diagnostics.The_return_type_of_a_property_decorator_function_must_be_either_void_or_any);break;case 156:case 158:case 159:var l=TC(n.parent),u=tu(l);a=Au([u,Ee]);break;default:return e.Debug.fail()}wp(r,a,n,o,function(){return i})}}function VT(n){BT(n&&e.getEntityNameFromTypeNode(n))}function BT(e){if(e){var n=wx(e),t=2097152|(72===e.kind?67897832:1920),r=Zt(n,n.escapedText,t,void 0,void 0,!0);r&&2097152&r.flags&&la(r)&&!HC(Nr(r))&&wr(r)}}function KT(n){var t=jT(n);t&&e.isEntityName(t)&&BT(t)}function jT(e){if(e)switch(e.kind){case 174:case 173:return UT(e.types);case 175:return UT([e.trueType,e.falseType]);case 177:return jT(e.type);case 164:return e.typeName}}function UT(n){for(var t,r=0,a=n;r<a.length;r++){var i=a[r];while(177===i.kind)i=i.type;if(132!==i.kind&&(w||96!==i.kind&&141!==i.kind)){var o=jT(i);if(!o)return;if(t){if(!e.isIdentifier(t)||!e.isIdentifier(o)||t.escapedText!==o.escapedText)return}else t=o}}return t}function HT(n){var t=e.getEffectiveTypeAnnotationNode(n);return e.isRestParameter(n)?e.getRestParameterElementType(t):t}function WT(n){if(n.decorators&&e.nodeCanBeDecorated(n,n.parent,n.parent.parent)){L.experimentalDecorators||It(n,e.Diagnostics.Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_to_remove_this_warning);var t=n.decorators[0];if(_A(t,8),151===n.kind&&_A(t,32),L.emitDecoratorMetadata)switch(_A(t,16),n.kind){case 240:var r=e.getFirstConstructorWithBody(n);if(r)for(var a=0,i=r.parameters;a<i.length;a++){var o=i[a];KT(HT(o))}break;case 158:case 159:var s=158===n.kind?159:158,c=e.getDeclarationOfKind(ra(n),s);KT(Ri(n)||c&&Ri(c));break;case 156:for(var l=0,u=n.parameters;l<u.length;l++){o=u[l];KT(HT(o))}KT(e.getEffectiveReturnTypeNode(n));break;case 154:KT(e.getEffectiveTypeAnnotationNode(n));break;case 151:KT(HT(n));for(var d=n.parent,p=0,m=d.parameters;p<m.length;p++){o=m[p];KT(HT(o))}break}e.forEach(n.decorators,GT)}}function qT(e){a&&(eS(e),HA(e),bS(e,e.name),ES(e,e.name))}function zT(n){n.typeExpression||It(n.name,e.Diagnostics.JSDoc_typedef_tag_should_either_have_a_type_annotation_or_be_followed_by_property_or_member_tags),n.name&&tx(n.name,e.Diagnostics.Type_alias_name_cannot_be_0),qx(n.typeExpression)}function JT(e){qx(e.constraint);for(var n=0,t=e.typeParameters;n<t.length;n++){var r=t[n];qx(r)}}function XT(e){qx(e.typeExpression)}function YT(n){if(qx(n.typeExpression),!e.getParameterSymbolFromJSDoc(n)){var t=e.getHostSignatureFromJSDoc(n);if(t){var r=e.getJSDocTags(t).filter(e.isJSDocParameterTag).indexOf(n);if(r>-1&&r<t.parameters.length&&e.isBindingPattern(t.parameters[r].name))return;kc(t)?e.findLast(e.getJSDocTags(t),e.isJSDocParameterTag)===n&&n.typeExpression&&n.typeExpression.type&&!Om(Rd(n.typeExpression.type))&&It(n.name,e.Diagnostics.JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_It_would_match_arguments_if_it_had_an_array_type,e.idText(148===n.name.kind?n.name.right:n.name)):e.isQualifiedName(n.name)?It(n.name,e.Diagnostics.Qualified_name_0_is_not_allowed_without_a_leading_param_object_1,e.entityNameToString(n.name),e.entityNameToString(n.name.left)):It(n.name,e.Diagnostics.JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name,e.idText(n.name))}}}function QT(n){!a||n.type||e.isJSDocConstructSignature(n)||Sf(n,re),eT(n)}function ZT(n){var t=e.getJSDocHost(n);if(e.isClassDeclaration(t)||e.isClassExpression(t)){var r=e.getJSDocTags(t).filter(e.isJSDocAugmentsTag);e.Debug.assert(r.length>0),r.length>1&&It(r[1],e.Diagnostics.Class_declarations_cannot_have_more_than_one_augments_or_extends_tag);var a=$T(n.class.expression),i=e.getClassExtendsHeritageElement(t);if(i){var o=$T(i.expression);o&&a.escapedText!==o.escapedText&&It(a,e.Diagnostics.JSDoc_0_1_does_not_match_the_extends_2_clause,e.idText(n.tagName),e.idText(a),e.idText(o))}}else It(t,e.Diagnostics.JSDoc_0_is_not_attached_to_a_class,e.idText(n.tagName))}function $T(e){switch(e.kind){case 72:return e;case 189:return e.name;default:return}}function eS(n){WT(n),eT(n);var t=e.getFunctionFlags(n);if(n.name&&149===n.name.kind&&Hv(n.name),!Ko(n)){var r=ra(n),i=n.localSymbol||r,o=e.find(i.declarations,function(e){return e.kind===n.kind&&!(65536&e.flags)});n===o&&IT(i),r.parent&&e.getDeclarationOfKind(r,n.kind)===n&&IT(r)}var s=155===n.kind?void 0:n.body;if(qx(s),0===(1&t)){var c=Zb(n,t);Yb(n,c)}if(a&&!e.getEffectiveReturnTypeNode(n)&&(e.nodeIsMissing(s)&&!kT(n)&&Sf(n,re),1&t&&e.nodeIsPresent(s)&&Fc(Cc(n))),e.isInJSFile(n)){var l=e.getJSDocTypeTag(n);l&&l.typeExpression&&!kv(Rd(l.typeExpression),n)&&It(l,e.Diagnostics.The_type_of_a_function_declaration_must_match_the_function_s_signature)}}function nS(n){if(a&&!(4194304&n.flags)){var t=e.getSourceFileOfNode(n),r=On.get(t.path);r||(r=[],On.set(t.path,r)),r.push(n)}}function tS(n,t){for(var r=0,a=n;r<a.length;r++){var i=a[r];switch(i.kind){case 240:case 209:iS(i,t),sS(i,t);break;case 279:case 244:case 218:case 246:case 225:case 226:case 227:dS(i,t);break;case 157:case 196:case 239:case 197:case 156:case 158:case 159:i.body&&dS(i,t),sS(i,t);break;case 155:case 160:case 161:case 165:case 166:case 242:case 241:sS(i,t);break;case 176:oS(i,t);break;default:e.Debug.assertNever(i,"Node should not have been registered for unused identifiers check")}}}function rS(n,t,r){var a=e.getNameOfDeclaration(n)||n,i=cC(n)?e.Diagnostics._0_is_declared_but_never_used:e.Diagnostics._0_is_declared_but_its_value_is_never_read;r(n,0,e.createDiagnosticForNode(a,i,t))}function aS(n){return e.isIdentifier(n)&&95===e.idText(n).charCodeAt(0)}function iS(n,t){for(var r=0,a=n.members;r<a.length;r++){var i=a[r];switch(i.kind){case 156:case 154:case 158:case 159:if(159===i.kind&&32768&i.symbol.flags)break;var o=ra(i);!o.isReferenced&&e.hasModifier(i,8)&&t(i,0,e.createDiagnosticForNode(i.name,e.Diagnostics._0_is_declared_but_its_value_is_never_read,Pa(o)));break;case 157:for(var s=0,c=i.parameters;s<c.length;s++){var l=c[s];!l.symbol.isReferenced&&e.hasModifier(l,8)&&t(l,0,e.createDiagnosticForNode(l.name,e.Diagnostics.Property_0_is_declared_but_its_value_is_never_read,e.symbolName(l.symbol)))}break;case 162:case 217:break;default:e.Debug.fail()}}}function oS(n,t){var r=n.typeParameter;cS(r)&&t(n,1,e.createDiagnosticForNode(n,e.Diagnostics._0_is_declared_but_its_value_is_never_read,e.idText(r.name)))}function sS(n,t){if(e.last(ra(n).declarations)===n)for(var r=e.getEffectiveTypeParameterDeclarations(n),a=new e.NodeSet,i=0,o=r;i<o.length;i++){var s=o[i];if(cS(s)){var c=e.idText(s.name),l=s.parent;if(176!==l.kind&&l.typeParameters.every(cS)){if(a.tryAdd(l)){var u=e.isJSDocTemplateTag(l)?e.rangeOfNode(l):e.rangeOfTypeParameters(l.typeParameters),d=1===r.length,p=d?e.Diagnostics._0_is_declared_but_its_value_is_never_read:e.Diagnostics.All_type_parameters_are_unused,m=d?c:void 0;t(s,1,e.createFileDiagnostic(e.getSourceFileOfNode(l),u.pos,u.end-u.pos,p,m))}}else t(s,1,e.createDiagnosticForNode(s,e.Diagnostics._0_is_declared_but_its_value_is_never_read,c))}}}function cS(e){return!(262144&ta(e.symbol).isReferenced)&&!aS(e.name)}function lS(e,n,t,r){var a=String(r(n)),i=e.get(a);i?i[1].push(t):e.set(a,[n,[t]])}function uS(n){return e.tryCast(e.getRootDeclaration(n),e.isParameter)}function dS(n,t){if(!(4194304&n.flags)){var r=e.createMap(),a=e.createMap(),i=e.createMap();n.locals.forEach(function(n){if(262144&n.flags?3&n.flags&&!(3&n.isReferenced):!n.isReferenced&&!n.exportSymbol)for(var o=0,s=n.declarations;o<s.length;o++){var c=s[o];if(!(e.isAmbientModule(c)||(e.isVariableDeclaration(c)&&e.isForInOrOfStatement(c.parent.parent)||mS(c))&&aS(c.name)))if(mS(c))lS(r,fS(c),c,l);else if(e.isBindingElement(c)&&e.isObjectBindingPattern(c.parent)){var u=e.last(c.parent.elements);c!==u&&e.last(c.parent.elements).dotDotDotToken||lS(a,c.parent,c,l)}else if(e.isVariableDeclaration(c))lS(i,c.parent,c,l);else{var d=n.valueDeclaration&&uS(n.valueDeclaration),p=n.valueDeclaration&&e.getNameOfDeclaration(n.valueDeclaration);d&&p?e.isParameterPropertyDeclaration(d)||e.parameterIsThisKeyword(d)||aS(p)||t(d,1,e.createDiagnosticForNode(p,e.Diagnostics._0_is_declared_but_its_value_is_never_read,e.symbolName(n))):rS(c,e.symbolName(n),t)}}}),r.forEach(function(n){var r=n[0],a=n[1],i=r.parent,o=(r.name?1:0)+(r.namedBindings?251===r.namedBindings.kind?1:r.namedBindings.elements.length:0);if(o===a.length)t(i,0,1===a.length?e.createDiagnosticForNode(i,e.Diagnostics._0_is_declared_but_its_value_is_never_read,e.idText(e.first(a).name)):e.createDiagnosticForNode(i,e.Diagnostics.All_imports_in_import_declaration_are_unused));else for(var s=0,c=a;s<c.length;s++){var l=c[s];rS(l,e.idText(l.name),t)}}),a.forEach(function(n){var r=n[0],a=n[1],o=uS(r.parent)?1:0;if(r.elements.length===a.length)1===a.length&&237===r.parent.kind&&238===r.parent.parent.kind?lS(i,r.parent.parent,r.parent,l):t(r,o,1===a.length?e.createDiagnosticForNode(r,e.Diagnostics._0_is_declared_but_its_value_is_never_read,pS(e.first(a).name)):e.createDiagnosticForNode(r,e.Diagnostics.All_destructured_elements_are_unused));else for(var s=0,c=a;s<c.length;s++){var u=c[s];t(u,o,e.createDiagnosticForNode(u,e.Diagnostics._0_is_declared_but_its_value_is_never_read,pS(u.name)))}}),i.forEach(function(n){var r=n[0],a=n[1];if(r.declarations.length===a.length)t(r,0,1===a.length?e.createDiagnosticForNode(e.first(a).name,e.Diagnostics._0_is_declared_but_its_value_is_never_read,pS(e.first(a).name)):e.createDiagnosticForNode(219===r.parent.kind?r.parent:r,e.Diagnostics.All_variables_are_unused));else for(var i=0,o=a;i<o.length;i++){var s=o[i];t(s,0,e.createDiagnosticForNode(s,e.Diagnostics._0_is_declared_but_its_value_is_never_read,pS(s.name)))}})}}function pS(n){switch(n.kind){case 72:return e.idText(n);case 185:case 184:return pS(e.cast(e.first(n.elements),e.isBindingElement).name);default:return e.Debug.assertNever(n)}}function mS(e){return 250===e.kind||253===e.kind||251===e.kind}function fS(e){return 250===e.kind?e:251===e.kind?e.parent:e.parent.parent}function _S(n){if(218===n.kind&&CD(n),e.isFunctionOrModuleBlock(n)){var t=Gn;e.forEach(n.statements,qx),Gn=t}else e.forEach(n.statements,qx);n.locals&&nS(n)}function gS(n){k>=2||L.noEmit||!e.hasRestParameter(n)||4194304&n.flags||e.nodeIsMissing(n.body)||e.forEach(n.parameters,function(n){n.name&&!e.isBindingPattern(n.name)&&n.name.escapedText===H.escapedName&&It(n,e.Diagnostics.Duplicate_identifier_arguments_Compiler_uses_arguments_to_initialize_rest_parameters)})}function vS(n,t,r){if(!t||t.escapedText!==r)return!1;if(154===n.kind||153===n.kind||156===n.kind||155===n.kind||158===n.kind||159===n.kind)return!1;if(4194304&n.flags)return!1;var a=e.getRootDeclaration(n);return 151!==a.kind||!e.nodeIsMissing(a.parent.body)}function yS(n){e.findAncestor(n,function(t){if(4&QC(t)){var r=72!==n.kind;return r?It(e.getNameOfDeclaration(n),e.Diagnostics.Duplicate_identifier_this_Compiler_uses_variable_declaration_this_to_capture_this_reference):It(n,e.Diagnostics.Expression_resolves_to_variable_declaration_this_that_compiler_uses_to_capture_this_reference),!0}return!1})}function hS(n){e.findAncestor(n,function(t){if(8&QC(t)){var r=72!==n.kind;return r?It(e.getNameOfDeclaration(n),e.Diagnostics.Duplicate_identifier_newTarget_Compiler_uses_variable_declaration_newTarget_to_capture_new_target_meta_property_reference):It(n,e.Diagnostics.Expression_resolves_to_variable_declaration_newTarget_that_compiler_uses_to_capture_new_target_meta_property_reference),!0}return!1})}function bS(n,t){if(!(N>=e.ModuleKind.ES2015||L.noEmit)&&(vS(n,t,"require")||vS(n,t,"exports"))&&(!e.isModuleDeclaration(n)||1===e.getModuleInstanceState(n))){var r=ei(n);279===r.kind&&e.isExternalOrCommonJsModule(r)&&It(t,e.Diagnostics.Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module,e.declarationNameToString(t),e.declarationNameToString(t))}}function ES(n,t){if(!(k>=4||L.noEmit)&&vS(n,t,"Promise")&&(!e.isModuleDeclaration(n)||1===e.getModuleInstanceState(n))){var r=ei(n);279===r.kind&&e.isExternalOrCommonJsModule(r)&&1024&r.flags&&It(t,e.Diagnostics.Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_containing_async_functions,e.declarationNameToString(t),e.declarationNameToString(t))}}function TS(n){if(0===(3&e.getCombinedNodeFlags(n))&&!e.isParameterDeclaration(n)&&(237!==n.kind||n.initializer)){var t=ra(n);if(1&t.flags){if(!e.isIdentifier(n.name))return e.Debug.fail();var r=Zt(n,n.name.escapedText,3,void 0,void 0,!1);if(r&&r!==t&&2&r.flags&&3&Ay(r)){var a=e.getAncestor(r.valueDeclaration,238),i=219===a.parent.kind&&a.parent.parent?a.parent.parent:void 0,o=i&&(218===i.kind&&e.isFunctionLike(i.parent)||245===i.kind||244===i.kind||279===i.kind);if(!o){var s=Pa(r);It(n,e.Diagnostics.Cannot_initialize_outer_scoped_variable_0_in_the_same_scope_as_block_scoped_declaration_1,s,s)}}}}}function SS(n){if(151===e.getRootDeclaration(n).kind){var t=e.getContainingFunction(n);r(n.initializer)}function r(a){if(!e.isTypeNode(a)&&!e.isDeclarationName(a)){if(189===a.kind)return r(a.expression);if(72!==a.kind)return e.forEachChild(a,r);var i=Zt(a,a.escapedText,69317567,void 0,void 0,!1);if(i&&i!==ne&&i.valueDeclaration)if(i.valueDeclaration!==n){var o=e.getEnclosingBlockScopeContainer(i.valueDeclaration);if(o===t){if(151===i.valueDeclaration.kind||186===i.valueDeclaration.kind){if(i.valueDeclaration.pos<n.pos)return;if(e.findAncestor(a,function(t){return t===n.initializer?"quit":e.isFunctionLike(t.parent)||154===t.parent.kind&&!e.hasModifier(t.parent,32)&&e.isClassLike(t.parent.parent)}))return}It(a,e.Diagnostics.Initializer_of_parameter_0_cannot_reference_identifier_1_declared_after_it,e.declarationNameToString(n.name),e.declarationNameToString(a))}}else It(a,e.Diagnostics.Parameter_0_cannot_be_referenced_in_its_initializer,e.declarationNameToString(n.name))}}}function xS(e){return e===ae?re:e===nn?en:e}function CS(n){if(WT(n),e.isBindingElement(n)||qx(n.type),n.name){if(149===n.name.kind&&(Hv(n.name),n.initializer&&RE(n.initializer)),186===n.kind){184===n.parent.kind&&k<6&&_A(n,4),n.propertyName&&149===n.propertyName.kind&&Hv(n.propertyName);var t=n.parent.parent,r=ii(t),a=n.propertyName||n.name;if(!e.isBindingPattern(a)&&r){var i=Bu(a);if(Fo(i)){var o=Uo(i),s=sc(r,o);s&&(Yy(s,void 0,!1),Ly(t,!!t.initializer&&98===t.initializer.kind,r,s))}}}if(e.isBindingPattern(n.name)&&(185===n.name.kind&&k<2&&L.downlevelIteration&&_A(n,512),e.forEach(n.name.elements,qx)),n.initializer&&151===e.getRootDeclaration(n).kind&&e.nodeIsMissing(e.getContainingFunction(n).body))It(n,e.Diagnostics.A_parameter_initializer_is_only_allowed_in_a_function_or_constructor_implementation);else if(e.isBindingPattern(n.name)){if(n.initializer&&226!==n.parent.parent.kind){var c=RE(n.initializer);w&&0===n.name.elements.length?My(c,n):Mp(c,Ni(n),n,n.initializer),SS(n)}}else{var l=ra(n),u=xS(Xi(l));if(n===l.valueDeclaration){var d=e.getEffectiveInitializer(n);if(d){var p=e.isInJSFile(n)&&e.isObjectLiteralExpression(d)&&(0===d.properties.length||e.isPrototypeAccess(n.name))&&e.hasEntries(l.exports);p||226===n.parent.parent.kind||(Mp(RE(d),u,n,d,void 0),SS(n))}l.declarations.length>1&&e.some(l.declarations,function(t){return t!==n&&e.isVariableLike(t)&&!DS(t,n)})&&It(n.name,e.Diagnostics.All_declarations_of_0_must_have_identical_modifiers,e.declarationNameToString(n.name))}else{var m=xS(Ni(n));u===oe||m===oe||Sp(u,m)||67108864&l.flags||AS(u,n,m),n.initializer&&Mp(RE(n.initializer),m,n,n.initializer,void 0),DS(n,l.valueDeclaration)||It(n.name,e.Diagnostics.All_declarations_of_0_must_have_identical_modifiers,e.declarationNameToString(n.name))}154!==n.kind&&153!==n.kind&&(wT(n),237!==n.kind&&186!==n.kind||TS(n),bS(n,n.name),ES(n,n.name))}}}function AS(n,t,r){var a=e.getNameOfDeclaration(t),i=154===t.kind||153===t.kind?e.Diagnostics.Subsequent_property_declarations_must_have_the_same_type_Property_0_must_be_of_type_1_but_here_has_type_2:e.Diagnostics.Subsequent_variable_declarations_must_have_the_same_type_Variable_0_must_be_of_type_1_but_here_has_type_2;It(a,i,e.declarationNameToString(a),Fa(n),Fa(r))}function DS(n,t){if(151===n.kind&&237===t.kind||237===n.kind&&151===t.kind)return!0;if(e.hasQuestionToken(n)!==e.hasQuestionToken(t))return!1;var r=504;return e.getSelectedModifierFlags(n,r)===e.getSelectedModifierFlags(t,r)}function LS(e){return cD(e),CS(e)}function kS(e){return rD(e),CS(e)}function NS(n){yA(n)||dD(n.declarationList)||mD(n),e.forEach(n.declarationList.declarations,qx)}function IS(e){CD(e),qE(e.expression)}function wS(n){CD(n),PS(n.expression),qx(n.thenStatement),220===n.thenStatement.kind&&It(n.thenStatement,e.Diagnostics.The_body_of_an_if_statement_cannot_be_the_empty_statement),qx(n.elseStatement)}function MS(e){CD(e),qx(e.statement),PS(e.expression)}function OS(e){CD(e),PS(e.expression),qx(e.statement)}function PS(n,t){var r=qE(n,t);return 16384&r.flags&&It(n,e.Diagnostics.An_expression_of_type_void_cannot_be_tested_for_truthiness),r}function RS(n){CD(n)||n.initializer&&238===n.initializer.kind&&dD(n.initializer),n.initializer&&(238===n.initializer.kind?e.forEach(n.initializer.declarations,LS):qE(n.initializer)),n.condition&&PS(n.condition),n.incrementor&&qE(n.incrementor),qx(n.statement),n.locals&&nS(n)}function FS(n){if(XA(n),n.awaitModifier){var t=e.getFunctionFlags(e.getContainingFunction(n));2===(6&t)&&k<6&&_A(n,16384)}else L.downlevelIteration&&k<2&&_A(n,256);if(238===n.initializer.kind)VS(n);else{var r=n.initializer,a=BS(n.expression,n.awaitModifier);if(187===r.kind||188===r.kind)xE(r,a||oe);else{var i=qE(r);iE(r,e.Diagnostics.The_left_hand_side_of_a_for_of_statement_must_be_a_variable_or_a_property_access),a&&Mp(a,i,r,n.expression)}}qx(n.statement),n.locals&&nS(n)}function GS(n){XA(n);var t=wy(qE(n.expression));if(238===n.initializer.kind){var r=n.initializer.declarations[0];r&&e.isBindingPattern(r.name)&&It(r.name,e.Diagnostics.The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern),VS(n)}else{var a=n.initializer,i=qE(a);187===a.kind||188===a.kind?It(a,e.Diagnostics.The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern):Lp(zu(t),i)?iE(a,e.Diagnostics.The_left_hand_side_of_a_for_in_statement_must_be_a_variable_or_a_property_access):It(a,e.Diagnostics.The_left_hand_side_of_a_for_in_statement_must_be_of_type_string_or_any)}t!==Te&&fE(t,126091264)||It(n.expression,e.Diagnostics.The_right_hand_side_of_a_for_in_statement_must_be_of_type_any_an_object_type_or_a_type_parameter_but_here_has_type_0,Fa(t)),qx(n.statement),n.locals&&nS(n)}function VS(e){var n=e.initializer;if(n.declarations.length>=1){var t=n.declarations[0];LS(t)}}function BS(e,n){var t=Iy(e);return KS(t,e,!0,void 0!==n)}function KS(e,n,t,r){return ai(e)?e:jS(e,n,t,r,!0)||re}function jS(n,t,r,a,i){if(n!==Te){var o=k>=2,s=!o&&L.downlevelIteration;if(o||s||a){var c=US(n,o?t:void 0,a,!0,i);if(c||o)return c}var l=n,u=!1,d=!1;if(r){if(1048576&l.flags){var p=n.types,m=e.filter(p,function(e){return!(132&e.flags)});m!==p&&(l=Au(m,2))}else 132&l.flags&&(l=Te);if(d=l!==n,d&&(k<1&&t&&(It(t,e.Diagnostics.Using_a_string_in_a_for_of_statement_is_only_supported_in_ECMAScript_5_and_higher),u=!0),131072&l.flags))return pe}if(!Fm(l)){if(t&&!u){var f=!!US(n,void 0,a,!0,i),_=!r||d?s?e.Diagnostics.Type_0_is_not_an_array_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator:f?e.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterating_of_iterators:e.Diagnostics.Type_0_is_not_an_array_type:s?e.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator:f?e.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterating_of_iterators:e.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type;It(t,_,Fa(l))}return d?pe:void 0}var g=mc(l,1);return d&&g?132&g.flags?pe:Au([g,pe],2):g}HS(t,n,a)}function US(n,t,r,a,i){if(!ai(n))return X_(n,o);function o(n){var o=n;if(r){if(o.iteratedTypeOfAsyncIterable)return o.iteratedTypeOfAsyncIterable;if(Yi(n,Hl(!1))||Yi(n,ql(!1)))return o.iteratedTypeOfAsyncIterable=n.typeArguments[0]}if(a){if(o.iteratedTypeOfIterable)return r?o.iteratedTypeOfAsyncIterable=RT(o.iteratedTypeOfIterable):o.iteratedTypeOfIterable;if(Yi(n,zl(!1))||Yi(n,Xl(!1)))return r?o.iteratedTypeOfAsyncIterable=RT(n.typeArguments[0]):o.iteratedTypeOfIterable=n.typeArguments[0]}var s=r&&ti(n,e.getPropertyNameForKnownSymbolName("asyncIterator")),c=s||(a?ti(n,e.getPropertyNameForKnownSymbolName("iterator")):void 0);if(!ai(c)){var l=c?lc(c,0):void 0;if(e.some(l)){var u=Au(e.map(l,Fc),2),d=WS(u,t,!!s);return i&&t&&d&&wp(n,s?ru(d):iu(d),t),d?r?o.iteratedTypeOfAsyncIterable=s?d:RT(d):o.iteratedTypeOfIterable=d:void 0}t&&(HS(t,n,r),t=void 0)}}}function HS(n,t,r){It(n,r?e.Diagnostics.Type_0_must_have_a_Symbol_asyncIterator_method_that_returns_an_async_iterator:e.Diagnostics.Type_0_must_have_a_Symbol_iterator_method_that_returns_an_iterator,Fa(t))}function WS(n,t,r){if(!ai(n)){var a=n;if(r?a.iteratedTypeOfAsyncIterator:a.iteratedTypeOfIterator)return r?a.iteratedTypeOfAsyncIterator:a.iteratedTypeOfIterator;var i=r?Wl:Jl;if(Yi(n,i(!1)))return r?a.iteratedTypeOfAsyncIterator=n.typeArguments[0]:a.iteratedTypeOfIterator=n.typeArguments[0];var o=ti(n,"next");if(!ai(o)){var s=o?lc(o,0):e.emptyArray;if(0!==s.length){var c=Au(e.map(s,Fc),2);if(!ai(c)&&(!r||(c=MT(c,t,e.Diagnostics.The_type_returned_by_the_next_method_of_an_async_iterator_must_be_a_promise_for_a_type_with_a_value_property),!ai(c)))){var l=c&&ti(c,"value");if(l)return r?a.iteratedTypeOfAsyncIterator=l:a.iteratedTypeOfIterator=l;t&&It(t,r?e.Diagnostics.The_type_returned_by_the_next_method_of_an_async_iterator_must_be_a_promise_for_a_type_with_a_value_property:e.Diagnostics.The_type_returned_by_the_next_method_of_an_iterator_must_have_a_value_property)}}else t&&It(t,r?e.Diagnostics.An_async_iterator_must_have_a_next_method:e.Diagnostics.An_iterator_must_have_a_next_method)}}}function qS(e,n){if(!ai(e))return US(e,void 0,n,!n,!1)||WS(e,void 0,n)}function zS(e){CD(e)||tD(e)}function JS(n,t){var r=2===(3&e.getFunctionFlags(n))?OT(t):t;return!!r&&mE(r,16387)}function XS(n){if(!CD(n)){var t=e.getContainingFunction(n);if(t){var r=Cc(t),a=Fc(r),i=e.getFunctionFlags(t),o=1&i;if(w||n.expression||131072&a.flags){var s=n.expression?RE(n.expression):ce;if(o)return;if(159===t.kind)n.expression&&It(n,e.Diagnostics.Setters_cannot_return_a_value);else if(157===t.kind)n.expression&&!Mp(s,a,n,n.expression)&&It(n,e.Diagnostics.Return_type_of_constructor_signature_must_be_assignable_to_the_instance_type_of_the_class);else if(Gc(t))if(2&i){var c=OT(a),l=PT(s,n,e.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member);c&&wp(l,c,n)}else Mp(s,a,n,n.expression)}else 157===t.kind||!L.noImplicitReturns||JS(t,a)||o||It(n,e.Diagnostics.Not_all_code_paths_return_a_value)}else gD(n,e.Diagnostics.A_return_statement_can_only_be_used_within_a_function_body)}}function YS(n){CD(n)||16384&n.flags&&gD(n,e.Diagnostics.with_statements_are_not_allowed_in_an_async_function_block),qE(n.expression);var t=e.getSourceFileOfNode(n);if(!_D(t)){var r=e.getSpanOfTokenAtPosition(t,n.pos).start,a=n.statement.pos;vD(t,r,a-r,e.Diagnostics.The_with_statement_is_not_supported_All_symbols_in_a_with_block_will_have_type_any)}}function QS(n){var t;CD(n);var r=!1,i=qE(n.expression),o=Hm(i);e.forEach(n.caseBlock.clauses,function(s){if(272===s.kind&&!r)if(void 0===t)t=s;else{var c=e.getSourceFileOfNode(n),l=e.skipTrivia(c.text,s.pos),u=s.statements.length>0?s.statements[0].pos:s.end;vD(c,l,u-l,e.Diagnostics.A_default_clause_cannot_appear_more_than_once_in_a_switch_statement),r=!0}if(a&&271===s.kind){var d=qE(s.expression),p=Hm(d),m=i;p&&o||(d=p?Wm(d):d,m=Wm(i)),DE(m,d)||Jp(d,m,s.expression,void 0)}e.forEach(s.statements,qx)}),n.caseBlock.locals&&nS(n.caseBlock)}function ZS(n){CD(n)||e.findAncestor(n.parent,function(t){return e.isFunctionLike(t)?"quit":233===t.kind&&t.label.escapedText===n.label.escapedText&&(yD(n.label,e.Diagnostics.Duplicate_label_0,e.getTextOfNode(n.label)),!0)}),qx(n.statement)}function $S(n){CD(n)||void 0===n.expression&&LD(n,e.Diagnostics.Line_break_not_permitted_here),n.expression&&qE(n.expression)}function ex(n){CD(n),_S(n.tryBlock);var t=n.catchClause;if(t){if(t.variableDeclaration)if(t.variableDeclaration.type)gD(t.variableDeclaration.type,e.Diagnostics.Catch_clause_variable_cannot_have_a_type_annotation);else if(t.variableDeclaration.initializer)gD(t.variableDeclaration.initializer,e.Diagnostics.Catch_clause_variable_cannot_have_an_initializer);else{var r=t.block.locals;r&&e.forEachKey(t.locals,function(n){var t=r.get(n);t&&0!==(2&t.flags)&&yD(t.valueDeclaration,e.Diagnostics.Cannot_redeclare_identifier_0_in_catch_clause,n)})}_S(t.block)}n.finallyBlock&&_S(n.finallyBlock)}function nx(n){var t,r=$c(n.symbol,1),a=$c(n.symbol,0),i=mc(n,0),o=mc(n,1);if(i||o){e.forEach(Ms(n),function(e){var t=Xi(e);f(e,t,n,a,i,0),f(e,t,n,r,o,1)});var s=n.symbol.valueDeclaration;if(1&e.getObjectFlags(n)&&e.isClassLike(s))for(var c=0,l=s.members;c<l.length;c++){var u=l[c];if(!e.hasModifier(u,32)&&Ko(u)){var d=ra(u),p=Xi(d);f(d,p,n,a,i,0),f(d,p,n,r,o,1)}}}if(i&&o&&(t=r||a,!t&&2&e.getObjectFlags(n))){var m=e.forEach(uo(n),function(e){return mc(e,0)&&mc(e,1)});t=m?void 0:n.symbol.declarations[0]}function f(n,t,r,a,i,o){if(i&&!e.isKnownSymbol(n)){var s=n.valueDeclaration,c=s&&e.getNameOfDeclaration(s);if(1!==o||(c?Bv(c):Uv(n.escapedName))){var l;if(s&&c&&(204===s.kind||149===c.kind||n.parent===r.symbol))l=s;else if(a)l=a;else if(2&e.getObjectFlags(r)){var u=e.forEach(uo(r),function(e){return Os(e,n.escapedName)&&mc(e,o)});l=u?void 0:r.symbol.declarations[0]}if(l&&!Lp(t,i)){var d=0===o?e.Diagnostics.Property_0_of_type_1_is_not_assignable_to_string_index_type_2:e.Diagnostics.Property_0_of_type_1_is_not_assignable_to_numeric_index_type_2;It(l,d,Pa(n),Fa(t),Fa(i))}}}}t&&!Lp(o,i)&&It(t,e.Diagnostics.Numeric_index_type_0_is_not_assignable_to_string_index_type_1,Fa(o),Fa(i))}function tx(e,n){switch(e.escapedText){case"any":case"unknown":case"number":case"bigint":case"boolean":case"string":case"symbol":case"void":case"object":It(e,n,e.escapedText)}}function rx(n){1===k&&"Object"===n.escapedText&&N!==e.ModuleKind.ES2015&&N!==e.ModuleKind.ESNext&&It(n,e.Diagnostics.Class_name_cannot_be_Object_when_targeting_ES5_with_module_0,e.ModuleKind[N])}function ax(n){if(n)for(var t=!1,r=0;r<n.length;r++){var i=n[r];if(XE(i),a){i.default?(t=!0,ix(i.default,n,r)):t&&It(i,e.Diagnostics.Required_type_parameters_may_not_follow_optional_type_parameters);for(var o=0;o<r;o++)n[o].symbol===i.symbol&&It(i.name,e.Diagnostics.Duplicate_identifier_0,e.declarationNameToString(i.name))}}}function ix(n,t,r){function a(n){if(164===n.kind){var i=Dl(n);if(262144&i.flags)for(var o=r;o<t.length;o++)i.symbol===ra(t[o])&&It(n,e.Diagnostics.Type_parameter_defaults_can_only_reference_previously_declared_type_parameters)}e.forEachChild(n,a)}a(n)}function ox(n){if(1!==n.declarations.length){var t=qt(n);if(!t.typeParametersChecked){t.typeParametersChecked=!0;var r=_x(n);if(r.length<=1)return;var a=Do(n);if(!sx(r,a.localTypeParameters))for(var i=Pa(n),o=0,s=r;o<s.length;o++){var c=s[o];It(c.name,e.Diagnostics.All_declarations_of_0_must_have_identical_type_parameters,i)}}}}function sx(n,t){for(var r=e.length(t),a=Sc(t),i=0,o=n;i<o.length;i++){var s=o[i],c=e.getEffectiveTypeParameterDeclarations(s),l=c.length;if(l<a||l>r)return!1;for(var u=0;u<l;u++){var d=c[u],p=t[u];if(d.name.escapedText!==p.symbol.escapedName)return!1;var m=e.getEffectiveConstraintOfTypeParameter(d),f=m&&Rd(m),_=Bs(p);if(f&&(!_||!Sp(f,_)))return!1;var g=d.default&&Rd(d.default),v=Zs(p);if(g&&v&&!Sp(g,v))return!1}}return!0}function cx(e){return dx(e),Yx(e),Xi(ra(e))}function lx(n){e.forEach(n.members,qx),nS(n)}function ux(n){n.name||e.hasModifier(n,512)||gD(n,e.Diagnostics.A_class_declaration_without_the_default_modifier_must_have_a_name),dx(n),e.forEach(n.members,qx),nS(n)}function dx(n){IA(n),WT(n),n.name&&(tx(n.name,e.Diagnostics.Class_name_cannot_be_0),bS(n,n.name),ES(n,n.name),4194304&n.flags||rx(n.name)),ax(e.getEffectiveTypeParameterDeclarations(n)),wT(n);var t=ra(n),r=Do(t),i=Xo(r),o=Xi(t);ox(t),nT(n),4194304&n.flags||tT(n);var s=e.getEffectiveBaseTypeNode(n);if(s){k<2&&_A(s.parent,1);var c=uo(r);if(c.length&&a){var l=c[0],u=lo(r),d=rc(u);mx(d,s),qx(s.expression);var p=e.getClassExtendsHeritageElement(n);if(p&&p!==s&&qE(p.expression),e.some(s.typeArguments)){e.forEach(s.typeArguments,qx);for(var m=0,f=so(d,s.typeArguments,s);m<f.length;m++){var _=f[m];if(!pT(s,_.typeParameters))break}}var g=Xo(l,r.thisType);if(wp(i,g,void 0)?wp(o,Tp(d),n.name||n,e.Diagnostics.Class_static_side_0_incorrectly_extends_base_class_static_side_1):px(n,i,g,e.Diagnostics.Class_0_incorrectly_extends_base_class_1),8650752&u.flags&&!ao(o)&&It(n.name||n,e.Diagnostics.A_mixin_class_must_have_a_constructor_with_a_single_rest_parameter_of_type_any),!(d.symbol&&32&d.symbol.flags)&&!(8650752&u.flags)){var v=co(d,s.typeArguments,s);e.forEach(v,function(e){return!ib(e.declaration)&&Fc(e)!==l})&&It(s.expression,e.Diagnostics.Base_constructors_must_all_have_the_same_return_type)}gx(r,l)}}var y=e.getClassImplementsHeritageClauseElements(n);if(y)for(var h=0,b=y;h<b.length;h++){var E=b[h];if(e.isEntityNameExpression(E.expression)||It(E.expression,e.Diagnostics.A_class_can_only_implement_an_identifier_Slashqualified_name_with_optional_type_arguments),fT(E),a){var T=Rd(E);if(T!==oe)if(fo(T)){var S=T.symbol&&32&T.symbol.flags?e.Diagnostics.Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass:e.Diagnostics.Class_0_incorrectly_implements_interface_1;g=Xo(T,r.thisType);wp(i,g,void 0)||px(n,i,g,S)}else It(E,e.Diagnostics.A_class_can_only_implement_an_object_type_or_intersection_of_object_types_with_statically_known_members)}}a&&(nx(r),aT(n),yx(n))}function px(n,t,r,a){for(var i=!1,o=function(n){if(e.hasStaticModifier(n))return"continue";var a=n.name&&hC(n.name)||hC(n);if(a){var o=sc(t,a.escapedName),s=sc(r,a.escapedName);if(o&&s){var c=function(){return e.chainDiagnosticMessages(void 0,e.Diagnostics.Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2,Pa(a),Fa(t),Fa(r))};wp(Xi(o),Xi(s),n.name||n,void 0,c)||(i=!0)}}},s=0,c=n.members;s<c.length;s++){var l=c[s];o(l)}i||wp(t,r,n.name||n,a)}function mx(n,t){var r=lc(n,1);if(r.length){var a=r[0].declaration;if(a&&e.hasModifier(a,8)){var i=e.getClassLikeDeclarationOfSymbol(n.symbol);mC(t,i)||It(t,e.Diagnostics.Cannot_extend_a_class_0_Class_constructor_is_marked_as_private,Or(n.symbol))}}}function fx(n){return 1&e.getCheckFlags(n)?n.target:n}function _x(n){return e.filter(n.declarations,function(e){return 240===e.kind||241===e.kind})}function gx(n,t){for(var r=Rs(t),a=0,i=r;a<i.length;a++){var o=i[a],s=fx(o);if(!(4194304&s.flags)){var c=fx(Os(n,s.escapedName)),l=e.getDeclarationModifierFlagsFromSymbol(s);if(e.Debug.assert(!!c,"derived should point to something, even if it is the base class' declaration."),c)if(c===s){var u=e.getClassLikeDeclarationOfSymbol(n.symbol);!(128&l)||u&&e.hasModifier(u,128)||(209===u.kind?It(u,e.Diagnostics.Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1,Pa(o),Fa(t)):It(u,e.Diagnostics.Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2,Fa(n),Pa(o),Fa(t)))}else{var d=e.getDeclarationModifierFlagsFromSymbol(c);if(8&l||8&d)continue;if(Dy(s)||98308&s.flags&&98308&c.flags)continue;var p=void 0;p=Dy(s)?98304&c.flags?e.Diagnostics.Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_accessor:e.Diagnostics.Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_property:98304&s.flags?e.Diagnostics.Class_0_defines_instance_member_accessor_1_but_extended_class_2_defines_it_as_instance_member_function:e.Diagnostics.Class_0_defines_instance_member_property_1_but_extended_class_2_defines_it_as_instance_member_function,It(e.getNameOfDeclaration(c.valueDeclaration)||c.valueDeclaration,p,Fa(t),Pa(s),Fa(n))}}}}function vx(n,t){var r=uo(n);if(r.length<2)return!0;var a=e.createUnderscoreEscapedMap();e.forEach(Ro(n).declaredProperties,function(e){a.set(e.escapedName,{prop:e,containingType:n})});for(var i=!0,o=0,s=r;o<s.length;o++)for(var c=s[o],l=Rs(Xo(c,n.thisType)),u=0,d=l;u<d.length;u++){var p=d[u],m=a.get(p.escapedName);if(m){var f=m.containingType!==n;if(f&&!Cm(m.prop,p)){i=!1;var _=Fa(m.containingType),g=Fa(c),v=e.chainDiagnosticMessages(void 0,e.Diagnostics.Named_property_0_of_types_1_and_2_are_not_identical,Pa(p),_,g);v=e.chainDiagnosticMessages(v,e.Diagnostics.Interface_0_cannot_simultaneously_extend_types_1_and_2,Fa(n),_,g),it.add(e.createDiagnosticForNodeFromMessageChain(t,v))}}else a.set(p.escapedName,{prop:p,containingType:c})}return i}function yx(n){if(w&&P&&!(4194304&n.flags))for(var t=ua(n),r=0,a=n.members;r<a.length;r++){var i=a[r];if(hx(i)){var o=i.name;if(e.isIdentifier(o)){var s=Xi(ra(i));3&s.flags||32768&nf(s)||t&&bx(o,s,t)||It(i.name,e.Diagnostics.Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor,e.declarationNameToString(o))}}}}function hx(n){return 154===n.kind&&!e.hasModifier(n,160)&&!n.exclamationToken&&!n.initializer}function bx(n,t,r){var a=e.createPropertyAccess(e.createThis(),n);a.expression.parent=a,a.parent=r,a.flowNode=r.returnFlowNode;var i=fg(a,t,sf(t));return!(32768&nf(i))}function Ex(n){if(yA(n)||jA(n),ax(n.typeParameters),a){tx(n.name,e.Diagnostics.Interface_name_cannot_be_0),wT(n);var t=ra(n);ox(t);var r=e.getDeclarationOfKind(t,241);if(n===r){var i=Do(t),o=Xo(i);if(vx(i,n.name)){for(var s=0,c=uo(i);s<c.length;s++){var l=c[s];wp(o,Xo(l,i.thisType),n.name,e.Diagnostics.Interface_0_incorrectly_extends_interface_1)}nx(i)}}rT(n)}e.forEach(e.getInterfaceBaseTypeNodes(n),function(n){e.isEntityNameExpression(n.expression)||It(n.expression,e.Diagnostics.An_interface_can_only_extend_an_identifier_Slashqualified_name_with_optional_type_arguments),fT(n)}),e.forEach(n.members,qx),a&&(aT(n),nS(n))}function Tx(n){yA(n),tx(n.name,e.Diagnostics.Type_alias_name_cannot_be_0),ax(n.typeParameters),qx(n.type),nS(n)}function Sx(e){var n=zt(e);if(!(16384&n.flags)){n.flags|=16384;for(var t=0,r=0,a=e.members;r<a.length;r++){var i=a[r],o=xx(i,t);zt(i).enumMemberValue=o,t="number"===typeof o?o+1:void 0}}}function xx(n,t){if(oi(n.name))It(n.name,e.Diagnostics.Computed_property_names_are_not_allowed_in_enums);else{var r=e.getTextOfPropertyName(n.name);Uv(r)&&!jv(r)&&It(n.name,e.Diagnostics.An_enum_member_cannot_have_a_numeric_name)}return n.initializer?Cx(n):4194304&n.parent.flags&&!e.isEnumConst(n.parent)?void 0:void 0!==t?t:void It(n.name,e.Diagnostics.Enum_member_must_have_initializer)}function Cx(n){var t=Eo(ra(n.parent)),r=e.isEnumConst(n.parent),a=n.initializer,i=1!==t||bo(n)?o(a):void 0;if(void 0!==i)r&&"number"===typeof i&&!isFinite(i)&&It(a,isNaN(i)?e.Diagnostics.const_enum_member_initializer_was_evaluated_to_disallowed_value_NaN:e.Diagnostics.const_enum_member_initializer_was_evaluated_to_a_non_finite_value);else{if(1===t)return It(a,e.Diagnostics.Computed_values_are_not_permitted_in_an_enum_with_string_valued_members),0;r?It(a,e.Diagnostics.const_enum_member_initializers_can_only_contain_literal_values_and_other_computed_enum_values):4194304&n.parent.flags?It(a,e.Diagnostics.In_ambient_enum_declarations_member_initializer_must_be_constant_expression):wp(qE(a),Do(ra(n.parent)),a,void 0)}return i;function o(t){switch(t.kind){case 202:var r=o(t.operand);if("number"===typeof r)switch(t.operator){case 38:return r;case 39:return-r;case 53:return~r}break;case 204:var a=o(t.left),i=o(t.right);if("number"===typeof a&&"number"===typeof i)switch(t.operatorToken.kind){case 50:return a|i;case 49:return a&i;case 47:return a>>i;case 48:return a>>>i;case 46:return a<<i;case 51:return a^i;case 40:return a*i;case 42:return a/i;case 38:return a+i;case 39:return a-i;case 43:return a%i;case 41:return Math.pow(a,i)}else if("string"===typeof a&&"string"===typeof i&&38===t.operatorToken.kind)return a+i;break;case 10:return t.text;case 8:return AD(t),+t.text;case 195:return o(t.expression);case 72:var c=t;return jv(c.escapedText)?+c.escapedText:e.nodeIsMissing(t)?0:s(t,ra(n.parent),c.escapedText);case 190:case 189:var l=t;if(Ax(l)){var u=HE(l.expression);if(u.symbol&&384&u.symbol.flags){var d=void 0;if(189===l.kind)d=l.name.escapedText;else{var p=l.argumentExpression;e.Debug.assert(e.isLiteralExpression(p)),d=e.escapeLeadingUnderscores(p.text)}return s(t,u.symbol,d)}}break}}function s(t,r,a){var i=r.exports.get(a);if(i){var o=i.valueDeclaration;if(o!==n)return Qt(o,n)?ZC(o):(It(t,e.Diagnostics.A_member_initializer_in_a_enum_declaration_cannot_reference_members_declared_after_it_including_members_defined_in_other_enums),0)}}}function Ax(e){return 72===e.kind||189===e.kind&&Ax(e.expression)||190===e.kind&&Ax(e.expression)&&10===e.argumentExpression.kind}function Dx(n){if(a){yA(n),tx(n.name,e.Diagnostics.Enum_name_cannot_be_0),bS(n,n.name),ES(n,n.name),wT(n),Sx(n);var t=e.isEnumConst(n);L.isolatedModules&&t&&4194304&n.flags&&It(n.name,e.Diagnostics.Ambient_const_enums_are_not_allowed_when_the_isolatedModules_flag_is_provided);var r=ra(n),i=e.getDeclarationOfKind(r,n.kind);if(n===i){r.declarations.length>1&&e.forEach(r.declarations,function(n){e.isEnumDeclaration(n)&&e.isEnumConst(n)!==t&&It(e.getNameOfDeclaration(n),e.Diagnostics.Enum_declarations_must_all_be_const_or_non_const)});var o=!1;e.forEach(r.declarations,function(n){if(243!==n.kind)return!1;var t=n;if(!t.members.length)return!1;var r=t.members[0];r.initializer||(o?It(r.name,e.Diagnostics.In_an_enum_with_multiple_declarations_only_one_declaration_can_omit_an_initializer_for_its_first_enum_element):o=!0)})}}}function Lx(n){for(var t=n.declarations,r=0,a=t;r<a.length;r++){var i=a[r];if((240===i.kind||239===i.kind&&e.nodeIsPresent(i.body))&&!(4194304&i.flags))return i}}function kx(n,t){var r=e.getEnclosingBlockScopeContainer(n),a=e.getEnclosingBlockScopeContainer(t);return Jt(r)?Jt(a):!Jt(a)&&r===a}function Nx(n){if(a){var t=e.isGlobalScopeAugmentation(n),r=4194304&n.flags;t&&!r&&It(n.name,e.Diagnostics.Augmentations_for_the_global_scope_should_have_declare_modifier_unless_they_appear_in_already_ambient_context);var i=e.isAmbientModule(n),o=i?e.Diagnostics.An_ambient_module_declaration_is_only_allowed_at_the_top_level_in_a_file:e.Diagnostics.A_namespace_declaration_is_only_allowed_in_a_namespace_or_module;if(Vx(n,o))return;yA(n)||r||10!==n.name.kind||yD(n.name,e.Diagnostics.Only_ambient_modules_can_use_quoted_names),e.isIdentifier(n.name)&&(bS(n,n.name),ES(n,n.name)),wT(n);var s=ra(n);if(512&s.flags&&s.declarations.length>1&&!r&&d(n,!!L.preserveConstEnums||!!L.isolatedModules)){var c=Lx(s);c&&(e.getSourceFileOfNode(n)!==e.getSourceFileOfNode(c)?It(n.name,e.Diagnostics.A_namespace_declaration_cannot_be_in_a_different_file_from_a_class_or_function_with_which_it_is_merged):n.pos<c.pos&&It(n.name,e.Diagnostics.A_namespace_declaration_cannot_be_located_prior_to_a_class_or_function_with_which_it_is_merged));var l=e.getDeclarationOfKind(s,240);l&&kx(n,l)&&(zt(n).flags|=32768)}if(i)if(e.isExternalModuleAugmentation(n)){var u=t||33554432&ra(n).flags;if(u&&n.body)for(var p=0,m=n.body.statements;p<m.length;p++){var f=m[p];Ix(f,t)}}else Jt(n.parent)?t?It(n.name,e.Diagnostics.Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_declarations):e.isExternalModuleNameRelative(e.getTextOfIdentifierOrLiteral(n.name))&&It(n.name,e.Diagnostics.Ambient_module_declaration_cannot_specify_relative_module_name):It(n.name,t?e.Diagnostics.Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_declarations:e.Diagnostics.Ambient_modules_cannot_be_nested_in_other_modules_or_namespaces)}n.body&&(qx(n.body),e.isGlobalScopeAugmentation(n)||nS(n))}function Ix(n,t){switch(n.kind){case 219:for(var r=0,a=n.declarationList.declarations;r<a.length;r++){var i=a[r];Ix(i,t)}break;case 254:case 255:gD(n,e.Diagnostics.Exports_and_export_assignments_are_not_permitted_in_module_augmentations);break;case 248:case 249:gD(n,e.Diagnostics.Imports_are_not_permitted_in_module_augmentations_Consider_moving_them_to_the_enclosing_external_module);break;case 186:case 237:var o=n.name;if(e.isBindingPattern(o)){for(var s=0,c=o.elements;s<c.length;s++){var l=c[s];Ix(l,t)}break}case 240:case 243:case 239:case 241:case 244:case 242:if(t)return;var u=ra(n);if(u){var d=!(33554432&u.flags);d||(d=!!u.parent&&e.isExternalModuleAugmentation(u.parent.declarations[0]))}break}}function wx(e){switch(e.kind){case 72:return e;case 148:do{e=e.left}while(72!==e.kind);return e;case 189:do{e=e.expression}while(72!==e.kind);return e}}function Mx(n){var t=e.getExternalModuleName(n);if(!t||e.nodeIsMissing(t))return!1;if(!e.isStringLiteral(t))return It(t,e.Diagnostics.String_literal_expected),!1;var r=245===n.parent.kind&&e.isAmbientModule(n.parent.parent);return 279===n.parent.kind||r?!(r&&e.isExternalModuleNameRelative(t.text)&&!Ha(n))||(It(n,e.Diagnostics.Import_or_export_declaration_in_an_ambient_module_declaration_cannot_reference_module_through_relative_module_name),!1):(It(t,255===n.kind?e.Diagnostics.Export_declarations_are_not_permitted_in_a_namespace:e.Diagnostics.Import_declarations_in_a_namespace_cannot_reference_a_module),!1)}function Ox(n){var t=ra(n),r=Nr(t);if(r!==ne){var a=(68268991&t.flags?67220415:0)|(67897832&t.flags?67897832:0)|(1920&t.flags?1920:0);if(r.flags&a){var i=257===n.kind?e.Diagnostics.Export_declaration_conflicts_with_exported_declaration_of_0:e.Diagnostics.Import_declaration_conflicts_with_local_declaration_of_0;It(n,i,Pa(t))}!L.isolatedModules||257!==n.kind||67220415&r.flags||4194304&n.flags||It(n,e.Diagnostics.Cannot_re_export_a_type_when_the_isolatedModules_flag_is_provided)}}function Px(e){bS(e,e.name),ES(e,e.name),Ox(e)}function Rx(n){if(!Vx(n,e.Diagnostics.An_import_declaration_can_only_be_used_in_a_namespace_or_module)&&(!yA(n)&&e.hasModifiers(n)&&gD(n,e.Diagnostics.An_import_declaration_cannot_have_modifiers),Mx(n))){var t=n.importClause;if(t&&(t.name&&Px(t),t.namedBindings))if(251===t.namedBindings.kind)Px(t.namedBindings);else{var r=Vr(n,n.moduleSpecifier);r&&e.forEach(t.namedBindings.elements,Px)}}}function Fx(n){if(!Vx(n,e.Diagnostics.An_import_declaration_can_only_be_used_in_a_namespace_or_module)&&(yA(n),e.isInternalModuleImportEqualsDeclaration(n)||Mx(n)))if(Px(n),e.hasModifier(n,1)&&Ir(n),259!==n.moduleReference.kind){var t=Nr(ra(n));if(t!==ne){if(67220415&t.flags){var r=wx(n.moduleReference);1920&Pr(r,67221439).flags||It(r,e.Diagnostics.Module_0_is_hidden_by_a_local_declaration_with_the_same_name,e.declarationNameToString(r))}67897832&t.flags&&tx(n.name,e.Diagnostics.Import_name_cannot_be_0)}}else N>=e.ModuleKind.ES2015&&!(4194304&n.flags)&&yD(n,e.Diagnostics.Import_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_import_Asterisk_as_ns_from_mod_import_a_from_mod_import_d_from_mod_or_another_module_format_instead)}function Gx(n){if(!Vx(n,e.Diagnostics.An_export_declaration_can_only_be_used_in_a_module)&&(!yA(n)&&e.hasModifiers(n)&&gD(n,e.Diagnostics.An_export_declaration_cannot_have_modifiers),!n.moduleSpecifier||Mx(n)))if(n.exportClause){e.forEach(n.exportClause.elements,Bx);var t=245===n.parent.kind&&e.isAmbientModule(n.parent.parent),r=!t&&245===n.parent.kind&&!n.moduleSpecifier&&4194304&n.flags;279===n.parent.kind||t||r||It(n,e.Diagnostics.Export_declarations_are_not_permitted_in_a_namespace)}else{var a=Vr(n,n.moduleSpecifier);a&&zr(a)&&It(n.moduleSpecifier,e.Diagnostics.Module_0_uses_export_and_cannot_be_used_with_export_Asterisk,Pa(a)),N!==e.ModuleKind.System&&N!==e.ModuleKind.ES2015&&N!==e.ModuleKind.ESNext&&_A(n,32768)}}function Vx(e,n){var t=279===e.parent.kind||245===e.parent.kind||244===e.parent.kind;return t||gD(e,n),!t}function Bx(n){if(Ox(n),e.getEmitDeclarations(L)&&Xa(n.propertyName||n.name,!0),!n.parent.parent.moduleSpecifier){var t=n.propertyName||n.name,r=Zt(t,t.escapedText,70107135,void 0,void 0,!0);r&&(r===j||Jt(ei(r.declarations[0])))?It(t,e.Diagnostics.Cannot_export_0_Only_local_declarations_can_be_exported_from_a_module,e.idText(t)):Ir(n)}}function Kx(n){if(!Vx(n,e.Diagnostics.An_export_assignment_can_only_be_used_in_a_module)){var t=279===n.parent.kind?n.parent:n.parent.parent;244!==t.kind||e.isAmbientModule(t)?(!yA(n)&&e.hasModifiers(n)&&gD(n,e.Diagnostics.An_export_assignment_cannot_have_modifiers),72===n.expression.kind?(Ir(n),e.getEmitDeclarations(L)&&Xa(n.expression,!0)):RE(n.expression),Ux(t),4194304&n.flags&&!e.isEntityNameExpression(n.expression)&&yD(n.expression,e.Diagnostics.The_expression_of_an_export_assignment_must_be_an_identifier_or_qualified_name_in_an_ambient_context),!n.isExportEquals||4194304&n.flags||(N>=e.ModuleKind.ES2015?yD(n,e.Diagnostics.Export_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_export_default_or_another_module_format_instead):N===e.ModuleKind.System&&yD(n,e.Diagnostics.Export_assignment_is_not_supported_when_module_flag_is_system))):n.isExportEquals?It(n,e.Diagnostics.An_export_assignment_cannot_be_used_in_a_namespace):It(n,e.Diagnostics.A_default_export_can_only_be_used_in_an_ECMAScript_style_module)}}function jx(n){return e.forEachEntry(n.exports,function(e,n){return"export="!==n})}function Ux(n){var t=ra(n),r=qt(t);if(!r.exportsChecked){var a=t.exports.get("export=");if(a&&jx(t)){var i=pr(a)||a.valueDeclaration;Ha(i)||e.isInJSFile(i)||It(i,e.Diagnostics.An_export_assignment_cannot_be_used_in_a_module_with_other_exported_elements)}var o=$r(t);o&&o.forEach(function(n,t){var r=n.declarations,a=n.flags;if("__export"!==t&&!(1984&a)){var i=e.countWhere(r,At);if(!(524288&a&&i<=2)&&i>1)for(var o=0,s=r;o<s.length;o++){var c=s[o];Wx(c)&&it.add(e.createDiagnosticForNode(c,e.Diagnostics.Cannot_redeclare_exported_variable_0,e.unescapeLeadingUnderscores(t)))}}}),r.exportsChecked=!0}}function Hx(n){return!e.isAccessor(n)}function Wx(e){return 239!==e.kind&&156!==e.kind||!!e.body}function qx(n){if(n){e.isInJSFile(n)&&e.forEach(n.jsDoc,function(n){var t=n.tags;return e.forEach(t,qx)});var t=n.kind;if(p)switch(t){case 244:case 240:case 241:case 239:p.throwIfCancellationRequested()}switch(t){case 150:return XE(n);case 151:return YE(n);case 154:case 153:return iT(n);case 165:case 166:case 160:case 161:case 162:return eT(n);case 156:case 155:return oT(n);case 157:return sT(n);case 158:case 159:return cT(n);case 164:return fT(n);case 163:return QE(n);case 167:return gT(n);case 168:return vT(n);case 169:return yT(n);case 170:return hT(n);case 173:case 174:return bT(n);case 177:case 171:case 172:return qx(n.type);case 178:return xT(n);case 179:return CT(n);case 175:return AT(n);case 176:return DT(n);case 183:return LT(n);case 295:return ZT(n);case 304:case 297:return zT(n);case 303:return JT(n);case 302:return XT(n);case 299:return YT(n);case 289:QT(n);case 287:case 286:case 284:case 285:case 292:return zx(n),void e.forEachChild(n,qx);case 290:return void Jx(n);case 283:return qx(n.type);case 180:return TT(n);case 181:return ST(n);case 239:return qT(n);case 218:case 245:return _S(n);case 219:return NS(n);case 221:return IS(n);case 222:return wS(n);case 223:return MS(n);case 224:return OS(n);case 225:return RS(n);case 226:return GS(n);case 227:return FS(n);case 228:case 229:return zS(n);case 230:return XS(n);case 231:return YS(n);case 232:return QS(n);case 233:return ZS(n);case 234:return $S(n);case 235:return ex(n);case 237:return LS(n);case 186:return kS(n);case 240:return ux(n);case 241:return Ex(n);case 242:return Tx(n);case 243:return Dx(n);case 244:return Nx(n);case 249:return Rx(n);case 248:return Fx(n);case 255:return Gx(n);case 254:return Kx(n);case 220:case 236:return void CD(n);case 258:return uT(n)}}}function zx(n){e.isInJSFile(n)||yD(n,e.Diagnostics.JSDoc_types_can_only_be_used_inside_documentation_comments)}function Jx(n){zx(n),qx(n.type);var t=n.parent;if(e.isParameter(t)&&e.isJSDocFunctionType(t.parent))e.last(t.parent.parameters)!==t&&It(n,e.Diagnostics.A_rest_parameter_must_be_last_in_a_parameter_list);else{e.isJSDocTypeExpression(t)||It(n,e.Diagnostics.JSDoc_may_only_appear_in_the_last_parameter_of_a_signature);var r=n.parent.parent;if(e.isJSDocParameterTag(r)){var a=e.getParameterSymbolFromJSDoc(r);if(a){var i=e.getHostSignatureFromJSDoc(r);i&&e.last(i.parameters).symbol===a||It(n,e.Diagnostics.A_rest_parameter_must_be_last_in_a_parameter_list)}}else It(n,e.Diagnostics.JSDoc_may_only_appear_in_the_last_parameter_of_a_signature)}}function Xx(n){var t=Rd(n.type),r=n.parent,a=n.parent.parent;if(e.isJSDocTypeExpression(n.parent)&&e.isJSDocParameterTag(a)){var i=e.getHostSignatureFromJSDoc(a);if(i){var o=e.lastOrUndefined(i.parameters),s=e.getParameterSymbolFromJSDoc(a);if(!o||s&&o.symbol===s&&e.isRestParameter(o))return su(t)}}return e.isParameter(r)&&e.isJSDocFunctionType(r.parent)?su(t):vi(t)}function Yx(n){var t=e.getSourceFileOfNode(n),r=zt(t);if(!(1&r.flags)){r.deferredNodes=r.deferredNodes||e.createMap();var a=""+l(n);r.deferredNodes.set(a,n)}}function Qx(e){var n=zt(e);n.deferredNodes&&n.deferredNodes.forEach(function(e){switch(e.kind){case 196:case 197:case 156:case 155:$b(e);break;case 158:case 159:cT(e);break;case 209:lx(e);break;case 261:Xv(e);break;case 260:Qv(e);break}})}function Zx(n){e.performance.mark("beforeCheck"),nC(n),e.performance.mark("afterCheck"),e.performance.measure("Check","beforeCheck","afterCheck")}function $x(n){switch(n){case 0:return!!L.noUnusedLocals;case 1:return!!L.noUnusedParameters;default:return e.Debug.assertNever(n)}}function eC(n){return On.get(n.path)||e.emptyArray}function nC(n){var t=zt(n);if(!(1&t.flags)){if(e.skipTypeChecking(n,L))return;xD(n),e.clear(tt),e.clear(rt),e.forEach(n.statements,qx),qx(n.endOfFileToken),Qx(n),e.isExternalOrCommonJsModule(n)&&nS(n),n.isDeclarationFile||!L.noUnusedLocals&&!L.noUnusedParameters||tS(eC(n),function(n,t,r){!e.containsParseError(n)&&$x(t)&&it.add(r)}),e.isExternalOrCommonJsModule(n)&&Ux(n),tt.length&&(e.forEach(tt,yS),e.clear(tt)),rt.length&&(e.forEach(rt,hS),e.clear(rt)),t.flags|=1}}function tC(e,n){try{return p=n,rC(e)}finally{p=void 0}}function rC(n){if(iC(),n){var t=it.getGlobalDiagnostics(),a=t.length;Zx(n);var i=it.getDiagnostics(n.fileName),o=it.getGlobalDiagnostics();if(o!==t){var s=e.relativeComplement(t,o,e.compareDiagnostics);return e.concatenate(s,i)}return 0===a&&o.length>0?e.concatenate(o,i):i}return e.forEach(r.getSourceFiles(),Zx),it.getDiagnostics()}function aC(){return iC(),it.getGlobalDiagnostics()}function iC(){if(!a)throw new Error("Trying to get diagnostics from a type checker that does not produce them.")}function oC(n,t){if(8388608&n.flags)return[];var r=e.createSymbolTable(),a=!1;return i(),r.delete("this"),gc(r);function i(){while(n){switch(n.locals&&!Jt(n)&&s(n.locals,t),n.kind){case 279:if(!e.isExternalOrCommonJsModule(n))break;case 244:s(ra(n).exports,2623475&t);break;case 243:s(ra(n).exports,8&t);break;case 209:var r=n.name;r&&o(n.symbol,t);case 240:case 241:a||s(zo(ra(n)),67897832&t);break;case 196:var i=n.name;i&&o(n.symbol,t);break}e.introducesArgumentsExoticObject(n)&&o(H,t),a=e.hasModifier(n,32),n=n.parent}s(wn,t)}function o(n,t){if(e.getCombinedLocalAndExportSymbolFlags(n)&t){var a=n.escapedName;r.has(a)||r.set(a,n)}}function s(e,n){n&&e.forEach(function(e){o(e,n)})}}function sC(e){return 72===e.kind&&cC(e.parent)&&e.parent.name===e}function cC(e){switch(e.kind){case 150:case 240:case 241:case 242:case 243:return!0;default:return!1}}function lC(e){while(148===e.parent.kind)e=e.parent;return 164===e.parent.kind}function uC(e){while(189===e.parent.kind)e=e.parent;return 211===e.parent.kind}function dC(n,t){var r;while(1){if(n=e.getContainingClass(n),!n)break;if(r=t(n))break}return r}function pC(n){return!!e.findAncestor(n,function(n){return!!(e.isConstructorDeclaration(n)&&e.nodeIsPresent(n.body)||e.isPropertyDeclaration(n))||!(!e.isClassLike(n)&&!e.isFunctionLikeDeclaration(n))&&"quit"})}function mC(e,n){return!!dC(e,function(e){return e===n})}function fC(e){while(148===e.parent.kind)e=e.parent;return 248===e.parent.kind?e.parent.moduleReference===e?e.parent:void 0:254===e.parent.kind&&e.parent.expression===e?e.parent:void 0}function _C(e){return void 0!==fC(e)}function gC(n){var t=e.getAssignmentDeclarationKind(n.parent.parent);switch(t){case 1:case 3:return ra(n.parent);case 4:case 2:case 5:return ra(n.parent.parent)}}function vC(n){var t=n.parent;while(e.isQualifiedName(t))n=t,t=t.parent;if(t&&183===t.kind&&t.qualifier===n)return t}function yC(n){if(e.isDeclarationName(n))return ra(n.parent);if(e.isInJSFile(n)&&189===n.parent.kind&&n.parent===n.parent.parent.left){var t=gC(n);if(t)return t}if(254===n.parent.kind&&e.isEntityNameExpression(n)){var r=Pr(n,70107135,!0);if(r&&r!==ne)return r}else if(!e.isPropertyAccessExpression(n)&&_C(n)){var a=e.getAncestor(n,248);return e.Debug.assert(void 0!==a),Mr(n,!0)}if(!e.isPropertyAccessExpression(n)){var i=vC(n);if(i){Rd(i);var o=zt(n).resolvedSymbol;return o===ne?void 0:o}}while(e.isRightSideOfQualifiedNameOrPropertyAccess(n))n=n.parent;if(uC(n)){var s=0;211===n.parent.kind?(s=67897832,e.isExpressionWithTypeArgumentsInClassExtendsClause(n.parent)&&(s|=67220415)):s=1920,s|=2097152;var c=e.isEntityNameExpression(n)?Pr(n,s):void 0;if(c)return c}if(299===n.parent.kind)return e.getParameterSymbolFromJSDoc(n.parent);if(150===n.parent.kind&&303===n.parent.parent.kind){e.Debug.assert(!e.isInJSFile(n));var l=e.getTypeParameterFromJsDoc(n.parent);return l&&l.symbol}if(e.isExpressionNode(n)){if(e.nodeIsMissing(n))return;if(72===n.kind){if(e.isJSXTagName(n)&&ny(n)){var u=sy(n.parent);return u===ne?void 0:u}return Pr(n,67220415,!1,!0)}if(189===n.kind||148===n.kind){var d=zt(n);return d.resolvedSymbol?d.resolvedSymbol:(189===n.kind?Oy(n):Py(n),d.resolvedSymbol)}}else if(lC(n)){s=164===n.parent.kind?67897832:1920;return Pr(n,s,!1,!0)}return 163===n.parent.kind?Pr(n,1):void 0}function hC(n){if(279===n.kind)return e.isExternalModule(n)?ta(n.symbol):void 0;var t=n.parent,r=t.parent;if(!(8388608&n.flags)){if(m(n)){var a=ra(t);return e.isImportOrExportSpecifier(n.parent)&&n.parent.propertyName===n?qv(a):a}if(e.isLiteralComputedPropertyDeclarationName(n))return ra(t.parent);if(72===n.kind){if(_C(n))return yC(n);if(186===t.kind&&184===r.kind&&n===t.propertyName){var i=TC(r),o=sc(i,n.escapedText);if(o)return o}}switch(n.kind){case 72:case 189:case 148:return yC(n);case 100:var s=e.getThisContainer(n,!1);if(e.isFunctionLike(s)){var c=Cc(s);if(c.thisParameter)return c.thisParameter}if(e.isInExpressionContext(n))return qE(n).symbol;case 178:return Pd(n).symbol;case 98:return qE(n).symbol;case 124:var l=n.parent;return l&&157===l.kind?l.parent.symbol:void 0;case 10:case 14:if(e.isExternalModuleImportEqualsDeclaration(n.parent.parent)&&e.getExternalModuleImportEqualsDeclarationExpression(n.parent.parent)===n||(249===n.parent.kind||255===n.parent.kind)&&n.parent.moduleSpecifier===n||e.isInJSFile(n)&&e.isRequireCall(n.parent,!1)||e.isImportCall(n.parent)||e.isLiteralTypeNode(n.parent)&&e.isLiteralImportTypeNode(n.parent.parent)&&n.parent.parent.argument===n.parent)return Vr(n,n);if(e.isCallExpression(t)&&e.isBindableObjectDefinePropertyCall(t)&&t.arguments[1]===n)return ra(t);case 8:var u=e.isElementAccessExpression(t)?t.argumentExpression===n?HE(t.expression):void 0:e.isLiteralTypeNode(t)&&e.isIndexedAccessTypeNode(r)?Rd(r.objectType):void 0;return u&&sc(u,e.escapeLeadingUnderscores(n.text));case 80:case 90:case 37:case 76:return ra(n.parent);case 183:return e.isLiteralImportTypeNode(n)?hC(n.argument.literal):void 0;case 85:return e.isExportAssignment(n.parent)?e.Debug.assertDefined(n.parent.symbol):void 0;default:return}}}function bC(e){if(e&&276===e.kind)return Pr(e.name,69317567)}function EC(e){return e.parent.parent.moduleSpecifier?Tr(e.parent.parent,e):Pr(e.propertyName||e.name,70107135)}function TC(n){if(8388608&n.flags)return oe;var t=e.tryGetClassImplementingOrExtendingExpressionWithTypeArguments(n),r=t&&vo(ra(t.class));if(e.isPartOfTypeNode(n)){var a=Rd(n);return r?Xo(a,r.thisType):a}if(e.isExpressionNode(n))return CC(n);if(r&&!t.isImplements){var i=e.firstOrUndefined(uo(r));return i?Xo(i,r.thisType):oe}if(cC(n)){var o=ra(n);return Do(o)}if(sC(n)){o=hC(n);return o?Do(o):oe}if(e.isDeclaration(n)){o=ra(n);return Xi(o)}if(m(n)){o=hC(n);return o?Xi(o):oe}if(e.isBindingPattern(n))return yi(n.parent,!0)||oe;if(_C(n)){o=hC(n);if(o){var s=Do(o);return s!==oe?s:Xi(o)}}return oe}function SC(n){if(e.Debug.assert(188===n.kind||187===n.kind),227===n.parent.kind){var t=BS(n.parent.expression,n.parent.awaitModifier);return xE(n,t||oe)}if(204===n.parent.kind){t=HE(n.parent.right);return xE(n,t||oe)}if(275===n.parent.kind){var r=SC(n.parent.parent);return EE(r||oe,n.parent)}e.Debug.assert(187===n.parent.kind);var a=SC(n.parent),i=KS(a||oe,n.parent,!1,!1)||oe;return SE(n.parent,a,n.parent.elements.indexOf(n),i||oe)}function xC(e){var n=SC(e.parent.parent);return n&&sc(n,e.escapedText)}function CC(n){return e.isRightSideOfQualifiedNameOrPropertyAccess(n)&&(n=n.parent),Ld(HE(n))}function AC(n){var t=ra(n.parent);return e.hasModifier(n,32)?Xi(t):Do(t)}function DC(n){var t=n.name;switch(t.kind){case 72:return Nd(e.idText(t));case 8:case 10:return Nd(t.text);case 149:var r=Hv(t);return fE(r,12288)?r:pe;default:return e.Debug.fail("Unsupported property name."),oe}}function LC(n){n=rc(n);var t=e.createSymbolTable(Rs(n)),r=lc(n,0).length?We:lc(n,1).length?qe:void 0;return r&&e.forEach(Rs(r),function(e){t.has(e.escapedName)||t.set(e.escapedName,e)}),ya(t)}function kC(n){return e.typeHasCallOrConstructSignatures(n,q)}function NC(n){var t=IC(n);return t?e.flatMap(t,NC):[n]}function IC(n){if(6&e.getCheckFlags(n))return e.mapDefined(qt(n).containingType.types,function(e){return sc(e,n.escapedName)});if(33554432&n.flags){var t=n,r=t.leftSpread,a=t.rightSpread,i=t.syntheticOrigin;return r?[r,a]:i?[i]:e.singleElementArray(wC(n))}}function wC(e){var n,t=e;while(t=qt(t).target)n=t;return n}function MC(n){if(!e.isGeneratedIdentifier(n)){var t=e.getParseTreeNode(n,e.isIdentifier);if(t){var r=189===t.parent.kind&&t.parent.name===t;return!r&&sA(t)===H}}return!1}function OC(n){var t=Vr(n.parent,n);if(!t||e.isShorthandAmbientModuleSymbol(t))return!0;var r=zr(t);t=Hr(t);var a=qt(t);return void 0===a.exportsSomeValue&&(a.exportsSomeValue=r?!!(67220415&t.flags):e.forEachEntry($r(t),i)),a.exportsSomeValue;function i(e){return e=kr(e),e&&!!(67220415&e.flags)}}function PC(n){return e.isModuleOrEnumDeclaration(n.parent)&&n===n.parent.name}function RC(n,t){var r=e.getParseTreeNode(n,e.isIdentifier);if(r){var a=sA(r,PC(r));if(a){if(1048576&a.flags){var i=ta(a.exportSymbol);if(!t&&944&i.flags&&!(3&i.flags))return;a=i}var o=aa(a);if(o){if(512&o.flags&&279===o.valueDeclaration.kind){var s=o.valueDeclaration,c=e.getSourceFileOfNode(r),l=s!==c;return l?void 0:s}return e.findAncestor(r.parent,function(n){return e.isModuleOrEnumDeclaration(n)&&ra(n)===o})}}}}function FC(n){var t=e.getParseTreeNode(n,e.isIdentifier);if(t){var r=sA(t);if(Lr(r,67220415))return pr(r)}}function GC(n){if(418&n.flags){var t=qt(n);if(void 0===t.isDeclarationWithCollidingName){var r=e.getEnclosingBlockScopeContainer(n.valueDeclaration);if(e.isStatementWithLocals(r)){var a=zt(n.valueDeclaration);if(Zt(r.parent,n.escapedName,67220415,void 0,void 0,!1))t.isDeclarationWithCollidingName=!0;else if(262144&a.flags){var i=524288&a.flags,o=e.isIterationStatement(r,!1),s=218===r.kind&&e.isIterationStatement(r.parent,!1);t.isDeclarationWithCollidingName=!e.isBlockScopedContainerTopLevel(r)&&(!i||!o&&!s)}else t.isDeclarationWithCollidingName=!1}}return t.isDeclarationWithCollidingName}return!1}function VC(n){if(!e.isGeneratedIdentifier(n)){var t=e.getParseTreeNode(n,e.isIdentifier);if(t){var r=sA(t);if(r&&GC(r))return r.valueDeclaration}}}function BC(n){var t=e.getParseTreeNode(n,e.isDeclaration);if(t){var r=ra(t);if(r)return GC(r)}return!1}function KC(n){switch(n.kind){case 248:case 250:case 251:case 253:case 257:return UC(ra(n)||ne);case 255:var t=n.exportClause;return!!t&&e.some(t.elements,KC);case 254:return!n.expression||72!==n.expression.kind||UC(ra(n)||ne)}return!1}function jC(n){var t=e.getParseTreeNode(n,e.isImportEqualsDeclaration);if(void 0===t||279!==t.parent.kind||!e.isInternalModuleImportEqualsDeclaration(t))return!1;var r=UC(ra(t));return r&&t.moduleReference&&!e.nodeIsMissing(t.moduleReference)}function UC(e){var n=Nr(e);return n===ne||!!(67220415&n.flags)&&(L.preserveConstEnums||!HC(n))}function HC(e){return vE(e)||!!e.constEnumOnlyModule}function WC(n,t){if(e.isAliasSymbolDeclaration(n)){var r=ra(n);if(r&&qt(r).referenced)return!0;var a=qt(r).target;if(a&&1&e.getModifierFlags(n)&&67220415&a.flags&&(L.preserveConstEnums||!HC(a)))return!0}return!!t&&!!e.forEachChild(n,function(e){return WC(e,t)})}function qC(n){if(e.nodeIsPresent(n.body)){if(e.isGetAccessor(n)||e.isSetAccessor(n))return!1;var t=ra(n),r=Nc(t);return r.length>1||1===r.length&&r[0].declaration!==n}return!1}function zC(n){return!!w&&!hc(n)&&!e.isJSDocParameterTag(n)&&!!n.initializer&&!e.hasModifier(n,92)}function JC(n){return w&&hc(n)&&!n.initializer&&e.hasModifier(n,92)}function XC(n){var t=e.getParseTreeNode(n,e.isFunctionDeclaration);if(!t)return!1;var r=ra(t);return!!(r&&16&r.flags)&&!!e.forEachEntry(Zr(r),function(n){return 67220415&n.flags&&e.isPropertyAccessExpression(n.valueDeclaration)})}function YC(n){var t=e.getParseTreeNode(n,e.isFunctionDeclaration);if(!t)return e.emptyArray;var r=ra(t);return r&&Rs(Xi(r))||e.emptyArray}function QC(e){return zt(e).flags||0}function ZC(e){return Sx(e.parent),zt(e).enumMemberValue}function $C(e){switch(e.kind){case 278:case 189:case 190:return!0}return!1}function eA(n){if(278===n.kind)return ZC(n);var t=zt(n).resolvedSymbol;if(t&&8&t.flags){var r=t.valueDeclaration;if(e.isEnumConst(r.parent))return ZC(r)}}function nA(e){return!!(524288&e.flags)&&lc(e,0).length>0}function tA(n,t){var r=e.getParseTreeNode(n,e.isEntityName);if(!r)return e.TypeReferenceSerializationKind.Unknown;if(t&&(t=e.getParseTreeNode(t),!t))return e.TypeReferenceSerializationKind.Unknown;var a=Pr(r,67220415,!0,!1,t),i=Pr(r,67897832,!0,!1,t);if(a&&a===i){var o=jl(!1);if(o&&a===o)return e.TypeReferenceSerializationKind.Promise;var s=Xi(a);if(s&&io(s))return e.TypeReferenceSerializationKind.TypeWithConstructSignatureAndValue}if(!i)return e.TypeReferenceSerializationKind.Unknown;var c=Do(i);return c===oe?e.TypeReferenceSerializationKind.Unknown:3&c.flags?e.TypeReferenceSerializationKind.ObjectType:fE(c,245760)?e.TypeReferenceSerializationKind.VoidNullableOrNeverType:fE(c,528)?e.TypeReferenceSerializationKind.BooleanType:fE(c,296)?e.TypeReferenceSerializationKind.NumberLikeType:fE(c,2112)?e.TypeReferenceSerializationKind.BigIntLikeType:fE(c,132)?e.TypeReferenceSerializationKind.StringLikeType:Xm(c)?e.TypeReferenceSerializationKind.ArrayLikeType:fE(c,12288)?e.TypeReferenceSerializationKind.ESSymbolType:nA(c)?e.TypeReferenceSerializationKind.TypeWithCallSignature:Om(c)?e.TypeReferenceSerializationKind.ArrayLikeType:e.TypeReferenceSerializationKind.ObjectType}function rA(n,t,r,a,i){var o=e.getParseTreeNode(n,e.isVariableLikeOrAccessor);if(!o)return e.createToken(120);var s=ra(o),c=!s||133120&s.flags?oe:qm(Xi(s));return 8192&c.flags&&c.symbol===s&&(r|=1048576),i&&(c=sf(c)),K.typeToTypeNode(c,t,1024|r,a)}function aA(n,t,r,a){var i=e.getParseTreeNode(n,e.isFunctionLike);if(!i)return e.createToken(120);var o=Cc(i);return K.typeToTypeNode(Fc(o),t,1024|r,a)}function iA(n,t,r,a){var i=e.getParseTreeNode(n,e.isExpression);if(!i)return e.createToken(120);var o=bf(CC(i));return K.typeToTypeNode(o,t,1024|r,a)}function oA(n){return wn.has(e.escapeLeadingUnderscores(n))}function sA(n,t){var r=zt(n).resolvedSymbol;if(r)return r;var a=n;if(t){var i=n.parent;e.isDeclaration(i)&&n===i.name&&(a=ei(i))}return Zt(a,n.escapedText,70366143,void 0,void 0,!0)}function cA(n){if(!e.isGeneratedIdentifier(n)){var t=e.getParseTreeNode(n,e.isIdentifier);if(t){var r=sA(t);if(r)return ca(r).valueDeclaration}}}function lA(n){return!!(e.isDeclarationReadonly(n)||e.isVariableDeclaration(n)&&e.isVarConst(n))&&kd(Xi(ra(n)))}function uA(n,t,r){var a=1024&n.flags?K.symbolToExpression(n.symbol,67220415,t,void 0,r):n===ve?e.createTrue():n===_e&&e.createFalse();return a||e.createLiteral(n.value)}function dA(e,n){var t=Xi(ra(e));return uA(t,e,n)}function pA(){var n,t=r.getResolvedTypeReferenceDirectives();return t&&(n=e.createMap(),t.forEach(function(e,n){if(e&&e.resolvedFileName){var t=r.getSourceFile(e.resolvedFileName);c(t,n)}})),{getReferencedExportContainer:RC,getReferencedImportDeclaration:FC,getReferencedDeclarationWithCollidingName:VC,isDeclarationWithCollidingName:BC,isValueAliasDeclaration:function(n){return n=e.getParseTreeNode(n),!n||KC(n)},hasGlobalName:oA,isReferencedAliasDeclaration:function(n,t){return n=e.getParseTreeNode(n),!n||WC(n,t)},getNodeCheckFlags:function(n){return n=e.getParseTreeNode(n),n?QC(n):0},isTopLevelValueImportEqualsWithEntityName:jC,isDeclarationVisible:Ja,isImplementationOfOverload:qC,isRequiredInitializedParameter:zC,isOptionalUninitializedParameterProperty:JC,isExpandoFunctionDeclaration:XC,getPropertiesOfContainerFunction:YC,createTypeOfDeclaration:rA,createReturnTypeOfSignatureDeclaration:aA,createTypeOfExpression:iA,createLiteralConstValue:dA,isSymbolAccessible:ka,isEntityNameVisible:Oa,getConstantValue:function(n){var t=e.getParseTreeNode(n,$C);return t?eA(t):void 0},collectLinkedAliases:Xa,getReferencedValueDeclaration:cA,getTypeReferenceSerializationKind:tA,isOptionalParameter:hc,moduleExportsSomeValue:OC,isArgumentsLocalBinding:MC,getExternalModuleFileFromDeclaration:mA,getTypeReferenceDirectivesForEntityName:i,getTypeReferenceDirectivesForSymbol:o,isLiteralConstDeclaration:lA,isLateBound:function(n){var t=e.getParseTreeNode(n,e.isDeclaration),r=t&&ra(t);return!!(r&&2048&e.getCheckFlags(r))},getJsxFactoryEntity:function(n){return n?(Dt(n),e.getSourceFileOfNode(n).localJsxFactory||ct):ct},getAllAccessorDeclarations:function(n){n=e.getParseTreeNode(n,e.isGetOrSetAccessorDeclaration);var t=159===n.kind?158:159,r=e.getDeclarationOfKind(ra(n),t),a=r&&r.pos<n.pos?r:n,i=r&&r.pos<n.pos?n:r,o=159===n.kind?n:r,s=158===n.kind?n:r;return{firstAccessor:a,secondAccessor:i,setAccessor:o,getAccessor:s}},getSymbolOfExternalModuleSpecifier:function(e){return Br(e,e,void 0)},isBindingCapturedByNode:function(n,t){var r=e.getParseTreeNode(n),a=e.getParseTreeNode(t);return!!r&&!!a&&(e.isVariableDeclaration(a)||e.isBindingElement(a))&&Ng(r,a)}};function a(e){return e.parent&&211===e.parent.kind&&e.parent.parent&&273===e.parent.parent.kind}function i(e){if(n){var t=67899368;(72===e.kind&&n_(e)||189===e.kind&&!a(e))&&(t=68268991);var r=Pr(e,t,!0);return r&&r!==ne?o(r,t):void 0}}function o(t,r){if(n&&s(t)){for(var a,i=0,o=t.declarations;i<o.length;i++){var c=o[i];if(c.symbol&&c.symbol.flags&r){var l=e.getSourceFileOfNode(c),u=n.get(l.path);if(!u)return;(a||(a=[])).push(u)}}return a}}function s(t){if(!t.declarations)return!1;var r=t;while(1){var a=aa(r);if(!a)break;r=a}if(r.valueDeclaration&&279===r.valueDeclaration.kind&&512&r.flags)return!1;for(var i=0,o=t.declarations;i<o.length;i++){var s=o[i],c=e.getSourceFileOfNode(s);if(n.has(c.path))return!0}return!1}function c(t,a){if(!n.has(t.path)){n.set(t.path,a);for(var i=0,o=t.referencedFiles;i<o.length;i++){var s=o[i].fileName,l=e.resolveTripleslashReference(s,t.originalFileName),u=r.getSourceFile(l);u&&c(u,a)}}}}function mA(n){var t=244===n.kind?e.tryCast(n.name,e.isStringLiteral):e.getExternalModuleName(n),r=Br(t,t,void 0);if(r)return e.getDeclarationOfKind(r,279)}function fA(){for(var n=0,t=r.getSourceFiles();n<t.length;n++){var a=t[n];e.bindSourceFile(a,L)}var i;Be=e.createMap();for(var o=0,s=r.getSourceFiles();o<s.length;o++){a=s[o];if(!a.redirectInfo&&(e.isExternalOrCommonJsModule(a)||Ut(wn,a.locals),a.jsGlobalAugmentations&&Ut(wn,a.jsGlobalAugmentations),a.patternAmbientModules&&a.patternAmbientModules.length&&(je=e.concatenate(je,a.patternAmbientModules)),a.moduleAugmentations.length&&(i||(i=[])).push(a.moduleAugmentations),a.symbol&&a.symbol.globalExports)){var c=a.symbol.globalExports;c.forEach(function(e,n){wn.has(n)||wn.set(n,e)})}}if(i)for(var l=0,u=i;l<u.length;l++)for(var d=u[l],p=0,m=d;p<m.length;p++){var f=m[p];e.isGlobalScopeAugmentation(f.parent)&&Ht(f)}if(Wt(wn,Ct,e.Diagnostics.Declaration_name_conflicts_with_built_in_global_identifier_0),qt(j).type=le,qt(H).type=Ol("IArguments",0,!0),qt(ne).type=oe,ze=Ol("Array",1,!0),Ue=Ol("Object",0,!0),He=Ol("Function",0,!0),We=O&&Ol("CallableFunction",0,!0)||He,qe=O&&Ol("NewableFunction",0,!0)||He,Xe=Ol("String",0,!0),Ye=Ol("Number",0,!0),Qe=Ol("Boolean",0,!0),Ze=Ol("RegExp",0,!0),en=su(re),nn=su(ae),nn===ke&&(nn=ba(void 0,A,e.emptyArray,e.emptyArray,void 0,void 0)),Je=Yl("ReadonlyArray",1),tn=Je?nu(Je,[re]):en,$e=Yl("ThisType",1),i)for(var _=0,g=i;_<g.length;_++){d=g[_];for(var v=0,y=d;v<y.length;v++){f=y[v];e.isGlobalScopeAugmentation(f.parent)||Ht(f)}}Be.forEach(function(n){var t=n.firstFile,r=n.secondFile,a=n.conflictingSymbols;if(a.size<8)a.forEach(function(n,t){for(var r=n.isBlockScoped,a=n.firstFileLocations,i=n.secondFileLocations,o=r?e.Diagnostics.Cannot_redeclare_block_scoped_variable_0:e.Diagnostics.Duplicate_identifier_0,s=0,c=a;s<c.length;s++){var l=c[s];Kt(l,o,t,i)}for(var u=0,d=i;u<d.length;u++){l=d[u];Kt(l,o,t,a)}});else{var i=e.arrayFrom(a.keys()).join(", ");it.add(Nt(e.createDiagnosticForNode(t,e.Diagnostics.Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0,i),e.createDiagnosticForNode(r,e.Diagnostics.Conflicts_are_in_this_file))),it.add(Nt(e.createDiagnosticForNode(r,e.Diagnostics.Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0,i),e.createDiagnosticForNode(t,e.Diagnostics.Conflicts_are_in_this_file)))}}),Be=void 0}function _A(n,t){if((_&t)!==t&&L.importHelpers){var r=e.getSourceFileOfNode(n);if(e.isEffectiveExternalModule(r,L)&&!(4194304&n.flags)){var a=vA(r,n);if(a!==ne)for(var i=t&~_,o=1;o<=65536;o<<=1)if(i&o){var s=gA(o),c=Xt(a.exports,e.escapeLeadingUnderscores(s),67220415);c||It(n,e.Diagnostics.This_syntax_requires_an_imported_helper_named_1_but_module_0_has_no_exported_member_1,e.externalHelpersModuleNameText,s)}_|=t}}}function gA(n){switch(n){case 1:return"__extends";case 2:return"__assign";case 4:return"__rest";case 8:return"__decorate";case 16:return"__metadata";case 32:return"__param";case 64:return"__awaiter";case 128:return"__generator";case 256:return"__values";case 512:return"__read";case 1024:return"__spread";case 2048:return"__await";case 4096:return"__asyncGenerator";case 8192:return"__asyncDelegator";case 16384:return"__asyncValues";case 32768:return"__exportStar";case 65536:return"__makeTemplateObject";default:return e.Debug.fail("Unrecognized helper")}}function vA(n,t){return g||(g=Kr(n,e.externalHelpersModuleNameText,e.Diagnostics.This_syntax_requires_an_imported_helper_but_module_0_cannot_be_found,t)||ne),g}function yA(e){return hA(e)||bA(e)}function hA(n){if(!n.decorators)return!1;if(!e.nodeCanBeDecorated(n,n.parent,n.parent.parent))return 156!==n.kind||e.nodeIsPresent(n.body)?gD(n,e.Diagnostics.Decorators_are_not_valid_here):gD(n,e.Diagnostics.A_decorator_can_only_decorate_a_method_implementation_not_an_overload);if(158===n.kind||159===n.kind){var t=e.getAllAccessorDeclarations(n.parent.members,n);if(t.firstAccessor.decorators&&n===t.secondAccessor)return gD(n,e.Diagnostics.Decorators_cannot_be_applied_to_multiple_get_Slashset_accessors_of_the_same_name)}return!1}function bA(n){var t,r,a,i,o=EA(n);if(void 0!==o)return o;for(var s=0,c=0,l=n.modifiers;c<l.length;c++){var u=l[c];if(133!==u.kind){if(153===n.kind||155===n.kind)return yD(u,e.Diagnostics._0_modifier_cannot_appear_on_a_type_member,e.tokenToString(u.kind));if(162===n.kind)return yD(u,e.Diagnostics._0_modifier_cannot_appear_on_an_index_signature,e.tokenToString(u.kind))}switch(u.kind){case 77:if(243!==n.kind)return yD(n,e.Diagnostics.A_class_member_cannot_have_the_0_keyword,e.tokenToString(77));break;case 115:case 114:case 113:var d=ja(e.modifierToFlag(u.kind));if(28&s)return yD(u,e.Diagnostics.Accessibility_modifier_already_seen);if(32&s)return yD(u,e.Diagnostics._0_modifier_must_precede_1_modifier,d,"static");if(64&s)return yD(u,e.Diagnostics._0_modifier_must_precede_1_modifier,d,"readonly");if(256&s)return yD(u,e.Diagnostics._0_modifier_must_precede_1_modifier,d,"async");if(245===n.parent.kind||279===n.parent.kind)return yD(u,e.Diagnostics._0_modifier_cannot_appear_on_a_module_or_namespace_element,d);if(128&s)return 113===u.kind?yD(u,e.Diagnostics._0_modifier_cannot_be_used_with_1_modifier,d,"abstract"):yD(u,e.Diagnostics._0_modifier_must_precede_1_modifier,d,"abstract");s|=e.modifierToFlag(u.kind);break;case 116:if(32&s)return yD(u,e.Diagnostics._0_modifier_already_seen,"static");if(64&s)return yD(u,e.Diagnostics._0_modifier_must_precede_1_modifier,"static","readonly");if(256&s)return yD(u,e.Diagnostics._0_modifier_must_precede_1_modifier,"static","async");if(245===n.parent.kind||279===n.parent.kind)return yD(u,e.Diagnostics._0_modifier_cannot_appear_on_a_module_or_namespace_element,"static");if(151===n.kind)return yD(u,e.Diagnostics._0_modifier_cannot_appear_on_a_parameter,"static");if(128&s)return yD(u,e.Diagnostics._0_modifier_cannot_be_used_with_1_modifier,"static","abstract");s|=32,t=u;break;case 133:if(64&s)return yD(u,e.Diagnostics._0_modifier_already_seen,"readonly");if(154!==n.kind&&153!==n.kind&&162!==n.kind&&151!==n.kind)return yD(u,e.Diagnostics.readonly_modifier_can_only_appear_on_a_property_declaration_or_index_signature);s|=64,i=u;break;case 85:if(1&s)return yD(u,e.Diagnostics._0_modifier_already_seen,"export");if(2&s)return yD(u,e.Diagnostics._0_modifier_must_precede_1_modifier,"export","declare");if(128&s)return yD(u,e.Diagnostics._0_modifier_must_precede_1_modifier,"export","abstract");if(256&s)return yD(u,e.Diagnostics._0_modifier_must_precede_1_modifier,"export","async");if(240===n.parent.kind)return yD(u,e.Diagnostics._0_modifier_cannot_appear_on_a_class_element,"export");if(151===n.kind)return yD(u,e.Diagnostics._0_modifier_cannot_appear_on_a_parameter,"export");s|=1;break;case 80:var p=279===n.parent.kind?n.parent:n.parent.parent;if(244===p.kind&&!e.isAmbientModule(p))return yD(u,e.Diagnostics.A_default_export_can_only_be_used_in_an_ECMAScript_style_module);s|=512;break;case 125:if(2&s)return yD(u,e.Diagnostics._0_modifier_already_seen,"declare");if(256&s)return yD(u,e.Diagnostics._0_modifier_cannot_be_used_in_an_ambient_context,"async");if(240===n.parent.kind)return yD(u,e.Diagnostics._0_modifier_cannot_appear_on_a_class_element,"declare");if(151===n.kind)return yD(u,e.Diagnostics._0_modifier_cannot_appear_on_a_parameter,"declare");if(4194304&n.parent.flags&&245===n.parent.kind)return yD(u,e.Diagnostics.A_declare_modifier_cannot_be_used_in_an_already_ambient_context);s|=2,r=u;break;case 118:if(128&s)return yD(u,e.Diagnostics._0_modifier_already_seen,"abstract");if(240!==n.kind){if(156!==n.kind&&154!==n.kind&&158!==n.kind&&159!==n.kind)return yD(u,e.Diagnostics.abstract_modifier_can_only_appear_on_a_class_method_or_property_declaration);if(240!==n.parent.kind||!e.hasModifier(n.parent,128))return yD(u,e.Diagnostics.Abstract_methods_can_only_appear_within_an_abstract_class);if(32&s)return yD(u,e.Diagnostics._0_modifier_cannot_be_used_with_1_modifier,"static","abstract");if(8&s)return yD(u,e.Diagnostics._0_modifier_cannot_be_used_with_1_modifier,"private","abstract")}s|=128;break;case 121:if(256&s)return yD(u,e.Diagnostics._0_modifier_already_seen,"async");if(2&s||4194304&n.parent.flags)return yD(u,e.Diagnostics._0_modifier_cannot_be_used_in_an_ambient_context,"async");if(151===n.kind)return yD(u,e.Diagnostics._0_modifier_cannot_appear_on_a_parameter,"async");s|=256,a=u;break}}return 157===n.kind?32&s?yD(t,e.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration,"static"):128&s?yD(t,e.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration,"abstract"):256&s?yD(a,e.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration,"async"):!!(64&s)&&yD(i,e.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration,"readonly"):(249===n.kind||248===n.kind)&&2&s?yD(r,e.Diagnostics.A_0_modifier_cannot_be_used_with_an_import_declaration,"declare"):151===n.kind&&92&s&&e.isBindingPattern(n.name)?yD(n,e.Diagnostics.A_parameter_property_may_not_be_declared_using_a_binding_pattern):151===n.kind&&92&s&&n.dotDotDotToken?yD(n,e.Diagnostics.A_parameter_property_cannot_be_declared_using_a_rest_parameter):!!(256&s)&&xA(n,a)}function EA(n){return!!n.modifiers&&(TA(n)?gD(n,e.Diagnostics.Modifiers_cannot_appear_here):void 0)}function TA(n){switch(n.kind){case 158:case 159:case 157:case 154:case 153:case 156:case 155:case 162:case 244:case 249:case 248:case 255:case 254:case 196:case 197:case 151:return!1;default:if(245===n.parent.kind||279===n.parent.kind)return!1;switch(n.kind){case 239:return SA(n,121);case 240:return SA(n,118);case 241:case 219:case 242:return!0;case 243:return SA(n,77);default:return e.Debug.fail(),!1}}}function SA(e,n){return e.modifiers.length>1||e.modifiers[0].kind!==n}function xA(n,t){switch(n.kind){case 156:case 239:case 196:case 197:return!1}return yD(t,e.Diagnostics._0_modifier_cannot_be_used_here,"async")}function CA(n,t){return void 0===t&&(t=e.Diagnostics.Trailing_comma_not_allowed),!(!n||!n.hasTrailingComma)&&vD(n[0],n.end-",".length,",".length,t)}function AA(n,t){if(n&&0===n.length){var r=n.pos-"<".length,a=e.skipTrivia(t.text,n.end)+">".length;return vD(t,r,a-r,e.Diagnostics.Type_parameter_list_cannot_be_empty)}return!1}function DA(n){for(var t=!1,r=n.length,a=0;a<r;a++){var i=n[a];if(i.dotDotDotToken){if(a!==r-1)return yD(i.dotDotDotToken,e.Diagnostics.A_rest_parameter_must_be_last_in_a_parameter_list);if(4194304&i.flags||CA(n,e.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma),i.questionToken)return yD(i.questionToken,e.Diagnostics.A_rest_parameter_cannot_be_optional);if(i.initializer)return yD(i.name,e.Diagnostics.A_rest_parameter_cannot_have_an_initializer)}else if(i.questionToken){if(t=!0,i.initializer)return yD(i.name,e.Diagnostics.Parameter_cannot_have_question_mark_and_initializer)}else if(t&&!i.initializer)return yD(i.name,e.Diagnostics.A_required_parameter_cannot_follow_an_optional_parameter)}}function LA(n){return e.filter(n,function(n){return!!n.initializer||e.isBindingPattern(n.name)||e.isRestParameter(n)})}function kA(n){if(k>=3){var t=n.body&&e.isBlock(n.body)&&e.findUseStrictPrologue(n.body.statements);if(t){var r=LA(n.parameters);if(e.length(r)){e.forEach(r,function(n){Nt(It(n,e.Diagnostics.This_parameter_is_not_allowed_with_use_strict_directive),e.createDiagnosticForNode(t,e.Diagnostics.use_strict_directive_used_here))});var a=r.map(function(n,t){return 0===t?e.createDiagnosticForNode(n,e.Diagnostics.Non_simple_parameter_declared_here):e.createDiagnosticForNode(n,e.Diagnostics.and_here)});return Nt.apply(void 0,[It(t,e.Diagnostics.use_strict_directive_cannot_be_used_with_non_simple_parameter_list)].concat(a)),!0}}}return!1}function NA(n){var t=e.getSourceFileOfNode(n);return yA(n)||AA(n.typeParameters,t)||DA(n.parameters)||wA(n,t)||e.isFunctionLikeDeclaration(n)&&kA(n)}function IA(n){var t=e.getSourceFileOfNode(n);return KA(n)||AA(n.typeParameters,t)}function wA(n,t){if(!e.isArrowFunction(n))return!1;var r=n.equalsGreaterThanToken,a=e.getLineAndCharacterOfPosition(t,r.pos).line,i=e.getLineAndCharacterOfPosition(t,r.end).line;return a!==i&&yD(r,e.Diagnostics.Line_terminator_not_permitted_before_arrow)}function MA(n){var t=n.parameters[0];if(1!==n.parameters.length)return yD(t?t.name:n,e.Diagnostics.An_index_signature_must_have_exactly_one_parameter);if(t.dotDotDotToken)return yD(t.dotDotDotToken,e.Diagnostics.An_index_signature_cannot_have_a_rest_parameter);if(e.hasModifiers(t))return yD(t.name,e.Diagnostics.An_index_signature_parameter_cannot_have_an_accessibility_modifier);if(t.questionToken)return yD(t.questionToken,e.Diagnostics.An_index_signature_parameter_cannot_have_a_question_mark);if(t.initializer)return yD(t.name,e.Diagnostics.An_index_signature_parameter_cannot_have_an_initializer);if(!t.type)return yD(t.name,e.Diagnostics.An_index_signature_parameter_must_have_a_type_annotation);if(138!==t.type.kind&&135!==t.type.kind){var r=Rd(t.type);return 4&r.flags||8&r.flags?yD(t.name,e.Diagnostics.An_index_signature_parameter_type_cannot_be_a_type_alias_Consider_writing_0_Colon_1_Colon_2_instead,e.getTextOfNode(t.name),Fa(r),Fa(Rd(n.type))):1048576&r.flags&&_E(r,128,!0)?yD(t.name,e.Diagnostics.An_index_signature_parameter_type_cannot_be_a_union_type_Consider_using_a_mapped_object_type_instead):yD(t.name,e.Diagnostics.An_index_signature_parameter_type_must_be_string_or_number)}return!n.type&&yD(n,e.Diagnostics.An_index_signature_must_have_a_type_annotation)}function OA(e){return yA(e)||MA(e)}function PA(n,t){if(t&&0===t.length){var r=e.getSourceFileOfNode(n),a=t.pos-"<".length,i=e.skipTrivia(r.text,t.end)+">".length;return vD(r,a,i-a,e.Diagnostics.Type_argument_list_cannot_be_empty)}return!1}function RA(e,n){return CA(n)||PA(e,n)}function FA(n){if(n)for(var t=0,r=n;t<r.length;t++){var a=r[t];if(210===a.kind)return vD(a,a.pos,0,e.Diagnostics.Argument_expression_expected)}return!1}function GA(e){return FA(e)}function VA(n){var t=n.types;if(CA(t))return!0;if(t&&0===t.length){var r=e.tokenToString(n.token);return vD(n,t.pos,0,e.Diagnostics._0_list_cannot_be_empty,r)}return e.some(t,BA)}function BA(e){return RA(e,e.typeArguments)}function KA(n){var t=!1,r=!1;if(!yA(n)&&n.heritageClauses)for(var a=0,i=n.heritageClauses;a<i.length;a++){var o=i[a];if(86===o.token){if(t)return gD(o,e.Diagnostics.extends_clause_already_seen);if(r)return gD(o,e.Diagnostics.extends_clause_must_precede_implements_clause);if(o.types.length>1)return gD(o.types[1],e.Diagnostics.Classes_can_only_extend_a_single_class);t=!0}else{if(e.Debug.assert(109===o.token),r)return gD(o,e.Diagnostics.implements_clause_already_seen);r=!0}VA(o)}}function jA(n){var t=!1;if(n.heritageClauses)for(var r=0,a=n.heritageClauses;r<a.length;r++){var i=a[r];if(86!==i.token)return e.Debug.assert(109===i.token),gD(i,e.Diagnostics.Interface_declaration_cannot_have_implements_clause);if(t)return gD(i,e.Diagnostics.extends_clause_already_seen);t=!0,VA(i)}return!1}function UA(n){if(149!==n.kind)return!1;var t=n;return 204===t.expression.kind&&27===t.expression.operatorToken.kind&&yD(t.expression,e.Diagnostics.A_comma_expression_is_not_allowed_in_a_computed_property_name)}function HA(n){if(n.asteriskToken){if(e.Debug.assert(239===n.kind||196===n.kind||156===n.kind),4194304&n.flags)return yD(n.asteriskToken,e.Diagnostics.Generators_are_not_allowed_in_an_ambient_context);if(!n.body)return yD(n.asteriskToken,e.Diagnostics.An_overload_signature_cannot_be_declared_as_a_generator)}}function WA(e,n){return!!e&&yD(e,n)}function qA(e,n){return!!e&&yD(e,n)}function zA(n,t){var r;(function(e){e[e["Property"]=1]="Property",e[e["GetAccessor"]=2]="GetAccessor",e[e["SetAccessor"]=4]="SetAccessor",e[e["GetOrSetAccessor"]=6]="GetOrSetAccessor"})(r||(r={}));for(var a=e.createUnderscoreEscapedMap(),i=0,o=n.properties;i<o.length;i++){var s=o[i];if(277!==s.kind){var c=s.name;if(149===c.kind&&UA(c),276===s.kind&&!t&&s.objectAssignmentInitializer)return yD(s.equalsToken,e.Diagnostics.can_only_be_used_in_an_object_literal_property_inside_a_destructuring_assignment);if(s.modifiers)for(var l=0,u=s.modifiers;l<u.length;l++){var d=u[l];121===d.kind&&156===s.kind||yD(d,e.Diagnostics._0_modifier_cannot_be_used_here,e.getTextOfNode(d))}var p=void 0;switch(s.kind){case 276:qA(s.exclamationToken,e.Diagnostics.A_definite_assignment_assertion_is_not_permitted_in_this_context);case 275:WA(s.questionToken,e.Diagnostics.An_object_member_cannot_be_declared_optional),8===c.kind&&AD(c);case 156:p=1;break;case 158:p=2;break;case 159:p=4;break;default:throw e.Debug.assertNever(s,"Unexpected syntax kind:"+s.kind)}var m=e.getPropertyNameForPropertyNameNode(c);if(void 0!==m){var f=a.get(m);if(f)if(1===p&&1===f)yD(c,e.Diagnostics.Duplicate_identifier_0,e.getTextOfNode(c));else{if(!(6&p&&6&f))return yD(c,e.Diagnostics.An_object_literal_cannot_have_property_and_accessor_with_the_same_name);if(6===f||p===f)return yD(c,e.Diagnostics.An_object_literal_cannot_have_multiple_get_Slashset_accessors_with_the_same_name);a.set(m,p|f)}else a.set(m,p)}}}}function JA(n){RA(n,n.typeArguments);for(var t=e.createUnderscoreEscapedMap(),r=0,a=n.attributes.properties;r<a.length;r++){var i=a[r];if(269!==i.kind){var o=i.name,s=i.initializer;if(t.get(o.escapedText))return yD(o,e.Diagnostics.JSX_elements_cannot_have_multiple_attributes_with_the_same_name);if(t.set(o.escapedText,!0),s&&270===s.kind&&!s.expression)return yD(s,e.Diagnostics.JSX_attributes_must_only_be_assigned_a_non_empty_expression)}}}function XA(n){if(CD(n))return!0;if(227===n.kind&&n.awaitModifier&&0===(16384&n.flags))return yD(n.awaitModifier,e.Diagnostics.A_for_await_of_statement_is_only_allowed_within_an_async_function_or_async_generator);if(238===n.initializer.kind){var t=n.initializer;if(!dD(t)){var r=t.declarations;if(!r.length)return!1;if(r.length>1){var a=226===n.kind?e.Diagnostics.Only_a_single_variable_declaration_is_allowed_in_a_for_in_statement:e.Diagnostics.Only_a_single_variable_declaration_is_allowed_in_a_for_of_statement;return gD(t.declarations[1],a)}var i=r[0];if(i.initializer){a=226===n.kind?e.Diagnostics.The_variable_declaration_of_a_for_in_statement_cannot_have_an_initializer:e.Diagnostics.The_variable_declaration_of_a_for_of_statement_cannot_have_an_initializer;return yD(i.name,a)}if(i.type){a=226===n.kind?e.Diagnostics.The_left_hand_side_of_a_for_in_statement_cannot_use_a_type_annotation:e.Diagnostics.The_left_hand_side_of_a_for_of_statement_cannot_use_a_type_annotation;return yD(i,a)}}}return!1}function YA(n){var t=n.kind;if(k<1)return yD(n.name,e.Diagnostics.Accessors_are_only_available_when_targeting_ECMAScript_5_and_higher);if(4194304&n.flags)return yD(n.name,e.Diagnostics.An_accessor_cannot_be_declared_in_an_ambient_context);if(void 0===n.body&&!e.hasModifier(n,128))return vD(n,n.end-1,";".length,e.Diagnostics._0_expected,"{");if(n.body&&e.hasModifier(n,128))return yD(n,e.Diagnostics.An_abstract_accessor_cannot_have_an_implementation);if(n.typeParameters)return yD(n.name,e.Diagnostics.An_accessor_cannot_have_type_parameters);if(!QA(n))return yD(n.name,158===t?e.Diagnostics.A_get_accessor_cannot_have_parameters:e.Diagnostics.A_set_accessor_must_have_exactly_one_parameter);if(159===t){if(n.type)return yD(n.name,e.Diagnostics.A_set_accessor_cannot_have_a_return_type_annotation);var r=n.parameters[0];if(r.dotDotDotToken)return yD(r.dotDotDotToken,e.Diagnostics.A_set_accessor_cannot_have_rest_parameter);if(r.questionToken)return yD(r.questionToken,e.Diagnostics.A_set_accessor_cannot_have_an_optional_parameter);if(r.initializer)return yD(n.name,e.Diagnostics.A_set_accessor_parameter_cannot_have_an_initializer)}return!1}function QA(e){return ZA(e)||e.parameters.length===(158===e.kind?0:1)}function ZA(n){if(n.parameters.length===(158===n.kind?1:2))return e.getThisParameter(n)}function $A(n){if(142===n.operator){if(139!==n.type.kind)return yD(n.type,e.Diagnostics._0_expected,e.tokenToString(139));var t=e.walkUpParenthesizedTypes(n.parent);switch(t.kind){case 237:var r=t;if(72!==r.name.kind)return yD(n,e.Diagnostics.unique_symbol_types_may_not_be_used_on_a_variable_declaration_with_a_binding_name);if(!e.isVariableDeclarationInVariableStatement(r))return yD(n,e.Diagnostics.unique_symbol_types_are_only_allowed_on_variables_in_a_variable_statement);if(!(2&r.parent.flags))return yD(t.name,e.Diagnostics.A_variable_whose_type_is_a_unique_symbol_type_must_be_const);break;case 154:if(!e.hasModifier(t,32)||!e.hasModifier(t,64))return yD(t.name,e.Diagnostics.A_property_of_a_class_whose_type_is_a_unique_symbol_type_must_be_both_static_and_readonly);break;case 153:if(!e.hasModifier(t,64))return yD(t.name,e.Diagnostics.A_property_of_an_interface_or_type_literal_whose_type_is_a_unique_symbol_type_must_be_readonly);break;default:return yD(n,e.Diagnostics.unique_symbol_types_are_not_allowed_here)}}}function eD(e,n){if(jo(e))return yD(e,n)}function nD(n){if(NA(n))return!0;if(156===n.kind){if(188===n.parent.kind){if(n.modifiers&&(1!==n.modifiers.length||121!==e.first(n.modifiers).kind))return gD(n,e.Diagnostics.Modifiers_cannot_appear_here);if(WA(n.questionToken,e.Diagnostics.An_object_member_cannot_be_declared_optional))return!0;if(qA(n.exclamationToken,e.Diagnostics.A_definite_assignment_assertion_is_not_permitted_in_this_context))return!0;if(void 0===n.body)return vD(n,n.end-1,";".length,e.Diagnostics._0_expected,"{")}if(HA(n))return!0}if(e.isClassLike(n.parent)){if(4194304&n.flags)return eD(n.name,e.Diagnostics.A_computed_property_name_in_an_ambient_context_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type);if(156===n.kind&&!n.body)return eD(n.name,e.Diagnostics.A_computed_property_name_in_a_method_overload_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type)}else{if(241===n.parent.kind)return eD(n.name,e.Diagnostics.A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type);if(168===n.parent.kind)return eD(n.name,e.Diagnostics.A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type)}}function tD(n){var t=n;while(t){if(e.isFunctionLike(t))return yD(n,e.Diagnostics.Jump_target_cannot_cross_function_boundary);switch(t.kind){case 233:if(n.label&&t.label.escapedText===n.label.escapedText){var r=228===n.kind&&!e.isIterationStatement(t.statement,!0);return!!r&&yD(n,e.Diagnostics.A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement)}break;case 232:if(229===n.kind&&!n.label)return!1;break;default:if(e.isIterationStatement(t,!1)&&!n.label)return!1;break}t=t.parent}if(n.label){var a=229===n.kind?e.Diagnostics.A_break_statement_can_only_jump_to_a_label_of_an_enclosing_statement:e.Diagnostics.A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement;return yD(n,a)}a=229===n.kind?e.Diagnostics.A_break_statement_can_only_be_used_within_an_enclosing_iteration_or_switch_statement:e.Diagnostics.A_continue_statement_can_only_be_used_within_an_enclosing_iteration_statement;return yD(n,a)}function rD(n){if(n.dotDotDotToken){var t=n.parent.elements;if(n!==e.last(t))return yD(n,e.Diagnostics.A_rest_element_must_be_last_in_a_destructuring_pattern);if(CA(t,e.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma),n.propertyName)return yD(n.name,e.Diagnostics.A_rest_element_cannot_have_a_property_name);if(n.initializer)return vD(n,n.initializer.pos-1,1,e.Diagnostics.A_rest_element_cannot_have_an_initializer)}}function aD(e){return 10===e.kind||8===e.kind||202===e.kind&&39===e.operator&&8===e.operand.kind}function iD(e){return 9===e.kind||202===e.kind&&39===e.operator&&9===e.operand.kind}function oD(n){if((e.isPropertyAccessExpression(n)||e.isElementAccessExpression(n)&&aD(n.argumentExpression))&&e.isEntityNameExpression(n.expression))return!!(1024&RE(n).flags)}function sD(n){var t=n.initializer;if(t){var r=!(aD(t)||oD(t)||102===t.kind||87===t.kind||iD(t)),a=e.isDeclarationReadonly(n)||e.isVariableDeclaration(n)&&e.isVarConst(n);if(!a||n.type)return yD(t,e.Diagnostics.Initializers_are_not_allowed_in_ambient_contexts);if(r)return yD(t,e.Diagnostics.A_const_initializer_in_an_ambient_context_must_be_a_string_or_numeric_literal_or_literal_enum_reference);if(!a||r)return yD(t,e.Diagnostics.Initializers_are_not_allowed_in_ambient_contexts)}}function cD(n){if(226!==n.parent.parent.kind&&227!==n.parent.parent.kind)if(4194304&n.flags)sD(n);else if(!n.initializer){if(e.isBindingPattern(n.name)&&!e.isBindingPattern(n.parent))return yD(n,e.Diagnostics.A_destructuring_declaration_must_have_an_initializer);if(e.isVarConst(n))return yD(n,e.Diagnostics.const_declarations_must_be_initialized)}if(n.exclamationToken&&(219!==n.parent.parent.kind||!n.type||n.initializer||4194304&n.flags))return yD(n.exclamationToken,e.Diagnostics.A_definite_assignment_assertion_is_not_permitted_in_this_context);L.module===e.ModuleKind.ES2015||L.module===e.ModuleKind.ESNext||L.module===e.ModuleKind.System||L.noEmit||4194304&n.parent.parent.flags||!e.hasModifier(n.parent.parent,1)||lD(n.name);var t=e.isLet(n)||e.isVarConst(n);return t&&uD(n.name)}function lD(n){if(72===n.kind){if("__esModule"===e.idText(n))return yD(n,e.Diagnostics.Identifier_expected_esModule_is_reserved_as_an_exported_marker_when_transforming_ECMAScript_modules)}else for(var t=n.elements,r=0,a=t;r<a.length;r++){var i=a[r];if(!e.isOmittedExpression(i))return lD(i.name)}return!1}function uD(n){if(72===n.kind){if(111===n.originalKeywordKind)return yD(n,e.Diagnostics.let_is_not_allowed_to_be_used_as_a_name_in_let_or_const_declarations)}else for(var t=n.elements,r=0,a=t;r<a.length;r++){var i=a[r];e.isOmittedExpression(i)||uD(i.name)}return!1}function dD(n){var t=n.declarations;return!!CA(n.declarations)||!n.declarations.length&&vD(n,t.pos,t.end-t.pos,e.Diagnostics.Variable_declaration_list_cannot_be_empty)}function pD(e){switch(e.kind){case 222:case 223:case 224:case 231:case 225:case 226:case 227:return!1;case 233:return pD(e.parent)}return!0}function mD(n){if(!pD(n.parent)){if(e.isLet(n.declarationList))return yD(n,e.Diagnostics.let_declarations_can_only_be_declared_inside_a_block);if(e.isVarConst(n.declarationList))return yD(n,e.Diagnostics.const_declarations_can_only_be_declared_inside_a_block)}}function fD(n){var t=n.name.escapedText;switch(n.keywordToken){case 95:if("target"!==t)return yD(n.name,e.Diagnostics._0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2,n.name.escapedText,e.tokenToString(n.keywordToken),"target");break;case 92:if("meta"!==t)return yD(n.name,e.Diagnostics._0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2,n.name.escapedText,e.tokenToString(n.keywordToken),"meta");break}}function _D(e){return e.parseDiagnostics.length>0}function gD(n,t,r,a,i){var o=e.getSourceFileOfNode(n);if(!_D(o)){var s=e.getSpanOfTokenAtPosition(o,n.pos);return it.add(e.createFileDiagnostic(o,s.start,s.length,t,r,a,i)),!0}return!1}function vD(n,t,r,a,i,o,s){var c=e.getSourceFileOfNode(n);return!_D(c)&&(it.add(e.createFileDiagnostic(c,t,r,a,i,o,s)),!0)}function yD(n,t,r,a,i){var o=e.getSourceFileOfNode(n);return!_D(o)&&(it.add(e.createDiagnosticForNode(n,t,r,a,i)),!0)}function hD(n){var t=e.isInJSFile(n)?e.getJSDocTypeParameterDeclarations(n):void 0,r=n.typeParameters||t&&e.firstOrUndefined(t);if(r){var a=r.pos===r.end?r.pos:e.skipTrivia(e.getSourceFileOfNode(n).text,r.pos);return vD(n,a,r.end-a,e.Diagnostics.Type_parameters_cannot_appear_on_a_constructor_declaration)}}function bD(n){var t=e.getEffectiveReturnTypeNode(n);if(t)return yD(t,e.Diagnostics.Type_annotation_cannot_appear_on_a_constructor_declaration)}function ED(n){if(e.isClassLike(n.parent)){if(eD(n.name,e.Diagnostics.A_computed_property_name_in_a_class_property_declaration_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type))return!0}else if(241===n.parent.kind){if(eD(n.name,e.Diagnostics.A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type))return!0;if(n.initializer)return yD(n.initializer,e.Diagnostics.An_interface_property_cannot_have_an_initializer)}else if(168===n.parent.kind){if(eD(n.name,e.Diagnostics.A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type))return!0;if(n.initializer)return yD(n.initializer,e.Diagnostics.A_type_literal_property_cannot_have_an_initializer)}if(4194304&n.flags&&sD(n),e.isPropertyDeclaration(n)&&n.exclamationToken&&(!e.isClassLike(n.parent)||!n.type||n.initializer||4194304&n.flags||e.hasModifier(n,160)))return yD(n.exclamationToken,e.Diagnostics.A_definite_assignment_assertion_is_not_permitted_in_this_context)}function TD(n){return 241!==n.kind&&242!==n.kind&&249!==n.kind&&248!==n.kind&&255!==n.kind&&254!==n.kind&&247!==n.kind&&!e.hasModifier(n,515)&&gD(n,e.Diagnostics.A_declare_modifier_is_required_for_a_top_level_declaration_in_a_d_ts_file)}function SD(n){for(var t=0,r=n.statements;t<r.length;t++){var a=r[t];if((e.isDeclaration(a)||219===a.kind)&&TD(a))return!0}return!1}function xD(e){return!!(4194304&e.flags)&&SD(e)}function CD(n){if(4194304&n.flags){if(e.isAccessor(n.parent))return zt(n).hasReportedStatementInAmbientContext=!0;var t=zt(n);if(!t.hasReportedStatementInAmbientContext&&e.isFunctionLike(n.parent))return zt(n).hasReportedStatementInAmbientContext=gD(n,e.Diagnostics.An_implementation_cannot_be_declared_in_ambient_contexts);if(218===n.parent.kind||245===n.parent.kind||279===n.parent.kind){var r=zt(n.parent);if(!r.hasReportedStatementInAmbientContext)return r.hasReportedStatementInAmbientContext=gD(n,e.Diagnostics.Statements_are_not_allowed_in_ambient_contexts)}}return!1}function AD(n){if(32&n.numericLiteralFlags){var t=void 0;if(k>=1?t=e.Diagnostics.Octal_literals_are_not_available_when_targeting_ECMAScript_5_and_higher_Use_the_syntax_0:e.isChildOfNodeWithKind(n,182)?t=e.Diagnostics.Octal_literal_types_must_use_ES2015_syntax_Use_the_syntax_0:e.isChildOfNodeWithKind(n,278)&&(t=e.Diagnostics.Octal_literals_are_not_allowed_in_enums_members_initializer_Use_the_syntax_0),t){var r=e.isPrefixUnaryExpression(n.parent)&&39===n.parent.operator,a=(r?"-":"")+"0o"+n.text;return yD(r?n.parent:n,t,a)}}return!1}function DD(n){var t=e.isLiteralTypeNode(n.parent)||e.isPrefixUnaryExpression(n.parent)&&e.isLiteralTypeNode(n.parent.parent);return!!(!t&&k<6&&yD(n,e.Diagnostics.BigInt_literals_are_not_available_when_targeting_lower_than_ESNext))}function LD(n,t,r,a,i){var o=e.getSourceFileOfNode(n);if(!_D(o)){var s=e.getSpanOfTokenAtPosition(o,n.pos);return it.add(e.createFileDiagnostic(o,e.textSpanEnd(s),0,t,r,a,i)),!0}return!1}function kD(){return Ke||(Ke=[],wn.forEach(function(e,n){t.test(n)&&Ke.push(e)})),Ke}function ND(n){if(N===e.ModuleKind.ES2015)return yD(n,e.Diagnostics.Dynamic_import_is_only_supported_when_module_flag_is_commonjs_or_esNext);if(n.typeArguments)return yD(n,e.Diagnostics.Dynamic_import_cannot_have_type_arguments);var t=n.arguments;return 1!==t.length?yD(n,e.Diagnostics.Dynamic_import_must_have_one_specifier_as_an_argument):(CA(t),!!e.isSpreadElement(t[0])&&yD(t[0],e.Diagnostics.Specifier_of_dynamic_import_cannot_be_spread_element))}}function m(n){switch(n.parent.kind){case 253:case 257:return e.isIdentifier(n);default:return e.isDeclarationName(n)}}function f(e){switch(e.kind){case 250:case 248:case 251:case 253:return!0;case 72:return 253===e.parent.kind;default:return!1}}e.getNodeId=l,e.getSymbolId=u,e.isInstantiatedModule=d,e.createTypeChecker=p,function(e){e.JSX="JSX",e.IntrinsicElements="IntrinsicElements",e.ElementClass="ElementClass",e.ElementAttributesPropertyNameContainer="ElementAttributesProperty",e.ElementChildrenAttributeNameContainer="ElementChildrenAttribute",e.Element="Element",e.IntrinsicAttributes="IntrinsicAttributes",e.IntrinsicClassAttributes="IntrinsicClassAttributes",e.LibraryManagedAttributes="LibraryManagedAttributes"}(n||(n={}))}(u||(u={})),function(e){function n(n){var t=e.createNode(n,-1,-1);return t.flags|=8,t}function t(n,t){return n!==t&&(Oi(n,t),si(n,t),e.aggregateTransformFlags(n)),n}function r(n,t){if(n&&n!==e.emptyArray){if(e.isNodeArray(n))return n}else n=[];var r=n;return r.pos=-1,r.end=-1,r.hasTrailingComma=t,r}function a(e){if(void 0===e)return e;var t=n(e.kind);for(var r in t.flags|=e.flags,Oi(t,e),e)!t.hasOwnProperty(r)&&e.hasOwnProperty(r)&&(t[r]=e[r]);return t}function i(n,t){if("number"===typeof n)return o(n+"");if("object"===typeof n&&"base10Value"in n)return s(e.pseudoBigIntToString(n)+"n");if("boolean"===typeof n)return n?C():A();if(e.isString(n)){var r=c(n);return t&&(r.singleQuote=!0),r}return u(n)}function o(e){var t=n(8);return t.text=e,t.numericLiteralFlags=0,t}function s(e){var t=n(9);return t.text=e,t}function c(e){var t=n(10);return t.text=e,t}function l(e){var t=n(13);return t.text=e,t}function u(n){var t=c(e.getTextOfIdentifierOrLiteral(n));return t.textSourceNode=n,t}function d(t,a){var i=n(72);return i.escapedText=e.escapeLeadingUnderscores(t),i.originalKeywordKind=t?e.stringToToken(t):0,i.autoGenerateFlags=0,i.autoGenerateId=0,a&&(i.typeArguments=r(a)),i}function p(n,r){return n.typeArguments!==r?t(d(e.idText(n),r),n):n}e.updateNode=t,e.createNodeArray=r,e.getSynthesizedClone=a,e.createLiteral=i,e.createNumericLiteral=o,e.createBigIntLiteral=s,e.createStringLiteral=c,e.createRegularExpressionLiteral=l,e.createIdentifier=d,e.updateIdentifier=p;var m,f=0;function _(e,n){var t=d("");return t.autoGenerateFlags=1,t.autoGenerateId=f,f++,e&&e(t),n&&(t.autoGenerateFlags|=8),t}function g(){var e=d("");return e.autoGenerateFlags=2,e.autoGenerateId=f,f++,e}function v(e){var n=d(e);return n.autoGenerateFlags=3,n.autoGenerateId=f,f++,n}function y(e){var n=d(e);return n.autoGenerateFlags=19,n.autoGenerateId=f,f++,n}function h(e){var n=y(e);return n.autoGenerateFlags|=32,n}function b(n,t){var r=d(n&&e.isIdentifier(n)?e.idText(n):"");return r.autoGenerateFlags=4|t,r.autoGenerateId=f,r.original=n,f++,r}function E(e){return n(e)}function T(){return n(98)}function S(){return n(100)}function x(){return n(96)}function C(){return n(102)}function A(){return n(87)}function D(e){return E(e)}function L(e){var n=[];return 1&e&&n.push(D(85)),2&e&&n.push(D(125)),512&e&&n.push(D(80)),2048&e&&n.push(D(77)),4&e&&n.push(D(115)),8&e&&n.push(D(113)),16&e&&n.push(D(114)),128&e&&n.push(D(118)),32&e&&n.push(D(116)),64&e&&n.push(D(133)),256&e&&n.push(D(121)),n}function k(e,t){var r=n(148);return r.left=e,r.right=ni(t),r}function N(e,n,r){return e.left!==n||e.right!==r?t(k(n,r),e):e}function I(n){return e.isCommaSequence(n)?En(n):n}function w(e){var t=n(149);return t.expression=I(e),t}function M(e,n){return e.expression!==n?t(w(n),e):e}function O(e,t,r){var a=n(150);return a.name=ni(e),a.constraint=t,a.default=r,a}function P(e,n,r,a){return e.name!==n||e.constraint!==r||e.default!==a?t(O(n,r,a),e):e}function R(t,r,a,i,o,s,c){var l=n(151);return l.decorators=ri(t),l.modifiers=ri(r),l.dotDotDotToken=a,l.name=ni(i),l.questionToken=o,l.type=s,l.initializer=c?e.parenthesizeExpressionForList(c):void 0,l}function F(e,n,r,a,i,o,s,c){return e.decorators!==n||e.modifiers!==r||e.dotDotDotToken!==a||e.name!==i||e.questionToken!==o||e.type!==s||e.initializer!==c?t(R(n,r,a,i,o,s,c),e):e}function G(t){var r=n(152);return r.expression=e.parenthesizeForAccess(t),r}function V(e,n){return e.expression!==n?t(G(n),e):e}function B(e,t,r,a,i){var o=n(153);return o.modifiers=ri(e),o.name=ni(t),o.questionToken=r,o.type=a,o.initializer=i,o}function K(e,n,r,a,i,o){return e.modifiers!==n||e.name!==r||e.questionToken!==a||e.type!==i||e.initializer!==o?t(B(n,r,a,i,o),e):e}function j(e,t,r,a,i,o){var s=n(154);return s.decorators=ri(e),s.modifiers=ri(t),s.name=ni(r),s.questionToken=void 0!==a&&56===a.kind?a:void 0,s.exclamationToken=void 0!==a&&52===a.kind?a:void 0,s.type=i,s.initializer=o,s}function U(e,n,r,a,i,o,s){return e.decorators!==n||e.modifiers!==r||e.name!==a||e.questionToken!==(void 0!==i&&56===i.kind?i:void 0)||e.exclamationToken!==(void 0!==i&&52===i.kind?i:void 0)||e.type!==o||e.initializer!==s?t(j(n,r,a,i,o,s),e):e}function H(e,n,t,r,a){var i=oe(155,e,n,t);return i.name=ni(r),i.questionToken=a,i}function W(e,n,r,a,i,o){return e.typeParameters!==n||e.parameters!==r||e.type!==a||e.name!==i||e.questionToken!==o?t(H(n,r,a,i,o),e):e}function q(e,t,a,i,o,s,c,l,u){var d=n(156);return d.decorators=ri(e),d.modifiers=ri(t),d.asteriskToken=a,d.name=ni(i),d.questionToken=o,d.typeParameters=ri(s),d.parameters=r(c),d.type=l,d.body=u,d}function z(e,n,r,a,i,o,s,c,l,u){return e.decorators!==n||e.modifiers!==r||e.asteriskToken!==a||e.name!==i||e.questionToken!==o||e.typeParameters!==s||e.parameters!==c||e.type!==l||e.body!==u?t(q(n,r,a,i,o,s,c,l,u),e):e}function J(e,t,a,i){var o=n(157);return o.decorators=ri(e),o.modifiers=ri(t),o.typeParameters=void 0,o.parameters=r(a),o.type=void 0,o.body=i,o}function X(e,n,r,a,i){return e.decorators!==n||e.modifiers!==r||e.parameters!==a||e.body!==i?t(J(n,r,a,i),e):e}function Y(e,t,a,i,o,s){var c=n(158);return c.decorators=ri(e),c.modifiers=ri(t),c.name=ni(a),c.typeParameters=void 0,c.parameters=r(i),c.type=o,c.body=s,c}function Q(e,n,r,a,i,o,s){return e.decorators!==n||e.modifiers!==r||e.name!==a||e.parameters!==i||e.type!==o||e.body!==s?t(Y(n,r,a,i,o,s),e):e}function Z(e,t,a,i,o){var s=n(159);return s.decorators=ri(e),s.modifiers=ri(t),s.name=ni(a),s.typeParameters=void 0,s.parameters=r(i),s.body=o,s}function $(e,n,r,a,i,o){return e.decorators!==n||e.modifiers!==r||e.name!==a||e.parameters!==i||e.body!==o?t(Z(n,r,a,i,o),e):e}function ee(e,n,t){return oe(160,e,n,t)}function ne(e,n,t,r){return se(e,n,t,r)}function te(e,n,t){return oe(161,e,n,t)}function re(e,n,t,r){return se(e,n,t,r)}function ae(e,t,a,i){var o=n(162);return o.decorators=ri(e),o.modifiers=ri(t),o.parameters=r(a),o.type=i,o}function ie(e,n,r,a,i){return e.parameters!==a||e.type!==i||e.decorators!==n||e.modifiers!==r?t(ae(n,r,a,i),e):e}function oe(e,t,r,a,i){var o=n(e);return o.typeParameters=ri(t),o.parameters=ri(r),o.type=a,o.typeArguments=ri(i),o}function se(e,n,r,a){return e.typeParameters!==n||e.parameters!==r||e.type!==a?t(oe(e.kind,n,r,a),e):e}function ce(e){return n(e)}function le(e,t){var r=n(163);return r.parameterName=ni(e),r.type=t,r}function ue(e,n,r){return e.parameterName!==n||e.type!==r?t(le(n,r),e):e}function de(t,r){var a=n(164);return a.typeName=ni(t),a.typeArguments=r&&e.parenthesizeTypeParameters(r),a}function pe(e,n,r){return e.typeName!==n||e.typeArguments!==r?t(de(n,r),e):e}function me(e,n,t){return oe(165,e,n,t)}function fe(e,n,t,r){return se(e,n,t,r)}function _e(e,n,t){return oe(166,e,n,t)}function ge(e,n,t,r){return se(e,n,t,r)}function ve(e){var t=n(167);return t.exprName=e,t}function ye(e,n){return e.exprName!==n?t(ve(n),e):e}function he(e){var t=n(168);return t.members=r(e),t}function be(e,n){return e.members!==n?t(he(n),e):e}function Ee(t){var r=n(169);return r.elementType=e.parenthesizeArrayTypeMember(t),r}function Te(e,n){return e.elementType!==n?t(Ee(n),e):e}function Se(e){var t=n(170);return t.elementTypes=r(e),t}function xe(e,n){return e.elementTypes!==n?t(Se(n),e):e}function Ce(t){var r=n(171);return r.type=e.parenthesizeArrayTypeMember(t),r}function Ae(e,n){return e.type!==n?t(Ce(n),e):e}function De(e){var t=n(172);return t.type=e,t}function Le(e,n){return e.type!==n?t(De(n),e):e}function ke(e){return Me(173,e)}function Ne(e,n){return Oe(e,n)}function Ie(e){return Me(174,e)}function we(e,n){return Oe(e,n)}function Me(t,r){var a=n(t);return a.types=e.parenthesizeElementTypeMembers(r),a}function Oe(e,n){return e.types!==n?t(Me(e.kind,n),e):e}function Pe(t,r,a,i){var o=n(175);return o.checkType=e.parenthesizeConditionalTypeMember(t),o.extendsType=e.parenthesizeConditionalTypeMember(r),o.trueType=a,o.falseType=i,o}function Re(e,n,r,a,i){return e.checkType!==n||e.extendsType!==r||e.trueType!==a||e.falseType!==i?t(Pe(n,r,a,i),e):e}function Fe(e){var t=n(176);return t.typeParameter=e,t}function Ge(e,n){return e.typeParameter!==n?t(Fe(n),e):e}function Ve(e,t,r,a){var i=n(183);return i.argument=e,i.qualifier=t,i.typeArguments=ri(r),i.isTypeOf=a,i}function Be(e,n,r,a,i){return e.argument!==n||e.qualifier!==r||e.typeArguments!==a||e.isTypeOf!==i?t(Ve(n,r,a,i),e):e}function Ke(e){var t=n(177);return t.type=e,t}function je(e,n){return e.type!==n?t(Ke(n),e):e}function Ue(){return n(178)}function He(t,r){var a=n(179);return a.operator="number"===typeof t?t:129,a.type=e.parenthesizeElementTypeMember("number"===typeof t?r:t),a}function We(e,n){return e.type!==n?t(He(e.operator,n),e):e}function qe(t,r){var a=n(180);return a.objectType=e.parenthesizeElementTypeMember(t),a.indexType=r,a}function ze(e,n,r){return e.objectType!==n||e.indexType!==r?t(qe(n,r),e):e}function Je(e,t,r,a){var i=n(181);return i.readonlyToken=e,i.typeParameter=t,i.questionToken=r,i.type=a,i}function Xe(e,n,r,a,i){return e.readonlyToken!==n||e.typeParameter!==r||e.questionToken!==a||e.type!==i?t(Je(n,r,a,i),e):e}function Ye(e){var t=n(182);return t.literal=e,t}function Qe(e,n){return e.literal!==n?t(Ye(n),e):e}function Ze(e){var t=n(184);return t.elements=r(e),t}function $e(e,n){return e.elements!==n?t(Ze(n),e):e}function en(e){var t=n(185);return t.elements=r(e),t}function nn(e,n){return e.elements!==n?t(en(n),e):e}function tn(e,t,r,a){var i=n(186);return i.dotDotDotToken=e,i.propertyName=ni(t),i.name=ni(r),i.initializer=a,i}function rn(e,n,r,a,i){return e.propertyName!==r||e.dotDotDotToken!==n||e.name!==a||e.initializer!==i?t(tn(n,r,a,i),e):e}function an(t,a){var i=n(187);return i.elements=e.parenthesizeListElements(r(t)),a&&(i.multiLine=!0),i}function on(e,n){return e.elements!==n?t(an(n,e.multiLine),e):e}function sn(e,t){var a=n(188);return a.properties=r(e),t&&(a.multiLine=!0),a}function cn(e,n){return e.properties!==n?t(sn(n,e.multiLine),e):e}function ln(t,r){var a=n(189);return a.expression=e.parenthesizeForAccess(t),a.name=ni(r),ci(a,131072),a}function un(n,r,a){return n.expression!==r||n.name!==a?t(ci(ln(r,a),e.getEmitFlags(n)),n):n}function dn(t,r){var a=n(190);return a.expression=e.parenthesizeForAccess(t),a.argumentExpression=ti(r),a}function pn(e,n,r){return e.expression!==n||e.argumentExpression!==r?t(dn(n,r),e):e}function mn(t,a,i){var o=n(191);return o.expression=e.parenthesizeForAccess(t),o.typeArguments=ri(a),o.arguments=e.parenthesizeListElements(r(i)),o}function fn(e,n,r,a){return e.expression!==n||e.typeArguments!==r||e.arguments!==a?t(mn(n,r,a),e):e}function _n(t,a,i){var o=n(192);return o.expression=e.parenthesizeForNew(t),o.typeArguments=ri(a),o.arguments=i?e.parenthesizeListElements(r(i)):void 0,o}function gn(e,n,r,a){return e.expression!==n||e.typeArguments!==r||e.arguments!==a?t(_n(n,r,a),e):e}function vn(t,r,a){var i=n(193);return i.tag=e.parenthesizeForAccess(t),a?(i.typeArguments=ri(r),i.template=a):(i.typeArguments=void 0,i.template=r),i}function yn(e,n,r,a){return e.tag!==n||(a?e.typeArguments!==r||e.template!==a:void 0!==e.typeArguments||e.template!==r)?t(vn(n,r,a),e):e}function hn(t,r){var a=n(194);return a.type=t,a.expression=e.parenthesizePrefixOperand(r),a}function bn(e,n,r){return e.type!==n||e.expression!==r?t(hn(n,r),e):e}function En(e){var t=n(195);return t.expression=e,t}function Tn(e,n){return e.expression!==n?t(En(n),e):e}function Sn(e,t,a,i,o,s,c){var l=n(196);return l.modifiers=ri(e),l.asteriskToken=t,l.name=ni(a),l.typeParameters=ri(i),l.parameters=r(o),l.type=s,l.body=c,l}function xn(e,n,r,a,i,o,s,c){return e.name!==a||e.modifiers!==n||e.asteriskToken!==r||e.typeParameters!==i||e.parameters!==o||e.type!==s||e.body!==c?t(Sn(n,r,a,i,o,s,c),e):e}function Cn(t,a,i,o,s,c){var l=n(197);return l.modifiers=ri(t),l.typeParameters=ri(a),l.parameters=r(i),l.type=o,l.equalsGreaterThanToken=s||E(37),l.body=e.parenthesizeConciseBody(c),l}function An(e,n,r,a,i,o,s){return e.modifiers!==n||e.typeParameters!==r||e.parameters!==a||e.type!==i||e.equalsGreaterThanToken!==o||e.body!==s?t(Cn(n,r,a,i,o,s),e):e}function Dn(t){var r=n(198);return r.expression=e.parenthesizePrefixOperand(t),r}function Ln(e,n){return e.expression!==n?t(Dn(n),e):e}function kn(t){var r=n(199);return r.expression=e.parenthesizePrefixOperand(t),r}function Nn(e,n){return e.expression!==n?t(kn(n),e):e}function In(t){var r=n(200);return r.expression=e.parenthesizePrefixOperand(t),r}function wn(e,n){return e.expression!==n?t(In(n),e):e}function Mn(t){var r=n(201);return r.expression=e.parenthesizePrefixOperand(t),r}function On(e,n){return e.expression!==n?t(Mn(n),e):e}function Pn(t,r){var a=n(202);return a.operator=t,a.operand=e.parenthesizePrefixOperand(r),a}function Rn(e,n){return e.operand!==n?t(Pn(e.operator,n),e):e}function Fn(t,r){var a=n(203);return a.operand=e.parenthesizePostfixOperand(t),a.operator=r,a}function Gn(e,n){return e.operand!==n?t(Fn(n,e.operator),e):e}function Vn(t,r,a){var i=n(204),o=ai(r),s=o.kind;return i.left=e.parenthesizeBinaryOperand(s,t,!0,void 0),i.operatorToken=o,i.right=e.parenthesizeBinaryOperand(s,a,!1,i.left),i}function Bn(e,n,r,a){return e.left!==n||e.right!==r?t(Vn(n,a||e.operatorToken,r),e):e}function Kn(t,r,a,i,o){var s=n(205);return s.condition=e.parenthesizeForConditionalHead(t),s.questionToken=o?r:E(56),s.whenTrue=e.parenthesizeSubexpressionOfConditionalExpression(o?a:r),s.colonToken=o?i:E(57),s.whenFalse=e.parenthesizeSubexpressionOfConditionalExpression(o||a),s}function jn(e,n,r,a,i,o){return e.condition!==n||e.questionToken!==r||e.whenTrue!==a||e.colonToken!==i||e.whenFalse!==o?t(Kn(n,r,a,i,o),e):e}function Un(e,t){var a=n(206);return a.head=e,a.templateSpans=r(t),a}function Hn(e,n,r){return e.head!==n||e.templateSpans!==r?t(Un(n,r),e):e}function Wn(e){var t=n(15);return t.text=e,t}function qn(e){var t=n(16);return t.text=e,t}function zn(e){var t=n(17);return t.text=e,t}function Jn(e){var t=n(14);return t.text=e,t}function Xn(e,t){var r=n(207);return r.asteriskToken=e&&40===e.kind?e:void 0,r.expression=e&&40!==e.kind?e:t,r}function Yn(e,n,r){return e.expression!==r||e.asteriskToken!==n?t(Xn(n,r),e):e}function Qn(t){var r=n(208);return r.expression=e.parenthesizeExpressionForList(t),r}function Zn(e,n){return e.expression!==n?t(Qn(n),e):e}function $n(e,t,a,i,o){var s=n(209);return s.decorators=void 0,s.modifiers=ri(e),s.name=ni(t),s.typeParameters=ri(a),s.heritageClauses=ri(i),s.members=r(o),s}function et(e,n,r,a,i,o){return e.modifiers!==n||e.name!==r||e.typeParameters!==a||e.heritageClauses!==i||e.members!==o?t($n(n,r,a,i,o),e):e}function nt(){return n(210)}function tt(t,r){var a=n(211);return a.expression=e.parenthesizeForAccess(r),a.typeArguments=ri(t),a}function rt(e,n,r){return e.typeArguments!==n||e.expression!==r?t(tt(n,r),e):e}function at(e,t){var r=n(212);return r.expression=e,r.type=t,r}function it(e,n,r){return e.expression!==n||e.type!==r?t(at(n,r),e):e}function ot(t){var r=n(213);return r.expression=e.parenthesizeForAccess(t),r}function st(e,n){return e.expression!==n?t(ot(n),e):e}function ct(e,t){var r=n(214);return r.keywordToken=e,r.name=t,r}function lt(e,n){return e.name!==n?t(ct(e.keywordToken,n),e):e}function ut(e,t){var r=n(216);return r.expression=e,r.literal=t,r}function dt(e,n,r){return e.expression!==n||e.literal!==r?t(ut(n,r),e):e}function pt(){return n(217)}function mt(e,t){var a=n(218);return a.statements=r(e),t&&(a.multiLine=t),a}function ft(e,n){return e.statements!==n?t(mt(n,e.multiLine),e):e}function _t(t,r){var a=n(219);return a.decorators=void 0,a.modifiers=ri(t),a.declarationList=e.isArray(r)?Qt(r):r,a}function gt(e,n,r){return e.modifiers!==n||e.declarationList!==r?t(_t(n,r),e):e}function vt(){return n(220)}function yt(t){var r=n(221);return r.expression=e.parenthesizeExpressionForExpressionStatement(t),r}function ht(e,n){return e.expression!==n?t(yt(n),e):e}function bt(e,t,r){var a=n(222);return a.expression=e,a.thenStatement=t,a.elseStatement=r,a}function Et(e,n,r,a){return e.expression!==n||e.thenStatement!==r||e.elseStatement!==a?t(bt(n,r,a),e):e}function Tt(e,t){var r=n(223);return r.statement=e,r.expression=t,r}function St(e,n,r){return e.statement!==n||e.expression!==r?t(Tt(n,r),e):e}function xt(e,t){var r=n(224);return r.expression=e,r.statement=t,r}function Ct(e,n,r){return e.expression!==n||e.statement!==r?t(xt(n,r),e):e}function At(e,t,r,a){var i=n(225);return i.initializer=e,i.condition=t,i.incrementor=r,i.statement=a,i}function Dt(e,n,r,a,i){return e.initializer!==n||e.condition!==r||e.incrementor!==a||e.statement!==i?t(At(n,r,a,i),e):e}function Lt(e,t,r){var a=n(226);return a.initializer=e,a.expression=t,a.statement=r,a}function kt(e,n,r,a){return e.initializer!==n||e.expression!==r||e.statement!==a?t(Lt(n,r,a),e):e}function Nt(e,t,r,a){var i=n(227);return i.awaitModifier=e,i.initializer=t,i.expression=r,i.statement=a,i}function It(e,n,r,a,i){return e.awaitModifier!==n||e.initializer!==r||e.expression!==a||e.statement!==i?t(Nt(n,r,a,i),e):e}function wt(e){var t=n(228);return t.label=ni(e),t}function Mt(e,n){return e.label!==n?t(wt(n),e):e}function Ot(e){var t=n(229);return t.label=ni(e),t}function Pt(e,n){return e.label!==n?t(Ot(n),e):e}function Rt(e){var t=n(230);return t.expression=e,t}function Ft(e,n){return e.expression!==n?t(Rt(n),e):e}function Gt(e,t){var r=n(231);return r.expression=e,r.statement=t,r}function Vt(e,n,r){return e.expression!==n||e.statement!==r?t(Gt(n,r),e):e}function Bt(t,r){var a=n(232);return a.expression=e.parenthesizeExpressionForList(t),a.caseBlock=r,a}function Kt(e,n,r){return e.expression!==n||e.caseBlock!==r?t(Bt(n,r),e):e}function jt(e,t){var r=n(233);return r.label=ni(e),r.statement=t,r}function Ut(e,n,r){return e.label!==n||e.statement!==r?t(jt(n,r),e):e}function Ht(e){var t=n(234);return t.expression=e,t}function Wt(e,n){return e.expression!==n?t(Ht(n),e):e}function qt(e,t,r){var a=n(235);return a.tryBlock=e,a.catchClause=t,a.finallyBlock=r,a}function zt(e,n,r,a){return e.tryBlock!==n||e.catchClause!==r||e.finallyBlock!==a?t(qt(n,r,a),e):e}function Jt(){return n(236)}function Xt(t,r,a){var i=n(237);return i.name=ni(t),i.type=r,i.initializer=void 0!==a?e.parenthesizeExpressionForList(a):void 0,i}function Yt(e,n,r,a){return e.name!==n||e.type!==r||e.initializer!==a?t(Xt(n,r,a),e):e}function Qt(e,t){void 0===t&&(t=0);var a=n(238);return a.flags|=3&t,a.declarations=r(e),a}function Zt(e,n){return e.declarations!==n?t(Qt(n,e.flags),e):e}function $t(e,t,a,i,o,s,c,l){var u=n(239);return u.decorators=ri(e),u.modifiers=ri(t),u.asteriskToken=a,u.name=ni(i),u.typeParameters=ri(o),u.parameters=r(s),u.type=c,u.body=l,u}function er(e,n,r,a,i,o,s,c,l){return e.decorators!==n||e.modifiers!==r||e.asteriskToken!==a||e.name!==i||e.typeParameters!==o||e.parameters!==s||e.type!==c||e.body!==l?t($t(n,r,a,i,o,s,c,l),e):e}function nr(e,t,a,i,o,s){var c=n(240);return c.decorators=ri(e),c.modifiers=ri(t),c.name=ni(a),c.typeParameters=ri(i),c.heritageClauses=ri(o),c.members=r(s),c}function tr(e,n,r,a,i,o,s){return e.decorators!==n||e.modifiers!==r||e.name!==a||e.typeParameters!==i||e.heritageClauses!==o||e.members!==s?t(nr(n,r,a,i,o,s),e):e}function rr(e,t,a,i,o,s){var c=n(241);return c.decorators=ri(e),c.modifiers=ri(t),c.name=ni(a),c.typeParameters=ri(i),c.heritageClauses=ri(o),c.members=r(s),c}function ar(e,n,r,a,i,o,s){return e.decorators!==n||e.modifiers!==r||e.name!==a||e.typeParameters!==i||e.heritageClauses!==o||e.members!==s?t(rr(n,r,a,i,o,s),e):e}function ir(e,t,r,a,i){var o=n(242);return o.decorators=ri(e),o.modifiers=ri(t),o.name=ni(r),o.typeParameters=ri(a),o.type=i,o}function or(e,n,r,a,i,o){return e.decorators!==n||e.modifiers!==r||e.name!==a||e.typeParameters!==i||e.type!==o?t(ir(n,r,a,i,o),e):e}function sr(e,t,a,i){var o=n(243);return o.decorators=ri(e),o.modifiers=ri(t),o.name=ni(a),o.members=r(i),o}function cr(e,n,r,a,i){return e.decorators!==n||e.modifiers!==r||e.name!==a||e.members!==i?t(sr(n,r,a,i),e):e}function lr(e,t,r,a,i){void 0===i&&(i=0);var o=n(244);return o.flags|=532&i,o.decorators=ri(e),o.modifiers=ri(t),o.name=r,o.body=a,o}function ur(e,n,r,a,i){return e.decorators!==n||e.modifiers!==r||e.name!==a||e.body!==i?t(lr(n,r,a,i,e.flags),e):e}function dr(e){var t=n(245);return t.statements=r(e),t}function pr(e,n){return e.statements!==n?t(dr(n),e):e}function mr(e){var t=n(246);return t.clauses=r(e),t}function fr(e,n){return e.clauses!==n?t(mr(n),e):e}function _r(e){var t=n(247);return t.name=ni(e),t}function gr(e,n){return e.name!==n?t(_r(n),e):e}function vr(e,t,r,a){var i=n(248);return i.decorators=ri(e),i.modifiers=ri(t),i.name=ni(r),i.moduleReference=a,i}function yr(e,n,r,a,i){return e.decorators!==n||e.modifiers!==r||e.name!==a||e.moduleReference!==i?t(vr(n,r,a,i),e):e}function hr(e,t,r,a){var i=n(249);return i.decorators=ri(e),i.modifiers=ri(t),i.importClause=r,i.moduleSpecifier=a,i}function br(e,n,r,a,i){return e.decorators!==n||e.modifiers!==r||e.importClause!==a||e.moduleSpecifier!==i?t(hr(n,r,a,i),e):e}function Er(e,t){var r=n(250);return r.name=e,r.namedBindings=t,r}function Tr(e,n,r){return e.name!==n||e.namedBindings!==r?t(Er(n,r),e):e}function Sr(e){var t=n(251);return t.name=e,t}function xr(e,n){return e.name!==n?t(Sr(n),e):e}function Cr(e){var t=n(252);return t.elements=r(e),t}function Ar(e,n){return e.elements!==n?t(Cr(n),e):e}function Dr(e,t){var r=n(253);return r.propertyName=e,r.name=t,r}function Lr(e,n,r){return e.propertyName!==n||e.name!==r?t(Dr(n,r),e):e}function kr(t,r,a,i){var o=n(254);return o.decorators=ri(t),o.modifiers=ri(r),o.isExportEquals=a,o.expression=a?e.parenthesizeBinaryOperand(59,i,!1,void 0):e.parenthesizeDefaultExpression(i),o}function Nr(e,n,r,a){return e.decorators!==n||e.modifiers!==r||e.expression!==a?t(kr(n,r,e.isExportEquals,a),e):e}function Ir(e,t,r,a){var i=n(255);return i.decorators=ri(e),i.modifiers=ri(t),i.exportClause=r,i.moduleSpecifier=a,i}function wr(e,n,r,a,i){return e.decorators!==n||e.modifiers!==r||e.exportClause!==a||e.moduleSpecifier!==i?t(Ir(n,r,a,i),e):e}function Mr(e){var t=n(256);return t.elements=r(e),t}function Or(e,n){return e.elements!==n?t(Mr(n),e):e}function Pr(e,t){var r=n(257);return r.propertyName=ni(e),r.name=ni(t),r}function Rr(e,n,r){return e.propertyName!==n||e.name!==r?t(Pr(n,r),e):e}function Fr(e){var t=n(259);return t.expression=e,t}function Gr(e,n){return e.expression!==n?t(Fr(n),e):e}function Vr(e){var t=n(283);return t.type=e,t}function Br(e,n){var t=Hr(302,"type");return t.typeExpression=e,t.comment=n,t}function Kr(e,n){var t=Hr(300,"returns");return t.typeExpression=e,t.comment=n,t}function jr(e,n,t,r){var a=Hr(299,"param");return a.typeExpression=t,a.name=e,a.isBracketed=n,a.comment=r,a}function Ur(e,t){var r=n(291);return r.comment=e,r.tags=t,r}function Hr(e,t){var r=n(e);return r.tagName=d(t),r}function Wr(e,t,a){var i=n(260);return i.openingElement=e,i.children=r(t),i.closingElement=a,i}function qr(e,n,r,a){return e.openingElement!==n||e.children!==r||e.closingElement!==a?t(Wr(n,r,a),e):e}function zr(e,t,r){var a=n(261);return a.tagName=e,a.typeArguments=ri(t),a.attributes=r,a}function Jr(e,n,r,a){return e.tagName!==n||e.typeArguments!==r||e.attributes!==a?t(zr(n,r,a),e):e}function Xr(e,t,r){var a=n(262);return a.tagName=e,a.typeArguments=ri(t),a.attributes=r,a}function Yr(e,n,r,a){return e.tagName!==n||e.typeArguments!==r||e.attributes!==a?t(Xr(n,r,a),e):e}function Qr(e){var t=n(263);return t.tagName=e,t}function Zr(e,n){return e.tagName!==n?t(Qr(n),e):e}function $r(e,t,a){var i=n(264);return i.openingFragment=e,i.children=r(t),i.closingFragment=a,i}function ea(e,n,r,a){return e.openingFragment!==n||e.children!==r||e.closingFragment!==a?t($r(n,r,a),e):e}function na(e,t){var r=n(267);return r.name=e,r.initializer=t,r}function ta(e,n,r){return e.name!==n||e.initializer!==r?t(na(n,r),e):e}function ra(e){var t=n(268);return t.properties=r(e),t}function aa(e,n){return e.properties!==n?t(ra(n),e):e}function ia(e){var t=n(269);return t.expression=e,t}function oa(e,n){return e.expression!==n?t(ia(n),e):e}function sa(e,t){var r=n(270);return r.dotDotDotToken=e,r.expression=t,r}function ca(e,n){return e.expression!==n?t(sa(e.dotDotDotToken,n),e):e}function la(t,a){var i=n(271);return i.expression=e.parenthesizeExpressionForList(t),i.statements=r(a),i}function ua(e,n,r){return e.expression!==n||e.statements!==r?t(la(n,r),e):e}function da(e){var t=n(272);return t.statements=r(e),t}function pa(e,n){return e.statements!==n?t(da(n),e):e}function ma(e,t){var a=n(273);return a.token=e,a.types=r(t),a}function fa(e,n){return e.types!==n?t(ma(e.token,n),e):e}function _a(t,r){var a=n(274);return a.variableDeclaration=e.isString(t)?Xt(t):t,a.block=r,a}function ga(e,n,r){return e.variableDeclaration!==n||e.block!==r?t(_a(n,r),e):e}function va(t,r){var a=n(275);return a.name=ni(t),a.questionToken=void 0,a.initializer=e.parenthesizeExpressionForList(r),a}function ya(e,n,r){return e.name!==n||e.initializer!==r?t(va(n,r),e):e}function ha(t,r){var a=n(276);return a.name=ni(t),a.objectAssignmentInitializer=void 0!==r?e.parenthesizeExpressionForList(r):void 0,a}function ba(e,n,r){return e.name!==n||e.objectAssignmentInitializer!==r?t(ha(n,r),e):e}function Ea(t){var r=n(277);return r.expression=void 0!==t?e.parenthesizeExpressionForList(t):void 0,r}function Ta(e,n){return e.expression!==n?t(Ea(n),e):e}function Sa(t,r){var a=n(278);return a.name=ni(t),a.initializer=r&&e.parenthesizeExpressionForList(r),a}function xa(e,n,r){return e.name!==n||e.initializer!==r?t(Sa(n,r),e):e}function Ca(e,a,i,o,s,c,l){if(e.statements!==a||void 0!==i&&e.isDeclarationFile!==i||void 0!==o&&e.referencedFiles!==o||void 0!==s&&e.typeReferenceDirectives!==s||void 0!==l&&e.libReferenceDirectives!==l||void 0!==c&&e.hasNoDefaultLib!==c){var u=n(279);return u.flags|=e.flags,u.statements=r(a),u.endOfFileToken=e.endOfFileToken,u.fileName=e.fileName,u.path=e.path,u.text=e.text,u.isDeclarationFile=void 0===i?e.isDeclarationFile:i,u.referencedFiles=void 0===o?e.referencedFiles:o,u.typeReferenceDirectives=void 0===s?e.typeReferenceDirectives:s,u.hasNoDefaultLib=void 0===c?e.hasNoDefaultLib:c,u.libReferenceDirectives=void 0===l?e.libReferenceDirectives:l,void 0!==e.amdDependencies&&(u.amdDependencies=e.amdDependencies),void 0!==e.moduleName&&(u.moduleName=e.moduleName),void 0!==e.languageVariant&&(u.languageVariant=e.languageVariant),void 0!==e.renamedDependencies&&(u.renamedDependencies=e.renamedDependencies),void 0!==e.languageVersion&&(u.languageVersion=e.languageVersion),void 0!==e.scriptKind&&(u.scriptKind=e.scriptKind),void 0!==e.externalModuleIndicator&&(u.externalModuleIndicator=e.externalModuleIndicator),void 0!==e.commonJsModuleIndicator&&(u.commonJsModuleIndicator=e.commonJsModuleIndicator),void 0!==e.identifiers&&(u.identifiers=e.identifiers),void 0!==e.nodeCount&&(u.nodeCount=e.nodeCount),void 0!==e.identifierCount&&(u.identifierCount=e.identifierCount),void 0!==e.symbolCount&&(u.symbolCount=e.symbolCount),void 0!==e.parseDiagnostics&&(u.parseDiagnostics=e.parseDiagnostics),void 0!==e.bindDiagnostics&&(u.bindDiagnostics=e.bindDiagnostics),void 0!==e.bindSuggestionDiagnostics&&(u.bindSuggestionDiagnostics=e.bindSuggestionDiagnostics),void 0!==e.lineMap&&(u.lineMap=e.lineMap),void 0!==e.classifiableNames&&(u.classifiableNames=e.classifiableNames),void 0!==e.resolvedModules&&(u.resolvedModules=e.resolvedModules),void 0!==e.resolvedTypeReferenceDirectiveNames&&(u.resolvedTypeReferenceDirectiveNames=e.resolvedTypeReferenceDirectiveNames),void 0!==e.imports&&(u.imports=e.imports),void 0!==e.moduleAugmentations&&(u.moduleAugmentations=e.moduleAugmentations),void 0!==e.pragmas&&(u.pragmas=e.pragmas),void 0!==e.localJsxFactory&&(u.localJsxFactory=e.localJsxFactory),void 0!==e.localJsxNamespace&&(u.localJsxNamespace=e.localJsxNamespace),t(u,e)}return e}function Aa(e){var n=a(e);return n.pos=e.pos,n.end=e.end,n.parent=e.parent,n}function Da(e){var t=n(307);return t.original=e,si(t,e),t}function La(e){var t=n(311);return t.emitNode={},t.original=e,t}function ka(e){var t=n(310);return t.emitNode={},t.original=e,t}function Na(e,t){var r=n(308);return r.expression=e,r.original=t,si(r,t),r}function Ia(e,n){return e.expression!==n?t(Na(n,e.original),e):e}function wa(n){if(e.nodeIsSynthesized(n)&&!e.isParseTreeNode(n)&&!n.original&&!n.emitNode&&!n.id){if(309===n.kind)return n.elements;if(e.isBinaryExpression(n)&&27===n.operatorToken.kind)return[n.left,n.right]}return n}function Ma(t){var a=n(309);return a.elements=r(e.sameFlatMap(t,wa)),a}function Oa(e,n){return e.elements!==n?t(Ma(n),e):e}function Pa(n,t){void 0===t&&(t=e.emptyArray);var r=e.createNode(280);return r.prepends=t,r.sourceFiles=n,r}function Ra(n,t,r){var a=e.createNode(281);return e.isString(n)?(a.text=n,a.sourceMapPath=t,a.sourceMapText=r):(e.Debug.assert("js"===t||"dts"===t),a.fileName="js"===t?n.javascriptPath:n.declarationPath,a.sourceMapPath="js"===t?n.javascriptMapPath:n.declarationMapPath,Object.defineProperties(a,{text:{get:function(){return"js"===t?n.javascriptText:n.declarationText}},sourceMapText:{get:function(){return"js"===t?n.javascriptMapText:n.declarationMapText}}})),a}function Fa(n,t,r,a,i,o){var s=e.createNode(282);if(e.isString(n))s.javascriptText=n,s.javascriptMapPath=r,s.javascriptMapText=a,s.declarationText=t,s.declarationMapPath=i,s.declarationMapText=o;else{var c=e.createMap(),l=function(e){if(void 0!==e){var t=c.get(e);return void 0===t&&(t=n(e),c.set(e,void 0!==t&&t)),!1!==t?t:void 0}},u=function(e){var n=l(e);return void 0!==n?n:"/* Input file "+e+" was missing */\r\n"};s.javascriptPath=t,s.javascriptMapPath=r,s.declarationPath=e.Debug.assertDefined(a),s.declarationMapPath=i,Object.defineProperties(s,{javascriptText:{get:function(){return u(t)}},javascriptMapText:{get:function(){return l(r)}},declarationText:{get:function(){return u(e.Debug.assertDefined(a))}},declarationMapText:{get:function(){return l(i)}}})}return s}function Ga(n,t,r){return void 0===r&&(r=e.emptyArray),n.sourceFiles!==t||n.prepends!==r?Pa(t,r):n}function Va(e,n,t){return mn(Sn(void 0,void 0,void 0,void 0,n?[n]:[],void 0,mt(e,!0)),void 0,t?[t]:[])}function Ba(e,n,t){return mn(Cn(void 0,void 0,n?[n]:[],void 0,void 0,mt(e,!0)),void 0,t?[t]:[])}function Ka(e,n){return Vn(e,27,n)}function ja(e,n){return Vn(e,28,n)}function Ua(e,n){return Vn(e,59,n)}function Ha(e,n){return Vn(e,35,n)}function Wa(e,n){return Vn(e,36,n)}function qa(e,n){return Vn(e,38,n)}function za(e,n){return Vn(e,39,n)}function Ja(e){return Fn(e,44)}function Xa(e,n){return Vn(e,54,n)}function Ya(e,n){return Vn(e,55,n)}function Qa(e){return Pn(52,e)}function Za(){return In(i(0))}function $a(e){return kr(void 0,void 0,!1,e)}function ei(e){return Ir(void 0,void 0,Mr([Pr(void 0,e)]))}function ni(n){return e.isString(n)?d(n):n}function ti(n){return e.isString(n)||"number"===typeof n?i(n):n}function ri(e){return e?r(e):void 0}function ai(e){return"number"===typeof e?E(e):e}function ii(n){n=e.getSourceFileOfNode(e.getParseTreeNode(n));var t=n&&n.emitNode,r=t&&t.annotatedNodes;if(r)for(var a=0,i=r;a<i.length;a++){var o=i[a];o.emitNode=void 0}}function oi(n){if(!n.emitNode){if(e.isParseTreeNode(n)){if(279===n.kind)return n.emitNode={annotatedNodes:[n]};var t=e.getSourceFileOfNode(n);oi(t).annotatedNodes.push(n)}n.emitNode={}}return n.emitNode}function si(e,n){return n&&(e.pos=n.pos,e.end=n.end),e}function ci(e,n){return oi(e).flags=n,e}function li(e,n){var t=oi(e);return t.flags=t.flags|n,e}function ui(e){var n=e.emitNode;return n&&n.sourceMapRange||e}function di(e,n){return oi(e).sourceMapRange=n,e}function pi(n,t,r){return new(m||(m=e.objectAllocator.getSourceMapSourceConstructor()))(n,t,r)}function mi(e,n){var t=e.emitNode,r=t&&t.tokenSourceMapRanges;return r&&r[n]}function fi(e,n,t){var r=oi(e),a=r.tokenSourceMapRanges||(r.tokenSourceMapRanges=[]);return a[n]=t,e}function _i(e){var n=e.emitNode;return n&&n.startsOnNewLine}function gi(e,n){return oi(e).startsOnNewLine=n,e}function vi(e){var n=e.emitNode;return n&&n.commentRange||e}function yi(e,n){return oi(e).commentRange=n,e}function hi(e){var n=e.emitNode;return n&&n.leadingComments}function bi(e,n){return oi(e).leadingComments=n,e}function Ei(n,t,r,a){return bi(n,e.append(hi(n),{kind:t,pos:-1,end:-1,hasTrailingNewLine:a,text:r}))}function Ti(e){var n=e.emitNode;return n&&n.trailingComments}function Si(e,n){return oi(e).trailingComments=n,e}function xi(n,t,r,a){return Si(n,e.append(Ti(n),{kind:t,pos:-1,end:-1,hasTrailingNewLine:a,text:r}))}function Ci(e,n){bi(e,hi(n)),Si(e,Ti(n));var t=oi(n);return t.leadingComments=void 0,t.trailingComments=void 0,e}function Ai(e){var n=e.emitNode;return n&&n.constantValue}function Di(e,n){var t=oi(e);return t.constantValue=n,e}function Li(n,t){var r=oi(n);return r.helpers=e.append(r.helpers,t),n}function ki(n,t){if(e.some(t))for(var r=oi(n),a=0,i=t;a<i.length;a++){var o=i[a];r.helpers=e.appendIfUnique(r.helpers,o)}return n}function Ni(n,t){var r=n.emitNode;if(r){var a=r.helpers;if(a)return e.orderedRemoveItem(a,t)}return!1}function Ii(e){var n=e.emitNode;return n&&n.helpers}function wi(n,t,r){var a=n.emitNode,i=a&&a.helpers;if(e.some(i)){for(var o=oi(t),s=0,c=0;c<i.length;c++){var l=i[c];r(l)?(s++,o.helpers=e.appendIfUnique(o.helpers,l)):s>0&&(i[c-s]=l)}s>0&&(i.length-=s)}}function Mi(n,t){return n===t?0:n.priority===t.priority?0:void 0===n.priority?1:void 0===t.priority?-1:e.compareValues(n.priority,t.priority)}function Oi(e,n){if(e.original=n,n){var t=n.emitNode;t&&(e.emitNode=Pi(t,e.emitNode))}return e}function Pi(n,t){var r=n.flags,a=n.leadingComments,i=n.trailingComments,o=n.commentRange,s=n.sourceMapRange,c=n.tokenSourceMapRanges,l=n.constantValue,u=n.helpers,d=n.startsOnNewLine;return t||(t={}),a&&(t.leadingComments=e.addRange(a.slice(),t.leadingComments)),i&&(t.trailingComments=e.addRange(i.slice(),t.trailingComments)),r&&(t.flags=r),o&&(t.commentRange=o),s&&(t.sourceMapRange=s),c&&(t.tokenSourceMapRanges=Ri(c,t.tokenSourceMapRanges)),void 0!==l&&(t.constantValue=l),u&&(t.helpers=e.addRange(t.helpers,u)),void 0!==d&&(t.startsOnNewLine=d),t}function Ri(e,n){for(var t in n||(n=[]),e)n[t]=e[t];return n}e.createTempVariable=_,e.createLoopVariable=g,e.createUniqueName=v,e.createOptimisticUniqueName=y,e.createFileLevelUniqueName=h,e.getGeneratedNameForNode=b,e.createToken=E,e.createSuper=T,e.createThis=S,e.createNull=x,e.createTrue=C,e.createFalse=A,e.createModifier=D,e.createModifiersFromModifierFlags=L,e.createQualifiedName=k,e.updateQualifiedName=N,e.createComputedPropertyName=w,e.updateComputedPropertyName=M,e.createTypeParameterDeclaration=O,e.updateTypeParameterDeclaration=P,e.createParameter=R,e.updateParameter=F,e.createDecorator=G,e.updateDecorator=V,e.createPropertySignature=B,e.updatePropertySignature=K,e.createProperty=j,e.updateProperty=U,e.createMethodSignature=H,e.updateMethodSignature=W,e.createMethod=q,e.updateMethod=z,e.createConstructor=J,e.updateConstructor=X,e.createGetAccessor=Y,e.updateGetAccessor=Q,e.createSetAccessor=Z,e.updateSetAccessor=$,e.createCallSignature=ee,e.updateCallSignature=ne,e.createConstructSignature=te,e.updateConstructSignature=re,e.createIndexSignature=ae,e.updateIndexSignature=ie,e.createSignatureDeclaration=oe,e.createKeywordTypeNode=ce,e.createTypePredicateNode=le,e.updateTypePredicateNode=ue,e.createTypeReferenceNode=de,e.updateTypeReferenceNode=pe,e.createFunctionTypeNode=me,e.updateFunctionTypeNode=fe,e.createConstructorTypeNode=_e,e.updateConstructorTypeNode=ge,e.createTypeQueryNode=ve,e.updateTypeQueryNode=ye,e.createTypeLiteralNode=he,e.updateTypeLiteralNode=be,e.createArrayTypeNode=Ee,e.updateArrayTypeNode=Te,e.createTupleTypeNode=Se,e.updateTupleTypeNode=xe,e.createOptionalTypeNode=Ce,e.updateOptionalTypeNode=Ae,e.createRestTypeNode=De,e.updateRestTypeNode=Le,e.createUnionTypeNode=ke,e.updateUnionTypeNode=Ne,e.createIntersectionTypeNode=Ie,e.updateIntersectionTypeNode=we,e.createUnionOrIntersectionTypeNode=Me,e.createConditionalTypeNode=Pe,e.updateConditionalTypeNode=Re,e.createInferTypeNode=Fe,e.updateInferTypeNode=Ge,e.createImportTypeNode=Ve,e.updateImportTypeNode=Be,e.createParenthesizedType=Ke,e.updateParenthesizedType=je,e.createThisTypeNode=Ue,e.createTypeOperatorNode=He,e.updateTypeOperatorNode=We,e.createIndexedAccessTypeNode=qe,e.updateIndexedAccessTypeNode=ze,e.createMappedTypeNode=Je,e.updateMappedTypeNode=Xe,e.createLiteralTypeNode=Ye,e.updateLiteralTypeNode=Qe,e.createObjectBindingPattern=Ze,e.updateObjectBindingPattern=$e,e.createArrayBindingPattern=en,e.updateArrayBindingPattern=nn,e.createBindingElement=tn,e.updateBindingElement=rn,e.createArrayLiteral=an,e.updateArrayLiteral=on,e.createObjectLiteral=sn,e.updateObjectLiteral=cn,e.createPropertyAccess=ln,e.updatePropertyAccess=un,e.createElementAccess=dn,e.updateElementAccess=pn,e.createCall=mn,e.updateCall=fn,e.createNew=_n,e.updateNew=gn,e.createTaggedTemplate=vn,e.updateTaggedTemplate=yn,e.createTypeAssertion=hn,e.updateTypeAssertion=bn,e.createParen=En,e.updateParen=Tn,e.createFunctionExpression=Sn,e.updateFunctionExpression=xn,e.createArrowFunction=Cn,e.updateArrowFunction=An,e.createDelete=Dn,e.updateDelete=Ln,e.createTypeOf=kn,e.updateTypeOf=Nn,e.createVoid=In,e.updateVoid=wn,e.createAwait=Mn,e.updateAwait=On,e.createPrefix=Pn,e.updatePrefix=Rn,e.createPostfix=Fn,e.updatePostfix=Gn,e.createBinary=Vn,e.updateBinary=Bn,e.createConditional=Kn,e.updateConditional=jn,e.createTemplateExpression=Un,e.updateTemplateExpression=Hn,e.createTemplateHead=Wn,e.createTemplateMiddle=qn,e.createTemplateTail=zn,e.createNoSubstitutionTemplateLiteral=Jn,e.createYield=Xn,e.updateYield=Yn,e.createSpread=Qn,e.updateSpread=Zn,e.createClassExpression=$n,e.updateClassExpression=et,e.createOmittedExpression=nt,e.createExpressionWithTypeArguments=tt,e.updateExpressionWithTypeArguments=rt,e.createAsExpression=at,e.updateAsExpression=it,e.createNonNullExpression=ot,e.updateNonNullExpression=st,e.createMetaProperty=ct,e.updateMetaProperty=lt,e.createTemplateSpan=ut,e.updateTemplateSpan=dt,e.createSemicolonClassElement=pt,e.createBlock=mt,e.updateBlock=ft,e.createVariableStatement=_t,e.updateVariableStatement=gt,e.createEmptyStatement=vt,e.createExpressionStatement=yt,e.updateExpressionStatement=ht,e.createStatement=yt,e.updateStatement=ht,e.createIf=bt,e.updateIf=Et,e.createDo=Tt,e.updateDo=St,e.createWhile=xt,e.updateWhile=Ct,e.createFor=At,e.updateFor=Dt,e.createForIn=Lt,e.updateForIn=kt,e.createForOf=Nt,e.updateForOf=It,e.createContinue=wt,e.updateContinue=Mt,e.createBreak=Ot,e.updateBreak=Pt,e.createReturn=Rt,e.updateReturn=Ft,e.createWith=Gt,e.updateWith=Vt,e.createSwitch=Bt,e.updateSwitch=Kt,e.createLabel=jt,e.updateLabel=Ut,e.createThrow=Ht,e.updateThrow=Wt,e.createTry=qt,e.updateTry=zt,e.createDebuggerStatement=Jt,e.createVariableDeclaration=Xt,e.updateVariableDeclaration=Yt,e.createVariableDeclarationList=Qt,e.updateVariableDeclarationList=Zt,e.createFunctionDeclaration=$t,e.updateFunctionDeclaration=er,e.createClassDeclaration=nr,e.updateClassDeclaration=tr,e.createInterfaceDeclaration=rr,e.updateInterfaceDeclaration=ar,e.createTypeAliasDeclaration=ir,e.updateTypeAliasDeclaration=or,e.createEnumDeclaration=sr,e.updateEnumDeclaration=cr,e.createModuleDeclaration=lr,e.updateModuleDeclaration=ur,e.createModuleBlock=dr,e.updateModuleBlock=pr,e.createCaseBlock=mr,e.updateCaseBlock=fr,e.createNamespaceExportDeclaration=_r,e.updateNamespaceExportDeclaration=gr,e.createImportEqualsDeclaration=vr,e.updateImportEqualsDeclaration=yr,e.createImportDeclaration=hr,e.updateImportDeclaration=br,e.createImportClause=Er,e.updateImportClause=Tr,e.createNamespaceImport=Sr,e.updateNamespaceImport=xr,e.createNamedImports=Cr,e.updateNamedImports=Ar,e.createImportSpecifier=Dr,e.updateImportSpecifier=Lr,e.createExportAssignment=kr,e.updateExportAssignment=Nr,e.createExportDeclaration=Ir,e.updateExportDeclaration=wr,e.createNamedExports=Mr,e.updateNamedExports=Or,e.createExportSpecifier=Pr,e.updateExportSpecifier=Rr,e.createExternalModuleReference=Fr,e.updateExternalModuleReference=Gr,e.createJSDocTypeExpression=Vr,e.createJSDocTypeTag=Br,e.createJSDocReturnTag=Kr,e.createJSDocParamTag=jr,e.createJSDocComment=Ur,e.createJsxElement=Wr,e.updateJsxElement=qr,e.createJsxSelfClosingElement=zr,e.updateJsxSelfClosingElement=Jr,e.createJsxOpeningElement=Xr,e.updateJsxOpeningElement=Yr,e.createJsxClosingElement=Qr,e.updateJsxClosingElement=Zr,e.createJsxFragment=$r,e.updateJsxFragment=ea,e.createJsxAttribute=na,e.updateJsxAttribute=ta,e.createJsxAttributes=ra,e.updateJsxAttributes=aa,e.createJsxSpreadAttribute=ia,e.updateJsxSpreadAttribute=oa,e.createJsxExpression=sa,e.updateJsxExpression=ca,e.createCaseClause=la,e.updateCaseClause=ua,e.createDefaultClause=da,e.updateDefaultClause=pa,e.createHeritageClause=ma,e.updateHeritageClause=fa,e.createCatchClause=_a,e.updateCatchClause=ga,e.createPropertyAssignment=va,e.updatePropertyAssignment=ya,e.createShorthandPropertyAssignment=ha,e.updateShorthandPropertyAssignment=ba,e.createSpreadAssignment=Ea,e.updateSpreadAssignment=Ta,e.createEnumMember=Sa,e.updateEnumMember=xa,e.updateSourceFileNode=Ca,e.getMutableClone=Aa,e.createNotEmittedStatement=Da,e.createEndOfDeclarationMarker=La,e.createMergeDeclarationMarker=ka,e.createPartiallyEmittedExpression=Na,e.updatePartiallyEmittedExpression=Ia,e.createCommaList=Ma,e.updateCommaList=Oa,e.createBundle=Pa,e.createUnparsedSourceFile=Ra,e.createInputFiles=Fa,e.updateBundle=Ga,e.createImmediatelyInvokedFunctionExpression=Va,e.createImmediatelyInvokedArrowFunction=Ba,e.createComma=Ka,e.createLessThan=ja,e.createAssignment=Ua,e.createStrictEquality=Ha,e.createStrictInequality=Wa,e.createAdd=qa,e.createSubtract=za,e.createPostfixIncrement=Ja,e.createLogicalAnd=Xa,e.createLogicalOr=Ya,e.createLogicalNot=Qa,e.createVoidZero=Za,e.createExportDefault=$a,e.createExternalModuleExport=ei,e.disposeEmitNodes=ii,e.getOrCreateEmitNode=oi,e.setTextRange=si,e.setEmitFlags=ci,e.addEmitFlags=li,e.getSourceMapRange=ui,e.setSourceMapRange=di,e.createSourceMapSource=pi,e.getTokenSourceMapRange=mi,e.setTokenSourceMapRange=fi,e.getStartsOnNewLine=_i,e.setStartsOnNewLine=gi,e.getCommentRange=vi,e.setCommentRange=yi,e.getSyntheticLeadingComments=hi,e.setSyntheticLeadingComments=bi,e.addSyntheticLeadingComment=Ei,e.getSyntheticTrailingComments=Ti,e.setSyntheticTrailingComments=Si,e.addSyntheticTrailingComment=xi,e.moveSyntheticComments=Ci,e.getConstantValue=Ai,e.setConstantValue=Di,e.addEmitHelper=Li,e.addEmitHelpers=ki,e.removeEmitHelper=Ni,e.getEmitHelpers=Ii,e.moveEmitHelpers=wi,e.compareEmitHelpers=Mi,e.setOriginalNode=Oi}(u||(u={})),function(e){function n(n,t){return"undefined"===t?e.createStrictEquality(n,e.createVoidZero()):e.createStrictEquality(e.createTypeOf(n),e.createLiteral(t))}function t(n,t,r){if(e.isComputedPropertyName(t))return e.setTextRange(e.createElementAccess(n,t.expression),r);var a=e.setTextRange(e.isIdentifier(t)?e.createPropertyAccess(n,t):e.createElementAccess(n,t),t);return e.getOrCreateEmitNode(a).flags|=64,a}function r(n,t,r,a){return e.setTextRange(e.createCall(e.createPropertyAccess(n,"call"),void 0,[t].concat(r)),a)}function a(n,t,r,a){return e.setTextRange(e.createCall(e.createPropertyAccess(n,"apply"),void 0,[t,r]),a)}function i(n,t){var r=[];return void 0!==t&&r.push("number"===typeof t?e.createLiteral(t):t),e.createCall(e.createPropertyAccess(n,"slice"),void 0,r)}function o(n,t){return e.createCall(e.createPropertyAccess(n,"concat"),void 0,t)}function s(n,t,r){return e.setTextRange(e.createCall(e.createPropertyAccess(e.createIdentifier("Math"),"pow"),void 0,[n,t]),r)}function c(n,t){var r=e.createIdentifier(n||"React");return r.flags&=-9,r.parent=e.getParseTreeNode(t),r}function l(n,t){if(e.isQualifiedName(n)){var r=l(n.left,t),a=e.createIdentifier(e.idText(n.right));return a.escapedText=n.right.escapedText,e.createPropertyAccess(r,a)}return c(e.idText(n),t)}function u(n,t,r){return n?l(n,r):e.createPropertyAccess(c(t,r),"createElement")}function d(n,t,r,a,i,o,s){var c=[r];if(a&&c.push(a),i&&i.length>0)if(a||c.push(e.createNull()),i.length>1)for(var l=0,d=i;l<d.length;l++){var p=d[l];Ae(p),c.push(p)}else c.push(i[0]);return e.setTextRange(e.createCall(u(n,t,o),void 0,c),s)}function p(n,t,r,a,i){var o=e.createPropertyAccess(c(t,a),"Fragment"),s=[o];if(s.push(e.createNull()),r&&r.length>0)if(r.length>1)for(var l=0,d=r;l<d.length;l++){var p=d[l];Ae(p),s.push(p)}else s.push(r[0]);return e.setTextRange(e.createCall(u(n,t,a),void 0,s),i)}function m(n){return e.setEmitFlags(e.createIdentifier(n),4098)}e.nullTransformationContext={enableEmitNotification:e.noop,enableSubstitution:e.noop,endLexicalEnvironment:function(){},getCompilerOptions:e.notImplemented,getEmitHost:e.notImplemented,getEmitResolver:e.notImplemented,hoistFunctionDeclaration:e.noop,hoistVariableDeclaration:e.noop,isEmitNotificationEnabled:e.notImplemented,isSubstitutionEnabled:e.notImplemented,onEmitNode:e.noop,onSubstituteNode:e.notImplemented,readEmitHelpers:e.notImplemented,requestEmitHelper:e.noop,resumeLexicalEnvironment:e.noop,startLexicalEnvironment:e.noop,suspendLexicalEnvironment:e.noop,addDiagnostic:e.noop},e.createTypeCheck=n,e.createMemberAccessForPropertyName=t,e.createFunctionCall=r,e.createFunctionApply=a,e.createArraySlice=i,e.createArrayConcat=o,e.createMathPow=s,e.createExpressionForJsxElement=d,e.createExpressionForJsxFragment=p,e.getHelperName=m;var f={name:"typescript:values",scoped:!1,text:'\n            var __values = (this && this.__values) || function (o) {\n                var m = typeof Symbol === "function" && o[Symbol.iterator], i = 0;\n                if (m) return m.call(o);\n                return {\n                    next: function () {\n                        if (o && i >= o.length) o = void 0;\n                        return { value: o && o[i++], done: !o };\n                    }\n                };\n            };'};function _(n,t,r){return n.requestEmitHelper(f),e.setTextRange(e.createCall(m("__values"),void 0,[t]),r)}e.createValuesHelper=_;var g={name:"typescript:read",scoped:!1,text:'\n            var __read = (this && this.__read) || function (o, n) {\n                var m = typeof Symbol === "function" && o[Symbol.iterator];\n                if (!m) return o;\n                var i = m.call(o), r, ar = [], e;\n                try {\n                    while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);\n                }\n                catch (error) { e = { error: error }; }\n                finally {\n                    try {\n                        if (r && !r.done && (m = i["return"])) m.call(i);\n                    }\n                    finally { if (e) throw e.error; }\n                }\n                return ar;\n            };'};function v(n,t,r,a){return n.requestEmitHelper(g),e.setTextRange(e.createCall(m("__read"),void 0,void 0!==r?[t,e.createLiteral(r)]:[t]),a)}e.createReadHelper=v;var y={name:"typescript:spread",scoped:!1,text:"\n            var __spread = (this && this.__spread) || function () {\n                for (var ar = [], i = 0; i < arguments.length; i++) ar = ar.concat(__read(arguments[i]));\n                return ar;\n            };"};function h(n,t,r){return n.requestEmitHelper(g),n.requestEmitHelper(y),e.setTextRange(e.createCall(m("__spread"),void 0,t),r)}function b(n,t){if(e.isVariableDeclarationList(n)){var r=e.first(n.declarations),a=e.updateVariableDeclaration(r,r.name,void 0,t);return e.setTextRange(e.createVariableStatement(void 0,e.updateVariableDeclarationList(n,[a])),n)}var i=e.setTextRange(e.createAssignment(n,t),n);return e.setTextRange(e.createStatement(i),n)}function E(n,t){return e.isBlock(n)?e.updateBlock(n,e.setTextRange(e.createNodeArray([t].concat(n.statements)),n.statements)):e.createBlock(e.createNodeArray([n,t]),!0)}function T(n,t,r){if(!t)return n;var a=e.updateLabel(t,t.label,233===t.statement.kind?T(n,t.statement):n);return r&&r(t),a}function S(n,t){var r=e.skipParentheses(n);switch(r.kind){case 72:return t;case 100:case 8:case 9:case 10:return!1;case 187:var a=r.elements;return 0!==a.length;case 188:return r.properties.length>0;default:return!0}}function x(n,t,r,a){void 0===a&&(a=!1);var i,o,s=Ee(n,7);if(e.isSuperProperty(s))i=e.createThis(),o=s;else if(98===s.kind)i=e.createThis(),o=r<2?e.setTextRange(e.createIdentifier("_super"),s):s;else if(4096&e.getEmitFlags(s))i=e.createVoidZero(),o=oe(s);else switch(s.kind){case 189:S(s.expression,a)?(i=e.createTempVariable(t),o=e.createPropertyAccess(e.setTextRange(e.createAssignment(i,s.expression),s.expression),s.name),e.setTextRange(o,s)):(i=s.expression,o=s);break;case 190:S(s.expression,a)?(i=e.createTempVariable(t),o=e.createElementAccess(e.setTextRange(e.createAssignment(i,s.expression),s.expression),s.argumentExpression),e.setTextRange(o,s)):(i=s.expression,o=s);break;default:i=e.createVoidZero(),o=oe(n);break}return{target:o,thisArg:i}}function C(n){return n.length>10?e.createCommaList(n):e.reduceLeft(n,e.createComma)}function A(n){if(e.isQualifiedName(n)){var t=A(n.left),r=e.getMutableClone(n.right);return e.setTextRange(e.createPropertyAccess(t,r),n)}return e.getMutableClone(n)}function D(n){return e.isIdentifier(n)?e.createLiteral(n):e.isComputedPropertyName(n)?e.getMutableClone(n.expression):e.getMutableClone(n)}function L(e,n,t){switch(n.kind){case 158:case 159:return k(e.properties,n,t,!!e.multiLine);case 275:return N(n,t);case 276:return I(n,t);case 156:return w(n,t)}}function k(n,t,r,a){var i=e.getAllAccessorDeclarations(n,t),o=i.firstAccessor,s=i.getAccessor,c=i.setAccessor;if(t===o){var l=[];if(s){var u=e.createFunctionExpression(s.modifiers,void 0,void 0,void 0,s.parameters,void 0,s.body);e.setTextRange(u,s),e.setOriginalNode(u,s);var d=e.createPropertyAssignment("get",u);l.push(d)}if(c){var p=e.createFunctionExpression(c.modifiers,void 0,void 0,void 0,c.parameters,void 0,c.body);e.setTextRange(p,c),e.setOriginalNode(p,c);var m=e.createPropertyAssignment("set",p);l.push(m)}l.push(e.createPropertyAssignment("enumerable",e.createTrue())),l.push(e.createPropertyAssignment("configurable",e.createTrue()));var f=e.setTextRange(e.createCall(e.createPropertyAccess(e.createIdentifier("Object"),"defineProperty"),void 0,[r,D(t.name),e.createObjectLiteral(l,a)]),o);return e.aggregateTransformFlags(f)}}function N(n,r){return e.aggregateTransformFlags(e.setOriginalNode(e.setTextRange(e.createAssignment(t(r,n.name,n.name),n.initializer),n),n))}function I(n,r){return e.aggregateTransformFlags(e.setOriginalNode(e.setTextRange(e.createAssignment(t(r,n.name,n.name),e.getSynthesizedClone(n.name)),n),n))}function w(n,r){return e.aggregateTransformFlags(e.setOriginalNode(e.setTextRange(e.createAssignment(t(r,n.name,n.name),e.setOriginalNode(e.setTextRange(e.createFunctionExpression(n.modifiers,n.asteriskToken,void 0,void 0,n.parameters,void 0,n.body),n),n)),n),n))}function M(e,n,t){return B(e,n,t,49152)}function O(n){return 0!==(32768&e.getEmitFlags(n))}function P(e,n,t){return B(e,n,t,16384)}function R(n){return 0!==(16384&e.getEmitFlags(n))}function F(e,n,t){return B(e,n,t,8192)}function G(n){return 0!==(8192&e.getEmitFlags(n))}function V(e,n,t){return B(e,n,t)}function B(n,t,r,a){void 0===a&&(a=0);var i=e.getNameOfDeclaration(n);if(i&&e.isIdentifier(i)&&!e.isGeneratedIdentifier(i)){var o=e.getMutableClone(i);return a|=e.getEmitFlags(i),r||(a|=48),t||(a|=1536),a&&e.setEmitFlags(o,a),o}return e.getGeneratedNameForNode(n)}function K(n,t,r,a){return n&&e.hasModifier(t,1)?j(n,B(t),r,a):F(t,r,a)}function j(n,t,r,a){var i=e.createPropertyAccess(n,e.nodeIsSynthesized(t)?t:e.getSynthesizedClone(t));e.setTextRange(i,t);var o=0;return a||(o|=48),r||(o|=1536),o&&e.setEmitFlags(i,o),i}function U(n,t){return e.isBlock(n)?n:e.setTextRange(e.createBlock([e.setTextRange(e.createReturn(n),n)],t),n)}function H(n){if(!n.body)return e.Debug.fail();var t=e.createFunctionExpression(n.modifiers,n.asteriskToken,n.name,n.typeParameters,n.parameters,n.type,n.body);return e.setOriginalNode(t,n),e.setTextRange(t,n),e.getStartsOnNewLine(n)&&e.setStartsOnNewLine(t,!0),e.aggregateTransformFlags(t),t}function W(n){return e.isStringLiteral(n.expression)&&"use strict"===n.expression.text}function q(e,n,t,r){var a=z(e,n,t);return J(e,n,a,r)}function z(n,t,r){e.Debug.assert(0===n.length,"Prologue directives should be at the first statement in the target statements array");var a=!1,i=0,o=t.length;while(i<o){var s=t[i];if(!e.isPrologueDirective(s))break;W(s)&&(a=!0),n.push(s),i++}return r&&!a&&n.push(Ae(e.createStatement(e.createLiteral("use strict")))),i}function J(n,t,r,a){var i=t.length;while(void 0!==r&&r<i){var o=t[r];if(!(1048576&e.getEmitFlags(o)))break;e.append(n,a?e.visitNode(o,a,e.isStatement):o),r++}return r}function X(n){for(var t=0,r=n;t<r.length;t++){var a=r[t];if(!e.isPrologueDirective(a))break;if(W(a))return a}}function Y(n){var t=e.firstOrUndefined(n);return void 0!==t&&e.isPrologueDirective(t)&&W(t)}function Q(n){var t=X(n);return t?n:e.setTextRange(e.createNodeArray([Ae(e.createStatement(e.createLiteral("use strict")))].concat(n)),n)}function Z(n,t,r,a){var i=e.skipPartiallyEmittedExpressions(t);return 195===i.kind?t:$(n,t,r,a)?e.createParen(t):t}function $(n,t,r,a){var i=e.getOperatorPrecedence(204,n),o=e.getOperatorAssociativity(204,n),s=e.skipPartiallyEmittedExpressions(t);if(!r&&197===t.kind&&i>4)return!0;var c=e.getExpressionPrecedence(s);switch(e.compareValues(c,i)){case-1:return!(!r&&1===o&&207===t.kind);case 1:return!1;case 0:if(r)return 1===o;if(e.isBinaryExpression(s)&&s.operatorToken.kind===n){if(ee(n))return!1;if(38===n){var l=a?ne(a):0;if(e.isLiteralKind(l)&&l===ne(s))return!1}}var u=e.getExpressionAssociativity(s);return 0===u}}function ee(e){return 40===e||50===e||49===e||51===e}function ne(n){if(n=e.skipPartiallyEmittedExpressions(n),e.isLiteralKind(n.kind))return n.kind;if(204===n.kind&&38===n.operatorToken.kind){if(void 0!==n.cachedLiteralKind)return n.cachedLiteralKind;var t=ne(n.left),r=e.isLiteralKind(t)&&t===ne(n.right)?t:0;return n.cachedLiteralKind=r,r}return 0}function te(n){var t=e.getOperatorPrecedence(205,56),r=e.skipPartiallyEmittedExpressions(n),a=e.getExpressionPrecedence(r);return-1===e.compareValues(a,t)?e.createParen(n):n}function re(n){var t=e.skipPartiallyEmittedExpressions(n);return he(t)?e.createParen(n):n}function ae(n){var t=e.skipPartiallyEmittedExpressions(n),r=he(t);if(!r)switch(ve(t,!1).kind){case 209:case 196:r=!0}return r?e.createParen(n):n}function ie(n){var t=ve(n,!0);switch(t.kind){case 191:return e.createParen(n);case 192:return t.arguments?n:e.createParen(n)}return oe(n)}function oe(n){var t=e.skipPartiallyEmittedExpressions(n);return e.isLeftHandSideExpression(t)&&(192!==t.kind||t.arguments)?n:e.setTextRange(e.createParen(n),n)}function se(n){return e.isLeftHandSideExpression(n)?n:e.setTextRange(e.createParen(n),n)}function ce(n){return e.isUnaryExpression(n)?n:e.setTextRange(e.createParen(n),n)}function le(n){for(var t,r=0;r<n.length;r++){var a=ue(n[r]);void 0===t&&a===n[r]||(void 0===t&&(t=n.slice(0,r)),t.push(a))}return void 0!==t?e.setTextRange(e.createNodeArray(t,n.hasTrailingComma),n):n}function ue(n){var t=e.skipPartiallyEmittedExpressions(n),r=e.getExpressionPrecedence(t),a=e.getOperatorPrecedence(204,27);return r>a?n:e.setTextRange(e.createParen(n),n)}function de(n){var t=e.skipPartiallyEmittedExpressions(n);if(e.isCallExpression(t)){var r=t.expression,a=e.skipPartiallyEmittedExpressions(r).kind;if(196===a||197===a){var i=e.getMutableClone(t);return i.expression=e.setTextRange(e.createParen(r),r),Ce(n,i,4)}}var o=ve(t,!1).kind;return 188===o||196===o?e.setTextRange(e.createParen(n),n):n}function pe(n){return 175===n.kind?e.createParenthesizedType(n):n}function me(n){switch(n.kind){case 173:case 174:case 165:case 166:return e.createParenthesizedType(n)}return pe(n)}function fe(n){switch(n.kind){case 167:case 179:case 176:return e.createParenthesizedType(n)}return me(n)}function _e(n){return e.createNodeArray(e.sameMap(n,me))}function ge(n){if(e.some(n)){for(var t=[],r=0;r<n.length;++r){var a=n[r];t.push(0===r&&e.isFunctionOrConstructorTypeNode(a)&&a.typeParameters?e.createParenthesizedType(a):a)}return e.createNodeArray(t)}}function ve(e,n){while(1){switch(e.kind){case 203:e=e.operand;continue;case 204:e=e.left;continue;case 205:e=e.condition;continue;case 193:e=e.tag;continue;case 191:if(n)return e;case 212:case 190:case 189:case 213:case 308:e=e.expression;continue}return e}}function ye(n){return e.isBlock(n)||!he(n)&&188!==ve(n,!1).kind?n:e.setTextRange(e.createParen(n),n)}function he(e){return 204===e.kind&&27===e.operatorToken.kind||309===e.kind}function be(e,n){switch(void 0===n&&(n=7),e.kind){case 195:return 0!==(1&n);case 194:case 212:case 213:return 0!==(2&n);case 308:return 0!==(4&n)}return!1}function Ee(n,t){var r;void 0===t&&(t=7);do{r=n,1&t&&(n=e.skipParentheses(n)),2&t&&(n=Te(n)),4&t&&(n=e.skipPartiallyEmittedExpressions(n))}while(r!==n);return n}function Te(n){while(e.isAssertionExpression(n)||213===n.kind)n=n.expression;return n}function Se(n,t){switch(n.kind){case 195:return e.updateParen(n,t);case 194:return e.updateTypeAssertion(n,n.type,t);case 212:return e.updateAsExpression(n,t,n.type);case 213:return e.updateNonNullExpression(n,t);case 308:return e.updatePartiallyEmittedExpression(n,t)}}function xe(n){return 195===n.kind&&e.nodeIsSynthesized(n)&&e.nodeIsSynthesized(e.getSourceMapRange(n))&&e.nodeIsSynthesized(e.getCommentRange(n))&&!e.some(e.getSyntheticLeadingComments(n))&&!e.some(e.getSyntheticTrailingComments(n))}function Ce(e,n,t){return void 0===t&&(t=7),e&&be(e,t)&&!xe(e)?Se(e,Ce(e.expression,n)):n}function Ae(n){return e.setStartsOnNewLine(n,!0)}function De(n){var t=e.getOriginalNode(n,e.isSourceFile),r=t&&t.emitNode;return r&&r.externalHelpersModuleName}function Le(n,t,r,a){if(t.importHelpers&&e.isEffectiveExternalModule(n,t)){var i=De(n);if(i)return i;var o=e.getEmitModuleKind(t),s=(r||t.esModuleInterop&&a)&&o!==e.ModuleKind.System&&o!==e.ModuleKind.ES2015&&o!==e.ModuleKind.ESNext;if(!s){var c=e.getEmitHelpers(n);if(c)for(var l=0,u=c;l<u.length;l++){var d=u[l];if(!d.scoped){s=!0;break}}}if(s){var p=e.getOriginalNode(n,e.isSourceFile),m=e.getOrCreateEmitNode(p);return m.externalHelpersModuleName||(m.externalHelpersModuleName=e.createUniqueName(e.externalHelpersModuleNameText))}}}function ke(n,t){var r=e.getNamespaceDeclarationNode(n);if(r&&!e.isDefaultImport(n)){var a=r.name;return e.isGeneratedIdentifier(a)?a:e.createIdentifier(e.getSourceTextOfNodeFromSourceFile(t,a)||e.idText(a))}return 249===n.kind&&n.importClause?e.getGeneratedNameForNode(n):255===n.kind&&n.moduleSpecifier?e.getGeneratedNameForNode(n):void 0}function Ne(n,t,r,a,i){var o=e.getExternalModuleName(n);if(10===o.kind)return Me(n,r,a,i)||Ie(o,t)||e.getSynthesizedClone(o)}function Ie(n,t){var r=t.renamedDependencies&&t.renamedDependencies.get(n.text);return r&&e.createLiteral(r)}function we(n,t,r){if(n)return n.moduleName?e.createLiteral(n.moduleName):n.isDeclarationFile||!r.out&&!r.outFile?void 0:e.createLiteral(e.getExternalModuleNameFromPath(t,n.fileName))}function Me(e,n,t,r){return we(t.getExternalModuleFileFromDeclaration(e),n,r)}function Oe(n){if(e.isDeclarationBindingElement(n))return n.initializer;if(e.isPropertyAssignment(n)){var t=n.initializer;return e.isAssignmentExpression(t,!0)?t.right:void 0}return e.isShorthandPropertyAssignment(n)?n.objectAssignmentInitializer:e.isAssignmentExpression(n,!0)?n.right:e.isSpreadElement(n)?Oe(n.expression):void 0}function Pe(n){if(e.isDeclarationBindingElement(n))return n.name;if(!e.isObjectLiteralElementLike(n))return e.isAssignmentExpression(n,!0)?Pe(n.left):e.isSpreadElement(n)?Pe(n.expression):n;switch(n.kind){case 275:return Pe(n.initializer);case 276:return n.name;case 277:return Pe(n.expression)}}function Re(e){switch(e.kind){case 151:case 186:return e.dotDotDotToken;case 208:case 277:return e}}function Fe(n){switch(n.kind){case 186:if(n.propertyName){var t=n.propertyName;return e.isComputedPropertyName(t)&&Ge(t.expression)?t.expression:t}break;case 275:if(n.name){t=n.name;return e.isComputedPropertyName(t)&&Ge(t.expression)?t.expression:t}break;case 277:return n.name}var r=Pe(n);if(r&&e.isPropertyName(r))return e.isComputedPropertyName(r)&&Ge(r.expression)?r.expression:r;e.Debug.fail("Invalid property name for binding element.")}function Ge(e){var n=e.kind;return 10===n||8===n}function Ve(e){switch(e.kind){case 184:case 185:case 187:return e.elements;case 188:return e.properties}}function Be(n){if(e.isBindingElement(n)){if(n.dotDotDotToken)return e.Debug.assertNode(n.name,e.isIdentifier),e.setOriginalNode(e.setTextRange(e.createSpread(n.name),n),n);var t=We(n.name);return n.initializer?e.setOriginalNode(e.setTextRange(e.createAssignment(t,n.initializer),n),n):t}return e.Debug.assertNode(n,e.isExpression),n}function Ke(n){if(e.isBindingElement(n)){if(n.dotDotDotToken)return e.Debug.assertNode(n.name,e.isIdentifier),e.setOriginalNode(e.setTextRange(e.createSpreadAssignment(n.name),n),n);if(n.propertyName){var t=We(n.name);return e.setOriginalNode(e.setTextRange(e.createPropertyAssignment(n.propertyName,n.initializer?e.createAssignment(t,n.initializer):t),n),n)}return e.Debug.assertNode(n.name,e.isIdentifier),e.setOriginalNode(e.setTextRange(e.createShorthandPropertyAssignment(n.name,n.initializer),n),n)}return e.Debug.assertNode(n,e.isObjectLiteralElementLike),n}function je(e){switch(e.kind){case 185:case 187:return He(e);case 184:case 188:return Ue(e)}}function Ue(n){return e.isObjectBindingPattern(n)?e.setOriginalNode(e.setTextRange(e.createObjectLiteral(e.map(n.elements,Ke)),n),n):(e.Debug.assertNode(n,e.isObjectLiteralExpression),n)}function He(n){return e.isArrayBindingPattern(n)?e.setOriginalNode(e.setTextRange(e.createArrayLiteral(e.map(n.elements,Be)),n),n):(e.Debug.assertNode(n,e.isArrayLiteralExpression),n)}function We(n){return e.isBindingPattern(n)?je(n):(e.Debug.assertNode(n,e.isExpression),n)}e.createSpreadHelper=h,e.createForOfBindingStatement=b,e.insertLeadingStatement=E,e.restoreEnclosingLabel=T,e.createCallBinding=x,e.inlineExpressions=C,e.createExpressionFromEntityName=A,e.createExpressionForPropertyName=D,e.createExpressionForObjectLiteralElementLike=L,e.getInternalName=M,e.isInternalName=O,e.getLocalName=P,e.isLocalName=R,e.getExportName=F,e.isExportName=G,e.getDeclarationName=V,e.getExternalModuleOrNamespaceExportName=K,e.getNamespaceMemberName=j,e.convertToFunctionBody=U,e.convertFunctionDeclarationToExpression=H,e.addPrologue=q,e.addStandardPrologue=z,e.addCustomPrologue=J,e.findUseStrictPrologue=X,e.startsWithUseStrict=Y,e.ensureUseStrict=Q,e.parenthesizeBinaryOperand=Z,e.parenthesizeForConditionalHead=te,e.parenthesizeSubexpressionOfConditionalExpression=re,e.parenthesizeDefaultExpression=ae,e.parenthesizeForNew=ie,e.parenthesizeForAccess=oe,e.parenthesizePostfixOperand=se,e.parenthesizePrefixOperand=ce,e.parenthesizeListElements=le,e.parenthesizeExpressionForList=ue,e.parenthesizeExpressionForExpressionStatement=de,e.parenthesizeConditionalTypeMember=pe,e.parenthesizeElementTypeMember=me,e.parenthesizeArrayTypeMember=fe,e.parenthesizeElementTypeMembers=_e,e.parenthesizeTypeParameters=ge,e.parenthesizeConciseBody=ye,e.isCommaSequence=he,function(e){e[e["Parentheses"]=1]="Parentheses",e[e["Assertions"]=2]="Assertions",e[e["PartiallyEmittedExpressions"]=4]="PartiallyEmittedExpressions",e[e["All"]=7]="All"}(e.OuterExpressionKinds||(e.OuterExpressionKinds={})),e.isOuterExpression=be,e.skipOuterExpressions=Ee,e.skipAssertions=Te,e.recreateOuterExpressions=Ce,e.startOnNewLine=Ae,e.getExternalHelpersModuleName=De,e.getOrCreateExternalHelpersModuleNameIfNeeded=Le,e.getLocalNameForExternalImport=ke,e.getExternalModuleNameLiteral=Ne,e.tryGetModuleNameFromFile=we,e.getInitializerOfBindingOrAssignmentElement=Oe,e.getTargetOfBindingOrAssignmentElement=Pe,e.getRestIndicatorOfBindingOrAssignmentElement=Re,e.getPropertyNameOfBindingOrAssignmentElement=Fe,e.getElementsOfBindingOrAssignmentPattern=Ve,e.convertToArrayAssignmentElement=Be,e.convertToObjectAssignmentElement=Ke,e.convertToAssignmentPattern=je,e.convertToObjectAssignmentPattern=Ue,e.convertToArrayAssignmentPattern=He,e.convertToAssignmentElementTarget=We}(u||(u={})),function(e){var n=e.or(e.isTypeNode,e.isTypeParameterDeclaration);function t(n,t,r,a){if(void 0===n||void 0===t)return n;e.aggregateTransformFlags(n);var i,o=t(n);return o===n?n:void 0!==o?(i=e.isArray(o)?(a||c)(o):o,e.Debug.assertNode(i,r),e.aggregateTransformFlags(i),i):void 0}function r(n,t,r,a,i){if(void 0===n||void 0===t)return n;var o,s=n.length;(void 0===a||a<0)&&(a=0),(void 0===i||i>s-a)&&(i=s-a),(a>0||i<s)&&(o=e.createNodeArray([],n.hasTrailingComma&&a+i===s));for(var c=0;c<i;c++){var l=n[c+a];e.aggregateTransformFlags(l);var u=void 0!==l?t(l):void 0;if((void 0!==o||void 0===u||u!==l)&&(void 0===o&&(o=e.createNodeArray(n.slice(0,c),n.hasTrailingComma),e.setTextRange(o,n)),u))if(e.isArray(u))for(var d=0,p=u;d<p.length;d++){var m=p[d];e.Debug.assertNode(m,r),e.aggregateTransformFlags(m),o.push(m)}else e.Debug.assertNode(u,r),e.aggregateTransformFlags(u),o.push(u)}return o||n}function a(n,t,a,i,o){a.startLexicalEnvironment(),n=r(n,t,e.isStatement,i),o&&!e.startsWithUseStrict(n)&&(n=e.setTextRange(e.createNodeArray([e.createExpressionStatement(e.createLiteral("use strict"))].concat(n)),n));var s=a.endLexicalEnvironment();return e.setTextRange(e.createNodeArray(e.concatenate(s,n)),n)}function i(n,t,a,i){void 0===i&&(i=r),a.startLexicalEnvironment();var o=i(n,t,e.isParameterDeclaration);return a.suspendLexicalEnvironment(),o}function o(n,r,a){a.resumeLexicalEnvironment();var i=t(n,r,e.isConciseBody),o=a.endLexicalEnvironment();if(e.some(o)){var s=e.convertToFunctionBody(i),c=e.mergeLexicalEnvironment(s.statements,o);return e.updateBlock(s,c)}return i}function s(s,c,l,u,d){if(void 0===u&&(u=r),void 0!==s){var p=s.kind;if(p>0&&p<=147||178===p)return s;switch(p){case 72:return e.updateIdentifier(s,u(s.typeArguments,c,n));case 148:return e.updateQualifiedName(s,t(s.left,c,e.isEntityName),t(s.right,c,e.isIdentifier));case 149:return e.updateComputedPropertyName(s,t(s.expression,c,e.isExpression));case 150:return e.updateTypeParameterDeclaration(s,t(s.name,c,e.isIdentifier),t(s.constraint,c,e.isTypeNode),t(s.default,c,e.isTypeNode));case 151:return e.updateParameter(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.dotDotDotToken,d,e.isToken),t(s.name,c,e.isBindingName),t(s.questionToken,d,e.isToken),t(s.type,c,e.isTypeNode),t(s.initializer,c,e.isExpression));case 152:return e.updateDecorator(s,t(s.expression,c,e.isExpression));case 153:return e.updatePropertySignature(s,u(s.modifiers,c,e.isToken),t(s.name,c,e.isPropertyName),t(s.questionToken,d,e.isToken),t(s.type,c,e.isTypeNode),t(s.initializer,c,e.isExpression));case 154:return e.updateProperty(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.name,c,e.isPropertyName),t(s.questionToken,d,e.isToken),t(s.type,c,e.isTypeNode),t(s.initializer,c,e.isExpression));case 155:return e.updateMethodSignature(s,u(s.typeParameters,c,e.isTypeParameterDeclaration),u(s.parameters,c,e.isParameterDeclaration),t(s.type,c,e.isTypeNode),t(s.name,c,e.isPropertyName),t(s.questionToken,d,e.isToken));case 156:return e.updateMethod(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.asteriskToken,d,e.isToken),t(s.name,c,e.isPropertyName),t(s.questionToken,d,e.isToken),u(s.typeParameters,c,e.isTypeParameterDeclaration),i(s.parameters,c,l,u),t(s.type,c,e.isTypeNode),o(s.body,c,l));case 157:return e.updateConstructor(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),i(s.parameters,c,l,u),o(s.body,c,l));case 158:return e.updateGetAccessor(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.name,c,e.isPropertyName),i(s.parameters,c,l,u),t(s.type,c,e.isTypeNode),o(s.body,c,l));case 159:return e.updateSetAccessor(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.name,c,e.isPropertyName),i(s.parameters,c,l,u),o(s.body,c,l));case 160:return e.updateCallSignature(s,u(s.typeParameters,c,e.isTypeParameterDeclaration),u(s.parameters,c,e.isParameterDeclaration),t(s.type,c,e.isTypeNode));case 161:return e.updateConstructSignature(s,u(s.typeParameters,c,e.isTypeParameterDeclaration),u(s.parameters,c,e.isParameterDeclaration),t(s.type,c,e.isTypeNode));case 162:return e.updateIndexSignature(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),u(s.parameters,c,e.isParameterDeclaration),t(s.type,c,e.isTypeNode));case 163:return e.updateTypePredicateNode(s,t(s.parameterName,c),t(s.type,c,e.isTypeNode));case 164:return e.updateTypeReferenceNode(s,t(s.typeName,c,e.isEntityName),u(s.typeArguments,c,e.isTypeNode));case 165:return e.updateFunctionTypeNode(s,u(s.typeParameters,c,e.isTypeParameterDeclaration),u(s.parameters,c,e.isParameterDeclaration),t(s.type,c,e.isTypeNode));case 166:return e.updateConstructorTypeNode(s,u(s.typeParameters,c,e.isTypeParameterDeclaration),u(s.parameters,c,e.isParameterDeclaration),t(s.type,c,e.isTypeNode));case 167:return e.updateTypeQueryNode(s,t(s.exprName,c,e.isEntityName));case 168:return e.updateTypeLiteralNode(s,u(s.members,c,e.isTypeElement));case 169:return e.updateArrayTypeNode(s,t(s.elementType,c,e.isTypeNode));case 170:return e.updateTupleTypeNode(s,u(s.elementTypes,c,e.isTypeNode));case 171:return e.updateOptionalTypeNode(s,t(s.type,c,e.isTypeNode));case 172:return e.updateRestTypeNode(s,t(s.type,c,e.isTypeNode));case 173:return e.updateUnionTypeNode(s,u(s.types,c,e.isTypeNode));case 174:return e.updateIntersectionTypeNode(s,u(s.types,c,e.isTypeNode));case 175:return e.updateConditionalTypeNode(s,t(s.checkType,c,e.isTypeNode),t(s.extendsType,c,e.isTypeNode),t(s.trueType,c,e.isTypeNode),t(s.falseType,c,e.isTypeNode));case 176:return e.updateInferTypeNode(s,t(s.typeParameter,c,e.isTypeParameterDeclaration));case 183:return e.updateImportTypeNode(s,t(s.argument,c,e.isTypeNode),t(s.qualifier,c,e.isEntityName),r(s.typeArguments,c,e.isTypeNode),s.isTypeOf);case 177:return e.updateParenthesizedType(s,t(s.type,c,e.isTypeNode));case 179:return e.updateTypeOperatorNode(s,t(s.type,c,e.isTypeNode));case 180:return e.updateIndexedAccessTypeNode(s,t(s.objectType,c,e.isTypeNode),t(s.indexType,c,e.isTypeNode));case 181:return e.updateMappedTypeNode(s,t(s.readonlyToken,d,e.isToken),t(s.typeParameter,c,e.isTypeParameterDeclaration),t(s.questionToken,d,e.isToken),t(s.type,c,e.isTypeNode));case 182:return e.updateLiteralTypeNode(s,t(s.literal,c,e.isExpression));case 184:return e.updateObjectBindingPattern(s,u(s.elements,c,e.isBindingElement));case 185:return e.updateArrayBindingPattern(s,u(s.elements,c,e.isArrayBindingElement));case 186:return e.updateBindingElement(s,t(s.dotDotDotToken,d,e.isToken),t(s.propertyName,c,e.isPropertyName),t(s.name,c,e.isBindingName),t(s.initializer,c,e.isExpression));case 187:return e.updateArrayLiteral(s,u(s.elements,c,e.isExpression));case 188:return e.updateObjectLiteral(s,u(s.properties,c,e.isObjectLiteralElementLike));case 189:return e.updatePropertyAccess(s,t(s.expression,c,e.isExpression),t(s.name,c,e.isIdentifier));case 190:return e.updateElementAccess(s,t(s.expression,c,e.isExpression),t(s.argumentExpression,c,e.isExpression));case 191:return e.updateCall(s,t(s.expression,c,e.isExpression),u(s.typeArguments,c,e.isTypeNode),u(s.arguments,c,e.isExpression));case 192:return e.updateNew(s,t(s.expression,c,e.isExpression),u(s.typeArguments,c,e.isTypeNode),u(s.arguments,c,e.isExpression));case 193:return e.updateTaggedTemplate(s,t(s.tag,c,e.isExpression),r(s.typeArguments,c,e.isExpression),t(s.template,c,e.isTemplateLiteral));case 194:return e.updateTypeAssertion(s,t(s.type,c,e.isTypeNode),t(s.expression,c,e.isExpression));case 195:return e.updateParen(s,t(s.expression,c,e.isExpression));case 196:return e.updateFunctionExpression(s,u(s.modifiers,c,e.isModifier),t(s.asteriskToken,d,e.isToken),t(s.name,c,e.isIdentifier),u(s.typeParameters,c,e.isTypeParameterDeclaration),i(s.parameters,c,l,u),t(s.type,c,e.isTypeNode),o(s.body,c,l));case 197:return e.updateArrowFunction(s,u(s.modifiers,c,e.isModifier),u(s.typeParameters,c,e.isTypeParameterDeclaration),i(s.parameters,c,l,u),t(s.type,c,e.isTypeNode),t(s.equalsGreaterThanToken,c,e.isToken),o(s.body,c,l));case 198:return e.updateDelete(s,t(s.expression,c,e.isExpression));case 199:return e.updateTypeOf(s,t(s.expression,c,e.isExpression));case 200:return e.updateVoid(s,t(s.expression,c,e.isExpression));case 201:return e.updateAwait(s,t(s.expression,c,e.isExpression));case 202:return e.updatePrefix(s,t(s.operand,c,e.isExpression));case 203:return e.updatePostfix(s,t(s.operand,c,e.isExpression));case 204:return e.updateBinary(s,t(s.left,c,e.isExpression),t(s.right,c,e.isExpression),t(s.operatorToken,c,e.isToken));case 205:return e.updateConditional(s,t(s.condition,c,e.isExpression),t(s.questionToken,c,e.isToken),t(s.whenTrue,c,e.isExpression),t(s.colonToken,c,e.isToken),t(s.whenFalse,c,e.isExpression));case 206:return e.updateTemplateExpression(s,t(s.head,c,e.isTemplateHead),u(s.templateSpans,c,e.isTemplateSpan));case 207:return e.updateYield(s,t(s.asteriskToken,d,e.isToken),t(s.expression,c,e.isExpression));case 208:return e.updateSpread(s,t(s.expression,c,e.isExpression));case 209:return e.updateClassExpression(s,u(s.modifiers,c,e.isModifier),t(s.name,c,e.isIdentifier),u(s.typeParameters,c,e.isTypeParameterDeclaration),u(s.heritageClauses,c,e.isHeritageClause),u(s.members,c,e.isClassElement));case 211:return e.updateExpressionWithTypeArguments(s,u(s.typeArguments,c,e.isTypeNode),t(s.expression,c,e.isExpression));case 212:return e.updateAsExpression(s,t(s.expression,c,e.isExpression),t(s.type,c,e.isTypeNode));case 213:return e.updateNonNullExpression(s,t(s.expression,c,e.isExpression));case 214:return e.updateMetaProperty(s,t(s.name,c,e.isIdentifier));case 216:return e.updateTemplateSpan(s,t(s.expression,c,e.isExpression),t(s.literal,c,e.isTemplateMiddleOrTemplateTail));case 218:return e.updateBlock(s,u(s.statements,c,e.isStatement));case 219:return e.updateVariableStatement(s,u(s.modifiers,c,e.isModifier),t(s.declarationList,c,e.isVariableDeclarationList));case 221:return e.updateExpressionStatement(s,t(s.expression,c,e.isExpression));case 222:return e.updateIf(s,t(s.expression,c,e.isExpression),t(s.thenStatement,c,e.isStatement,e.liftToBlock),t(s.elseStatement,c,e.isStatement,e.liftToBlock));case 223:return e.updateDo(s,t(s.statement,c,e.isStatement,e.liftToBlock),t(s.expression,c,e.isExpression));case 224:return e.updateWhile(s,t(s.expression,c,e.isExpression),t(s.statement,c,e.isStatement,e.liftToBlock));case 225:return e.updateFor(s,t(s.initializer,c,e.isForInitializer),t(s.condition,c,e.isExpression),t(s.incrementor,c,e.isExpression),t(s.statement,c,e.isStatement,e.liftToBlock));case 226:return e.updateForIn(s,t(s.initializer,c,e.isForInitializer),t(s.expression,c,e.isExpression),t(s.statement,c,e.isStatement,e.liftToBlock));case 227:return e.updateForOf(s,t(s.awaitModifier,c,e.isToken),t(s.initializer,c,e.isForInitializer),t(s.expression,c,e.isExpression),t(s.statement,c,e.isStatement,e.liftToBlock));case 228:return e.updateContinue(s,t(s.label,c,e.isIdentifier));case 229:return e.updateBreak(s,t(s.label,c,e.isIdentifier));case 230:return e.updateReturn(s,t(s.expression,c,e.isExpression));case 231:return e.updateWith(s,t(s.expression,c,e.isExpression),t(s.statement,c,e.isStatement,e.liftToBlock));case 232:return e.updateSwitch(s,t(s.expression,c,e.isExpression),t(s.caseBlock,c,e.isCaseBlock));case 233:return e.updateLabel(s,t(s.label,c,e.isIdentifier),t(s.statement,c,e.isStatement,e.liftToBlock));case 234:return e.updateThrow(s,t(s.expression,c,e.isExpression));case 235:return e.updateTry(s,t(s.tryBlock,c,e.isBlock),t(s.catchClause,c,e.isCatchClause),t(s.finallyBlock,c,e.isBlock));case 237:return e.updateVariableDeclaration(s,t(s.name,c,e.isBindingName),t(s.type,c,e.isTypeNode),t(s.initializer,c,e.isExpression));case 238:return e.updateVariableDeclarationList(s,u(s.declarations,c,e.isVariableDeclaration));case 239:return e.updateFunctionDeclaration(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.asteriskToken,d,e.isToken),t(s.name,c,e.isIdentifier),u(s.typeParameters,c,e.isTypeParameterDeclaration),i(s.parameters,c,l,u),t(s.type,c,e.isTypeNode),o(s.body,c,l));case 240:return e.updateClassDeclaration(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.name,c,e.isIdentifier),u(s.typeParameters,c,e.isTypeParameterDeclaration),u(s.heritageClauses,c,e.isHeritageClause),u(s.members,c,e.isClassElement));case 241:return e.updateInterfaceDeclaration(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.name,c,e.isIdentifier),u(s.typeParameters,c,e.isTypeParameterDeclaration),u(s.heritageClauses,c,e.isHeritageClause),u(s.members,c,e.isTypeElement));case 242:return e.updateTypeAliasDeclaration(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.name,c,e.isIdentifier),u(s.typeParameters,c,e.isTypeParameterDeclaration),t(s.type,c,e.isTypeNode));case 243:return e.updateEnumDeclaration(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.name,c,e.isIdentifier),u(s.members,c,e.isEnumMember));case 244:return e.updateModuleDeclaration(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.name,c,e.isIdentifier),t(s.body,c,e.isModuleBody));case 245:return e.updateModuleBlock(s,u(s.statements,c,e.isStatement));case 246:return e.updateCaseBlock(s,u(s.clauses,c,e.isCaseOrDefaultClause));case 247:return e.updateNamespaceExportDeclaration(s,t(s.name,c,e.isIdentifier));case 248:return e.updateImportEqualsDeclaration(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.name,c,e.isIdentifier),t(s.moduleReference,c,e.isModuleReference));case 249:return e.updateImportDeclaration(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.importClause,c,e.isImportClause),t(s.moduleSpecifier,c,e.isExpression));case 250:return e.updateImportClause(s,t(s.name,c,e.isIdentifier),t(s.namedBindings,c,e.isNamedImportBindings));case 251:return e.updateNamespaceImport(s,t(s.name,c,e.isIdentifier));case 252:return e.updateNamedImports(s,u(s.elements,c,e.isImportSpecifier));case 253:return e.updateImportSpecifier(s,t(s.propertyName,c,e.isIdentifier),t(s.name,c,e.isIdentifier));case 254:return e.updateExportAssignment(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.expression,c,e.isExpression));case 255:return e.updateExportDeclaration(s,u(s.decorators,c,e.isDecorator),u(s.modifiers,c,e.isModifier),t(s.exportClause,c,e.isNamedExports),t(s.moduleSpecifier,c,e.isExpression));case 256:return e.updateNamedExports(s,u(s.elements,c,e.isExportSpecifier));case 257:return e.updateExportSpecifier(s,t(s.propertyName,c,e.isIdentifier),t(s.name,c,e.isIdentifier));case 259:return e.updateExternalModuleReference(s,t(s.expression,c,e.isExpression));case 260:return e.updateJsxElement(s,t(s.openingElement,c,e.isJsxOpeningElement),u(s.children,c,e.isJsxChild),t(s.closingElement,c,e.isJsxClosingElement));case 261:return e.updateJsxSelfClosingElement(s,t(s.tagName,c,e.isJsxTagNameExpression),u(s.typeArguments,c,e.isTypeNode),t(s.attributes,c,e.isJsxAttributes));case 262:return e.updateJsxOpeningElement(s,t(s.tagName,c,e.isJsxTagNameExpression),u(s.typeArguments,c,e.isTypeNode),t(s.attributes,c,e.isJsxAttributes));case 263:return e.updateJsxClosingElement(s,t(s.tagName,c,e.isJsxTagNameExpression));case 264:return e.updateJsxFragment(s,t(s.openingFragment,c,e.isJsxOpeningFragment),u(s.children,c,e.isJsxChild),t(s.closingFragment,c,e.isJsxClosingFragment));case 267:return e.updateJsxAttribute(s,t(s.name,c,e.isIdentifier),t(s.initializer,c,e.isStringLiteralOrJsxExpression));case 268:return e.updateJsxAttributes(s,u(s.properties,c,e.isJsxAttributeLike));case 269:return e.updateJsxSpreadAttribute(s,t(s.expression,c,e.isExpression));case 270:return e.updateJsxExpression(s,t(s.expression,c,e.isExpression));case 271:return e.updateCaseClause(s,t(s.expression,c,e.isExpression),u(s.statements,c,e.isStatement));case 272:return e.updateDefaultClause(s,u(s.statements,c,e.isStatement));case 273:return e.updateHeritageClause(s,u(s.types,c,e.isExpressionWithTypeArguments));case 274:return e.updateCatchClause(s,t(s.variableDeclaration,c,e.isVariableDeclaration),t(s.block,c,e.isBlock));case 275:return e.updatePropertyAssignment(s,t(s.name,c,e.isPropertyName),t(s.initializer,c,e.isExpression));case 276:return e.updateShorthandPropertyAssignment(s,t(s.name,c,e.isIdentifier),t(s.objectAssignmentInitializer,c,e.isExpression));case 277:return e.updateSpreadAssignment(s,t(s.expression,c,e.isExpression));case 278:return e.updateEnumMember(s,t(s.name,c,e.isPropertyName),t(s.initializer,c,e.isExpression));case 279:return e.updateSourceFileNode(s,a(s.statements,c,l));case 308:return e.updatePartiallyEmittedExpression(s,t(s.expression,c,e.isExpression));case 309:return e.updateCommaList(s,u(s.elements,c,e.isExpression));default:return s}}}function c(n){return e.Debug.assert(n.length<=1,"Too many nodes written to output."),e.singleOrUndefined(n)}e.visitNode=t,e.visitNodes=r,e.visitLexicalEnvironment=a,e.visitParameterList=i,e.visitFunctionBody=o,e.visitEachChild=s}(u||(u={})),function(e){function n(e,n,t){return e?n(t,e):t}function t(e,n,t){return e?n(t,e):t}function r(r,a,i,o){if(void 0===r)return a;var s=o?t:e.reduceLeft,c=o||i,l=r.kind;if(l>0&&l<=147)return a;if(l>=163&&l<=182)return a;var u=a;switch(r.kind){case 217:case 220:case 210:case 236:case 307:break;case 148:u=n(r.left,i,u),u=n(r.right,i,u);break;case 149:u=n(r.expression,i,u);break;case 151:u=s(r.decorators,c,u),u=s(r.modifiers,c,u),u=n(r.name,i,u),u=n(r.type,i,u),u=n(r.initializer,i,u);break;case 152:u=n(r.expression,i,u);break;case 153:u=s(r.modifiers,c,u),u=n(r.name,i,u),u=n(r.questionToken,i,u),u=n(r.type,i,u),u=n(r.initializer,i,u);break;case 154:u=s(r.decorators,c,u),u=s(r.modifiers,c,u),u=n(r.name,i,u),u=n(r.type,i,u),u=n(r.initializer,i,u);break;case 156:u=s(r.decorators,c,u),u=s(r.modifiers,c,u),u=n(r.name,i,u),u=s(r.typeParameters,c,u),u=s(r.parameters,c,u),u=n(r.type,i,u),u=n(r.body,i,u);break;case 157:u=s(r.modifiers,c,u),u=s(r.parameters,c,u),u=n(r.body,i,u);break;case 158:u=s(r.decorators,c,u),u=s(r.modifiers,c,u),u=n(r.name,i,u),u=s(r.parameters,c,u),u=n(r.type,i,u),u=n(r.body,i,u);break;case 159:u=s(r.decorators,c,u),u=s(r.modifiers,c,u),u=n(r.name,i,u),u=s(r.parameters,c,u),u=n(r.body,i,u);break;case 184:case 185:u=s(r.elements,c,u);break;case 186:u=n(r.propertyName,i,u),u=n(r.name,i,u),u=n(r.initializer,i,u);break;case 187:u=s(r.elements,c,u);break;case 188:u=s(r.properties,c,u);break;case 189:u=n(r.expression,i,u),u=n(r.name,i,u);break;case 190:u=n(r.expression,i,u),u=n(r.argumentExpression,i,u);break;case 191:u=n(r.expression,i,u),u=s(r.typeArguments,c,u),u=s(r.arguments,c,u);break;case 192:u=n(r.expression,i,u),u=s(r.typeArguments,c,u),u=s(r.arguments,c,u);break;case 193:u=n(r.tag,i,u),u=s(r.typeArguments,c,u),u=n(r.template,i,u);break;case 194:u=n(r.type,i,u),u=n(r.expression,i,u);break;case 196:u=s(r.modifiers,c,u),u=n(r.name,i,u),u=s(r.typeParameters,c,u),u=s(r.parameters,c,u),u=n(r.type,i,u),u=n(r.body,i,u);break;case 197:u=s(r.modifiers,c,u),u=s(r.typeParameters,c,u),u=s(r.parameters,c,u),u=n(r.type,i,u),u=n(r.body,i,u);break;case 195:case 198:case 199:case 200:case 201:case 207:case 208:case 213:u=n(r.expression,i,u);break;case 202:case 203:u=n(r.operand,i,u);break;case 204:u=n(r.left,i,u),u=n(r.right,i,u);break;case 205:u=n(r.condition,i,u),u=n(r.whenTrue,i,u),u=n(r.whenFalse,i,u);break;case 206:u=n(r.head,i,u),u=s(r.templateSpans,c,u);break;case 209:u=s(r.modifiers,c,u),u=n(r.name,i,u),u=s(r.typeParameters,c,u),u=s(r.heritageClauses,c,u),u=s(r.members,c,u);break;case 211:u=n(r.expression,i,u),u=s(r.typeArguments,c,u);break;case 212:u=n(r.expression,i,u),u=n(r.type,i,u);break;case 216:u=n(r.expression,i,u),u=n(r.literal,i,u);break;case 218:u=s(r.statements,c,u);break;case 219:u=s(r.modifiers,c,u),u=n(r.declarationList,i,u);break;case 221:u=n(r.expression,i,u);break;case 222:u=n(r.expression,i,u),u=n(r.thenStatement,i,u),u=n(r.elseStatement,i,u);break;case 223:u=n(r.statement,i,u),u=n(r.expression,i,u);break;case 224:case 231:u=n(r.expression,i,u),u=n(r.statement,i,u);break;case 225:u=n(r.initializer,i,u),u=n(r.condition,i,u),u=n(r.incrementor,i,u),u=n(r.statement,i,u);break;case 226:case 227:u=n(r.initializer,i,u),u=n(r.expression,i,u),u=n(r.statement,i,u);break;case 230:case 234:u=n(r.expression,i,u);break;case 232:u=n(r.expression,i,u),u=n(r.caseBlock,i,u);break;case 233:u=n(r.label,i,u),u=n(r.statement,i,u);break;case 235:u=n(r.tryBlock,i,u),u=n(r.catchClause,i,u),u=n(r.finallyBlock,i,u);break;case 237:u=n(r.name,i,u),u=n(r.type,i,u),u=n(r.initializer,i,u);break;case 238:u=s(r.declarations,c,u);break;case 239:u=s(r.decorators,c,u),u=s(r.modifiers,c,u),u=n(r.name,i,u),u=s(r.typeParameters,c,u),u=s(r.parameters,c,u),u=n(r.type,i,u),u=n(r.body,i,u);break;case 240:u=s(r.decorators,c,u),u=s(r.modifiers,c,u),u=n(r.name,i,u),u=s(r.typeParameters,c,u),u=s(r.heritageClauses,c,u),u=s(r.members,c,u);break;case 243:u=s(r.decorators,c,u),u=s(r.modifiers,c,u),u=n(r.name,i,u),u=s(r.members,c,u);break;case 244:u=s(r.decorators,c,u),u=s(r.modifiers,c,u),u=n(r.name,i,u),u=n(r.body,i,u);break;case 245:u=s(r.statements,c,u);break;case 246:u=s(r.clauses,c,u);break;case 248:u=s(r.decorators,c,u),u=s(r.modifiers,c,u),u=n(r.name,i,u),u=n(r.moduleReference,i,u);break;case 249:u=s(r.decorators,c,u),u=s(r.modifiers,c,u),u=n(r.importClause,i,u),u=n(r.moduleSpecifier,i,u);break;case 250:u=n(r.name,i,u),u=n(r.namedBindings,i,u);break;case 251:u=n(r.name,i,u);break;case 252:case 256:u=s(r.elements,c,u);break;case 253:case 257:u=n(r.propertyName,i,u),u=n(r.name,i,u);break;case 254:u=e.reduceLeft(r.decorators,i,u),u=e.reduceLeft(r.modifiers,i,u),u=n(r.expression,i,u);break;case 255:u=e.reduceLeft(r.decorators,i,u),u=e.reduceLeft(r.modifiers,i,u),u=n(r.exportClause,i,u),u=n(r.moduleSpecifier,i,u);break;case 259:u=n(r.expression,i,u);break;case 260:u=n(r.openingElement,i,u),u=e.reduceLeft(r.children,i,u),u=n(r.closingElement,i,u);break;case 264:u=n(r.openingFragment,i,u),u=e.reduceLeft(r.children,i,u),u=n(r.closingFragment,i,u);break;case 261:case 262:u=n(r.tagName,i,u),u=s(r.typeArguments,i,u),u=n(r.attributes,i,u);break;case 268:u=s(r.properties,c,u);break;case 263:u=n(r.tagName,i,u);break;case 267:u=n(r.name,i,u),u=n(r.initializer,i,u);break;case 269:u=n(r.expression,i,u);break;case 270:u=n(r.expression,i,u);break;case 271:u=n(r.expression,i,u);case 272:u=s(r.statements,c,u);break;case 273:u=s(r.types,c,u);break;case 274:u=n(r.variableDeclaration,i,u),u=n(r.block,i,u);break;case 275:u=n(r.name,i,u),u=n(r.initializer,i,u);break;case 276:u=n(r.name,i,u),u=n(r.objectAssignmentInitializer,i,u);break;case 277:u=n(r.expression,i,u);break;case 278:u=n(r.name,i,u),u=n(r.initializer,i,u);break;case 279:u=s(r.statements,c,u);break;case 308:u=n(r.expression,i,u);break;case 309:u=s(r.elements,c,u);break;default:break}return u}function a(n,t){return e.some(t)?e.isNodeArray(n)?e.setTextRange(e.createNodeArray(e.addStatementsAfterPrologue(n.slice(),t)),n):e.addStatementsAfterPrologue(n,t):n}function i(n){return p.assert(e.every(n,e.isStatement),"Cannot lift nodes to a Block."),e.singleOrUndefined(n)||e.createBlock(n)}function o(e){return s(e),e}function s(n){if(void 0===n)return 0;if(536870912&n.transformFlags)return n.transformFlags&~e.getTransformFlagsSubtreeExclusions(n.kind);var t=l(n);return e.computeTransformFlagsForNode(n,t)}function c(e){if(void 0===e)return 0;for(var n=0,t=0,r=0,a=e;r<a.length;r++){var i=a[r];n|=s(i),t|=-536870913&i.transformFlags}return e.transformFlags=536870912|t,n}function l(n){return e.hasModifier(n,2)||e.isTypeNode(n)&&211!==n.kind?0:r(n,0,u,d)}function u(e,n){return e|s(n)}function d(e,n){return e|c(n)}var p;e.reduceEachChild=r,e.mergeLexicalEnvironment=a,e.liftToBlock=i,e.aggregateTransformFlags=o,function(n){var t=!1;function r(t,a){return n.fail((a||"Unexpected node.")+"\r\nNode "+e.formatSyntaxKind(t.kind)+" was unexpected.",r)}function a(){if(!t){Object.defineProperties(e.objectAllocator.getSymbolConstructor().prototype,{__debugFlags:{get:function(){return e.formatSymbolFlags(this.flags)}}}),Object.defineProperties(e.objectAllocator.getTypeConstructor().prototype,{__debugFlags:{get:function(){return e.formatTypeFlags(this.flags)}},__debugObjectFlags:{get:function(){return 524288&this.flags?e.formatObjectFlags(this.objectFlags):""}},__debugTypeToString:{value:function(){return this.checker.typeToString(this)}}});for(var n=[e.objectAllocator.getNodeConstructor(),e.objectAllocator.getIdentifierConstructor(),e.objectAllocator.getTokenConstructor(),e.objectAllocator.getSourceFileConstructor()],r=0,a=n;r<a.length;r++){var i=a[r];i.prototype.hasOwnProperty("__debugKind")||Object.defineProperties(i.prototype,{__debugKind:{get:function(){return e.formatSyntaxKind(this.kind)}},__debugModifierFlags:{get:function(){return e.formatModifierFlags(e.getModifierFlagsNoCache(this))}},__debugTransformFlags:{get:function(){return e.formatTransformFlags(this.transformFlags)}},__debugEmitFlags:{get:function(){return e.formatEmitFlags(e.getEmitFlags(this))}},__debugGetText:{value:function(n){if(e.nodeIsSynthesized(this))return"";var t=e.getParseTreeNode(this),r=t&&e.getSourceFileOfNode(t);return r?e.getSourceTextOfNodeFromSourceFile(r,t,n):""}}})}t=!0}}n.failBadSyntaxKind=r,n.assertEachNode=n.shouldAssert(1)?function(t,r,a){return n.assert(void 0===r||e.every(t,r),a||"Unexpected node.",function(){return"Node array did not pass test '"+n.getFunctionName(r)+"'."},n.assertEachNode)}:e.noop,n.assertNode=n.shouldAssert(1)?function(t,r,a){return n.assert(void 0===r||r(t),a||"Unexpected node.",function(){return"Node "+e.formatSyntaxKind(t.kind)+" did not pass test '"+n.getFunctionName(r)+"'."},n.assertNode)}:e.noop,n.assertOptionalNode=n.shouldAssert(1)?function(t,r,a){return n.assert(void 0===r||void 0===t||r(t),a||"Unexpected node.",function(){return"Node "+e.formatSyntaxKind(t.kind)+" did not pass test '"+n.getFunctionName(r)+"'."},n.assertOptionalNode)}:e.noop,n.assertOptionalToken=n.shouldAssert(1)?function(t,r,a){return n.assert(void 0===r||void 0===t||t.kind===r,a||"Unexpected node.",function(){return"Node "+e.formatSyntaxKind(t.kind)+" was not a '"+e.formatSyntaxKind(r)+"' token."},n.assertOptionalToken)}:e.noop,n.assertMissingNode=n.shouldAssert(1)?function(t,r){return n.assert(void 0===t,r||"Unexpected node.",function(){return"Node "+e.formatSyntaxKind(t.kind)+" was unexpected'."},n.assertMissingNode)}:e.noop,n.enableDebugInfo=a}(p=e.Debug||(e.Debug={}))}(u||(u={})),function(e){function n(n,t,r,a,i){var o,s,c=i.extendedDiagnostics?e.performance.createTimer("Source Map","beforeSourcemap","afterSourcemap"):e.performance.nullTimer,u=c.enter,d=c.exit,p=[],m=[],_=e.createMap(),g=[],v="",y=0,h=0,b=0,E=0,T=0,S=0,x=!1,C=0,A=0,D=0,L=0,k=0,N=0,I=!1,w=!1,M=!1;return{getSources:function(){return p},addSource:O,setSourceContent:P,addName:R,addMapping:V,appendSourceMap:B,toJSON:U,toString:function(){return JSON.stringify(U())}};function O(t){u();var r=e.getRelativePathToDirectoryOrUrl(a,t,n.getCurrentDirectory(),n.getCanonicalFileName,!0),i=_.get(r);return void 0===i&&(i=m.length,m.push(r),p.push(t),_.set(r,i)),d(),i}function P(e,n){if(u(),null!==n){o||(o=[]);while(o.length<e)o.push(null);o[e]=n}d()}function R(n){u(),s||(s=e.createMap());var t=s.get(n);return void 0===t&&(t=g.length,g.push(n),s.set(n,t)),d(),t}function F(e,n){return!I||C!==e||A!==n}function G(e,n,t){return void 0!==e&&void 0!==n&&void 0!==t&&D===e&&(L>n||L===n&&k>t)}function V(n,t,r,a,i,o){e.Debug.assert(n>=C,"generatedLine cannot backtrack"),e.Debug.assert(t>=0,"generatedCharacter cannot be negative"),e.Debug.assert(void 0===r||r>=0,"sourceIndex cannot be negative"),e.Debug.assert(void 0===a||a>=0,"sourceLine cannot be negative"),e.Debug.assert(void 0===i||i>=0,"sourceCharacter cannot be negative"),u(),(F(n,t)||G(r,a,i))&&(j(),C=n,A=t,w=!1,M=!1,I=!0),void 0!==r&&void 0!==a&&void 0!==i&&(D=r,L=a,k=i,w=!0,void 0!==o&&(N=o,M=!0)),d()}function B(n,t,r,a){var i;e.Debug.assert(n>=C,"generatedLine cannot backtrack"),e.Debug.assert(t>=0,"generatedCharacter cannot be negative"),u();for(var o,s=[],c=l(r.mappings),p=c.next(),m=p.value,f=p.done;!f;i=c.next(),m=i.value,f=i.done,i){var _=void 0,g=void 0,v=void 0,y=void 0;if(void 0!==m.sourceIndex){if(_=s[m.sourceIndex],void 0===_){var h=r.sources[m.sourceIndex],b=r.sourceRoot?e.combinePaths(r.sourceRoot,h):h,E=e.combinePaths(e.getDirectoryPath(a),b);s[m.sourceIndex]=_=O(E),r.sourcesContent&&"string"===typeof r.sourcesContent[m.sourceIndex]&&P(_,r.sourcesContent[m.sourceIndex])}g=m.sourceLine,v=m.sourceCharacter,r.names&&void 0!==m.nameIndex&&(o||(o=[]),y=o[m.nameIndex],void 0===y&&(o[m.nameIndex]=y=R(r.names[m.nameIndex])))}var T=m.generatedLine+n,S=0===m.generatedLine?m.generatedCharacter+t:m.generatedCharacter;V(T,S,_,g,v,y)}d()}function K(){return!x||y!==C||h!==A||b!==D||E!==L||T!==k||S!==N}function j(){if(I&&K()){if(u(),y<C)do{v+=";",y++,h=0}while(y<C);else e.Debug.assertEqual(y,C,"generatedLine cannot backtrack"),x&&(v+=",");v+=f(A-h),h=A,w&&(v+=f(D-b),b=D,v+=f(L-E),E=L,v+=f(k-T),T=k,M&&(v+=f(N-S),S=N)),x=!0,d()}}function U(){return j(),{version:3,file:t,sourceRoot:r,sources:m,names:g,mappings:v,sourcesContent:o}}}e.createSourceMapGenerator=n;var t=/^\/\/[@#] source[M]appingURL=(.+)\s*$/,r=/^\s*(\/\/[@#] .*)?$/;function a(e,n){return{getLineCount:function(){return n.length},getLineText:function(t){return e.substring(n[t],n[t+1])}}}function i(e){for(var n=e.getLineCount()-1;n>=0;n--){var a=e.getLineText(n),i=t.exec(a);if(i)return i[1];if(!a.match(r))break}}function o(e){return"string"===typeof e||null===e}function s(n){return null!==n&&"object"===typeof n&&3===n.version&&"string"===typeof n.file&&"string"===typeof n.mappings&&e.isArray(n.sources)&&e.every(n.sources,e.isString)&&(void 0===n.sourceRoot||null===n.sourceRoot||"string"===typeof n.sourceRoot)&&(void 0===n.sourcesContent||null===n.sourcesContent||e.isArray(n.sourcesContent)&&e.every(n.sourcesContent,o))&&(void 0===n.names||null===n.names||e.isArray(n.names)&&e.every(n.names,e.isString))}function c(e){try{var n=JSON.parse(e);if(s(n))return n}catch(t){}}function l(e){var n,t=!1,r=0,a=0,i=0,o=0,s=0,c=0,l=0;return{get pos(){return r},get error(){return n},get state(){return u(!0,!0)},next:function(){while(!t&&r<e.length){var n=e.charCodeAt(r);if(59!==n){if(44!==n){var p=!1,m=!1;if(i+=v(),_())return d();if(i<0)return f("Invalid generatedCharacter found");if(!g()){if(p=!0,o+=v(),_())return d();if(o<0)return f("Invalid sourceIndex found");if(g())return f("Unsupported Format: No entries after sourceIndex");if(s+=v(),_())return d();if(s<0)return f("Invalid sourceLine found");if(g())return f("Unsupported Format: No entries after sourceLine");if(c+=v(),_())return d();if(c<0)return f("Invalid sourceCharacter found");if(!g()){if(m=!0,l+=v(),_())return d();if(l<0)return f("Invalid nameIndex found");if(!g())return f("Unsupported Error Format: Entries after nameIndex")}}return{value:u(p,m),done:t}}r++}else a++,i=0,r++}return d()}};function u(e,n){return{generatedLine:a,generatedCharacter:i,sourceIndex:e?o:void 0,sourceLine:e?s:void 0,sourceCharacter:e?c:void 0,nameIndex:n?l:void 0}}function d(){return t=!0,{value:void 0,done:!0}}function p(e){void 0===n&&(n=e)}function f(e){return p(e),d()}function _(){return void 0!==n}function g(){return r===e.length||44===e.charCodeAt(r)||59===e.charCodeAt(r)}function v(){for(var n=!0,t=0,a=0;n;r++){if(r>=e.length)return p("Error in decoding base64VLQFormatDecode, past the mapping string"),-1;var i=m(e.charCodeAt(r));if(-1===i)return p("Invalid character in VLQ"),-1;n=0!==(32&i),a|=(31&i)<<t,t+=5}return 0===(1&a)?a>>=1:(a>>=1,a=-a),a}}function u(e,n){return e===n||e.generatedLine===n.generatedLine&&e.generatedCharacter===n.generatedCharacter&&e.sourceIndex===n.sourceIndex&&e.sourceLine===n.sourceLine&&e.sourceCharacter===n.sourceCharacter&&e.nameIndex===n.nameIndex}function d(e){return void 0!==e.sourceIndex&&void 0!==e.sourceLine&&void 0!==e.sourceCharacter}function p(n){return n>=0&&n<26?65+n:n>=26&&n<52?97+n-26:n>=52&&n<62?48+n-52:62===n?43:63===n?47:e.Debug.fail(n+": not a base64 value")}function m(e){return e>=65&&e<=90?e-65:e>=97&&e<=122?e-97+26:e>=48&&e<=57?e-48+52:43===e?62:47===e?63:-1}function f(e){e<0?e=1+(-e<<1):e<<=1;var n="";do{var t=31&e;e>>=5,e>0&&(t|=32),n+=String.fromCharCode(p(t))}while(e>0);return n}function _(e){return void 0!==e.sourceIndex&&void 0!==e.sourcePosition}function g(e,n){return e.generatedPosition===n.generatedPosition&&e.sourceIndex===n.sourceIndex&&e.sourcePosition===n.sourcePosition}function v(n,t){return e.Debug.assert(n.sourceIndex===t.sourceIndex),e.compareValues(n.sourcePosition,t.sourcePosition)}function y(n,t){return e.compareValues(n.generatedPosition,t.generatedPosition)}function h(e){return e.sourcePosition}function b(e){return e.generatedPosition}function E(n,t,r){var a,i,o,s=e.getDirectoryPath(r),c=t.sourceRoot?e.getNormalizedAbsolutePath(t.sourceRoot,s):s,u=e.getNormalizedAbsolutePath(t.file,s),p=n.getSourceFileLike(u),m=t.sources.map(function(n){return e.getNormalizedAbsolutePath(n,c)}),f=e.createMapFromEntries(m.map(function(e,t){return[n.getCanonicalFileName(e),t]}));return{getSourcePosition:A,getGeneratedPosition:C};function E(r){var a,i,o=void 0!==p?e.getPositionOfLineAndCharacter(p,r.generatedLine,r.generatedCharacter,!0):-1;if(d(r)){var s=n.getSourceFileLike(m[r.sourceIndex]);a=t.sources[r.sourceIndex],i=void 0!==s?e.getPositionOfLineAndCharacter(s,r.sourceLine,r.sourceCharacter,!0):-1}return{generatedPosition:o,source:a,sourceIndex:r.sourceIndex,sourcePosition:i,nameIndex:r.nameIndex}}function T(){if(void 0===a){var r=l(t.mappings),i=e.arrayFrom(r,E);void 0!==r.error?(n.log&&n.log("Encountered error while decoding sourcemap: "+r.error),a=e.emptyArray):a=i}return a}function S(n){if(void 0===o){for(var t=[],r=0,a=T();r<a.length;r++){var i=a[r];if(_(i)){var s=t[i.sourceIndex];s||(t[i.sourceIndex]=s=[]),s.push(i)}}o=t.map(function(n){return e.sortAndDeduplicate(n,v,g)})}return o[n]}function x(){if(void 0===i){for(var n=[],t=0,r=T();t<r.length;t++){var a=r[t];n.push(a)}i=e.sortAndDeduplicate(n,y,g)}return i}function C(t){var r=f.get(n.getCanonicalFileName(t.fileName));if(void 0===r)return t;var a=S(r);if(!e.some(a))return t;var i=e.binarySearchKey(a,t.pos,h,e.compareValues);i<0&&(i=~i);var o=a[i];return void 0===o||o.sourceIndex!==r?t:{fileName:u,pos:o.generatedPosition}}function A(n){var t=x();if(!e.some(t))return n;var r=e.binarySearchKey(t,n.pos,b,e.compareValues);r<0&&(r=~r);var a=t[r];return void 0!==a&&_(a)?{fileName:m[a.sourceIndex],pos:a.sourcePosition}:n}}e.getLineInfo=a,e.tryGetSourceMappingURL=i,e.isRawSourceMap=s,e.tryParseRawSourceMap=c,e.decodeMappings=l,e.sameMapping=u,e.isSourceMapping=d,e.createDocumentPositionMapper=E,e.identitySourceMapConsumer={getSourcePosition:e.identity,getGeneratedPosition:e.identity}}(u||(u={})),function(e){function n(n){return n=e.getOriginalNode(n),n?e.getNodeId(n):0}function t(n){return!!n&&(!!e.isNamedImports(n)&&e.some(n.elements,r))}function r(e){return void 0!==e.propertyName&&"default"===e.propertyName.escapedText}function a(n){return t;function t(e){return 279===e.kind?n(e):r(e)}function r(t){return e.createBundle(e.map(t.sourceFiles,n),t.prepends)}}function i(n){if(e.getNamespaceDeclarationNode(n))return!0;var t=n.importClause&&n.importClause.namedBindings;if(!t)return!1;if(!e.isNamedImports(t))return!1;for(var a=0,i=0,o=t.elements;i<o.length;i++){var s=o[i];r(s)&&a++}return a>0&&a!==t.elements.length||!!(t.elements.length-a)&&e.isDefaultImport(n)}function o(n){return!i(n)&&(e.isDefaultImport(n)||!!n.importClause&&e.isNamedImports(n.importClause.namedBindings)&&t(n.importClause.namedBindings))}function s(t,r,a){for(var s,u,d=[],p=e.createMultiMap(),m=[],f=e.createMap(),_=!1,g=!1,v=!1,y=0,h=t.statements;y<h.length;y++){var b=h[y];switch(b.kind){case 249:d.push(b),v=v||i(b)||o(b);break;case 248:259===b.moduleReference.kind&&d.push(b);break;case 255:if(b.moduleSpecifier)b.exportClause?d.push(b):(d.push(b),g=!0);else for(var E=0,T=b.exportClause.elements;E<T.length;E++){var S=T[E];if(!f.get(e.idText(S.name))){var x=S.propertyName||S.name;p.add(e.idText(x),S);var C=r.getReferencedImportDeclaration(x)||r.getReferencedValueDeclaration(x);C&&l(m,n(C),S.name),f.set(e.idText(S.name),!0),s=e.append(s,S.name)}}break;case 254:b.isExportEquals&&!u&&(u=b);break;case 219:if(e.hasModifier(b,1))for(var A=0,D=b.declarationList.declarations;A<D.length;A++){C=D[A];s=c(C,f,s)}break;case 239:if(e.hasModifier(b,1))if(e.hasModifier(b,512))_||(l(m,n(b),e.getDeclarationName(b)),_=!0);else{x=b.name;f.get(e.idText(x))||(l(m,n(b),x),f.set(e.idText(x),!0),s=e.append(s,x))}break;case 240:if(e.hasModifier(b,1))if(e.hasModifier(b,512))_||(l(m,n(b),e.getDeclarationName(b)),_=!0);else{x=b.name;x&&!f.get(e.idText(x))&&(l(m,n(b),x),f.set(e.idText(x),!0),s=e.append(s,x))}break}}var L=e.getOrCreateExternalHelpersModuleNameIfNeeded(t,a,g,v),k=L&&e.createImportDeclaration(void 0,void 0,e.createImportClause(void 0,e.createNamespaceImport(L)),e.createLiteral(e.externalHelpersModuleNameText));return k&&(e.addEmitFlags(k,67108864),d.unshift(k)),{externalImports:d,exportSpecifiers:p,exportEquals:u,hasExportStarsToExportValues:g,exportedBindings:m,exportedNames:s,externalHelpersImportDeclaration:k}}function c(n,t,r){if(e.isBindingPattern(n.name))for(var a=0,i=n.name.elements;a<i.length;a++){var o=i[a];e.isOmittedExpression(o)||(r=c(o,t,r))}else if(!e.isGeneratedIdentifier(n.name)){var s=e.idText(n.name);t.get(s)||(t.set(s,!0),r=e.append(r,n.name))}return r}function l(e,n,t){var r=e[n];return r?r.push(t):e[n]=r=[t],r}function u(n){return e.isStringLiteralLike(n)||8===n.kind||e.isKeyword(n.kind)||e.isIdentifier(n)}function d(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];return function(t){for(var r="",a=0;a<n.length;a++)r+=e[a],r+=t(n[a]);return r+=e[e.length-1],r}}e.getOriginalNodeId=n,e.chainBundle=a,e.getImportNeedsImportStarHelper=i,e.getImportNeedsImportDefaultHelper=o,e.collectExternalModuleInfo=s,e.isSimpleCopiableExpression=u,e.helperString=d}(u||(u={})),function(e){function n(n,r,a,o,s,c){var l,d,m=n;if(e.isDestructuringAssignment(n)){l=n.right;while(e.isEmptyArrayLiteral(n.left)||e.isEmptyObjectLiteral(n.left)){if(!e.isDestructuringAssignment(l))return e.visitNode(l,r,e.isExpression);m=n=l,l=n.right}}var _={context:a,level:o,downlevelIteration:!!a.getCompilerOptions().downlevelIteration,hoistTempVariables:!0,emitExpression:v,emitBindingOrAssignment:y,createArrayBindingOrAssignmentPattern:p,createObjectBindingOrAssignmentPattern:f,createArrayBindingOrAssignmentElement:g,visitor:r};if(l&&(l=e.visitNode(l,r,e.isExpression),e.isIdentifier(l)&&t(n,l.escapedText)?l=u(_,l,!1,m):s?l=u(_,l,!0,m):e.nodeIsSynthesized(n)&&(m=l)),i(_,n,l,m,e.isDestructuringAssignment(n)),l&&s){if(!e.some(d))return l;d.push(l)}return e.aggregateTransformFlags(e.inlineExpressions(d))||e.createOmittedExpression();function v(n){e.setEmitFlags(n,64),e.aggregateTransformFlags(n),d=e.append(d,n)}function y(n,t,a,i){e.Debug.assertNode(n,c?e.isIdentifier:e.isExpression);var o=c?c(n,t,a):e.setTextRange(e.createAssignment(e.visitNode(n,r,e.isExpression),t),a);o.original=i,v(o)}}function t(n,t){var a=e.getTargetOfBindingOrAssignmentElement(n);return e.isBindingOrAssignmentPattern(a)?r(a,t):!!e.isIdentifier(a)&&a.escapedText===t}function r(n,r){for(var a=e.getElementsOfBindingOrAssignmentPattern(n),i=0,o=a;i<o.length;i++){var s=o[i];if(t(s,r))return!0}return!1}function a(n,r,a,o,s,c,l){var p;void 0===c&&(c=!1);var f=[],g=[],v={context:a,level:o,downlevelIteration:!!a.getCompilerOptions().downlevelIteration,hoistTempVariables:c,emitExpression:N,emitBindingOrAssignment:I,createArrayBindingOrAssignmentPattern:d,createObjectBindingOrAssignmentPattern:m,createArrayBindingOrAssignmentElement:_,visitor:r};if(e.isVariableDeclaration(n)){var y=e.getInitializerOfBindingOrAssignmentElement(n);y&&e.isIdentifier(y)&&t(n,y.escapedText)&&(y=u(v,y,!1,y),n=e.updateVariableDeclaration(n,n.name,n.type,y))}if(i(v,n,s,n,l),p){var h=e.createTempVariable(void 0);if(c){var b=e.inlineExpressions(p);p=void 0,I(h,b,void 0,void 0)}else{a.hoistVariableDeclaration(h);var E=e.last(f);E.pendingExpressions=e.append(E.pendingExpressions,e.createAssignment(h,E.value)),e.addRange(E.pendingExpressions,p),E.value=h}}for(var T=0,S=f;T<S.length;T++){var x=S[T],C=x.pendingExpressions,A=x.name,D=(b=x.value,x.location),L=x.original,k=e.createVariableDeclaration(A,void 0,C?e.inlineExpressions(e.append(C,b)):b);k.original=L,e.setTextRange(k,D),e.isIdentifier(A)&&e.setEmitFlags(k,64),e.aggregateTransformFlags(k),g.push(k)}return g;function N(n){p=e.append(p,n)}function I(n,t,r,a){e.Debug.assertNode(n,e.isBindingName),p&&(t=e.inlineExpressions(e.append(p,t)),p=void 0),f.push({pendingExpressions:p,name:n,value:t,location:r,original:a})}}function i(n,t,r,a,i){if(!i){var l=e.visitNode(e.getInitializerOfBindingOrAssignmentElement(t),n.visitor,e.isExpression);l?r=r?c(n,r,l,a):l:r||(r=e.createVoidZero())}var u=e.getTargetOfBindingOrAssignmentElement(t);e.isObjectBindingOrAssignmentPattern(u)?o(n,t,u,r,a):e.isArrayBindingOrAssignmentPattern(u)?s(n,t,u,r,a):n.emitBindingOrAssignment(u,r,a,t)}function o(n,t,r,a,o){var s,c,d=e.getElementsOfBindingOrAssignmentPattern(r),p=d.length;if(1!==p){var m=!e.isDeclarationBindingElement(t)||0!==p;a=u(n,a,m,o)}for(var f=0;f<p;f++){var _=d[f];if(e.getRestIndicatorOfBindingOrAssignmentElement(_)){if(f===p-1){s&&(n.emitBindingOrAssignment(n.createObjectBindingOrAssignmentPattern(s),a,o,r),s=void 0);v=y(n.context,a,d,c,r);i(n,_,v,_)}}else{var g=e.getPropertyNameOfBindingOrAssignmentElement(_);if(!(n.level>=1)||393216&_.transformFlags||393216&e.getTargetOfBindingOrAssignmentElement(_).transformFlags||e.isComputedPropertyName(g)){s&&(n.emitBindingOrAssignment(n.createObjectBindingOrAssignmentPattern(s),a,o,r),s=void 0);var v=l(n,a,g);e.isComputedPropertyName(g)&&(c=e.append(c,v.argumentExpression)),i(n,_,v,_)}else s=e.append(s,_)}}s&&n.emitBindingOrAssignment(n.createObjectBindingOrAssignmentPattern(s),a,o,r)}function s(n,t,r,a,o){var s,c,l=e.getElementsOfBindingOrAssignmentPattern(r),d=l.length;if(n.level<1&&n.downlevelIteration)a=u(n,e.createReadHelper(n.context,a,d>0&&e.getRestIndicatorOfBindingOrAssignmentElement(l[d-1])?void 0:d,o),!1,o);else if(1!==d&&(n.level<1||0===d)||e.every(l,e.isOmittedExpression)){var p=!e.isDeclarationBindingElement(t)||0!==d;a=u(n,a,p,o)}for(var m=0;m<d;m++){var f=l[m];if(n.level>=1)if(262144&f.transformFlags){var _=e.createTempVariable(void 0);n.hoistTempVariables&&n.context.hoistVariableDeclaration(_),c=e.append(c,[_,f]),s=e.append(s,n.createArrayBindingOrAssignmentElement(_))}else s=e.append(s,f);else{if(e.isOmittedExpression(f))continue;if(e.getRestIndicatorOfBindingOrAssignmentElement(f)){if(m===d-1){g=e.createArraySlice(a,m);i(n,f,g,f)}}else{var g=e.createElementAccess(a,m);i(n,f,g,f)}}}if(s&&n.emitBindingOrAssignment(n.createArrayBindingOrAssignmentPattern(s),a,o,r),c)for(var v=0,y=c;v<y.length;v++){var h=y[v],b=h[0];f=h[1];i(n,f,b,f)}}function c(n,t,r,a){return t=u(n,t,!0,a),e.createConditional(e.createTypeCheck(t,"undefined"),r,t)}function l(n,t,r){if(e.isComputedPropertyName(r)){var a=u(n,e.visitNode(r.expression,n.visitor),!1,r);return e.createElementAccess(t,a)}if(e.isStringOrNumericLiteralLike(r)){a=e.getSynthesizedClone(r);return a.text=a.text,e.createElementAccess(t,a)}var i=e.createIdentifier(e.idText(r));return e.createPropertyAccess(t,i)}function u(n,t,r,a){if(e.isIdentifier(t)&&r)return t;var i=e.createTempVariable(void 0);return n.hoistTempVariables?(n.context.hoistVariableDeclaration(i),n.emitExpression(e.setTextRange(e.createAssignment(i,t),a))):n.emitBindingOrAssignment(i,t,a,void 0),i}function d(n){return e.Debug.assertEachNode(n,e.isArrayBindingElement),e.createArrayBindingPattern(n)}function p(n){return e.createArrayLiteral(e.map(n,e.convertToArrayAssignmentElement))}function m(n){return e.Debug.assertEachNode(n,e.isBindingElement),e.createObjectBindingPattern(n)}function f(n){return e.createObjectLiteral(e.map(n,e.convertToObjectAssignmentElement))}function _(n){return e.createBindingElement(void 0,void 0,n)}function g(e){return e}(function(e){e[e["All"]=0]="All",e[e["ObjectRest"]=1]="ObjectRest"})(e.FlattenLevel||(e.FlattenLevel={})),e.flattenDestructuringAssignment=n,e.flattenDestructuringBinding=a;var v={name:"typescript:rest",scoped:!1,text:'\n            var __rest = (this && this.__rest) || function (s, e) {\n                var t = {};\n                for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)\n                    t[p] = s[p];\n                if (s != null && typeof Object.getOwnPropertySymbols === "function")\n                    for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)\n                        t[p[i]] = s[p[i]];\n                return t;\n            };'};function y(n,t,r,a,i){n.requestEmitHelper(v);for(var o=[],s=0,c=0;c<r.length-1;c++){var l=e.getPropertyNameOfBindingOrAssignmentElement(r[c]);if(l)if(e.isComputedPropertyName(l)){var u=a[s];s++,o.push(e.createConditional(e.createTypeCheck(u,"symbol"),u,e.createAdd(u,e.createLiteral(""))))}else o.push(e.createLiteral(l))}return e.createCall(e.getHelperName("__rest"),void 0,[t,e.setTextRange(e.createArrayLiteral(o),i)])}}(u||(u={})),function(e){var n,t,r=!1;function a(n){var t,a,o,c,u,d,p,m,f,_,g=n.startLexicalEnvironment,v=n.resumeLexicalEnvironment,y=n.endLexicalEnvironment,h=n.hoistVariableDeclaration,b=n.getEmitResolver(),E=n.getCompilerOptions(),T=e.getStrictOptionValue(E,"strictNullChecks"),S=e.getEmitScriptTarget(E),x=e.getEmitModuleKind(E),C=n.onEmitNode,A=n.onSubstituteNode;return n.onEmitNode=pt,n.onSubstituteNode=mt,n.enableSubstitution(189),n.enableSubstitution(190),D;function D(e){return 280===e.kind?L(e):k(e)}function L(n){return e.createBundle(n.sourceFiles.map(k),e.mapDefined(n.prepends,function(n){return 282===n.kind?e.createUnparsedSourceFile(n,"js"):n}))}function k(r){if(r.isDeclarationFile)return r;t=r;var a=N(r,U);return e.addEmitHelpers(a,n.readEmitHelpers()),t=void 0,a}function N(e,n){var t=c,r=u,a=d;I(e);var i=n(e);return c!==t&&(d=a),c=t,u=r,i}function I(n){switch(n.kind){case 279:case 246:case 245:case 218:c=n,u=void 0,d=void 0;break;case 240:case 239:if(e.hasModifier(n,2))break;n.name?Nn(n):e.Debug.assert(240===n.kind||e.hasModifier(n,512)),e.isClassDeclaration(n)&&(u=n);break}}function w(e){return N(e,M)}function M(t){return 1&t.transformFlags?j(t):2&t.transformFlags?e.visitEachChild(t,w,n):t}function O(e){return N(e,P)}function P(e){switch(e.kind){case 249:case 248:case 254:case 255:return R(e);default:return M(e)}}function R(t){var r=e.getParseTreeNode(t);if(r!==t)return 2&t.transformFlags?e.visitEachChild(t,w,n):t;switch(t.kind){case 249:return Fn(t);case 248:return qn(t);case 254:return Kn(t);case 255:return jn(t);default:e.Debug.fail("Unhandled ellided statement")}}function F(e){return N(e,G)}function G(t){if(255!==t.kind&&249!==t.kind&&250!==t.kind&&(248!==t.kind||259!==t.moduleReference.kind))return 1&t.transformFlags||e.hasModifier(t,1)?j(t):2&t.transformFlags?e.visitEachChild(t,w,n):t}function V(e){return N(e,B)}function B(n){switch(n.kind){case 157:return;case 154:case 162:case 158:case 159:case 156:return M(n);case 217:return n;default:return e.Debug.failBadSyntaxKind(n)}}function K(n){if(!(2270&e.modifierToFlag(n.kind))&&(!a||85!==n.kind))return n}function j(n){if(e.hasModifier(n,2)&&e.isStatement(n))return e.createNotEmittedStatement(n);switch(n.kind){case 85:case 80:return a?void 0:n;case 115:case 113:case 114:case 118:case 77:case 125:case 133:case 169:case 170:case 171:case 172:case 168:case 163:case 150:case 120:case 143:case 123:case 138:case 135:case 132:case 106:case 139:case 166:case 165:case 167:case 164:case 173:case 174:case 175:case 177:case 178:case 179:case 180:case 181:case 182:case 162:case 152:case 242:return;case 154:return rn(n);case 247:return;case 157:return an(n);case 241:return e.createNotEmittedStatement(n);case 240:return z(n);case 209:return Y(n);case 273:return en(n);case 211:return nn(n);case 156:return on(n);case 158:return cn(n);case 159:return ln(n);case 239:return un(n);case 196:return dn(n);case 197:return pn(n);case 151:return mn(n);case 195:return vn(n);case 194:case 212:return yn(n);case 191:return bn(n);case 192:return En(n);case 193:return Tn(n);case 213:return hn(n);case 243:return xn(n);case 219:return fn(n);case 237:return gn(n);case 244:return On(n);case 248:return qn(n);default:return e.Debug.failBadSyntaxKind(n)}}function U(t){var r=e.getStrictOptionValue(E,"alwaysStrict")&&!(e.isExternalModule(t)&&x>=e.ModuleKind.ES2015)&&!e.isJsonSourceFile(t);return e.updateSourceFileNode(t,e.visitLexicalEnvironment(t.statements,O,n,0,r))}function H(n){if(n.decorators&&n.decorators.length>0)return!0;var t=e.getFirstConstructorWithBody(n);return!!t&&e.forEach(t.parameters,W)}function W(e){return void 0!==e.decorators&&e.decorators.length>0}function q(n,t){var r=0;e.some(t)&&(r|=1);var a=e.getEffectiveBaseTypeNode(n);return a&&96!==e.skipOuterExpressions(a.expression).kind&&(r|=64),H(n)&&(r|=2),e.childIsDecorated(n)&&(r|=4),zn(n)?r|=8:Yn(n)?r|=32:Xn(n)&&(r|=16),S<=1&&7&r&&(r|=128),r}function z(r){var a=_;_=void 0;var i=ie(r,!0),o=q(r,i);128&o&&n.startLexicalEnvironment();var s=r.name||(5&o?e.getGeneratedNameForNode(r):void 0),c=2&o?X(r,s,o):J(r,s,o),l=[c];if(e.some(_)&&l.push(e.createExpressionStatement(e.inlineExpressions(_))),_=a,1&o&&le(l,i,128&o?e.getInternalName(r):e.getLocalName(r)),Se(l,r,!1),Se(l,r,!0),Ae(l,r),128&o){var u=e.createTokenRange(e.skipTrivia(t.text,r.members.end),19),d=e.getInternalName(r),p=e.createPartiallyEmittedExpression(d);p.end=u.end,e.setEmitFlags(p,1536);var m=e.createReturn(p);m.pos=u.pos,e.setEmitFlags(m,1920),l.push(m),e.addStatementsAfterPrologue(l,n.endLexicalEnvironment());var f=e.createImmediatelyInvokedArrowFunction(l);e.setEmitFlags(f,33554432);var g=e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(e.getLocalName(r,!1,!1),void 0,f)]));e.setOriginalNode(g,r),e.setCommentRange(g,r),e.setSourceMapRange(g,e.moveRangePastDecorators(r)),e.startOnNewLine(g),l=[g]}return 8&o?Zn(l,r):(128&o||2&o)&&(32&o?l.push(e.createExportDefault(e.getLocalName(r,!1,!0))):16&o&&l.push(e.createExternalModuleExport(e.getLocalName(r,!1,!0)))),l.length>1&&(l.push(e.createEndOfDeclarationMarker(r)),e.setEmitFlags(c,4194304|e.getEmitFlags(c))),e.singleOrMany(l)}function J(n,t,r){var a=128&r?void 0:e.visitNodes(n.modifiers,K,e.isModifier),i=e.createClassDeclaration(void 0,a,t,void 0,e.visitNodes(n.heritageClauses,w,e.isHeritageClause),Q(n,0!==(64&r))),o=e.getEmitFlags(n);return 1&r&&(o|=32),e.setTextRange(i,n),e.setOriginalNode(i,n),e.setEmitFlags(i,o),i}function X(n,t,r){var a=e.moveRangePastDecorators(n),i=at(n),o=e.getLocalName(n,!1,!0),s=e.visitNodes(n.heritageClauses,w,e.isHeritageClause),c=Q(n,0!==(64&r)),l=e.createClassExpression(void 0,t,void 0,s,c);e.setOriginalNode(l,n),e.setTextRange(l,a);var u=e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(o,void 0,i?e.createAssignment(i,l):l)],1));return e.setOriginalNode(u,n),e.setTextRange(u,a),e.setCommentRange(u,n),u}function Y(n){var t=_;_=void 0;var r=ie(n,!0),a=e.visitNodes(n.heritageClauses,w,e.isHeritageClause),i=Q(n,e.some(a,function(e){return 86===e.token})),o=e.createClassExpression(void 0,n.name,void 0,a,i);if(e.setOriginalNode(o,n),e.setTextRange(o,n),e.some(r)||e.some(_)){var s=[],c=16777216&b.getNodeCheckFlags(n),l=e.createTempVariable(h,!!c);if(c){ct();var u=e.getSynthesizedClone(l);u.autoGenerateFlags&=-9,m[e.getOriginalNodeId(n)]=u}return e.setEmitFlags(o,65536|e.getEmitFlags(o)),s.push(e.startOnNewLine(e.createAssignment(l,o))),e.addRange(s,e.map(_,e.startOnNewLine)),_=t,e.addRange(s,ue(r,l)),s.push(e.startOnNewLine(l)),e.inlineExpressions(s)}return _=t,o}function Q(n,t){var r=[],a=Z(n,t);return a&&r.push(a),e.addRange(r,e.visitNodes(n.members,V,e.isClassElement)),e.setTextRange(e.createNodeArray(r),n.members)}function Z(t,r){var a=e.getFirstConstructorWithBody(t),i=e.forEach(t.members,se),o=a&&4096&a.transformFlags&&e.forEach(a.parameters,re);if(!i&&!o)return e.visitEachChild(a,w,n);var s=$(a),c=ee(t,a,r);return e.startOnNewLine(e.setOriginalNode(e.setTextRange(e.createConstructor(void 0,void 0,s,c),a||t),a))}function $(t){return e.visitParameterList(t&&t.parameters,w,n)||[]}function ee(n,t,r){var a=[],i=0;if(v(),t){i=ne(t,a);var o=te(t);e.addRange(a,e.map(o,ae))}else r&&a.push(e.createExpressionStatement(e.createCall(e.createSuper(),void 0,[e.createSpread(e.createIdentifier("arguments"))])));var s=ie(n,!1);return le(a,s,e.createThis()),t&&e.addRange(a,e.visitNodes(t.body.statements,w,e.isStatement,i)),a=e.mergeLexicalEnvironment(a,y()),e.setTextRange(e.createBlock(e.setTextRange(e.createNodeArray(a),t?t.body.statements:n.members),!0),t?t.body:void 0)}function ne(n,t){if(n.body){var r=n.body.statements,a=e.addPrologue(t,r,!1,w);if(a===r.length)return a;var i=r[a];return 221===i.kind&&e.isSuperCall(i.expression)?(t.push(e.visitNode(i,w,e.isStatement)),a+1):a}return 0}function te(n){return e.filter(n.parameters,re)}function re(n){return e.hasModifier(n,92)&&e.isIdentifier(n.name)}function ae(n){e.Debug.assert(e.isIdentifier(n.name));var t=n.name,r=e.getMutableClone(t);e.setEmitFlags(r,1584);var a=e.getMutableClone(t);return e.setEmitFlags(a,1536),e.startOnNewLine(e.setEmitFlags(e.setTextRange(e.createExpressionStatement(e.createAssignment(e.setTextRange(e.createPropertyAccess(e.createThis(),r),n.name),a)),e.moveRangePos(n,-1)),1536))}function ie(n,t){return e.filter(n.members,t?oe:se)}function oe(e){return ce(e,!0)}function se(e){return ce(e,!1)}function ce(n,t){return 154===n.kind&&t===e.hasModifier(n,32)&&void 0!==n.initializer}function le(n,t,r){for(var a=0,i=t;a<i.length;a++){var o=i[a],s=e.createExpressionStatement(de(o,r));e.setSourceMapRange(s,e.moveRangePastModifiers(o)),e.setCommentRange(s,o),e.setOriginalNode(s,o),n.push(s)}}function ue(n,t){for(var r=[],a=0,i=n;a<i.length;a++){var o=i[a],s=de(o,t);e.startOnNewLine(s),e.setSourceMapRange(s,e.moveRangePastModifiers(o)),e.setCommentRange(s,o),e.setOriginalNode(s,o),r.push(s)}return r}function de(n,t){var r=e.isComputedPropertyName(n.name)&&!Ye(n.name.expression)?e.updateComputedPropertyName(n.name,e.getGeneratedNameForNode(n.name)):n.name,a=e.visitNode(n.initializer,w,e.isExpression),i=e.createMemberAccessForPropertyName(t,r,r);return e.createAssignment(i,a)}function pe(n,t){return e.filter(n.members,t?function(e){return me(e,n)}:function(e){return fe(e,n)})}function me(e,n){return _e(e,!0,n)}function fe(e,n){return _e(e,!1,n)}function _e(n,t,r){return e.nodeOrChildIsDecorated(n,r)&&t===e.hasModifier(n,32)}function ge(e){var n;if(e)for(var t=e.parameters,r=0;r<t.length;r++){var a=t[r];(n||a.decorators)&&(n||(n=new Array(t.length)),n[r]=a.decorators)}return n}function ve(n){var t=n.decorators,r=ge(e.getFirstConstructorWithBody(n));if(t||r)return{decorators:t,parameters:r}}function ye(e,n){switch(n.kind){case 158:case 159:return he(e,n);case 156:return be(n);case 154:return Ee(n);default:return}}function he(n,t){if(t.body){var r=e.getAllAccessorDeclarations(n.members,t),a=r.firstAccessor,i=r.secondAccessor,o=r.setAccessor,s=a.decorators?a:i&&i.decorators?i:void 0;if(s&&t===s){var c=s.decorators,l=ge(o);if(c||l)return{decorators:c,parameters:l}}}}function be(e){if(e.body){var n=e.decorators,t=ge(e);if(n||t)return{decorators:n,parameters:t}}}function Ee(e){var n=e.decorators;if(n)return{decorators:n}}function Te(n,t,r){if(r){var a=[];return e.addRange(a,e.map(r.decorators,Le)),e.addRange(a,e.flatMap(r.parameters,ke)),Ne(n,t,a),a}}function Se(n,t,r){e.addRange(n,e.map(xe(t,r),Qn))}function xe(e,n){for(var t,r=pe(e,n),a=0,i=r;a<i.length;a++){var o=i[a],s=Ce(e,o);s&&(t?t.push(s):t=[s])}return t}function Ce(t,r){var a=ye(t,r),o=Te(r,t,a);if(o){var s=ot(t,r),c=Qe(r,!0),l=S>0?154===r.kind?e.createVoidZero():e.createNull():void 0,u=i(n,o,s,c,l,e.moveRangePastDecorators(r));return e.setEmitFlags(u,1536),u}}function Ae(n,t){var r=De(t);r&&n.push(e.setOriginalNode(e.createExpressionStatement(r),t))}function De(t){var r=ve(t),a=Te(t,t,r);if(a){var o=m&&m[e.getOriginalNodeId(t)],s=e.getLocalName(t,!1,!0),c=i(n,a,s),l=e.createAssignment(s,o?e.createAssignment(o,c):c);return e.setEmitFlags(l,1536),e.setSourceMapRange(l,e.moveRangePastDecorators(t)),l}}function Le(n){return e.visitNode(n.expression,w,e.isExpression)}function ke(t,r){var a;if(t){a=[];for(var i=0,o=t;i<o.length;i++){var s=o[i],c=l(n,Le(s),r,s.expression);e.setEmitFlags(c,1536),a.push(c)}}return a}function Ne(e,n,t){r?we(e,n,t):Ie(e,n,t)}function Ie(e,t,r){E.emitDecoratorMetadata&&(Me(e)&&r.push(s(n,"design:type",Fe(e))),Pe(e)&&r.push(s(n,"design:paramtypes",Ge(e,t))),Oe(e)&&r.push(s(n,"design:returntype",Be(e))))}function we(t,r,a){if(E.emitDecoratorMetadata){var i=void 0;Me(t)&&(i||(i=[])).push(e.createPropertyAssignment("type",e.createArrowFunction(void 0,void 0,[],void 0,e.createToken(37),Fe(t)))),Pe(t)&&(i||(i=[])).push(e.createPropertyAssignment("paramTypes",e.createArrowFunction(void 0,void 0,[],void 0,e.createToken(37),Ge(t,r)))),Oe(t)&&(i||(i=[])).push(e.createPropertyAssignment("returnType",e.createArrowFunction(void 0,void 0,[],void 0,e.createToken(37),Be(t)))),i&&a.push(s(n,"design:typeinfo",e.createObjectLiteral(i,!0)))}}function Me(e){var n=e.kind;return 156===n||158===n||159===n||154===n}function Oe(e){return 156===e.kind}function Pe(n){switch(n.kind){case 240:case 209:return void 0!==e.getFirstConstructorWithBody(n);case 156:case 158:case 159:return!0}return!1}function Re(n){var t=b.getAllAccessorDeclarations(n);return t.setAccessor&&e.getSetAccessorTypeAnnotationNode(t.setAccessor)||t.getAccessor&&e.getEffectiveReturnTypeNode(t.getAccessor)}function Fe(n){switch(n.kind){case 154:case 151:return Ke(n.type);case 159:case 158:return Ke(Re(n));case 240:case 209:case 156:return e.createIdentifier("Function");default:return e.createVoidZero()}}function Ge(n,t){var r=e.isClassLike(n)?e.getFirstConstructorWithBody(n):e.isFunctionLike(n)&&e.nodeIsPresent(n.body)?n:void 0,a=[];if(r)for(var i=Ve(r,t),o=i.length,s=0;s<o;s++){var c=i[s];0===s&&e.isIdentifier(c.name)&&"this"===c.name.escapedText||(c.dotDotDotToken?a.push(Ke(e.getRestParameterElementType(c.type))):a.push(Fe(c)))}return e.createArrayLiteral(a)}function Ve(n,t){if(t&&158===n.kind){var r=e.getAllAccessorDeclarations(t.members,n).setAccessor;if(r)return r.parameters}return n.parameters}function Be(n){return e.isFunctionLike(n)&&n.type?Ke(n.type):e.isAsyncFunction(n)?e.createIdentifier("Promise"):e.createVoidZero()}function Ke(n){if(void 0===n)return e.createIdentifier("Object");switch(n.kind){case 106:case 141:case 96:case 132:return e.createVoidZero();case 177:return Ke(n.type);case 165:case 166:return e.createIdentifier("Function");case 169:case 170:return e.createIdentifier("Array");case 163:case 123:return e.createIdentifier("Boolean");case 138:return e.createIdentifier("String");case 136:return e.createIdentifier("Object");case 182:switch(n.literal.kind){case 10:return e.createIdentifier("String");case 202:case 8:return e.createIdentifier("Number");case 9:return Xe();case 102:case 87:return e.createIdentifier("Boolean");default:return e.Debug.failBadSyntaxKind(n.literal)}case 135:return e.createIdentifier("Number");case 146:return Xe();case 139:return S<2?Je():e.createIdentifier("Symbol");case 164:return Ue(n);case 174:case 173:return je(n.types);case 175:return je([n.trueType,n.falseType]);case 167:case 179:case 180:case 181:case 168:case 120:case 143:case 178:case 183:break;default:return e.Debug.failBadSyntaxKind(n)}return e.createIdentifier("Object")}function je(n){for(var t,r=0,a=n;r<a.length;r++){var i=a[r];while(177===i.kind)i=i.type;if(132!==i.kind&&(T||96!==i.kind&&141!==i.kind)){var o=Ke(i);if(e.isIdentifier(o)&&"Object"===o.escapedText)return o;if(t){if(!e.isIdentifier(t)||!e.isIdentifier(o)||t.escapedText!==o.escapedText)return e.createIdentifier("Object")}else t=o}}return t||e.createVoidZero()}function Ue(n){var t=b.getTypeReferenceSerializationKind(n.typeName,u||c);switch(t){case e.TypeReferenceSerializationKind.Unknown:if(e.findAncestor(n,function(n){return n.parent&&e.isConditionalTypeNode(n.parent)&&(n.parent.trueType===n||n.parent.falseType===n)}))return e.createIdentifier("Object");var r=We(n.typeName),a=e.createTempVariable(h);return e.createConditional(e.createTypeCheck(e.createAssignment(a,r),"function"),a,e.createIdentifier("Object"));case e.TypeReferenceSerializationKind.TypeWithConstructSignatureAndValue:return qe(n.typeName);case e.TypeReferenceSerializationKind.VoidNullableOrNeverType:return e.createVoidZero();case e.TypeReferenceSerializationKind.BigIntLikeType:return Xe();case e.TypeReferenceSerializationKind.BooleanType:return e.createIdentifier("Boolean");case e.TypeReferenceSerializationKind.NumberLikeType:return e.createIdentifier("Number");case e.TypeReferenceSerializationKind.StringLikeType:return e.createIdentifier("String");case e.TypeReferenceSerializationKind.ArrayLikeType:return e.createIdentifier("Array");case e.TypeReferenceSerializationKind.ESSymbolType:return S<2?Je():e.createIdentifier("Symbol");case e.TypeReferenceSerializationKind.TypeWithCallSignature:return e.createIdentifier("Function");case e.TypeReferenceSerializationKind.Promise:return e.createIdentifier("Promise");case e.TypeReferenceSerializationKind.ObjectType:return e.createIdentifier("Object");default:return e.Debug.assertNever(t)}}function He(n,t){return e.createLogicalAnd(e.createStrictInequality(e.createTypeOf(n),e.createLiteral("undefined")),t)}function We(n){if(72===n.kind){var t=qe(n);return He(t,t)}if(72===n.left.kind)return He(qe(n.left),qe(n));var r=We(n.left),a=e.createTempVariable(h);return e.createLogicalAnd(e.createLogicalAnd(r.left,e.createStrictInequality(e.createAssignment(a,r.right),e.createVoidZero())),e.createPropertyAccess(a,n.right))}function qe(n){switch(n.kind){case 72:var t=e.getMutableClone(n);return t.flags&=-9,t.original=void 0,t.parent=e.getParseTreeNode(c),t;case 148:return ze(n)}}function ze(n){return e.createPropertyAccess(qe(n.left),n.right)}function Je(){return e.createConditional(e.createTypeCheck(e.createIdentifier("Symbol"),"function"),e.createIdentifier("Symbol"),e.createIdentifier("Object"))}function Xe(){return S<6?e.createConditional(e.createTypeCheck(e.createIdentifier("BigInt"),"function"),e.createIdentifier("BigInt"),e.createIdentifier("Object")):e.createIdentifier("BigInt")}function Ye(n){return!e.isIdentifier(n)&&e.isSimpleCopiableExpression(n)||e.isWellKnownSymbolSyntactically(n)}function Qe(n,t){var r=n.name;return e.isComputedPropertyName(r)?t&&!Ye(r.expression)?e.getGeneratedNameForNode(r):r.expression:e.isIdentifier(r)?e.createLiteral(e.idText(r)):e.getSynthesizedClone(r)}function Ze(n,t,r){if(e.isComputedPropertyName(n)){var a=e.visitNode(n.expression,w,e.isExpression),i=e.skipPartiallyEmittedExpressions(a),o=Ye(i);if(!o&&t){var s=e.getGeneratedNameForNode(n);return h(s),e.createAssignment(s,a)}return r&&(o||e.isIdentifier(i))?void 0:a}}function $e(n){var t=n.name,r=Ze(t,e.some(n.decorators),!1);return r?(e.some(_)&&(r=e.inlineExpressions(_.concat([r])),_.length=0),e.updateComputedPropertyName(t,r)):t}function en(n){if(86===n.token){var t=e.visitNodes(n.types,w,e.isExpressionWithTypeArguments,0,1);return e.setTextRange(e.createHeritageClause(86,t),n)}}function nn(n){return e.updateExpressionWithTypeArguments(n,void 0,e.visitNode(n.expression,w,e.isLeftHandSideExpression))}function tn(n){return!e.nodeIsMissing(n.body)}function rn(n){var t=Ze(n.name,e.some(n.decorators)||!!n.initializer,!0);t&&!Ye(t)&&(_||(_=[])).push(t)}function an(t){if(tn(t))return e.updateConstructor(t,e.visitNodes(t.decorators,w,e.isDecorator),e.visitNodes(t.modifiers,w,e.isModifier),e.visitParameterList(t.parameters,w,n),e.visitFunctionBody(t.body,w,n))}function on(t){if(tn(t)){var r=e.updateMethod(t,void 0,e.visitNodes(t.modifiers,K,e.isModifier),t.asteriskToken,$e(t),void 0,void 0,e.visitParameterList(t.parameters,w,n),void 0,e.visitFunctionBody(t.body,w,n));return r!==t&&(e.setCommentRange(r,t),e.setSourceMapRange(r,e.moveRangePastDecorators(t))),r}}function sn(n){return!(e.nodeIsMissing(n.body)&&e.hasModifier(n,128))}function cn(t){if(sn(t)){var r=e.updateGetAccessor(t,void 0,e.visitNodes(t.modifiers,K,e.isModifier),$e(t),e.visitParameterList(t.parameters,w,n),void 0,e.visitFunctionBody(t.body,w,n)||e.createBlock([]));return r!==t&&(e.setCommentRange(r,t),e.setSourceMapRange(r,e.moveRangePastDecorators(t))),r}}function ln(t){if(sn(t)){var r=e.updateSetAccessor(t,void 0,e.visitNodes(t.modifiers,K,e.isModifier),$e(t),e.visitParameterList(t.parameters,w,n),e.visitFunctionBody(t.body,w,n)||e.createBlock([]));return r!==t&&(e.setCommentRange(r,t),e.setSourceMapRange(r,e.moveRangePastDecorators(t))),r}}function un(t){if(!tn(t))return e.createNotEmittedStatement(t);var r=e.updateFunctionDeclaration(t,void 0,e.visitNodes(t.modifiers,K,e.isModifier),t.asteriskToken,t.name,void 0,e.visitParameterList(t.parameters,w,n),void 0,e.visitFunctionBody(t.body,w,n)||e.createBlock([]));if(zn(t)){var a=[r];return Zn(a,t),a}return r}function dn(t){if(!tn(t))return e.createOmittedExpression();var r=e.updateFunctionExpression(t,e.visitNodes(t.modifiers,K,e.isModifier),t.asteriskToken,t.name,void 0,e.visitParameterList(t.parameters,w,n),void 0,e.visitFunctionBody(t.body,w,n)||e.createBlock([]));return r}function pn(t){var r=e.updateArrowFunction(t,e.visitNodes(t.modifiers,K,e.isModifier),void 0,e.visitParameterList(t.parameters,w,n),void 0,t.equalsGreaterThanToken,e.visitFunctionBody(t.body,w,n));return r}function mn(n){if(!e.parameterIsThisKeyword(n)){var t=e.createParameter(void 0,void 0,n.dotDotDotToken,e.visitNode(n.name,w,e.isBindingName),void 0,void 0,e.visitNode(n.initializer,w,e.isExpression));return e.setOriginalNode(t,n),e.setTextRange(t,e.moveRangePastModifiers(n)),e.setCommentRange(t,n),e.setSourceMapRange(t,e.moveRangePastModifiers(n)),e.setEmitFlags(t.name,32),t}}function fn(t){if(zn(t)){var r=e.getInitializedVariables(t.declarationList);if(0===r.length)return;return e.setTextRange(e.createExpressionStatement(e.inlineExpressions(e.map(r,_n))),t)}return e.visitEachChild(t,w,n)}function _n(t){var r=t.name;return e.isBindingPattern(r)?e.flattenDestructuringAssignment(t,w,n,0,!1,et):e.setTextRange(e.createAssignment(nt(r),e.visitNode(t.initializer,w,e.isExpression)),t)}function gn(n){return e.updateVariableDeclaration(n,e.visitNode(n.name,w,e.isBindingName),void 0,e.visitNode(n.initializer,w,e.isExpression))}function vn(r){var a=e.skipOuterExpressions(r.expression,-3);if(e.isAssertionExpression(a)){var i=e.visitNode(r.expression,w,e.isExpression);return e.length(e.getLeadingCommentRangesOfNode(i,t))?e.updateParen(r,i):e.createPartiallyEmittedExpression(i,r)}return e.visitEachChild(r,w,n)}function yn(n){var t=e.visitNode(n.expression,w,e.isExpression);return e.createPartiallyEmittedExpression(t,n)}function hn(n){var t=e.visitNode(n.expression,w,e.isLeftHandSideExpression);return e.createPartiallyEmittedExpression(t,n)}function bn(n){return e.updateCall(n,e.visitNode(n.expression,w,e.isExpression),void 0,e.visitNodes(n.arguments,w,e.isExpression))}function En(n){return e.updateNew(n,e.visitNode(n.expression,w,e.isExpression),void 0,e.visitNodes(n.arguments,w,e.isExpression))}function Tn(n){return e.updateTaggedTemplate(n,e.visitNode(n.tag,w,e.isExpression),void 0,e.visitNode(n.template,w,e.isExpression))}function Sn(n){return!e.isEnumConst(n)||E.preserveConstEnums||E.isolatedModules}function xn(n){if(Sn(n)){var r=[],a=2,i=Mn(r,n);i&&(x===e.ModuleKind.System&&c===t||(a|=512));var s=tt(n),l=rt(n),u=e.hasModifier(n,1)?e.getExternalModuleOrNamespaceExportName(o,n,!1,!0):e.getLocalName(n,!1,!0),d=e.createLogicalOr(u,e.createAssignment(u,e.createObjectLiteral()));if(kn(n)){var p=e.getLocalName(n,!1,!0);d=e.createAssignment(p,d)}var m=e.createExpressionStatement(e.createCall(e.createFunctionExpression(void 0,void 0,void 0,void 0,[e.createParameter(void 0,void 0,void 0,s)],void 0,Cn(n,l)),void 0,[d]));return e.setOriginalNode(m,n),i&&(e.setSyntheticLeadingComments(m,void 0),e.setSyntheticTrailingComments(m,void 0)),e.setTextRange(m,n),e.addEmitFlags(m,a),r.push(m),r.push(e.createEndOfDeclarationMarker(n)),r}}function Cn(n,t){var r=o;o=t;var a=[];g();var i=e.map(n.members,An);return e.addStatementsAfterPrologue(a,y()),e.addRange(a,i),o=r,e.createBlock(e.setTextRange(e.createNodeArray(a),n.members),!0)}function An(n){var t=Qe(n,!1),r=Dn(n),a=e.createAssignment(e.createElementAccess(o,t),r),i=10===r.kind?a:e.createAssignment(e.createElementAccess(o,a),t);return e.setTextRange(e.createExpressionStatement(e.setTextRange(i,n)),n)}function Dn(n){var t=b.getConstantValue(n);return void 0!==t?e.createLiteral(t):(st(),n.initializer?e.visitNode(n.initializer,w,e.isExpression):e.createVoidZero())}function Ln(n){return e.isInstantiatedModule(n,!!E.preserveConstEnums||!!E.isolatedModules)}function kn(n){return zn(n)||Jn(n)&&x!==e.ModuleKind.ES2015&&x!==e.ModuleKind.ESNext&&x!==e.ModuleKind.System}function Nn(n){d||(d=e.createUnderscoreEscapedMap());var t=wn(n);d.has(t)||d.set(t,n)}function In(e){if(d){var n=wn(e);return d.get(n)===e}return!0}function wn(n){return e.Debug.assertNode(n.name,e.isIdentifier),n.name.escapedText}function Mn(n,t){var r=e.createVariableStatement(e.visitNodes(t.modifiers,K,e.isModifier),e.createVariableDeclarationList([e.createVariableDeclaration(e.getLocalName(t,!1,!0))],279===c.kind?0:1));if(e.setOriginalNode(r,t),Nn(t),In(t))return 243===t.kind?e.setSourceMapRange(r.declarationList,t):e.setSourceMapRange(r,t),e.setCommentRange(r,t),e.addEmitFlags(r,4195328),n.push(r),!0;var a=e.createMergeDeclarationMarker(r);return e.setEmitFlags(a,4195840),n.push(a),!1}function On(n){if(!Ln(n))return e.createNotEmittedStatement(n);e.Debug.assertNode(n.name,e.isIdentifier,"A TypeScript namespace should have an Identifier name."),lt();var r=[],a=2,i=Mn(r,n);i&&(x===e.ModuleKind.System&&c===t||(a|=512));var s=tt(n),l=rt(n),u=e.hasModifier(n,1)?e.getExternalModuleOrNamespaceExportName(o,n,!1,!0):e.getLocalName(n,!1,!0),d=e.createLogicalOr(u,e.createAssignment(u,e.createObjectLiteral()));if(kn(n)){var p=e.getLocalName(n,!1,!0);d=e.createAssignment(p,d)}var m=e.createExpressionStatement(e.createCall(e.createFunctionExpression(void 0,void 0,void 0,void 0,[e.createParameter(void 0,void 0,void 0,s)],void 0,Pn(n,l)),void 0,[d]));return e.setOriginalNode(m,n),i&&(e.setSyntheticLeadingComments(m,void 0),e.setSyntheticTrailingComments(m,void 0)),e.setTextRange(m,n),e.addEmitFlags(m,a),r.push(m),r.push(e.createEndOfDeclarationMarker(n)),r}function Pn(n,t){var r=o,i=a,s=d;o=t,a=n,d=void 0;var c,l,u=[];g();var p=n.body;if(245===p.kind)N(p,function(n){return e.addRange(u,e.visitNodes(n.statements,F,e.isStatement))}),c=p.statements,l=p;else{var m=On(p);m&&(e.isArray(m)?e.addRange(u,m):u.push(m));var f=Rn(n).body;c=e.moveRangePos(f.statements,-1)}e.addStatementsAfterPrologue(u,y()),o=r,a=i,d=s;var _=e.createBlock(e.setTextRange(e.createNodeArray(u),c),!0);return e.setTextRange(_,l),245!==p.kind&&e.setEmitFlags(_,1536|e.getEmitFlags(_)),_}function Rn(e){if(244===e.body.kind){var n=Rn(e.body);return n||e.body}}function Fn(n){if(!n.importClause)return n;var t=e.visitNode(n.importClause,Gn,e.isImportClause);return t?e.updateImportDeclaration(n,void 0,void 0,t,n.moduleSpecifier):void 0}function Gn(n){var t=b.isReferencedAliasDeclaration(n)?n.name:void 0,r=e.visitNode(n.namedBindings,Vn,e.isNamedImportBindings);return t||r?e.updateImportClause(n,t,r):void 0}function Vn(n){if(251===n.kind)return b.isReferencedAliasDeclaration(n)?n:void 0;var t=e.visitNodes(n.elements,Bn,e.isImportSpecifier);return e.some(t)?e.updateNamedImports(n,t):void 0}function Bn(e){return b.isReferencedAliasDeclaration(e)?e:void 0}function Kn(t){return b.isValueAliasDeclaration(t)?e.visitEachChild(t,w,n):void 0}function jn(n){if(!n.exportClause)return E.isolatedModules||b.moduleExportsSomeValue(n.moduleSpecifier)?n:void 0;if(b.isValueAliasDeclaration(n)){var t=e.visitNode(n.exportClause,Un,e.isNamedExports);return t?e.updateExportDeclaration(n,void 0,void 0,t,n.moduleSpecifier):void 0}}function Un(n){var t=e.visitNodes(n.elements,Hn,e.isExportSpecifier);return e.some(t)?e.updateNamedExports(n,t):void 0}function Hn(e){return b.isValueAliasDeclaration(e)?e:void 0}function Wn(n){return b.isReferencedAliasDeclaration(n)||!e.isExternalModule(t)&&b.isTopLevelValueImportEqualsWithEntityName(n)}function qn(t){if(e.isExternalModuleImportEqualsDeclaration(t))return b.isReferencedAliasDeclaration(t)?e.visitEachChild(t,w,n):void 0;if(Wn(t)){var r=e.createExpressionFromEntityName(t.moduleReference);return e.setEmitFlags(r,3584),Xn(t)||!zn(t)?e.setOriginalNode(e.setTextRange(e.createVariableStatement(e.visitNodes(t.modifiers,K,e.isModifier),e.createVariableDeclarationList([e.setOriginalNode(e.createVariableDeclaration(t.name,void 0,r),t)])),t),t):e.setOriginalNode($n(t.name,r,t),t)}}function zn(n){return void 0!==a&&e.hasModifier(n,1)}function Jn(n){return void 0===a&&e.hasModifier(n,1)}function Xn(n){return Jn(n)&&!e.hasModifier(n,512)}function Yn(n){return Jn(n)&&e.hasModifier(n,512)}function Qn(n){return e.createExpressionStatement(n)}function Zn(n,t){var r=e.createAssignment(e.getExternalModuleOrNamespaceExportName(o,t,!1,!0),e.getLocalName(t));e.setSourceMapRange(r,e.createRange(t.name?t.name.pos:t.pos,t.end));var a=e.createExpressionStatement(r);e.setSourceMapRange(a,e.createRange(-1,t.end)),n.push(a)}function $n(n,t,r){return e.setTextRange(e.createExpressionStatement(e.createAssignment(e.getNamespaceMemberName(o,n,!1,!0),t)),r)}function et(n,t,r){return e.setTextRange(e.createAssignment(nt(n),t),r)}function nt(n){return e.getNamespaceMemberName(o,n,!1,!0)}function tt(n){var t=e.getGeneratedNameForNode(n);return e.setSourceMapRange(t,n.name),t}function rt(n){return e.getGeneratedNameForNode(n)}function at(n){if(16777216&b.getNodeCheckFlags(n)){ct();var t=e.createUniqueName(n.name&&!e.isGeneratedIdentifier(n.name)?e.idText(n.name):"default");return m[e.getOriginalNodeId(n)]=t,h(t),t}}function it(n){return e.createPropertyAccess(e.getDeclarationName(n),"prototype")}function ot(n,t){return e.hasModifier(t,32)?e.getDeclarationName(n):it(n)}function st(){0===(8&p)&&(p|=8,n.enableSubstitution(72))}function ct(){0===(1&p)&&(p|=1,n.enableSubstitution(72),m=[])}function lt(){0===(2&p)&&(p|=2,n.enableSubstitution(72),n.enableSubstitution(276),n.enableEmitNotification(244))}function ut(n){return 244===e.getOriginalNode(n).kind}function dt(n){return 243===e.getOriginalNode(n).kind}function pt(n,r,a){var i=f,o=t;e.isSourceFile(r)&&(t=r),2&p&&ut(r)&&(f|=2),8&p&&dt(r)&&(f|=8),C(n,r,a),f=i,t=o}function mt(n,t){return t=A(n,t),1===n?_t(t):e.isShorthandPropertyAssignment(t)?ft(t):t}function ft(n){if(2&p){var t=n.name,r=yt(t);if(r){if(n.objectAssignmentInitializer){var a=e.createAssignment(r,n.objectAssignmentInitializer);return e.setTextRange(e.createPropertyAssignment(t,a),n)}return e.setTextRange(e.createPropertyAssignment(t,r),n)}}return n}function _t(e){switch(e.kind){case 72:return gt(e);case 189:return ht(e);case 190:return bt(e)}return e}function gt(e){return vt(e)||yt(e)||e}function vt(n){if(1&p&&33554432&b.getNodeCheckFlags(n)){var t=b.getReferencedValueDeclaration(n);if(t){var r=m[t.id];if(r){var a=e.getSynthesizedClone(r);return e.setSourceMapRange(a,n),e.setCommentRange(a,n),a}}}}function yt(n){if(p&f&&!e.isGeneratedIdentifier(n)&&!e.isLocalName(n)){var t=b.getReferencedExportContainer(n,!1);if(t&&279!==t.kind){var r=2&f&&244===t.kind||8&f&&243===t.kind;if(r)return e.setTextRange(e.createPropertyAccess(e.getGeneratedNameForNode(t),n),n)}}}function ht(e){return Et(e)}function bt(e){return Et(e)}function Et(n){var t=Tt(n);if(void 0!==t){e.setConstantValue(n,t);var r=e.createLiteral(t);if(!E.removeComments){var a=e.isPropertyAccessExpression(n)?e.declarationNameToString(n.name):e.getTextOfNode(n.argumentExpression);e.addSyntheticTrailingComment(r,3," "+a+" ")}return r}return n}function Tt(n){if(!E.isolatedModules)return e.isPropertyAccessExpression(n)||e.isElementAccessExpression(n)?b.getConstantValue(n):void 0}}function i(n,t,r,a,i,s){var c=[];return c.push(e.createArrayLiteral(t,!0)),c.push(r),a&&(c.push(a),i&&c.push(i)),n.requestEmitHelper(o),e.setTextRange(e.createCall(e.getHelperName("__decorate"),void 0,c),s)}(function(e){e[e["ClassAliases"]=1]="ClassAliases",e[e["NamespaceExports"]=2]="NamespaceExports",e[e["NonQualifiedEnumMembers"]=8]="NonQualifiedEnumMembers"})(n||(n={})),function(e){e[e["None"]=0]="None",e[e["HasStaticInitializedProperties"]=1]="HasStaticInitializedProperties",e[e["HasConstructorDecorators"]=2]="HasConstructorDecorators",e[e["HasMemberDecorators"]=4]="HasMemberDecorators",e[e["IsExportOfNamespace"]=8]="IsExportOfNamespace",e[e["IsNamedExternalExport"]=16]="IsNamedExternalExport",e[e["IsDefaultExternalExport"]=32]="IsDefaultExternalExport",e[e["IsDerivedClass"]=64]="IsDerivedClass",e[e["UseImmediatelyInvokedFunctionExpression"]=128]="UseImmediatelyInvokedFunctionExpression",e[e["HasAnyDecorators"]=6]="HasAnyDecorators",e[e["NeedsName"]=5]="NeedsName",e[e["MayNeedImmediatelyInvokedFunctionExpression"]=7]="MayNeedImmediatelyInvokedFunctionExpression",e[e["IsExported"]=56]="IsExported"}(t||(t={})),e.transformTypeScript=a;var o={name:"typescript:decorate",scoped:!1,priority:2,text:'\n            var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {\n                var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;\n                if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);\n                else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;\n                return c > 3 && r && Object.defineProperty(target, key, r), r;\n            };'};function s(n,t,r){return n.requestEmitHelper(c),e.createCall(e.getHelperName("__metadata"),void 0,[e.createLiteral(t),r])}var c={name:"typescript:metadata",scoped:!1,priority:3,text:'\n            var __metadata = (this && this.__metadata) || function (k, v) {\n                if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);\n            };'};function l(n,t,r,a){return n.requestEmitHelper(u),e.setTextRange(e.createCall(e.getHelperName("__param"),void 0,[e.createLiteral(r),t]),a)}var u={name:"typescript:param",scoped:!1,priority:4,text:"\n            var __param = (this && this.__param) || function (paramIndex, decorator) {\n                return function (target, key) { decorator(target, key, paramIndex); }\n            };"}}(u||(u={})),function(e){var n;function t(n){var t,a,o,s,c=n.resumeLexicalEnvironment,l=n.endLexicalEnvironment,u=n.hoistVariableDeclaration,d=n.getEmitResolver(),p=n.getCompilerOptions(),m=e.getEmitScriptTarget(p),f=0,_=[],g=n.onEmitNode,v=n.onSubstituteNode;return n.onEmitNode=j,n.onSubstituteNode=U,e.chainBundle(y);function y(t){if(t.isDeclarationFile)return t;var r=e.visitEachChild(t,h,n);return e.addEmitHelpers(r,n.readEmitHelpers()),r}function h(t){if(0===(16&t.transformFlags))return t;switch(t.kind){case 121:return;case 201:return A(t);case 156:return D(t);case 239:return L(t);case 196:return k(t);case 197:return N(t);case 189:return o&&e.isPropertyAccessExpression(t)&&98===t.expression.kind&&o.set(t.name.escapedText,!0),e.visitEachChild(t,h,n);case 190:return o&&98===t.expression.kind&&(s=!0),e.visitEachChild(t,h,n);default:return e.visitEachChild(t,h,n)}}function b(t){if(e.isNodeWithPossibleHoistedDeclaration(t))switch(t.kind){case 219:return T(t);case 225:return C(t);case 226:return S(t);case 227:return x(t);case 274:return E(t);case 218:case 232:case 246:case 271:case 272:case 235:case 223:case 224:case 222:case 231:case 233:return e.visitEachChild(t,b,n);default:return e.Debug.assertNever(t,"Unhandled node.")}return h(t)}function E(t){var r,i=e.createUnderscoreEscapedMap();if(I(t.variableDeclaration,i),i.forEach(function(n,t){a.has(t)&&(r||(r=e.cloneMap(a)),r.delete(t))}),r){var o=a;a=r;var s=e.visitEachChild(t,b,n);return a=o,s}return e.visitEachChild(t,b,n)}function T(t){if(w(t.declarationList)){var r=M(t.declarationList,!1);return r?e.createExpressionStatement(r):void 0}return e.visitEachChild(t,h,n)}function S(n){return e.updateForIn(n,w(n.initializer)?M(n.initializer,!0):e.visitNode(n.initializer,h,e.isForInitializer),e.visitNode(n.expression,h,e.isExpression),e.visitNode(n.statement,b,e.isStatement,e.liftToBlock))}function x(n){return e.updateForOf(n,e.visitNode(n.awaitModifier,h,e.isToken),w(n.initializer)?M(n.initializer,!0):e.visitNode(n.initializer,h,e.isForInitializer),e.visitNode(n.expression,h,e.isExpression),e.visitNode(n.statement,b,e.isStatement,e.liftToBlock))}function C(n){var t=n.initializer;return e.updateFor(n,w(t)?M(t,!1):e.visitNode(n.initializer,h,e.isForInitializer),e.visitNode(n.condition,h,e.isExpression),e.visitNode(n.incrementor,h,e.isExpression),e.visitNode(n.statement,b,e.isStatement,e.liftToBlock))}function A(n){return e.setOriginalNode(e.setTextRange(e.createYield(void 0,e.visitNode(n.expression,h,e.isExpression)),n),n)}function D(t){return e.updateMethod(t,void 0,e.visitNodes(t.modifiers,h,e.isModifier),t.asteriskToken,t.name,void 0,void 0,e.visitParameterList(t.parameters,h,n),void 0,2&e.getFunctionFlags(t)?G(t):e.visitFunctionBody(t.body,h,n))}function L(t){return e.updateFunctionDeclaration(t,void 0,e.visitNodes(t.modifiers,h,e.isModifier),t.asteriskToken,t.name,void 0,e.visitParameterList(t.parameters,h,n),void 0,2&e.getFunctionFlags(t)?G(t):e.visitFunctionBody(t.body,h,n))}function k(t){return e.updateFunctionExpression(t,e.visitNodes(t.modifiers,h,e.isModifier),t.asteriskToken,t.name,void 0,e.visitParameterList(t.parameters,h,n),void 0,2&e.getFunctionFlags(t)?G(t):e.visitFunctionBody(t.body,h,n))}function N(t){return e.updateArrowFunction(t,e.visitNodes(t.modifiers,h,e.isModifier),void 0,e.visitParameterList(t.parameters,h,n),void 0,t.equalsGreaterThanToken,2&e.getFunctionFlags(t)?G(t):e.visitFunctionBody(t.body,h,n))}function I(n,t){var r=n.name;if(e.isIdentifier(r))t.set(r.escapedText,!0);else for(var a=0,i=r.elements;a<i.length;a++){var o=i[a];e.isOmittedExpression(o)||I(o,t)}}function w(n){return!!n&&e.isVariableDeclarationList(n)&&!(3&n.flags)&&n.declarations.some(F)}function M(n,t){O(n);var r=e.getInitializedVariables(n);return 0===r.length?t?e.visitNode(e.convertToAssignmentElementTarget(n.declarations[0].name),h,e.isExpression):void 0:e.inlineExpressions(e.map(r,R))}function O(n){e.forEach(n.declarations,P)}function P(n){var t=n.name;if(e.isIdentifier(t))u(t);else for(var r=0,a=t.elements;r<a.length;r++){var i=a[r];e.isOmittedExpression(i)||P(i)}}function R(n){var t=e.setSourceMapRange(e.createAssignment(e.convertToAssignmentElementTarget(n.name),n.initializer),n);return e.visitNode(t,h,e.isExpression)}function F(n){var t=n.name;if(e.isIdentifier(t))return a.has(t.escapedText);for(var r=0,i=t.elements;r<i.length;r++){var o=i[r];if(!e.isOmittedExpression(o)&&F(o))return!0}return!1}function G(t){c();var u=e.getOriginalNode(t,e.isFunctionLike),p=u.type,f=m<2?B(p):void 0,g=197===t.kind,v=0!==(8192&d.getNodeCheckFlags(t)),y=a;a=e.createUnderscoreEscapedMap();for(var b=0,E=t.parameters;b<E.length;b++){var T=E[b];I(T,a)}var S,x=o,C=s;if(o=e.createUnderscoreEscapedMap(),s=!1,g){var A=i(n,v,f,V(t.body)),D=l();if(e.some(D)){M=e.convertToFunctionBody(A);S=e.updateBlock(M,e.setTextRange(e.createNodeArray(e.concatenate(D,M.statements)),M.statements))}else S=A}else{var L=[],k=e.addPrologue(L,t.body.statements,!1,h);L.push(e.createReturn(i(n,v,f,V(t.body,k)))),e.addStatementsAfterPrologue(L,l());var N=m>=2&&6144&d.getNodeCheckFlags(t);if(N){K();var w=r(d,t,o);_[e.getNodeId(w)]=!0,e.addStatementsAfterPrologue(L,[w])}var M=e.createBlock(L,!0);e.setTextRange(M,t.body),N&&s&&(4096&d.getNodeCheckFlags(t)?e.addEmitHelper(M,e.advancedAsyncSuperHelper):2048&d.getNodeCheckFlags(t)&&e.addEmitHelper(M,e.asyncSuperHelper)),S=M}return a=y,o=x,s=C,S}function V(n,t){return e.isBlock(n)?e.updateBlock(n,e.visitNodes(n.statements,b,e.isStatement,t)):e.convertToFunctionBody(e.visitNode(n,b,e.isConciseBody))}function B(n){var t=n&&e.getEntityNameFromTypeNode(n);if(t&&e.isEntityName(t)){var r=d.getTypeReferenceSerializationKind(t);if(r===e.TypeReferenceSerializationKind.TypeWithConstructSignatureAndValue||r===e.TypeReferenceSerializationKind.Unknown)return t}}function K(){0===(1&t)&&(t|=1,n.enableSubstitution(191),n.enableSubstitution(189),n.enableSubstitution(190),n.enableEmitNotification(240),n.enableEmitNotification(156),n.enableEmitNotification(158),n.enableEmitNotification(159),n.enableEmitNotification(157),n.enableEmitNotification(219))}function j(n,r,a){if(1&t&&J(r)){var i=6144&d.getNodeCheckFlags(r);if(i!==f){var o=f;return f=i,g(n,r,a),void(f=o)}}else if(t&&_[e.getNodeId(r)]){o=f;return f=0,g(n,r,a),void(f=o)}g(n,r,a)}function U(e,n){return n=v(e,n),1===e&&f?H(n):n}function H(e){switch(e.kind){case 189:return W(e);case 190:return q(e);case 191:return z(e)}return e}function W(n){return 98===n.expression.kind?e.setTextRange(e.createPropertyAccess(e.createFileLevelUniqueName("_super"),n.name),n):n}function q(e){return 98===e.expression.kind?X(e.argumentExpression,e):e}function z(n){var t=n.expression;if(e.isSuperProperty(t)){var r=e.isPropertyAccessExpression(t)?W(t):q(t);return e.createCall(e.createPropertyAccess(r,"call"),void 0,[e.createThis()].concat(n.arguments))}return n}function J(e){var n=e.kind;return 240===n||157===n||156===n||158===n||159===n}function X(n,t){return 4096&f?e.setTextRange(e.createPropertyAccess(e.createCall(e.createFileLevelUniqueName("_superIndex"),void 0,[n]),"value"),t):e.setTextRange(e.createCall(e.createFileLevelUniqueName("_superIndex"),void 0,[n]),t)}}function r(n,t,r){var a=0!==(4096&n.getNodeCheckFlags(t)),i=[];return r.forEach(function(n,t){var r=e.unescapeLeadingUnderscores(t),o=[];o.push(e.createPropertyAssignment("get",e.createArrowFunction(void 0,void 0,[],void 0,void 0,e.createPropertyAccess(e.createSuper(),r)))),a&&o.push(e.createPropertyAssignment("set",e.createArrowFunction(void 0,void 0,[e.createParameter(void 0,void 0,void 0,"v",void 0,void 0,void 0)],void 0,void 0,e.createAssignment(e.createPropertyAccess(e.createSuper(),r),e.createIdentifier("v"))))),i.push(e.createPropertyAssignment(r,e.createObjectLiteral(o)))}),e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(e.createFileLevelUniqueName("_super"),void 0,e.createCall(e.createPropertyAccess(e.createIdentifier("Object"),"create"),void 0,[e.createNull(),e.createObjectLiteral(i,!0)]))],2))}(function(e){e[e["AsyncMethodsWithSuper"]=1]="AsyncMethodsWithSuper"})(n||(n={})),e.transformES2017=t,e.createSuperAccessVariableStatement=r;var a={name:"typescript:awaiter",scoped:!1,priority:5,text:'\n            var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {\n                return new (P || (P = Promise))(function (resolve, reject) {\n                    function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }\n                    function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }\n                    function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }\n                    step((generator = generator.apply(thisArg, _arguments || [])).next());\n                });\n            };'};function i(n,t,r,i){n.requestEmitHelper(a);var o=e.createFunctionExpression(void 0,e.createToken(40),void 0,void 0,[],void 0,i);return(o.emitNode||(o.emitNode={})).flags|=786432,e.createCall(e.getHelperName("__awaiter"),void 0,[e.createThis(),t?e.createIdentifier("arguments"):e.createVoidZero(),r?e.createExpressionFromEntityName(r):e.createVoidZero(),o])}e.asyncSuperHelper={name:"typescript:async-super",scoped:!0,text:e.helperString(c(["\n            const "," = name => super[name];"],["\n            const "," = name => super[name];"]),"_superIndex")},e.advancedAsyncSuperHelper={name:"typescript:advanced-async-super",scoped:!0,text:e.helperString(c(["\n            const "," = (function (geti, seti) {\n                const cache = Object.create(null);\n                return name => cache[name] || (cache[name] = { get value() { return geti(name); }, set value(v) { seti(name, v); } });\n            })(name => super[name], (name, value) => super[name] = value);"],["\n            const "," = (function (geti, seti) {\n                const cache = Object.create(null);\n                return name => cache[name] || (cache[name] = { get value() { return geti(name); }, set value(v) { seti(name, v); } });\n            })(name => super[name], (name, value) => super[name] = value);"]),"_superIndex")}}(u||(u={})),function(e){var n;function t(n){var t=n.resumeLexicalEnvironment,r=n.endLexicalEnvironment,i=n.hoistVariableDeclaration,s=n.getEmitResolver(),l=n.getCompilerOptions(),d=e.getEmitScriptTarget(l),m=n.onEmitNode;n.onEmitNode=te;var f,_,g=n.onSubstituteNode;n.onSubstituteNode=re;var v,y,h=0,b=[];return e.chainBundle(E);function E(t){if(t.isDeclarationFile)return t;var r=e.visitEachChild(t,T,n);return e.addEmitHelpers(r,n.readEmitHelpers()),r}function T(e){return C(e,!1)}function S(e){return C(e,!0)}function x(e){if(121!==e.kind)return e}function C(t,r){if(0===(8&t.transformFlags))return t;switch(t.kind){case 201:return A(t);case 207:return D(t);case 230:return L(t);case 233:return k(t);case 188:return I(t);case 204:return P(t,r);case 237:return R(t);case 227:return V(t,void 0);case 225:return F(t);case 200:return G(t);case 157:return W(t);case 156:return J(t);case 158:return q(t);case 159:return z(t);case 239:return X(t);case 196:return Q(t);case 197:return Y(t);case 151:return H(t);case 221:return w(t);case 195:return M(t,r);case 274:return O(t);case 189:return v&&e.isPropertyAccessExpression(t)&&98===t.expression.kind&&v.set(t.name.escapedText,!0),e.visitEachChild(t,T,n);case 190:return v&&98===t.expression.kind&&(y=!0),e.visitEachChild(t,T,n);default:return e.visitEachChild(t,T,n)}}function A(t){return 2&_&&1&_?e.setOriginalNode(e.setTextRange(e.createYield(o(n,e.visitNode(t.expression,T,e.isExpression))),t),t):e.visitEachChild(t,T,n)}function D(t){if(2&_&&1&_){if(t.asteriskToken){var r=e.visitNode(t.expression,T,e.isExpression);return e.setOriginalNode(e.setTextRange(e.createYield(o(n,e.updateYield(t,t.asteriskToken,u(n,p(n,r,r),r)))),t),t)}return e.setOriginalNode(e.setTextRange(e.createYield(j(t.expression?e.visitNode(t.expression,T,e.isExpression):e.createVoidZero())),t),t)}return e.visitEachChild(t,T,n)}function L(t){return 2&_&&1&_?e.updateReturn(t,j(t.expression?e.visitNode(t.expression,T,e.isExpression):e.createVoidZero())):e.visitEachChild(t,T,n)}function k(t){if(2&_){var r=e.unwrapInnermostStatementOfLabel(t);return 227===r.kind&&r.awaitModifier?V(r,t):e.restoreEnclosingLabel(e.visitEachChild(r,T,n),t)}return e.visitEachChild(t,T,n)}function N(n){for(var t,r=[],a=0,i=n;a<i.length;a++){var o=i[a];if(277===o.kind){t&&(r.push(e.createObjectLiteral(t)),t=void 0);var s=o.expression;r.push(e.visitNode(s,T,e.isExpression))}else t=e.append(t,275===o.kind?e.createPropertyAssignment(o.name,e.visitNode(o.initializer,T,e.isExpression)):e.visitNode(o,T,e.isObjectLiteralElementLike))}return t&&r.push(e.createObjectLiteral(t)),r}function I(t){if(262144&t.transformFlags){var r=N(t.properties);return r.length&&188!==r[0].kind&&r.unshift(e.createObjectLiteral()),a(n,r)}return e.visitEachChild(t,T,n)}function w(t){return e.visitEachChild(t,S,n)}function M(t,r){return e.visitEachChild(t,r?S:T,n)}function O(t){return t.variableDeclaration?e.visitEachChild(t,T,n):e.updateCatchClause(t,e.createVariableDeclaration(e.createTempVariable(void 0)),e.visitNode(t.block,T,e.isBlock))}function P(t,r){return e.isDestructuringAssignment(t)&&262144&t.left.transformFlags?e.flattenDestructuringAssignment(t,T,n,1,!r):27===t.operatorToken.kind?e.updateBinary(t,e.visitNode(t.left,S,e.isExpression),e.visitNode(t.right,r?S:T,e.isExpression)):e.visitEachChild(t,T,n)}function R(t){return e.isBindingPattern(t.name)&&262144&t.name.transformFlags?e.flattenDestructuringBinding(t,T,n,1):e.visitEachChild(t,T,n)}function F(n){return e.updateFor(n,e.visitNode(n.initializer,S,e.isForInitializer),e.visitNode(n.condition,T,e.isExpression),e.visitNode(n.incrementor,T,e.isExpression),e.visitNode(n.statement,T,e.isStatement))}function G(t){return e.visitEachChild(t,S,n)}function V(t,r){return 262144&t.initializer.transformFlags&&(t=B(t)),t.awaitModifier?U(t,r):e.restoreEnclosingLabel(e.visitEachChild(t,T,n),r)}function B(n){var t=e.skipParentheses(n.initializer);if(e.isVariableDeclarationList(t)||e.isAssignmentPattern(t)){var r=void 0,a=void 0,i=e.createTempVariable(void 0),o=[e.createForOfBindingStatement(t,i)];return e.isBlock(n.statement)?(e.addRange(o,n.statement.statements),r=n.statement,a=n.statement.statements):n.statement&&(e.append(o,n.statement),r=n.statement,a=n.statement),e.updateForOf(n,n.awaitModifier,e.setTextRange(e.createVariableDeclarationList([e.setTextRange(e.createVariableDeclaration(i),n.initializer)],1),n.initializer),n.expression,e.setTextRange(e.createBlock(e.setTextRange(e.createNodeArray(o),a),!0),r))}return n}function K(n,t){var r,a,i=e.createForOfBindingStatement(n.initializer,t),o=[e.visitNode(i,T,e.isStatement)],s=e.visitNode(n.statement,T,e.isStatement);return e.isBlock(s)?(e.addRange(o,s.statements),r=s,a=s.statements):o.push(s),e.setEmitFlags(e.setTextRange(e.createBlock(e.setTextRange(e.createNodeArray(o),a),!0),r),432)}function j(t){return 1&_?e.createYield(void 0,o(n,t)):e.createAwait(t)}function U(t,r){var a=e.visitNode(t.expression,T,e.isExpression),o=e.isIdentifier(a)?e.getGeneratedNameForNode(a):e.createTempVariable(void 0),s=e.isIdentifier(a)?e.getGeneratedNameForNode(o):e.createTempVariable(void 0),c=e.createUniqueName("e"),l=e.getGeneratedNameForNode(c),u=e.createTempVariable(void 0),d=p(n,a,t.expression),m=e.createCall(e.createPropertyAccess(o,"next"),void 0,[]),f=e.createPropertyAccess(s,"done"),_=e.createPropertyAccess(s,"value"),g=e.createFunctionCall(u,o,[]);i(c),i(u);var v=e.setEmitFlags(e.setTextRange(e.createFor(e.setEmitFlags(e.setTextRange(e.createVariableDeclarationList([e.setTextRange(e.createVariableDeclaration(o,void 0,d),t.expression),e.createVariableDeclaration(s)]),t.expression),2097152),e.createComma(e.createAssignment(s,j(m)),e.createLogicalNot(f)),void 0,K(t,_)),t),256);return e.createTry(e.createBlock([e.restoreEnclosingLabel(v,r)]),e.createCatchClause(e.createVariableDeclaration(l),e.setEmitFlags(e.createBlock([e.createExpressionStatement(e.createAssignment(c,e.createObjectLiteral([e.createPropertyAssignment("error",l)])))]),1)),e.createBlock([e.createTry(e.createBlock([e.setEmitFlags(e.createIf(e.createLogicalAnd(e.createLogicalAnd(s,e.createLogicalNot(f)),e.createAssignment(u,e.createPropertyAccess(o,"return"))),e.createExpressionStatement(j(g))),1)]),void 0,e.setEmitFlags(e.createBlock([e.setEmitFlags(e.createIf(c,e.createThrow(e.createPropertyAccess(c,"error"))),1)]),1))]))}function H(t){return 262144&t.transformFlags?e.updateParameter(t,void 0,void 0,t.dotDotDotToken,e.getGeneratedNameForNode(t),void 0,void 0,e.visitNode(t.initializer,T,e.isExpression)):e.visitEachChild(t,T,n)}function W(t){var r=_;_=0;var a=e.updateConstructor(t,void 0,t.modifiers,e.visitParameterList(t.parameters,T,n),$(t));return _=r,a}function q(t){var r=_;_=0;var a=e.updateGetAccessor(t,void 0,t.modifiers,e.visitNode(t.name,T,e.isPropertyName),e.visitParameterList(t.parameters,T,n),void 0,$(t));return _=r,a}function z(t){var r=_;_=0;var a=e.updateSetAccessor(t,void 0,t.modifiers,e.visitNode(t.name,T,e.isPropertyName),e.visitParameterList(t.parameters,T,n),$(t));return _=r,a}function J(t){var r=_;_=e.getFunctionFlags(t);var a=e.updateMethod(t,void 0,1&_?e.visitNodes(t.modifiers,x,e.isModifier):t.modifiers,2&_?void 0:t.asteriskToken,e.visitNode(t.name,T,e.isPropertyName),e.visitNode(void 0,T,e.isToken),void 0,e.visitParameterList(t.parameters,T,n),void 0,2&_&&1&_?Z(t):$(t));return _=r,a}function X(t){var r=_;_=e.getFunctionFlags(t);var a=e.updateFunctionDeclaration(t,void 0,1&_?e.visitNodes(t.modifiers,x,e.isModifier):t.modifiers,2&_?void 0:t.asteriskToken,t.name,void 0,e.visitParameterList(t.parameters,T,n),void 0,2&_&&1&_?Z(t):$(t));return _=r,a}function Y(t){var r=_;_=e.getFunctionFlags(t);var a=e.updateArrowFunction(t,t.modifiers,void 0,e.visitParameterList(t.parameters,T,n),void 0,t.equalsGreaterThanToken,$(t));return _=r,a}function Q(t){var r=_;_=e.getFunctionFlags(t);var a=e.updateFunctionExpression(t,1&_?e.visitNodes(t.modifiers,x,e.isModifier):t.modifiers,2&_?void 0:t.asteriskToken,t.name,void 0,e.visitParameterList(t.parameters,T,n),void 0,2&_&&1&_?Z(t):$(t));return _=r,a}function Z(a){t();var i=[],o=e.addPrologue(i,a.body.statements,!1,T);ee(i,a);var l=v,u=y;v=e.createUnderscoreEscapedMap(),y=!1;var p=e.createReturn(c(n,e.createFunctionExpression(void 0,e.createToken(40),a.name&&e.getGeneratedNameForNode(a.name),void 0,[],void 0,e.updateBlock(a.body,e.visitLexicalEnvironment(a.body.statements,T,n,o))))),m=d>=2&&6144&s.getNodeCheckFlags(a);if(m){ne();var f=e.createSuperAccessVariableStatement(s,a,v);b[e.getNodeId(f)]=!0,e.addStatementsAfterPrologue(i,[f])}i.push(p),e.addStatementsAfterPrologue(i,r());var _=e.updateBlock(a.body,i);return m&&y&&(4096&s.getNodeCheckFlags(a)?e.addEmitHelper(_,e.advancedAsyncSuperHelper):2048&s.getNodeCheckFlags(a)&&e.addEmitHelper(_,e.asyncSuperHelper)),v=l,y=u,_}function $(n){t();var a=0,i=[],o=e.visitNode(n.body,T,e.isConciseBody);e.isBlock(o)&&(a=e.addPrologue(i,o.statements,!1,T)),e.addRange(i,ee(void 0,n));var s=r();if(a>0||e.some(i)||e.some(s)){var c=e.convertToFunctionBody(o,!0);return e.addStatementsAfterPrologue(i,s),e.addRange(i,c.statements.slice(a)),e.updateBlock(c,e.setTextRange(e.createNodeArray(i),c.statements))}return o}function ee(t,r){for(var a=0,i=r.parameters;a<i.length;a++){var o=i[a];if(262144&o.transformFlags){var s=e.getGeneratedNameForNode(o),c=e.flattenDestructuringBinding(o,T,n,1,s,!1,!0);if(e.some(c)){var l=e.createVariableStatement(void 0,e.createVariableDeclarationList(c));e.setEmitFlags(l,1048576),t=e.append(t,l)}}}return t}function ne(){0===(1&f)&&(f|=1,n.enableSubstitution(191),n.enableSubstitution(189),n.enableSubstitution(190),n.enableEmitNotification(240),n.enableEmitNotification(156),n.enableEmitNotification(158),n.enableEmitNotification(159),n.enableEmitNotification(157),n.enableEmitNotification(219))}function te(n,t,r){if(1&f&&ce(t)){var a=6144&s.getNodeCheckFlags(t);if(a!==h){var i=h;return h=a,m(n,t,r),void(h=i)}}else if(f&&b[e.getNodeId(t)]){i=h;return h=0,m(n,t,r),void(h=i)}m(n,t,r)}function re(e,n){return n=g(e,n),1===e&&h?ae(n):n}function ae(e){switch(e.kind){case 189:return ie(e);case 190:return oe(e);case 191:return se(e)}return e}function ie(n){return 98===n.expression.kind?e.setTextRange(e.createPropertyAccess(e.createFileLevelUniqueName("_super"),n.name),n):n}function oe(e){return 98===e.expression.kind?le(e.argumentExpression,e):e}function se(n){var t=n.expression;if(e.isSuperProperty(t)){var r=e.isPropertyAccessExpression(t)?ie(t):oe(t);return e.createCall(e.createPropertyAccess(r,"call"),void 0,[e.createThis()].concat(n.arguments))}return n}function ce(e){var n=e.kind;return 240===n||157===n||156===n||158===n||159===n}function le(n,t){return 4096&h?e.setTextRange(e.createPropertyAccess(e.createCall(e.createIdentifier("_superIndex"),void 0,[n]),"value"),t):e.setTextRange(e.createCall(e.createIdentifier("_superIndex"),void 0,[n]),t)}}(function(e){e[e["AsyncMethodsWithSuper"]=1]="AsyncMethodsWithSuper"})(n||(n={})),e.transformESNext=t;var r={name:"typescript:assign",scoped:!1,priority:1,text:"\n            var __assign = (this && this.__assign) || function () {\n                __assign = Object.assign || function(t) {\n                    for (var s, i = 1, n = arguments.length; i < n; i++) {\n                        s = arguments[i];\n                        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))\n                            t[p] = s[p];\n                    }\n                    return t;\n                };\n                return __assign.apply(this, arguments);\n            };"};function a(n,t){return n.getCompilerOptions().target>=2?e.createCall(e.createPropertyAccess(e.createIdentifier("Object"),"assign"),void 0,t):(n.requestEmitHelper(r),e.createCall(e.getHelperName("__assign"),void 0,t))}e.createAssignHelper=a;var i={name:"typescript:await",scoped:!1,text:"\n            var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }"};function o(n,t){return n.requestEmitHelper(i),e.createCall(e.getHelperName("__await"),void 0,[t])}var s={name:"typescript:asyncGenerator",scoped:!1,text:'\n            var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {\n                if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");\n                var g = generator.apply(thisArg, _arguments || []), i, q = [];\n                return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;\n                function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }\n                function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }\n                function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }\n                function fulfill(value) { resume("next", value); }\n                function reject(value) { resume("throw", value); }\n                function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }\n            };'};function c(n,t){return n.requestEmitHelper(i),n.requestEmitHelper(s),(t.emitNode||(t.emitNode={})).flags|=262144,e.createCall(e.getHelperName("__asyncGenerator"),void 0,[e.createThis(),e.createIdentifier("arguments"),t])}var l={name:"typescript:asyncDelegator",scoped:!1,text:'\n            var __asyncDelegator = (this && this.__asyncDelegator) || function (o) {\n                var i, p;\n                return i = {}, verb("next"), verb("throw", function (e) { throw e; }), verb("return"), i[Symbol.iterator] = function () { return this; }, i;\n                function verb(n, f) { i[n] = o[n] ? function (v) { return (p = !p) ? { value: __await(o[n](v)), done: n === "return" } : f ? f(v) : v; } : f; }\n            };'};function u(n,t,r){return n.requestEmitHelper(i),n.requestEmitHelper(l),e.setTextRange(e.createCall(e.getHelperName("__asyncDelegator"),void 0,[t]),r)}var d={name:"typescript:asyncValues",scoped:!1,text:'\n            var __asyncValues = (this && this.__asyncValues) || function (o) {\n                if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");\n                var m = o[Symbol.asyncIterator], i;\n                return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);\n                function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }\n                function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }\n            };'};function p(n,t,r){return n.requestEmitHelper(d),e.setTextRange(e.createCall(e.getHelperName("__asyncValues"),void 0,[t]),r)}}(u||(u={})),function(e){function n(n){var r,a=n.getCompilerOptions();return e.chainBundle(i);function i(t){if(t.isDeclarationFile)return t;r=t;var a=e.visitEachChild(t,o,n);return e.addEmitHelpers(a,n.readEmitHelpers()),a}function o(e){return 4&e.transformFlags?s(e):e}function s(t){switch(t.kind){case 260:return l(t,!1);case 261:return u(t,!1);case 264:return d(t,!1);case 270:return x(t);default:return e.visitEachChild(t,o,n)}}function c(n){switch(n.kind){case 11:return v(n);case 270:return x(n);case 260:return l(n,!0);case 261:return u(n,!0);case 264:return d(n,!0);default:return e.Debug.failBadSyntaxKind(n)}}function l(e,n){return p(e.openingElement,e.children,n,e)}function u(e,n){return p(e,void 0,n,e)}function d(e,n){return m(e.openingFragment,e.children,n,e)}function p(t,i,o,s){var l,u=T(t),d=t.attributes.properties;if(0===d.length)l=e.createNull();else{var p=e.flatten(e.spanMap(d,e.isJsxSpreadAttribute,function(n,t){return t?e.map(n,f):e.createObjectLiteral(e.map(n,_))}));e.isJsxSpreadAttribute(d[0])&&p.unshift(e.createObjectLiteral()),l=e.singleOrUndefined(p),l||(l=e.createAssignHelper(n,p))}var m=e.createExpressionForJsxElement(n.getEmitResolver().getJsxFactoryEntity(r),a.reactNamespace,u,l,e.mapDefined(i,c),t,s);return o&&e.startOnNewLine(m),m}function m(t,i,o,s){var l=e.createExpressionForJsxFragment(n.getEmitResolver().getJsxFactoryEntity(r),a.reactNamespace,e.mapDefined(i,c),t,s);return o&&e.startOnNewLine(l),l}function f(n){return e.visitNode(n.expression,o,e.isExpression)}function _(n){var t=S(n),r=g(n.initializer);return e.createPropertyAssignment(t,r)}function g(n){if(void 0===n)return e.createTrue();if(10===n.kind){var t=e.createLiteral(E(n.text)||n.text);return t.singleQuote=void 0!==n.singleQuote?n.singleQuote:!e.isStringDoubleQuoted(n,r),e.setTextRange(t,n)}return 270===n.kind?void 0===n.expression?e.createTrue():x(n):e.Debug.failBadSyntaxKind(n)}function v(n){var t=y(e.getTextOfNode(n,!0));return void 0===t?void 0:e.createLiteral(t)}function y(n){for(var t,r=0,a=-1,i=0;i<n.length;i++){var o=n.charCodeAt(i);e.isLineBreak(o)?(-1!==r&&-1!==a&&(t=h(t,n.substr(r,a-r+1))),r=-1):e.isWhiteSpaceSingleLine(o)||(a=i,-1===r&&(r=i))}return-1!==r?h(t,n.substr(r)):t}function h(e,n){var t=b(n);return void 0===e?t:e+" "+t}function b(e){return e.replace(/&((#((\d+)|x([\da-fA-F]+)))|(\w+));/g,function(e,n,r,a,i,o,s){if(i)return String.fromCharCode(parseInt(i,10));if(o)return String.fromCharCode(parseInt(o,16));var c=t.get(s);return c?String.fromCharCode(c):e})}function E(e){var n=b(e);return n===e?void 0:n}function T(n){if(260===n.kind)return T(n.openingElement);var t=n.tagName;return e.isIdentifier(t)&&e.isIntrinsicJsxName(t.escapedText)?e.createLiteral(e.idText(t)):e.createExpressionFromEntityName(t)}function S(n){var t=n.name,r=e.idText(t);return/^[A-Za-z_]\w*$/.test(r)?t:e.createLiteral(r)}function x(n){return e.visitNode(n.expression,o,e.isExpression)}}e.transformJsx=n;var t=e.createMapFromTemplate({quot:34,amp:38,apos:39,lt:60,gt:62,nbsp:160,iexcl:161,cent:162,pound:163,curren:164,yen:165,brvbar:166,sect:167,uml:168,copy:169,ordf:170,laquo:171,not:172,shy:173,reg:174,macr:175,deg:176,plusmn:177,sup2:178,sup3:179,acute:180,micro:181,para:182,middot:183,cedil:184,sup1:185,ordm:186,raquo:187,frac14:188,frac12:189,frac34:190,iquest:191,Agrave:192,Aacute:193,Acirc:194,Atilde:195,Auml:196,Aring:197,AElig:198,Ccedil:199,Egrave:200,Eacute:201,Ecirc:202,Euml:203,Igrave:204,Iacute:205,Icirc:206,Iuml:207,ETH:208,Ntilde:209,Ograve:210,Oacute:211,Ocirc:212,Otilde:213,Ouml:214,times:215,Oslash:216,Ugrave:217,Uacute:218,Ucirc:219,Uuml:220,Yacute:221,THORN:222,szlig:223,agrave:224,aacute:225,acirc:226,atilde:227,auml:228,aring:229,aelig:230,ccedil:231,egrave:232,eacute:233,ecirc:234,euml:235,igrave:236,iacute:237,icirc:238,iuml:239,eth:240,ntilde:241,ograve:242,oacute:243,ocirc:244,otilde:245,ouml:246,divide:247,oslash:248,ugrave:249,uacute:250,ucirc:251,uuml:252,yacute:253,thorn:254,yuml:255,OElig:338,oelig:339,Scaron:352,scaron:353,Yuml:376,fnof:402,circ:710,tilde:732,Alpha:913,Beta:914,Gamma:915,Delta:916,Epsilon:917,Zeta:918,Eta:919,Theta:920,Iota:921,Kappa:922,Lambda:923,Mu:924,Nu:925,Xi:926,Omicron:927,Pi:928,Rho:929,Sigma:931,Tau:932,Upsilon:933,Phi:934,Chi:935,Psi:936,Omega:937,alpha:945,beta:946,gamma:947,delta:948,epsilon:949,zeta:950,eta:951,theta:952,iota:953,kappa:954,lambda:955,mu:956,nu:957,xi:958,omicron:959,pi:960,rho:961,sigmaf:962,sigma:963,tau:964,upsilon:965,phi:966,chi:967,psi:968,omega:969,thetasym:977,upsih:978,piv:982,ensp:8194,emsp:8195,thinsp:8201,zwnj:8204,zwj:8205,lrm:8206,rlm:8207,ndash:8211,mdash:8212,lsquo:8216,rsquo:8217,sbquo:8218,ldquo:8220,rdquo:8221,bdquo:8222,dagger:8224,Dagger:8225,bull:8226,hellip:8230,permil:8240,prime:8242,Prime:8243,lsaquo:8249,rsaquo:8250,oline:8254,frasl:8260,euro:8364,image:8465,weierp:8472,real:8476,trade:8482,alefsym:8501,larr:8592,uarr:8593,rarr:8594,darr:8595,harr:8596,crarr:8629,lArr:8656,uArr:8657,rArr:8658,dArr:8659,hArr:8660,forall:8704,part:8706,exist:8707,empty:8709,nabla:8711,isin:8712,notin:8713,ni:8715,prod:8719,sum:8721,minus:8722,lowast:8727,radic:8730,prop:8733,infin:8734,ang:8736,and:8743,or:8744,cap:8745,cup:8746,int:8747,there4:8756,sim:8764,cong:8773,asymp:8776,ne:8800,equiv:8801,le:8804,ge:8805,sub:8834,sup:8835,nsub:8836,sube:8838,supe:8839,oplus:8853,otimes:8855,perp:8869,sdot:8901,lceil:8968,rceil:8969,lfloor:8970,rfloor:8971,lang:9001,rang:9002,loz:9674,spades:9824,clubs:9827,hearts:9829,diams:9830})}(u||(u={})),function(e){function n(n){var t=n.hoistVariableDeclaration;return e.chainBundle(r);function r(t){return t.isDeclarationFile?t:e.visitEachChild(t,a,n)}function a(t){if(0===(32&t.transformFlags))return t;switch(t.kind){case 204:return i(t);default:return e.visitEachChild(t,a,n)}}function i(t){switch(t.operatorToken.kind){case 63:return o(t);case 41:return s(t);default:return e.visitEachChild(t,a,n)}}function o(n){var r,i,o=e.visitNode(n.left,a,e.isExpression),s=e.visitNode(n.right,a,e.isExpression);if(e.isElementAccessExpression(o)){var c=e.createTempVariable(t),l=e.createTempVariable(t);r=e.setTextRange(e.createElementAccess(e.setTextRange(e.createAssignment(c,o.expression),o.expression),e.setTextRange(e.createAssignment(l,o.argumentExpression),o.argumentExpression)),o),i=e.setTextRange(e.createElementAccess(c,l),o)}else if(e.isPropertyAccessExpression(o)){c=e.createTempVariable(t);r=e.setTextRange(e.createPropertyAccess(e.setTextRange(e.createAssignment(c,o.expression),o.expression),o.name),o),i=e.setTextRange(e.createPropertyAccess(c,o.name),o)}else r=o,i=o;return e.setTextRange(e.createAssignment(r,e.createMathPow(i,s,n)),n)}function s(n){var t=e.visitNode(n.left,a,e.isExpression),r=e.visitNode(n.right,a,e.isExpression);return e.createMathPow(t,r,n)}}e.transformES2016=n}(u||(u={})),function(e){var n,t,r,a,i,o;function s(n){var t,r,a,i,o,s,u=n.startLexicalEnvironment,d=n.resumeLexicalEnvironment,p=n.endLexicalEnvironment,m=n.hoistVariableDeclaration,f=n.getCompilerOptions(),_=n.getEmitResolver(),g=n.onSubstituteNode,v=n.onEmitNode;function y(n){i=e.append(i,e.createVariableDeclaration(n))}return n.onEmitNode=qn,n.onSubstituteNode=Xn,e.chainBundle(h);function h(o){if(o.isDeclarationFile)return o;t=o,r=o.text;var s=L(o);return e.addEmitHelpers(s,n.readEmitHelpers()),t=void 0,r=void 0,i=void 0,a=0,s}function b(e,n){var t=a;return a=16383&(a&~e|n),t}function E(e,n,t){a=-16384&(a&~n|t)|e}function T(e){return 0!==(4096&a)&&230===e.kind&&!e.expression}function S(n){return 0!==(128&n.transformFlags)||void 0!==o||4096&a&&(e.isStatement(n)||218===n.kind)||e.isIterationStatement(n,!1)&&Ue(n)||0!==(33554432&e.getEmitFlags(n))}function x(e){return S(e)?D(e):e}function C(e){return S(e)?_e(e,!0):e}function A(e){return 98===e.kind?Hn(!0):x(e)}function D(t){switch(t.kind){case 116:return;case 240:return R(t);case 209:return F(t);case 151:return J(t);case 239:return de(t);case 197:return le(t);case 196:return ue(t);case 237:return xe(t);case 72:return O(t);case 238:return be(t);case 232:return k(t);case 246:return N(t);case 218:return _e(t,!1);case 229:case 228:return P(t);case 233:return De(t);case 223:case 224:return Ne(t,void 0);case 225:return Ie(t,void 0);case 226:return we(t,void 0);case 227:return Me(t,void 0);case 221:return ge(t);case 188:return Ge(t);case 274:return gn(t);case 276:return bn(t);case 149:return En(t);case 187:return Sn(t);case 191:return xn(t);case 192:return Ln(t);case 195:return ve(t,!0);case 204:return ye(t,!0);case 14:case 15:case 16:case 17:return Pn(t);case 10:return Rn(t);case 8:return Fn(t);case 193:return Gn(t);case 206:return Bn(t);case 207:return Tn(t);case 208:return Mn(t);case 98:return Hn(!1);case 100:return M(t);case 214:return Wn(t);case 156:return yn(t);case 158:case 159:return hn(t);case 219:return he(t);case 230:return w(t);default:return e.visitEachChild(t,x,n)}}function L(n){var t=b(3968,64),r=[];u();var a=e.addStandardPrologue(r,n.statements,!1);return ne(r,n),a=e.addCustomPrologue(r,n.statements,a,x),e.addRange(r,e.visitNodes(n.statements,x,e.isStatement,a)),i&&r.push(e.createVariableStatement(void 0,e.createVariableDeclarationList(i))),e.addStatementsAfterPrologue(r,p()),E(t,0,0),e.updateSourceFileNode(n,e.setTextRange(e.createNodeArray(r),n.statements))}function k(t){if(void 0!==o){var r=o.allowedNonLabeledJumps;o.allowedNonLabeledJumps|=2;var a=e.visitEachChild(t,x,n);return o.allowedNonLabeledJumps=r,a}return e.visitEachChild(t,x,n)}function N(t){var r=b(4032,0),a=e.visitEachChild(t,x,n);return E(r,0,0),a}function I(n){return e.setOriginalNode(e.createReturn(e.createFileLevelUniqueName("_this")),n)}function w(t){return o?(o.nonLocalJumps|=8,T(t)&&(t=I(t)),e.createReturn(e.createObjectLiteral([e.createPropertyAssignment(e.createIdentifier("value"),t.expression?e.visitNode(t.expression,x,e.isExpression):e.createVoidZero())]))):T(t)?I(t):e.visitEachChild(t,x,n)}function M(n){return o?2&a?(o.containsLexicalThis=!0,n):o.thisName||(o.thisName=e.createUniqueName("this")):n}function O(n){return o?e.isGeneratedIdentifier(n)?n:"arguments"===n.escapedText&&_.isArgumentsLocalBinding(n)?o.argumentsName||(o.argumentsName=e.createUniqueName("arguments")):n:n}function P(t){if(o){var r=229===t.kind?2:4,a=t.label&&o.labels&&o.labels.get(e.idText(t.label))||!t.label&&o.allowedNonLabeledJumps&r;if(!a){var i=void 0,s=t.label;s?229===t.kind?(i="break-"+s.escapedText,ln(o,!0,e.idText(s),i)):(i="continue-"+s.escapedText,ln(o,!1,e.idText(s),i)):229===t.kind?(o.nonLocalJumps|=2,i="break"):(o.nonLocalJumps|=4,i="continue");var c=e.createLiteral(i);if(o.loopOutParameters.length){for(var l=o.loopOutParameters,u=void 0,d=0;d<l.length;d++){var p=an(l[d],1);u=0===d?p:e.createBinary(u,27,p)}c=e.createBinary(u,27,c)}return e.createReturn(c)}}return e.visitEachChild(t,x,n)}function R(n){var t=e.createVariableDeclaration(e.getLocalName(n,!0),void 0,G(n));e.setOriginalNode(t,n);var r=[],a=e.createVariableStatement(void 0,e.createVariableDeclarationList([t]));if(e.setOriginalNode(a,n),e.setTextRange(a,n),e.startOnNewLine(a),r.push(a),e.hasModifier(n,1)){var i=e.hasModifier(n,512)?e.createExportDefault(e.getLocalName(n)):e.createExternalModuleExport(e.getLocalName(n));e.setOriginalNode(i,a),r.push(i)}var o=e.getEmitFlags(n);return 0===(4194304&o)&&(r.push(e.createEndOfDeclarationMarker(n)),e.setEmitFlags(a,4194304|o)),e.singleOrMany(r)}function F(e){return G(e)}function G(n){n.name&&zn();var t=e.getClassExtendsHeritageElement(n),a=e.createFunctionExpression(void 0,void 0,void 0,void 0,t?[e.createParameter(void 0,void 0,void 0,e.createFileLevelUniqueName("_super"))]:[],void 0,V(n,t));e.setEmitFlags(a,65536&e.getEmitFlags(n)|524288);var i=e.createPartiallyEmittedExpression(a);i.end=n.end,e.setEmitFlags(i,1536);var o=e.createPartiallyEmittedExpression(i);o.end=e.skipTrivia(r,n.pos),e.setEmitFlags(o,1536);var s=e.createParen(e.createCall(o,void 0,t?[e.visitNode(t.expression,x,e.isExpression)]:[]));return e.addSyntheticLeadingComment(s,3,"* @class "),s}function V(n,t){var a=[];u(),B(a,n,t),K(a,n,t),ae(a,n);var i=e.createTokenRange(e.skipTrivia(r,n.members.end),19),o=e.getInternalName(n),s=e.createPartiallyEmittedExpression(o);s.end=i.end,e.setEmitFlags(s,1536);var c=e.createReturn(s);c.pos=i.pos,e.setEmitFlags(c,1920),a.push(c),e.addStatementsAfterPrologue(a,p());var l=e.createBlock(e.setTextRange(e.createNodeArray(a),n.members),!0);return e.setEmitFlags(l,1536),l}function B(t,r,a){a&&t.push(e.setTextRange(e.createExpressionStatement(c(n,e.getInternalName(r))),a))}function K(n,t,r){var a=o;o=void 0;var i=b(16278,73),s=e.getFirstConstructorWithBody(t),c=rt(s,void 0!==r),l=e.createFunctionDeclaration(void 0,void 0,void 0,e.getInternalName(t),void 0,j(s,c),void 0,U(s,t,r,c));e.setTextRange(l,s||t),r&&e.setEmitFlags(l,8),n.push(l),E(i,49152,0),o=a}function j(t,r){return e.visitParameterList(t&&!r?t.parameters:void 0,x,n)||[]}function U(n,t,r,i){var o=[];d();var s=-1;i?s=0:n&&(s=e.addStandardPrologue(o,n.body.statements,!1)),n&&(Y(o,n),ee(o,n,i),i||(s=e.addCustomPrologue(o,n.body.statements,s,x)),e.Debug.assert(s>=0,"statementOffset not initialized correctly!"));var c=!!r&&96!==e.skipOuterExpressions(r.expression).kind,l=W(o,n,c,i,s);1!==l&&2!==l||s++,n&&(1===l&&(a|=4096),e.addRange(o,e.visitNodes(n.body.statements,x,e.isStatement,s))),!c||2===l||n&&H(n.body)||o.push(e.createReturn(e.createFileLevelUniqueName("_this"))),e.addStatementsAfterPrologue(o,p()),n&&re(o,n,!1);var u=e.createBlock(e.setTextRange(e.createNodeArray(o),n?n.body.statements:t.members),!0);return e.setTextRange(u,n?n.body:t),n||e.setEmitFlags(u,1536),u}function H(n){if(230===n.kind)return!0;if(222===n.kind){var t=n;if(t.elseStatement)return H(t.thenStatement)&&H(t.elseStatement)}else if(218===n.kind){var r=e.lastOrUndefined(n.statements);if(r&&H(r))return!0}return!1}function W(n,t,r,a,i){if(!r)return t&&ne(n,t),0;if(!t)return n.push(e.createReturn(z())),2;if(a)return te(n,t,z()),Jn(),1;var o,s,c=t.body.statements;if(i<c.length&&(o=c[i],221===o.kind&&e.isSuperCall(o.expression)&&(s=An(o.expression))),s&&i===c.length-1&&!(24576&t.transformFlags)){var l=e.createReturn(s);return 204===s.kind&&191===s.left.kind||e.Debug.fail("Assumed generated super call would have form 'super.call(...) || this'."),e.setCommentRange(l,e.getCommentRange(e.setEmitFlags(s.left,1536))),n.push(l),2}return te(n,t,s||q()),s?1:0}function q(){return e.setEmitFlags(e.createThis(),4)}function z(){return e.createLogicalOr(e.createLogicalAnd(e.createStrictInequality(e.createFileLevelUniqueName("_super"),e.createNull()),e.createFunctionApply(e.createFileLevelUniqueName("_super"),q(),e.createIdentifier("arguments"))),q())}function J(n){return n.dotDotDotToken?void 0:e.isBindingPattern(n.name)?e.setOriginalNode(e.setTextRange(e.createParameter(void 0,void 0,void 0,e.getGeneratedNameForNode(n),void 0,void 0,void 0),n),n):n.initializer?e.setOriginalNode(e.setTextRange(e.createParameter(void 0,void 0,void 0,n.name,void 0,void 0,void 0),n),n):n}function X(e){return 0!==(65536&e.transformFlags)}function Y(n,t){if(X(t))for(var r=0,a=t.parameters;r<a.length;r++){var i=a[r],o=i.name,s=i.initializer,c=i.dotDotDotToken;c||(e.isBindingPattern(o)?Q(n,i,o,s):s&&Z(n,i,o,s))}}function Q(t,r,a,i){var o=e.getGeneratedNameForNode(r);a.elements.length>0?t.push(e.setEmitFlags(e.createVariableStatement(void 0,e.createVariableDeclarationList(e.flattenDestructuringBinding(r,x,n,0,o))),1048576)):i&&t.push(e.setEmitFlags(e.createExpressionStatement(e.createAssignment(o,e.visitNode(i,x,e.isExpression))),1048576))}function Z(n,t,r,a){a=e.visitNode(a,x,e.isExpression);var i=e.createIf(e.createTypeCheck(e.getSynthesizedClone(r),"undefined"),e.setEmitFlags(e.setTextRange(e.createBlock([e.createExpressionStatement(e.setEmitFlags(e.setTextRange(e.createAssignment(e.setEmitFlags(e.getMutableClone(r),48),e.setEmitFlags(a,1584|e.getEmitFlags(a))),t),1536))]),t),1953));e.startOnNewLine(i),e.setTextRange(i,t),e.setEmitFlags(i,1050528),n.push(i)}function $(e,n){return!(!e||!e.dotDotDotToken||n)}function ee(t,r,a){var i=e.lastOrUndefined(r.parameters);if($(i,a)){var o=72===i.name.kind?e.getMutableClone(i.name):e.createTempVariable(void 0);e.setEmitFlags(o,48);var s=72===i.name.kind?e.getSynthesizedClone(i.name):o,c=r.parameters.length-1,l=e.createLoopVariable();t.push(e.setEmitFlags(e.setTextRange(e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(o,void 0,e.createArrayLiteral([]))])),i),1048576));var u=e.createFor(e.setTextRange(e.createVariableDeclarationList([e.createVariableDeclaration(l,void 0,e.createLiteral(c))]),i),e.setTextRange(e.createLessThan(l,e.createPropertyAccess(e.createIdentifier("arguments"),"length")),i),e.setTextRange(e.createPostfixIncrement(l),i),e.createBlock([e.startOnNewLine(e.setTextRange(e.createExpressionStatement(e.createAssignment(e.createElementAccess(s,0===c?l:e.createSubtract(l,e.createLiteral(c))),e.createElementAccess(e.createIdentifier("arguments"),l))),i))]));e.setEmitFlags(u,1048576),e.startOnNewLine(u),t.push(u),72!==i.name.kind&&t.push(e.setEmitFlags(e.setTextRange(e.createVariableStatement(void 0,e.createVariableDeclarationList(e.flattenDestructuringBinding(i,x,n,0,s))),i),1048576))}}function ne(n,t){16384&t.transformFlags&&197!==t.kind&&te(n,t,e.createThis())}function te(n,t,r){Jn();var a=e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(e.createFileLevelUniqueName("_this"),void 0,r)]));e.setEmitFlags(a,1050112),e.setSourceMapRange(a,t),n.push(a)}function re(n,t,r){if(16384&a){var i=void 0;switch(t.kind){case 197:return n;case 156:case 158:case 159:i=e.createVoidZero();break;case 157:i=e.createPropertyAccess(e.setEmitFlags(e.createThis(),4),"constructor");break;case 239:case 196:i=e.createConditional(e.createLogicalAnd(e.setEmitFlags(e.createThis(),4),e.createBinary(e.setEmitFlags(e.createThis(),4),94,e.getLocalName(t))),e.createPropertyAccess(e.setEmitFlags(e.createThis(),4),"constructor"),e.createVoidZero());break;default:return e.Debug.failBadSyntaxKind(t)}var o=e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(e.createFileLevelUniqueName("_newTarget"),void 0,i)]));if(r)return[o].concat(n);n.unshift(o)}return n}function ae(n,t){for(var r=0,a=t.members;r<a.length;r++){var i=a[r];switch(i.kind){case 217:n.push(ie(i));break;case 156:n.push(oe(tt(t,i),i,t));break;case 158:case 159:var o=e.getAllAccessorDeclarations(t.members,i);i===o.firstAccessor&&n.push(se(tt(t,i),o,t));break;case 157:break;default:e.Debug.failBadSyntaxKind(t);break}}}function ie(n){return e.setTextRange(e.createEmptyStatement(),n)}function oe(n,t,r){var i=b(0,0),o=e.getCommentRange(t),s=e.getSourceMapRange(t),c=e.createMemberAccessForPropertyName(n,e.visitNode(t.name,x,e.isPropertyName),t.name),l=pe(t,t,void 0,r);e.setEmitFlags(l,1536),e.setSourceMapRange(l,s);var u=e.setTextRange(e.createExpressionStatement(e.createAssignment(c,l)),t);return e.setOriginalNode(u,t),e.setCommentRange(u,o),e.setEmitFlags(u,48),E(i,49152,49152&a?16384:0),u}function se(n,t,r){var a=e.createExpressionStatement(ce(n,t,r,!1));return e.setEmitFlags(a,1536),e.setSourceMapRange(a,e.getSourceMapRange(t.firstAccessor)),a}function ce(n,t,r,i){var o=t.firstAccessor,s=t.getAccessor,c=t.setAccessor,l=b(0,0),u=e.getMutableClone(n);e.setEmitFlags(u,1568),e.setSourceMapRange(u,o.name);var d=e.createExpressionForPropertyName(e.visitNode(o.name,x,e.isPropertyName));e.setEmitFlags(d,1552),e.setSourceMapRange(d,o.name);var p=[];if(s){var m=pe(s,void 0,void 0,r);e.setSourceMapRange(m,e.getSourceMapRange(s)),e.setEmitFlags(m,512);var f=e.createPropertyAssignment("get",m);e.setCommentRange(f,e.getCommentRange(s)),p.push(f)}if(c){var _=pe(c,void 0,void 0,r);e.setSourceMapRange(_,e.getSourceMapRange(c)),e.setEmitFlags(_,512);var g=e.createPropertyAssignment("set",_);e.setCommentRange(g,e.getCommentRange(c)),p.push(g)}p.push(e.createPropertyAssignment("enumerable",e.createTrue()),e.createPropertyAssignment("configurable",e.createTrue()));var v=e.createCall(e.createPropertyAccess(e.createIdentifier("Object"),"defineProperty"),void 0,[u,d,e.createObjectLiteral(p,!0)]);return i&&e.startOnNewLine(v),E(l,49152,49152&a?16384:0),v}function le(t){8192&t.transformFlags&&Jn();var r=o;o=void 0;var a=b(16256,66),i=e.createFunctionExpression(void 0,void 0,void 0,void 0,e.visitParameterList(t.parameters,x,n),void 0,me(t));return e.setTextRange(i,t),e.setOriginalNode(i,t),e.setEmitFlags(i,8),E(a,0,0),o=r,i}function ue(t){var r=262144&e.getEmitFlags(t)?b(16278,69):b(16286,65),i=o;o=void 0;var s=e.visitParameterList(t.parameters,x,n),c=64&t.transformFlags?me(t):fe(t),l=16384&a?e.getLocalName(t):t.name;return E(r,49152,0),o=i,e.updateFunctionExpression(t,void 0,t.asteriskToken,l,void 0,s,void 0,c)}function de(t){var r=o;o=void 0;var i=b(16286,65),s=e.visitParameterList(t.parameters,x,n),c=64&t.transformFlags?me(t):fe(t),l=16384&a?e.getLocalName(t):t.name;return E(i,49152,0),o=r,e.updateFunctionDeclaration(t,void 0,e.visitNodes(t.modifiers,x,e.isModifier),t.asteriskToken,l,void 0,s,void 0,c)}function pe(t,r,i,s){var c=o;o=void 0;var l=s&&e.isClassLike(s)&&!e.hasModifier(t,32)?b(16286,73):b(16286,65),u=e.visitParameterList(t.parameters,x,n),d=me(t);return 16384&a&&!i&&(239===t.kind||196===t.kind)&&(i=e.getGeneratedNameForNode(t)),E(l,49152,0),o=c,e.setOriginalNode(e.setTextRange(e.createFunctionExpression(void 0,t.asteriskToken,i,void 0,u,void 0,d),r),t)}function me(r){var a,i,o,s=!1,c=!1,l=[],u=[],p=r.body;if(d(),e.isBlock(p)&&(o=e.addStandardPrologue(l,p.statements,!1)),ne(l,r),Y(l,r),ee(l,r,!1),e.isBlock(p))o=e.addCustomPrologue(l,p.statements,o,x),a=p.statements,e.addRange(u,e.visitNodes(p.statements,x,e.isStatement,o)),!s&&p.multiLine&&(s=!0);else{e.Debug.assert(197===r.kind),a=e.moveRangeEnd(p,-1);var m=r.equalsGreaterThanToken;e.nodeIsSynthesized(m)||e.nodeIsSynthesized(p)||(e.rangeEndIsOnSameLineAsRangeStart(m,p,t)?c=!0:s=!0);var f=e.visitNode(p,x,e.isExpression),_=e.createReturn(f);e.setTextRange(_,p),e.moveSyntheticComments(_,p),e.setEmitFlags(_,1440),u.push(_),i=p}var g=n.endLexicalEnvironment();e.addStatementsAfterPrologue(u,g),re(u,r,!1),(e.some(l)||e.some(g))&&(s=!0);var v=e.createBlock(e.setTextRange(e.createNodeArray(l.concat(u)),a),s);return e.setTextRange(v,r.body),!s&&c&&e.setEmitFlags(v,1),i&&e.setTokenSourceMapRange(v,19,i),e.setOriginalNode(v,r.body),v}function fe(t){var r=e.visitFunctionBody(t.body,C,n);return e.updateBlock(r,e.setTextRange(e.createNodeArray(re(r.statements,t,!0)),r.statements))}function _e(t,r){if(r)return e.visitEachChild(t,x,n);var i=256&a?b(4032,512):b(3904,128),o=e.visitEachChild(t,x,n);return E(i,0,0),o}function ge(t){switch(t.expression.kind){case 195:return e.updateExpressionStatement(t,ve(t.expression,!1));case 204:return e.updateExpressionStatement(t,ye(t.expression,!1))}return e.visitEachChild(t,x,n)}function ve(t,r){if(!r)switch(t.expression.kind){case 195:return e.updateParen(t,ve(t.expression,!1));case 204:return e.updateParen(t,ye(t.expression,!1))}return e.visitEachChild(t,x,n)}function ye(t,r){return e.isDestructuringAssignment(t)?e.flattenDestructuringAssignment(t,x,n,0,r):e.visitEachChild(t,x,n)}function he(t){var r,a=b(0,e.hasModifier(t,1)?32:0);if(o&&0===(3&t.declarationList.flags)){for(var i=void 0,s=0,c=t.declarationList.declarations;s<c.length;s++){var l=c[s];if(We(o,l),l.initializer){var u=void 0;e.isBindingPattern(l.name)?u=e.flattenDestructuringAssignment(l,x,n,0):(u=e.createBinary(l.name,59,e.visitNode(l.initializer,x,e.isExpression)),e.setTextRange(u,l)),i=e.append(i,u)}}r=i?e.setTextRange(e.createExpressionStatement(e.inlineExpressions(i)),t):void 0}else r=e.visitEachChild(t,x,n);return E(a,0,0),r}function be(t){if(64&t.transformFlags){3&t.flags&&zn();var r=e.flatMap(t.declarations,1&t.flags?Se:xe),a=e.createVariableDeclarationList(r);return e.setOriginalNode(a,t),e.setTextRange(a,t),e.setCommentRange(a,t),2097152&t.transformFlags&&(e.isBindingPattern(t.declarations[0].name)||e.isBindingPattern(e.last(t.declarations).name))&&e.setSourceMapRange(a,Ee(r)),a}return e.visitEachChild(t,x,n)}function Ee(n){for(var t=-1,r=-1,a=0,i=n;a<i.length;a++){var o=i[a];t=-1===t?o.pos:-1===o.pos?t:Math.min(t,o.pos),r=Math.max(r,o.end)}return e.createRange(t,r)}function Te(e){var n=_.getNodeCheckFlags(e),t=262144&n,r=524288&n,i=0!==(64&a)||t&&r&&0!==(512&a),o=!i&&0===(2048&a)&&(!_.isDeclarationWithCollidingName(e)||r&&!t&&0===(3072&a));return o}function Se(t){var r=t.name;if(e.isBindingPattern(r))return xe(t);if(!t.initializer&&Te(t)){var a=e.getMutableClone(t);return a.initializer=e.createVoidZero(),a}return e.visitEachChild(t,x,n)}function xe(t){var r,a=b(32,0);return r=e.isBindingPattern(t.name)?e.flattenDestructuringBinding(t,x,n,0,void 0,0!==(32&a)):e.visitEachChild(t,x,n),E(a,0,0),r}function Ce(n){o.labels.set(e.idText(n.label),!0)}function Ae(n){o.labels.set(e.idText(n.label),!1)}function De(n){o&&!o.labels&&(o.labels=e.createMap());var t=e.unwrapInnermostStatementOfLabel(n,o&&Ce);return e.isIterationStatement(t,!1)?Le(t,n):e.restoreEnclosingLabel(e.visitNode(t,x,e.isStatement,e.liftToBlock),n,o&&Ae)}function Le(e,n){switch(e.kind){case 223:case 224:return Ne(e,n);case 225:return Ie(e,n);case 226:return we(e,n);case 227:return Me(e,n)}}function ke(e,n,t,r,a){var i=b(e,n),o=qe(t,r,a);return E(i,0,0),o}function Ne(e,n){return ke(0,256,e,n)}function Ie(e,n){return ke(3008,1280,e,n)}function we(e,n){return ke(1984,2304,e,n)}function Me(e,n){return ke(1984,2304,e,n,f.downlevelIteration?Fe:Re)}function Oe(t,r,a){var i=[],o=t.initializer;if(e.isVariableDeclarationList(o)){3&t.initializer.flags&&zn();var s=e.firstOrUndefined(o.declarations);if(s&&e.isBindingPattern(s.name)){var c=e.flattenDestructuringBinding(s,x,n,0,r),l=e.setTextRange(e.createVariableDeclarationList(c),t.initializer);e.setOriginalNode(l,t.initializer),e.setSourceMapRange(l,e.createRange(c[0].pos,e.last(c).end)),i.push(e.createVariableStatement(void 0,l))}else i.push(e.setTextRange(e.createVariableStatement(void 0,e.setOriginalNode(e.setTextRange(e.createVariableDeclarationList([e.createVariableDeclaration(s?s.name:e.createTempVariable(void 0),void 0,r)]),e.moveRangePos(o,-1)),o)),e.moveRangeEnd(o,-1)))}else{var u=e.createAssignment(o,r);e.isDestructuringAssignment(u)?(e.aggregateTransformFlags(u),i.push(e.createExpressionStatement(ye(u,!1)))):(u.end=o.end,i.push(e.setTextRange(e.createExpressionStatement(e.visitNode(u,x,e.isExpression)),e.moveRangeEnd(o,-1))))}if(a)return Pe(e.addRange(i,a));var d=e.visitNode(t.statement,x,e.isStatement,e.liftToBlock);return e.isBlock(d)?e.updateBlock(d,e.setTextRange(e.createNodeArray(e.concatenate(i,d.statements)),d.statements)):(i.push(d),Pe(i))}function Pe(n){return e.setEmitFlags(e.createBlock(e.createNodeArray(n),!0),432)}function Re(n,t,r){var a=e.visitNode(n.expression,x,e.isExpression),i=e.createLoopVariable(),s=e.isIdentifier(a)?e.getGeneratedNameForNode(a):e.createTempVariable(void 0);e.setEmitFlags(a,48|e.getEmitFlags(a));var c=e.setTextRange(e.createFor(e.setEmitFlags(e.setTextRange(e.createVariableDeclarationList([e.setTextRange(e.createVariableDeclaration(i,void 0,e.createLiteral(0)),e.moveRangePos(n.expression,-1)),e.setTextRange(e.createVariableDeclaration(s,void 0,a),n.expression)]),n.expression),2097152),e.setTextRange(e.createLessThan(i,e.createPropertyAccess(s,"length")),n.expression),e.setTextRange(e.createPostfixIncrement(i),n.expression),Oe(n,e.createElementAccess(s,i),r)),n);return e.setEmitFlags(c,256),e.setTextRange(c,n),e.restoreEnclosingLabel(c,t,o&&Ae)}function Fe(t,r,a){var i=e.visitNode(t.expression,x,e.isExpression),s=e.isIdentifier(i)?e.getGeneratedNameForNode(i):e.createTempVariable(void 0),c=e.isIdentifier(i)?e.getGeneratedNameForNode(s):e.createTempVariable(void 0),l=e.createUniqueName("e"),u=e.getGeneratedNameForNode(l),d=e.createTempVariable(void 0),p=e.createValuesHelper(n,i,t.expression),f=e.createCall(e.createPropertyAccess(s,"next"),void 0,[]);m(l),m(d);var _=e.setEmitFlags(e.setTextRange(e.createFor(e.setEmitFlags(e.setTextRange(e.createVariableDeclarationList([e.setTextRange(e.createVariableDeclaration(s,void 0,p),t.expression),e.createVariableDeclaration(c,void 0,f)]),t.expression),2097152),e.createLogicalNot(e.createPropertyAccess(c,"done")),e.createAssignment(c,f),Oe(t,e.createPropertyAccess(c,"value"),a)),t),256);return e.createTry(e.createBlock([e.restoreEnclosingLabel(_,r,o&&Ae)]),e.createCatchClause(e.createVariableDeclaration(u),e.setEmitFlags(e.createBlock([e.createExpressionStatement(e.createAssignment(l,e.createObjectLiteral([e.createPropertyAssignment("error",u)])))]),1)),e.createBlock([e.createTry(e.createBlock([e.setEmitFlags(e.createIf(e.createLogicalAnd(e.createLogicalAnd(c,e.createLogicalNot(e.createPropertyAccess(c,"done"))),e.createAssignment(d,e.createPropertyAccess(s,"return"))),e.createExpressionStatement(e.createFunctionCall(d,s,[]))),1)]),void 0,e.setEmitFlags(e.createBlock([e.setEmitFlags(e.createIf(l,e.createThrow(e.createPropertyAccess(l,"error"))),1)]),1))]))}function Ge(t){for(var r=t.properties,i=r.length,o=i,s=i,c=0;c<i;c++){var l=r[c];if(4194304&l.transformFlags&&4&a&&c<s&&(s=c),149===l.name.kind){o=c;break}}if(o!==i){s<o&&(o=s);var u=e.createTempVariable(m),d=[],p=e.createAssignment(u,e.setEmitFlags(e.createObjectLiteral(e.visitNodes(r,x,e.isObjectLiteralElementLike,0,o),t.multiLine),65536));return t.multiLine&&e.startOnNewLine(p),d.push(p),pn(d,t,u,o),d.push(t.multiLine?e.startOnNewLine(e.getMutableClone(u)):u),e.inlineExpressions(d)}return e.visitEachChild(t,x,n)}function Ve(e){return 0!==(131072&_.getNodeCheckFlags(e))}function Be(n){return e.isForStatement(n)&&!!n.initializer&&Ve(n.initializer)}function Ke(n){return e.isForStatement(n)&&!!n.condition&&Ve(n.condition)}function je(n){return e.isForStatement(n)&&!!n.incrementor&&Ve(n.incrementor)}function Ue(e){return He(e)||Be(e)}function He(e){return 0!==(65536&_.getNodeCheckFlags(e))}function We(n,t){function r(t){if(72===t.kind)n.hoistedLocalVariables.push(t);else for(var a=0,i=t.elements;a<i.length;a++){var o=i[a];e.isOmittedExpression(o)||r(o.name)}}n.hoistedLocalVariables||(n.hoistedLocalVariables=[]),r(t.name)}function qe(t,r,a){if(!Ue(t)){var i=void 0;o&&(i=o.allowedNonLabeledJumps,o.allowedNonLabeledJumps=6);var s=a?a(t,r,void 0):e.restoreEnclosingLabel(e.visitEachChild(t,x,n),r,o&&Ae);return o&&(o.allowedNonLabeledJumps=i),s}var c=$e(t),l=[],u=o;o=c;var d,p=Be(t)?tn(t,c):void 0,m=He(t)?rn(t,c,u):void 0;if(o=u,p&&l.push(p.functionDeclaration),m&&l.push(m.functionDeclaration),en(l,c,u),p&&l.push(sn(p.functionName,p.containsYield)),m)if(a)d=a(t,r,m.part);else{var f=ze(t,p,e.createBlock(m.part,!0));e.aggregateTransformFlags(f),d=e.restoreEnclosingLabel(f,r,o&&Ae)}else{var _=ze(t,p,e.visitNode(t.statement,x,e.isStatement,e.liftToBlock));e.aggregateTransformFlags(_),d=e.restoreEnclosingLabel(_,r,o&&Ae)}return l.push(d),l}function ze(n,t,r){switch(n.kind){case 225:return Je(n,t,r);case 226:return Ye(n,r);case 227:return Xe(n,r);case 223:return Qe(n,r);case 224:return Ze(n,r);default:return e.Debug.failBadSyntaxKind(n,"IterationStatement expected")}}function Je(n,t,r){var a=n.condition&&Ve(n.condition),i=a||n.incrementor&&Ve(n.incrementor);return e.updateFor(n,e.visitNode(t?t.part:n.initializer,x,e.isForInitializer),e.visitNode(a?void 0:n.condition,x,e.isExpression),e.visitNode(i?void 0:n.incrementor,x,e.isExpression),r)}function Xe(n,t){return e.updateForOf(n,void 0,e.visitNode(n.initializer,x,e.isForInitializer),e.visitNode(n.expression,x,e.isExpression),t)}function Ye(n,t){return e.updateForIn(n,e.visitNode(n.initializer,x,e.isForInitializer),e.visitNode(n.expression,x,e.isExpression),t)}function Qe(n,t){return e.updateDo(n,t,e.visitNode(n.expression,x,e.isExpression))}function Ze(n,t){return e.updateWhile(n,e.visitNode(n.expression,x,e.isExpression),t)}function $e(n){var t;switch(n.kind){case 225:case 226:case 227:var r=n.initializer;r&&238===r.kind&&(t=r);break}var a=[],i=[];if(t&&3&e.getCombinedNodeFlags(t))for(var s=Be(n),c=0,l=t.declarations;c<l.length;c++){var u=l[c];dn(n,u,a,i,s)}var d={loopParameters:a,loopOutParameters:i};return o&&(o.argumentsName&&(d.argumentsName=o.argumentsName),o.thisName&&(d.thisName=o.thisName),o.hoistedLocalVariables&&(d.hoistedLocalVariables=o.hoistedLocalVariables)),d}function en(n,t,r){var a;if(t.argumentsName&&(r?r.argumentsName=t.argumentsName:(a||(a=[])).push(e.createVariableDeclaration(t.argumentsName,void 0,e.createIdentifier("arguments")))),t.thisName&&(r?r.thisName=t.thisName:(a||(a=[])).push(e.createVariableDeclaration(t.thisName,void 0,e.createIdentifier("this")))),t.hoistedLocalVariables)if(r)r.hoistedLocalVariables=t.hoistedLocalVariables;else{a||(a=[]);for(var i=0,o=t.hoistedLocalVariables;i<o.length;i++){var s=o[i];a.push(e.createVariableDeclaration(s))}}if(t.loopOutParameters.length){a||(a=[]);for(var c=0,l=t.loopOutParameters;c<l.length;c++){var u=l[c];a.push(e.createVariableDeclaration(u.outParamName))}}t.conditionVariable&&(a||(a=[]),a.push(e.createVariableDeclaration(t.conditionVariable,void 0,e.createFalse()))),a&&n.push(e.createVariableStatement(void 0,e.createVariableDeclarationList(a)))}function nn(n){return e.createVariableDeclaration(n.originalName,void 0,n.outParamName)}function tn(n,t){var r=e.createUniqueName("_loop_init"),i=0!==(4194304&n.initializer.transformFlags),o=0;t.containsLexicalThis&&(o|=8),i&&4&a&&(o|=262144);var s=[];s.push(e.createVariableStatement(void 0,n.initializer)),on(t.loopOutParameters,2,1,s);var c=e.createVariableStatement(void 0,e.setEmitFlags(e.createVariableDeclarationList([e.createVariableDeclaration(r,void 0,e.setEmitFlags(e.createFunctionExpression(void 0,i?e.createToken(40):void 0,void 0,void 0,void 0,void 0,e.visitNode(e.createBlock(s,!0),x,e.isBlock)),o))]),2097152)),l=e.createVariableDeclarationList(e.map(t.loopOutParameters,nn));return{functionName:r,containsYield:i,functionDeclaration:c,part:l}}function rn(n,t,r){var i=e.createUniqueName("_loop");u();var o=e.visitNode(n.statement,x,e.isStatement,e.liftToBlock),s=p(),c=[];(Ke(n)||je(n))&&(t.conditionVariable=e.createUniqueName("inc"),c.push(e.createIf(t.conditionVariable,e.createStatement(e.visitNode(n.incrementor,x,e.isExpression)),e.createStatement(e.createAssignment(t.conditionVariable,e.createTrue())))),Ke(n)&&c.push(e.createIf(e.createPrefix(52,e.visitNode(n.condition,x,e.isExpression)),e.visitNode(e.createBreak(),x,e.isStatement)))),e.isBlock(o)?e.addRange(c,o.statements):c.push(o),on(t.loopOutParameters,1,1,c),e.addStatementsAfterPrologue(c,s);var l=e.createBlock(c,!0);e.isBlock(o)&&e.setOriginalNode(l,o);var d=0!==(4194304&n.statement.transformFlags),m=0;t.containsLexicalThis&&(m|=8),d&&0!==(4&a)&&(m|=262144);var f=e.createVariableStatement(void 0,e.setEmitFlags(e.createVariableDeclarationList([e.createVariableDeclaration(i,void 0,e.setEmitFlags(e.createFunctionExpression(void 0,d?e.createToken(40):void 0,void 0,void 0,t.loopParameters,void 0,l),m))]),2097152)),_=cn(i,t,r,d);return{functionName:i,containsYield:d,functionDeclaration:f,part:_}}function an(n,t){var r=0===t?n.outParamName:n.originalName,a=0===t?n.originalName:n.outParamName;return e.createBinary(a,59,r)}function on(n,t,r,a){for(var i=0,o=n;i<o.length;i++){var s=o[i];s.flags&t&&a.push(e.createExpressionStatement(an(s,r)))}}function sn(n,t){var r=e.createCall(n,void 0,[]),a=t?e.createYield(e.createToken(40),e.setEmitFlags(r,8388608)):r;return e.createStatement(a)}function cn(n,t,r,a){var i=[],o=!(-5&t.nonLocalJumps)&&!t.labeledNonLocalBreaks&&!t.labeledNonLocalContinues,s=e.createCall(n,void 0,e.map(t.loopParameters,function(e){return e.name})),c=a?e.createYield(e.createToken(40),e.setEmitFlags(s,8388608)):s;if(o)i.push(e.createExpressionStatement(c)),on(t.loopOutParameters,1,0,i);else{var l=e.createUniqueName("state"),u=e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(l,void 0,c)]));if(i.push(u),on(t.loopOutParameters,1,0,i),8&t.nonLocalJumps){var d=void 0;r?(r.nonLocalJumps|=8,d=e.createReturn(l)):d=e.createReturn(e.createPropertyAccess(l,"value")),i.push(e.createIf(e.createBinary(e.createTypeOf(l),35,e.createLiteral("object")),d))}if(2&t.nonLocalJumps&&i.push(e.createIf(e.createBinary(l,35,e.createLiteral("break")),e.createBreak())),t.labeledNonLocalBreaks||t.labeledNonLocalContinues){var p=[];un(t.labeledNonLocalBreaks,!0,l,r,p),un(t.labeledNonLocalContinues,!1,l,r,p),i.push(e.createSwitch(l,e.createCaseBlock(p)))}}return i}function ln(n,t,r,a){t?(n.labeledNonLocalBreaks||(n.labeledNonLocalBreaks=e.createMap()),n.labeledNonLocalBreaks.set(r,a)):(n.labeledNonLocalContinues||(n.labeledNonLocalContinues=e.createMap()),n.labeledNonLocalContinues.set(r,a))}function un(n,t,r,a,i){n&&n.forEach(function(n,o){var s=[];if(!a||a.labels&&a.labels.get(o)){var c=e.createIdentifier(o);s.push(t?e.createBreak(c):e.createContinue(c))}else ln(a,t,o,n),s.push(e.createReturn(r));i.push(e.createCaseClause(e.createLiteral(n),s))})}function dn(n,t,r,a,i){var o=t.name;if(e.isBindingPattern(o))for(var s=0,c=o.elements;s<c.length;s++){var l=c[s];e.isOmittedExpression(l)||dn(n,l,r,a,i)}else{r.push(e.createParameter(void 0,void 0,void 0,o));var u=_.getNodeCheckFlags(t);if(4194304&u||i){var d=e.createUniqueName("out_"+e.idText(o)),p=0;4194304&u&&(p|=1),e.isForStatement(n)&&n.initializer&&_.isBindingCapturedByNode(n.initializer,t)&&(p|=2),a.push({flags:p,originalName:o,outParamName:d})}}}function pn(n,t,r,a){for(var i=t.properties,o=i.length,s=a;s<o;s++){var c=i[s];switch(c.kind){case 158:case 159:var l=e.getAllAccessorDeclarations(t.properties,c);c===l.firstAccessor&&n.push(ce(r,l,t,!!t.multiLine));break;case 156:n.push(_n(c,r,t,t.multiLine));break;case 275:n.push(mn(c,r,t.multiLine));break;case 276:n.push(fn(c,r,t.multiLine));break;default:e.Debug.failBadSyntaxKind(t);break}}}function mn(n,t,r){var a=e.createAssignment(e.createMemberAccessForPropertyName(t,e.visitNode(n.name,x,e.isPropertyName)),e.visitNode(n.initializer,x,e.isExpression));return e.setTextRange(a,n),r&&e.startOnNewLine(a),a}function fn(n,t,r){var a=e.createAssignment(e.createMemberAccessForPropertyName(t,e.visitNode(n.name,x,e.isPropertyName)),e.getSynthesizedClone(n.name));return e.setTextRange(a,n),r&&e.startOnNewLine(a),a}function _n(n,t,r,i){var o=b(0,0),s=e.createAssignment(e.createMemberAccessForPropertyName(t,e.visitNode(n.name,x,e.isPropertyName)),pe(n,n,void 0,r));return e.setTextRange(s,n),i&&e.startOnNewLine(s),E(o,49152,49152&a?16384:0),s}function gn(t){var r,a=b(4032,0);if(e.Debug.assert(!!t.variableDeclaration,"Catch clause variable should always be present when downleveling ES2015."),e.isBindingPattern(t.variableDeclaration.name)){var i=e.createTempVariable(void 0),o=e.createVariableDeclaration(i);e.setTextRange(o,t.variableDeclaration);var s=e.flattenDestructuringBinding(t.variableDeclaration,x,n,0,i),c=e.createVariableDeclarationList(s);e.setTextRange(c,t.variableDeclaration);var l=e.createVariableStatement(void 0,c);r=e.updateCatchClause(t,o,vn(t.block,l))}else r=e.visitEachChild(t,x,n);return E(a,0,0),r}function vn(n,t){var r=e.visitNodes(n.statements,x,e.isStatement);return e.updateBlock(n,[t].concat(r))}function yn(n){e.Debug.assert(!e.isComputedPropertyName(n.name));var t=pe(n,e.moveRangePos(n,-1),void 0,void 0);return e.setEmitFlags(t,512|e.getEmitFlags(t)),e.setTextRange(e.createPropertyAssignment(n.name,t),n)}function hn(t){e.Debug.assert(!e.isComputedPropertyName(t.name));var r=o;o=void 0;var a,i=b(16286,65),s=e.visitParameterList(t.parameters,x,n),c=16512&t.transformFlags?me(t):fe(t);return a=158===t.kind?e.updateGetAccessor(t,t.decorators,t.modifiers,t.name,s,t.type,c):e.updateSetAccessor(t,t.decorators,t.modifiers,t.name,s,c),E(i,49152,0),o=r,a}function bn(n){return e.setTextRange(e.createPropertyAssignment(n.name,e.getSynthesizedClone(n.name)),n)}function En(t){var r=b(0,8192),i=e.visitEachChild(t,x,n);return E(r,49152,49152&a?32768:0),i}function Tn(t){return e.visitEachChild(t,x,n)}function Sn(t){return 64&t.transformFlags?kn(t.elements,!0,!!t.multiLine,!!t.elements.hasTrailingComma):e.visitEachChild(t,x,n)}function xn(n){return 33554432&e.getEmitFlags(n)?Cn(n):64&n.transformFlags?Dn(n,!0):e.updateCall(n,e.visitNode(n.expression,A,e.isExpression),void 0,e.visitNodes(n.arguments,x,e.isExpression))}function Cn(n){var t=e.cast(e.cast(e.skipOuterExpressions(n.expression),e.isArrowFunction).body,e.isBlock),r=function(n){return e.isVariableStatement(n)&&!!e.first(n.declarationList.declarations).initializer},a=e.visitNodes(t.statements,x,e.isStatement),i=e.filter(a,r),o=e.filter(a,function(e){return!r(e)}),s=e.cast(e.first(i),e.isVariableStatement),c=s.declarationList.declarations[0],l=e.skipOuterExpressions(c.initializer),u=e.tryCast(l,e.isAssignmentExpression),d=e.cast(u?e.skipOuterExpressions(u.right):l,e.isCallExpression),p=e.cast(e.skipOuterExpressions(d.expression),e.isFunctionExpression),m=p.body.statements,f=0,_=-1,g=[];if(u){var v=e.tryCast(m[f],e.isExpressionStatement);v&&(g.push(v),f++),g.push(m[f]),f++,g.push(e.createExpressionStatement(e.createAssignment(u.left,e.cast(c.name,e.isIdentifier))))}while(!e.isReturnStatement(e.elementAt(m,_)))_--;return e.addRange(g,m,f,_),_<-1&&e.addRange(g,m,_+1),e.addRange(g,o),e.addRange(g,i,1),e.recreateOuterExpressions(n.expression,e.recreateOuterExpressions(c.initializer,e.recreateOuterExpressions(u&&u.right,e.updateCall(d,e.recreateOuterExpressions(d.expression,e.updateFunctionExpression(p,void 0,void 0,void 0,void 0,p.parameters,void 0,e.updateBlock(p.body,g))),void 0,d.arguments))))}function An(e){return Dn(e,!1)}function Dn(t,r){if(131072&t.transformFlags||98===t.expression.kind||e.isSuperProperty(e.skipOuterExpressions(t.expression))){var a=e.createCallBinding(t.expression,m),i=a.target,o=a.thisArg;98===t.expression.kind&&e.setEmitFlags(o,4);var s=void 0;if(s=131072&t.transformFlags?e.createFunctionApply(e.visitNode(i,A,e.isExpression),e.visitNode(o,x,e.isExpression),kn(t.arguments,!1,!1,!1)):e.createFunctionCall(e.visitNode(i,A,e.isExpression),e.visitNode(o,x,e.isExpression),e.visitNodes(t.arguments,x,e.isExpression),t),98===t.expression.kind){var c=e.createThis();e.setEmitFlags(c,4);var l=e.createLogicalOr(s,c);s=r?e.createAssignment(e.createFileLevelUniqueName("_this"),l):l}return e.setOriginalNode(s,t)}return e.visitEachChild(t,x,n)}function Ln(t){if(131072&t.transformFlags){var r=e.createCallBinding(e.createPropertyAccess(t.expression,"bind"),m),a=r.target,i=r.thisArg;return e.createNew(e.createFunctionApply(e.visitNode(a,x,e.isExpression),i,kn(e.createNodeArray([e.createVoidZero()].concat(t.arguments)),!1,!1,!1)),void 0,[])}return e.visitEachChild(t,x,n)}function kn(t,r,a,i){var o=t.length,s=e.flatten(e.spanMap(t,Nn,function(e,n,t,r){return n(e,a,i&&r===o)}));if(f.downlevelIteration){if(1===s.length){var c=s[0];if(e.isCallExpression(c)&&e.isIdentifier(c.expression)&&4096&e.getEmitFlags(c.expression)&&"___spread"===c.expression.escapedText)return s[0]}return e.createSpreadHelper(n,s)}if(1===s.length){var l=t[0];return r&&e.isSpreadElement(l)&&187!==l.expression.kind?e.createArraySlice(s[0]):s[0]}return e.createArrayConcat(s.shift(),s)}function Nn(n){return e.isSpreadElement(n)?In:wn}function In(n){return e.map(n,On)}function wn(n,t,r){return e.createArrayLiteral(e.visitNodes(e.createNodeArray(n,r),x,e.isExpression),t)}function Mn(n){return e.visitNode(n.expression,x,e.isExpression)}function On(n){return e.visitNode(n.expression,x,e.isExpression)}function Pn(n){return e.setTextRange(e.createLiteral(n.text),n)}function Rn(n){return n.hasExtendedUnicodeEscape?e.setTextRange(e.createLiteral(n.text),n):n}function Fn(n){return 384&n.numericLiteralFlags?e.setTextRange(e.createNumericLiteral(n.text),n):n}function Gn(r){var a=e.visitNode(r.tag,x,e.isExpression),i=[void 0],o=[],s=[],c=r.template;if(e.isNoSubstitutionTemplateLiteral(c))o.push(e.createLiteral(c.text)),s.push(Vn(c));else{o.push(e.createLiteral(c.head.text)),s.push(Vn(c.head));for(var u=0,d=c.templateSpans;u<d.length;u++){var p=d[u];o.push(e.createLiteral(p.literal.text)),s.push(Vn(p.literal)),i.push(e.visitNode(p.expression,x,e.isExpression))}}var m=l(n,e.createArrayLiteral(o),e.createArrayLiteral(s));if(e.isExternalModule(t)){var f=e.createUniqueName("templateObject");y(f),i[0]=e.createLogicalOr(f,e.createAssignment(f,m))}else i[0]=m;return e.createCall(a,void 0,i)}function Vn(n){var r=e.getSourceTextOfNodeFromSourceFile(t,n),a=14===n.kind||17===n.kind;return r=r.substring(1,r.length-(a?1:2)),r=r.replace(/\r\n?/g,"\n"),e.setTextRange(e.createLiteral(r),n)}function Bn(n){var t=[];jn(t,n),Un(t,n);var r=e.reduceLeft(t,e.createAdd);return e.nodeIsSynthesized(r)&&(r.pos=n.pos,r.end=n.end),r}function Kn(n){return e.Debug.assert(0!==n.templateSpans.length),0!==n.head.text.length||0===n.templateSpans[0].literal.text.length}function jn(n,t){Kn(t)&&n.push(e.createLiteral(t.head.text))}function Un(n,t){for(var r=0,a=t.templateSpans;r<a.length;r++){var i=a[r];n.push(e.visitNode(i.expression,x,e.isExpression)),0!==i.literal.text.length&&n.push(e.createLiteral(i.literal.text))}}function Hn(n){return 8&a&&!n?e.createPropertyAccess(e.createFileLevelUniqueName("_super"),"prototype"):e.createFileLevelUniqueName("_super")}function Wn(n){return 95===n.keywordToken&&"target"===n.name.escapedText?(a|=8192&a?32768:16384,e.createFileLevelUniqueName("_newTarget")):n}function qn(n,t,r){if(1&s&&e.isFunctionLike(t)){var a=b(16286,8&e.getEmitFlags(t)?81:65);return v(n,t,r),void E(a,0,0)}v(n,t,r)}function zn(){0===(2&s)&&(s|=2,n.enableSubstitution(72))}function Jn(){0===(1&s)&&(s|=1,n.enableSubstitution(100),n.enableEmitNotification(157),n.enableEmitNotification(156),n.enableEmitNotification(158),n.enableEmitNotification(159),n.enableEmitNotification(197),n.enableEmitNotification(196),n.enableEmitNotification(239))}function Xn(n,t){return t=g(n,t),1===n?Zn(t):e.isIdentifier(t)?Yn(t):t}function Yn(n){if(2&s&&!e.isInternalName(n)){var t=e.getParseTreeNode(n,e.isIdentifier);if(t&&Qn(t))return e.setTextRange(e.getGeneratedNameForNode(t),n)}return n}function Qn(e){switch(e.parent.kind){case 186:case 240:case 243:case 237:return e.parent.name===e&&_.isDeclarationWithCollidingName(e.parent)}return!1}function Zn(e){switch(e.kind){case 72:return $n(e);case 100:return nt(e)}return e}function $n(n){if(2&s&&!e.isInternalName(n)){var t=_.getReferencedDeclarationWithCollidingName(n);if(t&&(!e.isClassLike(t)||!et(t,n)))return e.setTextRange(e.getGeneratedNameForNode(e.getNameOfDeclaration(t)),n)}return n}function et(n,t){var r=e.getParseTreeNode(t);if(!r||r===n||r.end<=n.pos||r.pos>=n.end)return!1;var a=e.getEnclosingBlockScopeContainer(n);while(r){if(r===a||r===n)return!1;if(e.isClassElement(r)&&r.parent===n)return!0;r=r.parent}return!1}function nt(n){return 1&s&&16&a?e.setTextRange(e.createFileLevelUniqueName("_this"),n):n}function tt(n,t){return e.hasModifier(t,32)?e.getInternalName(n):e.createPropertyAccess(e.getInternalName(n),"prototype")}function rt(n,t){if(!n||!t)return!1;if(e.some(n.parameters))return!1;var r=e.firstOrUndefined(n.body.statements);if(!r||!e.nodeIsSynthesized(r)||221!==r.kind)return!1;var a=r.expression;if(!e.nodeIsSynthesized(a)||191!==a.kind)return!1;var i=a.expression;if(!e.nodeIsSynthesized(i)||98!==i.kind)return!1;var o=e.singleOrUndefined(a.arguments);if(!o||!e.nodeIsSynthesized(o)||208!==o.kind)return!1;var s=o.expression;return e.isIdentifier(s)&&"arguments"===s.escapedText}}function c(n,t){return n.requestEmitHelper(u),e.createCall(e.getHelperName("__extends"),void 0,[t,e.createFileLevelUniqueName("_super")])}function l(n,t,r){return n.requestEmitHelper(d),e.createCall(e.getHelperName("__makeTemplateObject"),void 0,[t,r])}(function(e){e[e["CapturedThis"]=1]="CapturedThis",e[e["BlockScopedBindings"]=2]="BlockScopedBindings"})(n||(n={})),function(e){e[e["Body"]=1]="Body",e[e["Initializer"]=2]="Initializer"}(t||(t={})),function(e){e[e["ToOriginal"]=0]="ToOriginal",e[e["ToOutParameter"]=1]="ToOutParameter"}(r||(r={})),function(e){e[e["Break"]=2]="Break",e[e["Continue"]=4]="Continue",e[e["Return"]=8]="Return"}(a||(a={})),function(e){e[e["NoReplacement"]=0]="NoReplacement",e[e["ReplaceSuperCapture"]=1]="ReplaceSuperCapture",e[e["ReplaceWithReturn"]=2]="ReplaceWithReturn"}(i||(i={})),function(e){e[e["None"]=0]="None",e[e["Function"]=1]="Function",e[e["ArrowFunction"]=2]="ArrowFunction",e[e["AsyncFunctionBody"]=4]="AsyncFunctionBody",e[e["NonStaticClassElement"]=8]="NonStaticClassElement",e[e["CapturesThis"]=16]="CapturesThis",e[e["ExportedVariableStatement"]=32]="ExportedVariableStatement",e[e["TopLevel"]=64]="TopLevel",e[e["Block"]=128]="Block",e[e["IterationStatement"]=256]="IterationStatement",e[e["IterationStatementBlock"]=512]="IterationStatementBlock",e[e["ForStatement"]=1024]="ForStatement",e[e["ForInOrForOfStatement"]=2048]="ForInOrForOfStatement",e[e["ConstructorWithCapturedSuper"]=4096]="ConstructorWithCapturedSuper",e[e["ComputedPropertyName"]=8192]="ComputedPropertyName",e[e["AncestorFactsMask"]=16383]="AncestorFactsMask",e[e["BlockScopeIncludes"]=0]="BlockScopeIncludes",e[e["BlockScopeExcludes"]=4032]="BlockScopeExcludes",e[e["SourceFileIncludes"]=64]="SourceFileIncludes",e[e["SourceFileExcludes"]=3968]="SourceFileExcludes",e[e["FunctionIncludes"]=65]="FunctionIncludes",e[e["FunctionExcludes"]=16286]="FunctionExcludes",e[e["AsyncFunctionBodyIncludes"]=69]="AsyncFunctionBodyIncludes",e[e["AsyncFunctionBodyExcludes"]=16278]="AsyncFunctionBodyExcludes",e[e["ArrowFunctionIncludes"]=66]="ArrowFunctionIncludes",e[e["ArrowFunctionExcludes"]=16256]="ArrowFunctionExcludes",e[e["ConstructorIncludes"]=73]="ConstructorIncludes",e[e["ConstructorExcludes"]=16278]="ConstructorExcludes",e[e["DoOrWhileStatementIncludes"]=256]="DoOrWhileStatementIncludes",e[e["DoOrWhileStatementExcludes"]=0]="DoOrWhileStatementExcludes",e[e["ForStatementIncludes"]=1280]="ForStatementIncludes",e[e["ForStatementExcludes"]=3008]="ForStatementExcludes",e[e["ForInOrForOfStatementIncludes"]=2304]="ForInOrForOfStatementIncludes",e[e["ForInOrForOfStatementExcludes"]=1984]="ForInOrForOfStatementExcludes",e[e["BlockIncludes"]=128]="BlockIncludes",e[e["BlockExcludes"]=3904]="BlockExcludes",e[e["IterationStatementBlockIncludes"]=512]="IterationStatementBlockIncludes",e[e["IterationStatementBlockExcludes"]=4032]="IterationStatementBlockExcludes",e[e["ComputedPropertyNameIncludes"]=8192]="ComputedPropertyNameIncludes",e[e["ComputedPropertyNameExcludes"]=0]="ComputedPropertyNameExcludes",e[e["NewTarget"]=16384]="NewTarget",e[e["NewTargetInComputedPropertyName"]=32768]="NewTargetInComputedPropertyName",e[e["SubtreeFactsMask"]=-16384]="SubtreeFactsMask",e[e["PropagateNewTargetMask"]=49152]="PropagateNewTargetMask"}(o||(o={})),e.transformES2015=s;var u={name:"typescript:extends",scoped:!1,priority:0,text:"\n            var __extends = (this && this.__extends) || (function () {\n                var extendStatics = function (d, b) {\n                    extendStatics = Object.setPrototypeOf ||\n                        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||\n                        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };\n                    return extendStatics(d, b);\n                };\n\n                return function (d, b) {\n                    extendStatics(d, b);\n                    function __() { this.constructor = d; }\n                    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());\n                };\n            })();"},d={name:"typescript:makeTemplateObject",scoped:!1,priority:0,text:'\n            var __makeTemplateObject = (this && this.__makeTemplateObject) || function (cooked, raw) {\n                if (Object.defineProperty) { Object.defineProperty(cooked, "raw", { value: raw }); } else { cooked.raw = raw; }\n                return cooked;\n            };'}}(u||(u={})),function(e){function n(n){var t,r,a=n.getCompilerOptions();1!==a.jsx&&3!==a.jsx||(t=n.onEmitNode,n.onEmitNode=s,n.enableEmitNotification(262),n.enableEmitNotification(263),n.enableEmitNotification(261),r=[]);var i=n.onSubstituteNode;return n.onSubstituteNode=c,n.enableSubstitution(189),n.enableSubstitution(275),e.chainBundle(o);function o(e){return e}function s(n,a,i){switch(a.kind){case 262:case 263:case 261:var o=a.tagName;r[e.getOriginalNodeId(o)]=!0;break}t(n,a,i)}function c(n,t){return t.id&&r&&r[t.id]?i(n,t):(t=i(n,t),e.isPropertyAccessExpression(t)?l(t):e.isPropertyAssignment(t)?u(t):t)}function l(n){var t=d(n.name);return t?e.setTextRange(e.createElementAccess(n.expression,t),n):n}function u(n){var t=e.isIdentifier(n.name)&&d(n.name);return t?e.updatePropertyAssignment(n,t,n.initializer):n}function d(n){var t=n.originalKeywordKind||(e.nodeIsSynthesized(n)?e.stringToToken(e.idText(n)):void 0);if(void 0!==t&&t>=73&&t<=108)return e.setTextRange(e.createLiteral(n),n)}}e.transformES5=n}(u||(u={})),function(e){var n,t,r,a,i;function o(e){switch(e){case 2:return"return";case 3:return"break";case 4:return"yield";case 5:return"yield*";case 7:return"endfinally";default:return}}function s(n){var t,r,a,i,s,l,u,d,p,m,f=n.resumeLexicalEnvironment,_=n.endLexicalEnvironment,g=n.hoistFunctionDeclaration,v=n.hoistVariableDeclaration,y=n.getCompilerOptions(),h=e.getEmitScriptTarget(y),b=n.getEmitResolver(),E=n.onSubstituteNode;n.onSubstituteNode=Ke;var T,S,x,C,A,D,L,k,N,I,w,M,O=1,P=0,R=0;return e.chainBundle(F);function F(t){if(t.isDeclarationFile||0===(512&t.transformFlags))return t;var r=e.visitEachChild(t,G,n);return e.addEmitHelpers(r,n.readEmitHelpers()),r}function G(t){var r=t.transformFlags;return i?V(t):a?B(t):256&r?j(t):512&r?e.visitEachChild(t,G,n):t}function V(e){switch(e.kind){case 223:return he(e);case 224:return Ee(e);case 232:return Oe(e);case 233:return Re(e);default:return B(e)}}function B(t){switch(t.kind){case 239:return U(t);case 196:return H(t);case 158:case 159:return W(t);case 219:return z(t);case 225:return Se(t);case 226:return Ce(t);case 229:return ke(t);case 228:return De(t);case 230:return Ie(t);default:return 4194304&t.transformFlags?K(t):8389120&t.transformFlags?e.visitEachChild(t,G,n):t}}function K(t){switch(t.kind){case 204:return J(t);case 205:return ne(t);case 207:return te(t);case 187:return re(t);case 188:return ie(t);case 190:return oe(t);case 191:return se(t);case 192:return ce(t);default:return e.visitEachChild(t,G,n)}}function j(n){switch(n.kind){case 239:return U(n);case 196:return H(n);default:return e.Debug.failBadSyntaxKind(n)}}function U(t){if(t.asteriskToken)t=e.setOriginalNode(e.setTextRange(e.createFunctionDeclaration(void 0,t.modifiers,void 0,t.name,void 0,e.visitParameterList(t.parameters,G,n),void 0,q(t.body)),t),t);else{var r=a,o=i;a=!1,i=!1,t=e.visitEachChild(t,G,n),a=r,i=o}return a?void g(t):t}function H(t){if(t.asteriskToken)t=e.setOriginalNode(e.setTextRange(e.createFunctionExpression(void 0,void 0,t.name,void 0,e.visitParameterList(t.parameters,G,n),void 0,q(t.body)),t),t);else{var r=a,o=i;a=!1,i=!1,t=e.visitEachChild(t,G,n),a=r,i=o}return t}function W(t){var r=a,o=i;return a=!1,i=!1,t=e.visitEachChild(t,G,n),a=r,i=o,t}function q(n){var t=[],r=a,o=i,c=s,g=l,v=u,y=d,h=p,b=m,E=O,A=T,D=S,L=x,k=C;a=!0,i=!1,s=void 0,l=void 0,u=void 0,d=void 0,p=void 0,m=void 0,O=1,T=void 0,S=void 0,x=void 0,C=e.createTempVariable(void 0),f();var N=e.addPrologue(t,n.statements,!1,G);le(n.statements,N);var I=Fn();return e.addStatementsAfterPrologue(t,_()),t.push(e.createReturn(I)),a=r,i=o,s=c,l=g,u=v,d=y,p=h,m=b,O=E,T=A,S=D,x=L,C=k,e.setTextRange(e.createBlock(t,n.multiLine),n)}function z(n){if(4194304&n.transformFlags)_e(n.declarationList);else{if(1048576&e.getEmitFlags(n))return n;for(var t=0,r=n.declarationList.declarations;t<r.length;t++){var a=r[t];v(a.name)}var i=e.getInitializedVariables(n.declarationList);if(0!==i.length)return e.setSourceMapRange(e.createExpressionStatement(e.inlineExpressions(e.map(i,ge))),n)}}function J(n){var t=e.getExpressionAssociativity(n);switch(t){case 0:return Z(n);case 1:return Q(n);default:return e.Debug.assertNever(t)}}function X(e){return e>=60&&e<=71}function Y(e){switch(e){case 60:return 38;case 61:return 39;case 62:return 40;case 63:return 41;case 64:return 42;case 65:return 43;case 66:return 46;case 67:return 47;case 68:return 48;case 69:return 49;case 70:return 50;case 71:return 51}}function Q(t){var r=t.left,a=t.right;if(Ve(a)){var i=void 0;switch(r.kind){case 189:i=e.updatePropertyAccess(r,He(e.visitNode(r.expression,G,e.isLeftHandSideExpression)),r.name);break;case 190:i=e.updateElementAccess(r,He(e.visitNode(r.expression,G,e.isLeftHandSideExpression)),He(e.visitNode(r.argumentExpression,G,e.isExpression)));break;default:i=e.visitNode(r,G,e.isExpression);break}var o=t.operatorToken.kind;return X(o)?e.setTextRange(e.createAssignment(i,e.setTextRange(e.createBinary(He(i),Y(o),e.visitNode(a,G,e.isExpression)),t)),t):e.updateBinary(t,i,e.visitNode(a,G,e.isExpression))}return e.visitEachChild(t,G,n)}function Z(t){if(Ve(t.right)){if(e.isLogicalOperator(t.operatorToken.kind))return $(t);if(27===t.operatorToken.kind)return ee(t);var r=e.getMutableClone(t);return r.left=He(e.visitNode(t.left,G,e.isExpression)),r.right=e.visitNode(t.right,G,e.isExpression),r}return e.visitEachChild(t,G,n)}function $(n){var t=qe(),r=We();return Dn(r,e.visitNode(n.left,G,e.isExpression),n.left),54===n.operatorToken.kind?Nn(t,r,n.left):kn(t,r,n.left),Dn(r,e.visitNode(n.right,G,e.isExpression),n.right),ze(t),r}function ee(n){var t=[];return r(n.left),r(n.right),e.inlineExpressions(t);function r(n){e.isBinaryExpression(n)&&27===n.operatorToken.kind?(r(n.left),r(n.right)):(Ve(n)&&t.length>0&&(Rn(1,[e.createExpressionStatement(e.inlineExpressions(t))]),t=[]),t.push(e.visitNode(n,G,e.isExpression)))}}function ne(t){if(Ve(t.whenTrue)||Ve(t.whenFalse)){var r=qe(),a=qe(),i=We();return Nn(r,e.visitNode(t.condition,G,e.isExpression),t.condition),Dn(i,e.visitNode(t.whenTrue,G,e.isExpression),t.whenTrue),Ln(a),ze(r),Dn(i,e.visitNode(t.whenFalse,G,e.isExpression),t.whenFalse),ze(a),i}return e.visitEachChild(t,G,n)}function te(t){var r=qe(),a=e.visitNode(t.expression,G,e.isExpression);if(t.asteriskToken){var i=0===(8388608&e.getEmitFlags(t.expression))?e.createValuesHelper(n,a,t):a;In(i,t)}else wn(a,t);return ze(r),xn(t)}function re(e){return ae(e.elements,void 0,void 0,e.multiLine)}function ae(n,t,r,a){var i,o=Be(n);if(o>0){i=We();var s=e.visitNodes(n,G,e.isExpression,0,o);Dn(i,e.createArrayLiteral(t?[t].concat(s):s)),t=void 0}var c=e.reduceLeft(n,l,[],o);return i?e.createArrayConcat(i,[e.createArrayLiteral(c,a)]):e.setTextRange(e.createArrayLiteral(t?[t].concat(c):c,a),r);function l(n,r){if(Ve(r)&&n.length>0){var o=void 0!==i;i||(i=We()),Dn(i,o?e.createArrayConcat(i,[e.createArrayLiteral(n,a)]):e.createArrayLiteral(t?[t].concat(n):n,a)),t=void 0,n=[]}return n.push(e.visitNode(r,G,e.isExpression)),n}}function ie(n){var t=n.properties,r=n.multiLine,a=Be(t),i=We();Dn(i,e.createObjectLiteral(e.visitNodes(t,G,e.isObjectLiteralElementLike,0,a),r));var o=e.reduceLeft(t,s,[],a);return o.push(r?e.startOnNewLine(e.getMutableClone(i)):i),e.inlineExpressions(o);function s(t,a){Ve(a)&&t.length>0&&(An(e.createExpressionStatement(e.inlineExpressions(t))),t=[]);var o=e.createExpressionForObjectLiteralElementLike(n,a,i),s=e.visitNode(o,G,e.isExpression);return s&&(r&&e.startOnNewLine(s),t.push(s)),t}}function oe(t){if(Ve(t.argumentExpression)){var r=e.getMutableClone(t);return r.expression=He(e.visitNode(t.expression,G,e.isLeftHandSideExpression)),r.argumentExpression=e.visitNode(t.argumentExpression,G,e.isExpression),r}return e.visitEachChild(t,G,n)}function se(t){if(!e.isImportCall(t)&&e.forEach(t.arguments,Ve)){var r=e.createCallBinding(t.expression,v,h,!0),a=r.target,i=r.thisArg;return e.setOriginalNode(e.createFunctionApply(He(e.visitNode(a,G,e.isLeftHandSideExpression)),i,ae(t.arguments),t),t)}return e.visitEachChild(t,G,n)}function ce(t){if(e.forEach(t.arguments,Ve)){var r=e.createCallBinding(e.createPropertyAccess(t.expression,"bind"),v),a=r.target,i=r.thisArg;return e.setOriginalNode(e.setTextRange(e.createNew(e.createFunctionApply(He(e.visitNode(a,G,e.isExpression)),i,ae(t.arguments,e.createVoidZero())),void 0,[]),t),t)}return e.visitEachChild(t,G,n)}function le(e,n){void 0===n&&(n=0);for(var t=e.length,r=n;r<t;r++)de(e[r])}function ue(n){e.isBlock(n)?le(n.statements):de(n)}function de(e){var n=i;i||(i=Ve(e)),pe(e),i=n}function pe(n){switch(n.kind){case 218:return me(n);case 221:return fe(n);case 222:return ve(n);case 223:return ye(n);case 224:return be(n);case 225:return Te(n);case 226:return xe(n);case 228:return Ae(n);case 229:return Le(n);case 230:return Ne(n);case 231:return we(n);case 232:return Me(n);case 233:return Pe(n);case 234:return Fe(n);case 235:return Ge(n);default:return An(e.visitNode(n,G,e.isStatement))}}function me(n){Ve(n)?le(n.statements):An(e.visitNode(n,G,e.isStatement))}function fe(n){An(e.visitNode(n,G,e.isStatement))}function _e(n){for(var t=0,r=n.declarations;t<r.length;t++){var a=r[t],i=e.getSynthesizedClone(a.name);e.setCommentRange(i,a.name),v(i)}var o=e.getInitializedVariables(n),s=o.length,c=0,l=[];while(c<s){for(var u=c;u<s;u++){a=o[u];if(Ve(a.initializer)&&l.length>0)break;l.push(ge(a))}l.length&&(An(e.createExpressionStatement(e.inlineExpressions(l))),c+=l.length,l=[])}}function ge(n){return e.setSourceMapRange(e.createAssignment(e.setSourceMapRange(e.getSynthesizedClone(n.name),n.name),e.visitNode(n.initializer,G,e.isExpression)),n)}function ve(n){if(Ve(n))if(Ve(n.thenStatement)||Ve(n.elseStatement)){var t=qe(),r=n.elseStatement?qe():void 0;Nn(n.elseStatement?r:t,e.visitNode(n.expression,G,e.isExpression),n.expression),ue(n.thenStatement),n.elseStatement&&(Ln(t),ze(r),ue(n.elseStatement)),ze(t)}else An(e.visitNode(n,G,e.isStatement));else An(e.visitNode(n,G,e.isStatement))}function ye(n){if(Ve(n)){var t=qe(),r=qe();on(t),ze(r),ue(n.statement),ze(t),kn(r,e.visitNode(n.expression,G,e.isExpression)),sn()}else An(e.visitNode(n,G,e.isStatement))}function he(t){return i?(an(),t=e.visitEachChild(t,G,n),sn(),t):e.visitEachChild(t,G,n)}function be(n){if(Ve(n)){var t=qe(),r=on(t);ze(t),Nn(r,e.visitNode(n.expression,G,e.isExpression)),ue(n.statement),Ln(t),sn()}else An(e.visitNode(n,G,e.isStatement))}function Ee(t){return i?(an(),t=e.visitEachChild(t,G,n),sn(),t):e.visitEachChild(t,G,n)}function Te(n){if(Ve(n)){var t=qe(),r=qe(),a=on(r);if(n.initializer){var i=n.initializer;e.isVariableDeclarationList(i)?_e(i):An(e.setTextRange(e.createExpressionStatement(e.visitNode(i,G,e.isExpression)),i))}ze(t),n.condition&&Nn(a,e.visitNode(n.condition,G,e.isExpression)),ue(n.statement),ze(r),n.incrementor&&An(e.setTextRange(e.createExpressionStatement(e.visitNode(n.incrementor,G,e.isExpression)),n.incrementor)),Ln(t),sn()}else An(e.visitNode(n,G,e.isStatement))}function Se(t){i&&an();var r=t.initializer;if(r&&e.isVariableDeclarationList(r)){for(var a=0,o=r.declarations;a<o.length;a++){var s=o[a];v(s.name)}var c=e.getInitializedVariables(r);t=e.updateFor(t,c.length>0?e.inlineExpressions(e.map(c,ge)):void 0,e.visitNode(t.condition,G,e.isExpression),e.visitNode(t.incrementor,G,e.isExpression),e.visitNode(t.statement,G,e.isStatement,e.liftToBlock))}else t=e.visitEachChild(t,G,n);return i&&sn(),t}function xe(n){if(Ve(n)){var t=We(),r=We(),a=e.createLoopVariable(),i=n.initializer;v(a),Dn(t,e.createArrayLiteral()),An(e.createForIn(r,e.visitNode(n.expression,G,e.isExpression),e.createExpressionStatement(e.createCall(e.createPropertyAccess(t,"push"),void 0,[r])))),Dn(a,e.createLiteral(0));var o=qe(),s=qe(),c=on(s);ze(o),Nn(c,e.createLessThan(a,e.createPropertyAccess(t,"length")));var l=void 0;if(e.isVariableDeclarationList(i)){for(var u=0,d=i.declarations;u<d.length;u++){var p=d[u];v(p.name)}l=e.getSynthesizedClone(i.declarations[0].name)}else l=e.visitNode(i,G,e.isExpression),e.Debug.assert(e.isLeftHandSideExpression(l));Dn(l,e.createElementAccess(t,a)),ue(n.statement),ze(s),An(e.createExpressionStatement(e.createPostfixIncrement(a))),Ln(o),sn()}else An(e.visitNode(n,G,e.isStatement))}function Ce(t){i&&an();var r=t.initializer;if(e.isVariableDeclarationList(r)){for(var a=0,o=r.declarations;a<o.length;a++){var s=o[a];v(s.name)}t=e.updateForIn(t,r.declarations[0].name,e.visitNode(t.expression,G,e.isExpression),e.visitNode(t.statement,G,e.isStatement,e.liftToBlock))}else t=e.visitEachChild(t,G,n);return i&&sn(),t}function Ae(n){var t=hn(n.label?e.idText(n.label):void 0);t>0?Ln(t,n):An(n)}function De(t){if(i){var r=hn(t.label&&e.idText(t.label));if(r>0)return Tn(r,t)}return e.visitEachChild(t,G,n)}function Le(n){var t=yn(n.label?e.idText(n.label):void 0);t>0?Ln(t,n):An(n)}function ke(t){if(i){var r=yn(t.label&&e.idText(t.label));if(r>0)return Tn(r,t)}return e.visitEachChild(t,G,n)}function Ne(n){Mn(e.visitNode(n.expression,G,e.isExpression),n)}function Ie(n){return Sn(e.visitNode(n.expression,G,e.isExpression),n)}function we(n){Ve(n)?(Ze(He(e.visitNode(n.expression,G,e.isExpression))),ue(n.statement),$e()):An(e.visitNode(n,G,e.isStatement))}function Me(n){if(Ve(n.caseBlock)){for(var t=n.caseBlock,r=t.clauses.length,a=ln(),i=He(e.visitNode(n.expression,G,e.isExpression)),o=[],s=-1,c=0;c<r;c++){var l=t.clauses[c];o.push(qe()),272===l.kind&&-1===s&&(s=c)}var u=0,d=[];while(u<r){var p=0;for(c=u;c<r;c++){l=t.clauses[c];if(271===l.kind){if(Ve(l.expression)&&d.length>0)break;d.push(e.createCaseClause(e.visitNode(l.expression,G,e.isExpression),[Tn(o[c],l.expression)]))}else p++}d.length&&(An(e.createSwitch(i,e.createCaseBlock(d))),u+=d.length,d=[]),p>0&&(u+=p,p=0)}Ln(s>=0?o[s]:a);for(c=0;c<r;c++)ze(o[c]),le(t.clauses[c].statements);un()}else An(e.visitNode(n,G,e.isStatement))}function Oe(t){return i&&cn(),t=e.visitEachChild(t,G,n),i&&un(),t}function Pe(n){Ve(n)?(pn(e.idText(n.label)),ue(n.statement),mn()):An(e.visitNode(n,G,e.isStatement))}function Re(t){return i&&dn(e.idText(t.label)),t=e.visitEachChild(t,G,n),i&&mn(),t}function Fe(n){On(e.visitNode(n.expression,G,e.isExpression),n)}function Ge(t){Ve(t)?(en(),ue(t.tryBlock),t.catchClause&&(nn(t.catchClause.variableDeclaration),ue(t.catchClause.block)),t.finallyBlock&&(tn(),ue(t.finallyBlock)),rn()):An(e.visitEachChild(t,G,n))}function Ve(e){return!!e&&0!==(4194304&e.transformFlags)}function Be(e){for(var n=e.length,t=0;t<n;t++)if(Ve(e[t]))return t;return-1}function Ke(e,n){return n=E(e,n),1===e?je(n):n}function je(n){return e.isIdentifier(n)?Ue(n):n}function Ue(n){if(!e.isGeneratedIdentifier(n)&&t&&t.has(e.idText(n))){var a=e.getOriginalNode(n);if(e.isIdentifier(a)&&a.parent){var i=b.getReferencedValueDeclaration(a);if(i){var o=r[e.getOriginalNodeId(i)];if(o){var s=e.getMutableClone(o);return e.setSourceMapRange(s,n),e.setCommentRange(s,n),s}}}}return n}function He(n){var t;return e.isGeneratedIdentifier(n)||4096&e.getEmitFlags(n)?n:(t=e.createTempVariable(v),Dn(t,n,n),t)}function We(n){var t=n?e.createUniqueName(n):e.createTempVariable(void 0);return v(t),t}function qe(){p||(p=[]);var e=O;return O++,p[e]=-1,e}function ze(n){e.Debug.assert(void 0!==p,"No labels were defined."),p[n]=T?T.length:0}function Je(e){s||(s=[],u=[],l=[],d=[]);var n=u.length;return u[n]=0,l[n]=T?T.length:0,s[n]=e,d.push(e),n}function Xe(){var n=Ye();if(void 0===n)return e.Debug.fail("beginBlock was never called.");var t=u.length;return u[t]=1,l[t]=T?T.length:0,s[t]=n,d.pop(),n}function Ye(){return e.lastOrUndefined(d)}function Qe(){var e=Ye();return e&&e.kind}function Ze(e){var n=qe(),t=qe();ze(n),Je({kind:1,expression:e,startLabel:n,endLabel:t})}function $e(){e.Debug.assert(1===Qe());var n=Xe();ze(n.endLabel)}function en(){var e=qe(),n=qe();return ze(e),Je({kind:0,state:0,startLabel:e,endLabel:n}),Cn(),n}function nn(a){var i;if(e.Debug.assert(0===Qe()),e.isGeneratedIdentifier(a.name))i=a.name,v(a.name);else{var o=e.idText(a.name);i=We(o),t||(t=e.createMap(),r=[],n.enableSubstitution(72)),t.set(o,!0),r[e.getOriginalNodeId(a)]=i}var s=Ye();e.Debug.assert(s.state<1);var c=s.endLabel;Ln(c);var l=qe();ze(l),s.state=1,s.catchVariable=i,s.catchLabel=l,Dn(i,e.createCall(e.createPropertyAccess(C,"sent"),void 0,[])),Cn()}function tn(){e.Debug.assert(0===Qe());var n=Ye();e.Debug.assert(n.state<2);var t=n.endLabel;Ln(t);var r=qe();ze(r),n.state=2,n.finallyLabel=r}function rn(){e.Debug.assert(0===Qe());var n=Xe(),t=n.state;t<2?Ln(n.endLabel):Pn(),ze(n.endLabel),Cn(),n.state=3}function an(){Je({kind:3,isScript:!0,breakLabel:-1,continueLabel:-1})}function on(e){var n=qe();return Je({kind:3,isScript:!1,breakLabel:n,continueLabel:e}),n}function sn(){e.Debug.assert(3===Qe());var n=Xe(),t=n.breakLabel;n.isScript||ze(t)}function cn(){Je({kind:2,isScript:!0,breakLabel:-1})}function ln(){var e=qe();return Je({kind:2,isScript:!1,breakLabel:e}),e}function un(){e.Debug.assert(2===Qe());var n=Xe(),t=n.breakLabel;n.isScript||ze(t)}function dn(e){Je({kind:4,isScript:!0,labelText:e,breakLabel:-1})}function pn(e){var n=qe();Je({kind:4,isScript:!1,labelText:e,breakLabel:n})}function mn(){e.Debug.assert(4===Qe());var n=Xe();n.isScript||ze(n.breakLabel)}function fn(e){return 2===e.kind||3===e.kind}function _n(e){return 4===e.kind}function gn(e){return 3===e.kind}function vn(e,n){for(var t=n;t>=0;t--){var r=d[t];if(!_n(r))break;if(r.labelText===e)return!0}return!1}function yn(e){if(d)if(e)for(var n=d.length-1;n>=0;n--){var t=d[n];if(_n(t)&&t.labelText===e)return t.breakLabel;if(fn(t)&&vn(e,n-1))return t.breakLabel}else for(n=d.length-1;n>=0;n--){t=d[n];if(fn(t))return t.breakLabel}return 0}function hn(e){if(d)if(e)for(var n=d.length-1;n>=0;n--){var t=d[n];if(gn(t)&&vn(e,n-1))return t.continueLabel}else for(n=d.length-1;n>=0;n--){t=d[n];if(gn(t))return t.continueLabel}return 0}function bn(n){if(void 0!==n&&n>0){void 0===m&&(m=[]);var t=e.createLiteral(-1);return void 0===m[n]?m[n]=[t]:m[n].push(t),t}return e.createOmittedExpression()}function En(n){var t=e.createLiteral(n);return e.addSyntheticTrailingComment(t,3,o(n)),t}function Tn(n,t){return e.Debug.assertLessThan(0,n,"Invalid label"),e.setTextRange(e.createReturn(e.createArrayLiteral([En(3),bn(n)])),t)}function Sn(n,t){return e.setTextRange(e.createReturn(e.createArrayLiteral(n?[En(2),n]:[En(2)])),t)}function xn(n){return e.setTextRange(e.createCall(e.createPropertyAccess(C,"sent"),void 0,[]),n)}function Cn(){Rn(0)}function An(e){e?Rn(1,[e]):Cn()}function Dn(e,n,t){Rn(2,[e,n],t)}function Ln(e,n){Rn(3,[e],n)}function kn(e,n,t){Rn(4,[e,n],t)}function Nn(e,n,t){Rn(5,[e,n],t)}function In(e,n){Rn(7,[e],n)}function wn(e,n){Rn(6,[e],n)}function Mn(e,n){Rn(8,[e],n)}function On(e,n){Rn(9,[e],n)}function Pn(){Rn(10)}function Rn(e,n,t){void 0===T&&(T=[],S=[],x=[]),void 0===p&&ze(qe());var r=T.length;T[r]=e,S[r]=n,x[r]=t}function Fn(){P=0,R=0,A=void 0,D=!1,L=!1,k=void 0,N=void 0,I=void 0,w=void 0,M=void 0;var t=Gn();return c(n,e.setEmitFlags(e.createFunctionExpression(void 0,void 0,void 0,void 0,[e.createParameter(void 0,void 0,void 0,C)],void 0,e.createBlock(t,t.length>0)),524288))}function Gn(){if(T){for(var n=0;n<T.length;n++)qn(n);Bn(T.length)}else Bn(0);if(k){var t=e.createPropertyAccess(C,"label"),r=e.createSwitch(t,e.createCaseBlock(k));return[e.startOnNewLine(r)]}return N||[]}function Vn(){N&&(jn(!D),D=!1,L=!1,R++)}function Bn(e){Kn(e)&&(Un(e),M=void 0,Yn(void 0,void 0)),N&&k&&jn(!1),Hn()}function Kn(e){if(!L)return!0;if(!p||!m)return!1;for(var n=0;n<p.length;n++)if(p[n]===e&&m[n])return!0;return!1}function jn(n){if(k||(k=[]),N){if(M)for(var t=M.length-1;t>=0;t--){var r=M[t];N=[e.createWith(r.expression,e.createBlock(N))]}if(w){var a=w.startLabel,i=w.catchLabel,o=w.finallyLabel,s=w.endLabel;N.unshift(e.createExpressionStatement(e.createCall(e.createPropertyAccess(e.createPropertyAccess(C,"trys"),"push"),void 0,[e.createArrayLiteral([bn(a),bn(i),bn(o),bn(s)])]))),w=void 0}n&&N.push(e.createExpressionStatement(e.createAssignment(e.createPropertyAccess(C,"label"),e.createLiteral(R+1))))}k.push(e.createCaseClause(e.createLiteral(R),N||[])),N=void 0}function Un(e){if(p)for(var n=0;n<p.length;n++)p[n]===e&&(Vn(),void 0===A&&(A=[]),void 0===A[R]?A[R]=[n]:A[R].push(n))}function Hn(){if(void 0!==m&&void 0!==A)for(var e=0;e<A.length;e++){var n=A[e];if(void 0!==n)for(var t=0,r=n;t<r.length;t++){var a=r[t],i=m[a];if(void 0!==i)for(var o=0,s=i;o<s.length;o++){var c=s[o];c.text=String(e)}}}}function Wn(e){if(s)for(;P<u.length&&l[P]<=e;P++){var n=s[P],t=u[P];switch(n.kind){case 0:0===t?(I||(I=[]),N||(N=[]),I.push(w),w=n):1===t&&(w=I.pop());break;case 1:0===t?(M||(M=[]),M.push(n)):1===t&&M.pop();break}}}function qn(e){if(Un(e),Wn(e),!D){D=!1,L=!1;var n=T[e];if(0!==n){if(10===n)return tt();var t=S[e];if(1===n)return zn(t[0]);var r=x[e];switch(n){case 2:return Jn(t[0],t[1],r);case 3:return Qn(t[0],r);case 4:return Zn(t[0],t[1],r);case 5:return $n(t[0],t[1],r);case 6:return et(t[0],r);case 7:return nt(t[0],r);case 8:return Yn(t[0],r);case 9:return Xn(t[0],r)}}}}function zn(e){e&&(N?N.push(e):N=[e])}function Jn(n,t,r){zn(e.setTextRange(e.createExpressionStatement(e.createAssignment(n,t)),r))}function Xn(n,t){D=!0,L=!0,zn(e.setTextRange(e.createThrow(n),t))}function Yn(n,t){D=!0,L=!0,zn(e.setEmitFlags(e.setTextRange(e.createReturn(e.createArrayLiteral(n?[En(2),n]:[En(2)])),t),384))}function Qn(n,t){D=!0,zn(e.setEmitFlags(e.setTextRange(e.createReturn(e.createArrayLiteral([En(3),bn(n)])),t),384))}function Zn(n,t,r){zn(e.setEmitFlags(e.createIf(t,e.setEmitFlags(e.setTextRange(e.createReturn(e.createArrayLiteral([En(3),bn(n)])),r),384)),1))}function $n(n,t,r){zn(e.setEmitFlags(e.createIf(e.createLogicalNot(t),e.setEmitFlags(e.setTextRange(e.createReturn(e.createArrayLiteral([En(3),bn(n)])),r),384)),1))}function et(n,t){D=!0,zn(e.setEmitFlags(e.setTextRange(e.createReturn(e.createArrayLiteral(n?[En(4),n]:[En(4)])),t),384))}function nt(n,t){D=!0,zn(e.setEmitFlags(e.setTextRange(e.createReturn(e.createArrayLiteral([En(5),n])),t),384))}function tt(){D=!0,zn(e.createReturn(e.createArrayLiteral([En(7)])))}}function c(n,t){return n.requestEmitHelper(l),e.createCall(e.getHelperName("__generator"),void 0,[e.createThis(),t])}(function(e){e[e["Nop"]=0]="Nop",e[e["Statement"]=1]="Statement",e[e["Assign"]=2]="Assign",e[e["Break"]=3]="Break",e[e["BreakWhenTrue"]=4]="BreakWhenTrue",e[e["BreakWhenFalse"]=5]="BreakWhenFalse",e[e["Yield"]=6]="Yield",e[e["YieldStar"]=7]="YieldStar",e[e["Return"]=8]="Return",e[e["Throw"]=9]="Throw",e[e["Endfinally"]=10]="Endfinally"})(n||(n={})),function(e){e[e["Open"]=0]="Open",e[e["Close"]=1]="Close"}(t||(t={})),function(e){e[e["Exception"]=0]="Exception",e[e["With"]=1]="With",e[e["Switch"]=2]="Switch",e[e["Loop"]=3]="Loop",e[e["Labeled"]=4]="Labeled"}(r||(r={})),function(e){e[e["Try"]=0]="Try",e[e["Catch"]=1]="Catch",e[e["Finally"]=2]="Finally",e[e["Done"]=3]="Done"}(a||(a={})),function(e){e[e["Next"]=0]="Next",e[e["Throw"]=1]="Throw",e[e["Return"]=2]="Return",e[e["Break"]=3]="Break",e[e["Yield"]=4]="Yield",e[e["YieldStar"]=5]="YieldStar",e[e["Catch"]=6]="Catch",e[e["Endfinally"]=7]="Endfinally"}(i||(i={})),e.transformGenerators=s;var l={name:"typescript:generator",scoped:!1,priority:6,text:'\n            var __generator = (this && this.__generator) || function (thisArg, body) {\n                var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;\n                return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;\n                function verb(n) { return function (v) { return step([n, v]); }; }\n                function step(op) {\n                    if (f) throw new TypeError("Generator is already executing.");\n                    while (_) try {\n                        if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;\n                        if (y = 0, t) op = [op[0] & 2, t.value];\n                        switch (op[0]) {\n                            case 0: case 1: t = op; break;\n                            case 4: _.label++; return { value: op[1], done: false };\n                            case 5: _.label++; y = op[1]; op = [0]; continue;\n                            case 7: op = _.ops.pop(); _.trys.pop(); continue;\n                            default:\n                                if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }\n                                if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }\n                                if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }\n                                if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }\n                                if (t[2]) _.ops.pop();\n                                _.trys.pop(); continue;\n                        }\n                        op = body.call(thisArg, _);\n                    } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }\n                    if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };\n                }\n            };'}}(u||(u={})),function(e){function n(n){function s(n){switch(n){case e.ModuleKind.AMD:return D;case e.ModuleKind.UMD:return L;default:return A}}var c=n.startLexicalEnvironment,l=n.endLexicalEnvironment,u=n.hoistVariableDeclaration,d=n.getCompilerOptions(),p=n.getEmitResolver(),m=n.getEmitHost(),f=e.getEmitScriptTarget(d),_=e.getEmitModuleKind(d),g=n.onSubstituteNode,v=n.onEmitNode;n.onSubstituteNode=me,n.onEmitNode=pe,n.enableSubstitution(72),n.enableSubstitution(204),n.enableSubstitution(202),n.enableSubstitution(203),n.enableSubstitution(276),n.enableEmitNotification(279);var y,h,b,E,T=[],S=[];return e.chainBundle(x);function x(n){if(n.isDeclarationFile||!(e.isEffectiveExternalModule(n,d)||16777216&n.transformFlags||e.isJsonSourceFile(n)&&e.hasJsonModuleEmitEnabled(d)&&(d.out||d.outFile)))return n;y=n,h=e.collectExternalModuleInfo(n,p,d),T[e.getOriginalNodeId(n)]=h;var t=s(_),r=t(n);return y=void 0,h=void 0,E=!1,e.aggregateTransformFlags(r)}function C(){return!(h.exportEquals||!e.isExternalModule(y))}function A(r){c();var a=[],i=e.getStrictOptionValue(d,"alwaysStrict")||!d.noImplicitUseStrict&&e.isExternalModule(y),o=e.addPrologue(a,r.statements,i,M);C()&&e.append(a,ce()),e.append(a,e.visitNode(h.externalHelpersImportDeclaration,M,e.isStatement)),e.addRange(a,e.visitNodes(r.statements,M,e.isStatement,o)),w(a,!1),e.addStatementsAfterPrologue(a,l());var s=e.updateSourceFileNode(r,e.setTextRange(e.createNodeArray(a),r.statements));return h.hasExportStarsToExportValues&&!d.importHelpers&&e.addEmitHelper(s,t),e.addEmitHelpers(s,n.readEmitHelpers()),s}function D(t){var r=e.createIdentifier("define"),a=e.tryGetModuleNameFromFile(t,m,d),i=e.isJsonSourceFile(t)&&t,o=k(t,!0),s=o.aliasedModuleNames,c=o.unaliasedModuleNames,l=o.importAliasNames,u=e.updateSourceFileNode(t,e.setTextRange(e.createNodeArray([e.createExpressionStatement(e.createCall(r,void 0,(a?[a]:[]).concat([e.createArrayLiteral(i?e.emptyArray:[e.createLiteral("require"),e.createLiteral("exports")].concat(s,c)),i?i.statements.length?i.statements[0].expression:e.createObjectLiteral():e.createFunctionExpression(void 0,void 0,void 0,void 0,[e.createParameter(void 0,void 0,void 0,"require"),e.createParameter(void 0,void 0,void 0,"exports")].concat(l),void 0,I(t))])))]),t.statements));return e.addEmitHelpers(u,n.readEmitHelpers()),u}function L(t){var r=k(t,!1),a=r.aliasedModuleNames,i=r.unaliasedModuleNames,o=r.importAliasNames,s=e.tryGetModuleNameFromFile(t,m,d),c=e.createFunctionExpression(void 0,void 0,void 0,void 0,[e.createParameter(void 0,void 0,void 0,"factory")],void 0,e.setTextRange(e.createBlock([e.createIf(e.createLogicalAnd(e.createTypeCheck(e.createIdentifier("module"),"object"),e.createTypeCheck(e.createPropertyAccess(e.createIdentifier("module"),"exports"),"object")),e.createBlock([e.createVariableStatement(void 0,[e.createVariableDeclaration("v",void 0,e.createCall(e.createIdentifier("factory"),void 0,[e.createIdentifier("require"),e.createIdentifier("exports")]))]),e.setEmitFlags(e.createIf(e.createStrictInequality(e.createIdentifier("v"),e.createIdentifier("undefined")),e.createExpressionStatement(e.createAssignment(e.createPropertyAccess(e.createIdentifier("module"),"exports"),e.createIdentifier("v")))),1)]),e.createIf(e.createLogicalAnd(e.createTypeCheck(e.createIdentifier("define"),"function"),e.createPropertyAccess(e.createIdentifier("define"),"amd")),e.createBlock([e.createExpressionStatement(e.createCall(e.createIdentifier("define"),void 0,(s?[s]:[]).concat([e.createArrayLiteral([e.createLiteral("require"),e.createLiteral("exports")].concat(a,i)),e.createIdentifier("factory")])))])))],!0),void 0)),l=e.updateSourceFileNode(t,e.setTextRange(e.createNodeArray([e.createExpressionStatement(e.createCall(c,void 0,[e.createFunctionExpression(void 0,void 0,void 0,void 0,[e.createParameter(void 0,void 0,void 0,"require"),e.createParameter(void 0,void 0,void 0,"exports")].concat(o),void 0,I(t))]))]),t.statements));return e.addEmitHelpers(l,n.readEmitHelpers()),l}function k(n,t){for(var r=[],a=[],i=[],o=0,s=n.amdDependencies;o<s.length;o++){var c=s[o];c.name?(r.push(e.createLiteral(c.path)),i.push(e.createParameter(void 0,void 0,void 0,c.name))):a.push(e.createLiteral(c.path))}for(var l=0,u=h.externalImports;l<u.length;l++){var f=u[l],_=e.getExternalModuleNameLiteral(f,y,m,p,d),g=e.getLocalNameForExternalImport(f,y);_&&(t&&g?(e.setEmitFlags(g,4),r.push(_),i.push(e.createParameter(void 0,void 0,void 0,g))):a.push(_))}return{aliasedModuleNames:r,unaliasedModuleNames:a,importAliasNames:i}}function N(n){if(!e.isImportEqualsDeclaration(n)&&!e.isExportDeclaration(n)&&e.getExternalModuleNameLiteral(n,y,m,p,d)){var t=e.getLocalNameForExternalImport(n,y),r=K(n,t);if(r!==t)return e.createExpressionStatement(e.createAssignment(t,r))}}function I(n){c();var r=[],i=e.addPrologue(r,n.statements,!d.noImplicitUseStrict,M);C()&&e.append(r,ce()),e.append(r,e.visitNode(h.externalHelpersImportDeclaration,M,e.isStatement)),_===e.ModuleKind.AMD&&e.addRange(r,e.mapDefined(h.externalImports,N)),e.addRange(r,e.visitNodes(n.statements,M,e.isStatement,i)),w(r,!0),e.addStatementsAfterPrologue(r,l());var o=e.createBlock(r,!0);return h.hasExportStarsToExportValues&&!d.importHelpers&&e.addEmitHelper(o,t),E&&e.addEmitHelper(o,a),o}function w(n,t){if(h.exportEquals){var r=e.visitNode(h.exportEquals.expression,O);if(r)if(t){var a=e.createReturn(r);e.setTextRange(a,h.exportEquals),e.setEmitFlags(a,1920),n.push(a)}else{a=e.createExpressionStatement(e.createAssignment(e.createPropertyAccess(e.createIdentifier("module"),"exports"),r));e.setTextRange(a,h.exportEquals),e.setEmitFlags(a,1536),n.push(a)}}}function M(t){switch(t.kind){case 249:return j(t);case 248:return H(t);case 255:return W(t);case 254:return q(t);case 219:return X(t);case 239:return z(t);case 240:return J(t);case 310:return Z(t);case 311:return ee(t);default:return e.visitEachChild(t,O,n)}}function O(t){return 16777216&t.transformFlags||2048&t.transformFlags?e.isImportCall(t)?F(t):1024&t.transformFlags&&e.isBinaryExpression(t)?R(t):e.visitEachChild(t,O,n):t}function P(n){if(e.isObjectLiteralExpression(n))for(var t=0,r=n.properties;t<r.length;t++){var a=r[t];switch(a.kind){case 275:if(P(a.initializer))return!0;break;case 276:if(P(a.name))return!0;break;case 277:if(P(a.expression))return!0;break;case 156:case 158:case 159:return!1;default:e.Debug.assertNever(a,"Unhandled object member kind")}}else if(e.isArrayLiteralExpression(n))for(var i=0,o=n.elements;i<o.length;i++){a=o[i];if(e.isSpreadElement(a)){if(P(a.expression))return!0}else if(P(a))return!0}else if(e.isIdentifier(n))return e.length(he(n))>(e.isExportName(n)?1:0);return!1}function R(t){return P(t.left)?e.flattenDestructuringAssignment(t,O,n,0,!1,Y):e.visitEachChild(t,O,n)}function F(n){var t=e.visitNode(e.firstOrUndefined(n.arguments),O),r=!!(8192&n.transformFlags);switch(d.module){case e.ModuleKind.AMD:return V(t,r);case e.ModuleKind.UMD:return G(t,r);case e.ModuleKind.CommonJS:default:return B(t,r)}}function G(n,t){if(E=!0,e.isSimpleCopiableExpression(n)){var r=e.isGeneratedIdentifier(n)?n:e.isStringLiteral(n)?e.createLiteral(n):e.setEmitFlags(e.setTextRange(e.getSynthesizedClone(n),n),1536);return e.createConditional(e.createIdentifier("__syncRequire"),B(n,t),V(r,t))}var a=e.createTempVariable(u);return e.createComma(e.createAssignment(a,n),e.createConditional(e.createIdentifier("__syncRequire"),B(a,t),V(a,t)))}function V(t,r){var a,o=e.createUniqueName("resolve"),s=e.createUniqueName("reject"),c=[e.createParameter(void 0,void 0,void 0,o),e.createParameter(void 0,void 0,void 0,s)],l=e.createBlock([e.createExpressionStatement(e.createCall(e.createIdentifier("require"),void 0,[e.createArrayLiteral([t||e.createOmittedExpression()]),o,s]))]);f>=2?a=e.createArrowFunction(void 0,void 0,c,void 0,void 0,l):(a=e.createFunctionExpression(void 0,void 0,void 0,void 0,c,void 0,l),r&&e.setEmitFlags(a,8));var u=e.createNew(e.createIdentifier("Promise"),void 0,[a]);return d.esModuleInterop?(n.requestEmitHelper(i),e.createCall(e.createPropertyAccess(u,e.createIdentifier("then")),void 0,[e.getHelperName("__importStar")])):u}function B(t,r){var a,o=e.createCall(e.createPropertyAccess(e.createIdentifier("Promise"),"resolve"),void 0,[]),s=e.createCall(e.createIdentifier("require"),void 0,t?[t]:[]);return d.esModuleInterop&&(n.requestEmitHelper(i),s=e.createCall(e.getHelperName("__importStar"),void 0,[s])),f>=2?a=e.createArrowFunction(void 0,void 0,[],void 0,void 0,s):(a=e.createFunctionExpression(void 0,void 0,void 0,void 0,[],void 0,e.createBlock([e.createReturn(s)])),r&&e.setEmitFlags(a,8)),e.createCall(e.createPropertyAccess(o,"then"),void 0,[a])}function K(t,r){return!d.esModuleInterop||67108864&e.getEmitFlags(t)?r:e.getImportNeedsImportStarHelper(t)?(n.requestEmitHelper(i),e.createCall(e.getHelperName("__importStar"),void 0,[r])):e.getImportNeedsImportDefaultHelper(t)?(n.requestEmitHelper(o),e.createCall(e.getHelperName("__importDefault"),void 0,[r])):r}function j(n){var t,r=e.getNamespaceDeclarationNode(n);if(_!==e.ModuleKind.AMD){if(!n.importClause)return e.setOriginalNode(e.setTextRange(e.createExpressionStatement(U(n)),n),n);var a=[];r&&!e.isDefaultImport(n)?a.push(e.createVariableDeclaration(e.getSynthesizedClone(r.name),void 0,K(n,U(n)))):(a.push(e.createVariableDeclaration(e.getGeneratedNameForNode(n),void 0,K(n,U(n)))),r&&e.isDefaultImport(n)&&a.push(e.createVariableDeclaration(e.getSynthesizedClone(r.name),void 0,e.getGeneratedNameForNode(n)))),t=e.append(t,e.setOriginalNode(e.setTextRange(e.createVariableStatement(void 0,e.createVariableDeclarationList(a,f>=2?2:0)),n),n))}else r&&e.isDefaultImport(n)&&(t=e.append(t,e.createVariableStatement(void 0,e.createVariableDeclarationList([e.setOriginalNode(e.setTextRange(e.createVariableDeclaration(e.getSynthesizedClone(r.name),void 0,e.getGeneratedNameForNode(n)),n),n)],f>=2?2:0))));if($(n)){var i=e.getOriginalNodeId(n);S[i]=ne(S[i],n)}else t=ne(t,n);return e.singleOrMany(t)}function U(n){var t=e.getExternalModuleNameLiteral(n,y,m,p,d),r=[];return t&&r.push(t),e.createCall(e.createIdentifier("require"),void 0,r)}function H(n){var t;if(e.Debug.assert(e.isExternalModuleImportEqualsDeclaration(n),"import= for internal module references should be handled in an earlier transformer."),_!==e.ModuleKind.AMD?t=e.hasModifier(n,1)?e.append(t,e.setOriginalNode(e.setTextRange(e.createExpressionStatement(ue(n.name,U(n))),n),n)):e.append(t,e.setOriginalNode(e.setTextRange(e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(e.getSynthesizedClone(n.name),void 0,U(n))],f>=2?2:0)),n),n)):e.hasModifier(n,1)&&(t=e.append(t,e.setOriginalNode(e.setTextRange(e.createExpressionStatement(ue(e.getExportName(n),e.getLocalName(n))),n),n))),$(n)){var r=e.getOriginalNodeId(n);S[r]=te(S[r],n)}else t=te(t,n);return e.singleOrMany(t)}function W(t){if(t.moduleSpecifier){var a=e.getGeneratedNameForNode(t);if(t.exportClause){var i=[];_!==e.ModuleKind.AMD&&i.push(e.setOriginalNode(e.setTextRange(e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(a,void 0,U(t))])),t),t));for(var o=0,s=t.exportClause.elements;o<s.length;o++){var c=s[o],l=e.createPropertyAccess(a,c.propertyName||c.name);i.push(e.setOriginalNode(e.setTextRange(e.createExpressionStatement(ue(e.getExportName(c),l)),c),c))}return e.singleOrMany(i)}return e.setOriginalNode(e.setTextRange(e.createExpressionStatement(r(n,_!==e.ModuleKind.AMD?U(t):a)),t),t)}}function q(n){if(!n.isExportEquals){var t,r=n.original;if(r&&$(r)){var a=e.getOriginalNodeId(n);S[a]=se(S[a],e.createIdentifier("default"),e.visitNode(n.expression,O),n,!0)}else t=se(t,e.createIdentifier("default"),e.visitNode(n.expression,O),n,!0);return e.singleOrMany(t)}}function z(t){var r;if(r=e.hasModifier(t,1)?e.append(r,e.setOriginalNode(e.setTextRange(e.createFunctionDeclaration(void 0,e.visitNodes(t.modifiers,de,e.isModifier),t.asteriskToken,e.getDeclarationName(t,!0,!0),void 0,e.visitNodes(t.parameters,O),void 0,e.visitEachChild(t.body,O,n)),t),t)):e.append(r,e.visitEachChild(t,O,n)),$(t)){var a=e.getOriginalNodeId(t);S[a]=ie(S[a],t)}else r=ie(r,t);return e.singleOrMany(r)}function J(t){var r;if(r=e.hasModifier(t,1)?e.append(r,e.setOriginalNode(e.setTextRange(e.createClassDeclaration(void 0,e.visitNodes(t.modifiers,de,e.isModifier),e.getDeclarationName(t,!0,!0),void 0,e.visitNodes(t.heritageClauses,O),e.visitNodes(t.members,O)),t),t)):e.append(r,e.visitEachChild(t,O,n)),$(t)){var a=e.getOriginalNodeId(t);S[a]=ie(S[a],t)}else r=ie(r,t);return e.singleOrMany(r)}function X(t){var r,a,i;if(e.hasModifier(t,1)){for(var o=void 0,s=0,c=t.declarationList.declarations;s<c.length;s++){var l=c[s];e.isIdentifier(l.name)&&e.isLocalName(l.name)?(o||(o=e.visitNodes(t.modifiers,de,e.isModifier)),a=e.append(a,l)):l.initializer&&(i=e.append(i,Q(l)))}a&&(r=e.append(r,e.updateVariableStatement(t,o,e.updateVariableDeclarationList(t.declarationList,a)))),i&&(r=e.append(r,e.setOriginalNode(e.setTextRange(e.createExpressionStatement(e.inlineExpressions(i)),t),t)))}else r=e.append(r,e.visitEachChild(t,O,n));if($(t)){var u=e.getOriginalNodeId(t);S[u]=re(S[u],t)}else r=re(r,t);return e.singleOrMany(r)}function Y(n,t,r){var a=he(n);if(a){for(var i=e.isExportName(n)?t:e.createAssignment(n,t),o=0,s=a;o<s.length;o++){var c=s[o];e.setEmitFlags(i,4),i=ue(c,i,r)}return i}return e.createAssignment(n,t)}function Q(t){return e.isBindingPattern(t.name)?e.flattenDestructuringAssignment(e.visitNode(t,O),void 0,n,0,!1,Y):e.createAssignment(e.setTextRange(e.createPropertyAccess(e.createIdentifier("exports"),t.name),t.name),e.visitNode(t.initializer,O))}function Z(n){if($(n)&&219===n.original.kind){var t=e.getOriginalNodeId(n);S[t]=re(S[t],n.original)}return n}function $(n){return 0!==(4194304&e.getEmitFlags(n))}function ee(n){var t=e.getOriginalNodeId(n),r=S[t];return r?(delete S[t],e.append(r,n)):n}function ne(e,n){if(h.exportEquals)return e;var t=n.importClause;if(!t)return e;t.name&&(e=oe(e,t));var r=t.namedBindings;if(r)switch(r.kind){case 251:e=oe(e,r);break;case 252:for(var a=0,i=r.elements;a<i.length;a++){var o=i[a];e=oe(e,o)}break}return e}function te(e,n){return h.exportEquals?e:oe(e,n)}function re(e,n){if(h.exportEquals)return e;for(var t=0,r=n.declarationList.declarations;t<r.length;t++){var a=r[t];e=ae(e,a)}return e}function ae(n,t){if(h.exportEquals)return n;if(e.isBindingPattern(t.name))for(var r=0,a=t.name.elements;r<a.length;r++){var i=a[r];e.isOmittedExpression(i)||(n=ae(n,i))}else e.isGeneratedIdentifier(t.name)||(n=oe(n,t));return n}function ie(n,t){if(h.exportEquals)return n;if(e.hasModifier(t,1)){var r=e.hasModifier(t,512)?e.createIdentifier("default"):e.getDeclarationName(t);n=se(n,r,e.getLocalName(t),t)}return t.name&&(n=oe(n,t)),n}function oe(n,t){var r=e.getDeclarationName(t),a=h.exportSpecifiers.get(e.idText(r));if(a)for(var i=0,o=a;i<o.length;i++){var s=o[i];n=se(n,s.name,r,s.name)}return n}function se(n,t,r,a,i){return n=e.append(n,le(t,r,a,i)),n}function ce(){var n;return n=0===f?e.createExpressionStatement(ue(e.createIdentifier("__esModule"),e.createLiteral(!0))):e.createExpressionStatement(e.createCall(e.createPropertyAccess(e.createIdentifier("Object"),"defineProperty"),void 0,[e.createIdentifier("exports"),e.createLiteral("__esModule"),e.createObjectLiteral([e.createPropertyAssignment("value",e.createLiteral(!0))])])),e.setEmitFlags(n,1048576),n}function le(n,t,r,a){var i=e.setTextRange(e.createExpressionStatement(ue(n,t)),r);return e.startOnNewLine(i),a||e.setEmitFlags(i,1536),i}function ue(n,t,r){return e.setTextRange(e.createAssignment(e.createPropertyAccess(e.createIdentifier("exports"),e.getSynthesizedClone(n)),t),r)}function de(e){switch(e.kind){case 85:case 80:return}return e}function pe(n,t,r){279===t.kind?(y=t,h=T[e.getOriginalNodeId(y)],b=[],v(n,t,r),y=void 0,h=void 0,b=void 0):v(n,t,r)}function me(n,t){return t=g(n,t),t.id&&b[t.id]?t:1===n?_e(t):e.isShorthandPropertyAssignment(t)?fe(t):t}function fe(n){var t=n.name,r=ge(t);if(r!==t){if(n.objectAssignmentInitializer){var a=e.createAssignment(r,n.objectAssignmentInitializer);return e.setTextRange(e.createPropertyAssignment(t,a),n)}return e.setTextRange(e.createPropertyAssignment(t,r),n)}return n}function _e(e){switch(e.kind){case 72:return ge(e);case 204:return ve(e);case 203:case 202:return ye(e)}return e}function ge(n){if(4096&e.getEmitFlags(n)){var t=e.getExternalHelpersModuleName(y);return t?e.createPropertyAccess(t,n):n}if(!e.isGeneratedIdentifier(n)&&!e.isLocalName(n)){var r=p.getReferencedExportContainer(n,e.isExportName(n));if(r&&279===r.kind)return e.setTextRange(e.createPropertyAccess(e.createIdentifier("exports"),e.getSynthesizedClone(n)),n);var a=p.getReferencedImportDeclaration(n);if(a){if(e.isImportClause(a))return e.setTextRange(e.createPropertyAccess(e.getGeneratedNameForNode(a.parent),e.createIdentifier("default")),n);if(e.isImportSpecifier(a)){var i=a.propertyName||a.name;return e.setTextRange(e.createPropertyAccess(e.getGeneratedNameForNode(a.parent.parent.parent),e.getSynthesizedClone(i)),n)}}}return n}function ve(n){if(e.isAssignmentOperator(n.operatorToken.kind)&&e.isIdentifier(n.left)&&!e.isGeneratedIdentifier(n.left)&&!e.isLocalName(n.left)&&!e.isDeclarationNameOfEnumOrNamespace(n.left)){var t=he(n.left);if(t){for(var r=n,a=0,i=t;a<i.length;a++){var o=i[a];b[e.getNodeId(r)]=!0,r=ue(o,r,n)}return r}}return n}function ye(n){if((44===n.operator||45===n.operator)&&e.isIdentifier(n.operand)&&!e.isGeneratedIdentifier(n.operand)&&!e.isLocalName(n.operand)&&!e.isDeclarationNameOfEnumOrNamespace(n.operand)){var t=he(n.operand);if(t){for(var r=203===n.kind?e.setTextRange(e.createBinary(n.operand,e.createToken(44===n.operator?60:61),e.createLiteral(1)),n):n,a=0,i=t;a<i.length;a++){var o=i[a];b[e.getNodeId(r)]=!0,r=ue(o,r)}return r}}return n}function he(n){if(!e.isGeneratedIdentifier(n)){var t=p.getReferencedImportDeclaration(n)||p.getReferencedValueDeclaration(n);if(t)return h&&h.exportedBindings[e.getOriginalNodeId(t)]}}}e.transformModule=n;var t={name:"typescript:export-star",scoped:!0,text:"\n            function __export(m) {\n                for (var p in m) if (!exports.hasOwnProperty(p)) exports[p] = m[p];\n            }"};function r(n,t){var r=n.getCompilerOptions();return r.importHelpers?e.createCall(e.getHelperName("__exportStar"),void 0,[t,e.createIdentifier("exports")]):e.createCall(e.createIdentifier("__export"),void 0,[t])}var a={name:"typescript:dynamicimport-sync-require",scoped:!0,text:'\n            var __syncRequire = typeof module === "object" && typeof module.exports === "object";'},i={name:"typescript:commonjsimportstar",scoped:!1,text:'\nvar __importStar = (this && this.__importStar) || function (mod) {\n    if (mod && mod.__esModule) return mod;\n    var result = {};\n    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];\n    result["default"] = mod;\n    return result;\n};'},o={name:"typescript:commonjsimportdefault",scoped:!1,text:'\nvar __importDefault = (this && this.__importDefault) || function (mod) {\n    return (mod && mod.__esModule) ? mod : { "default": mod };\n};'}}(u||(u={})),function(e){function n(n){var t=n.startLexicalEnvironment,r=n.endLexicalEnvironment,a=n.hoistVariableDeclaration,i=n.getCompilerOptions(),o=n.getEmitResolver(),s=n.getEmitHost(),c=n.onSubstituteNode,l=n.onEmitNode;n.onSubstituteNode=Ee,n.onEmitNode=be,n.enableSubstitution(72),n.enableSubstitution(276),n.enableSubstitution(204),n.enableSubstitution(202),n.enableSubstitution(203),n.enableEmitNotification(279);var u,d,p,m,f,_,g,v=[],y=[],h=[],b=[];return e.chainBundle(E);function E(n){if(n.isDeclarationFile||!(e.isEffectiveExternalModule(n,i)||16777216&n.transformFlags))return n;var t=e.getOriginalNodeId(n);u=n,_=n,d=v[t]=e.collectExternalModuleInfo(n,o,i),p=e.createUniqueName("exports"),h[t]=p,m=e.createUniqueName("context");var r=T(d.externalImports),a=S(n,r),c=e.createFunctionExpression(void 0,void 0,void 0,void 0,[e.createParameter(void 0,void 0,void 0,p),e.createParameter(void 0,void 0,void 0,m)],void 0,a),l=e.tryGetModuleNameFromFile(n,s,i),y=e.createArrayLiteral(e.map(r,function(e){return e.name})),E=e.setEmitFlags(e.updateSourceFileNode(n,e.setTextRange(e.createNodeArray([e.createExpressionStatement(e.createCall(e.createPropertyAccess(e.createIdentifier("System"),"register"),void 0,l?[l,y,c]:[y,c]))]),n.statements)),1024);return i.outFile||i.out||e.moveEmitHelpers(E,a,function(e){return!e.scoped}),g&&(b[t]=g,g=void 0),u=void 0,d=void 0,p=void 0,m=void 0,f=void 0,_=void 0,e.aggregateTransformFlags(E)}function T(n){for(var t=e.createMap(),r=[],a=0,c=n;a<c.length;a++){var l=c[a],d=e.getExternalModuleNameLiteral(l,u,s,o,i);if(d){var p=d.text,m=t.get(p);void 0!==m?r[m].externalImports.push(l):(t.set(p,r.length),r.push({name:d,externalImports:[l]}))}}return r}function S(n,a){var o=[];t();var s=e.getStrictOptionValue(i,"alwaysStrict")||!i.noImplicitUseStrict&&e.isExternalModule(u),c=e.addPrologue(o,n.statements,s,D);o.push(e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration("__moduleName",void 0,e.createLogicalAnd(m,e.createPropertyAccess(m,"id")))]))),e.visitNode(d.externalHelpersImportDeclaration,D,e.isStatement);var l=e.visitNodes(n.statements,D,e.isStatement,c);e.addRange(o,f),e.addStatementsAfterPrologue(o,r());var p=x(o),_=e.createObjectLiteral([e.createPropertyAssignment("setters",A(p,a)),e.createPropertyAssignment("execute",e.createFunctionExpression(void 0,void 0,void 0,void 0,[],void 0,e.createBlock(l,!0)))]);return _.multiLine=!0,o.push(e.createReturn(_)),e.createBlock(o,!0)}function x(n){if(d.hasExportStarsToExportValues){if(!d.exportedNames&&0===d.exportSpecifiers.size){for(var t=!1,r=0,a=d.externalImports;r<a.length;r++){var i=a[r];if(255===i.kind&&i.exportClause){t=!0;break}}if(!t){var o=C(void 0);return n.push(o),o.name}}var s=[];if(d.exportedNames)for(var c=0,l=d.exportedNames;c<l.length;c++){var u=l[c];"default"!==u.escapedText&&s.push(e.createPropertyAssignment(e.createLiteral(u),e.createTrue()))}for(var p=0,m=d.externalImports;p<m.length;p++){i=m[p];if(255===i.kind&&i.exportClause)for(var f=0,_=i.exportClause.elements;f<_.length;f++){var g=_[f];s.push(e.createPropertyAssignment(e.createLiteral(e.idText(g.name||g.propertyName)),e.createTrue()))}}var v=e.createUniqueName("exportedNames");n.push(e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(v,void 0,e.createObjectLiteral(s,!0))])));var y=C(v);return n.push(y),y.name}}function C(n){var t=e.createUniqueName("exportStar"),r=e.createIdentifier("m"),a=e.createIdentifier("n"),i=e.createIdentifier("exports"),o=e.createStrictInequality(a,e.createLiteral("default"));return n&&(o=e.createLogicalAnd(o,e.createLogicalNot(e.createCall(e.createPropertyAccess(n,"hasOwnProperty"),void 0,[a])))),e.createFunctionDeclaration(void 0,void 0,void 0,t,void 0,[e.createParameter(void 0,void 0,void 0,r)],void 0,e.createBlock([e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(i,void 0,e.createObjectLiteral([]))])),e.createForIn(e.createVariableDeclarationList([e.createVariableDeclaration(a,void 0)]),r,e.createBlock([e.setEmitFlags(e.createIf(o,e.createExpressionStatement(e.createAssignment(e.createElementAccess(i,a),e.createElementAccess(r,a)))),1)])),e.createExpressionStatement(e.createCall(p,void 0,[i]))],!0))}function A(n,t){for(var r=[],a=0,i=t;a<i.length;a++){for(var o=i[a],s=e.forEach(o.externalImports,function(n){return e.getLocalNameForExternalImport(n,u)}),c=s?e.getGeneratedNameForNode(s):e.createUniqueName(""),l=[],d=0,m=o.externalImports;d<m.length;d++){var f=m[d],_=e.getLocalNameForExternalImport(f,u);switch(f.kind){case 249:if(!f.importClause)break;case 248:e.Debug.assert(void 0!==_),l.push(e.createExpressionStatement(e.createAssignment(_,c)));break;case 255:if(e.Debug.assert(void 0!==_),f.exportClause){for(var g=[],v=0,y=f.exportClause.elements;v<y.length;v++){var h=y[v];g.push(e.createPropertyAssignment(e.createLiteral(e.idText(h.name)),e.createElementAccess(c,e.createLiteral(e.idText(h.propertyName||h.name)))))}l.push(e.createExpressionStatement(e.createCall(p,void 0,[e.createObjectLiteral(g,!0)])))}else l.push(e.createExpressionStatement(e.createCall(n,void 0,[c])));break}}r.push(e.createFunctionExpression(void 0,void 0,void 0,void 0,[e.createParameter(void 0,void 0,void 0,c)],void 0,e.createBlock(l,!0)))}return e.createArrayLiteral(r,!0)}function D(e){switch(e.kind){case 249:return L(e);case 248:return k(e);case 255:return;case 254:return N(e);default:return Z(e)}}function L(n){var t;if(n.importClause&&a(e.getLocalNameForExternalImport(n,u)),K(n)){var r=e.getOriginalNodeId(n);y[r]=U(y[r],n)}else t=U(t,n);return e.singleOrMany(t)}function k(n){var t;if(e.Debug.assert(e.isExternalModuleImportEqualsDeclaration(n),"import= for internal module references should be handled in an earlier transformer."),a(e.getLocalNameForExternalImport(n,u)),K(n)){var r=e.getOriginalNodeId(n);y[r]=H(y[r],n)}else t=H(t,n);return e.singleOrMany(t)}function N(n){if(!n.isExportEquals){var t=e.visitNode(n.expression,_e,e.isExpression),r=n.original;if(!r||!K(r))return Y(e.createIdentifier("default"),t,!0);var a=e.getOriginalNodeId(n);y[a]=X(y[a],e.createIdentifier("default"),t,!0)}}function I(t){if(f=e.hasModifier(t,1)?e.append(f,e.updateFunctionDeclaration(t,t.decorators,e.visitNodes(t.modifiers,he,e.isModifier),t.asteriskToken,e.getDeclarationName(t,!0,!0),void 0,e.visitNodes(t.parameters,_e,e.isParameterDeclaration),void 0,e.visitNode(t.body,_e,e.isBlock))):e.append(f,e.visitEachChild(t,_e,n)),K(t)){var r=e.getOriginalNodeId(t);y[r]=z(y[r],t)}else f=z(f,t)}function w(n){var t,r=e.getLocalName(n);if(a(r),t=e.append(t,e.setTextRange(e.createExpressionStatement(e.createAssignment(r,e.setTextRange(e.createClassExpression(void 0,n.name,void 0,e.visitNodes(n.heritageClauses,_e,e.isHeritageClause),e.visitNodes(n.members,_e,e.isClassElement)),n))),n)),K(n)){var i=e.getOriginalNodeId(n);y[i]=z(y[i],n)}else t=z(t,n);return e.singleOrMany(t)}function M(n){if(!P(n.declarationList))return e.visitNode(n,_e,e.isStatement);for(var t,r,a=e.hasModifier(n,1),i=K(n),o=0,s=n.declarationList.declarations;o<s.length;o++){var c=s[o];c.initializer?t=e.append(t,R(c,a&&!i)):O(c)}if(t&&(r=e.append(r,e.setTextRange(e.createExpressionStatement(e.inlineExpressions(t)),n))),i){var l=e.getOriginalNodeId(n);y[l]=W(y[l],n,a)}else r=W(r,n,!1);return e.singleOrMany(r)}function O(n){if(e.isBindingPattern(n.name))for(var t=0,r=n.name.elements;t<r.length;t++){var i=r[t];e.isOmittedExpression(i)||O(i)}else a(e.getSynthesizedClone(n.name))}function P(n){return 0===(2097152&e.getEmitFlags(n))&&(279===_.kind||0===(3&e.getOriginalNode(n).flags))}function R(t,r){var a=r?F:G;return e.isBindingPattern(t.name)?e.flattenDestructuringAssignment(t,_e,n,0,!1,a):t.initializer?a(t.name,e.visitNode(t.initializer,_e,e.isExpression)):t.name}function F(e,n,t){return V(e,n,t,!0)}function G(e,n,t){return V(e,n,t,!1)}function V(n,t,r,i){return a(e.getSynthesizedClone(n)),i?Q(n,ke(e.setTextRange(e.createAssignment(n,t),r))):ke(e.setTextRange(e.createAssignment(n,t),r))}function B(n){if(K(n)&&219===n.original.kind){var t=e.getOriginalNodeId(n),r=e.hasModifier(n.original,1);y[t]=W(y[t],n.original,r)}return n}function K(n){return 0!==(4194304&e.getEmitFlags(n))}function j(n){var t=e.getOriginalNodeId(n),r=y[t];if(r)return delete y[t],e.append(r,n);var a=e.getOriginalNode(n);return e.isModuleOrEnumDeclaration(a)?e.append(J(r,a),n):n}function U(e,n){if(d.exportEquals)return e;var t=n.importClause;if(!t)return e;t.name&&(e=J(e,t));var r=t.namedBindings;if(r)switch(r.kind){case 251:e=J(e,r);break;case 252:for(var a=0,i=r.elements;a<i.length;a++){var o=i[a];e=J(e,o)}break}return e}function H(e,n){return d.exportEquals?e:J(e,n)}function W(e,n,t){if(d.exportEquals)return e;for(var r=0,a=n.declarationList.declarations;r<a.length;r++){var i=a[r];(i.initializer||t)&&(e=q(e,i,t))}return e}function q(n,t,r){if(d.exportEquals)return n;if(e.isBindingPattern(t.name))for(var a=0,i=t.name.elements;a<i.length;a++){var o=i[a];e.isOmittedExpression(o)||(n=q(n,o,r))}else if(!e.isGeneratedIdentifier(t.name)){var s=void 0;r&&(n=X(n,t.name,e.getLocalName(t)),s=e.idText(t.name)),n=J(n,t,s)}return n}function z(n,t){if(d.exportEquals)return n;var r;if(e.hasModifier(t,1)){var a=e.hasModifier(t,512)?e.createLiteral("default"):t.name;n=X(n,a,e.getLocalName(t)),r=e.getTextOfIdentifierOrLiteral(a)}return t.name&&(n=J(n,t,r)),n}function J(n,t,r){if(d.exportEquals)return n;var a=e.getDeclarationName(t),i=d.exportSpecifiers.get(e.idText(a));if(i)for(var o=0,s=i;o<s.length;o++){var c=s[o];c.name.escapedText!==r&&(n=X(n,c.name,a))}return n}function X(n,t,r,a){return n=e.append(n,Y(t,r,a)),n}function Y(n,t,r){var a=e.createExpressionStatement(Q(n,t));return e.startOnNewLine(a),r||e.setEmitFlags(a,1536),a}function Q(n,t){var r=e.isIdentifier(n)?e.createLiteral(n):n;return e.setEmitFlags(t,1536|e.getEmitFlags(t)),e.setCommentRange(e.createCall(p,void 0,[r,t]),t)}function Z(e){switch(e.kind){case 219:return M(e);case 239:return I(e);case 240:return w(e);case 225:return $(e);case 226:return ee(e);case 227:return ne(e);case 223:return ae(e);case 224:return ie(e);case 233:return oe(e);case 231:return se(e);case 232:return ce(e);case 246:return le(e);case 271:return ue(e);case 272:return de(e);case 235:return pe(e);case 274:return me(e);case 218:return fe(e);case 310:return B(e);case 311:return j(e);default:return _e(e)}}function $(n){var t=_;return _=n,n=e.updateFor(n,n.initializer&&re(n.initializer),e.visitNode(n.condition,_e,e.isExpression),e.visitNode(n.incrementor,_e,e.isExpression),e.visitNode(n.statement,Z,e.isStatement)),_=t,n}function ee(n){var t=_;return _=n,n=e.updateForIn(n,re(n.initializer),e.visitNode(n.expression,_e,e.isExpression),e.visitNode(n.statement,Z,e.isStatement,e.liftToBlock)),_=t,n}function ne(n){var t=_;return _=n,n=e.updateForOf(n,n.awaitModifier,re(n.initializer),e.visitNode(n.expression,_e,e.isExpression),e.visitNode(n.statement,Z,e.isStatement,e.liftToBlock)),_=t,n}function te(n){return e.isVariableDeclarationList(n)&&P(n)}function re(t){if(te(t)){for(var r=void 0,a=0,i=t.declarations;a<i.length;a++){var o=i[a];r=e.append(r,R(o,!1)),o.initializer||O(o)}return r?e.inlineExpressions(r):e.createOmittedExpression()}return e.visitEachChild(t,Z,n)}function ae(n){return e.updateDo(n,e.visitNode(n.statement,Z,e.isStatement,e.liftToBlock),e.visitNode(n.expression,_e,e.isExpression))}function ie(n){return e.updateWhile(n,e.visitNode(n.expression,_e,e.isExpression),e.visitNode(n.statement,Z,e.isStatement,e.liftToBlock))}function oe(n){return e.updateLabel(n,n.label,e.visitNode(n.statement,Z,e.isStatement,e.liftToBlock))}function se(n){return e.updateWith(n,e.visitNode(n.expression,_e,e.isExpression),e.visitNode(n.statement,Z,e.isStatement,e.liftToBlock))}function ce(n){return e.updateSwitch(n,e.visitNode(n.expression,_e,e.isExpression),e.visitNode(n.caseBlock,Z,e.isCaseBlock))}function le(n){var t=_;return _=n,n=e.updateCaseBlock(n,e.visitNodes(n.clauses,Z,e.isCaseOrDefaultClause)),_=t,n}function ue(n){return e.updateCaseClause(n,e.visitNode(n.expression,_e,e.isExpression),e.visitNodes(n.statements,Z,e.isStatement))}function de(t){return e.visitEachChild(t,Z,n)}function pe(t){return e.visitEachChild(t,Z,n)}function me(n){var t=_;return _=n,n=e.updateCatchClause(n,n.variableDeclaration,e.visitNode(n.block,Z,e.isBlock)),_=t,n}function fe(t){var r=_;return _=t,t=e.visitEachChild(t,Z,n),_=r,t}function _e(t){return 1024&t.transformFlags&&204===t.kind?ve(t):e.isImportCall(t)?ge(t):2048&t.transformFlags||16777216&t.transformFlags?e.visitEachChild(t,_e,n):t}function ge(n){return e.createCall(e.createPropertyAccess(m,e.createIdentifier("import")),void 0,e.some(n.arguments)?[e.visitNode(n.arguments[0],_e)]:[])}function ve(t){return ye(t.left)?e.flattenDestructuringAssignment(t,_e,n,0,!0):e.visitEachChild(t,_e,n)}function ye(n){if(e.isAssignmentExpression(n,!0))return ye(n.left);if(e.isSpreadElement(n))return ye(n.expression);if(e.isObjectLiteralExpression(n))return e.some(n.properties,ye);if(e.isArrayLiteralExpression(n))return e.some(n.elements,ye);if(e.isShorthandPropertyAssignment(n))return ye(n.name);if(e.isPropertyAssignment(n))return ye(n.initializer);if(e.isIdentifier(n)){var t=o.getReferencedExportContainer(n);return void 0!==t&&279===t.kind}return!1}function he(e){switch(e.kind){case 85:case 80:return}return e}function be(n,t,r){if(279===t.kind){var a=e.getOriginalNodeId(t);u=t,d=v[a],p=h[a],g=b[a],g&&delete b[a],l(n,t,r),u=void 0,d=void 0,p=void 0,g=void 0}else l(n,t,r)}function Ee(e,n){return n=c(e,n),Ne(n)?n:1===e?xe(n):4===e?Te(n):n}function Te(e){switch(e.kind){case 276:return Se(e)}return e}function Se(n){var t=n.name;if(!e.isGeneratedIdentifier(t)&&!e.isLocalName(t)){var r=o.getReferencedImportDeclaration(t);if(r){if(e.isImportClause(r))return e.setTextRange(e.createPropertyAssignment(e.getSynthesizedClone(t),e.createPropertyAccess(e.getGeneratedNameForNode(r.parent),e.createIdentifier("default"))),n);if(e.isImportSpecifier(r))return e.setTextRange(e.createPropertyAssignment(e.getSynthesizedClone(t),e.createPropertyAccess(e.getGeneratedNameForNode(r.parent.parent.parent),e.getSynthesizedClone(r.propertyName||r.name))),n)}}return n}function xe(e){switch(e.kind){case 72:return Ce(e);case 204:return Ae(e);case 202:case 203:return De(e)}return e}function Ce(n){if(4096&e.getEmitFlags(n)){var t=e.getExternalHelpersModuleName(u);return t?e.createPropertyAccess(t,n):n}if(!e.isGeneratedIdentifier(n)&&!e.isLocalName(n)){var r=o.getReferencedImportDeclaration(n);if(r){if(e.isImportClause(r))return e.setTextRange(e.createPropertyAccess(e.getGeneratedNameForNode(r.parent),e.createIdentifier("default")),n);if(e.isImportSpecifier(r))return e.setTextRange(e.createPropertyAccess(e.getGeneratedNameForNode(r.parent.parent.parent),e.getSynthesizedClone(r.propertyName||r.name)),n)}}return n}function Ae(n){if(e.isAssignmentOperator(n.operatorToken.kind)&&e.isIdentifier(n.left)&&!e.isGeneratedIdentifier(n.left)&&!e.isLocalName(n.left)&&!e.isDeclarationNameOfEnumOrNamespace(n.left)){var t=Le(n.left);if(t){for(var r=n,a=0,i=t;a<i.length;a++){var o=i[a];r=Q(o,ke(r))}return r}}return n}function De(n){if((44===n.operator||45===n.operator)&&e.isIdentifier(n.operand)&&!e.isGeneratedIdentifier(n.operand)&&!e.isLocalName(n.operand)&&!e.isDeclarationNameOfEnumOrNamespace(n.operand)){var t=Le(n.operand);if(t){for(var r=203===n.kind?e.setTextRange(e.createPrefix(n.operator,n.operand),n):n,a=0,i=t;a<i.length;a++){var o=i[a];r=Q(o,ke(r))}return 203===n.kind&&(r=44===n.operator?e.createSubtract(ke(r),e.createLiteral(1)):e.createAdd(ke(r),e.createLiteral(1))),r}}return n}function Le(n){var t;if(!e.isGeneratedIdentifier(n)){var r=o.getReferencedImportDeclaration(n)||o.getReferencedValueDeclaration(n);if(r){var a=o.getReferencedExportContainer(n,!1);a&&279===a.kind&&(t=e.append(t,e.getDeclarationName(r))),t=e.addRange(t,d&&d.exportedBindings[e.getOriginalNodeId(r)])}}return t}function ke(n){return void 0===g&&(g=[]),g[e.getNodeId(n)]=!0,n}function Ne(e){return g&&e.id&&g[e.id]}}e.transformSystemModule=n}(u||(u={})),function(e){function n(n){var t,r=n.getCompilerOptions(),a=n.onEmitNode,i=n.onSubstituteNode;return n.onEmitNode=l,n.onSubstituteNode=u,n.enableEmitNotification(279),n.enableSubstitution(72),e.chainBundle(o);function o(t){if(t.isDeclarationFile)return t;if(e.isExternalModule(t)||r.isolatedModules){var a=e.getOrCreateExternalHelpersModuleNameIfNeeded(t,r);if(a){var i=[],o=e.addPrologue(i,t.statements),c=e.createImportDeclaration(void 0,void 0,e.createImportClause(void 0,e.createNamespaceImport(a)),e.createLiteral(e.externalHelpersModuleNameText));return e.addEmitFlags(c,67108864),e.append(i,c),e.addRange(i,e.visitNodes(t.statements,s,e.isStatement,o)),e.updateSourceFileNode(t,e.setTextRange(e.createNodeArray(i),t.statements))}return e.visitEachChild(t,s,n)}return t}function s(e){switch(e.kind){case 248:return;case 254:return c(e)}return e}function c(e){return e.isExportEquals?void 0:e}function l(n,r,i){e.isSourceFile(r)?(t=r,a(n,r,i),t=void 0):a(n,r,i)}function u(n,t){return t=i(n,t),e.isIdentifier(t)&&1===n?d(t):t}function d(n){if(4096&e.getEmitFlags(n)){var r=e.getExternalHelpersModuleName(t);if(r)return e.createPropertyAccess(r,n)}return n}}e.transformES2015Module=n}(u||(u={})),function(e){function n(n){return e.isVariableDeclaration(n)||e.isPropertyDeclaration(n)||e.isPropertySignature(n)||e.isBindingElement(n)||e.isSetAccessor(n)||e.isGetAccessor(n)||e.isConstructSignatureDeclaration(n)||e.isCallSignatureDeclaration(n)||e.isMethodDeclaration(n)||e.isMethodSignature(n)||e.isFunctionDeclaration(n)||e.isParameter(n)||e.isTypeParameterDeclaration(n)||e.isExpressionWithTypeArguments(n)||e.isImportEqualsDeclaration(n)||e.isTypeAliasDeclaration(n)||e.isConstructorDeclaration(n)||e.isIndexSignatureDeclaration(n)}function t(n){return e.isSetAccessor(n)||e.isGetAccessor(n)?t:e.isMethodSignature(n)||e.isMethodDeclaration(n)?i:r(n);function t(e){var t=a(e);return void 0!==t?{diagnosticMessage:t,errorNode:n,typeName:n.name}:void 0}function a(t){return e.hasModifier(n,32)?t.errorModuleName?2===t.accessibility?e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_private_name_1:240===n.parent.kind?t.errorModuleName?2===t.accessibility?e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_private_name_1:t.errorModuleName?e.Diagnostics.Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Property_0_of_exported_interface_has_or_is_using_private_name_1}function i(e){var t=o(e);return void 0!==t?{diagnosticMessage:t,errorNode:n,typeName:n.name}:void 0}function o(t){return e.hasModifier(n,32)?t.errorModuleName?2===t.accessibility?e.Diagnostics.Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_static_method_0_of_exported_class_has_or_is_using_private_name_1:240===n.parent.kind?t.errorModuleName?2===t.accessibility?e.Diagnostics.Public_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_method_0_of_exported_class_has_or_is_using_private_name_1:t.errorModuleName?e.Diagnostics.Method_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Method_0_of_exported_interface_has_or_is_using_private_name_1}}function r(n){return e.isVariableDeclaration(n)||e.isPropertyDeclaration(n)||e.isPropertySignature(n)||e.isBindingElement(n)||e.isConstructorDeclaration(n)?r:e.isSetAccessor(n)||e.isGetAccessor(n)?a:e.isConstructSignatureDeclaration(n)||e.isCallSignatureDeclaration(n)||e.isMethodDeclaration(n)||e.isMethodSignature(n)||e.isFunctionDeclaration(n)||e.isIndexSignatureDeclaration(n)?i:e.isParameter(n)?e.isParameterPropertyDeclaration(n)&&e.hasModifier(n.parent,8)?r:o:e.isTypeParameterDeclaration(n)?c:e.isExpressionWithTypeArguments(n)?l:e.isImportEqualsDeclaration(n)?u:e.isTypeAliasDeclaration(n)?d:e.Debug.assertNever(n,"Attempted to set a declaration diagnostic context for unhandled node kind: "+e.SyntaxKind[n.kind]);function t(t){return 237===n.kind||186===n.kind?t.errorModuleName?2===t.accessibility?e.Diagnostics.Exported_variable_0_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Exported_variable_0_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Exported_variable_0_has_or_is_using_private_name_1:154===n.kind||153===n.kind||151===n.kind&&e.hasModifier(n.parent,8)?e.hasModifier(n,32)?t.errorModuleName?2===t.accessibility?e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_private_name_1:240===n.parent.kind||151===n.kind?t.errorModuleName?2===t.accessibility?e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_private_name_1:t.errorModuleName?e.Diagnostics.Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Property_0_of_exported_interface_has_or_is_using_private_name_1:void 0}function r(e){var r=t(e);return void 0!==r?{diagnosticMessage:r,errorNode:n,typeName:n.name}:void 0}function a(t){var r;return r=159===n.kind?e.hasModifier(n,32)?t.errorModuleName?e.Diagnostics.Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_private_name_1:t.errorModuleName?e.Diagnostics.Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_private_name_1:e.hasModifier(n,32)?t.errorModuleName?2===t.accessibility?e.Diagnostics.Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_private_name_1:t.errorModuleName?2===t.accessibility?e.Diagnostics.Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Return_type_of_public_getter_0_from_exported_class_has_or_is_using_private_name_1,{diagnosticMessage:r,errorNode:n.name,typeName:n.name}}function i(t){var r;switch(n.kind){case 161:r=t.errorModuleName?e.Diagnostics.Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_0;break;case 160:r=t.errorModuleName?e.Diagnostics.Return_type_of_call_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_call_signature_from_exported_interface_has_or_is_using_private_name_0;break;case 162:r=t.errorModuleName?e.Diagnostics.Return_type_of_index_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_index_signature_from_exported_interface_has_or_is_using_private_name_0;break;case 156:case 155:r=e.hasModifier(n,32)?t.errorModuleName?2===t.accessibility?e.Diagnostics.Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:e.Diagnostics.Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_public_static_method_from_exported_class_has_or_is_using_private_name_0:240===n.parent.kind?t.errorModuleName?2===t.accessibility?e.Diagnostics.Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:e.Diagnostics.Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_public_method_from_exported_class_has_or_is_using_private_name_0:t.errorModuleName?e.Diagnostics.Return_type_of_method_from_exported_interface_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_method_from_exported_interface_has_or_is_using_private_name_0;break;case 239:r=t.errorModuleName?2===t.accessibility?e.Diagnostics.Return_type_of_exported_function_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:e.Diagnostics.Return_type_of_exported_function_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_exported_function_has_or_is_using_private_name_0;break;default:return e.Debug.fail("This is unknown kind for signature: "+n.kind)}return{diagnosticMessage:r,errorNode:n.name||n}}function o(e){var t=s(e);return void 0!==t?{diagnosticMessage:t,errorNode:n,typeName:n.name}:void 0}function s(t){switch(n.parent.kind){case 157:return t.errorModuleName?2===t.accessibility?e.Diagnostics.Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_constructor_from_exported_class_has_or_is_using_private_name_1;case 161:case 166:return t.errorModuleName?e.Diagnostics.Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1;case 160:return t.errorModuleName?e.Diagnostics.Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1;case 162:return t.errorModuleName?e.Diagnostics.Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_private_name_1;case 156:case 155:return e.hasModifier(n.parent,32)?t.errorModuleName?2===t.accessibility?e.Diagnostics.Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1:240===n.parent.parent.kind?t.errorModuleName?2===t.accessibility?e.Diagnostics.Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1:t.errorModuleName?e.Diagnostics.Parameter_0_of_method_from_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1;case 239:case 165:return t.errorModuleName?2===t.accessibility?e.Diagnostics.Parameter_0_of_exported_function_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Parameter_0_of_exported_function_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_exported_function_has_or_is_using_private_name_1;default:return e.Debug.fail("Unknown parent for parameter: "+e.SyntaxKind[n.parent.kind])}}function c(){var t;switch(n.parent.kind){case 240:t=e.Diagnostics.Type_parameter_0_of_exported_class_has_or_is_using_private_name_1;break;case 241:t=e.Diagnostics.Type_parameter_0_of_exported_interface_has_or_is_using_private_name_1;break;case 166:case 161:t=e.Diagnostics.Type_parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1;break;case 160:t=e.Diagnostics.Type_parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1;break;case 156:case 155:t=e.hasModifier(n.parent,32)?e.Diagnostics.Type_parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1:240===n.parent.parent.kind?e.Diagnostics.Type_parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1:e.Diagnostics.Type_parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1;break;case 165:case 239:t=e.Diagnostics.Type_parameter_0_of_exported_function_has_or_is_using_private_name_1;break;case 242:t=e.Diagnostics.Type_parameter_0_of_exported_type_alias_has_or_is_using_private_name_1;break;default:return e.Debug.fail("This is unknown parent for type parameter: "+n.parent.kind)}return{diagnosticMessage:t,errorNode:n,typeName:n.name}}function l(){var t;return t=240===n.parent.parent.kind?e.isHeritageClause(n.parent)&&109===n.parent.token?e.Diagnostics.Implements_clause_of_exported_class_0_has_or_is_using_private_name_1:e.Diagnostics.extends_clause_of_exported_class_0_has_or_is_using_private_name_1:e.Diagnostics.extends_clause_of_exported_interface_0_has_or_is_using_private_name_1,{diagnosticMessage:t,errorNode:n,typeName:e.getNameOfDeclaration(n.parent.parent)}}function u(){return{diagnosticMessage:e.Diagnostics.Import_declaration_0_is_using_private_name_1,errorNode:n,typeName:n.name}}function d(){return{diagnosticMessage:e.Diagnostics.Exported_type_alias_0_has_or_is_using_private_name_1,errorNode:n.type,typeName:n.name}}}e.canProduceDiagnostics=n,e.createGetSymbolAccessibilityDiagnosticForNodeName=t,e.createGetSymbolAccessibilityDiagnosticForNode=r}(u||(u={})),function(e){function n(n,t,a){if(a&&e.isSourceFileJS(a))return[];var i=n.getCompilerOptions(),o=e.transformNodes(t,n,i,a?[a]:e.filter(n.getSourceFiles(),e.isSourceFileNotJS),[r],!1);return o.diagnostics}e.getDeclarationDiagnostics=n;var t=531469;function r(n){var r,d,p,m,f,_,g,v,y,h,b=function(){return e.Debug.fail("Diagnostic emitted without context")},E=b,T=!0,S=!1,x=!1,C=!1,A=!1,D=n.getEmitHost(),L={trackSymbol:G,reportInaccessibleThisError:K,reportInaccessibleUniqueSymbolError:B,reportPrivateInBaseOfClassExpression:V,reportLikelyUnsafeImportRequiredError:j,moduleResolverHost:D,trackReferencedAmbientModule:P,trackExternalModuleSymbolOfImportTypeNode:F},k=n.getEmitResolver(),N=n.getCompilerOptions(),I=e.getNewLineCharacter(N),w=N.noResolve,M=N.stripInternal;return U;function O(n){if(n){d=d||e.createMap();for(var t=0,r=n;t<r.length;t++){var a=r[t];d.set(a,!0)}}}function P(n,t){var r=k.getTypeReferenceDirectivesForSymbol(t,67108863);if(e.length(r))return O(r);var a=e.getSourceFileOfNode(n);y.set(""+e.getOriginalNodeId(a),a)}function R(t){if(0===t.accessibility){if(t&&t.aliasesToMakeVisible)if(p)for(var r=0,a=t.aliasesToMakeVisible;r<a.length;r++){var i=a[r];e.pushIfUnique(p,i)}else p=t.aliasesToMakeVisible}else{var o=E(t);o&&(o.typeName?n.addDiagnostic(e.createDiagnosticForNode(t.errorNode||o.errorNode,o.diagnosticMessage,e.getTextOfNode(o.typeName),t.errorSymbolName,t.errorModuleName)):n.addDiagnostic(e.createDiagnosticForNode(t.errorNode||o.errorNode,o.diagnosticMessage,t.errorSymbolName,t.errorModuleName)))}}function F(e){S||(_||(_=[])).push(e)}function G(e,n,t){262144&e.flags||(R(k.isSymbolAccessible(e,n,t,!0)),O(k.getTypeReferenceDirectivesForSymbol(e,t)))}function V(t){g&&n.addDiagnostic(e.createDiagnosticForNode(g,e.Diagnostics.Property_0_of_exported_class_expression_may_not_be_private_or_protected,t))}function B(){g&&n.addDiagnostic(e.createDiagnosticForNode(g,e.Diagnostics.The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary,e.declarationNameToString(g),"unique symbol"))}function K(){g&&n.addDiagnostic(e.createDiagnosticForNode(g,e.Diagnostics.The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary,e.declarationNameToString(g),"this"))}function j(t){g&&n.addDiagnostic(e.createDiagnosticForNode(g,e.Diagnostics.The_inferred_type_of_0_cannot_be_named_without_a_reference_to_1_This_is_likely_not_portable_A_type_annotation_is_necessary,e.declarationNameToString(g),t))}function U(t){if(279===t.kind&&(t.isDeclarationFile||e.isSourceFileJS(t)))return t;if(280===t.kind){S=!0,y=e.createMap(),h=e.createMap();var a=!1,i=e.createBundle(e.map(t.sourceFiles,function(t){if(!t.isDeclarationFile&&!e.isSourceFileJS(t)){if(a=a||t.hasNoDefaultLib,v=t,r=t,p=void 0,f=!1,m=e.createMap(),E=b,C=!1,A=!1,H(t,y),W(t,h),e.isExternalModule(t)){x=!1,T=!1;var i=e.visitNodes(t.statements,pe),o=e.updateSourceFileNode(t,[e.createModuleDeclaration([],[e.createModifier(125)],e.createLiteral(e.getResolvedExternalModuleName(n.getEmitHost(),t)),e.createModuleBlock(e.setTextRange(e.createNodeArray(se(i)),t.statements)))],!0,[],[],!1,[]);return o}T=!0;var s=e.visitNodes(t.statements,pe);return e.updateSourceFileNode(t,se(s),!0,[],[],!1,[])}}),e.mapDefined(t.prepends,function(n){if(282===n.kind)return e.createUnparsedSourceFile(n,"dts")}));i.syntheticFileReferences=[],i.syntheticTypeReferences=w(),i.syntheticLibReferences=I(),i.hasNoDefaultLib=a;var o=e.getDirectoryPath(e.normalizeSlashes(e.getOutputPathsFor(t,D,!0).declarationFilePath)),s=O(i.syntheticFileReferences,o);return y.forEach(s),i}T=!0,C=!1,A=!1,r=t,v=t,E=b,S=!1,x=!1,f=!1,p=void 0,m=e.createMap(),d=void 0,y=H(v,e.createMap()),h=W(v,e.createMap());var c=[],l=e.getDirectoryPath(e.normalizeSlashes(e.getOutputPathsFor(t,D,!0).declarationFilePath)),u=O(c,l),g=e.visitNodes(t.statements,pe),L=e.setTextRange(e.createNodeArray(se(g)),t.statements);y.forEach(u);var k=e.filter(L,e.isAnyImportSyntax);e.isExternalModule(t)&&(!x||C&&!A)&&(L=e.setTextRange(e.createNodeArray(L.concat([e.createExportDeclaration(void 0,void 0,e.createNamedExports([]),void 0)])),L));var N=e.updateSourceFileNode(t,L,!0,c,w(),t.hasNoDefaultLib,I());return N.exportedModulesFromDeclarationEmit=_,N;function I(){return e.map(e.arrayFrom(h.keys()),function(e){return{fileName:e,pos:-1,end:-1}})}function w(){return d?e.mapDefined(e.arrayFrom(d.keys()),M):[]}function M(n){if(k)for(var t=0,r=k;t<r.length;t++){var a=r[t];if(e.isImportEqualsDeclaration(a)&&e.isExternalModuleReference(a.moduleReference)){var i=a.moduleReference.expression;if(e.isStringLiteralLike(i)&&i.text===n)return}else if(e.isImportDeclaration(a)&&e.isStringLiteral(a.moduleSpecifier)&&a.moduleSpecifier.text===n)return}return{fileName:n,pos:-1,end:-1}}function O(n,r){return function(a){var i;if(a.isDeclarationFile)i=a.fileName;else{if(S&&e.contains(t.sourceFiles,a))return;var o=e.getOutputPathsFor(a,D,!0);i=o.declarationFilePath||o.jsFilePath||a.fileName}if(i){var s=e.getRelativePathToDirectoryOrUrl(r,i,D.getCurrentDirectory(),D.getCanonicalFileName,!1);if(e.startsWith(s,"./")&&e.hasExtension(s)&&(s=s.substring(2)),e.startsWith(s,"node_modules/")||-1!==s.indexOf("/node_modules/"))return;n.push({pos:-1,end:-1,fileName:s})}}}}function H(n,t){return w||e.isSourceFileJS(n)?t:(e.forEach(n.referencedFiles,function(r){var a=e.tryResolveScriptReference(D,n,r);a&&t.set(""+e.getOriginalNodeId(a),a)}),t)}function W(n,t){return e.forEach(n.libReferenceDirectives,function(e){var n=D.getLibFileFromReference(e);n&&t.set(e.fileName.toLocaleLowerCase(),!0)}),t}function q(n){return 72===n.kind?n:185===n.kind?e.updateArrayBindingPattern(n,e.visitNodes(n.elements,t)):e.updateObjectBindingPattern(n,e.visitNodes(n.elements,t));function t(n){return 210===n.kind?n:e.updateBindingElement(n,n.dotDotDotToken,n.propertyName,q(n.name),J(n)?n.initializer:void 0)}}function z(n,t){var r;f||(r=E,E=e.createGetSymbolAccessibilityDiagnosticForNode(n));var a=e.updateParameter(n,void 0,i(n,t),n.dotDotDotToken,q(n.name),k.isOptionalParameter(n)?n.questionToken||e.createToken(56):void 0,Y(n,n.type,!0),X(n));return f||(E=r),a}function J(n){return c(n)&&k.isLiteralConstDeclaration(e.getParseTreeNode(n))}function X(n){if(J(n))return k.createLiteralConstValue(e.getParseTreeNode(n),L)}function Y(n,a,i){if((i||!e.hasModifier(n,8))&&!J(n)){var o,s=151===n.kind&&(k.isRequiredInitializedParameter(n)||k.isOptionalUninitializedParameterProperty(n));return a&&!s?e.visitNode(a,ue):e.getParseTreeNode(n)?159===n.kind?e.createKeywordTypeNode(120):(g=n.name,f||(o=E,E=e.createGetSymbolAccessibilityDiagnosticForNode(n)),237===n.kind||186===n.kind?c(k.createTypeOfDeclaration(n,r,t,L)):151===n.kind||154===n.kind||153===n.kind?n.initializer?c(k.createTypeOfDeclaration(n,r,t,L,s)||k.createTypeOfExpression(n.initializer,r,t,L)):c(k.createTypeOfDeclaration(n,r,t,L,s)):c(k.createReturnTypeOfSignatureDeclaration(n,r,t,L))):a?e.visitNode(a,ue):e.createKeywordTypeNode(120)}function c(n){return g=void 0,f||(E=o),n||e.createKeywordTypeNode(120)}}function Q(n){switch(n=e.getParseTreeNode(n),n.kind){case 239:case 244:case 241:case 240:case 242:case 243:return!k.isDeclarationVisible(n);case 237:return!Z(n);case 248:case 249:case 255:case 254:return!1}return!1}function Z(n){return!e.isOmittedExpression(n)&&(e.isBindingPattern(n.name)?e.some(n.name.elements,Z):k.isDeclarationVisible(n))}function $(n,t,r){if(!e.hasModifier(n,8)){var a=e.map(t,function(e){return z(e,r)});if(a)return e.createNodeArray(a,t.hasTrailingComma)}}function ee(n,t){return e.hasModifier(n,8)?void 0:e.visitNodes(t,ue)}function ne(n){return e.isSourceFile(n)||e.isTypeAliasDeclaration(n)||e.isModuleDeclaration(n)||e.isClassDeclaration(n)||e.isInterfaceDeclaration(n)||e.isFunctionLike(n)||e.isIndexSignatureDeclaration(n)||e.isMappedTypeNode(n)}function te(e,n){var t=k.isEntityNameVisible(e,n);R(t),O(k.getTypeReferenceDirectivesForEntityName(e))}function re(n,t){return e.hasJSDocNodes(n)&&e.hasJSDocNodes(t)&&(n.jsDoc=t.jsDoc),e.setCommentRange(n,e.getCommentRange(t))}function ae(t,r){if(r){if(x=x||244!==t.kind&&183!==t.kind,e.isStringLiteralLike(r))if(S){var a=e.getExternalModuleNameFromDeclaration(n.getEmitHost(),k,t);if(a)return e.createLiteral(a)}else{var i=k.getSymbolOfExternalModuleSpecifier(r);i&&(_||(_=[])).push(i)}return r}}function ie(n){if(k.isDeclarationVisible(n)){if(259===n.moduleReference.kind){var t=e.getExternalModuleImportEqualsDeclarationExpression(n);return e.updateImportEqualsDeclaration(n,void 0,n.modifiers,n.name,e.updateExternalModuleReference(n.moduleReference,ae(n,t)))}var a=E;return E=e.createGetSymbolAccessibilityDiagnosticForNode(n),te(n.moduleReference,r),E=a,n}}function oe(n){if(!n.importClause)return e.updateImportDeclaration(n,void 0,n.modifiers,n.importClause,ae(n,n.moduleSpecifier));var t=n.importClause&&n.importClause.name&&k.isDeclarationVisible(n.importClause)?n.importClause.name:void 0;if(!n.importClause.namedBindings)return t&&e.updateImportDeclaration(n,void 0,n.modifiers,e.updateImportClause(n.importClause,t,void 0),ae(n,n.moduleSpecifier));if(251===n.importClause.namedBindings.kind){var r=k.isDeclarationVisible(n.importClause.namedBindings)?n.importClause.namedBindings:void 0;return t||r?e.updateImportDeclaration(n,void 0,n.modifiers,e.updateImportClause(n.importClause,t,r),ae(n,n.moduleSpecifier)):void 0}var a=e.mapDefined(n.importClause.namedBindings.elements,function(e){return k.isDeclarationVisible(e)?e:void 0});return a&&a.length||t?e.updateImportDeclaration(n,void 0,n.modifiers,e.updateImportClause(n.importClause,t,a&&a.length?e.updateNamedImports(n.importClause.namedBindings,a):void 0),ae(n,n.moduleSpecifier)):void 0}function se(n){while(e.length(p)){var t=p.shift();if(!e.isLateVisibilityPaintedStatement(t))return e.Debug.fail("Late replaced statement was found which is not handled by the declaration transformer!: "+(e.SyntaxKind?e.SyntaxKind[t.kind]:t.kind));var r=T;T=t.parent&&e.isSourceFile(t.parent)&&!(e.isExternalModule(t.parent)&&S);var a=me(t,!0);T=r,m.set(""+e.getOriginalNodeId(t),a)}return e.visitNodes(n,i);function i(n){if(e.isLateVisibilityPaintedStatement(n)){var t=""+e.getOriginalNodeId(n);if(m.has(t)){var r=m.get(t);return m.delete(t),r&&e.isSourceFile(n.parent)&&((e.isArray(r)?e.some(r,le):le(r))&&(C=!0),(e.isArray(r)?e.some(r,ce):ce(r))&&(x=!0)),r}}return n}}function ce(n){return e.isAnyImportOrReExport(n)||e.isExportAssignment(n)||e.hasModifier(n,1)}function le(n){return!e.isAnyImportOrReExport(n)&&!e.isExportAssignment(n)&&!e.hasModifier(n,1)&&!e.isAmbientModule(n)}function ue(t){if(!he(t)){if(e.isDeclaration(t)){if(Q(t))return;if(e.hasDynamicName(t)&&!k.isLateBound(e.getParseTreeNode(t)))return}if((!e.isFunctionLike(t)||!k.isImplementationOfOverload(t))&&!e.isSemicolonClassElement(t)){var a;ne(t)&&(a=r,r=t);var i=E;if((e.isMethodDeclaration(t)||e.isMethodSignature(t))&&e.hasModifier(t,8)){if(t.symbol&&t.symbol.declarations&&t.symbol.declarations[0]!==t)return;return T(e.createProperty(void 0,Te(t),t.name,void 0,void 0,void 0))}var o=e.canProduceDiagnostics(t);o&&!f&&(E=e.createGetSymbolAccessibilityDiagnosticForNode(t)),e.isTypeQueryNode(t)&&te(t.exprName,r);var s=f,c=(168===t.kind||181===t.kind)&&242!==t.parent.kind;if(c&&(f=!0),u(t))switch(t.kind){case 211:(e.isEntityName(t.expression)||e.isEntityNameExpression(t.expression))&&te(t.expression,r);var l=e.visitEachChild(t,ue,n);return T(e.updateExpressionWithTypeArguments(l,e.parenthesizeTypeParameters(l.typeArguments),l.expression));case 164:te(t.typeName,r);l=e.visitEachChild(t,ue,n);return T(e.updateTypeReferenceNode(l,l.typeName,e.parenthesizeTypeParameters(l.typeArguments)));case 161:return T(e.updateConstructSignature(t,ee(t,t.typeParameters),$(t,t.parameters),Y(t,t.type)));case 157:var d=e.hasModifier(t,8),p=e.createSignatureDeclaration(157,d?void 0:ee(t,t.typeParameters),d?void 0:$(t,t.parameters,0),void 0);return p.modifiers=e.createNodeArray(Te(t)),T(p);case 156:var m=e.createSignatureDeclaration(155,ee(t,t.typeParameters),$(t,t.parameters),Y(t,t.type));return m.name=t.name,m.modifiers=e.createNodeArray(Te(t)),m.questionToken=t.questionToken,T(m);case 158:var _=xe(t);return T(_);case 159:_=xe(t);return T(_);case 154:return T(e.updateProperty(t,void 0,Te(t),t.name,t.questionToken,e.hasModifier(t,8)?void 0:Y(t,t.type),X(t)));case 153:return T(e.updatePropertySignature(t,Te(t),t.name,t.questionToken,e.hasModifier(t,8)?void 0:Y(t,t.type),X(t)));case 155:return T(e.updateMethodSignature(t,ee(t,t.typeParameters),$(t,t.parameters),Y(t,t.type),t.name,t.questionToken));case 160:return T(e.updateCallSignature(t,ee(t,t.typeParameters),$(t,t.parameters),Y(t,t.type)));case 162:return T(e.updateIndexSignature(t,void 0,Te(t),$(t,t.parameters),e.visitNode(t.type,ue)||e.createKeywordTypeNode(120)));case 237:return e.isBindingPattern(t.name)?_e(t.name):(c=!0,f=!0,T(e.updateVariableDeclaration(t,t.name,Y(t,t.type),X(t))));case 150:return de(t)&&(t.default||t.constraint)?T(e.updateTypeParameterDeclaration(t,t.name,void 0,void 0)):T(e.visitEachChild(t,ue,n));case 175:var g=e.visitNode(t.checkType,ue),v=e.visitNode(t.extendsType,ue),y=r;r=t.trueType;var h=e.visitNode(t.trueType,ue);r=y;var b=e.visitNode(t.falseType,ue);return T(e.updateConditionalTypeNode(t,g,v,h,b));case 165:return T(e.updateFunctionTypeNode(t,e.visitNodes(t.typeParameters,ue),$(t,t.parameters),e.visitNode(t.type,ue)));case 166:return T(e.updateConstructorTypeNode(t,e.visitNodes(t.typeParameters,ue),$(t,t.parameters),e.visitNode(t.type,ue)));case 183:return e.isLiteralImportTypeNode(t)?T(e.updateImportTypeNode(t,e.updateLiteralTypeNode(t.argument,ae(t,t.argument.literal)),t.qualifier,e.visitNodes(t.typeArguments,ue,e.isTypeNode),t.isTypeOf)):T(t);default:e.Debug.assertNever(t,"Attempted to process unhandled node kind: "+e.SyntaxKind[t.kind])}return T(e.visitEachChild(t,ue,n))}}function T(n){return n&&o&&e.hasDynamicName(t)&&ve(t),ne(t)&&(r=a),o&&!f&&(E=i),c&&(f=s),n===t?n:n&&e.setOriginalNode(re(n,t),t)}}function de(n){return 156===n.parent.kind&&e.hasModifier(n.parent,8)}function pe(n){if(l(n)&&!he(n)){switch(n.kind){case 255:return e.isSourceFile(n.parent)&&(x=!0,A=!0),e.updateExportDeclaration(n,void 0,n.modifiers,n.exportClause,ae(n,n.moduleSpecifier));case 254:if(e.isSourceFile(n.parent)&&(x=!0,A=!0),72===n.expression.kind)return n;var r=e.createOptimisticUniqueName("_default");E=function(){return{diagnosticMessage:e.Diagnostics.Default_export_of_the_module_has_or_is_using_private_name_0,errorNode:n}};var a=e.createVariableDeclaration(r,k.createTypeOfExpression(n.expression,n,t,L),void 0),i=e.createVariableStatement(T?[e.createModifier(125)]:[],e.createVariableDeclarationList([a],2));return[i,e.updateExportAssignment(n,n.decorators,n.modifiers,r)]}var o=me(n);return m.set(""+e.getOriginalNodeId(n),o),n}}function me(n,a){if(!he(n)){switch(n.kind){case 248:return ie(n);case 249:return oe(n)}if((!e.isDeclaration(n)||!Q(n))&&(!e.isFunctionLike(n)||!k.isImplementationOfOverload(n))){var i;ne(n)&&(i=r,r=n);var o=e.canProduceDiagnostics(n),s=E;o&&(E=e.createGetSymbolAccessibilityDiagnosticForNode(n));var c=T;switch(n.kind){case 242:return O(e.updateTypeAliasDeclaration(n,void 0,Te(n,a),n.name,e.visitNodes(n.typeParameters,ue,e.isTypeParameterDeclaration),e.visitNode(n.type,ue,e.isTypeNode)));case 241:return O(e.updateInterfaceDeclaration(n,void 0,Te(n,a),n.name,ee(n,n.typeParameters),Ce(n.heritageClauses),e.visitNodes(n.members,ue)));case 239:var l=O(e.updateFunctionDeclaration(n,void 0,Te(n,a),void 0,n.name,ee(n,n.typeParameters),$(n,n.parameters),Y(n,n.type),void 0));if(l&&k.isExpandoFunctionDeclaration(n)){var u=e.mapDefined(k.getPropertiesOfContainerFunction(n),function(n){if(e.isPropertyAccessExpression(n.valueDeclaration)){var a=k.createTypeOfDeclaration(n.valueDeclaration,r,t,L),i=e.createVariableDeclaration(e.unescapeLeadingUnderscores(n.escapedName),a,void 0);return e.createVariableStatement(void 0,e.createVariableDeclarationList([i]))}}),d=e.createModuleDeclaration(void 0,Te(n,a),n.name,e.createModuleBlock(u),16);return[l,d]}return l;case 244:T=!1;var p=n.body;if(p&&245===p.kind){var f=e.visitNodes(p.statements,pe),_=e.updateModuleBlock(p,se(f));T=c;var g=Te(n,a);return O(e.updateModuleDeclaration(n,void 0,g,e.isExternalModuleAugmentation(n)?ae(n,n.name):n.name,_))}T=c;g=Te(n,a);T=!1,e.visitNode(p,pe);var v=""+e.getOriginalNodeId(p);_=m.get(v);return m.delete(v),O(e.updateModuleDeclaration(n,void 0,g,n.name,_));case 240:var y=e.createNodeArray(Te(n,a)),h=ee(n,n.typeParameters),b=e.getFirstConstructorWithBody(n),S=void 0;if(b){var x=E;S=e.compact(e.flatMap(b.parameters,function(n){if(e.hasModifier(n,92))return E=e.createGetSymbolAccessibilityDiagnosticForNode(n),72===n.name.kind?re(e.createProperty(void 0,Te(n),n.name,n.questionToken,Y(n,n.type),X(n)),n):t(n.name);function t(r){for(var a,i=0,o=r.elements;i<o.length;i++){var s=o[i];e.isOmittedExpression(s)||(e.isBindingPattern(s.name)&&(a=e.concatenate(a,t(s.name))),a=a||[],a.push(e.createProperty(void 0,Te(n),s.name,void 0,Y(s,void 0),void 0)))}return a}})),E=x}var C=e.createNodeArray(e.concatenate(S,e.visitNodes(n.members,ue))),A=e.getEffectiveBaseTypeNode(n);if(A&&!e.isEntityNameExpression(A.expression)&&96!==A.expression.kind){var D=n.name?e.unescapeLeadingUnderscores(n.name.escapedText):"default",N=e.createOptimisticUniqueName(D+"_base");E=function(){return{diagnosticMessage:e.Diagnostics.extends_clause_of_exported_class_0_has_or_is_using_private_name_1,errorNode:A,typeName:n.name}};var I=e.createVariableDeclaration(N,k.createTypeOfExpression(A.expression,n,t,L),void 0),w=e.createVariableStatement(T?[e.createModifier(125)]:[],e.createVariableDeclarationList([I],2)),M=e.createNodeArray(e.map(n.heritageClauses,function(n){if(86===n.token){var t=E;E=e.createGetSymbolAccessibilityDiagnosticForNode(n.types[0]);var r=e.updateHeritageClause(n,e.map(n.types,function(n){return e.updateExpressionWithTypeArguments(n,e.visitNodes(n.typeArguments,ue),N)}));return E=t,r}return e.updateHeritageClause(n,e.visitNodes(e.createNodeArray(e.filter(n.types,function(n){return e.isEntityNameExpression(n.expression)||96===n.expression.kind})),ue))}));return[w,O(e.updateClassDeclaration(n,void 0,y,n.name,h,M,C))]}M=Ce(n.heritageClauses);return O(e.updateClassDeclaration(n,void 0,y,n.name,h,M,C));case 219:return O(fe(n,a));case 243:return O(e.updateEnumDeclaration(n,void 0,e.createNodeArray(Te(n,a)),n.name,e.createNodeArray(e.mapDefined(n.members,function(n){if(!he(n)){var t=k.getConstantValue(n);return re(e.updateEnumMember(n,n.name,void 0!==t?e.createLiteral(t):void 0),n)}}))))}return e.Debug.assertNever(n,"Unhandled top-level node in declaration emit: "+e.SyntaxKind[n.kind])}}function O(t){return ne(n)&&(r=i),o&&(E=s),244===n.kind&&(T=c),t===n?t:t&&e.setOriginalNode(re(t,n),n)}}function fe(n,t){if(e.forEach(n.declarationList.declarations,Z)){var r=e.visitNodes(n.declarationList.declarations,ue);if(e.length(r))return e.updateVariableStatement(n,e.createNodeArray(Te(n,t)),e.updateVariableDeclarationList(n.declarationList,r))}}function _e(n){return e.flatten(e.mapDefined(n.elements,function(e){return ge(e)}))}function ge(n){if(210!==n.kind&&n.name){if(!Z(n))return;return e.isBindingPattern(n.name)?_e(n.name):e.createVariableDeclaration(n.name,Y(n,void 0),void 0)}}function ve(n){var t;f||(t=E,E=e.createGetSymbolAccessibilityDiagnosticForNodeName(n)),g=n.name,e.Debug.assert(k.isLateBound(e.getParseTreeNode(n)));var a=n,i=a.name.expression;te(i,r),f||(E=t),g=void 0}function ye(n){var t=v.text.substring(n.pos,n.end);return e.stringContains(t,"@internal")}function he(n){if(M&&n){var t=e.getLeadingCommentRangesOfNode(e.getParseTreeNode(n),v);if(e.forEach(t,ye))return!0}return!1}function be(n){return e.isExportAssignment(n)||e.isExportDeclaration(n)}function Ee(n){return!!e.isModuleBlock(n)&&e.some(n.statements,be)}function Te(n,t){var r=e.getModifierFlags(n),a=Se(n,t);return r===a?n.modifiers:e.createModifiersFromModifierFlags(a)}function Se(n,t){var r=2811,i=T&&!a(n)?2:0,s=279===n.parent.kind;return(!s||S&&s&&e.isExternalModule(n.parent))&&(r^=2|(t||S&&s||Ee(n.parent)?0:1),i=0),o(n,r,i)}function xe(n){var t=k.getAllAccessorDeclarations(n);if(n.kind===t.firstAccessor.kind){var r=s(n);!r&&t.secondAccessor&&(r=s(t.secondAccessor),E=e.createGetSymbolAccessibilityDiagnosticForNode(t.secondAccessor));var a=e.createProperty(void 0,i(n,void 0,t.setAccessor?0:64),n.name,n.questionToken,Y(n,r),void 0),o=t.secondAccessor&&e.getLeadingCommentRangesOfNode(t.secondAccessor,v);if(o)for(var c=function(n){if(3===n.kind){var t=v.text.slice(n.pos+2,n.end-2),r=t.split(/\r\n?|\n/g);if(r.length>1){var i=r.slice(1),o=e.guessIndentation(i);t=[r[0]].concat(e.map(i,function(e){return e.slice(o)})).join(I)}e.addSyntheticLeadingComment(a,n.kind,t,n.hasTrailingNewLine)}},l=0,u=o;l<u.length;l++){var d=u[l];c(d)}return a}}function Ce(n){return e.createNodeArray(e.filter(e.map(n,function(n){return e.updateHeritageClause(n,e.visitNodes(e.createNodeArray(e.filter(n.types,function(t){return e.isEntityNameExpression(t.expression)||86===n.token&&96===t.expression.kind})),ue))}),function(e){return e.types&&!!e.types.length}))}}function a(e){return 241===e.kind}function i(n,t,r){return e.createModifiersFromModifierFlags(o(n,t,r))}function o(n,t,r){void 0===t&&(t=3067),void 0===r&&(r=0);var a=e.getModifierFlags(n)&t|r;return 512&a&&!(1&a)&&(a^=1),512&a&&2&a&&(a^=2),a}function s(e){if(e)return 158===e.kind?e.type:e.parameters.length>0?e.parameters[0].type:void 0}function c(n){switch(n.kind){case 154:case 153:return!e.hasModifier(n,8);case 151:case 237:return!0}return!1}function l(e){switch(e.kind){case 239:case 244:case 248:case 241:case 240:case 242:case 243:case 219:case 249:case 255:case 254:return!0}return!1}function u(e){switch(e.kind){case 161:case 157:case 156:case 158:case 159:case 154:case 153:case 155:case 160:case 162:case 237:case 150:case 211:case 164:case 175:case 165:case 166:case 183:return!0}return!1}e.transformDeclarations=r}(u||(u={})),function(e){function n(n){switch(n){case e.ModuleKind.ESNext:case e.ModuleKind.ES2015:return e.transformES2015Module;case e.ModuleKind.System:return e.transformSystemModule;default:return e.transformModule}}var t,r;function a(t,r){var a=t.jsx,i=e.getEmitScriptTarget(t),o=e.getEmitModuleKind(t),s=[];return e.addRange(s,r&&r.before),s.push(e.transformTypeScript),2===a&&s.push(e.transformJsx),i<6&&s.push(e.transformESNext),i<4&&s.push(e.transformES2017),i<3&&s.push(e.transformES2016),i<2&&(s.push(e.transformES2015),s.push(e.transformGenerators)),s.push(n(o)),i<1&&s.push(e.transformES5),e.addRange(s,r&&r.after),s}function i(e,n){return n}function o(e,n,t){t(e,n)}function s(n,t,r,a,s,c){for(var l,u,d,p=new Array(312),m=[],f=[],_=0,g=!1,v=i,y=o,h=0,b=[],E={getCompilerOptions:function(){return r},getEmitResolver:function(){return n},getEmitHost:function(){return t},startLexicalEnvironment:R,suspendLexicalEnvironment:F,resumeLexicalEnvironment:G,endLexicalEnvironment:V,hoistVariableDeclaration:O,hoistFunctionDeclaration:P,requestEmitHelper:B,readEmitHelpers:K,enableSubstitution:L,enableEmitNotification:I,isSubstitutionEnabled:k,isEmitNotificationEnabled:w,get onSubstituteNode(){return v},set onSubstituteNode(n){e.Debug.assert(h<1,"Cannot modify transformation hooks after initialization has completed."),e.Debug.assert(void 0!==n,"Value must not be 'undefined'"),v=n},get onEmitNode(){return y},set onEmitNode(n){e.Debug.assert(h<1,"Cannot modify transformation hooks after initialization has completed."),e.Debug.assert(void 0!==n,"Value must not be 'undefined'"),y=n},addDiagnostic:function(e){b.push(e)}},T=0,S=a;T<S.length;T++){var x=S[T];e.disposeEmitNodes(e.getSourceFileOfNode(e.getParseTreeNode(x)))}e.performance.mark("beforeTransform");var C=e.chain.apply(void 0,s)(E);h=1;var A=e.map(a,c?C:D);return h=2,e.performance.mark("afterTransform"),e.performance.measure("transformTime","beforeTransform","afterTransform"),{transformed:A,substituteNode:N,emitNodeWithNotification:M,dispose:j,diagnostics:b};function D(n){return!n||e.isSourceFile(n)&&n.isDeclarationFile?n:C(n)}function L(n){e.Debug.assert(h<2,"Cannot modify the transformation context after transformation has completed."),p[n]|=1}function k(n){return 0!==(1&p[n.kind])&&0===(4&e.getEmitFlags(n))}function N(n,t){return e.Debug.assert(h<3,"Cannot substitute a node after the result is disposed."),t&&k(t)&&v(n,t)||t}function I(n){e.Debug.assert(h<2,"Cannot modify the transformation context after transformation has completed."),p[n]|=2}function w(n){return 0!==(2&p[n.kind])||0!==(2&e.getEmitFlags(n))}function M(n,t,r){e.Debug.assert(h<3,"Cannot invoke TransformationResult callbacks after the result is disposed."),t&&(w(t)?y(n,t,r):r(n,t))}function O(n){e.Debug.assert(h>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(h<2,"Cannot modify the lexical environment after transformation has completed.");var t=e.setEmitFlags(e.createVariableDeclaration(n),64);l?l.push(t):l=[t]}function P(n){e.Debug.assert(h>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(h<2,"Cannot modify the lexical environment after transformation has completed."),u?u.push(n):u=[n]}function R(){e.Debug.assert(h>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(h<2,"Cannot modify the lexical environment after transformation has completed."),e.Debug.assert(!g,"Lexical environment is suspended."),m[_]=l,f[_]=u,_++,l=void 0,u=void 0}function F(){e.Debug.assert(h>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(h<2,"Cannot modify the lexical environment after transformation has completed."),e.Debug.assert(!g,"Lexical environment is already suspended."),g=!0}function G(){e.Debug.assert(h>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(h<2,"Cannot modify the lexical environment after transformation has completed."),e.Debug.assert(g,"Lexical environment is not suspended."),g=!1}function V(){var n;if(e.Debug.assert(h>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(h<2,"Cannot modify the lexical environment after transformation has completed."),e.Debug.assert(!g,"Lexical environment is suspended."),(l||u)&&(u&&(n=u.slice()),l)){var t=e.createVariableStatement(void 0,e.createVariableDeclarationList(l));n?n.push(t):n=[t]}return _--,l=m[_],u=f[_],0===_&&(m=[],f=[]),n}function B(n){e.Debug.assert(h>0,"Cannot modify the transformation context during initialization."),e.Debug.assert(h<2,"Cannot modify the transformation context after transformation has completed."),e.Debug.assert(!n.scoped,"Cannot request a scoped emit helper."),d=e.append(d,n)}function K(){e.Debug.assert(h>0,"Cannot modify the transformation context during initialization."),e.Debug.assert(h<2,"Cannot modify the transformation context after transformation has completed.");var n=d;return d=void 0,n}function j(){if(h<3){for(var n=0,t=a;n<t.length;n++){var r=t[n];e.disposeEmitNodes(e.getSourceFileOfNode(e.getParseTreeNode(r)))}l=void 0,m=void 0,u=void 0,f=void 0,v=void 0,y=void 0,d=void 0,h=3}}}(function(e){e[e["Uninitialized"]=0]="Uninitialized",e[e["Initialized"]=1]="Initialized",e[e["Completed"]=2]="Completed",e[e["Disposed"]=3]="Disposed"})(t||(t={})),function(e){e[e["Substitution"]=1]="Substitution",e[e["EmitNotifications"]=2]="EmitNotifications"}(r||(r={})),e.getTransformers=a,e.noEmitSubstitution=i,e.noEmitNotification=o,e.transformNodes=s}(u||(u={})),function(e){var n,t,r=".tsbundleinfo",a=f(),i={pos:-1,end:-1};function o(n,t,r,a){void 0===a&&(a=!1);var i=e.isArray(r)?r:e.getSourceFilesToEmit(n,r),o=n.getCompilerOptions();if(o.outFile||o.out){if(i.length){var s=e.createBundle(i,n.getPrependNodes()),l=t(c(s,n,a),s);if(l)return l}}else for(var u=0,d=i;u<d.length;u++){var p=d[u];l=t(c(p,n,a),p);if(l)return l}}function s(n,t){var a=n.outFile||n.out,i=n.emitDeclarationOnly?void 0:a,o=i&&l(i,n),s=t||e.getEmitDeclarations(n)?e.removeFileExtension(a)+".d.ts":void 0,c=s&&e.getAreDeclarationMapsEnabled(n)?s+".map":void 0,u=n.references&&i?e.removeFileExtension(i)+r:void 0;return{jsFilePath:i,sourceMapFilePath:o,declarationFilePath:s,declarationMapPath:c,bundleInfoPath:u}}function c(n,t,r){var a=t.getCompilerOptions();if(280===n.kind)return s(a,r);var i=e.getOwnEmitOutputFilePath(n.fileName,t,d(n,a)),o=e.isJsonSourceFile(n)&&0===e.comparePaths(n.fileName,i,t.getCurrentDirectory(),!t.useCaseSensitiveFileNames()),c=a.emitDeclarationOnly||o?void 0:i,u=!c||e.isJsonSourceFile(n)?void 0:l(c,a),p=e.isSourceFileJS(n),m=!r&&!e.getEmitDeclarations(a)||p?void 0:e.getDeclarationEmitOutputFilePath(n.fileName,t),f=m&&e.getAreDeclarationMapsEnabled(a)?m+".map":void 0;return{jsFilePath:c,sourceMapFilePath:u,declarationFilePath:m,declarationMapPath:f,bundleInfoPath:void 0}}function l(e,n){return n.sourceMap&&!n.inlineSourceMap?e+".map":void 0}function u(){return{originalOffset:-1,totalLength:-1}}function d(n,t){if(e.isJsonSourceFile(n))return".json";if(1===t.jsx)if(e.isSourceFileJS(n)){if(e.fileExtensionIs(n.fileName,".jsx"))return".jsx"}else if(1===n.languageVariant)return".jsx";return".js"}function p(n,t,r,a,i,s){var c,l=t.getCompilerOptions(),d=l.sourceMap||l.inlineSourceMap||e.getAreDeclarationMapsEnabled(l)?[]:void 0,p=l.listEmittedFiles?[]:void 0,f=e.createDiagnosticCollection(),_=e.getNewLineCharacter(l,function(){return t.getNewLine()}),g=e.createTextWriter(_),v=e.performance.createTimer("printTime","beforePrint","afterPrint"),y=v.enter,h=v.exit,b=u(),E=!1;return y(),o(t,T,e.getSourceFilesToEmit(t,r),a),h(),{emitSkipped:E,diagnostics:f.getDiagnostics(),emittedFiles:p,sourceMaps:d,exportedModulesFromDeclarationEmit:c};function T(e,n){var t=e.jsFilePath,r=e.sourceMapFilePath,i=e.declarationFilePath,o=e.declarationMapPath,s=e.bundleInfoPath;S(n,t,r,s),x(n,i,o),!E&&p&&(a||(t&&p.push(t),r&&p.push(r),s&&p.push(s)),i&&p.push(i),o&&p.push(o))}function S(r,o,s,c){if(!a&&o)if(o&&t.isEmitBlocked(o)||l.noEmit)E=!0;else{var u=e.transformNodes(n,t,l,[r],i,!1),d={removeComments:l.removeComments,newLine:l.newLine,noEmitHelpers:l.noEmitHelpers,module:l.module,target:l.target,sourceMap:l.sourceMap,inlineSourceMap:l.inlineSourceMap,inlineSources:l.inlineSources,extendedDiagnostics:l.extendedDiagnostics},p=m(d,{hasGlobalName:n.hasGlobalName,onEmitNode:u.emitNodeWithNotification,substituteNode:u.substituteNode});e.Debug.assert(1===u.transformed.length,"Should only see one output from the transform"),A(o,s,u.transformed[0],c,p,l),u.dispose()}}function x(r,i,o){if(i&&!e.isInJSFile(r)){var u=e.isSourceFile(r)?[r]:r.sourceFiles,d=e.filter(u,e.isSourceFileNotJS),p=l.outFile||l.out?[e.createBundle(d,e.isSourceFile(r)?void 0:r.prepends)]:d;a&&!e.getEmitDeclarations(l)&&d.forEach(C);var _=e.transformNodes(n,t,l,p,e.concatenate([e.transformDeclarations],s),!1);if(e.length(_.diagnostics))for(var g=0,v=_.diagnostics;g<v.length;g++){var y=v[g];f.add(y)}var h={removeComments:l.removeComments,newLine:l.newLine,noEmitHelpers:!0,module:l.module,target:l.target,sourceMap:l.sourceMap,inlineSourceMap:l.inlineSourceMap,extendedDiagnostics:l.extendedDiagnostics,onlyPrintJsDocStyle:!0},b=m(h,{hasGlobalName:n.hasGlobalName,onEmitNode:_.emitNodeWithNotification,substituteNode:_.substituteNode}),T=!!_.diagnostics&&!!_.diagnostics.length||!!t.isEmitBlocked(i)||!!l.noEmit;if(E=E||T,(!T||a)&&(e.Debug.assert(1===_.transformed.length,"Should only see one output from the decl transform"),A(i,o,_.transformed[0],void 0,b,{sourceMap:l.declarationMap,sourceRoot:l.sourceRoot,mapRoot:l.mapRoot,extendedDiagnostics:l.extendedDiagnostics}),a&&279===_.transformed[0].kind)){var S=_.transformed[0];c=S.exportedModulesFromDeclarationEmit}_.dispose()}}function C(t){e.isExportAssignment(t)?72===t.expression.kind&&n.collectLinkedAliases(t.expression,!0):e.isExportSpecifier(t)?n.collectLinkedAliases(t.propertyName||t.name,!0):e.forEachChild(t,C)}function A(n,r,a,i,o,s){var c,p=280===a.kind?a:void 0,m=279===a.kind?a:void 0,v=p?p.sourceFiles:[m];if(D(s,a)&&(c=e.createSourceMapGenerator(t,e.getBaseFileName(e.normalizeSlashes(n)),L(s),k(s,n,m),s)),p?o.writeBundle(p,b,g,c):o.writeFile(m,g,c),c){d&&d.push({inputSourceFileNames:c.getSources(),sourceMap:c.toJSON()});var y=N(s,c,n,r,m);if(y&&(g.isAtStartOfLine()||g.rawWrite(_),g.writeComment("//# sourceMappingURL="+y)),r){var h=c.toString();e.writeFile(t,f,r,h,!1,v)}}else g.writeLine();e.writeFile(t,f,n,g.getText(),!!l.emitBOM,v),i&&(b.totalLength=g.getTextPos(),e.writeFile(t,f,i,JSON.stringify(b,void 0,2),!1)),g.clear(),b=u()}function D(n,t){return(n.sourceMap||n.inlineSourceMap)&&(279!==t.kind||!e.fileExtensionIs(t.fileName,".json"))}function L(n){var t=e.normalizeSlashes(n.sourceRoot||"");return t?e.ensureTrailingDirectorySeparator(t):t}function k(n,r,a){if(n.sourceRoot)return t.getCommonSourceDirectory();if(n.mapRoot){var i=e.normalizeSlashes(n.mapRoot);return a&&(i=e.getDirectoryPath(e.getSourceFilePathInNewDir(a.fileName,t,i))),0===e.getRootLength(i)&&(i=e.combinePaths(t.getCommonSourceDirectory(),i)),i}return e.getDirectoryPath(e.normalizePath(r))}function N(n,r,a,i,o){if(n.inlineSourceMap){var s=r.toString(),c=e.base64encode(e.sys,s);return"data:application/json;base64,"+c}var l=e.getBaseFileName(e.normalizeSlashes(e.Debug.assertDefined(i)));if(n.mapRoot){var u=e.normalizeSlashes(n.mapRoot);return o&&(u=e.getDirectoryPath(e.getSourceFilePathInNewDir(o.fileName,t,u))),0===e.getRootLength(u)?(u=e.combinePaths(t.getCommonSourceDirectory(),u),e.getRelativePathToDirectoryOrUrl(e.getDirectoryPath(e.normalizePath(a)),e.combinePaths(u,l),t.getCurrentDirectory(),t.getCanonicalFileName,!0)):e.combinePaths(u,l)}return l}}function m(n,t){void 0===n&&(n={}),void 0===t&&(t={});var r,a,o,s,c,l,u,d,p,m,f,v,y,h,b,E=t.hasGlobalName,T=t.onEmitNode,S=void 0===T?e.noEmitNotification:T,x=t.substituteNode,C=void 0===x?e.noEmitSubstitution:x,A=t.onBeforeEmitNodeArray,D=t.onAfterEmitNodeArray,L=t.onBeforeEmitToken,k=t.onAfterEmitToken,N=!!n.extendedDiagnostics,I=e.getNewLineCharacter(n),w=e.getEmitModuleKind(n),M=e.createMap(),O=aa,P=!0,R=-1,F=-1,G=-1,V=-1,B=!1,K=!!n.removeComments,j=e.performance.createTimerIf(N,"commentTime","beforeComment","afterComment"),U=j.enter,H=j.exit;return oe(),{printNode:W,printList:q,printFile:J,printBundle:z,writeNode:Y,writeList:Q,writeFile:ee,writeBundle:Z};function W(n,t,r){switch(n){case 0:e.Debug.assert(e.isSourceFile(t),"Expected a SourceFile node.");break;case 2:e.Debug.assert(e.isIdentifier(t),"Expected an Identifier node.");break;case 1:e.Debug.assert(e.isExpression(t),"Expected an Expression node.");break}switch(t.kind){case 279:return J(t);case 280:return z(t);case 281:return X(t)}return Y(n,t,r,ne()),te()}function q(e,n,t){return Q(e,n,t,ne()),te()}function z(e){return Z(e,void 0,ne(),void 0),te()}function J(e){return ee(e,ne(),void 0),te()}function X(e){return $(e,ne()),te()}function Y(e,n,t,r){var a=p;ie(r,void 0),re(e,n,t),oe(),p=a}function Q(e,n,t,r){var a=p;ie(r,void 0),t&&ae(t),Zr(i,n,e),oe(),p=a}function Z(e,n,t,r){f=!1;var a=p;ie(t,r),Pr(e),Or(e),ve(e),Dr(e);for(var i=0,o=e.prepends;i<o.length;i++){var s=o[i];fa(),re(4,s,void 0)}n&&(n.originalOffset=p.getTextPos());for(var c=0,l=e.sourceFiles;c<l.length;c++){var u=l[c];re(0,u,u)}oe(),p=a}function $(e,n){var t=p;ie(n,void 0),re(4,e,void 0),oe(),p=t}function ee(e,n,t){f=!0;var r=p;ie(n,t),Pr(e),Or(e),re(0,e,e),oe(),p=r}function ne(){return m||(m=e.createTextWriter(I))}function te(){var e=m.getText();return m.clear(),e}function re(e,n,t){t&&ae(t);var r=de(0,n);r(e,n)}function ae(e){r=e,h=void 0,b=void 0,e&&Li(e)}function ie(t,r){t&&n.omitTrailingSemicolon&&(t=e.getTrailingSemicolonOmittingWriter(t)),p=t,v=r,P=!p||!v}function oe(){a=[],o=[],s=e.createMap(),c=[],l=0,u=[],r=void 0,h=void 0,b=void 0,ie(void 0,void 0)}function se(){return h||(h=e.getLineStarts(r))}function ce(e){if(void 0!==e){var n=de(0,e);n(4,e)}}function le(e){if(void 0!==e){var n=de(0,e);n(2,e)}}function ue(e){if(void 0!==e){var n=de(0,e);n(1,e)}}function de(n,t){switch(n){case 0:if(S!==e.noEmitNotification)return me;case 1:if(C!==e.noEmitSubstitution)return ge;case 2:if(!K&&279!==t.kind)return ni;case 3:if(!P&&279!==t.kind&&!e.isInJsonFile(t))return Si;case 4:return fe;default:return e.Debug.assertNever(n)}}function pe(e,n){return de(e+1,n)}function me(e,n){var t=pe(0,n);S(e,n,t)}function fe(n,t){if(0===n)return Ar(e.cast(t,e.isSourceFile));if(2===n)return Ee(e.cast(t,e.isIdentifier));if(3===n)return _e(e.cast(t,e.isTypeParameterDeclaration));if(5===n)return e.Debug.assertNode(t,e.isEmptyStatement),zn(!0);if(4===n){if(e.isKeyword(t.kind))return ya(t,ca);switch(t.kind){case 15:case 16:case 17:return he(t);case 281:return be(t);case 72:return Ee(t);case 148:return Te(t);case 149:return xe(t);case 150:return Ce(t);case 151:return Ae(t);case 152:return De(t);case 153:return Le(t);case 154:return ke(t);case 155:return Ne(t);case 156:return Ie(t);case 157:return we(t);case 158:case 159:return Me(t);case 160:return Oe(t);case 161:return Pe(t);case 162:return Re(t);case 163:return Ge(t);case 164:return Ve(t);case 165:return Be(t);case 289:return Ke(t);case 166:return We(t);case 167:return qe(t);case 168:return ze(t);case 169:return Je(t);case 170:return Ye(t);case 171:return Qe(t);case 173:return Ze(t);case 174:return $e(t);case 175:return en(t);case 176:return nn(t);case 177:return tn(t);case 211:return Vn(t);case 178:return rn();case 179:return an(t);case 180:return on(t);case 181:return sn(t);case 182:return cn(t);case 183:return ln(t);case 284:return void oa("*");case 285:return void oa("?");case 286:return je(t);case 287:return Ue(t);case 288:return He(t);case 172:case 290:return Xe(t);case 184:return un(t);case 185:return dn(t);case 186:return pn(t);case 216:return Un(t);case 217:return Fe();case 218:return Hn(t);case 219:return qn(t);case 220:return zn(!1);case 221:return Jn(t);case 222:return Xn(t);case 223:return Qn(t);case 224:return Zn(t);case 225:return $n(t);case 226:return et(t);case 227:return nt(t);case 228:return rt(t);case 229:return at(t);case 230:return ot(t);case 231:return st(t);case 232:return ct(t);case 233:return lt(t);case 234:return ut(t);case 235:return dt(t);case 236:return pt(t);case 237:return mt(t);case 238:return ft(t);case 239:return _t(t);case 240:return xt(t);case 241:return At(t);case 242:return Dt(t);case 243:return Lt(t);case 244:return kt(t);case 245:return Nt(t);case 246:return It(t);case 247:return Kt(t);case 248:return wt(t);case 249:return Ot(t);case 250:return Pt(t);case 251:return Rt(t);case 252:return Ft(t);case 253:return Gt(t);case 254:return Vt(t);case 255:return Bt(t);case 256:return jt(t);case 257:return Ut(t);case 258:return;case 259:return qt(t);case 11:return Qt(t);case 262:case 265:return Yt(t);case 263:case 266:return Zt(t);case 267:return er(t);case 268:return $t(t);case 269:return nr(t);case 270:return tr(t);case 271:return ar(t);case 272:return ir(t);case 273:return sr(t);case 274:return cr(t);case 275:return lr(t);case 276:return ur(t);case 277:return dr(t);case 278:return pr(t);case 299:case 305:return Tr(t);case 300:case 302:case 301:case 298:return fr(t);case 295:return _r(t);case 303:return gr(t);case 304:return vr(t);case 297:return yr(t);case 293:return Er(t);case 292:return br(t);case 296:case 294:return hr(t);case 291:return mr(t)}if(e.isExpression(t))n=1,C!==e.noEmitSubstitution&&(t=C(n,t));else if(e.isToken(t))return ya(t,oa)}if(1===n)switch(t.kind){case 8:case 9:return ye(t);case 10:case 13:case 14:return he(t);case 72:return Ee(t);case 87:case 96:case 98:case 102:case 100:case 92:return void ya(t,ca);case 187:return mn(t);case 188:return fn(t);case 189:return _n(t);case 190:return vn(t);case 191:return yn(t);case 192:return hn(t);case 193:return bn(t);case 194:return En(t);case 195:return Tn(t);case 196:return Sn(t);case 197:return xn(t);case 198:return An(t);case 199:return Dn(t);case 200:return Ln(t);case 201:return kn(t);case 202:return Nn(t);case 203:return wn(t);case 204:return Mn(t);case 205:return On(t);case 206:return Pn(t);case 207:return Rn(t);case 208:return Fn(t);case 209:return Gn(t);case 210:return;case 212:return Bn(t);case 213:return Kn(t);case 214:return jn(t);case 260:return zt(t);case 261:return Jt(t);case 264:return Xt(t);case 308:return Ir(t);case 309:return wr(t)}}function _e(e){ce(e.name),pa(),ca("in"),pa(),ce(e.constraint)}function ge(e,n){var t=pe(1,n);t(e,C(e,n))}function ve(t){var a=!1,i=280===t.kind?t:void 0;if(!i||w!==e.ModuleKind.None){for(var o=i?i.sourceFiles.length:1,s=0;s<o;s++){var c=i?i.sourceFiles[s]:t,l=e.isSourceFile(c)?c:r,u=n.noEmitHelpers||void 0!==e.getExternalHelpersModuleName(l),d=e.isSourceFile(c)&&!f,p=e.getEmitHelpers(c);if(p)for(var m=0,_=e.stableSort(p,e.compareEmitHelpers);m<_.length;m++){var g=_[m];if(g.scoped){if(i)continue}else{if(u)continue;if(d){if(M.get(g.name))continue;M.set(g.name,!0)}}"string"===typeof g.text?Ea(g.text):Ea(g.text(qa)),a=!0}}return a}}function ye(e){he(e)}function he(t){var r=wa(t,n.neverAsciiEscape);!n.sourceMap&&!n.inlineSourceMap||10!==t.kind&&!e.isTemplateLiteralKind(t.kind)?ra(r):ta(r)}function be(e){p.rawWrite(e.text)}function Ee(e){var n=e.symbol?ia:O;n(Ia(e,!1),e.symbol),Zr(e,e.typeArguments,53776)}function Te(e){Se(e.left),oa("."),ce(e.right)}function Se(e){72===e.kind?ue(e):ce(e)}function xe(e){oa("["),ue(e.expression),oa("]")}function Ce(e){ce(e.name),e.constraint&&(pa(),ca("extends"),pa(),ce(e.constraint)),e.default&&(pa(),la("="),pa(),ce(e.default))}function Ae(e){Wr(e,e.decorators),Fr(e,e.modifiers),ce(e.dotDotDotToken),Rr(e.name,ua),ce(e.questionToken),e.parent&&289===e.parent.kind&&!e.name?ce(e.type):Gr(e.type),Vr(e.initializer,e.type?e.type.end:e.questionToken?e.questionToken.end:e.name?e.name.end:e.modifiers?e.modifiers.end:e.decorators?e.decorators.end:e.pos,e)}function De(e){oa("@"),ue(e.expression)}function Le(e){Wr(e,e.decorators),Fr(e,e.modifiers),Rr(e.name,ma),ce(e.questionToken),Gr(e.type),sa()}function ke(e){Wr(e,e.decorators),Fr(e,e.modifiers),ce(e.name),ce(e.questionToken),ce(e.exclamationToken),Gr(e.type),Vr(e.initializer,e.type?e.type.end:e.questionToken?e.questionToken.end:e.name.end,e),sa()}function Ne(e){Ma(e),Wr(e,e.decorators),Fr(e,e.modifiers),ce(e.name),ce(e.questionToken),zr(e,e.typeParameters),Jr(e,e.parameters),Gr(e.type),sa(),Oa(e)}function Ie(e){Wr(e,e.decorators),Fr(e,e.modifiers),ce(e.asteriskToken),ce(e.name),ce(e.questionToken),yt(e,ht)}function we(e){Fr(e,e.modifiers),ca("constructor"),yt(e,ht)}function Me(e){Wr(e,e.decorators),Fr(e,e.modifiers),ca(158===e.kind?"get":"set"),pa(),ce(e.name),yt(e,ht)}function Oe(e){Ma(e),Wr(e,e.decorators),Fr(e,e.modifiers),zr(e,e.typeParameters),Jr(e,e.parameters),Gr(e.type),sa(),Oa(e)}function Pe(e){Ma(e),Wr(e,e.decorators),Fr(e,e.modifiers),ca("new"),pa(),zr(e,e.typeParameters),Jr(e,e.parameters),Gr(e.type),sa(),Oa(e)}function Re(e){Wr(e,e.decorators),Fr(e,e.modifiers),Qr(e,e.parameters),Gr(e.type),sa()}function Fe(){sa()}function Ge(e){ce(e.parameterName),pa(),ca("is"),pa(),ce(e.type)}function Ve(e){ce(e.typeName),qr(e,e.typeArguments)}function Be(e){Ma(e),zr(e,e.typeParameters),Yr(e,e.parameters),pa(),oa("=>"),pa(),ce(e.type),Oa(e)}function Ke(e){ca("function"),Jr(e,e.parameters),oa(":"),ce(e.type)}function je(e){oa("?"),ce(e.type)}function Ue(e){oa("!"),ce(e.type)}function He(e){ce(e.type),oa("=")}function We(e){Ma(e),ca("new"),pa(),zr(e,e.typeParameters),Jr(e,e.parameters),pa(),oa("=>"),pa(),ce(e.type),Oa(e)}function qe(e){ca("typeof"),pa(),ce(e.exprName)}function ze(n){oa("{");var t=1&e.getEmitFlags(n)?768:32897;Zr(n,n.members,524288|t),oa("}")}function Je(e){ce(e.elementType),oa("["),oa("]")}function Xe(e){oa("..."),ce(e.type)}function Ye(e){oa("["),Zr(e,e.elementTypes,528),oa("]")}function Qe(e){ce(e.type),oa("?")}function Ze(e){Zr(e,e.types,516)}function $e(e){Zr(e,e.types,520)}function en(e){ce(e.checkType),pa(),ca("extends"),pa(),ce(e.extendsType),pa(),oa("?"),pa(),ce(e.trueType),pa(),oa(":"),pa(),ce(e.falseType)}function nn(e){ca("infer"),pa(),ce(e.typeParameter)}function tn(e){oa("("),ce(e.type),oa(")")}function rn(){ca("this")}function an(e){ha(e.operator,ca),pa(),ce(e.type)}function on(e){ce(e.objectType),oa("["),ce(e.indexType),oa("]")}function sn(n){var t=e.getEmitFlags(n);oa("{"),1&t?pa():(fa(),_a()),n.readonlyToken&&(ce(n.readonlyToken),133!==n.readonlyToken.kind&&ca("readonly"),pa()),oa("[");var r=de(0,n.typeParameter);r(3,n.typeParameter),oa("]"),n.questionToken&&(ce(n.questionToken),56!==n.questionToken.kind&&oa("?")),oa(":"),pa(),ce(n.type),sa(),1&t?pa():(fa(),ga()),oa("}")}function cn(e){ue(e.literal)}function ln(e){e.isTypeOf&&(ca("typeof"),pa()),ca("import"),oa("("),ce(e.argument),oa(")"),e.qualifier&&(oa("."),ce(e.qualifier)),qr(e,e.typeArguments)}function un(e){oa("{"),Zr(e,e.elements,525136),oa("}")}function dn(e){oa("["),Zr(e,e.elements,524880),oa("]")}function pn(e){ce(e.dotDotDotToken),e.propertyName&&(ce(e.propertyName),oa(":"),pa()),ce(e.name),Vr(e.initializer,e.name.end,e)}function mn(e){var n=e.elements,t=e.multiLine?65536:0;$r(e,n,8914|t)}function fn(n){e.forEach(n.properties,Fa);var t=65536&e.getEmitFlags(n);t&&_a();var a=n.multiLine?65536:0,i=r.languageVersion>=1&&!e.isJsonSourceFile(r)?64:0;Zr(n,n.properties,526226|i|a),t&&ga()}function _n(n){var t=!1,a=!1,i=e.skipTrivia(r.text,n.expression.end,!1,!0),o=e.skipTrivia(r.text,i),s=o+1;if(!(131072&e.getEmitFlags(n))){var c=e.createToken(24);c.pos=n.expression.end,c.end=s,t=La(n,n.expression,c),a=La(n,c,n.name)}ue(n.expression),Ta(t,!1);var l=i!==o,u=!t&&gn(n.expression,l);u&&oa("."),it(24,n.expression.end,oa,n),Ta(a,!1),ce(n.name),Sa(t,a)}function gn(t,r){if(t=e.skipPartiallyEmittedExpressions(t),e.isNumericLiteral(t)){var a=wa(t,!0);return!t.numericLiteralFlags&&!e.stringContains(a,e.tokenToString(24))&&(!r||n.removeComments)}if(e.isPropertyAccessExpression(t)||e.isElementAccessExpression(t)){var i=e.getConstantValue(t);return"number"===typeof i&&isFinite(i)&&Math.floor(i)===i&&n.removeComments}}function vn(e){ue(e.expression),it(22,e.expression.end,oa,e),ue(e.argumentExpression),it(23,e.argumentExpression.end,oa,e)}function yn(e){ue(e.expression),qr(e,e.typeArguments),$r(e,e.arguments,2576)}function hn(e){it(95,e.pos,ca,e),pa(),ue(e.expression),qr(e,e.typeArguments),$r(e,e.arguments,18960)}function bn(e){ue(e.tag),qr(e,e.typeArguments),pa(),ue(e.template)}function En(e){oa("<"),ce(e.type),oa(">"),ue(e.expression)}function Tn(e){var n=it(20,e.pos,oa,e);ue(e.expression),it(21,e.expression?e.expression.end:n,oa,e)}function Sn(e){Ga(e.name),gt(e)}function xn(e){Wr(e,e.decorators),Fr(e,e.modifiers),yt(e,Cn)}function Cn(e){zr(e,e.typeParameters),Yr(e,e.parameters),Gr(e.type),pa(),ce(e.equalsGreaterThanToken)}function An(e){it(81,e.pos,ca,e),pa(),ue(e.expression)}function Dn(e){it(104,e.pos,ca,e),pa(),ue(e.expression)}function Ln(e){it(106,e.pos,ca,e),pa(),ue(e.expression)}function kn(e){it(122,e.pos,ca,e),pa(),ue(e.expression)}function Nn(e){ha(e.operator,la),In(e)&&pa(),ue(e.operand)}function In(e){var n=e.operand;return 202===n.kind&&(38===e.operator&&(38===n.operator||44===n.operator)||39===e.operator&&(39===n.operator||45===n.operator))}function wn(e){ue(e.operand),ha(e.operator,la)}function Mn(e){var n=27!==e.operatorToken.kind,t=La(e,e.left,e.operatorToken),r=La(e,e.operatorToken,e.right);ue(e.left),Ta(t,n),di(e.operatorToken.pos),ya(e.operatorToken,93===e.operatorToken.kind?ca:la),fi(e.operatorToken.end,!0),Ta(r,!0),ue(e.right),Sa(t,r)}function On(e){var n=La(e,e.condition,e.questionToken),t=La(e,e.questionToken,e.whenTrue),r=La(e,e.whenTrue,e.colonToken),a=La(e,e.colonToken,e.whenFalse);ue(e.condition),Ta(n,!0),ce(e.questionToken),Ta(t,!0),ue(e.whenTrue),Sa(n,t),Ta(r,!0),ce(e.colonToken),Ta(a,!0),ue(e.whenFalse),Sa(r,a)}function Pn(e){ce(e.head),Zr(e,e.templateSpans,262144)}function Rn(e){it(117,e.pos,ca,e),ce(e.asteriskToken),jr(e.expression)}function Fn(e){it(25,e.pos,oa,e),ue(e.expression)}function Gn(e){Ga(e.name),Ct(e)}function Vn(e){ue(e.expression),qr(e,e.typeArguments)}function Bn(e){ue(e.expression),e.type&&(pa(),ca("as"),pa(),ce(e.type))}function Kn(e){ue(e.expression),la("!")}function jn(e){va(e.keywordToken,e.pos,oa),oa("."),ce(e.name)}function Un(e){ue(e.expression),ce(e.literal)}function Hn(e){Wn(e,!e.multiLine&&ka(e))}function Wn(n,t){it(18,n.pos,oa,n);var r=t||1&e.getEmitFlags(n)?768:129;Zr(n,n.statements,r),it(19,n.statements.end,oa,n,!!(1&r))}function qn(e){Fr(e,e.modifiers),ce(e.declarationList),sa()}function zn(e){e?oa(";"):sa()}function Jn(n){ue(n.expression),e.isJsonSourceFile(r)&&!e.nodeIsSynthesized(n.expression)||sa()}function Xn(e){var n=it(91,e.pos,ca,e);pa(),it(20,n,oa,e),ue(e.expression),it(21,e.expression.end,oa,e),Hr(e,e.thenStatement),e.elseStatement&&(ba(e),it(83,e.thenStatement.end,ca,e),222===e.elseStatement.kind?(pa(),ce(e.elseStatement)):Hr(e,e.elseStatement))}function Yn(e,n){var t=it(107,n,ca,e);pa(),it(20,t,oa,e),ue(e.expression),it(21,e.expression.end,oa,e)}function Qn(n){it(82,n.pos,ca,n),Hr(n,n.statement),e.isBlock(n.statement)?pa():ba(n),Yn(n,n.statement.end),oa(";")}function Zn(e){Yn(e,e.pos),Hr(e,e.statement)}function $n(e){var n=it(89,e.pos,ca,e);pa();var t=it(20,n,oa,e);tt(e.initializer),t=it(26,e.initializer?e.initializer.end:t,oa,e),jr(e.condition),t=it(26,e.condition?e.condition.end:t,oa,e),jr(e.incrementor),it(21,e.incrementor?e.incrementor.end:t,oa,e),Hr(e,e.statement)}function et(e){var n=it(89,e.pos,ca,e);pa(),it(20,n,oa,e),tt(e.initializer),pa(),it(93,e.initializer.end,ca,e),pa(),ue(e.expression),it(21,e.expression.end,oa,e),Hr(e,e.statement)}function nt(e){var n=it(89,e.pos,ca,e);pa(),Ur(e.awaitModifier),it(20,n,oa,e),tt(e.initializer),pa(),it(147,e.initializer.end,ca,e),pa(),ue(e.expression),it(21,e.expression.end,oa,e),Hr(e,e.statement)}function tt(e){void 0!==e&&(238===e.kind?ce(e):ue(e))}function rt(e){it(78,e.pos,ca,e),Kr(e.label),sa()}function at(e){it(73,e.pos,ca,e),Kr(e.label),sa()}function it(n,t,a,i,o){var s=e.getParseTreeNode(i),c=s&&s.kind===i.kind,l=t;if(c&&(t=e.skipTrivia(r.text,t)),di&&c&&i.pos!==l){var u=o&&!e.positionsAreOnSameLine(l,t,r);u&&_a(),di(l),u&&ga()}return t=ha(n,a,t),fi&&c&&i.end!==t&&fi(t,!0),t}function ot(e){it(97,e.pos,ca,e),jr(e.expression),sa()}function st(e){var n=it(108,e.pos,ca,e);pa(),it(20,n,oa,e),ue(e.expression),it(21,e.expression.end,oa,e),Hr(e,e.statement)}function ct(e){var n=it(99,e.pos,ca,e);pa(),it(20,n,oa,e),ue(e.expression),it(21,e.expression.end,oa,e),pa(),ce(e.caseBlock)}function lt(e){ce(e.label),it(57,e.label.end,oa,e),pa(),ce(e.statement)}function ut(e){it(101,e.pos,ca,e),jr(e.expression),sa()}function dt(e){it(103,e.pos,ca,e),pa(),ce(e.tryBlock),e.catchClause&&(ba(e),ce(e.catchClause)),e.finallyBlock&&(ba(e),it(88,(e.catchClause||e.tryBlock).end,ca,e),pa(),ce(e.finallyBlock))}function pt(e){va(79,e.pos,ca),sa()}function mt(e){ce(e.name),Gr(e.type),Vr(e.initializer,e.type?e.type.end:e.name.end,e)}function ft(n){ca(e.isLet(n)?"let":e.isVarConst(n)?"const":"var"),pa(),Zr(n,n.declarations,528)}function _t(e){gt(e)}function gt(e){Wr(e,e.decorators),Fr(e,e.modifiers),ca("function"),ce(e.asteriskToken),pa(),le(e.name),yt(e,ht)}function vt(e,n){Et(n)}function yt(n,t){var r=n.body;if(r)if(e.isBlock(r)){var a=65536&e.getEmitFlags(n);a&&_a(),Ma(n),e.forEach(n.parameters,Ra),Ra(n.body),t(n),S?S(4,r,vt):Et(r),Oa(n),a&&ga()}else t(n),pa(),ue(r);else t(n),sa()}function ht(e){zr(e,e.typeParameters),Jr(e,e.parameters),Gr(e.type)}function bt(n){if(1&e.getEmitFlags(n))return!0;if(n.multiLine)return!1;if(!e.nodeIsSynthesized(n)&&!e.rangeIsOnSingleLine(n,r))return!1;if(xa(n,n.statements,2)||Aa(n,n.statements,2))return!1;for(var t,a=0,i=n.statements;a<i.length;a++){var o=i[a];if(Ca(t,o,2))return!1;t=o}return!0}function Et(e){pa(),oa("{"),_a();var n=bt(e)?Tt:St;oi?oi(e,e.statements,n):n(e),ga(),va(19,e.statements.end,oa,e)}function Tt(e){St(e,!0)}function St(e,n){var t=Mr(e.statements,!0),r=p.getTextPos();ve(e),0===t&&r===p.getTextPos()&&n?(ga(),Zr(e,e.statements,768),_a()):Zr(e,e.statements,1,t)}function xt(e){Ct(e)}function Ct(n){e.forEach(n.members,Fa),Wr(n,n.decorators),Fr(n,n.modifiers),ca("class"),n.name&&(pa(),le(n.name));var t=65536&e.getEmitFlags(n);t&&_a(),zr(n,n.typeParameters),Zr(n,n.heritageClauses,0),pa(),oa("{"),Zr(n,n.members,129),oa("}"),t&&ga()}function At(e){Wr(e,e.decorators),Fr(e,e.modifiers),ca("interface"),pa(),ce(e.name),zr(e,e.typeParameters),Zr(e,e.heritageClauses,512),pa(),oa("{"),Zr(e,e.members,129),oa("}")}function Dt(e){Wr(e,e.decorators),Fr(e,e.modifiers),ca("type"),pa(),ce(e.name),zr(e,e.typeParameters),pa(),oa("="),pa(),ce(e.type),sa()}function Lt(e){Fr(e,e.modifiers),ca("enum"),pa(),ce(e.name),pa(),oa("{"),Zr(e,e.members,145),oa("}")}function kt(e){Fr(e,e.modifiers),512&~e.flags&&(ca(16&e.flags?"namespace":"module"),pa()),ce(e.name);var n=e.body;if(!n)return sa();while(244===n.kind)oa("."),ce(n.name),n=n.body;pa(),ce(n)}function Nt(n){Ma(n),e.forEach(n.statements,Ra),Wn(n,ka(n)),Oa(n)}function It(e){it(18,e.pos,oa,e),Zr(e,e.clauses,129),it(19,e.clauses.end,oa,e,!0)}function wt(e){Fr(e,e.modifiers),it(92,e.modifiers?e.modifiers.end:e.pos,ca,e),pa(),ce(e.name),pa(),it(59,e.name.end,oa,e),pa(),Mt(e.moduleReference),sa()}function Mt(e){72===e.kind?ue(e):ce(e)}function Ot(e){Fr(e,e.modifiers),it(92,e.modifiers?e.modifiers.end:e.pos,ca,e),pa(),e.importClause&&(ce(e.importClause),pa(),it(144,e.importClause.end,ca,e),pa()),ue(e.moduleSpecifier),sa()}function Pt(e){ce(e.name),e.name&&e.namedBindings&&(it(27,e.name.end,oa,e),pa()),ce(e.namedBindings)}function Rt(e){var n=it(40,e.pos,oa,e);pa(),it(119,n,ca,e),pa(),ce(e.name)}function Ft(e){Ht(e)}function Gt(e){Wt(e)}function Vt(e){var n=it(85,e.pos,ca,e);pa(),e.isExportEquals?it(59,n,la,e):it(80,n,ca,e),pa(),ue(e.expression),sa()}function Bt(e){var n=it(85,e.pos,ca,e);if(pa(),e.exportClause?ce(e.exportClause):n=it(40,n,oa,e),e.moduleSpecifier){pa();var t=e.exportClause?e.exportClause.end:n;it(144,t,ca,e),pa(),ue(e.moduleSpecifier)}sa()}function Kt(e){var n=it(85,e.pos,ca,e);pa(),n=it(119,n,ca,e),pa(),n=it(131,n,ca,e),pa(),ce(e.name),sa()}function jt(e){Ht(e)}function Ut(e){Wt(e)}function Ht(e){oa("{"),Zr(e,e.elements,525136),oa("}")}function Wt(e){e.propertyName&&(ce(e.propertyName),pa(),it(119,e.propertyName.end,ca,e),pa()),ce(e.name)}function qt(e){ca("require"),oa("("),ue(e.expression),oa(")")}function zt(e){ce(e.openingElement),Zr(e,e.children,262144),ce(e.closingElement)}function Jt(e){oa("<"),rr(e.tagName),qr(e,e.typeArguments),pa(),ce(e.attributes),oa("/>")}function Xt(e){ce(e.openingFragment),Zr(e,e.children,262144),ce(e.closingFragment)}function Yt(n){oa("<"),e.isJsxOpeningElement(n)&&(rr(n.tagName),qr(n,n.typeArguments),n.attributes.properties&&n.attributes.properties.length>0&&pa(),ce(n.attributes)),oa(">")}function Qt(e){p.writeLiteral(Ia(e,!0))}function Zt(n){oa("</"),e.isJsxClosingElement(n)&&rr(n.tagName),oa(">")}function $t(e){Zr(e,e.properties,262656)}function er(e){ce(e.name),Br("=",oa,e.initializer,ce)}function nr(e){oa("{..."),ue(e.expression),oa("}")}function tr(e){e.expression&&(oa("{"),ce(e.dotDotDotToken),ue(e.expression),oa("}"))}function rr(e){72===e.kind?ue(e):ce(e)}function ar(e){it(74,e.pos,ca,e),pa(),ue(e.expression),or(e,e.statements,e.expression.end)}function ir(e){var n=it(80,e.pos,ca,e);or(e,e.statements,n)}function or(n,t,a){var i=1===t.length&&(e.nodeIsSynthesized(n)||e.nodeIsSynthesized(t[0])||e.rangeStartPositionsAreOnSameLine(n,t[0],r)),o=163969;i?(va(57,a,oa,n),pa(),o&=-130):it(57,a,oa,n),Zr(n,t,o)}function sr(e){pa(),ha(e.token,ca),pa(),Zr(e,e.types,528)}function cr(e){var n=it(75,e.pos,ca,e);pa(),e.variableDeclaration&&(it(20,n,oa,e),ce(e.variableDeclaration),it(21,e.variableDeclaration.end,oa,e),pa()),ce(e.block)}function lr(n){ce(n.name),oa(":"),pa();var t=n.initializer;if(fi&&0===(512&e.getEmitFlags(t))){var r=e.getCommentRange(t);fi(r.pos)}ue(t)}function ur(e){ce(e.name),e.objectAssignmentInitializer&&(pa(),oa("="),pa(),ue(e.objectAssignmentInitializer))}function dr(e){e.expression&&(it(25,e.pos,oa,e),ue(e.expression))}function pr(e){ce(e.name),Vr(e.initializer,e.name.end,e)}function mr(e){if(O("/**"),e.comment)for(var n=e.comment.split(/\r\n?|\n/g),t=0,r=n;t<r.length;t++){var a=r[t];fa(),pa(),oa("*"),pa(),O(a)}e.tags&&(1!==e.tags.length||302!==e.tags[0].kind||e.comment?Zr(e,e.tags,33):(pa(),ce(e.tags[0]))),pa(),O("*/")}function fr(e){Sr(e.tagName),Cr(e.typeExpression),xr(e.comment)}function _r(e){Sr(e.tagName),pa(),oa("{"),ce(e.class),oa("}"),xr(e.comment)}function gr(e){Sr(e.tagName),Cr(e.constraint),pa(),Zr(e,e.typeParameters,528),xr(e.comment)}function vr(e){Sr(e.tagName),e.typeExpression&&(283===e.typeExpression.kind?Cr(e.typeExpression):(pa(),oa("{"),O("Object"),e.typeExpression.isArrayType&&(oa("["),oa("]")),oa("}"))),e.fullName&&(pa(),ce(e.fullName)),xr(e.comment),e.typeExpression&&292===e.typeExpression.kind&&br(e.typeExpression)}function yr(e){Sr(e.tagName),e.name&&(pa(),ce(e.name)),xr(e.comment),Er(e.typeExpression)}function hr(e){Sr(e.tagName),xr(e.comment)}function br(n){Zr(n,e.createNodeArray(n.jsDocPropertyTags),33)}function Er(n){n.typeParameters&&Zr(n,e.createNodeArray(n.typeParameters),33),n.parameters&&Zr(n,e.createNodeArray(n.parameters),33),n.type&&(fa(),pa(),oa("*"),pa(),ce(n.type))}function Tr(e){Sr(e.tagName),Cr(e.typeExpression),pa(),e.isBracketed&&oa("["),ce(e.name),e.isBracketed&&oa("]"),xr(e.comment)}function Sr(e){oa("@"),ce(e)}function xr(e){e&&(pa(),O(e))}function Cr(e){e&&(pa(),oa("{"),ce(e.type),oa("}"))}function Ar(n){fa();var t=n.statements;if(oi){var r=0===t.length||!e.isPrologueDirective(t[0])||e.nodeIsSynthesized(t[0]);if(r)return void oi(n,t,Nr)}Nr(n)}function Dr(e){kr(!!e.hasNoDefaultLib,e.syntheticFileReferences||[],e.syntheticTypeReferences||[],e.syntheticLibReferences||[])}function Lr(e){e.isDeclarationFile&&kr(e.hasNoDefaultLib,e.referencedFiles,e.typeReferenceDirectives,e.libReferenceDirectives)}function kr(e,n,t,a){if(e&&(da('/// <reference no-default-lib="true"/>'),fa()),r&&r.moduleName&&(da('/// <amd-module name="'+r.moduleName+'" />'),fa()),r&&r.amdDependencies)for(var i=0,o=r.amdDependencies;i<o.length;i++){var s=o[i];s.name?da('/// <amd-dependency name="'+s.name+'" path="'+s.path+'" />'):da('/// <amd-dependency path="'+s.path+'" />'),fa()}for(var c=0,l=n;c<l.length;c++){var u=l[c];da('/// <reference path="'+u.fileName+'" />'),fa()}for(var d=0,p=t;d<p.length;d++){u=p[d];da('/// <reference types="'+u.fileName+'" />'),fa()}for(var m=0,f=a;m<f.length;m++){u=f[m];da('/// <reference lib="'+u.fileName+'" />'),fa()}}function Nr(n){var t=n.statements;Ma(n),e.forEach(n.statements,Ra),ve(n);var r=e.findIndex(t,function(n){return!e.isPrologueDirective(n)});Lr(n),Zr(n,t,1,-1===r?t.length:r),Oa(n)}function Ir(e){ue(e.expression)}function wr(e){$r(e,e.elements,528)}function Mr(n,t,r){for(var a=0;a<n.length;a++){var i=n[a];if(!e.isPrologueDirective(i))return a;var o=!r||!r.has(i.expression.text);o&&((t||a>0)&&fa(),ce(i),r&&r.set(i.expression.text,!0))}return n.length}function Or(n){if(e.isSourceFile(n))ae(n),Mr(n.statements);else{for(var t=e.createMap(),r=0,a=n.sourceFiles;r<a.length;r++){var i=a[r];ae(i),Mr(i.statements,!0,t)}ae(void 0)}}function Pr(n){if(e.isSourceFile(n)){var t=e.getShebang(n.text);if(t)return da(t),fa(),!0}else for(var r=0,a=n.sourceFiles;r<a.length;r++){var i=a[r];if(Pr(i))break}}function Rr(e,n){if(e){var t=O;O=n,ce(e),O=t}}function Fr(e,n){n&&n.length&&(Zr(e,n,262656),pa())}function Gr(e){e&&(oa(":"),pa(),ce(e))}function Vr(e,n,t){e&&(pa(),it(59,n,la,t),pa(),ue(e))}function Br(e,n,t,r){t&&(n(e),r(t))}function Kr(e){e&&(pa(),ce(e))}function jr(e){e&&(pa(),ue(e))}function Ur(e){e&&(ce(e),pa())}function Hr(n,t){if(e.isBlock(t)||1&e.getEmitFlags(n))pa(),ce(t);else{if(fa(),_a(),e.isEmptyStatement(t)){var r=de(0,t);r(5,t)}else ce(t);ga()}}function Wr(e,n){Zr(e,n,49153)}function qr(e,n){Zr(e,n,53776)}function zr(n,t){if(e.isFunctionLike(n)&&n.typeArguments)return qr(n,n.typeArguments);Zr(n,t,53776)}function Jr(e,n){Zr(e,n,2576)}function Xr(n,t){var r=e.singleOrUndefined(t);return r&&r.pos===n.pos&&e.isArrowFunction(n)&&!n.type&&!e.some(n.decorators)&&!e.some(n.modifiers)&&!e.some(n.typeParameters)&&!e.some(r.decorators)&&!e.some(r.modifiers)&&!r.dotDotDotToken&&!r.questionToken&&!r.type&&!r.initializer&&e.isIdentifier(r.name)}function Yr(e,n){Xr(e,n)?Zr(e,n,528):Jr(e,n)}function Qr(e,n){Zr(e,n,8848)}function Zr(e,n,t,r,a){na(ce,e,n,t,r,a)}function $r(e,n,t,r,a){na(ue,e,n,t,r,a)}function ea(e){switch(60&e){case 0:break;case 16:oa(",");break;case 4:pa(),oa("|");break;case 32:pa(),oa("*"),pa();break;case 8:pa(),oa("&");break}}function na(n,t,r,a,i,o){void 0===i&&(i=0),void 0===o&&(o=r?r.length-i:0);var s=void 0===r;if(!(s&&16384&a)){var c=void 0===r||i>=r.length||0===o;if(c&&32768&a)return A&&A(r),void(D&&D(r));if(15360&a&&(oa(_(a)),c&&!s&&fi(r.pos,!0)),A&&A(r),c)1&a?fa():256&a&&!(524288&a)&&pa();else{var l=0===(262144&a),u=l;xa(t,r,a)?(fa(),u=!1):256&a&&pa(),128&a&&_a();for(var d=void 0,p=!1,m=0;m<o;m++){var f=r[i+m];if(32&a?(fa(),ea(a)):d&&(60&a&&d.end!==t.end&&di(d.end),ea(a),Ca(d,f,a)?(0===(131&a)&&(_a(),p=!0),fa(),u=!1):d&&512&a&&pa()),u){if(fi){var v=e.getCommentRange(f);fi(v.pos)}}else u=l;n(f),p&&(ga(),p=!1),d=f}var y=64&a&&r.hasTrailingComma;16&a&&y&&oa(","),d&&60&a&&d.end!==t.end&&!(1024&e.getEmitFlags(d))&&di(d.end),128&a&&ga(),Aa(t,r,a)?fa():256&a&&pa()}D&&D(r),15360&a&&(c&&!s&&di(r.end),oa(g(a)))}}function ta(e){p.writeLiteral(e)}function ra(e){p.writeStringLiteral(e)}function aa(e){p.write(e)}function ia(e,n){p.writeSymbol(e,n)}function oa(e){p.writePunctuation(e)}function sa(){p.writeTrailingSemicolon(";")}function ca(e){p.writeKeyword(e)}function la(e){p.writeOperator(e)}function ua(e){p.writeParameter(e)}function da(e){p.writeComment(e)}function pa(){p.writeSpace(" ")}function ma(e){p.writeProperty(e)}function fa(){p.writeLine()}function _a(){p.increaseIndent()}function ga(){p.decreaseIndent()}function va(e,n,t,r){return P?ha(e,t,n):Di(r,e,t,n,ha)}function ya(n,t){L&&L(n),t(e.tokenToString(n.kind)),k&&k(n)}function ha(n,t,r){var a=e.tokenToString(n);return t(a),r<0?r:r+a.length}function ba(n){1&e.getEmitFlags(n)?pa():fa()}function Ea(n){for(var t=n.split(/\r\n?|\n/g),r=e.guessIndentation(t),a=0,i=t;a<i.length;a++){var o=i[a],s=r?o.slice(r):o;s.length&&(fa(),O(s),p.rawWrite(I))}}function Ta(e,n){e?(_a(),fa()):n&&pa()}function Sa(e,n){e&&ga(),n&&ga()}function xa(n,t,a){if(1&a)return!0;if(2&a){if(65536&a)return!0;var i=t[0];return void 0===i?!e.rangeIsOnSingleLine(n,r):e.positionIsSynthesized(n.pos)||e.nodeIsSynthesized(i)?Da(i,a):!e.rangeStartPositionsAreOnSameLine(n,i,r)}return!1}function Ca(n,t,a){return!!(1&a)||(2&a?void 0!==n&&void 0!==t&&(e.nodeIsSynthesized(n)||e.nodeIsSynthesized(t)?Da(n,a)||Da(t,a):!e.rangeEndIsOnSameLineAsRangeStart(n,t,r)):e.getStartsOnNewLine(t))}function Aa(n,t,a){if(1&a)return 0===(131072&a);if(2&a){if(65536&a)return!0;var i=e.lastOrUndefined(t);return void 0===i?!e.rangeIsOnSingleLine(n,r):e.positionIsSynthesized(n.pos)||e.nodeIsSynthesized(i)?Da(i,a):!e.rangeEndPositionsAreOnSameLine(n,i,r)}return!1}function Da(n,t){if(e.nodeIsSynthesized(n)){var r=e.getStartsOnNewLine(n);return void 0===r?0!==(65536&t):r}return 0!==(65536&t)}function La(n,t,a){return n=Na(n),t=Na(t),a=Na(a),!!e.getStartsOnNewLine(a)||!e.nodeIsSynthesized(n)&&!e.nodeIsSynthesized(t)&&!e.nodeIsSynthesized(a)&&!e.rangeEndIsOnSameLineAsRangeStart(t,a,r)}function ka(n){return 0===n.statements.length&&e.rangeEndIsOnSameLineAsRangeStart(n,n,r)}function Na(n){while(195===n.kind&&e.nodeIsSynthesized(n))n=n.expression;return n}function Ia(n,t){return e.isGeneratedIdentifier(n)?Va(n):e.isIdentifier(n)&&(e.nodeIsSynthesized(n)||!n.parent||!r||n.parent&&r&&e.getSourceFileOfNode(n)!==e.getOriginalNode(r))?e.idText(n):10===n.kind&&n.textSourceNode?Ia(n.textSourceNode,t):!e.isLiteralExpression(n)||!e.nodeIsSynthesized(n)&&n.parent?e.getSourceTextOfNodeFromSourceFile(r,n,t):n.text}function wa(n,t){if(10===n.kind&&n.textSourceNode){var a=n.textSourceNode;return e.isIdentifier(a)?t||16777216&e.getEmitFlags(n)?'"'+e.escapeString(Ia(a))+'"':'"'+e.escapeNonAsciiString(Ia(a))+'"':wa(a,t)}return e.getLiteralText(n,r,t)}function Ma(n){n&&524288&e.getEmitFlags(n)||(c.push(l),l=0,u.push(d))}function Oa(n){n&&524288&e.getEmitFlags(n)||(l=c.pop(),d=u.pop())}function Pa(n){d&&d!==e.lastOrUndefined(u)||(d=e.createMap()),d.set(n,!0)}function Ra(n){if(n)switch(n.kind){case 218:e.forEach(n.statements,Ra);break;case 233:case 231:case 223:case 224:Ra(n.statement);break;case 222:Ra(n.thenStatement),Ra(n.elseStatement);break;case 225:case 227:case 226:Ra(n.initializer),Ra(n.statement);break;case 232:Ra(n.caseBlock);break;case 246:e.forEach(n.clauses,Ra);break;case 271:case 272:e.forEach(n.statements,Ra);break;case 235:Ra(n.tryBlock),Ra(n.catchClause),Ra(n.finallyBlock);break;case 274:Ra(n.variableDeclaration),Ra(n.block);break;case 219:Ra(n.declarationList);break;case 238:e.forEach(n.declarations,Ra);break;case 237:case 151:case 186:case 240:Ga(n.name);break;case 239:Ga(n.name),524288&e.getEmitFlags(n)&&(e.forEach(n.parameters,Ra),Ra(n.body));break;case 184:case 185:e.forEach(n.elements,Ra);break;case 249:Ra(n.importClause);break;case 250:Ga(n.name),Ra(n.namedBindings);break;case 251:Ga(n.name);break;case 252:e.forEach(n.elements,Ra);break;case 253:Ga(n.propertyName||n.name);break}}function Fa(e){if(e)switch(e.kind){case 275:case 276:case 154:case 156:case 158:case 159:Ga(e.name);break}}function Ga(n){n&&(e.isGeneratedIdentifier(n)?Va(n):e.isBindingPattern(n)&&Ra(n))}function Va(e){if(4===(7&e.autoGenerateFlags))return Ba(ei(e),e.autoGenerateFlags);var n=e.autoGenerateId;return o[n]||(o[n]=$a(e))}function Ba(n,t){var r=e.getNodeId(n);return a[r]||(a[r]=Za(n,t))}function Ka(e){return ja(e)&&!s.has(e)&&!(d&&d.has(e))}function ja(n){return!r||e.isFileLevelUniqueName(r,n,E)}function Ua(n,t){for(var r=t;e.isNodeDescendantOf(r,t);r=r.nextContainer)if(r.locals){var a=r.locals.get(e.escapeLeadingUnderscores(n));if(a&&70366143&a.flags)return!1}return!0}function Ha(e,n){if(e&&!(l&e)){var t=268435456===e?"_i":"_n";if(Ka(t))return l|=e,n&&Pa(t),t}while(1){var r=268435455&l;if(l++,8!==r&&13!==r){t=r<26?"_"+String.fromCharCode(97+r):"_"+(r-26);if(Ka(t))return n&&Pa(t),t}}}function Wa(e,n,t,r){if(void 0===n&&(n=Ka),t&&n(e))return r?Pa(e):s.set(e,!0),e;95!==e.charCodeAt(e.length-1)&&(e+="_");var a=1;while(1){var i=e+a;if(n(i))return r?Pa(i):s.set(i,!0),i;a++}}function qa(e){return Wa(e,ja,!0)}function za(e){var n=Ia(e.name);return Ua(n,e)?n:Wa(n)}function Ja(n){var t=e.getExternalModuleName(n),r=e.isStringLiteral(t)?e.makeIdentifierFromModuleName(t.text):"module";return Wa(r)}function Xa(){return Wa("default")}function Ya(){return Wa("class")}function Qa(n){return e.isIdentifier(n.name)?Ba(n.name):Ha(0)}function Za(e,n){switch(e.kind){case 72:return Wa(Ia(e),Ka,!!(16&n),!!(8&n));case 244:case 243:return za(e);case 249:case 255:return Ja(e);case 239:case 240:case 254:return Xa();case 209:return Ya();case 156:case 158:case 159:return Qa(e);default:return Ha(0)}}function $a(n){switch(7&n.autoGenerateFlags){case 1:return Ha(0,!!(8&n.autoGenerateFlags));case 2:return Ha(268435456,!!(8&n.autoGenerateFlags));case 3:return Wa(e.idText(n),32&n.autoGenerateFlags?ja:Ka,!!(16&n.autoGenerateFlags),!!(8&n.autoGenerateFlags))}return e.Debug.fail("Unsupported GeneratedIdentifierKind.")}function ei(n){var t=n.autoGenerateId,r=n,a=r.original;while(a){if(r=a,e.isIdentifier(r)&&4&r.autoGenerateFlags&&r.autoGenerateId!==t)break;a=r.original}return r}function ni(n,t){U(),B=!1;var r=e.getEmitFlags(t),a=e.getCommentRange(t),i=a.pos,o=a.end,s=307!==t.kind,c=i<0||0!==(512&r)||11===t.kind,l=o<0||0!==(1024&r)||11===t.kind,u=F,d=G,p=V;(i>0||o>0)&&i!==o&&(c||si(i,s),(!c||i>=0&&0!==(512&r))&&(F=i),(!l||o>=0&&0!==(1024&r))&&(G=o,238===t.kind&&(V=o))),e.forEach(e.getSyntheticLeadingComments(t),ti),H();var m=pe(2,t);2048&r?(K=!0,m(n,t),K=!1):m(n,t),U(),e.forEach(e.getSyntheticTrailingComments(t),ri),(i>0||o>0)&&i!==o&&(F=u,G=d,V=p,!l&&s&&pi(o)),H()}function ti(e){2===e.kind&&p.writeLine(),ai(e),e.hasTrailingNewLine||2===e.kind?p.writeLine():p.writeSpace(" ")}function ri(e){p.isAtStartOfLine()||p.writeSpace(" "),ai(e),e.hasTrailingNewLine&&p.writeLine()}function ai(n){var t=ii(n),r=3===n.kind?e.computeLineStarts(t):void 0;e.writeCommentRange(t,r,p,0,t.length,I)}function ii(e){return 3===e.kind?"/*"+e.text+"*/":"//"+e.text}function oi(n,t,r){U();var a=t.pos,i=t.end,o=e.getEmitFlags(n),s=a<0||0!==(512&o),c=K||i<0||0!==(1024&o);s||bi(t),H(),2048&o&&!K?(K=!0,r(n),K=!1):r(n),U(),c||(si(t.end,!0),B&&!p.isAtStartOfLine()&&p.writeLine()),H()}function si(e,n){B=!1,n?gi(e,ui):0===e&&gi(e,ci)}function ci(e,n,t,r,a){Ti(e,n)&&ui(e,n,t,r,a)}function li(t,r){return!n.onlyPrintJsDocStyle||(e.isJSDocLikeText(t,r)||e.isPinnedComment(t,r))}function ui(n,t,a,i,o){li(r.text,n)&&(B||(e.emitNewLineBeforeLeadingCommentOfPosition(se(),p,o,n),B=!0),Ci(n),e.writeCommentRange(r.text,se(),p,n,t,I),Ci(t),i?p.writeLine():3===a&&p.writeSpace(" "))}function di(e){K||-1===e||si(e,!0)}function pi(e){vi(e,mi)}function mi(n,t,a,i){li(r.text,n)&&(p.isAtStartOfLine()||p.writeSpace(" "),Ci(n),e.writeCommentRange(r.text,se(),p,n,t,I),Ci(t),i&&p.writeLine())}function fi(e,n){K||(U(),vi(e,n?mi:_i),H())}function _i(n,t,a,i){Ci(n),e.writeCommentRange(r.text,se(),p,n,t,I),Ci(t),i?p.writeLine():p.writeSpace(" ")}function gi(n,t){!r||-1!==F&&n===F||(yi(n)?hi(t):e.forEachLeadingCommentRange(r.text,n,t,n))}function vi(n,t){r&&(-1===G||n!==G&&n!==V)&&e.forEachTrailingCommentRange(r.text,n,t)}function yi(n){return void 0!==b&&e.last(b).nodePos===n}function hi(n){var t=e.last(b).detachedCommentEndPos;b.length-1?b.pop():b=void 0,e.forEachLeadingCommentRange(r.text,t,n,t)}function bi(n){var t=e.emitDetachedComments(r.text,se(),p,Ei,n,I,K);t&&(b?b.push(t):b=[t])}function Ei(n,t,a,i,o,s){li(r.text,i)&&(Ci(i),e.writeCommentRange(n,t,a,i,o,s),Ci(o))}function Ti(n,t){return e.isRecognizedTripleSlashComment(r.text,n,t)}function Si(n,t){var r=pe(3,t);if(e.isUnparsedSource(t)&&void 0!==t.sourceMapText){var a=e.tryParseRawSourceMap(t.sourceMapText);a&&v.appendSourceMap(p.getLine(),p.getColumn(),a,t.sourceMapPath),r(n,t)}else{var i=e.getSourceMapRange(t),o=i.pos,s=i.end,c=i.source,l=void 0===c?y:c,u=e.getEmitFlags(t);307!==t.kind&&0===(16&u)&&o>=0&&Ai(l,xi(l,o)),64&u?(P=!0,r(n,t),P=!1):r(n,t),307!==t.kind&&0===(32&u)&&s>=0&&Ai(l,s)}}function xi(n,t){return n.skipTrivia?n.skipTrivia(t):e.skipTrivia(n.text,t)}function Ci(n){if(!(P||e.positionIsSynthesized(n)||ki(y))){var t=e.getLineAndCharacterOfPosition(y,n),r=t.line,a=t.character;v.addMapping(p.getLine(),p.getColumn(),R,r,a,void 0)}}function Ai(e,n){if(e!==y){var t=y;Li(e),Ci(n),Li(t)}else Ci(n)}function Di(n,t,r,a,i){if(P||n&&e.isInJsonFile(n))return i(t,r,a);var o=n&&n.emitNode,s=o&&o.flags||0,c=o&&o.tokenSourceMapRanges&&o.tokenSourceMapRanges[t],l=c&&c.source||y;return a=xi(l,c?c.pos:a),0===(128&s)&&a>=0&&Ai(l,a),a=i(t,r,a),c&&(a=c.end),0===(256&s)&&a>=0&&Ai(l,a),a}function Li(e){P||(y=e,ki(e)||(R=v.addSource(e.fileName),n.inlineSources&&v.setSourceContent(R,e.text)))}function ki(n){return e.fileExtensionIs(n.fileName,".json")}}function f(){var e=[];return e[1024]=["{","}"],e[2048]=["(",")"],e[4096]=["<",">"],e[8192]=["[","]"],e}function _(e){return a[15360&e][0]}function g(e){return a[15360&e][1]}e.forEachEmittedFile=o,e.getOutputPathsForBundle=s,e.getOutputPathsFor=c,e.getOutputExtension=d,e.emitFiles=p,function(e){e[e["Notification"]=0]="Notification",e[e["Substitution"]=1]="Substitution",e[e["Comments"]=2]="Comments",e[e["SourceMaps"]=3]="SourceMaps",e[e["Emit"]=4]="Emit"}(n||(n={})),e.createPrinter=m,function(e){e[e["Auto"]=0]="Auto",e[e["CountMask"]=268435455]="CountMask",e[e["_i"]=268435456]="_i"}(t||(t={}))}(u||(u={})),function(e){function n(n,t,r){if(n.getDirectories&&n.readDirectory){var a=e.createMap(),i=e.createGetCanonicalFileName(r);return{useCaseSensitiveFileNames:r,fileExists:g,readFile:function(e,t){return n.readFile(e,t)},directoryExists:n.directoryExists&&v,getDirectories:h,readDirectory:b,createDirectory:n.createDirectory&&y,writeFile:n.writeFile&&_,addOrDeleteFileOrDirectory:E,addOrDeleteFile:T,clearCache:x}}function o(n){return e.toPath(n,t,i)}function s(n){return a.get(e.ensureTrailingDirectorySeparator(n))}function c(n){return s(e.getDirectoryPath(n))}function l(n){return e.getBaseFileName(e.normalizePath(n))}function u(t,r){var i={files:e.map(n.readDirectory(t,void 0,void 0,["*.*"]),l)||[],directories:n.getDirectories(t)||[]};return a.set(e.ensureTrailingDirectorySeparator(r),i),i}function d(n,t){t=e.ensureTrailingDirectorySeparator(t);var r=s(t);if(r)return r;try{return u(n,t)}catch(i){return void e.Debug.assert(!a.has(e.ensureTrailingDirectorySeparator(t)))}}function p(e,n){return i(e)===i(n)}function m(n,t){return e.some(n,function(e){return p(e,t)})}function f(n,t,r){if(m(n,t)){if(!r)return e.filterMutate(n,function(e){return!p(e,t)})}else if(r)return n.push(t)}function _(e,t,r){var a=o(e),i=c(a);return i&&S(i,l(e),!0),n.writeFile(e,t,r)}function g(e){var t=o(e),r=c(t);return r&&m(r.files,l(e))||n.fileExists(e)}function v(t){var r=o(t);return a.has(e.ensureTrailingDirectorySeparator(r))||n.directoryExists(t)}function y(e){var t=o(e),r=c(t),a=l(e);r&&f(r.directories,a,!0),n.createDirectory(e)}function h(e){var t=o(e),r=d(e,t);return r?r.directories.slice():n.getDirectories(e)}function b(a,i,s,c,l){var u=o(a),p=d(a,u);return p?e.matchFiles(a,i,s,c,r,t,l,m):n.readDirectory(a,i,s,c,l);function m(n){var t=o(n);return t===u?p:d(n,t)||e.emptyFileSystemEntries}}function E(e,t){var r=s(t);if(r)x();else{var a=c(t);if(a){if(n.directoryExists){var i=l(e),o={fileExists:n.fileExists(t),directoryExists:n.directoryExists(t)};return o.directoryExists||m(a.directories,i)?x():S(a,i,o.fileExists),o}x()}}}function T(n,t,r){if(r!==e.FileWatcherEventKind.Changed){var a=c(t);a&&S(a,l(n),r===e.FileWatcherEventKind.Created)}}function S(e,n,t){f(e.files,n,t)}function x(){a.clear()}}var t;function r(n,t,r){var a=n.getMissingFilePaths(),i=e.arrayToSet(a);e.mutateMap(t,i,{createNewValue:r,onDeleteValue:e.closeFileWatcher})}function a(n,t,r){function a(e,n){return{watcher:r(e,n),flags:n}}function i(e,t,r){e.flags!==t&&(e.watcher.close(),n.set(r,a(r,t)))}e.mutateMap(n,t,{createNewValue:a,onDeleteValue:_,onExistingValue:i})}function i(e,n){return!!e&&e.isEmittedFile(n)}function o(e,n,t){return s(e,n,t,c,l)}function s(e,n,r,a,i){var o=u(e,a),s=e===t.None?l:o,c=u(e,i);return{watchFile:function(e,t,i,s,c,l){return o(e,t,i,s,void 0,c,l,a,n,"FileWatcher",r)},watchFilePath:function(e,t,i,o,c,l,u){return s(e,t,i,o,c,l,u,a,n,"FileWatcher",r)},watchDirectory:function(e,t,a,o,s,l){return c(e,t,a,o,void 0,s,l,i,n,"DirectoryWatcher",r)}};function l(e,n,t,r,i){return a(e,n,function(e,n){return t(e,n,i)},r)}}function c(e,n,t,r){return e.watchFile(n,t,r)}function l(e,n,t,r){return e.watchDirectory(n,t,0!==(1&r))}function u(e,n){switch(e){case t.None:return n;case t.TriggerOnly:return m;case t.Verbose:return n===l?p:d}}function d(e,n,t,r,a,i,o,s,c,l,u){c(l+":: Added:: "+f(n,r,i,o,u));var d=m(e,n,t,r,a,i,o,s,c,l,u);return{close:function(){c(l+":: Close:: "+f(n,r,i,o,u)),d.close()}}}function p(n,t,r,a,i,o,s,c,l,u,d){var p=u+":: Added:: "+f(t,a,o,s,d);l(p);var _=e.timestamp(),g=m(n,t,r,a,i,o,s,c,l,u,d),v=e.timestamp()-_;return l("Elapsed:: "+v+"ms "+p),{close:function(){var n=u+":: Close:: "+f(t,a,o,s,d);l(n);var r=e.timestamp();g.close();var i=e.timestamp()-r;l("Elapsed:: "+i+"ms "+n)}}}function m(n,t,r,a,i,o,s,c,l,u,d){return c(n,t,function(n,c){var p=u+":: Triggered with "+n+" "+(void 0!==c?c:"")+":: "+f(t,a,o,s,d);l(p);var m=e.timestamp();r(n,c,i);var _=e.timestamp()-m;l("Elapsed:: "+_+"ms "+p)},a)}function f(e,n,t,r,a){return"WatchInfo: "+e+" "+n+" "+(a?a(t,r):void 0===r?t:t+" "+r)}function _(e){e.watcher.close()}e.createCachedDirectoryStructureHost=n,function(e){e[e["None"]=0]="None",e[e["Partial"]=1]="Partial",e[e["Full"]=2]="Full"}(e.ConfigFileProgramReloadLevel||(e.ConfigFileProgramReloadLevel={})),e.updateMissingFilePathsWatch=r,e.updateWatchingWildcardDirectories=a,e.isEmittedFileOfProgram=i,function(e){e[e["None"]=0]="None",e[e["TriggerOnly"]=1]="TriggerOnly",e[e["Verbose"]=2]="Verbose"}(t=e.WatchLogLevel||(e.WatchLogLevel={})),e.getWatchFactory=o,e.closeFileWatcherOf=_}(u||(u={})),function(e){var n,t=/(^\s*$)|(^\s*\/\/\/?\s*(@ts-ignore)?)/;function r(n,t,r){return void 0===r&&(r="tsconfig.json"),e.forEachAncestorDirectory(n,function(n){var a=e.combinePaths(n,r);return t(a)?a:void 0})}function a(n,t){var r=e.getDirectoryPath(t),a=e.isRootedDiskPath(n)?n:e.combinePaths(r,n);return e.normalizePath(a)}function i(n,t,r){var a,i=e.forEach(n,function(n){var i=e.getNormalizedPathComponents(n,t);if(i.pop(),a){for(var o=Math.min(a.length,i.length),s=0;s<o;s++)if(r(a[s])!==r(i[s])){if(0===s)return!0;a.length=s;break}i.length<a.length&&(a.length=i.length)}else a=i});return i?"":a?e.getPathFromPathComponents(a):t}function s(e,n){return c(e,n)}function c(n,t,r){void 0===r&&(r=e.sys);var a,i=e.createMap();function o(e){return r.useCaseSensitiveFileNames?e:e.toLowerCase()}function s(n,r,a){var i;try{e.performance.mark("beforeIORead"),i=_.readFile(n),e.performance.mark("afterIORead"),e.performance.measure("I/O Read","beforeIORead","afterIORead")}catch(o){a&&a(o.message),i=""}return void 0!==i?e.createSourceFile(n,i,r,t):void 0}function c(e){return!!i.has(e)||!!r.directoryExists(e)&&(i.set(e,!0),!0)}function l(n){if(n.length>e.getRootLength(n)&&!c(n)){var t=e.getDirectoryPath(n);l(t),_.createDirectory?_.createDirectory(n):r.createDirectory(n)}}function u(n,t,i){a||(a=e.createMap());var o=r.createHash(t),s=r.getModifiedTime(n);if(s){var c=a.get(n);if(c&&c.byteOrderMark===i&&c.hash===o&&c.mtime.getTime()===s.getTime())return}r.writeFile(n,t,i);var l=r.getModifiedTime(n)||e.missingFileModifiedTime;a.set(n,{hash:o,byteOrderMark:i,mtime:l})}function d(t,a,i,o){try{e.performance.mark("beforeIOWrite"),l(e.getDirectoryPath(e.normalizePath(t))),e.isWatchSet(n)&&r.createHash&&r.getModifiedTime?u(t,a,i):r.writeFile(t,a,i),e.performance.mark("afterIOWrite"),e.performance.measure("I/O Write","beforeIOWrite","afterIOWrite")}catch(s){o&&o(s.message)}}function p(){return e.getDirectoryPath(e.normalizePath(r.getExecutingFilePath()))}var m=e.getNewLineCharacter(n,function(){return r.newLine}),f=r.realpath&&function(e){return r.realpath(e)},_={getSourceFile:s,getDefaultLibLocation:p,getDefaultLibFileName:function(n){return e.combinePaths(p(),e.getDefaultLibFileName(n))},writeFile:d,getCurrentDirectory:e.memoize(function(){return r.getCurrentDirectory()}),useCaseSensitiveFileNames:function(){return r.useCaseSensitiveFileNames},getCanonicalFileName:o,getNewLine:function(){return m},fileExists:function(e){return r.fileExists(e)},readFile:function(e){return r.readFile(e)},trace:function(e){return r.write(e+m)},directoryExists:function(e){return r.directoryExists(e)},getEnvironmentVariable:function(e){return r.getEnvironmentVariable?r.getEnvironmentVariable(e):""},getDirectories:function(e){return r.getDirectories(e)},realpath:f,readDirectory:function(e,n,t,a,i){return r.readDirectory(e,n,t,a,i)},createDirectory:function(e){return r.createDirectory(e)}};return _}function l(n,t,r){var a=n.readFile,i=n.fileExists,o=n.directoryExists,s=n.createDirectory,c=n.writeFile,l=e.createMap(),u=e.createMap(),d=e.createMap(),p=e.createMap(),m=function(e){var n=t(e),r=l.get(n);return void 0!==r?r||void 0:f(n,e)},f=function(e,t){var r=a.call(n,t);return l.set(e,r||!1),r};n.readFile=function(r){var i=t(r),o=l.get(i);return void 0!==o?o:e.fileExtensionIs(r,".json")?f(i,r):a.call(n,r)};var _=r?function(n,a,i,o){var s=t(n),c=p.get(s);if(c)return c;var l=r(n,a,i,o);return l&&(e.isDeclarationFileName(n)||e.fileExtensionIs(n,".json"))&&p.set(s,l),l}:void 0;return n.fileExists=function(e){var r=t(e),a=u.get(r);if(void 0!==a)return a;var o=i.call(n,e);return u.set(r,!!o),o},c&&(n.writeFile=function(e,r,a,i,o){var s=t(e);u.delete(s);var d=l.get(s);if(d&&d!==r)l.delete(s),p.delete(s);else if(_){var m=p.get(s);m&&m.text!==r&&p.delete(s)}c.call(n,e,r,a,i,o)}),o&&s&&(n.directoryExists=function(e){var r=t(e),a=d.get(r);if(void 0!==a)return a;var i=o.call(n,e);return d.set(r,!!i),i},n.createDirectory=function(e){var r=t(e);d.delete(r),s.call(n,e)}),{originalReadFile:a,originalFileExists:i,originalDirectoryExists:o,originalCreateDirectory:s,originalWriteFile:c,getSourceFileWithCache:_,readFileWithCache:m}}function u(n,t,r){var a=n.getConfigFileParsingDiagnostics().concat(n.getOptionsDiagnostics(r),n.getSyntacticDiagnostics(t,r),n.getGlobalDiagnostics(r),n.getSemanticDiagnostics(t,r));return e.getEmitDeclarations(n.getCompilerOptions())&&e.addRange(a,n.getDeclarationDiagnostics(t,r)),e.sortAndDeduplicateDiagnostics(a)}function d(e,n){for(var t="",r=0,a=e;r<a.length;r++){var i=a[r];t+=p(i,n)}return t}function p(n,t){var r=e.diagnosticCategoryName(n)+" TS"+n.code+": "+C(n.messageText,t.getNewLine())+t.getNewLine();if(n.file){var a=e.getLineAndCharacterOfPosition(n.file,n.start),i=a.line,o=a.character,s=n.file.fileName,c=e.convertToRelativePath(s,t.getCurrentDirectory(),function(e){return t.getCanonicalFileName(e)});return c+"("+(i+1)+","+(o+1)+"): "+r}return r}e.findConfigFile=r,e.resolveTripleslashReference=a,e.computeCommonSourceDirectoryOfFilenames=i,e.createCompilerHost=s,e.createCompilerHostWorker=c,e.changeCompilerHostLikeToUseCache=l,e.getPreEmitDiagnostics=u,e.formatDiagnostics=d,e.formatDiagnostic=p,function(e){e["Grey"]="[90m",e["Red"]="[91m",e["Yellow"]="[93m",e["Blue"]="[94m",e["Cyan"]="[96m"}(n=e.ForegroundColorEscapeSequences||(e.ForegroundColorEscapeSequences={}));var m="[7m",f=" ",_="[0m",g="...",v="  ",y="    ";function h(t){switch(t){case e.DiagnosticCategory.Error:return n.Red;case e.DiagnosticCategory.Warning:return n.Yellow;case e.DiagnosticCategory.Suggestion:return e.Debug.fail("Should never get an Info diagnostic on the command line.");case e.DiagnosticCategory.Message:return n.Blue}}function b(e,n){return n+e+_}function E(e,n){while(e.length<n)e=" "+e;return e}function T(n,t,r,a,i,o){var s=e.getLineAndCharacterOfPosition(n,t),c=s.line,l=s.character,u=e.getLineAndCharacterOfPosition(n,t+r),d=u.line,p=u.character,v=e.getLineAndCharacterOfPosition(n,n.text.length).line,y=d-c>=4,h=(d+1+"").length;y&&(h=Math.max(g.length,h));for(var T="",S=c;S<=d;S++){T+=o.getNewLine(),y&&c+1<S&&S<d-1&&(T+=a+b(E(g,h),m)+f+o.getNewLine(),S=d-1);var x=e.getPositionOfLineAndCharacter(n,S,0),C=S<v?e.getPositionOfLineAndCharacter(n,S+1,0):n.text.length,A=n.text.slice(x,C);if(A=A.replace(/\s+$/g,""),A=A.replace("\t"," "),T+=a+b(E(S+1+"",h),m)+f,T+=A+o.getNewLine(),T+=a+b(E("",h),m)+f,T+=i,S===c){var D=S===d?p:void 0;T+=A.slice(0,l).replace(/\S/g," "),T+=A.slice(l,D).replace(/./g,"~")}else T+=S===d?A.slice(0,p).replace(/./g,"~"):A.replace(/./g,"~");T+=_}return T}function S(t,r,a,i){void 0===i&&(i=b);var o=e.getLineAndCharacterOfPosition(t,r),s=o.line,c=o.character,l=a?e.convertToRelativePath(t.fileName,a.getCurrentDirectory(),function(e){return a.getCanonicalFileName(e)}):t.fileName,u="";return u+=i(l,n.Cyan),u+=":",u+=i(""+(s+1),n.Yellow),u+=":",u+=i(""+(c+1),n.Yellow),u}function x(t,r){for(var a="",i=0,o=t;i<o.length;i++){var s=o[i];if(s.file){var c=s.file,l=s.start;a+=S(c,l,r),a+=" - "}if(a+=b(e.diagnosticCategoryName(s),h(s.category)),a+=b(" TS"+s.code+": ",n.Grey),a+=C(s.messageText,r.getNewLine()),s.file&&(a+=r.getNewLine(),a+=T(s.file,s.start,s.length,"",h(s.category),r),s.relatedInformation)){a+=r.getNewLine();for(var u=0,d=s.relatedInformation;u<d.length;u++){var p=d[u],m=(c=p.file,l=p.start,p.length),f=p.messageText;c&&(a+=r.getNewLine(),a+=v+S(c,l,r),a+=T(c,l,m,y,n.Cyan,r)),a+=r.getNewLine(),a+=y+C(f,r.getNewLine())}}a+=r.getNewLine()}return a}function C(n,t){if(e.isString(n))return n;var r=n,a="",i=0;while(r){if(i){a+=t;for(var o=0;o<i;o++)a+="  "}a+=r.messageText,i++,r=r.next}return a}function A(n,t,r,a){if(0===n.length)return[];for(var i=[],o=e.createMap(),s=0,c=n;s<c.length;s++){var l=c[s],u=void 0;o.has(l)?u=o.get(l):o.set(l,u=a(l,t,r)),i.push(u)}return i}function D(n,t,r,a,i,o,s,c){if(!n||s)return!1;if(n.getRootFileNames().length!==t.length)return!1;var l;if(!e.arrayIsEqualTo(n.getProjectReferences(),c,m))return!1;if(n.getSourceFiles().some(d))return!1;if(n.getMissingFilePaths().some(i))return!1;var u=n.getCompilerOptions();return!!e.compareDataObjects(u,r)&&(!u.configFile||!r.configFile||u.configFile.text===r.configFile.text);function d(e){return!p(e)||o(e.path)}function p(e){return e.version===a(e.resolvedPath)}function m(t,r,a){return!!e.projectReferenceIsEqualTo(t,r)&&f(n.getResolvedProjectReferences()[a],t)}function f(n,t){return n?!!e.contains(l,n)||!!p(n.sourceFile)&&((l||(l=[])).push(n),!e.forEach(n.references,function(e,t){return!f(e,n.commandLine.projectReferences[t])})):!i(M(t))}}function L(e){return e.options.configFile?e.options.configFile.parseDiagnostics.concat(e.errors):e.errors}function k(n,t){if(!n)return!1;var r=n.getCompilerOptions();return!!e.sourceFileAffectingCompilerOptions.some(function(n){return!e.isJsonEqual(e.getCompilerOptionValue(r,n),e.getCompilerOptionValue(t,n))})}function N(e,n,t,r,a){return{rootNames:e,options:n,host:t,oldProgram:r,configFileParsingDiagnostics:a}}function I(n,r,c,l,u){var d,p,m,f,_,g,v,y,h=e.isArray(n)?N(n,r,c,l,u):n,b=h.rootNames,E=h.options,T=h.configFileParsingDiagnostics,S=h.projectReferences,x=h.oldProgram,C=e.createMap(),D={},L={},I=e.createMap(),F=e.createDiagnosticCollection(),G="number"===typeof E.maxNodeModuleJsDepth?E.maxNodeModuleJsDepth:0,V=0,B=e.createMap(),K=e.createMap();e.performance.mark("beforeProgram");var j,U,H,W,q=h.host||s(E),z=w(q),J=E.noLib,X=e.memoize(function(){return q.getDefaultLibFileName(E)}),Y=q.getDefaultLibLocation?q.getDefaultLibLocation():e.getDirectoryPath(X()),Q=e.createDiagnosticCollection(),Z=q.getCurrentDirectory(),$=e.getSupportedExtensions(E),ee=e.getSuppoertedExtensionsWithJsonIfResolveJsonModule(E,$),ne=e.createMap(),te=q.hasInvalidatedResolution||e.returnFalse;if(q.resolveModuleNames)H=function(n,t,r,a){return q.resolveModuleNames(e.Debug.assertEachDefined(n),t,r,a).map(function(n){if(!n||void 0!==n.extension)return n;var t=e.clone(n);return t.extension=e.extensionFromPath(n.resolvedFileName),t})};else{U=e.createModuleResolutionCache(Z,function(e){return q.getCanonicalFileName(e)});var re=function(n,t,r){return e.resolveModuleName(n,t,E,q,U,r).resolvedModule};H=function(n,t,r,a){return A(e.Debug.assertEachDefined(n),t,a,re)}}if(q.resolveTypeReferenceDirectives)W=function(n,t,r){return q.resolveTypeReferenceDirectives(e.Debug.assertEachDefined(n),t,r)};else{var ae=function(n,t,r){return e.resolveTypeReferenceDirective(n,t,E,q,r).resolvedTypeReferenceDirective};W=function(n,t,r){return A(e.Debug.assertEachDefined(n),t,r,ae)}}var ie,oe,se,ce,le=e.createMap(),ue=e.createMap(),de=e.createMultiMap(),pe=e.createMap(),me=q.useCaseSensitiveFileNames()?e.createMap():void 0,fe=k(x,E),_e=Be();if(2!==_e){if(p=[],m=[],S&&(oe||(oe=S.map(zn)),b.length))for(var ge=0,ve=oe;ge<ve.length;ge++){var ye=ve[ge];if(ye){var he=ye.commandLine.options.outFile||ye.commandLine.options.out;if(he){var be=e.changeExtension(he,".d.ts");Ln(be,!1,!1,void 0)}}}e.forEach(b,function(e){return En(e,!1,!1)});var Ee=b.length?e.getAutomaticTypeDirectiveNames(E,q):e.emptyArray;if(Ee.length)for(var Te=E.configFilePath?e.getDirectoryPath(E.configFilePath):q.getCurrentDirectory(),Se=e.combinePaths(Te,"__inferred type names__.ts"),xe=W(Ee,Se),Ce=0;Ce<Ee.length;Ce++)Bn(Ee[Ce],xe[Ce]);if(b.length&&!J){var Ae=X();!E.lib&&Ae?En(Ae,!0,!1):e.forEach(E.lib,function(n){En(e.combinePaths(Y,n),!0,!1)})}ie=e.arrayFrom(e.mapDefinedIterator(pe.entries(),function(e){var n=e[0],t=e[1];return void 0===t?n:void 0})),f=e.stableSort(p,we).concat(m),p=void 0,m=void 0}if(e.Debug.assert(!!ie),x&&q.onReleaseOldSourceFile){for(var De=x.getSourceFiles(),Le=0,ke=De;Le<ke.length;Le++){var Ne=ke[Le],Ie=en(Ne.resolvedPath);(fe||!Ie||Ne.resolvedPath===Ne.path&&Ie.resolvedPath!==Ne.path)&&q.onReleaseOldSourceFile(Ne,x.getCompilerOptions(),!!en(Ne.path))}x.forEachResolvedProjectReference(function(e,n){e&&!Fn(n)&&q.onReleaseOldSourceFile(e.sourceFile,x.getCompilerOptions(),!1)})}return x=void 0,d={getRootFileNames:function(){return b},getSourceFile:$e,getSourceFileByPath:en,getSourceFiles:function(){return f},getMissingFilePaths:function(){return ie},getCompilerOptions:function(){return E},getSyntacticDiagnostics:tn,getOptionsDiagnostics:vn,getGlobalDiagnostics:hn,getSemanticDiagnostics:rn,getSuggestionDiagnostics:un,getDeclarationDiagnostics:an,getTypeChecker:Xe,getClassifiableNames:Fe,getDiagnosticsProducingTypeChecker:ze,getCommonSourceDirectory:Re,emit:Ye,getCurrentDirectory:function(){return Z},getNodeCount:function(){return ze().getNodeCount()},getIdentifierCount:function(){return ze().getIdentifierCount()},getSymbolCount:function(){return ze().getSymbolCount()},getTypeCount:function(){return ze().getTypeCount()},getFileProcessingDiagnostics:function(){return F},getResolvedTypeReferenceDirectives:function(){return I},isSourceFileFromExternalLibrary:We,isSourceFileDefaultLibrary:qe,dropDiagnosticsProducingTypeChecker:Je,getSourceFileFromReference:An,getLibFileFromReference:Cn,sourceFileToPackageName:ue,redirectTargetsMap:de,isEmittedFile:st,getConfigFileParsingDiagnostics:bn,getResolvedModuleWithFailedLookupLocationsFromCache:Oe,getProjectReferences:Ue,getResolvedProjectReferences:je,getProjectReferenceRedirect:Mn,getResolvedProjectReferenceToRedirect:On,getResolvedProjectReferenceByPath:Fn,forEachResolvedProjectReference:Pn},Jn(),e.performance.mark("afterProgram"),e.performance.measure("Program","beforeProgram","afterProgram"),d;function we(n,t){return e.compareValues(Me(n),Me(t))}function Me(n){if(e.containsPath(Y,n.fileName,!1)){var t=e.getBaseFileName(n.fileName);if("lib.d.ts"===t||"lib.es6.d.ts"===t)return 0;var r=e.removeSuffix(e.removePrefix(t,"lib."),".d.ts"),a=e.libs.indexOf(r);if(-1!==a)return a+1}return e.libs.length+2}function Oe(n,t){return U&&e.resolveModuleNameFromCache(n,t,U)}function Pe(n){return e.toPath(n,Z,Un)}function Re(){if(void 0===_){var n=e.filter(f,function(n){return e.sourceFileMayBeEmitted(n,E,We)});E.rootDir&&qn(n,E.rootDir)?_=e.getNormalizedAbsolutePath(E.rootDir,Z):E.composite&&E.configFilePath?(_=e.getDirectoryPath(e.normalizeSlashes(E.configFilePath)),qn(n,_)):_=Wn(n),_&&_[_.length-1]!==e.directorySeparator&&(_+=e.directorySeparator)}return _}function Fe(){if(!y){Xe(),y=e.createUnderscoreEscapedMap();for(var n=0,t=f;n<t.length;n++){var r=t[n];e.copyEntries(r.classifiableNames,y)}}return y}function Ge(n,t,r){if(0===_e&&!r.ambientModuleNames.length)return H(n,t,void 0,On(r.originalFileName));var a,i,o,s=x&&x.getSourceFile(t);if(s!==r&&r.resolvedModules){for(var c=[],l=0,u=n;l<u.length;l++){var d=u[l],p=r.resolvedModules.get(d);c.push(p)}return c}for(var m={},f=0;f<n.length;f++){d=n[f];if(r===s&&!te(s.path)){var _=s&&s.resolvedModules.get(d);if(_){e.isTraceEnabled(E,q)&&e.trace(q,e.Diagnostics.Reusing_resolution_of_module_0_to_file_1_from_old_program,d,t),(i||(i=new Array(n.length)))[f]=_,(o||(o=[])).push(d);continue}}var g=!1;e.contains(r.ambientModuleNames,d)?(g=!0,e.isTraceEnabled(E,q)&&e.trace(q,e.Diagnostics.Module_0_was_resolved_as_locally_declared_ambient_module_in_file_1,d,t)):g=h(d),g?(i||(i=new Array(n.length)))[f]=m:(a||(a=[])).push(d)}var v=a&&a.length?H(a,t,o,On(r.originalFileName)):e.emptyArray;if(!i)return e.Debug.assert(v.length===n.length),v;var y=0;for(f=0;f<i.length;f++)i[f]?i[f]===m&&(i[f]=void 0):(i[f]=v[y],y++);return e.Debug.assert(y===v.length),i;function h(n){var t=e.getResolvedModule(s,n),r=t&&x.getSourceFile(t.resolvedFileName);if(t&&r&&!r.externalModuleIndicator)return!1;var a=C.get(n);return!!a&&(e.isTraceEnabled(E,q)&&e.trace(q,e.Diagnostics.Module_0_was_resolved_as_ambient_module_declared_in_1_since_this_file_was_not_modified,n,a),!0)}}function Ve(){return!Rn(x.getProjectReferences(),x.getResolvedProjectReferences(),function(e,n,t){var r=(t?t.commandLine.projectReferences:S)[n],a=zn(r);return e?!a||a.sourceFile!==e.sourceFile:void 0!==a},function(n,t){var r=t?Fn(t.sourceFile.path).commandLine.projectReferences:S;return!e.arrayIsEqualTo(n,r,e.projectReferenceIsEqualTo)})}function Be(){if(!x)return 0;var n=x.getCompilerOptions();if(e.changesAffectModuleResolution(n,E))return x.structureIsReused=0;e.Debug.assert(!(3&x.structureIsReused));var t=x.getRootFileNames();if(!e.arrayIsEqualTo(t,b))return x.structureIsReused=0;if(!e.arrayIsEqualTo(E.types,n.types))return x.structureIsReused=0;if(!Ve())return x.structureIsReused=0;S&&(oe=S.map(zn));var r=[],a=[];if(x.structureIsReused=2,x.getMissingFilePaths().some(function(e){return q.fileExists(e)}))return x.structureIsReused=0;var i,o=x.getSourceFiles();(function(e){e[e["Exists"]=0]="Exists",e[e["Modified"]=1]="Modified"})(i||(i={}));for(var s=e.createMap(),c=0,l=o;c<l.length;c++){var u=l[c],d=q.getSourceFileByPath?q.getSourceFileByPath(u.fileName,u.resolvedPath,E.target,void 0,fe):q.getSourceFile(u.fileName,E.target,void 0,fe);if(!d)return x.structureIsReused=0;e.Debug.assert(!d.redirectInfo,"Host should not return a redirect source file from `getSourceFile`");var p=void 0;if(u.redirectInfo){if(d!==u.redirectInfo.unredirected)return x.structureIsReused=0;p=!1,d=u}else if(x.redirectTargetsMap.has(u.path)){if(d!==u)return x.structureIsReused=0;p=!1}else p=d!==u;d.path=u.path,d.originalFileName=u.originalFileName,d.resolvedPath=u.resolvedPath,d.fileName=u.fileName;var m=x.sourceFileToPackageName.get(u.path);if(void 0!==m){var _=s.get(m),g=p?1:0;if(void 0!==_&&1===g||1===_)return x.structureIsReused=0;s.set(m,g)}if(p){if(!e.arrayIsEqualTo(u.libReferenceDirectives,d.libReferenceDirectives,Tn))return x.structureIsReused=0;u.hasNoDefaultLib!==d.hasNoDefaultLib&&(x.structureIsReused=1),e.arrayIsEqualTo(u.referencedFiles,d.referencedFiles,Tn)||(x.structureIsReused=1),xn(d),e.arrayIsEqualTo(u.imports,d.imports,Sn)||(x.structureIsReused=1),e.arrayIsEqualTo(u.moduleAugmentations,d.moduleAugmentations,Sn)||(x.structureIsReused=1),(1572864&u.flags)!==(1572864&d.flags)&&(x.structureIsReused=1),e.arrayIsEqualTo(u.typeReferenceDirectives,d.typeReferenceDirectives,Tn)||(x.structureIsReused=1),a.push({oldFile:u,newFile:d})}else te(u.path)&&(x.structureIsReused=1,a.push({oldFile:u,newFile:d}));r.push(d)}if(2!==x.structureIsReused)return x.structureIsReused;for(var v=a.map(function(e){return e.oldFile}),y=0,h=o;y<h.length;y++){var T=h[y];if(!e.contains(v,T))for(var A=0,D=T.ambientModuleNames;A<D.length;A++){var L=D[A];C.set(L,T.fileName)}}for(var k=0,N=a;k<N.length;k++){var w=N[k],M=(u=w.oldFile,d=w.newFile,e.getNormalizedAbsolutePath(d.originalFileName,Z));if(H){var O=R(d),P=Ge(O,M,d),G=e.hasChangesInResolutions(O,P,u.resolvedModules,e.moduleResolutionIsEqualTo);G?(x.structureIsReused=1,d.resolvedModules=e.zipToMap(O,P)):d.resolvedModules=u.resolvedModules}if(W){var V=e.map(d.typeReferenceDirectives,function(e){return e.fileName.toLocaleLowerCase()});P=W(V,M,On(d.originalFileName)),G=e.hasChangesInResolutions(V,P,u.resolvedTypeReferenceDirectiveNames,e.typeDirectiveIsEqualTo);G?(x.structureIsReused=1,d.resolvedTypeReferenceDirectiveNames=e.zipToMap(V,P)):d.resolvedTypeReferenceDirectiveNames=u.resolvedTypeReferenceDirectiveNames}}if(2!==x.structureIsReused)return x.structureIsReused;if(q.hasChangedAutomaticTypeDirectiveNames)return x.structureIsReused=1;ie=x.getMissingFilePaths();for(var B=0,j=r;B<j.length;B++){d=j[B];var U=d.path;wn(d,U,d.resolvedPath),x.isSourceFileFromExternalLibrary(x.getSourceFileByPath(U))&&K.set(U,!0)}f=r,F=x.getFileProcessingDiagnostics();for(var z=0,J=a;z<J.length;z++){var X=J[z];F.reattachFileDiagnostics(X.newFile)}return I=x.getResolvedTypeReferenceDirectives(),ue=x.sourceFileToPackageName,de=x.redirectTargetsMap,x.structureIsReused=2}function Ke(n){return o({getPrependNodes:He,getCanonicalFileName:Un,getCommonSourceDirectory:d.getCommonSourceDirectory,getCompilerOptions:d.getCompilerOptions,getCurrentDirectory:function(){return Z},getNewLine:function(){return q.getNewLine()},getSourceFile:d.getSourceFile,getSourceFileByPath:d.getSourceFileByPath,getSourceFiles:d.getSourceFiles,getLibFileFromReference:d.getLibFileFromReference,isSourceFileFromExternalLibrary:We,writeFile:n||function(e,n,t,r,a){return q.writeFile(e,n,t,r,a)},isEmitBlocked:Qe,readFile:function(e){return q.readFile(e)},fileExists:function(n){var t=Pe(n);return!!en(t)||!e.contains(ie,t)&&q.fileExists(n)}},q.directoryExists?{directoryExists:function(e){return q.directoryExists(e)}}:{},{useCaseSensitiveFileNames:function(){return q.useCaseSensitiveFileNames()}})}function je(){return oe}function Ue(){return S}function He(){if(!S)return e.emptyArray;for(var n=[],t=0;t<S.length;t++){var r=S[t],a=oe[t].commandLine;if(r.prepend&&a&&a.options){var i=a.options.outFile||a.options.out;if(!i)continue;var o=e.getOutputPathsForBundle(a.options,!0),s=o.jsFilePath,c=o.sourceMapFilePath,l=o.declarationFilePath,u=o.declarationMapPath,d=e.createInputFiles(function(e){var n=Pe(e),t=en(n);return t?t.text:pe.has(n)?void 0:q.readFile(n)},s,c,l,u);n.push(d)}}return n}function We(e){return!!K.get(e.path)}function qe(n){if(n.hasNoDefaultLib)return!0;if(!E.noLib)return!1;var t=q.useCaseSensitiveFileNames()?e.equateStringsCaseSensitive:e.equateStringsCaseInsensitive;return E.lib?e.some(E.lib,function(r){return t(n.fileName,e.combinePaths(Y,r))}):t(n.fileName,X())}function ze(){return g||(g=e.createTypeChecker(d,!0))}function Je(){g=void 0}function Xe(){return v||(v=e.createTypeChecker(d,!1))}function Ye(e,n,t,r,a){return sn(function(){return Ze(d,e,n,t,r,a)})}function Qe(e){return ne.has(Pe(e))}function Ze(n,t,r,a,i,o){var s=[];if(!i){if(E.noEmit)return{diagnostics:s,sourceMaps:void 0,emittedFiles:void 0,emitSkipped:!0};if(E.noEmitOnError){var c=n.getOptionsDiagnostics(a).concat(n.getSyntacticDiagnostics(t,a),n.getGlobalDiagnostics(a),n.getSemanticDiagnostics(t,a));if(0===c.length&&e.getEmitDeclarations(n.getCompilerOptions())&&(s=n.getDeclarationDiagnostics(void 0,a)),c.length>0||s.length>0)return{diagnostics:e.concatenate(c,s),sourceMaps:void 0,emittedFiles:void 0,emitSkipped:!0}}}var l=ze().getEmitResolver(E.outFile||E.out?void 0:t,a);e.performance.mark("beforeEmit");var u=i?[]:e.getTransformers(E,o),d=e.emitFiles(l,Ke(r),t,i,u,o&&o.afterDeclarations);return e.performance.mark("afterEmit"),e.performance.measure("Emit","beforeEmit","afterEmit"),d}function $e(e){return en(Pe(e))}function en(e){return pe.get(e)||void 0}function nn(n,t,r){return n?t(n,r):e.sortAndDeduplicateDiagnostics(e.flatMap(d.getSourceFiles(),function(e){return r&&r.throwIfCancellationRequested(),t(e,r)}))}function tn(e,n){return nn(e,on,n)}function rn(e,n){return nn(e,cn,n)}function an(e,n){var t=d.getCompilerOptions();return!e||t.out||t.outFile?mn(e,n):nn(e,gn,n)}function on(n){return e.isSourceFileJS(n)?(n.additionalSyntacticDiagnostics||(n.additionalSyntacticDiagnostics=pn(n)),e.concatenate(n.additionalSyntacticDiagnostics,n.parseDiagnostics)):n.parseDiagnostics}function sn(n){try{return n()}catch(t){throw t instanceof e.OperationCanceledException&&(v=void 0,g=void 0),t}}function cn(e,n){return _n(e,n,D,ln)}function ln(n,t){return sn(function(){if(e.skipTypeChecking(n,E))return e.emptyArray;var r=ze();e.Debug.assert(!!n.bindDiagnostics);for(var a,i=e.isCheckJsEnabledForFile(n,E),o=3===n.scriptKind||4===n.scriptKind||5===n.scriptKind||i||7===n.scriptKind,s=o?n.bindDiagnostics:e.emptyArray,c=o?r.getDiagnostics(n,t):e.emptyArray,l=F.getDiagnostics(n.fileName),u=Q.getDiagnostics(n.fileName),d=0,p=[s,c,l,u,i?n.jsDocDiagnostics:void 0];d<p.length;d++){var m=p[d];if(m)for(var f=0,_=m;f<_.length;f++){var g=_[f];dn(g)&&(a=e.append(a,g))}}return a})}function un(e,n){return sn(function(){return ze().getSuggestionDiagnostics(e,n)})}function dn(n){var r=n.file,a=n.start;if(r){var i=e.getLineStarts(r),o=e.computeLineAndCharacterOfPosition(i,a).line;while(o>0){var s=r.text.slice(i[o-1],i[o]),c=t.exec(s);if(!c)return!0;if(c[3])return!1;o--}}return!0}function pn(n){return sn(function(){var t=[],r=n;return a(n),t;function a(n){switch(r.kind){case 151:case 154:if(r.questionToken===n)return void t.push(c(n,e.Diagnostics._0_can_only_be_used_in_a_ts_file,"?"));case 156:case 155:case 157:case 158:case 159:case 196:case 239:case 197:case 237:if(r.type===n)return void t.push(c(n,e.Diagnostics.types_can_only_be_used_in_a_ts_file))}switch(n.kind){case 248:return void t.push(c(n,e.Diagnostics.import_can_only_be_used_in_a_ts_file));case 254:if(n.isExportEquals)return void t.push(c(n,e.Diagnostics.export_can_only_be_used_in_a_ts_file));break;case 273:var o=n;if(109===o.token)return void t.push(c(n,e.Diagnostics.implements_clauses_can_only_be_used_in_a_ts_file));break;case 241:return void t.push(c(n,e.Diagnostics.interface_declarations_can_only_be_used_in_a_ts_file));case 244:return void t.push(c(n,e.Diagnostics.module_declarations_can_only_be_used_in_a_ts_file));case 242:return void t.push(c(n,e.Diagnostics.type_aliases_can_only_be_used_in_a_ts_file));case 243:return void t.push(c(n,e.Diagnostics.enum_declarations_can_only_be_used_in_a_ts_file));case 213:return void t.push(c(n,e.Diagnostics.non_null_assertions_can_only_be_used_in_a_ts_file));case 212:return void t.push(c(n.type,e.Diagnostics.type_assertion_expressions_can_only_be_used_in_a_ts_file));case 194:e.Debug.fail()}var s=r;r=n,e.forEachChild(n,a,i),r=s}function i(n){switch(r.decorators!==n||E.experimentalDecorators||t.push(c(r,e.Diagnostics.Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_to_remove_this_warning)),r.kind){case 240:case 156:case 155:case 157:case 158:case 159:case 196:case 239:case 197:if(n===r.typeParameters)return void t.push(s(n,e.Diagnostics.type_parameter_declarations_can_only_be_used_in_a_ts_file));case 219:if(n===r.modifiers)return o(n,219===r.kind);break;case 154:if(n===r.modifiers){for(var i=0,l=n;i<l.length;i++){var u=l[i];116!==u.kind&&t.push(c(u,e.Diagnostics._0_can_only_be_used_in_a_ts_file,e.tokenToString(u.kind)))}return}break;case 151:if(n===r.modifiers)return void t.push(s(n,e.Diagnostics.parameter_modifiers_can_only_be_used_in_a_ts_file));break;case 191:case 192:case 211:case 261:case 262:if(n===r.typeArguments)return void t.push(s(n,e.Diagnostics.type_arguments_can_only_be_used_in_a_ts_file));break}for(var d=0,p=n;d<p.length;d++){var m=p[d];a(m)}}function o(n,r){for(var a=0,i=n;a<i.length;a++){var o=i[a];switch(o.kind){case 77:if(r)continue;case 115:case 113:case 114:case 133:case 125:case 118:t.push(c(o,e.Diagnostics._0_can_only_be_used_in_a_ts_file,e.tokenToString(o.kind)));break;case 116:case 85:case 80:}}}function s(t,r,a,i,o){var s=t.pos;return e.createFileDiagnostic(n,s,t.end-s,r,a,i,o)}function c(t,r,a,i,o){return e.createDiagnosticForNodeInSourceFile(n,t,r,a,i,o)}})}function mn(e,n){return _n(e,n,L,fn)}function fn(n,t){return sn(function(){var r=ze().getEmitResolver(n,t);return e.getDeclarationDiagnostics(Ke(e.noop),r,n)})}function _n(n,t,r,a){var i=n?r.perFile&&r.perFile.get(n.path):r.allDiagnostics;if(i)return i;var o=a(n,t)||e.emptyArray;return n?(r.perFile||(r.perFile=e.createMap()),r.perFile.set(n.path,o)):r.allDiagnostics=o,o}function gn(e,n){return e.isDeclarationFile?[]:mn(e,n)}function vn(){return e.sortAndDeduplicateDiagnostics(e.concatenate(F.getGlobalDiagnostics(),e.concatenate(Q.getGlobalDiagnostics(),yn())))}function yn(){if(!E.configFile)return e.emptyArray;var n=Q.getDiagnostics(E.configFile.fileName);return Pn(function(t){t&&(n=e.concatenate(n,Q.getDiagnostics(t.sourceFile.fileName)))}),n}function hn(){return b.length?e.sortAndDeduplicateDiagnostics(ze().getGlobalDiagnostics().slice()):e.emptyArray}function bn(){return T||e.emptyArray}function En(n,t,r){Ln(e.normalizePath(n),t,r,void 0)}function Tn(e,n){return e.fileName===n.fileName}function Sn(e,n){return 72===e.kind?72===n.kind&&e.escapedText===n.escapedText:10===n.kind&&e.text===n.text}function xn(n){if(!n.imports){var t,r,a,i=e.isSourceFileJS(n),o=e.isExternalModule(n);if(E.importHelpers&&(E.isolatedModules||o)&&!n.isDeclarationFile){var s=e.createLiteral(e.externalHelpersModuleNameText),c=e.createImportDeclaration(void 0,void 0,void 0,s);e.addEmitFlags(c,67108864),s.parent=c,c.parent=n,t=[s]}for(var l=0,u=n.statements;l<u.length;l++){var d=u[l];p(d,!1)}return(524288&n.flags||i)&&m(n),n.imports=t||e.emptyArray,n.moduleAugmentations=r||e.emptyArray,void(n.ambientModuleNames=a||e.emptyArray)}function p(i,s){if(e.isAnyImportOrReExport(i)){var c=e.getExternalModuleName(i);!(c&&e.isStringLiteral(c)&&c.text)||s&&e.isExternalModuleNameRelative(c.text)||(t=e.append(t,c))}else if(e.isModuleDeclaration(i)&&e.isAmbientModule(i)&&(s||e.hasModifier(i,2)||n.isDeclarationFile)){var l=e.getTextOfIdentifierOrLiteral(i.name);if(o||s&&!e.isExternalModuleNameRelative(l))(r||(r=[])).push(i.name);else if(!s){n.isDeclarationFile&&(a||(a=[])).push(l);var u=i.body;if(u)for(var d=0,m=u.statements;d<m.length;d++){var f=m[d];p(f,!0)}}}}function m(n){var r=/import|require/g;while(null!==r.exec(n.text)){var a=f(n,r.lastIndex);e.isRequireCall(a,!0)?t=e.append(t,a.arguments[0]):e.isImportCall(a)&&1===a.arguments.length&&e.isStringLiteralLike(a.arguments[0])?t=e.append(t,a.arguments[0]):e.isLiteralImportTypeNode(a)&&(t=e.append(t,a.argument.literal))}}function f(n,t){var r=n,a=function(e){if(e.pos<=t&&(t<e.end||t===e.end&&1===e.kind))return e};while(1){var o=i&&e.hasJSDocNodes(r)&&e.forEach(r.jsDoc,a)||e.forEachChild(r,a);if(!o)return r;r=o}}}function Cn(n){var t=n.fileName.toLocaleLowerCase(),r=e.libMap.get(t);if(r)return $e(e.combinePaths(Y,r))}function An(e,n){return Dn(a(n.fileName,e.fileName),function(e){return pe.get(Pe(e))||void 0})}function Dn(n,t,r,a){if(e.hasExtension(n)){if(!E.allowNonTsExtensions&&!e.forEach(ee,function(t){return e.fileExtensionIs(q.getCanonicalFileName(n),t)}))return void(r&&r(e.Diagnostics.File_0_has_unsupported_extension_The_only_supported_extensions_are_1,n,"'"+$.join("', '")+"'"));var i=t(n);if(r)if(i)a&&q.getCanonicalFileName(n)===q.getCanonicalFileName(a.fileName)&&r(e.Diagnostics.A_file_cannot_have_a_reference_to_itself);else{var o=Mn(n);o?r(e.Diagnostics.Output_file_0_has_not_been_built_from_source_file_1,o,n):r(e.Diagnostics.File_0_not_found,n)}return i}var s=E.allowNonTsExtensions&&t(n);if(s)return s;if(!r||!E.allowNonTsExtensions){var c=e.forEach($,function(e){return t(n+e)});return r&&!c&&r(e.Diagnostics.File_0_not_found,n+".ts"),c}r(e.Diagnostics.File_0_not_found,n)}function Ln(n,t,r,a,i,o,s){Dn(n,function(e){return In(e,Pe(e),t,r,i,o,s,a)},function(n){for(var t=[],r=1;r<arguments.length;r++)t[r-1]=arguments[r];F.add(void 0!==i&&void 0!==s&&void 0!==o?e.createFileDiagnostic.apply(void 0,[i,o,s-o,n].concat(t)):e.createCompilerDiagnostic.apply(void 0,[n].concat(t)))},i)}function kn(n,t,r,a,i){void 0!==r&&void 0!==a&&void 0!==i?F.add(e.createFileDiagnostic(r,a,i-a,e.Diagnostics.File_name_0_differs_from_already_included_file_name_1_only_in_casing,n,t)):F.add(e.createCompilerDiagnostic(e.Diagnostics.File_name_0_differs_from_already_included_file_name_1_only_in_casing,n,t))}function Nn(e,n,t,r,a,i){var o=Object.create(e);return o.fileName=t,o.path=r,o.resolvedPath=a,o.originalFileName=i,o.redirectInfo={redirectTarget:e,unredirected:n},K.set(r,V>0),Object.defineProperties(o,{id:{get:function(){return this.redirectInfo.redirectTarget.id},set:function(e){this.redirectInfo.redirectTarget.id=e}},symbol:{get:function(){return this.redirectInfo.redirectTarget.symbol},set:function(e){this.redirectInfo.redirectTarget.symbol=e}}}),o}function In(n,t,r,a,i,o,s,c){var l,u=n;if(pe.has(t)){var d=pe.get(t);if(d&&E.forceConsistentCasingInFileNames){var f=n,_=d.fileName,g=Pe(_)!==Pe(f);g&&(f=Mn(n)||n),e.getNormalizedAbsolutePath(_,Z)!==e.getNormalizedAbsolutePath(f,Z)&&kn(f,_,i,o,s)}return d&&K.get(d.path)&&0===V?(K.set(d.path,!1),E.noResolve||(Gn(d,r),Vn(d)),E.noLib||Kn(d),B.set(d.path,!1),Hn(d)):d&&B.get(d.path)&&V<G&&(B.set(d.path,!1),Hn(d)),d||void 0}if(i){var v=Mn(n);v&&((i.redirectedReferences||(i.redirectedReferences=[])).push(n),n=v,l=Pe(v))}var y=q.getSourceFile(n,E.target,function(t){void 0!==i&&void 0!==o&&void 0!==s?F.add(e.createFileDiagnostic(i,o,s-o,e.Diagnostics.Cannot_read_file_0_Colon_1,n,t)):F.add(e.createCompilerDiagnostic(e.Diagnostics.Cannot_read_file_0_Colon_1,n,t))},fe);if(c){var h=e.packageIdToString(c),b=le.get(h);if(b){var T=Nn(b,y,n,t,Pe(n),u);return de.add(b.path,n),wn(T,t,l),ue.set(t,c.name),m.push(T),T}y&&(le.set(h,y),ue.set(t,c.name))}if(wn(y,t,l),y){if(K.set(t,V>0),y.path=t,y.resolvedPath=Pe(n),y.originalFileName=u,q.useCaseSensitiveFileNames()){var S=t.toLowerCase(),x=me.get(S);x?kn(n,x.fileName,i,o,s):me.set(S,y)}J=J||y.hasNoDefaultLib&&!a,E.noResolve||(Gn(y,r),Vn(y)),E.noLib||Kn(y),Hn(y),r?p.push(y):m.push(y)}return y}function wn(e,n,t){t?(pe.set(t,e),pe.set(n,e||!1)):pe.set(n,e)}function Mn(n){if(oe&&oe.length&&!e.fileExtensionIs(n,".d.ts")&&e.fileExtensionIsOneOf(n,e.supportedTSExtensions)){var t=On(n);if(t){var r=t.commandLine.options.outFile||t.commandLine.options.out;return r?e.changeExtension(r,".d.ts"):e.getOutputDeclarationFileName(n,t.commandLine)}}}function On(n){void 0===ce&&(ce=e.createMap(),Pn(function(e,n){e&&Pe(E.configFilePath)!==n&&e.commandLine.fileNames.forEach(function(e){return ce.set(Pe(e),n)})}));var t=ce.get(Pe(n));return t&&Fn(t)}function Pn(e){return Rn(S,oe,function(n,t,r){var a=(r?r.commandLine.projectReferences:S)[t],i=Pe(M(a));return e(n,i)})}function Rn(n,t,r,a){var i;return o(n,t,void 0,r,a);function o(n,t,r,a,s){if(s){var c=s(n,r);if(c)return c}return e.forEach(t,function(n,t){if(!e.contains(i,n)){var c=a(n,t,r);if(c)return c;if(n)return(i||(i=[])).push(n),o(n.commandLine.projectReferences,n.references,n,a,s)}})}}function Fn(e){if(se)return se.get(e)||void 0}function Gn(n,t){e.forEach(n.referencedFiles,function(e){var r=a(e.fileName,n.originalFileName);Ln(r,t,!1,void 0,n,e.pos,e.end)})}function Vn(n){var t=e.map(n.typeReferenceDirectives,function(e){return e.fileName.toLocaleLowerCase()});if(t)for(var r=W(t,n.originalFileName,On(n.originalFileName)),a=0;a<t.length;a++){var i=n.typeReferenceDirectives[a],o=r[a],s=i.fileName.toLocaleLowerCase();e.setResolvedTypeReferenceDirective(n,s,o),Bn(s,o,n,i.pos,i.end)}}function Bn(n,t,r,a,i){var o=I.get(n);if(!o||!o.primary){var s=!0;if(t){if(t.isExternalLibraryImport&&V++,t.primary)Ln(t.resolvedFileName,!1,!1,t.packageId,r,a,i);else if(o){if(t.resolvedFileName!==o.resolvedFileName){var c=q.readFile(t.resolvedFileName);c!==$e(o.resolvedFileName).text&&F.add(jn(r,a,i,e.Diagnostics.Conflicting_definitions_for_0_found_at_1_and_2_Consider_installing_a_specific_version_of_this_library_to_resolve_the_conflict,n,t.resolvedFileName,o.resolvedFileName))}s=!1}else Ln(t.resolvedFileName,!1,!1,t.packageId,r,a,i);t.isExternalLibraryImport&&V--}else F.add(jn(r,a,i,e.Diagnostics.Cannot_find_type_definition_file_for_0,n));s&&I.set(n,t)}}function Kn(n){e.forEach(n.libReferenceDirectives,function(t){var r=t.fileName.toLocaleLowerCase(),a=e.libMap.get(r);if(a)En(e.combinePaths(Y,a),!0,!0);else{var i=e.removeSuffix(e.removePrefix(r,"lib."),".d.ts"),o=e.getSpellingSuggestion(i,e.libs,e.identity),s=o?e.Diagnostics.Cannot_find_lib_definition_for_0_Did_you_mean_1:e.Diagnostics.Cannot_find_lib_definition_for_0;F.add(jn(n,t.pos,t.end,s,r,o))}})}function jn(n,t,r,a){for(var i=[],o=4;o<arguments.length;o++)i[o-4]=arguments[o];return void 0===n||void 0===t||void 0===r?e.createCompilerDiagnostic.apply(void 0,[a].concat(i)):e.createFileDiagnostic.apply(void 0,[n,t,r-t,a].concat(i))}function Un(e){return q.getCanonicalFileName(e)}function Hn(n){if(xn(n),n.imports.length||n.moduleAugmentations.length){var t=R(n),r=Ge(t,e.getNormalizedAbsolutePath(n.originalFileName,Z),n);e.Debug.assert(r.length===t.length);for(var a=0;a<t.length;a++){var i=r[a];if(e.setResolvedModule(n,t[a],i),i){var o=i.isExternalLibraryImport,s=!e.resolutionExtensionIsTSOrJson(i.extension),c=o&&s,l=i.resolvedFileName;o&&V++;var u=c&&V>G,d=l&&!P(E,i)&&!E.noResolve&&a<n.imports.length&&!u&&!(s&&!E.allowJs)&&(e.isInJSFile(n.imports[a])||!(2097152&n.imports[a].flags));if(u)B.set(n.path,!0);else if(d){var p=Pe(l),m=e.skipTrivia(n.text,n.imports[a].pos);In(l,p,!1,!1,n,m,n.imports[a].end,i.packageId)}o&&V--}}}else n.resolvedModules=void 0}function Wn(n){var t=e.mapDefined(n,function(e){return e.isDeclarationFile?void 0:e.fileName});return i(t,Z,Un)}function qn(n,t){for(var r=!0,a=q.getCanonicalFileName(e.getNormalizedAbsolutePath(t,Z)),i=0,o=n;i<o.length;i++){var s=o[i];if(!s.isDeclarationFile){var c=q.getCanonicalFileName(e.getNormalizedAbsolutePath(s.fileName,Z));0!==c.indexOf(a)&&(Q.add(e.createCompilerDiagnostic(e.Diagnostics.File_0_is_not_under_rootDir_1_rootDir_is_expected_to_contain_all_source_files,s.fileName,t)),r=!1)}}return r}function zn(n){se||(se=e.createMap());var t=M(n),r=Pe(t),a=se.get(r);if(void 0!==a)return a||void 0;var i=e.getNormalizedAbsolutePath(e.getDirectoryPath(t),q.getCurrentDirectory()),o=q.getSourceFile(t,100);if(wn(o,r,void 0),void 0!==o){o.path=r,o.resolvedPath=r,o.originalFileName=t;var s=e.parseJsonSourceFileConfigFileContent(o,z,i,void 0,t),c={commandLine:s,sourceFile:o};return se.set(r,c),s.projectReferences&&(c.references=s.projectReferences.map(zn)),c}se.set(r,!1)}function Jn(){if(E.strictPropertyInitialization&&!e.getStrictOptionValue(E,"strictNullChecks")&&et(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1,"strictPropertyInitialization","strictNullChecks"),E.isolatedModules&&(e.getEmitDeclarations(E)&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,O(E),"isolatedModules"),E.noEmitOnError&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"noEmitOnError","isolatedModules"),E.out&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"out","isolatedModules"),E.outFile&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"outFile","isolatedModules")),E.inlineSourceMap&&(E.sourceMap&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"sourceMap","inlineSourceMap"),E.mapRoot&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"mapRoot","inlineSourceMap")),E.paths&&void 0===E.baseUrl&&et(e.Diagnostics.Option_paths_cannot_be_used_without_specifying_baseUrl_option,"paths"),E.composite&&!1===E.declaration&&et(e.Diagnostics.Composite_projects_may_not_disable_declaration_emit,"declaration"),Xn(),E.composite){var n=f.filter(function(e){return!e.isDeclarationFile});if(b.length<n.length)for(var t=b.map(function(n){return e.normalizePath(n).toLowerCase()}),r=0,a=n.map(function(n){return e.normalizePath(n.path).toLowerCase()});r<a.length;r++){var i=a[r];-1===t.indexOf(i)&&Q.add(e.createCompilerDiagnostic(e.Diagnostics.File_0_is_not_in_project_file_list_Projects_must_list_all_files_or_use_an_include_pattern,i))}}if(E.paths)for(var o in E.paths)if(e.hasProperty(E.paths,o))if(e.hasZeroOrOneAsteriskCharacter(o)||Qn(!0,o,e.Diagnostics.Pattern_0_can_have_at_most_one_Asterisk_character,o),e.isArray(E.paths[o])){var s=E.paths[o].length;0===s&&Qn(!1,o,e.Diagnostics.Substitutions_for_pattern_0_shouldn_t_be_an_empty_array,o);for(var c=0;c<s;c++){var l=E.paths[o][c],u=typeof l;"string"===u?e.hasZeroOrOneAsteriskCharacter(l)||Yn(o,c,e.Diagnostics.Substitution_0_in_pattern_1_in_can_have_at_most_one_Asterisk_character,l,o):Yn(o,c,e.Diagnostics.Substitution_0_for_pattern_1_has_incorrect_type_expected_string_got_2,l,o,u)}}else Qn(!1,o,e.Diagnostics.Substitutions_for_pattern_0_should_be_an_array,o);E.sourceMap||E.inlineSourceMap||(E.inlineSources&&et(e.Diagnostics.Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided,"inlineSources"),E.sourceRoot&&et(e.Diagnostics.Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided,"sourceRoot")),E.out&&E.outFile&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"out","outFile"),!E.mapRoot||E.sourceMap||E.declarationMap||et(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2,"mapRoot","sourceMap","declarationMap"),E.declarationDir&&(e.getEmitDeclarations(E)||et(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2,"declarationDir","declaration","composite"),(E.out||E.outFile)&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"declarationDir",E.out?"out":"outFile")),E.declarationMap&&!e.getEmitDeclarations(E)&&et(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2,"declarationMap","declaration","composite"),E.lib&&E.noLib&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"lib","noLib"),E.noImplicitUseStrict&&e.getStrictOptionValue(E,"alwaysStrict")&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"noImplicitUseStrict","alwaysStrict");var d=E.target||0,p=E.outFile||E.out,m=e.find(f,function(n){return e.isExternalModule(n)&&!n.isDeclarationFile});if(E.isolatedModules){E.module===e.ModuleKind.None&&d<2&&et(e.Diagnostics.Option_isolatedModules_can_only_be_used_when_either_option_module_is_provided_or_option_target_is_ES2015_or_higher,"isolatedModules","target");var _=e.find(f,function(n){return!e.isExternalModule(n)&&!n.isDeclarationFile&&6!==n.scriptKind});if(_){var g=e.getErrorSpanForNode(_,_);Q.add(e.createFileDiagnostic(_,g.start,g.length,e.Diagnostics.Cannot_compile_namespaces_when_the_isolatedModules_flag_is_provided))}}else if(m&&d<2&&E.module===e.ModuleKind.None){g=e.getErrorSpanForNode(m,m.externalModuleIndicator);Q.add(e.createFileDiagnostic(m,g.start,g.length,e.Diagnostics.Cannot_use_imports_exports_or_module_augmentations_when_module_is_none))}if(p&&!E.emitDeclarationOnly)if(E.module&&E.module!==e.ModuleKind.AMD&&E.module!==e.ModuleKind.System)et(e.Diagnostics.Only_amd_and_system_modules_are_supported_alongside_0,E.out?"out":"outFile","module");else if(void 0===E.module&&m){g=e.getErrorSpanForNode(m,m.externalModuleIndicator);Q.add(e.createFileDiagnostic(m,g.start,g.length,e.Diagnostics.Cannot_compile_modules_using_option_0_unless_the_module_flag_is_amd_or_system,E.out?"out":"outFile"))}if(E.resolveJsonModule&&(e.getEmitModuleResolutionKind(E)!==e.ModuleResolutionKind.NodeJs?et(e.Diagnostics.Option_resolveJsonModule_cannot_be_specified_without_node_module_resolution_strategy,"resolveJsonModule"):e.hasJsonModuleEmitEnabled(E)||et(e.Diagnostics.Option_resolveJsonModule_can_only_be_specified_when_module_code_generation_is_commonjs_amd_es2015_or_esNext,"resolveJsonModule","module")),E.outDir||E.sourceRoot||E.mapRoot){var v=Re();E.outDir&&""===v&&f.some(function(n){return e.getRootLength(n.fileName)>1})&&et(e.Diagnostics.Cannot_find_the_common_subdirectory_path_for_the_input_files,"outDir")}if(!E.noEmit&&E.allowJs&&e.getEmitDeclarations(E)&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"allowJs",O(E)),E.checkJs&&!E.allowJs&&Q.add(e.createCompilerDiagnostic(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1,"checkJs","allowJs")),E.emitDeclarationOnly&&(e.getEmitDeclarations(E)||et(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2,"emitDeclarationOnly","declaration","composite"),E.noEmit&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"emitDeclarationOnly","noEmit")),E.emitDecoratorMetadata&&!E.experimentalDecorators&&et(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1,"emitDecoratorMetadata","experimentalDecorators"),E.jsxFactory?(E.reactNamespace&&et(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"reactNamespace","jsxFactory"),e.parseIsolatedEntityName(E.jsxFactory,d)||nt("jsxFactory",e.Diagnostics.Invalid_value_for_jsxFactory_0_is_not_a_valid_identifier_or_qualified_name,E.jsxFactory)):E.reactNamespace&&!e.isIdentifierText(E.reactNamespace,d)&&nt("reactNamespace",e.Diagnostics.Invalid_value_for_reactNamespace_0_is_not_a_valid_identifier,E.reactNamespace),!E.noEmit&&!E.suppressOutputPathCheck){var y=Ke(),h=e.createMap();e.forEachEmittedFile(y,function(e){E.emitDeclarationOnly||T(e.jsFilePath,h),T(e.declarationFilePath,h)})}function T(n,t){if(n){var r,a=Pe(n);if(pe.has(a))E.configFilePath||(r=e.chainDiagnosticMessages(void 0,e.Diagnostics.Adding_a_tsconfig_json_file_will_help_organize_projects_that_contain_both_TypeScript_and_JavaScript_files_Learn_more_at_https_Colon_Slash_Slashaka_ms_Slashtsconfig)),r=e.chainDiagnosticMessages(r,e.Diagnostics.Cannot_write_file_0_because_it_would_overwrite_input_file,n),ot(n,e.createCompilerDiagnosticFromMessageChain(r));var i=q.useCaseSensitiveFileNames()?a:a.toLocaleLowerCase();t.has(i)?ot(n,e.createCompilerDiagnostic(e.Diagnostics.Cannot_write_file_0_because_it_would_be_overwritten_by_multiple_input_files,n)):t.set(i,!0)}}}function Xn(){Rn(S,oe,function(n,t,r){var a=(r?r.commandLine.projectReferences:S)[t],i=r&&r.sourceFile;if(n){var o=n.commandLine.options;if(!o.composite){var s=r?r.commandLine.fileNames:b;s.length&&tt(i,t,e.Diagnostics.Referenced_project_0_must_have_setting_composite_Colon_true,a.path)}if(a.prepend){var c=o.outFile||o.out;c?q.fileExists(c)||tt(i,t,e.Diagnostics.Output_file_0_from_project_1_does_not_exist,c,a.path):tt(i,t,e.Diagnostics.Cannot_prepend_project_0_because_it_does_not_have_outFile_set,a.path)}}else tt(i,t,e.Diagnostics.File_0_not_found,a.path)})}function Yn(n,t,r,a,i,o){for(var s=!0,c=$n(),l=0,u=c;l<u.length;l++){var d=u[l];if(e.isObjectLiteralExpression(d.initializer))for(var p=0,m=e.getPropertyAssignment(d.initializer,n);p<m.length;p++){var f=m[p],_=f.initializer;e.isArrayLiteralExpression(_)&&_.elements.length>t&&(Q.add(e.createDiagnosticForNodeInSourceFile(E.configFile,_.elements[t],r,a,i,o)),s=!1)}}s&&Q.add(e.createCompilerDiagnostic(r,a,i,o))}function Qn(n,t,r,a){for(var i=!0,o=$n(),s=0,c=o;s<c.length;s++){var l=c[s];e.isObjectLiteralExpression(l.initializer)&&it(l.initializer,n,t,void 0,r,a)&&(i=!1)}i&&Q.add(e.createCompilerDiagnostic(r,a))}function Zn(n){var t=at();if(t)return e.getPropertyAssignment(t,n)}function $n(){return Zn("paths")||e.emptyArray}function et(e,n,t,r){rt(!0,n,t,e,n,t,r)}function nt(e,n,t){rt(!1,e,void 0,n,t)}function tt(n,t,r,a,i){var o=e.firstDefined(e.getTsConfigPropArray(n||E.configFile,"references"),function(n){return e.isArrayLiteralExpression(n.initializer)?n.initializer:void 0});o&&o.elements.length>t?Q.add(e.createDiagnosticForNodeInSourceFile(n||E.configFile,o.elements[t],r,a,i)):Q.add(e.createCompilerDiagnostic(r,a,i))}function rt(n,t,r,a,i,o,s){var c=at(),l=!c||!it(c,n,t,r,a,i,o,s);l&&Q.add(e.createCompilerDiagnostic(a,i,o,s))}function at(){if(void 0===j){j=null;var n=e.getTsConfigObjectLiteralExpression(E.configFile);if(n)for(var t=0,r=e.getPropertyAssignment(n,"compilerOptions");t<r.length;t++){var a=r[t];if(e.isObjectLiteralExpression(a.initializer)){j=a.initializer;break}}}return j}function it(n,t,r,a,i,o,s,c){for(var l=e.getPropertyAssignment(n,r,a),u=0,d=l;u<d.length;u++){var p=d[u];Q.add(e.createDiagnosticForNodeInSourceFile(E.configFile,t?p.name:p.initializer,i,o,s,c))}return!!l.length}function ot(e,n){ne.set(Pe(e),!0),Q.add(n)}function st(n){if(E.noEmit)return!1;var t=Pe(n);if(en(t))return!1;var r=E.outFile||E.out;if(r)return ct(t,r)||ct(t,e.removeFileExtension(r)+".d.ts");if(E.declarationDir&&e.containsPath(E.declarationDir,t,Z,!q.useCaseSensitiveFileNames()))return!0;if(E.outDir)return e.containsPath(E.outDir,t,Z,!q.useCaseSensitiveFileNames());if(e.fileExtensionIsOneOf(t,e.supportedJSExtensions)||e.fileExtensionIs(t,".d.ts")){var a=e.removeFileExtension(t);return!!en(a+".ts")||!!en(a+".tsx")}return!1}function ct(n,t){return 0===e.comparePaths(n,t,Z,!q.useCaseSensitiveFileNames())}}function w(n,t){return void 0===t&&(t=n),{fileExists:function(e){return t.fileExists(e)},readDirectory:function(n,r,a,i,o){return e.Debug.assertDefined(t.readDirectory,"'CompilerHost.readDirectory' must be implemented to correctly process 'projectReferences'"),t.readDirectory(n,r,a,i,o)},readFile:function(e){return t.readFile(e)},useCaseSensitiveFileNames:n.useCaseSensitiveFileNames(),getCurrentDirectory:function(){return n.getCurrentDirectory()},onUnRecoverableConfigFileDiagnostic:n.onUnRecoverableConfigFileDiagnostic||function(){},trace:n.trace?function(e){return n.trace(e)}:void 0}}function M(n,t){var r=t||n;return e.resolveConfigFileProjectName(r.path)}function O(e){return e.declaration?"declaration":"composite"}function P(n,t){var r=t.extension;switch(r){case".ts":case".d.ts":return;case".tsx":return a();case".jsx":return a()||i();case".js":return i();case".json":return o()}function a(){return n.jsx?void 0:e.Diagnostics.Module_0_was_resolved_to_1_but_jsx_is_not_set}function i(){return n.allowJs||!e.getStrictOptionValue(n,"noImplicitAny")?void 0:e.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type}function o(){return n.resolveJsonModule?void 0:e.Diagnostics.Module_0_was_resolved_to_1_but_resolveJsonModule_is_not_used}}function R(e){for(var n=e.imports,t=e.moduleAugmentations,r=n.map(function(e){return e.text}),a=0,i=t;a<i.length;a++){var o=i[a];10===o.kind&&r.push(o.text)}return r}e.formatColorAndReset=b,e.formatLocation=S,e.formatDiagnosticsWithColorAndContext=x,e.flattenDiagnosticMessageText=C,e.isProgramUptoDate=D,e.getConfigFileParsingDiagnostics=L,e.createProgram=I,e.parseConfigHostFromCompilerHostLike=w,e.resolveProjectReferencePath=M,e.getResolutionDiagnostic=P}(u||(u={})),function(e){function n(e,n,t,r,a){var i=[],o=e.emit(n,s,r,t,a);return{outputFiles:i,emitSkipped:o.emitSkipped,exportedModulesFromDeclarationEmit:o.exportedModulesFromDeclarationEmit};function s(e,n,t){i.push({name:e,writeByteOrderMark:t,text:n})}}function t(n){return n?e.cloneMap(n):void 0}e.getFileEmitOutput=n,e.cloneMapOrUndefined=t}(u||(u={})),function(e){(function(n){function t(n){if(n.declarations&&n.declarations[0]){var t=e.getSourceFileOfNode(n.declarations[0]);return t&&t.resolvedPath}}function r(e,n){var r=e.getSymbolAtLocation(n);return r&&t(r)}function a(n,t,r,a){return e.toPath(n.getProjectReferenceRedirect(t)||t,r,a)}function i(n,t,i){var o;if(t.imports&&t.imports.length>0)for(var s=n.getTypeChecker(),c=0,l=t.imports;c<l.length;c++){var u=l[c],d=r(s,u);d&&C(d)}var p=e.getDirectoryPath(t.path);if(t.referencedFiles&&t.referencedFiles.length>0)for(var m=0,f=t.referencedFiles;m<f.length;m++){var _=f[m],g=a(n,_.fileName,p,i);C(g)}if(t.resolvedTypeReferenceDirectiveNames&&t.resolvedTypeReferenceDirectiveNames.forEach(function(e){if(e){var t=e.resolvedFileName,r=a(n,t,p,i);C(r)}}),t.moduleAugmentations.length){s=n.getTypeChecker();for(var v=0,y=t.moduleAugmentations;v<y.length;v++){var h=y[v];if(e.isStringLiteral(h)){var b=s.getSymbolAtLocation(h);b&&x(b)}}}for(var E=0,T=n.getTypeChecker().getAmbientModules();E<T.length;E++){var S=T[E];S.declarations.length>1&&x(S)}return o;function x(n){for(var r=0,a=n.declarations;r<a.length;r++){var i=a[r],o=e.getSourceFileOfNode(i);o&&o!==t&&C(o.resolvedPath)}}function C(n){o||(o=e.createMap()),o.set(n,!0)}}function s(e,n){return n&&!n.referencedMap===!e}function c(n,t,r){for(var a=e.createMap(),o=n.getCompilerOptions().module!==e.ModuleKind.None?e.createMap():void 0,c=o?e.createMap():void 0,l=e.createMap(),u=s(o,r),d=0,p=n.getSourceFiles();d<p.length;d++){var m=p[d],f=m.version,_=u?r.fileInfos.get(m.path):void 0;if(o){var g=i(n,m,t);if(g&&o.set(m.path,g),u){var v=r.exportedModulesMap.get(m.path);v&&c.set(m.path,v)}}a.set(m.path,{version:f,signature:_&&_.signature})}return{fileInfos:a,referencedMap:o,exportedModulesMap:c,hasCalledUpdateShapeSignature:l}}function l(e){e.allFilesExcludingDefaultLibraryFile=void 0,e.allFileNames=void 0}function u(n){var t=e.createMap();return n.fileInfos.forEach(function(e,n){t.set(n,o({},e))}),{fileInfos:t,referencedMap:e.cloneMapOrUndefined(n.referencedMap),exportedModulesMap:e.cloneMapOrUndefined(n.exportedModulesMap),hasCalledUpdateShapeSignature:e.cloneMap(n.hasCalledUpdateShapeSignature)}}function d(n,t,r,a,i,o,s){var c=o||e.createMap(),l=t.getSourceFileByPath(r);if(!l)return e.emptyArray;if(!m(n,t,l,c,a,i,s))return[l];var u=(n.referencedMap?x:S)(n,t,l,c,a,i,s);return o||p(n,c),u}function p(e,n){n.forEach(function(n,t){e.fileInfos.get(t).signature=n,e.hasCalledUpdateShapeSignature.set(t,!0)})}function m(n,t,r,a,i,o,s){if(e.Debug.assert(!!r),e.Debug.assert(!s||!!n.exportedModulesMap,"Compute visible to outside map only if visibleToOutsideReferencedMap present in the state"),n.hasCalledUpdateShapeSignature.has(r.path)||a.has(r.path))return!1;var c=n.fileInfos.get(r.path);if(!c)return e.Debug.fail();var l,u=c.signature;if(r.isDeclarationFile){if(l=r.version,s&&l!==u){var d=n.referencedMap?n.referencedMap.get(r.path):void 0;s.set(r.path,d||!1)}}else{var p=e.getFileEmitOutput(t,r,!0,i);p.outputFiles&&p.outputFiles.length>0?(l=o(p.outputFiles[0].text),s&&l!==u&&f(r,p.exportedModulesFromDeclarationEmit,s)):l=u}return a.set(r.path,l),!u||l!==u}function f(n,r,a){var i;r?(r.forEach(function(e){return o(t(e))}),a.set(n.path,i||!1)):a.set(n.path,!1);function o(n){n&&(i||(i=e.createMap()),i.set(n,!0))}}function _(n,t){t&&(e.Debug.assert(!!n.exportedModulesMap),t.forEach(function(e,t){e?n.exportedModulesMap.set(t,e):n.exportedModulesMap.delete(t)}))}function g(n,t,r){var a,i=t.getCompilerOptions();if(i.outFile||i.out)return v(n,t);if(!n.referencedMap||E(r))return v(n,t);var o=e.createMap(),s=[r.path];while(s.length){var c=s.pop();if(!o.has(c)){o.set(c,!0);var l=n.referencedMap.get(c);if(l)for(var u=l.keys(),d=u.next(),p=d.value,m=d.done;!m;a=u.next(),p=a.value,m=a.done,a)s.push(p)}}return e.arrayFrom(e.mapDefinedIterator(o.keys(),function(e){var n=t.getSourceFileByPath(e);return n?n.fileName:e}))}function v(n,t){if(!n.allFileNames){var r=t.getSourceFiles();n.allFileNames=r===e.emptyArray?e.emptyArray:r.map(function(e){return e.fileName})}return n.allFileNames}function y(n,t){return e.arrayFrom(e.mapDefinedIterator(n.referencedMap.entries(),function(e){var n=e[0],r=e[1];return r.has(t)?n:void 0}))}function h(n){for(var t=0,r=n.statements;t<r.length;t++){var a=r[t];if(!e.isModuleWithStringLiteralName(a))return!1}return!0}function b(n){return e.some(n.moduleAugmentations,function(n){return e.isGlobalScopeAugmentation(n.parent)})}function E(n){return b(n)||!e.isExternalModule(n)&&!h(n)}function T(n,t,r){if(n.allFilesExcludingDefaultLibraryFile)return n.allFilesExcludingDefaultLibraryFile;var a;c(r);for(var i=0,o=t.getSourceFiles();i<o.length;i++){var s=o[i];s!==r&&c(s)}return n.allFilesExcludingDefaultLibraryFile=a||e.emptyArray,n.allFilesExcludingDefaultLibraryFile;function c(e){t.isSourceFileDefaultLibrary(e)||(a||(a=[])).push(e)}}function S(e,n,t){var r=n.getCompilerOptions();return r&&(r.out||r.outFile)?[t]:T(e,n,t)}function x(n,t,r,a,i,o,s){if(E(r))return T(n,t,r);var c=t.getCompilerOptions();if(c&&(c.isolatedModules||c.out||c.outFile))return[r];var l=e.createMap();l.set(r.path,r);var u=y(n,r.resolvedPath);while(u.length>0){var d=u.pop();if(!l.has(d)){var p=t.getSourceFileByPath(d);l.set(d,p),p&&m(n,t,p,a,i,o,s)&&u.push.apply(u,y(n,p.resolvedPath))}}return e.arrayFrom(e.mapDefinedIterator(l.values(),function(e){return e}))}n.canReuseOldState=s,n.create=c,n.releaseCache=l,n.clone=u,n.getFilesAffectedBy=d,n.updateSignaturesFromCache=p,n.updateExportedFilesMapFromCache=_,n.getAllDependencies=g})(e.BuilderState||(e.BuilderState={}))}(u||(u={})),function(e){function n(n,t){return n===t||void 0!==n&&void 0!==t&&n.size===t.size&&!e.forEachKey(n,function(e){return!t.has(e)})}function t(t,r,a){var i=e.BuilderState.create(t,r,a);i.program=t;var o=t.getCompilerOptions();i.compilerOptions=o,o.outFile||o.out||o.isolatedModules||(i.semanticDiagnosticsPerFile=e.createMap()),i.changedFilesSet=e.createMap();var s=e.BuilderState.canReuseOldState(i.referencedMap,a),c=s?a.compilerOptions:void 0,l=s&&a.semanticDiagnosticsPerFile&&!!i.semanticDiagnosticsPerFile&&!e.compilerOptionsAffectSemanticDiagnostics(o,c);if(s){if(!a.currentChangedFilePath){var u=a.currentAffectedFilesSignatures;e.Debug.assert(!a.affectedFiles&&(!u||!u.size),"Cannot reuse if only few affected files of currentChangedFile were iterated")}l&&e.Debug.assert(!e.forEachKey(a.changedFilesSet,function(e){return a.semanticDiagnosticsPerFile.has(e)}),"Semantic diagnostics shouldnt be available for changed files"),e.copyEntries(a.changedFilesSet,i.changedFilesSet),o.outFile||o.out||!a.affectedFilesPendingEmit||(i.affectedFilesPendingEmit=a.affectedFilesPendingEmit,i.affectedFilesPendingEmitIndex=a.affectedFilesPendingEmitIndex)}var d=i.referencedMap,p=s?a.referencedMap:void 0,m=l&&!o.skipLibCheck===!c.skipLibCheck,f=m&&!o.skipDefaultLibCheck===!c.skipDefaultLibCheck;return i.fileInfos.forEach(function(r,o){var c,u;if(!s||!(c=a.fileInfos.get(o))||c.version!==r.version||!n(u=d&&d.get(o),p&&p.get(o))||u&&e.forEachKey(u,function(e){return!i.fileInfos.has(e)&&a.fileInfos.has(e)}))i.changedFilesSet.set(o,!0);else if(l){var _=t.getSourceFileByPath(o);if(_.isDeclarationFile&&!m)return;if(_.hasNoDefaultLib&&!f)return;var g=a.semanticDiagnosticsPerFile.get(o);g&&(i.semanticDiagnosticsPerFile.set(o,g),i.semanticDiagnosticsFromOldState||(i.semanticDiagnosticsFromOldState=e.createMap()),i.semanticDiagnosticsFromOldState.set(o,!0))}}),i}function r(n){e.BuilderState.releaseCache(n),n.program=void 0}function a(n){var t=e.BuilderState.clone(n);return t.semanticDiagnosticsPerFile=e.cloneMapOrUndefined(n.semanticDiagnosticsPerFile),t.changedFilesSet=e.cloneMap(n.changedFilesSet),t.affectedFiles=n.affectedFiles,t.affectedFilesIndex=n.affectedFilesIndex,t.currentChangedFilePath=n.currentChangedFilePath,t.currentAffectedFilesSignatures=e.cloneMapOrUndefined(n.currentAffectedFilesSignatures),t.currentAffectedFilesExportedModulesMap=e.cloneMapOrUndefined(n.currentAffectedFilesExportedModulesMap),t.seenAffectedFiles=e.cloneMapOrUndefined(n.seenAffectedFiles),t.cleanedDiagnosticsOfLibFiles=n.cleanedDiagnosticsOfLibFiles,t.semanticDiagnosticsFromOldState=e.cloneMapOrUndefined(n.semanticDiagnosticsFromOldState),t.program=n.program,t.compilerOptions=n.compilerOptions,t.affectedFilesPendingEmit=n.affectedFilesPendingEmit,t.affectedFilesPendingEmitIndex=n.affectedFilesPendingEmitIndex,t.seenEmittedFiles=e.cloneMapOrUndefined(n.seenEmittedFiles),t.programEmitComplete=n.programEmitComplete,t}function i(n,t){e.Debug.assert(!t||!n.affectedFiles||n.affectedFiles[n.affectedFilesIndex-1]!==t||!n.semanticDiagnosticsPerFile.has(t.path))}function o(n,t,r){while(1){var a=n.affectedFiles;if(a){var i=n.seenAffectedFiles,o=n.affectedFilesIndex;while(o<a.length){var s=a[o];if(!i.has(s.path))return n.affectedFilesIndex=o,c(n,s),s;i.set(s.path,!0),o++}n.changedFilesSet.delete(n.currentChangedFilePath),n.currentChangedFilePath=void 0,e.BuilderState.updateSignaturesFromCache(n,n.currentAffectedFilesSignatures),n.currentAffectedFilesSignatures.clear(),e.BuilderState.updateExportedFilesMapFromCache(n,n.currentAffectedFilesExportedModulesMap),n.affectedFiles=void 0}var l=n.changedFilesSet.keys().next();if(l.done)return;var u=e.Debug.assertDefined(n.program),d=u.getCompilerOptions();if(d.outFile||d.out)return e.Debug.assert(!n.semanticDiagnosticsPerFile),u;n.currentAffectedFilesSignatures=n.currentAffectedFilesSignatures||e.createMap(),n.exportedModulesMap&&(n.currentAffectedFilesExportedModulesMap=n.currentAffectedFilesExportedModulesMap||e.createMap()),n.affectedFiles=e.BuilderState.getFilesAffectedBy(n,u,l.value,t,r,n.currentAffectedFilesSignatures,n.currentAffectedFilesExportedModulesMap),n.currentChangedFilePath=l.value,n.affectedFilesIndex=0,n.seenAffectedFiles=n.seenAffectedFiles||e.createMap()}}function s(n){var t=n.affectedFilesPendingEmit;if(t){for(var r=n.seenEmittedFiles||(n.seenEmittedFiles=e.createMap()),a=n.affectedFilesPendingEmitIndex;a<t.length;a++){var i=e.Debug.assertDefined(n.program).getSourceFileByPath(t[a]);if(i&&!r.has(i.path))return n.affectedFilesPendingEmitIndex=a,i}n.affectedFilesPendingEmit=void 0,n.affectedFilesPendingEmitIndex=void 0}}function c(n,t){if(!d(n,t.path)){if(n.allFilesExcludingDefaultLibraryFile===n.affectedFiles&&!n.cleanedDiagnosticsOfLibFiles){n.cleanedDiagnosticsOfLibFiles=!0;var r=e.Debug.assertDefined(n.program),a=r.getCompilerOptions();if(e.forEach(r.getSourceFiles(),function(t){return r.isSourceFileDefaultLibrary(t)&&!e.skipTypeChecking(t,a)&&d(n,t.path)}))return}if(n.exportedModulesMap&&1!==n.affectedFiles.length&&n.changedFilesSet.has(t.path)){e.Debug.assert(!!n.currentAffectedFilesExportedModulesMap);var i=e.createMap();e.forEachEntry(n.currentAffectedFilesExportedModulesMap,function(e,r){return e&&e.has(t.path)&&l(n,r,i)})||e.forEachEntry(n.exportedModulesMap,function(e,r){return!n.currentAffectedFilesExportedModulesMap.has(r)&&e.has(t.path)&&l(n,r,i)})}}}function l(n,t,r){return e.forEachEntry(n.referencedMap,function(e,a){return e.has(t)&&u(n,a,r)})}function u(n,t,r){return!!e.addToSeen(r,t)&&(!!d(n,t)||(e.Debug.assert(!!n.currentAffectedFilesExportedModulesMap),!!e.forEachEntry(n.currentAffectedFilesExportedModulesMap,function(e,a){return e&&e.has(t)&&u(n,a,r)})||(!!e.forEachEntry(n.exportedModulesMap,function(e,a){return!n.currentAffectedFilesExportedModulesMap.has(a)&&e.has(t)&&u(n,a,r)})||!!e.forEachEntry(n.referencedMap,function(e,a){return e.has(t)&&!r.has(a)&&d(n,a)}))))}function d(e,n){return!e.semanticDiagnosticsFromOldState||(e.semanticDiagnosticsFromOldState.delete(n),e.semanticDiagnosticsPerFile.delete(n),!e.semanticDiagnosticsFromOldState.size)}function p(e,n,t){n===e.program?(e.changedFilesSet.clear(),e.programEmitComplete=!0):(e.seenAffectedFiles.set(n.path,!0),t?e.affectedFilesPendingEmitIndex++:e.affectedFilesIndex++)}function m(e,n,t,r){return p(e,t,r),{result:n,affected:t}}function f(n,t,r){var a=t.path;if(n.semanticDiagnosticsPerFile){var i=n.semanticDiagnosticsPerFile.get(a);if(i)return i}var o=e.Debug.assertDefined(n.program).getSemanticDiagnostics(t,r);return n.semanticDiagnosticsPerFile&&n.semanticDiagnosticsPerFile.set(a,o),o}var _;function g(n,t,r,a,i,o){var s,c,l;return void 0===n?(e.Debug.assert(void 0===t),s=r,l=a,e.Debug.assert(!!l),c=l.getProgram()):e.isArray(n)?(l=a,c=e.createProgram({rootNames:n,options:t,host:r,oldProgram:l&&l.getProgramOrUndefined(),configFileParsingDiagnostics:i,projectReferences:o}),s=r):(c=n,s=t,l=r,i=a),{host:s,newProgram:c,oldProgram:l,configFileParsingDiagnostics:i||e.emptyArray}}function v(n,c){var l=c.newProgram,u=c.host,d=c.oldProgram,g=c.configFileParsingDiagnostics,v=d&&d.getState();if(v&&l===v.program&&g===l.getConfigFileParsingDiagnostics())return l=void 0,v=void 0,d;var h,b=e.createGetCanonicalFileName(u.useCaseSensitiveFileNames()),E=u.createHash||e.generateDjb2Hash,T=t(l,b,v);l=void 0,d=void 0,v=void 0;var S=y(T,g);return S.getState=function(){return T},S.backupState=function(){e.Debug.assert(void 0===h),h=a(T)},S.restoreState=function(){T=e.Debug.assertDefined(h),h=void 0},S.getAllDependencies=function(n){return e.BuilderState.getAllDependencies(T,e.Debug.assertDefined(T.program),n)},S.getSemanticDiagnostics=D,S.emit=C,S.releaseProgram=function(){r(T),h=void 0},n===_.SemanticDiagnosticsBuilderProgram?S.getSemanticDiagnosticsOfNextAffectedFile=A:n===_.EmitAndSemanticDiagnosticsBuilderProgram?S.emitNextAffectedFile=x:e.notImplemented(),S;function x(n,t,r,a){var i=o(T,t,E),c=!1;if(!i)if(T.compilerOptions.out||T.compilerOptions.outFile){var l=e.Debug.assertDefined(T.program);if(T.programEmitComplete||!e.some(l.getProjectReferences(),function(e){return!!e.prepend}))return void(T.programEmitComplete=!0);i=l}else{if(i=s(T),!i)return;c=!0}return T.affectedFilesPendingEmit&&T.program!==i&&(T.seenEmittedFiles||(T.seenEmittedFiles=e.createMap())).set(i.path,!0),m(T,e.Debug.assertDefined(T.program).emit(i===T.program?void 0:i,n||u.writeFile,t,r,a),i,c)}function C(t,r,a,o,s){if(n===_.EmitAndSemanticDiagnosticsBuilderProgram&&(i(T,t),!t)){var c=[],l=!1,d=void 0,p=[],m=void 0;while(m=x(r,a,o,s))l=l||m.result.emitSkipped,d=e.addRange(d,m.result.diagnostics),p=e.addRange(p,m.result.emittedFiles),c=e.addRange(c,m.result.sourceMaps);return{emitSkipped:l,diagnostics:d||e.emptyArray,emittedFiles:p,sourceMaps:c}}return e.Debug.assertDefined(T.program).emit(t,r||u.writeFile,a,o,s)}function A(e,n){while(1){var t=o(T,e,E);if(!t)return;if(t===T.program)return m(T,T.program.getSemanticDiagnostics(void 0,e),t);if(!n||!n(t))return m(T,f(T,t,e),t);p(T,t)}}function D(t,r){i(T,t);var a,s,c,l=e.Debug.assertDefined(T.program).getCompilerOptions();if(l.outFile||l.out)return e.Debug.assert(!T.semanticDiagnosticsPerFile),e.Debug.assertDefined(T.program).getSemanticDiagnostics(t,r);if(t)return f(T,t,r);while(a=o(T,r,E))a!==T.program&&n===_.EmitAndSemanticDiagnosticsBuilderProgram&&(s||(s=[])).push(a.path),p(T,a);s&&(T.affectedFilesPendingEmit=e.concatenate(T.affectedFilesPendingEmit,s),void 0===T.affectedFilesPendingEmitIndex&&(T.affectedFilesPendingEmitIndex=0));for(var u=0,d=e.Debug.assertDefined(T.program).getSourceFiles();u<d.length;u++){var m=d[u];c=e.addRange(c,f(T,m,r))}return c||e.emptyArray}}function y(n,t){return{getState:e.notImplemented,backupState:e.noop,restoreState:e.noop,getProgram:r,getProgramOrUndefined:function(){return n.program},releaseProgram:function(){return n.program=void 0},getCompilerOptions:function(){return n.compilerOptions},getSourceFile:function(e){return r().getSourceFile(e)},getSourceFiles:function(){return r().getSourceFiles()},getOptionsDiagnostics:function(e){return r().getOptionsDiagnostics(e)},getGlobalDiagnostics:function(e){return r().getGlobalDiagnostics(e)},getConfigFileParsingDiagnostics:function(){return t},getSyntacticDiagnostics:function(e,n){return r().getSyntacticDiagnostics(e,n)},getDeclarationDiagnostics:function(e,n){return r().getDeclarationDiagnostics(e,n)},getSemanticDiagnostics:function(e,n){return r().getSemanticDiagnostics(e,n)},emit:function(e,n,t,a,i){return r().emit(e,n,t,a,i)},getAllDependencies:e.notImplemented,getCurrentDirectory:function(){return r().getCurrentDirectory()}};function r(){return e.Debug.assertDefined(n.program)}}(function(e){e[e["SemanticDiagnosticsBuilderProgram"]=0]="SemanticDiagnosticsBuilderProgram",e[e["EmitAndSemanticDiagnosticsBuilderProgram"]=1]="EmitAndSemanticDiagnosticsBuilderProgram"})(_=e.BuilderProgramKind||(e.BuilderProgramKind={})),e.getBuilderCreationParameters=g,e.createBuilderProgram=v,e.createRedirectedBuilderProgram=y}(u||(u={})),function(e){function n(n,t,r,a,i,o){return e.createBuilderProgram(e.BuilderProgramKind.SemanticDiagnosticsBuilderProgram,e.getBuilderCreationParameters(n,t,r,a,i,o))}function t(n,t,r,a,i,o){return e.createBuilderProgram(e.BuilderProgramKind.EmitAndSemanticDiagnosticsBuilderProgram,e.getBuilderCreationParameters(n,t,r,a,i,o))}function r(n,t,r,a,i,o){var s=e.getBuilderCreationParameters(n,t,r,a,i,o),c=s.newProgram,l=s.configFileParsingDiagnostics;return e.createRedirectedBuilderProgram({program:c,compilerOptions:c.getCompilerOptions()},l)}e.createSemanticDiagnosticsBuilderProgram=n,e.createEmitAndSemanticDiagnosticsBuilderProgram=t,e.createAbstractBuilder=r}(u||(u={})),function(e){function n(n){return e.stringContains(n,"/node_modules/.")}function t(t,r,a){var i,o,s,c=!1,l=e.createMultiMap(),u=e.memoize(function(){return t.getCurrentDirectory()}),d=t.getCachedDirectoryStructureHost(),p=e.createMap(),m=e.createCacheWithRedirects(),f=e.createCacheWithRedirects(),_=e.createModuleResolutionCacheWithMaps(m,f,u(),t.getCanonicalFileName),g=e.createMap(),v=e.createCacheWithRedirects(),y=[".ts",".tsx",".js",".jsx",".json"],h=e.createMap(),b=e.createMap(),E=r&&e.removeTrailingDirectorySeparator(e.getNormalizedAbsolutePath(r,u())),T=E&&t.toPath(E),S=e.createMap();return{startRecordingFilesWithChangedResolutions:L,finishRecordingFilesWithChangedResolutions:k,startCachingPerDirectoryResolution:w,finishCachingPerDirectoryResolution:M,resolveModuleNames:F,getResolvedModuleWithFailedLookupLocationsFromCache:G,resolveTypeReferenceDirectives:R,removeResolutionsFromProjectReferenceRedirects:ee,removeResolutionsOfFile:ne,invalidateResolutionOfFile:ie,setFilesWithInvalidatedNonRelativeUnresolvedImports:oe,createHasInvalidatedResolution:I,updateTypeRootsWatch:de,closeTypeRootsWatch:ce,clear:D};function x(e){return e.resolvedModule}function C(e){return e.resolvedTypeReferenceDirective}function A(n,t){return!(void 0===n||t.length<=n.length)&&(e.startsWith(t,n)&&t[n.length]===e.directorySeparator)}function D(){e.clearMap(b,e.closeFileWatcherOf),h.clear(),l.clear(),ce(),p.clear(),g.clear(),c=!1,w()}function L(){i=[]}function k(){var e=i;return i=void 0,e}function N(e){if(!s)return!1;var n=s.get(e);return!!n&&!!n.length}function I(n){if(c||n)return o=void 0,e.returnTrue;var t=o;return o=void 0,function(e){return!!t&&t.has(e)||N(e)}}function w(){m.clear(),f.clear(),v.clear(),l.forEach(J),l.clear()}function M(){c=!1,s=void 0,w(),b.forEach(function(e,n){0===e.refCount&&(b.delete(n),e.watcher.close())})}function O(n,r,a,i,o){var s=e.resolveModuleName(n,r,a,i,_,o);if(!t.getGlobalCache)return s;var c=t.getGlobalCache();if(void 0!==c&&!e.isExternalModuleNameRelative(n)&&(!s.resolvedModule||!e.extensionIsTS(s.resolvedModule.extension))){var l=e.loadModuleFromGlobalCache(n,t.projectName,a,i,c),u=l.resolvedModule,d=l.failedLookupLocations;if(u)return{resolvedModule:u,failedLookupLocations:e.addRange(s.failedLookupLocations,d)}}return s}function P(n,r,a,o,s,l,u,d,p,m){var f=t.toPath(r),_=o.get(f)||o.set(f,e.createMap()).get(f),g=e.getDirectoryPath(f),v=s.getOrCreateMapOfCacheRedirects(a),y=v.get(g);y||(y=e.createMap(),v.set(g,y));for(var h=[],b=t.getCompilationSettings(),E=m&&N(f),T=t.getCurrentProgram(),S=T&&T.getResolvedProjectReferenceToRedirect(r),x=S?!a||a.sourceFile.path!==S.sourceFile.path:!!a,C=e.createMap(),A=0,D=n;A<D.length;A++){var L=D[A],k=_.get(L);if(!C.has(L)&&c||x||!k||k.isInvalidated||E&&!e.isExternalModuleNameRelative(L)&&d(k)){var I=k,w=y.get(L);w?k=w:(k=l(L,r,b,t,a),y.set(L,k)),_.set(L,k),W(L,k),I&&Y(I),m&&i&&!M(I,k)&&(i.push(f),m=!1)}e.Debug.assert(void 0!==k&&!k.isInvalidated),C.set(L,!0),h.push(u(k))}return _.forEach(function(n,t){C.has(t)||e.contains(p,t)||(Y(n),_.delete(t))}),h;function M(e,n){if(e===n)return!0;if(!e||!n)return!1;var t=u(e),r=u(n);return t===r||!(!t||!r)&&t.resolvedFileName===r.resolvedFileName}}function R(n,t,r){return P(n,t,r,g,v,e.resolveTypeReferenceDirective,C,function(e){return void 0===e.resolvedTypeReferenceDirective},void 0,!1)}function F(n,t,r,i){return P(n,t,i,p,m,O,x,function(n){return!n.resolvedModule||!e.resolutionExtensionIsTSOrJson(n.resolvedModule.extension)},r,a)}function G(e,n){var r=p.get(t.toPath(n));return r&&r.get(e)}function V(n){return e.endsWith(n,"/node_modules")}function B(n){return e.endsWith(n,"/node_modules/@types")}function K(n){var t=e.getRootLength(n);if(n.length===t)return!1;var r=n.indexOf(e.directorySeparator,t);if(-1===r)return!1;if(47!==n.charCodeAt(0)&&-1===n.substr(t,r).search(/users/i))return!0;for(var a=r+1,i=2;i>0;i--)if(a=n.indexOf(e.directorySeparator,a)+1,0===a)return!1;return!0}function j(n,t){if(A(T,t)){n=e.isRootedDiskPath(n)?e.normalizePath(n):e.getNormalizedAbsolutePath(n,u()),e.Debug.assert(n.length===t.length,"FailedLookup: "+n+" failedLookupLocationPath: "+t);var r=t.indexOf(e.directorySeparator,T.length+1);return-1!==r?{dir:n.substr(0,r),dirPath:t.substr(0,r)}:{dir:E,dirPath:T,nonRecursive:!1}}return U(e.getDirectoryPath(e.getNormalizedAbsolutePath(n,u())),e.getDirectoryPath(t))}function U(n,t){while(e.pathContainsNodeModules(t))n=e.getDirectoryPath(n),t=e.getDirectoryPath(t);if(V(t))return K(e.getDirectoryPath(t))?{dir:n,dirPath:t}:void 0;var r,a,i=!0;if(void 0!==T)while(!A(t,T)){var o=e.getDirectoryPath(t);if(o===t)break;i=!1,r=t,a=n,t=o,n=e.getDirectoryPath(n)}return K(t)?{dir:a||n,dirPath:r||t,nonRecursive:i}:void 0}function H(n){return e.fileExtensionIsOneOf(n,y)}function W(n,t){t.failedLookupLocations&&t.failedLookupLocations.length&&(t.refCount?t.refCount++:(t.refCount=1,e.isExternalModuleNameRelative(n)?q(t):l.add(n,t)))}function q(n){e.Debug.assert(!!n.refCount);for(var r=n.failedLookupLocations,a=!1,i=0,o=r;i<o.length;i++){var s=o[i],c=t.toPath(s),l=j(s,c);if(l){var u=l.dir,d=l.dirPath,p=l.nonRecursive;if(!H(c)){var m=h.get(c)||0;h.set(c,m+1)}d===T?(e.Debug.assert(!p),a=!0):X(u,d,p)}}a&&X(E,T,!0)}function z(e){e.refCount=void 0}function J(e,n){var r=t.getCurrentProgram(),a=r&&r.getTypeChecker().tryFindAmbientModuleWithoutAugmentations(n)?z:q;e.forEach(a)}function X(n,t,r){var a=b.get(t);a?(e.Debug.assert(!!r===!!a.nonRecursive),a.refCount++):b.set(t,{watcher:Z(n,t,r),refCount:1,nonRecursive:r})}function Y(n){if(n.refCount&&(n.refCount--,!n.refCount)){for(var r=n.failedLookupLocations,a=!1,i=0,o=r;i<o.length;i++){var s=o[i],c=t.toPath(s),l=j(s,c);if(l){var u=l.dirPath,d=h.get(c);d&&(1===d?h.delete(c):(e.Debug.assert(d>1),h.set(c,d-1))),u===T?a=!0:Q(u)}}a&&Q(T)}}function Q(e){var n=b.get(e);n.refCount--}function Z(e,n,r){return t.watchDirectoryOfFailedLookupLocation(e,function(e){var r=t.toPath(e);d&&d.addOrDeleteFileOrDirectory(e,r),!c&&se(r,n===r)&&t.onInvalidatedResolution()},r?0:1)}function $(e,n){var t=e.get(n);t&&(t.forEach(Y),e.delete(n))}function ee(n){if(e.fileExtensionIs(n,".json")){var r=t.getCurrentProgram();if(r){var a=r.getResolvedProjectReferenceByPath(n);a&&a.commandLine.fileNames.forEach(function(e){return ne(t.toPath(e))})}}}function ne(e){$(p,e),$(g,e)}function te(n,t,r){var a=e.createMap();n.forEach(function(n,i){var s=e.getDirectoryPath(i),c=a.get(s);c||(c=e.createMap(),a.set(s,c)),n.forEach(function(n,a){c.has(a)||(c.set(a,!0),!n.isInvalidated&&t(n,r)&&(n.isInvalidated=!0,(o||(o=e.createMap())).set(i,!0)))})})}function re(){var n=t.maxNumberOfFilesToIterateForInvalidation||e.maxNumberOfFilesToIterateForInvalidation;return p.size>n||g.size>n}function ae(e){re()?c=!0:(te(p,e,x),te(g,e,C))}function ie(e){ne(e),ae(function(n,r){var a=r(n);return!!a&&t.toPath(a.resolvedFileName)===e})}function oe(n){e.Debug.assert(s===n||void 0===s),s=n}function se(r,a){var i;if(a)i=function(e){return A(r,t.toPath(e))};else{if(n(r))return!1;var s=e.getDirectoryPath(r);if(B(r)||V(r)||B(s)||V(s))i=function(n){var a=t.toPath(n);return a===r||e.startsWith(t.toPath(n),r)};else{if(!H(r)&&!h.has(r))return!1;if(e.isEmittedFileOfProgram(t.getCurrentProgram(),r))return!1;i=function(e){return t.toPath(e)===r}}}var l=function(n){return e.some(n.failedLookupLocations,i)},u=o&&o.size;return ae(l),c||o&&o.size!==u}function ce(){e.clearMap(S,e.closeFileWatcher)}function le(e,n){if(!c){if(A(T,n))return T;var t=U(e,n);return t&&b.has(t.dirPath)?t.dirPath:void 0}}function ue(e,n){return t.watchTypeRootsDirectory(n,function(r){var a=t.toPath(r);d&&d.addOrDeleteFileOrDirectory(r,a),t.onChangedAutomaticTypeDirectiveNames();var i=le(n,e);i&&se(a,i===a)&&t.onInvalidatedResolution()},1)}function de(){var n=t.getCompilationSettings();if(n.types)ce();else{var r=e.getEffectiveTypeRoots(n,{directoryExists:pe,getCurrentDirectory:u});r?e.mutateMap(S,e.arrayToMap(r,function(e){return t.toPath(e)}),{createNewValue:ue,onDeleteValue:e.closeFileWatcher}):ce()}}function pe(n){var r=e.getDirectoryPath(e.getDirectoryPath(n)),a=t.toPath(r);return a===T||K(a)}}e.isPathInNodeModulesStartingWithDot=n,e.maxNumberOfFilesToIterateForInvalidation=256,e.createResolutionCache=t}(u||(u={})),function(e){(function(n){var t,r;function a(n,t,r){var a=n.importModuleSpecifierPreference,i=n.importModuleSpecifierEnding;return{relativePreference:"relative"===a?0:"non-relative"===a?1:2,ending:o()};function o(){switch(i){case"minimal":return 0;case"index":return 1;case"js":return 2;default:return m(r)?2:e.getEmitModuleResolutionKind(t)!==e.ModuleResolutionKind.NodeJs?1:0}}}function i(n,t){return{relativePreference:e.isExternalModuleNameRelative(t)?0:1,ending:e.hasJSOrJsonFileExtension(t)?2:e.getEmitModuleResolutionKind(n)!==e.ModuleResolutionKind.NodeJs||e.endsWith(t,"index")?1:0}}function o(e,n,t,r,a,o,s){var l=c(e,n,t,r,a,o,i(e,s));if(l!==s)return l}function s(e,n,t,r,i,o,s,l){return void 0===s&&(s={}),c(e,t,r,i,o,l,a(s,e,n))}function c(n,t,r,a,i,o,s){var c=u(t,a),l=y(i,t,r,c.getCanonicalFileName,a,o);return e.firstDefined(l,function(e){return T(e,c,a,n)})||d(r,c,n,s)}function l(n,t,r,i,o,s,c){var l=h(n);if(l)return[l];var p=u(r.path,i),m=e.getSourceFileOfNode(n.valueDeclaration||e.getNonAugmentationDeclaration(n)),f=y(o,r.path,m.fileName,p.getCanonicalFileName,i,c),_=a(s,t,r),g=e.mapDefined(f,function(e){return T(e,p,i,t)});return g.length?g:f.map(function(e){return d(e,p,t,_)})}function u(n,t){var r=e.createGetCanonicalFileName(!t.useCaseSensitiveFileNames||t.useCaseSensitiveFileNames()),a=e.getDirectoryPath(n);return{getCanonicalFileName:r,sourceDirectory:a}}function d(n,t,r,a){var i=t.getCanonicalFileName,o=t.sourceDirectory,s=a.ending,c=a.relativePreference,l=r.baseUrl,u=r.paths,d=r.rootDirs,m=d&&E(d,n,o,i)||A(e.ensurePathIsNonModuleName(e.getRelativePathFromDirectory(o,n,i)),s,r);if(!l||0===c)return m;var f=L(n,l,i);if(!f)return m;var _=A(f,s,r),g=u&&b(e.removeFileExtension(f),_,u),v=void 0===g?_:g;return 1===c?v:(2!==c&&e.Debug.assertNever(c),k(v)||p(m)<p(v)?m:v)}function p(n){for(var t=0,r=e.startsWith(n,"./")?2:0;r<n.length;r++)47===n.charCodeAt(r)&&t++;return t}function m(n){var t=n.imports;return e.firstDefined(t,function(n){var t=n.text;return e.pathIsRelative(t)?e.hasJSOrJsonFileExtension(t):void 0})||!1}function f(e,n,t){return t(e)===t(n)}function _(n,t){return"node_modules"===t(n)||e.startsWith(n,"@")}function g(n,t,r,a){var i=e.getPathComponents(e.toPath(n,r,a)),o=e.getPathComponents(e.toPath(t,r,a));while(!_(i[i.length-2],a)&&!_(o[o.length-2],a)&&f(i[i.length-1],o[o.length-1],a))i.pop(),o.pop();return[e.getPathFromPathComponents(i),e.getPathFromPathComponents(o)]}function v(n,t,r){for(var a=e.createMap(),i=e.mapDefined(n,function(n){return n.resolvedModules&&e.firstDefinedIterator(n.resolvedModules.values(),function(e){return e&&e.originalPath&&e.resolvedFileName!==e.originalPath?[e.resolvedFileName,e.originalPath]:void 0})}),o=0,s=i;o<s.length;o++){var c=s[o],l=c[0],u=c[1],d=g(l,u,r,t),p=d[0],m=d[1];a.set(m,p)}return a}function y(n,t,r,a,i,o){var s=o.get(r),c=s?s.concat([r]):[r],l=i.getCurrentDirectory?i.getCurrentDirectory():"",u=c.map(function(n){return e.getNormalizedAbsolutePath(n,l)}),d=v(n,a,l),p=[],m=!i.useCaseSensitiveFileNames||i.useCaseSensitiveFileNames()?e.compareStringsCaseSensitive:e.compareStringsCaseInsensitive;return d.forEach(function(n,r){if(!e.startsWithDirectory(t,n,a)){var o=e.find(u,function(e){return 0===m(e.slice(0,n.length+1),n+"/")});if(void 0!==o){var s=e.getRelativePathFromDirectory(n,o,a),c=e.resolvePath(r,s);i.fileExists&&!i.fileExists(c)||p.push(c)}}}),p.push.apply(p,u),p}function h(n){var t=e.find(n.declarations,function(n){return e.isNonGlobalAmbientModule(n)&&(!e.isExternalModuleAugmentation(n)||!e.isExternalModuleNameRelative(e.getTextOfIdentifierOrLiteral(n.name)))});if(t)return t.name.text}function b(n,t,r){for(var a in r)for(var i=0,o=r[a];i<o.length;i++){var s=o[i],c=e.removeFileExtension(e.normalizePath(s)),l=c.indexOf("*");if(-1!==l){var u=c.substr(0,l),d=c.substr(l+1);if(t.length>=u.length+d.length&&e.startsWith(t,u)&&e.endsWith(t,d)||!d&&t===e.removeTrailingDirectorySeparator(u)){var p=t.substr(u.length,t.length-d.length);return a.replace("*",p)}}else if(c===t||c===n)return a}}function E(n,t,r,a){var i=C(t,n,a);if(void 0!==i){var o=C(r,n,a),s=void 0!==o?e.ensurePathIsNonModuleName(e.getRelativePathFromDirectory(o,i,a)):i;return e.removeFileExtension(s)}}function T(n,t,r,a){var i=t.getCanonicalFileName,o=t.sourceDirectory;if(r.fileExists&&r.readFile){var s=x(n);if(s){var c=n.substring(0,s.packageRootIndex),l=e.combinePaths(c,"package.json"),u=r.fileExists(l)?JSON.parse(r.readFile(l)):void 0,d=u&&u.typesVersions?e.getPackageJsonTypesVersionsPaths(u.typesVersions):void 0;if(d){var p=n.slice(s.packageRootIndex+1),m=b(e.removeFileExtension(p),A(p,0,a),d.paths);void 0!==m&&(n=e.combinePaths(n.slice(0,s.packageRootIndex),m))}var f=v(n);if(e.startsWith(o,i(f.substring(0,s.topLevelNodeModulesIndex)))){var _=f.substring(s.topLevelPackageNameIndex+1),g=e.getPackageNameFromTypesPackageName(_);return e.getEmitModuleResolutionKind(a)!==e.ModuleResolutionKind.NodeJs&&g===_?void 0:g}}}function v(n){if(u){var t=u.typings||u.types||u.main;if(t){var a=e.toPath(t,c,i);if(e.removeFileExtension(a)===e.removeFileExtension(i(n)))return c}}var o=e.removeFileExtension(n);return"/index"!==i(o.substring(s.fileNameIndex))||S(r,o.substring(0,s.fileNameIndex))?o:o.substring(0,s.fileNameIndex)}}function S(n,t){if(n.fileExists)for(var r=e.getSupportedExtensions({allowJs:!0},[{extension:"node",isMixedContent:!1},{extension:"json",isMixedContent:!1,scriptKind:6}]),a=0,i=r;a<i.length;a++){var o=i[a],s=t+o;if(n.fileExists(s))return s}}function x(n){var t,r=0,a=0,i=0,o=0;(function(e){e[e["BeforeNodeModules"]=0]="BeforeNodeModules",e[e["NodeModules"]=1]="NodeModules",e[e["Scope"]=2]="Scope",e[e["PackageContent"]=3]="PackageContent"})(t||(t={}));var s=0,c=0,l=0;while(c>=0)switch(s=c,c=n.indexOf("/",s+1),l){case 0:n.indexOf(e.nodeModulesPathPart,s)===s&&(r=s,a=c,l=1);break;case 1:case 2:1===l&&"@"===n.charAt(s+1)?l=2:(i=c,l=3);break;case 3:l=n.indexOf(e.nodeModulesPathPart,s)===s?1:3;break}return o=s,l>1?{topLevelNodeModulesIndex:r,topLevelPackageNameIndex:a,packageRootIndex:i,fileNameIndex:o}:void 0}function C(n,t,r){return e.firstDefined(t,function(e){var t=L(n,e,r);return k(t)?void 0:t})}function A(n,t,r){if(e.fileExtensionIs(n,".json"))return n;var a=e.removeFileExtension(n);switch(t){case 0:return e.removeSuffix(a,"/index");case 1:return a;case 2:return a+D(n,r);default:return e.Debug.assertNever(t)}}function D(n,t){var r=e.extensionFromPath(n);switch(r){case".ts":case".d.ts":return".js";case".tsx":return 1===t.jsx?".jsx":".js";case".js":case".jsx":case".json":return r;default:return e.Debug.assertNever(r)}}function L(n,t,r){var a=e.getRelativePathToDirectoryOrUrl(t,n,t,r,!1);return e.isRootedDiskPath(a)?void 0:a}function k(n){return e.startsWith(n,"..")}(function(e){e[e["Relative"]=0]="Relative",e[e["NonRelative"]=1]="NonRelative",e[e["Auto"]=2]="Auto"})(t||(t={})),function(e){e[e["Minimal"]=0]="Minimal",e[e["Index"]=1]="Index",e[e["JsExtension"]=2]="JsExtension"}(r||(r={})),n.updateModuleSpecifier=o,n.getModuleSpecifier=s,n.getModuleSpecifiers=l,n.countPathComponents=p})(e.moduleSpecifiers||(e.moduleSpecifiers={}))}(u||(u={})),function(e){var n=e.sys?{getCurrentDirectory:function(){return e.sys.getCurrentDirectory()},getNewLine:function(){return e.sys.newLine},getCanonicalFileName:e.createGetCanonicalFileName(e.sys.useCaseSensitiveFileNames)}:void 0;function t(t,r){var a=t===e.sys?n:{getCurrentDirectory:function(){return t.getCurrentDirectory()},getNewLine:function(){return t.newLine},getCanonicalFileName:e.createGetCanonicalFileName(t.useCaseSensitiveFileNames)};if(!r)return function(n){return t.write(e.formatDiagnostic(n,a))};var i=new Array(1);return function(n){i[0]=n,t.write(e.formatDiagnosticsWithColorAndContext(i,a)+a.getNewLine()),i[0]=void 0}}function r(n,t,r){return!(!n.clearScreen||r.preserveWatchOutput||r.extendedDiagnostics||r.diagnostics||!e.contains(e.screenStartingMessageCodes,t.code))&&(n.clearScreen(),!0)}function a(n,t){return e.contains(e.screenStartingMessageCodes,n.code)?t+t:t}function i(n,t){return t?function(t,a,i){r(n,t,i);var o="["+e.formatColorAndReset((new Date).toLocaleTimeString(),e.ForegroundColorEscapeSequences.Grey)+"] ";o+=""+e.flattenDiagnosticMessageText(t.messageText,n.newLine)+(a+a),n.write(o)}:function(t,i,o){var s="";r(n,t,o)||(s+=i),s+=(new Date).toLocaleTimeString()+" - ",s+=""+e.flattenDiagnosticMessageText(t.messageText,n.newLine)+a(t,i),n.write(s)}}function o(n,t,r,a){var i=r;i.onUnRecoverableConfigFileDiagnostic=function(n){return v(e.sys,a,n)};var o=e.getParsedCommandLineOfConfigFile(n,t,i);return i.onUnRecoverableConfigFileDiagnostic=void 0,o}function s(n){return e.countWhere(n,function(n){return n.category===e.DiagnosticCategory.Error})}function c(n){return 1===n?e.Diagnostics.Found_1_error_Watching_for_file_changes:e.Diagnostics.Found_0_errors_Watching_for_file_changes}function l(n,t){if(0===n)return"";var r=e.createCompilerDiagnostic(1===n?e.Diagnostics.Found_1_error:e.Diagnostics.Found_0_errors,n);return""+t+e.flattenDiagnosticMessageText(r.messageText,t)+t+t}function u(n,t,r,a,i){var o=n.getConfigFileParsingDiagnostics().slice(),c=o.length;e.addRange(o,n.getSyntacticDiagnostics());var l=!1;o.length===c&&(e.addRange(o,n.getOptionsDiagnostics()),e.addRange(o,n.getGlobalDiagnostics()),o.length===c&&(l=!0));var u=n.emit(void 0,i),d=u.emittedFiles,p=u.emitSkipped,m=u.diagnostics;if(e.addRange(o,m),l&&e.addRange(o,n.getSemanticDiagnostics()),e.sortAndDeduplicateDiagnostics(o).forEach(t),r){var f=n.getCurrentDirectory();e.forEach(d,function(n){var t=e.getNormalizedAbsolutePath(n,f);r("TSFILE: "+t)}),n.getCompilerOptions().listFiles&&e.forEach(n.getSourceFiles(),function(e){r(e.fileName)})}return a&&a(s(o)),p&&o.length>0?e.ExitStatus.DiagnosticsPresent_OutputsSkipped:o.length>0?e.ExitStatus.DiagnosticsPresent_OutputsGenerated:e.ExitStatus.Success}e.createDiagnosticReporter=t,e.screenStartingMessageCodes=[e.Diagnostics.Starting_compilation_in_watch_mode.code,e.Diagnostics.File_change_detected_Starting_incremental_compilation.code],e.createWatchStatusReporter=i,e.parseConfigFileWithSystem=o,e.getErrorCountForSummary=s,e.getWatchErrorSummaryDiagnosticMessage=c,e.getErrorSummaryText=l,e.emitFilesAndReportErrors=u;var d={close:e.noop};function p(n,t){void 0===n&&(n=e.sys);var r=t||i(n);return{onWatchStatusChange:r,watchFile:e.maybeBind(n,n.watchFile)||function(){return d},watchDirectory:e.maybeBind(n,n.watchDirectory)||function(){return d},setTimeout:e.maybeBind(n,n.setTimeout)||e.noop,clearTimeout:e.maybeBind(n,n.clearTimeout)||e.noop}}function m(n,t){var r=n.trace?t.extendedDiagnostics?e.WatchLogLevel.Verbose:t.diagnostics?e.WatchLogLevel.TriggerOnly:e.WatchLogLevel.None:e.WatchLogLevel.None,a=r!==e.WatchLogLevel.None?function(e){return n.trace(e)}:e.noop,i=e.getWatchFactory(r,a);return i.writeLog=a,i}function f(n,t,r){void 0===r&&(r=n);var a=n.useCaseSensitiveFileNames(),i=e.memoize(function(){return n.getNewLine()});return{getSourceFile:function(r,a,i){var o;try{e.performance.mark("beforeIORead"),o=n.readFile(r,t().charset),e.performance.mark("afterIORead"),e.performance.measure("I/O Read","beforeIORead","afterIORead")}catch(s){i&&i(s.message),o=""}return void 0!==o?e.createSourceFile(r,o,a):void 0},getDefaultLibLocation:e.maybeBind(n,n.getDefaultLibLocation),getDefaultLibFileName:function(e){return n.getDefaultLibFileName(e)},writeFile:s,getCurrentDirectory:e.memoize(function(){return n.getCurrentDirectory()}),useCaseSensitiveFileNames:function(){return a},getCanonicalFileName:e.createGetCanonicalFileName(a),getNewLine:function(){return e.getNewLineCharacter(t(),i)},fileExists:function(e){return n.fileExists(e)},readFile:function(e){return n.readFile(e)},trace:e.maybeBind(n,n.trace),directoryExists:e.maybeBind(r,r.directoryExists),getDirectories:e.maybeBind(r,r.getDirectories),realpath:e.maybeBind(n,n.realpath),getEnvironmentVariable:e.maybeBind(n,n.getEnvironmentVariable)||function(){return""},createHash:e.maybeBind(n,n.createHash),readDirectory:e.maybeBind(n,n.readDirectory)};function o(t){if(t.length>e.getRootLength(t)&&!n.directoryExists(t)){var r=e.getDirectoryPath(t);o(r),n.createDirectory&&n.createDirectory(t)}}function s(t,r,a,i){try{e.performance.mark("beforeIOWrite"),o(e.getDirectoryPath(e.normalizePath(t))),n.writeFile(t,r,a),e.performance.mark("afterIOWrite"),e.performance.measure("I/O Write","beforeIOWrite","afterIOWrite")}catch(s){i&&i(s.message)}}}function _(n,t){var r=e.memoize(function(){return e.getDirectoryPath(e.normalizePath(n.getExecutingFilePath()))});return{useCaseSensitiveFileNames:function(){return n.useCaseSensitiveFileNames},getNewLine:function(){return n.newLine},getCurrentDirectory:e.memoize(function(){return n.getCurrentDirectory()}),getDefaultLibLocation:r,getDefaultLibFileName:function(n){return e.combinePaths(r(),e.getDefaultLibFileName(n))},fileExists:function(e){return n.fileExists(e)},readFile:function(e,t){return n.readFile(e,t)},directoryExists:function(e){return n.directoryExists(e)},getDirectories:function(e){return n.getDirectories(e)},readDirectory:function(e,t,r,a,i){return n.readDirectory(e,t,r,a,i)},realpath:e.maybeBind(n,n.realpath),getEnvironmentVariable:e.maybeBind(n,n.getEnvironmentVariable),trace:function(e){return n.write(e+n.newLine)},createDirectory:function(e){return n.createDirectory(e)},writeFile:function(e,t,r){return n.writeFile(e,t,r)},onCachedDirectoryStructureHostCreate:function(e){return e||n},createHash:e.maybeBind(n,n.createHash),createProgram:t}}function g(n,t,r,a){void 0===n&&(n=e.sys);var i=function(e){return n.write(e+n.newLine)},o=_(n,t||e.createEmitAndSemanticDiagnosticsBuilderProgram);return e.copyProperties(o,p(n,a)),o.afterProgramCreate=function(t){var a=t.getCompilerOptions(),s=e.getNewLineCharacter(a,function(){return n.newLine});u(t,r,i,function(n){return o.onWatchStatusChange(e.createCompilerDiagnostic(c(n),n),s,a)})},o}function v(n,t,r){t(r),n.exit(e.ExitStatus.DiagnosticsPresent_OutputsSkipped)}function y(e,n,r,a,i,o){var s=i||t(r),c=g(r,a,s,o);return c.onUnRecoverableConfigFileDiagnostic=function(e){return v(r,s,e)},c.configFileName=e,c.optionsToExtend=n,c}function h(e,n,r,a,i,o,s){var c=g(r,a,i||t(r),o);return c.rootFiles=e,c.options=n,c.projectReferences=s,c}e.createWatchHost=p,function(e){e["ConfigFile"]="Config file",e["SourceFile"]="Source file",e["MissingFile"]="Missing file",e["WildcardDirectory"]="Wild card directory",e["FailedLookupLocations"]="Failed Lookup Locations",e["TypeRoots"]="Type roots"}(e.WatchType||(e.WatchType={})),e.createWatchFactory=m,e.createCompilerHostFromProgramHost=f,e.createProgramHost=_,e.createWatchCompilerHostOfConfigFile=y,e.createWatchCompilerHostOfFilesAndCompilerOptions=h}(u||(u={})),function(e){function n(n,t,r,a,i,o,s){return e.isArray(n)?e.createWatchCompilerHostOfFilesAndCompilerOptions(n,t,r,a,i,o,s):e.createWatchCompilerHostOfConfigFile(n,t,r,a,i,o)}e.createWatchCompilerHost=n;var t=1;function r(n){var r,a,i,o,s,c,l,u,d=e.createMap(),p=!1,m=!1,f=n.useCaseSensitiveFileNames(),_=n.getCurrentDirectory(),g=n.configFileName,v=n.optionsToExtend,y=void 0===v?{}:v,h=n.createProgram,b=n.rootFiles,E=n.options,T=n.projectReferences,S=!1,x=!1,C=void 0===g?void 0:e.createCachedDirectoryStructureHost(n,_,f);C&&n.onCachedDirectoryStructureHostCreate&&n.onCachedDirectoryStructureHostCreate(C);var A=C||n,D=e.parseConfigHostFromCompilerHostLike(n,A),L=H();g&&n.configFileParsingResult&&(oe(n.configFileParsingResult),L=H()),$(e.Diagnostics.Starting_compilation_in_watch_mode),g&&!n.configFileParsingResult&&(L=e.getNewLineCharacter(y,function(){return n.getNewLine()}),e.Debug.assert(!b),ie(),L=H());var k=e.createWatchFactory(n,E),N=k.watchFile,I=k.watchFilePath,w=k.watchDirectory,M=k.writeLog,O=e.createGetCanonicalFileName(f);M("Current directory: "+_+" CaseSensitiveFileNames: "+f),g&&N(n,g,ne,e.PollingInterval.High,"Config file");var P=e.createCompilerHostFromProgramHost(n,function(){return E},A),R=P.getSourceFile;P.getSourceFile=function(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];return X.apply(void 0,[e,W(e)].concat(n))},P.getSourceFileByPath=X,P.getNewLine=function(){return L},P.fileExists=J,P.onReleaseOldSourceFile=Z,P.toPath=W,P.getCompilationSettings=function(){return E},P.watchDirectoryOfFailedLookupLocation=function(e,t,r){return w(n,e,t,r,"Failed Lookup Locations")},P.watchTypeRootsDirectory=function(e,t,r){return w(n,e,t,r,"Type roots")},P.getCachedDirectoryStructureHost=function(){return C},P.onInvalidatedResolution=ee,P.onChangedAutomaticTypeDirectiveNames=function(){m=!0,ee()},P.maxNumberOfFilesToIterateForInvalidation=n.maxNumberOfFilesToIterateForInvalidation,P.getCurrentProgram=B,P.writeLog=M;var F=e.createResolutionCache(P,g?e.getDirectoryPath(e.getNormalizedAbsolutePath(g,_)):_,!1);P.resolveModuleNames=n.resolveModuleNames?function(e,t,r,a){return n.resolveModuleNames(e,t,r,a)}:function(e,n,t,r){return F.resolveModuleNames(e,n,t,r)},P.resolveTypeReferenceDirectives=n.resolveTypeReferenceDirectives?function(e,t,r){return n.resolveTypeReferenceDirectives(e,t,r)}:function(e,n,t){return F.resolveTypeReferenceDirectives(e,n,t)};var G=!!n.resolveModuleNames||!!n.resolveTypeReferenceDirectives;return K(),de(),g?{getCurrentProgram:V,getProgram:K}:{getCurrentProgram:V,getProgram:K,updateRootFileNames:U};function V(){return r}function B(){return r&&r.getProgram()}function K(){M("Synchronizing program");var t=B();p&&(L=H(),t&&e.changesAffectModuleResolution(t.getCompilerOptions(),E)&&F.clear());var a=F.createHasInvalidatedResolution(G);return e.isProgramUptoDate(B(),b,E,Q,J,a,m,T)?x&&(r=h(void 0,void 0,P,r,u,T),x=!1):j(t,a),n.afterProgramCreate&&n.afterProgramCreate(r),r}function j(n,t){M("CreatingProgramWith::"),M("  roots: "+JSON.stringify(b)),M("  options: "+JSON.stringify(E));var a=p||!n;if(p=!1,x=!1,F.startCachingPerDirectoryResolution(),P.hasInvalidatedResolution=t,P.hasChangedAutomaticTypeDirectiveNames=m,r=h(b,E,P,r,u,T),F.finishCachingPerDirectoryResolution(),e.updateMissingFilePathsWatch(r.getProgram(),i||(i=e.createMap()),le),a&&F.updateTypeRootsWatch(),c){for(var o=0,s=c;o<s.length;o++){var l=s[o];i.has(l)||d.delete(l)}c=void 0}}function U(n){e.Debug.assert(!g,"Cannot update root file names with config file watch mode"),b=n,ee()}function H(){return e.getNewLineCharacter(E||y,function(){return n.getNewLine()})}function W(n){return e.toPath(n,_,O)}function q(e){return"number"===typeof e}function z(e){return!!e.sourceFile}function J(e){var n=W(e);return!!q(d.get(n))||A.fileExists(e)}function X(r,a,i,o,s){var c=d.get(a);if(!q(c)){if(!c||s||!z(c)||c.version.toString()!==c.sourceFile.version){var l=R(r,i,o);if(c)s&&c.version++,l?(c.sourceFile=l,l.version=c.version.toString(),c.fileWatcher||(c.fileWatcher=I(n,r,se,e.PollingInterval.Low,a,"Source file"))):(z(c)&&c.fileWatcher.close(),d.set(a,c.version));else if(l){l.version=t.toString();var u=I(n,r,se,e.PollingInterval.Low,a,"Source file");d.set(a,{sourceFile:l,version:t,fileWatcher:u})}else d.set(a,t);return l}return c.sourceFile}}function Y(e){var n=d.get(e);void 0!==n&&(q(n)?d.set(e,{version:Number(n)+1}):n.version++)}function Q(e){var n=d.get(e);return!n||q(n)?void 0:n.version.toString()}function Z(e,n,t){var r=d.get(e.resolvedPath);r&&(q(r)?(c||(c=[])).push(e.path):r.sourceFile===e&&(r.fileWatcher&&r.fileWatcher.close(),d.delete(e.resolvedPath),t||F.removeResolutionsOfFile(e.path)))}function $(t){n.onWatchStatusChange&&n.onWatchStatusChange(e.createCompilerDiagnostic(t),L,E||y)}function ee(){n.setTimeout&&n.clearTimeout&&(s&&n.clearTimeout(s),M("Scheduling update"),s=n.setTimeout(te,250))}function ne(){e.Debug.assert(!!g),a=e.ConfigFileProgramReloadLevel.Full,ee()}function te(){switch(s=void 0,$(e.Diagnostics.File_change_detected_Starting_incremental_compilation),a){case e.ConfigFileProgramReloadLevel.Partial:return re();case e.ConfigFileProgramReloadLevel.Full:return ae();default:return void K()}}function re(){M("Reloading new file names and options");var n=e.getFileNamesFromConfigSpecs(l,e.getDirectoryPath(g),E,D);e.updateErrorForNoInputFiles(n,g,l,u,S)&&(x=!0),b=n.fileNames,K()}function ae(){M("Reloading config file: "+g),a=e.ConfigFileProgramReloadLevel.None,C&&C.clearCache(),ie(),p=!0,K(),de()}function ie(){oe(e.getParsedCommandLineOfConfigFile(g,y,D))}function oe(n){b=n.fileNames,E=n.options,l=n.configFileSpecs,T=n.projectReferences,u=e.getConfigFileParsingDiagnostics(n).slice(),S=e.canJsonReportNoInutFiles(n.raw),x=!0}function se(n,t,r){ce(n,r,t),t===e.FileWatcherEventKind.Deleted&&d.get(r)&&F.invalidateResolutionOfFile(r),F.removeResolutionsFromProjectReferenceRedirects(r),Y(r),ee()}function ce(e,n,t){C&&C.addOrDeleteFile(e,n,t)}function le(t){return I(n,t,ue,e.PollingInterval.Medium,t,"Missing file")}function ue(n,t,r){ce(n,r,t),t===e.FileWatcherEventKind.Created&&i.has(r)&&(i.get(r).close(),i.delete(r),Y(r),ee())}function de(){l?e.updateWatchingWildcardDirectories(o||(o=e.createMap()),e.createMapFromTemplate(l.wildcardDirectories),pe):o&&e.clearMap(o,e.closeFileWatcherOf)}function pe(t,r){return w(n,t,function(n){e.Debug.assert(!!g);var r=W(n);C&&C.addOrDeleteFileOrDirectory(n,r),Y(r),e.isPathInNodeModulesStartingWithDot(r)||(r===t||!e.hasExtension(r)||e.isSupportedSourceFileName(n,E)?a!==e.ConfigFileProgramReloadLevel.Full&&(a=e.ConfigFileProgramReloadLevel.Partial,ee()):M("Project: "+g+" Detected file add/remove of non supported extension: "+n))},r,"Wild card directory")}}e.createWatchProgram=r}(u||(u={})),function(e){var n,t,r=new Date(-864e13),a=new Date(864e13);function i(n){var t=e.createMap();return{setValue:o,getValue:s,removeKey:i,forEach:r,hasKey:a,getSize:c,clear:l};function r(e){t.forEach(e)}function a(e){return t.has(n(e))}function i(e){t.delete(n(e))}function o(e,r){t.set(n(e),r)}function s(e){return t.get(n(e))}function c(){return t.size}function l(){t.clear()}}function o(e,n,t){var r,a=e.getValue(n);return a||(r=t(),e.setValue(n,r)),a||r}function s(n,t){return o(n,t,e.createMap)}function c(n,t){var r=e.getRelativePathFromDirectory(p(t.options,t.options.configFilePath),n,!0),a=e.resolvePath(t.options.declarationDir||t.options.outDir||e.getDirectoryPath(t.options.configFilePath),r);return e.changeExtension(a,".d.ts")}function l(n,t){var r=e.getRelativePathFromDirectory(p(t.options,t.options.configFilePath),n,!0),a=e.resolvePath(t.options.outDir||e.getDirectoryPath(t.options.configFilePath),r),i=e.fileExtensionIs(n,".json")?".json":e.fileExtensionIs(n,".tsx")&&1===t.options.jsx?".jsx":".js";return e.changeExtension(a,i)}function u(n,t){if(t.options.outFile||t.options.out||e.fileExtensionIs(n,".d.ts"))return e.emptyArray;var r=[],a=l(n,t);if(r.push(a),t.options.sourceMap&&r.push(a+".map"),e.getEmitDeclarations(t.options)&&!e.fileExtensionIs(n,".json")){var i=c(n,t);r.push(i),t.options.declarationMap&&r.push(i+".map")}return r}function d(n){var t=n.options.outFile||n.options.out;if(!t)return e.Debug.fail("outFile must be set");var r=[];if(r.push(t),n.options.sourceMap&&r.push(t+".map"),e.getEmitDeclarations(n.options)){var a=e.changeExtension(t,".d.ts");r.push(a),n.options.declarationMap&&r.push(a+".map")}return r}function p(n,t){return n.rootDir||e.getDirectoryPath(t)}function m(e,n){return n>e?n:e}function f(n){return e.fileExtensionIs(n,".d.ts")}function _(n,t){return function(r){var a=t?"["+e.formatColorAndReset((new Date).toLocaleTimeString(),e.ForegroundColorEscapeSequences.Grey)+"] ":(new Date).toLocaleTimeString()+" - ";a+=""+e.flattenDiagnosticMessageText(r.messageText,n.newLine)+(n.newLine+n.newLine),n.write(a)}}function g(n,t,r,a){var i=e.createProgramHost(n,t);return i.getModifiedTime=n.getModifiedTime?function(e){return n.getModifiedTime(e)}:function(){},i.setModifiedTime=n.setModifiedTime?function(e,t){return n.setModifiedTime(e,t)}:e.noop,i.deleteFile=n.deleteFile?function(e){return n.deleteFile(e)}:e.noop,i.reportDiagnostic=r||e.createDiagnosticReporter(n),i.reportSolutionBuilderStatus=a||_(n),i}function v(n,t,r,a,i){void 0===n&&(n=e.sys);var o=g(n,t||e.createAbstractBuilder,r,a);return o.reportErrorSummary=i,o}function y(n,t,r,a,i){void 0===n&&(n=e.sys);var o=g(n,t||e.createEmitAndSemanticDiagnosticsBuilderProgram,r,a),s=e.createWatchHost(n,i);return e.copyProperties(o,s),o}function h(n){var t={};return e.commonOptionsWithBuild.forEach(function(e){t[e.name]=n[e.name]}),t}function b(c,l,u){var d,p=c,_=c.getCurrentDirectory(),g=e.createGetCanonicalFileName(c.useCaseSensitiveFileNames()),v=e.parseConfigHostFromCompilerHostLike(c),y=u,b=h(y),C=i(Q),A=i(Q),D=i(Q),L=e.createMap(),k=function(e){return c.trace&&c.trace(e)},N=function(e){return c.readFile(e)},I=b,w=e.createCompilerHostFromProgramHost(c,function(){return I}),M=w.getSourceFile,O=c.createHash||e.generateDjb2Hash;$();var P,R=i(Q),F=i(Q),G=i(Q),V=i(Q),B=[],K=0,j=!1,U=e.createWatchFactory(c,y),H=U.watchFile,W=U.watchFilePath,q=U.watchDirectory,z=U.writeLog,J=i(Q),X=i(Q),Y=i(Q);return{buildAllProjects:Oe,getUpToDateStatusOfFile:de,cleanAllProjects:Ie,resetBuildContext:Z,getBuildGraph:pe,invalidateProject:ge,buildInvalidatedProject:Te,resolveProjectName:we,startWatching:ae};function Q(n){return e.toPath(n,_,g)}function Z(n){void 0===n&&(n=u),y=n,b=h(y),C.clear(),A.clear(),D.clear(),L.clear(),d=void 0,F.clear(),G.clear(),V.clear(),B.length=0,K=0,P&&(clearTimeout(P),P=void 0),j=!1,e.clearMap(J,function(n){return e.clearMap(n,e.closeFileWatcherOf)}),e.clearMap(X,function(n){return e.clearMap(n,e.closeFileWatcher)}),e.clearMap(Y,e.closeFileWatcher),R.clear(),$()}function $(){y.watch?w.getSourceFile===M&&(w.getSourceFile=function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];var t=M.call.apply(M,[w].concat(e));return t&&y.watch&&(t.version=O.call(c,t.text)),t}):w.getSourceFile!==M&&(w.getSourceFile=M)}function ee(e){return!!e.options}function ne(n){var t,r=C.getValue(n);if(r)return ee(r)?r:void 0;v.onUnRecoverableConfigFileDiagnostic=function(e){return t=e};var a=e.getParsedCommandLineOfConfigFile(n,b,v);return v.onUnRecoverableConfigFileDiagnostic=e.noop,C.setValue(n,a||t),a}function te(n){for(var t=[],r=1;r<arguments.length;r++)t[r-1]=arguments[r];c.reportSolutionBuilderStatus(e.createCompilerDiagnostic.apply(void 0,[n].concat(t)))}function re(n){for(var t=[],r=1;r<arguments.length;r++)t[r-1]=arguments[r];p.onWatchStatusChange&&p.onWatchStatusChange(e.createCompilerDiagnostic.apply(void 0,[n].concat(t)),c.getNewLine(),b)}function ae(){for(var e=me(),n=0,t=e.buildQueue;n<t.length;n++){var r=t[n];ie(r);var a=ne(r);a&&(oe(r,a),se(r,a))}}function ie(n){y.watch&&!Y.hasKey(n)&&Y.setValue(n,H(p,n,function(){ue(n,e.ConfigFileProgramReloadLevel.Full)},e.PollingInterval.High,"Config file",n))}function oe(n,t){y.watch&&e.updateWatchingWildcardDirectories(s(J,n),e.createMapFromTemplate(t.configFileSpecs.wildcardDirectories),function(r,a){return q(p,r,function(a){var i=Q(a);i===Q(r)||!e.hasExtension(i)||e.isSupportedSourceFileName(a,t.options)?ce(a,t)?z(a+" is output file"):ue(n,e.ConfigFileProgramReloadLevel.Partial):z("Project: "+n+" Detected file add/remove of non supported extension: "+a)},a,"Wild card directory",n)})}function se(n,t){y.watch&&e.mutateMap(s(X,n),e.arrayToMap(t.fileNames,Q),{createNewValue:function(t,r){return W(p,r,function(){return ue(n,e.ConfigFileProgramReloadLevel.None)},e.PollingInterval.Low,t,"Source file",n)},onDeleteValue:e.closeFileWatcher})}function ce(n,t){if(t.options.noEmit)return!1;if(!e.fileExtensionIs(n,".d.ts")&&(e.fileExtensionIs(n,".ts")||e.fileExtensionIs(n,".tsx")))return!1;var r=t.options.outFile||t.options.out;return!(!r||!le(n,r)&&!le(n,e.removeFileExtension(r)+".d.ts"))||(!(!t.options.declarationDir||!e.containsPath(t.options.declarationDir,n,_,!c.useCaseSensitiveFileNames()))||(!(!t.options.outDir||!e.containsPath(t.options.outDir,n,_,!c.useCaseSensitiveFileNames()))||!e.forEach(t.fileNames,function(e){return le(n,e)})))}function le(n,t){return 0===e.comparePaths(n,t,_,!c.useCaseSensitiveFileNames())}function ue(e,n){j=!0,ve(e,n),Ee()}function de(e){return fe(ne(e))}function pe(e){return Ce(Me(e))}function me(){return d||(d=pe(l))}function fe(e){if(void 0===e)return{type:t.Unbuildable,reason:"File deleted mid-build"};var n=D.getValue(e.options.configFilePath);if(void 0!==n)return n;var r=_e(e);return D.setValue(e.options.configFilePath,r),r}function _e(n){for(var i=void 0,o=r,s=0,l=n.fileNames;s<l.length;s++){var u=l[s];if(!c.fileExists(u))return{type:t.Unbuildable,reason:u+" does not exist"};var d=c.getModifiedTime(u)||e.missingFileModifiedTime;d>o&&(i=u,o=d)}var p=T(n);if(0===p.length)return{type:t.ContainerOnly};for(var _,g="(none)",v=a,y="(none)",h=r,b=r,E=!1,S=0,x=p;S<x.length;S++){var C=x[S];if(!c.fileExists(C)){_=C;break}var L=c.getModifiedTime(C)||e.missingFileModifiedTime;if(L<v&&(v=L,g=C),L<o){E=!0;break}if(L>h&&(h=L,y=C),f(C)){var k=A.getValue(C);if(void 0!==k)b=m(k,b);else{var N=c.getModifiedTime(C)||e.missingFileModifiedTime;b=m(b,N)}}}var I,w=!1,M=!1;if(n.projectReferences){D.setValue(n.options.configFilePath,{type:t.ComputingUpstream});for(var O=0,P=n.projectReferences;O<P.length;O++){var R=P[O];M=M||!!R.prepend;var F=e.resolveProjectReferencePath(R),G=fe(ne(F));if(G.type!==t.ComputingUpstream){if(G.type===t.Unbuildable)return{type:t.UpstreamBlocked,upstreamProjectName:R.path};if(G.type!==t.UpToDate)return{type:t.UpstreamOutOfDate,upstreamProjectName:R.path};if(!(G.newestInputFileTime&&G.newestInputFileTime<=v)){if(!(G.newestDeclarationFileContentChangedTime&&G.newestDeclarationFileContentChangedTime<=v))return e.Debug.assert(void 0!==g,"Should have an oldest output filename here"),{type:t.OutOfDateWithUpstream,outOfDateOutputFileName:g,newerProjectName:R.path};w=!0,I=R.path}}}}return void 0!==_?{type:t.OutputMissing,missingOutputFileName:_}:E?{type:t.OutOfDateWithSelf,outOfDateOutputFileName:g,newerInputFileName:i}:M&&w?{type:t.OutOfDateWithUpstream,outOfDateOutputFileName:g,newerProjectName:I}:{type:w?t.UpToDateWithUpstreamTypes:t.UpToDate,newestDeclarationFileContentChangedTime:b,newestInputFileTime:o,newestOutputFileTime:h,newestInputFileName:i,newestOutputFileName:y,oldestOutputFileName:g}}function ge(e,n){ve(we(e),n)}function ve(n,t){t===e.ConfigFileProgramReloadLevel.Full&&(C.removeKey(n),d=void 0),D.removeKey(n),F.removeKey(n),ye(n,t)}function ye(n,t){var r=G.getValue(n);void 0===r?(G.setValue(n,t||e.ConfigFileProgramReloadLevel.None),B.push(n)):r<(t||e.ConfigFileProgramReloadLevel.None)&&G.setValue(n,t||e.ConfigFileProgramReloadLevel.None)}function he(){if(K<B.length){var e=B[K];K++;var n=G.getValue(e);return G.removeKey(e),G.getSize()||(B.length=0,K=0),{project:e,reloadLevel:n}}}function be(){return!!G.getSize()}function Ee(){p.setTimeout&&p.clearTimeout&&(P&&p.clearTimeout(P),P=p.setTimeout(Te,250))}function Te(){P=void 0,j&&(j=!1,V.clear(),re(e.Diagnostics.File_change_detected_Starting_incremental_compilation));var n=he();n&&(xe(n.project,n.reloadLevel),be()?y.watch&&!P&&Ee():Se())}function Se(){if(y.watch||c.reportErrorSummary){me().buildQueue.forEach(function(n){V.hasKey(n)||Fe(F.getValue(n)||e.emptyArray)});var n=0;F.forEach(function(t){return n+=e.getErrorCountForSummary(t)}),y.watch?re(e.getWatchErrorSummaryDiagnosticMessage(n),n):c.reportErrorSummary(n)}}function xe(r,a){var i=ne(r);if(i){if(a===e.ConfigFileProgramReloadLevel.Full)ie(r),oe(r,i),se(r,i);else if(a===e.ConfigFileProgramReloadLevel.Partial){var o=e.getFileNamesFromConfigSpecs(i.configFileSpecs,e.getDirectoryPath(r),i.options,v);e.updateErrorForNoInputFiles(o,r,i.configFileSpecs,i.errors,e.canJsonReportNoInutFiles(i.raw)),i.fileNames=o.fileNames,se(r,i)}var s=fe(i);if(Be(r,s),s.type!==t.UpstreamBlocked)if(s.type!==t.UpToDateWithUpstreamTypes){var c=Ae(r);if(!(c&n.AnyErrors)){var l=me(),u=l.referencingProjectsMap,d=l.buildQueue,p=u.getValue(r);if(p)for(var m=d.indexOf(r)+1;m<d.length;m++){var f=d[m],_=p.getValue(f);if(void 0!==_){var g=D.getValue(f);!_&&c&n.DeclarationOutputUnchanged?g&&g.type===t.UpToDate&&(g.type=t.UpToDateWithUpstreamTypes):!g||g.type!==t.UpToDate&&g.type!==t.UpToDateWithUpstreamTypes||D.setValue(f,{type:t.OutOfDateWithUpstream,outOfDateOutputFileName:g.oldestOutputFileName,newerProjectName:r}),ye(f)}}}}else Le(i);else y.verbose&&te(e.Diagnostics.Skipping_build_of_project_0_because_its_dependency_1_has_errors,r,s.upstreamProjectName)}else Pe(r)}function Ce(n){for(var t=i(Q),r=i(Q),a=[],s=[],c=i(Q),l=0,u=n;l<u.length;l++){var d=u[l];p(d)}return{buildQueue:s,referencingProjectsMap:c};function p(n,l){if(!r.hasKey(n))if(t.hasKey(n))l||te(e.Diagnostics.Project_references_may_not_form_a_circular_graph_Cycle_detected_Colon_0,a.join("\r\n"));else{t.setValue(n,!0),a.push(n);var u=ne(n);if(u&&u.projectReferences)for(var d=0,m=u.projectReferences;d<m.length;d++){var f=m[d],_=we(f.path);p(_,l||f.circular);var g=o(c,_,function(){return i(Q)});g.setValue(n,!!f.prepend)}a.pop(),r.setValue(n,!0),s.push(n)}}}function Ae(o){if(y.dry)return te(e.Diagnostics.A_non_dry_build_would_build_project_0,o),n.Success;y.verbose&&te(e.Diagnostics.Building_project_0,o);var s=n.None;s|=n.DeclarationOutputUnchanged;var l=ne(o);if(!l)return s|=n.ConfigFileErrors,Pe(o),D.setValue(o,{type:t.Unbuildable,reason:"Config file errors"}),s;if(0===l.fileNames.length)return Re(o,l.errors),n.None;I=l.options;var u=c.createProgram(l.fileNames,l.options,w,R.getValue(o),l.errors,l.projectReferences);I=b;var d=u.getConfigFileParsingDiagnostics().concat(u.getOptionsDiagnostics(),u.getGlobalDiagnostics(),u.getSyntacticDiagnostics());if(d.length)return M(d,n.SyntaxErrors,"Syntactic");var p=u.getSemanticDiagnostics();if(p.length)return M(p,n.TypeErrors,"Semantic");u.backupState();var _,g=r,v=!1,h=function(e){return(_||(_=[])).push(e)},E=[];if(e.emitFilesAndReportErrors(u,h,k,void 0,function(e,n,t){return E.push({name:e,text:n,writeByteOrderMark:t})}),_)return u.restoreState(),M(_,n.DeclarationEmitErrors,"Declaration file");var T=e.createDiagnosticCollection(),x=i(Q);E.forEach(function(t){var r,a=t.name,i=t.text,o=t.writeByteOrderMark;!v&&f(a)&&(c.fileExists(a)&&N(a)===i?r=c.getModifiedTime(a):(s&=~n.DeclarationOutputUnchanged,v=!0)),x.setValue(a,!0),e.writeFile(w,T,a,i,o),void 0!==r&&(g=m(r,g),A.setValue(a,r))});var C=T.getDiagnostics();if(C.length)return M(C,n.EmitErrors,"Emit");g=ke(l,g,e.Diagnostics.Updating_unchanged_output_timestamps_of_project_0,x);var L={type:t.UpToDate,newestDeclarationFileContentChangedTime:v?a:g,oldestOutputFileName:E.length?E[0].name:S(l)};return F.removeKey(o),D.setValue(o,L),De(o,u),s;function M(e,n,r){return s|=n,Re(o,e),D.setValue(o,{type:t.Unbuildable,reason:r+" errors"}),De(o,u),s}}function De(e,n){c.afterProgramEmitAndDiagnostics&&c.afterProgramEmitAndDiagnostics(n),y.watch&&(n.releaseProgram(),R.setValue(e,n))}function Le(n){if(y.dry)return te(e.Diagnostics.A_non_dry_build_would_build_project_0,n.options.configFilePath);var a=ke(n,r,e.Diagnostics.Updating_output_timestamps_of_project_0);D.setValue(n.options.configFilePath,{type:t.UpToDate,newestDeclarationFileContentChangedTime:a})}function ke(n,t,r,a){var i=T(n);if(!a||i.length!==a.getSize()){y.verbose&&te(r,n.options.configFilePath);for(var o=c.now?c.now():new Date,s=0,l=i;s<l.length;s++){var u=l[s];a&&a.hasKey(u)||(f(u)&&(t=m(t,c.getModifiedTime(u)||e.missingFileModifiedTime)),c.setModifiedTime(u,o),n.options.listEmittedFiles&&k("TSFILE: "+u))}}return t}function Ne(){for(var e=me(),n=[],t=0,r=e.buildQueue;t<r.length;t++){var a=r[t],i=ne(a);if(void 0!==i)for(var o=T(i),s=0,l=o;s<l.length;s++){var u=l[s];c.fileExists(u)&&n.push(u)}else Pe(a)}return n}function Ie(){var n=Ne();if(y.dry)return te(e.Diagnostics.A_non_dry_build_would_delete_the_following_files_Colon_0,n.map(function(e){return"\r\n * "+e}).join("")),e.ExitStatus.Success;for(var t=0,r=n;t<r.length;t++){var a=r[t];c.deleteFile(a)}return e.ExitStatus.Success}function we(n){return E(e.resolvePath(c.getCurrentDirectory(),n))}function Me(e){return e.map(we)}function Oe(){y.watch&&re(e.Diagnostics.Starting_compilation_in_watch_mode);var r=N,a=w.getSourceFile,i=e.changeCompilerHostLikeToUseCache(c,Q,function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return a.call.apply(a,[w].concat(e))}),o=i.originalReadFile,s=i.originalFileExists,l=i.originalDirectoryExists,d=i.originalCreateDirectory,p=i.originalWriteFile,m=i.getSourceFileWithCache,f=i.readFileWithCache;N=f,w.getSourceFile=m;var _=me();Ge(_);for(var g=!1,v=0,h=_.buildQueue;v<h.length;v++){var b=h[v],E=ne(b);if(void 0===E){Pe(b),g=!0;break}var T=E.errors,S=fe(E);Be(b,S);var x=E.options.configFilePath;if(S.type!==t.UpToDate||y.force)if(S.type!==t.UpToDateWithUpstreamTypes||y.force)if(S.type!==t.UpstreamBlocked)if(S.type!==t.ContainerOnly){var C=Ae(b);g=g||!!(C&n.AnyErrors)}else Re(b,T);else Re(b,T),y.verbose&&te(e.Diagnostics.Skipping_build_of_project_0_because_its_dependency_1_has_errors,x,S.upstreamProjectName);else Re(b,T),Le(E);else Re(b,T),u.dry&&te(e.Diagnostics.Project_0_is_up_to_date,x)}return Se(),c.readFile=o,c.fileExists=s,c.directoryExists=l,c.createDirectory=d,c.writeFile=p,w.getSourceFile=a,N=r,g?e.ExitStatus.DiagnosticsPresent_OutputsSkipped:e.ExitStatus.Success}function Pe(e){Re(e,[C.getValue(e)])}function Re(e,n){Fe(n),V.setValue(e,!0),F.setValue(e,n)}function Fe(e){e.forEach(function(e){return c.reportDiagnostic(e)})}function Ge(n){y.verbose&&te(e.Diagnostics.Projects_in_this_build_Colon_0,n.buildQueue.map(function(e){return"\r\n    * "+Ve(e)}).join(""))}function Ve(n){return e.convertToRelativePath(n,c.getCurrentDirectory(),function(e){return w.getCanonicalFileName(e)})}function Be(e,n){if(y.verbose)return x(e,n,Ve,te)}}function E(n){return e.fileExtensionIs(n,".json")?n:e.combinePaths(n,"tsconfig.json")}function T(e){if(e.options.outFile||e.options.out)return d(e);for(var n=[],t=0,r=e.fileNames;t<r.length;t++){var a=r[t];n.push.apply(n,u(a,e))}return n}function S(n){if(n.options.outFile||n.options.out)return e.first(d(n));for(var t=0,r=n.fileNames;t<r.length;t++){var a=r[t],i=u(a,n);if(i.length)return e.first(i)}return e.Debug.fail("project "+n.options.configFilePath+" expected to have at least one output")}function x(n,r,a,i){switch(r.type){case t.OutOfDateWithSelf:return i(e.Diagnostics.Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2,a(n),a(r.outOfDateOutputFileName),a(r.newerInputFileName));case t.OutOfDateWithUpstream:return i(e.Diagnostics.Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2,a(n),a(r.outOfDateOutputFileName),a(r.newerProjectName));case t.OutputMissing:return i(e.Diagnostics.Project_0_is_out_of_date_because_output_file_1_does_not_exist,a(n),a(r.missingOutputFileName));case t.UpToDate:if(void 0!==r.newestInputFileTime)return i(e.Diagnostics.Project_0_is_up_to_date_because_newest_input_1_is_older_than_oldest_output_2,a(n),a(r.newestInputFileName||""),a(r.oldestOutputFileName||""));break;case t.UpToDateWithUpstreamTypes:return i(e.Diagnostics.Project_0_is_up_to_date_with_d_ts_files_from_its_dependencies,a(n));case t.UpstreamOutOfDate:return i(e.Diagnostics.Project_0_is_out_of_date_because_its_dependency_1_is_out_of_date,a(n),a(r.upstreamProjectName));case t.UpstreamBlocked:return i(e.Diagnostics.Project_0_can_t_be_built_because_its_dependency_1_has_errors,a(n),a(r.upstreamProjectName));case t.Unbuildable:return i(e.Diagnostics.Failed_to_parse_file_0_Colon_1,a(n),r.reason);case t.ContainerOnly:case t.ComputingUpstream:break;default:e.assertType(r)}}(function(e){e[e["None"]=0]="None",e[e["Success"]=1]="Success",e[e["DeclarationOutputUnchanged"]=2]="DeclarationOutputUnchanged",e[e["ConfigFileErrors"]=4]="ConfigFileErrors",e[e["SyntaxErrors"]=8]="SyntaxErrors",e[e["TypeErrors"]=16]="TypeErrors",e[e["DeclarationEmitErrors"]=32]="DeclarationEmitErrors",e[e["EmitErrors"]=64]="EmitErrors",e[e["AnyErrors"]=124]="AnyErrors"})(n||(n={})),function(e){e[e["Unbuildable"]=0]="Unbuildable",e[e["UpToDate"]=1]="UpToDate",e[e["UpToDateWithUpstreamTypes"]=2]="UpToDateWithUpstreamTypes",e[e["OutputMissing"]=3]="OutputMissing",e[e["OutOfDateWithSelf"]=4]="OutOfDateWithSelf",e[e["OutOfDateWithUpstream"]=5]="OutOfDateWithUpstream",e[e["UpstreamOutOfDate"]=6]="UpstreamOutOfDate",e[e["UpstreamBlocked"]=7]="UpstreamBlocked",e[e["ComputingUpstream"]=8]="ComputingUpstream",e[e["ContainerOnly"]=9]="ContainerOnly"}(t=e.UpToDateStatusType||(e.UpToDateStatusType={})),e.getOutputDeclarationFileName=c,e.createBuilderStatusReporter=_,e.createSolutionBuilderHost=v,e.createSolutionBuilderWithWatchHost=y,e.createSolutionBuilder=b,e.resolveConfigFileProjectName=E,e.getAllProjectOutputs=T,e.formatUpToDateStatus=x}(u||(u={})),function(e){function n(n){return t(e.removeFileExtension(e.getBaseFileName(n)),y(n))}function t(e,n){return i(e,n,a())}function a(){var e=[],n=[];return function(t,r,a,i){if(-1!==e.indexOf(t)||n.length>4)return i(-1!==e.indexOf(t),n);e.push(t),n.push(r);var o=a();return n.pop(),e.pop(),o}}function i(n,t,r){return r(t,n,function(){if("function"===typeof t)return o(t,n,r);if("object"===typeof t){var a=c(n,t,r);if(void 0!==a)return a;var s=p(t),l=Object.getPrototypeOf(t)!==Object.prototype,u=e.flatMap(s,function(e){var n=e.key,t=e.value;return i(n,t,r)});return{kind:3,name:n,hasNontrivialPrototype:l,members:u}}return{kind:0,name:n,typeName:_(t)?"any":typeof t}},function(e,t){return g(n," "+(e?"Circular reference":"Too-deep object hierarchy")+" from "+t.join("."))})}function o(n,t,r){var a=l(n,r),o=e.flatMap(p(n),function(e){var n=e.key,t=e.value;return i(n,t,r)}),s=e.cast(Function.prototype.toString.call(n),e.isString),c=e.stringContains(s,"{ [native code] }")?m(n):s;return{kind:2,name:t,source:c,namespaceMembers:o,prototypeMembers:a}}(function(e){e[e["Const"]=0]="Const",e[e["Array"]=1]="Array",e[e["FunctionOrClass"]=2]="FunctionOrClass",e[e["Object"]=3]="Object"})(e.ValueKind||(e.ValueKind={})),e.inspectModule=n,e.inspectValue=t;var s=e.memoize(function(){for(var n=e.createMap(),t=0,a=p(r);t<a.length;t++){var i=a[t],o=i.key,s=i.value;"function"===typeof s&&"object"===typeof s.prototype&&s!==Object&&n.set(o,s)}return n});function c(n,t,r){return e.isArray(t)?{name:n,kind:1,inner:t.length&&i("element",e.first(t),r)||g(n)}:e.forEachEntry(s(),function(e,r){return t instanceof e?{kind:0,name:n,typeName:r}:void 0})}function l(n,t){var r=n.prototype;return"object"!==typeof r||null===r?e.emptyArray:e.mapDefined(p(r),function(e){var n=e.key,r=e.value;return"constructor"===n?void 0:i(n,r,t)})}var u=["arguments","caller","constructor","eval","super_"],d=Object.getOwnPropertyNames(e.noop);function p(n){var t=e.createMap(),r=[],a=n;while(!_(a)&&a!==Object.prototype&&a!==Function.prototype){for(var i=0,o=Object.getOwnPropertyNames(a);i<o.length;i++){var s=o[i];if(!v(s)&&-1===u.indexOf(s)&&("function"!==typeof n||-1===d.indexOf(s))&&e.addToSeen(t,s)){var c=f(a,s);"toString"===s&&"function"===typeof c&&0===c.length||r.push({key:s,value:c})}}a=Object.getPrototypeOf(a)}return r.sort(function(n,t){return e.compareStringsCaseSensitive(n.key,t.key)})}function m(n){return e.tryCast(f(n,"length"),e.isNumber)||0}function f(e,n){var t=Object.getOwnPropertyDescriptor(e,n);return t&&t.value}function _(e){return null==e}function g(e,n){return{kind:0,name:e,typeName:"any",comment:n}}function v(n){return e.startsWith(n,"_")}function y(e){try{return}catch(n){return}}e.isJsPrivate=v}(u||(u={})),function(e){(function(n){function t(n){return e.sys.args.indexOf(n)>=0}function r(n){var t=e.sys.args.indexOf(n);return t>=0&&t<e.sys.args.length-1?e.sys.args[t+1]:void 0}function a(){var e=new Date;return e.getHours()+":"+e.getMinutes()+":"+e.getSeconds()+"."+e.getMilliseconds()}n.ActionSet="action::set",n.ActionInvalidate="action::invalidate",n.ActionPackageInstalled="action::packageInstalled",n.ActionValueInspected="action::valueInspected",n.EventTypesRegistry="event::typesRegistry",n.EventBeginInstallTypes="event::beginInstallTypes",n.EventEndInstallTypes="event::endInstallTypes",n.EventInitializationFailed="event::initializationFailed",function(e){e.GlobalCacheLocation="--globalTypingsCacheLocation",e.LogFile="--logFile",e.EnableTelemetry="--enableTelemetry",e.TypingSafeListLocation="--typingSafeListLocation",e.TypesMapLocation="--typesMapLocation",e.NpmLocation="--npmLocation"}(n.Arguments||(n.Arguments={})),n.hasArgument=t,n.findArgument=r,n.nowString=a})(e.server||(e.server={}))}(u||(u={})),function(e){(function(n){function t(n,t){var r=new e.Version(e.getProperty(t,"ts"+e.versionMajorMinor)||e.getProperty(t,"latest"));return r.compareTo(n.version)<=0}function r(n,t){var r=e.readConfigFile(t,function(e){return n.readFile(e)});return e.createMapFromTemplate(r.config)}function a(n,t){var r=e.readConfigFile(t,function(e){return n.readFile(e)});if(r.config)return e.createMapFromTemplate(r.config.simpleMap)}function i(r,a,i,o,s,c,l,u,d){if(!l||!l.enable)return{cachedTypingPaths:[],newTypingNames:[],filesToWatch:[]};var p=e.createMap();i=e.mapDefined(i,function(n){var t=e.normalizePath(n);if(e.hasJSFileExtension(t))return t});var m=[];l.include&&C(l.include,"Explicitly included types");var f=l.exclude||[],_=e.arrayToSet(i,e.getDirectoryPath);if(_.set(o,!0),_.forEach(function(n,t){var r=e.combinePaths(t,"package.json");A(r,m);var a=e.combinePaths(t,"bower.json");A(a,m);var i=e.combinePaths(t,"bower_components");L(i,m);var o=e.combinePaths(t,"node_modules");L(o,m)}),D(i),u){var g=e.deduplicate(u.map(function(e){return n.nodeCoreModules.has(e)?"node":e}),e.equateStringsCaseSensitive,e.compareStringsCaseSensitive);C(g,"Inferred typings from unresolved imports")}c.forEach(function(e,n){var r=d.get(n);p.has(n)&&void 0===p.get(n)&&void 0!==r&&t(e,r)&&p.set(n,e.typingLocation)});for(var v=0,y=f;v<y.length;v++){var h=y[v],b=p.delete(h);b&&a&&a("Typing for "+h+" is in exclude list, will be ignored.")}var E=[],T=[];p.forEach(function(e,n){void 0!==e?T.push(e):E.push(n)});var S={cachedTypingPaths:T,newTypingNames:E,filesToWatch:m};return a&&a("Result: "+JSON.stringify(S)),S;function x(e){p.has(e)||p.set(e,void 0)}function C(n,t){a&&a(t+": "+JSON.stringify(n)),e.forEach(n,x)}function A(n,t){if(r.fileExists(n)){t.push(n);var a=e.readConfigFile(n,function(e){return r.readFile(e)}).config,i=e.flatMap([a.dependencies,a.devDependencies,a.optionalDependencies,a.peerDependencies],e.getOwnKeys);C(i,"Typing names in '"+n+"' dependencies")}}function D(n){var t=e.mapDefined(n,function(n){if(e.hasJSFileExtension(n)){var t=e.removeFileExtension(e.getBaseFileName(n.toLowerCase())),r=e.removeMinAndVersionNumbers(t);return s.get(r)}});t.length&&C(t,"Inferred typings from file names");var r=e.some(n,function(n){return e.fileExtensionIs(n,".jsx")});r&&(a&&a("Inferred 'react' typings due to presence of '.jsx' extension"),x("react"))}function L(n,t){if(t.push(n),r.directoryExists(n)){var i=r.readDirectory(n,[".json"],void 0,void 0,2);a&&a("Searching for typing names in "+n+"; all files: "+JSON.stringify(i));for(var o=[],s=0,c=i;s<c.length;s++){var l=c[s],u=e.normalizePath(l),d=e.getBaseFileName(u);if("package.json"===d||"bower.json"===d){var m=e.readConfigFile(u,function(e){return r.readFile(e)}),f=m.config;if(("package.json"!==d||!f._requiredBy||0!==e.filter(f._requiredBy,function(e){return"#"===e[0]||"/"===e}).length)&&f.name){var _=f.types||f.typings;if(_){var g=e.getNormalizedAbsolutePath(_,e.getDirectoryPath(u));a&&a("    Package '"+f.name+"' provides its own types."),p.set(f.name,g)}else o.push(f.name)}}}C(o,"    Found package names")}}}n.isTypingUpToDate=t,n.nodeCoreModuleList=["assert","async_hooks","buffer","child_process","cluster","console","constants","crypto","dgram","dns","domain","events","fs","http","https","http2","inspector","net","os","path","perf_hooks","process","punycode","querystring","readline","repl","stream","string_decoder","timers","tls","tty","url","util","v8","vm","zlib"],n.nodeCoreModules=e.arrayToSet(n.nodeCoreModuleList),n.loadSafeList=r,n.loadTypesMap=a,n.discoverTypings=i,function(e){e[e["Ok"]=0]="Ok",e[e["ScopedPackagesNotSupported"]=1]="ScopedPackagesNotSupported",e[e["EmptyName"]=2]="EmptyName",e[e["NameTooLong"]=3]="NameTooLong",e[e["NameStartsWithDot"]=4]="NameStartsWithDot",e[e["NameStartsWithUnderscore"]=5]="NameStartsWithUnderscore",e[e["NameContainsNonURISafeCharacters"]=6]="NameContainsNonURISafeCharacters"}(n.PackageNameValidationResult||(n.PackageNameValidationResult={}));var o=214;function s(e){return e?e.length>o?3:46===e.charCodeAt(0)?4:95===e.charCodeAt(0)?5:/^@[^/]+\/[^/]+$/.test(e)?1:encodeURIComponent(e)!==e?6:0:2}function c(n,t){switch(n){case 2:return"Package name '"+t+"' cannot be empty";case 3:return"Package name '"+t+"' should be less than "+o+" characters";case 4:return"Package name '"+t+"' cannot start with '.'";case 5:return"Package name '"+t+"' cannot start with '_'";case 1:return"Package '"+t+"' is scoped and currently is not supported";case 6:return"Package name '"+t+"' contains non URI safe characters";case 0:return e.Debug.fail();default:throw e.Debug.assertNever(n)}}n.validatePackageName=s,n.renderPackageNameValidationFailure=c})(e.JsTyping||(e.JsTyping={}))}(u||(u={})),function(e){var n;function t(e){return{indentSize:4,tabSize:4,newLineCharacter:e||"\n",convertTabsToSpaces:!0,indentStyle:n.Smart,insertSpaceAfterConstructor:!1,insertSpaceAfterCommaDelimiter:!0,insertSpaceAfterSemicolonInForStatements:!0,insertSpaceBeforeAndAfterBinaryOperators:!0,insertSpaceAfterKeywordsInControlFlowStatements:!0,insertSpaceAfterFunctionKeywordForAnonymousFunctions:!1,insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis:!1,insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets:!1,insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces:!0,insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces:!1,insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces:!1,insertSpaceBeforeFunctionParenthesis:!1,placeOpenBraceOnNewLineForFunctions:!1,placeOpenBraceOnNewLineForControlBlocks:!1}}(function(e){var n=function(){function e(e){this.text=e}return e.prototype.getText=function(e,n){return 0===e&&n===this.text.length?this.text:this.text.substring(e,n)},e.prototype.getLength=function(){return this.text.length},e.prototype.getChangeRange=function(){},e}();function t(e){return new n(e)}e.fromString=t})(e.ScriptSnapshot||(e.ScriptSnapshot={})),e.emptyOptions={},function(e){e["none"]="none",e["definition"]="definition",e["reference"]="reference",e["writtenReference"]="writtenReference"}(e.HighlightSpanKind||(e.HighlightSpanKind={})),function(e){e[e["None"]=0]="None",e[e["Block"]=1]="Block",e[e["Smart"]=2]="Smart"}(n=e.IndentStyle||(e.IndentStyle={})),e.getDefaultFormatCodeSettings=t,e.testFormatSettings=t("\n"),function(e){e[e["aliasName"]=0]="aliasName",e[e["className"]=1]="className",e[e["enumName"]=2]="enumName",e[e["fieldName"]=3]="fieldName",e[e["interfaceName"]=4]="interfaceName",e[e["keyword"]=5]="keyword",e[e["lineBreak"]=6]="lineBreak",e[e["numericLiteral"]=7]="numericLiteral",e[e["stringLiteral"]=8]="stringLiteral",e[e["localName"]=9]="localName",e[e["methodName"]=10]="methodName",e[e["moduleName"]=11]="moduleName",e[e["operator"]=12]="operator",e[e["parameterName"]=13]="parameterName",e[e["propertyName"]=14]="propertyName",e[e["punctuation"]=15]="punctuation",e[e["space"]=16]="space",e[e["text"]=17]="text",e[e["typeParameterName"]=18]="typeParameterName",e[e["enumMemberName"]=19]="enumMemberName",e[e["functionName"]=20]="functionName",e[e["regularExpressionLiteral"]=21]="regularExpressionLiteral"}(e.SymbolDisplayPartKind||(e.SymbolDisplayPartKind={})),function(e){e["Comment"]="comment",e["Region"]="region",e["Code"]="code",e["Imports"]="imports"}(e.OutliningSpanKind||(e.OutliningSpanKind={})),function(e){e[e["JavaScript"]=0]="JavaScript",e[e["SourceMap"]=1]="SourceMap",e[e["Declaration"]=2]="Declaration"}(e.OutputFileType||(e.OutputFileType={})),function(e){e[e["None"]=0]="None",e[e["InMultiLineCommentTrivia"]=1]="InMultiLineCommentTrivia",e[e["InSingleQuoteStringLiteral"]=2]="InSingleQuoteStringLiteral",e[e["InDoubleQuoteStringLiteral"]=3]="InDoubleQuoteStringLiteral",e[e["InTemplateHeadOrNoSubstitutionTemplate"]=4]="InTemplateHeadOrNoSubstitutionTemplate",e[e["InTemplateMiddleOrTail"]=5]="InTemplateMiddleOrTail",e[e["InTemplateSubstitutionPosition"]=6]="InTemplateSubstitutionPosition"}(e.EndOfLineState||(e.EndOfLineState={})),function(e){e[e["Punctuation"]=0]="Punctuation",e[e["Keyword"]=1]="Keyword",e[e["Operator"]=2]="Operator",e[e["Comment"]=3]="Comment",e[e["Whitespace"]=4]="Whitespace",e[e["Identifier"]=5]="Identifier",e[e["NumberLiteral"]=6]="NumberLiteral",e[e["BigIntLiteral"]=7]="BigIntLiteral",e[e["StringLiteral"]=8]="StringLiteral",e[e["RegExpLiteral"]=9]="RegExpLiteral"}(e.TokenClass||(e.TokenClass={})),function(e){e["unknown"]="",e["warning"]="warning",e["keyword"]="keyword",e["scriptElement"]="script",e["moduleElement"]="module",e["classElement"]="class",e["localClassElement"]="local class",e["interfaceElement"]="interface",e["typeElement"]="type",e["enumElement"]="enum",e["enumMemberElement"]="enum member",e["variableElement"]="var",e["localVariableElement"]="local var",e["functionElement"]="function",e["localFunctionElement"]="local function",e["memberFunctionElement"]="method",e["memberGetAccessorElement"]="getter",e["memberSetAccessorElement"]="setter",e["memberVariableElement"]="property",e["constructorImplementationElement"]="constructor",e["callSignatureElement"]="call",e["indexSignatureElement"]="index",e["constructSignatureElement"]="construct",e["parameterElement"]="parameter",e["typeParameterElement"]="type parameter",e["primitiveType"]="primitive type",e["label"]="label",e["alias"]="alias",e["constElement"]="const",e["letElement"]="let",e["directory"]="directory",e["externalModuleName"]="external module name",e["jsxAttribute"]="JSX attribute",e["string"]="string"}(e.ScriptElementKind||(e.ScriptElementKind={})),function(e){e["none"]="",e["publicMemberModifier"]="public",e["privateMemberModifier"]="private",e["protectedMemberModifier"]="protected",e["exportedModifier"]="export",e["ambientModifier"]="declare",e["staticModifier"]="static",e["abstractModifier"]="abstract",e["optionalModifier"]="optional",e["dtsModifier"]=".d.ts",e["tsModifier"]=".ts",e["tsxModifier"]=".tsx",e["jsModifier"]=".js",e["jsxModifier"]=".jsx",e["jsonModifier"]=".json"}(e.ScriptElementKindModifier||(e.ScriptElementKindModifier={})),function(e){e["comment"]="comment",e["identifier"]="identifier",e["keyword"]="keyword",e["numericLiteral"]="number",e["bigintLiteral"]="bigint",e["operator"]="operator",e["stringLiteral"]="string",e["whiteSpace"]="whitespace",e["text"]="text",e["punctuation"]="punctuation",e["className"]="class name",e["enumName"]="enum name",e["interfaceName"]="interface name",e["moduleName"]="module name",e["typeParameterName"]="type parameter name",e["typeAliasName"]="type alias name",e["parameterName"]="parameter name",e["docCommentTagName"]="doc comment tag name",e["jsxOpenTagName"]="jsx open tag name",e["jsxCloseTagName"]="jsx close tag name",e["jsxSelfClosingTagName"]="jsx self closing tag name",e["jsxAttribute"]="jsx attribute",e["jsxText"]="jsx text",e["jsxAttributeStringLiteralValue"]="jsx attribute string literal value"}(e.ClassificationTypeNames||(e.ClassificationTypeNames={})),function(e){e[e["comment"]=1]="comment",e[e["identifier"]=2]="identifier",e[e["keyword"]=3]="keyword",e[e["numericLiteral"]=4]="numericLiteral",e[e["operator"]=5]="operator",e[e["stringLiteral"]=6]="stringLiteral",e[e["regularExpressionLiteral"]=7]="regularExpressionLiteral",e[e["whiteSpace"]=8]="whiteSpace",e[e["text"]=9]="text",e[e["punctuation"]=10]="punctuation",e[e["className"]=11]="className",e[e["enumName"]=12]="enumName",e[e["interfaceName"]=13]="interfaceName",e[e["moduleName"]=14]="moduleName",e[e["typeParameterName"]=15]="typeParameterName",e[e["typeAliasName"]=16]="typeAliasName",e[e["parameterName"]=17]="parameterName",e[e["docCommentTagName"]=18]="docCommentTagName",e[e["jsxOpenTagName"]=19]="jsxOpenTagName",e[e["jsxCloseTagName"]=20]="jsxCloseTagName",e[e["jsxSelfClosingTagName"]=21]="jsxSelfClosingTagName",e[e["jsxAttribute"]=22]="jsxAttribute",e[e["jsxText"]=23]="jsxText",e[e["jsxAttributeStringLiteralValue"]=24]="jsxAttributeStringLiteralValue",e[e["bigintLiteral"]=25]="bigintLiteral"}(e.ClassificationType||(e.ClassificationType={}))}(u||(u={})),function(e){function n(n){switch(n.kind){case 237:return e.isInJSFile(n)&&e.getJSDocEnumTag(n)?7:1;case 151:case 186:case 154:case 153:case 275:case 276:case 156:case 155:case 157:case 158:case 159:case 239:case 196:case 197:case 274:case 267:return 1;case 150:case 241:case 242:case 168:return 2;case 304:return void 0===n.name?3:2;case 278:case 240:return 3;case 244:return e.isAmbientModule(n)?5:1===e.getModuleInstanceState(n)?5:4;case 243:case 252:case 253:case 248:case 249:case 254:case 255:return 7;case 279:return 5}return 7}function t(t){return 279===t.kind?1:254===t.parent.kind||259===t.parent.kind?7:a(t)?r(t):e.isDeclarationName(t)?n(t.parent):c(t)?2:i(t)?4:e.isTypeParameterDeclaration(t.parent)?(e.Debug.assert(e.isJSDocTemplateTag(t.parent.parent)),2):e.isLiteralTypeNode(t.parent)?3:1}function r(n){var t=148===n.kind?n:e.isQualifiedName(n.parent)&&n.parent.right===n?n.parent:void 0;return t&&248===t.parent.kind?7:4}function a(n){while(148===n.parent.kind)n=n.parent;return e.isInternalModuleImportEqualsDeclaration(n.parent)&&n.parent.moduleReference===n}function i(e){return o(e)||s(e)}function o(e){var n=e,t=!0;if(148===n.parent.kind){while(n.parent&&148===n.parent.kind)n=n.parent;t=n.right===e}return 164===n.parent.kind&&!t}function s(e){var n=e,t=!0;if(189===n.parent.kind){while(n.parent&&189===n.parent.kind)n=n.parent;t=n.name===e}if(!t&&211===n.parent.kind&&273===n.parent.parent.kind){var r=n.parent.parent.parent;return 240===r.kind&&109===n.parent.parent.token||241===r.kind&&86===n.parent.parent.token}return!1}function c(n){switch(e.isRightSideOfQualifiedNameOrPropertyAccess(n)&&(n=n.parent),n.kind){case 100:return!e.isExpressionNode(n);case 178:return!0}switch(n.parent.kind){case 164:return!0;case 183:return!n.parent.isTypeOf;case 211:return!e.isExpressionWithTypeArgumentsInClassExtendsClause(n.parent)}return!1}function l(n){return p(n,e.isCallExpression)}function u(n){return p(n,e.isNewExpression)}function d(n){return p(n,e.isCallOrNewExpression)}function p(e,n){var t=m(e);return!!t&&!!t.parent&&n(t.parent)&&t.parent.expression===t}function m(e){return E(e)?e.parent:e}function f(e,n){while(e){if(233===e.kind&&e.label.escapedText===n)return e.label;e=e.parent}}function _(n,t){return!!e.isPropertyAccessExpression(n.expression)&&n.expression.name.text===t}function g(n){return 72===n.kind&&e.isBreakOrContinueStatement(n.parent)&&n.parent.label===n}function v(n){return 72===n.kind&&e.isLabeledStatement(n.parent)&&n.parent.label===n}function y(e){return v(e)||g(e)}function h(n){return e.isJSDocTag(n.parent)&&n.parent.tagName===n}function b(e){return 148===e.parent.kind&&e.parent.right===e}function E(e){return e&&e.parent&&189===e.parent.kind&&e.parent.name===e}function T(e){return 244===e.parent.kind&&e.parent.name===e}function S(n){return 72===n.kind&&e.isFunctionLike(n.parent)&&n.parent.name===n}function x(n){switch(n.parent.kind){case 154:case 153:case 275:case 278:case 156:case 155:case 158:case 159:case 244:return e.getNameOfDeclaration(n.parent)===n;case 190:return n.parent.argumentExpression===n;case 149:return!0;case 182:return 180===n.parent.parent.kind;default:return!1}}function C(n){return e.isExternalModuleImportEqualsDeclaration(n.parent.parent)&&e.getExternalModuleImportEqualsDeclarationExpression(n.parent.parent)===n}function A(n){e.isJSDocTypeAlias(n)&&(n=n.parent.parent);while(1){if(n=n.parent,!n)return;switch(n.kind){case 279:case 156:case 155:case 239:case 196:case 158:case 159:case 240:case 241:case 243:case 244:return n}}}function D(n){switch(n.kind){case 279:return e.isExternalModule(n)?"module":"script";case 244:return"module";case 240:case 209:return"class";case 241:return"interface";case 242:case 297:case 304:return"type";case 243:return"enum";case 237:return i(n);case 186:return i(e.getRootDeclaration(n));case 197:case 239:case 196:return"function";case 158:return"getter";case 159:return"setter";case 156:case 155:return"method";case 154:case 153:return"property";case 162:return"index";case 161:return"construct";case 160:return"call";case 157:return"constructor";case 150:return"type parameter";case 278:return"enum member";case 151:return e.hasModifier(n,92)?"property":"parameter";case 248:case 253:case 257:case 251:return"alias";case 204:var t=e.getAssignmentDeclarationKind(n),r=n.right;switch(t){case 7:case 8:case 9:case 0:return"";case 1:case 2:var a=D(r);return""===a?"const":a;case 3:return e.isFunctionExpression(r)?"method":"property";case 4:return"property";case 5:return e.isFunctionExpression(r)?"method":"property";case 6:return"local class";default:return e.assertType(t),""}case 72:return e.isImportClause(n.parent)?"alias":"";default:return""}function i(n){return e.isVarConst(n)?"const":e.isLet(n)?"let":"var"}}function L(n){switch(n.kind){case 100:return!0;case 72:return e.identifierIsThisKeyword(n)&&151===n.parent.kind;default:return!1}}e.scanner=e.createScanner(6,!0),function(e){e[e["None"]=0]="None",e[e["Value"]=1]="Value",e[e["Type"]=2]="Type",e[e["Namespace"]=4]="Namespace",e[e["All"]=7]="All"}(e.SemanticMeaning||(e.SemanticMeaning={})),e.getMeaningFromDeclaration=n,e.getMeaningFromLocation=t,e.isInRightSideOfInternalImportEqualsDeclaration=a,e.isCallExpressionTarget=l,e.isNewExpressionTarget=u,e.isCallOrNewExpressionTarget=d,e.climbPastPropertyAccess=m,e.getTargetLabel=f,e.hasPropertyAccessExpressionWithName=_,e.isJumpStatementTarget=g,e.isLabelOfLabeledStatement=v,e.isLabelName=y,e.isTagName=h,e.isRightSideOfQualifiedName=b,e.isRightSideOfPropertyAccess=E,e.isNameOfModuleDeclaration=T,e.isNameOfFunctionDeclaration=S,e.isLiteralNameOfPropertyDeclarationOrIndexAccess=x,e.isExpressionOfExternalModuleImportEqualsDeclaration=C,e.getContainerNode=A,e.getNodeKind=D,e.isThis=L;var k=/^\/\/\/\s*</;function N(n,t){var r=e.getLineStarts(t),a=t.getLineAndCharacterOfPosition(n).line;return r[a]}function I(e,n){return P(e.pos,e.end,n)}function w(e,n){return O(e,n.pos)&&O(e,n.end)}function M(e,n){return e.pos<=n&&n<=e.end}function O(e,n){return e.pos<n&&n<e.end}function P(e,n,t){return e<=t.pos&&n>=t.end}function R(e,n,t){return e.pos<=n&&e.end>=t}function F(e,n,t){return V(e.pos,e.end,n,t)}function G(e,n,t,r){return V(e.getStart(n),e.end,t,r)}function V(e,n,t,r){var a=Math.max(e,t),i=Math.min(n,r);return a<i}function B(n,t,r){return e.Debug.assert(n.pos<=t),t<n.end||!K(n,r)}function K(n,t){if(void 0===n||e.nodeIsMissing(n))return!1;switch(n.kind){case 240:case 241:case 243:case 188:case 184:case 168:case 218:case 245:case 246:case 252:case 256:return j(n,19,t);case 274:return K(n.block,t);case 192:if(!n.arguments)return!0;case 191:case 195:case 177:return j(n,21,t);case 165:case 166:return K(n.type,t);case 157:case 158:case 159:case 239:case 196:case 156:case 155:case 161:case 160:case 197:return n.body?K(n.body,t):n.type?K(n.type,t):H(n,21,t);case 244:return!!n.body&&K(n.body,t);case 222:return n.elseStatement?K(n.elseStatement,t):K(n.thenStatement,t);case 221:return K(n.expression,t)||H(n,26,t);case 187:case 185:case 190:case 149:case 170:return j(n,23,t);case 162:return n.type?K(n.type,t):H(n,23,t);case 271:case 272:return!1;case 225:case 226:case 227:case 224:return K(n.statement,t);case 223:return H(n,107,t)?j(n,21,t):K(n.statement,t);case 167:return K(n.exprName,t);case 199:case 198:case 200:case 207:case 208:var r=n;return K(r.expression,t);case 193:return K(n.template,t);case 206:var a=e.lastOrUndefined(n.templateSpans);return K(a,t);case 216:return e.nodeIsPresent(n.literal);case 255:case 249:return e.nodeIsPresent(n.moduleSpecifier);case 202:return K(n.operand,t);case 204:return K(n.right,t);case 205:return K(n.whenFalse,t);default:return!0}}function j(n,t,r){var a=n.getChildren(r);if(a.length){var i=e.last(a);if(i.kind===t)return!0;if(26===i.kind&&1!==a.length)return a[a.length-2].kind===t}return!1}function U(n){var t=q(n);if(t){var r=t.getChildren(),a=e.indexOfNode(r,n);return{listItemIndex:a,list:t}}}function H(e,n,t){return!!W(e,n,t)}function W(n,t,r){return e.find(n.getChildren(r),function(e){return e.kind===t})}function q(n){var t=e.find(n.parent.getChildren(),function(t){return e.isSyntaxList(t)&&I(t,n)});return e.Debug.assert(!t||e.contains(t.getChildren(),n)),t}function z(n,t){return J(n,t,function(n){return e.isPropertyNameLiteral(n)||e.isKeyword(n.kind)})}function J(e,n,t){return Y(e,n,!1,t,!1)}function X(e,n){return Y(e,n,!0,void 0,!1)}function Y(e,n,t,r,a){var i=e;e:while(1){for(var o=0,s=i.getChildren(e);o<s.length;o++){var c=s[o],l=t?c.getFullStart():c.getStart(e,!0);if(l>n)break;var u=c.getEnd();if(n<u||n===u&&(1===c.kind||a)){i=c;continue e}if(r&&u===n){var d=$(n,e,c);if(d&&r(d))return d}}return i}}function Q(n,t){var r=X(n,t);return e.isToken(r)&&t>r.getStart(n)&&t<r.getEnd()?r:$(t,n)}function Z(n,t,r){return a(t);function a(t){return e.isToken(t)&&t.pos===n.end?t:e.firstDefined(t.getChildren(),function(e){var t=e.pos<=n.pos&&e.end>n.end||e.pos===n.end;return t&&fe(e,r)?a(e):void 0})}}function $(n,t,r,a){var i=o(r||t);return e.Debug.assert(!(i&&ie(i))),i;function o(i){if(ee(i)&&1!==i.kind)return i;for(var s=i.getChildren(t),c=0;c<s.length;c++){var l=s[c];if(n<l.end){var u=l.getStart(t,!a),d=u>=n||!fe(l,t)||ie(l);if(d){var p=te(s,c,t);return p&&ne(p,t)}return o(l)}}e.Debug.assert(void 0!==r||279===i.kind||1===i.kind||e.isJSDocCommentContainingNode(i));var m=te(s,s.length,t);return m&&ne(m,t)}}function ee(n){return e.isToken(n)&&!ie(n)}function ne(e,n){if(ee(e))return e;var t=e.getChildren(n),r=te(t,t.length,n);return r&&ne(r,n)}function te(n,t,r){for(var a=t-1;a>=0;a--){var i=n[a];if(ie(i))e.Debug.assert(a>0,"`JsxText` tokens should not be the first child of `JsxElement | JsxSelfClosingElement`");else if(fe(n[a],r))return n[a]}}function re(n,t,r){if(void 0===r&&(r=$(t,n)),r&&e.isStringTextContainingNode(r)){var a=r.getStart(n),i=r.getEnd();if(a<t&&t<i)return!0;if(t===i)return!!r.isUnterminated}return!1}function ae(e,n){var t=X(e,n);return!!t&&(11===t.kind||(28===t.kind&&11===t.parent.kind||(28===t.kind&&270===t.parent.kind||(!(!t||19!==t.kind||270!==t.parent.kind)||28===t.kind&&263===t.parent.kind))))}function ie(n){return e.isJsxText(n)&&n.containsOnlyWhiteSpaces}function oe(n,t){var r=X(n,t);return e.isTemplateLiteralKind(r.kind)&&t>r.getStart(n)}function se(n,t){var r=X(n,t);return!!e.isJsxText(r)||(!(18!==r.kind||!e.isJsxExpression(r.parent)||!e.isJsxElement(r.parent.parent))||!(28!==r.kind||!e.isJsxOpeningLikeElement(r.parent)||!e.isJsxElement(r.parent.parent)))}function ce(e,n,t){var r=e.kind,a=0;while(1){var i=$(e.getFullStart(),t);if(!i)return;if(e=i,e.kind===n){if(0===a)return e;a--}else e.kind===r&&a++}}function le(n,t,r){var a=de(n,t);return void 0!==a&&(e.isPartOfTypeNode(a.called)||0!==ue(a.called,a.nTypeArguments,r).length||le(a.called,t,r))}function ue(n,t,r){var a=r.getTypeAtLocation(n),i=e.isNewExpression(n.parent)?a.getConstructSignatures():a.getCallSignatures();return i.filter(function(e){return!!e.typeParameters&&e.typeParameters.length>=t})}function de(n,t){var r=n,a=0,i=0;while(r){switch(r.kind){case 28:if(r=$(r.getFullStart(),t),!r||!e.isIdentifier(r))return;if(!a)return e.isDeclarationName(r)?void 0:{called:r,nTypeArguments:i};a--;break;case 48:a=3;break;case 47:a=2;break;case 30:a++;break;case 19:if(r=ce(r,18,t),!r)return;break;case 21:if(r=ce(r,20,t),!r)return;break;case 23:if(r=ce(r,22,t),!r)return;break;case 27:i++;break;case 37:case 72:case 10:case 8:case 9:case 102:case 87:case 104:case 86:case 129:case 24:case 50:case 56:case 57:break;default:if(e.isTypeNode(r))break;return}r=$(r.getFullStart(),t)}}function pe(n,t,r){return e.formatting.getRangeOfEnclosingComment(n,t,void 0,r)}function me(n,t){var r=X(n,t);return!!e.findAncestor(r,e.isJSDoc)}function fe(e,n){return 1===e.kind?!!e.jsDoc:0!==e.getWidth(n)}function _e(n){var t=e.isDeclaration(n)?e.getCombinedModifierFlags(n):0,r=[];return 8&t&&r.push("private"),16&t&&r.push("protected"),4&t&&r.push("public"),32&t&&r.push("static"),128&t&&r.push("abstract"),1&t&&r.push("export"),4194304&n.flags&&r.push("declare"),r.length>0?r.join(","):""}function ge(n){return 164===n.kind||191===n.kind?n.typeArguments:e.isFunctionLike(n)||240===n.kind||241===n.kind?n.typeParameters:void 0}function ve(e){return 2===e||3===e}function ye(n){return!(10!==n&&13!==n&&!e.isTemplateLiteralKind(n))}function he(e){return 18<=e&&e<=71}function be(n,t,r){return e.isTemplateLiteralKind(n.kind)&&n.getStart(r)<t&&t<n.end||!!n.isUnterminated&&t===n.end}function Ee(e){switch(e){case 115:case 113:case 114:return!0}return!1}function Te(n){var t=e.clone(n);return e.setConfigFileInOptions(t,n&&n.configFile),t}function Se(e){if(187===e.kind||188===e.kind){if(204===e.parent.kind&&e.parent.left===e&&59===e.parent.operatorToken.kind)return!0;if(227===e.parent.kind&&e.parent.initializer===e)return!0;if(Se(275===e.parent.kind?e.parent.parent:e.parent))return!0}return!1}function xe(e,n){return Ae(e,n,!0)}function Ce(e,n){return Ae(e,n,!1)}function Ae(e,n,t){var r=pe(e,n,void 0);return!!r&&t===k.test(e.text.substring(r.pos,r.end))}function De(n,t){return e.createTextSpanFromBounds(n.getStart(t),n.getEnd())}function Le(n,t){return e.createRange(n.getStart(t),n.end)}function ke(n){return e.createTextSpanFromBounds(n.pos,n.end)}function Ne(n){return e.createRange(n.start,n.start+n.length)}function Ie(n,t,r){return we(e.createTextSpan(n,t),r)}function we(e,n){return{span:e,newText:n}}function Me(n){return e.contains(e.typeKeywords,n)}function Oe(e){return!!(1536&e.flags)&&34===e.name.charCodeAt(0)}function Pe(){var n=[];return function(t){var r=e.getNodeId(t);return!n[r]&&(n[r]=!0)}}function Re(e){return e.getText(0,e.getLength())}function Fe(e,n){for(var t="",r=0;r<n;r++)t+=e;return t}function Ge(e){return e.isTypeParameter()&&e.getConstraint()||e}function Ve(n){return 149===n.kind?e.isStringOrNumericLiteralLike(n.expression)?n.expression.text:void 0:e.getTextOfIdentifierOrLiteral(n)}function Be(e){return e.getSourceFiles().some(function(n){return!n.isDeclarationFile&&!e.isSourceFileFromExternalLibrary(n)&&!!n.externalModuleIndicator})}function Ke(e){return!!e.module||e.target>=2||!!e.noEmit}function je(e){return!!e.useCaseSensitiveFileNames&&e.useCaseSensitiveFileNames()}function Ue(n){return e.createGetCanonicalFileName(je(n))}function He(e,n,t,r){return e||n&&n.length?We(e,n,t,r):void 0}function We(n,t,r,a){return e.createImportDeclaration(void 0,void 0,n||t?e.createImportClause(n,t&&t.length?e.createNamedImports(t):void 0):void 0,"string"===typeof r?qe(r,a):r)}function qe(n,t){return e.createLiteral(n,0===t)}function ze(n,t){return e.isStringDoubleQuoted(n,t)?1:0}function Je(n,t){if(t.quotePreference&&"auto"!==t.quotePreference)return"single"===t.quotePreference?0:1;var r=n.imports&&e.find(n.imports,e.isStringLiteral);return r?ze(r,n):1}function Xe(n){switch(n){case 0:return"'";case 1:return'"';default:return e.Debug.assertNever(n)}}function Ye(n){var t=Qe(n);return void 0===t?void 0:e.unescapeLeadingUnderscores(t)}function Qe(n){return"default"!==n.escapedName?n.escapedName:e.firstDefined(n.declarations,function(n){var t=e.getNameOfDeclaration(n);return t&&72===t.kind?t.escapedText:void 0})}function Ze(n){return e.isBindingElement(n)&&e.isObjectBindingPattern(n.parent)&&e.isIdentifier(n.name)&&!n.propertyName}function $e(e,n){var t=e.getTypeAtLocation(n.parent);return t&&e.getPropertyOfType(t,n.name.text)}function en(n,t,r,a){var i=e.createMap();return o(n);function o(n){if(96&n.flags&&e.addToSeen(i,e.getSymbolId(n)))return e.firstDefined(n.declarations,function(n){return e.firstDefined(e.getAllSuperTypeNodes(n),function(n){var i=r.getTypeAtLocation(n),s=i&&i.symbol&&r.getPropertyOfType(i,t);return i&&s&&(e.firstDefined(r.getRootSymbols(s),a)||o(i.symbol))})})}}function nn(e,n){return en(e.parent,e.name,n,function(e){return!0})||!1}function tn(n,t,r){if(n)while(n.parent){if(e.isSourceFile(n.parent)||!rn(r,n.parent,t))return n;n=n.parent}}function rn(n,t,r){return e.textSpanContainsPosition(n,t.getStart(r))&&t.getEnd()<=e.textSpanEnd(n)}function an(n,t){return n.modifiers&&e.find(n.modifiers,function(e){return e.kind===t})}function on(n,t,r){var a=e.findLast(t.statements,e.isAnyImportSyntax);a?n.insertNodeAfter(t,a,r):n.insertNodeAtTopOfFile(t,r,!0)}function sn(e,n){return!!e&&!!n&&e.start===n.start&&e.length===n.length}function cn(e,n){return e.fileName===n.fileName&&sn(e.textSpan,n.textSpan)}e.getLineStartPositionForPosition=N,e.rangeContainsRange=I,e.rangeContainsRangeExclusive=w,e.rangeContainsPosition=M,e.rangeContainsPositionExclusive=O,e.startEndContainsRange=P,e.rangeContainsStartEnd=R,e.rangeOverlapsWithStartEnd=F,e.nodeOverlapsWithStartEnd=G,e.startEndOverlapsWithStartEnd=V,e.positionBelongsToNode=B,e.findListItemInfo=U,e.hasChildOfKind=H,e.findChildOfKind=W,e.findContainingList=q,e.getTouchingPropertyName=z,e.getTouchingToken=J,e.getTokenAtPosition=X,e.findTokenOnLeftOfPosition=Q,e.findNextToken=Z,e.findPrecedingToken=$,e.isInString=re,e.isInsideJsxElementOrAttribute=ae,e.isInTemplateString=oe,e.isInJSXText=se,e.findPrecedingMatchingToken=ce,e.isPossiblyTypeArgumentPosition=le,e.getPossibleGenericSignatures=ue,e.getPossibleTypeArgumentsInfo=de,e.isInComment=pe,e.hasDocComment=me,e.getNodeModifiers=_e,e.getTypeArgumentOrTypeParameterList=ge,e.isComment=ve,e.isStringOrRegularExpressionOrTemplateLiteral=ye,e.isPunctuation=he,e.isInsideTemplateLiteral=be,e.isAccessibilityModifier=Ee,e.cloneCompilerOptions=Te,e.isArrayLiteralOrObjectLiteralDestructuringPattern=Se,e.isInReferenceComment=xe,e.isInNonReferenceComment=Ce,e.createTextSpanFromNode=De,e.createTextRangeFromNode=Le,e.createTextSpanFromRange=ke,e.createTextRangeFromSpan=Ne,e.createTextChangeFromStartLength=Ie,e.createTextChange=we,e.typeKeywords=[120,146,123,87,129,132,96,135,136,138,139,102,106,141,142,143],e.isTypeKeyword=Me,e.isExternalModuleSymbol=Oe,e.nodeSeenTracker=Pe,e.getSnapshotText=Re,e.repeatString=Fe,e.skipConstraint=Ge,e.getNameFromPropertyName=Ve,e.programContainsEs6Modules=Be,e.compilerOptionsIndicateEs6Modules=Ke,e.hostUsesCaseSensitiveFileNames=je,e.hostGetCanonicalFileName=Ue,e.makeImportIfNecessary=He,e.makeImport=We,e.makeStringLiteral=qe,function(e){e[e["Single"]=0]="Single",e[e["Double"]=1]="Double"}(e.QuotePreference||(e.QuotePreference={})),e.quotePreferenceFromString=ze,e.getQuotePreference=Je,e.getQuoteFromPreference=Xe,e.symbolNameNoDefault=Ye,e.symbolEscapedNameNoDefault=Qe,e.isObjectBindingElementWithoutPropertyName=Ze,e.getPropertySymbolFromBindingElement=$e,e.getPropertySymbolsFromBaseTypes=en,e.isMemberSymbolInBaseType=nn,e.getParentNodeInSpan=tn,e.findModifier=an,e.insertImport=on,e.textSpansEqual=sn,e.documentSpansEqual=cn}(u||(u={})),function(e){function n(e){return e.declarations&&e.declarations.length>0&&151===e.declarations[0].kind}e.isFirstDeclarationOfSymbolParameter=n;var t=r();function r(){var n,t,r,o,s=10*e.defaultMaximumTruncationLength;m();var c=function(n){return u(n,e.SymbolDisplayPartKind.text)};return{displayParts:function(){var t=n.length&&n[n.length-1].text;return o>s&&t&&"..."!==t&&(e.isWhiteSpaceLike(t.charCodeAt(t.length-1))||n.push(i(" ",e.SymbolDisplayPartKind.space)),n.push(i("...",e.SymbolDisplayPartKind.punctuation))),n},writeKeyword:function(n){return u(n,e.SymbolDisplayPartKind.keyword)},writeOperator:function(n){return u(n,e.SymbolDisplayPartKind.operator)},writePunctuation:function(n){return u(n,e.SymbolDisplayPartKind.punctuation)},writeTrailingSemicolon:function(n){return u(n,e.SymbolDisplayPartKind.punctuation)},writeSpace:function(n){return u(n,e.SymbolDisplayPartKind.space)},writeStringLiteral:function(n){return u(n,e.SymbolDisplayPartKind.stringLiteral)},writeParameter:function(n){return u(n,e.SymbolDisplayPartKind.parameterName)},writeProperty:function(n){return u(n,e.SymbolDisplayPartKind.propertyName)},writeLiteral:function(n){return u(n,e.SymbolDisplayPartKind.stringLiteral)},writeSymbol:d,writeLine:p,write:c,writeComment:c,getText:function(){return""},getTextPos:function(){return 0},getColumn:function(){return 0},getLine:function(){return 0},isAtStartOfLine:function(){return!1},rawWrite:e.notImplemented,getIndent:function(){return r},increaseIndent:function(){r++},decreaseIndent:function(){r--},clear:m,trackSymbol:e.noop,reportInaccessibleThisError:e.noop,reportInaccessibleUniqueSymbolError:e.noop,reportPrivateInBaseOfClassExpression:e.noop};function l(){if(!(o>s)&&t){var a=e.getIndentString(r);a&&(o+=a.length,n.push(i(a,e.SymbolDisplayPartKind.space))),t=!1}}function u(e,t){o>s||(l(),o+=e.length,n.push(i(e,t)))}function d(e,t){o>s||(l(),o+=e.length,n.push(a(e,t)))}function p(){o>s||(o+=1,n.push(f()),t=!0)}function m(){n=[],t=!0,r=0,o=0}}function a(t,r){return i(t,a(r));function a(t){var r=t.flags;return 3&r?n(t)?e.SymbolDisplayPartKind.parameterName:e.SymbolDisplayPartKind.localName:4&r?e.SymbolDisplayPartKind.propertyName:32768&r?e.SymbolDisplayPartKind.propertyName:65536&r?e.SymbolDisplayPartKind.propertyName:8&r?e.SymbolDisplayPartKind.enumMemberName:16&r?e.SymbolDisplayPartKind.functionName:32&r?e.SymbolDisplayPartKind.className:64&r?e.SymbolDisplayPartKind.interfaceName:384&r?e.SymbolDisplayPartKind.enumName:1536&r?e.SymbolDisplayPartKind.moduleName:8192&r?e.SymbolDisplayPartKind.methodName:262144&r?e.SymbolDisplayPartKind.typeParameterName:524288&r?e.SymbolDisplayPartKind.aliasName:2097152&r?e.SymbolDisplayPartKind.aliasName:e.SymbolDisplayPartKind.text}}function i(n,t){return{text:n,kind:e.SymbolDisplayPartKind[t]}}function o(){return i(" ",e.SymbolDisplayPartKind.space)}function s(n){return i(e.tokenToString(n),e.SymbolDisplayPartKind.keyword)}function c(n){return i(e.tokenToString(n),e.SymbolDisplayPartKind.punctuation)}function l(n){return i(e.tokenToString(n),e.SymbolDisplayPartKind.operator)}function u(n){var t=e.stringToToken(n);return void 0===t?d(n):s(t)}function d(n){return i(n,e.SymbolDisplayPartKind.text)}e.symbolPart=a,e.displayPart=i,e.spacePart=o,e.keywordPart=s,e.punctuationPart=c,e.operatorPart=l,e.textOrKeywordPart=u,e.textPart=d;var p="\r\n";function m(e,n){return n&&n.newLineCharacter||e.getNewLine&&e.getNewLine()||p}function f(){return i("\n",e.SymbolDisplayPartKind.lineBreak)}function _(e){try{return e(t),t.displayParts()}finally{t.clear()}}function g(e,n,t,r){return void 0===r&&(r=0),_(function(a){e.writeType(n,t,17408|r,a)})}function v(e,n,t,r,a){return void 0===a&&(a=0),_(function(i){e.writeSymbol(n,t,r,8|a,i)})}function y(e,n,t,r){return void 0===r&&(r=0),r|=25632,_(function(a){e.writeSignature(n,t,r,void 0,a)})}function h(n){return!!n.parent&&e.isImportOrExportSpecifier(n.parent)&&n.parent.propertyName===n}function b(e){var n=e.length;return n>=2&&e.charCodeAt(0)===e.charCodeAt(n-1)&&E(e)?e.substring(1,n-1):e}function E(n){return e.isSingleOrDoubleQuote(n.charCodeAt(0))}function T(n,t){for(var r=[],a=2;a<arguments.length;a++)r[a-2]=arguments[a];var i=S(n,t);return e.some(r,function(e){return e===i})}function S(n,t){return e.ensureScriptKind(n,t&&t.getScriptKind&&t.getScriptKind(n))}function x(n,t){return e.getSymbolId(e.skipAlias(n,t))}function C(n,t){while(e.isWhiteSpaceLike(n.charCodeAt(t)))t+=1;return t}function A(n,t){while(t>-1&&e.isWhiteSpaceSingleLine(n.charCodeAt(t)))t-=1;return t+1}function D(e,n){void 0===n&&(n=!0);var t=e&&k(e);return t&&!n&&I(t),t}function L(n,t,r,a,i){var o;if(void 0===t&&(t=!0),e.isIdentifier(n)&&r&&a){var s=a.getSymbolAtLocation(n),c=s&&r.get(String(e.getSymbolId(s)));c&&(o=e.createIdentifier(c.text))}return o||(o=k(n,r,a,i)),o&&!t&&I(o),i&&o&&i(n,o),o}function k(n,t,r,a){var i=t||r||a?e.visitEachChild(n,s,e.nullTransformationContext):e.visitEachChild(n,D,e.nullTransformationContext);if(i===n){var o=e.getSynthesizedClone(n);return e.isStringLiteral(o)?o.textSourceNode=n:e.isNumericLiteral(o)&&(o.numericLiteralFlags=n.numericLiteralFlags),e.setTextRange(o,n)}return i.parent=void 0,i;function s(e){return L(e,!0,t,r,a)}}function N(n,t){return void 0===t&&(t=!0),n&&e.createNodeArray(n.map(function(e){return D(e,t)}),n.hasTrailingComma)}function I(e){w(e),M(e)}function w(e){O(e,512,P)}function M(n){O(n,1024,e.getLastChild)}function O(n,t,r){e.addEmitFlags(n,t);var a=r(n);a&&O(a,t,r)}function P(e){return e.forEachChild(function(e){return e})}function R(n,t){for(var r=n,a=1;!e.isFileLevelUniqueName(t,r);a++)r=n+"_"+a;return r}function F(n,t,r,a){for(var i=0,o=-1,s=0,c=n;s<c.length;s++){var l=c[s],u=l.fileName,d=l.textChanges;e.Debug.assert(u===t);for(var p=0,m=d;p<m.length;p++){var f=m[p],_=f.span,g=f.newText,v=V(g,r);if(-1!==v&&(o=_.start+i+v,!a))return o;i+=g.length-_.length}}return e.Debug.assert(a),e.Debug.assert(o>=0),o}function G(n,t,r,a,i){e.forEachLeadingCommentRange(r.text,n.pos,function(n,o,s,c){3===s?(n+=2,o-=2):n+=2,e.addSyntheticLeadingComment(t,a||s,r.text.slice(n,o),void 0!==i?i:c)})}function V(n,t){if(e.startsWith(n,t))return 0;var r=n.indexOf(" "+t);return-1===r&&(r=n.indexOf("."+t)),-1===r&&(r=n.indexOf('"'+t)),-1===r?-1:r+1}function B(e,n){var t=e.parent;switch(t.kind){case 192:return n.getContextualType(t);case 204:var r=t,a=r.left,i=r.operatorToken,o=r.right;return j(i.kind)?n.getTypeAtLocation(e===o?a:o):n.getContextualType(e);case 271:return t.expression===e?W(t,n):void 0;default:return n.getContextualType(e)}}function K(n,t){if(/^\d+$/.test(n))return n;var r=t.quotePreference||"auto",a=JSON.stringify(n);switch(r){case"auto":case"double":return a;case"single":return"'"+b(a).replace("'","\\'").replace('\\"','"')+"'";default:return e.Debug.assertNever(r)}}function j(e){switch(e){case 35:case 33:case 36:case 34:return!0;default:return!1}}function U(e){switch(e.kind){case 10:case 14:case 206:case 193:return!0;default:return!1}}function H(e){return!!e.getStringIndexType()||!!e.getNumberIndexType()}function W(e,n){return n.getTypeAtLocation(e.parent.parent.expression)}e.getNewLineOrDefaultFromHost=m,e.lineBreakPart=f,e.mapToDisplayParts=_,e.typeToDisplayParts=g,e.symbolToDisplayParts=v,e.signatureToDisplayParts=y,e.isImportOrExportSpecifierName=h,e.stripQuotes=b,e.startsWithQuote=E,e.scriptKindIs=T,e.getScriptKind=S,e.getUniqueSymbolId=x,e.getFirstNonSpaceCharacterPosition=C,e.getPrecedingNonSpaceCharacterPosition=A,e.getSynthesizedDeepClone=D,e.getSynthesizedDeepCloneWithRenames=L,e.getSynthesizedDeepClones=N,e.suppressLeadingAndTrailingTrivia=I,e.suppressLeadingTrivia=w,e.suppressTrailingTrivia=M,e.getUniqueName=R,e.getRenameLocation=F,e.copyComments=G,e.getContextualTypeFromParent=B,e.quote=K,e.isEqualityOperatorKind=j,e.isStringLiteralOrTemplate=U,e.hasIndexSignature=H,e.getSwitchedType=W}(u||(u={})),function(e){function n(){var n=e.createScanner(6,!1);function o(e,n,t){return i(l(e,n,t),e)}function l(i,o,l){var u=0,p=0,m=[],f=c(o),_=f.prefix,g=f.pushTemplate;i=_+i;var v=_.length;g&&m.push(15),n.setText(i);var y=0,h=[],b=0;do{u=n.scan(),e.isTrivia(u)||(S(),p=u);var E=n.getTextPos();if(a(n.getTokenPos(),E,v,d(u),h),E>=i.length){var T=r(n,u,e.lastOrUndefined(m));void 0!==T&&(y=T)}}while(1!==u);function S(){switch(u){case 42:case 64:t[p]||13!==n.reScanSlashToken()||(u=13);break;case 28:72===p&&b++;break;case 30:b>0&&b--;break;case 120:case 138:case 135:case 123:case 139:b>0&&!l&&(u=72);break;case 15:m.push(u);break;case 18:m.length>0&&m.push(u);break;case 19:if(m.length>0){var r=e.lastOrUndefined(m);15===r?(u=n.reScanTemplateToken(),17===u?m.pop():e.Debug.assertEqual(u,16,"Should have been a template middle.")):(e.Debug.assertEqual(r,18,"Should have been an open brace"),m.pop())}break;default:if(!e.isKeyword(u))break;24===p?u=72:e.isKeyword(p)&&e.isKeyword(u)&&!s(p,u)&&(u=72)}}return{endOfLineState:y,spans:h}}return{getClassificationsForLine:o,getEncodedLexicalClassifications:l}}e.createClassifier=n;var t=e.arrayToNumericMap([72,10,8,9,13,100,44,45,21,23,19,102,87],function(e){return e},function(){return!0});function r(n,t,r){switch(t){case 10:if(!n.isUnterminated())return;var a=n.getTokenText(),i=a.length-1,o=0;while(92===a.charCodeAt(i-o))o++;if(0===(1&o))return;return 34===a.charCodeAt(0)?3:2;case 3:return n.isUnterminated()?1:void 0;default:if(e.isTemplateLiteralKind(t)){if(!n.isUnterminated())return;switch(t){case 17:return 5;case 14:return 4;default:return e.Debug.fail("Only 'NoSubstitutionTemplateLiteral's and 'TemplateTail's can be unterminated; got SyntaxKind #"+t)}}return 15===r?6:void 0}}function a(e,n,t,r,a){if(8!==r){0===e&&t>0&&(e+=t);var i=n-e;i>0&&a.push(e-t,i,r)}}function i(n,t){for(var r=[],a=n.spans,i=0,s=0;s<a.length;s+=3){var c=a[s],l=a[s+1],u=a[s+2];if(i>=0){var d=c-i;d>0&&r.push({length:d,classification:e.TokenClass.Whitespace})}r.push({length:l,classification:o(u)}),i=c+l}var p=t.length-i;return p>0&&r.push({length:p,classification:e.TokenClass.Whitespace}),{entries:r,finalLexState:n.endOfLineState}}function o(n){switch(n){case 1:return e.TokenClass.Comment;case 3:return e.TokenClass.Keyword;case 4:return e.TokenClass.NumberLiteral;case 25:return e.TokenClass.BigIntLiteral;case 5:return e.TokenClass.Operator;case 6:return e.TokenClass.StringLiteral;case 8:return e.TokenClass.Whitespace;case 10:return e.TokenClass.Punctuation;case 2:case 11:case 12:case 13:case 14:case 15:case 16:case 9:case 17:return e.TokenClass.Identifier;default:return}}function s(n,t){if(!e.isAccessibilityModifier(n))return!0;switch(t){case 126:case 137:case 124:case 116:return!0;default:return!1}}function c(n){switch(n){case 3:return{prefix:'"\\\n'};case 2:return{prefix:"'\\\n"};case 1:return{prefix:"/*\n"};case 4:return{prefix:"`\n"};case 5:return{prefix:"}\n",pushTemplate:!0};case 6:return{prefix:"",pushTemplate:!0};case 0:return{prefix:""};default:return e.Debug.assertNever(n)}}function l(e){switch(e){case 40:case 42:case 43:case 38:case 39:case 46:case 47:case 48:case 28:case 30:case 31:case 32:case 94:case 93:case 119:case 33:case 34:case 35:case 36:case 49:case 51:case 50:case 54:case 55:case 70:case 69:case 71:case 66:case 67:case 68:case 60:case 61:case 62:case 64:case 65:case 59:case 27:return!0;default:return!1}}function u(e){switch(e){case 38:case 39:case 53:case 52:case 44:case 45:return!0;default:return!1}}function d(n){if(e.isKeyword(n))return 3;if(l(n)||u(n))return 5;if(n>=18&&n<=71)return 10;switch(n){case 8:return 4;case 9:return 25;case 10:return 6;case 13:return 7;case 7:case 3:case 2:return 1;case 5:case 4:return 8;case 72:default:return e.isTemplateLiteralKind(n)?6:2}}function p(e,n,t,r,a){return y(f(e,n,t,r,a))}function m(e,n){switch(n){case 244:case 240:case 241:case 239:e.throwIfCancellationRequested()}}function f(n,t,r,a,i){var o=[];return r.forEachChild(function o(c){if(c&&e.textSpanIntersectsWith(i,c.pos,c.getFullWidth())){if(m(t,c.kind),e.isIdentifier(c)&&!e.nodeIsMissing(c)&&a.has(c.escapedText)){var l=n.getSymbolAtLocation(c),u=l&&_(l,e.getMeaningFromLocation(c),n);u&&s(c.getStart(r),c.getEnd(),u)}c.forEachChild(o)}}),{spans:o,endOfLineState:0};function s(e,n,t){o.push(e),o.push(n-e),o.push(t)}}function _(e,n,t){var r=e.getFlags();return 0===(2885600&r)?void 0:32&r?11:384&r?12:524288&r?16:1536&r?4&n||1&n&&g(e)?14:void 0:2097152&r?_(t.getAliasedSymbol(e),n,t):2&n?64&r?13:262144&r?15:void 0:void 0}function g(n){return e.some(n.declarations,function(n){return e.isModuleDeclaration(n)&&1===e.getModuleInstanceState(n)})}function v(e){switch(e){case 1:return"comment";case 2:return"identifier";case 3:return"keyword";case 4:return"number";case 25:return"bigint";case 5:return"operator";case 6:return"string";case 8:return"whitespace";case 9:return"text";case 10:return"punctuation";case 11:return"class name";case 12:return"enum name";case 13:return"interface name";case 14:return"module name";case 15:return"type parameter name";case 16:return"type alias name";case 17:return"parameter name";case 18:return"doc comment tag name";case 19:return"jsx open tag name";case 20:return"jsx close tag name";case 21:return"jsx self closing tag name";case 22:return"jsx attribute";case 23:return"jsx text";case 24:return"jsx attribute string literal value";default:return}}function y(n){e.Debug.assert(n.spans.length%3===0);for(var t=n.spans,r=[],a=0;a<t.length;a+=3)r.push({textSpan:e.createTextSpan(t[a],t[a+1]),classificationType:v(t[a+2])});return r}function h(e,n,t){return y(b(e,n,t))}function b(n,t,r){var a=r.start,i=r.length,o=e.createScanner(6,!1,t.languageVariant,t.text),s=e.createScanner(6,!1,t.languageVariant,t.text),c=[];return E(t),{spans:c,endOfLineState:0};function l(e,n,t){c.push(e),c.push(n),c.push(t)}function u(n){o.setTextPos(n.pos);while(1){var r=o.getTextPos();if(!e.couldStartTrivia(t.text,r))return r;var a=o.scan(),i=o.getTextPos(),s=i-r;if(!e.isTrivia(a))return r;switch(a){case 4:case 5:continue;case 2:case 3:d(n,a,r,s),o.setTextPos(i);continue;case 7:var c=t.text,u=c.charCodeAt(r);if(60===u||62===u){l(r,s,1);continue}e.Debug.assert(124===u||61===u),g(c,r,i);break;case 6:break;default:e.Debug.assertNever(a)}}}function d(n,r,a,i){if(3===r){var o=e.parseIsolatedJSDocComment(t.text,a,i);if(o&&o.jsDoc)return o.jsDoc.parent=n,void f(o.jsDoc)}p(a,i)}function p(e,n){l(e,n,1)}function f(e){var n=e.pos;if(e.tags)for(var t=0,r=e.tags;t<r.length;t++){var a=r[t];switch(a.pos!==n&&p(n,a.pos-n),l(a.pos,1,10),l(a.tagName.pos,a.tagName.end-a.tagName.pos,18),n=a.tagName.end,a.kind){case 299:i(a);break;case 303:_(a),n=a.end;break;case 302:E(a.typeExpression),n=a.end;break;case 300:E(a.typeExpression),n=a.end;break}}return void(n!==e.end&&p(n,e.end-n));function i(e){e.isNameFirst&&(p(n,e.name.pos-n),l(e.name.pos,e.name.end-e.name.pos,17),n=e.name.end),e.typeExpression&&(p(n,e.typeExpression.pos-n),E(e.typeExpression),n=e.typeExpression.end),e.isNameFirst||(p(n,e.name.pos-n),l(e.name.pos,e.name.end-e.name.pos,17),n=e.name.end)}}function _(e){for(var n=0,t=e.getChildren();n<t.length;n++){var r=t[n];E(r)}}function g(n,t,r){var a;for(a=t;a<r;a++)if(e.isLineBreak(n.charCodeAt(a)))break;l(t,a-t,1),s.setTextPos(a);while(s.getTextPos()<r)v()}function v(){var e=s.getTextPos(),n=s.scan(),t=s.getTextPos(),r=b(n);r&&l(e,t-e,r)}function y(n){if(e.isJSDoc(n))return!0;if(e.nodeIsMissing(n))return!0;var t=h(n);if(!e.isToken(n)&&11!==n.kind&&void 0===t)return!1;var r=11===n.kind?n.pos:u(n),a=n.end-r;if(e.Debug.assert(a>=0),a>0){var i=t||b(n.kind,n);i&&l(r,a,i)}return!0}function h(e){switch(e.parent&&e.parent.kind){case 262:if(e.parent.tagName===e)return 19;break;case 263:if(e.parent.tagName===e)return 20;break;case 261:if(e.parent.tagName===e)return 21;break;case 267:if(e.parent.name===e)return 22;break}}function b(n,t){if(e.isKeyword(n))return 3;if((28===n||30===n)&&t&&e.getTypeArgumentOrTypeParameterList(t.parent))return 10;if(e.isPunctuation(n)){if(t){var r=t.parent;if(59===n&&(237===r.kind||154===r.kind||151===r.kind||267===r.kind))return 5;if(204===r.kind||202===r.kind||203===r.kind||205===r.kind)return 5}return 10}if(8===n)return 4;if(9===n)return 25;if(10===n)return 267===t.parent.kind?24:6;if(13===n)return 6;if(e.isTemplateLiteralKind(n))return 6;if(11===n)return 23;if(72===n){if(t)switch(t.parent.kind){case 240:return t.parent.name===t?11:void 0;case 150:return t.parent.name===t?15:void 0;case 241:return t.parent.name===t?13:void 0;case 243:return t.parent.name===t?12:void 0;case 244:return t.parent.name===t?14:void 0;case 151:return t.parent.name===t?e.isThisIdentifier(t)?3:17:void 0}return 2}}function E(r){if(r&&e.decodedTextSpanIntersectsWith(a,i,r.pos,r.getFullWidth())){m(n,r.kind);for(var o=0,s=r.getChildren(t);o<s.length;o++){var c=s[o];y(c)||E(c)}}}}e.getSemanticClassifications=p,e.getEncodedSemanticClassifications=f,e.getSyntacticClassifications=h,e.getEncodedSyntacticClassifications=b}(u||(u={})),function(e){(function(n){(function(t){function r(n,t,r,i,o,c,l,d){if(e.isInReferenceComment(n,t)){var p=O(n,t,o,c);return p&&s(p)}if(e.isInString(n,t,r))return r&&e.isStringLiteralLike(r)?a(u(n,r,t,i,o,c),n,i,l,d):void 0}function a(t,r,a,i,o){if(void 0!==t)switch(t.kind){case 0:return s(t.paths);case 1:var c=[];return n.getCompletionEntriesFromSymbols(t.symbols,c,r,r,a,6,i,4,o),{isGlobalCompletion:!1,isMemberCompletion:!0,isNewIdentifierLocation:t.hasIndexSignature,entries:c};case 2:c=t.types.map(function(e){return{name:e.value,kindModifiers:"",kind:"string",sortText:"0"}});return{isGlobalCompletion:!1,isMemberCompletion:!1,isNewIdentifierLocation:t.isNewIdentifier,entries:c};default:return e.Debug.assertNever(t)}}function i(n,t,r,a,i,s,c,l){if(a&&e.isStringLiteralLike(a)){var d=u(t,a,r,i,s,c);return d&&o(n,a,d,t,i,l)}}function o(t,r,a,i,o,s){switch(a.kind){case 0:var l=e.find(a.paths,function(e){return e.name===t});return l&&n.createCompletionDetails(t,c(l.extension),l.kind,[e.textPart(t)]);case 1:l=e.find(a.symbols,function(e){return e.name===t});return l&&n.createCompletionDetailsForSymbol(l,o,i,r,s);case 2:return e.find(a.types,function(e){return e.value===t})?n.createCompletionDetails(t,"","type",[e.textPart(t)]):void 0;default:return e.Debug.assertNever(a)}}function s(e){var n=!1,t=!0,r=e.map(function(e){var n=e.name,t=e.kind,r=e.span,a=e.extension;return{name:n,kind:t,kindModifiers:c(a),sortText:"0",replacementSpan:r}});return{isGlobalCompletion:n,isMemberCompletion:!1,isNewIdentifierLocation:t,entries:r}}function c(n){switch(n){case".d.ts":return".d.ts";case".js":return".js";case".json":return".json";case".jsx":return".jsx";case".ts":return".ts";case".tsx":return".tsx";case void 0:return"";default:return e.Debug.assertNever(n)}}var l;function u(n,t,r,a,i,o){var s=t.parent;switch(s.kind){case 182:switch(s.parent.kind){case 164:return{kind:2,types:f(a.getTypeArgumentConstraint(s)),isNewIdentifier:!1};case 180:return m(a.getTypeFromTypeNode(s.parent.objectType));case 183:return{kind:0,paths:y(n,t,i,o,a)};case 173:if(!e.isTypeReferenceNode(s.parent.parent))return;var c=d(s.parent,s),l=f(a.getTypeArgumentConstraint(s.parent)).filter(function(n){return!e.contains(c,n.value)});return{kind:2,types:l,isNewIdentifier:!1};default:return}case 275:return e.isObjectLiteralExpression(s.parent)&&s.name===t?m(a.getContextualType(s.parent)):h();case 190:var u=s,_=u.expression,g=u.argumentExpression;return t===g?m(a.getTypeAtLocation(_)):void 0;case 191:case 192:if(!e.isRequireCall(s,!1)&&!e.isImportCall(s)){var v=e.SignatureHelp.getArgumentInfoForCompletions(t,r,n);return v?p(v,a):h()}case 249:case 255:case 259:return{kind:0,paths:y(n,t,i,o,a)};default:return h()}function h(){return{kind:2,types:f(e.getContextualTypeFromParent(t,a)),isNewIdentifier:!1}}}function d(n,t){return e.mapDefined(n.types,function(n){return n!==t&&e.isLiteralTypeNode(n)&&e.isStringLiteral(n.literal)?n.literal.text:void 0})}function p(n,t){var r=!1,a=e.createMap(),i=[];t.getResolvedSignature(n.invocation,i,n.argumentCount);var o=e.flatMap(i,function(e){if(e.hasRestParameter||!(n.argumentCount>e.parameters.length)){var i=t.getParameterType(e,n.argumentIndex);return r=r||!!(4&i.flags),f(i,a)}});return{kind:2,types:o,isNewIdentifier:r}}function m(n){return n&&{kind:1,symbols:n.getApparentProperties(),hasIndexSignature:e.hasIndexSignature(n)}}function f(n,t){return void 0===t&&(t=e.createMap()),n?(n=e.skipConstraint(n),n.isUnion()?e.flatMap(n.types,function(e){return f(e,t)}):!n.isStringLiteral()||1024&n.flags||!e.addToSeen(t,n.value)?e.emptyArray:[n]):e.emptyArray}function _(e,n,t){return{name:e,kind:n,extension:t}}function g(e){return _(e,"directory",void 0)}function v(e,n,t){var r=V(e,n);return t.map(function(e){var n=e.name,t=e.kind,a=e.extension;return{name:n,kind:t,extension:a,span:r}})}function y(e,n,t,r,a){return v(n.text,n.getStart(e)+1,h(e,n,t,r,a))}function h(n,t,r,a,i){var o=e.normalizeSlashes(t.text),s=n.path,c=e.getDirectoryPath(s);return B(o)||!r.baseUrl&&(e.isRootedDiskPath(o)||e.isUrl(o))?E(o,c,r,a,s):D(o,c,r,a,i)}function b(e,n){return void 0===n&&(n=!1),{extensions:T(e),includeExtensions:n}}function E(e,n,t,r,a){var i=b(t);return t.rootDirs?x(t.rootDirs,e,n,i,t,r,a):C(e,n,i,r,a)}function T(n){var t=e.getSupportedExtensions(n);return n.resolveJsonModule&&e.getEmitModuleResolutionKind(n)===e.ModuleResolutionKind.NodeJs?t.concat(".json"):t}function S(n,t,r,a){n=n.map(function(n){return e.normalizePath(e.isRootedDiskPath(n)?n:e.combinePaths(t,n))});var i=e.firstDefined(n,function(n){return e.containsPath(n,r,t,a)?r.substr(n.length):void 0});return e.deduplicate(n.map(function(n){return e.combinePaths(n,i)}).concat([r]),e.equateStringsCaseSensitive,e.compareStringsCaseSensitive)}function x(n,t,r,a,i,o,s){var c=i.project||o.getCurrentDirectory(),l=!(o.useCaseSensitiveFileNames&&o.useCaseSensitiveFileNames()),u=S(n,c,r,l);return e.flatMap(u,function(e){return C(t,e,a,o,s)})}function C(n,t,r,a,i,o){var s=r.extensions,c=r.includeExtensions;void 0===o&&(o=[]),void 0===n&&(n=""),n=e.normalizeSlashes(n),e.hasTrailingDirectorySeparator(n)||(n=e.getDirectoryPath(n)),""===n&&(n="."+e.directorySeparator),n=e.ensureTrailingDirectorySeparator(n);var l=e.resolvePath(t,n),u=e.hasTrailingDirectorySeparator(l)?l:e.getDirectoryPath(l),d=!(a.useCaseSensitiveFileNames&&a.useCaseSensitiveFileNames());if(!q(a,u))return o;var p=H(a,u,s,void 0,["./*"]);if(p){for(var m=e.createMap(),f=0,v=p;f<v.length;f++){var y=v[f];if(y=e.normalizePath(y),!i||0!==e.comparePaths(y,i,t,d)){var h=c||e.fileExtensionIs(y,".json")?e.getBaseFileName(y):e.removeFileExtension(e.getBaseFileName(y));m.set(h,e.tryGetExtensionFromPath(y))}}m.forEach(function(e,n){o.push(_(n,"script",e))})}var b=U(a,u);if(b)for(var E=0,T=b;E<T.length;E++){var S=T[E],x=e.getBaseFileName(e.normalizePath(S));"@types"!==x&&o.push(g(x))}var C=F(u,a);if(C){var D=e.readJson(C,a),L=D.typesVersions;if("object"===typeof L){var k=e.getPackageJsonTypesVersionsPaths(L),N=k&&k.paths,I=l.slice(e.ensureTrailingDirectorySeparator(u).length);N&&A(o,I,u,s,N,a)}}return o}function A(n,t,r,a,i,o){for(var s in i)if(e.hasProperty(i,s)){var c=i[s];if(c)for(var l=function(e,t,r){n.some(function(n){return n.name===e})||n.push(_(e,t,r))},u=0,d=k(s,c,t,r,a,o);u<d.length;u++){var p=d[u],m=p.name,f=p.kind,g=p.extension;l(m,f,g)}}}function D(n,t,r,a,i){var o=r.baseUrl,s=r.paths,c=[],l=b(r);if(o){var u=r.project||a.getCurrentDirectory(),d=e.normalizePath(e.combinePaths(u,o));C(n,d,l,a,void 0,c),s&&A(c,n,d,l.extensions,s,a)}for(var p=L(n),m=0,f=M(n,p,i);m<f.length;m++){var g=f[m];c.push(_(g,"external module name",void 0))}if(P(a,r,t,p,l,c),e.getEmitModuleResolutionKind(r)===e.ModuleResolutionKind.NodeJs){var v=!1;if(void 0===p)for(var y=function(e){c.some(function(n){return n.name===e})||(v=!0,c.push(_(e,"external module name",void 0)))},h=0,E=G(a,t);h<E.length;h++){var T=E[h];y(T)}v||e.forEachAncestorDirectory(t,function(t){var r=e.combinePaths(t,"node_modules");q(a,r)&&C(n,r,l,a,void 0,c)})}return c}function L(n){return X(n)?e.hasTrailingDirectorySeparator(n)?n:e.getDirectoryPath(n):void 0}function k(n,t,r,a,i,o){if(!e.endsWith(n,"*"))return e.stringContains(n,"*")?e.emptyArray:l(n);var s=n.slice(0,n.length-1),c=e.tryRemovePrefix(r,s);return void 0===c?l(s):e.flatMap(t,function(e){return N(c,a,e,i,o)});function l(n){return e.startsWith(n,r)?[g(n)]:e.emptyArray}}function N(n,t,r,a,i){if(i.readDirectory){var o=e.hasZeroOrOneAsteriskCharacter(r)?e.tryParsePattern(r):void 0;if(o){var s=e.resolvePath(o.prefix),c=e.hasTrailingDirectorySeparator(o.prefix)?s:e.getDirectoryPath(s),l=e.hasTrailingDirectorySeparator(o.prefix)?"":e.getBaseFileName(s),u=X(n),d=u?e.hasTrailingDirectorySeparator(n)?n:e.getDirectoryPath(n):void 0,p=u?e.combinePaths(c,l+d):c,m=e.normalizePath(o.suffix),f=e.normalizePath(e.combinePaths(t,p)),v=u?f:e.ensureTrailingDirectorySeparator(f)+l,y=m?"**/*":"./*",h=e.mapDefined(H(i,f,a,void 0,[y]),function(n){var t=e.tryGetExtensionFromPath(n),r=E(n);return void 0===r?void 0:_(e.removeFileExtension(r),"script",t)}),b=e.mapDefined(U(i,f).map(function(n){return e.combinePaths(f,n)}),function(e){var n=E(e);return void 0===n?void 0:g(n)});return h.concat(b)}}function E(n){var t=I(e.normalizePath(n),v,m);return void 0===t?void 0:w(t)}}function I(n,t,r){return e.startsWith(n,t)&&e.endsWith(n,r)?n.slice(t.length,n.length-r.length):void 0}function w(n){return n[0]===e.directorySeparator?n.slice(1):n}function M(n,t,r){var a=r.getAmbientModules().map(function(n){return e.stripQuotes(n.name)}),i=a.filter(function(t){return e.startsWith(t,n)});if(void 0!==t){var o=e.ensureTrailingDirectorySeparator(t);return i.map(function(n){return e.removePrefix(n,o)})}return i}function O(n,t,r,a){var i=e.getTokenAtPosition(n,t),o=e.getLeadingCommentRanges(n.text,i.pos),s=o&&e.find(o,function(e){return t>=e.pos&&t<=e.end});if(s){var c=n.text.slice(s.pos,t),l=K.exec(c);if(l){var u=l[1],d=l[2],p=l[3],m=e.getDirectoryPath(n.path),f="path"===d?C(p,m,b(r,!0),a,n.path):"types"===d?P(a,r,m,L(p),b(r)):e.Debug.fail();return v(p,s.pos+u.length,f)}}}function P(n,t,r,a,i,o){void 0===o&&(o=[]);for(var s=e.createMap(),c=J(function(){return e.getEffectiveTypeRoots(t,n)})||e.emptyArray,l=0,u=c;l<u.length;l++){var d=u[l];v(d)}for(var p=0,m=R(r,n);p<m.length;p++){var f=m[p],g=e.combinePaths(e.getDirectoryPath(f),"node_modules/@types");v(g)}return o;function v(r){if(q(n,r))for(var c=0,l=U(n,r);c<l.length;c++){var u=l[c],d=e.unmangleScopedPackageName(u);if(!t.types||e.contains(t.types,d))if(void 0===a)s.has(d)||(o.push(_(d,"external module name",void 0)),s.set(d,!0));else{var p=e.combinePaths(r,u),m=e.tryRemoveDirectoryPrefix(a,d,e.hostGetCanonicalFileName(n));void 0!==m&&C(m,p,i,n,void 0,o)}}}}function R(n,t){var r=[];return e.forEachAncestorDirectory(n,function(n){var a=e.findConfigFile(n,function(e){return W(t,e)},"package.json");if(!a)return!0;r.push(a)}),r}function F(n,t){var r;return e.forEachAncestorDirectory(n,function(n){return"node_modules"===n||(r=e.findConfigFile(n,function(e){return W(t,e)},"package.json"),!!r||void 0)}),r}function G(n,t){if(!n.readFile||!n.fileExists)return e.emptyArray;for(var r=[],a=0,i=R(t,n);a<i.length;a++)for(var o=i[a],s=e.readJson(o,n),c=0,l=j;c<l.length;c++){var u=l[c],d=s[u];if(d)for(var p in d)d.hasOwnProperty(p)&&!e.startsWith(p,"@types/")&&r.push(p)}return r}function V(n,t){var r=Math.max(n.lastIndexOf(e.directorySeparator),n.lastIndexOf("\\")),a=-1!==r?r+1:0,i=n.length-a;return 0===i||e.isIdentifierText(n.substr(a,i),6)?void 0:e.createTextSpan(t+a,i)}function B(e){if(e&&e.length>=2&&46===e.charCodeAt(0)){var n=e.length>=3&&46===e.charCodeAt(1)?2:1,t=e.charCodeAt(n);return 47===t||92===t}return!1}t.getStringLiteralCompletions=r,t.getStringLiteralCompletionDetails=i,function(e){e[e["Paths"]=0]="Paths",e[e["Properties"]=1]="Properties",e[e["Types"]=2]="Types"}(l||(l={}));var K=/^(\/\/\/\s*<reference\s+(path|types)\s*=\s*(?:'|"))([^\3"]*)$/,j=["dependencies","devDependencies","peerDependencies","optionalDependencies"];function U(e,n){return z(e,e.getDirectories,n)||[]}function H(n,t,r,a,i){return z(n,n.readDirectory,t,r,a,i)||e.emptyArray}function W(e,n){return z(e,e.fileExists,n)}function q(n,t){return J(function(){return e.directoryProbablyExists(t,n)})||!1}function z(e,n){for(var t=[],r=2;r<arguments.length;r++)t[r-2]=arguments[r];return J(function(){return n&&n.apply(e,t)})}function J(e){try{return e()}catch(n){return}}function X(n){return e.stringContains(n,e.directorySeparator)}})(n.StringCompletions||(n.StringCompletions={}))})(e.Completions||(e.Completions={}))}(u||(u={})),function(e){(function(n){var t,r,a;function i(e){return 2===e.kind||1===e.kind}function o(e){return 2===e.kind||3===e.kind}function s(t,r,a,i,o,s,d){var p=r.getTypeChecker(),m=r.getCompilerOptions(),f=e.findPrecedingToken(o,i);if(!d||X(i,d,f,o)){var _=n.StringCompletions.getStringLiteralCompletions(i,o,f,p,m,t,a,s);if(_)return _;if(f&&e.isBreakOrContinueStatement(f.parent)&&(73===f.kind||78===f.kind||72===f.kind))return E(f.parent);var g=P(r,a,i,u(i,m),o,s,void 0);if(g)switch(g.kind){case 0:return l(i,p,m,a,g,s);case 1:return c(e.JsDoc.getJSDocTagNameCompletions());case 2:return c(e.JsDoc.getJSDocTagCompletions());case 3:return c(e.JsDoc.getJSDocParameterNameCompletions(g.tag));default:return e.Debug.assertNever(g)}}}function c(e){return{isGlobalCompletion:!1,isMemberCompletion:!1,isNewIdentifierLocation:!1,entries:e}}function l(n,t,r,a,i,o){var s=i.symbols,c=i.completionKind,l=i.isInSnippetScope,m=i.isNewIdentifierLocation,f=i.location,g=i.propertyAccessToConvert,v=i.keywordFilters,y=i.literals,h=i.symbolToOriginInfoMap,E=i.recommendedCompletion,T=i.isJsxInitializer;if(f&&f.parent&&e.isJsxClosingElement(f.parent)){var S=f.parent.parent.openingElement.tagName,x=!!e.findChildOfKind(f.parent,30,n),C={name:S.getFullText(n)+(x?"":">"),kind:"class",kindModifiers:void 0,sortText:"0"};return{isGlobalCompletion:!1,isMemberCompletion:!0,isNewIdentifierLocation:!1,entries:[C]}}var A=[];if(u(n,r)){var D=b(s,A,f,n,t,r.target,a,c,o,g,T,E,h);p(n,f.pos,D,r.target,A)}else{if(!m&&(!s||0===s.length)&&0===v)return;b(s,A,f,n,t,r.target,a,c,o,g,T,E,h)}if(0!==v)for(var L=e.arrayToSet(A,function(e){return e.name}),k=0,N=V(v);k<N.length;k++){var I=N[k];L.has(I.name)||A.push(I)}for(var w=0,M=y;w<M.length;w++){var O=M[w];A.push(_(O))}return{isGlobalCompletion:l,isMemberCompletion:d(c),isNewIdentifierLocation:m,entries:A}}function u(n,t){return e.isSourceFileJS(n)&&!e.isCheckJsEnabledForFile(n,t)}function d(e){switch(e){case 0:case 3:case 2:return!0;default:return!1}}function p(n,t,r,a,i){e.getNameTable(n).forEach(function(n,o){if(n!==t){var s=e.unescapeLeadingUnderscores(o);e.addToSeen(r,s)&&e.isIdentifierText(s,a)&&i.push({name:s,kind:"warning",kindModifiers:"",sortText:"1"})}})}(function(e){e[e["ThisType"]=0]="ThisType",e[e["SymbolMemberNoExport"]=1]="SymbolMemberNoExport",e[e["SymbolMemberExport"]=2]="SymbolMemberExport",e[e["Export"]=3]="Export"})(t||(t={})),function(e){e[e["None"]=0]="None",e[e["All"]=1]="All",e[e["ClassElementKeywords"]=2]="ClassElementKeywords",e[e["InterfaceElementKeywords"]=3]="InterfaceElementKeywords",e[e["ConstructorParameterKeywords"]=4]="ConstructorParameterKeywords",e[e["FunctionLikeBodyKeywords"]=5]="FunctionLikeBodyKeywords",e[e["TypeKeywords"]=6]="TypeKeywords"}(r||(r={})),function(e){e[e["Continue"]=0]="Continue",e[e["Success"]=1]="Success",e[e["Fail"]=2]="Fail"}(a||(a={})),n.getCompletionsAtPosition=s;var m,f=function(n){return"object"===typeof n?e.pseudoBigIntToString(n)+"n":JSON.stringify(n)};function _(e){return{name:f(e),kind:"string",kindModifiers:"",sortText:"0"}}function g(n,t,r,a,s,c,l,u,d,p,m){var f=R(n,s,l,c);if(f){var _,g,b=f.name,E=f.needsConvertPropertyAccess;if(l&&0===l.kind)_=E?"this["+e.quote(b,m)+"]":"this."+b;else if((l&&i(l)||E)&&d){_=E?"["+e.quote(b,m)+"]":"["+b+"]";var T=e.findChildOfKind(d,24,r),S=e.startsWith(b,d.name.text)?d.name.end:T.end;g=e.createTextSpanFromBounds(T.getStart(r),S)}if(p&&(void 0===_&&(_=b),_="{"+_+"}","boolean"!==typeof p&&(g=e.createTextSpanFromNode(p,r))),void 0===_||m.includeCompletionsWithInsertText)return{name:b,kind:e.SymbolDisplay.getSymbolKind(a,n,t),kindModifiers:e.SymbolDisplay.getSymbolModifiers(n),sortText:"0",source:h(l),hasAction:y(!!l&&o(l)),isRecommended:y(v(n,u,a)),insertText:_,replacementSpan:g}}}function v(e,n,t){return e===n||!!(1048576&e.flags)&&t.getExportSymbolOfSymbol(e)===n}function y(e){return!!e||void 0}function h(n){return n&&o(n)?e.stripQuotes(n.moduleSymbol.name):void 0}function b(n,t,r,a,i,o,s,c,l,u,d,p,m){for(var f=e.timestamp(),_=e.createMap(),v=0,y=n;v<y.length;v++){var h=y[v],b=m?m[e.getSymbolId(h)]:void 0,E=g(h,r,a,i,o,c,b,p,u,d,l);if(E){var T=E.name;_.has(T)||(b||void 0===h.parent&&!e.some(h.declarations,function(e){return e.getSourceFile()===r.getSourceFile()})||_.set(T,!0),t.push(E))}}return s("getCompletionsAtPosition: getCompletionEntriesFromSymbols: "+(e.timestamp()-f)),_}function E(e){var n=T(e);if(n.length)return{isGlobalCompletion:!1,isMemberCompletion:!1,isNewIdentifierLocation:!1,entries:n}}function T(n){var t=[],r=e.createMap(),a=n;while(a){if(e.isFunctionLike(a))break;if(e.isLabeledStatement(a)){var i=a.label.text;r.has(i)||(r.set(i,!0),t.push({name:i,kindModifiers:"",kind:"label",sortText:"0"}))}a=a.parent}return t}function S(n,t,r,a,i){var o=n.getCompilerOptions(),s=P(n,t,r,u(r,o),a,{includeCompletionsForModuleExports:!0,includeCompletionsWithInsertText:!0},i);if(!s)return{type:"none"};if(0!==s.kind)return{type:"request",request:s};var c=s.symbols,l=s.literals,d=s.location,p=s.completionKind,m=s.symbolToOriginInfoMap,_=s.previousToken,g=s.isJsxInitializer,v=e.find(l,function(e){return f(e)===i.name});return void 0!==v?{type:"literal",literal:v}:e.firstDefined(c,function(n){var t=m[e.getSymbolId(n)],r=R(n,o.target,t,p);return r&&r.name===i.name&&h(t)===i.source?{type:"symbol",symbol:n,location:d,symbolToOriginInfoMap:m,previousToken:_,isJsxInitializer:g}:void 0})||{type:"none"}}function x(n,t,r){return t&&o(t)&&(t.isDefaultExport&&"default"===n.escapedName||"export="===n.escapedName)?e.firstDefined(n.declarations,function(n){return e.isExportAssignment(n)&&e.isIdentifier(n.expression)?n.expression.text:void 0})||e.codefix.moduleSymbolToValidIdentifier(t.moduleSymbol,r):n.name}function C(t,r,a,i,o,s,c,l,u){var d=t.getTypeChecker(),p=t.getCompilerOptions(),m=o.name,_=e.findPrecedingToken(i,a);if(e.isInString(a,i,_))return n.StringCompletions.getStringLiteralCompletionDetails(m,a,i,_,d,p,s,u);var g=S(t,r,a,i,o);switch(g.type){case"request":var v=g.request;switch(v.kind){case 1:return e.JsDoc.getJSDocTagNameCompletionDetails(m);case 2:return e.JsDoc.getJSDocTagCompletionDetails(m);case 3:return e.JsDoc.getJSDocParameterNameCompletionDetails(m);default:return e.Debug.assertNever(v)}case"symbol":var y=g.symbol,h=g.location,b=g.symbolToOriginInfoMap,E=g.previousToken,T=k(b,y,t,d,s,p,a,i,E,c,l),x=T.codeActions,C=T.sourceDisplay;return D(y,d,a,h,u,x,C);case"literal":var L=g.literal;return A(f(L),"string","string"===typeof L?e.SymbolDisplayPartKind.stringLiteral:e.SymbolDisplayPartKind.numericLiteral);case"none":return G().some(function(e){return e.name===m})?A(m,"keyword",e.SymbolDisplayPartKind.keyword):void 0;default:e.Debug.assertNever(g)}}function A(n,t,r){return L(n,"",t,[e.displayPart(n,r)])}function D(n,t,r,a,i,o,s){var c=t.runWithCancellationToken(i,function(t){return e.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(t,n,r,a,a,7)}),l=c.displayParts,u=c.documentation,d=c.symbolKind,p=c.tags;return L(n.name,e.SymbolDisplay.getSymbolModifiers(n),d,l,u,p,o,s)}function L(e,n,t,r,a,i,o,s){return{name:e,kindModifiers:n,kind:t,displayParts:r,documentation:a,tags:i,codeActions:o,source:s}}function k(n,t,r,a,i,s,c,l,u,d,p){var m=n[e.getSymbolId(t)];if(!m||!o(m))return{codeActions:void 0,sourceDisplay:void 0};var f=m.moduleSymbol,_=a.getMergedSymbol(e.skipAlias(t.exportSymbol||t,a)),g=e.codefix.getImportCompletionAction(_,f,c,x(t,m,s.target),i,r,d,u&&e.isIdentifier(u)?u.getStart(c):l,p),v=g.moduleSpecifier,y=g.codeAction;return{sourceDisplay:[e.textPart(v)],codeActions:[y]}}function N(e,n,t,r,a){var i=S(e,n,t,r,a);return"symbol"===i.type?i.symbol:void 0}function I(n,t,r){return e.firstDefined(t&&(t.isUnion()?t.types:[t]),function(t){var a=t&&t.symbol;return a&&424&a.flags&&!e.isAbstractConstructorSymbol(a)?M(a,n,r):void 0})}function w(n,t,r,a){var i=n.parent;switch(n.kind){case 72:return e.getContextualTypeFromParent(n,a);case 59:switch(i.kind){case 237:return a.getContextualType(i.initializer);case 204:return a.getTypeAtLocation(i.left);case 267:return a.getContextualTypeForJsxAttribute(i);default:return}case 95:return a.getContextualType(i);case 74:return e.getSwitchedType(e.cast(i,e.isCaseClause),a);case 18:return e.isJsxExpression(i)&&260!==i.parent.kind?a.getContextualTypeForJsxAttribute(i.parent):void 0;default:var o=e.SignatureHelp.getArgumentInfoForCompletions(n,t,r);return o?a.getContextualTypeForArgumentAtIndex(o.invocation,o.argumentIndex+(27===n.kind?1:0)):e.isEqualityOperatorKind(n.kind)&&e.isBinaryExpression(i)&&e.isEqualityOperatorKind(i.operatorToken.kind)?a.getTypeAtLocation(i.left):a.getContextualType(n)}}function M(n,t,r){var a=r.getAccessibleSymbolChain(n,t,67108863,!1);return a?e.first(a):n.parent&&(O(n.parent)?n:M(n.parent,t,r))}function O(e){return e.declarations.some(function(e){return 279===e.kind})}function P(n,t,r,a,i,o,s){var c=n.getTypeChecker(),l=e.timestamp(),u=e.getTokenAtPosition(r,i);t("getCompletionData: Get current token: "+(e.timestamp()-l)),l=e.timestamp();var d=e.isInComment(r,i,u);t("getCompletionData: Is inside comment: "+(e.timestamp()-l));var p=!1,m=!1;if(d){if(e.hasDocComment(r,i)){if(64===r.text.charCodeAt(i-1))return{kind:1};var f=e.getLineStartPositionForPosition(i,r);if(!r.text.substring(f,i).match(/[^\*|\s|(/\*\*)]/))return{kind:2}}var _=H(u,i);if(_){if(_.tagName.pos<=i&&i<=_.tagName.end)return{kind:1};if(Q(_)&&_.typeExpression&&283===_.typeExpression.kind&&(u=e.getTokenAtPosition(r,i),u&&(e.isDeclarationName(u)||305===u.parent.kind&&u.parent.name===u)||(p=Oe(_.typeExpression))),e.isJSDocParameterTag(_)&&(e.nodeIsMissing(_.name)||_.name.pos<=i&&i<=_.name.end))return{kind:3,tag:_}}if(!p)return void t("Returning an empty list because completion was inside a regular comment or plain text part of a JsDoc comment.")}l=e.timestamp();var g=e.findPrecedingToken(i,r,void 0);t("getCompletionData: Get previous token 1: "+(e.timestamp()-l));var v=g;if(v&&i<=v.end&&(e.isIdentifier(v)||e.isKeyword(v.kind))){var y=e.timestamp();v=e.findPrecedingToken(v.getFullStart(),r,void 0),t("getCompletionData: Get previous token 2: "+(e.timestamp()-y))}var h,b=u,E=!1,T=!1,S=!1,C=!1,A=e.getTouchingPropertyName(r,i);if(v){if(ge(v))return void t("Returning an empty list because completion was requested in an invalid position.");var D=v.parent;if(24===v.kind)switch(E=!0,D.kind){case 189:h=D,b=h.expression;break;case 148:b=D.left;break;case 244:b=D.name;break;case 183:case 214:b=D;break;default:return}else if(1===r.languageVariant){if(D&&189===D.kind&&(v=D,D=D.parent),u.parent===A)switch(u.kind){case 30:260!==u.parent.kind&&262!==u.parent.kind||(A=u);break;case 42:261===u.parent.kind&&(A=u);break}switch(D.kind){case 263:42===v.kind&&(S=!0,A=v);break;case 204:if(!Y(D))break;case 261:case 260:case 262:28===v.kind&&(T=!0,A=v);break;case 267:switch(g.kind){case 59:C=!0;break;case 72:D!==g.parent&&!D.initializer&&e.findChildOfKind(D,59,r)&&(C=g)}break}}}var L=e.timestamp(),k=5,N=!1,O=0,P=[],R=[];if(E)Z();else if(T){var F=e.Debug.assertEachDefined(c.getJsxIntrinsicTagNamesAt(A),"getJsxIntrinsicTagNames() should all be defined");te(),P=F.concat(P),k=3,O=0}else if(S){var G=v.parent.parent.openingElement.tagName,V=c.getSymbolAtLocation(G);V&&(P=[V]),k=3,O=0}else if(!te())return;t("getCompletionData: Semantic work: "+(e.timestamp()-L));var B=g&&w(g,i,r,c),j=e.mapDefined(B&&(B.isUnion()?B.types:[B]),function(e){return e.isLiteral()?e.value:void 0}),X=g&&B&&I(g,B,c);return{kind:0,symbols:P,completionKind:k,isInSnippetScope:m,propertyAccessToConvert:h,isNewIdentifierLocation:N,location:A,keywordFilters:O,literals:j,symbolToOriginInfoMap:R,recommendedCompletion:X,previousToken:g,isJsxInitializer:C};function Q(e){switch(e.kind){case 299:case 305:case 300:case 302:case 304:return!0;default:return!1}}function Z(){k=2;var n=e.isLiteralImportTypeNode(b),t=p||n&&!b.isTypeOf||e.isPartOfTypeNode(b.parent),a=e.isInRightSideOfInternalImportEqualsDeclaration(b),i=a||!t&&e.isPossiblyTypeArgumentPosition(v,r,c);if(e.isEntityName(b)||n){var o=e.isModuleDeclaration(b.parent);o&&(N=!0);var s=c.getSymbolAtLocation(b);if(s&&(s=e.skipAlias(s,c),1920&s.flags)){for(var l=e.Debug.assertEachDefined(c.getExportsOfModule(s),"getExportsOfModule() should all be defined"),u=function(e){return c.isValidPropertyAccess(n?b:b.parent,e.name)},d=function(e){return pe(e)},m=o?function(e){return!!(1920&e.flags)&&!e.declarations.every(function(e){return e.parent===b.parent})}:i?function(e){return d(e)||u(e)}:t?d:u,f=0,_=l;f<_.length;f++){var g=_[f];m(g)&&P.push(g)}return void(!t&&s.declarations.some(function(e){return 279!==e.kind&&244!==e.kind&&243!==e.kind})&&$(c.getTypeOfSymbolAtLocation(s,b)))}}if(!e.isMetaProperty(b)||95!==b.keywordToken&&92!==b.keywordToken)t||$(c.getTypeAtLocation(b));else{var y=95===b.keywordToken?"target":"meta";P.push(c.createSymbol(4,e.escapeLeadingUnderscores(y)))}}function $(e){if(N=!!e.getStringIndexType(),a)P.push.apply(P,q(e,c));else for(var n=0,t=e.getApparentProperties();n<t.length;n++){var r=t[n];c.isValidPropertyAccessForCompletions(183===b.kind?b:b.parent,e,r)&&ee(r)}}function ee(n){var t=e.firstDefined(n.declarations,function(n){return e.tryCast(e.getNameOfDeclaration(n),e.isComputedPropertyName)});if(t){var r=ne(t.expression),a=r&&c.getSymbolAtLocation(r),i=a&&M(a,v,c);if(i&&!R[e.getSymbolId(i)]){P.push(i);var o=i.parent;R[e.getSymbolId(i)]=o&&e.isExternalModuleSymbol(o)?{kind:2,moduleSymbol:o,isDefaultExport:!1}:{kind:1}}}else P.push(n)}function ne(n){return e.isIdentifier(n)?n:e.isPropertyAccessExpression(n)?ne(n.expression):void 0}function te(){var e=be()||Ee()||re()||Te()||ae()||(ie(),1);return 1===e}function re(){return Ce(v)?(k=5,N=!0,O=4,1):0}function ae(){var e=De(v),n=e&&c.getContextualType(e.attributes);return n?(P=Me(W(n,e.attributes,c),e.attributes.properties),k=3,N=!1,1):0}function ie(){O=Ae(v)?5:1,k=1,N=ye(v),g!==v&&e.Debug.assert(!!g,"Expected 'contextToken' to be defined when different from 'previousToken'.");var t=g!==v?g.getStart():i,a=_e(v,t,r)||r;m=se(a);var s=le(),l=69996520|(s?0:67220415);if(P=e.Debug.assertEachDefined(c.getSymbolsInScope(a,l),"getSymbolsInScope() should all be defined"),o.includeCompletionsWithInsertText&&279!==a.kind){var u=c.tryGetThisTypeAt(a);if(u)for(var d=0,p=q(u,c);d<p.length;d++){var f=p[d];R[e.getSymbolId(f)]={kind:0},P.push(f)}}oe()&&me(P,g&&e.isIdentifier(g)?g.text:"",n.getCompilerOptions().target),ce(P)}function oe(){return!!o.includeCompletionsForModuleExports&&(!!r.externalModuleIndicator||!r.commonJsModuleIndicator&&(!!e.compilerOptionsIndicateEs6Modules(n.getCompilerOptions())||e.programContainsEs6Modules(n)))}function se(n){switch(n.kind){case 279:case 206:case 270:case 218:return!0;default:return e.isStatement(n)}}function ce(n){var t=le(),a=t||!ue(v)&&e.isPossiblyTypeArgumentPosition(v,r,c);t&&(O=6),e.filterMutate(n,function(n){if(!e.isSourceFile(A)){if(e.isExportAssignment(A.parent))return!0;if(n=e.skipAlias(n,c),e.isInRightSideOfInternalImportEqualsDeclaration(A))return!!(1920&n.flags);if(a){var r=pe(n);if(r||t)return r}}return!!(67220415&e.getCombinedLocalAndExportSymbolFlags(n))})}function le(){return p||!ue(v)&&(e.isPartOfTypeNode(A)||de(v))}function ue(n){return n&&104===n.kind&&(167===n.parent.kind||e.isTypeOfExpression(n.parent))}function de(n){if(n){var t=n.parent.kind;switch(n.kind){case 57:return 154===t||153===t||151===t||237===t||e.isFunctionLikeKind(t);case 59:return 242===t;case 119:return 212===t}}return!1}function pe(n,t){void 0===t&&(t=e.createMap());var r=e.skipAlias(n.exportSymbol||n,c);return!!(67897832&r.flags)||!!(1536&r.flags)&&e.addToSeen(t,e.getSymbolId(r))&&c.getExportsOfModule(r).some(function(e){return pe(e,t)})}function me(t,a,i){var o=a.toLowerCase(),l=e.createMap();e.codefix.forEachExternalModuleToImportFrom(c,r,n.getSourceFiles(),function(n){if(!s||!s.source||e.stripQuotes(n.name)===s.source){var r=c.resolveExternalModuleSymbol(n);if(e.addToSeen(l,e.getSymbolId(r))){r!==n&&e.some(r.declarations,function(e){return!!e.getSourceFile().externalModuleIndicator})&&(t.push(r),R[e.getSymbolId(r)]={kind:3,moduleSymbol:n,isDefaultExport:!1});for(var a=0,u=c.getExportsOfModule(n);a<u.length;a++){var d=u[a];if(c.getMergedSymbol(d.parent)===r&&!e.some(d.declarations,function(n){return e.isExportSpecifier(n)&&(n.propertyName?e.isIdentifierANonContextualKeyword(n.name):!!n.parent.parent.moduleSpecifier)})){var p="default"===d.escapedName;p&&(d=e.getLocalSymbolForExportDefault(d)||d);var m={kind:3,moduleSymbol:n,isDefaultExport:p};(s||fe(x(d,m,i).toLowerCase(),o))&&(t.push(d),R[e.getSymbolId(d)]=m)}}}}})}function fe(e,n){if(0===n.length)return!0;for(var t=0,r=0;r<e.length;r++)if(e.charCodeAt(r)===n.charCodeAt(t)&&(t++,t===n.length))return!0;return!1}function _e(n,t,r){var a=n;while(a&&!e.positionBelongsToNode(a,t,r))a=a.parent;return a}function ge(n){var r=e.timestamp(),a=he(n)||Le(n)||Ne(n)||ve(n);return t("getCompletionsAtPosition: isCompletionListBlocker: "+(e.timestamp()-r)),a}function ve(e){if(11===e.kind)return!0;if(30===e.kind&&e.parent){if(262===e.parent.kind)return!0;if(263===e.parent.kind||261===e.parent.kind)return!!e.parent.parent&&260===e.parent.parent.kind}return!1}function ye(e){if(e){var n=e.parent.kind;switch(U(e)){case 27:return 191===n||157===n||192===n||187===n||204===n||165===n;case 20:return 191===n||157===n||192===n||195===n||177===n;case 22:return 187===n||162===n||149===n;case 130:case 131:return!0;case 24:return 244===n;case 18:return 240===n;case 59:return 237===n||204===n;case 15:return 206===n;case 16:return 216===n;case 115:case 113:case 114:return 154===n}}return!1}function he(n){return(e.isRegularExpressionLiteral(n)||e.isStringTextContainingNode(n))&&(e.rangeContainsPositionExclusive(e.createTextRangeFromSpan(e.createTextSpanFromNode(n)),i)||i===n.end&&(!!n.isUnterminated||e.isRegularExpressionLiteral(n)))}function be(){var n,t,r=Se(v);if(!r)return 0;if(k=0,188===r.kind){var a=c.getContextualType(r);if(!a)return 2;N=e.hasIndexSignature(a),n=W(a,r,c),t=r.properties}else{e.Debug.assert(184===r.kind),N=!1;var i=e.getRootDeclaration(r.parent);if(!e.isVariableLike(i))return e.Debug.fail("Root declaration is not variable-like.");var o=e.hasInitializer(i)||e.hasType(i)||227===i.parent.parent.kind;if(o||151!==i.kind||(e.isExpression(i.parent)?o=!!c.getContextualType(i.parent):156!==i.parent.kind&&159!==i.parent.kind||(o=e.isExpression(i.parent.parent)&&!!c.getContextualType(i.parent.parent))),o){a=c.getTypeAtLocation(r);if(!a)return 2;n=c.getPropertiesOfType(a).filter(function(n){return!(24&e.getDeclarationModifierFlagsFromSymbol(n))}),t=r.elements}}return n&&n.length>0&&(P=Ie(n,e.Debug.assertDefined(t))),1}function Ee(){var n=!v||18!==v.kind&&27!==v.kind?void 0:e.tryCast(v.parent,e.isNamedImportsOrExports);if(!n)return 0;var t=(252===n.kind?n.parent.parent:n.parent).moduleSpecifier,r=c.getSymbolAtLocation(t);if(!r)return 2;k=3,N=!1;var a=c.getExportsAndPropertiesOfModule(r),i=e.arrayToSet(n.elements,function(e){return Oe(e)?void 0:(e.propertyName||e.name).escapedText});return P=a.filter(function(e){return"default"!==e.escapedName&&!i.get(e.escapedName)}),1}function Te(){var n=z(r,v,A);if(!n)return 0;if(k=3,N=!0,O=40===v.kind?0:e.isClassLike(n)?2:3,!e.isClassLike(n))return 1;var t=v.parent,a=e.isClassElement(t)?e.getModifierFlags(t):0;if(72===v.kind&&!Oe(v))switch(v.getText()){case"private":a|=8;break;case"static":a|=32;break}if(!(8&a)){var i=e.flatMap(e.getAllSuperTypeNodes(n),function(e){var t=c.getTypeAtLocation(e);return t&&c.getPropertiesOfType(32&a?c.getTypeOfSymbolAtLocation(t.symbol,n):t)});P=we(i,n.members,a)}return 1}function Se(n){if(n){var t=n.parent;switch(n.kind){case 18:case 27:if(e.isObjectLiteralExpression(t)||e.isObjectBindingPattern(t))return t;break;case 40:return e.isMethodDeclaration(t)?e.tryCast(t.parent,e.isObjectLiteralExpression):void 0;case 72:return"async"===n.text&&e.isShorthandPropertyAssignment(n.parent)?n.parent.parent:void 0}}}function xe(n){return!!n.parent&&e.isParameter(n.parent)&&e.isConstructorDeclaration(n.parent.parent)&&(e.isParameterPropertyModifier(n.kind)||e.isDeclarationName(n))}function Ce(n){if(n){var t=n.parent;switch(n.kind){case 20:case 27:return e.isConstructorDeclaration(n.parent)?n.parent:void 0;default:if(xe(n))return t.parent}}}function Ae(n){if(n){var t,r=e.findAncestor(n.parent,function(n){return e.isClassLike(n)?"quit":!(!e.isFunctionLikeDeclaration(n)||t!==n.body)||(t=n,!1)});return r&&r}}function De(n){if(n){var t=n.parent;switch(n.kind){case 30:case 29:case 42:case 72:case 189:case 268:case 267:case 269:if(t&&(261===t.kind||262===t.kind)){if(30===n.kind){var a=e.findPrecedingToken(n.pos,r,void 0);if(!t.typeArguments||a&&42===a.kind)break}return t}if(267===t.kind)return t.parent.parent;break;case 10:if(t&&(267===t.kind||269===t.kind))return t.parent.parent;break;case 19:if(t&&270===t.kind&&t.parent&&267===t.parent.kind)return t.parent.parent.parent;if(t&&269===t.kind)return t.parent.parent;break}}}function Le(n){var t=n.parent,r=t.kind;switch(n.kind){case 27:return 237===r||238===r||219===r||243===r||ke(r)||241===r||185===r||242===r||e.isClassLike(t)&&!!t.typeParameters&&t.typeParameters.end>=n.pos;case 24:return 185===r;case 57:return 186===r;case 22:return 185===r;case 20:return 274===r||ke(r);case 18:return 243===r;case 28:return 240===r||209===r||241===r||242===r||e.isFunctionLikeKind(r);case 116:return 154===r&&!e.isClassLike(t.parent);case 25:return 151===r||!!t.parent&&185===t.parent.kind;case 115:case 113:case 114:return 151===r&&!e.isConstructorDeclaration(t.parent);case 119:return 253===r||257===r||251===r;case 126:case 137:return!J(n);case 76:case 84:case 110:case 90:case 105:case 92:case 111:case 77:case 117:case 140:return!0;case 40:return e.isFunctionLike(n.parent)&&!e.isMethodDeclaration(n.parent)}if(K(U(n))&&J(n))return!1;if(xe(n)&&(!e.isIdentifier(n)||e.isParameterPropertyModifier(U(n))||Oe(n)))return!1;switch(U(n)){case 118:case 76:case 77:case 125:case 84:case 90:case 110:case 111:case 113:case 114:case 115:case 116:case 105:case 117:return!0;case 121:return e.isPropertyDeclaration(n.parent)}return e.isDeclarationName(n)&&!e.isJsxAttribute(n.parent)&&!(e.isClassLike(n.parent)&&(n!==g||i>g.end))}function ke(n){return e.isFunctionLikeKind(n)&&157!==n}function Ne(e){if(8===e.kind){var n=e.getFullText();return"."===n.charAt(n.length-1)}return!1}function Ie(n,t){if(0===t.length)return n;for(var r=e.createUnderscoreEscapedMap(),a=0,i=t;a<i.length;a++){var o=i[a];if((275===o.kind||276===o.kind||186===o.kind||156===o.kind||158===o.kind||159===o.kind)&&!Oe(o)){var s=void 0;if(e.isBindingElement(o)&&o.propertyName)72===o.propertyName.kind&&(s=o.propertyName.escapedText);else{var c=e.getNameOfDeclaration(o);s=c&&e.isPropertyNameLiteral(c)?e.getEscapedTextOfIdentifierOrLiteral(c):void 0}r.set(s,!0)}}return n.filter(function(e){return!r.get(e.escapedName)})}function we(n,t,r){for(var a=e.createUnderscoreEscapedMap(),i=0,o=t;i<o.length;i++){var s=o[i];if((154===s.kind||156===s.kind||158===s.kind||159===s.kind)&&(!Oe(s)&&!e.hasModifier(s,8)&&e.hasModifier(s,32)===!!(32&r))){var c=e.getPropertyNameForPropertyNameNode(s.name);c&&a.set(c,!0)}}return n.filter(function(n){return!a.has(n.escapedName)&&!!n.declarations&&!(8&e.getDeclarationModifierFlagsFromSymbol(n))})}function Me(n,t){for(var r=e.createUnderscoreEscapedMap(),a=0,i=t;a<i.length;a++){var o=i[a];Oe(o)||267===o.kind&&r.set(o.name.escapedText,!0)}return n.filter(function(e){return!r.get(e.escapedName)})}function Oe(e){return e.getStart(r)<=i&&i<=e.getEnd()}}function R(n,t,r,a){var i=x(n,r,t);if(!(void 0===i||1536&n.flags&&e.startsWithQuote(i)||e.isKnownSymbol(n))){var o={name:i,needsConvertPropertyAccess:!1};if(e.isIdentifierText(i,t))return o;switch(a){case 3:return;case 0:return{name:JSON.stringify(i),needsConvertPropertyAccess:!1};case 2:case 1:return 32===i.charCodeAt(0)?void 0:{name:i,needsConvertPropertyAccess:!0};case 5:case 4:return o;default:e.Debug.assertNever(a)}}}n.getCompletionEntriesFromSymbols=b,n.getCompletionEntryDetails=C,n.createCompletionDetailsForSymbol=D,n.createCompletionDetails=L,n.getCompletionEntrySymbol=N,function(e){e[e["Data"]=0]="Data",e[e["JsDocTagName"]=1]="JsDocTagName",e[e["JsDocTag"]=2]="JsDocTag",e[e["JsDocParameterName"]=3]="JsDocParameterName"}(m||(m={})),function(e){e[e["ObjectPropertyDeclaration"]=0]="ObjectPropertyDeclaration",e[e["Global"]=1]="Global",e[e["PropertyAccess"]=2]="PropertyAccess",e[e["MemberLike"]=3]="MemberLike",e[e["String"]=4]="String",e[e["None"]=5]="None"}(n.CompletionKind||(n.CompletionKind={}));var F=[],G=e.memoize(function(){for(var n=[],t=73;t<=147;t++)n.push({name:e.tokenToString(t),kind:"keyword",kindModifiers:"",sortText:"0"});return n});function V(n){return F[n]||(F[n]=G().filter(function(t){var r=e.stringToToken(t.name);switch(n){case 0:return!1;case 1:return 121===r||122;case 2:return K(r);case 3:return B(r);case 4:return e.isParameterPropertyModifier(r);case 5:return j(r);case 6:return e.isTypeKeyword(r);default:return e.Debug.assertNever(n)}}))}function B(e){return 133===e}function K(n){switch(n){case 118:case 124:case 126:case 137:case 121:return!0;default:return e.isClassMemberModifier(n)}}function j(n){return 121===n||122===n||!e.isContextualKeyword(n)&&!K(n)}function U(n){return e.isIdentifier(n)?n.originalKeywordKind||0:n.kind}function H(n,t){var r=e.findAncestor(n,e.isJSDoc);return r&&r.tags&&(e.rangeContainsPosition(r,t)?e.findLast(r.tags,function(e){return e.pos<t}):void 0)}function W(n,t,r){return n.isUnion()?r.getAllPossiblePropertiesOfTypes(n.types.filter(function(n){return!(131068&n.flags||r.isArrayLikeType(n)||e.typeHasCallOrConstructSignatures(n,r)||r.isTypeInvalidDueToUnionDiscriminant(n,t))})):n.getApparentProperties()}function q(n,t){return n.isUnion()?e.Debug.assertEachDefined(t.getAllPossiblePropertiesOfTypes(n.types),"getAllPossiblePropertiesOfTypes() should all be defined"):e.Debug.assertEachDefined(n.getApparentProperties(),"getApparentProperties() should all be defined")}function z(n,t,r){switch(r.kind){case 306:return e.tryCast(r.parent,e.isObjectTypeDeclaration);case 1:var a=e.tryCast(e.lastOrUndefined(e.cast(r.parent,e.isSourceFile).statements),e.isObjectTypeDeclaration);if(a&&!e.findChildOfKind(a,19,n))return a}if(t)switch(t.kind){case 26:case 19:return J(r)&&r.parent.name===r?r.parent.parent:e.tryCast(r,e.isObjectTypeDeclaration);case 18:case 27:return e.tryCast(t.parent,e.isObjectTypeDeclaration);default:if(!J(t))return;var i=e.isClassLike(t.parent.parent)?K:B;return i(t.kind)||40===t.kind||e.isIdentifier(t)&&i(e.stringToToken(t.text))?t.parent.parent:void 0}}function J(n){return n.parent&&e.isClassOrTypeElement(n.parent)&&e.isObjectTypeDeclaration(n.parent.parent)}function X(n,t,r,a){switch(t){case".":case"@":return!0;case'"':case"'":case"`":return!!r&&e.isStringLiteralOrTemplate(r)&&a===r.getStart(n)+1;case"<":return!!r&&28===r.kind&&(!e.isBinaryExpression(r.parent)||Y(r.parent));case"/":return!!r&&(e.isStringLiteralLike(r)?!!e.tryGetImportFromModuleSpecifier(r):42===r.kind&&e.isJsxClosingElement(r.parent));default:return e.Debug.assertNever(t)}}function Y(n){var t=n.left;return e.nodeIsMissing(t)}})(e.Completions||(e.Completions={}))}(u||(u={})),function(e){(function(n){function t(n,t,o,s,c){var l=e.getTouchingPropertyName(o,s);if(l.parent&&(e.isJsxOpeningElement(l.parent)&&l.parent.tagName===l||e.isJsxClosingElement(l.parent))){var u=l.parent.parent,d=u.openingElement,p=u.closingElement,m=[d,p].map(function(e){var n=e.tagName;return r(n,o)});return[{fileName:o.fileName,highlightSpans:m}]}return a(s,l,n,t,c)||i(l,o)}function r(n,t){return{fileName:t.fileName,textSpan:e.createTextSpanFromNode(n,t),kind:"none"}}function a(n,t,r,a,i){var o=e.arrayToSet(i,function(e){return e.fileName}),s=e.FindAllReferences.getReferenceEntriesForNode(n,t,r,i,a,void 0,o);if(s){var c=e.arrayToMultiMap(s.map(e.FindAllReferences.toHighlightSpan),function(e){return e.fileName},function(e){return e.span});return e.arrayFrom(c.entries(),function(n){var t=n[0],a=n[1];if(!o.has(t)){e.Debug.assert(r.redirectTargetsMap.has(t));var s=r.getSourceFile(t),c=e.find(i,function(e){return!!e.redirectInfo&&e.redirectInfo.redirectTarget===s});t=c.fileName,e.Debug.assert(o.has(t))}return{fileName:t,highlightSpans:a}})}}function i(e,n){var t=o(e,n);return t&&[{fileName:n.fileName,highlightSpans:t}]}function o(n,t){switch(n.kind){case 91:case 83:return e.isIfStatement(n.parent)?C(n.parent,t):void 0;case 97:return o(n.parent,e.isReturnStatement,E);case 101:return o(n.parent,e.isThrowStatement,b);case 103:case 75:case 88:var a=75===n.kind?n.parent.parent:n.parent;return o(a,e.isTryStatement,h);case 99:return o(n.parent,e.isSwitchStatement,y);case 74:case 80:return o(n.parent.parent.parent,e.isSwitchStatement,y);case 73:case 78:return o(n.parent,e.isBreakOrContinueStatement,v);case 89:case 107:case 82:return o(n.parent,function(n){return e.isIterationStatement(n,!0)},g);case 124:return i(e.isConstructorDeclaration,[124]);case 126:case 137:return i(e.isAccessor,[126,137]);case 122:return o(n.parent,e.isAwaitExpression,T);case 121:return s(T(n));case 117:return s(S(n));default:return e.isModifierKind(n.kind)&&(e.isDeclaration(n.parent)||e.isVariableStatement(n.parent))?s(m(n.kind,n.parent)):void 0}function i(r,a){return o(n.parent,r,function(n){return e.mapDefined(n.symbol.declarations,function(n){return r(n)?e.find(n.getChildren(t),function(n){return e.contains(a,n.kind)}):void 0})})}function o(e,n,r){return n(e)?s(r(e,t)):void 0}function s(e){return e&&e.map(function(e){return r(e,t)})}}function s(n){return e.isThrowStatement(n)?[n]:e.isTryStatement(n)?e.concatenate(n.catchClause?s(n.catchClause):n.tryBlock&&s(n.tryBlock),n.finallyBlock&&s(n.finallyBlock)):e.isFunctionLike(n)?void 0:u(n,s)}function c(n){var t=n;while(t.parent){var r=t.parent;if(e.isFunctionBlock(r)||279===r.kind)return r;if(e.isTryStatement(r)&&r.tryBlock===t&&r.catchClause)return t;t=r}}function l(n){return e.isBreakOrContinueStatement(n)?[n]:e.isFunctionLike(n)?void 0:u(n,l)}function u(n,t){var r=[];return n.forEachChild(function(n){var a=t(n);void 0!==a&&r.push.apply(r,e.toArray(a))}),r}function d(e,n){var t=p(n);return!!t&&t===e}function p(n){return e.findAncestor(n,function(t){switch(t.kind){case 232:if(228===n.kind)return!1;case 225:case 226:case 227:case 224:case 223:return!n.label||D(t,n.label.escapedText);default:return e.isFunctionLike(t)&&"quit"}})}function m(n,t){return e.mapDefined(f(t,e.modifierToFlag(n)),function(t){return e.findModifier(t,n)})}function f(n,t){var r=n.parent;switch(r.kind){case 245:case 279:case 218:case 271:case 272:return 128&t&&e.isClassDeclaration(n)?n.members.concat([n]):r.statements;case 157:case 156:case 239:return r.parameters.concat(e.isClassLike(r.parent)?r.parent.members:[]);case 240:case 209:var a=r.members;if(28&t){var i=e.find(r.members,e.isConstructorDeclaration);if(i)return a.concat(i.parameters)}else if(128&t)return a.concat([r]);return a;default:e.Debug.assertNever(r,"Invalid container kind.")}}function _(n,t){for(var r=[],a=2;a<arguments.length;a++)r[a-2]=arguments[a];return!(!t||!e.contains(r,t.kind))&&(n.push(t),!0)}function g(n){var t=[];if(_(t,n.getFirstToken(),89,107,82)&&223===n.kind)for(var r=n.getChildren(),a=r.length-1;a>=0;a--)if(_(t,r[a],107))break;return e.forEach(l(n.statement),function(e){d(n,e)&&_(t,e.getFirstToken(),73,78)}),t}function v(e){var n=p(e);if(n)switch(n.kind){case 225:case 226:case 227:case 223:case 224:return g(n);case 232:return y(n)}}function y(n){var t=[];return _(t,n.getFirstToken(),99),e.forEach(n.caseBlock.clauses,function(r){_(t,r.getFirstToken(),74,80),e.forEach(l(r),function(e){d(n,e)&&_(t,e.getFirstToken(),73)})}),t}function h(n,t){var r=[];if(_(r,n.getFirstToken(),103),n.catchClause&&_(r,n.catchClause.getFirstToken(),75),n.finallyBlock){var a=e.findChildOfKind(n,88,t);_(r,a,88)}return r}function b(n,t){var r=c(n);if(r){var a=[];return e.forEach(s(r),function(n){a.push(e.findChildOfKind(n,101,t))}),e.isFunctionBlock(r)&&e.forEachReturnStatement(r,function(n){a.push(e.findChildOfKind(n,97,t))}),a}}function E(n,t){var r=e.getContainingFunction(n);if(r){var a=[];return e.forEachReturnStatement(e.cast(r.body,e.isBlock),function(n){a.push(e.findChildOfKind(n,97,t))}),e.forEach(s(r.body),function(n){a.push(e.findChildOfKind(n,101,t))}),a}}function T(n){var t=e.getContainingFunction(n);if(t){var r=[];return t.modifiers&&t.modifiers.forEach(function(e){_(r,e,121)}),e.forEachChild(t,function(n){x(n,function(n){e.isAwaitExpression(n)&&_(r,n.getFirstToken(),122)})}),r}}function S(n){var t=e.getContainingFunction(n);if(t){var r=[];return e.forEachChild(t,function(n){x(n,function(n){e.isYieldExpression(n)&&_(r,n.getFirstToken(),117)})}),r}}function x(n,t){t(n),e.isFunctionLike(n)||e.isClassLike(n)||e.isInterfaceDeclaration(n)||e.isModuleDeclaration(n)||e.isTypeAliasDeclaration(n)||e.isTypeNode(n)||e.forEachChild(n,function(e){return x(e,t)})}function C(n,t){for(var a=A(n,t),i=[],o=0;o<a.length;o++){if(83===a[o].kind&&o<a.length-1){for(var s=a[o],c=a[o+1],l=!0,u=c.getStart(t)-1;u>=s.end;u--)if(!e.isWhiteSpaceSingleLine(t.text.charCodeAt(u))){l=!1;break}if(l){i.push({fileName:t.fileName,textSpan:e.createTextSpanFromBounds(s.getStart(),c.end),kind:"reference"}),o++;continue}}i.push(r(a[o],t))}return i}function A(n,t){var r=[];while(e.isIfStatement(n.parent)&&n.parent.elseStatement===n)n=n.parent;while(1){var a=n.getChildren(t);_(r,a[0],91);for(var i=a.length-1;i>=0;i--)if(_(r,a[i],83))break;if(!n.elseStatement||!e.isIfStatement(n.elseStatement))break;n=n.elseStatement}return r}function D(n,t){return!!e.findAncestor(n.parent,function(n){return e.isLabeledStatement(n)?n.label.escapedText===t:"quit"})}n.getDocumentHighlights=t})(e.DocumentHighlights||(e.DocumentHighlights={}))}(u||(u={})),function(e){function n(e,n){return t(e,n)}function t(n,t,a){void 0===t&&(t="");var i=e.createMap(),o=e.createGetCanonicalFileName(!!n);function s(){var n=e.arrayFrom(i.keys()).filter(function(e){return e&&"_"===e.charAt(0)}).map(function(e){var n=i.get(e),t=[];return n.forEach(function(e,n){t.push({name:n,refCount:e.languageServiceRefCount})}),t.sort(function(e,n){return n.refCount-e.refCount}),{bucket:e,sourceFiles:t}});return JSON.stringify(n,void 0,2)}function c(n,a,i,s,c){var u=e.toPath(n,t,o),d=r(a);return l(n,u,a,d,i,s,c)}function l(e,n,t,r,a,i,o){return p(e,n,t,r,a,i,!0,o)}function u(n,a,i,s,c){var l=e.toPath(n,t,o),u=r(a);return d(n,l,a,u,i,s,c)}function d(e,n,t,r,a,i,o){return p(e,n,t,r,a,i,!1,o)}function p(n,t,r,o,s,c,l,u){var d=e.getOrUpdate(i,o,e.createMap),p=d.get(t),m=6===u?100:r.target||1;if(!p&&a){var f=a.getDocument(o,t);f&&(e.Debug.assert(l),p={sourceFile:f,languageServiceRefCount:0},d.set(t,p))}if(p)p.sourceFile.version!==c&&(p.sourceFile=e.updateLanguageServiceSourceFile(p.sourceFile,s,c,s.getChangeRange(p.sourceFile.scriptSnapshot)),a&&a.setDocument(o,t,p.sourceFile)),l&&p.languageServiceRefCount++;else{f=e.createLanguageServiceSourceFile(n,s,m,c,!1,u);a&&a.setDocument(o,t,f),p={sourceFile:f,languageServiceRefCount:1},d.set(t,p)}return e.Debug.assert(0!==p.languageServiceRefCount),p.sourceFile}function m(n,a){var i=e.toPath(n,t,o),s=r(a);return f(i,s)}function f(n,t){var r=e.Debug.assertDefined(i.get(t)),a=r.get(n);a.languageServiceRefCount--,e.Debug.assert(a.languageServiceRefCount>=0),0===a.languageServiceRefCount&&r.delete(n)}function _(n){return e.arrayFrom(i.entries(),function(e){var t=e[0],r=e[1],a=r.get(n);return[t,a&&a.languageServiceRefCount]})}return{acquireDocument:c,acquireDocumentWithKey:l,updateDocument:u,updateDocumentWithKey:d,releaseDocument:m,releaseDocumentWithKey:f,getLanguageServiceRefCounts:_,reportStats:s,getKeyForCompilationSettings:r}}function r(n){return e.sourceFileAffectingCompilerOptions.map(function(t){return e.getCompilerOptionValue(n,t)}).join("|")}e.createDocumentRegistry=n,e.createDocumentRegistryInternal=t}(u||(u={})),function(e){(function(n){function t(e,n,t,i){var s=c(e,t,i);return function(c,l,u){var d=r(e,n,s,l,t,i),p=d.directImports,m=d.indirectUsers;return o({indirectUsers:m},a(p,c,l.exportKind,t,u))}}function r(n,t,r,a,o,s){var c=a.exportingModuleSymbol,l=a.exportKind,u=e.nodeSeenTracker(),d=e.nodeSeenTracker(),p=[],m=!!c.globalExports,f=m?void 0:[];return g(c),{directImports:p,indirectUsers:_()};function _(){if(m)return n;for(var r=0,a=c.declarations;r<a.length;r++){var i=a[r];e.isExternalModuleAugmentation(i)&&t.has(i.getSourceFile().fileName)&&b(i)}return f.map(e.getSourceFileOfNode)}function g(n){var t=T(n);if(t)for(var r=0,a=t;r<a.length;r++){var i=a[r];if(u(i))switch(s&&s.throwIfCancellationRequested(),i.kind){case 191:if(!m){var c=i.parent;if(2===l&&237===c.kind){var d=c.name;if(72===d.kind){p.push(d);break}}b(i.getSourceFile())}break;case 72:break;case 248:h(i,i.name,e.hasModifier(i,1),!1);break;case 249:p.push(i);var f=i.importClause&&i.importClause.namedBindings;f&&251===f.kind?h(i,f.name,!1,!0):!m&&e.isDefaultImport(i)&&b(y(i));break;case 255:i.exportClause?p.push(i):g(v(i,o));break;case 183:p.push(i);break;default:e.Debug.assertNever(i,"Unexpected import kind: "+e.Debug.showSyntaxKind(i))}}}function h(n,t,r,a){if(2===l)a||p.push(n);else if(!m){var s=y(n);e.Debug.assert(279===s.kind||244===s.kind),r||i(s,t,o)?E(s):b(s)}}function b(n){e.Debug.assert(!m);var t=d(n);return t&&f.push(n),t}function E(n){if(b(n)){var t=o.getMergedSymbol(n.symbol);e.Debug.assert(!!(1536&t.flags));var r=T(t);if(r)for(var a=0,i=r;a<i.length;a++){var s=i[a];E(y(s))}}}function T(n){return r.get(e.getSymbolId(n).toString())}}function a(n,t,r,a,i){var o=[],s=[];function c(e,n){o.push([e,n])}if(n)for(var l=0,u=n;l<u.length;l++){var d=u[l];p(d)}return{importSearches:o,singleReferences:s};function p(n){if(248!==n.kind)if(72!==n.kind)if(183!==n.kind){if(10===n.moduleSpecifier.kind)if(255!==n.kind){var o=n.importClause||{name:void 0,namedBindings:void 0},l=o.name,u=o.namedBindings;if(u)switch(u.kind){case 251:m(u.name);break;case 252:0!==r&&1!==r||f(u);break;default:e.Debug.assertNever(u)}if(l&&(1===r||2===r)&&(!i||l.escapedText===e.symbolEscapedNameNoDefault(t))){var d=a.getSymbolAtLocation(l);c(l,d)}}else f(n.exportClause)}else n.qualifier?e.isIdentifier(n.qualifier)&&n.qualifier.escapedText===e.symbolName(t)&&s.push(n.qualifier):2===r&&s.push(n.argument.literal);else m(n);else b(n)&&m(n.name)}function m(e){2!==r||i&&!_(e.escapedText)||c(e,a.getSymbolAtLocation(e))}function f(e){if(e)for(var n=0,r=e.elements;n<r.length;n++){var o=r[n],l=o.name,u=o.propertyName;if(_((u||l).escapedText))if(u)s.push(u),i&&l.escapedText!==t.escapedName||c(l,a.getSymbolAtLocation(l));else{var d=257===o.kind&&o.propertyName?a.getExportSpecifierLocalTargetSymbol(o):a.getSymbolAtLocation(l);c(l,d)}}}function _(e){return e===t.escapedName||0!==r&&"default"===e}}function i(n,t,r){var a=r.getSymbolAtLocation(t);return!!l(n,function(n){if(e.isExportDeclaration(n)){var t=n.exportClause,i=n.moduleSpecifier;return!i&&t&&t.elements.some(function(e){return r.getExportSpecifierLocalTargetSymbol(e)===a})}})}function s(e,n,t){for(var r=[],a=e.getTypeChecker(),i=0,o=n;i<o.length;i++){var s=o[i],c=t.valueDeclaration;if(279===c.kind){for(var l=0,d=s.referencedFiles;l<d.length;l++){var p=d[l];e.getSourceFileFromReference(s,p)===c&&r.push({kind:"reference",referencingFile:s,ref:p})}for(var m=0,f=s.typeReferenceDirectives;m<f.length;m++){p=f[m];var _=e.getResolvedTypeReferenceDirectives().get(p.fileName);void 0!==_&&_.resolvedFileName===c.fileName&&r.push({kind:"reference",referencingFile:s,ref:p})}}u(s,function(e,n){var i=a.getSymbolAtLocation(n);i===t&&r.push({kind:"import",literal:n})})}return r}function c(n,t,r){for(var a=e.createMap(),i=0,o=n;i<o.length;i++){var s=o[i];r&&r.throwIfCancellationRequested(),u(s,function(n,r){var i=t.getSymbolAtLocation(r);if(i){var o=e.getSymbolId(i).toString(),s=a.get(o);s||a.set(o,s=[]),s.push(n)}})}return a}function l(n,t){return e.forEach(279===n.kind?n.statements:n.body.statements,function(n){return t(n)||h(n)&&e.forEach(n.body&&n.body.statements,t)})}function u(n,t){if(n.externalModuleIndicator||void 0!==n.imports)for(var r=0,a=n.imports;r<a.length;r++){var i=a[r];t(e.importFromModuleSpecifier(i),i)}else l(n,function(n){switch(n.kind){case 255:case 249:var r=n;r.moduleSpecifier&&e.isStringLiteral(r.moduleSpecifier)&&t(r,r.moduleSpecifier);break;case 248:r=n;b(r)&&t(r,r.moduleReference.expression);break}})}function d(n,t,r,a){return a?i():i()||o();function i(){var i=n.parent,o=i.parent;if(t.exportSymbol)return 189===i.kind?t.declarations.some(function(e){return e===i})&&e.isBinaryExpression(o)?p(o,!1):void 0:s(t.exportSymbol,c(i));var l=m(i,n);if(l&&e.hasModifier(l,1)){if(e.isImportEqualsDeclaration(l)&&l.moduleReference===n){if(a)return;var u=r.getSymbolAtLocation(l.name);return{kind:0,symbol:u}}return s(t,c(l))}if(e.isExportAssignment(i))return d(i);if(e.isExportAssignment(o))return d(o);if(e.isBinaryExpression(i))return p(i,!0);if(e.isBinaryExpression(o))return p(o,!0);if(e.isJSDocTypedefTag(i))return s(t,0);function d(n){var r=e.Debug.assertDefined(n.symbol.parent,"Expected export symbol to have a parent"),a=n.isExportEquals?2:1;return{kind:1,symbol:t,exportInfo:{exportingModuleSymbol:r,exportKind:a}}}function p(n,a){var i;switch(e.getAssignmentDeclarationKind(n)){case 1:i=0;break;case 2:i=2;break;default:return}var o=a?r.getSymbolAtLocation(e.cast(n.left,e.isPropertyAccessExpression).name):t;return!o||1536&r.getMergedSymbol(o.parent).flags||e.Debug.fail("Special property assignment kind does not have a module as its parent. Assignment is "+e.Debug.showSymbol(o)+", parent is "+e.Debug.showSymbol(o.parent)),o&&s(o,i)}}function o(){var a=f(n);if(a){var i=r.getImmediateAliasedSymbol(t);if(i){i=g(i,r),"export="===i.escapedName&&(i=p(i,r));var o=e.symbolEscapedNameNoDefault(i);return void 0===o||"default"===o||o===t.escapedName?{kind:0,symbol:i}:void 0}}}function s(e,n){var t=_(e,n,r);return t&&{kind:1,symbol:e,exportInfo:t}}function c(n){return e.hasModifier(n,512)?1:0}}function p(n,t){if(2097152&n.flags)return e.Debug.assertDefined(t.getImmediateAliasedSymbol(n));var r=n.valueDeclaration;return e.isExportAssignment(r)?e.Debug.assertDefined(r.expression.symbol):e.isBinaryExpression(r)?e.Debug.assertDefined(r.right.symbol):e.isSourceFile(r)?e.Debug.assertDefined(r.symbol):e.Debug.fail()}function m(e,n){if(237===e.kind){var t=e;return t.name!==n?void 0:274===t.parent.kind?void 0:219===t.parent.parent.kind?t.parent.parent:void 0}return e}function f(n){var t=n.parent;switch(t.kind){case 248:return t.name===n&&b(t);case 253:return!t.propertyName;case 250:case 251:return e.Debug.assert(t.name===n),!0;default:return!1}}function _(n,t,r){var a=n.parent;if(a){var i=r.getMergedSymbol(a);return e.isExternalModuleSymbol(i)?{exportingModuleSymbol:i,exportKind:t}:void 0}}function g(n,t){if(n.declarations)for(var r=0,a=n.declarations;r<a.length;r++){var i=a[r];if(e.isExportSpecifier(i)&&!i.propertyName&&!i.parent.parent.moduleSpecifier)return t.getExportSpecifierLocalTargetSymbol(i)}return n}function v(e,n){return n.getMergedSymbol(y(e).symbol)}function y(n){if(191===n.kind)return n.getSourceFile();var t=n.parent;return 279===t.kind?t:(e.Debug.assert(245===t.kind),e.cast(t.parent,h))}function h(e){return 244===e.kind&&10===e.name.kind}function b(e){return 259===e.moduleReference.kind&&10===e.moduleReference.expression.kind}n.createImportTracker=t,function(e){e[e["Named"]=0]="Named",e[e["Default"]=1]="Default",e[e["ExportEquals"]=2]="ExportEquals"}(n.ExportKind||(n.ExportKind={})),function(e){e[e["Import"]=0]="Import",e[e["Export"]=1]="Export"}(n.ImportExport||(n.ImportExport={})),n.findModuleReferences=s,n.getImportOrExportSymbol=d,n.getExportInfo=_})(e.FindAllReferences||(e.FindAllReferences={}))}(u||(u={})),function(e){(function(n){function t(e,n){return void 0===n&&(n=1),{kind:n,node:e.name||e}}function r(t,r,a,i,o){var s=e.getTouchingPropertyName(i,o),c=n.Core.getReferencedSymbolsForNode(o,s,t,a,r),l=t.getTypeChecker();return c&&c.length?e.mapDefined(c,function(e){var n=e.definition,t=e.references;return n&&{definition:l.runWithCancellationToken(r,function(e){return u(n,e,s)}),references:t.map(m)}}):void 0}function a(n,t,r,a,o){var s=e.getTouchingPropertyName(a,o),c=i(n,t,r,s,o),l=n.getTypeChecker();return e.map(c,function(e){return g(e,l)})}function i(r,a,i,o,s){if(279!==o.kind){var l=r.getTypeChecker();if(276===o.parent.kind){var u=[];return n.Core.getReferenceEntriesForShorthandPropertyAssignment(o,l,function(e){return u.push(t(e))}),u}if(98===o.kind||e.isSuperProperty(o.parent)){var d=l.getSymbolAtLocation(o);return d.valueDeclaration&&[t(d.valueDeclaration)]}return c(s,o,r,i,a,{implementations:!0})}}function s(t,r,a,i,o,s,c){return e.map(l(n.Core.getReferencedSymbolsForNode(o,i,t,a,r,s)),function(e){return c(e,i,t.getTypeChecker())})}function c(t,r,a,i,o,s,c){return void 0===s&&(s={}),void 0===c&&(c=e.arrayToSet(i,function(e){return e.fileName})),l(n.Core.getReferencedSymbolsForNode(t,r,a,i,o,s,c))}function l(n){return n&&e.flatMap(n,function(e){return e.references})}function u(n,t,r){var a=function(){switch(n.type){case 0:var a=n.symbol,i=d(a,t,r),o=i.displayParts,s=i.kind,c=o.map(function(e){return e.text}).join("");return{node:a.declarations?e.getNameOfDeclaration(e.first(a.declarations))||e.first(a.declarations):r,name:c,kind:s,displayParts:o};case 1:var l=n.node;return{node:l,name:l.text,kind:"label",displayParts:[e.displayPart(l.text,e.SymbolDisplayPartKind.text)]};case 2:var u=n.node,p=e.tokenToString(u.kind);return{node:u,name:p,kind:"keyword",displayParts:[{text:p,kind:"keyword"}]};case 3:var m=n.node,f=(a=t.getSymbolAtLocation(m),a&&e.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(t,a,m.getSourceFile(),e.getContainerNode(m),m).displayParts||[e.textPart("this")]);return{node:m,name:"this",kind:"var",displayParts:f};case 4:var _=n.node;return{node:_,name:_.text,kind:"var",displayParts:[e.displayPart(e.getTextOfNode(_),e.SymbolDisplayPartKind.stringLiteral)]};default:return e.Debug.assertNever(n)}}(),i=a.node,o=a.name,s=a.kind,c=a.displayParts,l=i.getSourceFile();return{containerKind:"",containerName:"",fileName:l.fileName,kind:s,name:o,textSpan:h(e.isComputedPropertyName(i)?i.expression:i,l),displayParts:c}}function d(t,r,a){var i=n.Core.getIntersectingMeaningFromDeclarations(a,t),o=t.declarations&&e.firstOrUndefined(t.declarations)||a,s=e.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(r,t,o.getSourceFile(),o,o,i),c=s.displayParts,l=s.symbolKind;return{displayParts:c,kind:l}}function p(e,n,t,r){return o({},f(e),r&&_(e,n,t))}function m(n){var t=f(n),r=t.textSpan,a=t.fileName;if(0===n.kind)return{textSpan:r,fileName:a,isWriteAccess:!1,isDefinition:!1};var i=n.kind,o=n.node;return{textSpan:r,fileName:a,isWriteAccess:E(o),isDefinition:80===o.kind||!!e.getDeclarationFromName(o)||e.isLiteralComputedPropertyDeclarationName(o),isInString:2===i||void 0}}function f(e){if(0===e.kind)return{textSpan:e.textSpan,fileName:e.fileName};var n=e.node.getSourceFile();return{textSpan:h(e.node,n),fileName:n.fileName}}function _(n,t,r){if(0!==n.kind&&e.isIdentifier(t)){var a=n.node,i=n.kind,o=t.text,s=e.isShorthandPropertyAssignment(a.parent);if(s||e.isObjectBindingElementWithoutPropertyName(a.parent)){var c={prefixText:o+": "},l={suffixText:": "+o};return 3===i?c:4===i?l:s?l:c}if(e.isImportSpecifier(n.node.parent)&&!n.node.parent.propertyName){var u=e.isExportSpecifier(t.parent)?r.getExportSpecifierLocalTargetSymbol(t.parent):r.getSymbolAtLocation(t);return e.contains(u.declarations,n.node.parent)?{prefixText:o+" as "}:e.emptyOptions}if(e.isExportSpecifier(n.node.parent)&&!n.node.parent.propertyName)return t===n.node?{prefixText:o+" as "}:{suffixText:" as "+o}}return e.emptyOptions}function g(e,n){if(0!==e.kind){var t=e.node,r=t.getSourceFile();return o({textSpan:h(t,r),fileName:r.fileName},v(t,n))}var a=e.textSpan,i=e.fileName;return{textSpan:a,fileName:i,kind:"",displayParts:[]}}function v(n,t){var r=t.getSymbolAtLocation(e.isDeclaration(n)&&n.name?n.name:n);return r?d(r,t,n):188===n.kind?{kind:"interface",displayParts:[e.punctuationPart(20),e.textPart("object literal"),e.punctuationPart(21)]}:209===n.kind?{kind:"local class",displayParts:[e.punctuationPart(20),e.textPart("anonymous local class"),e.punctuationPart(21)]}:{kind:e.getNodeKind(n),displayParts:[]}}function y(e){if(0===e.kind){var n=e.fileName,t=e.textSpan;return{fileName:n,span:{textSpan:t,kind:"reference"}}}var r=e.node,a=e.kind,i=r.getSourceFile(),o=E(r),s={textSpan:h(r,i),kind:o?"writtenReference":"reference",isInString:2===a||void 0};return{fileName:i.fileName,span:s}}function h(n,t){var r=n.getStart(t),a=n.getEnd();return 10===n.kind&&(r+=1,a-=1),e.createTextSpanFromBounds(r,a)}function b(e){return 0===e.kind?e.textSpan:h(e.node,e.node.getSourceFile())}function E(n){var t=e.getDeclarationFromName(n);return!!t&&T(t)||80===n.kind||e.isWriteAccess(n)}function T(n){if(4194304&n.flags)return!0;switch(n.kind){case 204:case 186:case 240:case 209:case 80:case 243:case 278:case 257:case 250:case 248:case 253:case 241:case 297:case 304:case 267:case 244:case 247:case 251:case 151:case 276:case 242:case 150:return!0;case 275:return!e.isArrayLiteralOrObjectLiteralDestructuringPattern(n.parent);case 239:case 196:case 157:case 156:case 158:case 159:return!!n.body;case 237:case 154:return!!n.initializer||e.isCatchClause(n.parent);case 155:case 153:case 305:case 299:return!1;default:return e.Debug.failBadSyntaxKind(n)}}(function(e){e[e["Symbol"]=0]="Symbol",e[e["Label"]=1]="Label",e[e["Keyword"]=2]="Keyword",e[e["This"]=3]="This",e[e["String"]=4]="String"})(n.DefinitionKind||(n.DefinitionKind={})),function(e){e[e["Span"]=0]="Span",e[e["Node"]=1]="Node",e[e["StringLiteral"]=2]="StringLiteral",e[e["SearchedLocalFoundProperty"]=3]="SearchedLocalFoundProperty",e[e["SearchedPropertyFoundLocal"]=4]="SearchedPropertyFoundLocal"}(n.EntryKind||(n.EntryKind={})),n.nodeEntry=t,n.findReferencedSymbols=r,n.getImplementationsAtPosition=a,n.findReferenceOrRenameEntries=s,n.getReferenceEntriesForNode=c,n.toRenameLocation=p,n.toReferenceEntry=m,n.toHighlightSpan=y,n.getTextSpanOfEntry=b})(e.FindAllReferences||(e.FindAllReferences={}))}(u||(u={})),function(e){(function(n){(function(t){function r(n,t,r,s,d,p,m){if(void 0===p&&(p={}),void 0===m&&(m=e.arrayToSet(s,function(e){return e.fileName})),e.isSourceFile(t)){var f=e.GoToDefinition.getReferenceAtPosition(t,n,r),_=f&&r.getTypeChecker().getMergedSymbol(f.file.symbol);return _&&c(r,_,!1,s,m)}if(!p.implementations){var g=l(t,s,d);if(g)return g}var v=r.getTypeChecker(),y=v.getSymbolAtLocation(t);if(!y)return!p.implementations&&e.isStringLiteral(t)?Z(t,s,d):void 0;if("export="===y.escapedName)return c(r,y.parent,!1,s,m);var h=i(y,r,s,d,p,m);if(h&&!(33554432&y.flags))return h;var b=a(t,y,v),E=b&&i(b,r,s,d,p,m),T=u(y,t,s,m,v,d,p);return o(r,h,T,E)}function a(n,t,r){if(n.parent&&e.isNamespaceExportDeclaration(n.parent)){var a=r.getAliasedSymbol(t),i=r.getMergedSymbol(a);if(a!==i)return i}}function i(n,t,r,a,i,s){var l=1536&n.flags?e.find(n.declarations,e.isSourceFile):void 0;if(l){var d=n.exports.get("export="),p=c(t,n,!!d,r,s);if(!d||!s.has(l.fileName))return p;var m=t.getTypeChecker();return n=e.skipAlias(d,m),o(t,p,u(n,void 0,r,s,m,a,i))}}function o(t){for(var r,a=[],i=1;i<arguments.length;i++)a[i-1]=arguments[i];for(var o=0,c=a;o<c.length;o++){var l=c[o];if(l&&l.length)if(r)for(var u=function(a){if(!a.definition||0!==a.definition.type)return r.push(a),"continue";var i=a.definition.symbol,o=e.findIndex(r,function(e){return!!e.definition&&0===e.definition.type&&e.definition.symbol===i});if(-1===o)return r.push(a),"continue";var c=r[o];r[o]={definition:c.definition,references:c.references.concat(a.references).sort(function(r,a){var i=s(t,r),o=s(t,a);if(i!==o)return e.compareValues(i,o);var c=n.getTextSpanOfEntry(r),l=n.getTextSpanOfEntry(a);return c.start!==l.start?e.compareValues(c.start,l.start):e.compareValues(c.length,l.length)})}},d=0,p=l;d<p.length;d++){var m=p[d];u(m)}else r=l}return r}function s(e,n){var t=0===n.kind?e.getSourceFile(n.fileName):n.node.getSourceFile();return e.getSourceFiles().indexOf(t)}function c(t,r,a,i,o){e.Debug.assert(!!r.valueDeclaration);for(var s=e.mapDefined(n.findModuleReferences(t,i,r),function(t){if("import"===t.kind){var r=t.literal.parent;if(e.isLiteralTypeNode(r)){var i=e.cast(r.parent,e.isImportTypeNode);if(a&&!i.qualifier)return}return n.nodeEntry(t.literal)}return{kind:0,fileName:t.referencingFile.fileName,textSpan:e.createTextSpanFromRange(t.ref)}}),c=0,l=r.declarations;c<l.length;c++){var u=l[c];switch(u.kind){case 279:break;case 244:o.has(u.getSourceFile().fileName)&&s.push(n.nodeEntry(u.name));break;default:e.Debug.assert(!!(33554432&r.flags),"Expected a module symbol to be declared by a SourceFile or ModuleDeclaration.")}}var d=r.exports.get("export=");if(d)for(var p=0,m=d.declarations;p<m.length;p++){u=m[p];var f=u.getSourceFile();if(o.has(f.fileName)){var _=e.isBinaryExpression(u)&&e.isPropertyAccessExpression(u.left)?u.left.expression:e.isExportAssignment(u)?e.Debug.assertDefined(e.findChildOfKind(u,85,f)):e.getNameOfDeclaration(u)||u;s.push(n.nodeEntry(_))}}return s.length?[{definition:{type:0,symbol:r},references:s}]:e.emptyArray}function l(n,t,r){if(e.isTypeKeyword(n.kind))return N(t,n.kind,r);if(e.isJumpStatementTarget(n)){var a=e.getTargetLabel(n.parent,n.text);return a&&L(a.parent,a)}return e.isLabelOfLabeledStatement(n)?L(n.parent,n):e.isThis(n)?Q(n,t,r):98===n.kind?X(n):void 0}function u(n,t,r,a,i,o,s){var c=t&&p(n,t,i,!ce(s))||n,l=t?te(t,c):7,u=[],m=new _(r,a,t?d(t):0,i,o,l,s,u),f=ce(s)?e.find(c.declarations,e.isExportSpecifier):void 0;if(f)P(f.name,c,f,m.createSearch(t,n,void 0),m,!0,!0);else if(t&&80===t.kind)B(t,c,m),g(t,c,{exportingModuleSymbol:e.Debug.assertDefined(c.parent,"Expected export symbol to have a parent"),exportKind:1},m);else{var v=m.createSearch(t,c,void 0,{allSearchSymbols:t?$(c,t,i,!!s.isForRename,!!s.providePrefixAndSuffixTextForRename,!!s.implementations):[c]}),y=T(c);if(y)w(y,y.getSourceFile(),v,m,!(e.isSourceFile(y)&&!e.contains(r,y)));else for(var h=0,E=m.sourceFiles;h<E.length;h++){var S=E[h];m.cancellationToken.throwIfCancellationRequested(),b(S,v,m)}}return u}function d(n){switch(n.kind){case 124:return 1;case 72:if(e.isClassLike(n.parent))return e.Debug.assert(n.parent.name===n),2;default:return 0}}function p(n,t,r,a){var i=t.parent;return e.isExportSpecifier(i)&&a?R(t,n,i,r):e.firstDefined(n.declarations,function(a){if(!a.parent){if(33554432&n.flags)return;e.Debug.fail("Unexpected symbol at "+e.Debug.showSyntaxKind(t)+": "+e.Debug.showSymbol(n))}return e.isTypeLiteralNode(a.parent)&&e.isUnionTypeNode(a.parent.parent)?r.getPropertyOfType(r.getTypeFromTypeNode(a.parent.parent),n.name):void 0})}var m;function f(n){if(33555968&n.flags){var t=n.declarations&&e.find(n.declarations,function(n){return!e.isSourceFile(n)&&!e.isModuleDeclaration(n)});return t&&t.symbol}}t.getReferencedSymbolsForNode=r,function(e){e[e["None"]=0]="None",e[e["Constructor"]=1]="Constructor",e[e["Class"]=2]="Class"}(m||(m={}));var _=function(){function t(n,t,r,a,i,o,s,c){this.sourceFiles=n,this.sourceFilesSet=t,this.specialSearchKind=r,this.checker=a,this.cancellationToken=i,this.searchMeaning=o,this.options=s,this.result=c,this.inheritsFromCache=e.createMap(),this.markSeenContainingTypeReference=e.nodeSeenTracker(),this.markSeenReExportRHS=e.nodeSeenTracker(),this.symbolIdToReferences=[],this.sourceFileToSeenSymbols=[]}return t.prototype.includesSourceFile=function(e){return this.sourceFilesSet.has(e.fileName)},t.prototype.getImportSearches=function(e,t){return this.importTracker||(this.importTracker=n.createImportTracker(this.sourceFiles,this.sourceFilesSet,this.checker,this.cancellationToken)),this.importTracker(e,t,!!this.options.isForRename)},t.prototype.createSearch=function(n,t,r,a){void 0===a&&(a={});var i=a.text,o=void 0===i?e.stripQuotes(e.unescapeLeadingUnderscores((e.getLocalSymbolForExportDefault(t)||f(t)||t).escapedName)):i,s=a.allSearchSymbols,c=void 0===s?[t]:s,l=e.escapeLeadingUnderscores(o),u=this.options.implementations&&n?se(n,t,this.checker):void 0;return{symbol:t,comingFrom:r,text:o,escapedText:l,parents:u,allSearchSymbols:c,includes:function(n){return e.contains(c,n)}}},t.prototype.referenceAdder=function(t){var r=e.getSymbolId(t),a=this.symbolIdToReferences[r];return a||(a=this.symbolIdToReferences[r]=[],this.result.push({definition:{type:0,symbol:t},references:a})),function(e,t){return a.push(n.nodeEntry(e,t))}},t.prototype.addStringOrCommentReference=function(e,n){this.result.push({definition:void 0,references:[{kind:0,fileName:e,textSpan:n}]})},t.prototype.markSearchedSymbols=function(n,t){for(var r=e.getNodeId(n),a=this.sourceFileToSeenSymbols[r]||(this.sourceFileToSeenSymbols[r]=e.createMap()),i=!1,o=0,s=t;o<s.length;o++){var c=s[o];i=e.addToSeen(a,e.getSymbolId(c))||i}return i},t}();function g(e,n,t,r){var a=r.getImportSearches(n,t),i=a.importSearches,o=a.singleReferences,s=a.indirectUsers;if(o.length)for(var c=r.referenceAdder(n),l=0,u=o;l<u.length;l++){var d=u[l];y(d,r)&&c(d)}for(var p=0,m=i;p<m.length;p++){var f=m[p],_=f[0],g=f[1];I(_.getSourceFile(),r.createSearch(_,g,1),r)}if(s.length){var v=void 0;switch(t.exportKind){case 0:v=r.createSearch(e,n,1);break;case 1:v=r.options.isForRename?void 0:r.createSearch(e,n,1,{text:"default"});break;case 2:break}if(v)for(var h=0,E=s;h<E.length;h++){var T=E[h];b(T,v,r)}}}function v(t,r,a,i,o,s,c,l){for(var u=n.createImportTracker(t,e.arrayToSet(t,function(e){return e.fileName}),r,a),d=u(i,{exportKind:c?1:0,exportingModuleSymbol:o},!1),p=d.importSearches,m=d.indirectUsers,f=0,_=p;f<_.length;f++){var g=_[f][0];l(g)}for(var v=0,y=m;v<y.length;v++)for(var h=y[v],b=0,E=A(h,c?"default":s);b<E.length;b++){var T=E[b];e.isIdentifier(T)&&!e.isImportOrExportSpecifier(T.parent)&&r.getSymbolAtLocation(T)===i&&l(T)}}function y(n,t){return!!M(n,t)&&(!t.options.isForRename||!!e.isIdentifier(n)&&!(e.isImportOrExportSpecifier(n.parent)&&"default"===n.escapedText))}function h(e,n){for(var t=0,r=e.declarations;t<r.length;t++){var a=r[t],i=a.getSourceFile();I(i,n.createSearch(a,e,0),n,n.includesSourceFile(i))}}function b(n,t,r){void 0!==e.getNameTable(n).get(t.escapedText)&&I(n,t,r)}function E(n,t){return e.isArrayLiteralOrObjectLiteralDestructuringPattern(n.parent.parent)?t.getPropertySymbolOfDestructuringAssignment(n):void 0}function T(n){var t=n.declarations,r=n.flags,a=n.parent,i=n.valueDeclaration;if(i&&(196===i.kind||209===i.kind))return i;if(t){if(8196&r){var o=e.find(t,function(n){return e.hasModifier(n,8)});return o?e.getAncestor(o,240):void 0}if(!t.some(e.isObjectBindingElementWithoutPropertyName)){var s=a&&!(262144&n.flags);if(!s||e.isExternalModuleSymbol(a)&&!a.globalExports){for(var c,l=0,u=t;l<u.length;l++){var d=u[l],p=e.getContainerNode(d);if(c&&c!==p)return;if(!p||279===p.kind&&!e.isExternalOrCommonJsModule(p))return;c=p}return s?c.getSourceFile():c}}}}function S(e,n,t){return x(e,n,t,function(){return!0})||!1}function x(n,t,r,a){var i=e.isParameterPropertyDeclaration(n.parent)?e.first(t.getSymbolsOfParameterPropertyDeclaration(n.parent,n.text)):t.getSymbolAtLocation(n);if(i)for(var o=0,s=A(r,i.name);o<s.length;o++){var c=s[o];if(e.isIdentifier(c)&&c!==n&&c.escapedText===n.escapedText){var l=t.getSymbolAtLocation(c);if(l===i||t.getShorthandAssignmentValueSymbol(c.parent)===i||e.isExportSpecifier(c.parent)&&R(c,l,c.parent,t)===i){var u=a(c);if(u)return u}}}}function C(n,t,r,a){if(n.name&&e.isIdentifier(n.name))for(var i=e.Debug.assertDefined(r.getSymbolAtLocation(n.name)),o=0,s=t;o<s.length;o++)for(var c=s[o],l=0,u=A(c,i.name);l<u.length;l++){var d=u[l];if(e.isIdentifier(d)&&d!==n.name&&d.escapedText===n.name.escapedText){var p=e.climbPastPropertyAccess(d),m=p.parent;if(e.isCallExpression(m)&&m.expression===p){var f=r.getSymbolAtLocation(d);f&&r.getRootSymbols(f).some(function(e){return e===i})&&a(m)}}}}function A(n,t,r){return void 0===r&&(r=n),D(n,t,r).map(function(t){return e.getTouchingPropertyName(n,t)})}function D(n,t,r){void 0===r&&(r=n);var a=[];if(!t||!t.length)return a;var i=n.text,o=i.length,s=t.length,c=i.indexOf(t,r.pos);while(c>=0){if(c>r.end)break;var l=c+s;0!==c&&e.isIdentifierPart(i.charCodeAt(c-1),6)||l!==o&&e.isIdentifierPart(i.charCodeAt(l),6)||a.push(c),c=i.indexOf(t,c+s+1)}return a}function L(t,r){var a=t.getSourceFile(),i=r.text,o=e.mapDefined(A(a,i,t),function(t){return t===r||e.isJumpStatementTarget(t)&&e.getTargetLabel(t,i)===r?n.nodeEntry(t):void 0});return[{definition:{type:1,node:r},references:o}]}function k(n,t){switch(n.kind){case 72:return n.text.length===t.length;case 10:var r=n;return(e.isLiteralNameOfPropertyDeclarationOrIndexAccess(r)||e.isNameOfModuleDeclaration(n)||e.isExpressionOfExternalModuleImportEqualsDeclaration(n)||e.isCallExpression(n.parent)&&e.isBindableObjectDefinePropertyCall(n.parent)&&n.parent.arguments[1]===n)&&r.text.length===t.length;case 8:return e.isLiteralNameOfPropertyDeclarationOrIndexAccess(n)&&n.text.length===t.length;case 80:return"default".length===t.length;default:return!1}}function N(t,r,a){var i=e.flatMap(t,function(t){return a.throwIfCancellationRequested(),e.mapDefined(A(t,e.tokenToString(r),t),function(e){return e.kind===r?n.nodeEntry(e):void 0})});return i.length?[{definition:{type:2,node:i[0].node},references:i}]:void 0}function I(e,n,t,r){return void 0===r&&(r=!0),t.cancellationToken.throwIfCancellationRequested(),w(e,e,n,t,r)}function w(e,n,t,r,a){if(r.markSearchedSymbols(n,t.allSearchSymbols))for(var i=0,o=D(n,t.text,e);i<o.length;i++){var s=o[i];O(n,s,t,r,a)}}function M(n,t){return!!(e.getMeaningFromLocation(n)&t.searchMeaning)}function O(n,t,r,a,i){var o=e.getTouchingPropertyName(n,t);if(k(o,r.text)){if(M(o,a)){var s=a.checker.getSymbolAtLocation(o);if(s){var c=o.parent;if(!e.isImportSpecifier(c)||c.propertyName!==o){if(e.isExportSpecifier(c))return e.Debug.assert(72===o.kind),void P(o,s,c,r,a,i);var l=ne(r,s,o,a);if(l){switch(a.specialSearchKind){case 0:i&&B(o,l,a);break;case 1:K(o,n,r,a);break;case 2:j(o,r,a);break;default:e.Debug.assertNever(a.specialSearchKind)}G(o,s,r,a)}else V(s,r,a)}}}}else!a.options.implementations&&(a.options.findInStrings&&e.isInString(n,t)||a.options.findInComments&&e.isInNonReferenceComment(n,t))&&a.addStringOrCommentReference(n.fileName,e.createTextSpan(t,r.text.length))}function P(t,r,a,i,o,s,c){e.Debug.assert(!c||!!o.options.providePrefixAndSuffixTextForRename,"If alwaysGetReferences is true, then prefix/suffix text must be enabled");var l=a.parent,u=a.propertyName,d=a.name,p=l.parent,m=R(t,r,a,o.checker);if(c||i.includes(m)){if(u?t===u?(p.moduleSpecifier||b(),s&&!o.options.isForRename&&o.markSeenReExportRHS(d)&&B(d,e.Debug.assertDefined(a.symbol),o)):o.markSeenReExportRHS(t)&&b():o.options.isForRename&&"default"===d.escapedText||b(),!ce(o.options)||c){var f=80===t.originalKeywordKind?1:0,_=e.Debug.assertDefined(a.symbol),v=e.Debug.assertDefined(n.getExportInfo(_,f,o.checker));g(t,_,v,o)}if(1!==i.comingFrom&&p.moduleSpecifier&&!u&&!ce(o.options)){var y=o.checker.getExportSpecifierLocalTargetSymbol(a);y&&h(y,o)}}function b(){s&&B(t,m,o)}}function R(e,n,t,r){return F(e,t)&&r.getExportSpecifierLocalTargetSymbol(t)||n}function F(n,t){var r=t.parent,a=t.propertyName,i=t.name;return e.Debug.assert(a===n||i===n),a?a===n:!r.parent.moduleSpecifier}function G(e,t,r,a){var i=n.getImportOrExportSymbol(e,t,a.checker,1===r.comingFrom);if(i){var o=i.symbol;0===i.kind?ce(a.options)||h(o,a):g(e,o,i.exportInfo,a)}}function V(n,t,r){var a=n.flags,i=n.valueDeclaration,o=r.checker.getShorthandAssignmentValueSymbol(i),s=i&&e.getNameOfDeclaration(i);33554432&a||!s||!t.includes(o)||B(s,o,r)}function B(e,n,t){var r="kind"in n?n:{kind:void 0,symbol:n},a=r.kind,i=r.symbol,o=t.referenceAdder(i);t.options.implementations?W(e,o,t):o(e,a)}function K(n,t,r,a){e.isNewExpressionTarget(n)&&B(n,r.symbol,a);var i=function(){return a.referenceAdder(r.symbol)};if(e.isClassLike(n.parent))e.Debug.assert(80===n.kind||n.parent.name===n),U(r.symbol,t,i());else{var o=oe(n);o&&H(o,i())}}function j(n,t,r){B(n,t.symbol,r);var a=n.parent;if(!r.options.isForRename&&e.isClassLike(a)){e.Debug.assert(a.name===n);for(var i=r.referenceAdder(t.symbol),o=0,s=a.members;o<s.length;o++){var c=s[o];e.isMethodOrAccessor(c)&&e.hasModifier(c,32)&&(c.body&&c.body.forEachChild(function n(t){100===t.kind?i(t):e.isFunctionLike(t)||e.isClassLike(t)||t.forEachChild(n)}))}}}function U(n,t,r){for(var a=0,i=n.members.get("__constructor").declarations;a<i.length;a++){var o=i[a],s=e.findChildOfKind(o,124,t);e.Debug.assert(157===o.kind&&!!s),r(s)}n.exports.forEach(function(n){var t=n.valueDeclaration;if(t&&156===t.kind){var a=t.body;a&&ie(a,100,function(n){e.isNewExpressionTarget(n)&&r(n)})}})}function H(n,t){var r=n.symbol.members.get("__constructor");if(r)for(var a=0,i=r.declarations;a<i.length;a++){var o=i[a];e.Debug.assert(157===o.kind);var s=o.body;s&&ie(s,98,function(n){e.isCallExpressionTarget(n)&&t(n)})}}function W(n,t,r){if(e.isDeclarationName(n)&&re(n.parent))t(n);else if(72===n.kind){276===n.parent.kind&&ae(n,r.checker,t);var a=q(n);if(a)t(a);else{var i=e.findAncestor(n,function(n){return!e.isQualifiedName(n.parent)&&!e.isTypeNode(n.parent)&&!e.isTypeElement(n.parent)}),o=i.parent;if(e.hasType(o)&&o.type===i&&r.markSeenContainingTypeReference(o))if(e.hasInitializer(o))c(o.initializer);else if(e.isFunctionLike(o)&&o.body){var s=o.body;218===s.kind?e.forEachReturnStatement(s,function(e){e.expression&&c(e.expression)}):c(s)}else e.isAssertionExpression(o)&&c(o.expression)}}function c(e){z(e)&&t(e)}}function q(n){return e.isIdentifier(n)||e.isPropertyAccessExpression(n)?q(n.parent):e.isExpressionWithTypeArguments(n)?e.tryCast(n.parent.parent,e.isClassLike):void 0}function z(e){switch(e.kind){case 195:return z(e.expression);case 197:case 196:case 188:case 209:case 187:return!0;default:return!1}}function J(n,t,r,a){if(n===t)return!0;var i=e.getSymbolId(n)+","+e.getSymbolId(t),o=r.get(i);if(void 0!==o)return o;r.set(i,!1);var s=n.declarations.some(function(n){return e.getAllSuperTypeNodes(n).some(function(e){var n=a.getTypeAtLocation(e);return!!n&&!!n.symbol&&J(n.symbol,t,r,a)})});return r.set(i,s),s}function X(t){var r=e.getSuperContainer(t,!1);if(r){var a=32;switch(r.kind){case 154:case 153:case 156:case 155:case 157:case 158:case 159:a&=e.getModifierFlags(r),r=r.parent;break;default:return}var i=r.getSourceFile(),o=e.mapDefined(A(i,"super",r),function(t){if(98===t.kind){var i=e.getSuperContainer(t,!1);return i&&(32&e.getModifierFlags(i))===a&&i.parent.symbol===r.symbol?n.nodeEntry(t):void 0}});return[{definition:{type:0,symbol:r.symbol},references:o}]}}function Y(e){return 72===e.kind&&151===e.parent.kind&&e.parent.name===e}function Q(t,r,a){var i=e.getThisContainer(t,!1),o=32;switch(i.kind){case 156:case 155:if(e.isObjectLiteralMethod(i))break;case 154:case 153:case 157:case 158:case 159:o&=e.getModifierFlags(i),i=i.parent;break;case 279:if(e.isExternalModule(i)||Y(t))return;case 239:case 196:break;default:return}var s=e.flatMap(279===i.kind?r:[i.getSourceFile()],function(n){return a.throwIfCancellationRequested(),A(n,"this",e.isSourceFile(i)?n:i).filter(function(n){if(!e.isThis(n))return!1;var t=e.getThisContainer(n,!1);switch(i.kind){case 196:case 239:return i.symbol===t.symbol;case 156:case 155:return e.isObjectLiteralMethod(i)&&i.symbol===t.symbol;case 209:case 240:return t.parent&&i.symbol===t.parent.symbol&&(32&e.getModifierFlags(t))===o;case 279:return 279===t.kind&&!e.isExternalModule(t)&&!Y(n)}})}).map(function(e){return n.nodeEntry(e)}),c=e.firstDefined(s,function(n){return e.isParameter(n.node.parent)?n.node:void 0});return[{definition:{type:3,node:c||t},references:s}]}function Z(t,r,a){var i=e.flatMap(r,function(r){return a.throwIfCancellationRequested(),e.mapDefined(A(r,t.text),function(r){return e.isStringLiteral(r)&&r.text===t.text?n.nodeEntry(r,2):void 0})});return[{definition:{type:4,node:t},references:i}]}function $(e,n,t,r,a,i){var o=[];return ee(e,n,t,r,!(r&&a),function(e,n,t){o.push(t||n||e)},function(){return!i}),o}function ee(n,t,r,i,o,s,c){var l=e.getContainingObjectLiteralElement(t);if(l){var u=r.getShorthandAssignmentValueSymbol(t.parent);if(u&&i)return s(u,void 0,void 0,3);var d=r.getContextualType(l.parent),p=d&&e.firstDefined(e.getPropertySymbolsFromContextualType(l,r,d,!0),function(e){return S(e,4)});if(p)return p;var m=E(t,r),f=m&&s(m,void 0,void 0,4);if(f)return f;var _=u&&s(u,void 0,void 0,3);if(_)return _}var g=a(t,n,r);if(g){var v=s(g,void 0,void 0,1);if(v)return v}var y=S(n);if(y)return y;if(n.valueDeclaration&&e.isParameterPropertyDeclaration(n.valueDeclaration)){var h=r.getSymbolsOfParameterPropertyDeclaration(e.cast(n.valueDeclaration,e.isParameter),n.name);return e.Debug.assert(2===h.length&&!!(1&h[0].flags)&&!!(4&h[1].flags)),S(1&n.flags?h[1]:h[0])}if(!i){var b=void 0;return b=o?e.isObjectBindingElementWithoutPropertyName(t.parent)?e.getPropertySymbolFromBindingElement(r,t.parent):void 0:x(n,r),b&&S(b,4)}e.Debug.assert(i);var T=o;if(T){b=x(n,r);return b&&S(b,4)}function S(n,t){return e.firstDefined(r.getRootSymbols(n),function(a){return s(n,a,void 0,t)||(a.parent&&96&a.parent.flags&&c(a)?e.getPropertySymbolsFromBaseTypes(a.parent,a.name,r,function(e){return s(n,a,e,t)}):void 0)})}function x(n,t){var r=e.getDeclarationOfKind(n,186);if(r&&e.isObjectBindingElementWithoutPropertyName(r))return e.getPropertySymbolFromBindingElement(t,r)}}function ne(n,t,r,a){var i=a.checker;return ee(t,r,i,!1,!a.options.isForRename||!!a.options.providePrefixAndSuffixTextForRename,function(t,r,a,i){return n.includes(a||r||t)?{symbol:!r||6&e.getCheckFlags(t)?t:r,kind:i}:void 0},function(e){return!(n.parents&&!n.parents.some(function(n){return J(e.parent,n,a.inheritsFromCache,i)}))})}function te(n,t){var r=e.getMeaningFromLocation(n),a=t.declarations;if(a){var i=void 0;do{i=r;for(var o=0,s=a;o<s.length;o++){var c=s[o],l=e.getMeaningFromDeclaration(c);l&r&&(r|=l)}}while(r!==i)}return r}function re(n){return 4194304&n.flags?!(e.isInterfaceDeclaration(n)||e.isTypeAliasDeclaration(n)):e.isVariableLike(n)?e.hasInitializer(n):e.isFunctionLikeDeclaration(n)?!!n.body:e.isClassLike(n)||e.isModuleOrEnumDeclaration(n)}function ae(n,t,r){var a=t.getSymbolAtLocation(n),i=t.getShorthandAssignmentValueSymbol(a.valueDeclaration);if(i)for(var o=0,s=i.getDeclarations();o<s.length;o++){var c=s[o];1&e.getMeaningFromDeclaration(c)&&r(c)}}function ie(n,t,r){e.forEachChild(n,function(e){e.kind===t&&r(e),ie(e,t,r)})}function oe(n){return e.tryGetClassExtendingExpressionWithTypeArguments(e.climbPastPropertyAccess(n).parent)}function se(n,t,r){var a=e.isRightSideOfPropertyAccess(n)?n.parent:void 0,i=a&&r.getTypeAtLocation(a.expression),o=e.mapDefined(i&&(i.isUnionOrIntersection()?i.types:i.symbol===t.parent?void 0:[i]),function(e){return e.symbol&&96&e.symbol.flags?e.symbol:void 0});return 0===o.length?void 0:o}function ce(e){return e.isForRename&&e.providePrefixAndSuffixTextForRename}t.eachExportReference=v,t.isSymbolReferencedInFile=S,t.eachSymbolReferenceInFile=x,t.eachSignatureCall=C,t.getIntersectingMeaningFromDeclarations=te,t.getReferenceEntriesForShorthandPropertyAssignment=ae})(n.Core||(n.Core={}))})(e.FindAllReferences||(e.FindAllReferences={}))}(u||(u={})),function(e){function n(n,r,o,s,c,l,u){var d=e.hostUsesCaseSensitiveFileNames(s),p=e.createGetCanonicalFileName(d),m=t(r,o,p,u),f=t(o,r,p,u);return e.textChanges.ChangeTracker.with({host:s,formatContext:c},function(e){a(n,e,m,r,o,s.getCurrentDirectory(),d),i(n,e,m,f,s,p)})}function t(n,t,a,i){var o=a(n);return function(e){var n=i&&i.tryGetSourcePosition({fileName:e,pos:0}),t=s(n?n.fileName:e);return n?void 0===t?void 0:r(n.fileName,t,e,a):t};function s(n){if(a(n)===o)return t;var r=e.tryRemoveDirectoryPrefix(n,o,a);return void 0===r?void 0:t+"/"+r}}function r(n,t,r,a){var i=e.getRelativePathFromFile(n,t,a);return s(e.getDirectoryPath(r),i)}function a(n,t,r,a,i,o,c){var l=n.getCompilerOptions().configFile;if(l){var u=e.getDirectoryPath(l.fileName),m=e.getTsConfigObjectLiteralExpression(l);m&&p(m,function(n,r){switch(r){case"files":case"include":case"exclude":var s=f(n);if(!s&&"include"===r&&e.isArrayLiteralExpression(n.initializer)){var d=e.mapDefined(n.initializer.elements,function(n){return e.isStringLiteral(n)?n.text:void 0}),m=e.getFileMatcherPatterns(u,[],d,c,o);e.getRegexFromPattern(e.Debug.assertDefined(m.includeFilePattern),c).test(a)&&!e.getRegexFromPattern(e.Debug.assertDefined(m.includeFilePattern),c).test(i)&&t.insertNodeAfter(l,e.last(n.initializer.elements),e.createStringLiteral(g(i)))}break;case"compilerOptions":p(n.initializer,function(n,t){var r=e.getOptionFromName(t);r&&(r.isFilePath||"list"===r.type&&r.element.isFilePath)?f(n):"paths"===t&&p(n.initializer,function(n){if(e.isArrayLiteralExpression(n.initializer))for(var t=0,r=n.initializer.elements;t<r.length;t++){var a=r[t];_(a)}})});break}})}function f(n){for(var t=e.isArrayLiteralExpression(n.initializer)?n.initializer.elements:[n.initializer],r=!1,a=0,i=t;a<i.length;a++){var o=i[a];r=_(o)||r}return r}function _(n){if(!e.isStringLiteral(n))return!1;var a=s(u,n.text),i=r(a);return void 0!==i&&(t.replaceRangeWithText(l,d(n,l),g(i)),!0)}function g(n){return e.getRelativePathFromDirectory(u,n,!c)}}function i(n,t,r,a,i,o){for(var d=n.getSourceFiles(),p=function(p){var m=r(p.path),f=void 0!==m?m:p.path,_=e.getDirectoryPath(f),g=a(p.fileName),v=g||p.fileName,y=e.getDirectoryPath(v),h=void 0!==m||void 0!==g;u(p,t,function(n){if(e.pathIsRelative(n)){var t=s(y,n),a=r(t);return void 0===a?void 0:e.ensurePathIsNonModuleName(e.getRelativePathFromDirectory(_,a,o))}},function(t){var a=n.getTypeChecker().getSymbolAtLocation(t);if(!a||!a.declarations.some(function(n){return e.isAmbientModule(n)})){var o=void 0!==g?l(e.resolveModuleName(t.text,v,n.getCompilerOptions(),i),r):c(a,t,p,n,i,r);return void 0!==o&&(o.updated||h&&e.pathIsRelative(t.text))?e.moduleSpecifiers.updateModuleSpecifier(n.getCompilerOptions(),f,o.newFileName,i,d,n.redirectTargetsMap,t.text):void 0}})},m=0,f=d;m<f.length;m++){var _=f[m];p(_)}}function o(n,t){return e.normalizePath(e.combinePaths(n,t))}function s(n,t){return e.ensurePathIsNonModuleName(o(n,t))}function c(n,t,r,a,i,o){if(n){var s=e.find(n.declarations,e.isSourceFile).fileName,c=o(s);return void 0===c?{newFileName:s,updated:!1}:{newFileName:c,updated:!0}}var u=i.resolveModuleNames?i.getResolvedModuleWithFailedLookupLocationsFromCache&&i.getResolvedModuleWithFailedLookupLocationsFromCache(t.text,r.fileName):a.getResolvedModuleWithFailedLookupLocationsFromCache(t.text,r.fileName);return l(u,o)}function l(n,t){if(n){if(n.resolvedModule){var r=o(n.resolvedModule.resolvedFileName);if(r)return r}var a=e.forEach(n.failedLookupLocations,i);return a||n.resolvedModule&&{newFileName:n.resolvedModule.resolvedFileName,updated:!1}}function i(n){return e.endsWith(n,"/package.json")?void 0:o(n)}function o(e){var n=t(e);return n&&{newFileName:n,updated:!0}}}function u(n,t,r,a){for(var i=0,o=n.referencedFiles||e.emptyArray;i<o.length;i++){var s=o[i],c=r(s.fileName);void 0!==c&&c!==n.text.slice(s.pos,s.end)&&t.replaceRangeWithText(n,s,c)}for(var l=0,u=n.imports;l<u.length;l++){var p=u[l];c=a(p);void 0!==c&&c!==p.text&&t.replaceRangeWithText(n,d(p,n),c)}}function d(n,t){return e.createRange(n.getStart(t)+1,n.end-1)}function p(n,t){if(e.isObjectLiteralExpression(n))for(var r=0,a=n.properties;r<a.length;r++){var i=a[r];e.isPropertyAssignment(i)&&e.isStringLiteral(i.name)&&t(i,i.name.text)}}e.getEditsForFileRename=n,e.getPathUpdater=t}(u||(u={})),function(e){(function(n){function t(n,t,i){var o=a(t,i,n);if(o)return[v(o.fileName,o.file.fileName)];var s=e.getTouchingPropertyName(t,i);if(s!==t){var c=s.parent;if(e.isJumpStatementTarget(s)){var d=e.getTargetLabel(s.parent,s.text);return d?[f(d,"label",s.text,void 0)]:void 0}var g=n.getTypeChecker(),y=u(s,g);if(!y)return l(s,g);var h=b(g,s);if(h&&(!e.isJsxOpeningLikeElement(s.parent)||!E(h))){var T=_(g,h);if(g.getRootSymbols(y).some(function(e){return r(e,h)})||y.declarations&&y.declarations.some(function(n){return e.isVariableDeclaration(n)&&!!n.initializer&&e.isRequireCall(n.initializer,!1)}))return[T];var S=p(g,y,s)||e.emptyArray;return 98===s.kind?[T].concat(S):S.concat([T])}if(276===s.parent.kind){var x=g.getShorthandAssignmentValueSymbol(y.valueDeclaration);return x?x.declarations.map(function(e){return m(e,g,x,s)}):[]}if(e.isPropertyName(s)&&e.isBindingElement(c)&&e.isObjectBindingPattern(c.parent)&&s===(c.propertyName||c.name)){var C=e.getNameFromPropertyName(s),A=g.getTypeAtLocation(c.parent);return void 0===C?e.emptyArray:e.flatMap(A.isUnion()?A.types:[A],function(e){var n=e.getProperty(C);return n&&p(g,n,s)})}var D=e.getContainingObjectLiteralElement(s);if(D){var L=D&&g.getContextualType(D.parent);if(L)return e.flatMap(e.getPropertySymbolsFromContextualType(D,g,L,!1),function(e){return p(g,e,s)})}return p(g,y,s)}}function r(n,t){return n===t.symbol||n===t.symbol.parent||!e.isCallLikeExpression(t.parent)&&n===t.parent.symbol}function a(n,t,r){var a=g(n.referencedFiles,t);if(a){var i=e.tryResolveScriptReference(r,n,a);return i&&{fileName:a.fileName,file:i}}var o=g(n.typeReferenceDirectives,t);if(o){var s=r.getResolvedTypeReferenceDirectives().get(o.fileName);i=s&&r.getSourceFile(s.resolvedFileName);return i&&{fileName:o.fileName,file:i}}var c=g(n.libReferenceDirectives,t);if(c){i=r.getLibFileFromReference(c);return i&&{fileName:c.fileName,file:i}}}function i(n,t,r){var a=e.getTouchingPropertyName(t,r);if(a!==t){var i=n.getSymbolAtLocation(a);if(i){var c=n.getTypeOfSymbolAtLocation(i,a),l=s(i,c,n),u=l&&o(l,n,a);return u&&0!==u.length?u:o(c,n,a)}}}function o(n,t,r){return e.flatMap(!n.isUnion()||32&n.flags?[n]:n.types,function(e){return e.symbol&&p(t,e.symbol,r)})}function s(n,t,r){if(t.symbol===n||n.valueDeclaration&&t.symbol&&e.isVariableDeclaration(n.valueDeclaration)&&n.valueDeclaration.initializer===t.symbol.valueDeclaration){var a=t.getCallSignatures();if(1===a.length)return r.getReturnTypeOfSignature(e.first(a))}}function c(n,r,a){var i=t(n,r,a);if(i&&0!==i.length){var o=g(r.referencedFiles,a)||g(r.typeReferenceDirectives,a)||g(r.libReferenceDirectives,a);if(o)return{definitions:i,textSpan:e.createTextSpanFromRange(o)};var s=e.getTouchingPropertyName(r,a),c=e.createTextSpan(s.getStart(),s.getWidth());return{definitions:i,textSpan:c}}}function l(n,t){if(e.isPropertyAccessExpression(n.parent)&&n.parent.name===n){var r=t.getTypeAtLocation(n.parent.expression);return e.mapDefined(r.isUnionOrIntersection()?r.types:[r],function(e){var n=t.getIndexInfoOfType(e,0);return n&&n.declaration&&_(t,n.declaration)})}}function u(e,n){var t=n.getSymbolAtLocation(e);if(t&&2097152&t.flags&&d(e,t.declarations[0])){var r=n.getAliasedSymbol(t);if(r.declarations)return r}return t}function d(e,n){if(72!==e.kind)return!1;if(e.parent===n)return!0;switch(n.kind){case 250:case 248:return!0;case 253:return 252===n.parent.kind;default:return!1}}function p(n,t,r){return a()||i()||e.map(t.declarations,function(e){return m(e,n,t,r)});function a(){if(32&t.flags&&(e.isNewExpressionTarget(r)||124===r.kind)){var n=e.find(t.declarations,e.isClassLike)||e.Debug.fail("Expected declaration to have at least one class-like declaration");return o(n.members,!0)}}function i(){return e.isCallOrNewExpressionTarget(r)||e.isNameOfFunctionDeclaration(r)?o(t.declarations,!1):void 0}function o(a,i){if(a){var o=a.filter(i?e.isConstructorDeclaration:e.isFunctionLike);return o.length?[m(e.find(o,function(e){return!!e.body})||e.last(o),n,t,r)]:void 0}}}function m(n,t,r,a){var i=t.symbolToString(r),o=e.SymbolDisplay.getSymbolKind(t,r,a),s=r.parent?t.symbolToString(r.parent,a):"";return f(n,o,i,s)}function f(n,t,r,a){var i=e.getNameOfDeclaration(n)||n,o=i.getSourceFile();return{fileName:o.fileName,textSpan:e.createTextSpanFromNode(i,o),kind:t,name:r,containerKind:void 0,containerName:a}}function _(e,n){return m(n,e,n.symbol,n)}function g(n,t){return e.find(n,function(n){return e.textRangeContainsPositionInclusive(n,t)})}function v(n,t){return{fileName:t,textSpan:e.createTextSpanFromBounds(0,0),kind:"script",name:n,containerName:void 0,containerKind:void 0}}function y(n){var t=h(n),r=t.parent;return r&&e.isCallLikeExpression(r)&&e.getInvokedExpression(r)===t?r:void 0}function h(n){return e.isRightSideOfPropertyAccess(n)?h(n.parent):n}function b(n,t){var r=y(t),a=r&&n.getResolvedSignature(r);return e.tryCast(a&&a.declaration,function(n){return e.isFunctionLike(n)&&!e.isFunctionTypeNode(n)})}function E(e){switch(e.kind){case 157:case 166:case 161:return!0;default:return!1}}n.getDefinitionAtPosition=t,n.getReferenceAtPosition=a,n.getTypeDefinitionAtPosition=i,n.getDefinitionAndBoundSpan=c,n.findReferenceInPosition=g})(e.GoToDefinition||(e.GoToDefinition={}))}(u||(u={})),function(e){(function(n){var t,r,a=["abstract","access","alias","argument","async","augments","author","borrows","callback","class","classdesc","constant","constructor","constructs","copyright","default","deprecated","description","emits","enum","event","example","exports","extends","external","field","file","fileoverview","fires","function","generator","global","hideconstructor","host","ignore","implements","inheritdoc","inner","instance","interface","kind","lends","license","listens","member","memberof","method","mixes","module","name","namespace","override","package","param","private","property","protected","public","readonly","requires","returns","see","since","static","summary","template","this","throws","todo","tutorial","type","typedef","var","variation","version","virtual","yields"];function i(n){var t=[];return l(n,function(n){for(var r=0,a=o(n);r<a.length;r++){var i=a[r].comment;void 0!==i&&(t.length&&t.push(e.lineBreakPart()),t.push(e.textPart(i)))}}),t}function o(n){switch(n.kind){case 299:case 305:return[n];case 297:case 304:return[n,n.parent];default:return e.getJSDocCommentsAndTags(n)}}function s(n){var t=[];return l(n,function(n){for(var r=0,a=e.getJSDocTags(n);r<a.length;r++){var i=a[r];t.push({name:i.tagName.text,text:c(i)})}}),t}function c(e){var n=e.comment;switch(e.kind){case 295:return r(e.class);case 303:return a(e.typeParameters);case 302:return r(e.typeExpression);case 304:case 297:case 305:case 299:var t=e.name;return t?r(t):n;default:return n}function r(e){return i(e.getText())}function a(e){return i(e.map(function(e){return e.getText()}).join(", "))}function i(e){return void 0===n?e:e+" "+n}}function l(e,n){if(e)for(var t=0;t<e.length;t++)if(e.indexOf(e[t])===t){var r=n(e[t],t);if(r)return r}}function u(){return t||(t=e.map(a,function(e){return{name:e,kind:"keyword",kindModifiers:"",sortText:"0"}}))}function d(){return r||(r=e.map(a,function(e){return{name:"@"+e,kind:"keyword",kindModifiers:"",sortText:"0"}}))}function p(n){return{name:n,kind:"",kindModifiers:"",displayParts:[e.textPart(n)],documentation:e.emptyArray,tags:void 0,codeActions:void 0}}function m(n){if(!e.isIdentifier(n.name))return e.emptyArray;var t=n.name.text,r=n.parent,a=r.parent;return e.isFunctionLike(a)?e.mapDefined(a.parameters,function(a){if(e.isIdentifier(a.name)){var i=a.name.text;if(!r.tags.some(function(t){return t!==n&&e.isJSDocParameterTag(t)&&e.isIdentifier(t.name)&&t.name.escapedText===i})&&(void 0===t||e.startsWith(i,t)))return{name:i,kind:"parameter",kindModifiers:"",sortText:"0"}}}):[]}function f(n){return{name:n,kind:"parameter",kindModifiers:"",displayParts:[e.textPart(n)],documentation:e.emptyArray,tags:void 0,codeActions:void 0}}function _(n,t,r){var a=e.getTokenAtPosition(t,r),i=e.findAncestor(a,e.isJSDoc);if(!i||void 0===i.comment&&!e.length(i.tags)){var o=a.getStart(t);if(i||!(o<r)){var s=y(a);if(s){var c=s.commentOwner,l=s.parameters;if(!(c.getStart(t)<r)){if(!l||0===l.length){var u="/** */";return{newText:u,caretOffset:3}}var d=g(t,r),p="/**"+n+d+" * ",m=p+n+v(l,e.hasJSFileExtension(t.fileName),d,n)+d+" */"+(o===r?n+d:"");return{newText:m,caretOffset:p.length}}}}}}function g(n,t){for(var r=n.text,a=e.getLineStartPositionForPosition(t,n),i=a;i<=t&&e.isWhiteSpaceSingleLine(r.charCodeAt(i));i++);return r.slice(a,i)}function v(e,n,t,r){return e.map(function(e,a){var i=e.name,o=e.dotDotDotToken,s=72===i.kind?i.text:"param"+a,c=n?o?"{...any} ":"{any} ":"";return t+" * @param "+c+s+r}).join("")}function y(n){return e.forEachAncestor(n,h)}function h(n){switch(n.kind){case 239:case 196:case 156:case 157:case 155:var t=n.parameters;return{commentOwner:n,parameters:t};case 275:return h(n.initializer);case 240:case 241:case 153:case 243:case 278:case 242:return{commentOwner:n};case 219:var r=n,a=r.declarationList.declarations,i=1===a.length&&a[0].initializer?b(a[0].initializer):void 0;return{commentOwner:n,parameters:i};case 279:return"quit";case 244:return 244===n.parent.kind?void 0:{commentOwner:n};case 204:var o=n;if(0===e.getAssignmentDeclarationKind(o))return"quit";var s=e.isFunctionLike(o.right)?o.right.parameters:e.emptyArray;return{commentOwner:n,parameters:s}}}function b(n){while(195===n.kind)n=n.expression;switch(n.kind){case 196:case 197:return n.parameters;case 209:var t=e.find(n.members,e.isConstructorDeclaration);return t?t.parameters:e.emptyArray}return e.emptyArray}n.getJsDocCommentsFromDeclarations=i,n.getJsDocTagsFromDeclarations=s,n.getJSDocTagNameCompletions=u,n.getJSDocTagNameCompletionDetails=p,n.getJSDocTagCompletions=d,n.getJSDocTagCompletionDetails=p,n.getJSDocParameterNameCompletions=m,n.getJSDocParameterNameCompletionDetails=f,n.getDocCommentTemplateAtPosition=_})(e.JsDoc||(e.JsDoc={}))}(u||(u={})),function(e){(function(n){function t(n,t,a,i,o,s){var c=e.createPatternMatcher(i);if(!c)return e.emptyArray;for(var d=[],p=function(e){if(a.throwIfCancellationRequested(),s&&e.isDeclarationFile)return"continue";e.getNamedDeclarations().forEach(function(n,a){r(c,a,n,t,e.fileName,d)})},m=0,f=n;m<f.length;m++){var _=f[m];p(_)}return d.sort(l),(void 0===o?d:d.slice(0,o)).map(u)}function r(e,n,t,r,i,o){var s=e.getMatchForLastSegmentOfPattern(n);if(s)for(var l=0,u=t;l<u.length;l++){var d=u[l];if(a(d,r))if(e.patternContainsDots){var p=e.getFullMatch(c(d),n);p&&o.push({name:n,fileName:i,matchKind:p.kind,isCaseSensitive:p.isCaseSensitive,declaration:d})}else o.push({name:n,fileName:i,matchKind:s.kind,isCaseSensitive:s.isCaseSensitive,declaration:d})}}function a(e,n){switch(e.kind){case 250:case 253:case 248:var t=n.getSymbolAtLocation(e.name),r=n.getAliasedSymbol(t);return t.escapedName!==r.escapedName;default:return!0}}function i(n,t){var r=e.getNameOfDeclaration(n);return!!r&&(s(r,t)||149===r.kind&&o(r.expression,t))}function o(n,t){return s(n,t)||e.isPropertyAccessExpression(n)&&(t.push(n.name.text),!0)&&o(n.expression,t)}function s(n,t){return e.isPropertyNameLiteral(n)&&(t.push(e.getTextOfIdentifierOrLiteral(n)),!0)}function c(n){var t=[],r=e.getNameOfDeclaration(n);if(r&&149===r.kind&&!o(r.expression,t))return e.emptyArray;t.shift();var a=e.getContainerNode(n);while(a){if(!i(a,t))return e.emptyArray;a=e.getContainerNode(a)}return t.reverse()}function l(n,t){return e.compareValues(n.matchKind,t.matchKind)||e.compareStringsCaseSensitiveUI(n.name,t.name)}function u(n){var t=n.declaration,r=e.getContainerNode(t),a=r&&e.getNameOfDeclaration(r);return{name:n.name,kind:e.getNodeKind(t),kindModifiers:e.getNodeModifiers(t),matchKind:e.PatternMatchKind[n.matchKind],isCaseSensitive:n.isCaseSensitive,fileName:n.fileName,textSpan:e.createTextSpanFromNode(t),containerName:a?a.text:"",containerKind:a?e.getNodeKind(r):""}}n.getNavigateToItems=t})(e.NavigateTo||(e.NavigateTo={}))}(u||(u={})),function(e){(function(n){var t,r,a,i=/\s+/g,o=[],s=[];function c(n,a){t=a,r=n;try{return e.map(I(f(n)),M)}finally{u()}}function l(e,n){t=n,r=e;try{return w(f(e))}finally{u()}}function u(){r=void 0,t=void 0,o=[],a=void 0,s=[]}function d(e){return e.getText(r)}function p(e){return e.node.kind}function m(e,n){e.children?e.children.push(n):e.children=[n]}function f(n){e.Debug.assert(!o.length);var t={node:n,name:void 0,additionalNodes:void 0,parent:void 0,children:void 0,indent:0};a=t;for(var r=0,i=n.statements;r<i.length;r++){var s=i[r];b(s)}return y(),e.Debug.assert(!a&&!o.length),t}function _(e){m(a,g(e))}function g(n){return{node:n,name:e.isDeclaration(n)||e.isExpression(n)?e.getNameOfDeclaration(n):void 0,additionalNodes:void 0,parent:a,children:void 0,indent:a.indent+1}}function v(e){var n=g(e);m(a,n),o.push(a),a=n}function y(){a.children&&(E(a.children,a),D(a.children)),a=o.pop()}function h(e,n){v(e),b(n),y()}function b(n){if(t.throwIfCancellationRequested(),n&&!e.isToken(n))switch(n.kind){case 157:var r=n;h(r,r.body);for(var a=0,i=r.parameters;a<i.length;a++){var o=i[a];e.isParameterPropertyDeclaration(o)&&_(o)}break;case 156:case 158:case 159:case 155:e.hasDynamicName(n)||h(n,n.body);break;case 154:case 153:e.hasDynamicName(n)||_(n);break;case 250:var s=n;s.name&&_(s.name);var c=s.namedBindings;if(c)if(251===c.kind)_(c);else for(var l=0,u=c.elements;l<u.length;l++){var d=u[l];_(d)}break;case 186:case 237:var p=n,m=p.name,f=p.initializer;e.isBindingPattern(m)?b(m):f&&j(f)?f.name?b(f):(v(n),e.forEachChild(f,b),y()):h(n,f);break;case 197:case 239:case 196:h(n,n.body);break;case 243:v(n);for(var g=0,E=n.members;g<E.length;g++){var T=E[g];F(T)||_(T)}y();break;case 240:case 209:case 241:v(n);for(var S=0,x=n.members;S<x.length;S++){T=x[S];b(T)}y();break;case 244:h(n,R(n).body);break;case 257:case 248:case 162:case 160:case 161:case 242:_(n);break;case 204:var C=e.getAssignmentDeclarationKind(n);switch(C){case 1:case 2:case 3:case 6:return void h(n,n.right);case 4:case 5:case 0:case 7:case 8:case 9:break;default:e.Debug.assertNever(C)}default:e.hasJSDocNodes(n)&&e.forEach(n.jsDoc,function(n){e.forEach(n.tags,function(n){e.isJSDocTypeAlias(n)&&_(n)})}),e.forEachChild(n,b)}}function E(n,t){var r=e.createMap();e.filterMutate(n,function(n){var a=e.getNameOfDeclaration(n.node),i=a&&d(a);if(!i)return!0;var o=r.get(i);if(!o)return r.set(i,n),!0;if(o instanceof Array){for(var s=0,c=o;s<c.length;s++){var l=c[s];if(T(l,n,t))return!1}return o.push(n),!0}l=o;return!T(l,n,t)&&(r.set(i,[l,n]),!0)})}function T(e,n,t){return!!S(e.node,n.node,t)&&(A(e,n),!0)}function S(n,t,r){if(n.kind!==t.kind||n.parent!==t.parent&&(!x(n,r)||!x(t,r)))return!1;switch(n.kind){case 154:case 156:case 158:case 159:return e.hasModifier(n,32)===e.hasModifier(t,32);case 244:return C(n,t);default:return!0}}function x(n,t){var r=e.isModuleBlock(n.parent)?n.parent.parent:n.parent;return r===t.node||e.contains(t.additionalNodes,r)}function C(e,n){return e.body.kind===n.body.kind&&(244!==e.body.kind||C(e.body,n.body))}function A(n,t){var r;n.additionalNodes=n.additionalNodes||[],n.additionalNodes.push(t.node),t.additionalNodes&&(r=n.additionalNodes).push.apply(r,t.additionalNodes),n.children=e.concatenate(n.children,t.children),n.children&&(E(n.children,n),D(n.children))}function D(e){e.sort(L)}function L(n,t){return e.compareStringsCaseSensitiveUI(k(n.node),k(t.node))||e.compareValues(p(n),p(t))}function k(n){if(244===n.kind)return P(n);var t=e.getNameOfDeclaration(n);if(t&&e.isPropertyName(t))return e.unescapeLeadingUnderscores(e.getPropertyNameForPropertyNameNode(t));switch(n.kind){case 196:case 197:case 209:return B(n);default:return}}function N(n,t){if(244===n.kind)return P(n);if(t){var r=d(t);if(r.length>0)return r}switch(n.kind){case 279:var a=n;return e.isExternalModule(a)?'"'+e.escapeString(e.getBaseFileName(e.removeFileExtension(e.normalizePath(a.fileName))))+'"':"<global>";case 197:case 239:case 196:case 240:case 209:return 512&e.getModifierFlags(n)?"default":B(n);case 157:return"constructor";case 161:return"new()";case 160:return"()";case 162:return"[]";default:return"<unknown>"}}function I(n){var t=[];function r(e){if(a(e)&&(t.push(e),e.children))for(var n=0,i=e.children;n<i.length;n++){var o=i[n];r(o)}}return r(n),t;function a(n){switch(p(n)){case 240:case 209:case 243:case 241:case 244:case 279:case 242:case 304:case 297:return!0;case 157:case 156:case 158:case 159:case 237:return r(n);case 197:case 239:case 196:return t(n);default:return!1}function t(e){if(!e.node.body)return!1;switch(p(e.parent)){case 245:case 279:case 156:case 157:return!0;default:return r(e)}}function r(n){return e.some(n.children,function(e){var n=p(e);return 237!==n&&186!==n})}}}function w(n){return{text:N(n.node,n.name),kind:e.getNodeKind(n.node),kindModifiers:V(n.node),spans:O(n),nameSpan:n.name&&G(n.name),childItems:e.map(n.children,w)}}function M(n){return{text:N(n.node,n.name),kind:e.getNodeKind(n.node),kindModifiers:V(n.node),spans:O(n),childItems:e.map(n.children,t)||s,indent:n.indent,bolded:!1,grayed:!1};function t(n){return{text:N(n.node,n.name),kind:e.getNodeKind(n.node),kindModifiers:e.getNodeModifiers(n.node),spans:O(n),childItems:s,indent:0,bolded:!1,grayed:!1}}}function O(e){var n=[G(e.node)];if(e.additionalNodes)for(var t=0,r=e.additionalNodes;t<r.length;t++){var a=r[t];n.push(G(a))}return n}function P(n){if(e.isAmbientModule(n))return e.getTextOfNode(n.name);var t=[];t.push(e.getTextOfIdentifierOrLiteral(n.name));while(n.body&&244===n.body.kind)n=n.body,t.push(e.getTextOfIdentifierOrLiteral(n.name));return t.join(".")}function R(n){return n.body&&e.isModuleDeclaration(n.body)?R(n.body):n}function F(e){return!e.name||149===e.name.kind}function G(n){return 279===n.kind?e.createTextSpanFromRange(n):e.createTextSpanFromNode(n,r)}function V(n){return n.parent&&237===n.parent.kind&&(n=n.parent),e.getNodeModifiers(n)}function B(n){var t=n.parent;if(n.name&&e.getFullWidth(n.name)>0)return e.declarationNameToString(n.name);if(e.isVariableDeclaration(t))return e.declarationNameToString(t.name);if(e.isBinaryExpression(t)&&59===t.operatorToken.kind)return d(t.left).replace(i,"");if(e.isPropertyAssignment(t))return d(t.name);if(512&e.getModifierFlags(n))return"default";if(e.isClassLike(n))return"<class>";if(e.isCallExpression(t)){var a=K(t.expression);if(void 0!==a){var o=e.mapDefined(t.arguments,function(n){return e.isStringLiteralLike(n)?n.getText(r):void 0}).join(", ");return a+"("+o+") callback"}}return"<function>"}function K(n){if(e.isIdentifier(n))return n.text;if(e.isPropertyAccessExpression(n)){var t=K(n.expression),r=n.name.text;return void 0===t?r:t+"."+r}}function j(e){switch(e.kind){case 197:case 196:case 209:return!0;default:return!1}}n.getNavigationBarItems=c,n.getNavigationTree=l})(e.NavigationBar||(e.NavigationBar={}))}(u||(u={})),function(e){(function(n){function t(n,t,s,c,u){var d=e.textChanges.ChangeTracker.fromContext({host:s,formatContext:t}),p=function(e){return i(r(e,n,c))},m=n.statements.filter(e.isImportDeclaration);b(m,p);var f=n.statements.filter(e.isExportDeclaration);b(f,o);for(var _=0,g=n.statements.filter(e.isAmbientModule);_<g.length;_++){var v=g[_];if(v.body){var y=v.body.statements.filter(e.isImportDeclaration);b(y,p);var h=v.body.statements.filter(e.isExportDeclaration);b(h,o)}}return d.getChanges();function b(r,i){if(0!==e.length(r)){e.suppressLeadingTrivia(r[0]);var o=e.group(r,function(e){return a(e.moduleSpecifier)}),c=e.stableSort(o,function(e,n){return l(e[0].moduleSpecifier,n[0].moduleSpecifier)}),u=e.flatMap(c,function(e){return a(e[0].moduleSpecifier)?i(e):e});0===u.length?d.delete(n,r[0]):d.replaceNodeWithNodes(n,r[0],u,{useNonAdjustedStartPosition:!0,useNonAdjustedEndPosition:!1,suffix:e.getNewLineOrDefaultFromHost(s,t.options)});for(var p=1;p<r.length;p++)d.delete(n,r[p])}}}function r(n,t,r){for(var a=r.getTypeChecker(),i=a.getJsxNamespace(t),o=!!(4&t.transformFlags),c=[],l=0,u=n;l<u.length;l++){var d=u[l],p=d.importClause;if(p){var m=p.name,f=p.namedBindings;if(m&&!g(m)&&(m=void 0),f)if(e.isNamespaceImport(f))g(f.name)||(f=void 0);else{var _=f.elements.filter(function(e){return g(e.name)});_.length<f.elements.length&&(f=_.length?e.updateNamedImports(f,_):void 0)}(m||f)&&c.push(s(d,m,f))}else c.push(d)}return c;function g(n){return o&&n.text===i||e.FindAllReferences.Core.isSymbolReferencedInFile(n,a,t)}}function a(n){return void 0!==n&&e.isStringLiteralLike(n)?n.text:void 0}function i(n){if(0===n.length)return n;var t=S(n),r=t.importWithoutClause,a=t.defaultImports,i=t.namespaceImports,o=t.namedImports,l=[];if(r&&l.push(r),1===a.length&&1===i.length&&0===o.length){var d=a[0];return l.push(s(d,d.importClause.name,i[0].importClause.namedBindings)),l}for(var p,m=e.stableSort(i,function(e,n){return u(e.importClause.namedBindings.name,n.importClause.namedBindings.name)}),f=0,_=m;f<_.length;f++){var g=_[f];l.push(s(g,void 0,g.importClause.namedBindings))}if(0===a.length&&0===o.length)return l;var v=[];if(1===a.length)p=a[0].importClause.name;else for(var y=0,h=a;y<h.length;y++){d=h[y];v.push(e.createImportSpecifier(e.createIdentifier("default"),d.importClause.name))}v.push.apply(v,e.flatMap(o,function(e){return e.importClause.namedBindings.elements}));var b=c(v),E=a.length>0?a[0]:o[0],T=0===b.length?p?void 0:e.createNamedImports(e.emptyArray):0===o.length?e.createNamedImports(b):e.updateNamedImports(o[0].importClause.namedBindings,b);return l.push(s(E,p,T)),l;function S(n){for(var t,r=[],a=[],i=[],o=0,s=n;o<s.length;o++){var c=s[o];if(void 0!==c.importClause){var l=c.importClause,u=l.name,d=l.namedBindings;u&&r.push(c),d&&(e.isNamespaceImport(d)?a.push(c):i.push(c))}else t=t||c}return{importWithoutClause:t,defaultImports:r,namespaceImports:a,namedImports:i}}}function o(n){if(0===n.length)return n;var t=u(n),r=t.exportWithoutClause,a=t.namedExports,i=[];if(r&&i.push(r),0===a.length)return i;var o=[];o.push.apply(o,e.flatMap(a,function(e){return e.exportClause.elements}));var s=c(o),l=a[0];return i.push(e.updateExportDeclaration(l,l.decorators,l.modifiers,e.updateNamedExports(l.exportClause,s),l.moduleSpecifier)),i;function u(e){for(var n,t=[],r=0,a=e;r<a.length;r++){var i=a[r];void 0===i.exportClause?n=n||i:t.push(i)}return{exportWithoutClause:n,namedExports:t}}}function s(n,t,r){return e.updateImportDeclaration(n,n.decorators,n.modifiers,e.updateImportClause(n.importClause,t,r),n.moduleSpecifier)}function c(n){return e.stableSort(n,function(e,n){return u(e.propertyName||e.name,n.propertyName||n.name)||u(e.name,n.name)})}function l(n,t){var r=a(n),i=a(t);return e.compareBooleans(void 0===r,void 0===i)||e.compareBooleans(e.isExternalModuleNameRelative(r),e.isExternalModuleNameRelative(i))||e.compareStringsCaseInsensitive(r,i)}function u(n,t){return e.compareStringsCaseInsensitive(n.text,t.text)}n.organizeImports=t,n.coalesceImports=i,n.coalesceExports=o,n.compareModuleSpecifiers=l})(e.OrganizeImports||(e.OrganizeImports={}))}(u||(u={})),function(e){(function(n){function t(e,n){var t=[];return r(e,n,t),a(e,t),t.sort(function(e,n){return e.textSpan.start-n.textSpan.start})}function r(n,t,r){var a=40,i=0,o=n.statements,u=o.length;while(i<u){while(i<u&&!e.isAnyImportSyntax(o[i]))m(o[i]),i++;if(i===u)break;var d=i;while(i<u&&e.isAnyImportSyntax(o[i]))s(o[i],n,t,r),i++;var p=i-1;p!==d&&r.push(c(e.findChildOfKind(o[d],92,n).getStart(n),o[p].getEnd(),"imports"))}function m(i){if(0!==a){t.throwIfCancellationRequested(),e.isDeclaration(i)&&s(i,n,t,r),f(i)&&s(i.parent.parent.parent,n,t,r);var o=l(i,n);o&&r.push(o),a--,e.isIfStatement(i)&&i.elseStatement&&e.isIfStatement(i.elseStatement)?(m(i.expression),m(i.thenStatement),a++,m(i.elseStatement),a--):i.forEachChild(m),a++}}function f(n){if(!e.isFunctionExpression(n)&&!e.isArrowFunction(n))return!1;var t=e.findAncestor(n,e.isVariableStatement);return!!t&&e.getSingleInitializerOfVariableStatementOrPropertyDeclaration(t)===n}}function a(n,t){for(var r=[],a=n.getLineStarts(),i=0;i<a.length;i++){var s=a[i],c=i+1===a.length?n.getEnd():a[i+1]-1,l=n.text.substring(s,c),d=o(l);if(d&&!e.isInComment(n,s))if(d[1]){var p=r.pop();p&&(p.textSpan.length=c-p.textSpan.start,p.hintSpan.length=c-p.textSpan.start,t.push(p))}else{var m=e.createTextSpanFromBounds(n.text.indexOf("//",s),c);r.push(u(m,"region",m,!1,d[2]||"#region"))}}}n.collectElements=t;var i=/^\s*\/\/\s*#(end)?region(?:\s+(.*))?(?:\r)?$/;function o(e){return i.exec(e)}function s(n,t,r,a){var i=e.getLeadingCommentRangesOfNode(n,t);if(i){for(var s=-1,l=-1,u=0,d=t.getFullText(),p=0,m=i;p<m.length;p++){var f=m[p],_=f.kind,g=f.pos,v=f.end;switch(r.throwIfCancellationRequested(),_){case 2:var y=d.slice(g,v);if(o(y)){h(),u=0;break}0===u&&(s=g),l=v,u++;break;case 3:h(),a.push(c(g,v,"comment")),u=0;break;default:e.Debug.assertNever(_)}}h()}function h(){u>1&&a.push(c(s,l,"comment"))}}function c(n,t,r){return u(e.createTextSpanFromBounds(n,t),r)}function l(n,t){switch(n.kind){case 218:if(e.isFunctionBlock(n))return s(n.parent,197!==n.parent.kind);switch(n.parent.kind){case 223:case 226:case 227:case 225:case 222:case 224:case 231:case 274:return s(n.parent);case 235:var r=n.parent;if(r.tryBlock===n)return s(n.parent);if(r.finallyBlock===n)return s(e.findChildOfKind(r,88,t));default:return u(e.createTextSpanFromNode(n,t),"code")}case 245:return s(n.parent);case 240:case 209:case 241:case 243:case 246:return s(n);case 188:return o(n);case 187:return o(n,22);case 260:return a(n);case 261:case 262:return i(n.attributes)}function a(n){var r=e.createTextSpanFromBounds(n.openingElement.getStart(t),n.closingElement.getEnd()),a=n.openingElement.tagName.getText(t),i="<"+a+">...</"+a+">";return u(r,"code",r,!1,i)}function i(e){if(0!==e.properties.length)return c(e.getStart(t),e.getEnd(),"code")}function o(n,t){return void 0===t&&(t=18),s(n,!1,!e.isArrayLiteralExpression(n.parent)&&!e.isCallExpression(n.parent),t)}function s(r,a,i,o){void 0===a&&(a=!1),void 0===i&&(i=!0),void 0===o&&(o=18);var s=e.findChildOfKind(n,o,t),c=18===o?19:23,l=e.findChildOfKind(n,c,t);if(s&&l){var d=e.createTextSpanFromBounds(i?s.getFullStart():s.getStart(t),l.getEnd());return u(d,"code",e.createTextSpanFromNode(r,t),a)}}}function u(e,n,t,r,a){return void 0===t&&(t=e),void 0===r&&(r=!1),void 0===a&&(a="..."),{textSpan:e,kind:n,hintSpan:t,bannerText:a,autoCollapse:r}}})(e.OutliningElementsCollector||(e.OutliningElementsCollector={}))}(u||(u={})),function(e){var n;function t(e,n){return{kind:e,isCaseSensitive:n}}function r(n){var t=e.createMap(),r=n.trim().split(".").map(function(e){return m(e.trim())});if(!r.some(function(e){return!e.subWordTextChunks.length}))return{getFullMatch:function(e,n){return a(e,n,r,t)},getMatchForLastSegmentOfPattern:function(n){return s(n,e.last(r),t)},patternContainsDots:r.length>1}}function a(n,t,r,a){var i=s(t,e.last(r),a);if(i&&!(r.length-1>n.length)){for(var o,l=r.length-2,u=n.length-1;l>=0;l-=1,u-=1)o=c(o,s(n[u],r[l],a));return o}}function i(e,n){var t=n.get(e);return t||n.set(e,t=S(e)),t}function o(r,a,o){var s=g(r,a.textLowerCase);if(0===s)return t(a.text.length===r.length?n.exact:n.prefix,e.startsWith(r,a.text));if(a.isLowerCase){if(-1===s)return;for(var c=i(r,o),l=0,d=c;l<d.length;l++){var m=d[l];if(u(r,m,a.text,!0))return t(n.substring,u(r,m,a.text,!1))}if(a.text.length<r.length&&f(r.charCodeAt(s)))return t(n.substring,!1)}else{if(r.indexOf(a.text)>0)return t(n.substring,!0);if(a.characterSpans.length>0){var _=i(r,o),v=!!p(r,_,a,!1)||!p(r,_,a,!0)&&void 0;if(void 0!==v)return t(n.camelCase,v)}}}function s(e,n,t){if(N(n.totalTextChunk.text,function(e){return 32!==e&&42!==e})){var r=o(e,n.totalTextChunk,t);if(r)return r}for(var a,i=n.subWordTextChunks,s=0,l=i;s<l.length;s++){var u=l[s];a=c(a,o(e,u,t))}return a}function c(n,t){return e.min(n,t,l)}function l(n,t){return void 0===n?1:void 0===t?-1:e.compareValues(n.kind,t.kind)||e.compareBooleans(!n.isCaseSensitive,!t.isCaseSensitive)}function u(e,n,t,r,a){return void 0===a&&(a={start:0,length:t.length}),a.length<=n.length&&k(0,a.length,function(i){return d(t.charCodeAt(a.start+i),e.charCodeAt(n.start+i),r)})}function d(e,n,t){return t?v(e)===v(n):e===n}function p(n,t,r,a){var i,o,s=r.characterSpans,c=0,l=0;while(1){if(l===s.length)return!0;if(c===t.length)return!1;for(var d=t[c],p=!1;l<s.length;l++){var m=s[l];if(p&&(!f(r.text.charCodeAt(s[l-1].start))||!f(r.text.charCodeAt(s[l].start))))break;if(!u(n,d,r.text,a,m))break;p=!0,i=void 0===i?c:i,o=void 0===o||o,d=e.createTextSpan(d.start+m.length,d.length-m.length)}p||void 0===o||(o=!1),c++}}function m(e){return{totalTextChunk:E(e),subWordTextChunks:b(e)}}function f(n){if(n>=65&&n<=90)return!0;if(n<127||!e.isUnicodeIdentifierStart(n,6))return!1;var t=String.fromCharCode(n);return t===t.toUpperCase()}function _(n){if(n>=97&&n<=122)return!0;if(n<127||!e.isUnicodeIdentifierStart(n,6))return!1;var t=String.fromCharCode(n);return t===t.toLowerCase()}function g(e,n){for(var t=e.length-n.length,r=function(t){if(N(n,function(n,r){return v(e.charCodeAt(r+t))===n}))return{value:t}},a=0;a<=t;a++){var i=r(a);if("object"===typeof i)return i.value}return-1}function v(e){return e>=65&&e<=90?e-65+97:e<127?e:String.fromCharCode(e).toLowerCase().charCodeAt(0)}function y(e){return e>=48&&e<=57}function h(e){return f(e)||_(e)||y(e)||95===e||36===e}function b(e){for(var n=[],t=0,r=0,a=0;a<e.length;a++){var i=e.charCodeAt(a);h(i)?(0===r&&(t=a),r++):r>0&&(n.push(E(e.substr(t,r))),r=0)}return r>0&&n.push(E(e.substr(t,r))),n}function E(e){var n=e.toLowerCase();return{text:e,textLowerCase:n,isLowerCase:e===n,characterSpans:T(e)}}function T(e){return x(e,!1)}function S(e){return x(e,!0)}function x(n,t){for(var r=[],a=0,i=1;i<n.length;i++){var o=y(n.charCodeAt(i-1)),s=y(n.charCodeAt(i)),c=L(n,t,i),l=t&&D(n,i,a);(C(n.charCodeAt(i-1))||C(n.charCodeAt(i))||o!==s||c||l)&&(A(n,a,i)||r.push(e.createTextSpan(a,i-a)),a=i)}return A(n,a,n.length)||r.push(e.createTextSpan(a,n.length-a)),r}function C(e){switch(e){case 33:case 34:case 35:case 37:case 38:case 39:case 40:case 41:case 42:case 44:case 45:case 46:case 47:case 58:case 59:case 63:case 64:case 91:case 92:case 93:case 95:case 123:case 125:return!0}return!1}function A(e,n,t){return N(e,function(e){return C(e)&&95!==e},n,t)}function D(e,n,t){return n!==t&&n+1<e.length&&f(e.charCodeAt(n))&&_(e.charCodeAt(n+1))&&N(e,f,t,n)}function L(e,n,t){var r=f(e.charCodeAt(t-1)),a=f(e.charCodeAt(t));return a&&(!n||!r)}function k(e,n,t){for(var r=e;r<n;r++)if(!t(r))return!1;return!0}function N(e,n,t,r){return void 0===t&&(t=0),void 0===r&&(r=e.length),k(t,r,function(t){return n(e.charCodeAt(t),t)})}(function(e){e[e["exact"]=0]="exact",e[e["prefix"]=1]="prefix",e[e["substring"]=2]="substring",e[e["camelCase"]=3]="camelCase"})(n=e.PatternMatchKind||(e.PatternMatchKind={})),e.createPatternMatcher=r,e.breakIntoCharacterSpans=T,e.breakIntoWordSpans=S}(u||(u={})),function(e){function n(n,t,r){void 0===t&&(t=!0),void 0===r&&(r=!1);var a,i,o,s={languageVersion:1,pragmas:void 0,checkJsDirective:void 0,referencedFiles:[],typeReferenceDirectives:[],libReferenceDirectives:[],amdDependencies:[],hasNoDefaultLib:void 0,moduleName:void 0},c=[],l=0,u=!1;function d(){return i=o,o=e.scanner.scan(),18===o?l++:19===o&&l--,o}function p(){var n=e.scanner.getTokenValue(),t=e.scanner.getTokenPos();return{fileName:n,pos:t,end:t+n.length}}function m(){a||(a=[]),a.push({ref:p(),depth:l})}function f(){c.push(p()),_()}function _(){0===l&&(u=!0)}function g(){var n=e.scanner.getToken();return 125===n&&(n=d(),130===n&&(n=d(),10===n&&m()),!0)}function v(){if(24===i)return!1;var n=e.scanner.getToken();if(92===n){if(n=d(),20===n){if(n=d(),10===n)return f(),!0}else{if(10===n)return f(),!0;if(72===n||e.isKeyword(n))if(n=d(),144===n){if(n=d(),10===n)return f(),!0}else if(59===n){if(h(!0))return!0}else{if(27!==n)return!0;n=d()}if(18===n){n=d();while(19!==n&&1!==n)n=d();19===n&&(n=d(),144===n&&(n=d(),10===n&&f()))}else 40===n&&(n=d(),119===n&&(n=d(),(72===n||e.isKeyword(n))&&(n=d(),144===n&&(n=d(),10===n&&f()))))}return!0}return!1}function y(){var n=e.scanner.getToken();if(85===n){if(_(),n=d(),18===n){n=d();while(19!==n&&1!==n)n=d();19===n&&(n=d(),144===n&&(n=d(),10===n&&f()))}else if(40===n)n=d(),144===n&&(n=d(),10===n&&f());else if(92===n&&(n=d(),(72===n||e.isKeyword(n))&&(n=d(),59===n&&h(!0))))return!0;return!0}return!1}function h(n){var t=n?d():e.scanner.getToken();return 134===t&&(t=d(),20===t&&(t=d(),10===t&&f()),!0)}function b(){var n=e.scanner.getToken();if(72===n&&"define"===e.scanner.getTokenValue()){if(n=d(),20!==n)return!0;if(n=d(),10===n){if(n=d(),27!==n)return!0;n=d()}if(22!==n)return!0;n=d();while(23!==n&&1!==n)10===n&&f(),n=d();return!0}return!1}function E(){e.scanner.setText(n),d();while(1){if(1===e.scanner.getToken())break;g()||v()||y()||r&&(h(!1)||b())||d()}e.scanner.setText(void 0)}if(t&&E(),e.processCommentPragmas(s,n),e.processPragmasIntoFields(s,e.noop),u){if(a)for(var T=0,S=a;T<S.length;T++){var x=S[T];c.push(x.ref)}return{referencedFiles:s.referencedFiles,typeReferenceDirectives:s.typeReferenceDirectives,libReferenceDirectives:s.libReferenceDirectives,importedFiles:c,isLibFile:!!s.hasNoDefaultLib,ambientExternalModules:void 0}}var C=void 0;if(a)for(var A=0,D=a;A<D.length;A++){x=D[A];0===x.depth?(C||(C=[]),C.push(x.ref.fileName)):c.push(x.ref)}return{referencedFiles:s.referencedFiles,typeReferenceDirectives:s.typeReferenceDirectives,libReferenceDirectives:s.libReferenceDirectives,importedFiles:c,isLibFile:!!s.hasNoDefaultLib,ambientExternalModules:C}}e.preProcessFile=n}(u||(u={})),function(e){(function(n){function t(n,t,a,i){var s=e.getTouchingPropertyName(t,a),l=s&&c(s)?r(s,n.getTypeChecker(),t,function(e){return n.isSourceFileDefaultLibrary(e.getSourceFile())},i):void 0;return l||o(e.Diagnostics.You_cannot_rename_this_element)}function r(n,t,r,s,c){var l=t.getSymbolAtLocation(n);if(l){var u=l.declarations;if(u&&0!==u.length){if(u.some(s))return o(e.Diagnostics.You_cannot_rename_elements_that_are_defined_in_the_standard_TypeScript_library);if(!(e.isIdentifier(n)&&80===n.originalKeywordKind&&1536&l.parent.flags)){if(e.isStringLiteralLike(n)&&e.tryGetImportFromModuleSpecifier(n))return c&&c.allowRenameOfImportPath?a(n,r,l):void 0;var d=e.SymbolDisplay.getSymbolKind(t,l,n),p=e.isImportOrExportSpecifierName(n)||e.isStringOrNumericLiteralLike(n)&&149===n.parent.kind?e.stripQuotes(e.getTextOfIdentifierOrLiteral(n)):void 0,m=p||t.symbolToString(l),f=p||t.getFullyQualifiedName(l);return i(m,f,d,e.SymbolDisplay.getSymbolModifiers(l),n,r)}}}}function a(n,t,r){if(!e.isExternalModuleNameRelative(n.text))return o(e.Diagnostics.You_cannot_rename_a_module_via_a_global_import);var a=e.find(r.declarations,e.isSourceFile);if(a){var i=e.endsWith(n.text,"/index")||e.endsWith(n.text,"/index.js")?void 0:e.tryRemoveSuffix(e.removeFileExtension(a.fileName),"/index"),s=void 0===i?a.fileName:i,c=void 0===i?"module":"directory",l=n.text.lastIndexOf("/")+1,u=e.createTextSpan(n.getStart(t)+1+l,n.text.length-l);return{canRename:!0,fileToRename:s,kind:c,displayName:s,fullDisplayName:s,kindModifiers:"",triggerSpan:u}}}function i(e,n,t,r,a,i){return{canRename:!0,fileToRename:void 0,kind:t,displayName:e,fullDisplayName:n,kindModifiers:r,triggerSpan:s(a,i)}}function o(n){return{canRename:!1,localizedErrorMessage:e.getLocaleSpecificMessage(n)}}function s(n,t){var r=n.getStart(t),a=n.getWidth(t);return 10===n.kind&&(r+=1,a-=2),e.createTextSpan(r,a)}function c(n){switch(n.kind){case 72:case 10:case 100:return!0;case 8:return e.isLiteralNameOfPropertyDeclarationOrIndexAccess(n);default:return!1}}n.getRenameInfo=t})(e.Rename||(e.Rename={}))}(u||(u={})),function(e){(function(n){var t,r;function a(n,t,r,a,o){var c=n.getTypeChecker(),l=e.findTokenOnLeftOfPosition(t,r);if(l){var u=!!a&&"characterTyped"===a.kind;if(!u||!e.isInString(t,r,l)&&!e.isInComment(t,r)){var d=!!a&&"invoked"===a.kind,p=C(l,r,t,c,d);if(p){o.throwIfCancellationRequested();var m=i(p,c,t,l,u);return o.throwIfCancellationRequested(),m?c.runWithCancellationToken(o,function(e){return 0===m.kind?N(m.candidates,m.resolvedSignature,p,t,e):I(m.symbol,p,t,e)}):e.isSourceFileJS(t)?s(p,n,o):void 0}}}}function i(n,t,r,a,i){var s=n.invocation,l=n.argumentCount;switch(s.kind){case 0:if(i&&!o(a,s.node,r))return;var u=[],d=t.getResolvedSignatureForSignatureHelp(s.node,u,l);return 0===u.length?void 0:{kind:0,candidates:u,resolvedSignature:d};case 1:var p=s.called;if(i&&!c(a,r,e.isIdentifier(p)?p.parent:p))return;u=e.getPossibleGenericSignatures(p,l,t);if(0!==u.length)return{kind:0,candidates:u,resolvedSignature:e.first(u)};var m=t.getSymbolAtLocation(p);return m&&{kind:1,symbol:m};case 2:return{kind:0,candidates:[s.signature],resolvedSignature:s.signature};default:return e.Debug.assertNever(s)}}function o(n,t,r){if(!e.isCallOrNewExpression(t))return!1;var a=t.getChildren(r);switch(n.kind){case 20:return e.contains(a,n);case 27:var i=e.findContainingList(n);return!!i&&e.contains(a,i);case 28:return c(n,r,t.expression);default:return!1}}function s(n,t,r){if(2!==n.invocation.kind){var a=D(n.invocation),i=e.isIdentifier(a)?a.text:e.isPropertyAccessExpression(a)?a.name.text:void 0,o=t.getTypeChecker();return void 0===i?void 0:e.firstDefined(t.getSourceFiles(),function(t){return e.firstDefined(t.getNamedDeclarations().get(i),function(e){var a=e.symbol&&o.getTypeOfSymbolAtLocation(e.symbol,e),i=a&&a.getCallSignatures();if(i&&i.length)return o.runWithCancellationToken(r,function(e){return N(i,i[0],n,t,e)})})})}}function c(n,t,r){var a=e.Debug.assertDefined(e.findPrecedingToken(n.getFullStart(),t,n.parent,!0));return e.rangeContainsRange(r,a)}function l(e,n,t){var r=p(e,n,t);return!r||r.isTypeParameterList||0!==r.invocation.kind?void 0:{invocation:r.invocation.node,argumentCount:r.argumentCount,argumentIndex:r.argumentIndex}}function u(n,t){var r=d(n,t);if(r){var a=r.list,i=r.argumentIndex,o=b(a);0!==i&&e.Debug.assertLessThan(i,o);var s=S(a,t);return{list:a,argumentIndex:i,argumentCount:o,argumentsSpan:s}}}function d(n,t){if(28===n.kind||20===n.kind)return{list:A(n.parent,n,t),argumentIndex:0};var r=e.findContainingList(n);return r&&{list:r,argumentIndex:h(r,n)}}function p(n,t,r){var a=n.parent;if(e.isCallOrNewExpression(a)){var i=a,o=u(n,r);if(!o)return;var s=o.list,c=o.argumentIndex,l=o.argumentCount,d=o.argumentsSpan,p=!!a.typeArguments&&a.typeArguments.pos===s.pos;return{isTypeParameterList:p,invocation:{kind:0,node:i},argumentsSpan:d,argumentIndex:c,argumentCount:l}}if(e.isNoSubstitutionTemplateLiteral(n)&&e.isTaggedTemplateExpression(a))return e.isInsideTemplateLiteral(n,t,r)?T(a,0,r):void 0;if(e.isTemplateHead(n)&&193===a.parent.kind){var m=a,f=m.parent;e.Debug.assert(206===m.kind);c=e.isInsideTemplateLiteral(n,t,r)?0:1;return T(f,c,r)}if(e.isTemplateSpan(a)&&e.isTaggedTemplateExpression(a.parent.parent)){var _=a;f=a.parent.parent;if(e.isTemplateTail(n)&&!e.isInsideTemplateLiteral(n,t,r))return;var g=_.parent.templateSpans.indexOf(_);c=E(g,n,t,r);return T(f,c,r)}if(e.isJsxOpeningLikeElement(a)){var v=a.attributes.pos,y=e.skipTrivia(r.text,a.attributes.end,!1);return{isTypeParameterList:!1,invocation:{kind:0,node:a},argumentsSpan:e.createTextSpan(v,y-v),argumentIndex:0,argumentCount:1}}var h=e.getPossibleTypeArgumentsInfo(n,r);if(h){var b=h.called,S=h.nTypeArguments;i={kind:1,called:b},d=e.createTextSpanFromBounds(b.getStart(r),n.end);return{isTypeParameterList:!0,invocation:i,argumentsSpan:d,argumentIndex:S,argumentCount:S+1}}}function m(e,n,t,r){return g(e,n,t,r)||p(e,n,t)}function f(n){return e.isBinaryExpression(n.parent)?f(n.parent):n}function _(n){return e.isBinaryExpression(n.left)?_(n.left)+1:2}function g(n,t,r,a){var i=v(n,r,a);if(i){var o=i.contextualType,s=i.argumentIndex,c=i.argumentCount,l=i.argumentsSpan,u=o.getCallSignatures();if(1===u.length){var d={kind:2,signature:e.first(u),node:n,symbol:y(o.symbol)};return{isTypeParameterList:!1,invocation:d,argumentsSpan:l,argumentIndex:s,argumentCount:c}}}}function v(n,t,r){if(20===n.kind||27===n.kind){var a=n.parent;switch(a.kind){case 195:case 156:case 196:case 197:var i=u(n,t);if(!i)return;var o=i.argumentIndex,s=i.argumentCount,c=i.argumentsSpan,l=e.isMethodDeclaration(a)?r.getContextualTypeForObjectLiteralElement(a):r.getContextualType(a);return l&&{contextualType:l,argumentIndex:o,argumentCount:s,argumentsSpan:c};case 204:var d=f(a),p=r.getContextualType(d),m=20===n.kind?0:_(a)-1,g=_(d);return p&&{contextualType:p,argumentIndex:m,argumentCount:g,argumentsSpan:e.createTextSpanFromNode(a)};default:return}}}function y(n){return"__type"===n.name&&e.firstDefined(n.declarations,function(n){return e.isFunctionTypeNode(n)?n.parent.symbol:void 0})||n}function h(e,n){for(var t=0,r=0,a=e.getChildren();r<a.length;r++){var i=a[r];if(i===n)break;27!==i.kind&&t++}return t}function b(n){var t=n.getChildren(),r=e.countWhere(t,function(e){return 27!==e.kind});return t.length>0&&27===e.last(t).kind&&r++,r}function E(n,t,r,a){return e.Debug.assert(r>=t.getStart(),"Assumed 'position' could not occur before node."),e.isTemplateLiteralToken(t)?e.isInsideTemplateLiteral(t,r,a)?0:n+2:n+1}function T(n,t,r){var a=e.isNoSubstitutionTemplateLiteral(n.template)?1:n.template.templateSpans.length+1;return 0!==t&&e.Debug.assertLessThan(t,a),{isTypeParameterList:!1,invocation:{kind:0,node:n},argumentsSpan:x(n,r),argumentIndex:t,argumentCount:a}}function S(n,t){var r=n.getFullStart(),a=e.skipTrivia(t.text,n.getEnd(),!1);return e.createTextSpan(r,a-r)}function x(n,t){var r=n.template,a=r.getStart(),i=r.getEnd();if(206===r.kind){var o=e.last(r.templateSpans);0===o.literal.getFullWidth()&&(i=e.skipTrivia(t.text,i,!1))}return e.createTextSpan(a,i-a)}function C(n,t,r,a,i){for(var o=function(n){e.Debug.assert(e.rangeContainsRange(n.parent,n),"Not a subspan",function(){return"Child: "+e.Debug.showSyntaxKind(n)+", parent: "+e.Debug.showSyntaxKind(n.parent)});var i=m(n,t,r,a);if(i)return{value:i}},s=n;!e.isSourceFile(s)&&(i||!e.isBlock(s));s=s.parent){var c=o(s);if("object"===typeof c)return c.value}}function A(n,t,r){var a=n.getChildren(r),i=a.indexOf(t);return e.Debug.assert(i>=0&&a.length>i+1),a[i+1]}function D(n){return 0===n.kind?e.getInvokedExpression(n.node):n.called}function L(e){return 0===e.kind?e.node:1===e.kind?e.called:e.node}(function(e){e[e["Call"]=0]="Call",e[e["TypeArgs"]=1]="TypeArgs",e[e["Contextual"]=2]="Contextual"})(t||(t={})),n.getSignatureHelpItems=a,function(e){e[e["Candidate"]=0]="Candidate",e[e["Type"]=1]="Type"}(r||(r={})),n.getArgumentInfoForCompletions=l;var k=70246400;function N(n,t,r,a,i){var o=r.isTypeParameterList,s=r.argumentCount,c=r.argumentsSpan,l=r.invocation,u=r.argumentIndex,d=L(l),p=2===l.kind?l.symbol:i.getSymbolAtLocation(D(l)),m=p?e.symbolToDisplayParts(i,p,void 0,void 0):e.emptyArray,f=n.map(function(e){return O(e,m,o,i,d,a)});0!==u&&e.Debug.assertLessThan(u,s);var _=n.indexOf(t);return e.Debug.assert(-1!==_),{items:f,applicableSpan:c,selectedItemIndex:_,argumentIndex:u,argumentCount:s}}function I(e,n,t,r){var a=n.argumentCount,i=n.argumentsSpan,o=n.invocation,s=n.argumentIndex,c=r.getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(e);if(c){var l=[w(e,c,r,L(o),t)];return{items:l,applicableSpan:i,selectedItemIndex:0,argumentIndex:s,argumentCount:a}}}function w(n,t,r,a,i){var o=e.symbolToDisplayParts(r,n),s=e.createPrinter({removeComments:!0}),c=t.map(function(e){return V(e,r,a,i,s)}),l=n.getDocumentationComment(r),u=n.getJsDocTags(),d=o.concat([e.punctuationPart(28)]);return{isVariadic:!1,prefixDisplayParts:d,suffixDisplayParts:[e.punctuationPart(30)],separatorDisplayParts:M,parameters:c,documentation:l,tags:u}}var M=[e.punctuationPart(27),e.spacePart()];function O(e,n,t,r,a,i){var o=(t?R:F)(e,r,a,i),s=o.isVariadic,c=o.parameters,l=o.prefix,u=o.suffix,d=n.concat(l),p=u.concat(P(e,a,r)),m=e.getDocumentationComment(r),f=e.getJsDocTags();return{isVariadic:s,prefixDisplayParts:d,suffixDisplayParts:p,separatorDisplayParts:M,parameters:c,documentation:m,tags:f}}function P(n,t,r){return e.mapToDisplayParts(function(e){e.writePunctuation(":"),e.writeSpace(" ");var a=r.getTypePredicateOfSignature(n);a?r.writeTypePredicate(a,t,void 0,e):r.writeType(r.getReturnTypeOfSignature(n),t,void 0,e)})}function R(n,t,r,a){var i=(n.target||n).typeParameters,o=e.createPrinter({removeComments:!0}),s=(i||e.emptyArray).map(function(e){return V(e,t,r,a,o)}),c=e.mapToDisplayParts(function(i){var s=n.thisParameter?[t.symbolToParameterDeclaration(n.thisParameter,r,k)]:[],c=e.createNodeArray(s.concat(n.parameters.map(function(e){return t.symbolToParameterDeclaration(e,r,k)})));o.writeList(2576,c,a,i)});return{isVariadic:!1,parameters:s,prefix:[e.punctuationPart(28)],suffix:[e.punctuationPart(30)].concat(c)}}function F(n,t,r,a){var i=n.hasRestParameter,o=e.createPrinter({removeComments:!0}),s=e.mapToDisplayParts(function(i){if(n.typeParameters&&n.typeParameters.length){var s=e.createNodeArray(n.typeParameters.map(function(e){return t.typeParameterToDeclaration(e,r)}));o.writeList(53776,s,a,i)}}),c=n.parameters.map(function(e){return G(e,t,r,a,o)});return{isVariadic:i,parameters:c,prefix:s.concat([e.punctuationPart(20)]),suffix:[e.punctuationPart(21)]}}function G(n,t,r,a,i){var o=e.mapToDisplayParts(function(e){var o=t.symbolToParameterDeclaration(n,r,k);i.writeNode(4,o,a,e)}),s=t.isOptionalParameter(n.valueDeclaration);return{name:n.name,documentation:n.getDocumentationComment(t),displayParts:o,isOptional:s}}function V(n,t,r,a,i){var o=e.mapToDisplayParts(function(e){var o=t.typeParameterToDeclaration(n,r);i.writeNode(4,o,a,e)});return{name:n.symbol.name,documentation:n.symbol.getDocumentationComment(t),displayParts:o,isOptional:!1}}})(e.SignatureHelp||(e.SignatureHelp={}))}(u||(u={})),function(e){var n=/^data:(?:application\/json(?:;charset=[uU][tT][fF]-8);base64,([A-Za-z0-9+\/=]+)$)?/;function t(n){var t=e.createGetCanonicalFileName(n.useCaseSensitiveFileNames()),r=n.getCurrentDirectory(),a=e.createMap(),o=e.createMap();return{tryGetSourcePosition:l,tryGetGeneratedPosition:u,toLineColumnOffset:f,clearCache:_};function s(n){return e.toPath(n,r,t)}function c(r,a){var i,c=s(r),l=o.get(c);if(l)return l;if(n.getDocumentPositionMapper)i=n.getDocumentPositionMapper(r,a);else if(n.readFile){var u=m(r);i=u&&e.getDocumentPositionMapper({getSourceFileLike:m,getCanonicalFileName:t,log:function(e){return n.log(e)}},r,e.getLineInfo(u.text,e.getLineStarts(u)),function(e){return!n.fileExists||n.fileExists(e)?n.readFile(e):void 0})}return o.set(c,i||e.identitySourceMapConsumer),i||e.identitySourceMapConsumer}function l(n){if(e.isDeclarationFileName(n.fileName)){var t=d(n.fileName);if(t){var r=c(n.fileName).getSourcePosition(n);return r&&r!==n?l(r)||r:void 0}}}function u(a){if(!e.isDeclarationFileName(a.fileName)){var i=d(a.fileName);if(i){var o=n.getProgram(),s=o.getCompilerOptions(),l=s.outFile||s.out,u=l?e.removeFileExtension(l)+".d.ts":e.getDeclarationEmitOutputFilePathWorker(a.fileName,o.getCompilerOptions(),r,o.getCommonSourceDirectory(),t);if(void 0!==u){var p=c(u,a.fileName).getGeneratedPosition(a);return p===a?void 0:p}}}}function d(e){var t=n.getProgram();if(t){var r=s(e),a=t.getSourceFileByPath(r);return a&&a.resolvedPath===r?a:void 0}}function p(e){var t=s(e),r=a.get(t);if(void 0!==r)return r||void 0;if(n.readFile&&(!n.fileExists||n.fileExists(t))){var o=n.readFile(t),c=!!o&&i(o);return a.set(t,c),c||void 0}a.set(t,!1)}function m(e){return n.getSourceFileLike?n.getSourceFileLike(e):d(e)||p(e)}function f(e,n){var t=m(e);return t.getLineAndCharacterOfPosition(n)}function _(){a.clear(),o.clear()}}function r(t,r,i,o){var s=e.tryGetSourceMappingURL(i);if(s){var c=n.exec(s);if(c){if(c[1]){var l=c[1];return a(t,e.base64decode(e.sys,l),r)}s=void 0}}var u=[];s&&u.push(s),u.push(r+".map");for(var d=s&&e.getNormalizedAbsolutePath(s,e.getDirectoryPath(r)),p=0,m=u;p<m.length;p++){var f=m[p],_=e.getNormalizedAbsolutePath(f,e.getDirectoryPath(r)),g=o(_,d);if(e.isString(g))return a(t,g,_);if(void 0!==g)return g||void 0}}function a(n,t,r){var a=e.tryParseRawSourceMap(t);if(a&&a.sources&&a.file&&a.mappings)return e.createDocumentPositionMapper(n,a,r)}function i(n,t){return{text:n,lineMap:t,getLineAndCharacterOfPosition:function(n){return e.computeLineAndCharacterOfPosition(e.getLineStarts(this),n)}}}e.getSourceMapper=t,e.getDocumentPositionMapper=r}(u||(u={})),function(e){var n=e.createMap();function t(t,a,s){a.getSemanticDiagnostics(t,s);var c=[],u=a.getTypeChecker();t.commonJsModuleIndicator&&(e.programContainsEs6Modules(a)||e.compilerOptionsIndicateEs6Modules(a.getCompilerOptions()))&&r(t)&&c.push(e.createDiagnosticForNode(l(t.commonJsModuleIndicator),e.Diagnostics.File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module));var d=e.isSourceFileJS(t);if(n.clear(),h(t),e.getAllowSyntheticDefaultImports(a.getCompilerOptions()))for(var p=0,m=t.imports;p<m.length;p++){var f=m[p],_=e.importFromModuleSpecifier(f),g=i(_);if(g){var v=e.getResolvedModule(t,f.text),y=v&&a.getSourceFile(v.resolvedFileName);y&&y.externalModuleIndicator&&e.isExportAssignment(y.externalModuleIndicator)&&y.externalModuleIndicator.isExportEquals&&c.push(e.createDiagnosticForNode(g,e.Diagnostics.Import_may_be_converted_to_a_default_import))}}return e.addRange(c,t.bindSuggestionDiagnostics),e.addRange(c,a.getSuggestionDiagnostics(t,s)),c.sort(function(e,n){return e.start-n.start});function h(n){if(d)switch(n.kind){case 196:var r=e.getDeclarationOfExpando(n);if(r){var a=r.symbol;if(a&&(a.exports&&a.exports.size||a.members&&a.members.size)){c.push(e.createDiagnosticForNode(e.isVariableDeclaration(n.parent)?n.parent.name:n,e.Diagnostics.This_constructor_function_may_be_converted_to_a_class_declaration));break}}case 239:var i=n.symbol;i.members&&i.members.size>0&&c.push(e.createDiagnosticForNode(e.isVariableDeclaration(n.parent)?n.parent.name:n,e.Diagnostics.This_constructor_function_may_be_converted_to_a_class_declaration));break}else{if(e.isVariableStatement(n)&&n.parent===t&&2&n.declarationList.flags&&1===n.declarationList.declarations.length){var s=n.declarationList.declarations[0].initializer;s&&e.isRequireCall(s,!0)&&c.push(e.createDiagnosticForNode(s,e.Diagnostics.require_call_may_be_converted_to_an_import))}e.codefix.parameterShouldGetTypeFromJSDoc(n)&&c.push(e.createDiagnosticForNode(n.name||n,e.Diagnostics.JSDoc_types_may_be_moved_to_TypeScript_types))}e.isFunctionLikeDeclaration(n)&&o(n,u,c),n.forEachChild(h)}}function r(n){return n.statements.some(function(n){switch(n.kind){case 219:return n.declarationList.declarations.some(function(n){return!!n.initializer&&e.isRequireCall(a(n.initializer),!0)});case 221:var t=n.expression;if(!e.isBinaryExpression(t))return e.isRequireCall(t,!0);var r=e.getAssignmentDeclarationKind(t);return 1===r||2===r;default:return!1}})}function a(n){return e.isPropertyAccessExpression(n)?a(n.expression):n}function i(n){switch(n.kind){case 249:var t=n.importClause,r=n.moduleSpecifier;return t&&!t.name&&t.namedBindings&&251===t.namedBindings.kind&&e.isStringLiteral(r)?t.namedBindings.name:void 0;case 248:return n.name;default:return}}function o(t,r,a){s(t,r)&&!n.has(_(t))&&a.push(e.createDiagnosticForNode(!t.name&&e.isVariableDeclaration(t.parent)&&e.isIdentifier(t.parent.name)?t.parent.name:t,e.Diagnostics.This_may_be_converted_to_an_async_function))}function s(n,t){return!e.isAsyncFunction(n)&&n.body&&e.isBlock(n.body)&&u(n.body)&&c(n,t)}function c(e,n){var t=n.getTypeAtLocation(e),r=n.getSignaturesOfType(t,0),a=r.length?n.getReturnTypeOfSignature(r[0]):void 0;return!!a&&!!n.getPromisedTypeOfPromise(a)}function l(n){return e.isBinaryExpression(n)?n.left:n}function u(n){return!!e.forEachReturnStatement(n,d)}function d(n){return e.isReturnStatement(n)&&!!n.expression&&p(n.expression)}function p(n){if(!m(n)||!n.arguments.every(f))return!1;var t=n.expression;while(m(t)||e.isPropertyAccessExpression(t)){if(e.isCallExpression(t)&&!t.arguments.every(f))return!1;t=t.expression}return!0}function m(n){return e.isCallExpression(n)&&(e.hasPropertyAccessExpressionWithName(n,"then")||e.hasPropertyAccessExpressionWithName(n,"catch"))}function f(e){switch(e.kind){case 239:case 196:case 197:n.set(_(e),!0);case 96:case 72:return!0;default:return!1}}function _(e){return e.pos.toString()+":"+e.end.toString()}e.computeSuggestionDiagnostics=t,e.isReturnStatementWithFixablePromiseHandler=d,e.isFixablePromiseHandler=p}(u||(u={})),function(e){(function(n){function t(n,t,a){var i=r(n,t,a);if(""!==i)return i;var o=e.getCombinedLocalAndExportSymbolFlags(t);return 32&o?e.getDeclarationOfKind(t,209)?"local class":"class":384&o?"enum":524288&o?"type":64&o?"interface":262144&o?"type parameter":262144&o?"type parameter":8&o?"enum member":2097152&o?"alias":1536&o?"module":i}function r(n,t,r){var a=n.getRootSymbols(t);if(1===a.length&&8192&e.first(a).flags&&0!==n.getTypeOfSymbolAtLocation(t,r).getNonNullableType().getCallSignatures().length)return"method";if(n.isUndefinedSymbol(t))return"var";if(n.isArgumentsSymbol(t))return"local var";if(100===r.kind&&e.isExpression(r))return"parameter";var i=e.getCombinedLocalAndExportSymbolFlags(t);if(3&i)return e.isFirstDeclarationOfSymbolParameter(t)?"parameter":t.valueDeclaration&&e.isVarConst(t.valueDeclaration)?"const":e.forEach(t.declarations,e.isLet)?"let":o(t)?"local var":"var";if(16&i)return o(t)?"local function":"function";if(32768&i)return"getter";if(65536&i)return"setter";if(8192&i)return"method";if(16384&i)return"constructor";if(4&i){if(33554432&i&&6&t.checkFlags){var s=e.forEach(n.getRootSymbols(t),function(n){var t=n.getFlags();if(98311&t)return"property";e.Debug.assert(!!(8208&t))});if(!s){var c=n.getTypeOfSymbolAtLocation(t,r);return c.getCallSignatures().length?"method":"property"}return s}switch(r.parent&&r.parent.kind){case 262:case 260:case 261:return 72===r.kind?"property":"JSX attribute";case 267:return"JSX attribute";default:return"property"}}return""}function a(n){var t=n&&n.declarations&&n.declarations.length>0?e.getNodeModifiers(n.declarations[0]):"",r=n&&16777216&n.flags?"optional":"";return t&&r?t+","+r:t||r}function i(n,a,o,s,c,l,u){void 0===l&&(l=e.getMeaningFromLocation(c));var d,p,m,f,_,g,v=[],y=e.getCombinedLocalAndExportSymbolFlags(a),h=1&l?r(n,a,c):"",b=!1,E=100===c.kind&&e.isInExpressionContext(c);if(100===c.kind&&!E)return{displayParts:[e.keywordPart(100)],documentation:[],symbolKind:"primitive type",tags:void 0};if(""!==h||32&y||2097152&y){"getter"!==h&&"setter"!==h||(h="property");var T=void 0;if(m=E?n.getTypeAtLocation(c):n.getTypeOfSymbolAtLocation(a.exportSymbol||a,c),c.parent&&189===c.parent.kind){var S=c.parent.name;(S===c||S&&0===S.getFullWidth())&&(c=c.parent)}var x=void 0;if(e.isCallOrNewExpression(c)?x=c:e.isCallExpressionTarget(c)||e.isNewExpressionTarget(c)?x=c.parent:c.parent&&e.isJsxOpeningLikeElement(c.parent)&&e.isFunctionLike(a.valueDeclaration)&&(x=c.parent),x){var C=[];T=n.getResolvedSignature(x,C);var A=192===x.kind||e.isCallExpression(x)&&98===x.expression.kind,D=A?m.getConstructSignatures():m.getCallSignatures();if(e.contains(D,T.target)||e.contains(D,T)||(T=D.length?D[0]:void 0),T){switch(A&&32&y?(h="constructor",X(m.symbol,h)):2097152&y?(h="alias",Y(h),v.push(e.spacePart()),A&&(v.push(e.keywordPart(95)),v.push(e.spacePart())),J(a)):X(a,h),h){case"JSX attribute":case"property":case"var":case"const":case"let":case"parameter":case"local var":v.push(e.punctuationPart(57)),v.push(e.spacePart()),16&e.getObjectFlags(m)||!m.symbol||(e.addRange(v,e.symbolToDisplayParts(n,m.symbol,s,void 0,5)),v.push(e.lineBreakPart())),A&&(v.push(e.keywordPart(95)),v.push(e.spacePart())),Q(T,D,262144);break;default:Q(T,D)}b=!0}}else if(e.isNameOfFunctionDeclaration(c)&&!(98304&y)||124===c.kind&&157===c.parent.kind){var L=c.parent,k=e.find(a.declarations,function(e){return e===(124===c.kind?L.parent:L)});if(k){D=157===L.kind?m.getNonNullableType().getConstructSignatures():m.getNonNullableType().getCallSignatures();T=n.isImplementationOfOverload(L)?D[0]:n.getSignatureFromDeclaration(L),157===L.kind?(h="constructor",X(m.symbol,h)):X(160!==L.kind||2048&m.symbol.flags||4096&m.symbol.flags?a:m.symbol,h),Q(T,D),b=!0}}}if(32&y&&!b&&!E&&(q(),e.getDeclarationOfKind(a,209)?Y("local class"):v.push(e.keywordPart(76)),v.push(e.spacePart()),J(a),Z(a,o)),64&y&&2&l&&(W(),v.push(e.keywordPart(110)),v.push(e.spacePart()),J(a),Z(a,o)),524288&y&&2&l&&(W(),v.push(e.keywordPart(140)),v.push(e.spacePart()),J(a),Z(a,o),v.push(e.spacePart()),v.push(e.operatorPart(59)),v.push(e.spacePart()),e.addRange(v,e.typeToDisplayParts(n,n.getDeclaredTypeOfSymbol(a),s,8388608))),384&y&&(W(),e.some(a.declarations,function(n){return e.isEnumDeclaration(n)&&e.isEnumConst(n)})&&(v.push(e.keywordPart(77)),v.push(e.spacePart())),v.push(e.keywordPart(84)),v.push(e.spacePart()),J(a)),1536&y){W();var N=e.getDeclarationOfKind(a,244),I=N&&N.name&&72===N.name.kind;v.push(e.keywordPart(I?131:130)),v.push(e.spacePart()),J(a)}if(262144&y&&2&l)if(W(),v.push(e.punctuationPart(20)),v.push(e.textPart("type parameter")),v.push(e.punctuationPart(21)),v.push(e.spacePart()),J(a),a.parent)z(),J(a.parent,s),Z(a.parent,s);else{var w=e.getDeclarationOfKind(a,150);if(void 0===w)return e.Debug.fail();N=w.parent;if(N)if(e.isFunctionLikeKind(N.kind)){z();T=n.getSignatureFromDeclaration(N);161===N.kind?(v.push(e.keywordPart(95)),v.push(e.spacePart())):160!==N.kind&&N.name&&J(N.symbol),e.addRange(v,e.signatureToDisplayParts(n,T,o,32))}else 242===N.kind&&(z(),v.push(e.keywordPart(140)),v.push(e.spacePart()),J(N.symbol),Z(N.symbol,o))}if(8&y){h="enum member",X(a,"enum member");N=a.declarations[0];if(278===N.kind){var M=n.getConstantValue(N);void 0!==M&&(v.push(e.spacePart()),v.push(e.operatorPart(59)),v.push(e.spacePart()),v.push(e.displayPart(e.getTextOfConstantValue(M),"number"===typeof M?e.SymbolDisplayPartKind.numericLiteral:e.SymbolDisplayPartKind.stringLiteral)))}}if(2097152&y){if(W(),!b){var O=n.getAliasedSymbol(a);if(O!==a&&O.declarations&&O.declarations.length>0){var P=O.declarations[0],R=e.getNameOfDeclaration(P);if(R){var F=e.isModuleWithStringLiteralName(P)&&e.hasModifier(P,2),G="default"!==a.name&&!F,V=i(n,O,e.getSourceFileOfNode(P),P,R,l,G?a:O);v.push.apply(v,V.displayParts),v.push(e.lineBreakPart()),_=V.documentation,g=V.tags}}}switch(a.declarations[0].kind){case 247:v.push(e.keywordPart(85)),v.push(e.spacePart()),v.push(e.keywordPart(131));break;case 254:v.push(e.keywordPart(85)),v.push(e.spacePart()),v.push(e.keywordPart(a.declarations[0].isExportEquals?59:80));break;case 257:v.push(e.keywordPart(85));break;default:v.push(e.keywordPart(92))}v.push(e.spacePart()),J(a),e.forEach(a.declarations,function(t){if(248===t.kind){var r=t;if(e.isExternalModuleImportEqualsDeclaration(r))v.push(e.spacePart()),v.push(e.operatorPart(59)),v.push(e.spacePart()),v.push(e.keywordPart(134)),v.push(e.punctuationPart(20)),v.push(e.displayPart(e.getTextOfNode(e.getExternalModuleImportEqualsDeclarationExpression(r)),e.SymbolDisplayPartKind.stringLiteral)),v.push(e.punctuationPart(21));else{var a=n.getSymbolAtLocation(r.moduleReference);a&&(v.push(e.spacePart()),v.push(e.operatorPart(59)),v.push(e.spacePart()),J(a,s))}return!0}})}if(!b)if(""!==h){if(m)if(E?(W(),v.push(e.keywordPart(100))):X(a,h),"property"===h||"JSX attribute"===h||3&y||"local var"===h||E)if(v.push(e.punctuationPart(57)),v.push(e.spacePart()),m.symbol&&262144&m.symbol.flags){var B=e.mapToDisplayParts(function(t){var r=n.typeParameterToDeclaration(m,s);H().writeNode(4,r,e.getSourceFileOfNode(e.getParseTreeNode(s)),t)});e.addRange(v,B)}else e.addRange(v,e.typeToDisplayParts(n,m,s));else if(16&y||8192&y||16384&y||131072&y||98304&y||"method"===h){D=m.getNonNullableType().getCallSignatures();D.length&&Q(D[0],D)}}else h=t(n,a,c);if(!d&&(d=a.getDocumentationComment(n),p=a.getJsDocTags(),0===d.length&&4&y&&a.parent&&e.forEach(a.parent.declarations,function(e){return 279===e.kind})))for(var K=0,j=a.declarations;K<j.length;K++){N=j[K];if(N.parent&&204===N.parent.kind){var U=n.getSymbolAtLocation(N.parent.right);if(U&&(d=U.getDocumentationComment(n),p=U.getJsDocTags(),d.length>0))break}}return 0===d.length&&_&&(d=_),0===p.length&&g&&(p=g),{displayParts:v,documentation:d,symbolKind:h,tags:0===p.length?void 0:p};function H(){return f||(f=e.createPrinter({removeComments:!0})),f}function W(){v.length&&v.push(e.lineBreakPart()),q()}function q(){u&&(Y("alias"),v.push(e.spacePart()))}function z(){v.push(e.spacePart()),v.push(e.keywordPart(93)),v.push(e.spacePart())}function J(t,r){u&&t===a&&(t=u);var i=e.symbolToDisplayParts(n,t,r||o,void 0,7);e.addRange(v,i),16777216&a.flags&&v.push(e.punctuationPart(56))}function X(n,t){W(),t&&(Y(t),n&&!e.some(n.declarations,function(n){return e.isArrowFunction(n)||(e.isFunctionExpression(n)||e.isClassExpression(n))&&!n.name})&&(v.push(e.spacePart()),J(n)))}function Y(n){switch(n){case"var":case"function":case"let":case"const":case"constructor":return void v.push(e.textOrKeywordPart(n));default:return v.push(e.punctuationPart(20)),v.push(e.textOrKeywordPart(n)),void v.push(e.punctuationPart(21))}}function Q(t,r,a){void 0===a&&(a=0),e.addRange(v,e.signatureToDisplayParts(n,t,s,32|a)),r.length>1&&(v.push(e.spacePart()),v.push(e.punctuationPart(20)),v.push(e.operatorPart(38)),v.push(e.displayPart((r.length-1).toString(),e.SymbolDisplayPartKind.numericLiteral)),v.push(e.spacePart()),v.push(e.textPart(2===r.length?"overload":"overloads")),v.push(e.punctuationPart(21)));var i=t.getDocumentationComment(n);d=0===i.length?void 0:i,p=t.getJsDocTags()}function Z(t,r){var a=e.mapToDisplayParts(function(a){var i=n.symbolToTypeParameterDeclarations(t,r);H().writeList(53776,i,e.getSourceFileOfNode(e.getParseTreeNode(r)),a)});e.addRange(v,a)}}function o(n){return!n.parent&&e.forEach(n.declarations,function(n){if(196===n.kind)return!0;if(237!==n.kind&&239!==n.kind)return!1;for(var t=n.parent;!e.isFunctionBlock(t);t=t.parent)if(279===t.kind||245===t.kind)return!1;return!0})}n.getSymbolKind=t,n.getSymbolModifiers=a,n.getSymbolDisplayPartsDocumentationAndSymbolKind=i})(e.SymbolDisplay||(e.SymbolDisplay={}))}(u||(u={})),function(e){function n(n,t){var r=[],i=t.compilerOptions?a(t.compilerOptions,r):e.getDefaultCompilerOptions();i.isolatedModules=!0,i.suppressOutputPathCheck=!0,i.allowNonTsExtensions=!0,i.noLib=!0,i.lib=void 0,i.types=void 0,i.noEmit=void 0,i.noEmitOnError=void 0,i.paths=void 0,i.rootDirs=void 0,i.declaration=void 0,i.composite=void 0,i.declarationDir=void 0,i.out=void 0,i.outFile=void 0,i.noResolve=!0;var o=t.fileName||(i.jsx?"module.tsx":"module.ts"),s=e.createSourceFile(o,n,i.target);t.moduleName&&(s.moduleName=t.moduleName),t.renamedDependencies&&(s.renamedDependencies=e.createMapFromTemplate(t.renamedDependencies));var c,l,u=e.getNewLineCharacter(i),d={getSourceFile:function(n){return n===e.normalizePath(o)?s:void 0},writeFile:function(n,t){e.fileExtensionIs(n,".map")?(e.Debug.assertEqual(l,void 0,"Unexpected multiple source map outputs, file:",n),l=t):(e.Debug.assertEqual(c,void 0,"Unexpected multiple outputs, file:",n),c=t)},getDefaultLibFileName:function(){return"lib.d.ts"},useCaseSensitiveFileNames:function(){return!1},getCanonicalFileName:function(e){return e},getCurrentDirectory:function(){return""},getNewLine:function(){return u},fileExists:function(e){return e===o},readFile:function(){return""},directoryExists:function(){return!0},getDirectories:function(){return[]}},p=e.createProgram([o],i,d);return t.reportDiagnostics&&(e.addRange(r,p.getSyntacticDiagnostics(s)),e.addRange(r,p.getOptionsDiagnostics())),p.emit(void 0,void 0,void 0,void 0,t.transformers),void 0===c?e.Debug.fail("Output generation failed"):{outputText:c,diagnostics:r,sourceMapText:l}}function t(t,r,a,i,o){var s=n(t,{compilerOptions:r,fileName:a,reportDiagnostics:!!i,moduleName:o});return e.addRange(i,s.diagnostics),s.outputText}var r;function a(n,t){r=r||e.filter(e.optionDeclarations,function(n){return"object"===typeof n.type&&!e.forEachEntry(n.type,function(e){return"number"!==typeof e})}),n=e.cloneCompilerOptions(n);for(var a=function(r){if(!e.hasProperty(n,r.name))return"continue";var a=n[r.name];e.isString(a)?n[r.name]=e.parseCustomTypeOption(r,a,t):e.forEachEntry(r.type,function(e){return e===a})||t.push(e.createCompilerDiagnosticForInvalidCustomType(r))},i=0,o=r;i<o.length;i++){var s=o[i];a(s)}return n}e.transpileModule=n,e.transpile=t,e.fixupCompilerOptions=a}(u||(u={})),function(e){(function(n){(function(e){e[e["FormatDocument"]=0]="FormatDocument",e[e["FormatSelection"]=1]="FormatSelection",e[e["FormatOnEnter"]=2]="FormatOnEnter",e[e["FormatOnSemicolon"]=3]="FormatOnSemicolon",e[e["FormatOnOpeningCurlyBrace"]=4]="FormatOnOpeningCurlyBrace",e[e["FormatOnClosingCurlyBrace"]=5]="FormatOnClosingCurlyBrace"})(n.FormattingRequestKind||(n.FormattingRequestKind={}));var t=function(){function n(e,n,t){this.sourceFile=e,this.formattingRequestKind=n,this.options=t}return n.prototype.updateContext=function(n,t,r,a,i){this.currentTokenSpan=e.Debug.assertDefined(n),this.currentTokenParent=e.Debug.assertDefined(t),this.nextTokenSpan=e.Debug.assertDefined(r),this.nextTokenParent=e.Debug.assertDefined(a),this.contextNode=e.Debug.assertDefined(i),this.contextNodeAllOnSameLine=void 0,this.nextNodeAllOnSameLine=void 0,this.tokensAreOnSameLine=void 0,this.contextNodeBlockIsOnOneLine=void 0,this.nextNodeBlockIsOnOneLine=void 0},n.prototype.ContextNodeAllOnSameLine=function(){return void 0===this.contextNodeAllOnSameLine&&(this.contextNodeAllOnSameLine=this.NodeIsOnOneLine(this.contextNode)),this.contextNodeAllOnSameLine},n.prototype.NextNodeAllOnSameLine=function(){return void 0===this.nextNodeAllOnSameLine&&(this.nextNodeAllOnSameLine=this.NodeIsOnOneLine(this.nextTokenParent)),this.nextNodeAllOnSameLine},n.prototype.TokensAreOnSameLine=function(){if(void 0===this.tokensAreOnSameLine){var e=this.sourceFile.getLineAndCharacterOfPosition(this.currentTokenSpan.pos).line,n=this.sourceFile.getLineAndCharacterOfPosition(this.nextTokenSpan.pos).line;this.tokensAreOnSameLine=e===n}return this.tokensAreOnSameLine},n.prototype.ContextNodeBlockIsOnOneLine=function(){return void 0===this.contextNodeBlockIsOnOneLine&&(this.contextNodeBlockIsOnOneLine=this.BlockIsOnOneLine(this.contextNode)),this.contextNodeBlockIsOnOneLine},n.prototype.NextNodeBlockIsOnOneLine=function(){return void 0===this.nextNodeBlockIsOnOneLine&&(this.nextNodeBlockIsOnOneLine=this.BlockIsOnOneLine(this.nextTokenParent)),this.nextNodeBlockIsOnOneLine},n.prototype.NodeIsOnOneLine=function(e){var n=this.sourceFile.getLineAndCharacterOfPosition(e.getStart(this.sourceFile)).line,t=this.sourceFile.getLineAndCharacterOfPosition(e.getEnd()).line;return n===t},n.prototype.BlockIsOnOneLine=function(n){var t=e.findChildOfKind(n,18,this.sourceFile),r=e.findChildOfKind(n,19,this.sourceFile);if(t&&r){var a=this.sourceFile.getLineAndCharacterOfPosition(t.getEnd()).line,i=this.sourceFile.getLineAndCharacterOfPosition(r.getStart(this.sourceFile)).line;return a===i}return!1},n}();n.FormattingContext=t})(e.formatting||(e.formatting={}))}(u||(u={})),function(e){(function(n){var t,r=e.createScanner(6,!1,0),a=e.createScanner(6,!1,1);function i(n,t,i,o,s){var c=1===t?a:r;c.setText(n),c.setTextPos(i);var l,u,d,p,m,f=!0,_=s({advance:g,readTokenInfo:S,isOnToken:C,getCurrentLeadingTrivia:function(){return l},lastTrailingTriviaWasNewLine:function(){return f},skipToEndOf:D});return m=void 0,c.setText(void 0),_;function g(){m=void 0;var n=c.getStartPos()!==i;n?f=!!u&&4===e.last(u).kind:c.scan(),l=void 0,u=void 0;var t=c.getStartPos();while(t<o){var r=c.getToken();if(!e.isTrivia(r))break;c.scan();var a={pos:t,end:c.getStartPos(),kind:r};t=c.getStartPos(),l=e.append(l,a)}d=c.getStartPos()}function v(e){switch(e.kind){case 32:case 67:case 68:case 48:case 47:return!0}return!1}function y(n){if(n.parent)switch(n.parent.kind){case 267:case 262:case 263:case 261:return e.isKeyword(n.kind)||72===n.kind}return!1}function h(e){return 11===e.kind}function b(e){return 13===e.kind}function E(e){return 16===e.kind||17===e.kind}function T(e){return 42===e||64===e}function S(n){e.Debug.assert(C());var t=v(n)?1:b(n)?2:E(n)?3:y(n)?4:h(n)?5:0;if(m&&t===p)return A(m,n);c.getStartPos()!==d&&(e.Debug.assert(void 0!==m),c.setTextPos(d),c.scan());var r=x(n,t),a={pos:c.getStartPos(),end:c.getTextPos(),kind:r};u&&(u=void 0);while(c.getStartPos()<o){if(r=c.scan(),!e.isTrivia(r))break;var i={pos:c.getStartPos(),end:c.getTextPos(),kind:r};if(u||(u=[]),u.push(i),4===r){c.scan();break}}return m={leadingTrivia:l,trailingTrivia:u,token:a},A(m,n)}function x(n,t){var r=c.getToken();switch(p=0,t){case 1:if(30===r){p=1;var a=c.reScanGreaterToken();return e.Debug.assert(n.kind===a),a}break;case 2:if(T(r)){p=2;a=c.reScanSlashToken();return e.Debug.assert(n.kind===a),a}break;case 3:if(19===r)return p=3,c.reScanTemplateToken();break;case 4:return p=4,c.scanJsxIdentifier();case 5:return p=5,c.reScanJsxToken();case 0:break;default:e.Debug.assertNever(t)}return r}function C(){var n=m?m.token.kind:c.getToken(),t=m?m.token.pos:c.getStartPos();return t<o&&1!==n&&!e.isTrivia(n)}function A(n,t){return e.isToken(t)&&n.token.kind!==t.kind&&(n.token.kind=t.kind),n}function D(e){c.setTextPos(e.end),d=c.getStartPos(),p=void 0,m=void 0,f=!1,l=void 0,u=void 0}}(function(e){e[e["Scan"]=0]="Scan",e[e["RescanGreaterThanToken"]=1]="RescanGreaterThanToken",e[e["RescanSlashToken"]=2]="RescanSlashToken",e[e["RescanTemplateToken"]=3]="RescanTemplateToken",e[e["RescanJsxIdentifier"]=4]="RescanJsxIdentifier",e[e["RescanJsxText"]=5]="RescanJsxText"})(t||(t={})),n.getFormattingScanner=i})(e.formatting||(e.formatting={}))}(u||(u={})),function(e){(function(n){n.anyContext=e.emptyArray,function(e){e[e["Ignore"]=1]="Ignore",e[e["Space"]=2]="Space",e[e["NewLine"]=4]="NewLine",e[e["Delete"]=8]="Delete"}(n.RuleAction||(n.RuleAction={})),function(e){e[e["None"]=0]="None",e[e["CanDeleteNewLines"]=1]="CanDeleteNewLines"}(n.RuleFlags||(n.RuleFlags={}))})(e.formatting||(e.formatting={}))}(u||(u={})),function(e){(function(n){function t(){for(var t=[],i=0;i<=147;i++)t.push(i);function S(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return{tokens:t.filter(function(n){return!e.some(function(e){return e===n})}),isSpecific:!1}}var x={tokens:t,isSpecific:!1},A=a(t.concat([3])),I=o(73,147),P=o(28,71),R=[93,94,147,119,128],Y=[44,45,53,52],te=[8,9,72,20,22,18,100,95],ce=[72,20,100,95],le=[72,21,23,95],ue=[72,20,100,95],de=[72,21,23,95],pe=[2,3],me=[72].concat(e.typeKeywords),fe=A,_e=a([72,3,76,85,92]),ge=a([21,3,82,103,88,83]),ve=[r("IgnoreBeforeComment",x,pe,n.anyContext,1),r("IgnoreAfterLineComment",2,x,n.anyContext,1),r("NotSpaceBeforeColon",x,57,[j,_,g],8),r("SpaceAfterColon",57,x,[j,_],2),r("NoSpaceBeforeQuestionMark",x,56,[j,_],8),r("SpaceAfterQuestionMarkInConditionalOperator",56,x,[j,y],2),r("NoSpaceAfterQuestionMark",56,x,[j],8),r("NoSpaceBeforeDot",x,24,[j],8),r("NoSpaceAfterDot",24,x,[j],8),r("NoSpaceBetweenImportParenInImportType",92,20,[j,K],8),r("NoSpaceAfterUnaryPrefixOperator",Y,te,[j,_],8),r("NoSpaceAfterUnaryPreincrementOperator",44,ce,[j],8),r("NoSpaceAfterUnaryPredecrementOperator",45,ue,[j],8),r("NoSpaceBeforeUnaryPostincrementOperator",le,44,[j],8),r("NoSpaceBeforeUnaryPostdecrementOperator",de,45,[j],8),r("SpaceAfterPostincrementWhenFollowedByAdd",44,38,[j,f],2),r("SpaceAfterAddWhenFollowedByUnaryPlus",38,38,[j,f],2),r("SpaceAfterAddWhenFollowedByPreincrement",38,44,[j,f],2),r("SpaceAfterPostdecrementWhenFollowedBySubtract",45,39,[j,f],2),r("SpaceAfterSubtractWhenFollowedByUnaryMinus",39,39,[j,f],2),r("SpaceAfterSubtractWhenFollowedByPredecrement",39,45,[j,f],2),r("NoSpaceAfterCloseBrace",19,[27,26],[j],8),r("NewLineBeforeCloseBraceInBlockContext",A,19,[T],4),r("SpaceAfterCloseBrace",19,S(21),[j,w],2),r("SpaceBetweenCloseBraceAndElse",19,83,[j],2),r("SpaceBetweenCloseBraceAndWhile",19,107,[j],2),r("NoSpaceBetweenEmptyBraceBrackets",18,19,[j,O],8),r("SpaceAfterConditionalClosingParen",21,22,[M],2),r("NoSpaceBetweenFunctionKeywordAndStar",90,40,[k],8),r("SpaceAfterStarInGeneratorDeclaration",40,[72,20],[k],2),r("SpaceAfterFunctionInFuncDecl",90,x,[D],2),r("NewLineAfterOpenBraceInBlockContext",18,x,[T],4),r("SpaceAfterGetSetInMember",[126,137],72,[D],2),r("NoSpaceBetweenYieldKeywordAndStar",117,40,[j,oe],8),r("SpaceBetweenYieldOrYieldStarAndOperand",[117,40],x,[j,oe],2),r("NoSpaceBetweenReturnAndSemicolon",97,26,[j],8),r("SpaceAfterCertainKeywords",[105,101,95,81,97,104,122],x,[j],2),r("SpaceAfterLetConstInVariableDeclaration",[111,77],x,[j,Q],2),r("NoSpaceBeforeOpenParenInFuncCall",x,20,[j,F,G],8),r("SpaceBeforeBinaryKeywordOperator",x,R,[j,f],2),r("SpaceAfterBinaryKeywordOperator",R,x,[j,f],2),r("SpaceAfterVoidOperator",106,x,[j,ie],2),r("SpaceBetweenAsyncAndOpenParen",121,20,[B,j],2),r("SpaceBetweenAsyncAndFunctionKeyword",121,90,[j],2),r("NoSpaceBetweenTagAndTemplateString",[72,21],[14,15],[j],8),r("SpaceBeforeJsxAttribute",x,72,[W,j],2),r("SpaceBeforeSlashInJsxOpeningElement",x,42,[z,j],2),r("NoSpaceBeforeGreaterThanTokenInJsxOpeningElement",42,30,[z,j],8),r("NoSpaceBeforeEqualInJsxAttribute",x,59,[q,j],8),r("NoSpaceAfterEqualInJsxAttribute",59,x,[q,j],8),r("NoSpaceAfterModuleImport",[130,134],20,[j],8),r("SpaceAfterCertainTypeScriptKeywords",[118,76,125,80,84,85,86,126,109,92,110,130,131,113,115,114,133,137,116,140,144,129,127],x,[j],2),r("SpaceBeforeCertainTypeScriptKeywords",x,[86,109,144],[j],2),r("SpaceAfterModuleName",10,18,[$],2),r("SpaceBeforeArrow",x,37,[j],2),r("SpaceAfterArrow",37,x,[j],2),r("NoSpaceAfterEllipsis",25,72,[j],8),r("NoSpaceAfterOptionalParameters",56,[21,27],[j,_],8),r("NoSpaceBetweenEmptyInterfaceBraceBrackets",18,19,[j,ee],8),r("NoSpaceBeforeOpenAngularBracket",me,28,[j,re],8),r("NoSpaceBetweenCloseParenAndAngularBracket",21,28,[j,re],8),r("NoSpaceAfterOpenAngularBracket",28,x,[j,re],8),r("NoSpaceBeforeCloseAngularBracket",x,30,[j,re],8),r("NoSpaceAfterCloseAngularBracket",30,[20,22,30,27],[j,re,L],8),r("SpaceBeforeAt",[21,72],58,[j],2),r("NoSpaceAfterAt",58,x,[j],8),r("SpaceAfterDecorator",x,[118,72,85,80,76,116,115,113,114,126,137,22,40],[X],2),r("NoSpaceBeforeNonNullAssertionOperator",x,52,[j,se],8),r("NoSpaceAfterNewKeywordOnConstructorSignature",95,20,[j,ne],8)],ye=[r("SpaceAfterConstructor",124,20,[s("insertSpaceAfterConstructor"),j],2),r("NoSpaceAfterConstructor",124,20,[l("insertSpaceAfterConstructor"),j],8),r("SpaceAfterComma",27,x,[s("insertSpaceAfterCommaDelimiter"),j,U,V],2),r("NoSpaceAfterComma",27,x,[l("insertSpaceAfterCommaDelimiter"),j,U],8),r("SpaceAfterAnonymousFunctionKeyword",90,20,[s("insertSpaceAfterFunctionKeywordForAnonymousFunctions"),D],2),r("NoSpaceAfterAnonymousFunctionKeyword",90,20,[l("insertSpaceAfterFunctionKeywordForAnonymousFunctions"),D],8),r("SpaceAfterKeywordInControl",I,20,[s("insertSpaceAfterKeywordsInControlFlowStatements"),M],2),r("NoSpaceAfterKeywordInControl",I,20,[l("insertSpaceAfterKeywordsInControlFlowStatements"),M],8),r("SpaceAfterOpenParen",20,x,[s("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"),j],2),r("SpaceBeforeCloseParen",x,21,[s("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"),j],2),r("SpaceBetweenOpenParens",20,20,[s("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"),j],2),r("NoSpaceBetweenParens",20,21,[j],8),r("NoSpaceAfterOpenParen",20,x,[l("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"),j],8),r("NoSpaceBeforeCloseParen",x,21,[l("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"),j],8),r("SpaceAfterOpenBracket",22,x,[s("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"),j],2),r("SpaceBeforeCloseBracket",x,23,[s("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"),j],2),r("NoSpaceBetweenBrackets",22,23,[j],8),r("NoSpaceAfterOpenBracket",22,x,[l("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"),j],8),r("NoSpaceBeforeCloseBracket",x,23,[l("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"),j],8),r("SpaceAfterOpenBrace",18,x,[d("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"),b],2),r("SpaceBeforeCloseBrace",x,19,[d("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"),b],2),r("NoSpaceBetweenEmptyBraceBrackets",18,19,[j,O],8),r("NoSpaceAfterOpenBrace",18,x,[c("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"),j],8),r("NoSpaceBeforeCloseBrace",x,19,[c("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"),j],8),r("SpaceAfterTemplateHeadAndMiddle",[15,16],x,[s("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"),j],2),r("SpaceBeforeTemplateMiddleAndTail",x,[16,17],[s("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"),j],2),r("NoSpaceAfterTemplateHeadAndMiddle",[15,16],x,[l("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"),j],8),r("NoSpaceBeforeTemplateMiddleAndTail",x,[16,17],[l("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"),j],8),r("SpaceAfterOpenBraceInJsxExpression",18,x,[s("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"),j,H],2),r("SpaceBeforeCloseBraceInJsxExpression",x,19,[s("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"),j,H],2),r("NoSpaceAfterOpenBraceInJsxExpression",18,x,[l("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"),j,H],8),r("NoSpaceBeforeCloseBraceInJsxExpression",x,19,[l("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"),j,H],8),r("SpaceAfterSemicolonInFor",26,x,[s("insertSpaceAfterSemicolonInForStatements"),j,p],2),r("NoSpaceAfterSemicolonInFor",26,x,[l("insertSpaceAfterSemicolonInForStatements"),j,p],8),r("SpaceBeforeBinaryOperator",x,P,[s("insertSpaceBeforeAndAfterBinaryOperators"),j,f],2),r("SpaceAfterBinaryOperator",P,x,[s("insertSpaceBeforeAndAfterBinaryOperators"),j,f],2),r("NoSpaceBeforeBinaryOperator",x,P,[l("insertSpaceBeforeAndAfterBinaryOperators"),j,f],8),r("NoSpaceAfterBinaryOperator",P,x,[l("insertSpaceBeforeAndAfterBinaryOperators"),j,f],8),r("SpaceBeforeOpenParenInFuncDecl",x,20,[s("insertSpaceBeforeFunctionParenthesis"),j,D],2),r("NoSpaceBeforeOpenParenInFuncDecl",x,20,[l("insertSpaceBeforeFunctionParenthesis"),j,D],8),r("NewLineBeforeOpenBraceInControl",ge,18,[s("placeOpenBraceOnNewLineForControlBlocks"),M,E],4,1),r("NewLineBeforeOpenBraceInFunction",fe,18,[s("placeOpenBraceOnNewLineForFunctions"),D,E],4,1),r("NewLineBeforeOpenBraceInTypeScriptDeclWithBlock",_e,18,[s("placeOpenBraceOnNewLineForFunctions"),N,E],4,1),r("SpaceAfterTypeAssertion",30,x,[s("insertSpaceAfterTypeAssertion"),j,ae],2),r("NoSpaceAfterTypeAssertion",30,x,[l("insertSpaceAfterTypeAssertion"),j,ae],8),r("SpaceBeforeTypeAnnotation",x,57,[s("insertSpaceBeforeTypeAnnotation"),j,v],2),r("NoSpaceBeforeTypeAnnotation",x,57,[l("insertSpaceBeforeTypeAnnotation"),j,v],8)],he=[r("NoSpaceBeforeSemicolon",x,26,[j],8),r("SpaceBeforeOpenBraceInControl",ge,18,[u("placeOpenBraceOnNewLineForControlBlocks"),M,Z,h],2,1),r("SpaceBeforeOpenBraceInFunction",fe,18,[u("placeOpenBraceOnNewLineForFunctions"),D,C,Z,h],2,1),r("SpaceBeforeOpenBraceInTypeScriptDeclWithBlock",_e,18,[u("placeOpenBraceOnNewLineForFunctions"),N,Z,h],2,1),r("NoSpaceBeforeComma",x,27,[j],8),r("NoSpaceBeforeOpenBracket",S(121,74),22,[j],8),r("NoSpaceAfterCloseBracket",23,x,[j,J],8),r("SpaceAfterSemicolon",26,x,[j],2),r("SpaceBetweenForAndAwaitKeyword",89,122,[j],2),r("SpaceBetweenStatements",[21,82,83,74],x,[j,U,m],2),r("SpaceAfterTryFinally",[103,88],18,[j],2)];return ve.concat(ye,he)}function r(e,n,t,r,a,o){return void 0===o&&(o=0),{leftTokenRange:i(n),rightTokenRange:i(t),rule:{debugName:e,context:r,action:a,flags:o}}}function a(e){return{tokens:e,isSpecific:!0}}function i(n){return"number"===typeof n?a([n]):e.isArray(n)?a(n):n}function o(n,t,r){void 0===r&&(r=[]);for(var i=[],o=n;o<=t;o++)e.contains(r,o)||i.push(o);return a(i)}function s(e){return function(n){return n.options&&n.options.hasOwnProperty(e)&&!!n.options[e]}}function c(e){return function(n){return n.options&&n.options.hasOwnProperty(e)&&!n.options[e]}}function l(e){return function(n){return!n.options||!n.options.hasOwnProperty(e)||!n.options[e]}}function u(e){return function(n){return!n.options||!n.options.hasOwnProperty(e)||!n.options[e]||n.TokensAreOnSameLine()}}function d(e){return function(n){return!n.options||!n.options.hasOwnProperty(e)||!!n.options[e]}}function p(e){return 225===e.contextNode.kind}function m(e){return!p(e)}function f(e){switch(e.contextNode.kind){case 204:case 205:case 175:case 212:case 257:case 253:case 163:case 173:case 174:return!0;case 186:case 242:case 248:case 237:case 151:case 278:case 154:case 153:return 59===e.currentTokenSpan.kind||59===e.nextTokenSpan.kind;case 226:case 150:return 93===e.currentTokenSpan.kind||93===e.nextTokenSpan.kind;case 227:return 147===e.currentTokenSpan.kind||147===e.nextTokenSpan.kind}return!1}function _(e){return!f(e)}function g(e){return!v(e)}function v(n){var t=n.contextNode.kind;return 154===t||153===t||151===t||237===t||e.isFunctionLikeKind(t)}function y(e){return 205===e.contextNode.kind||175===e.contextNode.kind}function h(e){return e.TokensAreOnSameLine()||C(e)}function b(e){return 184===e.contextNode.kind||181===e.contextNode.kind||S(e)}function E(e){return C(e)&&!(e.NextNodeAllOnSameLine()||e.NextNodeBlockIsOnOneLine())}function T(e){return x(e)&&!(e.ContextNodeAllOnSameLine()||e.ContextNodeBlockIsOnOneLine())}function S(e){return x(e)&&(e.ContextNodeAllOnSameLine()||e.ContextNodeBlockIsOnOneLine())}function x(e){return A(e.contextNode)}function C(e){return A(e.nextTokenParent)}function A(e){if(I(e))return!0;switch(e.kind){case 218:case 246:case 188:case 245:return!0}return!1}function D(e){switch(e.contextNode.kind){case 239:case 156:case 155:case 158:case 159:case 160:case 196:case 157:case 197:case 241:return!0}return!1}function L(e){return!D(e)}function k(e){return 239===e.contextNode.kind||196===e.contextNode.kind}function N(e){return I(e.contextNode)}function I(e){switch(e.kind){case 240:case 209:case 241:case 243:case 168:case 244:case 255:case 256:case 249:case 252:return!0}return!1}function w(e){switch(e.currentTokenParent.kind){case 240:case 244:case 243:case 274:case 245:case 232:return!0;case 218:var n=e.currentTokenParent.parent;if(!n||197!==n.kind&&196!==n.kind)return!0}return!1}function M(e){switch(e.contextNode.kind){case 222:case 232:case 225:case 226:case 227:case 224:case 235:case 223:case 231:case 274:return!0;default:return!1}}function O(e){return 188===e.contextNode.kind}function P(e){return 191===e.contextNode.kind}function R(e){return 192===e.contextNode.kind}function F(e){return P(e)||R(e)}function G(e){return 27!==e.currentTokenSpan.kind}function V(e){return 23!==e.nextTokenSpan.kind}function B(e){return 197===e.contextNode.kind}function K(e){return 183===e.contextNode.kind}function j(e){return e.TokensAreOnSameLine()&&11!==e.contextNode.kind}function U(e){return 260!==e.contextNode.kind&&264!==e.contextNode.kind}function H(e){return 270===e.contextNode.kind||269===e.contextNode.kind}function W(e){return 267===e.nextTokenParent.kind}function q(e){return 267===e.contextNode.kind}function z(e){return 261===e.contextNode.kind}function J(e){return!D(e)&&!C(e)}function X(e){return e.TokensAreOnSameLine()&&!!e.contextNode.decorators&&Y(e.currentTokenParent)&&!Y(e.nextTokenParent)}function Y(n){while(e.isExpressionNode(n))n=n.parent;return 152===n.kind}function Q(e){return 238===e.currentTokenParent.kind&&e.currentTokenParent.getStart(e.sourceFile)===e.currentTokenSpan.pos}function Z(e){return 2!==e.formattingRequestKind}function $(e){return 244===e.contextNode.kind}function ee(e){return 168===e.contextNode.kind}function ne(e){return 161===e.contextNode.kind}function te(e,n){if(28!==e.kind&&30!==e.kind)return!1;switch(n.kind){case 164:case 194:case 242:case 240:case 209:case 241:case 239:case 196:case 197:case 156:case 155:case 160:case 161:case 191:case 192:case 211:return!0;default:return!1}}function re(e){return te(e.currentTokenSpan,e.currentTokenParent)||te(e.nextTokenSpan,e.nextTokenParent)}function ae(e){return 194===e.contextNode.kind}function ie(e){return 106===e.currentTokenSpan.kind&&200===e.currentTokenParent.kind}function oe(e){return 207===e.contextNode.kind&&void 0!==e.contextNode.expression}function se(e){return 213===e.contextNode.kind}n.getAllRules=t})(e.formatting||(e.formatting={}))}(u||(u={})),function(e){(function(n){function t(e){return{options:e,getRule:a()}}var r;function a(){return void 0===r&&(r=i(n.getAllRules())),r}function i(n){var t=o(n);return function(n){var r=t[s(n.currentTokenSpan.kind,n.nextTokenSpan.kind)];return r&&e.find(r,function(t){return e.every(t.context,function(e){return e(n)})})}}function o(e){for(var n=new Array(d*d),t=new Array(n.length),r=0,a=e;r<a.length;r++)for(var i=a[r],o=i.leftTokenRange.isSpecific&&i.rightTokenRange.isSpecific,c=0,l=i.leftTokenRange.tokens;c<l.length;c++)for(var u=l[c],m=0,f=i.rightTokenRange.tokens;m<f.length;m++){var _=f[m],g=s(u,_),v=n[g];void 0===v&&(v=n[g]=[]),p(v,i.rule,o,t,g)}return n}function s(n,t){return e.Debug.assert(n<=147&&t<=147,"Must compute formatting context from tokens"),n*d+t}n.getFormatContext=t;var c,l=5,u=31,d=148;function p(e,t,r,a,i){var o=1===t.action?r?c.IgnoreRulesSpecific:c.IgnoreRulesAny:t.context!==n.anyContext?r?c.ContextRulesSpecific:c.ContextRulesAny:r?c.NoContextRulesSpecific:c.NoContextRulesAny,s=a[i]||0;e.splice(m(s,o),0,t),a[i]=f(s,o)}function m(e,n){for(var t=0,r=0;r<=n;r+=l)t+=e&u,e>>=l;return t}function f(n,t){var r=1+(n>>t&u);return e.Debug.assert((r&u)===r,"Adding more rules into the sub-bucket than allowed. Maximum allowed is 32 rules."),n&~(u<<t)|r<<t}(function(e){e[e["IgnoreRulesSpecific"]=0]="IgnoreRulesSpecific",e[e["IgnoreRulesAny"]=1*l]="IgnoreRulesAny",e[e["ContextRulesSpecific"]=2*l]="ContextRulesSpecific",e[e["ContextRulesAny"]=3*l]="ContextRulesAny",e[e["NoContextRulesSpecific"]=4*l]="NoContextRulesSpecific",e[e["NoContextRulesAny"]=5*l]="NoContextRulesAny"})(c||(c={}))})(e.formatting||(e.formatting={}))}(u||(u={})),function(e){(function(n){var t,r,a,i,o;function s(n,t,r){var a=t.getLineAndCharacterOfPosition(n).line;if(0===a)return[];var i=e.getEndLinePosition(a,t);while(e.isWhiteSpaceSingleLine(t.text.charCodeAt(i)))i--;e.isLineBreak(t.text.charCodeAt(i))&&i--;var o={pos:e.getStartPositionOfLine(a-1,t),end:i+1};return T(o,t,r,2)}function c(e,n,t){var r=m(e,26,n);return E(f(r),n,t,3)}function l(n,t,r){var a=m(n,18,t);if(!a)return[];var i=a.parent,o=f(i),s={pos:e.getLineStartPositionForPosition(o.getStart(t),t),end:n};return T(s,t,r,4)}function u(e,n,t){var r=m(e,19,n);return E(f(r),n,t,5)}function d(e,n){var t={pos:0,end:e.text.length};return T(t,e,n,0)}function p(n,t,r,a){var i={pos:e.getLineStartPositionForPosition(n,r),end:t};return T(i,r,a,1)}function m(n,t,r){var a=e.findPrecedingToken(n,r);return a&&a.kind===t&&n===a.getEnd()?a:void 0}function f(e){var n=e;while(n&&n.parent&&n.parent.end===e.end&&!_(n.parent,n))n=n.parent;return n}function _(n,t){switch(n.kind){case 240:case 241:return e.rangeContainsRange(n.members,t);case 244:var r=n.body;return!!r&&245===r.kind&&e.rangeContainsRange(r.statements,t);case 279:case 218:case 245:return e.rangeContainsRange(n.statements,t);case 274:return e.rangeContainsRange(n.block.statements,t)}return!1}function g(n,t){return r(t);function r(a){var i=e.forEachChild(a,function(r){return e.startEndContainsRange(r.getStart(t),r.end,n)&&r});if(i){var o=r(i);if(o)return o}return a}}function v(n,t){if(!n.length)return i;var r=n.filter(function(n){return e.rangeOverlapsWithStartEnd(t,n.start,n.start+n.length)}).sort(function(e,n){return e.start-n.start});if(!r.length)return i;var a=0;return function(n){while(1){if(a>=r.length)return!1;var t=r[a];if(n.end<=t.start)return!1;if(e.startEndOverlapsWithStartEnd(n.pos,n.end,t.start,t.start+t.length))return!0;a++}};function i(){return!1}}function y(n,t,r){var a=n.getStart(r);if(a===t.pos&&n.end===t.end)return a;var i=e.findPrecedingToken(t.pos,r);return i?i.end>=t.pos?n.pos:i.end:n.pos}function h(e,t,r){var a,i=-1;while(e){var o=r.getLineAndCharacterOfPosition(e.getStart(r)).line;if(-1!==i&&o!==i)break;if(n.SmartIndenter.shouldIndentChildNode(t,e,a,r))return t.indentSize;i=o,a=e,e=e.parent}return 0}function b(e,t,r,a,i,o){var s={pos:0,end:t.text.length};return n.getFormattingScanner(t.text,r,s.pos,s.end,function(n){return S(s,e,a,i,n,o,1,function(e){return!1},t)})}function E(n,t,r,a){if(!n)return[];var i={pos:e.getLineStartPositionForPosition(n.getStart(t),t),end:n.end};return T(i,t,r,a)}function T(e,t,r,a){var i=g(e,t);return n.getFormattingScanner(t.text,t.languageVariant,y(i,e,t),e.end,function(o){return S(e,i,n.SmartIndenter.getIndentationForNode(i,e,t,r.options),h(i,r.options,t),o,r,a,v(t.parseDiagnostics,e),t)})}function S(t,r,a,i,o,s,c,l,u){var d,p,m,f,_=s.options,g=s.getRule,v=new n.FormattingContext(u,c,_),y=-1,h=[];if(o.advance(),o.isOnToken()){var b=u.getLineAndCharacterOfPosition(r.getStart(u)).line,E=b;r.decorators&&(E=u.getLineAndCharacterOfPosition(e.getNonDecoratorTokenPosOfNode(r,u)).line),N(r,r,b,E,a,i)}if(!o.isOnToken()){var T=o.getCurrentLeadingTrivia();T&&(I(T,a,!1,function(e){return M(e,u.getLineAndCharacterOfPosition(e.pos),r,r,void 0)}),K())}return h;function S(t,r,a,i,o){if(e.rangeOverlapsWithStartEnd(i,t,r)||e.rangeContainsStartEnd(i,t,r)){if(-1!==o)return o}else{var s=u.getLineAndCharacterOfPosition(t).line,c=e.getLineStartPositionForPosition(t,u),l=n.SmartIndenter.findFirstNonWhitespaceColumn(c,t,u,_);if(s!==a||t===l){var d=n.SmartIndenter.getBaseIndentation(_);return d>l?d:l}}return-1}function x(e,t,r,a,i,o){var s=n.SmartIndenter.shouldIndentChildNode(_,e)?_.indentSize:0;return o===t?{indentation:t===f?y:i.getIndentation(),delta:Math.min(_.indentSize,i.getDelta(e)+s)}:-1===r?20===e.kind&&t===f?{indentation:y,delta:i.getDelta(e)}:n.SmartIndenter.childStartsOnTheSameLineWithElseInIfStatement(a,e,t,u)?{indentation:i.getIndentation(),delta:s}:{indentation:i.getIndentation()+i.getDelta(e),delta:s}:{indentation:r,delta:s}}function L(n){if(n.modifiers&&n.modifiers.length)return n.modifiers[0].kind;switch(n.kind){case 240:return 76;case 241:return 110;case 239:return 90;case 243:return 243;case 158:return 126;case 159:return 137;case 156:if(n.asteriskToken)return 40;case 154:case 151:var t=e.getNameOfDeclaration(n);if(t)return t.kind}}function k(e,t,r,a){return{getIndentationForComment:function(e,n,t){switch(e){case 19:case 23:case 21:return r+o(t)}return-1!==n?n:r},getIndentationForToken:function(e,n,t,a){return!a&&i(e,n,t)?r+o(t):r},getIndentation:function(){return r},getDelta:o,recomputeIndentation:function(t){e.parent&&n.SmartIndenter.shouldIndentChildNode(_,e.parent,e,u)&&(r+=t?_.indentSize:-_.indentSize,a=n.SmartIndenter.shouldIndentChildNode(_,e)?_.indentSize:0)}};function i(n,r,a){switch(r){case 18:case 19:case 21:case 83:case 107:case 58:return!1;case 42:case 30:switch(a.kind){case 262:case 263:case 261:return!1}break;case 22:case 23:if(181!==a.kind)return!1;break}return t!==n&&!(e.decorators&&r===L(e))}function o(t){return n.SmartIndenter.nodeWillIndentChild(_,e,t,u,!0)?a:0}}function N(r,a,i,s,c,p){if(e.rangeOverlapsWithStartEnd(t,r.getStart(u),r.getEnd())){var m=k(r,i,c,p),g=a;e.forEachChild(r,function(e){h(e,-1,r,m,i,s,!1)},function(e){b(e,r,i,m)});while(o.isOnToken()){var v=o.readTokenInfo(r);if(v.token.end>r.end)break;E(v,r,m,r)}}function h(n,a,i,s,c,l,d,p){var m=n.getStart(u),f=u.getLineAndCharacterOfPosition(m).line,_=f;n.decorators&&(_=u.getLineAndCharacterOfPosition(e.getNonDecoratorTokenPosOfNode(n,u)).line);var v=-1;if(d&&e.rangeContainsRange(t,i)&&(v=S(m,n.end,c,t,a),-1!==v&&(a=v)),!e.rangeOverlapsWithStartEnd(t,n.pos,n.end))return n.end<t.pos&&o.skipToEndOf(n),a;if(0===n.getFullWidth())return a;while(o.isOnToken()){var y=o.readTokenInfo(r);if(y.token.end>m)break;E(y,r,s,r)}if(!o.isOnToken())return a;if(e.isToken(n)&&11!==n.kind){y=o.readTokenInfo(n);return e.Debug.assert(y.token.end===n.end,"Token end is child end"),E(y,r,s,n),a}var h=152===n.kind?f:l,b=x(n,f,v,r,s,h);if(N(n,g,f,_,b.indentation,b.delta),11===n.kind){var T={pos:n.getStart(),end:n.getEnd()};G(T,b.indentation,!0,!1)}return g=r,p&&187===i.kind&&-1===a&&(a=b.indentation),a}function b(t,a,i,s){e.Debug.assert(e.isNodeArray(t));var c=C(a,t),l=s,d=i;if(0!==c)while(o.isOnToken()){var p=o.readTokenInfo(a);if(p.token.end>t.pos)break;if(p.token.kind===c){d=u.getLineAndCharacterOfPosition(p.token.pos).line,E(p,a,s,a);var m=void 0;if(-1!==y)m=y;else{var f=e.getLineStartPositionForPosition(p.token.pos,u);m=n.SmartIndenter.findFirstNonWhitespaceColumn(f,p.token.pos,u,_)}l=k(a,i,m,_.indentSize)}else E(p,a,s,a)}for(var g=-1,v=0;v<t.length;v++){var b=t[v];g=h(b,g,r,l,d,d,!0,0===v)}var T=A(c);if(0!==T&&o.isOnToken()){p=o.readTokenInfo(a);27===p.token.kind&&e.isCallLikeExpression(a)&&(o.advance(),p=o.isOnToken()?o.readTokenInfo(a):void 0),p&&p.token.kind===T&&e.rangeContainsRange(a,p.token)&&E(p,a,l,a,!0)}}function E(n,r,a,i,s){e.Debug.assert(e.rangeContainsRange(r,n.token));var c=o.lastTrailingTriviaWasNewLine(),p=!1;n.leadingTrivia&&w(n.leadingTrivia,r,g,a);var m=0,_=e.rangeContainsRange(t,n.token),v=u.getLineAndCharacterOfPosition(n.token.pos);if(_){var h=l(n.token),b=d;if(m=M(n.token,v,r,g,a),!h)if(0===m){var E=b&&u.getLineAndCharacterOfPosition(b.end).line;p=c&&v.line!==E}else p=1===m}if(n.trailingTrivia&&w(n.trailingTrivia,r,g,a),p){var T=_&&!l(n.token)?a.getIndentationForToken(v.line,n.token.kind,i,!!s):-1,S=!0;if(n.leadingTrivia){var x=a.getIndentationForComment(n.token.kind,T,i);S=I(n.leadingTrivia,x,S,function(e){return P(e.pos,x,!1)})}-1!==T&&S&&(P(n.token.pos,T,1===m),f=v.line,y=T)}o.advance(),g=r}}function I(n,r,a,i){for(var o=0,s=n;o<s.length;o++){var c=s[o],l=e.rangeContainsRange(t,c);switch(c.kind){case 3:l&&G(c,r,!a),a=!1;break;case 2:a&&l&&i(c),a=!1;break;case 4:a=!0;break}}return a}function w(n,r,a,i){for(var o=0,s=n;o<s.length;o++){var c=s[o];if(e.isComment(c.kind)&&e.rangeContainsRange(t,c)){var l=u.getLineAndCharacterOfPosition(c.pos);M(c,l,r,a,i)}}}function M(e,n,r,a,i){var o=l(e),s=0;if(!o)if(d)s=O(e,n.line,r,d,m,p,a,i);else{var c=u.getLineAndCharacterOfPosition(t.pos);V(c.line,n.line)}return d=e,p=r,m=n.line,s}function O(n,t,r,a,i,o,s,c){v.updateContext(a,o,n,r,s);var l,d=g(v),p=0;if(d){switch(p=H(d,a,i,n,t),p){case 2:r.getStart(u)===n.pos&&c.recomputeIndentation(!1);break;case 1:r.getStart(u)===n.pos&&c.recomputeIndentation(!0);break;default:e.Debug.assert(0===p)}l=!(8&d.action)&&1!==d.flags}else l=!0;return t!==i&&l&&V(i,t,a),p}function P(n,t,r){var a=D(t,_);if(r)U(n,0,a);else{var i=u.getLineAndCharacterOfPosition(n),o=e.getStartPositionOfLine(i.line,u);(t!==R(o,i.character)||F(a,o))&&U(o,i.character,a)}}function R(e,n){for(var t=0,r=0;r<n;r++)9===u.text.charCodeAt(e+r)?t+=_.tabSize-t%_.tabSize:t++;return t}function F(e,n){return e!==u.text.substr(n,e.length)}function G(t,r,a,i){void 0===i&&(i=!0);var o=u.getLineAndCharacterOfPosition(t.pos).line,s=u.getLineAndCharacterOfPosition(t.end).line;if(o!==s){for(var c=[],l=t.pos,d=o;d<s;d++){var p=e.getEndLinePosition(d,u);c.push({pos:l,end:p}),l=e.getStartPositionOfLine(d+1,u)}if(i&&c.push({pos:l,end:t.end}),0!==c.length){var m=e.getStartPositionOfLine(o,u),f=n.SmartIndenter.findFirstNonWhitespaceCharacterAndColumn(m,c[0].pos,u,_);if(r!==f.column){var g=0;a&&(g=1,o++);for(var v=r-f.column,y=g;y<c.length;y++,o++){var h=e.getStartPositionOfLine(o,u),b=0===y?f:n.SmartIndenter.findFirstNonWhitespaceCharacterAndColumn(c[y].pos,c[y].end,u,_),E=b.column+v;if(E>0){var T=D(E,_);U(h,b.character,T)}else j(h,b.character)}}}}else a||P(t.pos,r,!1)}function V(n,t,r){for(var a=n;a<t;a++){var i=e.getStartPositionOfLine(a,u),o=e.getEndLinePosition(a,u);if(!(r&&(e.isComment(r.kind)||e.isStringOrRegularExpressionOrTemplateLiteral(r.kind))&&r.pos<=o&&r.end>o)){var s=B(i,o);-1!==s&&(e.Debug.assert(s===i||!e.isWhiteSpaceSingleLine(u.text.charCodeAt(s-1))),j(s,o+1-s))}}}function B(n,t){var r=t;while(r>=n&&e.isWhiteSpaceSingleLine(u.text.charCodeAt(r)))r--;return r!==t?r+1:-1}function K(){var e=d?d.end:t.pos,n=u.getLineAndCharacterOfPosition(e).line,r=u.getLineAndCharacterOfPosition(t.end).line;V(n,r+1,d)}function j(n,t){t&&h.push(e.createTextChangeFromStartLength(n,t,""))}function U(n,t,r){(t||r)&&h.push(e.createTextChangeFromStartLength(n,t,r))}function H(e,n,t,r,a){var i=a!==t;switch(e.action){case 1:return 0;case 8:if(n.end!==r.pos)return j(n.end,r.pos-n.end),i?2:0;break;case 4:if(1!==e.flags&&t!==a)return 0;var o=a-t;if(1!==o)return U(n.end,r.pos-n.end,_.newLineCharacter),i?0:1;break;case 2:if(1!==e.flags&&t!==a)return 0;var s=r.pos-n.end;if(1!==s||32!==u.text.charCodeAt(n.end))return U(n.end,r.pos-n.end," "),i?2:0}return 0}}function x(n,t,r,a){void 0===a&&(a=e.getTokenAtPosition(n,t));var i=e.findAncestor(a,e.isJSDoc);i&&(a=i.parent);var o=a.getStart(n);if(!(o<=t&&t<a.getEnd())){r=null===r?void 0:void 0===r?e.findPrecedingToken(t,n):r;var s=r&&e.getTrailingCommentRanges(n.text,r.end),c=e.getLeadingCommentRangesOfNode(a,n),l=e.concatenate(s,c);return l&&e.find(l,function(r){return e.rangeContainsPositionExclusive(r,t)||t===r.end&&(2===r.kind||t===n.getFullWidth())})}}function C(e,n){switch(e.kind){case 157:case 239:case 196:case 156:case 155:case 197:if(e.typeParameters===n)return 28;if(e.parameters===n)return 20;break;case 191:case 192:if(e.typeArguments===n)return 28;if(e.arguments===n)return 20;break;case 164:if(e.typeArguments===n)return 28;break;case 168:return 18}return 0}function A(e){switch(e){case 20:return 21;case 28:return 30;case 18:return 19}return 0}function D(n,t){var r=!a||a.tabSize!==t.tabSize||a.indentSize!==t.indentSize;if(r&&(a={tabSize:t.tabSize,indentSize:t.indentSize},i=o=void 0),t.convertTabsToSpaces){var s=void 0,c=Math.floor(n/t.indentSize),l=n%t.indentSize;return o||(o=[]),void 0===o[c]?(s=e.repeatString(" ",t.indentSize*c),o[c]=s):s=o[c],l?s+e.repeatString(" ",l):s}var u=Math.floor(n/t.tabSize),d=n-u*t.tabSize,p=void 0;return i||(i=[]),void 0===i[u]?i[u]=p=e.repeatString("\t",u):p=i[u],d?p+e.repeatString(" ",d):p}(function(e){e[e["Unknown"]=-1]="Unknown"})(t||(t={})),n.formatOnEnter=s,n.formatOnSemicolon=c,n.formatOnOpeningCurly=l,n.formatOnClosingCurly=u,n.formatDocument=d,n.formatSelection=p,n.formatNodeGivenIndentation=b,function(e){e[e["None"]=0]="None",e[e["LineAdded"]=1]="LineAdded",e[e["LineRemoved"]=2]="LineRemoved"}(r||(r={})),n.getRangeOfEnclosingComment=x,n.getIndentationString=D})(e.formatting||(e.formatting={}))}(u||(u={})),function(e){(function(n){(function(t){var r,a;function i(t,r,a,i){if(void 0===i&&(i=!1),t>r.text.length)return u(a);if(a.indentStyle===e.IndentStyle.None)return 0;var l=e.findPrecedingToken(t,r,void 0,!0),d=n.getRangeOfEnclosingComment(r,t,l||null);if(d&&3===d.kind)return o(r,t,a,d);if(!l)return u(a);var p=e.isStringOrRegularExpressionOrTemplateLiteral(l.kind);if(p&&l.getStart(r)<=t&&t<l.end)return 0;var f=r.getLineAndCharacterOfPosition(t).line;if(a.indentStyle===e.IndentStyle.Block)return s(r,t,a);if(27===l.kind&&204!==l.parent.kind){var _=m(l,r,a);if(-1!==_)return _}var g=b(t,l.parent,r);return g&&!e.rangeContainsRange(g,l)?S(g,r,a)+a.indentSize:c(r,t,l,f,i,a)}function o(n,t,r,a){var i=e.getLineAndCharacterOfPosition(n,t).line-1,o=e.getLineAndCharacterOfPosition(n,a.pos).line;if(e.Debug.assert(o>=0),i<=o)return L(e.getStartPositionOfLine(o,n),t,n,r);var s=e.getStartPositionOfLine(i,n),c=D(s,t,n,r),l=c.column,u=c.character;if(0===l)return l;var d=n.text.charCodeAt(s+u);return 42===d?l-1:l}function s(n,t,r){var a=t;while(a>0){var i=n.text.charCodeAt(a);if(!e.isWhiteSpaceLike(i))break;a--}var o=e.getLineStartPositionForPosition(a,n);return L(o,a,n,r)}function c(n,t,r,a,i,o){var s,c=r;while(c){if(e.positionBelongsToNode(c,t,n)&&I(o,c,s,n,!0)){var l=g(c,n),p=_(r,c,a,n),m=0!==p?i&&2===p?o.indentSize:0:a!==l.line?o.indentSize:0;return d(c,l,void 0,m,n,!0,o)}var f=x(c,n,o,!0);if(-1!==f)return f;s=c,c=c.parent}return u(o)}function l(e,n,t,r){var a=t.getLineAndCharacterOfPosition(e.getStart(t));return d(e,a,n,0,t,!1,r)}function u(e){return e.baseIndentSize||0}function d(e,n,t,r,a,i,o){var s=e.parent;while(s){var c=!0;if(t){var l=e.getStart(a);c=l<t.pos||l>t.end}var d=p(s,e,a),m=d.line===n.line||y(s,e,n.line,a);if(c){var _=x(e,a,o,!m);if(-1!==_)return _+r;if(_=f(e,s,n,m,a,o),-1!==_)return _+r}I(o,s,e,a,i)&&!m&&(r+=o.indentSize);var g=v(s,e,n.line,a);e=s,s=e.parent,n=g?a.getLineAndCharacterOfPosition(e.getStart(a)):d}return r+u(o)}function p(e,n,t){var r=h(n,t),a=r?r.pos:e.getStart(t);return t.getLineAndCharacterOfPosition(a)}function m(n,t,r){var a=e.findListItemInfo(n);return a&&a.listItemIndex>0?C(a.list.getChildren(),a.listItemIndex-1,t,r):-1}function f(n,t,r,a,i,o){var s=(e.isDeclaration(n)||e.isStatementButNotDeclaration(n))&&(279===t.kind||!a);return s?A(r,i,o):-1}function _(n,t,r,a){var i=e.findNextToken(n,t,a);if(!i)return 0;if(18===i.kind)return 1;if(19===i.kind){var o=g(i,a).line;return r===o?2:0}return 0}function g(e,n){return n.getLineAndCharacterOfPosition(e.getStart(n))}function v(n,t,r,a){if(!e.isCallExpression(n)||!e.contains(n.arguments,t))return!1;var i=n.expression.getEnd(),o=e.getLineAndCharacterOfPosition(a,i).line;return o===r}function y(n,t,r,a){if(222===n.kind&&n.elseStatement===t){var i=e.findChildOfKind(n,83,a);e.Debug.assert(void 0!==i);var o=g(i,a).line;return o===r}return!1}function h(e,n){return e.parent&&E(e.getStart(n),e.getEnd(),e.parent,n)}function b(e,n,t){return n&&E(e,e,n,t)}function E(n,t,r,a){switch(r.kind){case 164:return i(r.typeArguments);case 188:return i(r.properties);case 187:return i(r.elements);case 168:return i(r.members);case 239:case 196:case 197:case 156:case 155:case 160:case 157:case 166:case 161:return i(r.typeParameters)||i(r.parameters);case 240:case 209:case 241:case 242:case 303:return i(r.typeParameters);case 192:case 191:return i(r.typeArguments)||i(r.arguments);case 238:return i(r.declarations);case 252:case 256:return i(r.elements);case 184:case 185:return i(r.elements)}function i(i){return i&&e.rangeContainsStartEnd(T(r,i,a),n,t)?i:void 0}}function T(e,n,t){for(var r=e.getChildren(t),a=1;a<r.length-1;a++)if(r[a].pos===n.pos&&r[a].end===n.end)return{pos:r[a-1].end,end:r[a+1].getStart(t)};return n}function S(e,n,t){return e?A(n.getLineAndCharacterOfPosition(e.pos),n,t):-1}function x(e,n,t,r){if(e.parent&&238===e.parent.kind)return-1;var a=h(e,n);if(a){var i=a.indexOf(e);if(-1!==i){var o=C(a,i,n,t);if(-1!==o)return o}return S(a,n,t)+(r?t.indentSize:0)}return-1}function C(n,t,r,a){e.Debug.assert(t>=0&&t<n.length);for(var i=n[t],o=g(i,r),s=t-1;s>=0;s--)if(27!==n[s].kind){var c=r.getLineAndCharacterOfPosition(n[s].end).line;if(c!==o.line)return A(o,r,a);o=g(n[s],r)}return-1}function A(e,n,t){var r=n.getPositionOfLineAndCharacter(e.line,0);return L(r,r+e.character,n,t)}function D(n,t,r,a){for(var i=0,o=0,s=n;s<t;s++){var c=r.text.charCodeAt(s);if(!e.isWhiteSpaceSingleLine(c))break;9===c?o+=a.tabSize+o%a.tabSize:o++,i++}return{column:o,character:i}}function L(e,n,t,r){return D(e,n,t,r).column}function k(e,n,t,r,a){var i=t?t.kind:0;switch(n.kind){case 221:case 240:case 209:case 241:case 243:case 242:case 187:case 218:case 245:case 188:case 168:case 181:case 170:case 246:case 272:case 271:case 195:case 189:case 191:case 192:case 219:case 254:case 230:case 205:case 185:case 184:case 262:case 265:case 261:case 270:case 155:case 160:case 161:case 151:case 165:case 166:case 177:case 193:case 201:case 256:case 252:case 257:case 253:case 154:return!0;case 237:case 275:return!(!e.indentMultiLineObjectLiteralBeginningOnBlankLine&&r&&188===i)||w(r,t);case 223:case 224:case 226:case 227:case 225:case 222:case 239:case 196:case 156:case 197:case 157:case 158:case 159:return 218!==i;case 255:return 256!==i;case 249:return 250!==i||!!t.namedBindings&&252!==t.namedBindings.kind;case 260:return 263!==i;case 264:return 266!==i;case 174:case 173:if(168===i)return!1}return a}function N(e,n){switch(e){case 230:case 234:case 228:case 229:return 218!==n.kind;default:return!1}}function I(e,n,t,r,a){return void 0===a&&(a=!1),k(e,n,t,r,!1)&&!(a&&t&&N(t.kind,n))}function w(n,t){var r=e.skipTrivia(n.text,t.pos),a=n.getLineAndCharacterOfPosition(r).line,i=n.getLineAndCharacterOfPosition(t.end).line;return a===i}(function(e){e[e["Unknown"]=-1]="Unknown"})(r||(r={})),t.getIndentation=i,t.getIndentationForNode=l,t.getBaseIndentation=u,function(e){e[e["Unknown"]=0]="Unknown",e[e["OpenBrace"]=1]="OpenBrace",e[e["CloseBrace"]=2]="CloseBrace"}(a||(a={})),t.isArgumentAndStartLineOverlapsExpressionBeingCalled=v,t.childStartsOnTheSameLineWithElseInIfStatement=y,t.getContainingList=h,t.findFirstNonWhitespaceCharacterAndColumn=D,t.findFirstNonWhitespaceColumn=L,t.nodeWillIndentChild=k,t.shouldIndentChildNode=I})(n.SmartIndenter||(n.SmartIndenter={}))})(e.formatting||(e.formatting={}))}(u||(u={})),function(e){(function(n){function t(n){var t=n.__pos;return e.Debug.assert("number"===typeof t),t}function r(n,t){e.Debug.assert("number"===typeof t),n.__pos=t}function a(n){var t=n.__end;return e.Debug.assert("number"===typeof t),t}function i(n,t){e.Debug.assert("number"===typeof t),n.__end=t}var s,c;function l(n,t){return e.skipTrivia(n,t,!1,!0)}function u(n,t){var r=t;while(r<n.length){var a=n.charCodeAt(r);if(!e.isWhiteSpaceSingleLine(a))return 47===a;r++}return!1}function d(e,n,t,r){return{pos:p(e,n,r,s.Start),end:m(e,t,r)}}function p(n,t,r,a){if(r.useNonAdjustedStartPosition)return t.getStart(n);var i=t.getFullStart(),o=t.getStart(n);if(i===o)return o;var c=e.getLineStartPositionForPosition(i,n),u=e.getLineStartPositionForPosition(o,n);if(u===c)return a===s.Start?o:i;var d=i>0?1:0,p=e.getStartPositionOfLine(e.getLineOfLocalPosition(n,c)+d,n);return p=l(n.text,p),e.getStartPositionOfLine(e.getLineOfLocalPosition(n,p),n)}function m(n,t,r){var a=t.end;if(r.useNonAdjustedEndPosition||e.isExpression(t))return a;var i=e.skipTrivia(n.text,a,!0);return i!==a&&e.isLineBreak(n.text.charCodeAt(i-1))?i:a}function f(e,n){return!!n&&!!e.parent&&(27===n.kind||26===n.kind&&188===e.parent.kind)}function _(e){for(var n="",t=0;t<e;t++)n+=" ";return n}(function(e){e[e["FullStart"]=0]="FullStart",e[e["Start"]=1]="Start"})(s=n.Position||(n.Position={})),n.useNonAdjustedPositions={useNonAdjustedStartPosition:!0,useNonAdjustedEndPosition:!0},function(e){e[e["Remove"]=0]="Remove",e[e["ReplaceWithSingleNode"]=1]="ReplaceWithSingleNode",e[e["ReplaceWithMultipleNodes"]=2]="ReplaceWithMultipleNodes",e[e["Text"]=3]="Text"}(c||(c={}));var g,v=function(){function t(n,t){this.newLineCharacter=n,this.formatContext=t,this.changes=[],this.newFiles=[],this.classesWithNodesInsertedAtStart=e.createMap(),this.deletedNodes=[]}return t.fromContext=function(n){return new t(e.getNewLineOrDefaultFromHost(n.host,n.formatContext.options),n.formatContext)},t.with=function(e,n){var r=t.fromContext(e);return n(r),r.getChanges()},t.prototype.deleteRange=function(e,n){this.changes.push({kind:c.Remove,sourceFile:e,range:n})},t.prototype.delete=function(e,n){this.deletedNodes.push({sourceFile:e,node:n})},t.prototype.deleteModifier=function(n,t){this.deleteRange(n,{pos:t.getStart(n),end:e.skipTrivia(n.text,t.end,!0)})},t.prototype.deleteNodeRange=function(e,n,t,r){void 0===r&&(r={});var a=p(e,n,r,s.FullStart),i=m(e,t,r);this.deleteRange(e,{pos:a,end:i})},t.prototype.deleteNodeRangeExcludingEnd=function(e,n,t,r){void 0===r&&(r={});var a=p(e,n,r,s.FullStart),i=void 0===t?e.text.length:p(e,t,r,s.FullStart);this.deleteRange(e,{pos:a,end:i})},t.prototype.replaceRange=function(e,n,t,r){void 0===r&&(r={}),this.changes.push({kind:c.ReplaceWithSingleNode,sourceFile:e,range:n,options:r,node:t})},t.prototype.replaceNode=function(e,t,r,a){void 0===a&&(a=n.useNonAdjustedPositions),this.replaceRange(e,d(e,t,t,a),r,a)},t.prototype.replaceNodeRange=function(e,t,r,a,i){void 0===i&&(i=n.useNonAdjustedPositions),this.replaceRange(e,d(e,t,r,i),a,i)},t.prototype.replaceRangeWithNodes=function(e,n,t,r){void 0===r&&(r={}),this.changes.push({kind:c.ReplaceWithMultipleNodes,sourceFile:e,range:n,options:r,nodes:t})},t.prototype.replaceNodeWithNodes=function(e,t,r,a){void 0===a&&(a=n.useNonAdjustedPositions),this.replaceRangeWithNodes(e,d(e,t,t,a),r,a)},t.prototype.replaceNodeWithText=function(e,t,r){this.replaceRangeWithText(e,d(e,t,t,n.useNonAdjustedPositions),r)},t.prototype.replaceNodeRangeWithNodes=function(e,t,r,a,i){void 0===i&&(i=n.useNonAdjustedPositions),this.replaceRangeWithNodes(e,d(e,t,r,i),a,i)},t.prototype.nextCommaToken=function(n,t){var r=e.findNextToken(t,t.parent,n);return r&&27===r.kind?r:void 0},t.prototype.replacePropertyAssignment=function(e,n,t){var r=this.nextCommaToken(e,n)?"":","+this.newLineCharacter;this.replaceNode(e,n,t,{suffix:r})},t.prototype.insertNodeAt=function(n,t,r,a){void 0===a&&(a={}),this.replaceRange(n,e.createRange(t),r,a)},t.prototype.insertNodesAt=function(n,t,r,a){void 0===a&&(a={}),this.replaceRangeWithNodes(n,e.createRange(t),r,a)},t.prototype.insertNodeAtTopOfFile=function(n,t,r){var a=L(n);this.insertNodeAt(n,a,t,{prefix:0===a?void 0:this.newLineCharacter,suffix:(e.isLineBreak(n.text.charCodeAt(a))?"":this.newLineCharacter)+(r?this.newLineCharacter:"")})},t.prototype.insertNodeBefore=function(e,n,t,r){void 0===r&&(r=!1),this.insertNodeAt(e,p(e,n,{},s.Start),t,this.getOptionsForInsertNodeBefore(n,r))},t.prototype.insertModifierBefore=function(n,t,r){var a=r.getStart(n);this.insertNodeAt(n,a,e.createToken(t),{suffix:" "})},t.prototype.insertLastModifierBefore=function(n,t,r){if(r.modifiers){var a=r.modifiers.end;this.insertNodeAt(n,a,e.createToken(t),{prefix:" "})}else this.insertModifierBefore(n,t,r)},t.prototype.insertCommentBeforeLine=function(n,t,r,a){var i=e.getStartPositionOfLine(t,n),o=e.getFirstNonSpaceCharacterPosition(n.text,i),s=k(n,o),c=e.getTouchingToken(n,s?o:r),l=n.text.slice(i,o),u=(s?"":this.newLineCharacter)+"//"+a+this.newLineCharacter+l;this.insertText(n,c.getStart(n),u)},t.prototype.insertJsdocCommentBefore=function(n,t,r){var a=t.getStart(n);if(t.jsDoc)for(var i=0,o=t.jsDoc;i<o.length;i++){var s=o[i];this.deleteRange(n,{pos:e.getLineStartPositionForPosition(s.getStart(n),n),end:m(n,s,{})})}var c=e.getPrecedingNonSpaceCharacterPosition(n.text,a-1),l=n.text.slice(c,a);this.insertNodeAt(n,a,r,{preserveLeadingWhitespace:!1,suffix:this.newLineCharacter+l})},t.prototype.replaceRangeWithText=function(e,n,t){this.changes.push({kind:c.Text,sourceFile:e,range:n,text:t})},t.prototype.insertText=function(n,t,r){this.replaceRangeWithText(n,e.createRange(t),r)},t.prototype.tryInsertTypeAnnotation=function(n,t,r){var a;if(e.isFunctionLike(t)){if(a=e.findChildOfKind(t,21,n),!a){if(!e.isArrowFunction(t))return;a=e.first(t.parameters)}}else a=237!==t.kind&&t.questionToken?t.questionToken:t.name;this.insertNodeAt(n,a.end,r,{prefix:": "})},t.prototype.insertTypeParameters=function(n,t,r){var a=(e.findChildOfKind(t,20,n)||e.first(t.parameters)).getStart(n);this.insertNodesAt(n,a,r,{prefix:"<",suffix:">"})},t.prototype.getOptionsForInsertNodeBefore=function(n,t){return e.isStatement(n)||e.isClassElement(n)?{suffix:t?this.newLineCharacter+this.newLineCharacter:this.newLineCharacter}:e.isVariableDeclaration(n)?{suffix:", "}:e.isParameter(n)?{}:e.isStringLiteral(n)&&e.isImportDeclaration(n.parent)||e.isNamedImports(n)?{suffix:", "}:e.Debug.failBadSyntaxKind(n)},t.prototype.insertNodeAtConstructorStart=function(n,t,r){var a=e.firstOrUndefined(t.body.statements);a&&t.body.multiLine?this.insertNodeBefore(n,a,r):this.replaceConstructorBody(n,t,[r].concat(t.body.statements))},t.prototype.insertNodeAtConstructorEnd=function(n,t,r){var a=e.lastOrUndefined(t.body.statements);a&&t.body.multiLine?this.insertNodeAfter(n,a,r):this.replaceConstructorBody(n,t,t.body.statements.concat([r]))},t.prototype.replaceConstructorBody=function(n,t,r){this.replaceNode(n,t.body,e.createBlock(r,!0))},t.prototype.insertNodeAtEndOfScope=function(n,t,r){var a=p(n,t.getLastToken(),{},s.Start);this.insertNodeAt(n,a,r,{prefix:e.isLineBreak(n.text.charCodeAt(t.getLastToken().pos))?this.newLineCharacter:this.newLineCharacter+this.newLineCharacter,suffix:this.newLineCharacter})},t.prototype.insertNodeAtClassStart=function(e,n,t){this.insertNodeAtStartWorker(e,n,t)},t.prototype.insertNodeAtObjectStart=function(e,n,t){this.insertNodeAtStartWorker(e,n,t)},t.prototype.insertNodeAtStartWorker=function(n,t,r){var a=t.getStart(n),i=e.formatting.SmartIndenter.findFirstNonWhitespaceColumn(e.getLineStartPositionForPosition(a,n),a,n,this.formatContext.options)+this.formatContext.options.indentSize;this.insertNodeAt(n,b(t).pos,r,o({indentation:i},this.getInsertNodeAtStartPrefixSuffix(n,t)))},t.prototype.getInsertNodeAtStartPrefixSuffix=function(n,t){var r=e.isObjectLiteralExpression(t)?",":"";if(0===b(t).length){if(e.addToSeen(this.classesWithNodesInsertedAtStart,e.getNodeId(t),{node:t,sourceFile:n})){var a=e.positionsAreOnSameLine.apply(void 0,h(t,n).concat([n]));return{prefix:this.newLineCharacter,suffix:r+(a?this.newLineCharacter:"")}}return{prefix:"",suffix:r+this.newLineCharacter}}return{prefix:this.newLineCharacter,suffix:r}},t.prototype.insertNodeAfterComma=function(e,n,t){var r=this.insertNodeAfterWorker(e,this.nextCommaToken(e,n)||n,t);this.insertNodeAt(e,r,t,this.getInsertNodeAfterOptions(e,n))},t.prototype.insertNodeAfter=function(e,n,t){var r=this.insertNodeAfterWorker(e,n,t);this.insertNodeAt(e,r,t,this.getInsertNodeAfterOptions(e,n))},t.prototype.insertNodeAtEndOfList=function(e,n,t){this.insertNodeAt(e,n.end,t,{prefix:", "})},t.prototype.insertNodesAfter=function(n,t,r){var a=this.insertNodeAfterWorker(n,t,e.first(r));this.insertNodesAt(n,a,r,this.getInsertNodeAfterOptions(n,t))},t.prototype.insertNodeAfterWorker=function(n,t,r){N(t,r)&&59!==n.text.charCodeAt(t.end-1)&&this.replaceRange(n,e.createRange(t.end),e.createToken(26));var a=m(n,t,{});return a},t.prototype.getInsertNodeAfterOptions=function(n,t){var r=this.getInsertNodeAfterOptionsWorker(t);return o({},r,{prefix:t.end===n.end&&e.isStatement(t)?r.prefix?"\n"+r.prefix:"\n":r.prefix})},t.prototype.getInsertNodeAfterOptionsWorker=function(n){switch(n.kind){case 240:case 244:return{prefix:this.newLineCharacter,suffix:this.newLineCharacter};case 237:case 10:case 72:return{prefix:", "};case 275:return{suffix:","+this.newLineCharacter};case 85:return{prefix:" "};case 151:return{};default:return e.Debug.assert(e.isStatement(n)||e.isClassOrTypeElement(n)),{suffix:this.newLineCharacter}}},t.prototype.insertName=function(n,t,r){if(e.Debug.assert(!t.name),197===t.kind){var a=e.findChildOfKind(t,37,n),i=e.findChildOfKind(t,20,n);i?(this.insertNodesAt(n,i.getStart(n),[e.createToken(90),e.createIdentifier(r)],{joiner:" "}),I(this,n,a)):(this.insertText(n,e.first(t.parameters).getStart(n),"function "+r+"("),this.replaceRange(n,a,e.createToken(21))),218!==t.body.kind&&(this.insertNodesAt(n,t.body.getStart(n),[e.createToken(18),e.createToken(97)],{joiner:" ",suffix:" "}),this.insertNodesAt(n,t.body.end,[e.createToken(26),e.createToken(19)],{joiner:" "}))}else{var o=e.findChildOfKind(t,196===t.kind?90:76,n).end;this.insertNodeAt(n,o,e.createIdentifier(r),{prefix:" "})}},t.prototype.insertExportModifier=function(e,n){this.insertText(e,n.getStart(e),"export ")},t.prototype.insertNodeInListAfter=function(n,t,r,a){if(void 0===a&&(a=e.formatting.SmartIndenter.getContainingList(t,n)),a){var i=e.indexOfNode(a,t);if(!(i<0)){var o=t.getEnd();if(i!==a.length-1){var s=e.getTokenAtPosition(n,t.end);if(s&&f(t,s)){var c=e.getLineAndCharacterOfPosition(n,l(n.text,a[i+1].getFullStart())),d=e.getLineAndCharacterOfPosition(n,s.end),p=void 0,m=void 0;d.line===c.line?(m=s.end,p=_(c.character-d.character)):m=e.getStartPositionOfLine(c.line,n);var g=""+e.tokenToString(s.kind)+n.text.substring(s.end,a[i+1].getStart(n));this.replaceRange(n,e.createRange(m,a[i+1].getStart(n)),r,{prefix:p,suffix:g})}}else{var v=t.getStart(n),y=e.getLineStartPositionForPosition(v,n),h=void 0,b=!1;if(1===a.length)h=27;else{var E=e.findPrecedingToken(t.pos,n);h=f(t,E)?E.kind:27;var T=e.getLineStartPositionForPosition(a[i-1].getStart(n),n);b=T!==y}if(u(n.text,t.end)&&(b=!0),b){this.replaceRange(n,e.createRange(o),e.createToken(h));var S=e.formatting.SmartIndenter.findFirstNonWhitespaceColumn(y,v,n,this.formatContext.options),x=e.skipTrivia(n.text,o,!0,!1);x!==o&&e.isLineBreak(n.text.charCodeAt(x-1))&&x--,this.replaceRange(n,e.createRange(x),r,{indentation:S,prefix:this.newLineCharacter})}else this.replaceRange(n,e.createRange(o),r,{prefix:e.tokenToString(h)+" "})}}}else e.Debug.fail("node is not a list element")},t.prototype.finishClassesWithNodesInsertedAtStart=function(){var n=this;this.classesWithNodesInsertedAtStart.forEach(function(t){var r=t.node,a=t.sourceFile,i=h(r,a),o=i[0],s=i[1];e.positionsAreOnSameLine(o,s,a)&&o!==s-1&&n.deleteRange(a,e.createRange(o,s-1))})},t.prototype.finishDeleteDeclarations=function(){for(var n=this,t=new e.NodeSet,r=function(n,r){a.deletedNodes.some(function(t){return t.sourceFile===n&&e.rangeContainsRangeExclusive(t.node,r)})||(e.isArray(r)?a.deleteRange(n,e.rangeOfTypeParameters(r)):A.deleteDeclaration(a,t,n,r))},a=this,i=0,o=this.deletedNodes;i<o.length;i++){var s=o[i],c=s.sourceFile,l=s.node;r(c,l)}t.forEach(function(r){var a=r.getSourceFile(),i=e.formatting.SmartIndenter.getContainingList(r,a);if(r===e.last(i)){var o=e.findLastIndex(i,function(e){return!t.has(e)},i.length-2);-1!==o&&n.deleteRange(a,{pos:i[o].end,end:y(a,i[o+1])})}})},t.prototype.getChanges=function(e){this.finishDeleteDeclarations(),this.finishClassesWithNodesInsertedAtStart();for(var n=g.getTextChangesFromChanges(this.changes,this.newLineCharacter,this.formatContext,e),t=0,r=this.newFiles;t<r.length;t++){var a=r[t],i=a.oldFile,o=a.fileName,s=a.statements;n.push(g.newFileChanges(i,o,s,this.newLineCharacter,this.formatContext))}return n},t.prototype.createNewFile=function(e,n,t){this.newFiles.push({oldFile:e,fileName:n,statements:t})},t}();function y(n,t){return e.skipTrivia(n.text,p(n,t,{},s.FullStart),!1,!0)}function h(n,t){return[e.findChildOfKind(n,18,t).end,e.findChildOfKind(n,19,t).end]}function b(n){return e.isObjectLiteralExpression(n)?n.properties:n.members}function E(e,n,t,r){return g.newFileChangesWorker(void 0,n,e,t,r)}function T(n,t){for(var r=t.length-1;r>=0;r--){var a=t[r],i=a.span,o=a.newText;n=""+n.substring(0,i.start)+o+n.substring(e.textSpanEnd(i))}return n}function S(n){return e.skipTrivia(n,0)===n.length}function x(n){var r=e.visitEachChild(n,x,e.nullTransformationContext,C,x),i=e.nodeIsSynthesized(r)?r:Object.create(r);return i.pos=t(n),i.end=a(n),i}function C(n,r,i,o,s){var c=e.visitNodes(n,r,i,o,s);if(!c)return c;var l=c===n?e.createNodeArray(c.slice(0)):c;return l.pos=t(n),l.end=a(n),l}n.ChangeTracker=v,n.getNewFileText=E,function(n){function t(n,t,r,a){return e.group(n,function(e){return e.sourceFile.path}).map(function(n){for(var o=n[0].sourceFile,s=e.stableSort(n,function(e,n){return e.range.pos-n.range.pos||e.range.end-n.range.end}),c=function(n){e.Debug.assert(s[n].range.end<=s[n+1].range.pos,"Changes overlap",function(){return JSON.stringify(s[n].range)+" and "+JSON.stringify(s[n+1].range)})},l=0;l<s.length-1;l++)c(l);var u=s.map(function(n){return e.createTextChange(e.createTextSpanFromRange(n.range),i(n,o,t,r,a))});return{fileName:o.fileName,textChanges:u}})}function r(n,t,r,i,o){var s=a(n,e.getScriptKindFromFileName(t),r,i,o);return{fileName:t,textChanges:[e.createTextChange(e.createTextSpan(0,0),s)],isNewFile:!0}}function a(n,t,r,a,i){var o=r.map(function(e){return s(e,n,a).text}).join(a),c=e.createSourceFile("any file name",o,6,!0,t),l=e.formatting.formatDocument(c,i);return T(o,l)+a}function i(n,t,r,a,i){if(n.kind===c.Remove)return"";if(n.kind===c.Text)return n.text;var s=n.options,l=void 0===s?{}:s,u=n.range.pos,d=function(e){return o(e,t,u,l,r,a,i)},p=n.kind===c.ReplaceWithMultipleNodes?n.nodes.map(function(n){return e.removeSuffix(d(n),r)}).join(n.options.joiner||r):d(n.node),m=l.preserveLeadingWhitespace||void 0!==l.indentation||e.getLineStartPositionForPosition(u,t)===u?p:p.replace(/^\s+/,"");return(l.prefix||"")+m+(l.suffix||"")}function o(n,t,r,a,i,o,c){var l=a.indentation,u=a.prefix,d=a.delta,p=s(n,t,i),m=p.node,f=p.text;c&&c(m,f);var _=o.options,g=void 0!==l?l:e.formatting.SmartIndenter.getIndentation(r,t,_,u===i||e.getLineStartPositionForPosition(r,t)===r);void 0===d&&(d=e.formatting.SmartIndenter.shouldIndentChildNode(o.options,n)&&_.indentSize||0);var v={text:f,getLineAndCharacterOfPosition:function(n){return e.getLineAndCharacterOfPosition(this,n)}},y=e.formatting.formatNodeGivenIndentation(m,v,t.languageVariant,g,d,o);return T(f,y)}function s(n,t,r){var a=new D(r),i="\n"===r?1:0;return e.createPrinter({newLine:i,neverAsciiEscape:!0},a).writeNode(4,n,t,a),{text:a.getText(),node:x(n)}}n.getTextChangesFromChanges=t,n.newFileChanges=r,n.newFileChangesWorker=a,n.getNonformattedText=s}(g||(g={})),n.applyChanges=T;var A,D=function(){function n(n){var t=this;this.lastNonTriviaPosition=0,this.writer=e.createTextWriter(n),this.onEmitNode=function(e,n,a){n&&r(n,t.lastNonTriviaPosition),a(e,n),n&&i(n,t.lastNonTriviaPosition)},this.onBeforeEmitNodeArray=function(e){e&&r(e,t.lastNonTriviaPosition)},this.onAfterEmitNodeArray=function(e){e&&i(e,t.lastNonTriviaPosition)},this.onBeforeEmitToken=function(e){e&&r(e,t.lastNonTriviaPosition)},this.onAfterEmitToken=function(e){e&&i(e,t.lastNonTriviaPosition)}}return n.prototype.setLastNonTriviaPosition=function(n,t){if(t||!S(n)){this.lastNonTriviaPosition=this.writer.getTextPos();var r=0;while(e.isWhiteSpaceLike(n.charCodeAt(n.length-r-1)))r++;this.lastNonTriviaPosition-=r}},n.prototype.write=function(e){this.writer.write(e),this.setLastNonTriviaPosition(e,!1)},n.prototype.writeComment=function(e){this.writer.writeComment(e)},n.prototype.writeKeyword=function(e){this.writer.writeKeyword(e),this.setLastNonTriviaPosition(e,!1)},n.prototype.writeOperator=function(e){this.writer.writeOperator(e),this.setLastNonTriviaPosition(e,!1)},n.prototype.writePunctuation=function(e){this.writer.writePunctuation(e),this.setLastNonTriviaPosition(e,!1)},n.prototype.writeTrailingSemicolon=function(e){this.writer.writeTrailingSemicolon(e),this.setLastNonTriviaPosition(e,!1)},n.prototype.writeParameter=function(e){this.writer.writeParameter(e),this.setLastNonTriviaPosition(e,!1)},n.prototype.writeProperty=function(e){this.writer.writeProperty(e),this.setLastNonTriviaPosition(e,!1)},n.prototype.writeSpace=function(e){this.writer.writeSpace(e),this.setLastNonTriviaPosition(e,!1)},n.prototype.writeStringLiteral=function(e){this.writer.writeStringLiteral(e),this.setLastNonTriviaPosition(e,!1)},n.prototype.writeSymbol=function(e,n){this.writer.writeSymbol(e,n),this.setLastNonTriviaPosition(e,!1)},n.prototype.writeLine=function(){this.writer.writeLine()},n.prototype.increaseIndent=function(){this.writer.increaseIndent()},n.prototype.decreaseIndent=function(){this.writer.decreaseIndent()},n.prototype.getText=function(){return this.writer.getText()},n.prototype.rawWrite=function(e){this.writer.rawWrite(e),this.setLastNonTriviaPosition(e,!1)},n.prototype.writeLiteral=function(e){this.writer.writeLiteral(e),this.setLastNonTriviaPosition(e,!0)},n.prototype.getTextPos=function(){return this.writer.getTextPos()},n.prototype.getLine=function(){return this.writer.getLine()},n.prototype.getColumn=function(){return this.writer.getColumn()},n.prototype.getIndent=function(){return this.writer.getIndent()},n.prototype.isAtStartOfLine=function(){return this.writer.isAtStartOfLine()},n.prototype.clear=function(){this.writer.clear(),this.lastNonTriviaPosition=0},n}();function L(n){for(var t,r=0,a=n.statements;r<a.length;r++){var i=a[r];if(!e.isPrologueDirective(i))break;t=i}var o=0,s=n.text;if(t)return o=t.end,m(),o;var c=e.getShebang(s);void 0!==c&&(o=c.length,m());var l=e.getLeadingCommentRanges(s,o);if(!l)return o;l.length&&3===l[0].kind&&e.isPinnedComment(s,l[0].pos)&&(o=l[0].end,m(),l=l.slice(1));for(var u=0,d=l;u<d.length;u++){var p=d[u];if(2!==p.kind||!e.isRecognizedTripleSlashComment(s,p.pos,p.end))break;o=p.end,m()}return o;function m(){if(o<s.length){var n=s.charCodeAt(o);e.isLineBreak(n)&&(o++,o<s.length&&13===n&&10===s.charCodeAt(o)&&o++)}}}function k(n,t){return!e.isInComment(n,t)&&!e.isInString(n,t)&&!e.isInTemplateString(n,t)&&!e.isInJSXText(n,t)}function N(n,t){return(e.isPropertySignature(n)||e.isPropertyDeclaration(n))&&e.isClassOrTypeElement(t)&&149===t.name.kind||e.isStatementButNotDeclaration(n)&&e.isStatementButNotDeclaration(t)}function I(e,n,t,r){void 0===r&&(r={});var a=p(n,t,r,s.FullStart),i=m(n,t,r);e.deleteRange(n,{pos:a,end:i})}function w(n,t,r,a){var i=e.Debug.assertDefined(e.formatting.SmartIndenter.getContainingList(a,r)),o=e.indexOfNode(i,a);e.Debug.assert(-1!==o),1!==i.length?(e.Debug.assert(!t.has(a),"Deleting a node twice"),t.add(a),n.deleteRange(r,{pos:y(r,a),end:o===i.length-1?m(r,a,{}):y(r,i[o+1])})):I(n,r,a)}n.isValidLocationToAddComment=k,function(n){function t(n,t,o,s){switch(s.kind){case 151:var c=s.parent;e.isArrowFunction(c)&&1===c.parameters.length&&!e.findChildOfKind(c,20,o)?n.replaceNodeWithText(o,s,"()"):w(n,t,o,s);break;case 249:I(n,o,s,s===o.imports[0].parent?{useNonAdjustedStartPosition:!0,useNonAdjustedEndPosition:!1}:void 0);break;case 186:var l=s.parent,u=185===l.kind&&s!==e.last(l.elements);u?I(n,o,s):w(n,t,o,s);break;case 237:i(n,t,o,s);break;case 150:w(n,t,o,s);break;case 253:var d=s.parent;1===d.elements.length?a(n,o,d):w(n,t,o,s);break;case 251:a(n,o,s);break;default:e.isImportClause(s.parent)&&s.parent.name===s?r(n,o,s.parent):e.isCallLikeExpression(s.parent)?w(n,t,o,s):I(n,o,s,26===s.kind?{useNonAdjustedEndPosition:!0}:void 0)}}function r(n,t,r){if(r.namedBindings){var a=r.name.getStart(t),i=e.getTokenAtPosition(t,r.name.end);if(i&&27===i.kind){var o=e.skipTrivia(t.text,i.end,!1,!0);n.deleteRange(t,{pos:a,end:o})}else I(n,t,r.name)}else I(n,t,r.parent)}function a(n,t,r){if(r.parent.name){var a=e.Debug.assertDefined(e.getTokenAtPosition(t,r.pos-1));n.deleteRange(t,{pos:a.getStart(t),end:r.end})}else{var i=e.getAncestor(r,249);I(n,t,i)}}function i(n,t,r,a){var i=a.parent;if(274!==i.kind)if(1===i.declarations.length){var o=i.parent;switch(o.kind){case 227:case 226:n.replaceNode(r,a,e.createObjectLiteral());break;case 225:I(n,r,i);break;case 219:I(n,r,o);break;default:e.Debug.assertNever(o)}}else w(n,t,r,a);else n.deleteNodeRange(r,e.findChildOfKind(i,20,r),e.findChildOfKind(i,21,r))}n.deleteDeclaration=t}(A||(A={})),n.deleteNode=I})(e.textChanges||(e.textChanges={}))}(u||(u={})),function(e){(function(n){var t=e.createMultiMap(),r=e.createMap();function a(n){return e.isArray(n)?e.formatStringFromArgs(e.getLocaleSpecificMessage(n[0]),n.slice(1)):e.getLocaleSpecificMessage(n)}function i(e,n,t){return s(e,a(t),n,void 0,void 0)}function o(e,n,t,r,i,o){return s(e,a(t),n,r,a(i),o)}function s(e,n,t,r,a,i){return{fixName:e,description:n,changes:t,fixId:r,fixAllDescription:a,commands:i?[i]:void 0}}function c(n){for(var a=0,i=n.errorCodes;a<i.length;a++){var o=i[a];t.add(String(o),n)}if(n.fixIds)for(var s=0,c=n.fixIds;s<c.length;s++){var l=c[s];e.Debug.assert(!r.has(l)),r.set(l,n)}}function l(){return e.arrayFrom(t.keys())}function u(n){return e.flatMap(t.get(String(n.errorCode))||e.emptyArray,function(e){return e.getCodeActions(n)})}function d(n){return r.get(e.cast(n.fixId,e.isString)).getAllCodeActions(n)}function p(e,n){return{changes:e,commands:n}}function m(e,n){return{fileName:e,textChanges:n}}function f(n,t,r){var a=[],i=e.textChanges.ChangeTracker.with(n,function(e){return _(n,t,function(n){return r(e,n,a)})});return p(i,0===a.length?void 0:a)}function _(n,t,r){for(var a=n.program,i=n.sourceFile,o=n.cancellationToken,s=0,c=a.getSemanticDiagnostics(i,o).concat(e.computeSuggestionDiagnostics(i,a,o));s<c.length;s++){var l=c[s];e.contains(t,l.code)&&r(l)}}n.createCodeFixActionNoFixId=i,n.createCodeFixAction=o,n.registerCodeFix=c,n.getSupportedErrorCodes=l,n.getFixes=u,n.getAllFixes=d,n.createCombinedCodeActions=p,n.createFileTextChanges=m,n.codeFixAll=f,n.eachDiagnostic=_})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){function n(n){var t=n.startPosition,r=n.endPosition;return e.createTextSpanFromBounds(t,void 0===r?t:r)}(function(n){var t=e.createMap();function r(e,n){t.set(e,n)}function a(n){return e.arrayFrom(e.flatMapIterator(t.values(),function(e){return n.cancellationToken&&n.cancellationToken.isCancellationRequested()?void 0:e.getAvailableActions(n)}))}function i(e,n,r){var a=t.get(n);return a&&a.getEditsForAction(e,r)}n.registerRefactor=r,n.getApplicableRefactors=a,n.getEditsForRefactor=i})(e.refactor||(e.refactor={})),e.getRefactorContextSpan=n}(u||(u={})),function(e){(function(n){var t="addConvertToUnknownForNonOverlappingTypes",r=[e.Diagnostics.Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the_other_If_this_was_intentional_convert_the_expression_to_unknown_first.code];function a(n,t,r){var a=e.getTokenAtPosition(t,r),i=e.Debug.assertDefined(e.findAncestor(a,function(n){return e.isAsExpression(n)||e.isTypeAssertion(n)})),o=e.isAsExpression(i)?e.createAsExpression(i.expression,e.createKeywordTypeNode(143)):e.createTypeAssertion(e.createKeywordTypeNode(143),i.expression);n.replaceNode(t,i.expression,o)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,function(e){return a(e,r.sourceFile,r.span.start)});return[n.createCodeFixAction(t,i,e.Diagnostics.Add_unknown_conversion_for_non_overlapping_types,t,e.Diagnostics.Add_unknown_to_all_conversions_of_non_overlapping_types)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(e,n){return a(e,n.file,n.start)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="addMissingInvocationForDecorator",r=[e.Diagnostics._0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write_0.code];function a(n,t,r){var a=e.getTokenAtPosition(t,r),i=e.findAncestor(a,e.isDecorator);e.Debug.assert(!!i,"Expected position to be owned by a decorator.");var o=e.createCall(i.expression,void 0,void 0);n.replaceNode(t,i.expression,o)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,function(e){return a(e,r.sourceFile,r.span.start)});return[n.createCodeFixAction(t,i,e.Diagnostics.Call_decorator_expression,t,e.Diagnostics.Add_to_all_uncalled_decorators)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(e,n){return a(e,n.file,n.start)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="addNameToNamelessParameter",r=[e.Diagnostics.Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1.code];function a(n,t,r){var a=e.getTokenAtPosition(t,r);if(!e.isIdentifier(a))return e.Debug.fail("add-name-to-nameless-parameter operates on identifiers, but got a "+e.formatSyntaxKind(a.kind));var i=a.parent;if(!e.isParameter(i))return e.Debug.fail("Tried to add a parameter name to a non-parameter: "+e.formatSyntaxKind(a.kind));var o=i.parent.parameters.indexOf(i);e.Debug.assert(!i.type,"Tried to add a parameter name to a parameter that already had one."),e.Debug.assert(o>-1,"Parameter not found in parent parameter list.");var s=e.createParameter(void 0,i.modifiers,i.dotDotDotToken,"arg"+o,i.questionToken,e.createTypeReferenceNode(a,void 0),i.initializer);n.replaceNode(t,a,s)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,function(e){return a(e,r.sourceFile,r.span.start)});return[n.createCodeFixAction(t,i,e.Diagnostics.Add_parameter_name,t,e.Diagnostics.Add_names_to_all_parameters_without_names)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(e,n){return a(e,n.file,n.start)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="annotateWithTypeFromJSDoc",r=[e.Diagnostics.JSDoc_types_may_be_moved_to_TypeScript_types.code];function a(n,t){var r=e.getTokenAtPosition(n,t);return e.tryCast(e.isParameter(r.parent)?r.parent.parent:r.parent,i)}function i(e){return c(e)&&o(e)}function o(n){return e.isFunctionLikeDeclaration(n)?n.parameters.some(o)||!n.type&&!!e.getJSDocReturnType(n):!n.type&&!!e.getJSDocType(n)}function s(n,t,r){if(e.isFunctionLikeDeclaration(r)&&(e.getJSDocReturnType(r)||r.parameters.some(function(n){return!!e.getJSDocType(n)}))){if(!r.typeParameters){var a=e.getJSDocTypeParameterDeclarations(r);a.length&&n.insertTypeParameters(t,r,a)}var i=e.isArrowFunction(r)&&!e.findChildOfKind(r,20,t);i&&n.insertNodeBefore(t,e.first(r.parameters),e.createToken(20));for(var o=0,s=r.parameters;o<s.length;o++){var c=s[o];if(!c.type){var u=e.getJSDocType(c);u&&n.tryInsertTypeAnnotation(t,c,l(u))}}if(i&&n.insertNodeAfter(t,e.last(r.parameters),e.createToken(21)),!r.type){var d=e.getJSDocReturnType(r);d&&n.tryInsertTypeAnnotation(t,r,l(d))}}else{var p=e.Debug.assertDefined(e.getJSDocType(r));e.Debug.assert(!r.type),n.tryInsertTypeAnnotation(t,r,l(p))}}function c(n){return e.isFunctionLikeDeclaration(n)||237===n.kind||153===n.kind||154===n.kind}function l(n){switch(n.kind){case 284:case 285:return e.createTypeReferenceNode("any",e.emptyArray);case 288:return u(n);case 287:return l(n.type);case 286:return d(n);case 290:return p(n);case 289:return m(n);case 164:return _(n);default:var t=e.visitEachChild(n,l,void 0);return e.setEmitFlags(t,1),t}}function u(n){return e.createUnionTypeNode([e.visitNode(n.type,l),e.createTypeReferenceNode("undefined",e.emptyArray)])}function d(n){return e.createUnionTypeNode([e.visitNode(n.type,l),e.createTypeReferenceNode("null",e.emptyArray)])}function p(n){return e.createArrayTypeNode(e.visitNode(n.type,l))}function m(n){return e.createFunctionTypeNode(e.emptyArray,n.parameters.map(f),n.type)}function f(n){var t=n.parent.parameters.indexOf(n),r=290===n.type.kind&&t===n.parent.parameters.length-1,a=n.name||(r?"rest":"arg"+t),i=r?e.createToken(25):n.dotDotDotToken;return e.createParameter(n.decorators,n.modifiers,i,a,n.questionToken,e.visitNode(n.type,l),n.initializer)}function _(n){var t=n.typeName,r=n.typeArguments;if(e.isIdentifier(n.typeName)){if(e.isJSDocIndexSignature(n))return g(n);var a=n.typeName.text;switch(n.typeName.text){case"String":case"Boolean":case"Object":case"Number":a=a.toLowerCase();break;case"array":case"date":case"promise":a=a[0].toUpperCase()+a.slice(1);break}t=e.createIdentifier(a),r="Array"!==a&&"Promise"!==a||n.typeArguments?e.visitNodes(n.typeArguments,l):e.createNodeArray([e.createTypeReferenceNode("any",e.emptyArray)])}return e.createTypeReferenceNode(t,r)}function g(n){var t=e.createParameter(void 0,void 0,void 0,135===n.typeArguments[0].kind?"n":"s",void 0,e.createTypeReferenceNode(135===n.typeArguments[0].kind?"number":"string",[]),void 0),r=e.createTypeLiteralNode([e.createIndexSignature(void 0,void 0,[t],n.typeArguments[1])]);return e.setEmitFlags(r,1),r}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=a(r.sourceFile,r.span.start);if(i){var o=e.textChanges.ChangeTracker.with(r,function(e){return s(e,r.sourceFile,i)});return[n.createCodeFixAction(t,o,e.Diagnostics.Annotate_with_type_from_JSDoc,t,e.Diagnostics.Annotate_everything_with_types_from_JSDoc)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(e,n){var t=a(n.file,n.start);t&&s(e,n.file,t)})}}),n.parameterShouldGetTypeFromJSDoc=i})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t,r="inferFromUsage",a=[e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined.code,e.Diagnostics.Variable_0_implicitly_has_an_1_type.code,e.Diagnostics.Parameter_0_implicitly_has_an_1_type.code,e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code,e.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation.code,e.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type.code,e.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation.code,e.Diagnostics.Member_0_implicitly_has_an_1_type.code,e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage.code,e.Diagnostics.Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code,e.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code,e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage.code,e.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage.code,e.Diagnostics._0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage.code,e.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage.code,e.Diagnostics.Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code];function i(n,t){switch(n){case e.Diagnostics.Parameter_0_implicitly_has_an_1_type.code:case e.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:return e.isSetAccessorDeclaration(e.getContainingFunction(t))?e.Diagnostics.Infer_type_of_0_from_usage:e.Diagnostics.Infer_parameter_types_from_usage;case e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code:case e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Infer_parameter_types_from_usage;default:return e.Diagnostics.Infer_type_of_0_from_usage}}function o(n){switch(n){case e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined.code;case e.Diagnostics.Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Variable_0_implicitly_has_an_1_type.code;case e.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Parameter_0_implicitly_has_an_1_type.code;case e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code;case e.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage.code:return e.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation.code;case e.Diagnostics._0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type.code;case e.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage.code:return e.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation.code;case e.Diagnostics.Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Member_0_implicitly_has_an_1_type.code}return n}function s(n,t,r,a,i,s,p,f){if(e.isParameterPropertyModifier(r.kind)||72===r.kind||25===r.kind||100===r.kind){var g=r.parent;switch(a=o(a),a){case e.Diagnostics.Member_0_implicitly_has_an_1_type.code:case e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined.code:if(e.isVariableDeclaration(g)&&p(g)||e.isPropertyDeclaration(g)||e.isPropertySignature(g))return c(n,t,g,i,f,s),g;if(e.isPropertyAccessExpression(g)){var y=v(g.name,i,s),h=_(y,g,i,f);if(h){var b=e.createJSDocTypeTag(e.createJSDocTypeExpression(h),"");m(n,t,e.cast(g.parent.parent,e.isExpressionStatement),[b])}return g}return;case e.Diagnostics.Variable_0_implicitly_has_an_1_type.code:var E=i.getTypeChecker().getSymbolAtLocation(r);return E&&E.valueDeclaration&&e.isVariableDeclaration(E.valueDeclaration)&&p(E.valueDeclaration)?(c(n,t,E.valueDeclaration,i,f,s),E.valueDeclaration):void 0}var T=e.getContainingFunction(r);if(void 0!==T)switch(a){case e.Diagnostics.Parameter_0_implicitly_has_an_1_type.code:if(e.isSetAccessorDeclaration(T))return u(n,t,T,i,f,s),T;case e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code:if(p(T)){var S=e.cast(g,e.isParameter);return l(n,t,S,T,i,f,s),S}return;case e.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation.code:case e.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type.code:return e.isGetAccessorDeclaration(T)&&e.isIdentifier(T.name)?(d(n,t,T,v(T.name,i,s),i,f),T):void 0;case e.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation.code:return e.isSetAccessorDeclaration(T)?(u(n,t,T,i,f,s),T):void 0;default:return e.Debug.fail(String(a))}}}function c(n,t,r,a,i,o){e.isIdentifier(r.name)&&d(n,t,r,v(r.name,a,o),a,i)}function l(n,t,r,a,i,o,s){if(e.isIdentifier(r.name)){var c=y(a,t,i,s)||a.parameters.map(function(n){return{declaration:n,type:e.isIdentifier(n.name)?v(n.name,i,s):i.getTypeChecker().getAnyType()}});if(e.Debug.assert(a.parameters.length===c.length),e.isInJSFile(a))p(n,t,c,i,o);else{var l=e.isArrowFunction(a)&&!e.findChildOfKind(a,20,t);l&&n.insertNodeBefore(t,e.first(a.parameters),e.createToken(20));for(var u=0,m=c;u<m.length;u++){var f=m[u],_=f.declaration,g=f.type;!_||_.type||_.initializer||d(n,t,_,g,i,o)}l&&n.insertNodeAfter(t,e.last(a.parameters),e.createToken(21))}}}function u(n,t,r,a,i,o){var s=e.firstOrUndefined(r.parameters);if(s&&e.isIdentifier(r.name)&&e.isIdentifier(s.name)){var c=v(r.name,a,o);c===a.getTypeChecker().getAnyType()&&(c=v(s.name,a,o)),e.isInJSFile(r)?p(n,t,[{declaration:s,type:c}],a,i):d(n,t,s,c,a,i)}}function d(n,t,r,a,i,o){var s=_(a,r,i,o);if(s)if(e.isInJSFile(t)&&153!==r.kind){var c=e.isVariableDeclaration(r)?e.tryCast(r.parent.parent,e.isVariableStatement):r;if(!c)return;var l=e.createJSDocTypeExpression(s),u=e.isGetAccessorDeclaration(r)?e.createJSDocReturnTag(l,""):e.createJSDocTypeTag(l,"");m(n,t,c,[u])}else n.tryInsertTypeAnnotation(t,r,s)}function p(n,t,r,a,i){var o=r.length&&r[0].declaration.parent;if(o){var s=e.mapDefined(r,function(n){var t=n.declaration;if(!t.initializer&&!e.getJSDocType(t)&&e.isIdentifier(t.name)){var r=n.type&&_(n.type,t,a,i),o=e.getSynthesizedClone(t.name);return e.setEmitFlags(o,3584),r&&e.createJSDocParamTag(o,!!n.isOptional,e.createJSDocTypeExpression(r),"")}});m(n,t,o,s)}}function m(n,t,r,a){var i=e.mapDefined(r.jsDoc,function(e){return e.comment}),o=e.flatMapToMutable(r.jsDoc,function(e){return e.tags}),s=a.filter(function(e){return!o||!o.some(function(n,t){var r=f(n,e);return r&&(o[t]=r),!!r})}),c=e.createJSDocComment(i.join("\n"),e.createNodeArray((o||e.emptyArray).concat(s)));n.insertJsdocCommentBefore(t,r,c)}function f(n,t){if(n.kind===t.kind)switch(n.kind){case 299:var r=n,a=t;return e.isIdentifier(r.name)&&e.isIdentifier(a.name)&&r.name.escapedText===a.name.escapedText?e.createJSDocParamTag(a.name,a.isBracketed,a.typeExpression,r.comment):void 0;case 300:return e.createJSDocReturnTag(t.typeExpression,n.comment)}}function _(e,n,t,r){var a=t.getTypeChecker(),i=!0,o=function(){i=!1},s=a.typeToTypeNode(e,n,void 0,{trackSymbol:function(e,n,t){i=i&&0===a.isSymbolAccessible(e,n,t,!1).accessibility},reportInaccessibleThisError:o,reportPrivateInBaseOfClassExpression:o,reportInaccessibleUniqueSymbolError:o,moduleResolverHost:{readFile:r.readFile,fileExists:r.fileExists,directoryExists:r.directoryExists,getSourceFiles:t.getSourceFiles,getCurrentDirectory:t.getCurrentDirectory,getCommonSourceDirectory:t.getCommonSourceDirectory}});return i?s:void 0}function g(n,t,r){return e.mapDefined(e.FindAllReferences.getReferenceEntriesForNode(-1,n,t,t.getSourceFiles(),r),function(n){return 0!==n.kind?e.tryCast(n.node,e.isIdentifier):void 0})}function v(e,n,r){var a=g(e,n,r),i=n.getTypeChecker(),o=t.inferTypesFromReferences(a,i,r);return t.unifyFromContext(o,i)}function y(n,r,a,i){var o;switch(n.kind){case 157:o=e.findChildOfKind(n,124,r);break;case 197:case 196:var s=n.parent;o=e.isVariableDeclaration(s)&&e.isIdentifier(s.name)?s.name:n.name;break;case 239:case 156:o=n.name;break}if(o)return t.inferTypeForParametersFromReferences(g(o,a,i),n,a,i)}n.registerCodeFix({errorCodes:a,getCodeActions:function(t){var a,o=t.sourceFile,c=t.program,l=t.span.start,u=t.errorCode,d=t.cancellationToken,p=t.host,m=e.getTokenAtPosition(o,l),f=e.textChanges.ChangeTracker.with(t,function(n){a=s(n,o,m,u,c,d,e.returnTrue,p)}),_=a&&e.getNameOfDeclaration(a);return _&&0!==f.length?[n.createCodeFixAction(r,f,[i(u,m),_.getText(o)],r,e.Diagnostics.Infer_all_types_from_usage)]:void 0},fixIds:[r],getAllCodeActions:function(t){var r=t.sourceFile,i=t.program,o=t.cancellationToken,c=t.host,l=e.nodeSeenTracker();return n.codeFixAll(t,a,function(n,t){s(n,r,e.getTokenAtPosition(t.file,t.start),t.code,i,o,l,c)})}}),function(n){function t(e,n,t){for(var r={},i=0,o=e;i<o.length;i++){var s=o[i];t.throwIfCancellationRequested(),a(s,n,r)}return _(r,n)}function r(n,r,i,o){var s=i.getTypeChecker();if(0!==n.length&&r.parameters){for(var c={},l=0,u=n;l<u.length;l++){var d=u[l];o.throwIfCancellationRequested(),a(d,s,c)}var p=(c.constructContexts||[]).concat(c.callContexts||[]);return r.parameters.map(function(n,a){for(var c=[],l=e.isRestParameter(n),u=!1,d=0,f=p;d<f.length;d++){var _=f[d];if(_.argumentTypes.length<=a)u=e.isInJSFile(r),c.push(s.getUndefinedType());else if(l)for(var v=a;v<_.argumentTypes.length;v++)c.push(s.getBaseTypeOfLiteralType(_.argumentTypes[v]));else c.push(s.getBaseTypeOfLiteralType(_.argumentTypes[a]))}if(e.isIdentifier(n.name)){var y=t(g(n.name,i,o),s,o);c.push.apply(c,l?e.mapDefined(y,s.getElementTypeOfArrayType):y)}var h=m(c,s);return{type:l?s.createArrayType(h):h,isOptional:u&&!l,declaration:n}})}}function a(n,t,r){while(e.isRightSideOfQualifiedNameOrPropertyAccess(n))n=n.parent;switch(n.parent.kind){case 203:r.isNumber=!0;break;case 202:o(n.parent,r);break;case 204:s(n,n.parent,t,r);break;case 271:case 272:c(n.parent,t,r);break;case 191:case 192:n.parent.expression===n?l(n.parent,t,r):i(n,t,r);break;case 189:u(n.parent,t,r);break;case 190:d(n.parent,n,t,r);break;case 237:var a=n.parent,p=a.name,m=a.initializer;if(n===p){m&&h(r,t.getTypeAtLocation(m));break}default:return i(n,t,r)}}function i(n,t,r){e.isExpressionNode(n)&&h(r,t.getContextualType(n))}function o(e,n){switch(e.operator){case 44:case 45:case 39:case 53:n.isNumber=!0;break;case 38:n.isNumberOrString=!0;break}}function s(n,t,r,a){switch(t.operatorToken.kind){case 41:case 40:case 42:case 43:case 46:case 47:case 48:case 49:case 50:case 51:case 61:case 63:case 62:case 64:case 65:case 69:case 70:case 71:case 66:case 68:case 67:case 39:case 28:case 31:case 30:case 32:var i=r.getTypeAtLocation(t.left===n?t.right:t.left);1056&i.flags?h(a,i):a.isNumber=!0;break;case 60:case 38:var o=r.getTypeAtLocation(t.left===n?t.right:t.left);1056&o.flags?h(a,o):296&o.flags?a.isNumber=!0:132&o.flags?a.isString=!0:a.isNumberOrString=!0;break;case 59:case 33:case 35:case 36:case 34:h(a,r.getTypeAtLocation(t.left===n?t.right:t.left));break;case 93:n===t.left&&(a.isString=!0);break;case 55:n!==t.left||237!==n.parent.parent.kind&&!e.isAssignmentExpression(n.parent.parent,!0)||h(a,r.getTypeAtLocation(t.right));break;case 54:case 27:case 94:break}}function c(e,n,t){h(t,n.getTypeAtLocation(e.parent.parent.expression))}function l(e,n,t){var r={argumentTypes:[],returnType:{}};if(e.arguments)for(var i=0,o=e.arguments;i<o.length;i++){var s=o[i];r.argumentTypes.push(n.getTypeAtLocation(s))}a(e,n,r.returnType),191===e.kind?(t.callContexts||(t.callContexts=[])).push(r):(t.constructContexts||(t.constructContexts=[])).push(r)}function u(n,t,r){var i=e.escapeLeadingUnderscores(n.name.text);r.properties||(r.properties=e.createUnderscoreEscapedMap());var o=r.properties.get(i)||{};a(n,t,o),r.properties.set(i,o)}function d(e,n,t,r){if(n!==e.argumentExpression){var i=t.getTypeAtLocation(e.argumentExpression),o={};a(e,t,o),296&i.flags?r.numberIndexContext=o:r.stringIndexContext=o}else r.isNumberOrString=!0}function p(n,t){for(var r=[],a=0,i=n;a<i.length;a++)for(var o=i[a],s=0,c=t;s<c.length;s++){var l=c[s],u=l.high,d=l.low;u(o)&&(e.Debug.assert(!d(o)),r.push(d))}return n.filter(function(e){return r.every(function(n){return!n(e)})})}function m(e,n,t){if(void 0===t&&(t=n.getAnyType()),!e.length)return t;var r=n.getUnionType([n.getStringType(),n.getNumberType()]),a=[{high:function(e){return e===n.getStringType()||e===n.getNumberType()},low:function(e){return e===r}},{high:function(e){return!(16385&e.flags)},low:function(e){return!!(16385&e.flags)}},{high:function(e){return!(114689&e.flags)&&!(16&n.getObjectFlags(e))},low:function(e){return!!(16&n.getObjectFlags(e))}}],i=p(e,a),o=i.filter(function(e){return 16&n.getObjectFlags(e)});return o.length&&(i=i.filter(function(e){return!(16&n.getObjectFlags(e))}),i.push(f(o,n))),n.getWidenedType(n.getUnionType(i))}function f(n,t){if(1===n.length)return n[0];for(var r=[],a=[],i=[],o=[],s=!1,c=!1,l=e.createMultiMap(),u=0,d=n;u<d.length;u++){for(var p=d[u],m=0,f=t.getPropertiesOfType(p);m<f.length;m++){var _=f[m];l.add(_.name,t.getTypeOfSymbolAtLocation(_,_.valueDeclaration))}r.push.apply(r,t.getSignaturesOfType(p,0)),a.push.apply(a,t.getSignaturesOfType(p,1)),p.stringIndexInfo&&(i.push(p.stringIndexInfo.type),s=s||p.stringIndexInfo.isReadonly),p.numberIndexInfo&&(o.push(p.numberIndexInfo.type),c=c||p.numberIndexInfo.isReadonly)}var g=e.mapEntries(l,function(e,r){var a=r.length<n.length?16777216:0,i=t.createSymbol(4|a,e);return i.type=t.getUnionType(r),[e,i]});return t.createAnonymousType(n[0].symbol,g,r,a,i.length?t.createIndexInfo(t.getUnionType(i),s):void 0,o.length?t.createIndexInfo(t.getUnionType(o),c):void 0)}function _(n,t){var r=[];if(n.isNumber&&r.push(t.getNumberType()),n.isString&&r.push(t.getStringType()),n.isNumberOrString&&r.push(t.getUnionType([t.getStringType(),t.getNumberType()])),r.push.apply(r,(n.candidateTypes||[]).map(function(e){return t.getBaseTypeOfLiteralType(e)})),n.properties&&b(n.properties.get("then"))){var a=v(0,n.properties.get("then").callContexts,!1,t),i=a.getCallSignatures().map(function(e){return e.getReturnType()});i.push(t.createPromiseType(i.length?t.getUnionType(i,2):t.getAnyType()))}else n.properties&&b(n.properties.get("push"))&&r.push(t.createArrayType(v(0,n.properties.get("push").callContexts,!1,t)));if(n.numberIndexContext)r.push(t.createArrayType(E(n.numberIndexContext)));else if(n.properties||n.callContexts||n.constructContexts||n.stringIndexContext){var o=e.createUnderscoreEscapedMap(),s=[],c=[],l=void 0;if(n.properties&&n.properties.forEach(function(e,n){var r=t.createSymbol(4,n);r.type=E(e),o.set(n,r)}),n.callContexts)for(var u=0,d=n.callContexts;u<d.length;u++){var p=d[u];s.push(y(p,t))}if(n.constructContexts)for(var f=0,g=n.constructContexts;f<g.length;f++){var h=g[f];c.push(y(h,t))}n.stringIndexContext&&(l=t.createIndexInfo(E(n.stringIndexContext),!1)),r.push(t.createAnonymousType(void 0,o,s,c,l,void 0))}return r;function E(e){return m(_(e,t),t)}}function v(n,t,r,a){var i=[];if(t)for(var o=0,s=t;o<s.length;o++){var c=s[o];c.argumentTypes.length>n&&(r?i=e.concatenate(i,e.map(c.argumentTypes.slice(n),function(e){return a.getBaseTypeOfLiteralType(e)})):i.push(a.getBaseTypeOfLiteralType(c.argumentTypes[n])))}if(i.length){var l=a.getWidenedType(a.getUnionType(i,2));return r?a.createArrayType(l):l}}function y(n,t){for(var r=[],a=0;a<n.argumentTypes.length;a++){var i=t.createSymbol(1,e.escapeLeadingUnderscores("arg"+a));i.type=t.getWidenedType(t.getBaseTypeOfLiteralType(n.argumentTypes[a])),r.push(i)}var o=m(_(n.returnType,t),t,t.getVoidType());return t.createSignature(void 0,void 0,void 0,r,o,void 0,n.argumentTypes.length,!1,!1)}function h(e,n){!n||1&n.flags||131072&n.flags||(e.candidateTypes||(e.candidateTypes=[])).push(n)}function b(e){return!!e&&!!e.callContexts}n.inferTypesFromReferences=t,n.inferTypeForParametersFromReferences=r,n.unifyFromContext=m}(t||(t={}))})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="convertFunctionToEs6Class",r=[e.Diagnostics.This_constructor_function_may_be_converted_to_a_class_declaration.code];function a(n,t,r,a){var o=a.getSymbolAtLocation(e.getTokenAtPosition(t,r));if(o&&19&o.flags){var s,c,l=o.valueDeclaration;switch(l.kind){case 239:s=l,n.delete(t,l),c=p(l);break;case 237:s=l.parent.parent,c=d(l),1===l.parent.declarations.length?(e.copyComments(s,c,t),n.delete(t,s)):n.delete(t,l);break}c&&(e.copyComments(l,c,t),n.insertNodeAfter(t,s,c))}function u(r){var a=[];return r.members&&r.members.forEach(function(e){var n=s(e,void 0);n&&a.push(n)}),r.exports&&r.exports.forEach(function(n){var t=s(n,[e.createToken(116)]);t&&a.push(t)}),a;function o(n,t){return e.isFunctionLike(t)}function s(r,a){if(8192&r.flags){var s=r.valueDeclaration,c=s.parent;if(o(s,c.right)){var l=c.parent&&221===c.parent.kind?c.parent:c;if(n.delete(t,l),!c.right)return e.createProperty([],a,r.name,void 0,void 0,void 0);switch(c.right.kind){case 196:var u=c.right,d=e.concatenate(a,i(u,121)),p=e.createMethod(void 0,d,void 0,s.name,void 0,void 0,u.parameters,void 0,u.body);return e.copyComments(c,p,t),p;case 197:var m=c.right,f=m.body,_=void 0;_=218===f.kind?f:e.createBlock([e.createReturn(f)]);d=e.concatenate(a,i(m,121)),p=e.createMethod(void 0,d,void 0,s.name,void 0,void 0,m.parameters,void 0,_);return e.copyComments(c,p,t),p;default:if(e.isSourceFileJS(t))return;var g=e.createProperty(void 0,a,s.name,void 0,void 0,c.right);return e.copyComments(c.parent,g,t),g}}}}}function d(n){var t=n.initializer;if(t&&196===t.kind&&72===n.name.kind){var r=u(n.symbol);t.body&&r.unshift(e.createConstructor(void 0,void 0,t.parameters,t.body));var a=i(s,85),o=e.createClassDeclaration(void 0,a,n.name,void 0,void 0,r);return o}}function p(n){var t=u(o);n.body&&t.unshift(e.createConstructor(void 0,void 0,n.parameters,n.body));var r=i(n,85),a=e.createClassDeclaration(void 0,r,n.name,void 0,void 0,t);return a}}function i(n,t){return e.filter(n.modifiers,function(e){return e.kind===t})}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,function(e){return a(e,r.sourceFile,r.span.start,r.program.getTypeChecker())});return[n.createCodeFixAction(t,i,e.Diagnostics.Convert_function_to_an_ES2015_class,t,e.Diagnostics.Convert_all_constructor_functions_to_classes)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(n,t){return a(n,t.file,t.start,e.program.getTypeChecker())})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="convertToAsyncFunction",r=[e.Diagnostics.This_may_be_converted_to_an_async_function.code],a=!0;function i(n,t,r,a,i){var l,u=e.getTokenAtPosition(t,r);if(l=e.isIdentifier(u)&&e.isVariableDeclaration(u.parent)&&u.parent.initializer&&e.isFunctionLikeDeclaration(u.parent.initializer)?u.parent.initializer:e.tryCast(e.getContainingFunction(e.getTokenAtPosition(t,r)),e.isFunctionLikeDeclaration),l){var p=e.createMap(),m=e.createMap(),_=[],g=e.isInJSFile(l),v=c(l,a),y=d(l,a,p,i,v,m,_),h=s(p),b=y.body&&e.isBlock(y.body)?o(y.body):e.emptyArray,E={checker:a,synthNamesMap:p,allVarNames:_,setOfExpressionsToReturn:v,constIdentifiers:h,originalTypeMap:m,isInJSFile:g};if(b.length){n.insertLastModifierBefore(t,121,l);for(var T=function(n){e.forEachChild(n,function t(r){e.isCallExpression(r)?A(r,n):e.isFunctionLike(r)||e.forEachChild(r,t)})},S=0,x=b;S<x.length;S++){var C=x[S];T(C)}}}function A(e,r){var a=f(e,E,e);n.replaceNodeWithNodes(t,r,a)}}function o(n){var t=[];return e.forEachReturnStatement(n,function(n){e.isReturnStatementWithFixablePromiseHandler(n)&&t.push(n)}),t}function s(e){var n=[];return e.forEach(function(e){0===e.numberOfAssignmentsOriginal&&n.push(e.identifier)}),n}function c(n,t){if(!n.body)return e.createMap();var r=e.createMap();return e.forEachChild(n.body,function n(a){l(a,t,"then")?(r.set(e.getNodeId(a).toString(),!0),e.forEach(a.arguments,n)):l(a,t,"catch")?(r.set(e.getNodeId(a).toString(),!0),e.forEachChild(a,n)):l(a,t)?r.set(e.getNodeId(a).toString(),!0):e.forEachChild(a,n)}),r}function l(n,t,r){var a=r?e.isCallExpression(n):e.isExpression(n),i=a&&(!r||e.hasPropertyAccessExpressionWithName(n,r)),o=i&&t.getTypeAtLocation(n);return!(!o||!t.getPromisedTypeOfPromise(o))}function u(e,n){return e.valueDeclaration&&e.valueDeclaration.getSourceFile()===n}function d(n,t,r,a,i,o,s){var c=e.createMap(),l=e.createMap();return e.forEachChild(n,function n(i){if(e.isIdentifier(i)){var o=t.getSymbolAtLocation(i),f=o&&u(o,a.sourceFile);if(o&&f){var _=t.getTypeAtLocation(i),g=T(_,t),v=e.getSymbolId(o).toString();if(!g||e.isFunctionLikeDeclaration(i.parent)||r.has(v)){if(i.parent&&(e.isParameter(i.parent)||e.isVariableDeclaration(i.parent))){var y=i.text,h=l.get(y);if(h&&h.some(function(e){return e!==o})){var b=m(i,l);c.set(v,b.identifier),r.set(v,b),s.push({identifier:b.identifier,symbol:o}),p(l,y,o)}else{var E=e.getSynthesizedDeepClone(i);c.set(v,E),r.set(v,{identifier:E,types:[],numberOfAssignmentsOriginal:s.filter(function(e){return e.identifier.text===i.text}).length}),(e.isParameter(i.parent)&&d(i.parent.parent)||e.isVariableDeclaration(i.parent))&&(s.push({identifier:E,symbol:o}),p(l,y,o))}}}else{var S=e.firstOrUndefined(g.parameters),x=S&&e.isParameter(S.valueDeclaration)&&e.tryCast(S.valueDeclaration.name,e.isIdentifier)||e.createOptimisticUniqueName("result"),C=m(x,l);r.set(v,C),s.push({identifier:C.identifier,symbol:o}),p(l,x.text,o)}}}else e.forEachChild(i,n)}),e.getSynthesizedDeepCloneWithRenames(n,!0,c,t,f);function d(n){var r=n.parent;if(e.isCallExpression(r)||e.isIdentifier(r)&&!i.get(e.getNodeId(r).toString())){var a=t.getTypeAtLocation(r),o=a&&t.getPromisedTypeOfPromise(a);return!!o}return!1}function f(n,a){if(e.isIdentifier(n)){var s=t.getSymbolAtLocation(n),c=s&&e.getSymbolId(s).toString(),l=s&&r.get(c);if(l){var u=t.getTypeAtLocation(n);o.set(e.getNodeId(a).toString(),u)}}var d=i.get(e.getNodeId(n).toString());void 0!==d&&(i.delete(e.getNodeId(n).toString()),i.set(e.getNodeId(a).toString(),d))}}function p(e,n,t){e.has(n)?e.get(n).push(t):e.set(n,[t])}function m(n,t){var r=(t.get(n.text)||e.emptyArray).length,a=0,i=0===r?n:e.createIdentifier(n.text+"_"+r);return{identifier:i,types:[],numberOfAssignmentsOriginal:a}}function f(n,t,r,i){if(!n)return e.emptyArray;var o=e.isIdentifier(n)&&t.originalTypeMap.get(e.getNodeId(n).toString()),s=o||t.checker.getTypeAtLocation(n);return e.isCallExpression(n)&&e.hasPropertyAccessExpressionWithName(n,"then")&&s&&t.checker.getPromisedTypeOfPromise(s)?v(n,t,r,i):e.isCallExpression(n)&&e.hasPropertyAccessExpressionWithName(n,"catch")&&s&&t.checker.getPromisedTypeOfPromise(s)?_(n,t,i):e.isPropertyAccessExpression(n)?f(n.expression,t,r,i):s&&t.checker.getPromisedTypeOfPromise(s)?h(n,t,i):(a=!1,e.emptyArray)}function _(n,t,r){var a=n.arguments[0],i=C(a,t),o=t.setOfExpressionsToReturn.get(e.getNodeId(n).toString());r&&!o&&(r.numberOfAssignmentsOriginal=2,t.synthNamesMap.forEach(function(e,n){if(e.identifier.text===r.identifier.text){var a=g(r);t.synthNamesMap.set(n,a)}}),t.constIdentifiers.some(function(e){return e.text===r.identifier.text})&&t.constIdentifiers.push(g(r).identifier));var s,c=e.createBlock(f(n.expression,t,n,r)),l=E(a,r,i,n,t),u=i?i.identifier.text:"e",d=e.createCatchClause(u,e.createBlock(l));if(r&&!o){var p=r.types,m=t.checker.getUnionType(p,2),_=t.isInJSFile?void 0:t.checker.typeToTypeNode(m),v=[e.createVariableDeclaration(e.getSynthesizedDeepClone(r.identifier),_)];s=e.createVariableStatement(void 0,e.createVariableDeclarationList(v,1))}var y=e.createTry(c,d,void 0);return s?[s,y]:[y]}function g(n){var t=e.createOptimisticUniqueName(n.identifier.text),r={identifier:t,types:[],numberOfAssignmentsOriginal:0};return r}function v(n,t,r,a){var i=n.arguments,o=i[0],s=i[1];if(!o)return f(n.expression,t,r);var c=C(o,t),l=E(o,a,c,n,t);if(s){var u=C(s,t),d=e.createBlock(f(n.expression,t,n,c).concat(l)),p=E(s,a,u,n,t),m=u?u.identifier.text:"e",_=e.createCatchClause(m,e.createBlock(p));return[e.createTry(d,_,void 0)]}return f(n.expression,t,n,c).concat(l)}function y(e,n){var t=n.some(function(n){return n.text===e.text});return t?2:1}function h(n,t,r){var a=t.setOfExpressionsToReturn.get(e.getNodeId(n).toString()),i=n.original?n.original.parent:n.parent;return!r||a||i&&!e.isPropertyAccessExpression(i)?r||a||i&&!e.isPropertyAccessExpression(i)?[e.createReturn(e.getSynthesizedDeepClone(n))]:[e.createStatement(e.createAwait(n))]:b(r,e.createAwait(n),t)}function b(n,t,r){return n&&0!==n.identifier.text.length?n.types.length<n.numberOfAssignmentsOriginal?[e.createStatement(e.createAssignment(e.getSynthesizedDeepClone(n.identifier),t))]:[e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(e.getSynthesizedDeepClone(n.identifier),void 0,t)],y(n.identifier,r.constIdentifiers)))]:[e.createStatement(t)]}function E(n,t,r,i,o){var s=o.setOfExpressionsToReturn.get(e.getNodeId(i).toString());switch(n.kind){case 96:break;case 72:if(!r)break;var c=e.createCall(e.getSynthesizedDeepClone(n),void 0,[r.identifier]);if(s)return[e.createReturn(c)];var l=o.originalTypeMap.get(e.getNodeId(n).toString())||o.checker.getTypeAtLocation(n),u=o.checker.getSignaturesOfType(l,0);if(!u.length){a=!1;break}var d=u[0].getReturnType(),p=b(t,e.createAwait(c),o);return t&&t.types.push(d),p;case 196:case 197:var m=n.body;if(e.isBlock(m)){for(var f=[],_=!1,g=0,v=m.statements;g<v.length;g++){var y=v[g];e.isReturnStatement(y)&&(_=!0),e.isReturnStatementWithFixablePromiseHandler(y)?f=f.concat(x(o,[y],t)):f.push(y)}return s?f.map(function(n){return e.getSynthesizedDeepClone(n)}):S(f,void 0===t?void 0:t.identifier,o,_)}var h=e.isFixablePromiseHandler(m)?[e.createReturn(m)]:e.emptyArray,E=x(o,h,t);if(E.length>0)return E;var C=o.checker.getTypeAtLocation(n),A=T(C,o.checker).getReturnType(),D=e.getSynthesizedDeepClone(m),L=o.checker.getPromisedTypeOfPromise(A)?e.createAwait(D):D;if(s)return[e.createReturn(L)];var k=b(t,L,o);return t&&t.types.push(A),k;default:a=!1;break}return e.emptyArray}function T(n,t){var r=t.getSignaturesOfType(n,0);return e.lastOrUndefined(r)}function S(n,t,r,a){for(var i=[],o=0,s=n;o<s.length;o++){var c=s[o];if(e.isReturnStatement(c)){if(c.expression){var u=l(c.expression,r.checker)?e.createAwait(c.expression):c.expression;void 0===t?i.push(e.createExpressionStatement(u)):i.push(e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(t,void 0,u)],y(t,r.constIdentifiers))))}}else i.push(e.getSynthesizedDeepClone(c))}return a||void 0===t||i.push(e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(t,void 0,e.createIdentifier("undefined"))],y(t,r.constIdentifiers)))),i}function x(n,t,r){for(var a=[],i=0,o=t;i<o.length;i++){var s=o[i];e.forEachChild(s,function t(i){if(e.isCallExpression(i)){var o=f(i,n,i,r);if(a=a.concat(o),a.length>0)return}else e.isFunctionLike(i)||e.forEachChild(i,t)})}return a}function C(n,t){var r,a=0,i=[];if(e.isFunctionLikeDeclaration(n)){if(n.parameters.length>0){var o=n.parameters[0].name;r=s(o)}}else e.isIdentifier(n)&&(r=s(n));if(r&&"undefined"!==r.identifier.text)return r;function s(n){var r=l(n),o=c(r);if(!o)return{identifier:n,types:i,numberOfAssignmentsOriginal:a};var s=t.synthNamesMap.get(e.getSymbolId(o).toString());return s||{identifier:n,types:i,numberOfAssignmentsOriginal:a}}function c(e){return e.symbol?e.symbol:t.checker.getSymbolAtLocation(e)}function l(e){return e.original?e.original:e}}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){a=!0;var o=e.textChanges.ChangeTracker.with(r,function(e){return i(e,r.sourceFile,r.span.start,r.program.getTypeChecker(),r)});return a?[n.createCodeFixAction(t,o,e.Diagnostics.Convert_to_async_function,t,e.Diagnostics.Convert_all_to_async_functions)]:[]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(n,t){return i(n,t.file,t.start,e.program.getTypeChecker(),e)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){function t(n,t,r,a){for(var i=0,o=n.imports;i<o.length;i++){var s=o[i],c=e.getResolvedModule(n,s.text);if(c&&c.resolvedFileName===t.fileName){var l=e.importFromModuleSpecifier(s);switch(l.kind){case 248:r.replaceNode(n,l,e.makeImport(l.name,void 0,s,a));break;case 191:e.isRequireCall(l,!1)&&r.replaceNode(n,l,e.createPropertyAccess(e.getSynthesizedDeepClone(l),"default"));break}}}}function r(n,t,r,o,c){var l={original:E(n),additional:e.createMap()},u=a(n,t,l);i(n,u,r);for(var d=!1,p=0,m=n.statements;p<m.length;p++){var f=m[p],_=s(n,f,t,r,l,o,u,c);d=d||_}return d}function a(n,t,r){var a=e.createMap();return o(n,function(n){var i=n.name,o=i.text,s=i.originalKeywordKind;!a.has(o)&&(void 0!==s&&e.isNonContextualKeyword(s)||t.resolveName(n.name.text,n,67220415,!0))&&a.set(o,b("_"+o,r))}),a}function i(n,t,r){o(n,function(a,i){if(!i){var o=a.name.text;r.replaceNode(n,a,e.createIdentifier(t.get(o)||o))}})}function o(n,t){n.forEachChild(function r(a){if(e.isPropertyAccessExpression(a)&&e.isExportsOrModuleExportsOrAlias(n,a.expression)){var i=a.parent;t(a,e.isBinaryExpression(i)&&i.left===a&&59===i.operatorToken.kind)}a.forEachChild(r)})}function s(n,t,r,a,i,o,s,l){switch(t.kind){case 219:return c(n,t,a,r,i,o,l),!1;case 221:var d=t.expression;switch(d.kind){case 191:return e.isRequireCall(d,!0)&&a.replaceNode(n,t,e.makeImport(void 0,void 0,d.arguments[0],l)),!1;case 204:var p=d.operatorToken;return 59===p.kind&&u(n,r,d,a,s)}default:return!1}}function c(n,t,r,a,i,o,s){var c=t.declarationList,u=!1,d=e.flatMap(c.declarations,function(t){var d=t.name,p=t.initializer;if(p){if(e.isExportsOrModuleExportsOrAlias(n,p))return u=!0,[];if(e.isRequireCall(p,!0))return u=!0,y(n,d,p.arguments[0],r,a,i,o,s);if(e.isPropertyAccessExpression(p)&&e.isRequireCall(p.expression,!0))return u=!0,l(d,p.name.text,p.expression.arguments[0],i,s)}return e.createVariableStatement(void 0,e.createVariableDeclarationList([t],c.flags))});u&&r.replaceNodeWithNodes(n,t,d)}function l(n,t,r,a,i){switch(n.kind){case 184:case 185:var o=b(t,a);return[A(o,t,r,i),L(void 0,n,e.createIdentifier(o))];case 72:return[A(n.text,t,r,i)];default:return e.Debug.assertNever(n)}}function u(n,t,r,a,i){var o=r.left,s=r.right;if(!e.isPropertyAccessExpression(o))return!1;if(e.isExportsOrModuleExportsOrAlias(n,o)){if(!e.isExportsOrModuleExportsOrAlias(n,s)){var c=e.isObjectLiteralExpression(s)?d(s):e.isRequireCall(s,!0)?m(s.arguments[0],t):void 0;return c?(a.replaceNodeWithNodes(n,r.parent,c[0]),c[1]):(a.replaceRangeWithText(n,e.createRange(o.getStart(n),s.pos),"export default"),!0)}a.delete(n,r.parent)}else e.isExportsOrModuleExportsOrAlias(n,o.expression)&&p(n,r,a,i);return!1}function d(n){var t=e.mapAllOrFail(n.properties,function(n){switch(n.kind){case 158:case 159:case 276:case 277:return;case 275:return e.isIdentifier(n.name)?v(n.name.text,n.initializer):void 0;case 156:return e.isIdentifier(n.name)?x(n.name.text,[e.createToken(85)],n):void 0;default:e.Debug.assertNever(n)}});return t&&[t,!1]}function p(n,t,r,a){var i=t.left.name.text,o=a.get(i);if(void 0!==o){var s=[L(void 0,o,t.right),k([e.createExportSpecifier(o,i)])];r.replaceNodeWithNodes(n,t.parent,s)}else g(t,n,r)}function m(n,t){var r=n.text,a=t.getSymbolAtLocation(n),i=a?a.exports:e.emptyUnderscoreEscapedMap;return i.has("export=")?[[_(r)],!0]:i.has("default")?i.size>1?[[f(r),_(r)],!0]:[[_(r)],!0]:[[f(r)],!1]}function f(e){return k(void 0,e)}function _(n){return k([e.createExportSpecifier(void 0,"default")],n)}function g(n,t,r){var a=n.left,i=n.right,o=n.parent,s=a.name.text;if(!(e.isFunctionExpression(i)||e.isArrowFunction(i)||e.isClassExpression(i))||i.name&&i.name.text!==s)r.replaceNodeRangeWithNodes(t,a.expression,e.findChildOfKind(a,24,t),[e.createToken(85),e.createToken(77)],{joiner:" ",suffix:" "});else{r.replaceRange(t,{pos:a.getStart(t),end:i.getStart(t)},e.createToken(85),{suffix:" "}),i.name||r.insertName(t,i,s);var c=e.findChildOfKind(o,26,t);c&&r.delete(t,c)}}function v(n,t){var r=[e.createToken(85)];switch(t.kind){case 196:var a=t.name;if(a&&a.text!==n)return i();case 197:return x(n,r,t);case 209:return C(n,r,t);default:return i()}function i(){return L(r,e.createIdentifier(n),t)}}function y(t,r,a,i,o,s,c,l){switch(r.kind){case 184:var u=e.mapAllOrFail(r.elements,function(n){return n.dotDotDotToken||n.initializer||n.propertyName&&!e.isIdentifier(n.propertyName)||!e.isIdentifier(n.name)?void 0:D(n.propertyName&&n.propertyName.text,n.name.text)});if(u)return[e.makeImport(void 0,u,a,l)];case 185:var d=b(n.moduleSpecifierToValidIdentifier(a.text,c),s);return[e.makeImport(e.createIdentifier(d),void 0,a,l),L(void 0,e.getSynthesizedDeepClone(r),e.createIdentifier(d))];case 72:return h(t,r,a,i,o,s,l);default:return e.Debug.assertNever(r)}}function h(n,t,r,a,i,o,s){for(var c=i.getSymbolAtLocation(t),l=e.createMap(),u=!1,d=0,p=o.original.get(t.text);d<p.length;d++){var m=p[d];if(i.getSymbolAtLocation(m)===c&&m!==t){var f=m.parent;if(e.isPropertyAccessExpression(f)){var _=f.expression,g=f.name.text;e.Debug.assert(_===m);var v=l.get(g);void 0===v&&(v=b(g,o),l.set(g,v)),a.replaceNode(n,f,e.createIdentifier(v))}else u=!0}}var y=0===l.size?void 0:e.arrayFrom(e.mapIterator(l.entries(),function(n){var t=n[0],r=n[1];return e.createImportSpecifier(t===r?void 0:e.createIdentifier(t),e.createIdentifier(r))}));return y||(u=!0),[e.makeImport(u?e.getSynthesizedDeepClone(t):void 0,y,r,s)]}function b(e,n){while(n.original.has(e)||n.additional.has(e))e="_"+e;return n.additional.set(e,!0),e}function E(n){var t=e.createMultiMap();return T(n,function(e){return t.add(e.text,e)}),t}function T(n,t){e.isIdentifier(n)&&S(n)&&t(n),n.forEachChild(function(e){return T(e,t)})}function S(e){var n=e.parent;switch(n.kind){case 189:return n.name!==e;case 186:return n.propertyName!==e;case 253:return n.propertyName!==e;default:return!0}}function x(n,t,r){return e.createFunctionDeclaration(e.getSynthesizedDeepClones(r.decorators),e.concatenate(t,e.getSynthesizedDeepClones(r.modifiers)),e.getSynthesizedDeepClone(r.asteriskToken),n,e.getSynthesizedDeepClones(r.typeParameters),e.getSynthesizedDeepClones(r.parameters),e.getSynthesizedDeepClone(r.type),e.convertToFunctionBody(e.getSynthesizedDeepClone(r.body)))}function C(n,t,r){return e.createClassDeclaration(e.getSynthesizedDeepClones(r.decorators),e.concatenate(t,e.getSynthesizedDeepClones(r.modifiers)),n,e.getSynthesizedDeepClones(r.typeParameters),e.getSynthesizedDeepClones(r.heritageClauses),e.getSynthesizedDeepClones(r.members))}function A(n,t,r,a){return"default"===t?e.makeImport(e.createIdentifier(n),void 0,r,a):e.makeImport(void 0,[D(t,n)],r,a)}function D(n,t){return e.createImportSpecifier(void 0!==n&&n!==t?e.createIdentifier(n):void 0,e.createIdentifier(t))}function L(n,t,r){return e.createVariableStatement(n,e.createVariableDeclarationList([e.createVariableDeclaration(t,void 0,r)],2))}function k(n,t){return e.createExportDeclaration(void 0,void 0,n&&e.createNamedExports(n),void 0===t?void 0:e.createLiteral(t))}n.registerCodeFix({errorCodes:[e.Diagnostics.File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module.code],getCodeActions:function(a){var i=a.sourceFile,o=a.program,s=a.preferences,c=e.textChanges.ChangeTracker.with(a,function(n){var a=r(i,o.getTypeChecker(),n,o.getCompilerOptions().target,e.getQuotePreference(i,s));if(a)for(var c=0,l=o.getSourceFiles();c<l.length;c++){var u=l[c];t(u,i,n,e.getQuotePreference(u,s))}});return[n.createCodeFixActionNoFixId("convertToEs6Module",c,e.Diagnostics.Convert_to_ES6_module)]}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="correctQualifiedNameToIndexedAccessType",r=[e.Diagnostics.Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_property_1_in_0_with_0_1.code];function a(n,t){var r=e.findAncestor(e.getTokenAtPosition(n,t),e.isQualifiedName);return e.Debug.assert(!!r,"Expected position to be owned by a qualified name."),e.isIdentifier(r.left)?r:void 0}function i(n,t,r){var a=r.right.text,i=e.createIndexedAccessTypeNode(e.createTypeReferenceNode(r.left,void 0),e.createLiteralTypeNode(e.createLiteral(a)));n.replaceNode(t,r,i)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=a(r.sourceFile,r.span.start);if(o){var s=e.textChanges.ChangeTracker.with(r,function(e){return i(e,r.sourceFile,o)}),c=o.left.text+'["'+o.right.text+'"]';return[n.createCodeFixAction(t,s,[e.Diagnostics.Rewrite_as_the_indexed_access_type_0,c],t,e.Diagnostics.Rewrite_all_as_indexed_access_types)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(e,n){var t=a(n.file,n.start);t&&i(e,n.file,t)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t=[e.Diagnostics.Class_0_incorrectly_implements_interface_1.code,e.Diagnostics.Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass.code],r="fixClassIncorrectlyImplementsInterface";function a(n,t){return e.Debug.assertDefined(e.getContainingClass(e.getTokenAtPosition(n,t)))}function i(n){return!(8&e.getModifierFlags(n.valueDeclaration))}function o(t,r,a,o,c,l){var u=t.program.getTypeChecker(),d=s(o,u),p=u.getTypeAtLocation(r),m=u.getPropertiesOfType(p),f=m.filter(e.and(i,function(e){return!d.has(e.escapedName)})),_=u.getTypeAtLocation(o);function g(e,r){var i=u.getIndexInfoOfType(e,r);i&&c.insertNodeAtClassStart(a,o,u.indexInfoToIndexSignatureDeclaration(i,r,o,void 0,n.getNoopSymbolTrackerWithResolver(t)))}_.getNumberIndexType()||g(p,1),_.getStringIndexType()||g(p,0),n.createMissingMemberNodes(o,f,t,l,function(e){return c.insertNodeAtClassStart(a,o,e)})}function s(n,t){var r=e.getEffectiveBaseTypeNode(n);if(!r)return e.createSymbolTable();var a=t.getTypeAtLocation(r),o=t.getPropertiesOfType(a);return e.createSymbolTable(o.filter(i))}n.registerCodeFix({errorCodes:t,getCodeActions:function(t){var i=t.sourceFile,s=t.span,c=a(i,s.start);return e.mapDefined(e.getClassImplementsHeritageClauseElements(c),function(a){var s=e.textChanges.ChangeTracker.with(t,function(e){return o(t,a,i,c,e,t.preferences)});return 0===s.length?void 0:n.createCodeFixAction(r,s,[e.Diagnostics.Implement_interface_0,a.getText(i)],r,e.Diagnostics.Implement_all_unimplemented_interfaces)})},fixIds:[r],getAllCodeActions:function(r){var i=e.createMap();return n.codeFixAll(r,t,function(n,t){var s=a(t.file,t.start);if(e.addToSeen(i,e.getNodeId(s)))for(var c=0,l=e.getClassImplementsHeritageClauseElements(s);c<l.length;c++){var u=l[c];o(r,u,t.file,s,n,r.preferences)}})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){n.importFixId="fixMissingImport";var t,r,a=[e.Diagnostics.Cannot_find_name_0.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_1.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0.code,e.Diagnostics.Cannot_find_namespace_0.code,e.Diagnostics._0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead.code,e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here.code];function i(n,t,r,a,i,o,l,d,p){var m=c(n,t,a,r,o.getCompilerOptions(),o.getTypeChecker(),o.getSourceFiles());e.Debug.assert(m.some(function(e){return e.moduleSymbol===t}));var f=e.first(_(o,r,d,m,i,p)).moduleSpecifier,g=e.first(u(m,a,d,o,r,i,p));return{moduleSpecifier:f,codeAction:s(L({host:i,formatContext:l},r,a,g,e.getQuotePreference(r,p)))}}function s(e){var n=e.description,t=e.changes,r=e.commands;return{description:n,changes:t,commands:r}}function c(n,t,r,a,i,o,s){var c=[];return F(o,s,function(s,u){if(!u||s===t||!e.startsWith(a.fileName,e.getDirectoryPath(u.fileName))){var d=x(s,o,i);d&&d.name===r&&e.skipAlias(d.symbol,o)===n&&c.push({moduleSymbol:s,importKind:d.kind,exportedSymbolIsTypeOnly:l(d.symbol,o)});for(var p=0,m=o.getExportsOfModule(s);p<m.length;p++){var f=m[p];f.name===r&&e.skipAlias(f,o)===n&&c.push({moduleSymbol:s,importKind:0,exportedSymbolIsTypeOnly:l(f,o)})}}}),c}function l(n,t){return!(67220415&e.skipAlias(n,t).flags)}function u(n,t,r,a,i,o,s){var c=a.getTypeChecker(),l=e.flatMap(n,function(e){return f(e,c,i)}),u=void 0===r?void 0:d(l,t,r,c),m=p(l),_=m?[m]:g(n,l,a,i,r,o,s);return(u?[u]:e.emptyArray).concat(_)}function d(n,t,r,a){return e.firstDefined(n,function(n){var i=n.declaration,o=m(i);if(o){var s=a.getAliasedSymbol(a.getSymbolAtLocation(o));if(s&&s.exports.has(e.escapeLeadingUnderscores(t)))return{kind:0,namespacePrefix:o.text,position:r}}})}function p(n){return e.firstDefined(n,function(e){var n=e.declaration,t=e.importKind;if(249===n.kind){var r=n.importClause;if(r){var a=r.name,i=r.namedBindings;return 1===t&&!a||0===t&&(!i||252===i.kind)?{kind:2,importClause:r,importKind:t}:void 0}}})}function m(n){if(249===n.kind){var t=n.importClause&&e.isImportClause(n.importClause)&&n.importClause.namedBindings;return t&&251===t.kind?t.name:void 0}return n.name}function f(n,t,r){var a=n.moduleSymbol,i=n.importKind,o=n.exportedSymbolIsTypeOnly;return o&&e.isSourceFileJS(r)?e.emptyArray:e.mapDefined(r.imports,function(n){var r=e.importFromModuleSpecifier(n);return 249!==r.kind&&248!==r.kind||t.getSymbolAtLocation(n)!==a?void 0:{declaration:r,importKind:i}})}function _(n,t,r,a,i,o){var s=e.isSourceFileJS(t),c=e.flatMap(a,function(a){var c=a.moduleSymbol,l=a.importKind,u=a.exportedSymbolIsTypeOnly;return e.moduleSpecifiers.getModuleSpecifiers(c,n.getCompilerOptions(),t,i,n.getSourceFiles(),o,n.redirectTargetsMap).map(function(n){return u&&s?{kind:1,moduleSpecifier:n,position:e.Debug.assertDefined(r)}:{kind:3,moduleSpecifier:n,importKind:l}})});return e.sort(c,function(e,n){return e.moduleSpecifier.length-n.moduleSpecifier.length})}function g(n,t,r,a,i,o,s){var c=e.firstDefined(t,v);return c?[c]:_(r,a,i,n,o,s)}function v(n){var t=n.declaration,r=n.importKind,a=249===t.kind?t.moduleSpecifier:259===t.moduleReference.kind?t.moduleReference.expression:void 0;return a&&e.isStringLiteral(a)?{kind:3,moduleSpecifier:a.text,importKind:r}:void 0}function y(n,t,r){var a=e.getTokenAtPosition(n.sourceFile,r),i=t===e.Diagnostics._0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead.code?h(n,a):e.isIdentifier(a)?T(n,a):void 0;return i&&o({},i,{fixes:e.sort(i.fixes,function(e,n){return e.kind-n.kind})})}function h(n,t){var r=n.sourceFile,a=n.program,i=n.host,o=n.preferences,s=a.getTypeChecker(),c=b(t,s);if(c){var l=s.getAliasedSymbol(c),d=c.name,p=[{moduleSymbol:l,importKind:E(a.getCompilerOptions()),exportedSymbolIsTypeOnly:!1}],m=u(p,d,e.isIdentifier(t)?t.getStart(r):void 0,a,r,i,o);return{fixes:m,symbolName:d}}}function b(n,t){var r=e.isIdentifier(n)?t.getSymbolAtLocation(n):void 0;if(e.isUMDExportSymbol(r))return r;var a=n.parent;return e.isJsxOpeningLikeElement(a)&&a.tagName===n||e.isJsxOpeningFragment(a)?e.tryCast(t.resolveName(t.getJsxNamespace(a),e.isJsxOpeningLikeElement(a)?n:a,67220415,!1),e.isUMDExportSymbol):void 0}function E(n){if(e.getAllowSyntheticDefaultImports(n))return 1;var t=e.getEmitModuleKind(n);switch(t){case e.ModuleKind.AMD:case e.ModuleKind.CommonJS:case e.ModuleKind.UMD:return 3;case e.ModuleKind.System:case e.ModuleKind.ES2015:case e.ModuleKind.ESNext:case e.ModuleKind.None:return 2;default:return e.Debug.assertNever(t)}}function T(n,t){var r=n.sourceFile,a=n.program,i=n.cancellationToken,o=n.host,s=n.preferences,c=a.getTypeChecker(),l=e.isJsxOpeningLikeElement(t.parent)&&t.parent.tagName===t&&(e.isIntrinsicJsxName(t.text)||c.resolveName(t.text,t,67108863,!1))?c.getJsxNamespace(r):t.text;e.Debug.assert("default"!==l);var d=e.arrayFrom(e.flatMapIterator(S(l,e.getMeaningFromLocation(t),i,r,c,a).entries(),function(e){e[0];var n=e[1];return u(n,l,t.getStart(r),a,r,o,s)}));return{fixes:d,symbolName:l}}function S(n,t,r,a,i,o){var s=e.createMultiMap();function c(n,t,r){s.add(e.getUniqueSymbolId(t,i).toString(),{moduleSymbol:n,importKind:r,exportedSymbolIsTypeOnly:l(t,i)})}return R(i,a,o.getSourceFiles(),function(e){r.throwIfCancellationRequested();var a=x(e,i,o.getCompilerOptions());a&&a.name===n&&P(a.symbolForMeaning,t)&&c(e,a.symbol,a.kind);var s=i.tryGetMemberInModuleExportsAndProperties(n,e);s&&P(s,t)&&c(e,s,0)}),s}function x(e,n,t){var r=C(e,n);if(r){var a=r.symbol,i=r.kind,s=A(a,e,n,t);return s&&o({symbol:a,kind:i},s)}}function C(e,n){var t=n.tryGetMemberInModuleExports("default",e);if(t)return{symbol:t,kind:1};var r=n.resolveExternalModuleSymbol(e);return r===e?void 0:{symbol:r,kind:3}}function A(n,t,r,a){var i=e.getLocalSymbolForExportDefault(n);if(i)return{symbolForMeaning:i,name:i.name};var o=D(n);if(void 0!==o)return{symbolForMeaning:n,name:o};if(2097152&n.flags){var s=r.getImmediateAliasedSymbol(n);return s&&A(s,e.Debug.assertDefined(s.parent),r,a)}return{symbolForMeaning:n,name:V(t,a.target)}}function D(n){return n.declarations&&e.firstDefined(n.declarations,function(n){if(e.isExportAssignment(n)){if(e.isIdentifier(n.expression))return n.expression.text}else if(e.isExportSpecifier(n))return e.Debug.assert("default"===n.name.text),n.propertyName&&n.propertyName.text})}function L(t,r,a,i,o){var s,c=e.textChanges.ChangeTracker.with(t,function(e){s=k(e,r,a,i,o)});return n.createCodeFixAction("import",c,s,n.importFixId,e.Diagnostics.Add_all_missing_imports)}function k(n,t,r,a,i){switch(a.kind){case 0:return I(n,t,a),[e.Diagnostics.Change_0_to_1,r,a.namespacePrefix+"."+r];case 1:return w(n,t,a,i),[e.Diagnostics.Change_0_to_1,r,M(a.moduleSpecifier,i)+r];case 2:var o=a.importClause,s=a.importKind;N(n,t,o,1===s?r:void 0,0===s?[r]:e.emptyArray);var c=e.stripQuotes(o.parent.moduleSpecifier.getText());return[1===s?e.Diagnostics.Add_default_import_0_to_existing_import_declaration_from_1:e.Diagnostics.Add_0_to_existing_import_declaration_from_1,r,c];case 3:s=a.importKind;var l=a.moduleSpecifier;return O(n,t,l,i,1===s?{defaultImport:r,namedImports:e.emptyArray,namespaceLikeImport:void 0}:0===s?{defaultImport:void 0,namedImports:[r],namespaceLikeImport:void 0}:{defaultImport:void 0,namedImports:e.emptyArray,namespaceLikeImport:{importKind:s,name:r}}),[1===s?e.Diagnostics.Import_default_0_from_module_1:e.Diagnostics.Import_0_from_module_1,r,l];default:return e.Debug.assertNever(a)}}function N(n,t,r,a,i){if(a&&(e.Debug.assert(!r.name),n.insertNodeAt(t,r.getStart(t),e.createIdentifier(a),{suffix:", "})),i.length){var o=i.map(function(n){return e.createImportSpecifier(void 0,e.createIdentifier(n))});if(r.namedBindings&&e.cast(r.namedBindings,e.isNamedImports).elements.length)for(var s=0,c=o;s<c.length;s++){var l=c[s];n.insertNodeInListAfter(t,e.last(e.cast(r.namedBindings,e.isNamedImports).elements),l)}else if(o.length){var u=e.createNamedImports(o);r.namedBindings?n.replaceNode(t,r.namedBindings,u):n.insertNodeAfter(t,e.Debug.assertDefined(r.name),u)}}}function I(e,n,t){var r=t.namespacePrefix,a=t.position;e.insertText(n,a,r+".")}function w(e,n,t,r){var a=t.moduleSpecifier,i=t.position;e.insertText(n,i,M(a,r))}function M(n,t){var r=e.getQuoteFromPreference(t);return"import("+r+n+r+")."}function O(n,t,r,a,i){var o=i.defaultImport,s=i.namedImports,c=i.namespaceLikeImport,l=e.makeStringLiteral(r,a);(void 0!==o||s.length)&&e.insertImport(n,t,e.makeImport(void 0===o?void 0:e.createIdentifier(o),s.map(function(n){return e.createImportSpecifier(void 0,e.createIdentifier(n))}),r,a)),c&&e.insertImport(n,t,3===c.importKind?e.createImportEqualsDeclaration(void 0,void 0,e.createIdentifier(c.name),e.createExternalModuleReference(l)):e.createImportDeclaration(void 0,void 0,e.createImportClause(void 0,e.createNamespaceImport(e.createIdentifier(c.name))),l))}function P(n,t){var r=n.declarations;return e.some(r,function(n){return!!(e.getMeaningFromDeclaration(n)&t)})}function R(e,n,t,r){F(e,t,function(e,t){(void 0===t||t!==n&&G(n.fileName,t.fileName))&&r(e)})}function F(n,t,r){for(var a=0,i=n.getAmbientModules();a<i.length;a++){var o=i[a];r(o,void 0)}for(var s=0,c=t;s<c.length;s++){var l=c[s];e.isExternalOrCommonJsModule(l)&&r(n.getMergedSymbol(l.symbol),l)}}function G(n,t){var r=e.forEachAncestorDirectory(t,function(n){return"node_modules"===e.getBaseFileName(n)?n:void 0});return void 0===r||e.startsWith(n,e.getDirectoryPath(r))}function V(n,t){return B(e.removeFileExtension(e.stripQuotes(n.name)),t)}function B(n,t){var r=e.getBaseFileName(e.removeSuffix(n,"/index")),a="",i=!0,o=r.charCodeAt(0);e.isIdentifierStart(o,t)?a+=String.fromCharCode(o):i=!1;for(var s=1;s<r.length;s++){var c=r.charCodeAt(s),l=e.isIdentifierPart(c,t);if(l){var u=String.fromCharCode(c);i||(u=u.toUpperCase()),a+=u}i=l}return e.isStringANonContextualKeyword(a)?"_"+a:a||"_"}n.registerCodeFix({errorCodes:a,getCodeActions:function(n){var t=n.errorCode,r=n.preferences,a=n.sourceFile,i=n.span,o=y(n,t,i.start);if(o){var s=o.fixes,c=o.symbolName,l=e.getQuotePreference(a,r);return s.map(function(e){return L(n,a,c,e,l)})}},fixIds:[n.importFixId],getAllCodeActions:function(t){var r=t.sourceFile,i=t.preferences,o=[],s=[],c=e.createMap(),l=e.createMap();return n.eachDiagnostic(t,a,function(n){var r=y(t,n.code,n.start);if(r&&r.fixes.length){var a=r.fixes,i=r.symbolName,u=e.first(a);switch(u.kind){case 0:o.push(u);break;case 1:s.push(u);break;case 2:var d=u.importClause,p=u.importKind,m=String(e.getNodeId(d)),f=c.get(m);f||c.set(m,f={importClause:d,defaultImport:void 0,namedImports:[]}),0===p?e.pushIfUnique(f.namedImports,i):(e.Debug.assert(void 0===f.defaultImport||f.defaultImport===i),f.defaultImport=i);break;case 3:var _=u.moduleSpecifier;p=u.importKind,f=l.get(_);switch(f||l.set(_,f={defaultImport:void 0,namedImports:[],namespaceLikeImport:void 0}),p){case 1:e.Debug.assert(void 0===f.defaultImport||f.defaultImport===i),f.defaultImport=i;break;case 0:e.pushIfUnique(f.namedImports,i);break;case 3:case 2:e.Debug.assert(void 0===f.namespaceLikeImport||f.namespaceLikeImport.name===i),f.namespaceLikeImport={importKind:p,name:i};break}break;default:e.Debug.assertNever(u)}}}),n.createCombinedCodeActions(e.textChanges.ChangeTracker.with(t,function(n){for(var t=e.getQuotePreference(r,i),a=0,u=o;a<u.length;a++){var d=u[a];I(n,r,d)}for(var p=0,m=s;p<m.length;p++){d=m[p];w(n,r,d,t)}c.forEach(function(e){var t=e.importClause,a=e.defaultImport,i=e.namedImports;N(n,r,t,a,i)}),l.forEach(function(e,a){O(n,r,a,t,e)})}))}}),function(e){e[e["UseNamespace"]=0]="UseNamespace",e[e["ImportType"]=1]="ImportType",e[e["AddToExisting"]=2]="AddToExisting",e[e["AddNew"]=3]="AddNew"}(t||(t={})),function(e){e[e["Named"]=0]="Named",e[e["Default"]=1]="Default",e[e["Namespace"]=2]="Namespace",e[e["Equals"]=3]="Equals"}(r||(r={})),n.getImportCompletionAction=i,n.forEachExternalModuleToImportFrom=R,n.moduleSymbolToValidIdentifier=V,n.moduleSpecifierToValidIdentifier=B})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="fixSpelling",r=[e.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_1.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0.code,e.Diagnostics.Module_0_has_no_exported_member_1_Did_you_mean_2.code];function a(n,t,r){var a,i=e.getTokenAtPosition(n,t),c=r.program.getTypeChecker();if(e.isPropertyAccessExpression(i.parent)&&i.parent.name===i){e.Debug.assert(72===i.kind);var l=c.getTypeAtLocation(i.parent.expression);a=c.getSuggestionForNonexistentProperty(i,l)}else if(e.isImportSpecifier(i.parent)&&i.parent.name===i){e.Debug.assert(72===i.kind);var u=e.findAncestor(i,e.isImportDeclaration),d=s(n,r,u);d&&d.symbol&&(a=c.getSuggestionForNonexistentExport(i,d.symbol))}else{var p=e.getMeaningFromLocation(i),m=e.getTextOfNode(i);e.Debug.assert(void 0!==m,"name should be defined"),a=c.getSuggestionForNonexistentSymbol(i,m,o(p))}return void 0===a?void 0:{node:i,suggestion:a}}function i(n,t,r,a,i){!e.isIdentifierText(a,i)&&e.isPropertyAccessExpression(r.parent)?n.replaceNode(t,r.parent,e.createElementAccess(r.parent.expression,e.createLiteral(a))):n.replaceNode(t,r,e.createIdentifier(a))}function o(e){var n=0;return 4&e&&(n|=1920),2&e&&(n|=67897832),1&e&&(n|=67220415),n}function s(n,t,r){if(r&&e.isStringLiteralLike(r.moduleSpecifier)){var a=e.getResolvedModule(n,r.moduleSpecifier.text);if(a)return t.program.getSourceFile(a.resolvedFileName)}}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=a(o,r.span.start,r);if(s){var c=s.node,l=s.suggestion,u=r.host.getCompilationSettings().target,d=e.textChanges.ChangeTracker.with(r,function(e){return i(e,o,c,l,u)});return[n.createCodeFixAction("spelling",d,[e.Diagnostics.Change_spelling_to_0,l],t,e.Diagnostics.Fix_all_detected_spelling_errors)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(n,t){var r=a(t.file,t.start,e),o=e.host.getCompilationSettings().target;r&&i(n,e.sourceFile,r.node,r.suggestion,o)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t,r="addMissingMember",a=[e.Diagnostics.Property_0_does_not_exist_on_type_1.code,e.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2.code,e.Diagnostics.Property_0_is_missing_in_type_1_but_required_in_type_2.code,e.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2.code,e.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more.code],i="addMissingMember";function o(n,t){var r=[];while(n){var a=e.getClassExtendsHeritageElement(n),i=a&&t.getSymbolAtLocation(a.expression),o=i&&e.find(i.declarations,e.isClassLike);o&&r.push(o),n=o}return r}function s(n,t,r){var a=e.getTokenAtPosition(n,t);if(e.isIdentifier(a)){var i=a.parent;if(e.isPropertyAccessExpression(i)){var o=e.skipConstraint(r.getTypeAtLocation(i.expression)),s=o.symbol;if(s&&s.declarations){var c=e.find(s.declarations,e.isClassLike)||e.find(s.declarations,e.isInterfaceDeclaration);if(c){var l=(o.target||o)!==r.getDeclaredTypeOfSymbol(s),u=c.getSourceFile(),d=e.isSourceFileJS(u),p=e.tryCast(i.parent,e.isCallExpression);return{kind:1,token:a,parentDeclaration:c,makeStatic:l,declSourceFile:u,inJs:d,call:p}}var m=e.find(s.declarations,e.isEnumDeclaration);return m?{kind:0,token:a,parentDeclaration:m}:void 0}}}}function c(t,a,o,s,c){var u=e.textChanges.ChangeTracker.with(t,function(e){return l(e,a,o,s,c)});return 0===u.length?void 0:n.createCodeFixAction(r,u,[c?e.Diagnostics.Initialize_static_property_0:e.Diagnostics.Initialize_property_0_in_the_constructor,s],i,e.Diagnostics.Add_all_missing_members)}function l(n,t,r,a,i){if(i){if(209===r.kind)return;var o=r.name.getText(),s=u(e.createIdentifier(o),a);n.insertNodeAfter(t,r,s)}else{var c=e.getFirstConstructorWithBody(r);if(!c)return;var l=u(e.createThis(),a);n.insertNodeAtConstructorEnd(t,c,l)}}function u(n,t){return e.createStatement(e.createAssignment(e.createPropertyAccess(n,t),e.createIdentifier("undefined")))}function d(e,n,t,r,a){var i=p(e.program.getTypeChecker(),t,r),o=m(e,n,t,a,r.text,i);return a?[o]:[o,g(e,n,t,r.text,i)]}function p(n,t,r){var a;if(204===r.parent.parent.kind){var i=r.parent.parent,o=r.parent===i.left?i.right:i.left,s=n.getWidenedType(n.getBaseTypeOfLiteralType(n.getTypeAtLocation(o)));a=n.typeToTypeNode(s,t)}else{var c=n.getContextualType(r.parent);a=c?n.typeToTypeNode(c):void 0}return a||e.createKeywordTypeNode(120)}function m(t,a,o,s,c,l){var u=e.textChanges.ChangeTracker.with(t,function(e){return f(e,a,o,c,l,s)});return n.createCodeFixAction(r,u,[s?e.Diagnostics.Declare_static_property_0:e.Diagnostics.Declare_property_0,c],i,e.Diagnostics.Add_all_missing_members)}function f(n,t,r,a,i,o){var s=e.createProperty(void 0,o?[e.createToken(116)]:void 0,a,void 0,i,void 0),c=_(r);c?n.insertNodeAfter(t,c,s):n.insertNodeAtClassStart(t,r,s)}function _(n){for(var t,r=0,a=n.members;r<a.length;r++){var i=a[r];if(!e.isPropertyDeclaration(i))break;t=i}return t}function g(t,a,i,o,s){var c=e.createKeywordTypeNode(138),l=e.createParameter(void 0,void 0,void 0,"x",void 0,c,void 0),u=e.createIndexSignature(void 0,void 0,[l],s),d=e.textChanges.ChangeTracker.with(t,function(e){return e.insertNodeAtClassStart(a,i,u)});return n.createCodeFixActionNoFixId(r,d,[e.Diagnostics.Add_index_signature_for_property_0,o])}function v(t,a,o,s,c,l,u,d){var p=e.textChanges.ChangeTracker.with(t,function(e){return y(t,e,a,o,s,c,l,u,d)});return n.createCodeFixAction(r,p,[l?e.Diagnostics.Declare_static_method_0:e.Diagnostics.Declare_method_0,s.text],i,e.Diagnostics.Add_all_missing_members)}function y(t,r,a,i,o,s,c,l,u){var d=n.createMethodFromCallExpression(t,s,o.text,l,c,u,i),p=e.getAncestor(s,156);p&&p.parent===i?r.insertNodeAfter(a,p,d):r.insertNodeAtClassStart(a,i,d)}function h(n,t,r,a){var i=e.some(a.members,function(e){var n=t.getTypeAtLocation(e);return!!(n&&132&n.flags)}),o=e.createEnumMember(r,i?e.createStringLiteral(r.text):void 0);n.replaceNode(a.getSourceFile(),a,e.updateEnumDeclaration(a,a.decorators,a.modifiers,a.name,e.concatenate(a.members,e.singleElementArray(o))))}n.registerCodeFix({errorCodes:a,getCodeActions:function(t){var a=s(t.sourceFile,t.span.start,t.program.getTypeChecker());if(a){if(0===a.kind){var o=a.token,l=a.parentDeclaration,u=e.textChanges.ChangeTracker.with(t,function(e){return h(e,t.program.getTypeChecker(),o,l)});return[n.createCodeFixAction(r,u,[e.Diagnostics.Add_missing_enum_member_0,o.text],i,e.Diagnostics.Add_all_missing_members)]}var p=a.parentDeclaration,m=a.declSourceFile,f=a.inJs,_=a.makeStatic,g=a.token,y=a.call,b=y&&v(t,m,p,g,y,_,f,t.preferences),E=f&&!e.isInterfaceDeclaration(p)?e.singleElementArray(c(t,m,p,g.text,_)):d(t,m,p,g,_);return e.concatenate(e.singleElementArray(b),E)}},fixIds:[i],getAllCodeActions:function(t){var r=t.program,i=t.preferences,c=r.getTypeChecker(),u=e.createMap(),d=new e.NodeMap;return n.createCombinedCodeActions(e.textChanges.ChangeTracker.with(t,function(m){n.eachDiagnostic(t,a,function(n){var t=s(n.file,n.start,c);if(t&&e.addToSeen(u,e.getNodeId(t.parentDeclaration)+"#"+t.token.text))if(0===t.kind){var r=t.token,a=t.parentDeclaration;h(m,c,r,a)}else{a=t.parentDeclaration;var i=t.token,o=d.getOrUpdate(a,function(){return[]});o.some(function(e){return e.token.text===i.text})||o.push(t)}}),d.forEach(function(n,a){for(var s=o(a,c),u=function(n){if(s.some(function(e){var t=d.get(e);return!!t&&t.some(function(e){var t=e.token;return t.text===n.token.text})}))return"continue";var a=n.parentDeclaration,o=n.declSourceFile,c=n.inJs,u=n.makeStatic,_=n.token,g=n.call;if(g)y(t,m,o,a,_,g,u,c,i);else if(c&&!e.isInterfaceDeclaration(a))l(m,o,a,_.text,u);else{var v=p(r.getTypeChecker(),a,_);f(m,o,a,_.text,v,u)}},_=0,g=n;_<g.length;_++){var v=g[_];u(v)}})}))}}),function(e){e[e["Enum"]=0]="Enum",e[e["ClassOrInterface"]=1]="ClassOrInterface"}(t||(t={}))})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="addMissingNewOperator",r=[e.Diagnostics.Value_of_type_0_is_not_callable_Did_you_mean_to_include_new.code];function a(n,t,r){var a=e.cast(i(t,r),e.isCallExpression),o=e.createNew(a.expression,a.typeArguments,a.arguments);n.replaceNode(t,a,o)}function i(n,t){var r=e.getTokenAtPosition(n,t.start),a=e.textSpanEnd(t);while(r.end<a)r=r.parent;return r}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=r.sourceFile,o=r.span,s=e.textChanges.ChangeTracker.with(r,function(e){return a(e,i,o)});return[n.createCodeFixAction(t,s,e.Diagnostics.Add_missing_new_operator_to_call,t,e.Diagnostics.Add_missing_new_operator_to_all_calls)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(n,t){return a(n,e.sourceFile,t)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="fixCannotFindModule",r="installTypesPackage",a="generateTypes",i=e.Diagnostics.Cannot_find_module_0.code,o=[i,e.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type.code];function s(r,i){var o,s=e.textChanges.ChangeTracker.with(r,function(e){var n=l(e,r);o=void 0===n?void 0:c(n,i,r)});return o&&n.createCodeFixAction(t,s,[e.Diagnostics.Generate_types_for_0,i],a,e.Diagnostics.Generate_types_for_all_packages_without_types,o)}function c(n,t,r){var a=r.sourceFile.fileName,i=e.tryResolveJSModule(t,e.getDirectoryPath(a),r.host);if(void 0!==i){var o=e.resolvePath(e.getDirectoryPath(r.program.getCompilerOptions().configFile.fileName),n,t+".d.ts");if(!r.host.fileExists(o))return{type:"generate types",file:a,fileToGenerateTypesFor:i,outputFileName:o}}}function l(t,r){var a=r.program.getCompilerOptions().configFile;if(a){var i=e.getTsConfigObjectLiteralExpression(a);if(i){var o=n.findJsonProperty(i,"compilerOptions");if(!o){var s=e.createObjectLiteral([d(),_()]);return t.insertNodeAtObjectStart(a,i,n.createJsonPropertyAssignment("compilerOptions",s)),m}var c=o.initializer;if(!e.isObjectLiteralExpression(c))return m;var l=p(t,a,c),u=g(t,a,c);return e.combinePaths(l,u)}}}n.registerCodeFix({errorCodes:o,getCodeActions:function(a){var i=a.host,o=a.sourceFile,c=a.span.start,l=y(o,c);if(void 0!==l){var u=h(l,i,a.errorCode);return void 0===u?e.singleElementArray(s(a,l)):[n.createCodeFixAction(t,[],[e.Diagnostics.Install_0,u],r,e.Diagnostics.Install_all_missing_types_packages,v(o.fileName,u))]}},fixIds:[r,a],getAllCodeActions:function(t){var i=null;return n.codeFixAll(t,o,function(n,o,s){var u=y(o.file,o.start);if(void 0!==u)switch(t.fixId){case r:var d=h(u,t.host,o.code);d&&s.push(v(o.file.fileName,d));break;case a:var p=null!==i?i:i=l(n,t),m=void 0===p?void 0:c(p,u,t);m&&s.push(m);break;default:e.Debug.fail("Bad fixId: "+t.fixId)}})}});var u=".";function d(){return n.createJsonPropertyAssignment("baseUrl",e.createStringLiteral(u))}function p(t,r,a){var i=n.findJsonProperty(a,"baseUrl");return i?e.isStringLiteral(i.initializer)?i.initializer.text:u:(t.insertNodeAtObjectStart(r,a,d()),u)}var m="types";function f(){return n.createJsonPropertyAssignment("*",e.createArrayLiteral([e.createStringLiteral(m+"/*")]))}function _(){return n.createJsonPropertyAssignment("paths",e.createObjectLiteral([f()]))}function g(t,r,a){var i=n.findJsonProperty(a,"paths");if(!i||!e.isObjectLiteralExpression(i.initializer))return t.insertNodeAtObjectStart(r,a,_()),m;var o=e.firstDefined(i.initializer.properties,function(n){return e.isPropertyAssignment(n)&&e.isStringLiteral(n.name)&&"*"===n.name.text&&e.isArrayLiteralExpression(n.initializer)?e.firstDefined(n.initializer.elements,function(n){return e.isStringLiteral(n)?e.tryRemoveSuffix(n.text,"/*"):void 0}):void 0});return o||(t.insertNodeAtObjectStart(r,i.initializer,f()),m)}function v(e,n){return{type:"install package",file:e,packageName:n}}function y(n,t){var r=e.cast(e.getTokenAtPosition(n,t),e.isStringLiteral).text,a=e.parsePackageName(r).packageName;return e.isExternalModuleNameRelative(a)?void 0:a}function h(n,t,r){return r===i?e.JsTyping.nodeCoreModules.has(n)?"@types/node":void 0:t.isKnownTypesPackageName(n)?e.getTypesPackageName(n):void 0}})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t=[e.Diagnostics.Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2.code,e.Diagnostics.Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1.code],r="fixClassDoesntImplementInheritedAbstractMember";function a(n,t){var r=e.getTokenAtPosition(n,t);return e.cast(r.parent,e.isClassLike)}function i(t,r,a,i,s){var c=e.getEffectiveBaseTypeNode(t),l=a.program.getTypeChecker(),u=l.getTypeAtLocation(c),d=l.getPropertiesOfType(u).filter(o);n.createMissingMemberNodes(t,d,a,s,function(e){return i.insertNodeAtClassStart(r,t,e)})}function o(n){var t=e.getModifierFlags(e.first(n.getDeclarations()));return!(8&t)&&!!(128&t)}n.registerCodeFix({errorCodes:t,getCodeActions:function(t){var o=t.sourceFile,s=t.span,c=e.textChanges.ChangeTracker.with(t,function(e){return i(a(o,s.start),o,t,e,t.preferences)});return 0===c.length?void 0:[n.createCodeFixAction(r,c,e.Diagnostics.Implement_inherited_abstract_class,r,e.Diagnostics.Implement_all_inherited_abstract_classes)]},fixIds:[r],getAllCodeActions:function(r){var o=e.createMap();return n.codeFixAll(r,t,function(n,t){var s=a(t.file,t.start);e.addToSeen(o,e.getNodeId(s))&&i(s,r.sourceFile,r,n,r.preferences)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="classSuperMustPrecedeThisAccess",r=[e.Diagnostics.super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class.code];function a(e,n,t,r){e.insertNodeAtConstructorStart(n,t,r),e.delete(n,r)}function i(n,t){var r=e.getTokenAtPosition(n,t);if(100===r.kind){var a=e.getContainingFunction(r),i=o(a.body);return i&&!i.expression.arguments.some(function(n){return e.isPropertyAccessExpression(n)&&n.expression===r})?{constructor:a,superCall:i}:void 0}}function o(n){return e.isExpressionStatement(n)&&e.isSuperCall(n.expression)?n:e.isFunctionLike(n)?void 0:e.forEachChild(n,o)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=r.span,c=i(o,s.start);if(c){var l=c.constructor,u=c.superCall,d=e.textChanges.ChangeTracker.with(r,function(e){return a(e,o,l,u)});return[n.createCodeFixAction(t,d,e.Diagnostics.Make_super_call_the_first_statement_in_the_constructor,t,e.Diagnostics.Make_all_super_calls_the_first_statement_in_their_constructor)]}},fixIds:[t],getAllCodeActions:function(t){var o=t.sourceFile,s=e.createMap();return n.codeFixAll(t,r,function(n,t){var r=i(t.file,t.start);if(r){var c=r.constructor,l=r.superCall;e.addToSeen(s,e.getNodeId(c.parent))&&a(n,o,c,l)}})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="constructorForDerivedNeedSuperCall",r=[e.Diagnostics.Constructors_for_derived_classes_must_contain_a_super_call.code];function a(n,t){var r=e.getTokenAtPosition(n,t);return e.Debug.assert(124===r.kind),r.parent}function i(n,t,r){var a=e.createStatement(e.createCall(e.createSuper(),void 0,e.emptyArray));n.insertNodeAtConstructorStart(t,r,a)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=r.span,c=a(o,s.start),l=e.textChanges.ChangeTracker.with(r,function(e){return i(e,o,c)});return[n.createCodeFixAction(t,l,e.Diagnostics.Add_missing_super_call,t,e.Diagnostics.Add_all_missing_super_calls)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(n,t){return i(n,e.sourceFile,a(t.file,t.start))})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="enableExperimentalDecorators",r=[e.Diagnostics.Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_to_remove_this_warning.code];function a(t,r){n.setJsonCompilerOptionValue(t,r,"experimentalDecorators",e.createTrue())}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=r.program.getCompilerOptions().configFile;if(void 0!==i){var o=e.textChanges.ChangeTracker.with(r,function(e){return a(e,i)});return[n.createCodeFixActionNoFixId(t,o,e.Diagnostics.Enable_the_experimentalDecorators_option_in_your_configuration_file)]}},fixIds:[t]})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="extendsInterfaceBecomesImplements",r=[e.Diagnostics.Cannot_extend_an_interface_0_Did_you_mean_implements.code];function a(n,t){var r=e.getTokenAtPosition(n,t),a=e.getContainingClass(r).heritageClauses,i=a[0].getFirstToken();return 86===i.kind?{extendsToken:i,heritageClauses:a}:void 0}function i(n,t,r,a){if(n.replaceNode(t,r,e.createToken(109)),2===a.length&&86===a[0].token&&109===a[1].token){var i=a[1].getFirstToken(),o=i.getFullStart();n.replaceRange(t,{pos:o,end:o},e.createToken(27));var s=t.text,c=i.end;while(c<s.length&&e.isWhiteSpaceSingleLine(s.charCodeAt(c)))c++;n.deleteRange(t,{pos:i.getStart(),end:c})}}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=a(o,r.span.start);if(s){var c=s.extendsToken,l=s.heritageClauses,u=e.textChanges.ChangeTracker.with(r,function(e){return i(e,o,c,l)});return[n.createCodeFixAction(t,u,e.Diagnostics.Change_extends_to_implements,t,e.Diagnostics.Change_all_extended_interfaces_to_implements)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(e,n){var t=a(n.file,n.start);t&&i(e,n.file,t.extendsToken,t.heritageClauses)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="forgottenThisPropertyAccess",r=e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0.code,a=[e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0.code,r];function i(n,t,a){var i=e.getTokenAtPosition(n,t);if(e.isIdentifier(i))return{node:i,className:a===r?e.getContainingClass(i).name.text:void 0}}function o(n,t,r){var a=r.node,i=r.className;e.suppressLeadingAndTrailingTrivia(a),n.replaceNode(t,a,e.createPropertyAccess(i?e.createIdentifier(i):e.createThis(),a))}n.registerCodeFix({errorCodes:a,getCodeActions:function(r){var a=r.sourceFile,s=i(a,r.span.start,r.errorCode);if(s){var c=e.textChanges.ChangeTracker.with(r,function(e){return o(e,a,s)});return[n.createCodeFixAction(t,c,[e.Diagnostics.Add_0_to_unresolved_variable,s.className||"this"],t,e.Diagnostics.Add_qualifier_to_all_unresolved_variables_matching_a_member_name)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,a,function(n,t){var r=i(t.file,t.start,t.code);r&&o(n,e.sourceFile,r)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="unusedIdentifier",r="unusedIdentifier_prefix",a="unusedIdentifier_delete",i="unusedIdentifier_infer",o=[e.Diagnostics._0_is_declared_but_its_value_is_never_read.code,e.Diagnostics._0_is_declared_but_never_used.code,e.Diagnostics.Property_0_is_declared_but_its_value_is_never_read.code,e.Diagnostics.All_imports_in_import_declaration_are_unused.code,e.Diagnostics.All_destructured_elements_are_unused.code,e.Diagnostics.All_variables_are_unused.code,e.Diagnostics.All_type_parameters_are_unused.code];function s(n,t,r){n.replaceNode(t,r.parent,e.createKeywordTypeNode(143))}function c(r,i){return n.createCodeFixAction(t,r,i,a,e.Diagnostics.Delete_all_unused_declarations)}function l(n,t,r){n.delete(t,e.Debug.assertDefined(e.cast(r.parent,e.isDeclarationWithTypeParameterChildren).typeParameters))}function u(n){return 92===n.kind?e.tryCast(n.parent,e.isImportDeclaration):void 0}function d(n,t,r,a,i,o){if(18!==n.kind||!e.isObjectBindingPattern(n.parent))return!1;var s=n.parent.parent;return 151===s.kind?y(t,r,s,a,i,o):t.delete(r,s),!0}function p(n,t,r){var a=e.tryCast(t.parent,e.isVariableDeclarationList);return!(!a||a.getChildren(n)[0]!==t)&&(r.delete(n,219===a.parent.kind?a.parent:a),!0)}function m(n,t,r,a){t!==e.Diagnostics.Property_0_is_declared_but_its_value_is_never_read.code&&(127===a.kind&&(a=e.cast(a.parent,e.isInferTypeNode).typeParameter.name),e.isIdentifier(a)&&f(a)&&n.replaceNode(r,a,e.createIdentifier("_"+a.text)))}function f(e){switch(e.parent.kind){case 151:case 150:return!0;case 237:var n=e.parent;switch(n.parent.parent.kind){case 227:case 226:return!0}}return!1}function _(n,t,r,a,i,o){v(t,r,n,a,i,o),e.isIdentifier(t)&&g(r,n,t,a)}function g(n,t,r,a){e.FindAllReferences.Core.eachSymbolReferenceInFile(r,a,t,function(r){e.isPropertyAccessExpression(r.parent)&&r.parent.name===r&&(r=r.parent),e.isBinaryExpression(r.parent)&&e.isExpressionStatement(r.parent.parent)&&r.parent.left===r&&n.delete(t,r.parent.parent)})}function v(n,t,r,a,i,o){var s=n.parent;e.isParameter(s)?y(t,r,s,a,i,o):t.delete(r,e.isImportClause(s)?n:e.isComputedPropertyName(s)?s.parent:s)}function y(n,t,r,a,i,o){h(r,a,o)&&(r.modifiers&&r.modifiers.length>0&&(!e.isIdentifier(r.name)||e.FindAllReferences.Core.isSymbolReferencedInFile(r.name,a,t))?r.modifiers.forEach(function(e){n.deleteModifier(t,e)}):(n.delete(t,r),b(n,t,r,i,a)))}function h(n,t,r){var a=n.parent;switch(a.kind){case 156:var i=t.getSymbolAtLocation(a.name);if(e.isMemberSymbolInBaseType(i,t))return!1;case 157:case 239:return!0;case 196:case 197:var o=a.parameters,s=o.indexOf(n);return e.Debug.assert(-1!==s),r?o.slice(s+1).every(function(e){return 72===e.name.kind&&!e.symbol.isReferenced}):s===o.length-1;case 159:return!1;default:return e.Debug.failBadSyntaxKind(a)}}function b(n,t,r,a,i){e.FindAllReferences.Core.eachSignatureCall(r.parent,a,i,function(e){var a=r.parent.parameters.indexOf(r);e.arguments.length>a&&n.delete(t,e.arguments[a])})}n.registerCodeFix({errorCodes:o,getCodeActions:function(a){var o=a.errorCode,f=a.sourceFile,g=a.program,v=g.getTypeChecker(),y=g.getSourceFiles(),h=e.getTokenAtPosition(f,a.span.start);if(e.isJSDocTemplateTag(h))return[c(e.textChanges.ChangeTracker.with(a,function(e){return e.delete(f,h)}),e.Diagnostics.Remove_template_tag)];if(28===h.kind){var b=e.textChanges.ChangeTracker.with(a,function(e){return l(e,f,h)});return[c(b,e.Diagnostics.Remove_type_parameters)]}var E=u(h);if(E){b=e.textChanges.ChangeTracker.with(a,function(e){return e.delete(f,E)});return[c(b,[e.Diagnostics.Remove_import_from_0,e.showModuleSpecifier(E)])]}var T=e.textChanges.ChangeTracker.with(a,function(e){return d(h,e,f,v,y,!1)});if(T.length)return[c(T,e.Diagnostics.Remove_destructuring)];var S=e.textChanges.ChangeTracker.with(a,function(e){return p(f,h,e)});if(S.length)return[c(S,e.Diagnostics.Remove_variable_statement)];var x=[];if(127===h.kind){b=e.textChanges.ChangeTracker.with(a,function(e){return s(e,f,h)});var C=e.cast(h.parent,e.isInferTypeNode).typeParameter.name.text;x.push(n.createCodeFixAction(t,b,[e.Diagnostics.Replace_infer_0_with_unknown,C],i,e.Diagnostics.Replace_all_unused_infer_with_unknown))}else{var A=e.textChanges.ChangeTracker.with(a,function(e){return _(f,h,e,v,y,!1)});if(A.length){C=e.isComputedPropertyName(h.parent)?h.parent:h;x.push(c(A,[e.Diagnostics.Remove_declaration_for_Colon_0,C.getText(f)]))}}var D=e.textChanges.ChangeTracker.with(a,function(e){return m(e,o,f,h)});return D.length&&x.push(n.createCodeFixAction(t,D,[e.Diagnostics.Prefix_0_with_an_underscore,h.getText(f)],r,e.Diagnostics.Prefix_all_unused_declarations_with_where_possible)),x},fixIds:[r,a,i],getAllCodeActions:function(t){var c=t.sourceFile,f=t.program,g=f.getTypeChecker(),v=f.getSourceFiles();return n.codeFixAll(t,o,function(n,o){var f=e.getTokenAtPosition(c,o.start);switch(t.fixId){case r:m(n,o.code,c,f);break;case a:if(127===f.kind)break;var y=u(f);y?n.delete(c,y):e.isJSDocTemplateTag(f)?n.delete(c,f):28===f.kind?l(n,c,f):d(f,n,c,g,v,!0)||p(c,f,n)||_(c,f,n,g,v,!0);break;case i:127===f.kind&&s(n,c,f);break;default:e.Debug.fail(JSON.stringify(t.fixId))}})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="fixUnreachableCode",r=[e.Diagnostics.Unreachable_code_detected.code];function a(n,t,r,a){var o=e.getTokenAtPosition(t,r),s=e.findAncestor(o,e.isStatement);e.Debug.assert(s.getStart(t)===o.getStart(t));var c=(e.isBlock(s.parent)?s.parent:s).parent;if(!e.isBlock(s.parent)||s===e.first(s.parent.statements))switch(c.kind){case 222:if(c.elseStatement){if(e.isBlock(s.parent))break;return void n.replaceNode(t,s,e.createBlock(e.emptyArray))}case 224:case 225:return void n.delete(t,c)}if(e.isBlock(s.parent)){var l=r+a,u=e.Debug.assertDefined(i(e.sliceAfter(s.parent.statements,s),function(e){return e.pos<l}));n.deleteNodeRange(t,s,u)}else n.delete(t,s)}function i(e,n){for(var t,r=0,a=e;r<a.length;r++){var i=a[r];if(!n(i))break;t=i}return t}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,function(e){return a(e,r.sourceFile,r.span.start,r.span.length)});return[n.createCodeFixAction(t,i,e.Diagnostics.Remove_unreachable_code,t,e.Diagnostics.Remove_all_unreachable_code)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(e,n){return a(e,n.file,n.start,n.length)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="fixUnusedLabel",r=[e.Diagnostics.Unused_label.code];function a(n,t,r){var a=e.getTokenAtPosition(t,r),i=e.cast(a.parent,e.isLabeledStatement),o=a.getStart(t),s=i.statement.getStart(t),c=e.positionsAreOnSameLine(o,s,t)?s:e.skipTrivia(t.text,e.findChildOfKind(i,57,t).end,!0);n.deleteRange(t,{pos:o,end:c})}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,function(e){return a(e,r.sourceFile,r.span.start)});return[n.createCodeFixAction(t,i,e.Diagnostics.Remove_unused_label,t,e.Diagnostics.Remove_all_unused_labels)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(e,n){return a(e,n.file,n.start)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="fixJSDocTypes_plain",r="fixJSDocTypes_nullable",a=[e.Diagnostics.JSDoc_types_can_only_be_used_inside_documentation_comments.code];function i(e,n,t,r,a){e.replaceNode(n,t,a.typeToTypeNode(r,t))}function o(n,t,r){var a=e.findAncestor(e.getTokenAtPosition(n,t),s),i=a&&a.type;return i&&{typeNode:i,type:r.getTypeFromTypeNode(i)}}function s(e){switch(e.kind){case 212:case 160:case 161:case 239:case 158:case 162:case 181:case 156:case 155:case 151:case 154:case 153:case 159:case 242:case 194:case 237:return!0;default:return!1}}n.registerCodeFix({errorCodes:a,getCodeActions:function(a){var s=a.sourceFile,c=a.program.getTypeChecker(),l=o(s,a.span.start,c);if(l){var u=l.typeNode,d=l.type,p=u.getText(s),m=[f(d,t,e.Diagnostics.Change_all_jsdoc_style_types_to_TypeScript)];return 286===u.kind&&m.push(f(c.getNullableType(d,32768),r,e.Diagnostics.Change_all_jsdoc_style_types_to_TypeScript_and_add_undefined_to_nullable_types)),m}function f(t,r,o){var l=e.textChanges.ChangeTracker.with(a,function(e){return i(e,s,u,t,c)});return n.createCodeFixAction("jdocTypes",l,[e.Diagnostics.Change_0_to_1,p,c.typeToString(t)],r,o)}},fixIds:[t,r],getAllCodeActions:function(e){var t=e.fixId,s=e.program,c=e.sourceFile,l=s.getTypeChecker();return n.codeFixAll(e,a,function(e,n){var a=o(n.file,n.start,l);if(a){var s=a.typeNode,u=a.type,d=286===s.kind&&t===r?l.getNullableType(u,32768):u;i(e,c,s,d,l)}})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="fixAwaitInSyncFunction",r=[e.Diagnostics.await_expression_is_only_allowed_within_an_async_function.code,e.Diagnostics.A_for_await_of_statement_is_only_allowed_within_an_async_function_or_async_generator.code];function a(n){return n.type?n.type:e.isVariableDeclaration(n.parent)&&n.parent.type&&e.isFunctionTypeNode(n.parent.type)?n.parent.type.type:void 0}function i(n,t){var r=e.getTokenAtPosition(n,t),i=e.getContainingFunction(r);if(i){var o;switch(i.kind){case 156:o=i.name;break;case 239:case 196:o=e.findChildOfKind(i,90,n);break;case 197:o=e.findChildOfKind(i,20,n)||e.first(i.parameters);break;default:return}return o&&{insertBefore:o,returnType:a(i)}}}function o(n,t,r){var a=r.insertBefore,i=r.returnType;if(i){var o=e.getEntityNameFromTypeNode(i);o&&72===o.kind&&"Promise"===o.text||n.replaceNode(t,i,e.createTypeReferenceNode("Promise",e.createNodeArray([i])))}n.insertModifierBefore(t,121,a)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var a=r.sourceFile,s=r.span,c=i(a,s.start);if(c){var l=e.textChanges.ChangeTracker.with(r,function(e){return o(e,a,c)});return[n.createCodeFixAction(t,l,e.Diagnostics.Add_async_modifier_to_containing_function,t,e.Diagnostics.Add_all_missing_async_modifiers)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(n,t){var r=i(t.file,t.start);r&&o(n,e.sourceFile,r)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="disableJsDiagnostics",r="disableJsDiagnostics",a=e.mapDefined(Object.keys(e.Diagnostics),function(n){var t=e.Diagnostics[n];return t.category===e.DiagnosticCategory.Error?t.code:void 0});function i(n,t,r,a){var i=e.getLineAndCharacterOfPosition(t,r).line;a&&!e.addToSeen(a,i)||n.insertCommentBeforeLine(t,i,r," @ts-ignore")}n.registerCodeFix({errorCodes:a,getCodeActions:function(a){var o=a.sourceFile,s=a.program,c=a.span,l=a.host,u=a.formatContext;if(e.isInJSFile(o)&&e.isCheckJsEnabledForFile(o,s.getCompilerOptions())){var d=[n.createCodeFixActionNoFixId(t,[n.createFileTextChanges(o.fileName,[e.createTextChange(o.checkJsDirective?e.createTextSpanFromBounds(o.checkJsDirective.pos,o.checkJsDirective.end):e.createTextSpan(0,0),"// @ts-nocheck"+e.getNewLineOrDefaultFromHost(l,u.options))])],e.Diagnostics.Disable_checking_for_this_file)];return e.textChanges.isValidLocationToAddComment(o,c.start)&&d.unshift(n.createCodeFixAction(t,e.textChanges.ChangeTracker.with(a,function(e){return i(e,o,c.start)}),e.Diagnostics.Ignore_this_error_message,r,e.Diagnostics.Add_ts_ignore_to_all_error_messages)),d}},fixIds:[r],getAllCodeActions:function(t){var r=e.createMap();return n.codeFixAll(t,a,function(n,t){e.textChanges.isValidLocationToAddComment(t.file,t.start)&&i(n,t.file,t.start,r)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){function t(e,n,t,r,a){for(var o=e.symbol.members,s=0,c=n;s<c.length;s++){var l=c[s];o.has(l.escapedName)||i(l,e,t,r,a)}}function r(e){return{directoryExists:e.host.directoryExists?function(n){return e.host.directoryExists(n)}:void 0,fileExists:e.host.fileExists?function(n){return e.host.fileExists(n)}:void 0,getCurrentDirectory:e.host.getCurrentDirectory?function(){return e.host.getCurrentDirectory()}:void 0,readFile:e.host.readFile?function(n){return e.host.readFile(n)}:void 0,useCaseSensitiveFileNames:e.host.useCaseSensitiveFileNames?function(){return e.host.useCaseSensitiveFileNames()}:void 0,getSourceFiles:function(){return e.program.getSourceFiles()},getCommonSourceDirectory:function(){return e.program.getCommonSourceDirectory()}}}function a(n){return{trackSymbol:e.noop,moduleResolverHost:r(n)}}function i(n,t,r,i,s){var c=n.getDeclarations();if(c&&c.length){var u=r.program.getTypeChecker(),m=c[0],f=e.getSynthesizedDeepClone(e.getNameOfDeclaration(m),!1),_=p(e.getModifierFlags(m)),g=_?e.createNodeArray([_]):void 0,v=u.getWidenedType(u.getTypeOfSymbolAtLocation(n,t)),y=!!(16777216&n.flags);switch(m.kind){case 158:case 159:case 153:case 154:var h=u.typeToTypeNode(v,t,void 0,a(r));s(e.createProperty(void 0,g,f,y?e.createToken(56):void 0,h,void 0));break;case 155:case 156:var b=u.getSignaturesOfType(v,0);if(!e.some(b))break;if(1===c.length){e.Debug.assert(1===b.length);var E=b[0];x(E,g,f,d(i));break}for(var T=0,S=b;T<S.length;T++){E=S[T];x(E,e.getSynthesizedDeepClones(g,!1),e.getSynthesizedDeepClone(f,!1))}if(c.length>b.length){E=u.getSignatureFromDeclaration(c[c.length-1]);x(E,g,f,d(i))}else e.Debug.assert(c.length===b.length),s(l(b,f,y,g,i));break}}function x(e,n,a,i){var c=o(r,e,t,n,a,y,i);c&&s(c)}}function o(n,t,r,i,o,s,c){var l=n.program,u=l.getTypeChecker().signatureToSignatureDeclaration(t,156,r,257,a(n));if(u)return u.decorators=void 0,u.modifiers=i,u.name=o,u.questionToken=s?e.createToken(56):void 0,u.body=c,u}function s(n,t,r,i,o,s,l){var u=!e.isInterfaceDeclaration(l),p=t.typeArguments,m=t.arguments,f=t.parent,_=n.program.getTypeChecker(),g=a(n),v=e.map(m,function(e){return _.typeToTypeNode(_.getBaseTypeOfLiteralType(_.getTypeAtLocation(e)),l,void 0,g)}),y=e.map(m,function(n){return e.isIdentifier(n)?n.text:e.isPropertyAccessExpression(n)?n.name.text:void 0}),h=_.getContextualType(t),b=i?void 0:h&&_.typeToTypeNode(h,l,void 0,g)||e.createKeywordTypeNode(120);return e.createMethod(void 0,o?[e.createToken(116)]:void 0,e.isYieldExpression(f)?e.createToken(40):void 0,r,void 0,i?void 0:e.map(p,function(n,t){return e.createTypeParameterDeclaration(84+p.length-1<=90?String.fromCharCode(84+t):"T"+t)}),c(m.length,y,v,void 0,i),b,u?d(s):void 0)}function c(n,t,r,a,i){for(var o=[],s=0;s<n;s++){var c=e.createParameter(void 0,void 0,void 0,t&&t[s]||"arg"+s,void 0!==a&&s>=a?e.createToken(56):void 0,i?void 0:r&&r[s]||e.createKeywordTypeNode(120),void 0);o.push(c)}return o}function l(n,t,r,a,i){for(var o=n[0],s=n[0].minArgumentCount,l=!1,d=0,p=n;d<p.length;d++){var m=p[d];s=Math.min(m.minArgumentCount,s),m.hasRestParameter&&(l=!0),m.parameters.length>=o.parameters.length&&(!m.hasRestParameter||o.hasRestParameter)&&(o=m)}var f=o.parameters.length-(o.hasRestParameter?1:0),_=o.parameters.map(function(e){return e.name}),g=c(f,_,void 0,s,!1);if(l){var v=e.createArrayTypeNode(e.createKeywordTypeNode(120)),y=e.createParameter(void 0,void 0,e.createToken(25),_[f]||"rest",f>=s?e.createToken(56):void 0,v,void 0);g.push(y)}return u(a,t,r,void 0,g,void 0,i)}function u(n,t,r,a,i,o,s){return e.createMethod(void 0,n,void 0,t,r?e.createToken(56):void 0,a,i,o,d(s))}function d(n){return e.createBlock([e.createThrow(e.createNew(e.createIdentifier("Error"),void 0,[e.createLiteral("Method not implemented.","single"===n.quotePreference)]))],!0)}function p(n){return 4&n?e.createToken(115):16&n?e.createToken(114):void 0}function m(n,t,r,a){var i=e.getTsConfigObjectLiteralExpression(t);if(i){var o=_(i,"compilerOptions");if(void 0!==o){var s=o.initializer;if(e.isObjectLiteralExpression(s)){var c=_(s,r);void 0===c?n.insertNodeAtObjectStart(t,s,f(r,a)):n.replaceNode(t,c.initializer,a)}}else n.insertNodeAtObjectStart(t,i,f("compilerOptions",e.createObjectLiteral([f(r,a)])))}}function f(n,t){return e.createPropertyAssignment(e.createStringLiteral(n),t)}function _(n,t){return e.find(n.properties,function(n){return e.isPropertyAssignment(n)&&!!n.name&&e.isStringLiteral(n.name)&&n.name.text===t})}n.createMissingMemberNodes=t,n.getNoopSymbolTrackerWithResolver=a,n.createMethodFromCallExpression=s,n.setJsonCompilerOptionValue=m,n.createJsonPropertyAssignment=f,n.findJsonProperty=_})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="invalidImportSyntax";function r(n,t){var r=e.getSourceFileOfNode(t),i=e.getNamespaceDeclarationNode(t),o=n.program.getCompilerOptions(),s=[];return s.push(a(n,r,t,e.makeImport(i.name,void 0,t.moduleSpecifier,e.getQuotePreference(r,n.preferences)))),e.getEmitModuleKind(o)===e.ModuleKind.CommonJS&&s.push(a(n,r,t,e.createImportEqualsDeclaration(void 0,void 0,i.name,e.createExternalModuleReference(t.moduleSpecifier)))),s}function a(r,a,i,o){var s=e.textChanges.ChangeTracker.with(r,function(e){return e.replaceNode(a,i,o)});return n.createCodeFixActionNoFixId(t,s,[e.Diagnostics.Replace_import_with_0,s[0].textChanges[0].newText])}function i(n){var t=n.sourceFile,r=e.Diagnostics.Cannot_invoke_an_expression_whose_type_lacks_a_call_signature_Type_0_has_no_compatible_call_signatures.code===n.errorCode?191:192,a=e.findAncestor(e.getTokenAtPosition(t,n.span.start),function(e){return e.kind===r&&e.getStart()===n.span.start&&e.getEnd()===n.span.start+n.span.length});if(!a)return[];var i=a.expression;return s(n,i)}function o(n){var t=n.sourceFile,r=e.findAncestor(e.getTokenAtPosition(t,n.span.start),function(e){return e.getStart()===n.span.start&&e.getEnd()===n.span.start+n.span.length});return r?s(n,r):[]}function s(a,i){var o=a.program.getTypeChecker().getTypeAtLocation(i);if(!o.symbol||!o.symbol.originatingImport)return[];var s=[],c=o.symbol.originatingImport;if(e.isImportCall(c)||e.addRange(s,r(a,c)),e.isExpression(i)&&(!e.isNamedDeclaration(i.parent)||i.parent.name!==i)){var l=a.sourceFile,u=e.textChanges.ChangeTracker.with(a,function(n){return n.replaceNode(l,i,e.createPropertyAccess(i,"default"),{})});s.push(n.createCodeFixActionNoFixId(t,u,e.Diagnostics.Use_synthetic_default_member))}return s}n.registerCodeFix({errorCodes:[e.Diagnostics.Cannot_invoke_an_expression_whose_type_lacks_a_call_signature_Type_0_has_no_compatible_call_signatures.code,e.Diagnostics.Cannot_use_new_with_an_expression_whose_type_lacks_a_call_or_construct_signature.code],getCodeActions:i}),n.registerCodeFix({errorCodes:[e.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1.code,e.Diagnostics.Type_0_does_not_satisfy_the_constraint_1.code,e.Diagnostics.Type_0_is_not_assignable_to_type_1.code,e.Diagnostics.Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated.code,e.Diagnostics.Type_predicate_0_is_not_assignable_to_1.code,e.Diagnostics.Property_0_of_type_1_is_not_assignable_to_string_index_type_2.code,e.Diagnostics.Property_0_of_type_1_is_not_assignable_to_numeric_index_type_2.code,e.Diagnostics.Numeric_index_type_0_is_not_assignable_to_string_index_type_1.code,e.Diagnostics.Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2.code,e.Diagnostics.Property_0_in_type_1_is_not_assignable_to_type_2.code,e.Diagnostics.Property_0_of_JSX_spread_attribute_is_not_assignable_to_target_property.code,e.Diagnostics.The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1.code],getCodeActions:o})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="strictClassInitialization",r="addMissingPropertyDefiniteAssignmentAssertions",a="addMissingPropertyUndefinedType",i="addMissingPropertyInitializer",o=[e.Diagnostics.Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor.code];function s(n,t){var r=e.getTokenAtPosition(n,t);return e.isIdentifier(r)?e.cast(r.parent,e.isPropertyDeclaration):void 0}function c(a,i){var o=e.textChanges.ChangeTracker.with(a,function(e){return l(e,a.sourceFile,i)});return n.createCodeFixAction(t,o,[e.Diagnostics.Add_definite_assignment_assertion_to_property_0,i.getText()],r,e.Diagnostics.Add_definite_assignment_assertions_to_all_uninitialized_properties)}function l(n,t,r){var a=e.updateProperty(r,r.decorators,r.modifiers,r.name,e.createToken(52),r.type,r.initializer);n.replaceNode(t,r,a)}function u(r,i){var o=e.textChanges.ChangeTracker.with(r,function(e){return d(e,r.sourceFile,i)});return n.createCodeFixAction(t,o,[e.Diagnostics.Add_undefined_type_to_property_0,i.name.getText()],a,e.Diagnostics.Add_undefined_type_to_all_uninitialized_properties)}function d(n,t,r){var a=e.createKeywordTypeNode(141),i=r.type,o=e.isUnionTypeNode(i)?i.types.concat(a):[i,a];n.replaceNode(t,i,e.createUnionTypeNode(o))}function p(r,a){var o=r.program.getTypeChecker(),s=f(o,a);if(s){var c=e.textChanges.ChangeTracker.with(r,function(e){return m(e,r.sourceFile,a,s)});return n.createCodeFixAction(t,c,[e.Diagnostics.Add_initializer_to_property_0,a.name.getText()],i,e.Diagnostics.Add_initializers_to_all_uninitialized_properties)}}function m(n,t,r,a){var i=e.updateProperty(r,r.decorators,r.modifiers,r.name,r.questionToken,r.type,a);n.replaceNode(t,r,i)}function f(e,n){return _(e,e.getTypeFromTypeNode(n.type))}function _(n,t){if(512&t.flags)return t===n.getFalseType()||t===n.getFalseType(!0)?e.createFalse():e.createTrue();if(t.isLiteral())return e.createLiteral(t.value);if(t.isUnion())return e.firstDefined(t.types,function(e){return _(n,e)});if(t.isClass()){var r=e.getClassLikeDeclarationOfSymbol(t.symbol);if(!r||e.hasModifier(r,128))return;var a=e.getFirstConstructorWithBody(r);if(a&&a.parameters.length)return;return e.createNew(e.createIdentifier(t.symbol.name),void 0,void 0)}return n.isArrayLikeType(t)?e.createArrayLiteral():void 0}n.registerCodeFix({errorCodes:o,getCodeActions:function(n){var t=s(n.sourceFile,n.span.start);if(t){var r=[u(n,t),c(n,t)];return e.append(r,p(n,t)),r}},fixIds:[r,a,i],getAllCodeActions:function(t){return n.codeFixAll(t,o,function(n,o){var c=s(o.file,o.start);if(c)switch(t.fixId){case r:l(n,o.file,c);break;case a:d(n,o.file,c);break;case i:var u=t.program.getTypeChecker(),p=f(u,c);if(!p)return;m(n,o.file,c,p);break;default:e.Debug.fail(JSON.stringify(t.fixId))}})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){function n(e,n,t){return r(e,n,t,0)}function t(e,n,t){return r(e,n,t,3)}function r(n,t,r,i){return a(e.inspectValue(n,t),r,i)}function a(n,t,r){return void 0===r&&(r=0),e.textChanges.getNewFileText(s(n,r),3,t.newLineCharacter||"\n",e.formatting.getFormatContext(t))}var i;function o(e){return s(e,2)}function s(n,t){var r="default"===n.name,a=r?"_default":n.name;if(!E(a)||r&&1!==t)return e.emptyArray;var i=r&&2===n.kind?[e.createModifier(85),e.createModifier(80)]:3===t||0===t?[e.createModifier(125)]:1===t?[e.createModifier(85)]:void 0,u=function(){return 0===t?[c(n.name,!0)]:e.emptyArray},p=function(){return r?[c("_default",!1)]:e.emptyArray};switch(n.kind){case 2:return u().concat(l(i,a,n));case 3:var m=n.members,f=n.hasNontrivialPrototype;if(!f){if(0===t)return e.flatMap(m,function(e){return s(e,1)});if(m.some(function(e){return 2===e.kind}))return p().concat([x(i,a,e.flatMap(m,o))])}case 0:case 1:var _=0===n.kind?n.comment:void 0,g=e.createVariableStatement(i,e.createVariableDeclarationList([e.createVariableDeclaration(a,d(n))],2));return u().concat(p(),[T(g,_)]);default:return e.Debug.assertNever(n)}}function c(n,t){return e.createExportAssignment(void 0,void 0,t,e.createIdentifier(n))}function l(n,t,r){var a=r.source,i=r.prototypeMembers,o=r.namespaceMembers,c=v(a),l=void 0===c?{parameters:e.emptyArray,returnType:S()}:f(c),p=l.parameters,_=l.returnType,g=e.createMap();"object"===typeof c&&m(c,g);for(var y=0,h=i;y<h.length;y++){var b=h[y];if(2===b.kind){var C=u(b);C&&g.set(b.name,C)}}var A=0!==g.size||void 0===c||"number"!==typeof c&&157===c.kind?[]:void 0,D=e.flatMap(o,function(n){if(E(n.name)){if(A)switch(n.kind){case 3:if(n.members.some(function(e){return 2===e.kind}))break;case 1:case 0:return void A.push(T(e.createProperty(void 0,[e.createModifier(116)],n.name,void 0,d(n),void 0),0===n.kind?n.comment:void 0));case 2:if(!n.namespaceMembers.length){var t=u(n,[e.createModifier(116)]);if(t)return void A.push(t)}break;default:e.Debug.assertNever(n)}return s(n,2)}}),L=A?e.createClassDeclaration(void 0,n,t,void 0,void 0,A.concat(p.length?[e.createConstructor(void 0,void 0,p,void 0)]:e.emptyArray,e.arrayFrom(g.values()))):e.createFunctionDeclaration(void 0,n,void 0,t,void 0,p,_,void 0);return[L].concat(0===D.length?e.emptyArray:[x(n&&n.map(function(n){return e.getSynthesizedDeepClone(n)}),t,D)])}function u(n,t){var r=n.name,a=n.source;if(E(r)){var i=v(a);if(void 0!==i&&("number"===typeof i||157!==i.kind)){var o=f(i);return o&&e.createMethod(void 0,t,void 0,r,void 0,void 0,o.parameters,o.returnType,void 0)}}}function d(n){switch(n.kind){case 0:return e.createTypeReferenceNode(n.typeName,void 0);case 1:return e.createArrayTypeNode(d(n.inner));case 2:return e.createTypeReferenceNode("Function",void 0);case 3:return e.createTypeLiteralNode(n.members.map(function(n){return e.createPropertySignature(void 0,p(n.name),void 0,d(n),void 0)}));default:return e.Debug.assertNever(n)}}function p(n){return e.isIdentifierText(n,6)?e.createIdentifier(n):e.createStringLiteral(n)}function m(n,t){b(n,function(n){if(e.isAssignmentExpression(n,!0)&&e.isPropertyAccessExpression(n.left)&&100===n.left.expression.kind){var r=n.left.name.text;e.isJsPrivate(r)||e.getOrUpdate(t,r,function(){return e.createProperty(void 0,void 0,r,void 0,S(),void 0)})}})}function f(n){if("number"===typeof n)return{parameters:e.fill(n,function(e){return _("p"+e,S())}),returnType:S()};var t=!1,r=!1;b(n,function(n){t=t||e.isIdentifier(n)&&"arguments"===n.text,r=r||e.isReturnStatement(n)&&!!n.expression&&200!==n.expression.kind});var a=n.parameters.map(function(e){return _(""+e.name.getText(),h(n,e))}).concat(t?[g()]:e.emptyArray);return{parameters:a,returnType:r?S():e.createKeywordTypeNode(106)}}function _(n,t){return e.createParameter(void 0,void 0,void 0,n,void 0,t)}function g(){return e.createParameter(void 0,void 0,e.createToken(25),"args",void 0,e.createArrayTypeNode(S()))}function v(n){if("number"===typeof n)return n;var t=e.tryCast(y(n),function(n){return e.isFunctionExpression(n)||e.isArrowFunction(n)||e.isClassExpression(n)});return t?e.isClassExpression(t)?e.find(t.members,e.isConstructorDeclaration):t:e.cast(e.first(e.cast(y("{ "+n+" }"),e.isObjectLiteralExpression).properties),e.isMethodDeclaration)}function y(n){var t="const _ = "+n,r=e.createSourceFile("test.ts",t,6,!0);return e.first(e.cast(e.first(r.statements),e.isVariableStatement).declarationList.declarations).initializer}function h(e,n){return S()}function b(n,t){n.body.forEachChild(function n(r){t(r),e.isFunctionLike(r)||r.forEachChild(n)})}function E(n){var t=e.stringToToken(n);return!(t&&e.isNonContextualKeyword(t))&&e.isIdentifierText(n,6)}function T(n,t){return void 0!==t&&e.addSyntheticLeadingComment(n,2,t),n}function S(){return e.createKeywordTypeNode(120)}function x(n,t,r){return e.createModuleDeclaration(void 0,n,e.createIdentifier(t),e.createModuleBlock(r),16)}e.generateTypesForModule=n,e.generateTypesForGlobal=t,e.valueInfoToDeclarationFileText=a,function(e){e[e["ExportEquals"]=0]="ExportEquals",e[e["NamedExport"]=1]="NamedExport",e[e["NamespaceMember"]=2]="NamespaceMember",e[e["Global"]=3]="Global"}(i||(i={}))}(u||(u={})),function(e){(function(n){var t="requireInTs",r=[e.Diagnostics.require_call_may_be_converted_to_an_import.code];function a(n,t,r,a){var o=i(t,r),s=o.statement,c=o.name,l=o.required;n.replaceNode(t,s,e.getAllowSyntheticDefaultImports(a.getCompilerOptions())?e.createImportDeclaration(void 0,void 0,e.createImportClause(c,void 0),l):e.createImportEqualsDeclaration(void 0,void 0,c,e.createExternalModuleReference(l)))}function i(n,t){var r=e.getTokenAtPosition(n,t).parent;if(!e.isRequireCall(r,!0))throw e.Debug.failBadSyntaxKind(r);var a=e.cast(r.parent,e.isVariableDeclaration);return{statement:e.cast(a.parent.parent,e.isVariableStatement),name:e.cast(a.name,e.isIdentifier),required:r.arguments[0]}}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,function(e){return a(e,r.sourceFile,r.span.start,r.program)});return[n.createCodeFixAction(t,i,e.Diagnostics.Convert_require_to_import,t,e.Diagnostics.Convert_all_require_to_import)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(n,t){return a(n,t.file,t.start,e.program)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="useDefaultImport",r=[e.Diagnostics.Import_may_be_converted_to_a_default_import.code];function a(n,t){var r=e.getTokenAtPosition(n,t);if(e.isIdentifier(r)){var a=r.parent;if(e.isImportEqualsDeclaration(a)&&e.isExternalModuleReference(a.moduleReference))return{importNode:a,name:r,moduleSpecifier:a.moduleReference.expression};if(e.isNamespaceImport(a)){var i=a.parent.parent;return{importNode:i,name:r,moduleSpecifier:i.moduleSpecifier}}}}function i(n,t,r,a){n.replaceNode(t,r.importNode,e.makeImport(r.name,void 0,r.moduleSpecifier,e.getQuotePreference(t,a)))}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=r.span.start,c=a(o,s);if(c){var l=e.textChanges.ChangeTracker.with(r,function(e){return i(e,o,c,r.preferences)});return[n.createCodeFixAction(t,l,e.Diagnostics.Convert_to_default_import,t,e.Diagnostics.Convert_all_to_default_imports)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,function(n,t){var r=a(t.file,t.start);r&&i(n,t.file,r,e.preferences)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="fixAddModuleReferTypeMissingTypeof",r=t,a=[e.Diagnostics.Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0.code];function i(n,t){var r=e.getTokenAtPosition(n,t);return e.Debug.assert(92===r.kind),e.Debug.assert(183===r.parent.kind),r.parent}function o(n,t,r){var a=e.updateImportTypeNode(r,r.argument,r.qualifier,r.typeArguments,!0);n.replaceNode(t,r,a)}n.registerCodeFix({errorCodes:a,getCodeActions:function(t){var a=t.sourceFile,s=t.span,c=i(a,s.start),l=e.textChanges.ChangeTracker.with(t,function(e){return o(e,a,c)});return[n.createCodeFixAction(r,l,e.Diagnostics.Add_missing_typeof,r,e.Diagnostics.Add_missing_typeof)]},fixIds:[r],getAllCodeActions:function(e){return n.codeFixAll(e,a,function(n,t){return o(n,e.sourceFile,i(t.file,t.start))})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="fixConvertToMappedObjectType",r=t,a=[e.Diagnostics.An_index_signature_parameter_type_cannot_be_a_union_type_Consider_using_a_mapped_object_type_instead.code];function i(n,t){var r=e.getTokenAtPosition(n,t),a=e.cast(r.parent.parent,e.isIndexSignatureDeclaration);if(!e.isClassDeclaration(a.parent)){var i=e.isInterfaceDeclaration(a.parent)?a.parent:e.cast(a.parent.parent,e.isTypeAliasDeclaration);return{indexSignature:a,container:i}}}function o(n,t){return e.createTypeAliasDeclaration(n.decorators,n.modifiers,n.name,n.typeParameters,t)}function s(n,t,r){var a=r.indexSignature,i=r.container,s=e.isInterfaceDeclaration(i)?i.members:i.type.members,c=s.filter(function(n){return!e.isIndexSignatureDeclaration(n)}),l=e.first(a.parameters),u=e.createTypeParameterDeclaration(e.cast(l.name,e.isIdentifier),l.type),d=e.createMappedTypeNode(e.hasReadonlyModifier(a)?e.createModifier(133):void 0,u,a.questionToken,a.type),p=e.createIntersectionTypeNode(e.getAllSuperTypeNodes(i).concat([d],c.length?[e.createTypeLiteralNode(c)]:e.emptyArray));n.replaceNode(t,i,o(i,p))}n.registerCodeFix({errorCodes:a,getCodeActions:function(t){var a=t.sourceFile,o=t.span,c=i(a,o.start);if(c){var l=e.textChanges.ChangeTracker.with(t,function(e){return s(e,a,c)}),u=e.idText(c.container.name);return[n.createCodeFixAction(r,l,[e.Diagnostics.Convert_0_to_mapped_object_type,u],r,[e.Diagnostics.Convert_0_to_mapped_object_type,u])]}},fixIds:[r],getAllCodeActions:function(e){return n.codeFixAll(e,a,function(e,n){var t=i(n.file,n.start);t&&s(e,n.file,t)})}})})(e.codefix||(e.codefix={}))}(u||(u={})),function(e){(function(n){var t="Convert export",r="Convert default export to named export",a="Convert named export to default export";function i(n){var t=n.file,r=e.getRefactorContextSpan(n),a=e.getTokenAtPosition(t,r.start),i=e.getParentNodeInSpan(a,t,r);if(i&&(e.isSourceFile(i.parent)||e.isModuleBlock(i.parent)&&e.isAmbientModule(i.parent.parent))){var o=e.isSourceFile(i.parent)?i.parent.symbol:i.parent.parent.symbol,s=e.getModifierFlags(i),c=!!(512&s);if(1&s&&(c||!o.exports.has("default")))switch(i.kind){case 239:case 240:case 241:case 243:case 242:case 244:var l=i;return l.name&&e.isIdentifier(l.name)?{exportNode:l,exportName:l.name,wasDefault:c,exportingModuleSymbol:o}:void 0;case 219:var u=i;if(!(2&u.declarationList.flags)||1!==u.declarationList.declarations.length)return;var d=e.first(u.declarationList.declarations);if(!d.initializer)return;return e.Debug.assert(!c),e.isIdentifier(d.name)?{exportNode:u,exportName:d.name,wasDefault:c,exportingModuleSymbol:o}:void 0;default:return}}}function o(e,n,t,r,a){s(e,t,r,n.getTypeChecker()),c(n,t,r,a)}function s(n,t,r,a){var i=t.wasDefault,o=t.exportNode,s=t.exportName;if(i)r.delete(n,e.Debug.assertDefined(e.findModifier(o,80)));else{var c=e.Debug.assertDefined(e.findModifier(o,85));switch(o.kind){case 239:case 240:case 241:r.insertNodeAfter(n,c,e.createToken(80));break;case 219:if(!e.FindAllReferences.Core.isSymbolReferencedInFile(s,a,n)){r.replaceNode(n,o,e.createExportDefault(e.Debug.assertDefined(e.first(o.declarationList.declarations).initializer)));break}case 243:case 242:case 244:r.deleteModifier(n,c),r.insertNodeAfter(n,o,e.createExportDefault(e.createIdentifier(s.text)));break;default:e.Debug.assertNever(o)}}}function c(n,t,r,a){var i=t.wasDefault,o=t.exportName,s=t.exportingModuleSymbol,c=n.getTypeChecker(),d=e.Debug.assertDefined(c.getSymbolAtLocation(o));e.FindAllReferences.Core.eachExportReference(n.getSourceFiles(),c,a,d,s,o.text,i,function(e){var n=e.getSourceFile();i?l(n,e,r,o.text):u(n,e,r)})}function l(n,t,r,a){var i=t.parent;switch(i.kind){case 189:r.replaceNode(n,t,e.createIdentifier(a));break;case 253:case 257:var o=i;r.replaceNode(n,o,d(a,o.name.text));break;case 250:var s=i;e.Debug.assert(s.name===t);o=d(a,t.text);var c=s.namedBindings;if(c)if(251===c.kind){r.deleteRange(n,{pos:t.getStart(n),end:c.getStart(n)});var l=e.isStringLiteral(s.parent.moduleSpecifier)?e.quotePreferenceFromString(s.parent.moduleSpecifier,n):1,u=e.makeImport(void 0,[d(a,t.text)],s.parent.moduleSpecifier,l);r.insertNodeAfter(n,s.parent,u)}else r.delete(n,t),r.insertNodeAtEndOfList(n,c.elements,o);else r.replaceNode(n,t,e.createNamedImports([o]));break;default:e.Debug.failBadSyntaxKind(i)}}function u(n,t,r){var a=t.parent;switch(a.kind){case 189:r.replaceNode(n,t,e.createIdentifier("default"));break;case 253:var i=e.createIdentifier(a.name.text);1===a.parent.elements.length?r.replaceNode(n,a.parent,i):(r.delete(n,a),r.insertNodeBefore(n,a.parent,i));break;case 257:r.replaceNode(n,a,p("default",a.name.text));break;default:e.Debug.assertNever(a)}}function d(n,t){return e.createImportSpecifier(n===t?void 0:e.createIdentifier(n),e.createIdentifier(t))}function p(n,t){return e.createExportSpecifier(n===t?void 0:e.createIdentifier(n),e.createIdentifier(t))}n.registerRefactor(t,{getAvailableActions:function(n){var o=i(n);if(!o)return e.emptyArray;var s=o.wasDefault?e.Diagnostics.Convert_default_export_to_named_export.message:e.Diagnostics.Convert_named_export_to_default_export.message,c=o.wasDefault?r:a;return[{name:t,description:s,actions:[{name:c,description:s}]}]},getEditsForAction:function(n,t){e.Debug.assert(t===r||t===a);var s=e.textChanges.ChangeTracker.with(n,function(t){return o(n.file,n.program,e.Debug.assertDefined(i(n)),t,n.cancellationToken)});return{edits:s,renameFilename:void 0,renameLocation:void 0}}})})(e.refactor||(e.refactor={}))}(u||(u={})),function(e){(function(n){var t="Convert import",r="Convert namespace import to named imports",a="Convert named imports to namespace import";function i(n){var t=n.file,r=e.getRefactorContextSpan(n),a=e.getTokenAtPosition(t,r.start),i=e.getParentNodeInSpan(a,t,r);if(i&&e.isImportDeclaration(i)){var o=i.importClause;return o&&o.namedBindings}}function o(n,t,r,a){var i=t.getTypeChecker();251===a.kind?s(n,i,r,a,e.getAllowSyntheticDefaultImports(t.getCompilerOptions())):c(n,i,r,a)}function s(n,t,r,a,i){var o=!1,s=[],c=e.createMap();e.FindAllReferences.Core.eachSymbolReferenceInFile(a.name,t,n,function(n){if(e.isPropertyAccessExpression(n.parent)){var r=e.cast(n.parent,e.isPropertyAccessExpression),a=r.name.text;t.resolveName(a,n,67108863,!0)&&c.set(a,!0),e.Debug.assert(r.expression===n),s.push(r)}else o=!0});for(var u=e.createMap(),d=0,p=s;d<p.length;d++){var m=p[d],f=m.name.text,_=u.get(f);void 0===_&&u.set(f,_=c.has(f)?e.getUniqueName(f,n):f),r.replaceNode(n,m,e.createIdentifier(_))}var g=[];u.forEach(function(n,t){g.push(e.createImportSpecifier(n===t?void 0:e.createIdentifier(t),e.createIdentifier(n)))});var v=a.parent.parent;o&&!i?r.insertNodeAfter(n,v,l(v,void 0,g)):r.replaceNode(n,v,l(v,o?e.createIdentifier(a.name.text):void 0,g))}function c(n,t,r,a){for(var i=a.parent.parent,o=i.moduleSpecifier,s=o&&e.isStringLiteral(o)?e.codefix.moduleSpecifierToValidIdentifier(o.text,6):"module",c=a.elements.some(function(r){return e.FindAllReferences.Core.eachSymbolReferenceInFile(r.name,t,n,function(e){return!!t.resolveName(s,e,67108863,!0)})||!1}),u=c?e.getUniqueName(s,n):s,d=[],p=function(a){var i=(a.propertyName||a.name).text;e.FindAllReferences.Core.eachSymbolReferenceInFile(a.name,t,n,function(t){var o=e.createPropertyAccess(e.createIdentifier(u),i);e.isShorthandPropertyAssignment(t.parent)?r.replaceNode(n,t.parent,e.createPropertyAssignment(t.text,o)):e.isExportSpecifier(t.parent)&&!t.parent.propertyName?d.some(function(e){return e.name===a.name})||d.push(e.createImportSpecifier(a.propertyName&&e.createIdentifier(a.propertyName.text),e.createIdentifier(a.name.text))):r.replaceNode(n,t,o)})},m=0,f=a.elements;m<f.length;m++){var _=f[m];p(_)}r.replaceNode(n,a,e.createNamespaceImport(e.createIdentifier(u))),d.length&&r.insertNodeAfter(n,a.parent.parent,l(i,void 0,d))}function l(n,t,r){return e.createImportDeclaration(void 0,void 0,e.createImportClause(t,r&&r.length?e.createNamedImports(r):void 0),n.moduleSpecifier)}n.registerRefactor(t,{getAvailableActions:function(n){var o=i(n);if(!o)return e.emptyArray;var s=251===o.kind?e.Diagnostics.Convert_namespace_import_to_named_imports.message:e.Diagnostics.Convert_named_imports_to_namespace_import.message,c=251===o.kind?r:a;return[{name:t,description:s,actions:[{name:c,description:s}]}]},getEditsForAction:function(n,t){e.Debug.assert(t===r||t===a);var s=e.textChanges.ChangeTracker.with(n,function(t){return o(n.file,n.program,t,e.Debug.assertDefined(i(n)))});return{edits:s,renameFilename:void 0,renameLocation:void 0}}})})(e.refactor||(e.refactor={}))}(u||(u={})),function(e){(function(n){(function(t){var r,a,i,o,s="Extract Symbol";function c(n){var t=u(n.file,e.getRefactorContextSpan(n)),r=t.targetRange;if(void 0===r)return e.emptyArray;var a=g(r,n);if(void 0===a)return e.emptyArray;for(var i=[],o=e.createMap(),c=[],l=e.createMap(),d=0,p=0,m=a;p<m.length;p++){var f=m[p],_=f.functionExtraction,v=f.constantExtraction;if(0===_.errors.length){var y=_.description;o.has(y)||(o.set(y,!0),i.push({description:y,name:"function_scope_"+d}))}if(0===v.errors.length){y=v.description;l.has(y)||(l.set(y,!0),c.push({description:y,name:"constant_scope_"+d}))}d++}var h=[];return i.length&&h.push({name:s,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_function),actions:i}),c.length&&h.push({name:s,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_constant),actions:c}),h.length?h:e.emptyArray}function l(n,t){var r=u(n.file,e.getRefactorContextSpan(n)),a=r.targetRange,i=/^function_scope_(\d+)$/.exec(t);if(i){var o=+i[1];return e.Debug.assert(isFinite(o),"Expected to parse a finite number from the function scope index"),f(a,n,o)}var s=/^constant_scope_(\d+)$/.exec(t);if(s){o=+s[1];return e.Debug.assert(isFinite(o),"Expected to parse a finite number from the constant scope index"),_(a,n,o)}e.Debug.fail("Unrecognized action name")}function u(n,t){var i=t.length;if(0===i)return{errors:[e.createFileDiagnostic(n,t.start,i,r.cannotExtractEmpty)]};var o=e.getParentNodeInSpan(e.getTokenAtPosition(n,t.start),n,t),s=e.getParentNodeInSpan(e.findTokenOnLeftOfPosition(n,e.textSpanEnd(t)),n,t),c=[],l=a.None;if(!o||!s)return{errors:[e.createFileDiagnostic(n,t.start,i,r.cannotExtractRange)]};if(o.parent!==s.parent)return{errors:[e.createFileDiagnostic(n,t.start,i,r.cannotExtractRange)]};if(o!==s){if(!B(o.parent))return{errors:[e.createFileDiagnostic(n,t.start,i,r.cannotExtractRange)]};for(var u=[],p=o,m=0,f=p.parent.statements;m<f.length;m++){var _=f[m];if(_===o||u.length){var g=T(_);if(g)return{errors:g};u.push(_)}if(_===s)break}return u.length?{targetRange:{range:u,facts:l,declarations:c}}:{errors:[e.createFileDiagnostic(n,t.start,i,r.cannotExtractRange)]}}if(e.isReturnStatement(o)&&!o.expression)return{errors:[e.createFileDiagnostic(n,t.start,i,r.cannotExtractRange)]};var v=h(o),y=b(v)||T(v);return y?{errors:y}:{targetRange:{range:d(v),facts:l,declarations:c}};function h(n){if(e.isReturnStatement(n)){if(n.expression)return n.expression}else if(e.isVariableStatement(n)){for(var t=0,r=void 0,a=0,i=n.declarationList.declarations;a<i.length;a++){var o=i[a];o.initializer&&(t++,r=o.initializer)}if(1===t)return r}else if(e.isVariableDeclaration(n)&&n.initializer)return n.initializer;return n}function b(n){if(e.isIdentifier(e.isExpressionStatement(n)?n.expression:n))return[e.createDiagnosticForNode(n,r.cannotExtractIdentifier)]}function E(n,t){var r=n;while(r!==t){if(154===r.kind){e.hasModifier(r,32)&&(l|=a.InStaticRegion);break}if(151===r.kind){var i=e.getContainingFunction(r);157===i.kind&&(l|=a.InStaticRegion);break}156===r.kind&&e.hasModifier(r,32)&&(l|=a.InStaticRegion),r=r.parent}}function T(n){var i;if(function(e){e[e["None"]=0]="None",e[e["Break"]=1]="Break",e[e["Continue"]=2]="Continue",e[e["Return"]=4]="Return"}(i||(i={})),e.Debug.assert(n.pos<=n.end,"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809"),e.Debug.assert(!e.positionIsSynthesized(n.pos),"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809"),!e.isStatement(n)&&(!e.isExpressionNode(n)||!V(n)))return[e.createDiagnosticForNode(n,r.statementOrExpressionExpected)];if(4194304&n.flags)return[e.createDiagnosticForNode(n,r.cannotExtractAmbientBlock)];var o,s=e.getContainingClass(n);s&&E(n,s);var u,d=4;return p(n),o;function p(n){if(o)return!0;if(e.isDeclaration(n)){var i=237===n.kind?n.parent.parent:n;if(e.hasModifier(i,1))return(o||(o=[])).push(e.createDiagnosticForNode(n,r.cannotExtractExportedEntity)),!0;c.push(n.symbol)}switch(n.kind){case 249:return(o||(o=[])).push(e.createDiagnosticForNode(n,r.cannotExtractImport)),!0;case 98:if(191===n.parent.kind){var s=e.getContainingClass(n);if(s.pos<t.start||s.end>=t.start+t.length)return(o||(o=[])).push(e.createDiagnosticForNode(n,r.cannotExtractSuper)),!0}else l|=a.UsesThis;break}if(e.isFunctionLikeDeclaration(n)||e.isClassLike(n)){switch(n.kind){case 239:case 240:e.isSourceFile(n.parent)&&void 0===n.parent.externalModuleIndicator&&(o||(o=[])).push(e.createDiagnosticForNode(n,r.functionWillNotBeVisibleInTheNewScope));break}return!1}var m=d;switch(n.kind){case 222:d=0;break;case 235:d=0;break;case 218:n.parent&&235===n.parent.kind&&n.parent.finallyBlock===n&&(d=4);break;case 271:d|=1;break;default:e.isIterationStatement(n,!1)&&(d|=3);break}switch(n.kind){case 178:case 100:l|=a.UsesThis;break;case 233:var f=n.label;(u||(u=[])).push(f.escapedText),e.forEachChild(n,p),u.pop();break;case 229:case 228:f=n.label;f?e.contains(u,f.escapedText)||(o||(o=[])).push(e.createDiagnosticForNode(n,r.cannotExtractRangeContainingLabeledBreakOrContinueStatementWithTargetOutsideOfTheRange)):d&(229===n.kind?1:2)||(o||(o=[])).push(e.createDiagnosticForNode(n,r.cannotExtractRangeContainingConditionalBreakOrContinueStatements));break;case 201:l|=a.IsAsyncFunction;break;case 207:l|=a.IsGenerator;break;case 230:4&d?l|=a.HasReturn:(o||(o=[])).push(e.createDiagnosticForNode(n,r.cannotExtractRangeContainingConditionalReturnStatement));break;default:e.forEachChild(n,p);break}d=m}}}function d(n){return e.isStatement(n)?[n]:e.isExpressionNode(n)?e.isExpressionStatement(n.parent)?[n.parent]:n:void 0}function p(n){return e.isFunctionLikeDeclaration(n)||e.isSourceFile(n)||e.isModuleBlock(n)||e.isClassLike(n)}function m(n){var t=R(n.range)?e.first(n.range):n.range;if(n.facts&a.UsesThis){var r=e.getContainingClass(t);if(r){var i=e.findAncestor(t,e.isFunctionLikeDeclaration);return i?[i,r]:[r]}}var o=[];while(1)if(t=t.parent,151===t.kind&&(t=e.findAncestor(t,function(n){return e.isFunctionLikeDeclaration(n)}).parent),p(t)&&(o.push(t),279===t.kind))return o}function f(n,t,r){var a=v(n,t),i=a.scopes,o=a.readsAndWrites,s=o.target,c=o.usagesPerScope,l=o.functionErrorsPerScope,u=o.exposedVariableDeclarations;return e.Debug.assert(!l[r].length,"The extraction went missing? How?"),t.cancellationToken.throwIfCancellationRequested(),S(s,i[r],c[r],u,n,t)}function _(n,t,r){var a=v(n,t),i=a.scopes,o=a.readsAndWrites,s=o.target,c=o.usagesPerScope,l=o.constantErrorsPerScope,u=o.exposedVariableDeclarations;e.Debug.assert(!l[r].length,"The extraction went missing? How?"),e.Debug.assert(0===u.length,"Extract constant accepted a range containing a variable declaration?"),t.cancellationToken.throwIfCancellationRequested();var d=e.isExpression(s)?s:s.statements[0].expression;return x(d,i[r],c[r],n.facts,t)}function g(n,t){var r=v(n,t),a=r.scopes,i=r.readsAndWrites,o=i.functionErrorsPerScope,s=i.constantErrorsPerScope,c=a.map(function(n,t){var r,a,i=y(n),c=h(n),l=e.isFunctionLikeDeclaration(n)?b(n):e.isClassLike(n)?E(n):T(n);return 1===l?(r=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1_scope),[i,"global"]),a=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1_scope),[c,"global"])):0===l?(r=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1_scope),[i,"module"]),a=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1_scope),[c,"module"])):(r=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1),[i,l]),a=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1),[c,l])),0!==t||e.isClassLike(n)||(a=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_enclosing_scope),[c])),{functionExtraction:{description:r,errors:o[t]},constantExtraction:{description:a,errors:s[t]}}});return c}function v(e,n){var t=n.file,r=m(e),a=F(e,t),i=G(e,r,a,t,n.program.getTypeChecker(),n.cancellationToken);return{scopes:r,readsAndWrites:i}}function y(n){return e.isFunctionLikeDeclaration(n)?"inner function":e.isClassLike(n)?"method":"function"}function h(n){return e.isClassLike(n)?"readonly field":"constant"}function b(n){switch(n.kind){case 157:return"constructor";case 196:case 239:return n.name?"function '"+n.name.text+"'":"anonymous function";case 197:return"arrow function";case 156:return"method '"+n.name.getText()+"'";case 158:return"'get "+n.name.getText()+"'";case 159:return"'set "+n.name.getText()+"'";default:throw e.Debug.assertNever(n)}}function E(e){return 240===e.kind?e.name?"class '"+e.name.text+"'":"anonymous class declaration":e.name?"class expression '"+e.name.text+"'":"anonymous class expression"}function T(e){return 245===e.kind?"namespace '"+e.parent.name.getText()+"'":e.externalModuleIndicator?0:1}function S(n,t,r,i,o,s){var c,l,u=r.usages,d=r.typeParameterUsages,p=r.substitutions,m=s.program.getTypeChecker(),f=t.getSourceFile(),_=e.getUniqueName(e.isClassLike(t)?"newMethod":"newFunction",f),g=e.isInJSFile(t),v=e.createIdentifier(_),y=[],h=[];u.forEach(function(n,r){var a;if(!g){var i=m.getTypeOfSymbolAtLocation(n.symbol,n.node);i=m.getBaseTypeOfLiteralType(i),a=m.typeToTypeNode(i,t,1)}var o=e.createParameter(void 0,void 0,void 0,r,void 0,a);y.push(o),2===n.usage&&(l||(l=[])).push(n),h.push(e.createIdentifier(r))});var b=e.arrayFrom(d.values()).map(function(e){return{type:e,declaration:A(e)}}),E=b.sort(D),T=0===E.length?void 0:E.map(function(e){return e.declaration}),S=void 0!==T?T.map(function(n){return e.createTypeReferenceNode(n.name,void 0)}):void 0;if(e.isExpression(n)&&!g){var x=m.getContextualType(n);c=m.typeToTypeNode(x,t,1)}var C,N=k(n,i,l,p,!!(o.facts&a.HasReturn)),I=N.body,M=N.returnValueProperty;if(e.suppressLeadingAndTrailingTrivia(I),e.isClassLike(t)){var O=g?[]:[e.createToken(113)];o.facts&a.InStaticRegion&&O.push(e.createToken(116)),o.facts&a.IsAsyncFunction&&O.push(e.createToken(121)),C=e.createMethod(void 0,O.length?O:void 0,o.facts&a.IsGenerator?e.createToken(40):void 0,v,void 0,T,y,c,I)}else C=e.createFunctionDeclaration(void 0,o.facts&a.IsAsyncFunction?[e.createToken(121)]:void 0,o.facts&a.IsGenerator?e.createToken(40):void 0,v,T,y,c,I);var F=e.textChanges.ChangeTracker.fromContext(s),G=(R(o.range)?e.last(o.range):o.range).end,V=w(G,t);V?F.insertNodeBefore(s.file,V,C,!0):F.insertNodeAtEndOfScope(s.file,t,C);var B=[],K=L(t,o,_),j=e.createCall(K,S,h);if(o.facts&a.IsGenerator&&(j=e.createYield(e.createToken(40),j)),o.facts&a.IsAsyncFunction&&(j=e.createAwait(j)),i.length&&!l)if(e.Debug.assert(!M),e.Debug.assert(!(o.facts&a.HasReturn)),1===i.length){var U=i[0];B.push(e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(e.getSynthesizedDeepClone(U.name),e.getSynthesizedDeepClone(U.type),j)],U.parent.flags)))}else{for(var H=[],W=[],q=i[0].parent.flags,z=!1,J=0,X=i;J<X.length;J++){U=X[J];H.push(e.createBindingElement(void 0,void 0,e.getSynthesizedDeepClone(U.name)));var Y=m.typeToTypeNode(m.getBaseTypeOfLiteralType(m.getTypeAtLocation(U)),t,1);W.push(e.createPropertySignature(void 0,U.symbol.name,void 0,Y,void 0)),z=z||void 0!==U.type,q&=U.parent.flags}var Q=z?e.createTypeLiteralNode(W):void 0;Q&&e.setEmitFlags(Q,1),B.push(e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(e.createObjectBindingPattern(H),Q,j)],q)))}else if(i.length||l){if(i.length)for(var Z=0,$=i;Z<$.length;Z++){U=$[Z];var ee=U.parent.flags;2&ee&&(ee=-3&ee|1),B.push(e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(U.symbol.name,oe(U.type))],ee)))}M&&B.push(e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(M,oe(c))],1)));var ne=P(i,l);M&&ne.unshift(e.createShorthandPropertyAssignment(M)),1===ne.length?(e.Debug.assert(!M),B.push(e.createStatement(e.createAssignment(ne[0].name,j))),o.facts&a.HasReturn&&B.push(e.createReturn())):(B.push(e.createStatement(e.createAssignment(e.createObjectLiteral(ne),j))),M&&B.push(e.createReturn(e.createIdentifier(M))))}else o.facts&a.HasReturn?B.push(e.createReturn(j)):R(o.range)?B.push(e.createStatement(j)):B.push(j);R(o.range)?F.replaceNodeRangeWithNodes(s.file,e.first(o.range),e.last(o.range),B):F.replaceNodeWithNodes(s.file,o.range,B);var te=F.getChanges(),re=R(o.range)?e.first(o.range):o.range,ae=re.getSourceFile().fileName,ie=e.getRenameLocation(te,ae,_,!1);return{renameFilename:ae,renameLocation:ie,edits:te};function oe(n){if(void 0!==n){var t=e.getSynthesizedDeepClone(n),r=t;while(e.isParenthesizedTypeNode(r))r=r.type;return e.isUnionTypeNode(r)&&e.find(r.types,function(e){return 141===e.kind})?t:e.createUnionTypeNode([t,e.createKeywordTypeNode(141)])}}}function x(n,t,r,i,o){var s=r.substitutions,c=o.program.getTypeChecker(),l=t.getSourceFile(),u=e.getUniqueName(e.isClassLike(t)?"newProperty":"newLocal",l),d=e.isInJSFile(t),m=d||!c.isContextSensitive(n)?void 0:c.typeToTypeNode(c.getContextualType(n),t,1),f=N(n,s);e.suppressLeadingAndTrailingTrivia(f);var _=e.textChanges.ChangeTracker.fromContext(o);if(e.isClassLike(t)){e.Debug.assert(!d);var g=[];g.push(e.createToken(113)),i&a.InStaticRegion&&g.push(e.createToken(116)),g.push(e.createToken(133));var v=e.createProperty(void 0,g,u,void 0,m,f),y=e.createPropertyAccess(i&a.InStaticRegion?e.createIdentifier(t.name.getText()):e.createThis(),e.createIdentifier(u)),h=n.pos,b=M(h,t);_.insertNodeBefore(o.file,b,v,!0),_.replaceNode(o.file,n,y)}else{var E=e.createVariableDeclaration(u,m,f),T=C(n,t);if(T){_.insertNodeBefore(o.file,T,E);y=e.createIdentifier(u);_.replaceNode(o.file,n,y)}else if(221===n.parent.kind&&t===e.findAncestor(n,p)){var S=e.createVariableStatement(void 0,e.createVariableDeclarationList([E],2));_.replaceNode(o.file,n.parent,S)}else{S=e.createVariableStatement(void 0,e.createVariableDeclarationList([E],2)),b=O(n,t);if(0===b.pos?_.insertNodeAtTopOfFile(o.file,S,!1):_.insertNodeBefore(o.file,b,S,!1),221===n.parent.kind)_.delete(o.file,n.parent);else{y=e.createIdentifier(u);_.replaceNode(o.file,n,y)}}}var x=_.getChanges(),A=n.getSourceFile().fileName,D=e.getRenameLocation(x,A,u,!0);return{renameFilename:A,renameLocation:D,edits:x}}function C(n,t){var r;while(void 0!==n&&n!==t){if(e.isVariableDeclaration(n)&&n.initializer===r&&e.isVariableDeclarationList(n.parent)&&n.parent.declarations.length>1)return n;r=n,n=n.parent}}function A(e){var n,t=e.symbol;if(t&&t.declarations)for(var r=0,a=t.declarations;r<a.length;r++){var i=a[r];(void 0===n||i.pos<n.pos)&&(n=i)}return n}function D(n,t){var r=n.type,a=n.declaration,i=t.type,o=t.declaration;return e.compareProperties(a,o,"pos",e.compareValues)||e.compareStringsCaseSensitive(r.symbol?r.symbol.getName():"",i.symbol?i.symbol.getName():"")||e.compareValues(r.id,i.id)}function L(n,t,r){var i=e.createIdentifier(r);if(e.isClassLike(n)){var o=t.facts&a.InStaticRegion?e.createIdentifier(n.name.text):e.createThis();return e.createPropertyAccess(o,i)}return i}function k(n,t,r,a,i){var o,s=void 0!==r||t.length>0;if(e.isBlock(n)&&!s&&0===a.size)return{body:e.createBlock(n.statements,!0),returnValueProperty:void 0};var c=!1,l=e.createNodeArray(e.isBlock(n)?n.statements.slice(0):[e.isStatement(n)?n:e.createReturn(n)]);if(s||a.size){var u=e.visitNodes(l,p).slice();if(s&&!i&&e.isStatement(n)){var d=P(t,r);1===d.length?u.push(e.createReturn(d[0].name)):u.push(e.createReturn(e.createObjectLiteral(d)))}return{body:e.createBlock(u,!0),returnValueProperty:o}}return{body:e.createBlock(l,!0),returnValueProperty:void 0};function p(n){if(!c&&230===n.kind&&s){var i=P(t,r);return n.expression&&(o||(o="__return"),i.unshift(e.createPropertyAssignment(o,e.visitNode(n.expression,p)))),1===i.length?e.createReturn(i[0].name):e.createReturn(e.createObjectLiteral(i))}var l=c;c=c||e.isFunctionLikeDeclaration(n)||e.isClassLike(n);var u=a.get(e.getNodeId(n).toString()),d=u?e.getSynthesizedDeepClone(u):e.visitEachChild(n,p,e.nullTransformationContext);return c=l,d}}function N(n,t){return t.size?r(n):n;function r(n){var a=t.get(e.getNodeId(n).toString());return a?e.getSynthesizedDeepClone(a):e.visitEachChild(n,r,e.nullTransformationContext)}}function I(n){if(e.isFunctionLikeDeclaration(n)){var t=n.body;if(e.isBlock(t))return t.statements}else{if(e.isModuleBlock(n)||e.isSourceFile(n))return n.statements;if(e.isClassLike(n))return n.members;e.assertType(n)}return e.emptyArray}function w(n,t){return e.find(I(t),function(t){return t.pos>=n&&e.isFunctionLikeDeclaration(t)&&!e.isConstructorDeclaration(t)})}function M(n,t){var r,a=t.members;e.Debug.assert(a.length>0);for(var i=!0,o=0,s=a;o<s.length;o++){var c=s[o];if(c.pos>n)return r||a[0];if(i&&!e.isPropertyDeclaration(c)){if(void 0!==r)return c;i=!1}r=c}return void 0===r?e.Debug.fail():r}function O(n,t){var r;e.Debug.assert(!e.isClassLike(t));for(var a=n;a!==t;a=a.parent)p(a)&&(r=a);for(a=(r||n).parent;;a=a.parent){if(B(a)){for(var i=void 0,o=0,s=a.statements;o<s.length;o++){var c=s[o];if(c.pos>n.pos)break;i=c}return!i&&e.isCaseClause(a)?(e.Debug.assert(e.isSwitchStatement(a.parent.parent)),a.parent.parent):e.Debug.assertDefined(i)}e.Debug.assert(a!==t,"Didn't encounter a block-like before encountering scope")}}function P(n,t){var r=e.map(n,function(n){return e.createShorthandPropertyAssignment(n.symbol.name)}),a=e.map(t,function(n){return e.createShorthandPropertyAssignment(n.symbol.name)});return void 0===r?a:void 0===a?r:r.concat(a)}function R(n){return e.isArray(n)}function F(n,t){return R(n.range)?{pos:e.first(n.range).getStart(t),end:e.last(n.range).getEnd()}:n.range}function G(n,t,i,o,s,c){var l,u,d=e.createMap(),p=[],m=[],f=[],_=[],g=[],v=e.createMap(),y=[],h=R(n.range)?1===n.range.length&&e.isExpressionStatement(n.range[0])?n.range[0].expression:void 0:n.range;if(void 0===h){var b=n.range,E=e.first(b).getStart(),T=e.last(b).end;u=e.createFileDiagnostic(o,E,T-E,r.expressionExpected)}else 147456&s.getTypeAtLocation(h).flags&&(u=e.createDiagnosticForNode(h,r.uselessConstantType));for(var S=0,x=t;S<x.length;S++){var C=x[S];p.push({usages:e.createMap(),typeParameterUsages:e.createMap(),substitutions:e.createMap()}),m.push(e.createMap()),f.push(e.isFunctionLikeDeclaration(C)&&239!==C.kind?[e.createDiagnosticForNode(C,r.cannotExtractToOtherFunctionLike)]:[]);var A=[];u&&A.push(u),e.isClassLike(C)&&e.isInJSFile(C)&&A.push(e.createDiagnosticForNode(C,r.cannotExtractToJSClass)),e.isArrowFunction(C)&&!e.isBlock(C.body)&&A.push(e.createDiagnosticForNode(C,r.cannotExtractToExpressionArrowFunction)),_.push(A)}var D=e.createMap(),L=R(n.range)?e.createBlock(n.range):n.range,k=R(n.range)?e.first(n.range):n.range,N=U(k);if(W(L),N&&!R(n.range)){var I=s.getContextualType(n.range);H(I)}if(d.size>0){for(var w=e.createMap(),M=0,O=k;void 0!==O&&M<t.length;O=O.parent)if(O===t[M]&&(w.forEach(function(e,n){p[M].typeParameterUsages.set(n,e)}),M++),e.isDeclarationWithTypeParameters(O))for(var P=0,F=e.getEffectiveTypeParameterDeclarations(O);P<F.length;P++){var G=F[P],V=s.getTypeAtLocation(G);d.has(V.id.toString())&&w.set(V.id.toString(),V)}e.Debug.assert(M===t.length)}if(g.length){var B=e.isBlockScope(t[0],t[0].parent)?t[0]:e.getEnclosingBlockScopeContainer(t[0]);e.forEachChild(B,J)}for(var K=function(t){var a=p[t];if(t>0&&(a.usages.size>0||a.typeParameterUsages.size>0)){var i=R(n.range)?n.range[0]:n.range;_[t].push(e.createDiagnosticForNode(i,r.cannotAccessVariablesFromNestedScopes))}var o,s=!1;if(p[t].usages.forEach(function(n){2===n.usage&&(s=!0,106500&n.symbol.flags&&n.symbol.valueDeclaration&&e.hasModifier(n.symbol.valueDeclaration,64)&&(o=n.symbol.valueDeclaration))}),e.Debug.assert(R(n.range)||0===y.length),s&&!R(n.range)){var c=e.createDiagnosticForNode(n.range,r.cannotWriteInExpression);f[t].push(c),_[t].push(c)}else if(o&&t>0){c=e.createDiagnosticForNode(o,r.cannotExtractReadonlyPropertyInitializerOutsideConstructor);f[t].push(c),_[t].push(c)}else if(l){c=e.createDiagnosticForNode(l,r.cannotExtractExportedEntity);f[t].push(c),_[t].push(c)}},j=0;j<t.length;j++)K(j);return{target:L,usagesPerScope:p,functionErrorsPerScope:f,constantErrorsPerScope:_,exposedVariableDeclarations:y};function U(n){return!!e.findAncestor(n,function(n){return e.isDeclarationWithTypeParameters(n)&&0!==e.getEffectiveTypeParameterDeclarations(n).length})}function H(e){for(var n=s.getSymbolWalker(function(){return c.throwIfCancellationRequested(),!0}),t=n.walkType(e).visitedTypes,r=0,a=t;r<a.length;r++){var i=a[r];i.isTypeParameter()&&d.set(i.id.toString(),i)}}function W(n,t){if(void 0===t&&(t=1),N){var r=s.getTypeAtLocation(n);H(r)}if(e.isDeclaration(n)&&n.symbol&&g.push(n),e.isAssignmentExpression(n))W(n.left,2),W(n.right);else if(e.isUnaryExpressionWithWrite(n))W(n.operand,2);else if(e.isPropertyAccessExpression(n)||e.isElementAccessExpression(n))e.forEachChild(n,W);else if(e.isIdentifier(n)){if(!n.parent)return;if(e.isQualifiedName(n.parent)&&n!==n.parent.left)return;if(e.isPropertyAccessExpression(n.parent)&&n!==n.parent.expression)return;q(n,t,e.isPartOfTypeNode(n))}else e.forEachChild(n,W)}function q(n,r,a){var i=z(n,r,a);if(i)for(var o=0;o<t.length;o++){var s=m[o].get(i);s&&p[o].substitutions.set(e.getNodeId(n).toString(),s)}}function z(c,l,u){var d=X(c);if(d){var g=e.getSymbolId(d).toString(),v=D.get(g);if(v&&v>=l)return g;if(D.set(g,l),v){for(var y=0,h=p;y<h.length;y++){var b=h[y],E=b.usages.get(c.text);E&&b.usages.set(c.text,{usage:l,symbol:d,node:c})}return g}var T=d.getDeclarations(),S=T&&e.find(T,function(e){return e.getSourceFile()===o});if(S&&!e.rangeContainsStartEnd(i,S.getStart(),S.end)){if(n.facts&a.IsGenerator&&2===l){for(var x=e.createDiagnosticForNode(c,r.cannotExtractRangeThatContainsWritesToReferencesLocatedOutsideOfTheTargetRangeInGenerators),C=0,A=f;C<A.length;C++){var L=A[C];L.push(x)}for(var k=0,N=_;k<N.length;k++){L=N[k];L.push(x)}}for(var I=0;I<t.length;I++){var w=t[I],M=s.resolveName(d.name,w,d.flags,!1);if(M!==d&&!m[I].has(g)){var O=Y(d.exportSymbol||d,w,u);if(O)m[I].set(g,O);else if(u){if(!(262144&d.flags)){x=e.createDiagnosticForNode(c,r.typeWillNotBeVisibleInTheNewScope);f[I].push(x),_[I].push(x)}}else p[I].usages.set(c.text,{usage:l,symbol:d,node:c})}}return g}}}function J(t){if(!(t===n.range||R(n.range)&&n.range.indexOf(t)>=0)){var r=e.isIdentifier(t)?X(t):s.getSymbolAtLocation(t);if(r){var a=e.find(g,function(e){return e.symbol===r});if(a)if(e.isVariableDeclaration(a)){var i=a.symbol.id.toString();v.has(i)||(y.push(a),v.set(i,!0))}else l=l||a}e.forEachChild(t,J)}}function X(n){return n.parent&&e.isShorthandPropertyAssignment(n.parent)&&n.parent.name===n?s.getShorthandAssignmentValueSymbol(n.parent):s.getSymbolAtLocation(n)}function Y(n,t,r){if(n){var a=n.getDeclarations();if(a&&a.some(function(e){return e.parent===t}))return e.createIdentifier(n.name);var i=Y(n.parent,t,r);if(void 0!==i)return r?e.createQualifiedName(i,e.createIdentifier(n.name)):e.createPropertyAccess(i,n.name)}}}function V(e){var n=e.parent;switch(n.kind){case 278:return!1}switch(e.kind){case 10:return 249!==n.kind&&253!==n.kind;case 208:case 184:case 186:return!1;case 72:return 186!==n.kind&&253!==n.kind&&257!==n.kind}return!0}function B(e){switch(e.kind){case 218:case 279:case 245:case 271:return!0;default:return!1}}n.registerRefactor(s,{getAvailableActions:c,getEditsForAction:l}),t.getAvailableActions=c,t.getEditsForAction=l,function(n){function t(n){return{message:n,code:0,category:e.DiagnosticCategory.Message,key:n}}n.cannotExtractRange=t("Cannot extract range."),n.cannotExtractImport=t("Cannot extract import statement."),n.cannotExtractSuper=t("Cannot extract super call."),n.cannotExtractEmpty=t("Cannot extract empty range."),n.expressionExpected=t("expression expected."),n.uselessConstantType=t("No reason to extract constant of type."),n.statementOrExpressionExpected=t("Statement or expression expected."),n.cannotExtractRangeContainingConditionalBreakOrContinueStatements=t("Cannot extract range containing conditional break or continue statements."),n.cannotExtractRangeContainingConditionalReturnStatement=t("Cannot extract range containing conditional return statement."),n.cannotExtractRangeContainingLabeledBreakOrContinueStatementWithTargetOutsideOfTheRange=t("Cannot extract range containing labeled break or continue with target outside of the range."),n.cannotExtractRangeThatContainsWritesToReferencesLocatedOutsideOfTheTargetRangeInGenerators=t("Cannot extract range containing writes to references located outside of the target range in generators."),n.typeWillNotBeVisibleInTheNewScope=t("Type will not visible in the new scope."),n.functionWillNotBeVisibleInTheNewScope=t("Function will not visible in the new scope."),n.cannotExtractIdentifier=t("Select more than a single identifier."),n.cannotExtractExportedEntity=t("Cannot extract exported declaration"),n.cannotWriteInExpression=t("Cannot write back side-effects when extracting an expression"),n.cannotExtractReadonlyPropertyInitializerOutsideConstructor=t("Cannot move initialization of read-only class property outside of the constructor"),n.cannotExtractAmbientBlock=t("Cannot extract code from ambient contexts"),n.cannotAccessVariablesFromNestedScopes=t("Cannot access variables from nested scopes"),n.cannotExtractToOtherFunctionLike=t("Cannot extract method to a function-like scope that is not a function"),n.cannotExtractToJSClass=t("Cannot extract constant to a class scope in JS"),n.cannotExtractToExpressionArrowFunction=t("Cannot extract constant to an arrow function without a block")}(r=t.Messages||(t.Messages={})),function(e){e[e["None"]=0]="None",e[e["HasReturn"]=1]="HasReturn",e[e["IsGenerator"]=2]="IsGenerator",e[e["IsAsyncFunction"]=4]="IsAsyncFunction",e[e["UsesThis"]=8]="UsesThis",e[e["InStaticRegion"]=16]="InStaticRegion"}(a||(a={})),t.getRangeToExtract=u,function(e){e[e["Module"]=0]="Module",e[e["Global"]=1]="Global"}(i||(i={})),function(e){e[e["Read"]=1]="Read",e[e["Write"]=2]="Write"}(o||(o={}))})(n.extractSymbol||(n.extractSymbol={}))})(e.refactor||(e.refactor={}))}(u||(u={})),function(e){(function(n){(function(t){var r="Generate 'get' and 'set' accessors",a=e.Diagnostics.Generate_get_and_set_accessors.message;function i(n){return m(n)?[{name:r,description:a,actions:[{name:r,description:a}]}]:e.emptyArray}function o(n,t){var r=n.file,a=m(n);if(a){var i=e.isSourceFileJS(r),o=e.textChanges.ChangeTracker.fromContext(n),s=a.isStatic,c=a.isReadonly,l=a.fieldName,u=a.accessorName,p=a.originalName,g=a.type,v=a.container,E=a.declaration,T=a.renameAccessor;e.suppressLeadingAndTrailingTrivia(l),e.suppressLeadingAndTrailingTrivia(E),e.suppressLeadingAndTrailingTrivia(v);var S=e.isClassLike(v),x=-65&e.getModifierFlags(E),C=S?!x||8&x?d(i,s,115):e.createNodeArray(e.createModifiersFromModifierFlags(x)):void 0,A=S?d(i,s,113):void 0;y(o,r,E,l,A);var D=f(l,u,g,C,s,v);if(e.suppressLeadingAndTrailingTrivia(D),h(o,r,D,E,v),c){var L=e.getFirstConstructorWithBody(v);L&&b(o,r,L,l.text,p)}else{var k=_(l,u,g,C,s,v);e.suppressLeadingAndTrailingTrivia(k),h(o,r,k,E,v)}var N=o.getChanges(),I=r.fileName,w=T?u:l,M=e.isIdentifier(w)?0:-1,O=M+e.getRenameLocation(N,I,w.text,e.isParameter(E));return{renameFilename:I,renameLocation:O,edits:N}}}function s(n){return e.isIdentifier(n)||e.isStringLiteral(n)}function c(n){return e.isParameterPropertyDeclaration(n)||e.isPropertyDeclaration(n)||e.isPropertyAssignment(n)}function l(n,t){return e.isIdentifier(t)?e.createIdentifier(n):e.createLiteral(n)}function u(n,t,r){var a=t?r.name:e.createThis();return e.isIdentifier(n)?e.createPropertyAccess(a,n):e.createElementAccess(a,e.createLiteral(n))}function d(n,t,r){var a=e.append(n?void 0:[e.createToken(r)],t?e.createToken(116):void 0);return a&&e.createNodeArray(a)}function p(e){return 95===e.charCodeAt(0)}function m(n){var t=n.file,r=n.startPosition,a=n.endPosition,i=e.getTokenAtPosition(t,r),o=e.findAncestor(i.parent,c),u=124;if(o&&e.nodeOverlapsWithStartEnd(o.name,t,r,a)&&s(o.name)&&(e.getModifierFlags(o)|u)===u){var d=o.name.text,m=p(d),f=l(m?d:e.getUniqueName("_"+d,t),o.name),_=l(m?e.getUniqueName(d.substring(1),t):d,o.name);return{isStatic:e.hasStaticModifier(o),isReadonly:e.hasReadonlyModifier(o),type:e.getTypeAnnotationNode(o),container:151===o.kind?o.parent.parent:o.parent,originalName:o.name.text,declaration:o,fieldName:f,accessorName:_,renameAccessor:m}}}function f(n,t,r,a,i,o){return e.createGetAccessor(void 0,a,t,void 0,r,e.createBlock([e.createReturn(u(n,i,o))],!0))}function _(n,t,r,a,i,o){return e.createSetAccessor(void 0,a,t,[e.createParameter(void 0,void 0,void 0,e.createIdentifier("value"),void 0,r)],e.createBlock([e.createStatement(e.createAssignment(u(n,i,o),e.createIdentifier("value")))],!0))}function g(n,t,r,a,i){var o=e.updateProperty(r,r.decorators,i,a,r.questionToken||r.exclamationToken,r.type,r.initializer);n.replaceNode(t,r,o)}function v(n,t,r,a){var i=e.updatePropertyAssignment(r,a,r.initializer);n.replacePropertyAssignment(t,r,i)}function y(n,t,r,a,i){e.isPropertyDeclaration(r)?g(n,t,r,a,i):e.isPropertyAssignment(r)?v(n,t,r,a):n.replaceNode(t,r,e.updateParameter(r,r.decorators,i,r.dotDotDotToken,e.cast(a,e.isIdentifier),r.questionToken,r.type,r.initializer))}function h(n,t,r,a,i){e.isParameterPropertyDeclaration(a)?n.insertNodeAtClassStart(t,i,r):e.isPropertyAssignment(a)?n.insertNodeAfterComma(t,a,r):n.insertNodeAfter(t,a,r)}function b(n,t,r,a,i){r.body&&r.body.forEachChild(function r(o){e.isElementAccessExpression(o)&&100===o.expression.kind&&e.isStringLiteral(o.argumentExpression)&&o.argumentExpression.text===i&&e.isWriteAccess(o)&&n.replaceNode(t,o.argumentExpression,e.createStringLiteral(a)),e.isPropertyAccessExpression(o)&&100===o.expression.kind&&o.name.text===i&&e.isWriteAccess(o)&&n.replaceNode(t,o.name,e.createIdentifier(a)),e.isFunctionLike(o)||e.isClassLike(o)||o.forEachChild(r)})}n.registerRefactor(r,{getEditsForAction:o,getAvailableActions:i})})(n.generateGetAccessorAndSetAccessor||(n.generateGetAccessorAndSetAccessor={}))})(e.refactor||(e.refactor={}))}(u||(u={})),function(e){(function(n){var t="Move to a new file";function r(n){var t=n.file,r=e.createTextRangeFromSpan(e.getRefactorContextSpan(n)),a=t.statements,i=e.findIndex(a,function(e){return e.end>r.pos});if(-1!==i){var o=a[i];if(e.isNamedDeclaration(o)&&o.name&&e.rangeContainsRange(o.name,r))return{toMove:[a[i]],afterLast:a[i+1]};if(!(r.pos>o.getStart(t))){var s=e.findIndex(a,function(e){return e.end>r.end},i);if(-1===s||!(0===s||a[s].getStart(t)<r.end))return{toMove:a.slice(i,-1===s?a.length:s),afterLast:-1===s?void 0:a[s]}}}}function a(n,t,r,a,i,o){var l=t.getTypeChecker(),u=L(n,r.all,l),d=e.getDirectoryPath(n.fileName),p=e.extensionFromPath(n.fileName),m=A(D(u.movedSymbols),p,d,i),f=m+p;a.createNewFile(n,e.combinePaths(d,f),c(n,u,a,r,t,m,o)),s(t,a,n.fileName,f,e.hostGetCanonicalFileName(i))}function i(n){var t=r(n);if(void 0!==t){var a=[],i=[],s=t.toMove,c=t.afterLast;return e.getRangesWhere(s,function(e){return!o(e)},function(e,n){for(var t=e;t<n;t++)a.push(s[t]);i.push({first:s[e],afterLast:c})}),0===a.length?void 0:{all:a,ranges:i}}}function o(n){switch(n.kind){case 249:return!0;case 248:return!e.hasModifier(n,1);case 219:return n.declarationList.declarations.every(function(n){return!!n.initializer&&e.isRequireCall(n.initializer,!0)});default:return!1}}function s(n,t,r,a,i){var o=n.getCompilerOptions().configFile;if(o){var s=e.normalizePath(e.combinePaths(r,"..",a)),c=e.getRelativePathFromFile(o.fileName,s,i),l=o.statements[0]&&e.tryCast(o.statements[0].expression,e.isObjectLiteralExpression),u=l&&e.find(l.properties,function(n){return e.isPropertyAssignment(n)&&e.isStringLiteral(n.name)&&"files"===n.name.text});u&&e.isArrayLiteralExpression(u.initializer)&&t.insertNodeInListAfter(o,e.last(u.initializer.elements),e.createLiteral(c),u.initializer.elements)}}function c(n,t,r,a,i,o,s){var c=i.getTypeChecker();if(!n.externalModuleIndicator&&!n.commonJsModuleIndicator)return l(n,a.ranges,r),a.all;var p=!!n.externalModuleIndicator,m=e.getQuotePreference(n,s),f=v(t.oldFileImportsFromNewFile,o,p,m);return f&&e.insertImport(r,n,f),u(n,a.all,r,t.unusedImportsFromOldFile,c),l(n,a.ranges,r),d(r,i,n,t.movedSymbols,o),C(n,t.oldImportsNeededByNewFile,t.newFileImportsFromOldFile,r,c,p,m).concat(E(n,a.all,t.oldFileImportsFromNewFile,p))}function l(e,n,t){for(var r=0,a=n;r<a.length;r++){var i=a[r],o=i.first,s=i.afterLast;t.deleteNodeRangeExcludingEnd(e,o,s)}}function u(n,t,r,a,i){for(var o=0,s=n.statements;o<s.length;o++){var c=s[o];e.contains(t,c)||g(c,function(e){return T(n,e,r,function(e){return a.has(i.getSymbolAtLocation(e))})})}}function d(n,t,r,a,i){for(var o=t.getTypeChecker(),s=function(t){if(t===r)return"continue";for(var s=function(s){g(s,function(c){if(o.getSymbolAtLocation(_(c))===r.symbol){var l=function(n){var t=e.isBindingElement(n.parent)?e.getPropertySymbolFromBindingElement(o,n.parent):e.skipAlias(o.getSymbolAtLocation(n),o);return!!t&&a.has(t)};T(t,c,n,l);var u=e.combinePaths(e.getDirectoryPath(_(c).text),i),d=I(c,e.createLiteral(u),l);d&&n.insertNodeAfter(t,s,d);var f=p(c);f&&m(n,t,o,a,i,u,f,c)}})},c=0,l=t.statements;c<l.length;c++){var u=l[c];s(u)}},c=0,l=t.getSourceFiles();c<l.length;c++){var u=l[c];s(u)}}function p(n){switch(n.kind){case 249:return n.importClause&&n.importClause.namedBindings&&251===n.importClause.namedBindings.kind?n.importClause.namedBindings.name:void 0;case 248:return n.name;case 237:return e.tryCast(n.name,e.isIdentifier);default:return e.Debug.assertNever(n)}}function m(n,t,r,a,i,o,s,c){var l=e.codefix.moduleSpecifierToValidIdentifier(i,6),u=!1,d=[];if(e.FindAllReferences.Core.eachSymbolReferenceInFile(s,r,t,function(n){e.isPropertyAccessExpression(n.parent)&&(u=u||!!r.resolveName(l,n,67108863,!0),a.has(r.getSymbolAtLocation(n.parent.name))&&d.push(n))}),d.length){for(var p=u?e.getUniqueName(l,t):l,m=0,_=d;m<_.length;m++){var g=_[m];n.replaceNode(t,g,e.createIdentifier(p))}n.insertNodeAfter(t,c,f(c,i,o))}}function f(n,t,r){var a=e.createIdentifier(t),i=e.createLiteral(r);switch(n.kind){case 249:return e.createImportDeclaration(void 0,void 0,e.createImportClause(void 0,e.createNamespaceImport(a)),i);case 248:return e.createImportEqualsDeclaration(void 0,void 0,a,e.createExternalModuleReference(i));case 237:return e.createVariableDeclaration(a,void 0,b(i));default:return e.Debug.assertNever(n)}}function _(e){return 249===e.kind?e.moduleSpecifier:248===e.kind?e.moduleReference.expression:e.initializer.arguments[0]}function g(n,t){if(e.isImportDeclaration(n))e.isStringLiteral(n.moduleSpecifier)&&t(n);else if(e.isImportEqualsDeclaration(n))e.isExternalModuleReference(n.moduleReference)&&e.isStringLiteralLike(n.moduleReference.expression)&&t(n);else if(e.isVariableStatement(n))for(var r=0,a=n.declarationList.declarations;r<a.length;r++){var i=a[r];i.initializer&&e.isRequireCall(i.initializer,!0)&&t(i)}}function v(n,t,r,a){var i,o=[];return n.forEach(function(n){"default"===n.escapedName?i=e.createIdentifier(e.symbolNameNoDefault(n)):o.push(n.name)}),y(i,o,t,r,a)}function y(n,t,r,a,i){if(r=e.ensurePathIsNonModuleName(r),a){var o=t.map(function(n){return e.createImportSpecifier(void 0,e.createIdentifier(n))});return e.makeImportIfNecessary(n,o,r,i)}e.Debug.assert(!n);var s=t.map(function(n){return e.createBindingElement(void 0,void 0,n)});return s.length?h(e.createObjectBindingPattern(s),void 0,b(e.createLiteral(r))):void 0}function h(n,t,r,a){return void 0===a&&(a=2),e.createVariableStatement(void 0,e.createVariableDeclarationList([e.createVariableDeclaration(n,t,r)],a))}function b(n){return e.createCall(e.createIdentifier("require"),void 0,[n])}function E(n,t,r,a){return e.flatMap(t,function(t){if(G(t)&&!W(n,t,a)&&B(t,function(n){return r.has(e.Debug.assertDefined(n.symbol))})){var i=q(t,a);if(i)return i}return t})}function T(n,t,r,a){switch(t.kind){case 249:S(n,t,r,a);break;case 248:a(t.name)&&r.delete(n,t);break;case 237:x(n,t,r,a);break;default:e.Debug.assertNever(t)}}function S(e,n,t,r){if(n.importClause){var a=n.importClause,i=a.name,o=a.namedBindings,s=!i||r(i),c=!o||(251===o.kind?r(o.name):0!==o.elements.length&&o.elements.every(function(e){return r(e.name)}));if(s&&c)t.delete(e,n);else if(i&&s&&t.delete(e,i),o)if(c)t.delete(e,o);else if(252===o.kind)for(var l=0,u=o.elements;l<u.length;l++){var d=u[l];r(d.name)&&t.delete(e,d)}}}function x(n,t,r,a){var i=t.name;switch(i.kind){case 72:a(i)&&r.delete(n,i);break;case 185:break;case 184:if(i.elements.every(function(n){return e.isIdentifier(n.name)&&a(n.name)}))r.delete(n,e.isVariableDeclarationList(t.parent)&&1===t.parent.declarations.length?t.parent.parent:t);else for(var o=0,s=i.elements;o<s.length;o++){var c=s[o];e.isIdentifier(c.name)&&a(c.name)&&r.delete(n,c.name)}break}}function C(n,t,r,a,i,o,s){for(var c,l=[],u=0,d=n.statements;u<d.length;u++){var p=d[u];g(p,function(n){e.append(l,I(n,_(n),function(e){return t.has(i.getSymbolAtLocation(e))}))})}var m=[],f=e.nodeSeenTracker();return r.forEach(function(t){for(var r=0,i=t.declarations;r<i.length;r++){var s=i[r];if(R(s)){var l=j(s);if(l){var u=U(s);f(u)&&H(n,u,a,o),e.hasModifier(s,512)?c=l:m.push(l.text)}}}}),e.append(l,y(c,m,e.removeFileExtension(e.getBaseFileName(n.fileName)),o,s)),l}function A(n,t,r,a){for(var i=n,o=1;;o++){var s=e.combinePaths(r,i+t);if(!a.fileExists(s))return i;i=n+"."+o}}function D(n){return n.forEachEntry(e.symbolNameNoDefault)||"newFile"}function L(n,t,r){var a=new P,i=new P,o=new P,s=e.find(t,function(e){return!!(4&e.transformFlags)}),c=y(s);c&&i.add(c);for(var l=0,u=t;l<u.length;l++){var d=u[l];B(d,function(n){a.add(e.Debug.assertDefined(e.isExpressionStatement(n)?r.getSymbolAtLocation(n.expression.left):n.symbol))})}for(var p=0,m=t;p<m.length;p++){d=m[p];O(d,r,function(e){if(e.declarations)for(var t=0,r=e.declarations;t<r.length;t++){var s=r[t];k(s)?i.add(e):R(s)&&F(s)===n&&!a.has(e)&&o.add(e)}})}for(var f=i.clone(),_=new P,g=0,v=n.statements;g<v.length;g++){d=v[g];e.contains(t,d)||(c&&4&d.transformFlags&&f.delete(c),O(d,r,function(e){a.has(e)&&_.add(e),f.delete(e)}))}return{movedSymbols:a,newFileImportsFromOldFile:o,oldFileImportsFromNewFile:_,oldImportsNeededByNewFile:i,unusedImportsFromOldFile:f};function y(n){if(void 0!==n){var t=r.getJsxNamespace(n),a=r.resolveName(t,n,1920,!0);return a&&e.some(a.declarations,k)?a:void 0}}}function k(n){switch(n.kind){case 248:case 253:case 250:return!0;case 237:return N(n);case 186:return e.isVariableDeclaration(n.parent.parent)&&N(n.parent.parent);default:return!1}}function N(n){return e.isSourceFile(n.parent.parent.parent)&&!!n.initializer&&e.isRequireCall(n.initializer,!0)}function I(n,t,r){switch(n.kind){case 249:var a=n.importClause;if(!a)return;var i=a.name&&r(a.name)?a.name:void 0,o=a.namedBindings&&w(a.namedBindings,r);return i||o?e.createImportDeclaration(void 0,void 0,e.createImportClause(i,o),t):void 0;case 248:return r(n.name)?n:void 0;case 237:var s=M(n.name,r);return s?h(s,n.type,b(t),n.parent.flags):void 0;default:return e.Debug.assertNever(n)}}function w(n,t){if(251===n.kind)return t(n.name)?n:void 0;var r=n.elements.filter(function(e){return t(e.name)});return r.length?e.createNamedImports(r):void 0}function M(n,t){switch(n.kind){case 72:return t(n)?n:void 0;case 185:return n;case 184:var r=n.elements.filter(function(n){return n.propertyName||!e.isIdentifier(n.name)||t(n.name)});return r.length?e.createObjectBindingPattern(r):void 0}}function O(n,t,r){n.forEachChild(function n(a){if(e.isIdentifier(a)&&!e.isDeclarationName(a)){var i=t.getSymbolAtLocation(a);i&&r(i)}else a.forEachChild(n)})}n.registerRefactor(t,{getAvailableActions:function(n){if(!n.preferences.allowTextChangesInNewFiles||void 0===i(n))return e.emptyArray;var r=e.getLocaleSpecificMessage(e.Diagnostics.Move_to_a_new_file);return[{name:t,description:r,actions:[{name:t,description:r}]}]},getEditsForAction:function(n,r){e.Debug.assert(r===t);var o=e.Debug.assertDefined(i(n)),s=e.textChanges.ChangeTracker.with(n,function(e){return a(n.file,n.program,o,e,n.host,n.preferences)});return{edits:s,renameFilename:void 0,renameLocation:void 0}}});var P=function(){function n(){this.map=e.createMap()}return n.prototype.add=function(n){this.map.set(String(e.getSymbolId(n)),n)},n.prototype.has=function(n){return this.map.has(String(e.getSymbolId(n)))},n.prototype.delete=function(n){this.map.delete(String(e.getSymbolId(n)))},n.prototype.forEach=function(e){this.map.forEach(e)},n.prototype.forEachEntry=function(n){return e.forEachEntry(this.map,n)},n.prototype.clone=function(){var t=new n;return e.copyEntries(this.map,t.map),t},n}();function R(n){return V(n)&&e.isSourceFile(n.parent)||e.isVariableDeclaration(n)&&e.isSourceFile(n.parent.parent.parent)}function F(n){return e.isVariableDeclaration(n)?n.parent.parent.parent:n.parent}function G(n){return e.Debug.assert(e.isSourceFile(n.parent)),V(n)||e.isVariableStatement(n)}function V(e){switch(e.kind){case 239:case 240:case 244:case 243:case 242:case 241:case 248:return!0;default:return!1}}function B(n,t){switch(n.kind){case 239:case 240:case 244:case 243:case 242:case 241:case 248:return t(n);case 219:return e.firstDefined(n.declarationList.declarations,function(e){return K(e.name,t)});case 221:var r=n.expression;return e.isBinaryExpression(r)&&1===e.getAssignmentDeclarationKind(r)?t(n):void 0}}function K(n,t){switch(n.kind){case 72:return t(e.cast(n.parent,function(n){return e.isVariableDeclaration(n)||e.isBindingElement(n)}));case 185:case 184:return e.firstDefined(n.elements,function(n){return e.isOmittedExpression(n)?void 0:K(n.name,t)});default:return e.Debug.assertNever(n)}}function j(n){return e.isExpressionStatement(n)?n.expression.left.name:e.tryCast(n.name,e.isIdentifier)}function U(n){switch(n.kind){case 237:return n.parent.parent;case 186:return U(e.cast(n.parent.parent,function(n){return e.isVariableDeclaration(n)||e.isBindingElement(n)}));default:return n}}function H(n,t,r,a){if(!W(n,t,a))if(a)e.isExpressionStatement(t)||r.insertExportModifier(n,t);else{var i=X(t);0!==i.length&&r.insertNodesAfter(n,t,i.map(Y))}}function W(n,t,r){return r?!e.isExpressionStatement(t)&&e.hasModifier(t,1):X(t).some(function(t){return n.symbol.exports.has(e.escapeLeadingUnderscores(t))})}function q(e,n){return n?[z(e)]:J(e)}function z(n){var t=e.concatenate([e.createModifier(85)],n.modifiers);switch(n.kind){case 239:return e.updateFunctionDeclaration(n,n.decorators,t,n.asteriskToken,n.name,n.typeParameters,n.parameters,n.type,n.body);case 240:return e.updateClassDeclaration(n,n.decorators,t,n.name,n.typeParameters,n.heritageClauses,n.members);case 219:return e.updateVariableStatement(n,t,n.declarationList);case 244:return e.updateModuleDeclaration(n,n.decorators,t,n.name,n.body);case 243:return e.updateEnumDeclaration(n,n.decorators,t,n.name,n.members);case 242:return e.updateTypeAliasDeclaration(n,n.decorators,t,n.name,n.typeParameters,n.type);case 241:return e.updateInterfaceDeclaration(n,n.decorators,t,n.name,n.typeParameters,n.heritageClauses,n.members);case 248:return e.updateImportEqualsDeclaration(n,n.decorators,t,n.name,n.moduleReference);case 221:return e.Debug.fail();default:return e.Debug.assertNever(n)}}function J(e){return[e].concat(X(e).map(Y))}function X(n){switch(n.kind){case 239:case 240:return[n.name.text];case 219:return e.mapDefined(n.declarationList.declarations,function(n){return e.isIdentifier(n.name)?n.name.text:void 0});case 244:case 243:case 242:case 241:case 248:return e.emptyArray;case 221:return e.Debug.fail();default:return e.Debug.assertNever(n)}}function Y(n){return e.createExpressionStatement(e.createBinary(e.createPropertyAccess(e.createIdentifier("exports"),e.createIdentifier(n)),59,e.createIdentifier(n)))}})(e.refactor||(e.refactor={}))}(u||(u={})),function(e){(function(n){(function(t){var r="Add or remove braces in an arrow function",a=e.Diagnostics.Add_or_remove_braces_in_an_arrow_function.message,i="Add braces to arrow function",o="Remove braces from arrow function",s=e.Diagnostics.Add_braces_to_arrow_function.message,c=e.Diagnostics.Remove_braces_from_arrow_function.message;function l(n){var t=n.file,l=n.startPosition,u=p(t,l);return u?[{name:r,description:a,actions:[u.addBraces?{name:i,description:s}:{name:o,description:c}]}]:e.emptyArray}function u(n,t){var r=n.file,a=n.startPosition,s=p(r,a);if(s){var c,l=s.expression,u=s.returnStatement,m=s.func;if(t===i){var f=e.createReturn(l);c=e.createBlock([f],!0),e.suppressLeadingAndTrailingTrivia(c),e.copyComments(l,f,r,3,!0)}else if(t===o&&u){var _=l||e.createVoidZero();c=d(_)?e.createParen(_):_,e.suppressLeadingAndTrailingTrivia(c),e.copyComments(u,c,r,3,!1)}else e.Debug.fail("invalid action");var g=e.textChanges.ChangeTracker.with(n,function(e){return e.replaceNode(r,m.body,c)});return{renameFilename:void 0,renameLocation:void 0,edits:g}}}function d(n){return e.isBinaryExpression(n)&&27===n.operatorToken.kind||e.isObjectLiteralExpression(n)}function p(n,t){var r=e.getTokenAtPosition(n,t),a=e.getContainingFunction(r);if(a&&e.isArrowFunction(a)&&e.rangeContainsRange(a,r)&&!e.rangeContainsRange(a.body,r)){if(e.isExpression(a.body))return{func:a,addBraces:!0,expression:a.body};if(1===a.body.statements.length){var i=e.first(a.body.statements);if(e.isReturnStatement(i))return{func:a,addBraces:!1,expression:i.expression,returnStatement:i}}}}n.registerRefactor(r,{getEditsForAction:u,getAvailableActions:l})})(n.addOrRemoveBracesToArrowFunction||(n.addOrRemoveBracesToArrowFunction={}))})(e.refactor||(e.refactor={}))}(u||(u={})),function(e){function n(n,r,a,i){var o=e.isNodeKind(n)?new t(n,r,a):72===n?new p(72,r,a):new d(n,r,a);return o.parent=i,o.flags=12679168&i.flags,o}e.servicesVersion="0.8";var t=function(){function n(e,n,t){this.pos=n,this.end=t,this.flags=0,this.transformFlags=void 0,this.parent=void 0,this.kind=e}return n.prototype.assertHasRealPosition=function(n){e.Debug.assert(!e.positionIsSynthesized(this.pos)&&!e.positionIsSynthesized(this.end),n||"Node must have a real position for this operation")},n.prototype.getSourceFile=function(){return e.getSourceFileOfNode(this)},n.prototype.getStart=function(n,t){return this.assertHasRealPosition(),e.getTokenPosOfNode(this,n,t)},n.prototype.getFullStart=function(){return this.assertHasRealPosition(),this.pos},n.prototype.getEnd=function(){return this.assertHasRealPosition(),this.end},n.prototype.getWidth=function(e){return this.assertHasRealPosition(),this.getEnd()-this.getStart(e)},n.prototype.getFullWidth=function(){return this.assertHasRealPosition(),this.end-this.pos},n.prototype.getLeadingTriviaWidth=function(e){return this.assertHasRealPosition(),this.getStart(e)-this.pos},n.prototype.getFullText=function(e){return this.assertHasRealPosition(),(e||this.getSourceFile()).text.substring(this.pos,this.end)},n.prototype.getText=function(e){return this.assertHasRealPosition(),e||(e=this.getSourceFile()),e.text.substring(this.getStart(e),this.getEnd())},n.prototype.getChildCount=function(e){return this.getChildren(e).length},n.prototype.getChildAt=function(e,n){return this.getChildren(n)[e]},n.prototype.getChildren=function(e){return this.assertHasRealPosition("Node without a real position cannot be scanned and thus has no token nodes - use forEachChild and collect the result if that's fine"),this._children||(this._children=r(this,e))},n.prototype.getFirstToken=function(n){this.assertHasRealPosition();var t=this.getChildren(n);if(t.length){var r=e.find(t,function(e){return e.kind<283||e.kind>305});return r.kind<148?r:r.getFirstToken(n)}},n.prototype.getLastToken=function(n){this.assertHasRealPosition();var t=this.getChildren(n),r=e.lastOrUndefined(t);if(r)return r.kind<148?r:r.getLastToken(n)},n.prototype.forEachChild=function(n,t){return e.forEachChild(this,n,t)},n}();function r(n,t){if(!e.isNodeKind(n.kind))return e.emptyArray;var r=[];if(e.isJSDocCommentContainingNode(n))return n.forEachChild(function(e){r.push(e)}),r;e.scanner.setText((t||n.getSourceFile()).text);var a=n.pos,o=function(e){i(r,a,e.pos,n),r.push(e),a=e.end},c=function(e){i(r,a,e.pos,n),r.push(s(e,n)),a=e.end};return e.forEach(n.jsDoc,o),a=n.pos,n.forEachChild(o,c),i(r,a,n.end,n),e.scanner.setText(void 0),r}function i(t,r,a,i){e.scanner.setTextPos(r);while(r<a){var o=e.scanner.scan(),s=e.scanner.getTextPos();if(s<=a&&(72===o&&e.Debug.fail("Did not expect "+e.Debug.showSyntaxKind(i)+" to have an Identifier in its trivia"),t.push(n(o,r,s,i))),r=s,1===o)break}}function s(e,t){var r=n(306,e.pos,e.end,t);r._children=[];for(var a=e.pos,o=0,s=e;o<s.length;o++){var c=s[o];i(r._children,a,c.pos,t),r._children.push(c),a=c.end}return i(r._children,a,e.end,t),r}var c=function(){function n(e,n){this.pos=e,this.end=n,this.flags=0,this.parent=void 0}return n.prototype.getSourceFile=function(){return e.getSourceFileOfNode(this)},n.prototype.getStart=function(n,t){return e.getTokenPosOfNode(this,n,t)},n.prototype.getFullStart=function(){return this.pos},n.prototype.getEnd=function(){return this.end},n.prototype.getWidth=function(e){return this.getEnd()-this.getStart(e)},n.prototype.getFullWidth=function(){return this.end-this.pos},n.prototype.getLeadingTriviaWidth=function(e){return this.getStart(e)-this.pos},n.prototype.getFullText=function(e){return(e||this.getSourceFile()).text.substring(this.pos,this.end)},n.prototype.getText=function(e){return e||(e=this.getSourceFile()),e.text.substring(this.getStart(e),this.getEnd())},n.prototype.getChildCount=function(){return 0},n.prototype.getChildAt=function(){},n.prototype.getChildren=function(){return 1===this.kind&&this.jsDoc||e.emptyArray},n.prototype.getFirstToken=function(){},n.prototype.getLastToken=function(){},n.prototype.forEachChild=function(){},n}(),u=function(){function n(e,n){this.flags=e,this.escapedName=n}return n.prototype.getFlags=function(){return this.flags},Object.defineProperty(n.prototype,"name",{get:function(){return e.symbolName(this)},enumerable:!0,configurable:!0}),n.prototype.getEscapedName=function(){return this.escapedName},n.prototype.getName=function(){return this.name},n.prototype.getDeclarations=function(){return this.declarations},n.prototype.getDocumentationComment=function(n){return this.documentationComment||(this.documentationComment=e.emptyArray,this.documentationComment=g(this.declarations,n)),this.documentationComment},n.prototype.getJsDocTags=function(){return void 0===this.tags&&(this.tags=e.JsDoc.getJsDocTagsFromDeclarations(this.declarations)),this.tags},n}(),d=function(e){function n(n,t,r){var a=e.call(this,t,r)||this;return a.kind=n,a}return l(n,e),n}(c),p=function(n){function t(e,t,r){return n.call(this,t,r)||this}return l(t,n),Object.defineProperty(t.prototype,"text",{get:function(){return e.idText(this)},enumerable:!0,configurable:!0}),t}(c);p.prototype.kind=72;var m=function(){function n(e,n){this.checker=e,this.flags=n}return n.prototype.getFlags=function(){return this.flags},n.prototype.getSymbol=function(){return this.symbol},n.prototype.getProperties=function(){return this.checker.getPropertiesOfType(this)},n.prototype.getProperty=function(e){return this.checker.getPropertyOfType(this,e)},n.prototype.getApparentProperties=function(){return this.checker.getAugmentedPropertiesOfType(this)},n.prototype.getCallSignatures=function(){return this.checker.getSignaturesOfType(this,0)},n.prototype.getConstructSignatures=function(){return this.checker.getSignaturesOfType(this,1)},n.prototype.getStringIndexType=function(){return this.checker.getIndexTypeOfType(this,0)},n.prototype.getNumberIndexType=function(){return this.checker.getIndexTypeOfType(this,1)},n.prototype.getBaseTypes=function(){return this.isClassOrInterface()?this.checker.getBaseTypes(this):void 0},n.prototype.getNonNullableType=function(){return this.checker.getNonNullableType(this)},n.prototype.getConstraint=function(){return this.checker.getBaseConstraintOfType(this)},n.prototype.getDefault=function(){return this.checker.getDefaultFromTypeParameter(this)},n.prototype.isUnion=function(){return!!(1048576&this.flags)},n.prototype.isIntersection=function(){return!!(2097152&this.flags)},n.prototype.isUnionOrIntersection=function(){return!!(3145728&this.flags)},n.prototype.isLiteral=function(){return!!(384&this.flags)},n.prototype.isStringLiteral=function(){return!!(128&this.flags)},n.prototype.isNumberLiteral=function(){return!!(256&this.flags)},n.prototype.isTypeParameter=function(){return!!(262144&this.flags)},n.prototype.isClassOrInterface=function(){return!!(3&e.getObjectFlags(this))},n.prototype.isClass=function(){return!!(1&e.getObjectFlags(this))},n}(),f=function(){function n(e){this.checker=e}return n.prototype.getDeclaration=function(){return this.declaration},n.prototype.getTypeParameters=function(){return this.typeParameters},n.prototype.getParameters=function(){return this.parameters},n.prototype.getReturnType=function(){return this.checker.getReturnTypeOfSignature(this)},n.prototype.getDocumentationComment=function(){return this.documentationComment||(this.documentationComment=g(e.singleElementArray(this.declaration),this.checker))},n.prototype.getJsDocTags=function(){return void 0===this.jsDocTags&&(this.jsDocTags=this.declaration?e.JsDoc.getJsDocTagsFromDeclarations([this.declaration]):[]),this.jsDocTags},n}();function _(n){return e.getJSDocTags(n).some(function(e){return"inheritDoc"===e.tagName.text})}function g(n,t){if(!n)return e.emptyArray;var r=e.JsDoc.getJsDocCommentsFromDeclarations(n);if(0===r.length||n.some(_))for(var a=0,i=n;a<i.length;a++){var o=i[a],s=v(o,o.symbol.name,t);s&&(r=0===r.length?s.slice():s.concat(e.lineBreakPart(),r))}return r}function v(n,t,r){return e.firstDefined(n.parent?e.getAllSuperTypeNodes(n.parent):e.emptyArray,function(e){var n=r.getTypeAtLocation(e),a=n&&r.getPropertyOfType(n,t),i=a&&a.getDocumentationComment(r);return i&&i.length?i:void 0})}var y=function(n){function t(e,t,r){return n.call(this,e,t,r)||this}return l(t,n),t.prototype.update=function(n,t){return e.updateSourceFile(this,n,t)},t.prototype.getLineAndCharacterOfPosition=function(n){return e.getLineAndCharacterOfPosition(this,n)},t.prototype.getLineStarts=function(){return e.getLineStarts(this)},t.prototype.getPositionOfLineAndCharacter=function(n,t,r){return e.computePositionOfLineAndCharacter(e.getLineStarts(this),n,t,this.text,r)},t.prototype.getLineEndOfPosition=function(e){var n,t=this.getLineAndCharacterOfPosition(e).line,r=this.getLineStarts();t+1>=r.length&&(n=this.getEnd()),n||(n=r[t+1]-1);var a=this.getFullText();return"\n"===a[n]&&"\r"===a[n-1]?n-1:n},t.prototype.getNamedDeclarations=function(){return this.namedDeclarations||(this.namedDeclarations=this.computeNamedDeclarations()),this.namedDeclarations},t.prototype.computeNamedDeclarations=function(){var n=e.createMultiMap();return this.forEachChild(i),n;function t(e){var t=a(e);t&&n.add(t,e)}function r(e){var t=n.get(e);return t||n.set(e,t=[]),t}function a(n){var t=e.getNonAssignedNameOfDeclaration(n);return t&&(e.isComputedPropertyName(t)&&e.isPropertyAccessExpression(t.expression)?t.expression.name.text:e.isPropertyName(t)?e.getNameFromPropertyName(t):void 0)}function i(n){switch(n.kind){case 239:case 196:case 156:case 155:var o=n,s=a(o);if(s){var c=r(s),l=e.lastOrUndefined(c);l&&o.parent===l.parent&&o.symbol===l.symbol?o.body&&!l.body&&(c[c.length-1]=o):c.push(o)}e.forEachChild(n,i);break;case 240:case 209:case 241:case 242:case 243:case 244:case 248:case 257:case 253:case 250:case 251:case 158:case 159:case 168:t(n),e.forEachChild(n,i);break;case 151:if(!e.hasModifier(n,92))break;case 237:case 186:var u=n;if(e.isBindingPattern(u.name)){e.forEachChild(u.name,i);break}u.initializer&&i(u.initializer);case 278:case 154:case 153:t(n);break;case 255:n.exportClause&&e.forEach(n.exportClause.elements,i);break;case 249:var d=n.importClause;d&&(d.name&&t(d.name),d.namedBindings&&(251===d.namedBindings.kind?t(d.namedBindings):e.forEach(d.namedBindings.elements,i)));break;case 204:0!==e.getAssignmentDeclarationKind(n)&&t(n);default:e.forEachChild(n,i)}}},t}(t),h=function(){function n(e,n,t){this.fileName=e,this.text=n,this.skipTrivia=t}return n.prototype.getLineAndCharacterOfPosition=function(n){return e.getLineAndCharacterOfPosition(this,n)},n}();function b(){return{getNodeConstructor:function(){return t},getTokenConstructor:function(){return d},getIdentifierConstructor:function(){return p},getSourceFileConstructor:function(){return y},getSymbolConstructor:function(){return u},getTypeConstructor:function(){return m},getSignatureConstructor:function(){return f},getSourceMapSourceConstructor:function(){return h}}}function E(n){var t=!0;for(var r in n)if(e.hasProperty(n,r)&&!T(r)){t=!1;break}if(t)return n;var a={};for(var r in n)if(e.hasProperty(n,r)){var i=T(r)?r:r.charAt(0).toLowerCase()+r.substr(1);a[i]=n[r]}return a}function T(e){return!e.length||e.charAt(0)===e.charAt(0).toLowerCase()}function S(n){return n?e.map(n,function(e){return e.text}).join(""):""}function x(){return{target:1,jsx:1}}function C(){return e.codefix.getSupportedErrorCodes()}e.toEditorSettings=E,e.displayPartsToString=S,e.getDefaultCompilerOptions=x,e.getSupportedCodeFixes=C;var A=function(){function n(n,t){this.host=n,this.currentDirectory=n.getCurrentDirectory(),this.fileNameToEntry=e.createMap();for(var r=n.getScriptFileNames(),a=0,i=r;a<i.length;a++){var o=i[a];this.createEntry(o,e.toPath(o,this.currentDirectory,t))}this._compilationSettings=n.getCompilationSettings()||x()}return n.prototype.compilationSettings=function(){return this._compilationSettings},n.prototype.getProjectReferences=function(){return this.host.getProjectReferences&&this.host.getProjectReferences()},n.prototype.createEntry=function(n,t){var r,a=this.host.getScriptSnapshot(n);return r=a?{hostFileName:n,version:this.host.getScriptVersion(n),scriptSnapshot:a,scriptKind:e.getScriptKind(n,this.host)}:n,this.fileNameToEntry.set(t,r),r},n.prototype.getEntryByPath=function(e){return this.fileNameToEntry.get(e)},n.prototype.getHostFileInformation=function(n){var t=this.fileNameToEntry.get(n);return e.isString(t)?void 0:t},n.prototype.getOrCreateEntryByPath=function(n,t){var r=this.getEntryByPath(t)||this.createEntry(n,t);return e.isString(r)?void 0:r},n.prototype.getRootFileNames=function(){var n=[];return this.fileNameToEntry.forEach(function(t){e.isString(t)?n.push(t):6!==t.scriptKind&&n.push(t.hostFileName)}),n},n.prototype.getVersion=function(e){var n=this.getHostFileInformation(e);return n&&n.version},n.prototype.getScriptSnapshot=function(e){var n=this.getHostFileInformation(e);return n&&n.scriptSnapshot},n}(),D=function(){function n(e){this.host=e}return n.prototype.getCurrentSourceFile=function(n){var t=this.host.getScriptSnapshot(n);if(!t)throw new Error("Could not find file: '"+n+"'.");var r,a=e.getScriptKind(n,this.host),i=this.host.getScriptVersion(n);if(this.currentFileName!==n)r=k(n,t,6,i,!0,a);else if(this.currentFileVersion!==i){var o=t.getChangeRange(this.currentFileScriptSnapshot);r=N(this.currentSourceFile,t,i,o)}return r&&(this.currentFileVersion=i,this.currentFileName=n,this.currentFileScriptSnapshot=t,this.currentSourceFile=r),this.currentSourceFile},n}();function L(e,n,t){e.version=t,e.scriptSnapshot=n}function k(n,t,r,a,i,o){var s=e.createSourceFile(n,e.getSnapshotText(t),r,i,o);return L(s,t,a),s}function N(n,t,r,a,i){if(a&&r!==n.version&&!e.disableIncrementalParsing){var o=void 0,s=0!==a.span.start?n.text.substr(0,a.span.start):"",c=e.textSpanEnd(a.span)!==n.text.length?n.text.substr(e.textSpanEnd(a.span)):"";if(0===a.newLength)o=s&&c?s+c:s||c;else{var l=t.getText(a.span.start,a.span.start+a.newLength);o=s&&c?s+l+c:s?s+l:l+c}var u=e.updateSourceFile(n,o,a,i);return L(u,t,r),u.nameTable=void 0,n!==u&&n.scriptSnapshot&&(n.scriptSnapshot.dispose&&n.scriptSnapshot.dispose(),n.scriptSnapshot=void 0),u}return k(n.fileName,t,n.languageVersion,r,!0,n.scriptKind)}e.createLanguageServiceSourceFile=k,e.disableIncrementalParsing=!1,e.updateLanguageServiceSourceFile=N;var I=function(){function n(e){this.cancellationToken=e}return n.prototype.isCancellationRequested=function(){return!!this.cancellationToken&&this.cancellationToken.isCancellationRequested()},n.prototype.throwIfCancellationRequested=function(){if(this.isCancellationRequested())throw new e.OperationCanceledException},n}(),w=function(){function n(e,n){void 0===n&&(n=20),this.hostCancellationToken=e,this.throttleWaitMilliseconds=n,this.lastCancellationCheckTime=0}return n.prototype.isCancellationRequested=function(){var n=e.timestamp(),t=Math.abs(n-this.lastCancellationCheckTime);return t>=this.throttleWaitMilliseconds&&(this.lastCancellationCheckTime=n,this.hostCancellationToken.isCancellationRequested())},n.prototype.throwIfCancellationRequested=function(){if(this.isCancellationRequested())throw new e.OperationCanceledException},n}();function M(n,t,r){var a;void 0===t&&(t=e.createDocumentRegistry(n.useCaseSensitiveFileNames&&n.useCaseSensitiveFileNames(),n.getCurrentDirectory())),void 0===r&&(r=!1);var i,s,c=new D(n),l=0,u=new I(n.getCancellationToken&&n.getCancellationToken()),d=n.getCurrentDirectory();function p(e){n.log&&n.log(e)}!e.localizedDiagnosticMessages&&n.getLocalizedDiagnosticMessages&&(e.localizedDiagnosticMessages=n.getLocalizedDiagnosticMessages());var m=e.hostUsesCaseSensitiveFileNames(n),f=e.createGetCanonicalFileName(m),_=e.getSourceMapper({useCaseSensitiveFileNames:function(){return m},getCurrentDirectory:function(){return d},getProgram:y,fileExists:n.fileExists&&function(e){return n.fileExists(e)},readFile:n.readFile&&function(e,t){return n.readFile(e,t)},getDocumentPositionMapper:n.getDocumentPositionMapper&&function(e,t){return n.getDocumentPositionMapper(e,t)},getSourceFileLike:n.getSourceFileLike&&function(e){return n.getSourceFileLike(e)},log:p});function g(e){var n=i.getSourceFile(e);if(!n)throw new Error("Could not find file: '"+e+"'.");return n}function v(){if(e.Debug.assert(!r),n.getProjectVersion){var a=n.getProjectVersion();if(a){if(s===a&&!n.hasChangedAutomaticTypeDirectiveNames)return;s=a}}var o=n.getTypeRootsVersion?n.getTypeRootsVersion():0;l!==o&&(p("TypeRoots version has changed; provide new program"),i=void 0,l=o);var c=new A(n,f),g=c.getRootFileNames(),v=n.hasInvalidatedResolution||e.returnFalse,y=c.getProjectReferences();if(!e.isProgramUptoDate(i,g,c.compilationSettings(),function(e){return c.getVersion(e)},S,v,!!n.hasChangedAutomaticTypeDirectiveNames,y)){var h=c.compilationSettings(),b={getSourceFile:D,getSourceFileByPath:L,getCancellationToken:function(){return u},getCanonicalFileName:f,useCaseSensitiveFileNames:function(){return m},getNewLine:function(){return e.getNewLineCharacter(h,function(){return e.getNewLineOrDefaultFromHost(n)})},getDefaultLibFileName:function(e){return n.getDefaultLibFileName(e)},writeFile:e.noop,getCurrentDirectory:function(){return d},fileExists:S,readFile:x,realpath:n.realpath&&function(e){return n.realpath(e)},directoryExists:function(t){return e.directoryProbablyExists(t,n)},getDirectories:function(e){return n.getDirectories?n.getDirectories(e):[]},readDirectory:function(t,r,a,i,o){return e.Debug.assertDefined(n.readDirectory,"'LanguageServiceHost.readDirectory' must be implemented to correctly process 'projectReferences'"),n.readDirectory(t,r,a,i,o)},onReleaseOldSourceFile:C,hasInvalidatedResolution:v,hasChangedAutomaticTypeDirectiveNames:n.hasChangedAutomaticTypeDirectiveNames};n.trace&&(b.trace=function(e){return n.trace(e)}),n.resolveModuleNames&&(b.resolveModuleNames=function(e,t,r,a){return n.resolveModuleNames(e,t,r,a)}),n.resolveTypeReferenceDirectives&&(b.resolveTypeReferenceDirectives=function(e,t,r){return n.resolveTypeReferenceDirectives(e,t,r)});var E=t.getKeyForCompilationSettings(h),T={rootNames:g,options:h,host:b,oldProgram:i,projectReferences:y};return i=e.createProgram(T),c=void 0,_.clearCache(),void i.getTypeChecker()}function S(t){var r=e.toPath(t,d,f),a=c&&c.getEntryByPath(r);return a?!e.isString(a):!!n.fileExists&&n.fileExists(t)}function x(t){var r=e.toPath(t,d,f),a=c&&c.getEntryByPath(r);return a?e.isString(a)?void 0:e.getSnapshotText(a.scriptSnapshot):n.readFile&&n.readFile(t)}function C(e,n){var r=t.getKeyForCompilationSettings(n);t.releaseDocumentWithKey(e.resolvedPath,r)}function D(n,t,r,a){return L(n,e.toPath(n,d,f),t,r,a)}function L(n,r,a,o,s){e.Debug.assert(void 0!==c,"getOrCreateSourceFileByPath called after typical CompilerHost lifetime, check the callstack something with a reference to an old host.");var l=c&&c.getOrCreateEntryByPath(n,r);if(l){if(!s){var u=i&&i.getSourceFileByPath(r);if(u)return e.Debug.assertEqual(l.scriptKind,u.scriptKind,"Registered script kind should match new script kind.",r),t.updateDocumentWithKey(n,r,h,E,l.scriptSnapshot,l.version,l.scriptKind)}return t.acquireDocumentWithKey(n,r,h,E,l.scriptSnapshot,l.version,l.scriptKind)}}}function y(){if(!r)return v(),i;e.Debug.assert(void 0===i)}function h(){i=void 0}function b(){i&&(e.forEach(i.getSourceFiles(),function(e){return t.releaseDocument(e.fileName,i.getCompilerOptions())}),i=void 0),n=void 0}function T(e){return v(),i.getSyntacticDiagnostics(g(e),u).slice()}function S(n){v();var t=g(n),r=i.getSemanticDiagnostics(t,u);if(!e.getEmitDeclarations(i.getCompilerOptions()))return r.slice();var a=i.getDeclarationDiagnostics(t,u);return r.concat(a)}function x(n){return v(),e.computeSuggestionDiagnostics(g(n),i,u)}function C(){return v(),i.getOptionsDiagnostics(u).concat(i.getGlobalDiagnostics(u))}function L(t,r,a){void 0===a&&(a=e.emptyOptions);var s=o({},e.identity(a),{includeCompletionsForModuleExports:a.includeCompletionsForModuleExports||a.includeExternalModuleExports,includeCompletionsWithInsertText:a.includeCompletionsWithInsertText||a.includeInsertTextCompletions});return v(),e.Completions.getCompletionsAtPosition(n,i,p,g(t),r,s,a.triggerCharacter)}function k(t,r,a,o,s,c){return void 0===c&&(c=e.emptyOptions),v(),e.Completions.getCompletionEntryDetails(i,p,g(t),r,{name:a,source:s},n,o&&e.formatting.getFormatContext(o),c,u)}function N(n,t,r,a){return v(),e.Completions.getCompletionEntrySymbol(i,p,g(n),t,{name:r,source:a})}function w(n,t){v();var r=g(n),a=e.getTouchingPropertyName(r,t);if(a!==r){var o=i.getTypeChecker(),s=V(a,o);if(!s||o.isUnknownSymbol(s)){var c=M(r,a,t)?o.getTypeAtLocation(a):void 0;return c&&{kind:"",kindModifiers:"",textSpan:e.createTextSpanFromNode(a,r),displayParts:o.runWithCancellationToken(u,function(n){return e.typeToDisplayParts(n,c,e.getContainerNode(a))}),documentation:c.symbol?c.symbol.getDocumentationComment(o):void 0,tags:c.symbol?c.symbol.getJsDocTags():void 0}}var l=o.runWithCancellationToken(u,function(n){return e.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(n,s,r,e.getContainerNode(a),a)}),d=l.symbolKind,p=l.displayParts,m=l.documentation,f=l.tags;return{kind:d,kindModifiers:e.SymbolDisplay.getSymbolModifiers(s),textSpan:e.createTextSpanFromNode(a,r),displayParts:p,documentation:m,tags:f}}}function M(n,t,r){switch(t.kind){case 72:return!e.isLabelName(t)&&!e.isTagName(t);case 189:case 148:return!e.isInComment(n,r);case 100:case 178:case 98:return!0;default:return!1}}function O(n,t){return v(),e.GoToDefinition.getDefinitionAtPosition(i,g(n),t)}function P(n,t){return v(),e.GoToDefinition.getDefinitionAndBoundSpan(i,g(n),t)}function R(n,t){return v(),e.GoToDefinition.getTypeDefinitionAtPosition(i.getTypeChecker(),g(n),t)}function F(n,t){return v(),e.FindAllReferences.getImplementationsAtPosition(i,u,i.getSourceFiles(),g(n),t)}function G(n,t){return e.flatMap(B(n,t,[n]),function(e){return e.highlightSpans.map(function(n){return{fileName:e.fileName,textSpan:n.textSpan,isWriteAccess:"writtenReference"===n.kind,isDefinition:!1,isInString:n.isInString}})})}function B(n,t,r){var a=e.normalizePath(n);e.Debug.assert(r.some(function(n){return e.normalizePath(n)===a})),v();var o=r.map(g),s=g(n);return e.DocumentHighlights.getDocumentHighlights(i,u,s,t,o)}function K(n,t,r,a,i){v();var o=g(n),s=e.getTouchingPropertyName(o,t);if(e.isIdentifier(s)&&(e.isJsxOpeningElement(s.parent)||e.isJsxClosingElement(s.parent))&&e.isIntrinsicJsxName(s.escapedText)){var c=s.parent.parent,l=c.openingElement,u=c.closingElement;return[l,u].map(function(n){return{fileName:o.fileName,textSpan:e.createTextSpanFromNode(n.tagName,o)}})}return U(s,t,{findInStrings:r,findInComments:a,providePrefixAndSuffixTextForRename:i,isForRename:!0},function(n,t,r){return e.FindAllReferences.toRenameLocation(n,t,r,i||!1)})}function j(n,t){return v(),U(e.getTouchingPropertyName(g(n),t),t,{},e.FindAllReferences.toReferenceEntry)}function U(n,t,r,a){v();var o=r&&r.isForRename?i.getSourceFiles().filter(function(e){return!i.isSourceFileDefaultLibrary(e)}):i.getSourceFiles();return e.FindAllReferences.findReferenceOrRenameEntries(i,u,o,n,t,r,a)}function H(n,t){return v(),e.FindAllReferences.findReferencedSymbols(i,u,i.getSourceFiles(),g(n),t)}function W(n,t,r,a){void 0===a&&(a=!1),v();var o=r?[g(r)]:i.getSourceFiles();return e.NavigateTo.getNavigateToItems(o,i.getTypeChecker(),u,n,t,a)}function q(t,r){void 0===r&&(r=!1),v();var a=g(t),o=n.getCustomTransformers&&n.getCustomTransformers();return e.getFileEmitOutput(i,a,r,u,o)}function z(n,t,r){var a=(void 0===r?e.emptyOptions:r).triggerReason;v();var o=g(n);return e.SignatureHelp.getSignatureHelpItems(i,o,t,a,u)}function J(e){return c.getCurrentSourceFile(e)}function X(n,t,r){var a=c.getCurrentSourceFile(n),i=e.getTouchingPropertyName(a,t);if(i!==a){switch(i.kind){case 189:case 148:case 10:case 87:case 102:case 96:case 98:case 100:case 178:case 72:break;default:return}var o=i;while(1)if(e.isRightSideOfPropertyAccess(o)||e.isRightSideOfQualifiedName(o))o=o.parent;else{if(!e.isNameOfModuleDeclaration(o))break;if(244!==o.parent.parent.kind||o.parent.parent.body!==o.parent)break;o=o.parent.parent.name}return e.createTextSpanFromBounds(o.getStart(),i.getEnd())}}function Y(n,t){var r=c.getCurrentSourceFile(n);return e.BreakpointResolver.spanInSourceFileAtLocation(r,t)}function Q(n){return e.NavigationBar.getNavigationBarItems(c.getCurrentSourceFile(n),u)}function Z(n){return e.NavigationBar.getNavigationTree(c.getCurrentSourceFile(n),u)}function $(t){var r=e.getScriptKind(t,n);return 3===r||4===r}function ee(n,t){return $(n)?(v(),e.getSemanticClassifications(i.getTypeChecker(),u,g(n),i.getClassifiableNames(),t)):[]}function ne(n,t){return $(n)?(v(),e.getEncodedSemanticClassifications(i.getTypeChecker(),u,g(n),i.getClassifiableNames(),t)):{spans:[],endOfLineState:0}}function te(n,t){return e.getSyntacticClassifications(u,c.getCurrentSourceFile(n),t)}function re(n,t){return e.getEncodedSyntacticClassifications(u,c.getCurrentSourceFile(n),t)}function ae(n){var t=c.getCurrentSourceFile(n);return e.OutliningElementsCollector.collectElements(t,u)}var ie=e.createMapFromTemplate((a={},a[18]=19,a[20]=21,a[22]=23,a[30]=28,a));function oe(n,t){var r=c.getCurrentSourceFile(n),a=e.getTouchingToken(r,t),i=a.getStart(r)===t?ie.get(a.kind.toString()):void 0,o=i&&e.findChildOfKind(a.parent,i,r);return o?[e.createTextSpanFromNode(a,r),e.createTextSpanFromNode(o,r)].sort(function(e,n){return e.start-n.start}):e.emptyArray}function se(n,t,r){var a=e.timestamp(),i=E(r),o=c.getCurrentSourceFile(n);p("getIndentationAtPosition: getCurrentSourceFile: "+(e.timestamp()-a)),a=e.timestamp();var s=e.formatting.SmartIndenter.getIndentation(t,o,i);return p("getIndentationAtPosition: computeIndentation  : "+(e.timestamp()-a)),s}function ce(n,t,r,a){var i=c.getCurrentSourceFile(n);return e.formatting.formatSelection(t,r,i,e.formatting.getFormatContext(E(a)))}function le(n,t){return e.formatting.formatDocument(c.getCurrentSourceFile(n),e.formatting.getFormatContext(E(t)))}function ue(n,t,r,a){var i=c.getCurrentSourceFile(n),o=e.formatting.getFormatContext(E(a));if(!e.isInComment(i,t))switch(r){case"{":return e.formatting.formatOnOpeningCurly(t,i,o);case"}":return e.formatting.formatOnClosingCurly(t,i,o);case";":return e.formatting.formatOnSemicolon(t,i,o);case"\n":return e.formatting.formatOnEnter(t,i,o)}return[]}function de(t,r,a,o,s,c){void 0===c&&(c=e.emptyOptions),v();var l=g(t),d=e.createTextSpanFromBounds(r,a),p=e.formatting.getFormatContext(s);return e.flatMap(e.deduplicate(o,e.equateValues,e.compareValues),function(t){return u.throwIfCancellationRequested(),e.codefix.getFixes({errorCode:t,sourceFile:l,span:d,program:i,host:n,cancellationToken:u,formatContext:p,preferences:c})})}function pe(t,r,a,o){void 0===o&&(o=e.emptyOptions),v(),e.Debug.assert("file"===t.type);var s=g(t.fileName),c=e.formatting.getFormatContext(a);return e.codefix.getAllFixes({fixId:r,sourceFile:s,program:i,host:n,cancellationToken:u,formatContext:c,preferences:o})}function me(t,r,a){void 0===a&&(a=e.emptyOptions),v(),e.Debug.assert("file"===t.type);var o=g(t.fileName),s=e.formatting.getFormatContext(r);return e.OrganizeImports.organizeImports(o,s,n,i,a)}function fe(t,r,a,i){return void 0===i&&(i=e.emptyOptions),e.getEditsForFileRename(y(),t,r,n,e.formatting.getFormatContext(a),i,_)}function _e(n,t){var r="string"===typeof n?t:n,a="string"!==typeof n?t:void 0;return e.isArray(r)?Promise.all(r.map(function(e){return ge(e,a)})):ge(r,a)}function ge(t,r){var a=function(n){return e.toPath(n,d,f)};switch(t.type){case"install package":return n.installPackage?n.installPackage({fileName:a(t.file),packageName:t.packageName}):Promise.reject("Host does not implement `installPackage`");case"generate types":var i=t.fileToGenerateTypesFor,o=t.outputFileName;if(!n.inspectValue)return Promise.reject("Host does not implement `installPackage`");var s=n.inspectValue({fileNameToRequire:i});return s.then(function(t){var i=a(o);return n.writeFile(i,e.valueInfoToDeclarationFileText(t,r||e.testFormatSettings)),{successMessage:"Wrote types to '"+i+"'"}});default:return e.Debug.assertNever(t)}}function ve(t,r){return e.JsDoc.getDocCommentTemplateAtPosition(e.getNewLineOrDefaultFromHost(n),c.getCurrentSourceFile(t),r)}function ye(n,t,r){if(60===r)return!1;var a=c.getCurrentSourceFile(n);if(e.isInString(a,t))return!1;if(e.isInsideJsxElementOrAttribute(a,t))return 123===r;if(e.isInTemplateString(a,t))return!1;switch(r){case 39:case 34:case 96:return!e.isInComment(a,t)}return!0}function he(n,t){var r=c.getCurrentSourceFile(n),a=e.findPrecedingToken(t,r);if(a){var i=30===a.kind&&e.isJsxOpeningElement(a.parent)?a.parent.parent:e.isJsxText(a)?a.parent:void 0;return i&&be(i)?{newText:"</"+i.openingElement.tagName.getText(r)+">"}:void 0}}function be(n){var t=n.openingElement,r=n.closingElement,a=n.parent;return!e.tagNamesAreEquivalent(t.tagName,r.tagName)||e.isJsxElement(a)&&e.tagNamesAreEquivalent(t.tagName,a.openingElement.tagName)&&be(a)}function Ee(n,t,r){var a=c.getCurrentSourceFile(n),i=e.formatting.getRangeOfEnclosingComment(a,t);return!i||r&&3!==i.kind?void 0:e.createTextSpanFromRange(i)}function Te(n,t){v();var r=g(n);u.throwIfCancellationRequested();var a=r.text,i=[];if(t.length>0&&!b(r.fileName)){var o=y(),s=void 0;while(s=o.exec(a)){u.throwIfCancellationRequested();var c=3;e.Debug.assert(s.length===t.length+c);var l=s[1],d=s.index+l.length;if(e.isInComment(r,d)){for(var p=void 0,m=0;m<t.length;m++)s[m+c]&&(p=t[m]);if(void 0===p)return e.Debug.fail();if(!h(a.charCodeAt(d+p.text.length))){var f=s[2];i.push({descriptor:p,message:f,position:d})}}}}return i;function _(e){return e.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g,"\\$&")}function y(){var n=/(?:\/\/+\s*)/.source,r=/(?:\/\*+\s*)/.source,a=/(?:^(?:\s|\*)*)/.source,i="("+a+"|"+n+"|"+r+")",o="(?:"+e.map(t,function(e){return"("+_(e.text)+")"}).join("|")+")",s=/(?:$|\*\/)/.source,c=/(?:.*?)/.source,l="("+o+c+")",u=i+l+s;return new RegExp(u,"gim")}function h(e){return e>=97&&e<=122||e>=65&&e<=90||e>=48&&e<=57}function b(n){return e.stringContains(n,"/node_modules/")}}function Se(n,t,r){return v(),e.Rename.getRenameInfo(i,g(n),t,r)}function xe(t,r,a,i){var o="number"===typeof r?[r,void 0]:[r.pos,r.end],s=o[0],c=o[1];return{file:t,startPosition:s,endPosition:c,program:y(),host:n,formatContext:e.formatting.getFormatContext(i),cancellationToken:u,preferences:a}}function Ce(n,t,r){void 0===r&&(r=e.emptyOptions),v();var a=g(n);return e.refactor.getApplicableRefactors(xe(a,t,r))}function Ae(n,t,r,a,i,o){void 0===o&&(o=e.emptyOptions),v();var s=g(n);return e.refactor.getEditsForRefactor(xe(s,r,o,t),a,i)}return ie.forEach(function(e,n){return ie.set(e.toString(),Number(n))}),{dispose:b,cleanupSemanticCache:h,getSyntacticDiagnostics:T,getSemanticDiagnostics:S,getSuggestionDiagnostics:x,getCompilerOptionsDiagnostics:C,getSyntacticClassifications:te,getSemanticClassifications:ee,getEncodedSyntacticClassifications:re,getEncodedSemanticClassifications:ne,getCompletionsAtPosition:L,getCompletionEntryDetails:k,getCompletionEntrySymbol:N,getSignatureHelpItems:z,getQuickInfoAtPosition:w,getDefinitionAtPosition:O,getDefinitionAndBoundSpan:P,getImplementationAtPosition:F,getTypeDefinitionAtPosition:R,getReferencesAtPosition:j,findReferences:H,getOccurrencesAtPosition:G,getDocumentHighlights:B,getNameOrDottedNameSpan:X,getBreakpointStatementAtPosition:Y,getNavigateToItems:W,getRenameInfo:Se,findRenameLocations:K,getNavigationBarItems:Q,getNavigationTree:Z,getOutliningSpans:ae,getTodoComments:Te,getBraceMatchingAtPosition:oe,getIndentationAtPosition:se,getFormattingEditsForRange:ce,getFormattingEditsForDocument:le,getFormattingEditsAfterKeystroke:ue,getDocCommentTemplateAtPosition:ve,isValidBraceCompletionAtPosition:ye,getJsxClosingTagAtPosition:he,getSpanOfEnclosingComment:Ee,getCodeFixesAtPosition:de,getCombinedCodeFix:pe,applyCodeActionCommand:_e,organizeImports:me,getEditsForFileRename:fe,getEmitOutput:q,getNonBoundSourceFile:J,getProgram:y,getApplicableRefactors:Ce,getEditsForRefactor:Ae,toLineColumnOffset:_.toLineColumnOffset,getSourceMapper:function(){return _}}}function O(e){return e.nameTable||P(e),e.nameTable}function P(n){var t=n.nameTable=e.createUnderscoreEscapedMap();n.forEachChild(function n(r){if(e.isIdentifier(r)&&!e.isTagName(r)&&r.escapedText||e.isStringOrNumericLiteralLike(r)&&R(r)){var a=e.getEscapedTextOfIdentifierOrLiteral(r);t.set(a,void 0===t.get(a)?r.pos:-1)}if(e.forEachChild(r,n),e.hasJSDocNodes(r))for(var i=0,o=r.jsDoc;i<o.length;i++){var s=o[i];e.forEachChild(s,n)}})}function R(n){return e.isDeclarationName(n)||259===n.parent.kind||K(n)||e.isLiteralComputedPropertyDeclarationName(n)}function F(n){var t=G(n);return t&&(e.isObjectLiteralExpression(t.parent)||e.isJsxAttributes(t.parent))?t:void 0}function G(n){switch(n.kind){case 10:case 8:if(149===n.parent.kind)return e.isObjectLiteralElement(n.parent.parent)?n.parent.parent:void 0;case 72:return!e.isObjectLiteralElement(n.parent)||188!==n.parent.parent.kind&&268!==n.parent.parent.kind||n.parent.name!==n?void 0:n.parent}}function V(n,t){var r=F(n);if(r){var a=t.getContextualType(r.parent),i=a&&B(r,t,a,!1);if(i&&1===i.length)return e.first(i)}return t.getSymbolAtLocation(n)}function B(n,t,r,a){var i=e.getNameFromPropertyName(n.name);if(!i)return e.emptyArray;if(!r.isUnion()){var o=r.getProperty(i);return o?[o]:e.emptyArray}var s=e.mapDefined(r.types,function(r){return e.isObjectLiteralExpression(n.parent)&&t.isTypeInvalidDueToUnionDiscriminant(r,n.parent)?void 0:r.getProperty(i)});if(a&&(0===s.length||s.length===r.types.length)){o=r.getProperty(i);if(o)return[o]}return 0===s.length?e.mapDefined(r.types,function(e){return e.getProperty(i)}):s}function K(e){return e&&e.parent&&190===e.parent.kind&&e.parent.argumentExpression===e}function j(n){return a+e.directorySeparator+e.getDefaultLibFileName(n)}e.ThrottledCancellationToken=w,e.createLanguageService=M,e.getNameTable=O,e.getContainingObjectLiteralElement=F,e.getPropertySymbolsFromContextualType=B,e.getDefaultLibFilePath=j,e.objectAllocator=b()}(u||(u={})),function(e){(function(n){function t(n,t){if(!n.isDeclarationFile){var r=e.getTokenAtPosition(n,t),a=n.getLineAndCharacterOfPosition(t).line;if(n.getLineAndCharacterOfPosition(r.getStart(n)).line>a){var i=e.findPrecedingToken(r.pos,n);if(!i||n.getLineAndCharacterOfPosition(i.getEnd()).line!==a)return;r=i}if(!(4194304&r.flags))return p(r)}function o(t,r){var a=t.decorators?e.skipTrivia(n.text,t.decorators.end):t.getStart(n);return e.createTextSpanFromBounds(a,(r||t).getEnd())}function s(t,r){return o(t,e.findNextToken(r,r.parent,n))}function c(e,t){return e&&a===n.getLineAndCharacterOfPosition(e.getStart(n)).line?p(e):p(t)}function l(t){return e.createTextSpanFromBounds(e.skipTrivia(n.text,t.pos),t.end)}function u(t){return p(e.findPrecedingToken(t.pos,n))}function d(t){return p(e.findNextToken(t,t.parent,n))}function p(t){if(t){var r=t.parent;switch(t.kind){case 219:return y(t.declarationList.declarations[0]);case 237:case 154:case 153:return y(t);case 151:return b(t);case 239:case 156:case 155:case 158:case 159:case 157:case 196:case 197:return T(t);case 218:if(e.isFunctionBlock(t))return S(t);case 245:return x(t);case 274:return x(t.block);case 221:return o(t.expression);case 230:return o(t.getChildAt(0),t.expression);case 224:return s(t,t.expression);case 223:return p(t.statement);case 236:return o(t.getChildAt(0));case 222:return s(t,t.expression);case 233:return p(t.statement);case 229:case 228:return o(t.getChildAt(0),t.label);case 225:return A(t);case 226:return s(t,t.expression);case 227:return C(t);case 232:return s(t,t.expression);case 271:case 272:return p(t.statements[0]);case 235:return x(t.tryBlock);case 234:return o(t,t.expression);case 254:return o(t,t.expression);case 248:return o(t,t.moduleReference);case 249:return o(t,t.moduleSpecifier);case 255:return o(t,t.moduleSpecifier);case 244:if(1!==e.getModuleInstanceState(t))return;case 240:case 243:case 278:case 186:return o(t);case 231:return p(t.statement);case 152:return l(r.decorators);case 184:case 185:return D(t);case 241:case 242:return;case 26:case 1:return c(e.findPrecedingToken(t.pos,n));case 27:return u(t);case 18:return k(t);case 19:return N(t);case 23:return I(t);case 20:return w(t);case 21:return M(t);case 57:return O(t);case 30:case 28:return P(t);case 107:return R(t);case 83:case 75:case 88:return d(t);case 147:return F(t);default:if(e.isArrayLiteralOrObjectLiteralDestructuringPattern(t))return L(t);if((72===t.kind||208===t.kind||275===t.kind||276===t.kind)&&e.isArrayLiteralOrObjectLiteralDestructuringPattern(r))return o(t);if(204===t.kind){var a=t,i=a.left,m=a.operatorToken;if(e.isArrayLiteralOrObjectLiteralDestructuringPattern(i))return L(i);if(59===m.kind&&e.isArrayLiteralOrObjectLiteralDestructuringPattern(t.parent))return o(t);if(27===m.kind)return p(i)}if(e.isExpressionNode(t))switch(r.kind){case 223:return u(t);case 152:return p(t.parent);case 225:case 227:return o(t);case 204:if(27===t.parent.operatorToken.kind)return o(t);break;case 197:if(t.parent.body===t)return o(t);break}switch(t.parent.kind){case 275:if(t.parent.name===t&&!e.isArrayLiteralOrObjectLiteralDestructuringPattern(t.parent.parent))return p(t.parent.initializer);break;case 194:if(t.parent.type===t)return d(t.parent.type);break;case 237:case 151:var f=t.parent,_=f.initializer,g=f.type;if(_===t||g===t||e.isAssignmentOperator(t.kind))return u(t);break;case 204:i=t.parent.left;if(e.isArrayLiteralOrObjectLiteralDestructuringPattern(i)&&t!==i)return u(t);break;default:if(e.isFunctionLike(t.parent)&&t.parent.type===t)return u(t)}return p(t.parent)}}function v(t){return e.isVariableDeclarationList(t.parent)&&t.parent.declarations[0]===t?o(e.findPrecedingToken(t.pos,n,t.parent),t):o(t)}function y(t){if(226===t.parent.parent.kind)return p(t.parent.parent);var r=t.parent;return e.isBindingPattern(t.name)?D(t.name):t.initializer||e.hasModifier(t,1)||227===r.parent.kind?v(t):e.isVariableDeclarationList(t.parent)&&t.parent.declarations[0]!==t?p(e.findPrecedingToken(t.pos,n,t.parent)):void 0}function h(n){return!!n.initializer||void 0!==n.dotDotDotToken||e.hasModifier(n,12)}function b(n){if(e.isBindingPattern(n.name))return D(n.name);if(h(n))return o(n);var t=n.parent,r=t.parameters.indexOf(n);return e.Debug.assert(-1!==r),0!==r?b(t.parameters[r-1]):p(t.body)}function E(n){return e.hasModifier(n,1)||240===n.parent.kind&&157!==n.kind}function T(e){if(e.body)return E(e)?o(e):p(e.body)}function S(e){var n=e.statements.length?e.statements[0]:e.getLastToken();return E(e.parent)?c(e.parent,n):p(n)}function x(t){switch(t.parent.kind){case 244:if(1!==e.getModuleInstanceState(t.parent))return;case 224:case 222:case 226:return c(t.parent,t.statements[0]);case 225:case 227:return c(e.findPrecedingToken(t.pos,n,t.parent),t.statements[0])}return p(t.statements[0])}function C(e){if(238!==e.initializer.kind)return p(e.initializer);var n=e.initializer;return n.declarations.length>0?p(n.declarations[0]):void 0}function A(e){return e.initializer?C(e):e.condition?o(e.condition):e.incrementor?o(e.incrementor):void 0}function D(n){var t=e.forEach(n.elements,function(e){return 210!==e.kind?e:void 0});return t?p(t):186===n.parent.kind?o(n.parent):v(n.parent)}function L(n){e.Debug.assert(185!==n.kind&&184!==n.kind);var t=187===n.kind?n.elements:n.properties,r=e.forEach(t,function(e){return 210!==e.kind?e:void 0});return r?p(r):o(204===n.parent.kind?n.parent:n)}function k(t){switch(t.parent.kind){case 243:var r=t.parent;return c(e.findPrecedingToken(t.pos,n,t.parent),r.members.length?r.members[0]:r.getLastToken(n));case 240:var a=t.parent;return c(e.findPrecedingToken(t.pos,n,t.parent),a.members.length?a.members[0]:a.getLastToken(n));case 246:return c(t.parent.parent,t.parent.clauses[0])}return p(t.parent)}function N(n){switch(n.parent.kind){case 245:if(1!==e.getModuleInstanceState(n.parent.parent))return;case 243:case 240:return o(n);case 218:if(e.isFunctionBlock(n.parent))return o(n);case 274:return p(e.lastOrUndefined(n.parent.statements));case 246:var t=n.parent,r=e.lastOrUndefined(t.clauses);return r?p(e.lastOrUndefined(r.statements)):void 0;case 184:var a=n.parent;return p(e.lastOrUndefined(a.elements)||a);default:if(e.isArrayLiteralOrObjectLiteralDestructuringPattern(n.parent)){var i=n.parent;return o(e.lastOrUndefined(i.properties)||i)}return p(n.parent)}}function I(n){switch(n.parent.kind){case 185:var t=n.parent;return o(e.lastOrUndefined(t.elements)||t);default:if(e.isArrayLiteralOrObjectLiteralDestructuringPattern(n.parent)){var r=n.parent;return o(e.lastOrUndefined(r.elements)||r)}return p(n.parent)}}function w(e){return 223===e.parent.kind||191===e.parent.kind||192===e.parent.kind?u(e):195===e.parent.kind?d(e):p(e.parent)}function M(e){switch(e.parent.kind){case 196:case 239:case 197:case 156:case 155:case 158:case 159:case 157:case 224:case 223:case 225:case 227:case 191:case 192:case 195:return u(e);default:return p(e.parent)}}function O(n){return e.isFunctionLike(n.parent)||275===n.parent.kind||151===n.parent.kind?u(n):p(n.parent)}function P(e){return 194===e.parent.kind?d(e):p(e.parent)}function R(e){return 223===e.parent.kind?s(e,e.parent.expression):p(e.parent)}function F(e){return 227===e.parent.kind?d(e):p(e.parent)}}}n.spanInSourceFileAtLocation=t})(e.BreakpointResolver||(e.BreakpointResolver={}))}(u||(u={})),function(e){function n(n,t,r){var a=[];r=e.fixupCompilerOptions(r,a);var i=e.isArray(n)?n:[n],o=e.transformNodes(void 0,void 0,r,i,t,!0);return o.diagnostics=e.concatenate(o.diagnostics,a),o}e.transform=n}(u||(u={}));var u,d,p=function(){return this}();(function(e){function n(e,n){e&&e.log("*INTERNAL ERROR* - Exception in typescript services: "+n.message)}var t=function(){function n(e){this.scriptSnapshotShim=e}return n.prototype.getText=function(e,n){return this.scriptSnapshotShim.getText(e,n)},n.prototype.getLength=function(){return this.scriptSnapshotShim.getLength()},n.prototype.getChangeRange=function(n){var t=n,r=this.scriptSnapshotShim.getChangeRange(t.scriptSnapshotShim);if(null===r)return null;var a=JSON.parse(r);return e.createTextChangeRange(e.createTextSpan(a.span.start,a.span.length),a.newLength)},n.prototype.dispose=function(){"dispose"in this.scriptSnapshotShim&&this.scriptSnapshotShim.dispose()},n}(),r=function(){function n(n){var t=this;this.shimHost=n,this.loggingEnabled=!1,this.tracingEnabled=!1,"getModuleResolutionsForFile"in this.shimHost&&(this.resolveModuleNames=function(n,r){var a=JSON.parse(t.shimHost.getModuleResolutionsForFile(r));return e.map(n,function(n){var t=e.getProperty(a,n);return t?{resolvedFileName:t,extension:e.extensionFromPath(t),isExternalLibraryImport:!1}:void 0})}),"directoryExists"in this.shimHost&&(this.directoryExists=function(e){return t.shimHost.directoryExists(e)}),"getTypeReferenceDirectiveResolutionsForFile"in this.shimHost&&(this.resolveTypeReferenceDirectives=function(n,r){var a=JSON.parse(t.shimHost.getTypeReferenceDirectiveResolutionsForFile(r));return e.map(n,function(n){return e.getProperty(a,n)})})}return n.prototype.log=function(e){this.loggingEnabled&&this.shimHost.log(e)},n.prototype.trace=function(e){this.tracingEnabled&&this.shimHost.trace(e)},n.prototype.error=function(e){this.shimHost.error(e)},n.prototype.getProjectVersion=function(){if(this.shimHost.getProjectVersion)return this.shimHost.getProjectVersion()},n.prototype.getTypeRootsVersion=function(){return this.shimHost.getTypeRootsVersion?this.shimHost.getTypeRootsVersion():0},n.prototype.useCaseSensitiveFileNames=function(){return!!this.shimHost.useCaseSensitiveFileNames&&this.shimHost.useCaseSensitiveFileNames()},n.prototype.getCompilationSettings=function(){var e=this.shimHost.getCompilationSettings();if(null===e||""===e)throw Error("LanguageServiceShimHostAdapter.getCompilationSettings: empty compilationSettings");var n=JSON.parse(e);return n.allowNonTsExtensions=!0,n},n.prototype.getScriptFileNames=function(){var e=this.shimHost.getScriptFileNames();return JSON.parse(e)},n.prototype.getScriptSnapshot=function(e){var n=this.shimHost.getScriptSnapshot(e);return n&&new t(n)},n.prototype.getScriptKind=function(e){return"getScriptKind"in this.shimHost?this.shimHost.getScriptKind(e):0},n.prototype.getScriptVersion=function(e){return this.shimHost.getScriptVersion(e)},n.prototype.getLocalizedDiagnosticMessages=function(){var e=this.shimHost.getLocalizedDiagnosticMessages();if(null===e||""===e)return null;try{return JSON.parse(e)}catch(n){return this.log(n.description||"diagnosticMessages.generated.json has invalid JSON format"),null}},n.prototype.getCancellationToken=function(){var n=this.shimHost.getCancellationToken();return new e.ThrottledCancellationToken(n)},n.prototype.getCurrentDirectory=function(){return this.shimHost.getCurrentDirectory()},n.prototype.getDirectories=function(e){return JSON.parse(this.shimHost.getDirectories(e))},n.prototype.getDefaultLibFileName=function(e){return this.shimHost.getDefaultLibFileName(JSON.stringify(e))},n.prototype.readDirectory=function(n,t,r,a,i){var o=e.getFileMatcherPatterns(n,r,a,this.shimHost.useCaseSensitiveFileNames(),this.shimHost.getCurrentDirectory());return JSON.parse(this.shimHost.readDirectory(n,JSON.stringify(t),JSON.stringify(o.basePaths),o.excludePattern,o.includeFilePattern,o.includeDirectoryPattern,i))},n.prototype.readFile=function(e,n){return this.shimHost.readFile(e,n)},n.prototype.fileExists=function(e){return this.shimHost.fileExists(e)},n}();e.LanguageServiceShimHostAdapter=r;var a=function(){function n(e){var n=this;this.shimHost=e,this.useCaseSensitiveFileNames=!!this.shimHost.useCaseSensitiveFileNames&&this.shimHost.useCaseSensitiveFileNames(),"directoryExists"in this.shimHost?this.directoryExists=function(e){return n.shimHost.directoryExists(e)}:this.directoryExists=void 0,"realpath"in this.shimHost?this.realpath=function(e){return n.shimHost.realpath(e)}:this.realpath=void 0}return n.prototype.readDirectory=function(n,t,r,a,i){var o=e.getFileMatcherPatterns(n,r,a,this.shimHost.useCaseSensitiveFileNames(),this.shimHost.getCurrentDirectory());return JSON.parse(this.shimHost.readDirectory(n,JSON.stringify(t),JSON.stringify(o.basePaths),o.excludePattern,o.includeFilePattern,o.includeDirectoryPattern,i))},n.prototype.fileExists=function(e){return this.shimHost.fileExists(e)},n.prototype.readFile=function(e){return this.shimHost.readFile(e)},n.prototype.getDirectories=function(e){return JSON.parse(this.shimHost.getDirectories(e))},n}();function o(n,t,r,a){var i;a&&(n.log(t),i=e.timestamp());var o=r();if(a){var s=e.timestamp();if(n.log(t+" completed in "+(s-i)+" msec"),e.isString(o)){var c=o;c.length>128&&(c=c.substring(0,128)+"..."),n.log("  result.length="+c.length+", result='"+JSON.stringify(c)+"'")}}return o}function s(e,n,t,r){return c(e,n,!0,t,r)}function c(t,r,a,i,s){try{var c=o(t,r,i,s);return a?JSON.stringify({result:c}):c}catch(l){return l instanceof e.OperationCanceledException?JSON.stringify({canceled:!0}):(n(t,l),l.description=r,JSON.stringify({error:l}))}}e.CoreServicesShimHostAdapter=a;var u=function(){function e(e){this.factory=e,e.registerShim(this)}return e.prototype.dispose=function(e){this.factory.unregisterShim(this)},e}();function d(e,n){return e.map(function(e){return m(e,n)})}function m(n,t){return{message:e.flattenDiagnosticMessageText(n.messageText,t),start:n.start,length:n.length,category:e.diagnosticCategoryName(n),code:n.code,reportsUnnecessary:n.reportsUnnecessary}}e.realizeDiagnostics=d;var f=function(n){function t(e,t,r){var a=n.call(this,e)||this;return a.host=t,a.languageService=r,a.logPerformance=!1,a.logger=a.host,a}return l(t,n),t.prototype.forwardJSONCall=function(e,n){return s(this.logger,e,n,this.logPerformance)},t.prototype.dispose=function(e){this.logger.log("dispose()"),this.languageService.dispose(),this.languageService=null,p&&p.CollectGarbage&&(p.CollectGarbage(),this.logger.log("CollectGarbage()")),this.logger=null,n.prototype.dispose.call(this,e)},t.prototype.refresh=function(e){this.forwardJSONCall("refresh("+e+")",function(){return null})},t.prototype.cleanupSemanticCache=function(){var e=this;this.forwardJSONCall("cleanupSemanticCache()",function(){return e.languageService.cleanupSemanticCache(),null})},t.prototype.realizeDiagnostics=function(n){var t=e.getNewLineOrDefaultFromHost(this.host);return d(n,t)},t.prototype.getSyntacticClassifications=function(n,t,r){var a=this;return this.forwardJSONCall("getSyntacticClassifications('"+n+"', "+t+", "+r+")",function(){return a.languageService.getSyntacticClassifications(n,e.createTextSpan(t,r))})},t.prototype.getSemanticClassifications=function(n,t,r){var a=this;return this.forwardJSONCall("getSemanticClassifications('"+n+"', "+t+", "+r+")",function(){return a.languageService.getSemanticClassifications(n,e.createTextSpan(t,r))})},t.prototype.getEncodedSyntacticClassifications=function(n,t,r){var a=this;return this.forwardJSONCall("getEncodedSyntacticClassifications('"+n+"', "+t+", "+r+")",function(){return _(a.languageService.getEncodedSyntacticClassifications(n,e.createTextSpan(t,r)))})},t.prototype.getEncodedSemanticClassifications=function(n,t,r){var a=this;return this.forwardJSONCall("getEncodedSemanticClassifications('"+n+"', "+t+", "+r+")",function(){return _(a.languageService.getEncodedSemanticClassifications(n,e.createTextSpan(t,r)))})},t.prototype.getSyntacticDiagnostics=function(e){var n=this;return this.forwardJSONCall("getSyntacticDiagnostics('"+e+"')",function(){var t=n.languageService.getSyntacticDiagnostics(e);return n.realizeDiagnostics(t)})},t.prototype.getSemanticDiagnostics=function(e){var n=this;return this.forwardJSONCall("getSemanticDiagnostics('"+e+"')",function(){var t=n.languageService.getSemanticDiagnostics(e);return n.realizeDiagnostics(t)})},t.prototype.getSuggestionDiagnostics=function(e){var n=this;return this.forwardJSONCall("getSuggestionDiagnostics('"+e+"')",function(){return n.realizeDiagnostics(n.languageService.getSuggestionDiagnostics(e))})},t.prototype.getCompilerOptionsDiagnostics=function(){var e=this;return this.forwardJSONCall("getCompilerOptionsDiagnostics()",function(){var n=e.languageService.getCompilerOptionsDiagnostics();return e.realizeDiagnostics(n)})},t.prototype.getQuickInfoAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getQuickInfoAtPosition('"+e+"', "+n+")",function(){return t.languageService.getQuickInfoAtPosition(e,n)})},t.prototype.getNameOrDottedNameSpan=function(e,n,t){var r=this;return this.forwardJSONCall("getNameOrDottedNameSpan('"+e+"', "+n+", "+t+")",function(){return r.languageService.getNameOrDottedNameSpan(e,n,t)})},t.prototype.getBreakpointStatementAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getBreakpointStatementAtPosition('"+e+"', "+n+")",function(){return t.languageService.getBreakpointStatementAtPosition(e,n)})},t.prototype.getSignatureHelpItems=function(e,n,t){var r=this;return this.forwardJSONCall("getSignatureHelpItems('"+e+"', "+n+")",function(){return r.languageService.getSignatureHelpItems(e,n,t)})},t.prototype.getDefinitionAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getDefinitionAtPosition('"+e+"', "+n+")",function(){return t.languageService.getDefinitionAtPosition(e,n)})},t.prototype.getDefinitionAndBoundSpan=function(e,n){var t=this;return this.forwardJSONCall("getDefinitionAndBoundSpan('"+e+"', "+n+")",function(){return t.languageService.getDefinitionAndBoundSpan(e,n)})},t.prototype.getTypeDefinitionAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getTypeDefinitionAtPosition('"+e+"', "+n+")",function(){return t.languageService.getTypeDefinitionAtPosition(e,n)})},t.prototype.getImplementationAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getImplementationAtPosition('"+e+"', "+n+")",function(){return t.languageService.getImplementationAtPosition(e,n)})},t.prototype.getRenameInfo=function(e,n,t){var r=this;return this.forwardJSONCall("getRenameInfo('"+e+"', "+n+")",function(){return r.languageService.getRenameInfo(e,n,t)})},t.prototype.findRenameLocations=function(e,n,t,r,a){var i=this;return this.forwardJSONCall("findRenameLocations('"+e+"', "+n+", "+t+", "+r+", "+a+")",function(){return i.languageService.findRenameLocations(e,n,t,r,a)})},t.prototype.getBraceMatchingAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getBraceMatchingAtPosition('"+e+"', "+n+")",function(){return t.languageService.getBraceMatchingAtPosition(e,n)})},t.prototype.isValidBraceCompletionAtPosition=function(e,n,t){var r=this;return this.forwardJSONCall("isValidBraceCompletionAtPosition('"+e+"', "+n+", "+t+")",function(){return r.languageService.isValidBraceCompletionAtPosition(e,n,t)})},t.prototype.getSpanOfEnclosingComment=function(e,n,t){var r=this;return this.forwardJSONCall("getSpanOfEnclosingComment('"+e+"', "+n+")",function(){return r.languageService.getSpanOfEnclosingComment(e,n,t)})},t.prototype.getIndentationAtPosition=function(e,n,t){var r=this;return this.forwardJSONCall("getIndentationAtPosition('"+e+"', "+n+")",function(){var a=JSON.parse(t);return r.languageService.getIndentationAtPosition(e,n,a)})},t.prototype.getReferencesAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getReferencesAtPosition('"+e+"', "+n+")",function(){return t.languageService.getReferencesAtPosition(e,n)})},t.prototype.findReferences=function(e,n){var t=this;return this.forwardJSONCall("findReferences('"+e+"', "+n+")",function(){return t.languageService.findReferences(e,n)})},t.prototype.getOccurrencesAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getOccurrencesAtPosition('"+e+"', "+n+")",function(){return t.languageService.getOccurrencesAtPosition(e,n)})},t.prototype.getDocumentHighlights=function(n,t,r){var a=this;return this.forwardJSONCall("getDocumentHighlights('"+n+"', "+t+")",function(){var i=a.languageService.getDocumentHighlights(n,t,JSON.parse(r)),o=e.normalizeSlashes(n).toLowerCase();return e.filter(i,function(n){return e.normalizeSlashes(n.fileName).toLowerCase()===o})})},t.prototype.getCompletionsAtPosition=function(e,n,t){var r=this;return this.forwardJSONCall("getCompletionsAtPosition('"+e+"', "+n+", "+t+")",function(){return r.languageService.getCompletionsAtPosition(e,n,t)})},t.prototype.getCompletionEntryDetails=function(e,n,t,r,a,i){var o=this;return this.forwardJSONCall("getCompletionEntryDetails('"+e+"', "+n+", '"+t+"')",function(){var s=void 0===r?void 0:JSON.parse(r);return o.languageService.getCompletionEntryDetails(e,n,t,s,a,i)})},t.prototype.getFormattingEditsForRange=function(e,n,t,r){var a=this;return this.forwardJSONCall("getFormattingEditsForRange('"+e+"', "+n+", "+t+")",function(){var i=JSON.parse(r);return a.languageService.getFormattingEditsForRange(e,n,t,i)})},t.prototype.getFormattingEditsForDocument=function(e,n){var t=this;return this.forwardJSONCall("getFormattingEditsForDocument('"+e+"')",function(){var r=JSON.parse(n);return t.languageService.getFormattingEditsForDocument(e,r)})},t.prototype.getFormattingEditsAfterKeystroke=function(e,n,t,r){var a=this;return this.forwardJSONCall("getFormattingEditsAfterKeystroke('"+e+"', "+n+", '"+t+"')",function(){var i=JSON.parse(r);return a.languageService.getFormattingEditsAfterKeystroke(e,n,t,i)})},t.prototype.getDocCommentTemplateAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getDocCommentTemplateAtPosition('"+e+"', "+n+")",function(){return t.languageService.getDocCommentTemplateAtPosition(e,n)})},t.prototype.getNavigateToItems=function(e,n,t){var r=this;return this.forwardJSONCall("getNavigateToItems('"+e+"', "+n+", "+t+")",function(){return r.languageService.getNavigateToItems(e,n,t)})},t.prototype.getNavigationBarItems=function(e){var n=this;return this.forwardJSONCall("getNavigationBarItems('"+e+"')",function(){return n.languageService.getNavigationBarItems(e)})},t.prototype.getNavigationTree=function(e){var n=this;return this.forwardJSONCall("getNavigationTree('"+e+"')",function(){return n.languageService.getNavigationTree(e)})},t.prototype.getOutliningSpans=function(e){var n=this;return this.forwardJSONCall("getOutliningSpans('"+e+"')",function(){return n.languageService.getOutliningSpans(e)})},t.prototype.getTodoComments=function(e,n){var t=this;return this.forwardJSONCall("getTodoComments('"+e+"')",function(){return t.languageService.getTodoComments(e,JSON.parse(n))})},t.prototype.getEmitOutput=function(e){var n=this;return this.forwardJSONCall("getEmitOutput('"+e+"')",function(){return n.languageService.getEmitOutput(e)})},t.prototype.getEmitOutputObject=function(e){var n=this;return c(this.logger,"getEmitOutput('"+e+"')",!1,function(){return n.languageService.getEmitOutput(e)},this.logPerformance)},t}(u);function _(e){return{spans:e.spans.join(","),endOfLineState:e.endOfLineState}}var g=function(n){function t(t,r){var a=n.call(this,t)||this;return a.logger=r,a.logPerformance=!1,a.classifier=e.createClassifier(),a}return l(t,n),t.prototype.getEncodedLexicalClassifications=function(e,n,t){var r=this;return void 0===t&&(t=!1),s(this.logger,"getEncodedLexicalClassifications",function(){return _(r.classifier.getEncodedLexicalClassifications(e,n,t))},this.logPerformance)},t.prototype.getClassificationsForLine=function(e,n,t){void 0===t&&(t=!1);for(var r=this.classifier.getClassificationsForLine(e,n,t),a="",i=0,o=r.entries;i<o.length;i++){var s=o[i];a+=s.length+"\n",a+=s.classification+"\n"}return a+=r.finalLexState,a},t}(u),v=function(n){function t(e,t,r){var a=n.call(this,e)||this;return a.logger=t,a.host=r,a.logPerformance=!1,a}return l(t,n),t.prototype.forwardJSONCall=function(e,n){return s(this.logger,e,n,this.logPerformance)},t.prototype.resolveModuleName=function(n,t,r){var a=this;return this.forwardJSONCall("resolveModuleName('"+n+"')",function(){var i=JSON.parse(r),o=e.resolveModuleName(t,e.normalizeSlashes(n),i,a.host),s=o.resolvedModule?o.resolvedModule.resolvedFileName:void 0;return o.resolvedModule&&".ts"!==o.resolvedModule.extension&&".tsx"!==o.resolvedModule.extension&&".d.ts"!==o.resolvedModule.extension&&(s=void 0),{resolvedFileName:s,failedLookupLocations:o.failedLookupLocations}})},t.prototype.resolveTypeReferenceDirective=function(n,t,r){var a=this;return this.forwardJSONCall("resolveTypeReferenceDirective("+n+")",function(){var i=JSON.parse(r),o=e.resolveTypeReferenceDirective(t,e.normalizeSlashes(n),i,a.host);return{resolvedFileName:o.resolvedTypeReferenceDirective?o.resolvedTypeReferenceDirective.resolvedFileName:void 0,primary:!o.resolvedTypeReferenceDirective||o.resolvedTypeReferenceDirective.primary,failedLookupLocations:o.failedLookupLocations}})},t.prototype.getPreProcessedFileInfo=function(n,t){var r=this;return this.forwardJSONCall("getPreProcessedFileInfo('"+n+"')",function(){var n=e.preProcessFile(e.getSnapshotText(t),!0,!0);return{referencedFiles:r.convertFileReferences(n.referencedFiles),importedFiles:r.convertFileReferences(n.importedFiles),ambientExternalModules:n.ambientExternalModules,isLibFile:n.isLibFile,typeReferenceDirectives:r.convertFileReferences(n.typeReferenceDirectives),libReferenceDirectives:r.convertFileReferences(n.libReferenceDirectives)}})},t.prototype.getAutomaticTypeDirectiveNames=function(n){var t=this;return this.forwardJSONCall("getAutomaticTypeDirectiveNames('"+n+"')",function(){var r=JSON.parse(n);return e.getAutomaticTypeDirectiveNames(r,t.host)})},t.prototype.convertFileReferences=function(n){if(n){for(var t=[],r=0,a=n;r<a.length;r++){var i=a[r];t.push({path:e.normalizeSlashes(i.fileName),position:i.pos,length:i.end-i.pos})}return t}},t.prototype.getTSConfigFileInfo=function(n,t){var r=this;return this.forwardJSONCall("getTSConfigFileInfo('"+n+"')",function(){var a=e.parseJsonText(n,e.getSnapshotText(t)),i=e.normalizeSlashes(n),o=e.parseJsonSourceFileConfigFileContent(a,r.host,e.getDirectoryPath(i),{},i);return{options:o.options,typeAcquisition:o.typeAcquisition,files:o.fileNames,raw:o.raw,errors:d(a.parseDiagnostics.concat(o.errors),"\r\n")}})},t.prototype.getDefaultCompilationSettings=function(){return this.forwardJSONCall("getDefaultCompilationSettings()",function(){return e.getDefaultCompilerOptions()})},t.prototype.discoverTypings=function(n){var t=this,r=e.createGetCanonicalFileName(!1);return this.forwardJSONCall("discoverTypings()",function(){var a=JSON.parse(n);return void 0===t.safeList&&(t.safeList=e.JsTyping.loadSafeList(t.host,e.toPath(a.safeListPath,a.safeListPath,r))),e.JsTyping.discoverTypings(t.host,function(e){return t.logger.log(e)},a.fileNames,e.toPath(a.projectRootPath,a.projectRootPath,r),t.safeList,a.packageNameToTypingLocation,a.typeAcquisition,a.unresolvedImports,a.typesRegistry)})},t}(u),y=function(){function t(){this._shims=[]}return t.prototype.getServicesVersion=function(){return e.servicesVersion},t.prototype.createLanguageServiceShim=function(t){try{void 0===this.documentRegistry&&(this.documentRegistry=e.createDocumentRegistry(t.useCaseSensitiveFileNames&&t.useCaseSensitiveFileNames(),t.getCurrentDirectory()));var a=new r(t),i=e.createLanguageService(a,this.documentRegistry,!1);return new f(this,t,i)}catch(o){throw n(t,o),o}},t.prototype.createClassifierShim=function(e){try{return new g(this,e)}catch(t){throw n(e,t),t}},t.prototype.createCoreServicesShim=function(e){try{var t=new a(e);return new v(this,e,t)}catch(r){throw n(e,r),r}},t.prototype.close=function(){e.clear(this._shims),this.documentRegistry=void 0},t.prototype.registerShim=function(e){this._shims.push(e)},t.prototype.unregisterShim=function(e){for(var n=0;n<this._shims.length;n++)if(this._shims[n]===e)return void delete this._shims[n];throw new Error("Invalid operation")},t}();e.TypeScriptServicesFactory=y,i.exports&&(i.exports=e)})(u||(u={})),function(e){(function(e){e.TypeScriptServicesFactory=u.TypeScriptServicesFactory})(e.Services||(e.Services={}))}(d||(d={}));u.versionMajorMinor;u.createClassifier;const m=u.createLanguageService,f=(u.displayPartsToString,u.EndOfLineState,u.flattenDiagnosticMessageText,u.IndentStyle,u.ScriptKind),_=u.ScriptTarget;u.TokenClass}).call(this,t("f28c"),t("c8ba"),"/",t("dd40")(e))},c8ba:function(e,n){var t;t=function(){return this}();try{t=t||new Function("return this")()}catch(r){"object"===typeof window&&(t=window)}e.exports=t},d0c7:function(e,n,t){"use strict";t.r(n);var r=function(){function e(){this.listeners=[],this.unexpectedErrorHandler=function(e){setTimeout(function(){if(e.stack)throw new Error(e.message+"\n\n"+e.stack);throw e},0)}}return e.prototype.emit=function(e){this.listeners.forEach(function(n){n(e)})},e.prototype.onUnexpectedError=function(e){this.unexpectedErrorHandler(e),this.emit(e)},e.prototype.onUnexpectedExternalError=function(e){this.unexpectedErrorHandler(e)},e}(),a=new r;function i(e){c(e)||a.onUnexpectedError(e)}function o(e){if(e instanceof Error){var n=e.name,t=e.message,r=e.stacktrace||e.stack;return{$isError:!0,name:n,message:t,stack:r}}return e}var s="Canceled";function c(e){return e instanceof Error&&e.name===s&&e.message===s}function l(e){return e?new Error("Illegal argument: "+e):new Error("Illegal argument")}function u(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];return Array.isArray(e)?(e.forEach(function(e){return e&&e.dispose()}),[]):0===n.length?e?(e.dispose(),e):void 0:(u(e),u(n),[])}function d(e){return{dispose:function(){return u(e)}}}function p(e){return{dispose:function(){e()}}}var m=function(){function e(){this._toDispose=[],this._lifecycle_disposable_isDisposed=!1}return e.prototype.dispose=function(){this._lifecycle_disposable_isDisposed=!0,this._toDispose=u(this._toDispose)},e.prototype._register=function(e){return this._lifecycle_disposable_isDisposed?(console.warn("Registering disposable on object that has already been disposed."),e.dispose()):this._toDispose.push(e),e},e.None=Object.freeze({dispose:function(){}}),e}(),f=(function(){function e(e){this.object=e}e.prototype.dispose=function(){}}(),t("30db"));Object.prototype.hasOwnProperty;function _(e){var n=[],t=Object.getPrototypeOf(e);while(Object.prototype!==t)n=n.concat(Object.getOwnPropertyNames(t)),t=Object.getPrototypeOf(t);return n}var g=function(){var e=function(n,t){return e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,n){e.__proto__=n}||function(e,n){for(var t in n)n.hasOwnProperty(t)&&(e[t]=n[t])},e(n,t)};return function(n,t){function r(){this.constructor=n}e(n,t),n.prototype=null===t?Object.create(t):(r.prototype=t.prototype,new r)}}(),v="$initialize";var y=function(){function e(e){this._workerId=-1,this._handler=e,this._lastSentReq=0,this._pendingReplies=Object.create(null)}return e.prototype.setWorkerId=function(e){this._workerId=e},e.prototype.sendMessage=function(e,n){var t=this,r=String(++this._lastSentReq);return new Promise(function(a,i){t._pendingReplies[r]={resolve:a,reject:i},t._send({vsWorker:t._workerId,req:r,method:e,args:n})})},e.prototype.handleMessage=function(e){var n;try{n=JSON.parse(e)}catch(t){return}n&&n.vsWorker&&(-1!==this._workerId&&n.vsWorker!==this._workerId||this._handleMessage(n))},e.prototype._handleMessage=function(e){var n=this;if(e.seq){var t=e;if(!this._pendingReplies[t.seq])return void console.warn("Got reply to unknown seq");var r=this._pendingReplies[t.seq];if(delete this._pendingReplies[t.seq],t.err){var a=t.err;return t.err.$isError&&(a=new Error,a.name=t.err.name,a.message=t.err.message,a.stack=t.err.stack),void r.reject(a)}r.resolve(t.res)}else{var i=e,s=i.req,c=this._handler.handleMessage(i.method,i.args);c.then(function(e){n._send({vsWorker:n._workerId,seq:s,res:e,err:void 0})},function(e){e.detail instanceof Error&&(e.detail=o(e.detail)),n._send({vsWorker:n._workerId,seq:s,res:void 0,err:o(e)})})}},e.prototype._send=function(e){var n=JSON.stringify(e);this._handler.sendMessage(n)},e}(),h=(function(e){function n(n,t){var r=e.call(this)||this,a=null;r._worker=r._register(n.create("vs/base/common/worker/simpleWorker",function(e){r._protocol.handleMessage(e)},function(e){a&&a(e)})),r._protocol=new y({sendMessage:function(e){r._worker.postMessage(e)},handleMessage:function(e,n){return Promise.resolve(null)}}),r._protocol.setWorkerId(r._worker.getId());var i=null;"undefined"!==typeof self.require&&"function"===typeof self.require.getConfig?i=self.require.getConfig():"undefined"!==typeof self.requirejs&&(i=self.requirejs.s.contexts._.config),r._onModuleLoaded=r._protocol.sendMessage(v,[r._worker.getId(),t,i]),r._lazyProxy=new Promise(function(e,n){a=n,r._onModuleLoaded.then(function(n){for(var t={},r=0,a=n;r<a.length;r++){var i=a[r];t[i]=s(i,o)}e(t)},function(e){n(e),r._onError("Worker failed to load "+t,e)})});var o=function(e,n){return r._request(e,n)},s=function(e,n){return function(){var t=Array.prototype.slice.call(arguments,0);return n(e,t)}};return r}g(n,e),n.prototype.getProxyObject=function(){return this._lazyProxy},n.prototype._request=function(e,n){var t=this;return new Promise(function(r,a){t._onModuleLoaded.then(function(){t._protocol.sendMessage(e,n).then(r,a)},a)})},n.prototype._onError=function(e,n){console.error(e),console.info(n)}}(m),function(){function e(e,n){var t=this;this._requestHandler=n,this._protocol=new y({sendMessage:function(n){e(n)},handleMessage:function(e,n){return t._handleMessage(e,n)}})}return e.prototype.onmessage=function(e){this._protocol.handleMessage(e)},e.prototype._handleMessage=function(e,n){if(e===v)return this.initialize(n[0],n[1],n[2]);if(!this._requestHandler||"function"!==typeof this._requestHandler[e])return Promise.reject(new Error("Missing requestHandler or method: "+e));try{return Promise.resolve(this._requestHandler[e].apply(this._requestHandler,n))}catch(t){return Promise.reject(t)}},e.prototype.initialize=function(e,n,t){var r=this;if(this._protocol.setWorkerId(e),this._requestHandler){for(var a=[],i=0,o=_(this._requestHandler);i<o.length;i++){var s=o[i];"function"===typeof this._requestHandler[s]&&a.push(s)}return Promise.resolve(a)}return t&&("undefined"!==typeof t.baseUrl&&delete t["baseUrl"],"undefined"!==typeof t.paths&&"undefined"!==typeof t.paths.vs&&delete t.paths["vs"],t.catchError=!0,self.require.config(t)),new Promise(function(e,t){self.require([n],function(){for(var n=[],a=0;a<arguments.length;a++)n[a]=arguments[a];var i=n[0];if(r._requestHandler=i.create(),r._requestHandler){for(var o=[],s=0,c=_(r._requestHandler);s<c.length;s++){var l=c[s];"function"===typeof r._requestHandler[l]&&o.push(l)}e(o)}else t(new Error("No RequestHandler!"))},t)})},e}());function b(e,n){return T(e,n,0,e.length-1,[]),e}function E(e,n,t,r,a,i){for(var o=t,s=r+1,c=t;c<=a;c++)i[c]=e[c];for(c=t;c<=a;c++)o>r?e[c]=i[s++]:s>a?e[c]=i[o++]:n(i[s],i[o])<0?e[c]=i[s++]:e[c]=i[o++]}function T(e,n,t,r,a){if(!(r<=t)){var i=t+(r-t)/2|0;T(e,n,t,i,a),T(e,n,i+1,r,a),n(e[i],e[i+1])<=0||E(e,n,t,i,r,a)}}var S=function(){function e(e,n,t,r){this.originalStart=e,this.originalLength=n,this.modifiedStart=t,this.modifiedLength=r}return e.prototype.getOriginalEnd=function(){return this.originalStart+this.originalLength},e.prototype.getModifiedEnd=function(){return this.modifiedStart+this.modifiedLength},e}();function x(e){return{getLength:function(){return e.length},getElementAtIndex:function(n){return e.charCodeAt(n)}}}function C(e,n,t){return new I(x(e),x(n)).ComputeDiff(t)}var A,D=function(){function e(){}return e.Assert=function(e,n){if(!e)throw new Error(n)},e}(),L=function(){function e(){}return e.Copy=function(e,n,t,r,a){for(var i=0;i<a;i++)t[r+i]=e[n+i]},e}(),k=1447,N=function(){function e(){this.m_changes=[],this.m_originalStart=Number.MAX_VALUE,this.m_modifiedStart=Number.MAX_VALUE,this.m_originalCount=0,this.m_modifiedCount=0}return e.prototype.MarkNextChange=function(){(this.m_originalCount>0||this.m_modifiedCount>0)&&this.m_changes.push(new S(this.m_originalStart,this.m_originalCount,this.m_modifiedStart,this.m_modifiedCount)),this.m_originalCount=0,this.m_modifiedCount=0,this.m_originalStart=Number.MAX_VALUE,this.m_modifiedStart=Number.MAX_VALUE},e.prototype.AddOriginalElement=function(e,n){this.m_originalStart=Math.min(this.m_originalStart,e),this.m_modifiedStart=Math.min(this.m_modifiedStart,n),this.m_originalCount++},e.prototype.AddModifiedElement=function(e,n){this.m_originalStart=Math.min(this.m_originalStart,e),this.m_modifiedStart=Math.min(this.m_modifiedStart,n),this.m_modifiedCount++},e.prototype.getChanges=function(){return(this.m_originalCount>0||this.m_modifiedCount>0)&&this.MarkNextChange(),this.m_changes},e.prototype.getReverseChanges=function(){return(this.m_originalCount>0||this.m_modifiedCount>0)&&this.MarkNextChange(),this.m_changes.reverse(),this.m_changes},e}(),I=function(){function e(e,n,t){void 0===t&&(t=null),this.OriginalSequence=e,this.ModifiedSequence=n,this.ContinueProcessingPredicate=t,this.m_forwardHistory=[],this.m_reverseHistory=[]}return e.prototype.ElementsAreEqual=function(e,n){return this.OriginalSequence.getElementAtIndex(e)===this.ModifiedSequence.getElementAtIndex(n)},e.prototype.OriginalElementsAreEqual=function(e,n){return this.OriginalSequence.getElementAtIndex(e)===this.OriginalSequence.getElementAtIndex(n)},e.prototype.ModifiedElementsAreEqual=function(e,n){return this.ModifiedSequence.getElementAtIndex(e)===this.ModifiedSequence.getElementAtIndex(n)},e.prototype.ComputeDiff=function(e){return this._ComputeDiff(0,this.OriginalSequence.getLength()-1,0,this.ModifiedSequence.getLength()-1,e)},e.prototype._ComputeDiff=function(e,n,t,r,a){var i=[!1],o=this.ComputeDiffRecursive(e,n,t,r,i);return a?this.PrettifyChanges(o):o},e.prototype.ComputeDiffRecursive=function(e,n,t,r,a){a[0]=!1;while(e<=n&&t<=r&&this.ElementsAreEqual(e,t))e++,t++;while(n>=e&&r>=t&&this.ElementsAreEqual(n,r))n--,r--;if(e>n||t>r){var i=void 0;return t<=r?(D.Assert(e===n+1,"originalStart should only be one more than originalEnd"),i=[new S(e,0,t,r-t+1)]):e<=n?(D.Assert(t===r+1,"modifiedStart should only be one more than modifiedEnd"),i=[new S(e,n-e+1,t,0)]):(D.Assert(e===n+1,"originalStart should only be one more than originalEnd"),D.Assert(t===r+1,"modifiedStart should only be one more than modifiedEnd"),i=[]),i}var o=[0],s=[0],c=this.ComputeRecursionPoint(e,n,t,r,o,s,a),l=o[0],u=s[0];if(null!==c)return c;if(!a[0]){var d=this.ComputeDiffRecursive(e,l,t,u,a),p=[];return p=a[0]?[new S(l+1,n-(l+1)+1,u+1,r-(u+1)+1)]:this.ComputeDiffRecursive(l+1,n,u+1,r,a),this.ConcatenateChanges(d,p)}return[new S(e,n-e+1,t,r-t+1)]},e.prototype.WALKTRACE=function(e,n,t,r,a,i,o,s,c,l,u,d,p,m,f,_,g,v){var y,h=null,b=null,E=new N,T=n,x=t,C=p[0]-_[0]-r,A=Number.MIN_VALUE,D=this.m_forwardHistory.length-1;do{y=C+e,y===T||y<x&&c[y-1]<c[y+1]?(u=c[y+1],m=u-C-r,u<A&&E.MarkNextChange(),A=u,E.AddModifiedElement(u+1,m),C=y+1-e):(u=c[y-1]+1,m=u-C-r,u<A&&E.MarkNextChange(),A=u-1,E.AddOriginalElement(u,m+1),C=y-1-e),D>=0&&(c=this.m_forwardHistory[D],e=c[0],T=1,x=c.length-1)}while(--D>=-1);if(h=E.getReverseChanges(),v[0]){var L=p[0]+1,k=_[0]+1;if(null!==h&&h.length>0){var I=h[h.length-1];L=Math.max(L,I.getOriginalEnd()),k=Math.max(k,I.getModifiedEnd())}b=[new S(L,d-L+1,k,f-k+1)]}else{E=new N,T=i,x=o,C=p[0]-_[0]-s,A=Number.MAX_VALUE,D=g?this.m_reverseHistory.length-1:this.m_reverseHistory.length-2;do{y=C+a,y===T||y<x&&l[y-1]>=l[y+1]?(u=l[y+1]-1,m=u-C-s,u>A&&E.MarkNextChange(),A=u+1,E.AddOriginalElement(u+1,m+1),C=y+1-a):(u=l[y-1],m=u-C-s,u>A&&E.MarkNextChange(),A=u,E.AddModifiedElement(u+1,m+1),C=y-1-a),D>=0&&(l=this.m_reverseHistory[D],a=l[0],T=1,x=l.length-1)}while(--D>=-1);b=E.getChanges()}return this.ConcatenateChanges(h,b)},e.prototype.ComputeRecursionPoint=function(e,n,t,r,a,i,o){var s,c=0,l=0,u=0,d=0,p=0,m=0;e--,t--,a[0]=0,i[0]=0,this.m_forwardHistory=[],this.m_reverseHistory=[];var f,_,g=n-e+(r-t),v=g+1,y=new Array(v),h=new Array(v),b=r-t,E=n-e,T=e-t,x=n-r,C=E-b,A=C%2===0;for(y[b]=e,h[E]=n,o[0]=!1,s=1;s<=g/2+1;s++){var D=0,N=0;for(u=this.ClipDiagonalBound(b-s,s,b,v),d=this.ClipDiagonalBound(b+s,s,b,v),f=u;f<=d;f+=2){c=f===u||f<d&&y[f-1]<y[f+1]?y[f+1]:y[f-1]+1,l=c-(f-b)-T,_=c;while(c<n&&l<r&&this.ElementsAreEqual(c+1,l+1))c++,l++;if(y[f]=c,c+l>D+N&&(D=c,N=l),!A&&Math.abs(f-E)<=s-1&&c>=h[f])return a[0]=c,i[0]=l,_<=h[f]&&k>0&&s<=k+1?this.WALKTRACE(b,u,d,T,E,p,m,x,y,h,c,n,a,l,r,i,A,o):null}var I=(D-e+(N-t)-s)/2;if(null!==this.ContinueProcessingPredicate&&!this.ContinueProcessingPredicate(D,this.OriginalSequence,I))return o[0]=!0,a[0]=D,i[0]=N,I>0&&k>0&&s<=k+1?this.WALKTRACE(b,u,d,T,E,p,m,x,y,h,c,n,a,l,r,i,A,o):(e++,t++,[new S(e,n-e+1,t,r-t+1)]);for(p=this.ClipDiagonalBound(E-s,s,E,v),m=this.ClipDiagonalBound(E+s,s,E,v),f=p;f<=m;f+=2){c=f===p||f<m&&h[f-1]>=h[f+1]?h[f+1]-1:h[f-1],l=c-(f-E)-x,_=c;while(c>e&&l>t&&this.ElementsAreEqual(c,l))c--,l--;if(h[f]=c,A&&Math.abs(f-b)<=s&&c<=y[f])return a[0]=c,i[0]=l,_>=y[f]&&k>0&&s<=k+1?this.WALKTRACE(b,u,d,T,E,p,m,x,y,h,c,n,a,l,r,i,A,o):null}if(s<=k){var w=new Array(d-u+2);w[0]=b-u+1,L.Copy(y,u,w,1,d-u+1),this.m_forwardHistory.push(w),w=new Array(m-p+2),w[0]=E-p+1,L.Copy(h,p,w,1,m-p+1),this.m_reverseHistory.push(w)}}return this.WALKTRACE(b,u,d,T,E,p,m,x,y,h,c,n,a,l,r,i,A,o)},e.prototype.PrettifyChanges=function(e){for(var n=0;n<e.length;n++){var t=e[n],r=n<e.length-1?e[n+1].originalStart:this.OriginalSequence.getLength(),a=n<e.length-1?e[n+1].modifiedStart:this.ModifiedSequence.getLength(),i=t.originalLength>0,o=t.modifiedLength>0;while(t.originalStart+t.originalLength<r&&t.modifiedStart+t.modifiedLength<a&&(!i||this.OriginalElementsAreEqual(t.originalStart,t.originalStart+t.originalLength))&&(!o||this.ModifiedElementsAreEqual(t.modifiedStart,t.modifiedStart+t.modifiedLength)))t.originalStart++,t.modifiedStart++;var s=[null];n<e.length-1&&this.ChangesOverlap(e[n],e[n+1],s)&&(e[n]=s[0],e.splice(n+1,1),n--)}for(n=e.length-1;n>=0;n--){t=e[n],r=0,a=0;if(n>0){var c=e[n-1];c.originalLength>0&&(r=c.originalStart+c.originalLength),c.modifiedLength>0&&(a=c.modifiedStart+c.modifiedLength)}i=t.originalLength>0,o=t.modifiedLength>0;for(var l=0,u=this._boundaryScore(t.originalStart,t.originalLength,t.modifiedStart,t.modifiedLength),d=1;;d++){var p=t.originalStart-d,m=t.modifiedStart-d;if(p<r||m<a)break;if(i&&!this.OriginalElementsAreEqual(p,p+t.originalLength))break;if(o&&!this.ModifiedElementsAreEqual(m,m+t.modifiedLength))break;var f=this._boundaryScore(p,t.originalLength,m,t.modifiedLength);f>u&&(u=f,l=d)}t.originalStart-=l,t.modifiedStart-=l}return e},e.prototype._OriginalIsBoundary=function(e){if(e<=0||e>=this.OriginalSequence.getLength()-1)return!0;var n=this.OriginalSequence.getElementAtIndex(e);return"string"===typeof n&&/^\s*$/.test(n)},e.prototype._OriginalRegionIsBoundary=function(e,n){if(this._OriginalIsBoundary(e)||this._OriginalIsBoundary(e-1))return!0;if(n>0){var t=e+n;if(this._OriginalIsBoundary(t-1)||this._OriginalIsBoundary(t))return!0}return!1},e.prototype._ModifiedIsBoundary=function(e){if(e<=0||e>=this.ModifiedSequence.getLength()-1)return!0;var n=this.ModifiedSequence.getElementAtIndex(e);return"string"===typeof n&&/^\s*$/.test(n)},e.prototype._ModifiedRegionIsBoundary=function(e,n){if(this._ModifiedIsBoundary(e)||this._ModifiedIsBoundary(e-1))return!0;if(n>0){var t=e+n;if(this._ModifiedIsBoundary(t-1)||this._ModifiedIsBoundary(t))return!0}return!1},e.prototype._boundaryScore=function(e,n,t,r){var a=this._OriginalRegionIsBoundary(e,n)?1:0,i=this._ModifiedRegionIsBoundary(t,r)?1:0;return a+i},e.prototype.ConcatenateChanges=function(e,n){var t=[];if(0===e.length||0===n.length)return n.length>0?n:e;if(this.ChangesOverlap(e[e.length-1],n[0],t)){var r=new Array(e.length+n.length-1);return L.Copy(e,0,r,0,e.length-1),r[e.length-1]=t[0],L.Copy(n,1,r,e.length,n.length-1),r}r=new Array(e.length+n.length);return L.Copy(e,0,r,0,e.length),L.Copy(n,0,r,e.length,n.length),r},e.prototype.ChangesOverlap=function(e,n,t){if(D.Assert(e.originalStart<=n.originalStart,"Left change is not less than or equal to right change"),D.Assert(e.modifiedStart<=n.modifiedStart,"Left change is not less than or equal to right change"),e.originalStart+e.originalLength>=n.originalStart||e.modifiedStart+e.modifiedLength>=n.modifiedStart){var r=e.originalStart,a=e.originalLength,i=e.modifiedStart,o=e.modifiedLength;return e.originalStart+e.originalLength>=n.originalStart&&(a=n.originalStart+n.originalLength-e.originalStart),e.modifiedStart+e.modifiedLength>=n.modifiedStart&&(o=n.modifiedStart+n.modifiedLength-e.modifiedStart),t[0]=new S(r,a,i,o),!0}return t[0]=null,!1},e.prototype.ClipDiagonalBound=function(e,n,t,r){if(e>=0&&e<r)return e;var a=t,i=r-t-1,o=n%2===0;if(e<0){var s=a%2===0;return o===s?0:1}var c=i%2===0;return o===c?r-1:r-2},e}(),w=function(){var e=function(n,t){return e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,n){e.__proto__=n}||function(e,n){for(var t in n)n.hasOwnProperty(t)&&(e[t]=n[t])},e(n,t)};return function(n,t){function r(){this.constructor=n}e(n,t),n.prototype=null===t?Object.create(t):(r.prototype=t.prototype,new r)}}(),M={done:!0,value:void 0};(function(e){var n={next:function(){return M}};function t(){return n}function r(e,n,t){return void 0===n&&(n=0),void 0===t&&(t=e.length),{next:function(){return n>=t?M:{done:!1,value:e[n++]}}}}function a(n){return n?Array.isArray(n)?e.fromArray(n):n:e.empty()}function i(e,n){return{next:function(){var t=e.next();return t.done?M:{done:!1,value:n(t.value)}}}}function o(e,n){return{next:function(){while(1){var t=e.next();if(t.done)return M;if(n(t.value))return{done:!1,value:t.value}}}}}function s(e,n){for(var t=e.next();!t.done;t=e.next())n(t.value)}function c(e){var n=[];return s(e,function(e){return n.push(e)}),n}e.empty=t,e.fromArray=r,e.from=a,e.map=i,e.filter=o,e.forEach=s,e.collect=c})(A||(A={}));var O,P=function(){function e(e,n,t,r){void 0===n&&(n=0),void 0===t&&(t=e.length),void 0===r&&(r=n-1),this.items=e,this.start=n,this.end=t,this.index=r}return e.prototype.next=function(){return this.index=Math.min(this.index+1,this.end),this.current()},e.prototype.current=function(){return this.index===this.start-1||this.index===this.end?null:this.items[this.index]},e}(),R=(function(e){function n(n,t,r,a){return void 0===t&&(t=0),void 0===r&&(r=n.length),void 0===a&&(a=t-1),e.call(this,n,t,r,a)||this}w(n,e),n.prototype.current=function(){return e.prototype.current.call(this)},n.prototype.previous=function(){return this.index=Math.max(this.index-1,this.start-1),this.current()},n.prototype.first=function(){return this.index=this.start,this.current()},n.prototype.last=function(){return this.index=this.end-1,this.current()},n.prototype.parent=function(){return null}}(P),function(){function e(e,n){this.iterator=e,this.fn=n}e.prototype.next=function(){return this.fn(this.iterator.next())}}(),function(){var e=function(n,t){return e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,n){e.__proto__=n}||function(e,n){for(var t in n)n.hasOwnProperty(t)&&(e[t]=n[t])},e(n,t)};return function(n,t){function r(){this.constructor=n}e(n,t),n.prototype=null===t?Object.create(t):(r.prototype=t.prototype,new r)}}()),F=/^\w[\w\d+.-]*$/,G=/^\//,V=/^\/\//,B=!0;function K(e,n){if(!e.scheme){if(n||B)throw new Error('[UriError]: Scheme is missing: {scheme: "", authority: "'+e.authority+'", path: "'+e.path+'", query: "'+e.query+'", fragment: "'+e.fragment+'"}');console.warn('[UriError]: Scheme is missing: {scheme: "", authority: "'+e.authority+'", path: "'+e.path+'", query: "'+e.query+'", fragment: "'+e.fragment+'"}')}if(e.scheme&&!F.test(e.scheme))throw new Error("[UriError]: Scheme contains illegal characters.");if(e.path)if(e.authority){if(!G.test(e.path))throw new Error('[UriError]: If a URI contains an authority component, then the path component must either be empty or begin with a slash ("/") character')}else if(V.test(e.path))throw new Error('[UriError]: If a URI does not contain an authority component, then the path cannot begin with two slash characters ("//")')}function j(e,n){return n||B?e||H:(e||(console.trace("BAD uri lacks scheme, falling back to file-scheme."),e="file"),e)}function U(e,n){switch(e){case"https":case"http":case"file":n?n[0]!==W&&(n=W+n):n=W;break}return n}var H="",W="/",q=/^(([^:/?#]+?):)?(\/\/([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?/,z=function(){function e(e,n,t,r,a,i){void 0===i&&(i=!1),"object"===typeof e?(this.scheme=e.scheme||H,this.authority=e.authority||H,this.path=e.path||H,this.query=e.query||H,this.fragment=e.fragment||H):(this.scheme=j(e,i),this.authority=n||H,this.path=U(this.scheme,t||H),this.query=r||H,this.fragment=a||H,K(this,i))}return e.isUri=function(n){return n instanceof e||!!n&&("string"===typeof n.authority&&"string"===typeof n.fragment&&"string"===typeof n.path&&"string"===typeof n.query&&"string"===typeof n.scheme&&"function"===typeof n.fsPath&&"function"===typeof n.with&&"function"===typeof n.toString)},Object.defineProperty(e.prototype,"fsPath",{get:function(){return Z(this)},enumerable:!0,configurable:!0}),e.prototype.with=function(e){if(!e)return this;var n=e.scheme,t=e.authority,r=e.path,a=e.query,i=e.fragment;return void 0===n?n=this.scheme:null===n&&(n=H),void 0===t?t=this.authority:null===t&&(t=H),void 0===r?r=this.path:null===r&&(r=H),void 0===a?a=this.query:null===a&&(a=H),void 0===i?i=this.fragment:null===i&&(i=H),n===this.scheme&&t===this.authority&&r===this.path&&a===this.query&&i===this.fragment?this:new J(n,t,r,a,i)},e.parse=function(e,n){void 0===n&&(n=!1);var t=q.exec(e);return t?new J(t[2]||H,decodeURIComponent(t[4]||H),decodeURIComponent(t[5]||H),decodeURIComponent(t[7]||H),decodeURIComponent(t[9]||H),n):new J(H,H,H,H,H)},e.file=function(e){var n=H;if(f["c"]&&(e=e.replace(/\\/g,W)),e[0]===W&&e[1]===W){var t=e.indexOf(W,2);-1===t?(n=e.substring(2),e=W):(n=e.substring(2,t),e=e.substring(t)||W)}return new J("file",n,e,H,H)},e.from=function(e){return new J(e.scheme,e.authority,e.path,e.query,e.fragment)},e.prototype.toString=function(e){return void 0===e&&(e=!1),$(this,e)},e.prototype.toJSON=function(){return this},e.revive=function(n){if(n){if(n instanceof e)return n;var t=new J(n);return t._fsPath=n.fsPath,t._formatted=n.external,t}return n},e}(),J=function(e){function n(){var n=null!==e&&e.apply(this,arguments)||this;return n._formatted=null,n._fsPath=null,n}return R(n,e),Object.defineProperty(n.prototype,"fsPath",{get:function(){return this._fsPath||(this._fsPath=Z(this)),this._fsPath},enumerable:!0,configurable:!0}),n.prototype.toString=function(e){return void 0===e&&(e=!1),e?$(this,!0):(this._formatted||(this._formatted=$(this,!1)),this._formatted)},n.prototype.toJSON=function(){var e={$mid:1};return this._fsPath&&(e.fsPath=this._fsPath),this._formatted&&(e.external=this._formatted),this.path&&(e.path=this.path),this.scheme&&(e.scheme=this.scheme),this.authority&&(e.authority=this.authority),this.query&&(e.query=this.query),this.fragment&&(e.fragment=this.fragment),e},n}(z),X=(O={},O[58]="%3A",O[47]="%2F",O[63]="%3F",O[35]="%23",O[91]="%5B",O[93]="%5D",O[64]="%40",O[33]="%21",O[36]="%24",O[38]="%26",O[39]="%27",O[40]="%28",O[41]="%29",O[42]="%2A",O[43]="%2B",O[44]="%2C",O[59]="%3B",O[61]="%3D",O[32]="%20",O);function Y(e,n){for(var t=void 0,r=-1,a=0;a<e.length;a++){var i=e.charCodeAt(a);if(i>=97&&i<=122||i>=65&&i<=90||i>=48&&i<=57||45===i||46===i||95===i||126===i||n&&47===i)-1!==r&&(t+=encodeURIComponent(e.substring(r,a)),r=-1),void 0!==t&&(t+=e.charAt(a));else{void 0===t&&(t=e.substr(0,a));var o=X[i];void 0!==o?(-1!==r&&(t+=encodeURIComponent(e.substring(r,a)),r=-1),t+=o):-1===r&&(r=a)}}return-1!==r&&(t+=encodeURIComponent(e.substring(r))),void 0!==t?t:e}function Q(e){for(var n=void 0,t=0;t<e.length;t++){var r=e.charCodeAt(t);35===r||63===r?(void 0===n&&(n=e.substr(0,t)),n+=X[r]):void 0!==n&&(n+=e[t])}return void 0!==n?n:e}function Z(e){var n;return n=e.authority&&e.path.length>1&&"file"===e.scheme?"//"+e.authority+e.path:47===e.path.charCodeAt(0)&&(e.path.charCodeAt(1)>=65&&e.path.charCodeAt(1)<=90||e.path.charCodeAt(1)>=97&&e.path.charCodeAt(1)<=122)&&58===e.path.charCodeAt(2)?e.path[1].toLowerCase()+e.path.substr(2):e.path,f["c"]&&(n=n.replace(/\//g,"\\")),n}function $(e,n){var t=n?Q:Y,r="",a=e.scheme,i=e.authority,o=e.path,s=e.query,c=e.fragment;if(a&&(r+=a,r+=":"),(i||"file"===a)&&(r+=W,r+=W),i){var l=i.indexOf("@");if(-1!==l){var u=i.substr(0,l);i=i.substr(l+1),l=u.indexOf(":"),-1===l?r+=t(u,!1):(r+=t(u.substr(0,l),!1),r+=":",r+=t(u.substr(l+1),!1)),r+="@"}i=i.toLowerCase(),l=i.indexOf(":"),-1===l?r+=t(i,!1):(r+=t(i.substr(0,l),!1),r+=i.substr(l))}if(o){if(o.length>=3&&47===o.charCodeAt(0)&&58===o.charCodeAt(2)){var d=o.charCodeAt(1);d>=65&&d<=90&&(o="/"+String.fromCharCode(d+32)+":"+o.substr(3))}else if(o.length>=2&&58===o.charCodeAt(1)){d=o.charCodeAt(0);d>=65&&d<=90&&(o=String.fromCharCode(d+32)+":"+o.substr(2))}r+=t(o,!0)}return s&&(r+="?",r+=t(s,!1)),c&&(r+="#",r+=n?c:Y(c,!1)),r}var ee=function(){function e(e,n){this.lineNumber=e,this.column=n}return e.prototype.with=function(n,t){return void 0===n&&(n=this.lineNumber),void 0===t&&(t=this.column),n===this.lineNumber&&t===this.column?this:new e(n,t)},e.prototype.delta=function(e,n){return void 0===e&&(e=0),void 0===n&&(n=0),this.with(this.lineNumber+e,this.column+n)},e.prototype.equals=function(n){return e.equals(this,n)},e.equals=function(e,n){return!e&&!n||!!e&&!!n&&e.lineNumber===n.lineNumber&&e.column===n.column},e.prototype.isBefore=function(n){return e.isBefore(this,n)},e.isBefore=function(e,n){return e.lineNumber<n.lineNumber||!(n.lineNumber<e.lineNumber)&&e.column<n.column},e.prototype.isBeforeOrEqual=function(n){return e.isBeforeOrEqual(this,n)},e.isBeforeOrEqual=function(e,n){return e.lineNumber<n.lineNumber||!(n.lineNumber<e.lineNumber)&&e.column<=n.column},e.compare=function(e,n){var t=0|e.lineNumber,r=0|n.lineNumber;if(t===r){var a=0|e.column,i=0|n.column;return a-i}return t-r},e.prototype.clone=function(){return new e(this.lineNumber,this.column)},e.prototype.toString=function(){return"("+this.lineNumber+","+this.column+")"},e.lift=function(n){return new e(n.lineNumber,n.column)},e.isIPosition=function(e){return e&&"number"===typeof e.lineNumber&&"number"===typeof e.column},e}(),ne=function(){function e(e,n,t,r){e>t||e===t&&n>r?(this.startLineNumber=t,this.startColumn=r,this.endLineNumber=e,this.endColumn=n):(this.startLineNumber=e,this.startColumn=n,this.endLineNumber=t,this.endColumn=r)}return e.prototype.isEmpty=function(){return e.isEmpty(this)},e.isEmpty=function(e){return e.startLineNumber===e.endLineNumber&&e.startColumn===e.endColumn},e.prototype.containsPosition=function(n){return e.containsPosition(this,n)},e.containsPosition=function(e,n){return!(n.lineNumber<e.startLineNumber||n.lineNumber>e.endLineNumber)&&(!(n.lineNumber===e.startLineNumber&&n.column<e.startColumn)&&!(n.lineNumber===e.endLineNumber&&n.column>e.endColumn))},e.prototype.containsRange=function(n){return e.containsRange(this,n)},e.containsRange=function(e,n){return!(n.startLineNumber<e.startLineNumber||n.endLineNumber<e.startLineNumber)&&(!(n.startLineNumber>e.endLineNumber||n.endLineNumber>e.endLineNumber)&&(!(n.startLineNumber===e.startLineNumber&&n.startColumn<e.startColumn)&&!(n.endLineNumber===e.endLineNumber&&n.endColumn>e.endColumn)))},e.prototype.plusRange=function(n){return e.plusRange(this,n)},e.plusRange=function(n,t){var r,a,i,o;return t.startLineNumber<n.startLineNumber?(r=t.startLineNumber,a=t.startColumn):t.startLineNumber===n.startLineNumber?(r=t.startLineNumber,a=Math.min(t.startColumn,n.startColumn)):(r=n.startLineNumber,a=n.startColumn),t.endLineNumber>n.endLineNumber?(i=t.endLineNumber,o=t.endColumn):t.endLineNumber===n.endLineNumber?(i=t.endLineNumber,o=Math.max(t.endColumn,n.endColumn)):(i=n.endLineNumber,o=n.endColumn),new e(r,a,i,o)},e.prototype.intersectRanges=function(n){return e.intersectRanges(this,n)},e.intersectRanges=function(n,t){var r=n.startLineNumber,a=n.startColumn,i=n.endLineNumber,o=n.endColumn,s=t.startLineNumber,c=t.startColumn,l=t.endLineNumber,u=t.endColumn;return r<s?(r=s,a=c):r===s&&(a=Math.max(a,c)),i>l?(i=l,o=u):i===l&&(o=Math.min(o,u)),r>i?null:r===i&&a>o?null:new e(r,a,i,o)},e.prototype.equalsRange=function(n){return e.equalsRange(this,n)},e.equalsRange=function(e,n){return!!e&&!!n&&e.startLineNumber===n.startLineNumber&&e.startColumn===n.startColumn&&e.endLineNumber===n.endLineNumber&&e.endColumn===n.endColumn},e.prototype.getEndPosition=function(){return new ee(this.endLineNumber,this.endColumn)},e.prototype.getStartPosition=function(){return new ee(this.startLineNumber,this.startColumn)},e.prototype.toString=function(){return"["+this.startLineNumber+","+this.startColumn+" -> "+this.endLineNumber+","+this.endColumn+"]"},e.prototype.setEndPosition=function(n,t){return new e(this.startLineNumber,this.startColumn,n,t)},e.prototype.setStartPosition=function(n,t){return new e(n,t,this.endLineNumber,this.endColumn)},e.prototype.collapseToStart=function(){return e.collapseToStart(this)},e.collapseToStart=function(n){return new e(n.startLineNumber,n.startColumn,n.startLineNumber,n.startColumn)},e.fromPositions=function(n,t){return void 0===t&&(t=n),new e(n.lineNumber,n.column,t.lineNumber,t.column)},e.lift=function(n){return n?new e(n.startLineNumber,n.startColumn,n.endLineNumber,n.endColumn):null},e.isIRange=function(e){return e&&"number"===typeof e.startLineNumber&&"number"===typeof e.startColumn&&"number"===typeof e.endLineNumber&&"number"===typeof e.endColumn},e.areIntersectingOrTouching=function(e,n){return!(e.endLineNumber<n.startLineNumber||e.endLineNumber===n.startLineNumber&&e.endColumn<n.startColumn)&&!(n.endLineNumber<e.startLineNumber||n.endLineNumber===e.startLineNumber&&n.endColumn<e.startColumn)},e.areIntersecting=function(e,n){return!(e.endLineNumber<n.startLineNumber||e.endLineNumber===n.startLineNumber&&e.endColumn<=n.startColumn)&&!(n.endLineNumber<e.startLineNumber||n.endLineNumber===e.startLineNumber&&n.endColumn<=e.startColumn)},e.compareRangesUsingStarts=function(e,n){if(e&&n){var t=0|e.startLineNumber,r=0|n.startLineNumber;if(t===r){var a=0|e.startColumn,i=0|n.startColumn;if(a===i){var o=0|e.endLineNumber,s=0|n.endLineNumber;if(o===s){var c=0|e.endColumn,l=0|n.endColumn;return c-l}return o-s}return a-i}return t-r}var u=e?1:0,d=n?1:0;return u-d},e.compareRangesUsingEnds=function(e,n){return e.endLineNumber===n.endLineNumber?e.endColumn===n.endColumn?e.startLineNumber===n.startLineNumber?e.startColumn-n.startColumn:e.startLineNumber-n.startLineNumber:e.endColumn-n.endColumn:e.endLineNumber-n.endLineNumber},e.spansMultipleLines=function(e){return e.endLineNumber>e.startLineNumber},e}();function te(e){for(var n=0,t=e.length;n<t;n++){var r=e.charCodeAt(n);if(32!==r&&9!==r)return n}return-1}function re(e,n){void 0===n&&(n=e.length-1);for(var t=n;t>=0;t--){var r=e.charCodeAt(t);if(32!==r&&9!==r)return t}return-1}String.fromCharCode(65279);var ae=5e3,ie=3;function oe(e,n,t,r){var a=new I(e,n,t);return a.ComputeDiff(r)}var se=function(){function e(n){for(var t=[],r=[],a=0,i=n.length;a<i;a++)t[a]=e._getFirstNonBlankColumn(n[a],1),r[a]=e._getLastNonBlankColumn(n[a],1);this._lines=n,this._startColumns=t,this._endColumns=r}return e.prototype.getLength=function(){return this._lines.length},e.prototype.getElementAtIndex=function(e){return this._lines[e].substring(this._startColumns[e]-1,this._endColumns[e]-1)},e.prototype.getStartLineNumber=function(e){return e+1},e.prototype.getEndLineNumber=function(e){return e+1},e._getFirstNonBlankColumn=function(e,n){var t=te(e);return-1===t?n:t+1},e._getLastNonBlankColumn=function(e,n){var t=re(e);return-1===t?n:t+2},e.prototype.getCharSequence=function(e,n,t){for(var r=[],a=[],i=[],o=0,s=n;s<=t;s++)for(var c=this._lines[s],l=e?this._startColumns[s]:1,u=e?this._endColumns[s]:c.length+1,d=l;d<u;d++)r[o]=c.charCodeAt(d-1),a[o]=s+1,i[o]=d,o++;return new ce(r,a,i)},e}(),ce=function(){function e(e,n,t){this._charCodes=e,this._lineNumbers=n,this._columns=t}return e.prototype.getLength=function(){return this._charCodes.length},e.prototype.getElementAtIndex=function(e){return this._charCodes[e]},e.prototype.getStartLineNumber=function(e){return this._lineNumbers[e]},e.prototype.getStartColumn=function(e){return this._columns[e]},e.prototype.getEndLineNumber=function(e){return this._lineNumbers[e]},e.prototype.getEndColumn=function(e){return this._columns[e]+1},e}(),le=function(){function e(e,n,t,r,a,i,o,s){this.originalStartLineNumber=e,this.originalStartColumn=n,this.originalEndLineNumber=t,this.originalEndColumn=r,this.modifiedStartLineNumber=a,this.modifiedStartColumn=i,this.modifiedEndLineNumber=o,this.modifiedEndColumn=s}return e.createFromDiffChange=function(n,t,r){var a,i,o,s,c,l,u,d;return 0===n.originalLength?(a=0,i=0,o=0,s=0):(a=t.getStartLineNumber(n.originalStart),i=t.getStartColumn(n.originalStart),o=t.getEndLineNumber(n.originalStart+n.originalLength-1),s=t.getEndColumn(n.originalStart+n.originalLength-1)),0===n.modifiedLength?(c=0,l=0,u=0,d=0):(c=r.getStartLineNumber(n.modifiedStart),l=r.getStartColumn(n.modifiedStart),u=r.getEndLineNumber(n.modifiedStart+n.modifiedLength-1),d=r.getEndColumn(n.modifiedStart+n.modifiedLength-1)),new e(a,i,o,s,c,l,u,d)},e}();function ue(e){if(e.length<=1)return e;for(var n=[e[0]],t=n[0],r=1,a=e.length;r<a;r++){var i=e[r],o=i.originalStart-(t.originalStart+t.originalLength),s=i.modifiedStart-(t.modifiedStart+t.modifiedLength),c=Math.min(o,s);c<ie?(t.originalLength=i.originalStart+i.originalLength-t.originalStart,t.modifiedLength=i.modifiedStart+i.modifiedLength-t.modifiedStart):(n.push(i),t=i)}return n}var de=function(){function e(e,n,t,r,a){this.originalStartLineNumber=e,this.originalEndLineNumber=n,this.modifiedStartLineNumber=t,this.modifiedEndLineNumber=r,this.charChanges=a}return e.createFromDiffResult=function(n,t,r,a,i,o,s){var c,l,u,d,p=void 0;if(0===t.originalLength?(c=r.getStartLineNumber(t.originalStart)-1,l=0):(c=r.getStartLineNumber(t.originalStart),l=r.getEndLineNumber(t.originalStart+t.originalLength-1)),0===t.modifiedLength?(u=a.getStartLineNumber(t.modifiedStart)-1,d=0):(u=a.getStartLineNumber(t.modifiedStart),d=a.getEndLineNumber(t.modifiedStart+t.modifiedLength-1)),o&&0!==t.originalLength&&0!==t.modifiedLength&&i()){var m=r.getCharSequence(n,t.originalStart,t.originalStart+t.originalLength-1),f=a.getCharSequence(n,t.modifiedStart,t.modifiedStart+t.modifiedLength-1),_=oe(m,f,i,!0);s&&(_=ue(_)),p=[];for(var g=0,v=_.length;g<v;g++)p.push(le.createFromDiffChange(_[g],m,f))}return new e(c,l,u,d,p)},e}(),pe=function(){function e(e,n,t){this.shouldComputeCharChanges=t.shouldComputeCharChanges,this.shouldPostProcessCharChanges=t.shouldPostProcessCharChanges,this.shouldIgnoreTrimWhitespace=t.shouldIgnoreTrimWhitespace,this.shouldMakePrettyDiff=t.shouldMakePrettyDiff,this.maximumRunTimeMs=ae,this.originalLines=e,this.modifiedLines=n,this.original=new se(e),this.modified=new se(n)}return e.prototype.computeDiff=function(){if(1===this.original.getLength()&&0===this.original.getElementAtIndex(0).length)return[{originalStartLineNumber:1,originalEndLineNumber:1,modifiedStartLineNumber:1,modifiedEndLineNumber:this.modified.getLength(),charChanges:[{modifiedEndColumn:0,modifiedEndLineNumber:0,modifiedStartColumn:0,modifiedStartLineNumber:0,originalEndColumn:0,originalEndLineNumber:0,originalStartColumn:0,originalStartLineNumber:0}]}];if(1===this.modified.getLength()&&0===this.modified.getElementAtIndex(0).length)return[{originalStartLineNumber:1,originalEndLineNumber:this.original.getLength(),modifiedStartLineNumber:1,modifiedEndLineNumber:1,charChanges:[{modifiedEndColumn:0,modifiedEndLineNumber:0,modifiedStartColumn:0,modifiedStartLineNumber:0,originalEndColumn:0,originalEndLineNumber:0,originalStartColumn:0,originalStartLineNumber:0}]}];this.computationStartTime=(new Date).getTime();var e=oe(this.original,this.modified,this._continueProcessingPredicate.bind(this),this.shouldMakePrettyDiff);if(this.shouldIgnoreTrimWhitespace){for(var n=[],t=0,r=e.length;t<r;t++)n.push(de.createFromDiffResult(this.shouldIgnoreTrimWhitespace,e[t],this.original,this.modified,this._continueProcessingPredicate.bind(this),this.shouldComputeCharChanges,this.shouldPostProcessCharChanges));return n}for(var a=[],i=0,o=0,s=(t=-1,e.length);t<s;t++){var c=t+1<s?e[t+1]:null,l=c?c.originalStart:this.originalLines.length,u=c?c.modifiedStart:this.modifiedLines.length;while(i<l&&o<u){var d=this.originalLines[i],p=this.modifiedLines[o];if(d!==p){var m=se._getFirstNonBlankColumn(d,1),f=se._getFirstNonBlankColumn(p,1);while(m>1&&f>1){var _=d.charCodeAt(m-2),g=p.charCodeAt(f-2);if(_!==g)break;m--,f--}(m>1||f>1)&&this._pushTrimWhitespaceCharChange(a,i+1,1,m,o+1,1,f);var v=se._getLastNonBlankColumn(d,1),y=se._getLastNonBlankColumn(p,1),h=d.length+1,b=p.length+1;while(v<h&&y<b){_=d.charCodeAt(v-1),g=d.charCodeAt(y-1);if(_!==g)break;v++,y++}(v<h||y<b)&&this._pushTrimWhitespaceCharChange(a,i+1,v,h,o+1,y,b)}i++,o++}c&&(a.push(de.createFromDiffResult(this.shouldIgnoreTrimWhitespace,c,this.original,this.modified,this._continueProcessingPredicate.bind(this),this.shouldComputeCharChanges,this.shouldPostProcessCharChanges)),i+=c.originalLength,o+=c.modifiedLength)}return a},e.prototype._pushTrimWhitespaceCharChange=function(e,n,t,r,a,i,o){if(!this._mergeTrimWhitespaceCharChange(e,n,t,r,a,i,o)){var s=void 0;this.shouldComputeCharChanges&&(s=[new le(n,t,n,r,a,i,a,o)]),e.push(new de(n,n,a,a,s))}},e.prototype._mergeTrimWhitespaceCharChange=function(e,n,t,r,a,i,o){var s=e.length;if(0===s)return!1;var c=e[s-1];return 0!==c.originalEndLineNumber&&0!==c.modifiedEndLineNumber&&(c.originalEndLineNumber+1===n&&c.modifiedEndLineNumber+1===a&&(c.originalEndLineNumber=n,c.modifiedEndLineNumber=a,this.shouldComputeCharChanges&&c.charChanges.push(new le(n,t,n,r,a,i,a,o)),!0))},e.prototype._continueProcessingPredicate=function(){if(0===this.maximumRunTimeMs)return!0;var e=(new Date).getTime();return e-this.computationStartTime<this.maximumRunTimeMs},e}(),me=function(){function e(e,n,t){for(var r=new Uint8Array(e*n),a=0,i=e*n;a<i;a++)r[a]=t;this._data=r,this.rows=e,this.cols=n}return e.prototype.get=function(e,n){return this._data[e*this.cols+n]},e.prototype.set=function(e,n,t){this._data[e*this.cols+n]=t},e}();function fe(e){return e<0?0:e>255?255:0|e}function _e(e){return e<0?0:e>4294967295?4294967295:0|e}var ge=function(){function e(e,n){this.index=e,this.remainder=n}return e}(),ve=function(){function e(e){this.values=e,this.prefixSum=new Uint32Array(e.length),this.prefixSumValidIndex=new Int32Array(1),this.prefixSumValidIndex[0]=-1}return e.prototype.getCount=function(){return this.values.length},e.prototype.insertValues=function(e,n){e=_e(e);var t=this.values,r=this.prefixSum,a=n.length;return 0!==a&&(this.values=new Uint32Array(t.length+a),this.values.set(t.subarray(0,e),0),this.values.set(t.subarray(e),e+a),this.values.set(n,e),e-1<this.prefixSumValidIndex[0]&&(this.prefixSumValidIndex[0]=e-1),this.prefixSum=new Uint32Array(this.values.length),this.prefixSumValidIndex[0]>=0&&this.prefixSum.set(r.subarray(0,this.prefixSumValidIndex[0]+1)),!0)},e.prototype.changeValue=function(e,n){return e=_e(e),n=_e(n),this.values[e]!==n&&(this.values[e]=n,e-1<this.prefixSumValidIndex[0]&&(this.prefixSumValidIndex[0]=e-1),!0)},e.prototype.removeValues=function(e,n){e=_e(e),n=_e(n);var t=this.values,r=this.prefixSum;if(e>=t.length)return!1;var a=t.length-e;return n>=a&&(n=a),0!==n&&(this.values=new Uint32Array(t.length-n),this.values.set(t.subarray(0,e),0),this.values.set(t.subarray(e+n),e),this.prefixSum=new Uint32Array(this.values.length),e-1<this.prefixSumValidIndex[0]&&(this.prefixSumValidIndex[0]=e-1),this.prefixSumValidIndex[0]>=0&&this.prefixSum.set(r.subarray(0,this.prefixSumValidIndex[0]+1)),!0)},e.prototype.getTotalValue=function(){return 0===this.values.length?0:this._getAccumulatedValue(this.values.length-1)},e.prototype.getAccumulatedValue=function(e){return e<0?0:(e=_e(e),this._getAccumulatedValue(e))},e.prototype._getAccumulatedValue=function(e){if(e<=this.prefixSumValidIndex[0])return this.prefixSum[e];var n=this.prefixSumValidIndex[0]+1;0===n&&(this.prefixSum[0]=this.values[0],n++),e>=this.values.length&&(e=this.values.length-1);for(var t=n;t<=e;t++)this.prefixSum[t]=this.prefixSum[t-1]+this.values[t];return this.prefixSumValidIndex[0]=Math.max(this.prefixSumValidIndex[0],e),this.prefixSum[e]},e.prototype.getIndexOf=function(e){e=Math.floor(e),this.getTotalValue();var n=0,t=this.values.length-1,r=0,a=0,i=0;while(n<=t)if(r=n+(t-n)/2|0,a=this.prefixSum[r],i=a-this.values[r],e<i)t=r-1;else{if(!(e>=a))break;n=r+1}return new ge(r,e-i)},e}(),ye=(function(){function e(e){this._cacheAccumulatedValueStart=0,this._cache=null,this._actual=new ve(e),this._bustCache()}e.prototype._bustCache=function(){this._cacheAccumulatedValueStart=0,this._cache=null},e.prototype.insertValues=function(e,n){this._actual.insertValues(e,n)&&this._bustCache()},e.prototype.changeValue=function(e,n){this._actual.changeValue(e,n)&&this._bustCache()},e.prototype.removeValues=function(e,n){this._actual.removeValues(e,n)&&this._bustCache()},e.prototype.getTotalValue=function(){return this._actual.getTotalValue()},e.prototype.getAccumulatedValue=function(e){return this._actual.getAccumulatedValue(e)},e.prototype.getIndexOf=function(e){if(e=Math.floor(e),null!==this._cache){var n=e-this._cacheAccumulatedValueStart;if(n>=0&&n<this._cache.length)return this._cache[n]}return this._actual.getIndexOf(e)},e.prototype.warmUpCache=function(e,n){for(var t=[],r=e;r<=n;r++)t[r-e]=this.getIndexOf(r);this._cache=t,this._cacheAccumulatedValueStart=e}}(),function(){function e(e,n,t,r){this._uri=e,this._lines=n,this._eol=t,this._versionId=r,this._lineStarts=null}return e.prototype.dispose=function(){this._lines.length=0},e.prototype.getText=function(){return this._lines.join(this._eol)},e.prototype.onEvents=function(e){e.eol&&e.eol!==this._eol&&(this._eol=e.eol,this._lineStarts=null);for(var n=e.changes,t=0,r=n;t<r.length;t++){var a=r[t];this._acceptDeleteRange(a.range),this._acceptInsertText(new ee(a.range.startLineNumber,a.range.startColumn),a.text)}this._versionId=e.versionId},e.prototype._ensureLineStarts=function(){if(!this._lineStarts){for(var e=this._eol.length,n=this._lines.length,t=new Uint32Array(n),r=0;r<n;r++)t[r]=this._lines[r].length+e;this._lineStarts=new ve(t)}},e.prototype._setLineText=function(e,n){this._lines[e]=n,this._lineStarts&&this._lineStarts.changeValue(e,this._lines[e].length+this._eol.length)},e.prototype._acceptDeleteRange=function(e){if(e.startLineNumber!==e.endLineNumber)this._setLineText(e.startLineNumber-1,this._lines[e.startLineNumber-1].substring(0,e.startColumn-1)+this._lines[e.endLineNumber-1].substring(e.endColumn-1)),this._lines.splice(e.startLineNumber,e.endLineNumber-e.startLineNumber),this._lineStarts&&this._lineStarts.removeValues(e.startLineNumber,e.endLineNumber-e.startLineNumber);else{if(e.startColumn===e.endColumn)return;this._setLineText(e.startLineNumber-1,this._lines[e.startLineNumber-1].substring(0,e.startColumn-1)+this._lines[e.startLineNumber-1].substring(e.endColumn-1))}},e.prototype._acceptInsertText=function(e,n){if(0!==n.length){var t=n.split(/\r\n|\r|\n/);if(1!==t.length){t[t.length-1]+=this._lines[e.lineNumber-1].substring(e.column-1),this._setLineText(e.lineNumber-1,this._lines[e.lineNumber-1].substring(0,e.column-1)+t[0]);for(var r=new Uint32Array(t.length-1),a=1;a<t.length;a++)this._lines.splice(e.lineNumber+a-1,0,t[a]),r[a-1]=t[a].length+this._eol.length;this._lineStarts&&this._lineStarts.insertValues(e.lineNumber,r)}else this._setLineText(e.lineNumber-1,this._lines[e.lineNumber-1].substring(0,e.column-1)+t[0]+this._lines[e.lineNumber-1].substring(e.column-1))}},e}()),he="`~!@#$%^&*()-=+[{]}\\|;:'\",.<>/?";function be(e){void 0===e&&(e="");for(var n="(-?\\d*\\.\\d\\w*)|([^",t=0,r=he;t<r.length;t++){var a=r[t];e.indexOf(a)>=0||(n+="\\"+a)}return n+="\\s]+)",new RegExp(n,"g")}var Ee=be();function Te(e){var n=Ee;if(e&&e instanceof RegExp)if(e.global)n=e;else{var t="g";e.ignoreCase&&(t+="i"),e.multiline&&(t+="m"),e.unicode&&(t+="u"),n=new RegExp(e.source,t)}return n.lastIndex=0,n}function Se(e,n,t,r){var a,i=e-1-r,o=t.lastIndexOf(" ",i-1)+1;n.lastIndex=o;while(a=n.exec(t)){var s=a.index||0;if(s<=i&&n.lastIndex>=i)return{word:a[0],startColumn:r+1+s,endColumn:r+1+n.lastIndex}}return null}function xe(e,n,t,r){var a,i=e-1-r;n.lastIndex=0;while(a=n.exec(t)){var o=a.index||0;if(o>i)return null;if(n.lastIndex>=i)return{word:a[0],startColumn:r+1+o,endColumn:r+1+n.lastIndex}}return null}function Ce(e,n,t,r){n.lastIndex=0;var a=n.exec(t);if(!a)return null;var i=a[0].indexOf(" ")>=0?xe(e,n,t,r):Se(e,n,t,r);return n.lastIndex=0,i}var Ae=function(){function e(n){var t=fe(n);this._defaultValue=t,this._asciiMap=e._createAsciiMap(t),this._map=new Map}return e._createAsciiMap=function(e){for(var n=new Uint8Array(256),t=0;t<256;t++)n[t]=e;return n},e.prototype.set=function(e,n){var t=fe(n);e>=0&&e<256?this._asciiMap[e]=t:this._map.set(e,t)},e.prototype.get=function(e){return e>=0&&e<256?this._asciiMap[e]:this._map.get(e)||this._defaultValue},e}(),De=(function(){function e(){this._actual=new Ae(0)}e.prototype.add=function(e){this._actual.set(e,1)},e.prototype.has=function(e){return 1===this._actual.get(e)}}(),function(){function e(e){for(var n=0,t=0,r=0,a=e.length;r<a;r++){var i=e[r],o=i[0],s=i[1],c=i[2];s>n&&(n=s),o>t&&(t=o),c>t&&(t=c)}n++,t++;var l=new me(t,n,0);for(r=0,a=e.length;r<a;r++){var u=e[r];o=u[0],s=u[1],c=u[2];l.set(o,s,c)}this._states=l,this._maxCharCode=n}return e.prototype.nextState=function(e,n){return n<0||n>=this._maxCharCode?0:this._states.get(e,n)},e}()),Le=null;function ke(){return null===Le&&(Le=new De([[1,104,2],[1,72,2],[1,102,6],[1,70,6],[2,116,3],[2,84,3],[3,116,4],[3,84,4],[4,112,5],[4,80,5],[5,115,9],[5,83,9],[5,58,10],[6,105,7],[6,73,7],[7,108,8],[7,76,8],[8,101,9],[8,69,9],[9,58,10],[10,47,11],[11,47,12]])),Le}var Ne=null;function Ie(){if(null===Ne){Ne=new Ae(0);for(var e=" \t<>'\"、。｡､，．：；？！＠＃＄％＆＊‘“〈《「『【〔（［｛｢｣｝］）〕】』」》〉”’｀～…",n=0;n<e.length;n++)Ne.set(e.charCodeAt(n),1);var t=".,;";for(n=0;n<t.length;n++)Ne.set(t.charCodeAt(n),2)}return Ne}var we=function(){function e(){}return e._createLink=function(e,n,t,r,a){var i=a-1;do{var o=n.charCodeAt(i),s=e.get(o);if(2!==s)break;i--}while(i>r);if(r>0){var c=n.charCodeAt(r-1),l=n.charCodeAt(i);(40===c&&41===l||91===c&&93===l||123===c&&125===l)&&i--}return{range:{startLineNumber:t,startColumn:r+1,endLineNumber:t,endColumn:i+2},url:n.substring(r,i+1)}},e.computeLinks=function(n,t){void 0===t&&(t=ke());for(var r=Ie(),a=[],i=1,o=n.getLineCount();i<=o;i++){var s=n.getLineContent(i),c=s.length,l=0,u=0,d=0,p=1,m=!1,f=!1,_=!1;while(l<c){var g=!1,v=s.charCodeAt(l);if(13===p){var y=void 0;switch(v){case 40:m=!0,y=0;break;case 41:y=m?0:1;break;case 91:f=!0,y=0;break;case 93:y=f?0:1;break;case 123:_=!0,y=0;break;case 125:y=_?0:1;break;case 39:y=34===d||96===d?0:1;break;case 34:y=39===d||96===d?0:1;break;case 96:y=39===d||34===d?0:1;break;default:y=r.get(v)}1===y&&(a.push(e._createLink(r,s,i,u,l)),g=!0)}else if(12===p){y=void 0;91===v?(f=!0,y=0):y=r.get(v),1===y?g=!0:p=13}else p=t.nextState(p,v),0===p&&(g=!0);g&&(p=1,m=!1,f=!1,_=!1,u=l+1,d=v),l++}13===p&&a.push(e._createLink(r,s,i,u,c))}return a},e}();function Me(e){return e&&"function"===typeof e.getLineCount&&"function"===typeof e.getLineContent?we.computeLinks(e):[]}var Oe=function(){function e(){this._defaultValueSet=[["true","false"],["True","False"],["Private","Public","Friend","ReadOnly","Partial","Protected","WriteOnly"],["public","protected","private"]]}return e.prototype.navigateValueSet=function(e,n,t,r,a){if(e&&n){var i=this.doNavigateValueSet(n,a);if(i)return{range:e,value:i}}if(t&&r){i=this.doNavigateValueSet(r,a);if(i)return{range:t,value:i}}return null},e.prototype.doNavigateValueSet=function(e,n){var t=this.numberReplace(e,n);return null!==t?t:this.textReplace(e,n)},e.prototype.numberReplace=function(e,n){var t=Math.pow(10,e.length-(e.lastIndexOf(".")+1)),r=Number(e),a=parseFloat(e);return isNaN(r)||isNaN(a)||r!==a?null:0!==r||n?(r=Math.floor(r*t),r+=n?t:-t,String(r/t)):null},e.prototype.textReplace=function(e,n){return this.valueSetsReplace(this._defaultValueSet,e,n)},e.prototype.valueSetsReplace=function(e,n,t){for(var r=null,a=0,i=e.length;null===r&&a<i;a++)r=this.valueSetReplace(e[a],n,t);return r},e.prototype.valueSetReplace=function(e,n,t){var r=e.indexOf(n);return r>=0?(r+=t?1:-1,r<0?r=e.length-1:r%=e.length,e[r]):null},e.INSTANCE=new e,e}();t("5110");function Pe(e){var n,t=this,r=!1;return function(){return r?n:(r=!0,n=e.apply(t,arguments),n)}}var Re,Fe=function(){function e(n){this.element=n,this.next=e.Undefined,this.prev=e.Undefined}return e.Undefined=new e(void 0),e}(),Ge=function(){function e(){this._first=Fe.Undefined,this._last=Fe.Undefined,this._size=0}return Object.defineProperty(e.prototype,"size",{get:function(){return this._size},enumerable:!0,configurable:!0}),e.prototype.isEmpty=function(){return this._first===Fe.Undefined},e.prototype.clear=function(){this._first=Fe.Undefined,this._last=Fe.Undefined,this._size=0},e.prototype.unshift=function(e){return this._insert(e,!1)},e.prototype.push=function(e){return this._insert(e,!0)},e.prototype._insert=function(e,n){var t=this,r=new Fe(e);if(this._first===Fe.Undefined)this._first=r,this._last=r;else if(n){var a=this._last;this._last=r,r.prev=a,a.next=r}else{var i=this._first;this._first=r,r.next=i,i.prev=r}this._size+=1;var o=!1;return function(){o||(o=!0,t._remove(r))}},e.prototype.shift=function(){if(this._first!==Fe.Undefined){var e=this._first.element;return this._remove(this._first),e}},e.prototype._remove=function(e){if(e.prev!==Fe.Undefined&&e.next!==Fe.Undefined){var n=e.prev;n.next=e.next,e.next.prev=n}else e.prev===Fe.Undefined&&e.next===Fe.Undefined?(this._first=Fe.Undefined,this._last=Fe.Undefined):e.next===Fe.Undefined?(this._last=this._last.prev,this._last.next=Fe.Undefined):e.prev===Fe.Undefined&&(this._first=this._first.next,this._first.prev=Fe.Undefined);this._size-=1},e.prototype.iterator=function(){var e,n=this._first;return{next:function(){return n===Fe.Undefined?M:(e?e.value=n.element:e={done:!1,value:n.element},n=n.next,e)}}},e.prototype.toArray=function(){for(var e=[],n=this._first;n!==Fe.Undefined;n=n.next)e.push(n.element);return e},e}(),Ve=function(){var e=function(n,t){return e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,n){e.__proto__=n}||function(e,n){for(var t in n)n.hasOwnProperty(t)&&(e[t]=n[t])},e(n,t)};return function(n,t){function r(){this.constructor=n}e(n,t),n.prototype=null===t?Object.create(t):(r.prototype=t.prototype,new r)}}();(function(e){var n={dispose:function(){}};function t(e){return function(n,t,r){void 0===t&&(t=null);var a,i=!1;return a=e(function(e){if(!i)return a?a.dispose():i=!0,n.call(t,e)},null,r),i&&a.dispose(),a}}function r(e,n){return l(function(t,r,a){return void 0===r&&(r=null),e(function(e){return t.call(r,n(e))},null,a)})}function a(e,n){return l(function(t,r,a){return void 0===r&&(r=null),e(function(e){n(e),t.call(r,e)},null,a)})}function i(e,n){return l(function(t,r,a){return void 0===r&&(r=null),e(function(e){return n(e)&&t.call(r,e)},null,a)})}function o(e){return e}function s(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return function(n,t,r){return void 0===t&&(t=null),d(e.map(function(e){return e(function(e){return n.call(t,e)},null,r)}))}}function c(e,n,t){var a=t;return r(e,function(e){return a=n(a,e),a})}function l(e){var n,t=new Ue({onFirstListenerAdd:function(){n=e(t.fire,t)},onLastListenerRemove:function(){n.dispose()}});return t.event}function u(e,n,t,r,a){var i;void 0===t&&(t=100),void 0===r&&(r=!1);var o=void 0,s=void 0,c=0,l=new Ue({leakWarningThreshold:a,onFirstListenerAdd:function(){i=e(function(e){c++,o=n(o,e),r&&!s&&l.fire(o),clearTimeout(s),s=setTimeout(function(){var e=o;o=void 0,s=void 0,(!r||c>1)&&l.fire(e),c=0},t)})},onLastListenerRemove:function(){i.dispose()}});return l.event}function p(e){var n=(new Date).getTime();return r(t(e),function(e){return(new Date).getTime()-n})}function m(e){var n,t=!0;return i(e,function(e){var r=t||e!==n;return t=!1,n=e,r})}function f(e,n,t){void 0===n&&(n=!1),void 0===t&&(t=[]);var r=t.slice(),a=e(function(e){r?r.push(e):o.fire(e)}),i=function(){r&&r.forEach(function(e){return o.fire(e)}),r=null},o=new Ue({onFirstListenerAdd:function(){a||(a=e(function(e){return o.fire(e)}))},onFirstListenerDidAdd:function(){r&&(n?setTimeout(i):i())},onLastListenerRemove:function(){a&&a.dispose(),a=null}});return o.event}e.None=function(){return n},e.once=t,e.map=r,e.forEach=a,e.filter=i,e.signal=o,e.any=s,e.reduce=c,e.snapshot=l,e.debounce=u,e.stopwatch=p,e.latch=m,e.buffer=f;var _=function(){function e(e){this.event=e}return e.prototype.map=function(n){return new e(r(this.event,n))},e.prototype.forEach=function(n){return new e(a(this.event,n))},e.prototype.filter=function(n){return new e(i(this.event,n))},e.prototype.reduce=function(n,t){return new e(c(this.event,n,t))},e.prototype.latch=function(){return new e(m(this.event))},e.prototype.on=function(e,n,t){return this.event(e,n,t)},e.prototype.once=function(e,n,r){return t(this.event)(e,n,r)},e}();function g(e){return new _(e)}function v(e,n,t){void 0===t&&(t=function(e){return e});var r=function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return o.fire(t.apply(void 0,e))},a=function(){return e.on(n,r)},i=function(){return e.removeListener(n,r)},o=new Ue({onFirstListenerAdd:a,onLastListenerRemove:i});return o.event}function y(e){var n=new Ue,t=!1;return e.then(void 0,function(){return null}).then(function(){t?n.fire(void 0):setTimeout(function(){return n.fire(void 0)},0)}),t=!0,n.event}function h(e){return new Promise(function(n){return t(e)(n)})}e.chain=g,e.fromNodeEventEmitter=v,e.fromPromise=y,e.toPromise=h})(Re||(Re={}));var Be,Ke=-1,je=function(){function e(e,n){void 0===n&&(n=Math.random().toString(18).slice(2,5)),this.customThreshold=e,this.name=n,this._warnCountdown=0}return e.prototype.dispose=function(){this._stacks&&this._stacks.clear()},e.prototype.check=function(e){var n=this,t=Ke;if("number"===typeof this.customThreshold&&(t=this.customThreshold),!(t<=0||e<t)){this._stacks||(this._stacks=new Map);var r=(new Error).stack.split("\n").slice(3).join("\n"),a=this._stacks.get(r)||0;if(this._stacks.set(r,a+1),this._warnCountdown-=1,this._warnCountdown<=0){var i;this._warnCountdown=.5*t;var o=0;this._stacks.forEach(function(e,n){(!i||o<e)&&(i=n,o=e)}),console.warn("["+this.name+"] potential listener LEAK detected, having "+e+" listeners already. MOST frequent listener ("+o+"):"),console.warn(i)}return function(){var e=n._stacks.get(r)||0;n._stacks.set(r,e-1)}}},e}(),Ue=function(){function e(e){this._disposed=!1,this._options=e,this._leakageMon=Ke>0?new je(this._options&&this._options.leakWarningThreshold):void 0}return Object.defineProperty(e.prototype,"event",{get:function(){var n=this;return this._event||(this._event=function(t,r,a){n._listeners||(n._listeners=new Ge);var i=n._listeners.isEmpty();i&&n._options&&n._options.onFirstListenerAdd&&n._options.onFirstListenerAdd(n);var o,s,c=n._listeners.push(r?[t,r]:t);return i&&n._options&&n._options.onFirstListenerDidAdd&&n._options.onFirstListenerDidAdd(n),n._options&&n._options.onListenerDidAdd&&n._options.onListenerDidAdd(n,t,r),n._leakageMon&&(o=n._leakageMon.check(n._listeners.size)),s={dispose:function(){if(o&&o(),s.dispose=e._noop,!n._disposed&&(c(),n._options&&n._options.onLastListenerRemove)){var t=n._listeners&&!n._listeners.isEmpty();t||n._options.onLastListenerRemove(n)}}},Array.isArray(a)&&a.push(s),s}),this._event},enumerable:!0,configurable:!0}),e.prototype.fire=function(e){if(this._listeners){this._deliveryQueue||(this._deliveryQueue=new Ge);for(var n=this._listeners.iterator(),t=n.next();!t.done;t=n.next())this._deliveryQueue.push([t.value,e]);while(this._deliveryQueue.size>0){var r=this._deliveryQueue.shift(),a=r[0],o=r[1];try{"function"===typeof a?a.call(void 0,o):a[0].call(a[1],o)}catch(t){i(t)}}}},e.prototype.dispose=function(){this._listeners&&this._listeners.clear(),this._deliveryQueue&&this._deliveryQueue.clear(),this._leakageMon&&this._leakageMon.dispose(),this._disposed=!0},e._noop=function(){},e}(),He=(function(e){function n(n){var t=e.call(this,n)||this;return t._isPaused=0,t._eventQueue=new Ge,t._mergeFn=n&&n.merge,t}Ve(n,e),n.prototype.pause=function(){this._isPaused++},n.prototype.resume=function(){if(0!==this._isPaused&&0===--this._isPaused)if(this._mergeFn){var n=this._eventQueue.toArray();this._eventQueue.clear(),e.prototype.fire.call(this,this._mergeFn(n))}else while(!this._isPaused&&0!==this._eventQueue.size)e.prototype.fire.call(this,this._eventQueue.shift())},n.prototype.fire=function(n){this._listeners&&(0!==this._isPaused?this._eventQueue.push(n):e.prototype.fire.call(this,n))}}(Ue),function(){function e(){var e=this;this.hasListeners=!1,this.events=[],this.emitter=new Ue({onFirstListenerAdd:function(){return e.onFirstListenerAdd()},onLastListenerRemove:function(){return e.onLastListenerRemove()}})}Object.defineProperty(e.prototype,"event",{get:function(){return this.emitter.event},enumerable:!0,configurable:!0}),e.prototype.add=function(e){var n=this,t={event:e,listener:null};this.events.push(t),this.hasListeners&&this.hook(t);var r=function(){n.hasListeners&&n.unhook(t);var e=n.events.indexOf(t);n.events.splice(e,1)};return p(Pe(r))},e.prototype.onFirstListenerAdd=function(){var e=this;this.hasListeners=!0,this.events.forEach(function(n){return e.hook(n)})},e.prototype.onLastListenerRemove=function(){var e=this;this.hasListeners=!1,this.events.forEach(function(n){return e.unhook(n)})},e.prototype.hook=function(e){var n=this;e.listener=e.event(function(e){return n.emitter.fire(e)})},e.prototype.unhook=function(e){e.listener&&e.listener.dispose(),e.listener=null},e.prototype.dispose=function(){this.emitter.dispose()}}(),function(){function e(){this.buffers=[]}e.prototype.wrapEvent=function(e){var n=this;return function(t,r,a){return e(function(e){var a=n.buffers[n.buffers.length-1];a?a.push(function(){return t.call(r,e)}):t.call(r,e)},void 0,a)}},e.prototype.bufferEvents=function(e){var n=[];this.buffers.push(n);var t=e();return this.buffers.pop(),n.forEach(function(e){return e()}),t}}(),function(){function e(){var e=this;this.listening=!1,this.inputEvent=Re.None,this.inputEventListener=m.None,this.emitter=new Ue({onFirstListenerDidAdd:function(){e.listening=!0,e.inputEventListener=e.inputEvent(e.emitter.fire,e.emitter)},onLastListenerRemove:function(){e.listening=!1,e.inputEventListener.dispose()}}),this.event=this.emitter.event}Object.defineProperty(e.prototype,"input",{set:function(e){this.inputEvent=e,this.listening&&(this.inputEventListener.dispose(),this.inputEventListener=e(this.emitter.fire,this.emitter))},enumerable:!0,configurable:!0}),e.prototype.dispose=function(){this.inputEventListener.dispose(),this.emitter.dispose()}}(),Object.freeze(function(e,n){var t=setTimeout(e.bind(n),0);return{dispose:function(){clearTimeout(t)}}}));(function(e){function n(n){return n===e.None||n===e.Cancelled||(n instanceof qe||!(!n||"object"!==typeof n)&&("boolean"===typeof n.isCancellationRequested&&"function"===typeof n.onCancellationRequested))}e.isCancellationToken=n,e.None=Object.freeze({isCancellationRequested:!1,onCancellationRequested:Re.None}),e.Cancelled=Object.freeze({isCancellationRequested:!0,onCancellationRequested:He})})(Be||(Be={}));var We,qe=function(){function e(){this._isCancelled=!1,this._emitter=null}return e.prototype.cancel=function(){this._isCancelled||(this._isCancelled=!0,this._emitter&&(this._emitter.fire(void 0),this.dispose()))},Object.defineProperty(e.prototype,"isCancellationRequested",{get:function(){return this._isCancelled},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"onCancellationRequested",{get:function(){return this._isCancelled?He:(this._emitter||(this._emitter=new Ue),this._emitter.event)},enumerable:!0,configurable:!0}),e.prototype.dispose=function(){this._emitter&&(this._emitter.dispose(),this._emitter=null)},e}(),ze=function(){function e(e){this._token=void 0,this._parentListener=void 0,this._parentListener=e&&e.onCancellationRequested(this.cancel,this)}return Object.defineProperty(e.prototype,"token",{get:function(){return this._token||(this._token=new qe),this._token},enumerable:!0,configurable:!0}),e.prototype.cancel=function(){this._token?this._token instanceof qe&&this._token.cancel():this._token=Be.Cancelled},e.prototype.dispose=function(){this._parentListener&&this._parentListener.dispose(),this._token?this._token instanceof qe&&this._token.dispose():this._token=Be.None},e}(),Je=function(){function e(){this._keyCodeToStr=[],this._strToKeyCode=Object.create(null)}return e.prototype.define=function(e,n){this._keyCodeToStr[e]=n,this._strToKeyCode[n.toLowerCase()]=e},e.prototype.keyCodeToStr=function(e){return this._keyCodeToStr[e]},e.prototype.strToKeyCode=function(e){return this._strToKeyCode[e.toLowerCase()]||0},e}(),Xe=new Je,Ye=new Je,Qe=new Je;function Ze(e,n){var t=(65535&n)<<16>>>0;return(e|t)>>>0}(function(){function e(e,n,t,r){void 0===t&&(t=n),void 0===r&&(r=t),Xe.define(e,n),Ye.define(e,t),Qe.define(e,r)}e(0,"unknown"),e(1,"Backspace"),e(2,"Tab"),e(3,"Enter"),e(4,"Shift"),e(5,"Ctrl"),e(6,"Alt"),e(7,"PauseBreak"),e(8,"CapsLock"),e(9,"Escape"),e(10,"Space"),e(11,"PageUp"),e(12,"PageDown"),e(13,"End"),e(14,"Home"),e(15,"LeftArrow","Left"),e(16,"UpArrow","Up"),e(17,"RightArrow","Right"),e(18,"DownArrow","Down"),e(19,"Insert"),e(20,"Delete"),e(21,"0"),e(22,"1"),e(23,"2"),e(24,"3"),e(25,"4"),e(26,"5"),e(27,"6"),e(28,"7"),e(29,"8"),e(30,"9"),e(31,"A"),e(32,"B"),e(33,"C"),e(34,"D"),e(35,"E"),e(36,"F"),e(37,"G"),e(38,"H"),e(39,"I"),e(40,"J"),e(41,"K"),e(42,"L"),e(43,"M"),e(44,"N"),e(45,"O"),e(46,"P"),e(47,"Q"),e(48,"R"),e(49,"S"),e(50,"T"),e(51,"U"),e(52,"V"),e(53,"W"),e(54,"X"),e(55,"Y"),e(56,"Z"),e(57,"Meta"),e(58,"ContextMenu"),e(59,"F1"),e(60,"F2"),e(61,"F3"),e(62,"F4"),e(63,"F5"),e(64,"F6"),e(65,"F7"),e(66,"F8"),e(67,"F9"),e(68,"F10"),e(69,"F11"),e(70,"F12"),e(71,"F13"),e(72,"F14"),e(73,"F15"),e(74,"F16"),e(75,"F17"),e(76,"F18"),e(77,"F19"),e(78,"NumLock"),e(79,"ScrollLock"),e(80,";",";","OEM_1"),e(81,"=","=","OEM_PLUS"),e(82,",",",","OEM_COMMA"),e(83,"-","-","OEM_MINUS"),e(84,".",".","OEM_PERIOD"),e(85,"/","/","OEM_2"),e(86,"`","`","OEM_3"),e(110,"ABNT_C1"),e(111,"ABNT_C2"),e(87,"[","[","OEM_4"),e(88,"\\","\\","OEM_5"),e(89,"]","]","OEM_6"),e(90,"'","'","OEM_7"),e(91,"OEM_8"),e(92,"OEM_102"),e(93,"NumPad0"),e(94,"NumPad1"),e(95,"NumPad2"),e(96,"NumPad3"),e(97,"NumPad4"),e(98,"NumPad5"),e(99,"NumPad6"),e(100,"NumPad7"),e(101,"NumPad8"),e(102,"NumPad9"),e(103,"NumPad_Multiply"),e(104,"NumPad_Add"),e(105,"NumPad_Separator"),e(106,"NumPad_Subtract"),e(107,"NumPad_Decimal"),e(108,"NumPad_Divide")})(),function(e){function n(e){return Xe.keyCodeToStr(e)}function t(e){return Xe.strToKeyCode(e)}function r(e){return Ye.keyCodeToStr(e)}function a(e){return Qe.keyCodeToStr(e)}function i(e){return Ye.strToKeyCode(e)||Qe.strToKeyCode(e)}e.toString=n,e.fromString=t,e.toUserSettingsUS=r,e.toUserSettingsGeneral=a,e.fromUserSettings=i}(We||(We={}));(function(){function e(e,n,t,r,a){this.ctrlKey=e,this.shiftKey=n,this.altKey=t,this.metaKey=r,this.keyCode=a}e.prototype.equals=function(e){return this.ctrlKey===e.ctrlKey&&this.shiftKey===e.shiftKey&&this.altKey===e.altKey&&this.metaKey===e.metaKey&&this.keyCode===e.keyCode},e.prototype.isModifierKey=function(){return 0===this.keyCode||5===this.keyCode||57===this.keyCode||6===this.keyCode||4===this.keyCode},e.prototype.toChord=function(){return new Dn([this])},e.prototype.isDuplicateModifierCase=function(){return this.ctrlKey&&5===this.keyCode||this.shiftKey&&4===this.keyCode||this.altKey&&6===this.keyCode||this.metaKey&&57===this.keyCode}})();var $e,en,nn,tn,rn,an,on,sn,cn,ln,un,dn,pn,mn,fn,_n,gn,vn,yn,hn,bn,En,Tn,Sn,xn,Cn,An,Dn=function(){function e(e){if(0===e.length)throw l("parts");this.parts=e}return e.prototype.equals=function(e){if(null===e)return!1;if(this.parts.length!==e.parts.length)return!1;for(var n=0;n<this.parts.length;n++)if(!this.parts[n].equals(e.parts[n]))return!1;return!0},e}(),Ln=(function(){function e(e,n,t,r,a,i){this.ctrlKey=e,this.shiftKey=n,this.altKey=t,this.metaKey=r,this.keyLabel=a,this.keyAriaLabel=i}}(),function(){function e(){}}(),function(){var e=function(n,t){return e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,n){e.__proto__=n}||function(e,n){for(var t in n)n.hasOwnProperty(t)&&(e[t]=n[t])},e(n,t)};return function(n,t){function r(){this.constructor=n}e(n,t),n.prototype=null===t?Object.create(t):(r.prototype=t.prototype,new r)}}()),kn=function(e){function n(n,t,r,a){var i=e.call(this,n,t,r,a)||this;return i.selectionStartLineNumber=n,i.selectionStartColumn=t,i.positionLineNumber=r,i.positionColumn=a,i}return Ln(n,e),n.prototype.clone=function(){return new n(this.selectionStartLineNumber,this.selectionStartColumn,this.positionLineNumber,this.positionColumn)},n.prototype.toString=function(){return"["+this.selectionStartLineNumber+","+this.selectionStartColumn+" -> "+this.positionLineNumber+","+this.positionColumn+"]"},n.prototype.equalsSelection=function(e){return n.selectionsEqual(this,e)},n.selectionsEqual=function(e,n){return e.selectionStartLineNumber===n.selectionStartLineNumber&&e.selectionStartColumn===n.selectionStartColumn&&e.positionLineNumber===n.positionLineNumber&&e.positionColumn===n.positionColumn},n.prototype.getDirection=function(){return this.selectionStartLineNumber===this.startLineNumber&&this.selectionStartColumn===this.startColumn?0:1},n.prototype.setEndPosition=function(e,t){return 0===this.getDirection()?new n(this.startLineNumber,this.startColumn,e,t):new n(e,t,this.startLineNumber,this.startColumn)},n.prototype.getPosition=function(){return new ee(this.positionLineNumber,this.positionColumn)},n.prototype.setStartPosition=function(e,t){return 0===this.getDirection()?new n(e,t,this.endLineNumber,this.endColumn):new n(this.endLineNumber,this.endColumn,e,t)},n.fromPositions=function(e,t){return void 0===t&&(t=e),new n(e.lineNumber,e.column,t.lineNumber,t.column)},n.liftSelection=function(e){return new n(e.selectionStartLineNumber,e.selectionStartColumn,e.positionLineNumber,e.positionColumn)},n.selectionsArrEqual=function(e,n){if(e&&!n||!e&&n)return!1;if(!e&&!n)return!0;if(e.length!==n.length)return!1;for(var t=0,r=e.length;t<r;t++)if(!this.selectionsEqual(e[t],n[t]))return!1;return!0},n.isISelection=function(e){return e&&"number"===typeof e.selectionStartLineNumber&&"number"===typeof e.selectionStartColumn&&"number"===typeof e.positionLineNumber&&"number"===typeof e.positionColumn},n.createWithDirection=function(e,t,r,a,i){return 0===i?new n(e,t,r,a):new n(r,a,e,t)},n}(ne),Nn=function(){function e(e,n,t){this.offset=0|e,this.type=n,this.language=t}return e.prototype.toString=function(){return"("+this.offset+", "+this.type+")"},e}();(function(){function e(e,n){this.tokens=e,this.endState=n}})(),function(){function e(e,n){this.tokens=e,this.endState=n}}();(function(e){e[e["Unnecessary"]=1]="Unnecessary"})($e||($e={})),function(e){e[e["Hint"]=1]="Hint",e[e["Info"]=2]="Info",e[e["Warning"]=4]="Warning",e[e["Error"]=8]="Error"}(en||(en={})),function(e){e[e["Unknown"]=0]="Unknown",e[e["Backspace"]=1]="Backspace",e[e["Tab"]=2]="Tab",e[e["Enter"]=3]="Enter",e[e["Shift"]=4]="Shift",e[e["Ctrl"]=5]="Ctrl",e[e["Alt"]=6]="Alt",e[e["PauseBreak"]=7]="PauseBreak",e[e["CapsLock"]=8]="CapsLock",e[e["Escape"]=9]="Escape",e[e["Space"]=10]="Space",e[e["PageUp"]=11]="PageUp",e[e["PageDown"]=12]="PageDown",e[e["End"]=13]="End",e[e["Home"]=14]="Home",e[e["LeftArrow"]=15]="LeftArrow",e[e["UpArrow"]=16]="UpArrow",e[e["RightArrow"]=17]="RightArrow",e[e["DownArrow"]=18]="DownArrow",e[e["Insert"]=19]="Insert",e[e["Delete"]=20]="Delete",e[e["KEY_0"]=21]="KEY_0",e[e["KEY_1"]=22]="KEY_1",e[e["KEY_2"]=23]="KEY_2",e[e["KEY_3"]=24]="KEY_3",e[e["KEY_4"]=25]="KEY_4",e[e["KEY_5"]=26]="KEY_5",e[e["KEY_6"]=27]="KEY_6",e[e["KEY_7"]=28]="KEY_7",e[e["KEY_8"]=29]="KEY_8",e[e["KEY_9"]=30]="KEY_9",e[e["KEY_A"]=31]="KEY_A",e[e["KEY_B"]=32]="KEY_B",e[e["KEY_C"]=33]="KEY_C",e[e["KEY_D"]=34]="KEY_D",e[e["KEY_E"]=35]="KEY_E",e[e["KEY_F"]=36]="KEY_F",e[e["KEY_G"]=37]="KEY_G",e[e["KEY_H"]=38]="KEY_H",e[e["KEY_I"]=39]="KEY_I",e[e["KEY_J"]=40]="KEY_J",e[e["KEY_K"]=41]="KEY_K",e[e["KEY_L"]=42]="KEY_L",e[e["KEY_M"]=43]="KEY_M",e[e["KEY_N"]=44]="KEY_N",e[e["KEY_O"]=45]="KEY_O",e[e["KEY_P"]=46]="KEY_P",e[e["KEY_Q"]=47]="KEY_Q",e[e["KEY_R"]=48]="KEY_R",e[e["KEY_S"]=49]="KEY_S",e[e["KEY_T"]=50]="KEY_T",e[e["KEY_U"]=51]="KEY_U",e[e["KEY_V"]=52]="KEY_V",e[e["KEY_W"]=53]="KEY_W",e[e["KEY_X"]=54]="KEY_X",e[e["KEY_Y"]=55]="KEY_Y",e[e["KEY_Z"]=56]="KEY_Z",e[e["Meta"]=57]="Meta",e[e["ContextMenu"]=58]="ContextMenu",e[e["F1"]=59]="F1",e[e["F2"]=60]="F2",e[e["F3"]=61]="F3",e[e["F4"]=62]="F4",e[e["F5"]=63]="F5",e[e["F6"]=64]="F6",e[e["F7"]=65]="F7",e[e["F8"]=66]="F8",e[e["F9"]=67]="F9",e[e["F10"]=68]="F10",e[e["F11"]=69]="F11",e[e["F12"]=70]="F12",e[e["F13"]=71]="F13",e[e["F14"]=72]="F14",e[e["F15"]=73]="F15",e[e["F16"]=74]="F16",e[e["F17"]=75]="F17",e[e["F18"]=76]="F18",e[e["F19"]=77]="F19",e[e["NumLock"]=78]="NumLock",e[e["ScrollLock"]=79]="ScrollLock",e[e["US_SEMICOLON"]=80]="US_SEMICOLON",e[e["US_EQUAL"]=81]="US_EQUAL",e[e["US_COMMA"]=82]="US_COMMA",e[e["US_MINUS"]=83]="US_MINUS",e[e["US_DOT"]=84]="US_DOT",e[e["US_SLASH"]=85]="US_SLASH",e[e["US_BACKTICK"]=86]="US_BACKTICK",e[e["US_OPEN_SQUARE_BRACKET"]=87]="US_OPEN_SQUARE_BRACKET",e[e["US_BACKSLASH"]=88]="US_BACKSLASH",e[e["US_CLOSE_SQUARE_BRACKET"]=89]="US_CLOSE_SQUARE_BRACKET",e[e["US_QUOTE"]=90]="US_QUOTE",e[e["OEM_8"]=91]="OEM_8",e[e["OEM_102"]=92]="OEM_102",e[e["NUMPAD_0"]=93]="NUMPAD_0",e[e["NUMPAD_1"]=94]="NUMPAD_1",e[e["NUMPAD_2"]=95]="NUMPAD_2",e[e["NUMPAD_3"]=96]="NUMPAD_3",e[e["NUMPAD_4"]=97]="NUMPAD_4",e[e["NUMPAD_5"]=98]="NUMPAD_5",e[e["NUMPAD_6"]=99]="NUMPAD_6",e[e["NUMPAD_7"]=100]="NUMPAD_7",e[e["NUMPAD_8"]=101]="NUMPAD_8",e[e["NUMPAD_9"]=102]="NUMPAD_9",e[e["NUMPAD_MULTIPLY"]=103]="NUMPAD_MULTIPLY",e[e["NUMPAD_ADD"]=104]="NUMPAD_ADD",e[e["NUMPAD_SEPARATOR"]=105]="NUMPAD_SEPARATOR",e[e["NUMPAD_SUBTRACT"]=106]="NUMPAD_SUBTRACT",e[e["NUMPAD_DECIMAL"]=107]="NUMPAD_DECIMAL",e[e["NUMPAD_DIVIDE"]=108]="NUMPAD_DIVIDE",e[e["KEY_IN_COMPOSITION"]=109]="KEY_IN_COMPOSITION",e[e["ABNT_C1"]=110]="ABNT_C1",e[e["ABNT_C2"]=111]="ABNT_C2",e[e["MAX_VALUE"]=112]="MAX_VALUE"}(nn||(nn={})),function(e){e[e["LTR"]=0]="LTR",e[e["RTL"]=1]="RTL"}(tn||(tn={})),function(e){e[e["Auto"]=1]="Auto",e[e["Hidden"]=2]="Hidden",e[e["Visible"]=3]="Visible"}(rn||(rn={})),function(e){e[e["Left"]=1]="Left",e[e["Center"]=2]="Center",e[e["Right"]=4]="Right",e[e["Full"]=7]="Full"}(an||(an={})),function(e){e[e["TextDefined"]=0]="TextDefined",e[e["LF"]=1]="LF",e[e["CRLF"]=2]="CRLF"}(on||(on={})),function(e){e[e["LF"]=1]="LF",e[e["CRLF"]=2]="CRLF"}(sn||(sn={})),function(e){e[e["LF"]=0]="LF",e[e["CRLF"]=1]="CRLF"}(cn||(cn={})),function(e){e[e["AlwaysGrowsWhenTypingAtEdges"]=0]="AlwaysGrowsWhenTypingAtEdges",e[e["NeverGrowsWhenTypingAtEdges"]=1]="NeverGrowsWhenTypingAtEdges",e[e["GrowsOnlyWhenTypingBefore"]=2]="GrowsOnlyWhenTypingBefore",e[e["GrowsOnlyWhenTypingAfter"]=3]="GrowsOnlyWhenTypingAfter"}(ln||(ln={})),function(e){e[e["Smooth"]=0]="Smooth",e[e["Immediate"]=1]="Immediate"}(un||(un={})),function(e){e[e["NotSet"]=0]="NotSet",e[e["ContentFlush"]=1]="ContentFlush",e[e["RecoverFromMarkers"]=2]="RecoverFromMarkers",e[e["Explicit"]=3]="Explicit",e[e["Paste"]=4]="Paste",e[e["Undo"]=5]="Undo",e[e["Redo"]=6]="Redo"}(dn||(dn={})),function(e){e[e["None"]=0]="None",e[e["Small"]=1]="Small",e[e["Large"]=2]="Large",e[e["SmallBlocks"]=3]="SmallBlocks",e[e["LargeBlocks"]=4]="LargeBlocks"}(pn||(pn={})),function(e){e[e["None"]=0]="None",e[e["Same"]=1]="Same",e[e["Indent"]=2]="Indent",e[e["DeepIndent"]=3]="DeepIndent"}(mn||(mn={})),function(e){e[e["Hidden"]=0]="Hidden",e[e["Blink"]=1]="Blink",e[e["Smooth"]=2]="Smooth",e[e["Phase"]=3]="Phase",e[e["Expand"]=4]="Expand",e[e["Solid"]=5]="Solid"}(fn||(fn={})),function(e){e[e["Line"]=1]="Line",e[e["Block"]=2]="Block",e[e["Underline"]=3]="Underline",e[e["LineThin"]=4]="LineThin",e[e["BlockOutline"]=5]="BlockOutline",e[e["UnderlineThin"]=6]="UnderlineThin"}(_n||(_n={})),function(e){e[e["Off"]=0]="Off",e[e["On"]=1]="On",e[e["Relative"]=2]="Relative",e[e["Interval"]=3]="Interval",e[e["Custom"]=4]="Custom"}(gn||(gn={})),function(e){e[e["EXACT"]=0]="EXACT",e[e["ABOVE"]=1]="ABOVE",e[e["BELOW"]=2]="BELOW"}(vn||(vn={})),function(e){e[e["TOP_RIGHT_CORNER"]=0]="TOP_RIGHT_CORNER",e[e["BOTTOM_RIGHT_CORNER"]=1]="BOTTOM_RIGHT_CORNER",e[e["TOP_CENTER"]=2]="TOP_CENTER"}(yn||(yn={})),function(e){e[e["UNKNOWN"]=0]="UNKNOWN",e[e["TEXTAREA"]=1]="TEXTAREA",e[e["GUTTER_GLYPH_MARGIN"]=2]="GUTTER_GLYPH_MARGIN",e[e["GUTTER_LINE_NUMBERS"]=3]="GUTTER_LINE_NUMBERS",e[e["GUTTER_LINE_DECORATIONS"]=4]="GUTTER_LINE_DECORATIONS",e[e["GUTTER_VIEW_ZONE"]=5]="GUTTER_VIEW_ZONE",e[e["CONTENT_TEXT"]=6]="CONTENT_TEXT",e[e["CONTENT_EMPTY"]=7]="CONTENT_EMPTY",e[e["CONTENT_VIEW_ZONE"]=8]="CONTENT_VIEW_ZONE",e[e["CONTENT_WIDGET"]=9]="CONTENT_WIDGET",e[e["OVERVIEW_RULER"]=10]="OVERVIEW_RULER",e[e["SCROLLBAR"]=11]="SCROLLBAR",e[e["OVERLAY_WIDGET"]=12]="OVERLAY_WIDGET",e[e["OUTSIDE_EDITOR"]=13]="OUTSIDE_EDITOR"}(hn||(hn={})),function(e){e[e["None"]=0]="None",e[e["Indent"]=1]="Indent",e[e["IndentOutdent"]=2]="IndentOutdent",e[e["Outdent"]=3]="Outdent"}(bn||(bn={})),function(e){e[e["Method"]=0]="Method",e[e["Function"]=1]="Function",e[e["Constructor"]=2]="Constructor",e[e["Field"]=3]="Field",e[e["Variable"]=4]="Variable",e[e["Class"]=5]="Class",e[e["Struct"]=6]="Struct",e[e["Interface"]=7]="Interface",e[e["Module"]=8]="Module",e[e["Property"]=9]="Property",e[e["Event"]=10]="Event",e[e["Operator"]=11]="Operator",e[e["Unit"]=12]="Unit",e[e["Value"]=13]="Value",e[e["Constant"]=14]="Constant",e[e["Enum"]=15]="Enum",e[e["EnumMember"]=16]="EnumMember",e[e["Keyword"]=17]="Keyword",e[e["Text"]=18]="Text",e[e["Color"]=19]="Color",e[e["File"]=20]="File",e[e["Reference"]=21]="Reference",e[e["Customcolor"]=22]="Customcolor",e[e["Folder"]=23]="Folder",e[e["TypeParameter"]=24]="TypeParameter",e[e["Snippet"]=25]="Snippet"}(En||(En={})),function(e){e[e["KeepWhitespace"]=1]="KeepWhitespace",e[e["InsertAsSnippet"]=4]="InsertAsSnippet"}(Tn||(Tn={})),function(e){e[e["Invoke"]=0]="Invoke",e[e["TriggerCharacter"]=1]="TriggerCharacter",e[e["TriggerForIncompleteCompletions"]=2]="TriggerForIncompleteCompletions"}(Sn||(Sn={})),function(e){e[e["Invoke"]=1]="Invoke",e[e["TriggerCharacter"]=2]="TriggerCharacter",e[e["ContentChange"]=3]="ContentChange"}(xn||(xn={})),function(e){e[e["Text"]=0]="Text",e[e["Read"]=1]="Read",e[e["Write"]=2]="Write"}(Cn||(Cn={})),function(e){e[e["File"]=0]="File",e[e["Module"]=1]="Module",e[e["Namespace"]=2]="Namespace",e[e["Package"]=3]="Package",e[e["Class"]=4]="Class",e[e["Method"]=5]="Method",e[e["Property"]=6]="Property",e[e["Field"]=7]="Field",e[e["Constructor"]=8]="Constructor",e[e["Enum"]=9]="Enum",e[e["Interface"]=10]="Interface",e[e["Function"]=11]="Function",e[e["Variable"]=12]="Variable",e[e["Constant"]=13]="Constant",e[e["String"]=14]="String",e[e["Number"]=15]="Number",e[e["Boolean"]=16]="Boolean",e[e["Array"]=17]="Array",e[e["Object"]=18]="Object",e[e["Key"]=19]="Key",e[e["Null"]=20]="Null",e[e["EnumMember"]=21]="EnumMember",e[e["Struct"]=22]="Struct",e[e["Event"]=23]="Event",e[e["Operator"]=24]="Operator",e[e["TypeParameter"]=25]="TypeParameter"}(An||(An={}));var In=function(){function e(){}return e.chord=function(e,n){return Ze(e,n)},e.CtrlCmd=2048,e.Shift=1024,e.Alt=512,e.WinCtrl=256,e}();function wn(){return{editor:void 0,languages:void 0,CancellationTokenSource:ze,Emitter:Ue,KeyCode:nn,KeyMod:In,Position:ee,Range:ne,Selection:kn,SelectionDirection:tn,MarkerSeverity:en,MarkerTag:$e,Uri:z,Token:Nn}}var Mn=function(){var e=function(n,t){return e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,n){e.__proto__=n}||function(e,n){for(var t in n)n.hasOwnProperty(t)&&(e[t]=n[t])},e(n,t)};return function(n,t){function r(){this.constructor=n}e(n,t),n.prototype=null===t?Object.create(t):(r.prototype=t.prototype,new r)}}(),On=function(e){function n(){return null!==e&&e.apply(this,arguments)||this}return Mn(n,e),Object.defineProperty(n.prototype,"uri",{get:function(){return this._uri},enumerable:!0,configurable:!0}),Object.defineProperty(n.prototype,"version",{get:function(){return this._versionId},enumerable:!0,configurable:!0}),Object.defineProperty(n.prototype,"eol",{get:function(){return this._eol},enumerable:!0,configurable:!0}),n.prototype.getValue=function(){return this.getText()},n.prototype.getLinesContent=function(){return this._lines.slice(0)},n.prototype.getLineCount=function(){return this._lines.length},n.prototype.getLineContent=function(e){return this._lines[e-1]},n.prototype.getWordAtPosition=function(e,n){var t=Ce(e.column,Te(n),this._lines[e.lineNumber-1],0);return t?new ne(e.lineNumber,t.startColumn,e.lineNumber,t.endColumn):null},n.prototype.getWordUntilPosition=function(e,n){var t=this.getWordAtPosition(e,n);return t?{word:this._lines[e.lineNumber-1].substring(t.startColumn-1,e.column-1),startColumn:t.startColumn,endColumn:e.column}:{word:"",startColumn:e.column,endColumn:e.column}},n.prototype.createWordIterator=function(e){var n,t,r=this,a=0,i=0,o=[],s=function(){if(i<o.length){var c=t.substring(o[i].start,o[i].end);return i+=1,n?n.value=c:n={done:!1,value:c},n}return a>=r._lines.length?M:(t=r._lines[a],o=r._wordenize(t,e),i=0,a+=1,s())};return{next:s}},n.prototype.getLineWords=function(e,n){for(var t=this._lines[e-1],r=this._wordenize(t,n),a=[],i=0,o=r;i<o.length;i++){var s=o[i];a.push({word:t.substring(s.start,s.end),startColumn:s.start+1,endColumn:s.end+1})}return a},n.prototype._wordenize=function(e,n){var t,r=[];n.lastIndex=0;while(t=n.exec(e)){if(0===t[0].length)break;r.push({start:t.index,end:t.index+t[0].length})}return r},n.prototype.getValueInRange=function(e){if(e=this._validateRange(e),e.startLineNumber===e.endLineNumber)return this._lines[e.startLineNumber-1].substring(e.startColumn-1,e.endColumn-1);var n=this._eol,t=e.startLineNumber-1,r=e.endLineNumber-1,a=[];a.push(this._lines[t].substring(e.startColumn-1));for(var i=t+1;i<r;i++)a.push(this._lines[i]);return a.push(this._lines[r].substring(0,e.endColumn-1)),a.join(n)},n.prototype.offsetAt=function(e){return e=this._validatePosition(e),this._ensureLineStarts(),this._lineStarts.getAccumulatedValue(e.lineNumber-2)+(e.column-1)},n.prototype.positionAt=function(e){e=Math.floor(e),e=Math.max(0,e),this._ensureLineStarts();var n=this._lineStarts.getIndexOf(e),t=this._lines[n.index].length;return{lineNumber:1+n.index,column:1+Math.min(n.remainder,t)}},n.prototype._validateRange=function(e){var n=this._validatePosition({lineNumber:e.startLineNumber,column:e.startColumn}),t=this._validatePosition({lineNumber:e.endLineNumber,column:e.endColumn});return n.lineNumber!==e.startLineNumber||n.column!==e.startColumn||t.lineNumber!==e.endLineNumber||t.column!==e.endColumn?{startLineNumber:n.lineNumber,startColumn:n.column,endLineNumber:t.lineNumber,endColumn:t.column}:e},n.prototype._validatePosition=function(e){if(!ee.isIPosition(e))throw new Error("bad position");var n=e.lineNumber,t=e.column,r=!1;if(n<1)n=1,t=1,r=!0;else if(n>this._lines.length)n=this._lines.length,t=this._lines[n-1].length+1,r=!0;else{var a=this._lines[n-1].length+1;t<1?(t=1,r=!0):t>a&&(t=a,r=!0)}return r?{lineNumber:n,column:t}:e},n}(ye),Pn=function(){function e(e){this._foreignModuleFactory=e,this._foreignModule=null}return e.prototype.computeDiff=function(e,n,t){var r=this._getModel(e),a=this._getModel(n);if(!r||!a)return Promise.resolve(null);var i=r.getLinesContent(),o=a.getLinesContent(),s=new pe(i,o,{shouldComputeCharChanges:!0,shouldPostProcessCharChanges:!0,shouldIgnoreTrimWhitespace:t,shouldMakePrettyDiff:!0}),c=s.computeDiff(),l=!(c.length>0)&&this._modelsAreIdentical(r,a);return Promise.resolve({identical:l,changes:c})},e.prototype._modelsAreIdentical=function(e,n){var t=e.getLineCount(),r=n.getLineCount();if(t!==r)return!1;for(var a=1;a<=t;a++){var i=e.getLineContent(a),o=n.getLineContent(a);if(i!==o)return!1}return!0},e.prototype.computeMoreMinimalEdits=function(n,t){var r=this._getModel(n);if(!r)return Promise.resolve(t);var a=[],i=void 0;t=b(t,function(e,n){if(e.range&&n.range)return ne.compareRangesUsingStarts(e.range,n.range);var t=e.range?0:1,r=n.range?0:1;return t-r});for(var o=0,s=t;o<s.length;o++){var c=s[o],l=c.range,u=c.text,d=c.eol;if("number"===typeof d&&(i=d),!ne.isEmpty(l)||u){var p=r.getValueInRange(l);if(u=u.replace(/\r\n|\n|\r/g,r.eol),p!==u)if(Math.max(u.length,p.length)>e._diffLimit)a.push({range:l,text:u});else for(var m=C(p,u,!1),f=r.offsetAt(ne.lift(l).getStartPosition()),_=0,g=m;_<g.length;_++){var v=g[_],y=r.positionAt(f+v.originalStart),h=r.positionAt(f+v.originalStart+v.originalLength),E={text:u.substr(v.modifiedStart,v.modifiedLength),range:{startLineNumber:y.lineNumber,startColumn:y.column,endLineNumber:h.lineNumber,endColumn:h.column}};r.getValueInRange(E.range)!==E.text&&a.push(E)}}}return"number"===typeof i&&a.push({eol:i,text:"",range:{startLineNumber:0,startColumn:0,endLineNumber:0,endColumn:0}}),Promise.resolve(a)},e.prototype.computeLinks=function(e){var n=this._getModel(e);return n?Promise.resolve(Me(n)):Promise.resolve(null)},e.prototype.textualSuggest=function(n,t,r,a){var i=this._getModel(n);if(!i)return Promise.resolve(null);var o=Object.create(null),s=[],c=new RegExp(r,a),l=i.getWordUntilPosition(t,c),u=i.getWordAtPosition(t,c);u&&(o[i.getValueInRange(u)]=!0);for(var d=i.createWordIterator(c),p=d.next();!p.done&&s.length<=e._suggestionsLimit;p=d.next()){var m=p.value;o[m]||(o[m]=!0,isNaN(Number(m))&&s.push({kind:18,label:m,insertText:m,range:{startLineNumber:t.lineNumber,startColumn:l.startColumn,endLineNumber:t.lineNumber,endColumn:l.endColumn}}))}return Promise.resolve({suggestions:s})},e.prototype.computeWordRanges=function(e,n,t,r){var a=this._getModel(e);if(!a)return Promise.resolve(Object.create(null));for(var i=new RegExp(t,r),o=Object.create(null),s=n.startLineNumber;s<n.endLineNumber;s++)for(var c=a.getLineWords(s,i),l=0,u=c;l<u.length;l++){var d=u[l];if(isNaN(Number(d.word))){var p=o[d.word];p||(p=[],o[d.word]=p),p.push({startLineNumber:s,startColumn:d.startColumn,endLineNumber:s,endColumn:d.endColumn})}}return Promise.resolve(o)},e.prototype.navigateValueSet=function(e,n,t,r,a){var i=this._getModel(e);if(!i)return Promise.resolve(null);var o=new RegExp(r,a);n.startColumn===n.endColumn&&(n={startLineNumber:n.startLineNumber,startColumn:n.startColumn,endLineNumber:n.endLineNumber,endColumn:n.endColumn+1});var s=i.getValueInRange(n),c=i.getWordAtPosition({lineNumber:n.startLineNumber,column:n.startColumn},o);if(!c)return Promise.resolve(null);var l=i.getValueInRange(c),u=Oe.INSTANCE.navigateValueSet(n,s,c,l,t);return Promise.resolve(u)},e.prototype.loadForeignModule=function(e,n){var t=this,r={getMirrorModels:function(){return t._getModels()}};if(this._foreignModuleFactory){this._foreignModule=this._foreignModuleFactory(r,n);for(var a=[],i=0,o=_(this._foreignModule);i<o.length;i++){var s=o[i];"function"===typeof this._foreignModule[s]&&a.push(s)}return Promise.resolve(a)}return Promise.reject(new Error("Unexpected usage"))},e.prototype.fmr=function(e,n){if(!this._foreignModule||"function"!==typeof this._foreignModule[e])return Promise.reject(new Error("Missing requestHandler or method: "+e));try{return Promise.resolve(this._foreignModule[e].apply(this._foreignModule,n))}catch(t){return Promise.reject(t)}},e._diffLimit=1e5,e._suggestionsLimit=1e4,e}(),Rn=function(e){function n(n){var t=e.call(this,n)||this;return t._models=Object.create(null),t}return Mn(n,e),n.prototype.dispose=function(){this._models=Object.create(null)},n.prototype._getModel=function(e){return this._models[e]},n.prototype._getModels=function(){var e=this,n=[];return Object.keys(this._models).forEach(function(t){return n.push(e._models[t])}),n},n.prototype.acceptNewModel=function(e){this._models[e.url]=new On(z.parse(e.url),e.lines,e.EOL,e.versionId)},n.prototype.acceptModelChanged=function(e,n){if(this._models[e]){var t=this._models[e];t.onEvents(n)}},n.prototype.acceptRemovedModel=function(e){this._models[e]&&delete this._models[e]},n}(Pn);"function"===typeof importScripts&&(f["a"].monaco=wn());var Fn=!1;function Gn(e){if(!Fn){Fn=!0;var n=new Rn(e),t=new h(function(e){self.postMessage(e)},n);self.onmessage=function(e){t.onmessage(e.data)}}}self.onmessage=function(e){Fn||Gn(null)};var Vn=t("973d"),Bn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface SymbolConstructor {\n    /**\n     * A reference to the prototype.\n     */\n    readonly prototype: Symbol;\n\n    /**\n     * Returns a new unique Symbol value.\n     * @param  description Description of the new Symbol object.\n     */\n    (description?: string | number): symbol;\n\n    /**\n     * Returns a Symbol object from the global symbol registry matching the given key if found.\n     * Otherwise, returns a new symbol with this key.\n     * @param key key to search for.\n     */\n    for(key: string): symbol;\n\n    /**\n     * Returns a key from the global symbol registry matching the given Symbol if found.\n     * Otherwise, returns a undefined.\n     * @param sym Symbol to find the key for.\n     */\n    keyFor(sym: symbol): string | undefined;\n}\n\ndeclare var Symbol: SymbolConstructor;',Kn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n'+Bn+'\ninterface SymbolConstructor {\n    /**\n     * A method that determines if a constructor object recognizes an object as one of the\n     * constructor’s instances. Called by the semantics of the instanceof operator.\n     */\n    readonly hasInstance: symbol;\n\n    /**\n     * A Boolean value that if true indicates that an object should flatten to its array elements\n     * by Array.prototype.concat.\n     */\n    readonly isConcatSpreadable: symbol;\n\n    /**\n     * A regular expression method that matches the regular expression against a string. Called\n     * by the String.prototype.match method.\n     */\n    readonly match: symbol;\n\n    /**\n     * A regular expression method that replaces matched substrings of a string. Called by the\n     * String.prototype.replace method.\n     */\n    readonly replace: symbol;\n\n    /**\n     * A regular expression method that returns the index within a string that matches the\n     * regular expression. Called by the String.prototype.search method.\n     */\n    readonly search: symbol;\n\n    /**\n     * A function valued property that is the constructor function that is used to create\n     * derived objects.\n     */\n    readonly species: symbol;\n\n    /**\n     * A regular expression method that splits a string at the indices that match the regular\n     * expression. Called by the String.prototype.split method.\n     */\n    readonly split: symbol;\n\n    /**\n     * A method that converts an object to a corresponding primitive value.\n     * Called by the ToPrimitive abstract operation.\n     */\n    readonly toPrimitive: symbol;\n\n    /**\n     * A String value that is used in the creation of the default string description of an object.\n     * Called by the built-in method Object.prototype.toString.\n     */\n    readonly toStringTag: symbol;\n\n    /**\n     * An Object whose own property names are property names that are excluded from the \'with\'\n     * environment bindings of the associated objects.\n     */\n    readonly unscopables: symbol;\n}\n\ninterface Symbol {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Array<T> {\n    /**\n     * Returns an object whose properties have the value \'true\'\n     * when they will be absent when used in a \'with\' statement.\n     */\n    [Symbol.unscopables](): {\n        copyWithin: boolean;\n        entries: boolean;\n        fill: boolean;\n        find: boolean;\n        findIndex: boolean;\n        keys: boolean;\n        values: boolean;\n    };\n}\n\ninterface Date {\n    /**\n     * Converts a Date object to a string.\n     */\n    [Symbol.toPrimitive](hint: "default"): string;\n    /**\n     * Converts a Date object to a string.\n     */\n    [Symbol.toPrimitive](hint: "string"): string;\n    /**\n     * Converts a Date object to a number.\n     */\n    [Symbol.toPrimitive](hint: "number"): number;\n    /**\n     * Converts a Date object to a string or number.\n     *\n     * @param hint The strings "number", "string", or "default" to specify what primitive to return.\n     *\n     * @throws {TypeError} If \'hint\' was given something other than "number", "string", or "default".\n     * @returns A number if \'hint\' was "number", a string if \'hint\' was "string" or "default".\n     */\n    [Symbol.toPrimitive](hint: string): string | number;\n}\n\ninterface Map<K, V> {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface WeakMap<K extends object, V> {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Set<T> {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface WeakSet<T extends object> {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface JSON {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Function {\n    /**\n     * Determines whether the given value inherits from this function if this function was used\n     * as a constructor function.\n     *\n     * A constructor function can control which objects are recognized as its instances by\n     * \'instanceof\' by overriding this method.\n     */\n    [Symbol.hasInstance](value: any): boolean;\n}\n\ninterface GeneratorFunction {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Math {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Promise<T> {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface PromiseConstructor {\n    readonly [Symbol.species]: PromiseConstructor;\n}\n\ninterface RegExp {\n    /**\n     * Matches a string with this regular expression, and returns an array containing the results of\n     * that search.\n     * @param string A string to search within.\n     */\n    [Symbol.match](string: string): RegExpMatchArray | null;\n\n    /**\n     * Replaces text in a string, using this regular expression.\n     * @param string A String object or string literal whose contents matching against\n     *               this regular expression will be replaced\n     * @param replaceValue A String object or string literal containing the text to replace for every\n     *                     successful match of this regular expression.\n     */\n    [Symbol.replace](string: string, replaceValue: string): string;\n\n    /**\n     * Replaces text in a string, using this regular expression.\n     * @param string A String object or string literal whose contents matching against\n     *               this regular expression will be replaced\n     * @param replacer A function that returns the replacement text.\n     */\n    [Symbol.replace](string: string, replacer: (substring: string, ...args: any[]) => string): string;\n\n    /**\n     * Finds the position beginning first substring match in a regular expression search\n     * using this regular expression.\n     *\n     * @param string The string to search within.\n     */\n    [Symbol.search](string: string): number;\n\n    /**\n     * Returns an array of substrings that were delimited by strings in the original input that\n     * match against this regular expression.\n     *\n     * If the regular expression contains capturing parentheses, then each time this\n     * regular expression matches, the results (including any undefined results) of the\n     * capturing parentheses are spliced.\n     *\n     * @param string string value to split\n     * @param limit if not undefined, the output array is truncated so that it contains no more\n     * than \'limit\' elements.\n     */\n    [Symbol.split](string: string, limit?: number): string[];\n}\n\ninterface RegExpConstructor {\n    readonly [Symbol.species]: RegExpConstructor;\n}\n\ninterface String {\n    /**\n     * Matches a string an object that supports being matched against, and returns an array containing the results of that search.\n     * @param matcher An object that supports being matched against.\n     */\n    match(matcher: { [Symbol.match](string: string): RegExpMatchArray | null; }): RegExpMatchArray | null;\n\n    /**\n     * Replaces text in a string, using an object that supports replacement within a string.\n     * @param searchValue A object can search for and replace matches within a string.\n     * @param replaceValue A string containing the text to replace for every successful match of searchValue in this string.\n     */\n    replace(searchValue: { [Symbol.replace](string: string, replaceValue: string): string; }, replaceValue: string): string;\n\n    /**\n     * Replaces text in a string, using an object that supports replacement within a string.\n     * @param searchValue A object can search for and replace matches within a string.\n     * @param replacer A function that returns the replacement text.\n     */\n    replace(searchValue: { [Symbol.replace](string: string, replacer: (substring: string, ...args: any[]) => string): string; }, replacer: (substring: string, ...args: any[]) => string): string;\n\n    /**\n     * Finds the first substring match in a regular expression search.\n     * @param searcher An object which supports searching within a string.\n     */\n    search(searcher: { [Symbol.search](string: string): number; }): number;\n\n    /**\n     * Split a string into substrings using the specified separator and return them as an array.\n     * @param splitter An object that can split a string.\n     * @param limit A value used to limit the number of elements returned in the array.\n     */\n    split(splitter: { [Symbol.split](string: string, limit?: number): string[]; }, limit?: number): string[];\n}\n\ninterface ArrayBuffer {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface DataView {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Int8Array {\n    readonly [Symbol.toStringTag]: "Int8Array";\n}\n\ninterface Uint8Array {\n    readonly [Symbol.toStringTag]: "UInt8Array";\n}\n\ninterface Uint8ClampedArray {\n    readonly [Symbol.toStringTag]: "Uint8ClampedArray";\n}\n\ninterface Int16Array {\n    readonly [Symbol.toStringTag]: "Int16Array";\n}\n\ninterface Uint16Array {\n    readonly [Symbol.toStringTag]: "Uint16Array";\n}\n\ninterface Int32Array {\n    readonly [Symbol.toStringTag]: "Int32Array";\n}\n\ninterface Uint32Array {\n    readonly [Symbol.toStringTag]: "Uint32Array";\n}\n\ninterface Float32Array {\n    readonly [Symbol.toStringTag]: "Float32Array";\n}\n\ninterface Float64Array {\n    readonly [Symbol.toStringTag]: "Float64Array";\n}\n\ninterface ArrayConstructor {\n    readonly [Symbol.species]: ArrayConstructor;\n}\ninterface MapConstructor {\n    readonly [Symbol.species]: MapConstructor;\n}\ninterface SetConstructor {\n    readonly [Symbol.species]: SetConstructor;\n}\ninterface ArrayBufferConstructor {\n    readonly [Symbol.species]: ArrayBufferConstructor;\n}',jn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ndeclare namespace Reflect {\n    function apply(target: Function, thisArgument: any, argumentsList: ArrayLike<any>): any;\n    function construct(target: Function, argumentsList: ArrayLike<any>, newTarget?: any): any;\n    function defineProperty(target: object, propertyKey: PropertyKey, attributes: PropertyDescriptor): boolean;\n    function deleteProperty(target: object, propertyKey: PropertyKey): boolean;\n    function get(target: object, propertyKey: PropertyKey, receiver?: any): any;\n    function getOwnPropertyDescriptor(target: object, propertyKey: PropertyKey): PropertyDescriptor | undefined;\n    function getPrototypeOf(target: object): object;\n    function has(target: object, propertyKey: PropertyKey): boolean;\n    function isExtensible(target: object): boolean;\n    function ownKeys(target: object): PropertyKey[];\n    function preventExtensions(target: object): boolean;\n    function set(target: object, propertyKey: PropertyKey, value: any, receiver?: any): boolean;\n    function setPrototypeOf(target: object, proto: any): boolean;\n}\n',Un='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface ProxyHandler<T extends object> {\n    getPrototypeOf? (target: T): object | null;\n    setPrototypeOf? (target: T, v: any): boolean;\n    isExtensible? (target: T): boolean;\n    preventExtensions? (target: T): boolean;\n    getOwnPropertyDescriptor? (target: T, p: PropertyKey): PropertyDescriptor | undefined;\n    has? (target: T, p: PropertyKey): boolean;\n    get? (target: T, p: PropertyKey, receiver: any): any;\n    set? (target: T, p: PropertyKey, value: any, receiver: any): boolean;\n    deleteProperty? (target: T, p: PropertyKey): boolean;\n    defineProperty? (target: T, p: PropertyKey, attributes: PropertyDescriptor): boolean;\n    enumerate? (target: T): PropertyKey[];\n    ownKeys? (target: T): PropertyKey[];\n    apply? (target: T, thisArg: any, argArray?: any): any;\n    construct? (target: T, argArray: any, newTarget?: any): object;\n}\n\ninterface ProxyConstructor {\n    revocable<T extends object>(target: T, handler: ProxyHandler<T>): { proxy: T; revoke: () => void; };\n    new <T extends object>(target: T, handler: ProxyHandler<T>): T;\n}\ndeclare var Proxy: ProxyConstructor;\n',Hn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n'+Bn+"\ninterface SymbolConstructor {\n    /**\n     * A method that returns the default iterator for an object. Called by the semantics of the\n     * for-of statement.\n     */\n    readonly iterator: symbol;\n}\n\ninterface IteratorResult<T> {\n    done: boolean;\n    value: T;\n}\n\ninterface Iterator<T> {\n    next(value?: any): IteratorResult<T>;\n    return?(value?: any): IteratorResult<T>;\n    throw?(e?: any): IteratorResult<T>;\n}\n\ninterface Iterable<T> {\n    [Symbol.iterator](): Iterator<T>;\n}\n\ninterface IterableIterator<T> extends Iterator<T> {\n    [Symbol.iterator](): IterableIterator<T>;\n}\n\ninterface Array<T> {\n    /** Iterator */\n    [Symbol.iterator](): IterableIterator<T>;\n\n    /**\n     * Returns an iterable of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, T]>;\n\n    /**\n     * Returns an iterable of keys in the array\n     */\n    keys(): IterableIterator<number>;\n\n    /**\n     * Returns an iterable of values in the array\n     */\n    values(): IterableIterator<T>;\n}\n\ninterface ArrayConstructor {\n    /**\n     * Creates an array from an iterable object.\n     * @param iterable An iterable object to convert to an array.\n     */\n    from<T>(iterable: Iterable<T> | ArrayLike<T>): T[];\n\n    /**\n     * Creates an array from an iterable object.\n     * @param iterable An iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from<T, U>(iterable: Iterable<T> | ArrayLike<T>, mapfn: (v: T, k: number) => U, thisArg?: any): U[];\n}\n\ninterface ReadonlyArray<T> {\n    /** Iterator of values in the array. */\n    [Symbol.iterator](): IterableIterator<T>;\n\n    /**\n     * Returns an iterable of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, T]>;\n\n    /**\n     * Returns an iterable of keys in the array\n     */\n    keys(): IterableIterator<number>;\n\n    /**\n     * Returns an iterable of values in the array\n     */\n    values(): IterableIterator<T>;\n}\n\ninterface IArguments {\n    /** Iterator */\n    [Symbol.iterator](): IterableIterator<any>;\n}\n\ninterface Map<K, V> {\n    /** Returns an iterable of entries in the map. */\n    [Symbol.iterator](): IterableIterator<[K, V]>;\n\n    /**\n     * Returns an iterable of key, value pairs for every entry in the map.\n     */\n    entries(): IterableIterator<[K, V]>;\n\n    /**\n     * Returns an iterable of keys in the map\n     */\n    keys(): IterableIterator<K>;\n\n    /**\n     * Returns an iterable of values in the map\n     */\n    values(): IterableIterator<V>;\n}\n\ninterface ReadonlyMap<K, V> {\n    /** Returns an iterable of entries in the map. */\n    [Symbol.iterator](): IterableIterator<[K, V]>;\n\n    /**\n     * Returns an iterable of key, value pairs for every entry in the map.\n     */\n    entries(): IterableIterator<[K, V]>;\n\n    /**\n     * Returns an iterable of keys in the map\n     */\n    keys(): IterableIterator<K>;\n\n    /**\n     * Returns an iterable of values in the map\n     */\n    values(): IterableIterator<V>;\n}\n\ninterface MapConstructor {\n    new <K, V>(iterable: Iterable<[K, V]>): Map<K, V>;\n}\n\ninterface WeakMap<K extends object, V> { }\n\ninterface WeakMapConstructor {\n    new <K extends object, V>(iterable: Iterable<[K, V]>): WeakMap<K, V>;\n}\n\ninterface Set<T> {\n    /** Iterates over values in the set. */\n    [Symbol.iterator](): IterableIterator<T>;\n    /**\n     * Returns an iterable of [v,v] pairs for every value `v` in the set.\n     */\n    entries(): IterableIterator<[T, T]>;\n    /**\n     * Despite its name, returns an iterable of the values in the set,\n     */\n    keys(): IterableIterator<T>;\n\n    /**\n     * Returns an iterable of values in the set.\n     */\n    values(): IterableIterator<T>;\n}\n\ninterface ReadonlySet<T> {\n    /** Iterates over values in the set. */\n    [Symbol.iterator](): IterableIterator<T>;\n\n    /**\n     * Returns an iterable of [v,v] pairs for every value `v` in the set.\n     */\n    entries(): IterableIterator<[T, T]>;\n\n    /**\n     * Despite its name, returns an iterable of the values in the set,\n     */\n    keys(): IterableIterator<T>;\n\n    /**\n     * Returns an iterable of values in the set.\n     */\n    values(): IterableIterator<T>;\n}\n\ninterface SetConstructor {\n    new <T>(iterable: Iterable<T>): Set<T>;\n}\n\ninterface WeakSet<T extends object> { }\n\ninterface WeakSetConstructor {\n    new <T extends object = object>(iterable: Iterable<T>): WeakSet<T>;\n}\n\ninterface Promise<T> { }\n\ninterface PromiseConstructor {\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<TAll>(values: Iterable<TAll | PromiseLike<TAll>>): Promise<TAll[]>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T>(values: Iterable<T | PromiseLike<T>>): Promise<T>;\n}\n\ndeclare namespace Reflect {\n    function enumerate(target: object): IterableIterator<any>;\n}\n\ninterface String {\n    /** Iterator */\n    [Symbol.iterator](): IterableIterator<string>;\n}\n\ninterface Int8Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Int8ArrayConstructor {\n    new (elements: Iterable<number>): Int8Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Int8Array;\n}\n\ninterface Uint8Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Uint8ArrayConstructor {\n    new (elements: Iterable<number>): Uint8Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Uint8Array;\n}\n\ninterface Uint8ClampedArray {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Uint8ClampedArrayConstructor {\n    new (elements: Iterable<number>): Uint8ClampedArray;\n\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Uint8ClampedArray;\n}\n\ninterface Int16Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Int16ArrayConstructor {\n    new (elements: Iterable<number>): Int16Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Int16Array;\n}\n\ninterface Uint16Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Uint16ArrayConstructor {\n    new (elements: Iterable<number>): Uint16Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Uint16Array;\n}\n\ninterface Int32Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Int32ArrayConstructor {\n    new (elements: Iterable<number>): Int32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Int32Array;\n}\n\ninterface Uint32Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Uint32ArrayConstructor {\n    new (elements: Iterable<number>): Uint32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Uint32Array;\n}\n\ninterface Float32Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Float32ArrayConstructor {\n    new (elements: Iterable<number>): Float32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Float32Array;\n}\n\ninterface Float64Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Float64ArrayConstructor {\n    new (elements: Iterable<number>): Float64Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Float64Array;\n}\n",Wn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface PromiseConstructor {\n    /**\n     * A reference to the prototype.\n     */\n    readonly prototype: Promise<any>;\n\n    /**\n     * Creates a new Promise.\n     * @param executor A callback used to initialize the promise. This callback is passed two arguments:\n     * a resolve callback used to resolve the promise with a value or the result of another promise,\n     * and a reject callback used to reject the promise with a provided reason or error.\n     */\n    new <T>(executor: (resolve: (value?: T | PromiseLike<T>) => void, reject: (reason?: any) => void) => void): Promise<T>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5, T6, T7, T8, T9, T10>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike <T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>, T8 | PromiseLike<T8>, T9 | PromiseLike<T9>, T10 | PromiseLike<T10>]): Promise<[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5, T6, T7, T8, T9>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike <T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>, T8 | PromiseLike<T8>, T9 | PromiseLike<T9>]): Promise<[T1, T2, T3, T4, T5, T6, T7, T8, T9]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5, T6, T7, T8>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike <T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>, T8 | PromiseLike<T8>]): Promise<[T1, T2, T3, T4, T5, T6, T7, T8]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5, T6, T7>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike <T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>]): Promise<[T1, T2, T3, T4, T5, T6, T7]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5, T6>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike <T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>]): Promise<[T1, T2, T3, T4, T5, T6]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike <T4>, T5 | PromiseLike<T5>]): Promise<[T1, T2, T3, T4, T5]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike <T4>]): Promise<[T1, T2, T3, T4]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>]): Promise<[T1, T2, T3]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>]): Promise<[T1, T2]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T>(values: (T | PromiseLike<T>)[]): Promise<T[]>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T1, T2, T3, T4, T5, T6, T7, T8, T9, T10>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>, T8 | PromiseLike<T8>, T9 | PromiseLike<T9>, T10 | PromiseLike<T10>]): Promise<T1 | T2 | T3 | T4 | T5 | T6 | T7 | T8 | T9 | T10>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T1, T2, T3, T4, T5, T6, T7, T8, T9>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>, T8 | PromiseLike<T8>, T9 | PromiseLike<T9>]): Promise<T1 | T2 | T3 | T4 | T5 | T6 | T7 | T8 | T9>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T1, T2, T3, T4, T5, T6, T7, T8>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>, T8 | PromiseLike<T8>]): Promise<T1 | T2 | T3 | T4 | T5 | T6 | T7 | T8>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T1, T2, T3, T4, T5, T6, T7>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>]): Promise<T1 | T2 | T3 | T4 | T5 | T6 | T7>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T1, T2, T3, T4, T5, T6>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>]): Promise<T1 | T2 | T3 | T4 | T5 | T6>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T1, T2, T3, T4, T5>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>]): Promise<T1 | T2 | T3 | T4 | T5>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T1, T2, T3, T4>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>]): Promise<T1 | T2 | T3 | T4>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T1, T2, T3>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>]): Promise<T1 | T2 | T3>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T1, T2>(values: [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>]): Promise<T1 | T2>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T>(values: (T | PromiseLike<T>)[]): Promise<T>;\n\n    /**\n     * Creates a new rejected promise for the provided reason.\n     * @param reason The reason the promise was rejected.\n     * @returns A new rejected Promise.\n     */\n    reject<T = never>(reason?: any): Promise<T>;\n\n    /**\n     * Creates a new resolved promise for the provided value.\n     * @param value A promise.\n     * @returns A promise whose internal state matches the provided promise.\n     */\n    resolve<T>(value: T | PromiseLike<T>): Promise<T>;\n\n    /**\n     * Creates a new resolved promise .\n     * @returns A resolved promise.\n     */\n    resolve(): Promise<void>;\n}\n\ndeclare var Promise: PromiseConstructor;\n',qn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface Generator extends Iterator<any> { }\n\ninterface GeneratorFunction {\n    /**\n     * Creates a new Generator object.\n     * @param args A list of arguments the function accepts.\n     */\n    new (...args: any[]): Generator;\n    /**\n     * Creates a new Generator object.\n     * @param args A list of arguments the function accepts.\n     */\n    (...args: any[]): Generator;\n    /**\n     * The length of the arguments.\n     */\n    readonly length: number;\n    /**\n     * Returns the name of the function.\n     */\n    readonly name: string;\n    /**\n     * A reference to the prototype.\n     */\n    readonly prototype: Generator;\n}\n\ninterface GeneratorFunctionConstructor {\n    /**\n     * Creates a new Generator function.\n     * @param args A list of arguments the function accepts.\n     */\n    new (...args: string[]): GeneratorFunction;\n    /**\n     * Creates a new Generator function.\n     * @param args A list of arguments the function accepts.\n     */\n    (...args: string[]): GeneratorFunction;\n    /**\n     * The length of the arguments.\n     */\n    readonly length: number;\n    /**\n     * Returns the name of the function.\n     */\n    readonly name: string;\n    /**\n     * A reference to the prototype.\n     */\n    readonly prototype: GeneratorFunction;\n}\n',zn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface Map<K, V> {\n    clear(): void;\n    delete(key: K): boolean;\n    forEach(callbackfn: (value: V, key: K, map: Map<K, V>) => void, thisArg?: any): void;\n    get(key: K): V | undefined;\n    has(key: K): boolean;\n    set(key: K, value: V): this;\n    readonly size: number;\n}\n\ninterface MapConstructor {\n    new(): Map<any, any>;\n    new<K, V>(entries?: ReadonlyArray<[K, V]> | null): Map<K, V>;\n    readonly prototype: Map<any, any>;\n}\ndeclare var Map: MapConstructor;\n\ninterface ReadonlyMap<K, V> {\n    forEach(callbackfn: (value: V, key: K, map: ReadonlyMap<K, V>) => void, thisArg?: any): void;\n    get(key: K): V | undefined;\n    has(key: K): boolean;\n    readonly size: number;\n}\n\ninterface WeakMap<K extends object, V> {\n    delete(key: K): boolean;\n    get(key: K): V | undefined;\n    has(key: K): boolean;\n    set(key: K, value: V): this;\n}\n\ninterface WeakMapConstructor {\n    new <K extends object = object, V = any>(entries?: ReadonlyArray<[K, V]> | null): WeakMap<K, V>;\n    readonly prototype: WeakMap<object, any>;\n}\ndeclare var WeakMap: WeakMapConstructor;\n\ninterface Set<T> {\n    add(value: T): this;\n    clear(): void;\n    delete(value: T): boolean;\n    forEach(callbackfn: (value: T, value2: T, set: Set<T>) => void, thisArg?: any): void;\n    has(value: T): boolean;\n    readonly size: number;\n}\n\ninterface SetConstructor {\n    new <T = any>(values?: ReadonlyArray<T> | null): Set<T>;\n    readonly prototype: Set<any>;\n}\ndeclare var Set: SetConstructor;\n\ninterface ReadonlySet<T> {\n    forEach(callbackfn: (value: T, value2: T, set: ReadonlySet<T>) => void, thisArg?: any): void;\n    has(value: T): boolean;\n    readonly size: number;\n}\n\ninterface WeakSet<T extends object> {\n    add(value: T): this;\n    delete(value: T): boolean;\n    has(value: T): boolean;\n}\n\ninterface WeakSetConstructor {\n    new <T extends object = object>(values?: ReadonlyArray<T> | null): WeakSet<T>;\n    readonly prototype: WeakSet<object>;\n}\ndeclare var WeakSet: WeakSetConstructor;\n',Jn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface Array<T> {\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find<S extends T>(predicate: (this: void, value: T, index: number, obj: T[]) => value is S, thisArg?: any): S | undefined;\n    find(predicate: (value: T, index: number, obj: T[]) => boolean, thisArg?: any): T | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: T, index: number, obj: T[]) => boolean, thisArg?: any): number;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: T, start?: number, end?: number): this;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n}\n\ninterface ArrayConstructor {\n    /**\n     * Creates an array from an array-like object.\n     * @param arrayLike An array-like object to convert to an array.\n     */\n    from<T>(arrayLike: ArrayLike<T>): T[];\n\n    /**\n     * Creates an array from an iterable object.\n     * @param arrayLike An array-like object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of \'this\' used to invoke the mapfn.\n     */\n    from<T, U>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => U, thisArg?: any): U[];\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of<T>(...items: T[]): T[];\n}\n\ninterface DateConstructor {\n    new (value: number | string | Date): Date;\n}\n\ninterface Function {\n    /**\n     * Returns the name of the function. Function names are read-only and can not be changed.\n     */\n    readonly name: string;\n}\n\ninterface Math {\n    /**\n     * Returns the number of leading zero bits in the 32-bit binary representation of a number.\n     * @param x A numeric expression.\n     */\n    clz32(x: number): number;\n\n    /**\n     * Returns the result of 32-bit multiplication of two numbers.\n     * @param x First number\n     * @param y Second number\n     */\n    imul(x: number, y: number): number;\n\n    /**\n     * Returns the sign of the x, indicating whether x is positive, negative or zero.\n     * @param x The numeric expression to test\n     */\n    sign(x: number): number;\n\n    /**\n     * Returns the base 10 logarithm of a number.\n     * @param x A numeric expression.\n     */\n    log10(x: number): number;\n\n    /**\n     * Returns the base 2 logarithm of a number.\n     * @param x A numeric expression.\n     */\n    log2(x: number): number;\n\n    /**\n     * Returns the natural logarithm of 1 + x.\n     * @param x A numeric expression.\n     */\n    log1p(x: number): number;\n\n    /**\n     * Returns the result of (e^x - 1), which is an implementation-dependent approximation to\n     * subtracting 1 from the exponential function of x (e raised to the power of x, where e\n     * is the base of the natural logarithms).\n     * @param x A numeric expression.\n     */\n    expm1(x: number): number;\n\n    /**\n     * Returns the hyperbolic cosine of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    cosh(x: number): number;\n\n    /**\n     * Returns the hyperbolic sine of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    sinh(x: number): number;\n\n    /**\n     * Returns the hyperbolic tangent of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    tanh(x: number): number;\n\n    /**\n     * Returns the inverse hyperbolic cosine of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    acosh(x: number): number;\n\n    /**\n     * Returns the inverse hyperbolic sine of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    asinh(x: number): number;\n\n    /**\n     * Returns the inverse hyperbolic tangent of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    atanh(x: number): number;\n\n    /**\n     * Returns the square root of the sum of squares of its arguments.\n     * @param values Values to compute the square root for.\n     *     If no arguments are passed, the result is +0.\n     *     If there is only one argument, the result is the absolute value.\n     *     If any argument is +Infinity or -Infinity, the result is +Infinity.\n     *     If any argument is NaN, the result is NaN.\n     *     If all arguments are either +0 or −0, the result is +0.\n     */\n    hypot(...values: number[]): number;\n\n    /**\n     * Returns the integral part of the a numeric expression, x, removing any fractional digits.\n     * If x is already an integer, the result is x.\n     * @param x A numeric expression.\n     */\n    trunc(x: number): number;\n\n    /**\n     * Returns the nearest single precision float representation of a number.\n     * @param x A numeric expression.\n     */\n    fround(x: number): number;\n\n    /**\n     * Returns an implementation-dependent approximation to the cube root of number.\n     * @param x A numeric expression.\n     */\n    cbrt(x: number): number;\n}\n\ninterface NumberConstructor {\n    /**\n     * The value of Number.EPSILON is the difference between 1 and the smallest value greater than 1\n     * that is representable as a Number value, which is approximately:\n     * 2.2204460492503130808472633361816 x 10‍−‍16.\n     */\n    readonly EPSILON: number;\n\n    /**\n     * Returns true if passed value is finite.\n     * Unlike the global isFinite, Number.isFinite doesn\'t forcibly convert the parameter to a\n     * number. Only finite values of the type number, result in true.\n     * @param number A numeric value.\n     */\n    isFinite(number: number): boolean;\n\n    /**\n     * Returns true if the value passed is an integer, false otherwise.\n     * @param number A numeric value.\n     */\n    isInteger(number: number): boolean;\n\n    /**\n     * Returns a Boolean value that indicates whether a value is the reserved value NaN (not a\n     * number). Unlike the global isNaN(), Number.isNaN() doesn\'t forcefully convert the parameter\n     * to a number. Only values of the type number, that are also NaN, result in true.\n     * @param number A numeric value.\n     */\n    isNaN(number: number): boolean;\n\n    /**\n     * Returns true if the value passed is a safe integer.\n     * @param number A numeric value.\n     */\n    isSafeInteger(number: number): boolean;\n\n    /**\n     * The value of the largest integer n such that n and n + 1 are both exactly representable as\n     * a Number value.\n     * The value of Number.MAX_SAFE_INTEGER is 9007199254740991 2^53 − 1.\n     */\n    readonly MAX_SAFE_INTEGER: number;\n\n    /**\n     * The value of the smallest integer n such that n and n − 1 are both exactly representable as\n     * a Number value.\n     * The value of Number.MIN_SAFE_INTEGER is −9007199254740991 (−(2^53 − 1)).\n     */\n    readonly MIN_SAFE_INTEGER: number;\n\n    /**\n     * Converts a string to a floating-point number.\n     * @param string A string that contains a floating-point number.\n     */\n    parseFloat(string: string): number;\n\n    /**\n     * Converts A string to an integer.\n     * @param s A string to convert into a number.\n     * @param radix A value between 2 and 36 that specifies the base of the number in numString.\n     * If this argument is not supplied, strings with a prefix of \'0x\' are considered hexadecimal.\n     * All other strings are considered decimal.\n     */\n    parseInt(string: string, radix?: number): number;\n}\n\ninterface ObjectConstructor {\n    /**\n     * Copy the values of all of the enumerable own properties from one or more source objects to a\n     * target object. Returns the target object.\n     * @param target The target object to copy to.\n     * @param source The source object from which to copy properties.\n     */\n    assign<T, U>(target: T, source: U): T & U;\n\n    /**\n     * Copy the values of all of the enumerable own properties from one or more source objects to a\n     * target object. Returns the target object.\n     * @param target The target object to copy to.\n     * @param source1 The first source object from which to copy properties.\n     * @param source2 The second source object from which to copy properties.\n     */\n    assign<T, U, V>(target: T, source1: U, source2: V): T & U & V;\n\n    /**\n     * Copy the values of all of the enumerable own properties from one or more source objects to a\n     * target object. Returns the target object.\n     * @param target The target object to copy to.\n     * @param source1 The first source object from which to copy properties.\n     * @param source2 The second source object from which to copy properties.\n     * @param source3 The third source object from which to copy properties.\n     */\n    assign<T, U, V, W>(target: T, source1: U, source2: V, source3: W): T & U & V & W;\n\n    /**\n     * Copy the values of all of the enumerable own properties from one or more source objects to a\n     * target object. Returns the target object.\n     * @param target The target object to copy to.\n     * @param sources One or more source objects from which to copy properties\n     */\n    assign(target: object, ...sources: any[]): any;\n\n    /**\n     * Returns an array of all symbol properties found directly on object o.\n     * @param o Object to retrieve the symbols from.\n     */\n    getOwnPropertySymbols(o: any): symbol[];\n\n    /**\n     * Returns true if the values are the same value, false otherwise.\n     * @param value1 The first value.\n     * @param value2 The second value.\n     */\n    is(value1: any, value2: any): boolean;\n\n    /**\n     * Sets the prototype of a specified object o to  object proto or null. Returns the object o.\n     * @param o The object to change its prototype.\n     * @param proto The value of the new prototype or null.\n     */\n    setPrototypeOf(o: any, proto: object | null): any;\n}\n\ninterface ReadonlyArray<T> {\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find<S extends T>(predicate: (this: void, value: T, index: number, obj: ReadonlyArray<T>) => value is S, thisArg?: any): S | undefined;\n    find(predicate: (value: T, index: number, obj: ReadonlyArray<T>) => boolean, thisArg?: any): T | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: T, index: number, obj: ReadonlyArray<T>) => boolean, thisArg?: any): number;\n}\n\ninterface RegExp {\n    /**\n     * Returns a string indicating the flags of the regular expression in question. This field is read-only.\n     * The characters in this string are sequenced and concatenated in the following order:\n     *\n     *    - "g" for global\n     *    - "i" for ignoreCase\n     *    - "m" for multiline\n     *    - "u" for unicode\n     *    - "y" for sticky\n     *\n     * If no flags are set, the value is the empty string.\n     */\n    readonly flags: string;\n\n    /**\n     * Returns a Boolean value indicating the state of the sticky flag (y) used with a regular\n     * expression. Default is false. Read-only.\n     */\n    readonly sticky: boolean;\n\n    /**\n     * Returns a Boolean value indicating the state of the Unicode flag (u) used with a regular\n     * expression. Default is false. Read-only.\n     */\n    readonly unicode: boolean;\n}\n\ninterface RegExpConstructor {\n    new (pattern: RegExp, flags?: string): RegExp;\n    (pattern: RegExp, flags?: string): RegExp;\n}\n\ninterface String {\n    /**\n     * Returns a nonnegative integer Number less than 1114112 (0x110000) that is the code point\n     * value of the UTF-16 encoded code point starting at the string element at position pos in\n     * the String resulting from converting this object to a String.\n     * If there is no element at that position, the result is undefined.\n     * If a valid UTF-16 surrogate pair does not begin at pos, the result is the code unit at pos.\n     */\n    codePointAt(pos: number): number | undefined;\n\n    /**\n     * Returns true if searchString appears as a substring of the result of converting this\n     * object to a String, at one or more positions that are\n     * greater than or equal to position; otherwise, returns false.\n     * @param searchString search string\n     * @param position If position is undefined, 0 is assumed, so as to search all of the String.\n     */\n    includes(searchString: string, position?: number): boolean;\n\n    /**\n     * Returns true if the sequence of elements of searchString converted to a String is the\n     * same as the corresponding elements of this object (converted to a String) starting at\n     * endPosition – length(this). Otherwise returns false.\n     */\n    endsWith(searchString: string, endPosition?: number): boolean;\n\n    /**\n     * Returns the String value result of normalizing the string into the normalization form\n     * named by form as specified in Unicode Standard Annex #15, Unicode Normalization Forms.\n     * @param form Applicable values: "NFC", "NFD", "NFKC", or "NFKD", If not specified default\n     * is "NFC"\n     */\n    normalize(form: "NFC" | "NFD" | "NFKC" | "NFKD"): string;\n\n    /**\n     * Returns the String value result of normalizing the string into the normalization form\n     * named by form as specified in Unicode Standard Annex #15, Unicode Normalization Forms.\n     * @param form Applicable values: "NFC", "NFD", "NFKC", or "NFKD", If not specified default\n     * is "NFC"\n     */\n    normalize(form?: string): string;\n\n    /**\n     * Returns a String value that is made from count copies appended together. If count is 0,\n     * the empty string is returned.\n     * @param count number of copies to append\n     */\n    repeat(count: number): string;\n\n    /**\n     * Returns true if the sequence of elements of searchString converted to a String is the\n     * same as the corresponding elements of this object (converted to a String) starting at\n     * position. Otherwise returns false.\n     */\n    startsWith(searchString: string, position?: number): boolean;\n\n    /**\n     * Returns an <a> HTML anchor element and sets the name attribute to the text value\n     * @param name\n     */\n    anchor(name: string): string;\n\n    /** Returns a <big> HTML element */\n    big(): string;\n\n    /** Returns a <blink> HTML element */\n    blink(): string;\n\n    /** Returns a <b> HTML element */\n    bold(): string;\n\n    /** Returns a <tt> HTML element */\n    fixed(): string;\n\n    /** Returns a <font> HTML element and sets the color attribute value */\n    fontcolor(color: string): string;\n\n    /** Returns a <font> HTML element and sets the size attribute value */\n    fontsize(size: number): string;\n\n    /** Returns a <font> HTML element and sets the size attribute value */\n    fontsize(size: string): string;\n\n    /** Returns an <i> HTML element */\n    italics(): string;\n\n    /** Returns an <a> HTML element and sets the href attribute value */\n    link(url: string): string;\n\n    /** Returns a <small> HTML element */\n    small(): string;\n\n    /** Returns a <strike> HTML element */\n    strike(): string;\n\n    /** Returns a <sub> HTML element */\n    sub(): string;\n\n    /** Returns a <sup> HTML element */\n    sup(): string;\n}\n\ninterface StringConstructor {\n    /**\n     * Return the String value whose elements are, in order, the elements in the List elements.\n     * If length is 0, the empty string is returned.\n     */\n    fromCodePoint(...codePoints: number[]): string;\n\n    /**\n     * String.raw is intended for use as a tag function of a Tagged Template String. When called\n     * as such the first argument will be a well formed template call site object and the rest\n     * parameter will contain the substitution values.\n     * @param template A well-formed template string call site representation.\n     * @param substitutions A set of substitution values.\n     */\n    raw(template: TemplateStringsArray, ...substitutions: any[]): string;\n}\n',Xn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/////////////////////////////\n/// DOM Iterable APIs\n/////////////////////////////\n\ninterface AudioParamMap extends ReadonlyMap<string, AudioParam> {\n}\n\ninterface AudioTrackList {\n    [Symbol.iterator](): IterableIterator<AudioTrack>;\n}\n\ninterface CSSRuleList {\n    [Symbol.iterator](): IterableIterator<CSSRule>;\n}\n\ninterface CSSStyleDeclaration {\n    [Symbol.iterator](): IterableIterator<string>;\n}\n\ninterface ClientRectList {\n    [Symbol.iterator](): IterableIterator<ClientRect>;\n}\n\ninterface DOMRectList {\n    [Symbol.iterator](): IterableIterator<DOMRect>;\n}\n\ninterface DOMStringList {\n    [Symbol.iterator](): IterableIterator<string>;\n}\n\ninterface DOMTokenList {\n    [Symbol.iterator](): IterableIterator<string>;\n    entries(): IterableIterator<[number, string]>;\n    keys(): IterableIterator<number>;\n    values(): IterableIterator<string>;\n}\n\ninterface DataTransferItemList {\n    [Symbol.iterator](): IterableIterator<DataTransferItem>;\n}\n\ninterface FileList {\n    [Symbol.iterator](): IterableIterator<File>;\n}\n\ninterface FormData {\n    [Symbol.iterator](): IterableIterator<[string, FormDataEntryValue]>;\n    /**\n     * Returns an array of key, value pairs for every entry in the list.\n     */\n    entries(): IterableIterator<[string, FormDataEntryValue]>;\n    /**\n     * Returns a list of keys in the list.\n     */\n    keys(): IterableIterator<string>;\n    /**\n     * Returns a list of values in the list.\n     */\n    values(): IterableIterator<FormDataEntryValue>;\n}\n\ninterface HTMLAllCollection {\n    [Symbol.iterator](): IterableIterator<Element>;\n}\n\ninterface HTMLCollectionBase {\n    [Symbol.iterator](): IterableIterator<Element>;\n}\n\ninterface HTMLCollectionOf<T extends Element> {\n    [Symbol.iterator](): IterableIterator<T>;\n}\n\ninterface HTMLFormElement {\n    [Symbol.iterator](): IterableIterator<Element>;\n}\n\ninterface HTMLSelectElement {\n    [Symbol.iterator](): IterableIterator<Element>;\n}\n\ninterface Headers {\n    [Symbol.iterator](): IterableIterator<[string, string]>;\n    /**\n     * Returns an iterator allowing to go through all key/value pairs contained in this object.\n     */\n    entries(): IterableIterator<[string, string]>;\n    /**\n     * Returns an iterator allowing to go through all keys of the key/value pairs contained in this object.\n     */\n    keys(): IterableIterator<string>;\n    /**\n     * Returns an iterator allowing to go through all values of the key/value pairs contained in this object.\n     */\n    values(): IterableIterator<string>;\n}\n\ninterface MediaList {\n    [Symbol.iterator](): IterableIterator<string>;\n}\n\ninterface MimeTypeArray {\n    [Symbol.iterator](): IterableIterator<Plugin>;\n}\n\ninterface NamedNodeMap {\n    [Symbol.iterator](): IterableIterator<Attr>;\n}\n\ninterface NodeList {\n    [Symbol.iterator](): IterableIterator<Node>;\n    /**\n     * Returns an array of key, value pairs for every entry in the list.\n     */\n    entries(): IterableIterator<[number, Node]>;\n    /**\n     * Returns an list of keys in the list.\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the list.\n     */\n    values(): IterableIterator<Node>;\n}\n\ninterface NodeListOf<TNode extends Node> {\n    [Symbol.iterator](): IterableIterator<TNode>;\n    /**\n     * Returns an array of key, value pairs for every entry in the list.\n     */\n    entries(): IterableIterator<[number, TNode]>;\n    /**\n     * Returns an list of keys in the list.\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the list.\n     */\n    values(): IterableIterator<TNode>;\n}\n\ninterface Plugin {\n    [Symbol.iterator](): IterableIterator<MimeType>;\n}\n\ninterface PluginArray {\n    [Symbol.iterator](): IterableIterator<Plugin>;\n}\n\ninterface RTCStatsReport extends ReadonlyMap<string, any> {\n}\n\ninterface SVGLengthList {\n    [Symbol.iterator](): IterableIterator<SVGLength>;\n}\n\ninterface SVGNumberList {\n    [Symbol.iterator](): IterableIterator<SVGNumber>;\n}\n\ninterface SVGStringList {\n    [Symbol.iterator](): IterableIterator<string>;\n}\n\ninterface SourceBufferList {\n    [Symbol.iterator](): IterableIterator<SourceBuffer>;\n}\n\ninterface SpeechGrammarList {\n    [Symbol.iterator](): IterableIterator<SpeechGrammar>;\n}\n\ninterface SpeechRecognitionResult {\n    [Symbol.iterator](): IterableIterator<SpeechRecognitionAlternative>;\n}\n\ninterface SpeechRecognitionResultList {\n    [Symbol.iterator](): IterableIterator<SpeechRecognitionResult>;\n}\n\ninterface StyleSheetList {\n    [Symbol.iterator](): IterableIterator<StyleSheet>;\n}\n\ninterface TextTrackCueList {\n    [Symbol.iterator](): IterableIterator<TextTrackCue>;\n}\n\ninterface TextTrackList {\n    [Symbol.iterator](): IterableIterator<TextTrack>;\n}\n\ninterface TouchList {\n    [Symbol.iterator](): IterableIterator<Touch>;\n}\n\ninterface URLSearchParams {\n    [Symbol.iterator](): IterableIterator<[string, string]>;\n    /**\n     * Returns an array of key, value pairs for every entry in the search params.\n     */\n    entries(): IterableIterator<[string, string]>;\n    /**\n     * Returns a list of keys in the search params.\n     */\n    keys(): IterableIterator<string>;\n    /**\n     * Returns a list of values in the search params.\n     */\n    values(): IterableIterator<string>;\n}\n\ninterface VideoTrackList {\n    [Symbol.iterator](): IterableIterator<VideoTrack>;\n}\n',Yn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n\n\n/////////////////////////////\n/// Windows Script Host APIS\n/////////////////////////////\n\n\ninterface ActiveXObject {\n    new (s: string): any;\n}\ndeclare var ActiveXObject: ActiveXObject;\n\ninterface ITextWriter {\n    Write(s: string): void;\n    WriteLine(s: string): void;\n    Close(): void;\n}\n\ninterface TextStreamBase {\n    /**\n     * The column number of the current character position in an input stream.\n     */\n    Column: number;\n\n    /**\n     * The current line number in an input stream.\n     */\n    Line: number;\n\n    /**\n     * Closes a text stream.\n     * It is not necessary to close standard streams; they close automatically when the process ends. If\n     * you close a standard stream, be aware that any other pointers to that standard stream become invalid.\n     */\n    Close(): void;\n}\n\ninterface TextStreamWriter extends TextStreamBase {\n    /**\n     * Sends a string to an output stream.\n     */\n    Write(s: string): void;\n\n    /**\n     * Sends a specified number of blank lines (newline characters) to an output stream.\n     */\n    WriteBlankLines(intLines: number): void;\n\n    /**\n     * Sends a string followed by a newline character to an output stream.\n     */\n    WriteLine(s: string): void;\n}\n\ninterface TextStreamReader extends TextStreamBase {\n    /**\n     * Returns a specified number of characters from an input stream, starting at the current pointer position.\n     * Does not return until the ENTER key is pressed.\n     * Can only be used on a stream in reading mode; causes an error in writing or appending mode.\n     */\n    Read(characters: number): string;\n\n    /**\n     * Returns all characters from an input stream.\n     * Can only be used on a stream in reading mode; causes an error in writing or appending mode.\n     */\n    ReadAll(): string;\n\n    /**\n     * Returns an entire line from an input stream.\n     * Although this method extracts the newline character, it does not add it to the returned string.\n     * Can only be used on a stream in reading mode; causes an error in writing or appending mode.\n     */\n    ReadLine(): string;\n\n    /**\n     * Skips a specified number of characters when reading from an input text stream.\n     * Can only be used on a stream in reading mode; causes an error in writing or appending mode.\n     * @param characters Positive number of characters to skip forward. (Backward skipping is not supported.)\n     */\n    Skip(characters: number): void;\n\n    /**\n     * Skips the next line when reading from an input text stream.\n     * Can only be used on a stream in reading mode, not writing or appending mode.\n     */\n    SkipLine(): void;\n\n    /**\n     * Indicates whether the stream pointer position is at the end of a line.\n     */\n    AtEndOfLine: boolean;\n\n    /**\n     * Indicates whether the stream pointer position is at the end of a stream.\n     */\n    AtEndOfStream: boolean;\n}\n\ndeclare var WScript: {\n    /**\n     * Outputs text to either a message box (under WScript.exe) or the command console window followed by\n     * a newline (under CScript.exe).\n     */\n    Echo(s: any): void;\n\n    /**\n     * Exposes the write-only error output stream for the current script.\n     * Can be accessed only while using CScript.exe.\n     */\n    StdErr: TextStreamWriter;\n\n    /**\n     * Exposes the write-only output stream for the current script.\n     * Can be accessed only while using CScript.exe.\n     */\n    StdOut: TextStreamWriter;\n    Arguments: { length: number; Item(n: number): string; };\n\n    /**\n     *  The full path of the currently running script.\n     */\n    ScriptFullName: string;\n\n    /**\n     * Forces the script to stop immediately, with an optional exit code.\n     */\n    Quit(exitCode?: number): number;\n\n    /**\n     * The Windows Script Host build version number.\n     */\n    BuildVersion: number;\n\n    /**\n     * Fully qualified path of the host executable.\n     */\n    FullName: string;\n\n    /**\n     * Gets/sets the script mode - interactive(true) or batch(false).\n     */\n    Interactive: boolean;\n\n    /**\n     * The name of the host executable (WScript.exe or CScript.exe).\n     */\n    Name: string;\n\n    /**\n     * Path of the directory containing the host executable.\n     */\n    Path: string;\n\n    /**\n     * The filename of the currently running script.\n     */\n    ScriptName: string;\n\n    /**\n     * Exposes the read-only input stream for the current script.\n     * Can be accessed only while using CScript.exe.\n     */\n    StdIn: TextStreamReader;\n\n    /**\n     * Windows Script Host version\n     */\n    Version: string;\n\n    /**\n     * Connects a COM object\'s event sources to functions named with a given prefix, in the form prefix_event.\n     */\n    ConnectObject(objEventSource: any, strPrefix: string): void;\n\n    /**\n     * Creates a COM object.\n     * @param strProgiID\n     * @param strPrefix Function names in the form prefix_event will be bound to this object\'s COM events.\n     */\n    CreateObject(strProgID: string, strPrefix?: string): any;\n\n    /**\n     * Disconnects a COM object from its event sources.\n     */\n    DisconnectObject(obj: any): void;\n\n    /**\n     * Retrieves an existing object with the specified ProgID from memory, or creates a new one from a file.\n     * @param strPathname Fully qualified path to the file containing the object persisted to disk.\n     *                       For objects in memory, pass a zero-length string.\n     * @param strProgID\n     * @param strPrefix Function names in the form prefix_event will be bound to this object\'s COM events.\n     */\n    GetObject(strPathname: string, strProgID?: string, strPrefix?: string): any;\n\n    /**\n     * Suspends script execution for a specified length of time, then continues execution.\n     * @param intTime Interval (in milliseconds) to suspend script execution.\n     */\n    Sleep(intTime: number): void;\n};\n\n/**\n * WSH is an alias for WScript under Windows Script Host\n */\ndeclare var WSH: typeof WScript;\n\n/**\n * Represents an Automation SAFEARRAY\n */\ndeclare class SafeArray<T = any> {\n    private constructor();\n    private SafeArray_typekey: SafeArray<T>;\n}\n\n/**\n * Allows enumerating over a COM collection, which may not have indexed item access.\n */\ninterface Enumerator<T = any> {\n    /**\n     * Returns true if the current item is the last one in the collection, or the collection is empty,\n     * or the current item is undefined.\n     */\n    atEnd(): boolean;\n\n    /**\n     * Returns the current item in the collection\n     */\n    item(): T;\n\n    /**\n     * Resets the current item in the collection to the first item. If there are no items in the collection,\n     * the current item is set to undefined.\n     */\n    moveFirst(): void;\n\n    /**\n     * Moves the current item to the next item in the collection. If the enumerator is at the end of\n     * the collection or the collection is empty, the current item is set to undefined.\n     */\n    moveNext(): void;\n}\n\ninterface EnumeratorConstructor {\n    new <T = any>(safearray: SafeArray<T>): Enumerator<T>;\n    new <T = any>(collection: { Item(index: any): T }): Enumerator<T>;\n    new <T = any>(collection: any): Enumerator<T>;\n}\n\ndeclare var Enumerator: EnumeratorConstructor;\n\n/**\n * Enables reading from a COM safe array, which might have an alternate lower bound, or multiple dimensions.\n */\ninterface VBArray<T = any> {\n    /**\n     * Returns the number of dimensions (1-based).\n     */\n    dimensions(): number;\n\n    /**\n     * Takes an index for each dimension in the array, and returns the item at the corresponding location.\n     */\n    getItem(dimension1Index: number, ...dimensionNIndexes: number[]): T;\n\n    /**\n     * Returns the smallest available index for a given dimension.\n     * @param dimension 1-based dimension (defaults to 1)\n     */\n    lbound(dimension?: number): number;\n\n    /**\n     * Returns the largest available index for a given dimension.\n     * @param dimension 1-based dimension (defaults to 1)\n     */\n    ubound(dimension?: number): number;\n\n    /**\n     * Returns a Javascript array with all the elements in the VBArray. If there are multiple dimensions,\n     * each successive dimension is appended to the end of the array.\n     * Example: [[1,2,3],[4,5,6]] becomes [1,2,3,4,5,6]\n     */\n    toArray(): T[];\n}\n\ninterface VBArrayConstructor {\n    new <T = any>(safeArray: SafeArray<T>): VBArray<T>;\n}\n\ndeclare var VBArray: VBArrayConstructor;\n\n/**\n * Automation date (VT_DATE)\n */\ndeclare class VarDate {\n    private constructor();\n    private VarDate_typekey: VarDate;\n}\n\ninterface DateConstructor {\n    new (vd: VarDate): Date;\n}\n\ninterface Date {\n    getVarDate: () => VarDate;\n}\n',Qn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n\n/////////////////////////////\n/// WorkerGlobalScope APIs\n/////////////////////////////\n// These are only available in a Web Worker\ndeclare function importScripts(...urls: string[]): void;\n',Zn='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/////////////////////////////\n/// DOM APIs\n/////////////////////////////\n\ninterface Account {\n    displayName: string;\n    id: string;\n    imageURL?: string;\n    name?: string;\n    rpDisplayName: string;\n}\n\ninterface AddEventListenerOptions extends EventListenerOptions {\n    once?: boolean;\n    passive?: boolean;\n}\n\ninterface AesCbcParams extends Algorithm {\n    iv: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface AesCtrParams extends Algorithm {\n    counter: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    length: number;\n}\n\ninterface AesDerivedKeyParams extends Algorithm {\n    length: number;\n}\n\ninterface AesGcmParams extends Algorithm {\n    additionalData?: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    iv: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    tagLength?: number;\n}\n\ninterface AesKeyAlgorithm extends KeyAlgorithm {\n    length: number;\n}\n\ninterface AesKeyGenParams extends Algorithm {\n    length: number;\n}\n\ninterface Algorithm {\n    name: string;\n}\n\ninterface AnalyserOptions extends AudioNodeOptions {\n    fftSize?: number;\n    maxDecibels?: number;\n    minDecibels?: number;\n    smoothingTimeConstant?: number;\n}\n\ninterface AnimationEventInit extends EventInit {\n    animationName?: string;\n    elapsedTime?: number;\n    pseudoElement?: string;\n}\n\ninterface AnimationPlaybackEventInit extends EventInit {\n    currentTime?: number | null;\n    timelineTime?: number | null;\n}\n\ninterface AssertionOptions {\n    allowList?: ScopedCredentialDescriptor[];\n    extensions?: WebAuthnExtensions;\n    rpId?: string;\n    timeoutSeconds?: number;\n}\n\ninterface AssignedNodesOptions {\n    flatten?: boolean;\n}\n\ninterface AudioBufferOptions {\n    length: number;\n    numberOfChannels?: number;\n    sampleRate: number;\n}\n\ninterface AudioBufferSourceOptions {\n    buffer?: AudioBuffer | null;\n    detune?: number;\n    loop?: boolean;\n    loopEnd?: number;\n    loopStart?: number;\n    playbackRate?: number;\n}\n\ninterface AudioContextInfo {\n    currentTime?: number;\n    sampleRate?: number;\n}\n\ninterface AudioContextOptions {\n    latencyHint?: AudioContextLatencyCategory | number;\n    sampleRate?: number;\n}\n\ninterface AudioNodeOptions {\n    channelCount?: number;\n    channelCountMode?: ChannelCountMode;\n    channelInterpretation?: ChannelInterpretation;\n}\n\ninterface AudioParamDescriptor {\n    automationRate?: AutomationRate;\n    defaultValue?: number;\n    maxValue?: number;\n    minValue?: number;\n    name: string;\n}\n\ninterface AudioProcessingEventInit extends EventInit {\n    inputBuffer: AudioBuffer;\n    outputBuffer: AudioBuffer;\n    playbackTime: number;\n}\n\ninterface AudioTimestamp {\n    contextTime?: number;\n    performanceTime?: number;\n}\n\ninterface AudioWorkletNodeOptions extends AudioNodeOptions {\n    numberOfInputs?: number;\n    numberOfOutputs?: number;\n    outputChannelCount?: number[];\n    parameterData?: Record<string, number>;\n    processorOptions?: any;\n}\n\ninterface BiquadFilterOptions extends AudioNodeOptions {\n    Q?: number;\n    detune?: number;\n    frequency?: number;\n    gain?: number;\n    type?: BiquadFilterType;\n}\n\ninterface BlobPropertyBag {\n    endings?: EndingType;\n    type?: string;\n}\n\ninterface ByteLengthChunk {\n    byteLength?: number;\n}\n\ninterface CacheQueryOptions {\n    cacheName?: string;\n    ignoreMethod?: boolean;\n    ignoreSearch?: boolean;\n    ignoreVary?: boolean;\n}\n\ninterface CanvasRenderingContext2DSettings {\n    alpha?: boolean;\n}\n\ninterface ChannelMergerOptions extends AudioNodeOptions {\n    numberOfInputs?: number;\n}\n\ninterface ChannelSplitterOptions extends AudioNodeOptions {\n    numberOfOutputs?: number;\n}\n\ninterface ClientData {\n    challenge: string;\n    extensions?: WebAuthnExtensions;\n    hashAlg: string | Algorithm;\n    origin: string;\n    rpId: string;\n    tokenBinding?: string;\n}\n\ninterface ClientQueryOptions {\n    includeUncontrolled?: boolean;\n    type?: ClientTypes;\n}\n\ninterface CloseEventInit extends EventInit {\n    code?: number;\n    reason?: string;\n    wasClean?: boolean;\n}\n\ninterface CompositionEventInit extends UIEventInit {\n    data?: string;\n}\n\ninterface ComputedEffectTiming extends EffectTiming {\n    activeDuration?: number;\n    currentIteration?: number | null;\n    endTime?: number;\n    localTime?: number | null;\n    progress?: number | null;\n}\n\ninterface ComputedKeyframe {\n    composite: CompositeOperationOrAuto;\n    computedOffset: number;\n    easing: string;\n    offset: number | null;\n    [property: string]: string | number | null | undefined;\n}\n\ninterface ConfirmSiteSpecificExceptionsInformation extends ExceptionInformation {\n    arrayOfDomainStrings?: string[];\n}\n\ninterface ConstantSourceOptions {\n    offset?: number;\n}\n\ninterface ConstrainBooleanParameters {\n    exact?: boolean;\n    ideal?: boolean;\n}\n\ninterface ConstrainDOMStringParameters {\n    exact?: string | string[];\n    ideal?: string | string[];\n}\n\ninterface ConstrainDoubleRange extends DoubleRange {\n    exact?: number;\n    ideal?: number;\n}\n\ninterface ConstrainLongRange extends LongRange {\n    exact?: number;\n    ideal?: number;\n}\n\ninterface ConstrainVideoFacingModeParameters {\n    exact?: VideoFacingModeEnum | VideoFacingModeEnum[];\n    ideal?: VideoFacingModeEnum | VideoFacingModeEnum[];\n}\n\ninterface ConvolverOptions extends AudioNodeOptions {\n    buffer?: AudioBuffer | null;\n    disableNormalization?: boolean;\n}\n\ninterface CustomEventInit<T = any> extends EventInit {\n    detail?: T;\n}\n\ninterface DOMMatrix2DInit {\n    a?: number;\n    b?: number;\n    c?: number;\n    d?: number;\n    e?: number;\n    f?: number;\n    m11?: number;\n    m12?: number;\n    m21?: number;\n    m22?: number;\n    m41?: number;\n    m42?: number;\n}\n\ninterface DOMMatrixInit extends DOMMatrix2DInit {\n    is2D?: boolean;\n    m13?: number;\n    m14?: number;\n    m23?: number;\n    m24?: number;\n    m31?: number;\n    m32?: number;\n    m33?: number;\n    m34?: number;\n    m43?: number;\n    m44?: number;\n}\n\ninterface DOMPointInit {\n    w?: number;\n    x?: number;\n    y?: number;\n    z?: number;\n}\n\ninterface DOMQuadInit {\n    p1?: DOMPointInit;\n    p2?: DOMPointInit;\n    p3?: DOMPointInit;\n    p4?: DOMPointInit;\n}\n\ninterface DOMRectInit {\n    height?: number;\n    width?: number;\n    x?: number;\n    y?: number;\n}\n\ninterface DelayOptions extends AudioNodeOptions {\n    delayTime?: number;\n    maxDelayTime?: number;\n}\n\ninterface DeviceAccelerationDict {\n    x?: number | null;\n    y?: number | null;\n    z?: number | null;\n}\n\ninterface DeviceLightEventInit extends EventInit {\n    value?: number;\n}\n\ninterface DeviceMotionEventInit extends EventInit {\n    acceleration?: DeviceAccelerationDict | null;\n    accelerationIncludingGravity?: DeviceAccelerationDict | null;\n    interval?: number | null;\n    rotationRate?: DeviceRotationRateDict | null;\n}\n\ninterface DeviceOrientationEventInit extends EventInit {\n    absolute?: boolean;\n    alpha?: number | null;\n    beta?: number | null;\n    gamma?: number | null;\n}\n\ninterface DeviceRotationRateDict {\n    alpha?: number | null;\n    beta?: number | null;\n    gamma?: number | null;\n}\n\ninterface DocumentTimelineOptions {\n    originTime?: number;\n}\n\ninterface DoubleRange {\n    max?: number;\n    min?: number;\n}\n\ninterface DragEventInit extends MouseEventInit {\n    dataTransfer?: DataTransfer | null;\n}\n\ninterface DynamicsCompressorOptions extends AudioNodeOptions {\n    attack?: number;\n    knee?: number;\n    ratio?: number;\n    release?: number;\n    threshold?: number;\n}\n\ninterface EcKeyAlgorithm extends KeyAlgorithm {\n    namedCurve: NamedCurve;\n}\n\ninterface EcKeyGenParams extends Algorithm {\n    namedCurve: NamedCurve;\n}\n\ninterface EcKeyImportParams extends Algorithm {\n    namedCurve: NamedCurve;\n}\n\ninterface EcdhKeyDeriveParams extends Algorithm {\n    public: CryptoKey;\n}\n\ninterface EcdsaParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n}\n\ninterface EffectTiming {\n    delay?: number;\n    direction?: PlaybackDirection;\n    duration?: number | string;\n    easing?: string;\n    endDelay?: number;\n    fill?: FillMode;\n    iterationStart?: number;\n    iterations?: number;\n}\n\ninterface ElementDefinitionOptions {\n    extends?: string;\n}\n\ninterface ErrorEventInit extends EventInit {\n    colno?: number;\n    error?: any;\n    filename?: string;\n    lineno?: number;\n    message?: string;\n}\n\ninterface EventInit {\n    bubbles?: boolean;\n    cancelable?: boolean;\n    composed?: boolean;\n}\n\ninterface EventListenerOptions {\n    capture?: boolean;\n}\n\ninterface EventModifierInit extends UIEventInit {\n    altKey?: boolean;\n    ctrlKey?: boolean;\n    metaKey?: boolean;\n    modifierAltGraph?: boolean;\n    modifierCapsLock?: boolean;\n    modifierFn?: boolean;\n    modifierFnLock?: boolean;\n    modifierHyper?: boolean;\n    modifierNumLock?: boolean;\n    modifierOS?: boolean;\n    modifierScrollLock?: boolean;\n    modifierSuper?: boolean;\n    modifierSymbol?: boolean;\n    modifierSymbolLock?: boolean;\n    shiftKey?: boolean;\n}\n\ninterface ExceptionInformation {\n    domain?: string | null;\n}\n\ninterface FilePropertyBag extends BlobPropertyBag {\n    lastModified?: number;\n}\n\ninterface FocusEventInit extends UIEventInit {\n    relatedTarget?: EventTarget | null;\n}\n\ninterface FocusNavigationEventInit extends EventInit {\n    navigationReason?: string | null;\n    originHeight?: number;\n    originLeft?: number;\n    originTop?: number;\n    originWidth?: number;\n}\n\ninterface FocusNavigationOrigin {\n    originHeight?: number;\n    originLeft?: number;\n    originTop?: number;\n    originWidth?: number;\n}\n\ninterface FocusOptions {\n    preventScroll?: boolean;\n}\n\ninterface GainOptions extends AudioNodeOptions {\n    gain?: number;\n}\n\ninterface GamepadEventInit extends EventInit {\n    gamepad?: Gamepad;\n}\n\ninterface GetNotificationOptions {\n    tag?: string;\n}\n\ninterface GetRootNodeOptions {\n    composed?: boolean;\n}\n\ninterface HashChangeEventInit extends EventInit {\n    newURL?: string;\n    oldURL?: string;\n}\n\ninterface HkdfParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    info: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    salt: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface HmacImportParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    length?: number;\n}\n\ninterface HmacKeyAlgorithm extends KeyAlgorithm {\n    hash: KeyAlgorithm;\n    length: number;\n}\n\ninterface HmacKeyGenParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    length?: number;\n}\n\ninterface IDBIndexParameters {\n    multiEntry?: boolean;\n    unique?: boolean;\n}\n\ninterface IDBObjectStoreParameters {\n    autoIncrement?: boolean;\n    keyPath?: string | string[] | null;\n}\n\ninterface IDBVersionChangeEventInit extends EventInit {\n    newVersion?: number | null;\n    oldVersion?: number;\n}\n\ninterface IIRFilterOptions extends AudioNodeOptions {\n    feedback: number[];\n    feedforward: number[];\n}\n\ninterface IntersectionObserverEntryInit {\n    boundingClientRect: DOMRectInit;\n    intersectionRect: DOMRectInit;\n    isIntersecting: boolean;\n    rootBounds: DOMRectInit;\n    target: Element;\n    time: number;\n}\n\ninterface IntersectionObserverInit {\n    root?: Element | null;\n    rootMargin?: string;\n    threshold?: number | number[];\n}\n\ninterface JsonWebKey {\n    alg?: string;\n    crv?: string;\n    d?: string;\n    dp?: string;\n    dq?: string;\n    e?: string;\n    ext?: boolean;\n    k?: string;\n    key_ops?: string[];\n    kty?: string;\n    n?: string;\n    oth?: RsaOtherPrimesInfo[];\n    p?: string;\n    q?: string;\n    qi?: string;\n    use?: string;\n    x?: string;\n    y?: string;\n}\n\ninterface KeyAlgorithm {\n    name: string;\n}\n\ninterface KeyboardEventInit extends EventModifierInit {\n    code?: string;\n    key?: string;\n    location?: number;\n    repeat?: boolean;\n}\n\ninterface Keyframe {\n    composite?: CompositeOperationOrAuto;\n    easing?: string;\n    offset?: number | null;\n    [property: string]: string | number | null | undefined;\n}\n\ninterface KeyframeAnimationOptions extends KeyframeEffectOptions {\n    id?: string;\n}\n\ninterface KeyframeEffectOptions extends EffectTiming {\n    composite?: CompositeOperation;\n    iterationComposite?: IterationCompositeOperation;\n}\n\ninterface LongRange {\n    max?: number;\n    min?: number;\n}\n\ninterface MediaElementAudioSourceOptions {\n    mediaElement: HTMLMediaElement;\n}\n\ninterface MediaEncryptedEventInit extends EventInit {\n    initData?: ArrayBuffer | null;\n    initDataType?: string;\n}\n\ninterface MediaKeyMessageEventInit extends EventInit {\n    message?: ArrayBuffer | null;\n    messageType?: MediaKeyMessageType;\n}\n\ninterface MediaKeySystemConfiguration {\n    audioCapabilities?: MediaKeySystemMediaCapability[];\n    distinctiveIdentifier?: MediaKeysRequirement;\n    initDataTypes?: string[];\n    persistentState?: MediaKeysRequirement;\n    videoCapabilities?: MediaKeySystemMediaCapability[];\n}\n\ninterface MediaKeySystemMediaCapability {\n    contentType?: string;\n    robustness?: string;\n}\n\ninterface MediaQueryListEventInit extends EventInit {\n    matches?: boolean;\n    media?: string;\n}\n\ninterface MediaStreamAudioSourceOptions {\n    mediaStream: MediaStream;\n}\n\ninterface MediaStreamConstraints {\n    audio?: boolean | MediaTrackConstraints;\n    peerIdentity?: string;\n    video?: boolean | MediaTrackConstraints;\n}\n\ninterface MediaStreamErrorEventInit extends EventInit {\n    error?: MediaStreamError | null;\n}\n\ninterface MediaStreamEventInit extends EventInit {\n    stream?: MediaStream;\n}\n\ninterface MediaStreamTrackAudioSourceOptions {\n    mediaStreamTrack: MediaStreamTrack;\n}\n\ninterface MediaStreamTrackEventInit extends EventInit {\n    track?: MediaStreamTrack | null;\n}\n\ninterface MediaTrackCapabilities {\n    aspectRatio?: number | DoubleRange;\n    deviceId?: string;\n    echoCancellation?: boolean[];\n    facingMode?: string;\n    frameRate?: number | DoubleRange;\n    groupId?: string;\n    height?: number | LongRange;\n    sampleRate?: number | LongRange;\n    sampleSize?: number | LongRange;\n    volume?: number | DoubleRange;\n    width?: number | LongRange;\n}\n\ninterface MediaTrackConstraintSet {\n    aspectRatio?: number | ConstrainDoubleRange;\n    channelCount?: number | ConstrainLongRange;\n    deviceId?: string | string[] | ConstrainDOMStringParameters;\n    displaySurface?: string | string[] | ConstrainDOMStringParameters;\n    echoCancellation?: boolean | ConstrainBooleanParameters;\n    facingMode?: string | string[] | ConstrainDOMStringParameters;\n    frameRate?: number | ConstrainDoubleRange;\n    groupId?: string | string[] | ConstrainDOMStringParameters;\n    height?: number | ConstrainLongRange;\n    latency?: number | ConstrainDoubleRange;\n    logicalSurface?: boolean | ConstrainBooleanParameters;\n    sampleRate?: number | ConstrainLongRange;\n    sampleSize?: number | ConstrainLongRange;\n    volume?: number | ConstrainDoubleRange;\n    width?: number | ConstrainLongRange;\n}\n\ninterface MediaTrackConstraints extends MediaTrackConstraintSet {\n    advanced?: MediaTrackConstraintSet[];\n}\n\ninterface MediaTrackSettings {\n    aspectRatio?: number;\n    deviceId?: string;\n    echoCancellation?: boolean;\n    facingMode?: string;\n    frameRate?: number;\n    groupId?: string;\n    height?: number;\n    sampleRate?: number;\n    sampleSize?: number;\n    volume?: number;\n    width?: number;\n}\n\ninterface MediaTrackSupportedConstraints {\n    aspectRatio?: boolean;\n    deviceId?: boolean;\n    echoCancellation?: boolean;\n    facingMode?: boolean;\n    frameRate?: boolean;\n    groupId?: boolean;\n    height?: boolean;\n    sampleRate?: boolean;\n    sampleSize?: boolean;\n    volume?: boolean;\n    width?: boolean;\n}\n\ninterface MessageEventInit extends EventInit {\n    data?: any;\n    lastEventId?: string;\n    origin?: string;\n    ports?: MessagePort[];\n    source?: MessageEventSource | null;\n}\n\ninterface MouseEventInit extends EventModifierInit {\n    button?: number;\n    buttons?: number;\n    clientX?: number;\n    clientY?: number;\n    relatedTarget?: EventTarget | null;\n    screenX?: number;\n    screenY?: number;\n}\n\ninterface MutationObserverInit {\n    attributeFilter?: string[];\n    attributeOldValue?: boolean;\n    attributes?: boolean;\n    characterData?: boolean;\n    characterDataOldValue?: boolean;\n    childList?: boolean;\n    subtree?: boolean;\n}\n\ninterface NavigationPreloadState {\n    enabled?: boolean;\n    headerValue?: string;\n}\n\ninterface NotificationAction {\n    action: string;\n    icon?: string;\n    title: string;\n}\n\ninterface NotificationOptions {\n    actions?: NotificationAction[];\n    badge?: string;\n    body?: string;\n    data?: any;\n    dir?: NotificationDirection;\n    icon?: string;\n    image?: string;\n    lang?: string;\n    renotify?: boolean;\n    requireInteraction?: boolean;\n    silent?: boolean;\n    tag?: string;\n    timestamp?: number;\n    vibrate?: VibratePattern;\n}\n\ninterface OfflineAudioCompletionEventInit extends EventInit {\n    renderedBuffer: AudioBuffer;\n}\n\ninterface OfflineAudioContextOptions {\n    length: number;\n    numberOfChannels?: number;\n    sampleRate: number;\n}\n\ninterface OptionalEffectTiming {\n    delay?: number;\n    direction?: PlaybackDirection;\n    duration?: number | string;\n    easing?: string;\n    endDelay?: number;\n    fill?: FillMode;\n    iterationStart?: number;\n    iterations?: number;\n}\n\ninterface OscillatorOptions extends AudioNodeOptions {\n    detune?: number;\n    frequency?: number;\n    periodicWave?: PeriodicWave;\n    type?: OscillatorType;\n}\n\ninterface PannerOptions extends AudioNodeOptions {\n    coneInnerAngle?: number;\n    coneOuterAngle?: number;\n    coneOuterGain?: number;\n    distanceModel?: DistanceModelType;\n    maxDistance?: number;\n    orientationX?: number;\n    orientationY?: number;\n    orientationZ?: number;\n    panningModel?: PanningModelType;\n    positionX?: number;\n    positionY?: number;\n    positionZ?: number;\n    refDistance?: number;\n    rolloffFactor?: number;\n}\n\ninterface PaymentCurrencyAmount {\n    currency: string;\n    currencySystem?: string;\n    value: string;\n}\n\ninterface PaymentDetailsBase {\n    displayItems?: PaymentItem[];\n    modifiers?: PaymentDetailsModifier[];\n    shippingOptions?: PaymentShippingOption[];\n}\n\ninterface PaymentDetailsInit extends PaymentDetailsBase {\n    id?: string;\n    total: PaymentItem;\n}\n\ninterface PaymentDetailsModifier {\n    additionalDisplayItems?: PaymentItem[];\n    data?: any;\n    supportedMethods: string | string[];\n    total?: PaymentItem;\n}\n\ninterface PaymentDetailsUpdate extends PaymentDetailsBase {\n    error?: string;\n    total?: PaymentItem;\n}\n\ninterface PaymentItem {\n    amount: PaymentCurrencyAmount;\n    label: string;\n    pending?: boolean;\n}\n\ninterface PaymentMethodData {\n    data?: any;\n    supportedMethods: string | string[];\n}\n\ninterface PaymentOptions {\n    requestPayerEmail?: boolean;\n    requestPayerName?: boolean;\n    requestPayerPhone?: boolean;\n    requestShipping?: boolean;\n    shippingType?: string;\n}\n\ninterface PaymentRequestUpdateEventInit extends EventInit {\n}\n\ninterface PaymentShippingOption {\n    amount: PaymentCurrencyAmount;\n    id: string;\n    label: string;\n    selected?: boolean;\n}\n\ninterface Pbkdf2Params extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    iterations: number;\n    salt: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface PerformanceObserverInit {\n    buffered?: boolean;\n    entryTypes: string[];\n}\n\ninterface PeriodicWaveConstraints {\n    disableNormalization?: boolean;\n}\n\ninterface PeriodicWaveOptions extends PeriodicWaveConstraints {\n    imag?: number[] | Float32Array;\n    real?: number[] | Float32Array;\n}\n\ninterface PipeOptions {\n    preventAbort?: boolean;\n    preventCancel?: boolean;\n    preventClose?: boolean;\n}\n\ninterface PointerEventInit extends MouseEventInit {\n    height?: number;\n    isPrimary?: boolean;\n    pointerId?: number;\n    pointerType?: string;\n    pressure?: number;\n    tangentialPressure?: number;\n    tiltX?: number;\n    tiltY?: number;\n    twist?: number;\n    width?: number;\n}\n\ninterface PopStateEventInit extends EventInit {\n    state?: any;\n}\n\ninterface PositionOptions {\n    enableHighAccuracy?: boolean;\n    maximumAge?: number;\n    timeout?: number;\n}\n\ninterface ProgressEventInit extends EventInit {\n    lengthComputable?: boolean;\n    loaded?: number;\n    total?: number;\n}\n\ninterface PromiseRejectionEventInit extends EventInit {\n    promise: Promise<any>;\n    reason?: any;\n}\n\ninterface PropertyIndexedKeyframes {\n    composite?: CompositeOperationOrAuto | CompositeOperationOrAuto[];\n    easing?: string | string[];\n    offset?: number | (number | null)[];\n    [property: string]: string | string[] | number | null | (number | null)[] | undefined;\n}\n\ninterface PushSubscriptionJSON {\n    endpoint?: string;\n    expirationTime?: number | null;\n    keys?: Record<string, string>;\n}\n\ninterface PushSubscriptionOptionsInit {\n    applicationServerKey?: BufferSource | string | null;\n    userVisibleOnly?: boolean;\n}\n\ninterface QueuingStrategy<T = any> {\n    highWaterMark?: number;\n    size?: QueuingStrategySizeCallback<T>;\n}\n\ninterface RTCAnswerOptions extends RTCOfferAnswerOptions {\n}\n\ninterface RTCCertificateExpiration {\n    expires?: number;\n}\n\ninterface RTCConfiguration {\n    bundlePolicy?: RTCBundlePolicy;\n    certificates?: RTCCertificate[];\n    iceCandidatePoolSize?: number;\n    iceServers?: RTCIceServer[];\n    iceTransportPolicy?: RTCIceTransportPolicy;\n    peerIdentity?: string;\n    rtcpMuxPolicy?: RTCRtcpMuxPolicy;\n}\n\ninterface RTCDTMFToneChangeEventInit extends EventInit {\n    tone: string;\n}\n\ninterface RTCDataChannelEventInit extends EventInit {\n    channel: RTCDataChannel;\n}\n\ninterface RTCDataChannelInit {\n    id?: number;\n    maxPacketLifeTime?: number;\n    maxRetransmits?: number;\n    negotiated?: boolean;\n    ordered?: boolean;\n    priority?: RTCPriorityType;\n    protocol?: string;\n}\n\ninterface RTCDtlsFingerprint {\n    algorithm?: string;\n    value?: string;\n}\n\ninterface RTCDtlsParameters {\n    fingerprints?: RTCDtlsFingerprint[];\n    role?: RTCDtlsRole;\n}\n\ninterface RTCErrorEventInit extends EventInit {\n    error?: RTCError | null;\n}\n\ninterface RTCIceCandidateAttributes extends RTCStats {\n    addressSourceUrl?: string;\n    candidateType?: RTCStatsIceCandidateType;\n    ipAddress?: string;\n    portNumber?: number;\n    priority?: number;\n    transport?: string;\n}\n\ninterface RTCIceCandidateComplete {\n}\n\ninterface RTCIceCandidateDictionary {\n    foundation?: string;\n    ip?: string;\n    msMTurnSessionId?: string;\n    port?: number;\n    priority?: number;\n    protocol?: RTCIceProtocol;\n    relatedAddress?: string;\n    relatedPort?: number;\n    tcpType?: RTCIceTcpCandidateType;\n    type?: RTCIceCandidateType;\n}\n\ninterface RTCIceCandidateInit {\n    candidate?: string;\n    sdpMLineIndex?: number | null;\n    sdpMid?: string | null;\n    usernameFragment?: string;\n}\n\ninterface RTCIceCandidatePair {\n    local?: RTCIceCandidate;\n    remote?: RTCIceCandidate;\n}\n\ninterface RTCIceCandidatePairStats extends RTCStats {\n    availableIncomingBitrate?: number;\n    availableOutgoingBitrate?: number;\n    bytesReceived?: number;\n    bytesSent?: number;\n    localCandidateId?: string;\n    nominated?: boolean;\n    priority?: number;\n    readable?: boolean;\n    remoteCandidateId?: string;\n    roundTripTime?: number;\n    state?: RTCStatsIceCandidatePairState;\n    transportId?: string;\n    writable?: boolean;\n}\n\ninterface RTCIceGatherOptions {\n    gatherPolicy?: RTCIceGatherPolicy;\n    iceservers?: RTCIceServer[];\n}\n\ninterface RTCIceParameters {\n    password?: string;\n    usernameFragment?: string;\n}\n\ninterface RTCIceServer {\n    credential?: string | RTCOAuthCredential;\n    credentialType?: RTCIceCredentialType;\n    urls: string | string[];\n    username?: string;\n}\n\ninterface RTCIdentityProviderOptions {\n    peerIdentity?: string;\n    protocol?: string;\n    usernameHint?: string;\n}\n\ninterface RTCInboundRTPStreamStats extends RTCRTPStreamStats {\n    bytesReceived?: number;\n    fractionLost?: number;\n    jitter?: number;\n    packetsLost?: number;\n    packetsReceived?: number;\n}\n\ninterface RTCMediaStreamTrackStats extends RTCStats {\n    audioLevel?: number;\n    echoReturnLoss?: number;\n    echoReturnLossEnhancement?: number;\n    frameHeight?: number;\n    frameWidth?: number;\n    framesCorrupted?: number;\n    framesDecoded?: number;\n    framesDropped?: number;\n    framesPerSecond?: number;\n    framesReceived?: number;\n    framesSent?: number;\n    remoteSource?: boolean;\n    ssrcIds?: string[];\n    trackIdentifier?: string;\n}\n\ninterface RTCOAuthCredential {\n    accessToken: string;\n    macKey: string;\n}\n\ninterface RTCOfferAnswerOptions {\n    voiceActivityDetection?: boolean;\n}\n\ninterface RTCOfferOptions extends RTCOfferAnswerOptions {\n    iceRestart?: boolean;\n    offerToReceiveAudio?: boolean;\n    offerToReceiveVideo?: boolean;\n}\n\ninterface RTCOutboundRTPStreamStats extends RTCRTPStreamStats {\n    bytesSent?: number;\n    packetsSent?: number;\n    roundTripTime?: number;\n    targetBitrate?: number;\n}\n\ninterface RTCPeerConnectionIceErrorEventInit extends EventInit {\n    errorCode: number;\n    hostCandidate?: string;\n    statusText?: string;\n    url?: string;\n}\n\ninterface RTCPeerConnectionIceEventInit extends EventInit {\n    candidate?: RTCIceCandidate | null;\n    url?: string | null;\n}\n\ninterface RTCRTPStreamStats extends RTCStats {\n    associateStatsId?: string;\n    codecId?: string;\n    firCount?: number;\n    isRemote?: boolean;\n    mediaTrackId?: string;\n    mediaType?: string;\n    nackCount?: number;\n    pliCount?: number;\n    sliCount?: number;\n    ssrc?: string;\n    transportId?: string;\n}\n\ninterface RTCRtcpFeedback {\n    parameter?: string;\n    type?: string;\n}\n\ninterface RTCRtcpParameters {\n    cname?: string;\n    reducedSize?: boolean;\n}\n\ninterface RTCRtpCapabilities {\n    codecs: RTCRtpCodecCapability[];\n    headerExtensions: RTCRtpHeaderExtensionCapability[];\n}\n\ninterface RTCRtpCodecCapability {\n    channels?: number;\n    clockRate: number;\n    mimeType: string;\n    sdpFmtpLine?: string;\n}\n\ninterface RTCRtpCodecParameters {\n    channels?: number;\n    clockRate: number;\n    mimeType: string;\n    payloadType: number;\n    sdpFmtpLine?: string;\n}\n\ninterface RTCRtpCodingParameters {\n    rid?: string;\n}\n\ninterface RTCRtpContributingSource {\n    audioLevel?: number;\n    source: number;\n    timestamp: number;\n}\n\ninterface RTCRtpDecodingParameters extends RTCRtpCodingParameters {\n}\n\ninterface RTCRtpEncodingParameters extends RTCRtpCodingParameters {\n    active?: boolean;\n    codecPayloadType?: number;\n    dtx?: RTCDtxStatus;\n    maxBitrate?: number;\n    maxFramerate?: number;\n    priority?: RTCPriorityType;\n    ptime?: number;\n    scaleResolutionDownBy?: number;\n}\n\ninterface RTCRtpFecParameters {\n    mechanism?: string;\n    ssrc?: number;\n}\n\ninterface RTCRtpHeaderExtension {\n    kind?: string;\n    preferredEncrypt?: boolean;\n    preferredId?: number;\n    uri?: string;\n}\n\ninterface RTCRtpHeaderExtensionCapability {\n    uri?: string;\n}\n\ninterface RTCRtpHeaderExtensionParameters {\n    encrypted?: boolean;\n    id: number;\n    uri: string;\n}\n\ninterface RTCRtpParameters {\n    codecs: RTCRtpCodecParameters[];\n    headerExtensions: RTCRtpHeaderExtensionParameters[];\n    rtcp: RTCRtcpParameters;\n}\n\ninterface RTCRtpReceiveParameters extends RTCRtpParameters {\n    encodings: RTCRtpDecodingParameters[];\n}\n\ninterface RTCRtpRtxParameters {\n    ssrc?: number;\n}\n\ninterface RTCRtpSendParameters extends RTCRtpParameters {\n    degradationPreference?: RTCDegradationPreference;\n    encodings: RTCRtpEncodingParameters[];\n    transactionId: string;\n}\n\ninterface RTCRtpSynchronizationSource extends RTCRtpContributingSource {\n    voiceActivityFlag?: boolean;\n}\n\ninterface RTCRtpTransceiverInit {\n    direction?: RTCRtpTransceiverDirection;\n    sendEncodings?: RTCRtpEncodingParameters[];\n    streams?: MediaStream[];\n}\n\ninterface RTCRtpUnhandled {\n    muxId?: string;\n    payloadType?: number;\n    ssrc?: number;\n}\n\ninterface RTCSessionDescriptionInit {\n    sdp?: string;\n    type: RTCSdpType;\n}\n\ninterface RTCSrtpKeyParam {\n    keyMethod?: string;\n    keySalt?: string;\n    lifetime?: string;\n    mkiLength?: number;\n    mkiValue?: number;\n}\n\ninterface RTCSrtpSdesParameters {\n    cryptoSuite?: string;\n    keyParams?: RTCSrtpKeyParam[];\n    sessionParams?: string[];\n    tag?: number;\n}\n\ninterface RTCSsrcRange {\n    max?: number;\n    min?: number;\n}\n\ninterface RTCStats {\n    id: string;\n    timestamp: number;\n    type: RTCStatsType;\n}\n\ninterface RTCStatsEventInit extends EventInit {\n    report: RTCStatsReport;\n}\n\ninterface RTCStatsReport {\n}\n\ninterface RTCTrackEventInit extends EventInit {\n    receiver: RTCRtpReceiver;\n    streams?: MediaStream[];\n    track: MediaStreamTrack;\n    transceiver: RTCRtpTransceiver;\n}\n\ninterface RTCTransportStats extends RTCStats {\n    activeConnection?: boolean;\n    bytesReceived?: number;\n    bytesSent?: number;\n    localCertificateId?: string;\n    remoteCertificateId?: string;\n    rtcpTransportStatsId?: string;\n    selectedCandidatePairId?: string;\n}\n\ninterface RegistrationOptions {\n    scope?: string;\n    type?: WorkerType;\n    updateViaCache?: ServiceWorkerUpdateViaCache;\n}\n\ninterface RequestInit {\n    body?: BodyInit | null;\n    cache?: RequestCache;\n    credentials?: RequestCredentials;\n    headers?: HeadersInit;\n    integrity?: string;\n    keepalive?: boolean;\n    method?: string;\n    mode?: RequestMode;\n    redirect?: RequestRedirect;\n    referrer?: string;\n    referrerPolicy?: ReferrerPolicy;\n    signal?: AbortSignal | null;\n    window?: any;\n}\n\ninterface ResponseInit {\n    headers?: HeadersInit;\n    status?: number;\n    statusText?: string;\n}\n\ninterface RsaHashedImportParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n}\n\ninterface RsaHashedKeyAlgorithm extends RsaKeyAlgorithm {\n    hash: KeyAlgorithm;\n}\n\ninterface RsaHashedKeyGenParams extends RsaKeyGenParams {\n    hash: HashAlgorithmIdentifier;\n}\n\ninterface RsaKeyAlgorithm extends KeyAlgorithm {\n    modulusLength: number;\n    publicExponent: BigInteger;\n}\n\ninterface RsaKeyGenParams extends Algorithm {\n    modulusLength: number;\n    publicExponent: BigInteger;\n}\n\ninterface RsaOaepParams extends Algorithm {\n    label?: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface RsaOtherPrimesInfo {\n    d?: string;\n    r?: string;\n    t?: string;\n}\n\ninterface RsaPssParams extends Algorithm {\n    saltLength: number;\n}\n\ninterface SVGBoundingBoxOptions {\n    clipped?: boolean;\n    fill?: boolean;\n    markers?: boolean;\n    stroke?: boolean;\n}\n\ninterface ScopedCredentialDescriptor {\n    id: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null;\n    transports?: Transport[];\n    type: ScopedCredentialType;\n}\n\ninterface ScopedCredentialOptions {\n    excludeList?: ScopedCredentialDescriptor[];\n    extensions?: WebAuthnExtensions;\n    rpId?: string;\n    timeoutSeconds?: number;\n}\n\ninterface ScopedCredentialParameters {\n    algorithm: string | Algorithm;\n    type: ScopedCredentialType;\n}\n\ninterface ScrollIntoViewOptions extends ScrollOptions {\n    block?: ScrollLogicalPosition;\n    inline?: ScrollLogicalPosition;\n}\n\ninterface ScrollOptions {\n    behavior?: ScrollBehavior;\n}\n\ninterface ScrollToOptions extends ScrollOptions {\n    left?: number;\n    top?: number;\n}\n\ninterface SecurityPolicyViolationEventInit extends EventInit {\n    blockedURI?: string;\n    columnNumber?: number;\n    documentURI?: string;\n    effectiveDirective?: string;\n    lineNumber?: number;\n    originalPolicy?: string;\n    referrer?: string;\n    sourceFile?: string;\n    statusCode?: number;\n    violatedDirective?: string;\n}\n\ninterface ServiceWorkerMessageEventInit extends EventInit {\n    data?: any;\n    lastEventId?: string;\n    origin?: string;\n    ports?: MessagePort[] | null;\n    source?: ServiceWorker | MessagePort | null;\n}\n\ninterface StereoPannerOptions extends AudioNodeOptions {\n    pan?: number;\n}\n\ninterface StorageEstimate {\n    quota?: number;\n    usage?: number;\n}\n\ninterface StorageEventInit extends EventInit {\n    key?: string | null;\n    newValue?: string | null;\n    oldValue?: string | null;\n    storageArea?: Storage | null;\n    url?: string;\n}\n\ninterface StoreExceptionsInformation extends ExceptionInformation {\n    detailURI?: string | null;\n    explanationString?: string | null;\n    siteName?: string | null;\n}\n\ninterface StoreSiteSpecificExceptionsInformation extends StoreExceptionsInformation {\n    arrayOfDomainStrings?: string[];\n}\n\ninterface TextDecodeOptions {\n    stream?: boolean;\n}\n\ninterface TextDecoderOptions {\n    fatal?: boolean;\n    ignoreBOM?: boolean;\n}\n\ninterface TouchEventInit extends EventModifierInit {\n    changedTouches?: Touch[];\n    targetTouches?: Touch[];\n    touches?: Touch[];\n}\n\ninterface TouchInit {\n    altitudeAngle?: number;\n    azimuthAngle?: number;\n    clientX?: number;\n    clientY?: number;\n    force?: number;\n    identifier: number;\n    pageX?: number;\n    pageY?: number;\n    radiusX?: number;\n    radiusY?: number;\n    rotationAngle?: number;\n    screenX?: number;\n    screenY?: number;\n    target: EventTarget;\n    touchType?: TouchType;\n}\n\ninterface TrackEventInit extends EventInit {\n    track?: VideoTrack | AudioTrack | TextTrack | null;\n}\n\ninterface Transformer<I = any, O = any> {\n    flush?: TransformStreamDefaultControllerCallback<O>;\n    readableType?: undefined;\n    start?: TransformStreamDefaultControllerCallback<O>;\n    transform?: TransformStreamDefaultControllerTransformCallback<I, O>;\n    writableType?: undefined;\n}\n\ninterface TransitionEventInit extends EventInit {\n    elapsedTime?: number;\n    propertyName?: string;\n    pseudoElement?: string;\n}\n\ninterface UIEventInit extends EventInit {\n    detail?: number;\n    view?: Window | null;\n}\n\ninterface UnderlyingByteSource {\n    autoAllocateChunkSize?: number;\n    cancel?: ReadableStreamErrorCallback;\n    pull?: ReadableByteStreamControllerCallback;\n    start?: ReadableByteStreamControllerCallback;\n    type: "bytes";\n}\n\ninterface UnderlyingSink<W = any> {\n    abort?: WritableStreamErrorCallback;\n    close?: WritableStreamDefaultControllerCloseCallback;\n    start?: WritableStreamDefaultControllerStartCallback;\n    type?: undefined;\n    write?: WritableStreamDefaultControllerWriteCallback<W>;\n}\n\ninterface UnderlyingSource<R = any> {\n    cancel?: ReadableStreamErrorCallback;\n    pull?: ReadableStreamDefaultControllerCallback<R>;\n    start?: ReadableStreamDefaultControllerCallback<R>;\n    type?: undefined;\n}\n\ninterface VRDisplayEventInit extends EventInit {\n    display: VRDisplay;\n    reason?: VRDisplayEventReason;\n}\n\ninterface VRLayer {\n    leftBounds?: number[] | Float32Array | null;\n    rightBounds?: number[] | Float32Array | null;\n    source?: HTMLCanvasElement | null;\n}\n\ninterface VRStageParameters {\n    sittingToStandingTransform?: Float32Array;\n    sizeX?: number;\n    sizeY?: number;\n}\n\ninterface WaveShaperOptions extends AudioNodeOptions {\n    curve?: number[] | Float32Array;\n    oversample?: OverSampleType;\n}\n\ninterface WebAuthnExtensions {\n}\n\ninterface WebGLContextAttributes {\n    alpha?: GLboolean;\n    antialias?: GLboolean;\n    depth?: GLboolean;\n    failIfMajorPerformanceCaveat?: boolean;\n    powerPreference?: WebGLPowerPreference;\n    premultipliedAlpha?: GLboolean;\n    preserveDrawingBuffer?: GLboolean;\n    stencil?: GLboolean;\n}\n\ninterface WebGLContextEventInit extends EventInit {\n    statusMessage?: string;\n}\n\ninterface WheelEventInit extends MouseEventInit {\n    deltaMode?: number;\n    deltaX?: number;\n    deltaY?: number;\n    deltaZ?: number;\n}\n\ninterface WorkerOptions {\n    credentials?: RequestCredentials;\n    name?: string;\n    type?: WorkerType;\n}\n\ninterface WorkletOptions {\n    credentials?: RequestCredentials;\n}\n\ninterface EventListener {\n    (evt: Event): void;\n}\n\ninterface ANGLE_instanced_arrays {\n    drawArraysInstancedANGLE(mode: GLenum, first: GLint, count: GLsizei, primcount: GLsizei): void;\n    drawElementsInstancedANGLE(mode: GLenum, count: GLsizei, type: GLenum, offset: GLintptr, primcount: GLsizei): void;\n    vertexAttribDivisorANGLE(index: GLuint, divisor: GLuint): void;\n    readonly VERTEX_ATTRIB_ARRAY_DIVISOR_ANGLE: GLenum;\n}\n\ninterface AbortController {\n    /**\n     * Returns the AbortSignal object associated with this object.\n     */\n    readonly signal: AbortSignal;\n    /**\n     * Invoking this method will set this object\'s AbortSignal\'s aborted flag and\n     * signal to any observers that the associated activity is to be aborted.\n     */\n    abort(): void;\n}\n\ndeclare var AbortController: {\n    prototype: AbortController;\n    new(): AbortController;\n};\n\ninterface AbortSignalEventMap {\n    "abort": ProgressEvent;\n}\n\ninterface AbortSignal extends EventTarget {\n    /**\n     * Returns true if this AbortSignal\'s AbortController has signaled to abort, and false\n     * otherwise.\n     */\n    readonly aborted: boolean;\n    onabort: ((this: AbortSignal, ev: ProgressEvent) => any) | null;\n    addEventListener<K extends keyof AbortSignalEventMap>(type: K, listener: (this: AbortSignal, ev: AbortSignalEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AbortSignalEventMap>(type: K, listener: (this: AbortSignal, ev: AbortSignalEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var AbortSignal: {\n    prototype: AbortSignal;\n    new(): AbortSignal;\n};\n\ninterface AbstractRange {\n    readonly collapsed: boolean;\n    readonly endContainer: Node;\n    readonly endOffset: number;\n    readonly startContainer: Node;\n    readonly startOffset: number;\n}\n\ndeclare var AbstractRange: {\n    prototype: AbstractRange;\n    new(): AbstractRange;\n};\n\ninterface AbstractWorkerEventMap {\n    "error": ErrorEvent;\n}\n\ninterface AbstractWorker {\n    onerror: ((this: AbstractWorker, ev: ErrorEvent) => any) | null;\n    addEventListener<K extends keyof AbstractWorkerEventMap>(type: K, listener: (this: AbstractWorker, ev: AbstractWorkerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AbstractWorkerEventMap>(type: K, listener: (this: AbstractWorker, ev: AbstractWorkerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ninterface AesCfbParams extends Algorithm {\n    iv: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface AesCmacParams extends Algorithm {\n    length: number;\n}\n\ninterface AnalyserNode extends AudioNode {\n    fftSize: number;\n    readonly frequencyBinCount: number;\n    maxDecibels: number;\n    minDecibels: number;\n    smoothingTimeConstant: number;\n    getByteFrequencyData(array: Uint8Array): void;\n    getByteTimeDomainData(array: Uint8Array): void;\n    getFloatFrequencyData(array: Float32Array): void;\n    getFloatTimeDomainData(array: Float32Array): void;\n}\n\ndeclare var AnalyserNode: {\n    prototype: AnalyserNode;\n    new(context: BaseAudioContext, options?: AnalyserOptions): AnalyserNode;\n};\n\ninterface Animatable {\n    animate(keyframes: Keyframe[] | PropertyIndexedKeyframes | null, options?: number | KeyframeAnimationOptions): Animation;\n    getAnimations(): Animation[];\n}\n\ninterface AnimationEventMap {\n    "cancel": AnimationPlaybackEvent;\n    "finish": AnimationPlaybackEvent;\n}\n\ninterface Animation extends EventTarget {\n    currentTime: number | null;\n    effect: AnimationEffect | null;\n    readonly finished: Promise<Animation>;\n    id: string;\n    oncancel: ((this: Animation, ev: AnimationPlaybackEvent) => any) | null;\n    onfinish: ((this: Animation, ev: AnimationPlaybackEvent) => any) | null;\n    readonly pending: boolean;\n    readonly playState: AnimationPlayState;\n    playbackRate: number;\n    readonly ready: Promise<Animation>;\n    startTime: number | null;\n    timeline: AnimationTimeline | null;\n    cancel(): void;\n    finish(): void;\n    pause(): void;\n    play(): void;\n    reverse(): void;\n    updatePlaybackRate(playbackRate: number): void;\n    addEventListener<K extends keyof AnimationEventMap>(type: K, listener: (this: Animation, ev: AnimationEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AnimationEventMap>(type: K, listener: (this: Animation, ev: AnimationEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Animation: {\n    prototype: Animation;\n    new(effect?: AnimationEffect | null, timeline?: AnimationTimeline | null): Animation;\n};\n\ninterface AnimationEffect {\n    getComputedTiming(): ComputedEffectTiming;\n    getTiming(): EffectTiming;\n    updateTiming(timing?: OptionalEffectTiming): void;\n}\n\ndeclare var AnimationEffect: {\n    prototype: AnimationEffect;\n    new(): AnimationEffect;\n};\n\ninterface AnimationEvent extends Event {\n    readonly animationName: string;\n    readonly elapsedTime: number;\n    readonly pseudoElement: string;\n}\n\ndeclare var AnimationEvent: {\n    prototype: AnimationEvent;\n    new(type: string, animationEventInitDict?: AnimationEventInit): AnimationEvent;\n};\n\ninterface AnimationPlaybackEvent extends Event {\n    readonly currentTime: number | null;\n    readonly timelineTime: number | null;\n}\n\ndeclare var AnimationPlaybackEvent: {\n    prototype: AnimationPlaybackEvent;\n    new(type: string, eventInitDict?: AnimationPlaybackEventInit): AnimationPlaybackEvent;\n};\n\ninterface AnimationTimeline {\n    readonly currentTime: number | null;\n}\n\ndeclare var AnimationTimeline: {\n    prototype: AnimationTimeline;\n    new(): AnimationTimeline;\n};\n\ninterface ApplicationCacheEventMap {\n    "cached": Event;\n    "checking": Event;\n    "downloading": Event;\n    "error": Event;\n    "noupdate": Event;\n    "obsolete": Event;\n    "progress": ProgressEvent;\n    "updateready": Event;\n}\n\ninterface ApplicationCache extends EventTarget {\n    /** @deprecated */\n    oncached: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    onchecking: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    ondownloading: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    onerror: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    onnoupdate: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    onobsolete: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    onprogress: ((this: ApplicationCache, ev: ProgressEvent) => any) | null;\n    /** @deprecated */\n    onupdateready: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    readonly status: number;\n    /** @deprecated */\n    abort(): void;\n    /** @deprecated */\n    swapCache(): void;\n    /** @deprecated */\n    update(): void;\n    readonly CHECKING: number;\n    readonly DOWNLOADING: number;\n    readonly IDLE: number;\n    readonly OBSOLETE: number;\n    readonly UNCACHED: number;\n    readonly UPDATEREADY: number;\n    addEventListener<K extends keyof ApplicationCacheEventMap>(type: K, listener: (this: ApplicationCache, ev: ApplicationCacheEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ApplicationCacheEventMap>(type: K, listener: (this: ApplicationCache, ev: ApplicationCacheEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ApplicationCache: {\n    prototype: ApplicationCache;\n    new(): ApplicationCache;\n    readonly CHECKING: number;\n    readonly DOWNLOADING: number;\n    readonly IDLE: number;\n    readonly OBSOLETE: number;\n    readonly UNCACHED: number;\n    readonly UPDATEREADY: number;\n};\n\ninterface Attr extends Node {\n    readonly localName: string;\n    readonly name: string;\n    readonly namespaceURI: string | null;\n    readonly ownerElement: Element | null;\n    readonly prefix: string | null;\n    readonly specified: boolean;\n    value: string;\n}\n\ndeclare var Attr: {\n    prototype: Attr;\n    new(): Attr;\n};\n\ninterface AudioBuffer {\n    readonly duration: number;\n    readonly length: number;\n    readonly numberOfChannels: number;\n    readonly sampleRate: number;\n    copyFromChannel(destination: Float32Array, channelNumber: number, startInChannel?: number): void;\n    copyToChannel(source: Float32Array, channelNumber: number, startInChannel?: number): void;\n    getChannelData(channel: number): Float32Array;\n}\n\ndeclare var AudioBuffer: {\n    prototype: AudioBuffer;\n    new(options: AudioBufferOptions): AudioBuffer;\n};\n\ninterface AudioBufferSourceNode extends AudioScheduledSourceNode {\n    buffer: AudioBuffer | null;\n    readonly detune: AudioParam;\n    loop: boolean;\n    loopEnd: number;\n    loopStart: number;\n    readonly playbackRate: AudioParam;\n    start(when?: number, offset?: number, duration?: number): void;\n    addEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: AudioBufferSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: AudioBufferSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var AudioBufferSourceNode: {\n    prototype: AudioBufferSourceNode;\n    new(context: BaseAudioContext, options?: AudioBufferSourceOptions): AudioBufferSourceNode;\n};\n\ninterface AudioContext extends BaseAudioContext {\n    readonly baseLatency: number;\n    readonly outputLatency: number;\n    close(): Promise<void>;\n    createMediaElementSource(mediaElement: HTMLMediaElement): MediaElementAudioSourceNode;\n    createMediaStreamDestination(): MediaStreamAudioDestinationNode;\n    createMediaStreamSource(mediaStream: MediaStream): MediaStreamAudioSourceNode;\n    createMediaStreamTrackSource(mediaStreamTrack: MediaStreamTrack): MediaStreamTrackAudioSourceNode;\n    getOutputTimestamp(): AudioTimestamp;\n    suspend(): Promise<void>;\n    addEventListener<K extends keyof BaseAudioContextEventMap>(type: K, listener: (this: AudioContext, ev: BaseAudioContextEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof BaseAudioContextEventMap>(type: K, listener: (this: AudioContext, ev: BaseAudioContextEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var AudioContext: {\n    prototype: AudioContext;\n    new(contextOptions?: AudioContextOptions): AudioContext;\n};\n\ninterface AudioDestinationNode extends AudioNode {\n    readonly maxChannelCount: number;\n}\n\ndeclare var AudioDestinationNode: {\n    prototype: AudioDestinationNode;\n    new(): AudioDestinationNode;\n};\n\ninterface AudioListener {\n    readonly forwardX: AudioParam;\n    readonly forwardY: AudioParam;\n    readonly forwardZ: AudioParam;\n    readonly positionX: AudioParam;\n    readonly positionY: AudioParam;\n    readonly positionZ: AudioParam;\n    readonly upX: AudioParam;\n    readonly upY: AudioParam;\n    readonly upZ: AudioParam;\n    /** @deprecated */\n    setOrientation(x: number, y: number, z: number, xUp: number, yUp: number, zUp: number): void;\n    /** @deprecated */\n    setPosition(x: number, y: number, z: number): void;\n}\n\ndeclare var AudioListener: {\n    prototype: AudioListener;\n    new(): AudioListener;\n};\n\ninterface AudioNode extends EventTarget {\n    channelCount: number;\n    channelCountMode: ChannelCountMode;\n    channelInterpretation: ChannelInterpretation;\n    readonly context: BaseAudioContext;\n    readonly numberOfInputs: number;\n    readonly numberOfOutputs: number;\n    connect(destinationNode: AudioNode, output?: number, input?: number): AudioNode;\n    connect(destinationParam: AudioParam, output?: number): void;\n    disconnect(): void;\n    disconnect(output: number): void;\n    disconnect(destinationNode: AudioNode): void;\n    disconnect(destinationNode: AudioNode, output: number): void;\n    disconnect(destinationNode: AudioNode, output: number, input: number): void;\n    disconnect(destinationParam: AudioParam): void;\n    disconnect(destinationParam: AudioParam, output: number): void;\n}\n\ndeclare var AudioNode: {\n    prototype: AudioNode;\n    new(): AudioNode;\n};\n\ninterface AudioParam {\n    automationRate: AutomationRate;\n    readonly defaultValue: number;\n    readonly maxValue: number;\n    readonly minValue: number;\n    value: number;\n    cancelAndHoldAtTime(cancelTime: number): AudioParam;\n    cancelScheduledValues(cancelTime: number): AudioParam;\n    exponentialRampToValueAtTime(value: number, endTime: number): AudioParam;\n    linearRampToValueAtTime(value: number, endTime: number): AudioParam;\n    setTargetAtTime(target: number, startTime: number, timeConstant: number): AudioParam;\n    setValueAtTime(value: number, startTime: number): AudioParam;\n    setValueCurveAtTime(values: number[] | Float32Array, startTime: number, duration: number): AudioParam;\n}\n\ndeclare var AudioParam: {\n    prototype: AudioParam;\n    new(): AudioParam;\n};\n\ninterface AudioParamMap {\n    forEach(callbackfn: (value: AudioParam, key: string, parent: AudioParamMap) => void, thisArg?: any): void;\n}\n\ndeclare var AudioParamMap: {\n    prototype: AudioParamMap;\n    new(): AudioParamMap;\n};\n\ninterface AudioProcessingEvent extends Event {\n    readonly inputBuffer: AudioBuffer;\n    readonly outputBuffer: AudioBuffer;\n    readonly playbackTime: number;\n}\n\ndeclare var AudioProcessingEvent: {\n    prototype: AudioProcessingEvent;\n    new(type: string, eventInitDict: AudioProcessingEventInit): AudioProcessingEvent;\n};\n\ninterface AudioScheduledSourceNodeEventMap {\n    "ended": Event;\n}\n\ninterface AudioScheduledSourceNode extends AudioNode {\n    onended: ((this: AudioScheduledSourceNode, ev: Event) => any) | null;\n    start(when?: number): void;\n    stop(when?: number): void;\n    addEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: AudioScheduledSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: AudioScheduledSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var AudioScheduledSourceNode: {\n    prototype: AudioScheduledSourceNode;\n    new(): AudioScheduledSourceNode;\n};\n\ninterface AudioTrack {\n    enabled: boolean;\n    readonly id: string;\n    kind: string;\n    readonly label: string;\n    language: string;\n    readonly sourceBuffer: SourceBuffer;\n}\n\ndeclare var AudioTrack: {\n    prototype: AudioTrack;\n    new(): AudioTrack;\n};\n\ninterface AudioTrackListEventMap {\n    "addtrack": TrackEvent;\n    "change": Event;\n    "removetrack": TrackEvent;\n}\n\ninterface AudioTrackList extends EventTarget {\n    readonly length: number;\n    onaddtrack: ((this: AudioTrackList, ev: TrackEvent) => any) | null;\n    onchange: ((this: AudioTrackList, ev: Event) => any) | null;\n    onremovetrack: ((this: AudioTrackList, ev: TrackEvent) => any) | null;\n    getTrackById(id: string): AudioTrack | null;\n    item(index: number): AudioTrack;\n    addEventListener<K extends keyof AudioTrackListEventMap>(type: K, listener: (this: AudioTrackList, ev: AudioTrackListEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AudioTrackListEventMap>(type: K, listener: (this: AudioTrackList, ev: AudioTrackListEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n    [index: number]: AudioTrack;\n}\n\ndeclare var AudioTrackList: {\n    prototype: AudioTrackList;\n    new(): AudioTrackList;\n};\n\ninterface AudioWorklet extends Worklet {\n}\n\ndeclare var AudioWorklet: {\n    prototype: AudioWorklet;\n    new(): AudioWorklet;\n};\n\ninterface AudioWorkletNodeEventMap {\n    "processorerror": Event;\n}\n\ninterface AudioWorkletNode extends AudioNode {\n    onprocessorerror: ((this: AudioWorkletNode, ev: Event) => any) | null;\n    readonly parameters: AudioParamMap;\n    readonly port: MessagePort;\n    addEventListener<K extends keyof AudioWorkletNodeEventMap>(type: K, listener: (this: AudioWorkletNode, ev: AudioWorkletNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AudioWorkletNodeEventMap>(type: K, listener: (this: AudioWorkletNode, ev: AudioWorkletNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var AudioWorkletNode: {\n    prototype: AudioWorkletNode;\n    new(context: BaseAudioContext, name: string, options?: AudioWorkletNodeOptions): AudioWorkletNode;\n};\n\ninterface BarProp {\n    readonly visible: boolean;\n}\n\ndeclare var BarProp: {\n    prototype: BarProp;\n    new(): BarProp;\n};\n\ninterface BaseAudioContextEventMap {\n    "statechange": Event;\n}\n\ninterface BaseAudioContext extends EventTarget {\n    readonly audioWorklet: AudioWorklet;\n    readonly currentTime: number;\n    readonly destination: AudioDestinationNode;\n    readonly listener: AudioListener;\n    onstatechange: ((this: BaseAudioContext, ev: Event) => any) | null;\n    readonly sampleRate: number;\n    readonly state: AudioContextState;\n    createAnalyser(): AnalyserNode;\n    createBiquadFilter(): BiquadFilterNode;\n    createBuffer(numberOfChannels: number, length: number, sampleRate: number): AudioBuffer;\n    createBufferSource(): AudioBufferSourceNode;\n    createChannelMerger(numberOfInputs?: number): ChannelMergerNode;\n    createChannelSplitter(numberOfOutputs?: number): ChannelSplitterNode;\n    createConstantSource(): ConstantSourceNode;\n    createConvolver(): ConvolverNode;\n    createDelay(maxDelayTime?: number): DelayNode;\n    createDynamicsCompressor(): DynamicsCompressorNode;\n    createGain(): GainNode;\n    createIIRFilter(feedforward: number[], feedback: number[]): IIRFilterNode;\n    createOscillator(): OscillatorNode;\n    createPanner(): PannerNode;\n    createPeriodicWave(real: number[] | Float32Array, imag: number[] | Float32Array, constraints?: PeriodicWaveConstraints): PeriodicWave;\n    createScriptProcessor(bufferSize?: number, numberOfInputChannels?: number, numberOfOutputChannels?: number): ScriptProcessorNode;\n    createStereoPanner(): StereoPannerNode;\n    createWaveShaper(): WaveShaperNode;\n    decodeAudioData(audioData: ArrayBuffer, successCallback?: DecodeSuccessCallback | null, errorCallback?: DecodeErrorCallback | null): Promise<AudioBuffer>;\n    resume(): Promise<void>;\n    addEventListener<K extends keyof BaseAudioContextEventMap>(type: K, listener: (this: BaseAudioContext, ev: BaseAudioContextEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof BaseAudioContextEventMap>(type: K, listener: (this: BaseAudioContext, ev: BaseAudioContextEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var BaseAudioContext: {\n    prototype: BaseAudioContext;\n    new(): BaseAudioContext;\n};\n\ninterface BeforeUnloadEvent extends Event {\n    returnValue: any;\n}\n\ndeclare var BeforeUnloadEvent: {\n    prototype: BeforeUnloadEvent;\n    new(): BeforeUnloadEvent;\n};\n\ninterface BhxBrowser {\n    readonly lastError: DOMException;\n    checkMatchesGlobExpression(pattern: string, value: string): boolean;\n    checkMatchesUriExpression(pattern: string, value: string): boolean;\n    clearLastError(): void;\n    currentWindowId(): number;\n    fireExtensionApiTelemetry(functionName: string, isSucceeded: boolean, isSupported: boolean, errorString: string): void;\n    genericFunction(functionId: number, destination: any, parameters?: string, callbackId?: number): void;\n    genericSynchronousFunction(functionId: number, parameters?: string): string;\n    getExtensionId(): string;\n    getThisAddress(): any;\n    registerGenericFunctionCallbackHandler(callbackHandler: Function): void;\n    registerGenericListenerHandler(eventHandler: Function): void;\n    setLastError(parameters: string): void;\n    webPlatformGenericFunction(destination: any, parameters?: string, callbackId?: number): void;\n}\n\ndeclare var BhxBrowser: {\n    prototype: BhxBrowser;\n    new(): BhxBrowser;\n};\n\ninterface BiquadFilterNode extends AudioNode {\n    readonly Q: AudioParam;\n    readonly detune: AudioParam;\n    readonly frequency: AudioParam;\n    readonly gain: AudioParam;\n    type: BiquadFilterType;\n    getFrequencyResponse(frequencyHz: Float32Array, magResponse: Float32Array, phaseResponse: Float32Array): void;\n}\n\ndeclare var BiquadFilterNode: {\n    prototype: BiquadFilterNode;\n    new(context: BaseAudioContext, options?: BiquadFilterOptions): BiquadFilterNode;\n};\n\ninterface Blob {\n    readonly size: number;\n    readonly type: string;\n    slice(start?: number, end?: number, contentType?: string): Blob;\n}\n\ndeclare var Blob: {\n    prototype: Blob;\n    new(blobParts?: BlobPart[], options?: BlobPropertyBag): Blob;\n};\n\ninterface Body {\n    readonly body: ReadableStream<Uint8Array> | null;\n    readonly bodyUsed: boolean;\n    arrayBuffer(): Promise<ArrayBuffer>;\n    blob(): Promise<Blob>;\n    formData(): Promise<FormData>;\n    json(): Promise<any>;\n    text(): Promise<string>;\n}\n\ninterface BroadcastChannelEventMap {\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\ninterface BroadcastChannel extends EventTarget {\n    /**\n     * Returns the channel name (as passed to the constructor).\n     */\n    readonly name: string;\n    onmessage: ((this: BroadcastChannel, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: BroadcastChannel, ev: MessageEvent) => any) | null;\n    /**\n     * Closes the BroadcastChannel object, opening it up to garbage collection.\n     */\n    close(): void;\n    /**\n     * Sends the given message to other BroadcastChannel objects set up for this channel. Messages can be structured objects, e.g. nested objects and arrays.\n     */\n    postMessage(message: any): void;\n    addEventListener<K extends keyof BroadcastChannelEventMap>(type: K, listener: (this: BroadcastChannel, ev: BroadcastChannelEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof BroadcastChannelEventMap>(type: K, listener: (this: BroadcastChannel, ev: BroadcastChannelEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var BroadcastChannel: {\n    prototype: BroadcastChannel;\n    new(name: string): BroadcastChannel;\n};\n\ninterface BroadcastChannelEventMap {\n    message: MessageEvent;\n    messageerror: MessageEvent;\n}\n\ninterface ByteLengthQueuingStrategy extends QueuingStrategy<ArrayBufferView> {\n    highWaterMark: number;\n    size(chunk: ArrayBufferView): number;\n}\n\ndeclare var ByteLengthQueuingStrategy: {\n    prototype: ByteLengthQueuingStrategy;\n    new(options: { highWaterMark: number }): ByteLengthQueuingStrategy;\n};\n\ninterface CDATASection extends Text {\n}\n\ndeclare var CDATASection: {\n    prototype: CDATASection;\n    new(): CDATASection;\n};\n\ninterface CSS {\n    escape(value: string): string;\n    supports(property: string, value?: string): boolean;\n}\ndeclare var CSS: CSS;\n\ninterface CSSConditionRule extends CSSGroupingRule {\n    conditionText: string;\n}\n\ndeclare var CSSConditionRule: {\n    prototype: CSSConditionRule;\n    new(): CSSConditionRule;\n};\n\ninterface CSSFontFaceRule extends CSSRule {\n    readonly style: CSSStyleDeclaration;\n}\n\ndeclare var CSSFontFaceRule: {\n    prototype: CSSFontFaceRule;\n    new(): CSSFontFaceRule;\n};\n\ninterface CSSGroupingRule extends CSSRule {\n    readonly cssRules: CSSRuleList;\n    deleteRule(index: number): void;\n    insertRule(rule: string, index: number): number;\n}\n\ndeclare var CSSGroupingRule: {\n    prototype: CSSGroupingRule;\n    new(): CSSGroupingRule;\n};\n\ninterface CSSImportRule extends CSSRule {\n    readonly href: string;\n    readonly media: MediaList;\n    readonly styleSheet: CSSStyleSheet;\n}\n\ndeclare var CSSImportRule: {\n    prototype: CSSImportRule;\n    new(): CSSImportRule;\n};\n\ninterface CSSKeyframeRule extends CSSRule {\n    keyText: string;\n    readonly style: CSSStyleDeclaration;\n}\n\ndeclare var CSSKeyframeRule: {\n    prototype: CSSKeyframeRule;\n    new(): CSSKeyframeRule;\n};\n\ninterface CSSKeyframesRule extends CSSRule {\n    readonly cssRules: CSSRuleList;\n    name: string;\n    appendRule(rule: string): void;\n    deleteRule(select: string): void;\n    findRule(select: string): CSSKeyframeRule | null;\n}\n\ndeclare var CSSKeyframesRule: {\n    prototype: CSSKeyframesRule;\n    new(): CSSKeyframesRule;\n};\n\ninterface CSSMediaRule extends CSSConditionRule {\n    readonly media: MediaList;\n}\n\ndeclare var CSSMediaRule: {\n    prototype: CSSMediaRule;\n    new(): CSSMediaRule;\n};\n\ninterface CSSNamespaceRule extends CSSRule {\n    readonly namespaceURI: string;\n    readonly prefix: string;\n}\n\ndeclare var CSSNamespaceRule: {\n    prototype: CSSNamespaceRule;\n    new(): CSSNamespaceRule;\n};\n\ninterface CSSPageRule extends CSSRule {\n    readonly pseudoClass: string;\n    readonly selector: string;\n    selectorText: string;\n    readonly style: CSSStyleDeclaration;\n}\n\ndeclare var CSSPageRule: {\n    prototype: CSSPageRule;\n    new(): CSSPageRule;\n};\n\ninterface CSSRule {\n    cssText: string;\n    readonly parentRule: CSSRule | null;\n    readonly parentStyleSheet: CSSStyleSheet | null;\n    readonly type: number;\n    readonly CHARSET_RULE: number;\n    readonly FONT_FACE_RULE: number;\n    readonly IMPORT_RULE: number;\n    readonly KEYFRAMES_RULE: number;\n    readonly KEYFRAME_RULE: number;\n    readonly MEDIA_RULE: number;\n    readonly NAMESPACE_RULE: number;\n    readonly PAGE_RULE: number;\n    readonly STYLE_RULE: number;\n    readonly SUPPORTS_RULE: number;\n    readonly UNKNOWN_RULE: number;\n    readonly VIEWPORT_RULE: number;\n}\n\ndeclare var CSSRule: {\n    prototype: CSSRule;\n    new(): CSSRule;\n    readonly CHARSET_RULE: number;\n    readonly FONT_FACE_RULE: number;\n    readonly IMPORT_RULE: number;\n    readonly KEYFRAMES_RULE: number;\n    readonly KEYFRAME_RULE: number;\n    readonly MEDIA_RULE: number;\n    readonly NAMESPACE_RULE: number;\n    readonly PAGE_RULE: number;\n    readonly STYLE_RULE: number;\n    readonly SUPPORTS_RULE: number;\n    readonly UNKNOWN_RULE: number;\n    readonly VIEWPORT_RULE: number;\n};\n\ninterface CSSRuleList {\n    readonly length: number;\n    item(index: number): CSSRule | null;\n    [index: number]: CSSRule;\n}\n\ndeclare var CSSRuleList: {\n    prototype: CSSRuleList;\n    new(): CSSRuleList;\n};\n\ninterface CSSStyleDeclaration {\n    alignContent: string | null;\n    alignItems: string | null;\n    alignSelf: string | null;\n    alignmentBaseline: string | null;\n    animation: string;\n    animationDelay: string;\n    animationDirection: string;\n    animationDuration: string;\n    animationFillMode: string;\n    animationIterationCount: string;\n    animationName: string;\n    animationPlayState: string;\n    animationTimingFunction: string;\n    backfaceVisibility: string | null;\n    background: string | null;\n    backgroundAttachment: string | null;\n    backgroundClip: string | null;\n    backgroundColor: string | null;\n    backgroundImage: string | null;\n    backgroundOrigin: string | null;\n    backgroundPosition: string | null;\n    backgroundPositionX: string | null;\n    backgroundPositionY: string | null;\n    backgroundRepeat: string | null;\n    backgroundSize: string | null;\n    baselineShift: string | null;\n    border: string | null;\n    borderBottom: string | null;\n    borderBottomColor: string | null;\n    borderBottomLeftRadius: string | null;\n    borderBottomRightRadius: string | null;\n    borderBottomStyle: string | null;\n    borderBottomWidth: string | null;\n    borderCollapse: string | null;\n    borderColor: string | null;\n    borderImage: string | null;\n    borderImageOutset: string | null;\n    borderImageRepeat: string | null;\n    borderImageSlice: string | null;\n    borderImageSource: string | null;\n    borderImageWidth: string | null;\n    borderLeft: string | null;\n    borderLeftColor: string | null;\n    borderLeftStyle: string | null;\n    borderLeftWidth: string | null;\n    borderRadius: string | null;\n    borderRight: string | null;\n    borderRightColor: string | null;\n    borderRightStyle: string | null;\n    borderRightWidth: string | null;\n    borderSpacing: string | null;\n    borderStyle: string | null;\n    borderTop: string | null;\n    borderTopColor: string | null;\n    borderTopLeftRadius: string | null;\n    borderTopRightRadius: string | null;\n    borderTopStyle: string | null;\n    borderTopWidth: string | null;\n    borderWidth: string | null;\n    bottom: string | null;\n    boxShadow: string | null;\n    boxSizing: string | null;\n    breakAfter: string | null;\n    breakBefore: string | null;\n    breakInside: string | null;\n    captionSide: string | null;\n    clear: string | null;\n    clip: string | null;\n    clipPath: string | null;\n    clipRule: string | null;\n    color: string | null;\n    colorInterpolationFilters: string | null;\n    columnCount: any;\n    columnFill: string | null;\n    columnGap: any;\n    columnRule: string | null;\n    columnRuleColor: any;\n    columnRuleStyle: string | null;\n    columnRuleWidth: any;\n    columnSpan: string | null;\n    columnWidth: any;\n    columns: string | null;\n    content: string | null;\n    counterIncrement: string | null;\n    counterReset: string | null;\n    cssFloat: string | null;\n    cssText: string;\n    cursor: string | null;\n    direction: string | null;\n    display: string | null;\n    dominantBaseline: string | null;\n    emptyCells: string | null;\n    enableBackground: string | null;\n    fill: string | null;\n    fillOpacity: string | null;\n    fillRule: string | null;\n    filter: string | null;\n    flex: string | null;\n    flexBasis: string | null;\n    flexDirection: string | null;\n    flexFlow: string | null;\n    flexGrow: string | null;\n    flexShrink: string | null;\n    flexWrap: string | null;\n    floodColor: string | null;\n    floodOpacity: string | null;\n    font: string | null;\n    fontFamily: string | null;\n    fontFeatureSettings: string | null;\n    fontSize: string | null;\n    fontSizeAdjust: string | null;\n    fontStretch: string | null;\n    fontStyle: string | null;\n    fontVariant: string | null;\n    fontWeight: string | null;\n    gap: string | null;\n    glyphOrientationHorizontal: string | null;\n    glyphOrientationVertical: string | null;\n    grid: string | null;\n    gridArea: string | null;\n    gridAutoColumns: string | null;\n    gridAutoFlow: string | null;\n    gridAutoRows: string | null;\n    gridColumn: string | null;\n    gridColumnEnd: string | null;\n    gridColumnGap: string | null;\n    gridColumnStart: string | null;\n    gridGap: string | null;\n    gridRow: string | null;\n    gridRowEnd: string | null;\n    gridRowGap: string | null;\n    gridRowStart: string | null;\n    gridTemplate: string | null;\n    gridTemplateAreas: string | null;\n    gridTemplateColumns: string | null;\n    gridTemplateRows: string | null;\n    height: string | null;\n    imeMode: string | null;\n    justifyContent: string | null;\n    justifyItems: string | null;\n    justifySelf: string | null;\n    kerning: string | null;\n    layoutGrid: string | null;\n    layoutGridChar: string | null;\n    layoutGridLine: string | null;\n    layoutGridMode: string | null;\n    layoutGridType: string | null;\n    left: string | null;\n    readonly length: number;\n    letterSpacing: string | null;\n    lightingColor: string | null;\n    lineBreak: string | null;\n    lineHeight: string | null;\n    listStyle: string | null;\n    listStyleImage: string | null;\n    listStylePosition: string | null;\n    listStyleType: string | null;\n    margin: string | null;\n    marginBottom: string | null;\n    marginLeft: string | null;\n    marginRight: string | null;\n    marginTop: string | null;\n    marker: string | null;\n    markerEnd: string | null;\n    markerMid: string | null;\n    markerStart: string | null;\n    mask: string | null;\n    maskImage: string | null;\n    maxHeight: string | null;\n    maxWidth: string | null;\n    minHeight: string | null;\n    minWidth: string | null;\n    msContentZoomChaining: string | null;\n    msContentZoomLimit: string | null;\n    msContentZoomLimitMax: any;\n    msContentZoomLimitMin: any;\n    msContentZoomSnap: string | null;\n    msContentZoomSnapPoints: string | null;\n    msContentZoomSnapType: string | null;\n    msContentZooming: string | null;\n    msFlowFrom: string | null;\n    msFlowInto: string | null;\n    msFontFeatureSettings: string | null;\n    msGridColumn: any;\n    msGridColumnAlign: string | null;\n    msGridColumnSpan: any;\n    msGridColumns: string | null;\n    msGridRow: any;\n    msGridRowAlign: string | null;\n    msGridRowSpan: any;\n    msGridRows: string | null;\n    msHighContrastAdjust: string | null;\n    msHyphenateLimitChars: string | null;\n    msHyphenateLimitLines: any;\n    msHyphenateLimitZone: any;\n    msHyphens: string | null;\n    msImeAlign: string | null;\n    msOverflowStyle: string | null;\n    msScrollChaining: string | null;\n    msScrollLimit: string | null;\n    msScrollLimitXMax: any;\n    msScrollLimitXMin: any;\n    msScrollLimitYMax: any;\n    msScrollLimitYMin: any;\n    msScrollRails: string | null;\n    msScrollSnapPointsX: string | null;\n    msScrollSnapPointsY: string | null;\n    msScrollSnapType: string | null;\n    msScrollSnapX: string | null;\n    msScrollSnapY: string | null;\n    msScrollTranslation: string | null;\n    msTextCombineHorizontal: string | null;\n    msTextSizeAdjust: any;\n    msTouchAction: string | null;\n    msTouchSelect: string | null;\n    msUserSelect: string | null;\n    msWrapFlow: string;\n    msWrapMargin: any;\n    msWrapThrough: string;\n    objectFit: string | null;\n    objectPosition: string | null;\n    opacity: string | null;\n    order: string | null;\n    orphans: string | null;\n    outline: string | null;\n    outlineColor: string | null;\n    outlineOffset: string | null;\n    outlineStyle: string | null;\n    outlineWidth: string | null;\n    overflow: string | null;\n    overflowX: string | null;\n    overflowY: string | null;\n    padding: string | null;\n    paddingBottom: string | null;\n    paddingLeft: string | null;\n    paddingRight: string | null;\n    paddingTop: string | null;\n    pageBreakAfter: string | null;\n    pageBreakBefore: string | null;\n    pageBreakInside: string | null;\n    readonly parentRule: CSSRule;\n    penAction: string | null;\n    perspective: string | null;\n    perspectiveOrigin: string | null;\n    pointerEvents: string | null;\n    position: string | null;\n    quotes: string | null;\n    resize: string | null;\n    right: string | null;\n    rotate: string | null;\n    rowGap: string | null;\n    rubyAlign: string | null;\n    rubyOverhang: string | null;\n    rubyPosition: string | null;\n    scale: string | null;\n    scrollBehavior: string;\n    stopColor: string | null;\n    stopOpacity: string | null;\n    stroke: string | null;\n    strokeDasharray: string | null;\n    strokeDashoffset: string | null;\n    strokeLinecap: string | null;\n    strokeLinejoin: string | null;\n    strokeMiterlimit: string | null;\n    strokeOpacity: string | null;\n    strokeWidth: string | null;\n    tableLayout: string | null;\n    textAlign: string | null;\n    textAlignLast: string | null;\n    textAnchor: string | null;\n    textCombineUpright: string | null;\n    textDecoration: string | null;\n    textIndent: string | null;\n    textJustify: string | null;\n    textKashida: string | null;\n    textKashidaSpace: string | null;\n    textOverflow: string | null;\n    textShadow: string | null;\n    textTransform: string | null;\n    textUnderlinePosition: string | null;\n    top: string | null;\n    touchAction: string;\n    transform: string | null;\n    transformOrigin: string | null;\n    transformStyle: string | null;\n    transition: string;\n    transitionDelay: string;\n    transitionDuration: string;\n    transitionProperty: string;\n    transitionTimingFunction: string;\n    translate: string | null;\n    unicodeBidi: string | null;\n    userSelect: string | null;\n    verticalAlign: string | null;\n    visibility: string | null;\n    /** @deprecated */\n    webkitAlignContent: string;\n    /** @deprecated */\n    webkitAlignItems: string;\n    /** @deprecated */\n    webkitAlignSelf: string;\n    /** @deprecated */\n    webkitAnimation: string;\n    /** @deprecated */\n    webkitAnimationDelay: string;\n    /** @deprecated */\n    webkitAnimationDirection: string;\n    /** @deprecated */\n    webkitAnimationDuration: string;\n    /** @deprecated */\n    webkitAnimationFillMode: string;\n    /** @deprecated */\n    webkitAnimationIterationCount: string;\n    /** @deprecated */\n    webkitAnimationName: string;\n    /** @deprecated */\n    webkitAnimationPlayState: string;\n    /** @deprecated */\n    webkitAnimationTimingFunction: string;\n    /** @deprecated */\n    webkitAppearance: string;\n    /** @deprecated */\n    webkitBackfaceVisibility: string;\n    /** @deprecated */\n    webkitBackgroundClip: string;\n    /** @deprecated */\n    webkitBackgroundOrigin: string;\n    /** @deprecated */\n    webkitBackgroundSize: string;\n    /** @deprecated */\n    webkitBorderBottomLeftRadius: string;\n    /** @deprecated */\n    webkitBorderBottomRightRadius: string;\n    webkitBorderImage: string | null;\n    /** @deprecated */\n    webkitBorderRadius: string;\n    /** @deprecated */\n    webkitBorderTopLeftRadius: string;\n    /** @deprecated */\n    webkitBorderTopRightRadius: string;\n    /** @deprecated */\n    webkitBoxAlign: string;\n    webkitBoxDirection: string | null;\n    /** @deprecated */\n    webkitBoxFlex: string;\n    /** @deprecated */\n    webkitBoxOrdinalGroup: string;\n    webkitBoxOrient: string | null;\n    /** @deprecated */\n    webkitBoxPack: string;\n    /** @deprecated */\n    webkitBoxShadow: string;\n    /** @deprecated */\n    webkitBoxSizing: string;\n    webkitColumnBreakAfter: string | null;\n    webkitColumnBreakBefore: string | null;\n    webkitColumnBreakInside: string | null;\n    webkitColumnCount: any;\n    webkitColumnGap: any;\n    webkitColumnRule: string | null;\n    webkitColumnRuleColor: any;\n    webkitColumnRuleStyle: string | null;\n    webkitColumnRuleWidth: any;\n    webkitColumnSpan: string | null;\n    webkitColumnWidth: any;\n    webkitColumns: string | null;\n    /** @deprecated */\n    webkitFilter: string;\n    /** @deprecated */\n    webkitFlex: string;\n    /** @deprecated */\n    webkitFlexBasis: string;\n    /** @deprecated */\n    webkitFlexDirection: string;\n    /** @deprecated */\n    webkitFlexFlow: string;\n    /** @deprecated */\n    webkitFlexGrow: string;\n    /** @deprecated */\n    webkitFlexShrink: string;\n    /** @deprecated */\n    webkitFlexWrap: string;\n    /** @deprecated */\n    webkitJustifyContent: string;\n    /** @deprecated */\n    webkitMask: string;\n    /** @deprecated */\n    webkitMaskBoxImage: string;\n    /** @deprecated */\n    webkitMaskBoxImageOutset: string;\n    /** @deprecated */\n    webkitMaskBoxImageRepeat: string;\n    /** @deprecated */\n    webkitMaskBoxImageSlice: string;\n    /** @deprecated */\n    webkitMaskBoxImageSource: string;\n    /** @deprecated */\n    webkitMaskBoxImageWidth: string;\n    /** @deprecated */\n    webkitMaskClip: string;\n    /** @deprecated */\n    webkitMaskComposite: string;\n    /** @deprecated */\n    webkitMaskImage: string;\n    /** @deprecated */\n    webkitMaskOrigin: string;\n    /** @deprecated */\n    webkitMaskPosition: string;\n    /** @deprecated */\n    webkitMaskRepeat: string;\n    /** @deprecated */\n    webkitMaskSize: string;\n    /** @deprecated */\n    webkitOrder: string;\n    /** @deprecated */\n    webkitPerspective: string;\n    /** @deprecated */\n    webkitPerspectiveOrigin: string;\n    webkitTapHighlightColor: string | null;\n    /** @deprecated */\n    webkitTextFillColor: string;\n    /** @deprecated */\n    webkitTextSizeAdjust: string;\n    /** @deprecated */\n    webkitTextStroke: string;\n    /** @deprecated */\n    webkitTextStrokeColor: string;\n    /** @deprecated */\n    webkitTextStrokeWidth: string;\n    /** @deprecated */\n    webkitTransform: string;\n    /** @deprecated */\n    webkitTransformOrigin: string;\n    /** @deprecated */\n    webkitTransformStyle: string;\n    /** @deprecated */\n    webkitTransition: string;\n    /** @deprecated */\n    webkitTransitionDelay: string;\n    /** @deprecated */\n    webkitTransitionDuration: string;\n    /** @deprecated */\n    webkitTransitionProperty: string;\n    /** @deprecated */\n    webkitTransitionTimingFunction: string;\n    webkitUserModify: string | null;\n    webkitUserSelect: string | null;\n    webkitWritingMode: string | null;\n    whiteSpace: string | null;\n    widows: string | null;\n    width: string | null;\n    wordBreak: string | null;\n    wordSpacing: string | null;\n    wordWrap: string | null;\n    writingMode: string | null;\n    zIndex: string | null;\n    zoom: string | null;\n    getPropertyPriority(propertyName: string): string;\n    getPropertyValue(propertyName: string): string;\n    item(index: number): string;\n    removeProperty(propertyName: string): string;\n    setProperty(propertyName: string, value: string | null, priority?: string | null): void;\n    [index: number]: string;\n}\n\ndeclare var CSSStyleDeclaration: {\n    prototype: CSSStyleDeclaration;\n    new(): CSSStyleDeclaration;\n};\n\ninterface CSSStyleRule extends CSSRule {\n    selectorText: string;\n    readonly style: CSSStyleDeclaration;\n}\n\ndeclare var CSSStyleRule: {\n    prototype: CSSStyleRule;\n    new(): CSSStyleRule;\n};\n\ninterface CSSStyleSheet extends StyleSheet {\n    readonly cssRules: CSSRuleList;\n    /** @deprecated */\n    cssText: string;\n    /** @deprecated */\n    readonly id: string;\n    /** @deprecated */\n    readonly imports: StyleSheetList;\n    /** @deprecated */\n    readonly isAlternate: boolean;\n    /** @deprecated */\n    readonly isPrefAlternate: boolean;\n    readonly ownerRule: CSSRule | null;\n    /** @deprecated */\n    readonly owningElement: Element;\n    /** @deprecated */\n    readonly pages: any;\n    /** @deprecated */\n    readonly readOnly: boolean;\n    readonly rules: CSSRuleList;\n    /** @deprecated */\n    addImport(bstrURL: string, lIndex?: number): number;\n    /** @deprecated */\n    addPageRule(bstrSelector: string, bstrStyle: string, lIndex?: number): number;\n    addRule(bstrSelector: string, bstrStyle?: string, lIndex?: number): number;\n    deleteRule(index?: number): void;\n    insertRule(rule: string, index?: number): number;\n    /** @deprecated */\n    removeImport(lIndex: number): void;\n    removeRule(lIndex: number): void;\n}\n\ndeclare var CSSStyleSheet: {\n    prototype: CSSStyleSheet;\n    new(): CSSStyleSheet;\n};\n\ninterface CSSSupportsRule extends CSSConditionRule {\n}\n\ndeclare var CSSSupportsRule: {\n    prototype: CSSSupportsRule;\n    new(): CSSSupportsRule;\n};\n\ninterface Cache {\n    add(request: RequestInfo): Promise<void>;\n    addAll(requests: RequestInfo[]): Promise<void>;\n    delete(request: RequestInfo, options?: CacheQueryOptions): Promise<boolean>;\n    keys(request?: RequestInfo, options?: CacheQueryOptions): Promise<ReadonlyArray<Request>>;\n    match(request: RequestInfo, options?: CacheQueryOptions): Promise<Response | undefined>;\n    matchAll(request?: RequestInfo, options?: CacheQueryOptions): Promise<ReadonlyArray<Response>>;\n    put(request: RequestInfo, response: Response): Promise<void>;\n}\n\ndeclare var Cache: {\n    prototype: Cache;\n    new(): Cache;\n};\n\ninterface CacheStorage {\n    delete(cacheName: string): Promise<boolean>;\n    has(cacheName: string): Promise<boolean>;\n    keys(): Promise<string[]>;\n    match(request: RequestInfo, options?: CacheQueryOptions): Promise<Response | undefined>;\n    open(cacheName: string): Promise<Cache>;\n}\n\ndeclare var CacheStorage: {\n    prototype: CacheStorage;\n    new(): CacheStorage;\n};\n\ninterface CanvasCompositing {\n    globalAlpha: number;\n    globalCompositeOperation: string;\n}\n\ninterface CanvasDrawImage {\n    drawImage(image: CanvasImageSource, dx: number, dy: number): void;\n    drawImage(image: CanvasImageSource, dx: number, dy: number, dw: number, dh: number): void;\n    drawImage(image: CanvasImageSource, sx: number, sy: number, sw: number, sh: number, dx: number, dy: number, dw: number, dh: number): void;\n}\n\ninterface CanvasDrawPath {\n    beginPath(): void;\n    clip(fillRule?: CanvasFillRule): void;\n    clip(path: Path2D, fillRule?: CanvasFillRule): void;\n    fill(fillRule?: CanvasFillRule): void;\n    fill(path: Path2D, fillRule?: CanvasFillRule): void;\n    isPointInPath(x: number, y: number, fillRule?: CanvasFillRule): boolean;\n    isPointInPath(path: Path2D, x: number, y: number, fillRule?: CanvasFillRule): boolean;\n    isPointInStroke(x: number, y: number): boolean;\n    isPointInStroke(path: Path2D, x: number, y: number): boolean;\n    stroke(): void;\n    stroke(path: Path2D): void;\n}\n\ninterface CanvasFillStrokeStyles {\n    fillStyle: string | CanvasGradient | CanvasPattern;\n    strokeStyle: string | CanvasGradient | CanvasPattern;\n    createLinearGradient(x0: number, y0: number, x1: number, y1: number): CanvasGradient;\n    createPattern(image: CanvasImageSource, repetition: string): CanvasPattern | null;\n    createRadialGradient(x0: number, y0: number, r0: number, x1: number, y1: number, r1: number): CanvasGradient;\n}\n\ninterface CanvasFilters {\n    filter: string;\n}\n\ninterface CanvasGradient {\n    /**\n     * Adds a color stop with the given color to the gradient at the given offset. 0.0 is the offset\n     * at one end of the gradient, 1.0 is the offset at the other end.\n     * Throws an "IndexSizeError" DOMException if the offset\n     * is out of range. Throws a "SyntaxError" DOMException if\n     * the color cannot be parsed.\n     */\n    addColorStop(offset: number, color: string): void;\n}\n\ndeclare var CanvasGradient: {\n    prototype: CanvasGradient;\n    new(): CanvasGradient;\n};\n\ninterface CanvasImageData {\n    createImageData(sw: number, sh: number): ImageData;\n    createImageData(imagedata: ImageData): ImageData;\n    getImageData(sx: number, sy: number, sw: number, sh: number): ImageData;\n    putImageData(imagedata: ImageData, dx: number, dy: number): void;\n    putImageData(imagedata: ImageData, dx: number, dy: number, dirtyX: number, dirtyY: number, dirtyWidth: number, dirtyHeight: number): void;\n}\n\ninterface CanvasImageSmoothing {\n    imageSmoothingEnabled: boolean;\n    imageSmoothingQuality: ImageSmoothingQuality;\n}\n\ninterface CanvasPath {\n    arc(x: number, y: number, radius: number, startAngle: number, endAngle: number, anticlockwise?: boolean): void;\n    arcTo(x1: number, y1: number, x2: number, y2: number, radius: number): void;\n    bezierCurveTo(cp1x: number, cp1y: number, cp2x: number, cp2y: number, x: number, y: number): void;\n    closePath(): void;\n    ellipse(x: number, y: number, radiusX: number, radiusY: number, rotation: number, startAngle: number, endAngle: number, anticlockwise?: boolean): void;\n    lineTo(x: number, y: number): void;\n    moveTo(x: number, y: number): void;\n    quadraticCurveTo(cpx: number, cpy: number, x: number, y: number): void;\n    rect(x: number, y: number, w: number, h: number): void;\n}\n\ninterface CanvasPathDrawingStyles {\n    lineCap: CanvasLineCap;\n    lineDashOffset: number;\n    lineJoin: CanvasLineJoin;\n    lineWidth: number;\n    miterLimit: number;\n    getLineDash(): number[];\n    setLineDash(segments: number[]): void;\n}\n\ninterface CanvasPattern {\n    /**\n     * Sets the transformation matrix that will be used when rendering the pattern during a fill or\n     * stroke painting operation.\n     */\n    setTransform(transform?: DOMMatrix2DInit): void;\n}\n\ndeclare var CanvasPattern: {\n    prototype: CanvasPattern;\n    new(): CanvasPattern;\n};\n\ninterface CanvasRect {\n    clearRect(x: number, y: number, w: number, h: number): void;\n    fillRect(x: number, y: number, w: number, h: number): void;\n    strokeRect(x: number, y: number, w: number, h: number): void;\n}\n\ninterface CanvasRenderingContext2D extends CanvasState, CanvasTransform, CanvasCompositing, CanvasImageSmoothing, CanvasFillStrokeStyles, CanvasShadowStyles, CanvasFilters, CanvasRect, CanvasDrawPath, CanvasUserInterface, CanvasText, CanvasDrawImage, CanvasImageData, CanvasPathDrawingStyles, CanvasTextDrawingStyles, CanvasPath {\n    readonly canvas: HTMLCanvasElement;\n}\n\ndeclare var CanvasRenderingContext2D: {\n    prototype: CanvasRenderingContext2D;\n    new(): CanvasRenderingContext2D;\n};\n\ninterface CanvasShadowStyles {\n    shadowBlur: number;\n    shadowColor: string;\n    shadowOffsetX: number;\n    shadowOffsetY: number;\n}\n\ninterface CanvasState {\n    restore(): void;\n    save(): void;\n}\n\ninterface CanvasText {\n    fillText(text: string, x: number, y: number, maxWidth?: number): void;\n    measureText(text: string): TextMetrics;\n    strokeText(text: string, x: number, y: number, maxWidth?: number): void;\n}\n\ninterface CanvasTextDrawingStyles {\n    direction: CanvasDirection;\n    font: string;\n    textAlign: CanvasTextAlign;\n    textBaseline: CanvasTextBaseline;\n}\n\ninterface CanvasTransform {\n    getTransform(): DOMMatrix;\n    resetTransform(): void;\n    rotate(angle: number): void;\n    scale(x: number, y: number): void;\n    setTransform(a: number, b: number, c: number, d: number, e: number, f: number): void;\n    setTransform(transform?: DOMMatrix2DInit): void;\n    transform(a: number, b: number, c: number, d: number, e: number, f: number): void;\n    translate(x: number, y: number): void;\n}\n\ninterface CanvasUserInterface {\n    drawFocusIfNeeded(element: Element): void;\n    drawFocusIfNeeded(path: Path2D, element: Element): void;\n    scrollPathIntoView(): void;\n    scrollPathIntoView(path: Path2D): void;\n}\n\ninterface CaretPosition {\n    readonly offset: number;\n    readonly offsetNode: Node;\n    getClientRect(): DOMRect | null;\n}\n\ndeclare var CaretPosition: {\n    prototype: CaretPosition;\n    new(): CaretPosition;\n};\n\ninterface ChannelMergerNode extends AudioNode {\n}\n\ndeclare var ChannelMergerNode: {\n    prototype: ChannelMergerNode;\n    new(context: BaseAudioContext, options?: ChannelMergerOptions): ChannelMergerNode;\n};\n\ninterface ChannelSplitterNode extends AudioNode {\n}\n\ndeclare var ChannelSplitterNode: {\n    prototype: ChannelSplitterNode;\n    new(context: BaseAudioContext, options?: ChannelSplitterOptions): ChannelSplitterNode;\n};\n\ninterface CharacterData extends Node, NonDocumentTypeChildNode, ChildNode {\n    data: string;\n    readonly length: number;\n    appendData(data: string): void;\n    deleteData(offset: number, count: number): void;\n    insertData(offset: number, data: string): void;\n    replaceData(offset: number, count: number, data: string): void;\n    substringData(offset: number, count: number): string;\n}\n\ndeclare var CharacterData: {\n    prototype: CharacterData;\n    new(): CharacterData;\n};\n\ninterface ChildNode extends Node {\n    /**\n     * Inserts nodes just after node, while replacing strings in nodes with equivalent Text nodes.\n     * Throws a "HierarchyRequestError" DOMException if the constraints of\n     * the node tree are violated.\n     */\n    after(...nodes: (Node | string)[]): void;\n    /**\n     * Inserts nodes just before node, while replacing strings in nodes with equivalent Text nodes.\n     * Throws a "HierarchyRequestError" DOMException if the constraints of\n     * the node tree are violated.\n     */\n    before(...nodes: (Node | string)[]): void;\n    /**\n     * Removes node.\n     */\n    remove(): void;\n    /**\n     * Replaces node with nodes, while replacing strings in nodes with equivalent Text nodes.\n     * Throws a "HierarchyRequestError" DOMException if the constraints of\n     * the node tree are violated.\n     */\n    replaceWith(...nodes: (Node | string)[]): void;\n}\n\ninterface ClientRect {\n    bottom: number;\n    readonly height: number;\n    left: number;\n    right: number;\n    top: number;\n    readonly width: number;\n}\n\ndeclare var ClientRect: {\n    prototype: ClientRect;\n    new(): ClientRect;\n};\n\ninterface ClientRectList {\n    readonly length: number;\n    item(index: number): ClientRect;\n    [index: number]: ClientRect;\n}\n\ndeclare var ClientRectList: {\n    prototype: ClientRectList;\n    new(): ClientRectList;\n};\n\ninterface ClipboardEvent extends Event {\n    readonly clipboardData: DataTransfer;\n}\n\ndeclare var ClipboardEvent: {\n    prototype: ClipboardEvent;\n    new(type: string, eventInitDict?: ClipboardEventInit): ClipboardEvent;\n};\n\ninterface ClipboardEventInit extends EventInit {\n    data?: string;\n    dataType?: string;\n}\n\ninterface CloseEvent extends Event {\n    readonly code: number;\n    readonly reason: string;\n    readonly wasClean: boolean;\n    /** @deprecated */\n    initCloseEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, wasCleanArg: boolean, codeArg: number, reasonArg: string): void;\n}\n\ndeclare var CloseEvent: {\n    prototype: CloseEvent;\n    new(type: string, eventInitDict?: CloseEventInit): CloseEvent;\n};\n\ninterface Comment extends CharacterData {\n}\n\ndeclare var Comment: {\n    prototype: Comment;\n    new(data?: string): Comment;\n};\n\ninterface CompositionEvent extends UIEvent {\n    readonly data: string;\n    readonly locale: string;\n    initCompositionEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, dataArg: string, locale: string): void;\n}\n\ndeclare var CompositionEvent: {\n    prototype: CompositionEvent;\n    new(typeArg: string, eventInitDict?: CompositionEventInit): CompositionEvent;\n};\n\ninterface ConcatParams extends Algorithm {\n    algorithmId: Uint8Array;\n    hash?: string | Algorithm;\n    partyUInfo: Uint8Array;\n    partyVInfo: Uint8Array;\n    privateInfo?: Uint8Array;\n    publicInfo?: Uint8Array;\n}\n\ninterface Console {\n    memory: any;\n    assert(condition?: boolean, message?: string, ...data: any[]): void;\n    clear(): void;\n    count(label?: string): void;\n    debug(message?: any, ...optionalParams: any[]): void;\n    dir(value?: any, ...optionalParams: any[]): void;\n    dirxml(value: any): void;\n    error(message?: any, ...optionalParams: any[]): void;\n    exception(message?: string, ...optionalParams: any[]): void;\n    group(groupTitle?: string, ...optionalParams: any[]): void;\n    groupCollapsed(groupTitle?: string, ...optionalParams: any[]): void;\n    groupEnd(): void;\n    info(message?: any, ...optionalParams: any[]): void;\n    log(message?: any, ...optionalParams: any[]): void;\n    markTimeline(label?: string): void;\n    profile(reportName?: string): void;\n    profileEnd(reportName?: string): void;\n    table(...tabularData: any[]): void;\n    time(label?: string): void;\n    timeEnd(label?: string): void;\n    timeStamp(label?: string): void;\n    timeline(label?: string): void;\n    timelineEnd(label?: string): void;\n    trace(message?: any, ...optionalParams: any[]): void;\n    warn(message?: any, ...optionalParams: any[]): void;\n}\n\ndeclare var Console: {\n    prototype: Console;\n    new(): Console;\n};\n\ninterface ConstantSourceNode extends AudioScheduledSourceNode {\n    readonly offset: AudioParam;\n    addEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: ConstantSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: ConstantSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ConstantSourceNode: {\n    prototype: ConstantSourceNode;\n    new(context: BaseAudioContext, options?: ConstantSourceOptions): ConstantSourceNode;\n};\n\ninterface ConvolverNode extends AudioNode {\n    buffer: AudioBuffer | null;\n    normalize: boolean;\n}\n\ndeclare var ConvolverNode: {\n    prototype: ConvolverNode;\n    new(context: BaseAudioContext, options?: ConvolverOptions): ConvolverNode;\n};\n\ninterface Coordinates {\n    readonly accuracy: number;\n    readonly altitude: number | null;\n    readonly altitudeAccuracy: number | null;\n    readonly heading: number | null;\n    readonly latitude: number;\n    readonly longitude: number;\n    readonly speed: number | null;\n}\n\ninterface CountQueuingStrategy extends QueuingStrategy {\n    highWaterMark: number;\n    size(chunk: any): 1;\n}\n\ndeclare var CountQueuingStrategy: {\n    prototype: CountQueuingStrategy;\n    new(options: { highWaterMark: number }): CountQueuingStrategy;\n};\n\ninterface Crypto {\n    readonly subtle: SubtleCrypto;\n    getRandomValues<T extends Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | null>(array: T): T;\n}\n\ndeclare var Crypto: {\n    prototype: Crypto;\n    new(): Crypto;\n};\n\ninterface CryptoKey {\n    readonly algorithm: KeyAlgorithm;\n    readonly extractable: boolean;\n    readonly type: KeyType;\n    readonly usages: KeyUsage[];\n}\n\ndeclare var CryptoKey: {\n    prototype: CryptoKey;\n    new(): CryptoKey;\n};\n\ninterface CryptoKeyPair {\n    privateKey: CryptoKey;\n    publicKey: CryptoKey;\n}\n\ndeclare var CryptoKeyPair: {\n    prototype: CryptoKeyPair;\n    new(): CryptoKeyPair;\n};\n\ninterface CustomElementRegistry {\n    define(name: string, constructor: Function, options?: ElementDefinitionOptions): void;\n    get(name: string): any;\n    upgrade(root: Node): void;\n    whenDefined(name: string): Promise<void>;\n}\n\ndeclare var CustomElementRegistry: {\n    prototype: CustomElementRegistry;\n    new(): CustomElementRegistry;\n};\n\ninterface CustomEvent<T = any> extends Event {\n    /**\n     * Returns any custom data event was created with.\n     * Typically used for synthetic events.\n     */\n    readonly detail: T;\n    initCustomEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, detailArg: T): void;\n}\n\ndeclare var CustomEvent: {\n    prototype: CustomEvent;\n    new<T>(typeArg: string, eventInitDict?: CustomEventInit<T>): CustomEvent<T>;\n};\n\ninterface DOMError {\n    readonly name: string;\n    toString(): string;\n}\n\ndeclare var DOMError: {\n    prototype: DOMError;\n    new(): DOMError;\n};\n\ninterface DOMException {\n    readonly code: number;\n    readonly message: string;\n    readonly name: string;\n    readonly ABORT_ERR: number;\n    readonly DATA_CLONE_ERR: number;\n    readonly DOMSTRING_SIZE_ERR: number;\n    readonly HIERARCHY_REQUEST_ERR: number;\n    readonly INDEX_SIZE_ERR: number;\n    readonly INUSE_ATTRIBUTE_ERR: number;\n    readonly INVALID_ACCESS_ERR: number;\n    readonly INVALID_CHARACTER_ERR: number;\n    readonly INVALID_MODIFICATION_ERR: number;\n    readonly INVALID_NODE_TYPE_ERR: number;\n    readonly INVALID_STATE_ERR: number;\n    readonly NAMESPACE_ERR: number;\n    readonly NETWORK_ERR: number;\n    readonly NOT_FOUND_ERR: number;\n    readonly NOT_SUPPORTED_ERR: number;\n    readonly NO_DATA_ALLOWED_ERR: number;\n    readonly NO_MODIFICATION_ALLOWED_ERR: number;\n    readonly QUOTA_EXCEEDED_ERR: number;\n    readonly SECURITY_ERR: number;\n    readonly SYNTAX_ERR: number;\n    readonly TIMEOUT_ERR: number;\n    readonly TYPE_MISMATCH_ERR: number;\n    readonly URL_MISMATCH_ERR: number;\n    readonly VALIDATION_ERR: number;\n    readonly WRONG_DOCUMENT_ERR: number;\n}\n\ndeclare var DOMException: {\n    prototype: DOMException;\n    new(message?: string, name?: string): DOMException;\n    readonly ABORT_ERR: number;\n    readonly DATA_CLONE_ERR: number;\n    readonly DOMSTRING_SIZE_ERR: number;\n    readonly HIERARCHY_REQUEST_ERR: number;\n    readonly INDEX_SIZE_ERR: number;\n    readonly INUSE_ATTRIBUTE_ERR: number;\n    readonly INVALID_ACCESS_ERR: number;\n    readonly INVALID_CHARACTER_ERR: number;\n    readonly INVALID_MODIFICATION_ERR: number;\n    readonly INVALID_NODE_TYPE_ERR: number;\n    readonly INVALID_STATE_ERR: number;\n    readonly NAMESPACE_ERR: number;\n    readonly NETWORK_ERR: number;\n    readonly NOT_FOUND_ERR: number;\n    readonly NOT_SUPPORTED_ERR: number;\n    readonly NO_DATA_ALLOWED_ERR: number;\n    readonly NO_MODIFICATION_ALLOWED_ERR: number;\n    readonly QUOTA_EXCEEDED_ERR: number;\n    readonly SECURITY_ERR: number;\n    readonly SYNTAX_ERR: number;\n    readonly TIMEOUT_ERR: number;\n    readonly TYPE_MISMATCH_ERR: number;\n    readonly URL_MISMATCH_ERR: number;\n    readonly VALIDATION_ERR: number;\n    readonly WRONG_DOCUMENT_ERR: number;\n};\n\ninterface DOMImplementation {\n    createDocument(namespaceURI: string | null, qualifiedName: string | null, doctype: DocumentType | null): Document;\n    createDocumentType(qualifiedName: string, publicId: string, systemId: string): DocumentType;\n    createHTMLDocument(title?: string): Document;\n    /** @deprecated */\n    hasFeature(...args: any[]): true;\n}\n\ndeclare var DOMImplementation: {\n    prototype: DOMImplementation;\n    new(): DOMImplementation;\n};\n\ninterface DOML2DeprecatedColorProperty {\n    color: string;\n}\n\ninterface DOMMatrix extends DOMMatrixReadOnly {\n    a: number;\n    b: number;\n    c: number;\n    d: number;\n    e: number;\n    f: number;\n    m11: number;\n    m12: number;\n    m13: number;\n    m14: number;\n    m21: number;\n    m22: number;\n    m23: number;\n    m24: number;\n    m31: number;\n    m32: number;\n    m33: number;\n    m34: number;\n    m41: number;\n    m42: number;\n    m43: number;\n    m44: number;\n    invertSelf(): DOMMatrix;\n    multiplySelf(other?: DOMMatrixInit): DOMMatrix;\n    preMultiplySelf(other?: DOMMatrixInit): DOMMatrix;\n    rotateAxisAngleSelf(x?: number, y?: number, z?: number, angle?: number): DOMMatrix;\n    rotateFromVectorSelf(x?: number, y?: number): DOMMatrix;\n    rotateSelf(rotX?: number, rotY?: number, rotZ?: number): DOMMatrix;\n    scale3dSelf(scale?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    scaleSelf(scaleX?: number, scaleY?: number, scaleZ?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    setMatrixValue(transformList: string): DOMMatrix;\n    skewXSelf(sx?: number): DOMMatrix;\n    skewYSelf(sy?: number): DOMMatrix;\n    translateSelf(tx?: number, ty?: number, tz?: number): DOMMatrix;\n}\n\ndeclare var DOMMatrix: {\n    prototype: DOMMatrix;\n    new(init?: string | number[]): DOMMatrix;\n    fromFloat32Array(array32: Float32Array): DOMMatrix;\n    fromFloat64Array(array64: Float64Array): DOMMatrix;\n    fromMatrix(other?: DOMMatrixInit): DOMMatrix;\n};\n\ntype SVGMatrix = DOMMatrix;\ndeclare var SVGMatrix: typeof DOMMatrix;\n\ntype WebKitCSSMatrix = DOMMatrix;\ndeclare var WebKitCSSMatrix: typeof DOMMatrix;\n\ninterface DOMMatrixReadOnly {\n    readonly a: number;\n    readonly b: number;\n    readonly c: number;\n    readonly d: number;\n    readonly e: number;\n    readonly f: number;\n    readonly is2D: boolean;\n    readonly isIdentity: boolean;\n    readonly m11: number;\n    readonly m12: number;\n    readonly m13: number;\n    readonly m14: number;\n    readonly m21: number;\n    readonly m22: number;\n    readonly m23: number;\n    readonly m24: number;\n    readonly m31: number;\n    readonly m32: number;\n    readonly m33: number;\n    readonly m34: number;\n    readonly m41: number;\n    readonly m42: number;\n    readonly m43: number;\n    readonly m44: number;\n    flipX(): DOMMatrix;\n    flipY(): DOMMatrix;\n    inverse(): DOMMatrix;\n    multiply(other?: DOMMatrixInit): DOMMatrix;\n    rotate(rotX?: number, rotY?: number, rotZ?: number): DOMMatrix;\n    rotateAxisAngle(x?: number, y?: number, z?: number, angle?: number): DOMMatrix;\n    rotateFromVector(x?: number, y?: number): DOMMatrix;\n    scale(scaleX?: number, scaleY?: number, scaleZ?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    scale3d(scale?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    skewX(sx?: number): DOMMatrix;\n    skewY(sy?: number): DOMMatrix;\n    toFloat32Array(): Float32Array;\n    toFloat64Array(): Float64Array;\n    toJSON(): any;\n    transformPoint(point?: DOMPointInit): DOMPoint;\n    translate(tx?: number, ty?: number, tz?: number): DOMMatrix;\n}\n\ndeclare var DOMMatrixReadOnly: {\n    prototype: DOMMatrixReadOnly;\n    new(init?: string | number[]): DOMMatrixReadOnly;\n    fromFloat32Array(array32: Float32Array): DOMMatrixReadOnly;\n    fromFloat64Array(array64: Float64Array): DOMMatrixReadOnly;\n    fromMatrix(other?: DOMMatrixInit): DOMMatrixReadOnly;\n};\n\ninterface DOMParser {\n    parseFromString(str: string, type: SupportedType): Document;\n}\n\ndeclare var DOMParser: {\n    prototype: DOMParser;\n    new(): DOMParser;\n};\n\ninterface DOMPoint extends DOMPointReadOnly {\n    w: number;\n    x: number;\n    y: number;\n    z: number;\n}\n\ndeclare var DOMPoint: {\n    prototype: DOMPoint;\n    new(x?: number, y?: number, z?: number, w?: number): DOMPoint;\n    fromPoint(other?: DOMPointInit): DOMPoint;\n};\n\ntype SVGPoint = DOMPoint;\ndeclare var SVGPoint: typeof DOMPoint;\n\ninterface DOMPointReadOnly {\n    readonly w: number;\n    readonly x: number;\n    readonly y: number;\n    readonly z: number;\n    matrixTransform(matrix?: DOMMatrixInit): DOMPoint;\n    toJSON(): any;\n}\n\ndeclare var DOMPointReadOnly: {\n    prototype: DOMPointReadOnly;\n    new(x?: number, y?: number, z?: number, w?: number): DOMPointReadOnly;\n    fromPoint(other?: DOMPointInit): DOMPointReadOnly;\n};\n\ninterface DOMQuad {\n    readonly p1: DOMPoint;\n    readonly p2: DOMPoint;\n    readonly p3: DOMPoint;\n    readonly p4: DOMPoint;\n    getBounds(): DOMRect;\n    toJSON(): any;\n}\n\ndeclare var DOMQuad: {\n    prototype: DOMQuad;\n    new(p1?: DOMPointInit, p2?: DOMPointInit, p3?: DOMPointInit, p4?: DOMPointInit): DOMQuad;\n    fromQuad(other?: DOMQuadInit): DOMQuad;\n    fromRect(other?: DOMRectInit): DOMQuad;\n};\n\ninterface DOMRect extends DOMRectReadOnly {\n    height: number;\n    width: number;\n    x: number;\n    y: number;\n}\n\ndeclare var DOMRect: {\n    prototype: DOMRect;\n    new(x?: number, y?: number, width?: number, height?: number): DOMRect;\n    fromRect(other?: DOMRectInit): DOMRect;\n};\n\ntype SVGRect = DOMRect;\ndeclare var SVGRect: typeof DOMRect;\n\ninterface DOMRectList {\n    readonly length: number;\n    item(index: number): DOMRect | null;\n    [index: number]: DOMRect;\n}\n\ndeclare var DOMRectList: {\n    prototype: DOMRectList;\n    new(): DOMRectList;\n};\n\ninterface DOMRectReadOnly {\n    readonly bottom: number;\n    readonly height: number;\n    readonly left: number;\n    readonly right: number;\n    readonly top: number;\n    readonly width: number;\n    readonly x: number;\n    readonly y: number;\n    toJSON(): any;\n}\n\ndeclare var DOMRectReadOnly: {\n    prototype: DOMRectReadOnly;\n    new(x?: number, y?: number, width?: number, height?: number): DOMRectReadOnly;\n    fromRect(other?: DOMRectInit): DOMRectReadOnly;\n};\n\ninterface DOMSettableTokenList extends DOMTokenList {\n    value: string;\n}\n\ndeclare var DOMSettableTokenList: {\n    prototype: DOMSettableTokenList;\n    new(): DOMSettableTokenList;\n};\n\ninterface DOMStringList {\n    /**\n     * Returns the number of strings in strings.\n     */\n    readonly length: number;\n    /**\n     * Returns true if strings contains string, and false\n     * otherwise.\n     */\n    contains(string: string): boolean;\n    /**\n     * Returns the string with index index from strings.\n     */\n    item(index: number): string | null;\n    [index: number]: string;\n}\n\ndeclare var DOMStringList: {\n    prototype: DOMStringList;\n    new(): DOMStringList;\n};\n\ninterface DOMStringMap {\n    [name: string]: string | undefined;\n}\n\ndeclare var DOMStringMap: {\n    prototype: DOMStringMap;\n    new(): DOMStringMap;\n};\n\ninterface DOMTokenList {\n    /**\n     * Returns the number of tokens.\n     */\n    readonly length: number;\n    /**\n     * Returns the associated set as string.\n     * Can be set, to change the associated attribute.\n     */\n    value: string;\n    /**\n     * Adds all arguments passed, except those already present.\n     * Throws a "SyntaxError" DOMException if one of the arguments is the empty\n     * string.\n     * Throws an "InvalidCharacterError" DOMException if one of the arguments\n     * contains any ASCII whitespace.\n     */\n    add(...tokens: string[]): void;\n    /**\n     * Returns true if token is present, and false otherwise.\n     */\n    contains(token: string): boolean;\n    /**\n     * tokenlist[index]\n     */\n    item(index: number): string | null;\n    /**\n     * Removes arguments passed, if they are present.\n     * Throws a "SyntaxError" DOMException if one of the arguments is the empty\n     * string.\n     * Throws an "InvalidCharacterError" DOMException if one of the arguments\n     * contains any ASCII whitespace.\n     */\n    remove(...tokens: string[]): void;\n    /**\n     * Replaces token with newToken.\n     * Returns true if token was replaced with newToken, and false otherwise.\n     * Throws a "SyntaxError" DOMException if one of the arguments is the empty\n     * string.\n     * Throws an "InvalidCharacterError" DOMException if one of the arguments\n     * contains any ASCII whitespace.\n     */\n    replace(oldToken: string, newToken: string): void;\n    /**\n     * Returns true if token is in the associated attribute\'s supported tokens. Returns\n     * false otherwise.\n     * Throws a TypeError if the associated attribute has no supported tokens defined.\n     */\n    supports(token: string): boolean;\n    toggle(token: string, force?: boolean): boolean;\n    forEach(callbackfn: (value: string, key: number, parent: DOMTokenList) => void, thisArg?: any): void;\n    [index: number]: string;\n}\n\ndeclare var DOMTokenList: {\n    prototype: DOMTokenList;\n    new(): DOMTokenList;\n};\n\ninterface DataCue extends TextTrackCue {\n    data: ArrayBuffer;\n    addEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: DataCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: DataCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var DataCue: {\n    prototype: DataCue;\n    new(): DataCue;\n};\n\ninterface DataTransfer {\n    dropEffect: string;\n    effectAllowed: string;\n    /**\n     * Returns a FileList of the files being dragged, if any.\n     */\n    readonly files: FileList;\n    /**\n     * Returns a DataTransferItemList object, with the drag data.\n     */\n    readonly items: DataTransferItemList;\n    /**\n     * Returns a frozen array listing the formats that were set in the dragstart event. In addition, if any files are being\n     * dragged, then one of the types will be the string "Files".\n     */\n    readonly types: ReadonlyArray<string>;\n    /**\n     * Removes the data of the specified formats. Removes all data if the argument is omitted.\n     */\n    clearData(format?: string): void;\n    /**\n     * Returns the specified data. If there is no such data, returns the empty string.\n     */\n    getData(format: string): string;\n    /**\n     * Adds the specified data.\n     */\n    setData(format: string, data: string): void;\n    /**\n     * Uses the given element to update the drag feedback, replacing any previously specified\n     * feedback.\n     */\n    setDragImage(image: Element, x: number, y: number): void;\n}\n\ndeclare var DataTransfer: {\n    prototype: DataTransfer;\n    new(): DataTransfer;\n};\n\ninterface DataTransferItem {\n    /**\n     * Returns the drag data item kind, one of: "string",\n     * "file".\n     */\n    readonly kind: string;\n    /**\n     * Returns the drag data item type string.\n     */\n    readonly type: string;\n    /**\n     * Returns a File object, if the drag data item kind is File.\n     */\n    getAsFile(): File | null;\n    /**\n     * Invokes the callback with the string data as the argument, if the drag data item\n     * kind is Plain Unicode string.\n     */\n    getAsString(callback: FunctionStringCallback | null): void;\n    webkitGetAsEntry(): any;\n}\n\ndeclare var DataTransferItem: {\n    prototype: DataTransferItem;\n    new(): DataTransferItem;\n};\n\ninterface DataTransferItemList {\n    /**\n     * Returns the number of items in the drag data store.\n     */\n    readonly length: number;\n    /**\n     * Adds a new entry for the given data to the drag data store. If the data is plain\n     * text  then a type string has to be provided\n     * also.\n     */\n    add(data: string, type: string): DataTransferItem | null;\n    add(data: File): DataTransferItem | null;\n    /**\n     * Removes all the entries in the drag data store.\n     */\n    clear(): void;\n    item(index: number): DataTransferItem;\n    /**\n     * Removes the indexth entry in the drag data store.\n     */\n    remove(index: number): void;\n    [name: number]: DataTransferItem;\n}\n\ndeclare var DataTransferItemList: {\n    prototype: DataTransferItemList;\n    new(): DataTransferItemList;\n};\n\ninterface DeferredPermissionRequest {\n    readonly id: number;\n    readonly type: MSWebViewPermissionType;\n    readonly uri: string;\n    allow(): void;\n    deny(): void;\n}\n\ndeclare var DeferredPermissionRequest: {\n    prototype: DeferredPermissionRequest;\n    new(): DeferredPermissionRequest;\n};\n\ninterface DelayNode extends AudioNode {\n    readonly delayTime: AudioParam;\n}\n\ndeclare var DelayNode: {\n    prototype: DelayNode;\n    new(context: BaseAudioContext, options?: DelayOptions): DelayNode;\n};\n\ninterface DeviceAcceleration {\n    readonly x: number | null;\n    readonly y: number | null;\n    readonly z: number | null;\n}\n\ndeclare var DeviceAcceleration: {\n    prototype: DeviceAcceleration;\n    new(): DeviceAcceleration;\n};\n\ninterface DeviceLightEvent extends Event {\n    readonly value: number;\n}\n\ndeclare var DeviceLightEvent: {\n    prototype: DeviceLightEvent;\n    new(typeArg: string, eventInitDict?: DeviceLightEventInit): DeviceLightEvent;\n};\n\ninterface DeviceMotionEvent extends Event {\n    readonly acceleration: DeviceAcceleration | null;\n    readonly accelerationIncludingGravity: DeviceAcceleration | null;\n    readonly interval: number | null;\n    readonly rotationRate: DeviceRotationRate | null;\n    initDeviceMotionEvent(type: string, bubbles: boolean, cancelable: boolean, acceleration: DeviceAccelerationDict | null, accelerationIncludingGravity: DeviceAccelerationDict | null, rotationRate: DeviceRotationRateDict | null, interval: number | null): void;\n}\n\ndeclare var DeviceMotionEvent: {\n    prototype: DeviceMotionEvent;\n    new(typeArg: string, eventInitDict?: DeviceMotionEventInit): DeviceMotionEvent;\n};\n\ninterface DeviceOrientationEvent extends Event {\n    readonly absolute: boolean;\n    readonly alpha: number | null;\n    readonly beta: number | null;\n    readonly gamma: number | null;\n    initDeviceOrientationEvent(type: string, bubbles: boolean, cancelable: boolean, alpha: number | null, beta: number | null, gamma: number | null, absolute: boolean): void;\n}\n\ndeclare var DeviceOrientationEvent: {\n    prototype: DeviceOrientationEvent;\n    new(typeArg: string, eventInitDict?: DeviceOrientationEventInit): DeviceOrientationEvent;\n};\n\ninterface DeviceRotationRate {\n    readonly alpha: number | null;\n    readonly beta: number | null;\n    readonly gamma: number | null;\n}\n\ndeclare var DeviceRotationRate: {\n    prototype: DeviceRotationRate;\n    new(): DeviceRotationRate;\n};\n\ninterface DhImportKeyParams extends Algorithm {\n    generator: Uint8Array;\n    prime: Uint8Array;\n}\n\ninterface DhKeyAlgorithm extends KeyAlgorithm {\n    generator: Uint8Array;\n    prime: Uint8Array;\n}\n\ninterface DhKeyDeriveParams extends Algorithm {\n    public: CryptoKey;\n}\n\ninterface DhKeyGenParams extends Algorithm {\n    generator: Uint8Array;\n    prime: Uint8Array;\n}\n\ninterface DocumentEventMap extends GlobalEventHandlersEventMap, DocumentAndElementEventHandlersEventMap {\n    "fullscreenchange": Event;\n    "fullscreenerror": Event;\n    "readystatechange": ProgressEvent;\n    "visibilitychange": Event;\n}\n\ninterface Document extends Node, NonElementParentNode, DocumentOrShadowRoot, ParentNode, GlobalEventHandlers, DocumentAndElementEventHandlers {\n    /**\n     * Sets or gets the URL for the current document.\n     */\n    readonly URL: string;\n    /**\n     * Gets the object that has the focus when the parent document has focus.\n     */\n    readonly activeElement: Element | null;\n    /**\n     * Sets or gets the color of all active links in the document.\n     */\n    /** @deprecated */\n    alinkColor: string;\n    /**\n     * Returns a reference to the collection of elements contained by the object.\n     */\n    /** @deprecated */\n    readonly all: HTMLAllCollection;\n    /**\n     * Retrieves a collection of all a objects that have a name and/or id property. Objects in this collection are in HTML source order.\n     */\n    /** @deprecated */\n    readonly anchors: HTMLCollectionOf<HTMLAnchorElement>;\n    /**\n     * Retrieves a collection of all applet objects in the document.\n     */\n    /** @deprecated */\n    readonly applets: HTMLCollectionOf<HTMLAppletElement>;\n    /**\n     * Deprecated. Sets or retrieves a value that indicates the background color behind the object.\n     */\n    /** @deprecated */\n    bgColor: string;\n    /**\n     * Specifies the beginning and end of the document body.\n     */\n    body: HTMLElement;\n    /**\n     * Returns document\'s encoding.\n     */\n    readonly characterSet: string;\n    /**\n     * Gets or sets the character set used to encode the object.\n     */\n    readonly charset: string;\n    /**\n     * Gets a value that indicates whether standards-compliant mode is switched on for the object.\n     */\n    readonly compatMode: string;\n    /**\n     * Returns document\'s content type.\n     */\n    readonly contentType: string;\n    /**\n     * Returns the HTTP cookies that apply to the Document. If there are no cookies or\n     * cookies can\'t be applied to this resource, the empty string will be returned.\n     * Can be set, to add a new cookie to the element\'s set of HTTP cookies.\n     * If the contents are sandboxed into a\n     * unique origin (e.g. in an iframe with the sandbox attribute), a\n     * "SecurityError" DOMException will be thrown on getting\n     * and setting.\n     */\n    cookie: string;\n    /**\n     * Returns the script element, or the SVG script element,\n     * that is currently executing, as long as the element represents a classic script.\n     * In the case of reentrant script execution, returns the one that most recently started executing\n     * amongst those that have not yet finished executing.\n     * Returns null if the Document is not currently executing a script\n     * or SVG script element (e.g., because the running script is an event\n     * handler, or a timeout), or if the currently executing script or SVG\n     * script element represents a module script.\n     */\n    readonly currentScript: HTMLOrSVGScriptElement | null;\n    readonly defaultView: WindowProxy | null;\n    /**\n     * Sets or gets a value that indicates whether the document can be edited.\n     */\n    designMode: string;\n    /**\n     * Sets or retrieves a value that indicates the reading order of the object.\n     */\n    dir: string;\n    /**\n     * Gets an object representing the document type declaration associated with the current document.\n     */\n    readonly doctype: DocumentType | null;\n    /**\n     * Gets a reference to the root node of the document.\n     */\n    readonly documentElement: HTMLElement;\n    /**\n     * Returns document\'s URL.\n     */\n    readonly documentURI: string;\n    /**\n     * Sets or gets the security domain of the document.\n     */\n    domain: string;\n    /**\n     * Retrieves a collection of all embed objects in the document.\n     */\n    readonly embeds: HTMLCollectionOf<HTMLEmbedElement>;\n    /**\n     * Sets or gets the foreground (text) color of the document.\n     */\n    /** @deprecated */\n    fgColor: string;\n    /**\n     * Retrieves a collection, in source order, of all form objects in the document.\n     */\n    readonly forms: HTMLCollectionOf<HTMLFormElement>;\n    /** @deprecated */\n    readonly fullscreen: boolean;\n    /**\n     * Returns true if document has the ability to display elements fullscreen\n     * and fullscreen is supported, or false otherwise.\n     */\n    readonly fullscreenEnabled: boolean;\n    /**\n     * Returns the head element.\n     */\n    readonly head: HTMLHeadElement;\n    readonly hidden: boolean;\n    /**\n     * Retrieves a collection, in source order, of img objects in the document.\n     */\n    readonly images: HTMLCollectionOf<HTMLImageElement>;\n    /**\n     * Gets the implementation object of the current document.\n     */\n    readonly implementation: DOMImplementation;\n    /**\n     * Returns the character encoding used to create the webpage that is loaded into the document object.\n     */\n    readonly inputEncoding: string;\n    /**\n     * Gets the date that the page was last modified, if the page supplies one.\n     */\n    readonly lastModified: string;\n    /**\n     * Sets or gets the color of the document links.\n     */\n    /** @deprecated */\n    linkColor: string;\n    /**\n     * Retrieves a collection of all a objects that specify the href property and all area objects in the document.\n     */\n    readonly links: HTMLCollectionOf<HTMLAnchorElement | HTMLAreaElement>;\n    /**\n     * Contains information about the current URL.\n     */\n    location: Location;\n    onfullscreenchange: ((this: Document, ev: Event) => any) | null;\n    onfullscreenerror: ((this: Document, ev: Event) => any) | null;\n    /**\n     * Fires when the state of the object has changed.\n     * @param ev The event\n     */\n    onreadystatechange: ((this: Document, ev: ProgressEvent) => any) | null;\n    onvisibilitychange: ((this: Document, ev: Event) => any) | null;\n    /**\n     * Returns document\'s origin.\n     */\n    readonly origin: string;\n    /**\n     * Return an HTMLCollection of the embed elements in the Document.\n     */\n    readonly plugins: HTMLCollectionOf<HTMLEmbedElement>;\n    /**\n     * Retrieves a value that indicates the current state of the object.\n     */\n    readonly readyState: DocumentReadyState;\n    /**\n     * Gets the URL of the location that referred the user to the current page.\n     */\n    readonly referrer: string;\n    /**\n     * Retrieves a collection of all script objects in the document.\n     */\n    readonly scripts: HTMLCollectionOf<HTMLScriptElement>;\n    readonly scrollingElement: Element | null;\n    readonly timeline: DocumentTimeline;\n    /**\n     * Contains the title of the document.\n     */\n    title: string;\n    readonly visibilityState: VisibilityState;\n    /**\n     * Sets or gets the color of the links that the user has visited.\n     */\n    /** @deprecated */\n    vlinkColor: string;\n    /**\n     * Moves node from another document and returns it.\n     * If node is a document, throws a "NotSupportedError" DOMException or, if node is a shadow root, throws a\n     * "HierarchyRequestError" DOMException.\n     */\n    adoptNode<T extends Node>(source: T): T;\n    /** @deprecated */\n    captureEvents(): void;\n    caretPositionFromPoint(x: number, y: number): CaretPosition | null;\n    /** @deprecated */\n    caretRangeFromPoint(x: number, y: number): Range;\n    /** @deprecated */\n    clear(): void;\n    /**\n     * Closes an output stream and forces the sent data to display.\n     */\n    close(): void;\n    /**\n     * Creates an attribute object with a specified name.\n     * @param name String that sets the attribute object\'s name.\n     */\n    createAttribute(localName: string): Attr;\n    createAttributeNS(namespace: string | null, qualifiedName: string): Attr;\n    /**\n     * Returns a CDATASection node whose data is data.\n     */\n    createCDATASection(data: string): CDATASection;\n    /**\n     * Creates a comment object with the specified data.\n     * @param data Sets the comment object\'s data.\n     */\n    createComment(data: string): Comment;\n    /**\n     * Creates a new document.\n     */\n    createDocumentFragment(): DocumentFragment;\n    /**\n     * Creates an instance of the element for the specified tag.\n     * @param tagName The name of an element.\n     */\n    createElement<K extends keyof HTMLElementTagNameMap>(tagName: K, options?: ElementCreationOptions): HTMLElementTagNameMap[K];\n    /** @deprecated */\n    createElement<K extends keyof HTMLElementDeprecatedTagNameMap>(tagName: K, options?: ElementCreationOptions): HTMLElementDeprecatedTagNameMap[K];\n    createElement(tagName: string, options?: ElementCreationOptions): HTMLElement;\n    /**\n     * Returns an element with namespace namespace. Its namespace prefix will be everything before ":" (U+003E) in qualifiedName or null. Its local name will be everything after\n     * ":" (U+003E) in qualifiedName or qualifiedName.\n     * If localName does not match the Name production an\n     * "InvalidCharacterError" DOMException will be thrown.\n     * If one of the following conditions is true a "NamespaceError" DOMException will be thrown:\n     * localName does not match the QName production.\n     * Namespace prefix is not null and namespace is the empty string.\n     * Namespace prefix is "xml" and namespace is not the XML namespace.\n     * qualifiedName or namespace prefix is "xmlns" and namespace is not the XMLNS namespace.\n     * namespace is the XMLNS namespace and\n     * neither qualifiedName nor namespace prefix is "xmlns".\n     * When supplied, options\'s is can be used to create a customized built-in element.\n     */\n    createElementNS(namespaceURI: "http://www.w3.org/1999/xhtml", qualifiedName: string): HTMLElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "a"): SVGAElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "circle"): SVGCircleElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "clipPath"): SVGClipPathElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "componentTransferFunction"): SVGComponentTransferFunctionElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "cursor"): SVGCursorElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "defs"): SVGDefsElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "desc"): SVGDescElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "ellipse"): SVGEllipseElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feBlend"): SVGFEBlendElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feColorMatrix"): SVGFEColorMatrixElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feComponentTransfer"): SVGFEComponentTransferElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feComposite"): SVGFECompositeElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feConvolveMatrix"): SVGFEConvolveMatrixElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feDiffuseLighting"): SVGFEDiffuseLightingElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feDisplacementMap"): SVGFEDisplacementMapElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feDistantLight"): SVGFEDistantLightElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feFlood"): SVGFEFloodElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feFuncA"): SVGFEFuncAElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feFuncB"): SVGFEFuncBElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feFuncG"): SVGFEFuncGElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feFuncR"): SVGFEFuncRElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feGaussianBlur"): SVGFEGaussianBlurElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feImage"): SVGFEImageElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feMerge"): SVGFEMergeElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feMergeNode"): SVGFEMergeNodeElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feMorphology"): SVGFEMorphologyElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feOffset"): SVGFEOffsetElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "fePointLight"): SVGFEPointLightElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feSpecularLighting"): SVGFESpecularLightingElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feSpotLight"): SVGFESpotLightElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feTile"): SVGFETileElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "feTurbulence"): SVGFETurbulenceElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "filter"): SVGFilterElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "foreignObject"): SVGForeignObjectElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "g"): SVGGElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "image"): SVGImageElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "gradient"): SVGGradientElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "line"): SVGLineElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "linearGradient"): SVGLinearGradientElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "marker"): SVGMarkerElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "mask"): SVGMaskElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "path"): SVGPathElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "metadata"): SVGMetadataElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "pattern"): SVGPatternElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "polygon"): SVGPolygonElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "polyline"): SVGPolylineElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "radialGradient"): SVGRadialGradientElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "rect"): SVGRectElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "svg"): SVGSVGElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "script"): SVGScriptElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "stop"): SVGStopElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "style"): SVGStyleElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "switch"): SVGSwitchElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "symbol"): SVGSymbolElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "tspan"): SVGTSpanElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "textContent"): SVGTextContentElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "text"): SVGTextElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "textPath"): SVGTextPathElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "textPositioning"): SVGTextPositioningElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "title"): SVGTitleElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "use"): SVGUseElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: "view"): SVGViewElement;\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: string): SVGElement;\n    createElementNS(namespaceURI: string | null, qualifiedName: string, options?: ElementCreationOptions): Element;\n    createElementNS(namespace: string | null, qualifiedName: string, options?: string | ElementCreationOptions): Element;\n    createEvent(eventInterface: "AnimationEvent"): AnimationEvent;\n    createEvent(eventInterface: "AnimationPlaybackEvent"): AnimationPlaybackEvent;\n    createEvent(eventInterface: "AudioProcessingEvent"): AudioProcessingEvent;\n    createEvent(eventInterface: "BeforeUnloadEvent"): BeforeUnloadEvent;\n    createEvent(eventInterface: "ClipboardEvent"): ClipboardEvent;\n    createEvent(eventInterface: "CloseEvent"): CloseEvent;\n    createEvent(eventInterface: "CompositionEvent"): CompositionEvent;\n    createEvent(eventInterface: "CustomEvent"): CustomEvent;\n    createEvent(eventInterface: "DeviceLightEvent"): DeviceLightEvent;\n    createEvent(eventInterface: "DeviceMotionEvent"): DeviceMotionEvent;\n    createEvent(eventInterface: "DeviceOrientationEvent"): DeviceOrientationEvent;\n    createEvent(eventInterface: "DragEvent"): DragEvent;\n    createEvent(eventInterface: "ErrorEvent"): ErrorEvent;\n    createEvent(eventInterface: "Event"): Event;\n    createEvent(eventInterface: "Events"): Event;\n    createEvent(eventInterface: "FocusEvent"): FocusEvent;\n    createEvent(eventInterface: "FocusNavigationEvent"): FocusNavigationEvent;\n    createEvent(eventInterface: "GamepadEvent"): GamepadEvent;\n    createEvent(eventInterface: "HashChangeEvent"): HashChangeEvent;\n    createEvent(eventInterface: "IDBVersionChangeEvent"): IDBVersionChangeEvent;\n    createEvent(eventInterface: "KeyboardEvent"): KeyboardEvent;\n    createEvent(eventInterface: "ListeningStateChangedEvent"): ListeningStateChangedEvent;\n    createEvent(eventInterface: "MSGestureEvent"): MSGestureEvent;\n    createEvent(eventInterface: "MSMediaKeyMessageEvent"): MSMediaKeyMessageEvent;\n    createEvent(eventInterface: "MSMediaKeyNeededEvent"): MSMediaKeyNeededEvent;\n    createEvent(eventInterface: "MSPointerEvent"): MSPointerEvent;\n    createEvent(eventInterface: "MediaEncryptedEvent"): MediaEncryptedEvent;\n    createEvent(eventInterface: "MediaKeyMessageEvent"): MediaKeyMessageEvent;\n    createEvent(eventInterface: "MediaQueryListEvent"): MediaQueryListEvent;\n    createEvent(eventInterface: "MediaStreamErrorEvent"): MediaStreamErrorEvent;\n    createEvent(eventInterface: "MediaStreamEvent"): MediaStreamEvent;\n    createEvent(eventInterface: "MediaStreamTrackEvent"): MediaStreamTrackEvent;\n    createEvent(eventInterface: "MessageEvent"): MessageEvent;\n    createEvent(eventInterface: "MouseEvent"): MouseEvent;\n    createEvent(eventInterface: "MouseEvents"): MouseEvent;\n    createEvent(eventInterface: "MutationEvent"): MutationEvent;\n    createEvent(eventInterface: "MutationEvents"): MutationEvent;\n    createEvent(eventInterface: "OfflineAudioCompletionEvent"): OfflineAudioCompletionEvent;\n    createEvent(eventInterface: "OverflowEvent"): OverflowEvent;\n    createEvent(eventInterface: "PageTransitionEvent"): PageTransitionEvent;\n    createEvent(eventInterface: "PaymentRequestUpdateEvent"): PaymentRequestUpdateEvent;\n    createEvent(eventInterface: "PermissionRequestedEvent"): PermissionRequestedEvent;\n    createEvent(eventInterface: "PointerEvent"): PointerEvent;\n    createEvent(eventInterface: "PopStateEvent"): PopStateEvent;\n    createEvent(eventInterface: "ProgressEvent"): ProgressEvent;\n    createEvent(eventInterface: "PromiseRejectionEvent"): PromiseRejectionEvent;\n    createEvent(eventInterface: "RTCDTMFToneChangeEvent"): RTCDTMFToneChangeEvent;\n    createEvent(eventInterface: "RTCDataChannelEvent"): RTCDataChannelEvent;\n    createEvent(eventInterface: "RTCDtlsTransportStateChangedEvent"): RTCDtlsTransportStateChangedEvent;\n    createEvent(eventInterface: "RTCErrorEvent"): RTCErrorEvent;\n    createEvent(eventInterface: "RTCIceCandidatePairChangedEvent"): RTCIceCandidatePairChangedEvent;\n    createEvent(eventInterface: "RTCIceGathererEvent"): RTCIceGathererEvent;\n    createEvent(eventInterface: "RTCIceTransportStateChangedEvent"): RTCIceTransportStateChangedEvent;\n    createEvent(eventInterface: "RTCPeerConnectionIceErrorEvent"): RTCPeerConnectionIceErrorEvent;\n    createEvent(eventInterface: "RTCPeerConnectionIceEvent"): RTCPeerConnectionIceEvent;\n    createEvent(eventInterface: "RTCSsrcConflictEvent"): RTCSsrcConflictEvent;\n    createEvent(eventInterface: "RTCStatsEvent"): RTCStatsEvent;\n    createEvent(eventInterface: "RTCTrackEvent"): RTCTrackEvent;\n    createEvent(eventInterface: "SVGZoomEvent"): SVGZoomEvent;\n    createEvent(eventInterface: "SVGZoomEvents"): SVGZoomEvent;\n    createEvent(eventInterface: "SecurityPolicyViolationEvent"): SecurityPolicyViolationEvent;\n    createEvent(eventInterface: "ServiceWorkerMessageEvent"): ServiceWorkerMessageEvent;\n    createEvent(eventInterface: "SpeechRecognitionError"): SpeechRecognitionError;\n    createEvent(eventInterface: "SpeechRecognitionEvent"): SpeechRecognitionEvent;\n    createEvent(eventInterface: "SpeechSynthesisErrorEvent"): SpeechSynthesisErrorEvent;\n    createEvent(eventInterface: "SpeechSynthesisEvent"): SpeechSynthesisEvent;\n    createEvent(eventInterface: "StorageEvent"): StorageEvent;\n    createEvent(eventInterface: "TextEvent"): TextEvent;\n    createEvent(eventInterface: "TouchEvent"): TouchEvent;\n    createEvent(eventInterface: "TrackEvent"): TrackEvent;\n    createEvent(eventInterface: "TransitionEvent"): TransitionEvent;\n    createEvent(eventInterface: "UIEvent"): UIEvent;\n    createEvent(eventInterface: "UIEvents"): UIEvent;\n    createEvent(eventInterface: "VRDisplayEvent"): VRDisplayEvent;\n    createEvent(eventInterface: "VRDisplayEvent "): VRDisplayEvent ;\n    createEvent(eventInterface: "WebGLContextEvent"): WebGLContextEvent;\n    createEvent(eventInterface: "WheelEvent"): WheelEvent;\n    createEvent(eventInterface: string): Event;\n    /**\n     * Creates a NodeIterator object that you can use to traverse filtered lists of nodes or elements in a document.\n     * @param root The root element or node to start traversing on.\n     * @param whatToShow The type of nodes or elements to appear in the node list\n     * @param filter A custom NodeFilter function to use. For more information, see filter. Use null for no filter.\n     * @param entityReferenceExpansion A flag that specifies whether entity reference nodes are expanded.\n     */\n    createNodeIterator(root: Node, whatToShow?: number, filter?: NodeFilter | null): NodeIterator;\n    /**\n     * Returns a ProcessingInstruction node whose target is target and data is data.\n     * If target does not match the Name production an\n     * "InvalidCharacterError" DOMException will be thrown.\n     * If data contains "?>" an\n     * "InvalidCharacterError" DOMException will be thrown.\n     */\n    createProcessingInstruction(target: string, data: string): ProcessingInstruction;\n    /**\n     *  Returns an empty range object that has both of its boundary points positioned at the beginning of the document.\n     */\n    createRange(): Range;\n    /**\n     * Creates a text string from the specified value.\n     * @param data String that specifies the nodeValue property of the text node.\n     */\n    createTextNode(data: string): Text;\n    createTouch(view: WindowProxy, target: EventTarget, identifier: number, pageX: number, pageY: number, screenX: number, screenY: number): Touch;\n    createTouchList(...touches: Touch[]): TouchList;\n    /**\n     * Creates a TreeWalker object that you can use to traverse filtered lists of nodes or elements in a document.\n     * @param root The root element or node to start traversing on.\n     * @param whatToShow The type of nodes or elements to appear in the node list. For more information, see whatToShow.\n     * @param filter A custom NodeFilter function to use.\n     * @param entityReferenceExpansion A flag that specifies whether entity reference nodes are expanded.\n     */\n    createTreeWalker(root: Node, whatToShow?: number, filter?: NodeFilter | null): TreeWalker;\n    /** @deprecated */\n    createTreeWalker(root: Node, whatToShow: number, filter: NodeFilter | null, entityReferenceExpansion?: boolean): TreeWalker;\n    /**\n     * Returns the element for the specified x coordinate and the specified y coordinate.\n     * @param x The x-offset\n     * @param y The y-offset\n     */\n    elementFromPoint(x: number, y: number): Element | null;\n    elementsFromPoint(x: number, y: number): Element[];\n    evaluate(expression: string, contextNode: Node, resolver: XPathNSResolver | ((prefix: string) => string | null) | null, type: number, result: XPathResult | null): XPathResult;\n    /**\n     * Executes a command on the current document, current selection, or the given range.\n     * @param commandId String that specifies the command to execute. This command can be any of the command identifiers that can be executed in script.\n     * @param showUI Display the user interface, defaults to false.\n     * @param value Value to assign.\n     */\n    execCommand(commandId: string, showUI?: boolean, value?: string): boolean;\n    /**\n     * Stops document\'s fullscreen element from being displayed fullscreen and\n     * resolves promise when done.\n     */\n    exitFullscreen(): Promise<void>;\n    getAnimations(): Animation[];\n    /**\n     * Returns a reference to the first object with the specified value of the ID or NAME attribute.\n     * @param elementId String that specifies the ID value. Case-insensitive.\n     */\n    getElementById(elementId: string): HTMLElement | null;\n    /**\n     * collection = element . getElementsByClassName(classNames)\n     */\n    getElementsByClassName(classNames: string): HTMLCollectionOf<Element>;\n    /**\n     * Gets a collection of objects based on the value of the NAME or ID attribute.\n     * @param elementName Gets a collection of objects based on the value of the NAME or ID attribute.\n     */\n    getElementsByName(elementName: string): NodeListOf<HTMLElement>;\n    /**\n     * Retrieves a collection of objects based on the specified element name.\n     * @param name Specifies the name of an element.\n     */\n    getElementsByTagName<K extends keyof HTMLElementTagNameMap>(qualifiedName: K): HTMLCollectionOf<HTMLElementTagNameMap[K]>;\n    getElementsByTagName<K extends keyof SVGElementTagNameMap>(qualifiedName: K): HTMLCollectionOf<SVGElementTagNameMap[K]>;\n    getElementsByTagName(qualifiedName: string): HTMLCollectionOf<Element>;\n    /**\n     * If namespace and localName are\n     * "*" returns a HTMLCollection of all descendant elements.\n     * If only namespace is "*" returns a HTMLCollection of all descendant elements whose local name is localName.\n     * If only localName is "*" returns a HTMLCollection of all descendant elements whose namespace is namespace.\n     * Otherwise, returns a HTMLCollection of all descendant elements whose namespace is namespace and local name is localName.\n     */\n    getElementsByTagNameNS(namespaceURI: "http://www.w3.org/1999/xhtml", localName: string): HTMLCollectionOf<HTMLElement>;\n    getElementsByTagNameNS(namespaceURI: "http://www.w3.org/2000/svg", localName: string): HTMLCollectionOf<SVGElement>;\n    getElementsByTagNameNS(namespaceURI: string, localName: string): HTMLCollectionOf<Element>;\n    /**\n     * Gets a value indicating whether the object currently has focus.\n     */\n    hasFocus(): boolean;\n    importNode<T extends Node>(importedNode: T, deep: boolean): T;\n    /**\n     * Opens a new window and loads a document specified by a given URL. Also, opens a new window that uses the url parameter and the name parameter to collect the output of the write method and the writeln method.\n     * @param url Specifies a MIME type for the document.\n     * @param name Specifies the name of the window. This name is used as the value for the TARGET attribute on a form or an anchor element.\n     * @param features Contains a list of items separated by commas. Each item consists of an option and a value, separated by an equals sign (for example, "fullscreen=yes, toolbar=yes"). The following values are supported.\n     * @param replace Specifies whether the existing entry for the document is replaced in the history list.\n     */\n    open(url?: string, name?: string, features?: string, replace?: boolean): Document;\n    /**\n     * Returns a Boolean value that indicates whether a specified command can be successfully executed using execCommand, given the current state of the document.\n     * @param commandId Specifies a command identifier.\n     */\n    queryCommandEnabled(commandId: string): boolean;\n    /**\n     * Returns a Boolean value that indicates whether the specified command is in the indeterminate state.\n     * @param commandId String that specifies a command identifier.\n     */\n    queryCommandIndeterm(commandId: string): boolean;\n    /**\n     * Returns a Boolean value that indicates the current state of the command.\n     * @param commandId String that specifies a command identifier.\n     */\n    queryCommandState(commandId: string): boolean;\n    /**\n     * Returns a Boolean value that indicates whether the current command is supported on the current range.\n     * @param commandId Specifies a command identifier.\n     */\n    queryCommandSupported(commandId: string): boolean;\n    /**\n     * Returns the current value of the document, range, or current selection for the given command.\n     * @param commandId String that specifies a command identifier.\n     */\n    queryCommandValue(commandId: string): string;\n    /** @deprecated */\n    releaseEvents(): void;\n    /**\n     * Writes one or more HTML expressions to a document in the specified window.\n     * @param content Specifies the text and HTML tags to write.\n     */\n    write(...text: string[]): void;\n    /**\n     * Writes one or more HTML expressions, followed by a carriage return, to a document in the specified window.\n     * @param content The text and HTML tags to write.\n     */\n    writeln(...text: string[]): void;\n    /**\n     * Returns an object representing the current selection of the document that is loaded into the object displaying a webpage.\n     */\n    addEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: Document, ev: DocumentEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: Document, ev: DocumentEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Document: {\n    prototype: Document;\n    new(): Document;\n};\n\ninterface DocumentAndElementEventHandlersEventMap {\n    "copy": ClipboardEvent;\n    "cut": ClipboardEvent;\n    "paste": ClipboardEvent;\n}\n\ninterface DocumentAndElementEventHandlers {\n    oncopy: ((this: DocumentAndElementEventHandlers, ev: ClipboardEvent) => any) | null;\n    oncut: ((this: DocumentAndElementEventHandlers, ev: ClipboardEvent) => any) | null;\n    onpaste: ((this: DocumentAndElementEventHandlers, ev: ClipboardEvent) => any) | null;\n    addEventListener<K extends keyof DocumentAndElementEventHandlersEventMap>(type: K, listener: (this: DocumentAndElementEventHandlers, ev: DocumentAndElementEventHandlersEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof DocumentAndElementEventHandlersEventMap>(type: K, listener: (this: DocumentAndElementEventHandlers, ev: DocumentAndElementEventHandlersEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ninterface DocumentEvent {\n    createEvent(eventInterface: "AnimationEvent"): AnimationEvent;\n    createEvent(eventInterface: "AnimationPlaybackEvent"): AnimationPlaybackEvent;\n    createEvent(eventInterface: "AudioProcessingEvent"): AudioProcessingEvent;\n    createEvent(eventInterface: "BeforeUnloadEvent"): BeforeUnloadEvent;\n    createEvent(eventInterface: "ClipboardEvent"): ClipboardEvent;\n    createEvent(eventInterface: "CloseEvent"): CloseEvent;\n    createEvent(eventInterface: "CompositionEvent"): CompositionEvent;\n    createEvent(eventInterface: "CustomEvent"): CustomEvent;\n    createEvent(eventInterface: "DeviceLightEvent"): DeviceLightEvent;\n    createEvent(eventInterface: "DeviceMotionEvent"): DeviceMotionEvent;\n    createEvent(eventInterface: "DeviceOrientationEvent"): DeviceOrientationEvent;\n    createEvent(eventInterface: "DragEvent"): DragEvent;\n    createEvent(eventInterface: "ErrorEvent"): ErrorEvent;\n    createEvent(eventInterface: "Event"): Event;\n    createEvent(eventInterface: "Events"): Event;\n    createEvent(eventInterface: "FocusEvent"): FocusEvent;\n    createEvent(eventInterface: "FocusNavigationEvent"): FocusNavigationEvent;\n    createEvent(eventInterface: "GamepadEvent"): GamepadEvent;\n    createEvent(eventInterface: "HashChangeEvent"): HashChangeEvent;\n    createEvent(eventInterface: "IDBVersionChangeEvent"): IDBVersionChangeEvent;\n    createEvent(eventInterface: "KeyboardEvent"): KeyboardEvent;\n    createEvent(eventInterface: "ListeningStateChangedEvent"): ListeningStateChangedEvent;\n    createEvent(eventInterface: "MSGestureEvent"): MSGestureEvent;\n    createEvent(eventInterface: "MSMediaKeyMessageEvent"): MSMediaKeyMessageEvent;\n    createEvent(eventInterface: "MSMediaKeyNeededEvent"): MSMediaKeyNeededEvent;\n    createEvent(eventInterface: "MSPointerEvent"): MSPointerEvent;\n    createEvent(eventInterface: "MediaEncryptedEvent"): MediaEncryptedEvent;\n    createEvent(eventInterface: "MediaKeyMessageEvent"): MediaKeyMessageEvent;\n    createEvent(eventInterface: "MediaQueryListEvent"): MediaQueryListEvent;\n    createEvent(eventInterface: "MediaStreamErrorEvent"): MediaStreamErrorEvent;\n    createEvent(eventInterface: "MediaStreamEvent"): MediaStreamEvent;\n    createEvent(eventInterface: "MediaStreamTrackEvent"): MediaStreamTrackEvent;\n    createEvent(eventInterface: "MessageEvent"): MessageEvent;\n    createEvent(eventInterface: "MouseEvent"): MouseEvent;\n    createEvent(eventInterface: "MouseEvents"): MouseEvent;\n    createEvent(eventInterface: "MutationEvent"): MutationEvent;\n    createEvent(eventInterface: "MutationEvents"): MutationEvent;\n    createEvent(eventInterface: "OfflineAudioCompletionEvent"): OfflineAudioCompletionEvent;\n    createEvent(eventInterface: "OverflowEvent"): OverflowEvent;\n    createEvent(eventInterface: "PageTransitionEvent"): PageTransitionEvent;\n    createEvent(eventInterface: "PaymentRequestUpdateEvent"): PaymentRequestUpdateEvent;\n    createEvent(eventInterface: "PermissionRequestedEvent"): PermissionRequestedEvent;\n    createEvent(eventInterface: "PointerEvent"): PointerEvent;\n    createEvent(eventInterface: "PopStateEvent"): PopStateEvent;\n    createEvent(eventInterface: "ProgressEvent"): ProgressEvent;\n    createEvent(eventInterface: "PromiseRejectionEvent"): PromiseRejectionEvent;\n    createEvent(eventInterface: "RTCDTMFToneChangeEvent"): RTCDTMFToneChangeEvent;\n    createEvent(eventInterface: "RTCDataChannelEvent"): RTCDataChannelEvent;\n    createEvent(eventInterface: "RTCDtlsTransportStateChangedEvent"): RTCDtlsTransportStateChangedEvent;\n    createEvent(eventInterface: "RTCErrorEvent"): RTCErrorEvent;\n    createEvent(eventInterface: "RTCIceCandidatePairChangedEvent"): RTCIceCandidatePairChangedEvent;\n    createEvent(eventInterface: "RTCIceGathererEvent"): RTCIceGathererEvent;\n    createEvent(eventInterface: "RTCIceTransportStateChangedEvent"): RTCIceTransportStateChangedEvent;\n    createEvent(eventInterface: "RTCPeerConnectionIceErrorEvent"): RTCPeerConnectionIceErrorEvent;\n    createEvent(eventInterface: "RTCPeerConnectionIceEvent"): RTCPeerConnectionIceEvent;\n    createEvent(eventInterface: "RTCSsrcConflictEvent"): RTCSsrcConflictEvent;\n    createEvent(eventInterface: "RTCStatsEvent"): RTCStatsEvent;\n    createEvent(eventInterface: "RTCTrackEvent"): RTCTrackEvent;\n    createEvent(eventInterface: "SVGZoomEvent"): SVGZoomEvent;\n    createEvent(eventInterface: "SVGZoomEvents"): SVGZoomEvent;\n    createEvent(eventInterface: "SecurityPolicyViolationEvent"): SecurityPolicyViolationEvent;\n    createEvent(eventInterface: "ServiceWorkerMessageEvent"): ServiceWorkerMessageEvent;\n    createEvent(eventInterface: "SpeechRecognitionError"): SpeechRecognitionError;\n    createEvent(eventInterface: "SpeechRecognitionEvent"): SpeechRecognitionEvent;\n    createEvent(eventInterface: "SpeechSynthesisErrorEvent"): SpeechSynthesisErrorEvent;\n    createEvent(eventInterface: "SpeechSynthesisEvent"): SpeechSynthesisEvent;\n    createEvent(eventInterface: "StorageEvent"): StorageEvent;\n    createEvent(eventInterface: "TextEvent"): TextEvent;\n    createEvent(eventInterface: "TouchEvent"): TouchEvent;\n    createEvent(eventInterface: "TrackEvent"): TrackEvent;\n    createEvent(eventInterface: "TransitionEvent"): TransitionEvent;\n    createEvent(eventInterface: "UIEvent"): UIEvent;\n    createEvent(eventInterface: "UIEvents"): UIEvent;\n    createEvent(eventInterface: "VRDisplayEvent"): VRDisplayEvent;\n    createEvent(eventInterface: "VRDisplayEvent "): VRDisplayEvent ;\n    createEvent(eventInterface: "WebGLContextEvent"): WebGLContextEvent;\n    createEvent(eventInterface: "WheelEvent"): WheelEvent;\n    createEvent(eventInterface: string): Event;\n}\n\ninterface DocumentFragment extends Node, NonElementParentNode, ParentNode {\n    getElementById(elementId: string): HTMLElement | null;\n}\n\ndeclare var DocumentFragment: {\n    prototype: DocumentFragment;\n    new(): DocumentFragment;\n};\n\ninterface DocumentOrShadowRoot {\n    readonly activeElement: Element | null;\n    /**\n     * Retrieves a collection of styleSheet objects representing the style sheets that correspond to each instance of a link or style object in the document.\n     */\n    readonly styleSheets: StyleSheetList;\n    caretPositionFromPoint(x: number, y: number): CaretPosition | null;\n    /** @deprecated */\n    caretRangeFromPoint(x: number, y: number): Range;\n    elementFromPoint(x: number, y: number): Element | null;\n    elementsFromPoint(x: number, y: number): Element[];\n    getSelection(): Selection | null;\n}\n\ninterface DocumentTimeline extends AnimationTimeline {\n}\n\ndeclare var DocumentTimeline: {\n    prototype: DocumentTimeline;\n    new(options?: DocumentTimelineOptions): DocumentTimeline;\n};\n\ninterface DocumentType extends Node, ChildNode {\n    readonly name: string;\n    readonly publicId: string;\n    readonly systemId: string;\n}\n\ndeclare var DocumentType: {\n    prototype: DocumentType;\n    new(): DocumentType;\n};\n\ninterface DragEvent extends MouseEvent {\n    /**\n     * Returns the DataTransfer object for the event.\n     */\n    readonly dataTransfer: DataTransfer | null;\n}\n\ndeclare var DragEvent: {\n    prototype: DragEvent;\n    new(type: string, eventInitDict?: DragEventInit): DragEvent;\n};\n\ninterface DynamicsCompressorNode extends AudioNode {\n    readonly attack: AudioParam;\n    readonly knee: AudioParam;\n    readonly ratio: AudioParam;\n    readonly reduction: number;\n    readonly release: AudioParam;\n    readonly threshold: AudioParam;\n}\n\ndeclare var DynamicsCompressorNode: {\n    prototype: DynamicsCompressorNode;\n    new(context: BaseAudioContext, options?: DynamicsCompressorOptions): DynamicsCompressorNode;\n};\n\ninterface EXT_blend_minmax {\n    readonly MAX_EXT: GLenum;\n    readonly MIN_EXT: GLenum;\n}\n\ninterface EXT_frag_depth {\n}\n\ninterface EXT_sRGB {\n    readonly FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING_EXT: GLenum;\n    readonly SRGB8_ALPHA8_EXT: GLenum;\n    readonly SRGB_ALPHA_EXT: GLenum;\n    readonly SRGB_EXT: GLenum;\n}\n\ninterface EXT_shader_texture_lod {\n}\n\ninterface EXT_texture_filter_anisotropic {\n    readonly MAX_TEXTURE_MAX_ANISOTROPY_EXT: GLenum;\n    readonly TEXTURE_MAX_ANISOTROPY_EXT: GLenum;\n}\n\ninterface ElementEventMap {\n    "fullscreenchange": Event;\n    "fullscreenerror": Event;\n}\n\ninterface Element extends Node, ParentNode, NonDocumentTypeChildNode, ChildNode, Slotable, Animatable {\n    readonly assignedSlot: HTMLSlotElement | null;\n    readonly attributes: NamedNodeMap;\n    /**\n     * Allows for manipulation of element\'s class content attribute as a\n     * set of whitespace-separated tokens through a DOMTokenList object.\n     */\n    readonly classList: DOMTokenList;\n    /**\n     * Returns the value of element\'s class content attribute. Can be set\n     * to change it.\n     */\n    className: string;\n    readonly clientHeight: number;\n    readonly clientLeft: number;\n    readonly clientTop: number;\n    readonly clientWidth: number;\n    /**\n     * Returns the value of element\'s id content attribute. Can be set to\n     * change it.\n     */\n    id: string;\n    innerHTML: string;\n    /**\n     * Returns the local name.\n     */\n    readonly localName: string;\n    /**\n     * Returns the namespace.\n     */\n    readonly namespaceURI: string | null;\n    onfullscreenchange: ((this: Element, ev: Event) => any) | null;\n    onfullscreenerror: ((this: Element, ev: Event) => any) | null;\n    outerHTML: string;\n    /**\n     * Returns the namespace prefix.\n     */\n    readonly prefix: string | null;\n    readonly scrollHeight: number;\n    scrollLeft: number;\n    scrollTop: number;\n    readonly scrollWidth: number;\n    /**\n     * Returns element\'s shadow root, if any, and if shadow root\'s mode is "open", and null otherwise.\n     */\n    readonly shadowRoot: ShadowRoot | null;\n    /**\n     * Returns the value of element\'s slot content attribute. Can be set to\n     * change it.\n     */\n    slot: string;\n    /**\n     * Returns the HTML-uppercased qualified name.\n     */\n    readonly tagName: string;\n    /**\n     * Creates a shadow root for element and returns it.\n     */\n    attachShadow(shadowRootInitDict: ShadowRootInit): ShadowRoot;\n    /**\n     * Returns the first (starting at element) inclusive ancestor that matches selectors, and null otherwise.\n     */\n    closest<K extends keyof HTMLElementTagNameMap>(selector: K): HTMLElementTagNameMap[K] | null;\n    closest<K extends keyof SVGElementTagNameMap>(selector: K): SVGElementTagNameMap[K] | null;\n    closest(selector: string): Element | null;\n    /**\n     * Returns element\'s first attribute whose qualified name is qualifiedName, and null if there is no such attribute otherwise.\n     */\n    getAttribute(qualifiedName: string): string | null;\n    /**\n     * Returns element\'s attribute whose namespace is namespace and local name is localName, and null if there is\n     * no such attribute otherwise.\n     */\n    getAttributeNS(namespace: string | null, localName: string): string | null;\n    /**\n     * Returns the qualified names of all element\'s attributes.\n     * Can contain duplicates.\n     */\n    getAttributeNames(): string[];\n    getAttributeNode(name: string): Attr | null;\n    getAttributeNodeNS(namespaceURI: string, localName: string): Attr | null;\n    getBoundingClientRect(): ClientRect | DOMRect;\n    getClientRects(): ClientRectList | DOMRectList;\n    getElementsByClassName(classNames: string): HTMLCollectionOf<Element>;\n    getElementsByTagName<K extends keyof HTMLElementTagNameMap>(qualifiedName: K): HTMLCollectionOf<HTMLElementTagNameMap[K]>;\n    getElementsByTagName<K extends keyof SVGElementTagNameMap>(qualifiedName: K): HTMLCollectionOf<SVGElementTagNameMap[K]>;\n    getElementsByTagName(qualifiedName: string): HTMLCollectionOf<Element>;\n    getElementsByTagNameNS(namespaceURI: "http://www.w3.org/1999/xhtml", localName: string): HTMLCollectionOf<HTMLElement>;\n    getElementsByTagNameNS(namespaceURI: "http://www.w3.org/2000/svg", localName: string): HTMLCollectionOf<SVGElement>;\n    getElementsByTagNameNS(namespaceURI: string, localName: string): HTMLCollectionOf<Element>;\n    /**\n     * Returns true if element has an attribute whose qualified name is qualifiedName, and false otherwise.\n     */\n    hasAttribute(qualifiedName: string): boolean;\n    /**\n     * Returns true if element has an attribute whose namespace is namespace and local name is localName.\n     */\n    hasAttributeNS(namespace: string | null, localName: string): boolean;\n    /**\n     * Returns true if element has attributes, and false otherwise.\n     */\n    hasAttributes(): boolean;\n    hasPointerCapture(pointerId: number): boolean;\n    insertAdjacentElement(position: InsertPosition, insertedElement: Element): Element | null;\n    insertAdjacentHTML(where: InsertPosition, html: string): void;\n    insertAdjacentText(where: InsertPosition, text: string): void;\n    /**\n     * Returns true if matching selectors against element\'s root yields element, and false otherwise.\n     */\n    matches(selectors: string): boolean;\n    msGetRegionContent(): any;\n    releasePointerCapture(pointerId: number): void;\n    /**\n     * Removes element\'s first attribute whose qualified name is qualifiedName.\n     */\n    removeAttribute(qualifiedName: string): void;\n    /**\n     * Removes element\'s attribute whose namespace is namespace and local name is localName.\n     */\n    removeAttributeNS(namespace: string | null, localName: string): void;\n    removeAttributeNode(attr: Attr): Attr;\n    /**\n     * Displays element fullscreen and resolves promise when done.\n     */\n    requestFullscreen(): Promise<void>;\n    scroll(options?: ScrollToOptions): void;\n    scroll(x: number, y: number): void;\n    scrollBy(options?: ScrollToOptions): void;\n    scrollBy(x: number, y: number): void;\n    scrollIntoView(arg?: boolean | ScrollIntoViewOptions): void;\n    scrollTo(options?: ScrollToOptions): void;\n    scrollTo(x: number, y: number): void;\n    /**\n     * Sets the value of element\'s first attribute whose qualified name is qualifiedName to value.\n     */\n    setAttribute(qualifiedName: string, value: string): void;\n    /**\n     * Sets the value of element\'s attribute whose namespace is namespace and local name is localName to value.\n     */\n    setAttributeNS(namespace: string | null, qualifiedName: string, value: string): void;\n    setAttributeNode(attr: Attr): Attr | null;\n    setAttributeNodeNS(attr: Attr): Attr | null;\n    setPointerCapture(pointerId: number): void;\n    /**\n     * If force is not given, "toggles" qualifiedName, removing it if it is\n     * present and adding it if it is not present. If force is true, adds qualifiedName. If force is false, removes qualifiedName.\n     * Returns true if qualifiedName is now present, and false otherwise.\n     */\n    toggleAttribute(qualifiedName: string, force?: boolean): boolean;\n    webkitMatchesSelector(selectors: string): boolean;\n    addEventListener<K extends keyof ElementEventMap>(type: K, listener: (this: Element, ev: ElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ElementEventMap>(type: K, listener: (this: Element, ev: ElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Element: {\n    prototype: Element;\n    new(): Element;\n};\n\ninterface ElementCSSInlineStyle {\n    readonly style: CSSStyleDeclaration;\n}\n\ninterface ElementContentEditable {\n    contentEditable: string;\n    inputMode: string;\n    readonly isContentEditable: boolean;\n}\n\ninterface ElementCreationOptions {\n    is?: string;\n}\n\ninterface ErrorEvent extends Event {\n    readonly colno: number;\n    readonly error: any;\n    readonly filename: string;\n    readonly lineno: number;\n    readonly message: string;\n}\n\ndeclare var ErrorEvent: {\n    prototype: ErrorEvent;\n    new(type: string, eventInitDict?: ErrorEventInit): ErrorEvent;\n};\n\ninterface Event {\n    /**\n     * Returns true or false depending on how event was initialized. True if event goes through its target\'s ancestors in reverse tree order, and false otherwise.\n     */\n    readonly bubbles: boolean;\n    cancelBubble: boolean;\n    readonly cancelable: boolean;\n    /**\n     * Returns true or false depending on how event was initialized. True if event invokes listeners past a ShadowRoot node that is the root of its target, and false otherwise.\n     */\n    readonly composed: boolean;\n    /**\n     * Returns the object whose event listener\'s callback is currently being\n     * invoked.\n     */\n    readonly currentTarget: EventTarget | null;\n    readonly defaultPrevented: boolean;\n    readonly eventPhase: number;\n    /**\n     * Returns true if event was dispatched by the user agent, and\n     * false otherwise.\n     */\n    readonly isTrusted: boolean;\n    returnValue: boolean;\n    /** @deprecated */\n    readonly srcElement: Element | null;\n    /**\n     * Returns the object to which event is dispatched (its target).\n     */\n    readonly target: EventTarget | null;\n    /**\n     * Returns the event\'s timestamp as the number of milliseconds measured relative to\n     * the time origin.\n     */\n    readonly timeStamp: number;\n    /**\n     * Returns the type of event, e.g.\n     * "click", "hashchange", or\n     * "submit".\n     */\n    readonly type: string;\n    composedPath(): EventTarget[];\n    initEvent(type: string, bubbles?: boolean, cancelable?: boolean): void;\n    preventDefault(): void;\n    /**\n     * Invoking this method prevents event from reaching\n     * any registered event listeners after the current one finishes running and, when dispatched in a tree, also prevents event from reaching any\n     * other objects.\n     */\n    stopImmediatePropagation(): void;\n    /**\n     * When dispatched in a tree, invoking this method prevents event from reaching any objects other than the current object.\n     */\n    stopPropagation(): void;\n    readonly AT_TARGET: number;\n    readonly BUBBLING_PHASE: number;\n    readonly CAPTURING_PHASE: number;\n    readonly NONE: number;\n}\n\ndeclare var Event: {\n    prototype: Event;\n    new(type: string, eventInitDict?: EventInit): Event;\n    readonly AT_TARGET: number;\n    readonly BUBBLING_PHASE: number;\n    readonly CAPTURING_PHASE: number;\n    readonly NONE: number;\n};\n\ninterface EventListenerObject {\n    handleEvent(evt: Event): void;\n}\n\ninterface EventSource extends EventTarget {\n    readonly CLOSED: number;\n    readonly CONNECTING: number;\n    readonly OPEN: number;\n    onerror: (evt: MessageEvent) => any;\n    onmessage: (evt: MessageEvent) => any;\n    onopen: (evt: MessageEvent) => any;\n    readonly readyState: number;\n    readonly url: string;\n    readonly withCredentials: boolean;\n    close(): void;\n}\n\ndeclare var EventSource: {\n    prototype: EventSource;\n    new(url: string, eventSourceInitDict?: EventSourceInit): EventSource;\n};\n\ninterface EventSourceInit {\n    readonly withCredentials: boolean;\n}\n\ninterface EventTarget {\n    /**\n     * Appends an event listener for events whose type attribute value is type. The callback argument sets the callback that will be invoked when the event is dispatched.\n     * The options argument sets listener-specific options. For compatibility this can be a\n     * boolean, in which case the method behaves exactly as if the value was specified as options\'s capture.\n     * When set to true, options\'s capture prevents callback from being invoked when the event\'s eventPhase attribute value is BUBBLING_PHASE. When false (or not present), callback will not be invoked when event\'s eventPhase attribute value is CAPTURING_PHASE. Either way, callback will be invoked if event\'s eventPhase attribute value is AT_TARGET.\n     * When set to true, options\'s passive indicates that the callback will not cancel the event by invoking preventDefault(). This is used to enable performance optimizations described in §2.8 Observing event listeners.\n     * When set to true, options\'s once indicates that the callback will only be invoked once after which the event listener will\n     * be removed.\n     * The event listener is appended to target\'s event listener list and is not appended if it has the same type, callback, and capture.\n     */\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject | null, options?: boolean | AddEventListenerOptions): void;\n    /**\n     * Dispatches a synthetic event event to target and returns true\n     * if either event\'s cancelable attribute value is false or its preventDefault() method was not invoked, and false otherwise.\n     */\n    dispatchEvent(event: Event): boolean;\n    /**\n     * Removes the event listener in target\'s event listener list with the same type, callback, and options.\n     */\n    removeEventListener(type: string, callback: EventListenerOrEventListenerObject | null, options?: EventListenerOptions | boolean): void;\n}\n\ndeclare var EventTarget: {\n    prototype: EventTarget;\n    new(): EventTarget;\n};\n\ninterface ExtensionScriptApis {\n    extensionIdToShortId(extensionId: string): number;\n    fireExtensionApiTelemetry(functionName: string, isSucceeded: boolean, isSupported: boolean, errorString: string): void;\n    genericFunction(routerAddress: any, parameters?: string, callbackId?: number): void;\n    genericSynchronousFunction(functionId: number, parameters?: string): string;\n    genericWebRuntimeCallout(to: any, from: any, payload: string): void;\n    getExtensionId(): string;\n    registerGenericFunctionCallbackHandler(callbackHandler: Function): void;\n    registerGenericPersistentCallbackHandler(callbackHandler: Function): void;\n    registerWebRuntimeCallbackHandler(handler: Function): any;\n}\n\ndeclare var ExtensionScriptApis: {\n    prototype: ExtensionScriptApis;\n    new(): ExtensionScriptApis;\n};\n\ninterface External {\n    /** @deprecated */\n    AddSearchProvider(): void;\n    /** @deprecated */\n    IsSearchProviderInstalled(): void;\n}\n\ninterface File extends Blob {\n    readonly lastModified: number;\n    readonly name: string;\n}\n\ndeclare var File: {\n    prototype: File;\n    new(fileBits: BlobPart[], fileName: string, options?: FilePropertyBag): File;\n};\n\ninterface FileList {\n    readonly length: number;\n    item(index: number): File | null;\n    [index: number]: File;\n}\n\ndeclare var FileList: {\n    prototype: FileList;\n    new(): FileList;\n};\n\ninterface FileReaderEventMap {\n    "abort": ProgressEvent;\n    "error": ProgressEvent;\n    "load": ProgressEvent;\n    "loadend": ProgressEvent;\n    "loadstart": ProgressEvent;\n    "progress": ProgressEvent;\n}\n\ninterface FileReader extends EventTarget {\n    readonly error: DOMException | null;\n    onabort: ((this: FileReader, ev: ProgressEvent) => any) | null;\n    onerror: ((this: FileReader, ev: ProgressEvent) => any) | null;\n    onload: ((this: FileReader, ev: ProgressEvent) => any) | null;\n    onloadend: ((this: FileReader, ev: ProgressEvent) => any) | null;\n    onloadstart: ((this: FileReader, ev: ProgressEvent) => any) | null;\n    onprogress: ((this: FileReader, ev: ProgressEvent) => any) | null;\n    readonly readyState: number;\n    readonly result: string | ArrayBuffer | null;\n    abort(): void;\n    readAsArrayBuffer(blob: Blob): void;\n    readAsBinaryString(blob: Blob): void;\n    readAsDataURL(blob: Blob): void;\n    readAsText(blob: Blob, encoding?: string): void;\n    readonly DONE: number;\n    readonly EMPTY: number;\n    readonly LOADING: number;\n    addEventListener<K extends keyof FileReaderEventMap>(type: K, listener: (this: FileReader, ev: FileReaderEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof FileReaderEventMap>(type: K, listener: (this: FileReader, ev: FileReaderEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var FileReader: {\n    prototype: FileReader;\n    new(): FileReader;\n    readonly DONE: number;\n    readonly EMPTY: number;\n    readonly LOADING: number;\n};\n\ninterface FocusEvent extends UIEvent {\n    readonly relatedTarget: EventTarget;\n    initFocusEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, detailArg: number, relatedTargetArg: EventTarget): void;\n}\n\ndeclare var FocusEvent: {\n    prototype: FocusEvent;\n    new(typeArg: string, eventInitDict?: FocusEventInit): FocusEvent;\n};\n\ninterface FocusNavigationEvent extends Event {\n    readonly navigationReason: NavigationReason;\n    readonly originHeight: number;\n    readonly originLeft: number;\n    readonly originTop: number;\n    readonly originWidth: number;\n    requestFocus(): void;\n}\n\ndeclare var FocusNavigationEvent: {\n    prototype: FocusNavigationEvent;\n    new(type: string, eventInitDict?: FocusNavigationEventInit): FocusNavigationEvent;\n};\n\ninterface FormData {\n    append(name: string, value: string | Blob, fileName?: string): void;\n    delete(name: string): void;\n    get(name: string): FormDataEntryValue | null;\n    getAll(name: string): FormDataEntryValue[];\n    has(name: string): boolean;\n    set(name: string, value: string | Blob, fileName?: string): void;\n    forEach(callbackfn: (value: FormDataEntryValue, key: string, parent: FormData) => void, thisArg?: any): void;\n}\n\ndeclare var FormData: {\n    prototype: FormData;\n    new(form?: HTMLFormElement): FormData;\n};\n\ninterface GainNode extends AudioNode {\n    readonly gain: AudioParam;\n}\n\ndeclare var GainNode: {\n    prototype: GainNode;\n    new(context: BaseAudioContext, options?: GainOptions): GainNode;\n};\n\ninterface Gamepad {\n    readonly axes: number[];\n    readonly buttons: GamepadButton[];\n    readonly connected: boolean;\n    readonly displayId: number;\n    readonly hand: GamepadHand;\n    readonly hapticActuators: GamepadHapticActuator[];\n    readonly id: string;\n    readonly index: number;\n    readonly mapping: GamepadMappingType;\n    readonly pose: GamepadPose | null;\n    readonly timestamp: number;\n}\n\ndeclare var Gamepad: {\n    prototype: Gamepad;\n    new(): Gamepad;\n};\n\ninterface GamepadButton {\n    readonly pressed: boolean;\n    readonly touched: boolean;\n    readonly value: number;\n}\n\ndeclare var GamepadButton: {\n    prototype: GamepadButton;\n    new(): GamepadButton;\n};\n\ninterface GamepadEvent extends Event {\n    readonly gamepad: Gamepad;\n}\n\ndeclare var GamepadEvent: {\n    prototype: GamepadEvent;\n    new(typeArg: string, eventInitDict?: GamepadEventInit): GamepadEvent;\n};\n\ninterface GamepadHapticActuator {\n    readonly type: GamepadHapticActuatorType;\n    pulse(value: number, duration: number): Promise<boolean>;\n}\n\ndeclare var GamepadHapticActuator: {\n    prototype: GamepadHapticActuator;\n    new(): GamepadHapticActuator;\n};\n\ninterface GamepadPose {\n    readonly angularAcceleration: Float32Array | null;\n    readonly angularVelocity: Float32Array | null;\n    readonly hasOrientation: boolean;\n    readonly hasPosition: boolean;\n    readonly linearAcceleration: Float32Array | null;\n    readonly linearVelocity: Float32Array | null;\n    readonly orientation: Float32Array | null;\n    readonly position: Float32Array | null;\n}\n\ndeclare var GamepadPose: {\n    prototype: GamepadPose;\n    new(): GamepadPose;\n};\n\ninterface Geolocation {\n    clearWatch(watchId: number): void;\n    getCurrentPosition(successCallback: PositionCallback, errorCallback?: PositionErrorCallback, options?: PositionOptions): void;\n    watchPosition(successCallback: PositionCallback, errorCallback?: PositionErrorCallback, options?: PositionOptions): number;\n}\n\ninterface GetSVGDocument {\n    getSVGDocument(): Document;\n}\n\ninterface GlobalEventHandlersEventMap {\n    "abort": UIEvent;\n    "animationcancel": AnimationEvent;\n    "animationend": AnimationEvent;\n    "animationiteration": AnimationEvent;\n    "animationstart": AnimationEvent;\n    "auxclick": Event;\n    "blur": FocusEvent;\n    "cancel": Event;\n    "canplay": Event;\n    "canplaythrough": Event;\n    "change": Event;\n    "click": MouseEvent;\n    "close": Event;\n    "contextmenu": MouseEvent;\n    "cuechange": Event;\n    "dblclick": MouseEvent;\n    "drag": DragEvent;\n    "dragend": DragEvent;\n    "dragenter": DragEvent;\n    "dragexit": Event;\n    "dragleave": DragEvent;\n    "dragover": DragEvent;\n    "dragstart": DragEvent;\n    "drop": DragEvent;\n    "durationchange": Event;\n    "emptied": Event;\n    "ended": Event;\n    "error": ErrorEvent;\n    "focus": FocusEvent;\n    "gotpointercapture": PointerEvent;\n    "input": Event;\n    "invalid": Event;\n    "keydown": KeyboardEvent;\n    "keypress": KeyboardEvent;\n    "keyup": KeyboardEvent;\n    "load": Event;\n    "loadeddata": Event;\n    "loadedmetadata": Event;\n    "loadend": ProgressEvent;\n    "loadstart": Event;\n    "lostpointercapture": PointerEvent;\n    "mousedown": MouseEvent;\n    "mouseenter": MouseEvent;\n    "mouseleave": MouseEvent;\n    "mousemove": MouseEvent;\n    "mouseout": MouseEvent;\n    "mouseover": MouseEvent;\n    "mouseup": MouseEvent;\n    "pause": Event;\n    "play": Event;\n    "playing": Event;\n    "pointercancel": PointerEvent;\n    "pointerdown": PointerEvent;\n    "pointerenter": PointerEvent;\n    "pointerleave": PointerEvent;\n    "pointermove": PointerEvent;\n    "pointerout": PointerEvent;\n    "pointerover": PointerEvent;\n    "pointerup": PointerEvent;\n    "progress": ProgressEvent;\n    "ratechange": Event;\n    "reset": Event;\n    "resize": UIEvent;\n    "scroll": UIEvent;\n    "securitypolicyviolation": SecurityPolicyViolationEvent;\n    "seeked": Event;\n    "seeking": Event;\n    "select": UIEvent;\n    "stalled": Event;\n    "submit": Event;\n    "suspend": Event;\n    "timeupdate": Event;\n    "toggle": Event;\n    "touchcancel": TouchEvent;\n    "touchend": TouchEvent;\n    "touchmove": TouchEvent;\n    "touchstart": TouchEvent;\n    "transitioncancel": TransitionEvent;\n    "transitionend": TransitionEvent;\n    "transitionrun": TransitionEvent;\n    "transitionstart": TransitionEvent;\n    "volumechange": Event;\n    "waiting": Event;\n    "wheel": WheelEvent;\n}\n\ninterface GlobalEventHandlers {\n    /**\n     * Fires when the user aborts the download.\n     * @param ev The event.\n     */\n    onabort: ((this: GlobalEventHandlers, ev: UIEvent) => any) | null;\n    onanimationcancel: ((this: GlobalEventHandlers, ev: AnimationEvent) => any) | null;\n    onanimationend: ((this: GlobalEventHandlers, ev: AnimationEvent) => any) | null;\n    onanimationiteration: ((this: GlobalEventHandlers, ev: AnimationEvent) => any) | null;\n    onanimationstart: ((this: GlobalEventHandlers, ev: AnimationEvent) => any) | null;\n    onauxclick: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the object loses the input focus.\n     * @param ev The focus event.\n     */\n    onblur: ((this: GlobalEventHandlers, ev: FocusEvent) => any) | null;\n    oncancel: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when playback is possible, but would require further buffering.\n     * @param ev The event.\n     */\n    oncanplay: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    oncanplaythrough: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the contents of the object or selection have changed.\n     * @param ev The event.\n     */\n    onchange: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the user clicks the left mouse button on the object\n     * @param ev The mouse event.\n     */\n    onclick: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    onclose: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the user clicks the right mouse button in the client area, opening the context menu.\n     * @param ev The mouse event.\n     */\n    oncontextmenu: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    oncuechange: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the user double-clicks the object.\n     * @param ev The mouse event.\n     */\n    ondblclick: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Fires on the source object continuously during a drag operation.\n     * @param ev The event.\n     */\n    ondrag: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    /**\n     * Fires on the source object when the user releases the mouse at the close of a drag operation.\n     * @param ev The event.\n     */\n    ondragend: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    /**\n     * Fires on the target element when the user drags the object to a valid drop target.\n     * @param ev The drag event.\n     */\n    ondragenter: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    ondragexit: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires on the target object when the user moves the mouse out of a valid drop target during a drag operation.\n     * @param ev The drag event.\n     */\n    ondragleave: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    /**\n     * Fires on the target element continuously while the user drags the object over a valid drop target.\n     * @param ev The event.\n     */\n    ondragover: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    /**\n     * Fires on the source object when the user starts to drag a text selection or selected object.\n     * @param ev The event.\n     */\n    ondragstart: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    ondrop: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    /**\n     * Occurs when the duration attribute is updated.\n     * @param ev The event.\n     */\n    ondurationchange: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the media element is reset to its initial state.\n     * @param ev The event.\n     */\n    onemptied: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the end of playback is reached.\n     * @param ev The event\n     */\n    onended: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when an error occurs during object loading.\n     * @param ev The event.\n     */\n    onerror: ErrorEventHandler;\n    /**\n     * Fires when the object receives focus.\n     * @param ev The event.\n     */\n    onfocus: ((this: GlobalEventHandlers, ev: FocusEvent) => any) | null;\n    ongotpointercapture: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    oninput: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    oninvalid: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the user presses a key.\n     * @param ev The keyboard event\n     */\n    onkeydown: ((this: GlobalEventHandlers, ev: KeyboardEvent) => any) | null;\n    /**\n     * Fires when the user presses an alphanumeric key.\n     * @param ev The event.\n     */\n    onkeypress: ((this: GlobalEventHandlers, ev: KeyboardEvent) => any) | null;\n    /**\n     * Fires when the user releases a key.\n     * @param ev The keyboard event\n     */\n    onkeyup: ((this: GlobalEventHandlers, ev: KeyboardEvent) => any) | null;\n    /**\n     * Fires immediately after the browser loads the object.\n     * @param ev The event.\n     */\n    onload: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when media data is loaded at the current playback position.\n     * @param ev The event.\n     */\n    onloadeddata: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the duration and dimensions of the media have been determined.\n     * @param ev The event.\n     */\n    onloadedmetadata: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onloadend: ((this: GlobalEventHandlers, ev: ProgressEvent) => any) | null;\n    /**\n     * Occurs when Internet Explorer begins looking for media data.\n     * @param ev The event.\n     */\n    onloadstart: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onlostpointercapture: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    /**\n     * Fires when the user clicks the object with either mouse button.\n     * @param ev The mouse event.\n     */\n    onmousedown: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    onmouseenter: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    onmouseleave: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Fires when the user moves the mouse over the object.\n     * @param ev The mouse event.\n     */\n    onmousemove: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Fires when the user moves the mouse pointer outside the boundaries of the object.\n     * @param ev The mouse event.\n     */\n    onmouseout: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Fires when the user moves the mouse pointer into the object.\n     * @param ev The mouse event.\n     */\n    onmouseover: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Fires when the user releases a mouse button while the mouse is over the object.\n     * @param ev The mouse event.\n     */\n    onmouseup: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Occurs when playback is paused.\n     * @param ev The event.\n     */\n    onpause: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the play method is requested.\n     * @param ev The event.\n     */\n    onplay: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the audio or video has started playing.\n     * @param ev The event.\n     */\n    onplaying: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onpointercancel: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerdown: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerenter: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerleave: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointermove: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerout: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerover: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerup: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    /**\n     * Occurs to indicate progress while downloading media data.\n     * @param ev The event.\n     */\n    onprogress: ((this: GlobalEventHandlers, ev: ProgressEvent) => any) | null;\n    /**\n     * Occurs when the playback rate is increased or decreased.\n     * @param ev The event.\n     */\n    onratechange: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the user resets a form.\n     * @param ev The event.\n     */\n    onreset: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onresize: ((this: GlobalEventHandlers, ev: UIEvent) => any) | null;\n    /**\n     * Fires when the user repositions the scroll box in the scroll bar on the object.\n     * @param ev The event.\n     */\n    onscroll: ((this: GlobalEventHandlers, ev: UIEvent) => any) | null;\n    onsecuritypolicyviolation: ((this: GlobalEventHandlers, ev: SecurityPolicyViolationEvent) => any) | null;\n    /**\n     * Occurs when the seek operation ends.\n     * @param ev The event.\n     */\n    onseeked: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the current playback position is moved.\n     * @param ev The event.\n     */\n    onseeking: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the current selection changes.\n     * @param ev The event.\n     */\n    onselect: ((this: GlobalEventHandlers, ev: UIEvent) => any) | null;\n    /**\n     * Occurs when the download has stopped.\n     * @param ev The event.\n     */\n    onstalled: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onsubmit: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs if the load operation has been intentionally halted.\n     * @param ev The event.\n     */\n    onsuspend: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs to indicate the current playback position.\n     * @param ev The event.\n     */\n    ontimeupdate: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    ontoggle: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    ontouchcancel: ((this: GlobalEventHandlers, ev: TouchEvent) => any) | null;\n    ontouchend: ((this: GlobalEventHandlers, ev: TouchEvent) => any) | null;\n    ontouchmove: ((this: GlobalEventHandlers, ev: TouchEvent) => any) | null;\n    ontouchstart: ((this: GlobalEventHandlers, ev: TouchEvent) => any) | null;\n    ontransitioncancel: ((this: GlobalEventHandlers, ev: TransitionEvent) => any) | null;\n    ontransitionend: ((this: GlobalEventHandlers, ev: TransitionEvent) => any) | null;\n    ontransitionrun: ((this: GlobalEventHandlers, ev: TransitionEvent) => any) | null;\n    ontransitionstart: ((this: GlobalEventHandlers, ev: TransitionEvent) => any) | null;\n    /**\n     * Occurs when the volume is changed, or playback is muted or unmuted.\n     * @param ev The event.\n     */\n    onvolumechange: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when playback stops because the next frame of a video resource is not available.\n     * @param ev The event.\n     */\n    onwaiting: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onwheel: ((this: GlobalEventHandlers, ev: WheelEvent) => any) | null;\n    addEventListener<K extends keyof GlobalEventHandlersEventMap>(type: K, listener: (this: GlobalEventHandlers, ev: GlobalEventHandlersEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof GlobalEventHandlersEventMap>(type: K, listener: (this: GlobalEventHandlers, ev: GlobalEventHandlersEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ninterface GlobalFetch {\n    fetch(input: RequestInfo, init?: RequestInit): Promise<Response>;\n}\n\ninterface HTMLAllCollection {\n    /**\n     * Returns the number of elements in the collection.\n     */\n    readonly length: number;\n    /**\n     * element = collection(index)\n     */\n    item(nameOrIndex?: string): HTMLCollection | Element | null;\n    /**\n     * element = collection(name)\n     */\n    namedItem(name: string): HTMLCollection | Element | null;\n    [index: number]: Element;\n}\n\ndeclare var HTMLAllCollection: {\n    prototype: HTMLAllCollection;\n    new(): HTMLAllCollection;\n};\n\ninterface HTMLAnchorElement extends HTMLElement, HTMLHyperlinkElementUtils {\n    /**\n     * Sets or retrieves the character set used to encode the object.\n     */\n    /** @deprecated */\n    charset: string;\n    /**\n     * Sets or retrieves the coordinates of the object.\n     */\n    /** @deprecated */\n    coords: string;\n    download: string;\n    /**\n     * Sets or retrieves the language code of the object.\n     */\n    hreflang: string;\n    /**\n     * Sets or retrieves the shape of the object.\n     */\n    /** @deprecated */\n    name: string;\n    ping: string;\n    referrerPolicy: string;\n    /**\n     * Sets or retrieves the relationship between the object and the destination of the link.\n     */\n    rel: string;\n    readonly relList: DOMTokenList;\n    /**\n     * Sets or retrieves the relationship between the object and the destination of the link.\n     */\n    /** @deprecated */\n    rev: string;\n    /**\n     * Sets or retrieves the shape of the object.\n     */\n    /** @deprecated */\n    shape: string;\n    /**\n     * Sets or retrieves the window or frame at which to target content.\n     */\n    target: string;\n    /**\n     * Retrieves or sets the text of the object as a string.\n     */\n    text: string;\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAnchorElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAnchorElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLAnchorElement: {\n    prototype: HTMLAnchorElement;\n    new(): HTMLAnchorElement;\n};\n\ninterface HTMLAppletElement extends HTMLElement {\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves a text alternative to the graphic.\n     */\n    /** @deprecated */\n    alt: string;\n    /**\n     * Sets or retrieves a character string that can be used to implement your own archive functionality for the object.\n     */\n    /** @deprecated */\n    archive: string;\n    /** @deprecated */\n    code: string;\n    /**\n     * Sets or retrieves the URL of the component.\n     */\n    /** @deprecated */\n    codeBase: string;\n    readonly form: HTMLFormElement | null;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    /** @deprecated */\n    height: string;\n    /** @deprecated */\n    hspace: number;\n    /**\n     * Sets or retrieves the shape of the object.\n     */\n    /** @deprecated */\n    name: string;\n    /** @deprecated */\n    object: string;\n    /** @deprecated */\n    vspace: number;\n    /** @deprecated */\n    width: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAppletElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAppletElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLAppletElement: {\n    prototype: HTMLAppletElement;\n    new(): HTMLAppletElement;\n};\n\ninterface HTMLAreaElement extends HTMLElement, HTMLHyperlinkElementUtils {\n    /**\n     * Sets or retrieves a text alternative to the graphic.\n     */\n    alt: string;\n    /**\n     * Sets or retrieves the coordinates of the object.\n     */\n    coords: string;\n    download: string;\n    /**\n     * Sets or gets whether clicks in this region cause action.\n     */\n    /** @deprecated */\n    noHref: boolean;\n    ping: string;\n    referrerPolicy: string;\n    rel: string;\n    readonly relList: DOMTokenList;\n    /**\n     * Sets or retrieves the shape of the object.\n     */\n    shape: string;\n    /**\n     * Sets or retrieves the window or frame at which to target content.\n     */\n    target: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAreaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAreaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLAreaElement: {\n    prototype: HTMLAreaElement;\n    new(): HTMLAreaElement;\n};\n\ninterface HTMLAudioElement extends HTMLMediaElement {\n    addEventListener<K extends keyof HTMLMediaElementEventMap>(type: K, listener: (this: HTMLAudioElement, ev: HTMLMediaElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLMediaElementEventMap>(type: K, listener: (this: HTMLAudioElement, ev: HTMLMediaElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLAudioElement: {\n    prototype: HTMLAudioElement;\n    new(): HTMLAudioElement;\n};\n\ninterface HTMLBRElement extends HTMLElement {\n    /**\n     * Sets or retrieves the side on which floating objects are not to be positioned when any IHTMLBlockElement is inserted into the document.\n     */\n    /** @deprecated */\n    clear: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBRElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBRElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLBRElement: {\n    prototype: HTMLBRElement;\n    new(): HTMLBRElement;\n};\n\ninterface HTMLBaseElement extends HTMLElement {\n    /**\n     * Gets or sets the baseline URL on which relative links are based.\n     */\n    href: string;\n    /**\n     * Sets or retrieves the window or frame at which to target content.\n     */\n    target: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBaseElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBaseElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLBaseElement: {\n    prototype: HTMLBaseElement;\n    new(): HTMLBaseElement;\n};\n\ninterface HTMLBaseFontElement extends HTMLElement, DOML2DeprecatedColorProperty {\n    /**\n     * Sets or retrieves the current typeface family.\n     */\n    /** @deprecated */\n    face: string;\n    /**\n     * Sets or retrieves the font size of the object.\n     */\n    /** @deprecated */\n    size: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBaseFontElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBaseFontElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLBaseFontElement: {\n    prototype: HTMLBaseFontElement;\n    new(): HTMLBaseFontElement;\n};\n\ninterface HTMLBodyElementEventMap extends HTMLElementEventMap, WindowEventHandlersEventMap {\n    "orientationchange": Event;\n}\n\ninterface HTMLBodyElement extends HTMLElement, WindowEventHandlers {\n    /** @deprecated */\n    aLink: string;\n    /** @deprecated */\n    background: string;\n    /** @deprecated */\n    bgColor: string;\n    bgProperties: string;\n    /** @deprecated */\n    link: string;\n    /** @deprecated */\n    noWrap: boolean;\n    /** @deprecated */\n    onorientationchange: ((this: HTMLBodyElement, ev: Event) => any) | null;\n    /** @deprecated */\n    text: string;\n    /** @deprecated */\n    vLink: string;\n    addEventListener<K extends keyof HTMLBodyElementEventMap>(type: K, listener: (this: HTMLBodyElement, ev: HTMLBodyElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLBodyElementEventMap>(type: K, listener: (this: HTMLBodyElement, ev: HTMLBodyElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLBodyElement: {\n    prototype: HTMLBodyElement;\n    new(): HTMLBodyElement;\n};\n\ninterface HTMLButtonElement extends HTMLElement {\n    /**\n     * Provides a way to direct a user to a specific field when a document loads. This can provide both direction and convenience for a user, reducing the need to click or tab to a field when a page opens. This attribute is true when present on an element, and false when missing.\n     */\n    autofocus: boolean;\n    disabled: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Overrides the action attribute (where the data on a form is sent) on the parent form element.\n     */\n    formAction: string;\n    /**\n     * Used to override the encoding (formEnctype attribute) specified on the form element.\n     */\n    formEnctype: string;\n    /**\n     * Overrides the submit method attribute previously specified on a form element.\n     */\n    formMethod: string;\n    /**\n     * Overrides any validation or required attributes on a form or form elements to allow it to be submitted without validation. This can be used to create a "save draft"-type submit option.\n     */\n    formNoValidate: boolean;\n    /**\n     * Overrides the target attribute on a form element.\n     */\n    formTarget: string;\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    /**\n     * Gets the classification and default behavior of the button.\n     */\n    type: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /**\n     * Sets or retrieves the default or selected value of the control.\n     */\n    value: string;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLButtonElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLButtonElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLButtonElement: {\n    prototype: HTMLButtonElement;\n    new(): HTMLButtonElement;\n};\n\ninterface HTMLCanvasElement extends HTMLElement {\n    /**\n     * Gets or sets the height of a canvas element on a document.\n     */\n    height: number;\n    /**\n     * Gets or sets the width of a canvas element on a document.\n     */\n    width: number;\n    /**\n     * Returns an object that provides methods and properties for drawing and manipulating images and graphics on a canvas element in a document. A context object includes information about colors, line widths, fonts, and other graphic parameters that can be drawn on a canvas.\n     * @param contextId The identifier (ID) of the type of canvas to create. Internet Explorer 9 and Internet Explorer 10 support only a 2-D context using canvas.getContext("2d"); IE11 Preview also supports 3-D or WebGL context using canvas.getContext("experimental-webgl");\n     */\n    getContext(contextId: "2d", contextAttributes?: CanvasRenderingContext2DSettings): CanvasRenderingContext2D | null;\n    getContext(contextId: "webgl" | "experimental-webgl", contextAttributes?: WebGLContextAttributes): WebGLRenderingContext | null;\n    getContext(contextId: string, contextAttributes?: {}): CanvasRenderingContext2D | WebGLRenderingContext | null;\n    toBlob(callback: BlobCallback, type?: string, quality?: any): void;\n    /**\n     * Returns the content of the current canvas as an image that you can use as a source for another canvas or an HTML element.\n     * @param type The standard MIME type for the image format to return. If you do not specify this parameter, the default value is a PNG format image.\n     */\n    toDataURL(type?: string, quality?: any): string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLCanvasElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLCanvasElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLCanvasElement: {\n    prototype: HTMLCanvasElement;\n    new(): HTMLCanvasElement;\n};\n\ninterface HTMLCollectionBase {\n    /**\n     * Sets or retrieves the number of objects in a collection.\n     */\n    readonly length: number;\n    /**\n     * Retrieves an object from various collections.\n     */\n    item(index: number): Element | null;\n    [index: number]: Element;\n}\n\ninterface HTMLCollection extends HTMLCollectionBase {\n    /**\n     * Retrieves a select object or an object from an options collection.\n     */\n    namedItem(name: string): Element | null;\n}\n\ndeclare var HTMLCollection: {\n    prototype: HTMLCollection;\n    new(): HTMLCollection;\n};\n\ninterface HTMLCollectionOf<T extends Element> extends HTMLCollectionBase {\n    item(index: number): T | null;\n    namedItem(name: string): T | null;\n    [index: number]: T;\n}\n\ninterface HTMLDListElement extends HTMLElement {\n    /** @deprecated */\n    compact: boolean;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDListElement: {\n    prototype: HTMLDListElement;\n    new(): HTMLDListElement;\n};\n\ninterface HTMLDataElement extends HTMLElement {\n    value: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDataElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDataElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDataElement: {\n    prototype: HTMLDataElement;\n    new(): HTMLDataElement;\n};\n\ninterface HTMLDataListElement extends HTMLElement {\n    readonly options: HTMLCollectionOf<HTMLOptionElement>;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDataListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDataListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDataListElement: {\n    prototype: HTMLDataListElement;\n    new(): HTMLDataListElement;\n};\n\ninterface HTMLDetailsElement extends HTMLElement {\n    open: boolean;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDetailsElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDetailsElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDetailsElement: {\n    prototype: HTMLDetailsElement;\n    new(): HTMLDetailsElement;\n};\n\ninterface HTMLDialogElement extends HTMLElement {\n    open: boolean;\n    returnValue: string;\n    close(returnValue?: string): void;\n    show(): void;\n    showModal(): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDialogElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDialogElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDialogElement: {\n    prototype: HTMLDialogElement;\n    new(): HTMLDialogElement;\n};\n\ninterface HTMLDirectoryElement extends HTMLElement {\n    /** @deprecated */\n    compact: boolean;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDirectoryElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDirectoryElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDirectoryElement: {\n    prototype: HTMLDirectoryElement;\n    new(): HTMLDirectoryElement;\n};\n\ninterface HTMLDivElement extends HTMLElement {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDivElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDivElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDivElement: {\n    prototype: HTMLDivElement;\n    new(): HTMLDivElement;\n};\n\ninterface HTMLDocument extends Document {\n    addEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: HTMLDocument, ev: DocumentEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: HTMLDocument, ev: DocumentEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDocument: {\n    prototype: HTMLDocument;\n    new(): HTMLDocument;\n};\n\ninterface HTMLElementEventMap extends ElementEventMap, GlobalEventHandlersEventMap, DocumentAndElementEventHandlersEventMap {\n}\n\ninterface HTMLElement extends Element, GlobalEventHandlers, DocumentAndElementEventHandlers, ElementContentEditable, HTMLOrSVGElement, ElementCSSInlineStyle {\n    accessKey: string;\n    readonly accessKeyLabel: string;\n    autocapitalize: string;\n    dir: string;\n    draggable: boolean;\n    hidden: boolean;\n    innerText: string;\n    lang: string;\n    readonly offsetHeight: number;\n    readonly offsetLeft: number;\n    readonly offsetParent: Element | null;\n    readonly offsetTop: number;\n    readonly offsetWidth: number;\n    spellcheck: boolean;\n    title: string;\n    translate: boolean;\n    click(): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLElement: {\n    prototype: HTMLElement;\n    new(): HTMLElement;\n};\n\ninterface HTMLEmbedElement extends HTMLElement, GetSVGDocument {\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    height: string;\n    hidden: any;\n    /**\n     * Gets or sets whether the DLNA PlayTo device is available.\n     */\n    msPlayToDisabled: boolean;\n    /**\n     * Gets or sets the path to the preferred media source. This enables the Play To target device to stream the media content, which can be DRM protected, from a different location, such as a cloud media server.\n     */\n    msPlayToPreferredSourceUri: string;\n    /**\n     * Gets or sets the primary DLNA PlayTo device.\n     */\n    msPlayToPrimary: boolean;\n    /**\n     * Gets the source associated with the media element for use by the PlayToManager.\n     */\n    readonly msPlayToSource: any;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    /** @deprecated */\n    name: string;\n    /**\n     * Retrieves the palette used for the embedded document.\n     */\n    readonly palette: string;\n    /**\n     * Retrieves the URL of the plug-in used to view an embedded document.\n     */\n    readonly pluginspage: string;\n    readonly readyState: string;\n    /**\n     * Sets or retrieves a URL to be loaded by the object.\n     */\n    src: string;\n    /**\n     * Sets or retrieves the height and width units of the embed object.\n     */\n    units: string;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    width: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLEmbedElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLEmbedElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLEmbedElement: {\n    prototype: HTMLEmbedElement;\n    new(): HTMLEmbedElement;\n};\n\ninterface HTMLFieldSetElement extends HTMLElement {\n    disabled: boolean;\n    readonly elements: HTMLCollection;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    name: string;\n    readonly type: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFieldSetElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFieldSetElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLFieldSetElement: {\n    prototype: HTMLFieldSetElement;\n    new(): HTMLFieldSetElement;\n};\n\ninterface HTMLFontElement extends HTMLElement {\n    /** @deprecated */\n    color: string;\n    /**\n     * Sets or retrieves the current typeface family.\n     */\n    /** @deprecated */\n    face: string;\n    /** @deprecated */\n    size: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFontElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFontElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLFontElement: {\n    prototype: HTMLFontElement;\n    new(): HTMLFontElement;\n};\n\ninterface HTMLFormControlsCollection extends HTMLCollectionBase {\n    /**\n     * element = collection[name]\n     */\n    namedItem(name: string): RadioNodeList | Element | null;\n}\n\ndeclare var HTMLFormControlsCollection: {\n    prototype: HTMLFormControlsCollection;\n    new(): HTMLFormControlsCollection;\n};\n\ninterface HTMLFormElement extends HTMLElement {\n    /**\n     * Sets or retrieves a list of character encodings for input data that must be accepted by the server processing the form.\n     */\n    acceptCharset: string;\n    /**\n     * Sets or retrieves the URL to which the form content is sent for processing.\n     */\n    action: string;\n    /**\n     * Specifies whether autocomplete is applied to an editable text field.\n     */\n    autocomplete: string;\n    /**\n     * Retrieves a collection, in source order, of all controls in a given form.\n     */\n    readonly elements: HTMLFormControlsCollection;\n    /**\n     * Sets or retrieves the MIME encoding for the form.\n     */\n    encoding: string;\n    /**\n     * Sets or retrieves the encoding type for the form.\n     */\n    enctype: string;\n    /**\n     * Sets or retrieves the number of objects in a collection.\n     */\n    readonly length: number;\n    /**\n     * Sets or retrieves how to send the form data to the server.\n     */\n    method: string;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    /**\n     * Designates a form that is not validated when submitted.\n     */\n    noValidate: boolean;\n    /**\n     * Sets or retrieves the window or frame at which to target content.\n     */\n    target: string;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    /**\n     * Fires when the user resets a form.\n     */\n    reset(): void;\n    /**\n     * Fires when a FORM is about to be submitted.\n     */\n    submit(): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFormElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFormElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n    [index: number]: Element;\n    [name: string]: any;\n}\n\ndeclare var HTMLFormElement: {\n    prototype: HTMLFormElement;\n    new(): HTMLFormElement;\n};\n\ninterface HTMLFrameElement extends HTMLElement {\n    /**\n     * Retrieves the document object of the page or frame.\n     */\n    /** @deprecated */\n    readonly contentDocument: Document | null;\n    /**\n     * Retrieves the object of the specified.\n     */\n    /** @deprecated */\n    readonly contentWindow: WindowProxy | null;\n    /**\n     * Sets or retrieves whether to display a border for the frame.\n     */\n    /** @deprecated */\n    frameBorder: string;\n    /**\n     * Sets or retrieves a URI to a long description of the object.\n     */\n    /** @deprecated */\n    longDesc: string;\n    /**\n     * Sets or retrieves the top and bottom margin heights before displaying the text in a frame.\n     */\n    /** @deprecated */\n    marginHeight: string;\n    /**\n     * Sets or retrieves the left and right margin widths before displaying the text in a frame.\n     */\n    /** @deprecated */\n    marginWidth: string;\n    /**\n     * Sets or retrieves the frame name.\n     */\n    /** @deprecated */\n    name: string;\n    /**\n     * Sets or retrieves whether the user can resize the frame.\n     */\n    /** @deprecated */\n    noResize: boolean;\n    /**\n     * Sets or retrieves whether the frame can be scrolled.\n     */\n    /** @deprecated */\n    scrolling: string;\n    /**\n     * Sets or retrieves a URL to be loaded by the object.\n     */\n    /** @deprecated */\n    src: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFrameElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFrameElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLFrameElement: {\n    prototype: HTMLFrameElement;\n    new(): HTMLFrameElement;\n};\n\ninterface HTMLFrameSetElementEventMap extends HTMLElementEventMap, WindowEventHandlersEventMap {\n}\n\ninterface HTMLFrameSetElement extends HTMLElement, WindowEventHandlers {\n    /**\n     * Sets or retrieves the frame widths of the object.\n     */\n    /** @deprecated */\n    cols: string;\n    /**\n     * Sets or retrieves the frame heights of the object.\n     */\n    /** @deprecated */\n    rows: string;\n    addEventListener<K extends keyof HTMLFrameSetElementEventMap>(type: K, listener: (this: HTMLFrameSetElement, ev: HTMLFrameSetElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLFrameSetElementEventMap>(type: K, listener: (this: HTMLFrameSetElement, ev: HTMLFrameSetElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLFrameSetElement: {\n    prototype: HTMLFrameSetElement;\n    new(): HTMLFrameSetElement;\n};\n\ninterface HTMLHRElement extends HTMLElement {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    /** @deprecated */\n    color: string;\n    /**\n     * Sets or retrieves whether the horizontal rule is drawn with 3-D shading.\n     */\n    /** @deprecated */\n    noShade: boolean;\n    /** @deprecated */\n    size: string;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    /** @deprecated */\n    width: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHRElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHRElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLHRElement: {\n    prototype: HTMLHRElement;\n    new(): HTMLHRElement;\n};\n\ninterface HTMLHeadElement extends HTMLElement {\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHeadElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHeadElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLHeadElement: {\n    prototype: HTMLHeadElement;\n    new(): HTMLHeadElement;\n};\n\ninterface HTMLHeadingElement extends HTMLElement {\n    /**\n     * Sets or retrieves a value that indicates the table alignment.\n     */\n    /** @deprecated */\n    align: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHeadingElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHeadingElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLHeadingElement: {\n    prototype: HTMLHeadingElement;\n    new(): HTMLHeadingElement;\n};\n\ninterface HTMLHtmlElement extends HTMLElement {\n    /**\n     * Sets or retrieves the DTD version that governs the current document.\n     */\n    /** @deprecated */\n    version: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHtmlElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHtmlElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLHtmlElement: {\n    prototype: HTMLHtmlElement;\n    new(): HTMLHtmlElement;\n};\n\ninterface HTMLHyperlinkElementUtils {\n    hash: string;\n    host: string;\n    hostname: string;\n    href: string;\n    readonly origin: string;\n    password: string;\n    pathname: string;\n    port: string;\n    protocol: string;\n    search: string;\n    username: string;\n}\n\ninterface HTMLIFrameElement extends HTMLElement, GetSVGDocument {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    allowFullscreen: boolean;\n    allowPaymentRequest: boolean;\n    /**\n     * Retrieves the document object of the page or frame.\n     */\n    readonly contentDocument: Document | null;\n    /**\n     * Retrieves the object of the specified.\n     */\n    readonly contentWindow: Window | null;\n    /**\n     * Sets or retrieves whether to display a border for the frame.\n     */\n    /** @deprecated */\n    frameBorder: string;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    height: string;\n    /**\n     * Sets or retrieves a URI to a long description of the object.\n     */\n    /** @deprecated */\n    longDesc: string;\n    /**\n     * Sets or retrieves the top and bottom margin heights before displaying the text in a frame.\n     */\n    /** @deprecated */\n    marginHeight: string;\n    /**\n     * Sets or retrieves the left and right margin widths before displaying the text in a frame.\n     */\n    /** @deprecated */\n    marginWidth: string;\n    /**\n     * Sets or retrieves the frame name.\n     */\n    name: string;\n    readonly referrerPolicy: ReferrerPolicy;\n    readonly sandbox: DOMTokenList;\n    /**\n     * Sets or retrieves whether the frame can be scrolled.\n     */\n    /** @deprecated */\n    scrolling: string;\n    /**\n     * Sets or retrieves a URL to be loaded by the object.\n     */\n    src: string;\n    /**\n     * Sets or retrives the content of the page that is to contain.\n     */\n    srcdoc: string;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    width: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLIFrameElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLIFrameElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLIFrameElement: {\n    prototype: HTMLIFrameElement;\n    new(): HTMLIFrameElement;\n};\n\ninterface HTMLImageElement extends HTMLElement {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves a text alternative to the graphic.\n     */\n    alt: string;\n    /**\n     * Specifies the properties of a border drawn around an object.\n     */\n    /** @deprecated */\n    border: string;\n    /**\n     * Retrieves whether the object is fully loaded.\n     */\n    readonly complete: boolean;\n    crossOrigin: string | null;\n    readonly currentSrc: string;\n    decoding: "async" | "sync" | "auto";\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    height: number;\n    /**\n     * Sets or retrieves the width of the border to draw around the object.\n     */\n    /** @deprecated */\n    hspace: number;\n    /**\n     * Sets or retrieves whether the image is a server-side image map.\n     */\n    isMap: boolean;\n    /**\n     * Sets or retrieves a Uniform Resource Identifier (URI) to a long description of the object.\n     */\n    /** @deprecated */\n    longDesc: string;\n    /** @deprecated */\n    lowsrc: string;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    /** @deprecated */\n    name: string;\n    /**\n     * The original height of the image resource before sizing.\n     */\n    readonly naturalHeight: number;\n    /**\n     * The original width of the image resource before sizing.\n     */\n    readonly naturalWidth: number;\n    referrerPolicy: string;\n    sizes: string;\n    /**\n     * The address or URL of the a media resource that is to be considered.\n     */\n    src: string;\n    srcset: string;\n    /**\n     * Sets or retrieves the URL, often with a bookmark extension (#name), to use as a client-side image map.\n     */\n    useMap: string;\n    /**\n     * Sets or retrieves the vertical margin for the object.\n     */\n    /** @deprecated */\n    vspace: number;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    width: number;\n    readonly x: number;\n    readonly y: number;\n    decode(): Promise<void>;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLImageElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLImageElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLImageElement: {\n    prototype: HTMLImageElement;\n    new(): HTMLImageElement;\n};\n\ninterface HTMLInputElement extends HTMLElement {\n    /**\n     * Sets or retrieves a comma-separated list of content types.\n     */\n    accept: string;\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves a text alternative to the graphic.\n     */\n    alt: string;\n    /**\n     * Specifies whether autocomplete is applied to an editable text field.\n     */\n    autocomplete: string;\n    /**\n     * Provides a way to direct a user to a specific field when a document loads. This can provide both direction and convenience for a user, reducing the need to click or tab to a field when a page opens. This attribute is true when present on an element, and false when missing.\n     */\n    autofocus: boolean;\n    /**\n     * Sets or retrieves the state of the check box or radio button.\n     */\n    checked: boolean;\n    /**\n     * Sets or retrieves the state of the check box or radio button.\n     */\n    defaultChecked: boolean;\n    /**\n     * Sets or retrieves the initial contents of the object.\n     */\n    defaultValue: string;\n    dirName: string;\n    disabled: boolean;\n    /**\n     * Returns a FileList object on a file type input object.\n     */\n    files: FileList | null;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Overrides the action attribute (where the data on a form is sent) on the parent form element.\n     */\n    formAction: string;\n    /**\n     * Used to override the encoding (formEnctype attribute) specified on the form element.\n     */\n    formEnctype: string;\n    /**\n     * Overrides the submit method attribute previously specified on a form element.\n     */\n    formMethod: string;\n    /**\n     * Overrides any validation or required attributes on a form or form elements to allow it to be submitted without validation. This can be used to create a "save draft"-type submit option.\n     */\n    formNoValidate: boolean;\n    /**\n     * Overrides the target attribute on a form element.\n     */\n    formTarget: string;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    height: number;\n    indeterminate: boolean;\n    readonly labels: NodeListOf<HTMLLabelElement> | null;\n    /**\n     * Specifies the ID of a pre-defined datalist of options for an input element.\n     */\n    readonly list: HTMLElement | null;\n    /**\n     * Defines the maximum acceptable value for an input element with type="number".When used with the min and step attributes, lets you control the range and increment (such as only even numbers) that the user can enter into an input field.\n     */\n    max: string;\n    /**\n     * Sets or retrieves the maximum number of characters that the user can enter in a text control.\n     */\n    maxLength: number;\n    /**\n     * Defines the minimum acceptable value for an input element with type="number". When used with the max and step attributes, lets you control the range and increment (such as even numbers only) that the user can enter into an input field.\n     */\n    min: string;\n    minLength: number;\n    /**\n     * Sets or retrieves the Boolean value indicating whether multiple items can be selected from a list.\n     */\n    multiple: boolean;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    /**\n     * Gets or sets a string containing a regular expression that the user\'s input must match.\n     */\n    pattern: string;\n    /**\n     * Gets or sets a text string that is displayed in an input field as a hint or prompt to users as the format or type of information they need to enter.The text appears in an input field until the user puts focus on the field.\n     */\n    placeholder: string;\n    readOnly: boolean;\n    /**\n     * When present, marks an element that can\'t be submitted without a value.\n     */\n    required: boolean;\n    selectionDirection: string | null;\n    /**\n     * Gets or sets the end position or offset of a text selection.\n     */\n    selectionEnd: number | null;\n    /**\n     * Gets or sets the starting position or offset of a text selection.\n     */\n    selectionStart: number | null;\n    size: number;\n    /**\n     * The address or URL of the a media resource that is to be considered.\n     */\n    src: string;\n    /**\n     * Defines an increment or jump between values that you want to allow the user to enter. When used with the max and min attributes, lets you control the range and increment (for example, allow only even numbers) that the user can enter into an input field.\n     */\n    step: string;\n    /**\n     * Returns the content type of the object.\n     */\n    type: string;\n    /**\n     * Sets or retrieves the URL, often with a bookmark extension (#name), to use as a client-side image map.\n     */\n    /** @deprecated */\n    useMap: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /**\n     * Returns the value of the data at the cursor\'s current position.\n     */\n    value: string;\n    valueAsDate: any;\n    /**\n     * Returns the input field value as a number.\n     */\n    valueAsNumber: number;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    width: number;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    /**\n     * Makes the selection equal to the current object.\n     */\n    select(): void;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    setRangeText(replacement: string): void;\n    setRangeText(replacement: string, start: number, end: number, selectionMode?: SelectionMode): void;\n    /**\n     * Sets the start and end positions of a selection in a text field.\n     * @param start The offset into the text field for the start of the selection.\n     * @param end The offset into the text field for the end of the selection.\n     * @param direction The direction in which the selection is performed.\n     */\n    setSelectionRange(start: number, end: number, direction?: "forward" | "backward" | "none"): void;\n    /**\n     * Decrements a range input control\'s value by the value given by the Step attribute. If the optional parameter is used, it will decrement the input control\'s step value multiplied by the parameter\'s value.\n     * @param n Value to decrement the value by.\n     */\n    stepDown(n?: number): void;\n    /**\n     * Increments a range input control\'s value by the value given by the Step attribute. If the optional parameter is used, will increment the input control\'s value by that value.\n     * @param n Value to increment the value by.\n     */\n    stepUp(n?: number): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLInputElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLInputElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLInputElement: {\n    prototype: HTMLInputElement;\n    new(): HTMLInputElement;\n};\n\ninterface HTMLLIElement extends HTMLElement {\n    /** @deprecated */\n    type: string;\n    /**\n     * Sets or retrieves the value of a list item.\n     */\n    value: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLIElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLIElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLLIElement: {\n    prototype: HTMLLIElement;\n    new(): HTMLLIElement;\n};\n\ninterface HTMLLabelElement extends HTMLElement {\n    readonly control: HTMLElement | null;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Sets or retrieves the object to which the given label object is assigned.\n     */\n    htmlFor: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLabelElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLabelElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLLabelElement: {\n    prototype: HTMLLabelElement;\n    new(): HTMLLabelElement;\n};\n\ninterface HTMLLegendElement extends HTMLElement {\n    /** @deprecated */\n    align: string;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLegendElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLegendElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLLegendElement: {\n    prototype: HTMLLegendElement;\n    new(): HTMLLegendElement;\n};\n\ninterface HTMLLinkElement extends HTMLElement, LinkStyle {\n    as: string;\n    /**\n     * Sets or retrieves the character set used to encode the object.\n     */\n    /** @deprecated */\n    charset: string;\n    crossOrigin: string | null;\n    disabled: boolean;\n    /**\n     * Sets or retrieves a destination URL or an anchor point.\n     */\n    href: string;\n    /**\n     * Sets or retrieves the language code of the object.\n     */\n    hreflang: string;\n    integrity: string;\n    /**\n     * Sets or retrieves the media type.\n     */\n    media: string;\n    referrerPolicy: string;\n    /**\n     * Sets or retrieves the relationship between the object and the destination of the link.\n     */\n    rel: string;\n    readonly relList: DOMTokenList;\n    /**\n     * Sets or retrieves the relationship between the object and the destination of the link.\n     */\n    /** @deprecated */\n    rev: string;\n    readonly sizes: DOMTokenList;\n    /**\n     * Sets or retrieves the window or frame at which to target content.\n     */\n    /** @deprecated */\n    target: string;\n    /**\n     * Sets or retrieves the MIME type of the object.\n     */\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLinkElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLinkElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLLinkElement: {\n    prototype: HTMLLinkElement;\n    new(): HTMLLinkElement;\n};\n\ninterface HTMLMainElement extends HTMLElement {\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMainElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMainElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMainElement: {\n    prototype: HTMLMainElement;\n    new(): HTMLMainElement;\n};\n\ninterface HTMLMapElement extends HTMLElement {\n    /**\n     * Retrieves a collection of the area objects defined for the given map object.\n     */\n    readonly areas: HTMLCollection;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMapElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMapElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMapElement: {\n    prototype: HTMLMapElement;\n    new(): HTMLMapElement;\n};\n\ninterface HTMLMarqueeElementEventMap extends HTMLElementEventMap {\n    "bounce": Event;\n    "finish": Event;\n    "start": Event;\n}\n\ninterface HTMLMarqueeElement extends HTMLElement {\n    /** @deprecated */\n    behavior: string;\n    /** @deprecated */\n    bgColor: string;\n    /** @deprecated */\n    direction: string;\n    /** @deprecated */\n    height: string;\n    /** @deprecated */\n    hspace: number;\n    /** @deprecated */\n    loop: number;\n    /** @deprecated */\n    onbounce: ((this: HTMLMarqueeElement, ev: Event) => any) | null;\n    /** @deprecated */\n    onfinish: ((this: HTMLMarqueeElement, ev: Event) => any) | null;\n    /** @deprecated */\n    onstart: ((this: HTMLMarqueeElement, ev: Event) => any) | null;\n    /** @deprecated */\n    scrollAmount: number;\n    /** @deprecated */\n    scrollDelay: number;\n    /** @deprecated */\n    trueSpeed: boolean;\n    /** @deprecated */\n    vspace: number;\n    /** @deprecated */\n    width: string;\n    /** @deprecated */\n    start(): void;\n    /** @deprecated */\n    stop(): void;\n    addEventListener<K extends keyof HTMLMarqueeElementEventMap>(type: K, listener: (this: HTMLMarqueeElement, ev: HTMLMarqueeElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLMarqueeElementEventMap>(type: K, listener: (this: HTMLMarqueeElement, ev: HTMLMarqueeElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMarqueeElement: {\n    prototype: HTMLMarqueeElement;\n    new(): HTMLMarqueeElement;\n};\n\ninterface HTMLMediaElementEventMap extends HTMLElementEventMap {\n    "encrypted": MediaEncryptedEvent;\n    "msneedkey": Event;\n}\n\ninterface HTMLMediaElement extends HTMLElement {\n    /**\n     * Returns an AudioTrackList object with the audio tracks for a given video element.\n     */\n    readonly audioTracks: AudioTrackList;\n    /**\n     * Gets or sets a value that indicates whether to start playing the media automatically.\n     */\n    autoplay: boolean;\n    /**\n     * Gets a collection of buffered time ranges.\n     */\n    readonly buffered: TimeRanges;\n    /**\n     * Gets or sets a flag that indicates whether the client provides a set of controls for the media (in case the developer does not include controls for the player).\n     */\n    controls: boolean;\n    crossOrigin: string | null;\n    /**\n     * Gets the address or URL of the current media resource that is selected by IHTMLMediaElement.\n     */\n    readonly currentSrc: string;\n    /**\n     * Gets or sets the current playback position, in seconds.\n     */\n    currentTime: number;\n    defaultMuted: boolean;\n    /**\n     * Gets or sets the default playback rate when the user is not using fast forward or reverse for a video or audio resource.\n     */\n    defaultPlaybackRate: number;\n    /**\n     * Returns the duration in seconds of the current media resource. A NaN value is returned if duration is not available, or Infinity if the media resource is streaming.\n     */\n    readonly duration: number;\n    /**\n     * Gets information about whether the playback has ended or not.\n     */\n    readonly ended: boolean;\n    /**\n     * Returns an object representing the current error state of the audio or video element.\n     */\n    readonly error: MediaError | null;\n    /**\n     * Gets or sets a flag to specify whether playback should restart after it completes.\n     */\n    loop: boolean;\n    readonly mediaKeys: MediaKeys | null;\n    /**\n     * Specifies the purpose of the audio or video media, such as background audio or alerts.\n     */\n    msAudioCategory: string;\n    /**\n     * Specifies the output device id that the audio will be sent to.\n     */\n    msAudioDeviceType: string;\n    readonly msGraphicsTrustStatus: MSGraphicsTrust;\n    /**\n     * Gets the MSMediaKeys object, which is used for decrypting media data, that is associated with this media element.\n     */\n    /** @deprecated */\n    readonly msKeys: MSMediaKeys;\n    /**\n     * Gets or sets whether the DLNA PlayTo device is available.\n     */\n    msPlayToDisabled: boolean;\n    /**\n     * Gets or sets the path to the preferred media source. This enables the Play To target device to stream the media content, which can be DRM protected, from a different location, such as a cloud media server.\n     */\n    msPlayToPreferredSourceUri: string;\n    /**\n     * Gets or sets the primary DLNA PlayTo device.\n     */\n    msPlayToPrimary: boolean;\n    /**\n     * Gets the source associated with the media element for use by the PlayToManager.\n     */\n    readonly msPlayToSource: any;\n    /**\n     * Specifies whether or not to enable low-latency playback on the media element.\n     */\n    msRealTime: boolean;\n    /**\n     * Gets or sets a flag that indicates whether the audio (either audio or the audio track on video media) is muted.\n     */\n    muted: boolean;\n    /**\n     * Gets the current network activity for the element.\n     */\n    readonly networkState: number;\n    onencrypted: ((this: HTMLMediaElement, ev: MediaEncryptedEvent) => any) | null;\n    /** @deprecated */\n    onmsneedkey: ((this: HTMLMediaElement, ev: Event) => any) | null;\n    /**\n     * Gets a flag that specifies whether playback is paused.\n     */\n    readonly paused: boolean;\n    /**\n     * Gets or sets the current rate of speed for the media resource to play. This speed is expressed as a multiple of the normal speed of the media resource.\n     */\n    playbackRate: number;\n    /**\n     * Gets TimeRanges for the current media resource that has been played.\n     */\n    readonly played: TimeRanges;\n    /**\n     * Gets or sets the current playback position, in seconds.\n     */\n    preload: string;\n    readonly readyState: number;\n    /**\n     * Returns a TimeRanges object that represents the ranges of the current media resource that can be seeked.\n     */\n    readonly seekable: TimeRanges;\n    /**\n     * Gets a flag that indicates whether the client is currently moving to a new playback position in the media resource.\n     */\n    readonly seeking: boolean;\n    /**\n     * The address or URL of the a media resource that is to be considered.\n     */\n    src: string;\n    srcObject: MediaStream | MediaSource | Blob | null;\n    readonly textTracks: TextTrackList;\n    readonly videoTracks: VideoTrackList;\n    /**\n     * Gets or sets the volume level for audio portions of the media element.\n     */\n    volume: number;\n    addTextTrack(kind: TextTrackKind, label?: string, language?: string): TextTrack;\n    /**\n     * Returns a string that specifies whether the client can play a given media resource type.\n     */\n    canPlayType(type: string): CanPlayTypeResult;\n    /**\n     * Resets the audio or video object and loads a new media resource.\n     */\n    load(): void;\n    /**\n     * Clears all effects from the media pipeline.\n     */\n    msClearEffects(): void;\n    msGetAsCastingSource(): any;\n    /**\n     * Inserts the specified audio effect into media pipeline.\n     */\n    msInsertAudioEffect(activatableClassId: string, effectRequired: boolean, config?: any): void;\n    /** @deprecated */\n    msSetMediaKeys(mediaKeys: MSMediaKeys): void;\n    /**\n     * Specifies the media protection manager for a given media pipeline.\n     */\n    msSetMediaProtectionManager(mediaProtectionManager?: any): void;\n    /**\n     * Pauses the current playback and sets paused to TRUE. This can be used to test whether the media is playing or paused. You can also use the pause or play events to tell whether the media is playing or not.\n     */\n    pause(): void;\n    /**\n     * Loads and starts playback of a media resource.\n     */\n    play(): Promise<void>;\n    setMediaKeys(mediaKeys: MediaKeys | null): Promise<void>;\n    readonly HAVE_CURRENT_DATA: number;\n    readonly HAVE_ENOUGH_DATA: number;\n    readonly HAVE_FUTURE_DATA: number;\n    readonly HAVE_METADATA: number;\n    readonly HAVE_NOTHING: number;\n    readonly NETWORK_EMPTY: number;\n    readonly NETWORK_IDLE: number;\n    readonly NETWORK_LOADING: number;\n    readonly NETWORK_NO_SOURCE: number;\n    addEventListener<K extends keyof HTMLMediaElementEventMap>(type: K, listener: (this: HTMLMediaElement, ev: HTMLMediaElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLMediaElementEventMap>(type: K, listener: (this: HTMLMediaElement, ev: HTMLMediaElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMediaElement: {\n    prototype: HTMLMediaElement;\n    new(): HTMLMediaElement;\n    readonly HAVE_CURRENT_DATA: number;\n    readonly HAVE_ENOUGH_DATA: number;\n    readonly HAVE_FUTURE_DATA: number;\n    readonly HAVE_METADATA: number;\n    readonly HAVE_NOTHING: number;\n    readonly NETWORK_EMPTY: number;\n    readonly NETWORK_IDLE: number;\n    readonly NETWORK_LOADING: number;\n    readonly NETWORK_NO_SOURCE: number;\n};\n\ninterface HTMLMenuElement extends HTMLElement {\n    /** @deprecated */\n    compact: boolean;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMenuElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMenuElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMenuElement: {\n    prototype: HTMLMenuElement;\n    new(): HTMLMenuElement;\n};\n\ninterface HTMLMetaElement extends HTMLElement {\n    /**\n     * Gets or sets meta-information to associate with httpEquiv or name.\n     */\n    content: string;\n    /**\n     * Gets or sets information used to bind the value of a content attribute of a meta element to an HTTP response header.\n     */\n    httpEquiv: string;\n    /**\n     * Sets or retrieves the value specified in the content attribute of the meta object.\n     */\n    name: string;\n    /**\n     * Sets or retrieves a scheme to be used in interpreting the value of a property specified for the object.\n     */\n    /** @deprecated */\n    scheme: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMetaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMetaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMetaElement: {\n    prototype: HTMLMetaElement;\n    new(): HTMLMetaElement;\n};\n\ninterface HTMLMeterElement extends HTMLElement {\n    high: number;\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    low: number;\n    max: number;\n    min: number;\n    optimum: number;\n    value: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMeterElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMeterElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMeterElement: {\n    prototype: HTMLMeterElement;\n    new(): HTMLMeterElement;\n};\n\ninterface HTMLModElement extends HTMLElement {\n    /**\n     * Sets or retrieves reference information about the object.\n     */\n    cite: string;\n    /**\n     * Sets or retrieves the date and time of a modification to the object.\n     */\n    dateTime: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLModElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLModElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLModElement: {\n    prototype: HTMLModElement;\n    new(): HTMLModElement;\n};\n\ninterface HTMLOListElement extends HTMLElement {\n    /** @deprecated */\n    compact: boolean;\n    reversed: boolean;\n    /**\n     * The starting number.\n     */\n    start: number;\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLOListElement: {\n    prototype: HTMLOListElement;\n    new(): HTMLOListElement;\n};\n\ninterface HTMLObjectElement extends HTMLElement, GetSVGDocument {\n    /**\n     * Retrieves a string of the URL where the object tag can be found. This is often the href of the document that the object is in, or the value set by a base element.\n     */\n    readonly BaseHref: string;\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves a character string that can be used to implement your own archive functionality for the object.\n     */\n    /** @deprecated */\n    archive: string;\n    /** @deprecated */\n    border: string;\n    /**\n     * Sets or retrieves the URL of the file containing the compiled Java class.\n     */\n    /** @deprecated */\n    code: string;\n    /**\n     * Sets or retrieves the URL of the component.\n     */\n    /** @deprecated */\n    codeBase: string;\n    /**\n     * Sets or retrieves the Internet media type for the code associated with the object.\n     */\n    /** @deprecated */\n    codeType: string;\n    /**\n     * Retrieves the document object of the page or frame.\n     */\n    readonly contentDocument: Document | null;\n    /**\n     * Sets or retrieves the URL that references the data of the object.\n     */\n    data: string;\n    /** @deprecated */\n    declare: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    height: string;\n    /** @deprecated */\n    hspace: number;\n    /**\n     * Gets or sets whether the DLNA PlayTo device is available.\n     */\n    msPlayToDisabled: boolean;\n    /**\n     * Gets or sets the path to the preferred media source. This enables the Play To target device to stream the media content, which can be DRM protected, from a different location, such as a cloud media server.\n     */\n    msPlayToPreferredSourceUri: string;\n    /**\n     * Gets or sets the primary DLNA PlayTo device.\n     */\n    msPlayToPrimary: boolean;\n    /**\n     * Gets the source associated with the media element for use by the PlayToManager.\n     */\n    readonly msPlayToSource: any;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    readonly readyState: number;\n    /**\n     * Sets or retrieves a message to be displayed while an object is loading.\n     */\n    /** @deprecated */\n    standby: string;\n    /**\n     * Sets or retrieves the MIME type of the object.\n     */\n    type: string;\n    typemustmatch: boolean;\n    /**\n     * Sets or retrieves the URL, often with a bookmark extension (#name), to use as a client-side image map.\n     */\n    useMap: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /** @deprecated */\n    vspace: number;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    width: string;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLObjectElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLObjectElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLObjectElement: {\n    prototype: HTMLObjectElement;\n    new(): HTMLObjectElement;\n};\n\ninterface HTMLOptGroupElement extends HTMLElement {\n    disabled: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Sets or retrieves a value that you can use to implement your own label functionality for the object.\n     */\n    label: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOptGroupElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOptGroupElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLOptGroupElement: {\n    prototype: HTMLOptGroupElement;\n    new(): HTMLOptGroupElement;\n};\n\ninterface HTMLOptionElement extends HTMLElement {\n    /**\n     * Sets or retrieves the status of an option.\n     */\n    defaultSelected: boolean;\n    disabled: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Sets or retrieves the ordinal position of an option in a list box.\n     */\n    readonly index: number;\n    /**\n     * Sets or retrieves a value that you can use to implement your own label functionality for the object.\n     */\n    label: string;\n    /**\n     * Sets or retrieves whether the option in the list box is the default item.\n     */\n    selected: boolean;\n    /**\n     * Sets or retrieves the text string specified by the option tag.\n     */\n    text: string;\n    /**\n     * Sets or retrieves the value which is returned to the server when the form control is submitted.\n     */\n    value: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOptionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOptionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLOptionElement: {\n    prototype: HTMLOptionElement;\n    new(): HTMLOptionElement;\n};\n\ninterface HTMLOptionsCollection extends HTMLCollectionOf<HTMLOptionElement> {\n    /**\n     * Returns the number of elements in the collection.\n     * When set to a smaller number, truncates the number of option elements in the corresponding container.\n     * When set to a greater number, adds new blank option elements to that container.\n     */\n    length: number;\n    /**\n     * Returns the index of the first selected item, if any, or −1 if there is no selected\n     * item.\n     * Can be set, to change the selection.\n     */\n    selectedIndex: number;\n    /**\n     * Inserts element before the node given by before.\n     * The before argument can be a number, in which case element is inserted before the item with that number, or an element from the\n     * collection, in which case element is inserted before that element.\n     * If before is omitted, null, or a number out of range, then element will be added at the end of the list.\n     * This method will throw a "HierarchyRequestError" DOMException if\n     * element is an ancestor of the element into which it is to be inserted.\n     */\n    add(element: HTMLOptionElement | HTMLOptGroupElement, before?: HTMLElement | number | null): void;\n    /**\n     * Removes the item with index index from the collection.\n     */\n    remove(index: number): void;\n}\n\ndeclare var HTMLOptionsCollection: {\n    prototype: HTMLOptionsCollection;\n    new(): HTMLOptionsCollection;\n};\n\ninterface HTMLOrSVGElement {\n    readonly dataset: DOMStringMap;\n    nonce: string;\n    tabIndex: number;\n    blur(): void;\n    focus(options?: FocusOptions): void;\n}\n\ninterface HTMLOutputElement extends HTMLElement {\n    defaultValue: string;\n    readonly form: HTMLFormElement | null;\n    readonly htmlFor: DOMTokenList;\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    name: string;\n    readonly type: string;\n    readonly validationMessage: string;\n    readonly validity: ValidityState;\n    value: string;\n    readonly willValidate: boolean;\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    setCustomValidity(error: string): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOutputElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOutputElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLOutputElement: {\n    prototype: HTMLOutputElement;\n    new(): HTMLOutputElement;\n};\n\ninterface HTMLParagraphElement extends HTMLElement {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLParagraphElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLParagraphElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLParagraphElement: {\n    prototype: HTMLParagraphElement;\n    new(): HTMLParagraphElement;\n};\n\ninterface HTMLParamElement extends HTMLElement {\n    /**\n     * Sets or retrieves the name of an input parameter for an element.\n     */\n    name: string;\n    /**\n     * Sets or retrieves the content type of the resource designated by the value attribute.\n     */\n    /** @deprecated */\n    type: string;\n    /**\n     * Sets or retrieves the value of an input parameter for an element.\n     */\n    value: string;\n    /**\n     * Sets or retrieves the data type of the value attribute.\n     */\n    /** @deprecated */\n    valueType: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLParamElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLParamElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLParamElement: {\n    prototype: HTMLParamElement;\n    new(): HTMLParamElement;\n};\n\ninterface HTMLPictureElement extends HTMLElement {\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLPictureElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLPictureElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLPictureElement: {\n    prototype: HTMLPictureElement;\n    new(): HTMLPictureElement;\n};\n\ninterface HTMLPreElement extends HTMLElement {\n    /**\n     * Sets or gets a value that you can use to implement your own width functionality for the object.\n     */\n    /** @deprecated */\n    width: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLPreElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLPreElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLPreElement: {\n    prototype: HTMLPreElement;\n    new(): HTMLPreElement;\n};\n\ninterface HTMLProgressElement extends HTMLElement {\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    /**\n     * Defines the maximum, or "done" value for a progress element.\n     */\n    max: number;\n    /**\n     * Returns the quotient of value/max when the value attribute is set (determinate progress bar), or -1 when the value attribute is missing (indeterminate progress bar).\n     */\n    readonly position: number;\n    /**\n     * Sets or gets the current value of a progress element. The value must be a non-negative number between 0 and the max value.\n     */\n    value: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLProgressElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLProgressElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLProgressElement: {\n    prototype: HTMLProgressElement;\n    new(): HTMLProgressElement;\n};\n\ninterface HTMLQuoteElement extends HTMLElement {\n    /**\n     * Sets or retrieves reference information about the object.\n     */\n    cite: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLQuoteElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLQuoteElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLQuoteElement: {\n    prototype: HTMLQuoteElement;\n    new(): HTMLQuoteElement;\n};\n\ninterface HTMLScriptElement extends HTMLElement {\n    async: boolean;\n    /**\n     * Sets or retrieves the character set used to encode the object.\n     */\n    /** @deprecated */\n    charset: string;\n    crossOrigin: string | null;\n    /**\n     * Sets or retrieves the status of the script.\n     */\n    defer: boolean;\n    /**\n     * Sets or retrieves the event for which the script is written.\n     */\n    /** @deprecated */\n    event: string;\n    /**\n     * Sets or retrieves the object that is bound to the event script.\n     */\n    /** @deprecated */\n    htmlFor: string;\n    integrity: string;\n    noModule: boolean;\n    referrerPolicy: string;\n    /**\n     * Retrieves the URL to an external file that contains the source code or data.\n     */\n    src: string;\n    /**\n     * Retrieves or sets the text of the object as a string.\n     */\n    text: string;\n    /**\n     * Sets or retrieves the MIME type for the associated scripting engine.\n     */\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLScriptElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLScriptElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLScriptElement: {\n    prototype: HTMLScriptElement;\n    new(): HTMLScriptElement;\n};\n\ninterface HTMLSelectElement extends HTMLElement {\n    autocomplete: string;\n    /**\n     * Provides a way to direct a user to a specific field when a document loads. This can provide both direction and convenience for a user, reducing the need to click or tab to a field when a page opens. This attribute is true when present on an element, and false when missing.\n     */\n    autofocus: boolean;\n    disabled: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    /**\n     * Sets or retrieves the number of objects in a collection.\n     */\n    length: number;\n    /**\n     * Sets or retrieves the Boolean value indicating whether multiple items can be selected from a list.\n     */\n    multiple: boolean;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    readonly options: HTMLOptionsCollection;\n    /**\n     * When present, marks an element that can\'t be submitted without a value.\n     */\n    required: boolean;\n    /**\n     * Sets or retrieves the index of the selected option in a select object.\n     */\n    selectedIndex: number;\n    readonly selectedOptions: HTMLCollectionOf<HTMLOptionElement>;\n    /**\n     * Sets or retrieves the number of rows in the list box.\n     */\n    size: number;\n    /**\n     * Retrieves the type of select control based on the value of the MULTIPLE attribute.\n     */\n    readonly type: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /**\n     * Sets or retrieves the value which is returned to the server when the form control is submitted.\n     */\n    value: string;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Adds an element to the areas, controlRange, or options collection.\n     * @param element Variant of type Number that specifies the index position in the collection where the element is placed. If no value is given, the method places the element at the end of the collection.\n     * @param before Variant of type Object that specifies an element to insert before, or null to append the object to the collection.\n     */\n    add(element: HTMLOptionElement | HTMLOptGroupElement, before?: HTMLElement | number | null): void;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    /**\n     * Retrieves a select object or an object from an options collection.\n     * @param name Variant of type Number or String that specifies the object or collection to retrieve. If this parameter is an integer, it is the zero-based index of the object. If this parameter is a string, all objects with matching name or id properties are retrieved, and a collection is returned if more than one match is made.\n     * @param index Variant of type Number that specifies the zero-based index of the object to retrieve when a collection is returned.\n     */\n    item(index: number): Element | null;\n    /**\n     * Retrieves a select object or an object from an options collection.\n     * @param namedItem A String that specifies the name or id property of the object to retrieve. A collection is returned if more than one match is made.\n     */\n    namedItem(name: string): HTMLOptionElement | null;\n    /**\n     * Removes an element from the collection.\n     * @param index Number that specifies the zero-based index of the element to remove from the collection.\n     */\n    remove(): void;\n    remove(index: number): void;\n    reportValidity(): boolean;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSelectElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSelectElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n    [name: number]: HTMLOptionElement | HTMLOptGroupElement;\n}\n\ndeclare var HTMLSelectElement: {\n    prototype: HTMLSelectElement;\n    new(): HTMLSelectElement;\n};\n\ninterface HTMLSlotElement extends HTMLElement {\n    name: string;\n    assignedElements(options?: AssignedNodesOptions): Element[];\n    assignedNodes(options?: AssignedNodesOptions): Node[];\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSlotElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSlotElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLSlotElement: {\n    prototype: HTMLSlotElement;\n    new(): HTMLSlotElement;\n};\n\ninterface HTMLSourceElement extends HTMLElement {\n    /**\n     * Gets or sets the intended media type of the media source.\n     */\n    media: string;\n    sizes: string;\n    /**\n     * The address or URL of the a media resource that is to be considered.\n     */\n    src: string;\n    srcset: string;\n    /**\n     * Gets or sets the MIME type of a media resource.\n     */\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSourceElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSourceElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLSourceElement: {\n    prototype: HTMLSourceElement;\n    new(): HTMLSourceElement;\n};\n\ninterface HTMLSpanElement extends HTMLElement {\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSpanElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSpanElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLSpanElement: {\n    prototype: HTMLSpanElement;\n    new(): HTMLSpanElement;\n};\n\ninterface HTMLStyleElement extends HTMLElement, LinkStyle {\n    /**\n     * Sets or retrieves the media type.\n     */\n    media: string;\n    /**\n     * Retrieves the CSS language in which the style sheet is written.\n     */\n    /** @deprecated */\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLStyleElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLStyleElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLStyleElement: {\n    prototype: HTMLStyleElement;\n    new(): HTMLStyleElement;\n};\n\ninterface HTMLTableCaptionElement extends HTMLElement {\n    /**\n     * Sets or retrieves the alignment of the caption or legend.\n     */\n    /** @deprecated */\n    align: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableCaptionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableCaptionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableCaptionElement: {\n    prototype: HTMLTableCaptionElement;\n    new(): HTMLTableCaptionElement;\n};\n\ninterface HTMLTableCellElement extends HTMLElement {\n    /**\n     * Sets or retrieves abbreviated text for the object.\n     */\n    abbr: string;\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves a comma-delimited list of conceptual categories associated with the object.\n     */\n    /** @deprecated */\n    axis: string;\n    /** @deprecated */\n    bgColor: string;\n    /**\n     * Retrieves the position of the object in the cells collection of a row.\n     */\n    readonly cellIndex: number;\n    /** @deprecated */\n    ch: string;\n    /** @deprecated */\n    chOff: string;\n    /**\n     * Sets or retrieves the number columns in the table that the object should span.\n     */\n    colSpan: number;\n    /**\n     * Sets or retrieves a list of header cells that provide information for the object.\n     */\n    headers: string;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    /** @deprecated */\n    height: string;\n    /**\n     * Sets or retrieves whether the browser automatically performs wordwrap.\n     */\n    /** @deprecated */\n    noWrap: boolean;\n    /**\n     * Sets or retrieves how many rows in a table the cell should span.\n     */\n    rowSpan: number;\n    /**\n     * Sets or retrieves the group of cells in a table to which the object\'s information applies.\n     */\n    scope: string;\n    /** @deprecated */\n    vAlign: string;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    /** @deprecated */\n    width: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableCellElement: {\n    prototype: HTMLTableCellElement;\n    new(): HTMLTableCellElement;\n};\n\ninterface HTMLTableColElement extends HTMLElement {\n    /**\n     * Sets or retrieves the alignment of the object relative to the display or table.\n     */\n    /** @deprecated */\n    align: string;\n    /** @deprecated */\n    ch: string;\n    /** @deprecated */\n    chOff: string;\n    /**\n     * Sets or retrieves the number of columns in the group.\n     */\n    span: number;\n    /** @deprecated */\n    vAlign: string;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    /** @deprecated */\n    width: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableColElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableColElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableColElement: {\n    prototype: HTMLTableColElement;\n    new(): HTMLTableColElement;\n};\n\ninterface HTMLTableDataCellElement extends HTMLTableCellElement {\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableDataCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableDataCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableDataCellElement: {\n    prototype: HTMLTableDataCellElement;\n    new(): HTMLTableDataCellElement;\n};\n\ninterface HTMLTableElement extends HTMLElement {\n    /**\n     * Sets or retrieves a value that indicates the table alignment.\n     */\n    /** @deprecated */\n    align: string;\n    /** @deprecated */\n    bgColor: string;\n    /**\n     * Sets or retrieves the width of the border to draw around the object.\n     */\n    /** @deprecated */\n    border: string;\n    /**\n     * Retrieves the caption object of a table.\n     */\n    caption: HTMLTableCaptionElement | null;\n    /**\n     * Sets or retrieves the amount of space between the border of the cell and the content of the cell.\n     */\n    /** @deprecated */\n    cellPadding: string;\n    /**\n     * Sets or retrieves the amount of space between cells in a table.\n     */\n    /** @deprecated */\n    cellSpacing: string;\n    /**\n     * Sets or retrieves the way the border frame around the table is displayed.\n     */\n    /** @deprecated */\n    frame: string;\n    /**\n     * Sets or retrieves the number of horizontal rows contained in the object.\n     */\n    readonly rows: HTMLCollectionOf<HTMLTableRowElement>;\n    /**\n     * Sets or retrieves which dividing lines (inner borders) are displayed.\n     */\n    /** @deprecated */\n    rules: string;\n    /**\n     * Sets or retrieves a description and/or structure of the object.\n     */\n    /** @deprecated */\n    summary: string;\n    /**\n     * Retrieves a collection of all tBody objects in the table. Objects in this collection are in source order.\n     */\n    readonly tBodies: HTMLCollectionOf<HTMLTableSectionElement>;\n    /**\n     * Retrieves the tFoot object of the table.\n     */\n    tFoot: HTMLTableSectionElement | null;\n    /**\n     * Retrieves the tHead object of the table.\n     */\n    tHead: HTMLTableSectionElement | null;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    /** @deprecated */\n    width: string;\n    /**\n     * Creates an empty caption element in the table.\n     */\n    createCaption(): HTMLTableCaptionElement;\n    /**\n     * Creates an empty tBody element in the table.\n     */\n    createTBody(): HTMLTableSectionElement;\n    /**\n     * Creates an empty tFoot element in the table.\n     */\n    createTFoot(): HTMLTableSectionElement;\n    /**\n     * Returns the tHead element object if successful, or null otherwise.\n     */\n    createTHead(): HTMLTableSectionElement;\n    /**\n     * Deletes the caption element and its contents from the table.\n     */\n    deleteCaption(): void;\n    /**\n     * Removes the specified row (tr) from the element and from the rows collection.\n     * @param index Number that specifies the zero-based position in the rows collection of the row to remove.\n     */\n    deleteRow(index: number): void;\n    /**\n     * Deletes the tFoot element and its contents from the table.\n     */\n    deleteTFoot(): void;\n    /**\n     * Deletes the tHead element and its contents from the table.\n     */\n    deleteTHead(): void;\n    /**\n     * Creates a new row (tr) in the table, and adds the row to the rows collection.\n     * @param index Number that specifies where to insert the row in the rows collection. The default value is -1, which appends the new row to the end of the rows collection.\n     */\n    insertRow(index?: number): HTMLTableRowElement;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableElement: {\n    prototype: HTMLTableElement;\n    new(): HTMLTableElement;\n};\n\ninterface HTMLTableHeaderCellElement extends HTMLTableCellElement {\n    scope: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableHeaderCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableHeaderCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableHeaderCellElement: {\n    prototype: HTMLTableHeaderCellElement;\n    new(): HTMLTableHeaderCellElement;\n};\n\ninterface HTMLTableRowElement extends HTMLElement {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    /** @deprecated */\n    bgColor: string;\n    /**\n     * Retrieves a collection of all cells in the table row.\n     */\n    readonly cells: HTMLCollectionOf<HTMLTableDataCellElement | HTMLTableHeaderCellElement>;\n    /** @deprecated */\n    ch: string;\n    /** @deprecated */\n    chOff: string;\n    /**\n     * Retrieves the position of the object in the rows collection for the table.\n     */\n    readonly rowIndex: number;\n    /**\n     * Retrieves the position of the object in the collection.\n     */\n    readonly sectionRowIndex: number;\n    /** @deprecated */\n    vAlign: string;\n    /**\n     * Removes the specified cell from the table row, as well as from the cells collection.\n     * @param index Number that specifies the zero-based position of the cell to remove from the table row. If no value is provided, the last cell in the cells collection is deleted.\n     */\n    deleteCell(index: number): void;\n    /**\n     * Creates a new cell in the table row, and adds the cell to the cells collection.\n     * @param index Number that specifies where to insert the cell in the tr. The default value is -1, which appends the new cell to the end of the cells collection.\n     */\n    insertCell(index?: number): HTMLTableDataCellElement;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableRowElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableRowElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableRowElement: {\n    prototype: HTMLTableRowElement;\n    new(): HTMLTableRowElement;\n};\n\ninterface HTMLTableSectionElement extends HTMLElement {\n    /**\n     * Sets or retrieves a value that indicates the table alignment.\n     */\n    /** @deprecated */\n    align: string;\n    /** @deprecated */\n    ch: string;\n    /** @deprecated */\n    chOff: string;\n    /**\n     * Sets or retrieves the number of horizontal rows contained in the object.\n     */\n    readonly rows: HTMLCollectionOf<HTMLTableRowElement>;\n    /** @deprecated */\n    vAlign: string;\n    /**\n     * Removes the specified row (tr) from the element and from the rows collection.\n     * @param index Number that specifies the zero-based position in the rows collection of the row to remove.\n     */\n    deleteRow(index: number): void;\n    /**\n     * Creates a new row (tr) in the table, and adds the row to the rows collection.\n     * @param index Number that specifies where to insert the row in the rows collection. The default value is -1, which appends the new row to the end of the rows collection.\n     */\n    insertRow(index?: number): HTMLTableRowElement;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableSectionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableSectionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableSectionElement: {\n    prototype: HTMLTableSectionElement;\n    new(): HTMLTableSectionElement;\n};\n\ninterface HTMLTemplateElement extends HTMLElement {\n    readonly content: DocumentFragment;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTemplateElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTemplateElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTemplateElement: {\n    prototype: HTMLTemplateElement;\n    new(): HTMLTemplateElement;\n};\n\ninterface HTMLTextAreaElement extends HTMLElement {\n    autocomplete: string;\n    /**\n     * Provides a way to direct a user to a specific field when a document loads. This can provide both direction and convenience for a user, reducing the need to click or tab to a field when a page opens. This attribute is true when present on an element, and false when missing.\n     */\n    autofocus: boolean;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    cols: number;\n    /**\n     * Sets or retrieves the initial contents of the object.\n     */\n    defaultValue: string;\n    dirName: string;\n    disabled: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    /**\n     * Sets or retrieves the maximum number of characters that the user can enter in a text control.\n     */\n    maxLength: number;\n    minLength: number;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    /**\n     * Gets or sets a text string that is displayed in an input field as a hint or prompt to users as the format or type of information they need to enter.The text appears in an input field until the user puts focus on the field.\n     */\n    placeholder: string;\n    /**\n     * Sets or retrieves the value indicated whether the content of the object is read-only.\n     */\n    readOnly: boolean;\n    /**\n     * When present, marks an element that can\'t be submitted without a value.\n     */\n    required: boolean;\n    /**\n     * Sets or retrieves the number of horizontal rows contained in the object.\n     */\n    rows: number;\n    selectionDirection: string;\n    /**\n     * Gets or sets the end position or offset of a text selection.\n     */\n    selectionEnd: number;\n    /**\n     * Gets or sets the starting position or offset of a text selection.\n     */\n    selectionStart: number;\n    readonly textLength: number;\n    /**\n     * Retrieves the type of control.\n     */\n    readonly type: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /**\n     * Retrieves or sets the text in the entry field of the textArea element.\n     */\n    value: string;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Sets or retrieves how to handle wordwrapping in the object.\n     */\n    wrap: string;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    /**\n     * Highlights the input area of a form element.\n     */\n    select(): void;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    setRangeText(replacement: string): void;\n    setRangeText(replacement: string, start: number, end: number, selectionMode?: SelectionMode): void;\n    /**\n     * Sets the start and end positions of a selection in a text field.\n     * @param start The offset into the text field for the start of the selection.\n     * @param end The offset into the text field for the end of the selection.\n     * @param direction The direction in which the selection is performed.\n     */\n    setSelectionRange(start: number, end: number, direction?: "forward" | "backward" | "none"): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTextAreaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTextAreaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTextAreaElement: {\n    prototype: HTMLTextAreaElement;\n    new(): HTMLTextAreaElement;\n};\n\ninterface HTMLTimeElement extends HTMLElement {\n    dateTime: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTimeElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTimeElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTimeElement: {\n    prototype: HTMLTimeElement;\n    new(): HTMLTimeElement;\n};\n\ninterface HTMLTitleElement extends HTMLElement {\n    /**\n     * Retrieves or sets the text of the object as a string.\n     */\n    text: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTitleElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTitleElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTitleElement: {\n    prototype: HTMLTitleElement;\n    new(): HTMLTitleElement;\n};\n\ninterface HTMLTrackElement extends HTMLElement {\n    default: boolean;\n    kind: string;\n    label: string;\n    readonly readyState: number;\n    src: string;\n    srclang: string;\n    readonly track: TextTrack;\n    readonly ERROR: number;\n    readonly LOADED: number;\n    readonly LOADING: number;\n    readonly NONE: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTrackElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTrackElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTrackElement: {\n    prototype: HTMLTrackElement;\n    new(): HTMLTrackElement;\n    readonly ERROR: number;\n    readonly LOADED: number;\n    readonly LOADING: number;\n    readonly NONE: number;\n};\n\ninterface HTMLUListElement extends HTMLElement {\n    /** @deprecated */\n    compact: boolean;\n    /** @deprecated */\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLUListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLUListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLUListElement: {\n    prototype: HTMLUListElement;\n    new(): HTMLUListElement;\n};\n\ninterface HTMLUnknownElement extends HTMLElement {\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLUnknownElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLUnknownElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLUnknownElement: {\n    prototype: HTMLUnknownElement;\n    new(): HTMLUnknownElement;\n};\n\ninterface HTMLVideoElementEventMap extends HTMLMediaElementEventMap {\n    "MSVideoFormatChanged": Event;\n    "MSVideoFrameStepCompleted": Event;\n    "MSVideoOptimalLayoutChanged": Event;\n}\n\ninterface HTMLVideoElement extends HTMLMediaElement {\n    /**\n     * Gets or sets the height of the video element.\n     */\n    height: number;\n    msHorizontalMirror: boolean;\n    readonly msIsLayoutOptimalForPlayback: boolean;\n    readonly msIsStereo3D: boolean;\n    msStereo3DPackingMode: string;\n    msStereo3DRenderMode: string;\n    msZoom: boolean;\n    onMSVideoFormatChanged: ((this: HTMLVideoElement, ev: Event) => any) | null;\n    onMSVideoFrameStepCompleted: ((this: HTMLVideoElement, ev: Event) => any) | null;\n    onMSVideoOptimalLayoutChanged: ((this: HTMLVideoElement, ev: Event) => any) | null;\n    /**\n     * Gets or sets a URL of an image to display, for example, like a movie poster. This can be a still frame from the video, or another image if no video data is available.\n     */\n    poster: string;\n    /**\n     * Gets the intrinsic height of a video in CSS pixels, or zero if the dimensions are not known.\n     */\n    readonly videoHeight: number;\n    /**\n     * Gets the intrinsic width of a video in CSS pixels, or zero if the dimensions are not known.\n     */\n    readonly videoWidth: number;\n    readonly webkitDisplayingFullscreen: boolean;\n    readonly webkitSupportsFullscreen: boolean;\n    /**\n     * Gets or sets the width of the video element.\n     */\n    width: number;\n    getVideoPlaybackQuality(): VideoPlaybackQuality;\n    msFrameStep(forward: boolean): void;\n    msInsertVideoEffect(activatableClassId: string, effectRequired: boolean, config?: any): void;\n    msSetVideoRectangle(left: number, top: number, right: number, bottom: number): void;\n    webkitEnterFullScreen(): void;\n    webkitEnterFullscreen(): void;\n    webkitExitFullScreen(): void;\n    webkitExitFullscreen(): void;\n    addEventListener<K extends keyof HTMLVideoElementEventMap>(type: K, listener: (this: HTMLVideoElement, ev: HTMLVideoElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLVideoElementEventMap>(type: K, listener: (this: HTMLVideoElement, ev: HTMLVideoElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLVideoElement: {\n    prototype: HTMLVideoElement;\n    new(): HTMLVideoElement;\n};\n\ninterface HashChangeEvent extends Event {\n    readonly newURL: string;\n    readonly oldURL: string;\n}\n\ndeclare var HashChangeEvent: {\n    prototype: HashChangeEvent;\n    new(type: string, eventInitDict?: HashChangeEventInit): HashChangeEvent;\n};\n\ninterface Headers {\n    append(name: string, value: string): void;\n    delete(name: string): void;\n    get(name: string): string | null;\n    has(name: string): boolean;\n    set(name: string, value: string): void;\n    forEach(callbackfn: (value: string, key: string, parent: Headers) => void, thisArg?: any): void;\n}\n\ndeclare var Headers: {\n    prototype: Headers;\n    new(init?: HeadersInit): Headers;\n};\n\ninterface History {\n    readonly length: number;\n    scrollRestoration: ScrollRestoration;\n    readonly state: any;\n    back(): void;\n    forward(): void;\n    go(delta?: number): void;\n    pushState(data: any, title: string, url?: string | null): void;\n    replaceState(data: any, title: string, url?: string | null): void;\n}\n\ndeclare var History: {\n    prototype: History;\n    new(): History;\n};\n\ninterface HkdfCtrParams extends Algorithm {\n    context: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    hash: string | Algorithm;\n    label: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface IDBArrayKey extends Array<IDBValidKey> {\n}\n\ninterface IDBCursor {\n    /**\n     * Returns the direction ("next", "nextunique", "prev" or "prevunique")\n     * of the cursor.\n     */\n    readonly direction: IDBCursorDirection;\n    /**\n     * Returns the key of the cursor.\n     * Throws a "InvalidStateError" DOMException if the cursor is advancing or is finished.\n     */\n    readonly key: IDBValidKey | IDBKeyRange;\n    /**\n     * Returns the effective key of the cursor.\n     * Throws a "InvalidStateError" DOMException if the cursor is advancing or is finished.\n     */\n    readonly primaryKey: IDBValidKey | IDBKeyRange;\n    /**\n     * Returns the IDBObjectStore or IDBIndex the cursor was opened from.\n     */\n    readonly source: IDBObjectStore | IDBIndex;\n    /**\n     * Advances the cursor through the next count records in\n     * range.\n     */\n    advance(count: number): void;\n    /**\n     * Advances the cursor to the next record in range matching or\n     * after key.\n     */\n    continue(key?: IDBValidKey | IDBKeyRange): void;\n    /**\n     * Advances the cursor to the next record in range matching\n     * or after key and primaryKey. Throws an "InvalidAccessError" DOMException if the source is not an index.\n     */\n    continuePrimaryKey(key: IDBValidKey | IDBKeyRange, primaryKey: IDBValidKey | IDBKeyRange): void;\n    /**\n     * Delete the record pointed at by the cursor with a new value.\n     * If successful, request\'s result will be undefined.\n     */\n    delete(): IDBRequest<undefined>;\n    /**\n     * Updated the record pointed at by the cursor with a new value.\n     * Throws a "DataError" DOMException if the effective object store uses in-line keys and the key would have changed.\n     * If successful, request\'s result will be the record\'s key.\n     */\n    update(value: any): IDBRequest<IDBValidKey>;\n}\n\ndeclare var IDBCursor: {\n    prototype: IDBCursor;\n    new(): IDBCursor;\n};\n\ninterface IDBCursorWithValue extends IDBCursor {\n    /**\n     * Returns the cursor\'s current value.\n     */\n    readonly value: any;\n}\n\ndeclare var IDBCursorWithValue: {\n    prototype: IDBCursorWithValue;\n    new(): IDBCursorWithValue;\n};\n\ninterface IDBDatabaseEventMap {\n    "abort": Event;\n    "close": Event;\n    "error": Event;\n    "versionchange": IDBVersionChangeEvent;\n}\n\ninterface IDBDatabase extends EventTarget {\n    /**\n     * Returns the name of the database.\n     */\n    readonly name: string;\n    /**\n     * Returns a list of the names of object stores in the database.\n     */\n    readonly objectStoreNames: DOMStringList;\n    onabort: ((this: IDBDatabase, ev: Event) => any) | null;\n    onclose: ((this: IDBDatabase, ev: Event) => any) | null;\n    onerror: ((this: IDBDatabase, ev: Event) => any) | null;\n    onversionchange: ((this: IDBDatabase, ev: IDBVersionChangeEvent) => any) | null;\n    /**\n     * Returns the version of the database.\n     */\n    readonly version: number;\n    /**\n     * Closes the connection once all running transactions have finished.\n     */\n    close(): void;\n    /**\n     * Creates a new object store with the given name and options and returns a new IDBObjectStore.\n     * Throws a "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    createObjectStore(name: string, optionalParameters?: IDBObjectStoreParameters): IDBObjectStore;\n    /**\n     * Deletes the object store with the given name.\n     * Throws a "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    deleteObjectStore(name: string): void;\n    /**\n     * Returns a new transaction with the given mode ("readonly" or "readwrite")\n     * and scope which can be a single object store name or an array of names.\n     */\n    transaction(storeNames: string | string[], mode?: IDBTransactionMode): IDBTransaction;\n    addEventListener<K extends keyof IDBDatabaseEventMap>(type: K, listener: (this: IDBDatabase, ev: IDBDatabaseEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBDatabaseEventMap>(type: K, listener: (this: IDBDatabase, ev: IDBDatabaseEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBDatabase: {\n    prototype: IDBDatabase;\n    new(): IDBDatabase;\n};\n\ninterface IDBEnvironment {\n    readonly indexedDB: IDBFactory;\n}\n\ninterface IDBFactory {\n    /**\n     * Compares two values as keys. Returns -1 if key1 precedes key2, 1 if key2 precedes key1, and 0 if\n     * the keys are equal.\n     * Throws a "DataError" DOMException if either input is not a valid key.\n     */\n    cmp(first: any, second: any): number;\n    /**\n     * Attempts to delete the named database. If the\n     * database already exists and there are open connections that don\'t close in response to a versionchange event, the request will be blocked until all they close. If the request\n     * is successful request\'s result will be null.\n     */\n    deleteDatabase(name: string): IDBOpenDBRequest;\n    /**\n     * Attempts to open a connection to the named database with the specified version. If the database already exists\n     * with a lower version and there are open connections that don\'t close in response to a versionchange event, the request will be blocked until all they close, then an upgrade\n     * will occur. If the database already exists with a higher\n     * version the request will fail. If the request is\n     * successful request\'s result will\n     * be the connection.\n     */\n    open(name: string, version?: number): IDBOpenDBRequest;\n}\n\ndeclare var IDBFactory: {\n    prototype: IDBFactory;\n    new(): IDBFactory;\n};\n\ninterface IDBIndex {\n    readonly keyPath: string | string[];\n    readonly multiEntry: boolean;\n    /**\n     * Updates the name of the store to newName.\n     * Throws an "InvalidStateError" DOMException if not called within an upgrade\n     * transaction.\n     */\n    name: string;\n    /**\n     * Returns the IDBObjectStore the index belongs to.\n     */\n    readonly objectStore: IDBObjectStore;\n    readonly unique: boolean;\n    /**\n     * Retrieves the number of records matching the given key or key range in query.\n     * If successful, request\'s result will be the\n     * count.\n     */\n    count(key?: IDBValidKey | IDBKeyRange): IDBRequest<number>;\n    /**\n     * Retrieves the value of the first record matching the\n     * given key or key range in query.\n     * If successful, request\'s result will be the value, or undefined if there was no matching record.\n     */\n    get(key: IDBValidKey | IDBKeyRange): IDBRequest<any | undefined>;\n    /**\n     * Retrieves the values of the records matching the given key or key range in query (up to count if given).\n     * If successful, request\'s result will be an Array of the values.\n     */\n    getAll(query?: IDBValidKey | IDBKeyRange, count?: number): IDBRequest<any[]>;\n    /**\n     * Retrieves the keys of records matching the given key or key range in query (up to count if given).\n     * If successful, request\'s result will be an Array of the keys.\n     */\n    getAllKeys(query?: IDBValidKey | IDBKeyRange, count?: number): IDBRequest<IDBValidKey[]>;\n    /**\n     * Retrieves the key of the first record matching the\n     * given key or key range in query.\n     * If successful, request\'s result will be the key, or undefined if there was no matching record.\n     */\n    getKey(key: IDBValidKey | IDBKeyRange): IDBRequest<IDBValidKey | undefined>;\n    /**\n     * Opens a cursor over the records matching query,\n     * ordered by direction. If query is null, all records in index are matched.\n     * If successful, request\'s result will be an IDBCursorWithValue, or null if there were no matching records.\n     */\n    openCursor(range?: IDBValidKey | IDBKeyRange, direction?: IDBCursorDirection): IDBRequest<IDBCursorWithValue | null>;\n    /**\n     * Opens a cursor with key only flag set over the records matching query, ordered by direction. If query is null, all records in index are matched.\n     * If successful, request\'s result will be an IDBCursor, or null if there were no matching records.\n     */\n    openKeyCursor(range?: IDBValidKey | IDBKeyRange, direction?: IDBCursorDirection): IDBRequest<IDBCursor | null>;\n}\n\ndeclare var IDBIndex: {\n    prototype: IDBIndex;\n    new(): IDBIndex;\n};\n\ninterface IDBKeyRange {\n    /**\n     * Returns lower bound, or undefined if none.\n     */\n    readonly lower: any;\n    /**\n     * Returns true if the lower open flag is set, and false otherwise.\n     */\n    readonly lowerOpen: boolean;\n    /**\n     * Returns upper bound, or undefined if none.\n     */\n    readonly upper: any;\n    /**\n     * Returns true if the upper open flag is set, and false otherwise.\n     */\n    readonly upperOpen: boolean;\n    /**\n     * Returns true if key is included in the range, and false otherwise.\n     */\n    includes(key: any): boolean;\n}\n\ndeclare var IDBKeyRange: {\n    prototype: IDBKeyRange;\n    new(): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange spanning from lower to upper.\n     * If lowerOpen is true, lower is not included in the range.\n     * If upperOpen is true, upper is not included in the range.\n     */\n    bound(lower: any, upper: any, lowerOpen?: boolean, upperOpen?: boolean): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange starting at key with no\n     * upper bound. If open is true, key is not included in the\n     * range.\n     */\n    lowerBound(lower: any, open?: boolean): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange spanning only key.\n     */\n    only(value: any): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange with no lower bound and ending at key. If open is true, key is not included in the range.\n     */\n    upperBound(upper: any, open?: boolean): IDBKeyRange;\n};\n\ninterface IDBObjectStore {\n    /**\n     * Returns true if the store has a key generator, and false otherwise.\n     */\n    readonly autoIncrement: boolean;\n    /**\n     * Returns a list of the names of indexes in the store.\n     */\n    readonly indexNames: DOMStringList;\n    /**\n     * Returns the key path of the store, or null if none.\n     */\n    readonly keyPath: string | string[];\n    /**\n     * Updates the name of the store to newName.\n     * Throws "InvalidStateError" DOMException if not called within an upgrade\n     * transaction.\n     */\n    name: string;\n    /**\n     * Returns the associated transaction.\n     */\n    readonly transaction: IDBTransaction;\n    add(value: any, key?: IDBValidKey | IDBKeyRange): IDBRequest<IDBValidKey>;\n    /**\n     * Deletes all records in store.\n     * If successful, request\'s result will\n     * be undefined.\n     */\n    clear(): IDBRequest<undefined>;\n    /**\n     * Retrieves the number of records matching the\n     * given key or key range in query.\n     * If successful, request\'s result will be the count.\n     */\n    count(key?: IDBValidKey | IDBKeyRange): IDBRequest<number>;\n    /**\n     * Creates a new index in store with the given name, keyPath and options and returns a new IDBIndex. If the keyPath and options define constraints that cannot be\n     * satisfied with the data already in store the upgrade\n     * transaction will abort with\n     * a "ConstraintError" DOMException.\n     * Throws an "InvalidStateError" DOMException if not called within an upgrade\n     * transaction.\n     */\n    createIndex(name: string, keyPath: string | string[], options?: IDBIndexParameters): IDBIndex;\n    /**\n     * Deletes records in store with the given key or in the given key range in query.\n     * If successful, request\'s result will\n     * be undefined.\n     */\n    delete(key: IDBValidKey | IDBKeyRange): IDBRequest<undefined>;\n    /**\n     * Deletes the index in store with the given name.\n     * Throws an "InvalidStateError" DOMException if not called within an upgrade\n     * transaction.\n     */\n    deleteIndex(name: string): void;\n    /**\n     * Retrieves the value of the first record matching the\n     * given key or key range in query.\n     * If successful, request\'s result will be the value, or undefined if there was no matching record.\n     */\n    get(query: IDBValidKey | IDBKeyRange): IDBRequest<any | undefined>;\n    /**\n     * Retrieves the values of the records matching the\n     * given key or key range in query (up to count if given).\n     * If successful, request\'s result will\n     * be an Array of the values.\n     */\n    getAll(query?: IDBValidKey | IDBKeyRange, count?: number): IDBRequest<any[]>;\n    /**\n     * Retrieves the keys of records matching the\n     * given key or key range in query (up to count if given).\n     * If successful, request\'s result will\n     * be an Array of the keys.\n     */\n    getAllKeys(query?: IDBValidKey | IDBKeyRange, count?: number): IDBRequest<IDBValidKey[]>;\n    /**\n     * Retrieves the key of the first record matching the\n     * given key or key range in query.\n     * If successful, request\'s result will be the key, or undefined if there was no matching record.\n     */\n    getKey(query: IDBValidKey | IDBKeyRange): IDBRequest<IDBValidKey | undefined>;\n    index(name: string): IDBIndex;\n    /**\n     * Opens a cursor over the records matching query,\n     * ordered by direction. If query is null, all records in store are matched.\n     * If successful, request\'s result will be an IDBCursorWithValue pointing at the first matching record, or null if there were no matching records.\n     */\n    openCursor(range?: IDBValidKey | IDBKeyRange, direction?: IDBCursorDirection): IDBRequest<IDBCursorWithValue | null>;\n    /**\n     * Opens a cursor with key only flag set over the records matching query, ordered by direction. If query is null, all records in store are matched.\n     * If successful, request\'s result will be an IDBCursor pointing at the first matching record, or\n     * null if there were no matching records.\n     */\n    openKeyCursor(query?: IDBValidKey | IDBKeyRange, direction?: IDBCursorDirection): IDBRequest<IDBCursor | null>;\n    put(value: any, key?: IDBValidKey | IDBKeyRange): IDBRequest<IDBValidKey>;\n}\n\ndeclare var IDBObjectStore: {\n    prototype: IDBObjectStore;\n    new(): IDBObjectStore;\n};\n\ninterface IDBOpenDBRequestEventMap extends IDBRequestEventMap {\n    "blocked": Event;\n    "upgradeneeded": IDBVersionChangeEvent;\n}\n\ninterface IDBOpenDBRequest extends IDBRequest<IDBDatabase> {\n    onblocked: ((this: IDBOpenDBRequest, ev: Event) => any) | null;\n    onupgradeneeded: ((this: IDBOpenDBRequest, ev: IDBVersionChangeEvent) => any) | null;\n    addEventListener<K extends keyof IDBOpenDBRequestEventMap>(type: K, listener: (this: IDBOpenDBRequest, ev: IDBOpenDBRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBOpenDBRequestEventMap>(type: K, listener: (this: IDBOpenDBRequest, ev: IDBOpenDBRequestEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBOpenDBRequest: {\n    prototype: IDBOpenDBRequest;\n    new(): IDBOpenDBRequest;\n};\n\ninterface IDBRequestEventMap {\n    "error": Event;\n    "success": Event;\n}\n\ninterface IDBRequest<T = any> extends EventTarget {\n    /**\n     * When a request is completed, returns the error (a DOMException), or null if the request succeeded. Throws\n     * a "InvalidStateError" DOMException if the request is still pending.\n     */\n    readonly error: DOMException | null;\n    onerror: ((this: IDBRequest<T>, ev: Event) => any) | null;\n    onsuccess: ((this: IDBRequest<T>, ev: Event) => any) | null;\n    /**\n     * Returns "pending" until a request is complete,\n     * then returns "done".\n     */\n    readonly readyState: IDBRequestReadyState;\n    /**\n     * When a request is completed, returns the result,\n     * or undefined if the request failed. Throws a\n     * "InvalidStateError" DOMException if the request is still pending.\n     */\n    readonly result: T;\n    /**\n     * Returns the IDBObjectStore, IDBIndex, or IDBCursor the request was made against, or null if is was an open\n     * request.\n     */\n    readonly source: IDBObjectStore | IDBIndex | IDBCursor;\n    /**\n     * Returns the IDBTransaction the request was made within.\n     * If this as an open request, then it returns an upgrade transaction while it is running, or null otherwise.\n     */\n    readonly transaction: IDBTransaction | null;\n    addEventListener<K extends keyof IDBRequestEventMap>(type: K, listener: (this: IDBRequest<T>, ev: IDBRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBRequestEventMap>(type: K, listener: (this: IDBRequest<T>, ev: IDBRequestEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBRequest: {\n    prototype: IDBRequest;\n    new(): IDBRequest;\n};\n\ninterface IDBTransactionEventMap {\n    "abort": Event;\n    "complete": Event;\n    "error": Event;\n}\n\ninterface IDBTransaction extends EventTarget {\n    /**\n     * Returns the transaction\'s connection.\n     */\n    readonly db: IDBDatabase;\n    /**\n     * If the transaction was aborted, returns the\n     * error (a DOMException) providing the reason.\n     */\n    readonly error: DOMException;\n    /**\n     * Returns the mode the transaction was created with\n     * ("readonly" or "readwrite"), or "versionchange" for\n     * an upgrade transaction.\n     */\n    readonly mode: IDBTransactionMode;\n    /**\n     * Returns a list of the names of object stores in the\n     * transaction\'s scope. For an upgrade transaction this is all object stores in the database.\n     */\n    readonly objectStoreNames: DOMStringList;\n    onabort: ((this: IDBTransaction, ev: Event) => any) | null;\n    oncomplete: ((this: IDBTransaction, ev: Event) => any) | null;\n    onerror: ((this: IDBTransaction, ev: Event) => any) | null;\n    /**\n     * Aborts the transaction. All pending requests will fail with\n     * a "AbortError" DOMException and all changes made to the database will be\n     * reverted.\n     */\n    abort(): void;\n    /**\n     * Returns an IDBObjectStore in the transaction\'s scope.\n     */\n    objectStore(name: string): IDBObjectStore;\n    addEventListener<K extends keyof IDBTransactionEventMap>(type: K, listener: (this: IDBTransaction, ev: IDBTransactionEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBTransactionEventMap>(type: K, listener: (this: IDBTransaction, ev: IDBTransactionEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBTransaction: {\n    prototype: IDBTransaction;\n    new(): IDBTransaction;\n};\n\ninterface IDBVersionChangeEvent extends Event {\n    readonly newVersion: number | null;\n    readonly oldVersion: number;\n}\n\ndeclare var IDBVersionChangeEvent: {\n    prototype: IDBVersionChangeEvent;\n    new(type: string, eventInitDict?: IDBVersionChangeEventInit): IDBVersionChangeEvent;\n};\n\ninterface IIRFilterNode extends AudioNode {\n    getFrequencyResponse(frequencyHz: Float32Array, magResponse: Float32Array, phaseResponse: Float32Array): void;\n}\n\ndeclare var IIRFilterNode: {\n    prototype: IIRFilterNode;\n    new(context: BaseAudioContext, options: IIRFilterOptions): IIRFilterNode;\n};\n\ninterface ImageBitmap {\n    /**\n     * Returns the intrinsic height of the image, in CSS\n     * pixels.\n     */\n    readonly height: number;\n    /**\n     * Returns the intrinsic width of the image, in CSS\n     * pixels.\n     */\n    readonly width: number;\n    /**\n     * Releases imageBitmap\'s underlying bitmap data.\n     */\n    close(): void;\n}\n\ndeclare var ImageBitmap: {\n    prototype: ImageBitmap;\n    new(): ImageBitmap;\n};\n\ninterface ImageBitmapOptions {\n    colorSpaceConversion?: "none" | "default";\n    imageOrientation?: "none" | "flipY";\n    premultiplyAlpha?: "none" | "premultiply" | "default";\n    resizeHeight?: number;\n    resizeQuality?: "pixelated" | "low" | "medium" | "high";\n    resizeWidth?: number;\n}\n\ninterface ImageBitmapRenderingContext {\n    /**\n     * Returns the canvas element that the context is bound to.\n     */\n    readonly canvas: HTMLCanvasElement;\n    /**\n     * Replaces contents of the canvas element to which context\n     * is bound with a transparent black bitmap whose size corresponds to the width and height\n     * content attributes of the canvas element.\n     */\n    transferFromImageBitmap(bitmap: ImageBitmap | null): void;\n}\n\ndeclare var ImageBitmapRenderingContext: {\n    prototype: ImageBitmapRenderingContext;\n    new(): ImageBitmapRenderingContext;\n};\n\ninterface ImageData {\n    /**\n     * Returns the one-dimensional array containing the data in RGBA order, as integers in the\n     * range 0 to 255.\n     */\n    readonly data: Uint8ClampedArray;\n    /**\n     * Returns the actual dimensions of the data in the ImageData object, in\n     * pixels.\n     */\n    readonly height: number;\n    readonly width: number;\n}\n\ndeclare var ImageData: {\n    prototype: ImageData;\n    new(width: number, height: number): ImageData;\n    new(array: Uint8ClampedArray, width: number, height: number): ImageData;\n};\n\ninterface IntersectionObserver {\n    readonly root: Element | null;\n    readonly rootMargin: string;\n    readonly thresholds: number[];\n    disconnect(): void;\n    observe(target: Element): void;\n    takeRecords(): IntersectionObserverEntry[];\n    unobserve(target: Element): void;\n}\n\ndeclare var IntersectionObserver: {\n    prototype: IntersectionObserver;\n    new(callback: IntersectionObserverCallback, options?: IntersectionObserverInit): IntersectionObserver;\n};\n\ninterface IntersectionObserverEntry {\n    readonly boundingClientRect: ClientRect | DOMRect;\n    readonly intersectionRatio: number;\n    readonly intersectionRect: ClientRect | DOMRect;\n    readonly isIntersecting: boolean;\n    readonly rootBounds: ClientRect | DOMRect;\n    readonly target: Element;\n    readonly time: number;\n}\n\ndeclare var IntersectionObserverEntry: {\n    prototype: IntersectionObserverEntry;\n    new(intersectionObserverEntryInit: IntersectionObserverEntryInit): IntersectionObserverEntry;\n};\n\ninterface KeyboardEvent extends UIEvent {\n    readonly altKey: boolean;\n    /** @deprecated */\n    char: string;\n    /** @deprecated */\n    readonly charCode: number;\n    readonly code: string;\n    readonly ctrlKey: boolean;\n    readonly key: string;\n    /** @deprecated */\n    readonly keyCode: number;\n    readonly location: number;\n    readonly metaKey: boolean;\n    readonly repeat: boolean;\n    readonly shiftKey: boolean;\n    /** @deprecated */\n    readonly which: number;\n    getModifierState(keyArg: string): boolean;\n    /** @deprecated */\n    initKeyboardEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, keyArg: string, locationArg: number, modifiersListArg: string, repeat: boolean, locale: string): void;\n    readonly DOM_KEY_LOCATION_JOYSTICK: number;\n    readonly DOM_KEY_LOCATION_LEFT: number;\n    readonly DOM_KEY_LOCATION_MOBILE: number;\n    readonly DOM_KEY_LOCATION_NUMPAD: number;\n    readonly DOM_KEY_LOCATION_RIGHT: number;\n    readonly DOM_KEY_LOCATION_STANDARD: number;\n}\n\ndeclare var KeyboardEvent: {\n    prototype: KeyboardEvent;\n    new(typeArg: string, eventInitDict?: KeyboardEventInit): KeyboardEvent;\n    readonly DOM_KEY_LOCATION_JOYSTICK: number;\n    readonly DOM_KEY_LOCATION_LEFT: number;\n    readonly DOM_KEY_LOCATION_MOBILE: number;\n    readonly DOM_KEY_LOCATION_NUMPAD: number;\n    readonly DOM_KEY_LOCATION_RIGHT: number;\n    readonly DOM_KEY_LOCATION_STANDARD: number;\n};\n\ninterface KeyframeEffect extends AnimationEffect {\n    composite: CompositeOperation;\n    iterationComposite: IterationCompositeOperation;\n    target: Element | null;\n    getKeyframes(): ComputedKeyframe[];\n    setKeyframes(keyframes: Keyframe[] | PropertyIndexedKeyframes | null): void;\n}\n\ndeclare var KeyframeEffect: {\n    prototype: KeyframeEffect;\n    new(target: Element | null, keyframes: Keyframe[] | PropertyIndexedKeyframes | null, options?: number | KeyframeEffectOptions): KeyframeEffect;\n    new(source: KeyframeEffect): KeyframeEffect;\n};\n\ninterface LinkStyle {\n    readonly sheet: StyleSheet | null;\n}\n\ninterface ListeningStateChangedEvent extends Event {\n    readonly label: string;\n    readonly state: ListeningState;\n}\n\ndeclare var ListeningStateChangedEvent: {\n    prototype: ListeningStateChangedEvent;\n    new(): ListeningStateChangedEvent;\n};\n\ninterface Location {\n    /**\n     * Returns a DOMStringList object listing the origins of the ancestor browsing contexts, from the parent browsing\n     * context to the top-level browsing context.\n     */\n    readonly ancestorOrigins: DOMStringList;\n    /**\n     * Returns the Location object\'s URL\'s fragment (includes leading "#" if non-empty).\n     * Can be set, to navigate to the same URL with a changed fragment (ignores leading "#").\n     */\n    hash: string;\n    /**\n     * Returns the Location object\'s URL\'s host and port (if different from the default\n     * port for the scheme).\n     * Can be set, to navigate to the same URL with a changed host and port.\n     */\n    host: string;\n    /**\n     * Returns the Location object\'s URL\'s host.\n     * Can be set, to navigate to the same URL with a changed host.\n     */\n    hostname: string;\n    /**\n     * Returns the Location object\'s URL.\n     * Can be set, to navigate to the given URL.\n     */\n    href: string;\n    /**\n     * Returns the Location object\'s URL\'s origin.\n     */\n    readonly origin: string;\n    /**\n     * Returns the Location object\'s URL\'s path.\n     * Can be set, to navigate to the same URL with a changed path.\n     */\n    pathname: string;\n    /**\n     * Returns the Location object\'s URL\'s port.\n     * Can be set, to navigate to the same URL with a changed port.\n     */\n    port: string;\n    /**\n     * Returns the Location object\'s URL\'s scheme.\n     * Can be set, to navigate to the same URL with a changed scheme.\n     */\n    protocol: string;\n    /**\n     * Returns the Location object\'s URL\'s query (includes leading "?" if non-empty).\n     * Can be set, to navigate to the same URL with a changed query (ignores leading "?").\n     */\n    search: string;\n    /**\n     * Navigates to the given URL.\n     */\n    assign(url: string): void;\n    /**\n     * Reloads the current page.\n     */\n    reload(): void;\n    /** @deprecated */\n    reload(forcedReload: boolean): void;\n    /**\n     * Removes the current page from the session history and navigates to the given URL.\n     */\n    replace(url: string): void;\n}\n\ndeclare var Location: {\n    prototype: Location;\n    new(): Location;\n};\n\ninterface MSAssertion {\n    readonly id: string;\n    readonly type: MSCredentialType;\n}\n\ndeclare var MSAssertion: {\n    prototype: MSAssertion;\n    new(): MSAssertion;\n};\n\ninterface MSBlobBuilder {\n    append(data: any, endings?: string): void;\n    getBlob(contentType?: string): Blob;\n}\n\ndeclare var MSBlobBuilder: {\n    prototype: MSBlobBuilder;\n    new(): MSBlobBuilder;\n};\n\ninterface MSFIDOCredentialAssertion extends MSAssertion {\n    readonly algorithm: string | Algorithm;\n    readonly attestation: any;\n    readonly publicKey: string;\n    readonly transportHints: MSTransportType[];\n}\n\ndeclare var MSFIDOCredentialAssertion: {\n    prototype: MSFIDOCredentialAssertion;\n    new(): MSFIDOCredentialAssertion;\n};\n\ninterface MSFIDOSignature {\n    readonly authnrData: string;\n    readonly clientData: string;\n    readonly signature: string;\n}\n\ndeclare var MSFIDOSignature: {\n    prototype: MSFIDOSignature;\n    new(): MSFIDOSignature;\n};\n\ninterface MSFIDOSignatureAssertion extends MSAssertion {\n    readonly signature: MSFIDOSignature;\n}\n\ndeclare var MSFIDOSignatureAssertion: {\n    prototype: MSFIDOSignatureAssertion;\n    new(): MSFIDOSignatureAssertion;\n};\n\ninterface MSFileSaver {\n    msSaveBlob(blob: any, defaultName?: string): boolean;\n    msSaveOrOpenBlob(blob: any, defaultName?: string): boolean;\n}\n\ninterface MSGesture {\n    target: Element;\n    addPointer(pointerId: number): void;\n    stop(): void;\n}\n\ndeclare var MSGesture: {\n    prototype: MSGesture;\n    new(): MSGesture;\n};\n\ninterface MSGestureEvent extends UIEvent {\n    readonly clientX: number;\n    readonly clientY: number;\n    readonly expansion: number;\n    readonly gestureObject: any;\n    readonly hwTimestamp: number;\n    readonly offsetX: number;\n    readonly offsetY: number;\n    readonly rotation: number;\n    readonly scale: number;\n    readonly screenX: number;\n    readonly screenY: number;\n    readonly translationX: number;\n    readonly translationY: number;\n    readonly velocityAngular: number;\n    readonly velocityExpansion: number;\n    readonly velocityX: number;\n    readonly velocityY: number;\n    initGestureEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, detailArg: number, screenXArg: number, screenYArg: number, clientXArg: number, clientYArg: number, offsetXArg: number, offsetYArg: number, translationXArg: number, translationYArg: number, scaleArg: number, expansionArg: number, rotationArg: number, velocityXArg: number, velocityYArg: number, velocityExpansionArg: number, velocityAngularArg: number, hwTimestampArg: number): void;\n    readonly MSGESTURE_FLAG_BEGIN: number;\n    readonly MSGESTURE_FLAG_CANCEL: number;\n    readonly MSGESTURE_FLAG_END: number;\n    readonly MSGESTURE_FLAG_INERTIA: number;\n    readonly MSGESTURE_FLAG_NONE: number;\n}\n\ndeclare var MSGestureEvent: {\n    prototype: MSGestureEvent;\n    new(): MSGestureEvent;\n    readonly MSGESTURE_FLAG_BEGIN: number;\n    readonly MSGESTURE_FLAG_CANCEL: number;\n    readonly MSGESTURE_FLAG_END: number;\n    readonly MSGESTURE_FLAG_INERTIA: number;\n    readonly MSGESTURE_FLAG_NONE: number;\n};\n\ninterface MSGraphicsTrust {\n    readonly constrictionActive: boolean;\n    readonly status: string;\n}\n\ndeclare var MSGraphicsTrust: {\n    prototype: MSGraphicsTrust;\n    new(): MSGraphicsTrust;\n};\n\ninterface MSInputMethodContextEventMap {\n    "MSCandidateWindowHide": Event;\n    "MSCandidateWindowShow": Event;\n    "MSCandidateWindowUpdate": Event;\n}\n\ninterface MSInputMethodContext extends EventTarget {\n    readonly compositionEndOffset: number;\n    readonly compositionStartOffset: number;\n    oncandidatewindowhide: ((this: MSInputMethodContext, ev: Event) => any) | null;\n    oncandidatewindowshow: ((this: MSInputMethodContext, ev: Event) => any) | null;\n    oncandidatewindowupdate: ((this: MSInputMethodContext, ev: Event) => any) | null;\n    readonly target: HTMLElement;\n    getCandidateWindowClientRect(): ClientRect;\n    getCompositionAlternatives(): string[];\n    hasComposition(): boolean;\n    isCandidateWindowVisible(): boolean;\n    addEventListener<K extends keyof MSInputMethodContextEventMap>(type: K, listener: (this: MSInputMethodContext, ev: MSInputMethodContextEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MSInputMethodContextEventMap>(type: K, listener: (this: MSInputMethodContext, ev: MSInputMethodContextEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MSInputMethodContext: {\n    prototype: MSInputMethodContext;\n    new(): MSInputMethodContext;\n};\n\ninterface MSMediaKeyError {\n    readonly code: number;\n    readonly systemCode: number;\n    readonly MS_MEDIA_KEYERR_CLIENT: number;\n    readonly MS_MEDIA_KEYERR_DOMAIN: number;\n    readonly MS_MEDIA_KEYERR_HARDWARECHANGE: number;\n    readonly MS_MEDIA_KEYERR_OUTPUT: number;\n    readonly MS_MEDIA_KEYERR_SERVICE: number;\n    readonly MS_MEDIA_KEYERR_UNKNOWN: number;\n}\n\ndeclare var MSMediaKeyError: {\n    prototype: MSMediaKeyError;\n    new(): MSMediaKeyError;\n    readonly MS_MEDIA_KEYERR_CLIENT: number;\n    readonly MS_MEDIA_KEYERR_DOMAIN: number;\n    readonly MS_MEDIA_KEYERR_HARDWARECHANGE: number;\n    readonly MS_MEDIA_KEYERR_OUTPUT: number;\n    readonly MS_MEDIA_KEYERR_SERVICE: number;\n    readonly MS_MEDIA_KEYERR_UNKNOWN: number;\n};\n\ninterface MSMediaKeyMessageEvent extends Event {\n    readonly destinationURL: string | null;\n    readonly message: Uint8Array;\n}\n\ndeclare var MSMediaKeyMessageEvent: {\n    prototype: MSMediaKeyMessageEvent;\n    new(): MSMediaKeyMessageEvent;\n};\n\ninterface MSMediaKeyNeededEvent extends Event {\n    readonly initData: Uint8Array | null;\n}\n\ndeclare var MSMediaKeyNeededEvent: {\n    prototype: MSMediaKeyNeededEvent;\n    new(): MSMediaKeyNeededEvent;\n};\n\ninterface MSMediaKeySession extends EventTarget {\n    readonly error: MSMediaKeyError | null;\n    readonly keySystem: string;\n    readonly sessionId: string;\n    close(): void;\n    update(key: Uint8Array): void;\n}\n\ndeclare var MSMediaKeySession: {\n    prototype: MSMediaKeySession;\n    new(): MSMediaKeySession;\n};\n\ninterface MSMediaKeys {\n    readonly keySystem: string;\n    createSession(type: string, initData: Uint8Array, cdmData?: Uint8Array | null): MSMediaKeySession;\n}\n\ndeclare var MSMediaKeys: {\n    prototype: MSMediaKeys;\n    new(keySystem: string): MSMediaKeys;\n    isTypeSupported(keySystem: string, type?: string | null): boolean;\n    isTypeSupportedWithFeatures(keySystem: string, type?: string | null): string;\n};\n\ninterface MSNavigatorDoNotTrack {\n    confirmSiteSpecificTrackingException(args: ConfirmSiteSpecificExceptionsInformation): boolean;\n    confirmWebWideTrackingException(args: ExceptionInformation): boolean;\n    removeSiteSpecificTrackingException(args: ExceptionInformation): void;\n    removeWebWideTrackingException(args: ExceptionInformation): void;\n    storeSiteSpecificTrackingException(args: StoreSiteSpecificExceptionsInformation): void;\n    storeWebWideTrackingException(args: StoreExceptionsInformation): void;\n}\n\ninterface MSPointerEvent extends MouseEvent {\n    readonly currentPoint: any;\n    readonly height: number;\n    readonly hwTimestamp: number;\n    readonly intermediatePoints: any;\n    readonly isPrimary: boolean;\n    readonly pointerId: number;\n    readonly pointerType: any;\n    readonly pressure: number;\n    readonly rotation: number;\n    readonly tiltX: number;\n    readonly tiltY: number;\n    readonly width: number;\n    getCurrentPoint(element: Element): void;\n    getIntermediatePoints(element: Element): void;\n    initPointerEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, detailArg: number, screenXArg: number, screenYArg: number, clientXArg: number, clientYArg: number, ctrlKeyArg: boolean, altKeyArg: boolean, shiftKeyArg: boolean, metaKeyArg: boolean, buttonArg: number, relatedTargetArg: EventTarget, offsetXArg: number, offsetYArg: number, widthArg: number, heightArg: number, pressure: number, rotation: number, tiltX: number, tiltY: number, pointerIdArg: number, pointerType: any, hwTimestampArg: number, isPrimary: boolean): void;\n}\n\ndeclare var MSPointerEvent: {\n    prototype: MSPointerEvent;\n    new(typeArg: string, eventInitDict?: PointerEventInit): MSPointerEvent;\n};\n\ninterface MSStream {\n    readonly type: string;\n    msClose(): void;\n    msDetachStream(): any;\n}\n\ndeclare var MSStream: {\n    prototype: MSStream;\n    new(): MSStream;\n};\n\ninterface MediaDeviceInfo {\n    readonly deviceId: string;\n    readonly groupId: string;\n    readonly kind: MediaDeviceKind;\n    readonly label: string;\n}\n\ndeclare var MediaDeviceInfo: {\n    prototype: MediaDeviceInfo;\n    new(): MediaDeviceInfo;\n};\n\ninterface MediaDevicesEventMap {\n    "devicechange": Event;\n}\n\ninterface MediaDevices extends EventTarget {\n    ondevicechange: ((this: MediaDevices, ev: Event) => any) | null;\n    enumerateDevices(): Promise<MediaDeviceInfo[]>;\n    getSupportedConstraints(): MediaTrackSupportedConstraints;\n    getUserMedia(constraints: MediaStreamConstraints): Promise<MediaStream>;\n    addEventListener<K extends keyof MediaDevicesEventMap>(type: K, listener: (this: MediaDevices, ev: MediaDevicesEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MediaDevicesEventMap>(type: K, listener: (this: MediaDevices, ev: MediaDevicesEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MediaDevices: {\n    prototype: MediaDevices;\n    new(): MediaDevices;\n};\n\ninterface MediaElementAudioSourceNode extends AudioNode {\n    readonly mediaElement: HTMLMediaElement;\n}\n\ndeclare var MediaElementAudioSourceNode: {\n    prototype: MediaElementAudioSourceNode;\n    new(context: AudioContext, options: MediaElementAudioSourceOptions): MediaElementAudioSourceNode;\n};\n\ninterface MediaEncryptedEvent extends Event {\n    readonly initData: ArrayBuffer | null;\n    readonly initDataType: string;\n}\n\ndeclare var MediaEncryptedEvent: {\n    prototype: MediaEncryptedEvent;\n    new(type: string, eventInitDict?: MediaEncryptedEventInit): MediaEncryptedEvent;\n};\n\ninterface MediaError {\n    readonly code: number;\n    readonly message: string;\n    readonly msExtendedCode: number;\n    readonly MEDIA_ERR_ABORTED: number;\n    readonly MEDIA_ERR_DECODE: number;\n    readonly MEDIA_ERR_NETWORK: number;\n    readonly MEDIA_ERR_SRC_NOT_SUPPORTED: number;\n    readonly MS_MEDIA_ERR_ENCRYPTED: number;\n}\n\ndeclare var MediaError: {\n    prototype: MediaError;\n    new(): MediaError;\n    readonly MEDIA_ERR_ABORTED: number;\n    readonly MEDIA_ERR_DECODE: number;\n    readonly MEDIA_ERR_NETWORK: number;\n    readonly MEDIA_ERR_SRC_NOT_SUPPORTED: number;\n    readonly MS_MEDIA_ERR_ENCRYPTED: number;\n};\n\ninterface MediaKeyMessageEvent extends Event {\n    readonly message: ArrayBuffer;\n    readonly messageType: MediaKeyMessageType;\n}\n\ndeclare var MediaKeyMessageEvent: {\n    prototype: MediaKeyMessageEvent;\n    new(type: string, eventInitDict?: MediaKeyMessageEventInit): MediaKeyMessageEvent;\n};\n\ninterface MediaKeySession extends EventTarget {\n    readonly closed: Promise<void>;\n    readonly expiration: number;\n    readonly keyStatuses: MediaKeyStatusMap;\n    readonly sessionId: string;\n    close(): Promise<void>;\n    generateRequest(initDataType: string, initData: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null): Promise<void>;\n    load(sessionId: string): Promise<boolean>;\n    remove(): Promise<void>;\n    update(response: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null): Promise<void>;\n}\n\ndeclare var MediaKeySession: {\n    prototype: MediaKeySession;\n    new(): MediaKeySession;\n};\n\ninterface MediaKeyStatusMap {\n    readonly size: number;\n    forEach(callback: Function, thisArg?: any): void;\n    get(keyId: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null): MediaKeyStatus;\n    has(keyId: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null): boolean;\n}\n\ndeclare var MediaKeyStatusMap: {\n    prototype: MediaKeyStatusMap;\n    new(): MediaKeyStatusMap;\n};\n\ninterface MediaKeySystemAccess {\n    readonly keySystem: string;\n    createMediaKeys(): Promise<MediaKeys>;\n    getConfiguration(): MediaKeySystemConfiguration;\n}\n\ndeclare var MediaKeySystemAccess: {\n    prototype: MediaKeySystemAccess;\n    new(): MediaKeySystemAccess;\n};\n\ninterface MediaKeys {\n    createSession(sessionType?: MediaKeySessionType): MediaKeySession;\n    setServerCertificate(serverCertificate: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null): Promise<void>;\n}\n\ndeclare var MediaKeys: {\n    prototype: MediaKeys;\n    new(): MediaKeys;\n};\n\ninterface MediaList {\n    readonly length: number;\n    mediaText: string;\n    appendMedium(medium: string): void;\n    deleteMedium(medium: string): void;\n    item(index: number): string | null;\n    toString(): number;\n    [index: number]: string;\n}\n\ndeclare var MediaList: {\n    prototype: MediaList;\n    new(): MediaList;\n};\n\ninterface MediaQueryListEventMap {\n    "change": MediaQueryListEvent;\n}\n\ninterface MediaQueryList extends EventTarget {\n    readonly matches: boolean;\n    readonly media: string;\n    onchange: ((this: MediaQueryList, ev: MediaQueryListEvent) => any) | null;\n    /** @deprecated */\n    addListener(listener: ((this: MediaQueryList, ev: MediaQueryListEvent) => any) | null): void;\n    /** @deprecated */\n    removeListener(listener: ((this: MediaQueryList, ev: MediaQueryListEvent) => any) | null): void;\n    addEventListener<K extends keyof MediaQueryListEventMap>(type: K, listener: (this: MediaQueryList, ev: MediaQueryListEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MediaQueryListEventMap>(type: K, listener: (this: MediaQueryList, ev: MediaQueryListEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MediaQueryList: {\n    prototype: MediaQueryList;\n    new(): MediaQueryList;\n};\n\ninterface MediaQueryListEvent extends Event {\n    readonly matches: boolean;\n    readonly media: string;\n}\n\ndeclare var MediaQueryListEvent: {\n    prototype: MediaQueryListEvent;\n    new(type: string, eventInitDict?: MediaQueryListEventInit): MediaQueryListEvent;\n};\n\ninterface MediaSource extends EventTarget {\n    readonly activeSourceBuffers: SourceBufferList;\n    duration: number;\n    readonly readyState: ReadyState;\n    readonly sourceBuffers: SourceBufferList;\n    addSourceBuffer(type: string): SourceBuffer;\n    endOfStream(error?: EndOfStreamError): void;\n    removeSourceBuffer(sourceBuffer: SourceBuffer): void;\n}\n\ndeclare var MediaSource: {\n    prototype: MediaSource;\n    new(): MediaSource;\n    isTypeSupported(type: string): boolean;\n};\n\ninterface MediaStreamEventMap {\n    "active": Event;\n    "addtrack": MediaStreamTrackEvent;\n    "inactive": Event;\n    "removetrack": MediaStreamTrackEvent;\n}\n\ninterface MediaStream extends EventTarget {\n    readonly active: boolean;\n    readonly id: string;\n    onactive: ((this: MediaStream, ev: Event) => any) | null;\n    onaddtrack: ((this: MediaStream, ev: MediaStreamTrackEvent) => any) | null;\n    oninactive: ((this: MediaStream, ev: Event) => any) | null;\n    onremovetrack: ((this: MediaStream, ev: MediaStreamTrackEvent) => any) | null;\n    addTrack(track: MediaStreamTrack): void;\n    clone(): MediaStream;\n    getAudioTracks(): MediaStreamTrack[];\n    getTrackById(trackId: string): MediaStreamTrack | null;\n    getTracks(): MediaStreamTrack[];\n    getVideoTracks(): MediaStreamTrack[];\n    removeTrack(track: MediaStreamTrack): void;\n    stop(): void;\n    addEventListener<K extends keyof MediaStreamEventMap>(type: K, listener: (this: MediaStream, ev: MediaStreamEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MediaStreamEventMap>(type: K, listener: (this: MediaStream, ev: MediaStreamEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MediaStream: {\n    prototype: MediaStream;\n    new(): MediaStream;\n    new(stream: MediaStream): MediaStream;\n    new(tracks: MediaStreamTrack[]): MediaStream;\n};\n\ninterface MediaStreamAudioDestinationNode extends AudioNode {\n    readonly stream: MediaStream;\n}\n\ndeclare var MediaStreamAudioDestinationNode: {\n    prototype: MediaStreamAudioDestinationNode;\n    new(context: AudioContext, options?: AudioNodeOptions): MediaStreamAudioDestinationNode;\n};\n\ninterface MediaStreamAudioSourceNode extends AudioNode {\n    readonly mediaStream: MediaStream;\n}\n\ndeclare var MediaStreamAudioSourceNode: {\n    prototype: MediaStreamAudioSourceNode;\n    new(context: AudioContext, options: MediaStreamAudioSourceOptions): MediaStreamAudioSourceNode;\n};\n\ninterface MediaStreamError {\n    readonly constraintName: string | null;\n    readonly message: string | null;\n    readonly name: string;\n}\n\ndeclare var MediaStreamError: {\n    prototype: MediaStreamError;\n    new(): MediaStreamError;\n};\n\ninterface MediaStreamErrorEvent extends Event {\n    readonly error: MediaStreamError | null;\n}\n\ndeclare var MediaStreamErrorEvent: {\n    prototype: MediaStreamErrorEvent;\n    new(typeArg: string, eventInitDict?: MediaStreamErrorEventInit): MediaStreamErrorEvent;\n};\n\ninterface MediaStreamEvent extends Event {\n    readonly stream: MediaStream | null;\n}\n\ndeclare var MediaStreamEvent: {\n    prototype: MediaStreamEvent;\n    new(type: string, eventInitDict: MediaStreamEventInit): MediaStreamEvent;\n};\n\ninterface MediaStreamTrackEventMap {\n    "ended": MediaStreamErrorEvent;\n    "isolationchange": Event;\n    "mute": Event;\n    "overconstrained": MediaStreamErrorEvent;\n    "unmute": Event;\n}\n\ninterface MediaStreamTrack extends EventTarget {\n    enabled: boolean;\n    readonly id: string;\n    readonly isolated: boolean;\n    readonly kind: string;\n    readonly label: string;\n    readonly muted: boolean;\n    onended: ((this: MediaStreamTrack, ev: MediaStreamErrorEvent) => any) | null;\n    onisolationchange: ((this: MediaStreamTrack, ev: Event) => any) | null;\n    onmute: ((this: MediaStreamTrack, ev: Event) => any) | null;\n    onoverconstrained: ((this: MediaStreamTrack, ev: MediaStreamErrorEvent) => any) | null;\n    onunmute: ((this: MediaStreamTrack, ev: Event) => any) | null;\n    readonly readonly: boolean;\n    readonly readyState: MediaStreamTrackState;\n    readonly remote: boolean;\n    applyConstraints(constraints: MediaTrackConstraints): Promise<void>;\n    clone(): MediaStreamTrack;\n    getCapabilities(): MediaTrackCapabilities;\n    getConstraints(): MediaTrackConstraints;\n    getSettings(): MediaTrackSettings;\n    stop(): void;\n    addEventListener<K extends keyof MediaStreamTrackEventMap>(type: K, listener: (this: MediaStreamTrack, ev: MediaStreamTrackEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MediaStreamTrackEventMap>(type: K, listener: (this: MediaStreamTrack, ev: MediaStreamTrackEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MediaStreamTrack: {\n    prototype: MediaStreamTrack;\n    new(): MediaStreamTrack;\n};\n\ninterface MediaStreamTrackAudioSourceNode extends AudioNode {\n}\n\ndeclare var MediaStreamTrackAudioSourceNode: {\n    prototype: MediaStreamTrackAudioSourceNode;\n    new(context: AudioContext, options: MediaStreamTrackAudioSourceOptions): MediaStreamTrackAudioSourceNode;\n};\n\ninterface MediaStreamTrackEvent extends Event {\n    readonly track: MediaStreamTrack;\n}\n\ndeclare var MediaStreamTrackEvent: {\n    prototype: MediaStreamTrackEvent;\n    new(typeArg: string, eventInitDict?: MediaStreamTrackEventInit): MediaStreamTrackEvent;\n};\n\ninterface MessageChannel {\n    readonly port1: MessagePort;\n    readonly port2: MessagePort;\n}\n\ndeclare var MessageChannel: {\n    prototype: MessageChannel;\n    new(): MessageChannel;\n};\n\ninterface MessageEvent extends Event {\n    /**\n     * Returns the data of the message.\n     */\n    readonly data: any;\n    /**\n     * Returns the last event ID string, for\n     * server-sent events.\n     */\n    readonly lastEventId: string;\n    /**\n     * Returns the origin of the message, for server-sent events and\n     * cross-document messaging.\n     */\n    readonly origin: string;\n    /**\n     * Returns the MessagePort array sent with the message, for cross-document\n     * messaging and channel messaging.\n     */\n    readonly ports: ReadonlyArray<MessagePort>;\n    /**\n     * Returns the WindowProxy of the source window, for cross-document\n     * messaging, and the MessagePort being attached, in the connect event fired at\n     * SharedWorkerGlobalScope objects.\n     */\n    readonly source: MessageEventSource | null;\n}\n\ndeclare var MessageEvent: {\n    prototype: MessageEvent;\n    new(type: string, eventInitDict?: MessageEventInit): MessageEvent;\n};\n\ninterface MessagePortEventMap {\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\ninterface MessagePort extends EventTarget {\n    onmessage: ((this: MessagePort, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: MessagePort, ev: MessageEvent) => any) | null;\n    /**\n     * Disconnects the port, so that it is no longer active.\n     */\n    close(): void;\n    /**\n     * Posts a message through the channel. Objects listed in transfer are\n     * transferred, not just cloned, meaning that they are no longer usable on the sending side.\n     * Throws a "DataCloneError" DOMException if\n     * transfer contains duplicate objects or port, or if message\n     * could not be cloned.\n     */\n    postMessage(message: any, transfer?: Transferable[]): void;\n    /**\n     * Begins dispatching messages received on the port.\n     */\n    start(): void;\n    addEventListener<K extends keyof MessagePortEventMap>(type: K, listener: (this: MessagePort, ev: MessagePortEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MessagePortEventMap>(type: K, listener: (this: MessagePort, ev: MessagePortEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MessagePort: {\n    prototype: MessagePort;\n    new(): MessagePort;\n};\n\ninterface MimeType {\n    readonly description: string;\n    readonly enabledPlugin: Plugin;\n    readonly suffixes: string;\n    readonly type: string;\n}\n\ndeclare var MimeType: {\n    prototype: MimeType;\n    new(): MimeType;\n};\n\ninterface MimeTypeArray {\n    readonly length: number;\n    item(index: number): Plugin;\n    namedItem(type: string): Plugin;\n    [index: number]: Plugin;\n}\n\ndeclare var MimeTypeArray: {\n    prototype: MimeTypeArray;\n    new(): MimeTypeArray;\n};\n\ninterface MouseEvent extends UIEvent {\n    readonly altKey: boolean;\n    readonly button: number;\n    readonly buttons: number;\n    readonly clientX: number;\n    readonly clientY: number;\n    readonly ctrlKey: boolean;\n    /** @deprecated */\n    readonly fromElement: Element;\n    readonly layerX: number;\n    readonly layerY: number;\n    readonly metaKey: boolean;\n    readonly movementX: number;\n    readonly movementY: number;\n    readonly offsetX: number;\n    readonly offsetY: number;\n    readonly pageX: number;\n    readonly pageY: number;\n    readonly relatedTarget: EventTarget;\n    readonly screenX: number;\n    readonly screenY: number;\n    readonly shiftKey: boolean;\n    /** @deprecated */\n    readonly toElement: Element;\n    /** @deprecated */\n    readonly which: number;\n    readonly x: number;\n    readonly y: number;\n    getModifierState(keyArg: string): boolean;\n    initMouseEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, detailArg: number, screenXArg: number, screenYArg: number, clientXArg: number, clientYArg: number, ctrlKeyArg: boolean, altKeyArg: boolean, shiftKeyArg: boolean, metaKeyArg: boolean, buttonArg: number, relatedTargetArg: EventTarget | null): void;\n}\n\ndeclare var MouseEvent: {\n    prototype: MouseEvent;\n    new(typeArg: string, eventInitDict?: MouseEventInit): MouseEvent;\n};\n\ninterface MutationEvent extends Event {\n    readonly attrChange: number;\n    readonly attrName: string;\n    readonly newValue: string;\n    readonly prevValue: string;\n    readonly relatedNode: Node;\n    initMutationEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, relatedNodeArg: Node, prevValueArg: string, newValueArg: string, attrNameArg: string, attrChangeArg: number): void;\n    readonly ADDITION: number;\n    readonly MODIFICATION: number;\n    readonly REMOVAL: number;\n}\n\ndeclare var MutationEvent: {\n    prototype: MutationEvent;\n    new(): MutationEvent;\n    readonly ADDITION: number;\n    readonly MODIFICATION: number;\n    readonly REMOVAL: number;\n};\n\ninterface MutationObserver {\n    disconnect(): void;\n    /**\n     * Instructs the user agent to observe a given target (a node) and report any mutations based on\n     * the criteria given by options (an object).\n     * The options argument allows for setting mutation\n     * observation options via object members. These are the object members that\n     * can be used:\n     * childList\n     * Set to true if mutations to target\'s children are to be observed.\n     * attributes\n     * Set to true if mutations to target\'s attributes are to be observed. Can be omitted if attributeOldValue or attributeFilter is\n     * specified.\n     * characterData\n     * Set to true if mutations to target\'s data are to be observed. Can be omitted if characterDataOldValue is specified.\n     * subtree\n     * Set to true if mutations to not just target, but\n     * also target\'s descendants are to be\n     * observed.\n     * attributeOldValue\n     * Set to true if attributes is true or omitted\n     * and target\'s attribute value before the mutation\n     * needs to be recorded.\n     * characterDataOldValue\n     * Set to true if characterData is set to true or omitted and target\'s data before the mutation\n     * needs to be recorded.\n     * attributeFilter\n     * Set to a list of attribute local names (without namespace) if not all attribute mutations need to be\n     * observed and attributes is true\n     * or omitted.\n     */\n    observe(target: Node, options?: MutationObserverInit): void;\n    /**\n     * Empties the record queue and\n     * returns what was in there.\n     */\n    takeRecords(): MutationRecord[];\n}\n\ndeclare var MutationObserver: {\n    prototype: MutationObserver;\n    new(callback: MutationCallback): MutationObserver;\n};\n\ninterface MutationRecord {\n    readonly addedNodes: NodeList;\n    /**\n     * Returns the local name of the\n     * changed attribute, and null otherwise.\n     */\n    readonly attributeName: string | null;\n    /**\n     * Returns the namespace of the\n     * changed attribute, and null otherwise.\n     */\n    readonly attributeNamespace: string | null;\n    /**\n     * Return the previous and next sibling respectively\n     * of the added or removed nodes, and null otherwise.\n     */\n    readonly nextSibling: Node | null;\n    /**\n     * The return value depends on type. For\n     * "attributes", it is the value of the\n     * changed attribute before the change.\n     * For "characterData", it is the data of the changed node before the change. For\n     * "childList", it is null.\n     */\n    readonly oldValue: string | null;\n    readonly previousSibling: Node | null;\n    /**\n     * Return the nodes added and removed\n     * respectively.\n     */\n    readonly removedNodes: NodeList;\n    readonly target: Node;\n    /**\n     * Returns "attributes" if it was an attribute mutation.\n     * "characterData" if it was a mutation to a CharacterData node. And\n     * "childList" if it was a mutation to the tree of nodes.\n     */\n    readonly type: MutationRecordType;\n}\n\ndeclare var MutationRecord: {\n    prototype: MutationRecord;\n    new(): MutationRecord;\n};\n\ninterface NamedNodeMap {\n    readonly length: number;\n    getNamedItem(qualifiedName: string): Attr | null;\n    getNamedItemNS(namespace: string | null, localName: string): Attr | null;\n    item(index: number): Attr | null;\n    removeNamedItem(qualifiedName: string): Attr;\n    removeNamedItemNS(namespace: string | null, localName: string): Attr;\n    setNamedItem(attr: Attr): Attr | null;\n    setNamedItemNS(attr: Attr): Attr | null;\n    [index: number]: Attr;\n}\n\ndeclare var NamedNodeMap: {\n    prototype: NamedNodeMap;\n    new(): NamedNodeMap;\n};\n\ninterface NavigationPreloadManager {\n    disable(): Promise<void>;\n    enable(): Promise<void>;\n    getState(): Promise<NavigationPreloadState>;\n    setHeaderValue(value: string): Promise<void>;\n}\n\ndeclare var NavigationPreloadManager: {\n    prototype: NavigationPreloadManager;\n    new(): NavigationPreloadManager;\n};\n\ninterface Navigator extends NavigatorID, NavigatorOnLine, NavigatorContentUtils, NavigatorStorageUtils, MSNavigatorDoNotTrack, MSFileSaver, NavigatorBeacon, NavigatorConcurrentHardware, NavigatorUserMedia, NavigatorLanguage, NavigatorStorage, NavigatorAutomationInformation {\n    readonly activeVRDisplays: ReadonlyArray<VRDisplay>;\n    readonly authentication: WebAuthentication;\n    readonly cookieEnabled: boolean;\n    readonly doNotTrack: string | null;\n    gamepadInputEmulation: GamepadInputEmulationType;\n    readonly geolocation: Geolocation;\n    readonly maxTouchPoints: number;\n    readonly mimeTypes: MimeTypeArray;\n    readonly msManipulationViewsEnabled: boolean;\n    readonly msMaxTouchPoints: number;\n    readonly msPointerEnabled: boolean;\n    readonly plugins: PluginArray;\n    readonly pointerEnabled: boolean;\n    readonly serviceWorker: ServiceWorkerContainer;\n    readonly webdriver: boolean;\n    getGamepads(): (Gamepad | null)[];\n    getVRDisplays(): Promise<VRDisplay[]>;\n    javaEnabled(): boolean;\n    msLaunchUri(uri: string, successCallback?: MSLaunchUriCallback, noHandlerCallback?: MSLaunchUriCallback): void;\n    requestMediaKeySystemAccess(keySystem: string, supportedConfigurations: MediaKeySystemConfiguration[]): Promise<MediaKeySystemAccess>;\n    vibrate(pattern: number | number[]): boolean;\n}\n\ndeclare var Navigator: {\n    prototype: Navigator;\n    new(): Navigator;\n};\n\ninterface NavigatorAutomationInformation {\n    readonly webdriver: boolean;\n}\n\ninterface NavigatorBeacon {\n    sendBeacon(url: string, data?: Blob | Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | FormData | string | null): boolean;\n}\n\ninterface NavigatorConcurrentHardware {\n    readonly hardwareConcurrency: number;\n}\n\ninterface NavigatorContentUtils {\n}\n\ninterface NavigatorID {\n    readonly appCodeName: string;\n    readonly appName: string;\n    readonly appVersion: string;\n    readonly platform: string;\n    readonly product: string;\n    readonly productSub: string;\n    readonly userAgent: string;\n    readonly vendor: string;\n    readonly vendorSub: string;\n}\n\ninterface NavigatorLanguage {\n    readonly language: string;\n    readonly languages: ReadonlyArray<string>;\n}\n\ninterface NavigatorOnLine {\n    readonly onLine: boolean;\n}\n\ninterface NavigatorStorage {\n    readonly storage: StorageManager;\n}\n\ninterface NavigatorStorageUtils {\n}\n\ninterface NavigatorUserMedia {\n    readonly mediaDevices: MediaDevices;\n    getDisplayMedia(constraints: MediaStreamConstraints): Promise<MediaStream>;\n    getUserMedia(constraints: MediaStreamConstraints, successCallback: NavigatorUserMediaSuccessCallback, errorCallback: NavigatorUserMediaErrorCallback): void;\n}\n\ninterface Node extends EventTarget {\n    /**\n     * Returns node\'s node document\'s document base URL.\n     */\n    readonly baseURI: string;\n    /**\n     * Returns the children.\n     */\n    readonly childNodes: NodeListOf<ChildNode>;\n    /**\n     * Returns the first child.\n     */\n    readonly firstChild: ChildNode | null;\n    /**\n     * Returns true if node is connected and false otherwise.\n     */\n    readonly isConnected: boolean;\n    /**\n     * Returns the last child.\n     */\n    readonly lastChild: ChildNode | null;\n    /** @deprecated */\n    readonly namespaceURI: string | null;\n    /**\n     * Returns the next sibling.\n     */\n    readonly nextSibling: Node | null;\n    /**\n     * Returns a string appropriate for the type of node, as\n     * follows:\n     * Element\n     * Its HTML-uppercased qualified name.\n     * Attr\n     * Its qualified name.\n     * Text\n     * "#text".\n     * CDATASection\n     * "#cdata-section".\n     * ProcessingInstruction\n     * Its target.\n     * Comment\n     * "#comment".\n     * Document\n     * "#document".\n     * DocumentType\n     * Its name.\n     * DocumentFragment\n     * "#document-fragment".\n     */\n    readonly nodeName: string;\n    readonly nodeType: number;\n    nodeValue: string | null;\n    /**\n     * Returns the node document.\n     * Returns null for documents.\n     */\n    readonly ownerDocument: Document | null;\n    /**\n     * Returns the parent element.\n     */\n    readonly parentElement: HTMLElement | null;\n    /**\n     * Returns the parent.\n     */\n    readonly parentNode: Node & ParentNode | null;\n    /**\n     * Returns the previous sibling.\n     */\n    readonly previousSibling: Node | null;\n    textContent: string | null;\n    appendChild<T extends Node>(newChild: T): T;\n    /**\n     * Returns a copy of node. If deep is true, the copy also includes the node\'s descendants.\n     */\n    cloneNode(deep?: boolean): Node;\n    compareDocumentPosition(other: Node): number;\n    /**\n     * Returns true if other is an inclusive descendant of node, and false otherwise.\n     */\n    contains(other: Node | null): boolean;\n    /**\n     * Returns node\'s shadow-including root.\n     */\n    getRootNode(options?: GetRootNodeOptions): Node;\n    /**\n     * Returns whether node has children.\n     */\n    hasChildNodes(): boolean;\n    insertBefore<T extends Node>(newChild: T, refChild: Node | null): T;\n    isDefaultNamespace(namespace: string | null): boolean;\n    /**\n     * Returns whether node and otherNode have the same properties.\n     */\n    isEqualNode(otherNode: Node | null): boolean;\n    isSameNode(otherNode: Node | null): boolean;\n    lookupNamespaceURI(prefix: string | null): string | null;\n    lookupPrefix(namespace: string | null): string | null;\n    /**\n     * Removes empty exclusive Text nodes and concatenates the data of remaining contiguous exclusive Text nodes into the first of their nodes.\n     */\n    normalize(): void;\n    removeChild<T extends Node>(oldChild: T): T;\n    replaceChild<T extends Node>(newChild: Node, oldChild: T): T;\n    readonly ATTRIBUTE_NODE: number;\n    readonly CDATA_SECTION_NODE: number;\n    readonly COMMENT_NODE: number;\n    readonly DOCUMENT_FRAGMENT_NODE: number;\n    readonly DOCUMENT_NODE: number;\n    readonly DOCUMENT_POSITION_CONTAINED_BY: number;\n    readonly DOCUMENT_POSITION_CONTAINS: number;\n    readonly DOCUMENT_POSITION_DISCONNECTED: number;\n    readonly DOCUMENT_POSITION_FOLLOWING: number;\n    readonly DOCUMENT_POSITION_IMPLEMENTATION_SPECIFIC: number;\n    readonly DOCUMENT_POSITION_PRECEDING: number;\n    readonly DOCUMENT_TYPE_NODE: number;\n    readonly ELEMENT_NODE: number;\n    readonly ENTITY_NODE: number;\n    readonly ENTITY_REFERENCE_NODE: number;\n    readonly NOTATION_NODE: number;\n    readonly PROCESSING_INSTRUCTION_NODE: number;\n    readonly TEXT_NODE: number;\n}\n\ndeclare var Node: {\n    prototype: Node;\n    new(): Node;\n    readonly ATTRIBUTE_NODE: number;\n    readonly CDATA_SECTION_NODE: number;\n    readonly COMMENT_NODE: number;\n    readonly DOCUMENT_FRAGMENT_NODE: number;\n    readonly DOCUMENT_NODE: number;\n    readonly DOCUMENT_POSITION_CONTAINED_BY: number;\n    readonly DOCUMENT_POSITION_CONTAINS: number;\n    readonly DOCUMENT_POSITION_DISCONNECTED: number;\n    readonly DOCUMENT_POSITION_FOLLOWING: number;\n    readonly DOCUMENT_POSITION_IMPLEMENTATION_SPECIFIC: number;\n    readonly DOCUMENT_POSITION_PRECEDING: number;\n    readonly DOCUMENT_TYPE_NODE: number;\n    readonly ELEMENT_NODE: number;\n    readonly ENTITY_NODE: number;\n    readonly ENTITY_REFERENCE_NODE: number;\n    readonly NOTATION_NODE: number;\n    readonly PROCESSING_INSTRUCTION_NODE: number;\n    readonly TEXT_NODE: number;\n};\n\ninterface NodeFilter {\n    acceptNode(node: Node): number;\n}\n\ndeclare var NodeFilter: {\n    readonly FILTER_ACCEPT: number;\n    readonly FILTER_REJECT: number;\n    readonly FILTER_SKIP: number;\n    readonly SHOW_ALL: number;\n    readonly SHOW_ATTRIBUTE: number;\n    readonly SHOW_CDATA_SECTION: number;\n    readonly SHOW_COMMENT: number;\n    readonly SHOW_DOCUMENT: number;\n    readonly SHOW_DOCUMENT_FRAGMENT: number;\n    readonly SHOW_DOCUMENT_TYPE: number;\n    readonly SHOW_ELEMENT: number;\n    readonly SHOW_ENTITY: number;\n    readonly SHOW_ENTITY_REFERENCE: number;\n    readonly SHOW_NOTATION: number;\n    readonly SHOW_PROCESSING_INSTRUCTION: number;\n    readonly SHOW_TEXT: number;\n};\n\ninterface NodeIterator {\n    readonly filter: NodeFilter | null;\n    readonly pointerBeforeReferenceNode: boolean;\n    readonly referenceNode: Node;\n    readonly root: Node;\n    readonly whatToShow: number;\n    detach(): void;\n    nextNode(): Node | null;\n    previousNode(): Node | null;\n}\n\ndeclare var NodeIterator: {\n    prototype: NodeIterator;\n    new(): NodeIterator;\n};\n\ninterface NodeList {\n    /**\n     * Returns the number of nodes in the collection.\n     */\n    readonly length: number;\n    /**\n     * element = collection[index]\n     */\n    item(index: number): Node | null;\n    /**\n     * Performs the specified action for each node in an list.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the callbackfn function one time for each element in the list.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: Node, key: number, parent: NodeList) => void, thisArg?: any): void;\n    [index: number]: Node;\n}\n\ndeclare var NodeList: {\n    prototype: NodeList;\n    new(): NodeList;\n};\n\ninterface NodeListOf<TNode extends Node> extends NodeList {\n    length: number;\n    item(index: number): TNode;\n    /**\n     * Performs the specified action for each node in an list.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the callbackfn function one time for each element in the list.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: TNode, key: number, parent: NodeListOf<TNode>) => void, thisArg?: any): void;\n    [index: number]: TNode;\n}\n\ninterface NodeSelector {\n    querySelector<K extends keyof HTMLElementTagNameMap>(selectors: K): HTMLElementTagNameMap[K] | null;\n    querySelector<K extends keyof SVGElementTagNameMap>(selectors: K): SVGElementTagNameMap[K] | null;\n    querySelector<E extends Element = Element>(selectors: string): E | null;\n    querySelectorAll<K extends keyof HTMLElementTagNameMap>(selectors: K): NodeListOf<HTMLElementTagNameMap[K]>;\n    querySelectorAll<K extends keyof SVGElementTagNameMap>(selectors: K): NodeListOf<SVGElementTagNameMap[K]>;\n    querySelectorAll<E extends Element = Element>(selectors: string): NodeListOf<E>;\n}\n\ninterface NonDocumentTypeChildNode {\n    /**\n     * Returns the first following sibling that\n     * is an element, and null otherwise.\n     */\n    readonly nextElementSibling: Element | null;\n    /**\n     * Returns the first preceding sibling that\n     * is an element, and null otherwise.\n     */\n    readonly previousElementSibling: Element | null;\n}\n\ninterface NonElementParentNode {\n    /**\n     * Returns the first element within node\'s descendants whose ID is elementId.\n     */\n    getElementById(elementId: string): Element | null;\n}\n\ninterface NotificationEventMap {\n    "click": Event;\n    "close": Event;\n    "error": Event;\n    "show": Event;\n}\n\ninterface Notification extends EventTarget {\n    readonly actions: ReadonlyArray<NotificationAction>;\n    readonly badge: string;\n    readonly body: string;\n    readonly data: any;\n    readonly dir: NotificationDirection;\n    readonly icon: string;\n    readonly image: string;\n    readonly lang: string;\n    onclick: ((this: Notification, ev: Event) => any) | null;\n    onclose: ((this: Notification, ev: Event) => any) | null;\n    onerror: ((this: Notification, ev: Event) => any) | null;\n    onshow: ((this: Notification, ev: Event) => any) | null;\n    readonly renotify: boolean;\n    readonly requireInteraction: boolean;\n    readonly silent: boolean;\n    readonly tag: string;\n    readonly timestamp: number;\n    readonly title: string;\n    readonly vibrate: ReadonlyArray<number>;\n    close(): void;\n    addEventListener<K extends keyof NotificationEventMap>(type: K, listener: (this: Notification, ev: NotificationEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof NotificationEventMap>(type: K, listener: (this: Notification, ev: NotificationEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Notification: {\n    prototype: Notification;\n    new(title: string, options?: NotificationOptions): Notification;\n    readonly maxActions: number;\n    readonly permission: NotificationPermission;\n    requestPermission(deprecatedCallback?: NotificationPermissionCallback): Promise<NotificationPermission>;\n};\n\ninterface OES_element_index_uint {\n}\n\ninterface OES_standard_derivatives {\n    readonly FRAGMENT_SHADER_DERIVATIVE_HINT_OES: GLenum;\n}\n\ninterface OES_texture_float {\n}\n\ninterface OES_texture_float_linear {\n}\n\ninterface OES_texture_half_float {\n    readonly HALF_FLOAT_OES: GLenum;\n}\n\ninterface OES_texture_half_float_linear {\n}\n\ninterface OES_vertex_array_object {\n    bindVertexArrayOES(arrayObject: WebGLVertexArrayObjectOES | null): void;\n    createVertexArrayOES(): WebGLVertexArrayObjectOES | null;\n    deleteVertexArrayOES(arrayObject: WebGLVertexArrayObjectOES | null): void;\n    isVertexArrayOES(arrayObject: WebGLVertexArrayObjectOES | null): GLboolean;\n    readonly VERTEX_ARRAY_BINDING_OES: GLenum;\n}\n\ninterface OfflineAudioCompletionEvent extends Event {\n    readonly renderedBuffer: AudioBuffer;\n}\n\ndeclare var OfflineAudioCompletionEvent: {\n    prototype: OfflineAudioCompletionEvent;\n    new(type: string, eventInitDict: OfflineAudioCompletionEventInit): OfflineAudioCompletionEvent;\n};\n\ninterface OfflineAudioContextEventMap extends BaseAudioContextEventMap {\n    "complete": OfflineAudioCompletionEvent;\n}\n\ninterface OfflineAudioContext extends BaseAudioContext {\n    readonly length: number;\n    oncomplete: ((this: OfflineAudioContext, ev: OfflineAudioCompletionEvent) => any) | null;\n    startRendering(): Promise<AudioBuffer>;\n    suspend(suspendTime: number): Promise<void>;\n    addEventListener<K extends keyof OfflineAudioContextEventMap>(type: K, listener: (this: OfflineAudioContext, ev: OfflineAudioContextEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof OfflineAudioContextEventMap>(type: K, listener: (this: OfflineAudioContext, ev: OfflineAudioContextEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var OfflineAudioContext: {\n    prototype: OfflineAudioContext;\n    new(contextOptions: OfflineAudioContextOptions): OfflineAudioContext;\n    new(numberOfChannels: number, length: number, sampleRate: number): OfflineAudioContext;\n};\n\ninterface OscillatorNode extends AudioScheduledSourceNode {\n    readonly detune: AudioParam;\n    readonly frequency: AudioParam;\n    type: OscillatorType;\n    setPeriodicWave(periodicWave: PeriodicWave): void;\n    addEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: OscillatorNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: OscillatorNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var OscillatorNode: {\n    prototype: OscillatorNode;\n    new(context: BaseAudioContext, options?: OscillatorOptions): OscillatorNode;\n};\n\ninterface OverflowEvent extends UIEvent {\n    readonly horizontalOverflow: boolean;\n    readonly orient: number;\n    readonly verticalOverflow: boolean;\n    readonly BOTH: number;\n    readonly HORIZONTAL: number;\n    readonly VERTICAL: number;\n}\n\ndeclare var OverflowEvent: {\n    prototype: OverflowEvent;\n    new(): OverflowEvent;\n    readonly BOTH: number;\n    readonly HORIZONTAL: number;\n    readonly VERTICAL: number;\n};\n\ninterface PageTransitionEvent extends Event {\n    readonly persisted: boolean;\n}\n\ndeclare var PageTransitionEvent: {\n    prototype: PageTransitionEvent;\n    new(): PageTransitionEvent;\n};\n\ninterface PannerNode extends AudioNode {\n    coneInnerAngle: number;\n    coneOuterAngle: number;\n    coneOuterGain: number;\n    distanceModel: DistanceModelType;\n    maxDistance: number;\n    readonly orientationX: AudioParam;\n    readonly orientationY: AudioParam;\n    readonly orientationZ: AudioParam;\n    panningModel: PanningModelType;\n    readonly positionX: AudioParam;\n    readonly positionY: AudioParam;\n    readonly positionZ: AudioParam;\n    refDistance: number;\n    rolloffFactor: number;\n    /** @deprecated */\n    setOrientation(x: number, y: number, z: number): void;\n    /** @deprecated */\n    setPosition(x: number, y: number, z: number): void;\n}\n\ndeclare var PannerNode: {\n    prototype: PannerNode;\n    new(context: BaseAudioContext, options?: PannerOptions): PannerNode;\n};\n\ninterface ParentNode {\n    readonly childElementCount: number;\n    /**\n     * Returns the child elements.\n     */\n    readonly children: HTMLCollection;\n    /**\n     * Returns the first child that is an element, and null otherwise.\n     */\n    readonly firstElementChild: Element | null;\n    /**\n     * Returns the last child that is an element, and null otherwise.\n     */\n    readonly lastElementChild: Element | null;\n    /**\n     * Inserts nodes after the last child of node, while replacing\n     * strings in nodes with equivalent Text nodes.\n     * Throws a "HierarchyRequestError" DOMException if the constraints of\n     * the node tree are violated.\n     */\n    append(...nodes: (Node | string)[]): void;\n    /**\n     * Inserts nodes before the first child of node, while\n     * replacing strings in nodes with equivalent Text nodes.\n     * Throws a "HierarchyRequestError" DOMException if the constraints of\n     * the node tree are violated.\n     */\n    prepend(...nodes: (Node | string)[]): void;\n    /**\n     * Returns the first element that is a descendant of node that\n     * matches selectors.\n     */\n    querySelector<K extends keyof HTMLElementTagNameMap>(selectors: K): HTMLElementTagNameMap[K] | null;\n    querySelector<K extends keyof SVGElementTagNameMap>(selectors: K): SVGElementTagNameMap[K] | null;\n    querySelector<E extends Element = Element>(selectors: string): E | null;\n    /**\n     * Returns all element descendants of node that\n     * match selectors.\n     */\n    querySelectorAll<K extends keyof HTMLElementTagNameMap>(selectors: K): NodeListOf<HTMLElementTagNameMap[K]>;\n    querySelectorAll<K extends keyof SVGElementTagNameMap>(selectors: K): NodeListOf<SVGElementTagNameMap[K]>;\n    querySelectorAll<E extends Element = Element>(selectors: string): NodeListOf<E>;\n}\n\ninterface Path2D extends CanvasPath {\n    addPath(path: Path2D, transform?: DOMMatrix2DInit): void;\n}\n\ndeclare var Path2D: {\n    prototype: Path2D;\n    new(path?: Path2D | string): Path2D;\n};\n\ninterface PaymentAddress {\n    readonly addressLine: string[];\n    readonly city: string;\n    readonly country: string;\n    readonly dependentLocality: string;\n    readonly languageCode: string;\n    readonly organization: string;\n    readonly phone: string;\n    readonly postalCode: string;\n    readonly recipient: string;\n    readonly region: string;\n    readonly sortingCode: string;\n    toJSON(): any;\n}\n\ndeclare var PaymentAddress: {\n    prototype: PaymentAddress;\n    new(): PaymentAddress;\n};\n\ninterface PaymentRequestEventMap {\n    "shippingaddresschange": Event;\n    "shippingoptionchange": Event;\n}\n\ninterface PaymentRequest extends EventTarget {\n    readonly id: string;\n    onshippingaddresschange: ((this: PaymentRequest, ev: Event) => any) | null;\n    onshippingoptionchange: ((this: PaymentRequest, ev: Event) => any) | null;\n    readonly shippingAddress: PaymentAddress | null;\n    readonly shippingOption: string | null;\n    readonly shippingType: PaymentShippingType | null;\n    abort(): Promise<void>;\n    canMakePayment(): Promise<boolean>;\n    show(): Promise<PaymentResponse>;\n    addEventListener<K extends keyof PaymentRequestEventMap>(type: K, listener: (this: PaymentRequest, ev: PaymentRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof PaymentRequestEventMap>(type: K, listener: (this: PaymentRequest, ev: PaymentRequestEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var PaymentRequest: {\n    prototype: PaymentRequest;\n    new(methodData: PaymentMethodData[], details: PaymentDetailsInit, options?: PaymentOptions): PaymentRequest;\n};\n\ninterface PaymentRequestUpdateEvent extends Event {\n    updateWith(detailsPromise: Promise<PaymentDetailsUpdate>): void;\n}\n\ndeclare var PaymentRequestUpdateEvent: {\n    prototype: PaymentRequestUpdateEvent;\n    new(type: string, eventInitDict?: PaymentRequestUpdateEventInit): PaymentRequestUpdateEvent;\n};\n\ninterface PaymentResponse {\n    readonly details: any;\n    readonly methodName: string;\n    readonly payerEmail: string | null;\n    readonly payerName: string | null;\n    readonly payerPhone: string | null;\n    readonly requestId: string;\n    readonly shippingAddress: PaymentAddress | null;\n    readonly shippingOption: string | null;\n    complete(result?: PaymentComplete): Promise<void>;\n    toJSON(): any;\n}\n\ndeclare var PaymentResponse: {\n    prototype: PaymentResponse;\n    new(): PaymentResponse;\n};\n\ninterface PerfWidgetExternal {\n    readonly activeNetworkRequestCount: number;\n    readonly averageFrameTime: number;\n    readonly averagePaintTime: number;\n    readonly extraInformationEnabled: boolean;\n    readonly independentRenderingEnabled: boolean;\n    readonly irDisablingContentString: string;\n    readonly irStatusAvailable: boolean;\n    readonly maxCpuSpeed: number;\n    readonly paintRequestsPerSecond: number;\n    readonly performanceCounter: number;\n    readonly performanceCounterFrequency: number;\n    addEventListener(eventType: string, callback: Function): void;\n    getMemoryUsage(): number;\n    getProcessCpuUsage(): number;\n    getRecentCpuUsage(last: number | null): any;\n    getRecentFrames(last: number | null): any;\n    getRecentMemoryUsage(last: number | null): any;\n    getRecentPaintRequests(last: number | null): any;\n    removeEventListener(eventType: string, callback: Function): void;\n    repositionWindow(x: number, y: number): void;\n    resizeWindow(width: number, height: number): void;\n}\n\ndeclare var PerfWidgetExternal: {\n    prototype: PerfWidgetExternal;\n    new(): PerfWidgetExternal;\n};\n\ninterface PerformanceEventMap {\n    "resourcetimingbufferfull": Event;\n}\n\ninterface Performance extends EventTarget {\n    /** @deprecated */\n    readonly navigation: PerformanceNavigation;\n    onresourcetimingbufferfull: ((this: Performance, ev: Event) => any) | null;\n    readonly timeOrigin: number;\n    /** @deprecated */\n    readonly timing: PerformanceTiming;\n    clearMarks(markName?: string): void;\n    clearMeasures(measureName?: string): void;\n    clearResourceTimings(): void;\n    getEntries(): PerformanceEntryList;\n    getEntriesByName(name: string, type?: string): PerformanceEntryList;\n    getEntriesByType(type: string): PerformanceEntryList;\n    mark(markName: string): void;\n    measure(measureName: string, startMark?: string, endMark?: string): void;\n    now(): number;\n    setResourceTimingBufferSize(maxSize: number): void;\n    toJSON(): any;\n    addEventListener<K extends keyof PerformanceEventMap>(type: K, listener: (this: Performance, ev: PerformanceEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof PerformanceEventMap>(type: K, listener: (this: Performance, ev: PerformanceEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Performance: {\n    prototype: Performance;\n    new(): Performance;\n};\n\ninterface PerformanceEntry {\n    readonly duration: number;\n    readonly entryType: string;\n    readonly name: string;\n    readonly startTime: number;\n    toJSON(): any;\n}\n\ndeclare var PerformanceEntry: {\n    prototype: PerformanceEntry;\n    new(): PerformanceEntry;\n};\n\ninterface PerformanceMark extends PerformanceEntry {\n}\n\ndeclare var PerformanceMark: {\n    prototype: PerformanceMark;\n    new(): PerformanceMark;\n};\n\ninterface PerformanceMeasure extends PerformanceEntry {\n}\n\ndeclare var PerformanceMeasure: {\n    prototype: PerformanceMeasure;\n    new(): PerformanceMeasure;\n};\n\ninterface PerformanceNavigation {\n    readonly redirectCount: number;\n    readonly type: number;\n    toJSON(): any;\n    readonly TYPE_BACK_FORWARD: number;\n    readonly TYPE_NAVIGATE: number;\n    readonly TYPE_RELOAD: number;\n    readonly TYPE_RESERVED: number;\n}\n\ndeclare var PerformanceNavigation: {\n    prototype: PerformanceNavigation;\n    new(): PerformanceNavigation;\n    readonly TYPE_BACK_FORWARD: number;\n    readonly TYPE_NAVIGATE: number;\n    readonly TYPE_RELOAD: number;\n    readonly TYPE_RESERVED: number;\n};\n\ninterface PerformanceNavigationTiming extends PerformanceResourceTiming {\n    readonly domComplete: number;\n    readonly domContentLoadedEventEnd: number;\n    readonly domContentLoadedEventStart: number;\n    readonly domInteractive: number;\n    readonly loadEventEnd: number;\n    readonly loadEventStart: number;\n    readonly redirectCount: number;\n    readonly type: NavigationType;\n    readonly unloadEventEnd: number;\n    readonly unloadEventStart: number;\n    toJSON(): any;\n}\n\ndeclare var PerformanceNavigationTiming: {\n    prototype: PerformanceNavigationTiming;\n    new(): PerformanceNavigationTiming;\n};\n\ninterface PerformanceObserver {\n    disconnect(): void;\n    observe(options: PerformanceObserverInit): void;\n    takeRecords(): PerformanceEntryList;\n}\n\ndeclare var PerformanceObserver: {\n    prototype: PerformanceObserver;\n    new(callback: PerformanceObserverCallback): PerformanceObserver;\n};\n\ninterface PerformanceObserverEntryList {\n    getEntries(): PerformanceEntryList;\n    getEntriesByName(name: string, type?: string): PerformanceEntryList;\n    getEntriesByType(type: string): PerformanceEntryList;\n}\n\ndeclare var PerformanceObserverEntryList: {\n    prototype: PerformanceObserverEntryList;\n    new(): PerformanceObserverEntryList;\n};\n\ninterface PerformanceResourceTiming extends PerformanceEntry {\n    readonly connectEnd: number;\n    readonly connectStart: number;\n    readonly decodedBodySize: number;\n    readonly domainLookupEnd: number;\n    readonly domainLookupStart: number;\n    readonly encodedBodySize: number;\n    readonly fetchStart: number;\n    readonly initiatorType: string;\n    readonly nextHopProtocol: string;\n    readonly redirectEnd: number;\n    readonly redirectStart: number;\n    readonly requestStart: number;\n    readonly responseEnd: number;\n    readonly responseStart: number;\n    readonly secureConnectionStart: number;\n    readonly transferSize: number;\n    readonly workerStart: number;\n    toJSON(): any;\n}\n\ndeclare var PerformanceResourceTiming: {\n    prototype: PerformanceResourceTiming;\n    new(): PerformanceResourceTiming;\n};\n\ninterface PerformanceTiming {\n    readonly connectEnd: number;\n    readonly connectStart: number;\n    readonly domComplete: number;\n    readonly domContentLoadedEventEnd: number;\n    readonly domContentLoadedEventStart: number;\n    readonly domInteractive: number;\n    readonly domLoading: number;\n    readonly domainLookupEnd: number;\n    readonly domainLookupStart: number;\n    readonly fetchStart: number;\n    readonly loadEventEnd: number;\n    readonly loadEventStart: number;\n    readonly navigationStart: number;\n    readonly redirectEnd: number;\n    readonly redirectStart: number;\n    readonly requestStart: number;\n    readonly responseEnd: number;\n    readonly responseStart: number;\n    readonly secureConnectionStart: number;\n    readonly unloadEventEnd: number;\n    readonly unloadEventStart: number;\n    toJSON(): any;\n}\n\ndeclare var PerformanceTiming: {\n    prototype: PerformanceTiming;\n    new(): PerformanceTiming;\n};\n\ninterface PeriodicWave {\n}\n\ndeclare var PeriodicWave: {\n    prototype: PeriodicWave;\n    new(context: BaseAudioContext, options?: PeriodicWaveOptions): PeriodicWave;\n};\n\ninterface PermissionRequest extends DeferredPermissionRequest {\n    readonly state: MSWebViewPermissionState;\n    defer(): void;\n}\n\ndeclare var PermissionRequest: {\n    prototype: PermissionRequest;\n    new(): PermissionRequest;\n};\n\ninterface PermissionRequestedEvent extends Event {\n    readonly permissionRequest: PermissionRequest;\n}\n\ndeclare var PermissionRequestedEvent: {\n    prototype: PermissionRequestedEvent;\n    new(): PermissionRequestedEvent;\n};\n\ninterface Plugin {\n    readonly description: string;\n    readonly filename: string;\n    readonly length: number;\n    readonly name: string;\n    readonly version: string;\n    item(index: number): MimeType;\n    namedItem(type: string): MimeType;\n    [index: number]: MimeType;\n}\n\ndeclare var Plugin: {\n    prototype: Plugin;\n    new(): Plugin;\n};\n\ninterface PluginArray {\n    readonly length: number;\n    item(index: number): Plugin;\n    namedItem(name: string): Plugin;\n    refresh(reload?: boolean): void;\n    [index: number]: Plugin;\n}\n\ndeclare var PluginArray: {\n    prototype: PluginArray;\n    new(): PluginArray;\n};\n\ninterface PointerEvent extends MouseEvent {\n    readonly height: number;\n    readonly isPrimary: boolean;\n    readonly pointerId: number;\n    readonly pointerType: string;\n    readonly pressure: number;\n    readonly tangentialPressure: number;\n    readonly tiltX: number;\n    readonly tiltY: number;\n    readonly twist: number;\n    readonly width: number;\n}\n\ndeclare var PointerEvent: {\n    prototype: PointerEvent;\n    new(type: string, eventInitDict?: PointerEventInit): PointerEvent;\n};\n\ninterface PopStateEvent extends Event {\n    readonly state: any;\n}\n\ndeclare var PopStateEvent: {\n    prototype: PopStateEvent;\n    new(type: string, eventInitDict?: PopStateEventInit): PopStateEvent;\n};\n\ninterface Position {\n    readonly coords: Coordinates;\n    readonly timestamp: number;\n}\n\ninterface PositionError {\n    readonly code: number;\n    readonly message: string;\n    readonly PERMISSION_DENIED: number;\n    readonly POSITION_UNAVAILABLE: number;\n    readonly TIMEOUT: number;\n}\n\ninterface ProcessingInstruction extends CharacterData {\n    readonly target: string;\n}\n\ndeclare var ProcessingInstruction: {\n    prototype: ProcessingInstruction;\n    new(): ProcessingInstruction;\n};\n\ninterface ProgressEvent extends Event {\n    readonly lengthComputable: boolean;\n    readonly loaded: number;\n    readonly total: number;\n}\n\ndeclare var ProgressEvent: {\n    prototype: ProgressEvent;\n    new(type: string, eventInitDict?: ProgressEventInit): ProgressEvent;\n};\n\ninterface PromiseRejectionEvent extends Event {\n    readonly promise: Promise<any>;\n    readonly reason: any;\n}\n\ndeclare var PromiseRejectionEvent: {\n    prototype: PromiseRejectionEvent;\n    new(type: string, eventInitDict: PromiseRejectionEventInit): PromiseRejectionEvent;\n};\n\ninterface PushManager {\n    getSubscription(): Promise<PushSubscription | null>;\n    permissionState(options?: PushSubscriptionOptionsInit): Promise<PushPermissionState>;\n    subscribe(options?: PushSubscriptionOptionsInit): Promise<PushSubscription>;\n}\n\ndeclare var PushManager: {\n    prototype: PushManager;\n    new(): PushManager;\n    readonly supportedContentEncodings: ReadonlyArray<string>;\n};\n\ninterface PushSubscription {\n    readonly endpoint: string;\n    readonly expirationTime: number | null;\n    readonly options: PushSubscriptionOptions;\n    getKey(name: PushEncryptionKeyName): ArrayBuffer | null;\n    toJSON(): PushSubscriptionJSON;\n    unsubscribe(): Promise<boolean>;\n}\n\ndeclare var PushSubscription: {\n    prototype: PushSubscription;\n    new(): PushSubscription;\n};\n\ninterface PushSubscriptionOptions {\n    readonly applicationServerKey: ArrayBuffer | null;\n    readonly userVisibleOnly: boolean;\n}\n\ndeclare var PushSubscriptionOptions: {\n    prototype: PushSubscriptionOptions;\n    new(): PushSubscriptionOptions;\n};\n\ninterface RTCCertificate {\n    readonly expires: number;\n    getFingerprints(): RTCDtlsFingerprint[];\n}\n\ndeclare var RTCCertificate: {\n    prototype: RTCCertificate;\n    new(): RTCCertificate;\n    getSupportedAlgorithms(): AlgorithmIdentifier[];\n};\n\ninterface RTCDTMFSenderEventMap {\n    "tonechange": RTCDTMFToneChangeEvent;\n}\n\ninterface RTCDTMFSender extends EventTarget {\n    readonly canInsertDTMF: boolean;\n    ontonechange: ((this: RTCDTMFSender, ev: RTCDTMFToneChangeEvent) => any) | null;\n    readonly toneBuffer: string;\n    insertDTMF(tones: string, duration?: number, interToneGap?: number): void;\n    addEventListener<K extends keyof RTCDTMFSenderEventMap>(type: K, listener: (this: RTCDTMFSender, ev: RTCDTMFSenderEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCDTMFSenderEventMap>(type: K, listener: (this: RTCDTMFSender, ev: RTCDTMFSenderEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCDTMFSender: {\n    prototype: RTCDTMFSender;\n    new(): RTCDTMFSender;\n};\n\ninterface RTCDTMFToneChangeEvent extends Event {\n    readonly tone: string;\n}\n\ndeclare var RTCDTMFToneChangeEvent: {\n    prototype: RTCDTMFToneChangeEvent;\n    new(type: string, eventInitDict: RTCDTMFToneChangeEventInit): RTCDTMFToneChangeEvent;\n};\n\ninterface RTCDataChannelEventMap {\n    "bufferedamountlow": Event;\n    "close": Event;\n    "error": RTCErrorEvent;\n    "message": MessageEvent;\n    "open": Event;\n}\n\ninterface RTCDataChannel extends EventTarget {\n    binaryType: string;\n    readonly bufferedAmount: number;\n    bufferedAmountLowThreshold: number;\n    readonly id: number | null;\n    readonly label: string;\n    readonly maxPacketLifeTime: number | null;\n    readonly maxRetransmits: number | null;\n    readonly negotiated: boolean;\n    onbufferedamountlow: ((this: RTCDataChannel, ev: Event) => any) | null;\n    onclose: ((this: RTCDataChannel, ev: Event) => any) | null;\n    onerror: ((this: RTCDataChannel, ev: RTCErrorEvent) => any) | null;\n    onmessage: ((this: RTCDataChannel, ev: MessageEvent) => any) | null;\n    onopen: ((this: RTCDataChannel, ev: Event) => any) | null;\n    readonly ordered: boolean;\n    readonly priority: RTCPriorityType;\n    readonly protocol: string;\n    readonly readyState: RTCDataChannelState;\n    close(): void;\n    send(data: string): void;\n    send(data: Blob): void;\n    send(data: ArrayBuffer): void;\n    send(data: ArrayBufferView): void;\n    addEventListener<K extends keyof RTCDataChannelEventMap>(type: K, listener: (this: RTCDataChannel, ev: RTCDataChannelEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCDataChannelEventMap>(type: K, listener: (this: RTCDataChannel, ev: RTCDataChannelEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCDataChannel: {\n    prototype: RTCDataChannel;\n    new(): RTCDataChannel;\n};\n\ninterface RTCDataChannelEvent extends Event {\n    readonly channel: RTCDataChannel;\n}\n\ndeclare var RTCDataChannelEvent: {\n    prototype: RTCDataChannelEvent;\n    new(type: string, eventInitDict: RTCDataChannelEventInit): RTCDataChannelEvent;\n};\n\ninterface RTCDtlsTransportEventMap {\n    "error": RTCErrorEvent;\n    "statechange": Event;\n}\n\ninterface RTCDtlsTransport extends EventTarget {\n    onerror: ((this: RTCDtlsTransport, ev: RTCErrorEvent) => any) | null;\n    onstatechange: ((this: RTCDtlsTransport, ev: Event) => any) | null;\n    readonly state: RTCDtlsTransportState;\n    readonly transport: RTCIceTransport;\n    getRemoteCertificates(): ArrayBuffer[];\n    addEventListener<K extends keyof RTCDtlsTransportEventMap>(type: K, listener: (this: RTCDtlsTransport, ev: RTCDtlsTransportEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCDtlsTransportEventMap>(type: K, listener: (this: RTCDtlsTransport, ev: RTCDtlsTransportEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCDtlsTransport: {\n    prototype: RTCDtlsTransport;\n    new(): RTCDtlsTransport;\n};\n\ninterface RTCDtlsTransportStateChangedEvent extends Event {\n    readonly state: RTCDtlsTransportState;\n}\n\ndeclare var RTCDtlsTransportStateChangedEvent: {\n    prototype: RTCDtlsTransportStateChangedEvent;\n    new(): RTCDtlsTransportStateChangedEvent;\n};\n\ninterface RTCDtmfSenderEventMap {\n    "tonechange": RTCDTMFToneChangeEvent;\n}\n\ninterface RTCDtmfSender extends EventTarget {\n    readonly canInsertDTMF: boolean;\n    readonly duration: number;\n    readonly interToneGap: number;\n    ontonechange: ((this: RTCDtmfSender, ev: RTCDTMFToneChangeEvent) => any) | null;\n    readonly sender: RTCRtpSender;\n    readonly toneBuffer: string;\n    insertDTMF(tones: string, duration?: number, interToneGap?: number): void;\n    addEventListener<K extends keyof RTCDtmfSenderEventMap>(type: K, listener: (this: RTCDtmfSender, ev: RTCDtmfSenderEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCDtmfSenderEventMap>(type: K, listener: (this: RTCDtmfSender, ev: RTCDtmfSenderEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCDtmfSender: {\n    prototype: RTCDtmfSender;\n    new(sender: RTCRtpSender): RTCDtmfSender;\n};\n\ninterface RTCError extends Error {\n    errorDetail: string;\n    httpRequestStatusCode: number;\n    message: string;\n    name: string;\n    receivedAlert: number | null;\n    sctpCauseCode: number;\n    sdpLineNumber: number;\n    sentAlert: number | null;\n}\n\ndeclare var RTCError: {\n    prototype: RTCError;\n    new(errorDetail?: string, message?: string): RTCError;\n};\n\ninterface RTCErrorEvent extends Event {\n    readonly error: RTCError | null;\n}\n\ndeclare var RTCErrorEvent: {\n    prototype: RTCErrorEvent;\n    new(type: string, eventInitDict: RTCErrorEventInit): RTCErrorEvent;\n};\n\ninterface RTCIceCandidate {\n    readonly candidate: string;\n    readonly component: RTCIceComponent | null;\n    readonly foundation: string | null;\n    readonly ip: string | null;\n    readonly port: number | null;\n    readonly priority: number | null;\n    readonly protocol: RTCIceProtocol | null;\n    readonly relatedAddress: string | null;\n    readonly relatedPort: number | null;\n    readonly sdpMLineIndex: number | null;\n    readonly sdpMid: string | null;\n    readonly tcpType: RTCIceTcpCandidateType | null;\n    readonly type: RTCIceCandidateType | null;\n    readonly usernameFragment: string | null;\n    toJSON(): RTCIceCandidateInit;\n}\n\ndeclare var RTCIceCandidate: {\n    prototype: RTCIceCandidate;\n    new(candidateInitDict?: RTCIceCandidateInit): RTCIceCandidate;\n};\n\ninterface RTCIceCandidatePairChangedEvent extends Event {\n    readonly pair: RTCIceCandidatePair;\n}\n\ndeclare var RTCIceCandidatePairChangedEvent: {\n    prototype: RTCIceCandidatePairChangedEvent;\n    new(): RTCIceCandidatePairChangedEvent;\n};\n\ninterface RTCIceGathererEventMap {\n    "error": Event;\n    "localcandidate": RTCIceGathererEvent;\n}\n\ninterface RTCIceGatherer extends RTCStatsProvider {\n    readonly component: RTCIceComponent;\n    onerror: ((this: RTCIceGatherer, ev: Event) => any) | null;\n    onlocalcandidate: ((this: RTCIceGatherer, ev: RTCIceGathererEvent) => any) | null;\n    createAssociatedGatherer(): RTCIceGatherer;\n    getLocalCandidates(): RTCIceCandidateDictionary[];\n    getLocalParameters(): RTCIceParameters;\n    addEventListener<K extends keyof RTCIceGathererEventMap>(type: K, listener: (this: RTCIceGatherer, ev: RTCIceGathererEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCIceGathererEventMap>(type: K, listener: (this: RTCIceGatherer, ev: RTCIceGathererEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCIceGatherer: {\n    prototype: RTCIceGatherer;\n    new(options: RTCIceGatherOptions): RTCIceGatherer;\n};\n\ninterface RTCIceGathererEvent extends Event {\n    readonly candidate: RTCIceCandidateDictionary | RTCIceCandidateComplete;\n}\n\ndeclare var RTCIceGathererEvent: {\n    prototype: RTCIceGathererEvent;\n    new(): RTCIceGathererEvent;\n};\n\ninterface RTCIceTransportEventMap {\n    "gatheringstatechange": Event;\n    "selectedcandidatepairchange": Event;\n    "statechange": Event;\n}\n\ninterface RTCIceTransport extends EventTarget {\n    readonly component: RTCIceComponent;\n    readonly gatheringState: RTCIceGathererState;\n    ongatheringstatechange: ((this: RTCIceTransport, ev: Event) => any) | null;\n    onselectedcandidatepairchange: ((this: RTCIceTransport, ev: Event) => any) | null;\n    onstatechange: ((this: RTCIceTransport, ev: Event) => any) | null;\n    readonly role: RTCIceRole;\n    readonly state: RTCIceTransportState;\n    getLocalCandidates(): RTCIceCandidate[];\n    getLocalParameters(): RTCIceParameters | null;\n    getRemoteCandidates(): RTCIceCandidate[];\n    getRemoteParameters(): RTCIceParameters | null;\n    getSelectedCandidatePair(): RTCIceCandidatePair | null;\n    addEventListener<K extends keyof RTCIceTransportEventMap>(type: K, listener: (this: RTCIceTransport, ev: RTCIceTransportEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCIceTransportEventMap>(type: K, listener: (this: RTCIceTransport, ev: RTCIceTransportEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCIceTransport: {\n    prototype: RTCIceTransport;\n    new(): RTCIceTransport;\n};\n\ninterface RTCIceTransportStateChangedEvent extends Event {\n    readonly state: RTCIceTransportState;\n}\n\ndeclare var RTCIceTransportStateChangedEvent: {\n    prototype: RTCIceTransportStateChangedEvent;\n    new(): RTCIceTransportStateChangedEvent;\n};\n\ninterface RTCIdentityAssertion {\n    idp: string;\n    name: string;\n}\n\ndeclare var RTCIdentityAssertion: {\n    prototype: RTCIdentityAssertion;\n    new(idp: string, name: string): RTCIdentityAssertion;\n};\n\ninterface RTCPeerConnectionEventMap {\n    "connectionstatechange": Event;\n    "datachannel": RTCDataChannelEvent;\n    "icecandidate": RTCPeerConnectionIceEvent;\n    "icecandidateerror": RTCPeerConnectionIceErrorEvent;\n    "iceconnectionstatechange": Event;\n    "icegatheringstatechange": Event;\n    "negotiationneeded": Event;\n    "signalingstatechange": Event;\n    "statsended": RTCStatsEvent;\n    "track": RTCTrackEvent;\n}\n\ninterface RTCPeerConnection extends EventTarget {\n    readonly canTrickleIceCandidates: boolean | null;\n    readonly connectionState: RTCPeerConnectionState;\n    readonly currentLocalDescription: RTCSessionDescription | null;\n    readonly currentRemoteDescription: RTCSessionDescription | null;\n    readonly iceConnectionState: RTCIceConnectionState;\n    readonly iceGatheringState: RTCIceGatheringState;\n    readonly idpErrorInfo: string | null;\n    readonly idpLoginUrl: string | null;\n    readonly localDescription: RTCSessionDescription | null;\n    onconnectionstatechange: ((this: RTCPeerConnection, ev: Event) => any) | null;\n    ondatachannel: ((this: RTCPeerConnection, ev: RTCDataChannelEvent) => any) | null;\n    onicecandidate: ((this: RTCPeerConnection, ev: RTCPeerConnectionIceEvent) => any) | null;\n    onicecandidateerror: ((this: RTCPeerConnection, ev: RTCPeerConnectionIceErrorEvent) => any) | null;\n    oniceconnectionstatechange: ((this: RTCPeerConnection, ev: Event) => any) | null;\n    onicegatheringstatechange: ((this: RTCPeerConnection, ev: Event) => any) | null;\n    onnegotiationneeded: ((this: RTCPeerConnection, ev: Event) => any) | null;\n    onsignalingstatechange: ((this: RTCPeerConnection, ev: Event) => any) | null;\n    onstatsended: ((this: RTCPeerConnection, ev: RTCStatsEvent) => any) | null;\n    ontrack: ((this: RTCPeerConnection, ev: RTCTrackEvent) => any) | null;\n    readonly peerIdentity: Promise<RTCIdentityAssertion>;\n    readonly pendingLocalDescription: RTCSessionDescription | null;\n    readonly pendingRemoteDescription: RTCSessionDescription | null;\n    readonly remoteDescription: RTCSessionDescription | null;\n    readonly sctp: RTCSctpTransport | null;\n    readonly signalingState: RTCSignalingState;\n    addIceCandidate(candidate: RTCIceCandidateInit | RTCIceCandidate): Promise<void>;\n    addTrack(track: MediaStreamTrack, ...streams: MediaStream[]): RTCRtpSender;\n    addTransceiver(trackOrKind: MediaStreamTrack | string, init?: RTCRtpTransceiverInit): RTCRtpTransceiver;\n    close(): void;\n    createAnswer(options?: RTCOfferOptions): Promise<RTCSessionDescriptionInit>;\n    createDataChannel(label: string, dataChannelDict?: RTCDataChannelInit): RTCDataChannel;\n    createOffer(options?: RTCOfferOptions): Promise<RTCSessionDescriptionInit>;\n    getConfiguration(): RTCConfiguration;\n    getIdentityAssertion(): Promise<string>;\n    getReceivers(): RTCRtpReceiver[];\n    getSenders(): RTCRtpSender[];\n    getStats(selector?: MediaStreamTrack | null): Promise<RTCStatsReport>;\n    getTransceivers(): RTCRtpTransceiver[];\n    removeTrack(sender: RTCRtpSender): void;\n    setConfiguration(configuration: RTCConfiguration): void;\n    setIdentityProvider(provider: string, options?: RTCIdentityProviderOptions): void;\n    setLocalDescription(description: RTCSessionDescriptionInit): Promise<void>;\n    setRemoteDescription(description: RTCSessionDescriptionInit): Promise<void>;\n    addEventListener<K extends keyof RTCPeerConnectionEventMap>(type: K, listener: (this: RTCPeerConnection, ev: RTCPeerConnectionEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCPeerConnectionEventMap>(type: K, listener: (this: RTCPeerConnection, ev: RTCPeerConnectionEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCPeerConnection: {\n    prototype: RTCPeerConnection;\n    new(configuration?: RTCConfiguration): RTCPeerConnection;\n    generateCertificate(keygenAlgorithm: AlgorithmIdentifier): Promise<RTCCertificate>;\n    getDefaultIceServers(): RTCIceServer[];\n};\n\ninterface RTCPeerConnectionIceErrorEvent extends Event {\n    readonly errorCode: number;\n    readonly errorText: string;\n    readonly hostCandidate: string;\n    readonly url: string;\n}\n\ndeclare var RTCPeerConnectionIceErrorEvent: {\n    prototype: RTCPeerConnectionIceErrorEvent;\n    new(type: string, eventInitDict: RTCPeerConnectionIceErrorEventInit): RTCPeerConnectionIceErrorEvent;\n};\n\ninterface RTCPeerConnectionIceEvent extends Event {\n    readonly candidate: RTCIceCandidate | null;\n    readonly url: string | null;\n}\n\ndeclare var RTCPeerConnectionIceEvent: {\n    prototype: RTCPeerConnectionIceEvent;\n    new(type: string, eventInitDict?: RTCPeerConnectionIceEventInit): RTCPeerConnectionIceEvent;\n};\n\ninterface RTCRtpReceiver {\n    readonly rtcpTransport: RTCDtlsTransport | null;\n    readonly track: MediaStreamTrack;\n    readonly transport: RTCDtlsTransport | null;\n    getContributingSources(): RTCRtpContributingSource[];\n    getParameters(): RTCRtpReceiveParameters;\n    getStats(): Promise<RTCStatsReport>;\n    getSynchronizationSources(): RTCRtpSynchronizationSource[];\n}\n\ndeclare var RTCRtpReceiver: {\n    prototype: RTCRtpReceiver;\n    new(): RTCRtpReceiver;\n    getCapabilities(kind: string): RTCRtpCapabilities | null;\n};\n\ninterface RTCRtpSender {\n    readonly dtmf: RTCDTMFSender | null;\n    readonly rtcpTransport: RTCDtlsTransport | null;\n    readonly track: MediaStreamTrack | null;\n    readonly transport: RTCDtlsTransport | null;\n    getParameters(): RTCRtpSendParameters;\n    getStats(): Promise<RTCStatsReport>;\n    replaceTrack(withTrack: MediaStreamTrack | null): Promise<void>;\n    setParameters(parameters: RTCRtpSendParameters): Promise<void>;\n    setStreams(...streams: MediaStream[]): void;\n}\n\ndeclare var RTCRtpSender: {\n    prototype: RTCRtpSender;\n    new(): RTCRtpSender;\n    getCapabilities(kind: string): RTCRtpCapabilities | null;\n};\n\ninterface RTCRtpTransceiver {\n    readonly currentDirection: RTCRtpTransceiverDirection | null;\n    direction: RTCRtpTransceiverDirection;\n    readonly mid: string | null;\n    readonly receiver: RTCRtpReceiver;\n    readonly sender: RTCRtpSender;\n    readonly stopped: boolean;\n    setCodecPreferences(codecs: RTCRtpCodecCapability[]): void;\n    stop(): void;\n}\n\ndeclare var RTCRtpTransceiver: {\n    prototype: RTCRtpTransceiver;\n    new(): RTCRtpTransceiver;\n};\n\ninterface RTCSctpTransportEventMap {\n    "statechange": Event;\n}\n\ninterface RTCSctpTransport {\n    readonly maxChannels: number | null;\n    readonly maxMessageSize: number;\n    onstatechange: ((this: RTCSctpTransport, ev: Event) => any) | null;\n    readonly state: RTCSctpTransportState;\n    readonly transport: RTCDtlsTransport;\n    addEventListener<K extends keyof RTCSctpTransportEventMap>(type: K, listener: (this: RTCSctpTransport, ev: RTCSctpTransportEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCSctpTransportEventMap>(type: K, listener: (this: RTCSctpTransport, ev: RTCSctpTransportEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCSctpTransport: {\n    prototype: RTCSctpTransport;\n    new(): RTCSctpTransport;\n};\n\ninterface RTCSessionDescription {\n    readonly sdp: string;\n    readonly type: RTCSdpType;\n    toJSON(): any;\n}\n\ndeclare var RTCSessionDescription: {\n    prototype: RTCSessionDescription;\n    new(descriptionInitDict: RTCSessionDescriptionInit): RTCSessionDescription;\n};\n\ninterface RTCSrtpSdesTransportEventMap {\n    "error": Event;\n}\n\ninterface RTCSrtpSdesTransport extends EventTarget {\n    onerror: ((this: RTCSrtpSdesTransport, ev: Event) => any) | null;\n    readonly transport: RTCIceTransport;\n    addEventListener<K extends keyof RTCSrtpSdesTransportEventMap>(type: K, listener: (this: RTCSrtpSdesTransport, ev: RTCSrtpSdesTransportEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCSrtpSdesTransportEventMap>(type: K, listener: (this: RTCSrtpSdesTransport, ev: RTCSrtpSdesTransportEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCSrtpSdesTransport: {\n    prototype: RTCSrtpSdesTransport;\n    new(transport: RTCIceTransport, encryptParameters: RTCSrtpSdesParameters, decryptParameters: RTCSrtpSdesParameters): RTCSrtpSdesTransport;\n    getLocalParameters(): RTCSrtpSdesParameters[];\n};\n\ninterface RTCSsrcConflictEvent extends Event {\n    readonly ssrc: number;\n}\n\ndeclare var RTCSsrcConflictEvent: {\n    prototype: RTCSsrcConflictEvent;\n    new(): RTCSsrcConflictEvent;\n};\n\ninterface RTCStatsEvent extends Event {\n    readonly report: RTCStatsReport;\n}\n\ndeclare var RTCStatsEvent: {\n    prototype: RTCStatsEvent;\n    new(type: string, eventInitDict: RTCStatsEventInit): RTCStatsEvent;\n};\n\ninterface RTCStatsProvider extends EventTarget {\n    getStats(): Promise<RTCStatsReport>;\n    msGetStats(): Promise<RTCStatsReport>;\n}\n\ndeclare var RTCStatsProvider: {\n    prototype: RTCStatsProvider;\n    new(): RTCStatsProvider;\n};\n\ninterface RTCStatsReport {\n    forEach(callbackfn: (value: any, key: string, parent: RTCStatsReport) => void, thisArg?: any): void;\n}\n\ndeclare var RTCStatsReport: {\n    prototype: RTCStatsReport;\n    new(): RTCStatsReport;\n};\n\ninterface RTCTrackEvent extends Event {\n    readonly receiver: RTCRtpReceiver;\n    readonly streams: ReadonlyArray<MediaStream>;\n    readonly track: MediaStreamTrack;\n    readonly transceiver: RTCRtpTransceiver;\n}\n\ndeclare var RTCTrackEvent: {\n    prototype: RTCTrackEvent;\n    new(type: string, eventInitDict: RTCTrackEventInit): RTCTrackEvent;\n};\n\ninterface RadioNodeList extends NodeList {\n    value: string;\n}\n\ndeclare var RadioNodeList: {\n    prototype: RadioNodeList;\n    new(): RadioNodeList;\n};\n\ninterface RandomSource {\n    getRandomValues<T extends Int8Array | Uint8ClampedArray | Uint8Array | Int16Array | Uint16Array | Int32Array | Uint32Array>(array: T): T;\n}\n\ndeclare var RandomSource: {\n    prototype: RandomSource;\n    new(): RandomSource;\n};\n\ninterface Range extends AbstractRange {\n    /**\n     * Returns the node, furthest away from\n     * the document, that is an ancestor of both range\'s start node and end node.\n     */\n    readonly commonAncestorContainer: Node;\n    cloneContents(): DocumentFragment;\n    cloneRange(): Range;\n    collapse(toStart?: boolean): void;\n    compareBoundaryPoints(how: number, sourceRange: Range): number;\n    /**\n     * Returns −1 if the point is before the range, 0 if the point is\n     * in the range, and 1 if the point is after the range.\n     */\n    comparePoint(node: Node, offset: number): number;\n    createContextualFragment(fragment: string): DocumentFragment;\n    deleteContents(): void;\n    detach(): void;\n    extractContents(): DocumentFragment;\n    getBoundingClientRect(): ClientRect | DOMRect;\n    getClientRects(): ClientRectList | DOMRectList;\n    insertNode(node: Node): void;\n    /**\n     * Returns whether range intersects node.\n     */\n    intersectsNode(node: Node): boolean;\n    isPointInRange(node: Node, offset: number): boolean;\n    selectNode(node: Node): void;\n    selectNodeContents(node: Node): void;\n    setEnd(node: Node, offset: number): void;\n    setEndAfter(node: Node): void;\n    setEndBefore(node: Node): void;\n    setStart(node: Node, offset: number): void;\n    setStartAfter(node: Node): void;\n    setStartBefore(node: Node): void;\n    surroundContents(newParent: Node): void;\n    readonly END_TO_END: number;\n    readonly END_TO_START: number;\n    readonly START_TO_END: number;\n    readonly START_TO_START: number;\n}\n\ndeclare var Range: {\n    prototype: Range;\n    new(): Range;\n    readonly END_TO_END: number;\n    readonly END_TO_START: number;\n    readonly START_TO_END: number;\n    readonly START_TO_START: number;\n};\n\ninterface ReadableByteStreamController {\n    readonly byobRequest: ReadableStreamBYOBRequest | undefined;\n    readonly desiredSize: number | null;\n    close(): void;\n    enqueue(chunk: ArrayBufferView): void;\n    error(error?: any): void;\n}\n\ninterface ReadableStream<R = any> {\n    readonly locked: boolean;\n    cancel(reason?: any): Promise<void>;\n    getReader(options: { mode: "byob" }): ReadableStreamBYOBReader;\n    getReader(): ReadableStreamDefaultReader<R>;\n    pipeThrough<T>({ writable, readable }: { writable: WritableStream<R>, readable: ReadableStream<T> }, options?: PipeOptions): ReadableStream<T>;\n    pipeTo(dest: WritableStream<R>, options?: PipeOptions): Promise<void>;\n    tee(): [ReadableStream<R>, ReadableStream<R>];\n}\n\ndeclare var ReadableStream: {\n    prototype: ReadableStream;\n    new(underlyingSource: UnderlyingByteSource, strategy?: { highWaterMark?: number, size?: undefined }): ReadableStream<Uint8Array>;\n    new<R = any>(underlyingSource?: UnderlyingSource<R>, strategy?: QueuingStrategy<R>): ReadableStream<R>;\n};\n\ninterface ReadableStreamBYOBReader {\n    readonly closed: Promise<void>;\n    cancel(reason?: any): Promise<void>;\n    read<T extends ArrayBufferView>(view: T): Promise<ReadableStreamReadResult<T>>;\n    releaseLock(): void;\n}\n\ndeclare var ReadableStreamBYOBReader: {\n    prototype: ReadableStreamBYOBReader;\n    new(stream: ReadableStream<Uint8Array>): ReadableStreamBYOBReader;\n};\n\ninterface ReadableStreamBYOBRequest {\n    readonly view: ArrayBufferView;\n    respond(bytesWritten: number): void;\n    respondWithNewView(view: ArrayBufferView): void;\n}\n\ninterface ReadableStreamDefaultController<R = any> {\n    readonly desiredSize: number | null;\n    close(): void;\n    enqueue(chunk: R): void;\n    error(error?: any): void;\n}\n\ninterface ReadableStreamDefaultReader<R = any> {\n    readonly closed: Promise<void>;\n    cancel(reason?: any): Promise<void>;\n    read(): Promise<ReadableStreamReadResult<R>>;\n    releaseLock(): void;\n}\n\ninterface ReadableStreamReadResult<T> {\n    done: boolean;\n    value: T;\n}\n\ninterface ReadableStreamReader<R = any> {\n    cancel(): Promise<void>;\n    read(): Promise<ReadableStreamReadResult<R>>;\n    releaseLock(): void;\n}\n\ndeclare var ReadableStreamReader: {\n    prototype: ReadableStreamReader;\n    new(): ReadableStreamReader;\n};\n\ninterface Request extends Body {\n    /**\n     * Returns the cache mode associated with request, which is a string indicating\n     * how the request will interact with the browser\'s cache when fetching.\n     */\n    readonly cache: RequestCache;\n    /**\n     * Returns the credentials mode associated with request, which is a string\n     * indicating whether credentials will be sent with the request always, never, or only when sent to a\n     * same-origin URL.\n     */\n    readonly credentials: RequestCredentials;\n    /**\n     * Returns the kind of resource requested by request, e.g., "document" or\n     * "script".\n     */\n    readonly destination: RequestDestination;\n    /**\n     * Returns a Headers object consisting of the headers associated with request.\n     * Note that headers added in the network layer by the user agent will not be accounted for in this\n     * object, e.g., the "Host" header.\n     */\n    readonly headers: Headers;\n    /**\n     * Returns request\'s subresource integrity metadata, which is a cryptographic hash of\n     * the resource being fetched. Its value consists of multiple hashes separated by whitespace. [SRI]\n     */\n    readonly integrity: string;\n    /**\n     * Returns a boolean indicating whether or not request is for a history\n     * navigation (a.k.a. back-foward navigation).\n     */\n    readonly isHistoryNavigation: boolean;\n    /**\n     * Returns a boolean indicating whether or not request is for a reload navigation.\n     */\n    readonly isReloadNavigation: boolean;\n    /**\n     * Returns a boolean indicating whether or not request can outlive the global in which\n     * it was created.\n     */\n    readonly keepalive: boolean;\n    /**\n     * Returns request\'s HTTP method, which is "GET" by default.\n     */\n    readonly method: string;\n    /**\n     * Returns the mode associated with request, which is a string indicating\n     * whether the request will use CORS, or will be restricted to same-origin URLs.\n     */\n    readonly mode: RequestMode;\n    /**\n     * Returns the redirect mode associated with request, which is a string\n     * indicating how redirects for the request will be handled during fetching. A request will follow redirects by default.\n     */\n    readonly redirect: RequestRedirect;\n    /**\n     * Returns the referrer of request. Its value can be a same-origin URL if\n     * explicitly set in init, the empty string to indicate no referrer, and\n     * "about:client" when defaulting to the global\'s default. This is used during\n     * fetching to determine the value of the `Referer` header of the request being made.\n     */\n    readonly referrer: string;\n    /**\n     * Returns the referrer policy associated with request. This is used during\n     * fetching to compute the value of the request\'s referrer.\n     */\n    readonly referrerPolicy: ReferrerPolicy;\n    /**\n     * Returns the signal associated with request, which is an AbortSignal object indicating whether or not request has been aborted, and its abort\n     * event handler.\n     */\n    readonly signal: AbortSignal;\n    /**\n     * Returns the URL of request as a string.\n     */\n    readonly url: string;\n    clone(): Request;\n}\n\ndeclare var Request: {\n    prototype: Request;\n    new(input: RequestInfo, init?: RequestInit): Request;\n};\n\ninterface Response extends Body {\n    readonly headers: Headers;\n    readonly ok: boolean;\n    readonly redirected: boolean;\n    readonly status: number;\n    readonly statusText: string;\n    readonly trailer: Promise<Headers>;\n    readonly type: ResponseType;\n    readonly url: string;\n    clone(): Response;\n}\n\ndeclare var Response: {\n    prototype: Response;\n    new(body?: BodyInit | null, init?: ResponseInit): Response;\n    error(): Response;\n    redirect(url: string, status?: number): Response;\n};\n\ninterface SVGAElement extends SVGGraphicsElement, SVGURIReference {\n    readonly target: SVGAnimatedString;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGAElement: {\n    prototype: SVGAElement;\n    new(): SVGAElement;\n};\n\ninterface SVGAngle {\n    readonly unitType: number;\n    value: number;\n    valueAsString: string;\n    valueInSpecifiedUnits: number;\n    convertToSpecifiedUnits(unitType: number): void;\n    newValueSpecifiedUnits(unitType: number, valueInSpecifiedUnits: number): void;\n    readonly SVG_ANGLETYPE_DEG: number;\n    readonly SVG_ANGLETYPE_GRAD: number;\n    readonly SVG_ANGLETYPE_RAD: number;\n    readonly SVG_ANGLETYPE_UNKNOWN: number;\n    readonly SVG_ANGLETYPE_UNSPECIFIED: number;\n}\n\ndeclare var SVGAngle: {\n    prototype: SVGAngle;\n    new(): SVGAngle;\n    readonly SVG_ANGLETYPE_DEG: number;\n    readonly SVG_ANGLETYPE_GRAD: number;\n    readonly SVG_ANGLETYPE_RAD: number;\n    readonly SVG_ANGLETYPE_UNKNOWN: number;\n    readonly SVG_ANGLETYPE_UNSPECIFIED: number;\n};\n\ninterface SVGAnimateElement extends SVGAnimationElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGAnimateElement: {\n    prototype: SVGAnimateElement;\n    new(): SVGAnimateElement;\n};\n\ninterface SVGAnimateMotionElement extends SVGAnimationElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateMotionElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateMotionElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGAnimateMotionElement: {\n    prototype: SVGAnimateMotionElement;\n    new(): SVGAnimateMotionElement;\n};\n\ninterface SVGAnimateTransformElement extends SVGAnimationElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateTransformElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateTransformElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGAnimateTransformElement: {\n    prototype: SVGAnimateTransformElement;\n    new(): SVGAnimateTransformElement;\n};\n\ninterface SVGAnimatedAngle {\n    readonly animVal: SVGAngle;\n    readonly baseVal: SVGAngle;\n}\n\ndeclare var SVGAnimatedAngle: {\n    prototype: SVGAnimatedAngle;\n    new(): SVGAnimatedAngle;\n};\n\ninterface SVGAnimatedBoolean {\n    readonly animVal: boolean;\n    baseVal: boolean;\n}\n\ndeclare var SVGAnimatedBoolean: {\n    prototype: SVGAnimatedBoolean;\n    new(): SVGAnimatedBoolean;\n};\n\ninterface SVGAnimatedEnumeration {\n    readonly animVal: number;\n    baseVal: number;\n}\n\ndeclare var SVGAnimatedEnumeration: {\n    prototype: SVGAnimatedEnumeration;\n    new(): SVGAnimatedEnumeration;\n};\n\ninterface SVGAnimatedInteger {\n    readonly animVal: number;\n    baseVal: number;\n}\n\ndeclare var SVGAnimatedInteger: {\n    prototype: SVGAnimatedInteger;\n    new(): SVGAnimatedInteger;\n};\n\ninterface SVGAnimatedLength {\n    readonly animVal: SVGLength;\n    readonly baseVal: SVGLength;\n}\n\ndeclare var SVGAnimatedLength: {\n    prototype: SVGAnimatedLength;\n    new(): SVGAnimatedLength;\n};\n\ninterface SVGAnimatedLengthList {\n    readonly animVal: SVGLengthList;\n    readonly baseVal: SVGLengthList;\n}\n\ndeclare var SVGAnimatedLengthList: {\n    prototype: SVGAnimatedLengthList;\n    new(): SVGAnimatedLengthList;\n};\n\ninterface SVGAnimatedNumber {\n    readonly animVal: number;\n    baseVal: number;\n}\n\ndeclare var SVGAnimatedNumber: {\n    prototype: SVGAnimatedNumber;\n    new(): SVGAnimatedNumber;\n};\n\ninterface SVGAnimatedNumberList {\n    readonly animVal: SVGNumberList;\n    readonly baseVal: SVGNumberList;\n}\n\ndeclare var SVGAnimatedNumberList: {\n    prototype: SVGAnimatedNumberList;\n    new(): SVGAnimatedNumberList;\n};\n\ninterface SVGAnimatedPoints {\n    readonly animatedPoints: SVGPointList;\n    readonly points: SVGPointList;\n}\n\ninterface SVGAnimatedPreserveAspectRatio {\n    readonly animVal: SVGPreserveAspectRatio;\n    readonly baseVal: SVGPreserveAspectRatio;\n}\n\ndeclare var SVGAnimatedPreserveAspectRatio: {\n    prototype: SVGAnimatedPreserveAspectRatio;\n    new(): SVGAnimatedPreserveAspectRatio;\n};\n\ninterface SVGAnimatedRect {\n    readonly animVal: DOMRectReadOnly;\n    readonly baseVal: DOMRect;\n}\n\ndeclare var SVGAnimatedRect: {\n    prototype: SVGAnimatedRect;\n    new(): SVGAnimatedRect;\n};\n\ninterface SVGAnimatedString {\n    readonly animVal: string;\n    baseVal: string;\n}\n\ndeclare var SVGAnimatedString: {\n    prototype: SVGAnimatedString;\n    new(): SVGAnimatedString;\n};\n\ninterface SVGAnimatedTransformList {\n    readonly animVal: SVGTransformList;\n    readonly baseVal: SVGTransformList;\n}\n\ndeclare var SVGAnimatedTransformList: {\n    prototype: SVGAnimatedTransformList;\n    new(): SVGAnimatedTransformList;\n};\n\ninterface SVGAnimationElement extends SVGElement {\n    readonly targetElement: SVGElement;\n    getCurrentTime(): number;\n    getSimpleDuration(): number;\n    getStartTime(): number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimationElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimationElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGAnimationElement: {\n    prototype: SVGAnimationElement;\n    new(): SVGAnimationElement;\n};\n\ninterface SVGCircleElement extends SVGGraphicsElement {\n    readonly cx: SVGAnimatedLength;\n    readonly cy: SVGAnimatedLength;\n    readonly r: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGCircleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGCircleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGCircleElement: {\n    prototype: SVGCircleElement;\n    new(): SVGCircleElement;\n};\n\ninterface SVGClipPathElement extends SVGGraphicsElement {\n    readonly clipPathUnits: SVGAnimatedEnumeration;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGClipPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGClipPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGClipPathElement: {\n    prototype: SVGClipPathElement;\n    new(): SVGClipPathElement;\n};\n\ninterface SVGComponentTransferFunctionElement extends SVGElement {\n    readonly amplitude: SVGAnimatedNumber;\n    readonly exponent: SVGAnimatedNumber;\n    readonly intercept: SVGAnimatedNumber;\n    readonly offset: SVGAnimatedNumber;\n    readonly slope: SVGAnimatedNumber;\n    readonly tableValues: SVGAnimatedNumberList;\n    readonly type: SVGAnimatedEnumeration;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_DISCRETE: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_GAMMA: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_IDENTITY: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_LINEAR: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_TABLE: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGComponentTransferFunctionElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGComponentTransferFunctionElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGComponentTransferFunctionElement: {\n    prototype: SVGComponentTransferFunctionElement;\n    new(): SVGComponentTransferFunctionElement;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_DISCRETE: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_GAMMA: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_IDENTITY: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_LINEAR: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_TABLE: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_UNKNOWN: number;\n};\n\ninterface SVGCursorElement extends SVGElement {\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGCursorElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGCursorElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGCursorElement: {\n    prototype: SVGCursorElement;\n    new(): SVGCursorElement;\n};\n\ninterface SVGDefsElement extends SVGGraphicsElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGDefsElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGDefsElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGDefsElement: {\n    prototype: SVGDefsElement;\n    new(): SVGDefsElement;\n};\n\ninterface SVGDescElement extends SVGElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGDescElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGDescElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGDescElement: {\n    prototype: SVGDescElement;\n    new(): SVGDescElement;\n};\n\ninterface SVGElementEventMap extends ElementEventMap, GlobalEventHandlersEventMap, DocumentAndElementEventHandlersEventMap {\n}\n\ninterface SVGElement extends Element, GlobalEventHandlers, DocumentAndElementEventHandlers, SVGElementInstance, HTMLOrSVGElement, ElementCSSInlineStyle {\n    /** @deprecated */\n    readonly className: any;\n    readonly ownerSVGElement: SVGSVGElement | null;\n    readonly viewportElement: SVGElement | null;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGElement: {\n    prototype: SVGElement;\n    new(): SVGElement;\n};\n\ninterface SVGElementInstance extends EventTarget {\n    readonly correspondingElement: SVGElement;\n    readonly correspondingUseElement: SVGUseElement;\n}\n\ndeclare var SVGElementInstance: {\n    prototype: SVGElementInstance;\n    new(): SVGElementInstance;\n};\n\ninterface SVGElementInstanceList {\n    /** @deprecated */\n    readonly length: number;\n    /** @deprecated */\n    item(index: number): SVGElementInstance;\n}\n\ndeclare var SVGElementInstanceList: {\n    prototype: SVGElementInstanceList;\n    new(): SVGElementInstanceList;\n};\n\ninterface SVGEllipseElement extends SVGGraphicsElement {\n    readonly cx: SVGAnimatedLength;\n    readonly cy: SVGAnimatedLength;\n    readonly rx: SVGAnimatedLength;\n    readonly ry: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGEllipseElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGEllipseElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGEllipseElement: {\n    prototype: SVGEllipseElement;\n    new(): SVGEllipseElement;\n};\n\ninterface SVGFEBlendElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly in2: SVGAnimatedString;\n    readonly mode: SVGAnimatedEnumeration;\n    readonly SVG_FEBLEND_MODE_COLOR: number;\n    readonly SVG_FEBLEND_MODE_COLOR_BURN: number;\n    readonly SVG_FEBLEND_MODE_COLOR_DODGE: number;\n    readonly SVG_FEBLEND_MODE_DARKEN: number;\n    readonly SVG_FEBLEND_MODE_DIFFERENCE: number;\n    readonly SVG_FEBLEND_MODE_EXCLUSION: number;\n    readonly SVG_FEBLEND_MODE_HARD_LIGHT: number;\n    readonly SVG_FEBLEND_MODE_HUE: number;\n    readonly SVG_FEBLEND_MODE_LIGHTEN: number;\n    readonly SVG_FEBLEND_MODE_LUMINOSITY: number;\n    readonly SVG_FEBLEND_MODE_MULTIPLY: number;\n    readonly SVG_FEBLEND_MODE_NORMAL: number;\n    readonly SVG_FEBLEND_MODE_OVERLAY: number;\n    readonly SVG_FEBLEND_MODE_SATURATION: number;\n    readonly SVG_FEBLEND_MODE_SCREEN: number;\n    readonly SVG_FEBLEND_MODE_SOFT_LIGHT: number;\n    readonly SVG_FEBLEND_MODE_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEBlendElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEBlendElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEBlendElement: {\n    prototype: SVGFEBlendElement;\n    new(): SVGFEBlendElement;\n    readonly SVG_FEBLEND_MODE_COLOR: number;\n    readonly SVG_FEBLEND_MODE_COLOR_BURN: number;\n    readonly SVG_FEBLEND_MODE_COLOR_DODGE: number;\n    readonly SVG_FEBLEND_MODE_DARKEN: number;\n    readonly SVG_FEBLEND_MODE_DIFFERENCE: number;\n    readonly SVG_FEBLEND_MODE_EXCLUSION: number;\n    readonly SVG_FEBLEND_MODE_HARD_LIGHT: number;\n    readonly SVG_FEBLEND_MODE_HUE: number;\n    readonly SVG_FEBLEND_MODE_LIGHTEN: number;\n    readonly SVG_FEBLEND_MODE_LUMINOSITY: number;\n    readonly SVG_FEBLEND_MODE_MULTIPLY: number;\n    readonly SVG_FEBLEND_MODE_NORMAL: number;\n    readonly SVG_FEBLEND_MODE_OVERLAY: number;\n    readonly SVG_FEBLEND_MODE_SATURATION: number;\n    readonly SVG_FEBLEND_MODE_SCREEN: number;\n    readonly SVG_FEBLEND_MODE_SOFT_LIGHT: number;\n    readonly SVG_FEBLEND_MODE_UNKNOWN: number;\n};\n\ninterface SVGFEColorMatrixElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly type: SVGAnimatedEnumeration;\n    readonly values: SVGAnimatedNumberList;\n    readonly SVG_FECOLORMATRIX_TYPE_HUEROTATE: number;\n    readonly SVG_FECOLORMATRIX_TYPE_LUMINANCETOALPHA: number;\n    readonly SVG_FECOLORMATRIX_TYPE_MATRIX: number;\n    readonly SVG_FECOLORMATRIX_TYPE_SATURATE: number;\n    readonly SVG_FECOLORMATRIX_TYPE_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEColorMatrixElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEColorMatrixElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEColorMatrixElement: {\n    prototype: SVGFEColorMatrixElement;\n    new(): SVGFEColorMatrixElement;\n    readonly SVG_FECOLORMATRIX_TYPE_HUEROTATE: number;\n    readonly SVG_FECOLORMATRIX_TYPE_LUMINANCETOALPHA: number;\n    readonly SVG_FECOLORMATRIX_TYPE_MATRIX: number;\n    readonly SVG_FECOLORMATRIX_TYPE_SATURATE: number;\n    readonly SVG_FECOLORMATRIX_TYPE_UNKNOWN: number;\n};\n\ninterface SVGFEComponentTransferElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEComponentTransferElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEComponentTransferElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEComponentTransferElement: {\n    prototype: SVGFEComponentTransferElement;\n    new(): SVGFEComponentTransferElement;\n};\n\ninterface SVGFECompositeElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly in2: SVGAnimatedString;\n    readonly k1: SVGAnimatedNumber;\n    readonly k2: SVGAnimatedNumber;\n    readonly k3: SVGAnimatedNumber;\n    readonly k4: SVGAnimatedNumber;\n    readonly operator: SVGAnimatedEnumeration;\n    readonly SVG_FECOMPOSITE_OPERATOR_ARITHMETIC: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_ATOP: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_IN: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_OUT: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_OVER: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_UNKNOWN: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_XOR: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFECompositeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFECompositeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFECompositeElement: {\n    prototype: SVGFECompositeElement;\n    new(): SVGFECompositeElement;\n    readonly SVG_FECOMPOSITE_OPERATOR_ARITHMETIC: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_ATOP: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_IN: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_OUT: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_OVER: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_UNKNOWN: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_XOR: number;\n};\n\ninterface SVGFEConvolveMatrixElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly bias: SVGAnimatedNumber;\n    readonly divisor: SVGAnimatedNumber;\n    readonly edgeMode: SVGAnimatedEnumeration;\n    readonly in1: SVGAnimatedString;\n    readonly kernelMatrix: SVGAnimatedNumberList;\n    readonly kernelUnitLengthX: SVGAnimatedNumber;\n    readonly kernelUnitLengthY: SVGAnimatedNumber;\n    readonly orderX: SVGAnimatedInteger;\n    readonly orderY: SVGAnimatedInteger;\n    readonly preserveAlpha: SVGAnimatedBoolean;\n    readonly targetX: SVGAnimatedInteger;\n    readonly targetY: SVGAnimatedInteger;\n    readonly SVG_EDGEMODE_DUPLICATE: number;\n    readonly SVG_EDGEMODE_NONE: number;\n    readonly SVG_EDGEMODE_UNKNOWN: number;\n    readonly SVG_EDGEMODE_WRAP: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEConvolveMatrixElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEConvolveMatrixElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEConvolveMatrixElement: {\n    prototype: SVGFEConvolveMatrixElement;\n    new(): SVGFEConvolveMatrixElement;\n    readonly SVG_EDGEMODE_DUPLICATE: number;\n    readonly SVG_EDGEMODE_NONE: number;\n    readonly SVG_EDGEMODE_UNKNOWN: number;\n    readonly SVG_EDGEMODE_WRAP: number;\n};\n\ninterface SVGFEDiffuseLightingElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly diffuseConstant: SVGAnimatedNumber;\n    readonly in1: SVGAnimatedString;\n    readonly kernelUnitLengthX: SVGAnimatedNumber;\n    readonly kernelUnitLengthY: SVGAnimatedNumber;\n    readonly surfaceScale: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDiffuseLightingElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDiffuseLightingElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEDiffuseLightingElement: {\n    prototype: SVGFEDiffuseLightingElement;\n    new(): SVGFEDiffuseLightingElement;\n};\n\ninterface SVGFEDisplacementMapElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly in2: SVGAnimatedString;\n    readonly scale: SVGAnimatedNumber;\n    readonly xChannelSelector: SVGAnimatedEnumeration;\n    readonly yChannelSelector: SVGAnimatedEnumeration;\n    readonly SVG_CHANNEL_A: number;\n    readonly SVG_CHANNEL_B: number;\n    readonly SVG_CHANNEL_G: number;\n    readonly SVG_CHANNEL_R: number;\n    readonly SVG_CHANNEL_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDisplacementMapElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDisplacementMapElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEDisplacementMapElement: {\n    prototype: SVGFEDisplacementMapElement;\n    new(): SVGFEDisplacementMapElement;\n    readonly SVG_CHANNEL_A: number;\n    readonly SVG_CHANNEL_B: number;\n    readonly SVG_CHANNEL_G: number;\n    readonly SVG_CHANNEL_R: number;\n    readonly SVG_CHANNEL_UNKNOWN: number;\n};\n\ninterface SVGFEDistantLightElement extends SVGElement {\n    readonly azimuth: SVGAnimatedNumber;\n    readonly elevation: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDistantLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDistantLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEDistantLightElement: {\n    prototype: SVGFEDistantLightElement;\n    new(): SVGFEDistantLightElement;\n};\n\ninterface SVGFEFloodElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFloodElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFloodElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEFloodElement: {\n    prototype: SVGFEFloodElement;\n    new(): SVGFEFloodElement;\n};\n\ninterface SVGFEFuncAElement extends SVGComponentTransferFunctionElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncAElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncAElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEFuncAElement: {\n    prototype: SVGFEFuncAElement;\n    new(): SVGFEFuncAElement;\n};\n\ninterface SVGFEFuncBElement extends SVGComponentTransferFunctionElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncBElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncBElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEFuncBElement: {\n    prototype: SVGFEFuncBElement;\n    new(): SVGFEFuncBElement;\n};\n\ninterface SVGFEFuncGElement extends SVGComponentTransferFunctionElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEFuncGElement: {\n    prototype: SVGFEFuncGElement;\n    new(): SVGFEFuncGElement;\n};\n\ninterface SVGFEFuncRElement extends SVGComponentTransferFunctionElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncRElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncRElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEFuncRElement: {\n    prototype: SVGFEFuncRElement;\n    new(): SVGFEFuncRElement;\n};\n\ninterface SVGFEGaussianBlurElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly stdDeviationX: SVGAnimatedNumber;\n    readonly stdDeviationY: SVGAnimatedNumber;\n    setStdDeviation(stdDeviationX: number, stdDeviationY: number): void;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEGaussianBlurElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEGaussianBlurElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEGaussianBlurElement: {\n    prototype: SVGFEGaussianBlurElement;\n    new(): SVGFEGaussianBlurElement;\n};\n\ninterface SVGFEImageElement extends SVGElement, SVGFilterPrimitiveStandardAttributes, SVGURIReference {\n    readonly preserveAspectRatio: SVGAnimatedPreserveAspectRatio;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEImageElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEImageElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEImageElement: {\n    prototype: SVGFEImageElement;\n    new(): SVGFEImageElement;\n};\n\ninterface SVGFEMergeElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMergeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMergeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEMergeElement: {\n    prototype: SVGFEMergeElement;\n    new(): SVGFEMergeElement;\n};\n\ninterface SVGFEMergeNodeElement extends SVGElement {\n    readonly in1: SVGAnimatedString;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMergeNodeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMergeNodeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEMergeNodeElement: {\n    prototype: SVGFEMergeNodeElement;\n    new(): SVGFEMergeNodeElement;\n};\n\ninterface SVGFEMorphologyElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly operator: SVGAnimatedEnumeration;\n    readonly radiusX: SVGAnimatedNumber;\n    readonly radiusY: SVGAnimatedNumber;\n    readonly SVG_MORPHOLOGY_OPERATOR_DILATE: number;\n    readonly SVG_MORPHOLOGY_OPERATOR_ERODE: number;\n    readonly SVG_MORPHOLOGY_OPERATOR_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMorphologyElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMorphologyElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEMorphologyElement: {\n    prototype: SVGFEMorphologyElement;\n    new(): SVGFEMorphologyElement;\n    readonly SVG_MORPHOLOGY_OPERATOR_DILATE: number;\n    readonly SVG_MORPHOLOGY_OPERATOR_ERODE: number;\n    readonly SVG_MORPHOLOGY_OPERATOR_UNKNOWN: number;\n};\n\ninterface SVGFEOffsetElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly dx: SVGAnimatedNumber;\n    readonly dy: SVGAnimatedNumber;\n    readonly in1: SVGAnimatedString;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEOffsetElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEOffsetElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEOffsetElement: {\n    prototype: SVGFEOffsetElement;\n    new(): SVGFEOffsetElement;\n};\n\ninterface SVGFEPointLightElement extends SVGElement {\n    readonly x: SVGAnimatedNumber;\n    readonly y: SVGAnimatedNumber;\n    readonly z: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEPointLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEPointLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEPointLightElement: {\n    prototype: SVGFEPointLightElement;\n    new(): SVGFEPointLightElement;\n};\n\ninterface SVGFESpecularLightingElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly kernelUnitLengthX: SVGAnimatedNumber;\n    readonly kernelUnitLengthY: SVGAnimatedNumber;\n    readonly specularConstant: SVGAnimatedNumber;\n    readonly specularExponent: SVGAnimatedNumber;\n    readonly surfaceScale: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFESpecularLightingElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFESpecularLightingElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFESpecularLightingElement: {\n    prototype: SVGFESpecularLightingElement;\n    new(): SVGFESpecularLightingElement;\n};\n\ninterface SVGFESpotLightElement extends SVGElement {\n    readonly limitingConeAngle: SVGAnimatedNumber;\n    readonly pointsAtX: SVGAnimatedNumber;\n    readonly pointsAtY: SVGAnimatedNumber;\n    readonly pointsAtZ: SVGAnimatedNumber;\n    readonly specularExponent: SVGAnimatedNumber;\n    readonly x: SVGAnimatedNumber;\n    readonly y: SVGAnimatedNumber;\n    readonly z: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFESpotLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFESpotLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFESpotLightElement: {\n    prototype: SVGFESpotLightElement;\n    new(): SVGFESpotLightElement;\n};\n\ninterface SVGFETileElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFETileElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFETileElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFETileElement: {\n    prototype: SVGFETileElement;\n    new(): SVGFETileElement;\n};\n\ninterface SVGFETurbulenceElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly baseFrequencyX: SVGAnimatedNumber;\n    readonly baseFrequencyY: SVGAnimatedNumber;\n    readonly numOctaves: SVGAnimatedInteger;\n    readonly seed: SVGAnimatedNumber;\n    readonly stitchTiles: SVGAnimatedEnumeration;\n    readonly type: SVGAnimatedEnumeration;\n    readonly SVG_STITCHTYPE_NOSTITCH: number;\n    readonly SVG_STITCHTYPE_STITCH: number;\n    readonly SVG_STITCHTYPE_UNKNOWN: number;\n    readonly SVG_TURBULENCE_TYPE_FRACTALNOISE: number;\n    readonly SVG_TURBULENCE_TYPE_TURBULENCE: number;\n    readonly SVG_TURBULENCE_TYPE_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFETurbulenceElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFETurbulenceElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFETurbulenceElement: {\n    prototype: SVGFETurbulenceElement;\n    new(): SVGFETurbulenceElement;\n    readonly SVG_STITCHTYPE_NOSTITCH: number;\n    readonly SVG_STITCHTYPE_STITCH: number;\n    readonly SVG_STITCHTYPE_UNKNOWN: number;\n    readonly SVG_TURBULENCE_TYPE_FRACTALNOISE: number;\n    readonly SVG_TURBULENCE_TYPE_TURBULENCE: number;\n    readonly SVG_TURBULENCE_TYPE_UNKNOWN: number;\n};\n\ninterface SVGFilterElement extends SVGElement, SVGURIReference {\n    /** @deprecated */\n    readonly filterResX: SVGAnimatedInteger;\n    /** @deprecated */\n    readonly filterResY: SVGAnimatedInteger;\n    readonly filterUnits: SVGAnimatedEnumeration;\n    readonly height: SVGAnimatedLength;\n    readonly primitiveUnits: SVGAnimatedEnumeration;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    /** @deprecated */\n    setFilterRes(filterResX: number, filterResY: number): void;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFilterElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFilterElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFilterElement: {\n    prototype: SVGFilterElement;\n    new(): SVGFilterElement;\n};\n\ninterface SVGFilterPrimitiveStandardAttributes {\n    readonly height: SVGAnimatedLength;\n    readonly result: SVGAnimatedString;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n}\n\ninterface SVGFitToViewBox {\n    readonly preserveAspectRatio: SVGAnimatedPreserveAspectRatio;\n    readonly viewBox: SVGAnimatedRect;\n}\n\ninterface SVGForeignObjectElement extends SVGGraphicsElement {\n    readonly height: SVGAnimatedLength;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGForeignObjectElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGForeignObjectElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGForeignObjectElement: {\n    prototype: SVGForeignObjectElement;\n    new(): SVGForeignObjectElement;\n};\n\ninterface SVGGElement extends SVGGraphicsElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGGElement: {\n    prototype: SVGGElement;\n    new(): SVGGElement;\n};\n\ninterface SVGGeometryElement extends SVGGraphicsElement {\n    readonly pathLength: SVGAnimatedNumber;\n    getPointAtLength(distance: number): DOMPoint;\n    getTotalLength(): number;\n    isPointInFill(point?: DOMPointInit): boolean;\n    isPointInStroke(point?: DOMPointInit): boolean;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGeometryElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGeometryElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGGeometryElement: {\n    prototype: SVGGeometryElement;\n    new(): SVGGeometryElement;\n};\n\ninterface SVGGradientElement extends SVGElement, SVGURIReference {\n    readonly gradientTransform: SVGAnimatedTransformList;\n    readonly gradientUnits: SVGAnimatedEnumeration;\n    readonly spreadMethod: SVGAnimatedEnumeration;\n    readonly SVG_SPREADMETHOD_PAD: number;\n    readonly SVG_SPREADMETHOD_REFLECT: number;\n    readonly SVG_SPREADMETHOD_REPEAT: number;\n    readonly SVG_SPREADMETHOD_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGGradientElement: {\n    prototype: SVGGradientElement;\n    new(): SVGGradientElement;\n    readonly SVG_SPREADMETHOD_PAD: number;\n    readonly SVG_SPREADMETHOD_REFLECT: number;\n    readonly SVG_SPREADMETHOD_REPEAT: number;\n    readonly SVG_SPREADMETHOD_UNKNOWN: number;\n};\n\ninterface SVGGraphicsElement extends SVGElement, SVGTests {\n    readonly transform: SVGAnimatedTransformList;\n    getBBox(options?: SVGBoundingBoxOptions): DOMRect;\n    getCTM(): DOMMatrix | null;\n    getScreenCTM(): DOMMatrix | null;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGraphicsElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGraphicsElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGGraphicsElement: {\n    prototype: SVGGraphicsElement;\n    new(): SVGGraphicsElement;\n};\n\ninterface SVGImageElement extends SVGGraphicsElement, SVGURIReference {\n    readonly height: SVGAnimatedLength;\n    readonly preserveAspectRatio: SVGAnimatedPreserveAspectRatio;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGImageElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGImageElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGImageElement: {\n    prototype: SVGImageElement;\n    new(): SVGImageElement;\n};\n\ninterface SVGLength {\n    readonly unitType: number;\n    value: number;\n    valueAsString: string;\n    valueInSpecifiedUnits: number;\n    convertToSpecifiedUnits(unitType: number): void;\n    newValueSpecifiedUnits(unitType: number, valueInSpecifiedUnits: number): void;\n    readonly SVG_LENGTHTYPE_CM: number;\n    readonly SVG_LENGTHTYPE_EMS: number;\n    readonly SVG_LENGTHTYPE_EXS: number;\n    readonly SVG_LENGTHTYPE_IN: number;\n    readonly SVG_LENGTHTYPE_MM: number;\n    readonly SVG_LENGTHTYPE_NUMBER: number;\n    readonly SVG_LENGTHTYPE_PC: number;\n    readonly SVG_LENGTHTYPE_PERCENTAGE: number;\n    readonly SVG_LENGTHTYPE_PT: number;\n    readonly SVG_LENGTHTYPE_PX: number;\n    readonly SVG_LENGTHTYPE_UNKNOWN: number;\n}\n\ndeclare var SVGLength: {\n    prototype: SVGLength;\n    new(): SVGLength;\n    readonly SVG_LENGTHTYPE_CM: number;\n    readonly SVG_LENGTHTYPE_EMS: number;\n    readonly SVG_LENGTHTYPE_EXS: number;\n    readonly SVG_LENGTHTYPE_IN: number;\n    readonly SVG_LENGTHTYPE_MM: number;\n    readonly SVG_LENGTHTYPE_NUMBER: number;\n    readonly SVG_LENGTHTYPE_PC: number;\n    readonly SVG_LENGTHTYPE_PERCENTAGE: number;\n    readonly SVG_LENGTHTYPE_PT: number;\n    readonly SVG_LENGTHTYPE_PX: number;\n    readonly SVG_LENGTHTYPE_UNKNOWN: number;\n};\n\ninterface SVGLengthList {\n    readonly length: number;\n    readonly numberOfItems: number;\n    appendItem(newItem: SVGLength): SVGLength;\n    clear(): void;\n    getItem(index: number): SVGLength;\n    initialize(newItem: SVGLength): SVGLength;\n    insertItemBefore(newItem: SVGLength, index: number): SVGLength;\n    removeItem(index: number): SVGLength;\n    replaceItem(newItem: SVGLength, index: number): SVGLength;\n    [index: number]: SVGLength;\n}\n\ndeclare var SVGLengthList: {\n    prototype: SVGLengthList;\n    new(): SVGLengthList;\n};\n\ninterface SVGLineElement extends SVGGraphicsElement {\n    readonly x1: SVGAnimatedLength;\n    readonly x2: SVGAnimatedLength;\n    readonly y1: SVGAnimatedLength;\n    readonly y2: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGLineElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGLineElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGLineElement: {\n    prototype: SVGLineElement;\n    new(): SVGLineElement;\n};\n\ninterface SVGLinearGradientElement extends SVGGradientElement {\n    readonly x1: SVGAnimatedLength;\n    readonly x2: SVGAnimatedLength;\n    readonly y1: SVGAnimatedLength;\n    readonly y2: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGLinearGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGLinearGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGLinearGradientElement: {\n    prototype: SVGLinearGradientElement;\n    new(): SVGLinearGradientElement;\n};\n\ninterface SVGMarkerElement extends SVGElement, SVGFitToViewBox {\n    readonly markerHeight: SVGAnimatedLength;\n    readonly markerUnits: SVGAnimatedEnumeration;\n    readonly markerWidth: SVGAnimatedLength;\n    readonly orientAngle: SVGAnimatedAngle;\n    readonly orientType: SVGAnimatedEnumeration;\n    readonly refX: SVGAnimatedLength;\n    readonly refY: SVGAnimatedLength;\n    setOrientToAngle(angle: SVGAngle): void;\n    setOrientToAuto(): void;\n    readonly SVG_MARKERUNITS_STROKEWIDTH: number;\n    readonly SVG_MARKERUNITS_UNKNOWN: number;\n    readonly SVG_MARKERUNITS_USERSPACEONUSE: number;\n    readonly SVG_MARKER_ORIENT_ANGLE: number;\n    readonly SVG_MARKER_ORIENT_AUTO: number;\n    readonly SVG_MARKER_ORIENT_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMarkerElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMarkerElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGMarkerElement: {\n    prototype: SVGMarkerElement;\n    new(): SVGMarkerElement;\n    readonly SVG_MARKERUNITS_STROKEWIDTH: number;\n    readonly SVG_MARKERUNITS_UNKNOWN: number;\n    readonly SVG_MARKERUNITS_USERSPACEONUSE: number;\n    readonly SVG_MARKER_ORIENT_ANGLE: number;\n    readonly SVG_MARKER_ORIENT_AUTO: number;\n    readonly SVG_MARKER_ORIENT_UNKNOWN: number;\n};\n\ninterface SVGMaskElement extends SVGElement, SVGTests {\n    readonly height: SVGAnimatedLength;\n    readonly maskContentUnits: SVGAnimatedEnumeration;\n    readonly maskUnits: SVGAnimatedEnumeration;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMaskElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMaskElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGMaskElement: {\n    prototype: SVGMaskElement;\n    new(): SVGMaskElement;\n};\n\ninterface SVGMetadataElement extends SVGElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMetadataElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMetadataElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGMetadataElement: {\n    prototype: SVGMetadataElement;\n    new(): SVGMetadataElement;\n};\n\ninterface SVGNumber {\n    value: number;\n}\n\ndeclare var SVGNumber: {\n    prototype: SVGNumber;\n    new(): SVGNumber;\n};\n\ninterface SVGNumberList {\n    readonly length: number;\n    readonly numberOfItems: number;\n    appendItem(newItem: SVGNumber): SVGNumber;\n    clear(): void;\n    getItem(index: number): SVGNumber;\n    initialize(newItem: SVGNumber): SVGNumber;\n    insertItemBefore(newItem: SVGNumber, index: number): SVGNumber;\n    removeItem(index: number): SVGNumber;\n    replaceItem(newItem: SVGNumber, index: number): SVGNumber;\n    [index: number]: SVGNumber;\n}\n\ndeclare var SVGNumberList: {\n    prototype: SVGNumberList;\n    new(): SVGNumberList;\n};\n\ninterface SVGPathElement extends SVGGraphicsElement {\n    /** @deprecated */\n    readonly pathSegList: SVGPathSegList;\n    /** @deprecated */\n    createSVGPathSegArcAbs(x: number, y: number, r1: number, r2: number, angle: number, largeArcFlag: boolean, sweepFlag: boolean): SVGPathSegArcAbs;\n    /** @deprecated */\n    createSVGPathSegArcRel(x: number, y: number, r1: number, r2: number, angle: number, largeArcFlag: boolean, sweepFlag: boolean): SVGPathSegArcRel;\n    /** @deprecated */\n    createSVGPathSegClosePath(): SVGPathSegClosePath;\n    /** @deprecated */\n    createSVGPathSegCurvetoCubicAbs(x: number, y: number, x1: number, y1: number, x2: number, y2: number): SVGPathSegCurvetoCubicAbs;\n    /** @deprecated */\n    createSVGPathSegCurvetoCubicRel(x: number, y: number, x1: number, y1: number, x2: number, y2: number): SVGPathSegCurvetoCubicRel;\n    /** @deprecated */\n    createSVGPathSegCurvetoCubicSmoothAbs(x: number, y: number, x2: number, y2: number): SVGPathSegCurvetoCubicSmoothAbs;\n    /** @deprecated */\n    createSVGPathSegCurvetoCubicSmoothRel(x: number, y: number, x2: number, y2: number): SVGPathSegCurvetoCubicSmoothRel;\n    /** @deprecated */\n    createSVGPathSegCurvetoQuadraticAbs(x: number, y: number, x1: number, y1: number): SVGPathSegCurvetoQuadraticAbs;\n    /** @deprecated */\n    createSVGPathSegCurvetoQuadraticRel(x: number, y: number, x1: number, y1: number): SVGPathSegCurvetoQuadraticRel;\n    /** @deprecated */\n    createSVGPathSegCurvetoQuadraticSmoothAbs(x: number, y: number): SVGPathSegCurvetoQuadraticSmoothAbs;\n    /** @deprecated */\n    createSVGPathSegCurvetoQuadraticSmoothRel(x: number, y: number): SVGPathSegCurvetoQuadraticSmoothRel;\n    /** @deprecated */\n    createSVGPathSegLinetoAbs(x: number, y: number): SVGPathSegLinetoAbs;\n    /** @deprecated */\n    createSVGPathSegLinetoHorizontalAbs(x: number): SVGPathSegLinetoHorizontalAbs;\n    /** @deprecated */\n    createSVGPathSegLinetoHorizontalRel(x: number): SVGPathSegLinetoHorizontalRel;\n    /** @deprecated */\n    createSVGPathSegLinetoRel(x: number, y: number): SVGPathSegLinetoRel;\n    /** @deprecated */\n    createSVGPathSegLinetoVerticalAbs(y: number): SVGPathSegLinetoVerticalAbs;\n    /** @deprecated */\n    createSVGPathSegLinetoVerticalRel(y: number): SVGPathSegLinetoVerticalRel;\n    /** @deprecated */\n    createSVGPathSegMovetoAbs(x: number, y: number): SVGPathSegMovetoAbs;\n    /** @deprecated */\n    createSVGPathSegMovetoRel(x: number, y: number): SVGPathSegMovetoRel;\n    /** @deprecated */\n    getPathSegAtLength(distance: number): number;\n    getPointAtLength(distance: number): SVGPoint;\n    getTotalLength(): number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGPathElement: {\n    prototype: SVGPathElement;\n    new(): SVGPathElement;\n};\n\ninterface SVGPathSeg {\n    readonly pathSegType: number;\n    readonly pathSegTypeAsLetter: string;\n    readonly PATHSEG_ARC_ABS: number;\n    readonly PATHSEG_ARC_REL: number;\n    readonly PATHSEG_CLOSEPATH: number;\n    readonly PATHSEG_CURVETO_CUBIC_ABS: number;\n    readonly PATHSEG_CURVETO_CUBIC_REL: number;\n    readonly PATHSEG_CURVETO_CUBIC_SMOOTH_ABS: number;\n    readonly PATHSEG_CURVETO_CUBIC_SMOOTH_REL: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_ABS: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_REL: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL: number;\n    readonly PATHSEG_LINETO_ABS: number;\n    readonly PATHSEG_LINETO_HORIZONTAL_ABS: number;\n    readonly PATHSEG_LINETO_HORIZONTAL_REL: number;\n    readonly PATHSEG_LINETO_REL: number;\n    readonly PATHSEG_LINETO_VERTICAL_ABS: number;\n    readonly PATHSEG_LINETO_VERTICAL_REL: number;\n    readonly PATHSEG_MOVETO_ABS: number;\n    readonly PATHSEG_MOVETO_REL: number;\n    readonly PATHSEG_UNKNOWN: number;\n}\n\ndeclare var SVGPathSeg: {\n    prototype: SVGPathSeg;\n    new(): SVGPathSeg;\n    readonly PATHSEG_ARC_ABS: number;\n    readonly PATHSEG_ARC_REL: number;\n    readonly PATHSEG_CLOSEPATH: number;\n    readonly PATHSEG_CURVETO_CUBIC_ABS: number;\n    readonly PATHSEG_CURVETO_CUBIC_REL: number;\n    readonly PATHSEG_CURVETO_CUBIC_SMOOTH_ABS: number;\n    readonly PATHSEG_CURVETO_CUBIC_SMOOTH_REL: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_ABS: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_REL: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL: number;\n    readonly PATHSEG_LINETO_ABS: number;\n    readonly PATHSEG_LINETO_HORIZONTAL_ABS: number;\n    readonly PATHSEG_LINETO_HORIZONTAL_REL: number;\n    readonly PATHSEG_LINETO_REL: number;\n    readonly PATHSEG_LINETO_VERTICAL_ABS: number;\n    readonly PATHSEG_LINETO_VERTICAL_REL: number;\n    readonly PATHSEG_MOVETO_ABS: number;\n    readonly PATHSEG_MOVETO_REL: number;\n    readonly PATHSEG_UNKNOWN: number;\n};\n\ninterface SVGPathSegArcAbs extends SVGPathSeg {\n    angle: number;\n    largeArcFlag: boolean;\n    r1: number;\n    r2: number;\n    sweepFlag: boolean;\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegArcAbs: {\n    prototype: SVGPathSegArcAbs;\n    new(): SVGPathSegArcAbs;\n};\n\ninterface SVGPathSegArcRel extends SVGPathSeg {\n    angle: number;\n    largeArcFlag: boolean;\n    r1: number;\n    r2: number;\n    sweepFlag: boolean;\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegArcRel: {\n    prototype: SVGPathSegArcRel;\n    new(): SVGPathSegArcRel;\n};\n\ninterface SVGPathSegClosePath extends SVGPathSeg {\n}\n\ndeclare var SVGPathSegClosePath: {\n    prototype: SVGPathSegClosePath;\n    new(): SVGPathSegClosePath;\n};\n\ninterface SVGPathSegCurvetoCubicAbs extends SVGPathSeg {\n    x: number;\n    x1: number;\n    x2: number;\n    y: number;\n    y1: number;\n    y2: number;\n}\n\ndeclare var SVGPathSegCurvetoCubicAbs: {\n    prototype: SVGPathSegCurvetoCubicAbs;\n    new(): SVGPathSegCurvetoCubicAbs;\n};\n\ninterface SVGPathSegCurvetoCubicRel extends SVGPathSeg {\n    x: number;\n    x1: number;\n    x2: number;\n    y: number;\n    y1: number;\n    y2: number;\n}\n\ndeclare var SVGPathSegCurvetoCubicRel: {\n    prototype: SVGPathSegCurvetoCubicRel;\n    new(): SVGPathSegCurvetoCubicRel;\n};\n\ninterface SVGPathSegCurvetoCubicSmoothAbs extends SVGPathSeg {\n    x: number;\n    x2: number;\n    y: number;\n    y2: number;\n}\n\ndeclare var SVGPathSegCurvetoCubicSmoothAbs: {\n    prototype: SVGPathSegCurvetoCubicSmoothAbs;\n    new(): SVGPathSegCurvetoCubicSmoothAbs;\n};\n\ninterface SVGPathSegCurvetoCubicSmoothRel extends SVGPathSeg {\n    x: number;\n    x2: number;\n    y: number;\n    y2: number;\n}\n\ndeclare var SVGPathSegCurvetoCubicSmoothRel: {\n    prototype: SVGPathSegCurvetoCubicSmoothRel;\n    new(): SVGPathSegCurvetoCubicSmoothRel;\n};\n\ninterface SVGPathSegCurvetoQuadraticAbs extends SVGPathSeg {\n    x: number;\n    x1: number;\n    y: number;\n    y1: number;\n}\n\ndeclare var SVGPathSegCurvetoQuadraticAbs: {\n    prototype: SVGPathSegCurvetoQuadraticAbs;\n    new(): SVGPathSegCurvetoQuadraticAbs;\n};\n\ninterface SVGPathSegCurvetoQuadraticRel extends SVGPathSeg {\n    x: number;\n    x1: number;\n    y: number;\n    y1: number;\n}\n\ndeclare var SVGPathSegCurvetoQuadraticRel: {\n    prototype: SVGPathSegCurvetoQuadraticRel;\n    new(): SVGPathSegCurvetoQuadraticRel;\n};\n\ninterface SVGPathSegCurvetoQuadraticSmoothAbs extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegCurvetoQuadraticSmoothAbs: {\n    prototype: SVGPathSegCurvetoQuadraticSmoothAbs;\n    new(): SVGPathSegCurvetoQuadraticSmoothAbs;\n};\n\ninterface SVGPathSegCurvetoQuadraticSmoothRel extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegCurvetoQuadraticSmoothRel: {\n    prototype: SVGPathSegCurvetoQuadraticSmoothRel;\n    new(): SVGPathSegCurvetoQuadraticSmoothRel;\n};\n\ninterface SVGPathSegLinetoAbs extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegLinetoAbs: {\n    prototype: SVGPathSegLinetoAbs;\n    new(): SVGPathSegLinetoAbs;\n};\n\ninterface SVGPathSegLinetoHorizontalAbs extends SVGPathSeg {\n    x: number;\n}\n\ndeclare var SVGPathSegLinetoHorizontalAbs: {\n    prototype: SVGPathSegLinetoHorizontalAbs;\n    new(): SVGPathSegLinetoHorizontalAbs;\n};\n\ninterface SVGPathSegLinetoHorizontalRel extends SVGPathSeg {\n    x: number;\n}\n\ndeclare var SVGPathSegLinetoHorizontalRel: {\n    prototype: SVGPathSegLinetoHorizontalRel;\n    new(): SVGPathSegLinetoHorizontalRel;\n};\n\ninterface SVGPathSegLinetoRel extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegLinetoRel: {\n    prototype: SVGPathSegLinetoRel;\n    new(): SVGPathSegLinetoRel;\n};\n\ninterface SVGPathSegLinetoVerticalAbs extends SVGPathSeg {\n    y: number;\n}\n\ndeclare var SVGPathSegLinetoVerticalAbs: {\n    prototype: SVGPathSegLinetoVerticalAbs;\n    new(): SVGPathSegLinetoVerticalAbs;\n};\n\ninterface SVGPathSegLinetoVerticalRel extends SVGPathSeg {\n    y: number;\n}\n\ndeclare var SVGPathSegLinetoVerticalRel: {\n    prototype: SVGPathSegLinetoVerticalRel;\n    new(): SVGPathSegLinetoVerticalRel;\n};\n\ninterface SVGPathSegList {\n    readonly numberOfItems: number;\n    appendItem(newItem: SVGPathSeg): SVGPathSeg;\n    clear(): void;\n    getItem(index: number): SVGPathSeg;\n    initialize(newItem: SVGPathSeg): SVGPathSeg;\n    insertItemBefore(newItem: SVGPathSeg, index: number): SVGPathSeg;\n    removeItem(index: number): SVGPathSeg;\n    replaceItem(newItem: SVGPathSeg, index: number): SVGPathSeg;\n}\n\ndeclare var SVGPathSegList: {\n    prototype: SVGPathSegList;\n    new(): SVGPathSegList;\n};\n\ninterface SVGPathSegMovetoAbs extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegMovetoAbs: {\n    prototype: SVGPathSegMovetoAbs;\n    new(): SVGPathSegMovetoAbs;\n};\n\ninterface SVGPathSegMovetoRel extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegMovetoRel: {\n    prototype: SVGPathSegMovetoRel;\n    new(): SVGPathSegMovetoRel;\n};\n\ninterface SVGPatternElement extends SVGElement, SVGTests, SVGFitToViewBox, SVGURIReference {\n    readonly height: SVGAnimatedLength;\n    readonly patternContentUnits: SVGAnimatedEnumeration;\n    readonly patternTransform: SVGAnimatedTransformList;\n    readonly patternUnits: SVGAnimatedEnumeration;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPatternElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPatternElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGPatternElement: {\n    prototype: SVGPatternElement;\n    new(): SVGPatternElement;\n};\n\ninterface SVGPointList {\n    readonly numberOfItems: number;\n    appendItem(newItem: SVGPoint): SVGPoint;\n    clear(): void;\n    getItem(index: number): SVGPoint;\n    initialize(newItem: SVGPoint): SVGPoint;\n    insertItemBefore(newItem: SVGPoint, index: number): SVGPoint;\n    removeItem(index: number): SVGPoint;\n    replaceItem(newItem: SVGPoint, index: number): SVGPoint;\n}\n\ndeclare var SVGPointList: {\n    prototype: SVGPointList;\n    new(): SVGPointList;\n};\n\ninterface SVGPolygonElement extends SVGGraphicsElement, SVGAnimatedPoints {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPolygonElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPolygonElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGPolygonElement: {\n    prototype: SVGPolygonElement;\n    new(): SVGPolygonElement;\n};\n\ninterface SVGPolylineElement extends SVGGraphicsElement, SVGAnimatedPoints {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPolylineElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPolylineElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGPolylineElement: {\n    prototype: SVGPolylineElement;\n    new(): SVGPolylineElement;\n};\n\ninterface SVGPreserveAspectRatio {\n    align: number;\n    meetOrSlice: number;\n    readonly SVG_MEETORSLICE_MEET: number;\n    readonly SVG_MEETORSLICE_SLICE: number;\n    readonly SVG_MEETORSLICE_UNKNOWN: number;\n    readonly SVG_PRESERVEASPECTRATIO_NONE: number;\n    readonly SVG_PRESERVEASPECTRATIO_UNKNOWN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMIN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMIN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMIN: number;\n}\n\ndeclare var SVGPreserveAspectRatio: {\n    prototype: SVGPreserveAspectRatio;\n    new(): SVGPreserveAspectRatio;\n    readonly SVG_MEETORSLICE_MEET: number;\n    readonly SVG_MEETORSLICE_SLICE: number;\n    readonly SVG_MEETORSLICE_UNKNOWN: number;\n    readonly SVG_PRESERVEASPECTRATIO_NONE: number;\n    readonly SVG_PRESERVEASPECTRATIO_UNKNOWN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMIN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMIN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMIN: number;\n};\n\ninterface SVGRadialGradientElement extends SVGGradientElement {\n    readonly cx: SVGAnimatedLength;\n    readonly cy: SVGAnimatedLength;\n    readonly fx: SVGAnimatedLength;\n    readonly fy: SVGAnimatedLength;\n    readonly r: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGRadialGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGRadialGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGRadialGradientElement: {\n    prototype: SVGRadialGradientElement;\n    new(): SVGRadialGradientElement;\n};\n\ninterface SVGRectElement extends SVGGraphicsElement {\n    readonly height: SVGAnimatedLength;\n    readonly rx: SVGAnimatedLength;\n    readonly ry: SVGAnimatedLength;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGRectElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGRectElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGRectElement: {\n    prototype: SVGRectElement;\n    new(): SVGRectElement;\n};\n\ninterface SVGSVGElementEventMap extends SVGElementEventMap {\n    "SVGUnload": Event;\n    "SVGZoom": SVGZoomEvent;\n}\n\ninterface SVGSVGElement extends SVGGraphicsElement, DocumentEvent, SVGFitToViewBox, SVGZoomAndPan {\n    /** @deprecated */\n    contentScriptType: string;\n    /** @deprecated */\n    contentStyleType: string;\n    currentScale: number;\n    readonly currentTranslate: SVGPoint;\n    readonly height: SVGAnimatedLength;\n    onunload: ((this: SVGSVGElement, ev: Event) => any) | null;\n    onzoom: ((this: SVGSVGElement, ev: SVGZoomEvent) => any) | null;\n    /** @deprecated */\n    readonly pixelUnitToMillimeterX: number;\n    /** @deprecated */\n    readonly pixelUnitToMillimeterY: number;\n    /** @deprecated */\n    readonly screenPixelToMillimeterX: number;\n    /** @deprecated */\n    readonly screenPixelToMillimeterY: number;\n    /** @deprecated */\n    readonly viewport: SVGRect;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    checkEnclosure(element: SVGElement, rect: SVGRect): boolean;\n    checkIntersection(element: SVGElement, rect: SVGRect): boolean;\n    createSVGAngle(): SVGAngle;\n    createSVGLength(): SVGLength;\n    createSVGMatrix(): SVGMatrix;\n    createSVGNumber(): SVGNumber;\n    createSVGPoint(): SVGPoint;\n    createSVGRect(): SVGRect;\n    createSVGTransform(): SVGTransform;\n    createSVGTransformFromMatrix(matrix: SVGMatrix): SVGTransform;\n    deselectAll(): void;\n    /** @deprecated */\n    forceRedraw(): void;\n    getComputedStyle(elt: Element, pseudoElt?: string | null): CSSStyleDeclaration;\n    /** @deprecated */\n    getCurrentTime(): number;\n    getElementById(elementId: string): Element;\n    getEnclosureList(rect: SVGRect, referenceElement: SVGElement): NodeListOf<SVGCircleElement | SVGEllipseElement | SVGImageElement | SVGLineElement | SVGPathElement | SVGPolygonElement | SVGPolylineElement | SVGRectElement | SVGTextElement | SVGUseElement>;\n    getIntersectionList(rect: SVGRect, referenceElement: SVGElement): NodeListOf<SVGCircleElement | SVGEllipseElement | SVGImageElement | SVGLineElement | SVGPathElement | SVGPolygonElement | SVGPolylineElement | SVGRectElement | SVGTextElement | SVGUseElement>;\n    /** @deprecated */\n    pauseAnimations(): void;\n    /** @deprecated */\n    setCurrentTime(seconds: number): void;\n    /** @deprecated */\n    suspendRedraw(maxWaitMilliseconds: number): number;\n    /** @deprecated */\n    unpauseAnimations(): void;\n    /** @deprecated */\n    unsuspendRedraw(suspendHandleID: number): void;\n    /** @deprecated */\n    unsuspendRedrawAll(): void;\n    addEventListener<K extends keyof SVGSVGElementEventMap>(type: K, listener: (this: SVGSVGElement, ev: SVGSVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGSVGElementEventMap>(type: K, listener: (this: SVGSVGElement, ev: SVGSVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGSVGElement: {\n    prototype: SVGSVGElement;\n    new(): SVGSVGElement;\n    readonly SVG_ZOOMANDPAN_DISABLE: number;\n    readonly SVG_ZOOMANDPAN_MAGNIFY: number;\n    readonly SVG_ZOOMANDPAN_UNKNOWN: number;\n};\n\ninterface SVGScriptElement extends SVGElement, SVGURIReference {\n    type: string;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGScriptElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGScriptElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGScriptElement: {\n    prototype: SVGScriptElement;\n    new(): SVGScriptElement;\n};\n\ninterface SVGStopElement extends SVGElement {\n    readonly offset: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGStopElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGStopElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGStopElement: {\n    prototype: SVGStopElement;\n    new(): SVGStopElement;\n};\n\ninterface SVGStringList {\n    readonly length: number;\n    readonly numberOfItems: number;\n    appendItem(newItem: string): string;\n    clear(): void;\n    getItem(index: number): string;\n    initialize(newItem: string): string;\n    insertItemBefore(newItem: string, index: number): string;\n    removeItem(index: number): string;\n    replaceItem(newItem: string, index: number): string;\n    [index: number]: string;\n}\n\ndeclare var SVGStringList: {\n    prototype: SVGStringList;\n    new(): SVGStringList;\n};\n\ninterface SVGStyleElement extends SVGElement {\n    disabled: boolean;\n    media: string;\n    title: string;\n    type: string;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGStyleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGStyleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGStyleElement: {\n    prototype: SVGStyleElement;\n    new(): SVGStyleElement;\n};\n\ninterface SVGSwitchElement extends SVGGraphicsElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGSwitchElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGSwitchElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGSwitchElement: {\n    prototype: SVGSwitchElement;\n    new(): SVGSwitchElement;\n};\n\ninterface SVGSymbolElement extends SVGElement, SVGFitToViewBox {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGSymbolElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGSymbolElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGSymbolElement: {\n    prototype: SVGSymbolElement;\n    new(): SVGSymbolElement;\n};\n\ninterface SVGTSpanElement extends SVGTextPositioningElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTSpanElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTSpanElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTSpanElement: {\n    prototype: SVGTSpanElement;\n    new(): SVGTSpanElement;\n};\n\ninterface SVGTests {\n    readonly requiredExtensions: SVGStringList;\n    readonly systemLanguage: SVGStringList;\n}\n\ninterface SVGTextContentElement extends SVGGraphicsElement {\n    readonly lengthAdjust: SVGAnimatedEnumeration;\n    readonly textLength: SVGAnimatedLength;\n    getCharNumAtPosition(point: SVGPoint): number;\n    getComputedTextLength(): number;\n    getEndPositionOfChar(charnum: number): SVGPoint;\n    getExtentOfChar(charnum: number): SVGRect;\n    getNumberOfChars(): number;\n    getRotationOfChar(charnum: number): number;\n    getStartPositionOfChar(charnum: number): SVGPoint;\n    getSubStringLength(charnum: number, nchars: number): number;\n    selectSubString(charnum: number, nchars: number): void;\n    readonly LENGTHADJUST_SPACING: number;\n    readonly LENGTHADJUST_SPACINGANDGLYPHS: number;\n    readonly LENGTHADJUST_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextContentElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextContentElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTextContentElement: {\n    prototype: SVGTextContentElement;\n    new(): SVGTextContentElement;\n    readonly LENGTHADJUST_SPACING: number;\n    readonly LENGTHADJUST_SPACINGANDGLYPHS: number;\n    readonly LENGTHADJUST_UNKNOWN: number;\n};\n\ninterface SVGTextElement extends SVGTextPositioningElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTextElement: {\n    prototype: SVGTextElement;\n    new(): SVGTextElement;\n};\n\ninterface SVGTextPathElement extends SVGTextContentElement, SVGURIReference {\n    readonly method: SVGAnimatedEnumeration;\n    readonly spacing: SVGAnimatedEnumeration;\n    readonly startOffset: SVGAnimatedLength;\n    readonly TEXTPATH_METHODTYPE_ALIGN: number;\n    readonly TEXTPATH_METHODTYPE_STRETCH: number;\n    readonly TEXTPATH_METHODTYPE_UNKNOWN: number;\n    readonly TEXTPATH_SPACINGTYPE_AUTO: number;\n    readonly TEXTPATH_SPACINGTYPE_EXACT: number;\n    readonly TEXTPATH_SPACINGTYPE_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTextPathElement: {\n    prototype: SVGTextPathElement;\n    new(): SVGTextPathElement;\n    readonly TEXTPATH_METHODTYPE_ALIGN: number;\n    readonly TEXTPATH_METHODTYPE_STRETCH: number;\n    readonly TEXTPATH_METHODTYPE_UNKNOWN: number;\n    readonly TEXTPATH_SPACINGTYPE_AUTO: number;\n    readonly TEXTPATH_SPACINGTYPE_EXACT: number;\n    readonly TEXTPATH_SPACINGTYPE_UNKNOWN: number;\n};\n\ninterface SVGTextPositioningElement extends SVGTextContentElement {\n    readonly dx: SVGAnimatedLengthList;\n    readonly dy: SVGAnimatedLengthList;\n    readonly rotate: SVGAnimatedNumberList;\n    readonly x: SVGAnimatedLengthList;\n    readonly y: SVGAnimatedLengthList;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextPositioningElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextPositioningElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTextPositioningElement: {\n    prototype: SVGTextPositioningElement;\n    new(): SVGTextPositioningElement;\n};\n\ninterface SVGTitleElement extends SVGElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTitleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTitleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTitleElement: {\n    prototype: SVGTitleElement;\n    new(): SVGTitleElement;\n};\n\ninterface SVGTransform {\n    readonly angle: number;\n    readonly matrix: SVGMatrix;\n    readonly type: number;\n    setMatrix(matrix: SVGMatrix): void;\n    setRotate(angle: number, cx: number, cy: number): void;\n    setScale(sx: number, sy: number): void;\n    setSkewX(angle: number): void;\n    setSkewY(angle: number): void;\n    setTranslate(tx: number, ty: number): void;\n    readonly SVG_TRANSFORM_MATRIX: number;\n    readonly SVG_TRANSFORM_ROTATE: number;\n    readonly SVG_TRANSFORM_SCALE: number;\n    readonly SVG_TRANSFORM_SKEWX: number;\n    readonly SVG_TRANSFORM_SKEWY: number;\n    readonly SVG_TRANSFORM_TRANSLATE: number;\n    readonly SVG_TRANSFORM_UNKNOWN: number;\n}\n\ndeclare var SVGTransform: {\n    prototype: SVGTransform;\n    new(): SVGTransform;\n    readonly SVG_TRANSFORM_MATRIX: number;\n    readonly SVG_TRANSFORM_ROTATE: number;\n    readonly SVG_TRANSFORM_SCALE: number;\n    readonly SVG_TRANSFORM_SKEWX: number;\n    readonly SVG_TRANSFORM_SKEWY: number;\n    readonly SVG_TRANSFORM_TRANSLATE: number;\n    readonly SVG_TRANSFORM_UNKNOWN: number;\n};\n\ninterface SVGTransformList {\n    readonly numberOfItems: number;\n    appendItem(newItem: SVGTransform): SVGTransform;\n    clear(): void;\n    consolidate(): SVGTransform;\n    createSVGTransformFromMatrix(matrix: SVGMatrix): SVGTransform;\n    getItem(index: number): SVGTransform;\n    initialize(newItem: SVGTransform): SVGTransform;\n    insertItemBefore(newItem: SVGTransform, index: number): SVGTransform;\n    removeItem(index: number): SVGTransform;\n    replaceItem(newItem: SVGTransform, index: number): SVGTransform;\n}\n\ndeclare var SVGTransformList: {\n    prototype: SVGTransformList;\n    new(): SVGTransformList;\n};\n\ninterface SVGURIReference {\n    readonly href: SVGAnimatedString;\n}\n\ninterface SVGUnitTypes {\n    readonly SVG_UNIT_TYPE_OBJECTBOUNDINGBOX: number;\n    readonly SVG_UNIT_TYPE_UNKNOWN: number;\n    readonly SVG_UNIT_TYPE_USERSPACEONUSE: number;\n}\n\ndeclare var SVGUnitTypes: {\n    prototype: SVGUnitTypes;\n    new(): SVGUnitTypes;\n    readonly SVG_UNIT_TYPE_OBJECTBOUNDINGBOX: number;\n    readonly SVG_UNIT_TYPE_UNKNOWN: number;\n    readonly SVG_UNIT_TYPE_USERSPACEONUSE: number;\n};\n\ninterface SVGUseElement extends SVGGraphicsElement, SVGURIReference {\n    readonly animatedInstanceRoot: SVGElementInstance | null;\n    readonly height: SVGAnimatedLength;\n    readonly instanceRoot: SVGElementInstance | null;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGUseElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGUseElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGUseElement: {\n    prototype: SVGUseElement;\n    new(): SVGUseElement;\n};\n\ninterface SVGViewElement extends SVGElement, SVGFitToViewBox, SVGZoomAndPan {\n    /** @deprecated */\n    readonly viewTarget: SVGStringList;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGViewElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGViewElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGViewElement: {\n    prototype: SVGViewElement;\n    new(): SVGViewElement;\n    readonly SVG_ZOOMANDPAN_DISABLE: number;\n    readonly SVG_ZOOMANDPAN_MAGNIFY: number;\n    readonly SVG_ZOOMANDPAN_UNKNOWN: number;\n};\n\ninterface SVGZoomAndPan {\n    readonly zoomAndPan: number;\n}\n\ndeclare var SVGZoomAndPan: {\n    readonly SVG_ZOOMANDPAN_DISABLE: number;\n    readonly SVG_ZOOMANDPAN_MAGNIFY: number;\n    readonly SVG_ZOOMANDPAN_UNKNOWN: number;\n};\n\ninterface SVGZoomEvent extends UIEvent {\n    readonly newScale: number;\n    readonly newTranslate: SVGPoint;\n    readonly previousScale: number;\n    readonly previousTranslate: SVGPoint;\n    readonly zoomRectScreen: SVGRect;\n}\n\ndeclare var SVGZoomEvent: {\n    prototype: SVGZoomEvent;\n    new(): SVGZoomEvent;\n};\n\ninterface ScopedCredential {\n    readonly id: ArrayBuffer;\n    readonly type: ScopedCredentialType;\n}\n\ndeclare var ScopedCredential: {\n    prototype: ScopedCredential;\n    new(): ScopedCredential;\n};\n\ninterface ScopedCredentialInfo {\n    readonly credential: ScopedCredential;\n    readonly publicKey: CryptoKey;\n}\n\ndeclare var ScopedCredentialInfo: {\n    prototype: ScopedCredentialInfo;\n    new(): ScopedCredentialInfo;\n};\n\ninterface Screen {\n    readonly availHeight: number;\n    readonly availWidth: number;\n    readonly colorDepth: number;\n    readonly height: number;\n    readonly orientation: ScreenOrientation;\n    readonly pixelDepth: number;\n    readonly width: number;\n}\n\ndeclare var Screen: {\n    prototype: Screen;\n    new(): Screen;\n};\n\ninterface ScreenOrientationEventMap {\n    "change": Event;\n}\n\ninterface ScreenOrientation extends EventTarget {\n    readonly angle: number;\n    onchange: ((this: ScreenOrientation, ev: Event) => any) | null;\n    readonly type: OrientationType;\n    lock(orientation: OrientationLockType): Promise<void>;\n    unlock(): void;\n    addEventListener<K extends keyof ScreenOrientationEventMap>(type: K, listener: (this: ScreenOrientation, ev: ScreenOrientationEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ScreenOrientationEventMap>(type: K, listener: (this: ScreenOrientation, ev: ScreenOrientationEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ScreenOrientation: {\n    prototype: ScreenOrientation;\n    new(): ScreenOrientation;\n};\n\ninterface ScriptProcessorNodeEventMap {\n    "audioprocess": AudioProcessingEvent;\n}\n\ninterface ScriptProcessorNode extends AudioNode {\n    /** @deprecated */\n    readonly bufferSize: number;\n    /** @deprecated */\n    onaudioprocess: ((this: ScriptProcessorNode, ev: AudioProcessingEvent) => any) | null;\n    addEventListener<K extends keyof ScriptProcessorNodeEventMap>(type: K, listener: (this: ScriptProcessorNode, ev: ScriptProcessorNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ScriptProcessorNodeEventMap>(type: K, listener: (this: ScriptProcessorNode, ev: ScriptProcessorNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ScriptProcessorNode: {\n    prototype: ScriptProcessorNode;\n    new(): ScriptProcessorNode;\n};\n\ninterface SecurityPolicyViolationEvent extends Event {\n    readonly blockedURI: string;\n    readonly columnNumber: number;\n    readonly documentURI: string;\n    readonly effectiveDirective: string;\n    readonly lineNumber: number;\n    readonly originalPolicy: string;\n    readonly referrer: string;\n    readonly sourceFile: string;\n    readonly statusCode: number;\n    readonly violatedDirective: string;\n}\n\ndeclare var SecurityPolicyViolationEvent: {\n    prototype: SecurityPolicyViolationEvent;\n    new(type: string, eventInitDict?: SecurityPolicyViolationEventInit): SecurityPolicyViolationEvent;\n};\n\ninterface Selection {\n    readonly anchorNode: Node;\n    readonly anchorOffset: number;\n    readonly baseNode: Node;\n    readonly baseOffset: number;\n    readonly extentNode: Node;\n    readonly extentOffset: number;\n    readonly focusNode: Node;\n    readonly focusOffset: number;\n    readonly isCollapsed: boolean;\n    readonly rangeCount: number;\n    readonly type: string;\n    addRange(range: Range): void;\n    collapse(parentNode: Node, offset: number): void;\n    collapseToEnd(): void;\n    collapseToStart(): void;\n    containsNode(node: Node, partlyContained: boolean): boolean;\n    deleteFromDocument(): void;\n    empty(): void;\n    extend(newNode: Node, offset: number): void;\n    getRangeAt(index: number): Range;\n    removeAllRanges(): void;\n    removeRange(range: Range): void;\n    selectAllChildren(parentNode: Node): void;\n    setBaseAndExtent(baseNode: Node, baseOffset: number, extentNode: Node, extentOffset: number): void;\n    setPosition(parentNode: Node, offset: number): void;\n    toString(): string;\n}\n\ndeclare var Selection: {\n    prototype: Selection;\n    new(): Selection;\n};\n\ninterface ServiceUIFrameContext {\n    getCachedFrameMessage(key: string): string;\n    postFrameMessage(key: string, data: string): void;\n}\ndeclare var ServiceUIFrameContext: ServiceUIFrameContext;\n\ninterface ServiceWorkerEventMap extends AbstractWorkerEventMap {\n    "statechange": Event;\n}\n\ninterface ServiceWorker extends EventTarget, AbstractWorker {\n    onstatechange: ((this: ServiceWorker, ev: Event) => any) | null;\n    readonly scriptURL: string;\n    readonly state: ServiceWorkerState;\n    postMessage(message: any, transfer?: Transferable[]): void;\n    addEventListener<K extends keyof ServiceWorkerEventMap>(type: K, listener: (this: ServiceWorker, ev: ServiceWorkerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ServiceWorkerEventMap>(type: K, listener: (this: ServiceWorker, ev: ServiceWorkerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ServiceWorker: {\n    prototype: ServiceWorker;\n    new(): ServiceWorker;\n};\n\ninterface ServiceWorkerContainerEventMap {\n    "controllerchange": Event;\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\ninterface ServiceWorkerContainer extends EventTarget {\n    readonly controller: ServiceWorker | null;\n    oncontrollerchange: ((this: ServiceWorkerContainer, ev: Event) => any) | null;\n    onmessage: ((this: ServiceWorkerContainer, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: ServiceWorkerContainer, ev: MessageEvent) => any) | null;\n    readonly ready: Promise<ServiceWorkerRegistration>;\n    getRegistration(clientURL?: string): Promise<ServiceWorkerRegistration | undefined>;\n    getRegistrations(): Promise<ReadonlyArray<ServiceWorkerRegistration>>;\n    register(scriptURL: string, options?: RegistrationOptions): Promise<ServiceWorkerRegistration>;\n    startMessages(): void;\n    addEventListener<K extends keyof ServiceWorkerContainerEventMap>(type: K, listener: (this: ServiceWorkerContainer, ev: ServiceWorkerContainerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ServiceWorkerContainerEventMap>(type: K, listener: (this: ServiceWorkerContainer, ev: ServiceWorkerContainerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ServiceWorkerContainer: {\n    prototype: ServiceWorkerContainer;\n    new(): ServiceWorkerContainer;\n};\n\ninterface ServiceWorkerMessageEvent extends Event {\n    readonly data: any;\n    readonly lastEventId: string;\n    readonly origin: string;\n    readonly ports: ReadonlyArray<MessagePort> | null;\n    readonly source: ServiceWorker | MessagePort | null;\n}\n\ndeclare var ServiceWorkerMessageEvent: {\n    prototype: ServiceWorkerMessageEvent;\n    new(type: string, eventInitDict?: ServiceWorkerMessageEventInit): ServiceWorkerMessageEvent;\n};\n\ninterface ServiceWorkerRegistrationEventMap {\n    "updatefound": Event;\n}\n\ninterface ServiceWorkerRegistration extends EventTarget {\n    readonly active: ServiceWorker | null;\n    readonly installing: ServiceWorker | null;\n    readonly navigationPreload: NavigationPreloadManager;\n    onupdatefound: ((this: ServiceWorkerRegistration, ev: Event) => any) | null;\n    readonly pushManager: PushManager;\n    readonly scope: string;\n    readonly sync: SyncManager;\n    readonly updateViaCache: ServiceWorkerUpdateViaCache;\n    readonly waiting: ServiceWorker | null;\n    getNotifications(filter?: GetNotificationOptions): Promise<Notification[]>;\n    showNotification(title: string, options?: NotificationOptions): Promise<void>;\n    unregister(): Promise<boolean>;\n    update(): Promise<void>;\n    addEventListener<K extends keyof ServiceWorkerRegistrationEventMap>(type: K, listener: (this: ServiceWorkerRegistration, ev: ServiceWorkerRegistrationEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ServiceWorkerRegistrationEventMap>(type: K, listener: (this: ServiceWorkerRegistration, ev: ServiceWorkerRegistrationEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ServiceWorkerRegistration: {\n    prototype: ServiceWorkerRegistration;\n    new(): ServiceWorkerRegistration;\n};\n\ninterface ShadowRoot extends DocumentOrShadowRoot, DocumentFragment, DocumentOrShadowRoot {\n    readonly host: Element;\n    innerHTML: string;\n    readonly mode: ShadowRootMode;\n}\n\ninterface ShadowRootInit {\n    delegatesFocus?: boolean;\n    mode: "open" | "closed";\n}\n\ninterface Slotable {\n    readonly assignedSlot: HTMLSlotElement | null;\n}\n\ninterface SourceBuffer extends EventTarget {\n    appendWindowEnd: number;\n    appendWindowStart: number;\n    readonly audioTracks: AudioTrackList;\n    readonly buffered: TimeRanges;\n    mode: AppendMode;\n    readonly textTracks: TextTrackList;\n    timestampOffset: number;\n    readonly updating: boolean;\n    readonly videoTracks: VideoTrackList;\n    abort(): void;\n    appendBuffer(data: ArrayBuffer | Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | null): void;\n    appendStream(stream: MSStream, maxSize?: number): void;\n    remove(start: number, end: number): void;\n}\n\ndeclare var SourceBuffer: {\n    prototype: SourceBuffer;\n    new(): SourceBuffer;\n};\n\ninterface SourceBufferList extends EventTarget {\n    readonly length: number;\n    item(index: number): SourceBuffer;\n    [index: number]: SourceBuffer;\n}\n\ndeclare var SourceBufferList: {\n    prototype: SourceBufferList;\n    new(): SourceBufferList;\n};\n\ninterface SpeechGrammar {\n    src: string;\n    weight: number;\n}\n\ndeclare var SpeechGrammar: {\n    prototype: SpeechGrammar;\n    new(): SpeechGrammar;\n};\n\ninterface SpeechGrammarList {\n    readonly length: number;\n    addFromString(string: string, weight?: number): void;\n    addFromURI(src: string, weight?: number): void;\n    item(index: number): SpeechGrammar;\n    [index: number]: SpeechGrammar;\n}\n\ndeclare var SpeechGrammarList: {\n    prototype: SpeechGrammarList;\n    new(): SpeechGrammarList;\n};\n\ninterface SpeechRecognitionEventMap {\n    "audioend": Event;\n    "audiostart": Event;\n    "end": Event;\n    "error": SpeechRecognitionError;\n    "nomatch": SpeechRecognitionEvent;\n    "result": SpeechRecognitionEvent;\n    "soundend": Event;\n    "soundstart": Event;\n    "speechend": Event;\n    "speechstart": Event;\n    "start": Event;\n}\n\ninterface SpeechRecognition extends EventTarget {\n    continuous: boolean;\n    grammars: SpeechGrammarList;\n    interimResults: boolean;\n    lang: string;\n    maxAlternatives: number;\n    onaudioend: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onaudiostart: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onend: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onerror: ((this: SpeechRecognition, ev: SpeechRecognitionError) => any) | null;\n    onnomatch: ((this: SpeechRecognition, ev: SpeechRecognitionEvent) => any) | null;\n    onresult: ((this: SpeechRecognition, ev: SpeechRecognitionEvent) => any) | null;\n    onsoundend: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onsoundstart: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onspeechend: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onspeechstart: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onstart: ((this: SpeechRecognition, ev: Event) => any) | null;\n    serviceURI: string;\n    abort(): void;\n    start(): void;\n    stop(): void;\n    addEventListener<K extends keyof SpeechRecognitionEventMap>(type: K, listener: (this: SpeechRecognition, ev: SpeechRecognitionEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SpeechRecognitionEventMap>(type: K, listener: (this: SpeechRecognition, ev: SpeechRecognitionEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SpeechRecognition: {\n    prototype: SpeechRecognition;\n    new(): SpeechRecognition;\n};\n\ninterface SpeechRecognitionAlternative {\n    readonly confidence: number;\n    readonly transcript: string;\n}\n\ndeclare var SpeechRecognitionAlternative: {\n    prototype: SpeechRecognitionAlternative;\n    new(): SpeechRecognitionAlternative;\n};\n\ninterface SpeechRecognitionError extends Event {\n    readonly error: SpeechRecognitionErrorCode;\n    readonly message: string;\n}\n\ndeclare var SpeechRecognitionError: {\n    prototype: SpeechRecognitionError;\n    new(): SpeechRecognitionError;\n};\n\ninterface SpeechRecognitionEvent extends Event {\n    readonly emma: Document | null;\n    readonly interpretation: any;\n    readonly resultIndex: number;\n    readonly results: SpeechRecognitionResultList;\n}\n\ndeclare var SpeechRecognitionEvent: {\n    prototype: SpeechRecognitionEvent;\n    new(): SpeechRecognitionEvent;\n};\n\ninterface SpeechRecognitionResult {\n    readonly isFinal: boolean;\n    readonly length: number;\n    item(index: number): SpeechRecognitionAlternative;\n    [index: number]: SpeechRecognitionAlternative;\n}\n\ndeclare var SpeechRecognitionResult: {\n    prototype: SpeechRecognitionResult;\n    new(): SpeechRecognitionResult;\n};\n\ninterface SpeechRecognitionResultList {\n    readonly length: number;\n    item(index: number): SpeechRecognitionResult;\n    [index: number]: SpeechRecognitionResult;\n}\n\ndeclare var SpeechRecognitionResultList: {\n    prototype: SpeechRecognitionResultList;\n    new(): SpeechRecognitionResultList;\n};\n\ninterface SpeechSynthesisEventMap {\n    "voiceschanged": Event;\n}\n\ninterface SpeechSynthesis extends EventTarget {\n    onvoiceschanged: ((this: SpeechSynthesis, ev: Event) => any) | null;\n    readonly paused: boolean;\n    readonly pending: boolean;\n    readonly speaking: boolean;\n    cancel(): void;\n    getVoices(): SpeechSynthesisVoice[];\n    pause(): void;\n    resume(): void;\n    speak(utterance: SpeechSynthesisUtterance): void;\n    addEventListener<K extends keyof SpeechSynthesisEventMap>(type: K, listener: (this: SpeechSynthesis, ev: SpeechSynthesisEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SpeechSynthesisEventMap>(type: K, listener: (this: SpeechSynthesis, ev: SpeechSynthesisEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SpeechSynthesis: {\n    prototype: SpeechSynthesis;\n    new(): SpeechSynthesis;\n};\n\ninterface SpeechSynthesisErrorEvent extends SpeechSynthesisEvent {\n    readonly error: SpeechSynthesisErrorCode;\n}\n\ndeclare var SpeechSynthesisErrorEvent: {\n    prototype: SpeechSynthesisErrorEvent;\n    new(): SpeechSynthesisErrorEvent;\n};\n\ninterface SpeechSynthesisEvent extends Event {\n    readonly charIndex: number;\n    readonly elapsedTime: number;\n    readonly name: string;\n    readonly utterance: SpeechSynthesisUtterance;\n}\n\ndeclare var SpeechSynthesisEvent: {\n    prototype: SpeechSynthesisEvent;\n    new(): SpeechSynthesisEvent;\n};\n\ninterface SpeechSynthesisUtteranceEventMap {\n    "boundary": SpeechSynthesisEvent;\n    "end": SpeechSynthesisEvent;\n    "error": SpeechSynthesisErrorEvent;\n    "mark": SpeechSynthesisEvent;\n    "pause": SpeechSynthesisEvent;\n    "resume": SpeechSynthesisEvent;\n    "start": SpeechSynthesisEvent;\n}\n\ninterface SpeechSynthesisUtterance extends EventTarget {\n    lang: string;\n    onboundary: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    onend: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    onerror: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisErrorEvent) => any) | null;\n    onmark: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    onpause: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    onresume: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    onstart: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    pitch: number;\n    rate: number;\n    text: string;\n    voice: SpeechSynthesisVoice;\n    volume: number;\n    addEventListener<K extends keyof SpeechSynthesisUtteranceEventMap>(type: K, listener: (this: SpeechSynthesisUtterance, ev: SpeechSynthesisUtteranceEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SpeechSynthesisUtteranceEventMap>(type: K, listener: (this: SpeechSynthesisUtterance, ev: SpeechSynthesisUtteranceEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SpeechSynthesisUtterance: {\n    prototype: SpeechSynthesisUtterance;\n    new(): SpeechSynthesisUtterance;\n    new(text: string): SpeechSynthesisUtterance;\n};\n\ninterface SpeechSynthesisVoice {\n    readonly default: boolean;\n    readonly lang: string;\n    readonly localService: boolean;\n    readonly name: string;\n    readonly voiceURI: string;\n}\n\ndeclare var SpeechSynthesisVoice: {\n    prototype: SpeechSynthesisVoice;\n    new(): SpeechSynthesisVoice;\n};\n\ninterface StaticRange extends AbstractRange {\n}\n\ndeclare var StaticRange: {\n    prototype: StaticRange;\n    new(): StaticRange;\n};\n\ninterface StereoPannerNode extends AudioNode {\n    readonly pan: AudioParam;\n}\n\ndeclare var StereoPannerNode: {\n    prototype: StereoPannerNode;\n    new(context: BaseAudioContext, options?: StereoPannerOptions): StereoPannerNode;\n};\n\ninterface Storage {\n    /**\n     * Returns the number of key/value pairs currently present in the list associated with the\n     * object.\n     */\n    readonly length: number;\n    /**\n     * Empties the list associated with the object of all key/value pairs, if there are any.\n     */\n    clear(): void;\n    /**\n     * value = storage[key]\n     */\n    getItem(key: string): string | null;\n    /**\n     * Returns the name of the nth key in the list, or null if n is greater\n     * than or equal to the number of key/value pairs in the object.\n     */\n    key(index: number): string | null;\n    /**\n     * delete storage[key]\n     */\n    removeItem(key: string): void;\n    /**\n     * storage[key] = value\n     */\n    setItem(key: string, value: string): void;\n    [name: string]: any;\n}\n\ndeclare var Storage: {\n    prototype: Storage;\n    new(): Storage;\n};\n\ninterface StorageEvent extends Event {\n    /**\n     * Returns the key of the storage item being changed.\n     */\n    readonly key: string | null;\n    /**\n     * Returns the new value of the key of the storage item whose value is being changed.\n     */\n    readonly newValue: string | null;\n    /**\n     * Returns the old value of the key of the storage item whose value is being changed.\n     */\n    readonly oldValue: string | null;\n    /**\n     * Returns the Storage object that was affected.\n     */\n    readonly storageArea: Storage | null;\n    /**\n     * Returns the URL of the document whose storage item changed.\n     */\n    readonly url: string;\n}\n\ndeclare var StorageEvent: {\n    prototype: StorageEvent;\n    new(type: string, eventInitDict?: StorageEventInit): StorageEvent;\n};\n\ninterface StorageManager {\n    estimate(): Promise<StorageEstimate>;\n    persist(): Promise<boolean>;\n    persisted(): Promise<boolean>;\n}\n\ndeclare var StorageManager: {\n    prototype: StorageManager;\n    new(): StorageManager;\n};\n\ninterface StyleMedia {\n    readonly type: string;\n    matchMedium(mediaquery: string): boolean;\n}\n\ndeclare var StyleMedia: {\n    prototype: StyleMedia;\n    new(): StyleMedia;\n};\n\ninterface StyleSheet {\n    disabled: boolean;\n    readonly href: string | null;\n    readonly media: MediaList;\n    readonly ownerNode: Node;\n    readonly parentStyleSheet: StyleSheet | null;\n    readonly title: string | null;\n    readonly type: string;\n}\n\ndeclare var StyleSheet: {\n    prototype: StyleSheet;\n    new(): StyleSheet;\n};\n\ninterface StyleSheetList {\n    readonly length: number;\n    item(index: number): StyleSheet | null;\n    [index: number]: StyleSheet;\n}\n\ndeclare var StyleSheetList: {\n    prototype: StyleSheetList;\n    new(): StyleSheetList;\n};\n\ninterface SubtleCrypto {\n    decrypt(algorithm: string | RsaOaepParams | AesCtrParams | AesCbcParams | AesCmacParams | AesGcmParams | AesCfbParams, key: CryptoKey, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): PromiseLike<ArrayBuffer>;\n    deriveBits(algorithm: string | EcdhKeyDeriveParams | DhKeyDeriveParams | ConcatParams | HkdfCtrParams | Pbkdf2Params, baseKey: CryptoKey, length: number): PromiseLike<ArrayBuffer>;\n    deriveKey(algorithm: string | EcdhKeyDeriveParams | DhKeyDeriveParams | ConcatParams | HkdfCtrParams | Pbkdf2Params, baseKey: CryptoKey, derivedKeyType: string | AesDerivedKeyParams | HmacImportParams | ConcatParams | HkdfCtrParams | Pbkdf2Params, extractable: boolean, keyUsages: string[]): PromiseLike<CryptoKey>;\n    digest(algorithm: string | Algorithm, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): PromiseLike<ArrayBuffer>;\n    encrypt(algorithm: string | RsaOaepParams | AesCtrParams | AesCbcParams | AesCmacParams | AesGcmParams | AesCfbParams, key: CryptoKey, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): PromiseLike<ArrayBuffer>;\n    exportKey(format: "jwk", key: CryptoKey): PromiseLike<JsonWebKey>;\n    exportKey(format: "raw" | "pkcs8" | "spki", key: CryptoKey): PromiseLike<ArrayBuffer>;\n    exportKey(format: string, key: CryptoKey): PromiseLike<JsonWebKey | ArrayBuffer>;\n    generateKey(algorithm: string, extractable: boolean, keyUsages: string[]): PromiseLike<CryptoKeyPair | CryptoKey>;\n    generateKey(algorithm: RsaHashedKeyGenParams | EcKeyGenParams | DhKeyGenParams, extractable: boolean, keyUsages: string[]): PromiseLike<CryptoKeyPair>;\n    generateKey(algorithm: AesKeyGenParams | HmacKeyGenParams | Pbkdf2Params, extractable: boolean, keyUsages: string[]): PromiseLike<CryptoKey>;\n    importKey(format: "jwk", keyData: JsonWebKey, algorithm: string | RsaHashedImportParams | EcKeyImportParams | HmacImportParams | DhImportKeyParams | AesKeyAlgorithm, extractable: boolean, keyUsages: string[]): PromiseLike<CryptoKey>;\n    importKey(format: "raw" | "pkcs8" | "spki", keyData: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, algorithm: string | RsaHashedImportParams | EcKeyImportParams | HmacImportParams | DhImportKeyParams | AesKeyAlgorithm, extractable: boolean, keyUsages: string[]): PromiseLike<CryptoKey>;\n    importKey(format: string, keyData: JsonWebKey | Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, algorithm: string | RsaHashedImportParams | EcKeyImportParams | HmacImportParams | DhImportKeyParams | AesKeyAlgorithm, extractable: boolean, keyUsages: string[]): PromiseLike<CryptoKey>;\n    sign(algorithm: string | RsaPssParams | EcdsaParams | AesCmacParams, key: CryptoKey, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): PromiseLike<ArrayBuffer>;\n    unwrapKey(format: string, wrappedKey: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, unwrappingKey: CryptoKey, unwrapAlgorithm: string | Algorithm, unwrappedKeyAlgorithm: string | Algorithm, extractable: boolean, keyUsages: string[]): PromiseLike<CryptoKey>;\n    verify(algorithm: string | RsaPssParams | EcdsaParams | AesCmacParams, key: CryptoKey, signature: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): PromiseLike<boolean>;\n    wrapKey(format: string, key: CryptoKey, wrappingKey: CryptoKey, wrapAlgorithm: string | Algorithm): PromiseLike<ArrayBuffer>;\n}\n\ndeclare var SubtleCrypto: {\n    prototype: SubtleCrypto;\n    new(): SubtleCrypto;\n};\n\ninterface SyncManager {\n    getTags(): Promise<string[]>;\n    register(tag: string): Promise<void>;\n}\n\ndeclare var SyncManager: {\n    prototype: SyncManager;\n    new(): SyncManager;\n};\n\ninterface Text extends CharacterData, Slotable {\n    readonly assignedSlot: HTMLSlotElement | null;\n    /**\n     * Returns the combined data of all direct Text node siblings.\n     */\n    readonly wholeText: string;\n    /**\n     * Splits data at the given offset and returns the remainder as Text node.\n     */\n    splitText(offset: number): Text;\n}\n\ndeclare var Text: {\n    prototype: Text;\n    new(data?: string): Text;\n};\n\ninterface TextDecoder {\n    /**\n     * Returns encoding\'s name, lowercased.\n     */\n    readonly encoding: string;\n    /**\n     * Returns true if error mode is "fatal", and false\n     * otherwise.\n     */\n    readonly fatal: boolean;\n    /**\n     * Returns true if ignore BOM flag is set, and false otherwise.\n     */\n    readonly ignoreBOM: boolean;\n    /**\n     * Returns the result of running encoding\'s decoder. The\n     * method can be invoked zero or more times with options\'s stream set to\n     * true, and then once without options\'s stream (or set to false), to process\n     * a fragmented stream. If the invocation without options\'s stream (or set to\n     * false) has no input, it\'s clearest to omit both arguments.\n     * var string = "", decoder = new TextDecoder(encoding), buffer;\n     * while(buffer = next_chunk()) {\n     * string += decoder.decode(buffer, {stream:true});\n     * }\n     * string += decoder.decode(); // end-of-stream\n     * If the error mode is "fatal" and encoding\'s decoder returns error, throws a TypeError.\n     */\n    decode(input?: BufferSource, options?: TextDecodeOptions): string;\n}\n\ndeclare var TextDecoder: {\n    prototype: TextDecoder;\n    new(label?: string, options?: TextDecoderOptions): TextDecoder;\n};\n\ninterface TextEncoder {\n    /**\n     * Returns "utf-8".\n     */\n    readonly encoding: string;\n    /**\n     * Returns the result of running UTF-8\'s encoder.\n     */\n    encode(input?: string): Uint8Array;\n}\n\ndeclare var TextEncoder: {\n    prototype: TextEncoder;\n    new(): TextEncoder;\n};\n\ninterface TextEvent extends UIEvent {\n    readonly data: string;\n    initTextEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, dataArg: string, inputMethod: number, locale: string): void;\n    readonly DOM_INPUT_METHOD_DROP: number;\n    readonly DOM_INPUT_METHOD_HANDWRITING: number;\n    readonly DOM_INPUT_METHOD_IME: number;\n    readonly DOM_INPUT_METHOD_KEYBOARD: number;\n    readonly DOM_INPUT_METHOD_MULTIMODAL: number;\n    readonly DOM_INPUT_METHOD_OPTION: number;\n    readonly DOM_INPUT_METHOD_PASTE: number;\n    readonly DOM_INPUT_METHOD_SCRIPT: number;\n    readonly DOM_INPUT_METHOD_UNKNOWN: number;\n    readonly DOM_INPUT_METHOD_VOICE: number;\n}\n\ndeclare var TextEvent: {\n    prototype: TextEvent;\n    new(): TextEvent;\n    readonly DOM_INPUT_METHOD_DROP: number;\n    readonly DOM_INPUT_METHOD_HANDWRITING: number;\n    readonly DOM_INPUT_METHOD_IME: number;\n    readonly DOM_INPUT_METHOD_KEYBOARD: number;\n    readonly DOM_INPUT_METHOD_MULTIMODAL: number;\n    readonly DOM_INPUT_METHOD_OPTION: number;\n    readonly DOM_INPUT_METHOD_PASTE: number;\n    readonly DOM_INPUT_METHOD_SCRIPT: number;\n    readonly DOM_INPUT_METHOD_UNKNOWN: number;\n    readonly DOM_INPUT_METHOD_VOICE: number;\n};\n\ninterface TextMetrics {\n    readonly actualBoundingBoxAscent: number;\n    readonly actualBoundingBoxDescent: number;\n    readonly actualBoundingBoxLeft: number;\n    readonly actualBoundingBoxRight: number;\n    readonly alphabeticBaseline: number;\n    readonly emHeightAscent: number;\n    readonly emHeightDescent: number;\n    readonly fontBoundingBoxAscent: number;\n    readonly fontBoundingBoxDescent: number;\n    readonly hangingBaseline: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly ideographicBaseline: number;\n    readonly width: number;\n}\n\ndeclare var TextMetrics: {\n    prototype: TextMetrics;\n    new(): TextMetrics;\n};\n\ninterface TextTrackEventMap {\n    "cuechange": Event;\n    "error": Event;\n    "load": Event;\n}\n\ninterface TextTrack extends EventTarget {\n    readonly activeCues: TextTrackCueList;\n    readonly cues: TextTrackCueList;\n    readonly inBandMetadataTrackDispatchType: string;\n    readonly kind: string;\n    readonly label: string;\n    readonly language: string;\n    mode: TextTrackMode | number;\n    oncuechange: ((this: TextTrack, ev: Event) => any) | null;\n    onerror: ((this: TextTrack, ev: Event) => any) | null;\n    onload: ((this: TextTrack, ev: Event) => any) | null;\n    readonly readyState: number;\n    addCue(cue: TextTrackCue): void;\n    removeCue(cue: TextTrackCue): void;\n    readonly DISABLED: number;\n    readonly ERROR: number;\n    readonly HIDDEN: number;\n    readonly LOADED: number;\n    readonly LOADING: number;\n    readonly NONE: number;\n    readonly SHOWING: number;\n    addEventListener<K extends keyof TextTrackEventMap>(type: K, listener: (this: TextTrack, ev: TextTrackEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof TextTrackEventMap>(type: K, listener: (this: TextTrack, ev: TextTrackEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var TextTrack: {\n    prototype: TextTrack;\n    new(): TextTrack;\n    readonly DISABLED: number;\n    readonly ERROR: number;\n    readonly HIDDEN: number;\n    readonly LOADED: number;\n    readonly LOADING: number;\n    readonly NONE: number;\n    readonly SHOWING: number;\n};\n\ninterface TextTrackCueEventMap {\n    "enter": Event;\n    "exit": Event;\n}\n\ninterface TextTrackCue extends EventTarget {\n    endTime: number;\n    id: string;\n    onenter: ((this: TextTrackCue, ev: Event) => any) | null;\n    onexit: ((this: TextTrackCue, ev: Event) => any) | null;\n    pauseOnExit: boolean;\n    startTime: number;\n    text: string;\n    readonly track: TextTrack;\n    getCueAsHTML(): DocumentFragment;\n    addEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: TextTrackCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: TextTrackCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var TextTrackCue: {\n    prototype: TextTrackCue;\n    new(startTime: number, endTime: number, text: string): TextTrackCue;\n};\n\ninterface TextTrackCueList {\n    readonly length: number;\n    getCueById(id: string): TextTrackCue;\n    item(index: number): TextTrackCue;\n    [index: number]: TextTrackCue;\n}\n\ndeclare var TextTrackCueList: {\n    prototype: TextTrackCueList;\n    new(): TextTrackCueList;\n};\n\ninterface TextTrackListEventMap {\n    "addtrack": TrackEvent;\n}\n\ninterface TextTrackList extends EventTarget {\n    readonly length: number;\n    onaddtrack: ((this: TextTrackList, ev: TrackEvent) => any) | null;\n    item(index: number): TextTrack;\n    addEventListener<K extends keyof TextTrackListEventMap>(type: K, listener: (this: TextTrackList, ev: TextTrackListEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof TextTrackListEventMap>(type: K, listener: (this: TextTrackList, ev: TextTrackListEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n    [index: number]: TextTrack;\n}\n\ndeclare var TextTrackList: {\n    prototype: TextTrackList;\n    new(): TextTrackList;\n};\n\ninterface TimeRanges {\n    readonly length: number;\n    end(index: number): number;\n    start(index: number): number;\n}\n\ndeclare var TimeRanges: {\n    prototype: TimeRanges;\n    new(): TimeRanges;\n};\n\ninterface Touch {\n    readonly altitudeAngle: number;\n    readonly azimuthAngle: number;\n    readonly clientX: number;\n    readonly clientY: number;\n    readonly force: number;\n    readonly identifier: number;\n    readonly pageX: number;\n    readonly pageY: number;\n    readonly radiusX: number;\n    readonly radiusY: number;\n    readonly rotationAngle: number;\n    readonly screenX: number;\n    readonly screenY: number;\n    readonly target: EventTarget;\n    readonly touchType: TouchType;\n}\n\ndeclare var Touch: {\n    prototype: Touch;\n    new(touchInitDict: TouchInit): Touch;\n};\n\ninterface TouchEvent extends UIEvent {\n    readonly altKey: boolean;\n    readonly changedTouches: TouchList;\n    readonly ctrlKey: boolean;\n    readonly metaKey: boolean;\n    readonly shiftKey: boolean;\n    readonly targetTouches: TouchList;\n    readonly touches: TouchList;\n}\n\ndeclare var TouchEvent: {\n    prototype: TouchEvent;\n    new(type: string, eventInitDict?: TouchEventInit): TouchEvent;\n};\n\ninterface TouchList {\n    readonly length: number;\n    item(index: number): Touch | null;\n    [index: number]: Touch;\n}\n\ndeclare var TouchList: {\n    prototype: TouchList;\n    new(): TouchList;\n};\n\ninterface TrackEvent extends Event {\n    readonly track: VideoTrack | AudioTrack | TextTrack | null;\n}\n\ndeclare var TrackEvent: {\n    prototype: TrackEvent;\n    new(typeArg: string, eventInitDict?: TrackEventInit): TrackEvent;\n};\n\ninterface TransformStream<I = any, O = any> {\n    readonly readable: ReadableStream<O>;\n    readonly writable: WritableStream<I>;\n}\n\ndeclare var TransformStream: {\n    prototype: TransformStream;\n    new<I = any, O = any>(transformer?: Transformer<I, O>, writableStrategy?: QueuingStrategy<I>, readableStrategy?: QueuingStrategy<O>): TransformStream<I, O>;\n};\n\ninterface TransformStreamDefaultController<O = any> {\n    readonly desiredSize: number | null;\n    enqueue(chunk: O): void;\n    error(reason?: any): void;\n    terminate(): void;\n}\n\ninterface TransitionEvent extends Event {\n    readonly elapsedTime: number;\n    readonly propertyName: string;\n    readonly pseudoElement: string;\n}\n\ndeclare var TransitionEvent: {\n    prototype: TransitionEvent;\n    new(type: string, transitionEventInitDict?: TransitionEventInit): TransitionEvent;\n};\n\ninterface TreeWalker {\n    currentNode: Node;\n    readonly filter: NodeFilter | null;\n    readonly root: Node;\n    readonly whatToShow: number;\n    firstChild(): Node | null;\n    lastChild(): Node | null;\n    nextNode(): Node | null;\n    nextSibling(): Node | null;\n    parentNode(): Node | null;\n    previousNode(): Node | null;\n    previousSibling(): Node | null;\n}\n\ndeclare var TreeWalker: {\n    prototype: TreeWalker;\n    new(): TreeWalker;\n};\n\ninterface UIEvent extends Event {\n    readonly detail: number;\n    readonly view: Window;\n    initUIEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, detailArg: number): void;\n}\n\ndeclare var UIEvent: {\n    prototype: UIEvent;\n    new(typeArg: string, eventInitDict?: UIEventInit): UIEvent;\n};\n\ninterface URL {\n    hash: string;\n    host: string;\n    hostname: string;\n    href: string;\n    readonly origin: string;\n    password: string;\n    pathname: string;\n    port: string;\n    protocol: string;\n    search: string;\n    readonly searchParams: URLSearchParams;\n    username: string;\n    toJSON(): string;\n}\n\ndeclare var URL: {\n    prototype: URL;\n    new(url: string, base?: string | URL): URL;\n    createObjectURL(object: any): string;\n    revokeObjectURL(url: string): void;\n};\n\ntype webkitURL = URL;\ndeclare var webkitURL: typeof URL;\n\ninterface URLSearchParams {\n    /**\n     * Appends a specified key/value pair as a new search parameter.\n     */\n    append(name: string, value: string): void;\n    /**\n     * Deletes the given search parameter, and its associated value, from the list of all search parameters.\n     */\n    delete(name: string): void;\n    /**\n     * Returns the first value associated to the given search parameter.\n     */\n    get(name: string): string | null;\n    /**\n     * Returns all the values association with a given search parameter.\n     */\n    getAll(name: string): string[];\n    /**\n     * Returns a Boolean indicating if such a search parameter exists.\n     */\n    has(name: string): boolean;\n    /**\n     * Sets the value associated to a given search parameter to the given value. If there were several values, delete the others.\n     */\n    set(name: string, value: string): void;\n    sort(): void;\n    forEach(callbackfn: (value: string, key: string, parent: URLSearchParams) => void, thisArg?: any): void;\n}\n\ndeclare var URLSearchParams: {\n    prototype: URLSearchParams;\n    new(init?: string[][] | Record<string, string> | string | URLSearchParams): URLSearchParams;\n};\n\ninterface VRDisplay extends EventTarget {\n    readonly capabilities: VRDisplayCapabilities;\n    depthFar: number;\n    depthNear: number;\n    readonly displayId: number;\n    readonly displayName: string;\n    readonly isConnected: boolean;\n    readonly isPresenting: boolean;\n    readonly stageParameters: VRStageParameters | null;\n    cancelAnimationFrame(handle: number): void;\n    exitPresent(): Promise<void>;\n    getEyeParameters(whichEye: string): VREyeParameters;\n    getFrameData(frameData: VRFrameData): boolean;\n    getLayers(): VRLayer[];\n    /** @deprecated */\n    getPose(): VRPose;\n    requestAnimationFrame(callback: FrameRequestCallback): number;\n    requestPresent(layers: VRLayer[]): Promise<void>;\n    resetPose(): void;\n    submitFrame(pose?: VRPose): void;\n}\n\ndeclare var VRDisplay: {\n    prototype: VRDisplay;\n    new(): VRDisplay;\n};\n\ninterface VRDisplayCapabilities {\n    readonly canPresent: boolean;\n    readonly hasExternalDisplay: boolean;\n    readonly hasOrientation: boolean;\n    readonly hasPosition: boolean;\n    readonly maxLayers: number;\n}\n\ndeclare var VRDisplayCapabilities: {\n    prototype: VRDisplayCapabilities;\n    new(): VRDisplayCapabilities;\n};\n\ninterface VRDisplayEvent extends Event {\n    readonly display: VRDisplay;\n    readonly reason: VRDisplayEventReason | null;\n}\n\ndeclare var VRDisplayEvent: {\n    prototype: VRDisplayEvent;\n    new(type: string, eventInitDict: VRDisplayEventInit): VRDisplayEvent;\n};\n\ninterface VREyeParameters {\n    /** @deprecated */\n    readonly fieldOfView: VRFieldOfView;\n    readonly offset: Float32Array;\n    readonly renderHeight: number;\n    readonly renderWidth: number;\n}\n\ndeclare var VREyeParameters: {\n    prototype: VREyeParameters;\n    new(): VREyeParameters;\n};\n\ninterface VRFieldOfView {\n    readonly downDegrees: number;\n    readonly leftDegrees: number;\n    readonly rightDegrees: number;\n    readonly upDegrees: number;\n}\n\ndeclare var VRFieldOfView: {\n    prototype: VRFieldOfView;\n    new(): VRFieldOfView;\n};\n\ninterface VRFrameData {\n    readonly leftProjectionMatrix: Float32Array;\n    readonly leftViewMatrix: Float32Array;\n    readonly pose: VRPose;\n    readonly rightProjectionMatrix: Float32Array;\n    readonly rightViewMatrix: Float32Array;\n    readonly timestamp: number;\n}\n\ndeclare var VRFrameData: {\n    prototype: VRFrameData;\n    new(): VRFrameData;\n};\n\ninterface VRPose {\n    readonly angularAcceleration: Float32Array | null;\n    readonly angularVelocity: Float32Array | null;\n    readonly linearAcceleration: Float32Array | null;\n    readonly linearVelocity: Float32Array | null;\n    readonly orientation: Float32Array | null;\n    readonly position: Float32Array | null;\n    readonly timestamp: number;\n}\n\ndeclare var VRPose: {\n    prototype: VRPose;\n    new(): VRPose;\n};\n\ninterface VTTCue extends TextTrackCue {\n    align: AlignSetting;\n    line: LineAndPositionSetting;\n    lineAlign: LineAlignSetting;\n    position: LineAndPositionSetting;\n    positionAlign: PositionAlignSetting;\n    region: VTTRegion | null;\n    size: number;\n    snapToLines: boolean;\n    text: string;\n    vertical: DirectionSetting;\n    getCueAsHTML(): DocumentFragment;\n    addEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: VTTCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: VTTCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var VTTCue: {\n    prototype: VTTCue;\n    new(startTime: number, endTime: number, text: string): VTTCue;\n};\n\ninterface VTTRegion {\n    id: string;\n    lines: number;\n    regionAnchorX: number;\n    regionAnchorY: number;\n    scroll: ScrollSetting;\n    viewportAnchorX: number;\n    viewportAnchorY: number;\n    width: number;\n}\n\ndeclare var VTTRegion: {\n    prototype: VTTRegion;\n    new(): VTTRegion;\n};\n\ninterface ValidityState {\n    readonly badInput: boolean;\n    readonly customError: boolean;\n    readonly patternMismatch: boolean;\n    readonly rangeOverflow: boolean;\n    readonly rangeUnderflow: boolean;\n    readonly stepMismatch: boolean;\n    readonly tooLong: boolean;\n    readonly tooShort: boolean;\n    readonly typeMismatch: boolean;\n    readonly valid: boolean;\n    readonly valueMissing: boolean;\n}\n\ndeclare var ValidityState: {\n    prototype: ValidityState;\n    new(): ValidityState;\n};\n\ninterface VideoPlaybackQuality {\n    readonly corruptedVideoFrames: number;\n    readonly creationTime: number;\n    readonly droppedVideoFrames: number;\n    readonly totalFrameDelay: number;\n    readonly totalVideoFrames: number;\n}\n\ndeclare var VideoPlaybackQuality: {\n    prototype: VideoPlaybackQuality;\n    new(): VideoPlaybackQuality;\n};\n\ninterface VideoTrack {\n    readonly id: string;\n    kind: string;\n    readonly label: string;\n    language: string;\n    selected: boolean;\n    readonly sourceBuffer: SourceBuffer;\n}\n\ndeclare var VideoTrack: {\n    prototype: VideoTrack;\n    new(): VideoTrack;\n};\n\ninterface VideoTrackListEventMap {\n    "addtrack": TrackEvent;\n    "change": Event;\n    "removetrack": TrackEvent;\n}\n\ninterface VideoTrackList extends EventTarget {\n    readonly length: number;\n    onaddtrack: ((this: VideoTrackList, ev: TrackEvent) => any) | null;\n    onchange: ((this: VideoTrackList, ev: Event) => any) | null;\n    onremovetrack: ((this: VideoTrackList, ev: TrackEvent) => any) | null;\n    readonly selectedIndex: number;\n    getTrackById(id: string): VideoTrack | null;\n    item(index: number): VideoTrack;\n    addEventListener<K extends keyof VideoTrackListEventMap>(type: K, listener: (this: VideoTrackList, ev: VideoTrackListEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof VideoTrackListEventMap>(type: K, listener: (this: VideoTrackList, ev: VideoTrackListEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n    [index: number]: VideoTrack;\n}\n\ndeclare var VideoTrackList: {\n    prototype: VideoTrackList;\n    new(): VideoTrackList;\n};\n\ninterface WEBGL_color_buffer_float {\n    readonly FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE_EXT: GLenum;\n    readonly RGBA32F_EXT: GLenum;\n    readonly UNSIGNED_NORMALIZED_EXT: GLenum;\n}\n\ninterface WEBGL_compressed_texture_astc {\n    getSupportedProfiles(): string[];\n    readonly COMPRESSED_RGBA_ASTC_10x10_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_10x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_10x6_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_10x8_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_12x10_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_12x12_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_4x4_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_5x4_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_5x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_6x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_6x6_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_8x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_8x6_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_8x8_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x10_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x6_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x8_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_12x10_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_12x12_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_4x4_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_5x4_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_5x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_6x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_6x6_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_8x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_8x6_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_8x8_KHR: GLenum;\n}\n\ninterface WEBGL_compressed_texture_s3tc {\n    readonly COMPRESSED_RGBA_S3TC_DXT1_EXT: GLenum;\n    readonly COMPRESSED_RGBA_S3TC_DXT3_EXT: GLenum;\n    readonly COMPRESSED_RGBA_S3TC_DXT5_EXT: GLenum;\n    readonly COMPRESSED_RGB_S3TC_DXT1_EXT: GLenum;\n}\n\ninterface WEBGL_compressed_texture_s3tc_srgb {\n    readonly COMPRESSED_SRGB_ALPHA_S3TC_DXT1_EXT: GLenum;\n    readonly COMPRESSED_SRGB_ALPHA_S3TC_DXT3_EXT: GLenum;\n    readonly COMPRESSED_SRGB_ALPHA_S3TC_DXT5_EXT: GLenum;\n    readonly COMPRESSED_SRGB_S3TC_DXT1_EXT: GLenum;\n}\n\ninterface WEBGL_debug_renderer_info {\n    readonly UNMASKED_RENDERER_WEBGL: GLenum;\n    readonly UNMASKED_VENDOR_WEBGL: GLenum;\n}\n\ninterface WEBGL_debug_shaders {\n    getTranslatedShaderSource(shader: WebGLShader): string;\n}\n\ninterface WEBGL_depth_texture {\n    readonly UNSIGNED_INT_24_8_WEBGL: GLenum;\n}\n\ninterface WEBGL_draw_buffers {\n    drawBuffersWEBGL(buffers: GLenum[]): void;\n    readonly COLOR_ATTACHMENT0_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT10_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT11_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT12_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT13_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT14_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT15_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT1_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT2_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT3_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT4_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT5_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT6_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT7_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT8_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT9_WEBGL: GLenum;\n    readonly DRAW_BUFFER0_WEBGL: GLenum;\n    readonly DRAW_BUFFER10_WEBGL: GLenum;\n    readonly DRAW_BUFFER11_WEBGL: GLenum;\n    readonly DRAW_BUFFER12_WEBGL: GLenum;\n    readonly DRAW_BUFFER13_WEBGL: GLenum;\n    readonly DRAW_BUFFER14_WEBGL: GLenum;\n    readonly DRAW_BUFFER15_WEBGL: GLenum;\n    readonly DRAW_BUFFER1_WEBGL: GLenum;\n    readonly DRAW_BUFFER2_WEBGL: GLenum;\n    readonly DRAW_BUFFER3_WEBGL: GLenum;\n    readonly DRAW_BUFFER4_WEBGL: GLenum;\n    readonly DRAW_BUFFER5_WEBGL: GLenum;\n    readonly DRAW_BUFFER6_WEBGL: GLenum;\n    readonly DRAW_BUFFER7_WEBGL: GLenum;\n    readonly DRAW_BUFFER8_WEBGL: GLenum;\n    readonly DRAW_BUFFER9_WEBGL: GLenum;\n    readonly MAX_COLOR_ATTACHMENTS_WEBGL: GLenum;\n    readonly MAX_DRAW_BUFFERS_WEBGL: GLenum;\n}\n\ninterface WEBGL_lose_context {\n    loseContext(): void;\n    restoreContext(): void;\n}\n\ninterface WaveShaperNode extends AudioNode {\n    curve: Float32Array | null;\n    oversample: OverSampleType;\n}\n\ndeclare var WaveShaperNode: {\n    prototype: WaveShaperNode;\n    new(context: BaseAudioContext, options?: WaveShaperOptions): WaveShaperNode;\n};\n\ninterface WebAuthentication {\n    getAssertion(assertionChallenge: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null, options?: AssertionOptions): Promise<WebAuthnAssertion>;\n    makeCredential(accountInformation: Account, cryptoParameters: ScopedCredentialParameters[], attestationChallenge: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null, options?: ScopedCredentialOptions): Promise<ScopedCredentialInfo>;\n}\n\ndeclare var WebAuthentication: {\n    prototype: WebAuthentication;\n    new(): WebAuthentication;\n};\n\ninterface WebAuthnAssertion {\n    readonly authenticatorData: ArrayBuffer;\n    readonly clientData: ArrayBuffer;\n    readonly credential: ScopedCredential;\n    readonly signature: ArrayBuffer;\n}\n\ndeclare var WebAuthnAssertion: {\n    prototype: WebAuthnAssertion;\n    new(): WebAuthnAssertion;\n};\n\ninterface WebGLActiveInfo {\n    readonly name: string;\n    readonly size: GLint;\n    readonly type: GLenum;\n}\n\ndeclare var WebGLActiveInfo: {\n    prototype: WebGLActiveInfo;\n    new(): WebGLActiveInfo;\n};\n\ninterface WebGLBuffer extends WebGLObject {\n}\n\ndeclare var WebGLBuffer: {\n    prototype: WebGLBuffer;\n    new(): WebGLBuffer;\n};\n\ninterface WebGLContextEvent extends Event {\n    readonly statusMessage: string;\n}\n\ndeclare var WebGLContextEvent: {\n    prototype: WebGLContextEvent;\n    new(type: string, eventInit?: WebGLContextEventInit): WebGLContextEvent;\n};\n\ninterface WebGLFramebuffer extends WebGLObject {\n}\n\ndeclare var WebGLFramebuffer: {\n    prototype: WebGLFramebuffer;\n    new(): WebGLFramebuffer;\n};\n\ninterface WebGLObject {\n}\n\ndeclare var WebGLObject: {\n    prototype: WebGLObject;\n    new(): WebGLObject;\n};\n\ninterface WebGLProgram extends WebGLObject {\n}\n\ndeclare var WebGLProgram: {\n    prototype: WebGLProgram;\n    new(): WebGLProgram;\n};\n\ninterface WebGLRenderbuffer extends WebGLObject {\n}\n\ndeclare var WebGLRenderbuffer: {\n    prototype: WebGLRenderbuffer;\n    new(): WebGLRenderbuffer;\n};\n\ninterface WebGLRenderingContext extends WebGLRenderingContextBase {\n}\n\ndeclare var WebGLRenderingContext: {\n    prototype: WebGLRenderingContext;\n    new(): WebGLRenderingContext;\n    readonly ACTIVE_ATTRIBUTES: GLenum;\n    readonly ACTIVE_TEXTURE: GLenum;\n    readonly ACTIVE_UNIFORMS: GLenum;\n    readonly ALIASED_LINE_WIDTH_RANGE: GLenum;\n    readonly ALIASED_POINT_SIZE_RANGE: GLenum;\n    readonly ALPHA: GLenum;\n    readonly ALPHA_BITS: GLenum;\n    readonly ALWAYS: GLenum;\n    readonly ARRAY_BUFFER: GLenum;\n    readonly ARRAY_BUFFER_BINDING: GLenum;\n    readonly ATTACHED_SHADERS: GLenum;\n    readonly BACK: GLenum;\n    readonly BLEND: GLenum;\n    readonly BLEND_COLOR: GLenum;\n    readonly BLEND_DST_ALPHA: GLenum;\n    readonly BLEND_DST_RGB: GLenum;\n    readonly BLEND_EQUATION: GLenum;\n    readonly BLEND_EQUATION_ALPHA: GLenum;\n    readonly BLEND_EQUATION_RGB: GLenum;\n    readonly BLEND_SRC_ALPHA: GLenum;\n    readonly BLEND_SRC_RGB: GLenum;\n    readonly BLUE_BITS: GLenum;\n    readonly BOOL: GLenum;\n    readonly BOOL_VEC2: GLenum;\n    readonly BOOL_VEC3: GLenum;\n    readonly BOOL_VEC4: GLenum;\n    readonly BROWSER_DEFAULT_WEBGL: GLenum;\n    readonly BUFFER_SIZE: GLenum;\n    readonly BUFFER_USAGE: GLenum;\n    readonly BYTE: GLenum;\n    readonly CCW: GLenum;\n    readonly CLAMP_TO_EDGE: GLenum;\n    readonly COLOR_ATTACHMENT0: GLenum;\n    readonly COLOR_BUFFER_BIT: GLenum;\n    readonly COLOR_CLEAR_VALUE: GLenum;\n    readonly COLOR_WRITEMASK: GLenum;\n    readonly COMPILE_STATUS: GLenum;\n    readonly COMPRESSED_TEXTURE_FORMATS: GLenum;\n    readonly CONSTANT_ALPHA: GLenum;\n    readonly CONSTANT_COLOR: GLenum;\n    readonly CONTEXT_LOST_WEBGL: GLenum;\n    readonly CULL_FACE: GLenum;\n    readonly CULL_FACE_MODE: GLenum;\n    readonly CURRENT_PROGRAM: GLenum;\n    readonly CURRENT_VERTEX_ATTRIB: GLenum;\n    readonly CW: GLenum;\n    readonly DECR: GLenum;\n    readonly DECR_WRAP: GLenum;\n    readonly DELETE_STATUS: GLenum;\n    readonly DEPTH_ATTACHMENT: GLenum;\n    readonly DEPTH_BITS: GLenum;\n    readonly DEPTH_BUFFER_BIT: GLenum;\n    readonly DEPTH_CLEAR_VALUE: GLenum;\n    readonly DEPTH_COMPONENT: GLenum;\n    readonly DEPTH_COMPONENT16: GLenum;\n    readonly DEPTH_FUNC: GLenum;\n    readonly DEPTH_RANGE: GLenum;\n    readonly DEPTH_STENCIL: GLenum;\n    readonly DEPTH_STENCIL_ATTACHMENT: GLenum;\n    readonly DEPTH_TEST: GLenum;\n    readonly DEPTH_WRITEMASK: GLenum;\n    readonly DITHER: GLenum;\n    readonly DONT_CARE: GLenum;\n    readonly DST_ALPHA: GLenum;\n    readonly DST_COLOR: GLenum;\n    readonly DYNAMIC_DRAW: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER_BINDING: GLenum;\n    readonly EQUAL: GLenum;\n    readonly FASTEST: GLenum;\n    readonly FLOAT: GLenum;\n    readonly FLOAT_MAT2: GLenum;\n    readonly FLOAT_MAT3: GLenum;\n    readonly FLOAT_MAT4: GLenum;\n    readonly FLOAT_VEC2: GLenum;\n    readonly FLOAT_VEC3: GLenum;\n    readonly FLOAT_VEC4: GLenum;\n    readonly FRAGMENT_SHADER: GLenum;\n    readonly FRAMEBUFFER: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_NAME: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL: GLenum;\n    readonly FRAMEBUFFER_BINDING: GLenum;\n    readonly FRAMEBUFFER_COMPLETE: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_DIMENSIONS: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_UNSUPPORTED: GLenum;\n    readonly FRONT: GLenum;\n    readonly FRONT_AND_BACK: GLenum;\n    readonly FRONT_FACE: GLenum;\n    readonly FUNC_ADD: GLenum;\n    readonly FUNC_REVERSE_SUBTRACT: GLenum;\n    readonly FUNC_SUBTRACT: GLenum;\n    readonly GENERATE_MIPMAP_HINT: GLenum;\n    readonly GEQUAL: GLenum;\n    readonly GREATER: GLenum;\n    readonly GREEN_BITS: GLenum;\n    readonly HIGH_FLOAT: GLenum;\n    readonly HIGH_INT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_FORMAT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_TYPE: GLenum;\n    readonly INCR: GLenum;\n    readonly INCR_WRAP: GLenum;\n    readonly INT: GLenum;\n    readonly INT_VEC2: GLenum;\n    readonly INT_VEC3: GLenum;\n    readonly INT_VEC4: GLenum;\n    readonly INVALID_ENUM: GLenum;\n    readonly INVALID_FRAMEBUFFER_OPERATION: GLenum;\n    readonly INVALID_OPERATION: GLenum;\n    readonly INVALID_VALUE: GLenum;\n    readonly INVERT: GLenum;\n    readonly KEEP: GLenum;\n    readonly LEQUAL: GLenum;\n    readonly LESS: GLenum;\n    readonly LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_NEAREST: GLenum;\n    readonly LINES: GLenum;\n    readonly LINE_LOOP: GLenum;\n    readonly LINE_STRIP: GLenum;\n    readonly LINE_WIDTH: GLenum;\n    readonly LINK_STATUS: GLenum;\n    readonly LOW_FLOAT: GLenum;\n    readonly LOW_INT: GLenum;\n    readonly LUMINANCE: GLenum;\n    readonly LUMINANCE_ALPHA: GLenum;\n    readonly MAX_COMBINED_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_CUBE_MAP_TEXTURE_SIZE: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_VECTORS: GLenum;\n    readonly MAX_RENDERBUFFER_SIZE: GLenum;\n    readonly MAX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_TEXTURE_SIZE: GLenum;\n    readonly MAX_VARYING_VECTORS: GLenum;\n    readonly MAX_VERTEX_ATTRIBS: GLenum;\n    readonly MAX_VERTEX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_VECTORS: GLenum;\n    readonly MAX_VIEWPORT_DIMS: GLenum;\n    readonly MEDIUM_FLOAT: GLenum;\n    readonly MEDIUM_INT: GLenum;\n    readonly MIRRORED_REPEAT: GLenum;\n    readonly NEAREST: GLenum;\n    readonly NEAREST_MIPMAP_LINEAR: GLenum;\n    readonly NEAREST_MIPMAP_NEAREST: GLenum;\n    readonly NEVER: GLenum;\n    readonly NICEST: GLenum;\n    readonly NONE: GLenum;\n    readonly NOTEQUAL: GLenum;\n    readonly NO_ERROR: GLenum;\n    readonly ONE: GLenum;\n    readonly ONE_MINUS_CONSTANT_ALPHA: GLenum;\n    readonly ONE_MINUS_CONSTANT_COLOR: GLenum;\n    readonly ONE_MINUS_DST_ALPHA: GLenum;\n    readonly ONE_MINUS_DST_COLOR: GLenum;\n    readonly ONE_MINUS_SRC_ALPHA: GLenum;\n    readonly ONE_MINUS_SRC_COLOR: GLenum;\n    readonly OUT_OF_MEMORY: GLenum;\n    readonly PACK_ALIGNMENT: GLenum;\n    readonly POINTS: GLenum;\n    readonly POLYGON_OFFSET_FACTOR: GLenum;\n    readonly POLYGON_OFFSET_FILL: GLenum;\n    readonly POLYGON_OFFSET_UNITS: GLenum;\n    readonly RED_BITS: GLenum;\n    readonly RENDERBUFFER: GLenum;\n    readonly RENDERBUFFER_ALPHA_SIZE: GLenum;\n    readonly RENDERBUFFER_BINDING: GLenum;\n    readonly RENDERBUFFER_BLUE_SIZE: GLenum;\n    readonly RENDERBUFFER_DEPTH_SIZE: GLenum;\n    readonly RENDERBUFFER_GREEN_SIZE: GLenum;\n    readonly RENDERBUFFER_HEIGHT: GLenum;\n    readonly RENDERBUFFER_INTERNAL_FORMAT: GLenum;\n    readonly RENDERBUFFER_RED_SIZE: GLenum;\n    readonly RENDERBUFFER_STENCIL_SIZE: GLenum;\n    readonly RENDERBUFFER_WIDTH: GLenum;\n    readonly RENDERER: GLenum;\n    readonly REPEAT: GLenum;\n    readonly REPLACE: GLenum;\n    readonly RGB: GLenum;\n    readonly RGB565: GLenum;\n    readonly RGB5_A1: GLenum;\n    readonly RGBA: GLenum;\n    readonly RGBA4: GLenum;\n    readonly SAMPLER_2D: GLenum;\n    readonly SAMPLER_CUBE: GLenum;\n    readonly SAMPLES: GLenum;\n    readonly SAMPLE_ALPHA_TO_COVERAGE: GLenum;\n    readonly SAMPLE_BUFFERS: GLenum;\n    readonly SAMPLE_COVERAGE: GLenum;\n    readonly SAMPLE_COVERAGE_INVERT: GLenum;\n    readonly SAMPLE_COVERAGE_VALUE: GLenum;\n    readonly SCISSOR_BOX: GLenum;\n    readonly SCISSOR_TEST: GLenum;\n    readonly SHADER_TYPE: GLenum;\n    readonly SHADING_LANGUAGE_VERSION: GLenum;\n    readonly SHORT: GLenum;\n    readonly SRC_ALPHA: GLenum;\n    readonly SRC_ALPHA_SATURATE: GLenum;\n    readonly SRC_COLOR: GLenum;\n    readonly STATIC_DRAW: GLenum;\n    readonly STENCIL_ATTACHMENT: GLenum;\n    readonly STENCIL_BACK_FAIL: GLenum;\n    readonly STENCIL_BACK_FUNC: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_BACK_REF: GLenum;\n    readonly STENCIL_BACK_VALUE_MASK: GLenum;\n    readonly STENCIL_BACK_WRITEMASK: GLenum;\n    readonly STENCIL_BITS: GLenum;\n    readonly STENCIL_BUFFER_BIT: GLenum;\n    readonly STENCIL_CLEAR_VALUE: GLenum;\n    readonly STENCIL_FAIL: GLenum;\n    readonly STENCIL_FUNC: GLenum;\n    readonly STENCIL_INDEX8: GLenum;\n    readonly STENCIL_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_REF: GLenum;\n    readonly STENCIL_TEST: GLenum;\n    readonly STENCIL_VALUE_MASK: GLenum;\n    readonly STENCIL_WRITEMASK: GLenum;\n    readonly STREAM_DRAW: GLenum;\n    readonly SUBPIXEL_BITS: GLenum;\n    readonly TEXTURE: GLenum;\n    readonly TEXTURE0: GLenum;\n    readonly TEXTURE1: GLenum;\n    readonly TEXTURE10: GLenum;\n    readonly TEXTURE11: GLenum;\n    readonly TEXTURE12: GLenum;\n    readonly TEXTURE13: GLenum;\n    readonly TEXTURE14: GLenum;\n    readonly TEXTURE15: GLenum;\n    readonly TEXTURE16: GLenum;\n    readonly TEXTURE17: GLenum;\n    readonly TEXTURE18: GLenum;\n    readonly TEXTURE19: GLenum;\n    readonly TEXTURE2: GLenum;\n    readonly TEXTURE20: GLenum;\n    readonly TEXTURE21: GLenum;\n    readonly TEXTURE22: GLenum;\n    readonly TEXTURE23: GLenum;\n    readonly TEXTURE24: GLenum;\n    readonly TEXTURE25: GLenum;\n    readonly TEXTURE26: GLenum;\n    readonly TEXTURE27: GLenum;\n    readonly TEXTURE28: GLenum;\n    readonly TEXTURE29: GLenum;\n    readonly TEXTURE3: GLenum;\n    readonly TEXTURE30: GLenum;\n    readonly TEXTURE31: GLenum;\n    readonly TEXTURE4: GLenum;\n    readonly TEXTURE5: GLenum;\n    readonly TEXTURE6: GLenum;\n    readonly TEXTURE7: GLenum;\n    readonly TEXTURE8: GLenum;\n    readonly TEXTURE9: GLenum;\n    readonly TEXTURE_2D: GLenum;\n    readonly TEXTURE_BINDING_2D: GLenum;\n    readonly TEXTURE_BINDING_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Z: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Z: GLenum;\n    readonly TEXTURE_MAG_FILTER: GLenum;\n    readonly TEXTURE_MIN_FILTER: GLenum;\n    readonly TEXTURE_WRAP_S: GLenum;\n    readonly TEXTURE_WRAP_T: GLenum;\n    readonly TRIANGLES: GLenum;\n    readonly TRIANGLE_FAN: GLenum;\n    readonly TRIANGLE_STRIP: GLenum;\n    readonly UNPACK_ALIGNMENT: GLenum;\n    readonly UNPACK_COLORSPACE_CONVERSION_WEBGL: GLenum;\n    readonly UNPACK_FLIP_Y_WEBGL: GLenum;\n    readonly UNPACK_PREMULTIPLY_ALPHA_WEBGL: GLenum;\n    readonly UNSIGNED_BYTE: GLenum;\n    readonly UNSIGNED_INT: GLenum;\n    readonly UNSIGNED_SHORT: GLenum;\n    readonly UNSIGNED_SHORT_4_4_4_4: GLenum;\n    readonly UNSIGNED_SHORT_5_5_5_1: GLenum;\n    readonly UNSIGNED_SHORT_5_6_5: GLenum;\n    readonly VALIDATE_STATUS: GLenum;\n    readonly VENDOR: GLenum;\n    readonly VERSION: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_BUFFER_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_ENABLED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_NORMALIZED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_POINTER: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_SIZE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_STRIDE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_TYPE: GLenum;\n    readonly VERTEX_SHADER: GLenum;\n    readonly VIEWPORT: GLenum;\n    readonly ZERO: GLenum;\n};\n\ninterface WebGLRenderingContextBase {\n    readonly canvas: HTMLCanvasElement;\n    readonly drawingBufferHeight: GLsizei;\n    readonly drawingBufferWidth: GLsizei;\n    activeTexture(texture: GLenum): void;\n    attachShader(program: WebGLProgram, shader: WebGLShader): void;\n    bindAttribLocation(program: WebGLProgram, index: GLuint, name: string): void;\n    bindBuffer(target: GLenum, buffer: WebGLBuffer | null): void;\n    bindFramebuffer(target: GLenum, framebuffer: WebGLFramebuffer | null): void;\n    bindRenderbuffer(target: GLenum, renderbuffer: WebGLRenderbuffer | null): void;\n    bindTexture(target: GLenum, texture: WebGLTexture | null): void;\n    blendColor(red: GLclampf, green: GLclampf, blue: GLclampf, alpha: GLclampf): void;\n    blendEquation(mode: GLenum): void;\n    blendEquationSeparate(modeRGB: GLenum, modeAlpha: GLenum): void;\n    blendFunc(sfactor: GLenum, dfactor: GLenum): void;\n    blendFuncSeparate(srcRGB: GLenum, dstRGB: GLenum, srcAlpha: GLenum, dstAlpha: GLenum): void;\n    bufferData(target: GLenum, size: GLsizeiptr, usage: GLenum): void;\n    bufferData(target: GLenum, data: BufferSource | null, usage: GLenum): void;\n    bufferSubData(target: GLenum, offset: GLintptr, data: BufferSource): void;\n    checkFramebufferStatus(target: GLenum): GLenum;\n    clear(mask: GLbitfield): void;\n    clearColor(red: GLclampf, green: GLclampf, blue: GLclampf, alpha: GLclampf): void;\n    clearDepth(depth: GLclampf): void;\n    clearStencil(s: GLint): void;\n    colorMask(red: GLboolean, green: GLboolean, blue: GLboolean, alpha: GLboolean): void;\n    compileShader(shader: WebGLShader): void;\n    compressedTexImage2D(target: GLenum, level: GLint, internalformat: GLenum, width: GLsizei, height: GLsizei, border: GLint, data: ArrayBufferView): void;\n    compressedTexSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, data: ArrayBufferView): void;\n    copyTexImage2D(target: GLenum, level: GLint, internalformat: GLenum, x: GLint, y: GLint, width: GLsizei, height: GLsizei, border: GLint): void;\n    copyTexSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    createBuffer(): WebGLBuffer | null;\n    createFramebuffer(): WebGLFramebuffer | null;\n    createProgram(): WebGLProgram | null;\n    createRenderbuffer(): WebGLRenderbuffer | null;\n    createShader(type: GLenum): WebGLShader | null;\n    createTexture(): WebGLTexture | null;\n    cullFace(mode: GLenum): void;\n    deleteBuffer(buffer: WebGLBuffer | null): void;\n    deleteFramebuffer(framebuffer: WebGLFramebuffer | null): void;\n    deleteProgram(program: WebGLProgram | null): void;\n    deleteRenderbuffer(renderbuffer: WebGLRenderbuffer | null): void;\n    deleteShader(shader: WebGLShader | null): void;\n    deleteTexture(texture: WebGLTexture | null): void;\n    depthFunc(func: GLenum): void;\n    depthMask(flag: GLboolean): void;\n    depthRange(zNear: GLclampf, zFar: GLclampf): void;\n    detachShader(program: WebGLProgram, shader: WebGLShader): void;\n    disable(cap: GLenum): void;\n    disableVertexAttribArray(index: GLuint): void;\n    drawArrays(mode: GLenum, first: GLint, count: GLsizei): void;\n    drawElements(mode: GLenum, count: GLsizei, type: GLenum, offset: GLintptr): void;\n    enable(cap: GLenum): void;\n    enableVertexAttribArray(index: GLuint): void;\n    finish(): void;\n    flush(): void;\n    framebufferRenderbuffer(target: GLenum, attachment: GLenum, renderbuffertarget: GLenum, renderbuffer: WebGLRenderbuffer | null): void;\n    framebufferTexture2D(target: GLenum, attachment: GLenum, textarget: GLenum, texture: WebGLTexture | null, level: GLint): void;\n    frontFace(mode: GLenum): void;\n    generateMipmap(target: GLenum): void;\n    getActiveAttrib(program: WebGLProgram, index: GLuint): WebGLActiveInfo | null;\n    getActiveUniform(program: WebGLProgram, index: GLuint): WebGLActiveInfo | null;\n    getAttachedShaders(program: WebGLProgram): WebGLShader[] | null;\n    getAttribLocation(program: WebGLProgram, name: string): GLint;\n    getBufferParameter(target: GLenum, pname: GLenum): any;\n    getContextAttributes(): WebGLContextAttributes | null;\n    getError(): GLenum;\n    getExtension(extensionName: "EXT_blend_minmax"): EXT_blend_minmax | null;\n    getExtension(extensionName: "EXT_texture_filter_anisotropic"): EXT_texture_filter_anisotropic | null;\n    getExtension(extensionName: "EXT_frag_depth"): EXT_frag_depth | null;\n    getExtension(extensionName: "EXT_shader_texture_lod"): EXT_shader_texture_lod | null;\n    getExtension(extensionName: "EXT_sRGB"): EXT_sRGB | null;\n    getExtension(extensionName: "OES_vertex_array_object"): OES_vertex_array_object | null;\n    getExtension(extensionName: "WEBGL_color_buffer_float"): WEBGL_color_buffer_float | null;\n    getExtension(extensionName: "WEBGL_compressed_texture_astc"): WEBGL_compressed_texture_astc | null;\n    getExtension(extensionName: "WEBGL_compressed_texture_s3tc_srgb"): WEBGL_compressed_texture_s3tc_srgb | null;\n    getExtension(extensionName: "WEBGL_debug_shaders"): WEBGL_debug_shaders | null;\n    getExtension(extensionName: "WEBGL_draw_buffers"): WEBGL_draw_buffers | null;\n    getExtension(extensionName: "WEBGL_lose_context"): WEBGL_lose_context | null;\n    getExtension(extensionName: "WEBGL_depth_texture"): WEBGL_depth_texture | null;\n    getExtension(extensionName: "WEBGL_debug_renderer_info"): WEBGL_debug_renderer_info | null;\n    getExtension(extensionName: "WEBGL_compressed_texture_s3tc"): WEBGL_compressed_texture_s3tc | null;\n    getExtension(extensionName: "OES_texture_half_float_linear"): OES_texture_half_float_linear | null;\n    getExtension(extensionName: "OES_texture_half_float"): OES_texture_half_float | null;\n    getExtension(extensionName: "OES_texture_float_linear"): OES_texture_float_linear | null;\n    getExtension(extensionName: "OES_texture_float"): OES_texture_float | null;\n    getExtension(extensionName: "OES_standard_derivatives"): OES_standard_derivatives | null;\n    getExtension(extensionName: "OES_element_index_uint"): OES_element_index_uint | null;\n    getExtension(extensionName: "ANGLE_instanced_arrays"): ANGLE_instanced_arrays | null;\n    getExtension(extensionName: string): any;\n    getFramebufferAttachmentParameter(target: GLenum, attachment: GLenum, pname: GLenum): any;\n    getParameter(pname: GLenum): any;\n    getProgramInfoLog(program: WebGLProgram): string | null;\n    getProgramParameter(program: WebGLProgram, pname: GLenum): any;\n    getRenderbufferParameter(target: GLenum, pname: GLenum): any;\n    getShaderInfoLog(shader: WebGLShader): string | null;\n    getShaderParameter(shader: WebGLShader, pname: GLenum): any;\n    getShaderPrecisionFormat(shadertype: GLenum, precisiontype: GLenum): WebGLShaderPrecisionFormat | null;\n    getShaderSource(shader: WebGLShader): string | null;\n    getSupportedExtensions(): string[] | null;\n    getTexParameter(target: GLenum, pname: GLenum): any;\n    getUniform(program: WebGLProgram, location: WebGLUniformLocation): any;\n    getUniformLocation(program: WebGLProgram, name: string): WebGLUniformLocation | null;\n    getVertexAttrib(index: GLuint, pname: GLenum): any;\n    getVertexAttribOffset(index: GLuint, pname: GLenum): GLintptr;\n    hint(target: GLenum, mode: GLenum): void;\n    isBuffer(buffer: WebGLBuffer | null): GLboolean;\n    isContextLost(): boolean;\n    isEnabled(cap: GLenum): GLboolean;\n    isFramebuffer(framebuffer: WebGLFramebuffer | null): GLboolean;\n    isProgram(program: WebGLProgram | null): GLboolean;\n    isRenderbuffer(renderbuffer: WebGLRenderbuffer | null): GLboolean;\n    isShader(shader: WebGLShader | null): GLboolean;\n    isTexture(texture: WebGLTexture | null): GLboolean;\n    lineWidth(width: GLfloat): void;\n    linkProgram(program: WebGLProgram): void;\n    pixelStorei(pname: GLenum, param: GLint): void;\n    polygonOffset(factor: GLfloat, units: GLfloat): void;\n    readPixels(x: GLint, y: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    renderbufferStorage(target: GLenum, internalformat: GLenum, width: GLsizei, height: GLsizei): void;\n    sampleCoverage(value: GLclampf, invert: GLboolean): void;\n    scissor(x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    shaderSource(shader: WebGLShader, source: string): void;\n    stencilFunc(func: GLenum, ref: GLint, mask: GLuint): void;\n    stencilFuncSeparate(face: GLenum, func: GLenum, ref: GLint, mask: GLuint): void;\n    stencilMask(mask: GLuint): void;\n    stencilMaskSeparate(face: GLenum, mask: GLuint): void;\n    stencilOp(fail: GLenum, zfail: GLenum, zpass: GLenum): void;\n    stencilOpSeparate(face: GLenum, fail: GLenum, zfail: GLenum, zpass: GLenum): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, border: GLint, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texParameterf(target: GLenum, pname: GLenum, param: GLfloat): void;\n    texParameteri(target: GLenum, pname: GLenum, param: GLint): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    uniform1f(location: WebGLUniformLocation | null, x: GLfloat): void;\n    uniform1fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform1i(location: WebGLUniformLocation | null, x: GLint): void;\n    uniform1iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniform2f(location: WebGLUniformLocation | null, x: GLfloat, y: GLfloat): void;\n    uniform2fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform2i(location: WebGLUniformLocation | null, x: GLint, y: GLint): void;\n    uniform2iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniform3f(location: WebGLUniformLocation | null, x: GLfloat, y: GLfloat, z: GLfloat): void;\n    uniform3fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform3i(location: WebGLUniformLocation | null, x: GLint, y: GLint, z: GLint): void;\n    uniform3iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniform4f(location: WebGLUniformLocation | null, x: GLfloat, y: GLfloat, z: GLfloat, w: GLfloat): void;\n    uniform4fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform4i(location: WebGLUniformLocation | null, x: GLint, y: GLint, z: GLint, w: GLint): void;\n    uniform4iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniformMatrix2fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Float32List): void;\n    uniformMatrix3fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Float32List): void;\n    uniformMatrix4fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Float32List): void;\n    useProgram(program: WebGLProgram | null): void;\n    validateProgram(program: WebGLProgram): void;\n    vertexAttrib1f(index: GLuint, x: GLfloat): void;\n    vertexAttrib1fv(index: GLuint, values: Float32List): void;\n    vertexAttrib2f(index: GLuint, x: GLfloat, y: GLfloat): void;\n    vertexAttrib2fv(index: GLuint, values: Float32List): void;\n    vertexAttrib3f(index: GLuint, x: GLfloat, y: GLfloat, z: GLfloat): void;\n    vertexAttrib3fv(index: GLuint, values: Float32List): void;\n    vertexAttrib4f(index: GLuint, x: GLfloat, y: GLfloat, z: GLfloat, w: GLfloat): void;\n    vertexAttrib4fv(index: GLuint, values: Float32List): void;\n    vertexAttribPointer(index: GLuint, size: GLint, type: GLenum, normalized: GLboolean, stride: GLsizei, offset: GLintptr): void;\n    viewport(x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    readonly ACTIVE_ATTRIBUTES: GLenum;\n    readonly ACTIVE_TEXTURE: GLenum;\n    readonly ACTIVE_UNIFORMS: GLenum;\n    readonly ALIASED_LINE_WIDTH_RANGE: GLenum;\n    readonly ALIASED_POINT_SIZE_RANGE: GLenum;\n    readonly ALPHA: GLenum;\n    readonly ALPHA_BITS: GLenum;\n    readonly ALWAYS: GLenum;\n    readonly ARRAY_BUFFER: GLenum;\n    readonly ARRAY_BUFFER_BINDING: GLenum;\n    readonly ATTACHED_SHADERS: GLenum;\n    readonly BACK: GLenum;\n    readonly BLEND: GLenum;\n    readonly BLEND_COLOR: GLenum;\n    readonly BLEND_DST_ALPHA: GLenum;\n    readonly BLEND_DST_RGB: GLenum;\n    readonly BLEND_EQUATION: GLenum;\n    readonly BLEND_EQUATION_ALPHA: GLenum;\n    readonly BLEND_EQUATION_RGB: GLenum;\n    readonly BLEND_SRC_ALPHA: GLenum;\n    readonly BLEND_SRC_RGB: GLenum;\n    readonly BLUE_BITS: GLenum;\n    readonly BOOL: GLenum;\n    readonly BOOL_VEC2: GLenum;\n    readonly BOOL_VEC3: GLenum;\n    readonly BOOL_VEC4: GLenum;\n    readonly BROWSER_DEFAULT_WEBGL: GLenum;\n    readonly BUFFER_SIZE: GLenum;\n    readonly BUFFER_USAGE: GLenum;\n    readonly BYTE: GLenum;\n    readonly CCW: GLenum;\n    readonly CLAMP_TO_EDGE: GLenum;\n    readonly COLOR_ATTACHMENT0: GLenum;\n    readonly COLOR_BUFFER_BIT: GLenum;\n    readonly COLOR_CLEAR_VALUE: GLenum;\n    readonly COLOR_WRITEMASK: GLenum;\n    readonly COMPILE_STATUS: GLenum;\n    readonly COMPRESSED_TEXTURE_FORMATS: GLenum;\n    readonly CONSTANT_ALPHA: GLenum;\n    readonly CONSTANT_COLOR: GLenum;\n    readonly CONTEXT_LOST_WEBGL: GLenum;\n    readonly CULL_FACE: GLenum;\n    readonly CULL_FACE_MODE: GLenum;\n    readonly CURRENT_PROGRAM: GLenum;\n    readonly CURRENT_VERTEX_ATTRIB: GLenum;\n    readonly CW: GLenum;\n    readonly DECR: GLenum;\n    readonly DECR_WRAP: GLenum;\n    readonly DELETE_STATUS: GLenum;\n    readonly DEPTH_ATTACHMENT: GLenum;\n    readonly DEPTH_BITS: GLenum;\n    readonly DEPTH_BUFFER_BIT: GLenum;\n    readonly DEPTH_CLEAR_VALUE: GLenum;\n    readonly DEPTH_COMPONENT: GLenum;\n    readonly DEPTH_COMPONENT16: GLenum;\n    readonly DEPTH_FUNC: GLenum;\n    readonly DEPTH_RANGE: GLenum;\n    readonly DEPTH_STENCIL: GLenum;\n    readonly DEPTH_STENCIL_ATTACHMENT: GLenum;\n    readonly DEPTH_TEST: GLenum;\n    readonly DEPTH_WRITEMASK: GLenum;\n    readonly DITHER: GLenum;\n    readonly DONT_CARE: GLenum;\n    readonly DST_ALPHA: GLenum;\n    readonly DST_COLOR: GLenum;\n    readonly DYNAMIC_DRAW: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER_BINDING: GLenum;\n    readonly EQUAL: GLenum;\n    readonly FASTEST: GLenum;\n    readonly FLOAT: GLenum;\n    readonly FLOAT_MAT2: GLenum;\n    readonly FLOAT_MAT3: GLenum;\n    readonly FLOAT_MAT4: GLenum;\n    readonly FLOAT_VEC2: GLenum;\n    readonly FLOAT_VEC3: GLenum;\n    readonly FLOAT_VEC4: GLenum;\n    readonly FRAGMENT_SHADER: GLenum;\n    readonly FRAMEBUFFER: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_NAME: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL: GLenum;\n    readonly FRAMEBUFFER_BINDING: GLenum;\n    readonly FRAMEBUFFER_COMPLETE: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_DIMENSIONS: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_UNSUPPORTED: GLenum;\n    readonly FRONT: GLenum;\n    readonly FRONT_AND_BACK: GLenum;\n    readonly FRONT_FACE: GLenum;\n    readonly FUNC_ADD: GLenum;\n    readonly FUNC_REVERSE_SUBTRACT: GLenum;\n    readonly FUNC_SUBTRACT: GLenum;\n    readonly GENERATE_MIPMAP_HINT: GLenum;\n    readonly GEQUAL: GLenum;\n    readonly GREATER: GLenum;\n    readonly GREEN_BITS: GLenum;\n    readonly HIGH_FLOAT: GLenum;\n    readonly HIGH_INT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_FORMAT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_TYPE: GLenum;\n    readonly INCR: GLenum;\n    readonly INCR_WRAP: GLenum;\n    readonly INT: GLenum;\n    readonly INT_VEC2: GLenum;\n    readonly INT_VEC3: GLenum;\n    readonly INT_VEC4: GLenum;\n    readonly INVALID_ENUM: GLenum;\n    readonly INVALID_FRAMEBUFFER_OPERATION: GLenum;\n    readonly INVALID_OPERATION: GLenum;\n    readonly INVALID_VALUE: GLenum;\n    readonly INVERT: GLenum;\n    readonly KEEP: GLenum;\n    readonly LEQUAL: GLenum;\n    readonly LESS: GLenum;\n    readonly LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_NEAREST: GLenum;\n    readonly LINES: GLenum;\n    readonly LINE_LOOP: GLenum;\n    readonly LINE_STRIP: GLenum;\n    readonly LINE_WIDTH: GLenum;\n    readonly LINK_STATUS: GLenum;\n    readonly LOW_FLOAT: GLenum;\n    readonly LOW_INT: GLenum;\n    readonly LUMINANCE: GLenum;\n    readonly LUMINANCE_ALPHA: GLenum;\n    readonly MAX_COMBINED_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_CUBE_MAP_TEXTURE_SIZE: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_VECTORS: GLenum;\n    readonly MAX_RENDERBUFFER_SIZE: GLenum;\n    readonly MAX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_TEXTURE_SIZE: GLenum;\n    readonly MAX_VARYING_VECTORS: GLenum;\n    readonly MAX_VERTEX_ATTRIBS: GLenum;\n    readonly MAX_VERTEX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_VECTORS: GLenum;\n    readonly MAX_VIEWPORT_DIMS: GLenum;\n    readonly MEDIUM_FLOAT: GLenum;\n    readonly MEDIUM_INT: GLenum;\n    readonly MIRRORED_REPEAT: GLenum;\n    readonly NEAREST: GLenum;\n    readonly NEAREST_MIPMAP_LINEAR: GLenum;\n    readonly NEAREST_MIPMAP_NEAREST: GLenum;\n    readonly NEVER: GLenum;\n    readonly NICEST: GLenum;\n    readonly NONE: GLenum;\n    readonly NOTEQUAL: GLenum;\n    readonly NO_ERROR: GLenum;\n    readonly ONE: GLenum;\n    readonly ONE_MINUS_CONSTANT_ALPHA: GLenum;\n    readonly ONE_MINUS_CONSTANT_COLOR: GLenum;\n    readonly ONE_MINUS_DST_ALPHA: GLenum;\n    readonly ONE_MINUS_DST_COLOR: GLenum;\n    readonly ONE_MINUS_SRC_ALPHA: GLenum;\n    readonly ONE_MINUS_SRC_COLOR: GLenum;\n    readonly OUT_OF_MEMORY: GLenum;\n    readonly PACK_ALIGNMENT: GLenum;\n    readonly POINTS: GLenum;\n    readonly POLYGON_OFFSET_FACTOR: GLenum;\n    readonly POLYGON_OFFSET_FILL: GLenum;\n    readonly POLYGON_OFFSET_UNITS: GLenum;\n    readonly RED_BITS: GLenum;\n    readonly RENDERBUFFER: GLenum;\n    readonly RENDERBUFFER_ALPHA_SIZE: GLenum;\n    readonly RENDERBUFFER_BINDING: GLenum;\n    readonly RENDERBUFFER_BLUE_SIZE: GLenum;\n    readonly RENDERBUFFER_DEPTH_SIZE: GLenum;\n    readonly RENDERBUFFER_GREEN_SIZE: GLenum;\n    readonly RENDERBUFFER_HEIGHT: GLenum;\n    readonly RENDERBUFFER_INTERNAL_FORMAT: GLenum;\n    readonly RENDERBUFFER_RED_SIZE: GLenum;\n    readonly RENDERBUFFER_STENCIL_SIZE: GLenum;\n    readonly RENDERBUFFER_WIDTH: GLenum;\n    readonly RENDERER: GLenum;\n    readonly REPEAT: GLenum;\n    readonly REPLACE: GLenum;\n    readonly RGB: GLenum;\n    readonly RGB565: GLenum;\n    readonly RGB5_A1: GLenum;\n    readonly RGBA: GLenum;\n    readonly RGBA4: GLenum;\n    readonly SAMPLER_2D: GLenum;\n    readonly SAMPLER_CUBE: GLenum;\n    readonly SAMPLES: GLenum;\n    readonly SAMPLE_ALPHA_TO_COVERAGE: GLenum;\n    readonly SAMPLE_BUFFERS: GLenum;\n    readonly SAMPLE_COVERAGE: GLenum;\n    readonly SAMPLE_COVERAGE_INVERT: GLenum;\n    readonly SAMPLE_COVERAGE_VALUE: GLenum;\n    readonly SCISSOR_BOX: GLenum;\n    readonly SCISSOR_TEST: GLenum;\n    readonly SHADER_TYPE: GLenum;\n    readonly SHADING_LANGUAGE_VERSION: GLenum;\n    readonly SHORT: GLenum;\n    readonly SRC_ALPHA: GLenum;\n    readonly SRC_ALPHA_SATURATE: GLenum;\n    readonly SRC_COLOR: GLenum;\n    readonly STATIC_DRAW: GLenum;\n    readonly STENCIL_ATTACHMENT: GLenum;\n    readonly STENCIL_BACK_FAIL: GLenum;\n    readonly STENCIL_BACK_FUNC: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_BACK_REF: GLenum;\n    readonly STENCIL_BACK_VALUE_MASK: GLenum;\n    readonly STENCIL_BACK_WRITEMASK: GLenum;\n    readonly STENCIL_BITS: GLenum;\n    readonly STENCIL_BUFFER_BIT: GLenum;\n    readonly STENCIL_CLEAR_VALUE: GLenum;\n    readonly STENCIL_FAIL: GLenum;\n    readonly STENCIL_FUNC: GLenum;\n    readonly STENCIL_INDEX8: GLenum;\n    readonly STENCIL_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_REF: GLenum;\n    readonly STENCIL_TEST: GLenum;\n    readonly STENCIL_VALUE_MASK: GLenum;\n    readonly STENCIL_WRITEMASK: GLenum;\n    readonly STREAM_DRAW: GLenum;\n    readonly SUBPIXEL_BITS: GLenum;\n    readonly TEXTURE: GLenum;\n    readonly TEXTURE0: GLenum;\n    readonly TEXTURE1: GLenum;\n    readonly TEXTURE10: GLenum;\n    readonly TEXTURE11: GLenum;\n    readonly TEXTURE12: GLenum;\n    readonly TEXTURE13: GLenum;\n    readonly TEXTURE14: GLenum;\n    readonly TEXTURE15: GLenum;\n    readonly TEXTURE16: GLenum;\n    readonly TEXTURE17: GLenum;\n    readonly TEXTURE18: GLenum;\n    readonly TEXTURE19: GLenum;\n    readonly TEXTURE2: GLenum;\n    readonly TEXTURE20: GLenum;\n    readonly TEXTURE21: GLenum;\n    readonly TEXTURE22: GLenum;\n    readonly TEXTURE23: GLenum;\n    readonly TEXTURE24: GLenum;\n    readonly TEXTURE25: GLenum;\n    readonly TEXTURE26: GLenum;\n    readonly TEXTURE27: GLenum;\n    readonly TEXTURE28: GLenum;\n    readonly TEXTURE29: GLenum;\n    readonly TEXTURE3: GLenum;\n    readonly TEXTURE30: GLenum;\n    readonly TEXTURE31: GLenum;\n    readonly TEXTURE4: GLenum;\n    readonly TEXTURE5: GLenum;\n    readonly TEXTURE6: GLenum;\n    readonly TEXTURE7: GLenum;\n    readonly TEXTURE8: GLenum;\n    readonly TEXTURE9: GLenum;\n    readonly TEXTURE_2D: GLenum;\n    readonly TEXTURE_BINDING_2D: GLenum;\n    readonly TEXTURE_BINDING_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Z: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Z: GLenum;\n    readonly TEXTURE_MAG_FILTER: GLenum;\n    readonly TEXTURE_MIN_FILTER: GLenum;\n    readonly TEXTURE_WRAP_S: GLenum;\n    readonly TEXTURE_WRAP_T: GLenum;\n    readonly TRIANGLES: GLenum;\n    readonly TRIANGLE_FAN: GLenum;\n    readonly TRIANGLE_STRIP: GLenum;\n    readonly UNPACK_ALIGNMENT: GLenum;\n    readonly UNPACK_COLORSPACE_CONVERSION_WEBGL: GLenum;\n    readonly UNPACK_FLIP_Y_WEBGL: GLenum;\n    readonly UNPACK_PREMULTIPLY_ALPHA_WEBGL: GLenum;\n    readonly UNSIGNED_BYTE: GLenum;\n    readonly UNSIGNED_INT: GLenum;\n    readonly UNSIGNED_SHORT: GLenum;\n    readonly UNSIGNED_SHORT_4_4_4_4: GLenum;\n    readonly UNSIGNED_SHORT_5_5_5_1: GLenum;\n    readonly UNSIGNED_SHORT_5_6_5: GLenum;\n    readonly VALIDATE_STATUS: GLenum;\n    readonly VENDOR: GLenum;\n    readonly VERSION: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_BUFFER_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_ENABLED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_NORMALIZED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_POINTER: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_SIZE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_STRIDE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_TYPE: GLenum;\n    readonly VERTEX_SHADER: GLenum;\n    readonly VIEWPORT: GLenum;\n    readonly ZERO: GLenum;\n}\n\ninterface WebGLShader extends WebGLObject {\n}\n\ndeclare var WebGLShader: {\n    prototype: WebGLShader;\n    new(): WebGLShader;\n};\n\ninterface WebGLShaderPrecisionFormat {\n    readonly precision: GLint;\n    readonly rangeMax: GLint;\n    readonly rangeMin: GLint;\n}\n\ndeclare var WebGLShaderPrecisionFormat: {\n    prototype: WebGLShaderPrecisionFormat;\n    new(): WebGLShaderPrecisionFormat;\n};\n\ninterface WebGLTexture extends WebGLObject {\n}\n\ndeclare var WebGLTexture: {\n    prototype: WebGLTexture;\n    new(): WebGLTexture;\n};\n\ninterface WebGLUniformLocation {\n}\n\ndeclare var WebGLUniformLocation: {\n    prototype: WebGLUniformLocation;\n    new(): WebGLUniformLocation;\n};\n\ninterface WebGLVertexArrayObjectOES extends WebGLObject {\n}\n\ninterface WebKitPoint {\n    x: number;\n    y: number;\n}\n\ndeclare var WebKitPoint: {\n    prototype: WebKitPoint;\n    new(x?: number, y?: number): WebKitPoint;\n};\n\ninterface WebSocketEventMap {\n    "close": CloseEvent;\n    "error": Event;\n    "message": MessageEvent;\n    "open": Event;\n}\n\ninterface WebSocket extends EventTarget {\n    binaryType: BinaryType;\n    readonly bufferedAmount: number;\n    readonly extensions: string;\n    onclose: ((this: WebSocket, ev: CloseEvent) => any) | null;\n    onerror: ((this: WebSocket, ev: Event) => any) | null;\n    onmessage: ((this: WebSocket, ev: MessageEvent) => any) | null;\n    onopen: ((this: WebSocket, ev: Event) => any) | null;\n    readonly protocol: string;\n    readonly readyState: number;\n    readonly url: string;\n    close(code?: number, reason?: string): void;\n    send(data: string | ArrayBufferLike | Blob | ArrayBufferView): void;\n    readonly CLOSED: number;\n    readonly CLOSING: number;\n    readonly CONNECTING: number;\n    readonly OPEN: number;\n    addEventListener<K extends keyof WebSocketEventMap>(type: K, listener: (this: WebSocket, ev: WebSocketEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof WebSocketEventMap>(type: K, listener: (this: WebSocket, ev: WebSocketEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var WebSocket: {\n    prototype: WebSocket;\n    new(url: string, protocols?: string | string[]): WebSocket;\n    readonly CLOSED: number;\n    readonly CLOSING: number;\n    readonly CONNECTING: number;\n    readonly OPEN: number;\n};\n\ninterface WheelEvent extends MouseEvent {\n    readonly deltaMode: number;\n    readonly deltaX: number;\n    readonly deltaY: number;\n    readonly deltaZ: number;\n    getCurrentPoint(element: Element): void;\n    initWheelEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, detailArg: number, screenXArg: number, screenYArg: number, clientXArg: number, clientYArg: number, buttonArg: number, relatedTargetArg: EventTarget, modifiersListArg: string, deltaXArg: number, deltaYArg: number, deltaZArg: number, deltaMode: number): void;\n    readonly DOM_DELTA_LINE: number;\n    readonly DOM_DELTA_PAGE: number;\n    readonly DOM_DELTA_PIXEL: number;\n}\n\ndeclare var WheelEvent: {\n    prototype: WheelEvent;\n    new(typeArg: string, eventInitDict?: WheelEventInit): WheelEvent;\n    readonly DOM_DELTA_LINE: number;\n    readonly DOM_DELTA_PAGE: number;\n    readonly DOM_DELTA_PIXEL: number;\n};\n\ninterface WindowEventMap extends GlobalEventHandlersEventMap, WindowEventHandlersEventMap {\n    "abort": UIEvent;\n    "afterprint": Event;\n    "beforeprint": Event;\n    "beforeunload": BeforeUnloadEvent;\n    "blur": FocusEvent;\n    "canplay": Event;\n    "canplaythrough": Event;\n    "change": Event;\n    "click": MouseEvent;\n    "compassneedscalibration": Event;\n    "contextmenu": MouseEvent;\n    "dblclick": MouseEvent;\n    "devicelight": DeviceLightEvent;\n    "devicemotion": DeviceMotionEvent;\n    "deviceorientation": DeviceOrientationEvent;\n    "drag": DragEvent;\n    "dragend": DragEvent;\n    "dragenter": DragEvent;\n    "dragleave": DragEvent;\n    "dragover": DragEvent;\n    "dragstart": DragEvent;\n    "drop": DragEvent;\n    "durationchange": Event;\n    "emptied": Event;\n    "ended": Event;\n    "error": ErrorEvent;\n    "focus": FocusEvent;\n    "hashchange": HashChangeEvent;\n    "input": Event;\n    "invalid": Event;\n    "keydown": KeyboardEvent;\n    "keypress": KeyboardEvent;\n    "keyup": KeyboardEvent;\n    "load": Event;\n    "loadeddata": Event;\n    "loadedmetadata": Event;\n    "loadstart": Event;\n    "message": MessageEvent;\n    "mousedown": MouseEvent;\n    "mouseenter": MouseEvent;\n    "mouseleave": MouseEvent;\n    "mousemove": MouseEvent;\n    "mouseout": MouseEvent;\n    "mouseover": MouseEvent;\n    "mouseup": MouseEvent;\n    "mousewheel": Event;\n    "MSGestureChange": Event;\n    "MSGestureDoubleTap": Event;\n    "MSGestureEnd": Event;\n    "MSGestureHold": Event;\n    "MSGestureStart": Event;\n    "MSGestureTap": Event;\n    "MSInertiaStart": Event;\n    "MSPointerCancel": Event;\n    "MSPointerDown": Event;\n    "MSPointerEnter": Event;\n    "MSPointerLeave": Event;\n    "MSPointerMove": Event;\n    "MSPointerOut": Event;\n    "MSPointerOver": Event;\n    "MSPointerUp": Event;\n    "offline": Event;\n    "online": Event;\n    "orientationchange": Event;\n    "pagehide": PageTransitionEvent;\n    "pageshow": PageTransitionEvent;\n    "pause": Event;\n    "play": Event;\n    "playing": Event;\n    "popstate": PopStateEvent;\n    "progress": ProgressEvent;\n    "ratechange": Event;\n    "readystatechange": ProgressEvent;\n    "reset": Event;\n    "resize": UIEvent;\n    "scroll": UIEvent;\n    "seeked": Event;\n    "seeking": Event;\n    "select": UIEvent;\n    "stalled": Event;\n    "storage": StorageEvent;\n    "submit": Event;\n    "suspend": Event;\n    "timeupdate": Event;\n    "unload": Event;\n    "volumechange": Event;\n    "vrdisplayactivate": Event;\n    "vrdisplayblur": Event;\n    "vrdisplayconnect": Event;\n    "vrdisplaydeactivate": Event;\n    "vrdisplaydisconnect": Event;\n    "vrdisplayfocus": Event;\n    "vrdisplaypointerrestricted": Event;\n    "vrdisplaypointerunrestricted": Event;\n    "vrdisplaypresentchange": Event;\n    "waiting": Event;\n}\n\ninterface Window extends EventTarget, WindowTimers, WindowSessionStorage, WindowLocalStorage, WindowConsole, GlobalEventHandlers, IDBEnvironment, WindowBase64, GlobalFetch, WindowOrWorkerGlobalScope, WindowEventHandlers {\n    Blob: typeof Blob;\n    URL: typeof URL;\n    URLSearchParams: typeof URLSearchParams;\n    readonly applicationCache: ApplicationCache;\n    readonly caches: CacheStorage;\n    readonly clientInformation: Navigator;\n    readonly closed: boolean;\n    readonly crypto: Crypto;\n    customElements: CustomElementRegistry;\n    defaultStatus: string;\n    readonly devicePixelRatio: number;\n    readonly doNotTrack: string;\n    readonly document: Document;\n    readonly event: Event | undefined;\n    /** @deprecated */\n    readonly external: External;\n    readonly frameElement: Element;\n    readonly frames: Window;\n    readonly history: History;\n    readonly innerHeight: number;\n    readonly innerWidth: number;\n    readonly isSecureContext: boolean;\n    readonly length: number;\n    location: Location;\n    readonly locationbar: BarProp;\n    readonly menubar: BarProp;\n    readonly msContentScript: ExtensionScriptApis;\n    name: string;\n    readonly navigator: Navigator;\n    offscreenBuffering: string | boolean;\n    oncompassneedscalibration: ((this: Window, ev: Event) => any) | null;\n    ondevicelight: ((this: Window, ev: DeviceLightEvent) => any) | null;\n    ondevicemotion: ((this: Window, ev: DeviceMotionEvent) => any) | null;\n    ondeviceorientation: ((this: Window, ev: DeviceOrientationEvent) => any) | null;\n    onmousewheel: ((this: Window, ev: Event) => any) | null;\n    onmsgesturechange: ((this: Window, ev: Event) => any) | null;\n    onmsgesturedoubletap: ((this: Window, ev: Event) => any) | null;\n    onmsgestureend: ((this: Window, ev: Event) => any) | null;\n    onmsgesturehold: ((this: Window, ev: Event) => any) | null;\n    onmsgesturestart: ((this: Window, ev: Event) => any) | null;\n    onmsgesturetap: ((this: Window, ev: Event) => any) | null;\n    onmsinertiastart: ((this: Window, ev: Event) => any) | null;\n    onmspointercancel: ((this: Window, ev: Event) => any) | null;\n    onmspointerdown: ((this: Window, ev: Event) => any) | null;\n    onmspointerenter: ((this: Window, ev: Event) => any) | null;\n    onmspointerleave: ((this: Window, ev: Event) => any) | null;\n    onmspointermove: ((this: Window, ev: Event) => any) | null;\n    onmspointerout: ((this: Window, ev: Event) => any) | null;\n    onmspointerover: ((this: Window, ev: Event) => any) | null;\n    onmspointerup: ((this: Window, ev: Event) => any) | null;\n    /** @deprecated */\n    onorientationchange: ((this: Window, ev: Event) => any) | null;\n    onreadystatechange: ((this: Window, ev: ProgressEvent) => any) | null;\n    onvrdisplayactivate: ((this: Window, ev: Event) => any) | null;\n    onvrdisplayblur: ((this: Window, ev: Event) => any) | null;\n    onvrdisplayconnect: ((this: Window, ev: Event) => any) | null;\n    onvrdisplaydeactivate: ((this: Window, ev: Event) => any) | null;\n    onvrdisplaydisconnect: ((this: Window, ev: Event) => any) | null;\n    onvrdisplayfocus: ((this: Window, ev: Event) => any) | null;\n    onvrdisplaypointerrestricted: ((this: Window, ev: Event) => any) | null;\n    onvrdisplaypointerunrestricted: ((this: Window, ev: Event) => any) | null;\n    onvrdisplaypresentchange: ((this: Window, ev: Event) => any) | null;\n    opener: any;\n    /** @deprecated */\n    readonly orientation: string | number;\n    readonly outerHeight: number;\n    readonly outerWidth: number;\n    readonly pageXOffset: number;\n    readonly pageYOffset: number;\n    readonly parent: Window;\n    readonly performance: Performance;\n    readonly personalbar: BarProp;\n    readonly screen: Screen;\n    readonly screenLeft: number;\n    readonly screenTop: number;\n    readonly screenX: number;\n    readonly screenY: number;\n    readonly scrollX: number;\n    readonly scrollY: number;\n    readonly scrollbars: BarProp;\n    readonly self: Window;\n    readonly speechSynthesis: SpeechSynthesis;\n    status: string;\n    readonly statusbar: BarProp;\n    readonly styleMedia: StyleMedia;\n    readonly toolbar: BarProp;\n    readonly top: Window;\n    readonly window: Window;\n    alert(message?: any): void;\n    blur(): void;\n    cancelAnimationFrame(handle: number): void;\n    /** @deprecated */\n    captureEvents(): void;\n    close(): void;\n    confirm(message?: string): boolean;\n    departFocus(navigationReason: NavigationReason, origin: FocusNavigationOrigin): void;\n    focus(): void;\n    getComputedStyle(elt: Element, pseudoElt?: string | null): CSSStyleDeclaration;\n    getMatchedCSSRules(elt: Element, pseudoElt?: string | null): CSSRuleList;\n    getSelection(): Selection;\n    matchMedia(query: string): MediaQueryList;\n    moveBy(x: number, y: number): void;\n    moveTo(x: number, y: number): void;\n    msWriteProfilerMark(profilerMarkName: string): void;\n    open(url?: string, target?: string, features?: string, replace?: boolean): Window | null;\n    postMessage(message: any, targetOrigin: string, transfer?: Transferable[]): void;\n    print(): void;\n    prompt(message?: string, _default?: string): string | null;\n    /** @deprecated */\n    releaseEvents(): void;\n    requestAnimationFrame(callback: FrameRequestCallback): number;\n    resizeBy(x: number, y: number): void;\n    resizeTo(x: number, y: number): void;\n    scroll(options?: ScrollToOptions): void;\n    scroll(x: number, y: number): void;\n    scrollBy(options?: ScrollToOptions): void;\n    scrollBy(x: number, y: number): void;\n    scrollTo(options?: ScrollToOptions): void;\n    scrollTo(x: number, y: number): void;\n    stop(): void;\n    webkitCancelAnimationFrame(handle: number): void;\n    webkitConvertPointFromNodeToPage(node: Node, pt: WebKitPoint): WebKitPoint;\n    webkitConvertPointFromPageToNode(node: Node, pt: WebKitPoint): WebKitPoint;\n    webkitRequestAnimationFrame(callback: FrameRequestCallback): number;\n    addEventListener<K extends keyof WindowEventMap>(type: K, listener: (this: Window, ev: WindowEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof WindowEventMap>(type: K, listener: (this: Window, ev: WindowEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Window: {\n    prototype: Window;\n    new(): Window;\n};\n\ninterface WindowBase64 {\n    atob(encodedString: string): string;\n    btoa(rawString: string): string;\n}\n\ninterface WindowConsole {\n    readonly console: Console;\n}\n\ninterface WindowEventHandlersEventMap {\n    "afterprint": Event;\n    "beforeprint": Event;\n    "beforeunload": BeforeUnloadEvent;\n    "hashchange": HashChangeEvent;\n    "languagechange": Event;\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n    "offline": Event;\n    "online": Event;\n    "pagehide": PageTransitionEvent;\n    "pageshow": PageTransitionEvent;\n    "popstate": PopStateEvent;\n    "rejectionhandled": Event;\n    "storage": StorageEvent;\n    "unhandledrejection": PromiseRejectionEvent;\n    "unload": Event;\n}\n\ninterface WindowEventHandlers {\n    onafterprint: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    onbeforeprint: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    onbeforeunload: ((this: WindowEventHandlers, ev: BeforeUnloadEvent) => any) | null;\n    onhashchange: ((this: WindowEventHandlers, ev: HashChangeEvent) => any) | null;\n    onlanguagechange: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    onmessage: ((this: WindowEventHandlers, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: WindowEventHandlers, ev: MessageEvent) => any) | null;\n    onoffline: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    ononline: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    onpagehide: ((this: WindowEventHandlers, ev: PageTransitionEvent) => any) | null;\n    onpageshow: ((this: WindowEventHandlers, ev: PageTransitionEvent) => any) | null;\n    onpopstate: ((this: WindowEventHandlers, ev: PopStateEvent) => any) | null;\n    onrejectionhandled: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    onstorage: ((this: WindowEventHandlers, ev: StorageEvent) => any) | null;\n    onunhandledrejection: ((this: WindowEventHandlers, ev: PromiseRejectionEvent) => any) | null;\n    onunload: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    addEventListener<K extends keyof WindowEventHandlersEventMap>(type: K, listener: (this: WindowEventHandlers, ev: WindowEventHandlersEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof WindowEventHandlersEventMap>(type: K, listener: (this: WindowEventHandlers, ev: WindowEventHandlersEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ninterface WindowLocalStorage {\n    readonly localStorage: Storage;\n}\n\ninterface WindowOrWorkerGlobalScope {\n    readonly caches: CacheStorage;\n    readonly crypto: Crypto;\n    readonly indexedDB: IDBFactory;\n    readonly origin: string;\n    readonly performance: Performance;\n    atob(data: string): string;\n    btoa(data: string): string;\n    clearInterval(handle?: number): void;\n    clearTimeout(handle?: number): void;\n    createImageBitmap(image: ImageBitmapSource): Promise<ImageBitmap>;\n    createImageBitmap(image: ImageBitmapSource, sx: number, sy: number, sw: number, sh: number): Promise<ImageBitmap>;\n    fetch(input: RequestInfo, init?: RequestInit): Promise<Response>;\n    queueMicrotask(callback: Function): void;\n    setInterval(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\n    setTimeout(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\n}\n\ninterface WindowSessionStorage {\n    readonly sessionStorage: Storage;\n}\n\ninterface WindowTimers {\n}\n\ninterface WorkerEventMap extends AbstractWorkerEventMap {\n    "message": MessageEvent;\n}\n\ninterface Worker extends EventTarget, AbstractWorker {\n    onmessage: ((this: Worker, ev: MessageEvent) => any) | null;\n    postMessage(message: any, transfer?: Transferable[]): void;\n    terminate(): void;\n    addEventListener<K extends keyof WorkerEventMap>(type: K, listener: (this: Worker, ev: WorkerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof WorkerEventMap>(type: K, listener: (this: Worker, ev: WorkerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Worker: {\n    prototype: Worker;\n    new(stringUrl: string, options?: WorkerOptions): Worker;\n};\n\ninterface Worklet {\n    addModule(moduleURL: string, options?: WorkletOptions): Promise<void>;\n}\n\ndeclare var Worklet: {\n    prototype: Worklet;\n    new(): Worklet;\n};\n\ninterface WritableStream<W = any> {\n    readonly locked: boolean;\n    abort(reason?: any): Promise<void>;\n    getWriter(): WritableStreamDefaultWriter<W>;\n}\n\ndeclare var WritableStream: {\n    prototype: WritableStream;\n    new<W = any>(underlyingSink?: UnderlyingSink<W>, strategy?: QueuingStrategy<W>): WritableStream<W>;\n};\n\ninterface WritableStreamDefaultController {\n    error(error?: any): void;\n}\n\ninterface WritableStreamDefaultWriter<W = any> {\n    readonly closed: Promise<void>;\n    readonly desiredSize: number | null;\n    readonly ready: Promise<void>;\n    abort(reason?: any): Promise<void>;\n    close(): Promise<void>;\n    releaseLock(): void;\n    write(chunk: W): Promise<void>;\n}\n\ninterface XMLDocument extends Document {\n    addEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: XMLDocument, ev: DocumentEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: XMLDocument, ev: DocumentEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var XMLDocument: {\n    prototype: XMLDocument;\n    new(): XMLDocument;\n};\n\ninterface XMLHttpRequestEventMap extends XMLHttpRequestEventTargetEventMap {\n    "readystatechange": Event;\n}\n\ninterface XMLHttpRequest extends XMLHttpRequestEventTarget {\n    onreadystatechange: ((this: XMLHttpRequest, ev: Event) => any) | null;\n    /**\n     * Returns client\'s state.\n     */\n    readonly readyState: number;\n    /**\n     * Returns the response\'s body.\n     */\n    readonly response: any;\n    /**\n     * Returns the text response.\n     * Throws an "InvalidStateError" DOMException if responseType is not the empty string or "text".\n     */\n    readonly responseText: string;\n    /**\n     * Returns the response type.\n     * Can be set to change the response type. Values are:\n     * the empty string (default),\n     * "arraybuffer",\n     * "blob",\n     * "document",\n     * "json", and\n     * "text".\n     * When set: setting to "document" is ignored if current global object is not a Window object.\n     * When set: throws an "InvalidStateError" DOMException if state is loading or done.\n     * When set: throws an "InvalidAccessError" DOMException if the synchronous flag is set and current global object is a Window object.\n     */\n    responseType: XMLHttpRequestResponseType;\n    readonly responseURL: string;\n    /**\n     * Returns the document response.\n     * Throws an "InvalidStateError" DOMException if responseType is not the empty string or "document".\n     */\n    readonly responseXML: Document | null;\n    readonly status: number;\n    readonly statusText: string;\n    /**\n     * Can be set to a time in milliseconds. When set to a non-zero value will cause fetching to terminate after the given time has passed. When the time has passed, the\n     * request has not yet completed, and the synchronous flag is unset, a timeout event will then be dispatched, or a\n     * "TimeoutError" DOMException will be thrown otherwise (for the send() method).\n     * When set: throws an "InvalidAccessError" DOMException if the synchronous flag is set and current global object is a Window object.\n     */\n    timeout: number;\n    /**\n     * Returns the associated XMLHttpRequestUpload object. It can be used to gather transmission information when data is\n     * transferred to a server.\n     */\n    readonly upload: XMLHttpRequestUpload;\n    /**\n     * True when credentials are to be included in a cross-origin request. False when they are\n     * to be excluded in a cross-origin request and when cookies are to be ignored in its response.\n     * Initially false.\n     * When set: throws an "InvalidStateError" DOMException if state is not unsent or opened, or if the send() flag is set.\n     */\n    withCredentials: boolean;\n    /**\n     * Cancels any network activity.\n     */\n    abort(): void;\n    getAllResponseHeaders(): string;\n    getResponseHeader(name: string): string | null;\n    /**\n     * Sets the request method, request URL, and synchronous flag.\n     * Throws a "SyntaxError" DOMException if either method is not a\n     * valid HTTP method or url cannot be parsed.\n     * Throws a "SecurityError" DOMException if method is a\n     * case-insensitive match for `CONNECT`, `TRACE`, or `TRACK`.\n     * Throws an "InvalidAccessError" DOMException if async is false, current global object is a Window object, and the timeout attribute is not zero or the responseType attribute is not the empty string.\n     */\n    open(method: string, url: string): void;\n    open(method: string, url: string, async: boolean, username?: string | null, password?: string | null): void;\n    /**\n     * Acts as if the `Content-Type` header value for response is mime.\n     * (It does not actually change the header though.)\n     * Throws an "InvalidStateError" DOMException if state is loading or done.\n     */\n    overrideMimeType(mime: string): void;\n    /**\n     * Initiates the request. The optional argument provides the request body. The argument is ignored if request method is GET or HEAD.\n     * Throws an "InvalidStateError" DOMException if either state is not opened or the send() flag is set.\n     */\n    send(body?: Document | BodyInit | null): void;\n    /**\n     * Combines a header in author request headers.\n     * Throws an "InvalidStateError" DOMException if either state is not opened or the send() flag is set.\n     * Throws a "SyntaxError" DOMException if name is not a header name\n     * or if value is not a header value.\n     */\n    setRequestHeader(name: string, value: string): void;\n    readonly DONE: number;\n    readonly HEADERS_RECEIVED: number;\n    readonly LOADING: number;\n    readonly OPENED: number;\n    readonly UNSENT: number;\n    addEventListener<K extends keyof XMLHttpRequestEventMap>(type: K, listener: (this: XMLHttpRequest, ev: XMLHttpRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof XMLHttpRequestEventMap>(type: K, listener: (this: XMLHttpRequest, ev: XMLHttpRequestEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var XMLHttpRequest: {\n    prototype: XMLHttpRequest;\n    new(): XMLHttpRequest;\n    readonly DONE: number;\n    readonly HEADERS_RECEIVED: number;\n    readonly LOADING: number;\n    readonly OPENED: number;\n    readonly UNSENT: number;\n};\n\ninterface XMLHttpRequestEventTargetEventMap {\n    "abort": ProgressEvent;\n    "error": ProgressEvent;\n    "load": ProgressEvent;\n    "loadend": ProgressEvent;\n    "loadstart": ProgressEvent;\n    "progress": ProgressEvent;\n    "timeout": ProgressEvent;\n}\n\ninterface XMLHttpRequestEventTarget extends EventTarget {\n    onabort: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onerror: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onload: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onloadend: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onloadstart: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onprogress: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    ontimeout: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    addEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestEventTarget, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestEventTarget, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var XMLHttpRequestEventTarget: {\n    prototype: XMLHttpRequestEventTarget;\n    new(): XMLHttpRequestEventTarget;\n};\n\ninterface XMLHttpRequestUpload extends XMLHttpRequestEventTarget {\n    addEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestUpload, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestUpload, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var XMLHttpRequestUpload: {\n    prototype: XMLHttpRequestUpload;\n    new(): XMLHttpRequestUpload;\n};\n\ninterface XMLSerializer {\n    serializeToString(root: Node): string;\n}\n\ndeclare var XMLSerializer: {\n    prototype: XMLSerializer;\n    new(): XMLSerializer;\n};\n\ninterface XPathEvaluator {\n    createExpression(expression: string, resolver: XPathNSResolver): XPathExpression;\n    createNSResolver(nodeResolver?: Node): XPathNSResolver;\n    evaluate(expression: string, contextNode: Node, resolver: XPathNSResolver | ((prefix: string) => string | null) | null, type: number, result: XPathResult | null): XPathResult;\n}\n\ndeclare var XPathEvaluator: {\n    prototype: XPathEvaluator;\n    new(): XPathEvaluator;\n};\n\ninterface XPathExpression {\n    evaluate(contextNode: Node, type: number, result: XPathResult | null): XPathResult;\n}\n\ndeclare var XPathExpression: {\n    prototype: XPathExpression;\n    new(): XPathExpression;\n};\n\ninterface XPathNSResolver {\n    lookupNamespaceURI(prefix: string): string | null;\n}\n\ndeclare var XPathNSResolver: {\n    prototype: XPathNSResolver;\n    new(): XPathNSResolver;\n};\n\ninterface XPathResult {\n    readonly booleanValue: boolean;\n    readonly invalidIteratorState: boolean;\n    readonly numberValue: number;\n    readonly resultType: number;\n    readonly singleNodeValue: Node;\n    readonly snapshotLength: number;\n    readonly stringValue: string;\n    iterateNext(): Node;\n    snapshotItem(index: number): Node;\n    readonly ANY_TYPE: number;\n    readonly ANY_UNORDERED_NODE_TYPE: number;\n    readonly BOOLEAN_TYPE: number;\n    readonly FIRST_ORDERED_NODE_TYPE: number;\n    readonly NUMBER_TYPE: number;\n    readonly ORDERED_NODE_ITERATOR_TYPE: number;\n    readonly ORDERED_NODE_SNAPSHOT_TYPE: number;\n    readonly STRING_TYPE: number;\n    readonly UNORDERED_NODE_ITERATOR_TYPE: number;\n    readonly UNORDERED_NODE_SNAPSHOT_TYPE: number;\n}\n\ndeclare var XPathResult: {\n    prototype: XPathResult;\n    new(): XPathResult;\n    readonly ANY_TYPE: number;\n    readonly ANY_UNORDERED_NODE_TYPE: number;\n    readonly BOOLEAN_TYPE: number;\n    readonly FIRST_ORDERED_NODE_TYPE: number;\n    readonly NUMBER_TYPE: number;\n    readonly ORDERED_NODE_ITERATOR_TYPE: number;\n    readonly ORDERED_NODE_SNAPSHOT_TYPE: number;\n    readonly STRING_TYPE: number;\n    readonly UNORDERED_NODE_ITERATOR_TYPE: number;\n    readonly UNORDERED_NODE_SNAPSHOT_TYPE: number;\n};\n\ninterface XSLTProcessor {\n    clearParameters(): void;\n    getParameter(namespaceURI: string, localName: string): any;\n    importStylesheet(style: Node): void;\n    removeParameter(namespaceURI: string, localName: string): void;\n    reset(): void;\n    setParameter(namespaceURI: string, localName: string, value: any): void;\n    transformToDocument(source: Node): Document;\n    transformToFragment(source: Node, document: Document): DocumentFragment;\n}\n\ndeclare var XSLTProcessor: {\n    prototype: XSLTProcessor;\n    new(): XSLTProcessor;\n};\n\ninterface webkitRTCPeerConnection extends RTCPeerConnection {\n    addEventListener<K extends keyof RTCPeerConnectionEventMap>(type: K, listener: (this: webkitRTCPeerConnection, ev: RTCPeerConnectionEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCPeerConnectionEventMap>(type: K, listener: (this: webkitRTCPeerConnection, ev: RTCPeerConnectionEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var webkitRTCPeerConnection: {\n    prototype: webkitRTCPeerConnection;\n    new(configuration: RTCConfiguration): webkitRTCPeerConnection;\n};\n\ndeclare type EventListenerOrEventListenerObject = EventListener | EventListenerObject;\n\ninterface BlobCallback {\n    (blob: Blob | null): void;\n}\n\ninterface DecodeErrorCallback {\n    (error: DOMException): void;\n}\n\ninterface DecodeSuccessCallback {\n    (decodedData: AudioBuffer): void;\n}\n\ninterface ErrorEventHandler {\n    (event: Event | string, source?: string, fileno?: number, columnNumber?: number, error?: Error): void;\n}\n\ninterface EventHandlerNonNull {\n    (event: Event): any;\n}\n\ninterface ForEachCallback {\n    (keyId: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null, status: MediaKeyStatus): void;\n}\n\ninterface FrameRequestCallback {\n    (time: number): void;\n}\n\ninterface FunctionStringCallback {\n    (data: string): void;\n}\n\ninterface IntersectionObserverCallback {\n    (entries: IntersectionObserverEntry[], observer: IntersectionObserver): void;\n}\n\ninterface MSLaunchUriCallback {\n    (): void;\n}\n\ninterface MutationCallback {\n    (mutations: MutationRecord[], observer: MutationObserver): void;\n}\n\ninterface NavigatorUserMediaErrorCallback {\n    (error: MediaStreamError): void;\n}\n\ninterface NavigatorUserMediaSuccessCallback {\n    (stream: MediaStream): void;\n}\n\ninterface NotificationPermissionCallback {\n    (permission: NotificationPermission): void;\n}\n\ninterface OnBeforeUnloadEventHandlerNonNull {\n    (event: Event): string | null;\n}\n\ninterface OnErrorEventHandlerNonNull {\n    (event: Event | string, source?: string, lineno?: number, colno?: number, error?: any): any;\n}\n\ninterface PerformanceObserverCallback {\n    (entries: PerformanceObserverEntryList, observer: PerformanceObserver): void;\n}\n\ninterface PositionCallback {\n    (position: Position): void;\n}\n\ninterface PositionErrorCallback {\n    (positionError: PositionError): void;\n}\n\ninterface QueuingStrategySizeCallback<T = any> {\n    (chunk: T): number;\n}\n\ninterface RTCPeerConnectionErrorCallback {\n    (error: DOMException): void;\n}\n\ninterface RTCSessionDescriptionCallback {\n    (description: RTCSessionDescriptionInit): void;\n}\n\ninterface RTCStatsCallback {\n    (report: RTCStatsReport): void;\n}\n\ninterface ReadableByteStreamControllerCallback {\n    (controller: ReadableByteStreamController): void | PromiseLike<void>;\n}\n\ninterface ReadableStreamDefaultControllerCallback<R> {\n    (controller: ReadableStreamDefaultController<R>): void | PromiseLike<void>;\n}\n\ninterface ReadableStreamErrorCallback {\n    (reason: any): void | PromiseLike<void>;\n}\n\ninterface TransformStreamDefaultControllerCallback<O> {\n    (controller: TransformStreamDefaultController<O>): void | PromiseLike<void>;\n}\n\ninterface TransformStreamDefaultControllerTransformCallback<I, O> {\n    (chunk: I, controller: TransformStreamDefaultController<O>): void | PromiseLike<void>;\n}\n\ninterface VoidFunction {\n    (): void;\n}\n\ninterface WritableStreamDefaultControllerCloseCallback {\n    (): void | PromiseLike<void>;\n}\n\ninterface WritableStreamDefaultControllerStartCallback {\n    (controller: WritableStreamDefaultController): void | PromiseLike<void>;\n}\n\ninterface WritableStreamDefaultControllerWriteCallback<W> {\n    (chunk: W, controller: WritableStreamDefaultController): void | PromiseLike<void>;\n}\n\ninterface WritableStreamErrorCallback {\n    (reason: any): void | PromiseLike<void>;\n}\n\ninterface HTMLElementTagNameMap {\n    "a": HTMLAnchorElement;\n    "abbr": HTMLElement;\n    "address": HTMLElement;\n    "applet": HTMLAppletElement;\n    "area": HTMLAreaElement;\n    "article": HTMLElement;\n    "aside": HTMLElement;\n    "audio": HTMLAudioElement;\n    "b": HTMLElement;\n    "base": HTMLBaseElement;\n    "basefont": HTMLBaseFontElement;\n    "bdo": HTMLElement;\n    "blockquote": HTMLQuoteElement;\n    "body": HTMLBodyElement;\n    "br": HTMLBRElement;\n    "button": HTMLButtonElement;\n    "canvas": HTMLCanvasElement;\n    "caption": HTMLTableCaptionElement;\n    "cite": HTMLElement;\n    "code": HTMLElement;\n    "col": HTMLTableColElement;\n    "colgroup": HTMLTableColElement;\n    "data": HTMLDataElement;\n    "datalist": HTMLDataListElement;\n    "dd": HTMLElement;\n    "del": HTMLModElement;\n    "details": HTMLDetailsElement;\n    "dfn": HTMLElement;\n    "dialog": HTMLDialogElement;\n    "dir": HTMLDirectoryElement;\n    "div": HTMLDivElement;\n    "dl": HTMLDListElement;\n    "dt": HTMLElement;\n    "em": HTMLElement;\n    "embed": HTMLEmbedElement;\n    "fieldset": HTMLFieldSetElement;\n    "figcaption": HTMLElement;\n    "figure": HTMLElement;\n    "font": HTMLFontElement;\n    "footer": HTMLElement;\n    "form": HTMLFormElement;\n    "frame": HTMLFrameElement;\n    "frameset": HTMLFrameSetElement;\n    "h1": HTMLHeadingElement;\n    "h2": HTMLHeadingElement;\n    "h3": HTMLHeadingElement;\n    "h4": HTMLHeadingElement;\n    "h5": HTMLHeadingElement;\n    "h6": HTMLHeadingElement;\n    "head": HTMLHeadElement;\n    "header": HTMLElement;\n    "hgroup": HTMLElement;\n    "hr": HTMLHRElement;\n    "html": HTMLHtmlElement;\n    "i": HTMLElement;\n    "iframe": HTMLIFrameElement;\n    "img": HTMLImageElement;\n    "input": HTMLInputElement;\n    "ins": HTMLModElement;\n    "kbd": HTMLElement;\n    "label": HTMLLabelElement;\n    "legend": HTMLLegendElement;\n    "li": HTMLLIElement;\n    "link": HTMLLinkElement;\n    "map": HTMLMapElement;\n    "mark": HTMLElement;\n    "marquee": HTMLMarqueeElement;\n    "menu": HTMLMenuElement;\n    "meta": HTMLMetaElement;\n    "meter": HTMLMeterElement;\n    "nav": HTMLElement;\n    "noscript": HTMLElement;\n    "object": HTMLObjectElement;\n    "ol": HTMLOListElement;\n    "optgroup": HTMLOptGroupElement;\n    "option": HTMLOptionElement;\n    "output": HTMLOutputElement;\n    "p": HTMLParagraphElement;\n    "param": HTMLParamElement;\n    "picture": HTMLPictureElement;\n    "pre": HTMLPreElement;\n    "progress": HTMLProgressElement;\n    "q": HTMLQuoteElement;\n    "rt": HTMLElement;\n    "ruby": HTMLElement;\n    "s": HTMLElement;\n    "samp": HTMLElement;\n    "script": HTMLScriptElement;\n    "section": HTMLElement;\n    "select": HTMLSelectElement;\n    "slot": HTMLSlotElement;\n    "small": HTMLElement;\n    "source": HTMLSourceElement;\n    "span": HTMLSpanElement;\n    "strong": HTMLElement;\n    "style": HTMLStyleElement;\n    "sub": HTMLElement;\n    "sup": HTMLElement;\n    "table": HTMLTableElement;\n    "tbody": HTMLTableSectionElement;\n    "td": HTMLTableDataCellElement;\n    "template": HTMLTemplateElement;\n    "textarea": HTMLTextAreaElement;\n    "tfoot": HTMLTableSectionElement;\n    "th": HTMLTableHeaderCellElement;\n    "thead": HTMLTableSectionElement;\n    "time": HTMLTimeElement;\n    "title": HTMLTitleElement;\n    "tr": HTMLTableRowElement;\n    "track": HTMLTrackElement;\n    "u": HTMLElement;\n    "ul": HTMLUListElement;\n    "var": HTMLElement;\n    "video": HTMLVideoElement;\n    "wbr": HTMLElement;\n}\n\ninterface HTMLElementDeprecatedTagNameMap {\n    "listing": HTMLPreElement;\n    "xmp": HTMLPreElement;\n}\n\ninterface SVGElementTagNameMap {\n    "circle": SVGCircleElement;\n    "clipPath": SVGClipPathElement;\n    "defs": SVGDefsElement;\n    "desc": SVGDescElement;\n    "ellipse": SVGEllipseElement;\n    "feBlend": SVGFEBlendElement;\n    "feColorMatrix": SVGFEColorMatrixElement;\n    "feComponentTransfer": SVGFEComponentTransferElement;\n    "feComposite": SVGFECompositeElement;\n    "feConvolveMatrix": SVGFEConvolveMatrixElement;\n    "feDiffuseLighting": SVGFEDiffuseLightingElement;\n    "feDisplacementMap": SVGFEDisplacementMapElement;\n    "feDistantLight": SVGFEDistantLightElement;\n    "feFlood": SVGFEFloodElement;\n    "feFuncA": SVGFEFuncAElement;\n    "feFuncB": SVGFEFuncBElement;\n    "feFuncG": SVGFEFuncGElement;\n    "feFuncR": SVGFEFuncRElement;\n    "feGaussianBlur": SVGFEGaussianBlurElement;\n    "feImage": SVGFEImageElement;\n    "feMerge": SVGFEMergeElement;\n    "feMergeNode": SVGFEMergeNodeElement;\n    "feMorphology": SVGFEMorphologyElement;\n    "feOffset": SVGFEOffsetElement;\n    "fePointLight": SVGFEPointLightElement;\n    "feSpecularLighting": SVGFESpecularLightingElement;\n    "feSpotLight": SVGFESpotLightElement;\n    "feTile": SVGFETileElement;\n    "feTurbulence": SVGFETurbulenceElement;\n    "filter": SVGFilterElement;\n    "foreignObject": SVGForeignObjectElement;\n    "g": SVGGElement;\n    "image": SVGImageElement;\n    "line": SVGLineElement;\n    "linearGradient": SVGLinearGradientElement;\n    "marker": SVGMarkerElement;\n    "mask": SVGMaskElement;\n    "metadata": SVGMetadataElement;\n    "path": SVGPathElement;\n    "pattern": SVGPatternElement;\n    "polygon": SVGPolygonElement;\n    "polyline": SVGPolylineElement;\n    "radialGradient": SVGRadialGradientElement;\n    "rect": SVGRectElement;\n    "stop": SVGStopElement;\n    "svg": SVGSVGElement;\n    "switch": SVGSwitchElement;\n    "symbol": SVGSymbolElement;\n    "text": SVGTextElement;\n    "textPath": SVGTextPathElement;\n    "tspan": SVGTSpanElement;\n    "use": SVGUseElement;\n    "view": SVGViewElement;\n}\n\n/** @deprecated Directly use HTMLElementTagNameMap or SVGElementTagNameMap as appropriate, instead. */\ninterface ElementTagNameMap extends HTMLElementTagNameMap, SVGElementTagNameMap { }\n\ndeclare var Audio: {\n    new(src?: string): HTMLAudioElement;\n};\ndeclare var Image: {\n    new(width?: number, height?: number): HTMLImageElement;\n};\ndeclare var Option: {\n    new(text?: string, value?: string, defaultSelected?: boolean, selected?: boolean): HTMLOptionElement;\n};\ndeclare var Blob: typeof Blob;\ndeclare var URL: typeof URL;\ndeclare var URLSearchParams: typeof URLSearchParams;\ndeclare var applicationCache: ApplicationCache;\ndeclare var caches: CacheStorage;\ndeclare var clientInformation: Navigator;\ndeclare var closed: boolean;\ndeclare var crypto: Crypto;\ndeclare var customElements: CustomElementRegistry;\ndeclare var defaultStatus: string;\ndeclare var devicePixelRatio: number;\ndeclare var doNotTrack: string;\ndeclare var document: Document;\ndeclare var event: Event | undefined;\n/** @deprecated */\ndeclare var external: External;\ndeclare var frameElement: Element;\ndeclare var frames: Window;\ndeclare var history: History;\ndeclare var innerHeight: number;\ndeclare var innerWidth: number;\ndeclare var isSecureContext: boolean;\ndeclare var length: number;\ndeclare var location: Location;\ndeclare var locationbar: BarProp;\ndeclare var menubar: BarProp;\ndeclare var msContentScript: ExtensionScriptApis;\ndeclare const name: never;\ndeclare var navigator: Navigator;\ndeclare var offscreenBuffering: string | boolean;\ndeclare var oncompassneedscalibration: ((this: Window, ev: Event) => any) | null;\ndeclare var ondevicelight: ((this: Window, ev: DeviceLightEvent) => any) | null;\ndeclare var ondevicemotion: ((this: Window, ev: DeviceMotionEvent) => any) | null;\ndeclare var ondeviceorientation: ((this: Window, ev: DeviceOrientationEvent) => any) | null;\ndeclare var onmousewheel: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgesturechange: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgesturedoubletap: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgestureend: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgesturehold: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgesturestart: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgesturetap: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsinertiastart: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointercancel: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerdown: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerenter: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerleave: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointermove: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerout: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerover: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerup: ((this: Window, ev: Event) => any) | null;\n/** @deprecated */\ndeclare var onorientationchange: ((this: Window, ev: Event) => any) | null;\ndeclare var onreadystatechange: ((this: Window, ev: ProgressEvent) => any) | null;\ndeclare var onvrdisplayactivate: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplayblur: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplayconnect: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplaydeactivate: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplaydisconnect: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplayfocus: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplaypointerrestricted: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplaypointerunrestricted: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplaypresentchange: ((this: Window, ev: Event) => any) | null;\ndeclare var opener: any;\n/** @deprecated */\ndeclare var orientation: string | number;\ndeclare var outerHeight: number;\ndeclare var outerWidth: number;\ndeclare var pageXOffset: number;\ndeclare var pageYOffset: number;\ndeclare var parent: Window;\ndeclare var performance: Performance;\ndeclare var personalbar: BarProp;\ndeclare var screen: Screen;\ndeclare var screenLeft: number;\ndeclare var screenTop: number;\ndeclare var screenX: number;\ndeclare var screenY: number;\ndeclare var scrollX: number;\ndeclare var scrollY: number;\ndeclare var scrollbars: BarProp;\ndeclare var self: Window;\ndeclare var speechSynthesis: SpeechSynthesis;\ndeclare var status: string;\ndeclare var statusbar: BarProp;\ndeclare var styleMedia: StyleMedia;\ndeclare var toolbar: BarProp;\ndeclare var top: Window;\ndeclare var window: Window;\ndeclare function alert(message?: any): void;\ndeclare function blur(): void;\ndeclare function cancelAnimationFrame(handle: number): void;\n/** @deprecated */\ndeclare function captureEvents(): void;\ndeclare function close(): void;\ndeclare function confirm(message?: string): boolean;\ndeclare function departFocus(navigationReason: NavigationReason, origin: FocusNavigationOrigin): void;\ndeclare function focus(): void;\ndeclare function getComputedStyle(elt: Element, pseudoElt?: string | null): CSSStyleDeclaration;\ndeclare function getMatchedCSSRules(elt: Element, pseudoElt?: string | null): CSSRuleList;\ndeclare function getSelection(): Selection;\ndeclare function matchMedia(query: string): MediaQueryList;\ndeclare function moveBy(x: number, y: number): void;\ndeclare function moveTo(x: number, y: number): void;\ndeclare function msWriteProfilerMark(profilerMarkName: string): void;\ndeclare function open(url?: string, target?: string, features?: string, replace?: boolean): Window | null;\ndeclare function postMessage(message: any, targetOrigin: string, transfer?: Transferable[]): void;\ndeclare function print(): void;\ndeclare function prompt(message?: string, _default?: string): string | null;\n/** @deprecated */\ndeclare function releaseEvents(): void;\ndeclare function requestAnimationFrame(callback: FrameRequestCallback): number;\ndeclare function resizeBy(x: number, y: number): void;\ndeclare function resizeTo(x: number, y: number): void;\ndeclare function scroll(options?: ScrollToOptions): void;\ndeclare function scroll(x: number, y: number): void;\ndeclare function scrollBy(options?: ScrollToOptions): void;\ndeclare function scrollBy(x: number, y: number): void;\ndeclare function scrollTo(options?: ScrollToOptions): void;\ndeclare function scrollTo(x: number, y: number): void;\ndeclare function stop(): void;\ndeclare function webkitCancelAnimationFrame(handle: number): void;\ndeclare function webkitConvertPointFromNodeToPage(node: Node, pt: WebKitPoint): WebKitPoint;\ndeclare function webkitConvertPointFromPageToNode(node: Node, pt: WebKitPoint): WebKitPoint;\ndeclare function webkitRequestAnimationFrame(callback: FrameRequestCallback): number;\ndeclare function toString(): string;\n/**\n * Dispatches a synthetic event event to target and returns true\n * if either event\'s cancelable attribute value is false or its preventDefault() method was not invoked, and false otherwise.\n */\ndeclare function dispatchEvent(event: Event): boolean;\ndeclare var sessionStorage: Storage;\ndeclare var localStorage: Storage;\ndeclare var console: Console;\n/**\n * Fires when the user aborts the download.\n * @param ev The event.\n */\ndeclare var onabort: ((this: Window, ev: UIEvent) => any) | null;\ndeclare var onanimationcancel: ((this: Window, ev: AnimationEvent) => any) | null;\ndeclare var onanimationend: ((this: Window, ev: AnimationEvent) => any) | null;\ndeclare var onanimationiteration: ((this: Window, ev: AnimationEvent) => any) | null;\ndeclare var onanimationstart: ((this: Window, ev: AnimationEvent) => any) | null;\ndeclare var onauxclick: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the object loses the input focus.\n * @param ev The focus event.\n */\ndeclare var onblur: ((this: Window, ev: FocusEvent) => any) | null;\ndeclare var oncancel: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when playback is possible, but would require further buffering.\n * @param ev The event.\n */\ndeclare var oncanplay: ((this: Window, ev: Event) => any) | null;\ndeclare var oncanplaythrough: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the contents of the object or selection have changed.\n * @param ev The event.\n */\ndeclare var onchange: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the user clicks the left mouse button on the object\n * @param ev The mouse event.\n */\ndeclare var onclick: ((this: Window, ev: MouseEvent) => any) | null;\ndeclare var onclose: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the user clicks the right mouse button in the client area, opening the context menu.\n * @param ev The mouse event.\n */\ndeclare var oncontextmenu: ((this: Window, ev: MouseEvent) => any) | null;\ndeclare var oncuechange: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the user double-clicks the object.\n * @param ev The mouse event.\n */\ndeclare var ondblclick: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Fires on the source object continuously during a drag operation.\n * @param ev The event.\n */\ndeclare var ondrag: ((this: Window, ev: DragEvent) => any) | null;\n/**\n * Fires on the source object when the user releases the mouse at the close of a drag operation.\n * @param ev The event.\n */\ndeclare var ondragend: ((this: Window, ev: DragEvent) => any) | null;\n/**\n * Fires on the target element when the user drags the object to a valid drop target.\n * @param ev The drag event.\n */\ndeclare var ondragenter: ((this: Window, ev: DragEvent) => any) | null;\ndeclare var ondragexit: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires on the target object when the user moves the mouse out of a valid drop target during a drag operation.\n * @param ev The drag event.\n */\ndeclare var ondragleave: ((this: Window, ev: DragEvent) => any) | null;\n/**\n * Fires on the target element continuously while the user drags the object over a valid drop target.\n * @param ev The event.\n */\ndeclare var ondragover: ((this: Window, ev: DragEvent) => any) | null;\n/**\n * Fires on the source object when the user starts to drag a text selection or selected object.\n * @param ev The event.\n */\ndeclare var ondragstart: ((this: Window, ev: DragEvent) => any) | null;\ndeclare var ondrop: ((this: Window, ev: DragEvent) => any) | null;\n/**\n * Occurs when the duration attribute is updated.\n * @param ev The event.\n */\ndeclare var ondurationchange: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the media element is reset to its initial state.\n * @param ev The event.\n */\ndeclare var onemptied: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the end of playback is reached.\n * @param ev The event\n */\ndeclare var onended: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when an error occurs during object loading.\n * @param ev The event.\n */\ndeclare var onerror: ErrorEventHandler;\n/**\n * Fires when the object receives focus.\n * @param ev The event.\n */\ndeclare var onfocus: ((this: Window, ev: FocusEvent) => any) | null;\ndeclare var ongotpointercapture: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var oninput: ((this: Window, ev: Event) => any) | null;\ndeclare var oninvalid: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the user presses a key.\n * @param ev The keyboard event\n */\ndeclare var onkeydown: ((this: Window, ev: KeyboardEvent) => any) | null;\n/**\n * Fires when the user presses an alphanumeric key.\n * @param ev The event.\n */\ndeclare var onkeypress: ((this: Window, ev: KeyboardEvent) => any) | null;\n/**\n * Fires when the user releases a key.\n * @param ev The keyboard event\n */\ndeclare var onkeyup: ((this: Window, ev: KeyboardEvent) => any) | null;\n/**\n * Fires immediately after the browser loads the object.\n * @param ev The event.\n */\ndeclare var onload: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when media data is loaded at the current playback position.\n * @param ev The event.\n */\ndeclare var onloadeddata: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the duration and dimensions of the media have been determined.\n * @param ev The event.\n */\ndeclare var onloadedmetadata: ((this: Window, ev: Event) => any) | null;\ndeclare var onloadend: ((this: Window, ev: ProgressEvent) => any) | null;\n/**\n * Occurs when Internet Explorer begins looking for media data.\n * @param ev The event.\n */\ndeclare var onloadstart: ((this: Window, ev: Event) => any) | null;\ndeclare var onlostpointercapture: ((this: Window, ev: PointerEvent) => any) | null;\n/**\n * Fires when the user clicks the object with either mouse button.\n * @param ev The mouse event.\n */\ndeclare var onmousedown: ((this: Window, ev: MouseEvent) => any) | null;\ndeclare var onmouseenter: ((this: Window, ev: MouseEvent) => any) | null;\ndeclare var onmouseleave: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Fires when the user moves the mouse over the object.\n * @param ev The mouse event.\n */\ndeclare var onmousemove: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Fires when the user moves the mouse pointer outside the boundaries of the object.\n * @param ev The mouse event.\n */\ndeclare var onmouseout: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Fires when the user moves the mouse pointer into the object.\n * @param ev The mouse event.\n */\ndeclare var onmouseover: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Fires when the user releases a mouse button while the mouse is over the object.\n * @param ev The mouse event.\n */\ndeclare var onmouseup: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Occurs when playback is paused.\n * @param ev The event.\n */\ndeclare var onpause: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the play method is requested.\n * @param ev The event.\n */\ndeclare var onplay: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the audio or video has started playing.\n * @param ev The event.\n */\ndeclare var onplaying: ((this: Window, ev: Event) => any) | null;\ndeclare var onpointercancel: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerdown: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerenter: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerleave: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointermove: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerout: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerover: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerup: ((this: Window, ev: PointerEvent) => any) | null;\n/**\n * Occurs to indicate progress while downloading media data.\n * @param ev The event.\n */\ndeclare var onprogress: ((this: Window, ev: ProgressEvent) => any) | null;\n/**\n * Occurs when the playback rate is increased or decreased.\n * @param ev The event.\n */\ndeclare var onratechange: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the user resets a form.\n * @param ev The event.\n */\ndeclare var onreset: ((this: Window, ev: Event) => any) | null;\ndeclare var onresize: ((this: Window, ev: UIEvent) => any) | null;\n/**\n * Fires when the user repositions the scroll box in the scroll bar on the object.\n * @param ev The event.\n */\ndeclare var onscroll: ((this: Window, ev: UIEvent) => any) | null;\ndeclare var onsecuritypolicyviolation: ((this: Window, ev: SecurityPolicyViolationEvent) => any) | null;\n/**\n * Occurs when the seek operation ends.\n * @param ev The event.\n */\ndeclare var onseeked: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the current playback position is moved.\n * @param ev The event.\n */\ndeclare var onseeking: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the current selection changes.\n * @param ev The event.\n */\ndeclare var onselect: ((this: Window, ev: UIEvent) => any) | null;\n/**\n * Occurs when the download has stopped.\n * @param ev The event.\n */\ndeclare var onstalled: ((this: Window, ev: Event) => any) | null;\ndeclare var onsubmit: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs if the load operation has been intentionally halted.\n * @param ev The event.\n */\ndeclare var onsuspend: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs to indicate the current playback position.\n * @param ev The event.\n */\ndeclare var ontimeupdate: ((this: Window, ev: Event) => any) | null;\ndeclare var ontoggle: ((this: Window, ev: Event) => any) | null;\ndeclare var ontouchcancel: ((this: Window, ev: TouchEvent) => any) | null;\ndeclare var ontouchend: ((this: Window, ev: TouchEvent) => any) | null;\ndeclare var ontouchmove: ((this: Window, ev: TouchEvent) => any) | null;\ndeclare var ontouchstart: ((this: Window, ev: TouchEvent) => any) | null;\ndeclare var ontransitioncancel: ((this: Window, ev: TransitionEvent) => any) | null;\ndeclare var ontransitionend: ((this: Window, ev: TransitionEvent) => any) | null;\ndeclare var ontransitionrun: ((this: Window, ev: TransitionEvent) => any) | null;\ndeclare var ontransitionstart: ((this: Window, ev: TransitionEvent) => any) | null;\n/**\n * Occurs when the volume is changed, or playback is muted or unmuted.\n * @param ev The event.\n */\ndeclare var onvolumechange: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when playback stops because the next frame of a video resource is not available.\n * @param ev The event.\n */\ndeclare var onwaiting: ((this: Window, ev: Event) => any) | null;\ndeclare var onwheel: ((this: Window, ev: WheelEvent) => any) | null;\ndeclare var indexedDB: IDBFactory;\ndeclare function atob(encodedString: string): string;\ndeclare function btoa(rawString: string): string;\ndeclare function fetch(input: RequestInfo, init?: RequestInit): Promise<Response>;\ndeclare var caches: CacheStorage;\ndeclare var crypto: Crypto;\ndeclare var indexedDB: IDBFactory;\ndeclare var origin: string;\ndeclare var performance: Performance;\ndeclare function atob(data: string): string;\ndeclare function btoa(data: string): string;\ndeclare function clearInterval(handle?: number): void;\ndeclare function clearTimeout(handle?: number): void;\ndeclare function createImageBitmap(image: ImageBitmapSource): Promise<ImageBitmap>;\ndeclare function createImageBitmap(image: ImageBitmapSource, sx: number, sy: number, sw: number, sh: number): Promise<ImageBitmap>;\ndeclare function fetch(input: RequestInfo, init?: RequestInit): Promise<Response>;\ndeclare function queueMicrotask(callback: Function): void;\ndeclare function setInterval(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\ndeclare function setTimeout(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\ndeclare var sessionStorage: Storage;\ndeclare var localStorage: Storage;\ndeclare var onafterprint: ((this: Window, ev: Event) => any) | null;\ndeclare var onbeforeprint: ((this: Window, ev: Event) => any) | null;\ndeclare var onbeforeunload: ((this: Window, ev: BeforeUnloadEvent) => any) | null;\ndeclare var onhashchange: ((this: Window, ev: HashChangeEvent) => any) | null;\ndeclare var onlanguagechange: ((this: Window, ev: Event) => any) | null;\ndeclare var onmessage: ((this: Window, ev: MessageEvent) => any) | null;\ndeclare var onmessageerror: ((this: Window, ev: MessageEvent) => any) | null;\ndeclare var onoffline: ((this: Window, ev: Event) => any) | null;\ndeclare var ononline: ((this: Window, ev: Event) => any) | null;\ndeclare var onpagehide: ((this: Window, ev: PageTransitionEvent) => any) | null;\ndeclare var onpageshow: ((this: Window, ev: PageTransitionEvent) => any) | null;\ndeclare var onpopstate: ((this: Window, ev: PopStateEvent) => any) | null;\ndeclare var onrejectionhandled: ((this: Window, ev: Event) => any) | null;\ndeclare var onstorage: ((this: Window, ev: StorageEvent) => any) | null;\ndeclare var onunhandledrejection: ((this: Window, ev: PromiseRejectionEvent) => any) | null;\ndeclare var onunload: ((this: Window, ev: Event) => any) | null;\ndeclare function addEventListener<K extends keyof WindowEventMap>(type: K, listener: (this: Window, ev: WindowEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\ndeclare function addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\ndeclare function removeEventListener<K extends keyof WindowEventMap>(type: K, listener: (this: Window, ev: WindowEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\ndeclare function removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\ntype BlobPart = BufferSource | Blob | string;\ntype HeadersInit = Headers | string[][] | Record<string, string>;\ntype BodyInit = Blob | BufferSource | FormData | URLSearchParams | ReadableStream<Uint8Array> | string;\ntype RequestInfo = Request | string;\ntype DOMHighResTimeStamp = number;\ntype RenderingContext = CanvasRenderingContext2D | ImageBitmapRenderingContext | WebGLRenderingContext;\ntype HTMLOrSVGImageElement = HTMLImageElement | SVGImageElement;\ntype CanvasImageSource = HTMLOrSVGImageElement | HTMLVideoElement | HTMLCanvasElement | ImageBitmap;\ntype MessageEventSource = WindowProxy | MessagePort | ServiceWorker;\ntype HTMLOrSVGScriptElement = HTMLScriptElement | SVGScriptElement;\ntype ImageBitmapSource = CanvasImageSource | Blob | ImageData;\ntype OnErrorEventHandler = OnErrorEventHandlerNonNull | null;\ntype OnBeforeUnloadEventHandler = OnBeforeUnloadEventHandlerNonNull | null;\ntype TimerHandler = string | Function;\ntype PerformanceEntryList = PerformanceEntry[];\ntype VibratePattern = number | number[];\ntype AlgorithmIdentifier = string | Algorithm;\ntype HashAlgorithmIdentifier = AlgorithmIdentifier;\ntype BigInteger = Uint8Array;\ntype NamedCurve = string;\ntype GLenum = number;\ntype GLboolean = boolean;\ntype GLbitfield = number;\ntype GLint = number;\ntype GLsizei = number;\ntype GLintptr = number;\ntype GLsizeiptr = number;\ntype GLuint = number;\ntype GLfloat = number;\ntype GLclampf = number;\ntype TexImageSource = ImageBitmap | ImageData | HTMLImageElement | HTMLCanvasElement | HTMLVideoElement;\ntype Float32List = Float32Array | GLfloat[];\ntype Int32List = Int32Array | GLint[];\ntype BufferSource = ArrayBufferView | ArrayBuffer;\ntype DOMTimeStamp = number;\ntype LineAndPositionSetting = number | AutoKeyword;\ntype FormDataEntryValue = File | string;\ntype InsertPosition = "beforebegin" | "afterbegin" | "beforeend" | "afterend";\ntype IDBValidKey = number | string | Date | BufferSource | IDBArrayKey;\ntype MutationRecordType = "attributes" | "characterData" | "childList";\ntype ConstrainBoolean = boolean | ConstrainBooleanParameters;\ntype ConstrainDOMString = string | string[] | ConstrainDOMStringParameters;\ntype ConstrainDouble = number | ConstrainDoubleRange;\ntype ConstrainLong = number | ConstrainLongRange;\ntype IDBKeyPath = string;\ntype Transferable = ArrayBuffer | MessagePort | ImageBitmap;\ntype RTCIceGatherCandidate = RTCIceCandidateDictionary | RTCIceCandidateComplete;\ntype RTCTransport = RTCDtlsTransport | RTCSrtpSdesTransport;\n/** @deprecated */\ntype MouseWheelEvent = WheelEvent;\ntype WindowProxy = Window;\ntype AlignSetting = "start" | "center" | "end" | "left" | "right";\ntype AnimationPlayState = "idle" | "running" | "paused" | "finished";\ntype AppendMode = "segments" | "sequence";\ntype AudioContextLatencyCategory = "balanced" | "interactive" | "playback";\ntype AudioContextState = "suspended" | "running" | "closed";\ntype AutoKeyword = "auto";\ntype AutomationRate = "a-rate" | "k-rate";\ntype BinaryType = "blob" | "arraybuffer";\ntype BiquadFilterType = "lowpass" | "highpass" | "bandpass" | "lowshelf" | "highshelf" | "peaking" | "notch" | "allpass";\ntype CanPlayTypeResult = "" | "maybe" | "probably";\ntype CanvasDirection = "ltr" | "rtl" | "inherit";\ntype CanvasFillRule = "nonzero" | "evenodd";\ntype CanvasLineCap = "butt" | "round" | "square";\ntype CanvasLineJoin = "round" | "bevel" | "miter";\ntype CanvasTextAlign = "start" | "end" | "left" | "right" | "center";\ntype CanvasTextBaseline = "top" | "hanging" | "middle" | "alphabetic" | "ideographic" | "bottom";\ntype ChannelCountMode = "max" | "clamped-max" | "explicit";\ntype ChannelInterpretation = "speakers" | "discrete";\ntype ClientTypes = "window" | "worker" | "sharedworker" | "all";\ntype CompositeOperation = "replace" | "add" | "accumulate";\ntype CompositeOperationOrAuto = "replace" | "add" | "accumulate" | "auto";\ntype DirectionSetting = "" | "rl" | "lr";\ntype DisplayCaptureSurfaceType = "monitor" | "window" | "application" | "browser";\ntype DistanceModelType = "linear" | "inverse" | "exponential";\ntype DocumentReadyState = "loading" | "interactive" | "complete";\ntype EndOfStreamError = "network" | "decode";\ntype EndingType = "transparent" | "native";\ntype FillMode = "none" | "forwards" | "backwards" | "both" | "auto";\ntype GamepadHand = "" | "left" | "right";\ntype GamepadHapticActuatorType = "vibration";\ntype GamepadInputEmulationType = "mouse" | "keyboard" | "gamepad";\ntype GamepadMappingType = "" | "standard";\ntype IDBCursorDirection = "next" | "nextunique" | "prev" | "prevunique";\ntype IDBRequestReadyState = "pending" | "done";\ntype IDBTransactionMode = "readonly" | "readwrite" | "versionchange";\ntype ImageSmoothingQuality = "low" | "medium" | "high";\ntype IterationCompositeOperation = "replace" | "accumulate";\ntype KeyFormat = "raw" | "spki" | "pkcs8" | "jwk";\ntype KeyType = "public" | "private" | "secret";\ntype KeyUsage = "encrypt" | "decrypt" | "sign" | "verify" | "deriveKey" | "deriveBits" | "wrapKey" | "unwrapKey";\ntype LineAlignSetting = "start" | "center" | "end";\ntype ListeningState = "inactive" | "active" | "disambiguation";\ntype MSCredentialType = "FIDO_2_0";\ntype MSTransportType = "Embedded" | "USB" | "NFC" | "BT";\ntype MSWebViewPermissionState = "unknown" | "defer" | "allow" | "deny";\ntype MSWebViewPermissionType = "geolocation" | "unlimitedIndexedDBQuota" | "media" | "pointerlock" | "webnotifications";\ntype MediaDeviceKind = "audioinput" | "audiooutput" | "videoinput";\ntype MediaKeyMessageType = "license-request" | "license-renewal" | "license-release" | "individualization-request";\ntype MediaKeySessionType = "temporary" | "persistent-license" | "persistent-release-message";\ntype MediaKeyStatus = "usable" | "expired" | "output-downscaled" | "output-not-allowed" | "status-pending" | "internal-error";\ntype MediaKeysRequirement = "required" | "optional" | "not-allowed";\ntype MediaStreamTrackState = "live" | "ended";\ntype NavigationReason = "up" | "down" | "left" | "right";\ntype NavigationType = "navigate" | "reload" | "back_forward" | "prerender";\ntype NotificationDirection = "auto" | "ltr" | "rtl";\ntype NotificationPermission = "default" | "denied" | "granted";\ntype OrientationLockType = "any" | "natural" | "landscape" | "portrait" | "portrait-primary" | "portrait-secondary" | "landscape-primary" | "landscape-secondary";\ntype OrientationType = "portrait-primary" | "portrait-secondary" | "landscape-primary" | "landscape-secondary";\ntype OscillatorType = "sine" | "square" | "sawtooth" | "triangle" | "custom";\ntype OverSampleType = "none" | "2x" | "4x";\ntype PanningModelType = "equalpower" | "HRTF";\ntype PaymentComplete = "success" | "fail" | "unknown";\ntype PaymentShippingType = "shipping" | "delivery" | "pickup";\ntype PlaybackDirection = "normal" | "reverse" | "alternate" | "alternate-reverse";\ntype PositionAlignSetting = "line-left" | "center" | "line-right" | "auto";\ntype PushEncryptionKeyName = "p256dh" | "auth";\ntype PushPermissionState = "denied" | "granted" | "prompt";\ntype RTCBundlePolicy = "balanced" | "max-compat" | "max-bundle";\ntype RTCDataChannelState = "connecting" | "open" | "closing" | "closed";\ntype RTCDegradationPreference = "maintain-framerate" | "maintain-resolution" | "balanced";\ntype RTCDtlsRole = "auto" | "client" | "server";\ntype RTCDtlsTransportState = "new" | "connecting" | "connected" | "closed" | "failed";\ntype RTCDtxStatus = "disabled" | "enabled";\ntype RTCErrorDetailType = "data-channel-failure" | "dtls-failure" | "fingerprint-failure" | "idp-bad-script-failure" | "idp-execution-failure" | "idp-load-failure" | "idp-need-login" | "idp-timeout" | "idp-tls-failure" | "idp-token-expired" | "idp-token-invalid" | "sctp-failure" | "sdp-syntax-error" | "hardware-encoder-not-available" | "hardware-encoder-error";\ntype RTCIceCandidateType = "host" | "srflx" | "prflx" | "relay";\ntype RTCIceComponent = "rtp" | "rtcp";\ntype RTCIceConnectionState = "new" | "checking" | "connected" | "completed" | "disconnected" | "failed" | "closed";\ntype RTCIceCredentialType = "password" | "oauth";\ntype RTCIceGatherPolicy = "all" | "nohost" | "relay";\ntype RTCIceGathererState = "new" | "gathering" | "complete";\ntype RTCIceGatheringState = "new" | "gathering" | "complete";\ntype RTCIceProtocol = "udp" | "tcp";\ntype RTCIceRole = "controlling" | "controlled";\ntype RTCIceTcpCandidateType = "active" | "passive" | "so";\ntype RTCIceTransportPolicy = "relay" | "all";\ntype RTCIceTransportState = "new" | "checking" | "connected" | "completed" | "disconnected" | "failed" | "closed";\ntype RTCPeerConnectionState = "new" | "connecting" | "connected" | "disconnected" | "failed" | "closed";\ntype RTCPriorityType = "very-low" | "low" | "medium" | "high";\ntype RTCRtcpMuxPolicy = "negotiate" | "require";\ntype RTCRtpTransceiverDirection = "sendrecv" | "sendonly" | "recvonly" | "inactive";\ntype RTCSctpTransportState = "connecting" | "connected" | "closed";\ntype RTCSdpType = "offer" | "pranswer" | "answer" | "rollback";\ntype RTCSignalingState = "stable" | "have-local-offer" | "have-remote-offer" | "have-local-pranswer" | "have-remote-pranswer" | "closed";\ntype RTCStatsIceCandidatePairState = "frozen" | "waiting" | "inprogress" | "failed" | "succeeded" | "cancelled";\ntype RTCStatsIceCandidateType = "host" | "serverreflexive" | "peerreflexive" | "relayed";\ntype RTCStatsType = "inboundrtp" | "outboundrtp" | "session" | "datachannel" | "track" | "transport" | "candidatepair" | "localcandidate" | "remotecandidate";\ntype ReadyState = "closed" | "open" | "ended";\ntype ReferrerPolicy = "" | "no-referrer" | "no-referrer-when-downgrade" | "origin-only" | "origin-when-cross-origin" | "unsafe-url";\ntype RequestCache = "default" | "no-store" | "reload" | "no-cache" | "force-cache" | "only-if-cached";\ntype RequestCredentials = "omit" | "same-origin" | "include";\ntype RequestDestination = "" | "audio" | "audioworklet" | "document" | "embed" | "font" | "image" | "manifest" | "object" | "paintworklet" | "report" | "script" | "sharedworker" | "style" | "track" | "video" | "worker" | "xslt";\ntype RequestMode = "navigate" | "same-origin" | "no-cors" | "cors";\ntype RequestRedirect = "follow" | "error" | "manual";\ntype ResponseType = "basic" | "cors" | "default" | "error" | "opaque" | "opaqueredirect";\ntype ScopedCredentialType = "ScopedCred";\ntype ScrollBehavior = "auto" | "smooth";\ntype ScrollLogicalPosition = "start" | "center" | "end" | "nearest";\ntype ScrollRestoration = "auto" | "manual";\ntype ScrollSetting = "" | "up";\ntype SelectionMode = "select" | "start" | "end" | "preserve";\ntype ServiceWorkerState = "installing" | "installed" | "activating" | "activated" | "redundant";\ntype ServiceWorkerUpdateViaCache = "imports" | "all" | "none";\ntype ShadowRootMode = "open" | "closed";\ntype SpeechRecognitionErrorCode = "no-speech" | "aborted" | "audio-capture" | "network" | "not-allowed" | "service-not-allowed" | "bad-grammar" | "language-not-supported";\ntype SpeechSynthesisErrorCode = "canceled" | "interrupted" | "audio-busy" | "audio-hardware" | "network" | "synthesis-unavailable" | "synthesis-failed" | "language-unavailable" | "voice-unavailable" | "text-too-long" | "invalid-argument";\ntype SupportedType = "text/html" | "text/xml" | "application/xml" | "application/xhtml+xml" | "image/svg+xml";\ntype TextTrackKind = "subtitles" | "captions" | "descriptions" | "chapters" | "metadata";\ntype TextTrackMode = "disabled" | "hidden" | "showing";\ntype TouchType = "direct" | "stylus";\ntype Transport = "usb" | "nfc" | "ble";\ntype VRDisplayEventReason = "mounted" | "navigation" | "requested" | "unmounted";\ntype VideoFacingModeEnum = "user" | "environment" | "left" | "right";\ntype VisibilityState = "hidden" | "visible" | "prerender";\ntype WebGLPowerPreference = "default" | "low-power" | "high-performance";\ntype WorkerType = "classic" | "module";\ntype XMLHttpRequestResponseType = "" | "arraybuffer" | "blob" | "document" | "json" | "text";\n',$n="/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the \"License\"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib=\"true\"/>\n\n\n/////////////////////////////\n/// ECMAScript APIs\n/////////////////////////////\n\ndeclare const NaN: number;\ndeclare const Infinity: number;\n\n/**\n  * Evaluates JavaScript code and executes it.\n  * @param x A String value that contains valid JavaScript code.\n  */\ndeclare function eval(x: string): any;\n\n/**\n  * Converts A string to an integer.\n  * @param s A string to convert into a number.\n  * @param radix A value between 2 and 36 that specifies the base of the number in numString.\n  * If this argument is not supplied, strings with a prefix of '0x' are considered hexadecimal.\n  * All other strings are considered decimal.\n  */\ndeclare function parseInt(s: string, radix?: number): number;\n\n/**\n  * Converts a string to a floating-point number.\n  * @param string A string that contains a floating-point number.\n  */\ndeclare function parseFloat(string: string): number;\n\n/**\n  * Returns a Boolean value that indicates whether a value is the reserved value NaN (not a number).\n  * @param number A numeric value.\n  */\ndeclare function isNaN(number: number): boolean;\n\n/**\n  * Determines whether a supplied number is finite.\n  * @param number Any numeric value.\n  */\ndeclare function isFinite(number: number): boolean;\n\n/**\n  * Gets the unencoded version of an encoded Uniform Resource Identifier (URI).\n  * @param encodedURI A value representing an encoded URI.\n  */\ndeclare function decodeURI(encodedURI: string): string;\n\n/**\n  * Gets the unencoded version of an encoded component of a Uniform Resource Identifier (URI).\n  * @param encodedURIComponent A value representing an encoded URI component.\n  */\ndeclare function decodeURIComponent(encodedURIComponent: string): string;\n\n/**\n  * Encodes a text string as a valid Uniform Resource Identifier (URI)\n  * @param uri A value representing an encoded URI.\n  */\ndeclare function encodeURI(uri: string): string;\n\n/**\n  * Encodes a text string as a valid component of a Uniform Resource Identifier (URI).\n  * @param uriComponent A value representing an encoded URI component.\n  */\ndeclare function encodeURIComponent(uriComponent: string): string;\n\n/**\n  * Computes a new string in which certain characters have been replaced by a hexadecimal escape sequence.\n  * @param string A string value\n  */\ndeclare function escape(string: string): string;\n\n/**\n  * Computes a new string in which hexadecimal escape sequences are replaced with the character that it represents.\n  * @param string A string value\n  */\ndeclare function unescape(string: string): string;\n\ninterface Symbol {\n  /** Returns a string representation of an object. */\n  toString(): string;\n\n  /** Returns the primitive value of the specified object. */\n  valueOf(): symbol;\n}\n\ndeclare type PropertyKey = string | number | symbol;\n\ninterface PropertyDescriptor {\n    configurable?: boolean;\n    enumerable?: boolean;\n    value?: any;\n    writable?: boolean;\n    get?(): any;\n    set?(v: any): void;\n}\n\ninterface PropertyDescriptorMap {\n    [s: string]: PropertyDescriptor;\n}\n\ninterface Object {\n    /** The initial value of Object.prototype.constructor is the standard built-in Object constructor. */\n    constructor: Function;\n\n    /** Returns a string representation of an object. */\n    toString(): string;\n\n    /** Returns a date converted to a string using the current locale. */\n    toLocaleString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): Object;\n\n    /**\n      * Determines whether an object has a property with the specified name.\n      * @param v A property name.\n      */\n    hasOwnProperty(v: PropertyKey): boolean;\n\n    /**\n      * Determines whether an object exists in another object's prototype chain.\n      * @param v Another object whose prototype chain is to be checked.\n      */\n    isPrototypeOf(v: Object): boolean;\n\n    /**\n      * Determines whether a specified property is enumerable.\n      * @param v A property name.\n      */\n    propertyIsEnumerable(v: PropertyKey): boolean;\n}\n\ninterface ObjectConstructor {\n    new(value?: any): Object;\n    (): any;\n    (value: any): any;\n\n    /** A reference to the prototype for a class of objects. */\n    readonly prototype: Object;\n\n    /**\n      * Returns the prototype of an object.\n      * @param o The object that references the prototype.\n      */\n    getPrototypeOf(o: any): any;\n\n    /**\n      * Gets the own property descriptor of the specified object.\n      * An own property descriptor is one that is defined directly on the object and is not inherited from the object's prototype.\n      * @param o Object that contains the property.\n      * @param p Name of the property.\n    */\n    getOwnPropertyDescriptor(o: any, p: PropertyKey): PropertyDescriptor | undefined;\n\n    /**\n      * Returns the names of the own properties of an object. The own properties of an object are those that are defined directly\n      * on that object, and are not inherited from the object's prototype. The properties of an object include both fields (objects) and functions.\n      * @param o Object that contains the own properties.\n      */\n    getOwnPropertyNames(o: any): string[];\n\n    /**\n      * Creates an object that has the specified prototype or that has null prototype.\n      * @param o Object to use as a prototype. May be null.\n      */\n    create(o: object | null): any;\n\n    /**\n      * Creates an object that has the specified prototype, and that optionally contains specified properties.\n      * @param o Object to use as a prototype. May be null\n      * @param properties JavaScript object that contains one or more property descriptors.\n      */\n    create(o: object | null, properties: PropertyDescriptorMap & ThisType<any>): any;\n\n    /**\n      * Adds a property to an object, or modifies attributes of an existing property.\n      * @param o Object on which to add or modify the property. This can be a native JavaScript object (that is, a user-defined object or a built in object) or a DOM object.\n      * @param p The property name.\n      * @param attributes Descriptor for the property. It can be for a data property or an accessor property.\n      */\n    defineProperty(o: any, p: PropertyKey, attributes: PropertyDescriptor & ThisType<any>): any;\n\n    /**\n      * Adds one or more properties to an object, and/or modifies attributes of existing properties.\n      * @param o Object on which to add or modify the properties. This can be a native JavaScript object or a DOM object.\n      * @param properties JavaScript object that contains one or more descriptor objects. Each descriptor object describes a data property or an accessor property.\n      */\n    defineProperties(o: any, properties: PropertyDescriptorMap & ThisType<any>): any;\n\n    /**\n      * Prevents the modification of attributes of existing properties, and prevents the addition of new properties.\n      * @param o Object on which to lock the attributes.\n      */\n    seal<T>(o: T): T;\n\n    /**\n      * Prevents the modification of existing property attributes and values, and prevents the addition of new properties.\n      * @param o Object on which to lock the attributes.\n      */\n    freeze<T>(a: T[]): ReadonlyArray<T>;\n\n    /**\n      * Prevents the modification of existing property attributes and values, and prevents the addition of new properties.\n      * @param o Object on which to lock the attributes.\n      */\n    freeze<T extends Function>(f: T): T;\n\n    /**\n      * Prevents the modification of existing property attributes and values, and prevents the addition of new properties.\n      * @param o Object on which to lock the attributes.\n      */\n    freeze<T>(o: T): Readonly<T>;\n\n    /**\n      * Prevents the addition of new properties to an object.\n      * @param o Object to make non-extensible.\n      */\n    preventExtensions<T>(o: T): T;\n\n    /**\n      * Returns true if existing property attributes cannot be modified in an object and new properties cannot be added to the object.\n      * @param o Object to test.\n      */\n    isSealed(o: any): boolean;\n\n    /**\n      * Returns true if existing property attributes and values cannot be modified in an object, and new properties cannot be added to the object.\n      * @param o Object to test.\n      */\n    isFrozen(o: any): boolean;\n\n    /**\n      * Returns a value that indicates whether new properties can be added to an object.\n      * @param o Object to test.\n      */\n    isExtensible(o: any): boolean;\n\n    /**\n      * Returns the names of the enumerable properties and methods of an object.\n      * @param o Object that contains the properties and methods. This can be an object that you created or an existing Document Object Model (DOM) object.\n      */\n    keys(o: {}): string[];\n}\n\n/**\n  * Provides functionality common to all JavaScript objects.\n  */\ndeclare const Object: ObjectConstructor;\n\n/**\n  * Creates a new function.\n  */\ninterface Function {\n    /**\n      * Calls the function, substituting the specified object for the this value of the function, and the specified array for the arguments of the function.\n      * @param thisArg The object to be used as the this object.\n      * @param argArray A set of arguments to be passed to the function.\n      */\n    apply(this: Function, thisArg: any, argArray?: any): any;\n\n    /**\n      * Calls a method of an object, substituting another object for the current object.\n      * @param thisArg The object to be used as the current object.\n      * @param argArray A list of arguments to be passed to the method.\n      */\n    call(this: Function, thisArg: any, ...argArray: any[]): any;\n\n    /**\n      * For a given function, creates a bound function that has the same body as the original function.\n      * The this object of the bound function is associated with the specified object, and has the specified initial parameters.\n      * @param thisArg An object to which the this keyword can refer inside the new function.\n      * @param argArray A list of arguments to be passed to the new function.\n      */\n    bind(this: Function, thisArg: any, ...argArray: any[]): any;\n\n    /** Returns a string representation of a function. */\n    toString(): string;\n\n    prototype: any;\n    readonly length: number;\n\n    // Non-standard extensions\n    arguments: any;\n    caller: Function;\n}\n\ninterface FunctionConstructor {\n    /**\n      * Creates a new function.\n      * @param args A list of arguments the function accepts.\n      */\n    new(...args: string[]): Function;\n    (...args: string[]): Function;\n    readonly prototype: Function;\n}\n\ndeclare const Function: FunctionConstructor;\n\n/**\n * Extracts the type of the 'this' parameter of a function type, or 'unknown' if the function type has no 'this' parameter.\n */\ntype ThisParameterType<T> = T extends (this: unknown, ...args: any[]) => any ? unknown : T extends (this: infer U, ...args: any[]) => any ? U : unknown;\n\n/**\n * Removes the 'this' parameter from a function type.\n */\ntype OmitThisParameter<T> = unknown extends ThisParameterType<T> ? T : T extends (...args: infer A) => infer R ? (...args: A) => R : T;\n\ninterface CallableFunction extends Function {\n    /**\n      * Calls the function with the specified object as the this value and the elements of specified array as the arguments.\n      * @param thisArg The object to be used as the this object.\n      * @param args An array of argument values to be passed to the function.\n      */\n    apply<T, R>(this: (this: T) => R, thisArg: T): R;\n    apply<T, A extends any[], R>(this: (this: T, ...args: A) => R, thisArg: T, args: A): R;\n\n    /**\n      * Calls the function with the specified object as the this value and the specified rest arguments as the arguments.\n      * @param thisArg The object to be used as the this object.\n      * @param args Argument values to be passed to the function.\n      */\n    call<T, A extends any[], R>(this: (this: T, ...args: A) => R, thisArg: T, ...args: A): R;\n\n    /**\n      * For a given function, creates a bound function that has the same body as the original function.\n      * The this object of the bound function is associated with the specified object, and has the specified initial parameters.\n      * @param thisArg The object to be used as the this object.\n      * @param args Arguments to bind to the parameters of the function.\n      */\n    bind<T>(this: T, thisArg: ThisParameterType<T>): OmitThisParameter<T>;\n    bind<T, A0, A extends any[], R>(this: (this: T, arg0: A0, ...args: A) => R, thisArg: T, arg0: A0): (...args: A) => R;\n    bind<T, A0, A1, A extends any[], R>(this: (this: T, arg0: A0, arg1: A1, ...args: A) => R, thisArg: T, arg0: A0, arg1: A1): (...args: A) => R;\n    bind<T, A0, A1, A2, A extends any[], R>(this: (this: T, arg0: A0, arg1: A1, arg2: A2, ...args: A) => R, thisArg: T, arg0: A0, arg1: A1, arg2: A2): (...args: A) => R;\n    bind<T, A0, A1, A2, A3, A extends any[], R>(this: (this: T, arg0: A0, arg1: A1, arg2: A2, arg3: A3, ...args: A) => R, thisArg: T, arg0: A0, arg1: A1, arg2: A2, arg3: A3): (...args: A) => R;\n    bind<T, AX, R>(this: (this: T, ...args: AX[]) => R, thisArg: T, ...args: AX[]): (...args: AX[]) => R;\n}\n\ninterface NewableFunction extends Function {\n    /**\n      * Calls the function with the specified object as the this value and the elements of specified array as the arguments.\n      * @param thisArg The object to be used as the this object.\n      * @param args An array of argument values to be passed to the function.\n      */\n    apply<T>(this: new () => T, thisArg: T): void;\n    apply<T, A extends any[]>(this: new (...args: A) => T, thisArg: T, args: A): void;\n\n    /**\n      * Calls the function with the specified object as the this value and the specified rest arguments as the arguments.\n      * @param thisArg The object to be used as the this object.\n      * @param args Argument values to be passed to the function.\n      */\n    call<T, A extends any[]>(this: new (...args: A) => T, thisArg: T, ...args: A): void;\n\n    /**\n      * For a given function, creates a bound function that has the same body as the original function.\n      * The this object of the bound function is associated with the specified object, and has the specified initial parameters.\n      * @param thisArg The object to be used as the this object.\n      * @param args Arguments to bind to the parameters of the function.\n      */\n    bind<T>(this: T, thisArg: any): T;\n    bind<A0, A extends any[], R>(this: new (arg0: A0, ...args: A) => R, thisArg: any, arg0: A0): new (...args: A) => R;\n    bind<A0, A1, A extends any[], R>(this: new (arg0: A0, arg1: A1, ...args: A) => R, thisArg: any, arg0: A0, arg1: A1): new (...args: A) => R;\n    bind<A0, A1, A2, A extends any[], R>(this: new (arg0: A0, arg1: A1, arg2: A2, ...args: A) => R, thisArg: any, arg0: A0, arg1: A1, arg2: A2): new (...args: A) => R;\n    bind<A0, A1, A2, A3, A extends any[], R>(this: new (arg0: A0, arg1: A1, arg2: A2, arg3: A3, ...args: A) => R, thisArg: any, arg0: A0, arg1: A1, arg2: A2, arg3: A3): new (...args: A) => R;\n    bind<AX, R>(this: new (...args: AX[]) => R, thisArg: any, ...args: AX[]): new (...args: AX[]) => R;\n}\n\ninterface IArguments {\n    [index: number]: any;\n    length: number;\n    callee: Function;\n}\n\ninterface String {\n    /** Returns a string representation of a string. */\n    toString(): string;\n\n    /**\n      * Returns the character at the specified index.\n      * @param pos The zero-based index of the desired character.\n      */\n    charAt(pos: number): string;\n\n    /**\n      * Returns the Unicode value of the character at the specified location.\n      * @param index The zero-based index of the desired character. If there is no character at the specified index, NaN is returned.\n      */\n    charCodeAt(index: number): number;\n\n    /**\n      * Returns a string that contains the concatenation of two or more strings.\n      * @param strings The strings to append to the end of the string.\n      */\n    concat(...strings: string[]): string;\n\n    /**\n      * Returns the position of the first occurrence of a substring.\n      * @param searchString The substring to search for in the string\n      * @param position The index at which to begin searching the String object. If omitted, search starts at the beginning of the string.\n      */\n    indexOf(searchString: string, position?: number): number;\n\n    /**\n      * Returns the last occurrence of a substring in the string.\n      * @param searchString The substring to search for.\n      * @param position The index at which to begin searching. If omitted, the search begins at the end of the string.\n      */\n    lastIndexOf(searchString: string, position?: number): number;\n\n    /**\n      * Determines whether two strings are equivalent in the current locale.\n      * @param that String to compare to target string\n      */\n    localeCompare(that: string): number;\n\n    /**\n      * Matches a string with a regular expression, and returns an array containing the results of that search.\n      * @param regexp A variable name or string literal containing the regular expression pattern and flags.\n      */\n    match(regexp: string | RegExp): RegExpMatchArray | null;\n\n    /**\n      * Replaces text in a string, using a regular expression or search string.\n      * @param searchValue A string to search for.\n      * @param replaceValue A string containing the text to replace for every successful match of searchValue in this string.\n      */\n    replace(searchValue: string | RegExp, replaceValue: string): string;\n\n    /**\n      * Replaces text in a string, using a regular expression or search string.\n      * @param searchValue A string to search for.\n      * @param replacer A function that returns the replacement text.\n      */\n    replace(searchValue: string | RegExp, replacer: (substring: string, ...args: any[]) => string): string;\n\n    /**\n      * Finds the first substring match in a regular expression search.\n      * @param regexp The regular expression pattern and applicable flags.\n      */\n    search(regexp: string | RegExp): number;\n\n    /**\n      * Returns a section of a string.\n      * @param start The index to the beginning of the specified portion of stringObj.\n      * @param end The index to the end of the specified portion of stringObj. The substring includes the characters up to, but not including, the character indicated by end.\n      * If this value is not specified, the substring continues to the end of stringObj.\n      */\n    slice(start?: number, end?: number): string;\n\n    /**\n      * Split a string into substrings using the specified separator and return them as an array.\n      * @param separator A string that identifies character or characters to use in separating the string. If omitted, a single-element array containing the entire string is returned.\n      * @param limit A value used to limit the number of elements returned in the array.\n      */\n    split(separator: string | RegExp, limit?: number): string[];\n\n    /**\n      * Returns the substring at the specified location within a String object.\n      * @param start The zero-based index number indicating the beginning of the substring.\n      * @param end Zero-based index number indicating the end of the substring. The substring includes the characters up to, but not including, the character indicated by end.\n      * If end is omitted, the characters from start through the end of the original string are returned.\n      */\n    substring(start: number, end?: number): string;\n\n    /** Converts all the alphabetic characters in a string to lowercase. */\n    toLowerCase(): string;\n\n    /** Converts all alphabetic characters to lowercase, taking into account the host environment's current locale. */\n    toLocaleLowerCase(): string;\n\n    /** Converts all the alphabetic characters in a string to uppercase. */\n    toUpperCase(): string;\n\n    /** Returns a string where all alphabetic characters have been converted to uppercase, taking into account the host environment's current locale. */\n    toLocaleUpperCase(): string;\n\n    /** Removes the leading and trailing white space and line terminator characters from a string. */\n    trim(): string;\n\n    /** Returns the length of a String object. */\n    readonly length: number;\n\n    // IE extensions\n    /**\n      * Gets a substring beginning at the specified location and having the specified length.\n      * @param from The starting position of the desired substring. The index of the first character in the string is zero.\n      * @param length The number of characters to include in the returned substring.\n      */\n    substr(from: number, length?: number): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): string;\n\n    readonly [index: number]: string;\n}\n\ninterface StringConstructor {\n    new(value?: any): String;\n    (value?: any): string;\n    readonly prototype: String;\n    fromCharCode(...codes: number[]): string;\n}\n\n/**\n  * Allows manipulation and formatting of text strings and determination and location of substrings within strings.\n  */\ndeclare const String: StringConstructor;\n\ninterface Boolean {\n    /** Returns the primitive value of the specified object. */\n    valueOf(): boolean;\n}\n\ninterface BooleanConstructor {\n    new(value?: any): Boolean;\n    (value?: any): boolean;\n    readonly prototype: Boolean;\n}\n\ndeclare const Boolean: BooleanConstructor;\n\ninterface Number {\n    /**\n      * Returns a string representation of an object.\n      * @param radix Specifies a radix for converting numeric values to strings. This value is only used for numbers.\n      */\n    toString(radix?: number): string;\n\n    /**\n      * Returns a string representing a number in fixed-point notation.\n      * @param fractionDigits Number of digits after the decimal point. Must be in the range 0 - 20, inclusive.\n      */\n    toFixed(fractionDigits?: number): string;\n\n    /**\n      * Returns a string containing a number represented in exponential notation.\n      * @param fractionDigits Number of digits after the decimal point. Must be in the range 0 - 20, inclusive.\n      */\n    toExponential(fractionDigits?: number): string;\n\n    /**\n      * Returns a string containing a number represented either in exponential or fixed-point notation with a specified number of digits.\n      * @param precision Number of significant digits. Must be in the range 1 - 21, inclusive.\n      */\n    toPrecision(precision?: number): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): number;\n}\n\ninterface NumberConstructor {\n    new(value?: any): Number;\n    (value?: any): number;\n    readonly prototype: Number;\n\n    /** The largest number that can be represented in JavaScript. Equal to approximately 1.79E+308. */\n    readonly MAX_VALUE: number;\n\n    /** The closest number to zero that can be represented in JavaScript. Equal to approximately 5.00E-324. */\n    readonly MIN_VALUE: number;\n\n    /**\n      * A value that is not a number.\n      * In equality comparisons, NaN does not equal any value, including itself. To test whether a value is equivalent to NaN, use the isNaN function.\n      */\n    readonly NaN: number;\n\n    /**\n      * A value that is less than the largest negative number that can be represented in JavaScript.\n      * JavaScript displays NEGATIVE_INFINITY values as -infinity.\n      */\n    readonly NEGATIVE_INFINITY: number;\n\n    /**\n      * A value greater than the largest number that can be represented in JavaScript.\n      * JavaScript displays POSITIVE_INFINITY values as infinity.\n      */\n    readonly POSITIVE_INFINITY: number;\n}\n\n/** An object that represents a number of any kind. All JavaScript numbers are 64-bit floating-point numbers. */\ndeclare const Number: NumberConstructor;\n\ninterface TemplateStringsArray extends ReadonlyArray<string> {\n    readonly raw: ReadonlyArray<string>;\n}\n\n/**\n * The type of `import.meta`.\n *\n * If you need to declare that a given property exists on `import.meta`,\n * this type may be augmented via interface merging.\n */\ninterface ImportMeta {\n}\n\ninterface Math {\n    /** The mathematical constant e. This is Euler's number, the base of natural logarithms. */\n    readonly E: number;\n    /** The natural logarithm of 10. */\n    readonly LN10: number;\n    /** The natural logarithm of 2. */\n    readonly LN2: number;\n    /** The base-2 logarithm of e. */\n    readonly LOG2E: number;\n    /** The base-10 logarithm of e. */\n    readonly LOG10E: number;\n    /** Pi. This is the ratio of the circumference of a circle to its diameter. */\n    readonly PI: number;\n    /** The square root of 0.5, or, equivalently, one divided by the square root of 2. */\n    readonly SQRT1_2: number;\n    /** The square root of 2. */\n    readonly SQRT2: number;\n    /**\n      * Returns the absolute value of a number (the value without regard to whether it is positive or negative).\n      * For example, the absolute value of -5 is the same as the absolute value of 5.\n      * @param x A numeric expression for which the absolute value is needed.\n      */\n    abs(x: number): number;\n    /**\n      * Returns the arc cosine (or inverse cosine) of a number.\n      * @param x A numeric expression.\n      */\n    acos(x: number): number;\n    /**\n      * Returns the arcsine of a number.\n      * @param x A numeric expression.\n      */\n    asin(x: number): number;\n    /**\n      * Returns the arctangent of a number.\n      * @param x A numeric expression for which the arctangent is needed.\n      */\n    atan(x: number): number;\n    /**\n      * Returns the angle (in radians) from the X axis to a point.\n      * @param y A numeric expression representing the cartesian y-coordinate.\n      * @param x A numeric expression representing the cartesian x-coordinate.\n      */\n    atan2(y: number, x: number): number;\n    /**\n      * Returns the smallest integer greater than or equal to its numeric argument.\n      * @param x A numeric expression.\n      */\n    ceil(x: number): number;\n    /**\n      * Returns the cosine of a number.\n      * @param x A numeric expression that contains an angle measured in radians.\n      */\n    cos(x: number): number;\n    /**\n      * Returns e (the base of natural logarithms) raised to a power.\n      * @param x A numeric expression representing the power of e.\n      */\n    exp(x: number): number;\n    /**\n      * Returns the greatest integer less than or equal to its numeric argument.\n      * @param x A numeric expression.\n      */\n    floor(x: number): number;\n    /**\n      * Returns the natural logarithm (base e) of a number.\n      * @param x A numeric expression.\n      */\n    log(x: number): number;\n    /**\n      * Returns the larger of a set of supplied numeric expressions.\n      * @param values Numeric expressions to be evaluated.\n      */\n    max(...values: number[]): number;\n    /**\n      * Returns the smaller of a set of supplied numeric expressions.\n      * @param values Numeric expressions to be evaluated.\n      */\n    min(...values: number[]): number;\n    /**\n      * Returns the value of a base expression taken to a specified power.\n      * @param x The base value of the expression.\n      * @param y The exponent value of the expression.\n      */\n    pow(x: number, y: number): number;\n    /** Returns a pseudorandom number between 0 and 1. */\n    random(): number;\n    /**\n      * Returns a supplied numeric expression rounded to the nearest number.\n      * @param x The value to be rounded to the nearest number.\n      */\n    round(x: number): number;\n    /**\n      * Returns the sine of a number.\n      * @param x A numeric expression that contains an angle measured in radians.\n      */\n    sin(x: number): number;\n    /**\n      * Returns the square root of a number.\n      * @param x A numeric expression.\n      */\n    sqrt(x: number): number;\n    /**\n      * Returns the tangent of a number.\n      * @param x A numeric expression that contains an angle measured in radians.\n      */\n    tan(x: number): number;\n}\n/** An intrinsic object that provides basic mathematics functionality and constants. */\ndeclare const Math: Math;\n\n/** Enables basic storage and retrieval of dates and times. */\ninterface Date {\n    /** Returns a string representation of a date. The format of the string depends on the locale. */\n    toString(): string;\n    /** Returns a date as a string value. */\n    toDateString(): string;\n    /** Returns a time as a string value. */\n    toTimeString(): string;\n    /** Returns a value as a string value appropriate to the host environment's current locale. */\n    toLocaleString(): string;\n    /** Returns a date as a string value appropriate to the host environment's current locale. */\n    toLocaleDateString(): string;\n    /** Returns a time as a string value appropriate to the host environment's current locale. */\n    toLocaleTimeString(): string;\n    /** Returns the stored time value in milliseconds since midnight, January 1, 1970 UTC. */\n    valueOf(): number;\n    /** Gets the time value in milliseconds. */\n    getTime(): number;\n    /** Gets the year, using local time. */\n    getFullYear(): number;\n    /** Gets the year using Universal Coordinated Time (UTC). */\n    getUTCFullYear(): number;\n    /** Gets the month, using local time. */\n    getMonth(): number;\n    /** Gets the month of a Date object using Universal Coordinated Time (UTC). */\n    getUTCMonth(): number;\n    /** Gets the day-of-the-month, using local time. */\n    getDate(): number;\n    /** Gets the day-of-the-month, using Universal Coordinated Time (UTC). */\n    getUTCDate(): number;\n    /** Gets the day of the week, using local time. */\n    getDay(): number;\n    /** Gets the day of the week using Universal Coordinated Time (UTC). */\n    getUTCDay(): number;\n    /** Gets the hours in a date, using local time. */\n    getHours(): number;\n    /** Gets the hours value in a Date object using Universal Coordinated Time (UTC). */\n    getUTCHours(): number;\n    /** Gets the minutes of a Date object, using local time. */\n    getMinutes(): number;\n    /** Gets the minutes of a Date object using Universal Coordinated Time (UTC). */\n    getUTCMinutes(): number;\n    /** Gets the seconds of a Date object, using local time. */\n    getSeconds(): number;\n    /** Gets the seconds of a Date object using Universal Coordinated Time (UTC). */\n    getUTCSeconds(): number;\n    /** Gets the milliseconds of a Date, using local time. */\n    getMilliseconds(): number;\n    /** Gets the milliseconds of a Date object using Universal Coordinated Time (UTC). */\n    getUTCMilliseconds(): number;\n    /** Gets the difference in minutes between the time on the local computer and Universal Coordinated Time (UTC). */\n    getTimezoneOffset(): number;\n    /**\n      * Sets the date and time value in the Date object.\n      * @param time A numeric value representing the number of elapsed milliseconds since midnight, January 1, 1970 GMT.\n      */\n    setTime(time: number): number;\n    /**\n      * Sets the milliseconds value in the Date object using local time.\n      * @param ms A numeric value equal to the millisecond value.\n      */\n    setMilliseconds(ms: number): number;\n    /**\n      * Sets the milliseconds value in the Date object using Universal Coordinated Time (UTC).\n      * @param ms A numeric value equal to the millisecond value.\n      */\n    setUTCMilliseconds(ms: number): number;\n\n    /**\n      * Sets the seconds value in the Date object using local time.\n      * @param sec A numeric value equal to the seconds value.\n      * @param ms A numeric value equal to the milliseconds value.\n      */\n    setSeconds(sec: number, ms?: number): number;\n    /**\n      * Sets the seconds value in the Date object using Universal Coordinated Time (UTC).\n      * @param sec A numeric value equal to the seconds value.\n      * @param ms A numeric value equal to the milliseconds value.\n      */\n    setUTCSeconds(sec: number, ms?: number): number;\n    /**\n      * Sets the minutes value in the Date object using local time.\n      * @param min A numeric value equal to the minutes value.\n      * @param sec A numeric value equal to the seconds value.\n      * @param ms A numeric value equal to the milliseconds value.\n      */\n    setMinutes(min: number, sec?: number, ms?: number): number;\n    /**\n      * Sets the minutes value in the Date object using Universal Coordinated Time (UTC).\n      * @param min A numeric value equal to the minutes value.\n      * @param sec A numeric value equal to the seconds value.\n      * @param ms A numeric value equal to the milliseconds value.\n      */\n    setUTCMinutes(min: number, sec?: number, ms?: number): number;\n    /**\n      * Sets the hour value in the Date object using local time.\n      * @param hours A numeric value equal to the hours value.\n      * @param min A numeric value equal to the minutes value.\n      * @param sec A numeric value equal to the seconds value.\n      * @param ms A numeric value equal to the milliseconds value.\n      */\n    setHours(hours: number, min?: number, sec?: number, ms?: number): number;\n    /**\n      * Sets the hours value in the Date object using Universal Coordinated Time (UTC).\n      * @param hours A numeric value equal to the hours value.\n      * @param min A numeric value equal to the minutes value.\n      * @param sec A numeric value equal to the seconds value.\n      * @param ms A numeric value equal to the milliseconds value.\n      */\n    setUTCHours(hours: number, min?: number, sec?: number, ms?: number): number;\n    /**\n      * Sets the numeric day-of-the-month value of the Date object using local time.\n      * @param date A numeric value equal to the day of the month.\n      */\n    setDate(date: number): number;\n    /**\n      * Sets the numeric day of the month in the Date object using Universal Coordinated Time (UTC).\n      * @param date A numeric value equal to the day of the month.\n      */\n    setUTCDate(date: number): number;\n    /**\n      * Sets the month value in the Date object using local time.\n      * @param month A numeric value equal to the month. The value for January is 0, and other month values follow consecutively.\n      * @param date A numeric value representing the day of the month. If this value is not supplied, the value from a call to the getDate method is used.\n      */\n    setMonth(month: number, date?: number): number;\n    /**\n      * Sets the month value in the Date object using Universal Coordinated Time (UTC).\n      * @param month A numeric value equal to the month. The value for January is 0, and other month values follow consecutively.\n      * @param date A numeric value representing the day of the month. If it is not supplied, the value from a call to the getUTCDate method is used.\n      */\n    setUTCMonth(month: number, date?: number): number;\n    /**\n      * Sets the year of the Date object using local time.\n      * @param year A numeric value for the year.\n      * @param month A zero-based numeric value for the month (0 for January, 11 for December). Must be specified if numDate is specified.\n      * @param date A numeric value equal for the day of the month.\n      */\n    setFullYear(year: number, month?: number, date?: number): number;\n    /**\n      * Sets the year value in the Date object using Universal Coordinated Time (UTC).\n      * @param year A numeric value equal to the year.\n      * @param month A numeric value equal to the month. The value for January is 0, and other month values follow consecutively. Must be supplied if numDate is supplied.\n      * @param date A numeric value equal to the day of the month.\n      */\n    setUTCFullYear(year: number, month?: number, date?: number): number;\n    /** Returns a date converted to a string using Universal Coordinated Time (UTC). */\n    toUTCString(): string;\n    /** Returns a date as a string value in ISO format. */\n    toISOString(): string;\n    /** Used by the JSON.stringify method to enable the transformation of an object's data for JavaScript Object Notation (JSON) serialization. */\n    toJSON(key?: any): string;\n}\n\ninterface DateConstructor {\n    new(): Date;\n    new(value: number | string): Date;\n    new(year: number, month: number, date?: number, hours?: number, minutes?: number, seconds?: number, ms?: number): Date;\n    (): string;\n    readonly prototype: Date;\n    /**\n      * Parses a string containing a date, and returns the number of milliseconds between that date and midnight, January 1, 1970.\n      * @param s A date string\n      */\n    parse(s: string): number;\n    /**\n      * Returns the number of milliseconds between midnight, January 1, 1970 Universal Coordinated Time (UTC) (or GMT) and the specified date.\n      * @param year The full year designation is required for cross-century date accuracy. If year is between 0 and 99 is used, then year is assumed to be 1900 + year.\n      * @param month The month as an number between 0 and 11 (January to December).\n      * @param date The date as an number between 1 and 31.\n      * @param hours Must be supplied if minutes is supplied. An number from 0 to 23 (midnight to 11pm) that specifies the hour.\n      * @param minutes Must be supplied if seconds is supplied. An number from 0 to 59 that specifies the minutes.\n      * @param seconds Must be supplied if milliseconds is supplied. An number from 0 to 59 that specifies the seconds.\n      * @param ms An number from 0 to 999 that specifies the milliseconds.\n      */\n    UTC(year: number, month: number, date?: number, hours?: number, minutes?: number, seconds?: number, ms?: number): number;\n    now(): number;\n}\n\ndeclare const Date: DateConstructor;\n\ninterface RegExpMatchArray extends Array<string> {\n    index?: number;\n    input?: string;\n}\n\ninterface RegExpExecArray extends Array<string> {\n    index: number;\n    input: string;\n}\n\ninterface RegExp {\n    /**\n      * Executes a search on a string using a regular expression pattern, and returns an array containing the results of that search.\n      * @param string The String object or string literal on which to perform the search.\n      */\n    exec(string: string): RegExpExecArray | null;\n\n    /**\n      * Returns a Boolean value that indicates whether or not a pattern exists in a searched string.\n      * @param string String on which to perform the search.\n      */\n    test(string: string): boolean;\n\n    /** Returns a copy of the text of the regular expression pattern. Read-only. The regExp argument is a Regular expression object. It can be a variable name or a literal. */\n    readonly source: string;\n\n    /** Returns a Boolean value indicating the state of the global flag (g) used with a regular expression. Default is false. Read-only. */\n    readonly global: boolean;\n\n    /** Returns a Boolean value indicating the state of the ignoreCase flag (i) used with a regular expression. Default is false. Read-only. */\n    readonly ignoreCase: boolean;\n\n    /** Returns a Boolean value indicating the state of the multiline flag (m) used with a regular expression. Default is false. Read-only. */\n    readonly multiline: boolean;\n\n    lastIndex: number;\n\n    // Non-standard extensions\n    compile(): this;\n}\n\ninterface RegExpConstructor {\n    new(pattern: RegExp | string): RegExp;\n    new(pattern: string, flags?: string): RegExp;\n    (pattern: RegExp | string): RegExp;\n    (pattern: string, flags?: string): RegExp;\n    readonly prototype: RegExp;\n\n    // Non-standard extensions\n    $1: string;\n    $2: string;\n    $3: string;\n    $4: string;\n    $5: string;\n    $6: string;\n    $7: string;\n    $8: string;\n    $9: string;\n    lastMatch: string;\n}\n\ndeclare const RegExp: RegExpConstructor;\n\ninterface Error {\n    name: string;\n    message: string;\n    stack?: string;\n}\n\ninterface ErrorConstructor {\n    new(message?: string): Error;\n    (message?: string): Error;\n    readonly prototype: Error;\n}\n\ndeclare const Error: ErrorConstructor;\n\ninterface EvalError extends Error {\n}\n\ninterface EvalErrorConstructor {\n    new(message?: string): EvalError;\n    (message?: string): EvalError;\n    readonly prototype: EvalError;\n}\n\ndeclare const EvalError: EvalErrorConstructor;\n\ninterface RangeError extends Error {\n}\n\ninterface RangeErrorConstructor {\n    new(message?: string): RangeError;\n    (message?: string): RangeError;\n    readonly prototype: RangeError;\n}\n\ndeclare const RangeError: RangeErrorConstructor;\n\ninterface ReferenceError extends Error {\n}\n\ninterface ReferenceErrorConstructor {\n    new(message?: string): ReferenceError;\n    (message?: string): ReferenceError;\n    readonly prototype: ReferenceError;\n}\n\ndeclare const ReferenceError: ReferenceErrorConstructor;\n\ninterface SyntaxError extends Error {\n}\n\ninterface SyntaxErrorConstructor {\n    new(message?: string): SyntaxError;\n    (message?: string): SyntaxError;\n    readonly prototype: SyntaxError;\n}\n\ndeclare const SyntaxError: SyntaxErrorConstructor;\n\ninterface TypeError extends Error {\n}\n\ninterface TypeErrorConstructor {\n    new(message?: string): TypeError;\n    (message?: string): TypeError;\n    readonly prototype: TypeError;\n}\n\ndeclare const TypeError: TypeErrorConstructor;\n\ninterface URIError extends Error {\n}\n\ninterface URIErrorConstructor {\n    new(message?: string): URIError;\n    (message?: string): URIError;\n    readonly prototype: URIError;\n}\n\ndeclare const URIError: URIErrorConstructor;\n\ninterface JSON {\n    /**\n      * Converts a JavaScript Object Notation (JSON) string into an object.\n      * @param text A valid JSON string.\n      * @param reviver A function that transforms the results. This function is called for each member of the object.\n      * If a member contains nested objects, the nested objects are transformed before the parent object is.\n      */\n    parse(text: string, reviver?: (key: any, value: any) => any): any;\n    /**\n      * Converts a JavaScript value to a JavaScript Object Notation (JSON) string.\n      * @param value A JavaScript value, usually an object or array, to be converted.\n      * @param replacer A function that transforms the results.\n      * @param space Adds indentation, white space, and line break characters to the return-value JSON text to make it easier to read.\n      */\n    stringify(value: any, replacer?: (key: string, value: any) => any, space?: string | number): string;\n    /**\n      * Converts a JavaScript value to a JavaScript Object Notation (JSON) string.\n      * @param value A JavaScript value, usually an object or array, to be converted.\n      * @param replacer An array of strings and numbers that acts as a approved list for selecting the object properties that will be stringified.\n      * @param space Adds indentation, white space, and line break characters to the return-value JSON text to make it easier to read.\n      */\n    stringify(value: any, replacer?: (number | string)[] | null, space?: string | number): string;\n}\n\n/**\n  * An intrinsic object that provides functions to convert JavaScript values to and from the JavaScript Object Notation (JSON) format.\n  */\ndeclare const JSON: JSON;\n\n\n/////////////////////////////\n/// ECMAScript Array API (specially handled by compiler)\n/////////////////////////////\n\ninterface ReadonlyArray<T> {\n    /**\n      * Gets the length of the array. This is a number one higher than the highest element defined in an array.\n      */\n    readonly length: number;\n    /**\n      * Returns a string representation of an array.\n      */\n    toString(): string;\n    /**\n      * Returns a string representation of an array. The elements are converted to string using their toLocalString methods.\n      */\n    toLocaleString(): string;\n    /**\n      * Combines two or more arrays.\n      * @param items Additional items to add to the end of array1.\n      */\n    concat(...items: ConcatArray<T>[]): T[];\n    /**\n      * Combines two or more arrays.\n      * @param items Additional items to add to the end of array1.\n      */\n    concat(...items: (T | ConcatArray<T>)[]): T[];\n    /**\n      * Adds all the elements of an array separated by the specified separator string.\n      * @param separator A string used to separate one element of an array from the next in the resulting String. If omitted, the array elements are separated with a comma.\n      */\n    join(separator?: string): string;\n    /**\n      * Returns a section of an array.\n      * @param start The beginning of the specified portion of the array.\n      * @param end The end of the specified portion of the array.\n      */\n    slice(start?: number, end?: number): T[];\n    /**\n      * Returns the index of the first occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the search starts at index 0.\n      */\n    indexOf(searchElement: T, fromIndex?: number): number;\n    /**\n      * Returns the index of the last occurrence of a specified value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the search starts at the last index in the array.\n      */\n    lastIndexOf(searchElement: T, fromIndex?: number): number;\n    /**\n      * Determines whether all the members of an array satisfy the specified test.\n      * @param callbackfn A function that accepts up to three arguments. The every method calls the callbackfn function for each element in array1 until the callbackfn returns false, or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n      */\n    every(callbackfn: (value: T, index: number, array: ReadonlyArray<T>) => boolean, thisArg?: any): boolean;\n    /**\n      * Determines whether the specified callback function returns true for any element of an array.\n      * @param callbackfn A function that accepts up to three arguments. The some method calls the callbackfn function for each element in array1 until the callbackfn returns true, or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n      */\n    some(callbackfn: (value: T, index: number, array: ReadonlyArray<T>) => boolean, thisArg?: any): boolean;\n    /**\n      * Performs the specified action for each element in an array.\n      * @param callbackfn  A function that accepts up to three arguments. forEach calls the callbackfn function one time for each element in the array.\n      * @param thisArg  An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n      */\n    forEach(callbackfn: (value: T, index: number, array: ReadonlyArray<T>) => void, thisArg?: any): void;\n    /**\n      * Calls a defined callback function on each element of an array, and returns an array that contains the results.\n      * @param callbackfn A function that accepts up to three arguments. The map method calls the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n      */\n    map<U>(callbackfn: (value: T, index: number, array: ReadonlyArray<T>) => U, thisArg?: any): U[];\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param callbackfn A function that accepts up to three arguments. The filter method calls the callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n     */\n    filter<S extends T>(callbackfn: (value: T, index: number, array: ReadonlyArray<T>) => value is S, thisArg?: any): S[];\n    /**\n      * Returns the elements of an array that meet the condition specified in a callback function.\n      * @param callbackfn A function that accepts up to three arguments. The filter method calls the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n      */\n    filter(callbackfn: (value: T, index: number, array: ReadonlyArray<T>) => any, thisArg?: any): T[];\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n      */\n    reduce(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: ReadonlyArray<T>) => T): T;\n    reduce(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: ReadonlyArray<T>) => T, initialValue: T): T;\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n      */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: T, currentIndex: number, array: ReadonlyArray<T>) => U, initialValue: U): U;\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n      */\n    reduceRight(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: ReadonlyArray<T>) => T): T;\n    reduceRight(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: ReadonlyArray<T>) => T, initialValue: T): T;\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n      */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: T, currentIndex: number, array: ReadonlyArray<T>) => U, initialValue: U): U;\n\n    readonly [n: number]: T;\n}\n\ninterface ConcatArray<T> {\n    readonly length: number;\n    readonly [n: number]: T;\n    join(separator?: string): string;\n    slice(start?: number, end?: number): T[];\n}\n\ninterface Array<T> {\n    /**\n      * Gets or sets the length of the array. This is a number one higher than the highest element defined in an array.\n      */\n    length: number;\n    /**\n      * Returns a string representation of an array.\n      */\n    toString(): string;\n    /**\n      * Returns a string representation of an array. The elements are converted to string using their toLocalString methods.\n      */\n    toLocaleString(): string;\n    /**\n      * Removes the last element from an array and returns it.\n      */\n    pop(): T | undefined;\n    /**\n      * Appends new elements to an array, and returns the new length of the array.\n      * @param items New elements of the Array.\n      */\n    push(...items: T[]): number;\n    /**\n      * Combines two or more arrays.\n      * @param items Additional items to add to the end of array1.\n      */\n    concat(...items: ConcatArray<T>[]): T[];\n    /**\n      * Combines two or more arrays.\n      * @param items Additional items to add to the end of array1.\n      */\n    concat(...items: (T | ConcatArray<T>)[]): T[];\n    /**\n      * Adds all the elements of an array separated by the specified separator string.\n      * @param separator A string used to separate one element of an array from the next in the resulting String. If omitted, the array elements are separated with a comma.\n      */\n    join(separator?: string): string;\n    /**\n      * Reverses the elements in an Array.\n      */\n    reverse(): T[];\n    /**\n      * Removes the first element from an array and returns it.\n      */\n    shift(): T | undefined;\n    /**\n      * Returns a section of an array.\n      * @param start The beginning of the specified portion of the array.\n      * @param end The end of the specified portion of the array.\n      */\n    slice(start?: number, end?: number): T[];\n    /**\n      * Sorts an array.\n      * @param compareFn The name of the function used to determine the order of the elements. If omitted, the elements are sorted in ascending, ASCII character order.\n      */\n    sort(compareFn?: (a: T, b: T) => number): this;\n    /**\n      * Removes elements from an array and, if necessary, inserts new elements in their place, returning the deleted elements.\n      * @param start The zero-based location in the array from which to start removing elements.\n      * @param deleteCount The number of elements to remove.\n      */\n    splice(start: number, deleteCount?: number): T[];\n    /**\n      * Removes elements from an array and, if necessary, inserts new elements in their place, returning the deleted elements.\n      * @param start The zero-based location in the array from which to start removing elements.\n      * @param deleteCount The number of elements to remove.\n      * @param items Elements to insert into the array in place of the deleted elements.\n      */\n    splice(start: number, deleteCount: number, ...items: T[]): T[];\n    /**\n      * Inserts new elements at the start of an array.\n      * @param items  Elements to insert at the start of the Array.\n      */\n    unshift(...items: T[]): number;\n    /**\n      * Returns the index of the first occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the search starts at index 0.\n      */\n    indexOf(searchElement: T, fromIndex?: number): number;\n    /**\n      * Returns the index of the last occurrence of a specified value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the search starts at the last index in the array.\n      */\n    lastIndexOf(searchElement: T, fromIndex?: number): number;\n    /**\n      * Determines whether all the members of an array satisfy the specified test.\n      * @param callbackfn A function that accepts up to three arguments. The every method calls the callbackfn function for each element in array1 until the callbackfn returns false, or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n      */\n    every(callbackfn: (value: T, index: number, array: T[]) => boolean, thisArg?: any): boolean;\n    /**\n      * Determines whether the specified callback function returns true for any element of an array.\n      * @param callbackfn A function that accepts up to three arguments. The some method calls the callbackfn function for each element in array1 until the callbackfn returns true, or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n      */\n    some(callbackfn: (value: T, index: number, array: T[]) => boolean, thisArg?: any): boolean;\n    /**\n      * Performs the specified action for each element in an array.\n      * @param callbackfn  A function that accepts up to three arguments. forEach calls the callbackfn function one time for each element in the array.\n      * @param thisArg  An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n      */\n    forEach(callbackfn: (value: T, index: number, array: T[]) => void, thisArg?: any): void;\n    /**\n      * Calls a defined callback function on each element of an array, and returns an array that contains the results.\n      * @param callbackfn A function that accepts up to three arguments. The map method calls the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n      */\n    map<U>(callbackfn: (value: T, index: number, array: T[]) => U, thisArg?: any): U[];\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param callbackfn A function that accepts up to three arguments. The filter method calls the callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n     */\n    filter<S extends T>(callbackfn: (value: T, index: number, array: T[]) => value is S, thisArg?: any): S[];\n    /**\n      * Returns the elements of an array that meet the condition specified in a callback function.\n      * @param callbackfn A function that accepts up to three arguments. The filter method calls the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n      */\n    filter(callbackfn: (value: T, index: number, array: T[]) => any, thisArg?: any): T[];\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n      */\n    reduce(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: T[]) => T): T;\n    reduce(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: T[]) => T, initialValue: T): T;\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n      */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: T, currentIndex: number, array: T[]) => U, initialValue: U): U;\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n      */\n    reduceRight(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: T[]) => T): T;\n    reduceRight(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: T[]) => T, initialValue: T): T;\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n      */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: T, currentIndex: number, array: T[]) => U, initialValue: U): U;\n\n    [n: number]: T;\n}\n\ninterface ArrayConstructor {\n    new(arrayLength?: number): any[];\n    new <T>(arrayLength: number): T[];\n    new <T>(...items: T[]): T[];\n    (arrayLength?: number): any[];\n    <T>(arrayLength: number): T[];\n    <T>(...items: T[]): T[];\n    isArray(arg: any): arg is Array<any>;\n    readonly prototype: Array<any>;\n}\n\ndeclare const Array: ArrayConstructor;\n\ninterface TypedPropertyDescriptor<T> {\n    enumerable?: boolean;\n    configurable?: boolean;\n    writable?: boolean;\n    value?: T;\n    get?: () => T;\n    set?: (value: T) => void;\n}\n\ndeclare type ClassDecorator = <TFunction extends Function>(target: TFunction) => TFunction | void;\ndeclare type PropertyDecorator = (target: Object, propertyKey: string | symbol) => void;\ndeclare type MethodDecorator = <T>(target: Object, propertyKey: string | symbol, descriptor: TypedPropertyDescriptor<T>) => TypedPropertyDescriptor<T> | void;\ndeclare type ParameterDecorator = (target: Object, propertyKey: string | symbol, parameterIndex: number) => void;\n\ndeclare type PromiseConstructorLike = new <T>(executor: (resolve: (value?: T | PromiseLike<T>) => void, reject: (reason?: any) => void) => void) => PromiseLike<T>;\n\ninterface PromiseLike<T> {\n    /**\n     * Attaches callbacks for the resolution and/or rejection of the Promise.\n     * @param onfulfilled The callback to execute when the Promise is resolved.\n     * @param onrejected The callback to execute when the Promise is rejected.\n     * @returns A Promise for the completion of which ever callback is executed.\n     */\n    then<TResult1 = T, TResult2 = never>(onfulfilled?: ((value: T) => TResult1 | PromiseLike<TResult1>) | undefined | null, onrejected?: ((reason: any) => TResult2 | PromiseLike<TResult2>) | undefined | null): PromiseLike<TResult1 | TResult2>;\n}\n\n/**\n * Represents the completion of an asynchronous operation\n */\ninterface Promise<T> {\n    /**\n     * Attaches callbacks for the resolution and/or rejection of the Promise.\n     * @param onfulfilled The callback to execute when the Promise is resolved.\n     * @param onrejected The callback to execute when the Promise is rejected.\n     * @returns A Promise for the completion of which ever callback is executed.\n     */\n    then<TResult1 = T, TResult2 = never>(onfulfilled?: ((value: T) => TResult1 | PromiseLike<TResult1>) | undefined | null, onrejected?: ((reason: any) => TResult2 | PromiseLike<TResult2>) | undefined | null): Promise<TResult1 | TResult2>;\n\n    /**\n     * Attaches a callback for only the rejection of the Promise.\n     * @param onrejected The callback to execute when the Promise is rejected.\n     * @returns A Promise for the completion of the callback.\n     */\n    catch<TResult = never>(onrejected?: ((reason: any) => TResult | PromiseLike<TResult>) | undefined | null): Promise<T | TResult>;\n}\n\ninterface ArrayLike<T> {\n    readonly length: number;\n    readonly [n: number]: T;\n}\n\n/**\n * Make all properties in T optional\n */\ntype Partial<T> = {\n    [P in keyof T]?: T[P];\n};\n\n/**\n * Make all properties in T required\n */\ntype Required<T> = {\n    [P in keyof T]-?: T[P];\n};\n\n/**\n * Make all properties in T readonly\n */\ntype Readonly<T> = {\n    readonly [P in keyof T]: T[P];\n};\n\n/**\n * From T, pick a set of properties whose keys are in the union K\n */\ntype Pick<T, K extends keyof T> = {\n    [P in K]: T[P];\n};\n\n/**\n * Construct a type with a set of properties K of type T\n */\ntype Record<K extends keyof any, T> = {\n    [P in K]: T;\n};\n\n/**\n * Exclude from T those types that are assignable to U\n */\ntype Exclude<T, U> = T extends U ? never : T;\n\n/**\n * Extract from T those types that are assignable to U\n */\ntype Extract<T, U> = T extends U ? T : never;\n\n/**\n * Exclude null and undefined from T\n */\ntype NonNullable<T> = T extends null | undefined ? never : T;\n\n/**\n * Obtain the parameters of a function type in a tuple\n */\ntype Parameters<T extends (...args: any[]) => any> = T extends (...args: infer P) => any ? P : never;\n\n/**\n * Obtain the parameters of a constructor function type in a tuple\n */\ntype ConstructorParameters<T extends new (...args: any[]) => any> = T extends new (...args: infer P) => any ? P : never;\n\n/**\n * Obtain the return type of a function type\n */\ntype ReturnType<T extends (...args: any[]) => any> = T extends (...args: any[]) => infer R ? R : any;\n\n/**\n * Obtain the return type of a constructor function type\n */\ntype InstanceType<T extends new (...args: any[]) => any> = T extends new (...args: any[]) => infer R ? R : any;\n\n/**\n * Marker for contextual 'this' type\n */\ninterface ThisType<T> { }\n\n/**\n  * Represents a raw buffer of binary data, which is used to store data for the\n  * different typed arrays. ArrayBuffers cannot be read from or written to directly,\n  * but can be passed to a typed array or DataView Object to interpret the raw\n  * buffer as needed.\n  */\ninterface ArrayBuffer {\n    /**\n      * Read-only. The length of the ArrayBuffer (in bytes).\n      */\n    readonly byteLength: number;\n\n    /**\n      * Returns a section of an ArrayBuffer.\n      */\n    slice(begin: number, end?: number): ArrayBuffer;\n}\n\n/**\n * Allowed ArrayBuffer types for the buffer of an ArrayBufferView and related Typed Arrays.\n */\ninterface ArrayBufferTypes {\n    ArrayBuffer: ArrayBuffer;\n}\ntype ArrayBufferLike = ArrayBufferTypes[keyof ArrayBufferTypes];\n\ninterface ArrayBufferConstructor {\n    readonly prototype: ArrayBuffer;\n    new(byteLength: number): ArrayBuffer;\n    isView(arg: any): arg is ArrayBufferView;\n}\ndeclare const ArrayBuffer: ArrayBufferConstructor;\n\ninterface ArrayBufferView {\n    /**\n      * The ArrayBuffer instance referenced by the array.\n      */\n    buffer: ArrayBufferLike;\n\n    /**\n      * The length in bytes of the array.\n      */\n    byteLength: number;\n\n    /**\n      * The offset in bytes of the array.\n      */\n    byteOffset: number;\n}\n\ninterface DataView {\n    readonly buffer: ArrayBuffer;\n    readonly byteLength: number;\n    readonly byteOffset: number;\n    /**\n      * Gets the Float32 value at the specified byte offset from the start of the view. There is\n      * no alignment constraint; multi-byte values may be fetched from any offset.\n      * @param byteOffset The place in the buffer at which the value should be retrieved.\n      */\n    getFloat32(byteOffset: number, littleEndian?: boolean): number;\n\n    /**\n      * Gets the Float64 value at the specified byte offset from the start of the view. There is\n      * no alignment constraint; multi-byte values may be fetched from any offset.\n      * @param byteOffset The place in the buffer at which the value should be retrieved.\n      */\n    getFloat64(byteOffset: number, littleEndian?: boolean): number;\n\n    /**\n      * Gets the Int8 value at the specified byte offset from the start of the view. There is\n      * no alignment constraint; multi-byte values may be fetched from any offset.\n      * @param byteOffset The place in the buffer at which the value should be retrieved.\n      */\n    getInt8(byteOffset: number): number;\n\n    /**\n      * Gets the Int16 value at the specified byte offset from the start of the view. There is\n      * no alignment constraint; multi-byte values may be fetched from any offset.\n      * @param byteOffset The place in the buffer at which the value should be retrieved.\n      */\n    getInt16(byteOffset: number, littleEndian?: boolean): number;\n    /**\n      * Gets the Int32 value at the specified byte offset from the start of the view. There is\n      * no alignment constraint; multi-byte values may be fetched from any offset.\n      * @param byteOffset The place in the buffer at which the value should be retrieved.\n      */\n    getInt32(byteOffset: number, littleEndian?: boolean): number;\n\n    /**\n      * Gets the Uint8 value at the specified byte offset from the start of the view. There is\n      * no alignment constraint; multi-byte values may be fetched from any offset.\n      * @param byteOffset The place in the buffer at which the value should be retrieved.\n      */\n    getUint8(byteOffset: number): number;\n\n    /**\n      * Gets the Uint16 value at the specified byte offset from the start of the view. There is\n      * no alignment constraint; multi-byte values may be fetched from any offset.\n      * @param byteOffset The place in the buffer at which the value should be retrieved.\n      */\n    getUint16(byteOffset: number, littleEndian?: boolean): number;\n\n    /**\n      * Gets the Uint32 value at the specified byte offset from the start of the view. There is\n      * no alignment constraint; multi-byte values may be fetched from any offset.\n      * @param byteOffset The place in the buffer at which the value should be retrieved.\n      */\n    getUint32(byteOffset: number, littleEndian?: boolean): number;\n\n    /**\n      * Stores an Float32 value at the specified byte offset from the start of the view.\n      * @param byteOffset The place in the buffer at which the value should be set.\n      * @param value The value to set.\n      * @param littleEndian If false or undefined, a big-endian value should be written,\n      * otherwise a little-endian value should be written.\n      */\n    setFloat32(byteOffset: number, value: number, littleEndian?: boolean): void;\n\n    /**\n      * Stores an Float64 value at the specified byte offset from the start of the view.\n      * @param byteOffset The place in the buffer at which the value should be set.\n      * @param value The value to set.\n      * @param littleEndian If false or undefined, a big-endian value should be written,\n      * otherwise a little-endian value should be written.\n      */\n    setFloat64(byteOffset: number, value: number, littleEndian?: boolean): void;\n\n    /**\n      * Stores an Int8 value at the specified byte offset from the start of the view.\n      * @param byteOffset The place in the buffer at which the value should be set.\n      * @param value The value to set.\n      */\n    setInt8(byteOffset: number, value: number): void;\n\n    /**\n      * Stores an Int16 value at the specified byte offset from the start of the view.\n      * @param byteOffset The place in the buffer at which the value should be set.\n      * @param value The value to set.\n      * @param littleEndian If false or undefined, a big-endian value should be written,\n      * otherwise a little-endian value should be written.\n      */\n    setInt16(byteOffset: number, value: number, littleEndian?: boolean): void;\n\n    /**\n      * Stores an Int32 value at the specified byte offset from the start of the view.\n      * @param byteOffset The place in the buffer at which the value should be set.\n      * @param value The value to set.\n      * @param littleEndian If false or undefined, a big-endian value should be written,\n      * otherwise a little-endian value should be written.\n      */\n    setInt32(byteOffset: number, value: number, littleEndian?: boolean): void;\n\n    /**\n      * Stores an Uint8 value at the specified byte offset from the start of the view.\n      * @param byteOffset The place in the buffer at which the value should be set.\n      * @param value The value to set.\n      */\n    setUint8(byteOffset: number, value: number): void;\n\n    /**\n      * Stores an Uint16 value at the specified byte offset from the start of the view.\n      * @param byteOffset The place in the buffer at which the value should be set.\n      * @param value The value to set.\n      * @param littleEndian If false or undefined, a big-endian value should be written,\n      * otherwise a little-endian value should be written.\n      */\n    setUint16(byteOffset: number, value: number, littleEndian?: boolean): void;\n\n    /**\n      * Stores an Uint32 value at the specified byte offset from the start of the view.\n      * @param byteOffset The place in the buffer at which the value should be set.\n      * @param value The value to set.\n      * @param littleEndian If false or undefined, a big-endian value should be written,\n      * otherwise a little-endian value should be written.\n      */\n    setUint32(byteOffset: number, value: number, littleEndian?: boolean): void;\n}\n\ninterface DataViewConstructor {\n    new(buffer: ArrayBufferLike, byteOffset?: number, byteLength?: number): DataView;\n}\ndeclare const DataView: DataViewConstructor;\n\n/**\n  * A typed array of 8-bit integer values. The contents are initialized to 0. If the requested\n  * number of bytes could not be allocated an exception is raised.\n  */\ninterface Int8Array {\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * The ArrayBuffer instance referenced by the array.\n      */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n      * The length in bytes of the array.\n      */\n    readonly byteLength: number;\n\n    /**\n      * The offset in bytes of the array.\n      */\n    readonly byteOffset: number;\n\n    /**\n      * Returns the this object after copying a section of the array identified by start and end\n      * to the same array starting at position target\n      * @param target If target is negative, it is treated as length+target where length is the\n      * length of the array.\n      * @param start If start is negative, it is treated as length+start. If end is negative, it\n      * is treated as length+end.\n      * @param end If not specified, length of the this object is used as its default value.\n      */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n      * Determines whether all the members of an array satisfy the specified test.\n      * @param callbackfn A function that accepts up to three arguments. The every method calls\n      * the callbackfn function for each element in array1 until the callbackfn returns false,\n      * or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    every(callbackfn: (value: number, index: number, array: Int8Array) => boolean, thisArg?: any): boolean;\n\n    /**\n        * Returns the this object after filling the section identified by start and end with value\n        * @param value value to fill array section with\n        * @param start index to start filling the array at. If start is negative, it is treated as\n        * length+start where length is the length of the array.\n        * @param end index to stop filling the array at. If end is negative, it is treated as\n        * length+end.\n        */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n      * Returns the elements of an array that meet the condition specified in a callback function.\n      * @param callbackfn A function that accepts up to three arguments. The filter method calls\n      * the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    filter(callbackfn: (value: number, index: number, array: Int8Array) => any, thisArg?: any): Int8Array;\n\n    /**\n      * Returns the value of the first element in the array where predicate is true, and undefined\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found, find\n      * immediately returns that element value. Otherwise, find returns undefined.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    find(predicate: (value: number, index: number, obj: Int8Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n      * Returns the index of the first element in the array where predicate is true, and -1\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found,\n      * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    findIndex(predicate: (value: number, index: number, obj: Int8Array) => boolean, thisArg?: any): number;\n\n    /**\n      * Performs the specified action for each element in an array.\n      * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    forEach(callbackfn: (value: number, index: number, array: Int8Array) => void, thisArg?: any): void;\n\n    /**\n      * Returns the index of the first occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      *  search starts at index 0.\n      */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * Adds all the elements of an array separated by the specified separator string.\n      * @param separator A string used to separate one element of an array from the next in the\n      * resulting String. If omitted, the array elements are separated with a comma.\n      */\n    join(separator?: string): string;\n\n    /**\n      * Returns the index of the last occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      * search starts at index 0.\n      */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * The length of the array.\n      */\n    readonly length: number;\n\n    /**\n      * Calls a defined callback function on each element of an array, and returns an array that\n      * contains the results.\n      * @param callbackfn A function that accepts up to three arguments. The map method calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    map(callbackfn: (value: number, index: number, array: Int8Array) => number, thisArg?: any): Int8Array;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int8Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int8Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int8Array) => U, initialValue: U): U;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an\n      * argument instead of an array value.\n      */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int8Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int8Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int8Array) => U, initialValue: U): U;\n\n    /**\n      * Reverses the elements in an Array.\n      */\n    reverse(): Int8Array;\n\n    /**\n      * Sets a value or an array of values.\n      * @param array A typed or untyped array of values to set.\n      * @param offset The index in the current array at which the values are to be written.\n      */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n      * Returns a section of an array.\n      * @param start The beginning of the specified portion of the array.\n      * @param end The end of the specified portion of the array.\n      */\n    slice(start?: number, end?: number): Int8Array;\n\n    /**\n      * Determines whether the specified callback function returns true for any element of an array.\n      * @param callbackfn A function that accepts up to three arguments. The some method calls the\n      * callbackfn function for each element in array1 until the callbackfn returns true, or until\n      * the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    some(callbackfn: (value: number, index: number, array: Int8Array) => boolean, thisArg?: any): boolean;\n\n    /**\n      * Sorts an array.\n      * @param compareFn The name of the function used to determine the order of the elements. If\n      * omitted, the elements are sorted in ascending, ASCII character order.\n      */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n      * Gets a new Int8Array view of the ArrayBuffer store for this array, referencing the elements\n      * at begin, inclusive, up to end, exclusive.\n      * @param begin The index of the beginning of the array.\n      * @param end The index of the end of the array.\n      */\n    subarray(begin: number, end?: number): Int8Array;\n\n    /**\n      * Converts a number to a string by using the current locale.\n      */\n    toLocaleString(): string;\n\n    /**\n      * Returns a string representation of an array.\n      */\n    toString(): string;\n\n    [index: number]: number;\n}\ninterface Int8ArrayConstructor {\n    readonly prototype: Int8Array;\n    new(length: number): Int8Array;\n    new(arrayOrArrayBuffer: ArrayLike<number> | ArrayBufferLike): Int8Array;\n    new(buffer: ArrayBufferLike, byteOffset: number, length?: number): Int8Array;\n\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * Returns a new array from a set of elements.\n      * @param items A set of elements to include in the new array object.\n      */\n    of(...items: number[]): Int8Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      */\n    from(arrayLike: ArrayLike<number>): Int8Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      * @param mapfn A mapping function to call on every element of the array.\n      * @param thisArg Value of 'this' used to invoke the mapfn.\n      */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Int8Array;\n\n\n}\ndeclare const Int8Array: Int8ArrayConstructor;\n\n/**\n  * A typed array of 8-bit unsigned integer values. The contents are initialized to 0. If the\n  * requested number of bytes could not be allocated an exception is raised.\n  */\ninterface Uint8Array {\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * The ArrayBuffer instance referenced by the array.\n      */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n      * The length in bytes of the array.\n      */\n    readonly byteLength: number;\n\n    /**\n      * The offset in bytes of the array.\n      */\n    readonly byteOffset: number;\n\n    /**\n      * Returns the this object after copying a section of the array identified by start and end\n      * to the same array starting at position target\n      * @param target If target is negative, it is treated as length+target where length is the\n      * length of the array.\n      * @param start If start is negative, it is treated as length+start. If end is negative, it\n      * is treated as length+end.\n      * @param end If not specified, length of the this object is used as its default value.\n      */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n      * Determines whether all the members of an array satisfy the specified test.\n      * @param callbackfn A function that accepts up to three arguments. The every method calls\n      * the callbackfn function for each element in array1 until the callbackfn returns false,\n      * or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    every(callbackfn: (value: number, index: number, array: Uint8Array) => boolean, thisArg?: any): boolean;\n\n    /**\n        * Returns the this object after filling the section identified by start and end with value\n        * @param value value to fill array section with\n        * @param start index to start filling the array at. If start is negative, it is treated as\n        * length+start where length is the length of the array.\n        * @param end index to stop filling the array at. If end is negative, it is treated as\n        * length+end.\n        */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n      * Returns the elements of an array that meet the condition specified in a callback function.\n      * @param callbackfn A function that accepts up to three arguments. The filter method calls\n      * the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    filter(callbackfn: (value: number, index: number, array: Uint8Array) => any, thisArg?: any): Uint8Array;\n\n    /**\n      * Returns the value of the first element in the array where predicate is true, and undefined\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found, find\n      * immediately returns that element value. Otherwise, find returns undefined.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    find(predicate: (value: number, index: number, obj: Uint8Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n      * Returns the index of the first element in the array where predicate is true, and -1\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found,\n      * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    findIndex(predicate: (value: number, index: number, obj: Uint8Array) => boolean, thisArg?: any): number;\n\n    /**\n      * Performs the specified action for each element in an array.\n      * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    forEach(callbackfn: (value: number, index: number, array: Uint8Array) => void, thisArg?: any): void;\n\n    /**\n      * Returns the index of the first occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      *  search starts at index 0.\n      */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * Adds all the elements of an array separated by the specified separator string.\n      * @param separator A string used to separate one element of an array from the next in the\n      * resulting String. If omitted, the array elements are separated with a comma.\n      */\n    join(separator?: string): string;\n\n    /**\n      * Returns the index of the last occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      * search starts at index 0.\n      */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * The length of the array.\n      */\n    readonly length: number;\n\n    /**\n      * Calls a defined callback function on each element of an array, and returns an array that\n      * contains the results.\n      * @param callbackfn A function that accepts up to three arguments. The map method calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    map(callbackfn: (value: number, index: number, array: Uint8Array) => number, thisArg?: any): Uint8Array;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint8Array) => U, initialValue: U): U;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an\n      * argument instead of an array value.\n      */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint8Array) => U, initialValue: U): U;\n\n    /**\n      * Reverses the elements in an Array.\n      */\n    reverse(): Uint8Array;\n\n    /**\n      * Sets a value or an array of values.\n      * @param array A typed or untyped array of values to set.\n      * @param offset The index in the current array at which the values are to be written.\n      */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n      * Returns a section of an array.\n      * @param start The beginning of the specified portion of the array.\n      * @param end The end of the specified portion of the array.\n      */\n    slice(start?: number, end?: number): Uint8Array;\n\n    /**\n      * Determines whether the specified callback function returns true for any element of an array.\n      * @param callbackfn A function that accepts up to three arguments. The some method calls the\n      * callbackfn function for each element in array1 until the callbackfn returns true, or until\n      * the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    some(callbackfn: (value: number, index: number, array: Uint8Array) => boolean, thisArg?: any): boolean;\n\n    /**\n      * Sorts an array.\n      * @param compareFn The name of the function used to determine the order of the elements. If\n      * omitted, the elements are sorted in ascending, ASCII character order.\n      */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n      * Gets a new Uint8Array view of the ArrayBuffer store for this array, referencing the elements\n      * at begin, inclusive, up to end, exclusive.\n      * @param begin The index of the beginning of the array.\n      * @param end The index of the end of the array.\n      */\n    subarray(begin: number, end?: number): Uint8Array;\n\n    /**\n      * Converts a number to a string by using the current locale.\n      */\n    toLocaleString(): string;\n\n    /**\n      * Returns a string representation of an array.\n      */\n    toString(): string;\n\n    [index: number]: number;\n}\n\ninterface Uint8ArrayConstructor {\n    readonly prototype: Uint8Array;\n    new(length: number): Uint8Array;\n    new(arrayOrArrayBuffer: ArrayLike<number> | ArrayBufferLike): Uint8Array;\n    new(buffer: ArrayBufferLike, byteOffset: number, length?: number): Uint8Array;\n\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * Returns a new array from a set of elements.\n      * @param items A set of elements to include in the new array object.\n      */\n    of(...items: number[]): Uint8Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      */\n    from(arrayLike: ArrayLike<number>): Uint8Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      * @param mapfn A mapping function to call on every element of the array.\n      * @param thisArg Value of 'this' used to invoke the mapfn.\n      */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Uint8Array;\n\n}\ndeclare const Uint8Array: Uint8ArrayConstructor;\n\n/**\n  * A typed array of 8-bit unsigned integer (clamped) values. The contents are initialized to 0.\n  * If the requested number of bytes could not be allocated an exception is raised.\n  */\ninterface Uint8ClampedArray {\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * The ArrayBuffer instance referenced by the array.\n      */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n      * The length in bytes of the array.\n      */\n    readonly byteLength: number;\n\n    /**\n      * The offset in bytes of the array.\n      */\n    readonly byteOffset: number;\n\n    /**\n      * Returns the this object after copying a section of the array identified by start and end\n      * to the same array starting at position target\n      * @param target If target is negative, it is treated as length+target where length is the\n      * length of the array.\n      * @param start If start is negative, it is treated as length+start. If end is negative, it\n      * is treated as length+end.\n      * @param end If not specified, length of the this object is used as its default value.\n      */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n      * Determines whether all the members of an array satisfy the specified test.\n      * @param callbackfn A function that accepts up to three arguments. The every method calls\n      * the callbackfn function for each element in array1 until the callbackfn returns false,\n      * or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    every(callbackfn: (value: number, index: number, array: Uint8ClampedArray) => boolean, thisArg?: any): boolean;\n\n    /**\n        * Returns the this object after filling the section identified by start and end with value\n        * @param value value to fill array section with\n        * @param start index to start filling the array at. If start is negative, it is treated as\n        * length+start where length is the length of the array.\n        * @param end index to stop filling the array at. If end is negative, it is treated as\n        * length+end.\n        */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n      * Returns the elements of an array that meet the condition specified in a callback function.\n      * @param callbackfn A function that accepts up to three arguments. The filter method calls\n      * the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    filter(callbackfn: (value: number, index: number, array: Uint8ClampedArray) => any, thisArg?: any): Uint8ClampedArray;\n\n    /**\n      * Returns the value of the first element in the array where predicate is true, and undefined\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found, find\n      * immediately returns that element value. Otherwise, find returns undefined.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    find(predicate: (value: number, index: number, obj: Uint8ClampedArray) => boolean, thisArg?: any): number | undefined;\n\n    /**\n      * Returns the index of the first element in the array where predicate is true, and -1\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found,\n      * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    findIndex(predicate: (value: number, index: number, obj: Uint8ClampedArray) => boolean, thisArg?: any): number;\n\n    /**\n      * Performs the specified action for each element in an array.\n      * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    forEach(callbackfn: (value: number, index: number, array: Uint8ClampedArray) => void, thisArg?: any): void;\n\n    /**\n      * Returns the index of the first occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      *  search starts at index 0.\n      */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * Adds all the elements of an array separated by the specified separator string.\n      * @param separator A string used to separate one element of an array from the next in the\n      * resulting String. If omitted, the array elements are separated with a comma.\n      */\n    join(separator?: string): string;\n\n    /**\n      * Returns the index of the last occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      * search starts at index 0.\n      */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * The length of the array.\n      */\n    readonly length: number;\n\n    /**\n      * Calls a defined callback function on each element of an array, and returns an array that\n      * contains the results.\n      * @param callbackfn A function that accepts up to three arguments. The map method calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    map(callbackfn: (value: number, index: number, array: Uint8ClampedArray) => number, thisArg?: any): Uint8ClampedArray;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => U, initialValue: U): U;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an\n      * argument instead of an array value.\n      */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => U, initialValue: U): U;\n\n    /**\n      * Reverses the elements in an Array.\n      */\n    reverse(): Uint8ClampedArray;\n\n    /**\n      * Sets a value or an array of values.\n      * @param array A typed or untyped array of values to set.\n      * @param offset The index in the current array at which the values are to be written.\n      */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n      * Returns a section of an array.\n      * @param start The beginning of the specified portion of the array.\n      * @param end The end of the specified portion of the array.\n      */\n    slice(start?: number, end?: number): Uint8ClampedArray;\n\n    /**\n      * Determines whether the specified callback function returns true for any element of an array.\n      * @param callbackfn A function that accepts up to three arguments. The some method calls the\n      * callbackfn function for each element in array1 until the callbackfn returns true, or until\n      * the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    some(callbackfn: (value: number, index: number, array: Uint8ClampedArray) => boolean, thisArg?: any): boolean;\n\n    /**\n      * Sorts an array.\n      * @param compareFn The name of the function used to determine the order of the elements. If\n      * omitted, the elements are sorted in ascending, ASCII character order.\n      */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n      * Gets a new Uint8ClampedArray view of the ArrayBuffer store for this array, referencing the elements\n      * at begin, inclusive, up to end, exclusive.\n      * @param begin The index of the beginning of the array.\n      * @param end The index of the end of the array.\n      */\n    subarray(begin: number, end?: number): Uint8ClampedArray;\n\n    /**\n      * Converts a number to a string by using the current locale.\n      */\n    toLocaleString(): string;\n\n    /**\n      * Returns a string representation of an array.\n      */\n    toString(): string;\n\n    [index: number]: number;\n}\n\ninterface Uint8ClampedArrayConstructor {\n    readonly prototype: Uint8ClampedArray;\n    new(length: number): Uint8ClampedArray;\n    new(arrayOrArrayBuffer: ArrayLike<number> | ArrayBufferLike): Uint8ClampedArray;\n    new(buffer: ArrayBufferLike, byteOffset: number, length?: number): Uint8ClampedArray;\n\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * Returns a new array from a set of elements.\n      * @param items A set of elements to include in the new array object.\n      */\n    of(...items: number[]): Uint8ClampedArray;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      */\n    from(arrayLike: ArrayLike<number>): Uint8ClampedArray;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      * @param mapfn A mapping function to call on every element of the array.\n      * @param thisArg Value of 'this' used to invoke the mapfn.\n      */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Uint8ClampedArray;\n}\ndeclare const Uint8ClampedArray: Uint8ClampedArrayConstructor;\n\n/**\n  * A typed array of 16-bit signed integer values. The contents are initialized to 0. If the\n  * requested number of bytes could not be allocated an exception is raised.\n  */\ninterface Int16Array {\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * The ArrayBuffer instance referenced by the array.\n      */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n      * The length in bytes of the array.\n      */\n    readonly byteLength: number;\n\n    /**\n      * The offset in bytes of the array.\n      */\n    readonly byteOffset: number;\n\n    /**\n      * Returns the this object after copying a section of the array identified by start and end\n      * to the same array starting at position target\n      * @param target If target is negative, it is treated as length+target where length is the\n      * length of the array.\n      * @param start If start is negative, it is treated as length+start. If end is negative, it\n      * is treated as length+end.\n      * @param end If not specified, length of the this object is used as its default value.\n      */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n      * Determines whether all the members of an array satisfy the specified test.\n      * @param callbackfn A function that accepts up to three arguments. The every method calls\n      * the callbackfn function for each element in array1 until the callbackfn returns false,\n      * or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    every(callbackfn: (value: number, index: number, array: Int16Array) => boolean, thisArg?: any): boolean;\n\n    /**\n        * Returns the this object after filling the section identified by start and end with value\n        * @param value value to fill array section with\n        * @param start index to start filling the array at. If start is negative, it is treated as\n        * length+start where length is the length of the array.\n        * @param end index to stop filling the array at. If end is negative, it is treated as\n        * length+end.\n        */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n      * Returns the elements of an array that meet the condition specified in a callback function.\n      * @param callbackfn A function that accepts up to three arguments. The filter method calls\n      * the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    filter(callbackfn: (value: number, index: number, array: Int16Array) => any, thisArg?: any): Int16Array;\n\n    /**\n      * Returns the value of the first element in the array where predicate is true, and undefined\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found, find\n      * immediately returns that element value. Otherwise, find returns undefined.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    find(predicate: (value: number, index: number, obj: Int16Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n      * Returns the index of the first element in the array where predicate is true, and -1\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found,\n      * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    findIndex(predicate: (value: number, index: number, obj: Int16Array) => boolean, thisArg?: any): number;\n\n    /**\n      * Performs the specified action for each element in an array.\n      * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    forEach(callbackfn: (value: number, index: number, array: Int16Array) => void, thisArg?: any): void;\n    /**\n      * Returns the index of the first occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      *  search starts at index 0.\n      */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * Adds all the elements of an array separated by the specified separator string.\n      * @param separator A string used to separate one element of an array from the next in the\n      * resulting String. If omitted, the array elements are separated with a comma.\n      */\n    join(separator?: string): string;\n\n    /**\n      * Returns the index of the last occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      * search starts at index 0.\n      */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * The length of the array.\n      */\n    readonly length: number;\n\n    /**\n      * Calls a defined callback function on each element of an array, and returns an array that\n      * contains the results.\n      * @param callbackfn A function that accepts up to three arguments. The map method calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    map(callbackfn: (value: number, index: number, array: Int16Array) => number, thisArg?: any): Int16Array;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int16Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int16Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int16Array) => U, initialValue: U): U;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an\n      * argument instead of an array value.\n      */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int16Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int16Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int16Array) => U, initialValue: U): U;\n\n    /**\n      * Reverses the elements in an Array.\n      */\n    reverse(): Int16Array;\n\n    /**\n      * Sets a value or an array of values.\n      * @param array A typed or untyped array of values to set.\n      * @param offset The index in the current array at which the values are to be written.\n      */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n      * Returns a section of an array.\n      * @param start The beginning of the specified portion of the array.\n      * @param end The end of the specified portion of the array.\n      */\n    slice(start?: number, end?: number): Int16Array;\n\n    /**\n      * Determines whether the specified callback function returns true for any element of an array.\n      * @param callbackfn A function that accepts up to three arguments. The some method calls the\n      * callbackfn function for each element in array1 until the callbackfn returns true, or until\n      * the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    some(callbackfn: (value: number, index: number, array: Int16Array) => boolean, thisArg?: any): boolean;\n\n    /**\n      * Sorts an array.\n      * @param compareFn The name of the function used to determine the order of the elements. If\n      * omitted, the elements are sorted in ascending, ASCII character order.\n      */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n      * Gets a new Int16Array view of the ArrayBuffer store for this array, referencing the elements\n      * at begin, inclusive, up to end, exclusive.\n      * @param begin The index of the beginning of the array.\n      * @param end The index of the end of the array.\n      */\n    subarray(begin: number, end?: number): Int16Array;\n\n    /**\n      * Converts a number to a string by using the current locale.\n      */\n    toLocaleString(): string;\n\n    /**\n      * Returns a string representation of an array.\n      */\n    toString(): string;\n\n    [index: number]: number;\n}\n\ninterface Int16ArrayConstructor {\n    readonly prototype: Int16Array;\n    new(length: number): Int16Array;\n    new(arrayOrArrayBuffer: ArrayLike<number> | ArrayBufferLike): Int16Array;\n    new(buffer: ArrayBufferLike, byteOffset: number, length?: number): Int16Array;\n\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * Returns a new array from a set of elements.\n      * @param items A set of elements to include in the new array object.\n      */\n    of(...items: number[]): Int16Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      */\n    from(arrayLike: ArrayLike<number>): Int16Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      * @param mapfn A mapping function to call on every element of the array.\n      * @param thisArg Value of 'this' used to invoke the mapfn.\n      */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Int16Array;\n\n\n}\ndeclare const Int16Array: Int16ArrayConstructor;\n\n/**\n  * A typed array of 16-bit unsigned integer values. The contents are initialized to 0. If the\n  * requested number of bytes could not be allocated an exception is raised.\n  */\ninterface Uint16Array {\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * The ArrayBuffer instance referenced by the array.\n      */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n      * The length in bytes of the array.\n      */\n    readonly byteLength: number;\n\n    /**\n      * The offset in bytes of the array.\n      */\n    readonly byteOffset: number;\n\n    /**\n      * Returns the this object after copying a section of the array identified by start and end\n      * to the same array starting at position target\n      * @param target If target is negative, it is treated as length+target where length is the\n      * length of the array.\n      * @param start If start is negative, it is treated as length+start. If end is negative, it\n      * is treated as length+end.\n      * @param end If not specified, length of the this object is used as its default value.\n      */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n      * Determines whether all the members of an array satisfy the specified test.\n      * @param callbackfn A function that accepts up to three arguments. The every method calls\n      * the callbackfn function for each element in array1 until the callbackfn returns false,\n      * or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    every(callbackfn: (value: number, index: number, array: Uint16Array) => boolean, thisArg?: any): boolean;\n\n    /**\n        * Returns the this object after filling the section identified by start and end with value\n        * @param value value to fill array section with\n        * @param start index to start filling the array at. If start is negative, it is treated as\n        * length+start where length is the length of the array.\n        * @param end index to stop filling the array at. If end is negative, it is treated as\n        * length+end.\n        */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n      * Returns the elements of an array that meet the condition specified in a callback function.\n      * @param callbackfn A function that accepts up to three arguments. The filter method calls\n      * the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    filter(callbackfn: (value: number, index: number, array: Uint16Array) => any, thisArg?: any): Uint16Array;\n\n    /**\n      * Returns the value of the first element in the array where predicate is true, and undefined\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found, find\n      * immediately returns that element value. Otherwise, find returns undefined.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    find(predicate: (value: number, index: number, obj: Uint16Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n      * Returns the index of the first element in the array where predicate is true, and -1\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found,\n      * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    findIndex(predicate: (value: number, index: number, obj: Uint16Array) => boolean, thisArg?: any): number;\n\n    /**\n      * Performs the specified action for each element in an array.\n      * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    forEach(callbackfn: (value: number, index: number, array: Uint16Array) => void, thisArg?: any): void;\n\n    /**\n      * Returns the index of the first occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      *  search starts at index 0.\n      */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * Adds all the elements of an array separated by the specified separator string.\n      * @param separator A string used to separate one element of an array from the next in the\n      * resulting String. If omitted, the array elements are separated with a comma.\n      */\n    join(separator?: string): string;\n\n    /**\n      * Returns the index of the last occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      * search starts at index 0.\n      */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * The length of the array.\n      */\n    readonly length: number;\n\n    /**\n      * Calls a defined callback function on each element of an array, and returns an array that\n      * contains the results.\n      * @param callbackfn A function that accepts up to three arguments. The map method calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    map(callbackfn: (value: number, index: number, array: Uint16Array) => number, thisArg?: any): Uint16Array;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint16Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint16Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint16Array) => U, initialValue: U): U;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an\n      * argument instead of an array value.\n      */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint16Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint16Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint16Array) => U, initialValue: U): U;\n\n    /**\n      * Reverses the elements in an Array.\n      */\n    reverse(): Uint16Array;\n\n    /**\n      * Sets a value or an array of values.\n      * @param array A typed or untyped array of values to set.\n      * @param offset The index in the current array at which the values are to be written.\n      */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n      * Returns a section of an array.\n      * @param start The beginning of the specified portion of the array.\n      * @param end The end of the specified portion of the array.\n      */\n    slice(start?: number, end?: number): Uint16Array;\n\n    /**\n      * Determines whether the specified callback function returns true for any element of an array.\n      * @param callbackfn A function that accepts up to three arguments. The some method calls the\n      * callbackfn function for each element in array1 until the callbackfn returns true, or until\n      * the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    some(callbackfn: (value: number, index: number, array: Uint16Array) => boolean, thisArg?: any): boolean;\n\n    /**\n      * Sorts an array.\n      * @param compareFn The name of the function used to determine the order of the elements. If\n      * omitted, the elements are sorted in ascending, ASCII character order.\n      */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n      * Gets a new Uint16Array view of the ArrayBuffer store for this array, referencing the elements\n      * at begin, inclusive, up to end, exclusive.\n      * @param begin The index of the beginning of the array.\n      * @param end The index of the end of the array.\n      */\n    subarray(begin: number, end?: number): Uint16Array;\n\n    /**\n      * Converts a number to a string by using the current locale.\n      */\n    toLocaleString(): string;\n\n    /**\n      * Returns a string representation of an array.\n      */\n    toString(): string;\n\n    [index: number]: number;\n}\n\ninterface Uint16ArrayConstructor {\n    readonly prototype: Uint16Array;\n    new(length: number): Uint16Array;\n    new(arrayOrArrayBuffer: ArrayLike<number> | ArrayBufferLike): Uint16Array;\n    new(buffer: ArrayBufferLike, byteOffset: number, length?: number): Uint16Array;\n\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * Returns a new array from a set of elements.\n      * @param items A set of elements to include in the new array object.\n      */\n    of(...items: number[]): Uint16Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      */\n    from(arrayLike: ArrayLike<number>): Uint16Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      * @param mapfn A mapping function to call on every element of the array.\n      * @param thisArg Value of 'this' used to invoke the mapfn.\n      */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Uint16Array;\n\n\n}\ndeclare const Uint16Array: Uint16ArrayConstructor;\n/**\n  * A typed array of 32-bit signed integer values. The contents are initialized to 0. If the\n  * requested number of bytes could not be allocated an exception is raised.\n  */\ninterface Int32Array {\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * The ArrayBuffer instance referenced by the array.\n      */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n      * The length in bytes of the array.\n      */\n    readonly byteLength: number;\n\n    /**\n      * The offset in bytes of the array.\n      */\n    readonly byteOffset: number;\n\n    /**\n      * Returns the this object after copying a section of the array identified by start and end\n      * to the same array starting at position target\n      * @param target If target is negative, it is treated as length+target where length is the\n      * length of the array.\n      * @param start If start is negative, it is treated as length+start. If end is negative, it\n      * is treated as length+end.\n      * @param end If not specified, length of the this object is used as its default value.\n      */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n      * Determines whether all the members of an array satisfy the specified test.\n      * @param callbackfn A function that accepts up to three arguments. The every method calls\n      * the callbackfn function for each element in array1 until the callbackfn returns false,\n      * or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    every(callbackfn: (value: number, index: number, array: Int32Array) => boolean, thisArg?: any): boolean;\n\n    /**\n        * Returns the this object after filling the section identified by start and end with value\n        * @param value value to fill array section with\n        * @param start index to start filling the array at. If start is negative, it is treated as\n        * length+start where length is the length of the array.\n        * @param end index to stop filling the array at. If end is negative, it is treated as\n        * length+end.\n        */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n      * Returns the elements of an array that meet the condition specified in a callback function.\n      * @param callbackfn A function that accepts up to three arguments. The filter method calls\n      * the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    filter(callbackfn: (value: number, index: number, array: Int32Array) => any, thisArg?: any): Int32Array;\n\n    /**\n      * Returns the value of the first element in the array where predicate is true, and undefined\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found, find\n      * immediately returns that element value. Otherwise, find returns undefined.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    find(predicate: (value: number, index: number, obj: Int32Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n      * Returns the index of the first element in the array where predicate is true, and -1\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found,\n      * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    findIndex(predicate: (value: number, index: number, obj: Int32Array) => boolean, thisArg?: any): number;\n\n    /**\n      * Performs the specified action for each element in an array.\n      * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    forEach(callbackfn: (value: number, index: number, array: Int32Array) => void, thisArg?: any): void;\n\n    /**\n      * Returns the index of the first occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      *  search starts at index 0.\n      */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * Adds all the elements of an array separated by the specified separator string.\n      * @param separator A string used to separate one element of an array from the next in the\n      * resulting String. If omitted, the array elements are separated with a comma.\n      */\n    join(separator?: string): string;\n\n    /**\n      * Returns the index of the last occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      * search starts at index 0.\n      */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * The length of the array.\n      */\n    readonly length: number;\n\n    /**\n      * Calls a defined callback function on each element of an array, and returns an array that\n      * contains the results.\n      * @param callbackfn A function that accepts up to three arguments. The map method calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    map(callbackfn: (value: number, index: number, array: Int32Array) => number, thisArg?: any): Int32Array;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int32Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int32Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int32Array) => U, initialValue: U): U;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an\n      * argument instead of an array value.\n      */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int32Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int32Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int32Array) => U, initialValue: U): U;\n\n    /**\n      * Reverses the elements in an Array.\n      */\n    reverse(): Int32Array;\n\n    /**\n      * Sets a value or an array of values.\n      * @param array A typed or untyped array of values to set.\n      * @param offset The index in the current array at which the values are to be written.\n      */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n      * Returns a section of an array.\n      * @param start The beginning of the specified portion of the array.\n      * @param end The end of the specified portion of the array.\n      */\n    slice(start?: number, end?: number): Int32Array;\n\n    /**\n      * Determines whether the specified callback function returns true for any element of an array.\n      * @param callbackfn A function that accepts up to three arguments. The some method calls the\n      * callbackfn function for each element in array1 until the callbackfn returns true, or until\n      * the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    some(callbackfn: (value: number, index: number, array: Int32Array) => boolean, thisArg?: any): boolean;\n\n    /**\n      * Sorts an array.\n      * @param compareFn The name of the function used to determine the order of the elements. If\n      * omitted, the elements are sorted in ascending, ASCII character order.\n      */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n      * Gets a new Int32Array view of the ArrayBuffer store for this array, referencing the elements\n      * at begin, inclusive, up to end, exclusive.\n      * @param begin The index of the beginning of the array.\n      * @param end The index of the end of the array.\n      */\n    subarray(begin: number, end?: number): Int32Array;\n\n    /**\n      * Converts a number to a string by using the current locale.\n      */\n    toLocaleString(): string;\n\n    /**\n      * Returns a string representation of an array.\n      */\n    toString(): string;\n\n    [index: number]: number;\n}\n\ninterface Int32ArrayConstructor {\n    readonly prototype: Int32Array;\n    new(length: number): Int32Array;\n    new(arrayOrArrayBuffer: ArrayLike<number> | ArrayBufferLike): Int32Array;\n    new(buffer: ArrayBufferLike, byteOffset: number, length?: number): Int32Array;\n\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * Returns a new array from a set of elements.\n      * @param items A set of elements to include in the new array object.\n      */\n    of(...items: number[]): Int32Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      */\n    from(arrayLike: ArrayLike<number>): Int32Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      * @param mapfn A mapping function to call on every element of the array.\n      * @param thisArg Value of 'this' used to invoke the mapfn.\n      */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Int32Array;\n\n}\ndeclare const Int32Array: Int32ArrayConstructor;\n\n/**\n  * A typed array of 32-bit unsigned integer values. The contents are initialized to 0. If the\n  * requested number of bytes could not be allocated an exception is raised.\n  */\ninterface Uint32Array {\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * The ArrayBuffer instance referenced by the array.\n      */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n      * The length in bytes of the array.\n      */\n    readonly byteLength: number;\n\n    /**\n      * The offset in bytes of the array.\n      */\n    readonly byteOffset: number;\n\n    /**\n      * Returns the this object after copying a section of the array identified by start and end\n      * to the same array starting at position target\n      * @param target If target is negative, it is treated as length+target where length is the\n      * length of the array.\n      * @param start If start is negative, it is treated as length+start. If end is negative, it\n      * is treated as length+end.\n      * @param end If not specified, length of the this object is used as its default value.\n      */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n      * Determines whether all the members of an array satisfy the specified test.\n      * @param callbackfn A function that accepts up to three arguments. The every method calls\n      * the callbackfn function for each element in array1 until the callbackfn returns false,\n      * or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    every(callbackfn: (value: number, index: number, array: Uint32Array) => boolean, thisArg?: any): boolean;\n\n    /**\n        * Returns the this object after filling the section identified by start and end with value\n        * @param value value to fill array section with\n        * @param start index to start filling the array at. If start is negative, it is treated as\n        * length+start where length is the length of the array.\n        * @param end index to stop filling the array at. If end is negative, it is treated as\n        * length+end.\n        */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n      * Returns the elements of an array that meet the condition specified in a callback function.\n      * @param callbackfn A function that accepts up to three arguments. The filter method calls\n      * the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    filter(callbackfn: (value: number, index: number, array: Uint32Array) => any, thisArg?: any): Uint32Array;\n\n    /**\n      * Returns the value of the first element in the array where predicate is true, and undefined\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found, find\n      * immediately returns that element value. Otherwise, find returns undefined.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    find(predicate: (value: number, index: number, obj: Uint32Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n      * Returns the index of the first element in the array where predicate is true, and -1\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found,\n      * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    findIndex(predicate: (value: number, index: number, obj: Uint32Array) => boolean, thisArg?: any): number;\n\n    /**\n      * Performs the specified action for each element in an array.\n      * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    forEach(callbackfn: (value: number, index: number, array: Uint32Array) => void, thisArg?: any): void;\n    /**\n      * Returns the index of the first occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      *  search starts at index 0.\n      */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * Adds all the elements of an array separated by the specified separator string.\n      * @param separator A string used to separate one element of an array from the next in the\n      * resulting String. If omitted, the array elements are separated with a comma.\n      */\n    join(separator?: string): string;\n\n    /**\n      * Returns the index of the last occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      * search starts at index 0.\n      */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * The length of the array.\n      */\n    readonly length: number;\n\n    /**\n      * Calls a defined callback function on each element of an array, and returns an array that\n      * contains the results.\n      * @param callbackfn A function that accepts up to three arguments. The map method calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    map(callbackfn: (value: number, index: number, array: Uint32Array) => number, thisArg?: any): Uint32Array;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint32Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint32Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint32Array) => U, initialValue: U): U;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an\n      * argument instead of an array value.\n      */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint32Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint32Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint32Array) => U, initialValue: U): U;\n\n    /**\n      * Reverses the elements in an Array.\n      */\n    reverse(): Uint32Array;\n\n    /**\n      * Sets a value or an array of values.\n      * @param array A typed or untyped array of values to set.\n      * @param offset The index in the current array at which the values are to be written.\n      */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n      * Returns a section of an array.\n      * @param start The beginning of the specified portion of the array.\n      * @param end The end of the specified portion of the array.\n      */\n    slice(start?: number, end?: number): Uint32Array;\n\n    /**\n      * Determines whether the specified callback function returns true for any element of an array.\n      * @param callbackfn A function that accepts up to three arguments. The some method calls the\n      * callbackfn function for each element in array1 until the callbackfn returns true, or until\n      * the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    some(callbackfn: (value: number, index: number, array: Uint32Array) => boolean, thisArg?: any): boolean;\n\n    /**\n      * Sorts an array.\n      * @param compareFn The name of the function used to determine the order of the elements. If\n      * omitted, the elements are sorted in ascending, ASCII character order.\n      */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n      * Gets a new Uint32Array view of the ArrayBuffer store for this array, referencing the elements\n      * at begin, inclusive, up to end, exclusive.\n      * @param begin The index of the beginning of the array.\n      * @param end The index of the end of the array.\n      */\n    subarray(begin: number, end?: number): Uint32Array;\n\n    /**\n      * Converts a number to a string by using the current locale.\n      */\n    toLocaleString(): string;\n\n    /**\n      * Returns a string representation of an array.\n      */\n    toString(): string;\n\n    [index: number]: number;\n}\n\ninterface Uint32ArrayConstructor {\n    readonly prototype: Uint32Array;\n    new(length: number): Uint32Array;\n    new(arrayOrArrayBuffer: ArrayLike<number> | ArrayBufferLike): Uint32Array;\n    new(buffer: ArrayBufferLike, byteOffset: number, length?: number): Uint32Array;\n\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * Returns a new array from a set of elements.\n      * @param items A set of elements to include in the new array object.\n      */\n    of(...items: number[]): Uint32Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      */\n    from(arrayLike: ArrayLike<number>): Uint32Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      * @param mapfn A mapping function to call on every element of the array.\n      * @param thisArg Value of 'this' used to invoke the mapfn.\n      */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Uint32Array;\n\n}\ndeclare const Uint32Array: Uint32ArrayConstructor;\n\n/**\n  * A typed array of 32-bit float values. The contents are initialized to 0. If the requested number\n  * of bytes could not be allocated an exception is raised.\n  */\ninterface Float32Array {\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * The ArrayBuffer instance referenced by the array.\n      */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n      * The length in bytes of the array.\n      */\n    readonly byteLength: number;\n\n    /**\n      * The offset in bytes of the array.\n      */\n    readonly byteOffset: number;\n\n    /**\n      * Returns the this object after copying a section of the array identified by start and end\n      * to the same array starting at position target\n      * @param target If target is negative, it is treated as length+target where length is the\n      * length of the array.\n      * @param start If start is negative, it is treated as length+start. If end is negative, it\n      * is treated as length+end.\n      * @param end If not specified, length of the this object is used as its default value.\n      */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n      * Determines whether all the members of an array satisfy the specified test.\n      * @param callbackfn A function that accepts up to three arguments. The every method calls\n      * the callbackfn function for each element in array1 until the callbackfn returns false,\n      * or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    every(callbackfn: (value: number, index: number, array: Float32Array) => boolean, thisArg?: any): boolean;\n\n    /**\n        * Returns the this object after filling the section identified by start and end with value\n        * @param value value to fill array section with\n        * @param start index to start filling the array at. If start is negative, it is treated as\n        * length+start where length is the length of the array.\n        * @param end index to stop filling the array at. If end is negative, it is treated as\n        * length+end.\n        */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n      * Returns the elements of an array that meet the condition specified in a callback function.\n      * @param callbackfn A function that accepts up to three arguments. The filter method calls\n      * the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    filter(callbackfn: (value: number, index: number, array: Float32Array) => any, thisArg?: any): Float32Array;\n\n    /**\n      * Returns the value of the first element in the array where predicate is true, and undefined\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found, find\n      * immediately returns that element value. Otherwise, find returns undefined.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    find(predicate: (value: number, index: number, obj: Float32Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n      * Returns the index of the first element in the array where predicate is true, and -1\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found,\n      * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    findIndex(predicate: (value: number, index: number, obj: Float32Array) => boolean, thisArg?: any): number;\n\n    /**\n      * Performs the specified action for each element in an array.\n      * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    forEach(callbackfn: (value: number, index: number, array: Float32Array) => void, thisArg?: any): void;\n\n    /**\n      * Returns the index of the first occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      *  search starts at index 0.\n      */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * Adds all the elements of an array separated by the specified separator string.\n      * @param separator A string used to separate one element of an array from the next in the\n      * resulting String. If omitted, the array elements are separated with a comma.\n      */\n    join(separator?: string): string;\n\n    /**\n      * Returns the index of the last occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      * search starts at index 0.\n      */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * The length of the array.\n      */\n    readonly length: number;\n\n    /**\n      * Calls a defined callback function on each element of an array, and returns an array that\n      * contains the results.\n      * @param callbackfn A function that accepts up to three arguments. The map method calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    map(callbackfn: (value: number, index: number, array: Float32Array) => number, thisArg?: any): Float32Array;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float32Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float32Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Float32Array) => U, initialValue: U): U;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an\n      * argument instead of an array value.\n      */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float32Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float32Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Float32Array) => U, initialValue: U): U;\n\n    /**\n      * Reverses the elements in an Array.\n      */\n    reverse(): Float32Array;\n\n    /**\n      * Sets a value or an array of values.\n      * @param array A typed or untyped array of values to set.\n      * @param offset The index in the current array at which the values are to be written.\n      */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n      * Returns a section of an array.\n      * @param start The beginning of the specified portion of the array.\n      * @param end The end of the specified portion of the array.\n      */\n    slice(start?: number, end?: number): Float32Array;\n\n    /**\n      * Determines whether the specified callback function returns true for any element of an array.\n      * @param callbackfn A function that accepts up to three arguments. The some method calls the\n      * callbackfn function for each element in array1 until the callbackfn returns true, or until\n      * the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    some(callbackfn: (value: number, index: number, array: Float32Array) => boolean, thisArg?: any): boolean;\n\n    /**\n      * Sorts an array.\n      * @param compareFn The name of the function used to determine the order of the elements. If\n      * omitted, the elements are sorted in ascending, ASCII character order.\n      */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n      * Gets a new Float32Array view of the ArrayBuffer store for this array, referencing the elements\n      * at begin, inclusive, up to end, exclusive.\n      * @param begin The index of the beginning of the array.\n      * @param end The index of the end of the array.\n      */\n    subarray(begin: number, end?: number): Float32Array;\n\n    /**\n      * Converts a number to a string by using the current locale.\n      */\n    toLocaleString(): string;\n\n    /**\n      * Returns a string representation of an array.\n      */\n    toString(): string;\n\n    [index: number]: number;\n}\n\ninterface Float32ArrayConstructor {\n    readonly prototype: Float32Array;\n    new(length: number): Float32Array;\n    new(arrayOrArrayBuffer: ArrayLike<number> | ArrayBufferLike): Float32Array;\n    new(buffer: ArrayBufferLike, byteOffset: number, length?: number): Float32Array;\n\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * Returns a new array from a set of elements.\n      * @param items A set of elements to include in the new array object.\n      */\n    of(...items: number[]): Float32Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      */\n    from(arrayLike: ArrayLike<number>): Float32Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      * @param mapfn A mapping function to call on every element of the array.\n      * @param thisArg Value of 'this' used to invoke the mapfn.\n      */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Float32Array;\n\n\n}\ndeclare const Float32Array: Float32ArrayConstructor;\n\n/**\n  * A typed array of 64-bit float values. The contents are initialized to 0. If the requested\n  * number of bytes could not be allocated an exception is raised.\n  */\ninterface Float64Array {\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * The ArrayBuffer instance referenced by the array.\n      */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n      * The length in bytes of the array.\n      */\n    readonly byteLength: number;\n\n    /**\n      * The offset in bytes of the array.\n      */\n    readonly byteOffset: number;\n\n    /**\n      * Returns the this object after copying a section of the array identified by start and end\n      * to the same array starting at position target\n      * @param target If target is negative, it is treated as length+target where length is the\n      * length of the array.\n      * @param start If start is negative, it is treated as length+start. If end is negative, it\n      * is treated as length+end.\n      * @param end If not specified, length of the this object is used as its default value.\n      */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n      * Determines whether all the members of an array satisfy the specified test.\n      * @param callbackfn A function that accepts up to three arguments. The every method calls\n      * the callbackfn function for each element in array1 until the callbackfn returns false,\n      * or until the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    every(callbackfn: (value: number, index: number, array: Float64Array) => boolean, thisArg?: any): boolean;\n\n    /**\n        * Returns the this object after filling the section identified by start and end with value\n        * @param value value to fill array section with\n        * @param start index to start filling the array at. If start is negative, it is treated as\n        * length+start where length is the length of the array.\n        * @param end index to stop filling the array at. If end is negative, it is treated as\n        * length+end.\n        */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n      * Returns the elements of an array that meet the condition specified in a callback function.\n      * @param callbackfn A function that accepts up to three arguments. The filter method calls\n      * the callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    filter(callbackfn: (value: number, index: number, array: Float64Array) => any, thisArg?: any): Float64Array;\n\n    /**\n      * Returns the value of the first element in the array where predicate is true, and undefined\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found, find\n      * immediately returns that element value. Otherwise, find returns undefined.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    find(predicate: (value: number, index: number, obj: Float64Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n      * Returns the index of the first element in the array where predicate is true, and -1\n      * otherwise.\n      * @param predicate find calls predicate once for each element of the array, in ascending\n      * order, until it finds one where predicate returns true. If such an element is found,\n      * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n      * @param thisArg If provided, it will be used as the this value for each invocation of\n      * predicate. If it is not provided, undefined is used instead.\n      */\n    findIndex(predicate: (value: number, index: number, obj: Float64Array) => boolean, thisArg?: any): number;\n\n    /**\n      * Performs the specified action for each element in an array.\n      * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    forEach(callbackfn: (value: number, index: number, array: Float64Array) => void, thisArg?: any): void;\n\n    /**\n      * Returns the index of the first occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      *  search starts at index 0.\n      */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * Adds all the elements of an array separated by the specified separator string.\n      * @param separator A string used to separate one element of an array from the next in the\n      * resulting String. If omitted, the array elements are separated with a comma.\n      */\n    join(separator?: string): string;\n\n    /**\n      * Returns the index of the last occurrence of a value in an array.\n      * @param searchElement The value to locate in the array.\n      * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n      * search starts at index 0.\n      */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n      * The length of the array.\n      */\n    readonly length: number;\n\n    /**\n      * Calls a defined callback function on each element of an array, and returns an array that\n      * contains the results.\n      * @param callbackfn A function that accepts up to three arguments. The map method calls the\n      * callbackfn function one time for each element in the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    map(callbackfn: (value: number, index: number, array: Float64Array) => number, thisArg?: any): Float64Array;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float64Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float64Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array. The return value of\n      * the callback function is the accumulated result, and is provided as an argument in the next\n      * call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n      * callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Float64Array) => U, initialValue: U): U;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an\n      * argument instead of an array value.\n      */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float64Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float64Array) => number, initialValue: number): number;\n\n    /**\n      * Calls the specified callback function for all the elements in an array, in descending order.\n      * The return value of the callback function is the accumulated result, and is provided as an\n      * argument in the next call to the callback function.\n      * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n      * the callbackfn function one time for each element in the array.\n      * @param initialValue If initialValue is specified, it is used as the initial value to start\n      * the accumulation. The first call to the callbackfn function provides this value as an argument\n      * instead of an array value.\n      */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Float64Array) => U, initialValue: U): U;\n\n    /**\n      * Reverses the elements in an Array.\n      */\n    reverse(): Float64Array;\n\n    /**\n      * Sets a value or an array of values.\n      * @param array A typed or untyped array of values to set.\n      * @param offset The index in the current array at which the values are to be written.\n      */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n      * Returns a section of an array.\n      * @param start The beginning of the specified portion of the array.\n      * @param end The end of the specified portion of the array.\n      */\n    slice(start?: number, end?: number): Float64Array;\n\n    /**\n      * Determines whether the specified callback function returns true for any element of an array.\n      * @param callbackfn A function that accepts up to three arguments. The some method calls the\n      * callbackfn function for each element in array1 until the callbackfn returns true, or until\n      * the end of the array.\n      * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n      * If thisArg is omitted, undefined is used as the this value.\n      */\n    some(callbackfn: (value: number, index: number, array: Float64Array) => boolean, thisArg?: any): boolean;\n\n    /**\n      * Sorts an array.\n      * @param compareFn The name of the function used to determine the order of the elements. If\n      * omitted, the elements are sorted in ascending, ASCII character order.\n      */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n      * Gets a new Float64Array view of the ArrayBuffer store for this array, referencing the elements\n      * at begin, inclusive, up to end, exclusive.\n      * @param begin The index of the beginning of the array.\n      * @param end The index of the end of the array.\n      */\n    subarray(begin: number, end?: number): Float64Array;\n\n    /**\n      * Converts a number to a string by using the current locale.\n      */\n    toLocaleString(): string;\n\n    /**\n      * Returns a string representation of an array.\n      */\n    toString(): string;\n\n    [index: number]: number;\n}\n\ninterface Float64ArrayConstructor {\n    readonly prototype: Float64Array;\n    new(length: number): Float64Array;\n    new(arrayOrArrayBuffer: ArrayLike<number> | ArrayBufferLike): Float64Array;\n    new(buffer: ArrayBufferLike, byteOffset: number, length?: number): Float64Array;\n\n    /**\n      * The size in bytes of each element in the array.\n      */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n      * Returns a new array from a set of elements.\n      * @param items A set of elements to include in the new array object.\n      */\n    of(...items: number[]): Float64Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      */\n    from(arrayLike: ArrayLike<number>): Float64Array;\n\n    /**\n      * Creates an array from an array-like or iterable object.\n      * @param arrayLike An array-like or iterable object to convert to an array.\n      * @param mapfn A mapping function to call on every element of the array.\n      * @param thisArg Value of 'this' used to invoke the mapfn.\n      */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Float64Array;\n\n}\ndeclare const Float64Array: Float64ArrayConstructor;\n\n/////////////////////////////\n/// ECMAScript Internationalization API\n/////////////////////////////\n\ndeclare namespace Intl {\n    interface CollatorOptions {\n        usage?: string;\n        localeMatcher?: string;\n        numeric?: boolean;\n        caseFirst?: string;\n        sensitivity?: string;\n        ignorePunctuation?: boolean;\n    }\n\n    interface ResolvedCollatorOptions {\n        locale: string;\n        usage: string;\n        sensitivity: string;\n        ignorePunctuation: boolean;\n        collation: string;\n        caseFirst: string;\n        numeric: boolean;\n    }\n\n    interface Collator {\n        compare(x: string, y: string): number;\n        resolvedOptions(): ResolvedCollatorOptions;\n    }\n    var Collator: {\n        new(locales?: string | string[], options?: CollatorOptions): Collator;\n        (locales?: string | string[], options?: CollatorOptions): Collator;\n        supportedLocalesOf(locales: string | string[], options?: CollatorOptions): string[];\n    };\n\n    interface NumberFormatOptions {\n        localeMatcher?: string;\n        style?: string;\n        currency?: string;\n        currencyDisplay?: string;\n        useGrouping?: boolean;\n        minimumIntegerDigits?: number;\n        minimumFractionDigits?: number;\n        maximumFractionDigits?: number;\n        minimumSignificantDigits?: number;\n        maximumSignificantDigits?: number;\n    }\n\n    interface ResolvedNumberFormatOptions {\n        locale: string;\n        numberingSystem: string;\n        style: string;\n        currency?: string;\n        currencyDisplay?: string;\n        minimumIntegerDigits: number;\n        minimumFractionDigits: number;\n        maximumFractionDigits: number;\n        minimumSignificantDigits?: number;\n        maximumSignificantDigits?: number;\n        useGrouping: boolean;\n    }\n\n    interface NumberFormat {\n        format(value: number): string;\n        resolvedOptions(): ResolvedNumberFormatOptions;\n    }\n    var NumberFormat: {\n        new(locales?: string | string[], options?: NumberFormatOptions): NumberFormat;\n        (locales?: string | string[], options?: NumberFormatOptions): NumberFormat;\n        supportedLocalesOf(locales: string | string[], options?: NumberFormatOptions): string[];\n    };\n\n    interface DateTimeFormatOptions {\n        localeMatcher?: string;\n        weekday?: string;\n        era?: string;\n        year?: string;\n        month?: string;\n        day?: string;\n        hour?: string;\n        minute?: string;\n        second?: string;\n        timeZoneName?: string;\n        formatMatcher?: string;\n        hour12?: boolean;\n        timeZone?: string;\n    }\n\n    interface ResolvedDateTimeFormatOptions {\n        locale: string;\n        calendar: string;\n        numberingSystem: string;\n        timeZone: string;\n        hour12?: boolean;\n        weekday?: string;\n        era?: string;\n        year?: string;\n        month?: string;\n        day?: string;\n        hour?: string;\n        minute?: string;\n        second?: string;\n        timeZoneName?: string;\n    }\n\n    interface DateTimeFormat {\n        format(date?: Date | number): string;\n        resolvedOptions(): ResolvedDateTimeFormatOptions;\n    }\n    var DateTimeFormat: {\n        new(locales?: string | string[], options?: DateTimeFormatOptions): DateTimeFormat;\n        (locales?: string | string[], options?: DateTimeFormatOptions): DateTimeFormat;\n        supportedLocalesOf(locales: string | string[], options?: DateTimeFormatOptions): string[];\n    };\n}\n\ninterface String {\n    /**\n      * Determines whether two strings are equivalent in the current or specified locale.\n      * @param that String to compare to target string\n      * @param locales A locale string or array of locale strings that contain one or more language or locale tags. If you include more than one locale string, list them in descending order of priority so that the first entry is the preferred locale. If you omit this parameter, the default locale of the JavaScript runtime is used. This parameter must conform to BCP 47 standards; see the Intl.Collator object for details.\n      * @param options An object that contains one or more properties that specify comparison options. see the Intl.Collator object for details.\n      */\n    localeCompare(that: string, locales?: string | string[], options?: Intl.CollatorOptions): number;\n}\n\ninterface Number {\n    /**\n      * Converts a number to a string by using the current or specified locale.\n      * @param locales A locale string or array of locale strings that contain one or more language or locale tags. If you include more than one locale string, list them in descending order of priority so that the first entry is the preferred locale. If you omit this parameter, the default locale of the JavaScript runtime is used.\n      * @param options An object that contains one or more properties that specify comparison options.\n      */\n    toLocaleString(locales?: string | string[], options?: Intl.NumberFormatOptions): string;\n}\n\ninterface Date {\n    /**\n      * Converts a date and time to a string by using the current or specified locale.\n      * @param locales A locale string or array of locale strings that contain one or more language or locale tags. If you include more than one locale string, list them in descending order of priority so that the first entry is the preferred locale. If you omit this parameter, the default locale of the JavaScript runtime is used.\n      * @param options An object that contains one or more properties that specify comparison options.\n      */\n    toLocaleString(locales?: string | string[], options?: Intl.DateTimeFormatOptions): string;\n    /**\n      * Converts a date to a string by using the current or specified locale.\n      * @param locales A locale string or array of locale strings that contain one or more language or locale tags. If you include more than one locale string, list them in descending order of priority so that the first entry is the preferred locale. If you omit this parameter, the default locale of the JavaScript runtime is used.\n      * @param options An object that contains one or more properties that specify comparison options.\n      */\n    toLocaleDateString(locales?: string | string[], options?: Intl.DateTimeFormatOptions): string;\n\n    /**\n      * Converts a time to a string by using the current or specified locale.\n      * @param locales A locale string or array of locale strings that contain one or more language or locale tags. If you include more than one locale string, list them in descending order of priority so that the first entry is the preferred locale. If you omit this parameter, the default locale of the JavaScript runtime is used.\n      * @param options An object that contains one or more properties that specify comparison options.\n      */\n    toLocaleTimeString(locales?: string | string[], options?: Intl.DateTimeFormatOptions): string;\n}\n",et='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n'+$n+Jn+zn+qn+Wn+Hn+Un+jn+Bn+Kn,nt='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n'+et+Zn+Xn+Qn+Yn,tt='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved. \nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0  \n \nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, \nMERCHANTABLITY OR NON-INFRINGEMENT. \n \nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n'+$n+Zn+Qn+Yn,rt={NAME:"defaultLib:lib.d.ts",CONTENTS:tt},at={NAME:"defaultLib:lib.es6.d.ts",CONTENTS:nt},it=function(){function e(e,n){this._extraLibs=Object.create(null),this._languageService=Vn["c"](this),this._ctx=e,this._compilerOptions=n.compilerOptions,this._extraLibs=n.extraLibs}return e.prototype.getCompilationSettings=function(){return this._compilerOptions},e.prototype.getScriptFileNames=function(){var e=this._ctx.getMirrorModels().map(function(e){return e.uri.toString()});return e.concat(Object.keys(this._extraLibs))},e.prototype._getModel=function(e){for(var n=this._ctx.getMirrorModels(),t=0;t<n.length;t++)if(n[t].uri.toString()===e)return n[t];return null},e.prototype.getScriptVersion=function(e){var n=this._getModel(e);return n?n.version.toString():this.isDefaultLibFileName(e)?"1":e in this._extraLibs?String(this._extraLibs[e].version):void 0},e.prototype.getScriptSnapshot=function(e){var n,t=this._getModel(e);if(t)n=t.getValue();else if(e in this._extraLibs)n=this._extraLibs[e].content;else if(e===rt.NAME)n=rt.CONTENTS;else{if(e!==at.NAME)return;n=at.CONTENTS}return{getText:function(e,t){return n.substring(e,t)},getLength:function(){return n.length},getChangeRange:function(){}}},e.prototype.getScriptKind=function(e){var n=e.substr(e.lastIndexOf(".")+1);switch(n){case"ts":return Vn["a"].TS;case"tsx":return Vn["a"].TSX;case"js":return Vn["a"].JS;case"jsx":return Vn["a"].JSX;default:return this.getCompilationSettings().allowJs?Vn["a"].JS:Vn["a"].TS}},e.prototype.getCurrentDirectory=function(){return""},e.prototype.getDefaultLibFileName=function(e){return e.target<=Vn["b"].ES5?rt.NAME:at.NAME},e.prototype.isDefaultLibFileName=function(e){return e===this.getDefaultLibFileName(this._compilerOptions)},e.clearFiles=function(e){e.forEach(function(e){e.file=void 0;var n=e.relatedInformation;n&&n.forEach(function(e){return e.file=void 0})})},e.prototype.getSyntacticDiagnostics=function(n){var t=this._languageService.getSyntacticDiagnostics(n);return e.clearFiles(t),Promise.resolve(t)},e.prototype.getSemanticDiagnostics=function(n){var t=this._languageService.getSemanticDiagnostics(n);return e.clearFiles(t),Promise.resolve(t)},e.prototype.getCompilerOptionsDiagnostics=function(n){var t=this._languageService.getCompilerOptionsDiagnostics();return e.clearFiles(t),Promise.resolve(t)},e.prototype.getCompletionsAtPosition=function(e,n){return Promise.resolve(this._languageService.getCompletionsAtPosition(e,n,void 0))},e.prototype.getCompletionEntryDetails=function(e,n,t){return Promise.resolve(this._languageService.getCompletionEntryDetails(e,n,t,void 0,void 0,void 0))},e.prototype.getSignatureHelpItems=function(e,n){return Promise.resolve(this._languageService.getSignatureHelpItems(e,n,void 0))},e.prototype.getQuickInfoAtPosition=function(e,n){return Promise.resolve(this._languageService.getQuickInfoAtPosition(e,n))},e.prototype.getOccurrencesAtPosition=function(e,n){return Promise.resolve(this._languageService.getOccurrencesAtPosition(e,n))},e.prototype.getDefinitionAtPosition=function(e,n){return Promise.resolve(this._languageService.getDefinitionAtPosition(e,n))},e.prototype.getReferencesAtPosition=function(e,n){return Promise.resolve(this._languageService.getReferencesAtPosition(e,n))},e.prototype.getNavigationBarItems=function(e){return Promise.resolve(this._languageService.getNavigationBarItems(e))},e.prototype.getFormattingEditsForDocument=function(e,n){return Promise.resolve(this._languageService.getFormattingEditsForDocument(e,n))},e.prototype.getFormattingEditsForRange=function(e,n,t,r){return Promise.resolve(this._languageService.getFormattingEditsForRange(e,n,t,r))},e.prototype.getFormattingEditsAfterKeystroke=function(e,n,t,r){return Promise.resolve(this._languageService.getFormattingEditsAfterKeystroke(e,n,t,r))},e.prototype.getEmitOutput=function(e){return Promise.resolve(this._languageService.getEmitOutput(e))},e.prototype.updateExtraLibs=function(e){this._extraLibs=e},e}();self.onmessage=function(){Gn(function(e,n){return new it(e,n)})}},dd40:function(e,n){e.exports=function(e){if(!e.webpackPolyfill){var n=Object.create(e);n.children||(n.children=[]),Object.defineProperty(n,"loaded",{enumerable:!0,get:function(){return n.l}}),Object.defineProperty(n,"id",{enumerable:!0,get:function(){return n.i}}),Object.defineProperty(n,"exports",{enumerable:!0}),n.webpackPolyfill=1}return n}},f28c:function(e,n){var t,r,a=e.exports={};function i(){throw new Error("setTimeout has not been defined")}function o(){throw new Error("clearTimeout has not been defined")}function s(e){if(t===setTimeout)return setTimeout(e,0);if((t===i||!t)&&setTimeout)return t=setTimeout,setTimeout(e,0);try{return t(e,0)}catch(n){try{return t.call(null,e,0)}catch(n){return t.call(this,e,0)}}}function c(e){if(r===clearTimeout)return clearTimeout(e);if((r===o||!r)&&clearTimeout)return r=clearTimeout,clearTimeout(e);try{return r(e)}catch(n){try{return r.call(null,e)}catch(n){return r.call(this,e)}}}(function(){try{t="function"===typeof setTimeout?setTimeout:i}catch(e){t=i}try{r="function"===typeof clearTimeout?clearTimeout:o}catch(e){r=o}})();var l,u=[],d=!1,p=-1;function m(){d&&l&&(d=!1,l.length?u=l.concat(u):p=-1,u.length&&f())}function f(){if(!d){var e=s(m);d=!0;var n=u.length;while(n){l=u,u=[];while(++p<n)l&&l[p].run();p=-1,n=u.length}l=null,d=!1,c(e)}}function _(e,n){this.fun=e,this.array=n}function g(){}a.nextTick=function(e){var n=new Array(arguments.length-1);if(arguments.length>1)for(var t=1;t<arguments.length;t++)n[t-1]=arguments[t];u.push(new _(e,n)),1!==u.length||d||s(f)},_.prototype.run=function(){this.fun.apply(null,this.array)},a.title="browser",a.browser=!0,a.env={},a.argv=[],a.version="",a.versions={},a.on=g,a.addListener=g,a.once=g,a.off=g,a.removeListener=g,a.removeAllListeners=g,a.emit=g,a.prependListener=g,a.prependOnceListener=g,a.listeners=function(e){return[]},a.binding=function(e){throw new Error("process.binding is not supported")},a.cwd=function(){return"/"},a.chdir=function(e){throw new Error("process.chdir is not supported")},a.umask=function(){return 0}}});